// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// client.h

// Main Header file for Xplatform code in the client
// In general keep this file as small as possible.

// IMPORTANT
// Inclusion of this header should not make you link with another library
// This header needs to be safely included without changing dependencies

#ifndef __client_h
#define __client_h

#include "clientdefs.h"
#include "clientenum.h"
#include "clientdebug.h"

// Minimal platform differences are hidden inside these files

#if defined( WIN32 )
#include "clientwin32.h"
#elif defined( ZK_LINUX )
#include "clientlinux.h"
#else
#error "Please provide platform specific header file in client.h"
// TODO - Mac and other platform headers
#endif


#define ZKSTD	std



#ifndef DONT_INCLUDE_STL
#include "clientstl.h"
#endif

#include <zks.h>

// WARNING - please do not put header files here as
// a practice, this is temporary
#include <frNetworkDefaults.h>

#ifdef _WIN32
#pragma warning( disable : 4275 )
#endif

////////////////////////////////////////////////////////////////////////////////
// STRUCTURES, Types
////////////////////////////////////////////////////////////////////////////////



typedef struct SecuritySettings{
	int nReserved;	// size of this structure
	int	nHops;		// MIN_HOPS to MAX_HOPS hops when browsing
}SecuritySettings;


// Structure used with CS_CONNECTION_INFO
// Return value from app contains
// bits 0-15	loword, FirewallAction enumeration equivalent
// bits 16-31	hiword, Local port (host byte order) to switch the connection to if the enumeration was FirewallForwardSockaddr
typedef struct connection_info
{
	ESOCKETAPI		api;			// socket api
	char			app[MAX_PATH];	// application name, may be freedom itself
	unsigned short	Localport;		// local port in host byte order, may be 0
	unsigned short  DstPort;		// destination port in host byte order, may be 0
	bool			bIsFreedomProcess; // true if within freedom process, false otherwise
	unsigned int	DstIp;			// destination ip in network byte order, may be 0
	int				s;				// socket number
	int				socket_type;	// socket type, SOCK_DGRAM. SOCK_STREAM, 0 is an error
	unsigned long	tid;			// thread id

	bool
	OutgoingTraffic() const
	{
		if(epresendto == api|| epreconnect == api)
		{
			return true;
		}
		return false;
	}

	bool 
	IncomingTraffic() const
	{
		switch(api)
		{
		case eprelisten:	
			return true;
			break;

		case eprebind:
			if(SOCK_DGRAM == socket_type)
			{
				return true;;
			}
			break;

		default:
			return false;
		};

		return false;
	}

	bool
	IsUIOnlyApi() const
	{
		return ( (eprelisten == api) || (eprebind == api) );
	}
}connection_info;

// Structure used with CS_POSTCONNECTION_INFO
typedef struct postconnection_info
{
	unsigned int	DstIp;			// destination ip in network byte order
	unsigned short  DstPort;		// destination port in network byte order
	bool			Ok;				// True if connect succeeded, false otherwise
	int				s;				// socket number
	unsigned short  LocalPort;		// Only valid when connection succeeded, network byte order
	int				socket_type;	// socket type, SOCK_DGRAM. SOCK_STREAM, 0 is an error
	unsigned long	tid;			// thread id
	char			app[MAX_PATH];	// application name
}postconnection_info;



// Structure used with CS_POSTSENDTO_INFO
typedef struct postsendto_info{
	int				s;				// socket number
	unsigned int	DstIp;			// destination in network byte order
	unsigned short	DstPort;		// destination port in network byte order
	char			app[MAX_PATH];	// application name
	unsigned short	LocalPort;		// local port in network byte order, may be zero
	int				socket_type;	// socket type, SOCK_DGRAM. SOCK_STREAM, 0 is an error
	bool			Ok;				// true if sendto sent > 0
	unsigned long	tid;			// thread id
}postsendto_info;



typedef struct remotehost
{
	unsigned int	ip;						// in network byte order, keep in same position for CompareHosts to work
	unsigned short	port;					// in network byte order, keep in same position for CompareHosts to work
	char			application[MAX_PATH];	// not strictly required but usefull for debugging which app is in passive connect state
	unsigned short	LocalPort;				// May be 0. in network byte order.
	int				s;						// socket number
	unsigned long	tid;					// thread id
}remotehost;

typedef struct endpoint
{
	eTcpState		state;
	unsigned int	localip;		// in host byte order
	unsigned short	localport;		// in host byte order
	unsigned int	remoteip;		// in host byte order
	unsigned short	remoteport;		// in host byte order
}endpoint;

typedef struct endpointinfo
{
	struct endpointinfo*	prev;
	struct endpointinfo*	next;
	struct endpoint			ep;
}endpointinfo;

struct endpoint_equal_localport
{
	unsigned short x_;

	endpoint_equal_localport(unsigned short x) : x_(x) {}

	bool operator()(const endpoint& in_rh) const
	{ 
		return in_rh.localport == x_; 
	}
};

struct rh_equal_socket
{
	int x_;

	rh_equal_socket(int x) : x_(x) {}

	bool operator()(const remotehost& in_rh) const
	{ 
		return in_rh.s == x_; 
	}
};


struct rh_equal_LocalPort
{
	unsigned short x_;

	rh_equal_LocalPort(unsigned short  x) : x_(x) {}

	bool operator()(const remotehost& in_rh) const
	{ 
		return in_rh.LocalPort == x_; 
	}
};

// Compare only the LocalPort in the remotehost structure
struct CompareLocalPorts
{
  bool operator()(const remotehost& lhs, const remotehost& rhs) const
  {
    return (lhs.LocalPort < rhs.LocalPort);
  }
};

// Compare only the ip and port in the remotehost structure
struct CompareHosts
{
  bool operator()(const remotehost& lhs, const remotehost& rhs) const
  {
    return memcmp(&lhs, &rhs, 6) < 0;
  }
};

// Compare only the socket number in the remotehost structure
struct CompareSockets
{
	bool operator()(const remotehost& lhs, const remotehost& rhs) const
	{
		return lhs.s < rhs.s;
	}
};

// Compare the socket number and the thread id in the remotehost structure
struct CompareSocketsAndThreadId
{
	bool operator()(const remotehost& lhs, const remotehost& rhs) const
	{
		return memcmp(&lhs.s, &rhs.s, 8) < 0; // ! Note, 4 bytes socket + 4 bytes tid == 8 bytes bug # 7141
	}
};

typedef struct FirewallEvent_UsersChoice
{	
	bool		bRemember;			// If true don't ask the user again	
	eSelection	Sel;				// What the user has specified, block or allow
} FirewallEvent_UsersChoice;


typedef struct FirewallAppPersist
{
	char						app[MAX_PATH];	// Entire application path
	FirewallEvent_UsersChoice	incoming;		// block, allow, ask for incoming traffic 
	FirewallEvent_UsersChoice	outgoing;		// block, allow, ask for outgoing traffic
	FirewallEvent_UsersChoice	special_warn;	// block, allow, ask for special warning	

	eSelection 
	GetOutgoingSelection() const
	{
		return outgoing.Sel;
	}

	eSelection 
	GetIncomingSelection() const
	{
		return incoming.Sel;
	}

	eSelection 
	GetSpecialWarnSelection() const
	{
		return special_warn.Sel;
	}

}FirewallAppPersist;

struct charPtrLess
{
   bool operator()(const char* in_x, const char* in_y) const
   {
	   return strncmp(in_x, in_y, MAX_PATH) < 0;
   }
};


#define DEFAULT_REPLYBLOCK_LATENCY    60
#define DEFAULT_REPLYBLOCK_CRYPTO_ALG FRCRYPT_SYMALG_BLOWFISH 

#ifdef _UNICODE 
#define tstring ZKSTD::wstring
#else
#define tstring ZKSTD::string
#endif

#ifndef PURE
	#define PURE  = 0
#endif

#define DNS_PORT 53

#define LOCAL_POP3_PORT				WINDOWS_TCP_AND_UDP_PORT + 100 +  1
#define LOCAL_SMTP_PORT				WINDOWS_TCP_AND_UDP_PORT + 100 +  2
#define LOCAL_HTTP_PORT				WINDOWS_TCP_AND_UDP_PORT + 100 +  3
#define LOCAL_NNTP_PORT				WINDOWS_TCP_AND_UDP_PORT + 100 +  4
#define LOCAL_IRC_PORT				WINDOWS_TCP_AND_UDP_PORT + 100 +  5
#define LOCAL_TELNET_PORT			WINDOWS_TCP_AND_UDP_PORT + 100 +  6
#define LOCAL_SSL_PORT				WINDOWS_TCP_AND_UDP_PORT + 100 +  7
#define LOCAL_IPC_HTTP_SERVER_PORT	        WINDOWS_TCP_AND_UDP_PORT + 100 +  8

////////////////////////////////////////////////////////////////////////////////
// MACROS
////////////////////////////////////////////////////////////////////////////////


// Safe cleanup macros - NULL is a valid argument
// NOTE: it is often better to use Smart Pointers and smart allocators

// Free memory allocated with malloc
#define SAFEFREE(x)				if ((x) != NULL)					{ free((x)) ; (x) = NULL ;}

// Delete memory allocated with new (C++ allows delete on NULL)
#define SAFEDELETE(x)												{ delete((x)) ; (x) = NULL ;}

// Delete memory allocated with new[] (C++ allows delete on NULL)
#define SAFEDELETEARRAY(x)											{ delete[] (x); (x) = NULL; }

// Release a reference to a COM object
#define SAFERELEASEPTR(x)		if ( (x) != NULL )					{ (x)->Release() ; (x) = NULL ; }

// Close a socket if it isn't INVALID_SOCKET

#ifdef WIN32
#define SAFECLOSESOCKET(x)  if( (x) != INVALID_SOCKET) { closesocket(x); (x) = INVALID_SOCKET; }
#else
#define SAFECLOSESOCKET(x)  if( (x) != INVALID_SOCKET) { close(x); (x) = INVALID_SOCKET; }
#endif

#ifdef WIN32
#define XP_LAST_SOCKET_ERR (WSAGetLastError())
#else
#include <errno.h>
#define XP_LAST_SOCKET_ERR (errno)
#endif

// Cross-platform case-insensitive string comparison

#ifdef WIN32
#define XP_STRNICMP(x, y, z) (strnicmp(x, y, z))
#else
#define XP_STRNICMP(x, y, z) (strncasecmp(x, y, z))
#endif

#endif // __client_h
