// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// ClientChunkIds.h

// This header file defines the Chunk Ids used when
// presisting the client.  

//A centralized list provides unique chunk ids across the
//project and by the use of symbolic names, the easy ability to
//grep for code reading or writing a particular chunk

//For more information on Chunk Ids and other elements of the
//zks chunk format refer to zkchunk.h

//WARNING - don't change ID values of existing chunks - you'll break
//existing persistence!

//PLEASE keep the heirarchy information upto date as you add new chunks

/*
Freedom Persistence Format

This section describes the actual structure of the file that we save.
In includes data stored by various object in zkcentral - include ZKDocument
and IClientPersist.  To try to help keep this information current I
keep its definition close to the code that actually defines the chunk IDS.


Freedom.nym Top Level
  ZKDOCUMENT_HEADER_CHUNK_ID        -Strings describing the file and system, clear text, 
									 contents must reveal any private info
  SALT_CHUNK_ID                     -Salt is required to decrypt buffers
  PASSPHRASETEST_CHUNK_ID           -A small known value encrypted with passphrase to help 
									 errorhandling when loading file.
  DATA_CHUNK_ID					    -Encrypted data stored as an array in this chunk



Once DATA_CHUNK_ID contents are decrypted we have the following structure
  CLIENT_CENTRAL_CHUNK_ID		-Contents of the Client Central, which is basically 
								 everthing that doesn't have bearing to individual nyms
  NYM_CENTRAL_CHUNK_ID		    -Persistence of the nyms
  NETWORK_PERF_CHUNK_ID         -Persistent statistics about network usage
  FIREWALL_CHUNK_ID             -Firewall settings



CLIENT_CENTRAL_CHUNK_ID contents
  FRREMOTENYM_CHUNK_ID		    -frRemoteNym Chunk contains the Tokens
  PREF_CHUNK_ID                 -Global preference flags
								-and num minutes until re-ask passwords
  WALLET_DISABLED_DOMAINS_ID	-Disabled domains for the wallet

NYM_CENTRAL_CHUNK_ID contains all the nyms one after another
  NYM_CHUNK_ID
  NYM_CHUNK_ID
  NYM_CHUNK_ID
  ...


NYM_CHUNK_ID contains
  FRREMOTENYM_CHUNK_ID
  COOKIEJAR_CHUNK_ID
  WALLET_CHUNK_ID
  MAIL_CHUNK_ID
  ICON_CHUNK_ID
  NYM_ROUTE_CHUNK_ID


WALLET_CHUNK_ID contains
	WALLET_LOGIN_INFO_CHUNK_ID
	WALLET_PERSONALINFO_CHUNK_ID
	

COOKIEJAR_CHUNK_ID contains
	COOKIE_CHUNK_ID
	COOKIE_CHUNK_ID
	COOKIE_CHUNK_ID
	....

NETWORK_PERF_CHUNK_ID
	PERF_AIPS_CHUNK_ID		
	PERF_ROUTE_CREATE_CHUNK_ID
	PERF_BAD_ROUTES_CHUNK_ID

*/

#ifndef _CLIENTCHUNKIDS_H
#define _CLIENTCHUNKIDS_H

//Chunk IDs are stored as 3 bytes, hence maximum value is 2^24
//The minimum legal value is CHUNK_BASE_ID

#define CHUNK_MAX_ID  0xFFFFFF

//The Base Id is used purely to help detect corrupt buffers,
//There are still over 12 million possible chunk ids.

#define CHUNK_BASE_ID 0x3F2E1D

// Top level chunks in a document
#define ZKDOCUMENT_HEADER_CHUNK_ID		CHUNK_BASE_ID + 1 
#define SALT_CHUNK_ID					CHUNK_BASE_ID + 2
#define PASSPHRASETEST_CHUNK_ID			CHUNK_BASE_ID + 3
#define DATA_CHUNK_ID					CHUNK_BASE_ID + 4


// DATA CHUNK Contents
#define CLIENT_CENTRAL_CHUNK_ID			CHUNK_BASE_ID + 101
#define NYM_CENTRAL_CHUNK_ID			CHUNK_BASE_ID + 102
#define NETWORK_PERF_CHUNK_ID			CHUNK_BASE_ID + 103
#define FIREWALL_CHUNK_ID               CHUNK_BASE_ID + 104

// NYM CENTRAL Chunk Conents
// Nym Central Chunk is made up of a series of Nym Chunks
// Note: It is an accident that this is the same as CLIENT_CENTRAL_CHUNK_ID but
// we can't change this easily
#define NYM_CHUNK_ID					CHUNK_BASE_ID + 101

// NYM Chunk Contents
#define FRREMOTENYM_CHUNK_ID			CHUNK_BASE_ID + 200
#define COOKIEJAR_CHUNK_ID				CHUNK_BASE_ID + 201
#define WALLET_CHUNK_ID					CHUNK_BASE_ID + 202
#define MAIL_CHUNK_ID					CHUNK_BASE_ID + 203
#define ICON_CHUNK_ID					CHUNK_BASE_ID + 206
#define NYM_ROUTE_CHUNK_ID				CHUNK_BASE_ID + 207
#define NYM_WEB_CACHE_MGR_CHUNK_ID		CHUNK_BASE_ID + 208
// Client Central chunk contents
#define PREF_CHUNK_ID					CHUNK_BASE_ID + 300
#define WALLET_DISABLED_DOMAINS_ID		CHUNK_BASE_ID + 301
#define AD_BLOCKER_CHUNK_ID				CHUNK_BASE_ID + 302

#define WALLET_LOGININFO_CHUNK_ID		CHUNK_BASE_ID + 400
#define WALLET_PERSONALINFO_CHUNK_ID	CHUNK_BASE_ID + 401

#define COOKIE_CHUNK_ID					CHUNK_BASE_ID + 500

#define PERF_ROUTE_CREATE_CHUNK_ID      CHUNK_BASE_ID + 600
#define PERF_AIPS_CHUNK_ID				CHUNK_BASE_ID + 601
#define PERF_BAD_ROUTES_CHUNK_ID		CHUNK_BASE_ID + 602

#define FIREWALL_PREF_CHUNK_ID          CHUNK_BASE_ID + 700
// Beta 1.1, 1.2 of version 2.0 used this chunk id with a
// different format, and we didn't bother being backward compatible
#define OLD_FIREWALL_APP_RULES_CHUNK_ID CHUNK_BASE_ID + 701
#define FIREWALL_PORT_RULES_CHUNK_ID    CHUNK_BASE_ID + 702
#define FIREWALL_SPECIAL_WARN_CHUNK_ID	CHUNK_BASE_ID + 703
#define FIREWALL_APP_RULES_CHUNK_ID     CHUNK_BASE_ID + 704

#define TEST1_CHUNK_ID					CHUNK_BASE_ID + 1000
#define TEST2_CHUNK_ID					CHUNK_BASE_ID + 1001
#define TEST3_CHUNK_ID					CHUNK_BASE_ID + 1003
#define TEST4_CHUNK_ID					CHUNK_BASE_ID + 1004
#define TEST5_CHUNK_ID					CHUNK_BASE_ID + 1005
#define TEST6_CHUNK_ID					CHUNK_BASE_ID + 1006

// This chunk is saved in its own cache file
#define TOKEN_CACHE_CHUNK				CHUNK_BASE_ID + 2000

#endif