// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// Debugging macros and helpers for Windows
// Does not need or make use of MFC

#ifndef __clientdebug_h
#define __clientdebug_h

#ifdef _WIN32

#include <crtdbg.h>
#include <winsock2.h>

// This definition should be used throughout
// the code so that MFC dependency can be
// avoided, yet the full functionality of ASSERT be retained  
// Its definition should move elsewhere
// so that it can be used by all code including
// core c libraries

#define ZKASSERT	_ASSERTE

// Following is to break the build when ASSERT or VERIFY is used
// ZKASSERT or ZKVERIFY should be used instead.
#undef ASSERT
#define ASSERT		#error

#undef VERIFY
#define VERIFY		#error

#undef DEBUG_NEW
#define DEBUG_NEW	#error


#ifdef DEBUG
# define NODEFAULT   ZKASSERT(0)
#else
# define NODEFAULT   __assume(0)
#endif 



// The point of this foolishness is to try to create
// an ASSERT that works acts as an ASSERT in release mode as well

#ifdef _DEBUG

#define ZKTEST(f) ZKASSERT("ZKTEST FAILED "&&(f))
#define ZKVERIFY(f) ZKASSERT(f)

#else

#define ZKVERIFY(f)          ((void)(f))

#define ZKTEST( f )  \
{ \
	if( !(f) ) \
	{ \
		_CrtDbgBreak(); \
	} \
}


#endif

// Following is to break the build when TRACE is used
// ZKTRACEx should be used instead
#undef TRACE
#undef TRACE0
#undef TRACE1
#define TRACE		#error
#define TRACE0		#error
#define TRACE1		#error

//#define NT_WORKAROUND_FOR_TRACES

// Sometimes, the OutputDebugString() call the following line ends up calling
// from _CrtDbgReport() in dbgrpt.c, seems to freeze...may be related to Q173260 
// since this happens under NT only. NT SP1 & MSVC SP4 do not fix this.
#if defined(_DEBUG) && defined(NT_WORKAROUND_FOR_TRACES)
#include <stdio.h> // for vsnprintf

extern inline void 
ZkSafeTrace(const char* lpszFormat, ...)
{
	va_list args;
	va_start(args, lpszFormat);

	int nBuf;
	char szBuffer[1024] = { 0 };

	nBuf = _vsnprintf(szBuffer, 750, lpszFormat, args);

	if(nBuf > 0)
	{
		Sleep(0); 
		OutputDebugString( szBuffer ); 
	}

	va_end(args);
}
#else
extern inline void ZkSafeTrace(const char* lpszFormat, ...) {}
#endif //_DEBUG


// These definition should be used in place of TRACE
#ifndef NT_WORKAROUND_FOR_TRACES
	#define ZKTRACE0(sz)					_RPT0( _CRT_WARN, sz );
	#define ZKTRACE1(sz, p1)				_RPT1( _CRT_WARN, sz, p1 );
	#define ZKTRACE2(sz, p1, p2)			_RPT2( _CRT_WARN, sz, p1, p2 );
	#define ZKTRACE3(sz, p1, p2, p3)		_RPT3( _CRT_WARN, sz, p1, p2, p3 );
	#define ZKTRACE4(sz, p1, p2, p3, p4)	_RPT4( _CRT_WARN, sz, p1, p2, p3, p4 );
#else
	#define ZKTRACE0(sz)					ZkSafeTrace( sz );
	#define ZKTRACE1(sz, p1)				ZkSafeTrace( sz, p1 );
	#define ZKTRACE2(sz, p1, p2)			ZkSafeTrace( sz, p1, p2 );
	#define ZKTRACE3(sz, p1, p2, p3)		ZkSafeTrace( sz, p1, p2, p3 );
	#define ZKTRACE4(sz, p1, p2, p3, p4)	ZkSafeTrace( sz, p1, p2, p3, p4 );
#endif


// These definition are used to trace only if the condition 'cond' is verified.
#define ZKCTRACE0( cond, sz )					{ if( cond ) { ZKTRACE0( sz ); } }
#define ZKCTRACE1( cond, sz, p1 )				{ if( cond ) { ZKTRACE1( sz, p1 ); } }
#define ZKCTRACE2( cond, sz, p1, p2 )			{ if( cond ) { ZKTRACE2( sz, p1, p2  ); } }
#define ZKCTRACE3( cond, sz, p1, p2, p3 )		{ if( cond ) { ZKTRACE3( sz, p1, p2, p3 ); } }
#define ZKCTRACE4( cond, sz, p1, p2, p3, p4 )	{ if( cond ) { ZKTRACE4( sz, p1, p2, p3, p4 ); } }

// These definition will execute the function 'func' even in release build.
// The function is assumed to return a char*.
#define ZKxTRACE( func ) \
{ \
	const char* pszRet = func; \
	ZKTRACE0( pszRet ); \
}

#define ZKxTRACE0( sz, func ) \
{ \
	const char* pszRet = func; \
	ZKTRACE1( sz, pszRet ); \
}


// Returns true if the passed parameter points
// to at least nBytes of accessible memory. If bReadWrite is TRUE,
// the memory must be writeable; if bReadWrite is FALSE, the memory may be const.
extern inline bool ZKIsValidAddress(const void* lp, unsigned int nBytes, bool bReadWrite = true )
{
	// simple version using Win-32 APIs for pointer validation.
	return (lp != NULL && !IsBadReadPtr(lp, nBytes) &&
		(!bReadWrite || !IsBadWritePtr((void far*)lp, nBytes)));
}


#define ZKASSERT_POINTER(p, type) \
	ZKASSERT(((p) != NULL) && ZKIsValidAddress((p), sizeof(type), false))


// ZKIsValidString() returns true if the passed pointer
// references a string of at least the given length in characters.
// A length of -1 (the default parameter) means that the string
// buffer's minimum length isn't known, and the function will
// return TRUE no matter how long the string is. The memory
// used by the string can be read-only.
extern inline bool ZKIsValidString(LPCSTR lpsz, int nLength  = -1 )
{
	if (lpsz == NULL)
		return false;

	return ::IsBadStringPtrA(lpsz, nLength) == 0;
}


#define ZKASSERT_NULL_OR_POINTER(p, type) \
	ZKASSERT(((p) == NULL) || ZKIsValidAddress((p), sizeof(type), false))

#endif	// _WIN32

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Linux section -- ol
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifdef ZK_LINUX
#include <stdio.h>
#include <stdlib.h>


// In order to get ZKASSERT to work, you now have to link with frcerrorhandling,
// where this  function is defined. This allows for a UI popup on an assert.
extern void HandleAssert( const char* msg, const char* fileName, int lineNumber );

// Temporary macros. These are pretty sh*tty. XXX fix them

#define _ZK_ASSERTION_FAIL "ASSERTION FAILED "

#define DEBUG_REPORT printf


#if defined(DEBUG) | defined(_DEBUG)
// Hmm... not sure about #expr
#define _ASSERTE(expr) \
	do { if (!(expr)) { \
        HandleAssert( #expr, __FILE__, __LINE__ );\
	}} while (0)
#else
/* Way f'd up.
#define _ASSERTE(expr) \
	do { if (!(expr))  \
	DEBUG_REPORT(_ZK_ASSERTION_FAIL "at %s:%d (%s)\n", __FILE__, __LINE__, #expr);\
	 } while (0)
*/
#define _ASSERTE(expr)
#endif

#define ZKASSERT	_ASSERTE

// Following is to break the build when ASSERT or VERIFY is used
// ZKASSERT or ZKVERIFY should be used instead.
#undef ASSERT
#define ASSERT		#error

#undef VERIFY
#define VERIFY		#error

#undef DEBUG_NEW
#define DEBUG_NEW	#error


#if defined(DEBUG) | defined(_DEBUG)
# define NODEFAULT   ZKASSERT(0)
#else
# define NODEFAULT   
#endif 



// The point of this foolishness is to try to create
// an ASSERT that works acts as an ASSERT in release mode as well

#ifdef _DEBUG

#define ZKTEST ZKASSERT
#define ZKVERIFY(f) ZKASSERT(f)

#else

#define ZKVERIFY(f)          ((void)(f))

// XXX: Not sure if we should continue through the failure. Risky?
#define ZKTEST(f) \
do { \
if (!(f)) { \
	DEBUG_REPORT("ZKTEST FAILED at %s:%d (%s)\n", __FILE__, __LINE__, #f ); \
}} while (0) 
#endif

// Following is to break the build when TRACE is used
// ZKTRACEx should be used instead
#undef TRACE
#undef TRACE0
#undef TRACE1
#define TRACE		#error
#define TRACE0		#error
#define TRACE1		#error


#if defined(DEBUG) | defined (_DEBUG)
  // These definition should be used in place of TRACE
#  define ZKTRACE0(sz)					DEBUG_REPORT( sz ); fflush(stdout);
#  define ZKTRACE1(sz, p1)				DEBUG_REPORT( sz, p1 ); fflush(stdout);
#  define ZKTRACE2(sz, p1, p2)			DEBUG_REPORT( sz, p1, p2 ); fflush(stdout);
#  define ZKTRACE3(sz, p1, p2, p3)		DEBUG_REPORT( sz, p1, p2, p3 ); fflush(stdout);
#  define ZKTRACE4(sz, p1, p2, p3, p4)	DEBUG_REPORT( sz, p1, p2, p3, p4 ); fflush(stdout);

  // These definition are used to trace only if the condition 'cond' is verified.
#  define ZKCTRACE0( cond, sz )					{ if( cond ) ZKTRACE0( sz ); }
#  define ZKCTRACE1( cond, sz, p1 )				{ if( cond ) ZKTRACE1( sz, p1 ); }
#  define ZKCTRACE2( cond, sz, p1, p2 )			{ if( cond ) ZKTRACE2( sz, p1, p2  ); }
#  define ZKCTRACE3( cond, sz, p1, p2, p3 )		{ if( cond ) ZKTRACE3( sz, p1, p2, p3 ); }
#  define ZKCTRACE4( cond, sz, p1, p2, p3, p4 )	{ if( cond ) ZKTRACE4( sz, p1, p2, p3, p4 ); }

// These definition will execute the function 'func' even in release build.
// The function is assumed to return a char*.
#  define ZKxTRACE( func ) \
{ \
	const char* pszRet = func; \
	ZKTRACE0( pszRet ); \
}

#  define ZKxTRACE0( sz, func ) \
{ \
	const char* pszRet = func; \
	ZKTRACE1( sz, pszRet ); \
}

#else 

#  define ZKTRACE0(sz)
#  define ZKTRACE1(sz, p1)
#  define ZKTRACE2(sz, p1, p2)
#  define ZKTRACE3(sz, p1, p2, p3)
#  define ZKTRACE4(sz, p1, p2, p3, p4)
#  define ZKCTRACE0( cond, sz )
#  define ZKCTRACE1( cond, sz, p1 )
#  define ZKCTRACE2( cond, sz, p1, p2 )
#  define ZKCTRACE3( cond, sz, p1, p2, p3 )
#  define ZKCTRACE4( cond, sz, p1, p2, p3, p4 )
#  define ZKxTRACE( func ) 
#  define ZKxTRACE0( sz, func ) 

#endif // DEBUG

// Returns true if the passed parameter points
// to at least nBytes of accessible memory. If bReadWrite is TRUE,
// the memory must be writeable; if bReadWrite is FALSE, the memory may be const.
extern inline bool ZKIsValidAddress(const void* lp, unsigned int nBytes, bool bReadWrite = true )
{
//	return (lp != NULL && !IsBadReadPtr(lp, nBytes) &&
//		(!bReadWrite || !IsBadWritePtr((LPVOID)lp, nBytes)));
// XXX: I'm not sure if there's an equivalent on Linux. mprotect would work, except that it would also _CHANGE_ the protection
	(void)lp;(void) nBytes;(void)bReadWrite; // shut up GCC
	return true;
}


#define ZKASSERT_POINTER(p, type) \
	ZKASSERT(((p) != NULL) && ZKIsValidAddress((p), sizeof(type), false))


// ZKIsValidString() returns true if the passed pointer
// references a string of at least the given length in characters.
// A length of -1 (the default parameter) means that the string
// buffer's minimum length isn't known, and the function will
// return TRUE no matter how long the string is. The memory
// used by the string can be read-only.
extern inline bool ZKIsValidString(const char *lpsz, int nLength  = -1 )
{
	(void)nLength; // shut up GCC
	if (lpsz == NULL)
		return false;
	return true;
//	return ::IsBadStringPtrA(lpsz, nLength) == 0;
//	XXX: FixMe
}


#define ZKASSERT_NULL_OR_POINTER(p, type) \
	ZKASSERT(((p) == NULL) || ZKIsValidAddress((p), sizeof(type), false))

#endif // ZK_LINUX

#endif	// __clientdebug_h
