/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef ZKS_INCL_RZKPOPUP
#define ZKS_INCL_RZKPOPUP

/*===========================================================================
 * EXTERNAL DECLARATIONS
 *=========================================================================*/

#ifndef ZKS_INCL_IZKPOPUP
#include <IZkPopup.h>
#endif

#include <memory>

/*===========================================================================
 * CLASS ZkPopupHandler
 *=========================================================================*/

/*
  This base class is used to implement the object that will do the actual
  UI work. Although it is a bit more convoluted than putting the UI code 
  right in here, it has the added advantage of putting the UI dependencies
  where they belong.

  It also allows to develop a pure text based popup mechanism if we need to 
  do any tests with no UI.

  For the actual ui implementation of popups, have a look in 
  zkui/zkcommonui/linux/UiPopup.h
  
  More convoluted, but more flexible.....
*/
class ZkPopupHandler {

public:

	virtual void openMessageBox( 
        const tstring& msg, 
        const tstring& title, 
        IconType ) = 0;

	virtual bool openQuestionBox( 
        const tstring& msg,         
        BtnType, 
        IconType, 
        bool ) = 0;

	virtual bool openMessageCheckBox(	const tstring&,
										const tstring& in_CheckMsg,
										bool		in_bDefaultChecked, 
										IconType) = 0;

	virtual bool openQuestionCheckBox(	const tstring& in_strQuestion,
                                        bool& out_bChecked,
                                        uint32_t in_btn1textId,
                                        uint32_t in_btn2textId,
                                        uint32_t in_checkMsgId,
                                        bool in_bDefaultChecked, 
                                        IconType ) = 0;

	virtual tstring openRequestStringBox(	const tstring& in_strTitle,
                                            const tstring& in_strPrompt,
                                            const tstring& in_strInitialValue,
                                            int in_nMinLength,
                                            PopupPosition in_popupPos) = 0;

};

/*===========================================================================
 * CLASS RZkPopup
 *=========================================================================*/

/*
  Linux version of the IZkPopup class. Please see ../xp/IZkPopup.h for details.

  As an argument, the constructor expects a pointer to the object that will
  do the actual work of poping a message to the user.

  Asks for an auto_ptr so the RZkPopup can safely delete the object when it goes
  out of scope.
*/

class RZkPopup : public IZkPopup {

public: 

    /*----- member functions -----*/

    // Builds the single instance of the RZkPopup singleton.
    // Absolutely requires a ZkPopupHandler at construction.
    // auto_ptr used to make obvious the fact you have to give away your pointer.
    // The given object will be deleted by RZkPopup.
    RZkPopup( ZKSTD::auto_ptr< ZkPopupHandler > handler );
    virtual ~RZkPopup();

protected:

    /*----- member functions -----*/
      
    virtual void openMessageBox( uint32_t msgId, IconType, PopupPosition, uint32_t titleId );
	virtual void openMessageBox( const tstring&, IconType, PopupPosition, const tstring& );

    virtual bool openQuestionBox( uint32_t msgId, BtnType, PopupPosition );
	virtual bool openQuestionBox( const tstring&, BtnType, IconType, bool, PopupPosition );    

	virtual bool openMessageCheckBox(	uint32_t msgId, 
										uint32_t	in_checkMsgId,
										bool		in_bDefaultChecked, 
										IconType,
										PopupPosition );    
	virtual bool openMessageCheckBox(	const tstring&,
										uint32_t	in_checkMsgId,
										bool		in_bDefaultChecked, 
										IconType,
										PopupPosition );    

	virtual bool openQuestionCheckBox(	uint32_t in_msgId,
									bool& out_bChecked,
									uint32_t in_btn1textId,
									uint32_t in_btn2textId,
									uint32_t in_checkMsgId,
									bool in_bDefaultChecked, 
									IconType,
									PopupPosition);
	virtual bool openQuestionCheckBox(	const tstring& in_strQuestion,
									bool& out_bChecked,
									uint32_t in_btn1textId,
									uint32_t in_btn2textId,
									uint32_t in_checkMsgId,
									bool in_bDefaultChecked, 
									IconType,
									PopupPosition);

	virtual tstring openRequestStringBox(	const tstring& in_strTitle,
                                            const tstring& in_strPrompt,
                                            const tstring& in_strInitialValue,
                                            int in_nMinLength,
											PopupPosition in_popupPos);

private:

    /*----- data members -----*/

    ZKSTD::auto_ptr< ZkPopupHandler >  m_handler;
};
#endif
