// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "RZkRegistry.h"

#include "clientdebug.h"

#include <unistd.h>
#include <iostream>

static const char  ZERO_REG_KEY[] = "SOFTWARE/Zero Knowledge/Freedom";

RZkRegistry::RZkRegistry(const char *fileName)
    : m_Registry( 0 )
{
    init(fileName);
}

void
RZkRegistry::init(const char *fileName)
{
    int count = 0;
    REGERR err;

    if ( m_Registry != 0 ) {
        ZKASSERT("RZkRegistry::init called after registry initialization." && 0);
        return;
    }

    // we try to delete and reopen the registry as much as twice.
    // There's no particular reason to do it twice, but a second try
    // if the file was initially corrupt seems like not such a bad
    // idea --ol
    while (count <= 3) {
        err = NR_RegOpen(const_cast<char *>(fileName), &m_Registry);
        if (err != REGERR_OK) {
            ZKSTD::cerr << "Registry Initilization failed" << ZKSTD::endl;
            unlink(fileName);
            count++;
            m_Registry = 0;
            continue;
        } 

        err = NR_RegAddKey(m_Registry, ROOTKEY_COMMON, 
                       const_cast<char *>(ZERO_REG_KEY) , &m_FreedomKey);
        if (err != REGERR_OK) {
            ZKSTD::cerr << "Registry Initilization failed" << ZKSTD::endl;
            m_Registry = 0;
            NR_RegClose(m_Registry);
            unlink(fileName);
            count++;
            continue;
        } else break;

    }
    if (m_Registry == 0) {
        return;
    }
}

RZkRegistry::~RZkRegistry()
{
    if ( m_Registry == 0 ) {
        return;
    } else {
        NR_RegClose(m_Registry);
    }
}

bool
RZkRegistry::onGetFlag( const char* keyName )
{

    unsigned long intval;
    if (!onGetDWORD(keyName, intval))
	return false;

    return (intval == 1);

}

void
RZkRegistry::onSetFlag( const char* keyName, bool value )
{
    onSetDWORD(keyName, (unsigned long) value);
}

bool
RZkRegistry::onGetDWORD( const char* keyName, unsigned long& value )
{
    REGERR err;
    uint32 len = sizeof value;
    uint32 lvalue = value;
    err = NR_RegGetEntry( m_Registry, m_FreedomKey, const_cast<char *>(keyName), &lvalue,
			  &len );
    value = lvalue;
    return ( err == REGERR_OK );
}

void
RZkRegistry::onSetDWORD( const char* keyName, unsigned long value )
{
    NR_RegSetEntry( m_Registry, m_FreedomKey, const_cast<char *>(keyName), REGTYPE_ENTRY_BYTES,
                    (void *)&value, sizeof value);
}

void 
RZkRegistry::onGetShortArray( const char* keyName, ZKSTD::vector<uint16_t >&  out_vec )
{
    /* getValueLength reports the size in bytes. */
    long length = getValueLength(keyName);
    if (length <= 0)
        return;
    ZKASSERT("no partial shorts in array:" && (length % 2 == 0));
    uint16_t *buf = new uint16_t[length / 2];

    uint32 reglen = (uint32) length;
    REGERR err = NR_RegGetEntry(m_Registry, m_FreedomKey, const_cast<char *>(keyName), (void *)buf,
                                &reglen);
    if (err != REGERR_OK)
        return;                 // yay, no error reporting
    
    // copy the buffer into out_vec
    out_vec.clear();
    ZKSTD::copy( buf, buf + (length / 2), ZKSTD::back_inserter(out_vec));
    delete [] buf;
}

void
RZkRegistry::onSetShortArray( const char* keyName, const ZKSTD::vector<uint16_t > in_vec )
{
    unsigned long size = in_vec.size(), length = size * sizeof(uint16_t);
    uint16_t *buf = new uint16_t[size];

    for (size_t i = 0; i < size; i++)
        buf[i] = in_vec[i];
    
    REGERR err = NR_RegSetEntry(m_Registry, m_FreedomKey, const_cast<char *>(keyName),
                                REGTYPE_ENTRY_BYTES, (void *)buf, length);
    ZKASSERT( "registry store succeeded: " && err == REGERR_OK);
}

bool
RZkRegistry::onGetString( const char* keyName, ZKSTD::string& value)
{
    long length = getValueLength(keyName);
    if (length < 0)
        return false;
    if (length == 0)
        return true;
    
    length++;                   // trailing NUL
    char *buf = new char[length];
    REGERR err = NR_RegGetEntryString(m_Registry, m_FreedomKey, const_cast<char *>(keyName),
                                   buf, length);
    if (err == REGERR_OK)
        value = buf;
    delete [] buf;
    return (err == REGERR_OK);
}

void
RZkRegistry::onSetString( const char* keyName, const char* value )
{
    NR_RegSetEntryString( m_Registry, m_FreedomKey,
                          const_cast<char *>(keyName),
                          const_cast<char *>(value));
}

long
RZkRegistry::onDeleteValue(const char* keyName )
{
    // calling code appears to expect some windows success code (ERROR_SUCCESS). Tough.
    return NR_RegDeleteEntry(m_Registry, m_FreedomKey,
                             const_cast<char *>(keyName)) != REGERR_OK;
}

long
RZkRegistry::getValueLength(const char *keyName)
{
    REGERR err;
    REGINFO info = { sizeof info, 0, 0 };
    err = NR_RegGetEntryInfo( m_Registry, m_FreedomKey, const_cast<char *>(keyName),
			      &info );
    if (err != REGERR_OK)
	return -1;
    return info.entryLength;
}
