/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef ZKS_INCL_ZKPIPE
#define ZKS_INCL_ZKPIPE

#include <AutoInstCount.h>


/*===========================================================================
 * CLASS ZkPipe
 *=========================================================================*/

/* Simple class for managing Unixy IO redirection and pipes
 * Nothing fancy, just gives the various syscalls a C++ feel
 * --ol
 */

class ZkPipe AUTO_INSTANCE_COUNTING1( ZkPipe ) 
{

public: 

    /*----- member functions -----*/

    ZkPipe();
    ~ZkPipe();
    
	// Purpose:  Make the pipe represented by this class into a read-only pipe
    //           by closing write end.
	// Requires: Ctor was called
    void MakeReadOnly();

	// Purpose:  Make the pipe represented by this class into a write-only pipe
    //           by closing read end
	// Requires: Ctor was called
    void MakeWriteOnly();

	// Purpose:  Redirect the open fd used by this pipe to an existing fd passed as an argumnt
    //           usful for stdin/stdout redirection.
    //           if close_local is set to TRUE, the pipe will close it's copy of the redirected
    //           fd (and will cease to be useful for anything)
	// Requires: m_state is not BIDIR_PIPE (either MakeRead or MakWrite have been called)
    void RedirectRead(int in_fd, bool close_local = false);
    void RedirectWrite(int in_fd, bool close_local = false);

	// Purpose:  Close this pipe (called by Dtor)
    void Close();

	// Purpose:  Conversion to int, gets active FD for syscalls.
//    operator int() { return m_fdset[which_fd()]; };

    // Purpose: get the read fd for this pipe.
    // Requires: Pipe has been created
    int GetRead() {
        ZKASSERT("Read is valid" && ( m_state == READ_PIPE || m_state == BIDIR_PIPE ) );
        return m_fdset[0];
    }

    // Purpose: get the read fd for this pipe.
    // Requires: Pipe has been created
    int GetWrite() {
        ZKASSERT("Write is valid" && ( m_state == WRITE_PIPE || m_state == BIDIR_PIPE ) );
        return m_fdset[1];
    }

private: 

    /*----- constants -----*/

    enum PARENT_DIRECTION {
        BIDIR_PIPE,
        READ_PIPE,
        WRITE_PIPE,
        CLOSED_PIPE
    };

    /*----- member functions -----*/

    int which_fd() { 
        switch ( m_state ) {
            case READ_PIPE : 
                return 0;
            case WRITE_PIPE :
                return 1;
            default:
                ZKASSERT("Both ends of pipe are open" && 0 );                
        }
        ZKASSERT( "Impossible" && 0 );
        return 12345; // nevr reached, kill warning
    };
            

	// Prohibited funcs - not implemented.
	ZkPipe( const ZkPipe& );
	const ZkPipe& operator=( const ZkPipe& );
    
    /*----- data members -----*/

    int m_fdset[2];
    int m_state;

};

      

#endif


