// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):



//===========================================================================
// EXTERNAL DECLARATIONS
//===========================================================================

#include <ZkSmallSharpTool.h>

#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <signal.h>

//===========================================================================
// CLASS ZkSmallSharpTool
//===========================================================================

// Statics
ZkMutex ZkSmallSharpTool::m_Mutex = NULL;
ZkITimer ZkSmallSharpTool::m_Timer = NULL;
ZkSmallSharpTool *ZkSmallSharpTool::m_ActiveChild = NULL;


//---------------------------------------------------------------------------
//
ZkSmallSharpTool::ZkSmallSharpTool( char * in_command )
   : m_InputPipe( NULL )
   , m_OutputPipe( NULL )
   , m_ChildPID( 0 )
   , m_Timeout( ZK_INFINITE )
   , m_Command( in_command )
   , m_State( STOPPED )
{
    if (!m_Mutex) 
    {
        // We leak this mutex intentionally. (cleanup on exit)
        ZkError err;
        ZKVERIFY( zkErrorNew(&err) );
        ZKVERIFY( zkMutexNew(err, &m_Mutex) );
        ZKVERIFY( zkErrorFree(&err) );
    }
}

//---------------------------------------------------------------------------
//
ZkSmallSharpTool::~ZkSmallSharpTool( )
{
    if (m_State == RUNNING) {
        Terminate();
    }
}

//---------------------------------------------------------------------------
//
void ZkSmallSharpTool::Execute( ZKSTD::vector<ZKSTD::string> &args,
                                unsigned int in_io_flags = 0,
                                unsigned long in_timeout = ZK_INFINITE) throw ( RZkException )
{
    // Get a process wide lock. This class is not a singleton, the static mutex
    // makes sure we execute children lock step. This is really just to get the
    // termination timer to work properly ... we could change the timer API to pass
    // contect data if we need to support running more than one child at a time
    // there is no foreseeable need for this functionality, however.
    // --ol
    ZkError err;
    ZKASSERT( m_Mutex != NULL );
    ZKASSERT( m_State != RUNNING );
    ZKVERIFY( zkErrorNew(&err) );
    ZKVERIFY( zkMutexLock(err, m_Mutex) );

    // Set up argv for child ( we do this in the parent so we can catch errors _before_ forking )
    char **argv;
    ZKVERIFY( argv = (char **)malloc((args.size() + 2) * sizeof(char*)) );
    argv[0] = const_cast<char *>(m_Command.c_str());
    
    int i = 1;
    for (ZKSTD::vector<ZKSTD::string>::const_iterator it = args.begin();
         it != args.end(); ++it) {

        argv[i++] = const_cast<char*>((*it).c_str());
    }
    argv[args.size() + 1] = NULL;

    // Set up pipes
    ZKASSERT( m_InputPipe == NULL );
    if (in_io_flags & IO_INPUT) {
        m_InputPipe = new ZkPipe();
    }

    ZKASSERT( m_OutputPipe == NULL );
    if (in_io_flags & IO_OUTPUT) {
        m_OutputPipe = new ZkPipe();
    }

    m_Timeout = in_timeout;

    // Do the 'ol fork. XXX needs better eror handling.
    ZKVERIFY( (m_ChildPID = fork()) != -1 );

    // Er, well this probably looks strange if your a windows person.
    // fork branches us into two process spaces, the existing (parent)
    // one that called fork, and a child. Fork is returning twice, once in each
    // space, and giving us different values so we can differentiate. Too cool for win32
    if (m_ChildPID == 0)
    {
        // Child's address space.
        // Confused? The pipes 're reversed for the child process.
        if (m_InputPipe) 
        {
            m_InputPipe->MakeWriteOnly();
            m_InputPipe->RedirectWrite(STDOUT_FILENO, true);
        }
        if (m_OutputPipe)
        {
            m_OutputPipe->MakeReadOnly();
            m_OutputPipe->RedirectRead(STDIN_FILENO, true);
        }
        ChildFunction(argv);        
    } 
    else 
    {
        // Parent's address space.
        free(argv);
        if (m_InputPipe) 
        {
            m_InputPipe->MakeReadOnly();            
        }
        if (m_OutputPipe)
        {
            m_OutputPipe->MakeWriteOnly();
        }
        m_State = RUNNING;
        SetTimer();
    }
}

//---------------------------------------------------------------------------
//
void ZkSmallSharpTool::ChildFunction(char **argv) throw ( RZkException )
{
    char *_envp = NULL;
//    char **envp = &_envp;
    // pipes are, well, piped. exec our child application
    // Note that at jwilkins suggestion we just hose the entire environment.
    // This will likely break some stuff, but this is really only intended for 
    // running little light utils apps.
#if defined(DEBUG) || defined(_DEBUG)
    fprintf(stderr, "About to execute: %s - %s\n", m_Command.c_str(), argv[0]);
#endif

    execv(m_Command.c_str(), argv);

    // if exec worked, we will not get here.

    // stderr should still be shared with our parent, so the error will
    // come out on a console or whatnot.
    perror("Unable To Execute Child Process");
    
    exit(-1);
}


//---------------------------------------------------------------------------
//
int ZkSmallSharpTool::Wait() throw ( RZkException )
{
    ZKASSERT( m_ChildPID );
    ZKASSERT( m_State == RUNNING );

    pid_t rpid = 0;
    int rval = 0;
    ZKVERIFY( ((rpid = waitpid(m_ChildPID, &rval, WUNTRACED)) > 0) );
    // Should be ZKTRACE perhaps.
    if (rpid <= 0) perror("Unable To Join Child Process.");

    KillTimer();

    if (m_InputPipe) 
    {
        m_InputPipe->Close();
        delete m_InputPipe;
        m_InputPipe = NULL;
    }
    if (m_OutputPipe)
    {
        m_OutputPipe->Close();
        delete m_OutputPipe;
        m_OutputPipe = NULL;
    }
    m_State = STOPPED;
    m_ChildPID = 0;
    // Unlock our mutex.
    ZkError err;
    ZKVERIFY( zkErrorNew(&err) );
    ZKVERIFY( zkMutexUnlock(err, m_Mutex) );
    
    return WEXITSTATUS( rval );
    
}

//---------------------------------------------------------------------------
//
void ZkSmallSharpTool::Terminate() throw ( RZkException )
{
    ZKASSERT( m_ChildPID );
    ZKASSERT( m_State == RUNNING );

    pid_t deadDuck = m_ChildPID;

    ZKTRACE1( "Sending TERM to child PID %d\n", deadDuck);

    ZKVERIFY( kill(deadDuck, SIGTERM) == 0);

    if (  kill(deadDuck, SIGTERM) == -1 && errno == ESRCH )
    {
        sleep(1);
        ZKTRACE1( "Sending KILL to derilect child PID %d!\n", deadDuck);
        kill(deadDuck, SIGKILL);
        // Oh yeah, _real_ dead at this point.
    }

}

//---------------------------------------------------------------------------
//
void ZK_ITIMER_CALLBACK ZkSmallSharpTool::timerProc( void *, unsigned int, 
                                   unsigned int, unsigned long millis )
{
    ZkSmallSharpTool::m_ActiveChild->Terminate();
}

//---------------------------------------------------------------------------
//
void ZkSmallSharpTool::SetTimer()
{
    ZKASSERT( m_Timeout > 100 || m_Timeout == ZK_INFINITE );
    if (m_Timeout == ZK_INFINITE) ZKTRACE0("ZkSmallSarpTool not setting timer == INFINITE");
    if (m_Timeout != ZK_INFINITE) 
    {
        ZkError err;
        ZKVERIFY( zkErrorNew( &err ) );
        m_ActiveChild = this;
        ZKVERIFY( zkITimerNew( err, &m_Timer, timerProc, m_Timeout ) );
        zkErrorFree( &err );
    }
}

//---------------------------------------------------------------------------
//
void ZkSmallSharpTool::KillTimer()
{
    if (m_Timeout != ZK_INFINITE)
    {
        ZkError err;
        ZKVERIFY( zkErrorNew( &err ) );
        m_ActiveChild = NULL;
        ZKVERIFY( zkITimerFree( err, &m_Timer ) );
        zkErrorFree( &err );
        m_Timeout = 0;
    }
}
