// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "ZkUnicode.h"

#include <memory>
#include <clientdebug.h>
#include <iconv.h>
#include <unistd.h>
#include <iconv.h>
#include <locale.h>
#define __USE_XOPEN
#include <langinfo.h>

#include <zkError.h>

#include <string.h> // for swab


#define ZKUNICODE_DEBUG 0
#if ZKUNICODE_DEBUG
ZKUNICODE_ZKTRACE1(sz,p1) ZKTRACE1(sz,p1)
#else
#define ZKUNICODE_ZKTRACE1(sz,p1) 
#endif

// Used to notify what byte order the Unicode data is saved in
#define BYTE_ORDER_MARK            0xFEFF
#define BYTE_ORDER_MARK_SWAPPED    0xFFFE


// For linux are there equivalences to active code pages?
// If only standard ascii code page supported we can keep that.
uint32_t ZkUnicode::m_WindowsCodePage = 0 ;


int ZkUnicode::StorageUnicodeBufferToString (
	const uint16_t * in_pUnicode, 
	size_t in_BufferSize, 
	ZKSTD::string& out_str )

{
  iconv_t cd;  
  int retval = 0;
  
	// FIXME Check for null pointer
        if (!in_pUnicode || !in_BufferSize)
	{
	       ZKASSERT( "Unaccepted values." && 0 ) ;
	       return retval;
	        
	}    
         //Note: this is BACKWARDS. Linux UTH16 conversion works on big endian unicode data --ol
	if ( in_pUnicode[ 0 ] == BYTE_ORDER_MARK_SWAPPED )
	{
		// Skip past the BYTE_ORDER_MARK and convert the rest of the buffer
		retval = ZkUnicode::UnicodeBufferToString( &in_pUnicode[ 1 ], in_BufferSize - 2, out_str ) ;
	} 
	else if ( in_pUnicode[ 0 ] == BYTE_ORDER_MARK )
	{
		// This must have been saved on a machine with the different endian order
		// so we swap the bytes of each character manually before converting
	        // this is the patch for newer versions of the glibc in linux
	        //
	  
                ZKTRACE0(" ZkUnicode::StorageUnicodeBufferToString: swapping byte order in string\n");             
                uint16_t * pSwappedBuffer = NULL ;
		zkMemAlloc( NULL, (void**) &pSwappedBuffer, in_BufferSize - 2 ) ;

		swab((char *)&in_pUnicode[1], reinterpret_cast<char *>(pSwappedBuffer), in_BufferSize - 2);

		retval = ZkUnicode::UnicodeBufferToString( (uint16_t*) pSwappedBuffer, in_BufferSize - 2, out_str ) ;
		zkMemFree( NULL, (void**) &pSwappedBuffer ) ;
	}
	else
	{
		// UTF-16LE does not require Byte Order Mark,  
		// this is newer version of glibc do not do th econversion
	        retval = ZkUnicode::UnicodeBufferToString( in_pUnicode , in_BufferSize , out_str);
                //ZKASSERT( "No BOM or BSBOM in unicode string." && 0 ) ;
	}
	return retval;
}

int
ZkUnicode::StringToUnicodeBuffer( 
	const ZKSTD::string& in_str, 
	uint16_t **       io_ppUnicode, 
	size_t &          out_BufferSize )
{
    iconv_t cd;
    size_t insize, outsize, nconv;
    char *inbuf = const_cast<char*>(in_str.c_str());

    ZKASSERT("Only system code page is suppoted" && m_WindowsCodePage == 0 );

    outsize = insize = in_str.size() + 1;
    outsize *= sizeof(uint16_t); // (double byte)

    char *outbuf = (char*) malloc( outsize ) ;
    ZKASSERT( outbuf != NULL ) ;
    memset( outbuf, 0 , outsize ) ;
    
    *io_ppUnicode = (uint16_t*)(outbuf);


    char *charset_name = nl_langinfo(CODESET);
    ZKASSERT("We have a valid codepage" && charset_name);
    
    cd = iconv_open("UTF16", charset_name);

    if (cd == (iconv_t) -1)
    {
      // try UTF-16LE , this is added for newer glibc
      
      cd = iconv_open("UTF-16LE", charset_name);
      if (cd == (iconv_t) -1)
	{
	  ZKASSERT("iconv_open failed" && 0);
	  // the caller will free the memory regardless of the return value
	  //free(outbuf);
	  return 0;
	}
    } 
    char *wrptr = outbuf;
    
    ZKUNICODE_ZKTRACE1("ZkUnicode::StringToUnicodeBuffer insizee %d: \n",insize);
    ZKUNICODE_ZKTRACE1("ZkUnicode::StringToUnicodeBuffer outsize %d: \n",outsize);
    ZKUNICODE_ZKTRACE1("ZkUnicode::StringToUnicodeBuffer wrptr %8x: \n",wrptr);
    ZKUNICODE_ZKTRACE1("ZkUnicode::StringToUnicodeBuffer inbuf %8x: \n",inbuf);

    nconv = iconv(cd, const_cast<const char**>(&inbuf), &insize, &wrptr, &outsize);
    
    ZKUNICODE_ZKTRACE1("ZkUnicode::StringToUnicodeBuffer insizee %d: \n",insize);
    ZKUNICODE_ZKTRACE1("ZkUnicode::StringToUnicodeBuffer outsize %d: \n",outsize);
    ZKUNICODE_ZKTRACE1("ZkUnicode::StringToUnicodeBuffer wrptr %8x: \n",wrptr);
    ZKUNICODE_ZKTRACE1("ZkUnicode::StringToUnicodeBuffer inbuf %8x: \n",inbuf);

    if (nconv == (size_t)-1) {

	ZKUNICODE_ZKTRACE1("ZkUnicode::StringStorageBufferToString Error:%8x \n",errno);

        ZKASSERT("Conversion failed" && 0);
	// FIXME the caller does not check for the error returned
        //free(outbuf);
	iconv_close( cd );
        return 0;
    }


    out_BufferSize = wrptr - outbuf;

    iconv_close( cd );

    return out_BufferSize;
}

int
ZkUnicode::StringToStorageUnicodeBuffer( 
	const ZKSTD::string& in_str, 
	uint16_t **       io_ppUnicode, 
	size_t &          out_BufferSize )
{
    iconv_t cd;
    size_t insize, outsize, nconv;
    char *wrptr;
    char *inbuf = const_cast<char*>(in_str.c_str());

    ZKASSERT("Only system code page is suppoted" && m_WindowsCodePage == 0 );

    outsize = insize = in_str.size() + 1;
    outsize *= sizeof(uint16_t); // (double byte)
    outsize += sizeof(uint16_t); // headroom for BOM, required for UTF16 conv. 

    char *outbuf = (char*) malloc( outsize );
    ZKASSERT( outbuf != NULL ) ;
    memset( outbuf, 0 , outsize ) ;
    
    *io_ppUnicode = (uint16_t*)(outbuf);

    
    char *charset_name = nl_langinfo(CODESET);
    ZKASSERT("We have a valid codepage" && charset_name);
    cd = iconv_open("UTF16", charset_name);
    if (cd == (iconv_t) -1)
    {
      // try UTF-16LE , this is added for newer glibc
      
      cd = iconv_open("UTF-16LE", charset_name);
      if (cd == (iconv_t) -1)
	{
	  ZKASSERT("iconv_open failed" && 0);
	  // the caller will free the memory regardless of the return value
	  //free(outbuf);
	  return 0;
	}
      // Does not require Byte Order Mark, Always LE
      wrptr = outbuf;
    }  
    else
    { 
      *io_ppUnicode[0] = BYTE_ORDER_MARK_SWAPPED;
      wrptr = outbuf + sizeof(uint16_t); //start conversion after BOM
    }
    // iconv increments the pointer and decrements the size as it converts
    ZKUNICODE_ZKTRACE1("ZkUnicode::StringToStorageUnicodeBuffer insize %d: \n",insize);
    ZKUNICODE_ZKTRACE1("ZkUnicode::StringToStorageUnicodeBuffer outsize %d: \n",outsize);
    ZKUNICODE_ZKTRACE1("ZkUnicode::StringToStorageUnicodeBuffer wrptr %8x: \n",wrptr);
    ZKUNICODE_ZKTRACE1("ZkUnicode::StringToStorageUnicodeBuffer inbuf %8x: \n",inbuf);

    nconv = iconv(cd, const_cast<const char**>(&inbuf), &insize, &wrptr, &outsize);

    ZKUNICODE_ZKTRACE1("ZkUnicode::StringToStorageUnicodeBuffer iconv returned %d: \n",nconv);
    
    ZKUNICODE_ZKTRACE1("ZkUnicode::StringToStorageUnicodeBuffer insize %d: \n",insize);
    ZKUNICODE_ZKTRACE1("ZkUnicode::StringToStorageUnicodeBuffer outsize %d: \n",outsize);
    ZKUNICODE_ZKTRACE1("ZkUnicode::StringToStorageUnicodeBuffer wrptr %8x: \n",wrptr);
    ZKUNICODE_ZKTRACE1("ZkUnicode::StringToStorageUnicodeBuffer inbuf %8x: \n",inbuf);


    if (nconv == (size_t)-1) {
	ZKUNICODE_ZKTRACE1("ZkUnicode::StringStorageBufferToString Error:%8x \n",errno);

        ZKASSERT("Conversion failed" && 0);
	// The caller will free the momory regardless of the ret value
        //free(outbuf);
	iconv_close(cd);
        return 0;
    }

    //*io_ppUnicode = (uint16_t*)(outbuf);
    //*io_ppUnicode[0] = BYTE_ORDER_MARK_SWAPPED;
    out_BufferSize = wrptr - outbuf;

    iconv_close(cd);

    return out_BufferSize;
}

int 
ZkUnicode::UnicodeBufferToString( 
	const uint16_t * in_pUnicode, 
	size_t in_BufferSize, 
	ZKSTD::string& out_str )
{
    iconv_t cd;
    size_t insize, outsize, nconv,converted_size;
    size_t ret=0;

    ZKASSERT("No half unicode chars" && in_BufferSize % 2 == 0);
    ZKASSERT("Only system code page is supported" && m_WindowsCodePage == 0 );

    outsize = insize = in_BufferSize;
    outsize = outsize / 2 + 1 ; // (bytes)
    converted_size = outsize;
    // Pad so we are certain we have a null
    ZKSTD::auto_ptr<char> buf( new char[outsize] );

    memset( buf.get(), 0, outsize );


    char *charset_name = nl_langinfo(CODESET);
    ZKASSERT("We have a valid codepage" && charset_name);

    cd = iconv_open(charset_name, "UTF16");
    if (cd == (iconv_t) -1)
      {
	// try another time for newer glibc
	cd = iconv_open(charset_name, "UTF-16LE");
	if (cd == (iconv_t) -1)
	  {
      
	    ZKUNICODE_ZKTRACE1("ZkUnicode::UnicodeBufferToString Error:%8x \n",errno);
	    ZKASSERT("iconv_open failed" && 0);

	    return ret;
	    
	  }  
	  
      }	

    // These are because iconv modifies it's arguments.
    char *wrptr = buf.get();
    char *rdptr = (char*)in_pUnicode;
   

    ZKUNICODE_ZKTRACE1("ZkUnicode::UnicodeBufferToString insize:%d \n",insize);
    ZKUNICODE_ZKTRACE1("ZkUnicode::UnicodeBufferToString outsize:%d \n",outsize);
    ZKUNICODE_ZKTRACE1("ZkUnicode::UnicodeBufferToString rdptr:%8x \n",rdptr);
    ZKUNICODE_ZKTRACE1("ZkUnicode::UnicodeBufferToString wrptr:%8x \n",wrptr);
    // nconv moves the rdptr and wrptr ahead and decrements outsize and insize
    nconv = iconv(cd, (const char**)&rdptr, &insize, &wrptr, &outsize);
    ZKUNICODE_ZKTRACE1("ZkUnicode::UnicodeBufferToString insize:%d \n",insize);
    ZKUNICODE_ZKTRACE1("ZkUnicode::UnicodeBufferToString outsize:%d \n",outsize);
    ZKUNICODE_ZKTRACE1("ZkUnicode::UnicodeBufferToString rdptr:%8x \n",rdptr);
    ZKUNICODE_ZKTRACE1("ZkUnicode::UnicodeBufferToString wrptr:%8x \n",wrptr);

    //ZKASSERT(nconv != (size_t)-1);
    if (nconv == (size_t)-1){
      
	ZKUNICODE_ZKTRACE1("ZkUnicode::UnicodeBufferToString Error:%8x \n",errno);
        ZKASSERT("Conversion failed" && 0);
        iconv_close(cd);
	return ret;
	    
    }
    // outsize is not necessarily zero
    ret = converted_size - outsize; 
    out_str = buf.get() ;

    iconv_close(cd);
    return ret ;
}
