// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "IZkFileHelp.h"

#include <client.h>

#ifdef WIN32
#  include <io.h> // for _access
#  define W_OK 02
#else
#  include <unistd.h>
#endif


IZkFileHelp::IZkFileHelp()
{
}

// Functions with platform specific implementation.

int
IZkFileHelp::FolderPicker( const char* in_pszTitle, ZKSTD::string& out_strPickedFile ) 
{
	return get().onFolderPicker(in_pszTitle, out_strPickedFile);
}

bool
IZkFileHelp::IsNetWorkDrive(const ZKSTD::string& in_strPath)
{
	return get().onIsNetWorkDrive(in_strPath);
}

bool
IZkFileHelp::DoesFileOrDirExist( const char* in_pszFile ) throw(RZkException) 
{
	return get().onDoesFileOrDirExist(in_pszFile);
}

void
IZkFileHelp::CreateDirectory( const char* in_pszDirectory ) throw(RZkException )
{
	get().onCreateDirectory(in_pszDirectory);
}


ZKSTD::string
IZkFileHelp::BackupAndErase( const ZKSTD::string& in_rstrFileName )
{
	return get().onBackupAndErase(in_rstrFileName);
}


// Functions with XP implementation from zkbase/general/xp/file_hlp.h
// Moved here to fix 5771 nym file corruption which was contributed to
// by overzealous function inlining --ol

/* Helper functions for File routines - Mario Contestabile 18 March 2000
- Gabriel Duford 13 April 2000 : ExtractFileName

	Current helpers are:
	FileCopy      - Copy contents of a file into another given two FILE pointers
	FileCopy      - Copies contents of one file into another given their names
	FileCopywithverification - Copies contents of one file into another given their names, you decide whether to overwrite or not
	GetTempFileName  - Use to get a valid temporary file name
	FileExists	  - Verify if a given file name exists
	FileCopytodir - Copies contents of one file into another directory using the same file name
	ReadTextFile  - Places the contents of a text file into a string separated by \n
	ReadBinaryFile - Places the contents of a binary file into a string
	GetFileSize	  - Returns the size in bytes of a file
	GetLinesExceedingWidth - Returns the lines in a file exceeding the indicated length
	ExtractFileName - Extract the file name from a path (absolute || relative || none).
	WriteBinaryFile - Write the buffer into the specified file

*/


// PURPOSE:	Get a string which can represent a valid and unique file name
//			uses the P_tmpdir directory from stdio.h (which may be "\\")
//			so a name may be in the form "\\d9sdjs9"
//			do not call free() on the string
//			Use a buffer at least L_tmpnam * 2 size
//			returns a NULL string if name cannot be generated
//			If you're using a FILE* you may try using tmpfile() in place of this function
//			since the file will get deleted when the app terminates.
// REQUIRE:	
// PROMISE: 
// THROWS:	Nothing
// EXAMPLE: See Test15
void IZkFileHelp::GetTempFileName(char* out_pszFileName ) throw()
{
	ZKASSERT(out_pszFileName);
	tmpnam(out_pszFileName);	
	ZKASSERT(out_pszFileName);
}


// PURPOSE:	Verifies byte by byte that two files are equal
// REQUIRE:	
// PROMISE: Neither file is not changed by this call
//			However FILE* are pointing to an undetermined location
//			(Use fseek to return to a known place)
// THROWS:	Nothing

bool IZkFileHelp::AreFilesEqual( FILE* in_f1, FILE* in_f2 ) throw()
{	
	char	buf1[BUFSIZ] = { 0 };
	char	buf2[BUFSIZ] = { 0 };

	if(!in_f1 || !in_f2) return false ;

	fseek(in_f1, 0, SEEK_SET );
	fseek(in_f2, 0, SEEK_SET );

	while(!feof(in_f1))
	{
		const size_t count1 = fread(buf1, 1, BUFSIZ, in_f1);
		const size_t count2 = fread(buf2, 1, BUFSIZ, in_f2);

		if(ferror(in_f1))
		{	
			return false ;
		}

		if(ferror(in_f2))
		{	
			return false ;
		}


		if ( count1 != count2 )
		{	
			return false ;
		}

		if ( 0 != memcmp( buf1, buf2, count1 ) )
		{
			return false ;
		}
	}
	return true;
}



// PURPOSE:	Copy contents of a file into another given two FILE pointers
//			Returns 0 if either pointer is NULL
//			Returns 1 on success, 0 on error
// REQUIRE:	in_source is open for reading (r or w+)
// PROMISE: Rewinds in_source file pointer to beginning after verifying both pointers aren't NULL
// THROWS:	Nothing
// EXAMPLE: See Test15
int IZkFileHelp::FileCopy( FILE* in_dest,  FILE* in_source ) throw()
{	
	char	buf[BUFSIZ] = { 0 };

	if(!in_dest || !in_source) return 0;

	fseek(in_source, 0, SEEK_SET);

	while(!feof(in_source))
	{
		const size_t count = fread(buf, 1, BUFSIZ, in_source);
		if(ferror(in_source))
			return 0;
		if(fwrite(buf, 1, count, in_dest) != count)
			return 0;
	}
	return 1;
}


// PURPOSE:	Verifies byte by byte that two files are equal
// REQUIRE:	
// PROMISE: Neither file is not changed by this call
// THROWS:	Nothing

bool IZkFileHelp::AreFilesEqual(const char* in_dest, const char* in_source) throw()
{
	if(!in_source || !in_dest) return 0;

	bool bRet = 0;

	FILE* fFrom = fopen(in_source, "rb");
	FILE* fTo   = fopen(in_dest,   "rb");

	if(fFrom && fTo)
		bRet = AreFilesEqual(fTo, fFrom);

	if(fFrom) fclose(fFrom); fFrom = NULL;
	if(fTo)   fclose(fTo);   fTo   = NULL;

	return bRet;
}




// PURPOSE:	Copies contents of one file into another given their names
// REQUIRE:	
// PROMISE: Returns 1 on success, 0 on error
// THROWS:	Nothing
// EXAMPLE: See Test15
int IZkFileHelp::FileCopy(const char* in_dest, const char* in_source) throw()
{
	if(!in_source || !in_dest) return 0;

	int nRet = 0;

	FILE* fFrom = fopen(in_source, "rb");
	FILE* fTo   = fopen(in_dest,   "wb");

	if(fFrom && fTo)
		nRet = FileCopy(fTo, fFrom);

	if(fFrom) fclose(fFrom); fFrom = NULL;
	if(fTo)   fclose(fTo);   fTo   = NULL;



#ifdef INTERNAL_BUILD
#ifdef ZK_LINUX
	// We were having some trouble with corrupt files on
	// linux, possibly caused by this code - this will
	// try to flag the problem as it occurs - talk to Andrew (see #5771)

	if ( nRet )
	{
		nRet = AreFilesEqual( in_dest, in_source ) ;
		ZKASSERT( nRet ) ;
	}
#endif
#endif

	return nRet;
}


// PURPOSE:	Verify if a file can be writen to, calling this does not change the
//			contents in any way.  Returning "true" implies that the
//			file exists, but does not imply that the file can be "read".
// REQUIRE:	A valid file
// PROMISE: Returns true if the file can be opened in "write" mode
// THROWS:	Nothing
bool IZkFileHelp::IsFileWritable(const char* in_dest) throw()
{
	ZKASSERT( in_dest != NULL ) ;

	if (( in_dest == NULL ) || ( 0 == strlen( in_dest) ))
	{
		ZKASSERT( false ) ;
		return false ;
	}

	// 0 for success
	return !access(in_dest, W_OK);
}

// PURPOSE:	Verify if a given file name exists
// REQUIRE:
// PROMISE: Returns true if file exists (tries to open it to find out)
// EXAMPLE: See Test15
bool IZkFileHelp::FileExists(const char *in_FileName) throw()
{
	// REVIEW - see 6185
	// The file can exist but not have "read" permissions for the current user
	// This function should distinguish correctly - it may required "if/def"
	// platform specific code.
	
    struct stat stbuf = { 0 };
    // 0 == success 1 == file doesn't exist
    return !stat( in_FileName, &stbuf ); 
}


// PURPOSE: Exactly like filecopy() but verifies
//			if the destination file exists beforehand
//			and depeneding on that you specified, proceeds
//			with the copying, or not
// PROMISE: Returns  1 on success, 0 on error [passing true & the file exists also returns 0]
// EXAMPLE: See Test16
int IZkFileHelp::FileCopyWithVerification(
	const char* in_pszExistingFileName, 
	const char* in_pszNewFileName,
	bool        in_bFailIfExists ) throw()
{

	if(in_bFailIfExists && FileExists(in_pszNewFileName) )
		return 0;

	return FileCopy( /*dest*/ in_pszNewFileName, in_pszExistingFileName /* source*/ );
}


// PURPOSE:	Copies contents of one file into another directory using the same file name
// REQUIRE:
// PROMISE: Use posix calls. Returns 1 on success, 0 on error
//			Will append forward slash to given 'in_file' filename
// THROWS:	nothing
// EXAMPLE: See Test15
int IZkFileHelp::FileCopyToDir(const char* in_file, const char* in_dir) throw()
{
	ZKASSERT(in_file);
	ZKASSERT(in_dir);

	struct stat finfo = { 0 };

	if( (0 == stat(in_dir, &finfo)) && (finfo.st_mode & S_IFDIR) )
	{
		char newfile[FILENAME_MAX] = { 0 };
#ifdef WIN32
        char fnbuf[_MAX_FNAME] = { 0 };
        char extbuf[_MAX_EXT] = { 0 };
        _splitpath(in_file, NULL, NULL, fnbuf, extbuf);
		sprintf(newfile, "%s/%s%s", in_dir, fnbuf, extbuf );
#else
        // Hmm I think this is a bug on windows too .. if in_file cotains a path
        // it must be stripped before we can turn it into a new file name.
        // (yes it was, fixed above) --ol
		sprintf(newfile, "%s/%s", in_dir, basename(in_file));
#endif
		return FileCopy( newfile, in_file );
	}
	return 0;
}



// PURPOSE: Returns the length of a file
// REQUIRE: An existing filename
// RETURNS: 0 if inexistant file or unable to obtain information
unsigned long IZkFileHelp::GetFileSize(const char in_file[]) throw()
{
	struct stat finfo = { 0 };
	if(0 == stat(in_file, &finfo))
	{
		return finfo.st_size;
	}
	return 0;
}


// PURPOSE:	Read a text file into a string.
// REQUIRE: The name of the file to read.
//			A string to receive the content of the file.
//			The delimiter wanted in the output string.
//			A delimiter of line to break lines on.
// PROMISE: Returns true if filename was successfully opened for reading
//			Does not clear the output string for you
//			A maximum of str::max_size()-1 characters are read per line
//			Removes duplicates of the set of delimiters you want to have
//			in your output string if they are found before the delimiter
//			on which you break lines; 
//			e.g. readtextfile( "hello\r\nworld\n", str, "\r\n", '\n' ) will return
//			"hello\r\nworld\r\n" instead of "hello\r\r\nworld" if a simple
//			find and replace was done.
//			See the unit test for more examples.
// EXAMPLE: See Test16
bool IZkFileHelp::ReadTextFile(
	const ZKSTD::string& in_rstrFileName, 
	ZKSTD::string& out_rstrFileContent,
	const ZKSTD::string& in_rstrDelimiterToAdd /* = "\n" */,
	const char in_aDelimiter /* = '\n' */ )  throw()
{
	bool bSuccess = false;
	try
	{		
		const unsigned long ulFileSize = GetFileSize( in_rstrFileName.c_str() );
		if( ulFileSize ) 
		{
			ZKSTD::auto_ptr<char> buf( new char[ulFileSize+1] );
			ZKSTD::ifstream in( in_rstrFileName.c_str() );
			if( in ) 
			{
				bSuccess = true;				

				// getline() discards the delimiter
				while( in.getline( buf.get(), ulFileSize+1, in_aDelimiter ) )
				{	
					// get() does not relinquish ownership
					ZKSTD::string strLastLine( buf.get() );
					// find_last_not_of(..) return -1 == npos if it finds nothing.
					const unsigned int posOfAlreadyThereDelim = 
										strLastLine.find_last_not_of( in_rstrDelimiterToAdd ) + 1;
					// if it is not the last charactor of the string.
					// we make sure we do not duplicate eol delimiter.
					if( posOfAlreadyThereDelim != strLastLine.length() )
						out_rstrFileContent += strLastLine.substr( 0, posOfAlreadyThereDelim );
					else
						out_rstrFileContent += strLastLine;
					out_rstrFileContent += in_rstrDelimiterToAdd;
				}
			}			
		}
	}
	catch (...) 
	{
        ZKTRACE0("ReadTextFile -- exception\n");
	}

	return bSuccess;
}

// PURPOSE:	Read a binary file into a string
// REQUIRE: That you free() the memory if it returns true
// PROMISE: Returns true if filename was successfully opened for reading
//			To use malloc() to allocate the memory
//			To return false if malloc returns NULL
// EXAMPLE: See Test16
bool IZkFileHelp::ReadBinaryFile(const char* in_filename, unsigned char** out_buf, size_t& out_siz) throw()
{
	ZKASSERT( out_buf );
	ZKASSERT( in_filename );
	
	FILE* fp = fopen(in_filename, "rb");
	
	if (fp != NULL)
	{
		long	filesize;
		
		fseek(fp, 0, SEEK_END);
		filesize = ftell(fp);
		fseek(fp, 0, SEEK_SET);
		
		if (filesize >= 0)
		{
			void*	buff = malloc(filesize);
			
			if (buff != NULL)
			{
				if (fread(buff, filesize, 1, fp) == 1)
				{
					*out_buf	= (unsigned char *) buff;
					out_siz		= (size_t) filesize;
					
					fclose(fp);
					
					return (true);
				}
				
				free(buff);
			}
		}
		
		fclose(fp);
	}
	
	return (false);
}

bool IZkFileHelp::WriteBinaryFile(
	const char*				in_FileName, 
	const unsigned char *	in_pBuffer, 
	const size_t			in_szBuffer )
{
	FILE* fp = fopen(in_FileName, "wb");
	
	if (fp != NULL)
	{
		if (fwrite(in_pBuffer, in_szBuffer, 1, fp) == 1)
		{
			fclose(fp);
			
			return (true);
		}
		
		fclose(fp);
	}
	
	return (false);
}


// PURPOSE: Returns the line number, and the line itself, in the given file if it 
//			exceeds 'in_nMax' in length
// REQUIRE: An empty 'out_linemap', it does not empty the map for you beforehand
//			size_t is the maximum number of lines possible, entries in map will 
//			just wrap around map if exceeds
// THROWS:  Allocates the file size, so may throw bad_alloc exception if operator new fails
// RETURNS: Total Number of offensive lines found
size_t IZkFileHelp::GetLinesExceedingWidth(
	const char in_file[],
	const size_t in_nMax, 	
	ZKSTD::map<size_t, ZKSTD::string>& out_linemap,
	const char in_delimiter /* = '\n' */ )
{
	ZKTEST(out_linemap.empty());

	ZKSTD::ifstream infile(in_file);

	if(infile)
	{
		unsigned long lSize = GetFileSize(in_file);
		if(lSize)
		{
			ZKSTD::auto_ptr<char>buf( new char[lSize] );
			size_t lineno = 1;
			while(infile.getline(buf.get(), lSize, in_delimiter))
			{
				if( strlen( buf.get() ) > in_nMax)
				{
					out_linemap[lineno] = buf.get();
				}
				lineno++;
			}
			return out_linemap.size();
		}
	}
	return 0;
}

// PURPOSE: Extract the file name from a path.
// REQUIRE: A vector of file name containing either the absolute path
//			or the single name.
//			A vector of string that will receive only the file name.
// PROMISE: The returned vector will be emptied.
//			The returned vector will be the same size as the first parameter.
//			Do not check if the file exists.
void IZkFileHelp::ExtractFileName(
	const ZKSTD::vector<ZKSTD::string>& in_rFileNameVector,
	ZKSTD::vector<ZKSTD::string>& out_rFileNameOnlyVector )
{
#ifdef WIN32

	out_rFileNameOnlyVector.clear();

	ZKSTD::vector<ZKSTD::string>::const_iterator it = in_rFileNameVector.begin();
	for( ; it != in_rFileNameVector.end(); ++it )
	{
		if( (*it).find( "\\" ) != (*it).npos )
		{
			char strFullPath[MAX_PATH] = { 0 };
			if( (*it).find( ":" ) == (*it).npos )	// it is a relative path.
				_makepath( strFullPath, "c:", NULL, (*it).c_str(), NULL );
			else
				strncpy( strFullPath, (*it).c_str(), MAX_PATH );

			char strFileName[_MAX_FNAME] = { 0 };
			_splitpath( strFullPath, NULL, NULL, strFileName, NULL );
			out_rFileNameOnlyVector.push_back( strFileName );
		}
		else
			out_rFileNameOnlyVector.push_back( (*it) );
	}
	
	ZKASSERT( in_rFileNameVector.size() == out_rFileNameOnlyVector.size() );

#else
    out_rFileNameOnlyVector.clear();

    ZKSTD::vector<ZKSTD::string>::const_iterator it = in_rFileNameVector.begin();
    for ( ; it != in_rFileNameVector.end(); ++it )
    {
//        char strFileName[MAX_PATH] = { 0 };
        char *fname = basename((*it).c_str());
        out_rFileNameOnlyVector.push_back( fname );
    }

	ZKASSERT( in_rFileNameVector.size() == out_rFileNameOnlyVector.size() );

#endif
}


