/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef IZkFileHelp_H__
#define IZkFileHelp_H__

#include "ZkSingleton.h"	// for base class
#include "ClientBaseInc.h"	// for CLIENTBASE_DLL
#include <client.h>			// for ZKSTD
#include <clientdebug.h>    // strictly for ZKASSERT

#include <string>

#include <fstream>
#include <memory>
#include <vector>
#include <map>
// C
#include <stdio.h>
// Posix
#include <sys/types.h>
#include <sys/stat.h>



class RZkException;


class CLIENTBASE_DLL IZkFileHelp : public ZkSingleton< IZkFileHelp >
{

public:
	IZkFileHelp();
	virtual ~IZkFileHelp() {}

	static int FolderPicker( const char* in_pszTitle, ZKSTD::string& out_strPickedFile ) ;

	
	// PURPOSE:	Returns true if 'in_strPath' is on the network
	//					
	// REQUIRE: 'in_strPath' be greater or equal than 2 characters in length
	//			
	// PROMISE: Not to throw exceptions
	//  
	// EXAMPLES:
	// IsNetWorkDrive("G:");
	// IsNetWorkDrive("G:\");
	// IsNetWorkDrive("H:");
	// IsNetWorkDrive("H:\");
	// IsNetWorkDrive("\\fileserver\vss");
	static bool IsNetWorkDrive(const ZKSTD::string& in_strPath);


	static bool DoesFileOrDirExist( const char* in_pszFile ) throw(RZkException) ;
	static void CreateDirectory( const char* in_pszDirectory ) throw(RZkException ) ;


    // Functions implemented Xplatform --ol

	// automatic rename of the file with a time stamp
	static ZKSTD::string BackupAndErase( const ZKSTD::string& in_rstrFileName );



// PURPOSE:	Get a string which can represent a valid and unique file name
//			uses the P_tmpdir directory from stdio.h (which may be "\\")
//			so a name may be in the form "\\d9sdjs9"
//			do not call free() on the string
//			Use a buffer at least L_tmpnam * 2 size
//			returns a NULL string if name cannot be generated
//			If you're using a FILE* you may try using tmpfile() in place of this function
//			since the file will get deleted when the app terminates.
// REQUIRE:	
// PROMISE: 
// THROWS:	Nothing
// EXAMPLE: See Test15
    static void GetTempFileName(char* out_pszFileName ) throw();


// PURPOSE:	Verifies byte by byte that two files are equal
// REQUIRE:	
// PROMISE: Neither file is not changed by this call
//			However FILE* are pointing to an undetermined location
//			(Use fseek to return to a known place)
// THROWS:	Nothing

    static bool AreFilesEqual( FILE* in_f1, FILE* in_f2 ) throw();


// PURPOSE:	Copy contents of a file into another given two FILE pointers
//			Returns 0 if either pointer is NULL
//			Returns 1 on success, 0 on error
// REQUIRE:	in_source is open for reading (r or w+)
// PROMISE: Rewinds in_source file pointer to beginning after verifying both pointers aren't NULL
// THROWS:	Nothing
// EXAMPLE: See Test15
    static int FileCopy( FILE* in_dest,  FILE* in_source ) throw();


// PURPOSE:	Verifies byte by byte that two files are equal
// REQUIRE:	
// PROMISE: Neither file is not changed by this call
// THROWS:	Nothing

    static bool AreFilesEqual(const char* in_dest, const char* in_source) throw();


// PURPOSE:	Copies contents of one file into another given their names
// REQUIRE:	
// PROMISE: Returns 1 on success, 0 on error
// THROWS:	Nothing
// EXAMPLE: See Test15
    static int FileCopy(const char* in_dest, const char* in_source) throw();


// PURPOSE:	Verify if a file can be writen to, calling this does not change the
//			contents in any way.  Returning "true" implies that the
//			file exists, but does not imply that the file can be "read".
// REQUIRE:	A valid file
// PROMISE: Returns true if the file can be opened in "write" mode
// THROWS:	Nothing
    static bool IsFileWritable(const char* in_dest) throw();


// PURPOSE:	Verify if a given file name exists
// REQUIRE:
// PROMISE: Returns true if file exists (tries to open it to find out)
// EXAMPLE: See Test15
    static bool FileExists(const char *in_FileName) throw();


// PURPOSE: Exactly like filecopy() but verifies
//			if the destination file exists beforehand
//			and depeneding on that you specified, proceeds
//			with the copying, or not
// PROMISE: Returns  1 on success, 0 on error [passing true & the file exists also returns 0]
// EXAMPLE: See Test16
    static int FileCopyWithVerification(
        const char* in_pszExistingFileName, 
        const char* in_pszNewFileName,
        bool        in_bFailIfExists ) throw();


// PURPOSE:	Copies contents of one file into another directory using the same file name
// REQUIRE:
// PROMISE: Use posix calls. Returns 1 on success, 0 on error
//			Will append forward slash to given 'in_file' filename
// THROWS:	nothing
// EXAMPLE: See Test15
    static int FileCopyToDir(const char* in_file, const char* in_dir) throw();


// PURPOSE: Returns the length of a file
// REQUIRE: An existing filename
// RETURNS: 0 if inexistant file or unable to obtain information
    static unsigned long GetFileSize(const char in_file[]) throw();


// PURPOSE:	Read a text file into a string.
// REQUIRE: The name of the file to read.
//			A string to receive the content of the file.
//			The delimiter wanted in the output string.
//			A delimiter of line to break lines on.
// PROMISE: Returns true if filename was successfully opened for reading
//			Does not clear the output string for you
//			A maximum of str::max_size()-1 characters are read per line
//			Removes duplicates of the set of delimiters you want to have
//			in your output string if they are found before the delimiter
//			on which you break lines; 
//			e.g. readtextfile( "hello\r\nworld\n", str, "\r\n", '\n' ) will return
//			"hello\r\nworld\r\n" instead of "hello\r\r\nworld" if a simple
//			find and replace was done.
//			See the unit test for more examples.
// EXAMPLE: See Test16
    static bool ReadTextFile(
        const ZKSTD::string& in_rstrFileName, 
        ZKSTD::string& out_rstrFileContent,
        const ZKSTD::string& in_rstrDelimiterToAdd = "\n",
        const char in_aDelimiter = '\n' )  throw();


// PURPOSE:	Read a binary file into a string
// REQUIRE: That you free() the memory if it returns true
// PROMISE: Returns true if filename was successfully opened for reading
//			To use malloc() to allocate the memory
//			To return false if malloc returns NULL
// EXAMPLE: See Test16
    static bool ReadBinaryFile(const char* in_filename, unsigned char** out_buf, size_t& out_siz) throw();


    static bool WriteBinaryFile(
        const char*				in_FileName, 
        const unsigned char *	in_pBuffer, 
        const size_t			in_szBuffer );
       

// PURPOSE: Returns the line number, and the line itself, in the given file if it 
//			exceeds 'in_nMax' in length
// REQUIRE: An empty 'out_linemap', it does not empty the map for you beforehand
//			size_t is the maximum number of lines possible, entries in map will 
//			just wrap around map if exceeds
// THROWS:  Allocates the file size, so may throw bad_alloc exception if operator new fails
// RETURNS: Total Number of offensive lines found
    static size_t GetLinesExceedingWidth(
        const char in_file[],
        const size_t in_nMax, 	
        ZKSTD::map<size_t, ZKSTD::string>& out_linemap,
        const char in_delimiter = '\n');


// PURPOSE: Extract the file name from a path.
// REQUIRE: A vector of file name containing either the absolute path
//			or the single name.
//			A vector of string that will receive only the file name.
// PROMISE: The returned vector will be emptied.
//			The returned vector will be the same size as the first parameter.
//			Do not check if the file exists.
    static void ExtractFileName(
        const ZKSTD::vector<ZKSTD::string>& in_rFileNameVector,
        ZKSTD::vector<ZKSTD::string>& out_rFileNameOnlyVector );

protected:

	virtual int onFolderPicker( const char* in_pszTitle, ZKSTD::string& out_strPickedFile ) = 0;
	virtual bool onIsNetWorkDrive(const ZKSTD::string& in_strPath) = 0;

	virtual bool onDoesFileOrDirExist( const char* in_pszFile ) throw(RZkException) = 0;
	virtual void onCreateDirectory( const char* in_pszDirectory ) throw(RZkException )  = 0;

	virtual ZKSTD::string onBackupAndErase( const ZKSTD::string& in_rstrFileName ) = 0;
};

#endif	// IZkFileHelp_H__
