/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef ZKS_INCL_IZKPOPUP
#define ZKS_INCL_IZKPOPUP

/*
  IZkPopup provides a cross-platform api to open basic dialog boxes.
  Dialogs currently supported include:

  MessageBox  : A small window containing a message along with an 'OK' button.
  QuestionBox : A dialog with a Question along with 'Yes' 'No' or 'OK' Cancel'
                buttons.

  Note: This class is derived from ZkSingleton so the a platform specific
        instance can be create at run time. This method is used instead of 
        a define so that we can instanciate a test version. (I guess a define
        may work just as well, if that is the case we can always simplify this
        later without changing the api).

  To Do: Will eventually support a more generic api with the choice of 
         two or three buttons and an icon.
         (you can quote me on this if you need it, EricB )

*/

/*===========================================================================
 * EXTERNAL DECLARATIONS
 *=========================================================================*/

#include <client.h>
#include <zkIntTypes.h>

#include "ClientBaseInc.h"
#include "ConsMessages.h"
#include "MiscMessages.h"


#include "ZkSingleton.h"

#include <string>

/*===========================================================================
 * IZkPopup enums
 *=========================================================================*/

typedef enum {
	ZK_OK		= 0x00000000L,
    ZK_OKCANCEL = 0x00000001L,
    ZK_YESNO	= 0x00000004L
} BtnType;

typedef enum{
	ZK_MESSAGE	= 0,	
	ZK_ERROR	= 0x00000010L,
	ZK_QUESTION = 0x00000020L,
	ZK_WARNING	= 0x00000030L,
	ZK_INFORMATION = 0x00000040L
} IconType;

enum PopupPosition
{
	OVER_FREEDOM,
	OVER_LAST_ACTIVE,
	OVER_CENTER
};

/*===========================================================================
 * Class IZkPopup 
 *=========================================================================*/

class CLIENTBASE_DLL IZkPopup : public ZkSingleton< IZkPopup > {

public: 


    /*----- static member functions -----*/

	// Only possible BtnType is by default ZK_OK
    static void MessageBox( uint32_t msgId, 
							IconType = ZK_MESSAGE, 
							PopupPosition = OVER_FREEDOM,
                            uint32_t titleMsgId = IDS_DEFAULT_TITLE );
    static void MessageBox( const tstring&, 
							IconType = ZK_MESSAGE, 
							PopupPosition = OVER_FREEDOM,
                            const tstring& titleMsg = "Freedom" );

	// BtnType has to be (ZK_OKCANCEL) || (ZK_YESNO)	
    static bool QuestionBox( uint32_t msgId, 
								BtnType = ZK_OKCANCEL, 
								PopupPosition = OVER_FREEDOM );
	static bool QuestionBox( const tstring&, 
								BtnType, 
								IconType = ZK_QUESTION, 
								bool bDefault = true, 
								PopupPosition = OVER_FREEDOM );

	// A MessageCheckBox is a MessageBox-like dialog containing a string (1st param),
	// a checkbox (with text in_checkMsgId) that is by default checked if in_bDefaultChecked,
	// an icon of type 4th param, and an OK button.
	// These members return true if the checkbox is checked and false otherwise.
	static bool MessageCheckBox(	uint32_t	in_msgId,
									uint32_t	in_checkMsgId = IDS_DONT_SHOW_AGAIN_DEFAULT,
									bool		in_bDefaultChecked = false, 
									IconType = ZK_INFORMATION,
									PopupPosition = OVER_FREEDOM );
    static bool MessageCheckBox(	const tstring&,
									uint32_t	in_checkMsgId = IDS_DONT_SHOW_AGAIN_DEFAULT,
									bool		in_bDefaultChecked = false, 
									IconType = ZK_INFORMATION,
									PopupPosition = OVER_FREEDOM );

	// A QuestionCheckBox is like a MessageCheckBox but have 2 buttons instead of one.  It ask
	// a question and offer a suplemental check box to remember the answer (example).  The user
	// of this function must do by itself the work to save the preference.  The return value is
	// true if the first button (left) is pressed and false if the second button (right) is 
	// pressed.  The second parameter is an output value that return the state of the checkbox.
	static bool	QuestionCheckBox(	uint32_t in_msgId,
									bool& out_bChecked,
									uint32_t in_btn1textId = IDS_STOCK_YES,
									uint32_t in_btn2textId = IDS_STOCK_NO,
									uint32_t in_checkMsgId = IDS_DONT_SHOW_AGAIN_DEFAULT,
									bool in_bDefaultChecked = false, 
									IconType = ZK_QUESTION,
									PopupPosition = OVER_FREEDOM );
	static bool	QuestionCheckBox(	const tstring& in_strQuestion,
									bool& out_bChecked,
									uint32_t in_btn1textId = IDS_STOCK_YES,
									uint32_t in_btn2textId = IDS_STOCK_NO,
									uint32_t in_checkMsgId = IDS_DONT_SHOW_AGAIN_DEFAULT,
									bool in_bDefaultChecked = false, 
									IconType = ZK_QUESTION,
									PopupPosition = OVER_FREEDOM );


	// A RequestStringBox is a dialog that contains a title, OK/Cancel boxes, 
	// and prompts for a string, which the user enters into an edit box.
	// The edit box has inital value in_strInitialValue.  The string entered 
	// is returned.  If the user hits Cancel, an empty string is returned.
	static tstring RequestStringBox(const tstring& in_strTitle,
									const tstring& in_strPrompt,
									const tstring& in_strInitialValue = "",
									int in_nMinLength = -1,
									PopupPosition = OVER_FREEDOM );

    /*----- member functions -----*/

    IZkPopup();
    virtual ~IZkPopup();


protected:

	virtual void openMessageBox( uint32_t msgId, IconType, PopupPosition, uint32_t ) = 0;    
	virtual void openMessageBox( const tstring&, IconType, PopupPosition, const tstring& ) = 0;

    virtual bool openQuestionBox( uint32_t msgId, BtnType, PopupPosition ) = 0;
	virtual bool openQuestionBox( const tstring&, BtnType, IconType, bool, PopupPosition ) = 0;

	virtual bool openMessageCheckBox(	uint32_t msgId, 
										uint32_t	in_checkMsgId,
										bool		in_bDefaultChecked, 
										IconType,
										PopupPosition) = 0;    
	virtual bool openMessageCheckBox(	const tstring&,
										uint32_t	in_checkMsgId,
										bool		in_bDefaultChecked, 
										IconType,
										PopupPosition) = 0;

	virtual bool openQuestionCheckBox(	uint32_t in_msgId,
									bool& out_bChecked,
									uint32_t in_btn1textId,
									uint32_t in_btn2textId,
									uint32_t in_checkMsgId,
									bool in_bDefaultChecked, 
									IconType,
									PopupPosition) = 0;
	virtual bool openQuestionCheckBox(	const tstring& in_strQuestion,
									bool& out_bChecked,
									uint32_t in_btn1textId,
									uint32_t in_btn2textId,
									uint32_t in_checkMsgId,
									bool in_bDefaultChecked, 
									IconType,
									PopupPosition) = 0;


	virtual tstring openRequestStringBox(	const tstring& in_strTitle,
											const tstring& in_strPrompt,
											const tstring& in_strInitialValue,
											int in_nMinLength,
											PopupPosition) = 0;


};

#endif	// ZKS_INCL_IZKPOPUP
