// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// ZKConfiguration.cpp 
// Management of the configuration - in general this means controlling
// what network Freedom client connects to

#include "ZKConfiguration.h"
#include "ZKFileConfiguration.h"
#include "IZkRegistry.h"
#include "IZkModuleVer.h"
#if defined(ZK_LINUX)
#   include <zkVerInfo.h>
#endif

#ifdef INTERNAL_BUILD
#include <time.h>
#endif

#define NIQS_SERVER_PREFIX "niqs."            // Niqs server takes the format niqs.<network domain>, eg niqs.pray.freedom.net
#define REGISTRY_NETWORK_ID_KEY "Network"
#define REGISTRY_NETWORK_NAME   "NetworkName"

// Order of these string tables must match the enum in the header file
// We only use hardcoded network names temporarily....
static const char* NetworkNames[] = { "freedom.net",
									"pray.freedom.net",
									"test.freedom.net",
									"alpha.freedom.net",

									// These networks can also be specified in the 
									// general way (RegistrySpecifiedNetwork )
									// but we support these
									// enums to understand v1.1 settings
									"area51.freedom.net",
									"aussie.freedom.net",
									"groomlake.freedom.net",
									"beta.freedom.net",
									"delta.freedom.net",
									"gamma.freedom.net",
									"zeta.freedom.net",
									"beta1.freedom.net",
									"beta2.freedom.net"
								} ;

// This is the page to go to get a trial serial number
static const char* TrialScalper[] = { "http://www.freedom.net/STD/?nw=prod&snid=0",
								  "http://scalper.freedom.net/scalper?nw=pray&snid=trial",
								  "http://cabbage.test.freedom.net/cgi-bin/trial.scalper.cgi",
								  "http://scalper.freedom.net/scalper?nw=alpha&snid=trial",
								  "http://scalper.freedom.net/scalper?nw=area51&snid=trial",
								  "http://scalper.freedom.net/scalper?nw=aussie&snid=trial",
								  "http://scalper.freedom.net/scalper?nw=groomlake&snid=trial",
								  "http://scalper.freedom.net/scalper?nw=beta&snid=trial",
								  "http://scalper.freedom.net/scalper?nw=delta&snid=trial",
								  "http://scalper.freedom.net/scalper?nw=gamma&snid=trial",
								  "http://scalper.freedom.net/scalper?nw=zeta&snid=trial",
								  "http://scalper.freedom.net/scalper?nw=beta1&snid=trial",
								  "http://scalper.freedom.net/scalper?nw=beta2&snid=trial"
									} ;




// This is the page to go to get (purchase if external network) a "full" serial number

static const char* Type1Scalper[] = { "http://www.freedom.net/STD/?nw=prod&snid=500",
								 "http://scalper.freedom.net/scalper?nw=pray&snid=standard",
								 "http://cabbage.test.freedom.net/cgi-bin/std.scalper.cgi",
								  "http://scalper.freedom.net/scalper?nw=alpha&snid=standard",
								  "http://scalper.freedom.net/scalper?nw=area51&snid=standard",
								  "http://scalper.freedom.net/scalper?nw=aussie&snid=standard",
								  "http://scalper.freedom.net/scalper?nw=groomlake&snid=standard",
								  "http://scalper.freedom.net/scalper?nw=beta&snid=standard&user=yes",
								  "http://scalper.freedom.net/scalper?nw=delta&snid=standard",
								  "http://scalper.freedom.net/scalper?nw=gamma&snid=standard",
								  "http://scalper.freedom.net/scalper?nw=zeta&snid=standard",
								  "http://scalper.freedom.net/scalper?nw=beta1&snid=standard",
								  "http://scalper.freedom.net/scalper?nw=beta2&snid=standard"
								} ;

static const char* Type2Scalper[] = { "http://www.freedom.net/STD/?nw=prod&snid=501",
								  "http://scalper.freedom.net/scalper?nw=pray&snid=501",
								  "http://cabbage.test.freedom.net/cgi-bin/type2serial.scalper.cgi",
								  "http://scalper.freedom.net/scalper?nw=alpha&snid=501",
								  "http://scalper.freedom.net/scalper?nw=area51&snid=501",
								  "http://scalper.freedom.net/scalper?nw=aussie&snid=501",
								  "http://scalper.freedom.net/scalper?nw=groomlake&snid=501",
								  "http://scalper.freedom.net/scalper?nw=beta&snid=501",
								  "http://scalper.freedom.net/scalper?nw=delta&snid=501",
								  "http://scalper.freedom.net/scalper?nw=gamma&snid=501",
								  "http://scalper.freedom.net/scalper?nw=zeta&snid=501",
								  "http://scalper.freedom.net/scalper?nw=beta1&snid=501",
								  "http://scalper.freedom.net/scalper?nw=beta2&snid=501"
									} ;

// Free Nyms use seperate Create versus Extend tokens, hence two different scalpers
static const char* Type3CreateScalper[] = { "http://www.freedom.net/STD/?nw=prod&snid=502",
								  "http://scalper.freedom.net/scalper?nw=pray&snid=502",
								  "http://cabbage.test.freedom.net/cgi-bin/type3createserial.scalper.cgi",
								  "http://scalper.freedom.net/scalper?nw=alpha&snid=502",
								  "http://scalper.freedom.net/scalper?nw=area51&snid=502",
								  "http://scalper.freedom.net/scalper?nw=aussie&snid=502",
								  "http://scalper.freedom.net/scalper?nw=groomlake&snid=502",
								  "http://scalper.freedom.net/scalper?nw=beta&snid=502",
								  "http://scalper.freedom.net/scalper?nw=delta&snid=502",
								  "http://scalper.freedom.net/scalper?nw=gamma&snid=502",
								  "http://scalper.freedom.net/scalper?nw=zeta&snid=502",
								  "http://scalper.freedom.net/scalper?nw=beta1&snid=502",
								  "http://scalper.freedom.net/scalper?nw=beta2&snid=502"
									} ;

static const char* Type3ExtendScalper[] = { "http://www.freedom.net/STD/?nw=prod&snid=503",
								  "http://scalper.freedom.net/scalper?nw=pray&snid=503",
								  "http://cabbage.test.freedom.net/cgi-bin/type3extendserial.scalper.cgi",
								  "http://scalper.freedom.net/scalper?nw=alpha&snid=503",
								  "http://scalper.freedom.net/scalper?nw=area51&snid=503",
								  "http://scalper.freedom.net/scalper?nw=aussie&snid=503",
								  "http://scalper.freedom.net/scalper?nw=groomlake&snid=503",
								  "http://scalper.freedom.net/scalper?nw=beta&snid=503",
								  "http://scalper.freedom.net/scalper?nw=delta&snid=503",
								  "http://scalper.freedom.net/scalper?nw=gamma&snid=503",
								  "http://scalper.freedom.net/scalper?nw=zeta&snid=503",
								  "http://scalper.freedom.net/scalper?nw=beta1&snid=503",
								  "http://scalper.freedom.net/scalper?nw=beta2&snid=503"
									} ;

// Clearly the "hardcoding" technique above is hard to maintain
// This single formula for building the scalper URL can support an unlimited
// number of different test networks and serial number types
// The other URLs are left in for convenience and compatibility
static const char TestScalperUrl[] = "http://scalper.freedom.net/scalper?nw=%s&snid=%d" ;

static const int  TrialSerialNumberID = 0 ;

bool ZKConfiguration::s_bUseProductionNetwork = false;

ZKSTD::string ZKConfiguration::GetNiqsAddress()
{
	ZKSTD::string strAddress( NIQS_SERVER_PREFIX ) ;
	strAddress += GetNetworkName() ;
	return strAddress ;
}

unsigned long ZKConfiguration::GetChosenNetwork()
{
	if (s_bUseProductionNetwork)
	{
		return ProductionFreedomNetwork;
	}

	// We use default value if nothing set in registry
	unsigned long dwNetwork = DefaultFreedomNetwork ;	

	IZkRegistry::GetDWORD( REGISTRY_NETWORK_ID_KEY, dwNetwork ) ;

	if ( ( dwNetwork != RegistrySpecifiedNetwork ) &&
	     ( dwNetwork >= BaseFreedomNetworkCount ) ) 
	{
		ZKASSERT( 0 ) ;
		dwNetwork = DefaultFreedomNetwork ;
	}

	return dwNetwork ;
}


void ZKConfiguration::ChooseNetwork( unsigned long in_dwNetwork, const char* in_pszNetworkName )
{	
	if ( in_dwNetwork == RegistrySpecifiedNetwork ) 
	{
		// Network name is remembered inside the registry
		ZKASSERT( in_pszNetworkName != NULL ) ;
		IZkRegistry::SetString( REGISTRY_NETWORK_NAME, in_pszNetworkName ) ;
	}
	else if ( in_dwNetwork >= BaseFreedomNetworkCount ) 
	{
		// Invalid Network Value
		ZKASSERT( 0 ) ;

		in_dwNetwork = DefaultFreedomNetwork ;
	}

	IZkRegistry::SetDWORD( REGISTRY_NETWORK_ID_KEY, in_dwNetwork ) ;
}


// Build a scalper URL for internal networks.
// Depending on the serial number id different types of tokens are available
ZKSTD::string 
ZKConfiguration::BuildInternalScalperUrl( 
	const ZKSTD::string& in_rstrNetwork, 
	const int in_snid )
{

	// For example t1
	ZKSTD::string strJustNetworkName( in_rstrNetwork.substr( 0, in_rstrNetwork.find( '.' )));

	char strUrl[ 256 ] = { 0 } ;
	snprintf( strUrl, sizeof( strUrl), TestScalperUrl, strJustNetworkName.c_str(), in_snid ) ;

	return ZKSTD::string( strUrl ) ;
}





ZKSTD::string 
ZKConfiguration::GetTrialSerialNumberUrl()
{
	unsigned long dwNetwork = GetChosenNetwork() ;

	if ( dwNetwork == RegistrySpecifiedNetwork ) 
	{
		// For registry specified networks we build a standard URL
		// For example t1.dev.zks.net
		ZKSTD::string strNetwork( GetNetworkName() );		

		return BuildInternalScalperUrl( strNetwork, TrialSerialNumberID ) ;
	}
	else if ( dwNetwork == ProductionFreedomNetwork ) 
	{
		// Scalper URL might be in Registry
		ZKSTD::string strScalperInRegistry ;
		IZkRegistry::GetString( "TrialScalper", strScalperInRegistry ) ;

		if ( !strScalperInRegistry.empty() ) 
		{
			return strScalperInRegistry ;
		}
		else 
		{
			return ZKSTD::string( TrialScalper[dwNetwork] ) ;
		}
	}
	else 
	{
		return ZKSTD::string( TrialScalper[dwNetwork] ) ;
	}
}


ZKSTD::string 
ZKConfiguration::GetSerialNumberUrl(SerialNumberTypes serialtype)
{
	unsigned long dwNetwork = GetChosenNetwork() ;
	ZKSTD::string strSerialNumberUrl;

	// For registry specified networks we build a standard URL
	if ( dwNetwork == RegistrySpecifiedNetwork ) 
	{
		// For example t1.dev.zks.net
		ZKSTD::string strNetwork( GetNetworkName() );		

		if (serialtype == NYM_TYPE1_CREATE)
		{
			strSerialNumberUrl =  BuildInternalScalperUrl( strNetwork, NYM_TYPE1_CREATE ) ;
		}
		else if (serialtype == NYM_TYPE1_EXTEND)
		{
			strSerialNumberUrl = BuildInternalScalperUrl( strNetwork, NYM_TYPE1_EXTEND ) ;
		}
		else if (serialtype == NYM_TYPE2_CREATE)
		{
			strSerialNumberUrl = BuildInternalScalperUrl( strNetwork, NYM_TYPE2_CREATE ) ;
		}
		else if (serialtype == NYM_TYPE2_EXTEND)
		{
			strSerialNumberUrl = BuildInternalScalperUrl( strNetwork, NYM_TYPE2_EXTEND ) ;
		}
		else if (serialtype == NYM_TYPE3_CREATE)
		{
			strSerialNumberUrl = BuildInternalScalperUrl( strNetwork, NYM_TYPE3_CREATE ) ;
		}
		else if (serialtype == NYM_TYPE3_EXTEND)
		{
			strSerialNumberUrl = BuildInternalScalperUrl( strNetwork, NYM_TYPE3_EXTEND ) ;
		}
		else
		{
			ZKASSERT(false);
		}
	}

	// For non-registry specified networks we check the
	// registry to see if the url has been provided

	else if ( dwNetwork == ProductionFreedomNetwork ) 
	{
		// Scalper URL might be in Registry
		if (serialtype == NYM_TYPE1_CREATE)
		{
			IZkRegistry::GetString( "NYM_TYPE1_CREATE", strSerialNumberUrl ) ;
		}
		else if (serialtype == NYM_TYPE1_EXTEND)
		{
			IZkRegistry::GetString( "NYM_TYPE1_EXTEND", strSerialNumberUrl ) ;
		}
		else if (serialtype == NYM_TYPE2_CREATE)
		{
			IZkRegistry::GetString( "NYM_TYPE2_CREATE", strSerialNumberUrl ) ;
		}
		else if (serialtype == NYM_TYPE2_EXTEND)
		{
			IZkRegistry::GetString( "NYM_TYPE2_EXTEND", strSerialNumberUrl ) ;
		}
		else if (serialtype == NYM_TYPE3_CREATE)
		{
			IZkRegistry::GetString( "NYM_TYPE3_CREATE", strSerialNumberUrl ) ;
		}
		else if (serialtype == NYM_TYPE3_EXTEND)
		{
			IZkRegistry::GetString( "NYM_TYPE3_EXTEND", strSerialNumberUrl ) ;
		}		
	}

	// We couldn't build the url automatically, or read it from the
	// network, so it is time to check out the hardcoded values
	if ( strSerialNumberUrl.empty() )
	{
		if ( dwNetwork < BaseFreedomNetworkCount )
		{
			// ok, use the hardcoded values
			if (serialtype == NYM_TYPE1_CREATE)
			{
				strSerialNumberUrl = ( Type1Scalper[dwNetwork] ) ;
			}
			else if (serialtype == NYM_TYPE1_EXTEND)
			{
				strSerialNumberUrl = ( Type1Scalper[dwNetwork] ) ;
			}
			else if (serialtype == NYM_TYPE2_CREATE)
			{
				strSerialNumberUrl = ( Type2Scalper[dwNetwork] ) ;
			}
			else if (serialtype == NYM_TYPE2_EXTEND)
			{
				strSerialNumberUrl = ( Type2Scalper[dwNetwork] ) ;
			}
			else if (serialtype == NYM_TYPE3_CREATE)
			{
				strSerialNumberUrl = ( Type3CreateScalper[dwNetwork] ) ;
			}
			else if (serialtype == NYM_TYPE3_EXTEND)
			{
				strSerialNumberUrl = ( Type3ExtendScalper[dwNetwork] ) ;
			}
			else
			{
				ZKASSERT(false);
			}
		}
		else
		{
			ZKASSERT(false);
		}
	}

	return strSerialNumberUrl;
}

ZKSTD::string ZKConfiguration::GetSerialNumberUrl()
{
	// REVIEW: I think this is and should be
	// equiv to calling GetSerialNumberUrl( NYM_TYPE1_CREATE )
		
	unsigned long dwNetwork = GetChosenNetwork() ;

	if ( dwNetwork == RegistrySpecifiedNetwork ) 
	{
		// For registry specified networks we build a standard URL

		// For example t1.dev.zks.net

		ZKSTD::string strNetwork( GetNetworkName() );		

		return BuildInternalScalperUrl( strNetwork, NYM_TYPE1_CREATE ) ;
	}
	else if ( dwNetwork == ProductionFreedomNetwork ) 
	{
		// Scalper URL might be in Registry
		ZKSTD::string strScalperInRegistry ;

		IZkRegistry::GetString( "Scalper", strScalperInRegistry ) ;

		if ( !strScalperInRegistry.empty() ) 
		{
			return strScalperInRegistry ;
		}
		else 
		{
			return ZKSTD::string( Type1Scalper[dwNetwork] ) ;
		}
	}
	else 
	{
		return ZKSTD::string( Type1Scalper[dwNetwork] ) ;
	}

}





ZKSTD::string ZKConfiguration::GetNetworkName()
{
	unsigned long dwNetwork = GetChosenNetwork() ;
	ZKSTD::string strNetworkName ;

	if ( dwNetwork == RegistrySpecifiedNetwork ) 
	{
		if ( !IZkRegistry::GetString( REGISTRY_NETWORK_NAME, strNetworkName) ) 
		{
			ZKASSERT( false ) ; //Bogus configuration
			strNetworkName = "BogusRegistryConfigurationRunFreedomCfgTool" ;
		}
	}
	else 
	{
		strNetworkName = NetworkNames[ dwNetwork ] ;
	}

	ZKASSERT( !strNetworkName.empty() ) ;
	return strNetworkName ;
}





ZKSTD::string 

ZKConfiguration::GetFreedomPlatform()

{

#if defined(WIN32)

	const ZKSTD::string strPlatform( "Windows" ) ;

#elif defined(macintosh)

	const ZKSTD::string strPlatform( "Macintosh" ) ;

#elif defined(ZK_LINUX)

	const ZKSTD::string strPlatform( "Linux" ) ;

#else

	ZKASSERT( false ) ;

	const ZKSTD::string strPlatform( "Unknown Platform" ) ;

#endif



	return strPlatform ;

}



ZKSTD::string 
ZKConfiguration::GetFreedomProductName()
{
	ZKSTD::string strProductName("Zero Knowledge Freedom Client") ;

#if defined(WIN32)

	// We pull this stuff out of the resource 	
	if ( IZkModuleVer::GetFileVersionInfo( ZKFileConfiguration::GetFreedomExeFilePath().c_str() ) ) 
	{
		strProductName = IZkModuleVer::GetValue( "ProductName" ) ;
	}
#elif defined(ZK_LINUX)
    // Get it from clientlinux.h
    strProductName = FREEDOM_PRODUCT_NAME_STRING;
#endif

	return strProductName ;
}



ZKSTD::string 

ZKConfiguration::GetFreedomProductVersion()

{

	ZKSTD::string strProductVersion( "" ) ;



#if defined(WIN32)

	// We pull this stuff out of the resource 	



	if ( IZkModuleVer::GetFileVersionInfo( ZKFileConfiguration::GetFreedomExeFilePath().c_str() ) ) 

	{

		strProductVersion = IZkModuleVer::GetValue( "ProductVersion" ) ;

	}

#elif defined(ZK_LINUX)

    // Get it from clientlinux.h

    strProductVersion = zkVerInfoGetString();

#endif



	return strProductVersion ;

}



#ifdef INTERNAL_BUILD



void ZKConfiguration::DoUnitTest()

{

	ZKTEST( GetNiqsAddress().find( NIQS_SERVER_PREFIX ) != ZKSTD::string::npos );



	ZKTEST( !GetTrialSerialNumberUrl().empty() );

	ZKTEST( !GetSerialNumberUrl().empty() );

	// const unsigned long dwCurrentNet = GetChosenNetwork();

	::srand( time( NULL ) );

	const unsigned long dwRandomNet = static_cast<unsigned long>( ::rand() % BaseFreedomNetworkCount );



	// Grab the existing setting so that we can reset it when

	// we are done

	unsigned long dwNetwork = GetChosenNetwork() ;

	ZKSTD::string strNetwork( GetNetworkName() );		

		

	{

		// Now do some simple config tests

		ChooseNetwork( dwRandomNet, NULL );

		unsigned long dwNewNet = 0;

		IZkRegistry::GetDWORD( REGISTRY_NETWORK_ID_KEY, dwNewNet );

		ZKTEST( dwNewNet == dwRandomNet );

		ZKTEST( GetNetworkName() == NetworkNames[dwRandomNet] ) ;

		ZKSTD::string strMyNet( "MyNetwork" );

		ZKSTD::string strNewNet;

		ChooseNetwork( RegistrySpecifiedNetwork, strMyNet.c_str() );

		IZkRegistry::GetString( REGISTRY_NETWORK_NAME, strNewNet );

		ZKTEST( strMyNet == strNewNet );

		ZKTEST( GetNetworkName() == strNewNet ) ;

		// ChooseNetwork( dwCurrentNet, NULL ); this is causing a ZKASSERT in ChooseNetwork
		ZKTEST( !BuildInternalScalperUrl( GetNetworkName(), TrialSerialNumberID ).empty() );
	}

	ZKTEST( !GetFreedomPlatform().empty() ) ;
	ZKTEST( !GetFreedomProductName().empty() ) ;
	ZKTEST( !GetFreedomProductVersion().empty() ) ;


	// ok, reset the settings, if any
	ChooseNetwork( dwNetwork, strNetwork.c_str() );
}



#endif //INTERNAL_BUILD
