// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// ZKFileConfiguration.cpp


#include "ZKConfiguration.h"
#include "ZKFileConfiguration.h"
#include "IZkRegistry.h"
#include "IZkFileHelp.h"
#include <string_hlp.h>
#include <ZkErrorHandler.h>
#include <IZKErrorReporter.h>
#include <test_registryKeys.h>
#include <sstream>
#include <stdlib.h>

// FIXME - RZkException should be the same for everyone. See PaulL.

#ifdef WIN32
#  include <../win/RZkException.h>
#  include <shlobj.h>
#elif defined(ZK_LINUX)
#include <../linux/RZkException.h>
//#include <glib.h> // needed for g_getenv
#endif

#define TRACE_FILE_CONF	0


#  define NIQS_CACHE_POSTFIX "NIQS.cache"			  // NIQS Cache file is stored as <network domain>.NIQS.Cache
#  define TOKEN_DESC_CACHE_POSTFIX "TokenDesc.cache"  // NIQS Cache file is stored as <network domain>.NIQS.Cache

#ifdef WIN32

#  define NYM_FILE_POSTFIX "freedom.nym"
#  define KEY_CACHE_POSTFIX  "Keys.cache"
#  define ENTROPY_SEED_FILE  "Freedom.seed"

// This might need to be resources for the sake of different
// language versions of Windows 95
#  define APPLICATION_DATA "\\Application Data"

// Sub directory with in the user data directory for Freedom data
// eg. C:\Documents and Settings\Andrew.ZEROK\Application Data\Freedom
// Warning code currently won't work if we want more than one level nested, 
// eg "\\Zero Knowledge\\Freedom" 
#  define FREEDOM_USERDATA_SUBDIR			"\\Freedom\\"
#  define FREEDOM_USERDATA_BACKUP_SUBDIR	"\\Backup\\"

// The subdirectory for the freedom wallet scripts.
// These scripts are used by all users of the machine.
#  define FREEDOM_WALLET_SCRIPTS_SUBDIR	"FormFillerScripts/"

#  define CONSOLE_LOG_DIR "logs\\"

#elif defined (ZK_LINUX)

// Lower case is more standard on unixes
#  define NYM_FILE_POSTFIX "freedom.nym"
#  define KEY_CACHE_POSTFIX  "keys.cache"
#  define ENTROPY_SEED_FILE  "freedom.seed"
#  define APPLICATION_DATA "/Application Data"
#  define FREEDOM_USERDATA_SUBDIR			"/.freedom/"
#  define FREEDOM_USERDATA_BACKUP_SUBDIR	"/.freedom-backup/"
#  define FREEDOM_WALLET_SCRIPTS_SUBDIR	"scripts/" // Maybe we'll need it someday --ol

// The subdirectory in which to find the XML UI description.
#  define FREEDOM_XML_SUBDIR "/xml/"

// The subdirectory in which to find all pixmaps used in the UI
#  define FREEDOM_PIXMAP_SUBDIR "/pixmaps/"

#  define FREEDOM_INSTALL_DIR       "/usr/share/freedom/"
#  define FREEDOM_HELP_DIR          "/doc/"
#  define FREEDOM_HELP_NYM_WIZARD   "html/basi7lbh.htm"
#  define FREEDOM_HELP_SETUP_WIZARD "html/conf5n7b.htm" 
#  define FREEDOM_HELP_ROUTE_WIZARD "html/enha7en9.htm" 

#  define CONSOLE_LOG_DIR "/logs/"

char *ZKFileConfiguration::m_ExeFilePath = NULL;

#endif

// The main files we use for persistence are defined here
const char ZKFileConfiguration::szNymFileBackupName[] = "nym";
const char ZKFileConfiguration::szOldNymFileBackupName[] = "Freedom";
const char ZKFileConfiguration::szBackupExt[] = ".backup";

ZKSTD::string 
ZKFileConfiguration::GetFreedomExeFilePath()
{

// seems there is no such define as _MAX_PATH on Linux (ericb)
// (it's PATH_MAX, or MAX_PATH (which is win32 friendly) if you include clientlinux.h) --ol
#if defined (WIN32)
	char szFullPath[MAX_PATH] = { 0 } ;
	::GetModuleFileName(NULL, szFullPath, _MAX_PATH);
	ZKASSERT( strlen( szFullPath ) > 0 ) ;
	return szFullPath ;
#else
    // Note, since we're using argv0 this could be a relative path, and assumes we never change freedom's
    // cwd. This is the case right now, but this could break in the future if such is not the case --ol
    ZKASSERT("Valid string" && m_ExeFilePath);
    return (m_ExeFilePath);
#endif
}



ZKSTD::string 
ZKFileConfiguration::GetKeyCacheFilePath()
{
	ZKSTD::string strPath( GetFreedomCommonDataDir() ) ;

	unsigned long dwNetwork = ZKConfiguration::GetChosenNetwork() ;

	if ( dwNetwork == ZKConfiguration::ProductionFreedomNetwork ) {
		// Special case - production network doesn't have network name in it
		strPath += KEY_CACHE_POSTFIX ;
	}
	else {
		strPath += ZKConfiguration::GetNetworkName() ;
		strPath += '.' ;
		strPath += KEY_CACHE_POSTFIX ;
	}
	
	ZKASSERT( !strPath.empty() ) ;

	return strPath ;
}

ZKSTD::string 
ZKFileConfiguration::GetNiqsCacheFilePath()
{
	ZKSTD::string strPath( GetFreedomCommonDataDir() ) ;

	unsigned long dwNetwork = ZKConfiguration::GetChosenNetwork() ;

	if ( dwNetwork == ZKConfiguration::ProductionFreedomNetwork ) {
		// Special case - production network doesn't have network name in it
		strPath += NIQS_CACHE_POSTFIX ;
	}
	else {
		strPath += ZKConfiguration::GetNetworkName() ;
		strPath += '.' ;
		strPath += NIQS_CACHE_POSTFIX ;
	}
	
	ZKASSERT( !strPath.empty() ) ;

	return strPath ;
}

ZKSTD::string 
ZKFileConfiguration::GetTokenDescFilePath()
{
	ZKSTD::string strPath( GetFreedomUserDataDir() ) ;

	strPath += ZKConfiguration::GetNetworkName() ;
	strPath += '.' ;
	strPath += TOKEN_DESC_CACHE_POSTFIX ;
	
	ZKASSERT( !strPath.empty() ) ;

	return strPath ;
}


ZKSTD::string 
ZKFileConfiguration::GetEntropySeedFilePath()
{
	ZKSTD::string strPath( GetFreedomCommonDataDir() ) ;

	strPath += ENTROPY_SEED_FILE ;

	ZKASSERT( !strPath.empty() ) ;

	return strPath ;
}

ZKSTD::string 
ZKFileConfiguration::GetFreedomInstallDir()
{
// FIXME - ifdefed out until we replace this on Linux
#if defined (WIN32)

	char szDir[_MAX_DIR];	  // "\xx\xx\"
	char szDrive[_MAX_DRIVE]; // "C:"

	_splitpath(GetFreedomExeFilePath().c_str(), szDrive, szDir, NULL, NULL);

	ZKSTD::ostringstream ostr;
	ostr << szDrive << szDir;
	ZKCTRACE1(TRACE_FILE_CONF, "Freedom installation directory is [%s]\n", ostr.str().c_str());
	return ostr.str();

#elif defined(ZK_LINUX)

    // Check whether or not the FREEDOM_HOME environment variable has been ste
    // if not, use some harcoded value. 
    // FIXME - We will have to sync this function with some eventual 
    //         installation directory thing...
    ZKSTD::string path;
    char* homePath = getenv( "FREEDOM_HOME" ); /* audit: ok */

    if( homePath != 0 ) 
    {
        path = homePath;
    }
    else if( (homePath = getenv( "ZK_FREEDOM_ROOT" )) != 0 ) /* audit: ok */
    {
        path = homePath;
        path += "/client/zkui/freedomclient/linux/";
    }
    else
    {
        path = FREEDOM_INSTALL_DIR;
    }
    
    return path;

#endif
}

// Note: however that there may be problems even with SHGetSpecialFolderLocation
// on Win95 - we may have to special case it.
// From my experiments it is fine even on Windows 95 - it returns
// something like "C:\Windows\Application Data" just like Windows 98

ZKSTD::string 
ZKFileConfiguration::GetFreedomUserDataDir()
{

#if defined (WIN32)

	// SHGetFolderLocation is newer version, not necessarily available on
	// all machines (with v5 of Shell32)

    LPITEMIDLIST ppidl = NULL ;
	HRESULT hr = SHGetSpecialFolderLocation( NULL, CSIDL_APPDATA, &ppidl ) ;

	if ( FAILED( hr ) ) 
	{
		// By default we need to use the installation directory
		// We still need to determine if this can ever happen

		char strBuffer[ MAX_PATH ] ;
		ZKVERIFY( GetWindowsDirectory( strBuffer, MAX_PATH ) ) ;

		ZKSTD::string strWin95Location( strBuffer ) ;

		strWin95Location += APPLICATION_DATA ;

		// Just in case the "Application Data" is named something else in
		// the win95 directory, ie because of language - this is pretty 
		// much a hack.
		IZkFileHelp::CreateDirectory( strWin95Location.c_str() ) ;

		strWin95Location += FREEDOM_USERDATA_SUBDIR ;
		
		return strWin95Location ;
	}

	ZKASSERT( ppidl != NULL ) ;

	char szAppDataPath[ MAX_PATH ] ;

	ZKVERIFY( SHGetPathFromIDList( ppidl, szAppDataPath ) );


	ZKSTD::ostringstream ostr;
	ostr << szAppDataPath << FREEDOM_USERDATA_SUBDIR ;

	ZKCTRACE1(TRACE_FILE_CONF, "Freedom data is located in [%s]\n", ostr.str().c_str());
		
	// Some stupid cleanup

    LPMALLOC pMalloc = NULL;    
	SHGetMalloc(&pMalloc);    
    
	if(ppidl)
    {        
		pMalloc->Free(ppidl);    
	}   
	pMalloc->Release();

	return ostr.str();
#else
    ZKSTD::string userDir;
    char *homeDir = getenv("HOME");    /* audit: ok */
    ZKASSERT(homeDir);
    userDir = homeDir;
    userDir += FREEDOM_USERDATA_SUBDIR;
    return userDir.c_str();
#endif	// defined (WIN32)
}

ZKSTD::string 
ZKFileConfiguration::GetWalletScriptsDir()
{
	ZKSTD::string strPath( GetFreedomCommonDataDir() ) ;

	strPath += FREEDOM_WALLET_SCRIPTS_SUBDIR;

	return (strPath) ;
}

ZKSTD::string
ZKFileConfiguration::GetWalletScriptPath(const ZKSTD::string& in_strDomain)
{
	ZKSTD::string strFileName = ZKS::ExtractDomain(in_strDomain);
	ZKS::MakeLower(strFileName);

	return (GetWalletScriptsDir() + strFileName + ".dat");
}


ZKSTD::string 
ZKFileConfiguration::GetFreedomCommonDataDir()
{
	// Strictly this should be CSIDL_COMMON_APPDATA on windows 2000
	// Similar for Linux, Mac etc.
	
#ifdef WIN32
	// These files should probably go into the user data directory,
	// however at the moment the installer puts in the default instance
	// so this is easier for the moment - clean up as soon as we
	// can alter the installer.
	return GetFreedomInstallDir() ;

#elif defined(ZK_LINUX)
    // We don't really want multiple users to be able
    // to mess with each others key or topology caches (or even look at them, really).
    // This data is not shared on *nix, so it goes inside the home directory.
    return GetFreedomUserDataDir() ;
#endif
}

ZKSTD::string 
ZKFileConfiguration::GetOldFreedomDataFilePath()
{
	ZKSTD::string strCompletePath = GetFreedomInstallDir() ;
	strCompletePath += GetFreedomDataFileName() ;

	return strCompletePath ;
}

ZKSTD::string 
ZKFileConfiguration::GetFreedomDataFilePathProd()
{
	ZKSTD::string strCompletePath = GetFreedomUserDataDir() ;
	strCompletePath += NYM_FILE_POSTFIX ;

	return strCompletePath ;
}

ZKSTD::string 
ZKFileConfiguration::GetFreedomDataFilePath()
{
	ZKSTD::string strCompletePath = GetFreedomUserDataDir() ;
	strCompletePath += GetFreedomDataFileName() ;

	return strCompletePath ;
}

ZKSTD::string 
ZKFileConfiguration::GetFreedomDataFileName()
{
	unsigned long dwNetwork = ZKConfiguration::GetChosenNetwork() ;

	if ( dwNetwork == ZKConfiguration::ProductionFreedomNetwork ) {
		return ZKSTD::string( NYM_FILE_POSTFIX ) ;
	}
	else {
		ZKSTD::string strNetwork( ZKConfiguration::GetNetworkName() );		
		ZKSTD::string strFile( strNetwork.substr( 0, strNetwork.find( '.' ) ) ) ;

		strFile += '.' ;
		strFile += NYM_FILE_POSTFIX ;
		return strFile ;
	}
}


ZKSTD::string 
ZKFileConfiguration::GetOldFreedomDataBackupFileName()
{
	ZKSTD::string strFileName( GetFreedomDataFileName() ) ; // network.freedom.nym
	ZKSTD::string postFix( NYM_FILE_POSTFIX );
	strFileName = strFileName.substr( 0, strFileName.size() - postFix.size() ); // "network." or nothing

	// bug 5016 - not call that file *.freedom.*
	strFileName += szOldNymFileBackupName;
	strFileName += szBackupExt ;
	return strFileName;
} 


ZKSTD::string 
ZKFileConfiguration::GetFreedomDataBackupFileName()
{
	ZKSTD::string strFileName( GetFreedomDataFileName() ) ; // network.freedom.nym
	ZKSTD::string postFix( NYM_FILE_POSTFIX );
	strFileName = strFileName.substr( 0, strFileName.size() - postFix.size() ); // "network." or nothing

	// bug 5016 - not call that file *.freedom.*
	strFileName += szNymFileBackupName;
	strFileName += szBackupExt ;
	return strFileName;
} 


ZKSTD::string
ZKFileConfiguration::GetFreedomAutoBackupFilePath()
{
	ZKSTD::string strDataDir = GetFreedomUserDataDir() ;
//	strDataDir += FREEDOM_USERDATA_BACKUP_SUBDIR ;
	strDataDir += ZKFileConfiguration::GetFreedomDataBackupFileName() ;

	return strDataDir ;
}

ZKSTD::string
ZKFileConfiguration::GetFreedomTestDataDir()
{
	ZKSTD::string strDir = "";

#ifdef WIN32
	if( ! IZkRegistry::GetString( TEST_DATA_DIR_KEY, strDir ) )
	{
		strDir = "C:\\Dev\\Bin\\";
	}
	return strDir;
#else
    // XXX FIXME this is probably not the best place.
    return GetFreedomUserDataDir();
#endif
}


ZKSTD::string
ZKFileConfiguration::GetFreedomConsoleLogDir()
{
    return GetFreedomUserDataDir() + CONSOLE_LOG_DIR; 
}

#ifdef ZK_LINUX

ZKSTD::string 
ZKFileConfiguration::GetFreedomHelpDir()
{
    // FIXME - ugly and hardcoded for now...

    ZKSTD::string path;
    char* homePath = getenv( "FREEDOM_HOME" ); /* audit: ok */

    if( homePath != 0 ) 
    {
        path = homePath;
        path += FREEDOM_HELP_DIR;
    }
    else if( (homePath = getenv( "ZK_FREEDOM_ROOT" )) != 0 ) /* audit: ok */
    {
        path = homePath;
        path += "/client/zkinstall/linux/userdocs/";
    }
    else
    {
        path = FREEDOM_INSTALL_DIR;
        path += FREEDOM_HELP_DIR;
    }
    
    return path;
}

ZKSTD::string 
ZKFileConfiguration::GetFreedomNymWizardHelp()
{
    ZKSTD::string helpFile( ZKFileConfiguration::GetFreedomHelpDir() );
    helpFile += FREEDOM_HELP_NYM_WIZARD;

    return helpFile;
}

ZKSTD::string 
ZKFileConfiguration::GetFreedomRouteWizardHelp()
{
    ZKSTD::string helpFile( ZKFileConfiguration::GetFreedomHelpDir() );
    helpFile += FREEDOM_HELP_ROUTE_WIZARD;

    return helpFile;
}

ZKSTD::string 
ZKFileConfiguration::GetFreedomSetupWizardHelp()
{
    ZKSTD::string helpFile( ZKFileConfiguration::GetFreedomHelpDir() );
    helpFile += FREEDOM_HELP_SETUP_WIZARD;

    return helpFile;
}


ZKSTD::string 
ZKFileConfiguration::GetFreedomXMLDir()
{
    // Check whether or not the FREEDOM_HOME environment variable has been ste
    // if not, use some harcoded value. 
    // FIXME - We will have to sync this function with some eventual 
    //         installation directory thing...
    ZKSTD::string path = GetFreedomInstallDir();

    path += FREEDOM_XML_SUBDIR;
    
    return path;
}

ZKSTD::string 
ZKFileConfiguration::GetFreedomPixmapDir()
{
    // Check whether or not the FREEDOM_HOME environment variable has been ste
    // if not, use some harcoded value. 
    // FIXME - We will have to sync this function with some eventual 
    //         installation directory thing...
    ZKSTD::string path = GetFreedomInstallDir();

    path += FREEDOM_PIXMAP_SUBDIR;
    
    return path;
}

void 
ZKFileConfiguration::SetFreedomExeFilePath(char *argv0)
{
    ZKASSERT("Valid Argument" && argv0);
    m_ExeFilePath = argv0;
}


#endif // ZK_LINUX

void 
ZKFileConfiguration::EnsureFileConfigurationValid() throw (RZkException)
{
	ZKSTD::string strUserDataDir = GetFreedomUserDataDir() ;
	ZKASSERT( !strUserDataDir.empty() ) ;

	try
	{
		if ( !IZkFileHelp::DoesFileOrDirExist( strUserDataDir.c_str() ) )
		{
			// This will throw if it can't succeed
			// Freedom should abort if we can't access the correct directory
			IZkFileHelp::CreateDirectory( strUserDataDir.c_str() ) ;
		}
        // Make sure we have a valid directory where we can put console logs
        if ( !IZkFileHelp::DoesFileOrDirExist( GetFreedomConsoleLogDir().c_str() ) )
            IZkFileHelp::CreateDirectory( GetFreedomConsoleLogDir().c_str() ) ;
 	}
	catch( IZkException& ex )
	{
		throw RZkException (
							"ZKFileConfiguration::EnsureFileConfigurationValid", 
							ZkErrorHandler::GetIZkErrorReporter()->GetErrorString(ex), 
							IDS_FILE_CONFIG_NOT_VALID_WORK,
							strUserDataDir.c_str()
							) ;
	}
}

#ifdef INTERNAL_BUILD

void 
ZKFileConfiguration::DoUnitTest()
{
	ZKTEST( !GetFreedomInstallDir().empty() ) ;
	ZKTEST( !GetFreedomUserDataDir().empty() ) ;
	ZKTEST( !GetFreedomCommonDataDir().empty() ) ;
	ZKTEST( !GetFreedomExeFilePath().empty() ) ;
	ZKTEST( GetKeyCacheFilePath().find( KEY_CACHE_POSTFIX ) != ZKSTD::string::npos );
	ZKTEST( GetNiqsCacheFilePath().find( NIQS_CACHE_POSTFIX ) != ZKSTD::string::npos );
	ZKTEST( !GetFreedomDataFilePath().empty() );
	ZKTEST( GetFreedomDataFileName().find( NYM_FILE_POSTFIX ) != ZKSTD::string::npos );
	ZKTEST( !GetFreedomTestDataDir().empty() );

#if defined (WIN32)
	// Only windows version is concerned with the old location for files
	ZKTEST( !GetOldFreedomDataFilePath().empty() );

	// Part of our upgrade from v1.1 to new version depends on
	// freedom.nym file being in a different directory
	ZKTEST( GetOldFreedomDataFilePath() != GetFreedomDataFilePath() ) ;
#endif

}

#endif //INTERNAL_BUILD

