// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom(R) code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):


//===========================================================================
// EXTERNAL DECLARATIONS
//===========================================================================

#include "ZkConsoleLogger.h"

#include <time.h>

#include <ZkConsoleHint.h>
#include <clientdebug.h>
#include <IWordScan.h>

//===========================================================================
// CLASS ZkConsoleLogger
//===========================================================================

const char * const ZkConsoleLogger::SeverityStrings[] = {
        "VERYIMPORTANT\t",
        "IMPORTANT\t",
        "normal\t",
        "ui\t\t",
        "trivial\t",
        "debug\t\t"
    };

//---------------------------------------------------------------------------
//
ZkConsoleLogger::ZkConsoleLogger( )
    : m_file( NULL )
	, m_pWordScan( NULL )
{
}


//---------------------------------------------------------------------------
//

void ZkConsoleLogger::SetFileName( string logFileName )
{
    CloseFile();
    m_fileName = logFileName;
}


//---------------------------------------------------------------------------
//

void ZkConsoleLogger::SetIWordScan( IWordScan* pWordScan  )
{
    m_pWordScan = pWordScan;
}


//---------------------------------------------------------------------------
//

void ZkConsoleLogger::update( ZkHint* hint )
{
    ZkConsoleHint* cHint = dynamic_cast< ZkConsoleHint* >( hint );
    
    // These calls are not thread safe, but should not collide
    // very often. I'm taking the chance and not putting in a mutex
    // because it's probably not important, would reduce efficiency
    // and can easily be added later.
    if (!m_file && !OpenFile()) return;

    if ( cHint->getMessage().empty() ) {
        // empty message. bail.
        return;
    }

    time_t currentTime;
    time( &currentTime );
    struct tm *currentTimeTm = localtime( &currentTime );
    char timeString[ 100 ] = "";
    strftime( timeString, 100, "%D %H:%M:%S", currentTimeTm );

	ZKSTD::string strMessageSanitized(cHint->getMessage());
	ZKASSERT(m_pWordScan);
	if (m_pWordScan) {
		strMessageSanitized = m_pWordScan->Sanitize(strMessageSanitized);
	}
	if (strMessageSanitized.empty() ) {
        ZKASSERT(false);
        return;
    }
    fprintf( m_file, "%s %s:  %s\n", timeString, 
		SeverityStrings[cHint->getSeverity()], strMessageSanitized.c_str() );
    fflush( m_file );
}


//---------------------------------------------------------------------------
//


bool ZkConsoleLogger::OpenFile()
{
    ZKASSERT( "File Pointer is null" && !m_file );
    ZKASSERT( "We have a file name of some sort" && !m_fileName.empty() );

    m_file = fopen( m_fileName.c_str(), "a" );

    ZKASSERT( "Console log file opened" && m_file );
    return m_file != NULL;
}


//---------------------------------------------------------------------------
//

void ZkConsoleLogger::CloseFile()
{
    if (m_file) {
        fclose(m_file);
        m_file = NULL;
    }
}

//---------------------------------------------------------------------------
//
ZkConsoleLogger::~ZkConsoleLogger( )
{
    if (m_file) CloseFile();
}

        
