/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef ZKS_INCL_ZKSINGLETON
#define ZKS_INCL_ZKSINGLETON

#include <clientdebug.h>

/*===========================================================================
 * CLASS Singleton
 *=========================================================================*/

/*
  This class can be used to implement a class with a single global instance.
  It is not a true blue singleton in that it has to be explicitly instanciated
  before being used. (usually a singleton gets instanciated by the first call
  to "getInstance")

  The reason for this is that singletons usually do not provide any mechanism
  to get destructed. Also the type of singleton instance being created often
  depends on the plaform and it is sinpler to simply create the needed instance 
  right in the main.

  The class provides the initialization of the static instance in the constructor
  and its clearing in the destructor. It also provides the get() function to 
  obtain a reference to instance.
  
  The make a class a singleton, derive from it while setting the template to the
  derived type. i.e.

  class A : public ZkSingleton< A > {};

*/

template< class Base >
class  ZkSingleton {

public:
    
    /*----- static member functions -----*/

    static Base& get()
    {
        // ZkSingletons must be explicitly instantiated
        ZKASSERT( m_instance != 0 );
        
        return *m_instance;
    }

    static bool TestInstance() 
    {
        return ( m_instance != 0 );
    }
      
protected:

    /*----- member functions -----*/

    // Constructing a singleton instantiates the static global instance.
    // This class is not a true blue singleton in that the instance must be
    // explicitly created in the application inittilization phase.
    ZkSingleton()
    { 
        // This assertion should NOT be ignored.
        // It basically means a second instance of a singleton is being created,
        // which will cause a leak. It will also create a very unpleasant
        // situation when the first instance gets out of scope and the second is 
        // still in use.
		ZKASSERT( m_instance == 0 );

        m_instance = static_cast< Base* >( this ); 
    }
    
    // Resets the global pointr to zero before destructing the only instance.
    virtual ~ZkSingleton()
    { 
        m_instance = 0; 
    }
    
private:

    /*----- static data members -----*/

    static Base* m_instance;            
};

template<typename Base>
Base* ZkSingleton<Base>::m_instance = 0;



#endif
