// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "ZkSubject.h"
#include "ZkObserver.h"
#include "MutexGuard.h"
#include <clientdebug.h>
#include <algorithm>


ZkSubject::ZkSubject( )
	:m_error(NULL)
{

    ZKVERIFY( zkErrorNew(&m_error) );

    if( zkMutexNew( m_error, &m_lock ) == ZK_FALSE ) 
    {
        // FIXME - throw something intelligent here....
        throw;
    }
}

//---------------------------------------------------------------------------
//
ZkSubject::~ZkSubject( )
{
    // Make sure we remove ourselves from our observers.
    ObserverVector::iterator iter;

    for( iter = m_observers.begin(); iter != m_observers.end(); ++iter )
    {
        (*iter)->deregisterSubject( this );
    }

    // Then get rid of the vector contents (not that it matters anyway..)
    m_observers.clear();

    // Finally cleanup the mutex.

    if( zkMutexFree( m_error, &m_lock ) == ZK_FALSE ) 
    {
        ZKASSERT(0);
		// Never throw in DTORs
    }
   
	ZKVERIFY( zkErrorFree(&m_error) );
}

//---------------------------------------------------------------------------
//
void ZkSubject::connect( ZkObserver* obs )
{
    ZKASSERT( obs != 0 );

    MutexGuard guard( m_lock );

    // Connecting twice has no effect.
    ObserverVector::iterator iter;
    iter = ZKSTD::find( m_observers.begin(), m_observers.end(), obs );

    if( iter == m_observers.end() ) {
        m_observers.push_back( obs );

        // Now connect back to the observer.
        obs->registerSubject( this );
    }
}

//---------------------------------------------------------------------------
//
void ZkSubject::disconnect( ZkObserver* obs )
{
    ZKASSERT( obs != 0 );

    MutexGuard guard( m_lock );

    // Disconnecting has no effect if not in the list. 
    // (FIXME - could assert, this is not a normal situation)
    ObserverVector::iterator iter;
    iter = ZKSTD::find( m_observers.begin(), m_observers.end(), obs );

    if( iter != m_observers.end() ) {
        m_observers.erase( iter );

        // Now disconnect back to the observer.
        obs->deregisterSubject( this );
    }
}

//---------------------------------------------------------------------------
//
void ZkSubject::notify( ZkHint* hint )
{
    MutexGuard guard( m_lock );

    if( !m_observers.empty() ) {
        ObserverVector::iterator iter;

        for( iter = m_observers.begin(); iter != m_observers.end(); ++iter )
        {
            (*iter)->update( hint );
        }    
    }
}

