/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef ZKS_INCL_ZKSUBJECT
#define ZKS_INCL_ZKSUBJECT

/*===========================================================================
 * EXTERNAL DECLARATIONS
 *=========================================================================*/

#include <vector>
#include "ClientBaseInc.h"
#include <zkThread.h>
#include <clientstl.h>

/*----- forward declare -----*/

class  ZkObserver;
class  ZkHint;

#ifdef WIN32
#pragma warning( disable : 4251 )
#endif

/*
  Any object that needs to notify a few others when its state changes can
  use this class or derive from it.
  A Subject takes care of a list of registered "Observers". Observers connect to 
  the subject and receive "hints" when the subject state changes. Hints are
  anything the subject wishes to send, they are simply any class derived from
  ZkHint.

  API:

  connect:    allows observers to register themselves.
  disconnect: when an observer goes out of scope or is no longer interested in 
              the subject, it simply disconnects.
  notify:     whoever uses the subject call notify to send a hint to all 
              observers
*/


class CLIENTBASE_DLL ZkSubject {

public: 

    /*----- types and enumerations ----*/

    typedef ZKSTD::vector< ZkObserver* > ObserverVector;

    /*----- member functions -----*/
    
    ZkSubject() ;
    virtual ~ZkSubject();

	// Purpose:  Used by ZkObservers to tell the subject they want to receive its hints.
    //           Connecting multiple times will get you only one connection.
	// Promise:  ZkSubject will connect back with the observer.
	// Requires: A valid pointer, other it will just assert on you.
    void connect( ZkObserver* );

	// Purpose:  Used by ZkObservers to remove themselves from the ZkSubject dispatch
    //           list. Disconnecting an unconnected observer has no effect for now.
	// Promise:  ZkSubject will clear itself from ZkObserver's subject list.
    void disconnect( ZkObserver* );

	// Purpose:  Goes through the list of connected observers and sends them the
    //           given hint through the ZkObserver::update() function.
	// Promise:  HOlds the lock while notifying.
    void notify( ZkHint* );
    
	// Purpose:  For tests only.
	// Promise:  I'll kick your butt if you use it.
    const ZkSubject::ObserverVector& getObserverVector( ) const;


private: 

    /*----- member functions -----*/

	// Prohibited funcs - not implemented.
	ZkSubject( const ZkSubject& );
	const ZkSubject& operator=( const ZkSubject& );


    /*----- data members -----*/
	ZkError			m_error;

    ObserverVector	m_observers;
    ZkMutex			m_lock;
    

};

//---------------------------------------------------------------------------
//
inline const ZkSubject::ObserverVector& ZkSubject::getObserverVector( ) const
{
    return m_observers;
}


#endif
