/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/*
	A concrete case of IZkException. Along with the mandatory
	GetErrorCode and GetErrorStringId overrides, this class
	provides accessors to the underlying exception object including
	the DZkError structure. At the bottom of this translation unit
	you will find some macros to throw some simple errors.
*/

#ifndef ZkException_H__
#define ZkException_H__


#include "IZkException.h"
#include "ErrorHandling.h"
#include "DZkError.h"
#include "ExceptionMessages.h"

#include <AutoInstCount.h>

class ZKEXCEPTION_DLL RZkException : public IZkException AUTO_INSTANCE_COUNTING2( RZkException )
{
public:

	// PURPOSE: Store the error information for later retrieval
	//			by someone. Since there can be variable arguments
	//			those N parameters are evaluated in the ctor
	//			and the resultant string is stored in GetFormattedString
	// REQUIRE:	in_MessageId be added to the errorhandling resources
	// PROMISE: To use RZKStringLoader to load strings from resources	
	RZkException(
		const DZkError&			in_err,					
		const ZKSTD::string		in_strCallerFunction,	
		const ZKSTD::string		in_strDebugDescription,	
		const uint32_t			in_MessageId,
        ...);

	// PURPOSE: Store the error information for later retrieval
	//			by someone. Since there can be variable arguments
	//			those N parameters are evaluated in the ctor
	//			and the resultant string is stored in GetFormattedString
	// REQUIRE:	in_MessageId be added to the errorhandling resources
	// PROMISE: To use RZKStringLoader to load strings from resources
	RZkException(
		const ZKSTD::string		in_strCallerFunction,	
		const ZKSTD::string		in_strDebugDescription,	
		const uint32_t			in_MessageId,
        ...);


	virtual ~RZkException();

	// PURPOSE: override. Returns geterrcode from the underlying DZkError object
    virtual uint32_t GetErrorCode() const throw();

	// PURPOSE: override. Returns in_MessageId
    virtual uint32_t GetErrorStringId() const throw();

	const DZkError& GetDZkError() const;

	// PURPOSE: Returns in_strCallerFunction
	virtual const ZKSTD::string& GetCallerFunctionName() const throw();

	// PURPOSE: Returns in_strDescription
	virtual const ZKSTD::string& GetDescription() const throw();

	// PURPOSE: Returns the string as loaded and formatted by the CTOR
	// PROMISE: Length will be max MAXIMUM_STRING_LENGTH
	virtual const ZKSTD::string& GetFormattedString() const throw();

private:
	RZkException();
	RZkException& operator=(const RZkException&);

	// ctor member init	
	const DZkError			m_zkError;
	const ZKSTD::string		m_strCallerFunction;
	const ZKSTD::string		m_strDescription;
	const uint32_t			m_nMessageId;

	ZKSTD::string			m_strFormatted;

	static const unsigned int kMAXIMUM_STRING_LENGTH;
};

// We throw this special error internally if we fail to load 
// just because the passphrase is incorrect.
class /*ZKEXCEPTION_DLL*/ DPassphraseError : public RZkException
{
public:
	DPassphraseError(const char * strFct) 
		: RZkException( strFct, "Passphrase Failure", IDS_CANT_LOAD_PASSWORD_INVALID ) { ; }
} ;


// This is used if a mail operation fails because of an authentication error.
// It can be caught specially, and then a Certificate Update can optionally be 
// performed.
class DMailAuthException : public RZkException
{
public:
	DMailAuthException(const char * strFct) 
		: RZkException( strFct, "Mail Authentication Failure", IDS_MAIL_AUTHENTICATION_FAILED_NO_NYM_NAME ) { ; }
	DMailAuthException(const char * strFct, const char * /*strNymName*/ ) 
		: RZkException( strFct, "Mail Authentication Failure", IDS_MAIL_AUTHENTICATION_FAILED_NO_NYM_NAME ) { ; }
} ;

#define ZKFunctionEntry(fnName) static const char * __function = fnName;
 
// Simplest testing throw call, use for functions that should not fail
#define ZKTestAndThrow(bStatus) { if(!( bStatus)) { throw RZkException(__function,"Unexpected Failure",IDS_UNEXPECTED_FAIL); }}

#define ZKTestAndThrow2(zkerr, bStatus) { if(!( bStatus)) { throw RZkException(zkerr, __function,"Unexpected Failure",IDS_UNEXPECTED_FAIL); }}
 
// Use this to test the return value of a function call that doesn't throw, eg call to a C API that returns int
#define ZKTestAndThrowMsg(bStatus, strProblemDesc, ErrorStringID) { if(!( bStatus)) { throw RZkException(__function,(strProblemDesc),(ErrorStringID)); }}
 
#define ZKThrowRuntime(strProblemDesc, ErrorStringID) throw RZkException(__function,(strProblemDesc),(ErrorStringID));

// Use this to test pointer input arguments
#define ZKTestPointer(pt) { if((pt)==NULL){ throw RZkException(__function,"Null Pointer",IDS_NULL_POINTER);}}
 
#endif // ZkException_H__
