/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#ifndef AUTOINSTCOUNT_H__
#define AUTOINSTCOUNT_H__

#include <clientdefs.h>

#ifdef INTERNAL_BUILD
	#ifndef INSTANCE_COUNTING
		#define INSTANCE_COUNTING
	#endif
#endif

#include <client.h>
#include <lockguard.h>

#ifdef WIN32
#  include <clientwin32.h>	// for ZKCriticalSection
#else
#  include <clientlinux.h>
#endif

#include <time.h>
#include <fstream>

#ifdef INSTANCE_COUNTING
// PURPOSE: Lock used in the instance counter process to assure thread safety.
// REQUIRE: Nothing
// PROMISE: This class implement de singleton pattern.
class RInstanceCounterLock : public ZKCriticalSection
{
public:
	RInstanceCounterLock()
		: ZKCriticalSection()
	{}

	~RInstanceCounterLock()
	{}

	static RInstanceCounterLock& GetGlobalLock()
	{
		static RInstanceCounterLock s_instCounter;
		return s_instCounter;
	}
};

// PURPOSE: Simple helper to keep a global singleton to know if a function
//			has been called at least once.  It is used in the class
//			RInstanceCount<...> to know if we have to write the time in
//			the file. Because RInstanceCount is a template, it is impossible
//			to implement this functionnality within that class even with a static variable.
// REQUIRE: Nothing
// PROMISE: FirstTimeCalled implements the singleton and will always return the same bool.
//			IMPORTANT NOTE: It seems that the singleton is a singleton only in a module.
//			When the instance counter is used in multiple modules, we'll have multiple
//			copy of it.  I couldn't find a way to make it global in all the process' address
//			space.
extern inline bool& FirstTimeCalled()
{
	static bool s_bFirstCall = true;
	return s_bFirstCall;
}

// PURPOSE: Keep track of the number of instance created and destroyed.
//			This class is used as a static data member in the instance counter
//			class : RInstanceCounter.
// REQUIRE: Write access in the directory/file named in s_strFileName.
// PROMISE: Create/append the result in the file s_strFileName with the date
//			and time of the execution.  The string
//			"!! Memory Leak !!" will be appended to each line where there's mem leak.
//			The ouput include the name of the class, the number of instanciation of this
//			class, the number desctruction of this class and the difference.
// NOTE:	I would have liked to use the Lock_Guard in the destructor and in the other
//			functions of RInstanceCount but it is causing crashes (Access Violation)
//			at the end of the app because the singleton (static instance) of ZKCriticalSection
//			can be destroyed BEFORE the global variables and it is used in the destructor of
//			these global variables.  I think this is caused by the templates and could be
//			compiler dependent.
template<typename T>
class RInstanceCount
{
public:
	RInstanceCount<T>()
		: m_createdInstCount( 0 )
		, m_destroyedInstCount( 0 )
	{}

	~RInstanceCount<T>()
	{
		std::ofstream outputFile( s_strFileName, std::ios::out | std::ios::app );

		if( FirstTimeCalled() )
		{
			time_t theUTCTime;
			::time( &theUTCTime );
			tm* theLocalTime = ::localtime( &theUTCTime );

			outputFile << ::asctime( theLocalTime );
#ifdef WIN32
			MEMORY_BASIC_INFORMATION mbi = { 0 };
			::VirtualQuery( this, &mbi, sizeof mbi );
			HINSTANCE hModule = static_cast<HINSTANCE>( mbi.AllocationBase );
			char szModuleName[MAX_PATH] = { 0 };
			::GetModuleFileName( hModule, szModuleName, MAX_PATH );

			outputFile << "\t" << szModuleName << ZKSTD::endl;
#endif // _WIN32
			FirstTimeCalled() = false;
		}

		outputFile << typeid( T ).name() << "\t";
		outputFile << "Number of instance created: " << HowManyCreated();
		outputFile << "\t";
		outputFile << "Number of instance destroyed: " << HowManyDestroyed();
		outputFile << "\t";
		outputFile << "Difference : " << HowManyCreated() - HowManyDestroyed();
		
		if( ( HowManyCreated() - HowManyDestroyed() ) != 0 )
			outputFile << "\t" << "!! Memory Leak !!";
		outputFile << std::endl << std::endl;

		outputFile.close();
	}

	const RInstanceCount<T>& operator++()
	{
		ZKS::Lock_Guard<RInstanceCounterLock> lock( RInstanceCounterLock::GetGlobalLock() );
		++m_createdInstCount;
		return *this;
	}

	const RInstanceCount<T>& operator--()
	{
		ZKS::Lock_Guard<RInstanceCounterLock> lock( RInstanceCounterLock::GetGlobalLock() );
		++m_destroyedInstCount;
		return *this;
	}

	size_t HowManyCreated()
	{
		return m_createdInstCount;
	}

	size_t HowManyDestroyed()
	{
		return m_destroyedInstCount;
	}

private:
	size_t m_createdInstCount;
	size_t m_destroyedInstCount;

	static char s_strFileName[MAX_PATH];
};

#ifdef WIN32
template<typename T>
char RInstanceCount<T>::s_strFileName[] = { "c:\\memleak.txt" };
#else
template<typename T>
char RInstanceCount<T>::s_strFileName[] = { "memleak.txt" };
#endif

// PURPOSE: Class from which every class that want to be instance counted
//			should inherits.
// REQUIRE: Privately inherit from this class to use instance counting.
// PROMISE: Will count the number of instance of your inherited class
//			you created and you destroyed.
template<typename T>
class RInstanceCounter
{
public:
	RInstanceCounter<T>()
	{
		++s_instCount;
	}

	RInstanceCounter<T>( const RInstanceCounter<T>& )
	{
		++s_instCount;
	}

	~RInstanceCounter<T>()
	{
		--s_instCount;
	}

	static size_t HowMany()
	{
		return s_instCount.HowManyCreated() - s_instCount.HowManyDestroyed();
	}

private:
//	RInstanceCounter<T>& operator=( const RInstanceCounter<T>& );

	static RInstanceCount<T> s_instCount;
};

template<typename T>
RInstanceCount<T> RInstanceCounter<T>::s_instCount;


#define AUTO_INSTANCE_COUNTING1( classname ) \
	: private RInstanceCounter<classname>

#define AUTO_INSTANCE_COUNTING2( classname ) \
	, private RInstanceCounter<classname>

#define AUTO_INSTANCE_COPY_CTOR1( classname, copy ) \
	: RInstanceCounter<classname>((const RInstanceCounter<classname>&)copy)

#define AUTO_INSTANCE_COPY_CTOR2( classname, copy ) \
	, RInstanceCounter<classname>((const RInstanceCounter<classname>&)copy)

#else // INSTANCE_COUNTING

#define AUTO_INSTANCE_COUNTING1( classname ) 
#define AUTO_INSTANCE_COUNTING2( classname ) 
#define AUTO_INSTANCE_COPY_CTOR1( classname, copy )
#define AUTO_INSTANCE_COPY_CTOR2( classname, copy ) 

#endif // INSTANCE_COUNTING



#endif // AUTOINSTCOUNT_H__
