// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom(R) code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):


// See Test34 for usage examples.

#ifndef CIRACULARVECTOR_H__
#define CIRACULARVECTOR_H__

#include <client.h>		// For ZKSTD
#include <math.h>		// For ::ceil

namespace ZKS
{

// This class implements a fixed size queue as a circular array.
// It's usage is mainly to be used as an effective FIFO list by avoiding
// allocation and deallocation on each push-back/pop-head call.
// The caller should provide the most likely maximum size for the "list"
// but the class will reallocate memory with a factor of GROWING_FACTOR if needed.
// The caller can ask for a reallocation whenever needed with "Reserve".

// The data member m_uiBeginIndex is the index of the first element of the "list"
// if m_uiNumberOfElem greater than 0.  The data member m_uiEndIndex is the index
// where the next element will be put.  If m_uiBeginIndex == m_uiEndIndex, the list is
// either empty or full (depending on m_uiNumberOfElem).

// The vector used to hold the data is dynamically allocated to allow easier dynamic growth

#define GROWING_FACTOR 1.5			// This must be > 1.
#define DEFAULT_INIT_SIZE 10		// This must be > 0.

template <typename type>
class CircularVector
{
public:
	CircularVector( unsigned int in_uiInitialSize = DEFAULT_INIT_SIZE )
		: m_pVector( new ZKSTD::vector<type> )
		, m_uiBeginIndex( 0 )
		, m_uiEndIndex( 0 )
		, m_uiNumberOfElem( 0 )
	{
		ZKASSERT( m_pVector );
		m_pVector->reserve( in_uiInitialSize );
	}

	~CircularVector()
	{
		SAFEDELETE( m_pVector );
	}

	// Return true if the CircularVector is empty.
	bool IsEmpty() const
	{
		return m_uiNumberOfElem == 0;
	}

	// Return the number of element contained.
	size_t Size() const
	{
		return m_uiNumberOfElem;
	}

	// Return the capacity of the circular vector.
	size_t Capacity() const
	{
		ZKASSERT( m_pVector );
		return m_pVector->capacity();
	}

	// Resize the cicular vector to ensure in_size element can fit in.
	// It won't reallocate a smaller vector.
	void Reserve( unsigned int in_size )
	{
		if( !m_pVector )
		{
			ZKASSERT( false );
			return;
		}

		if( in_size > m_pVector->capacity() )
		{
			ZKSTD::vector<type>* pNewVec = new ZKSTD::vector<type>;
			ZKASSERT( pNewVec );

			pNewVec->reserve( in_size );

			unsigned int uiNumberOfElem = m_uiNumberOfElem; 
			for( unsigned int i = 0; i < uiNumberOfElem; ++i )
				(*pNewVec)[i] = PopHead();

			m_uiBeginIndex = 0;
			m_uiEndIndex = m_uiNumberOfElem = uiNumberOfElem;

			SAFEDELETE( m_pVector );
			m_pVector = pNewVec;
		}
	}

	// Get the element at the head of the list and removes it.
	// Will cycle the begin index to the beginning of the vector if it gets at the end of it.
	// If the CircularVector is empty, we return a default, unitialized value.
	const type& PopHead()
	{
		if( m_uiNumberOfElem == 0 )
		{
			ZKASSERT( !"CircularVector is empty.  Returning a default value." );
			return s_defaultVal;
		}

		// An unsigned int will get to max_int if --0;
		ZKVERIFY( --m_uiNumberOfElem <= m_pVector->capacity() );

		const type& retVal = (*m_pVector)[m_uiBeginIndex];
		
		if( ++m_uiBeginIndex == m_pVector->capacity() )
			m_uiBeginIndex = 0;
		
		return retVal;
	}

	// Add an element at the end of the "list".
	// It will make the internal vector grow by a factor of GROWING_FACTOR if
	// we are running out of space.
	// Will cycle the end index to the beginning of the vector if it gets at the end of it.
	void PushBack( const type& in_rElem )
	{
		if( m_uiNumberOfElem == m_pVector->capacity() )
		{
			ZKTRACE0( "CircularVector is full.  Re-allocating." );
			ZKASSERT( GROWING_FACTOR > 1.0 );
			Reserve( static_cast<unsigned int>( ::ceil( m_pVector->capacity() * GROWING_FACTOR ) ) );
		}

		++m_uiNumberOfElem;
		ZKASSERT( m_uiNumberOfElem <= m_pVector->capacity() );

		(*m_pVector)[m_uiEndIndex] = in_rElem;

		if( ++m_uiEndIndex == m_pVector->capacity() )
			m_uiEndIndex = 0;
	}


private:
	CircularVector( const CircularVector& );
	CircularVector& operator=( const CircularVector& );

	ZKSTD::vector<type>* m_pVector;
	unsigned int m_uiBeginIndex;
	unsigned int m_uiEndIndex;
	unsigned int m_uiNumberOfElem;

	static type s_defaultVal;
};

template <typename type>
type CircularVector<type>::s_defaultVal;


} // namespace ZKS

#endif // CIRACULARVECTOR_H__
