// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// DynamicLibrary.h: Wraps dynamic linking capabilities
//                   of Win32, Solaris, and Linux.
// Taken from C/C++ User's Journal

#ifndef DYNAMICLIBRARY_H_INCLUDED
#define DYNAMICLIBRARY_H_INCLUDED

#include <vector>

#ifdef ZK_LINUX
#include <dlfcn.h>
#endif

namespace ZKS
{

typedef void(*DLPROC)();

class RDynamicLibrary  
{
public:

	// PURPOSE: Default constructor.
	// REQUIRE: Nothing
	// PROMISE: Nothing
	RDynamicLibrary()
		: m_libHandle( 0 ) {}

	// PURPOSE: Standard constructor.
	// REQUIRE: Nothing
	// PROMISE: Throw if cannot load the lib.
	RDynamicLibrary( const char* in_strLibName ) throw ( ZKSTD::exception );

	// PURPOSE: Desctructor.
	// REQUIRE: Nothing
	// PROMISE: Nothing
	virtual ~RDynamicLibrary();

	// PURPOSE: Load the library.
	// REQUIRE: Nothing
	// PROMISE: Will load a lib if no other lib has been loaded before.
	bool Load( const char* in_strLibName );
	
	// PURPOSE: Unload the lib.
	// REQUIRE: Nothing
	// PROMISE: Unload the lib (if any loaded) and clear the proc cache.
	bool Unload();

	// PURPOSE: Check if a library is loaded.
	// REQUIRE: Nothing
	// PROMISE: Return true if a lib is loaded.
	bool IsLoaded() const { return ( m_libHandle != 0 ); }

	// PURPOSE: Get the address of a procedure implemented in the lib.
	// REQUIRE: A valid proc name.
	// PROMISE: A valid pointer to the procedure or NULL.
	DLPROC GetProcAddr( const char* in_strProcName ) const;
	
	// PURPOSE: Get the address of a procedure implemented in the lib with
	//			the caching feature.
	// REQUIRE: A valid name to the procedure and an ID for that proc.
	// PROMISE: A valid pointer to the proc or NULL.
	//			If the ID is new, it will be initialized to be sucessfully
	//			used in later call.
	DLPROC GetProcAddrCached( const char* in_strProcName, unsigned int in_uiProcId );

#ifdef _WIN32
	// PURPOSE: Get the address of a procedure implemented in the lib.
	// REQUIRE: A valid ordinal value inside the dll.
	// PROMISE: A valid pointer to the procedure represented by the ordinal
	//			parameter.
	DLPROC GetProcAddr( unsigned int in_uiOrdinal ) const;

	// PURPOSE: Get the address of a procedure implemented in the lib with
	//			the caching feature.
	// REQUIRE: A valid ordinal value to the procedure and an ID for that proc.
	// PROMISE: A valid pointer to the proc or NULL.
	//			If the ID is new, it will be initialized to be sucessfully
	//			used in later call.
	DLPROC GetProcAddrCached( unsigned int in_uiOrdinal );
#endif

private:
	struct RCacheInfo
	{
		DLPROC	procAddr;
		bool	testFlag;

		RCacheInfo() 
			: procAddr( 0 )
			, testFlag( false )
		{}

		~RCacheInfo()
		{}
	};

private:
	RDynamicLibrary( const RDynamicLibrary& );
	RDynamicLibrary& operator=( const RDynamicLibrary& );

#ifdef _WIN32
	HINSTANCE m_libHandle;
#else
	void* m_libHandle;
#endif

	ZKSTD::vector<RCacheInfo> m_procCache;
};

inline RDynamicLibrary::RDynamicLibrary( const char* in_strLibName )
	: m_libHandle( 0 )
{
	if ( !Load( in_strLibName ) ) 
		throw ZKSTD::exception();
}

inline RDynamicLibrary::~RDynamicLibrary()
{
	Unload();
}

inline bool RDynamicLibrary::Load( const char* in_strLibName )
{
	if( in_strLibName != NULL && ::strlen( in_strLibName ) > 0 )
	{
		if( m_libHandle == NULL )
		{
#ifdef _WIN32
			m_libHandle = ::LoadLibrary( in_strLibName );
#else
			m_libHandle = ::dlopen( in_strLibName, RTLD_LAZY );
#endif
		}
	}

	return ( m_libHandle != NULL );
}

inline bool RDynamicLibrary::Unload()
{
	if( m_libHandle != NULL )
	{
#ifdef _WIN32
		::FreeLibrary( m_libHandle );
#else
		::dlclose( m_libHandle );
#endif
	}

	m_libHandle = NULL;

	m_procCache.clear();

	return true;
}

inline DLPROC RDynamicLibrary::GetProcAddr( const char* in_strProcName ) const
{
	DLPROC pProc = NULL;

	if( m_libHandle != NULL )
	{
#ifdef _WIN32
		pProc = reinterpret_cast<DLPROC>( ::GetProcAddress( m_libHandle, in_strProcName ) );
#else
		pProc = reinterpret_cast<DLPROC>( ::dlsym( m_libHandle, in_strProcName ) );
#endif
	}

	return pProc;
}

inline DLPROC RDynamicLibrary::GetProcAddrCached( const char* in_strProcName, unsigned int in_uiProcId )
{
	DLPROC pProc = NULL;

	if( m_libHandle != NULL )
	{
		if( ( in_uiProcId >= m_procCache.size() ) || ( m_procCache.size() < 1 ) )
			m_procCache.resize( in_uiProcId + 1 );

		RCacheInfo& rCacheInfo = m_procCache[in_uiProcId];

		if( rCacheInfo.procAddr == NULL && !rCacheInfo.testFlag )
		{
			rCacheInfo.testFlag = true;
			rCacheInfo.procAddr = GetProcAddr( in_strProcName );
		}

		pProc = rCacheInfo.procAddr;
	}

	return pProc;
}

#ifdef _WIN32
inline DLPROC RDynamicLibrary::GetProcAddr( unsigned int in_uiOrdinal ) const
{
	DLPROC pProc = NULL;

	if (m_libHandle != NULL)
		pProc = reinterpret_cast<DLPROC>( ::GetProcAddress( m_libHandle,
															MAKEINTRESOURCE( in_uiOrdinal ) ) );

	return pProc;
}

inline DLPROC RDynamicLibrary::GetProcAddrCached( unsigned int in_uiOrdinal )
{
	return GetProcAddrCached( MAKEINTRESOURCE( in_uiOrdinal ), in_uiOrdinal );
}

#endif //_WIN_32
}

#endif // DYNAMICLIBRARY_H_INCLUDED
