/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

// ZKInverseSemaphore.h

/*	The ZKInverseSemaphore class provides a means for waiting until multiple events
	are signalled.

	It is typically used when one wants to spawn multiple threads and know when all threads
	are finished.
	

	Here is an example of how to use the ZKInverseSemaphore:

	void ThreadFcn(void* pVoid)
	{
		ZKInverseSemaphore* pInvSem = (ZKInverseSemaphore*) pVoid;
		ZKInverseSemaphoreLock gate(*pInvSem);
		// some code
	}

	ZKInverseSemaphore zkInverseSem;
	_beginthread(ThreadFcn, 0, (void*) &zkInverseSem);
	_beginthread(ThreadFcn, 0, (void*) &zkInverseSem);

	// Must sleep here for some small amount of time.
	
	zkInverseSem.WaitUntilCountIsZero(); // Call intelligently blocks until both threads
	are complete.

	Also see Test29
*/


//#pragma warning (disable: 4786)
//#pragma warning (disable: 4251)

#if !defined(_INVERSE_SEMAPHORE_H_)
#define _INVERSE_SEMAPHORE_H_


#include <DZkError.h>
#ifdef WIN32
#include <clientwin32.h>	// for ZKCriticalSection (which should shortly be made xp)
#else 
#include <clientlinux.h>
#endif

#include <zkThread.h>
#include "lockguard.h"


class ZKInverseSemaphore AUTO_INSTANCE_COUNTING1(ZKInverseSemaphore)
{
	friend class ZKInverseSemaphoreLock;

public:
	ZKInverseSemaphore();
	~ZKInverseSemaphore();
	
	// timeout in milliseconds
	// returns false if timeout is reached.  True otherwise
	bool WaitUntilCountIsZero(unsigned long timeout = ZK_INFINITE);

private:
	// Prohibited - not defined
	ZKInverseSemaphore(const ZKInverseSemaphore& control);			
	ZKInverseSemaphore& operator=(const ZKInverseSemaphore& rhs);
	
	void IncrementCount();
	void DecrementCount();

	unsigned long			m_dwCount;
	ZkEvent					m_noCountEvent;
	ZKCriticalSection		m_cs;
	DZkError				m_err;
};

//////////////////////////////////////////////////////////////

class ZKInverseSemaphoreLock AUTO_INSTANCE_COUNTING1(ZKInverseSemaphoreLock)
{
	
public:
	ZKInverseSemaphoreLock(ZKInverseSemaphore& in_inverseSemaphore)	
		: m_inverseSemaphore(in_inverseSemaphore)
		{ m_inverseSemaphore.IncrementCount(); };

	~ZKInverseSemaphoreLock()
		{ m_inverseSemaphore.DecrementCount(); };

private:
	// Prohibited - not defined
	ZKInverseSemaphoreLock();
	ZKInverseSemaphoreLock(const ZKInverseSemaphoreLock& control);			
	ZKInverseSemaphoreLock& operator=(const ZKInverseSemaphoreLock& rhs);

	ZKInverseSemaphore& m_inverseSemaphore;
};


inline ZKInverseSemaphore::ZKInverseSemaphore()
	: m_dwCount(0)
{
	// Create the "no clients" signal event object
	ZKVERIFY( zkEventNew(	&m_err,
							&m_noCountEvent,	
							true ));	// Manual reset
						
	ZKVERIFY( zkEventSet( &m_err, m_noCountEvent ));
}

inline ZKInverseSemaphore::~ZKInverseSemaphore()
{
	ZKVERIFY( zkEventFree( &m_err, &m_noCountEvent));
}

inline void ZKInverseSemaphore::IncrementCount()
{
	// Not thread safe (Reset/Set Event)
	ZKS::Lock_Guard<ZKCriticalSection> gate(m_cs);

	m_dwCount++;
	ZKVERIFY( zkEventReset( &m_err, m_noCountEvent ));
}

inline void ZKInverseSemaphore::DecrementCount()
{
	// Not thread safe (Reset/Set Event)
	ZKS::Lock_Guard<ZKCriticalSection> gate(m_cs);

	ZKASSERT( m_dwCount > 0 );
	
	if (m_dwCount > 0) 
	{
		m_dwCount--;
	}
	if (m_dwCount < 1) 
	{
		ZKVERIFY( zkEventSet( &m_err, m_noCountEvent));
	}
}

inline bool ZKInverseSemaphore::WaitUntilCountIsZero(unsigned long timeout)
{
	return (ZK_TRUE == zkEventWait( &m_err, m_noCountEvent, timeout));
}


#endif //_INVERSE_SEMAPHORE_H_
