/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef _ZKStringSearch_H__
#define _ZKStringSearch_H__

// The header file is not properly defined.
//
// -You can't include it in the same project
// more than once because of the "lowervec" definition
// that it includes (never include definitions in header files!
// -Everything is inline


#include <vector>
#include <string>

#include <climits>
#include <clientdebug.h> // for ZKASSERT

class ZKStringSearch
{
public:
	// This function is not thread safe, you must ensure it is - eg
	// by not using it as a static and moving global variables in as members
	// PURPOSE: Find in a buffer the words that are in a set of word.
	// REQUIRE: This function is not thread safe, you must ensure it is - eg
	//			by not using it as a static and moving global variables in as members
	//			Param: Buffer in which to scan.  Size of this buffer.
	//			Vector of words to look for.  Vector to receive words found.
	// PROMISE: Return true if at least a word has been found.
	static bool ScanBufferForWords(	unsigned char* in_pText, 
									const size_t in_dwSize, 
									const ZKSTD::vector<ZKSTD::string>& in_vecstrOffensiveWords, 
									ZKSTD::vector<ZKSTD::string>& out_vecstrOffensiveWordsFound );

private:
	// Initializes the bahamas
	static void bmha_init( const char* in_pszPattern );

	// Searches the bahamas
	static char* bmha_search( const char* in_pszString, const int in_stringlen );
};



//  Boyer-Moore-Horspool in_pszPattern match
//  Case-insensitive with accented character translation
//  limitation: pattern length + subject length must be less than 32767


typedef unsigned char uchar;

#define LOWER_ACCENTED_CHARS

unsigned char lowervec[UCHAR_MAX+1] = {                     /* rdg 10/93 */
  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
 32,'!','"','#','$','%','&','\'','(',')','*','+',',','-','.','/',
'0','1','2','3','4','5','6','7','8','9',':',';','<','=','>','?',
'@','a','b','c','d','e','f','g','h','i','j','k','l','m','n','o',
'p','q','r','s','t','u','v','w','x','y','z','[','\\',']','^','_',
'`','a','b','c','d','e','f','g','h','i','j','k','l','m','n','o',
'p','q','r','s','t','u','v','w','x','y','z','{','|','}','~',127,
#ifdef LOWER_ACCENTED_CHARS
'c','u','e','a','a','a','a','c','e','e','e','i','i','i','a','a',
'e',145,146,'o','o','o','u','u','y','o','u',155,156,157,158,159,
'a','i','o','u','n','n',166,167,168,169,170,171,172,173,174,175,
#else
128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,
144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,159,
160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,
#endif
176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191,
192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,
208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,
224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,
240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255,
};

#define lowerc(c) lowervec[(uchar)(c)]

#define LARGE 32767

static int patlen;
static int skip[UCHAR_MAX+1];                               /* rdg 10/93 */
static int skip2;
static uchar *pat;

inline bool ZKStringSearch::ScanBufferForWords(
	unsigned char* in_pText, 
	const size_t in_dwSize, 
	const ZKSTD::vector<ZKSTD::string>& in_vecstrOffensiveWords, 
	ZKSTD::vector<ZKSTD::string>& out_vecstrOffensiveWordsFound)
{

// We are using straight C++ here since we don't want to import
// os specific assertion mechanisms
#ifndef NDEBUG
	ZKASSERT( in_pText );
#endif

	bool bReturnValue = false;
	
	// All trigger information is global.
	// In the future we will probably have trigger information
	// associated with the Active Nym and the User Nym and combine both	

	out_vecstrOffensiveWordsFound.clear();

	size_t chunkSize = 0;
	unsigned char* pChunk = NULL;
	size_t totalScannedSoFar = 0;

	do
	{	
		totalScannedSoFar = totalScannedSoFar + chunkSize;
		pChunk = in_pText + totalScannedSoFar;

		// Determine how large this chunk will be
		int nRemainder = in_dwSize - totalScannedSoFar;

		if( nRemainder < 0x7000) 
			chunkSize = nRemainder;
		else 
			chunkSize = 0x7000; // 28 672

		size_t counter = 0;

		// Extend the chunk a bit to the end of the
		// next word, to avoid missing words that
		// extend over boundaries
		while( ( chunkSize + totalScannedSoFar < in_dwSize )	// End of the total buffer
				&& ( pChunk[chunkSize] != ' ' ) 					// end of a word
				&& ( counter < 512 ) )							// Stop in case of binary data
		{	
			++counter;
			++chunkSize;
		}

		for( ZKSTD::vector<ZKSTD::string>::const_iterator theIterator = in_vecstrOffensiveWords.begin();
				theIterator != in_vecstrOffensiveWords.end();
				++theIterator )
		{
			bmha_init( (*theIterator).c_str() );
			char* pszWordfind = NULL;
			pszWordfind = bmha_search((const char*)pChunk, chunkSize);
			
			if( pszWordfind != NULL )
			{
				ZKSTD::vector<ZKSTD::string>::const_iterator it;
				for( it = out_vecstrOffensiveWordsFound.begin(); 
						it != out_vecstrOffensiveWordsFound.end();
						++it )
				{
					if( *it == *theIterator )
						break;
				}

				if( it == out_vecstrOffensiveWordsFound.end() )
				{
					// Didn't already add this.
					out_vecstrOffensiveWordsFound.push_back( *theIterator );
				}		
				bReturnValue = true;
			}
		}

	}while( (totalScannedSoFar + chunkSize) < in_dwSize );


	return bReturnValue;
}

inline void ZKStringSearch::bmha_init( const char* in_pszPattern )
{
      int i = 0, j = 0;
      pat = (uchar*)in_pszPattern;
      patlen = strlen( in_pszPattern );

      for( i = 0; i <= UCHAR_MAX; ++i )
	  {
            skip[i] = patlen;
            for( j = patlen - 1; j >= 0; --j )
			{
                  if ( lowerc( i ) == lowerc( pat[j] ) )
                        break;
            }
            
			if( j >= 0 )
                  skip[i] = patlen - j - 1;
            if( j == patlen - 1 )
                  skip[i] = LARGE;
      }

      skip2 = patlen;
      
	  for( i = 0; i < patlen - 1; ++i )
	  {
            if ( lowerc( pat[i] ) == lowerc( pat[patlen - 1] ) )
                  skip2 = patlen - i - 1;
      }

}

inline char* ZKStringSearch::bmha_search( const char* in_pszString, const int in_stringlen )
{
      int i = 0, j = 0;
      char* s = NULL;

      i = patlen - 1 - in_stringlen;
      if( i >= 0 )
            return NULL;

      in_pszString += in_stringlen;
      for( ;; )
	  {
            while( ( i += skip[((uchar*)in_pszString)[i]] ) < 0 )
                  ;                           /* mighty fast inner loop */
            if( i < (LARGE - in_stringlen) )
                  return NULL;
          
			i -= LARGE;
            j = patlen - 1;
            s = (char *)in_pszString + (i - j);
            
			while( --j >= 0 && lowerc( s[j] ) == lowerc( pat[j] ) )
                  ;
            if( j < 0 )                                    /* rdg 10/93 */
                  return s;                                 /* rdg 10/93 */
            
			if( (i += skip2) >= 0 )                        /* rdg 10/93 */
                  return NULL;                              /* rdg 10/93 */
      }
}


#endif	// _ZKStringSearch_H__