/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/* Helper functions for bit manipulations - Mario Contestabile 2 Jan 2000

	Current helpers are:
	is_bit_on()		- Returns true if a bit is set at position in_pos
	set_bit_on()	- Set a bit on
	set_bit_off()	- Set a bit off
	getbits()		- Gets n bits starting at p

*/

#ifndef _ZKS_BITSHLP_H__
#define _ZKS_BITSHLP_H__

namespace ZKS{ 

// PURPOSE:	Determine if a certain bit is on. Bits start at position 0, least significant
// REQUIRE: Unsigned int or lower range
// PROMISE: Returns boolean true if bit is on
// THROWS:	Nothing
// EXAMPLE:	See the code below for a complete example
#ifdef WIN32
#pragma warning(disable:4800)
#endif
extern inline bool is_bit_on(unsigned int in_ui, int in_pos){
	return in_ui & (1<<in_pos);
}

// PURPOSE:	Set one bit on. Bits start at position 0, least significant
// REQUIRE: Unsigned int or lower range
// PROMISE: 
// THROWS:	Nothing
// EXAMPLE:	See the code below for a complete example
extern inline void set_bit_on(unsigned int& in_ui, int in_pos){
	in_ui |= 1<<in_pos;
}

// PURPOSE:	Set one bit off. Bits start at position 0, least significant
// REQUIRE: Unsigned int or lower range
// PROMISE: 
// THROWS:	Nothing
// EXAMPLE:	See the code below for a complete example
extern inline void set_bit_off(unsigned int& in_ui, int in_pos){
	in_ui &= ~(1<<in_pos);
}


// PURPOSE:	Get n bits starting at position p
//			n bits are numbered from 0 (least significant) up
//			position p starts at 1 and is read from right to left
// REQUIRE: Unsigned int or lower range
// PROMISE: 
// THROWS:	Nothing
// EXAMPLE:	See the code below for a complete example
extern inline unsigned int getbits(unsigned int in_x, int in_p, int in_n)
{
    return (in_x >> (in_p+1 - in_n)) & ~(~0 << in_n);
}



}		// name space ZKS

#endif	// _ZKS_BITSHLP_H__

/* Example Usage

void Test5(){

	ZKASSERT(ZKS::is_bit_on(18,   1));
	ZKASSERT(ZKS::is_bit_on(18,   4));
	ZKASSERT(ZKS::is_bit_on(64,   6));
	ZKASSERT(ZKS::is_bit_on(0xFF, 0));
	ZKASSERT(ZKS::is_bit_on(0xFF, 1));
	ZKASSERT(ZKS::is_bit_on(0xFF, 2));
	ZKASSERT(ZKS::is_bit_on(0xFF, 3));
	ZKASSERT(ZKS::is_bit_on(0xFF, 4));
	ZKASSERT(ZKS::is_bit_on(0xFF, 5));
	ZKASSERT(ZKS::is_bit_on(0xFF, 6));
	ZKASSERT(ZKS::is_bit_on(0xFF, 7));	
}

void Test6(){

	unsigned int x = 0;

	ZKS::set_bit_on(x, 0);
	ZKASSERT(ZKS::is_bit_on(x, 0));

	ZKS::set_bit_on(x, 1);
	ZKASSERT(ZKS::is_bit_on(x, 1));

	ZKS::set_bit_on(x, 2);
	ZKASSERT(ZKS::is_bit_on(x, 2));

	ZKS::set_bit_on(x, 3);
	ZKASSERT(ZKS::is_bit_on(x, 3));

	ZKS::set_bit_on(x, 4);
	ZKASSERT(ZKS::is_bit_on(x, 4));

	ZKS::set_bit_on(x, 32);
	ZKASSERT(ZKS::is_bit_on(x, 32));

	ZKS::set_bit_off(x, 0);
	ZKS::set_bit_off(x, 1);
	ZKS::set_bit_off(x, 2);
	ZKS::set_bit_off(x, 3);
	ZKS::set_bit_off(x, 4);
	ZKS::set_bit_off(x, 32);

	ZKASSERT(!ZKS::is_bit_on(x, 0));
	ZKASSERT(!ZKS::is_bit_on(x, 1));
	ZKASSERT(!ZKS::is_bit_on(x, 2));
	ZKASSERT(!ZKS::is_bit_on(x, 3));
	ZKASSERT(!ZKS::is_bit_on(x, 4));
	ZKASSERT(!ZKS::is_bit_on(x, 32));
}

void Test7(){

	unsigned int x = 16; // 10000
	ZKASSERT(ZKS::getbits(x, 4, 1) == 1);
	ZKASSERT(ZKS::getbits(x, 3, 1) == 0);
	ZKASSERT(ZKS::getbits(x, 2, 1) == 0);
	ZKASSERT(ZKS::getbits(x, 1, 1) == 0);
	ZKASSERT(ZKS::getbits(x, 0, 1) == 0);

	x = 1; // 00001
	ZKASSERT(ZKS::getbits(x, 0, 1) == 1);
	ZKASSERT(ZKS::getbits(x, 1, 1) == 0);
	ZKASSERT(ZKS::getbits(x, 2, 1) == 0);
	ZKASSERT(ZKS::getbits(x, 3, 1) == 0);
	ZKASSERT(ZKS::getbits(x, 4, 1) == 0);

	x = 21; // 10101
	ZKASSERT(ZKS::getbits(x, 0, 1) == 1);
	ZKASSERT(ZKS::getbits(x, 1, 1) == 0);
	ZKASSERT(ZKS::getbits(x, 2, 1) == 1);
	ZKASSERT(ZKS::getbits(x, 3, 1) == 0);
	ZKASSERT(ZKS::getbits(x, 4, 1) == 1);

	ZKASSERT(ZKS::getbits(x, 2, 3) == 5); // 101 LS
	ZKASSERT(ZKS::getbits(x, 1, 2) == 1); //  01 LS
	ZKASSERT(ZKS::getbits(x, 4, 5) == 21);
	ZKASSERT(ZKS::getbits(x, 4, 3) == 5); // 101 MS
}
*/