
/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s): Mario Contestabile

 */

#ifndef _CERT_HELPERS_HPP__
#define _CERT_HELPERS_HPP__

#include <clientdebug.h> // for ZKASSERT

#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#include <openssl/rsa.h>


const char	PROXYNAME[] = "SSL proxy";


// PURPOSE: Create a new certificate file
//			'in_certfile' is the root certificate which is in the browser
//			The private key for that certificate is in 'in_keyfile'
//
//			If the certificate in the browser changes, or if the 'in_keyfile' file
//			is lost, we need to re-install a new root certificate in the browser.
//
//			Generally this is called once, possibly at application installation time.
//			It is important that each user have a different root cert, so this code
//			should probably be run during the installation.
//
//		    The certificate will be good for 20 years
//			The issuer will be PROXYNAME and the subject will also be PROXYNAME
//
// THROWS:	Nothing
//
// REQUIRES:The pseudo-random number generator MUST be seeded prior to being called.
//			Since this can be called at installation, you have to be carefull!
//			Perhaps the ZkStartSeed dialog can be usefull.
//			On Windows a simple call to RAND_screen(); will do
//
// RETURNS: True on success
//			False on failure: 
//				- out_errorstr will contain the english meaning of the error
//				- out_errorval will contain the numeric error code
//
// PROMISE: If returns True, leaves out_errorstr and out_errorval intact
// DEBUG:	If you want to see the result, send in a valid file pointer for in_fileptr_debug

extern inline bool
MakeRootCert(
			 const char*	in_certfile, 
			 const char*	in_keyfile, 
			 unsigned long& out_errorval,
			 char*			out_errorstr, 
			 const int		in_errlen,
			 FILE*			in_fileptr_debug) throw()
{
	
	ZKASSERT(in_certfile);
	ZKASSERT(in_keyfile);
	ZKASSERT(out_errorstr);
	ZKASSERT(in_errlen);

	SSL_load_error_strings(); // Loads crypto and ssl strings

	BIO * rootkey = BIO_new_file(in_keyfile, "w");
	
	BIO * rootcert = BIO_new_file(in_certfile, "w");	
	
	EVP_PKEY *pkey = EVP_PKEY_new();
	
	X509 * x509 = X509_new();
	
	BIGNUM * num = BN_new();
	
	RSA *rsa = RSA_generate_key(512, 3, 0, 0);
	
	if(!rsa)
	{
		ZKTRACE0("RSA_generate_key failed!\n");
		out_errorval = ERR_get_error();
		strncpy(out_errorstr, ERR_error_string(out_errorval, NULL), in_errlen);
		ERR_free_strings();
		return false;
	}
	
	if (!EVP_PKEY_assign_RSA(pkey, rsa))
	{
		ZKTRACE0("EVP_PKEY_assign_RSA failed!\n");
		out_errorval = ERR_get_error();
		strncpy(out_errorstr, ERR_error_string(out_errorval, NULL), in_errlen);
		ERR_free_strings();
		return false;
	}
	
	X509_set_pubkey(x509, pkey);
	
	X509_NAME * subj = X509_get_subject_name(x509);
	X509_NAME_add_entry_by_txt(subj, "CN", MBSTRING_ASC, (unsigned char*)PROXYNAME, -1, -1, 0);
	
	X509_NAME * issuer = X509_get_issuer_name(x509);
	X509_NAME_add_entry_by_txt(issuer, "CN", MBSTRING_ASC, (unsigned char*)PROXYNAME, -1, -1, 0);
	
	/* Make The serial number a 128 bit random number */ 
	BN_rand(num, 128, 0, 0);
	BN_to_ASN1_INTEGER(num, X509_get_serialNumber(x509));
	
	X509_gmtime_adj(X509_get_notBefore(x509), 0);
	X509_gmtime_adj(X509_get_notAfter(x509), 20*365*60*60*24);
	
	if (!X509_sign(x509, pkey, EVP_md5()))
	{
		out_errorval = ERR_get_error();
		strncpy(out_errorstr, ERR_error_string(out_errorval, NULL), in_errlen);
		ERR_free_strings();
		return false;
	}
	
	if(1 != PEM_write_bio_PrivateKey(rootkey, pkey, 0, 0, 0, 0, 0))
	{
		ZKTRACE0("PEM_write_bio_PrivateKey() failed in MakeRootCert\n");
		return false;
	}
	
	PEM_write_bio_X509(rootcert, x509);
	
	BIO_free(rootkey);
	BIO_free(rootcert);
	
	EVP_PKEY_free(pkey);


	if(in_fileptr_debug) 
	{
		X509_print_fp(in_fileptr_debug, x509);
	}


	X509_free(x509);

	return true;
}

// PURPOSE: SSL_get_verify_result returns an error code for a certificate
//			If no peer certificate was presented, the returned result code is X509_V_OK. 
//			This is because no verification error occured, it does however not indicate 
//			success. SSL_get_verify_result() is only useful in connection with SSL_get_peer_certificate
//			
extern inline char* 
GetCertVerificationError(const long in_err) throw()
{

	ZKASSERT( X509_V_OK != in_err );

	switch(in_err)
	{
	case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT:
		return "UNABLE_TO_GET_ISSUER_CERT";
		break;
	case X509_V_ERR_UNABLE_TO_GET_CRL:
		return "UNABLE_TO_GET_CRL";
		break;
	case X509_V_ERR_UNABLE_TO_DECRYPT_CERT_SIGNATURE:
		return "UNABLE_TO_DECRYPT_CERT_SIGNATURE";
		break;
	case X509_V_ERR_UNABLE_TO_DECRYPT_CRL_SIGNATURE:
		return "UNABLE_TO_DECRYPT_CRL_SIGNATURE";
		break;
	case X509_V_ERR_UNABLE_TO_DECODE_ISSUER_PUBLIC_KEY:
		return "UNABLE_TO_DECODE_ISSUER_PUBLIC_KEY";
		break;
	case X509_V_ERR_CERT_SIGNATURE_FAILURE:
		return "CERT_SIGNATURE_FAILURE";
		break;
	case X509_V_ERR_CRL_SIGNATURE_FAILURE:
		return "CRL_SIGNATURE_FAILURE";
		break;
	case X509_V_ERR_CERT_NOT_YET_VALID:
		return "CERT_NOT_YET_VALID";
		break;
	case X509_V_ERR_CERT_HAS_EXPIRED:
		return "CERT_HAS_EXPIRED";
		break;
	case X509_V_ERR_CRL_NOT_YET_VALID:
		return "CRL_NOT_YET_VALID";
		break;
	case X509_V_ERR_CRL_HAS_EXPIRED:
		return "CRL_HAS_EXPIRED";
		break;
	case X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD:
		return "ERROR_IN_CERT_NOT_BEFORE_FIELD";
		break;
	case X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD:
		return "ERROR_IN_CERT_NOT_AFTER_FIELD";
		break;
	case X509_V_ERR_ERROR_IN_CRL_LAST_UPDATE_FIELD:
		return "ERROR_IN_CRL_LAST_UPDATE_FIELD";
		break;
	case X509_V_ERR_ERROR_IN_CRL_NEXT_UPDATE_FIELD:
		return "ERROR_IN_CRL_NEXT_UPDATE_FIELD";
		break;
	case X509_V_ERR_OUT_OF_MEM:
		return "OUT_OF_MEM";
		break;
	case X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT:
		return "DEPTH_ZERO_SELF_SIGNED_CERT";
		break;
	case X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN:
		return "SELF_SIGNED_CERT_IN_CHAIN";
		break;
	case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY:
		return "UNABLE_TO_GET_ISSUER_CERT_LOCALLY";
		break;
	case X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE:
		return "UNABLE_TO_VERIFY_LEAF_SIGNATURE";
		break;
	case X509_V_ERR_CERT_CHAIN_TOO_LONG:
		return "CERT_CHAIN_TOO_LONG";
		break;
	case X509_V_ERR_CERT_REVOKED:
		return "CERT_REVOKED";
		break;
	case X509_V_ERR_INVALID_CA:
		return "INVALID_CA";
		break;
	case X509_V_ERR_PATH_LENGTH_EXCEEDED:
		return "PATH_LENGTH_EXCEEDED";
		break;
	case X509_V_ERR_INVALID_PURPOSE:
		return "INVALID_PURPOSE";
		break;
	case X509_V_ERR_CERT_UNTRUSTED:
		return "CERT_UNTRUSTED";
		break;
	case X509_V_ERR_CERT_REJECTED:
		return "CERT_REJECTED";
		break;
	case X509_V_ERR_APPLICATION_VERIFICATION:
		return "The application is not happy";
		break;
	default:
		return "Unknown";
	};
}


#endif