/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/* Helper functions for Date routines - Mario Contestabile 2 Jan 2000

	Current helpers are:
	IsLeapYear() - Use to determine if year is a leap year
	DayOfWeek()  - Use to determine the day of the week for a given date
*/
#ifndef _ZKSDATE_HLP_H__
#define _ZKSDATE_HLP_H__


// for ZKASSERT
#include <clientdebug.h>
// for ZKSTD
#include <client.h>
#include <algorithm>
#include <string>
#include <string_hlp.h>
#include <sstream>

namespace ZKS{

// PURPOSE:	Determine if a year is a leap year
// REQUIRE: Nothing
// PROMISE: Not to throw
// THROWS:	Nothing
// EXAMPLE:	See the code below for a complete example
extern inline bool IsLeapYear(int y){
	return !(y%4) && ((y%100) || !(y%400));
}

// PURPOSE:	Determine the day of the week for any given date
// REQUIRE: Input date be greater than 14/09/1752
// PROMISE: Returns 0 for Monday, 1 for Tuesday, etc
// THROWS:	Nothing
// EXAMPLE:	See the code below for a complete example
extern inline int DayOfWeek(int in_Day, int in_Month, int n_Year){	

	if((in_Month==1) || (in_Month==2)){
		in_Month+= 12;
		n_Year--;
	}

	return (in_Day + 2*in_Month + 3*(in_Month+1)/5 + n_Year + n_Year/4 - n_Year/100 + n_Year/400) % 7;
}

extern inline bool 
GetSecondsCUT(const struct tm& tmExpires, long& lSeconds)
	// From a tm struct, get the number of seconds elapsed since 
	// midnight (00:00:00) jan 1, 1970, coordinated universal time.
	// BTW, COLEDateTime can't handle the format of the Date-Time strings
	// rec'd in cookies, so this type of (messy) solution will have to do.
{
	bool ReturnValue = true;
	long OneDay = 60*60*24;

	///////////////////////////////////////////////////////////////////////////////
	// # of years since 1970
	long lYears = (tmExpires.tm_year - 70);
	
	if ( lYears < 0 )
	{
		ZKASSERT( false ) ; // Shouldn't happen anymore now that I moved check to Parse code.... Need to test with
						  // site that uses this format - Andrew

		lYears = 30 + tmExpires.tm_year; // sometimes get: Wednesday, 09-Nov-36 23:12:40 GMT
	}

	long lLeapYearDays = (lYears / 4) ;

	long lDays = (lYears * 365) + lLeapYearDays ;

	///////////////////////////////////////////////////////////////////
	// Add in all the days in the completed months of the year

	static const long arrayTotalDaysUpUntilMonth[] = {  0,
													    31,
														31 + 28,
														31 + 28 + 31,
														31 + 28 + 31 + 30,
														31 + 28 + 31 + 30 + 31,
														31 + 28 + 31 + 30 + 31 + 30,
														31 + 28 + 31 + 30 + 31 + 30 + 31,
														31 + 28 + 31 + 30 + 31 + 30 + 31 + 31,
														31 + 28 + 31 + 30 + 31 + 30 + 31 + 31 + 31 ,
														31 + 28 + 31 + 30 + 31 + 30 + 31 + 31 + 31 + 31 ,
														31 + 28 + 31 + 30 + 31 + 30 + 31 + 31 + 31 + 31 + 31 } ;


	ZKASSERT( tmExpires.tm_mon < 12 ) ;
	
	lDays += arrayTotalDaysUpUntilMonth[ tmExpires.tm_mon ] ;
	
	// Add in the days of this month
	lDays += tmExpires.tm_mday ;

	// Convert to seconds
	lSeconds = lDays * OneDay;
	
	// Hours
	lSeconds = lSeconds + (tmExpires.tm_hour * 3600);

	// Minutes
	lSeconds = lSeconds + (tmExpires.tm_min * 60);

	// Seconds
	lSeconds = lSeconds + tmExpires.tm_sec;

	return ReturnValue;
}

// Format: Mon, 21 Feb 2000 20:54:17 GMT
extern inline void ConvertTmToString(ZKSTD::string& out_strDate, tm& in_time)
{
	char days[7][4] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
	char months[12][4] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
							"Aug", "Sep", "Oct", "Nov", "Dec"};

	char buffer[64];

	sprintf(buffer, "%s, %02d %s %d %d:%02d:%02d GMT",
					days[in_time.tm_wday], in_time.tm_mday,
					months[in_time.tm_mon], in_time.tm_year + 1900,
					in_time.tm_hour, in_time.tm_min, in_time.tm_sec);

	out_strDate = buffer;
}

// Takes out the first token (everything up to the first space)
// and puts it into the tm_mday field
extern inline void 
ParseDay(ZKSTD::string& io_strDate, tm& out_time)
{
	ZKSTD::istringstream istr(io_strDate);
	istr >> out_time.tm_mday;

	io_strDate.erase(0, io_strDate.find_first_of(' '));
}

// Takes out the first token (everything up to the first space)
// and puts it into the time fields
extern inline void 
ParseTime(ZKSTD::string& io_strDate, tm& out_time)
{
	ZKSTD::string temp(io_strDate.substr(0, io_strDate.find_first_of(' ')));
	ZKSTD::replace(temp.begin(), temp.end(), ':', ' ');
	ZKSTD::istringstream istr(temp);

	istr >> out_time.tm_hour;

	istr >> out_time.tm_min;

	istr >> out_time.tm_sec; 
	
	io_strDate.erase(0, io_strDate.find_first_of(' '));
}


// Takes out the first token (everything up to the first space)
// and puts it into the tm_year field
extern inline void 
ParseYear(ZKSTD::string& io_strDate, tm& out_time)
{
	ZKSTD::istringstream istr(io_strDate);

	istr >> out_time.tm_year;

	if(out_time.tm_year > 1900)
	{
		out_time.tm_year -= 1900;
	}
	if(out_time.tm_year < 70)
	{
		out_time.tm_year += 100;
	}
	if(out_time.tm_year > 137)
	{
		out_time.tm_year = 137;
	}
	
	io_strDate.erase(0, io_strDate.find_first_of(' '));
}

// Takes out the first token (everything up to the first space)
// and puts it into the tm_mon or tm_wday field
extern inline void 
ParseDayMonthOrZone(ZKSTD::string& io_strDate, tm& out_time)
{
	ZKASSERT(io_strDate.length() >= 3);

	char days[7][4] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
	char months[12][4] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
							"Aug", "Sep", "Oct", "Nov", "Dec"};

	bool found = false;
	int i;
	try
	{
		for(i = 0; i < 7; i++)
		{
			if(io_strDate.substr(0, 3) == days[i])
			{
				// Sunday == 0
				out_time.tm_wday = i;
				found = true;
				break;
			}
		}
		if(!found)
		{
			for(i = 0; i < 12; i++)
			{
				if(io_strDate.substr(0, 3) == months[i])
				{
					// January == 0
					out_time.tm_mon = i;
					found = true;
					break;
				}
			}
		}
		if(!found)
		{
			// Do we need to handle time zones here?
		}
	}
	catch(ZKSTD::out_of_range&)
	{
		ZKASSERT(0);
	}
	io_strDate.erase(0, io_strDate.find_first_of(' '));
}

/*
[Mon, ]20 Jun 82 23:59[:59] GMT			**Supported
Weekday, DD-Mon-YY HH:MM:SS TIMEZONE	**Supported
Wed Jan 02 02:03:55 1980				**Supported (Finally!)
*/
extern inline void 
ParseDateString(const ZKSTD::string& in_strDate, struct tm& out_time)
{
	ZKASSERT(!in_strDate.empty());
	ZKSTD::string strDate(in_strDate);

	memset(&out_time, 0, sizeof(out_time));


	ZKSTD::replace(strDate.begin(), strDate.end(), '-', ' ');
	ZKSTD::replace(strDate.begin(), strDate.end(), ',', ' ');
	ZKS::trimLeadingWhite(strDate);

	while(strDate.length()){

		if(ZKSTD::isalpha<ZKSTD::string::traits_type::char_type>(strDate[0], ZKSTD::locale())){

			ParseDayMonthOrZone(strDate, out_time);
		}
		else if(ZKSTD::isdigit<ZKSTD::string::traits_type::char_type>(strDate[0], ZKSTD::locale())){

			//The first number we see will be the day of the month.
			if(out_time.tm_mday == 0){

				ParseDay(strDate, out_time);
			}
			// We already have the day... If there is a ':', this is the time
			else if(strDate.find_first_of(':') < 3){

				ParseTime(strDate, out_time);
			}
			else if (out_time.tm_year == 0) {
				// This must be the year
				ParseYear(strDate, out_time);
			}
			else
			{
				// Don't know what it is, just get rid of it
				strDate.erase(0, strDate.find_first_of(' '));
			}
		} else {
            // Disard one character so we know that we will terminate sometime.
            strDate.erase(0,1);
        }
		
		ZKS::trimLeadingWhite(strDate);
	}// while
	
	if(out_time.tm_year == 0)
	{
		ZKTRACE1( "Could not parse year %s ---- Fudging value to 2037\n", in_strDate.c_str() ) ;
		ZKASSERT(false);
		out_time.tm_year = 2037;
	}
	if(out_time.tm_mday == 0)
	{
		ZKTRACE1( "Could not parse day %s ---- Fudging value to the 13th\n", in_strDate.c_str() ) ;
		ZKASSERT(false);
		out_time.tm_mday = 13;
	}
	

	// Validate our final output - does it make sense?
	ZKASSERT( out_time.tm_mon < 12 ) ;
	ZKASSERT( out_time.tm_year > 0 ) ;

}


}		// namespace ZKS

#endif	// _ZKSDATE_HLP_H__

/* Example Usage:

void Test3(){
	ZKASSERT(ZKS::IsLeapYear(2000));
	ZKASSERT(!ZKS::IsLeapYear(1999));
	ZKASSERT(!ZKS::IsLeapYear(1998));
	ZKASSERT(!ZKS::IsLeapYear(1997));
	ZKASSERT(ZKS::IsLeapYear(1996));
}

void Test4(){
	ZKASSERT(ZKS::DayOfWeek(2,  1,  2000) == 6); // sunday
	ZKASSERT(ZKS::DayOfWeek(28, 1,  1980) == 0); // monday
	ZKASSERT(ZKS::DayOfWeek(20, 9, 2080) == 4); // friday
	ZKASSERT(ZKS::DayOfWeek(9,  3, 2038) == 1); // tuesday
}

*/
