/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/*
	Mario Contestabile March 11, 2000

	This header is for big-little endian handling across platforms.
	LITTLE_ENDIAN or BIG_ENDIAN *must* be #defined prior to #including this header

	Functions are:

	- isLittleEndian()	Returns true if the machine is little endian
	- Two templates, big_endian and little_endian to aid in switching
	  values from/to different endianess

	Much of this work was presented in the C++ boost libraries
	done by Mark Borgerding

*/

#ifndef _ENDIAN_HELPERS__
#define _ENDIAN_HELPERS__

inline bool isLittleEndian() {
	int temp = 1;
	return (*(char*)&temp) != 0;
}


inline
unsigned int reverse_bytes(unsigned int d)
{
   ZKASSERT(sizeof(unsigned int) == 4);
   return ((d>>24)&0xff) | ((d&0xff)<<24) | ((d>>8)&0xff00) | ((d&0xff00)<<8);
}

inline
unsigned short reverse_bytes(unsigned short d)
{
   ZKASSERT(sizeof(unsigned short) == 2);
   return (d>>8) | (d<<8);
}

inline
unsigned char reverse_bytes(unsigned char d)
{
   ZKASSERT(sizeof(unsigned char) == 1);
   return d;
}

// general purpose function for swapping byte order (much slower than above fucntions)
template <class T>
inline
T reverse_bytes(T d)
{
   T orig(d);
   const char * pSrc = reinterpret_cast<const char *>(&orig);
   const char * pSrcEnd = pSrc + sizeof(T);
   char * pDest = reinterpret_cast<char *>(&d) + sizeof(T);

   for (;pSrc != pSrcEnd;++pSrc)
      *--pDest = *pSrc;
   return d;
};



#ifdef LITTLE_ENDIAN

template <class T>
struct big_endian 
{
   big_endian(T d=0):data(reverse_bytes(d)){}
   operator const T() const {return reverse_bytes(data);}
private:
   T data;
};

template <class T>
struct little_endian 
{
   little_endian(T d=0):data(d){}
   operator const T() const {return data;}
private:
   T data;
};

#elif  BIG_ENDIAN

template <class T>
struct big_endian 
{
   big_endian(T d=0):data(d){}
   operator const T() const {return data;}
private:
   T data;
};

template <class T>
struct little_endian 
{
   little_endian(T d=0):data(reverse_bytes(d)){}
   operator const T() const {return reverse_bytes(data);}
private:
   T data;
};

#else	// not LITTLE_ENDIAN or BIG_ENDIAN

#error "You need to #define LITTLE_ENDIAN or BIG_ENDIAN"

#endif	// LITTLE_ENDIAN or BIG_ENDIAN


#endif // _ENDIAN_HELPERS__
