/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/* Helper functions for File routines - Mario Contestabile 18 March 2000
- Gabriel Duford 13 April 2000 : ExtractFileName

	Current helpers are:
	FileCopy      - Copy contents of a file into another given two FILE pointers
	FileCopy      - Copies contents of one file into another given their names
	FileCopy      - Copies contents of one file into another given an output stream and an input stream
	FileCopywithverification - Copies contents of one file into another given their names, you decide whether to overwrite or not
	GetTempFileName  - Use to get a valid temporary file name
	FileExists	  - Verify if a given file name exists
	FileCopytodir - Copies contents of one file into another directory using the same file name
	ReadTextFile  - Places the contents of a text file into a string separated by \n
	ReadBinaryFile - Places the contents of a binary file into a string
	GetFileSize	  - Returns the size in bytes of a file
	GetLinesExceedingWidth - Returns the lines in a file exceeding the indicated length
	ExtractFileName - Extract the file name from a path (absolute || relative || none).
	WriteBinaryFile - Write the buffer into the specified file

*/

#ifndef _ZKSFILE_HLP_H__
#define _ZKSFILE_HLP_H__

// C++
#include <fstream>
#include <memory>
#include <vector>
#include <map>
// C
#include <stdio.h>
// Posix
#include <sys/types.h>
#include <sys/stat.h>

#ifdef ZK_LINUX
#  include <fcntl.h>
#endif

// strictly for ZKASSERT
#include <clientdebug.h>

namespace ZKS{

// PURPOSE:	Get a string which can represent a valid and unique file name
//			uses the P_tmpdir directory from stdio.h (which may be "\\")
//			so a name may be in the form "\\d9sdjs9"
//			do not call free() on the string
//			Use a buffer at least L_tmpnam * 2 size
//			returns a NULL string if name cannot be generated
//			If you're using a FILE* you may try using tmpfile() in place of this function
//			since the file will get deleted when the app terminates.
// REQUIRE:	
// PROMISE: 
// THROWS:	Nothing
// EXAMPLE: See Test15
extern inline void GetTempFileName(char* out_pszFileName ) throw()
{
	ZKASSERT(out_pszFileName);
	tmpnam(out_pszFileName);	
	ZKASSERT(out_pszFileName);
}


// PURPOSE:	Verifies byte by byte that two files are equal
// REQUIRE:	
// PROMISE: Neither file is not changed by this call
//			However FILE* are pointing to an undetermined location
//			(Use fseek to return to a known place)
// THROWS:	Nothing

extern inline bool AreFilesEqual( FILE* in_f1, FILE* in_f2 ) throw()
{	
	static char	buf1[BUFSIZ] = { 0 };
	static char	buf2[BUFSIZ] = { 0 };

	if(!in_f1 || !in_f2) return false ;

	fseek(in_f1, 0, SEEK_SET );
	fseek(in_f2, 0, SEEK_SET );

	while(!feof(in_f1))
	{
		const size_t count1 = fread(buf1, 1, BUFSIZ, in_f1);
		const size_t count2 = fread(buf2, 1, BUFSIZ, in_f2);

		if(ferror(in_f1))
		{	
			return false ;
		}

		if(ferror(in_f2))
		{	
			return false ;
		}


		if ( count1 != count2 )
		{	
			return false ;
		}

		if ( 0 != memcmp( buf1, buf2, count1 ) )
		{
			return false ;
		}
	}
	return true;
}



// PURPOSE:	Copy contents of a file into another given two FILE pointers
//			Returns 0 if either pointer is NULL
//			Returns 1 on success, 0 on error
// REQUIRE:	in_source is open for reading (r or w+)
// PROMISE: Rewinds in_source file pointer to beginning after verifying both pointers aren't NULL
// THROWS:	Nothing
// EXAMPLE: See Test15

#ifndef ZK_LINUX

extern inline int FileCopy( FILE* in_dest,  FILE* in_source ) throw()
{	
	static char	buf[BUFSIZ] = { 0 };

	if(!in_dest || !in_source) return 0;

	fseek(in_source, 0, SEEK_SET);

	while(!feof(in_source))
	{
		const size_t count = fread(buf, 1, BUFSIZ, in_source);
		if(ferror(in_source))
			return 0;
		if(fwrite(buf, 1, count, in_dest) != count)
			return 0;
	}
	return 1;
}

#else

// Safe version for Linux

extern inline int FileCopy( FILE* in_dest,  FILE* in_source ) throw()
{	
    ZKASSERT(" OLD VERSION " && 0);
	char buf[BUFSIZ] = { 0 };
    
	if(!in_dest || !in_source) return 0;

	fseek(in_source, 0, SEEK_SET);

	while(!feof(in_source))
	{
        size_t written = 0;
		size_t count = fread(buf, 1, BUFSIZ, in_source);

		if(ferror(in_source))
			return 0;
    retry:
		if( (written = fwrite(buf, 1, count, in_dest)) != count )
        {            
            if (errno == EINTR) 
            {
                // Well, this function is acting funny, so to be safe...
                if (written != 0) return 0;
                ZKTRACE0("I dunno if fwrite can set errno, the glibc sources look like they can, just in case");
                goto retry;
            }
            if ( ferror(in_dest) ) {
                ZKASSERT("WRITE ERROR IN FILECOPY" && 0);
                return 0;
            }
			ZKASSERT( "SHORT WRITE IN FileCopy !!!" && 0 );
            // We hsould do something smart here
            return 0;
        }
	}
	return 1;
}

#endif


// PURPOSE:	Verifies byte by byte that two files are equal
// REQUIRE:	
// PROMISE: Neither file is not changed by this call
// THROWS:	Nothing

extern inline bool AreFilesEqual(const char* in_dest, const char* in_source) throw()
{
	if(!in_source || !in_dest) return 0;

	bool bRet = 0;

	FILE* fFrom = fopen(in_source, "rb");
	FILE* fTo   = fopen(in_dest,   "rb");

	if(fFrom && fTo)
		bRet = AreFilesEqual(fTo, fFrom);

	if(fFrom) fclose(fFrom); fFrom = NULL;
	if(fTo)   fclose(fTo);   fTo   = NULL;

	return bRet;
}



#ifdef WIN32
// PURPOSE:	Copies contents of one file into another given their names
// REQUIRE:	
// PROMISE: Returns 1 on success, 0 on error
// THROWS:	Nothing
// EXAMPLE: See Test15
extern inline int FileCopy(const char* in_dest, const char* in_source) throw()
{
	if(!in_source || !in_dest) return 0;

	int nRet = 0;

	FILE* fFrom = fopen(in_source, "rb");
	FILE* fTo   = fopen(in_dest,   "wb");

	if(fFrom && fTo)
		nRet = FileCopy(fTo, fFrom);

	if(fFrom) fclose(fFrom); fFrom = NULL;
	if(fTo)   fclose(fTo);   fTo   = NULL;

	return nRet;
}

#else
inline int FileCopyRecursive(const char* in_dest, const char* in_source, int count) throw()
{
	if(!in_source || !in_dest) return 0;

// Recursive hackiness
    if (count >= 3) return 0;

	int nRet = 0;
    int fdfrom = -1, fdto = -1;
    char *buf = new char[BUFSIZ];
    ssize_t read_size, write_size;

    if ( ( fdfrom = open( in_source, O_RDONLY ) ) == -1 ) {
        ZKTRACE1( "Could not open %s.\n", in_source );
        delete[] buf;
        return false;
    }

    if ( ( fdto = open( in_dest, O_CREAT|O_RDWR|O_TRUNC, S_IRUSR | S_IWUSR ) ) == -1 ) {
        ZKTRACE1( "Could not open %s.\n", in_dest );
        close (fdfrom);
        delete[] buf;
        return false;
    }

 retry_read:
    while ( ( read_size = read( fdfrom, (void*)buf, BUFSIZ ) ) > 0 ) {
        write_size = 0;
        while ( read_size > 0 ) {
        retry_write:
            write_size = write( fdto, (const void *)&buf[write_size], read_size );
            if (write_size == -1) {
                if (errno == EINTR) goto retry_write;
                ZKTRACE1("File copy failed in write with errno %d\n", errno);
                ZKASSERT("Read the trace" && 0);
                delete[] buf;
                close (fdto);
                close (fdfrom);
                return 0;
            }
            read_size -= write_size;
            ZKASSERT( read_size >= 0 );
            ZKASSERT( sizeof(buf) - write_size >= read_size );
        }
    }
    if (read_size == -1) {
        if (errno == EINTR) goto retry_read;
        ZKTRACE1("File copy failed in read with errno %d\n", errno);
        ZKASSERT("Read the trace" && 0);
        delete[] buf;
        close (fdto);
        close (fdfrom);
        return 0;
    }

	// We were having some trouble with corrupt files on
	// linux, possibly caused by this code - this will
	// try to flag the problem as it occurs - talk to Andrew (see #5771)

    nRet = AreFilesEqual( in_dest, in_source ) ;
    ZKASSERT( nRet ) ;
    if (nRet == 0) {
        close (fdto);
        close (fdfrom);        
        delete[] buf;
        ZKTRACE1("Trying to resave count == %d\n", count);
        return FileCopyRecursive(in_dest, in_source, count+1);
    }

    delete[] buf;
    close (fdto);
    close (fdfrom);

	return nRet;
}

inline int FileCopy(const char* in_dest, const char* in_source) throw()
{
    return FileCopyRecursive(in_dest, in_source, 0);
}

#endif

// PURPOSE:	Verify if a file can be writen to, calling this does not change the
//			contents in any way.  Returning "true" implies that the
//			file exists, but does not imply that the file can be "read".
// REQUIRE:	A valid file
// PROMISE: Returns true if the file can be opened in "write" mode
// THROWS:	Nothing
extern inline bool IsFileWritable(const char* in_dest) throw()
{
	ZKASSERT( in_dest != NULL ) ;

	if (( in_dest == NULL ) || ( 0 == strlen( in_dest) ))
	{
		ZKASSERT( false ) ;
		return false ;
	}

	FILE* fFile = fopen(in_dest, "ab") ;

	if ( fFile != NULL )
	{
		fclose(fFile);
		return true ;
	}

	return false ;
}

// PURPOSE:	Verify if a given file name exists
// REQUIRE:
// PROMISE: Returns true if file exists (tries to open it to find out)
// EXAMPLE: See Test15
extern inline bool FileExists(const char *in_FileName) throw()
{
	// REVIEW - see 6185
	// The file can exist but not have "read" permissions for the current user
	// This function should distinguish correctly - it may required "if/def"
	// platform specific code.
	
    struct stat stbuf = { 0 };
    // 0 == success 1 == file doesn't exist
    return !stat( in_FileName, &stbuf ); 
}


// PURPOSE: Exactly like filecopy() but verifies
//			if the destination file exists beforehand
//			and depeneding on that you specified, proceeds
//			with the copying, or not
// PROMISE: Returns  1 on success, 0 on error [passing true & the file exists also returns 0]
// EXAMPLE: See Test16
extern inline int FileCopyWithVerification(
	const char* in_pszExistingFileName, 
	const char* in_pszNewFileName,
	bool        in_bFailIfExists ) throw()
{

	if(in_bFailIfExists && FileExists(in_pszNewFileName) )
		return 0;

	return FileCopy( /*dest*/ in_pszNewFileName, in_pszExistingFileName /* source*/ );
}


// PURPOSE:	Copies contents of one file into another given an output stream and an input stream
// REQUIRE:
// PROMISE: Returns 1 on success, 0 on error
// THROWS:	Possibly, if in_dest.operator<< throws or in_source.rdbuf() throws
// EXAMPLE: See Test15
extern inline int FileCopy(ZKSTD::ofstream& in_dest, const ZKSTD::ifstream& in_source) 
{

	if(!in_source || !in_dest) return 0;
	
	in_dest << in_source.rdbuf();

	if( in_dest.fail() || in_source.fail() ) return 0;

	return 1;
}


// PURPOSE:	Copies contents of one file into another directory using the same file name
// REQUIRE:
// PROMISE: Use posix calls. Returns 1 on success, 0 on error
//			Will append forward slash to given 'in_file' filename
// THROWS:	nothing
// EXAMPLE: See Test15
extern inline int FileCopyToDir(const char* in_file, const char* in_dir) throw()
{
	ZKASSERT(in_file);
	ZKASSERT(in_dir);

	struct stat finfo = { 0 };

	if( (0 == stat(in_dir, &finfo)) && (finfo.st_mode & S_IFDIR) )
	{
		char newfile[FILENAME_MAX] = { 0 };
#ifdef WIN32
        char fnbuf[_MAX_FNAME] = { 0 };
        char extbuf[_MAX_EXT] = { 0 };
        _splitpath(in_file, NULL, NULL, fnbuf, extbuf);
		sprintf(newfile, "%s/%s%s", in_dir, fnbuf, extbuf );
#else
        // Hmm I think this is a bug on windows too .. if in_file cotains a path
        // it must be stripped before we can turn it into a new file name.
        // (yes it was, fixed above) --ol
		sprintf(newfile, "%s/%s", in_dir, basename(in_file));
#endif
		return FileCopy( newfile, in_file );
	}
	return 0;
}



// PURPOSE: Returns the length of a file
// REQUIRE: An existing filename
// RETURNS: 0 if inexistant file or unable to obtain information
extern inline unsigned long GetFileSize(const char in_file[]) throw()
{
	struct stat finfo = { 0 };
	if(0 == stat(in_file, &finfo))
	{
		return finfo.st_size;
	}
	return 0;
}


// PURPOSE:	Read a text file into a string.
// REQUIRE: The name of the file to read.
//			A string to receive the content of the file.
//			The delimiter wanted in the output string.
//			A delimiter of line to break lines on.
// PROMISE: Returns true if filename was successfully opened for reading
//			Does not clear the output string for you
//			A maximum of str::max_size()-1 characters are read per line
//			Removes duplicates of the set of delimiters you want to have
//			in your output string if they are found before the delimiter
//			on which you break lines; 
//			e.g. readtextfile( "hello\r\nworld\n", str, "\r\n", '\n' ) will return
//			"hello\r\nworld\r\n" instead of "hello\r\r\nworld" if a simple
//			find and replace was done.
//			See the unit test for more examples.
// EXAMPLE: See Test16
extern inline bool ReadTextFile(
	const ZKSTD::string& in_rstrFileName, 
	ZKSTD::string& out_rstrFileContent,
	const ZKSTD::string& in_rstrDelimiterToAdd = "\n",
	const char in_aDelimiter = '\n' )  throw()
{
	bool bSuccess = false;
	try
	{		
		const unsigned long ulFileSize = GetFileSize( in_rstrFileName.c_str() );
		if( ulFileSize ) 
		{
			ZKSTD::auto_ptr<char> buf( new char[ulFileSize+1] );
			ZKSTD::ifstream in( in_rstrFileName.c_str() );
			if( in ) 
			{
				bSuccess = true;				

				// getline() discards the delimiter
				while( in.getline( buf.get(), ulFileSize+1, in_aDelimiter ) )
				{	
					// get() does not relinquish ownership
					ZKSTD::string strLastLine( buf.get() );
					// find_last_not_of(..) return -1 == npos if it finds nothing.
					const unsigned int posOfAlreadyThereDelim = 
										strLastLine.find_last_not_of( in_rstrDelimiterToAdd ) + 1;
					// if it is not the last charactor of the string.
					// we make sure we do not duplicate eol delimiter.
					if( posOfAlreadyThereDelim != strLastLine.length() )
						out_rstrFileContent += strLastLine.substr( 0, posOfAlreadyThereDelim );
					else
						out_rstrFileContent += strLastLine;
					out_rstrFileContent += in_rstrDelimiterToAdd;
				}
			}			
		}
	}
	catch (...) 
	{
        ZKTRACE0("ReadTextFile -- exception\n");
	}

	return bSuccess;
}

// PURPOSE:	Read a binary file into a string
// REQUIRE: That you free() the memory if it returns true
// PROMISE: Returns true if filename was successfully opened for reading
//			To use malloc() to allocate the memory
//			To return false if malloc returns NULL
// EXAMPLE: See Test16
extern inline bool ReadBinaryFile(const char* in_filename, unsigned char** out_buf, size_t& out_siz) throw()
{

	ZKASSERT( out_buf );
	ZKASSERT( in_filename );

	try
	{
		 out_siz = GetFileSize(in_filename);

		 if (out_siz) 
		 {		
			ZKSTD::ifstream in;			
			in.open(in_filename,ZKSTD::ios::binary);
			if (in.is_open()) 
			{

				unsigned char* buf = (unsigned char*) malloc( out_siz );
				if(!out_buf){
					return false;
				}

				in.read((char*)buf, out_siz );

				*out_buf = buf;

				return true;	
				
			}			
		}
	}
	catch(...)
	{
	}

	return false;
}

#ifndef ZK_LINUX

extern inline bool WriteBinaryFile(
	const char*				in_FileName, 
	const unsigned char *	in_pBuffer, 
	const size_t			in_szBuffer )
{
	ZKASSERT( ZKIsValidAddress(in_pBuffer, in_szBuffer, false) );
	ZKASSERT( in_szBuffer > 0 );
	ZKASSERT( in_FileName );

	try
	{
		ZKSTD::ofstream out;
		out.open(in_FileName, ZKSTD::ios::binary);
		if(out.is_open())
		{
			// This code does not check the error return,
			// which we suspect can fail on Linux
			out.write( (char*)in_pBuffer, in_szBuffer);
			return true;
		}
	}
	catch(...)
	{
	}

	return false;
}

#else

extern inline bool WriteBinaryFile(
	const char*				in_FileName, 
	const unsigned char *	in_pBuffer, 
	const size_t			in_szBuffer )
{
    ssize_t bytes_remaining;
    ssize_t wret;
    int wfd;

	ZKASSERT( ZKIsValidAddress(in_pBuffer, in_szBuffer, false) );
	ZKASSERT( in_szBuffer > 0 );
	ZKASSERT( in_FileName );

#ifndef WIN32
    if ( ( wfd = open( in_FileName, O_CREAT|O_RDWR|O_TRUNC, S_IRUSR | S_IWUSR ) ) == -1 ) 
#else
    if ( ( wfd = open( in_FileName, O_CREAT|O_RDWR|O_TRUNC, _S_IREAD | _S_IWRITE ) ) == -1 ) 
#endif
	{
        ZKTRACE1( "Could not open %s.\n", in_FileName );
        return false;
    }

    bytes_remaining = in_szBuffer;
    ZKASSERT("Writing something and didn't wrap to negative in conversion" && bytes_remaining > 0);
    
    while ( bytes_remaining > 0 ) 
	{
    retry:
        wret = write( wfd, (const void *)in_pBuffer, bytes_remaining );
        if ( wret == -1 ) 
		{
            if (errno == EINTR) goto retry;
            ZKTRACE1( "WriteBinaryFile failed errno == %d\n", errno );
            ZKASSERT( "Read the last message, bub." && 0);
            close(wfd);
            return false;
        } 
        bytes_remaining -= wret;
    }
    close(wfd);
    return true;
}

#endif


// PURPOSE: Returns the line number, and the line itself, in the given file if it 
//			exceeds 'in_nMax' in length
// REQUIRE: An empty 'out_linemap', it does not empty the map for you beforehand
//			size_t is the maximum number of lines possible, entries in map will 
//			just wrap around map if exceeds
// THROWS:  Allocates the file size, so may throw bad_alloc exception if operator new fails
// RETURNS: Total Number of offensive lines found
extern inline size_t GetLinesExceedingWidth(
	const char in_file[],
	const size_t in_nMax, 	
	ZKSTD::map<size_t, ZKSTD::string>& out_linemap,
	const char in_delimiter = '\n')
{
	ZKTEST(out_linemap.empty());

	ZKSTD::ifstream infile(in_file);

	if(infile)
	{
		unsigned long lSize = GetFileSize(in_file);
		if(lSize)
		{
			ZKSTD::auto_ptr<char>buf( new char[lSize] );
			size_t lineno = 1;
			while(infile.getline(buf.get(), lSize, in_delimiter))
			{
				if( strlen( buf.get() ) > in_nMax)
				{
					out_linemap[lineno] = buf.get();
				}
				lineno++;
			}
			return out_linemap.size();
		}
	}
	return 0;
}

// PURPOSE: Extract the file name from a path.
// REQUIRE: A vector of file name containing either the absolute path
//			or the single name.
//			A vector of string that will receive only the file name.
// PROMISE: The returned vector will be emptied.
//			The returned vector will be the same size as the first parameter.
//			Do not check if the file exists.
extern inline void ExtractFileName(
	const ZKSTD::vector<ZKSTD::string>& in_rFileNameVector,
	ZKSTD::vector<ZKSTD::string>& out_rFileNameOnlyVector )
{
#ifdef WIN32

	out_rFileNameOnlyVector.clear();

	ZKSTD::vector<ZKSTD::string>::const_iterator it = in_rFileNameVector.begin();
	for( ; it != in_rFileNameVector.end(); ++it )
	{
		if( (*it).find( "\\" ) != (*it).npos )
		{
			char strFullPath[MAX_PATH] = { 0 };
			if( (*it).find( ":" ) == (*it).npos )	// it is a relative path.
				_makepath( strFullPath, "c:", NULL, (*it).c_str(), NULL );
			else
				strncpy( strFullPath, (*it).c_str(), MAX_PATH );

			char strFileName[_MAX_FNAME] = { 0 };
			_splitpath( strFullPath, NULL, NULL, strFileName, NULL );
			out_rFileNameOnlyVector.push_back( strFileName );
		}
		else
			out_rFileNameOnlyVector.push_back( (*it) );
	}
	
	ZKASSERT( in_rFileNameVector.size() == out_rFileNameOnlyVector.size() );

#else
    out_rFileNameOnlyVector.clear();

    ZKSTD::vector<ZKSTD::string>::const_iterator it = in_rFileNameVector.begin();
    for ( ; it != in_rFileNameVector.end(); ++it )
    {
//        char strFileName[MAX_PATH] = { 0 };
        char *fname = basename((*it).c_str());
        out_rFileNameOnlyVector.push_back( fname );
    }

	ZKASSERT( in_rFileNameVector.size() == out_rFileNameOnlyVector.size() );

#endif
}

}		// namespace ZKS

#endif	// _ZKSFILE_HLP_H__
