/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/* Helper functions for Math routines - Mario Contestabile 2 Jan 2000

	Current helpers are:
	Duplicate   - Use to determine a duplicate value in a container
	sum         - Use to determine the sum total of 1 to N digits in constant time
	ZKSFACTORIAL- Use to obtain the factorial of a number at compile time
	factorial   - Use to obtain the factorial of a number at run time (only use if your compiler doesn't handle templates)
	gcd         - Use to obtain greatest common divisor of 2 ints (Euclid's algorithm)
*/

#ifndef _ZKSMATH_HLP_H__
#define _ZKSMATH_HLP_H__


namespace ZKS{

// PURPOSE:	Return the duplicate value found within a container
// REQUIRE: The container support begin(), end(), a const_iterator.
//			The values in the container must be in the range [1,N] and
//			each i in 1 <= i <= N must be in the container with only 1
//			j == i (only 1 value duplicated).
// PROMISE: Return 0 if no duplicate is found.
// THROWS:	Nothing
// EXAMPLE:	See the code below for an example using a vector<> and a list<>
template<typename Container>
typename Container::value_type Duplicate(Container& in_Container){

	typename Container::value_type sum = 0;
	typename Container::value_type numelements = 0;

	// sum the elements in the container
	for(typename Container::const_iterator it = in_Container.begin();
		it != in_Container.end(); 
		++it){

		sum += *it;
		++numelements;
	}

	return ( sum - (numelements * (numelements - 1) / 2) );
}

// PURPOSE:	Determine the total sum of digits between 1 and in_n
// REQUIRE: The type in_n not be float or double
// PROMISE: Constant time 
// THROWS:	Nothing
// EXAMPLE:	See the code below for a complete example
// Algorithm used
// even
//   sum = n + 1 * (n / 2);

// odd
//   sum = n * ((n / 2)+1)
template<typename T>
T sum(T in_n){
	
	const bool bEven = ((in_n % 2) == 0);

	T tmp = 0;

	if(bEven){
		tmp = in_n + 1;
		tmp *= (in_n / 2) ;

	}else{
		tmp = in_n;
		tmp *= ((in_n / 2) + 1);
	}

  return tmp;
}

// PURPOSE:	Determine the factorial of a number at compile time
// REQUIRE: A value less than 41 for MSVC, other compilers may accept more
// PROMISE: 0 runtime cost
// THROWS:	Nothing
// EXAMPLE:	See Test8
template<int N>
struct Factorial
{
  enum {value = N * Factorial<N-1>::value};
};

struct Factorial<0>
{
  enum {value = 1};
};

#define ZKSFACTORIAL(N) (ZKS::Factorial<N>::value)



// PURPOSE:	n! (n-factorial)
// REQUIRE: 
// PROMISE: 
// THROWS:	Nothing
// EXAMPLE:	See Test8
extern inline long factorial(long in_n)
{
	return (in_n <= 1) ? 1 : in_n * factorial(in_n - 1);
}

// PURPOSE: Find's greatest common divisor of 2 ints (Euclid's algorithm)
// REQUIRE: 2 ints as parameters
// PROMISE: 
// THROWS:	Nothing
// EXAMPLE:	See Test13
extern inline int gcd(int x, int y)
{
	while(y != 0)
	{
		const int rem = x % y;
		x = y;
		y = rem;
	}
	return x;
}

}		// namespace ZKS

#endif	// _ZKSMATH_HLP_H__
