/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/* Helper functions for ZKSTD::string - Mario Contestabile Nov 26 1999

	Current helpers are:
	filter()	- Use to remove elements from a string
	pad()		- Use to append blanks or user specified characters to a string
	trim()		- Use to remove trailing blanks from a string
	trimLeadingWhite() - Use to remove leading blanks from a string
	parse()		- Use to remove substring from a string
	MakeUpper()	- Use to make a string upper case
	MakeLower()	- Use to make a string lower case
	SplitOnCharacter() - Use to split one string into several substrings
	ascitohex()		   - Use to convert hex string to hex value
	replace_all()      - Use to replace a substring from a string with another substring (case sensitive)
	InsertInFront - Use to add a string in front of a string (work on vectors of string)
	LookForLine() - Use to find "\r\n" or "\n" in a string.
	ExtractDomain()		- Get the domain from an url.
	EscapeString()		- Prepare a string to be a parameter in an url
	UnEscapeString()	- Convert a string from an url parameter back to the original string
	is_palindrome()     - Returns true if the string is a palindrome

	The C++ syntax allows more than one definition of an inline function with external
	linkage, as long as all the definitions are identical.
*/

#ifndef _STRING_HELPERS_HPP__
#define _STRING_HELPERS_HPP__

#include <string>
#include <algorithm>
#include <locale>
#include <vector>
#include <stdexcept>

// for ZKASSERT
#include <clientdebug.h>
// for ZKSTD
#include <client.h>

#define ZK_LOCAL_COMPUTER	"local_computer"

namespace ZKS{
	

// PURPOSE:	Any elements in strFilt found in strResult will be removed.
//			Elements must be an exact match (case-insensitive)
// REQUIRE: Nothing
// PROMISE: Not to throw
// THROWS:	Although string::erase() can throw out_of_range, we can be assured
//			this particular function does not throw anything.
// EXAMPLE:	ZKS::filter("aeiou", strResult);
// Will remove vowels from strResult
extern inline void filter(const ZKSTD::string& strFilt, ZKSTD::string& strResult){
		
	ZKSTD::string::size_type pos = 0;

	while((pos = strResult.find_first_of(strFilt, pos)) != ZKSTD::string::npos)
		strResult.erase(pos,1);

}

// PURPOSE: pad string with blanks (or user specified character) to a specified length
// REQUIRE: Nothing 
// PROMISE: Not to touch the string if it's longer than l
// THROWS:	length_error if l > str.max_size() [unsigned integral type]
// EXAMPLE: 
//	strResult = "abcde";
//	ZKS::pad(strResult, 8);			// add three blanks
//	ZKS::pad(strResult, 10, 'X');	// add 2 'X'
//	ZKS::pad(strResult, 3);			// does nothing
//	ZKS::pad(strResult, 15);		// add 5 ' '
//	ZKS::trim(strResult);			// remove 5 ' '
//
extern inline void pad(ZKSTD::string& str, size_t l, char fillC = ' ') throw(ZKSTD::length_error){
	if(str.size() < l) str.resize(l,fillC);	
}


// PURPOSE:	strip trailing blanks from string
// REQUIRE: Nothing
// PROMISE: Not to throw
// THROWS:	length_error if l > str.max_size() [unsigned integral type]
//
extern inline void trim(ZKSTD::string& str) throw(ZKSTD::length_error){
	str.resize(str.find_last_not_of(' ')+1);
}

// PURPOSE:	strip leading blanks from string
// REQUIRE: Nothing
// PROMISE: Not to throw
// THROWS:	length_error if l > str.max_size() [unsigned integral type]
//
extern inline void trimLeadingWhite(ZKSTD::string& str) throw(ZKSTD::length_error){
	ZKSTD::string::size_type i = str.find_first_not_of(' ');
	if(i != ZKSTD::string::npos){
		str.erase(0, i);
	}
}


// PURPOSE:	Removes strPattern from strStripped
// REQUIRE: Nothing
// PROMISE: Not to throw
// THROWS:	Although string::erase() can throw out_of_range, we can be assured
//			this particular function does not throw anything.
// EXAMPLE: ZKS::parse(strResult, "CutThis");
// removes the string "CutThis" from strStripped
extern inline void parse(ZKSTD::string& strStripped, const ZKSTD::string& strPattern){

	ZKSTD::string::size_type pos = 0;

	while((pos = strStripped.find(strPattern, pos)) != ZKSTD::string::npos)
		strStripped.erase(pos, strPattern.size() );
}


template <typename T>
class Makeupper{
    public:
		T operator()(const T& p) { return ZKSTD::toupper<T>(p, ZKSTD::locale() ); }
};



// One can also use 
// return ZKSTD::tolower<T>(p, ZKSTD::locale() ); 
// But I am testing this for Corey's sake, seems IE crashes
// upon exit when this was called
template <typename T>
class Makelower{
	public:
#if _MSC_VER == 1200
		T operator()(const T& p) { 
			return ZKSTD::use_facet(ZKSTD::locale::empty(), (ZKSTD::ctype<T> *)0).tolower(p);
		}
#else
		T operator()(const T& p) { return ZKSTD::tolower<T>(p, ZKSTD::locale() ); }
#endif
};


// PURPOSE:	Make your string uppercase
// REQUIRE:	A positive attitude
// PROMISE: Use default 'C' locale
// THROWS:	Nothing
// EXAMPLE: ZKS::MakeUpper(strTemp);
extern inline void MakeUpper(ZKSTD::string& str){

	ZKSTD::transform (
		str.begin (), 
		str.end (), 
		str.begin (),
		Makeupper<ZKSTD::string::traits_type::char_type>()
		);
}


// PURPOSE:	Make your string lowercase
// REQUIRE:	A positive attitude
// PROMISE: Use default 'C' locale
// THROWS:	Nothing
// EXAMPLE: ZKS::MakeLower(strTemp);
extern inline void MakeLower(ZKSTD::string& str){

	ZKSTD::transform (
		str.begin (), 
		str.end (), 
		str.begin (),
		Makelower<ZKSTD::string::traits_type::char_type>()
		);
}


// PURPOSE:	Split one string into smaller strings
//			Adds the strings into the vector
//			(We don't empty it. Anything already in the
//			vector stays there). Splits on in_split, or
//			';' by default.
// PROMISE: Doesn't add empty strings.
// THROWS:	Nothing
extern inline void SplitOnCharacter(const ZKSTD::string& in_str, 
									ZKSTD::vector<ZKSTD::string>& out_vector,
									const char in_split = ';')
{	
	size_t i = 0;
	size_t j = 0;
	do{
		j = in_str.find_first_of(in_split, i);
		if(j == ZKSTD::string::npos) {
			j = in_str.length();
		}

		const ZKSTD::string strTemp(in_str.substr(i, j - i));
		if(strTemp.length()){
			out_vector.push_back(strTemp);
		}
		i = j + 1;
	}while(j != in_str.length());
}

// PURPOSE:	converts hex string to actual hex value
// REQUIRE:	non null input parm
// PROMISE: Return 0 if non hex string
// THROWS:	Nothing
// EXAMPLE: See Test12
extern inline long ascitohex(const char *in_s)
{
	ZKASSERT(in_s);
	return strtol(in_s, NULL, 16);
}

// PURPOSE:	Replace a substring from a string with another substring (case sensitive)
// REQUIRE:	
// PROMISE: 
// THROWS:	Nothing
// EXAMPLE: See Test14
extern inline ZKSTD::string replace_all(
				   const ZKSTD::string& in_s,
				   const ZKSTD::string& from,
				   const ZKSTD::string& to)
{

	ZKSTD::string buf = in_s;
	ZKSTD::string::size_type pos = 0;
	while((pos = buf.find(from, pos)) != ZKSTD::string::npos)
	{
		buf.replace(pos, from.size(), to);
		++pos;
	}
	return buf;
}

// PURPOSE: Insert a string in front of a string.
//			Works on a vector of string (batch job)
// REQUIRE: A vector of string to add the string in front of.
//			A vector of string to receive the modified strings.
// PROMISE: The output vector will be cleared.
// EXAMPLE: See Test22
extern inline void InsertInFront(
	const ZKSTD::vector<ZKSTD::string>& in_rOriginalStringVector,
	ZKSTD::vector<ZKSTD::string>& out_rModifiedStringVector,
	const ZKSTD::string& in_rStringToInsert = "__" )
{
	out_rModifiedStringVector.clear();
	
	ZKSTD::vector<ZKSTD::string>::const_iterator it = in_rOriginalStringVector.begin();
	for( ; it != in_rOriginalStringVector.end(); ++it )
		out_rModifiedStringVector.push_back( in_rStringToInsert + (*it) );

	ZKASSERT( in_rOriginalStringVector.size() == out_rModifiedStringVector.size() );
}

// PURPOSE: To find "\r\n" or "\n" in a string
// PROMISE: out_rstrALine returns with the contents of io_rstrWorkbuf when a line is not yet found
//			out_rstrALine is returned with the CRLF still attached when it is found
// REQUIRE: Call this function repeatedly with incoming buffers.  It will succeed each time 
//			a complete line is collected.  Output line contains CRLF.
//			This is not thread safe
extern inline bool LookForLine( 
	const char* in_pszBuffer, 
	const int in_nBufferSize, 
	ZKSTD::string& out_rstrALine, 
	ZKSTD::string& io_rstrWorkbuf ) throw()
{
	ZKASSERT( ZKIsValidString(in_pszBuffer, in_nBufferSize) );

	ZKASSERT( ZKIsValidAddress(in_pszBuffer, in_nBufferSize) );


	bool bReturnValue = false;
	
	try{

		io_rstrWorkbuf.append(in_pszBuffer, in_nBufferSize);

		out_rstrALine = io_rstrWorkbuf;

		// find the first CR in io_rstrWorkbuf
		const ZKSTD::string::size_type CR_Pos = io_rstrWorkbuf.find_first_of(0x0D/*CR*/, 0);
		if(CR_Pos != ZKSTD::string::npos)
		{
			if(io_rstrWorkbuf.at(CR_Pos+1) == '\n')
			{
				out_rstrALine = io_rstrWorkbuf.substr(0, CR_Pos+2);	// out_rstrALine is a string
				io_rstrWorkbuf.erase(0, io_rstrWorkbuf.size());			// clear io_rstrWorkbuf
				bReturnValue = true;
			}
		}
		// check for just a LF
		else 
		{
			ZKSTD::string::size_type LF_Pos = io_rstrWorkbuf.find_first_of(0x0A, 0);
			if (LF_Pos != ZKSTD::string::npos)
			{
				out_rstrALine = io_rstrWorkbuf.substr(0, LF_Pos);	// out_rstrALine is a string
				io_rstrWorkbuf.erase(0, (LF_Pos+1));		// remove the line from io_rstrWorkbuf
				bReturnValue = true;						// we got a line
			}
		}

	}
	catch(ZKSTD::out_of_range&){
		ZKTRACE0("ZKSTD::string out_of_range caught in ZKS::LookForLine()\n");
	}
		
	return bReturnValue;
}

// PURPOSE:	Find and return the lower case domain from an url
// REQUIRE:	Nothing
// PROMISE: If the url is screwy, this method will do its best to return the most domain-like
//			thing possible.  eg. ExtractDomain("http://devnet/bugfree") == "devnet"
// THROWS:	Never
// EXAMPLE: "http://www.xx.YAHOO.com:8080/punk" should return "YAHOO.com"
//			See test24
extern inline ZKSTD::string ExtractDomain(const ZKSTD::string& in_strUrl)
{
	ZKSTD::string strDomain = in_strUrl;

	ZKSTD::string::size_type pos;

	if (strDomain == "")
	{
		// This is the local computer
		strDomain = ZK_LOCAL_COMPUTER;
	}
	else
	{
		if ((pos = strDomain.find("//")) != strDomain.npos)
		{
			(strDomain.size() > pos + 2)
				? strDomain = strDomain.substr(pos + 2)
				: strDomain = "";
		}

		if ((pos = strDomain.find_first_of(":/")) != strDomain.npos)
		{
			strDomain = strDomain.substr(0, pos);
		}

		if (((pos = strDomain.find_last_of('.')) != strDomain.npos)
			&& ((pos = strDomain.find_last_of('.', pos - 1)) != strDomain.npos))
		{
			strDomain = strDomain.substr(pos + 1);
		}
	}
	return strDomain;
}

// PURPOSE:	Escapes a string to prepare it to be sent as a parameter
//			in an url.
//			Spaces are replaced by '+', letters and digits remain unchanged,
//			and all other characters are replaced by their hex representation.
// REQUIRE:	Nothing
// PROMISE: 
// THROWS:	Never
// EXAMPLE: "Hi there!" will return "Hi+there%21"
//			See Test25
extern inline ZKSTD::string EscapeString(const ZKSTD::string& in_str)
{
	ZKSTD::string strEscaped = in_str;

	ZKSTD::string::size_type i = 0;

	while (i < strEscaped.size())
	{
		char ch = strEscaped.at(i);

		if (ch == ' ')
		{
			strEscaped.at(i) = '+';
		}
		else if ((ch >= '0') && (ch <= '9'))
		{ }
		else if ((ch >= 'a') && (ch <= 'z'))
		{ }
		else if ((ch >= 'A') && (ch <= 'Z'))
		{ }
		else
		{
			char charInHex[3] = { '%', ' ', ' '};

			int firstCh = ch / 16;
			
			charInHex[1] = (firstCh <= 9) ? firstCh + '0' : firstCh - 10 + 'A';

			int secondCh = ch % 16;

			charInHex[2] = (secondCh <= 9) ? secondCh + '0' : secondCh - 10 + 'A';
		
			strEscaped.replace(i, 1, charInHex, 3);
			i += 2;
		}
		++i;
	}
	
	return strEscaped;
}

// PURPOSE:	Unescapes a string (that is typically a parameter in an url).
// REQUIRE:	Nothing
// PROMISE: If the string is screwy eg ("%GG" or "Hi%F" or "Hi%"),
//			"" will be returned
// THROWS:	Never
// EXAMPLE: "Hi+there%21" will return "Hi there!"
//			See Test26
extern inline ZKSTD::string UnEscapeString(const ZKSTD::string& in_str)
{
	ZKSTD::string strUnEscaped = in_str;

	ZKSTD::string::size_type i = 0;

	while (i < strUnEscaped.size())
	{
		char ch = strUnEscaped.at(i);

		if (ch == '+')
		{
			strUnEscaped.at(i) = ' ';
		}
		else if (ch == '%')
		{
			if (strUnEscaped.size() < i + 3)
			{
				strUnEscaped = "";
				break;
			}
			
			char realCh;

			char firstDigit = strUnEscaped.at(i + 1);
			char secondDigit = strUnEscaped.at(i + 2);

			if ((firstDigit >= '0') && (firstDigit <= '9'))
			{
				realCh = 16 * (firstDigit - '0');
			}
			else if ((firstDigit >= 'A') && (firstDigit <= 'F'))
			{
				realCh = 16 * (10 + firstDigit - 'A');
			}
			else
			{
				strUnEscaped = "";
				break;
			}

			if ((secondDigit >= '0') && (secondDigit <= '9'))
			{
				realCh += secondDigit - '0';
			}
			else if ((secondDigit >= 'A') && (secondDigit <= 'F'))
			{
				realCh += 10 + secondDigit - 'A';
			}
			else
			{
				strUnEscaped = "";
				break;
			}

			strUnEscaped.replace(i, 3, &realCh, 1);
		}
		++i;
	}
	
	return strUnEscaped;
}

extern inline void hideFieldInBuffer(char* in_strHeader,
									  size_t in_lenHeader,
									  const char* in_strFieldName)
{
	ZKSTD::string strLower = in_strFieldName;
	ZKSTD::string strUpper = in_strFieldName;
	MakeLower(strLower);
	MakeUpper(strUpper);

	for(int i = 0; i + strlen(in_strFieldName) - 1 < in_lenHeader; i++)
	{
		if(i == 0 || in_strHeader[i - 1] == '\n' ||
				(in_strHeader[i - 1] == '\r' && in_strHeader[i] != '\n') )
		{
			bool bFound = true;
			for(size_t j = 0; j < strlen(in_strFieldName); j++)
			{
				if(in_strHeader[i + j] != strLower[j] && in_strHeader[i + j] != strUpper[j])
				{
					bFound = false;
					break;
				}
			}
			if(bFound)	
			{
				memset(&in_strHeader[i], 'X', strlen(in_strFieldName));
			}
		}
	}
}


// Only use this function for hiding headers from client side
// software! (It doesn't actually remove the headers, just
// changes the field name to "XXXXXXXXXX...")
extern inline void hideHeaderField(char* in_strHeader,
									  size_t in_lenHeader,
									  const char* in_strFieldName)
{
	// Help out our callers...
	
	const char *pEndHeader = strstr(in_strHeader, "\r\n\r\n");
	const char *pEndHeader2 = strstr(in_strHeader, "\n\n");
	const char *pEndHeader3 = strstr(in_strHeader, "\n\r\n");
	if(pEndHeader2 && (!pEndHeader || pEndHeader2 < pEndHeader))
	{
		pEndHeader = pEndHeader2;
	}
	if(pEndHeader3 && (!pEndHeader || pEndHeader3 < pEndHeader))
	{
		pEndHeader = pEndHeader3;
	}

	

	if(pEndHeader)
	{
		if((size_t)(pEndHeader - in_strHeader) < in_lenHeader)
		{
			in_lenHeader = pEndHeader - in_strHeader;
		}
	}

	hideFieldInBuffer(in_strHeader, in_lenHeader, in_strFieldName);
}

extern inline bool extractHeaderField(const char* in_strHeader,
									  size_t in_lenHeader,
									  const char* in_strFieldName,
									  ZKSTD::string& out_strFieldValue)
{
	ZKSTD::string strLower = in_strFieldName;
	ZKSTD::string strUpper = in_strFieldName;
	MakeLower(strLower);
	MakeUpper(strUpper);

	// Help out our callers...
	
	const char *pEndHeader = strstr(in_strHeader, "\r\n\r\n");
	const char *pEndHeader2 = strstr(in_strHeader, "\n\n");
	const char *pEndHeader3 = strstr(in_strHeader, "\n\r\n");
	if(pEndHeader2 && (!pEndHeader || pEndHeader2 < pEndHeader))
	{
		pEndHeader = pEndHeader2;
	}
	if(pEndHeader3 && (!pEndHeader || pEndHeader3 < pEndHeader))
	{
		pEndHeader = pEndHeader3;
	}

	

	if(pEndHeader)
	{
		if((size_t)(pEndHeader - in_strHeader) < in_lenHeader)
		{
			in_lenHeader = pEndHeader - in_strHeader;
		}
	}

	for(int i = 0; i + strlen(in_strFieldName) - 1 < in_lenHeader; i++)
	{
		if(i == 0 || in_strHeader[i - 1] == '\n' ||
				(in_strHeader[i - 1] == '\r' && in_strHeader[i] != '\n') )
		{
			bool bFound = true;
			for(size_t j = 0; j < strlen(in_strFieldName); j++)
			{
				if(in_strHeader[i + j] != strLower[j] && in_strHeader[i + j] != strUpper[j])
				{
					bFound = false;
					break;
				}
			}
			if(bFound)	
			{
				// found it!
				char* pEnd = (char*)memchr(in_strHeader + i, '\n', in_lenHeader - i);
				if(!pEnd)
				{
					pEnd = (char*)memchr(in_strHeader + i, '\r', in_lenHeader - i);
				}

				while(pEnd && (*(pEnd + 1) == ' ' || *(pEnd + 1) == '\t'))
				{
					char *temp = pEnd;
					pEnd = (char*)memchr(pEnd + 1, '\n', in_lenHeader - (pEnd + 1 - in_strHeader) );
					if(!pEnd)
					{	
						pEnd = (char*)memchr(temp + 1, '\r', in_lenHeader - (temp + 1 - in_strHeader));
					}
				}
				
				if(!pEnd)
				{
					pEnd = (char *)in_strHeader + in_lenHeader;
				}
				while(pEnd >= in_strHeader + i && (*pEnd == '\r' || *pEnd == '\n'))
				{
					pEnd--;
				}

				ZKSTD::string strField(in_strHeader + i, pEnd - in_strHeader - i + 1);

				size_t k = strField.find_first_of(':');
				if(k != strField.npos)
				{
					k = strField.find_first_not_of(' ', k + 1);
					if(k != strField.npos)
					{
						out_strFieldValue = strField.substr(k);
					}
					else
					{
						out_strFieldValue = "";
					}
					return true;
				}
			}
		}
	}
	return false;
}

// See test32
extern inline bool
is_palindrome(const ZKSTD::string& in_s)
{
 ZKSTD::string s(in_s);

 ZKSTD::transform(s.begin(), s.end(), s.begin(), tolower);

 return ZKSTD::equal(s.begin(), s.end(), s.rbegin() );
}

// break length is ideal only.
// might actually break further than that.
extern inline void 
breakStringIntoLines( ZKSTD::string& out_str, 
                      unsigned int idealBreakLength = 60 )
{
    if( idealBreakLength > out_str.size()) 
    {
       return;
    }

    unsigned int curpos = 0;
    unsigned int currentStringLength = 0;

    while( curpos < out_str.size() )
    {
        //printf( "****************** %d of %d\n", 
        //   currentStringLength, idealBreakLength );

        if( out_str[ curpos ] == '\n' ) 
        {
            currentStringLength = 0;
        }

        if( out_str[ curpos ] == ' ' &&
            currentStringLength >= idealBreakLength )

        {
            // printf( "******** BREAKING A STRING\n" );
            out_str[ curpos ] = '\n';
            currentStringLength = 0;
        }
        curpos++;
        currentStringLength++;
    }
    // printf( "************ EXITING IN A HEALTHY MANNER\n" );
} 

}		// namespace ZKS
#endif	// _STRING_HELPERS_HPP__
