// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "EsmtpClient.h"
#include "ZkNetHelp.h"
#include <zkBase64.h>
#include <frmailmsg.h>

#define DATA_CMD "DATA\r\n"
#define RSET_CMD "RSET\r\n"

EsmtpClient::EsmtpClient(const ZKSTD::string& in_strFreedomDomain,
							const ZKSTD::string& in_strSenderAddress,
							const ZKSTD::string& in_strPassword,
							const ZkSocketAddress& in_RemoteAddr,
							FrCryptPubKey in_PrivSigningKey) throw (RZkException)
					: m_strFreedomDomain(in_strFreedomDomain)
					, m_strSenderAddress(in_strSenderAddress)
					, m_strPassword(in_strPassword)
					, m_PrivSigningKey(in_PrivSigningKey)
					, m_Socket(NULL)
					, m_bConnected(false)
{
	memcpy(&m_RemoteAddr, &in_RemoteAddr, sizeof(ZkSocketAddress));
}

EsmtpClient::~EsmtpClient()
{
	// Should send quit...
	if(m_Socket)
	{
		ZKVERIFY(zkSocketFree(&m_err, &m_Socket));
	}
}

void EsmtpClient::ConnectToRemoteServer() throw (RZkException)
{
	ZKASSERT(!m_bConnected);
	
	ZKVERIFY(zkSocketNew(&m_err, AF_INET, SOCK_STREAM, IPPROTO_TCP, &m_Socket));

// bug 6711
#ifndef _WIN32
	ZKVERIFY(zkSocketSetOption(&m_err, m_Socket, ZKSOCKET_OPTION_SendTimeout, 15 * 1000));
	ZKVERIFY(zkSocketSetOption(&m_err, m_Socket, ZKSOCKET_OPTION_RecvTimeout, 45 * 1000));
#endif

	ZKSTD::string strResponseLine;

	if(!zkSocketConnectAddr(&m_err, m_Socket, m_RemoteAddr))
	{
		throw RZkException(
				m_err,
				"EsmtpClient::ConnectToRemoteServer()",
				"zkSocketConnectAddr() failed", 
				IDS_FAILED_TO_CONNECT_TO_MAIL_SERVER);
	}

	if(!ZkNetHelp::ReceiveLine(&m_err, m_Socket, strResponseLine))
	{
		ZKASSERT(0);
		throw RZkException(
				m_err,
				"EsmtpClient::ConnectToRemoteServer()",
				"Failed to receive Outgoing Mail Server welcome banner", 
				IDS_FAILED_TO_CONTACT_OUTGOING_MAIL_SERVER);
	}
	
	if(!ZkNetHelp::SmtpResponseIsSuccess(strResponseLine))
	{
		ZKASSERT(0);
		ZKSTD::string strDetail = "The Mail Server returned a negative welcome banner ("
								+ strResponseLine + ").";

		throw RZkException(
				m_err,
				"EsmtpClient::ConnectToRemoteServer()",
				strDetail, 
				IDS_BAD_ESMTP_SERVER_RESPONSE,
				strResponseLine.c_str() );
	}

	// Should send HELO/EHLO!

	char strBase64NymName[128] = {0};

	size_t lenName = m_strSenderAddress.find_first_of('@');
	ZKASSERT(lenName);
	
	ZKVERIFY(zkConvertToBase64(&m_err, (const unsigned char*)m_strSenderAddress.c_str(), (unsigned char*)strBase64NymName, lenName));

	ZKSTD::string strAuthLoginCmd = "AUTH LOGIN ";
	strAuthLoginCmd += strBase64NymName;
	strAuthLoginCmd += "\r\n";
	ZKSTD::string strAuthCert = m_strPassword + "\r\n";

//bug 6666
#ifndef _WIN32	
	if(!zkSocketWriteAll(&m_err, m_Socket, strAuthLoginCmd.c_str(), strAuthLoginCmd.length()))
#else
	if(strAuthLoginCmd.length() != send(m_Socket->descriptor, strAuthLoginCmd.c_str(), strAuthLoginCmd.length(), 0) )
#endif
	{
		ZKASSERT(0);
		throw RZkException(
				m_err,
				"EsmtpClient::ConnectToRemoteServer()",
				"zkSocketWriteAll() failed", 
				IDS_FAILED_TO_CONTACT_OUTGOING_MAIL_SERVER);
	}
	
	if(!ZkNetHelp::ReceiveLine(&m_err, m_Socket, strResponseLine))
	{
		ZKASSERT(0);
		throw RZkException(
				m_err,
				"EsmtpClient::ConnectToRemoteServer()",
				"ZkNetHelp::ReceiveLine() failed", 
				IDS_FAILED_TO_CONTACT_OUTGOING_MAIL_SERVER);
	}
	
	if(!ZkNetHelp::SmtpResponseIsSuccess(strResponseLine))
	{
		ZKASSERT(0);
		ZKSTD::string strDetail = "Received negative response (" + strResponseLine + ")";
		throw RZkException(
				m_err,
				"EsmtpClient::ConnectToRemoteServer()",
				strDetail, 
				IDS_MAIL_AUTHENTICATION_FAILED,
				m_strSenderAddress.substr(0, lenName).c_str());
	}

//bug 6666
#ifndef _WIN32	
	if(!zkSocketWriteAll(&m_err, m_Socket, strAuthCert.c_str(), strAuthCert.length()))
#else
	if(strAuthCert.length() != send(m_Socket->descriptor, strAuthCert.c_str(), strAuthCert.length(), 0))
#endif
	{
		ZKASSERT(0);
		throw RZkException(
				m_err,
				"EsmtpClient::ConnectToRemoteServer()",
				"zkSocketWriteAll() failed", 
				IDS_FAILED_TO_CONTACT_OUTGOING_MAIL_SERVER);
	}
	
	if(!ZkNetHelp::ReceiveLine(&m_err, m_Socket, strResponseLine))
	{
		ZKASSERT(0);
		throw RZkException(
				m_err,
				"EsmtpClient::ConnectToRemoteServer()",
				"ZkNetHelp::ReceiveLine() failed", 
				IDS_FAILED_TO_CONTACT_OUTGOING_MAIL_SERVER);
	}
	
	if(!ZkNetHelp::SmtpResponseIsSuccess(strResponseLine))
	{
		ZKASSERT(0);
		throw DMailAuthException(
				"EsmtpClient::ConnectToRemoteServer()",
				m_strSenderAddress.substr(0, lenName).c_str());
	}

	m_bConnected = true;
}

void EsmtpClient::SendMailFromCmd() throw (RZkException)
{
	ZKSTD::string strMailFromCmd = "MAIL FROM:<" + m_strSenderAddress + ">\r\n";
	ZKSTD::string strResponseLine;

//bug 6666
#ifndef _WIN32	
	if(!zkSocketWriteAll(&m_err, m_Socket, strMailFromCmd.c_str(), strMailFromCmd.length()))
#else
	if(strMailFromCmd.length() != send(m_Socket->descriptor, strMailFromCmd.c_str(), strMailFromCmd.length(), 0))
#endif
	{
		ZKASSERT(0);
		throw RZkException(
				m_err,
				"EsmtpClient::SendMailFromCmd()",
				"zkSocketWriteAll() failed", 
				IDS_FAILED_TO_CONTACT_OUTGOING_MAIL_SERVER);
	}
	
	if(!ZkNetHelp::ReceiveLine(&m_err, m_Socket, strResponseLine))
	{
		ZKASSERT(0);
		throw RZkException(
				m_err,
				"EsmtpClient::SendMailFromCmd()",
				"ZkNetHelp::ReceiveLine() failed", 
				IDS_FAILED_TO_CONTACT_OUTGOING_MAIL_SERVER);
	}
	
	if(!ZkNetHelp::SmtpResponseIsSuccess(strResponseLine))
	{
		ZKASSERT(0);
		size_t lenName = m_strSenderAddress.find("@");
		if(!lenName)
		{
			lenName = m_strSenderAddress.length();
		}
		ZKSTD::string strDetail = "Received negative response (" + strResponseLine + ")";
		throw RZkException(
				m_err,
				"EsmtpClient::SendMailFromCmd()",
				strDetail, 
				IDS_MAIL_AUTHENTICATION_FAILED,
				m_strSenderAddress.substr(0, lenName).c_str());
	}
}

void EsmtpClient::AddRecipient(const ZKSTD::string& in_strRecipient) throw (RZkException)
{
	if(!m_bConnected)
	{
		ConnectToRemoteServer();
	}

	if(m_vecRecipients.size() == 0)
	{	// Before the first recipient, must send MAIL_FROM_CMD
		SendMailFromCmd();
	}
	
	ZKSTD::string strRcptLine = "RCPT TO:<" + in_strRecipient + ">\r\n";

//bug 6666
#ifndef _WIN32	
	if(!zkSocketWriteAll(&m_err, m_Socket, strRcptLine.c_str(), strRcptLine.length()))
#else
	if(strRcptLine.length() != send(m_Socket->descriptor, strRcptLine.c_str(), strRcptLine.length(), 0))
#endif
	{
		ZKASSERT(0);
		throw RZkException(
				m_err,
				"EsmtpClient::AddRecipient()",
				"zkSocketWriteAll() failed", 
				IDS_FAILED_TO_CONTACT_OUTGOING_MAIL_SERVER);
	}
	
	ZKSTD::string strResponseLine;
	
	if(!ZkNetHelp::ReceiveLine(&m_err, m_Socket, strResponseLine))
	{
		ZKASSERT(0);
		throw RZkException(
				m_err,
				"EsmtpClient::AddRecipient()",
				"ZkNetHelp::ReceiveLine() failed", 
				IDS_FAILED_TO_CONTACT_OUTGOING_MAIL_SERVER);
	}
	
	if(!ZkNetHelp::SmtpResponseIsSuccess(strResponseLine))
	{
		ZKASSERT(0);
		ZKSTD::string strDetail = "Recipient " + in_strRecipient + " was "
								+ "rejected by the Freedom Mail Server ("
								+ strResponseLine + ")";
		throw RZkException(
				m_err,
				"EsmtpClient::AddRecipient()",
				strDetail, 
				IDS_MAIL_SERVER_REJECTED_RECIPIENT,
				in_strRecipient.c_str());
	}

	m_vecRecipients.push_back(in_strRecipient);
}

void EsmtpClient::SendMsgBody(const char* in_strMsgBody, const size_t in_lenMsgBody)
{
	ZKASSERT(m_bConnected);
	ZKASSERT(m_vecRecipients.size());

//bug 6666
#ifndef _WIN32	
	if(!zkSocketWriteAll(&m_err, m_Socket, DATA_CMD, strlen(DATA_CMD)))
#else
	if(strlen(DATA_CMD) != send(m_Socket->descriptor, DATA_CMD, strlen(DATA_CMD), 0))
#endif
	{
		ZKASSERT(0);
		throw RZkException(
				m_err,
				"EsmtpClient::SendMsgBody()",
				"zkSocketWriteAll() failed", 
				IDS_FAILED_TO_CONTACT_OUTGOING_MAIL_SERVER);
	}

	ZKSTD::string strResponseLine;

	if(!ZkNetHelp::ReceiveLine(&m_err, m_Socket, strResponseLine))
	{
		ZKASSERT(0);
		throw RZkException(
				m_err,
				"EsmtpClient::SendMsgBody()",
				"ZkNetHelp::ReceiveLine() failed", 
				IDS_FAILED_TO_CONTACT_OUTGOING_MAIL_SERVER);
	}

	if(!ZkNetHelp::SmtpResponseIsSuccess(strResponseLine))
	{	
		ZKSTD::string strDetail = "Freedom Mail Server rejected message ("
								+ strResponseLine + ')';
		ZKASSERT(0);
		throw RZkException(
				m_err,
				"EsmtpClient::SendMsgBody()",
				strDetail, 
				IDS_BAD_ESMTP_SERVER_RESPONSE,
				strResponseLine.c_str() );
	}

	char *strSigFieldName = NULL;
	
	ZKTRACE0("Sending to Freedom ESMTP Server:\r\n");
	ZKVERIFY(frMailMsgGetEnvSigHeaderName(&m_err, (const char**)&strSigFieldName));
	bool bFoundFreedomRecipients = false;
	for(size_t i = 0; i < m_vecRecipients.size(); i++)
	{
		if(m_vecRecipients[i].find(m_strFreedomDomain) != ZKSTD::string::npos)
		{
			bFoundFreedomRecipients = true;
			continue;
		}
		
		ZKASSERT(!bFoundFreedomRecipients);

		if(!m_PrivSigningKey)
		{
			throw RZkException(
					m_err,
					"EsmtpClient::SendMsgBody()",
					"Need the Nym's Private Signature Key to send Internet Mail", 
					IDS_FAILED_TO_CONTACT_OUTGOING_MAIL_SERVER);
		}

		char strSig[FRMAIL_ENV_SIG_SIZE] = {0};
		ZKVERIFY(frMailEnvelopeAddrSign(&m_err, m_PrivSigningKey,
										m_strSenderAddress.c_str(),
										m_vecRecipients[i].c_str(), strSig));

		ZKSTD::string strEnvSigField = strSigFieldName;
		strEnvSigField				+= ": " + m_vecRecipients[i] + " ";
		strEnvSigField				+= strSig;
		strEnvSigField				+= "\r\n";

		ZKTRACE0(strEnvSigField.c_str());

//bug 6666
#ifndef _WIN32	
		if(!zkSocketWriteAll(&m_err, m_Socket, strEnvSigField.c_str(), strEnvSigField.length() ))
#else
		if(strEnvSigField.length() != send(m_Socket->descriptor, strEnvSigField.c_str(), strEnvSigField.length(), 0))
#endif
		{
			ZKASSERT(0);
			throw RZkException(
					m_err,
					"EsmtpClient::SendMsgBody()",
					"zkSocketWriteAll() failed", 
					IDS_FAILED_TO_CONTACT_OUTGOING_MAIL_SERVER);
		}
	}

	ZKTRACE1("%s\r\n.\r\n", ZKSTD::string(in_strMsgBody, in_lenMsgBody).c_str());

//bug 6666
#ifndef _WIN32	
	if(!zkSocketWriteAll(&m_err, m_Socket, in_strMsgBody, in_lenMsgBody))
#else
	if(in_lenMsgBody != send(m_Socket->descriptor, in_strMsgBody, in_lenMsgBody, 0))
#endif
	{
		ZKASSERT(0);
		throw RZkException(
				m_err,
				"EsmtpClient::SendMsgBody()",
				"zkSocketWriteAll() failed", 
				IDS_FAILED_TO_CONTACT_OUTGOING_MAIL_SERVER);
	}

//bug 6666
#ifndef _WIN32	
	if(!zkSocketWriteAll(&m_err, m_Socket, "\r\n.\r\n", 5))
#else
	if(5 != send(m_Socket->descriptor, "\r\n.\r\n", 5, 0))
#endif
	{
		ZKASSERT(0);
		throw RZkException(
				m_err,
				"EsmtpClient::SendMsgBody()",
				"zkSocketWriteAll() failed", 
				IDS_FAILED_TO_CONTACT_OUTGOING_MAIL_SERVER);
	}

	// bug 6711, which only affected Win32, but change should not affect Linux
	// so the ifdef can be avoided.
	// (30 minutes for eventual timeout because bug #1262)
	ZKVERIFY(zkSocketSetOption(NULL, m_Socket, ZKSOCKET_OPTION_RecvTimeout, 30 * 60 * 1000 )); 


	if(!ZkNetHelp::ReceiveLine(&m_err, m_Socket, strResponseLine))
	{
		ZKASSERT(0);
		throw RZkException(
				m_err,
				"EsmtpClient::AddRecipient()",
				"ZkNetHelp::ReceiveLine() failed", 
				IDS_FAILED_TO_CONTACT_OUTGOING_MAIL_SERVER);
	}
	
	if(!ZkNetHelp::SmtpResponseIsSuccess(strResponseLine))
	{
		ZKASSERT(0);
		throw RZkException(
				m_err,
				"EsmtpClient::AddRecipient()",
				"Received negative response", 
				IDS_BAD_ESMTP_SERVER_RESPONSE,
				strResponseLine.c_str() ); // !IDS_MAIL_SERVER_REJECTED_MSG);
	}

	bool bReset = false;

//bug 6666
#ifndef _WIN32	
	if(!zkSocketWriteAll(&m_err, m_Socket, RSET_CMD, strlen(RSET_CMD)))
#else
	if(strlen(RSET_CMD) != send(m_Socket->descriptor, RSET_CMD, strlen(RSET_CMD), 0))
#endif
	{	// This is not fatal. Just close the socket so a new connection
		// will be forced next time.
		ZKASSERT(0);
	}
	else if(!ZkNetHelp::ReceiveLine(&m_err, m_Socket, strResponseLine))
	{	//  ditto
		ZKASSERT(0);
	}
	else if(!ZkNetHelp::SmtpResponseIsSuccess(strResponseLine))
	{	// ..
		ZKASSERT(0);
	}
	else
	{
		bReset = true;
	}

	if(!bReset)
	{
		ZKVERIFY(zkSocketFree(&m_err, &m_Socket));
		m_bConnected = false;
	}

	m_vecRecipients.erase(m_vecRecipients.begin(), m_vecRecipients.end());
}
