/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#if !defined(AFX_ESMTPCLIENT_H__D4F316A6_50B1_11D2_8097_00104B6B4FA4__INCLUDED_)
#define AFX_ESMTPCLIENT_H__D4F316A6_50B1_11D2_8097_00104B6B4FA4__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "inethelp.h"
#include <zksocket.h>
#include <fr_pk.h>
#include <string>
#include <vector>
#include <clientstl.h>
#include <RZkException.h>
#include <DZkError.h>

#ifdef WIN32
#pragma warning ( disable : 4251)
#endif
// How To Use This Class
//
// This class is used to send mail from a nym to both Internet and Nym
// recipients. The class handles authentication, and deals with the ESMTP
// protocol. Unfortunately, since this class is very low on the dependency
// hierarchy, it does not have access to ZkService. Hence, you must pass in
// all the necessary data at construction.
//
// 1)	Construct with the root domain of the Freedom network (ie,
//		"freedom.net" or "beta.freedom.net"), the email address of the sending
//		Nym, the Nym's email password (base-64 encoded Mail cert), the address
//		of the NMTA, and optionally the Nym's private signature key. The
//		private signature key is required if sending mail to Internet users.
//
// 2)	Call AddRecipient() for each recipient of the message. Remember that
//		multiple Internet recipients are valid for a single mail, but only
//		one Nym may receive a given mail.
//
// 3)	Call SendMsgBody() to send the message. If the message is for a Nym,
//		the message must already be encrypted and include the bounce message.
//		If the message is destined for Internet recipients, then you DO NOT
//		need to add the signatures for each of the recipients. EsmtpClient
//		will do this for you.
//
// 4)	If you wish to send another message, simply go back to step (2). The
//		EsmtpClient automatically resets its state.

class INETHELP_DLL EsmtpClient
{
public:
	// PURPOSE: Constructor
	// REQUIRE:	in_PrivSigningKey is needed if you are sending to internet
	//			user(s)
	// PROMISE: Does nothing but store the data.
	EsmtpClient(const ZKSTD::string& in_strFreedomDomain,
					const ZKSTD::string& in_strSenderAddress,
					const ZKSTD::string& in_strPassword,
					const ZkSocketAddress& in_RemoteAddr,
					FrCryptPubKey in_PrivSigningKey = NULL) throw (RZkException);

	// PURPOSE: Clean up (disconnect from remote server if necessary)
	// REQUIRE: Nada.
	// PROMISE: Frees all resources, disconnects cleanly from server, if
	//			needed.
	~EsmtpClient();

	// PURPOSE: Send a RCPT TO:<...> line to the server. If the address is
	//			not in the freedom network, will store it and add an
	//			"X-Freedom-Enveloppe-Sig" header to the next message. On first call,
	//			will make connection to server and authenticate
	// REQUIRE:	User/Pass combo from Constructor is valid, otherwise will throw.
	// PROMISE:	The FormattedString of the exception will be suitable to return
	//			to an MUA (by prepending "-ERR ").
	void AddRecipient(const ZKSTD::string& in_strRecipient)  throw (RZkException);

	// PURPOSE: Send the actual body of the message (should already be
	//			encrypted!). If appropriate, add "X-Freedom-Enveloppe-Sig"
	//			headers for Internet recipients.
	// REQUIRE: At least one recipient added. If any call to AddRecipient fails
	//			(ie, even if you have one success and one failure) you must
	//			start again.
	// PROMISE: Sends the message, then RSET's the connection to prepare for
	//			another msg.
	void SendMsgBody(const char* in_strMsgBody, const size_t in_lenMsgBody) throw (RZkException);

private:
	void ConnectToRemoteServer() throw (RZkException);
	void SendMailFromCmd() throw (RZkException);
		
	ZKSTD::string m_strFreedomDomain;
	ZKSTD::string m_strSenderAddress;
	ZKSTD::string m_strPassword;

	FrCryptPubKey m_PrivSigningKey;

	ZKSTD::vector<ZKSTD::string> m_vecRecipients;

	ZkSocketAddress m_RemoteAddr;
	ZkSocket m_Socket;

	DZkError m_err;
	bool m_bConnected;
};

#endif // !defined(AFX_ESMTPCLIENT_H__D4F316A6_50B1_11D2_8097_00104B6B4FA4__INCLUDED_)
