// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "HttpHelper.h"
#include <zksocket.h>
#include <string>
#include <sstream>
#include <zkList.h>
#include <zkAlloc.h>
#include <zkCStr.h>

#define MAX_HOSTNAME_LENGTH 128
#define MAX_URL_LENGTH 1024
#define MAX_HEADER_SIZE 2048

#ifndef WIN32
#include <sys/time.h>
#endif

bool HttpHelper::GetRequest(ZkError in_err, const char* in_strURL, char** out_ppResponse,
							unsigned int* out_pResponseLen,	bool* in_pbCancel, int in_TimeoutSec)
{
	return Request( in_err, in_strURL, NULL, out_ppResponse, 
					out_pResponseLen, in_pbCancel, in_TimeoutSec);
}

bool HttpHelper::PostRequest(ZkError in_err, const char* in_strURL, const char *in_strPostBody,
					char** out_ppResponse, unsigned int* out_pResponseLen, bool* in_pbCancel, 
					int in_TimeoutSec)
{
	if (!in_strPostBody)
	{
		ZKASSERT(false);
		return false;
	}
	return Request( in_err, in_strURL, in_strPostBody, out_ppResponse, 
					out_pResponseLen, in_pbCancel, in_TimeoutSec);
}


bool HttpHelper::Request(ZkError in_err, const char* in_strURL, const char *in_strPostBody,
						 char** out_ppResponse, unsigned int* out_pResponseLen, bool* in_pbCancel, 
						 int in_TimeoutSec)
{
	if (!out_ppResponse || !in_strURL)
	{
		ZKASSERT(0);
		return false;
	}

	ZkSocket socket = NULL;
	ZkList listSockets = NULL;

	char strHost[256];
	char strResource[1024];
	int nPort;

	char strHostAndPort[256];
	bool ReturnValue = true;

	bool bDummy = false;

	if(!in_pbCancel)
	{
		in_pbCancel = &bDummy;
	}
	unsigned int temp;
	if (!out_pResponseLen)
	{
		out_pResponseLen = &temp;
	}
	*out_pResponseLen = 0;
	
	if(strlen(in_strURL) > MAX_URL_LENGTH)
	{
		ReturnValue = false;
		ZKASSERT(0);
	}
	else if(!SplitURL(in_strURL, &nPort, strHost, strResource))
	{
		ReturnValue = false;
		ZKASSERT(0);
	}
	else
	{
		snprintf(strHostAndPort, sizeof(strHostAndPort),
                         "%s:%d", strHost, nPort);
	}

	if(!ReturnValue)
	{
		// ...
	}
	else if(!zkSocketNew(in_err, AF_INET, SOCK_STREAM, IPPROTO_TCP, &socket) )
	{
		ReturnValue = false;
		ZKASSERT(0);
	}
	else if(!zkSocketSetOption(in_err, socket, ZKSOCKET_OPTION_SendTimeout, 15 * 1000) )
	{
		ReturnValue = false;
		ZKASSERT(0);
	}
	else if(!zkSocketSetOption(in_err, socket, ZKSOCKET_OPTION_RecvTimeout, 45 * 1000) )
	{
		ReturnValue = false;
		ZKASSERT(0);
	}
	else if( !zkSocketConnect(in_err, socket, strHostAndPort) )
	{
		ReturnValue = false;
	}
	else if( !zkListNew(in_err, &listSockets) )
	{
		ZKASSERT(0);
		ReturnValue = false;
	}
	else if(!zkListSetDataFreeFunction(in_err, listSockets, NULL, NULL))
	{
		ZKASSERT(0);
		ReturnValue = false;
	}
	else if(!zkListInsert(in_err, listSockets, socket, LIST_TAIL))
	{
		ZKASSERT(0);
		ReturnValue = false;
	}
	else
	{
		ZKSTD::ostringstream ossRequestHeader;

		if(in_strPostBody)
		{
			ossRequestHeader << "POST ";
		}
		else
		{
			ossRequestHeader << "GET ";
		}

		ossRequestHeader	<< strResource << " HTTP/1.0\r\n"
							<< "Host: " << strHost << "\r\n";
		
		if(in_strPostBody)
		{
			ossRequestHeader << "Content-Length: " << strlen(in_strPostBody) << "\r\n";
		}

		ossRequestHeader << "\r\n";


		if(!zkSocketWriteAll(in_err, socket, ossRequestHeader.str().c_str(), ossRequestHeader.str().length()))
		{
			ReturnValue = false;
			ZKASSERT(0);
		}
		else if(in_strPostBody)
		{
			if(*in_pbCancel)
			{
				ReturnValue = false;
			}
			else if(!zkSocketWriteAll(in_err, socket, in_strPostBody, strlen(in_strPostBody)))
			{
				ReturnValue = false;
				ZKASSERT(0);
			}
		}
	}
	
	long nBytesReceived = 0;
	long nContentLength = 0;
	char* pBuffer = NULL;
	if (!zkMemAlloc(in_err, (void**)&pBuffer, MAX_HEADER_SIZE))
	{
		ZKASSERT(false);
		ReturnValue = false;
	}
	size_t lenBuffer = MAX_HEADER_SIZE;

	while(ReturnValue)
	{
		
		struct timeval timeout = {0};
		timeout.tv_sec = in_TimeoutSec;
		int iSocket = -1;

		if(*in_pbCancel)
		{
			ReturnValue = false;
		}
        // XXX last param is close signal for linux. Do we need one here?
		else if(!zkSocketWaitReadableOrClosed(in_err, listSockets, &timeout, &iSocket, -1))
		{
			if(!nContentLength ||
				(nContentLength != -1 && nBytesReceived < nContentLength))
			{
				ReturnValue = false;
			}
			break;
		}

		// The WaitReadableOrClosed timed out
		else if(iSocket == -1)
		{
			// The connection is still good, just run through the loop again.
		}
		else if(iSocket == 0)
		{
			size_t nBytesRead = 0;

			if(!zkSocketRead(in_err, socket, pBuffer + nBytesReceived, lenBuffer - 1  - nBytesReceived, &nBytesRead))
			{
				// if there was no Content-Length, we assume this is the end of transmission
				if (nContentLength != -1)
				{
					ZKASSERT(0);
					ReturnValue = false;
				}
				break;
			}
			else
			{
				nBytesReceived += nBytesRead;
				pBuffer[nBytesReceived] = 0;
				if((nContentLength != 0) && ((size_t) nBytesReceived + 1 == lenBuffer))
				{
					if(!zkMemRealloc(in_err, (void**)&pBuffer, lenBuffer * 2))
					{
						ZKASSERT(0);
						ReturnValue = false;
					}
					lenBuffer *= 2;
				}
				if(nContentLength == 0)
				{
					if(!ResponseIsSuccess(pBuffer))
					{
						ZKASSERT(0);
						ReturnValue = false;
						break;
					}
					else if(strstr(pBuffer, "\r\n\r\n"))
					{
						if(!(nContentLength = GetContentLength(pBuffer)))
						{
							// No "Contennt-Length" field found.
							nContentLength = -1;
							char *pBody = strstr(pBuffer, "\r\n\r\n") + 4;
							nBytesReceived -= (pBody - pBuffer);
							memmove(pBuffer, pBody, nBytesReceived);
						}
						else
						{
							lenBuffer = nContentLength + 5;
							char *tempBuf = NULL;
							if (!zkMemAlloc(in_err, (void**)&tempBuf, lenBuffer))
							{
								ZKASSERT(false);
								ReturnValue = false;
								break;
							}
							memcpy(tempBuf, strstr(pBuffer, "\r\n\r\n") + 4, lenBuffer); 
							tempBuf[lenBuffer - 1] = 0;

							nBytesReceived = lenBuffer;
							zkMemFree(in_err, (void**)&pBuffer);
							pBuffer = tempBuf;
						}
					}
				}
			}
		}
		else
		{
			ZKASSERT(0);
			ReturnValue = false;
		}

		if(nContentLength > 0 && (nBytesReceived >= nContentLength))
		{
			break;
		}

	}

	if(!ReturnValue)
	{
		zkMemFree(in_err, (void**)&pBuffer);
	}
	else
	{
		// caller owns the buffer.
		*out_ppResponse = pBuffer;
		*out_pResponseLen = nBytesReceived;
	}

	if(socket)
	{
		zkSocketFree(in_err, &socket);
	}
	if(listSockets)
	{
		zkListFree(in_err, &listSockets);
	}

	return ReturnValue;
}

size_t HttpHelper::GetContentLength(const char* in_strResponse)
{
	const char* buffer = in_strResponse;

	for(unsigned int i = 0; i + 15 < strlen(in_strResponse); ++i)
	{
		if( (
			(*(buffer + i) == '\n') &&
			(*(buffer + i + 1) == 'C') ||  (*(buffer + i + 1) == 'c')) &&
			((*(buffer + i + 2) == 'O') || (*(buffer + (i + 2)) == 'o')) &&
			((*(buffer + i + 3) == 'N') || (*(buffer + (i + 3)) == 'n')) &&
			((*(buffer + i + 4) == 'T') || (*(buffer + (i + 4)) == 't')) &&
			((*(buffer + (i + 5)) == 'E') || (*(buffer + (i + 5)) == 'e')) &&
			((*(buffer + (i + 6)) == 'N') || (*(buffer + (i + 6)) == 'n')) &&
			((*(buffer + (i + 7)) == 'T') || (*(buffer + (i + 7)) == 't')) &&
			((*(buffer + (i + 8)) == '-') &&
			((*(buffer + (i + 9)) == 'L') || (*(buffer + (i + 9)) == 'l')) &&
			((*(buffer + (i + 10)) == 'E') || (*(buffer + (i + 10)) == 'e')) &&
			((*(buffer + (i + 11)) == 'N') || (*(buffer + (i + 11)) == 'n')) &&
			((*(buffer + (i + 12)) == 'G') || (*(buffer + (i + 12)) == 'g')) &&
			((*(buffer + (i + 13)) == 'T') || (*(buffer + (i + 13)) == 't')) &&
			((*(buffer + (i + 14)) == 'H') || (*(buffer + (i + 14)) == 'h'))
			))
		{
			char* pEnd = (char*)strchr(buffer + i, '\r');

			if(pEnd)
			{
				ZKSTD::string strField((buffer + i + 1), pEnd - buffer - i - 1);

				size_t j = strField.find_first_of(':');
				if(j != strField.npos)
				{
					j = strField.find_first_not_of(' ', j + 1);
					ZKSTD::string strValue(strField.substr(j));
					return atoi(strValue.c_str());
				}

			}

		}
	}
	return 0;
}

bool HttpHelper::ResponseIsSuccess(const char* in_strResponse)
{
	char *endline = strstr(in_strResponse, "\r\n");
	char *twohundred = strstr(in_strResponse, "200");
	if(endline && twohundred && twohundred < endline)
	{
		return true;
	}

	return false;
}

bool HttpHelper::SplitURL (const char * in_strUrl, int* out_Port, char *out_strHostName, char *out_strResource )
{
	const char * host = NULL;
	const char * port = NULL;
	const char * data = NULL;
	int end = 0; /* End offset for an extracted string from "data" */
	int i = 0; /* Loop index */

	ZKASSERT(in_strUrl);
	ZKASSERT(out_Port);
	ZKASSERT(out_strHostName);
	ZKASSERT(out_strResource);

	*out_Port = 80;
	if ((host = strstr(in_strUrl,"://")) == NULL)
	{
		host = in_strUrl;
	}
	else
	{
		host += 3;
	}

	while (host[i] != (char) 0)
	{
		if (host[i] == ':')
		{
			end = i;
			i++;
			port = &(host[i]);
			i = 0;
			*out_Port = 0;
			while (port[i] != (char) 0)
			{
				if (port[i] == '/') 
				{
					data = &(port[i]);
					break;
				}
				else if (isdigit(port[i]))
				{
					/* Warning: This is not very portable.. fix this */
					*out_Port = *out_Port * 10 + (port[i] - '0');
				}
				else
				{
					/* character was not a digit */
					return false;
				}
				i++;
			}
			break;
		}
		else if (host[i] == '/')
		{
			end = i;
			data = &(host[i]);
			break;
		} 
		i++;
		end = i;
	}
	if (end > MAX_HOSTNAME_LENGTH)
	{
		end = MAX_HOSTNAME_LENGTH;
	}
	strncpy(out_strHostName, host, end);
	out_strHostName[end] = (char) 0;
	if (data)
	{
		strncpy(out_strResource,data,MAX_URL_LENGTH);
		out_strResource[MAX_URL_LENGTH - 1] = (char) 0;
	}
	else
	{
		zkCStrlcpy(out_strResource,"/", MAX_URL_LENGTH);
	}
	return true;
}
