// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "PacketHelper.h"
#include "ZkNetMisc.h"		// for various string representation helpers of protocols
#include "PortProtocols.h"	// for various protocols and attacks
#include "ProtocolInfo.h"

PacketHelper::PacketHelper( 
		const int in_proto, 
		const unsigned int in_srcip,
		const unsigned int in_dstip,
		const uint8_t in_type, 
		const uint8_t in_code) 
	: m_protocol(in_proto)
	, m_srcip(in_srcip)
	, m_dstip(in_dstip)
	, m_Icmptype(in_type)
	, m_Icmpcode(in_code)
	, m_srcport(0)
	, m_dstport(0)
	, m_datetime(0)
{
}

PacketHelper::PacketHelper( 
		const int in_proto,
		const unsigned int in_dstip,
		const unsigned short in_dstport)
	: m_protocol(in_proto)
	, m_srcip(0)
	, m_dstip(in_dstip)
	, m_Icmptype(ZK_ICMP_MAXTYPE)
	, m_Icmpcode(ZK_ICMP_UNREACH_MAX)
	, m_srcport(0)
	, m_dstport(in_dstport)
	, m_datetime(0)
{
}

PacketHelper::PacketHelper( 
		const int in_proto,
		const unsigned int in_srcip,
		const unsigned int in_dstip,
		const unsigned short in_srcport,
		const unsigned short in_dstport)
	: m_protocol(in_proto)
	, m_srcip(in_srcip)
	, m_dstip(in_dstip)
	, m_Icmptype(ZK_ICMP_MAXTYPE)
	, m_Icmpcode(ZK_ICMP_UNREACH_MAX)
	, m_srcport(in_srcport)
	, m_dstport(in_dstport)
	, m_datetime(0)
{
}

PacketHelper::PacketHelper(const PacketHelper& in_pkh)
:	m_protocol(in_pkh.m_protocol),
	m_srcip(in_pkh.m_srcip),
	m_dstip(in_pkh.m_dstip),
	m_Icmptype(in_pkh.m_Icmptype),
	m_Icmpcode(in_pkh.m_Icmpcode),
	m_srcport(in_pkh.m_srcport),
	m_dstport(in_pkh.m_dstport),
	m_datetime(in_pkh.m_datetime)
{
}

PacketHelper& 
PacketHelper::operator=(const PacketHelper& in_rhs)
{
	if(this == &in_rhs) return *this; // address equality

	m_protocol = in_rhs.m_protocol;
	m_srcip    = in_rhs.m_srcip;
	m_dstip    = in_rhs.m_dstip;
	m_Icmptype = in_rhs.m_Icmptype;
	m_Icmpcode = in_rhs.m_Icmpcode;
	m_srcport  = in_rhs.m_srcport;
	m_dstport  = in_rhs.m_dstport;
	m_datetime = in_rhs.m_datetime;

	return *this;
}

int
PacketHelper::GetProtocol() const 
{
	return m_protocol;
}


ZKSTD::string 
PacketHelper::GetProtocolDesc_str() const 
{
	return ZKS::GetProtocolInfo(m_protocol);
}


ZKSTD::string
PacketHelper::GetProtocolPacketDesc_str() const 
{
	switch(m_protocol)
	{
	case IPPROTO_ICMP:
		{
			return ZKS::GetIcmpMessageType( m_Icmptype );
		}
	};
	return "";
}

// ICMP redirects are generated by routers, not by hosts
// Redirects are intended to be used by hosts, not routers
ZKSTD::string
PacketHelper::GetAdditionalPacketDesc_DestPort_str() const 
{
	switch(m_protocol)
	{
	case IPPROTO_ICMP:
		{
			switch(m_Icmptype)
			{
			case ZK_ICMP_UNREACH:
				return ZKS::GetIcmpUnreachCode(m_Icmpcode);
			case ZK_ICMP_REDIRECT:
				return ZKS::GetIcmpRedirectCode(m_Icmpcode);
			case ZK_ICMP_TIMXCEED:
				return ZKS::GetIcmpTimeExceededCode(m_Icmpcode);
			case ZK_ICMP_PARAMPROB:
				return ZKS::GetIcmpParamProbCode(m_Icmpcode);
			};
		};
		break;

	case IPPROTO_TCP:
	case IPPROTO_UDP:
		{
			return ProtocolInfo::theProtocolInfo().GetPortService_str(m_protocol, m_dstport);
		}
		break;

	};

	return "";
}

ZKSTD::string
PacketHelper::GetPossibleTrojans_DestPort_str() const
{
	return ProtocolInfo::theProtocolInfo().GetTrojans_str(m_dstport) ;
}

ZKSTD::string
PacketHelper::GetSourceIp_str() const 
{
	return ZKS::Cvt_int32toString(m_srcip);
}

ZKSTD::string
PacketHelper::GetDestinationIp_str() const
{
	return ZKS::Cvt_int32toString(m_dstip);
}

unsigned int	
PacketHelper::GetDestinationIp() const
{
	return m_dstip;
}

unsigned int	
PacketHelper::GetSourceIp() const
{
	return m_srcip;
}

unsigned short
PacketHelper::GetSourcePort() const
{
	return m_srcport;
}


unsigned short	
PacketHelper::GetDestinationPort() const
{
	return m_dstport;
}

uint8_t
PacketHelper::GetIcmpType() const
{
	return m_Icmptype;
}

ZKSTD::string
PacketHelper::GetAdditionalPacketDesc_SrcPort_str() const 
{

	switch(m_protocol)
	{
	case IPPROTO_ICMP:
		{
			switch(m_Icmptype)
			{
			case ZK_ICMP_UNREACH:
				return ZKS::GetIcmpUnreachCode(m_Icmpcode);
			case ZK_ICMP_REDIRECT:
				return ZKS::GetIcmpRedirectCode(m_Icmpcode);
			case ZK_ICMP_TIMXCEED:
				return ZKS::GetIcmpTimeExceededCode(m_Icmpcode);
			case ZK_ICMP_PARAMPROB:
				return ZKS::GetIcmpParamProbCode(m_Icmpcode);
			};
		};
		break;

	case IPPROTO_TCP:
	case IPPROTO_UDP:
		{
			return ProtocolInfo::theProtocolInfo().GetPortService_str(m_protocol, m_srcport);
		}
		break;

	};

	return "";
}

time_t
PacketHelper::GetDateTime()
{
	return m_datetime;
}

void PacketHelper::InitDateTime()
{
	::time(&m_datetime);
}
