// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include <string.h>
#include "RDnsMessage.h"
#include <functional>
#include <algorithm>


//---------------------------------------------------------------------------
// Struct DnsHeader
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//
bool RDnsMessage::DnsHeader::SanityCheck()
{
    // XXX ToDo: implement
    return ZK_TRUE;
}


//---------------------------------------------------------------------------
// Class DnsQuestion
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//
void RDnsMessage::DnsQuestion::Deserialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException )
{
    m_qname = RDnsMessage::EncodedDomainToString( buf, offset, length );
    if ( offset + sizeof( uint16_t ) * 2 > length ) 
    {
        throw RZkException(
            "RDnsMessage::DnsQuestion::Deserialize", 
            "Buffer overrun", 
            IDS_DEFAULT_ERROR);
    }
    // Hmm, this could be misaligned access?
    Unpack(m_qtype, buf, offset);
    Unpack(m_qclass, buf, offset);
}

//---------------------------------------------------------------------------
//
void RDnsMessage::DnsQuestion::Serialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException )
{
    RDnsMessage::StringToEncodedDomain( m_qname, buf, offset, length );
    if ( offset + sizeof( uint16_t ) * 2 > length ) 
    {
        throw RZkException(
            "RDnsMessage::DnsQuestion::Serialize", 
            "Buffer overrun", 
            IDS_DEFAULT_ERROR);
    }
    // Hmm, this could be misaligned access methinks.
    Pack(m_qtype, buf, offset);
    Pack(m_qclass, buf, offset);
}

//---------------------------------------------------------------------------
// Class DnsResourceRecord
//---------------------------------------------------------------------------

//--------------------------------------------------------------------------
//

// This is sort of class hierarchy incest. The copy ctor needs to have some intelligence about
// copying rdata since we do a deep copy. If we did a shallow copy life would be easy.
RDnsMessage::DnsResourceRecord::DnsResourceRecord( const RDnsMessage::DnsResourceRecord& peer )
   : m_rdata( NULL )
{
    SetName( peer.GetName() );
    SetClass( peer.GetClass() );
    SetTtl( peer.GetTtl() );

    SetType( peer.GetType() );

    // invoke assignment operator to copy data between rdata chunks
    //   the const_cast is to smoke the const (i.e. we're cheating here)
    if ( const_cast<DnsResourceRecord &>(peer).GetResourceData() ) 
    {
        // SetType above has constructed this object for us already.
        // Our type should now match our peer's
        switch ( GetType() )
        {
        case RDnsMessage::MX:
            *dynamic_cast<MXResourceData *>(m_rdata) = *dynamic_cast<MXResourceData *>(peer.m_rdata);
            break;
        case RDnsMessage::NS:
            *dynamic_cast<NSResourceData *>(m_rdata) = *dynamic_cast<NSResourceData *>(peer.m_rdata);
            break;
        case RDnsMessage::SOA:
            *dynamic_cast<SOAResourceData *>(m_rdata) = *dynamic_cast<SOAResourceData *>(peer.m_rdata);
            break;
        case RDnsMessage::A:
            *dynamic_cast<AResourceData *>(m_rdata) = *dynamic_cast<AResourceData *>(peer.m_rdata);
            break;
        default:
            *dynamic_cast<FooResourceData *>(m_rdata) = *dynamic_cast<FooResourceData *>(peer.m_rdata);
            break;
        }
        m_rdlength = peer.GetRDLength();
    }
}

//---------------------------------------------------------------------------
//

void RDnsMessage::DnsResourceRecord::SetType( const uint16_t in_type )
{
    if ( m_rdata ) delete m_rdata;
    m_type = in_type;
    switch( m_type )
    {
        case RDnsMessage::MX:
            m_rdata = new MXResourceData( this );
            break;
        case RDnsMessage::NS:
            m_rdata = new NSResourceData( this );
            break;
        case RDnsMessage::SOA:
            m_rdata = new SOAResourceData( this );
            break;
        case RDnsMessage::A:
            m_rdata = new AResourceData( this );
            break;
        default:
            m_rdata = new FooResourceData( this );
            break;
    };
}


//---------------------------------------------------------------------------
//
void RDnsMessage::DnsResourceRecord::Deserialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException )
{
    m_name = RDnsMessage::EncodedDomainToString( buf, offset, length );
    if ( offset + sizeof( uint16_t ) * 3 + sizeof(uint32_t) > length ) 
    {
        throw RZkException(
            "RDnsMessage::DnsResourceRecord::Deserialize", 
            "Buffer overrun", 
            IDS_DEFAULT_ERROR);
    }
    // Hmm, this could be misaligned access?
    uint16_t foo;
    Unpack( foo, buf, offset );
    SetType(foo);

    Unpack( m_rrclass, buf, offset );
  
    Unpack( m_ttl, buf, offset );
    m_ttl = ntohl( m_ttl );

    Unpack( m_rdlength, buf, offset );
    m_rdlength = ntohs( m_rdlength );

    m_rdata->Deserialize( buf, offset, length );
}


//---------------------------------------------------------------------------
//
void RDnsMessage::DnsResourceRecord::Serialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException )
{
    RDnsMessage::StringToEncodedDomain( m_name, buf, offset, length );
    char *pRDLength;
    int tmp;

    if ( offset + sizeof( uint16_t ) * 3 + sizeof(uint32_t) > length ) 
    {
        throw RZkException(
            "RDnsMessage::DnsQuestion::Serialize", 
            "Buffer overrun", 
            IDS_DEFAULT_ERROR);
    }
    // Hmm, this could be misaligned access?
    Pack( m_type, buf, offset );

    Pack( m_rrclass, buf, offset );
  
    Pack( htonl(m_ttl), buf, offset );

    if (!m_rdata) {
        throw RZkException(
            "RDnsMessage::DnsResourceRecord::Serialize", 
            "Attempt to serialize incomplete resource record.", 
            IDS_DEFAULT_ERROR);
    }
    // Store a pointer to the length word
    // this is just a trick so we don't need to precalculate the
    // length of the rdata.
    pRDLength = &buf[offset];
    offset += sizeof(uint16_t);
    tmp = offset;
    m_rdata->Serialize( buf, offset, length );
    *(uint16_t *)pRDLength = htons( offset - tmp );    
}


//---------------------------------------------------------------------------
// Class DnsResourceData
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//

RDnsMessage::DnsResourceData::DnsResourceData( DnsResourceRecord *in_parent )
    : parent( in_parent )
{
};


//---------------------------------------------------------------------------
//

RDnsMessage::DnsResourceData::~DnsResourceData()
{
};


//---------------------------------------------------------------------------
//
const RDnsMessage::DnsResourceData& RDnsMessage::DnsResourceData::operator=( const DnsResourceData& )
{
    return *this;
}

//---------------------------------------------------------------------------
// Class MXResourceData
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//
RDnsMessage::MXResourceData::MXResourceData(DnsResourceRecord *in_parent)
    : DnsResourceData( in_parent ) 
{
};

//---------------------------------------------------------------------------
//
RDnsMessage::MXResourceData::MXResourceData(DnsResourceRecord *in_parent, char * buf, size_t &offset, const size_t length)
    : DnsResourceData( in_parent )
{ 
    Deserialize( buf, offset, length ); 
};

//---------------------------------------------------------------------------
//
const RDnsMessage::MXResourceData& RDnsMessage::MXResourceData::operator=( const MXResourceData& peer)
{
    SetExchange( peer.GetExchange() );
    SetPreference( peer.GetPreference() );
    *(static_cast<RDnsMessage::DnsResourceData *>(this)) = static_cast<const RDnsMessage::DnsResourceData &>(peer);
    return (*this);
}


//---------------------------------------------------------------------------
//
RDnsMessage::MXResourceData::~MXResourceData()
{
};

//---------------------------------------------------------------------------
//
void RDnsMessage::MXResourceData::Deserialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException )
{
    ZKASSERT("Deserializing a parented DnsResourceData" && GetParent());
    unsigned int initialoffset = offset;

    if ( offset + sizeof(uint16_t) > length ) 
    {
        throw RZkException(
            "RDnsMessage::MXResourceData::Deserialize", 
            "Buffer overrun", 
            IDS_DEFAULT_ERROR);
    }

    Unpack( m_preference, buf, offset );
    m_preference = ntohs( m_preference );

    m_exchange = RDnsMessage::EncodedDomainToString(buf, offset, length);
    if ( GetParent() && GetParent()->GetRDLength() != offset - initialoffset )
    {
        throw RZkException(
            "RDnsMessage::MXResourceData::Deserialize", 
            "Actual RDLength does not match RR", 
            IDS_DEFAULT_ERROR);
    }
}


//---------------------------------------------------------------------------
//
void RDnsMessage::MXResourceData::Serialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException )
{
    if ( offset + sizeof(uint16_t) > length ) 
    {
        throw RZkException(
            "RDnsMessage::MXResourceData::Serialize", 
            "Buffer overrun", 
            IDS_DEFAULT_ERROR);
    }
    Pack( htons(m_preference), buf, offset );
    
    RDnsMessage::StringToEncodedDomain(m_exchange, buf, offset, length);
}

//---------------------------------------------------------------------------
// Class NSResourceData
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//
RDnsMessage::NSResourceData::NSResourceData(DnsResourceRecord *in_parent)
    : DnsResourceData( in_parent ) 
{
};

//---------------------------------------------------------------------------
//
RDnsMessage::NSResourceData::NSResourceData(DnsResourceRecord *in_parent, char * buf, size_t &offset, const size_t length)
    : DnsResourceData( in_parent )
{ 
    Deserialize( buf, offset, length ); 
};

//---------------------------------------------------------------------------
//
const RDnsMessage::NSResourceData& RDnsMessage::NSResourceData::operator=( const NSResourceData& peer)
{
    SetNameServer( peer.GetNameServer() );
    *(static_cast<RDnsMessage::DnsResourceData *>(this)) = static_cast<const RDnsMessage::DnsResourceData &>(peer);
    return (*this);
}


//---------------------------------------------------------------------------
//
RDnsMessage::NSResourceData::~NSResourceData()
{
};

//---------------------------------------------------------------------------
//
void RDnsMessage::NSResourceData::Deserialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException )
{
    ZKASSERT("Deserializing a parented DnsResourceData" && GetParent());
    unsigned int initialoffset = offset;

    
    m_nameserver = RDnsMessage::EncodedDomainToString( buf, offset, length );
    if ( GetParent() && GetParent()->GetRDLength() != offset - initialoffset )
    {
        throw RZkException(
            "RDnsMessage::NSResourceData::Deserialize", 
            "Actual RDLength does not match RR", 
            IDS_DEFAULT_ERROR);
    }
}


//---------------------------------------------------------------------------
//
void RDnsMessage::NSResourceData::Serialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException )
{
    RDnsMessage::StringToEncodedDomain( m_nameserver, buf, offset, length );
}

//---------------------------------------------------------------------------
// Class SOAResourceData
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//
RDnsMessage::SOAResourceData::SOAResourceData(DnsResourceRecord *in_parent)
    : DnsResourceData( in_parent ) 
{
};

//---------------------------------------------------------------------------
//
RDnsMessage::SOAResourceData::SOAResourceData(DnsResourceRecord *in_parent, char * buf, size_t &offset, const size_t length)
    : DnsResourceData( in_parent )
{ 
    Deserialize( buf, offset, length ); 
};

//---------------------------------------------------------------------------
//
const RDnsMessage::SOAResourceData& RDnsMessage::SOAResourceData::operator=( const SOAResourceData& peer)
{
    SetMName( peer.GetMName() );
    SetRName( peer.GetRName() );
    SetSerial( peer.GetSerial() );
    SetRefresh( peer.GetRefresh() );
    SetRetry( peer.GetRetry() );
    SetExpiry( peer.GetExpiry() );
    SetMinimum( peer.GetMinimum() );

    *(static_cast<RDnsMessage::DnsResourceData *>(this)) = static_cast<const RDnsMessage::DnsResourceData &>(peer);
    return (*this);
}


//---------------------------------------------------------------------------
//
RDnsMessage::SOAResourceData::~SOAResourceData()
{
};

//---------------------------------------------------------------------------
//
void RDnsMessage::SOAResourceData::Deserialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException )
{
    ZKASSERT("Deserializing a parented DnsResourceData" && GetParent());
    unsigned int initialoffset = offset;

    m_mname = RDnsMessage::EncodedDomainToString(buf, offset, length);
    m_rname = RDnsMessage::EncodedDomainToString(buf, offset, length);

    UnpackSafe(m_serial, buf, offset, length);
    m_serial = ntohl( m_serial );

    UnpackSafe(m_refresh, buf, offset, length);
    m_refresh = ntohl( m_refresh );

    UnpackSafe(m_retry, buf, offset, length);
    m_retry = ntohl( m_retry );

    UnpackSafe(m_expiry, buf, offset, length);
    m_expiry = ntohl( m_expiry );

    UnpackSafe(m_minimum, buf, offset, length);
    m_minimum = ntohl( m_minimum );

    if ( GetParent() && GetParent()->GetRDLength() != offset - initialoffset )
    {
        throw RZkException(
            "RDnsMessage::SOAResourceData::Deserialize", 
            "Actual RDLength does not match RR", 
            IDS_DEFAULT_ERROR);
    }
}


//---------------------------------------------------------------------------
//
void RDnsMessage::SOAResourceData::Serialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException )
{
    RDnsMessage::StringToEncodedDomain(m_mname, buf, offset, length);
    RDnsMessage::StringToEncodedDomain(m_rname, buf, offset, length);

    PackSafe( htonl( m_serial ), buf, offset, length);

    PackSafe( htonl( m_refresh ), buf, offset, length);

    PackSafe( htonl( m_retry ), buf, offset, length);

    PackSafe( htonl( m_expiry ), buf, offset, length);

    PackSafe( htonl( m_minimum ), buf, offset, length);
}

//---------------------------------------------------------------------------
// Class AResourceData
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//
RDnsMessage::AResourceData::AResourceData(DnsResourceRecord *in_parent)
    : DnsResourceData( in_parent ) 
{
};

//---------------------------------------------------------------------------
//
RDnsMessage::AResourceData::AResourceData(DnsResourceRecord *in_parent, char * buf, size_t &offset, const size_t length)
    : DnsResourceData( in_parent )
{ 
    Deserialize( buf, offset, length ); 
};

//---------------------------------------------------------------------------
//
const RDnsMessage::AResourceData& RDnsMessage::AResourceData::operator=( const AResourceData& peer)
{
    SetAddress( peer.GetAddress() );

    *(static_cast<RDnsMessage::DnsResourceData *>(this)) = static_cast<const RDnsMessage::DnsResourceData &>(peer);
    return (*this);
}


//---------------------------------------------------------------------------
//
RDnsMessage::AResourceData::~AResourceData()
{
};

//---------------------------------------------------------------------------
//
void RDnsMessage::AResourceData::Deserialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException )
{
    ZKASSERT("Deserializing a parented DnsResourceData" && GetParent());
    unsigned int initialoffset = offset;

    UnpackSafe(m_address, buf, offset, length);

    if ( GetParent() && GetParent()->GetRDLength() != offset - initialoffset )
    {
        throw RZkException(
            "RDnsMessage::SOAResourceData::Deserialize", 
            "Actual RDLength does not match RR", 
            IDS_DEFAULT_ERROR);
    }
}


//---------------------------------------------------------------------------
//
void RDnsMessage::AResourceData::Serialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException )
{
    PackSafe( m_address, buf, offset, length);
}


//---------------------------------------------------------------------------
// Class FooResourceData
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//
RDnsMessage::FooResourceData::FooResourceData(DnsResourceRecord *in_parent)           
    : DnsResourceData( in_parent ) 
   , m_buffer( NULL )
{
};

//---------------------------------------------------------------------------
//
RDnsMessage::FooResourceData::FooResourceData(DnsResourceRecord *in_parent, char * buf, size_t &offset, const size_t length)
    : DnsResourceData( in_parent ) 
   , m_buffer( NULL )
{
    Deserialize( buf, offset, length ); 
};

//---------------------------------------------------------------------------
//

const RDnsMessage::FooResourceData& RDnsMessage::FooResourceData::operator=( const RDnsMessage::FooResourceData& peer )
{
    if (m_buffer) delete[] m_buffer;
    SetBuffer( peer.m_buffer, peer.m_buflength );
    *(static_cast<RDnsMessage::DnsResourceData *>(this)) = static_cast<const RDnsMessage::DnsResourceData &>(peer);
    return (*this);
}

//---------------------------------------------------------------------------
//

void RDnsMessage::FooResourceData::SetBuffer( char *in_buffer, unsigned int in_length )
{ 
    m_buffer = new char[in_length];
    memcpy(m_buffer, in_buffer, in_length);
    m_buflength = in_length;
};


//---------------------------------------------------------------------------
//
void RDnsMessage::FooResourceData::Deserialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException )
{
    ZKASSERT("Deserializing a parented DnsResourceData" && GetParent());
    unsigned int inclen = GetParent()->GetRDLength();
    if ( offset + inclen > length ) 
    {
        throw RZkException(
            "RDnsMessage::FooResourceData::Deserialize", 
            "Buffer overrun", 
            IDS_DEFAULT_ERROR);
    }
    m_buffer = new char[inclen];
    m_buflength = inclen;
    memcpy( m_buffer, &buf[offset], inclen );
    offset += inclen;
}


//---------------------------------------------------------------------------
//
void RDnsMessage::FooResourceData::Serialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException )
{
    if (offset + m_buflength > length)
    {
        throw RZkException(
            "RDnsMessage::FooResourceData::Serialize", 
            "Buffer overrun", 
            IDS_DEFAULT_ERROR);
    }
    memcpy( &buf[offset], m_buffer, m_buflength );
    offset += m_buflength;
}

//---------------------------------------------------------------------------
//

RDnsMessage::FooResourceData::~FooResourceData()
{ 
    if (m_buffer) delete[] m_buffer; 
};


//---------------------------------------------------------------------------
// Class RDnsMessage
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//
RDnsMessage::RDnsMessage()
{
};

//---------------------------------------------------------------------------
//
// Deserialization ctor
RDnsMessage::RDnsMessage( char *data, const size_t length ) throw ( RZkException )
{
    Deserialize( data, length ); 
};


//---------------------------------------------------------------------------
//

RDnsMessage::DnsQuestion *RDnsMessage::GetQuestion( const unsigned int i )
{ 
    if ( i >= m_qdata.size() ) return NULL;
    else return m_qdata[i]; 
} ;

//---------------------------------------------------------------------------
//

void RDnsMessage::AddQuestion( const DnsQuestion &q )
{
    DnsQuestion *q2 = new DnsQuestion(q);
    m_qdata.push_back( q2 );
};

//---------------------------------------------------------------------------
//

RDnsMessage::DnsResourceRecord *RDnsMessage::GetAnswer( const unsigned int i )
{ 
    if ( i >= m_andata.size() ) return NULL;
    else return m_andata[i]; 
} ;

//---------------------------------------------------------------------------
//

void RDnsMessage::AddAnswer( const DnsResourceRecord &rr )
{
    DnsResourceRecord *rr2 = new DnsResourceRecord(rr);
    m_andata.push_back( rr2 );
};

//---------------------------------------------------------------------------
//

RDnsMessage::DnsResourceRecord *RDnsMessage::GetNameServer( const unsigned int i )
{ 
    if ( i >= m_nsdata.size() ) return NULL;
    else return m_nsdata[i]; 
} ;

//---------------------------------------------------------------------------
//

void RDnsMessage::AddNameServer( const RDnsMessage::DnsResourceRecord &rr )
{
    DnsResourceRecord *rr2 = new DnsResourceRecord(rr);
    m_nsdata.push_back( rr2 );
};

//---------------------------------------------------------------------------
//

RDnsMessage::DnsResourceRecord *RDnsMessage::GetAdditional( const unsigned int i )
{ 
    if ( i >= m_ardata.size() ) return NULL;
    else return m_ardata[i]; 
} ;

//---------------------------------------------------------------------------
//

void RDnsMessage::AddAdditional( const RDnsMessage::DnsResourceRecord &rr )
{
    DnsResourceRecord *rr2 = new DnsResourceRecord( rr );
    m_ardata.push_back( rr2 );
};

//---------------------------------------------------------------------------
// 

//STL voodoo, because you can... (probably shouldn't tho')
template <typename T> struct DieMoFo
    : public ZKSTD::unary_function<T*, void>
{
    void operator()(T* t_ptr)
    {
        delete t_ptr;
    }
};

RDnsMessage::~RDnsMessage()
{
    // Free ptr refs.
    ZKSTD::for_each(  m_qdata.begin(),  m_qdata.end(), DieMoFo<DnsQuestion>() );
    ZKSTD::for_each( m_andata.begin(), m_andata.end(), DieMoFo<DnsResourceRecord>() );
    ZKSTD::for_each( m_nsdata.begin(), m_nsdata.end(), DieMoFo<DnsResourceRecord>() );
    ZKSTD::for_each( m_ardata.begin(), m_ardata.end(), DieMoFo<DnsResourceRecord>() );
}

//---------------------------------------------------------------------------
//
void RDnsMessage::Deserialize( char *buf, const size_t length ) throw ( RZkException )
{
    size_t offset = 0;
    unsigned int i;

    if ( length < sizeof(DnsHeader) )
    {
			throw RZkException(
				"RDnsMessage::Deserialize", 
				"Supplied buffer is invalid", 
                IDS_DEFAULT_ERROR);
    }

    memcpy( &this->m_header, buf, sizeof(DnsHeader) );
    offset += sizeof(DnsHeader);

//    header.ToHostByteOrder();

    if ( !m_header.SanityCheck() )
    {
        throw RZkException(
            "RDnsMessage::Deserialize", 
            "Supplied buffer does not contain sane DNS header", 
            IDS_DEFAULT_ERROR);
    }

    // We rely on exceptions thrown by the deserialization constructor of these objects
    // to make this safe for malformed or short buffers.

    // Grok questions
    for ( i = 0; i < m_header.GetQDCount(); i++ ) 
    {
        DnsQuestion *q =  new DnsQuestion( buf, offset, length );
        m_qdata.push_back(q);
    }

    // Grok answers
    for ( i = 0; i < m_header.GetANCount(); i++ ) 
    {
        DnsResourceRecord *rr = new DnsResourceRecord( buf, offset, length );
        m_andata.push_back(rr);
    }

    // Grok name servers
    for ( i = 0; i < m_header.GetNSCount(); i++ ) 
    {
        DnsResourceRecord *rr = new DnsResourceRecord( buf, offset, length );
        m_nsdata.push_back(rr);
    }

    // Grok additional records
    for ( i = 0; i < m_header.GetARCount(); i++ ) 
    {
        DnsResourceRecord *rr = new DnsResourceRecord( buf, offset, length );
        m_ardata.push_back(rr);
    }
    
}

//---------------------------------------------------------------------------
//
size_t RDnsMessage::Serialize( char *buf, const size_t length ) throw ( RZkException )
{
    size_t offset = 0;
    unsigned int i;

    if ( length < sizeof(DnsHeader) )
    {
			throw RZkException(
				"RDnsMessage::Deserialize", 
				"Supplied buffer is invalid", 
                IDS_DEFAULT_ERROR);
    }

    m_header.SetQDCount( m_qdata.size() );
    m_header.SetANCount( m_andata.size() );
    m_header.SetNSCount( m_nsdata.size() );
    m_header.SetARCount( m_ardata.size() );

    memcpy( buf, &this->m_header, sizeof(DnsHeader) );
    offset += sizeof(DnsHeader);    

    // We rely on exceptions thrown by the serialization members of these objects
    // to make this safe for short buffers.

    // Put questions
    for ( i = 0; i < m_header.GetQDCount(); i++ ) 
    {
        m_qdata[i]->Serialize( buf, offset, length );
    }

    // Put answers
    for ( i = 0; i < m_header.GetANCount(); i++ ) 
    {
        m_andata[i]->Serialize( buf, offset, length );
    }

    // Put name servers
    for ( i = 0; i < m_header.GetNSCount(); i++ ) 
    {
        m_nsdata[i]->Serialize( buf, offset, length );
    }

    // Put additional records
    for ( i = 0; i < m_header.GetARCount(); i++ ) 
    {
        m_ardata[i]->Serialize( buf, offset, length );
    }

    return offset;
}


//---------------------------------------------------------------------------
//
// not very symmetrical. I hate returning through parameters tho'
ZKSTD::string RDnsMessage::EncodedDomainToString( char *buf, size_t &offset, const size_t length ) throw ( RZkException )
{
    unsigned int len = 0;
    char *lbuf = (char *)&buf[offset];
    ZKSTD::string rval("");
    bool first = ZK_TRUE, offset_updated = ZK_FALSE;
    
    if ( lbuf[0] == 0 ) {
        throw RZkException(
            "RDnsMessage::EncodedDomainToString", 
            "Domain has root at first byte. Buffer is corrupt.", 
            IDS_DEFAULT_ERROR);
    }

    while ( ZK_TRUE ) {
        len = lbuf[0];
        // C0 are the top two bits used to denote a compressed entry
        if ( len & 0xC0 ) 
        {
            // We hit a compression pointer in the string.
            if ( offset + 1 > length ) {
                throw RZkException(
                    "RDnsMessage::EncodedDomainToString", 
                    "Buffer overrun in compressed domain string", 
                    IDS_DEFAULT_ERROR);
            }
            uint16_t jump = ntohs( *(uint16_t*)&lbuf[0] );
            jump &= ~0xC000;
            if ( jump > length )
            {
                throw RZkException(
                    "RDnsMessage::EncodedDomainToString", 
                    "Buffer overrun in compressed domain string", 
                    IDS_DEFAULT_ERROR);
            }
            if (!offset_updated) {
                offset += lbuf - &buf[offset]; // String thus far.
                offset += 2;
                // prevent offset from getting updated from now on
                // Since we've finsihed with all the data that was inline
                offset_updated = ZK_TRUE;
            }
            lbuf = (char *)&buf[jump];
            continue;
        }
        // Normal (non-compressed) string handling
        if ( len == 0 ) 
        {
            // rewt, er, root
            break;
        }
        if ( offset + len > length ) {
            throw RZkException(
                "RDnsMessage::EncodedDomainToString", 
                "Buffer overrun", 
                IDS_DEFAULT_ERROR);
        }
        if ( !first ) {
            rval.append(".");
        } else {
            first = ZK_FALSE;
        }
        rval.append((char *)&lbuf[1], len);
        lbuf = lbuf + len + 1; // extra one is for length byte
    }
    lbuf++; // Move past root (0) (hmm, root is usually encoded as a '.'
    
    if (!offset_updated) offset += lbuf - &buf[offset];
    return rval;
}


//---------------------------------------------------------------------------
//
void RDnsMessage::StringToEncodedDomain( ZKSTD::string str, char *buf, size_t &offset, const size_t length ) throw ( RZkException )
{
    char cstr[255]; 
    char *lastdot = NULL;

    unsigned int i = 0, lasti = 0;
    char *lbuf = (char *)&buf[offset];

    if ( offset + str.length() + 1 > length )
    {
        throw RZkException(
            "RDnsMessage::StringToEncodedDomain", 
            "Buffer overrun", 
            IDS_DEFAULT_ERROR);
    }

    strncpy( cstr + 1, str.c_str(), 254 ); // truncate silently?
   
    lastdot = (char *)&cstr[0];
    for ( i = 1; true; i++ )
    {
        if ( cstr[i] == '.' || cstr[i] == 0) 
        {
            // The RFC is really unclear on whethre the limit is 255 or 64 here
            // #%@#%@##!!!
            if ( i - lasti > 255 ) // off by one?
            {
                throw RZkException(
                    "RDnsMessage::StringToEncodedDomain", 
                    "Domain label too long ( > 255 )", 
                    IDS_DEFAULT_ERROR);
            }
            if (lastdot) lastdot[0] = (char)(i - lasti - 1);
            lastdot = (char *)&cstr[i];
            lasti = i;
        }
        if ( cstr[i] == 0 ) break;
    }
    cstr[++i] = 0;
    memcpy(lbuf, cstr, i);
    offset += i;
    return;
}

//---------------------------------------------------------------------------
//

// This doesn't work how you think it would since we have no alignment guarantees.
// We need to look for a short which could be aligned in the array starting 
// at buf or buf + 1
// (May not work for big endian)
bool RDnsMessage::ScanBufferForShort( uint16_t in_needle, char *in_haystack, size_t in_haylen )
{
    char hibyte = in_needle >> 8;
    char lobyte = in_needle & 0xFF;
    size_t i = 0;

    for ( i = 1; i < in_haylen; i++ ) 
    {
        if ( in_haystack[i] == lobyte && in_haystack[i-1] == hibyte )
        {
            return ZK_TRUE;
        }
    }
    return ZK_FALSE;
}


//---------------------------------------------------------------------------
//
bool RDnsMessage::ScanBufferForString( char *in_needle, char *in_haystack, size_t in_haylen )
{
    size_t i = 0;

    for ( i = 0; i < in_haylen; i++ ) 
    {
        if ( in_haystack[i] == in_needle[0] )
        {
            if ( !strncmp( in_needle, &in_haystack[i], strlen(in_needle) ) ) return ZK_TRUE;
        }
    }
    return ZK_FALSE;
}
