/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef RDNSMESSAGE_H__
#define RDNSMESSAGE_H__

#include <client.h>	// for ZKSTD
#include "inethelp.h"
#include <zkError.h>
#include <vector>
#include <memory>
#include "RZkException.h"

//---------------------------------------------------------------------------
// Helper for RFC 1035 DNS queries 
//---------------------------------------------------------------------------


// XXX FIXME: No compression support for tx  (is supported for rx)
// XXX FIXME: Only some RRs are parsed, a dummy RR handles the others. More support could be added simply.
// XXX FIXME: No <character-string> (pascal string) support which is used in some RRs

/////
// NOTE: as a general rule, to minimize the amount of byte swapping the following
//       types of data are in the specified byte orders:
//
//    NETWORK ORDER:
//     - Constants
//     - Bitfields
//     - IP Addresses
//
//    HOST ORDER:
//     - All other numeric values.
/////

class INETHELP_DLL RDnsMessage
{

public:
    // Just read the big endian parts if you want this to make sense.
    enum Type {
#if defined(ZK_BIG_ENDIAN)
        A = 1, // Address
        NS = 2, // Authoratative name server
        MD = 3, // Mail destination (obs)
        MF = 4, // Mail forwarder (obs)
        CNAME = 5, // Canonical name 
        SOA = 6, // authority
        MB = 7, // Mailbox domain (x)
        MG = 8, // Mail group member (x)
        MR = 9, // Mail rename domain (x)
        TYPE_NULL = 10, // The guy who wrote this RFC was on crack
        WKS = 11, // Well known service desc
        PTR = 12, // Domain name pointer
        HINFO = 13, // Host info
        MINFO = 14, // Mailbox or mail list info
        MX = 15, // Mail exchange
        TXT = 16 // Text strings
#elif defined(ZK_LITTLE_ENDIAN)
        A = 256, // Address
        NS = 512, // Authoratative name server
        MD = 768, // Mail destination (obs)
        MF = 1024, // Mail forwarder (obs)
        CNAME = 1280, // Canonical name 
        SOA = 1536, // authority
        MB = 1792, // Mailbox domain (x)
        MG = 2048, // Mail group member (x)
        MR = 2304, // Mail rename domain (x)
        TYPE_NULL = 2560, // The guy who wrote this RFC was on crack
        WKS = 2816, // Well known service desc
        PTR = 3072, // Domain name pointer
        HINFO = 3328, // Host info
        MINFO = 3548, // Mailbox or mail list info
        MX = 3840, // Mail exchange
        TXT = 4096 // Text strings
#else
#error "You must define ZK_???_ENDIAN"
#endif       
    };

//---------------------------------------------------------------------------
//

    // Oh, if only you could subclass an enum.
    enum QType {
#if defined(ZK_BIG_ENDIAN)
        AXFR = 252,  // Full zone transfer
        MAILB = 253, // Mailbox related records
        MAILA = 254, // Mail agent RRs
        ALL = 255    // Everything
#elif defined(ZK_LITTLE_ENDIAN)
        AXFR = 64512,  // Full zone transfer
        MAILB = 64768, // Mailbox related records
        MAILA = 65024, // Mail agent RRs
        ALL = 65280    // Everything
#else
#  error "Ok so what are you trying? A pdp endian freedom port? Good luck."
#endif       
    };

//---------------------------------------------------------------------------
//

    enum Class {
#if defined(ZK_BIG_ENDIAN)
        INTERNET = 1, // Internet
        CSNET = 2, // CSNET (obs)
        CHAOS = 3, // CHAOS (obs)
        HESIOD = 4  // Hesiod (wtf?! is a hesiod)
#elif defined(ZK_LITTLE_ENDIAN)
        INTERNET = 256, // Internet
        CSNET = 512, // CSNET (obs)
        CHAOS = 768, // CHAOS (obs)
        HESIOD = 1024  // Hesiod (wtf?! is a hesiod)
#endif       
    };

//---------------------------------------------------------------------------
//
    enum OpCode {
#if defined(ZK_BIG_ENDIAN)
        QUERY = 1, // Query
        IQUERY = 2, // Inverse Query
        STATUS = 3 // Server Status
#elif defined(ZK_LITTLE_ENDIAN)
        QUERY = 256, // Query
        IQUERY = 512, // Inverse Query
        STATUS = 768 // Server Status
#endif       
    };

//---------------------------------------------------------------------------
//
    enum RCode {
#if defined(ZK_BIG_ENDIAN)
        NOERR = 0, // no error condition
        FORMATERR = 1, // formatting error
        SERVFAIL = 2, // server failure
        NAMEERR = 3, // name error
        NOTIMPL = 4, // query type not implemented
        REFUSED = 5 // policy error
#elif defined(ZK_LITTLE_ENDIAN)
        NOERR = 0, // no error condition
        FORMATERR = 256, // formatting error
        SERVFAIL = 512, // server failure
        NAMEERR = 768, // name error
        NOTIMPL = 1024, // query type not implemented
        REFUSED = 1280 // policy error
#endif               
    };


//---------------------------------------------------------------------------
//
    // The selectors and modifiers will make byte order conversion work for you.
    // It may make sense to make this a class and make it's data private
    // I figure the clinet can be wary and just not access the data and the
    // struct declaration makes it clear that this thing really shouldn't
    // have a vtable
#pragma pack(1)    
    typedef struct __tag_DnsHeader 
    {
        uint16_t GetID() const ;
        
        void SetID( const uint16_t in_id ) ;
        
        uint16_t GetQR() const ;
        
        void SetQR( const uint16_t in_qr ) ;
        
        uint16_t GetOpcode() const ;
        
        void SetOpcode( const uint16_t in_op ) ;
        
        uint16_t GetAA() const ;
        
        void SetAA( const uint16_t in_aa ) ;
        
        uint16_t GetTC() const ;
        
        void SetTC( const uint16_t in_tc ) ;

        uint16_t GetRD() const ;
        
        void SetRD( const uint16_t in_rd ) ;
        
        uint16_t GetRA() const ;
        
        void SetRA( const uint16_t in_ra ) ;
        
        uint16_t GetRCode() const ;
        
        void SetRCode( const uint16_t in_rcode ) ;
        
        uint16_t GetQDCount() const ;
        
        void SetQDCount( const uint16_t in_qdcount ) ;
        
        uint16_t GetANCount() const ;
        
        void SetANCount( const uint16_t in_ancount ) ;
        
        uint16_t GetNSCount() const ;
        
        void SetNSCount( const uint16_t in_nscount ) ;
        
        uint16_t GetARCount() const ;
        
        void SetARCount( const uint16_t in_arcount ) ;
        
        // check a header makes sense before we try to deserialize it
        bool SanityCheck() ;

        uint16_t id;
#if defined(ZK_BIG_ENDIAN)
        uint16_t qr:1;
        uint16_t opcode:4;
        uint16_t aa:1;
        uint16_t tc:1;
        uint16_t rd:1;
        uint16_t ra:1
        uint16_t z:3;
        uint16_t rcode:4;
#elif defined(ZK_LITTLE_ENDIAN)
        uint16_t rd:1;
        uint16_t tc:1;
        uint16_t aa:1;
        uint16_t opcode:4;
        uint16_t qr:1;
        uint16_t rcode:4;
        uint16_t z:3;
        uint16_t ra:1;
#endif
        uint16_t qdcount;
        uint16_t ancount;
        uint16_t nscount;
        uint16_t arcount;
        
    } DnsHeader ;
#pragma pack()



//---------------------------------------------------------------------------
//
    class DnsQuestion 
    {
    public:
        DnsQuestion();

        DnsQuestion( char *buf, size_t &offset, const size_t length );

// Use Defaults.
//        DnsQuestion( const DnsQuestion& );
//        const DnsQuestion& operator=( const DnsQuestion& );

        ~DnsQuestion();

        void Deserialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException );

        void Serialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException );

        ZKSTD::string GetName() const;

        void SetName( const ZKSTD::string name );

        uint16_t GetType() const;

        void SetType( const uint16_t type );

        uint16_t GetClass() const;

        void SetClass( const uint16_t in_class );

    private:
        ZKSTD::string m_qname;
        uint16_t m_qtype;
        uint16_t m_qclass;

    } ;


//---------------------------------------------------------------------------
//
    // forward decl for dependancy;
    class DnsResourceRecord;

    // For virtual function fun
    // base class.
    class DnsResourceData
    {
    public:
        // Friendship is for reparenting
        friend class DnsResourceRecord;

        DnsResourceData( DnsResourceRecord *in_parent );

        const DnsResourceData& operator=( const DnsResourceData& );
        
        virtual ~DnsResourceData();
        
        virtual void Deserialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException ) = 0;

        virtual void Serialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException ) = 0;

        DnsResourceRecord *GetParent( void ) const;

    protected:        

        void SetParent( DnsResourceRecord *in_parent );

    private:
        DnsResourceRecord *parent;

        // Prohibited funcs - not implemented.
        DnsResourceData( const DnsResourceData& );

    } ;


//---------------------------------------------------------------------------
//
    class DnsResourceRecord
    {
    public:
        DnsResourceRecord();

        DnsResourceRecord( char * buf, size_t &offset, const size_t length );

        DnsResourceRecord( const DnsResourceRecord& peer);

        ~DnsResourceRecord();

        void Deserialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException );

        void Serialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException );

        ZKSTD::string GetName() const;

        void SetName( const ZKSTD::string in_name );

        uint16_t GetType( void ) const;
        
        // This is a special modifier. Look at the definition.
        void SetType( const uint16_t in_type );

        uint16_t GetClass() const;

        void SetClass( const uint16_t in_class );

        uint16_t GetTtl() const;

        void SetTtl( const uint16_t in_ttl );

        // NULL could be returned if type is unset
        DnsResourceData *GetResourceData( void );

        uint16_t GetRDLength() const;

        /////
        // No way to directly set Resource Data, use SetType.
        // No way to set RDLengh, it's implied by the type and the contents of m_rdata
        /////

    private:
        ZKSTD::string m_name;
        uint16_t m_type;
        uint16_t m_rrclass;
        uint32_t m_ttl;
        uint16_t m_rdlength;
        DnsResourceData *m_rdata;

        // Prohibited funcs - not implemented.
        const DnsResourceRecord& operator=( const DnsResourceRecord& );
    };

//---------------------------------------------------------------------------
//
    // all the rdata types could be a union, but that would kind of suck.
    class MXResourceData
        : public DnsResourceData
    {
    public:
        MXResourceData(DnsResourceRecord *in_parent);

        MXResourceData(DnsResourceRecord *in_parent, char * buf, size_t &offset, const size_t length);

        const MXResourceData& operator=( const MXResourceData& );

        virtual ~MXResourceData();

        virtual void Deserialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException );

        virtual void Serialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException );

        uint16_t GetPreference( void ) const;

        void SetPreference( const uint16_t pref );

        ZKSTD::string GetExchange( void ) const;

        void SetExchange( const ZKSTD::string exch );
      
    private:
        uint16_t m_preference;
        ZKSTD::string m_exchange;

        // Prohibited funcs - not implemented.
        MXResourceData( const MXResourceData& );
    };

//---------------------------------------------------------------------------
//
    class NSResourceData
        : public DnsResourceData
    {
    public:
        NSResourceData(DnsResourceRecord *in_parent);

        NSResourceData(DnsResourceRecord *in_parent, char * buf, size_t &offset, const size_t length);

        const NSResourceData& operator=( const NSResourceData& );

        virtual ~NSResourceData();

        virtual void Deserialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException );

        virtual void Serialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException );

        ZKSTD::string GetNameServer( void ) const;

        void SetNameServer( const ZKSTD::string nsrv );
      
    private:
        ZKSTD::string m_nameserver;

        // Prohibited funcs - not implemented.
        NSResourceData( const NSResourceData& );
    };

//---------------------------------------------------------------------------
//
    class SOAResourceData
        : public DnsResourceData
    {
    public:
        SOAResourceData(DnsResourceRecord *in_parent);

        SOAResourceData(DnsResourceRecord *in_parent, char * buf, size_t &offset, const size_t length);

        const SOAResourceData& operator=( const SOAResourceData& );

        virtual ~SOAResourceData();

        virtual void Deserialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException );

        virtual void Serialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException );

        ZKSTD::string GetMName( void ) const;

        void SetMName( const ZKSTD::string in_mname );
      
        ZKSTD::string GetRName( void ) const;

        void SetRName( const ZKSTD::string in_rname );

        uint32_t GetSerial( void ) const;

        void SetSerial( const uint32_t in_serial );

        uint32_t GetRefresh( void ) const;

        void SetRefresh( const uint32_t in_refresh );
        
        uint32_t GetRetry( void ) const;

        void SetRetry( const uint32_t in_retry );

        uint32_t GetExpiry( void ) const;

        void SetExpiry( const uint32_t in_expiry );

        uint32_t GetMinimum( void ) const;

        void SetMinimum( const uint32_t in_minimum );

    private:
        ZKSTD::string m_mname;
        ZKSTD::string m_rname;
        uint32_t m_serial;
        uint32_t m_refresh;
        uint32_t m_retry;
        uint32_t m_expiry;
        uint32_t m_minimum;

        // Prohibited funcs - not implemented.
        SOAResourceData( const SOAResourceData& );
    };

//---------------------------------------------------------------------------
//
    class AResourceData
        : public DnsResourceData
    {
    public:
        AResourceData(DnsResourceRecord *in_parent);

        AResourceData(DnsResourceRecord *in_parent, char * buf, size_t &offset, const size_t length);

        const AResourceData& operator=( const AResourceData& );

        virtual ~AResourceData();

        virtual void Deserialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException );

        virtual void Serialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException );

        // NOTE: addresses are not byte swapped
        uint32_t GetAddress( void ) const;

        void SetAddress( const uint32_t in_address );
      
    private:
        uint32_t m_address;

        // Prohibited funcs - not implemented.
        AResourceData( const AResourceData& );
    };

//---------------------------------------------------------------------------
//
    // Place holder for unimplemented RRs 
    class FooResourceData
        : public DnsResourceData
    {
    public:
        FooResourceData(DnsResourceRecord *in_parent);

        FooResourceData(DnsResourceRecord *in_parent, char * buf, size_t &offset, const size_t length);

        const FooResourceData& operator=( const FooResourceData& );

        virtual ~FooResourceData();

        virtual void Deserialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException );

        virtual void Serialize( char *buf, size_t &offset, const size_t length ) throw ( RZkException );

        char *GetBuffer( void );

        void SetBuffer( char *in_buffer, unsigned int in_length );

    private:
        char *m_buffer;
        unsigned int m_buflength;

        // Prohibited funcs - not implemented.
        FooResourceData( const FooResourceData& );
    };



//---------------------------------------------------------------------------
// We're back in DnsMessage.

public:
    
    RDnsMessage();

    // Deserialization ctor
    RDnsMessage( char *data, const size_t length ) throw ( RZkException );

    // Smoke all dynamic questions and RRs
    ~RDnsMessage();

	// Purpose: Unpack a section of a DNS message from linear memory
	// Promise: Not to overrun available buffer
	// Requires: 
	// Throws: RZk... if the format is bad or the message is truncated (short buffer)
    void Deserialize( char *buf, const size_t length ) throw ( RZkException );

	// Purpose: Pack a section of a DNS message to linear memory
	// Promise: Not to overrun available buffer
	// Requires: 
	// Throws: RZk... if the message is truncated (short buffer)
    size_t Serialize( char *buf, const size_t length ) throw ( RZkException );

	// Purpose:  Parse the RFC <domain-name> format into a usable string. This is based on the assumption that
    //           this code isn't implementing a real DNS server where the indpendant sections would be more important
	// Throws:   RZk... if the domain is improperly formated or the buffer is short
    static ZKSTD::string EncodedDomainToString( char *buf, size_t &offset, const size_t length ) throw ( RZkException );

	// Purpose:  encode a usable sting in  RFC <domain-name> format. This is based on the assumption that
    //           this code isn't implementing a real DNS server where the indpendant sections would be more important
	// Throws:   RZk... if the string is improperly formated or the buffer is short
    static void StringToEncodedDomain( ZKSTD::string str, char *buf, size_t &offset, const size_t length ) throw ( RZkException );        

	// Purpose: Unpack a value from a linear array. val is filled in with the sizeof(T) bytes at buf[offset]
    // Requires: Instantiate with any scalar type
    // (wierdo first argument is so that you don't have to use the long instantiation syntax)
    template<typename T> static void Unpack( T &val, char *buf, size_t &offset )
	{
	    val =  *reinterpret_cast<T *>(&buf[offset]);
	    offset += sizeof(T);
	}

	// Purpose Pack a value into a linear array. buf[offset] is filled with sizeof(T) bytesfrom val:  
    // Requires: Instantiate with any scalar type
    template<typename T> static void Pack( T val, char *buf, size_t &offset )
	{
	    *reinterpret_cast<T *>(&buf[offset]) = val;
	    offset += sizeof(T);
	}

	// Purpose: Unpack a value from a linear array. val is filled in with the sizeof(T) bytes at buf[offset]
    // Throws: RZk... if bufer does not contain sizeof(T) additional bytes after offset 
    // (wierdo first argument is so that you don't have to use the long instantiation syntax)
    template<typename T> static void UnpackSafe( T &val, char *buf, size_t &offset, size_t length ) throw (RZkException)
	{
	    if ( offset + sizeof(T) > length )
	    {
			throw RZkException(
	            "RDnsMessage::DnsQuestion::Deserialize", 
				"Buffer overrun", 
				IDS_DEFAULT_ERROR);
		}
		Unpack( val, buf, offset );
	}

	// Purpose Pack a value into a linear array. buf[offset] is filled with sizeof(T) bytesfrom val
    // Throws: RZk... if bufer does not contain sizeof(T) additional bytes after offset 
    // Requires: Instantiate with any scalar type
    template<typename T> static void PackSafe( T val, char *buf, size_t &offset, size_t length ) throw (RZkException)
	{
	    if ( offset + sizeof(T) > length )
	    {
		    throw RZkException(
			    "RDnsMessage::DnsQuestion::Deserialize", 
				"Buffer overrun", 
	            IDS_DEFAULT_ERROR);
		}
		Pack( val, buf, offset );
	}

	// Purpose: Optimisation. This lets you do trivial rejection tests for DNS packets.
    //          If you want to dig into packets that contain SOA RRs, call this with the buffer
    //          and RDnsMessage::SOA as the needle. It will tell you if the packet MIGHT contain
    //          a SOA (i.e. it contains the SOA bit sequence somewhere). Since everyhting (including alignment)
    //          is variable in DNS this is as sure as you can get without parsing the whole shpiel.
	// Promise: Not to crash and burn. To work. etc. (Damn granola coding-standard #$%#%@)
	// Requires: needle should match the byte order of haystack. Meaning if you want to search
    //           a big endian buffer make sure your search parameter is also big endian, obviously.
    static bool ScanBufferForShort( uint16_t in_needle, char *in_haystack, size_t in_haylen );

	// Purpose: Optimisation. This lets you do trivial rejection tests for DNS packets.
    //          As above except that this searches for a string (consequently byte order is irrelevant)
	// Promise: Not to crash and burn. To work. etc. (Damn granola coding-standard #$%#%@)
    static bool RDnsMessage::ScanBufferForString( char *in_needle, char *in_haystack, size_t in_haylen );

    // Manipulate directly. Returns ref
    DnsHeader &GetHeader();

    DnsQuestion *GetQuestion( const unsigned int i );

    void AddQuestion( const DnsQuestion &q );

    DnsResourceRecord *GetAnswer( const unsigned int i );

    void AddAnswer( const DnsResourceRecord &rr );

    DnsResourceRecord *GetNameServer( const unsigned int i );

    void AddNameServer( const DnsResourceRecord &rr );
    
    DnsResourceRecord *GetAdditional( const unsigned int i );

    void AddAdditional( const DnsResourceRecord &rr );

private:
    DnsHeader m_header;
    ZKSTD::vector< DnsQuestion * > m_qdata; // question
    ZKSTD::vector< DnsResourceRecord * > m_andata; // answer
    ZKSTD::vector< DnsResourceRecord * > m_nsdata; // name server
    ZKSTD::vector< DnsResourceRecord * > m_ardata; // additional

};

////
// This code was actually readable before I decided to follow the
// ZKS coding standard for readability. --ol

//---------------------------------------------------------------------------
// RDnsMessage::DnsHeader Inline Members
//---------------------------------------------------------------------------
inline uint16_t RDnsMessage::DnsHeader::GetID() const 
{
    return id;
};

inline void RDnsMessage::DnsHeader::SetID( const uint16_t in_id ) 
{ 
    id = in_id; 
};

inline uint16_t RDnsMessage::DnsHeader::GetQR() const
{
    return qr; 
};

inline void RDnsMessage::DnsHeader::SetQR( const uint16_t in_qr ) 
{ 
    qr = in_qr; 
};

inline uint16_t RDnsMessage::DnsHeader::GetOpcode() const
{
    return opcode; 
}

inline void RDnsMessage::DnsHeader::SetOpcode( const uint16_t in_op ) 
{
    opcode = in_op; 
};

inline uint16_t RDnsMessage::DnsHeader::GetAA() const
{
    return aa; 
};

inline void RDnsMessage::DnsHeader::SetAA( const uint16_t in_aa ) 
{
    aa = in_aa; 
};
        
inline uint16_t RDnsMessage::DnsHeader::GetTC() const
{
    return tc; 
};

inline void RDnsMessage::DnsHeader::SetTC( const uint16_t in_tc ) 
{
    tc = in_tc; 
};

inline uint16_t RDnsMessage::DnsHeader::GetRD() const
{
    return rd; 
};

inline void RDnsMessage::DnsHeader::SetRD( const uint16_t in_rd ) 
{
    rd = in_rd; 
};

inline uint16_t RDnsMessage::DnsHeader::GetRA() const
{
    return ra; 
};

inline void RDnsMessage::DnsHeader::SetRA( const uint16_t in_ra ) 
{
    ra = in_ra; 
};

inline uint16_t RDnsMessage::DnsHeader::GetRCode() const
{
    return rcode; 
};

inline void RDnsMessage::DnsHeader::SetRCode( const uint16_t in_rcode ) 
{
    rcode = in_rcode; 
};

inline uint16_t RDnsMessage::DnsHeader::GetQDCount() const
{
    return ntohs( qdcount ); 
};

inline void RDnsMessage::DnsHeader::SetQDCount( const uint16_t in_qdcount ) 
{
    qdcount = htons( in_qdcount ); 
};

inline uint16_t RDnsMessage::DnsHeader::GetANCount() const
{
    return ntohs( ancount ); 
};
        
inline void RDnsMessage::DnsHeader::SetANCount( const uint16_t in_ancount ) 
{
    ancount = htons( in_ancount ); 
};

inline uint16_t RDnsMessage::DnsHeader::GetNSCount() const
{ 
    return ntohs( nscount ); 
};

inline void RDnsMessage::DnsHeader::SetNSCount( const uint16_t in_nscount ) 
{
    nscount = htons( in_nscount ); 
};

inline uint16_t RDnsMessage::DnsHeader::GetARCount() const
{
    return ntohs( arcount ); 
};

inline void RDnsMessage::DnsHeader::SetARCount( const uint16_t in_arcount ) 
{ 
    arcount = htons( in_arcount ); 
};


//---------------------------------------------------------------------------
// RDnsMessage::DnsQuestion Inline Members
//---------------------------------------------------------------------------

inline RDnsMessage::DnsQuestion::DnsQuestion()
{
};

inline RDnsMessage::DnsQuestion::DnsQuestion( char *buf, size_t &offset, const size_t length )
{ 
    Deserialize( buf, offset, length ); 
};

inline RDnsMessage::DnsQuestion::~DnsQuestion()
{
};

inline ZKSTD::string RDnsMessage::DnsQuestion::GetName() const
{ 
    return m_qname; 
};

inline void RDnsMessage::DnsQuestion::SetName( const ZKSTD::string name )
{ 
    m_qname = name; 
};

inline uint16_t RDnsMessage::DnsQuestion::GetType() const
{
    return m_qtype; 
};

inline void RDnsMessage::DnsQuestion::SetType( const uint16_t type )
{ 
    m_qtype = type; 
};

inline uint16_t RDnsMessage::DnsQuestion::GetClass() const
{ 
    return m_qclass; 
};

inline void RDnsMessage::DnsQuestion::SetClass( const uint16_t in_class )
{ 
    m_qclass = in_class; 
};


//---------------------------------------------------------------------------
// RDnsMessage::DnsResourceData Inline Members
//---------------------------------------------------------------------------

inline RDnsMessage::DnsResourceRecord *RDnsMessage::DnsResourceData::GetParent( void ) const
{
    return parent; 
};

inline void RDnsMessage::DnsResourceData::SetParent( DnsResourceRecord *in_parent )
{ 
    parent = in_parent; 
};


//---------------------------------------------------------------------------
// RDnsMessage::DnsResourceRecord Inline Members
//---------------------------------------------------------------------------

inline RDnsMessage::DnsResourceRecord::DnsResourceRecord()
    : m_rdata ( NULL )
{ 
};

inline RDnsMessage::DnsResourceRecord::DnsResourceRecord( char * buf, size_t &offset, const size_t length )
    : m_rdata ( NULL )
{ 
    Deserialize( buf, offset, length ); 
};

inline RDnsMessage::DnsResourceRecord::~DnsResourceRecord()
{ 
    if (m_rdata) delete m_rdata; 
};        

inline ZKSTD::string RDnsMessage::DnsResourceRecord::GetName() const
{ 
    return m_name; 
} ;

inline void RDnsMessage::DnsResourceRecord::SetName( const ZKSTD::string in_name )
{
    m_name = in_name; 
};

inline uint16_t RDnsMessage::DnsResourceRecord::GetType( void ) const
{
    return m_type; 
};
        
inline uint16_t RDnsMessage::DnsResourceRecord::GetClass() const
{ 
    return m_rrclass; 
};

inline void RDnsMessage::DnsResourceRecord::SetClass( const uint16_t in_class )
{
    m_rrclass = in_class; 
};

inline uint16_t RDnsMessage::DnsResourceRecord::GetTtl() const
{
    return m_ttl; 
};

inline void RDnsMessage::DnsResourceRecord::SetTtl( const uint16_t in_ttl )
{
    m_ttl = in_ttl; 
};

inline uint16_t RDnsMessage::DnsResourceRecord::GetRDLength() const
{
    return m_rdlength; 
};

// NULL could be returned if type is unset
inline RDnsMessage::DnsResourceData *RDnsMessage::DnsResourceRecord::GetResourceData( void )
{
    return m_rdata; 
};


//---------------------------------------------------------------------------
// RDnsMessage::MXResourceRecord Inline Members
//---------------------------------------------------------------------------

inline uint16_t RDnsMessage::MXResourceData::GetPreference( void ) const
{
    return m_preference; 
};

inline void RDnsMessage::MXResourceData::SetPreference( const uint16_t pref )
{ 
    m_preference = pref; 
};

inline ZKSTD::string RDnsMessage::MXResourceData::GetExchange( void ) const
{
    return m_exchange; 
}; 

inline void RDnsMessage::MXResourceData::SetExchange( const ZKSTD::string exch )
{ 
    m_exchange = exch; 
};

//---------------------------------------------------------------------------
// RDnsMessage::NSResourceRecord Inline Members
//---------------------------------------------------------------------------

inline ZKSTD::string RDnsMessage::NSResourceData::GetNameServer( void ) const
{
    return m_nameserver; 
}; 

inline void RDnsMessage::NSResourceData::SetNameServer( const ZKSTD::string nsrv )
{ 
    m_nameserver = nsrv; 
};


//---------------------------------------------------------------------------
// RDnsMessage::SOAResourceRecord Inline Members
//---------------------------------------------------------------------------

inline ZKSTD::string RDnsMessage::SOAResourceData::GetMName( void ) const
{
    return m_mname; 
}; 

inline void RDnsMessage::SOAResourceData::SetMName( const ZKSTD::string in_mname )
{ 
    m_mname = in_mname; 
};

inline ZKSTD::string RDnsMessage::SOAResourceData::GetRName( void ) const
{
    return m_rname; 
}; 

inline void RDnsMessage::SOAResourceData::SetRName( const ZKSTD::string in_rname )
{ 
    m_rname = in_rname; 
};

inline uint32_t RDnsMessage::SOAResourceData::GetSerial( void ) const
{
    return m_serial;
};

inline void RDnsMessage::SOAResourceData::SetSerial( const uint32_t in_serial )
{
    m_serial = in_serial;
};

inline uint32_t RDnsMessage::SOAResourceData::GetRefresh( void ) const
{
    return m_refresh;
};

inline void RDnsMessage::SOAResourceData::SetRefresh( const uint32_t in_refresh )
{
    m_refresh = in_refresh;
};

inline uint32_t RDnsMessage::SOAResourceData::GetRetry( void ) const
{
    return m_retry;
};

inline void RDnsMessage::SOAResourceData::SetRetry( const uint32_t in_retry )
{
    m_retry = in_retry;
};

inline uint32_t RDnsMessage::SOAResourceData::GetExpiry( void ) const
{
    return m_expiry;
};

inline void RDnsMessage::SOAResourceData::SetExpiry( const uint32_t in_expiry )
{
    m_expiry = in_expiry;
}

inline uint32_t RDnsMessage::SOAResourceData::GetMinimum( void ) const
{
    return m_minimum;
};

inline void RDnsMessage::SOAResourceData::SetMinimum( const uint32_t in_minimum )
{
    m_minimum = in_minimum;
};

//---------------------------------------------------------------------------
// RDnsMessage::FooResourceRecord Inline Members
//---------------------------------------------------------------------------
inline uint32_t RDnsMessage::AResourceData::GetAddress( void ) const
{
    return m_address;
};

inline void RDnsMessage::AResourceData::SetAddress( const uint32_t in_address )
{
    m_address = in_address;
};


//---------------------------------------------------------------------------
// RDnsMessage::FooResourceRecord Inline Members
//---------------------------------------------------------------------------

inline char *RDnsMessage::FooResourceData::GetBuffer( void )
{
    return m_buffer; 
};


//---------------------------------------------------------------------------
// RDnsMessage Inline Members
//---------------------------------------------------------------------------

// Manipulate directly. Returns ref
inline RDnsMessage::DnsHeader &RDnsMessage::GetHeader() 
{ 
    return m_header; 
};

#endif // RDNSMESSAGE_H__
