// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "ZkNetHelp.h"
#include <IZkConsoleMsg.h>
#include <zkDNS.h>

#ifndef _WIN32
#   include <netdb.h>
#   include <net/if.h>
#   include <sys/ioctl.h>
#   include <unistd.h>
#endif

ZKSTD::vector< ZKSTD::string > ZkNetHelp::m_vecStrIPs;
ZKSTD::vector< uint32_t > ZkNetHelp::m_vecIPs;
bool ZkNetHelp::m_bGrokedIPsAlready = false;


ZkNetHelp::ZkNetHelp()
	: m_socket( NULL )
{
	::memset( &m_realSockAddr, 0, sizeof( m_realSockAddr ) );
}


ZkNetHelp::~ZkNetHelp()
{
}

bool ZkNetHelp::InitRealSockAddr()
{
	ZKTRACE0( "ZkNetHelp::GetRealSockAddr()\n" );
	bool bReturnValue = true;
	char szBuffer[sizeof( sockaddr )];

	// Recv the real sockaddr
	if( 1 != zkSocketReadAll( &m_err, m_socket, szBuffer, sizeof (sockaddr) ) )
		bReturnValue = false;
	else 
	{		
			if( szBuffer[0] == AF_INET )
				memcpy( &GetOriginalDestinationSocketAddress(), szBuffer, sizeof (sockaddr) );
			else 
			{
				bReturnValue = false;
				ZKASSERT( false );
			}
	}
		
	return bReturnValue;
}

bool ZkNetHelp::SmtpResponseIsSuccess(const ZKSTD::string& in_strResponse)
{
	if(in_strResponse[0] != '2' && in_strResponse[0] != '3')
	{
		return false;
	}
	return true;
}

bool ZkNetHelp::IsRemoteHostDummyMailServer(ZkSocketAddress& in_RemoteAddr)
{
	bool bHostIsDummyServer = false;
	hostent*	pHostEntry = NULL;
 
	if ((pHostEntry = gethostbyaddr((const char *)&( (struct sockaddr_in*)&in_RemoteAddr)->sin_addr,
		sizeof(in_addr),
		AF_INET)) != NULL) 
	{
		if(!stricmp(pHostEntry->h_name, DUMMY_MAIL_SERVER))
		{
			bHostIsDummyServer = true;
		}
	}
	else
	{
		if ((pHostEntry = gethostbyname(DUMMY_MAIL_SERVER)) != NULL
			&& pHostEntry->h_addr_list[0])
		{
			for(size_t i = 0; pHostEntry->h_addr_list[i] != NULL; i++)
			{
				if(!memcmp(pHostEntry->h_addr_list[i],
					(const char *)&( (struct sockaddr_in*)&in_RemoteAddr )->sin_addr, 4))
				{
					bHostIsDummyServer = true;
					break;
				}
			}
			
		}
		else
		{
			IZkConsoleMsg::ShowMsg(CONSIMAGE_MAIL, IDS_CONS_MAIL_DNS_LOOKUP_FAILED, DUMMY_MAIL_SERVER);
		}
	}

	return bHostIsDummyServer;
}

bool ZkNetHelp::ReceiveLine(ZkError in_err, ZkSocket in_Socket, ZKSTD::string& out_strLine)
{
	bool ReturnValue = true;

	size_t ThisCallBytesRead = 0;
	size_t TotalBytesRead = 0;
	char ByteBuffer[256];

	while(true)
	{
		if( !zkSocketRead(
				in_err,
				in_Socket,
				ByteBuffer + TotalBytesRead,
				255 - TotalBytesRead,
				(unsigned int*)&ThisCallBytesRead))
		{
			ReturnValue = false;
			break;
		}
		else
		{
			TotalBytesRead += ThisCallBytesRead;
			ByteBuffer[TotalBytesRead] = '\0';
			if(strchr(ByteBuffer, '\r'))
			{
				out_strLine = ByteBuffer;
				break;
			}
		}
	}
	return ReturnValue;
}


//---------------------------------------------------------------------------
//
#ifdef ZK_LINUX
int ZkNetHelp::FetchIFAddr(struct ifreq *ifr, char *ifname)
{
    int fd;
    int rval = ZK_FALSE;

    /* IPv4 addresses */
    fd = socket(AF_INET, SOCK_STREAM, 0);

    if (fd >= 0) {
        zkCStrlcpy(ifr->ifr_name, ifname, sizeof(ifr->ifr_name));
        ifr->ifr_addr.sa_family = AF_INET;
        if (ioctl(fd, SIOCGIFADDR, ifr) == 0) {
            rval = ZK_TRUE;
        } 
        close (fd);
	} 
    return rval;
}

//---------------------------------------------------------------------------
//
#define PROC_NET_ROUTE_FNAME "/proc/net/route"
int ZkNetHelp::GetDefaultRouteIF( char *ifbuf, size_t ifbuflen )
{
    char buf[1024];
    FILE *fp;
    char ifname[128];
    char rtaddr[128];
    char gwaddr[128];
    int flags;
    int num;
    unsigned int i;
    fp = fopen(PROC_NET_ROUTE_FNAME, "r");
    if (fp) {
        while ( fgets(buf, 1023, fp) ) 
        {
            /* This should be safe, we limit our buffer sizes */
            num = sscanf(buf, "%128s %128s %128s %X", ifname, rtaddr, gwaddr, &flags);
            if ( num < 4 ) continue;
            /* well Ok, this is a bit dumbass, we don't check the mask
             * or look at the address's value as anything other than a string
             * still, it's all that should be necessary --ol
             */
            if ( !(flags & 1) ) continue;
            for ( i = 0; i < strlen(rtaddr) && rtaddr[i] == '0'; i++ );
            if (i >= strlen(rtaddr) - 1) {
                strncpy(ifbuf, ifname, ifbuflen - 1);
                fclose(fp);
                return ZK_TRUE;
            }            
        }
    } 
    fclose(fp);
    return ZK_FALSE;
}
#endif


//---------------------------------------------------------------------------
//
void ZkNetHelp::GetIPsForThisHost()
{
    if ( m_bGrokedIPsAlready ) return;
    else {
#ifdef ZK_LINUX
        struct ifreq ifr;
        char ifBuf[128];

        // The windows gethostbyname approach doesn't work on linux
        // We use it as a fall back, but we try to get the real internet
        // address of the machine first.

        if ( !GetDefaultRouteIF( ifBuf, 128 ) ) goto try2;
        if ( !FetchIFAddr( &ifr, ifBuf ) ) goto try2;
        m_vecIPs.push_back( ((struct sockaddr_in*)&ifr.ifr_addr)->sin_addr.s_addr );
        m_vecStrIPs.push_back( inet_ntoa( ((struct sockaddr_in*)&ifr.ifr_addr)->sin_addr ) );
		m_bGrokedIPsAlready = true;
        return;

    try2:
#endif
        char szBuf[1024];  //MAXGETHOSTSTRUCT
        struct hostent* pHostEntry = NULL;
        
        gethostname(szBuf, 1024-1); // used to retrieve name of local machine, is the only standard database function that does not block
        
        if (pHostEntry = gethostbyname(szBuf)) // always blocks no matter what type of socket used
        {
            ZKASSERT(pHostEntry);
            int j = 0 ;
            
            while ( pHostEntry->h_addr_list[j] ) 
            {
                struct in_addr * pAddr = (struct in_addr *) pHostEntry->h_addr_list[j] ;
                
                // Local host?
                
                // This is network byte order 
                m_vecStrIPs.push_back( inet_ntoa( *pAddr ) ) ;
                
                m_vecIPs.push_back( pAddr->s_addr ) ;
                
                // Try the other byte order too
                m_vecIPs.push_back( htonl(pAddr->s_addr) ) ;
                
                j++ ;
            }
            
            // We don't deallocate
            pHostEntry = NULL ;
        } 
        m_bGrokedIPsAlready = true;
    }       
}


//---------------------------------------------------------------------------
//
const ZKSTD::vector< ZKSTD::string > &ZkNetHelp::GetVectorOfStringIPs()
{
    if (!m_bGrokedIPsAlready) {
        GetIPsForThisHost();
    }
    return m_vecStrIPs;
}

//---------------------------------------------------------------------------
//
const ZKSTD::vector< uint32_t > &ZkNetHelp::GetVectorOfBinaryIPs()
{
    if (!m_bGrokedIPsAlready) {
        GetIPsForThisHost();
    }
    return m_vecIPs;
}

//---------------------------------------------------------------------------
//

#ifdef INTERNAL_BUILD
void ZkNetHelp::UnitTest()
{
	ZkNetHelp netHelper;
	ZKTEST( ( netHelper.m_socket == netHelper.GetSocket() )
			&& netHelper.m_socket == NULL );

	ZKTEST( ::zkSocketNew( &netHelper.m_err, 
							AF_INET, 
							SOCK_STREAM, 
							IPPROTO_TCP, 
							&netHelper.GetSocket() ) );
	ZKTEST( ( netHelper.m_socket == netHelper.GetSocket() )
			&& netHelper.m_socket != NULL );

	ZKTEST( ::zkSocketConnect( &netHelper.m_err, netHelper.GetSocket(), "caesar.zks.net:7" ) );
	sockaddr_in bufToSend = { AF_INET };
	uint32_t dwLen = 0;
	ZKTEST( ::zkSocketWrite( NULL, netHelper.GetSocket(), &bufToSend, sizeof bufToSend, &dwLen ) );

	// There must be no read on the socket before this call or it will fail.
	// A second call to this must assert.  We won't provoque it for the unit test.
	ZKTEST( netHelper.InitRealSockAddr() );
	ZKTEST( !::memcmp( &netHelper.GetError(), &netHelper.m_err, sizeof netHelper.m_err ) );

	ZkSocketAddress socketAddress = netHelper.GetOriginalDestinationSocketAddress();
	ZKTEST( !::memcmp( &socketAddress, &netHelper.m_realSockAddr, sizeof (ZkSocketAddress) ) );
//	ZKTEST( socketAddress.m_inet.sa.sin_port == 7 );
//	ZKTEST( ::inet_ntoa( socketAddress.m_inet.sa.sin_addr ) == "10.16.0.8" );

	ZKTEST( ::zkSocketFree( &netHelper.m_err, &netHelper.m_socket ) );
}
#endif // INTERNAL_BUILD
