/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/*
	This header provides little helper functions related to neworking in general
	NEVER #include anything which adds dependencies to this header!

	Cvt_int32toString	- Convert an ip address into a string without using DNS
	Ip2String			- Convert an ip address into a string, may use DNS
	GetIcmpMessageType	- Get a string representation of an icmp message type
	GetIcmpMessageCode	- Get a string represenation of an icmp code depending on the type
	GetIcmpUnreachCode	- Get a string representation for destination unreachable type
	GetIcmpRedirectCode - Get a string representation for redirect type
	GetIcmpTimeExceededCode - Get a string representation for time exceeded type
	GetIcmpParamProbCode	- Get a string represenation for the parameter problem type
	IsNetBiosPort			- Returns true if the port number corresponds to one of the 3 default netbios ports	
	IsDHCPPort				- Returns true if the port number is one of the two DHCP port numbers
	IsRPCPort				- Returns true if the port number is the RPC port
	IsIdentPort				- Returns true if the port number is Identification protocol
	IsProxyPortDefaultForTheCloud - Returns true if the port number is a _default_ port for a proxy which uses the cloud
	IsDefaultPop3Port		- Returns true if the port number is a default pop3 port
	IsDefaultSmtpPort		- Returns true if the port number is a default smtp port
	IsDefaultNewsPort		- Returns true if the port number is a default nntp port
	IsDefaultTelnetPort		- Returns true if the port number is a default telnet port
	IsDefaultSshPort		- Returns true if the port number is a default ssh port
	IsDefaultSslPort		- Returns true if the port number is a default ssl port
	IsDefaultWebPort		- Returns true if the port number is a default web port
	IsDefaultIrcPort		- Returns true if the port number is a default irc port
	IsZksTcpPortCloudOrNot	- Returns true if the port number is a port which we must go through the cloud to reach
							  it is not a proxy port, but rather a zks service port
	DumptcpHeader			- Dumps a tcp header using zktrace
	DumpUdpHeader			- Dumps a udp header using trace
	DumpUdpHeaderToFile		- Dumps a udp header into the passed ofstream
	DumptcpHeaderToFile		- Dumps a tcp header into the passed ofstream
	DumpIpHeaderToStream	- Dumps an ip header into the passed ostringstream
	GetSocketType			- Returns the socket type (SOCK_STREAM, etc) or 0 on failure
	ConvertFirewallAppPersistIntoString - Converts the structure into a human-readable string
*/

#ifndef WIN32
#include <netdb.h>				// for protoent structure
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif

#include <zkPacket.h>			// for ZK_ICMP_MAXTYPE
#include <time.h>
#include <frNetworkDefaults.h>	// for hard-coded compiled in default zks ports
#include <clientdebug.h>		// for zktrace
#include <client.h>				// for FirewallAppPersist

#include <sstream>				// for ostringstream
#include <fstream>				// for ofstream

namespace ZKS{

extern inline ZKSTD::string 
GetProtocolInfo(const int in_proto) throw()
{
	const protoent* prot = getprotobynumber(in_proto);

	if(prot)
	{
		return prot->p_name;
	}

	return "Unknown";
}

// Returns NULL if no service is returnd by the socket api
extern inline char*  
GetServiceInformation(const unsigned short in_port, const char* in_protocol) throw()
{
	const struct servent* ps = getservbyport( in_port, in_protocol );

	if(ps)
	{
		return ps->s_name;
	}
	return NULL;
}

extern inline ZKSTD::string 
Cvt_int32toString(const unsigned int in_ip)  throw()
{
	in_addr x;
	x.s_addr = in_ip;
	return ::inet_ntoa(x);
};

extern inline ZKSTD::string 
GetIcmpMessageType(const unsigned int in_type)  throw()
{
	static char* ttab[] = 
	{
        "Echo_Reply",
        "-Type-OUT-OF-RANGE-with-1",
        "-Type-OUT-OF-RANGE-with-2",
        "Dest_Unreachable",
        "Source_Quench",
        "Redirect",
        "-Type-OUT-OF-RANGE-with-6",
        "-Type-OUT-OF-RANGE-with-7",
        "Echo_Request",
        "RouterAdvert",
        "Router_Solicit",
        "Time_Exceeded",
        "Parameter_Problem",
        "Timestamp",
        "Timestamp_Reply",
        "Info_Request",
        "Info_Reply",
        "Mask_Request",
        "Mask_Reply"
    };
	   
	static char strBuff[50];
	
    if( in_type < 0 || in_type > ZK_ICMP_MAXTYPE) 
        snprintf( strBuff, 49, "[-Type-OUT-OF-RANGE-with-%d]", in_type );
    else 
        snprintf( strBuff, 49, "[%s]", ttab[in_type] );

    return strBuff;
}

extern inline ZKSTD::string 
GetIcmpUnreachCode( const unsigned int in_iIndex )  throw()
{
    static char	*ttab[] = 
	{
      "Net Unreachable",
      "Host Unreachable",
      "Protocol Unreachable",
      "Port Unreachable",
      "Fragmentation Needed/DF set",
      "Source_Route Failed",
      "DestNet Unknown",
      "DestHost Unknown",
      "Source Host Isolated",
      "Net Prohibited",
      "Host Prohibited",
      "Net TOS Unreachable",
      "Host TOS Unreachable",
      "Packet Filtered",
      "Precedence Violation",
      "Precedence Cut-Off"
	};
    static char strBuff[50];
	
    if( in_iIndex < 0 || in_iIndex > ZK_ICMP_UNREACH_MAX ) 
        snprintf( strBuff, 49, "[--Sub-Type-OUT-OF-RANGE-with-%d]", in_iIndex );
    else
        snprintf( strBuff, 49, "[%s]", ttab[in_iIndex] );
    
	return strBuff;
}

extern inline ZKSTD::string 
GetIcmpRedirectCode( const unsigned int in_iIndex )  throw()
{
    static char	*ttab[] = 
	{
        "Redirect Net",
        "Redirect Host",
        "Redirect Net for TOS",
        "Redirect Host for TOS"
    };
    static char strBuff[50];
	
    if( in_iIndex < 0 || in_iIndex > 3) 
        snprintf( strBuff, 49, "[-Sub-Type-OUT-OF-RANGE-with-%d]", in_iIndex );
    else 
        snprintf( strBuff, 49, "[%s]", ttab[in_iIndex] );

    return strBuff;
}


extern inline ZKSTD::string 
GetIcmpTimeExceededCode( const unsigned int in_iIndex)  throw()
{
    static char	*ttab[] = 
	{
        "TTL count exceeded in transit",
        "Fragment Reassembly time exceeded"
    };
    static char strBuff[50];
	
    if( in_iIndex < 0 || in_iIndex > 1 ) 
        snprintf( strBuff, 49, "[--Sub-Type-OUT-OF-RANGE-with-%d]", in_iIndex );
    else 
        snprintf( strBuff, 49, "[%s]", ttab[in_iIndex] );

    return strBuff;
}

extern inline ZKSTD::string 
GetIcmpParamProbCode( const unsigned int in_iIndex )  throw()
{
    static char	*ttab[] = 
	{
        "Misc parameter problem",
        "--Sub-Type-OUT-OF-RANGE-with-1",
        "Required option missing"
    };
    static char strBuff[50];
	
    if( in_iIndex < 0 || in_iIndex > 2 ) 
        snprintf( strBuff, 49, "[--Sub-Type-OUT-OF-RANGE-with-%d]", in_iIndex );
	else 
        snprintf( strBuff, 49, "[%s]", ttab[in_iIndex] );

    return strBuff;
}

extern inline ZKSTD::string 
GetIcmpMessageCode(const unsigned int in_type, const unsigned int in_code)  throw()
{
	switch(in_type)
	{
	case ZK_ICMP_UNREACH:
		return ZKS::GetIcmpUnreachCode(in_code);
	case ZK_ICMP_REDIRECT:
		return ZKS::GetIcmpRedirectCode(in_code);
	case ZK_ICMP_TIMXCEED:
		return ZKS::GetIcmpTimeExceededCode(in_code);
	case ZK_ICMP_PARAMPROB:
		return ZKS::GetIcmpParamProbCode(in_code);
	};
	return "unknown";
}

extern inline ZKSTD::string 
Ip2String( const bool in_bUseDns, const unsigned int in_ip, time_t& out_durationDns )
{	
	hostent*	lpHostEntry = NULL ; 
	in_addr		where;
	where.s_addr = in_ip;

	if(in_bUseDns)
	{
		// Unfortunately, although gethostbyaddr(), is really nice
		// because it might show a actual domain name (eg www.place.com) instead of
		// an IP, it can take too long when a nym is selected.
		// REVIEW - do we want this to happen when a user is active?

		time_t timeStart = 0;
		time_t timeEnd = 0;

		time( &timeStart );
		lpHostEntry = gethostbyaddr((const char *)&where, sizeof(where), AF_INET);
		time( &timeEnd ) ;

		out_durationDns = timeEnd - timeStart;
	}

	if(!lpHostEntry)
	{
		return inet_ntoa(where);
	}
	else 
	{
		return lpHostEntry->h_name;
	}
}

extern inline bool
IsIdentPort(const unsigned short in_port)
{
	if(in_port == 113)
	{
		return true;
	}
	return false;
}

extern inline bool
IsRPCPort(const unsigned short in_port)
{
	if(in_port == 135)
	{
		return true;
	}
	return false;
}

extern inline bool
IsNetBiosPort(const unsigned short in_port)
{
	if(in_port == 137 ||
	   in_port == 138 ||
	   in_port == 139
	   )
	{
		return true;
	}
	return false;
}

extern inline bool
IsDHCPPort(const unsigned short in_port)
{
	if(in_port == 67 ||
	   in_port == 68
	   )
	{
		return true;
	}
	return false;
}

extern inline bool
IsZksTcpInTheClearPort(const unsigned short in_port)
{
	if( (in_port == FR_DFLT_AIP_TCP_PORT) ||
		(in_port == FR_DFLT_SCRIPT_SERVER_TCP_PORT) ||
		(in_port == FR_DFLT_NIQS_SERVER_PORT)		
		)
	{
		return true;
	}
	return false;
}

extern inline bool
IsZksTcpThroughCloudPort(const unsigned short in_port)
{
	if( ( in_port == FR_DFLT_AIP_TCP_PORT )		||
	( in_port == FR_DFLT_TOKEN_SERVER_PORT )	||
	( in_port == FR_DFLT_FMGMAIP_SERVER_PORT )	||
	( in_port == FR_DFLT_NYM_SERVER_PORT )		||
	( in_port == FR_DFLT_POP_PORT )				||
	( in_port == FR_DFLT_NMTA_PORT)
	)
	{
		return true;
	}
	return false;
}

extern inline bool
IsZksUdpPort(const unsigned short in_port)
{
	if( (in_port == FR_DFLT_AIP_UDP_PORT )		||
		( in_port == FR_DFLT_KEYQRY_SERVER_PORT)
		)
	{
		return true;
	}
	return false;
}

extern inline bool
IsZksTcpPortCloudOrNot(const unsigned short in_port)
{
	if( IsZksTcpThroughCloudPort(in_port)		||
		(in_port == FR_DFLT_SCRIPT_SERVER_TCP_PORT)
		)
	{
		return true;
	}
	return false;
}

extern inline bool
IsDefaultIrcPort(const unsigned short in_port)
{
	switch(in_port)
	{
		case( 6660 ) :
		case( 6661 ) :
		case( 6662 ) :
		case( 6663 ) :
		case( 6664 ) :
		case( 6665 ) :
		case( 6666 ) :
		case( 6667 ) :
		case( 6668 ) :
		case( 6669 ) :
		case( 6670 ) :
		case( 7000 ) :
			return true;
	};

	return false;
}

extern inline bool
IsDefaultWebPort(const unsigned short in_port)
{
	switch(in_port)
	{
		case 80:
        case 81:
        case 88:
        case 8000:
        case 8008:
		case 8080:
        case 8088:
        case 8800:
        case 8808:
        case 8880:
        case 8888:
			return true;
	}
	return false;
}

extern inline bool
IsDefaultSslPort(const unsigned short in_port)
{
	return (443 == in_port);
}

extern inline bool
IsDefaultSshPort(const unsigned short in_port)
{
	return (22 == in_port);
}

extern inline bool
IsDefaultTelnetPort(const unsigned short in_port)
{
	return (23 == in_port);
}

extern inline bool
IsDefaultNewsPort(const unsigned short in_port)
{
	return (119 == in_port);
}

extern inline bool
IsDefaultSmtpPort(const unsigned short in_port)
{
	return (25 == in_port);
}

extern inline bool
IsDefaultPop3Port(const unsigned short in_port)
{
	return (110 == in_port);
}

extern inline bool
IsProxyPortDefaultForTheCloud(const unsigned short in_port)
{
	if(	ZKS::IsDefaultIrcPort(in_port) ||
		ZKS::IsDefaultWebPort(in_port) ||
		ZKS::IsDefaultSslPort(in_port) || 
		ZKS::IsDefaultSshPort(in_port) ||
		ZKS::IsDefaultTelnetPort(in_port)
		)
	{
		return true;
	}
	return false;
}

extern inline ZKSTD::string
ConvertFirewallAppPersistIntoString(const FirewallAppPersist& in_fap)
{
	ZKSTD::ostringstream ostr;

	ostr << "App [" << in_fap.app << "] ";

	ostr << "Incoming=> ";
	if(block == in_fap.incoming.Sel)
	{
		ostr << "block ";
	}
	else if(allow == in_fap.incoming.Sel)
	{
		ostr << "allow ";
	}
	else
	{
		ostr << "unknown ";
	}

	if(in_fap.incoming.bRemember)
	{
		"don't ask again ";
	}
	else
	{
		"ask again ";
	}

	ostr << "Outgoing=> ";
	if(block == in_fap.outgoing.Sel)
	{
		ostr << "block ";
	}
	else if(allow == in_fap.outgoing.Sel)
	{
		ostr << "allow ";
	}
	else
	{
		ostr << "unknown ";
	}

	if(in_fap.outgoing.bRemember)
	{
		"don't ask again ";
	}
	else
	{
		"ask again ";
	}

	return ostr.str();
}

extern inline void
DumpUdpHeader(const ZkUDPHeader* in_udph)
{
	ZkIPHeader* iph = (ZkIPHeader*)in_udph;

	ZKTRACE1("  IP Packet Length        %d bytes\n",iph->ihl * 4);
	ZKTRACE1("  Version                 %d\n",  iph->version);
	ZKTRACE1("  Type Of Service         %d\n",  iph->TypeOfSvr);
	ZKTRACE1("  Total Length            %d\n",  ntohs(iph->TotalLength));
	ZKTRACE1("  Identification          %d\n",  ntohs(iph->Identification));
	ZKTRACE1("  Flags                   0x%X\n",iph->Flags_FrameOffs);
	ZKTRACE1("  Time To Live            %d\n",  (unsigned short)iph->TimeToLive);
	ZKTRACE1("  Protocol                %d\n",  iph->Protocol);
	ZKTRACE1("  IP Header Checksum      %d\n",  (unsigned short)iph->HeaderChkSum);

	ZKTRACE1("  Source IP Address       %s\n",  ZKS::Cvt_int32toString(iph->SrcIpAddress).c_str());
	ZKTRACE1("  Destination IP Address  %s\n",  ZKS::Cvt_int32toString(iph->DstIpAddress).c_str());

	if(IPPROTO_UDP == iph->Protocol)
	{
		ZKTRACE1("  Source Port             %d\n",ntohs(in_udph->SrcPort));
		ZKTRACE1("  Destination Port        %d\n",ntohs(in_udph->DstPort));
	}
}

extern inline void
DumpIpHeaderToStream(const ZkIPHeader* in_iph, ZKSTD::ostringstream& out_ostr)
{		
	out_ostr << "  IP Packet Length         " << in_iph->ihl * 4 << "bytes"		<< ZKSTD::endl;
	out_ostr << "  Version                  " << (int)in_iph->version			<< ZKSTD::endl;
	out_ostr << "  Type Of Service          " << (int)in_iph->TypeOfSvr			<< ZKSTD::endl;
	out_ostr << "  Total Length             " << ntohs(in_iph->TotalLength)		<< ZKSTD::endl;

	out_ostr << "  Identification           " << ntohs(in_iph->Identification)			<< ZKSTD::endl;
	out_ostr << "  Flags                    " << in_iph->Flags_FrameOffs				<< ZKSTD::endl;
	out_ostr << "  Time To Live             " << (unsigned short)in_iph->TimeToLive		<< ZKSTD::endl;
	out_ostr << "  Protocol                 " << in_iph->Protocol						<< ZKSTD::endl;
	out_ostr << "  IP Header Checksum       " << (unsigned short)in_iph->HeaderChkSum	<< ZKSTD::endl;

	out_ostr << "  Source IP Address       " << ZKS::Cvt_int32toString(in_iph->SrcIpAddress).c_str()	<< ZKSTD::endl;
	out_ostr << "  Destination IP Address  " << ZKS::Cvt_int32toString(in_iph->DstIpAddress).c_str()	<< ZKSTD::endl;
}

extern inline void
DumpUdpHeaderToFile(const ZkUDPHeader* in_udph, ZKSTD::ofstream& out_os)
{
	ZkIPHeader* iph = (ZkIPHeader*)in_udph;

	ZKSTD::ostringstream ostr;

	ZKS::DumpIpHeaderToStream(iph, ostr);

	ostr << "  Source Port             " << ntohs(in_udph->SrcPort) << ZKSTD::endl;
	ostr << "  Destination Port        " << ntohs(in_udph->DstPort) << ZKSTD::endl;

	ostr << ZKSTD::endl << ZKSTD::endl;
	
	out_os << ostr.str();
}

extern inline void
DumptcpHeaderToFile(const ZkTCPHeader* in_tcph, ZKSTD::ofstream& out_os)
{
	ZkIPHeader* iph = (ZkIPHeader*)in_tcph;

	ZKSTD::ostringstream ostr;

	ZKS::DumpIpHeaderToStream(iph, ostr);

	static const char yn[2] = { ' ', 'X' };
	ostr << "  Source Port             " << ntohs(in_tcph->SrcPort)		<< ZKSTD::endl;
	ostr << "  Destination Port        " << ntohs(in_tcph->DstPort)		<< ZKSTD::endl;
	ostr << "  Sequence Number         " << ntohl(in_tcph->SequenceNumber)	<< ZKSTD::endl;
	ostr << "  Acknowledgement number  " << ntohl(in_tcph->AckNumber)		<< ZKSTD::endl;
	ostr << "  Reserved1               " << in_tcph->res1					<< ZKSTD::endl;
	ostr << "  TCP Header Length       " << (unsigned short)((unsigned short)in_tcph->doff * 4) << ZKSTD::endl;
	ostr << "  FIN flag                " << yn[(in_tcph->fin) & (1 << 0)]	<< ZKSTD::endl;
	ostr << "  SYN flag                " << yn[(in_tcph->syn) & (1 << 0)]	<< ZKSTD::endl;
	ostr << "  RST flag                " << yn[(in_tcph->rst) & (1 << 0)]	<< ZKSTD::endl;
	ostr << "  PSH flag                " << yn[(in_tcph->psh) & (1 << 0)]	<< ZKSTD::endl;
	ostr << "  ACK flag                " << yn[(in_tcph->ack) & (1 << 0)]	<< ZKSTD::endl;
	ostr << "  URG flag                " << yn[(in_tcph->urg) & (1 << 0)]	<< ZKSTD::endl;
	ostr << "  Reserved2               " << in_tcph->res2					<< ZKSTD::endl;
	ostr << "  Window                  " << htons(in_tcph->Window)			<< ZKSTD::endl;
	ostr << "  Header Checksum         " << (unsigned short)in_tcph->TcpCheckSum << ZKSTD::endl;
	ostr << "  Urgent Pointer          " << ntohs(in_tcph->UrgentPointer)	<< ZKSTD::endl;

	ostr << ZKSTD::endl << ZKSTD::endl;

	out_os << ostr.str();
}

extern inline void 
DumptcpHeader(const ZkTCPHeader* in_tcph)
{

	ZkIPHeader* iph = (ZkIPHeader*)in_tcph;

	ZKTRACE1("  IP Packet Length        %d bytes\n",iph->ihl * 4);
	ZKTRACE1("  Version                 %d\n",  iph->version);
	ZKTRACE1("  Type Of Service         %d\n",  iph->TypeOfSvr);
	ZKTRACE1("  Total Length            %d\n",  ntohs(iph->TotalLength));
	ZKTRACE1("  Identification          %d\n",  ntohs(iph->Identification));
	ZKTRACE1("  Flags                   0x%X\n",iph->Flags_FrameOffs);
	ZKTRACE1("  Time To Live            %d\n",  (unsigned short)iph->TimeToLive);
	ZKTRACE1("  Protocol                %d\n",  iph->Protocol);
	ZKTRACE1("  IP Header Checksum      %d\n",  (unsigned short)iph->HeaderChkSum);

	ZKTRACE1("  Source IP Address       %s\n",  ZKS::Cvt_int32toString(iph->SrcIpAddress).c_str());
	ZKTRACE1("  Destination IP Address  %s\n",  ZKS::Cvt_int32toString(iph->DstIpAddress).c_str());

	if(IPPROTO_TCP == iph->Protocol)
	{
		static const char yn[2] = { ' ', 'X' };
		ZKTRACE1("  Source Port             %d\n",ntohs(in_tcph->SrcPort));
		ZKTRACE1("  Destination Port        %d\n",ntohs(in_tcph->DstPort));
		ZKTRACE1("  Sequence Number         %lu\n",ntohl(in_tcph->SequenceNumber));
		ZKTRACE1("  Acknowledgement number  %lu\n",ntohl(in_tcph->AckNumber));
		ZKTRACE1("  Reserved1               %d\n",in_tcph->res1);
		ZKTRACE1("  TCP Header Length           %d\n",(unsigned short)((unsigned short)in_tcph->doff * 4) );
		ZKTRACE1("  FIN flag                %c\n",yn[(in_tcph->fin) & (1 << 0)]);
		ZKTRACE1("  SYN flag                %c\n",yn[(in_tcph->syn) & (1 << 0)]);
		ZKTRACE1("  RST flag                %c\n",yn[(in_tcph->rst) & (1 << 0)]);
		ZKTRACE1("  PSH flag                %c\n",yn[(in_tcph->psh) & (1 << 0)]);
		ZKTRACE1("  ACK flag                %c\n",yn[(in_tcph->ack) & (1 << 0)]);
		ZKTRACE1("  URG flag                %c\n",yn[(in_tcph->urg) & (1 << 0)]);
		ZKTRACE1("  Reserved2               %d\n",in_tcph->res2);
		ZKTRACE1("  Window                  %d\n",htons(in_tcph->Window));
		ZKTRACE1("  Header Checksum         %d\n",(unsigned short)in_tcph->TcpCheckSum);
		ZKTRACE1("  Urgent Pointer          %d\n",ntohs(in_tcph->UrgentPointer));
	}
}


extern inline int 
GetSocketType(const int in_sock)
{
	int type = 0;
	int size = sizeof(type);
	if( 0 == getsockopt(in_sock, SOL_SOCKET, SO_TYPE, (char*)&type, &size) )
	{
		ZKTRACE2("The socket [%d] is of type [%d]\n", in_sock, type);
	}
	return type;
}


}	// namespace ZKS
