// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):



//===========================================================================
// EXTERNAL DECLARATIONS
//===========================================================================

#include <ZkTest.h>
#include <clientdebug.h>

#include <fstream>
#include <strstream>

//===========================================================================
// CLASS ZkTest
//===========================================================================

ZkTestBase* ZkTest::s_pInitTest = NULL;

//---------------------------------------------------------------------------
//
ZkTest& ZkTest::Instance()
{
	static ZkTest theInstance;
	return theInstance;
}

//---------------------------------------------------------------------------
//
ZkTest::ZkTest( )
	: m_pCallBackFunc( NULL )
	, m_bAbortOnInit( true )
{
}

//---------------------------------------------------------------------------
//
ZkTest::~ZkTest( )
{
}

//---------------------------------------------------------------------------
//
void ZkTest::RegisterTest( ZkTestBase* in_pTest )
{
	if( !in_pTest )
	{
		ZKASSERT( in_pTest );
		return;
	}
	// Registering twice is perfectly legal.
	Instance().m_tests.push_back( in_pTest );
}

//---------------------------------------------------------------------------
//
void ZkTest::RegisterInit( ZkTestBase* in_pInitTest )
{
    TST_FATAL_ASSERT( s_pInitTest == NULL );

	s_pInitTest = in_pInitTest;
}

//---------------------------------------------------------------------------
//
long ZkTest::RunTests( ZkTestBase::TestType in_testType, void* in_pParam )
{
	ZkTestBase::SetTrace( false );

	ZKTRACE0( "\n\n" );

	// Perform any registered intialization
	RunInitTest();
	
	long lRetVal( 0 );

	TestList::iterator iter;
	for( iter = Instance().m_tests.begin(); iter != Instance().m_tests.end(); ++iter )
	{
		if( in_testType & (*iter)->GetType() )
		{
			try 
			{
				ZKTRACE1( "\n** Running Test: %s", (*iter)->GetName() );

				if( Instance().m_pCallBackFunc )
					Instance().m_pCallBackFunc( (*iter)->GetDescription() );

				lRetVal = (*iter)->Execute( in_pParam );

				if( lRetVal == 0 )
					throw ZkTestEx();

				ZKTRACE0( " - PASSED **\n" );
			}
			catch( ZkTestEx& ) 
			{
				ZKTRACE0( " - FAILED **\n" );
			}
			catch( ... ) 
			{
				ZKTRACE0( "Caught some unexpected exception.\n" );
			}
		}
		else
			ZKTRACE1( "\n** Test %s not run because it is not of the type specified.", 
						(*iter)->GetName() );
	}

	ZKTRACE0( "\n\n" );

	return lRetVal;
}

//---------------------------------------------------------------------------
//

long ZkTest::RunTest( const char* in_pTestName, void* in_pParam )
{
	ZKTRACE0( "\n\n" );

	// Perform any registered intialization
	RunInitTest();

	long lRetVal( 0 );
	bool bTestFound( false );

	TestList::iterator iter;
	for( iter = Instance().m_tests.begin(); iter != Instance().m_tests.end(); ++iter )
	{
		if( !::strcmp( in_pTestName, (*iter)->GetName() ) ) 
		{
			bTestFound = true;
			try 
			{
				ZKTRACE1( "\n** Running Test: %s **\n", (*iter)->GetName() );

				if( Instance().m_pCallBackFunc )
					Instance().m_pCallBackFunc( (*iter)->GetDescription() );

				lRetVal = (*iter)->Execute( in_pParam );

				if( lRetVal == 0 )
					throw ZkTestEx();
			}
			catch( ZkTestEx& ) 
			{
				ZKTRACE0( "** TEST FAILED **" );
			}
			catch( ... ) 
			{
				ZKTRACE0( "Caught some unexpected exception.\n** TEST FAILED **\n" );
			}
		}
	}
	
	ZKTRACE0( "\n\n" );

	if( !bTestFound )
		lRetVal = -1;	// Test not found.

	return lRetVal;
}

//---------------------------------------------------------------------------
//
long ZkTest::RunTestID( unsigned int in_testID, void* in_pParam )
{
	ZKSTD::ostrstream stringStream;

	ZKTRACE0( "\n\n" );

	// Perform any registered intialization
	long lRetVal = RunInitTest();
	if( lRetVal == 0 )
		return lRetVal;

	bool bTestFound( false );

	if( in_testID < Instance().m_tests.size() )
	{
		bTestFound = true;
		try 
		{
			ZKTRACE1( "\n** Running Test: %s **\n", Instance().m_tests[in_testID]->GetName() );

			if( Instance().m_pCallBackFunc )
				Instance().m_pCallBackFunc( Instance().m_tests[in_testID]->GetDescription() );

			lRetVal = Instance().m_tests[in_testID]->Execute( in_pParam );

			if( lRetVal == 0 )
				throw ZkTestEx();
		}
		catch( ZkTestEx& ) 
		{
			ZKTRACE0( "** TEST FAILED **" );
		}
		catch( ... ) 
		{
			ZKTRACE0( "Caught some unexpected exception.\n** TEST FAILED **" );
		}
	}

	if( !bTestFound )
		lRetVal = -1;

	return lRetVal;
}

//---------------------------------------------------------------------------
//
ZkTestBase *ZkTest::GetTestID( const unsigned int in_testID )
{
 	if( in_testID < Instance().m_tests.size() ) return Instance().m_tests[in_testID];
    else return NULL;
}

//---------------------------------------------------------------------------
//
long ZkTest::RunInitTest()
{
	// Make sure we return 1 if we do not execute de test. 
	long lRetVal = s_pInitTest ? 0 : 1;

	if( s_pInitTest ) 
	{
		try 
		{
			ZKTRACE0( "** Test Initialization." );

			lRetVal = s_pInitTest->Execute();

			if( lRetVal == 0 )
				throw ZkTestEx();

			ZKTRACE0( " - done **\n" );
		}
		catch( ZkTestEx& ) 
		{
			ZKTRACE0( " - failed **\nInitialization failed, so let's bail out.\n" );

			if( Instance().m_bAbortOnInit )
				::abort();
		}
		catch( ... )
		{
			ZKTRACE0( " - failed **\nCaught some unexpected exception during "
						"initialization failed, so let's bail out." );

			if( Instance().m_bAbortOnInit )
				::abort();
		}
	}

	return lRetVal;
}

//---------------------------------------------------------------------------
//
void ZkTest::ExportTestsToFile( const char* in_pFileName )
{
	ZKSTD::ofstream fileStream( in_pFileName ); 	// truncate the file by default.
	if( !fileStream.is_open() )
	{
		ZKASSERT( !"The directory doesn't exist: create the directory." );
		return;
	}

	for( unsigned int i = 0; i < Instance().m_tests.size(); ++i )
	{
		ExportTestToFile( Instance().m_tests[i], i, fileStream );
		fileStream << ZKSTD::endl;
	}

	fileStream.close();
}

//---------------------------------------------------------------------------
//
void ZkTest::ExportTestToFile( 
	const ZkTestBase* in_pBaseTest, 
	unsigned int in_testID, 
	ZKSTD::ostream& io_rStream )
{
	if( !in_pBaseTest )
	{
		ZKASSERT( false );
		return;
	}

	io_rStream << in_pBaseTest->GetName() << "\t" 
		<< in_pBaseTest->GetDescription() << "\t" 
		<< in_testID;
}

//---------------------------------------------------------------------------
//
void ZkTest::SetCallBackFunc( void (*in_pFunc)( const char*) )
{
	Instance().m_pCallBackFunc = in_pFunc;
}
//---------------------------------------------------------------------------
//
void ZkTest::SetAbortOnFailedInit( bool in_bAbortOnInit )
{
	Instance().m_bAbortOnInit = in_bAbortOnInit;
}

