/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef ZKS_INCL_ZKTEST
#define ZKS_INCL_ZKTEST

/*===========================================================================
 * EXTERNAL DECLARATIONS
 *=========================================================================*/

#include "LibZkTestInc.h"
#include "ZkTestBase.h"

#include <clientstl.h>
#include <vector>

//class ZkTestBase;

/*===========================================================================
 * CLASS ZkTest
 *=========================================================================*/

/*
  Zktest is a singleton that executes a series of unit tests. Unit tests
  register themselves through the registerTest function, which will create
  the ZkTest instance if it doesn't exist.

  ZkTest can either run all tests or run a single test by its name.
  
  Through a normal use of libZkTest you should never have to worry about
  ZkTest. ZkTestBase objects register themselves automatically, and the main()
  provided by the library takes care of calling runTest.
*/

class LIBZKTEST_DLL ZkTest 
{
public: 

    /*----- types and enumerations ----*/

    typedef ZKSTD::vector< ZkTestBase* > TestList;

    /*----- static member functions -----*/
      
    // PURPOSE: Get the one and single instance of ZkTest
	// REQUIRE: Nothing
	// PROMISE: Create the instance the first time it is called and return this
	//			created instance each subsequent time.
	static ZkTest& Instance();
    
	// PURPOSE: Register a ZkTestBase as a normal test to run.
	// REQUIRE: A valid pointer to a ZkTestBase.
	// PROMISE: Register the test in the list (if the pointer is valid).
	static void RegisterTest( ZkTestBase* in_pTest );
    
	// PURPOSE: Register a ZkTestBase as an initialization test.
	// REQUIRE: A valid ZkTestBase pointer.
	//			Only 1 initialization test can be registered.
	// PROMISE: Register the test as the test to run before any other test.
	static void RegisterInit( ZkTestBase* in_pInitTest );
    
	// PURPOSE: Run a test of the type "in_testType" passing it the parameter "in_pParam".
	// REQUIRE: Type of the test (default ALL).
	//			The parameter void* (default NULL)
	// PROMISE: Iterate and run each registered test of the type specified and call
	//			the set callback function before.
	//			Return 1 on success, 0 on failure.
	static long RunTests( ZkTestBase::TestType in_testType = ZkTestBase::ALL, 
							void* in_pParam = NULL );

    // PURPOSE: Run a test specified by its name.
	// REQUIRE: Name of an existing test and the parameter to pass.
	// PROMISE: Return 1 on success, 0 on failure, -1 if the test cannot be found.
	static long RunTest( const char* in_pTestName, void* in_pParam = NULL );
	
	// PURPOSE: Run a test specified by its ID (exported by "ExportTestsToFile").
	// REQUIRE: Valid test ID and the parameter to pass.
	// PROMISE: Return 1 on success, 0 on failure, -1 if the test cannot be found.
	static long RunTestID( unsigned int in_testID, void* in_pParam = NULL );

	// Purpose: Retrive a test by ID
	// Promise: Return valid pointer to test or NULL if paramter is out of range.
    static ZkTestBase *GetTestID( const unsigned int in_testID );
	
	// PURPOSE: Write the name/description/ID of each test to a file so it can
	//			be used by the test agent.
	// REQUIRE: A file name (and possibly a full path).  If the full path is given,
	//			then it MUST exist or the export will fail.
	// PROMISE: Export all the test to the file specified if the directory exists.
	static void ExportTestsToFile( const char* in_pFileName );
	
	// PURPOSE: Set the pointer to the function that will be called before running each test.
	// REQUIRE: A pointer to a function in the same address space that takes a char* and
	//			return void.
	// PROMISE: The function will be called before calling "Execute" on the test.  It will 
	//			pass the test description as the parameter.
	static void SetCallBackFunc( void (*in_pFunc)( const char*) );

	// PURPOSE: Set ZkTest to abort or not if the initialization test fail.
	// REQUIRE: True or false to make it abort or not.
	// PROMISE: Set the member.
	static void SetAbortOnFailedInit( bool in_bAbortOnInit );
    
    /*----- member functions -----*/
      
    // PURPOSE: Destructor.  Does nothing.
	// REQUIRE: Nothing
	// PROMISE: Nothing
	~ZkTest();

protected: 

    /*----- member functions -----*/

    ZkTest();

private: 

    /*----- member functions -----*/

	// PURPOSE: Run the initialization test.
	// REQUIRE: A registered init test or not.
	// PROMISE: Abort if the test fails if "SetAbortOnFailedInit" has been called with 
	//			true (true by default).
	//			Return 0 if the test fails (but "SetAbortOnFailedInit" has been called with 
	//			false), 1 if the test succeed.
	static long RunInitTest();
	
	// PURPOSE: Write one test in the stream.
	// REQUIRE: A valid pointer to a ZkTestBase to be written.
	//			A unique ID for the test.
	//			An ostream to write to.
	// PROMISE: Will write/format the information of the test into the stream.
	//			Info are full name, description and ID provided.
	static void ExportTestToFile( const ZkTestBase* in_pBaseTest,
									unsigned int in_testID,
									ZKSTD::ostream& io_rFileStream );

	/*----- data members -----*/

    TestList m_tests;

	void (*m_pCallBackFunc)( const char* pTestDesc );
	bool m_bAbortOnInit;

    static ZkTestBase* s_pInitTest;
};


#endif
