/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef ZKS_INCL_ZKTESTBASE
#define ZKS_INCL_ZKTESTBASE

/*===========================================================================
 * CLASS ZkTestBase
 *=========================================================================*/

/*
  This class provides an api to create automated unit tests. 
  A ZktestBase object automatically registers itself with a ZkTest singleton.
  ZkTest is then called to execute all registered tests.

  To create a test, simply derive from ZkTestBas and override the execute()
  function ( execute() should soon support runtime arguments ). Then you
  must create a static instance so that the test registers itself at runtime.

  If you do not need any data members and simply want to override execute(),
  you can use the ZKTEST_START and ZKTEST_END macros, they will take care of 
  the class declaration and function override, all you need to do is type the 
  content of execute()in between the two macros.

  For instance, to create a test named ClassATest
  Your .cpp would be something like:

  ---

  #include <ZkTestBase.h>

  ZKTEST_START( ClassATest )

	 // your test code here
	 ClassA instance;
	 instance.someCall();
	 
	 // some tests 


  ZKTEST_END

  ---

  You could also have more than one test in your file. For instance, you 
  could have test 1 and test 2 for your class.

  ---

  #include <ZkTestBase.h>

  ZKTEST_START( ClassATest1 )

  // your test 1 code here

  ZKTEST_END

  ZKTEST_START( ClassATest2 )

  // your test 2 code here.

  ZKTEST_END

  ---

  The advantage of this second example is that you can then 
  have more control over what you run. Lets say this your test was compile as
  UnitTests. You could then run all tests by calling:

  > UnitTests

  But you could also execute only the 2 by calling:

  > UnitTests ClassATest2
  

  Trace Utility:

  While in a test you should use the test library TST_TRACE and TST_ASSERT 
  macros. 
  - TST_TRACE uses ZKTRACE0 and takes care of the end of line.
  - TST_ASSERT throws a TstException instead of calling abort and allows to 
	run all tests before exit.
  - TST_FATAL_ASSERT can be used if an assert should terminate the program 
	because all remaining test depend on it.

*/
#include "LibZkTestInc.h"
#include <clientdebug.h>

#include <iostream>
#include <strstream>
#include <clientstl.h>

class LIBZKTEST_DLL ZkTestBase 
{
public: 
	/*----- types and enumerations ----*/

	// PURPOSE: Specifie the type of test.
	// PROMISE: These are flags that can be check with & operator.
	enum TestType
	{
		STANDARD = 0x1,
		INTERACTION = 0x2,
		ALL = static_cast<unsigned int>( -1 )
	};

	/*----- member functions -----*/

	// PURPOSE: Constructor.
	// REQUIRE: Name of the test that should be in the form "component/sub-component/TestName".
	//			Boolean set to true if this test is supposed to be run before anyone else 
	//			because it performs initialization.
	// PROMISE: Register this test to the ZkTest singleton.
	ZkTestBase( const char* in_pName, bool in_bInitialize = false );

	// PURPOSE: Constructor.
	// REQUIRE: Name of the test that should be in the form "component/sub-component/TestName".
	//			Description of the test.  It should be very explicit since it can be retreived 
	//			for display purposes (test agent for example).
	//			The type of this test.  Test that need the user to click on button to continue the 
	//			test should be of type INTERACTION.  Other test should be STANDARD.
	//			Boolean set to true if this test is supposed to be run before anyone else 
	//			because it performs initialization.
	// PROMISE: Register this test to the ZkTest singleton.
	ZkTestBase( const char* in_pName, 
				const char* in_pDescr,
				TestType in_testType = STANDARD,
				bool in_bInitialize = false );

	// PURPOSE: Destructor.  It does nothing.
	// REQUIRE: Nothing
	// PROMISE: Nothing
	virtual ~ZkTestBase( );

	// PURPOSE: Entry point in the class derived from ZkTestBase to run the test.
	// REQUIRE: Must be implemented in the derived class.
	//			The parameter can be used in the way the class wants but the 
	//			developer is responsible to make sure it is valid.
	// PROMISE: Depends on the implementation.
	virtual long Execute( void* in_pParam = NULL ) = 0; 	

	// PURPOSE: Retreive the name of the test.
	// REQUIRE: Nothing
	// PROMISE: A valid pointer to a char* (maybe an empty string).
	const char* GetName() const;

	// PURPOSE: Retreive the description of the test.
	// REQUIRE: Nothing
	// PROMISE: A valid pointer to a char* (maybe an empty string).
	const char* GetDescription() const;

	// PURPOSE: Retreive the type of the test.
	// REQUIRE: Nothing
	// PROMISE: The type of the type test.
	TestType GetType() const;

	// PURPOSE: Will trace occur inside the test?
	// REQUIRE: Nothing
	// PROMISE: The value of the member.
	static bool GetTrace( );

	// PURPOSE: Set the trace inside the test.
	// REQUIRE: True to set the trace || false to unset.
	// PROMISE: Set the member.
	static void SetTrace( bool in_bTrace );

private: 

	/*----- member functions -----*/

	// PURPOSE: Finalize initalization.
	// REQUIRE: Be called in the constructors.
	// PROMISE: Finalize initalization.
	void Init( bool in_bInitialize );

	/*----- data members -----*/

	const char* m_strName;
	const char* m_strDescription;
	const TestType m_testType;

	static bool s_bTrace; 
};

//---------------------------------------------------------------------------
//
class ZkTestEx {};

//---------------------------------------------------------------------------
//
inline void ZkTestBase::SetTrace( bool in_bTrace )
{
	s_bTrace = in_bTrace;
}

//---------------------------------------------------------------------------
//
inline bool ZkTestBase::GetTrace(  )
{
	return s_bTrace;
}

/*===========================================================================
 * Useful Macros
 *=========================================================================*/

/* 
   This will declare your test class for you and create the static instance
   to register with ZkTest.
   All you have to do now is override execute()
   
   NOTE: don't forget to close it with ZKTEST_END
*/

#define ZKTEST_START( name ) \
class name : public ZkTestBase {\
public:\
	name () : ZkTestBase( #name, false ){} \
	virtual long Execute( void* in_pParam = NULL );\
};\
name Instance##name;\
\
long name ::Execute( void* in_pParam ) \
{\
    (void)in_pParam;


/*
  This macro defines a test that you want registered as an initializer.
  An initailizer is not a test, it is used to set some states and initialize
  libraries common to all tests in the group.

   NOTE: don't forget to close it with ZKTEST_END
*/
#define ZKTEST_INIT( name ) \
class name : public ZkTestBase {\
public:\
	name () : ZkTestBase( #name , true ){} \
	virtual long Execute( void* in_pParam = NULL );\
};\
name Instance##name;\
\
long name ::Execute( void* in_pParam ) \
{ \
    (void)in_pParam;


/*
  Simply defines the end or your test. 
  (it actually translates to the end of the execute() function)
*/
#define ZKTEST_END }




#define TST_TRACE( os ) \
{ \
	if( ZkTestBase::GetTrace( ) ) \
	{ \
		ZKSTD::ostrstream stringStream; \
		stringStream << os << ZKSTD::endl; \
		ZKTRACE0( stringStream.str() ); \
	} \
} \

#define TST_ASSERT( cond ) \
	{if (!(cond)) { TST_TRACE(	"ASSERTION FAILED at " << __FILE__ << ":" <<  __LINE__	<< " (" #cond << ")" ); throw ZkTestEx(); }}
#define TST_FATAL_ASSERT( cond ) \
	{if (!(cond)) { TST_TRACE(	"ASSERTION FAILED (fatal) at " << __FILE__ << ":" <<  __LINE__	<< " (" #cond << ")" ); abort();}}
	  


#endif


/*
#ifdef INTERNAL_BUILD

#include <zkbase/libzktest/ZkTestBase.h>	// This _MUST_ be enclosed in INTERNAL_BUILD.

class YourClassName : public ZkTestBase
{
	YourClassName()
		: public ZkTestBase( "YourComponent/YourSubComponent/YourClassName",
								"Description of YourClassName role",
								ZkTestBase::typeStandard || typeInteraction )	// One of these.
	{}

	~YourClassName()
	{}

	virtual long Execute( void* )
	{ ... }
};

YourClassName g_yourClassName;	// Should be in a .cpp file to avoid multiple instanciation.

#endif // INTERNAL_BUILD

*/
