// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// AciInfo.h

// An ACI represents a Route.  In other words this 
// class tracks information about a particular route

// The information it contains works hand in hand with the
// information inside the frLocalAipCtx.  We try to keep this
// class as small as possible.

// The AciInfo tracks the life cycle of a route, its duration,
// its type (Authenticated versus Unauthenticated)

// REVIEW: Some of the information inside this class could be
// obtained by looking inside the frLocalAipCtx structures,
// by writing some accessor functions in that library we could
// reduce this class

#ifndef ACIINFO_H_
#define ACIINFO_H_

#include "ClientTypesInc.h"

#include <zkError.h>
#include <frAips.h>
#include <string>
#include <zksocket.h>
#include <client.h>	// for ZKSTD

#ifdef WIN32
#	if _MSC_VER >= 1000
#	pragma once
#	endif // _MSC_VER >= 1000
#endif

class ClientTypes_API AciInfo  
{

public:
	AciInfo();
	virtual ~AciInfo();

	// Reset the contents
	void Clear() ;

	// The Route goes through a life cycle as described in the 

	bool				IsValid() const		{ return m_RouteState != RouteInvalid ; } 
	bool				IsActive() const	{ return m_RouteState == RouteActive ; }
	RouteState			GetRouteState() const { return m_RouteState ; }

	// You can determine if a route is active or not with this method
	RouteType			GetRouteType() const { return m_RouteType ; }

	const time_t		GetStartTime()    const { return m_StartTime;    }
	const int			GetDuration()     const { return m_nDuration;    }
	const SOCKADDR_IN&	GetFirstAipAddr() const { return m_FirstAipAddr; }



	// Set Methods

	void				SetRouteValidNotActive() { m_RouteState = RouteValidNotActive ; }
	void				SetRouteActive()  { m_RouteState = RouteActive ; }
	void				SetRouteRetired() { m_RouteState = RouteRetired ; }

	void				SetRouteType( RouteType in_RouteType ) { m_RouteType = in_RouteType ; }

	void				SetStartTime( time_t in_StartTime ) { m_StartTime = in_StartTime ; }
	void				SetDuration( int in_Duration ) { m_nDuration = in_Duration ; }	// REVIEW - make it unsigned
	void				SetFirstAipAddr( sockaddr_in in_FirstAipAddr ) { m_FirstAipAddr = in_FirstAipAddr ; }


	// Authenticated Route Methods

	size_t				GetNymIndex() const { ZKASSERT( m_RouteType == AuthenticatedRoute ) ;
											  return m_iNym ; }
	void				SetNymIndex( size_t in_iNym ) { m_iNym = in_iNym ;}

	// Unauthenticated Route Methods
	uint16_t			GetDestinationPort() const { ZKASSERT( m_RouteType == UnauthenticatedRoute ) ;
													 return m_DestPort ; }

	void				SetDestinationPort( uint16_t in_DestPort ) { m_DestPort = in_DestPort ; } ;


	// Show the contents
	void				TraceState() const ;

    // Accessor & Mutator for m_RouteIdleCounter  
    void                SetRouteIdleCounter( const unsigned long count ) { m_RouteIdleCounter = count ; } ;
    unsigned long       GetRouteIdleCounter( ) const { return m_RouteIdleCounter ; } ;

private:

	RouteState			m_RouteState ;

	RouteType			m_RouteType ;		
	int					m_nDuration;		// in Seconds, default in BuildRoute() is DEFAULT_ROUTE_SECS
	time_t				m_StartTime;		// in Seconds, use this to calculate the existence of the context
	SOCKADDR_IN			m_FirstAipAddr;		// Address of first hop 

	// Members valid for Authenticated Routes only
	size_t				m_iNym ;

	// Members valid for Unauthenticated route only
	uint16_t            m_DestPort ;

    // Counter for route idle time. Used by RouteManager / ActiveRouteTable to tear down
    // routes which have been idle for a specified amount of time. The contents of this variable
    // are number of ticks, not any meaningful quantity.
    unsigned long       m_RouteIdleCounter ;
};

#endif

