// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "StdAfx.h"
#include <clientdefs.h>

#ifdef INTERNAL_BUILD

#include "WCoreTypes.h"
#include "WAutoBuffer.h"
#include "WReadBuffer.h"
#include "ClientTypesUnitTest.h"
#include "WCypherMailMsg.h"
#include "ZKChunk.h"
#include "ZKDocument.h"
#include "ZKConvert.h"
#include "WRoute.h"
#include "WNetInfoDatum.h"
#include "WNetInfo.h"
#include "WZkipClient.h"
#include "WNymIconInfo.h"
#include "IZkFileHelp.h"
#include <IZkException.h>
#include <zkipClient.h>

// Known values stored into chunk tests

const uint8_t        MagicUInt8 = 235 ;		// Test unsigned char support
const uint8_t        MagicUInt8Char = 'i' ;
const int8_t         MagicInt8 = -125 ;

const uint16_t       MagicUInt16 = 65530 ;
const int16_t        MagicInt16 = -32000 ;

const uint32_t       MagicUInt32 = 45 ;
const uint32_t       MagicUInt32Big = 3000000000 ;
const int32_t        MagicInt32		= -2000000000;

const uint64_t       MagicUInt64 = 18446744073709551610 ;  // number of bugs in bugfree by version 3
const int64_t        MagicInt64  = -9223372036854775800 ;

const uint32_t       Dummy  = 555 ;    // We won't read this

const uint8_t        MagicArray[]  = "Test This Carefully" ;
const ZKSTD::string    MagicString( "Do not eat the small green tablets" ) ;
//const ZKSTD::string    MagicUserString( "This could contain code-page strings" ) ;
const ZKSTD::string    MagicUserString( "Une le plutt agrable  voir en t" ) ;


const uint8_t        MagicBinaryArray[] = { 200, 201, 202, 100 } ;

// String containing some character from the upper ascii,
// which will be different depending on the code page
// Not sure how to test this - the goal is to prove
// that saving an accented e on the PC would end up accented
// e on the Mac also

const uint8_t         MagicLocalizedString[]= { 130, // e with accent
											  156, // English Pound
											  148, // o with double accent
											  't',
											  168, // upsidedown question mark
											  0 } ;

const ZKSTD::string	 hdrTestDesc( "Client Types Test File" ) ;
const ZKSTD::string	 hdrTestProduct( "Zero Knowledge Freedom Client" ) ;
const ZKSTD::string	 hdrTestVer( "recent version" ) ;
const ZKSTD::string	 hdrTestPlatform( "Windows" ) ;
const ZKSTD::string	 hdrTestNetwork( "broken.test.net" ) ;


void ClientTypesUnitTest::zkAllocationFunction1( unsigned char ** out_Buffer, size_t * out_Size )
{
	if (( out_Buffer == NULL ) || ( out_Size == NULL ) ) {
		ZKASSERT( false ) ;
		return ;
	}

	// Initialize output
	*out_Buffer = NULL ;
	*out_Size = 0 ;

	unsigned char * pBuffer = ( unsigned char* ) malloc ( 5 ) ;

	if ( pBuffer != NULL ) {

		// Fill with recognizable data
		pBuffer[0] = 0 ;
		pBuffer[1] = 1 ;
		pBuffer[2] = 2 ;
		pBuffer[3] = 3 ;
		pBuffer[4] = 4 ;

		*out_Buffer = pBuffer ;
		*out_Size = 5 ;
	}
}

void ClientTypesUnitTest::zkAllocationFunction2( char ** out_Buffer, size_t * out_Size )
{
	zkAllocationFunction1( (unsigned char**) out_Buffer, out_Size ) ;
}

void ClientTypesUnitTest::zkAllocationFunction3( void ** out_Buffer, size_t * out_Size )
{
	zkAllocationFunction1( (unsigned char**) out_Buffer, out_Size ) ;
}

void ClientTypesUnitTest::TestWAutoBuffer()
{
	WAutoBuffer theBuffer ;

	zkAllocationFunction1( &theBuffer, theBuffer.GetSizePointer() ) ;

	// print binary dump
	for ( unsigned char i = 0 ; i < theBuffer.Size() ; i++ ) {
		ZKTEST( theBuffer[ i ] == i ) ;
	}

	WAutoCharBuffer theBuffer2 ;

	zkAllocationFunction2( &theBuffer2, theBuffer2.GetSizePointer() ) ;

	// print binary dump
	for ( i = 0 ; i < theBuffer2.Size() ; i++ ) {
		ZKTEST( theBuffer2[ i ] == i ) ;
	}

	// Some function might want "void*"  Both
	// classes can appear as a void* buffer

	WAutoBuffer theBuffer3 ;

	zkAllocationFunction3( (void**)&theBuffer3, theBuffer3.GetSizePointer() ) ;

	// print binary dump
	for ( i = 0 ; i < theBuffer3.Size() ; i++ ) {
		ZKTEST( theBuffer3[ i ] == i ) ;
	}

	WAutoCharBuffer theBuffer4 ;

	zkAllocationFunction3( (void**)&theBuffer4, theBuffer4.GetSizePointer() ) ;

	// print binary dump
	for ( i = 0 ; i < theBuffer4.Size() ; i++ ) {
		ZKTEST( theBuffer4[ i ] == i ) ;
	}

	// Test assignment operator
	WAutoBuffer theBuffer5( theBuffer3 ) ;

	// Should do nothing, not leak
	theBuffer5 = theBuffer5 ;
}

void ClientTypesUnitTest::TestWZkipClient()
{
	{
		WZkipClient client;
	}
	{
		WZkipClient client;
		client.DisconnectAndReset();
	}
	{
		WZkipClient client;
		int fd = 0;

		if (zkipClientMakeBlocking(NULL, client, 45)
			&& zkipClientConnectToHost(NULL, client, "www.freedom.net", 80)
			&& zkipClientFd(NULL, client, &fd))
		{
			ZKTEST(zkipClientDisconnect(NULL, client));
			zkipClientDisconnect(NULL, client);
		}
		else
		{
			ZKTEST(false);
		}
	}
	{
		WZkipClient client;

		client.DisconnectAndReset();
		client.DisconnectAndReset();

	}
}

void ClientTypesUnitTest::TestWCypherMailMsg()
{
	WCypherMailMsg msg ;
	
	msg.SelfTest() ;
	msg.DumpToDebugOutput() ;

	// Fill in bogus values, this tests the APIs

	char FakeEncryptedBuffer[] = "Some stuff" ;

	msg.SetEncryptedText( (unsigned char*) FakeEncryptedBuffer, strlen( FakeEncryptedBuffer ) ) ;

	msg.SetFmgSignatureCheckResult( SigCheckSuccessful ) ;

	// Wolly sent a foger
	msg.SetSendingEntity( "Wolly" ) ;
	msg.SetSignatureCheckResult( SigCheckFailed ) ;

	msg.SetReplyBlock( 1 ) ;

	msg.SetRecipientNym( "Neo" ) ;

	msg.GetClearAutoBuffer().CopyBuffer( (unsigned char*)"A Result", 10 ) ;

	msg.SelfTest() ;
	msg.DumpToDebugOutput() ;

}

void ClientTypesUnitTest::WriteChunks( WMemBuf& io_Stream )
{
	// Writer (Need to use scope to ensure chunks closed)
	{
		ZKChunkWriter cw( io_Stream, TEST1_CHUNK_ID, 1 ) ;

		cw.WriteUInt64( MagicUInt64 ) ;
		cw.WriteInt64( MagicInt64 ) ;
		cw.WriteUInt32( MagicUInt32 ) ;
		cw.WriteUInt32( MagicUInt32Big ) ;
		cw.WriteInt32( MagicInt32 ) ;
		cw.WriteUInt16( MagicUInt16 ) ;
		cw.WriteInt16( MagicInt16 ) ;

		cw.WriteArray( MagicArray, sizeof( MagicArray ) ) ;

		cw.WriteNonUIString( MagicString ) ;

		{
			// Nested Chunk - test our "char" support
			ZKChunkWriter cw( io_Stream, TEST2_CHUNK_ID, 124 ) ;

			cw.WriteUInt8( MagicUInt8 ) ;
			cw.WriteUInt8( MagicUInt8Char ) ;
			cw.WriteInt8( MagicInt8 ) ;

			cw.WriteUInt32( Dummy ) ;
		}

		cw.WriteUInt32( Dummy ) ;
	}

	{
		// Demonstrate writing/reading of an unknown number of sub-chunks
		// Only restriction will be that there is no other chunk values

		ZKChunkWriter cw( io_Stream, TEST4_CHUNK_ID, 1 ) ;

		// Save five chunks, but we don't record the 5 anywhere
		// so reader will need to determine itself
		for ( size_t i = 0 ; i < 1 ; i++ )
		{
			ZKChunkWriter cw( io_Stream, TEST5_CHUNK_ID, 1 ) ;
			cw.WriteInt8( MagicInt8 ) ;
		}
	}

	{	
		ZKChunkWriter cw( io_Stream, TEST3_CHUNK_ID, 3 ) ;
		cw.WriteArray( MagicBinaryArray, sizeof( MagicBinaryArray ) ) ;
		cw.WriteString( MagicUserString ) ;

	}

}
void ClientTypesUnitTest::ReadChunks( unsigned char * in_pRawBuffer, size_t in_cntBuffer )
{
	WReadBuffer rb( in_pRawBuffer, in_cntBuffer ) ;

	{
		// Read the first chunk

		ZKChunkReader cr( rb ) ;

		ZKTEST( cr.GetChunkID() == TEST1_CHUNK_ID ) ;
		ZKTEST( cr.GetChunkSize( ) > CHUNK_HEADER_SIZE ) ;
		ZKTEST( cr.GetChunkVersion() == 1 ) ;

		ZKTEST( MagicUInt64 == cr.ReadUInt64() ) ;
		ZKTEST( MagicInt64 == cr.ReadInt64() ) ;

		ZKTEST( MagicUInt32 == cr.ReadUInt32() ) ;
		ZKTEST( MagicUInt32Big == cr.ReadUInt32() ) ;
		ZKTEST( MagicInt32 == cr.ReadInt32() ) ;

		ZKTEST( MagicUInt16 == cr.ReadUInt16() ) ;
		ZKTEST( MagicInt16 == cr.ReadInt16() ) ;

		const char * pArray = NULL ;
		size_t cntArray ;

		cr.ReadArray( (const void**)&pArray, &cntArray ) ;

		ZKTEST( cntArray = sizeof( MagicArray ) ) ;
		ZKTEST( 0 == strcmp( pArray, (char*) MagicArray ) ) ; // We know it is a terminated string

		ZKSTD::string str ;
		cr.ReadNonUIString( str ) ;

		ZKTEST( str == MagicString ) ;

		{
			// Nested Chunk

			ZKChunkReader cr( rb ) ;

			ZKTEST( cr.GetChunkID() == TEST2_CHUNK_ID ) ;
			ZKTEST( cr.GetChunkSize( ) == CHUNK_HEADER_SIZE + 7 ) ;
			ZKTEST( cr.GetChunkVersion() == 124 ) ;

			ZKTEST( MagicUInt8     == cr.ReadUInt8() ) ;
			ZKTEST( MagicUInt8Char == cr.ReadUInt8() ) ;
			ZKTEST( MagicInt8      == cr.ReadInt8() ) ;
			// Skip dummy
		}


		// Skip Dummy - We intentially don't finish reading everything from this stream
	}

	{
		ZKChunkReader cr2( rb ) ;

		ZKTEST( cr2.GetChunkID() == TEST4_CHUNK_ID ) ;

		while( !cr2.IsEndOfChunk() )
		{
			ZKChunkReader cr3( rb ) ;

			ZKTEST( cr3.GetChunkID() == TEST5_CHUNK_ID ) ;
			ZKTEST( 1 == cr3.GetCntRemainingBytesInChunk() ) ;
			ZKTEST( MagicInt8      == cr3.ReadInt8() ) ;

			// Verify that we are at end of chunk
			ZKTEST( 0 == cr3.GetCntRemainingBytesInChunk() ) ;
			ZKTEST( cr3.IsEndOfChunk() ) ;

			try
			{
				// Trying to read should throw, but
				// continuing should work.
				cr3.ReadUInt32() ;
				ZKTEST( false ) ;
			}
			catch( IZkException& )
			{	
			}
		}
	}
//	{
		ZKChunkReader cr( rb ) ;

		ZKTEST( cr.GetChunkID() == TEST3_CHUNK_ID ) ;

		size_t chunkSize = cr.GetChunkSize( ) ;

		// Unicode doesn't necessarily map as 1 byte ansi == 2 bytes unicode.
		// However for the string in question this is a reasonable assumption
		// as the lower bound (we add 2 for the byte order mark)
		size_t unicodeSize = ( MagicUserString.size() + 1 ) * 2 + 2;

		size_t estimatedChunkSize =   CHUNK_HEADER_SIZE + 
									  sizeof( uint32_t ) +                // Size of the array size value
									  sizeof( MagicBinaryArray ) +
									  sizeof( uint32_t ) +
									   + unicodeSize ;//User strings stored as wide chars

		ZKTEST( chunkSize >= estimatedChunkSize ) ; 
		ZKTEST( cr.GetChunkVersion() == 3 ) ;

		const uint8_t * pArray = NULL ;
		size_t cntArray ;

		cr.ReadArray( (const void**)&pArray, &cntArray ) ;

		ZKTEST( pArray != NULL ) ;
		ZKTEST( cntArray == sizeof( MagicBinaryArray ) ) ;
		ZKTEST( MagicBinaryArray[ 1 ] == pArray[ 1 ] ) ;
		ZKTEST( MagicBinaryArray[ 2 ] == pArray[ 2 ] ) ;


		ZKSTD::string testuserstr ;
		cr.ReadString( testuserstr ) ;

		ZKTEST( testuserstr.compare(MagicUserString) == 0 ) ;
//	}
}

void ClientTypesUnitTest::TestChunks()
{
	try 
	{
		// In this test we will stream out some chunk-structured data
		// into a buffer and then carefully read it out again to make
		// sure that the data was not corrupted.

		// This will hold the actual buffer with the 
		// data in it
		WMemBuf b ;
		WriteChunks( b ) ;

		// Test IN-MEMORY reading

		// Prepare the Read Buffer which basically
		// manages a "Read Pointer" through our buffer
		unsigned char * pRawBuffer = NULL ;
		size_t cntBuffer = 0 ;

		DZkError err ;

		zkMemBufGetData( &err, b, (void**)&pRawBuffer, &cntBuffer ) ;

		ZKTEST( pRawBuffer != NULL ) ; 
		ZKTEST( cntBuffer > 0 )  ; 

		ReadChunks( pRawBuffer, cntBuffer ) ;

		// Test Going to a file, the re-reading it

		
		IZkFileHelp::WriteBinaryFile( "chunktest.bin", pRawBuffer, cntBuffer ) ;

		WAutoBuffer RawBufferFromFile ;		

		ZKTEST( IZkFileHelp::ReadBinaryFile( "chunktest.bin", &RawBufferFromFile, RawBufferFromFile.GetSizeRef()) );

		// Permissible for file contents to be slightly larger, but not smaller
		ZKTEST( RawBufferFromFile.Size() >= cntBuffer ) ;

		ReadChunks( RawBufferFromFile, RawBufferFromFile.Size() ) ;
	}
	catch( IZkException& )
	{
		ZKTEST( false ) ;
	}

	// To Do - Encrypt it, Save it and then reload it, then test it,
	// This tests the cross platform encryption algorithms more than the
	// chunk code but it is more realistic
}

void ClientTypesUnitTest::TestWReadBuffer()
{
	// Test WReadBuffer

	// This class is pretty simple, but the idea is to make
	// sure it seeks and reads similar to basic C file i/o.

	unsigned char test[] = { 1, 2, 3, 4 } ;

	WReadBuffer rb( test, sizeof( test ) ) ;
	ZKTEST( !rb.IsEOF() ) ;

	unsigned char val ;
	rb.Read( &val, sizeof( val ) );
	ZKTEST( val == 1 ) ;

	rb.Read( &val, sizeof( val ) );
	ZKTEST( val == 2 ) ;

	rb.Seek( 3 ) ;

	rb.Read( &val, sizeof( val ) );
	ZKTEST( val == 4 ) ;

	ZKTEST( rb.IsEOF() ) ;

	// Try to read at end5 fails without crash
	ZKTEST( 0 == rb.Read( &val, sizeof( val ) ) );

	rb.Seek( 0 ) ;

	rb.Read( &val, sizeof( val ) );
	ZKTEST( val == 1 ) ;
}

void ClientTypesUnitTest::TestZKDocument()
{
	try 
	{	
		uint8_t slt[] = "Salt" ;
		uint8_t tst[] = "This would be an encrypted sample" ;
		uint8_t dat[] = "This would be the encrypted Data" ;

		// Now Create the file

		ZKDocumentWriter dw ;

		dw.SetHeader( hdrTestDesc, 
			          hdrTestProduct, 
					  hdrTestVer,
					  hdrTestPlatform,
					  hdrTestNetwork ) ;

		dw.SetSalt( slt, sizeof( slt ) ) ;
		dw.SetTest( tst, sizeof( tst ) ) ;
		dw.SetData( dat, sizeof( dat ) ) ;

		dw.WriteToFile( "TestFile.dat" ) ;		

		// Failure condition - this file is illegal bogus

		try
		{
			dw.WriteToFile( "" ) ;
			ZKTEST( false ) ;
		}
		catch( IZkException& )
		{
		}

		// Now read the file again

		ZKDocumentReader dr( "TestFile.dat" ) ;

		const uint8_t * pB = NULL ;
		size_t          cntB = 0 ;
		
		dr.GetSalt( &pB, cntB ) ;

		ZKTEST( 0 == strcmp( (char*)slt, (char*)pB ) ) ;

		dr.GetTestBuffer( &pB, cntB ) ;

		ZKTEST( 0 == strcmp( (char*)tst, (char*)pB ) ) ;

		dr.GetDataBuffer( &pB, cntB ) ;

		ZKTEST( 0 == strcmp( (char*)dat, (char*)pB ) ) ;

		// Verify the header

		ZKTEST( hdrTestDesc == dr.GetFileDesc() ) ;
		ZKTEST( hdrTestProduct == dr.GetProductNameDesc() ) ;
		ZKTEST( hdrTestPlatform == dr.GetPlatform() ) ;
		ZKTEST( hdrTestNetwork == dr.GetNetworkName() ) ;
		ZKTEST( hdrTestVer == dr.GetProductVersion() ) ;

	}
	catch( IZkException& )
	{
		ZKTEST( false ) ;
	}
	
	// Now test some nasty corrupt or truncated file conditions
	// and make sure we throw nicely
	// The ZKDocument object will throw since we are writing
	// a bogus size on the file

	unsigned char small_corrupt_contents[] = "X" ;
	unsigned char truncated_contents[] = {  0x3F, 0x2D, 0x1C,	// Header
											1,					// Version
											0, 0, 0, 64 }  ;	// Bogus Size

	unsigned char truncated_contents2[] = { 0x3F, 0x2D, 0x1C,	// Header
											1,					// Version
											0, 0, 0, 64,		// Bogus Size
											12,					// Bogus string size
											'P', 'a', 'r', 't' }  ;	// Partial String

	IZkFileHelp::WriteBinaryFile( "CorruptFile.dat", small_corrupt_contents, sizeof( small_corrupt_contents ) ) ;

	try 
	{
		ZKDocumentReader dr( "CorruptFile.dat" ) ;
		ZKTEST( false ) ;
	}
	catch( IZkException& )
	{
	}

	IZkFileHelp::WriteBinaryFile( "CorruptFile.dat", truncated_contents, sizeof( truncated_contents ) ) ;

	try 
	{
		ZKDocumentReader dr( "CorruptFile.dat" ) ;
		ZKTEST( false ) ;
	}
	catch( IZkException& )
	{
	}

	IZkFileHelp::WriteBinaryFile( "CorruptFile.dat", truncated_contents2, sizeof( truncated_contents2 ) ) ;

	try 
	{
		ZKDocumentReader dr( "CorruptFile.dat" ) ;
		ZKTEST( false ) ;
	}
	catch( IZkException& )
	{
	}

    unlink("CorruptFile.dat");
	try 
	{
		ZKDocumentReader dr( "CorruptFile.dat" ) ;
		ZKTEST( false ) ;
	}
	catch( IZkException& )
	{
	}
}

long ClientTypesUnitTest::Execute( void* in_pParam )
{
	TestWAutoBuffer() ;
	TestWCypherMailMsg() ;

	TestWReadBuffer() ;

	TestChunks() ;

	TestZKDocument() ;

	TestWNymIconInfo() ;

	TestWNetInfo();

	TestWZkipClient();

	WRoute::SelfTest();

	WNetInfoDatum::SelfTest() ;

	ZKTEST( FRENT_TYPE_NYM_SRV == ZKConvert::GetEntityTypeFromPort( FR_DFLT_NYM_SERVER_PORT ) ); 
	ZKTEST( FRENT_TYPE_NIQS == ZKConvert::GetEntityTypeFromPort( FR_DFLT_NIQS_SERVER_PORT ) ); 

	return 1;
}

void ClientTypesUnitTest::TestWNymIconInfo()
{
	WNymIconInfo info ;
	
	info.SetIconIndex( 4 ) ;

	ZKTEST( info.GetIconIndex() == 4 ) ;
	ZKTEST( info.GetIconType() == NymIconIndex ) ;
	
	ZKSTD::string strTestFile( "file.x" ) ;

	info.SetIconFilePath( strTestFile ) ;
	ZKTEST( info.GetIconType() == NymIconFilePath ) ;
	ZKTEST( strTestFile == info.GetFilePath() ) ;

	// Current behavior is that we get a default 
	// index value if we have established a file path
	ZKTEST( info.GetIconIndex() == DEFAULT_NYM_ICON_INDEX ) ;

	// Copy constructor test
	WNymIconInfo info2( info ) ;
	ZKTEST( info2.GetIconType() == NymIconFilePath ) ;
	ZKTEST( info2.GetIconIndex() == DEFAULT_NYM_ICON_INDEX ) ;
	ZKTEST( strTestFile == info2.GetFilePath() ) ;

	info2.SetIconIndex( 22 ) ;

	// Operator= test
	info = info2 ;

	ZKTEST( info.GetIconIndex() == 22 ) ;
}

void ClientTypesUnitTest::TestWNetInfo()
{
	WNetInfo test ;

	// test has an empty state.  Make sure that functions behave nicely

	ZKTEST( test.IsValid() ) ; // Always true

	WNetInfoDatum ent ;
	ZKTEST( false == test.GetEntityById( 500, ent ) ) ;
	ZKTEST( 0 == test.GetEntId( 1, 51107 ) ) ;
	ZKTEST( 0 == test.GetEntityTypeCount( FRENT_TYPE_TOK_SRV ) ) ;

	test.Flush() ;
}

#ifdef WIN32
ClientTypesUnitTest g_clientTypesUnitTest;
#endif

#endif
