// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// WAutoBuffer.h
//
// This is an autopointer-like class that prepresents an allocated
// buffer.  It primary purpose is to provide free memory deallocation,
// which prevents leaks and makes exception handling code simpler (eg fewer
// catch statements just to free memory.

// These is little or no overhead to using this class instead of
// raw pointers, so use this class!

// The class can be provided as an argument to the core ZKS libraries
// which will allocate and fill in the contents of the buffer.
// We except the allocation function to be the global ZKS Malloc function.
// This is not meant for buffers allocated by new()!  There is little or
// no advantage to allocating a buffer with new() anyway so this should
// never be an issue.

// There are two classes, one for unsigned char, one for char.  I intentionally
// didn't use a template because there would have only been two useages,
// and these are pretty simple classes.  If there is a demand for auto-arrays
// of multiple types then we can get funky.

// Example Usage:  See clienttypes unit test.
// 

#ifndef _WAutoBuffer_H
#define _WAutoBuffer_H

#include <zkAlloc.h>
#include <clientdebug.h>	// for ZKASSERT

#include <AutoInstCount.h>


class WAutoBuffer AUTO_INSTANCE_COUNTING1( WAutoBuffer )
{
public:
	WAutoBuffer() : m_pBuffer( NULL ), m_cntBuffer( 0 ) { ; }

	// Automatically free the buffer if any contents
	~WAutoBuffer() { Clear() ; }

	void Allocate( size_t in_cntBuffer ) ;
	void CopyBuffer( const unsigned char * in_pBufferToCopy, size_t in_cntBuffer ) ;

	void Clear() { if ( m_pBuffer != NULL ) { zkMemFree(NULL, (void**)&m_pBuffer); m_pBuffer = NULL ; m_cntBuffer = 0 ;} }

	size_t Size() const { return m_cntBuffer ; }
	bool IsValid() const { return ( NULL != m_pBuffer ) ; } 

	// Use this operators to gain access to the internal buffer
	operator unsigned char * () { return m_pBuffer ; }
	operator unsigned char * () const { return m_pBuffer ; }

	unsigned char * GetBuffer() { return m_pBuffer ; }

	// Safe copy constructor
	WAutoBuffer& operator= ( const WAutoBuffer& in_rhs ) ;
	WAutoBuffer( const WAutoBuffer& in_rhs ) ;

/*
	// Override this to get some free index checking.
	// If we consider this a problem that it can be ifdef'd to only
	// happen in debug mode.
	unsigned char operator [] ( int in_index ) { ZKASSERT( m_pBuffer ) ;
												   if ( (size_t) in_index >= m_cntBuffer ) {
													   ZKASSERT( false ) ;
													   return (unsigned char) 0 ;
												   }
												   return m_pBuffer[ in_index ] ; }

*/

	// Use these functions to provide the contents to a function
	// to do that allocation and creation of the buffer
	// If the asserts goes off then we are reallocing the buffer which
	// is a leak and presumably not what was intended
	// C++ only allows one operator& to be defined, so you need to
	// type cast if the caller wants char** or void**
	unsigned char** operator&() { ZKASSERT( m_pBuffer == NULL ) ; return &m_pBuffer ; }

	size_t * GetSizePointer() { return &m_cntBuffer ; }
	size_t & GetSizeRef()     { return m_cntBuffer ; }

private:
	unsigned char * m_pBuffer ;
	size_t          m_cntBuffer ;	

} ;

class WAutoCharBuffer AUTO_INSTANCE_COUNTING1( WAutoCharBuffer )
{
public:
	WAutoCharBuffer() : m_pBuffer( NULL ), m_cntBuffer( 0 ) { ; }

	// Automatically free the buffer if any contents
	~WAutoCharBuffer() { Clear() ; }

	void Allocate( size_t in_cntBuffer ) ;
	void CopyBuffer( const char * in_pBufferToCopy, size_t in_cntBuffer ) ;

	void Clear() { if ( m_pBuffer != NULL ) { zkMemFree(NULL, (void**)&m_pBuffer); m_pBuffer = NULL ; } }

	size_t Size() const { return m_cntBuffer ; }
	bool IsValid() const { return ( NULL != m_pBuffer ) ; } 

	// Use this operators to gain access to the internal buffer
	operator char * () { return m_pBuffer ; }
	operator char * () const { return m_pBuffer ; }

	// Safe copy constructor
	WAutoCharBuffer& operator= ( const WAutoCharBuffer& in_rhs ) ;
	WAutoCharBuffer( const WAutoCharBuffer& in_rhs ) ;


	// Use these functions to provide the contents to a function
	// to do that allocation and creation of the buffer
	// If the asserts goes off then we are reallocing the buffer which
	// is a leak and presumably not what was intended
	// C++ only allows one operator& to be defined, so you need to
	// type cast if the caller wants char** or void**
	char** operator&() { ZKASSERT( m_pBuffer == NULL ) ; return &m_pBuffer ; }

	// This gives someone the change to change the size of the buffer
	// Be Careful! This normally happens when you pass in to a funciton
	// that will perform the allocation, or perhaps decrease the size of the buffer (see zkConvertFromBase64)
	size_t * GetSizePointer() { return &m_cntBuffer ; }
	size_t & GetSizeRef()     { return m_cntBuffer ; }

private:
	char *  m_pBuffer ;
	size_t  m_cntBuffer ;


} ;

inline void WAutoBuffer::Allocate( size_t in_cntBuffer ) 
{ 
	Clear() ;

	ZKASSERT( m_pBuffer == NULL ) ;
	ZKASSERT( m_cntBuffer == 0 ) ;

	if ( in_cntBuffer > 0 ) {
		ZKVERIFY( zkMemAlloc( NULL, (void**)&m_pBuffer, in_cntBuffer )) ;
		m_cntBuffer = in_cntBuffer ;
		ZKASSERT( m_pBuffer != NULL ) ; 
	}
}

inline void WAutoBuffer::CopyBuffer( const unsigned char * in_pBufferToCopy, size_t in_cntBuffer )
{
	ZKASSERT( m_pBuffer == NULL ) ;
	ZKASSERT( m_cntBuffer == 0 ) ;

	Clear() ;

	if ( in_cntBuffer > 0 ) {

		ZKASSERT( in_pBufferToCopy != NULL ) ;
		ZKVERIFY( zkMemAlloc( NULL, (void**)&m_pBuffer, in_cntBuffer )) ;
		m_cntBuffer = in_cntBuffer ;
		ZKASSERT( m_pBuffer != NULL ) ; 
		memcpy( m_pBuffer, in_pBufferToCopy, in_cntBuffer ) ;
	}
}

inline WAutoBuffer::WAutoBuffer( const WAutoBuffer& in_rhs ) 
        AUTO_INSTANCE_COPY_CTOR1( WAutoBuffer, in_rhs )
{
	m_pBuffer = NULL;
	m_cntBuffer = 0;
	CopyBuffer( in_rhs, in_rhs.Size() ) ;
}

inline WAutoBuffer& WAutoBuffer::operator= ( const WAutoBuffer& in_rhs )
{
    // This is comparing the address of the buffer in in_rhs returned by operator & with
    // the buffer inside the this pointer. They should be unique values, so the same
	// affect as comparing the address of the object themselves.
	if ( (unsigned char*)in_rhs != (unsigned char*)*this )
	{	
		Clear() ;
		CopyBuffer( in_rhs, in_rhs.Size() ) ;
	}	
	return *this ; 
}

inline void WAutoCharBuffer::Allocate( size_t in_cntBuffer ) 
{ 
	Clear() ;

	ZKASSERT( m_pBuffer == NULL ) ;
	ZKASSERT( m_cntBuffer == 0 ) ;

	if ( in_cntBuffer > 0 ) {
		ZKVERIFY( zkMemAlloc( NULL, (void**)&m_pBuffer, in_cntBuffer )) ;
		m_cntBuffer = in_cntBuffer ;
		ZKASSERT( m_pBuffer != NULL ) ; 
	}
}

inline void WAutoCharBuffer::CopyBuffer( const char * in_pBufferToCopy, size_t in_cntBuffer )
{
	ZKASSERT( m_pBuffer == NULL ) ;
	ZKASSERT( m_cntBuffer == 0 ) ;

	Clear() ;

	if ( in_cntBuffer > 0 ) {

		ZKASSERT( in_pBufferToCopy != NULL ) ;
		ZKVERIFY( zkMemAlloc( NULL, (void**)&m_pBuffer, in_cntBuffer )) ;
		m_cntBuffer = in_cntBuffer ;
		ZKASSERT( m_pBuffer != NULL ) ; 
		memcpy( m_pBuffer, in_pBufferToCopy, in_cntBuffer ) ;
	}
}

inline WAutoCharBuffer::WAutoCharBuffer( const WAutoCharBuffer& in_rhs ) 
        AUTO_INSTANCE_COPY_CTOR1( WAutoCharBuffer, in_rhs )
{
	m_pBuffer = NULL;
	m_cntBuffer = 0;
	CopyBuffer( in_rhs, in_rhs.Size() ) ;	
}

inline WAutoCharBuffer& WAutoCharBuffer::operator= ( const WAutoCharBuffer& in_rhs )
{
    // See comment above
	if ( (char*)in_rhs != (char*)*this )
	{	
		CopyBuffer( in_rhs, in_rhs.Size() ) ;	
	}	
	return *this ; 
}


#endif
