// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// WCoreTypes.h
//
// This header file defines types to use instead of the
// underlying core LibZK or fr base types.  Always use these
// types instead of the raw C libraries. 
//
// If the typedef doesn't already exist then you can add it.
// However, specific, more powerful, wrappers to handle some of most 
// common and useful types exist outside this file in their own header
// files.
// 
// These types take care of automatically freeing, and in some
// cases auto-allocation of the underlying structures.  In effect
// they are autopointers.
//
// When adding any new types please respect the consistent naming scheme.

#ifndef WCORETYPES_H_
#define WCORETYPES_H_


// This are the include files necessary to 
// get at the all the basic types that are available
#include <clientdebug.h>
#include <zkList.h>
#include <zkMemBuf.h>
#include <zkMIME.h>
#include <zkEmail.h>
#include <frSec.h>
#include <frPktFmt.h>

#include <AutoInstCount.h>

// Forward Declarations
struct T_ZkError ;


// This template provides a "smart" pointer to 
// an "object" from the freedom fr and lib libraries
//
// Don't use this template directly, please use the typedefs
// below

template < typename T, 
		   int (__cdecl *pFreeFunc) ( struct T_ZkError * , T* ) >
class CoreLibTypeWrapper2 
{
protected:
	T m_pItem ;			//	Holds the real pointer to the object

private:	
	CoreLibTypeWrapper2( const T& )  ;	// Avoid assignment of these objects
	T& operator =( const T& )  ;			// because underlying core structure would also have
											// to duplicate itself

	void Clear() 
	{
		ZKASSERT( pFreeFunc != NULL ) ;

		// Free any associated resources
		if (( pFreeFunc != NULL ) && 
			( m_pItem != NULL ) )
		{
			// Call free function
			(pFreeFunc) ( NULL, &m_pItem ) ;

			// Free Function is expected to zero our pointer
			ZKASSERT( m_pItem == NULL ) ;
		}
	}

public:
	
	CoreLibTypeWrapper2() : m_pItem( NULL )
	{
	}
	
	~CoreLibTypeWrapper2() 
	{
		Clear() ;
	}

	operator T()		{ return m_pItem ; } ;
	operator T() const  { return m_pItem ; } ;

	T* operator&() 
	{	// We shouldn't have already allocated
		// if we are giving out the address 
		// Use a variant of the template that does not 
		// specify the NewFunc
		ZKASSERT( m_pItem == NULL ) ; 
		return &m_pItem ; 
	}

	// Can use to determine if the object has actually
	// been allocated yet
	bool IsValid() const { return ( m_pItem != NULL ) ; }

} ;


template < typename T, 
		   int (__cdecl *pFreeFunc) ( struct T_ZkError * , T* ), 
		   int (__cdecl *pNewFunc)  ( struct T_ZkError * , T* ) >
class CoreLibTypeWrapper1 : public CoreLibTypeWrapper2<T, pFreeFunc>
{
	
	CoreLibTypeWrapper1( const T& )  ;	// Avoid assignment of these objects
	T& operator =( const T& )  ;			// because underlying core structure would also have
											// to duplicate itself
public:
	
	CoreLibTypeWrapper1() : CoreLibTypeWrapper2<T, pFreeFunc>()        
	{
		if ( pNewFunc != NULL ) {
			(pNewFunc) ( NULL, &m_pItem ) ;
			ZKASSERT( m_pItem != NULL ) ;
		}
	}
	
	~CoreLibTypeWrapper1() 
	{
	}

} ;

// Use these types to access the autopointers

// Example usage:
// Case 1) The functions you will call ask that you pass in a
// initialized pointer to the core library type.
//	For example:
//  WZKList myList ;
//  zkEmailUtilAddToRecipientList(... myList....) ;
//  

// Case 2) The functions you will call ask that you pass address of the
// pointer to the core library type so that it can be created for you
//  For example:
//  WZKListPtr myListPtr ;
//	frRemoteNymGetStringListParameter(
//					&m_ZkError ,              
//					m_RemoteNym , 
//					NymScanWord,
//					myListPtr );
//   

// Case 3) The functions you will call ask that you pass address of the pointer
// to the core library type so that it can be set to point to an existing object.
//   In this case DO NOT use these types, because you don't own the instance that
//   you point to.


// CASE 1 types
// Use these types when the underlying object should be created before making
// calls
typedef CoreLibTypeWrapper1< ZkList,			zkListFree,			zkListNew >			WZkList ;
typedef CoreLibTypeWrapper1< ZkEmail,		zkEmailFree,		zkEmailNew >		WZkEmail ;
typedef CoreLibTypeWrapper1< FrSecPassInfo,	frSecPassInfoFree,	frSecPassInfoNew >	WSecPassInfo ;
typedef CoreLibTypeWrapper1< FrCryptSym,		frCryptSymFree,		frCryptSymNew >		WCryptSym ;
typedef CoreLibTypeWrapper1< ZkMemBuf,		zkMemBufFree,		zkMemBufNew	>		WMemBuf;

typedef CoreLibTypeWrapper1< FrPktAD,		frPktAipDataFree,	frPktAipDataNew >		WAipData;

// CASE 2 types
// Use these types to create objects that will be allocated inside
// a function that we call (so it is not allocated by the constructor)


typedef CoreLibTypeWrapper2< ZkList, zkListFree > WZkListNoAlloc ;

#endif
