// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// WCypherMailMsg.cpp

// This class represents a encrypted mail message intended 
// for a nym.  The clear text and other details of the message
// can be extracted from the object.


#ifndef WCYPHERMAILMSG_H_
#define WCYPHERMAILMSG_H_

#include "ClientTypesInc.h"
#include "WAutoBuffer.h"
#include <client.h>	// for ZKSTD
#include <string>


class ClientTypes_API WCypherMailMsg
{
public:
	WCypherMailMsg() ;
	~WCypherMailMsg() ;

	// Clear cleans out any existing settings.  Use this
	// if you wish to reuse a instance of the class for multiple decryptions
	void                  Clear() ;

	bool                  IsMsgDecrypted() { return GetClearTextSize() > 0 ; }

	// Use this to set up the message
	void				  SetEncryptedText( 
							const unsigned char * in_pBuf, 
							size_t in_szBuf ) ;
	
	// Set information - if we already know the nym and the reply block then
	// we can decrypt quickly

	void                  SetRecipientNym( const ZKSTD::string& in_strRecipientNym ) { m_strRecipientNym = in_strRecipientNym ; } 
	void                  SetReplyBlock( size_t in_nReplyBlock ) { m_nReplyBlock = in_nReplyBlock ; } 
	void                  SetSendingEntity( const ZKSTD::string& in_strSendingEntity ) { m_strSendingEntity = in_strSendingEntity ; }
	void                  SetFmgSignatureCheckResult( const SigCheckResult in_FmgSignatureResult ) { m_FmgSignatureResult = in_FmgSignatureResult ; }
	void                  SetSignatureCheckResult( const SigCheckResult in_SignatureResult ) { m_SignatureResult = in_SignatureResult ; }

	// Remember if we activated a reply block
	void                  SetMsgIsRBActivationToken() { ZKASSERT( m_nReplyBlock != UnknownValue ) ; m_bActivatedRB = true ; }

	// Index value for reply block when we don't know yet
	static const size_t UnknownValue ; 

	// Retrieve information

	const unsigned char * GetEncryptedText() const { return m_pEncryptedText ; }
	size_t                GetEncryptedSize() const { return m_cntEncryptedText ; }

	const unsigned char * GetClearText( )    const { return m_ClearText ; }
	size_t                GetClearTextSize() const { return m_ClearText.Size() ; }

	WAutoBuffer&          GetClearAutoBuffer() { return m_ClearText; } 

	void                  GetSendingEntity( ZKSTD::string& out_SendingEntity ) const { out_SendingEntity = m_strSendingEntity ; }
	SigCheckResult        GetFmgSigCheckResult() const { return m_FmgSignatureResult ; }
	SigCheckResult        GetSigCheckResult() const { return m_SignatureResult ; }

	// Returns WCypherMailMsg::UnknownValue if this has not been determined
	void                  GetRecipientNym( ZKSTD::string& out_RecipientNym ) const { out_RecipientNym = m_strRecipientNym ; }
	size_t                GetReplyBlock() const { ZKASSERT( ( !m_strRecipientNym.empty() ) ||
												 	        ( m_nReplyBlock == UnknownValue ) ) ;
												  return m_nReplyBlock ; }

	bool				  IsActivationRBMsg() { return m_bActivatedRB ; }

	void				  SelfTest() const ;
	void				  DumpToDebugOutput() const;

private:
	WCypherMailMsg( const WCypherMailMsg& ) ;
	WCypherMailMsg& operator= ( const WCypherMailMsg& ) ;

	WAutoBuffer		m_ClearText ;		 // This buffer contains the email message contents after decryption

	ZKSTD::string   m_strSendingEntity ; // Nym or Freedom Server that sent the mail
	ZKSTD::string   m_strRecipientNym ;	 // The nym which recieved the message
	size_t          m_nReplyBlock ;      // Index of the reply block which can be used to decrypt the message

	SigCheckResult  m_FmgSignatureResult ;  // Result of the FMG signature check
	SigCheckResult  m_SignatureResult ;     // Result of the signature check

	// This object doesn't own the encrypted text buffer
	const unsigned char * m_pEncryptedText ;
	size_t                m_cntEncryptedText ;

	// If the message was a reply block activation token, and the nym
	// server was successfully communicated with the active the RB, then this will be set
	bool				  m_bActivatedRB ;
} ;

#endif
