// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):


#include "StdAfx.h"
#include "WMaipRb.h"
#include "WAutoBuffer.h"

#include <RZkException.h>

static const int REPLY_BLOCK_VERSION = 1;

WMaipRb::WMaipRb()
	: m_FrMaipRbPriv(NULL)
{
	Allocate() ;
}

// Initialize based on an existing RB
WMaipRb::WMaipRb( const WMaipRb& in_rhs )
	: m_FrMaipRbPriv(NULL)	
	, m_RBBinary( in_rhs.m_RBBinary )
{
	Allocate() ;

	// Establish our state from the underlying buffer which we copied
	ZKVERIFY( frMaipRbPrivReadFromBuf(
				&m_Error, 
				m_FrMaipRbPriv,	 
				(void*)	m_RBBinary.GetPrivateBuffer(), 
				m_RBBinary.GetPrivateBufferSize() )) ;
}

WMaipRb& WMaipRb::operator= ( const WMaipRb & in_rhs )
{
	if ( &in_rhs != this )
	{	
		Clear() ;
		Allocate() ;
		m_RBBinary = in_rhs.m_RBBinary ;

		// Re-establish our state from the underlying buffer
		ZKVERIFY( frMaipRbPrivReadFromBuf(
					&m_Error, 
					m_FrMaipRbPriv,	 
					(void*)	m_RBBinary.GetPrivateBuffer(), 
					m_RBBinary.GetPrivateBufferSize() )) ;
	}

	return *this ;
}

WMaipRb::~WMaipRb()
{
	Clear() ;
	ZKASSERT( m_FrMaipRbPriv == NULL ) ;
}

void 
WMaipRb::Initialize( 
		const unsigned char * in_pBuffer, 
		const size_t in_cntBuffer ) throw(RZkException)
{
	ZKASSERT( m_FrMaipRbPriv != NULL ) ;
	ZKASSERT( in_pBuffer != NULL ) ;
	ZKASSERT( in_cntBuffer > 6 ) ;

	m_RBBinary.Initialize( in_pBuffer, in_cntBuffer ) ;
	
	int bStatus = frMaipRbPrivReadFromBuf(
				&m_Error, 
				m_FrMaipRbPriv,	 
				(void*)	m_RBBinary.GetPrivateBuffer(), 
				m_RBBinary.GetPrivateBufferSize() );

	if ( !bStatus ) {
		throw RZkException(m_Error,  "WMaipRb::Initialize", "Could not convert private buffer to a reply block", IDS_CORRUPT_BUFFER ) ;
	}
}

void
WMaipRb::Initialize( 
			const ZKSTD::string&			in_strRbName,
			const ZKSTD::string&			in_strUsersEmailAddr,
			const ZKSTD::string&			in_strNymName,
			const FrCryptCrypt			in_FMGPublicCryptKey,
			const ReplyBlockHopInfoVec& in_MaipHops,
			const unsigned int			in_nLatency /*= DEFAULT_LATENCY */,
			const FrCryptSymAlg			in_Alg /*= DEFAULT_CRYPTO_ALG*/ ) throw(RZkException)
{
	ZKASSERT( m_FrMaipRbPriv != NULL ) ;
	ZKASSERT( in_FMGPublicCryptKey != NULL ) ;
	ZKASSERT( in_MaipHops.size() >= 2 ) ;
	ZKASSERT( in_strRbName.size() > 0 ) ;
	ZKASSERT( in_strNymName.size() > 0 ) ;
	ZKASSERT( in_strUsersEmailAddr.size() > 0 ) ;

	// Build the reply block

    FrEntIdExp nymEntityExp;
	ZKVERIFY(frEntSetFromStr2(
				&m_Error, 
				&nymEntityExp, 
				"nym", 
				in_strNymName.c_str()) );

	ZKVERIFY( frMaipRbPrivSetEmail( &m_Error, m_FrMaipRbPriv, const_cast<char *>(in_strUsersEmailAddr.c_str()) ) ) ;
	ZKVERIFY( frMaipRbPrivSetName( &m_Error, m_FrMaipRbPriv, const_cast<char *>(in_strRbName.c_str()) ) ) ;
	ZKVERIFY( frMaipRbPrivSetNym(&m_Error, m_FrMaipRbPriv, const_cast<FrEntIdExp*>(&nymEntityExp)));

	// Now add the hops to the reply block

	bool bIsFirstHop = true ;

	for ( size_t iHop = 0 ; iHop < in_MaipHops.size() ; iHop++ ) {

		bIsFirstHop = ( iHop == 0 ) ;

		// Add a hop
		struct frmaip_rbpriv_set_hop hop = { 0 };

		// We use the previous hops key
		if ( iHop == 0 ) {
			hop.encrPubKey = in_FMGPublicCryptKey;
		}
		else {
			hop.encrPubKey = in_MaipHops[ iHop - 1 ].GetSigningKey() ;
		}

		hop.encrSymAlg  = in_Alg;  // Wrap RB when submitting to NymServer, fixed
		hop.replySymAlg = in_Alg;
		hop.port = 1;
		hop.latency = in_nLatency;      
	
		ZKSTD::string strMaipName ;
		in_MaipHops[ iHop ].GetEntityName( strMaipName ) ;

        // this is not the last hop so set the Maip information
		ZKVERIFY(frEntSetFromStr2(
			&m_Error, 
			&hop.maip, 
			"maip", 
			strMaipName.c_str()) ); 
				
		ZKVERIFY(frMaipRbPrivSetNextHop(
			&m_Error, 
			m_FrMaipRbPriv,
			bIsFirstHop,
			ZK_FALSE,			// not last hop
			&hop));
	}

	// Add one last hop with no hop information
	// This represents the email itself... talk to kent if you need to understand....
		
	struct frmaip_rbpriv_set_hop LastHop = { 0 };
	
	LastHop.encrPubKey  = in_MaipHops[ in_MaipHops.size() - 1 ].GetSigningKey();
	LastHop.encrSymAlg  = in_Alg;  
	LastHop.replySymAlg = in_Alg;
	LastHop.port		= 1;
	LastHop.latency		= in_nLatency;

	ZKVERIFY(frMaipRbPrivSetNextHop(
		&m_Error, 
		m_FrMaipRbPriv,
		ZK_FALSE,			//not first hop
		ZK_TRUE,			//yes last hop
		&LastHop));

	// This is a bit of tricky code,
	// so lets run through the result and
	// test some invariants

#ifdef _DEBUG

	int bTestIsLastHop = ZK_FALSE;
	frmaip_rbpriv_set_hop h;

	int iHopCount = 0 ;

	do {
		ZKVERIFY( frMaipRbPrivGetHop( &m_Error, m_FrMaipRbPriv, iHopCount, &bTestIsLastHop, &h) );

		ZKASSERT( h.encrSymAlg == in_Alg ) ;
		ZKASSERT( h.replySymAlg == in_Alg ) ;

		if ( !bTestIsLastHop ) {
			ZKASSERT( strlen( h.maip.name ) > 0 ) ;

			ZKSTD::string strMaipName ;
			in_MaipHops[iHopCount].GetEntityName( strMaipName ) ;
			ZKASSERT( 0 == strMaipName.compare( h.maip.name ) ) ;
		}

		++iHopCount ;
	}
	while ( !bTestIsLastHop ) ;

	ZKASSERT( (size_t) iHopCount == ( in_MaipHops.size() + 1 ) ) ;
#endif

	// Now set up the binary representation
	m_RBBinary.Initialize( m_FrMaipRbPriv ) ;
}


void 
WMaipRb::Allocate()
{
	ZKASSERT( m_FrMaipRbPriv == NULL ) ;
	ZKVERIFY( frMaipRbPrivNew(&m_Error, &m_FrMaipRbPriv) ) ;
}

void 
WMaipRb::Clear()
{
	frMaipRbPrivFree(&m_Error, &m_FrMaipRbPriv);
	m_FrMaipRbPriv = NULL;
	m_RBBinary.Clear() ;
}


unsigned int WMaipRb::GetMessageLatency() const
{
	frmaip_rbpriv_set_hop h;
	GetFirstHopInfo( h ) ;

	// TODO - in _DEBUG this is a good chance to verify that all the Hops have
	// the same latency

	return h.latency ;
}

void WMaipRb::GetFirstHopInfo( frmaip_rbpriv_set_hop& out_h ) const
{
	ZKASSERT ( m_FrMaipRbPriv != NULL ) ;

	memset( &out_h, 0, sizeof( out_h ) ) ;

	int bTestIsLastHop = ZK_FALSE;

	if(!(frMaipRbPrivGetHop(
			&m_Error, 
			m_FrMaipRbPriv, 
			0,			// Index of first hop
			&bTestIsLastHop, 
			&out_h)))
	{
		ZKASSERT(false);
	}
}

void
WMaipRb::GetName( ZKSTD::string & out_RBName ) const
{
	ZKASSERT ( m_FrMaipRbPriv != NULL ) ;
	
	out_RBName.erase() ;
	
	const char* buf = NULL;
		
	if(!(frMaipRbPrivGetName(
		&m_Error, 
		m_FrMaipRbPriv,
		&buf)))
	{
		ZKASSERT(false);
	}
	else {
		out_RBName = buf;
	}
}

void
WMaipRb::GetUserEmailAddress( ZKSTD::string & out_strAddress ) const
{
	ZKASSERT ( m_FrMaipRbPriv != NULL ) ;
	
	out_strAddress.erase() ;
	
	const char* buf = NULL;
		
	if(!(frMaipRbPrivGetEmail(
		&m_Error, 
		m_FrMaipRbPriv,
		&buf)))
	{
		ZKASSERT(false);
	}
	else {
		out_strAddress = buf;
	}
}

void
WMaipRb::GetNymName( ZKSTD::string & out_strNymName ) const
{
	ZKASSERT ( m_FrMaipRbPriv != NULL ) ;
	
	out_strNymName.erase() ;
	
	const FrEntIdExp * pNymIdExp = NULL ;
		
	if(!(frMaipRbPrivGetNym(
		&m_Error, 
		m_FrMaipRbPriv,
		&pNymIdExp)))
	{
		ZKASSERT(false);
	}
	else {
		out_strNymName = pNymIdExp->name ;
	}
}


FrCryptSymAlg
WMaipRb::GetReplyBlockWrapperAlgorithm() const
{
	frmaip_rbpriv_set_hop h;
	GetFirstHopInfo( h ) ;
	return h.encrSymAlg ;
}

FrCryptSymAlg
WMaipRb::GetReplyBlockCryptoAlgorithm() const
{
	frmaip_rbpriv_set_hop h;
	GetFirstHopInfo( h ) ;
	return h.replySymAlg ;
}

void
WMaipRb::GetMaipNamesOfHops( ZKSTD::vector<ZKSTD::string>& out_vecstrReplyBlockDomainNames ) 
{
	out_vecstrReplyBlockDomainNames.clear() ;

	int bTestIsLastHop = 0, i = 0;
	frmaip_rbpriv_set_hop h;

	do{
		memset(&h, 0, sizeof(frmaip_rbpriv_set_hop));

		ZKVERIFY(frMaipRbPrivGetHop(
			&m_Error, 
			GetPrivateReplyBlock(), 
			i,			
			&bTestIsLastHop, 
			&h));

		if ( strlen( h.maip.name ) > 0 ) {
			out_vecstrReplyBlockDomainNames.push_back(h.maip.name);
		}
		else {
			// Last hop has no maip name
			ZKASSERT( bTestIsLastHop == ZK_TRUE ) ;
		}

		++i;

	}while(bTestIsLastHop == 0);
}

void 
WMaipRb::GetActivationFlag( bool& out_bActivationFlag )
{
	unsigned int bFlag = ZK_TRUE ;
	ZKVERIFY(frMaipRbPubIsActive( &m_Error, (void*) GetPublicBuffer(), GetPublicBufferSize(), &bFlag ) ) ;
	out_bActivationFlag = ( bFlag != 0 ) ;
}	

void 
WMaipRb::SetActivationFlag( bool  in_bActivationFlag )
{		
	m_RBBinary.SetActivationFlag( in_bActivationFlag ) ;
}

void
WMaipRb::DumpToDebugOutput() const
{
#ifdef INTERNAL_BUILD
	
	ZKTRACE0( "----------REPLY BLOCK DUMP-------\n" ) ;

	if ( !IsValid() ) {
		// Nym is invalid
		ZKASSERT( ZK_FALSE ) ;
		return ;	
	}

	ZKSTD::string strName ;
	GetName( strName ) ;

	ZKSTD::string strEmail ;
	GetUserEmailAddress( strEmail ) ;

	ZKSTD::string strNymName ;
	GetNymName( strNymName ) ;

	ZKTRACE1( "Reply Block Name - %s\n", strName.c_str() ) ;
	ZKTRACE1( "Email Address    - %s\n", strEmail.c_str() ) ;
	ZKTRACE1( "Nym Name         - %s\n", strNymName.c_str() ) ;
	ZKTRACE1( "Crypto alg       - %d\n", GetReplyBlockWrapperAlgorithm() ) ;
	ZKTRACE1( "Wrapper alg      - %d\n", GetReplyBlockCryptoAlgorithm() ) ;
	ZKTRACE1( "Latency          - %d\n", GetMessageLatency() ) ;

	int bTestIsLastHop = ZK_FALSE;
	frmaip_rbpriv_set_hop h;

	int iHopCount = 0 ;

	do {
		ZKVERIFY( frMaipRbPrivGetHop( &m_Error, m_FrMaipRbPriv, iHopCount, &bTestIsLastHop, &h) );

		ZKTRACE1( "Hop %d - ", iHopCount ) ;

		if ( h.maip.name != NULL ) {
			ZKTRACE1( "\tMaip: %s\t", h.maip.name ) ;
		}

		ZKTRACE0( "\n" ) ;

		// TODO - show more internals

		++iHopCount ;
	}
	while ( !bTestIsLastHop ) ;

	ZKTRACE0( "\n\n\n" ) ;
#endif
}

void 
WMaipRb::SelfTest() const
{
#ifdef INTERNAL_BUILD

	// Verify that we have been either fully initialized or not at all
	if ( m_FrMaipRbPriv == NULL ) {
		ZKASSERT( m_RBBinary.GetRBAsBinarySize() == 0 ) ;
		ZKASSERT( m_RBBinary.GetRBAsBinary() == NULL ) ;
		return ;
	}
	else {
		ZKASSERT( m_RBBinary.GetRBAsBinarySize() > 0 ) ;
		ZKASSERT( m_RBBinary.GetRBAsBinary() != NULL ) ;

		ZKSTD::string strName ;
		GetName( strName ) ;
		ZKASSERT( !strName.empty() ) ;

		ZKSTD::string strEmail ;
		GetUserEmailAddress( strEmail ) ;
		ZKASSERT( !strEmail.empty() ) ;

		ZKSTD::string strNymName ;
		GetNymName( strNymName ) ;
		ZKASSERT( !strNymName.empty() ) ;
	}

#endif
}



// **********************************************************************
//
// WMaipRbBinary
//
// **********************************************************************

WMaipRbBinary::WMaipRbBinary( const WMaipRbBinary & in_rhs ) 
	: m_Buffer( in_rhs.m_Buffer ) 
{ 
	ParseBuffer() ;
}

WMaipRbBinary& 
WMaipRbBinary::operator =( const WMaipRbBinary & in_rhs )
{
	if ( &in_rhs != this )
	{	
		Clear() ;
		m_Buffer = in_rhs.m_Buffer ;
		ParseBuffer() ;
	}

	return *this ;
}

void 
WMaipRbBinary::ParseBuffer()
{
	if ( m_Buffer.Size() == 0 ) {
		m_szPrivateBuf = 0 ;
		m_szPublicBuf = 0 ;

		m_PrivateRBBuffer = NULL ;
		m_PublicRBBuffer = NULL ;

		return ;
	}
	
	// Now deconstruct the buffer into into components and
	// remember where they are.
	m_szPrivateBuf = m_Buffer[0] * 0x100 + m_Buffer[1] ;

	ZKASSERT( m_Buffer[2] == REPLY_BLOCK_VERSION ) ;

	// Private buffer starts at the third byte 
	m_PrivateRBBuffer = &(m_Buffer[3]);

	unsigned int public_buffer_location = m_szPrivateBuf + 4;
	m_szPublicBuf = m_Buffer[public_buffer_location] * 0x100 + m_Buffer[public_buffer_location + 1];
	
	ZKASSERT( m_Buffer[public_buffer_location + 2] == REPLY_BLOCK_VERSION ) ;

	m_PublicRBBuffer = &(m_Buffer[ public_buffer_location + 3 ]);

	// Post conditions
	ZKASSERT( m_szPublicBuf > 0 ) ;
	ZKASSERT( m_szPrivateBuf > 0 ) ;
	ZKASSERT( m_PrivateRBBuffer != NULL ) ;
	ZKASSERT( m_PublicRBBuffer != NULL ) ;
}


void 
WMaipRbBinary::Initialize( 
		const unsigned char * in_pBuffer, 
		const size_t in_cntBuffer )
{
	ZKASSERT( in_pBuffer != NULL ) ;
	ZKASSERT( in_cntBuffer > 6 ) ;

	m_Buffer.CopyBuffer( in_pBuffer, in_cntBuffer ) ;

	ParseBuffer() ;
}


void 
WMaipRbBinary::Initialize( const FrMaipRbPriv in_pRB )
{
	ZKASSERT( in_pRB != NULL ) ;

	WAutoBuffer privbuf ;
	WAutoBuffer pubbuf ;

	unsigned char* rb_buffer  = NULL;

	void* ptr_rb_buffer = NULL;
	void* ptr_pubbuf = NULL;

	// Make public version

	ZKVERIFY( frMaipRbPrivMakePublic (
				&m_Error, 
				in_pRB, 
				(void**)&pubbuf, 
				pubbuf.GetSizePointer()));

	ZKVERIFY( frMaipRbPubDeactivate( &m_Error, (void*)pubbuf, pubbuf.Size() ) ) ;

	ZKVERIFY( frMaipRbPrivWriteToBuf(
					&m_Error, 
					in_pRB, 
					(void**)&privbuf, 
					privbuf.GetSizePointer()));

	unsigned int sz_rb_buffer = privbuf.Size() + pubbuf.Size() + 7;

	m_Buffer.Allocate( sz_rb_buffer ) ;
	memset(m_Buffer, 0x00, sz_rb_buffer);
	
	// private buf
	m_Buffer[0] = privbuf.Size()/0x100; // size high
	m_Buffer[1] = privbuf.Size()%0x100; // size low
	m_Buffer[2] = REPLY_BLOCK_VERSION; // Version
	memcpy(m_Buffer + 3, privbuf, privbuf.Size());
	
	// public buf
	m_Buffer[privbuf.Size()+4] = pubbuf.Size()/0x100; // size high
	m_Buffer[privbuf.Size()+5] = pubbuf.Size()%0x100; // size low
	m_Buffer[privbuf.Size()+6] = REPLY_BLOCK_VERSION; // Version
	memcpy(m_Buffer + privbuf.Size() + 7, pubbuf, pubbuf.Size());

	ParseBuffer() ;
}

void 
WMaipRbBinary::SetActivationFlag( bool in_bActivationFlag )
{
	ZKASSERT( m_szPrivateBuf > 0 ) ; 
	ZKASSERT( m_szPublicBuf > 0 ) ;

	if ( in_bActivationFlag ) {
		ZKVERIFY( frMaipRbPubActivate( &m_Error, (void*) m_PublicRBBuffer, GetPublicBufferSize() ) );
	}
	else {
		ZKVERIFY( frMaipRbPubDeactivate( &m_Error, (void*) m_PublicRBBuffer, GetPublicBufferSize() ) );
	}
}
