// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// WMaipRb.h

// Reply Block Object

// Wraps the frMaipRbPriv structure
// Also handles binary representation a reply blocks

// A reply block represents a route for mail to pass between
// the FMG and the users POP box.  There are several forms that this
// information gets packed, most of this complexity is hidden in this
// class.  A reply block as a WMaipRB provides an API to the actual
// contents.  A WMaipRbBinary can provide both a private binary version
// which must be kept on the user machine, and a public version which
// is shared with the nym server

// WMaipRb is the main class, but it makes use of 
// ReplyBlockHopInfo, WMaipRbBinary


#ifndef _WMAIPRB_H
#define _WMAIPRB_H

#include "WNetInfoDatum.h"
#include "WAutoBuffer.h"

#include <frmaip_repblk.h>		// Defines underlying code structure
#include <DZkError.h>

#include <AutoInstCount.h>

class RZkException;

#include "ClientTypesInc.h"

// Class ReplyBlockHopInfo
//
// This defines a single hop in a reply block, containing the signing key
// and the entity name.

class ReplyBlockHopInfo AUTO_INSTANCE_COUNTING1( ReplyBlockHopInfo )
{
public:
	ReplyBlockHopInfo( ZKSTD::string in_strEntityName, FrCryptCrypt in_Key ) 
		: m_strEntityName( in_strEntityName )
		, m_MaipSigningKey( in_Key ) {}

	FrCryptCrypt GetSigningKey() const
			{ ZKASSERT( m_MaipSigningKey != NULL ) ; return m_MaipSigningKey ; }

	void GetEntityName( ZKSTD::string& out_strEntityName ) const 
			{ ZKASSERT( m_strEntityName.length() > 0 ) ; out_strEntityName = m_strEntityName ; }

private:
	ZKSTD::string  m_strEntityName ;
	FrCryptCrypt m_MaipSigningKey ;
} ;

#include <vector>
typedef ZKSTD::vector<ReplyBlockHopInfo> ReplyBlockHopInfoVec;


// Class WMaipRbBinary
//
// This represents our binary presentation of a reply block,
// including both public and private buffers
// This object holds the information which is necessary for storing
// a reply block in the remote nym
//
// Client will not deal with this object directly, but with the WMaipRb
// instead

class WMaipRbBinary AUTO_INSTANCE_COUNTING1( WMaipRbBinary )
{
public:
	WMaipRbBinary()	: m_PrivateRBBuffer( NULL ) 
					, m_szPrivateBuf( 0 )
					, m_PublicRBBuffer( NULL )
					, m_szPublicBuf( 0 ) { ; } 

	WMaipRbBinary( const WMaipRbBinary & in_rhs ) ;
	WMaipRbBinary& operator =( const WMaipRbBinary & in_rhs ) ;

	void Clear() { m_Buffer.Clear() ; 
			  m_PrivateRBBuffer = NULL ; 
			  m_szPrivateBuf = 0 ; 
			  m_PublicRBBuffer = NULL ; 
			  m_szPublicBuf = 0 ; }

	// Initialize from an already existing binary representation
	void Initialize( 
			const unsigned char * in_pBuffer, 
			const size_t in_cntBuffer ) throw(RZkException);

	// Initialize from FrMaipRbPriv object
	void Initialize( const FrMaipRbPriv in_pRB ) ;

	// This Buffer represents the Complete RB information (including
	// both private and public.  This buffer is stored in the NymPrivateReplyBlock
	// parameter on the Remote Nym
	const unsigned char*  GetRBAsBinary() const { return m_Buffer ; }
				   size_t GetRBAsBinarySize() const { return m_Buffer.Size() ; }

	// This Buffer is the "private" representation of a RB and can be used
	// to create an FrMaipRbPriv object
	const unsigned char*  GetPrivateBuffer() { return m_PrivateRBBuffer ; }
	               size_t GetPrivateBufferSize() const { return m_szPrivateBuf; }

	// This Buffer represents the public information that can be sent across the network
	// to create a reply block on the nym server
	const unsigned char*  GetPublicBuffer() { return m_PublicRBBuffer ; }
	               size_t GetPublicBufferSize() const { return m_szPublicBuf ; }

	void SetActivationFlag( bool  in_bActivationFlag ) ;
private:

	void  ParseBuffer() ;

	mutable	DZkError	 m_Error ;	
	WAutoBuffer			 m_Buffer ;			// Our binary representation, includes both Private and Public

	const unsigned char* m_PrivateRBBuffer;	//Points to the Private portion of buffer
	size_t				 m_szPrivateBuf;
	const unsigned char* m_PublicRBBuffer;	//Points to the public portion of the buffer
	size_t				 m_szPublicBuf;
} ;


// Class WMaipRB
// This wraps a MaipRB object.  
// Use the Initialize methods to establish the state of the object
// This object does not communicate with the network.

class ClientTypes_API WMaipRb AUTO_INSTANCE_COUNTING1( WMaipRb )
{
public:

	// Creates a reply block object that has no contents.
	// It must be initialized with a call to Initialize() 
	WMaipRb() ;
	~WMaipRb() ;

	// Initialize based on an existing RB
	WMaipRb( const WMaipRb& ) ;
	WMaipRb& operator= ( const WMaipRb & ) ;

	// Create a new reply block based on these arguments
	// No information is communicated to the server
	void Initialize( 
			const ZKSTD::string&			in_strRbName,
			const ZKSTD::string&			in_strUsersEmailAddr,
			const ZKSTD::string&			in_strNymName,
			const FrCryptCrypt			in_FMGPublicCryptKey,
			const ReplyBlockHopInfoVec& in_MaipHops,
			const unsigned int			in_nLatency = DEFAULT_REPLYBLOCK_LATENCY ,
			const FrCryptSymAlg			in_Alg = DEFAULT_REPLYBLOCK_CRYPTO_ALG ) throw(RZkException);

	// Create a reply block based on the buffer form (ie such
	// as that stored inside a frRemoteNym
	// No information is communicated to the server
	void Initialize( 
			const unsigned char * in_pBuffer, 
			const size_t in_cntBuffer ) throw(RZkException);

	// Access to internal reply block - use with caution.
	// Normally this is only used for calling Core library functions which
	// expect the FrMaipRbPriv

	FrMaipRbPriv GetPrivateReplyBlock()       { return m_FrMaipRbPriv ; }
	FrMaipRbPriv GetPrivateReplyBlock() const { return m_FrMaipRbPriv ; }

	bool IsValid() const { return ( m_FrMaipRbPriv != NULL ) ; } 

	// Access Information out of the reply block
	// There are no Set methods because we do not currently support
	// modifying existing reply blocks (expect for there activation state)
	void GetName(ZKSTD::string &) const ;
	void GetUserEmailAddress( ZKSTD::string & out_strAddress ) const ;
	void GetNymName( ZKSTD::string & out_strNymName ) const ;

	FrCryptSymAlg GetReplyBlockCryptoAlgorithm() const ;
	FrCryptSymAlg GetReplyBlockWrapperAlgorithm() const ;

	unsigned int GetMessageLatency() const;

	void GetMaipNamesOfHops( ZKSTD::vector<ZKSTD::string>& out_vecstrReplyBlockDomainNames ) ;

	void GetActivationFlag( bool& out_bActivationFlag ) ;
	void SetActivationFlag( bool  in_bActivationFlag ) ;

	// Get access to the buffers which represent this RB

	// The Private Buffer remains on the local machine
	const unsigned char*  GetPrivateBuffer() { return m_RBBinary.GetPrivateBuffer() ; }
	               size_t GetPrivateBufferSize() const { return m_RBBinary.GetPrivateBufferSize(); }

	// The Public Buffer gets sent over
	const unsigned char*  GetPublicBuffer() { return m_RBBinary.GetPublicBuffer() ; }
	               size_t GetPublicBufferSize() const { return m_RBBinary.GetPublicBufferSize() ; }

	// This is a buffer that contains the entire binary information (both
	// private and public
	const unsigned char*  GetRBAsBinary() const	  { return m_RBBinary.GetRBAsBinary() ; }
				   size_t GetRBAsBinarySize() const { return m_RBBinary.GetRBAsBinarySize() ; }

	// Debug assistance
	void DumpToDebugOutput() const;
	void SelfTest() const;

private:

	// Internal helpers

	void Allocate() ;
	void Clear() ;

	void GetFirstHopInfo( frmaip_rbpriv_set_hop& out_h ) const ;


				FrMaipRbPriv   m_FrMaipRbPriv;    // holds the actual data
				WMaipRbBinary  m_RBBinary ;
	mutable		DZkError	   m_Error ;
} ;

typedef ZKSTD::vector<WMaipRb> WMaipRbVec;

#endif
