// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):


#include "StdAfx.h"
#include "WNetInfo.h"
#include "ZKConvert.h"
#include <TraceControl.h>
#include <IZkConsoleMsg.h>

WNetInfo::WNetInfo()
	: m_pData( NULL ) 
{
	ZKVERIFY( frNetInfoNew( NULL, &m_pData ) ) ;
}

WNetInfo::~WNetInfo()
{
	ZKVERIFY( frNetInfoFree( NULL, &m_pData ) ) ;
}

bool 
WNetInfo::GetEntityById(const uint32_t in_Id, WNetInfoDatum& out_Entity) const throw()
{
	FrNetInfoDatum p = NULL;

	// Note that just because frNetInfoSearchById returns 1, it doesn't
	// mean anything, out_Entity needs to be checked for nullness
	if(1 == frNetInfoSearchById(&m_Error,
								m_pData, 
								in_Id,
								&p) )
	{
		if( p )
		{
			WNetInfoDatum ret(p);
			out_Entity = ret;

			return out_Entity.IsValid();
		}
	}

	return false;
}

uint32_t 
WNetInfo::GetEntId(const int32_t IP, const uint16_t in_Port) const throw()
{
	uint32_t		ent_ident = 0;

	struct in_addr	iaddr;
	iaddr.s_addr = IP;

	if ( !frNetInfoEntIdentFromIPandPort(&m_Error,
                                    m_pData,
                                    iaddr,
                                    in_Port,
                                    &ent_ident) )
	{
		// No entity exists for this IP/Port
		ent_ident = 0 ;
	}

	return ent_ident;
}

size_t 
WNetInfo::GetEntityTypeCount( const FrEntIdType in_entType ) const throw()
{	
	ZkList pList = NULL ;
	size_t cntServers = GetListOfEntityType( in_entType, pList ) ;

	return cntServers ;
}

size_t
WNetInfo::GetListOfEntityType( FrEntIdType in_type, ZkList & out_pList ) const throw()
{
	// WARNING - Access to this function must always be thread safe because ZkList is not
	// thread safe

	out_pList = NULL ;

	// Ask our NetInfo object for the list
	if(!frNetInfoGetListOf(&m_Error, m_pData, &out_pList, in_type))
	{
		ZKCTRACE0( TRACE_NETINFO, "frNetInfoGetListOf failed" ) ;
		return 0 ;
	}

	if(out_pList == NULL)
	{
		ZKCTRACE0( TRACE_NETINFO, "frNetInfoGetListOf succeeded but returned a NULL pointer" ) ;		
		return 0 ;
	}

	// Iterate our list to get our data out
	size_t listsize = 0;

	ZKVERIFY(zkListSize (
		&m_Error ,              
		out_pList ,
		&listsize));
	
	if(0 == listsize) 
	{
		ZKCTRACE0( TRACE_NETINFO, "frNetInfoGetListOf succeeded but zkListSize returned a list size of 0" ) ;
		return 0 ;
	}

	ZKVERIFY(zkListMoveCursor(&m_Error, out_pList, LIST_HEAD));

	return listsize ;
}

void 
WNetInfo::GetServersOfType(
    const FrEntIdType& in_type, 
    WNetInfoDatumVec& Servers, 
    bool bIncludeDownServers ) const throw(RZkException)
{	
	// The ZkList used by the underlying object is not thread safe -
	// a single iterator point is maintained.
	// This means that concurrent calls to GetServersOfType will 
	// crash or get unexpected results

	ZkList pList = NULL;	// frNetInfoGetListOf gives us a pointer to their internal
							// list, not a copy, so we must not free it
	
	// Clear any existing content
	Servers.clear() ;

	size_t listsize = GetListOfEntityType( in_type, pList ) ;

	for(size_t i = 0; i < listsize; ++i)
	{
		ZKASSERT( pList != NULL ) ;

		FrNetInfoDatum data = NULL ;	

		// Get a pointer to the List's copy of the server data
		ZKVERIFY(zkListGet(&m_Error, pList, (void **)&data, LIST_FORWARDS));

		WNetInfoDatum Server( data ) ;

		bool bIncludeThisOne = bIncludeDownServers ? true : Server.IsServerUp() ;

		if ( bIncludeThisOne ) 
		{
			// Copy the data into our own Datum object and
			// add it to the list
			Servers.push_back( Server ) ;
		}
	}
	
	// Throw if we don't have any information about this type
	// Network should always have at least one of each requested type
	if(Servers.empty())
	{
		throw RZkException(
				m_Error, 
				"WNetInfo::ThrowFailedGetServers", 
				"There is no information about this server type in our Netinfo", 
				IDS_FAILED_NETWORK_CONFIG, 
				ZKConvert::GetServerNameFromType( in_type ).c_str() ) ;
	}

	// We do not free the list
	pList = NULL ;
}

void 
WNetInfo::GetListOfAips(   
			bool in_bIncludeNonCoreAips,
			bool in_bIncludeCoreAips,
			bool in_bIncludeDownServers,
			WNetInfoDatumVec& out_nlist ) const
{

	// This function is very similar to GetServersOfType()
	// Because Aips are much more numerous than
	// other core servers this function is a candidate for
	// possible performance optimization work

	// The ZkList used by the underlying object is not thread safe -
	// a single iterator point is maintained.
	// This means that concurrent calls to GetServersOfType will 
	// crash or get unexpected results

	ZKASSERT( in_bIncludeNonCoreAips || in_bIncludeCoreAips ) ;


	ZkList pList = NULL;	// frNetInfoGetListOf gives us a pointer to their internal
							// list, not a copy, so we must not free it
	
	// Clear any existing content
	out_nlist.clear() ;

	size_t listsize = GetListOfEntityType( FRENT_TYPE_AIP, pList ) ;

	// We won't crash if there aren't any Aips, but what is going on?

	ZKASSERT( pList != NULL ) ;
	ZKASSERT( listsize > 0 ) ;

	FrNetInfoDatum data;	

	for(size_t i = 0; i < listsize; ++i)
	{
		// Get a pointer to the List's copy of the server data
		ZKVERIFY(zkListGet(&m_Error, pList, (void **)&data, LIST_FORWARDS));

		// REVIEW : This temporary copy of the underlying data is wasteful
		WNetInfoDatum Server( data ) ;

		if ( !in_bIncludeDownServers && !Server.IsServerUp() )
		{
			// Hard coded to never return down servers
			continue ;
		}

		if ( Server.IsCoreAip() )
		{
			if ( !in_bIncludeCoreAips )
			{
				continue ;
			}
		}
		else
		{
			if ( !in_bIncludeNonCoreAips )
			{
				continue ;
			}
		}

		// Copy the data into our own Datum object and
		// add it to the list
		out_nlist.push_back( Server ) ;
	}

	// We do not free the list
	pList = NULL ;
}

void
WNetInfo::GetServer( const FrEntIdType& in_entType, WNetInfoDatum& out_Server ) const
{
	// REVIEW - this implementation can be optimized - currently we get all the 
	// servers into a list just to take the first one!  Because number of
	// servers of a particular type is relatively small this is ok for the moment - Andrew

	WNetInfoDatumVec lst ;
	GetServersOfType(in_entType, lst, true ) ;	

	if ( lst.size() > 1 )
	{
		size_t index = 0 ;

		if (!zkRandGetUInt32InRange( &m_Error, 0, lst.size() - 1, &index)) 
		{
			/* Call above can fail on the first run of the client because
			   there might not be any entropy yet.  Because random selection of
			   servers is more a question of load balancing instead of 
			   security, and this only will fail the first time client is run,
			   AdamB has given permission for this careful use of rand().
			   See 5811 */
			srand( (unsigned)time( NULL ) );
			index = (rand() % lst.size() );
		}

		out_Server = lst[ index ] ;
	}
	else
	{
		out_Server = lst.front() ;
	}
}


// This function does not throw exceptions
// It takes a frNetInfo object and attempts to create a buffer from it
void 
WNetInfo::GetAsBuffer(void** out_pBuffer, size_t* out_pCntSize) const throw()
{
	if (( out_pBuffer == NULL ) ||
		( out_pCntSize == NULL ))
	{
		ZKASSERT( "Invalid args to WNetInfo::GetAsBuffer" && false ) ;
	}

	*out_pBuffer = NULL ;
	*out_pCntSize = 0 ;

	ZKVERIFY(frNetInfoStore(&m_Error, m_pData, out_pBuffer, out_pCntSize));
}

bool 
WNetInfo::InitFromBuffer( const unsigned char* in_pBuffer, const unsigned long in_cntBuffer )
{
	// REVIEW - if this fails what state is our WNetInfo in?  Should we force a
	// flush so it is not ambiguous.  Ideally only if it is successful should
	// it actually change any state (eg transactionally safe)

	unsigned char*  pBuffer = const_cast<unsigned char*>( in_pBuffer );
	return ( 0 != frNetInfoLoadNoSig( &m_Error, m_pData, pBuffer, in_cntBuffer ) ) ;
}

void
WNetInfo::Flush() throw()
{
	frNetInfoDeleteAll(&m_Error, m_pData);
}

void
WNetInfo::RandomizeServerList( const FrEntIdType in_type ) throw()
{
	ZKVERIFY( frNetInfoRandomizeList(&m_Error, m_pData, in_type) );
}

void 
WNetInfo::RoundRobinEntity(const FrEntIdType in_Entity) throw()
{
	ZKVERIFY( frNetInfoRoundRobinList(&m_Error, m_pData, in_Entity) );
}

void
WNetInfo::SendInfoAboutEntityTypeToConsole( FrEntIdType in_type, bool in_bVerbose ) const
{
	IZkConsoleMsg::ShowMsg( 
					CONS_DEBUG, 
					CONSIMAGE_COMP_TERM,
					IDS_CONS_SERVER_COUNT,
					GetEntityTypeCount( in_type ),
					ZKConvert::GetServerNameFromType( in_type ).c_str() ) ;

	// Verbose mode could list all the entities as well.....
}