// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

///////////////////////////////////////////////////////////////////////////////
//
//
//
// WNetInfo
//
//
// This is one of the key objects in the operation of the 
// Freedom Client.
// 
// If represents the Freedom Network topology,
// including all Aips and all Core Servers (Nym servers, Token servers...)
// 
// Once initialized with this data, this class provides convenient
// functions for querying the data.  
//
// The WNetInfoDatum is a closely related object - it presents a single
// "entity" (ie server) in the WNetInfo.
//
// This class holds the data and knows how to parse and save the data to
// binary.  It does not know how to get the data, and makes no network connections
// 
// This class wraps the frNetInfo library/structure.   All client code should
// use this class instead of the underlying objects.
// 
///////////////////////////////////////////////////////////////////////////////

#ifndef WNETINFO_H
#define WNETINFO_H

#include <frNetInfo.h>		// Defines underlying code structure
#include "WNetInfoDatum.h"
#include "ClientTypesInc.h"
#include <AutoInstCount.h>
#include <DZkError.h>
#include <RZkException.h>

class ClientTypes_API WNetInfo AUTO_INSTANCE_COUNTING1( WNetInfo )
{
public:

	WNetInfo() ;
	virtual ~WNetInfo() ;

	// Provide access to underlying structure - eg when
	// making calls that expect an FrNetInfo
	operator FrNetInfo()	   { return m_pData ; }
	operator FrNetInfo() const { return m_pData ; }

	bool IsValid() const { return ( m_pData != NULL ) ; }

	////////////////////////////////////////////////////////////////////////////
	// Low Level Query Services
	////////////////////////////////////////////////////////////////////////////

	// Look up an entity by it's id, which must be unique accross
	// all servers of all types.
	bool GetEntityById(const uint32_t in_Id, WNetInfoDatum& out_Entity) const throw() ;

	// Returns 0 if Entity does not exist for IP and Port
	uint32_t GetEntId(const int32_t IP, const uint16_t in_Port) const throw() ;

	////////////////////////////////////////////////////////////////////////////
	// High Level Query Services
	////////////////////////////////////////////////////////////////////////////

	void GetServersOfType(
			const FrEntIdType& type, 
			WNetInfoDatumVec& Servers, 
			bool bIncludeDownServers ) const ;

	void GetListOfAips(   
			bool in_bIncludeNonCoreAips,
			bool in_bIncludeCoreAips,
			bool in_bIncludeDownServers,
			WNetInfoDatumVec& out_nlist ) const /* throw (RZkException) */ ;

	// Returns the first server of a specific type
	void GetServer( 
			const FrEntIdType& in_entType, 
			WNetInfoDatum& out_Server ) const ;

	// Return number of servers of a particular type
	size_t GetEntityTypeCount( const FrEntIdType in_entType ) const throw() ;	

	////////////////////////////////////////////////////////////////////////////
	// Management Services
	////////////////////////////////////////////////////////////////////////////

	// Provide contents to the WNetInfo
	// No signature verification is done, caller is expected to validate this
	// themselves.
	// A copy of the buffer is made so caller doesn't have to keep a copy around
	// (The buffer format is the direct result of an "MxQuery", which is the
	// compressed and signed binary representation of a WNetInfo.
	// The WNetInfo keeps this represenation around as part of its
	// data)

	bool InitFromBuffer( const unsigned char* in_pBuffer, const unsigned long in_cntBuffer ) ;

	bool GetSignedBuffer( uint8_t ** pcbuf, size_t * pbsz) ;

	// Clear out the contents
	void Flush() throw() ;

	// When there are multiple servers in the list, the first is considered the 
	// actively used one.  This function randomizes the list so that not all 
	// clients use the same server by default.
	void RandomizeServerList( const FrEntIdType in_type ) throw() ;

	// Move to the next server in the list
	void RoundRobinEntity(const FrEntIdType in_type) throw() ;

	// Returns a binary represenation of the state, caller is expected to free
	// the memory
	void GetAsBuffer(void** out_pBuffer, size_t* out_cntSize) const throw() ;

	////////////////////////////////////////////////////////////////////////////
	// Debugging/Informational Services
	////////////////////////////////////////////////////////////////////////////

	void SendInfoAboutEntityTypeToConsole( FrEntIdType in_type, bool in_bVerbose = false ) const ;

private:
	WNetInfo( const WNetInfo& ) ;
	WNetInfo& operator= ( const WNetInfo & ) ;

	// PURPOSE: Helper function for handling frNetInfoGetListOf() error handling
	// PROMISE: Size of list is return value and pList points to the head of a non-empty list
	// Will throw if no servers
	size_t GetListOfEntityType( FrEntIdType in_type, ZkList & out_pList ) const throw() ;

				FrNetInfo	m_pData ;
	mutable		DZkError	m_Error ;
} ;

#endif