// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// WNetInfoDatum.cpp

#include "StdAfx.h"
#include "WNetInfoDatum.h"

// A core aip is only recognized as such by having this prefix.
// In fact the agreed on naming convention is "core-zks-"
// but the client is currently a bit more 

#define CORE_AIP_PREFIX "core"
#define CORE_AIP_PREFIX_LEN 4

#include <ci_string.h> // for strnicmp

// Object is created but is not initialized with any contents
WNetInfoDatum::WNetInfoDatum()
	: m_pDatum( NULL )
{
	Allocate() ;
}

// Create an object based on an existing FrNetInfoDatum
WNetInfoDatum::WNetInfoDatum( const FrNetInfoDatum in_pfrDatum )
	: m_pDatum( NULL )
{
	ZKASSERT( in_pfrDatum != NULL ) ;

	Allocate() ;
	CopyFromFrNetInfoDatumPtr( in_pfrDatum ) ;
}

WNetInfoDatum::WNetInfoDatum( const WNetInfoDatum& in_rhs )
	: m_pDatum( NULL )
{
	Allocate() ;
	CopyFromFrNetInfoDatumPtr( in_rhs ) ;
}

WNetInfoDatum&
WNetInfoDatum::operator = ( const WNetInfoDatum& in_rhs )
{
	if ( &in_rhs != this )
	{	
		Clear() ;
		Allocate() ;
		CopyFromFrNetInfoDatumPtr( in_rhs ) ;
	}

	return *this ;
}

WNetInfoDatum::~WNetInfoDatum()
{
	Clear() ;
	m_pDatum = NULL ;
}

void 
WNetInfoDatum::Allocate()
{
	ZKASSERT( m_pDatum == NULL ) ;
	ZKVERIFY ( frNetInfoDatumNew( &m_Error, &m_pDatum ) ) ;
	ZKASSERT( m_pDatum != NULL ) ;	
}

void 
WNetInfoDatum::Clear()
{
	if ( m_pDatum != NULL ) { 	 
		ZKVERIFY( frNetInfoDatumFree( &m_Error, &m_pDatum ) ) ;
		ZKASSERT( m_pDatum == NULL ) ;
	}
}


void 
WNetInfoDatum::CopyFromFrNetInfoDatumPtr( FrNetInfoDatum in_pDatumToCopy )
{
	ZKASSERT( m_pDatum != NULL ) ;
	ZKASSERT( IsValid() ) ;

	unsigned char * pBuffer = NULL ;
	size_t          bsz = 0 ;

	ZKVERIFY( frNetInfoDatumStore ( &m_Error, in_pDatumToCopy, (void**) &pBuffer, &bsz ) );

	ZKASSERT( bsz > 0 ) ;
	ZKASSERT( pBuffer != NULL ) ;

	ZKVERIFY( frNetInfoDatumLoad( &m_Error, m_pDatum, (void*) pBuffer, bsz ) ) ;
	
	SAFEFREE( pBuffer ) ;
}

uint32_t	
WNetInfoDatum::GetId() const
{
	ZKASSERT( IsValid() ) ;

	uint32_t id;
	ZKVERIFY(frNetInfoDatumGetId(&m_Error, m_pDatum, &id));
	return id ;
}

int
WNetInfoDatum::GetStatus() const
{
	ZKASSERT( IsValid() ) ;

	int status;
	ZKVERIFY(frNetInfoDatumGetStatus(&m_Error, m_pDatum, &status));
	return status ;
}

FrEntIdExp	
WNetInfoDatum::GetEntity() const
{
	FrEntIdExp IdExp ;	// In this case the Fr type is a structure not pointer to structure
	memset( &IdExp, 0, sizeof( IdExp ) ) ;
	ZKVERIFY(frNetInfoDatumGetEntity(&m_Error, m_pDatum, &IdExp));
	return IdExp ;
}

void 
WNetInfoDatum::GetEntityName( ZKSTD::string& out_strName ) const
{
	ZKASSERT( IsValid() ) ;

	FrEntIdExp IdExp ;
	memset( &IdExp, 0, sizeof( IdExp ) ) ;
	ZKVERIFY(frNetInfoDatumGetEntity(&m_Error, m_pDatum, &IdExp));
	out_strName = IdExp.name ;
}

FrEntIdType	
WNetInfoDatum::GetEntityType() const
{
	ZKASSERT( IsValid() ) ;

	FrEntIdExp IdExp ;
	memset( &IdExp, 0, sizeof( IdExp ) ) ;
	ZKVERIFY(frNetInfoDatumGetEntity(&m_Error, m_pDatum, &IdExp));
	return IdExp.type ;
}

int			
WNetInfoDatum::GetRating() const
{
	ZKASSERT( IsValid() ) ;

	int Rating ;
	ZKVERIFY( frNetInfoDatumGetRating(&m_Error, m_pDatum, &Rating ) );
	return Rating ;
}

void		
WNetInfoDatum::GetRegion( ZKSTD::string& out_strRegion ) const
{
	ZKASSERT( IsValid() ) ;

	const char * pRating = NULL ;
	ZKVERIFY( frNetInfoDatumGetRegion(&m_Error, m_pDatum, &pRating ) );
	if ( pRating == NULL ) {
		out_strRegion.erase() ;
	}
	else {
		out_strRegion = pRating ;
	}
}

void		
WNetInfoDatum::GetRegionCode( ZKSTD::string& out_strRegionCode ) const
{
	ZKASSERT( IsValid() ) ;

	// This gets the complete string - Country + Province + Long + Lat

	const char * pszRegionCode = NULL ;
	ZKVERIFY( frNetInfoDatumGetRegionCode(&m_Error, m_pDatum, &pszRegionCode ) );
	out_strRegionCode = pszRegionCode ;
}

void
WNetInfoDatum::GetCountryCode( ZKSTD::string& out_strCountryCode ) const
{
	ZKASSERT( IsValid() ) ;

	// Region Code is encoded as XXYY where XX is country code, YY is province number

	ZKSTD::string strRegionCode ;
	GetRegionCode( strRegionCode ) ;
	out_strCountryCode = strRegionCode.substr(0,CTRYCODE_LEN) ;
}

void       
WNetInfoDatum::GetProvinceCode( ZKSTD::string& out_strProvinceCode ) const
{
	ZKASSERT( IsValid() ) ;

	ZKSTD::string strRegionCode ;
	GetRegionCode( strRegionCode ) ;

	if ( strRegionCode.length() >= CTRYCODE_LEN + PROVCODE_LEN ) 
	{
		out_strProvinceCode = strRegionCode.substr(CTRYCODE_LEN,PROVCODE_LEN) ;
	}
	else {
		out_strProvinceCode.erase() ;
	}
}

void        
WNetInfoDatum::GetLatitude( ZKSTD::string& out_str ) const
{
	ZKASSERT( IsValid() ) ;

	ZKSTD::string strRegionCode ;
	GetRegionCode( strRegionCode ) ;

	if ( strRegionCode.length() == REGCODE_LEN ) 
	{
		out_str = strRegionCode.substr( CTRYCODE_LEN + PROVCODE_LEN, LATITUDE_LEN ) ;
	}
	else
	{
		// Information is not available on this network
		out_str.erase() ;
	}
}

void        
WNetInfoDatum::GetLongitude( ZKSTD::string& out_str ) const
{
	ZKASSERT( IsValid() ) ;

	ZKSTD::string strRegionCode ;
	GetRegionCode( strRegionCode ) ;

	if ( strRegionCode.length() == REGCODE_LEN ) 
	{
		out_str = strRegionCode.substr( CTRYCODE_LEN + PROVCODE_LEN + LATITUDE_LEN, LONGITUDE_LEN ) ;
	}
	else
	{
		// Information is not available on this network
		out_str.erase() ;
	}
}


// Operator name and ID are packed in the same string as a hack
// This class hides that detail but we have to extract the values here.

void		
WNetInfoDatum::GetOperator( ZKSTD::string& out_strOperator ) const
{
	ZKASSERT( IsValid() ) ;

	ConvertOperatorStringToOperatorNameAndOperatorId( &out_strOperator, NULL ) ;
}

int		
WNetInfoDatum::GetOperatorID() const
{
	ZKASSERT( IsValid() ) ;

	int id ;
	ConvertOperatorStringToOperatorNameAndOperatorId( NULL, &id ) ;
	return id ;
}

SOCKADDR_IN 
WNetInfoDatum::GetSockAddr() const
{
	ZKASSERT( IsValid() ) ;

	SOCKADDR_IN addr ;
	ZKVERIFY( frNetInfoDatumGetSockaddr(&m_Error, m_pDatum, &addr ) );
	return addr ;
}

void		
WNetInfoDatum::SetEntity( FrEntIdExp in_entity )
{
  ZKVERIFY( frNetInfoDatumSetEntity(&m_Error, 
								  m_pDatum, 
								  in_entity) ) ;
}

void
WNetInfoDatum::SetRating( int in_rating )
{
	ZKVERIFY( frNetInfoDatumSetRating(&m_Error, m_pDatum, in_rating ) );
}

void
WNetInfoDatum::SetRegion( const char *in_pszRegion ) 
{
	ZKVERIFY( frNetInfoDatumSetRegion(&m_Error, m_pDatum, in_pszRegion ) );
}

void
WNetInfoDatum::SetRegionCode( const char *in_pszRegion ) 
{  
	ZKVERIFY( frNetInfoDatumSetRegionCode(&m_Error, m_pDatum, in_pszRegion ) ) ;
}
  
void
WNetInfoDatum::SetOperator( const char * in_pszServoperator )
{
	ZKVERIFY( frNetInfoDatumSetOperator(&m_Error, m_pDatum, in_pszServoperator ) );
}

void
WNetInfoDatum::SetSockAddr( const struct sockaddr_in& in_s_addr )
{
	ZKVERIFY( frNetInfoDatumSetSockaddr( &m_Error, m_pDatum, in_s_addr ) );
}

void		
WNetInfoDatum::SetId( uint32_t in_id )
{
	ZKVERIFY( frNetInfoDatumSetId( &m_Error, m_pDatum, in_id ) );
}

void		
WNetInfoDatum::SetStatus( int in_Status )
{
	ZKVERIFY( frNetInfoDatumSetStatus( &m_Error, m_pDatum, in_Status ) );
}


bool
WNetInfoDatum::IsServerUp() const
{
	return ( GetStatus() != 0 ) ;
}

uint32_t ScoreChar( char x )
{	
	return x % 23 ;
}

void
WNetInfoDatum::GetImaginaryPosition( float& out_x, float& out_y ) const
{
	// Until we get some long/lat we calculate a fake but repeatable
	// location for the AIP based on its name

	std::string strEntity ;

	GetEntityName( strEntity ) ;

	if ( strEntity.size() < 4 ) 
	{
		// Just to avoid any problems
		strEntity += "    " ;
	}

	size_t cntStr = strEntity.size() ;

	uint32_t x = ( ScoreChar( strEntity[ 0 ] ) << 2 ) + 
		         ( ScoreChar( strEntity[ 1 ] ) << 1 ) + 
		           ScoreChar( strEntity[ 2 ] ) ;

	uint32_t y = ( ScoreChar( strEntity[ cntStr - 1 ] ) << 3 ) + 
				 ( ScoreChar( strEntity[ cntStr - 2 ] ) << 2 ) + 
				 ( ScoreChar( strEntity[ cntStr - 3 ] ) << 1 ) + 
				   ScoreChar( strEntity[ cntStr - 4 ] ) ;

	uint32_t factorx = ( 22 << 2 ) + ( 22 << 1 ) + 22 ;
	uint32_t factory = ( 22 << 3 ) + ( 22 << 2 ) + ( 22 << 1 ) + 22 ;

	ZKASSERT( x <= factorx && y <= factory ) ;

	out_x = ((float)x) / (float)factorx ;
	out_y = ((float)y) / (float)factory ;

	ZKASSERT( out_x <= 1.0 ) ;
	ZKASSERT( out_y <= 1.0 ) ;
}



int
WNetInfoDatum::GetValueFromString( const char* in_pstr, size_t in_size )
{
	if ( in_pstr == NULL )
	{
		ZKASSERT( false ) ;
		return 0 ;
	}

	if ( in_size >= 8 )
	{
		ZKASSERT( false ) ;
		return 0 ;
	}

	char     strVal[ 8 ] = { 0 } ;
	strncpy( strVal, in_pstr, in_size ) ;

	return atoi( strVal ) ;	
}

// REVIEW:
// Further research has show that a common format for
// Long/Lat representation is with floats with range ( -90.0 - 90.0, -180 to 180)
// I'm not sure exact convention for which of N/S, or E/W is positive,
// but if this is relatively standard then we should convert this
// parsing code to return that convention and move it to its own class
// or function so that it can be used by others.

void 
WNetInfoDatum::ConvertRegionCodeToPosition( const char* in_pstrRegion, float& out_x, float& out_y )
{
	// Default position (we can hard code this to closer to montreal
	out_x = 0.5 ;
	out_y = 0.5 ;

	if ( in_pstrRegion == NULL )
	{
		ZKTRACE0(" Invalid format on Region Code, returning default longitude/latitude\n" ) ;
		return ;
	}

	uint32_t offsetLocation = 0 ; // Location in the string of start of the long/lat

	if ( strlen( in_pstrRegion ) == REGCODE_SHORT_LEN )
	{
		offsetLocation = CTRYCODE_LEN ;
	}
	else if ( strlen( in_pstrRegion ) == REGCODE_LEN )
	{
		offsetLocation = CTRYCODE_LEN + PROVCODE_LEN ;
	}
	else
	{
		ZKTRACE0(" Invalid format on Region Code, returning default longitude/latitude\n" ) ;
		return ;
	}

	// We are very dependent on the exact format for this string
	// However the code should be careful NEVER to crash even with bad data
	// because any false imaginable contents could potentially come to use by accident

	// Calculate the latitude

	bool     bNorth = true ;
	uint32_t minutesLatitude = 0 ;

	const uint32_t maxMinutesLatitude = 90 * 60 ;

	minutesLatitude = GetValueFromString( in_pstrRegion + offsetLocation, LATITUDE_DEGREE_LEN ) * 60 ;

	uint32_t minutes = GetValueFromString( in_pstrRegion + offsetLocation + LATITUDE_DEGREE_LEN, MINUTE_LEN ) ;

	if ( minutes <= 60 )
	{
		minutesLatitude += minutes ;
	}
	else
	{
		ZKASSERT( false ) ;
	}

	if ( minutesLatitude > maxMinutesLatitude )
	{
		// Bogus data
		ZKASSERT( false ) ;
		minutesLatitude = 0 ;
	}

	char Hemisphere = in_pstrRegion[ offsetLocation + LATITUDE_DEGREE_LEN + MINUTE_LEN ] ;

	if (( Hemisphere == 'S' ) ||
		( Hemisphere == 's' ) )
	{
		bNorth = false ;
	}
	else
	{
		ZKASSERT( ( Hemisphere == 'N' ) || ( Hemisphere == 'n' ) ) ;
	}

	if ( bNorth )
	{
		// Give a number between 0 (at 90N) and 0.5
		out_y = ((float) ( maxMinutesLatitude - minutesLatitude )) / ((float)( maxMinutesLatitude * 2 )) ;
	}
	else
	{
		// 0.5 ( 0 degree ) and 1.0 ( 90S )
		out_y = 0.5f + ( (float) minutesLatitude  / ((float)( maxMinutesLatitude * 2 ) ) ) ;
	}

	// Calculate the longitude

	bool     bWest = true ;
	uint32_t minutesLongitude = 0 ;

	const uint32_t maxMinutesLongitude = 180 * 60 ;

	uint32_t offsetLongitude = offsetLocation + LONGITUDE_OFFSET_REL_LAT ;

	minutesLongitude = GetValueFromString( in_pstrRegion + offsetLongitude, LONGITUDE_DEGREE_LEN ) * 60 ;
	minutes = GetValueFromString( in_pstrRegion + offsetLongitude + LONGITUDE_DEGREE_LEN, MINUTE_LEN ) ;

	if ( minutes <= 60 )
	{
		minutesLongitude += minutes ;
	}
	else 
	{
		ZKASSERT( false ) ;
	}

	if ( minutesLongitude > maxMinutesLongitude )
	{
		// Bogus data
		ZKASSERT( false ) ;
		minutesLongitude = 0 ;
	}

	Hemisphere = in_pstrRegion[ offsetLongitude + LONGITUDE_DEGREE_LEN + MINUTE_LEN ] ;

	if (( Hemisphere == 'E' ) || ( Hemisphere == 'e' ))
	{
		bWest = false ;
	}
	else
	{
		ZKASSERT( ( Hemisphere == 'W' ) || ( Hemisphere == 'w' ) ) ;
	}


	if ( bWest )
	{
		// Give a number between 0 (at 180W) and 0.5 at 0 degrees
		out_x = ((float) ( maxMinutesLongitude - minutesLongitude )) / ((float)( maxMinutesLongitude * 2 )) ;
	}
	else
	{
		// Give a number between 0.5 at 0 degrees and 1.0 at 180E
		out_x = 0.5f + ( (float) minutesLongitude  / ((float)( maxMinutesLongitude * 2 ) ) ) ;
	}

	ZKASSERT( out_x <= 1.0 ) ;
	ZKASSERT( out_y <= 1.0 ) ;
}

void        
WNetInfoDatum::GetPhysicalPosition( float& out_x, float& out_y ) const
{
	// Default position (we can hard code this to closer to montreal

	const char * pszRegionCode = NULL ;
	ZKVERIFY( frNetInfoDatumGetRegionCode(&m_Error, m_pDatum, &pszRegionCode ) );

	ConvertRegionCodeToPosition( pszRegionCode, out_x, out_y )  ;
}


size_t      
WNetInfoDatum::GetCntConnections() const
{
	size_t retval ;
	ZKVERIFY( frNetInfoDatumGetNumNb( &m_Error, m_pDatum, &retval ) ) ;
	return retval ;
}

uint32_t
WNetInfoDatum::GetConnectionX( size_t in_whichConnection ) const
{
	ZKASSERT( in_whichConnection < GetCntConnections() ) ;

	uint32_t retval ;
	ZKVERIFY( frNetInfoDatumGetNbX( &m_Error, m_pDatum, in_whichConnection, &retval ) ) ;
	return retval ;
}

void
WNetInfoDatum::ConvertOperatorStringToOperatorNameAndOperatorId( 
	ZKSTD::string * out_pOptionalString,		//Optional
	int *		  out_pOptionalID ) const   //Optional
{
	// Doesn't make sense to call with both parameters NULL
	ZKASSERT( ( out_pOptionalString != NULL ) || ( out_pOptionalID != NULL ) ) ;
		
	// Get a pointer to the operator string
	const char * pszOperator = NULL ;
	ZKVERIFY( frNetInfoDatumGetOperator(&m_Error, m_pDatum, &pszOperator ) );

	// Operator string may include an operator number - find it, convert it to
	// an integer and remove it

	// For example "Acme Inc [45]" becomes "Acme Inc"

	ZKSTD::string strOperator( pszOperator ) ;
	int iOperatorNum = 0 ;

	size_t StartPos = strOperator.rfind( '[' ) ;

	if ( StartPos != ZKSTD::string::npos )
	{
		size_t EndPos = strOperator.rfind( ']' ) ;

		if ( StartPos < EndPos )
		{
			iOperatorNum = atoi( strOperator.substr( StartPos + 1, EndPos - StartPos - 1 ).c_str() ) ;
		}
	}

	// Fill out the parameters

	if ( out_pOptionalID ) {
		*out_pOptionalID = iOperatorNum ;
	}

	if ( out_pOptionalString ) {
		if ( iOperatorNum == 0 ) {
			*out_pOptionalString = pszOperator ;
		}
		else {

			if ( strOperator[ StartPos - 1 ] == ' ' ) 
			{
				// Don't include the space before the "[" symbol
				*out_pOptionalString = strOperator.substr( 0, StartPos - 1) ;
			}
			else
			{
				*out_pOptionalString = strOperator.substr( 0, StartPos - 1) ;
			}
		}
	}
}

/*

// This is what the function is supposed to be... removed because some networks still have
// the old hack....

bool 
WNetInfoDatum::IsCoreAip() const 
{
	// Core aip is destinguished based on the naming scheme - it
	// starts with a certain reserved prefix.  Eg core-zks-01 is a core aip

	ZKSTD::string strEntity ;
	GetEntityName( strEntity ) ;

	return ( 0 == strnicmp( strEntity.c_str(), CORE_AIP_PREFIX, CORE_AIP_PREFIX_LEN ) );
}

*/


bool 
WNetInfoDatum::IsCoreAip() const 
{
	// Experimental code for Core Aip being a simple naming scheme difference.
	ZKSTD::string strEntity ;
	GetEntityName( strEntity ) ;

	// TEMPORARY - host 5 and host 6 are just core
	// aips for the sake of internal testing (eg t12 network).....
	return (( 0 == strncmp( strEntity.c_str(), "host5", 5 ) ) ||
			( 0 == strncmp( strEntity.c_str(), "host6", 5 ) ) ||
			( 0 == strnicmp( strEntity.c_str(), CORE_AIP_PREFIX, CORE_AIP_PREFIX_LEN ) ) );
}



void
WNetInfoDatum::DumpToDebugOutput() const
{
#ifdef _DEBUG
	ZKTRACE0("---------------NETINFO DATUM DUMP------------------\n");

	
	ZKSTD::string Str ;
	GetEntityName( Str ) ;
	ZKTRACE1("GetEntityName() - %s\n\n", Str.c_str() );

	// For debug usage only - is there a vanilla sockets equivalent?
	const unsigned long bufLen = 100;
	char buf[bufLen] = { 0 };
	unsigned long nSize = bufLen;

	sockaddr_in addr = GetSockAddr() ;
	snprintf( buf, nSize, "%s:%hu", inet_ntoa( addr.sin_addr ), ntohs(addr.sin_port) ) ;

	ZKTRACE1("GetSockAddr() - %s\n\n", buf );
		
	ZKTRACE1("GetStatus() - %d\n", GetStatus());

	if ( IsServerUp() ) {
		ZKTRACE0("Server is Up\n" );
	}
	else {
		ZKTRACE0("Server is Down\n" );
	}

	ZKTRACE1("GetId() - %d\n", GetId());
	ZKTRACE1("GetRating() - %d\n", GetRating());

	ZKTRACE1("GetEntityType() - %d\n", GetEntityType());

	GetOperator( Str ) ;
	ZKTRACE1("GetOperator() - %s\n", Str.c_str() );

	ZKTRACE1("GetOperatorID() - %d\n", GetOperatorID());
	
	GetRegion( Str ) ;
	ZKTRACE1("GetRegion() - %s\n", Str.c_str() );

	GetRegionCode( Str ) ;
	ZKTRACE1("GetRegionCode() - %s\n", Str.c_str() );

	GetCountryCode( Str ) ;
	ZKTRACE1("GetCountryCode() - %s\n", Str.c_str() );

	GetProvinceCode( Str ) ;
	ZKTRACE1("GetProvinceCode() - %s\n", Str.c_str() );

	GetLatitude( Str ) ;
	ZKTRACE1("GetLatitudeCode() - %s\n", Str.c_str() );

	GetLongitude( Str ) ;
	ZKTRACE1("GetLongitude() - %s\n", Str.c_str() );

	float x1, x2, y1, y2 ;
	GetImaginaryPosition( x1, y1 ) ;
	GetPhysicalPosition( x2, y2 ) ;
	ZKTRACE4( "Imaginary coordinates( %f, %f ), Real Coordinates (%f, %f)", x1, y1, x2, y2 ) ;

	ZKTRACE0("----------------------------------------------------\n\n");
#endif
}



void		
WNetInfoDatum::SelfTest()
{
	ZKTEST( 5 == GetValueFromString( "5", 1 ) ) ;
	ZKTEST( 5 == GetValueFromString( "51", 1 ) ) ;
	ZKTEST( 5 == GetValueFromString( "5A", 1 ) ) ;
	ZKTEST( 51 == GetValueFromString( "51", 2 ) ) ;

	float x, y ;
	ConvertRegionCodeToPosition( "CA108905S10320W", x, y ) ;

	// Boundary conditions
	ConvertRegionCodeToPosition( "CA109000N18000W", x, y ) ;
	ZKTEST( x == 0.0 ) ;
	ZKTEST( y == 0.0 ) ;

	ConvertRegionCodeToPosition( "CA109000S18000E", x, y ) ;
	ZKTEST( x == 1.0 ) ;
	ZKTEST( y == 1.0 ) ;

	ConvertRegionCodeToPosition( "CA100000S00000E", x, y ) ;
	ZKTEST( x == 0.5 ) ;
	ZKTEST( y == 0.5 ) ;

	ConvertRegionCodeToPosition( "CA100000N00000W", x, y ) ;
	ZKTEST( x == 0.5 ) ;
	ZKTEST( y == 0.5 ) ;


	// Format without regions number

	ConvertRegionCodeToPosition( "CA8905S10320W", x, y ) ;

	// Boundary conditions
	ConvertRegionCodeToPosition( "CA9000N18000W", x, y ) ;
	ZKTEST( x == 0.0 ) ;
	ZKTEST( y == 0.0 ) ;

	ConvertRegionCodeToPosition( "CA9000S18000E", x, y ) ;
	ZKTEST( x == 1.0 ) ;
	ZKTEST( y == 1.0 ) ;

	ConvertRegionCodeToPosition( "CA0000S00000E", x, y ) ;
	ZKTEST( x == 0.5 ) ;
	ZKTEST( y == 0.5 ) ;

	ConvertRegionCodeToPosition( "CA0000N00000W", x, y ) ;
	ZKTEST( x == 0.5 ) ;
	ZKTEST( y == 0.5 ) ;

/*
	// Verify that we don't crash with invalid data
	// These cause some asserts to I run them during personal
	// testing but not as part of normal unit test

	ConvertRegionCodeToPosition( "CA10", x, y ) ;
	ConvertRegionCodeToPosition( "CA1", x, y ) ;
	ConvertRegionCodeToPosition( "CA108905S10", x, y ) ;
	ConvertRegionCodeToPosition( "CA108905S10320Y", x, y ) ;
	ConvertRegionCodeToPosition( "CA109599S10320Y", x, y ) ;
	ConvertRegionCodeToPosition( "CA108905S99920E", x, y ) ;
	ConvertRegionCodeToPosition( "CA108905S10099W", x, y ) ;
	ConvertRegionCodeToPosition( "CA108905S10", x, y ) ;
	ConvertRegionCodeToPosition( "CA8905S10320Y", x, y ) ;
	ConvertRegionCodeToPosition( "CA9599S10320Y", x, y ) ;
	ConvertRegionCodeToPosition( "CA8905S99920E", x, y ) ;
	ConvertRegionCodeToPosition( "CA8905S10099W", x, y ) ;
	ConvertRegionCodeToPosition( "RIGHTLENGHBUTNO", x, y ) ;
*/
}
