// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// WNetInfoDatum.h

// Wraps the T_frNetInfoDatum.h structure

// This is one of the most important structures kicking around
//
// It describes a particular server on the network - an Freedom Server such as
// the KeyQueryServer or a Maip or an Aip.  It hold such useful information
// as the location of the server (geographic and IP), the name of the server,
// information about its status (rating and whether it is up or down) etc.
// 

#ifndef _WNETINFODATUM_H
#define _WNETINFODATUM_H

#include <frNetInfo.h>		// Defines underlying code structure
#include <DZkError.h>

#include <AutoInstCount.h>

#include "ClientTypesInc.h"

#ifdef ZK_LINUX
#  include <clientlinux.h> // For SOCKADDR_IN
#endif

// Client is responsible for parsing the Region Code string
// The region string contains the Country Code, 
// The province code, then the long/lat
// Longitude and Latitude are encoded in a string with very strict
// formatting.  Eg 1020N10020W

#define REGCODE_LEN			   15		// The region is the Country + Province + Latitude + Long
#define REGCODE_SHORT_LEN      13       // The region is the Country + Latitude + Long
#define	CTRYCODE_LEN			2
#define	PROVCODE_LEN			2
#define GEOPOS_LEN             11       // Lat + Long
#define LATITUDE_LEN            5		// 2 characters degree, 2 characters minute, N/S
#define LONGITUDE_LEN           6       // 3 characters degree, 2 characters minute, E/W

#define LATITUDE_START_POS       CTRYCODE_LEN + PROVCODE_LEN
#define LONGITUDE_OFFSET_REL_LAT LATITUDE_DEGREE_LEN + MINUTE_LEN + 1  // Distance between start of Long/Lat and start of Long

#define MINUTE_LEN              2
#define LATITUDE_DEGREE_LEN     2
#define LONGITUDE_DEGREE_LEN    3


//class DZkError ;

class ClientTypes_API WNetInfoDatum AUTO_INSTANCE_COUNTING1( WNetInfoDatum )
{
public:

	WNetInfoDatum() ;

	// Initialize based on an existing FrNetInfoDatum
	explicit WNetInfoDatum( const FrNetInfoDatum ) ;

	WNetInfoDatum( const WNetInfoDatum& ) ;
	WNetInfoDatum& operator= ( const WNetInfoDatum & ) ;

	~WNetInfoDatum() ;

	// Provide access to underlying structure - eg when
	// making calls that expect an FrNetInfoDatum
	operator FrNetInfoDatum()		{ return m_pDatum ; }
	operator FrNetInfoDatum() const { return m_pDatum ; }

	bool IsValid() const { return ( m_pDatum != NULL ) ; } 

	// Data access

	uint32_t	GetId() const ;
	int			GetStatus() const ;
	FrEntIdExp	GetEntity() const ;		
	void		GetEntityName( ZKSTD::string& ) const ;	// Convenience above GetEntity()
	FrEntIdType	GetEntityType() const ;

	int			GetRating() const ;
	void		GetRegion( ZKSTD::string& ) const ;		// REVIEW - we could also provide char* pointers to our internal data
	void		GetRegionCode( ZKSTD::string& ) const ;

	void        GetCountryCode( ZKSTD::string& ) const ;		// The region code can be broken down into country code
	void        GetProvinceCode( ZKSTD::string& ) const ;		// and province number
	void        GetLatitude( ZKSTD::string& ) const ;		
	void        GetLongitude( ZKSTD::string& ) const ;		

	void		GetOperator( ZKSTD::string&) const ;
	int			GetOperatorID() const ;
	SOCKADDR_IN GetSockAddr()  const ;

	// Allows use to centralize the mechanism of distinguishing between a core aip and 
	// any other entity in a single location
	bool		IsCoreAip() const ;
		

	// Returns a coordinate for the aip in a distribution
	// that attempts to space the aips well.
	void        GetImaginaryPosition( float& out_x, float& out_y ) const ;

	// Returns a coordinate for the aip in a geographically
	// accurate distribution
	void        GetPhysicalPosition( float& out_x, float& out_y ) const ;

	// Provide access to the connection information
	size_t      GetCntConnections() const ;
	uint32_t	GetConnectionX( size_t in_whichConnection ) const ;

	void		SetId( uint32_t ) ;	
	void		SetStatus( int ) ;
	void		SetEntity( FrEntIdExp in_entity ) ;
	void		SetRating( int in_rating ) ;
	void		SetRegion( const char *in_pszRegion ) ;
	void		SetRegionCode( const char *in_pszRegion ) ;
	void		SetOperator( const char * in_pszServoperator ) ;
	void		SetSockAddr( const struct sockaddr_in& in_addr ) ;

	// Helpful function to quickly check the status of a server according
	// to our knowledge of the network coming from the NIQS
	bool        IsServerUp() const ;

	// Debug assistance
	static void	SelfTest() ;
	void        DumpToDebugOutput() const;

private:

	// Internal helpers

	void Allocate() ;
	void Clear() ;

	void CopyFromFrNetInfoDatumPtr( FrNetInfoDatum in_pDatumToCopy ) ;

	void ConvertOperatorStringToOperatorNameAndOperatorId( 
			ZKSTD::string * out_pOptionalString,			// Optional
			int *		  out_pOptionalID ) const ;     // Optional

	// Helper function - can also go in general string helper location
	// This function can read a value out of a string without
	// requiring that the number be followed by a \0
	// For example GetValueFromString( "20W", 2 ) returns 20

	static int GetValueFromString( const char* in_pstr, size_t in_size ) ;

	// Converts a well formed region code string into a mapping (0-1.0,0-1.0)
	// where (0.0, 0.0) is the position 90N, 180W on the map (eg top left hand
	// side).  Conversion assumes an even distance between degrees longitude and
	// latitude.  Other mapping require different algorithms
	static void ConvertRegionCodeToPosition( const char* in_pstrRegion, float& out_x, float& out_y ) ;

	// Member Data
				FrNetInfoDatum m_pDatum ;			// Points to the internal core type which
													// holds the actual data
	mutable		DZkError	   m_Error ;
} ;

#include <vector>
typedef ZKSTD::vector<WNetInfoDatum> WNetInfoDatumVec;

#endif
