// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// WNymIconInfo
//
// Class containing information about which Image (icon) represents the Nym on the UI
//
// This class takes care of hiding how we remember the nym icon
// Unfortunately there are currently some platform differences - 
//    Windows records an index value referring into a standard bitmap of nym icons (an array)
//	  Linux will record a file name
//	  Mac intends to record a complete icon file
//
// In general the most portable format is to record the icon directly inside this
// class in a safe cross platform image format (jpg? png?).  We do not have the time to do this
// immediately, but this class provides a centralized place to add this code.

// One important point: Nym files need to transfer between platforms.  Although
// because of the current platform differences, described above, prevent us currently from
// achieving cross platform nym images, we need to at least be able to revert to 
// "default" images.  One possibility was to allow multiple representations to
// reside in the same class - eg both a nym index _and_ and nym file name.  This allows
// the file to be tranfered back and forth between platforms without losing the image
// settings.  I don't think this is a high priority item so I'm not putting support it
// for this at the moment.  However as a compromise the code can be written so that it
// will not override the settings of another platform unless the value is changes
// from default on another platform.
// For example if the settings is "file.x" on linux and we transfer to windows
// we will show DEFAULT_NYM_ICON_INDEX.  On if the user changes the index then we lose the "file.x"
// when they move back to linux.

#ifndef WNYMICONINFO_H
#define WNYMICONINFO_H

#define DEFAULT_NYM_ICON_INDEX 1	// 0 is used for true identity look

#include "ZKChunk.h"
#include "ClientTypesInc.h"

enum NymIconType
{
	NymIconTypeUndefined = 0,
	NymIconIndex         = 1,
	NymIconFilePath      = 2
//  In future we will add capability to store raw Nym icon data
//  but we need to determine what format(s)
//	NymIconJpgData
} ;

class ClientTypes_API WNymIconInfo
{
public:

	WNymIconInfo() ;

	WNymIconInfo( const WNymIconInfo& in_rhs ) ;
	WNymIconInfo& operator=( const WNymIconInfo& in_rhs ) ;

	NymIconType   GetIconType() const { return m_IconType ; }
	ZKSTD::string GetFilePath() const ;
	size_t		  GetIconIndex() const ; 
	
	void SetIconIndex( size_t in_IconIndex ) ; 
	void SetIconFilePath( ZKSTD::string in_IconFilePath ) ; 

	void Save( WMemBuf& io_Buffer ) throw(RZkException) ;
	void Load( ZKChunkReader& io_ChunkReader ) throw(RZkException) ;

private:
	NymIconType    m_IconType ;
	size_t         m_IconIndex ;
	ZKSTD::string  m_strNymIconFilePath ;
} ;

#endif
