// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// WRemoteNym.cpp : 
//

#include "StdAfx.h"
#include "WRemoteNym.h"
#include <frNymClient.h>
#include <frNymUtil.h>
#include <zkDateTime.h>
#include <RZkException.h>

WRemoteNym::WRemoteNym()
{
	ZKVERIFY(frRemoteNymNew(NULL, &m_pFrRemoteNym)); 
}

WRemoteNym::~WRemoteNym()
{
	frRemoteNymSetPrivateSignatureKey(NULL, m_pFrRemoteNym, NULL);
	ZKVERIFY(frRemoteNymFree(NULL, &m_pFrRemoteNym));
}

bool WRemoteNym::GetBuffer( unsigned char** out_ppBuffer, 
						    size_t* cntBuffer ) const
{
	ZKASSERT( m_pFrRemoteNym ) ;
	ZKASSERT( out_ppBuffer != NULL ) ;

	bool ReturnValue = true;

	ZKVERIFY(frRemoteNymToBinary (
		&m_ZkError ,								
		m_pFrRemoteNym ,  
		out_ppBuffer ,										
		cntBuffer ));									

	return ReturnValue;
}

bool WRemoteNym::InitializeFromBuffer( const unsigned char* in_pBuffer, 
									   size_t in_cntBuffer)
{
	ZKASSERT( in_pBuffer != NULL ) ;

	bool ReturnValue = true;

	FrRemoteNym	pNewNym = NULL ;

	// Create a new Remote Nym structure and fill it with the
	// contents of the provided buffer
	if ( !frRemoteNymFromBinary (
				&m_ZkError ,								
				&pNewNym,  
				NULL, //FrCryptSig ,						
				in_pBuffer ,										
				in_cntBuffer )) {
		ZKASSERT( false ) ;
		ReturnValue = false ;
	}
	else {
		ZKASSERT( pNewNym != NULL ) ;

		// Replace our existing structure with this
		// new structure
		ZKVERIFY(frRemoteNymFree(NULL, &m_pFrRemoteNym));
		m_pFrRemoteNym = pNewNym ;
		pNewNym = NULL ;
	}	

	return ReturnValue;
}

bool WRemoteNym::DupFrRemoteNym(
	FrRemoteNym pOriginal, 
	FrRemoteNym* ppCopy)
{
	bool ReturnValue = true;
	unsigned char* NymBuf = NULL;
	size_t szNymBuf = 0;

	ZKVERIFY(frRemoteNymToBinary (
		&m_ZkError ,		
		pOriginal ,			
		&NymBuf ,			
		&szNymBuf ));		

	ZKVERIFY(frRemoteNymFromBinary (
		&m_ZkError ,		
		ppCopy ,				
		NULL ,				
		NymBuf ,			
		szNymBuf ));

	free( NymBuf );

	return ReturnValue;
}


bool WRemoteNym::DupFrRemoteNymWithSignatureKey(
	FrRemoteNym pOriginal, 
	FrRemoteNym* ppCopy)
{
	bool ReturnValue = true;

	FrCryptPubKey pEncryptKey = NULL ;
	FrCryptPubKey pSigKey = NULL ;
		
	unsigned char* NymBuf = NULL;
	size_t szNymBuf = 0;

	if (!frRemoteNymToBinary (
				&m_ZkError ,		
				pOriginal ,			
				&NymBuf ,			
				&szNymBuf )) {
		ZKASSERT( false ) ;
		ReturnValue = false ;
	}
	else if ( !frRemoteNymFromBinary (
				&m_ZkError ,		
				ppCopy ,				
				NULL ,				
				NymBuf ,			
				szNymBuf )) {
		ZKASSERT( false ) ;
		ReturnValue = false ;
	}
	else if ( !frRemoteNymGetPrivateKeys (
							&m_ZkError,
							pOriginal ,
							&pEncryptKey ,
							&pSigKey ) ) {
		ZKASSERT(false);
		ReturnValue = false;
	}
	else if ( !frRemoteNymSetPrivateSignatureKey(
					&m_ZkError,
					* ppCopy,
					pSigKey ) ) {
		ZKASSERT(false);
		ReturnValue = false;

		if ( pSigKey )
			frCryptPubKeyFree(NULL,&pSigKey);
	}

	// Cleanup

	free( NymBuf );

	if ( pEncryptKey )
		frCryptPubKeyFree(NULL,&pEncryptKey);

	return ReturnValue;
}

// Private method to pull a date out of the Server Parameters
void WRemoteNym::GetDateParam( enum FrRemoteNymParameter in_NymDateParameter, struct T_ZkDateTime& out_Date ) throw(RZkException)
{
	ZKFunctionEntry( "ZKNymServerSubSys::GetNymDate" ) ;

	const unsigned char* buf = NULL;
	size_t szbuf = 0;
	
	int bStatus = true;

	bStatus = frRemoteNymResetParameterSearch (&m_ZkError, m_pFrRemoteNym ) ;
	ZKTestAndThrow(bStatus);

	bStatus = frRemoteNymGetParameter ( 
					&m_ZkError , 
					m_pFrRemoteNym , 
					in_NymDateParameter , 
					&buf , 
					&szbuf ) ;
	ZKTestAndThrow(bStatus);

	// Convert buffer format of date into a ZKDateTime structure
	bStatus = zkDateTimeParse (
						&m_ZkError ,
						&out_Date ,
						(char*)buf ) ;
	ZKTestAndThrow(bStatus);
}

// Returns empty string if Parameter does not exist
void WRemoteNym::GetStringParam( enum FrRemoteNymParameter in_WhichParam, ZKSTD::string & out_string )
{
	out_string.erase() ;
		
	const unsigned char* strNymName = NULL;
	size_t szstrNymName = 0;

	ZKVERIFY(frRemoteNymResetParameterSearch (
		&m_ZkError ,                 
		m_pFrRemoteNym ));

	// If parameter doesn't exist we return an empty string

	if ( frRemoteNymGetParameter ( 
			&m_ZkError , 
			m_pFrRemoteNym , 
			in_WhichParam , 
			&strNymName , 
			&szstrNymName ))
	{
		if ( szstrNymName > 0 )
		{
			out_string = (char*) strNymName ;
		}
	}
}

void WRemoteNym::SetStringParam( enum FrRemoteNymParameter in_WhichParam, ZKSTD::string & in_string )
{
	// Make sure only one instance of this description type exists
	frRemoteNymDeleteParameter (
		&m_ZkError  ,                  
		m_pFrRemoteNym ,  
		in_WhichParam );

	ZKVERIFY(frRemoteNymAddParameter(         
		&m_ZkError, 
		m_pFrRemoteNym, 
		in_WhichParam, 
		(const unsigned char*)in_string.c_str(), 
		in_string.length() + 1) );
}

bool WRemoteNym::CopyReqdFrRemoteNymParamsToLocalFrRemoteNym(
	FrRemoteNym  ClientSideNym,
	FrRemoteNym  ServerSideNym)
{
	bool ReturnValue = true;
	
	// We copy (or overwrite) all parameters with the "FL_NYMSERVER"
	// flag from the server side copy to the client side copy.
	// This allows information such as the expiry data to travel
	// from the server.

	ZKVERIFY(frRemoteNymReplaceParameters ( 
		  &m_ZkError ,			
		  ServerSideNym ,		
		  ClientSideNym ,		
		  FL_NYMSERVER ));    

	return ReturnValue;
}

int DumpNymCallback (
	ZkError ,                  
	void * ,                   
	const char * pData,             
	size_t cntBuffer )
{
	// We can't trust pData to be NULL terminated
	char strOutput[ 256 ] = { 0 };
	
	if (cntBuffer > 255)
		cntBuffer = 255;

	memcpy( strOutput, pData, cntBuffer );
	ZKTRACE1("%s", strOutput);
	return 1 ;
}

// Show the contents of the Remote Nym in the Trace window
void WRemoteNym::DebugTraceRemoteNym( T_FrRemoteNym* in_pRemoteNym ) 
{
#ifdef _DEBUG
	ZKTRACE0( "\n-----------------------frRemoteNym Dump---------------------\n" ) ;

	ZkError	pZkError;
	ZKVERIFY(zkErrorNew(&pZkError));
	
	T_FrRemoteNym* pRemoteNym = in_pRemoteNym ;
	if ( pRemoteNym == NULL )
		pRemoteNym = m_pFrRemoteNym ;

	frRemoteNymVDump ( pZkError, pRemoteNym, DumpNymCallback, NULL ) ;
//	frRemoteNymDump2 ( pZkError, m_pFacade->GetFrRemoteNym(), DumpNymCallback, NULL ) ;

	ZKVERIFY(zkErrorFree(&pZkError));

	ZKTRACE0( "\n-----------------------End frRemoteNym Dump-----------------\n" ) ;
#endif
}

