// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "StdAfx.h"
#include "WRouteRules.h"

WRouteRules::WRouteRules()
	: m_First( 0 )
	, m_Last( 0 )
	, m_cntHops( DEFAULT_NUMBER_OF_HOPS )
	, m_bUseCoreAipAsExit( false )
{
}

WRouteRules::WRouteRules( size_t in_cntHops )
	: m_First( 0 )
	, m_Last( 0 )
	, m_cntHops( in_cntHops )
	, m_bUseCoreAipAsExit( false )
{
}

WRouteRules::WRouteRules( const MapRules& in_rules )
	: m_First( 0 )
	, m_Last( 0 )
	, m_cntHops( DEFAULT_NUMBER_OF_HOPS )
	, m_bUseCoreAipAsExit( false )
{
	SetRules( in_rules ) ;
}

void
WRouteRules::ClearRules()
{
	m_First = 0 ;
	m_Last = 0 ;
	ClearAvoidHops() ;
	m_bUseCoreAipAsExit = false ;

//  I'm not sure if ClearRule() should include number 
//  of hops
//	m_cntHops = DEFAULT_NUMBER_OF_HOPS ;
}

void 
WRouteRules::SetRules( const MapRules& in_rules )
{
	ClearRules() ;
	
	MapRules::const_iterator it ;
	
	for (it = in_rules.begin();
		 it != in_rules.end();
		 ++it)
	{	
		if ( (*it).second == RULE_FIRSTHOP) 
		{
			SetFirstHop( (*it).first ) ;
		}
		else if ( (*it).second == RULE_EXITFUNNEL) 
		{
			SetLastHop( (*it).first ) ;
		}
		else if ( (*it).second == RULE_AVOID) 
		{
			AddAvoidHop( (*it).first ) ; 
		}
	}

}

bool 
WRouteRules::DoesRoutePassRules( const WRoute& in_Route, size_t& out_FailedIndex ) const
{
	size_t cntHops = in_Route.GetSize() ;

	ZKASSERT( cntHops > 0 ) ;

	if ( m_cntHops != cntHops )
	{
		// This hop doesn't meet the basic 
		// number of hop criteria
		ZKASSERT( false ) ;
		return false ;
	}

	if (( cntHops == 1 ) && ( m_Last != 0 ) && ( m_First != 0 ))
	{
		// We have a dilemma - these particular rules only make good sense
		// for multiple hop routes.  We can work around by ignoring the 
		// first hop rule
		ZKASSERT( false ) ;	
	}

	// Test Avoid Hops

	for ( size_t k = 0 ; k < cntHops ; k++ )
	{
		uint32_t nextHop = in_Route.GetHop( k ) ;

		for ( size_t i = 0 ; i < m_AvoidHops.size() ; i++ )
		{
			if ( nextHop == m_AvoidHops[ i ] )
			{
				out_FailedIndex = k ;
				return false ;
			}
		}
	}

	// Test first hop

	if ( m_First != 0 )
	{
		if ( m_First != in_Route.GetHop( 0 ) )
		{
			// This can allow the caller to cull all
			// routes that start with this invalid first hop
			out_FailedIndex = 0 ;
			return false ;
		}
	}

	// Test last hop
	if ( m_Last != 0 )
	{
		if ( m_Last != in_Route.GetHop( cntHops - 1 ) )
		{
			out_FailedIndex = cntHops - 1 ;
			return false ;
		}
	}
	


	return true ;
}

void 
WRouteRules::AddAvoidHop( uint32_t in_EntityToAvoid)
{
#ifdef DEBUG
	for ( size_t i = 0 ; i < m_AvoidHops.size() ; i++ )
	{
		if ( m_AvoidHops[ i ] == in_EntityToAvoid)
		{
			// Already in the list, what is going on?
			ZKASSERT( false ) ;
			return ;
		}
	}

#endif

	m_AvoidHops.push_back( in_EntityToAvoid) ;
}

void 
WRouteRules::ClearAvoidHops()
{
	m_AvoidHops.clear() ;
}

bool 
WRouteRules::IsHopAvoided( uint32_t in_Aip ) const
{
	for ( size_t i = 0 ; i < m_AvoidHops.size() ; i++ )
	{
		if ( m_AvoidHops[ i ] == in_Aip )
		{
			return true ;
		}
	}

	return false ;
}


bool 
WRouteRules::AreRulesSet() const
{
	if (( m_AvoidHops.size() > 0 ) ||
		( m_Last != 0 ) ||
		( m_First != 0 ) )
	{
		return true ;
	}

	return false ;
}

void 
WRouteRules::SetHopCount( size_t in_cntHops )
{
	ZKASSERT( in_cntHops > 0 ) ;
	ZKASSERT( in_cntHops <= MAX_AIPS_IN_ROUTE ) ;

	m_cntHops = in_cntHops ;
}