// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// WRouteRules
// This object manages the rules that are used to
// filter rules.  Each nym potentially has different 
// rules.
// Rules can be used both based on user choice and
// automatically.  Cases where we use automatic rules include
// setting the exit funnel for unauthenticated routes
// and prolonging routes while maintinaing an existing exit funnel

#ifndef WROUTERULES_H_
#define WROUTERULES_H_

#include "ClientTypesInc.h"
#include "WRoute.h"
#include <clientstl.h>

#include <vector>

#define DEFAULT_NUMBER_OF_HOPS MAX_HOPS

class ClientTypes_API WRouteRules
{
public:
	WRouteRules() ;
	WRouteRules( size_t in_cntHops ) ;

	// Initialize based on the old v1.1 rule
	// format
	WRouteRules( const MapRules& in_rules ) ;

	void ClearRules() ;

	void SetRules( const MapRules& in_rules ) ;

	// Determines if a rule actually has been set
	// Having a number of hops set doesn't count
	bool AreRulesSet() const ;

	// Override the default number of hops
	void SetHopCount( size_t in_cntHops ) ;
	size_t GetHopCount() const { return m_cntHops ; }

	// Route rules is able to test a candidate route
	// to see if it passes the test
	// If the route fails the test the index of the failed hop
	// is returned.

	bool DoesRoutePassRules( const WRoute& in_Route, size_t& out_FailedIndex ) const ;

	// Rules can be configured through these methods

	void SetFirstHop( uint32_t in_New ) { m_First = in_New ; }
	void SetLastHop( uint32_t in_New ) { m_Last = in_New ; }

	void AddAvoidHop( uint32_t in_EntityToAvoid ) ;
	void ClearAvoidHops() ;		// Erase all hops that are to be avoided

	// Determine if the hop is not supposed to be avoided
	bool IsHopAvoided( uint32_t in_Hop ) const ;

	// Returns zero if no hop rule set
	uint32_t GetLastHop() const { return m_Last ; }
	uint32_t GetFirstHop() const { return m_First ; }	

	// CoreAips are a special case - certain routes
	// must exit with a core aip (hence the Entity type
	// is the deciding factor)
	// By default Last Hop is _not_ a core aip
	void SetCoreAipAsExitRule( bool in_bUseCoreAip ) { m_bUseCoreAipAsExit = in_bUseCoreAip ; }
	bool ShouldLastHopBeCoreAip() const { return m_bUseCoreAipAsExit ; }

private:

	// Not supported yet
	WRouteRules( const WRouteRules& ) ;
	WRouteRules& operator=( const WRouteRules& ) ;

	// Rules are stored as the entity id

	size_t   m_cntHops ;

	// Only a single exit funnel is supported currently
	uint32_t m_Last ;

	// In the case of a single hop route the 
	// last takes precendence over first
	uint32_t m_First ;

	ZKSTD::vector<uint32_t> m_AvoidHops ;

	bool m_bUseCoreAipAsExit ;
} ;

#endif