/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef ZK_WSYNCTYPES_H_INCLUDED
#define ZK_WSYNCTYPES_H_INCLUDED

#include <AutoInstCount.h>
#include <zkThread.h>
#include "ClientTypesInc.h"
#include <lockguard.h>
#include <DZkError.h>

/* Wrappers for libzktread synchronization objects
 * 02:06:00:ol <oliver@zks.net>
 */

#ifdef WIN32
#	pragma warning(disable: 4800)
#endif

class WZkMutex AUTO_INSTANCE_COUNTING1( WZkMutex )
{
private:
	ZkMutex m_mutex;

public:

	WZkMutex() 
	{
		if (!zkMutexNew(NULL, &m_mutex)) {
			ZKTRACE0("Mutex Allocation Failed. (There is something profoundly wrong!)");
			ZKASSERT(0);
		}
	};

	~WZkMutex()
	{
		zkMutexFree(NULL, &m_mutex);
	};

	void Lock( void ) {
		if (!zkMutexLock(NULL, m_mutex)) {
			ZKTRACE0("Mutex Lock Failed. Expect Problems");
			ZKASSERT(0);
		}
	};

	void Unlock( void ) {
		if (!zkMutexUnlock(NULL, m_mutex)) {
			ZKTRACE0("Mutex Unlock Failed. Expect Problems");
			ZKASSERT(0);
		}
	};

private:
	WZkMutex(const WZkMutex& c);// AUTO_INSTANCE_COPY_CTOR1(WZkMutex, c); // no copy
	WZkMutex& operator=(const WZkMutex&); // no assignment
};

class WZkSemaphore AUTO_INSTANCE_COUNTING1( WZkSemaphore )
{
private:
	ZkSemaphore m_sem;
public:

	WZkSemaphore(unsigned int value = 0) 
	{
		if (!zkSemaphoreNew(NULL, &m_sem, value)) {
			ZKTRACE0("Semaphore Allocation Failed.");
			ZKASSERT(0);
		}
	};

	~WZkSemaphore()
	{
		zkSemaphoreFree(NULL, &m_sem);
	};

	void Lock( void ) {
		if (!zkSemaphoreWait(NULL, m_sem)) {
			ZKTRACE0("Semaphore Lock Failed. Expect Problems");
			ZKASSERT(0);
		}
	};

	void Unlock( void ) {
		if (!zkSemaphorePost(NULL, m_sem)) {
			ZKTRACE0("Semaphore Unlock Failed. Expect Problems");
			ZKASSERT(0);
		}
	};

private:
	WZkSemaphore(const WZkSemaphore& c);// AUTO_INSTANCE_COPY_CTOR1(WZkSemaphore, c) {}; // no copy
	WZkSemaphore& operator=(const WZkSemaphore&); // no assignment
};

class WZkEvent AUTO_INSTANCE_COUNTING1( WZkEvent )
{
private:
	ZkEvent m_event;
public:

	WZkEvent(bool manual = ZK_FALSE)
	{
		if (!zkEventNew(NULL, &m_event, manual)) {
			ZKTRACE0("Event Allocation Failed.");
			ZKASSERT(0);
		}
	};

	~WZkEvent()
	{
		zkEventFree(NULL, &m_event);
	};

	bool Wait( unsigned long timeout = ZK_INFINITE) {
		int ret;
		if (!(ret = zkEventWait(NULL, m_event, timeout))) {
			ZKTRACE0("Event Lock/Wait Failed. Timeout?");
		}
		return ret;
	};

	void Lock() { Wait(); };

	void Pulse() {
		if (!zkEventPulse(NULL, m_event)) {
			ZKTRACE0("Event Pulse/Unlock Failed. Expect Problems");
			ZKASSERT(0);
		}
	};

	void Unlock( void ) { Pulse(); };

	void Set() {
		if (!zkEventSet(NULL, m_event)) {
			ZKTRACE0("Event Set Failed. Expect Problems");
			ZKASSERT(0);
		}
	};

	void Reset() {
		if (!zkEventReset(NULL, m_event)) {
			ZKTRACE0("Event Reset Failed. Expect Problems");
			ZKASSERT(0);
		}
	};

private:
	WZkEvent(const WZkEvent& c);// AUTO_INSTANCE_COPY_CTOR1(WZkEvent, c); // no copy
	WZkEvent& operator=(const WZkEvent&); // no assignment
};

//////////
// This class is based on the article:
//
//Compound Win32 Synchronization Objects
//Ruediger R. Asche
//Microsoft Developer Network Technology Group
//
//http://msdn.microsoft.com/library/techart/msdn_locktest.htm
// 
//
// --ol <oliver@zks.net>

#define ZK_WRITER 0
#define ZK_READER 1

class RRWLock  AUTO_INSTANCE_COUNTING1 (RRWLock)
{
public:
	RRWLock()
		: m_readerEvent(ZK_TRUE)
		, m_exclusionEvent(ZK_FALSE)
		, m_writerMutex()
		, m_countMutex()
		, m_count(-1)
	{
		// Needs to be initially set.
		// Note that other threads should not be allowed to access the object
		// until after it is constructed (which could actually be possible 
		// in a _really_ poorly implemented scenario where pointers to instances
		// of this class are being bandied about).
		m_exclusionEvent.Set();
	};

	~RRWLock() {}

	void Lock(bool which = ZK_READER)
	{
		if (which == ZK_WRITER)
		{
			m_writerMutex.Lock();
			m_exclusionEvent.Wait();
		}
		else
		{
			m_countMutex.Lock();
			if (++m_count == 0) 
			{
				m_countMutex.Unlock();
				m_exclusionEvent.Wait();
				m_readerEvent.Set();
			}
			else 
			{
				m_countMutex.Unlock();
			}
			m_readerEvent.Wait();
		}
	}

	void Unlock(bool which = ZK_READER)
	{
		if (which == ZK_WRITER)
		{
			m_exclusionEvent.Set();
			m_writerMutex.Unlock();
		}
		else
		{
			m_countMutex.Lock();
			if (--m_count < 0) 
			{
				m_countMutex.Unlock();
				m_readerEvent.Reset();
				m_exclusionEvent.Set();
			} 
			else 
			{
				m_countMutex.Unlock();
			}
		}
	}

private:
	RRWLock( RRWLock & ); // no copy
	RRWLock& operator=(const RRWLock&); // no assignment
	
	WZkEvent	m_readerEvent;
	WZkEvent	m_exclusionEvent;
	WZkMutex	m_writerMutex;
	WZkMutex	m_countMutex;

	int			m_count;
};

////////
// Andrew's idea for a simple MT testing class
// 
// --ol <oliver@zks.net>

class RThreadCollisionFinder AUTO_INSTANCE_COUNTING1 ( RThreadCollisionFinder )
{
public:
	RThreadCollisionFinder() : count(0) {}
	~RThreadCollisionFinder() {}

	void Lock() 
	{
		// Should probably use a mutex around this to be really safe.
		// XXX Test first
		if (++count > 1) {
			ZKTRACE1("THREAD COLLISION: %d threads are within a protected block!\n", count);
			ZKASSERT (count <= 1);
		}
	}

	void Unlock()
	{
		if (--count < 0) {
			ZKTRACE1("THREAD COLLISION number of locks != number unlocks. Difference is %d!\n", count);
			ZKASSERT (count >= 0);
		}
	}

private:
	int count;
};

#ifdef DEBUG

#	define COLLISION_FINDER_SCOPE(name) \
		ZKS::Lock_Guard<RThreadCollisionFinder> l_rtclf(name)

#	define COLLISION_FINDER_DECL(name) \
		RThreadCollisionFinder name

#	define COLLISION_FINDER_LOCK(name) \
		name.Lock()

#	define COLLISION_FINDER_UNLOCK(name) \
		name.Unlock()

#else // Release Build (no checks)

#	define COLLISION_FINDER_SCOPE(name) 

#	define COLLISION_FINDER_DECL(name) 

#	define COLLISION_FINDER_LOCK(name) 

#	define COLLISION_FINDER_UNLOCK(name) 

#endif // DEBUG

#endif //ZK_WSYNCTYPES_H_INCLUDED
