// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// ZKChunkWriter.cpp 

#include "StdAfx.h"
#include "ZKChunk.h"
#include "WReadBuffer.h"
#include "WAutoBuffer.h"

#include <zknbo.h>
#include <RZkException.h>
#include <ZkUnicode.h>
#include <TraceControl.h>

#include <vector>
#include <sstream>

//
// ZKChunkWriter Implementation
//
//
//

ZKChunkWriter::ZKChunkWriter( WMemBuf& in_Stream, uint32_t in_ChunkID, uint8_t in_Version )
	: m_posChunkStart( 0 )
	, m_Stream( in_Stream )
{
	ZKASSERT( in_ChunkID > CHUNK_BASE_ID) ;
	ZKASSERT( in_ChunkID < CHUNK_MAX_ID) ;
	ZKASSERT( in_Version > 0 ) ;

	// We take note of where the chunk starts
	ZKVERIFY( zkMemBufGetSize( &m_err, m_Stream, &m_posChunkStart ) ) ;

	// Write the chunk header
	// Format is strictly enforced

	uint32_t IdAndSize = in_ChunkID << 8 ;
	IdAndSize |= in_Version ;

	WriteUInt32( IdAndSize ) ;

	// Place Holder for the size - we jump back and write it in when we close the chunk
	WriteUInt32( 0 ) ;

	ZKCTRACE2( TRACE_PERSISTENCE,
			   "Creating Chunk - ID %d, Version %d\n", 
			   in_ChunkID - CHUNK_BASE_ID, 
			   in_Version ) ;
}

ZKChunkWriter::~ZKChunkWriter()
{
	// Time to close the chunk, which means we need to update the
	// size field

	// We don't use the "Write" methods because these
	// are designed to always append to the end of the buffer

	// Note: We rely on zkMemBuf giving us a pointer right into the internal
	// buffer and us being able to write to that memory

	unsigned char * pRawBuffer = NULL ;
	size_t cntBuffer = 0 ;

	ZKVERIFY( zkMemBufGetData( &m_err, m_Stream, (void**)&pRawBuffer, &cntBuffer ) );

	ZKASSERT( m_posChunkStart < cntBuffer ) ;

	size_t cntChunk = cntBuffer - m_posChunkStart ;

	// Size is stored at this exact position
	unsigned char * pSizePos = pRawBuffer + m_posChunkStart + CHUNK_ID_SIZE + CHUNK_VER_SIZE ;

	// Save size in Big Endian format
	
	zkNboFromHostInt32( (char*) pSizePos, cntChunk );
}

size_t ZKChunkWriter::GetChunkSize()
{
	size_t cntBuffer = 0 ;

	ZKVERIFY( zkMemBufGetSize( &m_err, m_Stream, &cntBuffer ) ) ;

	ZKASSERT( m_posChunkStart < cntBuffer ) ;

	return cntBuffer - m_posChunkStart;
}

void
ZKChunkWriter::WriteUInt64( uint64_t in_value )
{
	// Save explicitly in network byte order

	char nbo[8] ;
	zkNboFromHostUint64( nbo, in_value );
	
	ZKVERIFY( zkMemBufAppend( &m_err, m_Stream, nbo, sizeof( nbo ) ) ) ;
}

void
ZKChunkWriter::WriteInt64( int64_t in_value )
{
	// Save explicitly in network byte order

	char nbo[8] ;
	zkNboFromHostInt64( nbo, in_value );

	ZKVERIFY( zkMemBufAppend( &m_err, m_Stream, nbo, sizeof( nbo ) ) ) ;
}

void
ZKChunkWriter::WriteUInt32( uint32_t in_value )
{
	// Save explicitly in network byte order

	char nbo[4] ;
	zkNboFromHostUint32( nbo, in_value );
	
	ZKVERIFY( zkMemBufAppend( &m_err, m_Stream, nbo, sizeof( nbo ) ) ) ;
}

void
ZKChunkWriter::WriteInt32( int32_t in_value )
{
	// Save explicitly in network byte order

	char nbo[4] ;
	zkNboFromHostInt32( nbo, in_value );

	ZKVERIFY( zkMemBufAppend( &m_err, m_Stream, nbo, sizeof( nbo ) ) ) ;
}

void
ZKChunkWriter::WriteUInt16( uint16_t in_value )
{
	// Save explicitly in network byte order

	char nbo[2] ;
	zkNboFromHostUint16( nbo, in_value );
	
	ZKVERIFY( zkMemBufAppend( &m_err, m_Stream, nbo, sizeof( nbo ) ) ) ;
}

void
ZKChunkWriter::WriteInt16( int16_t in_value )
{
	// Save explicitly in network byte order

	char nbo[2] ;
	zkNboFromHostInt16( nbo, in_value );

	ZKVERIFY( zkMemBufAppend( &m_err, m_Stream, nbo, sizeof( nbo ) ) ) ;
}


void
ZKChunkWriter::WriteInt8( int8_t in_value )
{
	ZKVERIFY( zkMemBufAppend( &m_err, m_Stream, &in_value, sizeof( in_value ) ) ) ;
}

void
ZKChunkWriter::WriteUInt8( uint8_t in_value )
{
	ZKVERIFY( zkMemBufAppend( &m_err, m_Stream, &in_value, sizeof( in_value ) ) ) ;
}


void 
ZKChunkWriter::WriteArray( const void* in_pData, size_t in_cntData )
{
	ZKASSERT( in_pData != NULL ) ;
	ZKASSERT( in_cntData > 0 ) ;   // We can handle this but is it really desirable?

	// Remember the Size 
	WriteUInt32( in_cntData ) ;

	// Append the data
	if ( in_cntData > 0 ) 
	{
		ZKVERIFY( zkMemBufAppend( &m_err, m_Stream, in_pData, in_cntData ) ) ;
	}
}


void 
ZKChunkWriter::WriteNonUIString( const ZKSTD::string& in_str )
{
	// We save these strings just like binary data.  For simplicty
	// we include the termination character
	WriteArray( (uint8_t*) in_str.c_str(), in_str.size() + 1 ) ;
}

void 
ZKChunkWriter::WriteString( const ZKSTD::string& in_str )
{
	uint16_t * pUnicodeVersion = NULL;
	size_t     cntUnicodeVersion = 0 ;  // Size in Bytes

	ZkUnicode::StringToStorageUnicodeBuffer( in_str, &pUnicodeVersion, cntUnicodeVersion ) ;

	WriteUInt32( cntUnicodeVersion ) ;
	// Append the data
	if ( cntUnicodeVersion > 0 ) 
	{
		ZKVERIFY( zkMemBufAppend( &m_err, m_Stream, pUnicodeVersion, cntUnicodeVersion ) ) ;
	}

	SAFEFREE( pUnicodeVersion ) ;
}

//
// ZKChunkReader Implementation
//
//
//

ZKChunkReader::ZKChunkReader( WReadBuffer& in_Stream ) throw(RZkException)
	: m_posChunkStart( 0 )
	, m_Stream( in_Stream )
	, m_cntChunk( CHUNK_HEADER_SIZE )	// Important so that while reading the header we
										// don't think we are going past the chunk
{
	ZKASSERT( !m_Stream.IsEOF() ) ;
	m_posChunkStart = m_Stream.GetReadPos() ;

	// Read the Header
	uint32_t IdAndSize = ReadUInt32() ;

	m_ChunkID = ( IdAndSize >> 8 ) & 0x00FFFFFF ;

	m_ChunkVersion = ( IdAndSize & 0xFF ) ;

	m_cntChunk = ReadUInt32() ;

	ZKCTRACE3( TRACE_PERSISTENCE,
			   "Opened Chunk - ID %d, Version %d, Size %d\n", 
			   m_ChunkID - CHUNK_BASE_ID, 
			   m_ChunkVersion,
			   m_cntChunk ) ;

	// Look for Corrupt Stream

	if ( ( m_ChunkID < CHUNK_BASE_ID )||
		 ( m_ChunkVersion == 0 ) ||
		 ( m_cntChunk < CHUNK_HEADER_SIZE ) ||
		 ( m_cntChunk + m_posChunkStart > m_Stream.GetBufferSize() ) )
	{
		throw RZkException ("ZKChunkReader::ZKChunkReader", "Corrupt File", IDS_CORRUPT_BUFFER);
	}

}

ZKChunkReader::~ZKChunkReader()
{
	// Move past the end of the Chunk
	m_Stream.Seek( m_posChunkStart + m_cntChunk ) ;
}

uint64_t
ZKChunkReader::ReadUInt64() throw(RZkException)
{
	// Read value from stream (in Big Endian)
	uint64_t nboValue = 0 ;

	if ( GetCntRemainingBytesInChunk() < sizeof( uint64_t ) ||
		 !m_Stream.Read( &nboValue, sizeof( nboValue ) ) )
	{
		throw RZkException("ZKChunkReader::ZKChunkReader", 
							 "Corrupt File",
							 IDS_CORRUPT_BUFFER) ;		
	}

	return zkNboToHostUint64( (char *)&nboValue ) ;
}

int64_t
ZKChunkReader::ReadInt64() throw(RZkException)
{
	// Read value from stream (in Big Endian)
	int64_t nboValue = 0 ;
	if ( GetCntRemainingBytesInChunk() < sizeof( int64_t ) || 
		!m_Stream.Read( &nboValue, sizeof( nboValue ) ) )
	{
		throw RZkException("ZKChunkReader::ZKChunkReader", 
							 "Corrupt File",
							 IDS_CORRUPT_BUFFER) ;		
	}

	return zkNboToHostInt64( (char *)&nboValue ) ;
}


uint32_t
ZKChunkReader::ReadUInt32() throw(RZkException)
{
	// Read value from stream (in Big Endian)
	uint32_t nboValue = 0 ;

	if ( GetCntRemainingBytesInChunk() < sizeof( uint32_t ) || 
		!m_Stream.Read( &nboValue, sizeof( nboValue ) ) )
	{
		throw RZkException("ZKChunkReader::ZKChunkReader", 
							 "Corrupt File",
							 IDS_CORRUPT_BUFFER) ;		
	}

	return zkNboToHostUint32( (char *)&nboValue ) ;
}

int32_t
ZKChunkReader::ReadInt32() throw(RZkException)
{
	// Read value from stream (in Big Endian)
	int32_t nboValue = 0 ;
	if ( GetCntRemainingBytesInChunk() < sizeof( int32_t ) ||  
		!m_Stream.Read( &nboValue, sizeof( nboValue ) ) )
	{
		throw RZkException("ZKChunkReader::ZKChunkReader", 
							 "Corrupt File",
							 IDS_CORRUPT_BUFFER) ;		
	}

	return zkNboToHostInt32( (char *)&nboValue ) ;
}

uint16_t
ZKChunkReader::ReadUInt16() throw(RZkException)
{
	// Read value from stream (in Big Endian)
	uint16_t nboValue = 0 ;

	if ( GetCntRemainingBytesInChunk() < sizeof( uint16_t ) ||  
		!m_Stream.Read( &nboValue, sizeof( nboValue ) ) )
	{
		throw RZkException("ZKChunkReader::ZKChunkReader", 
							 "Corrupt File",
							 IDS_CORRUPT_BUFFER) ;		
	}

	return zkNboToHostUint16( (char *)&nboValue ) ;
}

int16_t
ZKChunkReader::ReadInt16() throw(RZkException)
{
	// Read value from stream (in Big Endian)
	int16_t nboValue = 0 ;
	if ( GetCntRemainingBytesInChunk() < sizeof( int16_t ) ||   
		!m_Stream.Read( &nboValue, sizeof( nboValue ) ) )
	{
		throw RZkException("ZKChunkReader::ZKChunkReader", 
							 "Corrupt File",
							 IDS_CORRUPT_BUFFER) ;		
	}

	return zkNboToHostInt16( (char *)&nboValue ) ;
}

int8_t 
ZKChunkReader::ReadInt8() throw(RZkException)
{
	int8_t val = 0 ;
	if ( IsEndOfChunk() ||
		 !m_Stream.Read( &val, sizeof( val ) ) )
	{
		throw RZkException("ZKChunkReader::ZKChunkReader", 
							 "Corrupt File",
							 IDS_CORRUPT_BUFFER) ;		
	}

	return val ;
}

uint8_t 
ZKChunkReader::ReadUInt8() throw(RZkException)
{
	uint8_t val = 0 ;

	if ( IsEndOfChunk() ||
		 !m_Stream.Read( &val, sizeof( val ) ) )
	{
		throw RZkException("ZKChunkReader::ZKChunkReader", 
							 "Corrupt File",
							 IDS_CORRUPT_BUFFER) ;		
	}

	return val ;
}


void 
ZKChunkReader::ReadArray( const void** out_pArray, size_t* out_pSize )
{
	// First the Array Size
	*out_pSize = ReadUInt32() ;

	// Look for corrupt buffer - array size should fit the stream
	if ( *out_pSize + m_Stream.GetReadPos() > m_Stream.GetBufferSize() )
	{
		throw RZkException("ZKChunkReader::ZKChunkReader", 
							 "Corrupt File",
								 IDS_CORRUPT_BUFFER) ;		
	}

	// Array should fit the chunk also
	if ( *out_pSize + m_Stream.GetReadPos() > m_posChunkStart + m_cntChunk )
	{
		throw RZkException("ZKChunkReader::ZKChunkReader", 
							 "Corrupt File",
								 IDS_CORRUPT_BUFFER) ;		
	}

	if ( *out_pSize == 0 )
	{
		*out_pArray = NULL ;
		return ;
	}

	// Point the caller to the correct place
	//*out_pArray = const_cast<uint8_t*>(m_Stream.GetBuffer() + m_Stream.GetReadPos() );

	*out_pArray = m_Stream.GetBuffer() + m_Stream.GetReadPos() ;

	// Move the read pointer past this position
	m_Stream.Seek( m_Stream.GetReadPos() + *out_pSize ) ;
}

void
ZKChunkReader::ReadString( ZKSTD::string& out_str ) throw(RZkException)
{
	// Attempt to convert from Wide characters to current code page,
	// might fail depending on user machine
	
	const uint8_t * pArray = NULL ;
	size_t          cntArray = 0 ;

	ReadArray( (const void**)&pArray, &cntArray ) ;
	ZkUnicode::StorageUnicodeBufferToString( (uint16_t*) pArray, cntArray, out_str ) ;
}

void 
ZKChunkReader::ReadNonUIString( ZKSTD::string& out_str ) throw(RZkException)
{

	const uint8_t * pArray = NULL ;
	size_t          cntArray = 0 ;

	ReadArray( (const void**)&pArray, &cntArray ) ;

	if (( pArray == NULL ) ||
		( cntArray == 0 ))
	{
		out_str.erase() ;
	}
	else
	{
		// We expect the string to be NULL terminated
		ZKASSERT( pArray[ cntArray - 1 ] == '\0' ) ;
		out_str = ( char* ) pArray ;
	}
}

bool        
ZKChunkReader::IsEndOfChunk()
{
	return ( GetCntRemainingBytesInChunk() == 0 );
}

uint32_t
ZKChunkReader::GetCntRemainingBytesInChunk()
{
	int32_t cntRemaining = ( m_posChunkStart + m_cntChunk - m_Stream.GetReadPos() ) ;

	if ( cntRemaining <= 0 )
	{
		return 0 ;
	}
	else 
	{
		return (uint32_t) cntRemaining ;
	}
}
