/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


/*
ZKChunk.h

This header provides the ability to read and write Binary data saved with a Chunk Structure.
It provides Cross Platform safe streaming by avoiding Endian encoding.
In the future it will provide platform safe user-readable string encoding using UNICODE.

A chunk system is used for easy versioning of persisted data at the component level.
Each "object" or other component can save its state into a seperate data chunk, then
re-initialize itself at a later data by reading the chunk.  By recognizing version numbers it
can continue to read old versions of the chunk in the future.

Chunks can be nested.  For example a nym chunk can contain multiple chunks.

Chunks are serialized our to a flat buffer.  When reading a chunk the code is responsible for
for reading the binary contents of the chunks and sending any nested chunks to the correct objects
to read.

Because of the overhead of the chunk header there is a trade off between size and flexibility
if too many chunks are used.

Our chunks have the following format:

Header
   Chunk ID [ 3 Bytes ]			- Binary Chunk value as defined in clientchunkids.h
   Chunk Version [ 1 Byte ]     - Start at 1, 
   Chunk Size [ 4 Bytes ]		- Size of the chunk, Including the Chunk Header
Binary Data

Note that this structure does not allow the "dumb" parsing of the tree structure of nested
chunks - it is not possible to distinguish child chunks from other content inside the
binary data.

Note - these classes do not provide Encryption or File I/O functionality.  All operations
occur on IN-MEMORY buffers which can be saved independently.

Please see the Unit Test Code for example Usage.

*/


#ifndef _ZKCHUNK_H
#define _ZKCHUNK_H

#include "ClientTypesInc.h"
#include "WCoreTypes.h"

#include <clientchunkids.h>
#include <DZkError.h>

#include <AutoInstCount.h>


class RZkException;

#define CHUNK_HEADER_SIZE  8

// Chunk header made up of the following ( in this
// order)
#define CHUNK_ID_SIZE	   3
#define CHUNK_VER_SIZE     1
#define CHUNK_SIZE_SIZE    4

class WReadBuffer ;


// ZKChunkWriter - use to stream data into a buffer with "chunk" structuring
// Life time of this object opens and then closes a chunk.  This means
// that the natural usage of "scope" will ensure that an object saves
// itself to disk.

class ClientTypes_API ZKChunkWriter AUTO_INSTANCE_COUNTING1( ZKChunkWriter )
{
public:
	// Opens the Chunk at the current position in the stream
	ZKChunkWriter( WMemBuf& in_Stream, uint32_t in_ChunkID, uint8_t in_Version ) ;

	// Closes the Chunk
	~ZKChunkWriter() ;

	// Find out how much data has been written so far.
	size_t GetChunkSize();

	// Call these methods to place data into the chunk
	// you are expected to remember the order that you place the items in
	// the stream, ie they must be read back in the same order.

	void WriteUInt64( uint64_t in_value ) ;
	void WriteInt64( int64_t in_value ) ;
	void WriteUInt32( uint32_t in_value ) ;
	void WriteInt32( int32_t in_value ) ;
	void WriteUInt16( uint16_t in_value ) ;
	void WriteInt16( int16_t in_value ) ;
	void WriteUInt8( uint8_t in_value ) ;
	void WriteInt8( int8_t in_value ) ;

	// Use for internal strings, for example binary data encoded
	// in strings.  Client is responsible to make sure that
	// contents of the array do not run into endian problems
	// on different platforms (for example this should not be used
	// to save an array of ints!)
	void WriteArray( const void* in_pData, size_t in_size ) ;
	
	// Use for potentially user readable string (important
	// so that string can be encoded in unicode)
	void WriteString( const ZKSTD::string& in_str ) ;

	void WriteNonUIString( const ZKSTD::string& in_str ) ;

private:

	size_t						m_posChunkStart ;
	WMemBuf&					m_Stream ;

	mutable DZkError			m_err;	
} ;


// ZKChunkReader - used to read data from a buffer that was saved with "chunk" structuring

class ClientTypes_API ZKChunkReader
{
public:

	// Open the Chunk and read its header
	// Will throw if the data is corrupt or we reach unexpected end of stream
	ZKChunkReader( WReadBuffer& in_Stream ) throw(RZkException);

	// Closes the Chunk Reader, which will move the 
	// stream position to point just past the end of this
	// chunk
	~ZKChunkReader() ;

	uint32_t    GetChunkID()      { return m_ChunkID ; }
	uint8_t		GetChunkVersion() { return m_ChunkVersion ; }	 // Use for version control
	uint32_t	GetChunkSize()	  { return m_cntChunk ; }		 // Normally not interesting for clients

	uint64_t	ReadUInt64( ) throw(RZkException);
	int64_t		ReadInt64( )  throw(RZkException);
	uint32_t	ReadUInt32( ) throw(RZkException);
	int32_t		ReadInt32( )  throw(RZkException);
	uint16_t	ReadUInt16( ) throw(RZkException);
	int16_t		ReadInt16( )  throw(RZkException);
	uint8_t		ReadUInt8( )  throw(RZkException);
	int8_t		ReadInt8( )   throw(RZkException);


	WReadBuffer& GetReadBuffer()  { return m_Stream ; }


	// Sets the pointer to the start of the array data 
	// and moves the read position past the array.  Does not
	// copy the data.
	void		ReadArray( const void** out_ppArray, size_t* out_pSize ) throw(RZkException);

	// Read a international-character set safe string from the stream
	void		ReadString( ZKSTD::string& out_str ) throw(RZkException);

	// Use this to save strings that are not shown to the user (eg
	// not subject to potential character set problems )
	void		ReadNonUIString( ZKSTD::string& out_str ) throw(RZkException);

	// Determine whether the Read Position has reached the end of the chunk.
	// Use this for while loops reading an array of contents in the chunk
	bool        IsEndOfChunk() ;

	// Determine if reading in_offset more bytes would take us
	// past the end of the chunk
	uint32_t    GetCntRemainingBytesInChunk() ;

	// Idea: PeekChunk() method would look at the current position and return
	// the chunk ID and size but NOT move the Read position.  This would
	// allow you to determine which send the chunk to the correct code to parse it.

private:

	size_t						m_posChunkStart ;
	WReadBuffer&				m_Stream ;
	
	// Chunk header information
	uint8_t                     m_ChunkVersion ;
	uint32_t					m_ChunkID ;
	uint32_t                    m_cntChunk ;

	mutable DZkError			m_err;	
} ;

#endif