// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "StdAfx.h"
#include "ZKDocument.h"
#include "WCoreTypes.h"
#include "WReadBuffer.h"
#include "ZKChunk.h"
#include <clientchunkids.h>
#include <RZkException.h>
#include <IZkFileHelp.h>
#include <TraceControl.h>

// See 5771
//#define VERIFY_SAVE_REALLY_WORKED

// (Version 1 didn't have network string)
#define ZKDOCUMENT_HEADER_CHUNK_VERSION 2

ZKDocumentWriter::ZKDocumentWriter()
	: m_pSalt( NULL )
	, m_cntSalt( 0 )
	, m_pTestBuffer( NULL )
	, m_cntTestBuffer( 0 )
	, m_pDataBuffer( NULL )
	, m_cntDataBuffer( 0 )
{

}

void 
ZKDocumentWriter::SetHeader( 
	const ZKSTD::string& in_strFileDescription, 
	const ZKSTD::string& in_strProductNameDesc, 
	const ZKSTD::string& in_strProductVersion, 
	const ZKSTD::string& in_strPlatform,
	const ZKSTD::string& in_strNetwork )
{
	m_strFileDesc = in_strFileDescription ;		
	m_strProductNameDesc = in_strProductNameDesc ;
	m_strProductVersion = in_strProductVersion ; 
	m_strPlatform = in_strPlatform ;
	m_strNetwork = in_strNetwork ;
}

void
ZKDocumentWriter::SetSalt( const uint8_t * in_pSalt, size_t in_cntSalt )
{
	m_pSalt = in_pSalt ;
	m_cntSalt = in_cntSalt ;
}


void
ZKDocumentWriter::SetTest( const uint8_t * in_pTest, size_t in_cntTest )
{
	m_pTestBuffer = in_pTest ;
	m_cntTestBuffer = in_cntTest ;
}

void
ZKDocumentWriter::SetData( const uint8_t * in_pData, size_t in_cntData )
{
	m_pDataBuffer = in_pData ;
	m_cntDataBuffer = in_cntData ;
}

void
ZKDocumentWriter::WriteChunksToBuffer( WMemBuf& io_Buf )
{
	ZKCTRACE0( TRACE_PERSISTENCE, "Saving A File ----- Header contents\n\n" ) ;
	ZKCTRACE1( TRACE_PERSISTENCE, "File Desc- %s\n", m_strFileDesc.c_str() ) ;
	ZKCTRACE1( TRACE_PERSISTENCE, "Product Name: %s\n", m_strProductNameDesc.c_str() ) ;
	ZKCTRACE1( TRACE_PERSISTENCE, "Product Version: %s\n", m_strProductVersion.c_str() ) ;
	ZKCTRACE1( TRACE_PERSISTENCE, "Platform: %s\n", m_strPlatform.c_str() ) ;
	ZKCTRACE1( TRACE_PERSISTENCE, "Network when File was saved- %s\n", m_strNetwork.c_str() ) ;	
	ZKCTRACE1( TRACE_PERSISTENCE, "Data Chunk Size %d\n", m_cntDataBuffer ) ;

	{
		ZKChunkWriter cw( io_Buf, ZKDOCUMENT_HEADER_CHUNK_ID, ZKDOCUMENT_HEADER_CHUNK_VERSION ) ;
		cw.WriteNonUIString( m_strFileDesc ) ;
		cw.WriteNonUIString( m_strProductNameDesc ) ;
		cw.WriteNonUIString( m_strProductVersion ) ;
		cw.WriteNonUIString( m_strPlatform ) ;
		cw.WriteNonUIString( m_strNetwork ) ;
	}

	if (( m_cntSalt > 0 ) && ( m_pSalt != NULL ))
	{
		ZKChunkWriter cw( io_Buf, SALT_CHUNK_ID, 1 ) ;
		cw.WriteArray( m_pSalt, m_cntSalt ) ;
	}

	if (( m_cntTestBuffer > 0 ) && ( m_pTestBuffer != NULL ))
	{
		ZKChunkWriter cw( io_Buf, PASSPHRASETEST_CHUNK_ID, 1 ) ;
		cw.WriteArray( m_pTestBuffer, m_cntTestBuffer ) ;
	}

	if (( m_cntDataBuffer > 0 ) && ( m_pDataBuffer != NULL ))
	{
		ZKChunkWriter cw( io_Buf, DATA_CHUNK_ID, 1 ) ;
		cw.WriteArray( m_pDataBuffer, m_cntDataBuffer ) ;
	}
}

void
ZKDocumentWriter::WriteToFile( const char* in_pszFileName, const char* in_pszBackupFile ) throw (RZkException)
{
	ZKFunctionEntry( "ZKDocumentWriter::WriteToFile" );

	// The Salt and Test buffer are optional, for example
	// in the case of plain test files, but the data is not
	ZKTestPointer( m_pDataBuffer ) ;

	WMemBuf mbuf ;

	WriteChunksToBuffer( mbuf ) ;

	unsigned char * pRawBuffer = NULL ;
	size_t cntBuffer = 0 ;

	DZkError err ;

	zkMemBufGetData( &err, mbuf, (void**)&pRawBuffer, &cntBuffer ) ;

	ZKTEST( pRawBuffer != NULL ) ; 
	ZKTEST( cntBuffer > 0 )  ; 

	if (( in_pszFileName == NULL ) ||
		( strlen( in_pszFileName ) == 0 ))
	{
		throw RZkException( "ZKDocumentWriter::WriteToFile", 
							 "Bogus File Name",
							 IDS_WRITE_FILENAME_INVALID ) ;
	}
	
	std::string strBackupFile ;
	
	if ( in_pszBackupFile != NULL )
	{
		strBackupFile = in_pszBackupFile ;
	}
	else
	{
		// Generate our own default backup file name
		strBackupFile = in_pszFileName ;
		strBackupFile += ".backup" ;
	}

	ZKASSERT( 0 != strcmp( strBackupFile.c_str(), in_pszFileName ) ) ;

	if(!IZkFileHelp::WriteBinaryFile( strBackupFile.c_str(), pRawBuffer, cntBuffer ) )
	{
		throw RZkException(	"ZKDocumentWriter::WriteToFile",
							"IZkFileHelp::WriteBinaryFile failed",
							IDS_FILE_WRITE_FAILED,
							strBackupFile.c_str() );	
	}

	// If the above hasn't thrown then we know we're now safely on disk,
	// this copy shouldn't fail
	if( !IZkFileHelp::FileCopyWithVerification( strBackupFile.c_str(), in_pszFileName, false ) )
	{
		throw RZkException( "ZKDocumentWriter::WriteToFile", 
							 "IZkFileHelp::FileCopyWithVerification failed",
							 IDS_WRITE_CANT_COPY,
							 strBackupFile.c_str(),
							 in_pszFileName) ;
	}


#ifdef VERIFY_SAVE_REALLY_WORKED

	// Try to detect Corruption happening on Linux
	// This should never fail, and is not normally a necessary check!
	// See 5771

	WAutoBuffer ActualFileContents ;

	if( !IZkFileHelp::ReadBinaryFile( in_pszFileName, &ActualFileContents, ActualFileContents.GetSizeRef() ) )
	{
		// Please collect the call stacks and submit

		ZKASSERT( false ) ;

		throw RZkException( "ZKDocumentWriter::WriteToFile", 
							 "Loading the file that we just saved failed",
							IDS_FILE_WRITE_FAILED,
							strBackupFile.c_str()) ;		
	}

	// (It is acceptable for the saved buffer to be slightly larger, we ignore any
	// white space added on the end
	
	if (( ActualFileContents.Size() != cntBuffer ) ||
	    ( 0 != memcmp( pRawBuffer, ActualFileContents, cntBuffer ) ) )
	{
		// Please collect the call stacks and submit

		ZKASSERT( false ) ;

		throw RZkException( "ZKDocumentWriter::WriteToFile", 
							 "Contents of the file that we just saved are corrupt!",
							IDS_FILE_WRITE_FAILED,
							strBackupFile.c_str()) ;		
	}
#endif

}


ZKDocumentReader::ZKDocumentReader( const char * in_pszFileName ) throw (RZkException)
	: m_pSalt( NULL )
	, m_cntSalt( 0 )
	, m_pTestBuffer( NULL )
	, m_cntTestBuffer( 0 )
	, m_pDataBuffer( NULL )
	, m_cntDataBuffer( 0 )
{	
	// Caller is expected to provide a valid file to read
	/*
	// See bug 

	if(! IZkFileHelp::FileExists( in_pszFileName ) )
	{
		throw RZkException("ZKDocumentReader::ZKDocumentReader()", 
							 "ZKDocumentReader called with invalid filename",
							 IDS_INVALID_ARG) ;
	}
	*/

	// This assumption is that we will want to write back to the file
	// any changes and that we check the file permissions immediately ( see 6180 )

	if(! IZkFileHelp::IsFileWritable( in_pszFileName ) )
	{
		throw RZkException("ZKDocumentReader::ZKDocumentReader()", 
							 "IZkFileHelp::IsFileWritable() returned false",
							 IDS_FREEDOM_DATA_PERMISSIONS ) ;
	}

	if(! IZkFileHelp::ReadBinaryFile( in_pszFileName, &m_bufFileContents, m_bufFileContents.GetSizeRef() ) )
	{
		// This can happen when the file has "write" permissions but not "read" permissions
		throw RZkException("ZKDocumentReader::ZKDocumentReader()", 
							 "IZkFileHelp::ReadBinaryFile() returned false",
							 IDS_CANT_OPEN_FREEDOM_DATA ) ;
	}


	WReadBuffer rb( m_bufFileContents, m_bufFileContents.Size() ) ;

	// This is an example of how to handle any possible order
	// of chunks, plus possible presence of future chunks

	while ( !rb.IsEOF() )
	{
		HandleNextChunk( rb ) ;
	}

	// Post Condition - We must have this

	if ( m_pDataBuffer == NULL )
	{
		// TODO - accurate error string
		throw RZkException("ZKChunkReader::ZKChunkReader", 
							 "Corrupt File",
							 IDS_CORRUPT_FREEDOM_DATA ) ;
	}
}

void 
ZKDocumentReader::HandleNextChunk( WReadBuffer& in_rb ) throw (RZkException)
{
	ZKChunkReader cr( in_rb ) ;

	switch( cr.GetChunkID() )
	{
		case( ZKDOCUMENT_HEADER_CHUNK_ID ) :
			ReadHeaderChunk( cr ) ;
			break ;
			
		case( SALT_CHUNK_ID ) :
			ReadSaltChunk( cr ) ;
			break ;

		case( PASSPHRASETEST_CHUNK_ID ) :
			ReadTestChunk( cr ) ;
			break ;

		case( DATA_CHUNK_ID ) :
			ReadDataChunk( cr ) ;
			break ;

		default :
			// Future chunks - just ignore them
			break ;
	}
}

void 
ZKDocumentReader::ReadHeaderChunk( ZKChunkReader& in_cr ) throw (RZkException)
{
	// Note that I don't check the version - this allows later versions to 
	// save more data afterwards and still be read by this code!

	ZKASSERT( ZKDOCUMENT_HEADER_CHUNK_ID == in_cr.GetChunkID() ) ;

	in_cr.ReadNonUIString( m_strFileDesc ) ;
	in_cr.ReadNonUIString( m_strProductNameDesc ) ;
	in_cr.ReadNonUIString( m_strProductVersion ) ;
	in_cr.ReadNonUIString( m_strPlatform ) ;

	if ( in_cr.GetChunkVersion() > 1 )
	{
		in_cr.ReadNonUIString( m_strNetwork ) ;
	}

	ZKCTRACE0( TRACE_PERSISTENCE, "Reading Header of File\n\n" ) ;
	ZKCTRACE1( TRACE_PERSISTENCE, "File Desc- %s\n", m_strFileDesc.c_str() ) ;
	ZKCTRACE1( TRACE_PERSISTENCE, "Product Name- %s\n", m_strProductNameDesc.c_str() ) ;
	ZKCTRACE1( TRACE_PERSISTENCE, "Product Version- %s\n", m_strProductVersion.c_str() ) ;
	ZKCTRACE1( TRACE_PERSISTENCE, "Platform- %s\n", m_strPlatform.c_str() ) ;
	ZKCTRACE1( TRACE_PERSISTENCE, "Network when File was saved- %s\n", m_strNetwork.c_str() ) ;
}

void 
ZKDocumentReader::ReadSaltChunk( ZKChunkReader& in_cr ) throw (RZkException)
{
	ZKASSERT( SALT_CHUNK_ID == in_cr.GetChunkID() ) ;
	in_cr.ReadArray( (const void**)&m_pSalt, &m_cntSalt ) ;
}
void 
ZKDocumentReader::ReadTestChunk( ZKChunkReader& in_cr ) throw (RZkException)
{
	ZKASSERT( PASSPHRASETEST_CHUNK_ID == in_cr.GetChunkID() ) ;
	in_cr.ReadArray( (const void**)&m_pTestBuffer, &m_cntTestBuffer ) ;
}

void 
ZKDocumentReader::ReadDataChunk( ZKChunkReader& in_cr ) throw (RZkException)
{
	ZKASSERT( DATA_CHUNK_ID == in_cr.GetChunkID() ) ;
	in_cr.ReadArray( (const void**)&m_pDataBuffer, &m_cntDataBuffer ) ;
}

void 
ZKDocumentReader::GetSalt( const uint8_t ** out_ppSalt, size_t & out_cntSalt )
{
	ZKASSERT( out_ppSalt != NULL ) ;
	ZKASSERT( m_pSalt != NULL ) ;
	ZKASSERT( m_cntSalt > 0 ) ;

	*out_ppSalt = m_pSalt ;
	out_cntSalt = m_cntSalt ;
}

void 
ZKDocumentReader::GetTestBuffer( const uint8_t ** out_ppTest, size_t & out_cntTest )
{
	ZKASSERT( out_ppTest != NULL ) ;
	ZKASSERT( m_pTestBuffer != NULL ) ;
	ZKASSERT( m_cntTestBuffer > 0 ) ;

	*out_ppTest = m_pTestBuffer ;
	out_cntTest = m_cntTestBuffer ;
}

void 
ZKDocumentReader::GetDataBuffer( const uint8_t ** out_ppData, size_t & out_cntData )
{
	ZKASSERT( out_ppData != NULL ) ;
	ZKASSERT( m_pDataBuffer != NULL ) ;
	ZKASSERT( m_cntDataBuffer > 0 ) ;

	*out_ppData = m_pDataBuffer ;
	out_cntData = m_cntDataBuffer ;
}
