// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):





///////////////////////////////////////////////////////////////////////////////
//
// ZKDocument provides the framework for reading and writing encrypted data 
// onto disk.

// Why not just dump the encrypted buffer into a file?

// Because we need a little file structure to ensure flexible reloading.

// This file format contains:
//    Chunk "HDR" - Clear-Text Header Identifying the file 
//    Chunk "SLT" - Clear-Text Salt Value (used when encrypting/decrypting)
//    Chunk "TST" - Chunk containing a small test buffer for fast verification of a passphrase
//					Because it contains a known string we can verify that the passphrase is valid
//    Chunk "DAT" - Data Chunk containing all the encrypted data

// Encryption is currently not provided by this class directly,
// (Review - the crypto sub-system is higher in the compilation order.  Some
// refactoring may fix this soon)

// This format can be enhanced via the chunk mechanism to store more data sections,
// for example multiple encrypted data sections.

// It is important that looking at this file provides no clues about the data inside, for example
// number of nyms, names of nyms etc can not be determined from the clear text (although
// currently file size can give some guesses)

// The SLT, and TST chunks are optional, in which case ZKDocument can be used for
// clear (non-encrypted) data, such as caches.

#ifndef _ZKDOCUMENT_H
#define _ZKDOCUMENT_H

#include "ClientTypesInc.h"
#include "WAutoBuffer.h"
#include "WCoreTypes.h"
#include <client.h>	// for ZKSTD

class RZkException;
class WReadBuffer ;
class ZKChunkReader ;

class ClientTypes_API ZKDocumentWriter AUTO_INSTANCE_COUNTING1( ZKDocumentWriter )
{
public:
	ZKDocumentWriter() ;

	void SetHeader( const ZKSTD::string& in_strFileDescription, 
					const ZKSTD::string& in_strProductNameDesc, 
					const ZKSTD::string& in_strProductVersion, 
					const ZKSTD::string& in_strPlatform,
					const ZKSTD::string& in_strNetwork ) ;

	// Optional data to be provided if the file is encrypted
	void SetSalt( const uint8_t * in_pSalt, size_t in_cntSalt ) ;
	void SetTest( const uint8_t * in_pTest, size_t in_cntTest ) ;

	// Data must be provided.  The Documentwriter does not make its
	// own copy of the data so the buffer must stay around for the life
	// time of this object.
	void SetData( const uint8_t * in_pData, size_t in_cntData ) ;


	// WriteToFile()
	// Any existing file is overwritten,

	// However if functions throws it is guaranteed that any existing file
	// was not damaged.

	// The file is saved first to a backup file, then once successful in_pszFileName
	// is overwritten.
	// The "in_pszBackupFile" argument allows you to override the file path that 
	// is used to save the intermediate file.  Otherwise in_pszFileName + ".backup" is used.
	// The backup file is left on disk as a form of auto "backup"

	void WriteToFile( const char * in_pszFileName, const char * in_pszBackupFile = NULL ) throw( RZkException ) ;

private:
	void WriteChunksToBuffer( WMemBuf& ) ;
	
	// Point to the provided data

	// Header Data
	ZKSTD::string m_strFileDesc ;		 // Describes what the file is for
	ZKSTD::string m_strProductNameDesc ; // Name of product saving file (helpful for debugging)
	ZKSTD::string m_strProductVersion ;  // Version number (preferably build) of product saving file
	ZKSTD::string m_strPlatform ;        // Platform Saving File (for informational purposes)
	ZKSTD::string m_strNetwork ;         // Network that data is valid for

	const uint8_t * m_pSalt ;
	size_t		    m_cntSalt ;

	const uint8_t * m_pTestBuffer ;
	size_t		    m_cntTestBuffer ;

	const uint8_t * m_pDataBuffer ;
	size_t		    m_cntDataBuffer ;

} ;


class ClientTypes_API ZKDocumentReader AUTO_INSTANCE_COUNTING1( ZKDocumentReader )
{
public:

	// File must be present and readable and correct format or else this will
	// vomit.  File is only open during the execution of the constructor,
	// then during the life time of the class the contents are held in memory
	ZKDocumentReader( const char * in_pszFileName ) throw(RZkException);

	ZKSTD::string GetFileDesc()		   { return m_strFileDesc ; }
	ZKSTD::string GetProductNameDesc() { return m_strProductNameDesc ; }
	ZKSTD::string GetProductVersion()  { return m_strProductVersion ; }
	ZKSTD::string GetPlatform()		   { return m_strPlatform ; }
	ZKSTD::string GetNetworkName()	   { return m_strNetwork ; }

	void GetSalt( const uint8_t ** out_ppSalt, size_t & out_cntSalt ) ;
	void GetTestBuffer( const uint8_t ** out_ppTest, size_t & out_cntTest ) ;
	void GetDataBuffer( const uint8_t ** out_ppData, size_t & out_cntData ) ;

private:

	void HandleNextChunk( WReadBuffer& in_rb ) throw(RZkException);
	void ReadHeaderChunk( ZKChunkReader& in_cr ) throw(RZkException);
	void ReadSaltChunk( ZKChunkReader& in_cr ) throw(RZkException);
	void ReadTestChunk( ZKChunkReader& in_cr ) throw(RZkException);
	void ReadDataChunk( ZKChunkReader& in_cr ) throw(RZkException);

	WAutoBuffer m_bufFileContents ;  // In memory copy of the file, 
								// Objects reading this should avoid duplication as much as 
								// possible


	// These are pointers into the File Contents buffer, marking
	// off the locations inside the file that these interesting sections
	// of memory are.

	const uint8_t * m_pSalt ;
	size_t		    m_cntSalt ;

	const uint8_t * m_pTestBuffer ;
	size_t		    m_cntTestBuffer ;

	const uint8_t * m_pDataBuffer ;
	size_t		    m_cntDataBuffer ;

	ZKSTD::string m_strFileDesc ;		 // Describes what the file is for
	ZKSTD::string m_strProductNameDesc ; // Name of product saving file (helpful for debugging)
	ZKSTD::string m_strProductVersion ;  // Version number (preferably build) of product saving file
	ZKSTD::string m_strPlatform ;        // Platform Saving File (for informational purposes)
	ZKSTD::string m_strNetwork ;         // Network that data is valid for
} ;

#endif