// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):



#include "StdAfx.h"

#include <math.h>
#include "ZKNymName.h"
#include <RZkException.h>

// Set class specific constants (see details/comments in header)
const int ZKNymName::MAX_ALT_NAMES = 100;
const int ZKNymName::MAX_NAME_LENGTH = 32;
const int ZKNymName::MAX_VALID_ALT_NAMES = 5;
const int ZKNymName::DEFAULT_ATTEMPT_VECTOR_SIZE = 20;
const int ZKNymName::MIN_NAMENUMBER_CEILING = 1;

ZKNymName::ZKNymName(ZKSTD::string& m_name) : m_status(ZKNNStatusUnknown),
	m_nymname(m_name)
{
	Init();
#ifdef _DEBUG
	testInvariant();
#endif
}

ZKNymName::~ZKNymName()
{

}

void ZKNymName::testInvariant()
{
	// We could do a check here to ensure that the name is syntactically valid
	// We could also check all the names in the various vectors
	if(m_nymname.empty())
	{
		throw RZkException("ZKNymName::testInvariant", "My nym name is invalid.", NULL ) ;
	}
}

void ZKNymName::Init()
{
	// Seed the random number generator (used to generate alternative names)
	srand((unsigned)time(NULL));
	// Create a list of names based on the requested name.
	GenerateAlternatives();
}

void ZKNymName::GenerateAlternatives (size_t numAlts)
{
	// Store the number we will add to the name
	char addNum[8];

	ZKSTD::string temp;

	// Used by the pig-latin conversion algorithm
	ZKSTD::string consonants("bcdfghjklmnpqrstvwxyzBCDFGHJKLMNPQRSTVWXYZ");

	// Pig latin conversion. If the first letter is a consonant, move it to the end.
	// Otherwise, put a 'y' on the end. Then, add 'ay'
	if(consonants.find(m_nymname[0]) != -1)
		temp = m_nymname.substr(1, m_nymname.size() - 1) +
			m_nymname[0] + "ay";
	else
		temp = m_nymname + "yay";

	AddNameToAltList(temp);
	
	// The meister algorithmmeister. Adds 'meister' to the end, and 'the_' to the
	// beginning, unless 'the' is already there.
	temp = m_nymname + "meister";
	if(temp.substr(0, 3) != "the")
	{
		temp = "the_" + temp;
	}

	AddNameToAltList(temp);

	// Add random numbers to the end of the name.
	// First, remove any numbers already there.
	temp = m_nymname;
	while(isdigit(*temp.end()))
	{
		temp.erase(temp.end());
	}

	// Fill up all the remaining space.
	while(m_vector_alternatives.size() < numAlts)
	{
		ZKSTD::string strNameAndNumber = temp;

#ifdef WIN32
		itoa(GetNumberForName(numAlts), addNum, 10);
#else
        // Could just use this version for all platforms.
        snprintf(addNum, 7, "%d", GetNumberForName(numAlts));
#endif

		if(strNameAndNumber.length() > 32 - strlen(addNum))
		{
			strNameAndNumber = strNameAndNumber.substr(0, 32 - strlen(addNum));
		}
		strNameAndNumber += addNum;

		AddNameToAltList(strNameAndNumber);
	}
	
	// Initialize our pointer. m_next_attempt always points to the next
	// name which will be put into an attempt vector and sent to the server.
	m_next_attempt = m_vector_alternatives.begin();
}

void ZKNymName::AddNameToAltList(ZKSTD::string& in_Name)
{
	if(in_Name.length() <= MAX_NAME_LENGTH)
	{
		m_vector_alternatives.push_back(in_Name);
	}
}
size_t ZKNymName::GetNumberForName(size_t numAlts)
{
	// Generates numbers of varying size, from 'MIN_NAMENUMBER_CEILING' digits
	// to 'MIN_NAMENUMBER_CEILING' + 3 digits
	return rand() % (size_t) (pow(10, MIN_NAMENUMBER_CEILING) * pow(10, (size_t) (m_vector_alternatives.size() * 4 / numAlts)));
}

ZKSTD::vector <ZKSTD::string>& ZKNymName::GetAttemptVector(ZKSTD::vector <ZKSTD::string>& v_attempt, size_t n_names)
{
	bool add;
	// We return an empty vector to indicate that we are done testing names
	// (either we went through all of m_vector_alternatives, or we have enough
	// alt names already). If someone calls with n_names == 0, then we will
	// return an empty vector. In order to avoid this ambiguous behavior, throw an
	// exception
	if(!n_names)
	{
		throw RZkException("ZKNymName::GetAttemptVector", "Called with n_names == 0", NULL );
	}

#ifdef _DEBUG
	testInvariant();
#endif

	// Clear the vector
	v_attempt.clear();

	if(m_vector_available.size() < MAX_VALID_ALT_NAMES)
	{
		while(v_attempt.size() < n_names && m_next_attempt != m_vector_alternatives.end())
		{
			add = true;
			// There may be duplicates in m_vector_alternatives
			// make sure we don't make duplicates in v_attempt
			for(size_t i = 0; i < v_attempt.size(); ++i)
			{
				if(v_attempt[i] == *m_next_attempt)
				{
					// It's already in there. Don't add it.
					add = false;
					break;
				}
			}
			if(add)
			{
				v_attempt.push_back(*m_next_attempt);
			}
			
			++m_next_attempt;
		}
	}
#ifdef _DEBUG
	testInvariant();
#endif
	return v_attempt;

}

void ZKNymName::ValidateNames(const ZKSTD::vector <ZKSTD::string>& v_names)
{
	bool add;
#ifdef _DEBUG
	testInvariant();
#endif

	for(size_t i = 0; i < m_vector_alternatives.size(); ++i)
	{
		for(size_t j = 0; j < v_names.size(); ++j)
		{
			if(m_vector_alternatives[i] == v_names[j])
			{
				add = true;
				// Check to make sure we haven't already validated
				// this name.
				for(size_t k = 0; k < m_vector_available.size(); ++k)
				{
					if(m_vector_available[k] == v_names[j])
					{
						// It's already in there. Don't add it.
						add = false;
						break;
					}
				}
				if(add)
				{
					m_vector_available.push_back(v_names[j]);
				}
				break;
			}
		}
	}

#ifdef _DEBUG
	testInvariant();
#endif
}
