// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// ZKNymName
//
// Declares a class used to represent the user's requested nym name
// and, in the case that the name is already taken, a list of alternative 
// suggestions.

#if !defined(ZKNYMNAME_Included_)
#define ZKNYMNAME_Included_

#include "ClientTypesInc.h"
#include <vector>
#include <string>
//#include <ErrorHandling/RZkException.h>
#include <client.h>

#include <AutoInstCount.h>

class RZkException;


enum ZKNymNameStatus {ZKNNStatusUnknown,
						ZKNNStatusNameAvailable,
						ZKNNStatusNameUnavailable};

class ClientTypes_API ZKNymName AUTO_INSTANCE_COUNTING1( ZKNymName )
{
private:

	////////////////////////////////////////////////////////////////
	// PURPOSE:	Prevent anyone from calling assignment operator,
	//			copy constructor, or default constructor.
	ZKNymName& operator= (const ZKNymName&);
	ZKNymName();
	ZKNymName(ZKNymName&);

	////////////////////////////////////////////////////////////////
	// PURPOSE:	Test the class invariants, and throw an exception if
	//			necessary
	//
	//			Right now, the primary invariant is that the nymname
	//			should be valid. The current testInvariant()
	//			simply ensures that it is of non-zero length.
	void testInvariant() throw(RZkException);
	
	////////////////////////////////////////////////////////////////
	// PURPOSE:	Number of alternative names to generate
	static const int MAX_ALT_NAMES;
	
	////////////////////////////////////////////////////////////////
	// PURPOSE:	Maximum Name Length
	static const int MAX_NAME_LENGTH;
	
	////////////////////////////////////////////////////////////////
	// PURPOSE:	Maximum number of alternative names we will show to
	//			user.
	static const int MAX_VALID_ALT_NAMES;

	////////////////////////////////////////////////////////////////
	// PURPOSE:	Default number of names to pass to server
	static const int DEFAULT_ATTEMPT_VECTOR_SIZE;

	////////////////////////////////////////////////////////////////
	// PURPOSE:	The initial number of digits we will add to a user's
	//			name (we will initially request names with small
	//			numbers, then bigger ones, in order to assure that
	//			we will find names)
	static const int MIN_NAMENUMBER_CEILING;

	////////////////////////////////////////////////////////////////
	// PURPOSE:	The user's requested name
	// REQUIRE:	The object is constructed with a syntactically
	//			correct name.
	// PROMISE: m_nymname represents the user's originally requested
	//			name.
	const ZKSTD::string m_nymname;
	
	////////////////////////////////////////////////////////////////
	// PURPOSE:	Stores a list of possible alternative names based on
	//			the user's requested name. These names have not yet
	//			been validated.
	// REQUIRE: m_nymname's requirements are met.
	// PROMISE: Contains a list of syntactically correct possible
	//			alternatives (may be zero-length) that have not yet
	//			been validated. They are ordered from most desirable
	//			names to least desirable.
	ZKSTD::vector <ZKSTD::string> m_vector_alternatives;
	
	////////////////////////////////////////////////////////////////
	// PURPOSE:	Points to the next string to put in an attempt
	//			vector.
	// REQUIRE: GenerateAlternatives() has been called.
	// PROMISE: Will point to an alternative string that has not yet
	//			been returned by GetAttemptVector(), or to
	//			m_vector_alternatives().end()
	ZKSTD::vector<ZKSTD::string>::const_iterator  m_next_attempt;

	////////////////////////////////////////////////////////////////
	// PURPOSE:	Stores a list of available alternative names based
	//			on the user's requested name. These names have been
	//			validated by the server.
	// REQUIRE: m_nymname's requirements are met. The requirements
	//			of ValidateNames() are always satisfied.
	// PROMISE: Contains a list of syntactically correct possible
	//			alternatives (may be zero-length) that have been
	//			validated. They are ordered from most desirable to
	//			least desirable.
	ZKSTD::vector <ZKSTD::string> m_vector_available;

	////////////////////////////////////////////////////////////////
	// PURPOSE: Indicate if the name is available, unavailable, or
	//			if we don't know yet.
	// REQUIRE: SetStatus() is called when the nym creation
	//			completes (or fails).
	// PROMISE: Will indicate the status of the nym name.
	ZKNymNameStatus m_status;

	////////////////////////////////////////////////////////////////
	// PURPOSE: General initialization. Seeds random number
	//			generator, and generates a list of alternative
	//			names.
	void Init();
	
	////////////////////////////////////////////////////////////////
	// PURPOSE:	Generate a list of alternative names and store them
	//			in m_vector_alternatives. First-choice alternatives
	//			should go in the front of the vector, followed by
	//			less desired names in order of desirability.
	// REQUIRE: Nothing.
	// PROMISE: m_vector_alternatives will contain up to numAlts
	//			syntactically correct alternative names.
	void GenerateAlternatives(size_t /* numAlts */ = MAX_ALT_NAMES);

	////////////////////////////////////////////////////////////////
	// PURPOSE: Generate a number to be added to a new nym name
	//			based on the total number of names we are generating
	//			(numAlts) and the number of names generated so far
	//			(from the size of m_vector_alternatives).
	// REQUIRE:	numAlts is the same int passed to
	//			GenerateAlternatives()
	// PROMISE:	The numbers generated will gradually increase in
	//			length as more names are generated.
	size_t GetNumberForName(size_t /* numAlts */);

	////////////////////////////////////////////////////////////////
	// PURPOSE: Adds a name to the alternative list, after checking
	//			its length.
	// REQUIRE:	nothing.
	// PROMISE:	The name will be added if it is less than
	//			MAX_NAME_LENGTH, or silently thrown out.
	void AddNameToAltList(ZKSTD::string& in_Name);

public:

	////////////////////////////////////////////////////////////////
	// PURPOSE:	Initialize with the user's requested name
	// REQUIRE:	The string is a syntactically valid nym name
	// PROMISE:	The object will be initialized with the name.
	ZKNymName(ZKSTD::string& /* m_name */ );

	~ZKNymName();
	////////////////////////////////////////////////////////////////
	// PURPOSE:	Return a constant reference to the user's requested
	//			name.
	// REQUIRE:	Nothing.
	// PROMISE:	Will return a constant reference to the user's
	//			requested name.
	const ZKSTD::string& GetRequestedName() const { return m_nymname; }

	////////////////////////////////////////////////////////////////
	// PURPOSE:	Return a reference to the vector of validated names.
	// REQUIRE:	Nothing.
	// PROMISE:	The vector contains names that have already been
	//			validated by the nym server.
	ZKSTD::vector <ZKSTD::string>& GetVectorOfValidNames() { return m_vector_available; }

	////////////////////////////////////////////////////////////////
	// PURPOSE:	Generate a vector of alternative names for the class
	//			user to validate with the server. Generates an empty
	//			vector when no more names should be validated. (If
	//			we have already run through our whole list, or if
	//			enough valid names have been found.)
	//			If v_attempt already has elements, we will clear
	//			them.
	// REQUIRE:	The vector reference is valid, and n_names > 0.
	// PROMISE:	The vector v_attempt will contain N names
	//			(0 <= N <= n_names). The return value is a reference
	//			to v_attempt.
	ZKSTD::vector <ZKSTD::string>&  GetAttemptVector(ZKSTD::vector <ZKSTD::string>& /* v_attempt */, size_t /* n_names */ = DEFAULT_ATTEMPT_VECTOR_SIZE) throw(RZkException);

	////////////////////////////////////////////////////////////////
	// PURPOSE:	Add a vector of alternative names to our list of
	//			validated names.
	// REQUIRE:	v_names is valid, the names in it were originally
	//			obtained from GetAttemptVector(), and they were all
	//			returned as 'available' by the server.
	// PROMISE:	The names will be added to our list of validated
	//			names.
	void ValidateNames(const ZKSTD::vector <ZKSTD::string>& /* v_names */ ) throw(RZkException);

	////////////////////////////////////////////////////////////////
	// PURPOSE:	Give access to the status field.
	// REQUIRE:	Nothing.
	// PROMISE:	Returns a reference to the status field.
	ZKNymNameStatus& Status() { return m_status; }
};

#endif