// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// DHKeyExchange.cpp: implementation of the DHKeyExchange class.
//
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"

#include <zks.h>
#include "DHKeyExchange.h"
#include <ZkErrorHandler.h>
#include <RZkException.h>

#include <IKeyQry.h>
#include <INiqs.h>
#include <ZkService.h>
#include <WDifHel.h>
#include <frNetworkDefaults.h>
#include <TraceControl.h>

DHKeyExchange::DHKeyExchange()
{

}

void 
DHKeyExchange::Exchange(
	unsigned int in_nRouteAci, 
	const sockaddr_in& in_firstAipAddr,
	const FrLocalAipCtx& ctx) throw(RZkException)
{
	// Will get key from key cache
	if(!frLAReadFirstHopPubSigKey(&m_err, ctx, in_nRouteAci))
	{
		throw RZkException(m_err, "DHKeyExchange::Exchange()",
					        "DLibs::LocalAIP_frLocalAipReadFirstHopPubSigKey() failed",
				            IDS_BUILDROUTE_FIRSTHOPSIG);
	}
	
	////////////////////// Make a TCP/IP connection to the first AIP on FIRST_AIP_TCP_PORT ////////////////
	// Always completes immediately, default used of blocking mode

//	SOCKET connectTCP = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	ZkSocket connectTCP;
	ZkSocketDescriptor connectTCPfd;
// XXX Doesn't use m_err for socket errors ... should I?
	DZkError err;
		
	if (!zkSocketNew(&err, AF_INET, SOCK_STREAM, IPPROTO_TCP, &connectTCP)
		|| !zkSocketGetFD(&err, connectTCP, &connectTCPfd)) 
	{
		throw RZkException("DHKeyExchange::Exchange()",
					        "socket() failed",
				            IDS_BUILDROUTE_SOCKET);
	}

	//////// Must explicitly state the local tcp port number, must be the same the as udp
	//////// local port number sending to the first hop for D-H key exchange to work
	SOCKADDR_IN saServer = in_firstAipAddr;
	
	// XXX this is a bit of a kludge AciInfo should probably start storing 
	// ZkSocketAddresses
	ZkSocketAddress zksaServer;
	saServer.sin_family  = AF_INET;	
	saServer.sin_port    = in_firstAipAddr.sin_port + htons(1);
	zksaServer.m_inet.sa = saServer;

	WDifHel				dh;
	FrAipsDifHelState	state = START;
	ZKVERIFY(frAipsDifHelInitState(&m_err, dh, RECV));

	fr_crypt_pubkey *	sigKey = NULL;
	if(!frAipsDifHelSetParms(&m_err, 
								dh, 
								sigKey, 
								FRCRYPT_SYMALG_BLOWFISH, 
								connectTCPfd, 
								WINDOWS_TCP_AND_UDP_PORT, // Actually local udp port in this case. we don't bind the tcp port
								60/*seconds*/, 
								saServer, 
								0/*current key version*/,
								ZkService::GetIKeyQry()->GetKeyQuerySocket(), //It will only do frKeyQryFindKey() [sendto()] using that socket
								saServer,				// Of the remote aip
								ZkService::GetIKeyQry()->GetKeyCache()
								))
		throw RZkException(m_err, "DHKeyExchange::Exchange()",
					        "frAipsDifHelSetParms() failed",
				            IDS_BUILDROUTE_SETPARMS);
	

	/*
		If at this stage you're on a modem, your dial-up dialog will appear.
	*/
	if (!zkSocketConnectAddr(&err, connectTCP, zksaServer)) {
		zkSocketFree(&err, &connectTCP);

#ifdef WIN32
		ZKCTRACE1( TRACE_ROUTEMANAGER, "DHKeyExchange::Exchange() connect() failed [%d]\n", WSAGetLastError());
#endif		
		throw RZkException("DHKeyExchange::Exchange()",
					        "connect() failed",
				            IDS_BUILDROUTE_CONNECT);		
	}
	////////////////////// End of Make a TCP/IP connection to the first AIP //////////////////////////////////////

	int timeout = 0;
	do{
		ZKVERIFY(frAipsDifHelTimeout(&m_err, dh, &timeout));
		if(!frAipsDifHelExchange(&m_err, dh, &state)){			

			zkSocketFree(&err, &connectTCP);

			char Buffer[100];
			unsigned long nSize = sizeof Buffer;
			snprintf( Buffer, nSize, "%s:%hu", inet_ntoa( saServer.sin_addr ), ntohs(saServer.sin_port) ) ;

			throw RZkException( m_err, 
								"DHKeyExchange::Exchange()",
					            "frAipsDifHelExchange() failed",
								IDS_BUILDROUTE_FIRSTHOP_DH,
								Buffer);

		}
		if(KEYLOOKUP == state)
		{ // should be first and last time we enter this state
			ZKCTRACE0( TRACE_ROUTEMANAGER, "Diffie-Hellman, gonna look for signature for first aip\n");
			fr_crypt_pubkey *	outkey = NULL;
			FrEntIdExp	entity;
			int			keyVer = 0;
			if(!frAipsDifHelGetPubKeyParms(&m_err, dh, &entity, &keyVer))
			{
				throw RZkException(m_err, "DHKeyExchange::Exchange()",
									"frAipsDifHelGetPubKeyParms() failed",
									IDS_BUILDROUTE_DIFHELGETPUBKEYPARMS);
			}

			if(!ZkService::GetIKeyQry()->GetKey(
				m_err,
				FRENT_TYPE_AIP, 
				3, // timeout
				entity.name, 
				keyVer, 
				FRCRYPT_KEYTYPE_SIGNING, 
				&outkey))
			{
					throw RZkException(m_err, 
									  "DHKeyExchange::Exchange()",
									  "GetKey() failed",
									  IDS_BUILDROUTE_AIPSIGNINGKEY,
									  entity.name );
			}
		}
	}while(!timeout && (DONE != state));

	if(DONE != state)
	{
		zkSocketFree(&err, &connectTCP);
		throw RZkException(m_err, "DHKeyExchange::Exchange()",
							"DONE != state",
							IDS_BUILDROUTE_DIFHELGETPUBKEYPARMS);
	}

	uint8_t	lekv;
	ZKVERIFY(frAipsDifHelGetLekv(&m_err, dh, &lekv));

	fr_crypt_sym *	symKey = NULL;
	if (!frLocalAipGetSymKey(&m_err, ctx, in_nRouteAci, lekv, &symKey))
	{
		ZKASSERT( false ) ;
		zkSocketFree(&err, &connectTCP);
		throw RZkException(m_err, "DHKeyExchange::Exchange()",
							"frLocalAipGetSymKey() failed.  Probably a threading problem"
							"has deleted the route underneath us",
							IDS_BUILDROUTE_FAIL);
	}

	SOCKADDR_IN	from;
	FrEntIdExp	entity;

	if ( !frAipsDifHelGetKey( &m_err, dh, &from, &entity, symKey ))
	{
		ZKASSERT( false ) ;
		zkSocketFree(&err, &connectTCP);
		throw RZkException(m_err, "DHKeyExchange::Exchange()",
							"frAipsDifHelGetKey() failed.  Probably a threading problem"
							"has deleted the route underneath us",
							IDS_BUILDROUTE_FAIL);
	}

	zkSocketFree(&err, &connectTCP);
}
