// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

////////////////////////////////////////////////////////////////////////////////
// IClientCentral.h

// Client Central represents the global settings
// of the client, such items which are not related
// to nyms but which are persisted.

// If the API for accessing one particular
// part of ClientCentral becomes large then
// it should be split off to its own interface,
// see ITokenService as an example

// XPlatform Note: Some settings are stored in the
// registry through a seperate ZKRegistry class.
////////////////////////////////////////////////////////////////////////////////

#ifndef _IClientCentral_H
#define _IClientCentral_H

////////////////////////////////////////////////////////////////////////////////
// Includes - (please keep to a minimum)
////////////////////////////////////////////////////////////////////////////////

#include <vector>
#include <set>
#include <ErrorHandling.h>
#include <WCoreTypes.h>
#include <client.h>	// for ZKSTD

////////////////////////////////////////////////////////////////////////////////
// Forward Declarations
////////////////////////////////////////////////////////////////////////////////

class ZKNymName ;
class WReadBuffer ;
class RZkException;

////////////////////////////////////////////////////////////////////////////////
// Enumerations
////////////////////////////////////////////////////////////////////////////////

// Client Preference Bit Fields
//
// All preferences for the client behavior that can be described with a bit are
// represented here.  
//
// Default state is always the "0" state for the field.
//
// Preferences are managed by NymCentral and persist from session to session
//
// To allow more than 32 possible preferences we use an enum of preference sets
//
// In the interest of backwards compatibility it is better to add new preferences
// than change the meaning of obsolete preferences.

enum ZK_PREFERENCE_SETS {
	ZK_TOOLBAR_PREF_SET     = 0,		// How the toolbar looks, how it reacts with the explorer
	ZK_SHOW_AGAIN_PREF_SET  = 1,		// Dialogs that have a "Don't show again" can remember this here
	ZK_CLIENT_PREF_SET		= 2,		// Log console into file (great for Beta period)
	ZK_FIREWALL_PREF_SET    = 3,        // Preferences for the behavior of the Firewall
	ZK_NUM_PREFERENCE_SETS  = 4
};

enum ZK_TOOLBAR_PREFS {
	PREF_NOT_ALWAYS_ON_TOP     = 1			// Whether freedom should float on top
};

enum ZK_SHOW_AGAIN_PREFS {
	PREF_OUTGOING_MAIL_ID		= 1,		// ??
	PREF_NO_SHOW_NEWS_OPTIONS	= 1 << 2,	// Whether to show the news reading warning again
	PREF_READNEWS_PRIV			= 1 << 3,	// store the users choice of news reading mechanism
	PREF_NO_SHOW_OUTGOING_MAIL  = 1 << 4,   // Don't show outgoing mail information dialog
	PREF_NO_SHOW_OUTGOING_NEWS  = 1 << 5,   // Don't show outgoing news information dialog

	PREF_NO_WARN_SIGNED_MAIL    = 1 << 6,	// Don't warn the user about outgoing mail
											// signed by the MUA
	PREF_NO_INFORM_DRAG_AND_DROP= 1 << 7,	// Don't inform the user how to pop up the Wallet
											// drag & drop dialog
	PREF_NO_SHOW_LOGIN_NORMALLY	= 1 << 8,	// Don't display the Wallet login normally dialog
	PREF_NO_SHOW_MAIL_DELETION_WARNING = 1 << 9, // Don't display a warning that Freedom doesn't support
												 // "Leave Mail on server"
	PREF_NO_SHOW_SSL_WARNING	= 1 << 10,	// Don't display the SSL warning again.
	PREF_NO_SHOW_MAIL_SELECT_NYM= 1 << 11,	// "...Please select a nym capable of receiving Freedom mail and 
											// try again.."
	PREF_NO_WARN_MAIL_CORRELATION=1 << 12,	// Explains the risk of correlation
											// when checking mail for all nyms.
    PREF_NO_SHOW_NO_MAIL_NYMS   = 1 << 13,  // You do not have any nyms capable of receiving Freedom mail..."
    PREF_NO_SHOW_NFS_WARNING    = 1 << 14,  // ...A directory mounted using NFS will not be anonymized."
	PREF_ALWAYS_FILL_IN_CC_INFO = 1 << 15,  // ...Your credit card information reveals your identity, even if you 
											// have a nym selected
	PREF_NO_SHOW_MAIL_SENT_AS_ACTIVE_NYM = 1 << 16	// This message will be sent as your active nym, because you 
													// selected "active nym only"... 
};

enum ZK_CLIENT_PREFS {
	PREF_LOG_CONSOLE				= 1,		// Send console output to freedomconsole.log Bug815
	PREF_SCAN_NYMNAMES				= 1 << 2,	// Whether to scan for nym names
	PREF_SCAN_IP					= 1 << 3,   // Whether to scan for the current IP addres
	PREF_SCAN_EMAIL_ADDR			= 1 << 4,	// Whether to scan for the provided email address
	PREF_NO_BLOCK_ADS				= 1 << 5,	// Disable the ad blocking feature
	PREF_NO_USER_COOKIE_MANAGEMENT	= 1 << 6,
	PREF_SHOW_LOGIN_INFO_FOR_ALL_NYMS= 1 << 7,	// In the Edit Login info page of Nym Preferences,
												// show login info for all nyms (1), or just the
												// nym of the current page (0).
	PREF_CHECK_MAIL_FOR_ACTIVE_NYM_ONLY= 1 << 8
};

enum ZK_FIREWALL_PREFS {
	PREF_BLOCK_NETBIOS				= 1,		    // Whether to block Netbios traffic
	PREF_BLOCK_RPC					= 1 << 2,       // Whether to block RPC traffic	
	PREF_BLOCK_DHCP					= 1 << 3,       // Whether to block DHCP traffic
	PREF_BLOCK_ICMP_ECHO			= 1 << 4,	    // Whether to block attempts to ping machine
	PREF_DISABLE_FIREWALL           = 1 << 5,       // Turn all firewall functionality off except
													// that required to ensure privacy features
	PREF_BLOCK_IDENT				= 1 << 6,		// Freedom will block ident request.
	PREF_NO_SHOW_BLOCKED_INCOMING	= 1 << 11		// Dialog for Incoming Traffic that we have blocked
};

#define ZK_CLIENT_NEVER_RE_ASK_PASSWORD -1

////////////////////////////////////////////////////////////////////////////////
// Interface
////////////////////////////////////////////////////////////////////////////////

class IClientCentral
{
public:		

	// Access to persisted preference flags.
	// Use to define the client behavior or other
	// data

	// Because this information is stored in Freedom persistence
	// instead of registry these preferences will ROAM
	// so no machine specific preference should be saved here

	// To allow more than 32 possible preferences there
	// are multiple sets of preferences (see ZK_PREFERENCE_SETS)
	virtual void SetPreference(
					unsigned long in_dwPrefSet, 
					unsigned long in_dwProperty, 
					bool in_bNewState) PURE ;

	virtual bool GetPreference(
					unsigned long in_dwPrefSet, 
					unsigned long in_dwProperty) const PURE ;

	// Access the Email Address of the User's own pop box,
	// it is a global setting, not associated with any particular
	// nym
	virtual void SetReturnEmailAddress( const ZKSTD::string& ) throw(RZkException) PURE ;
	virtual void GetReturnEmailAddress( ZKSTD::string& ) const PURE ;

    // These are used on Linux - see bug 5827
    virtual void    SetMailRelay( const ZKSTD::string& ) PURE;
    virtual void    GetMailRelay( ZKSTD::string& ) const PURE;

	// Save State into buffer
	// REQUIRE - Value Memory Buffer to stream into
	// PROMISE - Memory buffer will contain a chunk with the persisted data
	// such that the state can be retrieved by calling load.

	virtual void Save( WMemBuf& io_Buffer ) throw(RZkException) PURE ;


	// Load state from buffer
	// REQUIRE - Read Buffer pointing to the beginning of the data written by
	// this object in the past.
	// PROMISE - State will be loaded and Read buffer will have advanced to the
	// end of the data owned by Client Persist
	virtual void Load( WReadBuffer& io_Buffer ) throw(RZkException) PURE ;

	// These methods are replaced 
	// interface to avoid cluttering this interface


	// Provide a buffer to initialize the state of this object
	virtual bool InitializeFromBuffer_Obsolete( 
					const unsigned char* in_pBuffer, 
					size_t in_cntBuffer) PURE ;

	// Provide the preference settings as a buffer for this object
	virtual bool InitializePreferencesFromBuffer_Obsolete( 
					const unsigned char* in_pBuffer, 
					size_t in_cntBuffer) PURE ;

	// This retrieves a buffer with all the persistence except for
	// the preference bits wrapped up in a single buffer.
	// Note: This buffer contains the Tokens as well
	virtual bool GetBuffer_Obsolete( 
					unsigned char** out_ppBuffer, 
					size_t* cntBuffer ) const PURE ;

	virtual bool GetPreferenceBuffer_Obsolete( 
					unsigned char** out_ppBuffer, 
					size_t* cntBuffer ) const PURE ;

	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: To find out if the wallet is disabled for a domain.
	//
	// REQUIRE: Nothing.
	//
	// PROMISE:	Returns true if the wallet is disabled for this domain, and false otherwise.

	virtual bool IsWalletDisabledForDomain(
					const ZKSTD::string& in_strDomain) const PURE ;

	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: To mark a domain as enabled (if !in_bDisable), or disabled (if in_bDisable).
	//
	// REQUIRE: Nothing.
	//
	// PROMISE:	A subsequest call to IsWalletDisabled() for this domain will return in_bDisable.

	virtual void SetIsWalletDisabledForDomain(
					const ZKSTD::string& in_strDomain, 
					bool in_bDisable) PURE ;

	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: To set all the disabled domains in one fowl swoop. 
	//			if (in_strSetOfDomains.size() == 0), this is equivalent to enabling the wallet for
	//			all domains.
	//
	// REQUIRE: Nothing.
	//
	// PROMISE:	The strings in the set can be any urls (including 
	//			ZK_LOCAL_COMPUTER).  The domain will be stripped out of the url and lower-cased
	//			before being stored.
	//			For example, passing in "http://www.freedom.net:8080/", "MY.FREEDOM>NET", and 
	//			"freedom.net" are all equivalent.

	virtual void SetDomainsForWhichWalletIsDisabled(
					const ZKSTD::set<ZKSTD::string>& in_strSetOfDomains) PURE ;


	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: To get all the disabled domains in one fowl swoop. 
	//			if (out_strSetOfDomains.size() == 0), the wallet is enabled for all sites.
	//
	// REQUIRE: Nothing.
	//
	// PROMISE:	Any strings in out_strSetOfDomains will be cleared out before inserting the
	//			disabled domains.

	virtual void GetDomainsForWhichWalletIsDisabled(
					ZKSTD::set<ZKSTD::string>& out_strSetOfDomains) const PURE ;


	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: To discover if Freedom should show "sensitive" information to the user.  If 
	//			GetNumMinutesUntilReAskPassword() minutes have passed since the last call to
	//			this method (or since startup), false is returned.  Otherwise, true is returned
	//			and the time-until-this-method-will-return-false is extended to time now +
	//			GetNumMinutesUntilReAskPassword().
	//
	// REQUIRE: Nothing.
	//
	// PROMISE:	- If true is returned, future calls to IsPasswordRequired() between now and time 
	//			now + GetNumMinutesUntilReAskPassword() will return true.
	//			- If false is returned, all future calls to IsPasswordRequired() will return
	//			false until CorrectPasswordEntered() is called.

	virtual bool IsPasswordRequired() PURE ;

	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: To inform this class that the user entered the correct password.  This resets the
	//			time-until-this-method-will-return-false to time now + 
	//			GetNumMinutesUntilReAskPassword().
	//
	// REQUIRE: Nothing.
	//
	// PROMISE:	Future calls to IsPasswordRequired() between now and time now + 
	//			GetNumMinutesUntilReAskPassword() will return true.

	virtual void CorrectPasswordEntered() PURE ;


	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: To allow the UI to discover how many minutes must pass between calls to 
	//			either CorrectPasswordEntered() or IsPasswordRequired() before IsPasswordRequired()
	//			returns false.
	//
	// REQUIRE: Nothing.
	//
	// PROMISE:	Valid return values are all non-negative ints and ZK_CLIENT_NEVER_RE_ASK_PASSWORD.

	virtual int  GetNumMinutesUntilReAskPassword() PURE ;

	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: To set the number of minutes that must pass between calls to 
	//			either CorrectPasswordEntered() or IsPasswordRequired() before IsPasswordRequired()
	//			returns false.
	//
	// REQUIRE: in_nMinutes must be either a non-negative int or ZK_CLIENT_NEVER_RE_ASK_PASSWORD;.
	//
	// PROMISE:	- If (in_nMinutes == ZK_CLIENT_NEVER_RE_ASK_PASSWORD), IsPasswordRequired() will
	//			always return true.
	//			- If (in_nMinutes == 0), IsPasswordRequired() will always return false.

	virtual void SetNumMinutesUntilReAskPassword(const int in_nMinutes) PURE ;
} ;

#endif
