// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom(R) code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#ifndef ICLIENTINFO_H
#define ICLIENTINFO_H

// The ClientInfo is an infamous grab bag of configuration information
// that is provided to the client from the network.

// Over time this information may be split up into more sensible
// groupings and probably will come from different places on the network

#include <client.h> // Just for ZKSTD

class WClientInfo ;

class IClientInfo
{
public:

	////////////////////////////////////////////////////////////////////////////
	// Methods to control Client Info Service
	////////////////////////////////////////////////////////////////////////////

	// Use this function to set the active ClientInfo.
	// For example if we download a new ClientInfo from the
	// server then this function should be called with the 
	// new information.
	virtual void			SetClientInfo( const WClientInfo& ) PURE ;
	
	
	// This function determines whether we got a client info from the 
	// NIQS.  It indicates whether it is safe to call the other functions
	virtual bool			ClientInfoIsValid() PURE;
	

	////////////////////////////////////////////////////////////////////////////
	// Methods to Query Client Info Service
	////////////////////////////////////////////////////////////////////////////

	// This is based on the clients network configuration.
	// For example "beta.freedom.net"
	virtual void			GetFreedomDomain(	ZKSTD::string& out_strFreedomDomain ) PURE;


	// This is the time stamp returned from the NIQS.  It is
	// used to detect client machines whose clock is way off.
	// Obviously this value can only be considered valid and interesting
	// immediately after the Client Info has been downloaded, 
	// and should be considered accurate down the the exact second.
	// Its use in the client should be restricted to calculations
	// that will work even when off by a minute or so.
	virtual time_t			GetCurrentUTCFromClientInfo() PURE;

	// The NIQS tells us how often we should update
	// A default value is returned even if no contact was made with NIQS
	// so it is always safe to call this
	virtual int				GetNiqsUpdateFrequency() PURE;

	// The client provides a string describing its version and
	// the client info returns a delta for how long is left before
	// expiry.  If zero (or negative?) the client is considered expired,
	// otherwise this is considered for cases of warning the user in
	// advance of the expiry.
	virtual time_t			GetTimeToClientExpiry() PURE;

	// The message of the day is an optional string describing info
	// of interest to the whole network, for example notification of upcoming 
	// network downtime
	virtual ZKSTD::string	GetMessageOfTheDay() PURE;

	// WARNING: Not sure the semantics of this value, we only show MOTD if it is value
	// "1"
	virtual int				GetMessageOfTheDayUrgency() PURE;

	// This is the location of the Upgrade package.
	virtual ZKSTD::string	GetUpgradeUrl() PURE;

	// Client Info (and potentially other mechanisms such as the registry) can
	// allow us to hide the Free version
	virtual bool			IsFreeVersionAvailable() const PURE ;

	// Client Info (and potentially other mechanisms such as the registry) 
	// allow us to turn on the use of the next protocol use for Routing.
	// REVIEW - this API can be removed once we are completely migrated to
	// Aip 2.0
	virtual bool			IsAip2ProtocolEnabled() const PURE ;
} ;

#endif