// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// ICrpto Interface declaration

// This interface offers encryption services
// to the client.

// In particular it holds the main encryption symmetric key
// which was generated from the user's password

// It does _not_ give access to a Nym's or other entities
// keys.

#include <frNym.h>
#include <fr_entid.h>
#include <WCypherMailMsg.h>

// forward declarations
class DZkError;
class RZkException;
class DPassphraseError ;

// All passphrases should be this minimal length
#define	MIN_PASSPHRASE	 8


class ICrypto  
{
public:
	
	// Data stored on the client is protected by
	// encryption with a symmetric key based on 
	// a Passphrase and Salt

	// Initialize the State of the Crypto Service with
	// existing Passphrase and Salt.  This must be 
	// called before attempting to decrypt any existing buffers
	virtual void SetNewPassphraseAndSalt(
					IN const ZKSTD::string& strpass, 
					IN void* salt, 
					IN size_t saltsiz) throw(RZkException, DPassphraseError) PURE;
	
	// Call this the first time Freedom is run with the strOld empty.
	// by old we mean used previously during this session, we don't store the password
	virtual void ChangePassphrase( 
					const ZKSTD::string& strOld, 
					const ZKSTD::string& strNew) throw(RZkException) PURE ;

	// Get a clear text copy of the passphrase
	virtual bool GetPassphrase(ZKSTD::string& strPass) PURE ;

	// Get a clear text copy of the passphrase
	virtual void GetDefaultPassphrase(ZKSTD::string& strPass) PURE ;

	// Returns a copy of the salt buffer
	// Caller is responsible for freeing the provided 
	// buffer
	virtual bool GetSalt( void ** out_ppSalt, size_t& out_szSalt ) PURE ;
	
	virtual bool SymDecryptBuf(		
					const void* in_pSrc, 
					const size_t in_cntSrc, 
					void** out_ppDst, 
					size_t& out_cntDst ) PURE;

	virtual bool SymEncryptBuf(
					const void* in_pSrc, 
					const size_t in_cntDst, 
					void** out_ppDst, 
					size_t& out_cntDst) PURE;

	// Initialize the key cache based on the input file
	// TODO - I would prefer that knowledge of what file to read is
	// managed inside this interface and this this method is
	// is renamed to Init()
	virtual void KeyCacheFromFile(const ZKSTD::string& strFile) PURE;

	// Save the 
	virtual void GetKeyCacheIntoFile(const ZKSTD::string& strFile) PURE;

	// Given a File Path this function will attempt to decrypt it using the
	// current user password (and salt) and provide allocate the decrypted connected
	// into a buffer.  Caller is responsible to deallocate the buffer
	virtual bool ZkDecryptFile(
					const ZKSTD::string& in_strFile, 
					void*& out_pClearBuf, 
					size_t& out_szClear) PURE;

	// Method name is a bit misleading.  
	// This encrypt a buffer with the users passphrase and saves the results into a file
	virtual bool ZkEncryptFile(
					const ZKSTD::string& in_strFileToWrite, 
					void* in_buf, 
					size_t in_len) PURE;

	// CheckAndRemoveSignature
	// Given a signed buffer this function attempts to verify the signature.
	// Currently it does not fail if it can't verify the signature.
	//
	// If successful the signature is stripped from the buffer by an inplace
	// change, this adjusts the size of the buffer.
	// 
	// This function should only throw if there was an unexpected 
	// corrupt buffer or otherwise invalid input
	//
	// Note: This code currently will output information to the console.
	// This feedback should be more controlled by the called,
	// perhaps by returning more specific result information.

	virtual SigCheckResult CheckAndRemoveSignature( 
					unsigned char * io_pBuffer,
					size_t&         io_szBuffer,
					ZKSTD::string&    out_SigningEntity,
					FrEntIdType&	out_Entity) const throw(RZkException) PURE ;

	// Checks a signature but does not remove it (Buffer sent in is not altered)
	virtual SigCheckResult CheckSignature( 
					const unsigned char * in_pBuffer,
					size_t				  in_szBuffer,
					ZKSTD::string&		  out_SigningEntity,
					FrEntIdType&		  out_Entity) const throw(RZkException) PURE ;

	// Used to create the pub/private key pair for a nym
	// Note: implementated could move into Nym object directly - from
	// looking at the implementation there is not a lot of abstraction achieved
	// by this arangement
	virtual void NewCryptKey(
		const int eKeyVersion,
		const int eKeySize,
		const ZkClock crypt_expiration,	
		FrCryptPubKey& CryptPubKey,
		fr_entid_exp* ent) const throw(RZkException) PURE;

	virtual void NewSigKey(
		const int sKeyVersion,
		const int sKeySize,
		const ZkClock sign_expiration,	
		FrCryptPubKey& SigPubKey,
		fr_entid_exp* ent) const throw(RZkException) PURE;

	virtual int PutKeyInFile(
			DZkError& out_Err, 
			FrCryptPubKey key, 
			FILE *fp) PURE;

	virtual int SymmetricKeyToBuffer(
			DZkError& out_Err, 
			const FrCryptSym in_sym, 
			char* out_buf, 
			const size_t in_bs) PURE;
};
