// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// IFirewall

// This interface provides the logic for handling
// connections and packet traffic leaving the machine

// Potentially all such events are responded to in the following
// way:
// -Block them
// -Allow them out un-hindered
// -Reroute them.  In the case of TCP socket connection this means to reroute to a
//                 localhost proxy
//                 In the case of packets this means to send the packet as encrypted
//				   data through the Freedom Cloud

// The behavior of the firewall is partly configured by various preferences,
// UI and machine conditions.  The "business logic" of the firewall should find itself here,
// where as the platform specific code that carries out a firewall decision exists elsewhere.

#ifndef _IFirewall_H
#define _IFirewall_H

#ifdef ZK_LINUX
#	include <clientlinux.h>
#endif // ZK_LINUX

#include <clientenum.h>	// For FirewallAction
#include <zkIntTypes.h>
#include <ErrorHandling.h>
#include <WCoreTypes.h>			//WMemBuf.h
#include <client.h>
#include <vector>
#include <deque>
#include <ZkSubject.h>
#include <ZkHint.h>

struct	ZkNetFrame;
struct  ZkBuffer;
class	IPacketNotification;
class	PacketHelper;
class   WReadBuffer ;
class   RZkException ;
class   RPortToProxyMap ;

class IFirewall : public ZkSubject
{
public:

	// Initialize RFireWall - to be called once at startup
	virtual bool 
	Init() = 0 ;

	// Control the Firewall so that it can go into a completely 
	// passive mode (always allowing packets out)
	virtual void 
	EnableFirewall(bool in_bEnable) = 0 ;

	// PURPOSE: Having a firewall is fun when it tells you what it is doing
	//			This tells the firewall who to notify when interesting things happen
	virtual void 
	SendPacketNotificationsTo( IPacketNotification* ) = 0;

	// PURPOSE: Tell firewall when 
	//			a accept() completed successfully,
	//			a bind() completed successfully,
	//			prior to a connect() ion
	//			prior to a sendto()
	virtual unsigned int 
	GetAppFirewallOptions(const connection_info&) = 0;

	// PURPOSE: The firewall holds a temporary state concerning pending connections
	//			The purpose of this method is to ACK or NACK the connection
	//			so it can update its state
	virtual void
	PostConnectionResults(const postconnection_info&) = 0;

	// PURPOSE: Tell firewall when a sendto has completed
	virtual void
	PostSendToResults(const postsendto_info&) = 0;

	// PURPOSE: Tell the firewall about existing endpoints
	//			so it can remove some of its opened local ports
	//			Trapping closesocket is insufficient for this
	//			since closesocket won't block for the entire
	//			teardown sequence (that's a tcp/ip stack bug
	//			since it does not follow the Winsock standard).
	//			The good thing is that this allows us to cleanly handle apps which
	//			may of crashed, and would of neveer ended up calling
	//			closesocket anyways.
	virtual void
	VerifyLocalTcpPortsState(const ZKSTD::vector<endpoint>& in_v) = 0;

	// PURPOSE: Tell the firewall about existing endpoints
	//			so it can remove some of its opened local ports
	//			Trapping closesocket is insufficient for this
	//			since closesocket won't block for the entire
	//			teardown sequence (that's a tcp/ip stack bug
	//			since it does not follow the Winsock standard).
	//			The good thing is that this allows us to cleanly handle apps which
	//			may of crashed, and would of neveer ended up calling
	//			closesocket anyways.
	virtual void
	VerifyLocalUdpPortsState(const ZKSTD::vector<endpoint>& in_v) = 0;


	// We use various ports for our own communications, plus support standard
	// protocols through well known ports like 80.  For a give port number
	// this determines if the port fits that description.

	virtual bool 
	IsSupportedPort( uint16_t in_port ) = 0 ;

	// PURPOSE: Return true if the given port is one our proxies listen on.
	virtual bool 
	IsProxyPort(unsigned short) = 0;

	// PURPOSE: Return true if data for that port (	port is in host byte order )
	//			will be using the cloud at this time. This takes into consideration
	//			the type of nym selected, and checks both UDP and TCP type traffic.
	virtual bool 
	IsCloudData(const unsigned short in_port) const = 0;

	// PURPOSE: Verify that we have processed at least 1 outgoing packet 
	// This could indicate an incorrect configuration that might
	// compromise the users privacy
	virtual bool IsDriverSendingUpPackets() = 0 ;

	// return a copy of the log of inbound traffic blocked
	virtual void

	GetPacketsBackLog(ZKSTD::deque<PacketHelper>& out_log) = 0;

	// Convenience methods for accessing firewall
	// properties
	// For example PREF_BLOCK_NETBIOS
	virtual void 
	SetFirewallPreference(unsigned long in_dwProperty, bool in_bNewState) = 0 ;

	virtual bool 
	GetFirewallPreference(unsigned long in_dwProperty) const = 0;

	virtual void
	GetTcpPassiveStatePorts(ZKSTD::vector<remotehost>& out_v) = 0;

	virtual void
	GetTcpOpenedPorts(ZKSTD::vector<remotehost>& out_v) = 0;

	virtual void
	GetUdpOpenedPorts(ZKSTD::vector<remotehost>& out_v) = 0;

	// Test Code (for access from Unit Tests)
	virtual void 
	TestsWithNymActive() = 0 ;

	virtual void 
	TestsWithUserNymActive() = 0;

	// PURPOSE: Used in obtaining the entire persisted application settings
	virtual void
	GetApplicationSettings(ZKSTD::vector<FirewallAppPersist>& out_apps) = 0;

	virtual bool
	GetAppOptions(const char in_app[MAX_PATH], FirewallAppPersist& out_persisted_opts) = 0;	


	// PURPOSE : Allow UI or other force to change the rules for
	// an APP.
	// Overwrites any existing rules for the app or
	// creates a new entry if it does not exist yet.
	// Caller doesn't care.

	virtual void
	SetFirewallRulesForApplication( const FirewallAppPersist& in_Rule ) = 0 ;

	// PURPOSE : Allow UI or other force to remove an APP
	// from the rules
	virtual void
	EraseFirewallRulesForApplication( const char * in_pAppToErase ) = 0 ;

	// Returns copy of the rules for which extra
	// ports are supported
	virtual void
	GetExtraProxyPorts( RPortToProxyMap& out_Ports ) = 0 ;

	// Overwrite the rules for extra proxy ports 
	virtual void
	SetExtraProxyPorts( const RPortToProxyMap& in_Ports ) = 0 ;

	virtual void 
	Save( WMemBuf& io_Buffer ) throw(RZkException) = 0 ;

	virtual void 
	Load( WReadBuffer& io_Buffer ) throw(RZkException) = 0 ;

	virtual struct FirewallEvent_UsersChoice
	GetSpecificAppSettingsOut(const char in_AppName[MAX_PATH]) = 0;

	virtual struct FirewallEvent_UsersChoice
	GetSpecificAppSettingsIn(const char in_AppName[MAX_PATH]) = 0;

	virtual void
	SetSpecificAppSettingsOut(
		const char in_AppName[MAX_PATH], 
		const struct FirewallEvent_UsersChoice& in_uc,
		const bool in_bSpecialWarning) = 0;

	virtual void
	SetSpecificAppSettingsIn(const char in_AppName[MAX_PATH], const struct FirewallEvent_UsersChoice& in_uc) = 0;

	virtual void 
	FlushFirewallRules() const = 0;


	// Hint class returned by Firewall went events happen
    class FirewallHint 
        : public ZkHint
    {
    public:

        // Add other events which cause a change in route manager's state here
        enum {
            FIREWALL_APP_RULES_CHANGED,			// Signal that our app rules have changed
			FIREWALL_INCOMING_BLOCK_LOG_CHANGE  // Signal that the log of incoming packets has changed
        } ;
        
        FirewallHint( int stateChange )
            : m_StateChange( stateChange )
        { } ;

        ~FirewallHint() {} ;

        int GetStateChange() { return m_StateChange ; } ;
    private:
        int m_StateChange;
    } ;

} ;

#endif
