// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):


// INiqs Interface declaration

// This interface offers access to the NIQS service

// The Freedom NIQS is the main database server for information about the topology
// and "entities" inside the freedom network.  By "populating" or "updating" the
// NIQS you can get access to the latest status of the network.
// Client can then ask this service for the resulting information.

// NIQS Client Info is a particularly useful structure of data that gets sent
// to each connecting client.  It contains information about the expiry of the network,
// Message of the Day etc.


#ifndef _INIQS_H
#define _INIQS_H

#if defined(WIN32)

#elif defined(ZK_LINUX)
#	include <clientlinux.h>
#endif

#include <frNetInfo.h>
#include <WNetInfoDatum.h>

#define NYM_SERVER_TIMEOUT   90
#define TOKEN_SERVER_TIMEOUT 90

// forward declarations
class RZkException;
struct T_FrTokenType ;

class INiqs
{
public:

#if 0
	virtual bool		DownLoadMinimumKeys() PURE;
#endif

	////////////////////////////////////////////////////////////////
	// PURPOSE:	Gather information necessary to connect to the Freedom network
	//					
	// REQUIRE: Network connection or reasonably accurate NIQS cache on disk.
	//			
	// PROMISE: We will have information about all five core freedom servers
	//          We will have reasonably accurate MAIP and AIP information

	virtual void        Init(	const bool		in_bStartFromScratch, 
								const char *	in_strclient_ver ) throw(RZkException) PURE ;

	////////////////////////////////////////////////////////////////
	// PURPOSE:	Get an update of NIQS information			
	//				
	// REQUIRE: Network connection, NIQS server must be up
	//						
	// PROMISE: NIQS information will be updated to the latest information
	//			available from NIQS server

	virtual void		UpdateNiqsInformation() throw(RZkException) PURE ;


	// PURPOSE:	Retrieve a server of a specified type
	//				
	// REQUIRE: Network must have at least one of these servers
	//						
	// PROMISE: You will continue to get the same server each time you call this until
	//			RoundRobinEntity() is called on this entity type

	virtual void		GetServer( 
							const FrEntIdType& in_entType, 
							WNetInfoDatum& out_Server ) const throw(RZkException) PURE ;

	// PURPOSE:	Retrieve all server of a specified type
	//				
	// REQUIRE: Network must have at least one of these servers
	//						
	// PROMISE: To throw if no servers are found (this can happen if the Niqs 
	//			cannot be contacted).
	//			Just to be painfully clear, this function will never return an empty
	//			out_nList.  It would throw instead.

	virtual void		GetServersOfType(   
							const FrEntIdType& in_type,
							WNetInfoDatumVec& out_nlist, 
							bool bIncludeDownServers) const throw(RZkException) PURE;


	// PURPOSE:	Provide access to the list of Aips (both non-core aips and core aips)
	//				
	// REQUIRE: At least one of in_bIncludeNonCoreAips and in_bIncludeCoreAips must be true
	//			for the call to make sense.  Network must have Aips installed.
	//			You can optionally include "down" servers in the list.  These are servers
	//			the the NOC includes in the network which are marked as "down", possibly
	//			because they are not working properly or are being upgraded.
	//						
	// PROMISE: The most up to date list.  A copy of the data is returned, which
	//			means its contents will remain static.

	virtual void		GetListOfAips(   
							bool in_bIncludeNonCoreAips,
							bool in_bIncludeCoreAips,
							bool in_bIncludeDownServers,
							WNetInfoDatumVec& out_nlist ) const throw(RZkException) PURE;

	// Convenient filling in of the structures that are actually used to communicate with 
	// these common entities

	virtual void		GetNmtaAddr(SOCKADDR_IN *in_SA) const throw(RZkException) PURE;
	virtual void		GetPopServerAddr(SOCKADDR_IN *in_SA) const throw(RZkException) PURE;

	//TODO - to a similar helper for FrKeyQryHost
	//Calls frNetInfoEntIdentFromIPandPort
	virtual uint32_t	GetEntId(const int32_t in_IP, const uint16_t in_Port) throw() PURE;


	// PURPOSE:	Retrieve all information about an entity based on its ID
	//			all entities have unique IDs so this works without requiring
	//			the entity type specified.
	//				
	// REQUIRE: 
	//						
	// PROMISE: Current function returns false if entity is not found, in 
	//			future this should be changed to a throw.

	virtual bool		GetEntityById(const uint32_t, WNetInfoDatum&) throw() PURE;

	// This is a dangerous back door, releasing our internal structure to
	// the outside world, its usage should be avoided
	virtual FrNetInfo	GetNetInfo() PURE;

	// For debugging and unit test code - when turned on all connections to
	// our servers will fail (but internet connection is still active)
	virtual void        SetNetworkFailureDebuggingFlag( bool in_bFailConnections ) PURE ;

	// PURPOSE : Permits MIPAS Support, which is the ability to 
	//			 have multiple core servers of a particular type.
	//			 Calling this function _might_ cause subsequent 
	//			 calls to GetServer(), GetTokenServer()
	//			 to return a different server
	//
	// REQUIRES: Entropy for the randomness to work
	//
	virtual void			RoundRobinEntity(const FrEntIdType in_Entity) throw() PURE;



} ;


#endif