// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// INym.h

// This is the main interface for accessing the
// nym and its attributes.  A little coaxing and
// you might pull out a kitchen sink too.


#ifndef _INYM_H
#define _INYM_H

#include <guids.h> // All callers potentially need access to
					  // these guids in order to call methods like
					  // GetNym

#include <frNym.h>
#include <frNymClient.h>
#include <ErrorHandling.h>
#include <DZkError.h>
#include <WRemoteNym.h>
#include <ZKNymName.h>
#include <WCoreTypes.h>

class WReadBuffer ;
class RZkException ;
class WNymIconInfo ;

class INym
{
public:
	// Query Interface is COM style method of getting a different interface
	// from the INym pointer.  We don't follow the rules for HRESULT return 
	// value, instead we throw if the interface isn't supported.
	// We also don't use real GUIDs for the interface ids.
	// We also do not support "Addref'd" access to Nym objects - the
	// lifetime of NymCentral controls the lifetime of Nym objects
	// PROMISE: Set out_ppv to point to an object which is implementing the
	// interface in_iid.
	// REQUIRE: in_iid must be supported by INym

	virtual void    QueryInterface( uint32_t in_iid, void** out_ppv ) throw(RZkException) PURE;

	// PROMISE: Create a new nym on the Freedom Network
	// REQUIRE: Name does not already exist (otherwise we throw), Token
	//          is valid and has not been spent
	// NETWORK: Performs Nym Server Connection
	virtual void	CreateYourself(ZKNymName& name, const ZKSTD::string& token) PURE ;

	// PROMISE: Sets the Nym into a Disabled state.  Fine to call this if the
	//          Nym is already disabled
	// NETWORK: Performs Nym Server Connection
	virtual void	DisableNymOnNymServer() throw(RZkException) PURE ;

	// PROMISE: Update the information about a nym on the nym server with a new
	//          copy of the client side nym information.  Useful for updating changes
	//			to block list, reply blocks etc that the server needs to know about as well.
	// NETWORK: Performs Nym Server Connection
	virtual void	ModifyExistingNymOnNymServer() throw(RZkException) PURE ;

	// PROMISE: Spends a token on the Nym by updating the Nym Server
	// REQUIRE: That the token is valid and not already spent
	// NETWORK: Performs Nym Server Connection
	virtual void	AddTokenToNymOnNymServer(const ZKSTD::string& token) throw(RZkException) PURE ;

	// PROMISE: For a nym which can extend itself automatically, this 
	//          provides the return of a serial number 
	virtual void    ExtendYourself() throw(RZkException) PURE ;

	// VALID FOR USER NYM ONLY
	// PROMISE: Establish the name for the user nym
	// REVIEW: Clean this up, currently name is hardcoded anyway....
	// NETWORK: No access required
	virtual void	SetUserName(const ZKSTD::string&) PURE ;

	// PROMISE: Get and Set methods for the icon the nym 
	//			For example this might be an index into the array of nym bitmaps
	//			
	// NETWORK: No access required
	virtual void	GetNymIconInfo( WNymIconInfo& out_NymIcon ) PURE ;
	virtual void	SetNymIconInfo( const WNymIconInfo& in_NymIcon ) PURE ;

	// PROMISE: Get and Set methods for the "NymLongName" nym attribute
	// NETWORK: No access required
	virtual void	SetNymDescription(ZKSTD::string&)  PURE ;
	virtual void	GetNymDescription(ZKSTD::string&)  PURE ;

	// PROMISE: Get and Set methods for the "NymPrivateNote" nym attribute
	// NETWORK: No access required
	virtual void	SetNymPrivateNotes(ZKSTD::string&)  PURE ;
	virtual void	GetNymPrivateNotes(ZKSTD::string&)  PURE ;

	// PROMISE: Get and Set methods for the "NymPublicNote" nym attribute
	// NETWORK: No access required
	virtual void	SetNymPublicNotes(ZKSTD::string&)  PURE ;
	virtual void	GetNymPublicNotes(ZKSTD::string&)  PURE ;

	// PROMISE: Get and Set methods for the flag to determine whether to
	//			completely block cookies for a nym
	// NETWORK: No access required
	virtual bool	GetCookiesEnabled() PURE ;
	virtual void	SetCookiesEnabled(bool) PURE ;

	// PROMISE: Whether to hide the referer field or not
	// NETWORK: No access required
	virtual bool	GetHideRefererField() PURE ;
	virtual void	SetHideRefererField(bool) PURE ;

	virtual void	GetNymPreferencesBitfield(unsigned long&)  PURE ;
	virtual void	SetNymPreferencesBitfield(unsigned long, unsigned long)  PURE ;
	virtual bool	IsNymPreferenceSet( unsigned long in_PrefToTest ) PURE;

	// PROMISE: Return the Name of the Nym
	// NETWORK: No access required
	virtual bool	GetNymName(ZKSTD::string&) PURE ;

	// PROMISE: Return the Email Domain of the nym
	//			For example "test.freedom.net"
	// NETWORK: No access required
	virtual bool	GetNymEmailDomain(ZKSTD::string&) PURE ;	
	
	// PROMISE: Get/Set methods for the Nym's rules about 
	//			creating routes
	// NETWORK: No access required
	virtual void	SetRouteRulesOfNym(const MapRules&) PURE ;
	virtual void	GetRouteRulesOfNym(MapRules&) throw(RZkException) PURE ;

	// PROMISE: Get/Set methods for the security settings of the nym,
	//			For example number of hops
	// NETWORK: No access required
	virtual void	SetSecuritySettings(const SecuritySettings&) throw(RZkException) PURE ;
	virtual void	GetSecuritySettings(SecuritySettings&) throw(RZkException) PURE ;

	// Simplified access to Security Settings - never fails
	virtual size_t  GetNumberOfHops() PURE ;

	// PROMISE: Returns the date the nym was created in a formated string
	// NETWORK: No access required
	virtual void	GetNymBirthDate(ZKSTD::string&) throw(RZkException) PURE ;

	// PROMISE: Returns the date the nym was/will retire in a formated string
	// NETWORK: No access required
	virtual void	GetNymRetireDate(ZKSTD::string&) throw(RZkException) PURE ;

	// PROMISE: Returns the date the nym was/will "die" in a formated string
	// NETWORK: No access required
	virtual void	GetNymDeathDate(ZKSTD::string&) throw(RZkException) PURE ;

	// PROMISE: Returns the date the nym was created in a time_t format
	// NETWORK: No access required
	virtual void	GetNymBirthDate(time_t&) throw(RZkException) PURE ;

	// PROMISE: Returns the date the nym was/will retire in a time_t format
	// NETWORK: No access required
	virtual void	GetNymRetireDate(time_t&) throw(RZkException) PURE ;

	// PROMISE: Returns the date the nym was/will "die" in a time_t format
	// NETWORK: No access required
	virtual void	GetNymDeathDate(time_t&) throw(RZkException) PURE ;

	// PROMISE: Provide access to the private Crypto Key of the nym
	//			Note: User nym currently does not have a Crypto Key
	// NETWORK: No access required
	virtual void    GetCryptoKey(fr_crypt_pubkey**)throw(RZkException) PURE ;

	// PROMISE: Provide access to the private Signing Key of the nym
	//			Note: User nym currently does not have a Signature Key
	// NETWORK: No access required
	virtual void    GetSignatureKey(fr_crypt_pubkey**)throw(RZkException) PURE ;

	// PROMISE: Get/Set for the "Life Stage" of the nym, for example whether it is NYM_LIFE_STAGE_ACTIVE
	//			released.  Call "SetYourLifeStage" only with great caution.
	// NETWORK: No access required
	virtual void	GetYourLifeStage(unsigned long& out_LifeStage) throw(RZkException) PURE ;
	virtual void	SetYourLifeStage(unsigned long in_LifeStage )  PURE ;

	// PROMISE: Get/Set methods for the Offensive Word list
	//			The Offensive word list, aka word scan list, aka trigger info
	//			is applied globally
	//			Note: By convention the Offensive Word List on the user nym (index 0)
	//			holds the global offensive word list.
	// NETWORK: No access required
	virtual bool GetOffensiveWordList(ZKSTD::vector<ZKSTD::string>&) PURE ;
	virtual bool SetOffensiveWordList(const ZKSTD::vector<ZKSTD::string>&)PURE ;

	// Persistence Methods

	// PROMISE: Serialize the state into a chunked buffer
	// NETWORK: No access required
	virtual void Save( WMemBuf& io_Buffer ) throw(RZkException) PURE ;

	// PROMISE: Any existing state is completely lost and a new state is read from the
	//			stream
	// NETWORK: No access required
	virtual void Load( WReadBuffer& io_Buffer ) throw(RZkException) PURE ;

	// PROMISE: Will return true if the nym was created with a token that
	//			had the capability setting required to use the cloud.
	// NETWORK: No access required
	virtual bool DoesNymUseCloud() PURE;

	// PROMISE: Will return true if the nym was created with a token that
	//			had the capability setting required to use mail.
	// NETWORK: No access required
	virtual bool DoesNymUseMail() PURE;

	// PROMISE: Return true if the nym can be extended without user action
	//			- eg that the user will not need to pay for the serial number
	//			and a openly available "Extend" scalper exists.
	// NETWORK: No Access required
	virtual bool DoesNymAutoExtend() PURE;

} ;

#endif 
