// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// INymCentral.h

// NymCentral Interface provides access to the nyms that are available
// It also maintains the concept of "Active" nym

// A special case is the "User" nym which represents the actual user
// of the software

// Certain services which involve the nyms as a whole are also 
// handled here.  Some of these may move to their own interfaces in
// the future to help keep this interface clean. These include
// Message decryption and offensive word support


#ifndef _INYMCENTRAL_H
#define _INYMCENTRAL_H

#include <guids.h> // All callers potentially need access to
					  // these guids in order to call methods like
					  // GetNym

#include <WCypherMailMsg.h>
#include <WCoreTypes.h>
#include <vector>

class RZkException;


// Structure provides information about
// a WordScanSummary
class WordScanSummary 
{
public:
	WordScanSummary()
		: pScannedBuffer(NULL)
		, cntBuffer(0)
		, pFoundWords(NULL)
	{
		memset(&m_SockAddr, 0, sizeof(SOCKADDR));
	}


	unsigned char*					pScannedBuffer ;	// Structure does not own or allocated this memory
	size_t							cntBuffer ;
	ZKSTD::vector<ZKSTD::string> *  pFoundWords ;       // A vector of the words discovered in this buffer
	ZKSTD::string					strDomainName;
	SOCKADDR						m_SockAddr;
} ;

class ZKNymName ;
class WReadBuffer ;

class INymCentral
{
public:		

	////////////////////////////////////////////////////////////////////////////
	// Nym Access
	////////////////////////////////////////////////////////////////////////////

	// Determine the number of nyms 
	virtual size_t GetNumberOfNyms() PURE ;

	// Set a specific nym as the active nym
	// If USER_NYM_INDEX is specified then no nym is active
	virtual bool SetActiveNym(const size_t nIndex) PURE ;

	// REVIEW: Change to throw if interface not available....
	// Convenience access to any interface on the active nym
	virtual bool GetActiveNym(uint32_t in_iid, void** ppv)PURE ;

	// Returns the Index of the Active nym
	virtual size_t GetActiveNymIndex()PURE ;

	// Test if no nym is active, ie that the active nym is "User"
	virtual bool IsCurrentNymUser() const PURE;

	// Access to any interface from any nym by means of the nyms name
	virtual bool GetNym(
		uint32_t in_iid, 
		void** ppv, 
		const ZKSTD::string& strNymName) PURE ;

	// Access to any interface from any nym by means of the nyms index
	virtual bool GetNym(
		uint32_t in_iid, 
		void** ppv, 
		const size_t nIndex) PURE ;


	virtual void GetNymName( size_t in_indexNym, ZKSTD::string& out_strNymName ) PURE ;

    // PURPOSE: this function will return the nym index corresponding 
    //          to the given name. If the name isn;t found, the index will
    //          be set to USER_NYM_INDEX
	virtual void GetNymIndex( const ZKSTD::string& in_strNymName, size_t& out_nymIndex ) PURE ;

	// Convenience methods to get the active Nym name
	virtual void GetActiveNymName( ZKSTD::string& out_strNymName ) PURE ;


	virtual ZKSTD::string GetNoNymSelectedDisplayName() PURE;

	// Determine if the nym was created without the capability
	// to use the cloud for authenticated routes.  If a nym
	// has Cloud capability than it also has Mail capability.
	virtual bool DoesNymUseCloud(const size_t in_indexNym) PURE;

	// Determine if the nym was created without the capability
	// to use Freedom Mail.  Note that if a nym cannot use Mail
	// it cannot use the Cloud.
	virtual bool DoesNymUseMail(const size_t in_indexNym) PURE;

	virtual bool AtLeastOneNymUseMail() PURE;
	virtual bool AtLeastOneNymUseCloud() PURE;

	////////////////////////////////////////////////////////////////////////////
	// Nym Life Cycle
	////////////////////////////////////////////////////////////////////////////
	
	// Creates a new nym.  The token must be valid for nym creation,
	// the nym name must not exist on the network, the nym server must
	// be accessable - in other 
	virtual void CreateNym(	
					ZKNymName& NymName, 
					const ZKSTD::string& token ) throw(RZkException) PURE ;

	// Destroy a nym
	virtual void MarkNymAsDead(	const size_t nIndex ) throw(RZkException) PURE ;

	// Verify the life stage of the various nyms
	// This determines if any nyms have died yet
	// Call this when initializing Freedom, after ClientInfo has been downloaded (for
	// accurate time value)
	virtual void InitNymLifeCycle() PURE ;

	// Test the Nyms Life Cycle and see if any action is required:
	// Reasons for action:
	//  - Nym is nearing grace period
	//  - Nym is in grace period
	//  - Nym has expired and will be deleted permanently.
	// Possible actions include:
	//  - Return a warning message to show the user
	//  - Attempt an autoextend for nyms which can get free extend tokens
	// 
	// If no warning is necessary for this particular nym, no string is returned
	virtual void HandleNymsLifeCycleIssues(
						size_t in_nNymIndex, 
						ZKSTD::string& out_strWarning) throw(RZkException) PURE ;

	////////////////////////////////////////////////////////////////////////////
	// Email decryption
	////////////////////////////////////////////////////////////////////////////

	// This function decrypts a mail message.
	// The Message to decrypt should be set inside the WCypherMailMsg
	// If the intended nym is known that can be filled in also before
	// calling to speed up the process.  Even better, if the exact reply
	// block to use is known that can be filled in also.

	virtual void DecryptMessage( WCypherMailMsg&  io_MailMsg ) throw(RZkException) PURE ;


	////////////////////////////////////////////////////////////////////////////
	// Low level access to the state of NymCentral - for
	// usage by persistence
	////////////////////////////////////////////////////////////////////////////

	// Persistence Methods

	// Serialize the state into a chunked buffer
	virtual void Save( WMemBuf& io_Buffer ) throw(RZkException) PURE ;

	// Any existing state is completely lost and a new state is read from the
	// stream
	virtual void Load( WReadBuffer& io_Buffer ) throw(RZkException) PURE ;

	// Erase all nyms, including User nym.  No Network Access involved
	virtual void FlushNyms() PURE ;

	// Add a nym to nym central
	// used when loading up from a persisted state
	// Caller is responsible for memory management of the buffer (ie
	// it is not "freed" inside AddNym
	virtual bool AddNym_Obsolete( unsigned char* in_pBuffer, unsigned long in_ulBufferSize ) PURE ;

	virtual bool GetNymAsBuffer_Obsolete( size_t in_WhichNym, unsigned char** out_ppBuffer, size_t* cntBuffer ) PURE ;
} ;

#endif
