// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

////////////////////////////////////////////////////////////////////////////////
// INymWallet.h
////////////////////////////////////////////////////////////////////////////////

#ifndef _INYMWALLET_H
#define _INYMWALLET_H

#include <ErrorHandling.h>
#include <string>
#include <vector>
#include <client.h>
///////////////////////////////////////////////////////////////////////////////////////////////////
// Login Info

typedef struct LoginInfo {
	ZKSTD::string strNym;
	ZKSTD::string strUsername;
	ZKSTD::string strPassword;
	ZKSTD::string strDomain;
	LoginInfo(ZKSTD::string nym = "", 
		ZKSTD::string username = "", 
		ZKSTD::string password = "", 
		ZKSTD::string domain = "")
	{
		strNym = nym;
		strUsername = username;
		strPassword = password;
		strDomain = domain;
	};

	// Note the password is not compared to determine equality.
	// There should never be two login entries with all the same
	// parameters except password
	bool operator==(const LoginInfo& rhs) const
	{
		return( (strNym == rhs.strNym) &&
				(strUsername == rhs.strUsername) &&
				(strDomain == rhs.strDomain));			
	};

	bool operator<(const LoginInfo& rhs) const
	{
		ZKSTD::string strCombined1 = strNym + strUsername + strDomain;
		ZKSTD::string strCombined2 = rhs.strNym + rhs.strUsername + rhs.strDomain;
		return (strCombined1 < strCombined2);
	
	}
} LoginInfo;

///////////////////////////////////////////////////////////////////////////////////////////////////
// Name fields

enum {
	WALLET_NAME_TITLE = 101,			// Mr., Mrs., etc.
	WALLET_NAME_FIRST,			
	WALLET_NAME_MIDDLE,
	WALLET_NAME_LAST,
	WALLET_NAME_SUFFIX,			// Jr., Sr., III, etc.

	WALLET_BIRTH_DAY = 120,
	WALLET_BIRTH_MONTH,
	WALLET_BIRTH_YEAR,

	WALLET_GENDER = 130
};

static const char* WALLET_GENDERS[] = { "male", "female" };
static const int WALLET_GENDERS_NUM = sizeof(WALLET_GENDERS) / sizeof(*WALLET_GENDERS);

static const char* WALLET_NAME_TITLES[] = { "", "Mr.", "Mrs.", "Ms.", "Dr." };
static const int WALLET_NAME_TITLES_NUM = sizeof(WALLET_NAME_TITLES) 
											/ sizeof(*WALLET_NAME_TITLES);

static const char* WALLET_NAME_SUFFIXES[] = { "", "Jr.", "Sr.", "I", "II", "III" };
static const int WALLET_NAME_SUFFIXES_NUM = sizeof(WALLET_NAME_SUFFIXES) 
											/ sizeof(*WALLET_NAME_SUFFIXES);


///////////////////////////////////////////////////////////////////////////////////////////////////
// Address fields
//
// Each nym can have 3 addresses: Home, Work, and Other.
// Eg:	To get the home fax number, one would call:
//		GetField(WALLET_ADDR_START[WALLET_ADDR_HOME] + WALLET_FAX_OFFSET);

enum {
	WALLET_ADDR_HOME = 0,
	WALLET_ADDR_WORK,
	WALLET_ADDR_OTHER,
	WALLET_NUM_ADDR
};

const int WALLET_ADDR_START[] = {
	200, // HOME
	300, // WORK
	400  // OTHER
};

enum {
	WALLET_STREET_LINE1_OFFSET = 1,	
	WALLET_STREET_LINE2_OFFSET,		//If the street/apt is long, the user should split it up.
	WALLET_CITY_OFFSET,
	WALLET_STATEPROV_OFFSET,
	WALLET_POSTALCODE_OFFSET,
	WALLET_COUNTRY_OFFSET,
	WALLET_EMAIL_OFFSET,
	WALLET_TELEPHONE_INT_CODE_OFFSET,
	WALLET_TELEPHONE_LOC_CODE_OFFSET,
	WALLET_TELEPHONE_PREFIX_OFFSET,
	WALLET_TELEPHONE_SUFFIX_OFFSET,
	WALLET_TELEPHONE_EXTENSION_OFFSET,
	WALLET_FAX_INT_CODE_OFFSET,
	WALLET_FAX_LOC_CODE_OFFSET,
	WALLET_FAX_PREFIX_OFFSET,
	WALLET_FAX_SUFFIX_OFFSET,
	WALLET_ADDR_END
};


///////////////////////////////////////////////////////////////////////////////////////////////////
// Purchase info
//
// For now, purchase info consists of credit card info and applies only to the USER nym.
// Eventually, this may be extended to contain eCash.
// The user can store info for up to WALLET_NUM_CC credit cards.
// Eg:	To see if info has been entered for the i'th (0-based) credit card, check 
//		(GetField(WALLET_CC_START[i] + WALLET_CC_IN_USE_OFFSET) == CC_IS_IN_USE)
//		To get the CC type for the j'th card, call
//		GetField(WALLET_CC_START[j] + WALLET_CC_TYPE_OFFSET;
// It is okay and probably common for info to be stored for the i'th credit card, but not 
// for the (i-1)'th.

const int WALLET_CC_START[] = {
	1000,
	1100,
	1200,
	1300,
	1400,
	1500,
	1600,
	1700,
	1800,
	1900
};

const int WALLET_NUM_CC = sizeof(WALLET_CC_START) / sizeof(*WALLET_CC_START);

enum {
	WALLET_CC_IN_USE_OFFSET = 1,	// CC_IS_IN_USE if this CC has been entered
	WALLET_CC_FRIENDLY_NAME_OFFSET,
	WALLET_CC_NAME_OFFSET,
	WALLET_CC_TYPE_OFFSET,
	WALLET_CC_NUMBER_OFFSET,
	WALLET_CC_VERIFICATION_OFFSET,	// Not currently used
	WALLET_CC_EXPDATE_DAY_OFFSET,	// Not currently used
	WALLET_CC_EXPDATE_MONTH_OFFSET,
	WALLET_CC_EXPDATE_YEAR_OFFSET,
};

#define CC_IS_IN_USE	"IN USE"

///////////////////////////////////////////////////////////////////////////////////////////////////

typedef ZKSTD::map<int, std::string> Field2ValueMap;

// Interface
///////////////////////////////////////////////////////////////////////////////////////////////////

class INymWallet
{
public:
	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: To add personal info to the wallet. 
	//			e.g. SetField(WALLET_ADDR_START[WALLET_ADDR_HOME] + WALLET_COUNTRY_OFFSET, "Canada");
	//			e.g. SetField(WALLET_NAME_TITLE, "Mr.");
	//
	// REQUIRE: Nothing
	//
	// PROMISE:	If a value already exists for in_field, it will be overwritten.

	virtual void SetField(			const int in_field, 
									const ZKSTD::string& in_strValue) PURE;

	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: To get personal info from the wallet.  
	//
	// REQUIRE: Nothing
	//
	// PROMISE:	If no info is stored for this field, "" is returned.

	virtual ZKSTD::string GetField(	const int in_field) PURE;
	
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: To get a random string that makes sense for this field.  For example if the 
	//			requested field is an address, something like "123 Hfgkh St" will be returned
	//			* if (in_bPseudonymous), some fields will not be completely random.  For example,
	//			the email address will be the nym's email address and the name will be the nym's
	//			name.
	//			* Special case:  when requesting postal codes and provinces, the
	//			postal code and province should be requested one after the other to ensure the
	//			postal code and province matche.
	//
	// REQUIRE: Nothing
	//
	// PROMISE:	If an unknown field is passed in, a random string will be returned.

	virtual ZKSTD::string GetRandomField(const int in_field, bool in_bPseudonymous = true) PURE;


	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: Populate all personal info (not credit cards) with pseudo-random,
	//			or random data.  Pseudo-random data makes the email address be the nym's email
	//			address, the first name be the nym name, etc.  Random data creates a completely
	//			random email address, name, etc.
	//
	// REQUIRE: Nothing
	//
	// PROMISE:	Nothing.

	virtual void PopulatePersonalInfoWithRandomFields(bool in_bPseudonymous = true) PURE;


	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: To add a nym, username, password, domain combination to the wallet.  While the nym
	//			field may seem extraneous (because each instanciation of a NymWallet belongs to
	//			one nym), having only one structure for LoginInfo makes everything easier. 
	//
	// REQUIRE: The strNym field of the LoginInfo must be the same as the nym name of the INym
	//			that instanciates the NymWallet.
	//
	// PROMISE:	The strDomain field can be passed in as any valid url. This method will happilly 
	//			convert the url  from something like "http://www.FrEeDoM.net/asa"to something like
	//			"freedom.net" before storing it.
	//			If the domain/nym/username combo is found, the password will be changed to be
	//			in_loginInfo.strPassword
	//			Returns false if:
	//			1. The strUsername or strPassword field is blank.
	//			2. The strNym field doesn't match this nym's name
	//			3. The exact username/password/domain/nym combo is already stored.
	//			Otherwise (including the case where a matching username/domain/nym- but not
	//			password- combo is found) it returns true 

	virtual bool AddLoginInfo(const LoginInfo& in_loginInfo) PURE ;


	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: To get all of the LoginInfo for a given domain.
	//
	// REQUIRE: Nothing
	//
	// PROMISE:	outloginInfoVector will be cleared before inserting LoginInfo entires.
	//			- If (in_strDomain == 0), this will be interpreted to be the local computer.
	//			- If in_strDomain is a reasonable domain (like http://www.FrEeDoM.net/as) it will
	//			be converted to the domain (like freedom.net). 
	//			- If in_strDomain is a wacky domain, the entire in_strDomain will be used.
	//
	//			Returns an empty vector if there is no login info stored for the given domain.

	virtual void GetLoginInfo(	const ZKSTD::string in_strDomain,
								ZKSTD::vector<LoginInfo>& out_loginInfoVector) PURE;

		///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: To get all of the LoginInfo for all domains.
	//
	// REQUIRE: Nothing
	//
	// PROMISE:	outloginInfoVector will be cleared before inserting LoginInfo entires.
	//			Returns an empty vector if there is no login info stored.

	virtual void GetAllLoginInfo(ZKSTD::vector<LoginInfo>& out_loginInfoVector) PURE;

	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: To delete a specific LoginInfo entry.
	//
	// REQUIRE: Nothing.
	//
	// PROMISE:	The strDomain field can be passed in as any url. This method will happilly 
	//			convert the url  from something like "http://www.FrEeDoM.net/asa"to a domain like
	//			"freedom.net" before trying to delete the login entries.
	//			Returns false if the given loginInfo entry is not found.

	virtual bool DeleteLoginInfo(const LoginInfo& in_loginInfo) PURE;

	virtual void SelfTest() PURE;
};

#endif