// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// IRouteManager.h

#ifndef _IROUTEMANAGER_H
#define _IROUTEMANAGER_H

//#include <set>

#include <clientstl.h> // FOr Port2AipIds
#include <fr_entid.h>
#include <frAips.h>

#ifdef ZK_LINUX
#  include "ZkSubject.h"
#  include "ZkHint.h"
#endif // ZK_LINUX

class RRouteManager ;
class RZkException;
class WRoute ;
class DZkError;

class IRouteManager
#ifdef ZK_LINUX
: public ZkSubject
#endif
{
public:

	virtual void				Init() = 0 ;
	virtual void				Terminate() = 0 ;

	// Inform the Route of the time reported to us from the Client info
	// This isn't an argument to Init because that can be called before 
	// we have talked to Niqs.
	virtual void				SetFreedomTime(const time_t utc) = 0 ;

	// Start an authenticated route for specified nym
	// (System currently only allows a single authenticated route at a time)
	virtual int					StartAuthenticatedRoute(
										size_t in_iNym ) = 0 ;

	// Create an unauthenticated route to the specified freedom server
	// The port number defines which type of server it is.
	// The function does not return until it succeeds, it will perform multiple
	// retries is successful.  If there are no valid routes to the freedom server
	// or several retries have failed then it will return false.
	virtual bool				StartUnauthRoute(const uint16_t in_Port ) throw() = 0;

	virtual void				TerminateUnauthRoute(const uint16_t in_uPort) = 0 ;

	virtual void				KillAuthenticatedRoute() = 0 ;

	// Terminate NMTA and POP routes, if any exist.
	// Should be called whenever the active nym is changed.
	virtual void				TerminateMailRoutes() = 0;

	// Call this function when a nym's route has died.
	// It will attempt to build a new route.  May provoke UI if
	// interactive routing is enabled.
	// This function does nothing if no nym is active or
	// other conditions are not valid.
	virtual bool				AttemptRouteDeathRecovery() = 0 ;

	// Information API
	// These calls are intended for debugging/test code etc
	// Get the ACI of the Route ID 
	virtual ZkACI				GetAuthRouteID() = 0 ;
	virtual void				GetAuthenticatedRoutes(Port2AipIds&, size_t&) = 0 ;
	// Pass in NULL for out_pRoute, if the caller simply cares if a route exists or not.
	virtual bool				GetUnauthenticatedRoute(const uint16_t, unsigned int* out_pRoute = NULL) = 0 ;
	virtual void				GetUnauthenticatedRoutes(Port2AipIds&, size_t&) = 0;	

	// For debugging purposes, we allow the display of the route table.
	// This also displays extensive information in the debug traces.
//	virtual void				DisplayRouteTableInConsole() = 0 ;

	// Determines if we have a valid route to get to a certain destination
	// Used for security purposes
	//	CALLER : Firewall
	virtual bool				IsCloudTrafficAllowed( const u_long in_IP ) = 0 ;

	// Determine whether it is legal to send a TCP packet on the DH socket to
	// the IP in question
	virtual bool				IsDiffieHellmanTrafficAllowed( const u_long in_IP ) = 0 ;


	// Helper functions for getting a string description of a route
	// You might expect this to be a method on WRoute itself but WRoute isn't aware of
	// the user friendly names of the Aips in a route
	virtual ZKSTD::string		GetRouteDesc( const WRoute & in_routeToDescribe ) throw() = 0 ;
	virtual ZKSTD::string		GetRouteDesc( const FrEntIdType& in_entType ) throw() = 0 ;

	// Test code
	virtual void		SimulateAuthenticatedRouteDestroy() = 0 ;

	// Our own timers call these functions (because they are static methods)
	virtual void				ProlongNymRoute() = 0 ;
	virtual int					SendCoverTrafficPacketOnAllRoutes() = 0 ;

	virtual int	GetNymInformation(DZkError& out_Err, const ZkACI in_Route, char* out_nymName, FrCryptSig& out_key) = 0;

	virtual int GetAipSymKeys(DZkError& out_Err, const ZkACI in_nRouteId, size_t& out_nAipsInRoute, FrCryptSym* out_symkeys) = 0;

	virtual int GetFirstHopInRoutePublicSigningKey(DZkError& out_Err, const ZkACI in_Route, FrCryptSig& out_key) = 0;

    virtual unsigned long UpdateIdleCounter() = 0;
    virtual void          KillIdleRoutes() = 0;

#ifdef ZK_LINUX

    // Nested class used as hint to Observers when our implementation's state changes.
    // (Some people hate nested classes IMHO, if we don't start using them and the subject / observer
    // pattern becomes widely used, both the namespace and source tree will get awfully clutered)

    class RouteHint 
        : public ZkHint
    {
    public:

        // Add other events which cause a change in route manager's state here
        enum {
            ROUTE_DEATH_NOTIFICATION
        } ;
        
        RouteHint( int stateChange )
            : m_StateChange( stateChange )
        { } ;

        ~RouteHint() {} ;

        int GetStateChange() { return m_StateChange ; } ;
    private:
        int m_StateChange;
    } ;

#endif // ZK_LINUX
} ;

#endif
