// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// IRoutePicker.h
//
// RoutePicker contains the logic for finding routes based on the
// topology.  It relies on INetworkPerformance for performance info
// about the AIPs and on INiqs for the list of Aips and their topology.
// It does not actually build routes - IRouteManager will do that work.
//
// It roughly corresponds to the obsolete frNetRoute() library
//
// It is intended to meet the following goals;
// -Pick valid routes based on topology of the current network
// -Pick valid routes based on user rules
// -Pick routes in way that is difficult to predict (use randomness)
// -Pick a fast route
// -Be able to calculate multiple possible routes in a single pass
// -Be efficient

#ifndef IROUTEPICKER_H
#define IROUTEPICKER_H

#include <WRoute.h>
#include <WRouteRules.h>

// This is a maximum, if there aren't this many possible routes
// to match the graph then we search the whole graph and stop
// For single hop routes this means we collect 8 routes, for multiple
// hops we continue until we have collected routes started at this many
// different first hops
#define NUMBER_DISTINCT_FIRST_HOPS_IN_ALIST  8

// By default, if any of the hops have a rating below this value
// we consider it B-List.  Perhaps this is a value that should
// adapt.  In general it can't be too strict or we will search more
// of the graph, but it definitely should help avoid hops that might
// be down.
#define DEFAULT_MIN_A_LIST_RATING 100				

// To avoid excessive traversal of the tree we stop once
// the B-List reachs a size.  We want it fairly large to allow
// a good chance at finding lots of A list enteries
#define DEFAULT_BLIST_ROUTES_TO_COLLECT 50


class IRoutePicker 
{
public:

	// PROMISE: Will use the current topology and the rules and find some
	//			routes.  Randomness is an important element of the so repeated
	//			calls may return different results.  Routes are not necesssary the
	//			"best", but we attempt to find ones that might be be best and probably
	//			aren't the worst.
	//			
	//			Topology may not have as many routes as are requested, possibly even 
	//			no routes at all will satisfy the provided rules.  Caller should
	//			no assume that the size of the return vector is as large as in_cntDesiredRoutes
	// REQUIRE: INiqs should have enough topology information

	virtual void FindRoutes( 
				size_t in_cntDesiredRoutes, 
				const WRouteRules & in_RulesToApply,
				WRouteVec& out_FoundRoutes ) = 0;

	// Determines whether some rules are valid
	// Note: MapRules is obsolete in v2.0, it is not exposed to the user
	// Unless we choose to bring it back in the future this can be cleanedup
	// as obsolete.
	virtual bool CheckRules(const MapRules& in_Rules, const size_t in_numHops, size_t& out_numRules) throw() PURE ;

	// For debugging purposed we can vary some of the 
	// values in the heuristic (corresponding to DEFAULT_ defines above)

	virtual void SetMaxFirstHopsInAList( size_t in_newMax ) = 0 ;
	virtual void SetAListMinRating( size_t in_newMin ) = 0 ;
	virtual void SetBListMaxSize( size_t in_newMax ) = 0 ;

} ;

#endif