// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// ITokenService

// This service provides Serial Number and Token access

// A serial number can be cashed in for multiple tokens of a single
// type.

// The client can hold multiple tokens of multiple types.
// These tokens are persisted locally and should only 
// be erased after they are spent.

#ifndef _ITOKENSERVICE_H
#define _ITOKENSERVICE_H

class WTokenDescriptions ;
class RZkException;

struct T_FrTokenType;

#define ZKSERIAL_NUMBER_LENGTH 14

#ifdef WIN32
#  include <winsock2.h>
#endif

#include <WSerialNumber.h>
#include <WTokenDescriptions.h>
#include <vector>

class TokensOfType
{
public:

	TokensOfType( const WTokenDescriptions& rhs )
		:Desc( rhs )
	{ ; } 


	WTokenDescriptions Desc ; // All tokens in this instance of the class have identical description
	ZKSTD::vector<int>    Tokens ; //Indexes of tokens that match this description
								 //We need to track indexes because we cannot
								 //assume the array of tokens is sorted
} ;

class ITokenService
{
public:

	// Fill in a WTokenDescription object with the description
	// information of specified token
	virtual bool				GetDescriptionOfTokenX(
									const unsigned int, 
									WTokenDescriptions&) PURE;

	virtual unsigned int	GetTotalNumberOfTokens() PURE;


	// Returns a Base64 representation of a token
	virtual bool				GetTokenXinBase64(const unsigned int, 
												  ZKSTD::string&)  PURE;

	// Removes a spent or otherwise unwanted token
	virtual bool				RemoveTokenXFromArray(const unsigned int) PURE;
	virtual bool				RemoveTokenFromArray(const ZKSTD::string&)  PURE;	

	// Contact the Serial Number to determine the capabilities of a
	// particular serial number.  Serial number is expected to be
	// valid format and not to have been spent.  Provided it is in the
	// database on the server the WSerialNumber class is filled in with
	// information about how many tokens the serial number will produce,
	// whether it is a trial serial number etc
	virtual	void				GetSerialNumberCapabilities(
									WSerialNumber& io_SerialNumber ) throw(RZkException) PURE ;

	// Cash in a serial number for tokens
	// Serial number is expected to be valid format and present in the server capabilities.
	virtual void				RedeemSerialNumberForTokens( 
									const WSerialNumber& in_SerialNumber ) throw(RZkException) PURE ;

	// Cash in a serial number for tokens
	// Serial number is expected to be valid format and present in the server capabilities.
	// but fill the vector with the new tokens obtained
	virtual void				RedeemSerialNumberForTokens( 
									const WSerialNumber& in_SerialNumber, 
									ZKSTD::vector<tstring>& out_vStrTokens) throw(RZkException) PURE ;

    // Obtains a free serial number that is good for creating a free nym
	// and redeems it for tokens.
	virtual void				GetFreeNymCreateSerialNumberAndRedeemIt( ) throw(RZkException) PURE ;
    
	// Gets a serial number that can be used to extend and returns it (assumes
	// serial number only produces a single extend token)
	virtual void				GetFreeNymExtendSerialNumberAndRedeemIt( ZKSTD::string& out_token ) throw(RZkException) PURE ;


	// Get the tokens grouped by types.  Used by Ctrl on different platforms
	virtual void				GetTokensGroupedByTypeID(ZKSTD::vector<TokensOfType>& out_tokenTypes) PURE;

	// Get the number of different type of token
	virtual unsigned int		GetNumberOfDifferentTokenTypeID() PURE;

	// Determine the token capability to use the cloud
	virtual bool DoesTokenUseCloud(const unsigned int) PURE;
	virtual bool DoesTokenUseCloud(const ZKSTD::string&) PURE;

	// Determine the token capability to use the mail
	virtual bool DoesTokenUseMail(const unsigned int) PURE;
	virtual bool DoesTokenUseMail(const ZKSTD::string&) PURE;
} ;

#endif 
