// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "StdAfx.h"

#include "RActiveRouteTable.h"

#include "INiqs.h"
#include "INym.h"
#include "IRouteManager.h"
#include "INymCentral.h"
#include "ZkService.h"
#include <ZkErrorHandler.h>
#include <IZKErrorReporter.h>
#include <RZkException.h>
#include <ZKConvert.h>
#include <lockguard.h>
#include <IZkConsoleMsg.h>
#include <TraceControl.h>

#ifdef WIN32
#include <clientwin32WM.h>
#include <TrafficFilter.h>		// for 	AllowNetBiosInTheClear

// Marcp: this need to be fixed
#include "../win/ZKBaseUI.h"					// Temporary
#endif

#define ROUTE_EXPIRY_GRACE_PERIOD_SECONDS 30  // We give a little time before removing dead routes
											  // so that we have less mis-directed route destroy packets

RActiveRouteTable::RActiveRouteTable()
	: RRWLock()
	, m_Context( NULL )
{
}

RActiveRouteTable::~RActiveRouteTable()
{
	// We don't own this object
	m_Context = NULL ;
}

void
RActiveRouteTable::Init( FrLocalAipCtx in_Ctx ) throw( RZkException )
{
	ZKASSERT( in_Ctx != NULL ) ;
	ZKASSERT( m_Context == NULL ) ;

	m_Context = in_Ctx ;

	ValidateState() ;
}

size_t 
RActiveRouteTable::GetIndexOfNymOwningAuthenticatedRoute( ZkACI in_ACI ) throw ( RZkException )
{
	ZKS::Lock_Guard<RRWLock> gate(*this);

	ValidateAci( in_ACI ) ;
	ZKASSERT( IsAciValid( in_ACI ) ) ;

	// Currently only authenticated route is associated with a nym
	ZKASSERT( m_RouteTable[ in_ACI ].GetRouteType() == AuthenticatedRoute ) ;

	return m_RouteTable[ in_ACI ].GetNymIndex() ;
}


ZkACI
RActiveRouteTable::InitNewAuthenticatedRouteWithSameExitAip( 
    ZkACI in_RouteToCopy,
	int in_iNym,
	int in_Duration, 
	WRoute& in_route ) throw( RZkException )
{
	ZkACI newRouteAci = INVALID_ROUTE_ID ;
	
	if ( !IsAciValid( in_RouteToCopy ) )
	{
		// We can't prolong an invalid route
		return newRouteAci ;
	}
	
	Lock( ZK_WRITER );

	try
	{	
		sockaddr_in firstAipAddr ;

		if(!frLocalAipNewRouteReuseExit(
				&m_err, 
				m_Context, 
				in_Duration, 
				in_route.GetSize(), 
				in_route.GetHopArray(), 
				&firstAipAddr, 
				&newRouteAci, 
				in_RouteToCopy))
		{
			// This can fail when the user has changed route
			// preferences (eg number of hops etc)
			throw RZkException(m_err, "Route::ProlongRoute()",
								"frLocalAipNewRouteReuseExit()",
								IDS_PROLONG_ROUTE_CANT_REUSEEXIT);
		}

		ValidateAci( newRouteAci ) ;

		// REVIEW - frLocalAip doesn't allow us to get the First Aip Addr
		// later so we remember it ourselves.  API change will fix it
		m_RouteTable[ newRouteAci ].SetFirstAipAddr( firstAipAddr ) ;

		// REVIEW - frLocalAip doesn't allow us to get the duration
		// so we remember it ourselves.
		m_RouteTable[ newRouteAci ].SetDuration( in_Duration ) ;


		m_RouteTable[ newRouteAci ].SetRouteType( AuthenticatedRoute ) ;

		m_RouteTable[ newRouteAci ].SetRouteValidNotActive() ;

		SetNymInformation( newRouteAci, in_iNym ) ;

		ValidateState()	;

	}
	catch(IZkException&)
	{
		Unlock( ZK_WRITER ) ;
		throw ;
	}

	Unlock( ZK_WRITER ) ;

	return newRouteAci ;
}

ZkACI
RActiveRouteTable::InitNewRoute( 
	int in_Duration, 
	WRoute& in_route ) throw( RZkException )
{
	ZkACI newRouteAci = INVALID_ROUTE_ID ;
	sockaddr_in firstAipAddr ;

	if(!frLocalAipStartNewRoute(
			&m_err, 
			m_Context, 
			in_Duration, 
			in_route.GetSize(), 
			in_route.GetHopArray(), 
			&firstAipAddr, 
			&newRouteAci))
	{
		throw RZkException(m_err, "Route::BuildRoute()",
							"frLocalAipStartNewRoute() failed",
							IDS_BUILDROUTE_FAIL);
	}


	ValidateAci( newRouteAci ) ;

	// REVIEW - frLocalAip doesn't allow us to get the First Aip Addr
	// later so we remember it ourselves.  API change will fix it
	m_RouteTable[ newRouteAci ].SetFirstAipAddr( firstAipAddr ) ;

	// REVIEW - frLocalAip doesn't allow us to get the duration
	// so we remember it ourselves.
	m_RouteTable[ newRouteAci ].SetDuration( in_Duration ) ;

	m_RouteTable[ newRouteAci ].SetRouteValidNotActive() ;

    m_RouteTable[ newRouteAci ].SetRouteIdleCounter( 0 ) ; // Zero means never used.

	return newRouteAci ;
}

ZkACI 
RActiveRouteTable::InitNewUnauthenticatedRoute( 
	const uint16_t in_DestPort, 
	int in_Duration, 
	WRoute& in_route ) throw( RZkException )
{
	// This method adds an entry to the table 
	// so it needs a writer lock
	Lock( ZK_WRITER ) ;

	ZkACI newRouteAci = INVALID_ROUTE_ID ;

	try
	{

		ZKCTRACE0( TRACE_ROUTEMANAGER, 
				   "Adding new unauthenticated route to table\nTable contents:\n" ) ;
		ValidateState() ;

		newRouteAci = InitNewRoute( in_Duration, in_route ) ;
		m_RouteTable[ newRouteAci ].SetRouteType( UnauthenticatedRoute ) ;

		m_RouteTable[ newRouteAci ].SetDestinationPort( in_DestPort ) ;

		ZKCTRACE0( TRACE_ROUTEMANAGER, 
				   "Finished Adding new unauthenticated route to table\nTable contents:\n" ) ;
		ValidateState() ;
	}
	catch(IZkException&)
	{
		Unlock( ZK_WRITER ) ;
		throw ;
	}

	Unlock( ZK_WRITER ) ;

	return newRouteAci ;
}


ZkACI 
RActiveRouteTable::InitNewAuthenticatedRouteForNym( 
	int in_Duration, 
	WRoute& in_route,	
	int in_iNym
	) throw( RZkException )
{
	// This method adds an entry to the table 
	// so it needs a writer lock
	Lock( ZK_WRITER ) ;

	ZkACI newRouteAci = INVALID_ROUTE_ID ;

	try
	{

		ZKASSERT( in_iNym != USER_NYM_INDEX ) ;

		ZKCTRACE0( TRACE_ROUTEMANAGER, 
				   "Adding new authenticated route to table\nTable contents:\n" ) ;
		ValidateState() ;

		// TODO - for consistency check to make sure nym doesn't already
		// have a route up and running which we don't currently allow

		
		newRouteAci = InitNewRoute( in_Duration, in_route ) ;
		m_RouteTable[ newRouteAci ].SetRouteType( AuthenticatedRoute ) ;

		SetNymInformation( newRouteAci, in_iNym ) ;

		ZKCTRACE0( TRACE_ROUTEMANAGER, 
				   "Finished adding new unauthenticated route to table\nTable contents:\n" ) ;
		ValidateState() ;
	}
	catch(IZkException&)
	{
		Unlock( ZK_WRITER ) ;
		throw ;
	}

	Unlock( ZK_WRITER ) ;

	return newRouteAci ;
}

void
RActiveRouteTable::ValidateAci( ZkACI in_Route ) const throw( RZkException )
{
	// API is being called with invalid Route ID
	ZKASSERT( in_Route != INVALID_ROUTE_ID ) ;

	if ( in_Route >= MAX_ROUTES )
	{
		throw RZkException(m_err, "RActiveRouteTable::ValidateAci()",
							"Invalid Route Table Index",
							IDS_UNEXPECTED_FAIL);
	}
}

void 
RActiveRouteTable::RouteIsActive( ZkACI in_Route ) throw( RZkException )
{
	ZKS::Lock_Guard<RRWLock> gate(* const_cast<RActiveRouteTable*>(this));

	ValidateAci( in_Route ) ;
	m_RouteTable[ in_Route ].SetRouteActive() ;

	time_t	routestarttime;
	time( &routestarttime ); // seconds elapsed since 1 Jan 1970	

	m_RouteTable[ in_Route ].SetStartTime( routestarttime ) ;

#ifdef WIN32
	// For optimization purposes inform our driver of the first hop IPs so
	// that the packets can be efficiently allowed out to this destination.  
	// Linux version doesn't need this because it can route packets based on PID

	ZKSTD::vector< unsigned int > aipIPs ;

	for ( size_t i = 0 ; i < MAX_ROUTES ; i++ )
	{
		if ( m_RouteTable[ i ].GetRouteState() == RouteActive )
		{
			sockaddr_in addr ;
			GetFirstAipAddr( i, addr ) ;

			aipIPs.push_back( addr.sin_addr.s_addr /* ntohl(addr.sin_addr.s_addr) */) ;
		}
	}

	TrafficFilter::theTrafficFilter().SetRulesForFirstHopsInRoutes( aipIPs ) ;
#endif
}

bool
RActiveRouteTable::IsAciValid( ZkACI in_Route ) const throw( RZkException )
{
	ZKS::Lock_Guard<RRWLock> gate(* const_cast<RActiveRouteTable*>(this));

	ValidateAci( in_Route ) ;
	return m_RouteTable[ in_Route ].GetRouteType() != RouteTypeUnknown ;
}

bool 
RActiveRouteTable::IsRouteActive( ZkACI in_Route ) const throw( RZkException ) 
{	
	ZKS::Lock_Guard<RRWLock> gate(* const_cast<RActiveRouteTable*>(this));
	ValidateAci( in_Route ) ;
	return m_RouteTable[ in_Route ].IsActive() ;
}

RouteState
RActiveRouteTable::GetRouteState( ZkACI in_Route ) const throw( RZkException )
{
	ZKS::Lock_Guard<RRWLock> gate(* const_cast<RActiveRouteTable*>(this));
	ValidateAci( in_Route ) ;
	return m_RouteTable[ in_Route ].GetRouteState() ;
}

void 
RActiveRouteTable::DeleteDeadRoute( 
	ZkACI in_ACI, 
	ZkACI in_OutgoingACI, 
	bool in_bNotify ) throw( RZkException )
{
	if ( !IsAciValid( in_ACI ) )
	{
		// Presumably this route has already
		// been cleaned up
		return ;
	}

	Lock( ZK_READER ) ;

	ValidateAci( in_ACI ) ;

	ZKCTRACE1( TRACE_ROUTEMANAGER, 
			   "Deleting Route %d, table contents before delete:\n", in_ACI ) ;
	ValidateState() ;


	ZKCTRACE1( TRACE_ROUTEMANAGER, 
			   "Deleting Dead route from table %d\n",
			   in_ACI ) ;
	m_RouteTable[ in_ACI ].TraceState() ;

	// Collect some information before be remove everything
	bool bAuthenticatedRoute = ( m_RouteTable[ in_ACI ].GetRouteType() == AuthenticatedRoute ) ;
	bool bRouteActive = ( GetRouteState( in_ACI ) == RouteActive ) ;
	uint16_t uServerPort = 0 ;
	
	if ( !bAuthenticatedRoute )
	{
		uServerPort = m_RouteTable[ in_ACI ].GetDestinationPort() ;
	}

    // Show console messages if we want to notify the user what is going on
    if (( GetRouteType( in_ACI ) == AuthenticatedRoute ) &&
        bRouteActive )
    {
        ZKSTD::string strNymName ;
        ZkService::GetINymCentral()->GetNymName( m_RouteTable[ in_ACI ].GetNymIndex(), strNymName ); 

        IZkConsoleMsg::ShowMsg( 
            CONS_VERBOSE,
            CONSIMAGE_COMP_TERM, 
            IDS_CONS_AUTH_ROUTE_DELETE,
            strNymName.c_str() ) ;
		}
    else if ( GetRouteType( in_ACI ) == UnauthenticatedRoute ) 
    {
        IZkConsoleMsg::ShowMsg( 
            CONS_VERBOSE,
            CONSIMAGE_COMP_TERM, 
            IDS_CONS_UNAUTH_ROUTE_DELETE,
            ZKConvert::GetServerNameFromPort( GetDestinationPort( in_ACI ) ).c_str()
        ) ;
    }

	Unlock( ZK_READER ) ;

	// Lock the table just long enough to remove the ACI

	Lock( ZK_WRITER ) ;

	if ( in_OutgoingACI != 0 )
	{
		ZkACI currentOutgoingACI = 0 ;
		ZKVERIFY( frLocalAipGetOutgoingAci(&m_err, m_Context, in_ACI, &currentOutgoingACI )) ;

		if ( currentOutgoingACI != in_OutgoingACI )
		{
			// Presumably the route has been deleted from underneath us

			ZKCTRACE2( 
				TRACE_ROUTEMANAGER, 
				"Cancelling deletion of route because Outgoing ACI doesn't match (%d versus %d in table)\n",
				in_OutgoingACI,
				currentOutgoingACI ) ;					

			Unlock( ZK_WRITER ) ;
			return ;
		}
	}

	// Clean up the route

	m_RouteTable[ in_ACI ].Clear() ;

	ZKVERIFY( frLocalAipDeleteRoute(&m_err, m_Context, in_ACI ) );

	Unlock( ZK_WRITER ) ;

	ZKCTRACE0( TRACE_ROUTEMANAGER, 
			   "Delete complete table contents now:\n" ) ;
	ValidateState() ;
	

	// Notification

	// Is here the place to check for lack of nym route?
	// Maybe this isn't the right place but its here for the moment
	// to retain functionality

	if ( bAuthenticatedRoute && in_bNotify && bRouteActive )
	{
		if ( !ZkService::GetINymCentral()->IsCurrentNymUser() )
		{
			if ( INVALID_ROUTE_ID == GetActiveAuthenticatedRoute() )
			{
				// This needs to be replaced by a cross platform way to notify main UI
				// One possibility is to make the "UI" an observer on the ZkService::GetINymCentral()
				// or possibly on this object.

#ifdef WIN32
				::PostMessage(ZKBaseUI::GetFreedomBarHwnd(), WM_SELECTUSER_NYM, 0, 0);
#else
                IRouteManager::RouteHint hint(IRouteManager::RouteHint::ROUTE_DEATH_NOTIFICATION);
                ZkService::GetIRouteManager()->notify(&hint);
#endif
			}
		}
	}
}

void
RActiveRouteTable::GetFirstAipAddr( ZkACI in_ACI, sockaddr_in& out_addr ) const throw( RZkException )
{ 
	ZKS::Lock_Guard<RRWLock> gate(* const_cast<RActiveRouteTable*>(this));
	memset( &out_addr, 0, sizeof( sockaddr_in ) ) ;

	ValidateAci( in_ACI ) ;
	out_addr = m_RouteTable[ in_ACI ].GetFirstAipAddr() ;
}

void
RActiveRouteTable::GetExitAipAddr( ZkACI in_ACI, sockaddr_in& out_addr ) const 
{ 
	ZKS::Lock_Guard<RRWLock> gate(* const_cast<RActiveRouteTable*>(this));
	memset( &out_addr, 0, sizeof( sockaddr_in ) ) ;

	ValidateAci( in_ACI ) ;

	char * strHost = NULL ;
	ZKVERIFY( frLocalAipGetExit( &m_err, m_Context, in_ACI, &out_addr, &strHost ) ) ;

	SAFEFREE( strHost ) ;
}


void
RActiveRouteTable::GetExitAipHost( ZkACI in_ACI,  ZKSTD::string& out_str ) const 
{ 
	ZKS::Lock_Guard<RRWLock> gate(* const_cast<RActiveRouteTable*>(this));
	SOCKADDR_IN dummy ;
	ValidateAci( in_ACI ) ;

	char * strHost = NULL ;
	ZKVERIFY( frLocalAipGetExit( &m_err, m_Context, in_ACI, &dummy, &strHost ) ) ;

	out_str	= strHost ;

	SAFEFREE( strHost ) ;
}


void 
RActiveRouteTable::GetAipIds(
		const unsigned int in_ACI, 
		WRoute& out_Route ) const
{
	ZKS::Lock_Guard<RRWLock> gate(* const_cast<RActiveRouteTable*>(this));
	
	ValidateAci( in_ACI ) ;

	size_t numAips = MAX_AIPS_IN_ROUTE ;

	ZKVERIFY(frLocalAipGetAipIds(&m_err,
                             m_Context,
                             in_ACI,
                             out_Route.GetHopArray(),
                             &numAips));

	ZKASSERT( numAips == out_Route.GetSize() ) ;
}

void 
RActiveRouteTable::GetAipIds(
	const unsigned int in_ACI, 
	uint32_t out_AipIds[MAX_AIPS_IN_ROUTE], 
	size_t& io_numAips ) const
{
	ZKS::Lock_Guard<RRWLock> gate(* const_cast<RActiveRouteTable*>(this));
	ZKASSERT(io_numAips <= MAX_AIPS_IN_ROUTE);
	
	ValidateAci( in_ACI ) ;

	ZKVERIFY(frLocalAipGetAipIds(&m_err,
                             m_Context,
                             in_ACI,
                             out_AipIds,
                             &io_numAips));
}

bool 
RActiveRouteTable::IsIPFirstHopOfRoute( uint32_t in_IP, ZkACI * out_pACI )
{
	ZKS::Lock_Guard<RRWLock> gate(*this);
	bool retVal = false ;

	for ( size_t i = 0 ; i < MAX_ROUTES ; i++ )
	{
		if ( m_RouteTable[ i ].GetRouteType() == UnauthenticatedRoute )
		{
			// A route to a server ends at the server itself
			sockaddr_in addr ;
			GetFirstAipAddr( i, addr ) ;

			if ( in_IP == addr.sin_addr.s_addr )
			{
				if ( out_pACI != NULL )
				{
					*out_pACI = i ;
				}

				retVal = true ;
				break ;
			}
		}
	}

	return retVal ;
}

// See if we have any active routes to a particular Entity
// Type.  for example are there any Nym Server routes?
bool
RActiveRouteTable::IsRouteToServer( FrEntIdType in_DestinationEntity, ZkACI * out_pACI )
{
	ZKS::Lock_Guard<RRWLock> gate(*this);
	bool retVal = false ;

	for ( size_t i = 0 ; i < MAX_ROUTES ; i++ )
	{
		if ( m_RouteTable[ i ].GetRouteType() == UnauthenticatedRoute )
		{
			// A route to a server ends at the server itself
			sockaddr_in exit ;
			GetExitAipAddr( i, exit ) ;

			FrEntIdType type = ZKConvert::GetEntityTypeFromPort( ntohs( exit.sin_port ) ) ;

			if ( type == in_DestinationEntity )
			{
				if ( out_pACI != NULL )
				{
					*out_pACI = i ;
				}

				retVal = true ;
				break ;
			}
		}
	}

	return retVal ;
}

time_t 
RActiveRouteTable::GetExpireTime( ZkACI in_nRouteID )
{
	ZKS::Lock_Guard<RRWLock> gate(*this);

	ValidateAci( in_nRouteID ) ;

	if (( m_RouteTable[ in_nRouteID ].GetRouteState() == RouteActive ) ||
		( m_RouteTable[ in_nRouteID ].GetRouteState() == RouteRetired ))
	{
		ZKASSERT( m_RouteTable[ in_nRouteID ].GetStartTime() != 0 ) ;
		ZKASSERT( m_RouteTable[ in_nRouteID ].GetDuration() > 0 ) ;

		return ( m_RouteTable[ in_nRouteID ].GetStartTime() + 
				 m_RouteTable[ in_nRouteID ].GetDuration() ) ;
	}
	else
	{
		// This can happen in the period before the route is fully set up
		return 0 ;
	}
}

bool
RActiveRouteTable::IsRouteExpired(ZkACI in_nRouteID)
{
	ZKS::Lock_Guard<RRWLock> gate(*this);

	// Assume not expired by default
	bool bRet = false ;

	ValidateAci( in_nRouteID ) ;
	
	if (( m_RouteTable[ in_nRouteID ].GetRouteState() == RouteActive ) ||
		( m_RouteTable[ in_nRouteID ].GetRouteState() == RouteRetired ))
	{
		time_t currenttime;
		time( &currenttime );
	
		ZKASSERT( m_RouteTable[ in_nRouteID ].GetStartTime() != 0 ) ;
		ZKASSERT( m_RouteTable[ in_nRouteID ].GetDuration() > 0 ) ;

		if( currenttime > ( m_RouteTable[ in_nRouteID ].GetStartTime() + 
							m_RouteTable[ in_nRouteID ].GetDuration() +
							ROUTE_EXPIRY_GRACE_PERIOD_SECONDS ))
		{
			bRet = true ;
		}
	}

	return bRet ;
}

// Returns 1 on success
// Mostly used for debugging
int
RActiveRouteTable::GetNymInformation(DZkError& out_Err, const ZkACI in_Route, char* out_nymName, FrCryptSig& out_key)
{	

	return frLocalAipGetRouteNym(	&out_Err, 
									m_Context, 
									in_Route,
									out_nymName, 
									&out_key);

}

void
RActiveRouteTable::SetNymInformation( ZkACI in_Route, int in_iNym ) throw ( RZkException )
{
	fr_crypt_pubkey* keySignature = NULL;
	INym* pNym = NULL ;

	// We don't currently allow routes for "user"
	ZKASSERT( in_iNym != USER_NYM_INDEX ) ;

	try
	{
		ZkService::GetINymCentral()->GetNym( IID_INYM, (void**)&pNym, in_iNym );
		ZKSTD::string strName;
		ZKVERIFY(pNym->GetNymName(strName));

		pNym->GetSignatureKey( &keySignature );
		ZKASSERT( keySignature != NULL ) ;

		if ( !frLocalAipSetRouteNym(&m_err, m_Context, in_Route, strName.c_str(), keySignature) )
		{
			throw RZkException(	m_err, 
								"Route::CommonCoreRouting()",
								"frLocalAipSetRouteNym() failed",
								IDS_BUILDROUTE_SIGVERIF,
								strName.c_str());
		}

		m_RouteTable[ in_Route ].SetNymIndex( in_iNym ) ;

	}
	catch(IZkException&)
	{
		frCryptPubKeyFree( &m_err, &keySignature );
		throw;
	}

	frCryptPubKeyFree( &m_err, &keySignature );
}

ZkACI  
RActiveRouteTable::GetActiveAuthenticatedRoute( sockaddr_in * out_pFirstHopAddr /*= NULL*/ ) 
{
	ZKS::Lock_Guard<RRWLock> gate(*this);

	ZkACI retVal = INVALID_ROUTE_ID ;

	for ( size_t i = 0 ; i < MAX_ROUTES ; i++ )
	{
		if (( m_RouteTable[ i ].IsActive() ) &&
			( m_RouteTable[ i ].GetRouteType() == AuthenticatedRoute ))
		{
			retVal = i ;

			if ( out_pFirstHopAddr != NULL )
			{
				*out_pFirstHopAddr = m_RouteTable[ i ].GetFirstAipAddr() ;
			}

			break ;
		}
	}

	return retVal ;
}

ZkACI  
RActiveRouteTable::GetActiveRouteToFreedomServer( 
		uint16_t in_ServerPort,
		sockaddr_in * out_pFirstHopAddr /*= NULL*/ ) 
{
	ZKS::Lock_Guard<RRWLock> gate(*this);

	ZkACI retVal = INVALID_ROUTE_ID ;

	for ( size_t i = 0 ; i < MAX_ROUTES ; i++ )
	{
		if (( m_RouteTable[ i ].IsActive() ) &&
			( m_RouteTable[ i ].GetRouteType() == UnauthenticatedRoute ) && 
			( m_RouteTable[ i ].GetDestinationPort() == in_ServerPort ) )
		{
			retVal = i ;

			if ( out_pFirstHopAddr != NULL )
			{
				*out_pFirstHopAddr = m_RouteTable[ i ].GetFirstAipAddr() ;
			}

			break ;
		}
	}

	return retVal ;
}


RouteType 
RActiveRouteTable::GetRouteType( ZkACI in_RouteID ) const 
{ 
	ZKS::Lock_Guard<RRWLock> gate(* const_cast<RActiveRouteTable*>(this));
	ValidateAci( in_RouteID ) ;
	return m_RouteTable[ in_RouteID ].GetRouteType() ; 
}

uint16_t  
RActiveRouteTable::GetDestinationPort( ZkACI in_RouteID ) const throw(RZkException)
{
	ZKS::Lock_Guard<RRWLock> gate(* const_cast<RActiveRouteTable*>(this));
	ValidateAci( in_RouteID ) ;
	return m_RouteTable[ in_RouteID ].GetDestinationPort() ; 
}

void
RActiveRouteTable::RetireRoute( ZkACI in_RouteID )
{
	// In theory this is actually a "write" operation,
	// but this can safely be done with a read lock.

	ZKS::Lock_Guard<RRWLock> gate(*this);
	ValidateAci( in_RouteID ) ;

	// Normally only an active route would be retired
	ZKASSERT( m_RouteTable[ in_RouteID ].IsActive() ) ;

	m_RouteTable[ in_RouteID ].SetRouteRetired() ;

	ValidateState()	;
}

void
RActiveRouteTable::FlushExpiredRoutes()
{
	for ( size_t i = 0 ; i < MAX_ROUTES ; i++ )
	{
		if ( IsAciValid( i ) )
		{
			if ( IsRouteExpired( i ) )
			{
				ZKCTRACE1( TRACE_ROUTEMANAGER, " An expired route %d found, removing it from table\n", i ) ;

				ZkACI OutgoingAci = 0 ;
				ZKVERIFY( frLocalAipGetOutgoingAci(&m_err, m_Context, i, &OutgoingAci )) ;
 
				DeleteDeadRoute( i, OutgoingAci, true ) ;
			}
		}
	}
}

bool	  
RActiveRouteTable::IPisaFirstHopInaRoute( uint32_t in_IP )
{
	ZKS::Lock_Guard<RRWLock> gate(*this);
	for ( size_t i = 0 ; i < MAX_ROUTES ; i++ )
	{
		if ( IsAciValid( i ) )
		{
			SOCKADDR_IN addrFirstHop;
			GetFirstAipAddr( i, addrFirstHop );
	
			if ( in_IP == addrFirstHop.sin_addr.s_addr )
			{
				return true ;
			}
		}
	}

	return false ;
}

void
RActiveRouteTable::ValidateState()
{
#ifdef _DEBUG

	ZKCTRACE0( TRACE_ROUTEMANAGER, "\n----------------------------Active Route Table Dump-------------------\n" ) ;

	sockaddr_in addr ;
	
	for ( size_t i = 0 ; i < MAX_ROUTES ; i++ )
	{
		if ( m_RouteTable[ i ].GetRouteType() == RouteTypeUnknown )
		{
			// Make sure this Aci doesn't exist in the context
			char * data = NULL ;

			ZKASSERT( 0 == frLocalAipGetExit( &m_err,
								m_Context,
								i,
								&addr,
								&data ) );
			SAFEFREE( data ) ;

			// Invalid route can never be active
			ZKASSERT( !m_RouteTable[ i ].IsActive() ) ;
		}
		else 
		{
			// These routes might not be active yet

			ZKASSERT( ( m_RouteTable[ i ].GetRouteType() == AuthenticatedRoute ) ||
					  ( m_RouteTable[ i ].GetRouteType() == UnauthenticatedRoute ) ) ;


			if ( m_RouteTable[ i ].GetRouteType() == AuthenticatedRoute ) 
			{
				ZKASSERT( m_RouteTable[ i ].GetNymIndex() != USER_NYM_INDEX ) ;
			}

			if ( m_RouteTable[ i ].GetRouteType() == UnauthenticatedRoute ) 
			{
				ZKASSERT( m_RouteTable[ i ].GetDestinationPort() != 0 ) ;
			}

			// Make sure this Aci doesn't exist in the context

			char * data = NULL ;

			ZKASSERT( frLocalAipGetExit( &m_err,
								m_Context,
								i,
								&addr,
								&data ) );

			SAFEFREE( data ) ;

			ZKCTRACE2( TRACE_ROUTEMANAGER,  "Route ACI: %d IdleCounter %lu\n", i, GetRouteIdleCounter( i ) ) ;
			m_RouteTable[ i ].TraceState() ;
		}
	}

	ZKCTRACE0( TRACE_ROUTEMANAGER, "\n---------------------------------------------------\n" ) ;
#endif
}


unsigned long
RActiveRouteTable::GetRouteIdleCounter( const ZkACI in_routeID ) const
{    
	ValidateAci( in_routeID ) ;
    
	if (( m_RouteTable[ in_routeID ].GetRouteState() == RouteActive ) ||
		( m_RouteTable[ in_routeID ].GetRouteState() == RouteRetired ))
	{
		return  m_RouteTable[ in_routeID ].GetRouteIdleCounter() ;
    }
    else
	{
		// This can happen in the period before the route is fully set up
		return 0 ;
	}
}

void 
RActiveRouteTable::SetRouteIdleCounter( const ZkACI in_routeID, const unsigned long counter)
{
    ValidateAci( in_routeID ) ;

    if (( m_RouteTable[ in_routeID ].GetRouteState() == RouteActive ) ||
		( m_RouteTable[ in_routeID ].GetRouteState() == RouteRetired ))
	{
        m_RouteTable[ in_routeID ].SetRouteIdleCounter( counter ) ;
    }
}

void      
RActiveRouteTable::GetIdleRoutes( unsigned long idleSince, ZKSTD::vector<ZkACI> &out_routeList )
{
	ZKS::Lock_Guard<RRWLock> gate(*this);

	for ( ZkACI i = 0 ; i < MAX_ROUTES ; i++ )
	{
        // The condition below is the only thing that keeps idle authenticated routes
        // from being closed as well --ol
		if (( m_RouteTable[ i ].GetRouteType() == UnauthenticatedRoute ) 
            && (m_RouteTable[ i ].GetRouteState() == RouteActive )
            && ( m_RouteTable[ i ].GetRouteIdleCounter() <= idleSince )
            && ( m_RouteTable[ i ].GetRouteIdleCounter() != 0 ) // 0 is just initialized
        )
		{
            out_routeList.push_back(i);
		}
	}
}
