/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef _RACTIVEROUTETABLE_H
#define _RACTIVEROUTETABLE_H


#include <AciInfo.h>
#include <frLocalAip.h>
#include <DZkError.h>
#include <WSyncTypes.h>
#include <WRoute.h>

class RZkException ;


class RActiveRouteTable : public RRWLock
{
public:
	RActiveRouteTable() ;
	virtual ~RActiveRouteTable() ;

	// Prepare the Route Table by providing a shared instance of the
	// FrLocalAipCtx

	void Init( FrLocalAipCtx in_Ctx ) throw( RZkException );

	// Prepare our local state for a route which we will try to active
	// No Network Activity is provoked
	ZkACI InitNewUnauthenticatedRoute( 	
				const uint16_t in_DestPort, 
				int in_Duration,
				WRoute& in_route ) throw( RZkException );

	ZkACI InitNewAuthenticatedRouteForNym( 
			int in_Duration,
			WRoute& in_route,
			int in_iNym
			) throw( RZkException );

	// For use when prolonging a route
	// Caller has to determine the new rule, but
	// for some reason the underlying library wants
	// this call to be made...
	ZkACI InitNewAuthenticatedRouteWithSameExitAip( 
			ZkACI in_RouteToCopy,
			int in_iNym,
			int in_Duration, 
			WRoute& in_route ) throw( RZkException ) ;

	// Call this function once the route has actually been established
	void RouteIsActive( ZkACI in_Route ) throw( RZkException ) ;

	bool IsAciValid( ZkACI in_Route ) const throw( RZkException );

	// To determine if a route is active or not
	bool IsRouteActive( ZkACI in_Route ) const throw( RZkException );
	RouteState GetRouteState( ZkACI in_Route ) const throw( RZkException ) ;

	void GetFirstAipAddr( ZkACI in_ACI, sockaddr_in& out_addr  ) const throw( RZkException ) ;
	void GetExitAipAddr( ZkACI in_ACI, sockaddr_in& out_addr ) const  ;

	void GetExitAipHost( ZkACI in_ACI,  ZKSTD::string& out_str ) const ;

	void GetAipIds(
			const unsigned int in_ACI, 
			WRoute& out_Route ) const ;

	void GetAipIds(
			const unsigned int in_ACI, 
			uint32_t out_AipIds[MAX_AIPS_IN_ROUTE], 
			size_t& out_numAips ) const ;

	time_t GetExpireTime( ZkACI in_ACI ) ;

	// Remove the route from our table.  It is already assumed that the route
	// has been destroyed or has died on its own
	// in_OutgoingACI - this optional parameter allows the caller to provide
	//	an "outgoing ACI" value to confirm.  This prevents us from erasing
	//  the wrong route in certain multi-threading timing situations.
	// in_bNotify - this parameter can be set to notify 
	//	the UI if an route dies - this is particularly important when authenticated route
	// dies
	void DeleteDeadRoute( ZkACI in_ACI, ZkACI in_OutgoingACI = 0, bool in_bNotify = false ) throw( RZkException );

	size_t GetMaxRoutes() const { return MAX_ROUTES ; } 

	bool IsRouteToServer( FrEntIdType in_DestinationEntity, ZkACI * out_pACI ) ;
	bool IsIPFirstHopOfRoute( uint32_t in_IP, ZkACI * out_pACI ) ;

	bool IsRouteExpired(ZkACI in_RouteID) ;

	size_t GetIndexOfNymOwningAuthenticatedRoute( ZkACI in_ACI ) throw ( RZkException ) ;

	// These methods return INVALID_ROUTE_ID if no route currently exists of the 
	// requested type
	// As an optimization it is possible to get the first hop address
	// so that you don't have to access this table again to send a packet
	ZkACI  GetActiveAuthenticatedRoute(sockaddr_in * out_pFirstHopAddr = NULL) ;
	ZkACI  GetActiveRouteToFreedomServer( uint16_t in_ServerPort, sockaddr_in * out_pFirstHopAddr = NULL )  ;

	RouteType GetRouteType( ZkACI in_RouteID ) const  ;
	void	  RetireRoute( ZkACI in_RouteID ) ;


	// Check all active routes to make sure they haven't gone past their
	// expiry time.  We flush them if they have
	void	  FlushExpiredRoutes() ;

	uint16_t  GetDestinationPort( ZkACI in_RouteID ) const throw(RZkException) ;

	bool	  IPisaFirstHopInaRoute( uint32_t in_IP ) ;

	// Returns 1 on success
	int
	GetNymInformation(DZkError& out_Err, const ZkACI in_Route, char* out_nymName, FrCryptSig& out_key);


    unsigned long GetRouteIdleCounter( const ZkACI in_routeID ) const;

    void SetRouteIdleCounter( const ZkACI in_routeID, const unsigned long counter);

	// Get a vector of all unauthenticated route ACIs that have been idle since a specified time.
    //
	// Param idleSince: tick count prior to which routes are considered idle
	// Param out_routeList: vector into which all idle route ACIs will be inserted
    void      GetIdleRoutes( unsigned long idleSince, ZKSTD::vector<ZkACI> &out_routeList );

private:
	FrLocalAipCtx				m_Context;
	mutable DZkError			m_err;

	// m_Context keeps most of the state of routes
	// for us.  however it doesn't have everything we need so we 
	// use this parallel array to track extra data
	// This is a bit, ugly but completely hidden inside this structure
	AciInfo						m_RouteTable[ MAX_ROUTES ] ;


	ZkACI InitNewRoute( 
			int in_Duration, 
			WRoute& in_route ) throw( RZkException ) ;

	void ValidateAci( ZkACI in_Route ) const throw( RZkException ) ;

	void SetNymInformation( ZkACI in_Route, int in_iNym ) throw ( RZkException ) ;

	void ValidateState() ;
} ;




#endif
