// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RClientCentral.cpp

// Realization of the IClientCentral interface

#include "StdAfx.h"

#include "RClientCentral.h"

#include <frCrypt.h>
#include <frNymUtil.h>
#include <zkDateTime.h>
#include <ZKConvert.h>
#include <WTokenDescriptions.h>
#include <ZKConfiguration.h>
#include <INiqs.h>
#include <ICrypto.h>
#include <ZkService.h>
#include <ZKIP.h>
#include <RZkException.h>
#include <string_hlp.h>		// for ExtractDomain() 
#include <IAdBlocker.h>

#define BUFFER_HEADER_SIZE 3

#define CLIENTCENTRAL_PERSIST_VERSION 1

#define DEFAULT_NUM_MINUTES_UNTIL_RE_ASK_PASSWORD	15

RClientCentral::RClientCentral() 
	: m_TokenService(m_RemoteNym)
	, m_clockLastInteraction(clock())
	, m_nMinutesUntilReAskPassword(DEFAULT_NUM_MINUTES_UNTIL_RE_ASK_PASSWORD)
    , m_strMailRelay("127.0.0.1")
{	
	InitDefaultPreferenceValues() ;
}



RClientCentral::~RClientCentral()
{
}



void 
RClientCentral::DumpStateToTraceWindow()
{
#ifdef TRACE_SAVE_NYM_CENTRAL
#ifdef _DEBUG
	ZKTRACE0( "********************DUMP of RClientCentral STATE*********************\n\n");
	ZKTRACE0( "**RClientCentral REMOTE NYM**\n");

	m_RemoteNym.DebugTraceRemoteNym( NULL ) ;

	// TODO - dump preferences also
	
#endif
#endif
}



void 
RClientCentral::Load( WReadBuffer& io_Buffer ) throw(RZkException)
{
	// Flush all existing state
	m_setWalletDisabledDomains.clear();

	ZKChunkReader cr( io_Buffer ) ;

	if ( cr.GetChunkID() != CLIENT_CENTRAL_CHUNK_ID ) 
	{
		// Corruption or bug in the load code, we must load in the correct order
		ZKASSERT( false ) ;
		throw RZkException("ZKChunkReader::ZKChunkReader", 
							 "Corrupt File",
							 IDS_CORRUPT_FREEDOM_DATA ) ;
	}

	while ( !cr.IsEndOfChunk() )
	{
		ZKChunkReader cr2( cr.GetReadBuffer() );
		switch( cr2.GetChunkID() )
		{
			case( FRREMOTENYM_CHUNK_ID ) :
				LoadRemoteNymChunk( cr2 ) ;
				break ;
				
			case( PREF_CHUNK_ID ) :
				LoadPrefChunk( cr2 ) ;
				break ;

			case( WALLET_DISABLED_DOMAINS_ID ) :
				LoadWalletDisabledDomainsChunk( cr2 ) ;
				break ;

			case( AD_BLOCKER_CHUNK_ID ) :
				LoadAdBlockerChunk( cr2 ) ;
				break ;

			default :
				// Future chunks - just ignore them
				ZKASSERT( false ) ;
				break ;
		}
	}
}

void 
RClientCentral::LoadAdBlockerChunk( ZKChunkReader& io_ChunkReader ) throw(RZkException)
{
	if ( io_ChunkReader.GetChunkID() != AD_BLOCKER_CHUNK_ID ) 
	{
		// Corruption or bug in the load code, we must load in the correct order
		ZKASSERT( false ) ;
		throw RZkException("ZKChunkReader::ZKChunkReader", 
							 "Corrupt File",
							 IDS_CORRUPT_FREEDOM_DATA ) ;
	}

	ZkService::GetIAdBlocker()->SetNumberOfAdsBlocked(io_ChunkReader.ReadInt32());
}

void 
RClientCentral::LoadWalletDisabledDomainsChunk( ZKChunkReader& io_ChunkReader ) throw(RZkException)
{
	if ( io_ChunkReader.GetChunkID() != WALLET_DISABLED_DOMAINS_ID ) 
	{
		// Corruption or bug in the load code, we must load in the correct order
		ZKASSERT( false ) ;
		throw RZkException("ZKChunkReader::ZKChunkReader", 
							 "Corrupt File",
							 IDS_CORRUPT_FREEDOM_DATA ) ;
	}

	while( !io_ChunkReader.IsEndOfChunk() )
	{
		ZKSTD::string strDomain ;
		io_ChunkReader.ReadNonUIString( strDomain ) ;
		m_setWalletDisabledDomains.insert( strDomain ) ;

#if defined( TRACE_LOAD_NYM_CENTRAL )
		ZKTRACE1( "Loading Wallet Disabled Domains- %s.", strDomain.c_str() ) ;				
#endif
	}
}

void 
RClientCentral::LoadRemoteNymChunk( ZKChunkReader& io_ChunkReader ) throw(RZkException)
{
	if ( io_ChunkReader.GetChunkID() != FRREMOTENYM_CHUNK_ID ) 
	{
		// Corruption or bug in the load code, we must load in the correct order
		ZKASSERT( false ) ;
		throw RZkException("ZKChunkReader::ZKChunkReader", 
							 "Corrupt File",
							 IDS_CORRUPT_FREEDOM_DATA ) ;
	}

	const uint8_t* pRNymData = NULL ;
	size_t   cntRNymData = 0 ;

	io_ChunkReader.ReadArray( (const void**)&pRNymData, &cntRNymData ) ;

	if (( pRNymData == NULL ) ||
		( cntRNymData == 0 ) )
	{
		ZKASSERT( false ) ;
		throw RZkException("ZKChunkReader::ZKChunkReader", 
							 "Corrupt File",
							 IDS_CORRUPT_FREEDOM_DATA ) ;
	}

	m_RemoteNym.InitializeFromBuffer( pRNymData, cntRNymData ) ;
}

void 
RClientCentral::LoadPrefChunk( ZKChunkReader& io_ChunkReader ) throw(RZkException)
{
	if ( io_ChunkReader.GetChunkID() != PREF_CHUNK_ID ) 
	{
		// Corruption or bug in the load code, we must load in the correct order
		ZKASSERT( false ) ;
		throw RZkException("RClientCentral::LoadPrefChunk", 
							 "Corrupt File",
							 IDS_CORRUPT_FREEDOM_DATA ) ;
	}

	const uint8_t* pPrefData = NULL ;
	size_t   cntPrefData = 0 ;

	io_ChunkReader.ReadArray( (const void**)&pPrefData, &cntPrefData ) ;

	if (( pPrefData == NULL ) ||
		( cntPrefData == 0 ) )
	{
		ZKASSERT( false ) ;
		throw RZkException("ZKChunkReader::ZKChunkReader", 
							 "Corrupt File",
							 IDS_CORRUPT_FREEDOM_DATA ) ;
	}
	// Initialize defaults because the persisted array may be smaller than
	// the current number of supported preference sets
	InitDefaultPreferenceValues() ;
	memcpy(m_aPrefs, pPrefData, (cntPrefData < sizeof(m_aPrefs)) ? cntPrefData : sizeof(m_aPrefs));

	if (!io_ChunkReader.IsEndOfChunk())
	{
		m_nMinutesUntilReAskPassword = io_ChunkReader.ReadInt32();
	}

	if (!io_ChunkReader.IsEndOfChunk())
	{
		io_ChunkReader.ReadNonUIString(m_strMailRelay);
	}
}

void
RClientCentral::SaveWalletDisabledDomainsChunk( WMemBuf& io_Buffer ) throw(RZkException)
{
	ZKChunkWriter cw( io_Buffer, WALLET_DISABLED_DOMAINS_ID, 1 ) ;

	for (ZKSTD::set<ZKSTD::string>::const_iterator it = m_setWalletDisabledDomains.begin();
		it != m_setWalletDisabledDomains.end();
		++it)
	{
		cw.WriteNonUIString( (*it) ) ;
#ifdef TRACE_SAVE_NYM_CENTRAL
		ZKTRACE1( "Saving RClientCentral Wallet dsabled domain- %s.\n\n", (*it).c_str());
#endif
	}
}

void
RClientCentral::SaveAdBlockerChunk( WMemBuf& io_Buffer ) throw(RZkException)
{
	ZKChunkWriter cw( io_Buffer, AD_BLOCKER_CHUNK_ID, AD_BLOCKER_CHUNK_VERSION) ;

	cw.WriteUInt32(ZkService::GetIAdBlocker()->GetNumberOfAdsBlocked());

}

void
RClientCentral::SaveRemoteNymChunk( WMemBuf& io_Buffer ) throw(RZkException)
{
	ZKChunkWriter cw( io_Buffer, FRREMOTENYM_CHUNK_ID, 1 ) ;
	
	WAutoBuffer ab ;
	m_RemoteNym.GetBuffer( &ab, ab.GetSizePointer() ) ;

	ZKASSERT( ab.Size() > 0 ) ;

	cw.WriteArray( ab, ab.Size() ) ;


#ifdef TRACE_SAVE_NYM_CENTRAL
	ZKTRACE0( "Saving RClientCentral remoteNym\n\n");

	m_RemoteNym.DebugTraceRemoteNym( NULL ) ;
#endif
}

void
RClientCentral::SavePrefChunk( WMemBuf& io_Buffer ) throw(RZkException)
{
	ZKChunkWriter cw( io_Buffer, PREF_CHUNK_ID, 1 ) ;	
	cw.WriteArray( m_aPrefs, sizeof( m_aPrefs ) ) ;
	cw.WriteInt32( m_nMinutesUntilReAskPassword) ;
    cw.WriteNonUIString( m_strMailRelay );
}

void
RClientCentral::Save( WMemBuf& io_Buffer ) throw(RZkException)
{
	ZKChunkWriter cw( io_Buffer, CLIENT_CENTRAL_CHUNK_ID, CLIENTCENTRAL_PERSIST_VERSION ) ;
	SaveRemoteNymChunk( io_Buffer ) ;
	SavePrefChunk( io_Buffer ) ;
	SaveAdBlockerChunk( io_Buffer ) ;
	if (m_setWalletDisabledDomains.size() > 0)
	{
		SaveWalletDisabledDomainsChunk( io_Buffer );
	}
}

void 
RClientCentral::InitDefaultPreferenceValues()
{
	// Unless otherwise specified all values default to OFF
	memset( m_aPrefs, 0, sizeof( m_aPrefs ) );

	// Certain values to ON
	// This only affect new Nym Files, not existing Nym Files

	SetPreference( ZK_CLIENT_PREF_SET, PREF_SCAN_NYMNAMES, ZK_TRUE ) ;
	SetPreference( ZK_CLIENT_PREF_SET, PREF_SCAN_IP, ZK_TRUE ) ;
	SetPreference( ZK_CLIENT_PREF_SET, PREF_SCAN_EMAIL_ADDR, ZK_TRUE ) ;
	SetPreference( ZK_FIREWALL_PREF_SET, PREF_BLOCK_NETBIOS, ZK_TRUE ) ;
	SetPreference( ZK_CLIENT_PREF_SET, PREF_NO_BLOCK_ADS, ZK_TRUE );
	SetPreference(ZK_CLIENT_PREF_SET, PREF_NO_USER_COOKIE_MANAGEMENT, ZK_TRUE);
}

void 
RClientCentral::SetPreference(
	unsigned long in_dwPrefSet, 
	unsigned long in_dwProperty, 
	bool in_bNewState)
{
	if (in_dwPrefSet >= ZK_NUM_PREFERENCE_SETS )
	{
		ZKASSERT( false );
		throw RZkException("RClientCentral::SetPreference","Bogus argument",IDS_INVALID_ARG_RANGE);
	}

	if ( in_bNewState ) 
	{
		// Set the value
		m_aPrefs[in_dwPrefSet] |= in_dwProperty;
	}
	else 
	{
		// Make sure value is not set
		m_aPrefs[in_dwPrefSet] &= ~in_dwProperty;
	}
}

bool 
RClientCentral::GetPreference(unsigned long in_dwPrefSet, unsigned long in_dwProperty) const
{
	if (in_dwPrefSet >= ZK_NUM_PREFERENCE_SETS )
	{
		ZKASSERT( false );
		throw RZkException("RClientCentral::GetPreference","Bogus argument",IDS_INVALID_ARG_RANGE);
	}
	
	return (( m_aPrefs[in_dwPrefSet] & in_dwProperty ) != 0 );
}

void 
RClientCentral::SetReturnEmailAddress( const ZKSTD::string& in_strAddress )
{
	ZKFunctionEntry( "RClientCentral::SetReturnEmailAddress" );
	
	int bStatus;

	bStatus = frRemoteNymResetParameterSearch (
					&m_ZkError ,                 
					m_RemoteNym );

	ZKTestAndThrow(bStatus) ;

	// Delete any existing parameter
	frRemoteNymDeleteParameter (
				&m_ZkError  ,                     
				m_RemoteNym ,                  
				NymPrivateEmailAddr );	

	if ( in_strAddress.size() > 0 ) 
	{
		bStatus = frRemoteNymAddParameter(
							&m_ZkError,
							m_RemoteNym,
							NymPrivateEmailAddr,
							(unsigned char *)in_strAddress.c_str(),
							in_strAddress.length()
							);
		ZKTestAndThrow(bStatus);
	}
}


void 
RClientCentral::GetReturnEmailAddress( ZKSTD::string& out_strAddress ) const
{
	ZKFunctionEntry( "RClientCentral::GetReturnEmailAddress" );
	
	ZKVERIFY( frRemoteNymResetParameterSearch (
						&m_ZkError ,                 
						m_RemoteNym ) );

	const unsigned char * strData = NULL;
	size_t szstrData = 0;

	if(frRemoteNymGetParameter (
		&m_ZkError ,
		m_RemoteNym ,
		NymPrivateEmailAddr ,
		&strData ,
		&szstrData ))
	{			
		out_strAddress = (char*)strData;
	}
	else 
	{
		out_strAddress = "";
	}	
}

ITokenService* 
RClientCentral::GetTokenService() 
{
	ITokenService * pTokenService = NULL ;
	m_TokenService.GetInterface( &pTokenService );

	ZKASSERT( pTokenService != NULL ) ;
	return pTokenService ;
}



//
// Old Persistence Code
// Load code will need to stay around for continuing to read old v1.1 freedom.nym format
// Save code can be phased out once we are confident with new format.

bool 
RClientCentral::GetPreferenceBuffer_Obsolete( 
	unsigned char** out_ppBuffer, 
	size_t* out_cntBuffer ) const
{
	if ( out_ppBuffer == NULL ) 
	{
		ZKASSERT( false ) ;
		return false ;
	}

	*out_cntBuffer = sizeof( m_aPrefs ) ;
	*out_ppBuffer = (unsigned char*) malloc( sizeof( m_aPrefs ) ) ;
	memcpy( *out_ppBuffer, m_aPrefs, sizeof( m_aPrefs ) ) ;

	return true ;
}

bool 
RClientCentral::GetBuffer_Obsolete( 
	unsigned char** out_ppBuffer, 
	size_t* out_cntBuffer ) const
{
	return m_RemoteNym.GetBuffer( out_ppBuffer, out_cntBuffer ) ;
}

bool 
RClientCentral::InitializePreferencesFromBuffer_Obsolete( 
	const unsigned char* in_pBuffer, 
	size_t in_cntBuffer )
{	
	if (( in_pBuffer == NULL ) || 
		( in_cntBuffer == 0 )) 
	{
		ZKASSERT( false ) ;
		return false ;
	}

	// Initialize defaults because the persisted array may be smaller than
	// the current number of supported preference sets
	InitDefaultPreferenceValues() ;
	memcpy(m_aPrefs, in_pBuffer, (in_cntBuffer < sizeof(m_aPrefs)) ? in_cntBuffer : sizeof(m_aPrefs));

	return true ;
}

bool 
RClientCentral::InitializeFromBuffer_Obsolete( 
		const unsigned char* in_pBuffer, 
		size_t in_cntBuffer )
{	
	if (( in_pBuffer == NULL ) || 
		( in_cntBuffer == 0 )) 
	{
		ZKASSERT( false ) ;
		return false ;
	}

	return m_RemoteNym.InitializeFromBuffer( in_pBuffer, in_cntBuffer ) ;
}

void RClientCentral::SetIsWalletDisabledForDomain(	const ZKSTD::string& in_strDomain,
													bool in_bDisable)
{
	ZKS::Lock_Guard<WZkMutex> gate(m_csWalletDisabledDomains);

	ZKSTD::string strDomain(in_strDomain);
	ZKS::MakeLower(strDomain);

	ZKSTD::set<ZKSTD::string>::iterator it = m_setWalletDisabledDomains.find(strDomain);
	if (in_bDisable && (it == m_setWalletDisabledDomains.end()))
	{
		m_setWalletDisabledDomains.insert(strDomain);
	}
	else if (!in_bDisable && (it != m_setWalletDisabledDomains.end()))
	{
		m_setWalletDisabledDomains.erase(it);
	}
}
	
bool RClientCentral::IsWalletDisabledForDomain(const ZKSTD::string& in_strDomain) const
{
	ZKS::Lock_Guard<WZkMutex> gate(m_csWalletDisabledDomains);

	ZKSTD::string strDomain(in_strDomain);
	ZKS::MakeLower(strDomain);
		
	return (m_setWalletDisabledDomains.find(strDomain) != m_setWalletDisabledDomains.end());
}

void RClientCentral::SetDomainsForWhichWalletIsDisabled(
			const ZKSTD::set<ZKSTD::string>& in_strSetOfDomains)
{
	ZKS::Lock_Guard<WZkMutex> gate(m_csWalletDisabledDomains);

	m_setWalletDisabledDomains = in_strSetOfDomains;
}

void RClientCentral::GetDomainsForWhichWalletIsDisabled(
			ZKSTD::set<ZKSTD::string>& out_strSetOfDomains) const
{
	ZKS::Lock_Guard<WZkMutex> gate(m_csWalletDisabledDomains);

	out_strSetOfDomains = m_setWalletDisabledDomains;
}

bool RClientCentral::IsPasswordRequired()
{
	ZKS::Lock_Guard<WZkMutex> gate(m_csClock);

	bool bRet = true;

	const clock_t now = clock();
	const float nDeltaMinutes = (float) (now - m_clockLastInteraction) / (CLOCKS_PER_SEC * 60);
	if ((nDeltaMinutes < (float) m_nMinutesUntilReAskPassword) 
		|| (ZK_CLIENT_NEVER_RE_ASK_PASSWORD == m_nMinutesUntilReAskPassword))
	{
		bRet = false;
		m_clockLastInteraction = now;
	}

	return bRet;
}

void RClientCentral::CorrectPasswordEntered()
{
	ZKS::Lock_Guard<WZkMutex> gate(m_csClock);
	m_clockLastInteraction = clock();
}

int  RClientCentral::GetNumMinutesUntilReAskPassword()
{
	ZKS::Lock_Guard<WZkMutex> gate(m_csClock);
	return m_nMinutesUntilReAskPassword;
}

void RClientCentral::SetNumMinutesUntilReAskPassword(const int in_nMinutes)
{
	if ((in_nMinutes >= 0) || (in_nMinutes == ZK_CLIENT_NEVER_RE_ASK_PASSWORD))
	{
		ZKS::Lock_Guard<WZkMutex> gate(m_csClock);
		m_nMinutesUntilReAskPassword = in_nMinutes;
	}
	else
	{
		// Invalid in_nMinutes.
		ZKASSERT(false);
	}
}

void RClientCentral::SetMailRelay(const ZKSTD::string& in_strMailRelay)
{
    m_strMailRelay = in_strMailRelay;
}

void RClientCentral::GetMailRelay( ZKSTD::string& out_strMailRelay) const
{
    out_strMailRelay = m_strMailRelay;
}
