// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom(R) code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "StdAfx.h"
#include "RClientInfo.h"

#include <IZkRegistry.h>
#include <ZKConfiguration.h>
#include <string_hlp.h>

RClientInfo::RClientInfo()
{
}

RClientInfo::~RClientInfo()
{
}

void		
RClientInfo::SetClientInfo( const WClientInfo& in_ClientInfo )
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	// Make our own private copy
	m_ClientInfo = in_ClientInfo ;

	// Show contents of Client info
	ZKTRACE1( "CLIENTINFO - ClientID %s\n", m_ClientInfo.GetClientId().c_str() ) ;
	ZKTRACE1( "CLIENTINFO - MOTD %s\n", m_ClientInfo.GetMessageOfTheDay().c_str() ) ;
	ZKTRACE1( "CLIENTINFO - Upgrade Url %s\n", m_ClientInfo.GetUpgradeUrl().c_str() ) ;
	ZKTRACE1( "CLIENTINFO - Serial Url (aka misc flags) %s\n", m_ClientInfo.GetSerialNoUrl().c_str() ) ;
	ZKTRACE1( "CLIENTINFO - Niqs Update (seconds) %d\n", m_ClientInfo.GetNiqsUpdateFrequency() ) ;
	ZKTRACE1( "CLIENTINFO - Current Time %d\n", m_ClientInfo.GetCurrentUTC() ) ;
	ZKTRACE1( "CLIENTINFO - Seconds to Expiry %d\n", m_ClientInfo.ExpiresIn() ) ;	
}

bool	
RClientInfo::ClientInfoIsValid()
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	return m_ClientInfo.IsValid();
}

bool        
RClientInfo::IsFreeVersionAvailable() const
{
	if( IZkRegistry::GetFlag( "Premium" ) ) 
	{
		// We can put a registry key in, for example as part of
		// a CD installation, that will cause us to hide the Free Version
		// option.

		return false ;
	}

	// See bug #5170
	return !IsFlagSetInClientInfo( "no_free" ) ;
}

bool        
RClientInfo::IsAip2ProtocolEnabled() const
{
	if( IZkRegistry::GetFlag( "v2routes" ) ) 
	{
		// We can put a registry key in to force use of the
		// protocol
		return true ;
	}

	// See bug #8395
	return IsFlagSetInClientInfo( "v2routes" ) ;
}

bool
RClientInfo::IsFlagSetInClientInfo( const char * in_pToken ) const
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);

	if (( in_pToken == NULL ) || 
		( strlen( in_pToken ) == 0 ) )
	{
		// Don't match empty word
		return false ;
	}

	ZKSTD::string strToken( in_pToken ) ;
	ZKS::MakeLower( strToken ) ;

	// It is an ugly hack that we override the "SerialNumberUrl"
	// to hold flags for the client

	ZKSTD::string strInfo( m_ClientInfo.GetSerialNoUrl() ) ;
	ZKS::MakeLower( strInfo ) ;

	if ( ZKSTD::string::npos != strInfo.find( strToken ) )
	{
		return true ;		
	}

	return false ;
}


time_t RClientInfo::GetTimeToClientExpiry()
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	return m_ClientInfo.ExpiresIn();
}

ZKSTD::string RClientInfo::GetMessageOfTheDay()
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	return m_ClientInfo.GetMessageOfTheDay();
}

int RClientInfo::GetMessageOfTheDayUrgency()
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	return m_ClientInfo.GetMessageOfTheDayUrgency();
}

ZKSTD::string 
RClientInfo::GetUpgradeUrl()
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	return m_ClientInfo.GetUpgradeUrl();
}

int		
RClientInfo::GetNiqsUpdateFrequency()
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	// Provided we talk to the NIQS this is the server provided
	// frequency.  Otherwise it is the default value.
	int frequency = m_ClientInfo.GetNiqsUpdateFrequency();

	// Allow user to override in registry
	IZkRegistry::GetDWORD( "NiqsUpdateFrequency", (unsigned long&) frequency ) ;

	return frequency ;
}

time_t	
RClientInfo::GetCurrentUTCFromClientInfo()
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	return m_ClientInfo.GetCurrentUTC();
}

void 
RClientInfo::GetFreedomDomain(ZKSTD::string &out_strFreedomDomain)
{
	// #7604 - The client info doesn't always come to us, for example
	// if the NIQS is not available at startup.
	// We need to make some assumptions in that case, which actually
	// means that we don't really have the flexibility to change the domain
	// via the NIQS.


	out_strFreedomDomain = ZKConfiguration::GetNetworkName() ;

	ZKASSERT( !out_strFreedomDomain.empty() ) ;
}


