// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RClientPersist.cpp

// This takes care of client persistence.

#include "StdAfx.h"
#include "RClientPersist.h"
#include "INymCentral.h"
#include "IClientCentral.h"
#include "ICrypto.h"
#include "INetworkPerformance.h"
#include "IFirewall.h"
#include "ZkService.h"

#if defined(WIN32)
#include "RClientPersistOldWinVer.h"
#endif

#include <IZkConsoleMsg.h>
#include <ZKConfiguration.h>
#include <ZKFileConfiguration.h>
#include "IZkFileHelp.h"
#include <ZKDocument.h>
#include <WReadBuffer.h>
#include <ZkErrorHandler.h>
#include <IZKErrorReporter.h>
#include <RZkException.h>

#define PERIODIC_SAVE_TIMER_MS 1000 * 60 * 15
//#define PERIODIC_SAVE_TIMER_MS 1000 * 10			// Useful for debugging

static const char	szTestPhrase[]	= "TheseArentTheDroids" ;
static const std::string g_strFileDesc( "Client Data" ) ;



RClientPersist::RClientPersist() 
	: m_bNewFormat( true )
	, m_bDirtyFlag( false )
	, m_pZkTimer( NULL )
{

#if defined(WIN32)
	// Only windows version has to worry about old persistence
	UpgradePersistV1XHelper::GetDataFileToLoad( m_strLoadFilePath, m_bNewFormat ) ;
#else
	m_strLoadFilePath = ZKFileConfiguration::GetFreedomDataFilePath();

	if ( !IZkFileHelp::DoesFileOrDirExist( m_strLoadFilePath.c_str() ) )
	{
		// If the file doesn't exist we won't try to load it
		m_strLoadFilePath.erase() ;
	}
#endif

	// Turn on periodic save timer

	ZKVERIFY( zkITimerNew( &m_err, 
						   &m_pZkTimer,
						   RClientPersist::PeriodicSaveCallback,
						   PERIODIC_SAVE_TIMER_MS ) ) ;
}

RClientPersist::~RClientPersist() 
{
	if ( m_pZkTimer != NULL )
	{
		ZKVERIFY( zkITimerFree( &m_err, &m_pZkTimer ) );
	}
}

void 
RClientPersist::LoadClientState(
	const std::string& strPassphrase)  throw(RZkException)
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	
	ZKASSERT( !m_strLoadFilePath.empty() ) ;

	if ( m_bNewFormat )
	{
		ZKTRACE1( "Loading new file format %s\n", m_strLoadFilePath.c_str() ) ;
// At the moment load happens too early in the game for console messages...
//		ZKConsoleMsg::ShowMsg( CONS_DEBUG, CONSIMAGE_NORMAL, IDS_CONS_LOAD_NEW_FILEFMT, m_strLoadFilePath.c_str() ) ;
		LoadClientState( strPassphrase, m_strLoadFilePath ) ;
	}
	else
	{
#if defined( WIN32 )
		UpgradePersistV1XHelper::LoadWithOldPersistenceCode( strPassphrase, m_strLoadFilePath ) ;
#endif
	}
}

// Should only be called at initial login since this may take some time
void 
RClientPersist::LoadClientState(
	const std::string& strPassphrase,
	const std::string& in_strFileToLoad ) throw(RZkException)
{
	ZKASSERT( !in_strFileToLoad.empty() )  ;
	
	if(strPassphrase.length() < MIN_PASSPHRASE)
	{
		throw DPassphraseError( "RClientPersist::TestPassPhraseAgainstFile" );
	}

	try
	{
#ifdef TRACE_LOAD_NYM_CENTRAL
		ZKTRACE0( "\n\n---------------------------------------\n"
				 "Attempting Loading Client State\n\n" ) ;
		ZKTRACE1( "File Name- %s\n", in_strFileToLoad.c_str() ) ;
#endif

		ZKDocumentReader dr( in_strFileToLoad.c_str() ) ;

		// This is the network that the user used when saving.
		// Version 2.0 and before did not save this so this may be empty
		ZKSTD::string strNetwork = dr.GetNetworkName() ;

		if ( !strNetwork.empty() && ZKConfiguration::GetNetworkName() != strNetwork )
		{
			// In future we could throw or change the network configuration on the fly
			ZKASSERT( !"File was persisted with a different network that client is configured for" ) ;
		}
		
		const uint8_t * pSalt = NULL ;
		size_t cntSalt = 0 ;

		dr.GetSalt( &pSalt, cntSalt ) ;

		// Tell crypto subsystem what the value of SALT and Passphrase was, let it handle symmetric key creation with it
		ZkService::GetICrypto()->SetNewPassphraseAndSalt(strPassphrase, (void*)pSalt, cntSalt);

		// User might have give the wrong passphrase	
		const uint8_t * pTestBuffer = NULL ;
		size_t cntTestBuffer = 0 ;

		dr.GetTestBuffer( &pTestBuffer, cntTestBuffer ) ;
		TestPassPhraseAgainstFile( pTestBuffer, cntTestBuffer );

		// User might have give the wrong passphrase	
		const uint8_t * pDataBuffer = NULL ;
		size_t cntDataBuffer = 0 ;

		dr.GetDataBuffer( &pDataBuffer, cntDataBuffer ) ;
		DecryptAndInitializeClientState( pDataBuffer, cntDataBuffer ) ;

	}
	catch( DPassphraseError& )
	{
		// Error is already well formed
		throw ;
	}
	catch(IZkException& ex )
	{
		// Take the error and build it into general load failure error
		// with details
		char buf[1024] = { 0 };
		snprintf(buf, 
				 sizeof buf, 
				 "Load Failure on file %s, (%s)", 
				 in_strFileToLoad.c_str(), 
				 ZkErrorHandler::GetIZkErrorReporter()->GetErrorString(ex).c_str() );

		throw RZkException( m_err,
							"RClientPersist::LoadClientState", 
							buf, 
							IDS_LOAD_FREEDOM_DATA_FAILED) ;
	}
}


// Save state into file
void 
RClientPersist::SaveClientState(const char * in_pszFilePath)
{	
	ZKFunctionEntry( "RClientPersist::SaveClientState" );

	ZKDocumentWriter DocWriter ;

	WAutoBuffer    SaltStream ;
	WAutoBuffer    TestStream ;
	WAutoBuffer    DataStream ;

	bool           bSavedCriticalData = false ;

	try 
	{	

		ZKTRACE1( "Saving client state to %s\n", in_pszFilePath ) ;

		// Header Information

		DocWriter.SetHeader( g_strFileDesc, 
							 ZKConfiguration::GetFreedomProductName(), 
							 ZKConfiguration::GetFreedomProductVersion(), 
							 ZKConfiguration::GetFreedomPlatform(),
							 ZKConfiguration::GetNetworkName()) ;

		// Get the Salt from the Crypto Service
		// so that we can save it

		ZkService::GetICrypto()->GetSalt( (void**)&SaltStream, SaltStream.GetSizeRef() ) ;
		DocWriter.SetSalt( SaltStream, SaltStream.Size() ) ;

		// Decrypt a simple text in order to save it

		GenerateAndEncryptTestStream( (void**)&TestStream, TestStream.GetSizeRef() ) ;
		DocWriter.SetTest( TestStream, TestStream.Size() ) ;
		
		GenerateAndEncryptDataStream( (void**)&DataStream, DataStream.GetSizeRef() ) ;
		DocWriter.SetData( DataStream, DataStream.Size() ) ;
		
		// This function actually saves the data to disk
		// For historic reasons the backup file path should be named something 
		// slightly different the filepath + ".backup".
		DocWriter.WriteToFile( in_pszFilePath, ZKFileConfiguration::GetFreedomAutoBackupFilePath().c_str() ) ;

#ifdef TRACE_SAVE_NYM_CENTRAL
		// Doesn't include header size or top level chunk overhead
		uint32_t debug_size = DataStream.Size() + TestStream.Size() + SaltStream.Size() ;
		ZKTRACE1( "File Size (Approximately)(bytes): %d\n", debug_size ) ;
		ZKTRACE0( "Save operation complete successfully"
				  "\n\n--------------------------------------\n" ) ;
#endif

		bSavedCriticalData = true ;	// Rest of data is cache only and not mission critical

		// REVIEW: Because this is just Cache information we do not need to save this every time
		// that freedom is saved.  Instead we can save it when freedom exists as a seperate
		// method on NymCentral.  This will save some time and remove some complexity from
		// this method....
		ZkService::GetICrypto()->GetKeyCacheIntoFile(ZKFileConfiguration::GetKeyCacheFilePath().c_str()) ;
	}
	catch( IZkException& ex )
	{
		if ( !bSavedCriticalData ) 
		{
			// Save operation was not successful, we have to inform	callers
			// This adds a base message describing what we were doing, but
			// retains the details of the throw.

			// Review - I format message with nothing because the same
			// string is used in the OldWinVer.  This can be cleaned
			// once that persistence code is put to rest permanently.
			char buf[1024] = { 0 };
			snprintf(buf, sizeof buf, "Save Failure:%s", ZkErrorHandler::GetIZkErrorReporter()->GetErrorString(ex).c_str() );
			throw RZkException( m_err,
								"RClientPersist::SaveClientState", 
								buf, 
								IDS_FREEDOM_SAVE_FAIL ) ;

		}	
	}
}

// Save state into file
void 
RClientPersist::SaveClientState()
{	
	// Prevent concurrent saves (see #6467)
	// In general only the a single thread should be in charge of
	// saving client state but it is not possible to enforce
	// this completely.  For example on linux the periodic
	// save timer is running in its own thread and can conflict
	// with other saves - Andrew
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);

	// We save even if m_bDirtyFlag is not set to avoid possible
	// bugs resulting in data losses.

	SaveClientState( ZKFileConfiguration::GetFreedomDataFilePath().c_str() ) ;

	m_bDirtyFlag = false ;
}

bool 
RClientPersist::IsFirstTimeRun()
{
	// If we didn't determine a file to load then it must not exist
	return ( m_strLoadFilePath.empty() ) ;
}

void 
RClientPersist::GetActiveFilePath( std::string& out_fname )
{
	out_fname = ZKFileConfiguration::GetFreedomDataFilePath().c_str();
}

void	
RClientPersist::GenerateAndEncryptTestStream( void** out_ppStream, size_t& out_cntStream )
{
	// Encrypt a known string
	EncryptBuffer( szTestPhrase, strlen(szTestPhrase) + 1, out_ppStream, out_cntStream ) ;
}

void	
RClientPersist::GenerateAndEncryptDataStream( void** out_ppStream, size_t& out_cntStream )
{
	// Gather the data into out stream

	// All Data will be streamed into this in memory array
	WMemBuf mb ;

	ZkService::GetIClientCentral()->Save( mb ) ;
	ZkService::GetINymCentral()->Save( mb ) ;
	ZkService::GetINetworkPerformance()->Save( mb ) ;
	ZkService::GetIFirewall()->Save( mb ) ;

	// Get acccess to the clear text buffer with all the chunked data in it

	unsigned char * pRawBuffer = NULL ;
	size_t cntBuffer = 0 ;

	zkMemBufGetData( &m_err, mb, (void**)&pRawBuffer, &cntBuffer ) ;

	ZKASSERT( pRawBuffer != NULL ) ;
	ZKASSERT( cntBuffer > 0 ) ;

	// Encrypt the buffer so that it can be saved safely on disk
	EncryptBuffer( pRawBuffer, cntBuffer, out_ppStream, out_cntStream ) ;
}


void
RClientPersist::DecryptAndInitializeClientState( 
	const uint8_t * in_pBuffer,
	size_t in_cntBuffer ) 
{
	WAutoBuffer    DecryptedText ;
	DecryptBuffer( in_pBuffer, in_cntBuffer, (void**)&DecryptedText, DecryptedText.GetSizeRef() ) ;

	// Load up our objects based on the decrypted data

	// We should have put a master chunk at this level so that we can
	// support versioning at the top level.  This means that we don't have
	// as much flexibility for adding new chunks at this level

	WReadBuffer rb( DecryptedText, DecryptedText.Size() ) ;

	ZkService::GetIClientCentral()->Load( rb ) ;
	ZkService::GetINymCentral()->Load( rb ) ;

	if ( !rb.IsEOF() )
	{
		// This chunk was added later
		ZkService::GetINetworkPerformance()->Load( rb ) ;
	}

	if ( !rb.IsEOF() )
	{
		// This chunk was added even later
		ZkService::GetIFirewall()->Load( rb ) ;
	}
}


void
RClientPersist::TestPassPhraseAgainstFile( 
	const uint8_t * in_pTestBuffer,
	size_t in_cntTestBuffer ) 
{
	WAutoBuffer    DecryptedText ;

	DecryptBuffer( in_pTestBuffer, in_cntTestBuffer, (void**)&DecryptedText, DecryptedText.GetSizeRef() ) ;

	if ( strcmp((const char*)(const unsigned char*)DecryptedText, szTestPhrase) != 0 ) 
	{
		throw DPassphraseError( "RClientPersist::TestPassPhraseAgainstFile" );
	}
}

void 
RClientPersist::EncryptBuffer(
	const void * in_pClearBuffer, 
	size_t in_cntClearSize,
	void ** out_ppBuffer,
	size_t& out_cntBuffer ) 
{
	ZKFunctionEntry( "RClientPersist::EncryptBuffer" );
	ZKTestPointer( in_pClearBuffer ) ;
	ZKTestPointer( out_ppBuffer ) ;

	int bStatus = ZkService::GetICrypto()->SymEncryptBuf(in_pClearBuffer, 
													   in_cntClearSize, 
													   out_ppBuffer, 
													   out_cntBuffer) ;
	ZKTestAndThrow(bStatus);
}


void 
RClientPersist::DecryptBuffer(
	const void * in_pEncryptedBuffer, 
	size_t in_cntEncryptedSize,
	void ** out_ppBuffer,
	size_t& out_cntBuffer ) 
{
	ZKFunctionEntry( "RClientPersist::DecryptBuffer" );
	ZKTestPointer( in_pEncryptedBuffer ) ;
	ZKTestPointer( out_ppBuffer ) ;

	if ( !ZkService::GetICrypto()->SymDecryptBuf(in_pEncryptedBuffer, 
													   in_cntEncryptedSize, 
													   out_ppBuffer, 
													   out_cntBuffer) )
	{
		throw DPassphraseError( "RClientPersist::TestPassPhraseAgainstFile" );
	}
}

void 
RClientPersist::SetDirty()
{
	m_bDirtyFlag = true ;
}

bool
RClientPersist::IsDirty() 
{ 
	return m_bDirtyFlag ;
}

void ZK_ITIMER_CALLBACK 
RClientPersist::PeriodicSaveCallback(
	void *p, 
	unsigned int i1, 
	unsigned int i2, 
	unsigned long time)
{
	// As a static method we don't have the "this" state
	// so we must communicate through ZkService

	try
	{
		if ( ZkService::GetIClientPersist()->IsDirty() )
		{
			IZkConsoleMsg::ShowMsg( IDS_CONS_BCKGRND_SAVE ) ;
			ZkService::GetIClientPersist()->SaveClientState() ;
		}
	}
	catch( IZkException& )
	{
		ZKTRACE0( "Periodic save failed" ) ;
	}
}
