// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RClientPersistOldWinVer.cpp

// See .h for explaination

// This takes care of client persistence.
// This code is _not_ cross platform, it uses Windows Structured
// Storage for the persistence.

// Structured Storage provided ability to have multiple
// independent "streams" with in a file.  This makes
// incremental loading and saving very convenient (but
// we are not yet doing this) and keeps code simple

#include "StdAfx.h"
#include "RClientPersistOldWinVer.h"
#include "INymCentral.h"
#include "INymCookieMgr.h"
#include "IClientCentral.h"
#include "ICrypto.h"
#include "ZkService.h"
#include "resource.h"
#include <comdef.h> // For IStoragePtr autopointer
#include <ZKFileConfiguration.h>
#include "IZkFileHelp.h"
#include <RZkException.h>
#include <ZkStringMgr.h>
#include <MiscMessages.h>
#include <zkCStr.h>


#define MAGIC_NUMBER 0x10101010 // A known number to test loading,
								// can also server as a version number
static const char	szTestPhrase[]	= "TheseArentTheDroids" ;
													  													  
static const WCHAR  wszTempFileExt[] = L"_backup" ;
static const char   szTempFileExt[] = "_backup" ;

static const WCHAR  wszTestStream[] = L"Test" ;
static const WCHAR  wszClientCentralStream[] = L"NymCentral" ;
static const WCHAR  wszNymStream[] = L"Nym%d" ;
static const WCHAR  wszPreferencesStream[] = L"Preferences" ;
static const WCHAR	wszSaltStream[] = L"Salt";


// Upgrade files
static const char   szUpdateBackupDir[] = "v11 upgrade data backup\\" ;

RClientPersistOldWinVer::RClientPersistOldWinVer() 
{
	GetDefaultFileLocation() ;
}

RClientPersistOldWinVer::~RClientPersistOldWinVer() 
{
}


// Should only be called at initial login since this may take some time
void 
RClientPersistOldWinVer::LoadClientState(
	const std::string& pass) 
{
	LoadClientState( pass, m_strFilePath ) ;
}

void 
RClientPersistOldWinVer::LoadClientState(
	const std::string& pass,
	const std::string& in_FileToLoad ) 
{

	ZKASSERT( in_FileToLoad.length() > 0 ) ;

	IStoragePtr pFile ;
	
	if(pass.length() == 0)
	{
		TestAndThrowLoadHRESULT(IDS_CANT_LOAD_PASSWORD_INVALID, E_INVALIDARG);
	}

	WCHAR wszFile[ MAX_PATH ];

	MultiByteToWideChar(CP_ACP,  // ansi code page
						0, 
						in_FileToLoad.c_str(), 
						in_FileToLoad.length() + 1, 
						wszFile, 
						MAX_PATH ) ;

	// There are a whole lot of different error conditions that
	// can cause this, user needs messages for some of these

	HRESULT hr = StgOpenStorage( wszFile,
						 NULL, 
						 STGM_READWRITE | STGM_SHARE_EXCLUSIVE, 
						 NULL,
						 0,
						 &pFile ) ;

	TestAndThrowLoadHRESULT(IDS_LOAD_FREEDOM_DATA_FAILED, hr);

	// Read out the SALT from our Freedom.dat
	unsigned char*	pSaltBuffer = NULL;
	unsigned long			cntSaltBuffer = 0;
	IStreamPtr		pStream ;

	hr = pFile->OpenStream(wszSaltStream, 0, STGM_READ | STGM_SHARE_EXCLUSIVE, 0, &pStream ) ;

	if ( SUCCEEDED( hr ) ) 
	{
		hr = GetBufferFromStream(pStream, &pSaltBuffer, cntSaltBuffer) ;
	}
	
	TestAndThrowLoadHRESULT(IDS_CANT_READ_SALT_STREAM, hr);

	// Tell crypto subsystem what the value of SALT and Passphrase was, let it handle symmetric key creation with it
	ZkService::GetICrypto()->SetNewPassphraseAndSalt(pass, pSaltBuffer, cntSaltBuffer);
	SAFEFREE( pSaltBuffer ) ;

	hr = TestPassPhraseAgainstFile(pFile);

	TestAndThrowLoadHRESULT(IDS_CANT_LOAD_PASSWORD_INVALID, hr);

	// Load the Client Central State
	unsigned char* pClientCentralStateBuffer = NULL;
	unsigned long cntSize = 0 ;

	hr = LoadAndDecryptBuffer(pFile, 
							  wszClientCentralStream, 
							  &pClientCentralStateBuffer, 
							  cntSize ) ;
	TestAndThrowLoadHRESULT(IDS_LOAD_FREEDOM_DATA_FAILED, hr);
	ZKVERIFY(ZkService::GetIClientCentral()->InitializeFromBuffer_Obsolete( pClientCentralStateBuffer, cntSize ) ) ;
	free( pClientCentralStateBuffer );

	// Load the preferences

	
	unsigned char* pPreferenceBuffer = NULL;
	cntSize = 0 ;

	HRESULT hrPrefLoad;

	hrPrefLoad = LoadAndDecryptBuffer(pFile, 
							  wszPreferencesStream, 
							  &pPreferenceBuffer, 
							  cntSize) ;

	if ( SUCCEEDED( hrPrefLoad ) ) 
	{
		ZkService::GetIClientCentral()->InitializePreferencesFromBuffer_Obsolete( pPreferenceBuffer, cntSize ) ;
		free(pPreferenceBuffer);
	}
	else 
	{
		// Not fatal
		ZKTRACE0("Failed to load Preferences\n") ;
	}


	ZkService::GetINymCentral()->FlushNyms() ;

	// Step through the Nyms, loading them one by one
	// we don't persist the number of nyms so we go until we can't
	// find anymore in the file
	size_t nIndex = 0 ;

	HRESULT hrNymLoad = S_OK ;
	while( SUCCEEDED( hrNymLoad ) ) 
	{
		hrNymLoad = LoadNym(pFile, nIndex);
		nIndex++;
	}
	ZKASSERT( nIndex == ZkService::GetINymCentral()->GetNumberOfNyms() + 1 ) ;

	return;
}


// Save state into file
void 
RClientPersistOldWinVer::SaveClientState(WCHAR * in_wszFilePath)
{	
	ZKFunctionEntry( "RClientPersistOldWinVer::SaveClientState" );

	IStoragePtr	   pFile ;
	
	unsigned char* pClientCentralStateBuffer = NULL;
	size_t		   cntSize = 0 ;

	void *		   pSalt = NULL ;
	size_t		   cntSaltBuffer = 0 ;

	void *		   pClientPreferences = NULL ;
	size_t		   cntPreferenceBuffer = 0 ;

	
	bool           bSavedCriticalData = false ;

	try 
	{	
		HRESULT hr = StgCreateDocfile(in_wszFilePath,
						  STGM_CREATE | STGM_READWRITE | STGM_SHARE_EXCLUSIVE, 
						  0,
						  &pFile);

		TestAndThrowSaveHRESULT( IDS_FREEDOM_SAVE_FAIL, hr ) ;

		// Save the salt 
		IStreamPtr pStream ;
		hr = pFile->CreateStream( wszSaltStream, STGM_CREATE|STGM_READWRITE| STGM_SHARE_EXCLUSIVE, 0, 0, &pStream ) ;
		TestAndThrowSaveHRESULT( IDS_FREEDOM_SAVE_FAIL, hr );


		// Get the Salt from the Crypto Service
		// so that we can save it


		ZkService::GetICrypto()->GetSalt( &pSalt, cntSaltBuffer ) ;
		WriteBufferToStream( pStream, (unsigned char*)pSalt, cntSaltBuffer);

		// Save a standard phrase encrypted with the password for testing
		WriteTestStreamToFile(pFile) ;
		
		// Save the Nym central state
		ZKVERIFY(ZkService::GetIClientCentral()->GetBuffer_Obsolete( &pClientCentralStateBuffer, &cntSize ));

		if ( pClientCentralStateBuffer == NULL ) 
		{
			ZKASSERT( FALSE );
			ZKThrowRuntime("Unexpected failure", IDS_UNEXPECTED_FAIL);
		}

		EncryptAndSaveBuffer(pFile, wszClientCentralStream, pClientCentralStateBuffer, cntSize );
		
		// Save the preferences
		ZKVERIFY(ZkService::GetIClientCentral()->GetPreferenceBuffer_Obsolete( (unsigned char**) &pClientPreferences, &cntPreferenceBuffer ));

		if (( pClientPreferences != NULL ) &&
			( cntPreferenceBuffer != 0 )) 
		{
			EncryptAndSaveBuffer(pFile, wszPreferencesStream, (unsigned char*)pClientPreferences, cntPreferenceBuffer);
		}
					
		// Step through the Nyms, saving ones that are still alive
		size_t nNymIDForStreamName = 0 ;
		for( size_t nIndex = 0 ; nIndex < ZkService::GetINymCentral()->GetNumberOfNyms() ; nIndex++ ) 
		{
			INym * pNym = NULL ;
			ZkService::GetINymCentral()->GetNym( IID_INYM, (void**)&pNym, nIndex ) ;
			
			if ( pNym != NULL ) 
			{
				unsigned long dwLifeStage = 0 ;
				pNym->GetYourLifeStage( dwLifeStage ) ;

				if ( dwLifeStage != NYM_LIFE_STAGE_RELEASED ) 
				{
					SaveNym( pFile, nIndex, nNymIDForStreamName, pNym );
					++nNymIDForStreamName;
				}
			}
		}

		bSavedCriticalData = true ;	// Rest of data is cache only and not mission critical

		// REVIEW: Because this is just Cache information we do not need to save this every time
		// that freedom is saved.  Instead we can save it when freedom exists as a seperate
		// method on NymCentral.  This will save some time and remove some complexity from
		// this method....
		ZkService::GetICrypto()->GetKeyCacheIntoFile(ZKFileConfiguration::GetKeyCacheFilePath().c_str()) ;
	}
	catch( IZkException& ) 
	{
		SAFEFREE( pClientPreferences ) ;
		SAFEFREE( pClientCentralStateBuffer ) ;
		SAFEFREE( pSalt ) ;		

		if ( !bSavedCriticalData ) 
		{
			// Save operation was not successful, we have to inform	callers
			throw ;
		}
	}

	SAFEFREE( pClientPreferences ) ;
	SAFEFREE( pClientCentralStateBuffer ) ;
	SAFEFREE( pSalt ) ;	
}

// Save state into file
void 
RClientPersistOldWinVer::SaveClientState()
{		
	// We follow a two step saving process to avoid data loss if we crash during a save -
	// worst that can happen is loss of one sessions changes

	WCHAR wszTempFile[ MAX_PATH ];
	char  szTempFile[ MAX_PATH ] ;

	// Prepare Wide Character file path
	MultiByteToWideChar( 0, 0, m_strFilePath.c_str(), m_strFilePath.length() + 1, wszTempFile, MAX_PATH ) ;
	wcscat( wszTempFile, wszTempFileExt ) ;
	
	// Prepare normal char file path
	zkCStrlcpy( szTempFile, m_strFilePath.c_str(), sizeof(szTempFile)) ;
	zkCStrlcat( szTempFile, szTempFileExt, sizeof(szTempFile)) ;

	SaveClientState( wszTempFile ) ;

	// Now copy the successful update over top of the previous version	
	if(!IZkFileHelp::FileCopyWithVerification( szTempFile, m_strFilePath.c_str(), false ) )
	{
		throw RZkException( "RClientPersistOldWinVer::SaveClientState", 
							 "IZkFileHelp::FileCopyWithVerification() failed",
							 IDS_WRITE_CANT_COPY,
							 szTempFile,
							 m_strFilePath.c_str() ) ;
	}
	
}

void 
RClientPersistOldWinVer::TestAndThrowSaveHRESULT(unsigned int in_BaseErrorString, HRESULT in_hr )
{
	if ( SUCCEEDED( in_hr ) ) 
	{
		return ;
	}

	// We can give some vague information about where we are trying to save
	// ASSUMPTION - that we always save into the Freedom directory
  	char szFullPath[MAX_PATH];
	char szDrive[_MAX_DRIVE]; // "C:"

	::GetModuleFileName(NULL, szFullPath, MAX_PATH);
	_splitpath(szFullPath, szDrive, NULL, NULL, NULL);

	char strErr[ 256 ] = { 0 } ;

	switch( in_hr )
	{
		case( STG_E_ACCESSDENIED ) :
		case( STG_E_LOCKVIOLATION ) :
		case( STG_E_SHAREVIOLATION ) :
			ZkStringMgr::Load( strErr, 256, IDS_CANT_SAVE_ACCESSDENIED_OLD ) ;
			break ;

		case( STG_E_MEDIUMFULL ) :
		case( STG_E_WRITEFAULT ) :
			ZkStringMgr::Load( strErr, 256, IDS_CANT_SAVE_MEDIUMFULL_OLD, szDrive ) ;
			break;

		case( STG_E_INSUFFICIENTMEMORY ) :			
			ZkStringMgr::Load( strErr, 256, IDS_CANT_SAVE_NO_MEMORY_OLD ) ;
			break ;
		case( STG_E_INVALIDFLAG ) :			// Programming errors
		case( STG_E_INVALIDPOINTER ) :		
			ZKASSERT( FALSE ) ;
			// Fall through!

		case( STG_E_TOOMANYOPENFILES ) :	// Seems unlikely
		default:
			snprintf( strErr, sizeof(strErr), "The system reported error 0x%x", in_hr );
	}

	throw RZkException("RClientPersistOldWinVer::SaveClientState",strErr, in_BaseErrorString, (const char *)strErr ) ;
}

void 
RClientPersistOldWinVer::TestAndThrowLoadHRESULT( unsigned int in_BaseErrorString, HRESULT in_hr )
{
	if ( SUCCEEDED( in_hr ) ) 
	{
		return ;
	}

	// We can give some vague information about where we are trying to save
	// ASSUMPTION - that we always save into the Freedom directory
  	char szFullPath[MAX_PATH];
	char szDrive[_MAX_DRIVE]; // "C:"

	::GetModuleFileName(NULL, szFullPath, MAX_PATH);
	_splitpath(szFullPath, szDrive, NULL, NULL, NULL);

	char strErr[ 256 ] ;
	strErr[0] = 0 ;

	switch( in_hr )
	{
		case( STG_E_ACCESSDENIED ) :
		case( STG_E_LOCKVIOLATION ) :
		case( STG_E_SHAREVIOLATION ) :		
			ZkStringMgr::Load( strErr, 256, IDS_CANT_LOAD_ACCESSDENIED) ;
			break ;

		case( STG_E_MEDIUMFULL ) :
		case( STG_E_WRITEFAULT ) :
			ZkStringMgr::Load( strErr, 256, IDS_CANT_LOAD_MEDIUMFULL, szDrive ) ;
			break;

		case( STG_E_INSUFFICIENTMEMORY ) :			
			ZkStringMgr::Load( strErr, 256, IDS_CANT_LOAD_NO_MEMORY ) ;
			break ;

		case( STG_E_FILEALREADYEXISTS ) :
			ZkStringMgr::Load( strErr, 256, IDS_CANT_LOAD_FILE_CORRUPT ) ;
			break ;
		case( E_INVALIDARG ) :
			// Means a passphrase error.  We already have enough string
			// in the incoming message
			break ;

		case( STG_E_INVALIDFLAG ) :			// Programming errors
		case( STG_E_INVALIDPOINTER ) :		
			ZKASSERT( FALSE ) ;
			// Fall through!

		case( STG_E_TOOMANYOPENFILES ) :	// Seems unlikely
		default:
			snprintf( strErr, sizeof(strErr), "The system reported error 0x%x", in_hr );
	}

	throw RZkException("RClientPersistOldWinVer::SaveClientState",strErr, in_BaseErrorString, (const char *)strErr ) ;
}

void 
RClientPersistOldWinVer::GetDefaultFileLocation()
{
	m_strFilePath = ZKFileConfiguration::GetOldFreedomDataFilePath();
}

bool 
RClientPersistOldWinVer::IsFirstTimeRun()
{
	if(GetFileAttributes(m_strFilePath.c_str()) == (unsigned long)-1) 
	{
		return true ;
	}

	// File is already present
	return false ;
}

void 
RClientPersistOldWinVer::GetActiveFilePath( std::string& out_fname )
{
	if (m_strFilePath.empty() ) 
	{
		throw RZkException("RClientPersistOldWinVer::GetActiveFilePath","No active file path",IDS_UNEXPECTED_ERROR);
	}

	out_fname = m_strFilePath;
}

HRESULT 
RClientPersistOldWinVer::LoadNym( LPSTORAGE pFile,const size_t nIndex )
{
	HRESULT hr = S_OK ;
	unsigned char* pBuffer=NULL;
	unsigned long ulBufferSize=0;

	if ( pFile==NULL ) 
	{
		ZKASSERT(false);
		return E_INVALIDARG;
	}

	WCHAR wszStreamName[ 12 ] ;
	swprintf( wszStreamName, wszNymStream, nIndex );

	hr = LoadAndDecryptBuffer(pFile, wszStreamName, &pBuffer, ulBufferSize );
		
	if ( SUCCEEDED( hr )) 
	{
		ZkService::GetINymCentral()->AddNym_Obsolete( pBuffer, ulBufferSize ) ;
	}

	free( pBuffer );

	return hr ;
}

void 
RClientPersistOldWinVer::SaveNym(
	LPSTORAGE pFile,
	const size_t in_nNymIndex, 
	const size_t in_nStreamDigitName, 
	INym * in_pNymDebugging )
{
	ZKFunctionEntry( "RClientPersistOldWinVer::SaveNym" );
	ZKTestPointer( pFile ) ;
	ZKTestPointer( in_pNymDebugging ) ;

#ifdef _DEBUG
	std::string strName;
	in_pNymDebugging->GetNymName(strName);
	ZKTRACE1("Saving nym [%s] into persistant storage\n", strName.c_str() );
#endif

	ZKASSERT( in_nNymIndex < ZkService::GetINymCentral()->GetNumberOfNyms() ) ;

	// Determine the stream name for this nym
	// Note, we can't use the nym name as the stream name for security reasons
	WCHAR wszStreamName[ 12 ] ;
	swprintf( wszStreamName, L"Nym%d", in_nStreamDigitName );

	unsigned char* pNymBuffer = NULL;
	size_t cntSize = 0 ;

	if ( !ZkService::GetINymCentral()->GetNymAsBuffer_Obsolete(in_nNymIndex, &pNymBuffer, &cntSize) ) 
	{
		ZKThrowRuntime("Unexpected failure", IDS_UNEXPECTED_FAIL);		
	}		 

	try 
	{
		EncryptAndSaveBuffer(pFile, wszStreamName, pNymBuffer, cntSize ) ;
	}
	catch( IZkException& ) 
	{
		SAFEFREE( pNymBuffer ) ;
		throw ;
	}

	SAFEFREE( pNymBuffer );
}

// Write a short stream encrypted by the password that we will use as a test when
// we load again
void RClientPersistOldWinVer::WriteTestStreamToFile(LPSTORAGE pFile)
{
	EncryptAndSaveBuffer( pFile, wszTestStream, (unsigned char*) szTestPhrase, strlen(szTestPhrase) + 1 );
}

HRESULT 
RClientPersistOldWinVer::TestPassPhraseAgainstFile( LPSTORAGE pFile )
{
	unsigned char * pBuffer = NULL;
	unsigned long ulBufferSize = 0;

	HRESULT hr = LoadAndDecryptBuffer(pFile, wszTestStream, &pBuffer, ulBufferSize );
	
	if ( SUCCEEDED(hr) ) 
	{
		if ( strcmp((char*)pBuffer, szTestPhrase) != 0 ) 
		{
				// Invalid password provided
				hr = E_INVALIDARG ;
		}
	}

	free( pBuffer );
	return hr;
}

void 
RClientPersistOldWinVer::EncryptAndSaveBuffer(
	LPSTORAGE pFile, 
	LPCWSTR wszStreamName, 
	unsigned char* pBuffer, 
	size_t cntSize )
{
	ZKFunctionEntry( "RClientPersistOldWinVer::EncryptAndSaveBuffer" );
	ZKTestPointer( pFile ) ;
	ZKTestPointer( wszStreamName ) ;
	ZKTestPointer( pBuffer ) ;		

	IStreamPtr pStream ;

	// Encrypt a magic string with Passphrase
	unsigned char* pEncrypted = NULL;
	size_t cEncrypted = 0;	

	try 
	{
		int bStatus = ZkService::GetICrypto()->SymEncryptBuf((void*)pBuffer, 
														cntSize, 
														(void**)&pEncrypted, 
														cEncrypted);
		ZKTestAndThrow(bStatus);

		HRESULT hr = pFile->CreateStream( wszStreamName, STGM_CREATE|STGM_READWRITE| STGM_SHARE_EXCLUSIVE, 0, 0, &pStream ) ;
		TestAndThrowSaveHRESULT( IDS_FREEDOM_SAVE_FAIL, hr ) ;

		WriteBufferToStream( pStream, pEncrypted, cEncrypted );
	}
	catch( IZkException& ) 
	{
		free( pEncrypted ) ;
		throw ;
	}

	free( pEncrypted );
}

HRESULT 
RClientPersistOldWinVer::LoadAndDecryptBuffer(
	LPSTORAGE pFile, 
	LPCWSTR wszStreamName, 
	unsigned char** ppBuffer, 
	unsigned long & cntSize )
{
	HRESULT hr;

	if ( pFile==NULL ) 
	{
		ZKASSERT(false);
		return E_INVALIDARG;
	}

	if ( wszStreamName==NULL ) 
	{
		ZKASSERT(false);
		return E_INVALIDARG;
	}

	if ( ppBuffer==NULL ) 
	{
		ZKASSERT(false);
		return E_INVALIDARG;
	}

	// Initialize output
	*ppBuffer = NULL;
	cntSize = 0;

	IStreamPtr pStream ;

	unsigned char* pEncryptedBuffer = NULL;
	unsigned long cntEncryptedBuffer = 0;	

	unsigned char* pDecryptedBuffer = NULL;
	size_t cntDecryptedBuffer = 0;	

	hr = pFile->OpenStream( wszStreamName, 0, STGM_READ| STGM_SHARE_EXCLUSIVE, 0, &pStream ) ;

	if ( SUCCEEDED( hr ) ) 
	{
		hr = GetBufferFromStream(pStream, &pEncryptedBuffer, cntEncryptedBuffer) ;
	}
	
	if (SUCCEEDED( hr )) 
	{
		if(!ZkService::GetICrypto()->SymDecryptBuf((void*)pEncryptedBuffer, 
										cntEncryptedBuffer, 
										(void**)&pDecryptedBuffer, 
										cntDecryptedBuffer)) 
		{
			// Invalid password provided
			hr = E_INVALIDARG ;
		}
	}

	if ( SUCCEEDED(hr) ) 
	{
		*ppBuffer = pDecryptedBuffer ;
		cntSize = cntDecryptedBuffer ;
	}
	else 
	{
		free( pDecryptedBuffer );
	}

	free( pEncryptedBuffer ) ;

	return hr;
}

// Saves a Buffer into a stream with enough formating information that
// we can reliably load it again and detect file damage

HRESULT 
RClientPersistOldWinVer::GetBufferFromStream(
	LPSTREAM pStream, 
	unsigned char** ppBuffer, 
	unsigned long & ulBufferSize)
{
	if ( pStream==NULL ) 
	{
		ZKASSERT( false ) ;
		return false ;
	}

	HRESULT hr = S_OK ;

	// Initialize output
	*ppBuffer=NULL;
	ulBufferSize=0;

	// Check magic number
	unsigned long dwMagic = 0 ;
	unsigned long ulRead = 0 ;

	hr = pStream->Read(&dwMagic, sizeof( unsigned long), NULL);
	
	if (dwMagic != MAGIC_NUMBER) 
	{
		hr = E_FAIL ;
	}

	unsigned long ulBlobSize = 0;
	if (SUCCEEDED(hr)) 
	{
		hr = pStream->Read(&ulBlobSize, sizeof(unsigned long), NULL);
	}

	if (ulBlobSize==0)
	{
		hr=E_FAIL;
	}

	if (SUCCEEDED(hr)) 
	{
		*ppBuffer = ( unsigned char *) malloc( ulBlobSize );
		hr = pStream->Read(*ppBuffer, ulBlobSize, &ulRead);
	}

	if (ulRead != ulBlobSize) 
	{
		// Stream was truncated
		hr = E_UNEXPECTED;
	}

	// TODO: Check ending magic number also

    if (SUCCEEDED(hr))
	{
		ulBufferSize=ulBlobSize;
	}
	else 
	{
		free( *ppBuffer );
		*ppBuffer = NULL ;
	}

	return hr;
}

void 
RClientPersistOldWinVer::WriteBufferToStream(
	LPSTREAM pStream, 
	unsigned char* pBuffer, 
	unsigned long ulBufferSize)
{
	ZKFunctionEntry( "RClientPersistOldWinVer::WriteBufferToStream" );
	ZKTestPointer( pStream ) ;
	ZKTestPointer( pBuffer ) ;

	HRESULT hr = S_OK ;

	// Check magic number
	unsigned long dwMagic = MAGIC_NUMBER ;
	unsigned long ulWrite = 0 ;

	// most likely failure here is running out of disk space condition

	hr = pStream->Write(&dwMagic, sizeof(unsigned long), &ulWrite);
	TestAndThrowSaveHRESULT( IDS_FREEDOM_SAVE_FAIL, hr ) ;

	hr = pStream->Write(&ulBufferSize, sizeof(unsigned long), &ulWrite);
	TestAndThrowSaveHRESULT( IDS_FREEDOM_SAVE_FAIL, hr ) ;

	hr = pStream->Write(pBuffer, ulBufferSize, &ulWrite);
	TestAndThrowSaveHRESULT( IDS_FREEDOM_SAVE_FAIL, hr ) ;

	hr = pStream->Write(&dwMagic, sizeof(unsigned long), &ulWrite);
	TestAndThrowSaveHRESULT( IDS_FREEDOM_SAVE_FAIL, hr ) ;
}

void
UpgradePersistV1XHelper::GetDataFileToLoad( std::string& out_strFilePath, bool& out_bIsNewFormat )
{
	out_strFilePath.erase() ;
	out_bIsNewFormat = true ;

	// First check new format
	
	std::string strNewLocation = ZKFileConfiguration::GetFreedomDataFilePath() ;

	if ( IZkFileHelp::DoesFileOrDirExist( strNewLocation.c_str() ) )
	{
		// We're paranoid that someone might have restored or manually 
		// copied an old format to the new location

		out_strFilePath = strNewLocation ;
		out_bIsNewFormat = !IsDocFile( strNewLocation ) ;
	}
	else 
	{
		std::string strOldLocation = ZKFileConfiguration::GetOldFreedomDataFilePath() ;

		if ( IZkFileHelp::DoesFileOrDirExist( strOldLocation.c_str() ) )
		{
			// We're paranoid that someone might have restored or manually 
			// copied an old format to the new location

			out_strFilePath = strOldLocation ;
			out_bIsNewFormat = !IsDocFile( strOldLocation ) ;
		}
	}
	
}

bool
UpgradePersistV1XHelper::IsDocFile( const std::string& in_strCandidateFile )
{
	WCHAR wszFile[ MAX_PATH ];
	IStoragePtr pFile ;

	MultiByteToWideChar(CP_ACP,  // ansi code page
						0, 
						in_strCandidateFile.c_str(), 
						in_strCandidateFile.length() + 1, 
						wszFile, 
						MAX_PATH ) ;

	// There are a whole lot of different error conditions that
	// can cause this, user needs messages for some of these

	HRESULT hr = StgOpenStorage( wszFile,
						 NULL, 
						 STGM_READWRITE | STGM_SHARE_EXCLUSIVE, 
						 NULL,
						 0,
						 &pFile ) ;

	return ( SUCCEEDED( hr ) ) ;
}

void 
UpgradePersistV1XHelper::LoadWithOldPersistenceCode( 
	const std::string& in_strPassphrase, 
	const std::string& in_strLoadCandidate ) 
{
	// Make a special backup of the file in a place that support could find
	// it
	MakePreUpgradeBackup( in_strLoadCandidate ) ;

	ZKTRACE1( "Loading old file format %s\n", in_strLoadCandidate.c_str() ) ;

//      Should we show a message that the operation is happening?
//		ZKConsoleMsg::ShowMsg( CONS_DEBUG, CONSIMAGE_NORMAL, IDS_CONS_LOAD_OLD_FILEFMT, m_strLoadFilePath.c_str() ) ;

	RClientPersistOldWinVer OldLoadCode ;
	OldLoadCode.LoadClientState( in_strPassphrase, in_strLoadCandidate ) ;

	// Now take care of loading the cookies, but we won't allow this to fail our
	// upgrade

	try
	{
		for( size_t i = 1 ;	// User nym had no cookies for v1.X
			 i < ZkService::GetINymCentral()->GetNumberOfNyms() ;
			 i++ )

		{
			INymCookieMgr* pMgr = NULL ;
			ZkService::GetINymCentral()->GetNym( IID_INYMCOOKIEMGR, (void far**)&pMgr, i ) ;		
			pMgr->LoadOldV1XFormat() ;
		}			
	}
	catch( IZkException& )
	{
	}
}

void 
UpgradePersistV1XHelper::MakePreUpgradeBackup( const std::string& in_strLoadCandidate )
{
	try
	{
		// Backup the file into a new directory
		std::string strBackupDir( ZKFileConfiguration::GetFreedomUserDataDir() ) ;
		strBackupDir += szUpdateBackupDir ;

		if ( !IZkFileHelp::DoesFileOrDirExist( strBackupDir.c_str() ) )
		{
			IZkFileHelp::CreateDirectory( strBackupDir.c_str() ) ;
		}
	
		std::string strBackupFile( strBackupDir.c_str() ) ;
		strBackupFile += ZKFileConfiguration::GetFreedomDataBackupFileName() ;				

		IZkFileHelp::FileCopyWithVerification( in_strLoadCandidate.c_str(), strBackupFile.c_str(), false ) ;
	}
	catch( IZkException& )
	{
		// We don't want to throw just for failing the backup
		ZKTRACE0( "Creation of PreUpgrade backup failed Backup failed\n" ) ;
	} 
}
