// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "StdAfx.h"

#include "RCoreServerConnect.h"
#include "INiqs.h"
#include "IRouteManager.h"
#include "INetworkPerformance.h"
#include "ZkService.h"
#include <IZkConsoleMsg.h>
#include <ZKConvert.h>
#include <RZkException.h>
#include <IZkRegistry.h>
#include <WZkipClient.h>
#include <lockguard.h>
#include <ZKConfiguration.h>

// If there is no "CoreServerNumAttempts" registry key, this is the number of
// attempts SmartConnectThroughCloud() will make.  An attempt fails if either the
// route create or the socket connect fails.
static const size_t DEFAULT_NUM_ATTEMPTS	= 3;

// If this registry key exists, and SmartConnectThroughCloud)( is called with
// the default in_nAttempts, this value will override DEFAULT_NUM_ATTEMPTS.
static const char	REG_CORESERVER_NUM_ATTEMPTS[]		= "CoreServerNumAttempts";

// If this registry key exists, its value will override the timeout passed into
// SmartConnectThroughCloud()
static const char	REG_CORESERVER_CONNECT_TIMEOUT[]	= "CoreServerConnectTimeout";


RCoreServerConnect::RCoreServerConnect()
	: m_nAttempts(DEFAULT_NUM_ATTEMPTS)
	, m_nConnectTimeout(-1)
{
	IZkRegistry::GetDWORD(REG_CORESERVER_NUM_ATTEMPTS, (unsigned long&) m_nAttempts);
	IZkRegistry::GetDWORD(REG_CORESERVER_CONNECT_TIMEOUT, (unsigned long&) m_nConnectTimeout);
}

RCoreServerConnect::~RCoreServerConnect()
{
}

void RCoreServerConnect::SmartConnectThroughCloud(
	const FrEntIdType& in_entType,
	WZkipClient& out_client,
	const int& in_nTimeoutPerConnect,
	const size_t& in_nAttempts /*= UINT32_MAX*/) throw(RZkException) 
{
	bool bConnected = false;
	size_t nAttemptsRemaining = (in_nAttempts == UINT32_MAX) ? m_nAttempts : in_nAttempts;
	int nTimeout = (m_nConnectTimeout == -1) ? in_nTimeoutPerConnect : m_nConnectTimeout;
	bool bRouteBuildWorks = false;

	ZKASSERT(nTimeout > 0);
	if (nAttemptsRemaining == 0)
	{
		ZKASSERT(false);
		throw RZkException("RCoreServerConnect::SmartConnectThroughCloud", 
			"Bogus argument",
			IDS_INVALID_ARG_RANGE);
	}

	WNetInfoDatumVec vecServers;
	ZkService::GetINiqs()->GetServersOfType(in_entType, vecServers, false);

	uint32_t index = 0 ;
	{
		ZKS::Lock_Guard<WZkMutex> gate(m_csMapEntIdType2NextServerIndex);

		if (m_mapEntIdType2NextServerIndex.find(in_entType) != m_mapEntIdType2NextServerIndex.end()
			&& m_mapEntIdType2NextServerIndex[in_entType] < vecServers.size() )
		{
			index = m_mapEntIdType2NextServerIndex[in_entType];
		}
		else
		{
			ZKVERIFY( zkRandGetUInt32InRange(&m_err, 0, vecServers.size() - 1, &index) );
		}
	}

	while (nAttemptsRemaining-- && !bConnected)
	{
		// A route can only exist the first time though this loop because if the connect
		// fails, the route is torn down.  If this route exists, we want to try connecting
		// through it because it is probably good (we connected through it before).
		if (ZkService::GetIRouteManager()->GetUnauthenticatedRoute(
			ZKConvert::GetPortFromEntityType(in_entType))
			|| ZkService::GetIRouteManager()->StartUnauthRoute(ZKConvert::GetPortFromEntityType(in_entType)))
		{
			bRouteBuildWorks = true;

			ZkService::GetINetworkPerformance()->IncrementCoreServerConnectCloudAttempts() ;
			if (!ConnectToServer(vecServers[index].GetSockAddr(), 
												nTimeout,
												m_err,
												out_client))
			{
				ConnectThroughCloudFailed(vecServers[index], m_err);
				index = ++index % vecServers.size();
			}
			else
			{
				bConnected = true;
				ZkService::GetINetworkPerformance()->IncrementCoreServerConnectCloudSuccess() ;
			}
		}
	}
	{
		ZKS::Lock_Guard<WZkMutex> gate(m_csMapEntIdType2NextServerIndex);
		m_mapEntIdType2NextServerIndex[in_entType] = index;
	}

	if (!bConnected)
	{
		if (!bRouteBuildWorks)
		{
			throw RZkException("RCoreServerConnect::SmartConnectThroughCloud", 
				"Failed to build a route.",
				IDS_FAILED_TO_BUILD_ROUTE_TO_CORESERVEER,
				ZKConvert::GetServerNameFromType(in_entType).c_str());
		}
		else
		{
			throw RZkException(m_err,
				"RCoreServerConnect::SmartConnectThroughCloud", 
				"Successfully built route, but failed to connect.",
				IDS_FAILED_TO_CONNECT_TO_CORESERVEER,
				ZKConvert::GetServerNameFromType(in_entType).c_str());
		}
	}
}


bool RCoreServerConnect::ConnectToServer(	const sockaddr_in& in_addrServer, 
											const int& in_nTimeout,
											DZkError& out_err,
											WZkipClient& out_client) const throw()
{
	bool bRet = false;
	int fd = 0;

	if (zkipClientMakeBlocking(&out_err, out_client, in_nTimeout)
		&& zkipClientConnectToAddress(&out_err, out_client, in_addrServer)
		&& zkipClientFd(&out_err, out_client, &fd))
	{
		bRet = true;
	}
	else
	{
		out_client.DisconnectAndReset();
	}
	return bRet;
}

void RCoreServerConnect::ConnectThroughCloudFailed(const WNetInfoDatum& in_server, 
												   const DZkError& in_err) const throw()
{
	IZkConsoleMsg::ShowMsg(CONSIMAGE_COMP_TERM,
		IDS_FAILED_CONNECT_UNAUTH,
		ZKConvert::GetServerNameFromType(in_server.GetEntityType()).c_str(),
		inet_ntoa(in_server.GetSockAddr().sin_addr),
		ZkService::GetIRouteManager()->GetRouteDesc(in_server.GetEntityType()).c_str(),
		in_err.getdetail().c_str());

	ZkService::GetIRouteManager()->TerminateUnauthRoute(
		ZKConvert::GetPortFromEntityType(in_server.GetEntityType()));
}


void RCoreServerConnect::ConnectToNiqs(
			WZkipClient& out_client,
			const int& in_nTimeoutPerConnect,
			const size_t& in_nAttempts ) const throw(RZkException)
{
	// This code originally was found in frNetInfoHandleNiqsConnection

	size_t nAttemptsRemaining = (in_nAttempts == UINT32_MAX) ? m_nAttempts : in_nAttempts;

	if ( nAttemptsRemaining == 0 ) 
	{
		ZKASSERT( !"Please specify a reasonable number of tries" ) ;
		return ;
	}

	bool bConnected = false ;

	WNetInfoDatumVec niqsServers ;
	ZkService::GetINiqs()->GetServersOfType( FRENT_TYPE_NIQS, niqsServers, false ) ;

	// We will attempt to use the NIQS information from the NetInfo 
	// first.  We won't try any more than the size of the buffer
	// and we will leave the last attempt to be on DNS
	size_t nAttemptsFromNiqsInNetInfo = nAttemptsRemaining - 1 ;

	if ( nAttemptsFromNiqsInNetInfo > niqsServers.size() )
	{
		nAttemptsFromNiqsInNetInfo = niqsServers.size() ;
	}

	// In order to support load balancing we start at a random
	// index in the list
	size_t index = 0 ;

	if (!zkRandGetUInt32InRange( &m_err, 0, niqsServers.size() - 1, &index)) 
	{
		/* Call above can fail on the first run of the client because
		   there might not be any entropy yet.  Because random selection of
		   servers is more a question of load balancing instead of 
		   security, and this only will fail the first time client is run,
		   AdamB has given permission for this careful use of rand().
		   See 5811 */
		srand( (unsigned)time( NULL ) );
		index = (rand() % niqsServers.size() );
	}

    struct sockaddr_in saddr;

	while( !bConnected && nAttemptsFromNiqsInNetInfo ) 
	{
		// Wrap around the array
		if ( index >= niqsServers.size() )
		{
			index = 0 ;
		}

		saddr = niqsServers[ index ].GetSockAddr() ;

		ZKSTD::string strEntity ;
		niqsServers[ index ].GetEntityName( strEntity ) ;

		IZkConsoleMsg::ShowMsg( 
			CONS_DEBUG, 
			CONSIMAGE_CONNECTION, 
			IDS_CONS_CONNECT_ATTEMPT_DETAIL,
			strEntity.c_str(),
			inet_ntoa( saddr.sin_addr )) ;

        if( frNiqsConnect(
				&m_err, 
				&out_client,
                inet_ntoa( saddr.sin_addr ),
                FR_DFLT_NIQS_SERVER_PORT,
                in_nTimeoutPerConnect)) 
		{
			bConnected = true ;
			break ;
		}

		nAttemptsFromNiqsInNetInfo-- ;
		index++ ;
	}

	if ( !bConnected )
	{
		ZKSTD::string strNiqsAddress = ZKConfiguration::GetNiqsAddress().c_str() ;

		IZkConsoleMsg::ShowMsg( 
			CONS_DEBUG, 
			CONSIMAGE_CONNECTION, 
			IDS_CONS_CONNECT_ATTEMPT, 
			strNiqsAddress.c_str()) ;

        if ( frNiqsConnect(
				&m_err, 
				&out_client,
				strNiqsAddress.c_str(),
				FR_DFLT_NIQS_SERVER_PORT,
				in_nTimeoutPerConnect ) )
		{
			bConnected = true ;
		}
	}		

	if ( !bConnected )
	{
		throw RZkException(
				m_err, 
				"RCoreServerConnect::ConnectToNiqs()",
				"Connection failed",
				IDS_NIQS_UPDATE_FAILED_NO_CONNECT);
	}
}