// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RCryptoService.cpp: implementation of the RCryptoService class.
//
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "RCryptoService.h"
#include "IKeyQry.h"
#include "ZkService.h"


#include <IZkConsoleMsg.h>
#include <ZKConvert.h>
#include <WCoreTypes.h>
#include <RZkException.h>
#include <ZKFileConfiguration.h>

#include <fr_rand.h>

static const char szDefaultSymKey[] = "StarWars";

RCryptoService::RCryptoService() 
	: m_symkey(NULL) 
	, m_salt(NULL)
	, m_saltsiz(0)
{
	m_strSeedFile = ZKFileConfiguration::GetEntropySeedFilePath() ;
	frCryptInit( NULL, const_cast<char*>( m_strSeedFile.c_str()) );	
}

RCryptoService::~RCryptoService()
{	
	SAFEFREE( m_salt ) ;

	if(m_symkey) { frCryptSymFree(&m_err, &m_symkey); m_symkey = NULL; }

	// Close down the Crypto library.  After this point all crypto operations
	// are illegal
	// The entropy argument is not necessary because yarrow happens to remember
	// the path (see #8108)
	frCryptFinal(NULL, NULL);
}

// Can be used to quickly compare with a previous password
bool 
RCryptoService::GetPassphrase(ZKSTD::string& strPass)
{
	strPass = m_pass;
	return true;
}

void 
RCryptoService::GetDefaultPassphrase(ZKSTD::string& strPass)
{
	strPass = "StarWarsMeetSouthParkMeetTitanic";
}

bool 
RCryptoService::ZkDecryptFile(
	const ZKSTD::string& in_strFile, 
	void*& out_pClearBuf, 
	size_t& out_szClear)
{
	ZKASSERT(in_strFile.length() > 0 ) ;
	ZKASSERT(m_symkey);
	return 0 != frSecDecryptFromFile(&m_err, in_strFile.c_str(), m_symkey, &out_pClearBuf, &out_szClear);
}

bool 
RCryptoService::ZkEncryptFile(const ZKSTD::string& in_strFileToWrite, void* in_buf, size_t in_len)
{
	ZKASSERT( in_strFileToWrite.length() > 0 ) ;
	ZKASSERT( in_buf != NULL ) ;
	ZKASSERT( in_len != 0 ) ;
	ZKASSERT( m_symkey != NULL ) ;

	return 0 != frSecEncryptToFile(&m_err, in_strFileToWrite.c_str(), m_symkey, in_buf, in_len);
}


void 
RCryptoService::ChangePassphrase(
	const ZKSTD::string& strOld, // by old we mean used previously during this session, we don't store the password
	const ZKSTD::string& strNew) throw(RZkException)
	// The only other thing cryptoed with the passphrase
	// are the cookie files, so that will have to be managed
	// after a successful return from this call.

	// Call this the first time with the strOld empty.
{
	ZKASSERT( strNew.length() >= MIN_PASSPHRASE ) ;
	
	ZKSTD::string strVerifyOld;
	GetPassphrase(strVerifyOld);

	if( strVerifyOld == strOld )
	{
		SetNewPassphraseCreateSalt(strNew) ;
	}
	else 
	{
		throw RZkException("RCryptoService::ChangePassphrase","Old Passphrase incorrect",IDS_INCORRECT_PASSPHRASE);
	}
	
}

// Create a sym key but use a SALT passed IN (probably read in from Freedom.nym)
// It will be the symkey used until the password is changed
void
RCryptoService::SetNewPassphraseAndSalt(
	IN const ZKSTD::string& strpass, 
	IN void* in_salt, 
	IN size_t in_cntSalt) throw(RZkException, DPassphraseError)
{

	ZKASSERT( in_salt != NULL ) ;
	ZKASSERT( in_cntSalt > 0 ) ;

	if ( strpass.length() < MIN_PASSPHRASE ) 
	{
		// This isn't going to work, user probably typed in
		// wrong address
		
		throw DPassphraseError( "RCryptoService::SetNewPassphraseAndSalt");
	}

	WSecPassInfo    pinfo ;

	FrCryptSymAlg	symalg = FRCRYPT_SYMALG_BLOWFISH;
	FrCryptSym		symkey = NULL;	

	try
	{

		if(!frSecPassInfoSetFromBuf(&m_err, pinfo, in_salt, in_cntSalt))
		{
			ZKASSERT(0);

			throw RZkException( 
					m_err,
					"RCryptoService::SetNewPassphraseAndSalt",
					"Passphrase is too short",
					IDS_INCORRECT_PASSPHRASE );
		}

		if(!frCryptSymNew(&m_err, &symkey))
		{
			throw RZkException( 
					m_err,
					"RCryptoService::SetNewPassphraseAndSalt",
					"Passphrase is too short",
					IDS_INCORRECT_PASSPHRASE );
		}

		if(!frCryptSymSetAlg(&m_err, symkey, symalg))
		{
			ZKASSERT(0);
			throw RZkException( 
					m_err,
					"RCryptoService::SetNewPassphraseAndSalt",
					"Passphrase is too short",
					IDS_INCORRECT_PASSPHRASE );
		}

		if(!frSecPassVerifySetSymKey(&m_err, strpass.c_str(), symkey, pinfo))
		{
			// This can fail because user entered wrong password
			throw DPassphraseError( "RCryptoService::SetNewPassphraseAndSalt" );
		}
		m_pass = strpass;

		// Erase any existing sym key
		if(m_symkey) 
		{ 
			frCryptSymFree(&m_err, &m_symkey); m_symkey = NULL; 
		}

		m_symkey = symkey;
		symkey = NULL ;

		// Remember the provided salt
		SAFEFREE( m_salt ) ;
		m_salt = malloc( in_cntSalt ) ;
		memcpy( m_salt, in_salt, in_cntSalt ) ;

		m_saltsiz = in_cntSalt ;

	}
	catch(IZkException& )
	{
		if(symkey) 
		{ 
			frCryptSymFree(&m_err, &symkey); 
			symkey = NULL; 
		}

		throw ;
	}

	// Verify expected state
	ZKASSERT( m_salt != NULL ) ;
	ZKASSERT( m_saltsiz != 0 ) ;
	ZKASSERT( m_symkey != NULL ) ;
	ZKASSERT( m_pass.length() >= MIN_PASSPHRASE ) ;
}

// Create a sym key from this new passphrase, caller can then save the SALT as sees fit
void
RCryptoService::SetNewPassphraseCreateSalt(IN const ZKSTD::string& strpass) throw(RZkException)
{
	ZKASSERT( strpass.length() >= MIN_PASSPHRASE ) ;

	FrSecPassInfo	pinfo    = NULL;
	FrCryptSymAlg	symalg   = FRCRYPT_SYMALG_BLOWFISH;
	unsigned int	logiters = 7; // 15 == ~ 3 second wait on a PIII
	FrCryptSym		symkey	 = NULL;	
	
	try
	{
		// Free any existing salt
		SAFEFREE( m_salt ) ;

		if(!frSecPassInfoNew(&m_err, &pinfo) )
		{
			throw RZkException( 
					m_err,
					"RCryptoService::ChangePassphrase",
					"frSecPassInfoNew Failed",
					IDS_INCORRECT_PASSPHRASE );
		}

		if(!frCryptSymNew(&m_err, &symkey))
		{
			throw RZkException( 
					m_err,
					"RCryptoService::ChangePassphrase",
					"frCryptSymNew Failed",
					IDS_INCORRECT_PASSPHRASE );
		}

		if(!frCryptSymSetAlg(&m_err, symkey, symalg))
		{
			ZKASSERT(0);

			throw RZkException( 
					m_err,
					"RCryptoService::ChangePassphrase",
					"frCryptSymSetAlg Failed",
					IDS_INCORRECT_PASSPHRASE );
		}

		if(!frSecPassSetSymKey(&m_err, strpass.c_str(), logiters, symkey, pinfo))
		{
			ZKASSERT(0);

			throw RZkException( 
					m_err,
					"RCryptoService::ChangePassphrase",
					"frSecPassSetSymKey Failed",
					IDS_INCORRECT_PASSPHRASE );
		}

		if(!frSecPassInfoToMem(&m_err, pinfo, &m_salt, &m_saltsiz))
		{
			ZKASSERT(0);

			throw RZkException( 
					m_err,
					"RCryptoService::ChangePassphrase",
					"frSecPassInfoToMem Failed",
					IDS_INCORRECT_PASSPHRASE );
		}

		m_pass = strpass;

		// Transfer the sym key to our member variable
		if(m_symkey) 
		{ 
			frCryptSymFree(&m_err, &m_symkey); 
			m_symkey = NULL; 
		}

		m_symkey = symkey;
		symkey = NULL ;

		if(pinfo) 
		{ 
			frSecPassInfoFree(&m_err, &pinfo); 
			pinfo = NULL; 
		}
	}
	catch(IZkException& )
	{
		if(pinfo) 
		{ 
			frSecPassInfoFree(&m_err, &pinfo); pinfo = NULL; 
		}

		if(symkey) 
		{ 
			frCryptSymFree(&m_err, &symkey); symkey = NULL; 
		}

		throw ;
	}

	// Verify expected state
	ZKASSERT(m_symkey);
	ZKASSERT(m_salt);
	ZKASSERT(m_saltsiz);
	ZKASSERT(pinfo == NULL ) ;
	ZKASSERT(symkey == NULL ) ;
}

bool 
RCryptoService::GetSalt( void ** out_ppSalt, size_t& out_szSalt )
{
	if (( m_salt == NULL ) || 
		( m_saltsiz == 0 )) 
	{
		// Our internal state hasn't been set
		// yet, something is wrong is calling order
		ZKASSERT( false ); 
		return false ;
	}
	
	if ( out_ppSalt == NULL ) 
	{
		ZKASSERT( false ) ;
		return false ;
	}

	*out_ppSalt = malloc( m_saltsiz ) ;
	memcpy( *out_ppSalt, m_salt, m_saltsiz ) ;

	out_szSalt = m_saltsiz ;
	
	return true ;
}

void 
RCryptoService::KeyCacheFromFile(const ZKSTD::string& in_strFileToRead) 
{
	bool bReset = false;
	if(!m_symkey)
	{
		bReset = true;
		m_symkey = (FrCryptSym)szDefaultSymKey;
		ZKTRACE0("Reading cleartext keys.cache\n");
	}

	ZKTRACE2("Getting keys from file [%s] using [%s] passphrase\n", in_strFileToRead.c_str(), m_pass.c_str());

	FILE* istream = fopen(in_strFileToRead.c_str(), "rb" );

	// File might be corrupt or missing
	if( istream ) 
	{
		if ( !frKeyQryCacheRead(&m_err, ZkService::GetIKeyQry()->GetKeyCache(), m_symkey, istream) )
		{
			IZkConsoleMsg::ShowMsg( CONS_VERBOSE, CONSIMAGE_WARNING, IDS_CONS_NO_KEY_CACHE ) ;
		}

		ZKVERIFY(0==fclose(istream));
	}
	else
 	{
 		IZkConsoleMsg::ShowMsg( CONS_VERBOSE, CONSIMAGE_WARNING, IDS_CONS_NO_KEY_CACHE ) ;
	}

	if(bReset) m_symkey = NULL;
}


void 
RCryptoService::GetKeyCacheIntoFile(const ZKSTD::string& strFile) 
{
	if(!m_symkey)
	{
		ZKTRACE0("Not saving key cache file with new keys\n");
		return;
	}

	ZKTRACE2("Saving keys in file [%s] using [%s] passphrase\n", strFile.c_str(), m_pass.c_str());

	FILE* ostream = fopen(strFile.c_str(), "wb");
	ZKASSERT(ostream);

	if(ostream) 
	{ // If this asserts your file is probably read-only
		ZKVERIFY(frKeyQryCacheWrite(&m_err, ZkService::GetIKeyQry()->GetKeyCache(), m_symkey, ostream));
		ZKVERIFY(0==fclose(ostream));
	}
}

// free dst when you're done
// returns 1 on success
bool 
RCryptoService::SymDecryptBuf(
	const void* in_src, 
	const size_t in_ssz, 
	void** out_dst, 
	size_t& out_dsz)
{

	return 0 != frSecDecryptFromBuf(&m_err, const_cast<void*>(in_src), in_ssz, m_symkey, out_dst, &out_dsz);
}


// free dst when you're done
// returns 1 on success
bool 
RCryptoService::SymEncryptBuf(
	const void* in_src, 
	const size_t in_ssz, 
	void** out_dst, 
	size_t& out_dsz)
{
	
	return 0 != frSecEncryptToBuf(&m_err, const_cast<void*>(in_src), in_ssz, m_symkey, out_dst, &out_dsz);
	
}


// You must frCryptPubKeyFree() the FrCryptPubKey when you're done with it
void 
RCryptoService::NewCryptKey(
	const int eKeyVersion,
	const int eKeySize,
	const ZkClock crypt_expiration,	// use ~0 for never, else make a struct tm, and send me the result of mktime() on that struct
	FrCryptPubKey& CryptPubKey,
	fr_entid_exp* ent) const  throw(RZkException)
{
	newKey(FRCRYPT_KEYTYPE_CRYPTION, eKeyVersion, eKeySize, crypt_expiration, CryptPubKey, ent);
}


// You must frCryptPubKeyFree() the FrCryptPubKey when you're done with it
void 
RCryptoService::NewSigKey(
	const int sKeyVersion,
	const int sKeySize,
	const ZkClock sign_expiration,	// use ~0 for never, else make a struct tm, and send me the result of mktime() on that struct
	FrCryptPubKey& SigPubKey,
	fr_entid_exp* ent) const  throw(RZkException)
{
	newKey(FRCRYPT_KEYTYPE_SIGNING, sKeyVersion, sKeySize, sign_expiration, SigPubKey, ent);
}

void 
RCryptoService::newKey(
	const FrPubKeyType kt, 
	const int kv, 
	const int ks, 
	const ZkClock ke, 
	FrCryptPubKey& pk,
	fr_entid_exp* ent) const  throw(RZkException)
{

	
	// Create a new key object, and set various required parameters.
	
	if(!(frCryptPubKeyNew(&m_err, &pk) &&
		  frCryptPubKeySetType(&m_err, pk, kt) &&
		  frCryptPubKeySetSize(&m_err, pk, ks) &&
		  frCryptPubKeySetVersion(&m_err, pk, kv) &&
		  frCryptPubKeySetExpiration(&m_err, pk, ke) &&
		  frCryptPubKeySetEntity(&m_err, pk, ent) )) 
	{
		throw RZkException("RCryptoService::newKey()", 
							"frCryptPubKey...() failed", 
							IDS_FRCRYPT);
	}

	
	//  Generate the bignum parameters.
	

	switch (kt)
	{
		case FRCRYPT_KEYTYPE_CRYPTION:
			if(!(frCryptCryptMakeCrypt(&m_err, pk) &&
					frCryptCryptSetRandomKeys(&m_err, pk)))
			{
				throw RZkException("RCryptoService::newKey()", 
									"frCryptCryptMakeCrypt() or frCryptCryptSetRandomKeys() failed", 
									IDS_FRCRYPT);
			}
			
			break;

		case FRCRYPT_KEYTYPE_SIGNING:
			if(!(frCryptSigMakeSig(&m_err, pk) &&
					frCryptSigSetRandomKeys(&m_err, pk)))
			{
				throw RZkException("RCryptoService::newKey()", 
									"frCryptSigMakeSig() or frCryptSigSetRandomKeys() failed", 
									IDS_FRCRYPT);
			}

			break;

		default:
			ZKASSERT( false ) ;

			break ;
			
	}
}


SigCheckResult 
RCryptoService::CheckSignature( 
		const unsigned char * in_pBuffer,
		size_t          in_szBuffer,
		ZKSTD::string&  out_SigningEntity,
		FrEntIdType&	out_Entity) const throw(RZkException)
{
	ZKFunctionEntry( "RCryptoService::CheckSignature" );
	ZKTestPointer(in_pBuffer);

	FrEntIdExp		Entity;
	unsigned int	version = 0;
	FrCryptSig		sigKey = NULL;
	int				bValid = 0;
	SigCheckResult  Result = SigCheckNotAttempted ;
	int             bStatus = ZK_FALSE ;
	
	// Unfortunately frSec library doesn't use "const"
	// We have to carefully only call functions we know don't
	// touch the buffer to respect the const nature of the buffer
	unsigned char*  pBuffer = const_cast<unsigned char*>( in_pBuffer );

	bStatus = frSecSigGetEntity( &m_err, pBuffer, in_szBuffer, &Entity ) ;
	ZKTestAndThrow(bStatus);

	out_SigningEntity = Entity.name ;

	bStatus = frSecSigGetVersion( &m_err, pBuffer, in_szBuffer, &version) ;
	ZKTestAndThrow(bStatus);

	out_Entity = Entity.type ;

	// Contact the key server or used a cached key
	int bGotKey = ZkService::GetIKeyQry()->GetKey( m_err,
								 out_Entity,	// entity, 
								 3,				// timeout
								 Entity.name, 
								 version , 
								 FRCRYPT_KEYTYPE_SIGNING, 
								 &sigKey ) ;

	if ( bGotKey ) 
	{
		bStatus = frSecSigVerify(&m_err, sigKey, pBuffer, in_szBuffer, &bValid) ;

		if ( !bValid ) 
		{
			bStatus = ZK_FALSE ;
		}

		// Move these messages up
		if ( !bStatus ) 
		{
			// Signature verification failed - message might be forged.
			// We need policy about how to deal with this
			Result = SigCheckFailed ;
		}
		else 
		{
			// Caller wants us to give some information about what happened
			Result = SigCheckSuccessful ;
		}
	}
	else 
	{
		// Inform the user that we couldn't get the key
		Result = SigCheckKeyNotAvail ;
	}

	return Result ;
}


SigCheckResult 
RCryptoService::CheckAndRemoveSignature( 
	unsigned char *		io_pBuffer,
	size_t&				io_szBuffer,
	ZKSTD::string&		out_SigningEntity,
	FrEntIdType&		out_Entity) const throw(RZkException) 
{
	ZKFunctionEntry( "RCryptoService::CheckAndRemoveSignature" );
	ZKTestPointer(io_pBuffer);

	SigCheckResult  Result = SigCheckNotAttempted ;

	Result = CheckSignature( io_pBuffer, io_szBuffer, out_SigningEntity, out_Entity ) ;

	// This call alters our variables returning a buffer without the signature
	ZKVERIFY( frSecSigRemoveInSitu(&m_err, io_pBuffer, &io_szBuffer) ) ;

	return Result ;
}

// Return 1 on success
// Mostly used for debugging
int RCryptoService::PutKeyInFile(DZkError& out_Err, FrCryptPubKey key, FILE *fp)
{
	ZKASSERT(fp);
	ZKASSERT(key);

	return frCryptPubKeyPrint(&out_Err, key, fp);
}


// Return 1 on success
// Mostly used for debugging
int RCryptoService::SymmetricKeyToBuffer(DZkError& out_Err, const FrCryptSym in_sym, char* out_buf, const size_t in_bs)
{
	return frCryptSymDebugDump(&out_Err, in_sym, out_buf, in_bs);
}