// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "StdAfx.h"
#include "RFirewall.h"
#include "IRouteManager.h"
#include "INymCentral.h"
#include "IClientCentral.h"
#include "ZkService.h"
#include "IPacketNotification.h"
#include "IClientCentral.h"
#include <IZkAppState.h>
#include <TraceControl.h>
#include <IZkRegistry.h>
#include <zkPacket.h>			// for ZK_ICMP_ECHO
#include <algorithm>			// stl find functionality
#include <frNetworkDefaults.h>	// for FR_DFLT_AIP_TCP_PORT
#include <ZkNetMisc.h>			// for Cvt_int32toString
#include <PacketHelper.h>
#include <INym.h>
#include <TrafficFilter.h>		// for 	AllowNetBiosInTheClear
#include <ZKChunk.h>
#include <RZkException.h>

#if defined(INTERNAL_BUILD) && defined(WIN32)
#include <ZKBaseUI.h>	// bug 318
#endif

#define	FIREWALL_APP_RULES_CHUNK_VERSION  1
#define FIREWALL_CHUNK_VERSION          1


const unsigned int RFirewall::s_maxSizeOfBackLog = 200;

#if defined(INTERNAL_BUILD) && defined(WIN32)
	ZKSTD::ofstream incoming_packet_drop_file("C:\\incoming_dropped_packets.log");
	ZKSTD::ofstream outgoing_packet_drop_file("C:\\outgoing_dropped_packets.log");
#endif


RFirewall::RFirewall()
	:m_bAtLeastOneOutgoingPacketFromDriver(false)
	,m_bFirewallOn(false)
	,m_bAllowTelnetToCoreServers( false )
	,m_bHttpProxyForUserNym( true )
	,m_pHandler(NULL)
	,m_uPop3Port(LOCAL_POP3_PORT)
	,m_uSmtpPort(LOCAL_SMTP_PORT)
	,m_uHttpPort(LOCAL_HTTP_PORT)
	,m_uNntpPort(LOCAL_NNTP_PORT)
	,m_uIrcPort(LOCAL_IRC_PORT)
	,m_uTelnetPort(LOCAL_TELNET_PORT)
	,m_uSslPort(LOCAL_SSL_PORT)
{
}

RFirewall::~RFirewall()
{
}

bool 
RFirewall::Init()
{
	InitLocalProxyPorts();

	m_bAllowTelnetToCoreServers = IZkRegistry::GetFlag( "AllowTelnetCoreServers" ) ;
	m_bHttpProxyForUserNym		= !IZkRegistry::GetFlag( NO_HTTP_PROX_FOR_USER ) ;

	// ensure the registry values have already been read in at this point for GetNetbiosOption to return the right value
	// (Andrew doesn't understand pervious comment -> is it obsolete? Jan 2001)

	// Instruct the driver's settings based on firewall settings
	// Firewall settings are saved in the nym file, so this Init() method should only execute
	// after it has been loaded.

#ifdef WIN32
	TrafficFilter::theTrafficFilter().CaptureAllIncomingPackets( !GetFirewallPreference(PREF_DISABLE_FIREWALL) ); // see bug 7139
	TrafficFilter::theTrafficFilter().AllowNetBiosInTheClear( !GetFirewallPreference(PREF_BLOCK_NETBIOS) ); // see bug 318
	TrafficFilter::theTrafficFilter().AllowDhcpInTheClear( !GetFirewallPreference(PREF_BLOCK_DHCP) );
	TrafficFilter::theTrafficFilter().AllowRpcInTheClear( !GetFirewallPreference(PREF_BLOCK_RPC) );	
#endif

	return true;
}

void 
RFirewall::SendPacketNotificationsTo( IPacketNotification* in_pHandler)
{
	ZKCTRACE0( TRACE_MYINTAC, "RFirewall getting handler attached to it\n");
	
	ZKASSERT_POINTER(in_pHandler, IPacketNotification);
	ZKASSERT( NULL == m_pHandler);

	m_pHandler = in_pHandler;

	ZKASSERT_POINTER(m_pHandler, IPacketNotification);
}



bool
RFirewall::HandleIncomingPacket(const ZkNetFrame* in_buf)
{
	if( !m_bFirewallOn )
	{
		ZKTRACE0("Firewall mode is off, letting all packets in\n");
		return true;
	}

	if( !m_pHandler )
	{
		ZKTRACE0("Firewall mode is ON but the handler is NULL, letting all packets in, you're probably running Linux\n");
		return true;
	}

	ZKASSERT(in_buf);

	ZKASSERT( !GetFirewallPreference(PREF_DISABLE_FIREWALL) ); // see bug 7139

	if( 4 == in_buf->TCP.IP.version )
	{
		switch(in_buf->TCP.IP.Protocol)
		{
		case IPPROTO_ICMP:			
			{
				const ZkIPHeader* pIpHeader  = (ZkIPHeader*)( &in_buf->TCP );
				const int iIpHdrLen = pIpHeader->ihl << 2;								
				const ZkICMPHeader* pIcmp = reinterpret_cast<ZkICMPHeader*>( ((unsigned char*)&in_buf->TCP) + iIpHdrLen );


				ZKASSERT_POINTER(pIcmp, ZkICMPHeader);
				ZKCTRACE1( TRACE_MYINTAC, "Incoming ICMP packet from [%s]\n",	ZKS::Cvt_int32toString(in_buf->TCP.IP.SrcIpAddress).c_str() );
				ZKCTRACE1( TRACE_MYINTAC, "Destined for [%s]\n",				ZKS::Cvt_int32toString(in_buf->TCP.IP.DstIpAddress).c_str() );
				ZKCTRACE2( TRACE_MYINTAC, "The Icmp type [%s] and code [%s]\n",	ZKS::GetIcmpMessageType( pIcmp->Type).c_str(), 
																				ZKS::GetIcmpMessageCode( pIcmp->Type, pIcmp->code).c_str() );

				const bool bRet = ApplyIncomingIcmpRules(pIcmp);
				if(!bRet)
				{	// dropping packet, log it
					PacketHelper hint(IPPROTO_ICMP, in_buf->TCP.IP.SrcIpAddress, in_buf->TCP.IP.DstIpAddress, pIcmp->Type, pIcmp->code);
					hint.InitDateTime();
					LogAndNotifyDroppedIncomingPacket( hint ) ;
				}
				else
				{
					// accepting packet, ui may care
					PacketHelper hint(IPPROTO_ICMP, in_buf->TCP.IP.SrcIpAddress, in_buf->TCP.IP.DstIpAddress, pIcmp->Type, pIcmp->code);
					NotifyAcceptedIncomingPacket( hint ) ;
				}
				return bRet;
			}
			break;

		case IPPROTO_TCP:
			{

				const ZkIPHeader* pIpHeader  = (ZkIPHeader*)( &in_buf->TCP );
				const int iIpHdrLen = pIpHeader->ihl << 2;								
				const ZkTCPHeader* pTcp = reinterpret_cast<ZkTCPHeader*>( ((unsigned char*)&in_buf->TCP) );

				const unsigned short srcPort = ntohs( in_buf->TCP.SrcPort );
				const unsigned short dstPort = ntohs( in_buf->TCP.DstPort );

#if defined(INTERNAL_BUILD) && defined(WIN32)
				if( GetFirewallPreference(PREF_BLOCK_NETBIOS) )
				{
					ZKASSERT( !ZKS::IsNetBiosPort(srcPort) );
					ZKASSERT( !ZKS::IsNetBiosPort(dstPort) );
				}
#endif

				ZKASSERT_POINTER(pTcp, ZkTCPHeader);
				ZKCTRACE1( TRACE_MYINTAC, "Incoming tcp packet from [%s]\n",	ZKS::Cvt_int32toString(in_buf->TCP.IP.SrcIpAddress).c_str() );
				ZKCTRACE1( TRACE_MYINTAC, "Destined for [%s]\n",				ZKS::Cvt_int32toString(in_buf->TCP.IP.DstIpAddress).c_str() );
				ZKCTRACE2( TRACE_MYINTAC, "Source port [%d] dest port [%d]\n",  srcPort, dstPort);

				const bool bRet = ApplyIncomingTcpRules(pTcp);

				if(!bRet)
				{	// dropping packet, log it
					PacketHelper hint(IPPROTO_TCP, in_buf->TCP.IP.SrcIpAddress, in_buf->TCP.IP.DstIpAddress, srcPort, dstPort);
					hint.InitDateTime();
					LogAndNotifyDroppedIncomingPacket( hint ) ;
				}
				else
				{
					// accepting packet, ui may care
					PacketHelper hint(IPPROTO_TCP, in_buf->TCP.IP.SrcIpAddress, in_buf->TCP.IP.DstIpAddress, srcPort, dstPort);
					NotifyAcceptedIncomingPacket( hint ) ;
				}
				return bRet;
			}
			break;

		case IPPROTO_UDP:
			{
				const ZkIPHeader* pIpHeader  = (ZkIPHeader*)( &in_buf->TCP );
				const int iIpHdrLen = pIpHeader->ihl << 2;								
				const ZkUDPHeader* pUdp = reinterpret_cast<ZkUDPHeader*>( ((unsigned char*)&in_buf->UDP) );

				const unsigned short srcPort = ntohs( in_buf->UDP.SrcPort );
				const unsigned short dstPort = ntohs( in_buf->UDP.DstPort );

#if defined(INTERNAL_BUILD) && defined(WIN32)
				if( GetFirewallPreference( PREF_BLOCK_NETBIOS ) )
				{
					ZKASSERT( !ZKS::IsNetBiosPort(srcPort) );
					ZKASSERT( !ZKS::IsNetBiosPort(dstPort) );
				}
#endif

				ZKASSERT_POINTER(pUdp, ZkUDPHeader);
				ZKCTRACE1( TRACE_MYINTAC, "Incoming udp packet from [%s]\n",	ZKS::Cvt_int32toString(in_buf->TCP.IP.SrcIpAddress).c_str() );
				ZKCTRACE1( TRACE_MYINTAC, "Destined for [%s]\n",				ZKS::Cvt_int32toString(in_buf->TCP.IP.DstIpAddress).c_str() );
				ZKCTRACE2( TRACE_MYINTAC, "Source port [%d] dest port [%d]\n",  srcPort, dstPort);

				const bool bRet = ApplyIncomingUdpRules(pUdp);
			
				if(!bRet)
				{	// dropping packet, log it
					PacketHelper hint(IPPROTO_UDP, in_buf->TCP.IP.SrcIpAddress, in_buf->TCP.IP.DstIpAddress, srcPort, dstPort);
					hint.InitDateTime();
					LogAndNotifyDroppedIncomingPacket( hint ) ;
				}
				else
				{
					// accepting packet, ui may care
					PacketHelper hint(IPPROTO_UDP, in_buf->TCP.IP.SrcIpAddress, in_buf->TCP.IP.DstIpAddress, srcPort, dstPort);
					NotifyAcceptedIncomingPacket( hint ) ;
				}
				return bRet;
			}
			break;

		}; // switch on the protocol
	}
	else
	{
		ZKCTRACE0( TRACE_MYINTAC, "Received a non-IP packet, in it goes\n");
	}

	return true;
}

void
RFirewall::LogAndNotifyDroppedIncomingPacket( PacketHelper &hint )
{
	if( hint.GetDestinationPort() == FR_DFLT_AIP_TCP_PORT )
		return;

	if(m_pHandler)
	{
		// log it
		PushBackLogWithMax(hint) ;

		ZKASSERT_POINTER(m_pHandler, IPacketNotification);

		// and notify if preference is on
		if ( !GetFirewallPreference( PREF_NO_SHOW_BLOCKED_INCOMING ) )
		{	
			m_pHandler->NotifyDroppingIncomingPacket( &hint );
		}
	}
	else
	{
		ZKCTRACE0( TRACE_MYINTAC, "No one to notify, m_pHandler is NULL\n");
	}
}

void
RFirewall::NotifyAcceptedIncomingPacket( PacketHelper &hint )
{
	if(m_pHandler)
	{
		ZKASSERT_POINTER(m_pHandler, IPacketNotification);
		m_pHandler->NotifyAcceptingIncomingPacket( &hint );
	}
	else
	{
		ZKCTRACE0( TRACE_MYINTAC, "No one to notify, m_pHandler is NULL\n");
	}
}


// There are hard Icmp rules. Every type of Icmp message is let in, except 
// 5 - router redirect 
// 9 - router advertisement (someone may be trying a MIM by redirecting you through his router)
// 12 - parameter problem (you may be under attack)
// The
// 8 - echo request 
// is user configurable (see bug 5712) so this machine may be pingable
bool
RFirewall::ApplyIncomingIcmpRules(const ZkICMPHeader* in_pIcmpHeader)
{

	ZKASSERT( m_bFirewallOn );	

	if( ZK_ICMP_ECHO == in_pIcmpHeader->Type )
	{
		// bug 5712
		ZKCTRACE1(TRACE_MYINTAC, 
				  " Someone is trying to ping us - %s\n", 
				  GetFirewallPreference( PREF_BLOCK_ICMP_ECHO ) ? "Blocking" : "Allowing" );
		return !GetFirewallPreference( PREF_BLOCK_ICMP_ECHO ) ;
	}
	else if( ZK_ICMP_REDIRECT == in_pIcmpHeader->Type )
	{
		ZKCTRACE0(TRACE_MYINTAC, " Someone is trying to redirect your default router, returning false\n");
		return false;
	}
	else if( ZK_ICMP_ROUTERADVERT == in_pIcmpHeader->Type )
	{
		ZKCTRACE0(TRACE_MYINTAC, " Someone is trying to advertise a new router, returning false\n");
		return false;
	}
	else if( ZK_ICMP_PARAMPROB == in_pIcmpHeader->Type )
	{
		ZKCTRACE0(TRACE_MYINTAC, " Bad Icmp received, may be an attack, returning false\n");
		return false;
	}
	else
	{
		ZKCTRACE0(TRACE_MYINTAC, "Accepting incoming non-Echo-request Icmp packet\n");
	}
	return true;
}

// Returning true lets the packet in
bool
RFirewall::ApplyIncomingUdpRules(const ZkUDPHeader* pHdr)
{
	ZKASSERT( m_bFirewallOn );	

	if( 0 == pHdr->DstPort )
	{
		return false; // bad parm
	}

	const unsigned short dstPort = ntohs( pHdr->DstPort );
	const unsigned short srcPort = ntohs( pHdr->SrcPort );

	if((ZKS::IsNetBiosPort( dstPort ) || ZKS::IsNetBiosPort( srcPort )) && !GetFirewallPreference(PREF_BLOCK_NETBIOS))
	{
		ZKTRACE1("OK for incoming NetBios UDP packet [%d]\n", dstPort );
		return true;
	}
	else if((ZKS::IsDHCPPort( dstPort) || ZKS::IsDHCPPort( srcPort )) && !GetFirewallPreference(PREF_BLOCK_DHCP) )
	{
		ZKTRACE1("OK for incoming DHCP UDP packet [%d]\n", dstPort );
		return true;
	}
	else if((ZKS::IsRPCPort( dstPort ) || ZKS::IsRPCPort( srcPort )) && !GetFirewallPreference(PREF_BLOCK_RPC) )
	{
		ZKTRACE0("OK for incoming RPC UDP packet\n");
		return true;
	}
	else if (WINDOWS_TCP_AND_UDP_PORT == dstPort || FR_DFLT_KEYQRY_SERVER_PORT == srcPort /* bug 7099 */)
	{
		//ZKTRACE0("WINDOWS_TCP_AND_UDP_PORT coming in normally\n");
		return true;
	}
	else if (srcPort == DNS_PORT)
	{
		ZKTRACE1("Letting UDP DNS in from port [%d]\n", srcPort);
		return true;
	}
	else if(IsLocalUdpPortOpened(dstPort, pHdr->IP.SrcIpAddress, pHdr->SrcPort))
	{
		ZKTRACE1("A UDP packet for port [%d] is being let in, we've seen someone wanting it\n", dstPort );
		return true;
	}
	else if(IsLocalUdpPortInFastSendto(pHdr->IP.SrcIpAddress, pHdr->SrcPort))
	{
		ZKTRACE1("A UDP packet for port [%d] is being let in, we've seen someone wanting it\n", dstPort );
		return true;
	}

	ZKTRACE2("A incoming UDP packet for port [%d] from port [%d] is being dropped cause we ain't seen nothing there\n", dstPort, srcPort );

#if defined(INTERNAL_BUILD) && defined(WIN32)
	ZKS::DumpUdpHeader(pHdr);
	ZKS::DumpUdpHeaderToFile(pHdr, incoming_packet_drop_file);
#endif

	return false;
}

bool
RFirewall::IsLocalUdpPortInFastSendto(
	const unsigned int   in_source_ip, 
	const unsigned short in_source_port)
{
	ZKASSERT (m_bFirewallOn );

	ZKTRACE2("PartOfTcpConnectionExchange with HostSourcePort [%d] HostSourceIP [%d]\n", 
				ntohs(in_source_port),
				in_source_ip);

	ZKS::Lock_Guard<ZKCriticalSection> gate(m_TcpHostVeccs);

	for(HostVec::const_iterator it = m_UdpHostVec.begin(); it != m_UdpHostVec.end(); ++it)
	{
		ZKTRACE3("Looking in TcpHostVec ip [%d] port [%d] LocalPort [%d] (should be 0)\n", 
			(*it).DstIp, 
			ntohs((*it).DstPort), 
			ntohs((*it).LocalPort));

		if( ((*it).DstIp == in_source_ip) && ((*it).DstPort == in_source_port) )
		{
			ZKTRACE2("The IP [%d] is good to receive data to local TCP port [%d]\n", (*it).DstIp, ntohs( (*it).DstPort ) );
			return true;
		}
	}

	ZKTRACE2("Did not find host whom we previously attempted to TCP connect IP [%d] Port [%d]\n", in_source_ip, ntohs(in_source_port) );

	return false;
}


// in_port parameter is in host byte order
// in_source_ip is in NBO
// in_source_port is in NBO
bool
RFirewall::IsLocalUdpPortOpened(
	const unsigned short in_port, 
	const unsigned int   in_source_ip, 
	const unsigned short in_source_port)
{
	ZKTRACE3("Looking up an opened UDP local port [%d] IP [%d] port [%d]\n", in_port, in_source_ip, ntohs(in_source_port) );	

	ZKS::Lock_Guard<ZKCriticalSection> gate(m_UdpPortSetcs);

	PortSet::const_iterator it = ZKSTD::find_if(	
											m_UdpPortSet.begin(),
											m_UdpPortSet.end(),
											rh_equal_LocalPort( htons(in_port) ) 
											);
											
#ifndef NDEBUG
	if( it != m_UdpPortSet.end() )
	{
		ZKTRACE2("in udp port set ip is [%d] and port is [%d]\n", (*it).ip, ntohs((*it).port) );		
	}
#endif

	return (it != m_UdpPortSet.end());
}

// Returning true lets the packet in
bool
RFirewall::ApplyIncomingTcpRules(const ZkTCPHeader* pHdr)
{
	ZKASSERT( m_bFirewallOn );

	if( 0 == pHdr->DstPort )
	{
		return false; // bad parm
	}

	const unsigned short dstPort = ntohs( pHdr->DstPort );
	const unsigned short srcPort = ntohs( pHdr->SrcPort );
	
	if((ZKS::IsNetBiosPort( dstPort ) || ZKS::IsNetBiosPort( srcPort )) && !GetFirewallPreference( PREF_BLOCK_NETBIOS ))
	{
		ZKTRACE1("OK for incoming NetBios TCP packet [%d]\n", dstPort );
		return true;
	}
	else if((ZKS::IsDHCPPort( dstPort) || ZKS::IsDHCPPort( srcPort )) && !GetFirewallPreference( PREF_BLOCK_DHCP ) )
	{
		ZKTRACE1("OK for incoming DHCP TCP packet [%d]\n", dstPort );
		return true;
	}
	else if((ZKS::IsRPCPort( dstPort ) || ZKS::IsRPCPort( srcPort )) && !GetFirewallPreference( PREF_BLOCK_RPC ) )
	{
		ZKTRACE0("OK for incoming RPC TCP packet\n");
		return true;
	}
	else if(ZKS::IsIdentPort( dstPort ) && !GetFirewallPreference( PREF_BLOCK_IDENT ) )
	{
		ZKTRACE0("OK for incoming Identification protocol TCP packet\n");
		return true;
	}
	else if(PartOfTcpConnectionExchange(pHdr))
	{
		return true;
	}
	else if(IsLocalTcpPortOpened(dstPort, pHdr->IP.SrcIpAddress, pHdr->SrcPort))
	{
		ZKTRACE1("A TCP packet for port [%d] is being let in, we've seen someone wanting it\n", dstPort );
		return true;
	}

	ZKTRACE2("A incoming TCP packet for port [%d] from port [%d] is being dropped cause we ain't seen nothing there\n", dstPort, srcPort );

#if defined(INTERNAL_BUILD) && defined(WIN32)
	ZKS::DumptcpHeader(pHdr);
	ZKS::DumptcpHeaderToFile(pHdr, incoming_packet_drop_file);
#endif

	return false;
}

// in_port parameter is in host byte order
// in_source_ip is in NBO
// in_source_port is in NBO
bool
RFirewall::IsLocalTcpPortOpened(
	const unsigned short in_port,
	const unsigned int   in_source_ip, 
	const unsigned short in_source_port)
{
	ZKTRACE3("Looking up an opened TCP local port [%d] IP [%d] port [%d]\n", in_port, in_source_ip, ntohs(in_source_port) );

	ZKS::Lock_Guard<ZKCriticalSection> gate(m_TcpPortSetcs);

	PortSet::const_iterator it = ZKSTD::find_if(	
											m_TcpPortSet.begin(),
											m_TcpPortSet.end(),
											rh_equal_LocalPort( htons(in_port) )
											);

#ifndef NDEBUG
	if( it != m_TcpPortSet.end() )
	{
		ZKTRACE2("in tcp port set ip is [%d] and port is [%d] (These can be zero)\n", (*it).ip, ntohs((*it).port) );
	}
#endif

	return (it != m_TcpPortSet.end());
}

FirewallAction
RFirewall::GetActionForOutgoingTCPPacket( ZkBuffer *in_pbuf ) 
{
	ZKASSERT(IPPROTO_TCP == in_pbuf->Tcp.IP.Protocol) ;

	// The driver calls this when processing packets, so this
	// is our way of detecting if driver isn't working
	m_bAtLeastOneOutgoingPacketFromDriver = true ;

	if( !m_bFirewallOn )
	{
		ZKCTRACE0( TRACE_MYINTAC, "Freedom is loading, allowing normal traffic to leave,");
		ZKCTRACE0( TRACE_MYINTAC, "probably connecting to ISP and getting aip keys along with NIQS info\n");	
		return FirewallAllowOut;
	}
	
	// Determine the port and destination for the packet

	uint16_t DestPort = in_pbuf->Tcp.DstPort;
	uint32_t DestIP   = in_pbuf->Tcp.IP.DstIpAddress;

	// Save everybody down the line the trouble of converting the port into little endian
	const uint16_t hboDestPort = ntohs(DestPort);

	// Look for SYN packets and
	// if the destination ip and port are in our m_TcpHostVec
	// we will add the local port to that piece of information
	// Don't call PartOfTcpConnectionExchange since that is for incoming packets, not outgoing.
	
	const ZkTCPHeader* pTcp = reinterpret_cast<ZkTCPHeader*>( ((unsigned char*)&in_pbuf->Tcp) );

	if(pTcp->syn)
	{
		if(AddLocalPortToTcpHostVecEntry(DestIP, DestPort, pTcp->SrcPort))
		{				
			ZKCTRACE1(TRACE_MYINTAC,"Saving local port during tcp connection [%d]\n", ntohs(pTcp->SrcPort) );
		}
		else
		{
			ZKCTRACE1(TRACE_MYINTAC,"Not Saving local port during tcp connection [%d]\n", ntohs(pTcp->SrcPort) );			
			ZKCTRACE0(TRACE_MYINTAC,"This connection bypassed our LSP or bind()ed previously\n");
		}
	}
	
	// The key function call to decide the fate of this packet
	FirewallAction actionRet = PacketFirewallTCP( DestIP, hboDestPort ) ;
		
#if defined(INTERNAL_BUILD) && defined(WIN32) && defined(TRACE_MYINTAC)
	if(FirewallBlock == actionRet)
	{
		ZKS::DumptcpHeader(&in_pbuf->Tcp);
		ZKS::DumptcpHeaderToFile(&in_pbuf->Tcp, outgoing_packet_drop_file);
	}
#endif
		
	return actionRet;
}

FirewallAction
RFirewall::GetActionForOutgoingUDPPacket( ZkBuffer *in_pbuf )
{
	ZKASSERT(IPPROTO_UDP == in_pbuf->Udp.IP.Protocol) ;

	// The driver calls this when processing packets, so this
	// is our way of detecting if driver isn't working
	m_bAtLeastOneOutgoingPacketFromDriver = true ;

	if( !m_bFirewallOn )
	{
		ZKCTRACE0( TRACE_MYINTAC, "Freedom is loading, allowing normal traffic to leave,");
		ZKCTRACE0( TRACE_MYINTAC, "probably connecting to ISP and getting aip keys along with NIQS info\n");	
		return FirewallAllowOut;
	}	

	// Determine the port and destination for the packet

	uint16_t DestPort = in_pbuf->Udp.DstPort;
	uint32_t DestIP   = in_pbuf->Udp.IP.DstIpAddress;

	// Save everybody down the line the trouble of converting the port into little endian
	const uint16_t hboDestPort = ntohs(DestPort);

	// Handle special case of Fragmented Route Creates
	if(	( ntohs(in_pbuf->Tcp.IP.Flags_FrameOffs) & 0x3FFF) &&
		( FirewallAllowOut == PacketFirewallUDP(
											DestIP, 
											ntohs(FR_DFLT_AIP_UDP_PORT)) ) )
	{
		ZKCTRACE0(TRACE_MYINTAC, "Got an outgoing UDP packet which is fragmented, ");
		ZKCTRACE0(TRACE_MYINTAC, "IT IS a route create or directed to a first hop, letting it go out\n");
		return FirewallAllowOut;
	}
	
	FirewallAction actionRet = PacketFirewallUDP( DestIP, hboDestPort) ;
		
#if defined(INTERNAL_BUILD) && defined(WIN32) && defined(TRACE_MYINTAC)
	if(FirewallBlock == actionRet)
	{
		ZKS::DumpUdpHeader(&in_pbuf->Udp);
		ZKS::DumpUdpHeaderToFile(&in_pbuf->Udp, outgoing_packet_drop_file);
	}
#endif
		
	return actionRet;
}

FirewallAction 
RFirewall::GetActionForOutGoingIcmp(int in_type)
{	
	if( !m_bFirewallOn ) // firewall could be off at startup
	{
		return FirewallAllowOut;
	}

	FirewallAction retAction = FirewallBlock; // block all icmp by default

	switch(in_type)
	{
	case ZK_ICMP_ECHO:
		ZKCTRACE0( TRACE_MYINTAC, "Allowing outgoing icmp ping out\n");
		retAction = FirewallAllowOut;	// allow outgoing pings from this machine, with or without a nym selected
		break;
	case ZK_ICMP_ECHOREPLY:
		ZKCTRACE1(TRACE_MYINTAC, 
				  "Outgoing ICMP echo reply - %s\n",
				  GetFirewallPreference( PREF_BLOCK_ICMP_ECHO ) ? "Blocking" : "Allowing");

		retAction = GetFirewallPreference( PREF_BLOCK_ICMP_ECHO ) ? FirewallBlock : FirewallAllowOut ;

		break;
	default:
		ZKCTRACE1( TRACE_MYINTAC, "Blocking outgoing icmp [%d]\n", in_type);
		break;
	};

	return retAction;
}

FirewallAction 
RFirewall::PacketFirewallTCP(
	const u_long DestIP, 
	const uint16_t in_DestPort ) const
{
	ZKASSERT( m_bFirewallOn );

#ifdef WIN32
	// Windows has different default behavior because the firewall 
	// feature allows control per application
	// REVIEW - this is very ugly way to differentiate the platform
	// behavior
	FirewallAction Action = FirewallAllowOut; 
#else
    FirewallAction Action = FirewallBlock;
#endif

	if( !ZkService::GetINymCentral()->IsCurrentNymUser() ) 
	{
		// Override everything for certain "Supported" ports
		if( ZKS::IsProxyPortDefaultForTheCloud(in_DestPort) ||
			IsPortPartOfUserAddedProxyPortForCloud(in_DestPort)
			)
		{
			const size_t nIndex = ZkService::GetINymCentral()->GetActiveNymIndex();
			ZKASSERT (nIndex >= 0);
			const bool bAccessCloud = ( (nIndex != USER_NYM_INDEX) && ZkService::GetINymCentral()->DoesNymUseCloud(nIndex) );

			if (!bAccessCloud)				
			{
				Action = FirewallAllowOut;
			}
			else
			{
				Action = FirewallRouteThroughCloud;
			}
		}
		else if(ZKS::IsDefaultNewsPort(in_DestPort) ||
				ZKS::IsDefaultSmtpPort(in_DestPort) ||
				ZKS::IsDefaultPop3Port(in_DestPort)
			)
		{
			Action = FirewallAllowOut;
		}
	}

	if ( in_DestPort == FR_DFLT_AIP_TCP_PORT )
	{	
		if( ZkService::GetIRouteManager()->IsDiffieHellmanTrafficAllowed(DestIP) ) 
		{
			ZKCTRACE1( TRACE_MYINTAC, "DH packet going out. Port is [%d]\n", in_DestPort);
			Action = FirewallAllowOut ;
		}
		else
		{
			ZKCTRACE0( TRACE_MYINTAC, "Someone is trying to hack us!\n");
			Action = FirewallBlock;			
		}
	}
	else if( in_DestPort == FR_DFLT_KEYQRY_SERVER_PORT )
	{
		// Keyquery is UDP not TCP protocol
		Action = FirewallBlock ;
	}
	else if(( in_DestPort == FR_DFLT_TOKEN_SERVER_PORT ) ||
			( in_DestPort == FR_DFLT_FMGMAIP_SERVER_PORT ) ||
			( in_DestPort == FR_DFLT_NYM_SERVER_PORT ) ||
			( in_DestPort == FR_DFLT_POP_PORT ) ||
			( in_DestPort == FR_DFLT_NMTA_PORT ) )
	{
		Action = FirewallRouteThroughCloud ;
	}
	else if ( in_DestPort == FR_DFLT_NIQS_SERVER_PORT )
	{
		// Bug #8391 - NIQS allowed out in the clear
		Action = FirewallAllowOut ;
	}

	return Action;
}

FirewallAction 
RFirewall::PacketFirewallUDP(
	const u_long DestIP, 
	const uint16_t in_DestPort // in_DestPort is in host byte order
	) const
{
	ZKASSERT( m_bFirewallOn );

#ifdef WIN32
	FirewallAction Action = FirewallAllowOut; 
#else
    FirewallAction Action = FirewallBlock;
#endif

	if( in_DestPort == FR_DFLT_AIP_UDP_PORT ) // Data
	{	
		if( ZkService::GetIRouteManager()->IsCloudTrafficAllowed(DestIP) ) 
		{
			ZKCTRACE1( TRACE_MYINTAC, "Encrypting a UDP packet to first hop. Port is [%d]\n", in_DestPort);
			Action = FirewallAllowOut ;
		}
		else
		{
			ZKCTRACE0( TRACE_MYINTAC, "Someone is trying to hack us!\n");
			Action = FirewallBlock;			
		}

	}
	else if(in_DestPort == DNS_PORT)
	{
		ZKCTRACE0( TRACE_MYINTAC, "Use the cloud for DNS if selected Nym uses cloud\n");

		if( !ZkService::GetINymCentral()->IsCurrentNymUser() ) 
		{
			const size_t nIndex = ZkService::GetINymCentral()->GetActiveNymIndex();
			ZKASSERT (nIndex >= 0);
			const bool bAccessCloud = ( (nIndex != USER_NYM_INDEX) && ZkService::GetINymCentral()->DoesNymUseCloud(nIndex) );

			if (bAccessCloud)
			{
				ZKCTRACE0(TRACE_MYINTAC, "DNS going out over the cloud, selected nym uses cloud\n");
				Action = FirewallRouteThroughCloud;
			}
			else
			{
				ZKCTRACE0(TRACE_MYINTAC, "DNS going out in the clear, nym does not use cloud\n");
				Action = FirewallAllowOut;
			}
		}
		else
		{
			ZKCTRACE0(TRACE_MYINTAC, "DNS going out in the clear, no nym selected\n");
			Action = FirewallAllowOut;
		}
	}
	else if( in_DestPort == FR_DFLT_KEYQRY_SERVER_PORT )
	{
		// This is the only UDP protocol with our servers
		// Must use the cloud for key query server, 
		// We may still need keys later on for Emailing and initial queries may have failed
		Action = FirewallRouteThroughCloud;	
	}
	else if(( in_DestPort == FR_DFLT_TOKEN_SERVER_PORT ) ||
			( in_DestPort == FR_DFLT_FMGMAIP_SERVER_PORT ) ||
			( in_DestPort == FR_DFLT_NYM_SERVER_PORT ) ||
			( in_DestPort == FR_DFLT_NIQS_SERVER_PORT ) ||
			( in_DestPort == FR_DFLT_POP_PORT ) ||
			( in_DestPort == FR_DFLT_NMTA_PORT ) )
	{
		Action = FirewallBlock ;
	}

	return Action;
}


void
RFirewall::EnableFirewall( bool in_bEnable )
{
	m_bFirewallOn = in_bEnable ;

#ifdef WIN32
	IZkAppState::SetFirewall(in_bEnable);
#endif
}

void RFirewall::TestsWithUserNymActive()
{
	if ( !ZkService::GetINymCentral()->IsCurrentNymUser() ) 
	{
		ZKTEST( false ) ;
		return ;
	}

	unsigned long testIP = inet_addr( "100.100.100.100" ) ;

	// This is self-test code (calling internal private members)

	// Andrew - Removed tests here because now with differences
	// between Linux, Free Nyms etc it is hard to verify 
	// all the behavior differences.

	SelfTestCommon() ;
}

void RFirewall::TestsWithNymActive()
{
	SelfTestCommon() ;

	if ( ZkService::GetINymCentral()->IsCurrentNymUser() ) 
	{
		ZKTEST( false ) ;
		return ;
	}
}

void RFirewall::SelfTestCommon()
{
	// These are tests which apply at all Freedom states

	unsigned long testIP = inet_addr( "100.100.100.100" ) ;

	ZKTEST( FirewallBlock == PacketFirewallTCP(testIP, FR_DFLT_AIP_TCP_PORT) );	
	ZKTEST( FirewallBlock == PacketFirewallUDP(testIP, FR_DFLT_AIP_UDP_PORT) );	

	ZKTEST( FirewallRouteThroughCloud == PacketFirewallTCP(testIP, FR_DFLT_TOKEN_SERVER_PORT) );
	ZKTEST( FirewallBlock == PacketFirewallUDP(testIP, FR_DFLT_TOKEN_SERVER_PORT) );

	ZKTEST( FirewallAllowOut == PacketFirewallTCP(testIP, FR_DFLT_NIQS_SERVER_PORT) );
	ZKTEST( FirewallBlock == PacketFirewallUDP(testIP, FR_DFLT_NIQS_SERVER_PORT) );
}

// some of these ports go through the cloud, others don't
// but they are all 'special' in that we don't let the user
// configure our hard-coded ports
bool 
RFirewall::IsSupportedPort( uint16_t in_port )
{

	if( ZKS::IsDefaultIrcPort(in_port)				||
		ZKS::IsDefaultWebPort(in_port)				||
		ZKS::IsDefaultSslPort(in_port)				|| 
		ZKS::IsDefaultSshPort(in_port)				||
		ZKS::IsDefaultTelnetPort(in_port)			||
		ZKS::IsDefaultNewsPort(in_port)				||
		ZKS::IsDefaultSmtpPort(in_port)				||
		ZKS::IsDefaultPop3Port(in_port)				||
		ZKS::IsZksTcpThroughCloudPort( in_port )	|| 
		ZKS::IsZksUdpPort( in_port )				||
		FR_DFLT_SCRIPT_SERVER_TCP_PORT == in_port	||
		DNS_PORT == in_port							||
		65535 == in_port								// bug # 4999, access list problem on aips prevents us from using that port number
		)
	{
		return true;
	}

	// We don't check the Extra ports because these are not part of our 
	// defaults

	return false ;
}

// Returns true if the port was added by the user for the 
// HTTP, IRC, or TELNET proxies
bool
RFirewall::IsPortPartOfUserAddedProxyPortForCloud(
	const unsigned short in_Port) const
{
	return m_ExtraPortRules.IsProxyPort( in_Port ) ;
}

// includes user added ports (which applies to HTTP, IRC, and TELNET only (See firewall doc))
bool
RFirewall::IsProxyPort(unsigned short in_port)
{
	if( ZKS::IsDefaultIrcPort(in_port)			||
		ZKS::IsDefaultWebPort(in_port)			||
		ZKS::IsDefaultSslPort(in_port)			|| 
		ZKS::IsDefaultSshPort(in_port)			||
		ZKS::IsDefaultTelnetPort(in_port)		||
		ZKS::IsDefaultNewsPort(in_port)			||
		ZKS::IsDefaultSmtpPort(in_port)			||
		ZKS::IsDefaultPop3Port(in_port)
		)
	{
		return true;
	}

	// Check out map of extra proxy ports
	return m_ExtraPortRules.IsProxyPort( in_port ) ;
}


void RFirewall::GetPacketsBackLog(ZKSTD::deque<PacketHelper>& out_log)
{
	out_log = m_packetsBacklog;
}

// Returns value as indicated in connection_info structure comments
// loword, FirewallAction enumeration equivalent
// hiword, Local port to switch the connection to
unsigned int
RFirewall::SocketConnection(const connection_info&	in_ci)
{
	ZKASSERT (m_bFirewallOn );

	unsigned int ReturnValue	= FirewallAllowOut;
	bool bRunSwitch				= false;
	bool bIsCurrentNymUser		= ZkService::GetINymCentral()->IsCurrentNymUser() ;

	ZKASSERT( in_ci.DstIp != htonl(INADDR_LOOPBACK) );

	const unsigned short s_Destport = in_ci.DstPort;

	if( ZKS::IsDefaultIrcPort(s_Destport) || m_ExtraPortRules.IsIRCPort(s_Destport))
	{
		// Always wordscan IRC
		bRunSwitch = true;
		SwitchBoard(in_ci.bIsFreedomProcess, PROTOCOL_IRC, ReturnValue,  s_Destport, in_ci.DstIp );
	}
	else if(ZKS::IsDefaultWebPort(s_Destport)|| m_ExtraPortRules.IsHttpPort(s_Destport))
	{
		if(!bIsCurrentNymUser)
		{
			bRunSwitch = true;
			SwitchBoard(in_ci.bIsFreedomProcess, PROTOCOL_HTTP, ReturnValue, s_Destport, in_ci.DstIp );
		}
		else if( m_bHttpProxyForUserNym )
		{
			bRunSwitch = true;
			SwitchBoard(in_ci.bIsFreedomProcess, PROTOCOL_HTTP, ReturnValue, s_Destport, in_ci.DstIp );
		}
	}
	else if(ZKS::IsDefaultNewsPort(s_Destport))
	{
		// user chose not to see the news reader warning again
		bRunSwitch = true;
		SwitchBoard(in_ci.bIsFreedomProcess, PROTOCOL_NNTP1, ReturnValue,  s_Destport, in_ci.DstIp );
	}
	else if(ZKS::IsDefaultSslPort(s_Destport))
	{
		// We want to go through the SSL proxy even if no nym is selected so we
		// can pop up the SSL warning from a x-platform place (DInetD.cpp)
		bRunSwitch = true;
		SwitchBoard(in_ci.bIsFreedomProcess, PROTOCOL_SSL, ReturnValue,  s_Destport, in_ci.DstIp );
	}
	else if(ZKS::IsDefaultSshPort(s_Destport))
	{
		if(!bIsCurrentNymUser){
			bRunSwitch = true;
			SwitchBoard(in_ci.bIsFreedomProcess, PROTOCOL_SSH, ReturnValue,  s_Destport, in_ci.DstIp );
		}
	}
	else if(ZKS::IsDefaultTelnetPort(s_Destport) || m_ExtraPortRules.IsTelnetPort(s_Destport))
	{
		if(!bIsCurrentNymUser){
			bRunSwitch = true;
			SwitchBoard(in_ci.bIsFreedomProcess, PROTOCOL_TELNET, ReturnValue,  s_Destport, in_ci.DstIp );
		}
	}
	else if(ZKS::IsDefaultSmtpPort(s_Destport))
	{			
		bRunSwitch = true;
		SwitchBoard(in_ci.bIsFreedomProcess, PROTOCOL_SMTP, ReturnValue,  s_Destport, in_ci.DstIp );
	}
	else if(ZKS::IsDefaultPop3Port(s_Destport))
	{
		bRunSwitch = true;
		SwitchBoard(in_ci.bIsFreedomProcess, PROTOCOL_POP3, ReturnValue,  s_Destport, in_ci.DstIp );
	}
	else if(ZKS::IsZksTcpThroughCloudPort(s_Destport)) // bug # 1876
	{
		if( !in_ci.bIsFreedomProcess && !m_bAllowTelnetToCoreServers )
		{
			ZKCTRACE0( TRACE_MYINTAC, "Blocking connection to aip/token/fmgmaip/nym/niqs from a non-Freedom address space\n");
			// Returning immediately to prevent unnecessary message sending
			return FirewallBlock;
		}
	}
	else
	{
		// Unknown port, connection to foreign host.


	}

	if (!bRunSwitch) {
		// Message is not being routed through our proxies.  We will tell
		// Freedom for it to mention this fact to the user		

		if ( ZK_LOSHORT(ReturnValue) == FirewallAllowOut ) 
		{
			if(m_pHandler)
			{
				ZKASSERT_POINTER(m_pHandler, IPacketNotification);
				PacketHelper hint(IPPROTO_TCP, in_ci.DstIp, s_Destport );

				// We don't expect these connections because our proxies are active even when no nym is active
				// This CAN assert if you get mail using a localhost ssh proxy, don't be frightened
				ZKASSERT(s_Destport != 110 && s_Destport != 25 && s_Destport != 119);

				m_pHandler->NotifyAllowingSocketConnection( &hint );
			}
			else
			{
				ZKCTRACE0( TRACE_MYINTAC, "No one to notify about allowed socket connection, m_pHandler is NULL\n");				
			}

		}
		else if ( ZK_LOSHORT(ReturnValue) == FirewallBlock ) 
		{
			if(m_pHandler)
			{
				ZKASSERT_POINTER(m_pHandler, IPacketNotification);
				PacketHelper hint(IPPROTO_TCP, in_ci.DstIp, s_Destport );

				// We don't expect these connections because our proxies are active even when no nym is active
				ZKASSERT(s_Destport != 110 && s_Destport != 25 && s_Destport != 119);

				m_pHandler->NotifyDisallowingSocketConnection( &hint );
			}
			else
			{
				ZKCTRACE0( TRACE_MYINTAC, "No one to notify about disallowed socket connection, m_pHandler is NULL\n");				
			}
		}
	}

	if ( ZK_LOSHORT(ReturnValue) == FirewallAllowOut )	   
	{
		AllowTcpConnectionWithHost(
			in_ci.DstIp, 
			htons(s_Destport)	// convert port from host byte order to network byte order as per struct pending_connection comments
		);
	}

	return ReturnValue;
}

// Call this prior to a socket connect()ions reaching the base provider 
// (essentially before a SYN packet actually leaves)
// Later, when a SYN-ACK comes in, and the source ip and source port
// match the ip and port we are storing here, we will let the packet come into the stack
// LocalPort is zero at this point, it will be saved when the SYN packet leaves
// What we save is the destination ip, destination port, app name and socket number
void
RFirewall::AllowTcpConnectionWithHost(
	const unsigned int		in_ip,	// NBO
	const unsigned short	in_port	// NBO
	)
{
	ZKASSERT (m_bFirewallOn );
	ZKASSERT (in_ip != htonl(INADDR_LOOPBACK));

	ZKS::Lock_Guard<ZKCriticalSection> gate(m_TcpHostVeccs);

	const pending_connection pc = { in_ip, in_port, 0 };

	m_TcpHostVec.push_back(pc);

	ZKTRACE2("TCP packets from IP [%d] Port [%d] will be allowed in\n", 
		in_ip, 
		ntohs(in_port) 
		);
}


// Returns true if the packet's source IP and source Port were previously sent to AllowTcpConnectionWithHost
// Call for incoming TCP packets only
// You may think about calling this method only if the packet is a SYN-ACK
// but that may lead to a race condition where packets come in after the SYN-ACK quicker
// than the LSP can inform the firewall of the legitimacy of the local port
// You'll notice we can compare local port numbers because we saved it with the SYN
// packet in AddLocalPortToTcpHostVecEntry.
bool
RFirewall::PartOfTcpConnectionExchange(const ZkTCPHeader* in_phdr)
{
	ZKASSERT (m_bFirewallOn );

	const unsigned short LocalPort      = in_phdr->DstPort;
	const unsigned short HostSourcePort = in_phdr->SrcPort;
	const unsigned int   HostSourceIP   = in_phdr->IP.SrcIpAddress;

	ZKTRACE3("PartOfTcpConnectionExchange with LocalPort [%d] HostSourcePort [%d] HostSourceIP [%d]\n", 
				ntohs(LocalPort),
				ntohs(HostSourcePort),
				HostSourceIP);

	ZKS::Lock_Guard<ZKCriticalSection> gate(m_TcpHostVeccs);

	for(HostVec::const_iterator it = m_TcpHostVec.begin(); it != m_TcpHostVec.end(); ++it)
	{
		ZKTRACE3("Looking in TcpHostVec ip [%d] port [%d] LocalPort [%d]\n", 
			(*it).DstIp, 
			ntohs((*it).DstPort), 
			ntohs((*it).LocalPort));

		if( ((*it).DstIp == HostSourceIP) && ((*it).DstPort == HostSourcePort) && ((*it).LocalPort == LocalPort) )
		{
			ZKTRACE2("The IP [%d] is good to receive data to local TCP port [%d]\n", (*it).DstIp, ntohs( (*it).DstPort ) );
			return true;
		}
	}

	ZKTRACE2("Did not find host whom we previously attempted to TCP connect IP [%d] Port [%d]\n", HostSourceIP, ntohs(HostSourcePort) );

	return false;
}


// Sets "out_ReturnValue" to 
// loword, FirewallAction enumeration equivalent
// hiword, Local port to switch the connection to
void
RFirewall::SwitchBoard(
	const bool			in_bIsFreedomProcess,
	INET_PROTOCOLS		in_Protocol, 
	unsigned int&		out_ReturnValue, 
	const u_short		in_outgoingtoport,
	const unsigned int	in_outgoingtoip)
{

	if( in_bIsFreedomProcess )
	{

		out_ReturnValue = FirewallAllowOut; // loword

		if(in_Protocol == PROTOCOL_HTTP)
		{
			if(m_pHandler)
			{
				ZKASSERT_POINTER(m_pHandler, IPacketNotification);
				m_pHandler->NotifyConnectionHttp( in_outgoingtoip, in_outgoingtoport );
			}
			else
			{
				ZKCTRACE0( TRACE_MYINTAC, "No one to notify about http local proxy connection, m_pHandler is NULL\n");
				ZKASSERT(0); // this shouldn't happen, SendPacketNotificationsTo() should be called
			}
		}
		else if(in_Protocol == PROTOCOL_IRC)
		{
			if(m_pHandler)
			{
				ZKASSERT_POINTER(m_pHandler, IPacketNotification);
				m_pHandler->NotifyConnectionIrc( in_outgoingtoip, in_outgoingtoport );
			}
			else
			{
				ZKCTRACE0( TRACE_MYINTAC, "No one to notify about irc local proxy connection, m_pHandler is NULL\n");
				ZKASSERT(0); // this shouldn't happen, SendPacketNotificationsTo() should be called
			}
		}
		else if(in_Protocol == PROTOCOL_NNTP1)
		{
			if(m_pHandler)
			{
				ZKASSERT_POINTER(m_pHandler, IPacketNotification);
				m_pHandler->NotifyConnectionNntp( in_outgoingtoip, in_outgoingtoport );
			}
			else
			{
				ZKCTRACE0( TRACE_MYINTAC, "No one to notify about nntp local proxy connection, m_pHandler is NULL\n");
				ZKASSERT(0); // this shouldn't happen, SendPacketNotificationsTo() should be called
			}
		}
		else if( in_Protocol == PROTOCOL_SSL )
		{
			if( m_pHandler )
			{
				ZKASSERT_POINTER( m_pHandler, IPacketNotification );
				m_pHandler->NotifyConnectionSsl( in_outgoingtoip, in_outgoingtoport );
			}
			else
			{
				ZKCTRACE0( TRACE_MYINTAC, "No one to notify about ssl local proxy connection, m_pHandler is NULL\n");
				ZKASSERT(0); // this shouldn't happen, SendPacketNotificationsTo() should be called
			}
		}
		else if(in_Protocol == PROTOCOL_TELNET)
		{
			if(m_pHandler)
			{
				ZKASSERT_POINTER(m_pHandler, IPacketNotification);					
				m_pHandler->NotifyConnectionTelnet( in_outgoingtoip, in_outgoingtoport );					
			}
			else
			{
				ZKCTRACE0( TRACE_MYINTAC, "No one to notify about telnet local proxy connection, m_pHandler is NULL\n");
				ZKASSERT(0); // this shouldn't happen, SendPacketNotificationsTo() should be called
			}
		}
		else if(in_Protocol == PROTOCOL_SSH)
		{				
			if(m_pHandler)
			{
				ZKASSERT_POINTER(m_pHandler, IPacketNotification);					
				m_pHandler->NotifyConnectionSsh( in_outgoingtoip, in_outgoingtoport );					
			}
			else
			{
				ZKCTRACE0( TRACE_MYINTAC, "No one to notify about ssh over telnet local proxy connection, m_pHandler is NULL\n");
				ZKASSERT(0); // this shouldn't happen, SendPacketNotificationsTo() should be called
			}
		}
		else if(in_Protocol == PROTOCOL_SMTP)
		{
			if(m_pHandler)
			{
				ZKASSERT_POINTER(m_pHandler, IPacketNotification);
				m_pHandler->NotifyConnectionSmtp( in_outgoingtoip, in_outgoingtoport );
			}
			else
			{
				ZKCTRACE0( TRACE_MYINTAC, "No one to notify about smtp local proxy connection, m_pHandler is NULL\n");
				ZKASSERT(0); // this shouldn't happen, SendPacketNotificationsTo() should be called
			}
		}
		else if(in_Protocol == PROTOCOL_POP3)
		{
			if(m_pHandler)
			{
				ZKASSERT_POINTER(m_pHandler, IPacketNotification);
				m_pHandler->NotifyConnectionPop3( in_outgoingtoip, in_outgoingtoport );
			}
			else
			{
				ZKCTRACE0( TRACE_MYINTAC, "No one to notify about pop3 local proxy connection, m_pHandler is NULL\n");
				ZKASSERT(0); // this shouldn't happen, SendPacketNotificationsTo() should be called
			}
		}
		else
		{
			ZKASSERT(0);
		}

#ifndef NDEBUG
		in_addr trc;
		trc.s_addr = in_outgoingtoip;
		ZKCTRACE2( TRACE_MYINTAC, "RFirewall::SwitchBoard() port [%d] and IP [%s]\n", in_outgoingtoport, inet_ntoa(trc) );
#endif		
	}
	else
	{
		if( !IZkAppState::IsFreedomInitialized() )
		{
			ZKTRACE0("Freedom is not ready to switch things over to the proxies\n");
			out_ReturnValue = FirewallAllowOut;
		}
		else
		{
			uint16_t nLocalListeningPort = 0;

			if(in_Protocol == PROTOCOL_HTTP)			nLocalListeningPort = m_uHttpPort;
			else if (in_Protocol == PROTOCOL_POP3)		nLocalListeningPort = m_uPop3Port;
			else if (in_Protocol == PROTOCOL_SMTP)		nLocalListeningPort = m_uSmtpPort;
			else if (in_Protocol == PROTOCOL_IRC)		nLocalListeningPort = m_uIrcPort;
			else if (in_Protocol == PROTOCOL_NNTP1)		nLocalListeningPort = m_uNntpPort;
			else if (in_Protocol == PROTOCOL_TELNET)	nLocalListeningPort = m_uTelnetPort;
			else if (in_Protocol == PROTOCOL_SSH)		nLocalListeningPort = m_uTelnetPort;
			else if (in_Protocol == PROTOCOL_SSL)		nLocalListeningPort = m_uSslPort;
			else ZKASSERT(false);

			ZKASSERT(nLocalListeningPort <= 65535);
			ZKASSERT(nLocalListeningPort > 0);
					
			out_ReturnValue = FirewallForwardSockaddr; // loword

			// we are sending the new local port to the lsp dll in the hiword
			out_ReturnValue |= (nLocalListeningPort << 16); // hiword		

			ZKASSERT(nLocalListeningPort == ZK_HISHORT(out_ReturnValue) );
					
			ZKCTRACE2( TRACE_MYINTAC, "Going to Change destination port from [%d] to local proxy port [%d]\n", in_outgoingtoport, nLocalListeningPort );
		
			ZKCTRACE1( TRACE_MYINTAC, "RFirewall::SwitchBoard(): switched the WSPConnect port to %d\n", nLocalListeningPort);
		}
	}
}

// Call from successfull bind/accept
// Returns 0 if the local port wasn't saved
// Returns 1 otherwise
int
RFirewall::SaveLocalServerPort(const connection_info& in_ci)
{
	ZKASSERT (m_bFirewallOn);
	ZKASSERT (in_ci.s != 0);
	ZKASSERT (in_ci.Localport != 0);

	if(epostbind == in_ci.api)
	{
		ZKASSERT (in_ci.DstIp == 0);
		ZKASSERT (in_ci.DstPort == 0);
	}

	if(0 == in_ci.socket_type)
	{
		ZKASSERT(0); // How can you successfully call bind/accept but the socket type be unavailable?
		return 0;
	}

	if(SOCK_DGRAM == in_ci.socket_type)
	{
		if(in_ci.Localport != 0)
		{
			ZKTRACE2("Saving local UDP port [%d] for [%s]\n", in_ci.Localport, in_ci.app);

			remotehost rh = { 0, 0, "", htons(in_ci.Localport), in_ci.s, in_ci.tid };
			strncpy(rh.application, in_ci.app, MAX_PATH);
			InsertUdpPortSetEntry(rh);
			return 1;
		}
		else
		{	
			ZKTRACE2("Not Saving local UDP port of zero [%d] for [%s]\n", in_ci.Localport, in_ci.app);
			ZKASSERT(0);
			return 0;
		}		

	}

	ZKASSERT(SOCK_STREAM == in_ci.socket_type);

	// Following is for a local TCP port //

#ifdef _DEBUG
	if(in_ci.bIsFreedomProcess)
	{
		ZKTRACE2("SaveLocalServerPort freedom address space is true, TCP localport [%d] destPort should be zero for bind [%d]\n", in_ci.Localport, in_ci.DstPort);
	}
	else
	{
		ZKTRACE2("SaveLocalServerPort freedom address space is false, TCP localport [%d] destPort should be zero for bind [%d]\n", in_ci.Localport, in_ci.DstPort);
	}
#endif

	// Notice how the connection_info structure has LocalPort in HBO whereas
	// the remotehost structure has it in NBO.

	remotehost rh = { 0, 0, "", htons(in_ci.Localport), in_ci.s, in_ci.tid };
	strncpy(rh.application, in_ci.app, MAX_PATH);
	InsertTcpPortSetEntry(rh);

	ZKTRACE2("[%s] local TCP Port [%d] will be opened\n", rh.application, in_ci.Localport);

	return 1;
}

void
RFirewall::InitLocalProxyPorts()
{
	IZkRegistry::GetDWORD( "Pop3", (unsigned long&)m_uPop3Port);
				
	IZkRegistry::GetDWORD( "SMTP", (unsigned long&)m_uSmtpPort);
					
	IZkRegistry::GetDWORD( "HTTP", (unsigned long&)m_uHttpPort);
	
	IZkRegistry::GetDWORD( "NNTP", (unsigned long&)m_uNntpPort);
	
	IZkRegistry::GetDWORD( "IRC", (unsigned long&)m_uIrcPort);
	
	IZkRegistry::GetDWORD( "TELNET", (unsigned long&)m_uTelnetPort);
	
	IZkRegistry::GetDWORD( "SSL", (unsigned long&)m_uSslPort);
}


// You should reboot the machine if adding any of these entries in the registry
bool
RFirewall::IsLocalProxyPort(const unsigned short in_localPort)
{
	if( ( in_localPort == m_uPop3Port ) ||
		( in_localPort == m_uSmtpPort ) ||
	    ( in_localPort == m_uHttpPort ) ||
		( in_localPort == m_uNntpPort ) ||
	    ( in_localPort == m_uIrcPort ) ||
	    ( in_localPort == m_uTelnetPort ) ||
	    ( in_localPort == m_uSslPort )
		)
	{
		return true;
	}
	return false;
}

bool 
RFirewall::IsDriverSendingUpPackets()
{
#ifdef INTERNAL_BUILD
 	if(getenv("BYPASS_DRIVER")) /* audit: ok */
	{
		// We pretend the driver is running to allow
		// freedom to start
		return true;
	}
#endif

	return m_bAtLeastOneOutgoingPacketFromDriver ;
}

// We go through m_TcpHostVec and
// if the local port is not in endpoint remove it from m_TcpHostVec
// then do the same thing for the m_TcpPortSet
// If the port _is_ in the endpoint vector, we'll check if the tcp state is closed or time_wait 
// and remove it from the set if it is in that state in the endpoint vector
// See bug 5545
void
RFirewall::VerifyLocalTcpPortsState(const ZKSTD::vector<endpoint>& in_tcpvec)
{
	if( !m_bFirewallOn )
	{
		ZKTRACE0("Firewall is off, VerifyLocalTcpPortsState returning\n");
		return;
	}

	{
		ZKS::Lock_Guard<ZKCriticalSection> gate(m_TcpHostVeccs);

		for(HostVec::iterator it = m_TcpHostVec.begin(); it != m_TcpHostVec.end(); )
		{
			if(0 == (*it).LocalPort)
			{
				++it;
				continue; // we can have a 0 as the localPort in the host vector, but not in the port set
			}

			ZKSTD::vector<endpoint>::const_iterator vecIt = ZKSTD::find_if(
																	in_tcpvec.begin(), 
																	in_tcpvec.end(), 
																	endpoint_equal_localport( ntohs((*it).LocalPort) )
																	);

			if( vecIt == in_tcpvec.end() ) // local port is in m_TcpHostVec but is not in endpoint vector, remove it from m_TcpHostVec
			{
				ZKTRACE2("Removing entry in m_TcpHostVec of size [%d] for localPort [%d] it is not an endpoint\n", 
					m_TcpHostVec.size(),
					ntohs((*it).LocalPort)
					);
			
				m_TcpHostVec.erase(it);

				it = m_TcpHostVec.begin();

				ZKTRACE1("After attempted removal size is [%d]\n", m_TcpHostVec.size() );
			}
			else
			{
				++it;
			}
		}
	}

	{
		ZKS::Lock_Guard<ZKCriticalSection> gate(m_TcpPortSetcs);

		for(PortSet::iterator it = m_TcpPortSet.begin(); it != m_TcpPortSet.end(); )
		{
			ZKSTD::vector<endpoint>::const_iterator vecIt = ZKSTD::find_if(
																	in_tcpvec.begin(), 
																	in_tcpvec.end(), 
																	endpoint_equal_localport( ntohs((*it).LocalPort) )
																	);

			if( vecIt == in_tcpvec.end() ) // port in m_PortSet but is not in endpoint vector, remove it from set
			{
				ZKTRACE3("Removing entry in m_TcpPortSet of size [%d] for localPort [%d] app [%s] it is not an endpoint\n", 
					m_TcpPortSet.size(),
					ntohs((*it).LocalPort), 
					(*it).application 
					);
				
				PortSet::iterator tmp = it;
				++tmp;
				m_TcpPortSet.erase(it);
				it = tmp;

				ZKTRACE1("After attempted removal size is [%d]\n", m_TcpPortSet.size() );
			}
			else
			{
				// port is in m_TcpPortSet _and_ in the endpoint vector
				// if the state is closed or time_wait we'll remove it

				if( (closed == (*vecIt).state) || (time_wait == (*vecIt).state) ) 
				{
					ZKTRACE3("Removing entry in m_TcpPortSet of size [%d] for localPort [%d] app [%s] it is closed or time_wait\n", 
						m_TcpPortSet.size(),
						ntohs((*it).LocalPort), 
						(*it).application 
						);

					PortSet::iterator tmp = it;
					++tmp;
					m_TcpPortSet.erase(it);
					it = tmp;
					
					ZKTRACE1("After attempted removal size is [%d]\n", m_TcpPortSet.size() );
				}
				else
				{
					++it;
				}
			}
		}
	}
}


// We go through m_UdpPortSet and
// if the local port is not in endpoint remove it from m_UdpPortSet
// See bug 5545
void
RFirewall::VerifyLocalUdpPortsState(const ZKSTD::vector<endpoint>& in_udpvec)
{

	if( !m_bFirewallOn )
	{
		ZKTRACE0("Firewall is off, VerifyLocalUdpPortsState returning\n");
		return;
	}

	{
		ZKS::Lock_Guard<ZKCriticalSection> gate(m_UdpHostVeccs);	

		for(HostVec::iterator it = m_UdpHostVec.begin(); it != m_UdpHostVec.end(); )
		{
			ZKSTD::vector<endpoint>::const_iterator vecIt = ZKSTD::find_if(
														in_udpvec.begin(), 
														in_udpvec.end(), 
														endpoint_equal_pending_connection( (*it) ) 
														);

			ZKTRACE2("VerifyLocalUdpPortsState() in m_UdpHostVec UDP sendto IP [%d] Port [%d]\n", (*it).DstIp, ntohs((*it).DstPort) );

			if( vecIt == in_udpvec.end() ) // port in m_UdpHostVec but is not in endpoint vector, remove it from vector
			{	
				ZKTRACE1("Removing entry in m_UdpHostVec of size [%d] it is not an endpoint\n", m_UdpHostVec.size() );

				m_UdpHostVec.erase(it);

				it = m_UdpHostVec.begin();

				ZKTRACE1("After attemped removal size is [%d]\n", m_UdpHostVec.size() );
			}
			else
			{
				++it;
			}
		}
	}


	ZKS::Lock_Guard<ZKCriticalSection> gate(m_UdpPortSetcs);

	for(PortSet::iterator it = m_UdpPortSet.begin(); it != m_UdpPortSet.end(); )
	{
		ZKSTD::vector<endpoint>::const_iterator vecIt = ZKSTD::find_if(
																in_udpvec.begin(), 
																in_udpvec.end(), 
																endpoint_equal_localport( ntohs((*it).LocalPort) )
																);

		if( vecIt == in_udpvec.end() ) // port in m_UdpPortSet but is not in endpoint vector, remove it from set
		{			

			ZKTRACE3("Removing entry in m_UdpPortSet of size [%d] for localPort [%d] app [%s] it is not an endpoint\n", 
				m_UdpPortSet.size(),
				ntohs((*it).LocalPort), 
				(*it).application 
				);

			PortSet::iterator tmp = it;
			++tmp;
			m_UdpPortSet.erase(it);
			it = tmp;

			ZKTRACE1("After attempted removal size is [%d]\n", m_UdpPortSet.size() );
		}
		else
		{
			++it;
		}
	}
}

void
RFirewall::FlushFirewallRules() const
{
#ifdef WIN32
	TrafficFilter::theTrafficFilter().FlushAllRules();
#endif
}

// Generic Flag management methods in ClientCentral allow us to manage multiple
// flags with minimal code

void 
RFirewall::SetFirewallPreference(unsigned long in_dwProperty, bool in_bNewState)
{

#ifdef WIN32
	// Special cases for three check-boxes
	if( in_dwProperty == PREF_DISABLE_FIREWALL) // see bug 7668
	{
		TrafficFilter::theTrafficFilter().CaptureAllIncomingPackets( !in_bNewState ); // see bug 7139
	}
	else if ( in_dwProperty == PREF_BLOCK_NETBIOS )
	{		
		TrafficFilter::theTrafficFilter().AllowNetBiosInTheClear( !in_bNewState ); // see bug 318
	}
	else if( in_dwProperty == PREF_BLOCK_DHCP )
	{
		TrafficFilter::theTrafficFilter().AllowDhcpInTheClear( !in_bNewState );
	}
	else if( in_dwProperty == PREF_BLOCK_RPC)
	{
		TrafficFilter::theTrafficFilter().AllowRpcInTheClear( !in_bNewState );
	}	
#endif
	
	ZkService::GetIClientCentral()->SetPreference( ZK_FIREWALL_PREF_SET, in_dwProperty, in_bNewState ) ;
}

bool
RFirewall::GetFirewallPreference(unsigned long in_dwProperty) const
{
	return ZkService::GetIClientCentral()->GetPreference( ZK_FIREWALL_PREF_SET, in_dwProperty ) ;
}


void
RFirewall::PostSendToResults(const postsendto_info& in_psi)
{
	if(!m_bFirewallOn) return;

	if( in_psi.DstIp == htonl(INADDR_LOOPBACK) )
	{
		ZKTRACE3("ignoring localhost sendto ip [%d] port [%d] localPort [%d]\n", 
			in_psi.DstIp, 
			ntohs(in_psi.DstPort), 
			ntohs(in_psi.LocalPort) 
			);
		return;
	}

	if(in_psi.Ok)
	{
		if(SOCK_DGRAM == in_psi.socket_type) // may be implicit binding through sendto
		{
			remotehost rh = { in_psi.DstIp, in_psi.DstPort, "", in_psi.LocalPort, in_psi.s, in_psi.tid };
			strncpy(rh.application, in_psi.app, MAX_PATH);

			ZKS::Lock_Guard<ZKCriticalSection> gate(m_UdpPortSetcs);	

			PortSet::const_iterator it = m_UdpPortSet.find(rh);

			// If the socket number is there and the local port is zero save it
			if(it != m_UdpPortSet.end() && (0 == (*it).LocalPort) )
			{
				InsertUdpPortSetEntry(rh);
			}
			else if(it == m_UdpPortSet.end())
			{
				// implicit binding
				InsertUdpPortSetEntry(rh);
			}
			else
			{
				ZKTRACE3("PostSendToResults() Socket [%d] LocalPort [%d] DstPort[%] not being saved\n",
					in_psi.s,
					ntohs(in_psi.LocalPort),
					ntohs(in_psi.DstPort)
					);
			}
		}		
		else if(SOCK_STREAM != in_psi.socket_type)
		{
			// This could simply be an app using raw sockets
			// We can still display our pop up firewall dialogs
			// Note we do nothing about local ports
			// Local ports are handled only for UDP and TCP sockets
			// Unfortunately things like ping.exe and tracert
			// bypass our LSP completely...
			ZKTRACE2("PostSendToResults() not a tcp or udp socket [%d] for app [%s]\n", in_psi.s, in_psi.app);
		}
	}
	else
	{
		// if the sendto failed, nothing to do
	}
}

void
RFirewall::PostConnectionResults(const postconnection_info& in_pci)
{
	if(!m_bFirewallOn) return;


	if( (SOCK_DGRAM != in_pci.socket_type) && (SOCK_STREAM != in_pci.socket_type) )
	{
		ZKTRACE2("PostConnectionResults() socket [%d] is of type [%d] letting it go out\n", in_pci.s, in_pci.socket_type);
		return;
	}	


	if(  in_pci.DstIp == htonl(INADDR_LOOPBACK) )
	{
		ZKTRACE3("Since we don't save localhost information during connects, we don't want to handle post connects ip [%d] port [%d] localPort [%d]\n", in_pci.DstIp, ntohs(in_pci.DstPort), ntohs(in_pci.LocalPort) );
		return;
	}

	if(in_pci.Ok)
	{
		if(SOCK_DGRAM == in_pci.socket_type)
		{
			// Add the local port to that entry information, other fields are accurate
			remotehost rh = { in_pci.DstIp, in_pci.DstPort, "", in_pci.LocalPort, in_pci.s, in_pci.tid };
			ZKS::Lock_Guard<ZKCriticalSection> gate(m_UdpPortSetcs);	
			PortSet::const_iterator it = m_UdpPortSet.find(rh);

			remotehost rh_entry = *it;

			if(it != m_UdpPortSet.end())
			{	
				ZKASSERT(rh.s == rh_entry.s);			

				if(0 == rh_entry.ip)
				{
					rh_entry.ip = rh.ip;				// add any missing piece of information
				}
				else
				{
					ZKASSERT(rh.ip == rh_entry.ip);
				}

				if(0 == rh_entry.port)
				{
					rh_entry.port = rh.port;			// add any missing piece of information
				}
				else
				{
					ZKASSERT(rh.port == rh_entry.port);
				}

				if(0 == rh_entry.LocalPort)
				{
					rh_entry.LocalPort = in_pci.LocalPort; // add any missing piece of information
				}
				else
				{
					// app probably did a bind before the connect, in which case we already 
					// have the local port for that entry
					ZKASSERT( in_pci.LocalPort == rh_entry.LocalPort ); 
				}
				InsertUdpPortSetEntry(rh_entry);
			}			
			return;
		}

		ZKASSERT(SOCK_STREAM == in_pci.socket_type);			

		// when the connection succeeds
		// look up entry in m_TcpHostVec
		// only if found move the information into m_TcpPortSet

		ZKS::Lock_Guard<ZKCriticalSection> gate(m_TcpHostVeccs);

		remotehost rh = { in_pci.DstIp, in_pci.DstPort, "", in_pci.LocalPort, in_pci.s, in_pci.tid };
		strncpy(rh.application, in_pci.app, MAX_PATH);

		for(HostVec::const_iterator it = m_TcpHostVec.begin(); it != m_TcpHostVec.end(); ++it)
		{
			ZKTRACE3("Looking in TcpHostVec ip [%d] port [%d] LocalPort [%d]\n", 
				(*it).DstIp, 
				ntohs((*it).DstPort), 
				ntohs((*it).LocalPort));

			if( ((*it).DstIp == in_pci.DstIp) && ((*it).DstPort == in_pci.DstPort) && ((*it).LocalPort == in_pci.LocalPort) )
			{
				InsertTcpPortSetEntry( rh );
				return;
			}
		}

		ZKTRACE0("The connection succeeded, yet we never saw a syn packet which we extracted the local port\n");
	}
}

// Call when seen outgoing SYN
// so that we can associate a local port with a particular remote ip and port
bool
RFirewall::AddLocalPortToTcpHostVecEntry(
	const unsigned int		in_DstIp,	// NBO
	const unsigned short	in_DstPort,	// NBO
	const unsigned short	in_port)	// NBO
{
	ZKASSERT (m_bFirewallOn);
	ZKASSERT (0 != in_port);

	try{
		ZKS::Lock_Guard<ZKCriticalSection> gate(m_TcpHostVeccs);
		
		for(HostVec::iterator it = m_TcpHostVec.begin(); it!= m_TcpHostVec.end(); ++it)
		{
			pending_connection rh_entry = *it;

			ZKTRACE3("AddLocalPortToTcpHostVecEntry() Looking in TcpHostVec ip [%d] port [%d] LocalPort [%d]\n", 
				(*it).DstIp, 
				ntohs((*it).DstPort), 
				ntohs((*it).LocalPort));

			if( (rh_entry.DstIp == in_DstIp) && (rh_entry.DstPort == in_DstPort) && (rh_entry.LocalPort == 0) )
			{
				(*it).LocalPort = in_port;
				return true;
			}
		}

	}
	catch(...)
	{
		ZKTRACE0("C++ exception when attempting to modify existing m_TcpHostVec entry\n");
	}
	
	ZKTRACE0("AddLocalPortToTcpHostVecEntry didn't save anything\n");

	return false;
}



void
RFirewall::Dump_remotehost(const char* in_function, const remotehost& in_rh)
{
	ZKASSERT(in_function);

	ZKTRACE4(" [%s] dumping remotehost => ip [%s] port [%d] app [%s]\n", 
		in_function, 
		ZKS::Cvt_int32toString(in_rh.ip).c_str(), 
		ntohs(in_rh.port),
		in_rh.application);

	ZKTRACE2("LocalPort [%d] socket [%d]\n", ntohs(in_rh.LocalPort), in_rh.s);
}


void
RFirewall::InsertTcpPortSetEntry(const remotehost& in_rh)
{
	ZKASSERT (m_bFirewallOn );
	ZKASSERT (in_rh.s != 0);

	try{
		ZKS::Lock_Guard<ZKCriticalSection> gate(m_TcpPortSetcs);

		Dump_remotehost("InsertTcpPortSetEntry", in_rh);

		m_TcpPortSet.erase( in_rh );
		m_TcpPortSet.insert( in_rh );

#ifdef _DEBUG
		PortSet::const_iterator it = m_TcpPortSet.find(in_rh);
		ZKASSERT( it != m_TcpPortSet.end() );
		ZKASSERT( in_rh.ip == (*it).ip );
		ZKASSERT( in_rh.LocalPort == (*it).LocalPort );
		ZKASSERT( in_rh.port == (*it).port );
#endif

	}
	catch(...)
	{
		ZKTRACE0("C++ exception when attempting to insert m_TcpPortSet\n");
	}
}

void
RFirewall::InsertUdpPortSetEntry(const remotehost& in_rh)
{
	ZKASSERT (m_bFirewallOn );
	ZKASSERT (in_rh.s != 0);

	// The local port can be 0 when a UDP socket does a connect 

	try{
		ZKS::Lock_Guard<ZKCriticalSection> gate(m_UdpPortSetcs);
		Dump_remotehost("InsertUdpPortSetEntry parm==>", in_rh);

#ifndef NDEBUG
		PortSet::const_iterator it = m_UdpPortSet.find(in_rh);
		// May be in the set, (app does bind, then connect...)
		if( it != m_UdpPortSet.end() )
		{
			Dump_remotehost("InsertUdpPortSetEntry previous entry being overwritten==>", *it);		
			ZKASSERT(in_rh.s == (*it).s);
			ZKASSERT(in_rh.tid == (*it).tid);
		}
#endif
		m_UdpPortSet.erase( in_rh );
		m_UdpPortSet.insert( in_rh );
	}
	catch(...)
	{
		ZKTRACE0("C++ exception when attempting to insert entry into udp port set\n");
	}
}



bool
RFirewall::RemoveUdpPortSetEntry(const int in_s, const unsigned long in_tid)
{

	ZKASSERT (m_bFirewallOn );

	try{
		ZKS::Lock_Guard<ZKCriticalSection> gate(m_UdpPortSetcs);

		ZKTRACE2("Removing socket [%d] from UDP port set of size [%d]\n", in_s, m_UdpPortSet.size() );

		remotehost rh = { 0 }; 
		rh.s	= in_s;
		rh.tid	= in_tid;

		PortSet::const_iterator itNewEnd = m_UdpPortSet.find(rh);

		if( itNewEnd != m_UdpPortSet.end() )
		{
			ZKTRACE3("The following udp port set entry will be deleted app [%s] localport [%d] remote port [%d]\n",
							(*itNewEnd).application, ntohs( (*itNewEnd).LocalPort ), ntohs( (*itNewEnd).port ) );

			m_UdpPortSet.erase(rh);
		}

		ZKTRACE1("After attempted removal size is [%d]\n", m_UdpPortSet.size() );

		return true;
	}
	catch(...)
	{
		ZKTRACE0("C++ exception when attempting to remove a socket entry from the UDP port set\n");
	}
	return false;
}


// Returns
// loword, FirewallAction enumeration equivalent
// hiword, Local port to switch the connection to
unsigned int 
RFirewall::GetAppFirewallOptions(const connection_info& in_ci)
{

	if( !m_bFirewallOn )
	{	
		return FirewallAllowOut;
	}

	if( in_ci.DstIp == htonl(INADDR_LOOPBACK) )
	{
		// If it's a 127.0.0.1 connection we let it proceed normally
		// We should do this for the real machine's IP also (eg: 192.168.1.125)
		// This also corrects Netscape's double-connection feature,
		// see  http://www.stardust.com/cgi-bin/wa?A2=ind9807&L=WINSOC&P=R9959		

		ZKASSERT( !in_ci.bIsFreedomProcess || in_ci.DstPort == LOCAL_IPC_HTTP_SERVER_PORT );
		return FirewallAllowOut;
	}


	ZKASSERT(in_ci.api == eprelisten || in_ci.api == epostbind || in_ci.api == epresendto || in_ci.api == epreconnect || in_ci.api == epostaccept || in_ci.api == eprebind);

	if( (SOCK_DGRAM != in_ci.socket_type) && (SOCK_STREAM != in_ci.socket_type) )
	{
		ZKTRACE2("GetAppFirewallOptions() socket [%d] is of type [%d] letting it go out\n", in_ci.s, in_ci.socket_type);
		return FirewallAllowOut;
	}		

	if(!in_ci.bIsFreedomProcess)
	{
		if(in_ci.IncomingTraffic())
		{
			ZKTRACE1("The app [%s] wants to act as a server\n", in_ci.app );
			ZKASSERT_POINTER(m_pHandler, IPacketNotification);

			if( GetFirewallPreference(PREF_DISABLE_FIREWALL) ) // see bug 7139
			{
				return FirewallAllowOut;
			}

			const eSelection sel = SomeoneWantsToActAsAServer(in_ci) ;

			// Some Api's are only provided for UI purposes and we don't have any local port number to handle
			// So we return right here for those api's
			if(in_ci.IsUIOnlyApi())
			{
				ZKASSERT(unknownselection != sel);
				if(block == sel)
				{
					return FirewallBlock;
				}
				else
				{
					return FirewallAllowOut;
				}
			}
			// If users chooses No, return right away.
			else if(block == sel)
			{
				return FirewallBlock;
			}
		}
		else if(in_ci.OutgoingTraffic())
		{
			ZKTRACE1("The app [%s] wants to get to the internet\n", in_ci.app );
			ZKASSERT_POINTER(m_pHandler, IPacketNotification);

			const eSelection sel = GetFirewallPreference(PREF_DISABLE_FIREWALL) ? 
									allow : SomeoneWantsToConnectToNet(in_ci);

			// Some Api's are only provided for UI purposes and we don't have any local port number to handle
			// So we return right here for those api's
			if(in_ci.IsUIOnlyApi())
			{
				if(block == sel)
				{
					return FirewallBlock;
				}
				else
				{
					return FirewallAllowOut;
				}
			}
			// If users chooses No, return right away.
			else if(block == sel)
			{
				return FirewallBlock;
			}
		}
	}

	// Bug fix 6142
	if(in_ci.IsUIOnlyApi())
	{
		return FirewallAllowOut;
	}

	// store the local port as good-to-receive-packets ////////////////////////////////////	

	if(epostaccept == in_ci.api)
	{
		ZKASSERT( in_ci.DstIp != htonl(INADDR_LOOPBACK) );		
		ZKTRACE1("Saving a local port [%d] gotten from accept\n", in_ci.Localport);
		ZKASSERT(in_ci.Localport != 0);
		SaveLocalServerPort(in_ci); 		
		return FirewallAllowOut; // loword of return value
	}
	else if(epostbind == in_ci.api)
	{
		ZKTRACE0("Saving a local port gotten from bind\n");
		ZKASSERT(in_ci.Localport != 0);
		SaveLocalServerPort(in_ci);
		return FirewallAllowOut; // loword of return value
	}
	else if(epresendto == in_ci.api)
	{
		// bug 6066, optimization only, scenario described in bug 6561 cannot affect freedom
		// optimized in spi.cpp
		ZKASSERT(!in_ci.bIsFreedomProcess) ;
		
		ZKTRACE2("Saving dest ip and dest port in presendto ip[%d] port [%d]\n", in_ci.DstIp, in_ci.DstPort);

		ZKS::Lock_Guard<ZKCriticalSection> gate(m_UdpHostVeccs);

		const pending_connection pc = { in_ci.DstIp, htons(in_ci.DstPort), 0 }; // convert port from host byte order to network byte order as per struct pending_connection comments

		m_UdpHostVec.push_back(pc);		
		
		return FirewallAllowOut; // loword of return value
	}
	else if(epreconnect == in_ci.api)
	{
		// already asked about app accessing the net and answer was affirmative		

		if(SOCK_DGRAM == in_ci.socket_type)
		{
			remotehost rh = { in_ci.DstIp, htons(in_ci.DstPort), "", htons(in_ci.Localport), in_ci.s, in_ci.tid };
			strncpy(rh.application, in_ci.app, MAX_PATH);

			// Notice how the connection_info structure has LocalPort in HBO whereas
			// the remotehost structure has it in NBO.

			// Although bind may of been called previously, we lose no information
			// by re-inserting this socket information, the local port will be valid
			// here. In addition we now know the destination ip and port


			InsertUdpPortSetEntry(rh);
			return FirewallAllowOut; // loword of return value
		}
		else
		{
			// NO need to save local port on TCP connect, internal mechanism handles this
			ZKASSERT(SOCK_STREAM == in_ci.socket_type);
			// loword, FirewallAction enumeration equivalent
			// hiword, Local port to switch the connection to        
			const unsigned int nRet = SocketConnection(in_ci);
			return nRet;
		}
	}
	ZKASSERT(0);
	return FirewallBlock;
}

void
RFirewall::GetApplicationSettings(ZKSTD::vector<FirewallAppPersist>& out_apps)
{
	ZKS::Lock_Guard<ZKCriticalSection> gate(m_AppMapSetcs);

	for(AppMap::const_iterator the_it = m_AppMap.begin();  the_it != m_AppMap.end(); ++the_it)
	{
		out_apps.push_back((*the_it).second);
	}
}

// Uses ZKTRACE to print contents of structure
void
RFirewall::DumpFirewallAppPersist(const FirewallAppPersist& in_fwp)
{
	ZKTRACE1("DumpFirewallAppPersist=> [%s] \n", ZKS::ConvertFirewallAppPersistIntoString(in_fwp).c_str() );
}

void
RFirewall::SetFirewallRulesForApplication( const FirewallAppPersist& in_Rule )
{
	if ( strlen( in_Rule.app ) == 0 )
	{
		// Invalid data!
		ZKASSERT( false ) ;
		return ;
	}


	// If caller _should_ care or different behavior is
	// wanted then ModifyExistingAppEntry, 

	ZKS::Lock_Guard<ZKCriticalSection> gate(m_AppMapSetcs);
	m_AppMap[in_Rule.app] = in_Rule;
}

void
RFirewall::EraseFirewallRulesForApplication( const char * in_pAppToErase )
{
	if (( in_pAppToErase == NULL ) ||
		( strlen( in_pAppToErase ) == 0 ) )
	{
		// Invalid arguments
		ZKASSERT( false ) ;
		return ;
	}

	ZKS::Lock_Guard<ZKCriticalSection> gate(m_AppMapSetcs);

	AppMap::iterator it = m_AppMap.find(in_pAppToErase);

	if(it != m_AppMap.end())
	{
		m_AppMap.erase( it ) ;
	}
	else
	{
		ZKTRACE1( "Firewall - EraseFirewallRulesForApplication was called for %s but this is not in list\n",
				  in_pAppToErase ) ;
	}
}

// Given an application name we will find the existing entry for it in our
// map and modify it according to the FirewallAppPersist parameter
void
RFirewall::ModifyExistingAppEntry(
	const char					in_app[MAX_PATH], 
	const FirewallAppPersist&	in_opts)
{
	ZKS::Lock_Guard<ZKCriticalSection> gate(m_AppMapSetcs);

	FirewallAppPersist	persisted_opts = { 0 };
	ZKVERIFY( GetAppOptions(in_app, persisted_opts) );

#ifdef _DEBUG
	{
	ZKTRACE0("In ModifyExistingAppEntry() replacing the entry  =====>\n");
	DumpFirewallAppPersist(persisted_opts);
	ZKTRACE0("With the following entry  =====>\n");
	DumpFirewallAppPersist(in_opts);
	}
#endif
	
	// Avoid multiple calls by detecting whether anything actually changed
	// Doing a yucky binary comparison
	if ( 0 != memcmp( &persisted_opts, &in_opts, sizeof( in_opts) ) )
	{
		m_AppMap[in_app] = in_opts;

		// We assume that the UI does not call this method itself to persist changes
		IFirewall::FirewallHint hint(IFirewall::FirewallHint::FIREWALL_APP_RULES_CHANGED);
		notify(&hint);
	}
}

// Given an application name we will save a FirewallAppPersist structure
// in our m_AppMap. The application name must not already exist in our map
void
RFirewall::SaveAppSettings(const char in_app[MAX_PATH], const FirewallAppPersist& in_persisted_opts)
{
	ZKS::Lock_Guard<ZKCriticalSection> gate(m_AppMapSetcs);

#ifdef _DEBUG
	{
	FirewallAppPersist	persisted_opts = { 0 };
	if( GetAppOptions(in_app, persisted_opts) )
	{
		ZKTRACE0("In SaveAppSettings() replacing the entry  =====>\n");
		DumpFirewallAppPersist(persisted_opts);
		ZKTRACE0("With the following entry  =====>\n");
		DumpFirewallAppPersist(in_persisted_opts);
	}
	}
#endif

	ZKASSERT( !strcmp(in_app, in_persisted_opts.app) );

	m_AppMap[in_app] = in_persisted_opts;

	// We assume that the UI does not call this method itself to persist changes
	IFirewall::FirewallHint hint(IFirewall::FirewallHint::FIREWALL_APP_RULES_CHANGED);
	notify(&hint);
}
	

// Given an application name we will look up the application
// in our persisted info and return a FirewallAppPersist structure
// Returns true if application was previously persisted, false otherwise
bool
RFirewall::GetAppOptions(const char in_app[MAX_PATH], FirewallAppPersist& out_persisted_opts)
{
	ZKS::Lock_Guard<ZKCriticalSection> gate(m_AppMapSetcs);

	AppMap::const_iterator it = m_AppMap.find(in_app);

	if(it != m_AppMap.end())
	{
		out_persisted_opts = (*it).second;

		ZKTRACE3("In GetAppOptions() app [%s] incoming [%d] outgoing [%d]\n",
			out_persisted_opts.app,
			out_persisted_opts.incoming.Sel,
			out_persisted_opts.outgoing.Sel
			);

#ifdef _DEBUG
		if(unknownselection == out_persisted_opts.incoming.Sel)
		{
			// Doesn't make sense to remember "unknown selection"
			ZKASSERT( out_persisted_opts.incoming.bRemember == false ) ;
		}

		if(unknownselection == out_persisted_opts.outgoing.Sel)
		{
			// Doesn't make sense to remember "unknown selection"
			ZKASSERT( out_persisted_opts.outgoing.bRemember == false ) ;
		}
#endif

		return true;
	}
	else
	{
		ZKTRACE1("We've never saved anything on [%s]\n", in_app);
	}
	return false;
}

eSelection
RFirewall::SomeoneWantsToActAsAServer(const connection_info& in_ci)
{
	ZKASSERT (m_bFirewallOn );

	FirewallAppPersist	persisted_opts = { 0 };

	if( GetAppOptions(in_ci.app, persisted_opts) )
	{
		if(persisted_opts.incoming.bRemember)
		{
			ZKASSERT(unknownselection != persisted_opts.incoming.Sel);
			return persisted_opts.incoming.Sel;
		}
	}
	if(m_pHandler)
	{
		const eSelection Sel = m_pHandler->AppRequestsInNetAccess(in_ci.app);

		ZKASSERT(unknownselection != Sel);			

		return Sel;				
	}
	return block;
}



void
RFirewall::GetTcpPassiveStatePorts(ZKSTD::vector<remotehost>& out_v)
{
	ZKS::Lock_Guard<ZKCriticalSection> gate(m_TcpHostVeccs);

	for(HostVec::const_iterator it = m_TcpHostVec.begin(); it != m_TcpHostVec.end(); ++it)
	{
		const remotehost rh = { (*it).DstIp, (*it).DstPort, "n/a-passive", (*it).LocalPort, 0 };

		out_v.push_back(rh);
	}
}

void
RFirewall::GetTcpOpenedPorts(ZKSTD::vector<remotehost>& out_v)
{
	ZKS::Lock_Guard<ZKCriticalSection> gate(m_TcpPortSetcs);

	for(PortSet::const_iterator it = m_TcpPortSet.begin(); it != m_TcpPortSet.end(); ++it)
	{
		out_v.push_back(*it);
	}
}

void
RFirewall::GetUdpOpenedPorts(ZKSTD::vector<remotehost>& out_v)
{
	ZKS::Lock_Guard<ZKCriticalSection> gate(m_UdpPortSetcs);

	for(PortSet::const_iterator it = m_UdpPortSet.begin(); it != m_UdpPortSet.end(); ++it)
	{
		out_v.push_back(*it);
	}
}

// Returns true if data for a port _at this time_ and _for this nym_
// will got through the cloud. The PacketFirewall method
// already has all the smarts dealing with this, so we'll use that
// method
bool 
RFirewall::IsCloudData(
	const unsigned short in_port	// port is in host byte order
	) const
{
	return (
		FirewallRouteThroughCloud == PacketFirewallTCP(1, in_port) ||
		FirewallRouteThroughCloud == PacketFirewallUDP(1, in_port)  // bug 6828
		);
}

eSelection
RFirewall::SomeoneWantsToConnectToNet(const connection_info& in_ci)
{
	ZKASSERT (m_bFirewallOn );

	FirewallAppPersist	persisted_opts = { 0 };

	const size_t nIndex = ZkService::GetINymCentral()->GetActiveNymIndex();
	ZKASSERT (nIndex >= 0);	
	const bool bNymUsesCloud = ( (nIndex != USER_NYM_INDEX) && ZkService::GetINymCentral()->DoesNymUseCloud(nIndex) );

	if( GetAppOptions(in_ci.app, persisted_opts) )
	{
		if(IsCloudData(in_ci.DstPort))
		{
			if(persisted_opts.outgoing.bRemember)
			{
				ZKASSERT(unknownselection != persisted_opts.outgoing.Sel);
				return persisted_opts.outgoing.Sel;
			}
			else
			{
				const eSelection Sel = 	m_pHandler->AppRequestsOutNetAccess(in_ci.app, 
														in_ci.DstIp, 
														in_ci.DstPort, 
														false	// is it a special warning?
														);

				ZKASSERT(unknownselection != Sel);

				return Sel;
			}
		}
		else // non-cloud data
		{
			if( bNymUsesCloud )  
			{
				// premium nym selected yet this connection is not using the cloud, see bug 7138

				if( persisted_opts.special_warn.bRemember )
				{
					ZKASSERT(unknownselection != persisted_opts.special_warn.Sel);
					return persisted_opts.special_warn.Sel;
				}
				else
				{
					const eSelection Sel = m_pHandler->AppRequestsOutNetAccess(in_ci.app, 
														in_ci.DstIp, 
														in_ci.DstPort, 
														true	// is this a special warning?
														); 

					ZKASSERT(unknownselection != Sel);

					return Sel;
				}
			}
			else
			{
			
				// not a premium nym selected and app doesn't use the cloud

				const eSelection Sel = m_pHandler->AppRequestsOutNetAccess(in_ci.app, 
														in_ci.DstIp, 
														in_ci.DstPort, 
														false	// is this a special warning?
														); 

				ZKASSERT(unknownselection != Sel);

				return Sel;
			}
		}
	}
	else //  never asked about app.exe before
	{
		if(IsCloudData(in_ci.DstPort) )
		{
			const eSelection Sel = m_pHandler->AppRequestsOutNetAccess(in_ci.app, 
													in_ci.DstIp, 
													in_ci.DstPort, 
													false	// is this special warning?												
													);

			ZKASSERT(unknownselection != Sel);

			return Sel;
		}
		else // non cloud data
		{
			if( bNymUsesCloud )
			{
				// premium nym selected but app is about to go outside the cloud

				const eSelection Sel = m_pHandler->AppRequestsOutNetAccess(in_ci.app, 
													in_ci.DstIp, 
													in_ci.DstPort, 
													true	// is this special warning?												
													);

				ZKASSERT(unknownselection != Sel);

				return Sel;
			}
			else
			{
				// Not a premium nym selected and app is about to go outside the cloud

				const eSelection Sel = m_pHandler->AppRequestsOutNetAccess(in_ci.app, 
													in_ci.DstIp, 
													in_ci.DstPort, 
													false // is this special warning?
													);

				ZKASSERT(unknownselection != Sel);

				return Sel;
			}
		}
	}
}

void
RFirewall::GetExtraProxyPorts( RPortToProxyMap& out_Ports )
{
	out_Ports = m_ExtraPortRules ;
}

void
RFirewall::SetExtraProxyPorts( const RPortToProxyMap& in_Ports )
{
	m_ExtraPortRules = in_Ports ;
}

void 
RFirewall::Save( WMemBuf& io_Buffer ) throw(RZkException)
{
	ZKChunkWriter cw( io_Buffer, FIREWALL_CHUNK_ID, FIREWALL_CHUNK_VERSION ) ;

	if ( !m_AppMap.empty() )
	{
		SaveFirewallAppRules( io_Buffer ) ;
	}

	if ( !m_ExtraPortRules.empty() )
	{
		SaveExtraPortRules( io_Buffer ) ;
	}
}


void 
RFirewall::SaveExtraPortRules( WMemBuf& io_Buffer ) throw(RZkException)
{
	ZKChunkWriter cw( io_Buffer, FIREWALL_PORT_RULES_CHUNK_ID, 1 ) ;

	for(RPortToProxyMap::const_iterator it = m_ExtraPortRules.begin();  
		it != m_ExtraPortRules.end(); 
		++it)
	{
		// Verify data validity
		if ( IsSupportedPort( (*it).first ) )
		{
			ZKASSERT( false ) ;
			continue ;
		}

		if ( (*it).second == PortRuleUndefined )
		{
			ZKASSERT( false ) ;
			continue ;
		}

		cw.WriteUInt16( (*it).first ) ;
		cw.WriteUInt8( (uint8_t) (*it).second ) ;
	}
}

void 
RFirewall::LoadExtraPortRules( ZKChunkReader& io_cr ) throw(RZkException)
{
	if ( io_cr.GetChunkID() != FIREWALL_PORT_RULES_CHUNK_ID ) 
	{
		// Corruption or bug in the load code, we must load in the correct order
		ZKASSERT( false ) ;
		throw RZkException("RFirewall::LoadExtraPortRules", 
							 "Corrupt File",
							 IDS_CORRUPT_FREEDOM_DATA ) ;
	}

	// Erase all existing content
	m_ExtraPortRules.clear() ;

	while ( !io_cr.IsEndOfChunk() )
	{
		uint16_t port = io_cr.ReadUInt16() ;
		RuleForPort rule = (RuleForPort) io_cr.ReadUInt8() ;

		if (  IsSupportedPort( port ) || 
			  ( rule == PortRuleUndefined ))
		{
			// This might happen if we change the list of supported
			// ports, but normally should never happen
			ZKASSERT( false ) ;
			continue ;
		}

		m_ExtraPortRules[ port ] = rule ;
	}
}

void
RFirewall::SaveFirewallAppRules( WMemBuf& io_Buffer ) throw(RZkException) 
{
	ZKChunkWriter cw( io_Buffer, FIREWALL_APP_RULES_CHUNK_ID, FIREWALL_APP_RULES_CHUNK_VERSION ) ;

	for(AppMap::const_iterator the_it = m_AppMap.begin();  the_it != m_AppMap.end(); ++the_it)
	{
		const FirewallAppPersist& AppRule = (*the_it).second ;

		if ( strlen( AppRule.app ) == 0 )
		{
			// Don't save bogus data
			ZKASSERT( false ) ;
			continue ;
		}

		cw.WriteString( AppRule.app ) ;
		cw.WriteUInt8( AppRule.incoming.bRemember ) ;
		cw.WriteUInt8( AppRule.incoming.Sel ) ;

		cw.WriteUInt8( AppRule.outgoing.bRemember ) ;
		cw.WriteUInt8( AppRule.outgoing.Sel ) ;

		cw.WriteUInt8( AppRule.special_warn.bRemember ) ;
		cw.WriteUInt8( AppRule.special_warn.Sel );
	}
}

void
RFirewall::LoadFirewallAppRules( ZKChunkReader& io_cr ) throw(RZkException)
{
	if ( io_cr.GetChunkID() != FIREWALL_APP_RULES_CHUNK_ID ) 
	{
		// Corruption or bug in the load code, we must load in the correct order
		ZKASSERT( false ) ;
		throw RZkException("RFirewall::LoadFirewallAppRules", 
							 "Corrupt File",
							 IDS_CORRUPT_FREEDOM_DATA ) ;
	}

	// Erase all existing content
	m_AppMap.clear() ;

	// Load the content present
	while ( !io_cr.IsEndOfChunk() )
	{
		FirewallAppPersist AppRule = { 0 } ;

		ZKSTD::string strAppName ;
		io_cr.ReadString( strAppName ) ;

		if (( strAppName.size() == 0 ) ||
			( strAppName.length() >= MAX_PATH ))
		{
			// bogus data
			ZKASSERT( false ) ;
			return ;
		}

		strncpy( AppRule.app, strAppName.c_str(), MAX_PATH ) ;

		AppRule.incoming.bRemember = ( 0 != io_cr.ReadUInt8()) ;
		AppRule.incoming.Sel = ( eSelection	) io_cr.ReadUInt8() ;

		AppRule.outgoing.bRemember = ( 0 != io_cr.ReadUInt8()) ;
		AppRule.outgoing.Sel = ( eSelection	) io_cr.ReadUInt8() ;
		
		AppRule.special_warn.bRemember = ( 0 != io_cr.ReadUInt8()) ;
		AppRule.special_warn.Sel = ( eSelection	) io_cr.ReadUInt8() ;

#ifdef _DEBUG
		if(unknownselection == AppRule.incoming.Sel)
		{
			// Doesn't make sense to remember "unknown selection"
			ZKASSERT( AppRule.incoming.bRemember == false ) ;
		}

		if(unknownselection == AppRule.outgoing.Sel)
		{
			// Doesn't make sense to remember "unknown selection"
			ZKASSERT( AppRule.outgoing.bRemember == false ) ;
		}

		if(unknownselection == AppRule.special_warn.Sel)
		{
			// Doesn't make sense to remember "unknown selection"
			ZKASSERT( AppRule.special_warn.bRemember == false ) ;
		}
#endif

		SetFirewallRulesForApplication( AppRule ) ;
	}
}

void 
RFirewall::Load( WReadBuffer& io_Buffer ) throw(RZkException)
{	
	ZKChunkReader cr( io_Buffer ) ;

	if ( cr.GetChunkID() != FIREWALL_CHUNK_ID ) 
	{
		// Corruption or bug in the load code, we must load in the correct order
		ZKASSERT( false ) ;
		throw RZkException("RFirewall::Load", 
							 "Corrupt File - Firewall Chunk",
							 IDS_CORRUPT_FREEDOM_DATA ) ;
	}

	while ( !cr.IsEndOfChunk() )
	{
		ZKChunkReader cr2( cr.GetReadBuffer() );
		switch( cr2.GetChunkID() )
		{
			case( FIREWALL_APP_RULES_CHUNK_ID ) :
				LoadFirewallAppRules( cr2 ) ;
				break ;

			case( FIREWALL_PORT_RULES_CHUNK_ID ) :
				LoadExtraPortRules( cr2 ) ;
				break ;

			case ( OLD_FIREWALL_APP_RULES_CHUNK_ID ):
				// We throw away app rules beta1.1 and 1.2 of version 2.0
				// because the format changed
				ZKTRACE0("Loaded old app firewall chunk id, skipping it\n");
				break ;

			case ( FIREWALL_SPECIAL_WARN_CHUNK_ID ):
				ZKTRACE0("Loaded old firewall chunk id, skipping it\n");
				break;
				
			default :
				// Future chunks - just ignore them
				ZKASSERT( false ) ;
				break ;
		}
	}
}

void 
RFirewall::PushBackLogWithMax(const PacketHelper& in_packet)
{
	// assure that there is no more that s_maxSizeOfBackLog element
	// in the deque.  More than that could cause slow UI and too much
	// memory needed to run the process (swap).
	while( m_packetsBacklog.size() > s_maxSizeOfBackLog)
	{
		m_packetsBacklog.pop_front();
	}
	m_packetsBacklog.push_back(in_packet);

	// Tell any observers of this Log structure that
	// we've changed.  For example if a property page is showing
	// the log it can refresh
	IFirewall::FirewallHint hint(IFirewall::FirewallHint::FIREWALL_INCOMING_BLOCK_LOG_CHANGE);
	notify(&hint);
}


struct FirewallEvent_UsersChoice
RFirewall::GetSpecificAppSettingsOut(const char in_AppName[MAX_PATH])
{
	FirewallEvent_UsersChoice uc = { 0 };
	ZKASSERT(in_AppName);

	ZKS::Lock_Guard<ZKCriticalSection> gate(m_AppMapSetcs);

	AppMap::const_iterator it = m_AppMap.find(in_AppName);

	if(it != m_AppMap.end())
	{
		uc = (*it).second.outgoing;
	}
	return uc;
}

struct FirewallEvent_UsersChoice
RFirewall::GetSpecificAppSettingsIn(const char in_AppName[MAX_PATH])
{
	FirewallEvent_UsersChoice uc = { 0 };
	ZKASSERT(in_AppName);

	ZKS::Lock_Guard<ZKCriticalSection> gate(m_AppMapSetcs);

	AppMap::const_iterator it = m_AppMap.find(in_AppName);

	if(it != m_AppMap.end())
	{
		uc = (*it).second.incoming;
	}
	return uc;
}



// Unlike SetFirewallRulesForApplication this method is used for
// a user activated ui activity, such that we assert the selection
// is not unknownselection. If remember is false though, we won't
// save the selection value, instead we will save unknownselection.
// This is simply to avoid the case where bRemember is false
// for a specific app, yet the selection value is allow or block.
void
RFirewall::SetSpecificAppSettingsOut(
	const char								in_AppName[MAX_PATH], 
	const struct FirewallEvent_UsersChoice& in_uc,
	const bool								in_bSpecialWarning)
{
	ZKASSERT(in_AppName);
	ZKASSERT(unknownselection != in_uc.Sel);

	ZKS::Lock_Guard<ZKCriticalSection> gate(m_AppMapSetcs);

	AppMap::const_iterator it = m_AppMap.find(in_AppName);


	if(it != m_AppMap.end())
	{
		FirewallAppPersist ap = (*it).second;
		if(!in_bSpecialWarning)
		{
			ap.outgoing = in_uc;		
			if(!in_uc.bRemember)
			{
				ap.outgoing.Sel = unknownselection;
			}
		}
		else
		{
			ap.special_warn = in_uc;
			if(!in_uc.bRemember)
			{
				ap.special_warn.Sel = unknownselection;
			}
		}
		ModifyExistingAppEntry(in_AppName, ap);
	}
	else
	{
		FirewallAppPersist ap = { 0 };
		strncpy(ap.app, in_AppName, MAX_PATH);
		if(!in_bSpecialWarning)
		{
			ap.outgoing = in_uc;		
			if(!in_uc.bRemember)
			{
				ap.outgoing.Sel = unknownselection;
			}
		}
		else
		{
			ap.special_warn = in_uc;
			if(!in_uc.bRemember)
			{
				ap.special_warn.Sel = unknownselection;
			}
		}
		SaveAppSettings(in_AppName, ap);
	}
}

// Unlike SetFirewallRulesForApplication this method is used for
// a user activated ui activity, such that we assert the selection
// is not unknownselection. If remember is false though, we won't
// save the selection value, instead we will save unknownselection.
// This is simply to avoid the case where bRemeber is false
// for a specific app, yet the selection value is allow or block.
void
RFirewall::SetSpecificAppSettingsIn(const char in_AppName[MAX_PATH], const struct FirewallEvent_UsersChoice& in_uc)
{
	ZKASSERT(in_AppName);
	ZKASSERT(unknownselection != in_uc.Sel);

	ZKS::Lock_Guard<ZKCriticalSection> gate(m_AppMapSetcs);

	AppMap::const_iterator it = m_AppMap.find(in_AppName);

	if(it != m_AppMap.end())
	{
		FirewallAppPersist ap = (*it).second;
		ap.incoming = in_uc;
		if(!in_uc.bRemember)
		{
			ap.incoming.Sel = unknownselection;
		}
		ModifyExistingAppEntry(in_AppName, ap);
	}
	else
	{
		FirewallAppPersist ap = { 0 };
		strncpy(ap.app, in_AppName, MAX_PATH);
		ap.incoming = in_uc;
		if(!in_uc.bRemember)
		{
			ap.incoming.Sel = unknownselection;
		}
		SaveAppSettings(in_AppName, ap);
	}
}
