// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// Realization of the IFirewall interface

#ifndef _RFirewall_H
#define _RFirewall_H

#include "IFirewall.h"
#include "IPacketFirewall.h"
#include <client.h>
#include <zks.h>
#include <vector>
#include <deque>
#include <set>
#include <RPortToProxyMap.h>

struct ZkICMPHeader;
struct ZkUDPHeader;
struct ZkTCPHeader;
class ZKChunkReader;

typedef struct pending_connection
{
	unsigned int	DstIp;		// network byte order
	unsigned short	DstPort;	// network byte order
	unsigned short	LocalPort;	// network byte order
}pending_connection;


struct endpoint_equal_pending_connection
{
	pending_connection x_;

	endpoint_equal_pending_connection(const pending_connection&  x) : x_(x) {}

	bool operator()(const endpoint& in_rh) const
	{ 
		return( (htonl(in_rh.remoteip) == x_.DstIp) && (htons(in_rh.remoteport) == x_.DstPort) );
	}
};


typedef ZKSTD::vector<pending_connection>	HostVec;

typedef ZKSTD::set<remotehost, CompareSocketsAndThreadId> PortSet;

typedef ZKSTD::map<ZKSTD::string, FirewallAppPersist> AppMap;

class RFirewall : public IFirewall, public IPacketFirewall
{
public:
	RFirewall() ;
	virtual	~RFirewall() ;

	// Comments in base class
	bool				Init();

	// Comments in base class
	// PROMISE: To call ApplyIncomingIcmpRules for the return value
	//			To call the NotifyDroppingIncomingPacket method if we are dropping it
	//			To call the NotifyAcceptingIncomingPacket method if we are accepting it
	// REQUIRE: We are given a IPacketNotification pointer previously
	virtual bool HandleIncomingPacket(const ZkNetFrame* in_buf) throw();
	

	// Comments in base class
	virtual void 
	SendPacketNotificationsTo( IPacketNotification* ) throw();

	// See comments in base class
	// Returns
	// loword, FirewallAction enumeration equivalent
	// hiword, Local port to switch the connection to
	virtual unsigned int 
	GetAppFirewallOptions(const connection_info&);

	// Comments in base class
	virtual void
	PostConnectionResults(const postconnection_info&);

	// Comments in base class
	virtual void
	PostSendToResults(const postsendto_info&);

	// Comments in base class
	virtual void
	VerifyLocalUdpPortsState(const ZKSTD::vector<endpoint>& in_v) throw();

	// Comments in base class
	virtual void
	VerifyLocalTcpPortsState(const ZKSTD::vector<endpoint>& in_v) throw();

	// Main logic for deciding what we should do with a packet that it trying
	// to leave the machine, see comments in base class
	virtual FirewallAction	
	GetActionForOutgoingTCPPacket( ZkBuffer *in_pbuf ) ;	

	virtual FirewallAction	
	GetActionForOutgoingUDPPacket( ZkBuffer *in_pbuf ) ;	

	virtual FirewallAction  
	GetActionForOutGoingIcmp(int in_type);

	// Test Code (for access from Unit Tests)
	void                
	TestsWithNymActive() ;

	void
	TestsWithUserNymActive() ;

	// Turn on the Firewall - by default it passively
	// allows all packets out
	virtual void        EnableFirewall(bool in_bEnable) ;

	virtual bool	IsSupportedPort( uint16_t in_port ) ;

	bool			IsProxyPort(unsigned short in_port) throw();

	virtual bool IsDriverSendingUpPackets() ;

	// return a copy of the log of inbound traffic blocked
	void GetPacketsBackLog(ZKSTD::deque<PacketHelper>& out_log);

	int
	SaveLocalServerPort(const connection_info& in_ci);

	// PURPOSE: Debugging
	virtual void
	GetTcpPassiveStatePorts(ZKSTD::vector<remotehost>& out_v);

	// PURPOSE: Debugging
	virtual void
	GetTcpOpenedPorts(ZKSTD::vector<remotehost>& out_v);

	// PURPOSE: Debugging
	virtual void
	GetUdpOpenedPorts(ZKSTD::vector<remotehost>& out_v);

	virtual void
	GetApplicationSettings(ZKSTD::vector<FirewallAppPersist>& out_apps);

	virtual void
	SetFirewallRulesForApplication( const FirewallAppPersist& in_Rule );

	virtual void
	EraseFirewallRulesForApplication( const char * in_pAppToErase );

	// Convenience methods for accessing firewall
	// properties
	// For example PREF_BLOCK_NETBIOS
	virtual void 
	SetFirewallPreference(unsigned long in_dwProperty, bool in_bNewState) ;

	virtual bool 
	GetFirewallPreference(unsigned long in_dwProperty) const ;

	virtual void
	GetExtraProxyPorts( RPortToProxyMap& out_Ports ) ;

	// Overwrite the rules for extra proxy ports 
	virtual void
	SetExtraProxyPorts( const RPortToProxyMap& in_Ports ) ;

	// Persistence Methods
	virtual void 
	Save( WMemBuf& io_Buffer ) throw(RZkException) ;

	virtual void 
	Load( WReadBuffer& io_Buffer ) throw(RZkException) ;

	virtual struct FirewallEvent_UsersChoice
	GetSpecificAppSettingsOut(const char in_AppName[MAX_PATH]);

	virtual struct FirewallEvent_UsersChoice
	GetSpecificAppSettingsIn(const char in_AppName[MAX_PATH]);

	virtual void
	SetSpecificAppSettingsOut(
		const char								in_AppName[MAX_PATH], 
		const struct FirewallEvent_UsersChoice& in_uc,
		const bool								in_bSpecialWarning);

	virtual void
	SetSpecificAppSettingsIn(const char in_AppName[MAX_PATH], const struct FirewallEvent_UsersChoice& in_uc);

	virtual bool
	GetAppOptions(const char in_app[MAX_PATH], FirewallAppPersist& out_persisted_opts);

	virtual bool 
	IsCloudData(const unsigned short in_port) const; // port is in host byte order

	void
	FlushFirewallRules() const;

private:

	// ctor member init	
	bool					m_bAtLeastOneOutgoingPacketFromDriver ;

	bool                    m_bFirewallOn ;	// Take a deep breath, throw another log on the fire, and
											// let me explain the difference between m_bFirewallOn == FALSE
											// and GetFirewallPreference(PREF_DISABLE_FIREWALL) = TRUE
											// When m_bFirewall all packets leave in the clear - no routes, proxies
											// or anything - eg when first starting the client and contacting NIQS
											// in the clear...
											// When m_bFirewall is on but GetFirewallPreference(PREF_DISABLE_FIREWALL) = FALSE
											// then in theory all unsupported ports go out in the clear, all incoming
											// packets are allows but the routes and proxies also work

	bool					m_bAllowTelnetToCoreServers ;	// By default, for security reasons, only freedom can core servers
															// contact core servers.  However for debugging purposes it is
															// useful to connect to core servers via telnet
	bool					m_bHttpProxyForUserNym ;		// We can adjust the behavior of the HTTP proxy when
															// no nym is selected

	IPacketNotification*	m_pHandler;

	// We allow override via the registry of the port numbers that the proxies wait
	// on.  
	uint16_t				m_uPop3Port;  // LOCAL_POP3_PORT is default, "Pop3" is override reg key
	uint16_t				m_uSmtpPort;
	uint16_t				m_uHttpPort;
	uint16_t				m_uNntpPort;
	uint16_t				m_uIrcPort;
	uint16_t				m_uTelnetPort;
	uint16_t				m_uSslPort;

	ZKSTD::deque<PacketHelper>	m_packetsBacklog;
	PortSet						m_UdpPortSet;
	PortSet						m_TcpPortSet;
	HostVec						m_UdpHostVec;
	HostVec						m_TcpHostVec;
	AppMap						m_AppMap;	

	// Proxies can be configured to accept other
	// ports via this map
	RPortToProxyMap				m_ExtraPortRules ;

	// Protect our asses
	ZKCriticalSection			m_UdpPortSetcs;
	ZKCriticalSection			m_TcpPortSetcs;
	ZKCriticalSection			m_UdpHostVeccs;
	ZKCriticalSection			m_TcpHostVeccs;
	ZKCriticalSection			m_AppMapSetcs;

	// static members

	static const unsigned int s_maxSizeOfBackLog;

	// private functions

	FirewallAction	
	PacketFirewallTCP(const u_long, const uint16_t) const;
	
	FirewallAction	
	PacketFirewallUDP(const u_long, const uint16_t) const;
	
	void			
	SelfTestCommon() ;

	// PURPOSE: General incoming Icmp rules.
	//			Note these hard rules should be added to the driver initially
	//			(see RWin2kPacketHandler::ApplyInitialIncomingPacketRules() as an example)
	//			For example, we may apply rules such that when this method
	//			does get called, it always will return true, since we applied
	//			rules to the driver telling it give-me-these-kinds-of-icmp
	//			packets, and let all the other ones into the stack.
	// PROMISE: Returns true if the packet should be let in
	//			Returns false if the packet should be blocked
	bool		
	ApplyIncomingIcmpRules(const ZkICMPHeader* in_pIcmpHeader) throw();

	// PURPOSE: General udp rules
	bool		
	ApplyIncomingUdpRules(const ZkUDPHeader*) throw();

	// PURPOSE: General tcp rules
	bool		
	ApplyIncomingTcpRules(const ZkTCPHeader*) throw();

	// PURPOSE: When a connection is attempted to a port for which
	//			we have a proxy, this is called to get the local port number
	void		
	SwitchBoard(const bool			in_bIsFreedomProcess,
			INET_PROTOCOLS		in_Protocol, 
			unsigned int&		out_ReturnValue, 
			const u_short		in_outgoingtoport,
			const unsigned int	in_outgoingtoip) throw();


	// Initializes our member variables with the port number where our proxies are listening on
	void		
	InitLocalProxyPorts() throw();

	bool
	IsLocalProxyPort(const unsigned short in_localPort) throw();

	// When a connection is initiated, the remote ip and port
	// is remembered, and incoming data from that ip & port is
	// let in. After the connect call that ip and port
	// is removed from this temporary map and may move in our m_portset
	bool
	PartOfTcpConnectionExchange(const ZkTCPHeader* in_phdr) throw();


	void
	AllowTcpConnectionWithHost(
			const unsigned int		in_ip,	// NBO
			const unsigned short	in_port	// NBO
			) throw();


	// port is in host byte order
	bool
	IsLocalTcpPortOpened(
			const unsigned short in_port,
			const unsigned int   in_source_ip, 
			const unsigned short in_source_port) throw();

	// port is in host byte order
	bool
	IsLocalUdpPortOpened(
			const unsigned short in_port,
			const unsigned int   in_source_ip, 
			const unsigned short in_source_port) throw();

	void
	InsertTcpPortSetEntry(const remotehost& in_rh) throw();

	// Port is in network byte order
	bool
	AddLocalPortToTcpHostVecEntry(
		const unsigned int		in_DstIp,	// NBO
		const unsigned short	in_DstPort,	// NBO
		const unsigned short	in_port		// NBO
	) throw();	

	bool
	RemoveUdpPortSetEntry(const int in_s, const unsigned long in_tid) throw();

	unsigned int
	SocketConnection(const connection_info&	in_ci) throw();

	void
	InsertUdpPortSetEntry(const remotehost& in_rh) throw();

	// Uses ZKTRACE to print contents of remotehost structure
	// The first parameter is also printed out, usually the caller function name is sent there
	void
	Dump_remotehost(const char* in_function, const remotehost& in_rh) throw();

	// Returns true if the socket in_s is in the m_TcpPortSet
	bool
	SocketIsInTcpPortSet(const int in_s) throw();

	void
	SaveAppSettings(const char in_app[MAX_PATH], const FirewallAppPersist& in_persisted_opts);

	eSelection
	SomeoneWantsToActAsAServer(const connection_info& in_ci);

	eSelection
	SomeoneWantsToConnectToNet(const connection_info& in_ci);

	bool
	IsPortPartOfUserAddedProxyPortForCloud(const unsigned short in_Port) const;

	void
	DumpFirewallAppPersist(const FirewallAppPersist& in_fwp);

	void
	ModifyExistingAppEntry(const char in_app[MAX_PATH], const FirewallAppPersist& in_opts);

	void
	LogAndNotifyDroppedIncomingPacket( PacketHelper &hint ) ;

	void
	NotifyAcceptedIncomingPacket( PacketHelper &hint ) ;

	void 
	PushBackLogWithMax(const PacketHelper& in_packet);

	void SaveExtraPortRules( WMemBuf& io_Buffer ) throw(RZkException) ;
	void LoadExtraPortRules( ZKChunkReader& io_ChunkReader ) throw(RZkException) ;

	void SaveFirewallAppRules( WMemBuf& io_Buffer ) throw(RZkException) ;
	void LoadFirewallAppRules( ZKChunkReader& io_ChunkReader ) throw(RZkException) ;

	bool
	IsLocalUdpPortInFastSendto(
		const unsigned int   in_source_ip, 
		const unsigned short in_source_port);


} ;


#endif
