// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RKeyQry - implementation of CRKeyQry class
// This class acts as a singleton and serves key requests
// to the client

#include "StdAfx.h"

// This project's includes
#include "Rkeyqry.h"
#include "INiqs.h"

#include "ICrypto.h"
#include "ZkService.h"
#include <IZkConsoleMsg.h>
#include <ZKFileConfiguration.h>
#include <IZkRegistry.h>
#include <memory>

// C DLL includes
#include <fr_keyqry_fetchkey.h>
#include <fr_rand.h>

RKeyQry::RKeyQry() 
	: m_Socket(NULL)
	, m_kq(NULL)
{
	// Note, this should be run once in the course of a freedom execution
	ZKVERIFY(frKeyQryNew(&m_Error, &m_kq));
}

RKeyQry::~RKeyQry()
{
	if(NULL != m_Socket)
	{
		DZkError err;
		if (!zkSocketFree(&err, &m_Socket)) ZKVERIFY(m_Socket = NULL);
	}
	ZKVERIFY(frKeyQryFree(&m_Error, &m_kq));	
}

FrKeyQry 
RKeyQry::GetKeyCache()
{
	ZKASSERT(m_kq);
	return m_kq;
}

bool 
RKeyQry::GetKey(
	DZkError& err, 
	const FrEntIdType	in_entity, 
	const int			in_nTimeout, 
	const char*			in_strEntityName, 
	const unsigned int	in_uKeyVersion, 
	const FrPubKeyType  in_keytype, 
	FrCryptPubKey*		out_pKey ) throw (RZkException)
{
	if (( in_strEntityName == NULL ) ||
		( out_pKey == NULL ))
	{
		ZKASSERT( !"Invalid Arguments to RKeyQry::GetKey" ) ;
		return false ;
	}

	// Determine the IP of the KeyQryServer
	WNetInfoDatum KeyQrySrvr ;

	////
	// This should be put back in. Route creation caused by packets going to the key qeury server
	// from frKeyQryFetchPubKey can cause additinal key lookups that get blocked on ths mutex.
	// In effect we rely on the fact that we can call these non-threadsafe functions from multiple
	// threads simultaneously.  --ol
	/////
    //   ZKS::Lock_Guard< WZkMutex > scopeGuard(m_keymutex);

	ZkService::GetINiqs()->GetServer( FRENT_TYPE_KEYQRY_SRV, KeyQrySrvr );

	SOCKADDR_IN KQS_Addr = KeyQrySrvr.GetSockAddr() ;

	bool bRet = false; 	

	if(!StartKeyQueryListeningSocket())
	{
		return bRet ;
	}

	FrEntIdExp exp;
	memset( &exp, 0, sizeof( exp ) ) ;
	exp.type = in_entity;
	strncpy(exp.name, in_strEntityName, FRENT_NAME_SIZE);

	if ( !GetKeyFromCache( err, in_entity, in_strEntityName, in_uKeyVersion, in_keytype, out_pKey) )
	{
		FrKeyQryHost kqh;
		memset( &kqh, 0, sizeof( kqh ) ) ;
		kqh.fd = GetKeyQuerySocket();
		memcpy(&kqh.host, &KQS_Addr, sizeof(SOCKADDR_IN));
		int retval;

		// REVIEW - should we call frKeyQryFetchPubKeyWithStatus for more info?

		ZkBool bFound = false ;
		ZkBool bValid = false ;

		retval = frKeyQryFetchPubKeyWithStatus(&err, m_kq, in_uKeyVersion,
								   in_keytype, exp, kqh, 3, in_nTimeout,
								   out_pKey, &bFound, &bValid );

		if ( bFound && bValid )
		{
			IZkConsoleMsg::ShowMsg( 
				IDS_CONS_DOWNLOAD_KEYS_EX, 
				GetStringFromKeyType( in_keytype ).c_str(),
				in_strEntityName );
		}
		else if ( retval == 1 )
		{
			// KeyQuery Server contacted but key not found!

			IZkConsoleMsg::ShowMsg( 
				IDS_CONS_KEY_NOT_ON_SERVER, 
				GetStringFromKeyType( in_keytype ).c_str(), 
				in_strEntityName ) ;
		}	
		else
		{
			// Probably a network error reaching the server

			IZkConsoleMsg::ShowMsg( 
				IDS_CONS_DOWNLOAD_KEYS_FAIL, 
				GetStringFromKeyType( in_keytype ).c_str(), 
				in_strEntityName, 
				err.getdetail().c_str() ) ;
		}

		return ( bFound && bValid ) ;
	}

	return true ;
}


bool
RKeyQry::GetKeyFromCache( 
	DZkError& err, 
	const FrEntIdType	in_entity, 
	const char*			in_strEntityName, 
	const unsigned int	in_uKeyVersion, 
	const FrPubKeyType  in_keytype, 
	OUT FrCryptPubKey*	out_pKey )
{
	if (( in_strEntityName == NULL ) ||
		( out_pKey == NULL ))
	{
		ZKASSERT( !"Invalid Arguments to RKeyQry::GetKeyFromCache" ) ;
		return false ;
	}

	FrKeyQryState state;
	FrKeyQryQueryExp kqry;
	FrEntIdExp exp;
	memset( &kqry, 0, sizeof( kqry ) ) ;
	memset( &exp, 0, sizeof( exp ) ) ;

	exp.type = in_entity;
	strncpy(exp.name, in_strEntityName, FRENT_NAME_SIZE);

	memset(&kqry, 0, sizeof(kqry));
	kqry.keyVersion = in_uKeyVersion;
	kqry.keyType = in_keytype;
	memcpy(&kqry.keyOwner, &exp, sizeof(FrEntIdExp));

	if (!frKeyQryCacheFindKey(&err, 
							m_kq, 
							&kqry, 
							&state, 
							out_pKey))
	{   
		return false;
	}

	if (state != FRKEYQRY_ST_VALID) 
	{
		return false ;
	}

	return true ;
}

bool 
RKeyQry::StartKeyQueryListeningSocket()
{
	// Someone may have called StartListeningSocket() manually, instead of simply calling GetKey().
	// This may be the case if they needed the actualy SOCKET value beforehand
	DZkError err;

	if( m_Socket != NULL )
	{
		ZKASSERT("We are in fact initialised" && m_Socket != NULL);
		return true;
	}

	ZKVERIFY(zkSocketNew(&err, AF_INET, SOCK_DGRAM, IPPROTO_UDP, &m_Socket));

	if(m_Socket == NULL)
	{
		ZKASSERT("RKeyQry::StartKeyQueryListeningSocket() - m_Socket = NULL" && 0);
		return false ;
	}

	return true ;
}

ZkSocketDescriptor 
RKeyQry::GetKeyQuerySocket()
{
	if ( m_Socket == NULL )
	{
		StartKeyQueryListeningSocket() ;
	}

	DZkError err;
	ZkSocketDescriptor retval;

	if (m_Socket == NULL) 
	{
		ZKASSERT( false ) ;
		return INVALID_SOCKET;
	}

	if (!zkSocketGetFD(&err, m_Socket, &retval)) 
	{
		ZKASSERT(0);
	}

	return retval;
}

void
RKeyQry::GetMultipleKeys(FrEntIdType			 in_EntityType, 
						 const WNetInfoDatumVec& in_EntitiesToLookup, 
						 FrPubKeyType			 in_keytype            // Type of key to lookup
						 ) throw (RZkException)
{

	ZKASSERT( in_keytype != FRCRYPT_KEYTYPE_UNUSED_ ) ;

	FrCryptPubKey	unused = NULL;

	DZkError	err;
	for(WNetInfoDatumVec::const_iterator i = in_EntitiesToLookup.begin(); i != in_EntitiesToLookup.end(); ++i)
	{	
		ZKSTD::string strEntityName ;
		(*i).GetEntityName( strEntityName ) ;

		ZKASSERT( !strEntityName.empty() ) ;

		ZKTRACE2("Getting %s key for %s\n",
					( in_keytype == FRCRYPT_KEYTYPE_CRYPTION ) ? "Crypto" : "Signature", 
					strEntityName.c_str());

		if(!GetKey(err, in_EntityType, 1, strEntityName.c_str(), 0, in_keytype, &unused))
		{
			ZKTRACE2("Key lookup failed for %s! [%s]\n", strEntityName.c_str(), err.getdetail().c_str() );
		}
	}
}

void 
RKeyQry::DownLoadKeysIfRequired() throw (RZkException)
{
	time_t time1 = 0 ;
	time_t time2 = 0 ;

	if ( IZkRegistry::GetFlag( "SkipAipKeyLookupOnStartup" ) )
	{
		IZkConsoleMsg::ShowMsg( 
			CONS_DEBUG, 
			CONSIMAGE_CONNECTION, 
			IDS_CONS_JUST_STRING, 
			"Client is configured to skip downloading AIP keys at startup" ) ;

		return ;
	}

	IZkConsoleMsg::ShowMsgForceUpdate( CONS_IMPORTANT, CONSIMAGE_NORMAL, IDS_CONS_DOWNLOAD_KEYS );
	time( &time1 ) ;

	//	See bug#572, 690 for detailed explanation of why we don't put this in a seperate thread after
	//	CWinThread* myWinThread = AfxBeginThread(KeyProc, (void*)&m_KeyStuff, THREAD_PRIORITY_HIGHEST);
	//	if(myWinThread) WaitForSingleObject(myWinThread->m_hThread, INFINITE);


	// If I get only the keys for those aips which aren't marked as down, it will go faster since frKeyQryFetchPubKey() is pretty slow
	// We probably won't need those keys, since downed aips aren't used in routes.
	// The only issue that brings up is the case where:
	// - An aip is marked as down AND we don't have his key
	// - Some time later it is marked as up through an Niqs update
	// - If the key is not in our cache and it is used as a hop in a route create TO THE KEY QUERY SERVER, that route create will fail
	// So when would we get his key? If the aip is marked as up the next time the client was started.
	// we'll get his key then. But if the client runs for so long, and aips get added for which we don't have their key
	// yet, it could be problematic....

	// try to get all the keys for the aips using user specified key query server
	WNetInfoDatumVec lst ;
	ZkService::GetINiqs()->GetServersOfType( FRENT_TYPE_AIP, lst, false );

	GetMultipleKeys( FRENT_TYPE_AIP, lst, FRCRYPT_KEYTYPE_CRYPTION ) ;
	GetMultipleKeys( FRENT_TYPE_AIP, lst, FRCRYPT_KEYTYPE_SIGNING ) ;

	// Save the latest state of the Key cache now that we've made sure it is up to date.
	ZkService::GetICrypto()->GetKeyCacheIntoFile(ZKFileConfiguration::GetKeyCacheFilePath().c_str()) ;


	time( &time2 ) ;

	// This will put the time in the console
	IZkConsoleMsg::ShowMsg( CONS_DEBUG, CONSIMAGE_WARNING, IDS_CONS_AIP_KEY_TIME, time2 - time1 );
}


ZKSTD::string
RKeyQry::GetStringFromKeyType( const FrPubKeyType in_keytype ) const
{
	// Todo - move to a CAT file

	if ( in_keytype == FRCRYPT_KEYTYPE_CRYPTION )
	{
		return ZKSTD::string( "encryption" ) ;
	}
	else
	{
		ZKASSERT( in_keytype == FRCRYPT_KEYTYPE_SIGNING );
		return ZKSTD::string( "signature" ) ;
	}
}