// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "StdAfx.h"
#include "RNetworkPerformance.h"

#include <clientdebug.h>
#include <IZkConsoleMsg.h>

#include <IRouteManager.h>
#include <INymRoute.h>
#include <INymCentral.h>
#include <INiqs.h>
#include <ZkService.h>
#include <ZkErrorHandler.h>
#include <IZKErrorReporter.h>
#include <IZkException.h>
#include <RZkException.h>
#include <zkAlloc.h>
#include <WRoute.h>
#include <ZKChunk.h>

#include <TraceControl.h>
#include <IZkRegistry.h>		
#include <zkrand.h>

#include <ZkIcmp.h>			// Until Ping is available


#define MAX_PING_TIMEOUT 1000 

#define FAILED_TEST_TIME 0xFFFFFFFF

#define MAX_AIPS_IN_BLACKLIST 6

#define NETWORK_PERF_PERSIST_VERSION 1

#define MIN_SUCCEEDED_PERFORMANCE_RESULT 50 // We give this score to a server that fails with the worst
											// possible timeout, whereas a successful server gets
											// MAX_PERFORMANCE_RESULT

#define MIN_BACKGROUND_PING_START_TIME_MS 1000

#ifdef _DEBUG
#define MAX_BACKGROUND_PING_START_TIME_MS 1000 * 45		// Run the probe more quickly for testing purposes
#else
#define MAX_BACKGROUND_PING_START_TIME_MS 1000 * 60 * 15	// Some time in the first fifteen minutes of running we will perform the background ping
#endif


// Performance Task contains just the information necessary to perform
// and report the result of a ping test on an entity

class PerformanceTask
{
public:
	PerformanceTask() : m_entityID( 0 ), m_PerformanceResult( 0 ) { memset( &m_addr, 0, sizeof(sockaddr_in) ) ; }

	// Construct a ping task to perform on a provided Entity
	PerformanceTask( const EntityPerformanceInfo& in_entity ) 
				: m_entityID( in_entity.m_entityID )
				, m_PerformanceResult( 0 ) { memset( &m_addr, 0, sizeof(sockaddr_in) ) ; }


	PerformanceTask& operator=( const EntityPerformanceInfo& in_entity )
	{
		m_entityID = in_entity.m_entityID ;
		memset( &m_addr, 0, sizeof(sockaddr_in) ) ;
		return *this ;
	}
	
	void SetAddressOfServerToTest( sockaddr_in in_addr )
	{
		memcpy( &m_addr, &in_addr, sizeof( m_addr ) ) ;
	}

	uint32_t m_entityID ;
	sockaddr_in m_addr ;
	uint32_t m_PerformanceResult ;
} ;


RNetworkPerformance::RNetworkPerformance()
	: m_cntPerformanceTask( 0 )
	, m_pPerformanceTask( NULL )
	, m_maxPingTimeout( MAX_PING_TIMEOUT )
	, m_pNetworkProbeThread( NULL )
	, m_cntRouteCreateAttempts( 0 )
	, m_cntRouteCreateSuccess( 0 )
	, m_cntCoreServerConnectCloudAttempts( 0 )
	, m_cntCoreServerConnectCloudSuccess( 0 )
	, m_pBackgroundPingTimer(NULL)
{
	unsigned long timeout = MAX_PING_TIMEOUT ;
	IZkRegistry::GetDWORD( "PingTimeout", timeout ) ;

	// If what we've read in the registry is between 0 and MAX_PING_TIMEOUT
	// set m_maxPingTimeout to that value
	if (( timeout <= MAX_PING_TIMEOUT ) &&
		( timeout > 0 ) )
	{
		m_maxPingTimeout = timeout ;
	} 
	else
	{
		ZKASSERT( false ) ;
	}	
}

RNetworkPerformance::~RNetworkPerformance()
{
	KillBackgroundPingTimer() ;
	
	// Close down route create thread if it exists
	if (m_pNetworkProbeThread)
	{
		DZkError derr ;
		void *rval;
		ZKVERIFY(zkThreadJoin(&derr, &m_pNetworkProbeThread, &rval));
	}

	SAFEDELETEARRAY( m_pPerformanceTask ) ;
}

void
RNetworkPerformance::SelfTest()
{
	// Move this test code elsewhere
	ZKASSERT( CalculateScore( false, 0 ) == 0 ) ;
	ZKASSERT( CalculateScore( true, 0 ) == MAX_PERFORMANCE_RESULT ) ;
	ZKASSERT( CalculateScore( true, m_maxPingTimeout ) == MIN_SUCCEEDED_PERFORMANCE_RESULT ) ;
}


uint32_t 
RNetworkPerformance::GetRouteCreateStat( StatisticType in_whichStat )
{
	return m_routeCreateStats.GetStat( in_whichStat ) ;
}

void 
RNetworkPerformance::AddRouteCreateTime( uint32_t in_msCreateTime )
{
	m_routeCreateStats.AddDataPoint( in_msCreateTime ) ;
}

void 
RNetworkPerformance::IncrementRouteCreateAttempts()
{
	++m_cntRouteCreateAttempts ;
}

void
RNetworkPerformance::IncrementRouteCreateSuccess()
{
	++m_cntRouteCreateSuccess ;
}

void 
RNetworkPerformance::IncrementCoreServerConnectCloudAttempts()
{
	++m_cntCoreServerConnectCloudAttempts ;
}

void
RNetworkPerformance::IncrementCoreServerConnectCloudSuccess()
{
	++m_cntCoreServerConnectCloudSuccess ;
}

void 
RNetworkPerformance::ResetAllStats()
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs) ;

	m_EntityMap.clear() ;

	m_routeCreateStats.Reset() ;
	
	m_cntRouteCreateAttempts = 0 ;
	m_cntRouteCreateSuccess = 0 ;

	m_cntCoreServerConnectCloudAttempts = 0;
	m_cntCoreServerConnectCloudSuccess = 0;

	m_AipBlackList.clear() ;
}

void
RNetworkPerformance::ShowCurrentStats( ConsSeverity in_Severity )
{	
	ZKS::Lock_Guard<WZkMutex> gate(m_cs) ;

	ZKTRACE4( "Routing Statistics for this session (milliseconds) - Best %d, Average %d, Worst %d, Route Creates %d\n", 
			m_routeCreateStats.GetStat( MinDatum ),
			m_routeCreateStats.GetStat( AvgOfDatums ),
			m_routeCreateStats.GetStat( MaxDatum ),
			m_routeCreateStats.GetStat( CountDatums ) ) ;

	IZkConsoleMsg::ShowMsg( in_Severity, 
							CONSIMAGE_COMP_TERM,
							IDS_CONS_ROUTE_CREATE_STATS,
							m_routeCreateStats.GetStat( MinDatum ),
							m_routeCreateStats.GetStat( AvgOfDatums ),
							m_routeCreateStats.GetStat( MaxDatum ) ) ;

	if ( m_cntRouteCreateAttempts > 0 )
	{
		ZKASSERT( m_cntRouteCreateAttempts >= m_cntRouteCreateSuccess ) ;

		char strMsg[ 128 ] = { 0 } ;

		snprintf( 
			strMsg, 
			128,
			"Route Creation success rate %d%% (total %d attempts)", 
			( m_cntRouteCreateSuccess * 100 ) / m_cntRouteCreateAttempts,
			m_cntRouteCreateAttempts ) ;

		IZkConsoleMsg::ShowMsg( in_Severity, 
								CONSIMAGE_COMP_TERM,
								IDS_CONS_JUST_STRING,
								strMsg ) ;
	}

	if ( m_cntCoreServerConnectCloudAttempts > 0 )
	{
		ZKASSERT( m_cntCoreServerConnectCloudAttempts >= m_cntCoreServerConnectCloudSuccess ) ;

		char strMsg[ 128 ] = { 0 } ;

		snprintf( 
			strMsg, 
			128,
			"Connections through a route success rate %d%% (total %d attempts)", 
			( m_cntCoreServerConnectCloudSuccess * 100 ) / m_cntCoreServerConnectCloudAttempts,
			m_cntCoreServerConnectCloudAttempts ) ;

		IZkConsoleMsg::ShowMsg( in_Severity, 
								CONSIMAGE_COMP_TERM,
								IDS_CONS_JUST_STRING,
								strMsg ) ;
	}

	ShowAipBlackList( in_Severity ) ;


#if 0
	// As we may be saving this information into a log file (#8210)
	// we need a stronger justification for the security risk that
	// it posses.

	// Show Nym Last Successful Routes

	size_t cntNyms = ZkService::GetINymCentral()->GetNumberOfNyms() ;

	for ( size_t i = 1 ; i < cntNyms ; i++ )
	{
		WRoute lastRoute ;

		INymRoute* pNymRoute = NULL ; 
		ZkService::GetINymCentral()->GetNym( IID_INYMROUTE, (void**) &pNymRoute, i ) ;

		if ( pNymRoute != NULL )
		{
			pNymRoute->GetLastSuccessfulRoute( lastRoute ) ;

			if ( lastRoute.GetSize() > 0 )
			{			
				ZKSTD::string strNymName ;
				ZKSTD::string strRoute ;
				ZkService::GetINymCentral()->GetNymName( i, strNymName ) ;

				strRoute = ZkService::GetIRouteManager()->GetRouteDesc( lastRoute ) ;

				IZkConsoleMsg::ShowMsg(
					CONS_DEBUG,
					CONSIMAGE_COMP_TERM,
					IDS_CONS_LAST_NYM_ROUTE,
					strNymName.c_str(),
					strRoute.c_str() ) ;
			}
		}
	}

#endif

	// show the Aip Performance
	IZkConsoleMsg::ShowMsg( in_Severity, 
							CONSIMAGE_COMP_TERM,
							IDS_CONS_JUST_STRING,
							"Aip Performance Table Listing" ) ;

	for ( EntityPerformanceMap::iterator it = m_EntityMap.begin() ;
		  it != m_EntityMap.end() ;
		  it++ )
	{
		const EntityPerformanceInfo& entityInfo = (*it).second ;

		ZKSTD::string entityName ;		
		WNetInfoDatum entity ;
		if ( !ZkService::GetINiqs()->GetEntityById( entityInfo.m_entityID, entity ) )
		{
			ZKTRACE1( "Entity %d has not info in the NIQS info", entityInfo.m_entityID ) ;
			entityName = "Unknown Server" ;
		}
		else
		{
			entity.GetEntityName( entityName ) ;
		}

		ZKTRACE4( "Entity %d (%s) Attempts: %d, Score %d\n",
				entityInfo.m_entityID,
				entityName.c_str(),
				entityInfo.m_cntAttempts,
				entityInfo.GetScore() ) ;

		//Entity %d (%s) Connection results: Best %d, Average %d, Worst %d Failed %d

		IZkConsoleMsg::ShowMsg(
			in_Severity,
			CONSIMAGE_COMP_TERM,
			IDS_CONS_ENTITY_PERF,
			entityInfo.m_entityID,
			entityName.c_str(),
			entityInfo.m_cntAttempts,
			entityInfo.GetScore() ) ;
	}
}


void
RNetworkPerformance::PreparePerformanceTests()
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs) ;

	if ( m_EntityMap.size() == 0 )
	{
		// nothing to do
		return ;
	}

	// Copy the entities that we want to test into an
	// array

	SAFEDELETEARRAY( m_pPerformanceTask ) ;

	m_pPerformanceTask = new PerformanceTask[ m_EntityMap.size() ] ;

	size_t i = 0 ;

	for ( EntityPerformanceMap::iterator it = m_EntityMap.begin() ;
		  it != m_EntityMap.end() ;
		  it++ )
	{
		const EntityPerformanceInfo& entityInfo = (*it).second ;
		WNetInfoDatum entity ;

		m_pPerformanceTask[ i ] = (*it).second ;

		// Need to address of the entity
		if ( ZkService::GetINiqs()->GetEntityById( entityInfo.m_entityID, entity ) )		
		{
			m_pPerformanceTask[ i ].SetAddressOfServerToTest( entity.GetSockAddr() ) ;
		}
		else
		{
			// We shouldn't perform this test because we don't know where to go
			m_pPerformanceTask[ i ].m_entityID = 0 ;
		}
	
		i++ ;
	}

	m_cntPerformanceTask = m_EntityMap.size() ;

}

void
RNetworkPerformance::PerformTests()
{
	if (( m_pPerformanceTask == NULL ) ||
		( m_cntPerformanceTask == 0 ))
	{
		ZKASSERT( false ) ;
		return ;
	}	
	// Perform the test

	for ( size_t k = 0 ; k < m_cntPerformanceTask ; k++ )
	{
		PerformTest( m_pPerformanceTask[ k ] ) ;
	}

}

void
RNetworkPerformance::ProcessPerformanceResults()
{
	// We lock our status information just long enough to update our results
	ZKS::Lock_Guard<WZkMutex> gate(m_cs) ;

	if (( m_pPerformanceTask == NULL ) ||
		( m_cntPerformanceTask == 0 ))
	{
		ZKASSERT( false ) ;
		return ;
	}


	// Merge results back into table

	for ( size_t k = 0 ; k < m_cntPerformanceTask ; k++ )
	{
		if ( m_pPerformanceTask[k].m_entityID == 0 )
		{
			// Ignore this, we didn't know where the server
			// was
			continue ;
		}

		if ( m_pPerformanceTask[k].m_PerformanceResult == 0 )
		{
			// Guess we didn't complete the test?....
            // Or we got a zero ping --ol
			ZKTRACE1("PerformanceTask %d got a zero result. This is probably normal.\n", k);
		}

        EntityPerformanceMap::iterator foundEntity = m_EntityMap.find( m_pPerformanceTask[k].m_entityID ) ;
        
        if ( foundEntity == m_EntityMap.end() )
        {
				// Entity is no longer in the table.  This could
            // happen if the stats were reset during the update
            continue ;
        }					
        
        EntityPerformanceInfo& entityinfo = (*foundEntity).second ;
        bool bTestSucceeded = ( m_pPerformanceTask[k].m_PerformanceResult != FAILED_TEST_TIME ) ;
        
        entityinfo.AddPerformanceDatum( 
            CalculateScore( bTestSucceeded, m_pPerformanceTask[k].m_PerformanceResult ) );
    }
    

	// Clean up the task
	SAFEDELETEARRAY( m_pPerformanceTask ) ;
	m_cntPerformanceTask = 0 ;
}

uint32_t 
RNetworkPerformance::CalculateScore( bool in_bSuccess, uint32_t in_msPerformanceResult ) 
{
	if ( in_bSuccess )
	{
		const uint32_t Range = MAX_PERFORMANCE_RESULT - MIN_SUCCEEDED_PERFORMANCE_RESULT ;

		// Now convert the 0 - m_maxPingTimeout scale to a value
		// between MIN_SUCCEEDED_PERFORMANCE_RESULT and MAX_PERFORMANCE_RESULT
		uint32_t PingScore = ( in_msPerformanceResult * Range ) / m_maxPingTimeout ;

		ZKASSERT( PingScore <= Range ) ;

		// A high PingFactor is bad, so it takes away from the score we return

		uint32_t retVal = MAX_PERFORMANCE_RESULT - PingScore ;

		ZKASSERT( retVal >= MIN_SUCCEEDED_PERFORMANCE_RESULT ) ;
		ZKASSERT( retVal <= MAX_PERFORMANCE_RESULT ) ;

		ZKCTRACE2( 
			TRACE_ROUTEPICKER, 
			"Ping value %d - Score calculated %d\n", 
			in_msPerformanceResult, 
			retVal ) ;

		return retVal ;
	}
	else
	{
		return 0 ;
	}
}


void
RNetworkPerformance::PerformTest( PerformanceTask& io_task )
{
	if ( io_task.m_entityID == 0 )
	{
		// Invalid task
		io_task.m_PerformanceResult = FAILED_TEST_TIME ;
		return ;
	}

	ZkIcmp		Icmp;
	DZkError	err;

	unsigned long  pingResult = FAILED_TEST_TIME ;

	in_addr where ;
	where.s_addr = io_task.m_addr.sin_addr.s_addr ;

	if ( !Icmp.Ping(
				err, 
				inet_ntoa( where ), 
				pingResult, 
				m_maxPingTimeout ) )
	{	
		io_task.m_PerformanceResult = FAILED_TEST_TIME ;
	}
	else
	{
		ZKASSERT( pingResult <= m_maxPingTimeout ) ;
// Removed 0 is a very valid ping --ol       
//		ZKASSERT( pingResult > 0 ) ;
		io_task.m_PerformanceResult = pingResult ;
	}
}

void
RNetworkPerformance::UpdateEntityList()
{
	// New Entities might be added to the network, this
	// forces us to update our structures to recognize new 
	// entities.

	ZKS::Lock_Guard<WZkMutex> gate(m_cs) ;

	try
	{
		WNetInfoDatumVec aips ;

		// We get the list of Aips.
		// We include Core Aips.  Even though the
		// core aips are all probably at the same location relative to the
		// user this is needed for 1 hop Core Aip routes and the scores are interesting
		// information to collect from users having performance problems.
		ZkService::GetINiqs()->GetListOfAips( true, true, false, aips ) ;

		for(WNetInfoDatumVec::const_iterator it = aips.begin(); it != aips.end(); ++it)
		{
			const WNetInfoDatum& aip = *it ;

			if ( m_EntityMap.find( aip.GetId()) == m_EntityMap.end() )
			{
				// Entity is not already in the list
				m_EntityMap.insert( 
					EntityPerformanceMap::value_type( 
											aip.GetId(), 
											EntityPerformanceInfo( aip ) ) );
			}
		
			// We assume that entities that we already have don't need any kind of
			// update - if there was some actually useful status information then
			// that wouldn't be a correct assumption.
		}

		// Initially we never remove old entities that are taken out of the network.
	}
	catch( IZkException& ex )
	{
		// This can throw on test networks when there are no up aips
		// We can swallow up the error and keep going.
		ZKTRACE1("UpdateEntityList has a problem- %s",  ex.GetFormattedString().c_str() ) ;
	}
}

uint32_t 
RNetworkPerformance::GetEntityPerformanceResult( uint32_t in_Entity )
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs) ;

	EntityPerformanceMap::iterator foundEntity = m_EntityMap.find( in_Entity ) ;

	if ( foundEntity == m_EntityMap.end() )
	{
		return DEFAULT_PERFORMANCE_RESULT ;
	}					

	EntityPerformanceInfo& entityinfo = (*foundEntity).second ;

	return entityinfo.GetScore() ;
}


void * ZK_THREAD_CALLBACK 
RNetworkPerformance::BackgroundProbe(void * in_pThis)
{
	RNetworkPerformance* pThis = (RNetworkPerformance*) in_pThis ;

	time_t t1 ;
	time( &t1 ) ;

	pThis->PerformTests() ;
	pThis->ProcessPerformanceResults() ;

	time_t t2 ;
	time( &t2 ) ;

	IZkConsoleMsg::ShowMsg( CONSIMAGE_COMP_TERM, IDS_CONS_FINISH_PROBE, t2 - t1 ) ;

	return 0 ;
}

void 
RNetworkPerformance::CalculatePerformance( bool in_bStartImmediately )
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs) ;

	if ( m_pPerformanceTask != NULL )
	{
		ZKTRACE0( "We appear to aready have a background probe in progress, not going to start another\n" ) ;
		return ;
	}

	if ( !in_bStartImmediately && ( m_cntRouteCreateAttempts > 0 ) )
	{
		// See bug 4696 - security reasons not to ping right after
		// start up.  Note that the entitymap will be empty when
		// you run the setup wizard so we will start the test immediately

		uint32_t rn = 0 ;
		ZKVERIFY( zkRandGetUInt32InRange(&m_err, MIN_BACKGROUND_PING_START_TIME_MS, MAX_BACKGROUND_PING_START_TIME_MS, &rn) );

		StartBackgroundPingTimer( rn ) ;
		return ;
	}

	IZkConsoleMsg::ShowMsg( CONSIMAGE_COMP_TERM, IDS_CONS_START_PROBE ) ;

	PreparePerformanceTests() ;

	DZkError derr ;
	zkThreadCreate(&derr, &m_pNetworkProbeThread, &RNetworkPerformance::BackgroundProbe, this) ;	
}


void
RNetworkPerformance::GetAipIdsSortedByRating( 
	unsigned int ** out_pEntities, 
	size_t & out_cntEntities )
{
	// PERFORMANCE NOTE: This involves extra memory copies of network information
	// and potentially redundant sorting.  It is possible to optimize this
	// by keeping the sorted list and only erasing it when the entity list has
	// changed which is infrequent
	
	*out_pEntities = NULL ;
	out_cntEntities = 0 ;

	// Get the (unsorted) list of Aips from NIQS
	WNetInfoDatumVec aips ;

	// We don't include down servers in this list

	// We include core aips for the case of a single hop unauthenticated route

	ZkService::GetINiqs()->GetListOfAips( true, true, false, aips ) ;

	size_t cntAips = aips.size() ;
	
	if ( cntAips == 0 ) 
	{
		// crap, we don't have any aips to deal with - this
		// error can travel up
		ZKTRACE0("Can't sort because there aren't any aips\n" ) ;
		return ;
	}

	ZKVERIFY( zkMemAlloc( NULL, (void**)out_pEntities, sizeof( uint32_t ) * cntAips )) ;

	out_cntEntities = cntAips ;

	uint32_t * pAips = *out_pEntities ;

	ZKASSERT( pAips != NULL ) ;

	for ( size_t i = 0 ; i < cntAips ; i++ )
	{
		// Initialize as aips in the order
		pAips[ i ] = aips[ i ].GetId() ;
	}

	qsort( pAips, cntAips, sizeof( uint32_t ), RNetworkPerformance::CompareAipsByRating );

/*
	// Trace results....

	ZKTRACE0( "Results of sorting:\n\n" ) ;

	for ( i = 0 ; i < cntAips ; i++ )
	{
		uint32_t r1 = ZkService::GetINetworkPerformance()->GetEntityPerformanceResult( pAips[ i ] ) ;
		ZKTRACE2( "Entity %d Rating %d\n", pAips[ i ], r1 ) ;
	}
*/


}


int 
RNetworkPerformance::CompareAipsByRating( const void *elem1, const void *elem2 )
{
	uint32_t e1 = *(uint32_t*) elem1 ;
	uint32_t e2 = *(uint32_t*) elem2 ;

	ZKASSERT( e1 != 0 ) ;
	ZKASSERT( e2 != 0 ) ;

	uint32_t r1 = ZkService::GetINetworkPerformance()->GetEntityPerformanceResult( e1 ) ;
	uint32_t r2 = ZkService::GetINetworkPerformance()->GetEntityPerformanceResult( e2 ) ;

	if ( r1 > r2 )
	{
		return -1 ;
	}
	else if ( r1 < r2 )
	{
		return 1 ;
	}
	else
	{
		return 0 ;
	}
}


void     
RNetworkPerformance::AddRouteToBlackList( const WRoute& in_route )
{
	ZKASSERT( in_route.GetSize() > 0 ) ;

	ZKSTD::vector< uint32_t >::iterator it ;

	for ( size_t i = 0 ; i < in_route.GetSize() ; i++ )
	{
		// Is this Aip already in our list

		
		for ( it = m_AipBlackList.begin() ; it != m_AipBlackList.end() ; it++ )
		{
			if ( *it == in_route.GetHop( i ) )		
			{
				// We intentionally delete the existing hop rather than not inserting
				// the new hop because the position in the list acts as a aging
				// progress
				m_AipBlackList.erase( it ) ;
				break ;
			}
		}

		// Add the item
		m_AipBlackList.push_back( in_route.GetHop( i ) ) ;
	}

	// Truncate the oldest entries if we have too many hops in black list

	while( m_AipBlackList.size() > MAX_AIPS_IN_BLACKLIST )
	{
		it = m_AipBlackList.begin() ;
		m_AipBlackList.erase( it ) ;
	}

	if ( TRACE_ROUTEPICKER )
	{
		ShowAipBlackList() ;
	}
}

bool     
RNetworkPerformance::IsRouteBlackListed( const WRoute& in_route ) const
{
	// We return true if ANY of the hops in the route are in the black list
	
	ZKASSERT( in_route.GetSize() > 0 ) ;

	ZKSTD::vector< uint32_t >::const_iterator it ;

	for ( size_t i = 0 ; i < in_route.GetSize() ; i++ )
	{
		for ( it = m_AipBlackList.begin() ; it != m_AipBlackList.end() ; it++ )
		{
			if ( *it == in_route.GetHop( i ) )		
			{
				// Is this Aip already in our list
				return true ;
			}
		}
	}

	return false ;
}

void
RNetworkPerformance::ShowAipBlackList( ConsSeverity in_Severity )
{
	// Showing the black list
	if ( m_AipBlackList.size() > 0 )
	{
		ZKSTD::string strBlackList( "Black list: " ) ;

		for ( size_t i = 0 ; i < m_AipBlackList.size() ; i++ )
		{
			ZKSTD::string entityName ;		
			WNetInfoDatum entity ;
			if ( !ZkService::GetINiqs()->GetEntityById( m_AipBlackList[ i ], entity ) )
			{
				entityName = "Unknown Server" ;
			}
			else
			{
				entity.GetEntityName( entityName ) ;
			}

			strBlackList += entityName ; 
			strBlackList += " " ;
		}
		IZkConsoleMsg::ShowMsg( in_Severity, 
								CONSIMAGE_COMP_TERM,
								IDS_CONS_JUST_STRING,
								strBlackList.c_str() ) ;

	}
	else
	{
		IZkConsoleMsg::ShowMsg( in_Severity, 
								CONSIMAGE_COMP_TERM,
								IDS_CONS_JUST_STRING,
								"Aip Black List is empty" ) ;		
	}
}

void
RNetworkPerformance::SaveRouteCreateChunk( WMemBuf& io_Buffer ) throw(RZkException)
{
	ZKChunkWriter cw( io_Buffer, PERF_ROUTE_CREATE_CHUNK_ID, 1 ) ;

	// Size of elements
	cw.WriteUInt32( m_cntRouteCreateAttempts ) ;
	cw.WriteUInt32( m_cntRouteCreateSuccess ) ;
}

void
RNetworkPerformance::SaveAipPerformanceChunk( WMemBuf& io_Buffer ) throw(RZkException)
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs) ;
	ZKChunkWriter cw( io_Buffer, PERF_AIPS_CHUNK_ID, 1 ) ;

	// Size of elements
	cw.WriteUInt32( m_EntityMap.size() ) ;

	for ( EntityPerformanceMap::iterator it = m_EntityMap.begin() ;
		  it != m_EntityMap.end() ;
		  it++ )
	{
		const EntityPerformanceInfo& entityInfo = (*it).second ;

		cw.WriteUInt32( entityInfo.m_entityID ) ;
		cw.WriteUInt32( entityInfo.m_cntAttempts ) ;
		cw.WriteUInt32( entityInfo.m_Score ) ;
	}
}

void 
RNetworkPerformance::Save( WMemBuf& io_Buffer ) throw(RZkException)
{
	ZKChunkWriter cw( io_Buffer, NETWORK_PERF_CHUNK_ID, NETWORK_PERF_PERSIST_VERSION ) ;
	SaveAipPerformanceChunk( io_Buffer ) ;
	SaveRouteCreateChunk( io_Buffer ) ;
}

void 
RNetworkPerformance::LoadRouteCreateChunk( ZKChunkReader& io_ChunkReader ) throw(RZkException)
{
	// Clear out existing settings
	
	if ( io_ChunkReader.GetChunkID() != PERF_ROUTE_CREATE_CHUNK_ID ) 
	{
		// Corruption or bug in the load code, we must load in the correct order
		ZKASSERT( false ) ;
		throw RZkException("RNetworkPerformance::LoadRouteCreateChunk", 
							 "Corrupt File",
							 IDS_CORRUPT_FREEDOM_DATA ) ;
	}

	m_cntRouteCreateAttempts = io_ChunkReader.ReadUInt32() ;
	m_cntRouteCreateSuccess = io_ChunkReader.ReadUInt32() ;
}


void 
RNetworkPerformance::LoadAipPerformanceChunk( ZKChunkReader& io_ChunkReader ) throw(RZkException)
{
	// Clear out existing settings
	
	if ( io_ChunkReader.GetChunkID() != PERF_AIPS_CHUNK_ID ) 
	{
		// Corruption or bug in the load code, we must load in the correct order
		ZKASSERT( false ) ;
		throw RZkException("RNetworkPerformance::LoadAipPerformanceChunk", 
							 "Corrupt File",
							 IDS_CORRUPT_FREEDOM_DATA ) ;
	}

	// Version 1
	uint32_t cntAipPerfDatums = io_ChunkReader.ReadUInt32() ;

	for ( uint32_t i = 0 ; i < cntAipPerfDatums ; i++ )
	{
		// fill in the map

		EntityPerformanceInfo entityPerformance ;

		entityPerformance.m_entityID = io_ChunkReader.ReadUInt32() ;
		entityPerformance.m_cntAttempts = io_ChunkReader.ReadUInt32() ;
		entityPerformance.m_Score = io_ChunkReader.ReadUInt32() ;

		ZKASSERT( entityPerformance.m_entityID != 0 ) ;
		ZKASSERT( entityPerformance.m_Score <= MAX_PERFORMANCE_RESULT ) ;

		if ( m_EntityMap.find( entityPerformance.m_entityID) == m_EntityMap.end() )
		{
			// Entity is not already in the list
			m_EntityMap.insert( 
				EntityPerformanceMap::value_type( 
										entityPerformance.m_entityID, 
										entityPerformance) );
		}
		else
		{
			// Why is this aip already in the list?
			ZKASSERT( false ) ;
		}
	}
}

void 
RNetworkPerformance::Load( WReadBuffer& io_Buffer ) throw(RZkException)
{
	ResetAllStats() ;
	
	ZKChunkReader cr( io_Buffer ) ;

	if ( cr.GetChunkID() != NETWORK_PERF_CHUNK_ID ) 
	{
		// Corruption or bug in the load code, we must load in the correct order
		ZKASSERT( false ) ;
		throw RZkException("RNetworkPerformance::Load", 
							 "Corrupt File - Network Performance Chunk",
							 IDS_CORRUPT_FREEDOM_DATA ) ;
	}

	while ( !cr.IsEndOfChunk() )
	{
		ZKChunkReader cr2( cr.GetReadBuffer() );
		switch( cr2.GetChunkID() )
		{
			case( PERF_AIPS_CHUNK_ID ) :
				LoadAipPerformanceChunk( cr2 ) ;
				break ;

			case( PERF_ROUTE_CREATE_CHUNK_ID ) :
				LoadRouteCreateChunk( cr2 ) ;
				break ;

			default :
				// Future chunks - just ignore them
				ZKASSERT( false ) ;
				break ;
		}
	}
}

void
RNetworkPerformance::KillBackgroundPingTimer()
{	
	if( m_pBackgroundPingTimer )
	{
		// we could get a destroy _before_ our indicated timeout, handle that
		ZKVERIFY( zkITimerFree( &m_err, &m_pBackgroundPingTimer ) );
	}

	ZKASSERT( m_pBackgroundPingTimer == NULL ) ;
}

void
RNetworkPerformance::StartBackgroundPingTimer( uint32_t in_msTime )
{
	ZKASSERT( in_msTime != 0 ) ;

	if ( m_pBackgroundPingTimer != NULL )
	{
		// We already have a background ping in the queue- don't start another
		return  ;
	}

	ZKVERIFY( zkITimerNew( &m_err, 
						   &m_pBackgroundPingTimer,
						   RNetworkPerformance::BackgroundPingTimerCallback,
						   in_msTime ) ) ;
	ZKASSERT( m_pBackgroundPingTimer );
}

void 
RNetworkPerformance::HandleCalculatePerformanceTimer()
{
	KillBackgroundPingTimer() ;
	CalculatePerformance( true ) ;
}

void ZK_ITIMER_CALLBACK 
RNetworkPerformance::BackgroundPingTimerCallback(
					void *p, 
					unsigned int i1, 
					unsigned int i2, 
					unsigned long time)
{
	ZkService::GetINetworkPerformance()->HandleCalculatePerformanceTimer() ;
}

EntityPerformanceInfo::EntityPerformanceInfo() 
	: m_entityID( 0 )
	, m_Score( 0 )
	, m_cntAttempts( 0 )
{ 
} 

EntityPerformanceInfo::EntityPerformanceInfo( const WNetInfoDatum& in_Datum )
	: m_entityID( in_Datum.GetId())
	, m_Score( 0 )
	, m_cntAttempts( 0 )
{
}

void
EntityPerformanceInfo::AddPerformanceDatum( uint32_t in_NewScore ) 
{	
	m_cntAttempts++ ;

	if ( m_cntAttempts == 1 )
	{
		// First attempt - don't include historic value
		m_Score = in_NewScore ;
	}
	else
	{
		// We are weighing latest result as 50% of the score,
		// historic results as the other 50%.  This is because
		// the current situation is very relevant to the user

		m_Score = m_Score /2 + in_NewScore / 2 ;
	}

	ZKASSERT( m_Score <= MAX_PERFORMANCE_RESULT ) ;
}
