// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RNetworkPerformance.h

// The interface provides access to a service which 
// implements 


#ifndef RNETWORKPERFORMANCE_H_
#define RNETWORKPERFORMANCE_H_

#include "INetworkPerformance.h"
#include <RStatCollector.h>
#include <client.h>
#include <DZkError.h>
#include <WSyncTypes.h>
#include <WNetInfoDatum.h>

class PerformanceTask ;
class ZKChunkReader ;

// This class represents the performance information we have for a particular 
// Entity

class EntityPerformanceInfo
{
public:
	EntityPerformanceInfo() ; 

	// Constructor to create Performance Info for a particular NetInfoDatum
	EntityPerformanceInfo( const WNetInfoDatum& in_Datum ) ;

	// Record a result
	void AddPerformanceDatum( uint32_t in_NewScore ) ;

	uint32_t GetScore() const { return m_Score ; }

	uint32_t	m_entityID ;

	uint32_t    m_Score ;			// Current score, this is adjusted as new performance datums come in

	uint32_t    m_cntAttempts ;     // Total number of times that we have tried

	// Review:
	// Other stats could be collected - if we used it in a route, if it is
	// "fast" in practise, if we failed to create a route to it....

} ;


typedef ZKSTD::map<uint32_t, EntityPerformanceInfo>	EntityPerformanceMap ;		// aip id -> Performance

class RNetworkPerformance : public INetworkPerformance
{
public:
	RNetworkPerformance() ;
	virtual ~RNetworkPerformance() ;


	// Route Create time management

	// This function is used to feed the route create data to the 
	// service.
	// In future we may add argument for which hops were used
	// so that we collect route create times per aip
	virtual void	 AddRouteCreateTime( uint32_t in_msCreateTime ) ;
	virtual uint32_t GetRouteCreateStat( StatisticType in_whichStat ) ;

	// Route Create Statistics

	virtual void     IncrementRouteCreateAttempts() ;	// Total number of route creates attempted
	virtual void     IncrementRouteCreateSuccess() ;


	virtual void     IncrementCoreServerConnectCloudAttempts() ;
	virtual void     IncrementCoreServerConnectCloudSuccess() ;

	// Route Black List - this tracks routes that failed in the past so
	// that we can avoid the aips in future route creates

	virtual void     AddRouteToBlackList( const WRoute& in_route ) ;
	virtual bool     IsRouteBlackListed( const WRoute& in_route ) const ;

	// This functions sends the current network statistics to the console/trace window
	virtual void ShowCurrentStats( ConsSeverity in_Severity ) ;
	virtual void ResetAllStats() ;

	virtual void CalculatePerformance( bool in_bStartImmediately ) ;
	virtual void HandleCalculatePerformanceTimer() ;

	virtual void GetAipIdsSortedByRating( 
					unsigned int ** out_pEntities, 
					unsigned int & out_cntEntities ) ;

	// Entity performance management

	virtual void UpdateEntityList() ;
	virtual uint32_t GetEntityPerformanceResult( uint32_t in_Entity ) ;

	//virtual void BackgroundTestAips() ;

	// Persistence Methods
	virtual void Save( WMemBuf& io_Buffer ) throw(RZkException) ;
	virtual void Load( WReadBuffer& io_Buffer ) throw(RZkException) ;

	virtual void SelfTest() ;

private:
	void PreparePerformanceTests() ;
	void ProcessPerformanceResults() ;
	void PerformTests() ;
	void PerformTest( PerformanceTask& io_task ) ;

	static int CompareAipsByRating( const void *elem1, const void *elem2 ) ;
	static void * ZK_THREAD_CALLBACK BackgroundProbe(void * pThis);

	void ShowAipBlackList( ConsSeverity in_Severity = CONS_DEBUG ) ;
	void LoadAipPerformanceChunk( ZKChunkReader& io_ChunkReader ) throw(RZkException) ;
	void LoadRouteCreateChunk( ZKChunkReader& io_ChunkReader ) throw(RZkException) ;
	void SaveAipPerformanceChunk( WMemBuf& io_Buffer ) throw(RZkException) ;
	void SaveRouteCreateChunk( WMemBuf& io_Buffer ) throw(RZkException) ;

	uint32_t CalculateScore( bool in_bSuccess, uint32_t in_msPerformanceResult )  ;

	void StartBackgroundPingTimer( uint32_t in_msTime ) ;
	void KillBackgroundPingTimer() ;

	// Called to manage Route Life Time
	static void ZK_ITIMER_CALLBACK BackgroundPingTimerCallback(
					void *p, 
					unsigned int i1, 
					unsigned int i2, 
					unsigned long time) ;

	size_t				   m_cntPerformanceTask ;
	PerformanceTask *	   m_pPerformanceTask ;
	size_t                 m_maxPingTimeout ;

	// Network probing can take time and occurs in a
	// background thread for that purpose.
	ZkThread			   m_pNetworkProbeThread ;
	WZkMutex			   m_cs ;

	// This data structure contains all the performance
	// information we have collected about individual aips
	EntityPerformanceMap m_EntityMap ;

	// This is the collect data about Route Create Times 
	// It is _not_ currectly persisted, so the value applies
	// to the current session only
	RStatCollector m_routeCreateStats ;				

	// This pair of fails provides a measure of route successfulness
	// These values are persisted so they apply to the entire life
	// of the Freedom.nym file
	size_t         m_cntRouteCreateAttempts ;
	size_t         m_cntRouteCreateSuccess ;

	size_t         m_cntCoreServerConnectCloudAttempts ;
	size_t         m_cntCoreServerConnectCloudSuccess ;

	// These are aips that should be avoided because they might be bad
	// This list is not persisted because network conditions can
	// change rapidly
	ZKSTD::vector<uint32_t> m_AipBlackList ;

	// This timer will go off when it is time to perform a background
	// network probe
	ZkITimer				m_pBackgroundPingTimer ;

	mutable DZkError		m_err ;
} ;

#endif
