// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RNiqsService.cpp: implementation of the RNiqsService class.
//
//////////////////////////////////////////////////////////////////////


#include "StdAfx.h"
#include "RNiqsService.h"

#include <IZKErrorReporter.h>
#include <ZkErrorHandler.h>
#include <IZkFileHelp.h>
#include <IZkRegistry.h>
#include <ZKConfiguration.h>
#include <ZKFileConfiguration.h>
#include <IZkException.h>
#include <IZkConsoleMsg.h>

#include <WNetInfoDatum.h>
#include <ZKConvert.h>
#include <WClientInfo.h>
#include <WZkipClient.h>

#include "ICoreServerConnect.h"
#include "IKeyQry.h"
#include "ITokenDesc.h"
#include "ICrypto.h"
#include "IClientInfo.h"
#include "INetworkPerformance.h"
#include "ZkService.h"

#include <frDBOFuncs.h>			// frDBODriverInitAll

#ifdef ZK_LINUX
#	include <frNetworkDefaults.h>
#	include <netinet/in.h>
#	include <arpa/inet.h>
#endif

#define NIQS_SERVER_CLIENTINFO_CONNECT_TIMEOUT 30
#define NIQS_SERVER_CONNECT_TIMEOUT 20

// Used for debugging (and doom deathmatch) - forces a failed connection when used
#define BOGUS_PORT_VALUE 666

// Mipas allows multiple NIQS to exist, but we don't want to try to
// contact all of them if there are hundreds
#define MAX_NIQS_TO_CONTACT 2



RNiqsService::RNiqsService() 
	: m_bDebugFailConnections( false )
	, m_cs()
{
	// Note, this should be run only once in the course of a freedom execution
	ZKVERIFY( frDBODriverInitAll( &m_Error ) ) ;

	ZKASSERT( m_NetInfo.IsValid() ) ;
	ZKVERIFY(frNetInfoSetClientConsoleCallback(&m_Error, m_NetInfo, RNiqsService::zkClientConsoleCallback, (void*) this) );
}


RNiqsService::~RNiqsService()
{
	ZKASSERT( m_NetInfo.IsValid() ) ;
	ZKVERIFY(frNetInfoSetClientConsoleCallback(&m_Error, m_NetInfo, NULL, NULL ) );

	ZKVERIFY( frDBODriverClearAll( &m_Error ) ) ;	
}

FrNetInfo
RNiqsService::GetNetInfo()
{
	ZKASSERT(m_NetInfo.IsValid());
	return m_NetInfo;
}

#if 0
bool
RNiqsService::DownLoadMinimumKeys()
{
	// The purpose of this function would be to make sure we have enough
	// keys to get 

	// If you look at v1.52 of the Niqs Service you will see a partially
	// writen example that was obsoleted by the availability of core aips
	// and removal of frNetRoute project.

	// The idea here would be to make sure that we have enough AIP keys
	// to always be able to create a route to the Core Aips and hence
	// to be able to get more keys from the key query server.  (This is
	// called when we are still collecting in the clear)

	// Trouble is that for security reasons we don't want to query just the keys
	// of aips we will use for a route in the clear.  Also difficult to calculate
	// the minimum number that we should download....
}
#endif


bool 
RNiqsService::GetEntityById(const uint32_t in_Id, WNetInfoDatum& out_Entity) throw()
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	return m_NetInfo.GetEntityById( in_Id, out_Entity ) ;
}

uint32_t 
RNiqsService::GetEntId(const int32_t IP, const uint16_t Port) throw()
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	return m_NetInfo.GetEntId( IP, Port ) ;
}


// Goal of this function is to provide fast start up times -
// we only contact NIQS server if we absolutely have to.
// It can be contacted later on, but unless the user has a fast
// startup we look bad.

void
RNiqsService::Init( 
	const bool		in_bStartFromScratch, 
	const char *	in_strclient_ver ) throw(RZkException)
{
	//During the update the frNetInfo data is being updated
	//so any calls to functions like GetServersOfType() are not safe
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
		
	time_t time1 = 0 ;
	time_t time2 = 0 ;
	time( &time1 ) ;

	if ( in_bStartFromScratch ) 
	{
		// Backup any file that we have.  Only valid reason to call "Start from Scratch" is to 
		// bypass a corrupt NIQS file that should be sent to support!

		IZkFileHelp::BackupAndErase( ZKFileConfiguration::GetNiqsCacheFilePath() ) ;
	}

	// Attempt to load from cache
	LoadNIQSCache() ;

	// This call will actually contact the NIQS, retrieving Token Desc, Client Info
	// it also downloads an updated view of the network if the locally cached info
	// is out of date
	if ( !InitNetInfo( in_strclient_ver ) )
	{
		IZkConsoleMsg::ShowMsg( IDS_CONS_NONETWORK );
		SendErrorInfoToConsole() ;
	}
	else
	{
		// Time to update our on disk cache
		SaveNIQSCache() ;

		// Tell our performance monitor that we have update network info
		ZkService::GetINetworkPerformance()->UpdateEntityList() ;

		// Launch a lazy background probing of the network to update our view
		// of the status.  Because we specific false it will actually start at
		// some time within the next hour or so
		ZkService::GetINetworkPerformance()->CalculatePerformance( false ) ;
	}

	time( &time2 ) ;
	IZkConsoleMsg::ShowMsg( CONS_DEBUG, CONSIMAGE_WARNING, IDS_CONS_NIQS_TIME, time2 - time1 );


	VerifyEnoughNetworkInfo() ;
}

void
RNiqsService::SendErrorInfoToConsole()
{
	if ( !m_Error.getdetail().empty() )
	{
		// This is ugly technical information for the user but it might explain important detail of why 
		// a key NIQS command has failed
		IZkConsoleMsg::ShowMsg( CONS_VERBOSE, CONSIMAGE_WARNING, IDS_CONS_JUST_STRING, m_Error.getdetail().c_str() );
	}
}

bool
RNiqsService::InitNetInfo( const char *	in_strclient_ver )
{
	if ( IZkRegistry::GetFlag( "SkipNiqsOnStartup" ) )
	{
		IZkConsoleMsg::ShowMsg( 
			CONS_DEBUG, 
			CONSIMAGE_CONNECTION, 
			IDS_CONS_JUST_STRING, 
			"Client is configured to skip contacting NIQS server"
			"- no client info, token descriptions or topology will be downloaded" ) ;

		return false ;
	}

	FrTokenType	pTokenTypeDescriptions = NULL ;

	WClientInfo clientInfo ;

	// Make sure we start off accessing a random server
	m_NetInfo.RandomizeServerList( FRENT_TYPE_NIQS ) ;

	if( 1 != frNetInfoClientInitSequence(
					&m_Error,
					m_NetInfo,
					ZKConfiguration::GetNiqsAddress().c_str(), 
					FR_DFLT_NIQS_SERVER_PORT, 
					NIQS_SERVER_CONNECT_TIMEOUT,
					MAX_NIQS_TO_CONTACT,
					ZkService::GetIKeyQry()->GetKeyCache(),
					in_strclient_ver,
					(FrNiqsClientInfo*)clientInfo,						
					&pTokenTypeDescriptions ) )
	{

		SendErrorInfoToConsole() ;
		return false ;
	}
	else
	{
		// Provide the client info service with an updated state
		ZkService::GetIClientInfo()->SetClientInfo( clientInfo ) ;

		if ( pTokenTypeDescriptions != NULL )
		{
			IZkConsoleMsg::ShowMsg( 
				CONS_DEBUG, 
				CONSIMAGE_CONNECTION, 
				IDS_CONS_JUST_STRING, 
				"Using Token Descriptions downloaded from NIQS" ) ;

			// Provide this information to the Token Description service.
			// REVIEW - once we rewrite the NIQS API so that we can contact over an
			// open socket we can let the ITokenDesc take care of all the messy details
			ZkService::GetITokenDesc()->SetTokenDescriptions( pTokenTypeDescriptions ) ;
			pTokenTypeDescriptions = NULL ;
		}

		return true ;
	}
}


// This function does not throw exceptions
// It takes an arbitrary buffer and attempts to create a frNetInfo object from it
bool
RNiqsService::NiqsFromBuffer(unsigned char * pNiqsBuffer, const unsigned long cntSize) throw()
{
	bool bLoaded = false ;

	try
	{
		ZKASSERT(ZkService::GetIKeyQry()->GetKeyCache());

		// Load from the buffer (Signature is not checked yet)
		if ( m_NetInfo.InitFromBuffer( pNiqsBuffer, cntSize ) )
		{
			if ( ValidateNetInfoSignature() )
			{
				IZkConsoleMsg::ShowMsg( 
					CONS_DEBUG, 
					CONSIMAGE_WARNING, 
					IDS_CONS_JUST_STRING, 
					"NIQS cache file successfully loaded" );
				bLoaded = true;
			}
		}
	}
	catch(...)
	{
		// Data Corruption?
	}

	if ( !bLoaded )
	{
		IZkConsoleMsg::ShowMsg( 
			CONS_DEBUG, 
			CONSIMAGE_WARNING, 
			IDS_CONS_NIQS_LOAD_FAILED, 
			 m_Error.getdetail().c_str() );		

		// Make sure no contents left
		m_NetInfo.Flush() ;
	}

	return bLoaded ;
}







// Commit the NIQS cache to disk
void 
RNiqsService::SaveNIQSCache()
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);

	unsigned char* pNiqsBuffer = NULL;
	size_t		   cntSize      = 0 ;

	try 
	{
		m_NetInfo.GetAsBuffer( (void**)&pNiqsBuffer, &cntSize ) ;

		ZKASSERT( pNiqsBuffer ) ;
		ZKASSERT( cntSize > 0 ) ;

		if ( ( pNiqsBuffer != NULL ) && ( cntSize > 0 ) ) 
		{
			if(!IZkFileHelp::WriteBinaryFile( 
				ZKFileConfiguration::GetNiqsCacheFilePath().c_str(), 
				pNiqsBuffer, 
				cntSize ) )
			{
				throw RZkException(	"RNiqsService::SaveNIQSCache",
									"IZkFileHelp::WriteBinaryFile failed",
									IDS_FILE_WRITE_FAILED,
									ZKFileConfiguration::GetNiqsCacheFilePath().c_str() );	
			}
		}

	}
	catch( IZkException& ) 
	{
		// Not saving this NIQS cache is not fatal
		ZKTRACE0( "Failed to save NIQS cache!\n" ) ;
		SAFEFREE( pNiqsBuffer ) ;
	}
	SAFEFREE( pNiqsBuffer ) ;
}

bool
RNiqsService::LoadNIQSCache()
{
	bool   bLoadedNiqs = true ;

	// A registry setting can be set in order to avoid using cached
	// NIQS information.  If this registry value is set then the
	// complete NIQS information will get downloaded
	if( !IZkRegistry::GetFlag( "NoCacheNIQS" ) ) 
	{
		WAutoBuffer NiqsBuffer ;

		if(! IZkFileHelp::ReadBinaryFile(  ZKFileConfiguration::GetNiqsCacheFilePath().c_str(), 
												&NiqsBuffer, 
												NiqsBuffer.GetSizeRef() ) )
		{

			IZkConsoleMsg::ShowMsg( 
				CONS_VERBOSE, 
				CONSIMAGE_WARNING, 
				IDS_CONS_JUST_STRING, 
				"NIQS cache file is missing or could not be opened" );

			// Cache file is invalid or not available, we'll recover
			// by downloading all information from the server
			bLoadedNiqs = false ;
		}

		if ( bLoadedNiqs ) 
		{
			bLoadedNiqs = NiqsFromBuffer(NiqsBuffer, NiqsBuffer.Size());
		}
	}
	return bLoadedNiqs ;
}

// This function provides feedback from UNIX about what is happening
int 
RNiqsService::zkClientConsoleCallback( 
		  ZK_INOUT ZkError /*err*/ ,
		  ZK_IN    const char * /*prefix*/ ,
		  ZK_IN    ZkClientConsoleMessage  /*msg*/,
		  ZK_IN    ConsSeverity lvl ,
		  ZK_IN    const char * /*action*/ ,
		  ZK_IN    const char * msg_short ,
		  ZK_IN    const char * /*msg_long*/ ,
		  ZK_IN    void * this_pointer )
{
	if (( this_pointer == NULL ) || 
		( msg_short == NULL ) ) 
	{
		ZKASSERT( ZK_FALSE ) ;
		return 0 ;
	}

	if ( lvl != CONS_DEBUG )
	{
		IZkConsoleMsg::ShowMsgForceUpdate( 
			CONS_DEBUG, 
			CONSIMAGE_CONNECTION, 
			IDS_CONS_STARTUP_NIQS_DETAIL, 
			msg_short );
	}
	else
	{
		// Show a lot of debugging information in the console - not present in external build.
		IZkConsoleMsg::ShowMsg( CONS_DEBUG, CONSIMAGE_CONNECTION, IDS_CONS_JUST_STRING, msg_short ) ;
	}
	return 1 ;
}

void
RNiqsService::SetNetworkFailureDebuggingFlag( bool in_bFailConnections )
{

#ifdef INTERNAL_BUILD

	m_bDebugFailConnections = in_bFailConnections ;

#else
	ZKASSERT( false ) ;
	// This code shouldn't be called in external version
#endif
}

void
RNiqsService::DebugNetworkFailureCheck( WNetInfoDatum& io_Server ) const throw(RZkException)
{
#ifdef INTERNAL_BUILD
	if ( m_bDebugFailConnections ) 
	{
		SOCKADDR_IN addr = io_Server.GetSockAddr() ;
		addr.sin_port = htons(BOGUS_PORT_VALUE) ;
		io_Server.SetSockAddr( addr ) ;
	}
#endif
}

void
RNiqsService::GetServer( const FrEntIdType& in_entType, WNetInfoDatum& out_Server ) const throw(RZkException)
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);

	// This takes the first item
	m_NetInfo.GetServer( in_entType, out_Server ) ;
	DebugNetworkFailureCheck( out_Server ) ;
}

void
RNiqsService::GetPopServerAddr(SOCKADDR_IN *in_SA) const throw(RZkException)
{
	WNetInfoDatum PopServer ;
	GetServer( FRENT_TYPE_POP, PopServer );
	memcpy(in_SA, &PopServer.GetSockAddr(), sizeof(SOCKADDR_IN));
}

void
RNiqsService::GetNmtaAddr(SOCKADDR_IN *in_SA) const throw(RZkException)
{
	WNetInfoDatum Nmta ;
	GetServer( FRENT_TYPE_NMTA, Nmta);
	memcpy(in_SA, &Nmta.GetSockAddr(), sizeof(SOCKADDR_IN));
}


void 
RNiqsService::GetServersOfType(
    const FrEntIdType& type, 
    WNetInfoDatumVec& Servers, 
    bool bIncludeDownServers ) const throw(RZkException)
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);	
	m_NetInfo.GetServersOfType( type, Servers, bIncludeDownServers ) ;
}

void 
RNiqsService::GetListOfAips(   
			bool in_bIncludeNonCoreAips,
			bool in_bIncludeCoreAips,
			bool in_bIncludeDownServers,
			WNetInfoDatumVec& out_nlist ) const throw (RZkException) 
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	m_NetInfo.GetListOfAips( in_bIncludeNonCoreAips, in_bIncludeCoreAips, in_bIncludeDownServers, out_nlist ) ;
}

void 
RNiqsService::RoundRobinEntity(const FrEntIdType in_Entity) throw()
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	m_NetInfo.RoundRobinEntity( in_Entity ) ;
}

void 
RNiqsService::UpdateNiqsInformation()
{
	IZkConsoleMsg::ShowMsgForceUpdate( CONSIMAGE_COMP_TERM, IDS_CONS_UPDATING_NETCONFIG );

	time_t timeStart = 0 ;
	time_t timeEnd = 0 ;
	time( &timeStart ) ;

	bool bSucceeded = false ;

	try 
	{
		ContactNiqsServerForUpdate() ;

		// Time to update our on disk cache
		SaveNIQSCache() ;

		// Tell our performance monitor that we have update network info
		// REVIEW - this could happen via Subject observer to be
		// more elegant
		ZkService::GetINetworkPerformance()->UpdateEntityList() ;

		// Launch a background probing of the network to update our view
		// of the AIP status - by specifying false the calculation actually
		// occurs sometime from now - eg within 15 minutes or so.
		ZkService::GetINetworkPerformance()->CalculatePerformance( false ) ;

		bSucceeded = true ;
	}
	catch( IZkException& ex ) 
	{
		IZkConsoleMsg::ShowMsg( 
						CONS_VERBOSE, 
						CONSIMAGE_COMP_TERM, 
						IDS_CONS_NETWORK_UPDATE_FAILED, 
						ZkErrorHandler::GetIZkErrorReporter()->GetErrorString(ex).c_str() 
						) ;
        ZKTRACE1("NIQS error: %s\n", ZkErrorHandler::GetIZkErrorReporter()->GetErrorString(ex).c_str());
	}
	catch( ... )
	{
		// What happened, unexpected?
	}

	// We used to tear down the route whether or not we connect to the NIQS.
	// we have a problem that delayed packets might show up immediately following
	// the route tear-down, and end up creating a another brand new one.
	// This behavior was seen on the NIQS so we have removed it (#8077)


	if ( bSucceeded )
	{
		time( &timeEnd ) ;
		IZkConsoleMsg::ShowMsg( 
						CONS_NORMAL, 
						CONSIMAGE_COMP_TERM, 
						IDS_CONS_DONE_UPDATING_NETCONFIG_WITH_TIME, 
						1 + timeEnd - timeStart );
	}
}

void
RNiqsService::SendNetInfoToConsole()
{
	// Trace out the number of servers of each type
	m_NetInfo.SendInfoAboutEntityTypeToConsole( FRENT_TYPE_TOK_SRV ) ;
	m_NetInfo.SendInfoAboutEntityTypeToConsole( FRENT_TYPE_NYM_SRV ) ;
	m_NetInfo.SendInfoAboutEntityTypeToConsole( FRENT_TYPE_NIQS ) ;
	m_NetInfo.SendInfoAboutEntityTypeToConsole( FRENT_TYPE_KEYQRY_SRV ) ;
	m_NetInfo.SendInfoAboutEntityTypeToConsole( FRENT_TYPE_NMTA ) ;
	m_NetInfo.SendInfoAboutEntityTypeToConsole( FRENT_TYPE_POP ) ;
	m_NetInfo.SendInfoAboutEntityTypeToConsole( FRENT_TYPE_AIP ) ;
}


void
RNiqsService::VerifyEnoughNetworkInfo() throw (RZkException)
{
	SendNetInfoToConsole() ;

	// One case where this can fail is that we have no cache and
	// NIQS connection fails.
	// Another scenario is when the NIQS is not properly configured on
	// a test network.
	// We run this code to identify these scenarios quickly on test network.
	// On the external build we don't run the complete test, but we prevent
	// freedom from starting

	// TODO - refine the exact behavior and error messages that we should
	// show the user at startup if no topology is available.

#ifdef INTERNAL_BUILD
	
	// This will throw
	try
	{
		WNetInfoDatumVec lst ;

		//Aip is currently already being tested with "DownloadKeysIsNecessary"
		//ZkService::GetINiqs()->GetServersOfType( FRENT_TYPE_AIP, lst, true );
		ZkService::GetINiqs()->GetServersOfType( FRENT_TYPE_TOK_SRV, lst, true );
		ZkService::GetINiqs()->GetServersOfType( FRENT_TYPE_NYM_SRV, lst, true );
		ZkService::GetINiqs()->GetServersOfType( FRENT_TYPE_NIQS, lst, true );
		ZkService::GetINiqs()->GetServersOfType( FRENT_TYPE_KEYQRY_SRV, lst, true );

		ZkService::GetINiqs()->GetServersOfType( FRENT_TYPE_NMTA, lst, true );
		ZkService::GetINiqs()->GetServersOfType( FRENT_TYPE_POP, lst, true );
	}
	catch(IZkException& ex )
	{
		// Because this is an internal build we keep going so that
		// we done block anyone, however the network is clearly not fully functional
		// Depending on what fails it may or may not be usable at all
		ZkErrorHandler::GetIZkErrorReporter()->ShowError(ex);
	}
#else
	// For the external version we don't test so much, but we don't swallow the exception either
	// We prevent the client from starting....
	try
	{
		WNetInfoDatumVec lst ;
		ZkService::GetINiqs()->GetServersOfType( FRENT_TYPE_NYM_SRV, lst, true );
	}
	catch(IZkException& ex )
	{
		throw RZkException( m_Error,
							"RNiqsService::VerifyEnoughNetworkInfo", 
							ZkErrorHandler::GetIZkErrorReporter()->GetErrorString(ex).c_str(), 
							IDS_NO_NIQS_INFO ) ;
	}
#endif

}



bool
RNiqsService::ValidateNetInfoSignature()
{
	// Lock to protect contents of WNetInfo
	// REVIEW - this is not the same lock as inside WNetInfo!
	// Is the only way to fix this to get frNetInfoGetSignedBuffer to
	// actually copy the data for us?
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	
	bool retVal = true ;
	
	// Verify the signature 
	// We do it in this order so that we can use the key query server information
	// out of the NetInfo to check the signature.  

	// REVIEW: Clearly this scheme is a bit suspect and also hard to follow.  If we don't
	// trust the buffer how can we use information out of it in order to
	// try and test its validity?

	// REVIEW: this is requires more information about hte content of the
	// WNetInfo then we would ideally like to have

	unsigned char * pSignedBuffer = NULL ;
	size_t cntSignedBuffer = 0 ;

	if ( !frNetInfoGetSignedBuffer(
				&m_Error, 
				m_NetInfo, 
				&pSignedBuffer, 
				&cntSignedBuffer ) )
	{
		return false ;
	}

	std::string		strSigner ;
 	FrEntIdType		entityType ;

	SigCheckResult SigResult = ZkService::GetICrypto()->CheckSignature( 
															pSignedBuffer, 
															cntSignedBuffer, 
															strSigner, 
															entityType ) ;

	if ( SigResult != SigCheckSuccessful )
	{
		// REVIEW - we are pretty strict here, even
		// if we couldn't contact the key query server
		// we are assuming this is a fraud
		// Eg we could do different things depending on the 			
		// SigCheckResult

		IZkConsoleMsg::ShowMsg( 
			CONS_DEBUG, 
			CONSIMAGE_WARNING, 
			IDS_CONS_JUST_STRING, 
			"Signature check failed on Niqs buffer, ignoring its contents" );

		retVal = false ;
	}
	else if ( entityType != FRENT_TYPE_NIQS )
	{
		IZkConsoleMsg::ShowMsg( 
			CONS_DEBUG, 
			CONSIMAGE_WARNING, 
			IDS_CONS_STRING_AND_DETAIL, 
			"Signature on NIQS buffer signed by wrong Entity",
			ZKConvert::GetServerNameFromType( entityType ).c_str() );

		retVal = false ;
	}

	if ( !retVal )
	{
		// Throw away the contents, we will
		// be downloading fresh stuff from NIQS
		m_NetInfo.Flush() ;
	}

	return retVal ;
}

#ifndef USE_NEW_NIQS_LIB

void 
RNiqsService::ContactNiqsServerForUpdate() throw(RZkException)
{
	// We should already have an established NIQS state,
	// this function only makes an incremental update

	// During the update we need to protect our WNetInfo
	// from any other threads that might try to access it.
	// Because we connect through the clear we don't have the
	// trouble of a route creation causing problems.

	ZKS::Lock_Guard<WZkMutex> gate(m_cs);

	// Randomize the list to make sure we access a random
	// NIQS each time we do an update (has no affect if there is
	// only one)
	m_NetInfo.RandomizeServerList( FRENT_TYPE_NIQS ) ;

	int bStatus = frNetInfoClientUpdateSequence(
				&m_Error,
				m_NetInfo,
				ZKConfiguration::GetNiqsAddress().c_str(), 
				FR_DFLT_NIQS_SERVER_PORT, 
				NIQS_SERVER_CONNECT_TIMEOUT,
				MAX_NIQS_TO_CONTACT,
				ZkService::GetIKeyQry()->GetKeyCache() ) ;
	
	if ( !bStatus )
	{
		// Update with the old protocol (works on networks before NI updated to new
		// version, after all networks updated this code can be retired

		throw RZkException(
				m_Error, 
				"RNiqsService::ContactNiqsServerForUpdate()",
				"Periodical update of NIQS failed",
				IDS_NIQS_UPDATE_FAILED_NO_CONNECT);
	}
}

#else

// This is work in progress for taking more control of the NIQS 
// (eg #8363)

static FrEntIdType hostEnt[] = {
  FRENT_TYPE_AIP,
  FRENT_TYPE_NYM_SRV,
  FRENT_TYPE_SER_SRV,           
  FRENT_TYPE_TOK_SRV,
  FRENT_TYPE_KEYQRY_SRV,
  FRENT_TYPE_KEYUPD_SRV,
  FRENT_TYPE_NNTP_SRV,
  FRENT_TYPE_NIQS,
  FRENT_TYPE_NISS,
  FRENT_TYPE_POP,
  FRENT_TYPE_NMTA,
  FRENT_TYPE_IMEP,
  FRENT_TYPE_COREAIP
};

#define ASZ(a) (sizeof (a) / sizeof (a)[0])

void 
RNiqsService::ContactNiqsServerForUpdate() throw(RZkException)
{
	// We should already have an established NIQS state,
	// this function only makes an incremental update

	// During the update we need to protect our WNetInfo
	// from any other threads that might try to access it.
	// Because we connect through the clear we don't have the
	// trouble of a route creation causing problems.

	WZkipClient zkipClient ;

	ZkService::GetICoreServerConnect()->ConnectToNiqs( 
											zkipClient, 
											NIQS_SERVER_CONNECT_TIMEOUT, 
											MAX_NIQS_TO_CONTACT ) ;


	// Check if the remote NetInfo has changed

	WAutoBuffer mxqBuf ;

	if (!frNiqsMxQuery(
			&m_Error, 
			zkipClient, 
			FRNIQ_ATTR_DESCRIPTION, 
			0,
            ZKIP_ENCODING_BZIP2, 
			hostEnt, 
			ASZ(hostEnt),
			(void**)(unsigned char**)&mxqBuf,
			mxqBuf.GetSizePointer() ) )
	{
		throw RZkException(
				m_Error, 
				"RNiqsService::ContactNiqsServerForUpdate()",
				"Periodical update of NIQS failed",
				IDS_CONS_NIQS_MXQUERY_FAIL);
	}

	frNiqsDisconnect(&m_Error, &zkipClient) ;

	// Now we need to verify the signature

	ZKS::Lock_Guard<WZkMutex> gate(m_cs);

	// Now set the NetInfo -> this doesn't work because
	// the mxqBuf doesn't have the header that 
	// frNetInfoLoadNoSig expects
	// m_NetInfo.InitFromBuffer( mxqBuf, mxqBuf.Size() ) ;	
}


#endif

