// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RNiqsService.h: interface for the RNiqsService class.
//
//////////////////////////////////////////////////////////////////////

#ifndef _NIQSSERVICE_H
#define _NIQSSERVICE_H

#include "INiqs.h"
#include <DZkError.h>
#include <WNetInfo.h>
#include <WSyncTypes.h>


class RNiqsService : public INiqs
{

public:
	RNiqsService() ;
	
	virtual				~RNiqsService();

	virtual void        Init(
							const bool		in_bStartFromScratch, 
							const char *	in_strclient_ver ) throw(RZkException) ;

	// PURPOSE: This is used to do the initial Niqs update at application launch, then subsequently
	//			every x minutes afterwards (where x is in the clientinfo)
	// PROMISE: This will first attempt to use all the niqs' from the cache (if 'in_bUseCache' is true)
	//			and only if they all fail to connect will it try the hard coded host as returned from
	//			ZKConfiguration::GetNiqsAddress()
	virtual void		UpdateNiqsInformation() ;

	virtual FrNetInfo	GetNetInfo();

	virtual uint32_t	GetEntId(const int32_t, const uint16_t) throw();
	
	virtual void		GetServer( const FrEntIdType& in_entType, WNetInfoDatum& out_Server ) const throw(RZkException);

	virtual void        SetNetworkFailureDebuggingFlag( bool in_bFailConnections ) ;
	
	// PURPOSE: Searches the object's WNetInfo for a specific entity Id	
	// PROMISE:	Returns true if it exists, false otherwise
	// REQUIRES:One of the populate functions should probably be called prior,
	//			but won't die otherwise
	// THROWS:	Nothing
	virtual bool GetEntityById(const uint32_t in_Id, WNetInfoDatum& out_Entity) throw();

	// Call back for Console updates
	static int zkClientConsoleCallback( 
		  ZK_INOUT ZkError err ,
		  ZK_IN    const char * prefix ,
		  ZK_IN    ZkClientConsoleMessage  msg,
		  ZK_IN    ConsSeverity lvl ,
		  ZK_IN    const char * action ,
		  ZK_IN    const char * msg_short ,
		  ZK_IN    const char * msg_long ,
		  ZK_IN    void * extra );

	// PURPOSE: Shuffles the order of available servers of a certain entity type
	//			in a round robin fashion. This is to allow us to use multiple
	//			servers to do load balancing.
	// PROMISE:	To call frNetInfoRoundRobinList
	// THROWS:  Nothing
	void RoundRobinEntity(const FrEntIdType in_Entity) throw();

	// PURPOSE:		Gets all servers of type type	
	void GetServersOfType(	const FrEntIdType& type,
							WNetInfoDatumVec& nlist, 
							bool bIncludeDownServers)
							const throw(RZkException);
	void GetListOfAips(   
			bool in_bIncludeNonCoreAips,
			bool in_bIncludeCoreAips,
			bool in_bIncludeDownServers,
			WNetInfoDatumVec& out_nlist ) const throw(RZkException) ;

	virtual void		GetNmtaAddr(SOCKADDR_IN *in_SA)
							const throw(RZkException);
	virtual void		GetPopServerAddr(SOCKADDR_IN *in_SA)
							const throw(RZkException);

private:

	// PROMISE: This will use the hard coded host name for the niqs as returned from ZKConfiguration::GetNiqsAddress()
	void PopulateNiqsFromScratch() throw(RZkException) ;

	bool LoadNIQSCache() ;

	// PURPOSE: We save the state of the network topology when we
	//			download information from the Niqs.
	// PROMISE: To call InitializeNetInfo() if the buffer was successfully saved.
	void SaveNIQSCache() ;

	bool NiqsFromBuffer(unsigned char* in_pBuffer, const unsigned long in_cntBuffer ) throw();

	bool ValidateNetInfoSignature() ;

	RNiqsService& operator=(const RNiqsService&);

	void ContactNiqsServerForUpdate() throw(RZkException) ;

	void SendErrorInfoToConsole() ;

	bool InitNetInfo( const char *	in_strclient_ver ) ;

	// When Network Failure Debugging is turned on this function will munge network address
	// to force connection failures.
	void DebugNetworkFailureCheck( WNetInfoDatum& io_Server ) const throw(RZkException);

	// PURPOSE: Verify that we have enough NIQS information to successfully use the Freedom network
	void VerifyEnoughNetworkInfo() throw (RZkException) ;

	void SendNetInfoToConsole() ;

	// ctor member init
	bool				m_bDebugFailConnections;

	mutable DZkError	m_Error;
	mutable WZkMutex	m_cs;			// Protect access to frNetInfo
	WNetInfo			m_NetInfo;		// All network topology is here
};

#endif 
