// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RNym.cpp: implementation of the RNym class.
//
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "RNym.h"
#include "RNymMail.h"
#include "RNymWebCacheMgr.h"
#include "RNymWallet.h"
#include "RNymCookieMgr.h"
#include "RNymRoute.h"
#include "IRouteManager.h"
#include "ICoreServerConnect.h"
#include <frCrypt.h>
#include <frNymUtil.h>
#include <zkDateTime.h>
#include <ZKConvert.h>
#include <WCoreTypes.h>
#include <WAutoBuffer.h>
#include <ZKChunk.h>
#include <ICrypto.h>
#include <INiqs.h>			// for NYM_SERVER_TIMEOUT
#include <ITokenService.h>
#include <ZkService.h>
#include <IZkConsoleMsg.h>
#include <ZKIP.h>
#include <RZkException.h>
#include <WZkipClient.h>

#define BUFFER_HEADER_SIZE 3

// Adjust this value if we need to change the structure of persistence at the
// nym level
#define NYM_PERSIST_VERSION 1

RNym::RNym() 
	: m_pNymMail(NULL)
	, m_key1(NULL)
	, m_key2(NULL)
	, m_LifeStage( NYM_LIFE_STAGE_UNDEFINED )
	, m_nymCapability( FRNYM_CAPABILITY_INIT )
{	
	memset(&m_NymServerResponse, 0, sizeof(NymServerResponse));
	m_pNymMail = new RNymMail( this, m_RemoteNym ) ;
		
	m_pNymWallet  = new RNymWallet(this);

	m_pNymCookieMgr = new RNymCookieMgr(this);

	m_pNymRoute = new RNymRoute(this) ;

	m_pNymWebCacheMgr = new RNymWebCacheMgr;
}

RNym::~RNym()
{		
	SAFEDELETE( m_pNymMail ) ;
	SAFEDELETE( m_pNymWallet );
	SAFEDELETE( m_pNymCookieMgr );
	SAFEDELETE( m_pNymRoute ) ;
	SAFEDELETE( m_pNymWebCacheMgr ) ;

	if(m_key2 != NULL){
		ZKVERIFY(frCryptPubKeyFree(&m_ZkError, &m_key2));
	}
	if(m_key1 != NULL){
		ZKVERIFY(frCryptPubKeyFree(&m_ZkError, &m_key1));
	}
}

void
RNym::QueryInterface(uint32_t in_iid, void** ppv) throw(RZkException)
{
	if (in_iid == IID_INYM)
	{
		*ppv = static_cast<INym*>(this) ; 
	}
	else if (in_iid == IID_IMAIL)
	{
		ZKASSERT( m_pNymMail != NULL ) ;

		// Aggregation
		// However we are breaking COM rules by not having this
		// INymMail also derive from QueryInterface.
		*ppv = static_cast<INymMail*>(m_pNymMail) ; 
	}
	else if (in_iid == IID_INYMWALLET)
	{
		ZKASSERT( m_pNymWallet != NULL ) ;
		*ppv = static_cast<INymWallet*>(m_pNymWallet) ; 
	}
	else if (in_iid == IID_INYMCOOKIEMGR)
	{
		ZKASSERT( m_pNymCookieMgr != NULL ) ;
		*ppv = static_cast<INymCookieMgr*>(m_pNymCookieMgr) ; 
	}
	else if (in_iid == IID_INYMROUTE)
	{
		ZKASSERT( m_pNymRoute != NULL ) ;
		*ppv = static_cast<INymRoute*>(m_pNymRoute) ; 
	}
	else if (in_iid == IID_INYMWEBCACHEMGR)
	{
		ZKASSERT( m_pNymWebCacheMgr != NULL ) ;
		*ppv = static_cast<INymWebCacheMgr*>(m_pNymWebCacheMgr) ;
	}
	else
	{
		*ppv = NULL ;
		throw RZkException( "RNym::QueryInterface","Unexpected Failure",IDS_UNEXPECTED_FAIL ) ;
	}
}

bool
RNym::IsNymPreferenceSet( unsigned long in_PrefToTest )
{
	unsigned long dwFlags;
	GetNymPreferencesBitfield(dwFlags);
	return (dwFlags & in_PrefToTest);
}

bool 
RNym::GetCookiesEnabled()
{
	unsigned long dwFlags;
	GetNymPreferencesBitfield(dwFlags);
	return !(dwFlags & PREF_COOKIES_DISABLED);
}

void 
RNym::SetCookiesEnabled(bool bEnabled)
{
	SetNymPreferencesBitfield(bEnabled ? 0 : PREF_COOKIES_DISABLED,	PREF_COOKIES_DISABLED);
}

bool 
RNym::GetHideRefererField()
{
	unsigned long dwFlags;
	GetNymPreferencesBitfield(dwFlags);
	return !(dwFlags & PREF_NO_HIDE_REFERER);
}

void 
RNym::SetHideRefererField(bool bHide)
{
	SetNymPreferencesBitfield(bHide ? 0 : PREF_NO_HIDE_REFERER,
		PREF_NO_HIDE_REFERER);
}

void 
RNym::CreateYourself(ZKNymName& name, const std::string& token)
{
	ZkService::GetIRouteManager()->TerminateMailRoutes();
	// Validate parameters
	ZKASSERT( !token.empty() ) ;
	ZKASSERT( !name.GetRequestedName().empty() ) ;

	CreateNymOnNymServer( name, token);
	SetYourLifeStage( NYM_LIFE_STAGE_ACTIVE ) ;

	if(DoesNymUseMail())
	{
		ZKSTD::string strNymName;
		GetNymName(strNymName);
		IZkConsoleMsg::ShowImportantMsg(CONSIMAGE_MAIL, IDS_CONS_CREATING_NYM_MAIL_ACCOUNT,  strNymName.c_str());
		try
		{
			m_pNymMail->RetrieveNymMailCert();
			m_pNymMail->CreatePopAccountOnPopServer();
			m_pNymMail->SetSpamPreference(true);
		}
		catch(RZkException ex)
		{
			IZkConsoleMsg::ShowImportantMsg(CONSIMAGE_MAIL, IDS_CONS_JUST_STRING, ex.GetFormattedString().c_str());
		}
	}
}

void 
RNym::GetYourLifeStage(unsigned long& out_LifeStage) throw(RZkException)
{
#ifdef _DEBUG
	static const char* buf[] = {
	"NYM_LIFE_STAGE_UNDEFINED",
	"NYM_LIFE_STAGE_ACTIVE",
	"NYM_LIFE_STAGE_EXPIRED", 
	"NYM_LIFE_STAGE_RELEASED" };

	std::string strName;
	GetNymName(strName);
	// Sometimes, the OutputDebugString() call the following line ends up calling
	// from _CrtDbgReport() in dbgrpt.c, seems to freeze...may be related to Q173260 
	// since this happens under NT only
	// ZKTRACE2("RNym::GetYourLifeStage [%s] life stage returning [%s]\n", strName.c_str(), buf[m_LifeStage]);
#endif


	if ( m_LifeStage != NYM_LIFE_STAGE_UNDEFINED ) {
		out_LifeStage = m_LifeStage ;
		return ;
	}
	
	// All life stage values should be determined at start up
	// based on the NIQS time

	ZKASSERT( false ) ;

	out_LifeStage = NYM_LIFE_STAGE_UNDEFINED ;

	return ;
}

// Warning: use this function with extreme care - 
// changing life stage at this level will not
// change the internal facade data, or update
// the state of the server.
void 
RNym::SetYourLifeStage( unsigned long in_LifeStage )
{
	ZKASSERT( in_LifeStage != NYM_LIFE_STAGE_UNDEFINED ) ;
	m_LifeStage = in_LifeStage ;

#ifdef _DEBUG
	const char* buf[] = {
	"NYM_LIFE_STAGE_UNDEFINED",
	"NYM_LIFE_STAGE_ACTIVE",
	"NYM_LIFE_STAGE_EXPIRED", 
	"NYM_LIFE_STAGE_RELEASED" };

	std::string strName;
	GetNymName(strName);
	ZKTRACE2("RNym::SetYourLifeStage() [%s] setting life stage to [%s]\n", strName.c_str(), buf[in_LifeStage]);
#endif
}


void 
RNym::GetNymPreferencesBitfield(unsigned long& out_dwFlags)
{
	const unsigned char *data = NULL;
	size_t length = 0;

	out_dwFlags = 0;
	frRemoteNymResetParameterSearch(&m_ZkError, m_RemoteNym.GetFrRemoteNym());
	frRemoteNymGetParameter(&m_ZkError,
		m_RemoteNym.GetFrRemoteNym(),
		NymPreferences,
		&data, &length);

	if(length > 0)
	{
		out_dwFlags = *((unsigned long *)data);
	}

}

void RNym::SetNymPreferencesBitfield(unsigned long in_dwFlags, unsigned long in_dwMask)
{
	unsigned long dwPrefs;

	GetNymPreferencesBitfield(dwPrefs);

	dwPrefs &= ~in_dwMask;
	dwPrefs |= (in_dwMask & in_dwFlags);

    frRemoteNymDeleteParameter(&m_ZkError,
		m_RemoteNym.GetFrRemoteNym(),
		NymPreferences);
    frRemoteNymAddParameter(&m_ZkError,
		m_RemoteNym.GetFrRemoteNym(),
		NymPreferences,(const unsigned char *)&dwPrefs,sizeof(unsigned long));
}

void 
RNym::SetNymCryptoAndSigningKeys() throw(RZkException)
	// keys are stored in the nym's FrRemoteNym
{

	time_t currentTime ;

	time( &currentTime ) ;



	time_t timeExpiration = currentTime + DAYS_UNTIL_FREEDOM_KEYS_EXPIRE * 24 * 60 * 60 ;
	const ZkClock expiration = zkClockFromTime( timeExpiration );

	// entity
	fr_entid_exp frEntity;
	memset( &frEntity, 0, sizeof(fr_entid_exp) );
	const unsigned char* strNymName = NULL;
	size_t szstrNymName = 0;

	if(!frRemoteNymResetParameterSearch (
		&m_ZkError ,                 
		m_RemoteNym.GetFrRemoteNym() ))
	{
		throw RZkException("RNym::SetNymCryptoAndSigningKeys()", 
							"frRemoteNymResetParameterSearch() failed", 
							IDS_FRNYMPARAM);
	}


	if(!(frRemoteNymGetParameter ( 
		&m_ZkError , 
		m_RemoteNym.GetFrRemoteNym() , 
		NymName , 
		&strNymName , 
		&szstrNymName )))
	{
		throw RZkException("RNym::SetNymCryptoAndSigningKeys()", 
							"frRemoteNymGetParameter() failed", 
							IDS_FRNYMPARAM);
	}
	
	frEntity.type = FRENT_TYPE_NYM;
	memcpy(frEntity.name, strNymName, szstrNymName); 	


	ZkService::GetICrypto()->NewCryptKey(1, 1024, expiration, m_key1, &frEntity);

	ZkService::GetICrypto()->NewSigKey(1, 1024, expiration, m_key2, &frEntity);


	if(!frRemoteNymSetPrivateKeys (
			&m_ZkError ,
			m_RemoteNym.GetFrRemoteNym() ,
			m_key1 , /* Encryption key */
			m_key2 ))  /* Signature key */
	{
		throw RZkException("RNym::SetNymCryptoAndSigningKeys()", 
							"frRemoteNymSetPrivateKeys() failed", 
							IDS_FRNYMPARAM);
	}

	if(!frRemoteNymResetParameterSearch (
		&m_ZkError ,                 
		m_RemoteNym.GetFrRemoteNym() ))
	{
		throw RZkException("RNym::SetNymCryptoAndSigningKeys()", 
							"frRemoteNymSetPrivateKeys() failed", 
							IDS_FRNYMPARAM);
	}

	// Push keys on
	if(!frRemoteNymSetPubKeys (
		&m_ZkError, 
		m_RemoteNym.GetFrRemoteNym(),
		m_key1,
		m_key2))
	{
		throw RZkException("RNym::SetNymCryptoAndSigningKeys()", 
							"frRemoteNymSetPrivateKeys() failed", 
							IDS_FRNYMPARAM);
	}
}


void 
RNym::CreateNymOnNymServer(
	ZKNymName& NameOfNym, 
	const std::string& token)	throw(RZkException)
{
	ZKFunctionEntry( "RNym::CreateNymOnNymServer" ) ;	
	
	int bStatus = true;

	const unsigned char* strNymName = NULL;
	size_t szstrNymName = 0;

	FrRemoteNym ClientNym = m_RemoteNym.GetFrRemoteNym();	// Our Client Side representation of the Nym
	FrRemoteNym StrippedCopyClientSideNym = NULL;			// Client-side Nym with private information removed
	FrRemoteNym ServerSideNym = NULL;						// Server Side represenation of the Nym

	try 
	{
		// Name
		bStatus = frRemoteNymAddParameter(         
					&m_ZkError, 
					ClientNym, 
					NymName, 
					(const unsigned char*)NameOfNym.GetRequestedName().c_str(), 
					NameOfNym.GetRequestedName().length()) ;

		ZKTestAndThrow(bStatus);

		// Check Name
		bStatus = frRemoteNymVerifyName(&m_ZkError, ClientNym) ;

		ZKTestAndThrowMsg(bStatus, "Bad Nym Name", IDS_BAD_NYM_NAME)
			
		SetNymCryptoAndSigningKeys();

		// Dup
		bStatus = m_RemoteNym.DupFrRemoteNym(
					m_RemoteNym.GetFrRemoteNym(), 
					&StrippedCopyClientSideNym);

		ZKTestAndThrow(bStatus);

		// Strip
		bStatus = frRemoteNymStrip (
					&m_ZkError ,
					StrippedCopyClientSideNym ,
					FL_CLIENT_ONLY );

		ZKTestAndThrow(bStatus);
			
		bStatus = frRemoteNymResetParameterSearch (
						&m_ZkError ,                 
						StrippedCopyClientSideNym );
		ZKTestAndThrow(bStatus);

		bStatus = frRemoteNymGetParameter ( 
						&m_ZkError , 
						StrippedCopyClientSideNym , 
						NymName , 
						&strNymName , 
						&szstrNymName );
		ZKTestAndThrow(bStatus);			

		// Create

		WZkipClient client;
		// SmartConnectThroughCloud() throws if the connect is unsuccessful
		ZkService::GetICoreServerConnect()->SmartConnectThroughCloud(
						FRENT_TYPE_NYM_SRV, client, NYM_SERVER_TIMEOUT);
		
		bStatus = frRemoteNymCreateConnected(
						&m_ZkError, 
						StrippedCopyClientSideNym, 
						(char*)strNymName,					// well, Philippe's calls should use const,
						const_cast<char*>(token.c_str()),	// I'll fix that later.		
						client,
						NULL,
						&m_NymServerResponse,
						&ServerSideNym) ;

		client.DisconnectAndReset();
		
		if (!bStatus && ( m_NymServerResponse.message[0] != 0 ) &&	   
			( m_NymServerResponse.code == _COMMAND_FAILED_NYM_EXISTS )) 
		{
			
			// Nym Creation failed since the name was unavailable.
			// Get a list of alternative names and see if they're
			// available. Store them where the UI can find them,
			// then throw back up to the UI.

			std::vector <std::string> v_attempt;
			
			char * avail_name_str = NULL;
			size_t i;

			NameOfNym.Status() = ZKNNStatusNameUnavailable;
			
			while(NameOfNym.GetAttemptVector(v_attempt).size())
			{
				WZkList attempt_list;
				WZkListNoAlloc avail_list;
				for(i = 0; i < v_attempt.size(); ++i)
				{
					zkListInsert(&m_ZkError, attempt_list, (void *)v_attempt[i].c_str(), LIST_HEAD);
				}
				
				// SmartConnectThroughCloud() throws if the connect is unsuccessful
				ZkService::GetICoreServerConnect()->SmartConnectThroughCloud(
									FRENT_TYPE_NYM_SRV, client, NYM_SERVER_TIMEOUT);

				if(!frRemoteNymAvailableNamesConnected(&m_ZkError, attempt_list, client, 
									NULL, &m_NymServerResponse, &avail_list))
				{
					throw RZkException( "RNym::CreateNymOnNymServer",
									 "NymName Already Exists, frRemoteNymAvailableNamesConnected failed.", 
									 IDS_ALTNAMEFAILED);									 
				}
				else
				{
					v_attempt.clear();

					zkListMoveCursor(&m_ZkError,avail_list,LIST_HEAD);
					while (zkListGet(&m_ZkError,avail_list,(void **)&avail_name_str,LIST_FORWARDS)) 
					{
						v_attempt.push_back((std::string)avail_name_str);
					}

					NameOfNym.ValidateNames(v_attempt);
				}
			}
			// Throw an exception back up to the UI. 
			throw RZkException( "RNym::CreateNymOnNymServer",
								 "NymName Already Exists", 
								 IDS_NYM_NAME_ALREADY_EXISTS);
		}
		else if (!bStatus && ( m_NymServerResponse.code == _COMMAND_FAILED_TOKEN_SPENT )) 
        {
            throw RZkException( "RNym::CreateNymOnNymServer",
                                "Token Already Spent",
                                IDS_TOKEN_ALREADY_SPENT );
        }


		TestAndThrowServerResponse(bStatus) ;

		NameOfNym.Status() = ZKNNStatusNameAvailable;

		// Add the extra fields that we find in the Server
		// copy of the Nym back to the 
		bStatus = m_RemoteNym.CopyReqdFrRemoteNymParamsToLocalFrRemoteNym(
						ClientNym,
						ServerSideNym);
		ZKTestAndThrow(bStatus);

	}
	catch( IZkException& )
	{
		if(ServerSideNym != NULL){
			ZKVERIFY(frRemoteNymFree(&m_ZkError, &ServerSideNym));
		}
		
		if(StrippedCopyClientSideNym != NULL){
			ZKVERIFY(frRemoteNymFree(&m_ZkError, &StrippedCopyClientSideNym));
		}
		throw ;
	}

	if(ServerSideNym != NULL)
	{
		ZKVERIFY(frRemoteNymFree(&m_ZkError, &ServerSideNym));
	}

	if(StrippedCopyClientSideNym != NULL)
	{
		ZKVERIFY(frRemoteNymFree(&m_ZkError, &StrippedCopyClientSideNym));
	}
}


void 
RNym::DisableNymOnNymServer() throw(RZkException)
	// Client responsible for disabling in the App etc....
{
	ZKFunctionEntry( "RNym::DisableNymOnNymServer" ) ;	

	int bStatus = true;

	const unsigned char* strNymName = NULL;
	size_t szstrNymName = 0;

	FrRemoteNym StrippedCopyClientSideNym = NULL;

	try 
	{

		// Duplicate the Nym
		bStatus = m_RemoteNym.DupFrRemoteNymWithSignatureKey(
						m_RemoteNym.GetFrRemoteNym(), 
						&StrippedCopyClientSideNym);

		ZKTestAndThrow(bStatus);

		// Strip out private Client information
		bStatus = frRemoteNymStrip (
						&m_ZkError ,
						StrippedCopyClientSideNym ,
						FL_CLIENT_ONLY );
		ZKTestAndThrow(bStatus);


		bStatus = frRemoteNymGetParameter ( 
						&m_ZkError , 
						StrippedCopyClientSideNym , 
						NymName , 
						&strNymName , 
						&szstrNymName ) ;
		ZKTestAndThrow(bStatus);

		WZkipClient client;
		// SmartConnectThroughCloud() throws if the connect is unsuccessful
		ZkService::GetICoreServerConnect()->SmartConnectThroughCloud(
						FRENT_TYPE_NYM_SRV, client, NYM_SERVER_TIMEOUT);

		bStatus = frRemoteNymDisableConnected(
						&m_ZkError, 
						StrippedCopyClientSideNym, 
						(char *)strNymName , 
						client,
						NULL,
						&m_NymServerResponse ) ;

		if (!bStatus && ( m_NymServerResponse.message[0] != 0 ) &&	   
			( m_NymServerResponse.code == _COMMAND_FAILED_NYM_DISABLED )) 
		{
			
			// Couldn't disable because the nym was already disabled.
		}
		else
		{
			TestAndThrowServerResponse(bStatus) ;
		}
	}
	catch ( std::exception& )
	{
		if( StrippedCopyClientSideNym )
			ZKVERIFY(frRemoteNymFree(&m_ZkError, &StrippedCopyClientSideNym));

		throw ;
	}

	if( StrippedCopyClientSideNym )
		ZKVERIFY(frRemoteNymFree(&m_ZkError, &StrippedCopyClientSideNym)) ;
}


void 
RNym::ModifyExistingNymOnNymServer() throw(RZkException)
{
	ZKFunctionEntry( "RNym::ModifyExistingNymOnNymServer" ) ;	
	
	int bStatus = true;
	const unsigned char* strNymName = NULL;
	size_t szstrNymName = 0;
	FrRemoteNym ServerSideNym = NULL;
	FrRemoteNym StrippedCopyClientSideNym = NULL;

	try 
	{
		// Dup
		bStatus = m_RemoteNym.DupFrRemoteNymWithSignatureKey(
					m_RemoteNym.GetFrRemoteNym(), 
					&StrippedCopyClientSideNym);
		ZKTestAndThrow(bStatus);

		// Strip
		bStatus = frRemoteNymStrip (
					&m_ZkError ,
					StrippedCopyClientSideNym ,
					FL_CLIENT_ONLY );
		ZKTestAndThrow(bStatus);

		bStatus = frRemoteNymResetParameterSearch (
					&m_ZkError ,                 
					m_RemoteNym.GetFrRemoteNym() );
		ZKTestAndThrow(bStatus);

		// Pull name out
		bStatus = frRemoteNymGetParameter ( 
					&m_ZkError , 
					StrippedCopyClientSideNym , 
					NymName , 
					&strNymName , 
					&szstrNymName );
		ZKTestAndThrow(bStatus);

		WZkipClient client;
		// SmartConnectThroughCloud() throws if the connect is unsuccessful
		ZkService::GetICoreServerConnect()->SmartConnectThroughCloud(
							FRENT_TYPE_NYM_SRV, client, NYM_SERVER_TIMEOUT);

		// Update the Server with the Nym Name
		bStatus = frRemoteNymModifyConnected(
					&m_ZkError, 
					StrippedCopyClientSideNym, 
					(char *)strNymName, 
					client,
					NULL,
					&m_NymServerResponse,
					&ServerSideNym);

		TestAndThrowServerResponse(bStatus) ;

		// Fill in our Nym with any new parameters from
		// the server

		bStatus = m_RemoteNym.CopyReqdFrRemoteNymParamsToLocalFrRemoteNym(
						m_RemoteNym.GetFrRemoteNym(),
						ServerSideNym) ;
		ZKTestAndThrow(bStatus);
	}
	catch ( std::exception& )
	{
		// Free mem
		if(ServerSideNym != NULL)
		{
			ZKVERIFY(frRemoteNymFree(&m_ZkError, &ServerSideNym));
		}
		if(StrippedCopyClientSideNym != NULL)
		{
			ZKVERIFY(frRemoteNymFree(&m_ZkError, &StrippedCopyClientSideNym));
		}
		
		throw ;
	}


	// Free mem
	if(ServerSideNym != NULL)
	{
		ZKVERIFY(frRemoteNymFree(&m_ZkError, &ServerSideNym));
	}
	if(StrippedCopyClientSideNym != NULL)
	{
		ZKVERIFY(frRemoteNymFree(&m_ZkError, &StrippedCopyClientSideNym));
	}
}

void 
RNym::AddTokenToNymOnNymServer( const std::string& token ) throw(RZkException)
{
	ZKFunctionEntry( "RNym::AddTokenToNymOnNymServer" ) ;	
	
	int bStatus = true;
	
	const unsigned char* strNymName = NULL;
	size_t szstrNymName = 0;
	FrRemoteNym ServerSideNym = NULL;
	FrRemoteNym StrippedCopyClientSideNym = NULL;

	try 
	{	
		// Dup
		bStatus = m_RemoteNym.DupFrRemoteNymWithSignatureKey(
					m_RemoteNym.GetFrRemoteNym(), 
					&StrippedCopyClientSideNym);
		ZKTestAndThrow(bStatus);

		// Strip
		bStatus = frRemoteNymStrip (
					&m_ZkError ,
					StrippedCopyClientSideNym ,
					FL_CLIENT_ONLY );
		ZKTestAndThrow(bStatus);
	
		bStatus = frRemoteNymResetParameterSearch (
					&m_ZkError ,                 
					m_RemoteNym.GetFrRemoteNym());
		ZKTestAndThrow(bStatus);

		// Pull name out
		bStatus = frRemoteNymGetParameter ( 
					&m_ZkError , 
					StrippedCopyClientSideNym , 
					NymName , 
					&strNymName , 
					&szstrNymName );
		ZKTestAndThrow(bStatus);

		WZkipClient client;
		// SmartConnectThroughCloud() throws if the connect is unsuccessful
		ZkService::GetICoreServerConnect()->SmartConnectThroughCloud(
						FRENT_TYPE_NYM_SRV, client, NYM_SERVER_TIMEOUT);

		// Add Token
		bStatus = frRemoteNymAddTokenConnected (
					&m_ZkError ,                 
					StrippedCopyClientSideNym ,
		  			const_cast<char*>(token.c_str()) ,
					client,
					NULL,
					&m_NymServerResponse,     
					&ServerSideNym );

        // Eric B. - not proud of this one, but we need to have the same response as
        //           RNym::CreateNymOnNymServer when the token has already been spent.
        // FIXME - exception handling in this file needs quite a big overhaul.
        //         use exception classes to identify errors instead of strings.
		if (!bStatus && ( m_NymServerResponse.code == _COMMAND_FAILED_TOKEN_SPENT )) 
        {
            throw RZkException( "RNym::AddTokenToNymOnNymServer",
                                "Token Already Spent",
                                IDS_TOKEN_ALREADY_SPENT );
        }
        else
        {
            TestAndThrowServerResponse(bStatus) ;
        }

		// Add what we got from the server to our Nym
		bStatus = m_RemoteNym.CopyReqdFrRemoteNymParamsToLocalFrRemoteNym(
					m_RemoteNym.GetFrRemoteNym(),
					ServerSideNym);
		ZKTestAndThrow(bStatus);
	}
	catch ( std::exception& )
	{
		// Free mem
		if(ServerSideNym != NULL)
		{
			ZKVERIFY(frRemoteNymFree(&m_ZkError, &ServerSideNym));
		}
		if(StrippedCopyClientSideNym != NULL)
		{
			ZKVERIFY(frRemoteNymFree(&m_ZkError, &StrippedCopyClientSideNym));
		}

		throw;
	}

	// Free mem
	if(ServerSideNym != NULL)
	{
		ZKVERIFY(frRemoteNymFree(&m_ZkError, &ServerSideNym));
	}
	if(StrippedCopyClientSideNym != NULL)
	{
		ZKVERIFY(frRemoteNymFree(&m_ZkError, &StrippedCopyClientSideNym));
	}
}

bool 
RNym::GetNymEmailDomain( std::string& strEmail )
{
	bool ReturnValue = true;

	ZKVERIFY(frRemoteNymResetParameterSearch (
		&m_ZkError ,                 
		m_RemoteNym.GetFrRemoteNym() ));

	const unsigned char* strEmailDomain = NULL;
	size_t szstrEmailDomain = 0;

	if(!frRemoteNymGetParameter ( 
		&m_ZkError , 
		m_RemoteNym.GetFrRemoteNym() , 
		NymEmailDomain , 
		&strEmailDomain , 
		&szstrEmailDomain ))
	{
		ZKASSERT(false);
		ReturnValue = false;
	}
	else 
	{
		strEmail = (char*)strEmailDomain;

		// The nym should always return an email domain
		ZKASSERT( strEmail.size() > 0 ) ;
	}

	return ReturnValue;
}

void 
RNym::SetUserName( const std::string& NameOfNym )
{
	ZKFunctionEntry( "RNym::SetUserName" ) ;
	
	int bStatus ;

	// Name
	bStatus = frRemoteNymAddParameter(         
				&m_ZkError, 
				m_RemoteNym.GetFrRemoteNym(), 
				NymName, 
				(const unsigned char*)NameOfNym.c_str(), 
				NameOfNym.length()) ;

	ZKTestAndThrow(bStatus);

	// Check Name
	bStatus = frRemoteNymVerifyName(&m_ZkError, m_RemoteNym.GetFrRemoteNym()) ;

	ZKTestAndThrowMsg(bStatus, "Bad Nym Name", IDS_BAD_NYM_NAME);
}

void	
RNym::GetNymIconInfo( WNymIconInfo& out_NymIcon )
{
	if ( m_NymIconInfo.GetIconType() == NymIconTypeUndefined )
	{
		// Backward compatibility - read it out of the frRemote Nym

		const unsigned char* index = NULL;
		size_t szIndex = 0;

		ZKVERIFY(frRemoteNymResetParameterSearch (
			&m_ZkError ,                 
			m_RemoteNym.GetFrRemoteNym()) );

		if( frRemoteNymGetParameter (
				&m_ZkError ,
				m_RemoteNym.GetFrRemoteNym(),
				NymPhoto , // use for now as an index
				&index ,
				&szIndex )) 
		{
			m_NymIconInfo.SetIconIndex( *index ) ;
		}

		// Remove the parameter which was the old way to
		// save it just to clean stuff up

		ZKVERIFY(frRemoteNymResetParameterSearch (
			&m_ZkError ,                 
			m_RemoteNym.GetFrRemoteNym()));

		frRemoteNymDeleteParameter (
			&m_ZkError  ,                  
			m_RemoteNym.GetFrRemoteNym(),  
			NymPhoto );
	}

	out_NymIcon = m_NymIconInfo ;
}

void
RNym::SetNymIconInfo( const WNymIconInfo& in_NymIcon )
{
	m_NymIconInfo = in_NymIcon ;
}

// Returns empty string if Parameter does not exist
void 
RNym::GetNymStringParameter( 
	enum FrRemoteNymParameter in_WhichParam, 
	std::string & out_string )
{
	out_string.erase() ;
		
	const unsigned char* strNymName = NULL;
	size_t szstrNymName = 0;

	ZKVERIFY(frRemoteNymResetParameterSearch (
		&m_ZkError ,                 
		m_RemoteNym.GetFrRemoteNym() ));

	// If parameter doesn't exist we return an empty string

	if ( frRemoteNymGetParameter ( 
			&m_ZkError , 
			m_RemoteNym.GetFrRemoteNym() , 
			in_WhichParam , 
			&strNymName , 
			&szstrNymName ))
	{
		if ( szstrNymName > 0 )
		{
			out_string = (char*) strNymName ;
		}
	}
}

void 
RNym::SetNymStringParameter( 
	enum FrRemoteNymParameter in_WhichParam, 
	std::string & in_string )
{
	// Make sure only one instance of this description type exists
	frRemoteNymDeleteParameter (
		&m_ZkError  ,                  
		m_RemoteNym.GetFrRemoteNym() ,  
		in_WhichParam );

	ZKVERIFY(frRemoteNymAddParameter(         
		&m_ZkError, 
		m_RemoteNym.GetFrRemoteNym(), 
		in_WhichParam, 
		(const unsigned char*)in_string.c_str(), 
		in_string.length() + 1) );
}

void
RNym::SetNymPrivateNotes(std::string& in_str)
{
	SetNymStringParameter( NymPrivateNote, in_str ) ;
}

void
RNym::GetNymPrivateNotes(std::string& out_str) 
{
	GetNymStringParameter( NymPrivateNote, out_str ) ;
}

void
RNym::SetNymPublicNotes(std::string& in_str) 
{
	SetNymStringParameter( NymPublicNote, in_str ) ;
}

void
RNym::GetNymPublicNotes(std::string& out_str) 
{
	GetNymStringParameter( NymPublicNote, out_str ) ;
}

void 
RNym::SetNymDescription( std::string& in_str )
{	
	SetNymStringParameter( NymLongName, in_str ) ;
}

void 
RNym::GetNymDescription( std::string& out_str )
{	
	GetNymStringParameter( NymLongName, out_str ) ;
}

void 
RNym::GetSignatureKey(fr_crypt_pubkey** key) throw(RZkException)
{ 
	ZKFunctionEntry( "RNym::GetSignatureKey" ) ;
	
	int bStatus = true;

	// Initialize output
	*key = NULL;

	bStatus = frRemoteNymResetParameterSearch (
					&m_ZkError ,                 
					m_RemoteNym.GetFrRemoteNym() );


	ZKTestAndThrow(bStatus);

	FrCryptPubKey notused = NULL;
	
	bStatus = frRemoteNymGetPrivateKeys (
					&m_ZkError ,
					m_RemoteNym.GetFrRemoteNym() ,
					&notused,   /* Returns the private encryption key */
					key ) ;  /* Returns the private signature key */
	
	ZKTestAndThrow(bStatus);

    if (notused) 
	{
        frCryptPubKeyFree(NULL,&notused);
	}
}

void 
RNym::GetCryptoKey(fr_crypt_pubkey** key) throw(RZkException)
{ 
	ZKFunctionEntry( "RNym::GetCryptoKey" ) ;
	
	int bStatus = true;

	// Initialize output
	*key = NULL;

	bStatus = frRemoteNymResetParameterSearch (
					&m_ZkError ,                 
					m_RemoteNym.GetFrRemoteNym() );


	ZKTestAndThrow(bStatus);

	FrCryptPubKey notused = NULL;
	
	bStatus = frRemoteNymGetPrivateKeys (
					&m_ZkError ,
					m_RemoteNym.GetFrRemoteNym() ,
					key,   /* Returns the private encryption key */
					&notused) ;  /* Returns the private signature key */
	
	ZKTestAndThrow(bStatus);

    if (notused) 
	{
        frCryptPubKeyFree(NULL,&notused);
	}
}

bool 
RNym::GetNymName( std::string& strName )
{
	bool ReturnValue = true;

	const unsigned char* strNymName = NULL;
	size_t szstrNymName = 0;

	ZKVERIFY(frRemoteNymResetParameterSearch (
		&m_ZkError ,                 
		m_RemoteNym.GetFrRemoteNym() ));

	if(frRemoteNymGetParameter ( 
			&m_ZkError , 
			m_RemoteNym.GetFrRemoteNym() , 
			NymName , 
			&strNymName , 
			&szstrNymName ))
	{
		// make copy
		strName = (char*)strNymName;
		
		// not necessary to free mem
		strNymName = NULL;	
	}
	else 
	{
		ZKASSERT(false);
		ReturnValue = false;
	}
		

	return ReturnValue;
}

void 
RNym::GetNymBirthDate(std::string& strBirthDate) throw(RZkException)
{
	strBirthDate.erase() ;

	ZkDateTime zkdt;
	m_RemoteNym.GetDateParam( NymTimeStamp, zkdt ) ;
	ZKConvert::StringFromZKDateTime( zkdt, strBirthDate ) ;
}

void 
RNym::GetNymRetireDate(std::string& strRetireDate) throw(RZkException)
{
	strRetireDate.erase() ;

	ZkDateTime zkdt;
	m_RemoteNym.GetDateParam( NymExpiry, zkdt ) ;
	ZKConvert::StringFromZKDateTime( zkdt, strRetireDate ) ;
}

void 
RNym::GetNymDeathDate(std::string& strDate) throw(RZkException)
{
	strDate.erase() ;

	ZkDateTime zkdt;
	m_RemoteNym.GetDateParam( NymRelease, zkdt ) ;
	ZKConvert::StringFromZKDateTime( zkdt, strDate ) ;
}

void 
RNym::GetNymBirthDate(time_t& out_Date) throw(RZkException)
{
	memset( &out_Date, 0, sizeof( out_Date ) ) ;

	ZkDateTime zkdt;
	m_RemoteNym.GetDateParam( NymTimeStamp, zkdt ) ;
	ZKConvert::TimeTFromZKDateTime( zkdt, out_Date ) ;
}

void 
RNym::GetNymRetireDate(time_t& out_Date) throw(RZkException)
{
	memset( &out_Date, 0, sizeof( out_Date ) ) ;

	ZkDateTime zkdt;
	m_RemoteNym.GetDateParam( NymExpiry, zkdt ) ;
	ZKConvert::TimeTFromZKDateTime( zkdt, out_Date ) ;
}

void 
RNym::GetNymDeathDate(time_t& out_Date) throw(RZkException)
{
	memset( &out_Date, 0, sizeof( out_Date ) ) ;

	ZkDateTime zkdt;
	m_RemoteNym.GetDateParam( NymRelease, zkdt ) ;
	ZKConvert::TimeTFromZKDateTime( zkdt, out_Date ) ;
}


void 
RNym::SetRouteRulesOfNym(const MapRules& mapRules) 
{
	ZKFunctionEntry( "RNym::SetRouteRulesOfNym" ) ;

	int bStatus = true;

	WAutoBuffer buf ;

	bStatus = frRemoteNymDeleteParameter (
				&m_ZkError  ,                  
				m_RemoteNym.GetFrRemoteNym() ,  
				NymRules );

	ZKTestAndThrow(bStatus);
	
	if ( !mapRules.empty() )
	{

		MapRulesToBinary(
					mapRules, 
					&buf, 
					buf.GetSizePointer()) ;			

		bStatus = frRemoteNymAddParameter(         
					&m_ZkError, 
					m_RemoteNym.GetFrRemoteNym(), 
					NymRules, 
					buf, 
					buf.Size()) ;

		ZKTestAndThrow(bStatus);
	}


	DumpRouteRulesOfNym(mapRules);
}

void 
RNym::GetRouteRulesOfNym(MapRules& mapRules) throw(RZkException)
{
	ZKFunctionEntry( "RNym::GetRouteRulesOfNym" ) ;
	
	int bStatus = true;

	mapRules.clear();

	const unsigned char* buf = NULL;
	size_t szbuf = 0;

	bStatus	= frRemoteNymResetParameterSearch(
				&m_ZkError ,                 
				m_RemoteNym.GetFrRemoteNym());

	ZKTestAndThrow(bStatus);

	bStatus	= frRemoteNymGetParameter ( 
					&m_ZkError , 
					m_RemoteNym.GetFrRemoteNym() , 
					NymRules , 
					&buf , 
					&szbuf ) ;

	if ( bStatus ) 
	{
		MapRulesFromBinary(
				mapRules, 
				(unsigned char*)buf , 
				szbuf) ;
	}

	DumpRouteRulesOfNym(mapRules);
}

void 
RNym::SetSecuritySettings(const SecuritySettings& ss) throw(RZkException)
{
	ZKFunctionEntry( "RNym::SetSecuritySettings" ) ;

	ZKASSERT(ss.nHops>=MIN_HOPS && ss.nHops<=MAX_HOPS );

	WAutoBuffer buf ;

	int bStatus = frRemoteNymDeleteParameter (
				&m_ZkError  ,                  
				m_RemoteNym.GetFrRemoteNym() ,  
				NymSecuritySettings);

	ZKTestAndThrow2(m_ZkError, bStatus);
	
	SecuritySettingsToBinary(ss, &buf, buf.GetSizePointer() ) ;

	if ( buf.Size() > 0 ) 
	{
		bStatus = frRemoteNymAddParameter(         
					&m_ZkError, 
					m_RemoteNym.GetFrRemoteNym(), 
					NymSecuritySettings, 
					buf, 
					buf.Size()) ;

		ZKTestAndThrow(bStatus);
	}
}

size_t  
RNym::GetNumberOfHops()
{
	size_t nHops = MAX_HOPS ;

	try
	{
		SecuritySettings	ss;
		GetSecuritySettings(ss); // may throw a RZkException
		nHops = MAX_HOPS ;
		if ( ss.nHops !=-1 )
		{
			// Backward compatibility
			nHops = ss.nHops ;
		}
	}
	catch(IZkException&)
	{ 
		ZKASSERT( false ) ;
	}

	return nHops ;
}
	
void 
RNym::GetSecuritySettings(SecuritySettings& ss) throw(RZkException)
{
	ZKFunctionEntry( "RNym::GetSecuritySettings");
	
	// Initialize output
	ss.nHops = -1 ;
	ss.nReserved = 0 ;
	
	const unsigned char* buf = NULL;
	size_t szbuf = 0;

	int bStatus = frRemoteNymResetParameterSearch(
				&m_ZkError ,                 
				m_RemoteNym.GetFrRemoteNym());

	ZKTestAndThrow2(m_ZkError, bStatus);

	bStatus	= frRemoteNymGetParameter ( 
					&m_ZkError , 
					m_RemoteNym.GetFrRemoteNym() , 
					NymSecuritySettings , 
					&buf , 
					&szbuf ) ;

	if(bStatus)
	{
		SecuritySettingsFromBinary(ss, buf, szbuf) ;
	}
}





/////////////////////////////////////////////////////////////////////////////
//									private
/////////////////////////////////////////////////////////////////////////////

void 
RNym::MapRulesToBinary(
	const MapRules& mapRules, 
	unsigned char** binary, 
	size_t* szbinary)
{
	ZKASSERT(!mapRules.empty());

	// Initialize output
	*binary = NULL ;
	*szbinary = 0 ;

	// determine size and allocate
	size_t szbuf = mapRules.size() * sizeof(int) * 2 + BUFFER_HEADER_SIZE;
	unsigned char* buf = (unsigned char*) malloc( szbuf );

	MapRules::const_iterator theIterator;

	buf[0] = 0; // size
	buf[1] = 0; // size
	buf[2] = 1;    // Version

	size_t i = BUFFER_HEADER_SIZE;

	for (theIterator = mapRules.begin();
		theIterator != mapRules.end();
		++theIterator)
	{		
		memcpy(&(buf[i]), &(*theIterator).first, sizeof(int));
		i = i + sizeof(int) ;
		
		memcpy(&(buf[i]), &(*theIterator).second, sizeof(int));
		i = i + sizeof(int) ;
	}

	buf[0] = i/0x100; // size high
	buf[1] = i%0x100; // size low

	ZKASSERT( i == szbuf ) ;

	*binary = buf;
	*szbinary = i;
}


void 
RNym::MapRulesFromBinary(
	MapRules& mapRules, 
	unsigned char* buf, 
	size_t szbuf)
{
	ZKFunctionEntry( "RNym::MapRulesFromBinary" ) ;
	ZKTestPointer(buf) ;

	mapRules.clear() ;

	if ( ( buf[0] != szbuf/0x100 ) ||  
		 ( buf[1] != szbuf%0x100 ) ||
		 ( buf[2] != 1 ) )  
	{				
		ZKThrowRuntime( "Buffer contains incorrect size or parity", IDS_CORRUPT_BUFFER );
	}

	size_t i = BUFFER_HEADER_SIZE;
	
	while(i < szbuf)
	{		
		int ID = 0;
		RULEZ R = RULE_NONE ;

		memcpy(&ID, &buf[i], sizeof(ID) );
		i = i + sizeof(ID) ;

		memcpy(&R, &buf[i], sizeof(R) );
		i = i + sizeof(R) ;

		mapRules[ ID ] = R ;
	}

	ZKASSERT( i == szbuf ) ;
	ZKASSERT(!mapRules.empty());
}

#define SECURITY_SETTINGS_VER	8

// Called from SetSecuritySettings()
void 
RNym::SecuritySettingsToBinary(
	const SecuritySettings& ss,
	unsigned char** binary, 
	size_t* szbinary)
{
	ZKASSERT(ss.nHops>=MIN_HOPS && ss.nHops<=MAX_HOPS) ;
	ZKASSERT(ss.nReserved == SECURITY_SETTINGS_VER) ;

	// Initialize output
	*binary = NULL;
	*szbinary = 0;

	// size is determined, allocate	
	unsigned char* buf = (unsigned char*) malloc(2);

	buf[0] =  ss.nReserved; // this is SECURITY_SETTINGS_VER
	buf[1] =  ss.nHops;

	*binary = buf;
	*szbinary = ss.nReserved;
}

// Called from GetSecuritySettings()
void 
RNym::SecuritySettingsFromBinary(
	SecuritySettings& ss, 
	const unsigned char* buf, 
	size_t szbuf)
{
	ZKASSERT(ZKIsValidAddress(buf, szbuf));
	
	if ( buf[0] != SECURITY_SETTINGS_VER ) 
	{
		// Possibly corrupt data
		ZKASSERT( false ) ;
		return ;
	}

	ss.nReserved = buf[0];
	ss.nHops     = buf[1];

	ZKASSERT(ss.nHops>=MIN_HOPS && ss.nHops<=MAX_HOPS);
}


// If the server returns information it gets filled into
// a member structure.  This function allows the information
// to pass up to the called through exception handling
// techniques
void 
RNym::TestAndThrowServerResponse(int bStatus)
{
	if ( bStatus ) 
	{
		// No failure, nothing to do
		return ;
	}

	// General Server Message

	else if ( m_NymServerResponse.message[0] != 0 ) 
	{
		char buf[NSRMESG_SZ + 16]; // msg + code

		if (m_NymServerResponse.code != 0) 
		{
			snprintf(buf, sizeof(buf), "%s [%d]", m_NymServerResponse.message, m_NymServerResponse.code);
		}
		else
		{
			snprintf(buf, sizeof(buf), "%s", m_NymServerResponse.message);
		}

		throw RZkException( m_ZkError, "RNym::TestAndThrowServerResponse",buf, IDS_SERVER_ERROR, buf ) ;
	}

	// General Client Side Message
	else 
	{
		ZKASSERT( false ) ;
		throw RZkException( 
				m_ZkError,
				"RNym::TestAndThrowServerResponse", 
				"No Server Response Message Available", 
				IDS_SERVER_ERROR_UNKNOWN);
	}
}

// REVIEW - we could cache this information in our own
// vector of string.  This could save on allocations.

bool 
RNym::GetOffensiveWordList(
	std::vector<std::string>& vecstrBadWords)
{	
	bool ReturnValue = true;

	WZkListNoAlloc OffensiveList ;

	vecstrBadWords.clear();

	// Pull the list out of the Remote Nym structure
	if ( frRemoteNymGetStringListParameter(
						&m_ZkError ,              
						m_RemoteNym , 
						NymScanWord,
						&OffensiveList ) )
	{

		// Iterate through the ZkList to fill in the string vector
		size_t listsize = 0;

		ZKVERIFY(zkListSize (
			&m_ZkError ,              
			OffensiveList ,
			&listsize)); 

		ZKVERIFY(zkListMoveCursor (
			&m_ZkError ,               
			OffensiveList ,
			LIST_HEAD));

		void* badword = NULL;

		for(size_t i = 0; i < listsize; ++i)
		{			
			ZKVERIFY(zkListGet (
				&m_ZkError ,               
				OffensiveList ,
				&badword ,               
				LIST_FORWARDS ));  

			vecstrBadWords.push_back((char*)badword);
		}
	}

	return ReturnValue;
}

//TODO Take a look at RNymMail block list - the code is very close,
// in fact I suspect cut and past.  We should generalize conversion of
// string vector -> zkList ( and vis versa ) - Andrew

bool RNym::SetOffensiveWordList(
	const std::vector<std::string>& vecstrBadWords)
{
	bool ReturnValue = true ;

	// Erase any existing information
	frRemoteNymDeleteParameter (&m_ZkError, m_RemoteNym, NymScanWord );

	if ( vecstrBadWords.size() > 0)
	{
		WZkList OffensiveList ;
		zkListSetDataFreeFunction(&m_ZkError, OffensiveList, zkListDataFree, NULL );

		std::vector<std::string>::const_iterator theIterator ;
	
		for(theIterator = vecstrBadWords.begin() ;
			theIterator != vecstrBadWords.end();
			++theIterator)
		{
			// Fill in Offensive word list with the array
			void * pData = malloc( (*theIterator).length() + 1 ) ;
			memcpy( pData, (*theIterator).c_str(), (*theIterator).length() + 1 ) ;

			ZKVERIFY(zkListInsert (
				&m_ZkError ,
				OffensiveList ,
				pData , 
				LIST_HEAD ));		
		}
		
		ZKVERIFY(zkListMoveCursor(&m_ZkError, OffensiveList, LIST_HEAD));

		ZKVERIFY(frRemoteNymAddStringListParameter (
					&m_ZkError ,          
					m_RemoteNym ,
					NymScanWord,
					OffensiveList )) ;
	}

	return ReturnValue;
}

void 
RNym::DumpRouteRulesOfNym(const MapRules& in_mapRules)
{
#ifndef NDEBUG

	static char* table[RULE_EXITFUNNEL + 1];
	table[RULE_NONE]		= "None";
	table[RULE_RANDOM]		= "Random";
	table[RULE_SPECIFY]		= "Specify";
	table[RULE_FIRSTHOP]	= "First hop";
	table[RULE_AVOID]		= "Avoid";
	table[RULE_MUSTUSE]		= "Must use";
	table[RULE_EXITFUNNEL]	= "Exit funnel";

	ZKTRACE0("Dumping rule rules map\n");

	for(MapRules::const_iterator it = in_mapRules.begin(); it != in_mapRules.end(); ++it)
	{
		ZKTRACE3("For Aip ID [%d] ---> Rules are [%d] meaning [%s]\n", (*it).first, (*it).second, table[(*it).second] );		
	}

#endif
}

void
RNym::SaveRemoteNymChunk( WMemBuf& io_Buffer ) throw(RZkException)
{
	ZKChunkWriter cw( io_Buffer, FRREMOTENYM_CHUNK_ID, 1 ) ;
	
	WAutoBuffer ab ;
	m_RemoteNym.GetBuffer( &ab, ab.GetSizePointer() ) ;

	ZKASSERT( ab.Size() > 0 ) ;

	cw.WriteArray( ab, ab.Size() ) ;

#ifdef TRACE_SAVE_NYM_CENTRAL
	ZKTRACE0( "Saving Nym remoteNym\n\n");

	m_RemoteNym.DebugTraceRemoteNym( NULL ) ;
#endif
}

// Serialize the state into a chunked buffer
void 
RNym::Save( WMemBuf& io_Buffer ) throw(RZkException)
{
	ZKChunkWriter cw( io_Buffer, NYM_CHUNK_ID, NYM_PERSIST_VERSION ) ;
	SaveRemoteNymChunk( io_Buffer ) ;

	m_pNymWallet->Save( io_Buffer ) ;
	m_pNymCookieMgr->Save( io_Buffer ) ;
	m_pNymMail->Save( io_Buffer );

	m_pNymRoute->Save( io_Buffer ) ;

	m_pNymWebCacheMgr->Save( io_Buffer );

	if ( m_NymIconInfo.GetIconType() != NymIconTypeUndefined )
	{
		m_NymIconInfo.Save( io_Buffer ) ;
	}	
}

void 
RNym::LoadRemoteNymChunk( ZKChunkReader& io_ChunkReader ) throw(RZkException)
{
	ZKASSERT( io_ChunkReader.GetChunkID() == FRREMOTENYM_CHUNK_ID ) ;

	const uint8_t* pRNymData = NULL ;
	size_t   cntRNymData = 0 ;

	io_ChunkReader.ReadArray( (const void**)&pRNymData, &cntRNymData ) ;

	if (( pRNymData == NULL ) ||
		( cntRNymData == 0 ) )
	{
		ZKASSERT( false ) ;
		throw RZkException( "ZKChunkReader::ZKChunkReader", 
							 "Corrupt File",
							 IDS_CORRUPT_FREEDOM_DATA ) ;
	}

	m_RemoteNym.InitializeFromBuffer( pRNymData, cntRNymData ) ;
}

void 
RNym::Load( WReadBuffer& io_Buffer ) throw(RZkException)
{
	ZKChunkReader cr( io_Buffer ) ;

	if ( cr.GetChunkID() != NYM_CHUNK_ID )
	{
		// Corruption or bug in the load code, we must load in the correct order
		ZKASSERT( false ) ;
		throw RZkException( "RNym::Load", 
							 "Corrupt File",
							 IDS_CORRUPT_FREEDOM_DATA ) ;		
	}

	bool bFoundRemoteNymChunk = false ;

	while( !cr.IsEndOfChunk() )
	{
		ZKChunkReader crNested( io_Buffer ) ;

		switch( crNested.GetChunkID() )
		{
			case( FRREMOTENYM_CHUNK_ID ) :
				LoadRemoteNymChunk( crNested ) ;
				bFoundRemoteNymChunk = true ;
				break ;
				
			case( WALLET_CHUNK_ID ) :
				m_pNymWallet->Load( crNested ) ;
				break ;

			case( COOKIEJAR_CHUNK_ID ) :
				m_pNymCookieMgr->Load( crNested ) ;
				break ;

			case( MAIL_CHUNK_ID ) :
				m_pNymMail->Load( crNested ) ;
				break ;

			case( ICON_CHUNK_ID ) :
				m_NymIconInfo.Load( crNested ) ;
				break ;

			case( NYM_ROUTE_CHUNK_ID ) :
				m_pNymRoute->Load( crNested ) ;
				break ;

			case( NYM_WEB_CACHE_MGR_CHUNK_ID ) :
				m_pNymWebCacheMgr->Load( crNested ) ;
				break ;

			default :
				// Future chunks - just ignore them
				ZKASSERT( false ) ;
				break ;
		}		
	}		

	ZKASSERT( bFoundRemoteNymChunk ) ;

#ifdef TRACE_LOAD_NYM_CENTRAL
	ZKTRACE0( "Loading Nym remoteNym\n\n");

	m_RemoteNym.DebugTraceRemoteNym( NULL ) ;
#endif	
}

FrNymCapability
RNym::GetNymTypeCapability()
{
	// See Bug #6608.
	// This function is called by multiple thread and can fail (e.g. on amazon.com).
	// if we call it each time.  We assume by caching the capabilities that it won't change
	// over the time of execution.
	if( m_nymCapability == FRNYM_CAPABILITY_INIT )
	{
		ZKVERIFY(frRemoteNymGetCapability (
					&m_ZkError,
					m_RemoteNym.GetFrRemoteNym(),
					&m_nymCapability ));
	}
	
	return m_nymCapability;
}


bool 
RNym::DoesNymUseCloud()
{
	FrNymCapability capa = GetNymTypeCapability();

	return (capa & FRNYM_CAPABILITY_AUTHROUTES);
}


bool 
RNym::DoesNymUseMail()
{
	FrNymCapability capa = GetNymTypeCapability();

	return (capa & FRNYM_CAPABILITY_MAIL);
}


bool 
RNym::DoesNymAutoExtend()
{
	// Free Nyms are autoextended.
	return (!DoesNymUseMail() && !DoesNymUseCloud() ) ;
}

void
RNym::ExtendYourself()
{
	// Only a nym which has auto-extend capability can do this
	ZKASSERT( DoesNymAutoExtend() ) ;

	ZKSTD::string strToken ;

	try
	{
		ZkService::GetITokenService()->GetFreeNymExtendSerialNumberAndRedeemIt( strToken ) ;

		AddTokenToNymOnNymServer( strToken ) ;
	}
	catch( RZkException& )
	{
		// Don't leave an extend token in our array, we don't keep those
		// around even if we couldn't spend it
		if ( !strToken.empty() )
		{
			ZkService::GetITokenService()->RemoveTokenFromArray( strToken ) ;
		}

		throw ;
	}

	if ( !strToken.empty() )
	{
		ZkService::GetITokenService()->RemoveTokenFromArray( strToken ) ;
	}
}
