// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RNymCentral.cpp: implementation of the RNymCentral class.
//
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "RNym.h"
#include "INymMail.h"
#include "RNymCentral.h"
#include "INymWallet.h"
#include "IRouteManager.h"
//#include "resource.h"

#include <frCrypt.h>
#include <frNymUtil.h>
#include <zkDateTime.h>
#include <ZKConvert.h>
#include <WTokenDescriptions.h>
#include <IZkConsoleMsg.h>
#include <ZKChunk.h>
#include <IClientInfo.h>
#include <ICrypto.h>
#include <IClientCentral.h>
#include <ICookieManager.h>
#include <ZkService.h>
#include <ZKIP.h>
//#include <ZKStringSearch.h>
#include <lockguard.h>
#include <RZkException.h>
#include <IZKErrorReporter.h>
#include <ZkErrorHandler.h>
#include <ZkStringMgr.h>
#include <MiscMessages.h>
#include <zkDNS.h>
#include <ZkNetHelp.h>

#ifdef ZK_LINUX	
#include <netdb.h>		// for hostent
#include <../linux/RShimInterface.h> // For setting shim rules based on nym mode --ol
#endif


#define NYM_CENTRAL_PERSIST_VERSION 1

#define BUFFER_HEADER_SIZE 3

//#define FOUND_YOUR_IP "Your Internet Address (IP)"

static const char STR_TRUE_IDENTITY[] = "True Identity";
static const char USER_NYM_NAME[] = "User" ;                // Internal Name (not necessarily shown)


RNymCentral::RNymCentral() 
	: m_ActiveNym(USER_NYM_INDEX)
{	
	CreateUser() ;
}

RNymCentral::~RNymCentral()
{
	// Make sure all the Nyms are cleaned up
	FlushNyms() ;
}

void
RNymCentral::FlushNyms()
{
	ZKSTD::vector<RNym*>::iterator theIterator;
	
	for(theIterator = m_vecNymPointers.begin();
		theIterator != m_vecNymPointers.end();
		++theIterator)
	{
		RNym* pNym = (*theIterator);
		delete pNym;
	}

	m_vecNymPointers.clear();
}

void 
RNymCentral::GetNymIndex( 
    const ZKSTD::string& in_strNymName, 
    size_t& out_nymIndex )
{
	ZKSTD::vector<RNym*>::iterator theIterator;
	ZKSTD::string strName;
 
    // set the index to "user"
    out_nymIndex = USER_NYM_INDEX;
   
    int index = 0;
	for( theIterator = m_vecNymPointers.begin();
         theIterator != m_vecNymPointers.end();
         ++theIterator, index++ )
	{		
		ZKVERIFY((*theIterator)->GetNymName( strName ));
		
		if( in_strNymName == strName )
		{
            out_nymIndex = index;
			break ;
		}
	}

    ZKASSERT( "Name is in the Nym list" && 
              (theIterator != m_vecNymPointers.end() ) );
}


void 
RNymCentral::GetNymName( size_t in_indexNym, ZKSTD::string& out_strNymName )
{
	if ( in_indexNym < m_vecNymPointers.size() ) 
	{
		INym* pNym = NULL ;
		GetNym( IID_INYM, (void**)&pNym, in_indexNym ) ;
		pNym->GetNymName( out_strNymName ) ;
	}
	else 
	{
		out_strNymName.clear() ;
		ZKASSERT( false ) ;
	}
}

void 
RNymCentral::GetActiveNymName( ZKSTD::string& out_strNymName )
{
	GetNymName( m_ActiveNym, out_strNymName ) ;
}

ZKSTD::string 
RNymCentral::GetNoNymSelectedDisplayName()
{
	INymWallet* pNymWallet = NULL ;
	GetNym( IID_INYMWALLET, (void**)&pNymWallet, USER_NYM_INDEX ) ;
	ZKASSERT(pNymWallet);
	const ZKSTD::string strFirstName =  pNymWallet->GetField(WALLET_NAME_FIRST);
	const ZKSTD::string strLastName =  pNymWallet->GetField(WALLET_NAME_LAST);
	
	ZKSTD::string strDisplay;
	if ((strFirstName.size() > 0) && (strLastName.size() > 0))
	{
		strDisplay = strFirstName + " " + strLastName + " - " + STR_TRUE_IDENTITY;
	}
	else if (strFirstName.size() > 0)
	{
		strDisplay = strFirstName + " - " + STR_TRUE_IDENTITY;
	}
	else
	{
		strDisplay = STR_TRUE_IDENTITY;
	}
	return strDisplay;
}


// Nym Management

void 
RNymCentral::CreateNym(
		ZKNymName& NymName, 
		const ZKSTD::string& token ) throw(RZkException)
{
	RNym* pZKNewNym = new(ZKSTD::nothrow) RNym;

	if ( pZKNewNym == NULL ) 
	{
		throw RZkException("RNymCentral::CreateNym","NoMemory",IDS_OUT_OF_MEMORY ) ;
	}

	pZKNewNym->CreateYourself(NymName, token) ;

	m_vecNymPointers.push_back(pZKNewNym);

	// Populate the nym's wallet with pseudorandom personal info
	INymWallet* pNymWallet = NULL;
	pZKNewNym->QueryInterface(IID_INYMWALLET, (void**)&pNymWallet);
	if (!pNymWallet)
	{
		ZKASSERT(false);
	}
	else
	{
		pNymWallet->PopulatePersonalInfoWithRandomFields();
	}

#ifdef _DEBUG
	ZKSTD::string strName;
	pZKNewNym->GetNymName(strName);
	ZKTRACE1("Adding [%s] to m_vecNymPointers\n", strName.c_str() ); 
#endif
}

// These normally will only be called once, when the RNymCentral is being
// initialized by the setup wizard

void 
RNymCentral::CreateUser()
{
	// We are assuming that User is the first Nym to create
	ZKASSERT( GetNumberOfNyms() == 0 ) ;
	
	RNym* pZKNewNym = new(ZKSTD::nothrow) RNym;

	if ( pZKNewNym == NULL ) 
	{
		throw RZkException("RNymCentral::CreateUser","NoMemory",IDS_OUT_OF_MEMORY ) ;
	}	
	
	try 
	{
		pZKNewNym->SetUserName( USER_NYM_NAME );

		// User is always active
		pZKNewNym->SetYourLifeStage( NYM_LIFE_STAGE_ACTIVE );

		m_vecNymPointers.push_back(pZKNewNym);
	}
	catch( IZkException& ) 
	{
		delete pZKNewNym ;
		throw ;
	}
}


void RNymCentral::MarkNymAsDead(
		size_t in_nIndex)
{	
	if ( in_nIndex >= m_vecNymPointers.size() ||
		 in_nIndex == 0 )
	{
		// Bogus argument
		ZKASSERT( false ) ;
		return ;
	}

	// We don't take the nym out of vector,
	// in order to avoid invalidating other
	// nym index values.  However we mark the 
	// nym life stage so that it won't
	// be visible anymore.
	// It will be gone next time we load
	RNym* pNym = m_vecNymPointers[ in_nIndex ] ;
	pNym->SetYourLifeStage( NYM_LIFE_STAGE_RELEASED ) ;
}

bool RNymCentral::GetActiveNym(uint32_t in_iid, void** out_ppv)
{
	bool ReturnValue = true;
	*out_ppv = NULL ; // Initialize output

	RNym * pNym = m_vecNymPointers[ m_ActiveNym ] ;

	pNym->QueryInterface( in_iid, out_ppv );

	ZKASSERT( *out_ppv != NULL ) ;

	return true ;
}

size_t RNymCentral::GetActiveNymIndex()
{
	ZKASSERT( m_ActiveNym < GetNumberOfNyms() );	
	return m_ActiveNym ;
}

bool RNymCentral::IsCurrentNymUser() const
{
	return m_ActiveNym == USER_NYM_INDEX;
}

bool RNymCentral::SetActiveNym(const size_t nIndex)
{
	bool ReturnValue = true;
	
	if ( nIndex < m_vecNymPointers.size() ) 
	{
		m_ActiveNym = nIndex;
	}
	else 
	{
		ReturnValue = false ;
	}

	ZKASSERT( ReturnValue ) ;

	if(ZkService::GetIClientCentral()->GetPreference(ZK_CLIENT_PREF_SET, PREF_CHECK_MAIL_FOR_ACTIVE_NYM_ONLY))
	{
		ZkService::GetIRouteManager()->TerminateMailRoutes();
	}

#ifdef ZK_LINUX
    // Push rules into the linux kernel based on our current
    // active nym. --ol
    // Note: these can throw if the ioctl goes awry
    if ( IsCurrentNymUser() ) RShimInterface::getShimInterface().UserMode();
    else RShimInterface::getShimInterface().NymMode();    
#endif
	return ReturnValue;
}

bool RNymCentral::GetNym(
	uint32_t in_iid, 
	void** out_ppv, 
	const ZKSTD::string& strNymName)
{
	ZKASSERT( m_vecNymPointers.size() > 0 );

	bool ReturnValue = true;

	ZKSTD::vector<RNym*>::iterator theIterator;
	*out_ppv = NULL ;
	ZKSTD::string strName;

	for(theIterator = m_vecNymPointers.begin();
		theIterator != m_vecNymPointers.end();
		++theIterator)
	{		
		ZKVERIFY((*theIterator)->GetNymName(strName));
		
		if (strNymName == strName)
		{
			RNym * pNym = (*theIterator);
			pNym->QueryInterface( in_iid, out_ppv );

			ZKASSERT( *out_ppv != NULL ) ;
			break ;
		}
	}	

	return ReturnValue;
}


bool RNymCentral::GetNym(
	uint32_t in_iid, 
	void** out_ppv, 
	const size_t in_nIndex)
{
	ZKASSERT( m_vecNymPointers.size() > 0 );

	if ( out_ppv == NULL )
	{
		ZKASSERT( false ) ;
		return false ;
	}

	if ( in_nIndex >= m_vecNymPointers.size() )
	{
		// Bogus argument
		ZKASSERT( false ) ;
		return false ;
	}

	// Initialize output
	*out_ppv = NULL ;

	RNym* pNym = m_vecNymPointers[ in_nIndex ] ;

	pNym->QueryInterface( in_iid, out_ppv );

	ZKASSERT( *out_ppv != NULL ) ;

	return true ;
}

// Call this function each time freedom starts to determine whether the 
// the Nyms are active or not

void 
RNymCentral::InitNymLifeCycle() throw(RZkException)
{	
	time_t timeNow ;
	time_t timeRetire ;
	time_t timeDeath ;

	// We use NIQS time if we got it, otherwise have
	// to use the users machine time which might be way off
	time( &timeNow ) ;

	if(ZkService::GetIClientInfo()->GetCurrentUTCFromClientInfo() != 0)
	{
		timeNow = ZkService::GetIClientInfo()->GetCurrentUTCFromClientInfo();
	}

	size_t cntNyms = GetNumberOfNyms() ;

	INym* pNym = NULL ;
	ZKSTD::string strNymName ;

	// Look at each nym to see what stage of life it is at

	for( size_t nIndex = 1 ; nIndex < GetNumberOfNyms() ; nIndex++ )
	{
		GetNym( IID_INYM, (void**) &pNym, nIndex ) ;

		pNym->GetNymName(strNymName);
		pNym->GetNymRetireDate( timeRetire ) ;
		pNym->GetNymDeathDate( timeDeath ) ;

		ZKASSERT( timeDeath >= timeRetire ) ;

		if ( timeNow > timeDeath )
		{
			// No network communication necessary - just a client operation
			pNym->SetYourLifeStage( NYM_LIFE_STAGE_RELEASED ) ;
			ZKTRACE1( "%s is being removed because it is dead\n", strNymName.c_str() ) ;
		}
		else if ( timeNow > timeRetire )
		{
			pNym->SetYourLifeStage( NYM_LIFE_STAGE_EXPIRED ) ;
		}
		else
		{
			pNym->SetYourLifeStage( NYM_LIFE_STAGE_ACTIVE ) ;

		}
	}

	// Special case - User Nym is always active
	GetNym( IID_INYM, (void**) &pNym, USER_NYM_INDEX ) ;
	pNym->SetYourLifeStage( NYM_LIFE_STAGE_ACTIVE ) ;
}

void 
RNymCentral::HandleNymsLifeCycleIssues( 
	size_t in_nNymIndex, 
	ZKSTD::string& out_strWarning ) throw(RZkException)
{	
	// Initialize output
	out_strWarning.erase() ;

	time_t timeNow ;
	time_t timeRetire ;
	time_t timeDeath ;

	// We use NIQS time if we got it, otherwise have
	// to use the users machine time which might be way off
	time( &timeNow ) ;

	if(ZkService::GetIClientInfo()->GetCurrentUTCFromClientInfo() != 0)
	{
		timeNow = ZkService::GetIClientInfo()->GetCurrentUTCFromClientInfo();
	}

	ZKASSERT( in_nNymIndex < GetNumberOfNyms() ) ;

	INym* pNym = NULL ;
	GetNym( IID_INYM, (void**) &pNym, in_nNymIndex ) ;

	ZKSTD::string strNymName ;

	char strErr[ 256 ] ;
	strErr[0] = 0 ;

	pNym->GetNymName(strNymName);

	pNym->GetNymRetireDate( timeRetire ) ;
	pNym->GetNymDeathDate( timeDeath ) ;

	ZKASSERT( timeDeath >= timeRetire ) ;

	unsigned long dwLifeStage ;
	pNym->GetYourLifeStage( dwLifeStage ) ;

	bool bAutoextendNym = pNym->DoesNymAutoExtend() ;

	switch( dwLifeStage )
	{
		case( NYM_LIFE_STAGE_UNDEFINED ):
			// We need to initialize Nym Life cycle informaiton before
			// calling this
			ZKASSERT( false ) ;
			break ;

		case( NYM_LIFE_STAGE_ACTIVE ):
			{
				// Check how long is remaining and show a warning if it is close

				int DaysRemaining = (int) ( ( timeRetire - timeNow ) / ( 60 * 60 * 24 ) ) ;

				if ( bAutoextendNym )
				{
					HandleAutoExtendNym( pNym, strNymName, DaysRemaining, out_strWarning ) ;
				}
				else
				{
					// Warn the user to extend the nym if we are starting to get close

					switch ( DaysRemaining )
					{
						case( 25 ) :
						case( 15 ) :
  						case( 10 ) :
						case( 5 ) :
						case( 3 ) :
						case( 2 ) :
						case( 1 ) :

							if(DaysRemaining != 1)
							{
								ZkStringMgr::Load( out_strWarning, IDS_NYM_EXPIRE_WARNING, strNymName.c_str(), DaysRemaining ) ;
							}
							else
							{
								ZkStringMgr::Load( out_strWarning, IDS_NYM_EXPIRE_ONE_DAY_WARNING, strNymName.c_str() ) ;
							}

							// Shorter message for console
							if(DaysRemaining != 1)
							{
								IZkConsoleMsg::ShowMsg( IDS_CONS_NYM_EXPIRE_WARNING, strNymName.c_str(), DaysRemaining ) ;
							}
							else
							{
								IZkConsoleMsg::ShowMsg( IDS_CONS_NYM_EXPIRE_ONE_DAY_WARNING, strNymName.c_str()) ;
							}

							ZKTRACE2( "Nym %s will expire in %d days\n", strNymName.c_str(), DaysRemaining ) ;
							break ;

						case( 0 ) :

							ZkStringMgr::Load( out_strWarning, IDS_NYM_EXPIRE_TODAY_WARNING, strNymName.c_str() ) ;

							// Shorter message for console
							IZkConsoleMsg::ShowMsg( IDS_CONS_NYM_EXPIRE_TODAY_WARNING, strNymName.c_str()) ;

							ZKTRACE1( "Nym %s will expire today\n", strNymName.c_str() ) ;
							break ;
					}
				}
			}
			break ;


		case( NYM_LIFE_STAGE_EXPIRED ) :
			{
				// Nym has expired, we warn if the nym is close to 
				// retiring 

				int DaysRemaining = (int) ( ( timeDeath - timeNow ) / ( 60 * 60 * 24 ) ) ;

				if ( bAutoextendNym )
				{
					HandleAutoExtendNym( pNym, strNymName, 0, out_strWarning ) ;
				}
				else
				{				
					if ( DaysRemaining > 0 ) 
					{
						// Current policy is every day!

						ZkStringMgr::Load( out_strWarning, IDS_NYM_EXPIRED_WARNING, strNymName.c_str(), DaysRemaining ) ;

						// Shorter message for console
						IZkConsoleMsg::ShowMsg( IDS_CONS_NYM_EXPIRED_WARNING, strNymName.c_str(), DaysRemaining ) ;

						ZKTRACE2( "Expired Nym %s will die in %d days\n", strNymName.c_str(), DaysRemaining ) ;
					}
					else 
					{
						ZkStringMgr::Load( out_strWarning, IDS_NYM_EXPIRED_LASTDAY_WARNING, strNymName.c_str() ) ;

						// Shorter message for console
						IZkConsoleMsg::ShowMsg( IDS_CONS_NYM_EXPIRED_LASTDAY_WARNING, strNymName.c_str() ) ;

						ZKTRACE1( "Expired Nym %s will die today\n", strNymName.c_str() ) ;				
					}
				}
			}

			break ;

		case( NYM_LIFE_STAGE_RELEASED ) :
			{
				// Nym is dead on server and needs to be removed from client

				ZkStringMgr::Load( out_strWarning, IDS_NYM_DEAD_WARNING, strNymName.c_str() ) ;
				IZkConsoleMsg::ShowMsg( IDS_CONS_NYM_DEAD_WARNING, strNymName.c_str() ) ;
			}
			break ;

		default :
			ZKASSERT( false ) ;
			break ;
	}
}

void
RNymCentral::HandleAutoExtendNym( 
	INym* in_pNym, 
	const ZKSTD::string& in_strNym, 
	int in_DaysRemaining, 
	ZKSTD::string& out_strMsgToUser )
{
	if (( in_pNym == NULL ) ||
		( in_strNym.empty() ) )
	{
		ZKASSERT( false ) ;
		return ;
	}

	// We autoextend any nym with less than a month before
	// it expires.  This automatically includes "grace period" who
	// have a 0 (or conceptually negative) days remaining
	if ( in_DaysRemaining > 30 )
	{	
		// Nothing to do
		return ;
	}

	try
	{
		in_pNym->ExtendYourself() ;

		// I'm not we even need to tell the user that we were successful
		ZkStringMgr::Load( out_strMsgToUser, IDS_NYM_AUTOEXTEND_SUCCESSFUL, in_strNym.c_str() ) ;

		IZkConsoleMsg::ShowMsg( 
						CONS_VERBOSE, 
						CONSIMAGE_COMP_TERM, 
						IDS_CONS_AUTOEXTENDED, 
						in_strNym.c_str() ) ;

		// Make sure the nym is marked as active now (it might have been expired)
		in_pNym->SetYourLifeStage( NYM_LIFE_STAGE_ACTIVE ) ;
	}
	catch( IZkException& ex ) 
	{
		// We just show the message in the console
		IZkConsoleMsg::ShowMsg( 
						CONS_VERBOSE, 
						CONSIMAGE_COMP_TERM, 
						IDS_CONS_AUTOEXTENDFAIL, 
						in_strNym.c_str(),
						ZkErrorHandler::GetIZkErrorReporter()->GetErrorString(ex).c_str() 
						) ;
        ZKTRACE2(
			"Extend error: Nym %s, Error Details: %s\n", 
			in_strNym.c_str(),
			ZkErrorHandler::GetIZkErrorReporter()->GetErrorString(ex).c_str());
	}
}

size_t 
RNymCentral::GetNumberOfNyms()
{
	return m_vecNymPointers.size();
}

#ifdef WIN32 // XXX BROKN BUILD ol
void
RNymCentral::AttemptDecryptWithNym( 
	WCypherMailMsg&  io_MailMsg, 
	ZKSTD::string in_strNym ) throw(RZkException)
{
	ZKASSERT( !in_strNym.empty() ) ;

	INymMail* pNymMail = NULL ;
	GetNym( IID_IMAIL, (void**) &pNymMail, in_strNym ) ;

	pNymMail->DecryptMessage( io_MailMsg ) ;


	// Verify that expected information has been filled in
	ZKASSERT( io_MailMsg.GetReplyBlock() != WCypherMailMsg::UnknownValue ) ;

#ifdef _DEBUG
	ZKSTD::string RecievingNym ;
	io_MailMsg.GetRecipientNym(RecievingNym) ;

	ZKASSERT( RecievingNym.compare( in_strNym ) == 0 ) ;
#endif
}
#endif

void 
RNymCentral::DecryptMessage( WCypherMailMsg&  io_MailMsg ) throw(RZkException)
{
	ZKFunctionEntry( "RNymCentral::DecryptMessage" );

	bool bStatus = false;

	ZKASSERT( io_MailMsg.GetEncryptedSize() > 0 ) ;

	ZKSTD::string strNymName ;
	io_MailMsg.GetRecipientNym( strNymName )  ;

	if ( strNymName.empty() ) 
	{
		// We don't know which nym to use yet
		ZKSTD::vector<RNym*>::iterator theIterator;

		theIterator = m_vecNymPointers.begin();

		for(++theIterator;							// Skip past the user nym, for the moment
													// it has not key
			theIterator != m_vecNymPointers.end();
			++theIterator)
		{
			RNym *    pNym = (*theIterator);

			pNym->GetNymName( strNymName ) ;

			try 
			{
#ifdef WIN32 // XXX BROKEN BUILD ol
				AttemptDecryptWithNym( io_MailMsg, strNymName ) ;
#endif
				bStatus = true ;
				break ;
			}
			catch( IZkException& ) 
			{
				// We expect failures until we find the correct Nym
			}
		}

		if ( !bStatus ) 
		{
			ZKThrowRuntime("Could not decrypt message", IDS_COULDNOT_DECRYPT);		
		}
	}
	else 
	{
#ifdef WIN32 // XXX BROKEN BUILD ol
		AttemptDecryptWithNym( io_MailMsg, strNymName ) ;
#endif
	}
}

void RNymCentral::DumpStateToTraceWindow()
{
#ifdef TRACE_SAVE_NYM_CENTRAL
#ifdef _DEBUG
	ZKTRACE0( "********************DUMP of NYMCENTRAL STATE*********************\n\n");

	ZKTRACE0( "**NYM DUMP**\n");

	int i = 0 ;

	ZKSTD::vector<RNym*>::iterator theIterator;
	
	for(theIterator = m_vecNymPointers.begin();
		theIterator != m_vecNymPointers.end();
		++theIterator, ++i)
	{
		if ( i == 0 ) 
		{
			ZKTRACE0( "<<<<<<<<<<<<<<<<USER NYM (NYM 0)>>>>>>>>>>>>>>>>\n" ) ;
		}
		else 
		{
			ZKTRACE1( "<<<<<<<<<<<<<<<<NYM %d>>>>>>>>>>>>>>>>>>>>>>>>>>\n", i ) ;
		}
		
		(*theIterator)->RemoteNym().DebugTraceRemoteNym( NULL ) ;
	}
	
#endif
#endif
}

void 
RNymCentral::Load( WReadBuffer& io_Buffer ) throw(RZkException)
{
	ZKChunkReader cr( io_Buffer ) ;

	if ( cr.GetChunkID() != NYM_CENTRAL_CHUNK_ID )
	{
		// Corruption or bug in the load code, we must load in the correct order
		ZKASSERT( false ) ;
		throw RZkException("RNymCentral::Load", 
							 "Corrupt File",
							 IDS_CORRUPT_FREEDOM_DATA ) ;		
	}
	
	// Verify that there are no nyms in the index already
	FlushNyms() ;

	// We need to read the nym chunks out in order
	while( !cr.IsEndOfChunk() )
	{
		AddNym( io_Buffer ) ;
	}

	// We expect at least 1 nym (the user nym) in Nym central

	if ( GetNumberOfNyms() == 0 )
	{
		ZKASSERT( "Why did we load a file with no nyms?" && false ) ;
		CreateUser() ;
	}
	else
	{
		ZKSTD::string strNymName ;
		GetNymName( 0, strNymName ) ;
		ZKASSERT( strNymName.compare( USER_NYM_NAME ) == 0 ) ;
	}
}
	
void
RNymCentral::AddNym( WReadBuffer& io_Buffer ) throw(RZkException)
{
	RNym* pZKNewNym = new RNym;

	try
	{
		pZKNewNym->Load( io_Buffer ) ;
	}
	catch( IZkException& )
	{
		delete pZKNewNym;
		throw ;
	}

	m_vecNymPointers.push_back(pZKNewNym);

#ifdef _DEBUG
		ZKSTD::string strName;
		pZKNewNym->GetNymName(strName);
		ZKTRACE1("Adding [%s] to m_vecNymPointers\n", strName.c_str() ); 
#endif
}


// Serialize the state into a chunked buffer
void 
RNymCentral::Save( WMemBuf& io_Buffer ) throw(RZkException)
{
	ZKChunkWriter cw( io_Buffer, NYM_CENTRAL_CHUNK_ID, NYM_CENTRAL_PERSIST_VERSION ) ;

	// Step through the Nyms, saving ones that are still alive
	for( size_t nIndex = 0 ; nIndex < GetNumberOfNyms() ; nIndex++ ) 
	{
		INym * pNym = NULL ;
		GetNym( IID_INYM, (void**)&pNym, nIndex ) ;
		
		if ( pNym != NULL ) 
		{
			unsigned long dwLifeStage = 0 ;
			pNym->GetYourLifeStage( dwLifeStage ) ;

			if ( dwLifeStage != NYM_LIFE_STAGE_RELEASED ) 
			{
				pNym->Save( io_Buffer ) ;
			}
		}
	}
}



// Old persistence methods


bool 
RNymCentral::AddNym_Obsolete( unsigned char* in_pBuffer, unsigned long in_ulBufferSize )
{
	if (( in_pBuffer == NULL ) || 
		( in_ulBufferSize == 0 )) 
	{
		ZKASSERT( false ) ;
		return false ;
	}
	
	RNym* pZKNewNym = new RNym;

	if (!pZKNewNym->RemoteNym().InitializeFromBuffer( in_pBuffer, in_ulBufferSize)) 
	{
		delete pZKNewNym;
		return false ;
	}
	else 
	{
		m_vecNymPointers.push_back(pZKNewNym);

#ifdef _DEBUG
		ZKSTD::string strName;
		pZKNewNym->GetNymName(strName);
		ZKTRACE1("Adding [%s] to m_vecNymPointers\n", strName.c_str() ); 
#endif
	}

	return true ;
}

bool 
RNymCentral::GetNymAsBuffer_Obsolete( 
	size_t in_WhichNym, 
	unsigned char** out_ppBuffer, 
	size_t* cntBuffer )
{
	if ( in_WhichNym >= m_vecNymPointers.size() ) 
	{
		// Nym doesn't exist
		ZKASSERT( false ) ;
		return false ;
	}

	RNym * pNym = m_vecNymPointers[ in_WhichNym ] ;

#ifndef NDEBUG
	ZKSTD::string strName;
	pNym->GetNymName(strName);
	ZKTRACE1("In RNymCentral::GetNymAsBuffer() getting buffer for [%s]\n", strName.c_str() );
#endif

	return pNym->RemoteNym().GetBuffer( out_ppBuffer, cntBuffer ) ;
}


bool
RNymCentral::DoesNymUseCloud(const size_t in_indexNym)
{
	ZKASSERT(in_indexNym < m_vecNymPointers.size());

	INym* pNym = NULL ;
	GetNym( IID_INYM, (void**)&pNym, in_indexNym ) ;
	
	return (pNym->DoesNymUseCloud());
}


bool
RNymCentral::DoesNymUseMail(const size_t in_indexNym) 
{
	ZKASSERT(in_indexNym < m_vecNymPointers.size());

	INym* pNym = NULL ;
	GetNym( IID_INYM, (void**)&pNym, in_indexNym ) ;
	
	return (pNym->DoesNymUseMail());
}

bool 
RNymCentral::AtLeastOneNymUseMail()
{
	bool bRetVal = false;
	for( size_t i = 0; i < m_vecNymPointers.size(); ++i )
	{
		if( i != USER_NYM_INDEX && m_vecNymPointers[i]->DoesNymUseMail() )
		{
			bRetVal = true;
			break;
		}
	}

	return bRetVal;
}

bool 
RNymCentral::AtLeastOneNymUseCloud()
{
	bool bRetVal = false;
	for( size_t i = 0; i < m_vecNymPointers.size(); ++i )
	{
		if( i != USER_NYM_INDEX && m_vecNymPointers[i]->DoesNymUseCloud() )
		{
			bRetVal = true;
			break;
		}
	}

	return bRetVal;
}
