// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "StdAfx.h"
#include "RNym.h"
#include "RNymCookieMgr.h"

#ifdef _WIN32
#pragma warning (disable : 4786)
#endif

#include <sstream>
#include <errno.h> 
#include <zks.h>
#include <IZkConsoleMsg.h>
#include <ZKChunk.h>
#include <INym.h>
#include <ICrypto.h>
#include <INymCentral.h>
#include <IClientPersist.h>
#include <ZkService.h>
#include <string_hlp.h>
#include <ci_string.h> // for strnicmp
#include <lockguard.h>
#include <date_hlp.h>

static const char OLD_COOKIE_FILE_EXT[] = ".CookieJar";		// V1X used seperate files
const char * const SESSION_COOKIE  = "Session Cookie";

#define COOKIEJAR_PERSIST_VERSION 1
#define COOKIE_PERSIST_VERSION    1

RNymCookieMgr::RNymCookieMgr( RNym* pParent )
	: m_pParent(pParent)
{
}

RNymCookieMgr::~RNymCookieMgr()
{
}


bool 
RNymCookieMgr::DomainMatch(const ZKSTD::string& lhs, const ZKSTD::string& rhs)
{
	bool ReturnValue = false;

	if(lhs == rhs) 
	{
		ReturnValue = true;
	}
	else if(!DomainIsIP(lhs) && lhs.find('.') != lhs.npos)
	{	
		// remove partial domains until a match, or no more partial domains..
		// i.e., for a.b.c.com, we test .b.c.com, then .c.com
		// Note that we do not match only top-level domains (i.e. '.net')

		ZKSTD::string part = lhs;

		part = part.substr(part.find_first_of('.', 1), part.size() - 1);
		while(part.find_last_of('.') != part.find_first_of('.')) 
		{
			if(part == rhs) 
			{				
				// A partial match, lhs domain-matches rhs
				ReturnValue = true;
				break;
			}
			part = part.substr(part.find_first_of('.', 1), part.size() - 1);

		}
	}

	return ReturnValue;
}


bool 
RNymCookieMgr::DomainIsIP(const ZKSTD::string& in_strDomainName)
{
	unsigned long	address = INADDR_NONE;

	address = inet_addr(in_strDomainName.c_str());

	if (address == INADDR_NONE)
	{
		return false; // not an ip addr
	}
	else 
	{
		return true;
	}
}

bool 
RNymCookieMgr::BuildCookieField(const ZKSTD::string& in_strDomainName, ZKSTD::string& out_Cookie)
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);	
	
	bool bFoundCookiesForDomain = false;
	
	if(m_pParent->GetCookiesEnabled())
	{
		COOKIEJARMAP::iterator theIterator = m_NymCookieMap.begin();

		while(theIterator != m_NymCookieMap.end())
		{
			//ZKTRACE2( "RNymCookieMgr::BuildCookieField gonna compare domains [%s] [%s]\n", in_strDomainName.c_str(), (*theIterator).second.Domain.c_str() );
		
			if(DomainMatch(in_strDomainName, (*theIterator).second.Domain))
			{
				//ZKTRACE0( "They matched\n");

				//We have a match. Append it to our cookie field
				out_Cookie.append( (*theIterator).first );

				out_Cookie += "; ";

				bFoundCookiesForDomain = true;
			}
			else
			{
				//ZKTRACE0( "They didn't match\n");
			}

			++theIterator;
		}

		if(bFoundCookiesForDomain) 
		{
			// remove the trailing ';' to not confuse web servers
			out_Cookie.erase(out_Cookie.size() - 2, 2);
		}
	}

	if ( bFoundCookiesForDomain ) 
	{
		ZkService::GetICookieManager()->SignalCookieEvent() ;
	}

	//ZKTRACE1( "RNymCookieMgr::BuildCookieField returning [%s]\n", out_Cookie.c_str() );

	return bFoundCookiesForDomain;
}

bool 
RNymCookieMgr::SetNymCookieMapEntry(
	const ZKSTD::string& in_strNameValue, 
	const ZKSTD::string& in_strDomainName, 
	const ZKSTD::string& in_strExpires)
{

	ZKS::Lock_Guard<WZkMutex> gate(m_cs);

	// Returns false if we fail to set the Map Entry
	bool ReturnValue = true;

	if ( in_strNameValue.empty() ||
		 in_strDomainName.empty() ||
		 in_strExpires.empty() )
	{	
		ZKASSERT(false) ;
		ReturnValue = false ;
	}
	else
	{
		if (!m_pParent->GetCookiesEnabled()) 
		{
			// User doesn't want us to accept any cookies
			ReturnValue = false ;
		}	
		else
		{
			// Determine the details of the new cookie coming in
			DomainAndDate entry;
			long lCUTTime = 0;

			entry.Domain = in_strDomainName;
			entry.DateExpire = in_strExpires;

			time(&lCUTTime);
			ZKSTD::string DateRecd (ctime(&lCUTTime)); // changes to use local time
			entry.DateRecd = DateRecd.substr(0, 24); // ditch the '\n\0'
			 
			if( !IsSessionCookie( in_strExpires ))
			{
				long lSeconds = 0;
				struct tm tmExpires;
				memset(&tmExpires, 0, sizeof(tm));
				
				ZKS::ParseDateString(in_strExpires, tmExpires);

				if(!(ZKS::GetSecondsCUT(tmExpires, lSeconds)))
				{
					ReturnValue = false;
					ZKASSERT(false);
				}

				entry.SecondsExpire = lSeconds;
			}
			else 
			{
				entry.SecondsExpire = 0;
			}

			// Stick it in the map
			if(ReturnValue)
			{			
				// Check to see if a previous cookie with same
				// name from same domain is in cookie jar. If so,
				// remove the old entry, put in the new one.

				bool match = false;
				COOKIEJARMAP::iterator theIterator = m_NymCookieMap.begin();

				ZKSTD::string NameOnly(in_strNameValue.substr(0, in_strNameValue.find_first_of('=')));

				// setup for partial domain names (i.e. .part.com)
				ZKSTD::string::size_type pos1 = in_strDomainName.find_first_of('.');
				ZKSTD::string PartialDomain;
				if(pos1 != in_strDomainName.npos)
				{
					PartialDomain = in_strDomainName.substr(pos1, in_strDomainName.size()-pos1);
				}

				//Iterate through the whole map
				while(theIterator != m_NymCookieMap.end())
				{
					const ZKSTD::string   strCookieData = (*theIterator).first ;
					const DomainAndDate strCookieInfo = (*theIterator).second ;					

					// Only compare the actual name
					ZKSTD::string strCookieName(strCookieData.substr(0, strCookieData.find_first_of('='))) ;

					ZKTRACE2("Looking through cookie jar for a cookie [%s] currently looking at [%s]\n", NameOnly.c_str(), strCookieName.c_str() );

					if(strCookieName == NameOnly) 
					{
						if ( strCookieInfo.Domain == in_strDomainName || strCookieInfo.Domain == PartialDomain) 
						{
							m_NymCookieMap.erase(theIterator);

							DumpCookie(entry);

							// NOTE by georgesh
							// If we inserted right back where we deleted from, would our cookies come out in 
							// temporal order, the way IE's and Netscape's do, rather than in alphabetical order,
							// the way they do now?  That could be good

							// Handle the 'pre-expired cookie' case: we deleted the matching cookie, but we
							// won't put anything back in!
							if( !IsSessionCookie( entry.DateExpire ) && entry.SecondsExpire < lCUTTime)
							{
								ZKTRACE1("Cookie [%s] is already expired; matches and therefore deletes existing" \
									"cookie of same name.\n", in_strNameValue.c_str() );
								ReturnValue = false;
							}
							else 
							{							
								m_NymCookieMap.insert(COOKIEJARMAP::value_type(in_strNameValue, entry));
							}

							// Remember that we have updated a cookie
							ZkService::GetICookieManager()->SignalCookieEvent() ;

							// Remember that our state has changed
							ZkService::GetIClientPersist()->SetDirty() ;

							INymCookieMgr::NymCookieMgrHint hint(INymCookieMgr::NymCookieMgrHint::NYMCOOKIEMGR_UPDATED_COOKIE);
							notify(&hint);

							match = true;
							break;
						}

					}
					++theIterator;
				}
				
				if(!match )
				{ 
					// Make sure it isn't pre-expired!
					if( !IsSessionCookie( entry.DateExpire ) && entry.SecondsExpire < lCUTTime)
					{
						ZKTRACE1("Cookie [%s] is already expired\n", in_strNameValue.c_str() );
						ReturnValue = false;
					}
					else
					{
						// No match, just stick this cookie in the map
						DumpCookie(entry);
						m_NymCookieMap.insert(COOKIEJARMAP::value_type(in_strNameValue,entry));
						// Remember that we have added a cookie
						ZkService::GetICookieManager()->SignalCookieEvent() ;

						// Remember that our state has changed so that
						// we get re-persisted
						ZkService::GetIClientPersist()->SetDirty() ;

						INymCookieMgr::NymCookieMgrHint hint(INymCookieMgr::NymCookieMgrHint::NYMCOOKIEMGR_ADDED_COOKIE);
						notify(&hint);

						IZkConsoleMsg::ShowMsg( CONSIMAGE_COOKIE, IDS_CONS_COOKIE_NEW, in_strDomainName.c_str() ) ;
					}
				}				
			}
		}
	}

	return ReturnValue;
}

void 
RNymCookieMgr::Save( WMemBuf& io_Buffer )
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);	
	
	ZKChunkWriter cw( io_Buffer, COOKIEJAR_CHUNK_ID, COOKIEJAR_PERSIST_VERSION ) ;

	COOKIEJARMAP::iterator theIterator;

	long ltime;   
	time(&ltime);

	for(theIterator = m_NymCookieMap.begin(); theIterator != m_NymCookieMap.end(); ++theIterator)
	{			
		if( ( !IsSessionCookie( (*theIterator).second.DateExpire ) ) && ((*theIterator).second.SecondsExpire > ltime) )
		{
			SaveCookie( io_Buffer, (*theIterator).first, (*theIterator).second ) ;
		}
	}
}

void 
RNymCookieMgr::SaveCookie( 
	WMemBuf& io_Buffer, 
	const ZKSTD::string& in_strNameValue, 
	const DomainAndDate& in_DomainAndDate )
{
	ZKChunkWriter cw( io_Buffer, COOKIE_CHUNK_ID, COOKIE_PERSIST_VERSION ) ;

	cw.WriteNonUIString( in_strNameValue ) ;
	cw.WriteNonUIString( in_DomainAndDate.DateRecd ) ;
	cw.WriteNonUIString( in_DomainAndDate.DateExpire ) ;
	cw.WriteInt32( in_DomainAndDate.SecondsExpire ) ;
	cw.WriteNonUIString( in_DomainAndDate.Domain ) ;

} 

void 
RNymCookieMgr::LoadCookie( ZKChunkReader& io_ChunkReader )
{
	ZKASSERT( io_ChunkReader.GetChunkID() == COOKIE_CHUNK_ID ) ;

	ZKSTD::string strNameValue ;
	DomainAndDate data ;

	io_ChunkReader.ReadNonUIString( strNameValue ) ;
	io_ChunkReader.ReadNonUIString( data.DateRecd ) ;
	io_ChunkReader.ReadNonUIString( data.DateExpire ) ;
	data.SecondsExpire = io_ChunkReader.ReadInt32() ;
	io_ChunkReader.ReadNonUIString( data.Domain ) ;

	long ltime;   
	time(&ltime);

	if ( data.SecondsExpire > ltime )
	{
		DumpCookie(data);
		m_NymCookieMap.insert(COOKIEJARMAP::value_type(strNameValue, data));
	}
}

void 
RNymCookieMgr::Load( ZKChunkReader& io_ChunkReader )
{
	m_NymCookieMap.clear() ;

	// We demand that our chunk be sent to us

	ZKASSERT( io_ChunkReader.GetChunkID() == COOKIEJAR_CHUNK_ID ) ;

	// Process and nested wallet chunks that we find
	while ( !io_ChunkReader.IsEndOfChunk() )
	{
		ZKChunkReader cr( io_ChunkReader.GetReadBuffer() ) ;

		switch( cr.GetChunkID() )
		{
			case( COOKIE_CHUNK_ID ) :
				LoadCookie( cr ) ;
				break ;
			default:
				ZKASSERT( false ) ;
				break ;			
		}
	}
}




bool 
RNymCookieMgr::SetNymCookieMap(
	COOKIEJARMAP& in_CookieMap)
	// Replace any Map we currently have for Nym with a new map
{
	bool ReturnValue = true;
	
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);	

	m_NymCookieMap.clear();
					
	COOKIEJARMAP::iterator theIterator;
	for(theIterator = in_CookieMap.begin(); theIterator != in_CookieMap.end();
		++theIterator)
	{
		m_NymCookieMap.insert(COOKIEJARMAP::value_type(
			(*theIterator).first, (*theIterator).second));
	}

	// We could save at this point, to make sure the change gets written to disk

	return ReturnValue;
}


bool 
RNymCookieMgr::GetNymCookieMap(
	COOKIEJARMAP& out_CookieMap)
	// Based on the strNym param, fill in the CookieMap
{
	bool ReturnValue = true;

	ZKS::Lock_Guard<WZkMutex> gate(m_cs);	

	out_CookieMap.clear();
					
	COOKIEJARMAP::iterator theIterator;
	for(theIterator = m_NymCookieMap.begin(); theIterator != m_NymCookieMap.end();
		++theIterator)
	{
		out_CookieMap.insert(COOKIEJARMAP::value_type(
			(*theIterator).first, (*theIterator).second));
	}

	return ReturnValue;
}


bool
RNymCookieMgr::IsSessionCookie( const ZKSTD::string & in_ExpiryString )
{
	return ( 0 == in_ExpiryString.compare( SESSION_COOKIE ) ) ;
}

void
RNymCookieMgr::AddCookie( 
	const ZKSTD::string& in_cookie, 
	const ZKSTD::string& in_domain )
{

	ZKTRACE2("RNymCookieMgr::AddCookie() got cookie [%s] domain [%s]\n", in_cookie.c_str(), in_domain.c_str());

	ZKSTD::string namevaluefield;
	ZKSTD::string domain;
	ZKSTD::string expires;

	if ( ParseCookie( in_cookie, in_domain, namevaluefield, domain, expires ) ) 
	{
		ZKASSERT( !namevaluefield.empty() ) ;
		ZKASSERT( !domain.empty() ) ;
		ZKASSERT( !expires.empty() ) ;

		// We don't assert since for tests we purposely send pre-expired cookies
		SetNymCookieMapEntry( namevaluefield, domain, expires );
	}
	else 
	{
		// As of 07/27/00, we actually reject some cookies (which means ParseCookie sometimes returns false)
		//		Therefore, let's not assert here; if ParseCookie returns false it's not fatal.
		//		ZKASSERT( false ) ;
	}
}

bool
RNymCookieMgr::ParseCookie(
					const ZKSTD::string& in_cookie, 
					const ZKSTD::string& in_domain,
					ZKSTD::string& out_namevalue, 
					ZKSTD::string& out_domain,
					ZKSTD::string& out_expires)
{
	bool ReturnValue = true;

	char cookieBuf[4096] = { 0 };
	char tempBuf[4096] = { 0 };
	int i = 0;
	int j = 0;

	out_namevalue.clear();
	out_domain.clear();
	out_expires.clear();

	if(in_cookie.length() >= sizeof(cookieBuf))
	{
		ZKASSERT(false);
		ReturnValue = false;
	}
	else
	{
		memcpy(cookieBuf, in_cookie.c_str(), in_cookie.length());

		for(i = 0; i < sizeof(cookieBuf); ++i)
		{ // skip whitespace
			if(cookieBuf[i] != ' ')
			{
				break;
			}
		}

		// The first field must be [Name, '=', Value]
		// followed by zero or more [';', Attribute]
		// The following code finds the first ';', or
		// the end of the cookie if there is none.
		for(j = i; j < sizeof(cookieBuf); ++j)
		{
			if(cookieBuf[j] == ';')
			{
				break;
			}
		}

		strncpy(tempBuf, &(cookieBuf[i]), j - i);
		tempBuf[j - i] = '\0';

		out_namevalue = tempBuf;
		
		// if we stopped on a semicolon skip it
		// before the next search
		i = j + 1;
		
		while(i < sizeof(cookieBuf))
		{
			// Read through all the attributes
			for(; i < sizeof(cookieBuf); ++i)
			{
				if(cookieBuf[i] != ' ')
				{
					break;
				}
			}
			
			// We found actual text... or the end of the string
			// Parse up to the next delimeter ('=' or ';')
			// to find the attribute identifier
			for(j = i; j < sizeof(cookieBuf); ++j)
			{
				if(cookieBuf[j] == '=' ||
					cookieBuf[j] == ';')
				{
					break;
				}
			}
			strncpy(tempBuf, &(cookieBuf[i]), j - i);
			tempBuf[j - i] = '\0';

			i = j;

			// _strnicmp performs case-insensitive compare of two strings
			if(!strnicmp(tempBuf, "comment", 7))
			{
				
			}
			else if(!strnicmp(tempBuf, "domain", 6))
			{
				++i;
				
				// Now parse to the end of the value
				for(j = i; j < sizeof(cookieBuf); ++j)
				{
					if(cookieBuf[j] == ';')
					{
						break;
					}
				}
				strncpy(tempBuf, &(cookieBuf[i]), j - i);
				tempBuf[j - i] = '\0';

				out_domain = tempBuf;

				// Make sure domain field matches actual domain the cookie is coming from (fixes bug 4742)
				if( !DomainMatch( in_domain, tempBuf ) ) {
					ReturnValue = false;
					ZKTRACE1("RNymCookieMgr::ParseCookie REJECTED in_cookie [%s]\n", in_cookie.c_str() );
					break;
				}

			}
			else if(!strnicmp(tempBuf, "max-age", 7))
			{

			}
			else if(!strnicmp(tempBuf, "expires", 7))
			{
				++i;

				// Now parse to the end of the value
				for(j = i; j < sizeof(cookieBuf); ++j)
				{
					if(cookieBuf[j] == ';')
					{
						break;
					}
				}
				strncpy(tempBuf, &(cookieBuf[i]), j - i);
				tempBuf[j - i] = '\0';
				out_expires = tempBuf;
			}
			else if(!strnicmp(tempBuf, "path", 4))
			{

			}
			else if(!strnicmp(tempBuf, "secure", 6))
			{

			}
			else if(!strnicmp(tempBuf, "version", 7))
			{

			}
			// We might have stopped on an '=', (ie the path attribute)
			// Skip to the next ';'. (Does nothing if we already stopped
			// on a ';'.
			for(; i < sizeof(cookieBuf); ++i)
			{
				if(cookieBuf[i] == ';')
				{
					break;
				}
			}
			// skip the ';'
			++i;
		}

		if(ReturnValue)
		{
			if(out_expires.empty())
			{ // Default 'expires' value is "Session Cookie"
				out_expires = SESSION_COOKIE ;
			}
			if(out_domain.empty())
			{
				out_domain = in_domain;
			}
		}
	}

	if( ReturnValue )
	{
		ZKASSERT(!out_namevalue.empty());
		ZKASSERT(!out_domain.empty());
		ZKASSERT(!out_expires.empty());

		ZKTRACE2("RNymCookieMgr::ParseCookie got in_cookie [%s] out_namevalue is now [%s]\n", in_cookie.c_str(), out_namevalue.c_str());
		ZKTRACE2("RNymCookieMgr::ParseCookie got in_domain [%s] out_domain is now [%s]\n", in_domain.c_str(), out_domain.c_str());
		ZKTRACE1("RNymCookieMgr::ParseCookie got expires is [%s]\n", out_expires.c_str());
	}

	return ReturnValue;
}

void 
RNymCookieMgr::DumpCookie(const DomainAndDate& rhs)
{
	ZKTRACE1("Inserting cookie from [%s]\n", rhs.Domain.c_str());
	ZKTRACE3("Date Received [%s] Date Expires [%s] LifeTime[%ld]\n",
		rhs.DateRecd.c_str(), rhs.DateExpire.c_str(), rhs.SecondsExpire);
}

void 
RNymCookieMgr::DumpCookie2(COOKIEJARMAP::iterator& rhs)
{
	ZKTRACE1("Removing cookie from [%s]\n", (*rhs).second.Domain.c_str());
	ZKTRACE3("Date Received [%s] Date Expires [%s] LifeTime[%ld]\n",
		(*rhs).second.DateRecd.c_str(), (*rhs).second.DateExpire.c_str(), (*rhs).second.SecondsExpire);
}

void 
RNymCookieMgr::GetV1XCookieFilePath(
	ZKSTD::string& out_strNymFileName )
	// Locate the cookie file corresponding to the current nym,	so that it can be opened
	// File might not exist yet, this is not an error!
{

#ifdef WIN32

	// Only Windows version ever had old cookie format - this does not need to 

	// be ported


	ZKTRACE0("RNymCookieMgr::GetNymCookieFilePath\n");

	ZKSTD::string strNymName ;
	ZKVERIFY( m_pParent->GetNymName( strNymName ) );

	ZKASSERT(!strNymName.empty());
		
	// Determine the exe's directory - we put the cookies in the
	// same directory

	// TODO: There is a helper function to take care of getting the directory....

	char szFullPath[MAX_PATH];
	char szDir[_MAX_DIR];	  // "\xx\xx\"
	char szDrive[_MAX_DRIVE]; // "C:"

	::GetModuleFileName(NULL, szFullPath, MAX_PATH);

	_splitpath(szFullPath, szDrive, szDir, NULL, NULL);

	unsigned char HashedNymName[ FRCRYPT_HASH_SIZE ];
	memset(HashedNymName, 0, sizeof(HashedNymName));

	if (!frCryptHashGenerate(
			&m_ZkError, 
			(void*)strNymName.c_str(), 
			strNymName.size(), 
			HashedNymName, 
			FRCRYPT_HASH_SIZE)) 
	{
		ZKASSERT( ZK_FALSE ) ;
		// Yikes this failed.  We'll end up using a RANDOM cookie file name
		// so the user will effectively lose information
	}

	char strFileName[ FRCRYPT_HASH_SIZE * 2 + 1 ] ;
	memset(strFileName, 0, sizeof(strFileName));

	// Map each Byte of the Hash result into 2 Character Hex representation which
	// is safe as a file name
	for( int i = 0 ; i < FRCRYPT_HASH_SIZE ; i++ ) 
	{
		snprintf( &(strFileName[ 2 * i ]), (sizeof(strFileName)-(2*i)),
                          "%2x", HashedNymName[ i ] );
	}

	out_strNymFileName = szDrive ;
	out_strNymFileName += szDir ;
	out_strNymFileName += strFileName ;
	out_strNymFileName += OLD_COOKIE_FILE_EXT ;

	ZKTRACE1("RNymCookieMgr::GetNymCookieFilePath(): NymFileName = %s\n", out_strNymFileName.c_str());

#else
    (void)out_strNymFileName;
	ZKASSERT( false ) ;

#endif
}

void 
RNymCookieMgr::LoadOldV1XFormat()
{

#ifdef WIN32

	// Only Windows version ever had old cookie format - this does not need to 

	// be ported
	ZKSTD::string strFileName ;
	GetV1XCookieFilePath( strFileName ) ;

	if ( !strFileName.empty() )
	{
		ZKVERIFY( LoadNymCookieMapFromV1XFileFormat( strFileName, m_NymCookieMap ) ) ;
	}

#endif
}

bool 
RNymCookieMgr::LoadNymCookieMapFromV1XFileFormat(
	const ZKSTD::string& in_strNymFileName, 
	COOKIEJARMAP& CookieMap)
	// Fill in m_NymCookieMap with the cookie data from storage
{
#ifdef WIN32

	// Only Windows version ever had old cookie format - this does not need to 

	// be ported


	ZKASSERT(!in_strNymFileName.empty());
	ZKASSERT(CookieMap.empty());

	////////////////////////////////////////////////////
	// If no cookie file then we return empty map

	if ( GetFileAttributes( in_strNymFileName.c_str() ) == (unsigned long)-1 ) 
	{
		return true ;
	}

	////////////////////////////////////////////////////
	// Decrypt the buffer
	bool ReturnValue = true;
	void* pClearBuf = NULL;
	size_t szClear = 0;
	ReturnValue =  ZkService::GetICrypto()->ZkDecryptFile(in_strNymFileName.c_str(), pClearBuf, szClear);

	/////////////////////////////////////////////////////
	// Get each line in pClearBuf into the Cookiemap
	if(ReturnValue)
	{
		ZKASSERT(szClear >= 0);
		ZKASSERT(ZKIsValidAddress(pClearBuf, szClear, TRUE));
		
		ZKSTD::string parser((char*)pClearBuf);
		ZKASSERT(ZKIsValidString(parser.c_str(), -1));
		long ltime;   
		time(&ltime);
		DomainAndDate entry;
				
		ZKSTD::string::size_type posStart = 0;
		ZKSTD::string::size_type posEnd = 0;

		while(true)
		{
			bool chk = true;
			posEnd = parser.find_first_of('\n', posStart);
			if(posEnd == parser.npos)
			{
				break;
			}
			else
			{
				ZKSTD::string::size_type pos1 = parser.find_first_of(';', posStart);

				if(pos1 == parser.npos)
				{
					ZKTRACE1("RNymCookieMgr::LoadNymCookieMapFromV1XFileFormat(): %s could not be properly parsed, throwing out\n", parser.substr(0,400).c_str());
					chk = false;
					ZKASSERT(false);
					break;
				}

				ZKSTD::string::size_type pos2 = parser.find_first_of(';', pos1+1);
				if(pos2 == parser.npos)
				{
					ZKTRACE1("RNymCookieMgr::LoadNymCookieMapFromV1XFileFormat(): %s could not be properly parsed, throwing out\n", parser.substr(0,400).c_str());
					chk = false;
					ZKASSERT(false);
					break;
				}
				
				ZKSTD::string::size_type pos3 = parser.find_first_of(';', pos2+1);
				if(pos3 == parser.npos)
				{
					ZKTRACE1("RNymCookieMgr::LoadNymCookieMapFromV1XFileFormat(): %s could not be properly parsed, throwing out\n", parser.substr(0,400).c_str());
					chk = false;
					ZKASSERT(false);
					break;
				}

				ZKSTD::string::size_type pos4 = parser.find_first_of(';', pos3+1);
				if(pos4 == parser.npos)
				{
					ZKTRACE1("RNymCookieMgr::LoadNymCookieMapFromV1XFileFormat(): %s could not be properly parsed, throwing out\n", parser.substr(0,400).c_str());
					chk = false;
					ZKASSERT(false);
					break;
				}

				if(chk)
				{
					ZKSTD::string strNameValue(parser.substr(posStart, pos1-posStart));
					entry.DateRecd = parser.substr(pos1+1, pos2-pos1-1);
					entry.DateExpire = parser.substr(pos2+1, pos3-pos2-1);
					entry.SecondsExpire = atol( (parser.substr(pos3+1, pos4-pos3-1) ).c_str() ); 
					entry.Domain = parser.substr(pos4+1, posEnd - (pos4+2));

					if ( entry.SecondsExpire > ltime )
					{
						DumpCookie(entry);
						CookieMap.insert(COOKIEJARMAP::value_type(strNameValue, entry));
					}
				}
			posStart = posEnd+1;
			}
		}
	}
	/////////////////////////////////////////////////////
	// Free the mem allocated to *ppClearBuf in DecryptCipherBuf()

	SAFEFREE(pClearBuf);

	return ReturnValue;

#else
    (void)in_strNymFileName;
    (void)CookieMap;
	ZKASSERT( false ) ;

	return false ;

#endif
}


void 
RNymCookieMgr::SelfTest()
{
#ifdef INTERNAL_BUILD
	
	//   * both host names are IP addresses and their host name strings match
	//     exactly; or

	const ZKSTD::string IP1( "100.100.100.100" ) ;
	const ZKSTD::string IP2( "99.99.99.99" ) ;

	ZKTEST ( DomainMatch( IP1, IP1 ) );
	ZKTEST ( !DomainMatch( IP1, IP2 ) ) ;

	//   * both host names are FQDN strings and their host name strings match
	//     exactly; or

	const ZKSTD::string FQDN1( "x.y.com" ) ;

	ZKTEST ( DomainMatch( FQDN1, FQDN1 ) ) ;

	//   * A is a FQDN string and has the form NB, where N is a non-empty name
	//     string, B has the form .B', and B' is a FQDN string.  (So, x.y.com
	//     domain-matches .y.com but not y.com.)

	const ZKSTD::string FQDN2( ".y.com" ) ;
	const ZKSTD::string FQDN3( "y.com" ) ;

	ZKTEST ( DomainMatch( FQDN1, FQDN2 ) ) ;
	ZKTEST ( !DomainMatch( FQDN1, FQDN3 ) ) ;

	
	const ZKSTD::string strExpiry1 = "Wednesday, 09-Nov-2003 23:12:40 GMT" ;
	const ZKSTD::string TestDomain( "Free.Nelson.Mandela.com" ) ; 

	// HUH: Pre-expired cookies aren't really an error condition anymore, but I've left 
	// SetNymCookieMapEntry as returning false when it gets one.  This may need some
	// tweeking...   by: georgesh

	ZKTEST( SetNymCookieMapEntry( 
					ZKSTD::string( "Item=Another Test Cookie Content" ),
					TestDomain,
					strExpiry1 ) );

	// Try another Date format we claim to support
	const ZKSTD::string strAnExpiry = "Mon, 18 Jun 2007 04:20:21 GMT" ;

	ZKTEST( SetNymCookieMapEntry( 
					ZKSTD::string( "Item=More cookies than you can eat in one sitting" ),
					TestDomain,
					strAnExpiry ) );

	// And how about a session cookie
	ZKTEST( SetNymCookieMapEntry( 
					ZKSTD::string( "Item=This is a session cookie" ),
					TestDomain,
					SESSION_COOKIE ) );

#endif
}
