// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RNymMail.cpp: implementation of the RNymMail class.
//
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "RNym.h"
#include "RNymMail.h"
#include "ICoreServerConnect.h"
#include <frNym.h>
#include <frNymClient.h>
#include <frNymUtil.h>
#include <frSec.h>
#include <zkMemBuf.h>
#include <IZkConsoleMsg.h>
#include <IClientPersist.h>
#include <IClientCentral.h>
#include <IKeyQry.h>
#include <ICrypto.h>
#include <INiqs.h>
#include <IClientInfo.h>
#include <ZkService.h>
#include <WMaipRb.h>
#include <string_hlp.h>
#include <RZkException.h>
#include <ZKChunk.h>
#include <zkBase64.h>
#include <frMailAuth.h>
#include <EsmtpClient.h>
#include <frMailBlockList_client.h>
#include <ZkErrorHandler.h>
#include <IZKErrorReporter.h>
#include <ZKBusyAVI.h>
#include <WZkipClient.h>

static const size_t SMTP_BUF_SZ = 1024;
static const size_t ERROR_BUF_SZ = 1024;
static const std::string HELO("HELO");
static const std::string MAIL_FROM("MAIL FROM:");
static const std::string RCPT_TO("RCPT TO:");
static const std::string DATA("DATA");
static const std::string QUIT("QUIT");

#define AUTOREPLYBLOCKNAME "auto-created"

RNymMail::RNymMail(RNym * pParent, WRemoteNym& RemoteNym )
	: m_pParent(pParent)
	, m_RemoteNym(RemoteNym)
	, m_bBlockSpam(false)
	, m_NextMailAuthVersion(1)
	, m_MacKey(NULL)
	, m_bNymHadMailChunk(false)
	, m_bTagline(true)
        , m_nCurrentMailSentCount( 0 )
        , mailSentInitTime( 0 )
#ifdef WIN32
	, m_ReplyBlockCacheLoaded(false)
#endif
{
	ZKVERIFY(frCryptMacKeyNew(&m_ZkError, &m_MacKey));
}


RNymMail::~RNymMail()
{
	if(m_MacKey)
	{
		ZKVERIFY(frCryptMacKeyFree(&m_ZkError, &m_MacKey));
	}
}

bool RNymMail::UpdateMailCertificate()
{
	if(!m_pParent->DoesNymUseMail())
	{
		return false;
	}
	ZKSTD::string strNymName;
	m_pParent->GetNymName(strNymName);
	IZkConsoleMsg::ShowMsgForceUpdate(CONSIMAGE_MAIL, IDS_CONS_UPDATING_MAIL_CERT, strNymName.c_str());
	try
	{
		RetrieveNymMailCert();
	}
	catch(RZkException ex)
	{
		IZkConsoleMsg::ShowImportantMsg(CONSIMAGE_MAIL, IDS_CONS_JUST_STRING, ex.GetFormattedString().c_str());
		return false;
	}

	return true;
}

void RNymMail::GetMailCertificate(ZKSTD::string& out_strCert)
{	
	out_strCert = m_strMailCertificate;
}

FrCryptMacKey	RNymMail::GetMailMacKey()
{
	return m_MacKey;
}

void
RNymMail::SendControlMessageToServer(const std::string& in_strRecipient, const char* in_strMsgBody)
{
	ZKASSERT(m_pParent->DoesNymUseMail());

	FrCryptPubKey PrivSigningKey = NULL;
	INym * pNym = NULL ;
	INymMail *pNymMail = NULL;

	ZKSTD::string strSenderAddress;
	ZKSTD::string strNymFreedomDomain;

	try
	{
		m_pParent->GetSignatureKey( &PrivSigningKey );

		m_pParent->GetNymEmailDomain(strNymFreedomDomain);
		m_pParent->GetNymName(strSenderAddress);
		
		strSenderAddress += '@' + strNymFreedomDomain;

		ZKSTD::string strAuthSecretCert;
		GetMailCertificate(strAuthSecretCert);

		if(!strAuthSecretCert.length())
		{
			if(UpdateMailCertificate())
			{
				GetMailCertificate(strAuthSecretCert);
			}
		}

		if(!strAuthSecretCert.length())
		{
			ZKASSERT(0);
			throw DMailAuthException(
				"RNymMail::SendControlMessageToServer",
				strSenderAddress.substr(0, strSenderAddress.find('@')).c_str());
		}
		else
		{
			ZkSocketAddress RemoteAddr;
			SOCKADDR_IN sa;

			ZkService::GetINiqs()->GetNmtaAddr(&sa);
			memcpy( &RemoteAddr, &sa, sizeof (SOCKADDR_IN) );
			
			ZKSTD::string strFreedomDomain; // May be different from the nym freedom domain

			// The freedom domain is used to determine which recipients
			// are "nyms" and which are not.  This technique will not
			// work if there are multiple domains, except for the limited
			// case where all nym domains are build on top of the base domain,
			// eg trial.freedom.net, freedom.net
			ZkService::GetIClientInfo()->GetFreedomDomain(strFreedomDomain);

			EsmtpClient Mailer(strFreedomDomain, strSenderAddress, strAuthSecretCert,
								RemoteAddr,PrivSigningKey);

			Mailer.AddRecipient(in_strRecipient);

			Mailer.SendMsgBody(in_strMsgBody, strlen(in_strMsgBody));
		}
	}
	catch (...)
	{		
		if (PrivSigningKey) 
		{
			frCryptPubKeyFree(NULL,&PrivSigningKey);
		}
		throw;
	}
			
	if (PrivSigningKey) 
	{
		frCryptPubKeyFree(NULL,&PrivSigningKey);
	}
}

void RNymMail::SendEmailBlockListToServer(const std::vector<std::string>& vecAddresses)
{
	std::string strNymName;
	std::string strDomain;

	m_pParent->GetNymName(strNymName);
	m_pParent->GetNymEmailDomain(strDomain);
	std::string strMsg = "From: " + strNymName + '@' + strDomain + "\r\n\r\n";

	char buffer[1024] = {0};

	for(size_t i = 0; i < vecAddresses.size(); i++)
	{
		size_t lenEntry = 0;
		ZKVERIFY(frMailBlockList_CreateSenderEntry(&m_ZkError,
												vecAddresses[i].c_str(),
												buffer,
												1024,
												&lenEntry));
		ZKSTD::string strEntry(buffer, lenEntry);

		strMsg += strEntry + " ";
	}

	// We send the email to the same domain as the nyms email
	// domain.  For example if the nym is a trial.freedom.net
	// we send block list to config-blocklist@trial.freedom.net

	ZKSTD::string strRecip = "config-blocklist@" + strDomain;
	SendControlMessageToServer(strRecip, strMsg.c_str());
}

void RNymMail::SendSpamPreferenceToServer(bool in_bBlockSpam)
{
	std::string strNymName;
	std::string strDomain;

	m_pParent->GetNymName(strNymName);
	m_pParent->GetNymEmailDomain(strDomain);
	std::string strMsg =	"From: " + strNymName + '@' + strDomain + "\r\n\r\n";
	if(in_bBlockSpam)
	{
		strMsg += "On\r\n";
	}
	else
	{
		strMsg += "Off\r\n";
	}


	ZKSTD::string strRecip = "config-spam-control@" + strDomain;
	SendControlMessageToServer(strRecip, strMsg.c_str());
}

void RNymMail::SendTaglinePreferenceToServer(bool in_bTagline)
{
	std::string strNymName;
	std::string strDomain;

	m_pParent->GetNymName(strNymName);
	m_pParent->GetNymEmailDomain(strDomain);
	std::string strMsg =	"From: " + strNymName + '@' + strDomain + "\r\n\r\n";
	if(in_bTagline)
	{
		strMsg += "On\r\n";
	}
	else
	{
		strMsg += "Off\r\n";
	}


	ZKSTD::string strRecip = "config-tagline@" + strDomain;
	SendControlMessageToServer(strRecip, strMsg.c_str());
}

void 
RNymMail::TransitionIfNecessary()
{
	if(m_pParent->DoesNymUseMail() && !m_bNymHadMailChunk)
	{
		// we got here because we were loading a 1.1 nym file.
		// so we have to ask for this stuff in a 1.1 way.
	    unsigned long dwSpam = 0;
	    OldGetEmailAddressBlockList(m_vecBlockList);
	    OldGetSpamControls( dwSpam );
	    m_bBlockSpam = (dwSpam != 0);
		
		try
		{
			ZkService::GetIClientPersist()->SaveClientState();
		}
		catch (RZkException ex)
		{
			ZkErrorHandler::GetIZkErrorReporter()->ShowError(ex);
		}
	}	
}

void 
RNymMail::SetEmailAddressBlockList(
	const std::vector<std::string>& vecAddresses)
{
	SendEmailBlockListToServer(vecAddresses);
	m_vecBlockList = vecAddresses;

	ZkService::GetIClientPersist()->SetDirty();
}

void
RNymMail::SetSpamPreference(bool in_bBlockSpam)
{
	SendSpamPreferenceToServer(in_bBlockSpam);
	m_bBlockSpam = in_bBlockSpam;

	ZkService::GetIClientPersist()->SetDirty();
}

bool
RNymMail::NymHasTaglineChoice()
{
	return true;
}

void
RNymMail::SetTaglinePreference(bool in_bTagline)
{
	SendTaglinePreferenceToServer(in_bTagline);
	m_bTagline = in_bTagline;

	ZkService::GetIClientPersist()->SetDirty();
}

void 
RNymMail::GetEmailAddressBlockList(
	std::vector<std::string>& vecstremailaddr)
{
	vecstremailaddr = m_vecBlockList;
}

void 
RNymMail::OldGetEmailAddressBlockList( std::vector<std::string>& vecstremailaddr )
{
	// Initialize output
	vecstremailaddr.clear();

	ZkList BlockList      = NULL;

	frRemoteNymGetStringListParameter(
					&m_ZkError ,              
					RemoteNym() ,
					NymBlockList,					
					&BlockList );

	if(BlockList != NULL) 
	{
		// We have some items in the Email Block List

		size_t listsize = 0;

		ZKVERIFY(zkListSize (
			&m_ZkError ,              
			BlockList ,
			&listsize)); 

		ZKVERIFY(zkListMoveCursor (
			&m_ZkError ,               
			BlockList ,
			LIST_HEAD));

		void* emailaddr = NULL;

		for(size_t i = 0; i < listsize; ++i)
		{
			
			ZKVERIFY(zkListGet (
				&m_ZkError ,               
				BlockList ,
				&emailaddr ,               
				LIST_FORWARDS ));  

			vecstremailaddr.push_back((char*)emailaddr);
		}

		ZKVERIFY(zkListFree(&m_ZkError, &BlockList));
	}
}

void RNymMail::OldGetSpamControls( unsigned long & out_dwSpamControlFlags )
{
	uint32_t uiCtlFlags = 0 ;	
	frRemoteNymGetSpamControls ( &m_ZkError, RemoteNym(), &uiCtlFlags ) ;
	out_dwSpamControlFlags = uiCtlFlags ;
}

void 
RNymMail::TestAndThrowServerResponse(int bStatus)
{
	if ( bStatus ) 
	{
		// No failure, nothing to do
		return ;
	}

	if ( !m_strResponse.empty() ) 
	{
		char buf[NSRMESG_SZ + 16]; // msg + code
		snprintf(buf, sizeof(buf), "%s", m_strResponse.c_str());
		throw RZkException( m_ZkError, "ZKFreedomMailSubSy::ThrowServerResponse",buf, IDS_SERVER_ERROR, buf ) ;
	}
	else 
	{
		ZKASSERT( false ) ;
		throw RZkException(m_ZkError, "RNymMail::ThrowServerResponse", 
					"No Server Response Message Available", 
					IDS_SERVER_ERROR_UNKNOWN);
	}
}


bool RNymMail::RetrieveNymMailCert()  throw(RZkException)
{
	ZKBusyAVI BusyAVI;

	if(!m_pParent->DoesNymUseMail())
	{
		ZKASSERT(0);
		return false;
	}

	bool ReturnValue = false;


	NymServerResponse	theNymServerResponse;
	memset(&theNymServerResponse, 0, sizeof(theNymServerResponse));

	FrRemoteNym StrippedCopyClientSideNym = NULL;

	m_RemoteNym.DupFrRemoteNymWithSignatureKey(
						RemoteNym(), 
						&StrippedCopyClientSideNym);

	FrMailAuth theCert = NULL ;
	FrCryptMacKey theMacKey = NULL ;

	void *certBuffer = NULL;
	size_t certSize = 0;
	ZKSTD::string strNymName;
	m_pParent->GetNymName(strNymName);

	WAutoBuffer EncodedCert ;

	try
	{
		WZkipClient client;
		// SmartConnectThroughCloud throws if the connect fails.
		ZkService::GetICoreServerConnect()->SmartConnectThroughCloud(
						FRENT_TYPE_NYM_SRV, client, NYM_SERVER_TIMEOUT);

		if(!frRemoteNymMailCertConnected(&m_ZkError, StrippedCopyClientSideNym, client, NULL,
							m_NextMailAuthVersion++, &theNymServerResponse, &theCert, &theMacKey))
		{
			if(theNymServerResponse.message && strlen(theNymServerResponse.message))
			{
				throw RZkException( 
						m_ZkError, 
						"RNymMail::RetrieveNymMailCert()",
						"frRemoteNymMailCertConnected failed", 
						IDS_FAILED_TO_GET_MAIL_CERTIFICATE,
						strNymName.c_str(),
						theNymServerResponse.message
						);
			}
			else
			{
				throw RZkException( 
						m_ZkError, 
						"RNymMail::RetrieveNymMailCertConnected()",
						"frRemoteNymMailCertConnected failed", 
						IDS_FAILED_TO_GET_MAIL_CERTIFICATE,
						strNymName.c_str(),
						m_ZkError.getdetail()
						);
			}
		}

		if(!frMailAuthToBuf(&m_ZkError, theCert, &certBuffer, &certSize))
		{
			ZKASSERT(0);
			
			throw RZkException( 
					m_ZkError, 
					"RNymMail::RetrieveNymMailCert()",
					"frMailAuthToBuf failed", 
					IDS_FAILED_TO_CREATE_MAIL_ACCOUNT,
					strNymName.c_str());
		}

		EncodedCert.Allocate( 2 * certSize ) ;

		if(!zkConvertToBase64(&m_ZkError, (const unsigned char*)certBuffer, EncodedCert, certSize))
		{
			ZKASSERT(0);
			throw RZkException( 
					m_ZkError, 
					"RNymMail::RetrieveNymMailCert()",
					"zkConvertToBase64 failed", 
					IDS_FAILED_TO_CREATE_MAIL_ACCOUNT,
					strNymName.c_str());
		}

		m_strMailCertificate = (char*)EncodedCert.GetBuffer();

		if(m_MacKey)
		{
			ZKVERIFY(frCryptMacKeyFree(&m_ZkError, &m_MacKey));
			m_MacKey = theMacKey;
			theMacKey = NULL ;
		}
		ReturnValue = true;
	}
	catch(RZkException)
	{
		
		if(certBuffer)
		{
			zkMemFree(&m_ZkError, &certBuffer);
		}

		if(theMacKey)
		{
			ZKVERIFY(frCryptMacKeyFree(&m_ZkError, &theMacKey));
		}

		if(theCert)
		{
			ZKVERIFY(frMailAuthFree(&m_ZkError, &theCert));
		}

		throw;
	}
	
	if(certBuffer)
	{
		zkMemFree(&m_ZkError, &certBuffer);
	}

	if(theMacKey)
	{
		ZKVERIFY(frCryptMacKeyFree(&m_ZkError, &theMacKey));
	}

	if(theCert)
	{
		ZKVERIFY(frMailAuthFree(&m_ZkError, &theCert));
	}

	return ReturnValue;
}

T_FrRemoteNym*  
RNymMail::RemoteNym() 
{ 
	return m_RemoteNym.GetFrRemoteNym();
}

// Linux doesn't need backward compatibility
#ifdef WIN32
void
RNymMail::EnsureCacheLoaded()
{
 	if ( m_ReplyBlockCacheLoaded ) 
 	{
 		return ;
 	}
 
 	LoadReplyBlocks() ;
 
 	ZKASSERT( m_ReplyBlockCacheLoaded ) ;
}
 
bool 
RNymMail::LoadReplyBlocks()
{
	bool ReturnValue = true;
	
	// Empty out the cache
	m_ReplyBlocks.clear() ;
	
	// Traverse the remote nym adding the reply blocks
	ZKVERIFY(frRemoteNymResetParameterSearch (
				&m_ZkError ,                 
				RemoteNym() ));
	
	while(1)
	{
		const unsigned char* buf = NULL;
		size_t szbuf = 0;

		// NOTE: we never read the NymReplyBlock parameter that we added,
		// we only read the NymPrivateReplyBlock.  This is because we can't
		// safely assume that the order we read would read back these two sets of
		// values would be correctly ordered.  For this reason we save all the
		// necessary information (including the public data) in the 
		// NymPrivateReplyBlock member.

		if(!(frRemoteNymGetParameter(
			&m_ZkError ,
			RemoteNym() ,
			NymPrivateReplyBlock ,
			&buf ,
			&szbuf)))
		{
			break;
		}

		try 
		{
			WMaipRb MaipRb ;
			MaipRb.Initialize( buf, szbuf ) ;

			m_ReplyBlocks.push_back( MaipRb ) ;
		}
		catch(std::exception&)
		{	
			// Possibly a corrupt reply block
			// Perhaps there are valid ones so keep going
		}
	}

	ZKVERIFY(frRemoteNymResetParameterSearch (
		&m_ZkError ,                 
		RemoteNym() ));

	m_ReplyBlockCacheLoaded = true ;

	return ReturnValue;
}

size_t
RNymMail::GetNumberOfReplyBlocks()
{
	EnsureCacheLoaded() ;
	return m_ReplyBlocks.size();
}

void 
RNymMail::PeelReplyBlockFromMessage( 
	WCypherMailMsg&  io_MailMsg,
	unsigned char ** out_ppPeeledBuffer,
	size_t&          out_szPeeledBuffer  ) throw (RZkException)
{
	ZKFunctionEntry( "RNymMail::PeelReplyBlockFromMessage" );
	ZKASSERT(io_MailMsg.GetEncryptedSize() > 0 );

	EnsureCacheLoaded() ;

	int bStatus = true ;

	try 
	{
		size_t nReplyBlocks = GetNumberOfReplyBlocks();
		
		ZKASSERT( nReplyBlocks == m_ReplyBlocks.size() ) ;

		size_t j;
		for( j = 0; j < nReplyBlocks; ++j )
		{	
			bStatus = PeelReplyBlockXFromMessage( 
							j, 
							io_MailMsg.GetEncryptedText(),
							io_MailMsg.GetEncryptedSize(), 
							out_ppPeeledBuffer, 
							out_szPeeledBuffer ) ;

			if (bStatus) 
			{
				ZKTRACE1( "Message was decrypted using Reply block %d\n", j ) ;

				// This is the right reply block
				io_MailMsg.SetReplyBlock( j ) ;

				break;
			}
		}

		if ( j == nReplyBlocks ) 
		{
			// Couldn't find it
			ZKThrowRuntime("Could not decrypt with any reply block", IDS_UNEXPECTED_ERROR);
		}
	}
	catch(std::exception&)
	{
		throw;
	}
} 

bool 
RNymMail::PeelReplyBlockXFromMessage(
		size_t				   in_IndexOfRBToUse,
		const unsigned char *  in_pBuffer,
		size_t				   in_szBuffer,
		unsigned char **	   out_ppPeeledBuffer,
		size_t&				   out_szPeeledBuffer ) 
{
	ZKASSERT( out_ppPeeledBuffer != NULL ) ;
	ZKASSERT( in_pBuffer != NULL ) ;

	*out_ppPeeledBuffer = NULL ;
	out_szPeeledBuffer = 0 ;

	EnsureCacheLoaded() ;

#ifdef _DEBUG	
	size_t nReplyBlocks = GetNumberOfReplyBlocks();
	ZKASSERT( in_IndexOfRBToUse < nReplyBlocks ) ;
#endif

	FrMaipRbPriv RbPriv = NULL ;
	RbPriv = m_ReplyBlocks[ in_IndexOfRBToUse ].GetPrivateReplyBlock() ;

	ZKASSERT( RbPriv != NULL ) ;

	return ( 1 == frMaipRbPrivDecrypt(&m_ZkError, 
						RbPriv, 
						(void*)in_pBuffer, 
						in_szBuffer,
						(void**)out_ppPeeledBuffer,
						&out_szPeeledBuffer) );
}


void 
RNymMail::DecryptMessageEncryptedWithNymsPrivateKey( 
	unsigned char *  in_pBuffer,
	size_t           in_szBuffer,
	unsigned char ** out_ppClearBuffer,
	size_t&          out_szClearBuffer)  throw(RZkException)
{
	bool bSuccessful = false ;

	if (( in_pBuffer == NULL ) ||
		( in_szBuffer == 0 ) ||
		( out_ppClearBuffer == NULL ))
	{
		// holy invalid parameters batman!
		throw RZkException("RNymMail::DecryptMessageEncryptedWithNymsPrivateKey", 
							"Invalid Parameters", 
							IDS_NULL_POINTER);

	}	

	// Initialize output
	*out_ppClearBuffer = NULL ;
	out_szClearBuffer = 0 ;

	FrCryptPubKey PrivCryptoKey = NULL;
	FrCryptPubKey PrivSigningKey = NULL;

	if ( frRemoteNymGetPrivateKeys (
				&m_ZkError ,
				RemoteNym() ,
				&PrivCryptoKey ,	/* Returns the private encryption key */
				&PrivSigningKey ))   /* Returns the private signature key */
	{
		if ( frSecPubKeyDecryptData(
							&m_ZkError, 
							PrivCryptoKey,
							in_pBuffer,
							in_szBuffer,
							(void**)out_ppClearBuffer,
							&out_szClearBuffer) ) 
		{
			bSuccessful = true ;
		}
	}
	if( PrivCryptoKey )
	{
		frCryptPubKeyFree(&m_ZkError, &PrivCryptoKey);
	}

	if( PrivSigningKey )
	{
		frCryptPubKeyFree(&m_ZkError, &PrivSigningKey);
	}

	if ( !bSuccessful ) 
	{
		// TODO - create specific exception message
		throw RZkException(m_ZkError, "RNymMail::DecryptMessageEncryptedWithNymsPrivateKey", 
							"Could not decrypt buffer with nym's private key", 
							IDS_UNEXPECTED_FAIL);
	}
}

void    
RNymMail::DecryptMessage( WCypherMailMsg&  io_MailMsg )  throw(RZkException)
{
	ZKFunctionEntry( "RNymMail::DecryptMessage" );
	EnsureCacheLoaded() ;
 	
 	// Verify that we have a message to decrypt
 	if ( io_MailMsg.GetEncryptedSize() == 0 ) 
 	{
 		ZKASSERT( false ) ;
 		return; 
 	}
 
 	// Verify that this message hasn't already been decrypted
 	ZKASSERT( !io_MailMsg.IsMsgDecrypted() ) ;
 
 	// Not using WAutoBuffer because CheckAndRemoveSignature() changes the size in place
 	unsigned char * pPeeledBuffer = NULL ;
 	size_t         szPeeledBuffer = 0 ;
 
 	try
 	{
 		PeelReplyBlockFromMessage( io_MailMsg, &pPeeledBuffer, szPeeledBuffer ) ;
 
 		// If we were able to peel the message then this
 		// must have been meant for this nym
 
 		std::string strMyName ;
 		m_pParent->GetNymName( strMyName ) ;
 
 		io_MailMsg.SetRecipientNym( strMyName ) ;
 
 		ZKASSERT( pPeeledBuffer ) ;
 		ZKASSERT( szPeeledBuffer != 0 ) ;
 
 		// First signature is the FMG signature
 		std::string		strFmgEntity ;
 		FrEntIdType		FmgEntity;
 		SigCheckResult SigResult = ZkService::GetICrypto()->CheckAndRemoveSignature( pPeeledBuffer, szPeeledBuffer, strFmgEntity, FmgEntity ) ;
 
 		io_MailMsg.SetFmgSignatureCheckResult( SigResult ) ;
 
 		if ( ( SigResult != SigCheckSuccessful ) ||
 			 ( FRENT_TYPE_FMG != FmgEntity) ) 
 		{
 			// No way
 			ZKThrowRuntime("Message was not correctly signed by the Freedom Mail Gateway", IDS_FMG_SIG_CHECK_FAILED );
 		}
 
 		// Second signature is the Nym Signature for Nym to Nym mail and
 		// the FMG signature again for Internet User to Nym mail
 		std::string		strEntity ;
 		FrEntIdType		SecondEntity;
 
 		SigResult = ZkService::GetICrypto()->CheckAndRemoveSignature( pPeeledBuffer, szPeeledBuffer, strEntity, SecondEntity ) ;
 
 		// TODO - make sure we react to the signature check situation
 
 		io_MailMsg.SetSignatureCheckResult( SigResult ) ;
 		io_MailMsg.SetSendingEntity( strEntity ) ;
 
 		DecryptMessageEncryptedWithNymsPrivateKey( pPeeledBuffer,
 												   szPeeledBuffer,
 												   &io_MailMsg.GetClearAutoBuffer(),
 												   io_MailMsg.GetClearAutoBuffer().GetSizeRef() ) ; 
 	}
 	catch(std::exception& )
 	{
 		// Cleanup
 		SAFEFREE( pPeeledBuffer ) ;
 		throw ;
 	}
 
 	// Cleanup
 	SAFEFREE( pPeeledBuffer ) ;
}
#endif // WIN32

void	
RNymMail::SelfTest()
{
#ifdef INTERNAL_BUILD
	
	// Test Signature/From Field matching

	// CompareFromFieldWithSigningEntity() has been disabled
	// (see bug 3300) temporarily. Uncomment this code when
	// it is re-enabled.
/*
	WZkEmail			*DecryptedEmail ;
	ZkEmailHeader		header;
	std::string			testMsg = "From: nym1@freedom.net\r\n\r\nBody\r\n.\r\n";
	DecryptedEmail = new WZkEmail;
	zkEmailAddText(
		&m_ZkError, 
		*DecryptedEmail, 
		(const unsigned char *)testMsg.c_str(),
		testMsg.length());

	zkEmailParse(&m_ZkError, *DecryptedEmail);

	ZKVERIFY( zkEmailHeader(&m_ZkError, *DecryptedEmail, &header) );
	try
	{
		CompareFromFieldWithSigningEntity(header, "nym1");
	}
	catch(std::exception)
	{
		// Shouldn't have thrown...
		ZKTEST(0);
	}

	try
	{
		CompareFromFieldWithSigningEntity(header, "nym2");
		// Should never execute
		ZKTEST(0);
	}
	catch(std::exception)
	{// Should throw...
	}

	testMsg = "From: spambait@somewhereelse.net\r\n\r\nBody\r\n.\r\n";
	delete DecryptedEmail;
	DecryptedEmail = new WZkEmail;
	zkEmailAddText(
		&m_ZkError, 
		*DecryptedEmail, 
		(const unsigned char *)testMsg.c_str(),
		testMsg.length());

	zkEmailParse(&m_ZkError, *DecryptedEmail);

	ZKVERIFY( zkEmailHeader(&m_ZkError, *DecryptedEmail, &header) );
	try
	{
		CompareFromFieldWithSigningEntity(header, "FMG");
	}
	catch(std::exception)
	{// Shouldn't have thrown...
		ZKTEST(0);
	}

	
	testMsg = "From: spambait\r\n\r\nBody\r\n.\r\n";
	delete DecryptedEmail;
	DecryptedEmail = new WZkEmail;
	zkEmailAddText(
		&m_ZkError, 
		*DecryptedEmail, 
		(const unsigned char *)testMsg.c_str(),
		testMsg.length());

	zkEmailParse(&m_ZkError, *DecryptedEmail);

	ZKVERIFY( zkEmailHeader(&m_ZkError, *DecryptedEmail, &header) );

	try
	{
		CompareFromFieldWithSigningEntity(header, "FMG");
	}
	catch(std::exception)
	{// Shouldn't have thrown...
		ZKTEST(0);
	}

	
	testMsg = "Subject: Email without a from\r\n\r\nBody\r\n.\r\n";
	delete DecryptedEmail;
	DecryptedEmail = new WZkEmail;
	zkEmailAddText(
		&m_ZkError, 
		*DecryptedEmail, 
		(const unsigned char *)testMsg.c_str(),
		testMsg.length());

	zkEmailParse(&m_ZkError, *DecryptedEmail);

	ZKVERIFY( zkEmailHeader(&m_ZkError, *DecryptedEmail, &header) );
	try
	{
		CompareFromFieldWithSigningEntity(header, "FMG");
	}
	catch(std::exception)
	{// Shouldn't have thrown...
		ZKTEST(0);
	}
	delete DecryptedEmail;
	*/
#endif	
	
	// Try encrypting a message as it would come in and then make sure we can decrypt it
	// again

	// Unfortunately I don't know if we can test this because unfortunately I think we
	// need private keys of the maip to do the encryption


}


void 
RNymMail::Save( WMemBuf& io_Buffer ) throw(RZkException)
{
	if(!m_pParent->DoesNymUseMail())
	{
		return;
	}
	char MacKeyBuf[FRCRYPT_MACKEY_SIZE];

	ZKChunkWriter cw( io_Buffer, MAIL_CHUNK_ID, NYM_MAIL_PERSIST_VERSION) ;
	
	// Transition code was executing after setup wizard - #7671
	m_bNymHadMailChunk = true;

	uint32_t MailPrefs = 0;
	if(m_bBlockSpam)
	{
		MailPrefs = MAIL_PREF_BLOCK_SPAM;
	}
	if(!m_bTagline)
	{
		MailPrefs |= MAIL_PREF_NO_TAGLINE;
	}

	cw.WriteUInt32(MailPrefs);
	
	cw.WriteUInt16(m_NextMailAuthVersion );
	ZKVERIFY(frCryptMacKeyStore(&m_ZkError, m_MacKey, MacKeyBuf, FRCRYPT_MACKEY_SIZE));
	cw.WriteArray( MacKeyBuf, FRCRYPT_MACKEY_SIZE);
	cw.WriteNonUIString(m_strMailCertificate);
	cw.WriteInt16(m_vecBlockList.size());

	for(size_t i = 0 ; i < m_vecBlockList.size() ; i++)
	{
		cw.WriteNonUIString(m_vecBlockList[i]);
	}

    	// save the mail limit stuff
	cw.WriteUInt32( m_nCurrentMailSentCount );
    	cw.WriteUInt32((unsigned int) mailSentInitTime );
}

void 
RNymMail::Load( ZKChunkReader& io_ChunkReader ) throw(RZkException)
{
	m_bNymHadMailChunk = true;
	// Clear any state
	m_strMailCertificate.erase();
	void *MacKeyBuf = NULL;
	size_t MacKeySize = 0;

	m_vecBlockList.erase(m_vecBlockList.begin(), m_vecBlockList.end());
	
	// We demand that our chunk be sent to us

	ZKASSERT( io_ChunkReader.GetChunkID() == MAIL_CHUNK_ID ) ;

	uint32_t mailPrefs = io_ChunkReader.ReadUInt32( );
	m_bBlockSpam = ( mailPrefs & MAIL_PREF_BLOCK_SPAM);
	m_bTagline   =!( mailPrefs & MAIL_PREF_NO_TAGLINE);
	m_NextMailAuthVersion = io_ChunkReader.ReadUInt16( );
	io_ChunkReader.ReadArray( (const void**)&MacKeyBuf, &MacKeySize);
	ZKVERIFY(frCryptMacKeyLoad(&m_ZkError, m_MacKey, MacKeyBuf, FRCRYPT_MACKEY_SIZE));
	io_ChunkReader.ReadNonUIString( m_strMailCertificate ) ;
	int nBlockList = io_ChunkReader.ReadInt16();
	for(int i = 0; i < nBlockList; i++)
	{
		std::string strBlockListItem;
		io_ChunkReader.ReadNonUIString( strBlockListItem ) ;
		ZKASSERT( !strBlockListItem.empty() ) ;
		m_vecBlockList.push_back( strBlockListItem);
	}
	if(!io_ChunkReader.IsEndOfChunk())
	{// These were init'd to 0 by the constructor
		m_nCurrentMailSentCount = io_ChunkReader.ReadInt32() ;
		mailSentInitTime = (time_t)io_ChunkReader.ReadInt32() ;
	}
}

void RNymMail::CreatePopAccountOnPopServer() throw(RZkException)
{
        ZKBusyAVI BusyAVI;

	ZKSTD::string strNymName;
	ZKSTD::string strMailCert;

	GetMailCertificate(strMailCert);
	m_pParent->GetNymName(strNymName);

	ZkSocket socket;

	try
	{
		ZKVERIFY( zkSocketNew(&m_ZkError, AF_INET, SOCK_STREAM, IPPROTO_TCP, &socket) );
		ZKVERIFY( zkSocketSetOption(&m_ZkError, socket, ZKSOCKET_OPTION_SendTimeout, 15 * 1000) );
		ZKVERIFY( zkSocketSetOption(&m_ZkError, socket, ZKSOCKET_OPTION_RecvTimeout, 45 * 1000) );
		
		ZKSTD::string strUserCommand = "user " + strNymName + "\r\n";
		ZKSTD::string strPassCommand = "pass " + strMailCert + "\r\n";

		SOCKADDR_IN PopAddr;
		ZkSocketAddress RemoteAddr;
		
		// May throw..
		ZkService::GetINiqs()->GetPopServerAddr(&PopAddr);
		memcpy( &RemoteAddr, &PopAddr, sizeof (SOCKADDR_IN) );
		

		if( !zkSocketConnectAddr(&m_ZkError, socket, RemoteAddr) )
		{
			throw RZkException( 
					m_ZkError, 
					"RNymMail::CreatePopAccountOnPopServer()",
					"zkSocketConnect failed", 
					IDS_FAILED_TO_CREATE_MAIL_ACCOUNT,
					strNymName.c_str());
		}
		else if( !ReceiveAndTestPOPResponse(socket) )
		{
			throw RZkException( 
					m_ZkError, 
					"RNymMail::CreatePopAccountOnPopServer()",
					"Failed to receive POP welcome banner", 
					IDS_FAILED_TO_CREATE_MAIL_ACCOUNT,
					strNymName.c_str());
		}
		else if(!zkSocketWriteAll(&m_ZkError, socket, strUserCommand.c_str(),
									strUserCommand.length()))
		{
			ZKASSERT(false);
			throw RZkException( 
					m_ZkError, 
					"RNymMail::CreatePopAccountOnPopServer()",
					"Failed to write 'USER' command", 
					IDS_FAILED_TO_CREATE_MAIL_ACCOUNT,
					strNymName.c_str());
		}
		else if( !ReceiveAndTestPOPResponse(socket) )
		{
			throw RZkException( 
					m_ZkError, 
					"RNymMail::CreatePopAccountOnPopServer()",
					"Invalid or no response to 'USER' command", 
					IDS_FAILED_TO_CREATE_MAIL_ACCOUNT,
					strNymName.c_str());
		}
		else if(!zkSocketWriteAll(&m_ZkError, socket, strPassCommand.c_str(),
									strPassCommand.length()))
		{
			ZKASSERT(false);
			throw RZkException( 
					m_ZkError, 
					"RNymMail::CreatePopAccountOnPopServer()",
					"Failed to write 'PASS' command", 
					IDS_FAILED_TO_CREATE_MAIL_ACCOUNT,
					strNymName.c_str());
		}
		else if( !ReceiveAndTestPOPResponse(socket) )
		{
			throw RZkException( 
					m_ZkError, 
					"RNymMail::CreatePopAccountOnPopServer()",
					"Invalid or no response to 'PASS' command", 
					IDS_FAILED_TO_CREATE_MAIL_ACCOUNT,
					strNymName.c_str());
		}
				
		ZKVERIFY(zkSocketWriteAll(&m_ZkError, socket, "QUIT\r\n", 6));
	}
	catch (RZkException)
	{
		if(socket)
		{
			ZKVERIFY(zkSocketFree(&m_ZkError, &socket));
		}
		throw;
	}

	if(socket)
	{
		ZKVERIFY(zkSocketFree(&m_ZkError, &socket));
	}
}

bool RNymMail::ReceiveAndTestPOPResponse(ZkSocket socket)
{
	bool ReturnValue = true;

	char Buffer[512];

	size_t BytesReceived = 0;

	if( !zkSocketRead(&m_ZkError, socket, Buffer, 512, &BytesReceived))
	{
		ReturnValue = false;
		ZKASSERT(0);
	}
	else
	{

		if(Buffer[0] != '+')
		{
			m_ZkError.setdetail(Buffer);

			ReturnValue = false;
		}
	}
	return ReturnValue;
}

// Mail limit support

void RNymMail::SetCurrentMailSentCount( const unsigned int in_mailSentCount )
{
	m_nCurrentMailSentCount = in_mailSentCount;
	ZkService::GetIClientPersist()->SetDirty();
}
unsigned int RNymMail::GetCurrentMailSentCount() const
{
	return m_nCurrentMailSentCount;
}

void RNymMail::SetMailSentInitTime( const time_t in_MailSentInitTime )
{
	mailSentInitTime = in_MailSentInitTime;
	ZkService::GetIClientPersist()->SetDirty();
}

time_t RNymMail::GetMailSentInitTime() const
{
	return mailSentInitTime;
}
