// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RNymMail.h: interface for the RNymMail class.
//
// This class implements the INymMail interface

#ifndef _ZKNYMMAIL_H
#define _ZKNYMMAIL_H

#include "INymMail.h"
#include <zkError.h>
#include <zkList.h>
#include <frNym.h>
#include <DZkError.h>
#include <zksocket.h>
#include <WMaipRb.h>

// Forward Declarations

class ZKNym ;

#define NYM_MAIL_PERSIST_VERSION 1

// Class RNymMail
// Implements the INymMail interface for a particular nym.
// The persistent state is contained within the m_RemoteNym 
// object which is shared with the parent Nym object

class RNymMail : public INymMail
{
public:
	RNymMail(RNym* pParent, WRemoteNym& RemoteNym );
	virtual ~RNymMail();

	
	// 2.0 Mail Functionality

	//
	virtual void	TransitionIfNecessary();
	virtual bool	RetrieveNymMailCert() throw(RZkException);
	void CreatePopAccountOnPopServer() throw(RZkException);

	// Used to log in to POP/ESMTP servers
	virtual void	GetMailCertificate(ZKSTD::string& out_strCert);
	virtual bool	UpdateMailCertificate();
	// Used to validate FMG signature of Internet->Nym mail.
	virtual FrCryptMacKey	GetMailMacKey();

	virtual void	OldGetEmailAddressBlockList(ZKSTD::vector<ZKSTD::string>&);
	virtual void	OldGetSpamControls( unsigned long &);
	virtual void	GetEmailAddressBlockList(ZKSTD::vector<ZKSTD::string>&);
	virtual void	SetEmailAddressBlockList(const ZKSTD::vector<ZKSTD::string>&) throw(RZkException) ;
	virtual void	SetSpamPreference(bool) throw(RZkException) ;
	virtual bool	GetSpamPreference() { return m_bBlockSpam; }

	virtual bool	NymHasTaglineChoice();
	virtual void	SetTaglinePreference(bool) throw(RZkException) ;
	virtual bool	GetTaglinePreference() { return m_bTagline; }

	virtual void	SetCurrentMailSentCount( const unsigned int );
	virtual unsigned int	GetCurrentMailSentCount() const;

	virtual void	SetMailSentInitTime( const time_t );
	virtual time_t		GetMailSentInitTime() const;

	virtual void	SelfTest() ;

	virtual void Save( WMemBuf& io_Buffer ) throw(RZkException) ;

	virtual void Load( ZKChunkReader& io_ChunkReader ) throw(RZkException) ;
private:

	enum
	{
		MAIL_PREF_BLOCK_SPAM = 1,
		MAIL_PREF_NO_TAGLINE = 2
	}MailPreferences;
	virtual T_FrRemoteNym* GetfrRemoteNym() { return m_RemoteNym ; } ;

	RNymMail(const RNymMail& rhs);
	RNymMail& operator=(const RNymMail& rhs);

	bool m_bNymHadMailChunk;

	// ctor member init
	DZkError					m_ZkError;
	ZKSTD::string					m_strResponse;
	
	RNym*						m_pParent ;
	WRemoteNym&					m_RemoteNym ;

	bool						m_bMailAccountInitialized;
	int							m_NextMailAuthVersion;
	FrCryptMacKey				m_MacKey;
	bool						m_bBlockSpam;
	bool						m_bTagline;
	unsigned int				m_nCurrentMailSentCount;
	time_t					mailSentInitTime;
	ZKSTD::string				m_strMailCertificate;
	ZKSTD::vector<ZKSTD::string> m_vecBlockList;


	void SendControlMessageToServer(const std::string& in_strRecipient, const char* in_strMsgBody)throw(RZkException);
	void SendEmailBlockListToServer(const std::vector<std::string>& vecAddresses) throw(RZkException);
	void SendSpamPreferenceToServer(bool in_bBlockSpam) throw(RZkException);
	void SendTaglinePreferenceToServer(bool in_bTagline) throw(RZkException);
	bool ReceiveAndTestPOPResponse(ZkSocket socket);

	void TestAndThrowServerResponse(int bStatus);

#ifdef WIN32

	virtual size_t	GetNumberOfReplyBlocks();
	virtual void    DecryptMessage( 
					WCypherMailMsg&  io_MailMsg ) throw(RZkException);
	virtual void    DecryptMessageEncryptedWithNymsPrivateKey( 
						unsigned char *  in_pBuffer,
						size_t           in_szBuffer,
						unsigned char ** out_ppClearBuffer,
						size_t&          out_szClearBuffer) throw(RZkException);
	bool LoadReplyBlocks();
	void EnsureCacheLoaded() ;								// Any access to the m_ReplyBlock cache should
															// call this function to make sure the cache is initialized

	void PeelReplyBlockFromMessage( 
		WCypherMailMsg&  io_MailMsg,
		unsigned char ** out_ppPeeledBuffer,
		size_t&          out_szPeeledBuffer  ) throw (RZkException);
	bool PeelReplyBlockXFromMessage(
						size_t				   in_IndexOfRBToUse,
						const unsigned char *  in_pBuffer,
						size_t				   in_szBuffer,
						unsigned char **	   out_ppPeeledBuffer,
						size_t&				   out_szPeeledBuffer ) ;


	WMaipRbVec                  m_ReplyBlocks ;				// This vector of Reply block objects
															// is a CACHE of the information stored
															// inside the Remote Nym
	bool						m_ReplyBlockCacheLoaded ;	// Tracks whether the m_ReplyBlock cache has been
															// loaded
#endif // WIN32

	// Get access to the shared RemoteNym object
	T_FrRemoteNym* RemoteNym() ;
};

#endif
