// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "StdAfx.h"
#include "RNymWebCacheMgr.h"
#include <date_hlp.h>
#include <ZKChunk.h>

// If a nym has MAX_WEBCACHE_RECORDS,
// prune to no more than OPTIMUM_WEBCACHE_RECORDS
// (this way, the nym doesn't constantly push over the limit,
// forcing us to prune every time)
#define MAX_WEBCACHE_RECORDS 1000
#define OPTIMUM_WEBCACHE_RECORDS 800

// 60 secs/min * 60 min/hr * 24 hr/day * 14 days == 2 weeks
#define MAX_WEBCACHE_RECORD_AGE ( 60 * 60 * 24 * 14)

RNymWebCacheMgr::RNymWebCacheMgr()
{

}

RNymWebCacheMgr::~RNymWebCacheMgr()
{
}

void RNymWebCacheMgr::ProcessIncomingETag(const ZKSTD::string& in_strUrl,
								const ZKSTD::string& in_strETag)
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	WebCacheMap::iterator it = m_mapCacheInfo.find(in_strUrl);
	if(it != m_mapCacheInfo.end())
	{
		it->second.m_strEtag = in_strETag;
		it->second.m_lastAccessed = time(NULL);
	}
	else
	{
		ZKSTD::pair<ZKSTD::string, WebCacheInfo> thePair(in_strUrl, WebCacheInfo(in_strUrl, in_strETag));
		m_mapCacheInfo.insert(thePair);
	}
}

void RNymWebCacheMgr::ProcessIncomingLastDateModified(const ZKSTD::string& in_strUrl,
								const ZKSTD::string& in_strLastDateModified)
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	time_t lMod = 0;
	tm tmMod;
	memset(&tmMod, 0, sizeof(tm));

	ZKS::ParseDateString(in_strLastDateModified, tmMod);
	ZKS::GetSecondsCUT(tmMod, lMod);
	if(lMod)
	{
		WebCacheMap::iterator it = m_mapCacheInfo.find(in_strUrl);
		if(it != m_mapCacheInfo.end())
		{
			it->second.m_lastModified = lMod;
			it->second.m_lastAccessed = time(NULL);
		}
		else
		{
			ZKSTD::pair<ZKSTD::string, WebCacheInfo> thePair(in_strUrl, WebCacheInfo(in_strUrl, lMod));
			m_mapCacheInfo.insert(thePair);
		}
	}
}

bool RNymWebCacheMgr::GetOutgoingTagForUrl(const ZKSTD::string& in_strUrl, ZKSTD::string& out_strField, ZKSTD::string& out_strVal)
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	WebCacheMap::iterator it = m_mapCacheInfo.find(in_strUrl);
	if(it != m_mapCacheInfo.end())
	{
		if(it->second.m_strEtag.length())
		{
			out_strField = "If-None-Match";
			out_strVal = it->second.m_strEtag;
			it->second.m_lastAccessed = time(NULL);
			return true;
		}
		
		if(it->second.m_lastModified > 0)
		{
			out_strField = "If-Modified-Since";
			ZKS::ConvertTmToString(out_strVal, *gmtime(&it->second.m_lastModified));
			it->second.m_lastAccessed = time(NULL);
			return true;
		}
	}
	return false;
}
bool RNymWebCacheMgr::GetLastModifiedTagToGiveToBrowser(const ZKSTD::string& in_strUrl, ZKSTD::string& out_strTag)
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	WebCacheMap::iterator it = m_mapCacheInfo.find(in_strUrl);
	if(it != m_mapCacheInfo.end())
	{
		if(it->second.m_lastModified > time(NULL))
		{
			// Create a time two days later than today
			time_t t = time(NULL) + 60 * 60 * 24 * 2;
			ZKS::ConvertTmToString(out_strTag, *gmtime(&t));
			out_strTag = "Last-Modified: " + out_strTag; 
			return true;
		}

		if(it->second.m_lastModified > 0)
		{	// Greater than zero, but before the present...

			// Create a time two days before today
			time_t t = time(NULL) - 60 * 60 * 24 * 2;
			ZKS::ConvertTmToString(out_strTag, *gmtime(&t));
			out_strTag = "Last-Modified: " + out_strTag; 
			return true;
		}
	}
	return false;
}
bool RNymWebCacheMgr::GetETagToGiveToBrowser(const ZKSTD::string& in_strUrl, ZKSTD::string& out_strTag)
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	WebCacheMap::iterator it = m_mapCacheInfo.find(in_strUrl);
	if(it != m_mapCacheInfo.end())
	{
		if(it->second.m_strEtag.length())
		{
			out_strTag = "ETag: \"Freedomized\"";
			return true;
		}
	}
	return false;
}

void RNymWebCacheMgr::PruneCacheInfo()
{
	
	WebCacheMap::iterator it = m_mapCacheInfo.begin();
	while(it != m_mapCacheInfo.end())
	{
		if(it->second.m_lastAccessed + MAX_WEBCACHE_RECORD_AGE < time(NULL))
		{
			ZKTRACE2("Deleting WebCache record for %s, it was %d secs old.\r\n", it->second.m_strUrl.c_str(), time(NULL) - it->second.m_lastAccessed);
            WebCacheMap::iterator temp = it;
			it++;
            m_mapCacheInfo.erase(temp);
		}
		else
		{
			it++;
		}
	}

	if(m_mapCacheInfo.size() > MAX_WEBCACHE_RECORDS)
	{
		ZKSTD::multimap<size_t, WebCacheInfo> records;

		{
			it = m_mapCacheInfo.begin();
			while(it != m_mapCacheInfo.end())
			{
				records.insert(ZKSTD::pair<size_t, WebCacheInfo>(it->second.m_lastAccessed, it->second));
				it++;
			}

			m_mapCacheInfo.erase(m_mapCacheInfo.begin(), m_mapCacheInfo.end());
		}


		ZKSTD::multimap<size_t, WebCacheInfo>::iterator itByAge = records.end();
		itByAge--;
		for(size_t i = 0; i < OPTIMUM_WEBCACHE_RECORDS; i++)
		{
			m_mapCacheInfo.insert(ZKSTD::pair<ZKSTD::string, WebCacheInfo>(itByAge->second.m_strUrl, itByAge->second));
			itByAge--;
		}
	}
}

void RNymWebCacheMgr::Save( WMemBuf& io_Buffer )
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);

	ZKChunkWriter cw( io_Buffer, NYM_WEB_CACHE_MGR_CHUNK_ID, NYM_WEB_CACHE_MGR_CHUNK_VERSION ) ;
	
	PruneCacheInfo();

	int nRecords = m_mapCacheInfo.size();
	cw.WriteInt32( nRecords ) ;
	WebCacheMap::iterator it = m_mapCacheInfo.begin();

	int i = 0;

	while(it != m_mapCacheInfo.end())
	{
		cw.WriteNonUIString( it->second.m_strUrl ) ;
		cw.WriteInt32( it->second.m_lastAccessed ) ;

		cw.WriteInt32( it->second.m_lastModified ) ;
		cw.WriteNonUIString( it->second.m_strEtag ) ;

		i++;
		it++;
	}

	ZKASSERT(i == nRecords);
	
	ZKTRACE2("Saved %d webcache records for a total of %d bytes.\r\n", m_mapCacheInfo.size(), cw.GetChunkSize());
}

void RNymWebCacheMgr::Load( ZKChunkReader& io_ChunkReader )
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
	ZKASSERT( io_ChunkReader.GetChunkID() == NYM_WEB_CACHE_MGR_CHUNK_ID ) ;

	m_mapCacheInfo.erase(m_mapCacheInfo.begin(), m_mapCacheInfo.end());

	size_t nUrls = io_ChunkReader.ReadInt32() ;

	for(size_t i = 0; i < nUrls; i++)
	{
		ZKSTD::string strUrl;
		ZKSTD::string strEtag;

		io_ChunkReader.ReadNonUIString( strUrl ) ;
		size_t lastAccessed = io_ChunkReader.ReadInt32() ;
		size_t lastModified = io_ChunkReader.ReadInt32() ;
		io_ChunkReader.ReadNonUIString( strEtag ) ;

		m_mapCacheInfo.insert(ZKSTD::pair<ZKSTD::string, WebCacheInfo>
								(strUrl, WebCacheInfo(strUrl, lastAccessed,
												lastModified, strEtag)));
	}
}
