// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom(R) code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "StdAfx.h"
#define __STDC_LIMIT_MACROS

#include "RRouteManager.h"
#include "DHKeyExchange.h"
#include "RRoutePicker.h"

#include <IZkConsoleMsg.h>
#include <INetworkPerformance.h>
#include <INym.h>
#include <IRoutePacket.h>
#include <INiqs.h>
#include <IClientCentral.h>
#include <INymRoute.h>
#include <INymCentral.h>
#include <ZkService.h>
#include <IZkPopup.h>
#include <TraceControl.h>
#include <ZKConvert.h>
#include <guids.h>

#include <RCloudSocket.h>
#include <IRouteManager.h>
#include <IFirewall.h>
#include <IKeyQry.h>
#include <INiqs.h>
#include <IClientInfo.h>
#include <ZkService.h>

#include <ZkErrorHandler.h>
#include <IZKErrorReporter.h>
#include <RZkException.h>
#include <lockguard.h>

#include <zkError.h>
#include <frNetworkDefaults.h>

#include <IZkRegistry.h>
#include <ZkStringMgr.h>
#include <MiscMessages.h>

#include <IRouteDialogs.h>

// send cover traffic out every 30 seconds by default (to each first hop)
// Actually this could come from registry for more configuration control
// Note, this value is in milliseconds
#define COVER_TRAFFIC_INTERVAL					30000

// 2 seconds wait time when creating a route
// On a fast connection a much smaller value should be suitable
#define	TIME_TO_WAIT_FOR_RTE_CRT_ACK	2000

#define PROLONG_ROUTE_SECS 5 * 60	// We attempt to prolong the route 5 minutes before the route will die
#define PROLONG_SLEEP_WHEN_BUSY 30  // If the route manager is busy when we try to prolong we will sleep this amount
#define MAX_NUM_ROUTE_RETRIES 3		// Override in registry: "RouteRetries"
#define RETRY_WAIT_FACTOR 3			// Each retry we wait longer for the route create ack, increasing
									// exponentially by this value.  For example if initial wait is 1000 ms and
									// RETRY_WAIT_FACTOR is 3 we wait 1000, 3000, 9000,....
									// and if it is 4 we wait 1000, 4000, 16000

// Routes which have had no traffic sent on them for m_routeIDleTicks * COVER_TRAFFIC_INTERVAL
// will be automajickly killed. --ol
#define DEFAULT_ROUTE_IDLE_TICKS 8         // 8 * 30s = 4 mins

#define DEFAULT_SERVER_HOP_COUNT 2


#ifdef _WIN32
// WaitForSingleObject won't take this otherwise, I don't know why
struct __tag_ZkEvent
{
	HANDLE event;
};
#endif



bool RRouteManager::m_bBuilding = false;
ZkEvent RRouteManager::m_Win32CoverTrafficHandle = NULL;

RRouteManager::RRouteManager() 
	: m_pRouteProlongTimer(NULL)
	, m_pBackgroundRouteKeepAliveTimer(NULL)
	, m_pBackgroundRouteKeepAliveThread(NULL)
	, m_pRouteCreateThread( NULL )
	, m_durationNymRoute( 0 )
	, m_durationServerRoute( 0 )
	, m_timeProlongRoute( 0 )
	, m_MaxNumRouteRetries( MAX_NUM_ROUTE_RETRIES )
	, m_RouteRetryFactor( RETRY_WAIT_FACTOR )
	, m_Context( NULL )
	, m_RteCrtAckEvnt()
	, m_cs()
    , m_DestPortForRouteBeingCreated( 0 )
	, m_RouteToBeCreatedOutgoingAci( 0 )
    , m_idleCounterTicks( 1 )
	, m_cntServerHops( DEFAULT_SERVER_HOP_COUNT )
{
	m_durationNymRoute		= DEFAULT_NYM_ROUTE_SECS ;
	m_durationServerRoute	= DEFAULT_SERVER_ROUTE_SECS ;
    m_routeIdleTicks        = DEFAULT_ROUTE_IDLE_TICKS ;

	IZkRegistry::GetDWORD( "RouteNymDuration", m_durationNymRoute ) ;
	IZkRegistry::GetDWORD( "RouteServerDuration", m_durationServerRoute ) ;
	IZkRegistry::GetDWORD( "RouteRetries", m_MaxNumRouteRetries ) ;
	IZkRegistry::GetDWORD( "RouteRetryFactor", m_RouteRetryFactor ) ;
	IZkRegistry::GetDWORD( "RouteIdleTicks", m_routeIdleTicks ) ;

	// Each Nym currently has its own preference for number of hops,
	// but currently this is the only way to change the count of hops
	// used for unauthenticated routes to servers.
	IZkRegistry::GetDWORD( "RouteServerHops", m_cntServerHops ) ;

	if (( m_cntServerHops == 0 ) || ( m_cntServerHops > 5 ) )
	{
		ZKASSERT( !"Invalid Registry setting for number of server hops" ) ;
		m_cntServerHops = DEFAULT_SERVER_HOP_COUNT ;
	}

	ZKASSERT( m_durationNymRoute > 0 ) ;
	ZKASSERT( m_durationServerRoute > 0 ) ;
	ZKASSERT( m_MaxNumRouteRetries > 0 ) ;

	m_timeProlongRoute = PROLONG_ROUTE_SECS ;

	if ( m_durationNymRoute <= m_timeProlongRoute )
	{
		// Only debugging circumstances should ever 
		// deal with very short routes
		m_timeProlongRoute = m_durationNymRoute / 2 ;
	}

}

RRouteManager::~RRouteManager()
{	
	KillRouteProlongTimer() ;

	ZKASSERT( m_Context == NULL ) ;
}


void
RRouteManager::Init()
{
	ZKASSERT( m_Context == NULL ) ;

	// During migration from v1 and v2 of the Aip we need to support
	// both protocols.  This code can be removed once v2 is fully 
	// deployed.#8395

	if( ZkService::GetIClientInfo()->IsAip2ProtocolEnabled() )
	{
		IZkConsoleMsg::ShowMsg( 
						CONSIMAGE_COMP_TERM, 
						IDS_CONS_JUST_STRING,
						"Client will use Aip Version 2 protocol" );

		ZKVERIFY(frLocalAipCtxNew(&m_err, &m_Context, FRPKT_DATA_VERSION_2, FRPKT_ROUTE_VERSION_2, MAX_ROUTES, FRPKT_NUM_ROUTES_MAX, WINDOWS_TCP_AND_UDP_PORT));
	}
	else
	{
		IZkConsoleMsg::ShowMsg( 
						CONSIMAGE_COMP_TERM, 
						IDS_CONS_JUST_STRING,
						"Client will use Aip Version 1 protocol" );

		ZKVERIFY(frLocalAipCtxNew(&m_err, &m_Context, FRPKT_DATA_VERSION_1, FRPKT_ROUTE_VERSION_1, MAX_ROUTES, FRPKT_NUM_ROUTES_MAX, WINDOWS_TCP_AND_UDP_PORT));
	}


	ZKVERIFY(frLocalAipCtxSetKeyCache(&m_err, m_Context, ZkService::GetIKeyQry()->GetKeyCache()));
	ZKVERIFY(frLocalAipCtxSetNetInfo(&m_err, m_Context, ZkService::GetINiqs()->GetNetInfo()));

	m_RouteTable.Init( m_Context ) ;

	if ( ZkService::GetIClientInfo()->ClientInfoIsValid() )
	{
		// Inform the Aip Context of the time so that it knows what what delta
		// the Aip times will be at relative to the client time
		SetFreedomTime( ZkService::GetIClientInfo()->GetCurrentUTCFromClientInfo() ) ;
	}

#ifdef _WIN32
	ZKVERIFY( zkEventNew (&m_err, &m_Win32CoverTrafficHandle, 0));
	zkThreadCreate(&m_err, &m_pBackgroundRouteKeepAliveThread, &RRouteManager::BackgroundRouteKeepAliveThread, NULL) ;
#else
	StartCoverTrafficTimer( COVER_TRAFFIC_INTERVAL );
#endif

}

#ifndef _WIN32
bool
RRouteManager::StartCoverTrafficTimer(const unsigned long dwMil)
{
	if(m_pBackgroundRouteKeepAliveTimer)
	{
		zkITimerFree(&m_err, &m_pBackgroundRouteKeepAliveTimer);
	}
	
	if(!zkITimerNew(&m_err, &m_pBackgroundRouteKeepAliveTimer, &RRouteManager::BackgroundRouteKeepAliveTimer, dwMil))
	{
		return ZK_FALSE;
	}

	return ZK_TRUE;
}

void ZK_ITIMER_CALLBACK
RRouteManager::BackgroundRouteKeepAliveTimer(void*, unsigned int, unsigned int, unsigned long)
{
	try
	{
		ZkService::GetIRouteManager()->SendCoverTrafficPacketOnAllRoutes( ) ;
        RRouteManager::RouteIdleTimer(NULL, 0, 0, 0);
	}
	catch( IZkException& )
	{
		// We don't expect any throws, in fact
		// there may be unreleased locks in this case.
		ZKASSERT( false ) ;
	}
}
#endif

void ZK_ITIMER_CALLBACK 
RRouteManager::RouteIdleTimer(
    void *p, 
    unsigned int i1, 
    unsigned int i2, 
    unsigned long time)
{
    ZkService::GetIRouteManager()->UpdateIdleCounter();
    ZkService::GetIRouteManager()->KillIdleRoutes();
}


void
RRouteManager::Terminate()
{
	// Close down route create thread if it exists
	if (m_pRouteCreateThread)
	{
		DZkError derr ;
		void *rval;
		ZKVERIFY(zkThreadJoin(&derr, &m_pRouteCreateThread, &rval));
	}

	// Stop keep-alive thread/timer
#ifdef WIN32

	if ( m_Win32CoverTrafficHandle != NULL )
	{
		ZKVERIFY( zkEventSet(&m_err, m_Win32CoverTrafficHandle) );

		HANDLE BackgroundRouteKeepAliveThreadHandle = NULL ;
		zkThreadGetThreadNP( &m_err, m_pRouteCreateThread, &BackgroundRouteKeepAliveThreadHandle ) ;
		if(BackgroundRouteKeepAliveThreadHandle)
		{
			WaitForSingleObject( BackgroundRouteKeepAliveThreadHandle, 10000);
		}
		ZKVERIFY( zkEventFree(&m_err, &m_Win32CoverTrafficHandle) );
	}
	else
	{
		ZKASSERT( "RRouteManager::Terminate() called without associated call with Init" && false ) ;
	}
#else
	KillCoverTrafficTimer();
#endif
	
	// udp socket must only be closed afterwards obviously since RCloudSocket re-uses 'theUDPThreadSocket'
	KillAllRoutes(); // Can't expect the user to remember everything	

	if ( m_Context != NULL )
	{
		ZKVERIFY(frLocalAipCtxFree(&m_err, &m_Context));
	}
}

// Under Linux the timer is stopped when this is called from the dtor
#ifndef _WIN32
void
RRouteManager::KillCoverTrafficTimer()
{
	if(m_pBackgroundRouteKeepAliveTimer)
	{
		zkITimerFree(&m_err, &m_pBackgroundRouteKeepAliveTimer);
	}
}
#endif


IRoutePacket* 
RRouteManager::GetRoutePacketHandler()
{
	IRoutePacket* pRoute = (IRoutePacket*) this  ;

	ZKASSERT( pRoute != NULL ) ;
	return pRoute ;
}



void
RRouteManager::ProlongNymRoute()
{
	// Kill the timer - a new one will be started if we decide we need it
	KillRouteProlongTimer();	
	
	// The timer might go off even if the nym authenticated route is now gone

	if (( ZkService::GetINymCentral()->IsCurrentNymUser() ) ||
		( m_RouteTable.GetActiveAuthenticatedRoute() == INVALID_ROUTE_ID ))
	{
		ZKCTRACE0( TRACE_ROUTEMANAGER, "Not extending route because we have nothing to extend." ) ;
		return ;
	}

	if ( m_bBuilding )
	{
		// We appear to be busy building a different route, lets try back a little 
		// later - if this happens a few times then the underlying route is going
		// to expire
		ZKCTRACE0( TRACE_ROUTEMANAGER, "Route Manager is busy building another route - can't prolong nym route yet\n" ) ;
		StartRouteProlongTimer( PROLONG_SLEEP_WHEN_BUSY*1000 );
		return ;
	}

	try
	{
		ZKCTRACE0( TRACE_ROUTEMANAGER, "Gonna prolong authenticated route, getting nym rulez\n");

		WRouteVec routes ;
		size_t nHops = CalcRoutesForExtendingExistingNymRoute( m_MaxNumRouteRetries, routes ) ;

		ZKCTRACE1( TRACE_ROUTEMANAGER, "This nym wants [%d] hops in this route getting prolonged\n", nHops);

		if(routes.size() == 0)
		{
			// Instead of displaying AfxMessageBox(IDS_CANT_APPLY_RULE); we'll simply leave a message in the console
			// We'll use a yellow icon not to scare the user too much

			IZkConsoleMsg::ShowMsg( CONSIMAGE_COMP_TERM, IDS_CONS_CANT_PROLONG_NO_RULEZ ) ;
			return;
		}

		// Get the hops of the existing route so that we can compare		
		WRoute oldRoute ;

		m_RouteTable.GetAipIds( m_RouteTable.GetActiveAuthenticatedRoute(), oldRoute ) ;

		ZKSTD::string strNewHops ;

		m_bBuilding = true; // When this is true, it is the only time traffic will be let out to port 51102

		// Try with different route candidates until one seems to work
        size_t i;
		for (  i = 0 ; i < routes.size() ; i++ )
		{
			try
			{

				strNewHops = GetRouteDesc( routes[ i ] ) ;

				IZkConsoleMsg::ShowMsg( 
								CONSIMAGE_COMP_TERM, 
								IDS_CONS_ATTEMPT_PROLONG_ROUTE,
								strNewHops.c_str() );

				ProlongRoute(
							m_RouteTable.GetActiveAuthenticatedRoute(), 
							m_durationNymRoute, 
							routes[i],
							i+1 );
				ZKCTRACE1( TRACE_ROUTEMANAGER, "There are [%d] hops in this prolonged route\n", nHops);

				// We are successful
				break ;

			}
			catch(IZkException& ex)
			{
				if ( i == ( routes.size() - 1 ))
				{
					m_bBuilding = false;

					// We haven't been able to prolong the route despite having
					// retried several retries

					// In 3 minutes 'User' will get re-selected since we couldn't prolong route; the user will see a dialog
					// See ::PostMessage(msgDest, WM_SELECTUSER_NYM, nRouteID, 0); in this file for code which will execute in 3 minutes
					
					IZkConsoleMsg::ShowMsg( 
						CONSIMAGE_SERIOUS, 
						IDS_CONS_ROUTE_PROLONG_FAIL, 
						ZkErrorHandler::GetIZkErrorReporter()->GetErrorString(ex).c_str() ) ;

					return;
				}
			}
		}

		m_bBuilding = false;

		if ( oldRoute == routes[ i ] )
		{
			// Because we are forced to use the same exit hop, it
			// is very possible that the same route is selected twice
			// This is not a bad thing
			IZkConsoleMsg::ShowMsg( 
				CONSIMAGE_COMP_TERM, 
				IDS_CONS_ROUTE_PROLONG_SAME_HOPS, 
				strNewHops.c_str() ) ;
		}
		else
		{
			IZkConsoleMsg::ShowMsg( 
				CONSIMAGE_COMP_TERM, 
				IDS_CONS_ROUTE_PROLONG, 
				strNewHops.c_str() ) ;
		}	

		ZKCTRACE1( 
			TRACE_ROUTEMANAGER, 
			"Route prolonged, doing so again in [%d] milliseconds\n", 
			(m_durationNymRoute*1000) - m_timeProlongRoute*1000);

		// Prepare for the next prolong effort
		// subtract 3 minutes from initial duration
		StartRouteProlongTimer( (m_durationNymRoute*1000)-m_timeProlongRoute*1000 );
	}
	catch( ... )
	{
	}
}


void ZK_ITIMER_CALLBACK 
RRouteManager::RouteProlongTimerCallback(
					void *p, 
					unsigned int i1, 
					unsigned int i2, 
					unsigned long time)
{
	ZkService::GetIRouteManager()->ProlongNymRoute() ;
}

void
RRouteManager::KillRouteProlongTimer()
{
	if( m_pRouteProlongTimer )
	{
		// we could get a destroy _before_ our indicated timeout, handle that
		ZKVERIFY( zkITimerFree( &m_err, &m_pRouteProlongTimer ) );
	}

	ZKASSERT( m_pRouteProlongTimer == NULL ) ;
}

void
RRouteManager::StartRouteProlongTimer( uint32_t in_msTime )
{
	ZKASSERT( in_msTime != 0 ) ;

	ZKASSERT( m_pRouteProlongTimer == NULL ) ;

	ZKVERIFY( zkITimerNew( &m_err, 
						   &m_pRouteProlongTimer,
						   RRouteManager::RouteProlongTimerCallback,
						   in_msTime ) ) ;
	ZKASSERT( m_pRouteProlongTimer );
}


void
RRouteManager::MakeSureRouteManagerNotBusy() throw(RZkException)
{
	// This function attempts to wait until the
	// route manager is not busy any more.
	// It is not the ideal solution, nor is it necessarily
	// completely thread safe.  Because m_bBuilding is not
	// actually a lock we aren't sure that someone else won't also
	// steal m_bBuilding after we think we are safe.
	///Please view the document "RouteMgrThreadIssues.html"

	size_t cntRetries = 10 ;	// Wait five seconds at most

	if ( m_bBuilding )
	{
		ZKCTRACE0( TRACE_ROUTEMANAGER, "Route Manager is busy - probably building another route\n" ) ;

		while ( cntRetries )
		{
	#ifdef WIN32
			Sleep(500); 
	#else
			usleep(50);
	#endif

			ZKCTRACE1( 
				TRACE_ROUTEMANAGER, 
				"Testing if Route Manager is busy again - retries remaining %d\n", 
				cntRetries ) ;

			if ( !m_bBuilding )
			{	
				return ;
			}

			cntRetries-- ;
		}

		// we don't want to be trapped for ever so we eventually fail

		throw RZkException(
				"RRouteManager::MakeSureRouteManagerNotBusy", 
				"Route Manager is busy.  Try again or restart the application.", 
				IDS_BUILDROUTE_FAIL);
	}

	return ;
}

int 
RRouteManager::StartAuthenticatedRoute(
	size_t in_NymIndex)
{	
	size_t nHops = 0 ; 

	// Only real nyms should make this call
	ZKASSERT( ZkService::GetINymCentral()->DoesNymUseCloud( in_NymIndex ) ) ;
	ZKASSERT( in_NymIndex != USER_NYM_INDEX );

	try
	{
		MakeSureRouteManagerNotBusy() ;

		WRouteVec routes ;
		size_t nHops = CalcRoutesForNym( m_MaxNumRouteRetries, routes ) ;

		if( routes.size() == 0 )
		{

			return 0;
		}
		
		ZKSTD::string strNymName ;
		ZkService::GetINymCentral()->GetNymName( in_NymIndex, strNymName ) ;

		m_bBuilding = true; // When this is true, it is the only time traffic will be let out to port 51102

		for( size_t i = 0 ; i < routes.size() ; i++ )
		{
			try
			{
				IZkConsoleMsg::ShowMsg( 
								CONSIMAGE_COMP_TERM, 
								IDS_CONS_ATTEMPT_AUTH_ROUTE,
								GetRouteDesc( routes[ i ] ).c_str()
								);

				ZKASSERT(m_RouteTable.GetActiveAuthenticatedRoute() == INVALID_ROUTE_ID);

				BuildRouteForNymTraffic(
									routes[ i ], 
									in_NymIndex,
									m_durationNymRoute,
									i+1 );

				// Success
				break ;
			}
			catch( IZkException& )
			{
				ZKCTRACE1( 
					TRACE_ROUTEMANAGER, 
					"Route creation attempt %d has failed\n", 
					i ) ;

				// We only throw out when all of our attempts have failed
				if ( i == ( routes.size() - 1 ) )
				{
					throw ;
				}
			}
		}




		IZkConsoleMsg::ShowMsg( 
						CONSIMAGE_COMP_TERM, 
						IDS_CONS_AUTH_ROUTE_CREATE,
						strNymName.c_str() );
		
		ZKASSERT( m_RouteTable.GetActiveAuthenticatedRoute() != INVALID_ROUTE_ID );

		// I set the building variable to false _after_ inserting the IP in our set to avoid any race conditions
		RRouteManager::m_bBuilding = false;

		ZKCTRACE1( TRACE_ROUTEMANAGER, "Authenticated route built, will attempt auto prolong in [%d] milliseconds\n", (m_durationNymRoute*1000)-m_timeProlongRoute*1000);

		// subtract 3 minutes from initial duration

		KillRouteProlongTimer() ;
		StartRouteProlongTimer( (m_durationNymRoute*1000)- m_timeProlongRoute*1000 ) ;
		return 1;
	}
	catch(IZkException& ex)
	{
		IZkConsoleMsg::ShowMsg( 
						CONSIMAGE_WARNING, 
						IDS_CONS_AUTH_ROUTE_FAIL, 
						ZkErrorHandler::GetIZkErrorReporter()->GetErrorString(ex).c_str() );

		ZkErrorHandler::GetIZkErrorReporter()->ShowError(ex);
		RRouteManager::m_bBuilding = false;
	}

	return 0;
}


// The 'User' Nym was probably selected
// or user 'switched' to another one of his Nyms
void 
RRouteManager::KillAuthenticatedRoute()
{
	ZkACI activeAuthRoute = m_RouteTable.GetActiveAuthenticatedRoute() ;

	if ( activeAuthRoute != INVALID_ROUTE_ID)
	{
		KillSpecificRoute( activeAuthRoute ) ;
	}
}

bool 
RRouteManager::GetUnauthenticatedRoute(
	const uint16_t DestPort, 
	unsigned int* out_pRoute)
{
	bool bRetVal = false; 

	int nRouteID = m_RouteTable.GetActiveRouteToFreedomServer( DestPort ) ;

	if ( nRouteID != INVALID_ROUTE_ID )
	{
		bRetVal = true ;
	}

	if (out_pRoute)
	{
		*out_pRoute = nRouteID;
	}
	return bRetVal;
}



void 
RRouteManager::GetUnauthenticatedRoutes(
	Port2AipIds& mmap, 
	size_t& numRoutes )
{
	for ( size_t i = 0 ; i < m_RouteTable.GetMaxRoutes() ; i++ )
	{
		if (( m_RouteTable.IsRouteActive( i ) ) &&
			( m_RouteTable.GetRouteType( i ) == UnauthenticatedRoute ) )
		{
			size_t	numAips = MAX_AIPS_IN_ROUTE; 
			uint32_t aipIds[MAX_ROUTES][MAX_AIPS_IN_ROUTE];
			
			m_RouteTable.GetAipIds(i, aipIds[numRoutes], numAips);
			
			// For that port keep all the aip id's which make up the route
			for(size_t j=0; j<numAips; ++j)
			{
				mmap.insert(ZKSTD::pair<uint16_t, unsigned int>(m_RouteTable.GetDestinationPort( i ), 
																aipIds[numRoutes][j]));
			}

			++numRoutes;		
			ZKASSERT(numAips < MAX_AIPS_IN_ROUTE);
			ZKASSERT(numRoutes < MAX_ROUTES);
			ZKASSERT(0 < numAips); 
			
		}
	}
}


void 
RRouteManager::GetAuthenticatedRoutes(
	Port2AipIds& mmap, 
	size_t& numAuthRoutes )
{

	// Fill the 'AuthAipIds' array 'numAuthRoutes' times with 'numAips' aip id's in authenticated routes
	numAuthRoutes = 0;

	if ( m_RouteTable.GetActiveAuthenticatedRoute() != INVALID_ROUTE_ID ) 
	{
		size_t	numAips = MAX_AIPS_IN_ROUTE; 
		uint32_t AuthAipIds[MAX_AIPS_IN_ROUTE];

		m_RouteTable.GetAipIds(m_RouteTable.GetActiveAuthenticatedRoute(), AuthAipIds, numAips);

		// For that FAKE_PORT keep all the aip id's which make up the route
		for(size_t i=0; i<numAips; ++i)
		{
			mmap.insert(ZKSTD::pair<uint16_t, unsigned int>(FAKE_PORT, AuthAipIds[i]));
		}
		++numAuthRoutes;
	}

	ZKASSERT(numAuthRoutes < MAX_ROUTES);

	// Should only be one route returned max since only one Nym at a time can be selected presently
	ZKASSERT( (0 == numAuthRoutes) || (1 == numAuthRoutes) ); 
}


RouteType
RRouteManager::GetRouteTypeFromDestPort( const uint16_t DestPort )
{
	switch( DestPort )
	{
	case FR_DFLT_TOKEN_SERVER_PORT:
	case FR_DFLT_KEYQRY_SERVER_PORT:
	case FR_DFLT_NYM_SERVER_PORT:
	case FR_DFLT_NIQS_SERVER_PORT:
	case FR_DFLT_FMGMAIP_SERVER_PORT:
	case FR_DFLT_POP_PORT:
	case FR_DFLT_NMTA_PORT:
		return UnauthenticatedRoute;
	};
	return AuthenticatedRoute;
}


bool 
RRouteManager::StartUnauthRoute( 
	const uint16_t in_DestPort ) throw()
{
	// Initialized to false in the ctor, when this is true, it is the only time 
	// traffic will be let out to port 51102
	m_bBuilding = true; 

	size_t nHops = 0 ;
	try
	{
		WRouteVec routes ;

		CalcCoreAipRouteToServer( in_DestPort, m_MaxNumRouteRetries, routes ) ;
		
		if( routes.size() == 0 )
		{
			throw RZkException(
					"RRouteManager::StartUnauthRoute", 
					"No routes are currently available", 
					IDS_NO_UNAUTH_ROUTES_AVAIL);
		}	

		// Attempt to build a route until one succeeds or all fail
		// The timeout increases with each additional retry
		for ( size_t i = 0 ; i < routes.size() ; i++ )
		{
			try
			{

				IZkConsoleMsg::ShowMsg( 
								CONSIMAGE_COMP_TERM, 
								IDS_CONS_ATTEMPT_UNAUTH_ROUTE,
								ZKConvert::GetServerNameFromPort( in_DestPort ).c_str(),
								GetRouteDesc( routes[ i ] ).c_str()
								);
				
				BuildRouteToFreedomServer(
									in_DestPort,
									routes[ i ],
									m_durationServerRoute,
									i+1 );

				// Successful completion
				break ;
			}
			catch( IZkException& )
			{
				if ( i == ( routes.size() - 1 ) )
				{
					throw ;
				}
			}
		}

		IZkConsoleMsg::ShowMsg( 
			CONSIMAGE_COMP_TERM, 
			IDS_CONS_UNAUTH_ROUTE_CREATE, 
			ZKConvert::GetServerNameFromPort( in_DestPort ).c_str() ) ;

	}
	catch(IZkException& ex) 
	{
		// After printing the problematic route on the console, 
		// now we want to print the cause of the failure.
		IZkConsoleMsg::ShowMsg( 
			CONSIMAGE_SERIOUS, 
			IDS_CONS_UNAUTH_ROUTE_FAIL, 
			ZKConvert::GetServerNameFromPort( in_DestPort ).c_str(),
			ZkErrorHandler::GetIZkErrorReporter()->GetErrorString(ex).c_str() ) ;

		m_bBuilding = false ; 
		return false;
	}
	catch(...)
	{
		// This code should never be executed.
		ZKASSERT(false);
		IZkConsoleMsg::ShowMsg( 
			CONSIMAGE_SERIOUS, 
			IDS_CONS_UNAUTH_ROUTE_FAIL, 
			ZKConvert::GetServerNameFromPort( in_DestPort ).c_str(),
			"No error string" ) ;

		m_bBuilding = false ; 
		return false;
	}
	m_bBuilding = false ; 
	return true;
}

uint32_t 
RRouteManager::AipIdFromIpAndPort(const u_long DestIP, const uint16_t DestPort)
{
	return ZkService::GetINiqs()->GetEntId(DestIP, DestPort );
}

void * ZK_THREAD_CALLBACK 
RRouteManager::BuildRouteThread(void * in_pThis)
{
	RRouteManager* pThis = static_cast<RRouteManager*>( in_pThis ) ;

	m_bBuilding = true; 

	unsigned int nRouteID = INVALID_ROUTE_ID;
	pThis->StartUnauthRoute( pThis->m_DestPortForRouteBeingCreated );

	m_bBuilding = false; 
	return (void *)1;
}



// We'll only let out unencrypted TCP traffic if it's to one of our Aips 
// and even so, it has to be when a route create was initiated.
// TCP Port will cause frNetInfoEntIdentFromIPandPort() to fail since entities aren't indexed on the TCP port (Luc 15 Sept 99)
// ....
// As for UDP traffic, we have to check if it's going towards one of our first Aips in a route
// which should slow down web surfing unfortunately.
bool 
RRouteManager::IsCloudTrafficAllowed(const u_long in_IP)
{
	ZKCTRACE0( TRACE_ROUTEMANAGER, "Encrypted traffic heading for first Aip\n");
	if(m_bBuilding)
	{
		ZKCTRACE0( TRACE_ROUTEMANAGER, "Create packet leaving machine\n");

		// Make sure packet is actually going to one of our Aips
		return (AipIdFromIpAndPort(in_IP, FR_DFLT_AIP_UDP_PORT) != 0);;
	}
	else
	{
		// Route must already exist, make sure that
		// it is really an Aip that we have a route to
		// REVIEW - this is an expensive call to make per outgoing
		// cloud packet
		return m_RouteTable.IPisaFirstHopInaRoute(in_IP);
	}
}

bool
RRouteManager::IsDiffieHellmanTrafficAllowed( const u_long in_IP )
{
	// Verify that we are really building a route and that the
	// destination is really an Aip
	// See bug #635 for why this is important

	if( m_bBuilding )
	{		
		ZKCTRACE0( TRACE_ROUTEMANAGER, "D-Hellman leaving machine\n");

		// Make sure DH is actually going to one of our Aips
		return (AipIdFromIpAndPort(in_IP, FR_DFLT_AIP_UDP_PORT) != 0);
	}
	else
	{
		return false ;
	}
}

bool 
RRouteManager::GetRouteIdForThisDestination(
	const uint16_t in_DestPort, 
	unsigned int&  out_nRouteID, 
	sockaddr_in &  out_FirstHopAddr)
{
	bool bRet = false;

	ZKASSERT( FR_DFLT_AIP_TCP_PORT != in_DestPort ); // Bug#1566	

	out_nRouteID = INVALID_ROUTE_ID ;

	switch(GetRouteTypeFromDestPort(in_DestPort))
	{

	case UnauthenticatedRoute :
		{
			// We get the first aip address immediately so that this is our only access to the
			// table for each outgoing packet
			ZkACI RouteID = m_RouteTable.GetActiveRouteToFreedomServer( in_DestPort, &out_FirstHopAddr ) ;

			if ( RouteID != INVALID_ROUTE_ID )
			{
				out_nRouteID = RouteID ;
				ZKCTRACE2( TRACE_ROUTEMANAGER, "port [%d] already has an existing route [%d]\n", in_DestPort, out_nRouteID);
				bRet = true;
			}
			else if (RouteID == INVALID_ROUTE_ID
				&& in_DestPort == FR_DFLT_TOKEN_SERVER_PORT)
			{
				// Bug 8019 means this should never happen because the route is built first
				ZKASSERT(false && "No route to Token Server exists, but client is trying to connect.");
				bRet = false;
			}
			else if (RouteID == INVALID_ROUTE_ID
				&& in_DestPort == FR_DFLT_NYM_SERVER_PORT)
			{
				// Bug 8019 means this should never happen because the route is built first
				ZKASSERT(false && "No route to Nym Server exists, but client is trying to connect.");
				bRet = false;
			}
			else if(m_bBuilding)
			{
				ZKCTRACE0( TRACE_ROUTEMANAGER, "Currently trying to build a route, postponing further route create requests until previous attempt fails or succeeds\n");
				ZKCTRACE0( TRACE_ROUTEMANAGER, "This is entirely dependent on the number of retries and timeout value for connect() calls on this machine\n");
				bRet = false;
			}
			else if(!m_bBuilding) // Just in case it changes underneath us
			{
				ZKCTRACE1( TRACE_ROUTEMANAGER, "An unauthenticated route for [%d] will be built for you\n", in_DestPort);				

				m_DestPortForRouteBeingCreated = in_DestPort ;

				DZkError derr ;
				bRet = zkThreadCreate(&derr, &m_pRouteCreateThread, &RRouteManager::BuildRouteThread, this) ;

	#ifdef WIN32
				if ( m_pRouteCreateThread )
				{
					HANDLE hRouteCreateThread = NULL ;
					zkThreadGetThreadNP( &derr, m_pRouteCreateThread, &hRouteCreateThread ) ;
						
					SetThreadPriority( hRouteCreateThread, THREAD_PRIORITY_HIGHEST ) ;
				}
	#endif

				zkThreadYield( &derr ) ;
			}
		}
		break;

	case AuthenticatedRoute :
		ZKCTRACE1( 
			TRACE_ROUTEMANAGER, 
			"An AUTHENTICATED route seems to be necessary for port [%d].\n", in_DestPort );
		out_nRouteID = m_RouteTable.GetActiveAuthenticatedRoute( &out_FirstHopAddr ) ;
		if ( out_nRouteID == INVALID_ROUTE_ID ) 
		{
			bRet = false ;
		}
		break;
	}
	return bRet;
}

ZKSTD::string
RRouteManager::GetRouteDesc( const FrEntIdType& in_entType ) throw()
{
	ZKSTD::string strRetVal ;

	unsigned int nRouteID = m_RouteTable.GetActiveRouteToFreedomServer( 
		ZKConvert::GetPortFromEntityType(in_entType) ) ;

	if ( nRouteID == INVALID_ROUTE_ID )
	{
		ZKASSERT(false && "No route exists to to this entity type");
		ZkStringMgr::Load(strRetVal, IDS_NO_ROUTE_DESC);
	}
	else
	{
		WRoute route;
		m_RouteTable.GetAipIds(nRouteID, route);
		strRetVal = GetRouteDesc(route);
	}
	return strRetVal;
}


ZKSTD::string
RRouteManager::GetRouteDesc( const WRoute& in_Route ) throw()
{
	ZKSTD::string strRetVal ;

	WNetInfoDatumVec lst;
	try
	{
		for(size_t i=0; i< in_Route.GetSize(); ++i )
		{	// loop through array first so they'll be displayed in order
			
			ZKASSERT( in_Route.GetHop( i ) != 0 ) ;

			WNetInfoDatum NextHop ;
			ZkService::GetINiqs()->GetEntityById( in_Route.GetHop( i ), NextHop ) ;

			ZKSTD::string strEntityName ;
			NextHop.GetEntityName( strEntityName ) ;

			strRetVal += strEntityName ;

			if ( i < in_Route.GetSize() - 1 )
			{
				strRetVal += ", " ;
			}
		}		
	}
	catch(IZkException& ex)
	{
		ZkErrorHandler::GetIZkErrorReporter()->ShowError(ex);
	}	

	return strRetVal ;
}

void RRouteManager::TerminateMailRoutes()
{
	ZkACI unauthRoute = m_RouteTable.GetActiveRouteToFreedomServer( FR_DFLT_POP_PORT ) ;
	if(unauthRoute != INVALID_ROUTE_ID)
	{
		KillSpecificRoute(unauthRoute);
	}
	
	unauthRoute = m_RouteTable.GetActiveRouteToFreedomServer( FR_DFLT_NMTA_PORT ) ;
	if(unauthRoute != INVALID_ROUTE_ID)
	{
		KillSpecificRoute(unauthRoute);
	}
}


void 
RRouteManager::TerminateUnauthRoute(const uint16_t in_uPort)
{
	ZKASSERT( in_uPort != 0 ) ;
	ZKASSERT(UnauthenticatedRoute == GetRouteTypeFromDestPort( in_uPort) );
	
	ZkACI authRoute = m_RouteTable.GetActiveRouteToFreedomServer( in_uPort ) ;

	if ( authRoute != INVALID_ROUTE_ID )
	{	
		ZKCTRACE1( TRACE_ROUTEMANAGER, "Sending destroy on unauthentic route id  [%d]\n", authRoute);
		// Send a route destroy
		KillSpecificRoute(authRoute);
	}
	else
	{
		ZKCTRACE1( TRACE_ROUTEMANAGER, "No unauth route to port [%d] exists\n", in_uPort);
	}
}

ZkACI
RRouteManager::GetAuthRouteID()
{
	return m_RouteTable.GetActiveAuthenticatedRoute() ;
}

// Make sure there are not active routes
void 
RRouteManager::KillAllRoutes()
{
	for ( size_t i = 0 ; i < m_RouteTable.GetMaxRoutes() ; i++ )
	{
		KillSpecificRoute( i ) ;

	}
}

// Tears down route if it is active and removes local route information
void
RRouteManager::KillSpecificRoute( const ZkACI in_RouteID )
{
	if (( m_RouteTable.GetRouteState( in_RouteID ) == RouteActive ) ||
		( m_RouteTable.GetRouteState( in_RouteID ) == RouteRetired ))
	{

		// Tear down the route
		ZKVERIFY( ShutdownRoute( in_RouteID ) ) ;

		//  IsCloudTrafficAllowed() will be called when the route destroy packet leaves
		//  We need to wait a little while to make sure it goes out,
		//  otherwise we leave a orphan
#ifdef WIN32
		Sleep(1000); 
#else
        usleep(100);
#endif
	}

	// Make sure route is removed from table
	m_RouteTable.DeleteDeadRoute( in_RouteID ) ;
}

bool
RRouteManager::ShutdownRoute( ZkACI in_RouteID )
{		
	ZKASSERT( ( m_RouteTable.GetRouteState( in_RouteID ) == RouteActive ) ||
			  ( m_RouteTable.GetRouteState( in_RouteID ) == RouteRetired ) )  ;

	int			nRet = 0;
	char		data[FRPKT_BUF_MAX_SIZE];
	size_t		dlen = FRPKT_BUF_MAX_SIZE;
    WAipData    aipData;

	if(frLocalAipMakeDestroyPktUsingAipData(&m_err, m_Context, aipData, data, &dlen, in_RouteID))
	{
		if(frLocalAipLinkEncrypt(&m_err, 
								  m_Context, 
								  in_RouteID, 
								  FRPKTTYPE_AIP_ROUTE_DESTROY, 
								  data, 
								  dlen))
		{
			sockaddr_in firstAipAddr ;
			m_RouteTable.GetFirstAipAddr( in_RouteID, firstAipAddr ) ;

			if( RCloudSocket::theRCloudSocket().SendTo( data, dlen, firstAipAddr ))
			{
				ZKCTRACE1( TRACE_ROUTEMANAGER, "ACI [%d] successfully deleted\n", in_RouteID);
				return true;
			}
		}
	}

	return false ;
}


void
RRouteManager::VerifyKeysAvailableForAllHops( unsigned int in_nRouteId ) throw( RZkException )
{
	// We could actually verify this before calling frLocalAipStartNewRoute()
	// on our own array of Hop Ids, not using frLocalAipGetAipNames - then 
	// we don't have to remove the route if there is a failure.

	const unsigned int uKeyVersion = 0 ;
	char names[FRPKT_NUM_ROUTES_MAX][FRENT_NAME_SIZE];
	size_t num = FRPKT_NUM_ROUTES_MAX;
	if(frLocalAipGetAipNames(&m_err, 
							 m_Context, 
							 in_nRouteId, 
							 names, 
							 &num))
	{
		fr_crypt_pubkey*	unused = NULL;
		for(size_t i=0; i<num; ++i)
		{
			ZKCTRACE1( TRACE_ROUTEMANAGER, "[%s] is in this route\n", names[i]);
			if(!ZkService::GetIKeyQry()->GetKey(
										m_err, 
										FRENT_TYPE_AIP, 
										DEFAULT_MAXKEYQUERY_TIME_SECS, 
										names[i], 
										uKeyVersion, 
										FRCRYPT_KEYTYPE_CRYPTION, 
										&unused ))
			{
				throw RZkException(	m_err, 
									"RRouteManager::VerifyKeysAvailableForAllHops()",
									"GetKey() for Hop Crypto key failed",
									IDS_BUILDROUTE_AIPCRYPTIONKEY,
									names[i]);

			}

			if(!ZkService::GetIKeyQry()->GetKey(
										m_err, 
										FRENT_TYPE_AIP, 
										DEFAULT_MAXKEYQUERY_TIME_SECS, 
										names[i], 
										uKeyVersion, 
										FRCRYPT_KEYTYPE_SIGNING, 
										&unused ))
			{
				throw RZkException(	m_err, 
									"RRouteManager::VerifyKeysAvailableForAllHops()",
									"GetKey() for Hop Signing Key failed",
									IDS_BUILDROUTE_AIPSIGNINGKEY,
									names[i]);
			}
		}
	}
	else
	{
		throw RZkException(	m_err, 
							"RRouteManager::VerifyKeysAvailableForAllHops()",
							"frLocalAipGetAipNames() failed",
							IDS_BUILDROUTE_AIPNAME,
							"frLocalAipGetAipNames");
	}

}

/*
When a route expires, and we have to build
a new route, in the context of a to_send_or_not_to_send(), it will _not_ work since we use TCP
to exchange the session key. Since I call InTac::RouteIDisDead() when a route dies, the client should then proceed
to build another, inform the client, or prolong the old route beforehand.
*/

void 
RRouteManager::BuildRouteForNymTraffic(
	WRoute& in_routeAips, 
	size_t in_iNym,
	const int nDuration,
	size_t in_cntAttempt ) throw(RZkException)
{	
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);	// See bug 2123

	ZKASSERT( in_iNym != 0 ) ;

	ZKASSERT( in_routeAips.GetSize() > 0 );

	// This call prepares our structures for the route
	// but doesn't send anything over the wire yet

	ZkACI newRoute = INVALID_ROUTE_ID ;

	newRoute = m_RouteTable.InitNewAuthenticatedRouteForNym( nDuration, in_routeAips, in_iNym ); 
	
	ZKASSERT( newRoute != INVALID_ROUTE_ID ) ;

	CommonCoreRouting( newRoute, in_routeAips, in_cntAttempt ) ;

	// Route creation has been successful

	INymRoute* pNymRoute = NULL ;
	ZkService::GetINymCentral()->GetNym( IID_INYMROUTE, (void**)&pNymRoute, in_iNym ) ;
	
	if ( pNymRoute != NULL )
	{
		pNymRoute->SetSuccessfulRoute( in_routeAips ) ;
	}
}

void 
RRouteManager::BuildRouteToFreedomServer(
	const uint16_t in_DestPort, 
	WRoute& in_route, 
	const int nDuration,
	size_t in_cntAttempt) throw(RZkException)
{	
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);	// See bug 2123

	// This call prepares our structures for the route
	// but doesn't send anything over the wire yet

	ZkACI newRoute = INVALID_ROUTE_ID ;

	newRoute = m_RouteTable.InitNewUnauthenticatedRoute( in_DestPort, nDuration, in_route ) ;
	
	ZKASSERT( newRoute != INVALID_ROUTE_ID ) ;

	CommonCoreRouting( newRoute, in_route, in_cntAttempt );
}

/*
The data will start coming through the new aci, then there will
be a route destroy received on the old route (aci).
*/
void 
RRouteManager::ProlongRoute(
	const unsigned int uOldRouteID, 
	const int nDuration, 
	WRoute& in_route,
	size_t in_cntAttempt )
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);	// See bug 2123

	unsigned int uNewRouteID = INVALID_ROUTE_ID ;

	size_t iNym = m_RouteTable.GetIndexOfNymOwningAuthenticatedRoute( uOldRouteID ) ;

	uNewRouteID = m_RouteTable.InitNewAuthenticatedRouteWithSameExitAip( 
						uOldRouteID,
						iNym,
						nDuration, 
						in_route ) ;

	if ( uNewRouteID == INVALID_ROUTE_ID )
	{
		throw RZkException(m_err, "RRouteManager::ProlongRoute()",
							"Unable to create a new authenticated route to replace existing one",
							IDS_UNEXPECTED_FAIL);
	}

	CommonCoreRouting( uNewRouteID, in_route, in_cntAttempt, true ) ;

	// Mark the old route so that we don't use it anymore.  It will 
	// be destroyed soon
	m_RouteTable.RetireRoute( uOldRouteID ) ;
}

unsigned long
RRouteManager::GetAckWait( size_t in_cntAttempt )
{
	unsigned long dwAckWait = TIME_TO_WAIT_FOR_RTE_CRT_ACK;

	IZkRegistry::GetDWORD( "AckWait", dwAckWait) ;

	ZKASSERT( dwAckWait > 0 ) ;
	ZKASSERT( in_cntAttempt > 0 ) ;

	// The later the retries, the longer we will wait for the ack
	unsigned long factor = 1 ;
	for ( size_t i = 1 ; i < in_cntAttempt ; i++ )
	{
		factor *= m_RouteRetryFactor ;
	}
	
	dwAckWait *= factor ;

	IZkConsoleMsg::ShowMsg( CONS_DEBUG, 
						CONSIMAGE_COMP_TERM,
						IDS_CONS_ROUTE_TIMEOUT,
						dwAckWait ) ;

	return dwAckWait ;
}

void 
RRouteManager::CommonCoreRouting(
	unsigned int  in_nRouteID,
	const WRoute& in_routeAips, 
	size_t        in_cntAttempt,
	bool          in_bProlongExisting )
{
	try
	{
		ZKASSERT( m_RouteTable.GetRouteState( in_nRouteID ) == RouteValidNotActive ) ;

		ZkService::GetINetworkPerformance()->IncrementRouteCreateAttempts() ;

		VerifyKeysAvailableForAllHops( in_nRouteID ) ;

		// Remember the exact route ID that we are trying to create so that
		// we don't get signalled for the wrong one
		ZkACI OutgoingAci = 0 ;
		ZKVERIFY( frLocalAipGetOutgoingAci(&m_err, m_Context, in_nRouteID, &OutgoingAci )) ;
		m_RouteToBeCreatedOutgoingAci = OutgoingAci ;

		// DH Key Exchange 
		DHKeyExchange	dh;
		sockaddr_in aipAddr ;
		m_RouteTable.GetFirstAipAddr( in_nRouteID, aipAddr ) ;

		dh.Exchange(in_nRouteID, aipAddr, m_Context);
		
		// Prepare a route create packet

		char				data[FRPKT_BUF_MAX_SIZE];
		size_t				dlen = FRPKT_BUF_MAX_SIZE;	

		// Will retrieve the keys from the cache
		if(!frLocalAipMakeCreatePkt(&m_err, m_Context, data, &dlen, in_nRouteID, in_bProlongExisting, FRCRYPT_SYMALG_BLOWFISH))
		{
			throw RZkException(m_err, "RRouteManager::CommonCoreRouting()",
								"Freedom was unable to build route creation packet",
								IDS_BUILDROUTE_FAIL);
		}

		if(!frLocalAipLinkEncrypt(&m_err, m_Context, in_nRouteID, FRPKTTYPE_AIP_ROUTE_CREATE, data, dlen))
		{
			throw RZkException(m_err, "RRouteManager::CommonCoreRouting()",
								"Freedom was unable to link encrypt the route create packet",
								IDS_BUILDROUTE_FAIL);
		}

	#ifdef WIN32	
		// Bug #1028
		Sleep(666);
	#else
		// microseconds
		usleep(66);
	#endif	

#ifdef WIN32
		unsigned long start = ZKGetTickCount();
#endif

		// Fine as long as we send out udp traffic, since remember, right now we are in the context of a tcp packet
		// not getting placed on the wire, and if we send out another tcp packet with the same destination address and port
		// we would end up right here.

		unsigned long ackWait = GetAckWait( in_cntAttempt ) ;

		if(RCloudSocket::theRCloudSocket().SendTo(data, dlen, aipAddr))
		{
			if (m_RteCrtAckEvnt.Wait(ackWait))
			{
#ifdef WIN32
				// We've discovered that some route creates are so fast
				// that they seem to happen within the resolution
				// of the GetTickCount() function and we get a zero 
				// time difference.  Hence the +1 fudge.
				unsigned long end = ZKGetTickCount() + 1 ;

				ZKASSERT( end - start  > 0 ) ;
				ZkService::GetINetworkPerformance()->AddRouteCreateTime( end - start ) ;

				ZKTRACE1( "Route create took %d milliseconds", end - start ) ;
#endif

				ZkService::GetINetworkPerformance()->IncrementRouteCreateSuccess() ;

				m_RouteTable.RouteIsActive( in_nRouteID ) ;

                // Set the idle counter
                m_RouteTable.SetRouteIdleCounter( in_nRouteID, m_idleCounterTicks ) ;
				return;
			}
		}

		throw RZkException( "RRouteManager::CommonCoreRouting()",
							"Freedom was unable to create a private route through the Freedom network. "
							"This can be caused by a server in the route not responding fast enough, "
							"or by network problems preventing data from reaching a Freedom server. "
							"If this problem persists, check your network connection or run the "
							"Freedom diagnostics.",
							IDS_BUILDROUTE_FAIL);

	}
	catch(RZkException& ex)
	{
		// Remember that this route failed
		ZkService::GetINetworkPerformance()->AddRouteToBlackList( in_routeAips ) ;

		// Pull out the route if we couldn't create it
		if ( in_nRouteID != INVALID_ROUTE_ID )
		{
			m_RouteTable.DeleteDeadRoute( in_nRouteID );
		}

		// Explain why the failure happened in the console

		IZkConsoleMsg::ShowMsg( CONS_NORMAL, 
								CONSIMAGE_COMP_TERM,
								IDS_CONS_STRING_AND_DETAIL, 
								ex.GetFormattedString().c_str(),
								ex.GetDescription().c_str() );
		throw ex;
	}
}

bool
RRouteManager::AttemptRouteDeathRecovery()
{
	// Because of possible threading issues or delays in processing this
	// call first verify that the situation is still valid

	if ( ZkService::GetINymCentral()->IsCurrentNymUser() )
	{
		return true;
	}
	
	size_t iActiveNym = ZkService::GetINymCentral()->GetActiveNymIndex() ;

	INym* pNym = NULL ;
	ZkService::GetINymCentral()->GetNym( IID_INYM, (void**)&pNym, iActiveNym ) ;

	if (!pNym->DoesNymUseCloud())
	{
		// We don't have routes for free nym, has the selected nym
		// changed underneath us
		ZKASSERT( false ) ;
		return true ;
	}

	if ( INVALID_ROUTE_ID != ZkService::GetIRouteManager()->GetAuthRouteID() )
	{
		// We expect there there is no route for the nym
		ZKASSERT( false ) ;
		return true;
	}

	if ( pNym->IsNymPreferenceSet( PREF_INTERACTIVE_ROUTES ) )
	{
		// User is set up to actively choose his/her own
		// route so we tell them why the wizard is suddenly
		// going to pop up
		IZkPopup::MessageBox( IDS_ROUTE_HAS_BEEN_LOST, ZK_ERROR, OVER_LAST_ACTIVE );
	}

	bool bBuiltRoute = false ;

	try
	{
		// Attempt to rebuild the route
		// If this succeeds existing TCP connections will be broken
		// that is why we normally use "Prolong" functionality instead
		// of this flow of execution

		if ( ZkService::GetIRouteManager()->StartAuthenticatedRoute( iActiveNym ) )
		{
			ZKASSERT( INVALID_ROUTE_ID != ZkService::GetIRouteManager()->GetAuthRouteID() ) ;
			bBuiltRoute = true;
		}
	}
	catch (IZkException&)
	{
	}

	return bBuiltRoute ;
}


// Not called through user intervention, but rather indirectly from our UDPServer thread
void 
RRouteManager::CreatePacketWasReceived(const ZkACI in_OutgoingAci)
{	
	if ( m_RouteToBeCreatedOutgoingAci != in_OutgoingAci )
	{
		ZKCTRACE2( 
			TRACE_ROUTEMANAGER, 
			"ROUTING - We have a route create for the wrong route - ignore it ( %d versus %d )\n",
			m_RouteToBeCreatedOutgoingAci,
			in_OutgoingAci ) ;

		return ;
	}

#ifdef WIN32
	zkThreadYield(&m_err); // Avoid the KB Q173260 problem 
#endif

	m_RteCrtAckEvnt.Set();
}



bool		
RRouteManager::GetExitAipIPForDNS( sockaddr_in & out_addrExitAip )
{
	// DNS will get routed using our Authenticated route
	ZkACI RouteID = ZkService::GetIRouteManager()->GetAuthRouteID() ;

	if ( !m_RouteTable.IsAciValid( RouteID ) )
	{
		return false ;
	}

	m_RouteTable.GetExitAipAddr( RouteID, out_addrExitAip ) ;
	return true;
}


void 
RRouteManager::SetFreedomTime(const time_t utc)
{
	ZKVERIFY(frLocalAipCtxSetFrTime( &m_err, m_Context, utc) );
}


FrPktType
RRouteManager::ProcessIncomingCloudPacket( 
	sockaddr_in & in_AipAddr, 
	char * in_pBuffer, 
	const size_t in_cntBuffer, 
	ZkBuffer& out_DecryptedPacket )
{
	FrPktType	 type;
	unsigned int vers;
	ZkACI		 nAci = INVALID_ROUTE_ID;
	DZkError     derr ;
    WAipData     aipData;

	m_RouteTable.Lock( ZK_READER ) ;

	// Link decryption will return the aci of this route
	// We send it the source address of the packet to aid it in determining the route since there can be more than one route
	// Note also that this will add entropy to both SSLeay's and Yarrow's
	// pool through frPktLinkDecrypt()->zkRandEntropyAdd()
	if(!frLocalAipLinkDecryptUsingAipData(&derr, m_Context, aipData, in_pBuffer, in_cntBuffer, 
		in_AipAddr, &type, &vers, &nAci))
	{
		ZKCTRACE0( TRACE_ROUTEMANAGER, "frLocalAipLinkDecrypt() failed in UDPServer - possibly route does not exist locally\n");
		m_RouteTable.Unlock( ZK_READER ) ;
		return FRPKTTYPE_UNUSED_ ;
	}
			
	switch(type)
	{
	case FRPKTTYPE_AIP_DATA:
		{
			ZKCTRACE0( TRACE_ROUTEMANAGER, "Aip data was received\n");
			// allow data to bubble up the stack

			size_t i = sizeof( out_DecryptedPacket ) ;

			size_t sizDecryptedPacket = in_cntBuffer;

			if(!frLocalAipReadDataPktUsingAipData(	&derr, 
										m_Context, 
                                        aipData,
										nAci,
										in_pBuffer, 
										&sizDecryptedPacket, 
										(char*) &out_DecryptedPacket, 
										&i)){

				ZKCTRACE0( TRACE_ROUTEMANAGER, "Unable to frLocalAipReadDataPkt(), continuing to listen\n");
				m_RouteTable.Unlock( ZK_READER ) ;
				return FRPKTTYPE_UNUSED_;
			}

			ZKCTRACE2( TRACE_ROUTEMANAGER, "After calling frLocalAipReadDataPkt(), the size of the information is [%d] and size of packet is [%d]\n", i, sizDecryptedPacket);
            
            // Update the idle counter for this route.
            m_RouteTable.SetRouteIdleCounter( nAci, m_idleCounterTicks );

			m_RouteTable.Unlock( ZK_READER ) ;
			return FRPKTTYPE_AIP_DATA ;
		}

	case FRPKTTYPE_AIP_ROUTE_CRTACK:
		{
			ZkACI OutgoingAci = 0 ;
			ZKVERIFY( frLocalAipGetOutgoingAci(&derr, m_Context, nAci, &OutgoingAci )) ;

			// This function is necessary to copy the ACI of the exit funnel into our
			// local structures.   This is necessary when we try to prolong a route -
			// it needs to reuse this ACI.
			// Strictly speaking we should be passing the last argument differently depending
			// whether nAci is prolonged or not, however we don't know this.

			if ( !frLocalAipReadCreateAckPktUsingAipData(&derr, m_Context, aipData, nAci, in_pBuffer, in_cntBuffer, 0 ) ) 
			{
				// Discard this Route Create packet, something is wrong

				m_RouteTable.Unlock( ZK_READER ) ;
				return FRPKTTYPE_UNUSED_ ;
			}

			m_RouteTable.Unlock( ZK_READER ) ;

			ZKCTRACE1( TRACE_ROUTEMANAGER, "RRouteManager Create was acked [%d]\n", nAci);
			CreatePacketWasReceived(OutgoingAci);
			return FRPKTTYPE_AIP_ROUTE_CRTACK ;
		}
	case FRPKTTYPE_AIP_ROUTE_DESTROY:
		{
			// Route destroys are really tricky - threads management, reporting
			// to the UI etc all must be handled

			// We are going to need the Write Lock to actually
			// delete the route.  releaseing this lock and
			// taking the Writer lock inside DeleteDeadRoute may
			// require us to wait and have the route deleted by
			// someone else underneath us.
			// To prevent this from causing problems we will remember
			// the actual unique route identifier and make sure
			// we only delete that.

			ZkACI OutgoingAci = 0 ;
			ZKVERIFY( frLocalAipGetOutgoingAci(&derr, m_Context, nAci, &OutgoingAci )) ;

			TraceIncomingRouteDestroyPacketInfo( nAci, OutgoingAci ) ;

			m_RouteTable.Unlock( ZK_READER ) ;

			// Make sure route is removed from table
			m_RouteTable.DeleteDeadRoute( nAci, OutgoingAci, true ) ;

			// REVIEW - this can come back immediately when we are trying to create a route
			// In theory we should detect this case and signal the event - but how can
			// we safely detect that condition and not a route destroy coming back during
			// a different route create????

			// REVIEW - we might delete the wrong route....
		}

		return FRPKTTYPE_AIP_ROUTE_DESTROY ;
	default:
		m_RouteTable.Unlock( ZK_READER ) ;
		ZKCTRACE0( TRACE_ROUTEMANAGER, "Link decrypted something which wasn't a CRTACK, DATA or a DESTROY");
		return FRPKTTYPE_UNUSED_ ;
	}
	
}

void
RRouteManager::TraceIncomingRouteDestroyPacketInfo( ZkACI in_ACI, ZkACI in_OutgoingACI )
{
	// Purely debug info - this function should have no side effects (other than
	// unavoidable timing issues)
	
#if !defined(NDEBUG)

	if ( m_bBuilding )
	{
		// Currently we don't know exactly what is going on in co-relation with 
		// the route building operation, this is the best we can do for the moment
		ZKCTRACE0( 
			TRACE_ROUTEMANAGER, 
			"m_Building Flag is on - route destroy might be coming for a route we are trying to create\n" );		
	}

	ZKCTRACE2( 
		TRACE_ROUTEMANAGER, 
		"RRouteManager Destroy was received [%d], Outgoing Aci\n", 
		in_ACI, 
		in_OutgoingACI );

	
	// Is this route destroy coming in at an expected time, or are we
	// being deleted in advance?  Useful information to see in the trace
	time_t timeExpire = m_RouteTable.GetExpireTime( in_ACI ) ;
	time_t timeNow = 0 ;
	time( &timeNow ) ;

	if ( timeExpire == 0 )
	{
		ZKCTRACE0( 
			TRACE_ROUTEMANAGER, 
			"RRouteManager Destroy was for route that was not yet created - probably AIP is rejecting route creation\n" );
		ZKCTRACE0( 
			TRACE_ROUTEMANAGER, 
			"This can happen if client clock is invalid, aip is screwed up etc.?\n" );
	}
	else if ( timeNow < timeExpire )
	{
		ZKCTRACE1( 
			TRACE_ROUTEMANAGER, 
			"RRouteManager Destroy was for route that hasn't reached expected retire time [difference %d seconds]\n",
			timeExpire - timeNow 
			);				
	}
	else if ( timeNow > timeExpire )
	{
		ZKCTRACE1( 
			TRACE_ROUTEMANAGER, 
			"RRouteManager Destroy was for route that has already reached expected retire time [difference %d seconds]\n",
			timeNow - timeExpire 
			);				
	}
	else
	{
		ZKCTRACE0( 
			TRACE_ROUTEMANAGER, 
			"RRouteManager Destroy arrived exactly the second that route was expiring!\n"
			);	
	}
#endif			

}


bool 
RRouteManager::CreateOutgoingPacket( 
	ZkBuffer& in_Buffer,				// Buffer to encrypt
	char * out_pEncryptedPacket,		// Resulting packet
	size_t& io_cntEncryptedPacket,		// size out result
	sockaddr_in & out_FirstHopAddr )	// address of first hop of route that is suitable for this
{  
	// We lock the route table to prevent routes from disappearing underneath us
	// This should be the only lock per packet of the table
	ZKS::Lock_Guard<RRWLock> gate( m_RouteTable );

	// Determine which route we should send the 
	// packet on
	ZkACI RouteACI = INVALID_ROUTE_ID ;
	uint16_t hboDestPort = ntohs(in_Buffer.Tcp.DstPort);	// Actually might be UDP also

	GetRouteIdForThisDestination( hboDestPort, RouteACI, out_FirstHopAddr );

	if ( RouteACI == INVALID_ROUTE_ID )
	{
		// We don't have a route (yet) for this destination, so
		// we need to block the packet
		return false ;
	}
	
	ZKASSERT( out_pEncryptedPacket != NULL ) ;
	ZKASSERT( io_cntEncryptedPacket >= FRPKT_BUF_MAX_SIZE ) ;

	// Sometimes, the OutputDebugString() call the following line ends up calling
	// from _CrtDbgReport() in dbgrpt.c, seems to freeze...may be related to Q173260 
	// since this happens under NT only
	// ZKCTRACE1( TRACE_ROUTEMANAGER, "Gonna send frLocalAipMakeDataPkt() an Aci of [%d]\n", in_RouteID);

    {
        WAipData aipData;

        if(!frLocalAipMakeDataPktUsingAipData(
			&m_err, 
			m_Context, 
            aipData,
			RouteACI, 
			(char*)(&in_Buffer), 
			ntohs(in_Buffer.Tcp.IP.TotalLength), 
			out_pEncryptedPacket, 
			&io_cntEncryptedPacket))
        {
            // This might fail if the route has been erased underneath us, between the time
            // we determined which route to use and now that we actually try to use it
            ZKCTRACE0( TRACE_ROUTEMANAGER, "frLocalAipMakeDataPkt() failed in RRouteManager::CreateOutgoingPacket()\n");
            return false;
        }
    }
	// Sometimes, the OutputDebugString() call the following line ends up calling
	// from _CrtDbgReport() in dbgrpt.c, seems to freeze...may be related to Q173260 
	// since this happens under NT only
	//ZKCTRACE1( TRACE_ROUTEMANAGER, "After using frLocalAipMakeDataPkt(), size [%d]\n", io_cntEncryptedPacket );

	if(!frLocalAipLinkEncrypt(
			&m_err, 
			m_Context, 
			RouteACI, 
			FRPKTTYPE_AIP_DATA, 
			out_pEncryptedPacket, 
			io_cntEncryptedPacket))
	{
		ZKCTRACE0( TRACE_ROUTEMANAGER, "frLocalAipLinkEncrypt() failed in RRouteManager::CreateOutgoingPacket()\n");
		return false;
	}

    // Update the idle counter for this route.
    m_RouteTable.SetRouteIdleCounter( RouteACI, m_idleCounterTicks );

	return true ;
}



// -1 Couldn't make garbage packet
// -2 RCloudSocket().SendTo() failed
//  1 success
int 
RRouteManager::SendCoverTrafficPacketOnAllRoutes()
{	
	// Cleanup our route table - expired routes can accumulate if we miss the
	// route destory packet.  This happens for every packet, we might be
	// able to do this on a less frequent timer also

	m_RouteTable.FlushExpiredRoutes() ;

	char CoverTrafficBuf[FRPKT_BUF_MAX_SIZE];	
	int nRet = 1;

	for( size_t i = 0 ; i < m_RouteTable.GetMaxRoutes() ; i++ )
	{
		size_t	bufsiz = FRPKT_BUF_MAX_SIZE;

		// Make sure the route doesn't disappear 
		// underneath us

		m_RouteTable.Lock( ZK_READER ) ;

		if ( !m_RouteTable.IsRouteActive( i ) )
		{
			m_RouteTable.Unlock( ZK_READER ) ;
			continue ;
		}

        if(!frLocalAipMakeKeepAlivePkt(&m_err, m_Context, CoverTrafficBuf, &bufsiz))
        {
            m_RouteTable.Unlock( ZK_READER ) ;
            ZKASSERT(false);
            continue ;
        }	
        
		sockaddr_in firstAipAddr ;
		m_RouteTable.GetFirstAipAddr( i, firstAipAddr ) ;

		m_RouteTable.Unlock( ZK_READER ) ;

		if(RCloudSocket::theRCloudSocket().SendTo(CoverTrafficBuf, bufsiz, firstAipAddr))
		{
#if !defined(NDEBUG)
			char Buffer[100];
			unsigned long nSize = sizeof Buffer;

			snprintf( Buffer, nSize, "%s:%hu", inet_ntoa( firstAipAddr.sin_addr ), ntohs(firstAipAddr.sin_port) ) ;
			ZKCTRACE2( TRACE_ROUTEMANAGER, "RCloudSocket::SendTo() worked cover traffic on route [%d] towards [%s]\n", i, Buffer);
#endif
		}
		else
		{
#if !defined(NDEBUG)
			char Buffer[100];
			unsigned long nSize = sizeof Buffer;
			snprintf( Buffer, nSize, "%s:%hu", inet_ntoa( firstAipAddr.sin_addr ), ntohs(firstAipAddr.sin_port) ) ;

			ZKCTRACE2( TRACE_ROUTEMANAGER, "RCloudSocket::SendTo() failed to work cover traffic on route [%d]  [%s]\n", i, Buffer);
#endif
			// REVIEW REVIEW - shoudln't we consider closing this route and building another one?

			nRet = -2;
		}
	}
	return nRet;
}

// The default cover traffic is one garbage packet per route per second
// On linux they we use a timer
#ifdef _WIN32
void * ZK_THREAD_CALLBACK 
RRouteManager::BackgroundRouteKeepAliveThread(void * in_pParam)
{	

	do
	{
		if ( m_Win32CoverTrafficHandle == NULL )
		{
			// Probably a threading problem - object has gone away without us
			ZKASSERT( false ) ;
			return NULL ;
		}

		const DWORD ret = WaitForSingleObject(m_Win32CoverTrafficHandle->event, COVER_TRAFFIC_INTERVAL);

		if(WAIT_TIMEOUT != ret)
		{
			return NULL;
		}

		try
		{
			ZkService::GetIRouteManager()->SendCoverTrafficPacketOnAllRoutes( ) ;
		}
		catch( IZkException& )
		{
			// We don't expect any throws, in fact
			// there may be unreleased locks in this case.
			ZKASSERT( false ) ;
		}
        // Invoke the route idle timer function to shutdown routes that have had no traffic sent on them.
        RRouteManager::RouteIdleTimer(NULL, 0, 0, 0); 
            
	}while(1);

	return NULL;
}
#endif

size_t
RRouteManager::CalcCoreAipRouteToServer( 
	const uint16_t in_DestPort, 
	size_t in_cntRoutesToCalculate, 
	WRouteVec& out_routes )
{	
	// Destination should be the actual port of the server
	ZKASSERT( in_DestPort != FR_DFLT_AIP_UDP_PORT ) ; 

	ZKCTRACE0( 
		TRACE_ROUTEMANAGER, 
		"Building unauthenticated route to Server via Core Aip\n" ) ;

	WRouteRules rules ;

	rules.SetHopCount( m_cntServerHops ) ;

	// Tell rules that we want to use a core aip as the exit
	rules.SetCoreAipAsExitRule( true ) ;

	// This will calculate as many candidate routes as it can, up to "in_cntRoutesToCalculate"	
	ZkService::GetIRoutePicker()->FindRoutes( in_cntRoutesToCalculate, rules, out_routes ) ;

	return rules.GetHopCount() ;
}

size_t
RRouteManager::CalcRoutesForExtendingExistingNymRoute( size_t in_cntRoutesToCalculate, WRouteVec& out_routes )
{
	out_routes.clear() ;
	
	INym* pNym = NULL ;		
	ZkService::GetINymCentral()->GetActiveNym( IID_INYM, (void**)&pNym );

	size_t nHops = pNym->GetNumberOfHops() ;

	if ( pNym->IsNymPreferenceSet( PREF_INTERACTIVE_ROUTES ) )
	{	
		// User hand picked the route so we can't change it at all
		WRoute activeNymRoute ;

		m_RouteTable.GetAipIds( m_RouteTable.GetActiveAuthenticatedRoute(), activeNymRoute ) ;
		out_routes.push_back( activeNymRoute ) ;	
	}
	else
	{

		WRouteRules rules ;

		rules.SetHopCount( nHops ) ;

		// Prolonged routes have the same exit funnel

		SOCKADDR_IN	addr;
		m_RouteTable.GetExitAipAddr(
					m_RouteTable.GetActiveAuthenticatedRoute(),
					addr) ;
		
		const uint32_t exit_id = AipIdFromIpAndPort(
											addr.sin_addr.s_addr, 
											htons( addr.sin_port ) );

		ZKCTRACE1( TRACE_ROUTEMANAGER, "The exit funnel is [%d]\n", exit_id );
		ZKASSERT( exit_id != 0 ) ;

		rules.SetLastHop( exit_id ) ;

		ZkService::GetIRoutePicker()->FindRoutes( in_cntRoutesToCalculate, rules, out_routes ) ;

		ZKCTRACE1( 
			TRACE_ROUTEMANAGER, 
			"We've applied with the nym rules and will attempt"
			"to prolong with [%d] different routes\n", out_routes.size());
	}

	return nHops ;
}

size_t
RRouteManager::CalcRoutesForNym( size_t in_cntRoutesToCalculate, WRouteVec& out_routes )
{
	INym* pNym = NULL ;
	
	bool bAppliedRule = false ;	
	
	out_routes.clear() ;

	ZkService::GetINymCentral()->GetActiveNym( IID_INYM, (void**)&pNym );

	// This object will be filled with the nyms rules
	WRouteRules rules ;

	size_t nHops = pNym->GetNumberOfHops() ;
	rules.SetHopCount( nHops ) ;


	// See if we want to use a specific route

	WRoute selRoute ;

	if ( pNym->IsNymPreferenceSet( PREF_INTERACTIVE_ROUTES ) )
	{
		// Launch UI now to get the rules

		if ( !ZKRouteDialogs::GetInteractiveRouteRules( rules, selRoute ) )
		{	
			// User has cancelled route selection at this point
			return nHops ;
		}

		// It is legal for GetInteractiveRouteRules() to succeed but
		// return empty route - this means that we let RoutePicker find a route
		if ( selRoute.GetSize() == nHops )
		{
			out_routes.push_back( selRoute ) ;
			return nHops ;
		}
	}

	ZKASSERT ( out_routes.size() == 0 ) ;

	// Use the Route Picker to find some routes
	
	ZkService::GetIRoutePicker()->FindRoutes( in_cntRoutesToCalculate, rules, out_routes ) ;

	ZKCTRACE1( 
		TRACE_ROUTEMANAGER, 
		"We've applied with the nym rules and found [%d] routes to try\n", out_routes.size());

	// Normally we don't show any UI from ZkService - however in this
	// case it is hard to pass back the exact context of why the route failed
	// TODO - use the exception mechanism instead
	if ( out_routes.size() == 0 )
	{
		IZkPopup::MessageBox( IDS_CANT_APPLY_RULE, ZK_MESSAGE, OVER_LAST_ACTIVE );
	}

	return nHops ;
}

void
RRouteManager::SimulateAuthenticatedRouteDestroy()
{
	// Test code to simulate what we do when we get a route destory
	// for the nym route
	
	DZkError     derr ;
	ZkACI		 nAci = INVALID_ROUTE_ID ;

	m_RouteTable.Lock( ZK_READER ) ;
	
	nAci = GetAuthRouteID();

	ZKASSERT( nAci != INVALID_ROUTE_ID ) ;

	ZkACI OutgoingAci = 0 ;
	ZKVERIFY( frLocalAipGetOutgoingAci(&derr, m_Context, nAci, &OutgoingAci )) ;

	m_RouteTable.Unlock( ZK_READER ) ;

	// Make sure route is removed from table
	m_RouteTable.DeleteDeadRoute( nAci, OutgoingAci, true ) ;
}

// Returns 1 on success
int
RRouteManager::GetNymInformation(DZkError& out_Err, const ZkACI in_Route, char* out_nymName, FrCryptSig& out_key)
{
	ZKASSERT(m_Context);
	ZKASSERT(out_nymName);

	m_RouteTable.Lock( ZK_READER ) ;
	const int ret = m_RouteTable.GetNymInformation(out_Err, in_Route, out_nymName, out_key);
	m_RouteTable.Unlock( ZK_READER ) ;
	return ret;
}


// Returns 1 on success
int
RRouteManager::GetAipSymKeys(DZkError& out_Err, const ZkACI in_nRouteId, size_t& out_nAipsInRoute, FrCryptSym* out_symkeys)
{
	ZKASSERT(m_Context);
	ZKASSERT(out_symkeys);

	return frLocalAipGetRouteSymKeys(&out_Err, m_Context, in_nRouteId, &out_nAipsInRoute, out_symkeys);
}


// Returns 1 on success
int
RRouteManager::GetFirstHopInRoutePublicSigningKey(DZkError& out_Err, const ZkACI in_Route, FrCryptSig& out_key)
{
	ZKASSERT(m_Context);

	return frLocalAipGetFirstHopPubSignKey(&out_Err, m_Context, in_Route, &out_key);
}


unsigned long RRouteManager::UpdateIdleCounter()
{
    return ++m_idleCounterTicks;
}

void RRouteManager::KillIdleRoutes()
{
    ZKSTD::vector< ZkACI > aciIdleVec;

	ZKCTRACE1( TRACE_ROUTEMANAGER,  "Harvesting idle routes. Tick count %lu\n", m_idleCounterTicks ) ;

    // Boundary case, initially we haven't been around m_routeIdleTicks
    // timing intervals yet, so the subtraction below would do wierd things
    if (m_idleCounterTicks < m_routeIdleTicks ) return;

    // This should fill in the vector with all of our idle routes.
	m_RouteTable.Lock( ZK_READER ) ;
    m_RouteTable.GetIdleRoutes( m_idleCounterTicks - m_routeIdleTicks, aciIdleVec);
	m_RouteTable.Unlock( ZK_READER ) ;

    // Now we iterate through the vector bringing down each route.
    for (
        ZKSTD::vector< ZkACI >::iterator it = aciIdleVec.begin();
        it != aciIdleVec.end(); 
        ++it
    ) {
        KillSpecificRoute(*it);
    }
}
