/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef _RROUTEMANAGER_H
#define _RROUTEMANAGER_H

#include "IRouteManager.h"
#include "IRoutePacket.h"

#include "RActiveRouteTable.h"

#include <zkThread.h>

#include <frLocalAip.h>

#include <map>
#include <string>

#include <IRoutePacket.h>
#include <WSyncTypes.h>
#include <ErrorHandling.h>
#include <WCoreTypes.h>
#include <zkPacket.h>		// for ZkBuffer et alia
#include <WRoute.h>

class RRouteManager : public IRouteManager, public IRoutePacket
{
public:
	RRouteManager();	
	virtual ~RRouteManager();


	//
	// IRouteManager Interface Methods
	//

	void				Init() ;
	void				Terminate() ;
	void				SetFreedomTime(const time_t in_utc) ;

	int					StartAuthenticatedRoute( size_t in_NymIndex );
	bool				StartUnauthRoute(const uint16_t ) throw();	

	void				TerminateUnauthRoute(const uint16_t in_uPort);
	void				KillAuthenticatedRoute();

	void				TerminateMailRoutes();

	virtual bool		AttemptRouteDeathRecovery() ;

	virtual ZkACI		GetAuthRouteID() ;
	void				GetAuthenticatedRoutes(Port2AipIds&, size_t&);
	bool				GetUnauthenticatedRoute(const uint16_t, unsigned int*);
	void				GetUnauthenticatedRoutes(Port2AipIds&, size_t&);


	// This also builds the unauth route for you if it doesn't exist
	bool				GetRouteIdForThisDestination(
							const uint16_t in_DestPort, 
							unsigned int& out_nRouteID,
							sockaddr_in & out_FirstHopAddr ) ;  

	bool				IsCloudTrafficAllowed(const u_long in_IP);
	bool				IsDiffieHellmanTrafficAllowed( const u_long in_IP ) ;

	ZKSTD::string		GetRouteDesc( const WRoute & in_routeToDescribe ) throw();	
	ZKSTD::string		GetRouteDesc( const FrEntIdType& in_entType ) throw();

	// Test code
	virtual void		SimulateAuthenticatedRouteDestroy() ;

	// Methods for our own timers
	void				ProlongNymRoute() ;
	virtual int			SendCoverTrafficPacketOnAllRoutes();


	//
	// IPacket Interface
	//

	IRoutePacket*       GetRoutePacketHandler() ;

	virtual bool		GetExitAipIPForDNS( sockaddr_in & out_addrExitAip ) ;

	virtual FrPktType	ProcessIncomingCloudPacket( 
							sockaddr_in & in_AipAddr, 
							char * in_pBuffer, 
							const size_t in_cntBuffer, 
							ZkBuffer& out_DecryptedPacket ) ;

	// Returns false is something has gone wrong that
	// doesn't allow us to send the packet
    virtual bool		CreateOutgoingPacket( 
							ZkBuffer& in_Buffer,
							char * out_pEncryptedPacket,
							size_t& io_cntEncryptedPacket,
							sockaddr_in & out_FirstHopAddr ) ;
private:


	void	BuildRouteForNymTraffic(
					WRoute& in_routeAips, 
					IN size_t in_iNym,
					IN const int nDuration,
					size_t in_cntAttempt )
						throw(RZkException);

	void	BuildRouteToFreedomServer(
				const uint16_t in_DestPort, 
				WRoute& in_route,
				const int nDuration,
				size_t in_cntAttempt ) throw(RZkException) ;

	//	The data will start coming through the new aci, then there will
	//	be a route destroy received on the old route (aci).
	//	Perhaps oldnumber of hops != nHops since this is a new route, with only the exit funnel in common
	void	ProlongRoute(
				const unsigned int uOldRouteID, 
				const int nDuration, 
				WRoute& in_route, 
				size_t in_cntAttempt ) throw(RZkException);

	// This is the master function for building a route.
	// The route is already created in the local table, but
	// this function was be called to make the necessary communications
	// with the first Aip in the route in order to create the route.
	// PROMISE: Route is created.	Will throw if anything goes wrong,
	// in which case the route will have been cleaned up by CommonCoreRouting
	// REQUIRES: Route has been created in the table.  Supports both
	// authenticated routes, unauthenticated routes.
	void	CommonCoreRouting( 
				unsigned int in_nRoute, 
				const WRoute& in_routeAips, 
				size_t in_cntAttempt,
				bool in_bProlongExisting = false ) throw(RZkException);


	// Removes a route, no matter what state it is in
	// Takes care of tearing down the route if necessary
	void	KillSpecificRoute(const unsigned int in_ACI);

	// Sends out a destroy packet, does not remove the route information
	// locally immediately
	bool	ShutdownRoute( ZkACI in_RouteID ) ;

	// Close any open routes
	void	KillAllRoutes();

	// This is called when the client recieves a route creation ack
	// from the first hop
	void	CreatePacketWasReceived(const ZkACI aci);

	int		AddStartAip(IN const int /*Aip ID*/) ;
	int		AddRule(const RULEZ, const int);

	size_t	CalcRoutesForNym( size_t in_cntRoutesToCalculate, WRouteVec& out_routes ) ;
	size_t  CalcRoutesForExtendingExistingNymRoute( size_t in_cntRoutesToCalculate, WRouteVec& out_routes ) ;

	size_t CalcCoreAipRouteToServer( 
				const uint16_t in_DestPort, 
				size_t in_cntRoutesToCalculate, 
				WRouteVec& out_routes ) ;

	// Usefull if you have nym's with different rules, simply apply rules prior to creating a route for a nym
	// then clear them before applying another set of rules for another nym
	// Returns 1 if successfull
	// 0 otherwise
	int		ClearAllRules();

	// When you apply rules you may obtain many routes applicable.
	// To get the Aips in a route use this function. Determine which route you want and send that to BuildRoute()
	// Returns 1 if successfull
	// 0 otherwise

	int		GetRouteInfo(const int nRoute, uint32_t aips[MAX_AIPS_IN_ROUTE], size_t& nHops);

	// Add a rule that specifies which starting AIP the nym is interested in using
	// Returns 1 if successfull
	// 0 otherwise

	int AddExitAIP(IN const int /*Aip ID*/) ;
	// Add a rule that specifies which exit AIP the nym is interested in using
	// Returns 1 if successfull
	// 0 otherwise

	int AddAvoidAIP(IN const int /*Aip ID*/) ;

	void		GetNymRulesButDontApply() ;

	void		VerifyKeysAvailableForAllHops( unsigned int in_nRouteId ) throw( RZkException );

	static void * ZK_THREAD_CALLBACK 
	BackgroundRouteKeepAliveThread(void * in_pParam);

	static void ZK_ITIMER_CALLBACK 
	BackgroundRouteKeepAliveTimer(
					void *p, 
					unsigned int i1, 
					unsigned int i2, 
					unsigned long time) ;

    // This timer function is called periodically to mark routes that are idle so that
    // we can determine when it's safe to shut them down. Presently this is called from
    // BacgroundKeepAliveTimer every 30 secs.
	static void ZK_ITIMER_CALLBACK 
	RouteIdleTimer(
        void *p, 
        unsigned int i1, 
        unsigned int i2, 
        unsigned long time) ;

	bool
	StartCoverTrafficTimer(const unsigned long dwMil);

	void
	KillCoverTrafficTimer();

	// Called to manage Route Life Time
	static void ZK_ITIMER_CALLBACK RouteProlongTimerCallback(
					void *p, 
					unsigned int i1, 
					unsigned int i2, 
					unsigned long time) ;

	// Thread to take care of building Unauthenticated routes
	static void * ZK_THREAD_CALLBACK BuildRouteThread(void * pThis);

	void		KillRouteProlongTimer() ;
	void        StartRouteProlongTimer( uint32_t in_msTime ) ;

	// Firewall functionality
	size_t		DetermineWhichRoute2UseForNym(const size_t);
	uint32_t	AipIdFromIpAndPort(const u_long, const uint16_t);

	RouteType	GetRouteTypeFromDestPort(const uint16_t);

	RRouteManager& operator=(const RRouteManager&);

	// Handles waiting until m_bBuilding is not set, will throw
	// if this does not happen in a reasonable amount of time
	void MakeSureRouteManagerNotBusy() throw(RZkException) ;

	unsigned long GetAckWait( size_t in_cntAttempt ) ;

	void TraceIncomingRouteDestroyPacketInfo( ZkACI in_ACI, ZkACI in_OutgoingACI ) ;

	// Returns 1 on success
	int GetNymInformation(DZkError& out_Err, const ZkACI in_Route, char* out_nymName, FrCryptSig& out_key);

	// Returns 1 on success
	int GetAipSymKeys(DZkError& out_Err, const ZkACI in_nRouteId, size_t& out_nAipsInRoute, FrCryptSym* out_symkeys);

	// Returns 1 on success
	int GetFirstHopInRoutePublicSigningKey(DZkError& out_Err, const ZkACI in_Route, FrCryptSig& out_key);

    // Call periodically to increment the m_IdleCounterTicks member so we can find idle routes
    // returns the current counter value.
    virtual unsigned long UpdateIdleCounter();

    // Call to kill any routes who have been idle for ROUTE_IDLE_TICKS ticks 
    // (a tick == a call to UpdateIdleCounter)
    virtual void KillIdleRoutes();

private:

	// ctor member init
	ZkITimer				m_pRouteProlongTimer ;
	ZkITimer				m_pBackgroundRouteKeepAliveTimer;
	ZkThread				m_pBackgroundRouteKeepAliveThread;  // Cover traffic thread (keep alive)
	ZkThread				m_pRouteCreateThread ;
	unsigned long			m_durationNymRoute ;	// Duration to create authenticated routes for
	unsigned long			m_durationServerRoute ;	// Duration to create routes for
	unsigned long			m_timeProlongRoute ;	// Time in advance of route death that 
													// we start trying to prolong
	unsigned long           m_MaxNumRouteRetries ;
	unsigned long			m_RouteRetryFactor ;
	unsigned long			m_cntServerHops ;		// Number of hops to user for routes to core servers
	FrLocalAipCtx			m_Context;				// One Context for N routes, discussion with Kent Tse, 3 August 1998
	WZkEvent				m_RteCrtAckEvnt;		// Communication of route creation ack between threads
	WZkMutex				m_cs;			
	uint16_t				m_DestPortForRouteBeingCreated ;
	ZkACI					m_RouteToBeCreatedOutgoingAci ;  // We need to track the actual unique value to
															// prevent the wrong route create from triggering
															// us
	static ZkEvent			m_Win32CoverTrafficHandle;
	static bool				m_bBuilding;		   // Only one route create can be
												   // launched at a time, so this bool is very important
												   // to prevent deadlock
	mutable DZkError		m_err ;				   
	RActiveRouteTable		m_RouteTable;

    unsigned long           m_routeIdleTicks;
    unsigned long           m_idleCounterTicks;   // counter used to determine when routes are idle
};


#endif
