// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "StdAfx.h"

#include "RRoutePicker.h"
#include <clientdebug.h>

#include <INetworkPerformance.h>
#include <INiqs.h>
#include <ZkService.h>
#include <zkrand.h>


#include <IZkRegistry.h>


#include <TraceControl.h> // for TRACE_ROUTEPICKER_VERBOSE, TRACE_ROUTEPICKER

RRoutePicker::RRoutePicker()
	: m_cntHops( 0 )
	, m_cntMaxFirstHopsInAList( NUMBER_DISTINCT_FIRST_HOPS_IN_ALIST )
	, m_cntFirstHopsInAList( 0 ) 
	, m_cntBListRoutesToCollect( DEFAULT_BLIST_ROUTES_TO_COLLECT )
	, m_cntExaminedRoutes( 0 )
	, m_minAListRating( DEFAULT_MIN_A_LIST_RATING )
	, m_pRules( NULL )
{
	// We allow tweaking of the algorithm with the registry
	// Changing this value is a compromise between performance (only
	// using fast routes and only searching part of the graph) and
	// security ( randomness to the route selection ).  On small
	// networks this doesn't come into play at all
	unsigned long cntFromRegistry = NUMBER_DISTINCT_FIRST_HOPS_IN_ALIST ;
	IZkRegistry::GetDWORD( "RouteFirstHopsInAList", cntFromRegistry) ;

	if ( cntFromRegistry > 0 ) 
	{
		m_cntMaxFirstHopsInAList = cntFromRegistry ;
	}

	cntFromRegistry = DEFAULT_BLIST_ROUTES_TO_COLLECT ;
	IZkRegistry::GetDWORD( "RouteBListSize", cntFromRegistry) ;

	if ( cntFromRegistry > 0 ) 
	{
		m_cntBListRoutesToCollect = cntFromRegistry ;
	}


	unsigned long minFromRegistry = DEFAULT_MIN_A_LIST_RATING ;
	IZkRegistry::GetDWORD( "RouteAListMin", minFromRegistry ) ;

	if ( minFromRegistry < MAX_PERFORMANCE_RESULT )
	{
		m_minAListRating = minFromRegistry ;
	}

	ZKASSERT( m_cntBListRoutesToCollect >= m_cntMaxFirstHopsInAList ) ;
}

RRoutePicker::~RRoutePicker()
{
}

void 
RRoutePicker::FindRoutes( 
	size_t in_cntDesiredRoutes, 
	const WRouteRules & in_RulesToApply,
	WRouteVec& out_FoundRoutes )
{
	// Only a single thread can calculate routes at a time
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);
		
	ZKCTRACE0( TRACE_ROUTEPICKER, "\nRoute Picker is starting calculation\n" ) ;
	ZKASSERT( in_cntDesiredRoutes > 0 ) ;

	// Not strictly fatal, but the maximum number of routes
	// we expect to return is m_cntMaxFirstHopsInAList
//	ZKASSERT( in_cntDesiredRoutes <= m_cntMaxFirstHopsInAList ) ;

	m_cntHops = in_RulesToApply.GetHopCount() ;

	ZKASSERT( m_cntHops > 0 ) ;
	ZKASSERT( m_cntHops <= MAX_AIPS_IN_ROUTE ) ;

	m_pRules = &in_RulesToApply ;
	
	// Clear out any existing state
	m_AListRoutes.clear() ;
	m_BListRoutes.clear() ;
	out_FoundRoutes.clear() ;
	m_cntFirstHopsInAList = 0 ;
	ResetStats() ;

	m_SortedEntityIds.Clear() ;
	ZkService::GetINetworkPerformance()->GetAipIdsSortedByRating( 
					(uint32_t**)&m_SortedEntityIds, 
					m_SortedEntityIds.GetSizeRef() ) ;


	if ( m_SortedEntityIds.Size() == 0 ) 
	{
		// crap, we don't have any aips to deal with - this
		// error can travel up
		ZKTRACE0("Can't find any routes because there ain't no aips\n" ) ;
		return ;
	}

	uint32_t * pEntitiesIds = ( uint32_t * ) m_SortedEntityIds.GetBuffer() ;

	// Traverse the list of aips
	for ( size_t i = 0 ; i < m_SortedEntityIds.Size() ; i++ )
	{
		WRoute route ;

		route.AddHop( pEntitiesIds[ i ] ) ;

		if ( !IsAipSuitable( pEntitiesIds[ i ], m_cntHops == 1 ) )
		{
			continue ;
		}
						
		size_t badIndex = MAX_AIPS_IN_ROUTE ; // Start off with no bad 

		if ( !TraverseGraph( route, badIndex ) )
		{
			break ;
		}
	}

	PadAListWithBList() ;

	if ( in_cntDesiredRoutes > m_AListRoutes.size() )
	{
		ZKCTRACE0( TRACE_ROUTEPICKER,  "Route Picker found less routes than requested." ) ;
		ZKCTRACE2( TRACE_ROUTEPICKER,  
				  "Network is too small or rules too restrictive.  Asked for %d, got %d\n", 
				  in_cntDesiredRoutes, 
				  m_AListRoutes.size() ) ;

		// Copy anything we got into the output

		for ( size_t k = 0 ; k < m_AListRoutes.size() ; k++ )
		{
			// We juggle the ordering in the list just so that
			// in small network conditions we don't always keep
			// trying the same route over and over again.
			// For example on T12 network we were always returning host2, host1, host5
			// then host2, host1, host6, which were the only valid routes.
			// However host5 was actually broken - route create would succeed but
			// no traffic flow.  This meant the client could never contact the core servers!
			// This at least should help it contact 50% of the time.

			uint32_t rn = 0 ;
			ZKVERIFY( zkRandGetUInt32InRange(&m_Error, 0, 1, &rn) );

			if ( rn == 0 )
			{
				out_FoundRoutes.insert( out_FoundRoutes.begin(), m_AListRoutes[k] ) ;
			}
			else
			{
				out_FoundRoutes.push_back( m_AListRoutes[ k ] ) ;
			}
		}
	}
	else
	{
		// Pick random routes to fill into the output
		// REVIEW: Currently this is destructive to m_AListRoutes
		// because that was a fast way to get it running.  If we
		// want to reference the data later it can be done in other ways (eg seperate
		// vector)

		while( out_FoundRoutes.size() < in_cntDesiredRoutes )
		{
			// Proposed random number
			uint32_t rn = 0 ;
			ZKVERIFY( zkRandGetUInt32InRange(&m_Error, 0, m_AListRoutes.size() - 1, &rn) );

			if ( m_AListRoutes[ rn ].GetSize() != 0 )
			{
				out_FoundRoutes.push_back( m_AListRoutes[ rn ] ) ;
				
				// Mark the candidate route to make sure we don't
				// pick it again.
				m_AListRoutes[ rn ].Clear() ;
			}
		}
	}

	// Trace Stats
	TraceStats( in_cntDesiredRoutes, out_FoundRoutes ) ;
}

void
RRoutePicker::PadAListWithBList()
{
	ZKCTRACE1( TRACE_ROUTEPICKER, 
			   "Route A List Size %d\n", 
			   m_AListRoutes.size() ) ;

	ZKCTRACE1( TRACE_ROUTEPICKER, 
			   "Route B List Size %d\n", 
			   m_BListRoutes.size() ) ;

	ZKCTRACE2( TRACE_ROUTEPICKER, 
			   "Maximum first hops in A list was %d, Maximum B list size was %d,\n",
			   m_cntMaxFirstHopsInAList,
			   m_cntBListRoutesToCollect ) ;

	// This is where the BList comes into play - if
	// we didn't find enough AList routes we will promote
	// some BList items into the AList.
	// This scenario is very feasible: if the rules are very strict and force usage
	// of bad routes, if the user is cursed with bad conditions,
	// if the network is small
	// Its hard to know how many B list items to at to the list,
	// if we add too many then we get rid of the advantage of A and B lists.
	// If we take too few then we might decrease the randomness.
	// Some trial and error tweaking might be necessary here

	size_t minAListSize = m_cntMaxFirstHopsInAList ;

	if ( minAListSize > m_AListRoutes.size() )
	{
		size_t cntBadRoutesAdded = 0 ;

		// Fill the array with crappy routes to pad the list

		for ( size_t k = 0 ; k < m_BListRoutes.size() ; k++ )
		{				
			m_AListRoutes.push_back( m_BListRoutes[ k ] ) ;

			cntBadRoutesAdded++ ;

			if ( minAListSize <= m_AListRoutes.size() )
			{
				// We've filled enough
				break ;
			}
		}

		ZKCTRACE1( TRACE_ROUTEPICKER, "A list was padded with %d b-list routes\n", cntBadRoutesAdded ) ;
	}
	else
	{
		ZKCTRACE0( TRACE_ROUTEPICKER, "A List is full, B-List isn't needed\n" ) ;
	}
}

bool
RRoutePicker::IsRouteNewFirstHopInAList( const WRoute& in_Route )
{
	// Determine if the first hop of this route doesn't
	// already exist in the list

	if ( m_AListRoutes.size() == 0 ) 
	{
		return true ;
	}
	else
	{
		// This algorithm assumes we traverse the graph in order of first hops (eg
		// we scan all routes from a particular first hop before going to a different
		// first hop)

		if ( m_AListRoutes[ m_AListRoutes.size() - 1 ].GetHop( 0 ) != in_Route.GetHop( 0 ) )
		{
			return true ;
		}
	}

	return false ;
}

bool
RRoutePicker::TestCandidateRoute( const WRoute& in_Route, size_t& out_FailedIndex )
{
	ZKASSERT( m_cntFirstHopsInAList < m_cntMaxFirstHopsInAList ) ;

	// Only candidates with the correct number of hops should ever be considered
	ZKASSERT( in_Route.GetSize() == m_cntHops ) ;

	m_cntExaminedRoutes++ ;

	// Does it pass the rules?

	ZKASSERT( m_pRules != NULL ) ;

	if ( !m_pRules->DoesRoutePassRules( in_Route, out_FailedIndex ) )
	{
		ZKCTRACE0( TRACE_ROUTEPICKER_VERBOSE, "Candidate Route Failed the test--- " ) ;
		TraceRoute( in_Route, TRACE_ROUTEPICKER_VERBOSE ) ;

		// We tell the algorithm to keep searching
		return true ;
	}

	// Route passes the test, now lets decide what list it belongs on

	if ( IsABListRoute( in_Route ) )
	{
		// This route matches the rules but we don't like it
		// very much.  We'll only use it if we can't find enough
		// routes in the m_AListRoutes in a reasonable amount of
		// time.

		ZKCTRACE0( TRACE_ROUTEPICKER_VERBOSE, "Candidate Route joins the b-team --- " ) ;
		TraceRoute( in_Route, TRACE_ROUTEPICKER_VERBOSE ) ;

		m_BListRoutes.push_back( in_Route ) ;

		// We do stop searching eventually if the b-list gets really large

		return ( m_BListRoutes.size() < m_cntBListRoutesToCollect ) ;
	}
	else
	{
		ZKCTRACE0( TRACE_ROUTEPICKER_VERBOSE, "Candidate Route Passed the test--- " ) ;
		TraceRoute( in_Route, TRACE_ROUTEPICKER_VERBOSE ) ;

		if ( IsRouteNewFirstHopInAList( in_Route ) )
		{
			m_cntFirstHopsInAList++ ;
		}

		m_AListRoutes.push_back( in_Route ) ;
	}

	// We return false to stop the entire traversal of the graph
	// This happens as soon as we have enough different hops
	// Note a little inperfection - we don't traverse all possible routes
	// from the last hop that we consider - just having one is enough to
	// trigger this check.  For example if m_cntMaxFirstHopsInAList is 
	// 10 then there will only be a single route from the 10th first hop
	// that was considered even if more are possible.
		
	return ( m_cntFirstHopsInAList < m_cntMaxFirstHopsInAList ) ;
}

bool  
RRoutePicker::TraverseGraph( WRoute& io_Route, size_t& out_FailedIndex )
{
	bool bRet = true ;	// Be default we tell user to continue the search

//	ZKTRACE0( "Traverse Graph --- " ) ;
//	TraceRoute( io_Route, true ) ;

	size_t cntHops = io_Route.GetSize() ;

	// Must place at least one candidate hop in the route
	ZKASSERT( cntHops > 0 ) ;

	if ( cntHops == m_cntHops )
	{
		// We have a candidate route now,
		// test it
		bRet =  TestCandidateRoute( io_Route, out_FailedIndex ) ;
	}
	else
	{
		// Traverse the connections of the graph

		WNetInfoDatum lastHopSoFar ;

		if ( !ZkService::GetINiqs()->GetEntityById( io_Route.GetHop( cntHops - 1), lastHopSoFar ) )
		{
			// Entity doesn't exist, our topology information must be corrupt
			// Just keep going and hope for the best
			bRet = true ;
		}
		else
		{
			// Traverse the connections from this aip
			size_t cntAipConnections = lastHopSoFar.GetCntConnections() ;

			for ( size_t i = 0 ; i < cntAipConnections ; i++ )
			{
				uint32_t idConnectedAip = lastHopSoFar.GetConnectionX( i ) ;

				if ( io_Route.IsHopInRoute( idConnectedAip ) )
				{
					// be don't want routes to reuse aips
					continue ;
				}

				if ( !IsAipSuitable( idConnectedAip, cntHops + 1 == m_cntHops ) )
				{
					// This aip doesn't match our core aip requirements

					continue ;
				}
				
				io_Route.AddHop( idConnectedAip ) ;

				if ( !TraverseGraph( io_Route, out_FailedIndex ) )
				{
					// We've found enough routes, or have some
					// other reason to abort, stop the search.
					bRet = false ;
					break ;
				}

				if ( out_FailedIndex < cntHops )
				{
					// There is a bad aip at this level or higher, so stop searching
					// the children (but not the total graph traversal)
					bRet = true ;
					break ;
				}
			}
		}		
	}

	// We go up a level so remove the last aip
	io_Route.PopHop() ;
	return bRet ;
}

bool
RRoutePicker::DoesHopMatchCoreAipRestrictions( WNetInfoDatum& in_hopInfo, bool in_bIsLastHop )
{	
	bool bRet = false ; // Assume false until we prove otherwise

	// Handle core aip requirements

	bool bIsCoreAip = in_hopInfo.IsCoreAip() ;

	if ( bIsCoreAip )
	{
		if ( in_bIsLastHop )
		{
			if ( m_pRules->ShouldLastHopBeCoreAip() )
			{
				bRet = true ;
			}
		}
	}
	else
	{
		if ( in_bIsLastHop )
		{
			if ( !m_pRules->ShouldLastHopBeCoreAip() )
			{
				bRet = true ;
			}
		}
		else
		{
			// all the hops that aren't the last should
			// not be a Core Aip
			bRet = true ;
		}
	}

	return bRet ;
}

bool
RRoutePicker::IsAipSuitable( uint32_t in_Aip, bool in_bIsLastHop )
{
	WNetInfoDatum hopInfo ;

	if ( !ZkService::GetINiqs()->GetEntityById( in_Aip, hopInfo ) )
	{
		// Entity doesn't exist, our topology information must be corrupt
		// Just keep going and hope for the best
		return false ;
	}
	else
	{
		// Never accept Down servers
		if ( !hopInfo.IsServerUp() )
		{
			return false ;
		}

		return DoesHopMatchCoreAipRestrictions( hopInfo, in_bIsLastHop ) ;		
	}
}

void
RRoutePicker::ResetStats()
{
	m_cntExaminedRoutes = 0 ;
}


void
RRoutePicker::TraceRoute( const WRoute& in_Route, int in_bTrace ) const
{
	if ( !in_bTrace )
	{
		return ;
	}

	ZKTRACE0( "Route - " ) ;

	for ( size_t i = 0 ; i < in_Route.GetSize() ; i++ )
	{
		uint32_t hop = in_Route.GetHop( i ) ;

		ZKASSERT( hop != 0 ) ;

		ZKTRACE1( "%d ", hop ) ;

		// Calculate the ratings
		uint32_t entityRating = ZkService::GetINetworkPerformance()->GetEntityPerformanceResult( hop ) ;

		ZKTRACE1( "(Rating %d) ", entityRating ) ;
	}

	ZKTRACE0( "\n" ) ;
}

void
RRoutePicker::TraceStats( size_t in_cntDesiredRoutes, const WRouteVec& in_FoundRoutes ) const
{
	ZKCTRACE0( TRACE_ROUTEPICKER, "\n\nRoute Search Results: " ) ;
	ZKCTRACE1( TRACE_ROUTEPICKER,
			  "Number of routes tested %d \n", 
				m_cntExaminedRoutes ) ;

	ZKCTRACE1( TRACE_ROUTEPICKER,
			 "Number of Routes wanted: %d\n", in_cntDesiredRoutes ) ;

	ZKCTRACE1( TRACE_ROUTEPICKER,
			  "Number of Routes returned: %d\n", in_FoundRoutes.size() ) ;

	ZKCTRACE0( TRACE_ROUTEPICKER_VERBOSE, "Discovered Routes returned:\n" ) ;

	for ( size_t i = 0 ; i < in_FoundRoutes.size() ; i++ )
	{
		// Analysis of the results
		ZKCTRACE1( TRACE_ROUTEPICKER_VERBOSE, "Returned Route %d ", i ) ;

		// Show the route
		TraceRoute( in_FoundRoutes[ i ], TRACE_ROUTEPICKER_VERBOSE  ) ;
	}
	ZKCTRACE0( TRACE_ROUTEPICKER_VERBOSE, "\n\n" ) ;

}

bool
RRoutePicker::IsABListRoute( const WRoute& in_Route ) const
{

	// If any of the hops has really poor ping time / performance rating then
	// we don't trust it

	bool bRet = false ;

	for ( size_t i = 0 ; i < in_Route.GetSize() ; i++ )
	{
		uint32_t hop = in_Route.GetHop( i ) ;

		ZKASSERT( hop != 0 ) ;

		// Calculate the ratings
		uint32_t entityRating = ZkService::GetINetworkPerformance()->GetEntityPerformanceResult( hop ) ;

		if ( entityRating < m_minAListRating )
		{
			// This aip sucks

			bRet = true ;
			break ;
		}
	}

	// Now check the aip black list - this means that one of the aips was in a recent failed route create

	if ( !bRet )
	{
		bRet = ZkService::GetINetworkPerformance()->IsRouteBlackListed( in_Route ) ;
	}

	return bRet ;
}


// CheckRules - Currently we are not using MapRules for the user.
// Unless it comes back in the future this is an obsolete feature.
//
// Currently, the user has access to the route rules dialog only when a Nym is selected
// Consequently, some time later that Nym's route will be prolonged, ergo his rules
// will be cleared and re-applied. We could of re-used that FrNetRoute, but imagine the consequences
// when the timer occurs during the period of time where the user in the route rules dialog!
bool 
RRoutePicker::CheckRules(const MapRules& mr, const size_t nHops, size_t& nRoutes) throw ()
{
	ZKASSERT(nHops >= MIN_HOPS && nHops <= MAX_HOPS);

	nRoutes = 0;

	WRouteRules rules( mr ) ;
	rules.SetHopCount( nHops ) ;

	WRouteVec RoutesFound ;

	// To optimize this we can provide a low value for MAX_ROUTES_TO_CALCULATE_FOR_CHECK_RULES
	// eg say "There are at least 10 possible routes..." instead of calculating hundreds
	// of possible routes
	FindRoutes( MAX_ROUTES_TO_CALCULATE_FOR_CHECK_RULES, rules, RoutesFound ) ;

	nRoutes = RoutesFound.size() ;
	return true;
}
