// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// This class takes care of picking
// routes for the client
// See IRoutePicker.h for more documentation

#ifndef RROUTEPICKER_H
#define RROUTEPICKER_H

#include "IRoutePicker.h"

#include <clientdefs.h>
#include <zkIntTypes.h>
#include <DZkError.h>
#include <WAutoBuffer.h>
#include <WNetInfoDatum.h>
#include <WSyncTypes.h>

class RRoutePicker : public IRoutePicker
{
public:

	RRoutePicker() ;
    virtual ~RRoutePicker();

	// Tell the route picker to discover routes 
	virtual void FindRoutes( 
				size_t in_cntDesiredRoutes, 
				const WRouteRules & in_RulesToApply,
				WRouteVec& out_FoundRoutes ) ;

	virtual bool CheckRules(const MapRules& in_Rules, const size_t in_numHops, size_t& out_numRules) throw() ;

	virtual void SetMaxFirstHopsInAList( size_t in_newMax ) { m_cntMaxFirstHopsInAList = in_newMax ; }
	virtual void SetAListMinRating( size_t in_newMin ) { m_minAListRating = in_newMin ; }
	virtual void SetBListMaxSize( size_t in_newMax ) { m_cntBListRoutesToCollect = in_newMax ; }


private:
	// Recursive function for going through the
	// graph finding routes
	//
	// Returns false when we should stop traversing
	// the graph
	// Arguments: io_Route - Contains the route so
	//					 far that we are considering.  It acts as a
	//					 stack
	// out_FailedIndex - this is the index of the aip that failed
	//		the rule test.  We use this to prune the tree for
	//		bad nodes (for example, why consider all routes from
	//		a node that doesn't pass the rules).  Value is MAX_AIPS_IN_ROUTE
	//		if there was no failure

	bool  TraverseGraph( WRoute& io_Route, size_t& out_FailedIndex ) ;

	// Test a possible route. 
	// Return value is _not_ whether the route is good or not, it is
	// whether we should continue to search for routes for not
	bool  TestCandidateRoute( const WRoute& in_Route, size_t& out_FailedIndex ) ;

	// Debugging tool - show a route (with its performance)
	// on the Debug Output
	// Argument defines whether to TRACE or not (for convenience of
	// turning tracing on or off)
	void TraceRoute( const WRoute& in_Route, int in_bTrace ) const ;

	// Show statistics after the route search finished
	void TraceStats( size_t in_cntDesiredRoutes, const WRouteVec& in_FoundRoutes ) const ;

	// Fill in any missing spots on our list of candidate with bad route candidates
	//
	// We collect worse "B" list routes and use them if the "A" list isn't large enough
	// This ensures that we can find at least a crappy route if no good routes exist.
	void PadAListWithBList() ;

	// Contains heuristic for decided whether route belongs in the "B" list.
	// If false is returned then it belongs in the "A" list.
	bool IsABListRoute( const WRoute& in_Route ) const ;

	// We collect some stats for each route collection.  This flushes them
	void ResetStats() ;

	// Returns true if none of the routes in the A list start
	// with the same first hop as the provided candidate route.
	// Note: This is not an expensive calculation because we know that
	// we traverse the topology in the order of first hops.
	bool IsRouteNewFirstHopInAList( const WRoute& in_Route ) ;

	// Determines if we should ignore an AIP or not
	// This function handles two cases: 
	// 1. Down servers
	// 2. Core Aips 
	bool IsAipSuitable( uint32_t in_Aip, bool in_bIsLastHop ) ;

	// Core Aips are a special case - they are connected in the topology but 
	// are only used as the last hop in unauthenticated routes, and never
	// used for Authenticated routes. This function returns true is an Aip
	// passes these Core Aip requirements, false if it does not

	bool DoesHopMatchCoreAipRestrictions( WNetInfoDatum& in_hopInfo, bool in_bIsLastHop ) ;

	// We are building routes of this size
	size_t    m_cntHops ;

	// We collect candidate routes into this list,
	// then transfer some of them into the output list
	WRouteVec m_AListRoutes ;

	// Rather than collecting every single possible route
	// when we only need a few, we collect up to a certain number
	size_t    m_cntMaxFirstHopsInAList ;

	// Count for a particular route traversal of how many routes we have calculated
	size_t    m_cntFirstHopsInAList ;

	// One of the criteria for AList versus BList is
	// that no hops are below a certain rating.
	uint32_t  m_minAListRating ;

	// We also collect crappy routes that match the rules but
	// probably suck.  We only use these if we can't find enough good
	// ones in the candidate route list
	WRouteVec m_BListRoutes ;

	// Max number of BListRoutes that we will collect
	size_t    m_cntBListRoutesToCollect ;

	// Statistics 
	size_t    m_cntExaminedRoutes ;

	const WRouteRules *		   m_pRules ;
	WAutoBuffer				   m_SortedEntityIds ;
	mutable		DZkError	   m_Error ;

	WZkMutex				   m_cs ;
} ;


#endif
