// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RScriptServer.cpp
//
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "RScriptServer.h"
#include <ZkService.h>
#include <ZKFileConfiguration.h>
#include <IWallet.h>
#include <TraceControl.h>
#include <RZkException.h>
#include <string_hlp.h>
#include <IZkRegistry.h>
#include <HttpHelper.h>
#include <sstream>
#include <zkAlloc.h>
#include <IZkPopup.h>
#include <IZkFileHelp.h>
#include <string_hlp.h>
#include <bzlib.h>
#include <IAdBlocker.h>
#include <IZkConsoleMsg.h>
#include <ICrypto.h>
#include <IKeyQry.h>
#include <zkBase64.h>

//////////////////////////////////////////////////////////////////////
// Consts
//

static const char REG_DO_NOT_CHECK_SCRIPTSERVER_SIG[]	= "ScriptServerDoNotCheckSig";
static const char REG_LAST_WALLET_UPDATE[]				= "ScriptServerLastWalletUpdate";
static const char REG_LAST_ADBLOCKER_UPDATE[]			= "ScriptServerLastAdBlockerUpdate";
static const char REG_SCRIPTSERVER_HOSTNAME[]			= "scriptserver";

static const char THIS_WALLET_SCRIPT_UPDATE[]			= "thisScriptUpdate";
static const char THIS_ADBLOCKER_UPDATE[]				= "thisAdUpdate";

static const char DIVIDER_WALLET_SCRIPTS_START[]		= "---ScriptUpdate---\n\n";
static const char DIVIDER_WALLET_SCRIPTS_END[]			= "---EndScriptUpdate---";

static const char DIVIDER_ADBLOCKER_START[]				= "---AdBlockUpdate---\n";
static const char DIVIDER_ADBLOCKER_END[]				= "---EndAdBlockerUpdate---";

static const char DIVIDER_SIGNATURE_START[]				= "\n---Signature---\n";
static const char DIVIDER_SIGNATURE_END[]				= "\n---EndSignature---\n";

static const char WALLET_CONFLICT[]						= "CONFLICT:\"";
static const char WALLET_SUCCESS[]						= "SUCCESS";
static const char WRONG_PASSWORD[]						= "WRONG_PASSWORD";

static const char STR_SCRIPTSERVER_HOST[]				= "scriptserver.freedom.net";

static const char SCRIPTSERVER_ENTITY_NAME[]			= "ScriptServer";

static const char CGI_UPDATE_SINCE[]					= "UpdateSince.cgi";
static const char CGI_ADD_RULE[]						= "AddRule.cgi";
static const char CGI_ADD_SCRIPT[]						= "AddScript.cgi";

static const char PARAM_VALUE_ZIP_IT[]					= "zipIt=true";
static const char PARAM_VALUE_SIGN_IT[]					= "signIt=true";
static const char PARAM_VALUE_OVERWRITE[]				= "overwrite=yes";
static const char PARAM_LAST_AD_UPDATE[]				= "lastAdUpdate";
static const char PARAM_LAST_WALLET_UPDATE[]			= "lastScriptUpdate";
static const char PARAM_NONCE[]							= "nonce";
static const char PARAM_DATA[]							= "data";
static const char PARAM_DOMAIN[]						= "domain";
static const char PARAM_PASSWORD[]						= "password";

#define MAX_MAPPING_LENGTH		128
#define TIME_UNTIL_RETRY		1000 * 60 * 60
#define MIN(x, y) (x)<(y) ? (x):(y)

//////////////////////////////////////////////////////////////////////
// static members and methods
//

void * ZK_THREAD_CALLBACK 
RScriptServer::UpdateAllThread(void * in_pThis)
{
	ZKASSERT(in_pThis);
	RScriptServer* pThis = (RScriptServer*)in_pThis;
	pThis->UpdateAllWithRetries();
	return 0;
}

//////////////////////////////////////////////////////////////////////
// 
//

RScriptServer::RScriptServer()
	: m_pUpdateThread(NULL)
	, m_bUpdateThreadExists(false)
	, m_bAbortUpdate(false)
	, m_strScriptServerHost(STR_SCRIPTSERVER_HOST)
{
#ifdef INTERNAL_BUILD
	// This is a hook so internal people can contact the internal scriptserver
	// used for updating wallet scripts and ad blocker rules 
	ZKSTD::string strScriptServerHost;
	if (IZkRegistry::GetString(REG_SCRIPTSERVER_HOSTNAME, strScriptServerHost))
	{
		if (strScriptServerHost.empty())
		{
			IZkPopup::MessageBox("There is an empty scriptserver key in the registry.\n"
				"Using default scriptserver.");
		}
		else
		{
			m_strScriptServerHost = strScriptServerHost;
		}
	}
#endif //INTERNAL_BUILD

	DZkError err;
	ZKVERIFY(ZK_TRUE == zkEventNew(&err, &m_abortUpdateEvent, true));

	// A little shortcut so erasing the wallet scripts directory will cause a full update of
	// all scripts
	if(	!IZkFileHelp::DoesFileOrDirExist( ZKFileConfiguration::GetWalletScriptsDir().c_str()) )
	{
		IZkRegistry::SetString(REG_LAST_WALLET_UPDATE, "0");
	}
	ZKSTD::string strBlockListPath = ZKFileConfiguration::GetFreedomCommonDataDir();
	strBlockListPath += BLOCK_LIST_FILE_NAME;
	if(	!IZkFileHelp::DoesFileOrDirExist( strBlockListPath.c_str() ) )
	{
		IZkRegistry::SetString(REG_LAST_ADBLOCKER_UPDATE, "0");
	}
}

RScriptServer::~RScriptServer()
{
	ZKASSERT(!m_bUpdateThreadExists);
	AbortIfStillUpdating();
	DZkError err;
	ZKVERIFY(ZK_TRUE == zkEventFree(&err, &m_abortUpdateEvent));
}

bool 
RScriptServer::UpdateAllInNonBlockingThread()
{
	bool bRet = false;

	DZkError err;

	// Do not allow 2 simulaneous update threads
	if (!m_bUpdateThreadExists)
	{
		m_bUpdateThreadExists = true;
		if (ZK_TRUE == zkThreadCreate (&err, &m_pUpdateThread, UpdateAllThread, this))
		{
			bRet = true;
		}
	}
	return bRet;
}

void
RScriptServer::UpdateAllWithRetries()
{
	m_bUpdateThreadExists = true;
	m_bAbortUpdate = false;

	DZkError err;
	ZKVERIFY(ZK_TRUE == zkEventReset(&err, m_abortUpdateEvent));
    bool retval = false;
	
	while (true)
	{
// Wallet only implemented on Windows
        try {
#ifdef WIN32
            retval = Update();
#else
            retval = Update(false, true);
#endif
        } catch ( ... ) {
            retval = false;
        }
        if (retval)
		{
			break;
		}
		else
		{
			// The update failed due to a reason other than Freedom asking it to stop.
			// Try again later
			
			if (ZK_TRUE == zkEventWait(&err, m_abortUpdateEvent, TIME_UNTIL_RETRY))
			{
				break;
			}
		}
    }
	m_bUpdateThreadExists = false;
}

void 
RScriptServer::AbortIfStillUpdating()
{
	DZkError err;
	void *rval;
	m_bAbortUpdate = true;
	
	ZKVERIFY(ZK_TRUE == zkEventSet(&err, m_abortUpdateEvent));
	if (m_bUpdateThreadExists && m_pUpdateThread)
	{
#ifdef WIN32        
		ZKVERIFY(ZK_TRUE == zkThreadJoin(&err, &m_pUpdateThread, &rval)); 
#else
#  warning ScriptServer termination causes problems on linux.
#endif
	}
	m_bAbortUpdate = false;
}

bool
RScriptServer::Update(bool in_bUpdateWalletScripts, bool in_bUpdateAdBlockerRules)
{
	bool bRet = false;
	if (!in_bUpdateWalletScripts && !in_bUpdateAdBlockerRules)
	{
		ZKASSERT(false && "Invalid paramaters");
		return false;
	}
	DZkError err;
	ZKSTD::ostringstream ossRequest;
	ossRequest << m_strScriptServerHost << ':' << FR_DFLT_SCRIPT_SERVER_TCP_PORT
		<< '/' << CGI_UPDATE_SINCE << '?' << PARAM_VALUE_ZIP_IT << '&' << PARAM_VALUE_SIGN_IT;
	
	// Fix this to use a real Nonce
	ZKSTD::string strNonce = GetNonce();
	ossRequest << '&' << PARAM_NONCE << '=' << ZKS::EscapeString(strNonce);

	if (in_bUpdateWalletScripts)
	{
		ZKSTD::string strLastWalletScriptUpdate("0");
		IZkRegistry::GetString(REG_LAST_WALLET_UPDATE, strLastWalletScriptUpdate);
		ossRequest << '&' << PARAM_LAST_WALLET_UPDATE << '=' <<  strLastWalletScriptUpdate;
	}
	if (in_bUpdateAdBlockerRules)
	{
		ZKSTD::string strLastAdBlockerRulesUpdate("0");
		IZkRegistry::GetString(REG_LAST_ADBLOCKER_UPDATE, strLastAdBlockerRulesUpdate);
		ossRequest << '&' << PARAM_LAST_AD_UPDATE << '=' << strLastAdBlockerRulesUpdate;
	}

	ZKSTD::string strReturnedNonce;

	char* strResponse = NULL;
	unsigned int nResponseLen = 0;
	char* strResponseUnzipped = NULL;
	unsigned int nDataLen = 0;

	if (!HttpHelper::GetRequest(&err, ossRequest.str().c_str(), &strResponse, 
		&nResponseLen, &m_bAbortUpdate))
	{
		ZKASSERT(false);
	}
	else if (m_bAbortUpdate)
	{
	}
	else if (nResponseLen == 0)
	{
		ZKASSERT(false);
	}
	else if (!CheckSignature(strResponse, nResponseLen, nDataLen))
	{
		IZkConsoleMsg::ShowImportantMsg(IDS_CONS_SCRIPTSERVER_BAD_SIG);
	}
	else if (!Decompress(strResponse, nDataLen, &strResponseUnzipped))
	{
		ZKASSERT(false);
	}
	else if (!GetValue(strResponseUnzipped, PARAM_NONCE, strReturnedNonce))
	{
		ZKASSERT(false && "No nonce returned");
	}
	else if (strReturnedNonce != strNonce)
	{
		ZKASSERT(false && "Wrong nonce returned");
	}
	else
	{
		bRet = true;
		if (in_bUpdateWalletScripts)
		{
			ZKSTD::string strThisWalletScriptUpdate;
			if (!ProcessWalletScriptsUpdate(strResponseUnzipped))
			{
				ZKASSERT(false);
				bRet = false;
			}
			else if (!GetValue(strResponseUnzipped, THIS_WALLET_SCRIPT_UPDATE, strThisWalletScriptUpdate))
			{
				ZKASSERT(false);
				bRet = false;
			}
			else
			{
				IZkRegistry::SetString(REG_LAST_WALLET_UPDATE, 
					strThisWalletScriptUpdate.c_str());
				IZkConsoleMsg::ShowImportantMsg(IDS_CONS_UPDATED_FORM_FILLER_SCRIPTS);
			}

			if(!bRet)
			{
				IZkConsoleMsg::ShowImportantMsg(IDS_CONS_FORM_FILLER_UPDATE_FAILED);
			}
		}

		if (in_bUpdateAdBlockerRules)
		{

			if(bRet)
			{
				bRet = false;
				ZKSTD::string strThisAdBlockerRulesUpdate;

				if (!ProcessAdBlockerUpdate(strResponseUnzipped))
				{
					ZKASSERT(false);
				}
				else if (!GetValue(strResponseUnzipped, THIS_ADBLOCKER_UPDATE, strThisAdBlockerRulesUpdate))
				{
					ZKASSERT(false);
				}
				else
				{
		
					IZkRegistry::SetString( REG_LAST_ADBLOCKER_UPDATE,
											strThisAdBlockerRulesUpdate.c_str());
					bRet = true;
					IZkConsoleMsg::ShowImportantMsg(IDS_CONS_UPDATED_AD_BLOCKER);
				}
			}
			if(!bRet)
			{
				IZkConsoleMsg::ShowImportantMsg(IDS_CONS_AD_BLOCKER_UPDATE_FAILED);
			}
		}
	}

	if (strResponse)
	{
		zkMemFree(&err, (void**)&strResponse);
	}
	if (strResponseUnzipped)
	{
		zkMemFree(&err, (void**)&strResponseUnzipped);
	}
	if (!bRet)
	{
		IZkConsoleMsg::ShowImportantMsg(IDS_CONS_SCRIPTSERVER_UPDATE_FAILED);
	}
	return bRet;
}

ZKSTD::string
RScriptServer::GetNonce()
{
	ZKSTD::string strRet("No randomness");

	const size_t nBytesOfRandomness = 8;
	unsigned char pRandomData[nBytesOfRandomness] = { 0 };
	unsigned char pBase64edRandomData[(nBytesOfRandomness / 3 + 1) * 4 + 1] = { 0 };

	if (!zkRandGetData(NULL, (void*) pRandomData, nBytesOfRandomness))
	{
		ZKASSERT(false && "Failed to get random data");
	}
	else if (!zkConvertToBase64(NULL, pRandomData, pBase64edRandomData, nBytesOfRandomness))
	{
		ZKASSERT(false && "Failed to Base64 Encode buffer");
	}
	else
	{
		strRet = (char*) pBase64edRandomData;
	}
	return strRet;
}

bool
RScriptServer::CheckSignature(char* in_strDataAndSignature, 
							  const unsigned int in_nDataAndSignatureLen,
							  unsigned int& out_nDataLen)
{
	bool bRet = false;
	out_nDataLen = 0;

	if (!in_strDataAndSignature ||
		(in_nDataAndSignatureLen < 
			(strlen(DIVIDER_SIGNATURE_END) + strlen(DIVIDER_SIGNATURE_START))))
	{
		return false;
	}

	// Can't use strstr because '\0's appear in the buffer.
	char* strSignatureEnd = NULL;
	char* strSignatureStart = NULL;
	for (char* pBuf = in_strDataAndSignature + in_nDataAndSignatureLen - 1;
		pBuf >= in_strDataAndSignature;
		--pBuf)
	{
		if (!strSignatureEnd)
		{
			// Haven't found signature termination yet
			if (!memcmp((const void*)pBuf, 
						(const void*)DIVIDER_SIGNATURE_END,
						strlen(DIVIDER_SIGNATURE_END)))
			{
				strSignatureEnd = pBuf;
			}
		}
		else
		{
			// Have found signature termination, but not signature start
			if (!memcmp((const void*)pBuf, 
						(const void*)DIVIDER_SIGNATURE_START,
						strlen(DIVIDER_SIGNATURE_START)))
			{
				strSignatureStart = pBuf;
				break;
			}
		}
	}

	FrCryptPubKey	scriptServerPubSignKey = NULL;
	DZkError		err;
	int				nIsValidSignature = 0;

	if (!strSignatureEnd || !strSignatureStart || (strSignatureStart >= strSignatureEnd))
	{
		//Signature start or termination tag not found
	}
	else if (IZkRegistry::GetFlag(REG_DO_NOT_CHECK_SCRIPTSERVER_SIG))
	{
		// This is a hook.  On some networks, it is possible a scriptserver key
		// does not exist, so setting this registry value will make Freedom
		// pretend the sig check passes.
		bRet = true;
		out_nDataLen = strSignatureStart - in_strDataAndSignature;
	}
	else if(!ZkService::GetIKeyQry()->GetKey(
			err, 
			FRENT_TYPE_NIQS, 
			3, 
			SCRIPTSERVER_ENTITY_NAME, 
			0, 
			FRCRYPT_KEYTYPE_SIGNING, 
			&scriptServerPubSignKey))
	{
		IZkConsoleMsg::ShowImportantMsg(IDS_CONS_SCRIPTSERVER_KEY_NOT_FOUND);
	}
	else if (!frSecVerifySigBuffer(	&err, 
								scriptServerPubSignKey, 
								strSignatureStart + strlen(DIVIDER_SIGNATURE_START), 
								strSignatureEnd - strSignatureStart - strlen(DIVIDER_SIGNATURE_START),
								in_strDataAndSignature, 
								strSignatureStart - in_strDataAndSignature,
								&nIsValidSignature))
	{
		// ScriptServer signature check failed
	}
	else if (nIsValidSignature != 1)
	{
		// ScriptServer signature check failed
	}
	else
	{
		bRet = true;
		out_nDataLen = strSignatureStart - in_strDataAndSignature;
	}
	return bRet;
}

bool 
RScriptServer::Decompress(char* in_strZipped, 
						  unsigned int in_nZippedLen, 
						  char** out_pStrUnzipped)
{
	if (!out_pStrUnzipped || !in_strZipped || (in_nZippedLen == 0))
	{
		return false;
	}
        
	bool bRet = false;

	bz_stream strm;

	strm.bzalloc = NULL;
	strm.bzfree = NULL;
	strm.opaque = NULL;
	if (BZ_OK != BZ2_bzDecompressInit (&strm, 4, 0))
	{
		ZKASSERT(false);
	}
	else
	{
		const int bufLen = in_nZippedLen * 3;
		char* buf = NULL;
		ZKVERIFY(zkMemAlloc(NULL, (void**)&buf, bufLen + 1));
		
		*out_pStrUnzipped = NULL;

		strm.next_in = in_strZipped;
		strm.avail_in = in_nZippedLen;
		strm.next_out = buf;
		strm.avail_out = bufLen;

		size_t nNewSize = bufLen;
		int ret = 0;
		while(BZ_OK == (ret = BZ2_bzDecompress(&strm)))
		{
			// Our buffer wasn't big enough for the entire unzipped string
			ZKASSERT(strm.avail_out == 0);
			if (!*out_pStrUnzipped)
			{
				zkMemAlloc(NULL, (void**)out_pStrUnzipped, bufLen);
				memcpy(*out_pStrUnzipped, buf, bufLen);
			}
			else
			{
				zkMemAppend(NULL, (void**)out_pStrUnzipped, &nNewSize, (void*)buf, bufLen);
			}
			strm.avail_out = bufLen;
			strm.next_out = buf;
		}
			
		if (ret == BZ_STREAM_END)
		{
			//The decompression was successful
			if (!*out_pStrUnzipped)
			{
				nNewSize = (bufLen - strm.avail_out) + 1;
				zkMemAlloc(NULL, (void**)out_pStrUnzipped, nNewSize);
				memcpy(*out_pStrUnzipped, buf, nNewSize);
			}
			else
			{
				zkMemAppend(NULL, (void**)out_pStrUnzipped, &nNewSize, (void*)buf, 
					(bufLen - strm.avail_out) + 1);
			}

			(*out_pStrUnzipped)[nNewSize - 1] = '\0';
			bRet = true;
		}
		else if (!*out_pStrUnzipped)
		{
			zkMemFree(NULL, (void**)out_pStrUnzipped);
		}
		zkMemFree(NULL, (void**)&buf);

		BZ2_bzDecompressEnd ( &strm );
	}
	
	return bRet;
}

bool
RScriptServer::ProcessWalletScriptsUpdate(const char* in_strResponse)
{
/*	The Wallet Scripts update looks something like:
	---ScriptUpdate---

	yahoo.com

	em	Email
	fn	Name_First

	freedom.net

	cc_expmonth	CCard_ExpDate_Month

	---EndScriptUpdate---
*/
	bool bRet = false;

	char* strWalletUpdateStart = strstr(in_strResponse, DIVIDER_WALLET_SCRIPTS_START);
	char* strWalletUpdateEnd =  strstr(in_strResponse, DIVIDER_WALLET_SCRIPTS_END);

	if (!strWalletUpdateStart || !strWalletUpdateEnd 
		|| (strWalletUpdateEnd < strWalletUpdateStart))
	{
		ZKASSERT(false);
	}
	else
	{
		bRet = true;

		char* strCurPos = strWalletUpdateStart + strlen(DIVIDER_WALLET_SCRIPTS_START);
		while (strCurPos < strWalletUpdateEnd)
		{
			if (m_bAbortUpdate)
			{
				bRet = false;
				break;
			}
			char* strEndDomain = strstr(strCurPos, "\n\n");
			if (!strEndDomain)
			{
				bRet = false;
				ZKASSERT(false); //ScriptServer returned screwy response
				break;
			}
			else
			{
				char *strDomain = new char[strEndDomain - strCurPos + 1];
				memcpy(strDomain, strCurPos, strEndDomain - strCurPos);
				strDomain[strEndDomain - strCurPos] = '\0';

				char* strScriptStart = strEndDomain + strlen("\n\n");
				char* strScriptEnd = NULL;

				if (*strScriptStart == '\n')
				{
					//Special caes -> empty script
					strScriptEnd = strScriptStart;
				}
				else
				{
					strScriptEnd = strstr(strScriptStart, "\n\n");
					if (strScriptEnd)
					{
						++strScriptEnd; 
					}
				}

				if (!strScriptEnd 
					|| !ZkService::GetIWallet()->WriteScriptToDisk(
							strDomain, 
							strScriptStart, 
							strScriptEnd - strScriptStart))
				{
					bRet = false;
					SAFEDELETEARRAY(strDomain);
					ZKASSERT(false);
					break;
				}
				SAFEDELETEARRAY(strDomain);
				strCurPos = strScriptEnd + strlen("\n");
			}
		}
	}
	return bRet;
}
#ifdef INTERNAL_BUILD
bool RScriptServer::UploadAdBlockerRules(const ZKSTD::string& in_strRuleList)
{
#ifdef WIN32
	bool ReturnValue = false;

	if (!PromptForPassword())
	{
	}
	else
	{
		DZkError err;

		ZKSTD::ostringstream ossUrl;
		ossUrl << m_strScriptServerHost << ':' << FR_DFLT_SCRIPT_SERVER_TCP_PORT
			<< '/' << CGI_ADD_RULE;

		ZKSTD::ostringstream ossPost;
		ossPost << "password=" << ZKS::EscapeString(m_strScriptServerPassword)
			<< '&' << PARAM_DATA << '=' << ZKS::EscapeString(in_strRuleList) << "\r\n";
		char *strResponse = NULL;
		size_t nResponseLen = 0;
		if (!HttpHelper::PostRequest(&err, ossUrl.str().c_str(), ossPost.str().c_str(), &strResponse, &nResponseLen))
		{
			ZKASSERT(false);
		}
		else if (strstr(strResponse, WRONG_PASSWORD))
		{
			m_strScriptServerPassword.erase();
			if (IZkPopup::QuestionBox("Incorrect passphrase.\nTry again?", ZK_YESNO))
			{
				ReturnValue = UploadAdBlockerRules(in_strRuleList);
			}
		}
		else
		{
			ReturnValue = true;
		}

		if(strResponse)
		{
			zkMemFree(&err, (void**)&strResponse);
		}
	}
	return ReturnValue;
#else
	// Updating scripts not supported on non windows platforms.
	// We can support it on other platforms be adding a dialog that
	// asks for a password.
	return false;
#endif
}
#endif

#if defined(INTERNAL_BUILD) && defined(WIN32)
bool
RScriptServer::PromptForPassword()
{
	bool bRet = true;

	// The scripts on the real scriptserver are not modifyable.
	// For trained, internal people who want to edit the scripts, you must set a 
	// value in the registry with name "scriptserver" and value hostname. 
	if ((m_strScriptServerHost == STR_SCRIPTSERVER_HOST)
		&& (!IZkPopup::QuestionBox("You cannot update ad blocker rules of form filler scripts on scriptserver.\n"
				"You must specify the scriptupdate server in the registry under the \"scriptserver\" key.\n"
				"See Corey for more info.\n\n"
				"Continue anyway?",
				ZK_YESNO)))
	{
		bRet = false;
	}
	else if (m_strScriptServerPassword.empty())
	{
		m_strScriptServerPassword
			= IZkPopup::RequestStringBox("Modifying ScriptServer",
							"This is only for trained Zero Knowldge professionals.\n"
							"For a password, see Corey.\n"
							"Please enter the ScriptServer password.",
							"");
		if (m_strScriptServerPassword.empty())
		{
			bRet = false;
		}
	}
	return bRet;
}
#endif

#ifdef INTERNAL_BUILD
bool 
RScriptServer::AddScriptMappings(	const ZKSTD::string& in_strDomain, 
									const String2StringMap& in_realName2ZkNameMap, 
									bool in_bOverwrite)
{
#ifdef WIN32
	if (!PromptForPassword())
	{
		return false;
	}

	if (in_realName2ZkNameMap.size() == 0)
	{
		return true;
	}

	bool bRet = false;

	ZKSTD::ostringstream ossNewMappings;
	for (String2StringMap::const_iterator it = in_realName2ZkNameMap.begin();
		it != in_realName2ZkNameMap.end(); ++it)
	{
		if (!ZkService::GetIWallet()->IsMappingInDefaultScript((*it).first, (*it).second))
		{
			ossNewMappings << ZKS::EscapeString((*it).first) << '\t' 
				<< ZKS::EscapeString((*it).second) << '\n';
		}
	}

	DZkError err;
	ZKSTD::ostringstream ossRequest;
	ossRequest << m_strScriptServerHost << ':' << FR_DFLT_SCRIPT_SERVER_TCP_PORT
		<< '/' << CGI_ADD_SCRIPT;

	ZKSTD::ostringstream ossPost;
	ossPost << PARAM_PASSWORD << '=' << ZKS::EscapeString(m_strScriptServerPassword) 
		<< '&' << PARAM_DOMAIN << '=' << ZKS::EscapeString(ZKS::ExtractDomain(in_strDomain))
		<< '&' << PARAM_DATA << '=' << ZKS::EscapeString(ossNewMappings.str());

	if (in_bOverwrite)
	{
		ossPost << '&' << PARAM_VALUE_OVERWRITE;
	}

	char* strResponse = NULL;
	unsigned int nResponseLen = 0;

	if (!HttpHelper::PostRequest(&err, ossRequest.str().c_str(), ossPost.str().c_str(), 
		&strResponse, &nResponseLen))
	{
		ZKASSERT(false);
	}
	else if (strstr(strResponse, WRONG_PASSWORD))
	{
		m_strScriptServerPassword.erase();
		if (IZkPopup::QuestionBox("Incorrect passphrase.\nTry again?", ZK_YESNO))
		{
			bRet = AddScriptMappings(in_strDomain, in_realName2ZkNameMap, in_bOverwrite);
		}
	}
	else if (strstr(strResponse, WALLET_CONFLICT))
	{
		//There was a conflict
		bRet = true;
		DealWithConflicts(in_strDomain, strResponse);	
	}
	else if (strstr(strResponse, WALLET_SUCCESS))
	{
		bRet = true;
	}
	else
	{
		ZKASSERT(false);
	}
	Update(true, false);

	return bRet;
#else
	// Updating scripts not supported on non windows platforms.
	// We can support it on other platforms be adding a dialog that
	// asks for a password.
	return false;
#endif
}

void 
RScriptServer::DealWithConflicts(const ZKSTD::string& in_strDomain, const char* in_strResponse)
{
/*	in_strResponse should look something like:
	CONFLICT:"em\tnew":"em\told"\nCONFLIC...
*/
	
	if (!in_strResponse)
	{
		ZKASSERT(false);
		return;
	}
	char* strNextConflict = strstr(in_strResponse, WALLET_CONFLICT);
	
	String2StringMap realName2ZkNameMap;
	while (strNextConflict)
	{
		char* strFormElemNameStart = NULL;
		char* strFormElemNameEnd = NULL;
		char* strNewZkNameEnd = NULL;
		char* strOldZkNameStartMinus1 = NULL;
		char* strOldZkNameEnd = NULL;
		if (!(strFormElemNameStart = strNextConflict + strlen(WALLET_CONFLICT)))
		{
			ZKASSERT(false);
		}
		else if (!(strFormElemNameEnd = strchr(strFormElemNameStart, '\t')))
		{
			ZKASSERT(false);
		}
		else if (!(strNewZkNameEnd = strchr(strFormElemNameEnd, '\"')))
		{
			ZKASSERT(false);
		}
		else if (!(strOldZkNameStartMinus1 = (strchr(strNewZkNameEnd, '\t'))))
		{
			ZKASSERT(false);
		}
		else if (!(strOldZkNameEnd = strchr(strOldZkNameStartMinus1, '\"')))
		{
			ZKASSERT(false);
		}
		else
		{
			char strFormElemName[MAX_MAPPING_LENGTH + 1] = { 0 };
			char strNewZkName[MAX_MAPPING_LENGTH + 1] = { 0 };
			char strOldZkName[MAX_MAPPING_LENGTH + 1] = { 0 };

			strncpy(strFormElemName, strFormElemNameStart, 
				MIN(strFormElemNameEnd - strFormElemNameStart, MAX_MAPPING_LENGTH));
			strncpy(strNewZkName, strFormElemNameEnd + 1, 
				MIN(strNewZkNameEnd - strFormElemNameEnd - 1, MAX_MAPPING_LENGTH));
			strncpy(strOldZkName, strOldZkNameStartMinus1 + 1,
				MIN(strOldZkNameEnd - strOldZkNameStartMinus1 - 1 , MAX_MAPPING_LENGTH));

			ZKSTD::ostringstream ossQuestion;
			ossQuestion 
				<< "THERE IS A CONFLICT for the " << in_strDomain << " script.\n"
				<< '\"' << ZKS::UnEscapeString(strFormElemName) << '\"' << " is currently mapped to \"" 
				<< ZKS::UnEscapeString(strOldZkName) << "\".\n"
				<< "You say   -> " << ZKS::UnEscapeString(strNewZkName) << ".\n\n"
				<< "YES -> Make the change.  NO -> Leave as is.";
			if (IZkPopup::QuestionBox(ossQuestion.str().c_str(), ZK_YESNO, ZK_QUESTION, true, OVER_CENTER))
			{
				realName2ZkNameMap[ZKS::UnEscapeString(strFormElemName)]
					= ZKS::UnEscapeString(strNewZkName);
			}
		}

		strNextConflict = strstr(strOldZkNameEnd, WALLET_CONFLICT);
	}

	if (realName2ZkNameMap.size() > 0)
	{
		AddScriptMappings(in_strDomain, realName2ZkNameMap, true);
	}
}
#endif //INTERNAL_BUILD

bool
RScriptServer::ProcessAdBlockerUpdate(const char* in_strResponse)
{	
/*	The Ad blocker update looks something like:
	---AdBlockUpdate---
	~doubleclick.net
	@www.doubleclick.net
	!freedom.net
	---EndAdBlockerUpdate---
*/
	bool bRet = true;

	const char* strAdBlockerRulesStart = strstr(in_strResponse, DIVIDER_ADBLOCKER_START);
	const char* strAdBlockerRulesEnd =  strstr(in_strResponse, DIVIDER_ADBLOCKER_END);

	if (!strAdBlockerRulesStart || !strAdBlockerRulesEnd 
		|| (strAdBlockerRulesEnd < strAdBlockerRulesStart))
	{
		ZKASSERT(false);
		bRet = false;
	}
	else
	{
		strAdBlockerRulesStart += strlen(DIVIDER_ADBLOCKER_START);
		const char *strBeginRule = strAdBlockerRulesStart;
		bool bAddedRule = false;
		while(strBeginRule && strBeginRule < strAdBlockerRulesEnd)
		{
			const char *strEndRule = strchr(strBeginRule, '\n');
			if(strEndRule && strEndRule <= strAdBlockerRulesEnd)
			{
				ZKSTD::string strRule(strBeginRule, strEndRule - strBeginRule);

				// false to disable SaveToDisk - do it all at once afterward..
				ZkService::GetIAdBlocker()->AddEntry(ZKS::UnEscapeString(strRule), false);
				bAddedRule = true;
				strBeginRule = strEndRule + 1;
			}
			else
			{
				break;
			}
		}
		if(bAddedRule)
		{
			ZkService::GetIAdBlocker()->SaveToDisk();
		}
	}
	return bRet;
}

bool
RScriptServer::GetValue(const char* in_strResponse, 
						const char* in_strParam, 
						ZKSTD::string& out_strValue)
{
	bool bRet = false;

	char* strValueStart;
		
	strValueStart = strstr(in_strResponse, in_strParam);
	if (!strValueStart)
	{
		ZKASSERT(false);
	}
	else if (strlen(in_strResponse) - (strValueStart - in_strResponse) 
				< strlen(in_strParam) + strlen("="))
	{
		ZKASSERT(false);
	}
	else
	{
		strValueStart += strlen(in_strParam) + strlen("=");
		char* strValueEnd = strchr(strValueStart, '\n');
		if (!strValueEnd)
		{
			ZKASSERT(false);
		}
		else
		{
			out_strValue.clear();
			out_strValue.append(strValueStart, strValueEnd - strValueStart);
			bRet = true;
		}
	}
	return bRet;
}

void 
RScriptServer::SelfTest()
{
#ifdef INTERNAL_BUILD

	ZKSTD::string strNonce1 = GetNonce();
	ZKSTD::string strNonce2 = GetNonce();
	ZKTEST(strNonce1 != strNonce2);

	bool bOldValue = IZkRegistry::GetFlag(REG_DO_NOT_CHECK_SCRIPTSERVER_SIG);
	unsigned int nDataLen = 0;
	char strData1[] = "This data doesn't even contain a signature!";
	char strData2[] = "This data contains \n---Signature---\n, but not the termination";
	char strData3[] = "This data contains \n---EndSignature---\n, but not the start";
	char strData4[] = "This data contains \n---Signature---\n and \n---EndSignature---\n,"
					  "obviously the data isn't signed correctly";

	IZkRegistry::SetFlag(REG_DO_NOT_CHECK_SCRIPTSERVER_SIG, false);
	ZKTEST(!CheckSignature(strData1, strlen(strData1), nDataLen));
	ZKTEST(!CheckSignature(strData2, strlen(strData2), nDataLen));
	ZKTEST(!CheckSignature(strData3, strlen(strData3), nDataLen));
	ZKTEST(!CheckSignature(strData4, strlen(strData4), nDataLen));
					  
	IZkRegistry::SetFlag(REG_DO_NOT_CHECK_SCRIPTSERVER_SIG, true);
	ZKTEST(!CheckSignature(strData1, strlen(strData1), nDataLen));
	ZKTEST(!CheckSignature(strData2, strlen(strData2), nDataLen));
	ZKTEST(!CheckSignature(strData3, strlen(strData3), nDataLen));
	ZKTEST(CheckSignature(strData4, strlen(strData4), nDataLen));

	IZkRegistry::SetFlag(REG_DO_NOT_CHECK_SCRIPTSERVER_SIG, bOldValue);

	ZKTEST(UpdateAllInNonBlockingThread());
	ZKTEST(!UpdateAllInNonBlockingThread());
	AbortIfStillUpdating();

	const char* strResponse= "hfvbajfvjsf---ScriptUpdate---\n\n"
		"testDomain.com\n\n"
		"GARBALDIGOOK\n\n"
		"---EndScriptUpdate---";
	ZKTEST(ProcessWalletScriptsUpdate(strResponse));
	ZKSTD::ifstream ifs(ZKFileConfiguration::GetWalletScriptPath("testDomain.com").c_str());
	if (!ifs)
	{
		ZKTEST(false);
	}
	else
	{
		ZKSTD::ostringstream ossFileContents;
		ossFileContents << ifs.rdbuf();
		ZKTEST(0 == ossFileContents.str().find("GARBALDIGOOK\n"));
		ifs.close();
	}

#ifdef WIN32
	::DeleteFile(ZKFileConfiguration::GetWalletScriptPath("testDomain.com").c_str());
#else
#warning Need to Port DeleteFile call
#endif

	strResponse = "hfvbajfvjsf---ScriptUpdate---\n\n"
		"testDomain.com\n" //Note- should be another \n here
		"param1\tvalue1\n\n"
		"---EndScriptUpdate---";
	ZKTEST(!ProcessWalletScriptsUpdate(strResponse));

	strResponse = "hfvbajfvjsf---ScriptUpdate---\n\n"
		"testDomain.com\n\n"
		"param1value1\n\n" // Note- should be \t between param1 and value1
		"---EndScriptUpd";
	ZKTEST(!ProcessWalletScriptsUpdate(strResponse));

#endif
}
