// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "StdAfx.h"
#include "RTokenDesc.h"

#include <WAutoBuffer.h>
#include <IZkFileHelp.h>
#include <ZKFileConfiguration.h>
#include <ZKConfiguration.h>
#include <IZkConsoleMsg.h>

#include <clientchunkids.h>

#include <ZKDocument.h>
#include <ZKChunk.h>
#include <WReadBuffer.h>

static const std::string g_strFileDesc( "Token Type Descriptions" ) ;


RTokenDesc::RTokenDesc()
	: m_pTokenTypeDescriptions( NULL )
{
	InitFromCache( ZKFileConfiguration::GetTokenDescFilePath().c_str() ) ;
}

RTokenDesc::~RTokenDesc()
{
	// We could save at this point, but because the contents probably
	// only change after a "set" call we do it then instead

	if(m_pTokenTypeDescriptions != NULL)
	{
		ZKVERIFY(frTokenTypeFree(&m_ZkError, &m_pTokenTypeDescriptions)); 
	}
}

void			
RTokenDesc::InitFromCache( const char* in_strFile )
{
	try
	{		
		// Create a document reader.  The constructor will load
		// and parse the top level of the file.  Will throw
		// if file isn't well formated

		ZKDocumentReader dr( in_strFile ) ;

		// This is the network that the user used when saving.
		// Version 2.0 and before did not save this so this may be empty
		ZKSTD::string strNetwork = dr.GetNetworkName() ;

		if ( !strNetwork.empty() && ZKConfiguration::GetNetworkName() != strNetwork )
		{
			// Looks like the user renamed the token cache from a different network - we won't use it.
			ZKASSERT( !"File was persisted with a different network that client is configured for" ) ;
			return ;
		}

		if ( 0 != dr.GetFileDesc().compare( g_strFileDesc ) ) 
		{
			ZKASSERT( !"Wrong kind of file - we expect token description file" ) ;
			return ;
		}

		// Get pointers to the data contents
		const uint8_t * pDataBuffer = NULL ;
		size_t cntDataBuffer = 0 ;

		dr.GetDataBuffer( &pDataBuffer, cntDataBuffer ) ;

		// Create a read buffer to help us parse through the data
		WReadBuffer rb( pDataBuffer, cntDataBuffer ) ;

		// Data buffer is expected to be made up of a TOKEN_CACHE_CHUNK

		ZKChunkReader cr( rb ) ;

		if ( cr.GetChunkID() != TOKEN_CACHE_CHUNK )
		{
			// Corrupt file
			return ;
		}

		const uint8_t * pTokenDescBuffer = NULL ;
		size_t		    cntTokenDescBuffer = 0 ;
		
		cr.ReadArray( (const void**)&pTokenDescBuffer, &cntTokenDescBuffer ) ;

		// Attempt to parse the data into a temporary FrTokenType

		FrTokenType	pTokenTypeDescFromFile = NULL ;

		if ( frTokenTypeFromBinary ( 
					&m_ZkError, 
					&pTokenTypeDescFromFile,
					(void*) pTokenDescBuffer,
					cntTokenDescBuffer ) )
		{
			// We only overwrite any existing descriptions if
			// we see we are successful in parsing the file

			if ( m_pTokenTypeDescriptions != NULL)
			{
				ZKVERIFY(frTokenTypeFree(&m_ZkError, &m_pTokenTypeDescriptions)); 
			}

			ZKASSERT( pTokenTypeDescFromFile != NULL ) ;
			m_pTokenTypeDescriptions = pTokenTypeDescFromFile ;
			pTokenTypeDescFromFile = NULL ;

			ZKTRACE1( "Successfully loaded token description file %s\n", in_strFile ) ;

			IZkConsoleMsg::ShowMsg( 
				CONS_DEBUG, 
				CONSIMAGE_CONNECTION, 
				IDS_CONS_JUST_STRING, 
				"Successfully loaded token descriptions from file" ) ;
		}
		else
		{
			ZKTRACE2( "Token Description File %s was corrupt (%s)\n", in_strFile, m_ZkError.getdetail().c_str() ) ;
		}
	}
	catch( ... )
	{
		// Probably corrupt file
		ZKTRACE2( "Token Description File %s was corrupt (%s)\n", in_strFile, m_ZkError.getdetail().c_str() ) ;
	}
}

void			
RTokenDesc::SaveTokenDescriptions( const char* in_strFile )
{
	try
	{
		ZKTRACE1( "Saving Token Description Cache to %s", in_strFile ) ;
		TraceTokenDesc() ;

		if (( in_strFile == NULL ) ||
			( m_pTokenTypeDescriptions == NULL ))
		{
			// Nothing to save
			return ;
		}

		// We use a Zero knowledge document writer so that we have
		// a consistent top level header with useful information
		// This is not an encrypted file so we use no Salt or other crap

		ZKDocumentWriter DocWriter ;

		DocWriter.SetHeader( g_strFileDesc, 
					 ZKConfiguration::GetFreedomProductName(), 
					 ZKConfiguration::GetFreedomProductVersion(), 
					 ZKConfiguration::GetFreedomPlatform(),
					 ZKConfiguration::GetNetworkName()) ;

		WAutoBuffer tokenDescBuffer ;

		if ( frTokenTypeToBinary( 
				&m_ZkError, 
				m_pTokenTypeDescriptions, 
				(void**)&tokenDescBuffer, 
				tokenDescBuffer.GetSizePointer() ) )
		{
			WMemBuf fileInMemory ;

			{
				// The Data section currently contains 1 chunk, with the contents
				// of the token description.  If we ever wanted to change the
				// format of the file we would probably change the chunk id or chunk version
				// of this chunk

				ZKChunkWriter cw( fileInMemory, TOKEN_CACHE_CHUNK, 1 ) ;
				cw.WriteArray( tokenDescBuffer.GetBuffer(), tokenDescBuffer.Size() ) ;
			}

			unsigned char * pRawBuffer = NULL ;
			size_t cntBuffer = 0 ;

			DZkError err ;

			zkMemBufGetData( &err, fileInMemory, (void**)&pRawBuffer, &cntBuffer ) ;

			ZKTEST( pRawBuffer != NULL ) ; 
			ZKTEST( cntBuffer > 0 )  ; 

			DocWriter.SetData( pRawBuffer, cntBuffer ) ;
			DocWriter.WriteToFile( in_strFile ) ;
		}
		
	}
	catch( ... )
	{

	}
}


void 
RTokenDesc::SetTokenDescriptions( T_FrTokenType* in_pNewTokenDescriptions )
{
	if ( in_pNewTokenDescriptions == NULL )
	{
		// We can support this behavior, but why?
		ZKASSERT( !"Provided with NULL Token Descriptions" ) ;
		return ;
	}

	// Clear any existing content
	if ( m_pTokenTypeDescriptions != NULL)
	{
		ZKVERIFY(frTokenTypeFree(&m_ZkError, &m_pTokenTypeDescriptions)); 
	}

	// Take over the provided data for ourselves
	m_pTokenTypeDescriptions = in_pNewTokenDescriptions ;

	// Here is a good time to save because presumably we have the latest
	// information
	SaveTokenDescriptions( ZKFileConfiguration::GetTokenDescFilePath().c_str() ) ;
}

bool
RTokenDesc::GetTokenDesc( unsigned short in_TypeId, WTokenDescriptions& out_TokenDesc )
{
	// Initialise token description to the default value - even if we return false
	// at least there is something in the structure that can be used.
	out_TokenDesc = WTokenDescriptions();

	if ( m_pTokenTypeDescriptions == NULL )
	{
		ZKASSERT( false ) ;
		return false ;
	}

	// Type lookup
	const FrTokenTypeDescription* pTokenTypeDescription = NULL;
	if (!frTokenTypeLookup (
			&m_ZkError ,			  // Returns the error if any occured 
			m_pTokenTypeDescriptions ,			  // Token type structure 
			in_TypeId ,			  // Desired type ID
			&pTokenTypeDescription ))// Returns a pointer to the description
	{
		// Probably token type is not in the token description structure
		return false ;
	}			
	else 
	{
		out_TokenDesc = WTokenDescriptions( pTokenTypeDescription );
		return false ;
	}

}

void
RTokenDesc::TraceTokenDesc()
{
	if ( m_pTokenTypeDescriptions == NULL )
	{
		ZKTRACE0( "TokenDescriptionTrace - RTokenDesc is empty\n" ) ;
	}

/*
	// Its extremely hard to get a dump of the contents of this
	// data type - there are no calls to iterate through the various
	// descriptions, no way to know how many items there are....

	// This is just a yucky hack to try to show interesting contents
	// while I'm developing....  Its hit an miss what type IDs are valid
	// or even what range they would be it....

	// TODO - Write some access functions in the library underneath to
	// trace out the contents.

	for ( unsigned short typeID = 0 ; typeID < 100 ; typeID++ )
	{
		const FrTokenTypeDescription* pTokenTypeDescription = NULL;

		if ( frTokenTypeLookup( 					
						&m_ZkError, 
						m_pTokenTypeDescriptions,
						typeID,
						&pTokenTypeDescription ) )
		{
			
			WTokenDescriptions desc( pTokenTypeDescription ) ;
			desc.TraceContents() ;
		}
	}
*/
}
