// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RTokenService.cpp

// RTokenService.cpp: implementation of the RTokenService class.
//
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "RTokenService.h"
#include "ICoreServerConnect.h"
#include "RTokenDesc.h"
#include <sstream>
#include <frToken.h>
#include <frCapability.h>
#include <WTokenDescriptions.h>
#include <WRemoteNym.h>
#include <frNiqsTokenInfo.h>
#include <INiqs.h>
#include <IKeyQry.h>
#include <ZkService.h>
#include <WAutoBuffer.h>
#include <WCoreTypes.h>
#include <ZkErrorHandler.h>
#include <IZKErrorReporter.h>
#include <RZkException.h>
#include <HttpHelper.h>
#include <WZkipClient.h>

RTokenService::RTokenService(WRemoteNym& RemoteNym)
	: m_FrTokenCache(NULL)
	, m_RemoteNym(RemoteNym)
{
	memset(&m_TokenServerResponse, 0, sizeof(TokenServerResponse));
}

RTokenService::~RTokenService()
{
	
	if(m_FrTokenCache != NULL)
	{
		ZKVERIFY(frTokenCacheFree(&m_ZkError, &m_FrTokenCache));
	}
}


bool 
RTokenService::LoadTokenCache()
{
	bool ReturnValue = true;

	const unsigned char* buf = NULL;
	size_t szbuf = 0;

	if ( m_FrTokenCache != NULL )
	{
		ZKASSERT( false ) ;
		return true ;
	}

	ZKVERIFY( frRemoteNymResetParameterSearch(
				&m_ZkError ,                 
				m_RemoteNym.GetFrRemoteNym()));
	
	if(frRemoteNymGetParameter(
		&m_ZkError ,
		m_RemoteNym.GetFrRemoteNym() ,
		NymTokenCache,
		&buf ,
		&szbuf) )
	{

		ZKVERIFY(frTokenCacheFromBuffer ( 
			&m_ZkError, 
			&m_FrTokenCache ,  // free in DTOR
			(void*)buf , 
			szbuf ));

		buf = NULL;
		ZKASSERT(m_FrTokenCache);
	}
	else 
	{
		// These are the first tokens created, time to create the
		// token cache
		ZKVERIFY(frTokenCacheNew(&m_ZkError, &m_FrTokenCache));
	}

	return ( m_FrTokenCache != NULL ) ;
}


bool 
RTokenService::SaveTokenCache()
{
	ZKASSERT_POINTER(m_FrTokenCache, FrTokenCache);	
	bool ReturnValue = true;
	
	// The tokens are stored in the "frRemoteNym" object
	// This is historic, if we were to do it again we
	// would probably just save the token strings
	// into a chunk. (Andrew Dec 2000)

	// Delete the previous version of the
	// token buffer in the FrRemoteNym
	ZKVERIFY(frRemoteNymDeleteParameter (
		&m_ZkError  ,                    
		m_RemoteNym.GetFrRemoteNym() ,
		NymTokenCache ));  

	// Store newer version

	WAutoBuffer TokenCacheAsBuffer ;

	ZKVERIFY(frTokenCacheToBuffer ( 
		&m_ZkError , 
		m_FrTokenCache , 
		(void**)&TokenCacheAsBuffer , 
		TokenCacheAsBuffer.GetSizePointer()));

	ZKVERIFY(frRemoteNymAddParameter (
		&m_ZkError ,
		m_RemoteNym.GetFrRemoteNym() ,
		NymTokenCache, 
		TokenCacheAsBuffer ,
		TokenCacheAsBuffer.Size() ));	

	return ReturnValue;
}

// Returns false if unable to get token description
bool 
RTokenService::GetDescriptionOfTokenX(
	const unsigned int nTokenIndex, 
	WTokenDescriptions& TokenDescriptions )
{
	bool ReturnValue = true;

	if ( m_FrTokenCache == NULL )
	{
		LoadTokenCache() ;
	}

	if ( m_FrTokenCache != NULL )
	{
		// Sanity on index
		if (GetTotalNumberOfTokens() < nTokenIndex)
		{
			ZKASSERT(false);
			ReturnValue = false;
		}
		else 
		{
			// Get Base64 Token from cache
			const char* pToken = NULL;
			ZKVERIFY(frTokenCacheGetToken ( 
				&m_ZkError , 
				m_FrTokenCache , 
				nTokenIndex , 
				&pToken ));

			// Token New From Base64
			FrToken TokenToken = NULL;
			ZKVERIFY(frTokenNewFromBase64(
				&m_ZkError ,
				&TokenToken,
				(char*)pToken ,		
				NULL ) );

			// Token type
			unsigned short tokentypeid = 0;
			ZKVERIFY(frTokenTypeId (
				&m_ZkError ,
				TokenToken ,
				&tokentypeid));

			ReturnValue = ZkService::GetITokenDesc()->GetTokenDesc( tokentypeid, TokenDescriptions ) ;
		
			// Free
			ZKVERIFY(frTokenFree (&m_ZkError , &TokenToken));
			TokenToken = NULL;
		}
	}

	return ReturnValue;
}

unsigned int 
RTokenService::GetTotalNumberOfTokens()
{
	size_t temp = 0;

	if ( m_FrTokenCache == NULL ) 
	{
		LoadTokenCache();
	}

	if(m_FrTokenCache != NULL)
	{
		ZKVERIFY(frTokenCacheTokens ( 
			&m_ZkError , 
			m_FrTokenCache , 
			&temp ));
	}
	
	return temp;
}


// Verifies a serial number has correct form
// and whether it is a trial token.

void
RTokenService::GetSerialNumberCapabilities(
	WSerialNumber& io_SerialNumber ) throw(RZkException)
{
	if ( !io_SerialNumber.IsWellFormed() ) {
		// We already know this serial number is bogus,
		// no need talking to the network
		throw RZkException(
				"RTokenService::GetSerialNumberCapabilities",
				"Activation Code is Bad",
				IDS_SERIAL_NUMBER_INVALID ) ;
	}

	// Check whether the serial number is a trial serial number
	// We consider this non-fatal, If we can't get this information we assume
	// it is not a trial token.

	char   CapabilityBuffer[ 4096 ] ;
	size_t szBuffer = 4096 ;
	FrKeyQryHost kq_host ;
	kq_host.fd = ZkService::GetIKeyQry()->GetKeyQuerySocket();

	WNetInfoDatum KQS ;
	ZkService::GetINiqs()->GetServer( FRENT_TYPE_KEYQRY_SRV, KQS );

	SOCKADDR_IN KQS_Addr = KQS.GetSockAddr() ;

	memcpy(&kq_host.host, &KQS_Addr, sizeof(SOCKADDR_IN));	

	WZkipClient client;

	// SmartConnectThroughCloud() throws if the connect is unsuccessful
	ZkService::GetICoreServerConnect()->SmartConnectThroughCloud(
				FRENT_TYPE_TOK_SRV, client, TOKEN_SERVER_TIMEOUT);
	
	int bStatus = frSerialNumberGetCapabilityConnected( &m_ZkError, 								
								 kq_host,
								 ZkService::GetIKeyQry()->GetKeyCache(),
								 io_SerialNumber ,
								 client,
								 &m_TokenServerResponse,
								 CapabilityBuffer,
								 &szBuffer ) ;

	TestAndThrowServerResponse( bStatus, "RTokenService::GetSerialNumberCapabilities" ) ;

	io_SerialNumber.SetCapabilities( CapabilityBuffer, szBuffer ) ;
}

void
RTokenService::RedeemSerialNumberForTokens( const WSerialNumber& in_SerialNumber ) throw(RZkException)
{
	ZKSTD::vector<tstring> vNewTokens;
	RedeemSerialNumberForTokens( in_SerialNumber, vNewTokens );
}


void
RTokenService::RedeemSerialNumberForTokens( const WSerialNumber& in_SerialNumber,
										   ZKSTD::vector<tstring>& out_vStrTokens ) throw(RZkException)
{
	if ( !in_SerialNumber.IsWellFormed() ) 
	{
		// We already know this serial number is bogus,
		// no need talking to the network
		throw RZkException(
				"RTokenService::RedeemSerialNumberForTokens",
				"Activation Code is Bad",
				IDS_SERIAL_NUMBER_INVALID ) ;
	}

	// Make sure we have a token cache 
	// to put the tokens in before redeeming,
	// This may involve loading existing tokens
	if(!m_FrTokenCache)
	{
		if ( !LoadTokenCache() )
		{
			throw RZkException(
					"RTokenService::RedeemSerialNumberForTokens",
					"Token Cache could not be created",
					IDS_DEFAULT_ERROR ) ;
		}
	}


	char** ppchTokens = NULL;
	size_t szTokens = 0;
	WZkipClient client;

	// SmartConnectThroughCloud() throws if the connect is unsuccessful
	ZkService::GetICoreServerConnect()->SmartConnectThroughCloud(
					FRENT_TYPE_TOK_SRV, client, TOKEN_SERVER_TIMEOUT);
	
	int bStatus = frSerialNumberRedeemConnected (
						&m_ZkError ,                 /* Return the error if any occured */
						in_SerialNumber ,             /* The serial number structure to redeem */
						client ,			         
						NULL,
						&m_TokenServerResponse ,     /* Server response structure */
						&ppchTokens ,                /* Returns an array of tokens */
						&szTokens) ;				 /* Number of Tokens in array */
						
	TestAndThrowServerResponse( bStatus, "RTokenService::RedeemSerialNumberForTokens" ) ;
						
	// Store in m_FrTokenCache (in BASE64 format)
	unsigned int nIndex = 0;

	if (ppchTokens) 
	{
		for (size_t n = 0; n < szTokens; ++n) 
		{
			if (ppchTokens[n]) 
			{ 
				ZKVERIFY(frTokenCacheAddToken(
					&m_ZkError,
					m_FrTokenCache,
					ppchTokens[n],
					&nIndex));

				out_vStrTokens.push_back(ppchTokens[n]);
				SAFEFREE( ppchTokens[n] ) ;
			}
		}

		SAFEFREE( ppchTokens ) ;
	}

	// Store in FrRemoteNym (this doesn't actually save onto disk yet
	ZKVERIFY(SaveTokenCache());	
}


void
RTokenService::GetSerialNumberAndRedeemIt( 
	SerialNumberTypes in_type, 
	ZKSTD::vector<tstring>& out_vStrTokens) throw(RZkException)
{
    // Get a NYM_TYPE3_CREATE serial number
    DZkError io_err;
    ZKSTD::string strUrl( ZKConfiguration::GetSerialNumberUrl(in_type) );
    ZKASSERT( !strUrl.empty() ) ;

    WAutoCharBuffer ResponseBuf;

    char strserial[ZKSERIAL_NUMBER_LENGTH + 1] = {0};

    if (!HttpHelper::GetRequest(&io_err, 
                                strUrl.c_str(), 
                                &ResponseBuf) )
    {
        // http request failed
        throw RZkException(io_err, "ZKNymNamePage::OnWizardNext",
                           "HTTP request for free activation code failed",
                           IDS_HTTP_REQUEST_FAILED);
    }
    else
    {
        if( !ResponseBuf.IsValid() ||
            strlen(ResponseBuf) < ZKSERIAL_NUMBER_LENGTH) 
        {
            // what are we sending back these days ..
            ZKASSERT(false);
            throw RZkException(io_err, "ZKNymNamePage::OnWizardNext",
                               "badly formed activation code",
                               IDS_SERIAL_NUM_MAIMED);
        }
        else 
        {
            // grab the first ZKSERIAL_NUMBER_LENGTH chars
            // (http reqests typically come with \r\n\r\n
            // tagged on to the end ...)
            memcpy(strserial, ResponseBuf, ZKSERIAL_NUMBER_LENGTH);
        }
					
        WSerialNumber serialNumber( strserial );
				
        // Do some serial number distribution management
        ZkService::GetITokenService()->GetSerialNumberCapabilities( serialNumber ) ;
							
        // Redeem the serial number for tokens
        ZkService::GetITokenService()->RedeemSerialNumberForTokens( serialNumber, out_vStrTokens ) ;
    }
}

void
RTokenService::GetFreeNymExtendSerialNumberAndRedeemIt( ZKSTD::string& out_strToken ) throw(RZkException)
{
	out_strToken = "" ;

	ZKSTD::vector<tstring> vNewTokens;
	GetSerialNumberAndRedeemIt( NYM_TYPE3_EXTEND, vNewTokens ) ;

	ZKASSERT( vNewTokens.size() == 1 ) ;

	if ( vNewTokens.size() > 0 )
	{
		out_strToken = vNewTokens[ 0 ] ;
	}
}

void				
RTokenService::GetFreeNymCreateSerialNumberAndRedeemIt( ) throw(RZkException)
{
	ZKSTD::vector<tstring> vNewTokens;
	GetSerialNumberAndRedeemIt( NYM_TYPE3_CREATE, vNewTokens ) ;
}


// Use these to create a nym:
bool 
RTokenService::GetTokenXinBase64(
	const unsigned int nIndex, 
	ZKSTD::string& strToken)
	// Get a Token in BASE64, pass to NymSubSystem
{
	bool ReturnValue = true;
	
	if ( m_FrTokenCache == NULL ) 
	{
		ReturnValue = LoadTokenCache() ;
	}
	
	const char* Token = NULL;
	
	if ( m_FrTokenCache != NULL )
	{
		if ( !frTokenCacheGetToken ( 
						&m_ZkError , 
						m_FrTokenCache , 
						nIndex , 
						&Token ) ) 
		{
			// Probably the requested token index is bogus
			ZKASSERT( false ) ;
			ReturnValue = false ;
		}
		else 
		{
			strToken = Token;
		}
	}

	return ReturnValue;
}


bool 
RTokenService::RemoveTokenXFromArray(
	const unsigned int nIndex)
	// Remove the Token from the Array after nym creation

{
	bool ReturnValue = true;

	if ( m_FrTokenCache == NULL ) 
	{
		ReturnValue = LoadTokenCache() ;
	}

	if ( m_FrTokenCache != NULL )
	{	
		// Delete
		ZKVERIFY(frTokenCacheDeleteToken ( 
			&m_ZkError , 
			m_FrTokenCache , 
			nIndex ));

		// Dump
		ZKVERIFY(SaveTokenCache());
	}

	return ReturnValue;
}


bool 
RTokenService::RemoveTokenFromArray(const ZKSTD::string& strRemoveToken)
{
	bool ReturnValue = false;
	size_t nTokens = GetTotalNumberOfTokens();
	ZKSTD::string strToken;

	for(size_t i = 0; i < nTokens; ++i)
	{
		ZKVERIFY(GetTokenXinBase64(i, strToken));

		if(strToken == strRemoveToken)
		{
			ZKVERIFY(RemoveTokenXFromArray(i));
			ReturnValue = true;
			break;
		}
	}

	return ReturnValue;
}


void 
RTokenService::TestAndThrowServerResponse(
	int bStatus, 
	char* in_strOperation) throw (RZkException)
{
	char* pOperation = "RTokenService::TestAndThrowServerResponse" ;
	if ( in_strOperation != NULL ) {
		pOperation = in_strOperation ;
	}
	
	if ( bStatus ) 
	{
		// No failure, nothing to do
		return ;
	}

	// General Server Message

	else if ( m_TokenServerResponse.message[0] != 0 ) 
	{
		char buf[NSRMESG_SZ + 16]; // msg + code

		if (m_TokenServerResponse.code != 0) 
		{
			snprintf(buf, sizeof(buf), "%s [%d]", m_TokenServerResponse.message, m_TokenServerResponse.code);
		}
		else
		{
			snprintf(buf, sizeof(buf), "%s", m_TokenServerResponse.message);
		}


		throw RZkException(pOperation, buf,IDS_SERVER_ERROR, buf ) ;
	}
	// General Client Side Message
	else 
	{
		ZKASSERT( false ) ;
		throw RZkException(
				m_ZkError,
				pOperation, 
				"No Server Response Message Available", 
				IDS_SERVER_ERROR_UNKNOWN);
	}
}

void
RTokenService::GetTokensGroupedByTypeID(ZKSTD::vector<TokensOfType>& out_tokenTypes)
{
	out_tokenTypes.clear();

	// Traverse the tokens collecting the information to build the data structures		
	size_t cntTokens = GetTotalNumberOfTokens();

	// We assume that multiple tokens may exist with the same description,
	// we do not assume they are sorted in that order

	for ( size_t i = 0 ; i < cntTokens ; i++ )
	{
		WTokenDescriptions TokenDescription;

		// The combination "alot of tokens + Downed Niqs server" may cause this call to be too long
		GetDescriptionOfTokenX(i, TokenDescription);

		// Get the structure that refers to this token type
		// (will create one if this is the first time we see it)
		TokensOfType & refItemData = FindItemForTokenDesc( TokenDescription, out_tokenTypes ) ;

		// Remember that this token is on this type
		refItemData.Tokens.push_back( i ) ;
	}
}

// Returns the Item that refers to this type of token.  If it does not
// already exist it creates such an item.

TokensOfType & 
RTokenService::FindItemForTokenDesc(const WTokenDescriptions & in_Desc, 
									ZKSTD::vector<TokensOfType>& io_tokenTypes)
{
	// Find if this Token Type is already in the array
	for ( size_t i = 0 ; i < io_tokenTypes.size() ; i++ )
	{
		// REVIEW: Is there anything else we should compare to determine if
		// two tokens are the same?
		if ( io_tokenTypes[ i ].Desc.GetTypeId() == in_Desc.GetTypeId() )
		{
			return io_tokenTypes[i] ;
		}
	}

	// Wasn't found, create a new entry and return that

	TokensOfType NewTokenType( in_Desc ) ;

	io_tokenTypes.push_back( NewTokenType ) ;

	return io_tokenTypes[ io_tokenTypes.size() - 1 ] ;
}

unsigned int 
RTokenService::GetNumberOfDifferentTokenTypeID()
{
	ZKSTD::vector<TokensOfType> tokenTypes;
	GetTokensGroupedByTypeID(tokenTypes);
	return tokenTypes.size();
}

void 
RTokenService::GetInterface( ITokenService** out_ppITokenService )
{
	ZKASSERT( out_ppITokenService != NULL ) ;
	*out_ppITokenService = static_cast<ITokenService*>(this) ;
}

bool
RTokenService::DoesTokenUseCloud(const unsigned int in_index)
{
	ZKASSERT(in_index >= 0);
	WTokenDescriptions tokenDesc;
	GetDescriptionOfTokenX(in_index,tokenDesc);
	
	FrNymType nymType = FRNYM_TYPE_FROM_TOKEN_FLAGS( tokenDesc.GetFlags() );
	FrNymCapability nymCapability = FRNYM_CAPABILITY_INIT;
	ZKVERIFY(frRemoteNymGetCapabilityFromType(NULL, nymType, &nymCapability));

	return ( nymCapability & FRNYM_CAPABILITY_AUTHROUTES );
}

bool 
RTokenService::DoesTokenUseCloud( const ZKSTD::string& in_strToken )
{
	unsigned int i;
	ZKVERIFY( FindIndex( in_strToken, i ));
	return DoesTokenUseCloud( i );
}


bool
RTokenService::DoesTokenUseMail(const unsigned int in_index)
{
	ZKASSERT(in_index >= 0);
	WTokenDescriptions tokenDesc;
	GetDescriptionOfTokenX(in_index,tokenDesc);
	
	FrNymType nymType = FRNYM_TYPE_FROM_TOKEN_FLAGS( tokenDesc.GetFlags() );
	FrNymCapability nymCapability = FRNYM_CAPABILITY_INIT;
	ZKVERIFY(frRemoteNymGetCapabilityFromType(NULL, nymType, &nymCapability));

	return (( nymCapability & FRNYM_CAPABILITY_MAIL ) != 0);
}

bool 
RTokenService::DoesTokenUseMail( const ZKSTD::string& in_strToken )
{
	unsigned int i;
	ZKVERIFY( FindIndex( in_strToken, i ));
	return DoesTokenUseMail( i );
}



/////////////////////////////////////
// private
/////////////////////////////////////

bool
RTokenService::FindIndex( const ZKSTD::string& in_strToken, unsigned int& out_index )
{
	bool bReturnValue = false;
	size_t nTokens = GetTotalNumberOfTokens();
	ZKSTD::string strToken;

	for(size_t i = 0; i < nTokens; ++i)
	{
		ZKVERIFY( GetTokenXinBase64(i, strToken) );

		if(strToken == in_strToken)
		{
			out_index = i;
			bReturnValue = true;
			break;
		}
	}

	return bReturnValue;
}
