// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RWallet.cpp
//
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "RWallet.h"
#include <ZKFileConfiguration.h>
#include <TraceControl.h>
#include <RZkException.h>
#include <string_hlp.h>
#include <lockguard.h>
#include <ZKFileConfiguration.h>
#include <IZkFileHelp.h>
#include <string_hlp.h>
#include <fstream>
#include <sstream>

static const char DEFAULT_MAPPINGS[] = "default.reserved";

RWallet::RWallet()
{
	LoadDefaultMappings();
}

RWallet::~RWallet()
{
}

void RWallet::LoadDefaultMappings()
{
	ZKS::Lock_Guard<ZKCriticalSection> gate(m_cs);
	GetScriptMappings(DEFAULT_MAPPINGS, m_defaultRealNameLwr2ZkNameMap);
}

bool
RWallet::GetScriptMappings(	const ZKSTD::string& in_strDomain, 
							String2StringMap& out_realNameLwr2ZkNameMap)
{
	bool bIsDefaultScript = false;

	if (in_strDomain == DEFAULT_MAPPINGS)
	{
		out_realNameLwr2ZkNameMap.clear();
		bIsDefaultScript = true;
	}
	else
	{
		ZKS::Lock_Guard<ZKCriticalSection> gate(m_cs);
		out_realNameLwr2ZkNameMap = m_defaultRealNameLwr2ZkNameMap;
	}
	bool bRet = true;

	ZKSTD::string strFile = ZKFileConfiguration::GetWalletScriptPath(in_strDomain);

	std::ifstream ifs(strFile.c_str());
	// If file does not exist return empty multimap

	if (ifs)
	{
		// file exists - populate map from file

		ZKSTD::ostringstream oss;
		oss << ifs.rdbuf();

		char* strFileContents = new char[(oss.str().size() + 1)];
		oss.str().copy(strFileContents, oss.str().size());
		strFileContents[oss.str().size()] = '\0';
		if (strstr(strFileContents, "\t\n") != NULL)
		{
			// faulty file format
			bRet = false;
			ZKASSERT(false);
		}
		else
		{
			char strSeps[] = "\t\n";
			char* strRealName = strtok(strFileContents, strSeps);
			while(strRealName != NULL)
			{
				char* strZkNameEscaped = strtok(NULL, strSeps);
				if (!strZkNameEscaped)
				{
					// Faulty file format
					ZKASSERT(false);
				}
				else 
				{
					// Real name is always lwr cased
					// Zk name is left as is
					ZKSTD::string strRealNameLwr = ZKS::UnEscapeString(strRealName);
					ZKS::MakeLower(strRealNameLwr);
					ZKSTD::string strZkName = ZKS::UnEscapeString(strZkNameEscaped);
					
					// This overwrites if it exists.
					out_realNameLwr2ZkNameMap[strRealNameLwr] = strZkName; 
				}
				strRealName = strtok(NULL, strSeps);
				bRet = true;
			}
		}
		SAFEDELETEARRAY(strFileContents);
		ifs.close();
	}

	return bRet;
}

bool
RWallet::WriteScriptToDisk(const char* in_strDomain, 
						   const char* in_strFile, 
						   unsigned int in_nBytes)
{
	if (!in_strDomain || !in_strFile)
	{
		ZKASSERT(false);
		return false;
	}
	bool bRet = false;

	ZKSTD::string strFile = ZKFileConfiguration::GetWalletScriptPath(in_strDomain);

	try
	{
		if(	!IZkFileHelp::DoesFileOrDirExist( ZKFileConfiguration::GetWalletScriptsDir().c_str()) )
		{
			IZkFileHelp::CreateDirectory( ZKFileConfiguration::GetWalletScriptsDir().c_str() );
		}
	}
	catch(RZkException&)
	{
		ZKASSERT(false);
	}

	FILE* pFile = fopen(strFile.c_str(), "w");
	if (pFile)
	{
		if (in_nBytes != fwrite((void*) in_strFile, sizeof(char), in_nBytes, pFile))
		{
			ZKASSERT(false);
		}
		else
		{
			bRet = true;
		}
		ZKVERIFY(!fclose(pFile));
	}
	if (bRet && !strcmp(in_strDomain, DEFAULT_MAPPINGS))
	{
		LoadDefaultMappings();
	}

	return bRet;
}

#ifdef INTERNAL_BUILD
bool RWallet::IsMappingInDefaultScript(const ZKSTD::string& in_strRealName, 
									 const ZKSTD::string& in_strZkName)
{
	bool bRet = false;

	ZKSTD::string strRealNameLwr = in_strRealName;
	ZKS::MakeLower(strRealNameLwr);

	String2StringMap::const_iterator it = m_defaultRealNameLwr2ZkNameMap.find(in_strRealName);
	if (it != m_defaultRealNameLwr2ZkNameMap.end())
	{
		bRet = ((*it).second == strRealNameLwr);
	}
	return bRet;
}
#endif //INTERNAL_BUILD


void 
RWallet::SelfTest()
{
#ifdef INTERNAL_BUILD

#endif
}
