// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RWordScan.cpp: implementation of the RWordScan class.
//
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "RWordScan.h"
#include <frCrypt.h>
#include <frNymUtil.h>
#include <zkDateTime.h>
#include <ZKConvert.h>
#include <WTokenDescriptions.h>
#include <IZkConsoleMsg.h>
#include <ZKChunk.h>
#include <INiqs.h>
#include <ICrypto.h>
#include <IClientCentral.h>
#include <ICookieManager.h>
#include <ZKIP.h>
#include <ZKStringSearch.h>
#include <lockguard.h>
#include <RZkException.h>
#include <IZKErrorReporter.h>
#include <ZkErrorHandler.h>
#include <ZkStringMgr.h>
#include <MiscMessages.h>
#include <string_hlp.h>


#include <ZkNetHelp.h>
#include "INymWallet.h"
#include "INym.h"
#include "INymCentral.h"
#include "ZkService.h"
#include <sstream>


#define FOUND_YOUR_IP "Your Internet Address (IP)"
static const char STR_NYM_NAME_REPLACEMENT[] = "nym";

RWordScan::RWordScan()
{
}

RWordScan::~RWordScan()
{
}


// Typically, traffic to scan originates from the active nym
// (ie http, telnet). Some traffic, however, (email/nntp)
// may be originated from any nym. This function calls the
// more general override below.
bool 
RWordScan::WordScan(
	unsigned char* in_pBuf, 
	size_t	in_pBufSiz,
	ZKSTD::vector<ZKSTD::string>& out_vecstrOffensiveWordsFound)
{
	size_t indexActive = ZkService::GetINymCentral()->GetActiveNymIndex( );
	return WordScan(in_pBuf, in_pBufSiz, out_vecstrOffensiveWordsFound, indexActive);
}


// Gets the list of offending words for the nym 'User'
// In the future we may do this on a per-Nym basis
// 
// If the originating nym is "User", just search for
// nym names. Otherwise, use (according to preferences)
//	- Basic word scan list
//  - User email address
//  - nym name for any other nym besides originator
//  - IP address
bool 
RWordScan::WordScan(
	unsigned char* in_pBuf, 
	size_t	in_pBufSiz,
	ZKSTD::vector<ZKSTD::string>& out_vecstrOffensiveWordsFound,
	size_t indexOriginatingNym)
{
	ZKS::Lock_Guard<WZkMutex> gate(m_cs);	

	bool bFoundIPAsBinary = false ;

	ZKSTD::vector<ZKSTD::string> vecstrOffensiveWords;

	if(indexOriginatingNym != USER_NYM_INDEX)
	{
		INym* pNym = NULL;
		ZkService::GetINymCentral()->GetNym(IID_INYM, (void**) &pNym, USER_NYM_INDEX);	
		pNym->GetOffensiveWordList(vecstrOffensiveWords);
		AddWebQueryStringSpaceEquivs( vecstrOffensiveWords ) ;
	}


	// Add in the other nym names 

	if ( ZkService::GetIClientCentral()->GetPreference( ZK_CLIENT_PREF_SET, PREF_SCAN_NYMNAMES ) ) 
	{
		for ( size_t i = 1 ; i < ZkService::GetINymCentral()->GetNumberOfNyms() ; i++ ) 
		{

			if ( i == indexOriginatingNym ) 
			{
				continue ;
			}

			INym* pDifferentNym = NULL ;
			ZkService::GetINymCentral()->GetNym(IID_INYM, (void**) &pDifferentNym, i) ;

			if( pDifferentNym )
			{
				unsigned long dwLifeStage = NYM_LIFE_STAGE_UNDEFINED ;
				pDifferentNym->GetYourLifeStage( dwLifeStage ) ;

				if ( dwLifeStage != NYM_LIFE_STAGE_RELEASED ) 
				{
					ZKSTD::string strNym ;
					pDifferentNym->GetNymName( strNym ) ;

					vecstrOffensiveWords.push_back( strNym ) ;
				}
			}
			else
				ZKASSERT( !"Invalid INym returned (GetNym Failed)." );
		}
	}

	if ( indexOriginatingNym != USER_NYM_INDEX &&
		ZkService::GetIClientCentral()->GetPreference( ZK_CLIENT_PREF_SET, PREF_SCAN_EMAIL_ADDR ) ) 
	{
		ZKSTD::string strReturnAddr ;

		INymWallet *pNymWallet = NULL;
		ZkService::GetINymCentral()->GetNym(IID_INYMWALLET, (void**)&pNymWallet, USER_NYM_INDEX);

		strReturnAddr = pNymWallet->GetField(WALLET_ADDR_START[WALLET_ADDR_HOME] + WALLET_EMAIL_OFFSET);
		if ( !strReturnAddr.empty() ) 
		{			
			vecstrOffensiveWords.push_back( strReturnAddr ) ;
		}
		
		strReturnAddr = pNymWallet->GetField(WALLET_ADDR_START[WALLET_ADDR_WORK] + WALLET_EMAIL_OFFSET);
		if ( !strReturnAddr.empty() ) 
		{			
			vecstrOffensiveWords.push_back( strReturnAddr ) ;
		}
		
		strReturnAddr = pNymWallet->GetField(WALLET_ADDR_START[WALLET_ADDR_OTHER] + WALLET_EMAIL_OFFSET);
		if ( !strReturnAddr.empty() ) 
		{
			vecstrOffensiveWords.push_back( strReturnAddr ) ;
		}
	}

	if ( indexOriginatingNym != USER_NYM_INDEX &&
		ZkService::GetIClientCentral()->GetPreference( ZK_CLIENT_PREF_SET, PREF_SCAN_IP ) ) 
	{
		const ZKSTD::vector<ZKSTD::string> vecstrIps = ZkNetHelp::GetVectorOfStringIPs();
		ZKSTD::vector< uint32_t > vecdwIps =   ZkNetHelp::GetVectorOfBinaryIPs() ;

		ZKASSERT( vecstrIps.size() > 0 ) ;
		ZKASSERT( vecdwIps.size() > 0 ) ;

		size_t k, sz;
		// Add the string versions of the IPs
		for ( k = 0 ; k < vecstrIps.size() ; k++ ) 
		{
			vecstrOffensiveWords.push_back( vecstrIps[ k ] ) ;
		}

		for ( k = 0, sz = vecdwIps.size() ; k < sz ; k++ ) 
		{
			vecdwIps.push_back( htonl( vecdwIps[ k ] ) ) ;
		}

		// Look for the IPs
		if ( IsBinaryValueInBuffer( 
				in_pBuf, 
				in_pBufSiz, 
				vecdwIps ) ) 
		{
			bFoundIPAsBinary = true ;
		}
	}

	if( !vecstrOffensiveWords.empty() ) 
	{
		ZKStringSearch::ScanBufferForWords(
				in_pBuf, 
				in_pBufSiz, 
				vecstrOffensiveWords, 
				out_vecstrOffensiveWordsFound);
	}

	if ( bFoundIPAsBinary ) 
	{
		out_vecstrOffensiveWordsFound.push_back( FOUND_YOUR_IP ) ;
	}

	return !out_vecstrOffensiveWordsFound.empty();
}

bool 
RWordScan::IsBinaryValueInBuffer(
	unsigned char* in_pBuf, 
	size_t	in_pBufSiz,
	const ZKSTD::vector< uint32_t > & in_vecValuesToSearchFor )
{
	// Scan for each value
	for ( size_t i = 0 ; i < in_vecValuesToSearchFor.size() ; i++ ) 
	{
		unsigned long ulSearchVal = in_vecValuesToSearchFor[ i ]  ;

		// What are you doing seraching for 0?
		ZKASSERT( ulSearchVal != 0 ) ;

		// Traverse buffer
		for ( size_t pos = 0 ; pos + 3 < in_pBufSiz ; pos++ ) 
		{
			unsigned long ulCandidate = *(( uint32_t *)(in_pBuf + pos)) ;

			if ( ulCandidate == ulSearchVal ) 
			{
				// Stop searching once we find one instance,
				// if we wanted to return more info we should continue 
				return true ;
			}
		}
	}

	return false ;
}


void
RWordScan::AddWebQueryStringSpaceEquivs( ZKSTD::vector<ZKSTD::string>& io_vecstrOffensiveWords )
{	
	ZKSTD::vector<ZKSTD::string>::iterator it ;

	// We fill in equivalent versions of phrases with spaces
	// into this temporary vector
	ZKSTD::vector<ZKSTD::string> vecSpaceEquivs ;

	for ( it = io_vecstrOffensiveWords.begin() ;
		  it != io_vecstrOffensiveWords.end() ;
		  ++it ) 
	{	
		ZKSTD::string& strCandidate = (*it) ;

		ZKASSERT( !strCandidate.empty() ) ;

		if ( ZKSTD::string::npos != strCandidate.find( ' ' ) ) 
		{
			// there is a space so we need to create replacements
			ZKSTD::string strWithPlus ;
			ZKSTD::string strWithEscapedSpace ;

			for( size_t i = 0 ; i < strCandidate.size() ; i++ ) 
			{
				char nextChar = strCandidate[ i ] ;

				if ( nextChar == ' ' ) 
				{
					strWithPlus += '+' ;
					strWithEscapedSpace += "%20" ;
				}
				else 
				{
					strWithPlus += nextChar ;
					strWithEscapedSpace += nextChar ;
				}
			}

			ZKASSERT( strWithPlus.size() == strCandidate.size() ) ;
			ZKASSERT( strWithEscapedSpace.size() > strCandidate.size() ) ;

			vecSpaceEquivs.push_back( strWithPlus ) ;
			vecSpaceEquivs.push_back( strWithEscapedSpace ) ;

		}
	}

	size_t origSize = io_vecstrOffensiveWords.size() ;

	// Add any extra phrases to the word scan list	
	io_vecstrOffensiveWords.insert( io_vecstrOffensiveWords.end(), vecSpaceEquivs.begin(), vecSpaceEquivs.end() );

	ZKASSERT( origSize + vecSpaceEquivs.size() == io_vecstrOffensiveWords.size() ) ;
}

ZKSTD::string
RWordScan::Sanitize(const ZKSTD::string& in_strToSanitize) const throw()
{
	ZKSTD::string strRet(in_strToSanitize);
	ZKSTD::string strNymName;

	size_t nNyms = ZkService::GetINymCentral()->GetNumberOfNyms();
	for (size_t i = 0; i < nNyms; ++i)
	{
		if (i != USER_NYM_INDEX)
		{
			ZkService::GetINymCentral()->GetNymName(i, strNymName);
			ZKASSERT(!strNymName.empty());
			if (!strNymName.empty())
			{
				ZKSTD::ostringstream ossReplacement;
				ossReplacement << STR_NYM_NAME_REPLACEMENT << i;
				strRet = ZKS::replace_all(strRet, strNymName, ossReplacement.str());
			}
		}
	}

	return strRet;
}
