/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/*

   See document in VSS under /zk/freedom/client/zkui/win32/freedomclient/doc/RouteTester.txt

  Important Note 1
  For these tests to run, you must open port 51125 for ALL your Nym's using the personal firewall.

  Important Note 2
  This only works on the religious network. For these tests to work you should have
  the niqstool.cgi running on WEBNIQSTOOL, which at this time is http://faith.freedom.net:51125/fradmin/cgi-bin/niqstool.cgi

  Important Note 3
  Only one of these tests at a time, per network, can be run, since it consists of bringing up and
  down various aips.


*/

#ifndef ROUTETESTER_H
#define ROUTETESTER_H

#ifdef INTERNAL_BUILD

#include "ZkServiceInc.h"

#include <zkbase/libzktest/ZkTestBase.h>

#include <ErrorHandling.h>
#include <list>
#include <client.h>

static const char WEBNIQSTOOL[] = "faith.freedom.net";

class WNetInfoDatum;
class RZkException ;

class ZkService_API RouteTesterHelper
{
public:
	// PURPOSE:	Tests currently selected Nym's security settings
	//			If User nym is selected, only minimal assertions are made
	//			This checks the nym's security settings vs the actual hops in his route
	// REQUIRES:A Nym to be selected 
	// PROMISE:	To correctly handle if a user or a nym is selected.
//	void TestCurrentNymSecuritySettings() throw(RZkException);	

	// PURPOSE:	Marks the current nym's first server as down, forces an NIQS update, tries
	//			to build an authenticated route and verifies the output.
	//			Then this method does the same for an exit rule and an avoid rule.
	// REQUIRES:A Nym to be selected 
	// PROMISE: Asserts if the current nym is user
//	void TestAuthenticatedRoutesWithKnownDownAip() throw(RZkException);

	// PURPOSE: Build and destroy the various unauthenticated routes
	//			Requires a driver for part of the test to work, but useful
	//			even without
//	void TestUnauthenticatedRoutes() ;

	// PURPOSE: Shows what happens when the authenticated route dies (provokes
	//			a UI reaction so this cannot be part of silent smoke tests)
//	void SimulateAuthRouteDestroy();

	// PURPOSE: Similar to TestUnauthenticatedRoutes() except that this one
	//			leaves all the unauthenticated routes up.  Useful for testing
	//			other related areas of the code.
//	void BringUpAllUnauthenticatedRoutes() ;

//private:


	// PURPOSE:  Tests the building of routes when aips are down (and we know about it, meaning our netinfo is up to date)
	//			 The rules are parameterized, so you can test what happens if the rules
	//			 are first, exit, or avoid.
	//			 Your rule will be applied to a random aip, which will be marked a down. Afterwards, the nym will
	//			 be selected.
	// REQUIRES: A functionning Niqs (or more) is necessary to mark aip as down
	// PROMISE:	 Not to mark a Niqs as down
	static void TestAuthenticatedRoutesWithKnownDownAip(const size_t in_nNym, const RULEZ in_Rule) throw(RZkException);

	// Cross platform
	static void sleep( size_t in_milliTimeToSleep ) ;

	enum BLOCKING_MODE { NONBLOCKING, BLOCKING };

	static bool ChangeAvailabiliyOfAips(	const ZKSTD::list<ZKSTD::string>& in_aipsToSetAvailable, 
											const ZKSTD::list<ZKSTD::string>& in_aipsToSetUnavailable);

	//PURPOSE:  To apply specified rule to one of the available aips (except any of the Niqs)
	//REQUIRE:	0 is the user, 1 is the first nym
	//PROMISE:	Not to mark any of the Niqs as down (all other aips are fair game)
	//			Will assert if an niqs returned by ZkService::GetINiqs()->GetAllNiqs is not
	//			in the list of aips returned by ZkService::GetINiqs()->GetServersOfType() [comparing IP's]
	//			For the specified Nym, note that his current rules will be erased and modified.
	//			Any existing rules on that Nym are not taken into account.
	static bool ApplyRuleToRandomAipForNym(const unsigned int in_nymIndex, WNetInfoDatum& out_aipInfo, const RULEZ in_RuleToApply) throw(RZkException);

	static bool MarkAipUnavailable(const WNetInfoDatum& in_aipInfo);
	static bool MarkAipAvailable(const WNetInfoDatum& in_aipInfo);

	// PURPOSE: Send a message to freedom to set the active nym
	// PROMISE: Will assert if in_nymIndex is greater than or equal to the number of nyms
	//			Does this in a non-blocking fashion, so if you must know if it succeeded wait a couple of seconds
	// REQUIRE: 0 is the user, 1 is the first nym, etc.
	static bool MakeNymActive(unsigned int in_nymIndex, const BLOCKING_MODE in_Blocking);

	static ZKSTD::string CreateStringUsingParameter(	const ZKSTD::string& in_parameter, 
														const ZKSTD::list<ZKSTD::string>& in_aips);

	//PURPOSE: Compare the country and province of the exisiting authenticated route with that of the parameter
	//REQUIRE: A valid input parameter, AND an authentic route must exist
	//PROMISE: asserts if they are different
	static void VerifyRegions(const WNetInfoDatum& in_FirstAip);

	//PURPOSE: Compare two WNetInfoDatum's country's and province
	//REQUIRE: Valid input parameters
	//PROMISE: Returns true if both the country and the province are the same
	static bool CheckIfAipIsInRegionAsRules(const WNetInfoDatum& in_FirstAip, const WNetInfoDatum& in_CurrentFirstAip);

	static void EnsureRandomRouteWindowIsActiveAndDismiss(const int in_nDismissMode);

	static void TestResultsOfRoutes(const size_t in_nNym, const RULEZ in_Rule, const WNetInfoDatum& in_OriginalFirstAip);
	

	// Build an unauthenticated route by connecting to the server
	static void BuildUnauthenticatedRouteDirectly( int in_indexServerInfo ) ;

	// Build an unauthenticated route by directly forcing the route manager to build the route
	static void BuildUnauthenticatedRouteIndirectly( int in_indexServerInfo ) ;

	// Build an unauthenticated route by a call to ICoreServerConnect
	static void BuildUnauthenticatedRouteWithCoreServerConnect( int in_indexServerInfo ) ;

	// Helper function to make sure no authenticated route exists
	static void TearDownUnauthenticatedRoute( int in_indexServerInfo ) ;


};

class TestCurrentNymSecuritySettingsUnitTest : public ZkTestBase
{
public:
	TestCurrentNymSecuritySettingsUnitTest()
		: ZkTestBase( "zkcentral/zkservices/TestCurrentNymSecuritySettingsUnitTest",
						"Current Nym Security Settings Test" )
	{}

	~TestCurrentNymSecuritySettingsUnitTest()
	{}

	virtual long Execute( void* );

private:
	// PURPOSE:	Tests currently selected Nym's security settings
	//			If User nym is selected, only minimal assertions are made
	//			This checks the nym's security settings vs the actual hops in his route
	// REQUIRES:A Nym to be selected 
	// PROMISE:	To correctly handle if a user or a nym is selected.
	void TestCurrentNymSecuritySettings() throw( RZkException );	
};

class TestAuthenticatedRoutesWithKnownDownAipUnitTest : public ZkTestBase
{
public:
	TestAuthenticatedRoutesWithKnownDownAipUnitTest()
		: ZkTestBase( "zkcentral/zkservices/TestAuthenticatedRoutesWithKnownDownAip",
						"Authenticated Routes Test" )
	{}

	~TestAuthenticatedRoutesWithKnownDownAipUnitTest()
	{}

	virtual long Execute( void* );

private:
	// PURPOSE:	Marks the current nym's first server as down, forces an NIQS update, tries
	//			to build an authenticated route and verifies the output.
	//			Then this method does the same for an exit rule and an avoid rule.
	// REQUIRES:A Nym to be selected 
	// PROMISE: Asserts if the current nym is user
	void TestAuthenticatedRoutesWithKnownDownAip() throw( RZkException );
};



class SimulateAuthRouteDestroyUnitTest : public ZkTestBase
{
public:
	SimulateAuthRouteDestroyUnitTest()
		: ZkTestBase( "zkcentral/zkservices/SimulateAuthRouteDestroyUnitTest",
						"Simulate Route Destroy Test" )
	{}

	~SimulateAuthRouteDestroyUnitTest()
	{}

	virtual long Execute( void* );

private:
	// PURPOSE: Shows what happens when the authenticated route dies (provokes
	//			a UI reaction so this cannot be part of silent smoke tests)
	void SimulateAuthRouteDestroy();
};

class TestUnauthenticatedRoutesUnitTest : public ZkTestBase
{
public:
	TestUnauthenticatedRoutesUnitTest()
		: ZkTestBase( "zkcentral/zkservices/TestUnauthenticatedRoutesUnitTest",
						"Unauthenticated route Test" )
	{}

	~TestUnauthenticatedRoutesUnitTest()
	{}

	virtual long Execute( void* );

private:
	// PURPOSE: Build and destroy the various unauthenticated routes
	//			Requires a driver for part of the test to work, but useful
	//			even without
	void TestUnauthenticatedRoutes() ;
};


class BringUpAllUnauthenticatedRoutesUnitTest : public ZkTestBase
{
public:
	BringUpAllUnauthenticatedRoutesUnitTest()
		: ZkTestBase( "zkcentral/zkservices/BringUpAllUnauthenticatedRoutesUnitTest",
						"Build all unauthenticated routes" )
	{}

	~BringUpAllUnauthenticatedRoutesUnitTest()
	{}

	virtual long Execute( void* );

private:
	// PURPOSE: Similar to TestUnauthenticatedRoutes() except that this one
	//			leaves all the unauthenticated routes up.  Useful for testing
	//			other related areas of the code.
	void BringUpAllUnauthenticatedRoutes() ;

};

#endif // INTERNAL_BUILD

#endif
