// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// IZkContext.h

#ifndef _IZKCONTEXT_H
#define _IZKCONTEXT_H
#include <client.h>
#include <ErrorHandling.h>
#include <string>
#include <zksocket.h>
#include <zkMemBuf.h>

class DZkError;

class IZkContext
{
public:
	enum {	ZKPROXY_DO_NOTHING, 
			ZKPROXY_SEND_TO_CLIENT,
			ZKPROXY_SEND_TO_SERVER,
			ZKPROXY_CONNECT_TO_SERVER};

	// PURPOSE: This buffer should be used by the ZKProxy to receive raw data.
	//
	//			The ZKProxy should store the number of bytes actually received
	//			into ByteBufferLength().
	//
	//			The Protocol may SetProtocolOutput to point at this buffer.
	//			(See below.)
	//
	// PROMISE: The buffer is of size RealByteBufferSize.
	// REQUIRE: Nothing.
	virtual void*	ByteBuffer		() const PURE;

	// PURPOSE: The maximum amount of data that can be stored in ByteBuffer().
	// PROMISE: ByteBuffer() points to an allocated buffer of this size.
	// REQUIRE: Nothing.
	virtual size_t		ByteBufferSize	() const PURE;

	// PURPOSE: Used by the ZKProxy to tell the Protocol object how much
	//			data was actually received into ByteBuffer().
	// PROMISE:	Nothing.
	// REQUIRE: Nothing.
	virtual size_t		GetByteBufferLength() const PURE;
	virtual void		SetByteBufferLength(size_t) PURE;


	// PURPOSE: Used by the Protocol to tell the ZkProxy where to find
	//			the Freedomized output. The Protocol should
	//			SetProtocolOutputLength() to the length of the data to send.
	// PROMISE:	Nothing.
	// REQUIRE: Nothing.
	virtual void*		GetProtocolOutput() const	PURE;
	virtual void		SetProtocolOutput(void*)	PURE;

	// PURPOSE: Used by the Protocol to tell the ZkProxy how much data to
	//			send from GetProtocolOutput().
	// PROMISE:	Nothing.
	// REQUIRE: Nothing.
	virtual size_t		GetProtocolOutputLength() const PURE;
	virtual void		SetProtocolOutputLength(size_t) PURE;

	// PURPOSE: Tell the ZkProxy to close the connection after this transaction.
	// PROMISE:	Nothing.
	// REQUIRE: Nothing.
	virtual void		SetCloseConnection(bool) PURE;
	virtual bool		GetCloseConnection() const PURE;

	// PURPOSE: This ZkMemBuf is to be used by the protocol, if the 2048byte
	//			buffer is not sufficient. In this case, SetProtocolOutput()
	//			to point to the pointer from zkMemBufGetData() and
	//			SetProtocolOutputLength() to point to the length of this
	//			data.
	//
	//			The ZkMemBuf will not be allocated when the Context is passed
	//			to the Protocol. The Protocol should expect the ZkMemBuf
	//			to be deallocated between transactions (after the proxy sends
	//			the data, if any).
	//
	//			If the ZkMemBuf has been allocated, it will be freed by
	//			ResetContext()
	//			PROMISE: 
	virtual ZkMemBuf&	GetZkMemBuf() PURE;

	// PURPOSE:	Indicate the action the ZkProxy should take at the end
	//			of this transaction.
	//			ZKPROXY_DO_NOTHING,
	//			ZKPROXY_SEND_TO_CLIENT,
	//			ZKPROXY_SEND_TO_SERVER, or
	//			ZKPROXY_CONNECT_TO_SERVER
	//			This is set by the Protocol object, and read by the ZkProxy.
	//			This is reset by ResetContext() to ZKPROXY_DO_NOTHING.
	// PROMISE: Nothing.
	// REQUIRE: Nothing.
	virtual int			GetProxyAction		() const PURE;
	virtual void		SetProxyAction		(int) PURE;

	// PURPOSE: Reset:
	//				GetByteBufferLength()		= 0
	//				GetProxyAction()			= ZKPROXY_DO_NOTHING
	//				GetZkMemBuf()				= NULL
	//				GetProtocolOutputLength()	= 0
	//				GetProtocolOutput()			= NULL
	virtual void		ResetContext		() PURE;

	virtual void		SetRemoteAddress( ZkSocketAddress& ) PURE;
	virtual	ZkSocketAddress& GetRemoteAddress( ) PURE;
	// PURPOSE: Return the name of the remote host.
	// PROMISE: A best effort to get the domain name. Will try, in order:
	//				1) Return the name as set by the ProtocolLogic
	//					(ie, the "Host:" field in an HTTP Request)
	//					Only available if the ProtocolLogic has set this 
	//					value using SetRemoteDomainName()
	//				2) Return the name as determined by a reverse DNS lookup.
	//				3) Return the IP number.
	// REQUIRE: The Context should have been constructed with the proper
	//			ZkSocketAddress.
	virtual const ZKSTD::string& GetRemoteDomainName	() PURE;

	
	// PURPOSE: ProtocolLogic can use this to set the host field from
	//			its own source. (ie, the "Host:" field in an HTTP Request)
	// PROMISE: Nothing.
	// Require: Nothing.
	virtual void		SetRemoteDomainName	(const ZKSTD::string&) PURE;

	// PURPOSE: Return the ZKError object to be used by the Proxy and
	//			ProtocolLogic objects.
	// PROMISE: Nothing.
	// REQUIRE: Nothing.
	virtual DZkError*	GetZkError	() PURE;

	// PURPOSE: Ensure the internal state is valid. 
	// PROMISE: Nothing.
	// REQUIRE: Nothing.
	virtual void		VerifyState	() const PURE;
};

#endif // _IZKCONTEXT_H
