// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// IZKProtocol.h

#ifndef _IZKPROTOCOL_H
#define _IZKPROTOCOL_H
#include "IZkContext.h"

// An object of type IZKProtocol should implement protocol specific
// proxy functionality. The data will be received from the User Agent,
// and given to the Protocol object (via
// FreedomizeClientRequest/ServerResponse)
// The Protocol object should modify the data as it deems appropriate,
// before returning from these functions.

// The GetX(), SetX(), and ByteBuffer() functions are all in IZkContext

class IZkProtocol
{
public:

	// PURPOSE: Called when the local client is ready to receive data.
	//			At this time, the protocol should probably either send
	//			back a welcome banner or connect to the remote server.
	// PROMISE: The protocol will SetProxyAction() and (if appropriate)
	//			SetProtocolOutput()/SetProtocolOutputLength.
	// REQUIRE: 
	virtual bool ClientReady				(IZkContext&) PURE;

    // PURPOSE: Called after failing to connect to the remote server.
    //          The protocol can either do nothing or return an
    //          error or other msg to the client app.
    // PROMISE: The Protocol should properly SetProxyAction() and (if
    //          appropriate) SetProtocolOutput()/SetProtocolOutputLength()
    // Require: 
    virtual bool ConnectFailed              (IZkContext&) PURE;

	// PURPOSE: Data has been received from the local client.
	//			The Protocol should determine the appropriate action
	//			usually either passing on the (modified?) request to
	//			the server, or returning a response to the client
	//			if the request will be delt with/queued locally
	// PROMISE: The protocol will SetProxyAction() and (if appropriate)
	//			SetProtocolOutput()/SetProtocolOutputLength.
	// REQUIRE:
	virtual bool FreedomizeClientRequest	(IZkContext&) PURE;
	
	// PURPOSE: Data has been received from the remote server.
	//			The Protocol should determine the appropriate action
	//			usually passing on the (modified?) response to
	//			the client.
	// PROMISE: The protocol will SetProxyAction() and (if appropriate)
	//			SetProtocolOutput()/SetProtocolOutputLength.
	// REQUIRE:
	virtual bool FreedomizeServerResponse	(IZkContext&) PURE;
	
	// PURPOSE: The first function called on a Protocol object.
	// PROMISE: Protocol will be ready for use.
	// REQUIRE:
	virtual bool InitProtocolResources(IZkContext&) PURE;

	// PURPOSE: Last function called on a Protocol object.
	// PROMISE: Frees all resources.
	// REQUIRE:
	virtual bool FreeProtocolResources(IZkContext&) PURE;
};

#endif // _IZKPROTOCOL_H
