// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#ifdef WIN32
#pragma warning (disable : 4251)
#endif

#include "RZkHttpProtocol.h"
#include <vector>
#include <ZkProxyDialogs.h>
#include <ZkService.h>
#include <INymCentral.h>
#include <IClientCentral.h>
#include <ICookieManager.h>
#include <IAdBlocker.h>
#include <string_hlp.h>
#include <ci_string.h>
#include <algorithm>
#include <IWordScan.h>
#include <INym.h>
#include <INymWebCacheMgr.h>
#include <IZkConsoleMsg.h>

// Look for files named frrequestXXX.dat, frresponseXXX.dat, rwrequestXXX.dat, rwresponseXXX.dat
// in C:\ on windows and in \tmp\ on linux.
#ifdef _DEBUG
//#define TRACE_HTTP_TO_FILE
#endif

#ifdef WIN32
#define HTTP_LOG_DIRECTORY "C:\\"
#else
#define HTTP_LOG_DIRECTORY "/tmp/"
#endif

#define URL_TO_BE_BLOCKED "BLOCKED_BY_AD_BLOCKER"

RZkHttpProtocol::RZkHttpProtocol()
			: m_RequestHeader(NULL)
			, m_ResponseState(READING_RESPONSE_HEADER)
{
}

// Callback for PhilippeB's dump functions
extern "C" {
  int HTTPWriteBuf ( ZkError , void * , const unsigned char* ,int );
}	
int HTTPWriteBuf ( 
    ZkError err ,
    void * cdata ,
    const unsigned char * data ,
    int dlen )
{
    zkErrorClear(err);
    if (cdata && data && (dlen > 0)) {
		zkMemBufAppend(err, (ZkMemBuf)cdata, (void*)data, dlen);
	}
    return 1;
}

void RZkHttpProtocol::TraceOutgoing(IZkContext& in_Context)
{
#ifdef TRACE_HTTP_TO_FILE
	ZKSTD::string strFile = HTTP_LOG_DIRECTORY;
	if(in_Context.GetProxyAction() == IZkContext::ZKPROXY_SEND_TO_CLIENT)
	{
		strFile += "frresponse";
	}
	else if(in_Context.GetProxyAction() == IZkContext::ZKPROXY_SEND_TO_SERVER)
	{
		strFile += "frrequest";
	}
	else
	{
		return;
	}
	strFile += m_strDebugSuffix;

	FILE *pF = fopen(strFile.c_str(), "ab");

	fwrite(in_Context.GetProtocolOutput(), 1, in_Context.GetProtocolOutputLength(), pF);

	fclose(pF);
#endif
}

void RZkHttpProtocol::AppendWithCommentary(const char* in_strData)
{
	AppendWithCommentary(in_strData, strlen(in_strData));
}

void RZkHttpProtocol::AppendWithCommentary(const ZKSTD::string& in_strData)
{
	AppendWithCommentary(in_strData.c_str(), in_strData.length());
}

void RZkHttpProtocol::AppendWithCommentary(const char *in_pData, const size_t in_lenData)
{
#ifdef TRACE_HTTP_TO_FILE
	ZKSTD::string strFile = HTTP_LOG_DIRECTORY;
    strFile += "frcomments";
	
	strFile += m_strDebugSuffix;

	FILE *pF = fopen(strFile.c_str(), "ab");

	fwrite(in_pData, 1, in_lenData, pF);

	fclose(pF);
#endif
}

void RZkHttpProtocol::TraceIncomingRequest(IZkContext& in_Context)
{
#ifdef TRACE_HTTP_TO_FILE
	ZKSTD::string strFile = HTTP_LOG_DIRECTORY;
    strFile += "rwrequest";
	
	strFile += m_strDebugSuffix;

	FILE *pF = fopen(strFile.c_str(), "ab");

	fwrite(in_Context.ByteBuffer(), 1, in_Context.GetByteBufferLength(), pF);

	fclose(pF);
#endif
}

void RZkHttpProtocol::TraceIncomingResponse(IZkContext& in_Context)
{
#ifdef TRACE_HTTP_TO_FILE
	ZKSTD::string strFile = HTTP_LOG_DIRECTORY;
    strFile += "rwresponse";
	
	strFile += m_strDebugSuffix;

	FILE *pF = fopen(strFile.c_str(), "ab");

	fwrite(in_Context.ByteBuffer(), 1, in_Context.GetByteBufferLength(), pF);

	fclose(pF);
#endif
}

bool RZkHttpProtocol::InitProtocolResources(IZkContext& in_Context)
{
#ifdef TRACE_HTTP_TO_FILE
	srand(time(NULL));
	int i = rand();
	char buffer[32];
	
	snprintf(buffer, 32, "%d.dat", i);

	m_strDebugSuffix = buffer;
#endif

	bool ReturnValue = false;
	if(!zkHTTPHeaderNew(in_Context.GetZkError(), &m_RequestHeader))
	{
		ZKASSERT(0);
	}
	else if(!zkHTTPHeaderKeepBodyData(in_Context.GetZkError(), m_RequestHeader))
	{
		ZKASSERT(0);
	}
	else if(!zkHTTPHeaderSetType(in_Context.GetZkError(), m_RequestHeader, REQUEST_HEADER))
	{
		ZKASSERT(0);
	}
	else
	{
		ReturnValue = true;
	}

	return ReturnValue;

}

bool RZkHttpProtocol::ClientReady(IZkContext& in_Context)
{
	in_Context.SetProxyAction( IZkContext::ZKPROXY_CONNECT_TO_SERVER );
#ifndef WIN32
    static ZKSTD::string strPrevIP;
    ZKSTD::string strIP(inet_ntoa(in_Context.GetRemoteAddress().m_inet.sa.sin_addr));

    if(strIP != strPrevIP)
    {
        if(ZkService::GetINymCentral()->IsCurrentNymUser())
        {
            IZkConsoleMsg::ShowImportantMsg( IDS_CONS_HTTP_SEENIP_CONNECT, 
               strIP.c_str());
        }
        else
        {
            IZkConsoleMsg::ShowImportantMsg( IDS_CONS_HTTP_HIDENIP_CONNECT, 
               strIP.c_str());
        }
        strPrevIP = strIP;
    }
#endif

	return true;
}

bool RZkHttpProtocol::ConnectFailed(IZkContext& in_Context)
{
    in_Context.SetCloseConnection(true);
    return true;
}

bool RZkHttpProtocol::FreeProtocolResources(IZkContext& in_Context)
{
	if(m_listUris.size())
	{
		ZKTRACE1("The following were still in my list of pending HTTP responses from %s:\r\n", in_Context.GetRemoteDomainName().c_str());
	
		for(ZKSTD::list<ZKSTD::string>::iterator it = m_listUris.begin(); it != m_listUris.end(); it++)
		{
			ZKTRACE1("%s\r\n", it->c_str());
		}
	}

	if(zkHTTPHeaderFree(in_Context.GetZkError(), &m_RequestHeader))
	{
		return true;
	}

	return false;
}

bool RZkHttpProtocol::FreedomizeClientRequest	(IZkContext& in_Context)
{
	TraceIncomingRequest(in_Context);

	int iStatus = 0;
	bool ReturnValue = true;

	if (!zkHTTPHeaderAddText(in_Context.GetZkError(),
					m_RequestHeader, (unsigned char*)in_Context.ByteBuffer(),
					in_Context.GetByteBufferLength()))
	{
		ReturnValue = false;
		ZKASSERT(false);
	}
		
	else if (!zkHTTPHeaderStatus(in_Context.GetZkError(),
					m_RequestHeader, &iStatus))
	{
		ReturnValue = false;
		ZKASSERT(false);
	}
	else
	{
		switch(iStatus)
		{
		case PARSE_ERROR:
			ReturnValue = false;
			in_Context.SetProxyAction(IZkContext::ZKPROXY_DO_NOTHING);
			ResetRequestHeader(in_Context);
			break;
		case PARSE_COMPLETE:
			ProcessRequest(in_Context);
			ResetRequestHeader(in_Context);
			break;

		default:
			in_Context.SetProxyAction(IZkContext::ZKPROXY_DO_NOTHING);
			break;
		}
	}
	
	TraceOutgoing(in_Context);

	return ReturnValue;
}

void RZkHttpProtocol::AppendToFreedomizedResponse(IZkContext& in_Context, const char *in_pData, const size_t in_len)
{
	in_Context.SetProxyAction( IZkContext::ZKPROXY_SEND_TO_CLIENT );
	if(!in_Context.GetZkMemBuf())
	{
		ZKVERIFY(zkMemBufNew(in_Context.GetZkError(), &in_Context.GetZkMemBuf()));
	}

	ZKVERIFY(zkMemBufAppend(in_Context.GetZkError(), in_Context.GetZkMemBuf(), in_pData, in_len));
}


bool RZkHttpProtocol::FreedomizeServerResponse	(IZkContext& in_Context)
{
	bool ReturnValue = true;

	TraceIncomingResponse(in_Context);

	in_Context.SetProxyAction( IZkContext::ZKPROXY_DO_NOTHING );

	size_t iCurPos = 0;

	while(ReturnValue && iCurPos < in_Context.GetByteBufferLength())
	{
		switch(m_ResponseState)
		{
		case READING_FOOTER:
		case READING_RESPONSE_HEADER:
			{
				int len = FindHeaderEnd((char*)in_Context.ByteBuffer() + iCurPos,
										in_Context.GetByteBufferLength() - iCurPos);

				if(len == -1)
				{
					// no header end found; is there a header at all? (bug 8735)
					if( m_strCurPart.substr(0,4) != "HTTP" &&
						m_strCurPart.substr(0,5) != "\nHTTP" &&
						m_strCurPart.substr(0,6) != "\r\nHTTP" )
					{
						// ...Nope, no header.  Assume it's a successful response, and move on
#ifdef TRACE_HTTP_TO_FILE
						AppendWithCommentary("\r\n*** Uh oh, didn't see \"HTTP\" ***");
#endif
						m_ResponseState = READING_UNTERMINATED_RESPONSE;
						iCurPos = 0;	// we'll re-process
						break;
					}
					else
					{
						// ...Yup there's a header, but we didn't find the end - wait for more data
						m_strCurPart += ZKSTD::string((const char*)in_Context.ByteBuffer() + iCurPos,
													in_Context.GetByteBufferLength() - iCurPos);
						iCurPos = in_Context.GetByteBufferLength();
					}
				}
				else if(m_ResponseState == READING_FOOTER)
				{
					m_strCurPart += ZKSTD::string((char*)in_Context.ByteBuffer() + iCurPos, len);
					iCurPos += len;
					
					AppendToFreedomizedResponse(in_Context, m_strCurPart.c_str(), m_strCurPart.length());

#ifdef TRACE_HTTP_TO_FILE
					AppendWithCommentary("\r\n*** Response Footer ***\r\n");
					AppendWithCommentary(m_strCurPart);
#endif

					m_strCurPart = "";

					DoneWithOneResponse(in_Context);

					m_ResponseState = READING_RESPONSE_HEADER;
				}
				else
				{
					ZKSTD::string strValue;

					// Assume we don't find Content-Length or "Chunked"
					m_ResponseState = READING_UNTERMINATED_RESPONSE;
					m_strCurPart += ZKSTD::string((const char*)in_Context.ByteBuffer() + iCurPos, len);

					if(m_strCurPart.substr(0,4) != "HTTP" &&
						m_strCurPart.substr(0,5) != "\nHTTP" &&
						m_strCurPart.substr(0,6) != "\r\nHTTP" )
					{
						// Well well, we thought we found the end of a header but we don't
						// really have a header here.  Bug 8735...proceed as above.
#ifdef TRACE_HTTP_TO_FILE
						AppendWithCommentary("\r\n*** Uh oh, didn't see \"HTTP\" ***");
#endif
						iCurPos = 0;	// re-process as unterminated, headerless response
						break;
					}
					
					iCurPos += len;

#ifdef TRACE_HTTP_TO_FILE

					AppendWithCommentary("\r\n*** Response Header (unsanitized) ***\r\n");
					AppendWithCommentary(m_strCurPart);
#endif
					SanitizeIncomingHeader(in_Context, m_strCurPart);
					AppendToFreedomizedResponse(in_Context, m_strCurPart.c_str(), m_strCurPart.length());
					
					int iResponseCode = GetResponseStatusCode(m_strCurPart);

					if(!(iResponseCode >= 100 && iResponseCode < 200) && iResponseCode != 408)
					{
						// i.e. iResponseCode is NOT 'Informational' and NOT a time-out notice
						ZKASSERT(m_listUris.size());
						if(m_listUris.size())
						{
							m_listUris.pop_front();
						}
					}

					if( (iResponseCode >= 100 && iResponseCode < 200) ||
						iResponseCode == 204 || iResponseCode == 304)
					{
						// i.e. iResponseCode is 'Informational' or 'No Content' or 'Not Modified'
						DoneWithOneResponse(in_Context);
						m_ResponseState = READING_RESPONSE_HEADER;
					}
					// bug 8913... transfer-encoding takes precedence over content-length.
					else if(ZKS::extractHeaderField(m_strCurPart.c_str(), m_strCurPart.length(),
										"Transfer-Encoding", strValue))
					{
						ZKS::ci_string strCiValue = strValue.c_str();
						if(strCiValue == "chunked")
						{
							m_ResponseState = READING_CHUNK_HEADER;
						}
					}
					else if(ZKS::extractHeaderField(m_strCurPart.c_str(), m_strCurPart.length(),
										"Content-Length", strValue))
					{
						m_lenLeftToRead = atoi(strValue.c_str());
						m_ResponseState = READING_RESPONSE_WITH_CONTENT_LENGTH;
					}

					m_strCurPart = "";

				}
			}
			break;

		case READING_CHUNK_HEADER:
			{
				int len = FindLineEnd((char*)in_Context.ByteBuffer() + iCurPos,
										in_Context.GetByteBufferLength() - iCurPos);
				if(len == -1)
				{
					m_strCurPart += ZKSTD::string((char*)in_Context.ByteBuffer() + iCurPos,
												in_Context.GetByteBufferLength() - iCurPos);
					iCurPos = in_Context.GetByteBufferLength();
				}
				else
				{
					m_strCurPart += ZKSTD::string((char*)in_Context.ByteBuffer() + iCurPos, len);
					iCurPos += len;

#ifdef TRACE_HTTP_TO_FILE
					AppendWithCommentary("\r\n*** Chunk Header ***\r\n");
					AppendWithCommentary(m_strCurPart);
#endif
					AppendToFreedomizedResponse(in_Context, m_strCurPart.c_str(), m_strCurPart.length());

					m_lenLeftToRead = 0;
					sscanf(m_strCurPart.c_str(), " %x", &m_lenLeftToRead);
					if(m_lenLeftToRead)
					{
						m_lenLeftToRead += 2; // There's a CRLF after the chunk data
						m_ResponseState = READING_CHUNK_BODY;
					}
					else
					{
						m_ResponseState = READING_FOOTER;
					}
					m_strCurPart = "";
				}
			}
			break;

		case READING_CHUNK_BODY:
		case READING_RESPONSE_WITH_CONTENT_LENGTH:
			{
				size_t n = in_Context.GetByteBufferLength() - iCurPos;
				if(m_lenLeftToRead < in_Context.GetByteBufferLength() - iCurPos)
				{
					n = m_lenLeftToRead;
				}

#ifdef TRACE_HTTP_TO_FILE
					AppendWithCommentary("\r\n*** Body ***\r\n");
					AppendWithCommentary((const char*)in_Context.ByteBuffer() + iCurPos, n);
#endif
					
				SanitizeResponseBody((char*)in_Context.ByteBuffer() + iCurPos, n);
				AppendToFreedomizedResponse(in_Context, (char*)in_Context.ByteBuffer() + iCurPos, n);
				m_lenLeftToRead -= n;
				if(!m_lenLeftToRead)
				{
					if(m_ResponseState == READING_CHUNK_BODY)
					{
						m_ResponseState = READING_CHUNK_HEADER;
					}
					else
					{
						DoneWithOneResponse(in_Context);
						m_ResponseState = READING_RESPONSE_HEADER;
						m_strCurPart = "";
					}
				}
				iCurPos += n;
			}
			break;

		case READING_UNTERMINATED_RESPONSE:
#ifdef TRACE_HTTP_TO_FILE
					AppendWithCommentary("\r\n*** Unterminated Response ***\r\n");
					AppendWithCommentary((const char*)in_Context.ByteBuffer() + iCurPos, in_Context.GetByteBufferLength() - iCurPos);
#endif
			SanitizeResponseBody((char*)in_Context.ByteBuffer() + iCurPos, 
								in_Context.GetByteBufferLength() - iCurPos);
			AppendToFreedomizedResponse(in_Context, (char*)in_Context.ByteBuffer() + iCurPos,
							in_Context.GetByteBufferLength() - iCurPos);
			iCurPos = in_Context.GetByteBufferLength();
			break;
		}
	}


	void *pHttpData = NULL;
	size_t lenHttpData = 0;

	if(!ReturnValue)
	{

	}
	else if(in_Context.GetZkMemBuf())
	{	// There is data to return to the client
		if(!zkMemBufAppend(in_Context.GetZkError(), 
			in_Context.GetZkMemBuf(), "\0", 1))
		{
			ReturnValue = false;
			ZKASSERT(false);
		}			
		else if(!(zkMemBufGetData(in_Context.GetZkError(), 
			in_Context.GetZkMemBuf(), (void**)&pHttpData, &lenHttpData)))
		{
			ReturnValue = false;
			ZKASSERT(false);
		}
		else
		{
			in_Context.SetProtocolOutput(pHttpData);
			// -1 to remove the trailing NULL
			in_Context.SetProtocolOutputLength(lenHttpData - 1);
		}
	}

	TraceOutgoing(in_Context);
	return ReturnValue;
}

void RZkHttpProtocol::SanitizeResponseBody(char *in_pBody, size_t in_lenBody)
{
	
	if(!ZkService::GetINymCentral()->IsCurrentNymUser() ||
		!ZkService::GetIClientCentral()->GetPreference(ZK_CLIENT_PREF_SET, PREF_NO_USER_COOKIE_MANAGEMENT ) )
	{
		ZKS::hideFieldInBuffer(in_pBody, in_lenBody, "Set-Cookie");
	}
	ZKS::hideFieldInBuffer(in_pBody, in_lenBody, "Etag");
	ZKS::hideFieldInBuffer(in_pBody, in_lenBody, "Last-Modified");
}

void RZkHttpProtocol::DoneWithOneResponse(IZkContext& in_Context)
{
	if(m_listUris.size())
	{
		if(m_listUris.front() == URL_TO_BE_BLOCKED)
		{
			ZKTRACE0("Returning Bogus Img Response.\r\n\r\n");

			const char *pBuf = NULL;
			size_t bufSize = 0;

			ZkService::GetIAdBlocker()->GetBogusResponse(&pBuf, &bufSize);

			if(!pBuf)
			{
				ZKASSERT(0);
			}
			else
			{
				AppendToFreedomizedResponse(in_Context, pBuf, bufSize);
			}
			m_listUris.pop_front();
		}
	}
}

void RZkHttpProtocol::SanitizeIncomingHeader(IZkContext& in_Context, ZKSTD::string& io_strHdr)
{
#ifndef COOKIE_OBSERVER_MODE

	if(!ZkService::GetINymCentral()->IsCurrentNymUser() ||
		!ZkService::GetIClientCentral()->GetPreference(ZK_CLIENT_PREF_SET, PREF_NO_USER_COOKIE_MANAGEMENT ) )
	{
		const char* buffer = io_strHdr.c_str();

		for(unsigned int i = 0; i + 10 < io_strHdr.length(); ++i)
		{
			if( (
				(*(buffer + i) == '\n') &&
				(*(buffer + i + 1) == 'S') ||  (*(buffer + i + 1) == 's')) &&
				((*(buffer + i + 2) == 'E') || (*(buffer + (i + 2)) == 'e')) &&
				((*(buffer + i + 3) == 'T') || (*(buffer + (i + 3)) == 't')) &&
				(*(buffer + (i + 4)) == '-') &&
				((*(buffer + (i + 5)) == 'C') || (*(buffer + (i + 5)) == 'c')) &&
				((*(buffer + (i + 6)) == 'O') || (*(buffer + (i + 6)) == 'o')) &&
				((*(buffer + (i + 7)) == 'O') || (*(buffer + (i + 7)) == 'o')) &&
				((*(buffer + (i + 8)) == 'K') || (*(buffer + (i + 8)) == 'k')) &&
				((*(buffer + (i + 9)) == 'I') || (*(buffer + (i + 9)) == 'i')) &&
				((*(buffer + (i + 10)) == 'E') || (*(buffer + (i + 10)) == 'e'))
				)
			{

				// "+ 1" is to skip the '\n'
				size_t iBegin = i + 1;
				size_t iLenToDelete = io_strHdr.npos;

				i = io_strHdr.find(":", i);
				if(i != io_strHdr.npos)
				{
					i = io_strHdr.find_first_not_of(" \t", i + 1);
					if(i != io_strHdr.npos)
					{
						size_t j = io_strHdr.find("\r\n",i);
						size_t k = io_strHdr.find("\n", i);
					
						if( j == io_strHdr.npos || k <= j )
						{
							j = k;
							if(j != io_strHdr.npos)
							{
								// To include the \n
								iLenToDelete = j - iBegin + 1;
							}
						}
						else
						{
							// To include the \r\n
							iLenToDelete = j - iBegin + 2;
						}
						
						if(j != io_strHdr.npos)
						{
							ZKSTD::string strCookie = io_strHdr.substr(i, j - i);

							ZKTRACE1("COOKIE - Incoming Cookie Value:%s\n", strCookie.substr(0,128).c_str());
								
							ZkService::GetICookieManager()->AddCookie(strCookie, in_Context.GetRemoteDomainName() );
						}
					}
				}

				io_strHdr.erase(iBegin, iLenToDelete);

				// In case the next line is also "Set-Cookie", force us to re-evaluate this index.
				i = iBegin - 2;
				buffer = io_strHdr.c_str();
			}
		}
	}

	const char* buffer = io_strHdr.c_str();

	INymWebCacheMgr *pNymCacheMgr;
	ZkService::GetINymCentral()->GetActiveNym(IID_INYMWEBCACHEMGR, (void**)&pNymCacheMgr );

	for(unsigned int i = 0; i + 4 < io_strHdr.length(); ++i)
	{
		if( (
			(*(buffer + i) == '\n') &&
			(*(buffer + i + 1) == 'E') ||  (*(buffer + i + 1) == 'e')) &&
			((*(buffer + i + 2) == 'T') || (*(buffer + (i + 2)) == 't')) &&
			((*(buffer + i + 3) == 'A') || (*(buffer + (i + 3)) == 'a')) &&
			((*(buffer + (i + 4)) == 'G') || (*(buffer + (i + 4)) == 'g'))
			)
		{

			// "+ 1" is to skip the '\n'
			size_t iBegin = i + 1;
			size_t iLenToDelete = io_strHdr.npos;

			i = io_strHdr.find(":", i);
			if(i != io_strHdr.npos)
			{
				i = io_strHdr.find_first_not_of(" \t", i + 1);
				if(i != io_strHdr.npos)
				{
					size_t j = io_strHdr.find("\r\n",i);
					size_t k = io_strHdr.find("\n", i);
					
					if( j == io_strHdr.npos || k <= j )
					{
						j = k;
						if(j != io_strHdr.npos)
						{
							// To include the \n
							iLenToDelete = j - iBegin + 1;
						}
					}
					else
					{
						// To include the \r\n
						iLenToDelete = j - iBegin + 2;
					}

					if(j != io_strHdr.npos)
					{
						ZKSTD::string strEtag = io_strHdr.substr(i, j - i);

						ZKTRACE1("ETAG - Incoming Etag Value:%s\n", strEtag.substr(0,128).c_str());

						pNymCacheMgr->ProcessIncomingETag(m_listUris.front(), strEtag);
					}
				}
			}

			io_strHdr.erase(iBegin, iLenToDelete);

			i = iBegin - 2;
			buffer = io_strHdr.c_str();
		}
	}

	for(i = 0; i + 13 < io_strHdr.length(); ++i)
	{
		if((
			(*(buffer + i) == '\n') &&
			((*(buffer + i + 1) == 'L') || (*(buffer + i + 1) == 'l')) &&
			((*(buffer + i + 2) == 'A') || (*(buffer + i + 2) == 'a')) &&
			((*(buffer + i + 3) == 'S') || (*(buffer + i + 3) == 's')) &&
			((*(buffer + i + 4) == 'T') || (*(buffer + i + 4) == 't')) &&
			(*(buffer + i + 5) == '-') &&
			((*(buffer + i + 6) == 'M') || (*(buffer + i + 6) == 'm')) &&
			((*(buffer + i + 7) == 'O') || (*(buffer + i + 7) == 'o')) &&
			((*(buffer + i + 8) == 'D') || (*(buffer + i + 8) == 'd')) &&
			((*(buffer + i + 9) == 'I') || (*(buffer + i + 9) == 'i')) &&
			((*(buffer + i + 10) == 'F') || (*(buffer + i + 10) == 'f')) &&
			((*(buffer + i + 11) == 'I') || (*(buffer + i + 11) == 'i')) &&
			((*(buffer + i + 12) == 'E') || (*(buffer + i + 12) == 'e')) &&
			((*(buffer + i + 13) == 'D') || (*(buffer + i + 13) == 'd')) ))
		{

			// "+ 1" is to skip the '\n'
			size_t iBegin = i + 1;
			size_t iLenToDelete = io_strHdr.npos;

			i = io_strHdr.find(":", i);
			if(i != io_strHdr.npos)
			{
				i = io_strHdr.find_first_not_of(" \t", i + 1);
				if(i != io_strHdr.npos)
				{
					size_t j = io_strHdr.find("\r\n",i);
					size_t k = io_strHdr.find("\n", i);
				
					if( j == io_strHdr.npos || k <= j )
					{
						j = k;
						if(j != io_strHdr.npos)
						{
							// To include the \n
							iLenToDelete = j - iBegin + 1;
						}
					}
					else
					{
						// To include the \r\n
						iLenToDelete = j - iBegin + 2;
					}
					
					if(j != io_strHdr.npos)
					{
						ZKSTD::string strLastModified = io_strHdr.substr(i, j - i);

						ZKTRACE1("LAST-MODIFIED - Incoming Last-Modified Value:%s\n", strLastModified.substr(0,128).c_str());

						pNymCacheMgr->ProcessIncomingLastDateModified(m_listUris.front(), strLastModified);
					}
				}
			}

			io_strHdr.erase(iBegin, iLenToDelete);

			i = iBegin - 2;
			buffer = io_strHdr.c_str();
		}
	}

	ZKSTD::string strTagForBrowser;
	if(m_listUris.size() && pNymCacheMgr->GetETagToGiveToBrowser(m_listUris.front(), strTagForBrowser))
	{
		i = io_strHdr.find("\r\n\r\n");

		if(i != io_strHdr.npos)
		{
			strTagForBrowser = "\r\n" + strTagForBrowser;
			io_strHdr.insert(i, strTagForBrowser);
		}
		else
		{
			i = io_strHdr.find("\n\n");
			if( i != io_strHdr.npos )
			{
				strTagForBrowser = "\n" + strTagForBrowser;
				io_strHdr.insert( i, strTagForBrowser );
			}
			else
			{
				i = io_strHdr.find("\n\r\n");
				if( i != io_strHdr.npos )
				{
					strTagForBrowser = "\r\n" + strTagForBrowser;
					io_strHdr.insert( i, strTagForBrowser );
				}
				else
				{
					ZKASSERT( 0 );
				}
			}
		}
	}
	if(m_listUris.size() && pNymCacheMgr->GetLastModifiedTagToGiveToBrowser(m_listUris.front(), strTagForBrowser))
	{
		i = io_strHdr.find("\r\n\r\n");

		if(i != io_strHdr.npos)
		{
			strTagForBrowser = "\r\n" + strTagForBrowser;
			io_strHdr.insert(i, strTagForBrowser);
		}
		else
		{
			i = io_strHdr.find("\n\n");
			if( i != io_strHdr.npos )
			{
				strTagForBrowser = "\n" + strTagForBrowser;
				io_strHdr.insert( i, strTagForBrowser );
			}
			else
			{
				i = io_strHdr.find("\n\r\n");
				if( i != io_strHdr.npos )
				{
					strTagForBrowser = "\r\n" + strTagForBrowser;
					io_strHdr.insert( i, strTagForBrowser );
				}
				else
				{
					ZKASSERT( 0 );
				}
			}
		}
	}
#endif //#ifndef COOKIE_OBSERVER_MODE
}



int RZkHttpProtocol::FindHeaderEnd(const char *in_pData, const size_t in_len)
{

	if(m_strCurPart.length() > 4)
	{
		ZKSTD::string strTemp = m_strCurPart.substr(m_strCurPart.length() - 4) + ZKSTD::string(in_pData, 4);
		size_t i = strTemp.find("\r\n\r\n");
		if(i != strTemp.npos)
		{
			return i;
		}
		else
		{
			i = strTemp.find( "\n\n" );
			if( i != strTemp.npos )
			{
				return i - 2;
			}
			else
			{
				i = strTemp.find( "\n\r\n" );
				if( i != strTemp.npos )
				{
					return i - 1;
				}
			}
		}

	}

	for(size_t i = 0; i + 1 < in_len; ++i)
	{
		if( i + 3 < in_len && *(in_pData + i) == '\r' && *(in_pData + i + 1) == '\n' &&
			*(in_pData + i + 2) == '\r' && *(in_pData + i + 3) == '\n')
		{
			return i + 4;
		}

		if( *(in_pData + i) == '\n' && *(in_pData + i + 1) == '\n' )
		{
			return i + 2;
		}
		if( i + 2 < in_len && *(in_pData + i) == '\n' && *(in_pData + i + 1) == '\r' &&
			*(in_pData + i + 2) == '\n' )
		{
			return i + 3;
		}
	}

	if(m_ResponseState == READING_FOOTER &&	m_strCurPart.length() == 0 && in_len >= 2 &&
				in_pData[0] == '\r' && in_pData[1] == '\n')
	{
		return 2;
	}

	else if (m_ResponseState == READING_FOOTER && m_strCurPart.length() == 0 && in_len >= 1 &&
				in_pData[0] == '\n' )
	{
		return 1;
	}

	return -1;
}



int RZkHttpProtocol::FindLineEnd(const char *in_pData, const size_t in_len)
{
	
	if(m_strCurPart.length() && m_strCurPart[m_strCurPart.length() - 1] == '\r'
		&& *in_pData == '\n')
	{
		return 1;
	}

	for(size_t i = 0; i + 1 < in_len; ++i)
	{
		if( *(in_pData + i) == '\r' && *(in_pData + i + 1) == '\n' )
		{
			return i + 2;
		}
	}

	const char *pRes = (const char *)memchr(in_pData, '\n', in_len);
	if(pRes)
	{
		return pRes - in_pData + 1;
	}

	return -1;
	
}

bool RZkHttpProtocol::ProcessRequest(IZkContext& in_Context)
{
	bool ReturnValue = true;
	
	ZkHTTPField field = NULL;
	const unsigned char * value = NULL;

	ZKVERIFY(zkHTTPHeaderResetFieldSearch(in_Context.GetZkError(),
		m_RequestHeader));
	
	if(zkHTTPHeaderGetField(in_Context.GetZkError(),
		m_RequestHeader, Host, &field)) {

		if (!zkHTTPFieldValue(in_Context.GetZkError(), field, &value))
		{
			ZKASSERT(false);
		}
		else
		{
			in_Context.SetRemoteDomainName(std::string((char *)value));
		}
	}

	if(ShouldBlockRequest(in_Context))
	{
		if(m_ResponseState == READING_RESPONSE_HEADER && !m_strCurPart.length()
								&& !m_listUris.size())
		{	// Go ahead and respond to it now.
			ZKVERIFY(PlaceBogusResponseInContext(in_Context));
		}
		else
		{	// Pop a special token on the back, and it will be responded to in due time...
			m_listUris.push_back(ZKSTD::string(URL_TO_BE_BLOCKED));
			in_Context.SetProxyAction(IZkContext::ZKPROXY_DO_NOTHING);
		}

	}
	else
	{
		in_Context.SetProxyAction(IZkContext::ZKPROXY_SEND_TO_SERVER);

		bool bFoundConditional = false;

		ZKSTD::string strUri;
		GetRequestUri(strUri, in_Context);

		ZKASSERT(strUri.length());
		m_listUris.push_back(strUri);

		ZKVERIFY(zkHTTPHeaderResetFieldSearch(in_Context.GetZkError(),
			m_RequestHeader));
		
		if(zkHTTPHeaderGetField(in_Context.GetZkError(),
			m_RequestHeader, If_None_Match, &field))
		{
			bFoundConditional = true;
		}

		ZKVERIFY(zkHTTPHeaderResetFieldSearch(in_Context.GetZkError(),
			m_RequestHeader));
		
		if(zkHTTPHeaderGetField(in_Context.GetZkError(),
			m_RequestHeader, If_Modified_Since, &field))
		{
			bFoundConditional = true;
		}

		zkHTTPHeaderStripFieldByID(in_Context.GetZkError(),
								m_RequestHeader, If_Modified_Since);

		zkHTTPHeaderStripFieldByID(in_Context.GetZkError(),
								m_RequestHeader, If_Match);

		zkHTTPHeaderStripFieldByID(in_Context.GetZkError(),
								m_RequestHeader, If_None_Match);

		if(!ZkService::GetINymCentral()->IsCurrentNymUser() ||
				!ZkService::GetIClientCentral()->GetPreference(ZK_CLIENT_PREF_SET,
												PREF_NO_USER_COOKIE_MANAGEMENT ))
		{
			if(ZkService::GetINymCentral()->IsCurrentNymUser() )
			{
				// only strip cookies
				zkHTTPHeaderStripFieldByID(in_Context.GetZkError(),
										m_RequestHeader, Cookie);
			}

			// Strip ALL freedom offensive fields (including cookies)
			else
			{
				zkHTTPHeaderStripField(in_Context.GetZkError(),
							m_RequestHeader, DONT_STRIP_UNKNOWN);
			}

			if(ReturnValue)
			{
				// Call CookieFieldManager to add in any appropriate cookies
				if(!ManageOutgoingCookies(in_Context) )
				{
					ReturnValue = false;	
				}
			}
		}
		if(bFoundConditional && ReturnValue)
		{
			INymWebCacheMgr *pNymCacheMgr = NULL;
			ZkService::GetINymCentral()->GetActiveNym(IID_INYMWEBCACHEMGR, (void**)&pNymCacheMgr);

			ZKSTD::string strFieldName;
			ZKSTD::string strFieldValue;

			if(pNymCacheMgr->GetOutgoingTagForUrl(strUri, strFieldName, strFieldValue))
			{
				if(!zkHTTPHeaderAddFieldData(in_Context.GetZkError(),
								m_RequestHeader,
								(const unsigned char*)strFieldName.c_str(),
								(const unsigned char*)strFieldValue.c_str()))
				{
					ZKASSERT(0);
					ReturnValue = false;	
				}
			}
		}
		if(ReturnValue)
		{
			PlaceFreedomizedRequestInContext(in_Context);
		}


	}
	return ReturnValue;
}

bool RZkHttpProtocol::ManageOutgoingCookies(IZkContext& in_Context)
{
	bool ReturnValue = true;
	
	ZKSTD::string domain = in_Context.GetRemoteDomainName();

	ZkHTTPField field = NULL;

	bool FoundCookies = false;
	ZKSTD::string strCookie;

	FoundCookies = ZkService::GetICookieManager()->BuildCookieField(domain, strCookie);
	
	if(!FoundCookies)
	{
		
	}

	else if (!zkHTTPFieldNew(in_Context.GetZkError(),&field))
	{
		ReturnValue = false;
		ZKASSERT(false);
	}
				
	else if(!zkHTTPFieldSetId(in_Context.GetZkError(), field, Cookie))
	{
		ReturnValue = false;
		ZKASSERT(false);
	}

	else if(!zkHTTPFieldSetValue(in_Context.GetZkError(),
					field, (const unsigned char *)strCookie.c_str() ))
	{
		ReturnValue = false;
		ZKASSERT(false);
	}
	else if (!zkHTTPHeaderAddField(in_Context.GetZkError(),
					m_RequestHeader,field))
	{
		ReturnValue = false;
		ZKASSERT(false);
	}

	return ReturnValue;
}

bool RZkHttpProtocol::PlaceBogusResponseInContext(IZkContext& in_Context)
{
	bool ReturnValue = true;
	const char *pBuf = NULL;

	size_t bufSize = 0;
	ZkService::GetIAdBlocker()->GetBogusResponse(&pBuf, &bufSize);
	if(!pBuf)
	{
		ReturnValue = false;
		ZKASSERT(0);
	}
	else
	{
		ZKTRACE0("Returning Bogus Img Response.\r\n\r\n");

		in_Context.SetProxyAction( IZkContext::ZKPROXY_SEND_TO_CLIENT );
		in_Context.SetProtocolOutput((void*)pBuf);
		in_Context.SetProtocolOutputLength(bufSize);
	}
	return ReturnValue;
}

int RZkHttpProtocol::GetResponseStatusCode(ZKSTD::string& in_strHdr)
{
	size_t i = in_strHdr.find_first_not_of(' ');
	if(i != ZKSTD::string::npos)
	{
		i = in_strHdr.find_first_of(' ', i);
		if(i != ZKSTD::string::npos)
		{
			i = in_strHdr.find_first_not_of(' ', i);
			if(i != ZKSTD::string::npos)
			{
				size_t j = in_strHdr.find_first_of(' ', i);
				if(j != ZKSTD::string::npos)
				{
					j -= i;
				}
				return atoi(in_strHdr.substr(i, j).c_str());
				
			}
		}
	}

	return 0;
}

void RZkHttpProtocol::GetRequestUri(ZKSTD::string& out_strUri, IZkContext& in_Context)
{
	const char *strRequest = NULL;
	char *strUriInRequest = NULL;

	zkHTTPHeaderRequest(in_Context.GetZkError(), m_RequestHeader, &strRequest);
	if(strRequest)
	{
		strUriInRequest = strchr(strRequest, ' ');
	}

	if(strUriInRequest)
	{
		char *strEndUriInRequest;

		strUriInRequest++; // Skip White Space
		strEndUriInRequest= strchr(strUriInRequest, ' ');
		if(strEndUriInRequest)
		{
			out_strUri = in_Context.GetRemoteDomainName();
			out_strUri += ZKSTD::string(strUriInRequest, strEndUriInRequest - strUriInRequest);
		}
	}
}

bool RZkHttpProtocol::ShouldBlockRequest(IZkContext& in_Context)
{
	bool ReturnValue = false;

	if ( !ZkService::GetIClientCentral()->GetPreference(ZK_CLIENT_PREF_SET, PREF_NO_BLOCK_ADS))
	{
		ZKSTD::string strURI;
		GetRequestUri(strURI, in_Context);
		if(!strURI.length())
		{
			ZKASSERT(0);
			ReturnValue = true;
		}
		else
		{
			ReturnValue = ZkService::GetIAdBlocker()->UriIsBlocked(strURI.c_str());
		}
	}
	return ReturnValue;
	
}

bool RZkHttpProtocol::PlaceFreedomizedRequestInContext(IZkContext& in_Context)
{
	bool ReturnValue = true;

	size_t lenHttpData = 0;
	void *pHttpData = NULL;

	if(!zkMemBufNew(in_Context.GetZkError(), &in_Context.GetZkMemBuf() ))
	{

	}

	// Ok, time to dump out the Freedomized request
	else if(!zkHTTPHeaderDump(in_Context.GetZkError(), 
		m_RequestHeader, HTTPWriteBuf, in_Context.GetZkMemBuf()))
	{
		ReturnValue = false;
		ZKASSERT(false);
	}
	else if(!zkMemBufAppend(in_Context.GetZkError(), 
		in_Context.GetZkMemBuf(), "\0", 1))
	{
		ReturnValue = false;
		ZKASSERT(false);
	}
				
	else if(!(zkMemBufGetData(in_Context.GetZkError(), 
		in_Context.GetZkMemBuf(), (void**)&pHttpData, &lenHttpData)))
	{
		ReturnValue = false;
		ZKASSERT(false);
	}

	
					
	else
	{
		in_Context.SetProtocolOutput(pHttpData);
		// -1 to remove the trailing NULL
		in_Context.SetProtocolOutputLength(lenHttpData - 1);
		
		ZKSTD::vector<ZKSTD::string> vecstrOffensiveWordsFound;
					
		// Check for Freedom offensive material in the body of the Request											
		if( ZkService::GetIWordScan()->WordScan(
				(unsigned char*)in_Context.GetProtocolOutput(),
				in_Context.GetProtocolOutputLength(),
				vecstrOffensiveWordsFound) 
				)
		{
						
			WordScanSummary Result ;
			Result.pScannedBuffer = (unsigned char*)in_Context.GetProtocolOutput();
			Result.cntBuffer = in_Context.GetProtocolOutputLength();
			Result.pFoundWords = &vecstrOffensiveWordsFound ;
			Result.strDomainName = in_Context.GetRemoteDomainName();
			
			Result.m_SockAddr = *( (sockaddr*)(&in_Context.GetRemoteAddress().m_inet.sa) );

            if( !ZkProxyDialogs::LaunchWordScanMsgDlg( PROTOCOL_HTTP, &Result ) )
            {
				ReturnValue = false;
				in_Context.SetProxyAction(IZkContext::ZKPROXY_DO_NOTHING);
			}
		}
	}
	return ReturnValue;
}

bool RZkHttpProtocol::ResetRequestHeader(IZkContext& in_Context)
{
	bool ReturnValue = false;
	if(!zkHTTPHeaderFree(in_Context.GetZkError(), &m_RequestHeader))
	{
	}
	else if(!zkHTTPHeaderNew(in_Context.GetZkError(), &m_RequestHeader))
	{
	}
	else if(!zkHTTPHeaderKeepBodyData(in_Context.GetZkError(), m_RequestHeader))
	{
	}
	else if(!zkHTTPHeaderSetType(in_Context.GetZkError(), m_RequestHeader, REQUEST_HEADER))
	{
	}
	else
	{
		ReturnValue = true;
	}
	return ReturnValue;
}
