// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#pragma warning (disable: 4251)

#include "RZkNntpProtocol.h"
#include "ZkEmailHelp.h"
#include <ZkProxyDialogs.h>
#include <zkEmailUtil.h>
#include <zkMemBuf.h>
#include <ZkService.h>
#include <INymCentral.h>
#include <INym.h>
#include <IZKErrorReporter.h>
#include <IZkException.h>
#include <IZkConsoleMsg.h>
#include <IClientInfo.h>
#include <IWordScan.h>
#include <zkCStr.h>
// Marcp: this need to be fixed

#ifdef WIN32
#include "../win/ZKStatusAVI.h"
class CWnd;
#else
#include <ZkStatus.h>
#warning Make win version use new API.
#endif

#include <ZkErrorHandler.h>
#include <lockguard.h>

WZkMutex RZkNntpProtocol::m_gMutex;

#define POST_ACCEPTED_RESPONSE		"240 Article Posted OK\r\n"
#define CONTINUE_POST_RESPONSE		"340 Send Article\r\n"
#define POST_CMD_STRING				"POST\r\n"
#define POST_TERMINATION			"\r\n.\r\n"
#define USER_CANCELLED_RESPONSE		"441 User canceled posting operation\r\n"

#define SIMPLE_ERROR_STRING					"441 Freedom experienced a problem sending posting article. Article not sent.\r\n"
#define ERROR_DETAIL_FORMAT_STRING			"441 Freedom experienced a problem sending posting article. Article not sent (%s).\r\n"
#define ERROR_DETAIL_AND_FUNC_FORMAT_STRING	"441 Freedom experienced a problem sending posting article. Article not sent (%s in %s).\r\n"

extern "C" {
  int NNTPWriteBuf ( ZkError , void * , const unsigned char* ,int );
}	
int NNTPWriteBuf ( 
    ZkError err ,
    void * cdata ,
    const unsigned char * data ,
    int dlen )
{
    zkErrorClear(err);
    if (cdata && data && (dlen > 0)) {
		zkMemBufAppend(err, (ZkMemBuf)cdata, (void*)data, dlen);
	}
    return 1;
}

RZkNntpProtocol::RZkNntpProtocol()
				: m_Email(NULL)
				, m_State(FRESH_STATE)
{
}

bool RZkNntpProtocol::InitProtocolResources		(IZkContext& in_Context)
{
	bool ReturnValue = true;
	if( !zkEmailNew(in_Context.GetZkError(), &m_Email) )
	{
		ReturnValue = false;
		ZKASSERT(false);
	}
	else if( !zkEmailScanForSMTPTermination(in_Context.GetZkError(), m_Email) )
	{
		ReturnValue = false;
		ZKASSERT(false);
	}

	return ReturnValue;
}

bool RZkNntpProtocol::FreeProtocolResources		(IZkContext& in_Context)
{
	if( !zkEmailFree(in_Context.GetZkError(), &m_Email) )
	{
		ZKASSERT(false);

		return false;
	}
	return true;
}

bool RZkNntpProtocol::ConnectFailed(IZkContext& in_Context)
{
    in_Context.SetCloseConnection(true);
    return true;
}

bool RZkNntpProtocol::ClientReady(IZkContext& in_Context)
{
	in_Context.SetProxyAction( IZkContext::ZKPROXY_CONNECT_TO_SERVER );
	return true;
}

bool RZkNntpProtocol::FreedomizeClientRequest	(IZkContext& in_Context)
{
	ZKTRACE1("NNTP Client Says: %s\r\n", (char*)in_Context.ByteBuffer());
	in_Context.SetProxyAction(IZkContext::ZKPROXY_DO_NOTHING);

	if(m_State == RECEIVING_POST_DATA)
	{
		HandleMoreData(in_Context);
	}
	else if(!XP_STRNICMP((char*)in_Context.ByteBuffer(), "POST", 4))
	{
		HandlePostCommand(in_Context);
	}
	else
	{
		in_Context.SetProxyAction( IZkContext::ZKPROXY_SEND_TO_SERVER);

		in_Context.SetProtocolOutput(in_Context.ByteBuffer());
		in_Context.SetProtocolOutputLength(in_Context.GetByteBufferLength());
	}
	TraceOutgoingStream(in_Context);
	return true;
}

bool RZkNntpProtocol::FreedomizeServerResponse	(IZkContext& in_Context)
{
	ZKTRACE1("NNTP Server Says: %s\r\n", (char*)in_Context.ByteBuffer());
	in_Context.SetProxyAction( IZkContext::ZKPROXY_DO_NOTHING);

	switch(m_State)
	{
	case FRESH_STATE:
		// Pass most responses directly on to the client.
		PassResponseThroughToClient(in_Context);
		break;

	case SENT_POST_COMMAND:
		if(!XP_STRNICMP((char*)in_Context.ByteBuffer(), "340", 3))
		{ // 340, send article to be posted. End with <CR-LF>.<CR-LF>
			SendMessageBodyToServer(in_Context);
		}
		else
		{
			// Pass error directly on to the client.
			PassResponseThroughToClient(in_Context);
			ZKVERIFY(ResetState(in_Context));
		}
		break;

	case SENT_MESSAGE_BODY:
		// Pass Results directly on to the client.
		PassResponseThroughToClient(in_Context);
		ZKVERIFY(ResetState(in_Context));
		break;
		
	default:
		ErrorResponse(in_Context);
		ZKASSERT(0);
		break;
	}
	TraceOutgoingStream(in_Context);
	return true;
}

void RZkNntpProtocol::HandlePostCommand(IZkContext& in_Context)
{
	SetResponse(in_Context, CONTINUE_POST_RESPONSE);
	m_State = RECEIVING_POST_DATA;
}

void RZkNntpProtocol::HandleMoreData(IZkContext& in_Context)
{
	if( !zkEmailAddText(in_Context.GetZkError(), m_Email, 
				(unsigned char*)in_Context.ByteBuffer(),
				in_Context.GetByteBufferLength()) )
	{
		// Error response returns the ZkError->detail and resets our state
		ErrorResponse(in_Context);
		ZKASSERT(false);
	}
	else
	{
		int iStatus = 0;
		if( !zkEmailStatus(in_Context.GetZkError(), m_Email, &iStatus) )
		{
			ErrorResponse(in_Context);
			ZKASSERT(false);
		}
		else
		{
			if(iStatus == PARSE_COMPLETE)
			{
				HandleCompletePost(in_Context);
			}
		}
	}
}

void RZkNntpProtocol::HandleCompletePost(IZkContext& in_Context)
{
	ZKSTD::vector<ZKSTD::string> vecGroupsToPost;

	ZkEmailHeader	EmailHeader;
	ZkEmailField	NewsgroupsField;
	const char*		NewsgroupsValue;
	
	if(!zkEmailHeader(in_Context.GetZkError(), m_Email, &EmailHeader))
	{
		ErrorResponse(in_Context);
		ZKASSERT(false);
	}
	else if(!zkEmailHeaderGetField(in_Context.GetZkError(), EmailHeader, Newsgroups, &NewsgroupsField))
	{
		ErrorResponse(in_Context);
		ZKASSERT(false);
	}
	else if(!zkEmailFieldValue(in_Context.GetZkError(), NewsgroupsField, (const unsigned char**)&NewsgroupsValue))
	{
		ErrorResponse(in_Context);
		ZKASSERT(false);
	}
	else
	{
		ZKSTD::string::size_type pos = 0;
		ZKSTD::string::size_type i   = 0;
		ZKSTD::string			 strNewsgroupsValue = NewsgroupsValue;

		do
		{
			pos = strNewsgroupsValue.find_first_of(',', i);
			if(pos == ZKSTD::string::npos)
			{
				pos = strNewsgroupsValue.length();
			}
			vecGroupsToPost.push_back( strNewsgroupsValue.substr(i, pos - i) );

			i = pos + 1;
		}while( pos < strNewsgroupsValue.length() );
		
		if(vecGroupsToPost.empty() )
		{
			// Should this error explicitly say, no newsgroups specified?
			ErrorResponse(in_Context);
			ZKASSERT(false);
		}
        else 
        {
            long lResult = ZkProxyDialogs::LaunchOutgoingMsgDlg( &vecGroupsToPost,
                                                                 false );

            if( lResult == -1 )
            {
                SetResponse(in_Context, USER_CANCELLED_RESPONSE);
                ResetState(in_Context);
            }
            else if ( lResult == USER_NYM_INDEX )
            {
                PostAsUser(in_Context, vecGroupsToPost);
            }
            else 
            {
                PostAsNym(in_Context, vecGroupsToPost, lResult);
            }
        }
	}
}

void RZkNntpProtocol::NotifyConsoleOfPost(const ZKSTD::vector<ZKSTD::string>& in_vecGroups, const size_t in_iNym)
{
	ZKSTD::string strGroups;
	for(size_t i = 0; i < in_vecGroups.size() && i < 3; i++)
	{
		if(i > 0)
		{
			if(in_vecGroups.size() > 2)
			{
				strGroups += ", ";
			}
			else
			{
				strGroups += ' ';
			}
		}
		if(i + 1 == in_vecGroups.size() && in_vecGroups.size() > 1)
		{
			strGroups += "and ";
		}
		strGroups += in_vecGroups[i];
		if(i == 2 && in_vecGroups.size() > 3)
		{
			strGroups += "...";
		}
	}
	ZKSTD::string strNymName;
	ZkService::GetINymCentral()->GetNymName(in_iNym, strNymName);
	IZkConsoleMsg::ShowImportantMsg(CONSIMAGE_MAIL, IDS_CONS_POST_FROM_NYMNAME_TO_GROUPS, strNymName.c_str(), strGroups.c_str());
}

void RZkNntpProtocol::PostAsUser(IZkContext& in_Context, const ZKSTD::vector<ZKSTD::string>& in_vecGroups)
{
	if(!ScanPostForOffensiveWords(in_Context, USER_NYM_INDEX))
	{
		SetResponse(in_Context, USER_CANCELLED_RESPONSE);
		ResetState(in_Context);
	}
	else
	{
		NotifyConsoleOfPost(in_vecGroups, USER_NYM_INDEX);
		SendPostCommandToServer(in_Context);
	}
}

void RZkNntpProtocol::SendPostCommandToServer(IZkContext& in_Context)
{
	in_Context.SetProxyAction( IZkContext::ZKPROXY_SEND_TO_SERVER);

	in_Context.SetProtocolOutput(POST_CMD_STRING);
	in_Context.SetProtocolOutputLength(strlen(POST_CMD_STRING));

	m_State = SENT_POST_COMMAND;
}

bool RZkNntpProtocol::ResetState(IZkContext& in_Context)
{
	bool ReturnValue = true;

	m_State = FRESH_STATE;
	if(!zkEmailFree(in_Context.GetZkError(), &m_Email))
	{
		ReturnValue = false;
		ZKASSERT(false);
	}
	else if( !zkEmailNew(in_Context.GetZkError(), &m_Email) )
	{
		ReturnValue = false;
		ZKASSERT(false);
	}
	else if( !zkEmailScanForSMTPTermination(in_Context.GetZkError(), m_Email) )
	{
		ReturnValue = false;
		ZKASSERT(false);
	}

	return ReturnValue;
}

void RZkNntpProtocol::SetResponse(IZkContext& in_Context, const char* strResponse)
{
	in_Context.SetProtocolOutput((void *)strResponse);
	in_Context.SetProtocolOutputLength(strlen(strResponse));

	in_Context.SetProxyAction(IZkContext::ZKPROXY_SEND_TO_CLIENT);
}

void RZkNntpProtocol::ErrorResponse (IZkContext& in_Context)
{
	in_Context.SetProxyAction( IZkContext::ZKPROXY_SEND_TO_CLIENT );

	unsigned int lenDetail = strlen( in_Context.GetZkError()->getdetail().c_str() );
	unsigned int lenFunc = strlen( in_Context.GetZkError()->getfnc().c_str() );

	if ( lenDetail > 0 )
	{
		if ( lenFunc > 0 )
		{
			// Function name is a bit ugly to show to the user, we only show
			// this if we have one from the unix guys
			snprintf(	(char *)in_Context.ByteBuffer(), 
						in_Context.ByteBufferSize(), 
						ERROR_DETAIL_AND_FUNC_FORMAT_STRING,
						in_Context.GetZkError()->getdetail().c_str(), 
						in_Context.GetZkError()->getfnc().c_str() );
		}
		else
		{
			
			snprintf(	(char *)in_Context.ByteBuffer(), 
						in_Context.ByteBufferSize(), 
						ERROR_DETAIL_FORMAT_STRING,
						in_Context.GetZkError()->getdetail().c_str());
		}
	}
	else
	{
		zkCStrlcpy((char *)in_Context.ByteBuffer(), SIMPLE_ERROR_STRING, in_Context.ByteBufferSize());
	}

	in_Context.SetProtocolOutput(in_Context.ByteBuffer());
	in_Context.SetProtocolOutputLength(strlen((char*)in_Context.ByteBuffer()));

	ZKVERIFY(ResetState(in_Context));
}

void RZkNntpProtocol::PassResponseThroughToClient(IZkContext& in_Context)
{
	in_Context.SetProtocolOutput(in_Context.ByteBuffer());
	in_Context.SetProtocolOutputLength(in_Context.GetByteBufferLength() );
	in_Context.SetProxyAction( IZkContext::ZKPROXY_SEND_TO_CLIENT );
}

void RZkNntpProtocol::SendMessageBodyToServer(IZkContext& in_Context)
{
	ZkMemBuf memBuf;
	void* pPostData = NULL;
	unsigned int lenPostData = 0;

	if(!zkEmailParse(in_Context.GetZkError(), m_Email))
	{
		ErrorResponse(in_Context);
		ZKASSERT(false);
	}
	else if(!zkMemBufNew(in_Context.GetZkError(), &memBuf ))
	{
		ErrorResponse(in_Context);
		ZKASSERT(false);
	}
	else if(!zkEmailDump(in_Context.GetZkError(), m_Email, NNTPWriteBuf, memBuf))
	{
		ErrorResponse(in_Context);
		ZKASSERT(false);
	}
	else if(!zkMemBufAppend(in_Context.GetZkError(), memBuf, POST_TERMINATION, strlen(POST_TERMINATION)))
	{
		ErrorResponse(in_Context);
		ZKASSERT(false);
	}
	// For debugging, add a zero so it can be traced out
	// BUT remember to subtract 1 from the size when you give it to
	// SetProtocolOutputLength
	else if(!zkMemBufAppend(in_Context.GetZkError(), memBuf, "\0", 1))
	{
		ErrorResponse(in_Context);
		ZKASSERT(false);
	}
	else if(!zkMemBufGetData(in_Context.GetZkError(), memBuf, &pPostData, &lenPostData))
	{
		ErrorResponse(in_Context);
		ZKASSERT(false);
	}
	else
	{
		in_Context.SetProtocolOutput(pPostData);
		// Why -1? See above.
		in_Context.SetProtocolOutputLength(lenPostData - 1);
		in_Context.SetProxyAction(IZkContext::ZKPROXY_SEND_TO_SERVER);
		m_State = SENT_MESSAGE_BODY;

		// Let the proxy deallocate the memBuf when it's done sending the data.
		in_Context.GetZkMemBuf() = memBuf;

	}
}

void RZkNntpProtocol::PostAsNym(IZkContext& in_Context,
								const ZKSTD::vector<ZKSTD::string>& in_vecGroups,
								size_t iOriginatingNym)
{
#ifdef WIN32
	// Obsolete in the new windows UI
	//	ZKStatusAVI			BusyWithMail( ZK_STATUS_CRYPTO_BUSY ) ;
#else
	ZkStatus			BusyWithMail( ZK_STATUS_CRYPTO_BUSY ) ;
#endif

	INym*				pNym = NULL;
	ZKSTD::string		strOriginatingNym;
	ZKSTD::string		strFreedomDomain;
	ZKVERIFY(ZkService::GetINymCentral()->GetNym(IID_INYM, (void**)&pNym, iOriginatingNym));
	ZKVERIFY(pNym->GetNymEmailDomain(strFreedomDomain));
	ZKVERIFY(pNym->GetNymName( strOriginatingNym ));

	ZKSTD::string			strFrom( strOriginatingNym + "@" + strFreedomDomain);
	ZKSTD::string			strToFieldValue("mail2news");

	// May be different from Nym's (nym may be @trial.freedom.net..)
	ZkService::GetIClientInfo()->GetFreedomDomain(strFreedomDomain);

	strToFieldValue += "@" + strFreedomDomain;

	ZKSTD::string			strNewsgroupValue;
	
	ZkList				ListRecipients = NULL;
	
	ZkEmailHeader		EmailHeader = NULL;
	ZkEmailField		EmailField = NULL;
	
	// Should make this a member variable so it can be shared by the
	// wordscan function. I'll do it when the sun comes up.
	ZkMemBuf			MemBuf = NULL; 

	char *pData = NULL;
	size_t lenData = 0;
	if(!zkEmailHeader(in_Context.GetZkError(), m_Email, &EmailHeader))
	{
		ErrorResponse(in_Context);
		ZKASSERT(false);
	}
	else if(!zkEmailHeaderAddFieldData(in_Context.GetZkError(), EmailHeader, (const unsigned char*)"To", (const unsigned char*)strToFieldValue.c_str()))
	{ // Format a correct "To:" field using the mail2news gateway obtained from globfree
		ErrorResponse(in_Context);
		ZKASSERT(false);
	}
	else if(!zkEmailHeaderGetField(in_Context.GetZkError(), EmailHeader, From, &EmailField))
	{
		ErrorResponse(in_Context);
		ZKASSERT(false);
	}
	else if(!zkEmailFieldSetValue(in_Context.GetZkError(), EmailField, (const unsigned char*) strFrom.c_str()))
	{
		ErrorResponse(in_Context);
		ZKASSERT(false);
	}
	if(!zkEmailStripHeader(in_Context.GetZkError(), m_Email, DONT_STRIP_NEWSGROUPS) )
	{// Function removes malicious fields
		ErrorResponse(in_Context);
		ZKASSERT(false);
	}
	else if(!ScanPostForOffensiveWords(in_Context, iOriginatingNym))
	{
		SetResponse(in_Context, USER_CANCELLED_RESPONSE);
		ResetState(in_Context);
	}
	else if(!zkMemBufNew(in_Context.GetZkError(), &MemBuf ))
	{
		ZKASSERT(false);
	}
	else if(!zkEmailDump(in_Context.GetZkError(), m_Email,
						NNTPWriteBuf, MemBuf))
	{
		ZKASSERT(false);
	}
	// So that the Msg may be treated as a string by the ZkEmailHelp and EsmtpClient
	else if(!zkMemBufAppend(in_Context.GetZkError(), MemBuf, "\0", 1 ))
	{
		ZKASSERT(false);
	}
	else if(!zkMemBufGetData(in_Context.GetZkError(), MemBuf, (void**)&pData, &lenData))
	{
		ZKASSERT(false);
	}
	else
	{
		ZKSTD::vector<ZKSTD::string> vecInet;
		ZKSTD::vector<ZKSTD::string> vecFreedom;
		vecInet.push_back(strToFieldValue);
		ZKASSERT(pData[lenData - 1] == 0);

		NotifyConsoleOfPost(in_vecGroups, iOriginatingNym);

		if(ZkEmailHelp::SendEmailESMTP(in_Context.GetZkError(), vecInet, vecFreedom, pData, iOriginatingNym))
		{
			SetResponse(in_Context, POST_ACCEPTED_RESPONSE);
			ResetState(in_Context);

			IZkConsoleMsg::ShowImportantMsg( CONSIMAGE_MAIL, IDS_CONS_PROXY_NNTP_POST_DONE, strFrom.c_str() );
		}
		else
		{
			ErrorResponse(in_Context);
			ZKASSERT(false);

			IZkConsoleMsg::ShowImportantMsg( CONSIMAGE_MAIL, IDS_CONS_PROXY_NNTP_POST_FAIL, strFrom.c_str() );
		}
	}

	if(ListRecipients)
	{
		zkListFree(in_Context.GetZkError(), &ListRecipients);
	}
	if(MemBuf)
	{
		zkMemBufFree(in_Context.GetZkError(), &MemBuf);
	}
}

bool RZkNntpProtocol::ScanPostForOffensiveWords(IZkContext& in_Context, size_t in_NymIndex)
{
	ZKS::Lock_Guard<WZkMutex> gate(m_gMutex);	

	ZKSTD::vector<ZKSTD::string> vecStrOffensiveWordsFound;
	bool ReturnValue = true;
	WAutoBuffer EmailBuffer;

	if(!zkEmailParse(in_Context.GetZkError(), m_Email))
	{
		ZKASSERT(false);
		ReturnValue = false;
	}
	else if(!zkEmailBuffer(in_Context.GetZkError(), m_Email , (char **)&EmailBuffer, EmailBuffer.GetSizePointer()))
	{

		ZKASSERT(false);
		ReturnValue = false;
	}
	else if(ZkService::GetIWordScan()->WordScan(
				EmailBuffer,
				EmailBuffer.Size(),
				vecStrOffensiveWordsFound,
				in_NymIndex) )
	{ 			
		// Launch an Word Scan Alert Dialog or other UI reaction for the user

		WordScanSummary Result ;
		Result.pScannedBuffer = EmailBuffer;
		Result.cntBuffer = EmailBuffer.Size();
		Result.pFoundWords = &vecStrOffensiveWordsFound ;
		// This socket address is for the user's SMTP server
		// The message is actually going to the FMG.
		//Result.m_SockAddr = GetOriginalDestinationSocketAddress();

        if( !ZkProxyDialogs::LaunchWordScanMsgDlg( PROTOCOL_NNTP1, &Result ) )
        {
            ReturnValue = false;
        }
	}

	return ReturnValue;
}

void RZkNntpProtocol::TraceOutgoingStream(IZkContext& in_Context)
{
	switch(in_Context.GetProxyAction())
	{
	case IZkContext::ZKPROXY_SEND_TO_CLIENT:
		ZKTRACE1("Telling NNTP Client: %s\r\n", (char*)in_Context.GetProtocolOutput());
		break;
	case IZkContext::ZKPROXY_SEND_TO_SERVER:
		ZKTRACE1("Telling NNTP Server: %s\r\n", (char*)in_Context.GetProtocolOutput());
		break;
	default:
		break;
	}

}
