// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RZkPopProxy.cpp

#include "RZkPopProxy.h"
#include "RZkProxy.h"
#include <IZKErrorReporter.h>
#include <ZkErrorHandler.h>
#include <strstream>
#include <zksocket.h>
#include <string_hlp.h>
#include <INiqs.h>
#include <IClientInfo.h>
#include <zkAlloc.h>
#include <zkBase64.h>
#include <ZkService.h>
#include <INymCentral.h>
#include <frmailmsg.h>
#include <INym.h>
#include <IKeyQry.h>
#include <INymMail.h>
#include <WNetInfoDatum.h>
#include <IZkConsoleMsg.h>
#include <lockguard.h>
#include <IClientCentral.h>
#include <IZkRegistry.h>
#include <ZkProxyDialogs.h>
#include <IClientPersist.h>
#include <ConsMessages.h>
#include <zkCStr.h>
#include <ZkNetHelp.h>
#include <IZkPopup.h>
#include <IZkException.h>

#ifndef WIN32
#  include <sys/time.h>
#  include <netdb.h> // for hostent
#  include <RZkException.h>
#  include <../linux/RShimInterface.h>     // for VerifySocketConnection
#endif

static const char* DETAIL_KEY_NOT_FOUND			= "Key not found in query";
static const char* DETAIL_BASE64_DECODE_FAILED	= "Base64 decode failed";
static const char* DETAIL_NO_AUTH_TOKEN			= "Could not find authentication token";
static const char* DETAIL_NONE_AVAILABLE		= "No details available";
static const char* DETAIL_INVALID_SIG			= "Invalid signature";
static const char* DETAIL_DECRYPTION_FAILED		= "Decryption failed";
static const char* DETAIL_NOT_FROM_SIGNER		= "From field does not match signature";

static const char* USER_CMD_FORMAT_STRING					= "USER %s\r\n";
static const char* PASS_CMD_FORMAT_STRING					= "PASS_%s\r\n";
static const char* APOP_CMD_FORMAT_STRING					= "APOP %s %s\r\n";

static const char* UIDL_CMD_STRING							= "UIDL\r\n";
static const char* LIST_CMD_STRING							= "LIST\r\n";
static const char* RSET_CMD_STRING							= "RSET\r\n";
static const char* NOOP_CMD_STRING							= "NOOP\r\n";
static const char* STAT_CMD_STRING							= "STAT\r\n";

static const char* LIST_NUM_FORMAT_STRING					= "LIST %d\r\n";
static const char* RETR_NUM_FORMAT_STRING					= "RETR %d\r\n";
static const char* TOP_NUM_FORMAT_STRING					= "TOP %d %s\r\n";
static const char* DELE_NUM_FORMAT_STRING					= "DELE %d\r\n";
static const char* UIDL_NUM_FORMAT_STRING					= "UIDL %d\r\n";

static const char* QUIT_CMD_STRING							= "QUIT\r\n";

static const char* USER_CANCELLED_ERROR					= "-ERR User cancelled the operation\r\n";
static const char* TOP_UNSUPPORTED_ERROR				= "-ERR Freedom Does Not Support the TOP Command\r\n";
static const char* INVALID_SEQUENCE_ERROR				= "-ERR Invalid Sequence\r\n";
static const char* MESSAGE_DELETED_ERROR				= "-ERR Message Deleted\r\n";
static const char* MESSAGE_OUT_OF_RANGE_ERROR			= "-ERR Message Out Of Range\r\n";
static const char* TOO_MANY_PARAMETERS_ERROR			= "-ERR Too Many Parameters\r\n";
static const char* NOT_ENOUGH_PARAMETERS_ERROR			= "-ERR Not Enough Parameters\r\n";
static const char* UNKNOWN_COMMAND_ERROR				= "-ERR Command Unknown\r\n";
static const char* COULDNT_CONNECT_ERROR				= "-ERR Couldn't Connect To Remote Server\r\n";
static const char* SIMPLE_ERROR_STRING					= "-ERR Unknown Freedom Error\r\n";
static const char* ERROR_DETAIL_FORMAT_STRING			= "-ERR Freedom Error (%s)\r\n";
static const char* ERROR_DETAIL_AND_FUNC_FORMAT_STRING	= "-ERR Freedom Error(%s in %s)\r\n";

static const char* APOP_ACCEPTED_RESPONSE		="+OK Authenticated. Continue.\r\n";
static const char* USER_ACCEPTED_RESPONSE		="+OK Accepted. Send Password.\r\n";
static const char* PASS_ACCEPTED_RESPONSE		="+OK Authenticated. Continue.\r\n";
static const char* WELCOME_BANNER_RESPONSE		="+OK Freedom POP Server Ready.\r\n";
static const char* BOGUS_TOP_RESPONSE			="+OK Beginning Of Message Follows.\r\n" \
												"Subject: Freedom Encrypted Mail\r\n" \
												"From: Freedom User\r\n" \
												"To: Freedom User\r\n\r\n" \
												"This is a Freedom encrypted message. To see " \
												"the actual contents of this message,\r\nyou must " \
												"download the entire message.\r\n.\r\n";

WZkMutex RZkPopProxy::m_gMutex;
bool RZkPopProxy::m_gbNymAccountsLocked = false;
#define DONT_SEND_TO_SERVER -1
#define SEND_TO_ALL_SERVERS -2
#define USERS_POP_SERVER	0

RZkPopProxy::RZkPopProxy(ZkSocket in_LocalSocket)
		: m_LocalSocket(in_LocalSocket)
		, m_TotalPopDropSize(0)
		, m_NumAccounts(0)
		, m_iServerToContact(DONT_SEND_TO_SERVER)
		, m_iCommand(0)
		, m_ByteBuffer(NULL)
		, m_ByteBufferSize(0)
		, m_bCheckingUserAccount(false)
		, m_bHaveLockOnNymAccounts(false)
		, m_TotalNumMessages(0)
		, m_BytesReceived(0)
		, m_bUserAccountSupportsUidl(true)
		, m_bShowVerboseConsoleMsg(false)
        , m_listSockets(NULL) // XXX
{
	ZKVERIFY(zkMemAlloc(&m_err, (void**)&m_ByteBuffer, DEFAULT_POP_BUFFER_SIZE));
	m_ByteBufferSize = DEFAULT_POP_BUFFER_SIZE;
	DZkError derr;

    if( !zkListNew(&derr, &m_listSockets) )
	{
		ZKASSERT("Couldn't allocate list" && 0);
	}
	else if(!zkListSetDataFreeFunction(&derr,
								m_listSockets, NULL, NULL))
	{
		ZKASSERT("Couldn't set list free function" && 0);
	}
	else if(!zkListInsert(&derr, m_listSockets,
								m_LocalSocket, LIST_TAIL))
	{
		ZKASSERT("Couldn't insert socket into list" && 0);
	}

	if ( IZkRegistry::GetFlag(CONSOLE_VERBOSE_REGISTRY_KEY) ) 
	{
		m_bShowVerboseConsoleMsg = true ;
	}

	memset(&m_RemoteAddress, 0, sizeof(ZkSocketAddress));
}

RZkPopProxy::~RZkPopProxy()
{
    DZkError derr;

	if(m_bHaveLockOnNymAccounts )
	{
		m_gbNymAccountsLocked = false;
	}

    ZKVERIFY( zkListFree(&derr, &m_listSockets) );
	SAFEDELETEARRAY(m_ByteBuffer);
}

bool RZkPopProxy::RunProxy()
{
	struct timeval timeout = {0};
	timeout.tv_sec = 45;

	bool ReturnValue = true;
	try
	{
		if( !ReceiveRemoteAddress() )
		{
			ReturnValue = false;
		}
		else if(!BuildAccountList())
		{
			// User may be allowed to cancel at this time.
		}
		else if(!ConnectToRemoteServers())
		{
			ZKVERIFY(ErrorResponse());
		}
		// Will interact with MUA to get APOP/USER-PASS information
		else if(!AuthenticateWithRemoteServers())
		{// Handles returning errors internally
		}
		// Will send a STAT command to all accounts, amalgamate response
		else if(!BuildMessageMap())
		{
			ZKVERIFY(ErrorResponse());
		}
		else
		{
			while(ReturnValue && m_iCommand != QUIT_CMD)
			{
				if( !ReceiveRequestFromMUA() )
				{
					break;
				}
				
				// Determine which server(s) should receive the request
				else if( !ParseRequest() )
				{
					ZKVERIFY(ErrorResponse());
				}
				
				else
				{
					if(m_iCommand == RETR_NUM_CMD)
					{
						if( m_vecPopAccounts[m_iServerToContact].bUserAccount )
						{
							// It's the user
							IZkConsoleMsg::ShowImportantMsg(	CONSIMAGE_MAIL,
																IDS_CONS_PROXY_POP_USER_DOWNLOAD,
																m_vecPopAccounts[m_iServerToContact].strUserName.c_str(),
																atoi(m_strParamA.c_str()),
																m_TotalNumMessages
																) ;
						}
						else
						{
							int iMsgNum = atoi(m_strParamA.c_str());
							// It's a nym
							if( m_bShowVerboseConsoleMsg ) {
								// bug 8042: With "verbose console" on, we'll get better mail status info.
								IZkConsoleMsg::ShowMsg(
									CONS_VERBOSE,
									CONSIMAGE_MAIL,
									IDS_CONS_PROXY_POP_NYM_DOWNLOAD_START,
									m_vecPopAccounts[m_iServerToContact].strUserName.c_str(),
									iMsgNum,
									m_TotalNumMessages,
									m_vecMsgInfo[iMsgNum].SizeInBytes
									) ;
							}
							else
							{
								// this is for everyday users
								IZkConsoleMsg::ShowImportantMsg(	
									CONSIMAGE_MAIL,
									IDS_CONS_PROXY_POP_NYM_DOWNLOAD,
									m_vecPopAccounts[m_iServerToContact].strUserName.c_str(),
									iMsgNum,
									m_TotalNumMessages
									) ;
							}
						}
					}

					if(m_iCommand == QUIT_CMD)
					{
						// See bug 5954
						DeleteAllRetrievedNymMail();
						zkCStrlcpy(m_ByteBuffer, "QUIT\r\n", m_ByteBufferSize);
						m_iServerToContact = SEND_TO_ALL_SERVERS;
					}

					if( m_iServerToContact == DONT_SEND_TO_SERVER )
					{

					}
					else if( m_iServerToContact == SEND_TO_ALL_SERVERS)
					{
						if(!SendRequestToAllServers())
						{
							break;
						}
					
						if( !ReceiveAndProcessResponseFromAllServers() )
						{
							break;
						}
					}
					else
					{
						ZKASSERT((unsigned int)m_iServerToContact < m_vecPopAccounts.size());
						if(!SendRequestToRemoteServer(m_vecPopAccounts[m_iServerToContact]))
						{
							break;
						}
					
						if( !ReceiveAndProcessResponseFromOneServer(m_vecPopAccounts[m_iServerToContact]) )
						{
							break;
						}
					}
					
					if( !SendResponseToMUA() )
					{
						// bug 8322: trying to track mail download failures better...
						IZkConsoleMsg::ShowImportantMsg(	
							CONSIMAGE_MAIL,
							IDS_CONS_PROXY_POP_SEND_TO_MUA_FAILED,
							m_err.getdetail().c_str()
							) ;						
						break;
					}
				}
			}  // while
		}
	}
	catch(...)
	{
		CleanUpSockets();
		throw;
	}

	CleanUpSockets();	

	return ReturnValue;
}

void RZkPopProxy::CleanUpSockets()
{
	size_t i;
	for(i = 0; i < m_vecPopAccounts.size(); i++)
	{
		if(m_vecPopAccounts[i].RemoteSocket)
		{
			ZKVERIFY(zkSocketFree(&m_err, &m_vecPopAccounts[i].RemoteSocket));
			m_vecPopAccounts[i].RemoteSocket = NULL;
		}
	}	
}

bool RZkPopProxy::DeleteAllRetrievedNymMail()
{	// See bug 5954
	bool bShowWarning = !ZkService::GetIClientCentral()->GetPreference(ZK_SHOW_AGAIN_PREF_SET, PREF_NO_SHOW_MAIL_DELETION_WARNING);
	bool ReturnValue = true;

	for(size_t i = 1; i < m_vecMsgInfo.size() && ReturnValue; i++)
	{
		if(m_vecMsgInfo[i].bFreedomMsg && !m_vecMsgInfo[i].bDeleted &&
											m_vecMsgInfo[i].bRetrieved)
		{

			if (bShowWarning && IZkPopup::MessageCheckBox(
							IDS_MAIL_DELETION_WARNING))
			{
				ZkService::GetIClientCentral()->SetPreference(
					ZK_SHOW_AGAIN_PREF_SET, 
					PREF_NO_SHOW_MAIL_DELETION_WARNING,
					true);
			}
			bShowWarning = false;

			snprintf(m_ByteBuffer, m_ByteBufferSize, DELE_NUM_FORMAT_STRING, m_vecMsgInfo[i].MsgNumber);
			m_iServerToContact = m_vecMsgInfo[i].PopAccountIndex;
			if(!SendRequestToRemoteServer(m_vecPopAccounts[m_iServerToContact]))
			{
				ReturnValue = false;
			}
			else if(!ReceiveResponseFromRemoteServer(m_vecPopAccounts[m_iServerToContact]))
			{
				ReturnValue = false;
			}
			else if(!ResponseIsSuccess())
			{
				ReturnValue = false;
			}
			else
			{
				m_vecMsgInfo[i].bDeleted = true;
			}

		}
	}
	return ReturnValue;
}

bool RZkPopProxy::ResponseIsSuccess()
{
	
	if(m_ByteBufferSize && m_ByteBuffer[0] == '+')
	{
		return true;
	}
	return false;
}

bool RZkPopProxy::ErrorResponse()
{
	bool ReturnValue = true;

	unsigned int lenDetail = strlen( m_err.getdetail().c_str() );
	unsigned int lenFunc = strlen( m_err.getfnc().c_str() );

	if ( lenDetail > 0 )
	{
		if(!strncmp(m_err.getdetail().c_str(), "-ERR", 4))
		{   
			strncpy(m_ByteBuffer, m_err.getdetail().c_str(), m_ByteBufferSize);
		}
		else if ( lenFunc > 0 )
		{
			// Function name is a bit ugly to show to the user, we only show
			// this if we have one from the unix guys
			snprintf(m_ByteBuffer, m_ByteBufferSize, ERROR_DETAIL_AND_FUNC_FORMAT_STRING,
												m_err.getdetail().c_str(), m_err.getfnc().c_str() );
		}
		else
		{
			
			snprintf(m_ByteBuffer, m_ByteBufferSize, ERROR_DETAIL_FORMAT_STRING,
												m_err.getdetail().c_str());
		}
	}
	else
	{
		zkCStrlcpy(m_ByteBuffer, SIMPLE_ERROR_STRING, m_ByteBufferSize);
	}
	
	// bug 8042... tell the console (it'll only show in debug mode)
	ZKSTD::string strResponse = "[No response]";
	if(m_ByteBufferSize > 2)
	{
		strResponse = ZKSTD::string(m_ByteBuffer, strlen(m_ByteBuffer) - 2);
	}
	IZkConsoleMsg::ShowMsg( CONS_DEBUG, CONSIMAGE_MAIL,
					IDS_CONS_PROXY_POP_ERROR_RESPONSE, 
					strResponse.c_str());

	if(!zkSocketWriteAll(&m_err, m_LocalSocket,
			m_ByteBuffer, strlen(m_ByteBuffer) ))
	{
		ReturnValue = false;
	}

	return ReturnValue;
}

bool RZkPopProxy::SendResponseToMUA()
{
	ZKSTD::string strResponse;

	bool ReturnValue = true;
	int i;

	switch(m_iCommand)
	{
		
	case USER_CMD:
	case PASS_CMD:
	case APOP_CMD:
		zkCStrlcpy(m_ByteBuffer, INVALID_SEQUENCE_ERROR, m_ByteBufferSize);
		break;

	case TOP_NUM_CMD:
		// The response was already copied in by ParseRequest();
		break;

	case STAT_CMD:
		// If messages have been deleted, the total num and total size should decrease
		snprintf(m_ByteBuffer, m_ByteBufferSize, "+OK %d %d\r\n", m_TotalNumMessages, m_TotalPopDropSize);
		break;

	case UIDL_CMD:
		strResponse = "+OK\r\n";
		for(i = 1; i <= m_TotalNumMessages; i++)
		{
			if(!m_vecMsgInfo[i].bDeleted)
			{
				char buf[10] = {0};
				snprintf(buf, sizeof(buf), "%d", i);
				strResponse += buf;
				strResponse += ' ';
				strResponse += m_vecMsgInfo[i].strUIDL.c_str();
				strResponse += "\r\n";
			}
		}
		strResponse += ".\r\n";
		break;

	case LIST_CMD:
		strResponse = "+OK\r\n";
		for(i = 1; i <= m_TotalNumMessages; i++)
		{
			if(!m_vecMsgInfo[i].bDeleted)
			{
				ZKTRACE2("Listing message %d, size %d\n", i, m_vecMsgInfo[i].SizeInBytes);
				char buf[10] = {0};
				snprintf(buf, sizeof(buf),  "%d", i);
				strResponse += buf;
				strResponse += ' ';
				snprintf(buf, sizeof(buf), "%d", m_vecMsgInfo[i].SizeInBytes);
				strResponse += buf;
				strResponse += "\r\n";
			}
		}
		strResponse += ".\r\n";
		break;

	case RETR_NUM_CMD:
	case DELE_NUM_CMD:
		// If it was Freedom mail, it's already decrypted and in m_ByteBuffer
		// Pass through whatever the server gave us.
		break;

	case LIST_NUM_CMD:
		{
			size_t iMsgNum = atoi(m_strParamA.c_str());
			if( iMsgNum < m_vecMsgInfo.size() && iMsgNum > 0)
			{
				if(m_vecMsgInfo[iMsgNum].bDeleted)
				{
					zkCStrlcpy(m_ByteBuffer, MESSAGE_DELETED_ERROR, m_ByteBufferSize);
				}
				else
				{
					snprintf(m_ByteBuffer, m_ByteBufferSize, "+OK %d %d\r\n", iMsgNum, m_vecMsgInfo[iMsgNum].SizeInBytes);
				}
			}
			else
			{
				zkCStrlcpy(m_ByteBuffer, MESSAGE_OUT_OF_RANGE_ERROR, m_ByteBufferSize);
			}
		}
		break;

	case UIDL_NUM_CMD:
		{
			size_t iMsgNum = atoi(m_strParamA.c_str());
			if( iMsgNum < m_vecMsgInfo.size() && iMsgNum > 0)
			{
				if(m_vecMsgInfo[iMsgNum].bDeleted)
				{
					zkCStrlcpy(m_ByteBuffer, MESSAGE_DELETED_ERROR, m_ByteBufferSize);
				}
				else
				{
					snprintf(m_ByteBuffer, m_ByteBufferSize, "+OK %d %s\r\n", iMsgNum, m_vecMsgInfo[iMsgNum].strUIDL.c_str());
				}
			}
			else
			{
				zkCStrlcpy(m_ByteBuffer, MESSAGE_OUT_OF_RANGE_ERROR, m_ByteBufferSize);
			}
		}
		break;

	case RSET_CMD:
	case QUIT_CMD:
	case NOOP_CMD:
		// ReceiveAndProcessResponseFromAllServers has put the appropriate response in the buffer.
		break;

	// case OTHER_CMD?

	default:
		ZKASSERT(0);
		ReturnValue = false;
		break;
	}

	if(!ReturnValue)
	{
		ErrorResponse();
	}
	else if(m_iCommand == RETR_NUM_CMD && m_vecMsgInfo[atoi(m_strParamA.c_str())].bFreedomMsg)
	{	// Special case - add a +OK\r\n in front

		if(!zkSocketWriteAll(&m_err, m_LocalSocket,
				"+OK\r\n", 5 ))
		{
			ReturnValue = false;
		}
		// m_ByteBuffer is null terminated when we decrypt it
		// in ProcessFreedomMail()
		else if(!zkSocketWriteAll(&m_err, m_LocalSocket,
				m_ByteBuffer, strlen(m_ByteBuffer) ))
		{
			ReturnValue = false;
		}

		ZKTRACE1("Said to Client: OK\r\n%s\r\n", strResponse.c_str());
	}
	else if(strResponse.length())
	{
		// Send from the std::string

		if(!zkSocketWriteAll(&m_err, m_LocalSocket,
				strResponse.c_str(), strResponse.length() ))
		{
			ReturnValue = false;
		}

		ZKTRACE1("Said to Client: %s\r\n", strResponse.c_str());
	}
	else
	{
		if(!zkSocketWriteAll(&m_err, m_LocalSocket,
				m_ByteBuffer, strlen(m_ByteBuffer) ))
		{
			ReturnValue = false;
		}
		ZKTRACE1("Said to Client: %s\r\n", m_ByteBuffer);
	}

	return ReturnValue;
}

bool RZkPopProxy::PassResponseThroughToMUA()
{
	if(!zkSocketWriteAll(&m_err, m_LocalSocket,
			m_ByteBuffer, m_BytesReceived ))
	{
		return false;
	}
	return true;
}

bool RZkPopProxy::SendRequestToRemoteServer(PopAccount& in_PopAccount)
{
	if(!zkSocketWriteAll(&m_err, in_PopAccount.RemoteSocket,
			m_ByteBuffer, strlen(m_ByteBuffer) ))
	{
		in_PopAccount.strLastCommand = "[Send Failed]";
		return false;
	}
	// Crop off the \r\n
	in_PopAccount.strLastCommand = ZKSTD::string(m_ByteBuffer, strlen(m_ByteBuffer) - 2);
	ZKTRACE1("Said to Server: %s\r\n", m_ByteBuffer);
	return true;
}

bool RZkPopProxy::ReceiveAndProcessResponseFromAllServers()
{
	
	bool ReturnValue = true;
	size_t i;

	switch(m_iCommand)
	{
	// We create these responses without contacting anyone.
	case STAT_CMD:
	case UIDL_CMD:
	case LIST_CMD:
		ZKASSERT(0);
		break;

	// These only go to one server.
	case LIST_NUM_CMD:
	case TOP_NUM_CMD:
	case USER_CMD:
	case PASS_CMD:
	case APOP_CMD:
	case RETR_NUM_CMD:
	case DELE_NUM_CMD:
	case UIDL_NUM_CMD:
		ZKASSERT(0);
		break;

	case RSET_CMD:		
	case QUIT_CMD:
	case NOOP_CMD:
		
		for(i = 0; i < m_vecPopAccounts.size(); i++)
		{
			if(!ReceiveAndProcessResponseFromOneServer(m_vecPopAccounts[i]))
			{
				ReturnValue = false;
				break;
			}
		}
		break;

	default:
		ZKASSERT(0);
		break;
	}

	return ReturnValue;
}

bool RZkPopProxy::ReceiveAndProcessResponseFromOneServer(PopAccount& in_PopAccount)
{
	bool ReturnValue = true;

	switch(m_iCommand)
	{
	case USER_CMD:
	case PASS_CMD:
	case APOP_CMD:
	case LIST_NUM_CMD:
	case UIDL_NUM_CMD:
		ZKASSERT(0);
		break;

	case TOP_NUM_CMD:
		ZKASSERT( in_PopAccount.bUserAccount );
		if(!ReceiveResponseFromRemoteServer(in_PopAccount))
		{
			ZKASSERT(0);
			// not so descriptive, but what else to say?
			m_err.setdetail(COULDNT_CONNECT_ERROR);		
			ReturnValue = false;
		}
		// We're going to pass the user server's response through...
		break;

	case RETR_NUM_CMD:
		if(!ReceiveResponseFromRemoteServer(in_PopAccount))
		{
			ErrorResponse();
			ReturnValue = false;
		}
		else if(ResponseIsSuccess())
		{
			m_vecMsgInfo[atoi(m_strParamA.c_str())].bRetrieved = true;

			if(IsFreedomMail())
			{
				// bug 8042... "Message X has been downloaded; now decrypting..."
				IZkConsoleMsg::ShowMsg( 
					CONS_VERBOSE, 
					CONSIMAGE_MAIL,
					IDS_CONS_PROXY_DECRYPTING_MSG,
					atoi(m_strParamA.c_str()),
					in_PopAccount.strUserName.c_str() 
					);

				bool bKeyQryFailed = false;
				// bug 8042... we can report all sorts of encryption errors through this string
				ZKSTD::string strFailureDetail;

				if(ProcessFreedomMail(in_PopAccount.strUserName, bKeyQryFailed, strFailureDetail))
				{
					m_vecMsgInfo[atoi(m_strParamA.c_str())].bFreedomMsg = true;
					// bug 8042... "Message X has been decrypted."
					IZkConsoleMsg::ShowMsg( 
						CONS_VERBOSE,
						CONSIMAGE_MAIL,
						IDS_CONS_PROXY_POP_DECRYPT_SUCCESS,
						atoi(m_strParamA.c_str()),
						in_PopAccount.strUserName.c_str()
						);
				}
				else 
				{
					INym *pNym;
					ZkService::GetINymCentral()->GetNym(IID_INYM, (void**)&pNym,
														in_PopAccount.strUserName);
					ZKSTD::string strEmailDomain;
					pNym->GetNymEmailDomain(strEmailDomain);
					// bug 8042... report failure to console (note "important")
					if( !strFailureDetail.empty() ) 
					{
						IZkConsoleMsg::ShowImportantMsg( CONSIMAGE_MAIL,
											IDS_CONS_POP_DECRYPT_FAILED,
											in_PopAccount.strUserName.c_str(),
											strFailureDetail.c_str() ) ;
					}
					else
					{
						IZkConsoleMsg::ShowImportantMsg( CONSIMAGE_MAIL,
											IDS_CONS_POP_DECRYPT_FAILED_NO_DETAIL,
											in_PopAccount.strUserName.c_str()
											) ;
						
					}
						
					if(bKeyQryFailed)
					{
						ZKSTD::string strMessage = "+ERR Failed to query public "
													"signature key for the sending nym ("
													+ m_err.getdetail() + "). "
													"Try again later.\r\n";
						zkCStrlcpy(m_ByteBuffer, strMessage.c_str(), m_ByteBufferSize);
					}
					else
					{
						ZKSTD::string strMessage =	"+OK Decrypt Failed, Error Msg Follows\r\n"\
								"Subject: Message decryption failed\r\n" \
								"From: support@freedom.net\r\n" \
								"To: " + in_PopAccount.strUserName + "@" + strEmailDomain + "\r\n\r\n" \
								"This is an automatically generated message.\r\n\r\n" \
								"Freedom was unable to decrypt a message intended for " +
								in_PopAccount.strUserName + " because the message " \
								"was corrupted or damaged during delivery. Freedom has deleted the message from " \
								"your POP server.";
						if( !strFailureDetail.empty() ) { 
							strMessage += "\r\nError description: " + strFailureDetail;
						}
						strMessage += "\r\n\r\nIf you continue to experience problems, " \
									"exit Freedom and run the Freedom Diagnostic Utility.\r\n.\r\n";

						zkCStrlcpy(m_ByteBuffer, strMessage.c_str(), m_ByteBufferSize);
					}

				}
			}
		}

		break;

	case DELE_NUM_CMD:
		if(!ReceiveResponseFromRemoteServer(in_PopAccount))
		{
			ReturnValue = false;
		}
		else if(ResponseIsSuccess())
		{
			int iMsgNum = atoi(m_strParamA.c_str());
			snprintf(m_ByteBuffer, m_ByteBufferSize, "+OK Message %d Deleted\r\n", iMsgNum);
			m_vecMsgInfo[iMsgNum].bDeleted = true;
			// bug 8042... "Message X successfully deleted from server."
			IZkConsoleMsg::ShowMsg( 
				CONS_VERBOSE,
				CONSIMAGE_MAIL,
				IDS_CONS_PROXY_POP_DELETE_SUCCESS,
				iMsgNum,
				in_PopAccount.strUserName.c_str()
				);
		}
		else
		{
			// Pass through server error code
		}
		break;


	case UIDL_CMD:
	case LIST_CMD:
	case STAT_CMD:
		ZKASSERT(0);
		break;

	case RSET_CMD:
		if(!ReceiveResponseFromRemoteServer(in_PopAccount))
		{
			ReturnValue = false;
		}
		else if(ResponseIsSuccess())
		{
			size_t i;
			for(i = 1; i < in_PopAccount.vecRealToVirtualMsgNums.size(); i++)
			{
				size_t VirtualNum = in_PopAccount.vecRealToVirtualMsgNums[i];

				m_vecMsgInfo[VirtualNum].bDeleted = false;
			}
		}
		else
		{
			// What to do about the response code?
			// (Currently only return the most recent one)
		}
		break;

	case QUIT_CMD:
	case NOOP_CMD:
		ReceiveResponseFromRemoteServer(in_PopAccount);
		// What to do about the response codes?
		// (Currently only return the most recent one)
		break;

	default:
		ZKASSERT(0);
		break;
	}

	return ReturnValue;
}


bool RZkPopProxy::IsFreedomMail()
{

	// Find a line that begins with X-Freedom-Message-Auth
	char *pFreedomHeader = strstr(m_ByteBuffer, "\r\nX-Freedom-Message-Auth:");

	// Find two CRLF's in a row, indicating end of header
	char *pEndOfHeader = strstr(m_ByteBuffer, "\r\n\r\n");

	// If we found both of those things, and the X-Freedom-blah is before the
	// end of the header, this is a freedom message.

	if(pFreedomHeader && pEndOfHeader && pFreedomHeader < pEndOfHeader)
	{
		return true;
	}

	return false;
}

bool RZkPopProxy::GetAuthenticationToken(void **out_ppAuthToken,
											size_t *out_lenAuthToken)
{
	bool ReturnValue = true;

	char strFieldName[] = "\r\nX-Freedom-Message-Auth:";
	char *pBeginField = strstr(m_ByteBuffer, strFieldName);
	char *pEndField = NULL;
	if(!pBeginField)
	{
		ReturnValue = false;
	}
	else
	{
		pBeginField += strlen(strFieldName);

		while(*pBeginField == ' ')
		{
			pBeginField ++;
		}
		pEndField = strstr(pBeginField, "\r\n");
		while(pEndField && *(pEndField + 2) == ' ')
		{
			pEndField = strstr(pEndField + 2, "\r\n");
		}

		if(!pEndField)
		{
			ReturnValue = false;
		}
		else
		{
			*out_lenAuthToken = pEndField - pBeginField;			

			if(!zkMemAlloc(&m_err, out_ppAuthToken, *out_lenAuthToken))
			{
				ReturnValue = false;
			}
			else
			{
				*pEndField = 0; // Base64 routines require nul-termination
								// We know this was '\r', just remember to set it back
				if(!zkConvertFromBase64(&m_err, (const unsigned char*)pBeginField, (unsigned char*)*out_ppAuthToken,
											out_lenAuthToken))
				{
					ReturnValue = false;
				}

				*pEndField = '\r';
			}
		}
	}
	return ReturnValue;
}

bool RZkPopProxy::GetMimeBoundary(ZKSTD::string& out_strMimeBoundary)
{
	// This could potentially be more robust...

	bool ReturnValue = true;
	ZKSTD::string strContentType;
	if(!ZKS::extractHeaderField(m_ByteBuffer, strlen(m_ByteBuffer),
								"Content-Type", strContentType))
	{
		ReturnValue = false;
	}
	else
	{
		size_t i = strContentType.find("boundary=");
		if(!i)
		{
			ReturnValue = false;
		}
		else
		{
			out_strMimeBoundary = strContentType.substr(i + strlen("boundary="));
		}
	}

	return ReturnValue;
}

bool RZkPopProxy::GetEncryptedMsgBody(void **out_ppEncryptedBody,
										size_t *out_lenEncryptedBody)
{
	ZKSTD::string strMimeBoundary;

	bool ReturnValue = true;

	char *pBeginMsg = NULL;
	
	if(!GetMimeBoundary(strMimeBoundary))
	{
		ReturnValue = false;
	}
	else
	{
		// Mime boundary in mail is preceded and followed by '-'
		strMimeBoundary = "\r\n-" + strMimeBoundary + "-";
		char* pBeginSection = strstr(m_ByteBuffer, "\r\n\r\n");
		if(pBeginSection)
		{
			// so that the strstr will find the CR before the boundary
			pBeginSection -= 2;

			while((pBeginSection = strstr(pBeginSection, strMimeBoundary.c_str())))
			{
				ZKSTD::string strContentType;
				if(ZKS::extractHeaderField(pBeginSection, strlen(pBeginSection), "Content-Type", strContentType))
				{
					if(strContentType == "application/x-Freedom-Mail-Message-Body")
					{
						pBeginMsg = strstr(pBeginSection, "\r\n\r\n");
						if(pBeginMsg)
						{
							pBeginMsg += 4;
						}
						break;
					}
				}
				pBeginSection += 1; // So we won't find the same boundary
			}
		}
	}

	if(!pBeginMsg)
	{
		ReturnValue = false;
	}
	else
	{
		char* pEndMsg = strstr(pBeginMsg, strMimeBoundary.c_str());
		if(!pEndMsg)
		{
			ReturnValue = false;
		}
		else
		{
			*out_lenEncryptedBody = pEndMsg - pBeginMsg;

			if(!zkMemAlloc(&m_err, out_ppEncryptedBody, *out_lenEncryptedBody))
			{
				ReturnValue = false;
			}
			else
			{
				char temp = *pEndMsg;
				*pEndMsg = 0;// Base64 routines require nul-termination
							// record the previous value, just remember to set it back
				
				// Notice the address of out_lenEncryptedBody is sent in.
				if(!zkConvertFromBase64(&m_err, (const unsigned char*)pBeginMsg,
									(unsigned char*)*out_ppEncryptedBody, out_lenEncryptedBody))
				{
					ReturnValue = false;
				}
				*pEndMsg = temp;
			}
		}
	}

	return ReturnValue;
}

bool RZkPopProxy::ProcessFreedomMail(const ZKSTD::string& in_strNymName,
									bool& out_bKeyQryFailed,
									ZKSTD::string& out_strFailureDetail )
{
	out_bKeyQryFailed = false;
	out_strFailureDetail = "";
	bool ReturnValue = true;
	bool bDone = false;

	void *pDecodedAuthField = NULL;
	size_t lenAuthField = 0;
	void *pEncryptedBody;
	size_t lenEncryptedBody = 0;
	
	if(!GetAuthenticationToken(&pDecodedAuthField, &lenAuthField))
	{
		out_strFailureDetail = DETAIL_NO_AUTH_TOKEN;
		ReturnValue = false;
	}
	else if(!GetEncryptedMsgBody(&pEncryptedBody, &lenEncryptedBody))
	{
		out_strFailureDetail = DETAIL_BASE64_DECODE_FAILED;
		ReturnValue = false;
	}
	else
	{
		FrMailMsg MailMsg = NULL;

		FrEntIdExp sender;
		memset(&sender, 0, sizeof sender);

		FrCryptPubKey PubSigKey = NULL;
		FrCryptPubKey PrivCryptoKey = NULL;

		INym * pNym = NULL ;
		INymMail *pNymMail = NULL;

		ZkService::GetINymCentral()->GetNym( IID_INYM, (void**) &pNym, in_strNymName) ;
		pNym->GetCryptoKey(&PrivCryptoKey);
		pNym->QueryInterface(IID_IMAIL, (void**) &pNymMail);
		if(!frMailMsgNew(&m_err, &MailMsg))
		{
			ReturnValue = false;
		}
		else if(!frMailMsgSetFromHeader(&m_err, MailMsg, PrivCryptoKey,
										pDecodedAuthField, lenAuthField))
		{
			ReturnValue = false;
		}
		else if(!frMailMsgGetSender(&m_err, MailMsg, &sender))
		{
			ReturnValue = false;
		}
		else
		{
			int sigvalid = 0;
			int hashvalid = 0;
			char* pDecryptedMsg = 0;
			size_t lenDecryptedMsg = 0;

			if(sender.type == FRENT_TYPE_IMEP)
			{
				FrCryptMacKey MacKey = pNymMail->GetMailMacKey();
				if(!MacKey)
				{
					ZKASSERT(0);
					ReturnValue = false;
				}
				else if(!frMailMsgMacVerify(&m_err, MailMsg, MacKey, &sigvalid))
				{
					sigvalid = 0;
					ZKASSERT(0);
					ReturnValue = false;
				}
			}
			else
			{
				if(!ZkService::GetIKeyQry()->GetKey( m_err,
									 sender.type,	// entity, 
									 3,				// timeout
									 sender.name,
									 0,
									 FRCRYPT_KEYTYPE_SIGNING, 
									 &PubSigKey))
				{
					if(zkErrorCode(&m_err) == ZKENOTFOUND)
					{	// bug 8042... report this clearly
						out_strFailureDetail = DETAIL_KEY_NOT_FOUND;
					}
					else
					{	// failed for some reason besides
						// not finding the key.
						// This makes us return -ERR instead of bogus msg
						out_bKeyQryFailed = true;
					}
					ReturnValue = false;					
				}
				else if(!frMailMsgSigVerify(&m_err, MailMsg, PubSigKey, &sigvalid))
				{
					sigvalid = 0;
					ReturnValue = false;
				}
			}

			if(!sigvalid)
			{
				out_strFailureDetail = DETAIL_INVALID_SIG;
				ReturnValue = false;
			}
			else if(!frMailMsgDecrypt(&m_err, MailMsg,
									pEncryptedBody, lenEncryptedBody,
									(void**)&pDecryptedMsg, &lenDecryptedMsg))
			{
				out_strFailureDetail = DETAIL_DECRYPTION_FAILED;
				ReturnValue = false;
			}
			else if(!frMailMsgHashVerify(&m_err, MailMsg, pDecryptedMsg,
											lenDecryptedMsg, &hashvalid))
			{
				ReturnValue = false;
			}
			else
			{
				// Need to add SMTP term. and NULL term.
				ZKVERIFY(zkMemRealloc(&m_err, (void **)&pDecryptedMsg, lenDecryptedMsg + 6));
				zkCStrlcpy(pDecryptedMsg + lenDecryptedMsg, "\r\n.\r\n", 6);

				// Only use this function for hiding headers from client side
				// software! (It doesn't actually remove the headers, just
				// changes the field name to "XXXXXXXXXX...")
				ZKS::hideHeaderField(pDecryptedMsg, lenDecryptedMsg,
								"Disposition-Notification-To");

				ZKSTD::string strFromField;

				if(sender.type != FRENT_TYPE_IMEP &&
					(ZKS::extractHeaderField(pDecryptedMsg, lenDecryptedMsg,
														"From", strFromField) ||
					ZKS::extractHeaderField(pDecryptedMsg, lenDecryptedMsg,
														"Sender", strFromField)) )
				{
					ZKSTD::string strSigningNym(sender.name);
					ZKSTD::string strFreedomDomain;
					ZkService::GetIClientInfo()->GetFreedomDomain(strFreedomDomain);

					ZKS::MakeLower(strFromField);
					ZKS::MakeLower(strSigningNym);

					size_t pos = strFromField.find('@');
					size_t pos2 = strFromField.find(strFreedomDomain);

					if( pos == ZKSTD::string::npos ||
						 pos2 == ZKSTD::string::npos ||
						 (strFromField[pos2 - 1] != '@' &&
							strFromField[pos2 -1] != '.') )
					{
						// This isn't nym to nym mail - we have no way of validating
					}
					else
					{			
						std::string strFromName(strFromField.substr(0, pos));
						if(strFromName != strSigningNym)
						{
							out_strFailureDetail = DETAIL_NOT_FROM_SIGNER;
							ReturnValue = false;
						}
					}
				}
				if(ReturnValue)
				{
					zkMemFree(&m_err, (void**)&m_ByteBuffer);
					m_ByteBuffer = pDecryptedMsg;
					m_ByteBufferSize = lenDecryptedMsg + 6;
					m_BytesReceived = lenDecryptedMsg + 6; // null not included
					bDone = true;
				}
				else
				{
					zkMemFree(&m_err, (void**)&pDecryptedMsg);					
				}
			}


			if(PrivCryptoKey)
			{
				ZKVERIFY(frCryptPubKeyFree(&m_err, &PrivCryptoKey));
			}

			if(MailMsg)
			{
				ZKVERIFY(frMailMsgFree(&m_err, &MailMsg));
			}

		}
	}
	return ReturnValue;
}

bool RZkPopProxy::SendRequestToAllServers()
{
	bool ReturnValue = true;
	size_t i;
	for(i = 0; i < m_vecPopAccounts.size(); i++)
	{
		if(!SendRequestToRemoteServer(m_vecPopAccounts[i]))
		{
			ReturnValue = false;
			break;
		}
	}
	return ReturnValue;
}

bool RZkPopProxy::ParseRequest()
{
	bool ReturnValue = true;
	ZKVERIFY(zkErrorClear(&m_err));

	// Make sure we have a null terminated string
	ZKASSERT( ((char*)m_ByteBuffer)[m_BytesReceived] == 0 );
	
	ZKSTD::string strCommand((char *)m_ByteBuffer);

	m_iCommand = OTHER_CMD;
	m_iServerToContact = DONT_SEND_TO_SERVER;
	size_t i = strCommand.find_first_of(' ');

	m_strParamA = "";
	m_strParamB = "";

	if(i != strCommand.npos)
	{
		i = strCommand.find_first_not_of(' ', i);
		if(i != strCommand.npos)
		{
			size_t j = strCommand.find_first_of(" \r\n", i);
			ZKASSERT(j != strCommand.npos);

			m_strParamA = strCommand.substr(i, j - i);

			i = strCommand.find_first_not_of(" \r\n", j);
			if(i != strCommand.npos)
			{
				j = strCommand.find_first_of(" \r\n", i);
				ZKASSERT(j != strCommand.npos);
				m_strParamB = strCommand.substr(i, j -i);
			}
		}


	}

	ZKS::MakeLower(strCommand);

	if( !strncmp(strCommand.c_str(), "user", 4) )
	{
		if(m_strParamB.length())
		{
			m_err.setdetail(TOO_MANY_PARAMETERS_ERROR);
			ReturnValue = false;
		}
		else if(m_strParamA.length())
		{
			m_iCommand = USER_CMD;
			if(m_bCheckingUserAccount)
			{
				m_iServerToContact = USERS_POP_SERVER;
				m_vecPopAccounts[0].strUserName = m_strParamA;
			}
			else
			{
				m_iServerToContact = DONT_SEND_TO_SERVER;

			}
		}
		else
		{
			m_err.setdetail(NOT_ENOUGH_PARAMETERS_ERROR);
			ReturnValue = false;
		}
	}
	else if( !strncmp(strCommand.c_str(), "pass", 4) )
	{
		if(m_strParamB.length())
		{
			m_err.setdetail(TOO_MANY_PARAMETERS_ERROR);
			ReturnValue = false;
		}
		else if(m_strParamA.length())
		{
			m_iCommand = PASS_CMD;
			if(m_bCheckingUserAccount)
			{
				m_vecPopAccounts[0].strPassword = m_strParamA;
				m_iServerToContact = USERS_POP_SERVER;
			}
			else
			{
				m_iServerToContact = DONT_SEND_TO_SERVER;
			}
		}
		else
		{
			ReturnValue = false;
		}
	}
	else if( !strncmp(strCommand.c_str(), "uidl", 4) )
	{
		// Here's the fix for UIDL: we know if the user's server
		//	can handle it or not by this point -- if not, we
		//	pretend we don't know the command.
		if( !m_bUserAccountSupportsUidl )
		{
			m_err.setdetail(UNKNOWN_COMMAND_ERROR);
			m_iCommand = OTHER_CMD;
			m_iServerToContact = DONT_SEND_TO_SERVER;
			ReturnValue = false;
		}
		else if(m_strParamB.length())
		{
			m_err.setdetail(TOO_MANY_PARAMETERS_ERROR);
			ReturnValue = false;
		}
		else if(m_strParamA.length())
		{
			m_iCommand = UIDL_NUM_CMD;
			// Already have this data
			m_iServerToContact = DONT_SEND_TO_SERVER;
		}
		else
		{
			m_iCommand = UIDL_CMD;
			// Already have this data
			m_iServerToContact = DONT_SEND_TO_SERVER;
		}
	}
	else if( !strncmp(strCommand.c_str(), "list", 4) )
	{
		if(m_strParamB.length())
		{
			m_err.setdetail(TOO_MANY_PARAMETERS_ERROR);
			ReturnValue = false;
		}
		else if(m_strParamA.length())
		{
			m_iCommand = LIST_NUM_CMD;
			// Already have this data
			m_iServerToContact = DONT_SEND_TO_SERVER;
		}
		else
		{
			m_iCommand = LIST_CMD;
			// Already have this data
			m_iServerToContact = DONT_SEND_TO_SERVER;
		}
	}
	else if( !strncmp(strCommand.c_str(), "rset", 4) )
	{
		if(m_strParamA.length())
		{
			m_err.setdetail(TOO_MANY_PARAMETERS_ERROR);
			ReturnValue = false;
		}
		else
		{
			m_iCommand = RSET_CMD;
			m_iServerToContact = SEND_TO_ALL_SERVERS;
		}
	}
	else if( !strncmp(strCommand.c_str(), "noop", 4) )
	{
		if(m_strParamA.length())
		{
			m_err.setdetail(TOO_MANY_PARAMETERS_ERROR);
			ReturnValue = false;
		}
		else
		{
			m_iCommand = NOOP_CMD;
			m_iServerToContact = SEND_TO_ALL_SERVERS;
		}
	}
	else if( !strncmp(strCommand.c_str(), "stat", 4) )
	{
		if(m_strParamA.length())
		{
			m_err.setdetail(TOO_MANY_PARAMETERS_ERROR);
			ReturnValue = false;
		}
		else
		{
			m_iCommand = STAT_CMD;
			// Already have this data
			m_iServerToContact = DONT_SEND_TO_SERVER;
		}
	}
	else if( !strncmp(strCommand.c_str(), "retr", 4) )
	{
		if(m_strParamB.length())
		{
			m_err.setdetail(TOO_MANY_PARAMETERS_ERROR);
			ReturnValue = false;
		}
		else if(m_strParamA.length())
		{
			m_iCommand = RETR_NUM_CMD;
			ReturnValue = RenumberCommandAndPickServer();
		}
		else
		{
			m_err.setdetail(NOT_ENOUGH_PARAMETERS_ERROR);
			ReturnValue = false;
		}
	}
	else if( !strncmp(strCommand.c_str(), "dele", 4) )
	{
		if(m_strParamB.length())
		{
			m_err.setdetail(TOO_MANY_PARAMETERS_ERROR);
			ReturnValue = false;
		}
		else if(m_strParamA.length())
		{
			m_iCommand = DELE_NUM_CMD;
			ReturnValue = RenumberCommandAndPickServer();
		}
		else
		{
			m_err.setdetail(NOT_ENOUGH_PARAMETERS_ERROR);
			ReturnValue = false;
		}
	}
	else if( !strncmp(strCommand.c_str(), "top", 3) )
	{

		if(m_strParamB.length())
		{
			// There were two parameters, as expected
			m_iCommand = TOP_NUM_CMD;
			ReturnValue = RenumberCommandAndPickServer();
			if(!m_vecPopAccounts[m_iServerToContact].bUserAccount)
			{
				// Don't send TOP commands to Freedom servers
				// (the whole body is on one line, so the whole message will
				// be retrieved)
				m_iServerToContact = DONT_SEND_TO_SERVER;
				ZKASSERT(m_ByteBufferSize > strlen(BOGUS_TOP_RESPONSE));
				strncpy(m_ByteBuffer, BOGUS_TOP_RESPONSE, m_ByteBufferSize);
				m_ByteBuffer[m_ByteBufferSize - 1] = 0;
			}
			else {
				// It's the user account, so send it off to the user's server.
				m_iServerToContact = USERS_POP_SERVER;
				// what else?
			}
		}
		else
		{
			m_err.setdetail(NOT_ENOUGH_PARAMETERS_ERROR);
			ReturnValue = false;
		}
	}
	else if( !strncmp(strCommand.c_str(), "quit", 4) )
	{
		if(m_strParamA.length())
		{
			m_err.setdetail(TOO_MANY_PARAMETERS_ERROR);
			ReturnValue = false;
		}
		else
		{
			m_iCommand = QUIT_CMD;
			m_iServerToContact = SEND_TO_ALL_SERVERS;
		}
	}
	else
	{
		m_err.setdetail(UNKNOWN_COMMAND_ERROR);
		m_iServerToContact = DONT_SEND_TO_SERVER;
		ReturnValue = false;
	}
	
	return ReturnValue;
}

bool RZkPopProxy::RenumberCommandAndPickServer()
{
	bool ReturnValue = true;

	int iMsgNum = atoi(m_strParamA.c_str());

	switch(m_iCommand)
	{
	case TOP_NUM_CMD:
	case LIST_NUM_CMD:
	case RETR_NUM_CMD:
	case DELE_NUM_CMD:
	case UIDL_NUM_CMD:
		if(iMsgNum > m_TotalNumMessages || iMsgNum <= 0)
		{
			m_err.setdetail(MESSAGE_OUT_OF_RANGE_ERROR);
			ReturnValue = false;
		}
		else if(m_vecMsgInfo[iMsgNum].bDeleted)
		{
			m_err.setdetail(MESSAGE_DELETED_ERROR);
			ReturnValue = false;
		}
		else
		{
			m_iServerToContact = m_vecMsgInfo[iMsgNum].PopAccountIndex;
		}
		break;

	default:
		// Unknown command, or a command which doesn't apply to a specific msg
		ReturnValue = false;
		break;
	}
	if(ReturnValue)
	{
		switch(m_iCommand)
		{
		case TOP_NUM_CMD:
			snprintf((char*)m_ByteBuffer, m_ByteBufferSize, TOP_NUM_FORMAT_STRING, m_vecMsgInfo[iMsgNum].MsgNumber, m_strParamB.c_str());
			break;

		case LIST_NUM_CMD:
			snprintf((char*)m_ByteBuffer, m_ByteBufferSize, LIST_NUM_FORMAT_STRING, m_vecMsgInfo[iMsgNum].MsgNumber);
			break;

		case RETR_NUM_CMD:
			snprintf((char*)m_ByteBuffer, m_ByteBufferSize, RETR_NUM_FORMAT_STRING, m_vecMsgInfo[iMsgNum].MsgNumber);
			break;

		case DELE_NUM_CMD:
			snprintf((char*)m_ByteBuffer, m_ByteBufferSize, DELE_NUM_FORMAT_STRING, m_vecMsgInfo[iMsgNum].MsgNumber);
			break;

		case UIDL_NUM_CMD:
			snprintf((char*)m_ByteBuffer, m_ByteBufferSize, UIDL_NUM_FORMAT_STRING, m_vecMsgInfo[iMsgNum].MsgNumber);
			break;

		default:
			ZKASSERT(0);
			break;
		}
	}
	return ReturnValue;
}

bool RZkPopProxy::BuildMessageMap()
{
	unsigned int i;
	// precreate the 0th element, which doesn't map to a msg# (they start from 1)
	m_vecMsgInfo.resize(1);
	bool ReturnValue = true;
	
	m_iCommand = STAT_CMD;
	zkCStrlcpy(m_ByteBuffer, STAT_CMD_STRING, m_ByteBufferSize);

	for(i = 0; i < m_vecPopAccounts.size(); i++)
	{
		if(!SendRequestToRemoteServer(m_vecPopAccounts[i]))
		{
			ReturnValue = false;
			break;
		}
	}
	if(ReturnValue)
	{
		for(i = 0; i < m_vecPopAccounts.size(); i++)
		{
			if(!ReceiveResponseFromRemoteServer(m_vecPopAccounts[i]))
			{
				ReturnValue = false;
				break;
			}
			else if(!ResponseIsSuccess())
			{

			}
			else
			{
				ZKSTD::string strStatResponse(m_ByteBuffer, m_BytesReceived);

				size_t j = strStatResponse.find_first_of(' ');
				if(j == ZKSTD::string::npos)
				{
					ReturnValue = false;
					break;
				}

				j = strStatResponse.find_first_not_of(' ', j);

				if(j == ZKSTD::string::npos)
				{
					ReturnValue = false;
					break;
				}
			
				size_t k = strStatResponse.find_first_of(' ', j);
				
				if(k == ZKSTD::string::npos)
				{
					ReturnValue = false;
					break;
				}

				m_vecPopAccounts[i].NumMessages = atoi( strStatResponse.substr(j, k - j).c_str() );
				m_TotalNumMessages += m_vecPopAccounts[i].NumMessages;
				j = strStatResponse.find_first_not_of(' ', k);
				
				if(j == ZKSTD::string::npos)
				{
					ReturnValue = false;
					break;
				}

				k = strStatResponse.find_first_of(" \r", j);
				if(k == ZKSTD::string::npos)
				{
					ReturnValue = false;
					break;
				}
				size_t PopDropSize = atoi( strStatResponse.substr(j, k - j).c_str() );

				m_TotalPopDropSize += PopDropSize;

				j = m_vecMsgInfo.size();

				m_vecMsgInfo.resize(j + m_vecPopAccounts[i].NumMessages);

				// 0th index is just a place holder
				m_vecPopAccounts[i].vecRealToVirtualMsgNums.resize(m_vecPopAccounts[i].NumMessages + 1);
				for(k = 1; k <= m_vecPopAccounts[i].NumMessages; k++)
				{
					m_vecPopAccounts[i].vecRealToVirtualMsgNums[k] = j;
					m_vecMsgInfo[j].PopAccountIndex = i;
					m_vecMsgInfo[j].MsgNumber = k;
					j++;
				}
			}
		}
	}

	m_iCommand = LIST_CMD;
	zkCStrlcpy(m_ByteBuffer, LIST_CMD_STRING, m_ByteBufferSize);

	for(i = 0; i < m_vecPopAccounts.size(); i++)
	{
		if(m_vecPopAccounts[i].NumMessages == 0)
		{
			// We already know this popbox has no messages, and we don't want to
			// risk an -ERR responses from a grumpy server.
			continue;
		}
		if(!SendRequestToRemoteServer(m_vecPopAccounts[i]))
		{
			ReturnValue = false;
			break;
		}
	}
	if(ReturnValue)
	{
		for(i = 0; i < m_vecPopAccounts.size(); i++)
		{
			if(m_vecPopAccounts[i].NumMessages == 0)
			{
				// We already know this pop box has no messages.
				// Furthermore, we didn't send him anything,
				// so we won't expect him to reply either!
				continue;
			}
			if(!ReceiveResponseFromRemoteServer(m_vecPopAccounts[i]))
			{
				ReturnValue = false;
				break;
			}
			else if(!ResponseIsSuccess())
			{
				// Some POP servers (e.g. pop.test.0k.net) return -ERR if
				// you try to get a LIST when the box is empty.  For now,
				// we'll just not process that response; the problem is 
				// that any other ERR is now hidden from us and the user.
				ReturnValue = false;
				break;
				// Actually, we just won't ask in that case...
			}

			else if(!InterpretMultilineResponse(m_vecPopAccounts[i]))
			{
				ReturnValue = false;
				break;
			}
		}
	}

	m_iCommand = UIDL_CMD;
	zkCStrlcpy(m_ByteBuffer, UIDL_CMD_STRING, m_ByteBufferSize);

	for(i = 0; i < m_vecPopAccounts.size(); i++)
	{
		if(m_vecPopAccounts[i].NumMessages == 0)
		{
			// See comments in LIST, above.
			continue;
		}
		if(!SendRequestToRemoteServer(m_vecPopAccounts[i]))
		{
			ReturnValue = false;
			break;
		}
	}
	if(ReturnValue)
	{
		for(i = 0; i < m_vecPopAccounts.size(); i++)
		{
			if(m_vecPopAccounts[i].NumMessages == 0)
			{
				// See comments in LIST, above.
				continue;
			}
			if(!ReceiveResponseFromRemoteServer(m_vecPopAccounts[i]))
			{
				ReturnValue = false;
				break;
			}
			if(!ResponseIsSuccess())
			{
				ZKASSERT( m_vecPopAccounts[i].bUserAccount );
				// There is a better way to fail here...
				// If the user's account doesn't support UIDLs, then we should tell the MUA that
				// UIDLs aren't supported, and don't bother getting them for the nyms.  
				// Otherwise we go ahead as usual.
				m_bUserAccountSupportsUidl = false;
				// NOTE: the user account always 'happens' to be the first one in the m_vecPopAccounts.
				// We have to go around and receive the responses from all the servers, though,
				// or it screws up later communications! 
				continue;
			}

			if(!InterpretMultilineResponse(m_vecPopAccounts[i]))
			{
				ReturnValue = false;
				break;
			}
		}
	}

	for(i = 0; i < m_vecPopAccounts.size(); i++)
	{
		// bug 7271... report user account mailchecks as well.
		if(m_vecPopAccounts[i].bUserAccount)
		{
			if(m_vecPopAccounts[i].NumMessages > 1)
			{
				IZkConsoleMsg::ShowMsg(CONSIMAGE_MAIL, IDS_CONS_USER_HAS_X_MESSAGES,
					m_vecPopAccounts[i].strUserName.c_str(),
					m_vecPopAccounts[i].NumMessages);				
			}
			else if(m_vecPopAccounts[i].NumMessages == 1)
			{
				IZkConsoleMsg::ShowMsg(CONSIMAGE_MAIL, IDS_CONS_USER_HAS_ONE_MESSAGE,
					m_vecPopAccounts[i].strUserName.c_str());				
			}
			else
			{
				IZkConsoleMsg::ShowMsg(CONSIMAGE_MAIL, IDS_CONS_USER_HAS_NO_MESSAGES,
					m_vecPopAccounts[i].strUserName.c_str());
			}
		}
		else
		{
			if(m_vecPopAccounts[i].NumMessages > 1)
			{
				IZkConsoleMsg::ShowMsg(CONSIMAGE_MAIL, IDS_CONS_NYM_HAS_X_MESSAGES,
					m_vecPopAccounts[i].strUserName.c_str(),
					m_vecPopAccounts[i].NumMessages);				
			}
			else if(m_vecPopAccounts[i].NumMessages == 1)
			{
				IZkConsoleMsg::ShowMsg(CONSIMAGE_MAIL, IDS_CONS_NYM_HAS_ONE_MESSAGE,
					m_vecPopAccounts[i].strUserName.c_str());				
			}
			else
			{
				IZkConsoleMsg::ShowMsg(CONSIMAGE_MAIL, IDS_CONS_NYM_HAS_NO_MESSAGES,
					m_vecPopAccounts[i].strUserName.c_str());
			}
		}
	}
	return ReturnValue;
}

bool RZkPopProxy::InterpretMultilineResponse(PopAccount& in_PopAccount)
{
	bool ReturnValue = false;
	ZKSTD::istrstream ResponseStream(m_ByteBuffer, m_BytesReceived);

	char strBuffer[256] = {0};

	 // Get rid of the "+OK List follows\r\n"
	ResponseStream.ignore(256, '\n');

	while(true)
	{
		ResponseStream.getline(strBuffer, 256, '\r');
		ResponseStream.ignore(1); // ignore the \n which follows
		
		ZKSTD::string strLine(strBuffer);

		if(!strLine.length())
		{
			ZKASSERT(0);
			break;
		}

		if(strLine[0] == '.')
		{
			ReturnValue = true;
			break;
		}


		size_t i = strLine.find_first_of(' ');
		
		if(i == ZKSTD::string::npos)
		{
			break;
		}

		ZKSTD::string strMsgNumber (strLine.substr(0, i) );
		ZKSTD::string strMsgInfo   (strLine.substr(i + 1) );

		int iMsgNum = atoi(strMsgNumber.c_str());

		if(m_iCommand == LIST_CMD)
		{
			size_t m = in_PopAccount.vecRealToVirtualMsgNums[iMsgNum];
			size_t n = atoi(strMsgInfo.c_str());
			m_vecMsgInfo[m].SizeInBytes = n;
		}
		else if(m_iCommand == UIDL_CMD)
		{
			m_vecMsgInfo[in_PopAccount.vecRealToVirtualMsgNums[iMsgNum]].strUIDL = strMsgInfo;
		}
		else
		{
			ZKASSERT(false);
			break;
		}	
	}

	return ReturnValue;
}

bool RZkPopProxy::ReceiveRemoteAddress()
{
	bool ReturnValue = true;
#ifdef WIN32
	char szBuffer[sizeof( sockaddr )];

	// Recv the real sockaddr
	if( 1 != zkSocketReadAll(	&m_err,
								m_LocalSocket, szBuffer, sizeof (sockaddr) ) )
	{
		ReturnValue = false;
	}
	else 
	{		
		if( szBuffer[0] == AF_INET )
		{
			memcpy( &m_RemoteAddress, szBuffer, sizeof (sockaddr) );
		}
		else 
		{
			ReturnValue = false;
			ZKASSERT( false );
		}
	}
#else
    try {
        RShimInterface::getShimInterface().VerifySocketConnection(m_LocalSocket, m_RemoteAddress.m_inet.sa.sin_addr.s_addr, 
                                                                  m_RemoteAddress.m_inet.sa.sin_port);
        m_RemoteAddress.m_inet.sa.sin_family = AF_INET;
    } catch (...) {
        // VerifySocketConnection... throws to indicate errors
        return false;
    }
#endif
	return ReturnValue;
}

bool RZkPopProxy::AuthenticateWithRemoteServers()
{
	bool ReturnValue = true;

	if(!AuthenticateWithUserAccount())
	{
		ReturnValue = false;
	}
	else
	{
		ZKSTD::vector<PopAccount>::iterator it = m_vecPopAccounts.begin();

		bool bShouldKeepTryingForNyms = true;
		while(it != m_vecPopAccounts.end())
		{
			if(!it->bUserAccount)
			{	
				if(bShouldKeepTryingForNyms && 
					AuthenticateWithFreedomAccount(*it,
					bShouldKeepTryingForNyms) )
				{
					it++;
				}
				else
				{
					IZkConsoleMsg::ShowMsg(CONSIMAGE_MAIL, IDS_CONS_MAILCHECK_FAILED_FOR_NYM, it->strUserName.c_str() );

					ZKVERIFY(zkSocketFree(&m_err, &it->RemoteSocket));
					// Remove this element, and set 'it' to the next element
					
					it = m_vecPopAccounts.erase(it);
				}
			}
			else
			{
				it++;
			}
		}
		if(m_vecPopAccounts.size() == 0)
		{
			ZKVERIFY(zkSocketWriteAll(&m_err, m_LocalSocket, COULDNT_CONNECT_ERROR,
													strlen(COULDNT_CONNECT_ERROR) ));
			ReturnValue = false;
		}
	}
	return ReturnValue;
}

bool RZkPopProxy::AuthenticateWithFreedomAccount(PopAccount& in_PopAccount,
												bool& out_bShouldKeepTryingForNyms)
{
	bool ReturnValue = true;
	bool bLoggedIn = false;
	bool bSecondTry = false;
	
	out_bShouldKeepTryingForNyms = true;

	while(!bLoggedIn && ReturnValue)
	{
		if( !SendUserNameToServer(in_PopAccount) )
		{
			out_bShouldKeepTryingForNyms = false;
			ReturnValue = false;
			ZKASSERT(0);
		}
		else if( !ReceiveResponseFromRemoteServer(in_PopAccount) )
		{
			out_bShouldKeepTryingForNyms = false;
			ReturnValue = false;
			ZKASSERT(0);
		}
		else if( !ResponseIsSuccess() )
		{
			out_bShouldKeepTryingForNyms = false;
			ReturnValue = false;
			ZKASSERT(0);
		}
		else if( !SendPasswordToServer(in_PopAccount) )
		{
			out_bShouldKeepTryingForNyms = false;
			ReturnValue = false;
			ZKASSERT(0);
		}
		else if( !ReceiveResponseFromRemoteServer(in_PopAccount) )
		{
			out_bShouldKeepTryingForNyms = false;
			ReturnValue = false;
			ZKASSERT(0);
		}
		else if( !ResponseIsSuccess() )
		{
			ZKASSERT(0);
		}
		else
		{
			bLoggedIn = true;
		}

		if(!bLoggedIn && ReturnValue)
		{ // Not an outright failure....
			if(bSecondTry)
			{
				ReturnValue = false;
			}
			else
			{
				INymMail *pNymMail;
				ZKVERIFY(ZkService::GetINymCentral()->GetNym(IID_IMAIL, (void**)&pNymMail,
													in_PopAccount.strUserName));
				if(!pNymMail->UpdateMailCertificate())
				{
					ReturnValue = false;
				}
				else
				{
					// Updated the Cert successfully
					pNymMail->GetMailCertificate(in_PopAccount.strPassword);

					ZKVERIFY( zkSocketFree(&m_err, &in_PopAccount.RemoteSocket));
					ZKVERIFY( zkSocketNew(&m_err, AF_INET, SOCK_STREAM, IPPROTO_TCP, &in_PopAccount.RemoteSocket) );
					ZKVERIFY( zkSocketSetOption(&m_err, in_PopAccount.RemoteSocket, ZKSOCKET_OPTION_SendTimeout, 15 * 1000) );
					ZKVERIFY( zkSocketSetOption(&m_err, in_PopAccount.RemoteSocket, ZKSOCKET_OPTION_RecvTimeout, 45 * 1000) );

					if( !zkSocketConnectAddr(&m_err,
												in_PopAccount.RemoteSocket,
												in_PopAccount.RemoteAddress) )
					{
						ReturnValue = false;
					}
					// Get the welcome banner
					else if( !ReceiveResponseFromRemoteServer(in_PopAccount) )
					{
						ReturnValue = false;
					}
					else if( !ResponseIsSuccess() )
					{
						ReturnValue = false;
					}
					else
					{
						// Alright, we're connected, try again
						bSecondTry = true;
					}
				}
			}
		}
	}
	return ReturnValue;
}

bool RZkPopProxy::SendUserNameToServer(PopAccount& in_PopAccount)
{
	snprintf(m_ByteBuffer, m_ByteBufferSize, "USER %s\r\n", in_PopAccount.strUserName.c_str());
	return SendRequestToRemoteServer(in_PopAccount);
}

bool RZkPopProxy::SendPasswordToServer(PopAccount& in_PopAccount)
{
	snprintf(m_ByteBuffer, m_ByteBufferSize, "PASS %s\r\n", in_PopAccount.strPassword.c_str());
	return SendRequestToRemoteServer(in_PopAccount);
}

bool RZkPopProxy::AuthenticateWithUserAccount()
{
	bool ReturnValue = true;
	bool bLoggedIn = false;
	// User has already been given the welcome banner
	if(!m_bCheckingUserAccount)
	{
		while(ReturnValue && !bLoggedIn)
		{
			if( !ReceiveRequestFromMUA() )
			{
				ErrorResponse();
				ReturnValue = false;
			}
			else if( !ParseRequest() )
			{
				ZKVERIFY(ErrorResponse());
				// Allow them to try again - don't return false
			}
			else if( m_iCommand != USER_CMD && m_iCommand != APOP_CMD)
			{
				// Allow them to try again - don't return false
				m_err.setdetail(INVALID_SEQUENCE_ERROR);
				ErrorResponse();
			}
			else
			{
				switch(m_iCommand)
				{
				case APOP_CMD:
					
					if(!zkSocketWriteAll(&m_err, m_LocalSocket,
						APOP_ACCEPTED_RESPONSE, strlen(APOP_ACCEPTED_RESPONSE) ))
					{
						ReturnValue = false;
					}
					else
					{
						// We're done (only one transaction)
						bLoggedIn = true;
					}
					break;

				case USER_CMD:
					
					if(!zkSocketWriteAll(&m_err, m_LocalSocket,
						USER_ACCEPTED_RESPONSE, strlen(APOP_ACCEPTED_RESPONSE) ))
					{
						ReturnValue = false;
					}
					else if( !ReceiveRequestFromMUA() )
					{
						ZKVERIFY(ErrorResponse());
						ReturnValue = false;
					}

					// Alright: sometimes Outlook Express sends a blank password
					// (bug 6075), and we really aren't sure why.  FORTUNATELY,
					// we really don't care about the password, so let's just let it
					// slide regardless (I've removed the ParseRequest() part here).
					else if(!zkSocketWriteAll(&m_err, m_LocalSocket,
						PASS_ACCEPTED_RESPONSE, strlen(PASS_ACCEPTED_RESPONSE) ))
					{
						ReturnValue = false;
					}
					else
					{
						bLoggedIn = true;
					}
					break;
				}
			}
		}
	}
	else
	{
		while(ReturnValue && !bLoggedIn)
		{
			if( !ReceiveRequestFromMUA() )
			{
				ErrorResponse();
				ReturnValue = false;
			}
			else if( !ParseRequest() )
			{
				ZKVERIFY(ErrorResponse());
				// Allow them to try again - don't return false
			}
			else if( m_iCommand != USER_CMD && m_iCommand != APOP_CMD)
			{
				// Allow them to try again - don't return false
				m_err.setdetail(INVALID_SEQUENCE_ERROR);
				ErrorResponse();
			}
			else if(!SendRequestToRemoteServer(m_vecPopAccounts[0]))
			{ // Either APOP or USER
				ErrorResponse();
				ReturnValue = false;
			}
			else if( !ReceiveResponseFromRemoteServer(m_vecPopAccounts[0]) )
			{
				ErrorResponse();
				ReturnValue = false;
			}
			else if( !ResponseIsSuccess() )
			{
				// Pass the -err response through to MUA
				PassResponseThroughToMUA();
			}
			else if( !PassResponseThroughToMUA() )
			{
				ReturnValue = false;
			}
			else
			{
				switch(m_iCommand)
				{
				case APOP_CMD:
					// We're done (only one transaction)
					bLoggedIn = true;
					break;

				case USER_CMD:
					// There's also the PASS command to handle...
					if( !ReceiveRequestFromMUA() )
					{
						ZKVERIFY(ErrorResponse());
						ReturnValue = false;
					}
					else if( !ParseRequest() )
					{
						ZKVERIFY(ErrorResponse());
						// Allow them to try again - don't return false
					}
					else if( m_iCommand != PASS_CMD)
					{
						// Allow them to try again - don't return false
						m_err.setdetail(INVALID_SEQUENCE_ERROR);
						ErrorResponse();
					}
					else if(!SendRequestToRemoteServer(m_vecPopAccounts[0]))
					{
						ZKVERIFY(ErrorResponse());
						ReturnValue = false;
					}
					else if( !ReceiveResponseFromRemoteServer(m_vecPopAccounts[0]) )
					{
						ZKVERIFY(ErrorResponse());
						ReturnValue = false;
					}
					else if( !ResponseIsSuccess() )
					{
						// Allow them to try again - don't return false
						PassResponseThroughToMUA();
					}
					else if( !PassResponseThroughToMUA() )
					{
						ReturnValue = false;
					}
					else
					{
						bLoggedIn = true;
					}
					break;

				default:
					// Allow them to try again - don't return false
					m_err.setdetail(INVALID_SEQUENCE_ERROR);
					ErrorResponse();
					break;
				}
			}
		}
	}
	return ReturnValue;
}

bool RZkPopProxy::BuildAccountList()
{	
	bool ReturnValue = true;

	ZKS::Lock_Guard<WZkMutex> gate(m_gMutex);

	bool bCheckActiveOnly = ZkService::GetIClientCentral()->GetPreference(ZK_CLIENT_PREF_SET, 
		PREF_CHECK_MAIL_FOR_ACTIVE_NYM_ONLY);

	INym *pNym = NULL;
    ZkService::GetINymCentral()->GetActiveNym(IID_INYM, (void**)&pNym);

	m_bCheckingUserAccount = ( !ZkNetHelp::IsRemoteHostDummyMailServer(	m_RemoteAddress )
			// bug 7271... don't check user's box if you have a premium nym selected and want 
			// to check active nym only
			&& !( bCheckActiveOnly 
					&& ZkService::GetINymCentral()->GetActiveNymIndex() != USER_NYM_INDEX 
					&& pNym->DoesNymUseMail() 
					)
			);

	if(m_bCheckingUserAccount)
	{
		PopAccount UserAccount;
		memcpy( &UserAccount.RemoteAddress, &m_RemoteAddress, sizeof( ZkSocketAddress ) );
		UserAccount.bUserAccount = true;

		// This will be replaced by the user's pop login
		// but this will help us if an error occurs before
		// that point...
		UserAccount.strUserName = "[Real Identity]";

		m_vecPopAccounts.push_back(UserAccount);
	}
	else
	{
		if (!ZkService::GetINymCentral()->AtLeastOneNymUseMail())
		{
             if(!ZkService::GetIClientCentral()->GetPreference(ZK_SHOW_AGAIN_PREF_SET, PREF_NO_SHOW_NO_MAIL_NYMS))
			{
				if(IZkPopup::MessageCheckBox(IDS_MAIL_NO_MAIL_NYMS))
				{
					ZkService::GetIClientCentral()->SetPreference(ZK_SHOW_AGAIN_PREF_SET, PREF_NO_SHOW_NO_MAIL_NYMS, true);
				}
				ReturnValue = false;
			}
        }
        else
        {
		
            if(bCheckActiveOnly && (ZkService::GetINymCentral()->GetActiveNymIndex() == USER_NYM_INDEX
                                    || !pNym->DoesNymUseMail()) )
            {
                if(!ZkService::GetIClientCentral()->GetPreference(ZK_SHOW_AGAIN_PREF_SET, PREF_NO_SHOW_MAIL_SELECT_NYM))
                {
					if( IZkPopup::MessageCheckBox(IDS_MAIL_MUST_SELECT_NYM) )
					{
						ZkService::GetIClientCentral()->SetPreference(ZK_SHOW_AGAIN_PREF_SET, PREF_NO_SHOW_MAIL_SELECT_NYM, true);
					}
                    ReturnValue = false;
                }
            }
		}
	}

	if(!ReturnValue)
	{

	}
	else if(!m_gbNymAccountsLocked)
	{
		m_gbNymAccountsLocked = true;
		m_bHaveLockOnNymAccounts = true;

		try
		{
			//Bug 7853
			int nMailWorthyNyms = 0;
			if( m_bCheckingUserAccount ) {
				++nMailWorthyNyms;
			}
			for (size_t i = 1; i < ZkService::GetINymCentral()->GetNumberOfNyms(); ++i)
			{
				if ( ZkService::GetINymCentral()->DoesNymUseMail(i) )
				{	// bug 7271: treat "user Nym" as any other nym, for purposes of showing
					// the mail correlation dialog
					if (++nMailWorthyNyms >= 2)
					{
						break;
					}
				}
			}
			if(!bCheckActiveOnly &&
				nMailWorthyNyms >= 2 &&
				!ZkService::GetIClientCentral()->GetPreference(ZK_SHOW_AGAIN_PREF_SET, PREF_NO_WARN_MAIL_CORRELATION))
			{
				switch(ZkProxyDialogs::LaunchFirstMailCheckDlg())
				{
				case ZK_MAIL_CHECK_ACTIVE:
					ZkService::GetIClientCentral()->SetPreference(ZK_CLIENT_PREF_SET, PREF_CHECK_MAIL_FOR_ACTIVE_NYM_ONLY, true);
					ZkService::GetIClientCentral()->SetPreference(ZK_SHOW_AGAIN_PREF_SET, PREF_NO_WARN_MAIL_CORRELATION, true);
					ZkService::GetIClientPersist()->SetDirty();
					bCheckActiveOnly = true;
					break;

				case ZK_MAIL_CHECK_ALL:
					ZkService::GetIClientCentral()->SetPreference(ZK_SHOW_AGAIN_PREF_SET, PREF_NO_WARN_MAIL_CORRELATION, true);
					ZkService::GetIClientPersist()->SetDirty();
					break;

				case ZK_MAIL_CHECK_BLOCK:
					m_err.setdetail(USER_CANCELLED_ERROR);
					ErrorResponse();
					ReturnValue = false;
					break;

				default:
					ErrorResponse();
					ZKASSERT(0);
					ReturnValue = false;
					break;

				}
			}

			if(!ReturnValue)
			{

			}
			else if(bCheckActiveOnly)
			{
				size_t nIndex = ZkService::GetINymCentral()->GetActiveNymIndex();
				if(nIndex != USER_NYM_INDEX)
				{
					AddNymToAccountList(nIndex);
				}
			}
			else
			{
				for(size_t i = 1; i < ZkService::GetINymCentral()->GetNumberOfNyms(); i++)
				{
					AddNymToAccountList(i);
				}
			}
		}
		catch(RZkException&)
		{
			// Failed to find a POP server... So, we just check your regular account.
			IZkConsoleMsg::ShowMsg(CONSIMAGE_MAIL, IDS_CONS_NO_POP_SERVER);
		}
	}
	if(!m_vecPopAccounts.size())
	{
		ReturnValue = false;
	}

	return ReturnValue;
}

void RZkPopProxy::AddNymToAccountList(size_t in_index)
{
	INym *pNym = NULL;
	ZkService::GetINymCentral()->GetNym(IID_INYM, (void**)&pNym, in_index);
	if(!pNym->DoesNymUseMail())
	{
		return;
	}

	PopAccount Account;

	ZkService::GetINymCentral()->GetNymName(in_index, Account.strUserName );
	ZkService::GetINiqs()->GetPopServerAddr((sockaddr_in*)&Account.RemoteAddress);

	// Get the Nym's mail password
	INymMail *pNymMail = NULL;

	ZKVERIFY(ZkService::GetINymCentral()->GetNym(IID_IMAIL, (void**) &pNymMail, in_index));
	pNymMail->GetMailCertificate(Account.strPassword);

	if(!Account.strPassword.length())
	{
		if(pNymMail->UpdateMailCertificate())
		{
			pNymMail->GetMailCertificate(Account.strPassword);
		}
	}
	
	if(!Account.strUserName.length())
	{
		char buffer[64] = {0};
		snprintf(buffer, 64, "[Unknown Nym %d]", in_index);
		Account.strUserName = buffer;
	}
	if(Account.strPassword.length())
	{
		m_vecPopAccounts.push_back(Account);
	}
	else
	{
		IZkConsoleMsg::ShowMsg(CONSIMAGE_MAIL, IDS_CONS_MAILCHECK_FAILED_FOR_NYM, Account.strUserName.c_str() );
	}
}

bool RZkPopProxy::ConnectToRemoteServers()
{
	bool ReturnValue = true;

	// If we fail to connect to a nym mail server once, don't waste
	// our time trying again for another nym - it's the same server.
	bool bNymMailServerFailed = false;

	ZKASSERT(m_vecPopAccounts.size());
	ZKSTD::vector<PopAccount>::iterator it = m_vecPopAccounts.begin();
	
	while(it != m_vecPopAccounts.end())
	{
		bool bConnected = false;

		if(!bNymMailServerFailed || it->bUserAccount)
		{// Haven't yet failed to connect
			ZKVERIFY( zkSocketNew(&m_err, AF_INET, SOCK_STREAM, IPPROTO_TCP, &it->RemoteSocket) );
			ZKVERIFY( zkSocketSetOption(&m_err, it->RemoteSocket, ZKSOCKET_OPTION_SendTimeout, 15 * 1000) );
			ZKVERIFY( zkSocketSetOption(&m_err, it->RemoteSocket, ZKSOCKET_OPTION_RecvTimeout, 45 * 1000) );

			// For debugging.
			it->strLastCommand = "[Connected]";

			if( (zkSocketConnectAddr(&m_err,
										it->RemoteSocket,
										it->RemoteAddress) ))
			{
				bConnected = true;
				if( !ReceiveResponseFromRemoteServer(*it) )
				{
					IZkConsoleMsg::ShowImportantMsg(CONSIMAGE_MAIL,IDS_CONS_PROXY_POP_SERVER_BUSY);
					bConnected = false;
				}
				else if( !ResponseIsSuccess() )
				{
					bConnected = false;
				}
				else if(it->bUserAccount)
				{
					if(!PassResponseThroughToMUA() )
					{
						bConnected = false;
						ReturnValue = false;
					}
				}
			}
		}
		
		if(bConnected)
		{
			it++;
		}
		else
		{
			if(!it->bUserAccount)
			{
				// We won't try to connect again for the next nym.
				bNymMailServerFailed = true;
			}

			IZkConsoleMsg::ShowMsg(CONSIMAGE_MAIL, IDS_CONS_MAILCHECK_FAILED_FOR_NYM, it->strUserName.c_str() );
			if(it->RemoteSocket)
			{
				ZKVERIFY(zkSocketFree(&m_err, &it->RemoteSocket));
			}
			// Remove this element, and set 'it' to the next element
			
			it = m_vecPopAccounts.erase(it);
		}
	}

	if(!m_vecPopAccounts.size())
	{	
		// Didn't connect to any servers.
		ReturnValue = false;
	}
	else
	{
		if(m_bCheckingUserAccount)
		{
			if(!m_vecPopAccounts[0].bUserAccount)
			{
				// if we were supposed to connect to user's server,
				// we MUST connect to it, or fail
				ReturnValue = false;
			}
		}
		else
		{
			// We didn't need to contact the user server, but that means we haven't
			// yet sent a welcome banner. Do that now...
			if(!zkSocketWriteAll(&m_err, m_LocalSocket,
				WELCOME_BANNER_RESPONSE, strlen(WELCOME_BANNER_RESPONSE) ))
			{
				ReturnValue = false;
			}
		}
	}

	if(!ReturnValue)
	{
		// in case it was set by zkSocket*
		zkErrorClear(&m_err);
		m_err.setdetail(COULDNT_CONNECT_ERROR);
	}
	return ReturnValue;
}

bool RZkPopProxy::ReceiveRequestFromMUA()
{
	struct timeval timeout = {0};
	timeout.tv_sec = 45;

	bool ReturnValue = true;

	size_t BytesRead = 0;
	m_BytesReceived = 0;
	while(true)
	{
        DZkError derr;
        int iSocket = -1;

        if(!zkSocketWaitReadableOrClosed(&derr, m_listSockets, &timeout, &iSocket, 
			RZkProxy::GetSignalFD())) {
#ifdef WIN32
            int lastError = WSAGetLastError();
            if (WSAEINTR == lastError || WSAENOTSOCK == lastError) continue;
#else
            if (EINTR == errno) continue;
#endif            
			// On windows this is normal. One of the file descriptors was closed
			// underneath WaitReadableOrClosed. On linux we will exit through the 
			// path below ( == -2 )
            ReturnValue = false;
            break;
        }
        else if (iSocket < 0) {
            // if -1 then timeout and we just spin the loop
            // if -2 then signalled by DInetD and we want to quit.
            if ( iSocket == -2 ) {
                ReturnValue = false;
                break;            
            }
        }
		else if( !zkSocketRead(
				&m_err,
				m_LocalSocket,
				m_ByteBuffer + m_BytesReceived,
				m_ByteBufferSize - 1 - m_BytesReceived,
				(unsigned int*)&BytesRead))
		{
			ReturnValue = false;
			break;
		}
		else
		{
			m_BytesReceived += BytesRead;
			m_ByteBuffer[m_BytesReceived] = '\0';
			if(strchr(m_ByteBuffer, '\r'))
			{
				break;
			}
		}
	}

	ZKTRACE1("Client Said: %s\r\n", m_ByteBuffer);
	return ReturnValue;
}

bool RZkPopProxy::ReceiveResponseFromRemoteServer(PopAccount& in_Account)
{
	bool ReturnValue = true;
	bool bFinished = false;

	size_t BytesRead = 0;
	m_BytesReceived = 0;

	ResizeByteBuffer(in_Account);

	while(!bFinished)
	{
		if( m_BytesReceived + 1 >=  m_ByteBufferSize)
		{
			// message is longer than we thought
			// (see the discussion on byte-stuffing in ResizeByteBuffer)
			zkMemRealloc(&m_err, (void **)&m_ByteBuffer, (int)(m_ByteBufferSize * 1.5));
			m_ByteBufferSize = (int)(m_ByteBufferSize * 1.5);
		}

		if( !zkSocketRead(
				&m_err,
				in_Account.RemoteSocket,
				m_ByteBuffer + m_BytesReceived,
				m_ByteBufferSize - 1 - m_BytesReceived,
				(unsigned int*)&BytesRead))
		{
			ReturnValue = false;
			break;
		}
		else
		{
			m_BytesReceived += BytesRead;
			m_ByteBuffer[m_BytesReceived] = '\0';
			switch(m_iCommand)
			{
				// Multi-line responses:
				case TOP_NUM_CMD:
				case UIDL_CMD:
				case LIST_CMD:
				case RETR_NUM_CMD:
					if(ResponseIsSuccess())
					{
						if(strstr(m_ByteBuffer, "\r\n.\r\n"))
						{
							bFinished = true;
						}
					}
					else if(strstr(m_ByteBuffer, "\r\n"))
					{// Error responses are still single-line
						bFinished = true;
					}
					break;

				case USER_CMD:
				case PASS_CMD:
				case APOP_CMD:
				case STAT_CMD:
				case DELE_NUM_CMD:
				case LIST_NUM_CMD:
				case UIDL_NUM_CMD:
				case RSET_CMD:		
				case QUIT_CMD:
				case NOOP_CMD:
					if(strstr(m_ByteBuffer, "\r\n"))
					{
						bFinished = true;
					}
					break;

				default:
					ZKASSERT(0);
					bFinished = true;
					break;
			}
		}
	}

	// bug 8042... report POP error (from server) to debug Console
	if( !ResponseIsSuccess() )
	{
		ZKSTD::string strResponse = "[No response]";
		if(m_BytesReceived > 2)
		{
			strResponse = ZKSTD::string(m_ByteBuffer, strlen(m_ByteBuffer) - 2);
		}
		// bug 8568... don't show user's password in console
		if( m_iCommand == PASS_CMD || in_Account.strLastCommand.at(0) == 'P' 
			|| in_Account.strLastCommand.at(0) == 'p' )
		{
			in_Account.strLastCommand.erase();
			in_Account.strLastCommand.append( "PASS <hidden>" );
		}
		IZkConsoleMsg::ShowMsg(	CONSIMAGE_MAIL,
							IDS_CONS_PROXY_POP_VERBOSE_ERROR,
							in_Account.strLastCommand.c_str(), 
							strResponse.c_str());
	}

	if(ReturnValue)
	{
		ZKTRACE1("Server Said: %s\r\n", m_ByteBuffer);
	}
	return ReturnValue;
}

void RZkPopProxy::ResizeByteBuffer(PopAccount& in_Account)
{
	size_t DesiredSize = 0;
	switch(m_iCommand)
	{
		// Multi-line responses:
		case TOP_NUM_CMD:
			// We can only guess at this one - in theory, the whole msg could be on one
			// line (or even all in the headers) so we could retrieve an arbitrary amount
			// of data. We'll guess that it's less than DEFAULT_POP_BUFFER_SIZE (if not,
			// the buffer can be reallocated above.)
			DesiredSize = DEFAULT_POP_BUFFER_SIZE;
			break;

		case UIDL_CMD:
			//		Starting Line	(512 max CRLF incl.)
			// +	76 per line		("+OK " + 70 max UID + CRLF)
			// +	Termination		("." + CRLF)
			DesiredSize = 512 + (76 * in_Account.NumMessages) + 3;
			break;

		case LIST_CMD:
			//		Starting Line	(512 max CRLF incl.)
			// +	512 per line	(Arbitrary comments can come after message size)
			// +	Termination		("." + CRLF)
			DesiredSize = 512 + (512 * in_Account.NumMessages) + 3;
			break;

		case RETR_NUM_CMD:
			
			//		Starting Line	(512 max CRLF incl.)
			// +	Message Size
			// +	Termination		("." + CRLF)
			// Note that the message can actually be bigger. Lines that contain
			// only one or more '.' are "Byte-Stuffed" (have an extra '.' at the
			// beginning). This extra byte is not counted in the size. Therefore
			// the message could be as much as 33% bigger (if every line was
			// ".\r\n" they would become "..\r\n"
			// For this reason, we have an emergency reallocation strategy in
			// ReceiveResponseFromRemoteServer()
			DesiredSize = 512 + m_vecMsgInfo[atoi(m_strParamA.c_str())].SizeInBytes + 3;
			break;

		case USER_CMD:
		case PASS_CMD:
		case APOP_CMD:
		case STAT_CMD:
		case DELE_NUM_CMD:
		case LIST_NUM_CMD:
		case UIDL_NUM_CMD:
		case RSET_CMD:		
		case QUIT_CMD:
		case NOOP_CMD:
			// Single line, max size 512.
			DesiredSize = 512;
			break;

		default:
			ZKASSERT(0);
			break;
	}

	if(DesiredSize < DEFAULT_POP_BUFFER_SIZE)
	{
		DesiredSize = DEFAULT_POP_BUFFER_SIZE;
	}

	if(DesiredSize > m_ByteBufferSize)
	{
		ZKVERIFY(zkMemRealloc(&m_err, (void**)&m_ByteBuffer, DesiredSize));
		m_ByteBufferSize = DesiredSize;
	}
	else if(DesiredSize < m_ByteBufferSize)
	{
		// We don't necessarily need to free this buffer.
		// let's think about this.
		
		ZKVERIFY(zkMemRealloc(&m_err, (void**)&m_ByteBuffer, DesiredSize));
		m_ByteBufferSize = DesiredSize;
	}
}
