// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RZkPopProxy.h


///////////////////////////////////////////////////////////////
// For detailed information about the Freedom Pop Proxy, see //
// freedom/client/zkproxies/doc/RZkPopProxy.html             //
///////////////////////////////////////////////////////////////

#ifndef _RZKPOPPROXY_H
#define _RZKPOPPROXY_H
#pragma warning (disable : 4251)

#include "IZkProxy.h"
#include <zksocket.h>
#include "ZkProxiesInc.h"
#include <DZkError.h>
#include <vector>
#include <WSyncTypes.h>

// Default size of the receive buffer. The buffer may grow larger.
// See RZkPopProxy::ResizeByteBuffer()
#define DEFAULT_POP_BUFFER_SIZE 2048


// Information about a POP account
// The POP account may be the user's real POP account
// or the POP account of one of the user's Nyms.
//
// see freedom/client/zkproxies/doc/PopProxyInternals.JPG
typedef struct PopAccount
{
	PopAccount::PopAccount()
		: RemoteSocket(NULL)
		, bUserAccount(false)
		, NumMessages(0)
	{
		memset(&RemoteAddress, 0, sizeof(ZkSocketAddress) );
	}

	ZkSocketAddress RemoteAddress;
	ZKSTD::string	strUserName;
	ZKSTD::string	strPassword;
	ZkSocket		RemoteSocket;

	// Is this the user's normal account?
	// (Can be a special case)
	bool			bUserAccount;
	size_t			NumMessages;
	// if X is the index of the msg on this server,
	// vecRealToVirtualMsgNums[X] is what the MUA thinks the index is
	ZKSTD::vector<int>	vecRealToVirtualMsgNums;

	// For debugging purposes, keep track of the last thing
	// we said to this server.
	ZKSTD::string strLastCommand;
}PopAccount;

// Information about a particular Message
// The message may be on the user's real POP account or
// in the POP account of one of the user's Nyms.
//
// see freedom/client/zkproxies/doc/PopProxyInternals.JPG
typedef struct MsgInfo
{
	MsgInfo::MsgInfo()
		: PopAccountIndex(-1)
		, MsgNumber(-1)
		, SizeInBytes(0)
		, bDeleted(false)
		, bFreedomMsg(false)
		, bRetrieved(false)
	{
	}

	// Which pop account does the message actually exist on?
	// (index into RZkProxy::m_vecPopAccounts)
	int PopAccountIndex;

	// Real number of the Msg on that account
	int MsgNumber;
	int SizeInBytes;

	// Unique ID
	std::string strUIDL;
	// Has it been marked as deleted on the server?
	bool bDeleted;
	// Has the msg been retrieved?
	bool bRetrieved;
	// Is this a Freedom Encrypted Msg?
	bool bFreedomMsg;
}MsgInfo;

class ZkProxies_API RZkPopProxy : public IZkProxy
{
public:
	RZkPopProxy(ZkSocket in_LocalSocket);
	~RZkPopProxy();

	// Start the Proxy
	bool RunProxy();

private:
	// PURPOSE: Receive the address of the original
	//			remote host from the shim layer
	// PROMISE: Will set m_RemoteAddress, or return false
	// REQUIRE: m_LocalSocket is set. Shim sends the address properly.
	bool ReceiveRemoteAddress();
	
	// PURPOSE: Populates m_vecPopAccounts . Checks to see if the user is
	//			connecting to mail.freedom.net . If so, sets
	//			m_bCheckingUserAccount to true. Otherwise creates a PopAccount
	//			to represent the user's normal POP account. Determines which
	//			nyms to check mail for and calls AddNymToAccountList()
	bool BuildAccountList();
	
	// PURPOSE: Attempts to create an entry for the nym in m_vecPopAccounts.
	//			Checks to see if the nym actually has mail privileges and a
	//			mail certificate. If the nym has no mail certificate,
	//			attempts to retrieve one from the nym server.
	void AddNymToAccountList(size_t in_index);

	// PURPOSE: Connect to POP servers and read the welcome banner.
	//			Verify that it is Positive (+OK). If we connect to the
	//			user's normal account, return the banner from this account
	//			to the MUA.
	// PROMISE: If any connection fails, removes that account from the vector
	//			If all connections fail, returns false
	//
	//			Will put an appropriate error in m_ByteBuffer.
	// REQUIRE: m_vecPopAccounts has been setup
	bool ConnectToRemoteServers();

	// PURPOSE: Authenticate with each account. For user account, interactively
	//			requests the USER/PASS combo from the MUA and sends it to the remote
	//			server. For Freedom accounts, sends the Nym name and Mail Cert.
	//			(assuming they've been properly stored in the PopAccount structure)
	// PROMISE: If a Freedom account fails to authenticate once, update its
	//			mail certificate and try again. If the second attempt fails, remove
	//			it from the list. If no accounts are able to authenticate,
	//			returns false.
	//			Also, if we fail to connect to a nym account for network reasons
	//			(ie, not just a bad mail cert), won't try to connect for other
	//			nyms, since they will probably also fail and that's time consuming.
	//
	//			In the case of an error, will send it's own error response.
	// REQUIRE: m_vecPopAccounts has been setup
	bool AuthenticateWithRemoteServers();

	// PURPOSE: Accept the username and password from the MUA and pass it
	//			to the user's real mail server. If the user has no POP account,
	//			(ie, they connected to mail.freedom.net), then just accept
	//			whatever the MUA gives us and throw it away. Also supports
	//			APOP.
	bool AuthenticateWithUserAccount();

	// PURPOSE: Send the Nym name and Mail Certificate to the Freedom
	//			POP server. If the certificate is rejected, retrieve a new 
	//			mail cert from the nym server and try again.
	//
	//			If there is a network failure, sets the boolean to false.
	//			This tells the caller not to try to connect for other nyms,
	//			since these connections are also likely to fail.
	bool AuthenticateWithFreedomAccount(PopAccount& in_PopAccount,
									bool& out_bShouldKeepTryingForNyms);

	// PURPOSE: Build the "USER xxx\r\n" command line using the nym's name
	//			and send it to the Freedom POP server
	bool SendUserNameToServer(PopAccount& in_PopAccount);

	// PURPOSE: Build the "PASS xxx\r\n" command line using the nym's Mail
	//			Certificate and send it to the Freedom POP server
	bool SendPasswordToServer(PopAccount& in_PopAccount);

	// PURPOSE: Request message information from each account, setup the
	//			m_vecMsgInfo and add information to m_vecPopAccounts
	//			This function performs a STAT, UIDL, and LIST command on each
	//			account.
	// PROMISE: If any of these commands (or the interpretation of the results)
	//			fails, returns false;
	// REQUIRE: m_vecPopAccounts has been setup
	bool BuildMessageMap();

	// PURPOSE: Parses the POP command in m_ByteBuffer.
	//			Sets m_iCommand, m_iServerToContact, m_strParamA, and
	//			m_strParamB. If there is a parse error, puts the appropriate
	//			error string into the detail of m_err. This string should
	//			be returned to the MUA.
	//			If there is no error, m_ByteBuffer contains the reformatted
	//			command, and m_iServerToContact indicates which server to send
	//			the command to, either DONT_SEND_TO_SERVER,
	//			SEND_TO_ALL_SERVERS, USERS_POP_SERVER, or else an index into
	//			m_vecPopAccounts.
	bool ParseRequest();

	// PURPOSE: Called by ParseRequest() in the case of commands that
	//			refer to a specific message. Determines which server
	//			the message resides on and the index of the message
	//			on that server. Then determines what command should
	//			be executed in response to the MUA's request.
	//			Will set m_ByteBuffer and m_iServerToContact appropriately.
	bool RenumberCommandAndPickServer();
	
	// PURPOSE: Sends the command in m_ByteBuffer to all accounts listed
	//			in m_vecPopAccounts. Returns false if any one send fails.
	bool SendRequestToAllServers();

	// PURPOSE: Calls ReceiveAndProcessResponseFromOneServer() for each
	//			account listed in m_vecPopAccounts
	bool ReceiveAndProcessResponseFromAllServers();
	
	// PURPOSE: Receives a response from one server and processes it
	//			appropriately. Leaves a response in m_ByteBuffer that
	//			is appropriate to return to the user agent.
	bool ReceiveAndProcessResponseFromOneServer(PopAccount&);

	// PURPOSE: Receives a response from one server. Checks m_iCommand
	//			to determine if a multi-line response is expected
	//			(ie TOP, RETR, LIST, or UIDL). Calls ResizeByteBuffer()
	//			to make sure the byte buffer is big enough for the response.
	bool ReceiveResponseFromRemoteServer(PopAccount&);

	// PURPOSE: Checks m_iCommand to determine what type of response is
	//			expected, and how big it is expected to be. Resizes
	//			m_ByteBuffer to be the maximum likely size.
	//			The caller should be prepared for the actual response to
	//			be bigger, and respond appropriately.
	void ResizeByteBuffer(PopAccount& in_PopAccount);

	// PURPOSE: Attempts to receive a single line of data from the MUA.
	//			Stops as soon as it sees a '\r'
	bool ReceiveRequestFromMUA();

	// PURPOSE: Builds an appropriate response for the command issued by
	//			the MUA and sends it.
	bool SendResponseToMUA();

	// PURPOSE: Checks to see if m_ByteBuffer begins with a '+'.
	bool ResponseIsSuccess();

	// PURPOSE: Uses the detail and function name from m_err to build
	//			a meaningful POP error response, and sends it back to the
	//			MUA.
	bool ErrorResponse();

	// PURPOSE: Sends the data in m_ByteBuffer to the specified server.
	bool SendRequestToRemoteServer(PopAccount&);

	// PURPOSE: Handle the response from a UIDL or LIST command, use it
	//			to populate m_vecMsgInfo.
	bool InterpretMultilineResponse(PopAccount&);

	// PURPOSE: Send the data in m_ByteBuffer to MUA.
	bool PassResponseThroughToMUA();

	// PURPOSE: Looks for X-Freedom-Message-Auth header in m_ByteBuffer. If found, this
	//			is a freedom msg.
	bool IsFreedomMail();

	// PURPOSE: Looks for Content-Type header, and parses it to find the MIME boundary.
	bool GetMimeBoundary(ZKSTD::string& out_strMimeBoundary);

	// PURPOSE: Gets the contents of the X-Freedom-Message-Auth field, decodes
	//			it from base-64.
	bool GetAuthenticationToken(void **out_ppAuthToken,
								size_t *out_lenAuthToken);

	// PURPOSE: Iterates through all MIME parts in a message and finds the
	//			part labeled application/x-Freedom-Mail-Message-Body.
	//			If found, decodes it from base-64
	bool GetEncryptedMsgBody(void **out_ppEncryptedBody,
								size_t *out_lenEncryptedBody);

	// PURPOSE: Decrypts a Freedom message and verifies its authenticity.
	//			For Nym->Nym mail:
	//				Checks the signature
	//				Makes sure the From/Sender fields are the same as the
	//				signing Nym.
	//			For Internet->Nym mail:
	//				Checks the MAC authentication token
	//
	//			Also removes Disposition-Notification-To fields (to prevent
	//			Read-Receipts from being sent).
	bool ProcessFreedomMail(const ZKSTD::string& strNymName, 
						bool& out_bKeyQryFailed,
						ZKSTD::string& out_strFailureDetail
						);

	// PURPOSE: Frees all ZkSocket structures.
	void CleanUpSockets();

	
	// PURPOSE: Deletes any nym mail that has been retrieved.
	//			See bug 5954
	bool DeleteAllRetrievedNymMail();

	char *m_ByteBuffer;
	size_t m_ByteBufferSize;

	size_t m_BytesReceived;
	int m_iServerToContact;
	int m_iCommand;

	static WZkMutex m_gMutex;
	bool m_bHaveLockOnNymAccounts;
	static bool m_gbNymAccountsLocked;

	enum COMMANDS
	{
		USER_CMD,
		PASS_CMD,
		APOP_CMD,

		UIDL_CMD,
		LIST_CMD,
		RSET_CMD,
		NOOP_CMD,
		STAT_CMD,

		LIST_NUM_CMD,
		RETR_NUM_CMD,
		DELE_NUM_CMD,
		TOP_NUM_CMD,  // Will we bother supporting this?
		UIDL_NUM_CMD,

		QUIT_CMD,

		OTHER_CMD
	};
	bool m_bCheckingUserAccount;
	bool m_bUserAccountSupportsUidl;
	bool m_bShowVerboseConsoleMsg;	// bug 8042

	ZKSTD::vector<PopAccount> m_vecPopAccounts;
	ZKSTD::vector<MsgInfo> m_vecMsgInfo;
	
	int m_TotalPopDropSize;
	int m_TotalNumMessages;
	int m_NumAccounts;

	ZKSTD::string m_strParamA;
	ZKSTD::string m_strParamB;

	ZkSocketAddress m_RemoteAddress;
	ZkSocket m_LocalSocket;
	ZkList m_listSockets; // List to hold sockets for WaitReadableOrClosed

	DZkError m_err;

	RZkPopProxy() {};
	RZkPopProxy(RZkPopProxy&) {};
};

#endif // _RZKPOPPROXY_H
