// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RZkProxy.cpp

#include "RZkProxy.h"
#include "RZkPassThruProtocol.h"
#include "RZkHttpProtocol.h"
#include "RZkSmtpProtocol.h"
#include "RZkNntpProtocol.h"
#include "RZkIrcProtocol.h"


#ifndef WIN32
#  include <sys/time.h>           // for struct timeout (for zkSocketWaitReadable or closed)
#  include <sys/errno.h>
#  include <../linux/RShimInterface.h>     // for VerifySocketConnection
#endif

int RZkProxy::m_signalFD = -1;

// in_LocalSocket is the socket returned by accept from CDInetDApp::HandleConnections to one of our locahost proxies
// m_LocalSocket holds that value here
RZkProxy::RZkProxy(ZkSocket in_LocalSocket, IZkProxy::PROXY_PROTOCOLS in_WhichProtocol)
		: m_LocalSocket(in_LocalSocket)
		, m_RemoteSocket(NULL)
		, m_ZkProtocol(NULL)
		, m_bOpenedRemoteSocket(false)
		, m_listSockets(NULL)
        , m_bFailedOpenRemoteSocket(false)
		, m_nProtocol(in_WhichProtocol)
{
	memset(&m_RemoteAddress, 0, sizeof(ZkSocketAddress));

	ZKVERIFY( zkSocketNew(m_ZkContext.GetZkError(), AF_INET, SOCK_STREAM, IPPROTO_TCP, &m_RemoteSocket) );

// bug 6711
#ifndef _WIN32
	ZKVERIFY( zkSocketSetOption(m_ZkContext.GetZkError(), m_RemoteSocket, ZKSOCKET_OPTION_SendTimeout, 15 * 1000) );
	ZKVERIFY( zkSocketSetOption(m_ZkContext.GetZkError(), m_RemoteSocket, ZKSOCKET_OPTION_RecvTimeout, 45 * 1000) );
#endif

	switch(in_WhichProtocol)
	{

	case HTTP_PROTOCOL:
		m_ZkProtocol = new RZkHttpProtocol();
		break;
	
	case SMTP_PROTOCOL:
		m_ZkProtocol = new RZkSmtpProtocol();
		break;

	case NNTP_PROTOCOL:
		m_ZkProtocol = new RZkNntpProtocol();
		break;

	case SSL_PROTOCOL:
		m_ZkProtocol = new RZkPassThruProtocol();
		break;

	case TELNET_PROTOCOL:
		m_ZkProtocol = new RZkPassThruProtocol();
		break;

	case IRC_PROTOCOL:
		m_ZkProtocol = new RZkIrcProtocol();
		break;

	default:
		ZKASSERT(0);
		break;
	}
	m_ZkProtocol->InitProtocolResources(m_ZkContext);
}

RZkProxy::~RZkProxy()
{
	ZKVERIFY( zkSocketFree(m_ZkContext.GetZkError(), &m_RemoteSocket) );
	m_ZkProtocol->FreeProtocolResources(m_ZkContext);
	delete m_ZkProtocol;
}

bool RZkProxy::RunProxy()
{
	struct timeval timeout = {0};
	timeout.tv_sec = 45;
    
	bool ReturnValue = true;

	if( !ReceiveRemoteAddress() )
	{
		ReturnValue = false;
	}
	else if( !zkListNew(m_ZkContext.GetZkError(), &m_listSockets) )
	{
		ZKASSERT(0);
		ReturnValue = false;
	}
	else if(!zkListSetDataFreeFunction(m_ZkContext.GetZkError(),
								m_listSockets, NULL, NULL))
	{
		ZKASSERT(0);
		ReturnValue = false;
	}
	else if(!zkListInsert(m_ZkContext.GetZkError(), m_listSockets,
								m_LocalSocket, LIST_TAIL))
	{
		ZKASSERT(0);
		ReturnValue = false;
	}
	// In ClientReady(), proxy can either send a welcome banner to the client,
	// or open the connection to the server
	else if( !m_ZkProtocol->ClientReady(m_ZkContext) )
	{
		ZKASSERT(0);
		ReturnValue = false;
	}
	// If requested, SendFreedomizedData will open the connection to the server
	// and add the ZkSocket to our list
	// This will call ConnectToRemoteServer(), which will initialize m_RemoteSocket
	// IFF the m_ZkContext is set to ZKPROXY_CONNECT_TO_SERVER
	else if( !SendFreedomizedData() )
	{
		// Connect could have failed for any reason - don't assert
		// and don't return an error.
	}
	else
	{
		while(ReturnValue)
		{
				
			int iSocket = -1;

			// Last action was a failed connect
            // give the protocol a chance to act on this...
			if(m_bFailedOpenRemoteSocket)
            {
                m_bFailedOpenRemoteSocket = false;

                if(!m_ZkProtocol->ConnectFailed(m_ZkContext))
				{
					ReturnValue = false;
					ZKASSERT(0);
					break;
				}
				else if(!SendFreedomizedData())
				{
					break;
				}
            }
			// Wait for data to be available. The index of the readable socket
			// will be placed in iSocket. (index into the ZkList)
            else if(!zkSocketWaitReadableOrClosed(m_ZkContext.GetZkError(),
				m_listSockets, &timeout, &iSocket, RZkProxy::GetSignalFD()))
			{	
#ifdef WIN32
				int lastError = WSAGetLastError();
				ReturnValue = (WSAEINTR == lastError || WSAENOTSOCK == lastError);
#else
				ReturnValue = (EBADF == errno);
#endif

				ZKASSERT(ReturnValue);

				break;
			}

			// The WaitReadableOrClosed timed out
			else if(iSocket == -1){
				// The connection is still good, just run through the loop again.

			} 
            else if (iSocket == -2) {
                // Shutdown was signalled.
                ReturnValue = false;
                break;
            }
			// Local socket (because we inserted that socket into the ZKList first)
			else if(iSocket == 0)
			{
				if(!RecvRealRequest())
				{
					break;
				}
				else if(!m_ZkProtocol->FreedomizeClientRequest(m_ZkContext))
				{
					ReturnValue = false;
					ZKASSERT(0);
					break;
				}
				else if(!SendFreedomizedData())
				{
					break;
				}
			}
			
			// Remote socket (because we inserted that socket second)
			else if(iSocket == 1)
			{
				if(!RecvRealResponse())
				{
					break;
				}
				else if( !m_ZkProtocol->FreedomizeServerResponse(m_ZkContext) )
				{
					ReturnValue = false;
					ZKASSERT(0);
					break;
				}
				else if( !SendFreedomizedData() )
				{
					break;
				}
			}
			
			if(m_ZkContext.GetCloseConnection())
			{
				break;
			}
			m_ZkContext.ResetContext();
		}  // while
	}

	if(m_listSockets)
	{
		ZKVERIFY(zkListFree(m_ZkContext.GetZkError(), &m_listSockets) );
	}
	return ReturnValue;
}

bool RZkProxy::ReceiveRemoteAddress()
{
	bool ReturnValue = true;
#ifdef WIN32
	char szBuffer[sizeof( sockaddr )];

	// Recv the real sockaddr
	if( 1 != zkSocketReadAll(	m_ZkContext.GetZkError(),
								m_LocalSocket, szBuffer, sizeof (sockaddr) ) )
	{
		ReturnValue = false;
	}
	else 
	{		
		if( szBuffer[0] == AF_INET )
		{
			memcpy( &m_RemoteAddress, szBuffer, sizeof (sockaddr) );
			m_ZkContext.SetRemoteAddress(m_RemoteAddress);

#ifndef NDEBUG
			// see enum in izkproxy.h
			char *buf[] = { "http", "smtp", "nntp", "ssl", "telnet", "irc" };

			ZKTRACE2("RZkProxy::ReceiveRemoteAddress() ==> [%s] for [%s]\n", 
							inet_ntoa(m_ZkContext.GetRemoteAddress().m_inet.sa.sin_addr) 
							, buf[m_nProtocol]
							);
#endif
		}
		else 
		{
			ReturnValue = false;
			ZKASSERT( false );
		}
	}
#else
    try {
        RShimInterface::getShimInterface().VerifySocketConnection(m_LocalSocket, m_RemoteAddress.m_inet.sa.sin_addr.s_addr,
                                                                  m_RemoteAddress.m_inet.sa.sin_port);
        m_RemoteAddress.m_inet.sa.sin_family = AF_INET;
        m_ZkContext.SetRemoteAddress(m_RemoteAddress);        
    } catch (...) {
        // VerifySocketConnection... throws to indicate errors
        return false;
    }
#endif
	return ReturnValue;
}

bool RZkProxy::ConnectToRemoteServer()
{
	bool ReturnValue = true;

    // assume failure...
    m_bFailedOpenRemoteSocket = true;

	ZKASSERT(m_bOpenedRemoteSocket == false);
	if( !(zkSocketConnectAddr(m_ZkContext.GetZkError(),
								m_RemoteSocket,
								m_ZkContext.GetRemoteAddress() ) ))
	{
        // This is not a real failure... m_bFailedOpenRemoteSocket is set to true,
        // but don't return false.
	}
	else if(!zkListInsert(m_ZkContext.GetZkError(), m_listSockets,
									m_RemoteSocket, LIST_TAIL))
	{
		ZKASSERT(0);
		ReturnValue = false;
	}
	else
	{
        m_bFailedOpenRemoteSocket = false;
		m_bOpenedRemoteSocket = true;
	}
	return ReturnValue;
}

bool RZkProxy::RecvRealRequest()
{
	bool ReturnValue = true;

	int iBytesReceived = 0;

	if( !zkSocketRead(
			m_ZkContext.GetZkError(), 
			m_LocalSocket, 
			m_ZkContext.ByteBuffer(), 
			m_ZkContext.ByteBufferSize(), 
			(unsigned int*)&iBytesReceived))
	{
		ReturnValue = false;
	}
	else
	{
		m_ZkContext.SetByteBufferLength(iBytesReceived);
	}

	return ReturnValue;
}

bool RZkProxy::RecvRealResponse()
{
	bool ReturnValue = true;

	int iBytesReceived = 0;

	if( !zkSocketRead(
			m_ZkContext.GetZkError(), 
			m_RemoteSocket, 
			m_ZkContext.ByteBuffer(), 
			m_ZkContext.ByteBufferSize(), 
			(unsigned int*)&iBytesReceived))
	{
		ReturnValue = false;
	}
	else
	{
		m_ZkContext.SetByteBufferLength(iBytesReceived);
	}

	return ReturnValue;
}

bool RZkProxy::SendFreedomizedData()
{
	bool ReturnValue = true;

	switch (m_ZkContext.GetProxyAction())
	{
	case IZkContext::ZKPROXY_DO_NOTHING:
		break;

	case IZkContext::ZKPROXY_SEND_TO_CLIENT:
		ReturnValue = SendFreedomizedResponse();
		break;

	case IZkContext::ZKPROXY_SEND_TO_SERVER:
		ReturnValue = SendFreedomizedRequest();
		break;

	case IZkContext::ZKPROXY_CONNECT_TO_SERVER:	
		
		if(!ConnectToRemoteServer())
		{
			ReturnValue = false;
		}
		break;

	default:
		ReturnValue = false;
		ZKASSERT(0);
		break;

	}
	return ReturnValue;
}

bool RZkProxy::SendFreedomizedResponse()
{
	bool ReturnValue = true;

	if(!zkSocketWriteAll(m_ZkContext.GetZkError(), m_LocalSocket,
			m_ZkContext.GetProtocolOutput(),
			m_ZkContext.GetProtocolOutputLength() ))
	{
		ReturnValue = false;
	}

	return ReturnValue;
}

bool RZkProxy::SendFreedomizedRequest()
{
	bool ReturnValue = true;

	if(!m_bOpenedRemoteSocket)
	{
		ReturnValue = ConnectToRemoteServer();
	}
	if(ReturnValue)
	{
		if(!zkSocketWriteAll(m_ZkContext.GetZkError(), m_RemoteSocket,
				m_ZkContext.GetProtocolOutput(),
				m_ZkContext.GetProtocolOutputLength() ))
		{
			ReturnValue = false;
			ZKASSERT(false);
		}
	}

	return ReturnValue;
}

int RZkProxy::GetSignalFD()
{
	return m_signalFD;
}

void RZkProxy::SetSignalFD(int fd)
{
	m_signalFD = fd;
}
