// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#pragma warning (disable:4251)

#include "RZkSmtpProtocol.h"
#include "ZkEmailHelp.h"
#include <ZkService.h>
#include <ZkProxyDialogs.h>
#include <INymCentral.h>
#include <IClientCentral.h>
#include <IClientPersist.h>
#include <INym.h>
#include <string_hlp.h>
#include <frMaip.h>
#include <IZkConsoleMsg.h>
#include <IClientInfo.h>
#include <ci_string.h>
#include <string_hlp.h>
#include <ZkNetHelp.h>
#include <INymMail.h>
#include <IWordScan.h>
#include <zkCStr.h>
#include <IZkPopup.h>
#include <ZkStringMgr.h>

#ifdef WIN32
class CWnd;

// Marcp: this need to be fixed
#include "../win/ZKBaseUI.h"
#endif


// Mail limit support
#define MAIL_DAY_LIMIT 250
#define SECONDS_PER_DAY 86400
#define MAIL_SENT_LAG 300

#define DATA_CMD_STRING "DATA\r\n"
#define RSET_CMD_STRING "RSET\r\n"
#define QUIT_CMD_STRING "QUIT\r\n"
#define HELO_CMD_STRING	"HELO computer\r\n"
#define MAIL_CMD_FORMAT_STRING "MAIL FROM:<%s>\r\n"
#define RCPT_CMD_FORMAT_STRING "RCPT TO:<%s>\r\n"

#define PGP_SIGNED_MSG_HEADER "-----BEGIN PGP SIGNED MESSAGE-----"
#define FREEDOM_MAIL_TIMEOUT 240

#define WELCOME_BANNER_RESPONSE			"220 SMTP localhost, Freedom Baby\r\n"
#define UNSUPPORTED_CMD_RESPONSE		"502 Command not implemented\r\n"
#define HELO_RESPONSE					"250 Freedom Mail Server ready\r\n"
#define NOOP_RESPONSE					"250 OK\r\n"
#define RSET_RESPONSE					"220 Service ready\r\n"
#define QUIT_RESPONSE					"221 Service closing transmission channel\r\n"
#define USER_CANCELLED_RESPONSE			"554 User cancelled send operation\r\n"
#define MAIL_ACCEPTED_RESPONSE			"250 Mail accepted for delivery\r\n"
#define MAIL_FAILED_RESPONSE			"554 Hi, Freedom Mail system failed\r\n"
#define BAD_SEQUENCE_RESPONSE			"503 Bad sequence of commands\r\n"
#define BAD_SYNTAX_RESPONSE				"500 Hi, Syntax error, command unrecognized\r\n"
#define RCPT_RESPONSE					"250 Requested mail action okay, completed\r\n"
#define DATA_RESPONSE					"354 Start mail input; end with <CRLF>.<CRLF>\r\n"
#define MAIL_RESPONSE					"250 Requested mail action okay, completed\r\n"

#define SIMPLE_ERROR_STRING					"554 Freedom experienced a problem sending email. Email not sent.\r\n"
#define ERROR_DETAIL_FORMAT_STRING			"554 Freedom experienced a problem sending email. Email not sent (%s).\r\n"
#define ERROR_DETAIL_AND_FUNC_FORMAT_STRING	"554 Freedom experienced a problem sending email. Email not sent (%s in %s).\r\n"
#define ERROR_MAIL_SENT_COUNT		"554 Freedom experienced a problem sending email. Mail sent limit exceeded\r\n" 

extern "C" {
  int SMTPWriteBuf ( ZkError , void * , const unsigned char* ,int );
}	
int SMTPWriteBuf ( 
    ZkError err ,
    void * cdata ,
    const unsigned char * data ,
    int dlen )
{
    zkErrorClear(err);
    if (cdata && data && (dlen > 0)) {
		zkMemBufAppend(err, (ZkMemBuf)cdata, (void*)data, dlen);
	}
    return 1;
}

RZkSmtpProtocol::RZkSmtpProtocol()
				: m_Email(NULL)
				, m_MemBuf(NULL)
				, m_bAlreadyConnected(false)
				, m_State(FRESH_STATE)
				, m_bLinuxRelayHack(false)
{

}

bool RZkSmtpProtocol::InitProtocolResources (IZkContext& in_Context)
{
	bool ReturnValue = true;
	if( !zkEmailNew(in_Context.GetZkError(), &m_Email) )
	{
		ReturnValue = false;
		ZKASSERT(false);
	}
	else if( !zkEmailScanForSMTPTermination(in_Context.GetZkError(), m_Email) )
	{
		ReturnValue = false;
		ZKASSERT(false);
	}

	return ReturnValue;
}

bool RZkSmtpProtocol::FreeProtocolResources (IZkContext& in_Context)
{
	bool ReturnValue = true;
	if( !zkEmailFree(in_Context.GetZkError(), &m_Email) )
	{
		ReturnValue = false;
		ZKASSERT(false);
	}
	if( m_MemBuf )
	{
		if(!zkMemBufFree(in_Context.GetZkError(), &m_MemBuf) )
		{
			ReturnValue = false;
			ZKASSERT(false);
		}
	}
	return ReturnValue;
}

bool RZkSmtpProtocol::ConnectFailed(IZkContext& in_Context)
{
	// see bug 5827 
#ifdef LINUX
	if(m_bLinuxRelayHack)
	{
		ZKSTD::string strMailRelay;
		ZKSTD::string strMsg;

		ZkService::GetIClientCentral()->GetMailRelay(strMailRelay);
		ZkStringMgr::Load(strMsg, IDS_MAIL_RELAY_INVALID, strMailRelay.c_str());

		IZkPopup::MessageBox(strMsg);
	}
#endif

    in_Context.SetCloseConnection(true);
    return true;
}


bool RZkSmtpProtocol::ClientReady(IZkContext& in_Context)
{
	SetResponse(in_Context, WELCOME_BANNER_RESPONSE);
	TraceOutgoingStream(in_Context);
	return true;
}

bool RZkSmtpProtocol::FreedomizeClientRequest (IZkContext& in_Context)
{
	ZKTRACE1("SMTP Client Says: %s\r\n", (char*)in_Context.ByteBuffer());
	in_Context.SetProxyAction(IZkContext::ZKPROXY_DO_NOTHING);

	if(m_State == RECEIVING_DATA_STATE)
	{
		HandleMoreData( in_Context );
	}
	else
	{
		m_strLine += ZKSTD::string((char*)in_Context.ByteBuffer(),
			in_Context.GetByteBufferLength() );
		if(m_strLine.find("\r\n") != m_strLine.npos)
		{
			int iCommand = ParseCommand( in_Context );
			switch( iCommand )
			{
			// Unsupported Commands
			case EHLO_CMD:
			case VRFY_CMD:
			case HELP_CMD:
			case EXPN_CMD:
			case OTHER_CMD:
				SetResponse(in_Context, UNSUPPORTED_CMD_RESPONSE);
				break;

			// HELO and NOOP are supported, but do not change our state
			case HELO_CMD:
				SetResponse(in_Context, HELO_RESPONSE);
				break;

			case NOOP_CMD:
				SetResponse(in_Context, NOOP_RESPONSE);
				break;

			// RSET will revert our state to that of a fresh connection
			case RSET_CMD:				// RSET
				if(!ResetState(in_Context))
				{
					in_Context.SetCloseConnection(true);
					ErrorResponse(in_Context);
				}
				else
				{
					SetResponse(in_Context, RSET_RESPONSE);
				}
				break;

			// QUIT will terminate the connection
			case QUIT_CMD:				// QUIT
				SendQuitCommandToServer(in_Context);
				break;

			case MAIL_CMD:				// MAIL FROM:<sender>
				HandleMailCommand( in_Context );
				break;

			case RCPT_CMD:				// RCPT TO:<recipient>
				HandleRcptCommand( in_Context );
				break;

			case DATA_CMD:				// DATA
				HandleDataCommand( in_Context );
				break;
				
			default:
				ErrorResponse(in_Context);
				ZKASSERT(0);
				break;
			}
			m_strLine = "";
		}
	}

	TraceOutgoingStream(in_Context);
	return true;
}

bool RZkSmtpProtocol::FreedomizeServerResponse	(IZkContext& in_Context)
{
	ZKTRACE1("SMTP Server Says: %s\r\n", (char*)in_Context.ByteBuffer());
	in_Context.SetProxyAction(IZkContext::ZKPROXY_DO_NOTHING);

	if( ((char*)in_Context.ByteBuffer())[0] == '5' )
	{
		// Could send back the server error message to the client...
		SetResponse(in_Context, MAIL_FAILED_RESPONSE);
		ZKVERIFY(ResetState(in_Context));
	}
	else
	{
		switch(m_State)
		{
		case CONNECTING_TO_SERVER:
			SendHeloCommandToServer(in_Context);
			break;

		case SENT_HELO_CMD:
			SendMailCommandToServer(in_Context);
			break;

		case SENT_MAIL_FROM_CMD:
		case SENT_RCPT_TO_CMD:
			// SendRecipient Command will send one recipient name,
			// then remove it from the vector. When all the names
			// have been sent, the vector will be empty.
			if(m_vecRecipients.size())
			{
				SendRcptCommandToServer(in_Context);
			}
			else
			{
				SendDataCommandToServer(in_Context);
			}
			break;

		case SENT_DATA_CMD:
			SendMailBodyToServer(in_Context);
			break;

		case SENT_MAIL_BODY:
			SendRsetCommandToServer(in_Context);
			break;

		case SENT_RSET_CMD:
			SetResponse(in_Context, MAIL_ACCEPTED_RESPONSE);
			ZKVERIFY(ResetState(in_Context));
			break;
			
		case SENT_QUIT_CMD:
			SetResponse(in_Context, QUIT_RESPONSE);
			in_Context.SetCloseConnection(true);
			break;

		default:
			ErrorResponse(in_Context);
			ZKASSERT(0);
			break;
		}
	}

	TraceOutgoingStream(in_Context);
	return true;
}

void RZkSmtpProtocol::SetResponse(IZkContext& in_Context, const char* strResponse)
{
	in_Context.SetProtocolOutput((void *)strResponse);
	in_Context.SetProtocolOutputLength(strlen(strResponse));

	in_Context.SetProxyAction(IZkContext::ZKPROXY_SEND_TO_CLIENT);
}

void RZkSmtpProtocol::ErrorResponse (IZkContext& in_Context)
{
	in_Context.SetProxyAction( IZkContext::ZKPROXY_SEND_TO_CLIENT);

	unsigned int lenDetail = strlen(in_Context.GetZkError()->getdetail().c_str() );
	unsigned int lenFunc = strlen( in_Context.GetZkError()->getfnc().c_str() );

	if ( lenDetail > 0 )
	{
		if ( lenFunc > 0 )
		{
			// Function name is a bit ugly to show to the user, we only show
			// this if we have one from the unix guys
			snprintf(	(char *)in_Context.ByteBuffer(), 
						in_Context.ByteBufferSize(), 
						ERROR_DETAIL_AND_FUNC_FORMAT_STRING,
						in_Context.GetZkError()->getdetail().c_str(), 
						in_Context.GetZkError()->getfnc().c_str() );
		}
		else
		{
			
			snprintf((char *)in_Context.ByteBuffer(), in_Context.ByteBufferSize(), ERROR_DETAIL_FORMAT_STRING,
												in_Context.GetZkError()->getdetail().c_str());
		}
	}
	else
	{
		zkCStrlcpy((char *)in_Context.ByteBuffer(), SIMPLE_ERROR_STRING, in_Context.ByteBufferSize());
	}

	in_Context.SetProtocolOutput(in_Context.ByteBuffer());
	in_Context.SetProtocolOutputLength(strlen((char*)in_Context.ByteBuffer()));

	ZKVERIFY(ResetState(in_Context));
}

int RZkSmtpProtocol::ParseCommand(IZkContext& in_Context)
	// Match the request to a cmd
{
	int iCommand = OTHER_CMD;

	if(m_strLine[0] == 'H' || m_strLine[0] == 'h'){
		
		if(!(strnicmp(m_strLine.c_str(), "HELO", 4))){
			iCommand = HELO_CMD;
		}

		else if(!(strnicmp(m_strLine.c_str(), "HELP", 4))){
			iCommand = HELP_CMD;
		}

		else {
			iCommand = OTHER_CMD;
		}
	}

	else if(m_strLine[0] == 'E' || m_strLine[0] == 'e'){
		
		if(!(strnicmp(m_strLine.c_str(), "EHLO", 4))){
			iCommand = EHLO_CMD;
		}

		else if(!(strnicmp(m_strLine.c_str(), "EXPN", 4))){
			iCommand = EXPN_CMD;
		}

		else {
			iCommand = OTHER_CMD;
		}
	}

	else if(m_strLine[0] == 'M' || m_strLine[0] == 'm'){
	
		if(!(strnicmp(m_strLine.c_str(), "MAIL", 4))){
			iCommand = MAIL_CMD;
		}

		else {
			iCommand = OTHER_CMD;
		}
	}

	else if(m_strLine[0] == 'R' || m_strLine[0] == 'r'){
	
		if(!(strnicmp(m_strLine.c_str(), "RCPT", 4))){
			iCommand = RCPT_CMD;
		}

		else if(!(strnicmp(m_strLine.c_str(), "RSET", 4))){
			iCommand = RSET_CMD;
		}

		else {
			iCommand = OTHER_CMD;
		}
	}

	else if(m_strLine[0] == 'D' || m_strLine[0] == 'd'){
	
		if(!(strnicmp(m_strLine.c_str(), "DATA", 4))){
			iCommand = DATA_CMD;
		}

		else {
			iCommand = OTHER_CMD;
		}
	}

	else if(m_strLine[0] == 'V' || m_strLine[0] == 'v'){
	
		if(!(strnicmp(m_strLine.c_str(), "VRFY", 4))){
			iCommand = VRFY_CMD;
		}

		else {
			iCommand = OTHER_CMD;
		}
	}

	else if(m_strLine[0] == 'Q' || m_strLine[0] == 'q'){
	
		if(!(strnicmp(m_strLine.c_str(), "QUIT", 4))){
			iCommand = QUIT_CMD;
		}

		else {
			iCommand = OTHER_CMD;
		}
	}

	else if(m_strLine[0] == 'N' || m_strLine[0] == 'n'){
	
		if(!(strnicmp(m_strLine.c_str(), "NOOP", 4))){
			iCommand = NOOP_CMD;
		}

		else {
			iCommand = OTHER_CMD;
		}
	}

	else {
		iCommand = OTHER_CMD;
	}

	return iCommand;
}

bool RZkSmtpProtocol::ResetState(IZkContext& in_Context)
{
	bool ReturnValue = true;

	m_State = FRESH_STATE;
	if(!zkEmailFree(in_Context.GetZkError(), &m_Email))
	{
		ReturnValue = false;
		ZKASSERT(false);
	}
	else if( !zkEmailNew(in_Context.GetZkError(), &m_Email) )
	{
		ReturnValue = false;
		ZKASSERT(false);
	}
	else if( !zkEmailScanForSMTPTermination(in_Context.GetZkError(), m_Email) )
	{
		ReturnValue = false;
		ZKASSERT(false);
	}

	if( m_MemBuf )
	{
		if(!zkMemBufFree(in_Context.GetZkError(), &m_MemBuf) )
		{
			ReturnValue = false;
			ZKASSERT(false);
		}
		m_MemBuf = NULL;
	}

	m_vecRecipients.erase( m_vecRecipients.begin(), m_vecRecipients.end() );
	m_strMailFromArg = "";

	return ReturnValue;
}

void RZkSmtpProtocol::HandleMailCommand(IZkContext& in_Context)
{
	if(m_State != FRESH_STATE)
	{
		SetResponse(in_Context, BAD_SEQUENCE_RESPONSE);
	}
	// Stores argument into m_strMailFromArg
	else if(!GetSmtpCommandArguments(in_Context))
	{
		// Command was invalid
		SetResponse(in_Context, BAD_SYNTAX_RESPONSE);
	}
	else 
	{
		SetResponse(in_Context, MAIL_RESPONSE);
		m_State = RECEIVED_MAIL_CMD_STATE ;
	}

}

void RZkSmtpProtocol::HandleRcptCommand(IZkContext& in_Context)
{
	if(m_State != RECEIVED_MAIL_CMD_STATE && m_State != RECEIVED_RCPT_CMD_STATE )
	{
		SetResponse(in_Context, BAD_SEQUENCE_RESPONSE);
	}
	// stores argument into m_vecRecipients
	else if( !GetSmtpCommandArguments(in_Context) )
	{
		// Command was invalid
		SetResponse(in_Context, BAD_SYNTAX_RESPONSE);
	}
	else
	{
		SetResponse(in_Context, RCPT_RESPONSE);
		m_State = RECEIVED_RCPT_CMD_STATE;
	}
}

void RZkSmtpProtocol::HandleDataCommand(IZkContext& in_Context)
{
	if(m_State != RECEIVED_RCPT_CMD_STATE)
	{
		SetResponse(in_Context, BAD_SEQUENCE_RESPONSE);
	}
	else
	{
		SetResponse(in_Context, DATA_RESPONSE);
		m_State = RECEIVING_DATA_STATE;
	}
}

void RZkSmtpProtocol::HandleMoreData(IZkContext& in_Context)
{
	if( !m_MemBuf )
	{
		if(!zkMemBufNew(in_Context.GetZkError(), &m_MemBuf))
		{
			ErrorResponse(in_Context);
		}
	}

	if( m_MemBuf )
	{
		if( !zkMemBufAppend(in_Context.GetZkError(), m_MemBuf,
			in_Context.ByteBuffer(), in_Context.GetByteBufferLength()) )
		{
			// Error response returns the ZkError->detail and resets our state
			ErrorResponse(in_Context);
		}
		else
		{
			size_t len = 0;
			void *pData = NULL;
			ZKVERIFY(zkMemBufGetData(in_Context.GetZkError(), m_MemBuf, &pData, &len));
			if(len > 5)
			{
				ZKSTD::string strBufEnd((char*)pData + len - 5, 5);
				if( strBufEnd == "\r\n.\r\n" )
				{
					HandleCompleteMail(in_Context);
				}
			}
		}
	}
}

bool RZkSmtpProtocol::MsgIsReadReceipt(IZkContext &in_Context)
{

	char *pEmailData = NULL;
	unsigned int lenEmailData = 0;
	
	ZKVERIFY(zkMemBufGetData(in_Context.GetZkError(), m_MemBuf,
								(void**)&pEmailData, &lenEmailData));

	ZKSTD::string strContentType;
	if(ZKS::extractHeaderField(pEmailData, lenEmailData, "Content-Type", strContentType))
	{
		ZKS::MakeLower(strContentType);
		size_t i = strContentType.find("report-type");
		if(i != strContentType.npos)
		{
			i += strlen("report-type");

			size_t j = strContentType.find('=', i);
			if(j != strContentType.npos && j == strContentType.find_first_not_of(" \t", i))
			{
				i = strContentType.find_first_not_of(" \t", j + 1);
				if(i != strContentType.npos)
				{
					j = strContentType.find_first_of("; \t\r\n", i);
					size_t len = j - i;
					if(j == strContentType.npos)
					{
						len = strContentType.length() - i;
					}
					ZKSTD::string strReportType = strContentType.substr(i, len);
					if(strReportType == "disposition-notification")
					{
						return true;
					}
				}
			}
		}
	}

	return false;

}

void RZkSmtpProtocol::HandleCompleteMail(IZkContext& in_Context)
{
	long lResult = -1;
	if(MsgIsReadReceipt(in_Context))
	{
		if(ZkNetHelp::IsRemoteHostDummyMailServer(in_Context.GetRemoteAddress()))
		{
			lResult = -1;
		}
		else
		{
			lResult = USER_NYM_INDEX;
		}
	}
	else if(ZkNetHelp::IsRemoteHostDummyMailServer(in_Context.GetRemoteAddress()))
	{
		lResult = ZkProxyDialogs::LaunchOutgoingMsgDlg( &m_vecRecipients,
                                                         true);
	}
	else
	{
		lResult = ZkProxyDialogs::LaunchOutgoingMsgDlg( &m_vecRecipients,
                                                         true,
														 &m_strMailFromArg);
	}

	if(-1 == lResult)
	{
		// block this message, for now the message will remain in the
		// users outbox (if we want we can eat the message at the 
		// proxy level).
		SetResponse(in_Context, USER_CANCELLED_RESPONSE);
		ZKVERIFY(ResetState(in_Context));
	}
	else if ( lResult == USER_NYM_INDEX )
	{
		SendMailAsUser(in_Context);
	}
	else 
	{
		SendMailAsNym(in_Context, lResult);
	}
}

void RZkSmtpProtocol::NotifyConsoleOfMail(const size_t in_iNym)
{
	ZKSTD::string strRecipients;
	for(size_t i = 0; i < m_vecRecipients.size() && i < 3; i++)
	{
		if(i > 0)
		{
			if(m_vecRecipients.size() > 2)
			{
				strRecipients += ", ";
			}
			else
			{
				strRecipients += ' ';
			}
		}
		if(i + 1 == m_vecRecipients.size() && m_vecRecipients.size() > 1)
		{
			strRecipients += "and ";
		}
		strRecipients += m_vecRecipients[i];
		if(i == 2 && m_vecRecipients.size() > 3)
		{
			strRecipients += "...";
		}
	}
	ZKSTD::string strNymName;
	ZkService::GetINymCentral()->GetNymName(in_iNym, strNymName);
	IZkConsoleMsg::ShowImportantMsg(CONSIMAGE_MAIL, IDS_CONS_MAIL_FROM_NYMNAME_TO_RECIPIENTS, strNymName.c_str(), strRecipients.c_str());
}

bool RZkSmtpProtocol::GetSmtpCommandArguments(IZkContext& in_Context)
	// Called for multiple part commands
{
	bool ReturnValue = true;
	int iCommand = ParseCommand( in_Context );


	ZKSTD::string::size_type pos1 = m_strLine.find_first_not_of(' ', 0);
	ZKSTD::string::size_type pos2 = m_strLine.find_first_of(' ', pos1);
	pos1 = m_strLine.find_first_not_of(' ', pos2);

	if(pos1 == m_strLine.npos)
	{// all spaces following first command
		ReturnValue = false;
	}
	else
	{
		// find the first space _OR_ colon
		ZKS::ci_string strModifier;
		pos2 = m_strLine.find_first_of(" :", pos1);
	
		if(pos2 == m_strLine.npos)
		{//no colon or space
			ReturnValue = false;
		}
		else
		{
			strModifier = m_strLine.substr(pos1, pos2 - pos1).c_str();
			pos2 = m_strLine.find_first_of(':', pos1);
			
			if(pos2 == m_strLine.npos)
			{//no colon
				ReturnValue = false;
			}
		}

		if(ReturnValue)
		{
			// pos2 is the colon
			// check section between : and <
			ZKSTD::string::size_type pos3 = m_strLine.find_first_not_of(' ', (pos2+1)); //pos 2 is on the colon
			if(pos3 == m_strLine.npos)
			{ // all spaces
				ReturnValue = false;
			}

			else if(m_strLine[pos3] != '<')
			{
				ReturnValue = false;
			}

			else
			{
				// get section between < and >
				pos2 = m_strLine.find_first_of('>', pos3);
				if (pos2 == m_strLine.npos)
				{
					ReturnValue = false;
				}

				else if(iCommand == MAIL_CMD)
				{
					// require this arg if user decides to send via smtp
					if(strModifier == "FROM")
					{
						m_strMailFromArg = m_strLine.substr(pos3+1, (pos2-(pos3+1)));
					}
					else
					{
						ReturnValue = false;
					}
					//
				}
				
				else if (iCommand == RCPT_CMD)
				{ // ... several RCPT TO's
					if(strModifier == "TO")
					{
						m_vecRecipients.push_back(m_strLine.substr(pos3+1, (pos2-(pos3+1))));
					}
					else
					{
						ReturnValue = false;
					}
					
				}
			}

			// make sure there is no garbage following '>'
			pos3 = m_strLine.find_first_not_of(' ', (pos2+1));	// anything after '>' that is not a space is an error
			if( pos3 == m_strLine.npos || m_strLine.at(pos3) != '\r')
			{// if not CR, then junk
			ReturnValue = false;
			}
		}
	}

	return ReturnValue;
}

bool RZkSmtpProtocol::ScanEmailForOffensiveWords(IZkContext& in_Context, size_t in_NymIndex)
{
	bool ReturnValue = true;

	unsigned char *pEmailData = NULL;
	unsigned int lenEmailData = 0;
	ZKSTD::vector<ZKSTD::string> vecstrOffensiveWordsFound;

	if(!zkMemBufGetData(in_Context.GetZkError(), 
		m_MemBuf, (void**)&pEmailData, &lenEmailData))
	{
		// If we are unable to wordscan the email, block it from being
		// sent...
		ReturnValue = false;
		ZKASSERT(false);
	}
	else if(ZkService::GetIWordScan()->WordScan(pEmailData, lenEmailData,
										vecstrOffensiveWordsFound, in_NymIndex) )
	{ 			
		
		// Launch an Word Scan Alert Dialog or other UI reaction for the user

		WordScanSummary Result ;
		Result.pScannedBuffer = pEmailData;
		Result.cntBuffer = lenEmailData;
		Result.pFoundWords = &vecstrOffensiveWordsFound ;
		// This socket address is for the user's SMTP server
		// The message is actually going to the FMG.
		//Result.m_SockAddr = GetOriginalDestinationSocketAddress();

        if( !ZkProxyDialogs::LaunchWordScanMsgDlg( PROTOCOL_SMTP, &Result ) )
        {
            ReturnValue = false;
        }
	}
	return ReturnValue;
}

bool RZkSmtpProtocol::ChangeFromFieldToFreedom(IZkContext& in_Context, size_t in_NymIndex)
{
	bool ReturnValue = true;
	ZKSTD::string strNymAddress;
	ZKSTD::string     strNymDomain ;
	ZkEmailHeader	header;
	ZkEmailField	field;


	INym* pNym = NULL;
	ZKVERIFY(ZkService::GetINymCentral()->GetNym(IID_INYM, (void**)&pNym, in_NymIndex));
	ZKVERIFY(pNym->GetNymName(strNymAddress));

	ZKVERIFY(pNym->GetNymEmailDomain( strNymDomain ));

	strNymAddress.append("@");
	strNymAddress.append(strNymDomain);
	
	ZKVERIFY( zkEmailHeader(in_Context.GetZkError(), m_Email, &header) );
	
	if( !zkEmailHeaderGetField(in_Context.GetZkError(), header, From, &field) )
	{
		// Field doesn't exist... Create it.
		if(!zkEmailFieldNew(in_Context.GetZkError(), &field))
		{
			ZKASSERT(false);
			ReturnValue = false;
		}
		else if(!zkEmailFieldSetId(in_Context.GetZkError(), field, From))
		{
			ZKASSERT(false);
			ReturnValue = false;
		}
		else if( !zkEmailFieldSetValue(in_Context.GetZkError(), field, (const unsigned char*)strNymAddress.c_str() ))
		{
			ZKASSERT(false);
			ReturnValue = false;
		}
		else if( !zkEmailHeaderAddField(in_Context.GetZkError(), header,field))
		{
			ZKASSERT(false);
			ReturnValue = false;
		}
	}

	// Field was found - change it.
	else if( !zkEmailFieldSetValue(in_Context.GetZkError(), field, (const unsigned char*)strNymAddress.c_str() ))
	{ 
		ZKASSERT(false);
		ReturnValue = false;
	}

	return ReturnValue;
}

// return true if the user can send this mail
// without exceeding MAIL_DAY_LIMIT number of mails in the last 24 hours.
// handle the reset of the clock as well.

bool RZkSmtpProtocol::CheckMailCount(IZkContext & in_Context, size_t in_NymIndex ) {
	bool ReturnValue = true;

    INymMail * pNymMail = NULL;
    ZKVERIFY(ZkService::GetINymCentral()->GetNym(IID_IMAIL, (void**)&pNymMail, in_NymIndex));

    unsigned int actualMailSentCount = pNymMail->GetCurrentMailSentCount();
    time_t actualMailSentInitTime = pNymMail->GetMailSentInitTime();
    time_t actualTime; 
	time( & actualTime );
    int numAttemptMails = m_vecRecipients.size();

    // if their mailSentCount is 0, then reset their MailSentInitTime
    // to now. 
    // else if the last time the rollover time was set was more than 
    // a day ago, set the MailSentCount to 0, and set the rollover time to now

    if( actualMailSentCount == 0 )
	{
        pNymMail->SetMailSentInitTime( actualTime );
    }
	else  if( ( actualTime - actualMailSentInitTime ) > SECONDS_PER_DAY + MAIL_SENT_LAG )
	{
        pNymMail->SetCurrentMailSentCount( 0 );
        actualMailSentCount = 0;
        pNymMail->SetMailSentInitTime( actualTime );
    }
        
    // check the amount of mail they would have sent if this went through
    // and let them know if this is okay.
    if( ( actualMailSentCount + numAttemptMails ) <= MAIL_DAY_LIMIT )
	{
       ZKTRACE1( "** let the mail go through... this will bring them up to %d\n",
                 actualMailSentCount + numAttemptMails );
       ReturnValue = true;
    }
	else
	{
       // this dialog always returns false.
	   // the second argument is what gets displayed as the time
	   // when they can start sending mail again
       ZkProxyDialogs::LaunchMailSentCountDlg( 
                       actualMailSentInitTime + SECONDS_PER_DAY + MAIL_SENT_LAG,
                       & m_vecRecipients );
       ReturnValue = false;
    }

    return ReturnValue;
}

bool RZkSmtpProtocol::ScanForRealSignature(IZkContext& in_Context)
{
	ZkEmailField field = NULL;
	ZkEmailHeader header = NULL;
	const unsigned char *value = NULL;
	bool ReturnValue = true;
	bool bSigned = false;
	char *pEmailData = NULL;
	unsigned int lenEmailData = 0;

	if(!zkMemBufGetData(in_Context.GetZkError(), 
		m_MemBuf, (void**)&pEmailData, &lenEmailData))
	{
		// If we are unable to scan the email, block it from being
		// sent...
		ReturnValue = false;
		ZKASSERT(false);
	}

	else if(!zkEmailHeader(in_Context.GetZkError(), m_Email, &header))
	{
		ReturnValue = false;
		ZKASSERT(false);
	}
	else if(!zkEmailHeaderGetField(in_Context.GetZkError(), header,
										Content_Type, &field))
	{
	}
	else if(!zkEmailFieldValue(in_Context.GetZkError(), field, &value))
	{
	}
	else
	{
		ZKSTD::string strValue((char *)value);
		ZKS::MakeLower(strValue);
		if(strValue.find("signed") != ZKSTD::string::npos ||
			strstr(pEmailData, PGP_SIGNED_MSG_HEADER) != NULL)
		{
            ReturnValue = ZkProxyDialogs::LaunchWarnSignedMailDlg();
		}
	}
	return ReturnValue;
}

void RZkSmtpProtocol::SendMailAsNym(IZkContext& in_Context, size_t in_NymIndex )
{
	bool bSuccess = false;
	bool bUserCancelled = false;
	bool bMailSentCountExceeded = false;

	unsigned char *pEmailData = NULL;
	unsigned int lenEmailData = 0;
	
	ZKSTD::string	strNymAddress;
	ZKSTD::string	strNymDomain ;

	INym* pNym = NULL;
	ZKVERIFY(ZkService::GetINymCentral()->GetNym(IID_INYM, (void**)&pNym, in_NymIndex));
	ZKVERIFY(pNym->GetNymName(strNymAddress));

	ZKVERIFY(pNym->GetNymEmailDomain( strNymDomain ));

	strNymAddress.append("@");
	strNymAddress.append(strNymDomain);

	int iStatus = 0;

	if(!zkMemBufGetData(in_Context.GetZkError(), m_MemBuf, (void**)&pEmailData, &lenEmailData))
	{
		ZKASSERT(false);
	}
	else if( !zkEmailAddText(in_Context.GetZkError(), m_Email, 
				pEmailData, lenEmailData) )
	{
		ZKASSERT(false);
	}
	else if( ! zkMemBufFree(in_Context.GetZkError(), &m_MemBuf) )
	{
		ZKASSERT(false);
	}
	else if( !zkEmailStatus(in_Context.GetZkError(), m_Email, &iStatus) )
	{
		ZKASSERT(false);
	}
	else if( iStatus != PARSE_COMPLETE)
	{
		ZKASSERT(false);
	}
	else if( !zkEmailParse(in_Context.GetZkError(), m_Email) )
	{
		ZKASSERT(false);
	}

	// Bug # 1472 & 2365, Any field with the flag FL_DELETE will be deleted
	// (comment in zkemail.c) zkemailfield.c has the fields
	else if( !zkEmailStripHeader(in_Context.GetZkError(), m_Email, 0) )
	{ 
		ZKASSERT(false);
	}
	else if(!ChangeFromFieldToFreedom( in_Context, in_NymIndex))
	{
		ZKASSERT(false);
	}
	else if(!zkMemBufNew(in_Context.GetZkError(), &m_MemBuf ))
	{
		ZKASSERT(false);
	}
	else if(!zkEmailDump(in_Context.GetZkError(), m_Email,
						SMTPWriteBuf, m_MemBuf))
	{
		ZKASSERT(false);
	}
	else if(!zkMemBufAppend(in_Context.GetZkError(), m_MemBuf, "\0", 1))
	{
		ZKASSERT(false);
	}
        else if(!CheckMailCount(in_Context, in_NymIndex))
    	{
		bSuccess = false;
		bUserCancelled = false;
		bMailSentCountExceeded = true;
    	}
	else if(!ScanForRealSignature(in_Context))
	{
		// User signed mail with their
		// real signature.
		bSuccess = true;
		bUserCancelled = true;
	}
	else if(!ScanEmailForOffensiveWords(in_Context, in_NymIndex))
	{
		// User can choose to cancel if they see
		// information leaving their computer
		bSuccess = true;
		bUserCancelled = true;
	}
	else
	{
		ZKSTD::string strFreedomDomain;
		ZkService::GetIClientInfo()->GetFreedomDomain(strFreedomDomain);

		ZKSTD::vector<ZKSTD::string> vecInetRecipients;
		ZKSTD::vector<ZKSTD::string> vecFreedomRecipients;
		ZKSTD::vector<ZKSTD::string>::iterator it;

		for(it = m_vecRecipients.begin(); it != m_vecRecipients.end(); ++it)
		{
			// Make sure the recipient domain exists and isn't empty.
			size_t rcptDomainIndex = it->find("@");
			if(rcptDomainIndex == it->npos || rcptDomainIndex == it->length()-1)
			{
				ZKASSERT(!"Bogus recipient email address : skipping.");
				continue;
			}
			
			if(IsFreedomRecipient(strFreedomDomain, it->substr(rcptDomainIndex+1)))
			{
				vecFreedomRecipients.push_back(*it);
			}
			else
			{
				vecInetRecipients.push_back(*it);
			}
		}

		
		if(!zkMemBufGetData(in_Context.GetZkError(), 
				m_MemBuf, (void**)&pEmailData, &lenEmailData))
		{
			ZKASSERT(false);
		}
		else
		{
			NotifyConsoleOfMail(in_NymIndex);
			ZKASSERT(pEmailData[lenEmailData - 1] == 0);
			if(ZkEmailHelp::SendEmailESMTP(in_Context.GetZkError(), vecInetRecipients,
										vecFreedomRecipients, (char*)pEmailData, in_NymIndex))
			{
				bSuccess = true;
			}
			else
			{
				IZkConsoleMsg::ShowImportantMsg( CONSIMAGE_MAIL, IDS_CONS_JUST_STRING, in_Context.GetZkError()->getdetail().c_str() );
			}
		}
	}	

	if(bSuccess)
	{
		if(bUserCancelled)
		{
			SetResponse(in_Context, USER_CANCELLED_RESPONSE);
		}
		else
		{
			SetResponse(in_Context, MAIL_ACCEPTED_RESPONSE);

            // add on the mail limit stuff here
            // this is the only time we can guarantee that the mail
            // action is being mirrored on the server
            INymMail * pNymMail = NULL;
            ZKVERIFY(ZkService::GetINymCentral()->GetNym(IID_IMAIL, 
                    (void**)&pNymMail, in_NymIndex));
            pNymMail->SetCurrentMailSentCount( pNymMail->GetCurrentMailSentCount() 
                                               + m_vecRecipients.size());
		}
		ZKVERIFY(ResetState(in_Context));
	}
	else
	{
		if( bMailSentCountExceeded )
		{
			SetResponse( in_Context, ERROR_MAIL_SENT_COUNT );
        	}
		else
		{
			ErrorResponse(in_Context);
        	}
	}
}

// This function will return true if each segment of the sender's domain
// are the same as each _last_ segments of the recipient domain.
// The reason why we compare each segments instead of a plain string is
// because it makes it easier to trap:
// Sender: ...@freedom.net
// Rcpt: ...@notfreedom.net
// This function won't return true if can't find _all_ the sender's domain
// segment in the rcpt's domain, starting from the end.
bool RZkSmtpProtocol::IsFreedomRecipient(
	const ZKSTD::string& in_strSenderDomain, 
	const ZKSTD::string& in_strRcptDomain)
{
	ZKSTD::vector<ZKSTD::string> sendersDomainSections;
	ZKS::SplitOnCharacter(in_strSenderDomain, sendersDomainSections, '.');
	
	if(sendersDomainSections.empty())
	{
		ZKASSERT(!"Unable to split sender's domain on character.");
		return false;
	}

	ZKSTD::vector<ZKSTD::string> rcptsDomainSections;
	ZKS::SplitOnCharacter(in_strRcptDomain, rcptsDomainSections, '.');

	if(rcptsDomainSections.empty())
	{
		ZKASSERT(!"Unable to split recipient's domain on character.");
		return false;
	}

	ZKSTD::vector<ZKSTD::string>::reverse_iterator senderIt(sendersDomainSections.rbegin());
	ZKSTD::vector<ZKSTD::string>::reverse_iterator rcptIt(rcptsDomainSections.rbegin());

	while(senderIt != sendersDomainSections.rend() && rcptIt != rcptsDomainSections.rend())
	{
		if(ZKS::ci_string(senderIt->c_str()) != ZKS::ci_string(rcptIt->c_str()))
			break;

		++senderIt;
		++rcptIt;
	}

	bool bRetVal = false;
	if(senderIt == sendersDomainSections.rend())
		bRetVal = true;

	return bRetVal;
}


void RZkSmtpProtocol::SendMailAsUser(IZkContext& in_Context)
{
	// For debugging, add a zero so it can be traced out
	// BUT remember to subtract 1 from the size when you give it to
	// SetProtocolOutputLength
	if(!zkMemBufAppend(in_Context.GetZkError(), 
							m_MemBuf, "\0", 1))
	{
		ErrorResponse(in_Context);
		ZKASSERT(false);
	}
	else if(!ScanEmailForOffensiveWords(in_Context, USER_NYM_INDEX))
	{
		// User can choose to cancel if they see
		// information leaving their computer
		SetResponse(in_Context, USER_CANCELLED_RESPONSE);
		ZKVERIFY(ResetState(in_Context));
	}
	else if(!m_bAlreadyConnected)
	{
#ifdef ZK_LINUX
        in_addr addr = {0};
        if(!inet_aton("127.0.0.1", &addr))
        {
            ErrorResponse(in_Context);
            ZKASSERT(false);
        }
        else if( !memcmp(&addr, &in_Context.GetRemoteAddress().m_inet.sa.sin_addr, sizeof(in_addr)) )
        {
			m_bLinuxRelayHack = true;
            ZKSTD::string strMailRelay;
            ZkService::GetIClientCentral()->GetMailRelay(strMailRelay);
            ZkSocketAddress zkAddr = {0};
            if(!parse_inet_addr(in_Context.GetZkError(), strMailRelay.c_str(), &zkAddr.m_inet))
            {
                ErrorResponse(in_Context);
                ZKASSERT(false);
            }
            else
            {
                if(!zkAddr.m_inet.sa.sin_port)
                {// if user didn't specify a port in the Freedom mail page, set it to the default now.
                    zkAddr.m_inet.sa.sin_port = htons(25);
                }
                in_Context.SetRemoteAddress(zkAddr);
                in_Context.SetProxyAction( IZkContext::ZKPROXY_CONNECT_TO_SERVER );
                m_State = CONNECTING_TO_SERVER;
                m_bAlreadyConnected = true;
            }
                
        }
        else
        {
            in_Context.SetProxyAction( IZkContext::ZKPROXY_CONNECT_TO_SERVER );
            m_State = CONNECTING_TO_SERVER;
            m_bAlreadyConnected = true;
        }
#else
        in_Context.SetProxyAction( IZkContext::ZKPROXY_CONNECT_TO_SERVER );
        m_State = CONNECTING_TO_SERVER;
        m_bAlreadyConnected = true;
#endif

	}
	else
	{
		NotifyConsoleOfMail(USER_NYM_INDEX);
		SendMailCommandToServer(in_Context);
	}
}

void RZkSmtpProtocol::SendHeloCommandToServer(IZkContext& in_Context)
{

	in_Context.SetProxyAction( IZkContext::ZKPROXY_SEND_TO_SERVER);

	in_Context.SetProtocolOutput(HELO_CMD_STRING);
	in_Context.SetProtocolOutputLength(strlen(HELO_CMD_STRING));

	m_State = SENT_HELO_CMD;
}

void RZkSmtpProtocol::SendMailCommandToServer(IZkContext& in_Context)
{
	in_Context.SetProxyAction( IZkContext::ZKPROXY_SEND_TO_SERVER);
	snprintf((char *)in_Context.ByteBuffer(), in_Context.ByteBufferSize(), MAIL_CMD_FORMAT_STRING, m_strMailFromArg.c_str());
	in_Context.SetProtocolOutput(in_Context.ByteBuffer());
	in_Context.SetProtocolOutputLength( strlen((char*)in_Context.ByteBuffer()) );

	m_State = SENT_MAIL_FROM_CMD;
}

void RZkSmtpProtocol::SendRcptCommandToServer(IZkContext& in_Context)
{
	ZKASSERT( m_vecRecipients.size() );

	in_Context.SetProxyAction( IZkContext::ZKPROXY_SEND_TO_SERVER);

	snprintf((char *)in_Context.ByteBuffer(), in_Context.ByteBufferSize(), RCPT_CMD_FORMAT_STRING, m_vecRecipients[m_vecRecipients.size() - 1].c_str());
	m_vecRecipients.pop_back();

	in_Context.SetProtocolOutput(in_Context.ByteBuffer());
	in_Context.SetProtocolOutputLength( strlen((char*)in_Context.ByteBuffer()) );

	m_State = SENT_RCPT_TO_CMD;
}

void RZkSmtpProtocol::SendDataCommandToServer(IZkContext& in_Context)
{
	in_Context.SetProxyAction( IZkContext::ZKPROXY_SEND_TO_SERVER);

	in_Context.SetProtocolOutput(DATA_CMD_STRING);
	in_Context.SetProtocolOutputLength(strlen(DATA_CMD_STRING));

	m_State = SENT_DATA_CMD;
}

void RZkSmtpProtocol::SendMailBodyToServer(IZkContext& in_Context)
{
	char *pEmailData = NULL;
	unsigned int lenEmailData = 0;
	// Put the pointer here, so that ZkProxy will free it after sending the data
	in_Context.GetZkMemBuf() = m_MemBuf;
	m_MemBuf = NULL;
	ZKVERIFY(zkMemBufGetData(in_Context.GetZkError(), in_Context.GetZkMemBuf(),
								(void**)&pEmailData, &lenEmailData));

	in_Context.SetProtocolOutput(pEmailData);
	// Why -1? Because we added a null byte at the end to make tracing easier!
	// but don't send that null byte to the server
	in_Context.SetProtocolOutputLength(lenEmailData - 1);

	in_Context.SetProxyAction( IZkContext::ZKPROXY_SEND_TO_SERVER);
	m_State = SENT_MAIL_BODY;
}

void RZkSmtpProtocol::SendRsetCommandToServer(IZkContext& in_Context)
{
	in_Context.SetProxyAction( IZkContext::ZKPROXY_SEND_TO_SERVER);

	in_Context.SetProtocolOutput(RSET_CMD_STRING);
	in_Context.SetProtocolOutputLength(strlen(RSET_CMD_STRING));

	m_State = SENT_RSET_CMD;
}

void RZkSmtpProtocol::SendQuitCommandToServer(IZkContext& in_Context)
{
	if(m_bAlreadyConnected)
	{
		in_Context.SetProxyAction( IZkContext::ZKPROXY_SEND_TO_SERVER);

		in_Context.SetProtocolOutput(QUIT_CMD_STRING);
		in_Context.SetProtocolOutputLength(strlen(QUIT_CMD_STRING));
		
		m_State = SENT_QUIT_CMD;
	}
	else
	{
		// No one to disconnect from!
		SetResponse(in_Context, QUIT_RESPONSE);
		in_Context.SetCloseConnection(true);
	}
}

void RZkSmtpProtocol::TraceOutgoingStream(IZkContext& in_Context)
{
	switch(in_Context.GetProxyAction())
	{
	case IZkContext::ZKPROXY_SEND_TO_CLIENT:
		ZKTRACE1("Telling SMTP Client: %s\r\n", (char*)in_Context.GetProtocolOutput());
		break;
	case IZkContext::ZKPROXY_SEND_TO_SERVER:
		ZKTRACE1("Telling SMTP Server: %s\r\n", (char*)in_Context.GetProtocolOutput());
		break;
	default:
		break;
	}

}
