// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s): Mario Contestabile

#include "RZkSslProxy.h"
#include <openssl/err.h>
#include <ssl_hlp.h>
#include <cert_hlp.h>



RZkSslProxy::RZkSslProxy(ZkSocket in_LocalSocket)
	: m_LocalSocket(in_LocalSocket)
	, m_ctx(NULL)
	, m_sctx(NULL)
	, m_clientconn(NULL)
	, m_serverconn(NULL)
	, m_RemoteSocket(NULL)
	, m_clientssl(NULL)
	, m_serverssl(NULL)
	, m_bIgnoreCaProblems(false)
{
	/*
	SSL_library_init() only registers ciphers. Another important initialization is the seeding of the PRNG 
	(Pseudo Random Number Generator), which has to be performed separately. 
	*/
	SSL_library_init();

	SSL_load_error_strings();
	OpenSSL_add_ssl_algorithms();
}

RZkSslProxy::~RZkSslProxy()
{
	ZKVERIFY( zkSocketFree(&m_err, &m_RemoteSocket) );
}


bool
RZkSslProxy::RunProxy()
{
	ZKASSERT(m_ctx);
	ZKASSERT(m_sctx);

	if(!ReceiveRemoteAddress())	// Init m_RemoteAddress 
	{
		return false;
	}

	// If the close flag is set then the socket is shut down and closed when the BIO is freed. 
	m_clientconn = BIO_new_socket(m_LocalSocket->descriptor, BIO_CLOSE); // returns the newly allocated BIO or NULL if an error occurred

	if(!m_clientconn)
	{
		ZKTRACE2("BIO_new_socket failed with local socket [%d] [%s]", m_LocalSocket->descriptor, GetLastProxyError() );
		return false;
	}

	// Instead of creating a ZkSocket, I could probably use BIO calls directly
	// for example, Calling BIO_new(), then BIO_set_conn_ip() and BIO_set_conn_int_port() with m_RemoteAddress
	// and finally calling BIO_do_connect() to do the connection
	// Since m_LocalSocket has already accepted the connection in DInetD, we'll use another ZkSocket
	// with the BIO_new_socket to associate our socket with a BIO object.

	ZKVERIFY( zkSocketNew(&m_err, AF_INET, SOCK_STREAM, IPPROTO_TCP, &m_RemoteSocket) );

	if( !zkSocketConnectAddr(&m_err, m_RemoteSocket, m_RemoteAddress) )
	{
		ZKTRACE1("zkSocketConnectAddr failed to [%s]\n", inet_ntoa(m_RemoteAddress.m_inet.sa.sin_addr) )
		return false;
	}

	m_serverconn = BIO_new_socket(m_RemoteSocket->descriptor, BIO_CLOSE);

	if(!m_serverconn)
	{
		ZKTRACE2("BIO_new_socket failed with local socket [%d] [%s]", m_RemoteSocket->descriptor, GetLastProxyError() );
		return false;
	}

	if(!TalkSSL())
	{
		ZKTRACE0("TalkSSL() returned false\n");
		return false;
	}

	if(!SetTransferToNonBlocking())
	{
		ZKTRACE0("SetTransferToNonBlocking() returned false\n");
		return false;
	}

	if(!ExchangeData())
	{
		ZKTRACE0("ExchangeData() returned false\n");
		return false;
	}

	ZKTRACE0("SSL proxy returning normally\n");
	return true;
}


bool
RZkSslProxy::Init(const bool in_bIgnoreCaProblems, const char* in_CAFile, const char* in_CAFolder)
{

	ZKASSERT(in_bIgnoreCaProblems || in_CAFile || in_CAFolder);

	m_bIgnoreCaProblems = in_bIgnoreCaProblems;

	/*
	A TLS/SSL connection established with the 'SSLv23_' methods will understand the SSLv2, SSLv3, 
	and TLSv1 protocol. A client will send out SSLv2 client hello messages and will indicate that it 
	also understands SSLv3 and TLSv1. A server will understand SSLv2, SSLv3, and TLSv1 client hello 
	messages. This is the best choice when compatibility is a concern
	*/
	m_ctx = SSL_CTX_new(SSLv23_server_method());
	if (!m_ctx)
	{
		ZKTRACE1("SSL_CTX_new failed [%s]", GetLastProxyError() );
		return false;
	}

	m_sctx = SSL_CTX_new(SSLv23_client_method());
	if (!m_sctx)
	{
		ZKTRACE1("SSL_CTX_new failed [%s]", GetLastProxyError() );
		return false;
	}

	if(in_CAFile) // if not NULL, it points to a file of CA certificates in PEM format. The file can contain several CA certificates
	{
		if(1 != SSL_CTX_load_verify_locations(m_sctx, in_CAFile,0))
		{
			ZKTRACE2("SSL_CTX_load_verify_locations failed with PEM file [%s] [%s]", in_CAFile, GetLastProxyError() );

			if(!m_bIgnoreCaProblems)
				return false;
		}
					 
	}else if(in_CAFolder) // if not NULL, it points to a directory containing CA certificates in PEM format. The files each contain one CA certificate
	{
		if(1 != SSL_CTX_load_verify_locations(m_sctx, 0, in_CAFolder))
		{
			ZKTRACE2("SSL_CTX_load_verify_locations failed with folder [%s] [%s]", in_CAFolder, GetLastProxyError() );

			if(!m_bIgnoreCaProblems)
				return false;
		}
	}
	else if(!m_bIgnoreCaProblems)
	{
		ZKTRACE0("in_CAFile and in_CAFolder are NULL, yet m_bIgnoreCaProblems is not true, so returning false\n");
		return false;
	}
	
	if (!SSL_CTX_set_cipher_list(m_ctx, "ALL"))
	{
		ZKTRACE1("SSL_CTX_set_cipher_list failed [%s]", GetLastProxyError() );
		return false;
	}

	/*
	Sets the callback which will be called when a temporary private key is required. 
	*/

	SSL_CTX_set_tmp_rsa_callback(m_ctx, TmpRsa);

	return true;
}


char *
RZkSslProxy::GetLastProxyError()
{
	static char buf[1024] = { 0 };
	ERR_error_string( ERR_get_error(), buf );
	return buf;
}

RSA *
RZkSslProxy::TmpRsa(SSL * in_s, int is_export, int in_keylength)
{
	static RSA *rsa_tmp = NULL;

	/*
	RSA_generate_key generates a key pair and returns it in a newly allocated RSA structure. 
	The pseudo-random number generator must be seeded prior to calling RSA_generate_key(). 
	*/

	if (NULL == rsa_tmp)
	{
		rsa_tmp = RSA_generate_key(in_keylength, RSA_F4, NULL, NULL);
	}

	return(rsa_tmp);
}

bool
RZkSslProxy::ReceiveRemoteAddress()
{

	bool ReturnValue = true;

#ifdef WIN32
	char szBuffer[sizeof( sockaddr )];

	// Recv the real sockaddr
	if( 1 != zkSocketReadAll(&m_err, m_LocalSocket, szBuffer, sizeof (sockaddr) ) )
	{
		ReturnValue = false;
	}
	else 
	{		
		if( szBuffer[0] == AF_INET )
		{
			memcpy( &m_RemoteAddress, szBuffer, sizeof (sockaddr) );			
		}
		else 
		{
			ReturnValue = false;
			ZKASSERT( false );
		}
	}
#else
    try {
        RShimInterface::getShimInterface().VerifySocketConnection(	m_LocalSocket, 
																	m_RemoteAddress.m_inet.sa.sin_addr.s_addr,
																	m_RemoteAddress.m_inet.sa.sin_port);
        m_RemoteAddress.m_inet.sa.sin_family = AF_INET;              
    } catch (...) 
	{
        // VerifySocketConnection... throws to indicate errors
        return false;
    }
#endif

	return ReturnValue;
}


bool
RZkSslProxy::TalkSSL()
{
	ZKASSERT(m_serverconn);
	ZKASSERT(m_clientconn);

	m_serverssl = SSL_new(m_sctx); // create a new SSL structure for a connection 

	if(!m_serverssl)
	{
		ZKTRACE1("SSL_new failed [%s]", GetLastProxyError() );
		return false;
	}

	/* 
		Used to associate the network connection with the object 
		If a BIO is non-blocking, the ssl will also have non-blocking behaviour

		SSL_set_bio() connects the BIOs rbio and wbio for the read and write 
		operations of the TLS/SSL (encrypted) side of ssl. 

		The SSL engine inherits the behaviour of rbio and wbio, respectively. 
		If a BIO is non-blocking, the ssl will also have non-blocking behaviour. 

		If there was already a BIO connected to ssl, BIO_free() will be called 
		(for both the reading and writing side, if different). 
	*/

	SSL_set_bio(m_serverssl, m_serverconn, m_serverconn);

    /* initiate the TLS/SSL handshake 
	   the entropy pool must be seeded!
	*/
    if (SSL_connect(m_serverssl) <= 0) 
	{
		ZKTRACE1("SSL_connect failed [%s]", GetLastProxyError() );
		return false;
	}

#ifndef NDEBUG
	SslDump("\\remotessldump.txt", m_serverssl);
#endif

    /* SSL_get_verify_result() can only return one error code while the verification 
	   of a certificate can fail because of many reasons at the same time. Only the last 
	   verification error that occured during the processing is available from 
	   SSL_get_verify_result(). 

       If no peer certificate was presented, the returned result code is X509_V_OK. 
	   This is because no verification error occured, it does however not indicate 
	   success. SSL_get_verify_result() is only useful in connection with SSL_get_peer_certificate(3).
	*/

	const long certresult = SSL_get_verify_result(m_serverssl);

	// We can continue even if we can't find the root cert,
	// but the browser will display a warning.

    if ( (certresult != X509_V_OK) && !m_bIgnoreCaProblems) 
	{
		ZKTRACE1("SSL_get_verify_result did not return X509_V_OK, AND we're not ignoring CA problems [%s]", GetLastProxyError() );
		return false;
    }


	/* returns a pointer to the X509 certificate the peer presented. If the peer did not 
	   present a certificate, NULL is returned. 
	*/
    X509* servercert = SSL_get_peer_certificate(m_serverssl);

	// Lets create one on the fly for this session.

	X509*     x509 = X509_new();
	EVP_PKEY* pkey = EVP_PKEY_new();

	const int certdurationindays = 1;

    if(!MakeSiteCert(x509, servercert, certdurationindays, pkey))
	{
		X509_free(servercert);
		ZKTRACE1("MakeSiteCert failed [%s]\n", GetLastProxyError() );
		return false;
	}

    X509_free(servercert);

	// Now lets use the certificate we created on the fly,
	// by giving it to our client context

	if(1 != SSL_CTX_use_certificate(m_ctx, x509) ) // loads certificate into context
	{
		ZKTRACE1("SSL_CTX_use_certificate failed [%s]\n", GetLastProxyError() );
		return false;
	}

	if(1 != SSL_CTX_use_PrivateKey(m_ctx, pkey) ) // adds pkey as the private key to m_ctx
	{
		ZKTRACE1("SSL_CTX_use_PrivateKey failed [%s]\n", GetLastProxyError() );
		return false;
	}

	/*
	checks the consistency of a private key with the corresponding certificate loaded into ctx.
	If more than one key/certificate pair (RSA/DSA) is installed, the last item installed will 
	be checked. If e.g. the last item was a RSA certificate or key, the RSA key/certificate pair
	will be checked
	*/
	if(1 != SSL_CTX_check_private_key(m_ctx) )
	{
		ZKTRACE1("SSL_CTX_check_private_key failed [%s]\n", GetLastProxyError() );
		return false;
	}

	m_clientssl = SSL_new(m_ctx);

	if(!m_clientssl)
	{
		ZKTRACE1("SSL_new failed [%s]\n", GetLastProxyError() );
		return false;
	}

	SSL_set_bio(m_clientssl, m_clientconn, m_clientconn);


	return true;
}

bool 
RZkSslProxy::MakeSiteCert(X509 *in_x509, X509 *in_orig, const int in_days, EVP_PKEY *out_pkey)
{  
  
	BIGNUM *num = BN_new();

	BIO *rootkey = BIO_new_file("\\root.key", "r"); // this should match the file name used at installation

	EVP_PKEY *rootpkey = PEM_read_bio_PrivateKey(rootkey, 0, 0, 0);

	if (!rootpkey)  // This shouldn't fail if the installation succeeded
	{
		BIO_free(rootkey); 

		unsigned long lerrorval = 0;
		const int errlen = 512;
		char errorstr[errlen] = { 0 };

#ifndef NDEBUG
		FILE* fptr = fopen("\\rootcertdump.txt", "w");
#endif

		// but even if we can't read our root cert from installation, we'll make one on the fly
		// and start using it...hopefully the user will install our root cert
		// in his CA;-)


		if(!MakeRootCert(
			 "\\root.crt", 
			 "\\root.key",	// this should match the file name used with *rootkey pointer
			 lerrorval,
			 errorstr, 
			 errlen,
#ifndef NDEBUG
			 fptr))
#else
			 NULL))
#endif
		{
			ZKTRACE2("MakeRootCert failed [%d] [%s]\n", lerrorval, errorstr );
			return false;
		}

		// It needs to be BIO_free'd and re-opened unfortunately
		// Failing to do so will cause PEM_read_bio_PrivateKey to fail

		BIO *rootkey = BIO_new_file("\\root.key", "r"); 

		rootpkey = PEM_read_bio_PrivateKey(rootkey, 0, 0, 0);

		if (!rootpkey)
		{
			ZKTRACE1("PEM_read_bio_PrivateKey failed [%s] in MakeSiteCert\n", GetLastProxyError() );
			return false;
		}
	}

	/*
	The EVP library provides a high-level interface to cryptographic functions

	RSA_generate_key generates a key pair and returns it in a newly allocated RSA structure. 
	The pseudo-random number generator must be seeded prior to calling RSA_generate_key(). 
	If key generation fails, RSA_generate_key() returns NULL
	*/
	if (!EVP_PKEY_assign_RSA(out_pkey, RSA_generate_key(	512, /* The modulus size will be num bits */
													3,   /* The exponent is an odd number, typically 3 or 65535 */
													0,   /* No Callback */ 
													0)))
	{
		return false;
	}


	X509_set_pubkey(in_x509, out_pkey);

	X509_set_subject_name(in_x509, X509_get_subject_name(in_orig));
	//X509_NAME *subj = X509_get_subject_name(x509);
	//X509_NAME_add_entry_by_txt(subj, "CN", MBSTRING_ASC, sitename, -1, -1, 0);

	X509_NAME *issuer = X509_get_issuer_name(in_x509);

	X509_NAME_add_entry_by_txt(issuer, "CN", MBSTRING_ASC, (unsigned char*)PROXYNAME, -1, -1, 0);

	BN_rand(num, 128, 0, 0);
	BN_to_ASN1_INTEGER(num, X509_get_serialNumber(in_x509));

	X509_gmtime_adj(X509_get_notBefore(in_x509), 0);
	X509_gmtime_adj(X509_get_notAfter(in_x509), 60*60*24*in_days);

	if (!X509_sign(in_x509, rootpkey, EVP_md5()))
	{
		return false;
	}

	BIO_free(rootkey);

	return true;
}

bool
RZkSslProxy::ExchangeData()
{
	ZKASSERT(m_serverssl);
	ZKASSERT(m_clientssl);

	if(!m_serverssl || !m_clientssl)
	{
		return false;
	}

	bool good = true;
	do
	{
		char line[1024] = { 0 };

		const int nRead = SSL_read(m_clientssl, line, sizeof line); // read from IE

		if (SslClosed(m_clientssl, nRead)) 
		{
			SSL_shutdown(m_serverssl);
			good = false;
			break;
		}

		if (nRead > 0) 
		{
			const int nWritten = SSL_write(m_serverssl, line, nRead); // write to ibm
		
			if (SslClosed(m_serverssl, nWritten)) 
			{
				SSL_shutdown(m_clientssl);
				good = false;
				break;
			}
		}
  
		const int nReadRemote = SSL_read(m_serverssl, line, sizeof line); // read from ibm

		if (SslClosed(m_serverssl, nReadRemote)) 
		{
			SSL_shutdown(m_clientssl);
			good = false;
			break;
		}

		if (nReadRemote > 0) 
		{
			const int nWrittenLocal = SSL_write(m_clientssl, line, nReadRemote); // write to IE

			if (SslClosed(m_clientssl, nWrittenLocal)) 
			{
				SSL_shutdown(m_serverssl);
				good = false;
				break;
			}
		}
	}while(good);

	return true;
}


bool 
RZkSslProxy::SslClosed(SSL * in_s, int in_n)
{
	ZKASSERT(in_s);

	/*
	Don't bother checking when we know there wasn't an error
	From SSL_read()/SSL_write(), If >0, The read operation was successful; the return value is the number of 
	bytes actually read from the TLS/SSL connection. 
	*/
	if (in_n > 0) 
		return false;

	/*
	SSL_get_error() returns a result code (suitable for the C ``switch'' statement) for a preceding 
	call to SSL_connect(), SSL_accept(), SSL_read(), SSL_peek(), or SSL_write() on ssl. The value returned 
	by that TLS/SSL I/O function must be passed to SSL_get_error() in parameter ret. 
	*/
	const int e = SSL_get_error(in_s, in_n);

	/*
	The TLS/SSL connection has been closed. If the protocol version is SSL 3.0 or TLS 1.0, 
	this result code is returned only if a closure alert has occurred in the protocol, i.e. 
	if the connection has been closed cleanly. 
	*/
	if (e == SSL_ERROR_ZERO_RETURN) 
	{
		ZKTRACE1("SSL_ERROR_ZERO_RETURN the SSL connection is closed [%s]\n", GetLastProxyError() );
		return true;
	}
	else if (e == SSL_ERROR_SYSCALL) //  Some I/O error occurred
	{
		ZKTRACE1("SSL_ERROR_SYSCALL the SSL connection is closed [%s]\n", GetLastProxyError() );
		return true;
	}

	return false;
}

bool
RZkSslProxy::SetTransferToNonBlocking()
{
	ZKASSERT(m_serverssl);
	ZKASSERT(m_clientssl);

	{
		ZkSocketDescriptor d = SSL_get_fd(m_clientssl);

		if (d < 0)
		{
			ZKTRACE0("SSL_get_fd(m_clientssl) failed \n");
			return false;
		}
    
	#ifdef WIN32
		ULONG ul = 1;
		const int nRet = ioctlsocket(d, FIONBIO, &ul);	// returns SOCKET_ERROR (-1) if it fails
	#else
		const int nRet = fcntl(d, F_SETFL, O_NONBLOCK);
	#endif

		if (nRet < 0)
		{
			ZKTRACE1("Couldn't set socket [%d] to non-blocking\n", d);
			return false;
		}
	}

	{
		ZkSocketDescriptor m = SSL_get_fd(m_serverssl);

		if (m < 0)
		{
			ZKTRACE0("SSL_get_fd(m_serverssl) failed \n");
			return false;
		}

	#ifdef WIN32
		ULONG ul = 1;
		const int nRet = ioctlsocket(m, FIONBIO, &ul);	// returns SOCKET_ERROR (-1) if it fails
	#else
		const int nRet = fcntl(m, F_SETFL, O_NONBLOCK);
	#endif

		if (nRet < 0)
		{
			ZKTRACE1("Couldn't set socket [%d] to non-blocking\n", m);
			return false;
		}
	}

	return true;
}