// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s): Mario Contestabile

#include "IZkProxy.h"
#include "ZkProxiesInc.h"
#include <zksocket.h>
#include <DZkError.h>
#include <openssl/ssl.h>


class ZkProxies_API RZkSslProxy : public IZkProxy
{
public:
	RZkSslProxy(ZkSocket in_LocalSocket);
	~RZkSslProxy();

	// PURPOSE: This is a required interface inherited from IZkProxy
	// REQUIRE: The initialization function Init() MUST be called beforehand	
	// PROMISE:
	virtual bool RunProxy();

	// PURPOSE: The is used to specify what behavior to adopt
	//			and where to find some required data.
	//			If in_bIgnoreCaProblems is true, then Init will still return true even if
	//			a CA error occurs. Typical errors are sending an invalid file type in 'in_CAFile'
	//			or an empty folder name in 'in_CAFolder', or a folder name with invalid file types in it.
	//			The File types MUST be PEM format. This is what ssleay uses.
	//			Ideally, you'd want to extract all the root certs on a user's machine, put those
	//			in a file and add our root cert, and use that file in Freedom.
	//			You can do this as follows, by selecting all the root certs in IE
	//			(Content Tab-Certificates Button, select them all and hit export)
	//			You can obtain a file with the installed certs.
	//			You can convert the file with
	//			openssl pkcs7 -in t2.p7b -inform DER -print_certs > t2.pem
	//			t2.p7b is the filename you gave when exporting the certs from IE
	//			t2.pem could be sent to in_CAFile
	//			Additionally, you can send in "M:\\lib3rd\\openssl\\certs"
	//			in 'in_CAFolder', ssleay has root certs there.
	// PROMISE: If 'in_CAFile' and 'in_CAFolder' are NULL, 'in_bIgnoreCaProblems' will
	//			be used to determine the return value
	//			If 'in_bIgnoreCaProblems' is false yet both pointer parameters are false, an assertion will trigger
	//			If both in_CAFile and in_CAFolder are non-NULL, in_CAFile will be used.
	bool		Init(const bool in_bIgnoreCaProblems, const char* in_CAFile, const char* in_CAFolder);

private:
	ZkSocket	m_LocalSocket;
	SSL_CTX*	m_ctx;
	SSL_CTX*	m_sctx;
	BIO*		m_clientconn;
	BIO*		m_serverconn;
	ZkSocket	m_RemoteSocket;
	SSL*		m_clientssl;
	SSL*		m_serverssl;
	bool		m_bIgnoreCaProblems;

	ZkSocketAddress m_RemoteAddress;	// as sent to us by the LSP
	DZkError		m_err;
	
	char *		GetLastProxyError();

	// PURPOSE: To Generate a key of appropriate size (using ???) and return it. 
	//			The export flag will be set if the reason for needing a temp key is that an export 
	//			ciphersuite is in use, in which case, keylength will contain the required keylength in bits. 
	//			
	static RSA*	TmpRsa(SSL *in_s, int is_export, int in_keylength);


	// PURPOSE:	To retrieve the real SSL server ip the connection should go to.
	//			Initializes 'm_RemoteAddress'
	//			Should be one of the first thing the proxy calls
	//			and if returns false then we were unable to obtain the remote Ip,
	//			exiting the proxy would be a wise move if this returns false
	bool		ReceiveRemoteAddress();

	// PURPOSE: To execute the actual SSL protocol exchange
	// REQUIRE: m_serverconn and m_clientconn must be properly setup beforehand
	bool		TalkSSL();

	// PURPOSE:	For each SSL web site we will generate a new certificate signed by us to give
	//			back to the browser. We make this certificate with the smallest key size
	//			possible for better speed (RSA 512).
	//
	//			Given a pointer to an X509 structure 'x509', this will 
	//			- Change 'out_pkey' type to EVP_PKEY_RSA
	//			- Give 'out_pkey' a new 512 bit RSA key
	//			- Set the 'x509' public key to 'out_pkey'
	//			- Change the issuer name to PROXYNAME (from <cert_hlp.h>)
	//			- Change the validity to be 'in_days'
	//			- Sign the 'x509' cert using the root certificate
	//REQUIRES: 
	//			- The out_pkey pointer must be allocated using EVP_PKEY_new() beforehand
	//			- The root certificate must exist. If it does not exist one will be created.
	bool		MakeSiteCert(X509 *in_x509, X509 *in_orig, const int in_days, EVP_PKEY *out_pkey);

	// PURPOSE: Once everything is setup properly this reads the data from the real server
	//			and feeds it back into the real broser stream
	// REQUIRE: m_serverssl and m_clientssl must be properly setup beforehand
	//			Always returns true, only returns false if m_serverssl or m_clientssl is NULL
	bool		ExchangeData();

	// PURPOSE: Returns true if the in_s SSL connection has been closed
	bool		SslClosed(SSL * in_s, int in_n);

	// PURPOSE: Once we are ready to handle the data transfer we want to do so
	//			using non-blocking sockets or else the UI feedback will not be there.
	bool		SetTransferToNonBlocking();
};