// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// Wallet.h
//

#if !defined(_WALLET_H_INCLUDED_)
#define _WALLET_H_INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#ifdef DIPCHTTPSERVER_IMPL
   #define DIPCHTTPSERVER_DLL __declspec( dllexport ) 
#else
   #define DIPCHTTPSERVER_DLL __declspec( dllimport ) 
#endif

#include <client.h>
#include <clientstl.h>
#include <RZkException.h>
#include <ZkService.h>
#include <IWallet.h>
#include <INymWallet.h>			//For LoginInfo
#include <ZkSocket.h>
#include <string_hlp.h>

/////////////////////////////////////////////////////////////////////////////////

class HtmlFormElem {
public:
	ZKSTD::string strName;
	ZKSTD::string strNameLwr;
	ZKSTD::string strType;
	ZKSTD::string strValue;

	HtmlFormElem(ZKSTD::string in_strName = "", 
				 ZKSTD::string in_strType = "", 
				 ZKSTD::string in_strValue = "")
		: strName(in_strName)
		, strType(in_strType)
		, strValue(in_strValue)
	{ 
		strNameLwr = in_strName;
		ZKS::MakeLower(strNameLwr);
	};
};

/////////////////////////////////////////////////////////////////////////////////

class DIPCHTTPSERVER_DLL Wallet AUTO_INSTANCE_COUNTING1(Wallet)
{
public:

	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: Constructs a Wallet object with the given in_strRequest.
	//			Run() should be called after construction.
	//			in_strRequest should resemble "/command?param1=value1&param2=value2"
	//			See WalletConsts.h for acceptable commands
	//			This class should close in_sock if it is going to block and not write back
	//			to the socket.  (This is so the BHO doesn't sit around waiting for nothing).
	//			Otherwise, it does not have to close the socket.  It will be closed by the caller.
	//
	// REQUIRE: The parameters should be "Escaped" before being passed into the constructor
	//			"/command?param1=I & I Rastafari" is unacceptable
	//			"/command?param1=I+%26+I+Rastafari" is acceptable
	//			Use ZKS::EscapeString().
	//
	// PROMISE:	This class is quite kind.  The following requests will be intelligently handled:
	//			"/command?"
	//			"/command"
	//			"/command?param1=value1&"

	Wallet(const ZKSTD::string& in_strRequest, const ZkSocket& in_sock);

	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: Standard destructor
	//
	// REQUIRE: Nothing
	//
	// PROMISE:	Nothing

	~Wallet();


	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: This method should be called shortly after constructing the object.  This method
	//			acts on the request that was passed into the ctor.  Here are the possibilities
	//			(See WalletConsts.h)
	//			1. "KillWindow":	Closes a dialog if one is open.  Otherwise does nothing.
	//								out_strResponse === ""
	//			2. "AddLoginInfo":	Adds login info to the wallet.  The username, password,
	//								domain combo is passed in through parameters in in_strRequest.
	//								out_strResponse === ""
	//			3. "FillInForm":	Pops up the "Select login info" dialog if login info exists
	//								for the specified domain.  Otherwise, it pops up the 
	//								"Login normally" dialog.
	//								This method then waits for CallBackInformThatWindowClosed()
	//								to be called by the dialog.
    //								->out_strResponse is the login info if the user clicked log in
	//								->it is g_noLoginInfoStored if the user clicked cancel or if
	//								  no login info is stored for this domain
	//
	// REQUIRE: Nothing
	//
	// PROMISE:	Will return false if
	//			1. Some error occurred in ZkService (unlikely), or
	//			2. in_strRequest (passed in in ctor) is not understandable (more likely)

	bool Run(ZKSTD::string& out_strResponse);


	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: Called by the UI as soon as a dialog is popped up.  This provides this class with
	//			the ability to later close the dialog (which we must do if the command is 
	//			"KillWindow").
	//
	// REQUIRE: Nothing
	//
	// PROMISE:	This class will killany existing (wallet related) dialog before popping up another
	//			one

	void CallBackInformOfMyDialog(CDialog* in_pDialog);


	///////////////////////////////////////////////////////////////////////////////////////////////
	// PURPOSE: Called by the UI as soon as a dialog is closed (either by user action or because
	//			this class closed it).
	//			If the dialog is a login dialog and the user wants to login,  the first methid
	//			should be called.
	//			Otherwise, the second method should be called with m_bFillInForm == true if 
	//			the form should be filled in (if m_bFillInForm -> the UI changed the state of a
	//			parameter this class passed up to it.)
	//
	//			If the user clicked cancel, or the dialog was told to close (eg. by this class),
	//			CallBackInformThatWindowClosed(false) should be called
	//
	// REQUIRE: Nothing
	//
	// PROMISE:	To return back to the caller of Run() very soon.


	void CallBackInformThatWindowClosed(const LoginInfo& in_loginInfo);
	void CallBackInformThatWindowClosed(bool in_bFillInForm = false);

	ZKSTD::string CallBackGetDomain() { if (m_pStrDomain) {return (*m_pStrDomain);} else {return "";} };

private:
	Wallet(const Wallet& rhs);
	Wallet& operator=(const Wallet& rhs);

	bool GetCommand();
	bool GetParameters();
	bool ProcessCommand();

	bool CmdKillWindow();
	bool CmdAddLoginInfo();
	bool CmdFillInForm();
	int CmdKnowForm();

	bool IsAuthenticated();

#ifdef INTERNAL_BUILD
	bool CmdTeachForm();
#endif //INTERNAL_BUILD

	bool OneClickFillForm();
	bool FillInLoginForm(bool in_isReloginForm = false);
	bool InformOfDragAndDrop();

	static CDialog*				s_pDialog;
	static ZKCriticalSection	s_cs;
	static bool					s_bGeneralFormSeen;

	HANDLE						m_hDialogClosedEvent;	
	
	ZKSTD::string				m_strRequest;
	ZKSTD::string				m_strCommand;
	ZKSTD::string				m_strResponse;

	String2StringMap			m_param2ValueMap;

	ZKSTD::vector<HtmlFormElem>	m_formElemVect;
	String2StringMap			m_realNameLwr2ZkNameMap;

	LoginInfo					m_loginInfo;
	bool						m_bFillInForm;

	ZKSTD::string*				m_pStrDomain;
	const ZkSocket				m_sock;

#ifdef INTERNAL_BUILD 
public:
	void SelfTest();
	void ClearAllMembers();
#endif //INTERNAL_BUILD
};

#endif //_WALLET_H_INCLUDED_