/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

 */

/*******************************************************************************
; zkshim driver, cache management routines
; Written by Marc Boucher <marc@mbsi.ca>
*******************************************************************************/

#include "zkshim_priv.h"

static LIST_HEAD(zkShimDgramCacheEntries);

static KRWLOCK_T zkShimDgramCacheLock; /* Lock protecting zkShimDgramCacheEntries list */

static ZKTIMER_T zkShimDgramCacheExpTimer; /* Cache expiration timer */

int zkshim_cacheinterval = ZKSHIM_CACHE_INTERVAL;
int zkshim_cachetimeout = ZKSHIM_CACHE_TIMEOUT;

/* Statistics: */
int zkShimDgramCacheActive, zkShimDgramCacheHits, zkShimDgramCacheMisses, zkShimDgramCacheExpired;

#ifdef ZK_LINUX
MODULE_PARM(zkshim_cacheinterval, "i");
MODULE_PARM(zkshim_cachetimeout, "i");
EXPORT_SYMBOL(zkshim_cacheinterval);
EXPORT_SYMBOL(zkshim_cachetimeout);

EXPORT_SYMBOL(zkShimDgramCacheActive);
EXPORT_SYMBOL(zkShimDgramCacheHits);
EXPORT_SYMBOL(zkShimDgramCacheMisses);
EXPORT_SYMBOL(zkShimDgramCacheExpired);
#endif /* ZK_LINUX */

/***************************************************************************
;  Timeout-handler called at zkshim_cacheinterval to expire cache entries
;****************************************************************************/
static void
zkShimDgramCacheExpire(ZKTIMERFUNCARG_T data)
{
	ZkShimDgramCacheEntry *zce;
	struct list_head *lh;
#ifdef ZK_LINUX
	unsigned long flags;
#endif /* ZK_LINUX */

	ZKDBG(100, "entered\n");

#ifdef ZK_SOLARIS
	zkShimDgramCacheExpTimer = 0;
#endif
	WRITE_LOCK (&zkShimDgramCacheLock, flags);
	lh = zkShimDgramCacheEntries.next;
	while (lh != &zkShimDgramCacheEntries) {
		zce = list_entry(lh, ZkShimDgramCacheEntry, zce_list);

		ZKDBG(75, "zce=%p age=%ld\n", zce, ZK_TIME_CURRENT() - zce->zce_time);
		if((ZK_TIME_CURRENT() - zce->zce_time) >= zkshim_cachetimeout){
			struct list_head *lh2;

			lh2 = lh->next;
			list_del(lh);
			zkShimDgramCacheActive--;
			zkShimDgramCacheExpired++;
			ZKDBG(75, "KMFREE zce=%p\n", zce);
			KMFREE(zce);
			lh = lh2;
		} else
			lh = lh->next;
	}
	WRITE_UNLOCK (&zkShimDgramCacheLock, flags);

	ZK_SET_TIMER(&zkShimDgramCacheExpTimer, zkshim_cacheinterval,
			zkShimDgramCacheExpire, 0);
}

/***************************************************************************
;  Purge the cache
;****************************************************************************/
void
zkShimDgramCachePurge(void)
{
	ZkShimDgramCacheEntry *zce;
#ifdef ZK_LINUX
	unsigned long flags;
#endif /* ZK_LINUX */

	ZKDBG(20, "entered\n");

	WRITE_LOCK (&zkShimDgramCacheLock, flags);
	while(!list_empty(&zkShimDgramCacheEntries)) {
		zce = list_entry(zkShimDgramCacheEntries.next, ZkShimDgramCacheEntry, zce_list);
		list_del(&zce->zce_list);
		zkShimDgramCacheActive--;
		KMFREE(zce);
	}
	WRITE_UNLOCK (&zkShimDgramCacheLock, flags);
}

/***************************************************************************
;  Lookup key in cache and return corresponding rule entry
;****************************************************************************/
ZkShimRuleEntry *
zkShimDgramCacheLookup(ZkShimDgramCacheKey *zck)
{
	struct list_head *lh;
	ZkShimDgramCacheEntry *zce;

	ZKDBG(100, "entered\n");

	READ_LOCK (&zkShimDgramCacheLock);
	for(lh = zkShimDgramCacheEntries.next; lh != &zkShimDgramCacheEntries; lh = lh->next) {
		zce = list_entry(lh, ZkShimDgramCacheEntry, zce_list);

		if(ZK_KEYS_EQUAL(&zce->zce_cachekey, zck)) {
			zce->zce_time = ZK_TIME_CURRENT();
			zkShimDgramCacheHits++;
			READ_UNLOCK (&zkShimDgramCacheLock);
			ZKDBG(75, "HIT sport=%d dport=%d\n", (int)ntohs(zck->zck_sport), (int)ntohs(zck->zck_dport));
			return zce->zce_ruleentry;
		}
	}

	zkShimDgramCacheMisses++;
	READ_UNLOCK (&zkShimDgramCacheLock);
	ZKDBG(90, "MISS sport=%d dport=%d\n", (int)ntohs(zck->zck_sport), (int)ntohs(zck->zck_dport));
	return NULL;
}

/***************************************************************************
;  Update cache entry
;****************************************************************************/
void
zkShimDgramCacheUpdate(ZkShimDgramCacheKey *zck, ZkShimRuleEntry *zre)
{
	struct list_head *lh;
	ZkShimDgramCacheEntry *zce = NULL;
#ifdef ZK_LINUX
	unsigned long flags;
#endif /* ZK_LINUX */

	ZKDBG(100, "entered\n");

	WRITE_LOCK (&zkShimDgramCacheLock, flags);

	/* first, look for an existing identical entry */
	for(lh = zkShimDgramCacheEntries.next; lh != &zkShimDgramCacheEntries; lh = lh->next) {
		zce = list_entry(lh, ZkShimDgramCacheEntry, zce_list);

		if(ZK_KEYS_EQUAL(&zce->zce_cachekey, zck)) {
			break;
		}
	}

	if(lh != &zkShimDgramCacheEntries) {
		list_del(&zce->zce_list);
		ZKDBG(90, "EXISTING sport=%d dport=%d\n", (int)ntohs(zck->zck_sport), (int)ntohs(zck->zck_dport));
	} else {
		zce = (ZkShimDgramCacheEntry *) KMALLOC(sizeof(ZkShimDgramCacheEntry));
		if (!zce) {
			WRITE_UNLOCK (&zkShimDgramCacheLock, flags);
			cmn_err(CE_NOTE, "zkShimDgramCacheUpdate: malloc(%d) for ZkShimDgramCacheEntry failed\n", sizeof(ZkShimDgramCacheEntry));
			return;
		}

		bcopy(zck, &zce->zce_cachekey, sizeof(zce->zce_cachekey));
		zkShimDgramCacheActive++;
		ZKDBG(75, "NEW sport=%d dport=%d zce=%p\n", (int)ntohs(zck->zck_sport), (int)ntohs(zck->zck_dport), zce);
	}

	zce->zce_ruleentry = zre;
	zce->zce_time = ZK_TIME_CURRENT();
	list_add(&zce->zce_list, &zkShimDgramCacheEntries);

	WRITE_UNLOCK (&zkShimDgramCacheLock, flags);
}

/***************************************************************************
;  Initialize the cache
;****************************************************************************/
void
zkShimDgramCacheInit(void)
{
	ZKDBG(5, "entered\n");

	zkShimDgramCacheActive = zkShimDgramCacheHits = \
		zkShimDgramCacheMisses = zkShimDgramCacheExpired = 0;
	INIT_LOCK (&zkShimDgramCacheLock, "zkshim zkShimDgramCacheLock");
	INIT_LIST_HEAD(&zkShimDgramCacheEntries);

	ZK_SET_TIMER(&zkShimDgramCacheExpTimer, zkshim_cacheinterval,
			zkShimDgramCacheExpire, 0);
}

/***************************************************************************
;  Cleanup cache resources before unloading
;****************************************************************************/
void
zkShimDgramCacheCleanup(void)
{
	ZKDBG(5, "entered\n");

	ZK_CANCEL_TIMER(&zkShimDgramCacheExpTimer);
	zkShimDgramCachePurge();
	DESTROY_LOCK(&zkShimDgramCacheLock);
}


#ifdef CONFIG_PROC_FS
static int
zkShimDgramCacheProcInfo(char *buffer, char **start, off_t offset, int length, int reset)
{
	int len=0;
	off_t pos=0;
	char temp[129];

	struct list_head *lh;
	ZkShimDgramCacheEntry *zce;
    ZkShimRuleEntry *zre;

	char type='0';
	char action='0';


	pos = 128;

	if (offset<128)	{
		/* change to snprintf */
		sprintf(buffer,"%-127s\n", "Entry\t\tTYPE\tACTION\tIPARG\t\tPORTARG\tTOS\tProto\tSourceIP/Mask\t\tDest. IP/Mask\t\tSrcPort\tDstPort\tUID\tPID\tSID\t");

		len = 128;
  	}
	
  	
	READ_LOCK (&zkShimDgramCacheLock);
	start_bh_atomic();

	for(lh = zkShimDgramCacheEntries.next; lh != &zkShimDgramCacheEntries; lh = lh->next) {
		zce = list_entry(lh, ZkShimDgramCacheEntry, zce_list);
        zre = zce->zce_ruleentry;

        /*
         *	Spin through entries until we are ready
         */
        pos += 128;
        
        if (pos <= offset) {
            len = 0;
				continue;
        }
        
        if (zre->zre_rule.zru_type== ZRU_FILTER )
            type='F';
        else if (zre->zre_rule.zru_type== ZRU_SOCKET )
            type='S';
        if (zre->zre_rule.zru_action== ZSA_ALLOW )
            action='A';
			else if (zre->zre_rule.zru_action== ZSA_DENY )
				action='D';
        else if (zre->zre_rule.zru_action== ZSA_REDIRECT )
            action='R';
        else if (zre->zre_rule.zru_action== ZSA_REDIRECT_TXIP )
            action='T';

        
        sprintf(temp, "%08lX\t%c\t%c\t%08lX\t%02X\t%02X\t%02X\t%08lX/%08lX\t%08lX/%08lX\t%d\t%d\t%d\t%d\t%d",
				zce,
				type,
				action,
				(int)ntohl(zre->zre_rule.zru_ip_arg),
				(int)ntohs(zre->zre_rule.zru_port_arg),
				zre->zre_rule.zru_ip_tos,
				zre->zre_rule.zru_ip_p,
				ntohl(zre->zre_rule.zru_ip_src),
				ntohl(zre->zre_rule.zru_ip_src_mask),
				ntohl(zre->zre_rule.zru_ip_dst),
				ntohl(zre->zre_rule.zru_ip_dst_mask),
				(int)ntohs(zre->zre_rule.zru_sport),
				(int)ntohs(zre->zre_rule.zru_dport),
				(int)zre->zre_rule.zru_uid,
				(int)zre->zre_rule.zru_pid,
				(int)zre->zre_rule.zru_sid);
        
        sprintf(buffer+len,"%-127s\n",temp);
        len += 128;
        if (pos >= offset+length)
            goto done;
    }

done:
	end_bh_atomic();
	READ_UNLOCK (&zkShimDgramCacheLock);
  	
  	*start = buffer+len-(pos-offset);
  	len = pos-offset;
  	if (len>length)
  		len = length;
  	return len;
} 

#endif
