/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

 */

/*******************************************************************************
; zkshim Linux driver
; Written by Marc Boucher <marc@mbsi.ca>
*******************************************************************************/

/******************************************************************************
 * Extensions for Linux client by Oliver Lavery <oliver@zks.net>
 ******************************************************************************/

#define ZKSHIM_MAIN
#include "zkshim_priv.h"

KRWLOCK_DECL(zkShimLock); /* Lock protecting all global data */

ZkShimUnit *zkShimUnitList = NULL; /* List of open units */

#ifdef ZKSHIM_DEBUG
int zkshim_debug = ZKSHIM_DEBUG;
#else
int zkshim_debug = 0;
#endif

static struct proto_ops *shim_stream_ops;
static struct proto_ops *shim_dgram_ops;

int (*real_inet_stream_connect)(struct socket *sock, struct sockaddr * uaddr,
			int addr_len, int flags);

int (*real_inet_dgram_connect)(struct socket *sock, struct sockaddr * uaddr,
			int addr_len, int flags);

#ifdef MODULE
/* because the kernel doesn't export the init_task symbol
   (required by for_each_task), we have to determine it at run-time */
/*#if (defined(ZK_CALDERA) & ZK_CALDERA == 24) || (defined(ZK_SUSE) & ZK_SUSE == 70 ) */
#ifdef  __SMP__
/* this is just a patch, pre 2.4 kernels do not resolve tasklist_lock but 2.4 will */
/* also 2.4 will UNDEF __SMP__ and use CONFIG_SMP instead so hopefully this will work out*/
rwlock_t  tasklist_lock = RW_LOCK_UNLOCKED;
# endif /* defined(ZK_CALDERA) & ZK_CALDERA == 24  */ 
struct task_struct *zkInitTaskPtr;
#endif /* MODULE */

/***************************************************************************
;  Driver open entry-point.
;  Initializes new ZkShimUnit element and adds it to zkShimUnitList
;****************************************************************************/
static int
zkShimDevOpen (struct inode * inode, struct file * file)
{
	ZkShimUnit *zun;
	int i;
	unsigned long flags;
	

	WRITE_LOCK (&zkShimLock, flags);

	if (file->private_data) {
		WRITE_UNLOCK (&zkShimLock, flags);
		return -EBUSY;	/* already open */
	}

	if ((zun = (ZkShimUnit *) kmalloc(sizeof (ZkShimUnit), GFP_KERNEL)) == NULL) {
		WRITE_UNLOCK (&zkShimLock, flags);
		return -ENOMEM;
	}
	file->private_data = zun;

	zun->zun_minordev = MINOR(ZKSHIM_MINOR);
	zun->zun_file = file;
	zun->zun_waitq = NULL;
	zun->zun_fasync = NULL;
	skb_queue_head_init(&zun->zun_readq);
	zun->zun_rules = NULL;
	zun->zun_secret = NULL;
	
	zun->zun_nextp = zkShimUnitList;
	zkShimUnitList = zun;

	WRITE_UNLOCK (&zkShimLock, flags);

	MOD_INC_USE_COUNT;
	return 0;
} /* zkShimDevOpen */

/***************************************************************************
;  Driver fasync entry-point.
;****************************************************************************/
static int
zkShimDevFAsync (int fd, struct file *file, int on)
{
	ZkShimUnit *zun = (ZkShimUnit *) file->private_data;
	int retval;

	ZKDBG(150, "entered\n");

	retval = fasync_helper(fd, file, on, &zun->zun_fasync);
	if (retval < 0)
		return retval;

	return 0;
} /* zkShimDevFAsync */

/***************************************************************************
;  Driver close entry-point.
;  Removes ZkShimUnit element from zkShimUnitList and frees associated
;  resources.
;****************************************************************************/
static int
zkShimDevClose (struct inode * inode, struct file * file)
{
	ZkShimUnit *zun, **prevzun;
	ZkShimRuleEntry *ruleEntry, *nextRuleEntry;
	ZKMB_T *skb;
	unsigned long flags;

	ZKDBG(150, "entered\n");

	zkShimDevFAsync(-1, file, 0);

	WRITE_LOCK (&zkShimLock, flags);

	for (prevzun = &zkShimUnitList; (zun = *prevzun); prevzun = &zun->zun_nextp)
		if (zun == (ZkShimUnit *) file->private_data)
			break;
	*prevzun = zun->zun_nextp;

	for(ruleEntry = zun->zun_rules; ruleEntry; ruleEntry = nextRuleEntry) {
		nextRuleEntry = ruleEntry->zre_nextp;
		KMFREE(ruleEntry);
	}

	while((skb = skb_dequeue(&zun->zun_readq)) != NULL)
		kfree_skb(skb);

    if (zun->zun_secret) KMFREE(zun->zun_secret);
	kfree(zun);

	file->private_data = NULL;

#ifdef CACHE
	zkShimDgramCachePurge();
#endif
    zkShimTcpCacheShutdownUnit(zun);

	WRITE_UNLOCK (&zkShimLock, flags);

	MOD_DEC_USE_COUNT;
	return 0;
} /* zkShimDevClose */

/***************************************************************************
;  Driver ioctl entry-point.
;****************************************************************************/
static int
zkShimDevIoctl (struct inode *inode, struct file *file,
				unsigned int cmd, unsigned long arg)
{
	ZkShimUnit *zun = (ZkShimUnit *) file->private_data;
	ZkShimRule zkShimRule;

	ZkShimSecret zkShimSecret;
	
	ZKDBG(150, "entered, cmd=%ud\n", cmd);

	switch (cmd) {
	case ZKIOC_ADD_RULE:
		if(copy_from_user(&zkShimRule, (ZkShimRule *)arg, sizeof(ZkShimRule)))
			return -EFAULT;

		return -zkShimAddRule(zun, &zkShimRule);

	case ZKIOC_DEL_RULE:

		if(copy_from_user(&zkShimRule, (ZkShimRule *)arg, sizeof(ZkShimRule)))
			return -EFAULT;

		return -zkShimDelRule(zun, &zkShimRule);

        /* Clear all rules for the given unit.
         * This is meant as a helper for the freedom client (which would need to 
         * delete all the rules on a unit when nyms change etc.)
         */
	case ZKIOC_CLEAR_RULES:
		return -zkShimClearRules(zun);

	case ZKIOC_ADD_SECRET:

		if(copy_from_user(&zkShimSecret, (ZkShimSecret *)arg, sizeof(ZkShimSecret)))
			return -EFAULT;
	
		return -zkShimAddSecret(zun,&zkShimSecret);
		
	default:
		printk(KERN_ERR "zkShimDevIoctl: Unknown ioctl 0x%x\n", cmd);
		return -EINVAL;
	}
} /* zkShimDevIoctl */

/***************************************************************************
;  Driver write entry-point.
;****************************************************************************/
static ssize_t
zkShimDevWrite (struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	ZKMB_T *skb;
	iphdr_t *ip;
	struct rtable *rt;
    int hh_len;

	ZKDBG(150, "entered, count=%d\n", count);

	/* Can't seek (pwrite) on this device */
	if (ppos != &file->f_pos)
		return -ESPIPE;

	if (count < sizeof(iphdr_t)) {
		return -EINVAL;
	}

    if (verify_area(VERIFY_READ, buf, count)) {
		printk(KERN_ERR "zkShimDevWSrv: recieved bad user memory buffer %p %lu", buf, count);
        return -EFAULT;
    }

	ip = (iphdr_t *)buf;

	/* Try to guess incoming interface */
	if (ip_route_output(&rt, ip->daddr, ip->saddr, RT_TOS(ip->tos), 0)) {
		ZKDBG(150, "zkShimDevWSrv: unable to guess incoming interface for daddr = 0x%"INADDRLFMT"x\n", ntohl(ip->saddr));
//		return -EIO;
        if (ip_route_output(&rt, ip->saddr, ip->daddr, RT_TOS(ip->tos), 0)) {
            printk(KERN_ERR "zkShimDevWSrv: unable to guess outgoing interface for daddr = 0x%"INADDRLFMT"x\n", ntohl(ip->saddr));
            return -EIO;
        }

        // Incoming packet. Ineject via netif_rx
                
        skb = dev_alloc_skb(count);
        if(!skb) {
            printk(KERN_ERR "zkShimDevWrite: dev_alloc_skb(%d) failed\n", count);
            ip_rt_put(rt);
            return -ENOMEM;
        }
        
        skb->dev = rt->u.dst.dev;
        
        memcpy(skb_put(skb,count), buf, count);
        skb->protocol = __constant_htons(ETH_P_IP);
        skb->pkt_type = PACKET_HOST;
        skb->mac.raw=skb->data;
        ZKDBG(10, "injecting packet 0x%p via interface %s\n", skb, skb->dev->name);
        netif_rx(skb);

        return count;
	}

    // Outgoing packet. inject via ip_send
    hh_len = (rt->u.dst.dev->hard_header_len + 15)&~15;

	ZKDBG(100, "hardHeaderLength for outgoing packet is %d\n", hh_len);

	skb = alloc_skb(count + hh_len+15, GFP_ATOMIC);
	if(!skb) {
		printk(KERN_ERR "zkShimDevWrite: alloc_skb(%d) failed\n", count);
        ip_rt_put(rt);
		return -ENOMEM;
	}
    
    skb_reserve(skb, hh_len);

    skb->nh.raw = skb_put(skb, count);

    /* This is a little redundant since we already verify_area'd the memory --ol */
	if (copy_from_user((void*)skb->nh.iph, buf, count))
    {
		printk(KERN_ERR "zkShimDevWSrv: can not copy user memory buffer %p %lu", buf, count);
        ip_rt_put(rt);
        kfree_skb(skb);
        return -EFAULT;
    }

	skb->dst = &rt->u.dst ;// dst_clone(&rt->u.dst);
    ip_send(skb);
	return count;
} /* zkShimDevWrite */

/***************************************************************************
;  Driver read entry-point.
;****************************************************************************/
static ssize_t
zkShimDevRead (struct file *file, char *buf, size_t count, loff_t *ppos)
{
	ZkShimUnit *zun = (ZkShimUnit *) file->private_data;
	struct wait_queue wait = { current, NULL };
	int err;
    size_t i = 0;
	ZKMB_T *skb;

	ZKDBG(5, "entered, count=%d\n", count);

	/* Can't seek (pread) on this device */
	if (ppos != &file->f_pos)
		return -ESPIPE;

    if (!zun) return -EINVAL;

	if(skb_queue_empty(&zun->zun_readq)) {
		if (file->f_flags & O_NONBLOCK)
			return -EAGAIN;
		add_wait_queue(&zun->zun_waitq, &wait);
repeat:
        /* In case it was closed while we we're waiting */
        zun = file->private_data;
		current->state = TASK_INTERRUPTIBLE;
		if (zun && skb_queue_empty(&zun->zun_readq) && !signal_pending(current)) {
			schedule();
			goto repeat;
		}
		current->state = TASK_RUNNING;
		if (zun) remove_wait_queue(&zun->zun_waitq, &wait);
	}

    if (!zun) return -EINVAL;

	if(!skb_queue_empty(&zun->zun_readq) &&
			(skb = skb_dequeue(&zun->zun_readq))) {
		i = count;
		if(i > skb->len)
			i = skb->len;

		skb->h.raw = skb->nh.raw;
		err = skb_copy_datagram(skb, 0, buf, i);

		kfree_skb(skb);

		if(err) {
			return err;
		}
	}

    if (i) {
        file->f_dentry->d_inode->i_atime = CURRENT_TIME;
        return i;
    }

	if (signal_pending(current))
		return -ERESTARTSYS;

	return 0;
} /* zkShimDevRead */

/***************************************************************************
;  Driver poll entry-point.
;****************************************************************************/
static unsigned int
zkShimDevPoll (struct file *file, poll_table * wait)
{
	ZkShimUnit *zun = (ZkShimUnit *) file->private_data;

	ZKDBG(150, "entered\n");

	poll_wait(file, &zun->zun_waitq, wait);
	if (!skb_queue_empty(&zun->zun_readq))
		return POLLIN | POLLRDNORM;

	return 0;
} /* zkShimDevPoll */

/*---------------------------------------------------------------------------
 */
ZKMB_T *
zkShimCloneMb(ZKMB_T *skb)
{
	ZKMB_T *skbn;

	if(skb_cloned(skb)) /* can't we really not clone clones? */
		skbn = skb_copy(skb, GFP_ATOMIC);
	else
		skbn = skb_clone(skb, GFP_ATOMIC);

	if (!skbn) {
		printk(KERN_ERR "zkShimCloneMb: skb_clone/copy failed\n");
	}

	return skbn;
}

/*---------------------------------------------------------------------------
 */
static void
zkShimDivertToUser(ZkShimUnit *zun, ZKMB_T *skb, int linkhdrlen)
{
	ZKDBG(150, "entered zun=0x%p skb=0x%p\n", zun, skb);

	/* Under Linux, linkhdrlen is assumed to always be 0 */

	ZKDBG(10, "redirecting packet 0x%p to user zun=0x%p\n", skb, zun);

	/* Duplicate the skb for our own purpose while leaving the original
	   one untouched, which, by returning FW_QUEUE, we cause the kernel not
	   to send anymore. */

	skb = zkShimCloneMb(skb);
	if (!skb) {
		return /* XXX FW_BLOCK */;
	}

	skb_queue_tail(&zun->zun_readq, skb);
	if (zun->zun_fasync)
/* This is a specific workaround for the kernel sources that ship with RedHat 6.2
 * Change this conditional if it affects other kernels, but later versions of the canonical sources 
 * do not have the patches that are in 6.2s kernel and I don't want to break building on 'stock' kernels
 * to support a couple of patches a distro applies --ol
 */
#if defined(ZK_REDHAT) & ZK_REDHAT == 62 & defined(KERNEL_PATCH_LEVEL) & KERNEL_PATCH_LEVEL == 3 & defined(LINUX_VERSION_CODE) & LINUX_VERSION_CODE == KERNEL_VERSION(2,2,16) 
         kill_fasync(zun->zun_fasync, SIGIO, SI_SIGIO);
		
#elif defined(ZK_REDHAT) & ZK_REDHAT == 62 & defined(KERNEL_PATCH_LEVEL) & KERNEL_PATCH_LEVEL >= 5 & defined(LINUX_VERSION_CODE) & LINUX_VERSION_CODE == KERNEL_VERSION(2,2,14) 
         kill_fasync(zun->zun_fasync, SIGIO, SI_SIGIO);

#elif defined(ZK_MANDRAKE) & ZK_MANDRAKE == 71 & defined(KERNEL_PATCH_LEVEL) & KERNEL_PATCH_LEVEL == 4 & defined(LINUX_VERSION_CODE) & LINUX_VERSION_CODE == KERNEL_VERSION(2,2,15)
			kill_fasync(zun->zun_fasync, SIGIO, SI_SIGIO);

#elif defined(ZK_REDHAT) & ZK_REDHAT == 70 & defined(KERNEL_PATCH_LEVEL) & KERNEL_PATCH_LEVEL == 22 & defined(LINUX_VERSION_CODE) & LINUX_VERSION_CODE == KERNEL_VERSION(2,2,16)
			kill_fasync(zun->zun_fasync, SIGIO, SI_SIGIO);
#else
			kill_fasync(zun->zun_fasync, SIGIO);
#endif

	wake_up_interruptible(&zun->zun_waitq);
}

/*---------------------------------------------------------------------------
 */
static void
zkShimIfOutputProc(struct device *dev, ZKMB_T *skb)
{
	ZKDBG(80, "entered dev=%s skb=0x%p\n", dev->name, skb);

	if(((struct rtable *)(skb->dst))->u.dst.output(skb)) {
		/* error */
		ip_statistics.IpOutDiscards++;
		return /*B_FALSE*/;
	} else
		return /*B_TRUE*/;
}

/*---------------------------------------------------------------------------
 */
static void
zkShimIfDiscardProc(struct device *dev, ZKMB_T *skb)
{
	ZKDBG(150, "entered dev=%s skb=0x%p\n", dev->name, skb);

	kfree_skb(skb);
}

/***************************************************************************
;  This function does the work of zkShimFwOutputCheck/zkShimFwForwardCheck
;****************************************************************************/
static int
zkShimOutputCheck (struct device *dev, struct iphdr *ip, ZKMB_T **pskb, boolean_t forwarding)
{
	uint16_t sport, dport;
	uint8_t type,code;


	if (zkShimUnitList == NULL)
		return FW_SKIP;

    if (ip->saddr == ip->daddr) {
        ZKDBG(1, "Oooo. Got one of these wierdo packets with source and dest that match\n");
    }

	if((ntohs((u_short)ip->ip_off) & 0x1fff) == 0) {
		if(ip->ip_p == IPPROTO_TCP) {
			sport = ((struct tcphdr *)((__u32 *)ip+ip->ip_hl))->source;
			dport = ((struct tcphdr *)((__u32 *)ip+ip->ip_hl))->dest;
		} else if(ip->ip_p == IPPROTO_UDP) {
			sport = ((struct udphdr *)((__u32 *)ip+ip->ip_hl))->source;
			dport = ((struct udphdr *)((__u32 *)ip+ip->ip_hl))->dest;
		} else if(ip->ip_p == IPPROTO_ICMP){
			
			/* drop dest. unreachables just to test if this will work */
			type = ((struct icmphdr *)((__u32 *)ip+ip->ip_hl))->type;
			code = ((struct icmphdr *)((__u32 *)ip+ip->ip_hl))->code;
			if (type == ICMP_DEST_UNREACH){
				if (code == ICMP_PORT_UNREACH) {
					ZKDBG(5, "zkShimOutputCheck icmp port unreachable message type:%d ,code:%d \n", type,code);
				   return FW_BLOCK;
				}
			}
		}	
      else {
			sport = dport = 0;
		}
	} else
		sport = dport = 0;

	if(zkShimProcessOutgoing(*pskb, ip, sport, dport,
			zkShimDivertToUser,
			(void (*)(void *, void *))zkShimIfOutputProc,
			(void (*)(void *, void *))zkShimIfDiscardProc, dev, 0)) {
		return FW_QUEUE; /* packet was diverted/queued internally */
	} else {
		return FW_SKIP; /* not matched, process packet normally */
	}
}
/***************************************************************************
;  Firewall forward-check entry-point.
;****************************************************************************/
static int
zkShimFwForwardCheck (struct firewall_ops *this, int pf, struct device *dev,
					void *phdr, void *arg, ZKMB_T **pskb)
{
	ZKDBG(150, "entered dev=%s\n", dev->name);

	return zkShimOutputCheck (dev, (struct iphdr *)phdr, pskb, B_TRUE);
} /* zkShimFwForwardCheck */

/***************************************************************************
;  Firewall input-check entry-point.
;****************************************************************************/
static int
zkShimFwInputCheck (struct firewall_ops *this, int pf, struct device *dev,
					void *phdr, void *arg, ZKMB_T **pskb)
{
	ZKDBG(150, "entered dev=%s\n", dev->name);

	return FW_SKIP;
} /* zkShimFwInputCheck */

/***************************************************************************
;  Firewall output-check entry-point.
;****************************************************************************/
static int
zkShimFwOutputCheck (struct firewall_ops *this, int pf, struct device *dev,
					void *phdr, void *arg, ZKMB_T **pskb)
{
	ZKDBG(150, "entered dev=%s\n", dev->name);

	return zkShimOutputCheck (dev, (struct iphdr *)phdr, pskb, B_FALSE);
} /* zkShimFwOutputCheck */

static struct firewall_ops zkshim_fw_ops = {
	NULL,
	zkShimFwForwardCheck,
	zkShimFwInputCheck,
	zkShimFwOutputCheck,
	PF_INET,
	INT_MAX /* highest priority because we want to be first in the chain */
};

static struct file_operations zkshim_fops = {
	NULL,		/* zkshim_seek */
	zkShimDevRead,
	zkShimDevWrite,
	NULL, 		/* zkshim_readdir */
	zkShimDevPoll,
	zkShimDevIoctl,
	NULL,		/* zkshim_mmap */
	zkShimDevOpen,
	NULL,		/* flush */
	zkShimDevClose,
	NULL,		/* fsync */
	zkShimDevFAsync,
	NULL,		/* check_media_change */
	NULL,		/* revalidate */
	NULL		/* lock */
};

static struct miscdevice zkshim_device = {
	ZKSHIM_MINOR,ZKSHIM_NAME,&zkshim_fops
};

#ifdef CONFIG_PROC_FS
/***************************************************************************
;  /proc/net/zkshim entry-point.
;****************************************************************************/
static int
zkShimProcInfo(char *buffer, char **start, off_t offset, int length, int reset)
{
	int len=0;
	off_t pos=0;
	char temp[129];
	ZkShimUnit *zun;
	ZkShimRuleEntry *zre;

	char type='0';
	char action='0';


	pos = 128;

	if (offset<128)	{
		/* change to snprintf */
		sprintf(buffer,"%-127s\n", "Unit\t\tTYPE\tACTION\tIPARG\t\tPORTARG\tTOS\tProto\tSourceIP/Mask\t\tDest. IP/Mask\t\tSrcPort\tDstPort\tUID\tPID\tSID\t");

		len = 128;
  	}
	
  	
	READ_LOCK (&zkShimLock);
	start_bh_atomic();

	for (zun = zkShimUnitList; zun; zun = zun->zun_nextp) {

		for (zre = zun->zun_rules; zre; zre = zre->zre_nextp) {
			/*
			 *	Spin through entries until we are ready
			 */
			pos += 128;

			if (pos <= offset) {
				len = 0;
				continue;
			}

			if (zre->zre_rule.zru_type== ZRU_FILTER )
				type='F';
			else if (zre->zre_rule.zru_type== ZRU_SOCKET )
				type='S';
			if (zre->zre_rule.zru_action== ZSA_ALLOW )
				action='A';
			else if (zre->zre_rule.zru_action== ZSA_DENY )
				action='D';
				else if (zre->zre_rule.zru_action== ZSA_REDIRECT )
					action='R';
					else if (zre->zre_rule.zru_action== ZSA_REDIRECT_TXIP )
						action='T';

			
			sprintf(temp, "%08lX\t%c\t%c\t%08lX\t%02X\t%02X\t%02X\t%08lX/%08lX\t%08lX/%08lX\t%d\t%d\t%d\t%d\t%d",

				
				(unsigned long)zun->zun_file,
				type,
				action,
				(int)ntohl(zre->zre_rule.zru_ip_arg),
				(int)ntohs(zre->zre_rule.zru_port_arg),
				zre->zre_rule.zru_ip_tos,
				zre->zre_rule.zru_ip_p,
				ntohl(zre->zre_rule.zru_ip_src),
				ntohl(zre->zre_rule.zru_ip_src_mask),
				ntohl(zre->zre_rule.zru_ip_dst),
				ntohl(zre->zre_rule.zru_ip_dst_mask),
				(int)ntohs(zre->zre_rule.zru_sport),
				(int)ntohs(zre->zre_rule.zru_dport),
				(int)zre->zre_rule.zru_uid,
				(int)zre->zre_rule.zru_pid,
				(int)zre->zre_rule.zru_sid);
			
			sprintf(buffer+len,"%-127s\n",temp);
			len += 128;
			if (pos >= offset+length)
				goto done;
		}
        }

done:
	end_bh_atomic();
	READ_UNLOCK (&zkShimLock);
  	
  	*start = buffer+len-(pos-offset);
  	len = pos-offset;
  	if (len>length)
  		len = length;
  	return len;
} /* zkShimProcInfo */


static struct proc_dir_entry zkshim_proc_dir_entry = {
	PROC_NET_ZKSHIM, sizeof(ZKSHIM_NAME)-1,
	ZKSHIM_NAME, S_IFREG | S_IRUGO, 1, 0, 0,
	0, &proc_net_inode_operations, zkShimProcInfo
};

static struct proc_dir_entry zkshim_tcpcache_proc_dir_entry = {
	PROC_NET_ZKSHIM_TCP, sizeof(ZKSHIM_NAME "_tcp")-1,
	ZKSHIM_NAME "_tcp", S_IFREG | S_IRUGO, 1, 0, 0,
	0, &proc_net_inode_operations, zkShimTcpCacheProcInfo
};

#  ifdef CACHE
static struct proc_dir_entry zkshim_dgramcache_proc_dir_entry = {
	PROC_NET_ZKSHIM_DGRAM, sizeof(ZKSHIM_NAME "_dgram")-1,
	ZKSHIM_NAME "_dgram", S_IFREG | S_IRUGO, 1, 0, 0,
	0, &proc_net_inode_operations, zkShimDgramCacheProcInfo
};
#  endif /*CACHE*/

#endif

/* Write data to a socket.
 * net/socket.c doesn't export sock_write. This version does the same thing
 * but takes a socket as an arg instead of an fd 
 */
static ssize_t shim_sock_write(struct socket *sock, const char *ubuf,
                          size_t size)
{
        struct msghdr msg;
        struct iovec iov;
        
        /* Kept this around for no particular reason.        */
        if(size == 0)             /* Match SYS5 behaviour */
                return 0;

        /* valid for UDP? it should be connect()ed at this point so it should have
         * a valid default daddr, one would think
         */
        msg.msg_name = NULL;
        msg.msg_namelen = 0;
        msg.msg_iov = &iov;
        msg.msg_iovlen = 1;
        msg.msg_control = NULL;
        msg.msg_controllen = 0;
	
        /* Hmm, this might need to change (assuming this works at all)
         * my thinking is it will get this to DinetD before any more data is sent
         * which will allow the proxy dst socket to get opened faster.
         */
        msg.msg_flags = 0;
        iov.iov_base = (void *)ubuf;
        iov.iov_len=size;
        
	ZKDBG(1, "Attempting to kick a buf down a shimmed socket.\n");
	
        return sock_sendmsg(sock, &msg, size);
}


/* Shim function for TCP connect() socketcall.
 * This function will check outgoing connect() calls against ZRU_PACKET rules
 * and take action as appropriate.
 */
int shim_inet_stream_connect(struct socket *sock, struct sockaddr * uaddr,
			int addr_len, int flags)
{
	ZkShimRule *rule;
	struct sockaddr_in *uaddr_in = (struct sockaddr_in*)uaddr;
  	ZkShimUnit *pzun;
	
	ZKDBG(2, " entered.\n");

	rule = zkShimMatchOutgoingSocket(sock, uaddr_in, IPPROTO_TCP, &pzun);
	if (!rule) {
		ZKDBG(2, " no rule for this connection.\n");
		return real_inet_stream_connect(sock, uaddr, addr_len, flags);
	}

	if ( (pzun==NULL) || (pzun->zun_secret==NULL || pzun->zun_secret->cookie==NULL)){
		ZKDBG(2, " Error with the secret stuff, Maybe no cookie  \n");
        /* NOT backwars compatable. Set the cookie or connections will FAIL */
        return -ECONNREFUSED;
	}	
	
	switch (rule->zru_action) {
	case ZSA_ALLOW :
		return real_inet_stream_connect(sock, uaddr, addr_len, flags);
	default :
	case ZSA_DENY :
		ZKDBG(2, " denying connection.\n");	
		return -ECONNREFUSED;
	case ZSA_REDIRECT :
	case ZSA_REDIRECT_TXIP : 
        {
            int rval;
            int saved_flags;

            uint32_t real_addr = uaddr_in->sin_addr.s_addr;	
            uint16_t real_port = uaddr_in->sin_port;

            void *membytes = vmalloc(sizeof(uint32_t) + sizeof(uint16_t) + ZKSHIM_COOKIE_SIZE );
				
            if (membytes == NULL) goto die;

            ZKDBG(5, "original dest. port is %hd.\n", ntohs(real_port));

            uaddr_in->sin_addr.s_addr = (rule->zru_ip_arg)? rule->zru_ip_arg : real_addr;
            uaddr_in->sin_port = (rule->zru_port_arg)? rule->zru_port_arg : real_port; 
            
            /* First of all we must make the socket blocking if it is not.
             * this makes are TX of the IP and Port work on non-blocking sockets 
             * (like Mozilla uses). This solution is not so fantastic, and
             * we should analyze better ways of doing this. For the time being
             * it is adequate, however. I would prefer to add an ioctl or such that
             * the freedom proxies could use to retrieve this data, but this approach 
             * preserves compatability with the windows client and is straightforward 
             */
            saved_flags = flags;
            
            /* remove nonblock and fasync flags */
            flags = flags & ~(O_NONBLOCK | FASYNC);
            
            rval = real_inet_stream_connect(sock, uaddr, addr_len, flags);
            if (rval != 0) {
                ZKDBG(1, "TCP connection error.\n");
                vfree(membytes);
                return rval;
            } 
            if (rule->zru_action == ZSA_REDIRECT_TXIP) 
            {
                int rval2;
                unsigned long flags;
                mm_segment_t saved_ds;
                
                ZKDBG(2, "kicking real daddr down connect()ed sock.\n");
                
                ZKDBG(5, "original dest. port was %hd.\n", ntohs(real_port));
                /* set up membytes with our dest stuff */
                memcpy (membytes, &real_addr, sizeof(real_addr));
                memcpy (membytes + sizeof(real_addr), &real_port, sizeof(real_port));
                /* The cookie is not assumed to be null terminated, just stream of bytes*/
                memcpy (membytes + sizeof(real_addr) + sizeof(real_port), &pzun->zun_secret->cookie, ZKSHIM_COOKIE_SIZE );
				 
                ZKDBG(5, "memcpy finished.\n");

                /* Set fs to disable access_ok checks (or permit them for kernel data segment) */
                saved_ds = get_fs();
                set_fs(KERNEL_DS);

                rval2 = shim_sock_write(sock, membytes, sizeof(uint32_t) + sizeof(uint16_t) + ZKSHIM_COOKIE_SIZE);					 

                ZKDBG(2, "***shim_sock_write returned %d\n", rval2);
                set_fs(saved_ds);                
            }
            vfree(membytes);
            return rval;	
        }
	}
die:
	return -ECONNREFUSED;
}			

/* Shim function for UDP connect() socketcall.
 * This function will check outgoing connect() calls against ZRU_PACKET rules
 * and take action as appropriate.
 */

int shim_inet_dgram_connect(struct socket *sock, struct sockaddr * uaddr,
			int addr_len, int flags)
{
	ZkShimRule *rule;
	struct sockaddr_in *uaddr_in = (struct sockaddr_in*)uaddr;

  	ZkShimUnit *pzun;

    /* NB this will not work properly for fasync or nonblock
       (in fact this code is not used by any proxies at the moment
    */
	
	ZKDBG(2, " entered.\n");

	rule = zkShimMatchOutgoingSocket(sock, uaddr_in, IPPROTO_UDP,&pzun);
	
	if (!rule) {
		ZKDBG(2, " no rule for this connection.\n");
		return real_inet_dgram_connect(sock, uaddr, addr_len, flags);
	}

	if ( (pzun==NULL) || (pzun->zun_secret==NULL || pzun->zun_secret->cookie==NULL)){
		ZKDBG(2, " Error with the secret stuff, Maybe no cookie  \n");
        /* NOT backwars compatable. Set the cookie or connections will FAIL */
        return -ECONNREFUSED;		
	}	
	
	switch (rule->zru_action) {
	case ZSA_ALLOW :
		return real_inet_dgram_connect(sock, uaddr, addr_len, flags);
	default :
	case ZSA_DENY :
		ZKDBG(2, " denying connection.\n");	
		return -ECONNREFUSED;
	case ZSA_REDIRECT :
	case ZSA_REDIRECT_TXIP :
        {
            int rval;
            int saved_flags;

            uint32_t real_addr = uaddr_in->sin_addr.s_addr;	
            uint16_t real_port = uaddr_in->sin_port;
            void *membytes = vmalloc(sizeof(uint32_t) + sizeof(uint16_t) + ZKSHIM_COOKIE_SIZE );

            if (membytes == NULL) goto die;

            ZKDBG(5, "original dest. port is %hd.\n", ntohs(real_port));

            uaddr_in->sin_addr.s_addr = (rule->zru_ip_arg)? rule->zru_ip_arg : real_addr;
            uaddr_in->sin_port = (rule->zru_port_arg)? rule->zru_port_arg : real_port; 
            
            /* First of all we must make the socket blocking if it is not.
             * this makes are TX of the IP and Port work on non-blocking sockets 
             * (like Mozilla uses). This solution is not so fantastic, and
             * we should analyze better ways of doing this. For the time being
             * it is adequate, however. I would prefer to add an ioctl or such that
             * the freedom proxies could use to retrieve this data, but this approach 
             * preserves compatability with the windows client and is straightforward 
             */
            saved_flags = flags;
            
            /* remove nonblock and fasync flags */
            flags = flags & ~(O_NONBLOCK | FASYNC);
            
            rval = real_inet_dgram_connect(sock, uaddr, addr_len, flags);
            if (rval != 0) {
                ZKDBG(1, "TCP connection error.\n");
                vfree(membytes);
                return rval;
            } 
            if (rule->zru_action == ZSA_REDIRECT_TXIP) 
            {
                int rval2;
                unsigned long flags;
                mm_segment_t saved_ds;
                
                ZKDBG(2, "kicking real daddr down connect()ed sock.\n");
                
                ZKDBG(5, "original dest. port was %hd.\n", ntohs(real_port));
                /* set up membytes with our dest stuff */
                memcpy (membytes, &real_addr, sizeof(real_addr));
                memcpy (membytes + sizeof(real_addr), &real_port, sizeof(real_port));


					 /* The cookie is not assumed to be null terminated, just stream of bytes*/
					 memcpy (membytes + sizeof(real_addr) + ZKSHIM_COOKIE_SIZE, &pzun->zun_secret->cookie,ZKSHIM_COOKIE_SIZE );
				 
                ZKDBG(5, "memcpy finished.\n");

                /* Set fs to disable access_ok checks (or permit them for kernel data segment) */
                saved_ds = get_fs();
                set_fs(KERNEL_DS);

                rval2 = shim_sock_write(sock, membytes, sizeof(uint32_t) + sizeof(uint16_t)+ZKSHIM_COOKIE_SIZE);					 

                ZKDBG(2, "***shim_sock_write returned %d\n", rval2);
                set_fs(saved_ds);                
            }
            vfree(membytes);
            return rval;	
        }
	}
die:
	return -ECONNREFUSED;
}			

/***************************************************************************
;  Driver entry-point for initialization.
;****************************************************************************/
#ifdef MODULE
EXPORT_NO_SYMBOLS;
int init_module(void)
#else
__initfunc(int zkshim_init(void))
#endif
{
	int res;

	int retval;
	struct socket *sock;
	
#ifdef MODULE
	struct task_struct *p;
#endif /* MODULE */

	ZKDBG(150, "entered\n");

	zkShimFragInit();

#ifdef CACHE
	zkShimDgramCacheInit();
#endif
    zkShimTcpCacheInit();

	res = register_firewall(PF_INET, &zkshim_fw_ops);
	if(res) {
		printk(KERN_ERR "zkshim_init: register_firewall failed\n");
		return res;
	}
	/* modifications for teh misc device
	res = register_chrdev(ZKSHIM_MAJOR, ZKSHIM_MODINFO, &zkshim_fops);
	*/
	res = misc_register(&zkshim_device);
	if(res) {
		printk(KERN_ERR "zkshim_init: register_chrdev failed\n");
		unregister_firewall(PF_INET, &zkshim_fw_ops);
		return res;
	}

#ifdef MODULE
	/* because the kernel doesn't export the init_task symbol
           (required by for_each_task), we have to determine it at run-time */

	for( p = current; p; p=p->next_task) {
		if(p->pid == 0)
			break;
	}
	zkInitTaskPtr = p;
#endif

#ifdef CONFIG_PROC_FS
	proc_net_register(&zkshim_proc_dir_entry);
	proc_net_register(&zkshim_tcpcache_proc_dir_entry);
#  ifdef CACHE
	proc_net_register(&zkshim_dgramcache_proc_dir_entry);
#  endif
#endif


	ZKDBG(1, "Hijacking af_inet.c:inet_stream_ops ... pray.\n");

	/* this should snag us a stream socket
	 */
	retval = sock_create(AF_INET, SOCK_STREAM, 0, &sock);
	if (retval < 0) {
		printk(KERN_EMERG "Oops, can't create an inet stream socket!\n");
		return 0;
	}
	ZKDBG(1, "Got stream proto_ops at %p\n", sock->ops);

	if (!sock->ops) {
		printk(KERN_EMERG "Whoa, stream proto_ops is NULL. Spooky.\n");
		sock_release(sock);
		return 0;
	}
	
	shim_stream_ops = sock->ops;
	real_inet_stream_connect = shim_stream_ops->connect;	
	ZKDBG(1, "found real stream con func at %p\n", real_inet_stream_connect);
	shim_stream_ops->connect = shim_inet_stream_connect;
	ZKDBG(1, "installed stream shim connect at %p == %p\n",
		shim_stream_ops->connect, shim_inet_stream_connect);
	
	sock_release(sock);
	
	ZKDBG(1, "Hijacking af_inet.c:inet_dgram_ops ... pray more.\n");

	/* this should snag us a datagram socket
	 */
	retval = sock_create(AF_INET, SOCK_DGRAM, 0, &sock);
	if (retval < 0) {
		printk(KERN_EMERG "Oops, can't create an inet dgram socket!\n");
		return 0;
	}
	ZKDBG(1, "Got dgram proto_ops at %p\n", sock->ops);

	if (!sock->ops) {
		printk(KERN_EMERG "Whoa, dgram proto_ops is NULL. Spooky.\n");
		sock_release(sock);
		return 0;
	}
	
	shim_dgram_ops = sock->ops;
	real_inet_dgram_connect = shim_dgram_ops->connect;	
	ZKDBG(1, "found real dgram con func at %p\n", real_inet_dgram_connect);
	shim_dgram_ops->connect = shim_inet_dgram_connect;
	ZKDBG(1, "installed dgram shim connect at %p == %p\n",
		shim_dgram_ops->connect, shim_inet_dgram_connect);
	
	sock_release(sock);

	ZKDBG(50, "brk = %p\n", current->mm->brk);

	return 0;
} /* init_module | zkshim_init */



/***************************************************************************
;  Loadable module entry-point for unloading.
;****************************************************************************/
#ifdef MODULE
void cleanup_module(void)
{
	unsigned long flags;

	ZKDBG(150, "entered\n");

	WRITE_LOCK (&zkShimLock, flags);

	if (zkShimUnitList) {
		WRITE_UNLOCK (&zkShimLock, flags);
		return  EBUSY;
	}

#ifdef CONFIG_PROC_FS
	proc_net_unregister(PROC_NET_ZKSHIM);
	proc_net_unregister(PROC_NET_ZKSHIM_TCP);
#ifdef CACHE
	proc_net_unregister(PROC_NET_ZKSHIM_DGRAM);
#endif
#endif

	misc_deregister(&zkshim_device);
	/*unregister_chrdev(ZKSHIM_MAJOR, ZKSHIM_MODINFO);*/
	unregister_firewall(PF_INET, &zkshim_fw_ops);

#ifdef CACHE
	zkShimDgramCacheCleanup();
#endif
    zkShimTcpCacheCleanup();

	zkShimFragCleanup();

	WRITE_UNLOCK (&zkShimLock, flags);
	if (shim_stream_ops) {
		shim_stream_ops->connect = real_inet_stream_connect;
		ZKDBG(1, "uninstalled stream shim connect at %p == %p\n",
			shim_stream_ops->connect, real_inet_stream_connect);
	}

	if (shim_dgram_ops) {
		shim_dgram_ops->connect = real_inet_dgram_connect;
		ZKDBG(1, "uninstalled dgram shim connect at %p == %p\n",
			shim_dgram_ops->connect, real_inet_dgram_connect);
	}

} /* cleanup_module */
#endif /* MODULE */
