/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

 */

/*******************************************************************************
; zkshim driver, packet rule-matching routines
; Written by Marc Boucher <marc@mbsi.ca>
*******************************************************************************/

#include "zkshim_priv.h"
#ifdef ZK_SOLARIS
#include "sunos5/procsock.h"
#endif /* ZK_SOLARIS */


/***************************************************************************
;  Add a secret
;****************************************************************************/
int
zkShimAddSecret(ZkShimUnit *zun, ZkShimSecret *secret) {
	
#ifdef ZK_LINUX
	unsigned long flags;
#endif /* ZK_LINUX */
	int ret;

	ZKDBG(5, "entered\n");

	/* This is suppose to be a one shot thing? */
	WRITE_LOCK (&zkShimLock, flags);

	/* if the secret is at initialization state then accept new one */
	/* for now check the null pointer */
	if(!zun->zun_secret){

		/* allocate emory */
		zun->zun_secret  = (ZkShimSecret *) KMALLOC(sizeof(ZkShimSecret));
		if (!zun->zun_secret) {
            cmn_err(CE_NOTE, "zkShimAddSecret: malloc(%d) for ZkShimRuleSecret failed\n", sizeof(ZkShimSecret));
            ret = ENOMEM ; goto ret;
		}
		/* adsd the secret */
		
		ZKDBG(2, "adding secret \n" );
		bcopy(secret, zun->zun_secret, sizeof(ZkShimSecret));
		ret = 0; /* successful */
		goto ret;
	}	
	ZKDBG(2, "Secret already exist \n" );
   ret = EEXIST;

ret:	
	WRITE_UNLOCK (&zkShimLock, flags);
	return ret; 
}

/***************************************************************************
;  Add a redirection rule
;****************************************************************************/
/* XXX: For client the rules need better validation. Esp. since some combinations of values
 * are plain old illegal, and others are simply menaingless
 */
int
zkShimAddRule(ZkShimUnit *zun, ZkShimRule *rule)
{
	boolean_t privileged;
	uid_t euid;
	ZkShimRuleEntry *zre, **prevzre;
#ifdef ZK_LINUX
	unsigned long flags;
#endif /* ZK_LINUX */

	ZKDBG(5, "entered\n");

#ifdef ZK_SOLARIS
	euid = zun->zun_cred.cr_uid;
	privileged = !drv_priv(&zun->zun_cred);
#endif ZK_SOLARIS
#ifdef ZK_LINUX
	euid = current->euid;
	privileged = (current->euid == 0);
#endif /* ZK_LINUX */

	if(!privileged) { /* non-root user? */
		if(!rule->zru_uid) { /* treat 0 as meaning "my uid" */
			rule->zru_uid = euid;
		}

		if(rule->zru_uid != euid) {
			return EPERM;
		}

		/* XXX: check authorization for other elements, such as pid etc.. */
	}

	/* check if identical rule already exists */
	WRITE_LOCK (&zkShimLock, flags);
	for (prevzre = &zun->zun_rules; (zre = *prevzre); prevzre = &zre->zre_nextp) {
		if(!bcmp(&zre->zre_rule, rule, sizeof(zre->zre_rule))) {
			WRITE_UNLOCK (&zkShimLock, flags);
			return EEXIST; /* already exists */
		}
	}

	zre = (ZkShimRuleEntry *) KMALLOC(sizeof(ZkShimRuleEntry));
	if (!zre) {
		cmn_err(CE_NOTE, "zkShimAddRule: malloc(%d) for ZkShimRuleEntry failed\n", sizeof(ZkShimRuleEntry));
		WRITE_UNLOCK (&zkShimLock, flags);
		return ENOMEM;
	}
	
	ZKDBG(2, "adding rule %d %d 0x%"INADDRLFMT"x/0x%"INADDRLFMT"x 0x%"INADDRLFMT"x/0x%"INADDRLFMT"x %d %d %d %d %d\n",
		rule->zru_ip_tos,
		rule->zru_ip_p,
		ntohl(rule->zru_ip_src), ntohl(rule->zru_ip_src_mask),
		ntohl(rule->zru_ip_dst), ntohl(rule->zru_ip_dst_mask),
		ntohs(rule->zru_sport),
		ntohs(rule->zru_dport),
		rule->zru_uid,
		rule->zru_pid,
		rule->zru_sid);
		
	bcopy(rule, &zre->zre_rule, sizeof(zre->zre_rule));

	zre->zre_unitp = zun;

	zre->zre_nextp = zun->zun_rules;
	zun->zun_rules = zre;

#ifdef CACHE
	zkShimDgramCachePurge();
#endif

	WRITE_UNLOCK (&zkShimLock, flags);

	return 0; /* successful */
}

/***************************************************************************
;  Remove a redirection rule
;****************************************************************************/
int
zkShimDelRule(ZkShimUnit *zun, ZkShimRule *rule)
{
	boolean_t privileged;
	uid_t euid;
	ZkShimRuleEntry *zre, **prevzre;
#ifdef ZK_LINUX
	unsigned long flags;
#endif /* ZK_LINUX */

	ZKDBG(5, "entered\n");

#ifdef ZK_SOLARIS
	euid = zun->zun_cred.cr_uid;
	privileged = !drv_priv(&zun->zun_cred);
#endif ZK_SOLARIS
#ifdef ZK_LINUX
	euid = current->euid;
	privileged = (current->euid == 0);
#endif /* ZK_LINUX */

	if(!privileged) { /* non-root user? */
		if(!rule->zru_uid) { /* treat 0 as meaning "my uid" */
			rule->zru_uid = euid;
		}

		if(rule->zru_uid != euid) {
			return EPERM;
		}

		/* XXX: check authorization for other elements, such as pid etc.. */
	}

	WRITE_LOCK (&zkShimLock, flags);
	for (prevzre = &zun->zun_rules; (zre = *prevzre); prevzre = &zre->zre_nextp) {
		if(!bcmp(&zre->zre_rule, rule, sizeof(zre->zre_rule))) {
			*prevzre = zre->zre_nextp;
#ifdef CACHE
			zkShimDgramCachePurge();
#endif
			WRITE_UNLOCK (&zkShimLock, flags);
			KMFREE (zre);
			return 0; /* successful */
		}
	}
	WRITE_UNLOCK (&zkShimLock, flags);

	return ESRCH; /* not found */
}

/***************************************************************************
;  Remove all redirection rules for a given unit
;****************************************************************************/
int
zkShimClearRules(ZkShimUnit *zun)
{
	ZkShimRuleEntry *zre, **prevzre;
	unsigned long flags;
	WRITE_LOCK (&zkShimLock, flags);
	for (prevzre = &zun->zun_rules; (zre = *prevzre); prevzre = &zre->zre_nextp) {
		*prevzre = zre->zre_nextp;
		KMFREE (zre);
	}
	zun->zun_rules = NULL;
#ifdef CACHE
	zkShimDgramCachePurge();
#endif
	WRITE_UNLOCK (&zkShimLock, flags);
	return 0;
}

/***************************************************************************
;  Test a rule against an IP packet
;****************************************************************************/
static boolean_t
zkShimPacketRuleMatches(ZkShimRule *rule, ZKMB_T *mp, iphdr_t *ip,
			ushort sport, ushort dport)
{

	struct task_struct *p;
	boolean_t found = B_FALSE;
	int i;
	
	if(rule->zru_type != ZRU_FILTER) {
		/*ZKDBG(10, "zkShimPacketRuleMatches called with zru_type != ZRU_FILTER\n");*/
		return B_FALSE;
	}
	
	ZKDBG(30, "rule/packet: ip_tos=%d/%d ip_p=%d/%d 0x%"INADDRLFMT"x/0x%"INADDRLFMT"x 0x%"INADDRLFMT"x/0x%"INADDRLFMT"x %d/%d %d/%d\n",
		rule->zru_ip_tos, ip->ip_tos,
		rule->zru_ip_p, ip->ip_p,
		ntohl(rule->zru_ip_src&rule->zru_ip_src_mask), ntohl(ip->ip_saddr&rule->zru_ip_src_mask),
		ntohl(rule->zru_ip_dst&rule->zru_ip_dst_mask), ntohl(ip->ip_daddr&rule->zru_ip_dst_mask),
		ntohs(rule->zru_sport), ntohs(sport),
		ntohs(rule->zru_dport), ntohs(dport));
	
	if(rule->zru_ip_tos && (rule->zru_ip_tos != ip->ip_tos))
		return B_FALSE;

	if(rule->zru_ip_p && (rule->zru_ip_p != ip->ip_p))
		return B_FALSE;

	if(rule->zru_ip_src_mask && ((rule->zru_ip_src&rule->zru_ip_src_mask) != (ip->ip_saddr&rule->zru_ip_src_mask)))
		return B_FALSE;

	if(rule->zru_ip_dst_mask && ((rule->zru_ip_dst&rule->zru_ip_dst_mask) != (ip->ip_daddr&rule->zru_ip_dst_mask)))
		return B_FALSE;

	if(rule->zru_sport && (rule->zru_sport != sport))
		return B_FALSE;

	if(rule->zru_dport && (rule->zru_dport != dport))
		return B_FALSE;

#ifdef ZK_SOLARIS
	if(sport && (rule->zru_uid || rule->zru_pid || rule->zru_sid)) {
		struct sonode *lso;
		proc_t *p;

		lso = zkShimFindLocalSocket(ip->ip_p, ip->ip_saddr, sport);
		if(!lso)
			return B_FALSE;

		p = zkShimFindVnodeProc(&lso->so_vnode);
		if(!p)
			return B_FALSE;

		if(rule->zru_uid && (rule->zru_uid != p->p_cred->cr_uid))
			return B_FALSE;

		if(rule->zru_pid && (rule->zru_pid != p->p_pidp->pid_id))
			return B_FALSE;

		if(rule->zru_sid && (rule->zru_sid != p->p_sessp->s_sidp->pid_id))
			return B_FALSE;
	}
#elif defined(ZK_LINUX)
	if((rule->zru_uid || rule->zru_pid || rule->zru_sid) && (!mp->sk || !mp->sk->socket))
			return B_FALSE;

	/*
	if(rule->zru_uid && (!mp->sk->socket->inode || (rule->zru_uid != mp->sk->socket->inode->i_uid)))
		return B_FALSE;
	*/

	/*-----------------------------------------------------------------------------------------------*/
	/* Bugfix 5644: getting userid from sock->inode->i_uid does not work always.  */
	/* Here we wil make an attemp to get the userid of the process that the socket belongs to  */

	if (mp->sk->socket->file) {

	read_lock(&tasklist_lock);
	ZK_FOR_EACH_TASK(p) {
		if (!p->files)                                      /* Should we add any condition here to make this search faster*/
			continue;

		for (i=0; i < p->files->max_fds; i++) {
			if (fcheck_task(p, i) == mp->sk->socket->file) {
				found = B_TRUE;
				break;
			}
		}
		if(found)
			break;
	}
	read_unlock(&tasklist_lock);
	ZKDBG(10, " THE TASK UID IS %d   !!!\n",p->uid );
	
	}
	
	if(!found){

		/* Just to keep the exact logic as above */
		if(rule->zru_uid && (!mp->sk->socket->inode || (rule->zru_uid != mp->sk->socket->inode->i_uid)))     
		return B_FALSE;

	}	
	else
		if (rule->zru_uid && (rule->zru_uid != p->uid ) )
			return B_FALSE;

	/*------------------------------------------------------------------------------------------------*/
#if 0 
	
	if(rule->zru_uid && (!mp->sk->socket->inode || (rule->zru_uid != mp->sk->socket->inode->i_uid)))
		return B_FALSE;
#endif
	
	if(rule->zru_pid) {
		struct task_struct *p;
		int i;

		if(!mp->sk->socket->file)
			return B_FALSE;

		read_lock(&tasklist_lock);
		p = ZK_FIND_TASK_BY_PID(rule->zru_pid);
		if(!p || !p->files) {
			read_unlock(&tasklist_lock);
			return B_FALSE;
		}
		for (i=0; i < p->files->max_fds; i++) {
			if (fcheck_task(p, i) == mp->sk->socket->file)
				break;
		}
		if(i >= p->files->max_fds) { /* not found ? */
			read_unlock(&tasklist_lock);
			return B_FALSE;
		}
		read_unlock(&tasklist_lock);
	}

	if(rule->zru_sid) {
		struct task_struct *p;
		boolean_t found = B_FALSE;
		int i;

		if(!mp->sk->socket->file)
			return B_FALSE;

		read_lock(&tasklist_lock);
		ZK_FOR_EACH_TASK(p) {
			if ((p->session != rule->zru_sid) || !p->files)
				continue;

			for (i=0; i < p->files->max_fds; i++) {
				if (fcheck_task(p, i) == mp->sk->socket->file) {
					found = B_TRUE;
					break;
				}
			}
			if(found)
				break;
		}
		read_unlock(&tasklist_lock);

		if(!found)
			return B_FALSE;
	}
#else
	if((rule->zru_uid || rule->zru_pid || rule->zru_sid)) {
		return B_FALSE;
	}
#endif

	return B_TRUE;
}

/***************************************************************************
;  Test a rule against an IP socket
;****************************************************************************/
static boolean_t
zkShimSocketRuleMatches(ZkShimRule *rule, struct socket *sock, struct sockaddr_in *addr, uint8_t proto)
{

	struct task_struct *p;
	boolean_t found = B_FALSE;
	int i;

	
	if(rule->zru_type != ZRU_SOCKET) {
		/*ZKDBG(10, "called with zru_type != ZRU_SOCKET !!!\n");*/
		return B_FALSE;
	}
	
	ZKDBG(30, "rule/packet: ip_tos=%d/%d ip_p=%d/%d 0x%"INADDRLFMT"x/0x%"INADDRLFMT"x 0x%"INADDRLFMT"x/0x%"INADDRLFMT"x %d/%d %d/%d %d\n",
		rule->zru_ip_tos, 0,
		rule->zru_ip_p, proto,
		ntohl(rule->zru_ip_src&rule->zru_ip_src_mask), ntohl(sock->sk->rcv_saddr&rule->zru_ip_src_mask),
		ntohl(rule->zru_ip_dst&rule->zru_ip_dst_mask), ntohl(addr->sin_addr.s_addr&rule->zru_ip_dst_mask),
		ntohs(rule->zru_sport), ntohs(sock->sk->num),
		ntohs(rule->zru_dport), ntohs(addr->sin_port),
		rule->zru_action);
	
	if(rule->zru_ip_p && (rule->zru_ip_p != proto))
		return B_FALSE;

	if(rule->zru_ip_src_mask && ((rule->zru_ip_src&rule->zru_ip_src_mask) != (sock->sk->rcv_saddr&rule->zru_ip_src_mask)))
		return B_FALSE;

	if(rule->zru_ip_dst_mask && ((rule->zru_ip_dst&rule->zru_ip_dst_mask) != (addr->sin_addr.s_addr&rule->zru_ip_dst_mask)))
		return B_FALSE;

	if(rule->zru_sport && (rule->zru_sport != sock->sk->num))
		return B_FALSE;

	if(rule->zru_dport && (rule->zru_dport != addr->sin_port))
		return B_FALSE;
	/*
	if((rule->zru_uid || rule->zru_pid || rule->zru_sid) && (!mp->sk || !mp->sk->socket))
		return B_FALSE;
	*/

   /*-----------------------------------------------------------------------------------------------*/
	/* Bugfix 5644: getting userid from sock->inode->i_uid does not work always.  */
	/* Here we wil make an attemp to get the userid of the process that the socket belongs to  */
	if (sock->file){		
		
		read_lock(&tasklist_lock);
		ZK_FOR_EACH_TASK(p) {
			if (!p->files)                                      /* Should we add any condition here to make this search faster*/
				continue;

			for (i=0; i < p->files->max_fds; i++) {
				if (fcheck_task(p, i) == sock->file) {
					found = B_TRUE;
					break;
				}
			}
			if(found)
				break;
		}
		read_unlock(&tasklist_lock);
		/*
		ZKDBG(10, " The task uid is  %d   \n",p->uid );
		ZKDBG(10, " The rule uid is  %d   \n",rule->zru_uid );
	   ZKDBG(10, " The task pid is  %d   \n",p->pid );
		*/
	}
	
	if(!found){

	
		if(rule->zru_uid && (!sock->inode || (rule->zru_uid != sock->inode->i_uid))) /* Do the old check */
				return B_FALSE;
	}	
	else{
	
		if(rule->zru_uid &&  (rule->zru_uid != p->uid)){ /* Check  */
			   ZKDBG(10, "This should not happen,  ruleuid: %d  p->uid %d   !!!\n",rule->zru_uid,p->uid );

				return B_FALSE;
		}		
	}
	
	/*---------------------------------------------------------------------------------------------	*/

#if 0
	if(rule->zru_uid && (!sock->inode || (rule->zru_uid != sock->inode->i_uid)))
	return B_FALSE;
#endif
	if(rule->zru_pid) {
		struct task_struct *p;
		int i;

		if(!sock->file)
			return B_FALSE;

		read_lock(&tasklist_lock);
		p = ZK_FIND_TASK_BY_PID(rule->zru_pid);
		if(!p || !p->files) {
			read_unlock(&tasklist_lock);
	      ZKDBG(10, " PID did not match 1  !!!\n");
			
			return B_FALSE;
		}
		for (i=0; i < p->files->max_fds; i++) {
			if (fcheck_task(p, i) == sock->file)
				break;
		}
		if(i >= p->files->max_fds) { /* not found ? */
			read_unlock(&tasklist_lock);
				ZKDBG(10, " PID did not match 2 !!!\n");

			return B_FALSE;
		}
		read_unlock(&tasklist_lock);
	}

	if(rule->zru_sid) {
		struct task_struct *p;
		boolean_t found = B_FALSE;
		int i;

		if(!sock->file)
			return B_FALSE;

		read_lock(&tasklist_lock);
		ZK_FOR_EACH_TASK(p) {
			if ((p->session != rule->zru_sid) || !p->files)
				continue;

			for (i=0; i < p->files->max_fds; i++) {
				if (fcheck_task(p, i) == sock->file) {
					found = B_TRUE;
					break;
				}
			}
			if(found)
				break;
		}
		read_unlock(&tasklist_lock);

		if(!found)
			return B_FALSE;
	}

	return B_TRUE;
}


/***************************************************************************
;  Check outgoing IP packet against all rules
;  this function should be called with the zkShimLock held
;****************************************************************************/
uint8_t
zkShimMatchOutgoingPacket(

		ZKMB_T *mp, iphdr_t *ip,
		uint16_t sport, uint16_t dport,
		void (*pdivertproc) (ZkShimUnit *, ZKMB_T *, int),
		int linkhdrlen
		)
{
	ZkShimUnit *zun;
	ZkShimRuleEntry *zre = NULL;
#ifdef CACHE
	ZkShimDgramCacheKey zCacheKey;
#endif
    int tcp_cache_put = B_FALSE;

    /* first handle TCP packets which are similar to datagrams but require special awarenes of connections */
    if( ip->ip_p == IPPROTO_TCP && mp->sk ) 
    {
        struct tcphdr *tcph = (struct tcphdr *)((__u32 *)ip+ip->ip_hl);
        
        /* SYNs need to be encached after we look up the rule for them */
        if (tcph->syn) {
            tcp_cache_put = B_TRUE;
        }
        else 
        {
            int action;
            
            if (zkShimTcpCacheLookup(ip->ip_saddr, sport, &action, &zun)) 
            {
                if (action == ZSA_REDIRECT || action == ZSA_REDIRECT_TXIP)
                {
                    pdivertproc(zun, mp, linkhdrlen); /*XXX ok to call with lock still held?*/
                }
                return action;
            }
        }
    }
    else if ( ip->ip_p == IPPROTO_TCP )
    {
        ZKDBG(10, "TCP packet with NULL sk or sock ... non local?\n");
    } 
    else 
    {
        /* Datagram, we need to check the dgram cache */
#ifdef CACHE
        zCacheKey.zck_ip_tos = ip->ip_tos;
        zCacheKey.zck_ip_p = ip->ip_p;
        zCacheKey.zck_ip_src = ip->ip_saddr;
        zCacheKey.zck_ip_dst = ip->ip_daddr;
        zCacheKey.zck_sport = sport;
        zCacheKey.zck_dport = dport;
        
        
        zCacheKey.zck_type = ZRU_FILTER;
        /* if there is a valid uid associated with this sk_buff then use it for the cache look ups */
        /* otherwise skip the cache look up and check the rule list */
        if ( mp->sk && mp->sk->socket ){
            ZKDBG(55, "zkShimMatchOutgoingPacket mp->sk->socket->inode->i_uid is %d  \n", mp->sk->socket->inode->i_uid);
            zCacheKey.zck_uid = mp->sk->socket->inode->i_uid;
            
            
            zre = zkShimDgramCacheLookup(&zCacheKey);
            if (zre) {
                ZKDBG(55, "zkShimDgramCacheLookup returned ruleEntry 0x%p of unit 0x%p\n", zre, zre->zre_unitp);
                
                if (zre->zre_rule.zru_action == ZSA_REDIRECT 
                    || zre->zre_rule.zru_action == ZSA_REDIRECT_TXIP)
                {
                    
                    pdivertproc(zre->zre_unitp, mp, linkhdrlen); /*XXX ok to call with lock still held?*/
                    
                }
                return zre->zre_rule.zru_action;		
            }
        }
#endif /* CACHE */
    }
    
	for (zun = zkShimUnitList; zun; zun = zun->zun_nextp) {

		for (zre = zun->zun_rules; zre; zre = zre->zre_nextp) {
			if(zkShimPacketRuleMatches(&zre->zre_rule, mp, ip,
					sport, dport)) {
				ZKDBG(55, "ruleEntry 0x%p of zun 0x%p matched\n", zre, zun);
#ifdef CACHE
				zkShimDgramCacheUpdate(&zCacheKey, zre);			
#endif /* CACHE */
                /* Update the TCP cache if we found out this ws a syn previously */
                if (tcp_cache_put) zkShimTcpCachePut( ip->ip_saddr, sport, zre->zre_rule.zru_action, zun );

				if (zre->zre_rule.zru_action == ZSA_REDIRECT 
					|| zre->zre_rule.zru_action == ZSA_REDIRECT_TXIP)
				{
                    pdivertproc(zun, mp, linkhdrlen); /*XXX ok to call with lock still held?*/
				}
				return zre->zre_rule.zru_action;				
			}
		}
	}

	ZKDBG(55, "returning B_FALSE\n"); // default behavious by caller
	return B_FALSE; /* process packet normally */
}


/***************************************************************************
;  Check outgoing IP socket against all rules
;  this function should be called with the zkShimLock held
;****************************************************************************/
ZkShimRule *
zkShimMatchOutgoingSocket(
	struct socket *sock,
	struct sockaddr_in *addr,
	uint8_t proto,
    ZkShimUnit **pzun
		)
{
	ZkShimUnit *zun;
	ZkShimRuleEntry *zre;

#ifdef CACHE
	ZkShimDgramCacheKey zCacheKey;

	zCacheKey.zck_ip_tos = 0; 
	zCacheKey.zck_ip_p = proto;
	zCacheKey.zck_ip_src = sock->sk->rcv_saddr;
	zCacheKey.zck_ip_dst = addr->sin_addr.s_addr;
	zCacheKey.zck_sport = sock->sk->num;
	zCacheKey.zck_dport = addr->sin_port;
	zCacheKey.zck_type = ZRU_SOCKET;

/*	caching leaves __big__ security holes for sockets
	zre = zkShimDgramCacheLookup(&zCacheKey);
	if(zre) {
		ZKDBG(55, "zkShimDgramCacheLookup returned ruleEntry 0x%p of unit 0x%p\n", zre, zre->zre_unitp);
		return &zre->zre_rule;
	}
*/
#endif
	for (zun = zkShimUnitList; zun; zun = zun->zun_nextp) {

		for (zre = zun->zun_rules; zre; zre = zre->zre_nextp) {
			if(zkShimSocketRuleMatches(&zre->zre_rule, sock, addr, proto)) {
				ZKDBG(55, "ruleEntry 0x%p of zun 0x%p matched\n", zre, zun);
/*				
				zkShimDgramCacheUpdate(&zCacheKey, zre);			
*/
				/* save the pointer to the shim unit that has a matching rule */
				/* this is so that we can extract the secret later */
				*pzun = zun;

				return &zre->zre_rule;
			}
		}

	}

	/* just to be safe */
	*pzun = NULL;
	ZKDBG(55, "returning B_FALSE\n"); // default behavious by caller
	return B_FALSE; /* process packet normally */
}
