/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

 */

#include "zkshim_priv.h"

#if defined(DEBUG) || defined(_DEBUG)
#  define _INLINE
#else
#  define _INLINE inline
#endif

#define HASH_SIZE 1024

typedef struct cache_iterator 
{
    unsigned long bucket;
    ZkShimTcpCacheEntry *entry;
} cache_iterator;

static struct list_head zkShimTcpCacheHash[HASH_SIZE];

static KRWLOCK_T zkShimTcpCacheLock;


/* (SVR4 elf hash fn cribbed from Practical Algorithms for Programmers pp 71) */
/* There are probably more straightforward ways to do this, but this is the   */
/* hash fn I generally use, fix it if you have a better one  --ol             */
static _INLINE unsigned long 
hashcode( uint32_t saddr, uint16_t sport)
{
    unsigned long h = 0, g;
    unsigned int i = 0;
    unsigned char buf[sizeof( uint32_t ) + sizeof( uint16_t )];

    memcpy( &buf[0], &saddr,  sizeof( uint32_t ) );
    memcpy( &buf[sizeof( uint32_t )], &sport, sizeof( uint16_t ) );

    /* unroll? */
    while( i < sizeof( uint32_t ) + sizeof( uint16_t ) )
    {
        h = ( h << 4 ) + buf[i++];
        if ( g = h & 0xF0000000 )
            h ^= g >> 24;
        h &= ~g;        
    }
    return h % HASH_SIZE;
}

void 
zkShimTcpCacheInit ( void )
{
    unsigned int i;
    for ( i = 0; i < HASH_SIZE; i++ ) INIT_LIST_HEAD( &zkShimTcpCacheHash[i] );    
    INIT_LOCK( &zkShimTcpCacheLock, "zkshim zkShimTcpCacheLock" );
}

/* Impl, find an entry. Call w/ read lock held */
static _INLINE ZkShimTcpCacheEntry *
FindTcpEntry ( uint32_t saddr, uint16_t sport )
{
    struct list_head *lh;
    ZkShimTcpCacheEntry *ent;
    unsigned long bucket;
    
    bucket = hashcode(saddr, sport);

    for ( lh = zkShimTcpCacheHash[bucket].next; lh != &zkShimTcpCacheHash[bucket]; lh = lh->next) 
    {
        ent = list_entry(lh, ZkShimTcpCacheEntry, tcpcache_list);
        if (ent->saddr == saddr && ent->sport == sport) return ent;
    }
    return NULL;
}

/* These functions try to make the hashtable structure function sanely as a list
 * so we can iterate through the entries etc.
 * The parameter holds state info and consequently gets munged.
 */
static _INLINE ZkShimTcpCacheEntry *
FirstTcpEntry ( cache_iterator *it )
{
    struct list_head *lh;
    it->bucket = 0; it->entry = NULL;
    for ( it->bucket = 0; it->bucket < HASH_SIZE; it->bucket++ )
    {
        lh = zkShimTcpCacheHash[it->bucket].next;
        if ( lh == &zkShimTcpCacheHash[it->bucket] ) continue;
        it->entry = list_entry( lh, ZkShimTcpCacheEntry, tcpcache_list );
        return it->entry;
    }
    it->entry = NULL;
    return it->entry;
}

static _INLINE ZkShimTcpCacheEntry *
NextTcpEntry ( cache_iterator *it )
{
    struct list_head *lh;

    /* how sane are you? */
    if (it->entry == NULL) return NULL;
    if (it->bucket >= HASH_SIZE) return NULL;

    lh = &it->entry->tcpcache_list;
    do {
        lh = lh->next;

        if (lh != &zkShimTcpCacheHash[it->bucket] ) {
            it->entry = list_entry( lh, ZkShimTcpCacheEntry, tcpcache_list );
            return it->entry;
        } 
    } while ( ++it->bucket < HASH_SIZE && ( lh = &zkShimTcpCacheHash[it->bucket] ) );

    it->entry = NULL;
    return it->entry;
}

/* return TRUE for hit FALSE for miss. The two out args will contain the result of a cache hit  */
int
zkShimTcpCacheLookup ( uint32_t saddr, uint16_t sport, int *out_action, ZkShimUnit **out_zun )
{
    ZkShimTcpCacheEntry *ent;
    int rval = B_FALSE;

    if (!out_zun || !out_action) goto die_mofo;

    READ_LOCK( &zkShimTcpCacheLock );
    ent = FindTcpEntry( saddr, sport );
    if (ent) {
        *out_action = ent->action;
        *out_zun = ent->zun;        
        rval = B_TRUE;
    } 
    READ_UNLOCK( &zkShimTcpCacheLock );
 die_mofo:
    return rval;
}

/* Add an entry to the cache */
void
zkShimTcpCachePut ( uint32_t saddr, uint16_t sport, int action, ZkShimUnit *zun )
{
    ZkShimTcpCacheEntry *ent;
    unsigned long f;

    READ_LOCK( &zkShimTcpCacheLock );
    ent = FindTcpEntry( saddr, sport );
    READ_UNLOCK( &zkShimTcpCacheLock );
    if (ent) {
        /* overwrite matches, set new action in entry */
        WRITE_LOCK( &zkShimTcpCacheLock, f );
        ent->action = action;
        ent->zun = zun;
        WRITE_UNLOCK( &zkShimTcpCacheLock, f );
    } else {
        /* insert unmatched entries */
        unsigned long bucket = hashcode( saddr, sport );
        ent = (ZkShimTcpCacheEntry *) kmalloc( sizeof(ZkShimTcpCacheEntry), GFP_ATOMIC );
        ent->saddr = saddr; 
        ent->sport = sport;
        ent->action = action;
        ent->zun = zun;
        WRITE_LOCK( &zkShimTcpCacheLock, f );
        list_add( &ent->tcpcache_list, &zkShimTcpCacheHash[bucket] );
        WRITE_UNLOCK( &zkShimTcpCacheLock, f );
    }
}

/* Handle shutting down a ZkShimUnit
 */
void
zkShimTcpCacheShutdownUnit ( ZkShimUnit *zun )
{
    ZkShimTcpCacheEntry *ent, *ent2;
    unsigned long f;
    cache_iterator ci;

    WRITE_LOCK( &zkShimTcpCacheLock, f );

    ent = FirstTcpEntry(&ci);

    while ( ent )
    {
        if ( ent->zun == zun ) 
        {
            ent2 = NextTcpEntry(&ci);
            list_del(ent);
            kfree(ent);
            ent = ent2;
        } else {
            ent = NextTcpEntry(&ci);
        }
    }
    WRITE_UNLOCK( &zkShimTcpCacheLock, f );    
}

void 
zkShimTcpCacheCleanup ( void )
{
    /* This wil probably never actually happen since ref counting has kept
       us locked if their are any open devices, but it never hurts to be 
       anal.
    */
    ZkShimTcpCacheEntry *ent, *ent2;
    unsigned long f;
    cache_iterator ci;

    WRITE_LOCK( &zkShimTcpCacheLock, f );

    ent = FirstTcpEntry(&ci);
    while ( ent )
    {
        ent2 = NextTcpEntry(&ci);
        list_del(ent);
        kfree(ent);
        ent = ent2;
    }
    WRITE_UNLOCK( &zkShimTcpCacheLock, f );    
    
    DESTROY_LOCK( &zkShimTcpCacheLock )
}


#ifdef CONFIG_PROC_FS
/***
 * Handler for proc entry for the TCP cache. 
 * More or less the same as the other proc handlers.
 */
static int
zkShimTcpCacheProcInfo(char *buffer, char **start, off_t offset, int length, int reset)
{
	int len=0;
	off_t pos=0;
	char temp[129];
	ZkShimTcpCacheEntry *ent;
    cache_iterator ci;
	char action='0';


	pos = 128;

	if (offset<128)	{
		/* change to snprintf */
		sprintf(buffer,"%-127s\n", "IP\t\tPort\t\tAction\t\tUnit");

		len = 128;
  	}
	
  	
	READ_LOCK (&zkShimTcpCacheLock);
	start_bh_atomic();


    for ( ent = FirstTcpEntry(&ci); ent != NULL; ent = NextTcpEntry(&ci) ) 
    {
        /*
         *	Spin through entries until we are ready
         */
        pos += 128;
        
        if (pos <= offset) {
            len = 0;
            continue;
        }
        

        if ( ent->action == ZSA_ALLOW )
            action='A';
        else if ( ent->action == ZSA_DENY )
            action='D';
        else if ( ent->action == ZSA_REDIRECT )
            action='R';
        else if ( ent->action == ZSA_REDIRECT_TXIP )
            action='T';
        
        
        sprintf(temp, "%u\t%hu\t\t%c\t\t%08lX",
				ent->saddr, ent->sport, action, ent->zun);
        
        sprintf(buffer+len,"%-127s\n",temp);
        len += 128;
        if (pos >= offset+length)
            goto done;
    }
    
 done:
	end_bh_atomic();
	READ_UNLOCK (&zkShimTcpCacheLock);
  	
  	*start = buffer+len-(pos-offset);
  	len = pos-offset;
  	if (len>length)
  		len = length;
  	return len;
} 

#endif
