/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

 */

#ifndef __ZKSHIM_TCP_CACHE_INCLUDED
#define __ZKSHIM_TCP_CACHE_INCLUDED
#include "zkshim_priv.h"

/* This sturcture maintains the action of a shim rule across
 * all packets in a TCP stream. It caches the action of the original
 * rule that matched this stream, and the zkshim unit associated with
 * that action. The reason for this is that rules applied to TCP should
 * apply to the _stream_itslf_ not the individual datagrams which make up
 * the stream. Consequently if I close a socket, any other packets arriving
 * for that port while it's in FIN_WAITn should also be delivered to me 
 * (or denied etc.). Also rule transitions underneath open TCP streams should
 * _not_ affect those streams ... new rules should be applied to subsequent 
 * stream connections not datagrams which make up an existing connection.
 * So, while this is a cache-like mecanism, it is also a core piece of functionality.
 * (unlike the datagram cache which is really just for performance).
 *
 * Also note that since AIP 1.0 does not protect against packets being sent to us after 
 * a connection has terminated (until the route dies or a specified timeout is reached)
 * there is no harvesting of these entries at the moment, doing so would mean that we would
 * need to be smart about FIN and RST and start a timeout on shutdown streams. Even then I'm
 * not entirely convinced we could make it secure, and the largest amount of memory we will 
 * consume is sizeof(ZkShimTcpCacheEntry) * 2^16 ports * # of net interfaces
 *
 * --ol
 */
typedef struct ZkShimTcpCacheEntry {
    struct list_head tcpcache_list;
    uint32_t saddr;
    uint16_t sport;
    uint8_t action;
    /* no args since these are always
       asumed to be filtering rules */
    ZkShimUnit *zun;
} ZkShimTcpCacheEntry;


void 
zkShimTcpCacheInit ( void );

int
zkShimTcpCacheLookup ( uint32_t saddr, uint16_t sport, int *out_action, ZkShimUnit **out_zun );

void
zkShimTcpCachePut ( uint32_t saddr, uint16_t sport, int action, ZkShimUnit *zun );

void
zkShimTcpCacheShutdownUnit ( ZkShimUnit *zun );

void 
zkShimTcpCacheCleanup ( void );

#ifdef CONFIG_PROC_FS
static int
zkShimTcpCacheProcInfo(char *buffer, char **start, off_t offset, int length, int reset);
#endif

#endif /*__ZKSHIM_TCP_CACHE_INCLUDED*/
