// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "RLinuxPacketHandler.h"
#include <DZkError.h>
#include <RShimInterface.h>
#include <errno.h>
#include <signal.h>
#include <sys/poll.h>
#include "RDnsMessage.h"

bool RLinuxPacketHandler::bRunning = ZK_FALSE;

//---------------------------------------------------------------------------
//
RLinuxPacketHandler::RLinuxPacketHandler()
    : thread(NULL)
{
    m_DnsFilter.SetOwner(this);
}

//---------------------------------------------------------------------------
//
RLinuxPacketHandler::~RLinuxPacketHandler()
{
	if (bRunning) Terminate();
}

//---------------------------------------------------------------------------
//
// Return ZK_TRUE on success, throws std::runtime_error on error 
// Only call this once, it _WILL_ start multiple threads which is a deeply bad thing.
int  RLinuxPacketHandler::Init() throw ( RZkException )
{
//    DPacket::Init();
	if (!bRunning) {
		DZkError err;
#ifndef POLL_SHUTDOWN        
        signal(SIGUSR1, &signalHandler);
#endif
		RShimInterface::getShimInterface().Init();
		if (!zkThreadCreate(&err, &thread, &RLinuxPacketHandler::PacketThread, this)) 
			throw RZkException("RLinuxPacketHandler::Init()", "unable to start shim packet thread", IDS_FAILED_TO_CREATE_THREAD);
		else return bRunning = ZK_TRUE;
	}
	else return bRunning;
}

//---------------------------------------------------------------------------
//
void  RLinuxPacketHandler::Terminate()
{
    char c = '~';
	if (bRunning) {
		DZkError err;
		void * ret;
		ZKTRACE0("RLinuxPacketHandler: shutting down shim.\n");
		bRunning = ZK_FALSE;

        // Signal packet thread to shutdown.

#ifdef POLL_SHUTDOWN
        ZKTRACE0("Signaling packethandler shutdown.\n");
        write(m_pipe.GetWrite(), &c, 1);
#endif
		RShimInterface::getShimInterface().Shutdown();
#ifndef POLL_SHUTDOWN
        kill(0, SIGUSR1);
#endif
		if (thread && !zkThreadJoin(&err, &thread, &ret))
			throw RZkException("RLinuxPacketHandler::Terminate()", "unable to join shim packet thread", IDS_FAILED_TO_CREATE_THREAD);
	}
}

//---------------------------------------------------------------------------
//
DnsFilter& RLinuxPacketHandler::GetDnsFilter() 
{
    return m_DnsFilter;
}


//---------------------------------------------------------------------------
//

ZK_THREAD_CALLBACK void *RLinuxPacketHandler::PacketThread(void * arg)
{
	RLinuxPacketHandler *handler = (RLinuxPacketHandler *)arg;
	ZKASSERT(handler);
	unsigned char packet[FRAME_LENGTH];
	ssize_t length;
    
    do {
#ifdef POLL_SHUTDOWN
        struct pollfd fds[2];
        int rval;
    
        fds[0].fd = handler->m_pipe.GetRead();
        fds[0].events = POLLIN;
        fds[0].revents = 0;
        fds[1].fd = RShimInterface::getShimInterface().GetShimFD();
        fds[1].events = POLLIN;
        fds[1].revents = 0;

    retry: 
        if ((rval = poll(fds, 2, -1)) == -1) {
            if (errno == EINTR || rval == 0) goto retry;
            else ZKTRACE0("!!!Poll error\n");
        }

        if (fds[0].revents & POLLIN) {
            // told to shutdown
                ZKTRACE0("RLinuxPacketHandler Shutting down packet thread\n");            
                break;            
        }
        
        // We have data to read
#endif        
        try {
            length = RShimInterface::getShimInterface().ReadPacket(packet, FRAME_LENGTH);
        } catch (RZkException &e)
        {
            // this condition exists so we can clearly shutdown by signalling
          if (bRunning == ZK_FALSE) {
                ZKTRACE0("RLinuxPacketHandler Shutting down packet thread\n");            
                break;
          }
            ZKTRACE1("RLinuxPacketHandler::PacketThread bad read on shim device [%s]", e.GetFormattedString().c_str());
            continue;
        }

#ifdef DUMP_OUTGOING
        DPacket::DumptcpHeader((ZkTCPHeader*)packet);
#endif
		// we re-inject packets if to_send... returns true.
		if (handler->To_send_OR_not_to_send(0, (ZkBuffer *)&packet, ((ZkBuffer *)packet)->Tcp.IP.Protocol)) {
			try {
                // Need to inject a header checksum onto the packet. The linux shim has
                // captured it prior to ip csum calculation.
                ((ZkBuffer *)&packet)->Tcp.IP.HeaderChkSum = zkIPHeaderChecksum(&((ZkBuffer *)&packet)->Tcp.IP);
				RShimInterface::getShimInterface().WritePacket(packet, length);
			} catch (RZkException &e) {
				ZKTRACE1("RLinuxPacketHandler::PacketThread bad write on shim device [%s]", e.GetFormattedString().c_str());
				continue;
				// XXX Non fatal. This should be improved.
			}
		}
#ifdef ONE_PACKET
	} while (0);
	RShimInterface::getShimInterface().Shutdown();
#else
	} while (1);
#endif	
	return NULL;
}

//---------------------------------------------------------------------------
//

bool RLinuxPacketHandler::InjectPacketIntoProtocolStack(const ZkBuffer &buf, const unsigned int &nDevice)
{
    (void)nDevice; // Shut up gcc
    try {
#ifdef DUMP_INCOMING
        DPacket::DumptcpHeader((ZkTCPHeader*)&buf);
#endif
        RShimInterface::getShimInterface().WritePacket(
           reinterpret_cast<unsigned char *>(&const_cast<ZkBuffer &>(buf))
           , ntohs(buf.Tcp.IP.TotalLength)
           );
    } catch (RZkException &e) {
        ZKTRACE1("RLinuxPacketHandler::PacketThread bad write on shim device [%s]", e.GetFormattedString().c_str());
        return false;
        // XXX Non fatal. This should be improved.
    }
    return true;
}

//---------------------------------------------------------------------------
//

#ifndef POLL_SHUTDOWN
void RLinuxPacketHandler::signalHandler(int)
{
    ZKTRACE0("SIGUSR1 recieved by RLinuxPacketHandler");
}
#endif

//---------------------------------------------------------------------------
//

bool RLinuxPacketHandler::HandleOutGoingUDPPacket( const int nDevice, ZkBuffer *pbuf )
{
    // This adds extra DNS functionality to Linux to handle sendmail MX lookups
    if ( htons(DNS_PORT) == pbuf->Udp.DstPort || htons(DNS_PORT) == pbuf->Udp.SrcPort ) {
        // 8 is the length of a udp header. (sizeof doesn't work 'cause the struct is f*cked) --ol
        // This looks scary but it isn't. We scan the buffer for any ocurances of ALL or MX which are both DNS question types
        // that can cause MX RRs to be returned (so we need to hijack those bastards)
        if (
            RDnsMessage::ScanBufferForShort( RDnsMessage::ALL, reinterpret_cast<char*>(pbuf) + sizeof(ZkUDPHeader), ntohs(pbuf->Udp.UdpLength) - 8 )
            || RDnsMessage::ScanBufferForShort( RDnsMessage::MX, reinterpret_cast<char*>(pbuf) + sizeof(ZkUDPHeader), ntohs(pbuf->Udp.UdpLength) - 8 )
        ) {
            ZKTRACE0("Got a DNS packet with a MX or ALL lookup in it (I think).\n");
            // Turn it into a usable data structure.
            RDnsMessage DnsData(reinterpret_cast<char*>(pbuf) + sizeof(ZkUDPHeader), ntohs(pbuf->Udp.UdpLength) - 8);
            RDnsMessage::DnsQuestion *q;
            bool found = false;
            size_t i;
            
            for ( i = 0; i < DnsData.GetHeader().GetQDCount(); i++ ) {
                q = DnsData.GetQuestion(i);
                if ( q && ( q->GetType() == RDnsMessage::ALL || q->GetType() == RDnsMessage::MX ) ) {                    
                    // Got 'em
                    ZKTRACE0("Got a DNS packet with a MX or ALL lookup in it (I'm certain).\n");
                    found = 1;
                    break;
                }                
            }
            if ( found && q->GetClass() == RDnsMessage::INTERNET ) {
                // Now this starts to get phun.
                RDnsMessage DnsReply;
                // Setup the general DNS header.
                RDnsMessage::DnsHeader &hresponse =  DnsReply.GetHeader();
                RDnsMessage::DnsHeader &hrequest =  DnsData.GetHeader();

                hresponse.SetID( hrequest.GetID() ); // Copy ID
                hresponse.SetQR( 1 ); // 1 == response 0 == question
                hresponse.SetOpcode( hrequest.GetOpcode() ); // Copy opcode
                hresponse.SetAA( 0 ); // We are not authoratative
                hresponse.SetTC( 0 ); // no truncation
                hresponse.SetRD( hrequest.GetRD() );
                hresponse.SetRA( 0 ); // We do not support recursion
                hresponse.z = 0 ;
                hresponse.SetRCode( 0 ); // Green light status.
               
                DnsReply.AddQuestion( *q ); // We're responding to q

                RDnsMessage::DnsResourceRecord rr = RDnsMessage::DnsResourceRecord();
                
                rr.SetName( q->GetName() );
                rr.SetClass( RDnsMessage::INTERNET );
                rr.SetTtl( 0 );                 

                if ( q->GetType() == RDnsMessage::ALL ) {
                    // If the request is type ALL we need to send back some more info.
                    // one way to do this better might be to just include an MX and a authorative nameserver
                    // entry for the wormhole. We'll see.
                    rr.SetType( RDnsMessage::NS );
                    ((RDnsMessage::NSResourceData *)rr.GetResourceData())->SetNameServer("freedomized.safedns");
                    
                    DnsReply.AddAnswer( rr );
                    
                    rr.SetType( RDnsMessage::A );
                    ((RDnsMessage::AResourceData *)rr.GetResourceData())->SetAddress( inet_addr("127.0.0.1") );

                    DnsReply.AddAnswer( rr );
                }
                // Add MX response
                rr.SetType( RDnsMessage::MX );
                ((RDnsMessage::MXResourceData *)rr.GetResourceData())->SetPreference(10); // 10 is arbitrary, we just need a number
                ((RDnsMessage::MXResourceData *)rr.GetResourceData())->SetExchange("freedomized.safedns");
                
                DnsReply.AddAnswer( rr );

                // Add additional section for our phony freedomized.safedns address.
                rr.SetName( "freedomized.safedns" );
                rr.SetType( RDnsMessage::A );                
                ((RDnsMessage::AResourceData *)rr.GetResourceData())->SetAddress( inet_addr("127.0.0.1") );
                
                DnsReply.AddAdditional( rr );

                // Ok. It's built. Now we need to TCP/IPify the data.
                
                ZkBuffer OutputBuffer;
                uint32_t tmp32;
                uint16_t tmp16;
                size_t length;
                // Slap on the old header.
                memcpy(&OutputBuffer, pbuf, sizeof(ZkUDPHeader));

                // swap src and dst IPs
                tmp32 = OutputBuffer.Udp.IP.SrcIpAddress;
                OutputBuffer.Udp.IP.SrcIpAddress = OutputBuffer.Udp.IP.DstIpAddress;
                OutputBuffer.Udp.IP.DstIpAddress = tmp32;

                length = DnsReply.Serialize( ((char*)&OutputBuffer) + sizeof(ZkUDPHeader), ZKBUFFER_LEN );

                OutputBuffer.Udp.IP.TotalLength = htons( length + sizeof(ZkUDPHeader) );
                OutputBuffer.Udp.UdpLength = htons( length + 8 ); // 8 is length of UDP header only

                // swap src and dst ports
                tmp16 = OutputBuffer.Udp.SrcPort;
                OutputBuffer.Udp.SrcPort = OutputBuffer.Udp.DstPort;
                OutputBuffer.Udp.DstPort = tmp16;

                // calc csums
                OutputBuffer.Udp.UdpCheckSum = 0;
                OutputBuffer.Udp.UdpCheckSum = zkUDPHeaderChecksum(&OutputBuffer.Udp, length + 8 );
                OutputBuffer.Udp.IP.HeaderChkSum = zkIPHeaderChecksum(&OutputBuffer.Udp.IP);

                // cocked, locked, and ready to rock.
                InjectPacketIntoProtocolStack(OutputBuffer, 0);
                
                return ZK_FALSE;
            }

        } else if ( 
            RDnsMessage::ScanBufferForString("freedomized", reinterpret_cast<char*>(pbuf) + sizeof(ZkUDPHeader), ntohs(pbuf->Udp.UdpLength) - 8  ) 
            && RDnsMessage::ScanBufferForShort( RDnsMessage::A, reinterpret_cast<char*>(pbuf) + sizeof(ZkUDPHeader), ntohs(pbuf->Udp.UdpLength) - 8 )
        ) {
            // Turn it into a usable data structure.
            RDnsMessage DnsData(reinterpret_cast<char*>(pbuf) + sizeof(ZkUDPHeader), ntohs(pbuf->Udp.UdpLength) - 8);
            RDnsMessage::DnsQuestion *q;
            bool found = false;
            size_t i;
            
            for ( i = 0; i < DnsData.GetHeader().GetQDCount(); i++ ) {
                q = DnsData.GetQuestion(i);
                if ( q && ( q->GetType() == RDnsMessage::A ) ) {                    
                    // Got 'em
                    ZKTRACE0("Got a DNS packet with an A lookup in it for freedomized.safedns (I'm certain).\n");
                    found = 1;
                    break;
                }                
            }
            if ( found && q->GetClass() == RDnsMessage::INTERNET ) {
                // Now this starts to get phun.
                RDnsMessage DnsReply;
                // Setup the general DNS header.
                RDnsMessage::DnsHeader &hresponse =  DnsReply.GetHeader();
                RDnsMessage::DnsHeader &hrequest =  DnsData.GetHeader();

                hresponse.SetID( hrequest.GetID() ); // Copy ID
                hresponse.SetQR( 1 ); // 1 == response 0 == question
                hresponse.SetOpcode( hrequest.GetOpcode() ); // Copy opcode
                hresponse.SetAA( 0 ); // We are not authoratative
                hresponse.SetTC( 0 ); // no truncation
                hresponse.SetRD( hrequest.GetRD() );
                hresponse.SetRA( 0 ); // We do not support recursion
                hresponse.z = 0 ;
                hresponse.SetRCode( 0 ); // Green light status.
               
                DnsReply.AddQuestion( *q ); // We're responding to q

                RDnsMessage::DnsResourceRecord rr = RDnsMessage::DnsResourceRecord();
                
                rr.SetName( q->GetName() );
                rr.SetClass( RDnsMessage::INTERNET );
                rr.SetTtl( 0 );                 
                rr.SetType( RDnsMessage::A );                
                ((RDnsMessage::AResourceData *)rr.GetResourceData())->SetAddress( inet_addr("127.0.0.1") );
                
                DnsReply.AddAnswer( rr );

                // Ok. It's built. Now we need to TCP/IPify the data.
                
                ZkBuffer OutputBuffer;
                uint32_t tmp32;
                uint16_t tmp16;
                size_t length;
                // Slap on the old header.
                memcpy(&OutputBuffer, pbuf, sizeof(ZkUDPHeader));

                // swap src and dst IPs
                tmp32 = OutputBuffer.Udp.IP.SrcIpAddress;
                OutputBuffer.Udp.IP.SrcIpAddress = OutputBuffer.Udp.IP.DstIpAddress;
                OutputBuffer.Udp.IP.DstIpAddress = tmp32;

                length = DnsReply.Serialize( ((char*)&OutputBuffer) + sizeof(ZkUDPHeader), ZKBUFFER_LEN );

                OutputBuffer.Udp.IP.TotalLength = htons( length + sizeof(ZkUDPHeader) );
                OutputBuffer.Udp.UdpLength = htons( length + 8 ); // 8 is length of UDP header only

                // swap src and dst ports
                tmp16 = OutputBuffer.Udp.SrcPort;
                OutputBuffer.Udp.SrcPort = OutputBuffer.Udp.DstPort;
                OutputBuffer.Udp.DstPort = tmp16;

                // calc csums
                OutputBuffer.Udp.UdpCheckSum = 0;
                OutputBuffer.Udp.UdpCheckSum = zkUDPHeaderChecksum(&OutputBuffer.Udp, length + 8 );
                OutputBuffer.Udp.IP.HeaderChkSum = zkIPHeaderChecksum(&OutputBuffer.Udp.IP);

                // cocked, locked, and ready to rock.
                InjectPacketIntoProtocolStack(OutputBuffer, 0);
                
                return ZK_FALSE;
            }
        }
    }
    return DPacket::HandleOutGoingUDPPacket( nDevice, pbuf );
}
