// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include <RShimInterface.h>
#include <RShimRule.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <client.h>

//#include <clientdebug.h>
//#include <clientstl.h>


// This is a hack to get the main application pid without having
// to pass it around, basically because getpid() is nasty fubar and broken
// with LinuxThreads -ol
pid_t RShimInterface::m_globalPid = getpid();

const ZKSTD::string strZKSHIM_DEVNAME = ZKSTD::string(ZKSHIM_DEVNAME);

// Singleton instantiation thingamajig
RShimInterface &RShimInterface::getShimInterface()
{
	static RShimInterface theInstance;
	return theInstance;
}

// private constructor
RShimInterface::RShimInterface()
{ 
	shim_fd = 0;
	bInitialized = ZK_FALSE;
}

void RShimInterface::Init(const ZKSTD::string &strName) throw ( RZkException )
{
	if (!bInitialized) {
		if ((shim_fd = open(strName.c_str(), O_RDWR)) < 0) {
			throw RZkException("RShimInterface::Init",
                               strerror(errno),
                               IDS_FAILED_TO_OPEN_THE_ZKSHIM);
		}
		bInitialized = ZK_TRUE;
        InitializeSecret();
        InitMode();
	}
}

void RShimInterface::InitializeSecret() throw ( RZkException )
{
    int fd;
    ssize_t len;

    if ( ( fd = open("/dev/random", O_RDONLY) ) == -1 )
    {
        throw RZkException("RShimInterface::InitializeSecret",
                           "Unable to open /dev/random/",
                           IDS_DEFAULT_ERROR);
    }
    if ( ( len = read(fd, (void *)&m_secret.cookie, ZKSHIM_COOKIE_SIZE)) != sizeof(m_secret) )
    {
        throw RZkException("RShimInterface::InitializeSecret",
                           "Unable to read /dev/random/",
                           IDS_DEFAULT_ERROR);
    }
    if(ioctl(shim_fd, ZKIOC_ADD_SECRET, (void *)&m_secret, sizeof(m_secret)) < 0) {
        throw RZkException("RShimInterface::InitializeSecret",
                           strerror(errno),
                           IDS_DRIVER_IOCTL_FAILED);		
    }
}



// For the time being the rule set for Nym selected and User modes
// are identical. This should change at some point, since we really have
// no need to look at so many packets when no nym is selected.
void RShimInterface::NymMode() throw ( RZkException )
{
    ClearShimRules();
	ZKTRACE0("Adding Shim rules (Nym)\n");

    AddAllPacketInterceptionRules(true);
    AddAllProtocolProxyRules();
}

void RShimInterface::UserMode() throw ( RZkException )
{
    // Clear and rebuild user mode rules, we don't want to make assumptions about
    // what rules are already loaded.
	ZKTRACE0("Adding Shim rules (User)\n");

    ClearShimRules();

    AddAllPacketInterceptionRules(false);

/////////////////////////////
// POP3
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_POP3_PORT, REMOTE_POP3_PORT); 

/////////////////////////////
// SMTP
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_SMTP_PORT, REMOTE_SMTP_PORT); 

/////////////////////////////
// HTTP
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT2);
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT3);
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT4);
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT5);
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT6);
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT7);
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT8);
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT9);
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT10);
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT11);

/////////////////////////////
// SSL
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_SSL_PORT, REMOTE_SSL_PORT); 

/////////////////////////////
// NNTP
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_NNTP_PORT, REMOTE_NNTP_PORT); 

/////////////////////////////
// IRC
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT2); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT3); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT4); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT5); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT6); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT7); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT8); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT9); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT10); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT11); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT12); 

//    AddAllProtocolProxyRules();

}

void RShimInterface::InitMode() throw ( RZkException )
{
	ZKTRACE0("Adding Shim rules (Initialization)\n");
    ClearShimRules();

    // We do not have DInetD running at this point.
    AddAllPacketInterceptionRules(false);
}

void RShimInterface::AddAllProtocolProxyRules( void )
{
/////////////////////////////
// POP3
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_POP3_PORT, REMOTE_POP3_PORT); 

/////////////////////////////
// SMTP
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_SMTP_PORT, REMOTE_SMTP_PORT); 

/////////////////////////////
// HTTP
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT2);
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT3);
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT4);
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT5);
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT6);
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT7);
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT8);
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT9);
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT10);
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_HTTP_PORT, REMOTE_HTTP_PORT11);

/////////////////////////////
// NNTP
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_NNTP_PORT, REMOTE_NNTP_PORT); 

/////////////////////////////
// IRC
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT2); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT3); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT4); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT5); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT6); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT7); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT8); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT9); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT10); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT11);
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_IRC_PORT, REMOTE_IRC_PORT12); 

/////////////////////////////
// Telnet & SSH
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_TELNET_PORT, REMOTE_TELNET_PORT); 
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_TELNET_PORT, REMOTE_SSH_PORT); 

/////////////////////////////
// SSL
    AddProtocolProxyRule(IPPROTO_TCP, LOCAL_SSL_PORT, REMOTE_SSL_PORT); 
}

void RShimInterface::AddProtocolProxyRule(int protocol, uint16_t local_port, uint16_t remote_port)  throw ( RZkException )
{
	RShimRule rule;

    rule.setProto(protocol);
    rule.setType(ZRU_SOCKET);
    rule.setAction(ZSA_REDIRECT_TXIP);
    rule.setDestPort(htons(remote_port));
    rule.setPID(0);
    rule.setPortArg(htons(local_port));
    rule.setIPArg(htonl(INADDR_LOOPBACK));
    AddShimRule(rule);

    // This adds an exception for freedom's outbound connection
    rule.setIPArg(0);
    rule.setPortArg(0);
    rule.setAction(ZSA_ALLOW);
    rule.setPID(m_globalPid);
    AddShimRule(rule);
}

void RShimInterface::AddAllPacketInterceptionRules( bool AddDefaultRule )
{
	RShimRule rule;

    if ( AddDefaultRule )
    {
        // Freedom gets all packets sent by this UID (should be more selective)
        rule.setType(ZRU_FILTER);
        rule.setAction(ZSA_REDIRECT);
        AddShimRule(rule); 
    }

    rule.setType(ZRU_FILTER);
    rule.setAction(ZSA_REDIRECT);

    // Exception: Freedom does not get packets which are traversing loopback.

    rule.setSourceIP(htonl(INADDR_LOOPBACK), htonl(0xFF000000L));
    rule.setSourcePort(0);
    rule.setDestIP(htonl(INADDR_LOOPBACK), htonl(0xFF000000L));
    rule.setDestPort(0);
    rule.setAction(ZSA_ALLOW);
	AddShimRule(rule); 

    // Exception: Freedom does not get UDP packets destined for an AIP port
    // We don't actually validate IPs. If this is a security risk we can
    // not add this exception and pump the packets through RFirewall.
    // another alternative would be to dynamically add and remove these exception rules
    // whenever routes go up and down. This would be easy since these rules need to be
    // at the head of the list in any event. --ol

    rule.setSourceIP(0, 0);
    rule.setDestIP(0, 0);
    rule.setDestPort(htons(FR_DFLT_AIP_UDP_PORT));
    rule.setPID(m_globalPid);
    rule.setAction(ZSA_ALLOW);
    rule.setProto(IPPROTO_UDP);
    AddShimRule(rule);

    // Bug 5425 aka 1876
    // Non-Freedom processes not allowed to initiate connections which cause
    // unauthenticated route creation. 
    
    rule.setType(ZRU_FILTER);
    rule.setSourceIP(0, 0);
    rule.setDestIP(0, 0);
    rule.setDestPort(htons(FR_DFLT_TOKEN_SERVER_PORT));
    rule.setPID(0);
    rule.setAction(ZSA_DENY);
    rule.setProto(0); // any
    AddShimRule(rule);

    rule.setDestPort(htons(FR_DFLT_NYM_SERVER_PORT));
    AddShimRule(rule);

    rule.setDestPort(htons(FR_DFLT_NIQS_SERVER_PORT));
    AddShimRule(rule);

    rule.setDestPort(htons(FR_DFLT_FMGMAIP_SERVER_PORT)); // XXX do we still need this?
    AddShimRule(rule);

    rule.setDestPort(htons(FR_DFLT_POP_PORT));
    AddShimRule(rule);

    rule.setDestPort(htons(FR_DFLT_NMTA_PORT));
    AddShimRule(rule);

    rule.setDestPort(htons(FR_DFLT_KEYQRY_SERVER_PORT));
    AddShimRule(rule);

    // And corresponding exceptions for freedom's PID 
    rule.setAction(ZSA_REDIRECT);
    rule.setPID(m_globalPid);
        
    rule.setDestPort(htons(FR_DFLT_TOKEN_SERVER_PORT));
    AddShimRule(rule);

    rule.setDestPort(htons(FR_DFLT_NYM_SERVER_PORT));
    AddShimRule(rule);

    rule.setDestPort(htons(FR_DFLT_NIQS_SERVER_PORT));
    AddShimRule(rule);

    rule.setDestPort(htons(FR_DFLT_FMGMAIP_SERVER_PORT)); // XXX see above
    AddShimRule(rule);

    rule.setDestPort(htons(FR_DFLT_POP_PORT));
    AddShimRule(rule);

    rule.setDestPort(htons(FR_DFLT_NMTA_PORT));
    AddShimRule(rule);

    rule.setDestPort(htons(FR_DFLT_KEYQRY_SERVER_PORT));
    AddShimRule(rule);
}

void RShimInterface::Shutdown()
{
	if (bInitialized) {
		// Can't really fail (unless it's already closed)
		ZKTRACE0("RShimInterface: About to close shim device\n");
		close(shim_fd);
		bInitialized = ZK_FALSE;
		ZKTRACE0("RShimInterface: Just closed shim\n");
	}
}

void RShimInterface::AddShimRule(const RShimRule &rule) throw ( RZkException )
{
	if (bInitialized) {
		const ZkShimRule *rule_struct = rule.getRuleStruct();
		if(ioctl(shim_fd, ZKIOC_ADD_RULE, rule_struct, sizeof(ZkShimRule)) < 0) {
			throw RZkException("RShimInterface::AddShimRule",
                               strerror(errno),
                               IDS_DRIVER_IOCTL_FAILED);		
		}
	} else {
			throw RZkException("RShimInterface::AddShimRule",
                               "Driver not initialized",
                               IDS_FAILED_TO_OPEN_THE_ZKSHIM);		
	}	
}

void RShimInterface::DeleteShimRule(const RShimRule &rule) throw( RZkException )
{
	if (bInitialized) {
		const ZkShimRule *rule_struct = rule.getRuleStruct();
		if(ioctl(shim_fd, ZKIOC_DEL_RULE, rule_struct, sizeof(ZkShimRule)) < 0) {
			throw RZkException("RShimInterface::DeleteShimRule",
                               strerror(errno),
                               IDS_DRIVER_IOCTL_FAILED);		
		}
	} else {
			throw RZkException("RShimInterface::DeleteShimRule",
                               "Driver not initialized",
                               IDS_FAILED_TO_OPEN_THE_ZKSHIM);		
	}	
}

void RShimInterface::AddShimRuleSet(const RShimRuleSet &rule, bool reverse) throw( RZkException )
{
	if (reverse) {
		// this is handy since generally it makes more sense to specify rules in the order
		// you want to have them evaluated. The linked list structures in the shim insert new
		// rules at the list head, and evaluate them from the list head, which is a bit 
		// counterintuitive. Changing this at this point could break a lot of code, so the reverse
		// mode here is a helper.
		for (RShimRuleSet::const_reverse_iterator it = rule.rbegin(); it != rule.rend(); ++it) {
			AddShimRule(*it);
		}
	} else {
		for (RShimRuleSet::const_iterator it = rule.begin(); it != rule.end(); ++it) {
			AddShimRule(*it);
		}
	}
}

void RShimInterface::DeleteShimRuleSet(const RShimRuleSet &rule) throw( RZkException )
{
	for (RShimRuleSet::const_iterator it = rule.begin(); it != rule.end(); ++it) {
		DeleteShimRule(*it);
	}
}

void RShimInterface::ClearShimRules() throw( RZkException )
{
	if (bInitialized) {
		if(ioctl(shim_fd, ZKIOC_CLEAR_RULES) < 0) {
			throw RZkException("RShimInterface::ClearShimRules",
                               strerror(errno),
                               IDS_DRIVER_IOCTL_FAILED);		
		}
	} else {
			throw RZkException("RShimInterface::ClearShimRules",
                               "Driver not initialized",
                               IDS_FAILED_TO_OPEN_THE_ZKSHIM);		
	}	
}

ssize_t RShimInterface::WritePacket(unsigned char *buf, ssize_t buflen) throw ( RZkException )
{
	if (bInitialized) {
		ssize_t n;
retry:		// This is very simple, but all that's necessarry I believe.
		n = write(shim_fd, buf, buflen);
		if (n <= 0) {
            if (errno == EINTR) goto retry;
			throw RZkException("RShimInterface::WritePacket",
                               strerror(errno),
                               IDS_DRIVER_IO_FAILED);		
        }
		else return n;
	} else {
			throw RZkException("RShimInterface::WritePacket",
                               "Driver not initialized",
                               IDS_FAILED_TO_OPEN_THE_ZKSHIM);		
    }
}
	
ssize_t RShimInterface::ReadPacket(unsigned char *buf, ssize_t buflen) throw ( RZkException )
{
	if (bInitialized) {
		ssize_t n;

		n = read(shim_fd, buf, buflen);
		if (n <= 0) {          
            if (errno == EINTR) return -1; // workaround for shim behaviour in gdb
			throw RZkException("RShimInterface::ReadPacket",
                               strerror(errno),
                               IDS_DRIVER_IO_FAILED);		
        }
		else return n;
	} else {
			throw RZkException("RShimInterface::ReadPacket",
                               "Driver not initialized",
                               IDS_FAILED_TO_OPEN_THE_ZKSHIM);		
	}	
}

void RShimInterface::VerifySocketConnection(ZkSocket ZkSock, uint32_t &out_addr, uint16_t &out_port) throw ( RZkException )
{
    DZkError derr;
	char buf[6 + sizeof(ZkShimSecret)];
    
	// Recv the real sockaddr
	if( 1 != zkSocketReadAll(	&derr,
								ZkSock, buf, sizeof(uint32_t) + sizeof(uint16_t) + ZKSHIM_COOKIE_SIZE) )
	{
        throw RZkException("RShimInterface::VerifySocketConnection",
                           "Unable to read secret cookie from inbound connection.",
                           IDS_DEFAULT_ERROR);		
	}
	else 
	{
        if ( 0 != memcmp( (void *)&m_secret.cookie, (void *)&buf[sizeof(uint32_t) + sizeof(uint16_t)], ZKSHIM_COOKIE_SIZE ) )
        {
            throw RZkException("RShimInterface::VerifySocketConnection",
                               "Invalid secret cookie for inbound connection.",
                               IDS_DEFAULT_ERROR);		        
        }
        out_port = *((short*)&buf[sizeof(uint32_t)]);
        memcpy( &out_addr, buf, sizeof(uint32_t) );
	}
}
