// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// DPacket.cpp: implementation of the DPacket class.
//
//////////////////////////////////////////////////////////////////////

// Do not change this inclusion sequence.
#include "StdAfx.h"

#define __STDC_LIMIT_MACROS

#include <IPacketFirewall.h>
#include <IRoutePacket.h>
#include <zkError.h>
#include "DPacket.h"
#include "RCloudSocket.h"
#include <algorithm>
#include <ZkNetMisc.h>
#include <frNetworkDefaults.h> // for MAX_MSS

#ifdef ZK_LINUX
#	include <netdb.h>
#endif

#define DPACKET_TRACE 0

#ifdef WIN32
#define MARIO_HACKINGAWAY_BUG666
#endif


// You probably don't want to do anything which will try to connect the modem in CTORs...
DPacket::DPacket()
	: m_pFirewall( NULL )
	, m_pRoutePacket( NULL )
#ifdef WIN32
	, m_hsomethingtosend( NULL )
	, m_hThread( NULL )
	, m_vector( 100 )
	, m_lastKnownGoodIp(0)
#endif
{
	ZKCTRACE0(DPACKET_TRACE, "DPacket CTOR, since it's a static object, you should see this once!!\n");
	memset(&myIP, 0, sizeof(myIP) );
}

DPacket::~DPacket()
{	
#ifdef WIN32
    SAFECLOSEHANDLE( m_hsomethingtosend );
	SAFECLOSEHANDLE( m_hThread );
#endif
}


void 
DPacket::Terminate()
{
}

bool 
DPacket::HandleIncomingCloudICMPPacket(ZkICMPHeader* in_icmp)
{
	ZKASSERT(in_icmp);

	if(ZK_ICMP_UNREACH == in_icmp->Type)
	{
		ZKCTRACE2(DPACKET_TRACE,  "Next Hop MTU size:  %d (0x%4.4X)\n",
                        ntohs(in_icmp->icmp_NextHop_MTU),
                        ntohs(in_icmp->icmp_NextHop_MTU) );


		if(ZK_ICMP_FRAGMENTATION_NEEDED == in_icmp->code && in_icmp->icmp_NextHop_MTU)
		{
			uint16_t mss = in_icmp->icmp_NextHop_MTU;

			ZKCTRACE2(DPACKET_TRACE, "The PMTU of [%d] will be reduced if it is bigger than [%d]\n", ntohs(mss), MAX_MSS);

			if(ntohs(mss) > MAX_MSS)
			{
				in_icmp->icmp_NextHop_MTU = htons(MAX_MSS);

				ZKCTRACE2(DPACKET_TRACE,  "Next Hop MTU size was reduced to ==>:  %d (0x%4.4X)\n",
                        ntohs(in_icmp->icmp_NextHop_MTU),
                        ntohs(in_icmp->icmp_NextHop_MTU) );
				
			}
		}

		return true;
	}

	return false;
}

bool 
DPacket::ProcessIncomingCloudPacket(ZkBuffer &TCPwithData, unsigned int &nDevice)
{

	if( ((ZkIPHeader*)&TCPwithData)->Protocol == IPPROTO_ICMP )
	{
		const int iphl = ((ZkIPHeader*)&TCPwithData)->ihl * 4; // ip header length in bytes
		bool bRet = HandleIncomingCloudICMPPacket((ZkICMPHeader*)((char*)(&TCPwithData)+iphl));
		zkIPHeaderSetDestinationAddress((ZkIPHeader*)&TCPwithData, m_lastKnownGoodIp);

		return bRet;
	}

	// Bug 1541
	uint32_t ulSrcIp = 0;
	uint32_t ulDstIp = 0;
	if(!zkIPHeaderSourceAddress((ZkIPHeader*)&TCPwithData, &ulSrcIp))
	{
		ZKCTRACE0(DPACKET_TRACE, "zkIPHeaderSourceAddress() failed to retrieve source ip\n");
		return false;
	}

	if(!zkIPHeaderDestinationAddress((ZkIPHeader*)&TCPwithData, &ulDstIp))
	{
		ZKCTRACE0(DPACKET_TRACE, "zkIPHeaderDestinationAddress() failed to retrieve dest ip\n");
		return false;
	}

	if ( (((ZkIPHeader*)&TCPwithData)->Protocol == IPPROTO_UDP) && TCPwithData.Udp.SrcPort == htons(DNS_PORT))
	{
		if( ! GetDnsFilter().ProcessDnsResponse( (ZkBuffer*)&TCPwithData ) ) {
			ZKASSERT("DNS response recieved for unknown request" && 0);
			return false;
		}		
	}
	/* XXX DNS stuff
	if (dnsKey = GetIPKey(1, ulDstIp)) 
	{
		ZKCTRACE1(DPACKET_TRACE, "This packet has DNS key %hu\n", dnsKey);
		ZKASSERT( dnsKey > 0 );	// m_nDnsKeys is pre-inc somewhere.
		ZKVERIFY(ulSrcIp = m_dnsMap[dnsKey]);
		ZKCTRACE1(DPACKET_TRACE, "This packet is actually from DNS server %lu\n", ulSrcIp);
		zkIPHeaderSetSourceAddress((ZkIPHeader*)&TCPwithData, ulSrcIp);
	}
	*/
	if(!GetDestinationIpFromSourceIp(ulDstIp, ulSrcIp ))
	{
		ZKCTRACE1(DPACKET_TRACE, "GetDestinationIp() failed to get destination ip for [%u]\n", ulSrcIp);
#ifndef NDEBUG
		in_addr debugging_src;
		debugging_src.s_addr = ulSrcIp;
		ZKCTRACE1(DPACKET_TRACE, " the ip was [%s]\n", inet_ntoa(debugging_src) );
#endif	// NDEBUG
		return false;
	}

#ifndef NDEBUG
	in_addr debugging_dest;
	debugging_dest.s_addr = ulDstIp;

	in_addr debugging_src;
	debugging_src.s_addr = ulSrcIp;

	ZKCTRACE1(DPACKET_TRACE, "Setting destination ip (that's this machine) to [%s] for source ip of ", inet_ntoa(debugging_dest) );
	ZKCTRACE1(DPACKET_TRACE, "[%s]\n", inet_ntoa(debugging_src) );

#endif	// NDEBUG

	zkIPHeaderSetDestinationAddress((ZkIPHeader*)&TCPwithData, ulDstIp);
	return true;
}

bool
DPacket::HandleOutGoingICMPPacket(
	const int nDevice, 
	ZkBuffer *pbuf )
{
	const bool bBlockPacket = false;
	const bool bAllowPacket = true;



	ZkIPHeader* pIpHeader  = reinterpret_cast<ZkIPHeader*>( pbuf );
	int iIpHdrLen = pIpHeader->ihl << 2;
	ZkICMPHeader* pIcmpHeader = reinterpret_cast<ZkICMPHeader*>( ((unsigned char*)pbuf) + iIpHdrLen );

	FirewallAction   actionFirewall = m_pFirewall->GetActionForOutGoingIcmp(
														 pIcmpHeader->Type) ;

	ZKASSERT( FirewallRouteThroughCloud != actionFirewall );

	if ( actionFirewall == FirewallBlock )
	{
		ZKCTRACE0(DPACKET_TRACE, "Blocking outgoing ICMP packet\n");
		return bBlockPacket;

	}
	else
	{
		ZKCTRACE0(DPACKET_TRACE, "Allowing ICMP packet out\n");
		return bAllowPacket;
	}
}

#ifdef MARIO_HACKINGAWAY_BUG666

unsigned __stdcall 
foo_MARIO_HACKINGAWAY_BUG666( void * in_parm)
{
	DPacket* pDPacket = (DPacket*)in_parm;
	ZKASSERT(pDPacket);

	do
	{
		try
		{
			const DWORD ret = WaitForSingleObject(pDPacket->m_hsomethingtosend, INFINITE);

			while(!pDPacket->m_vector.IsEmpty())
			{

				mybuffer mybuf = { 0 };
				{
					ZKS::Lock_Guard<ZKCriticalSection> gate(pDPacket->m_gate);			
					mybuf = pDPacket->m_vector.PopHead();
				}

				// Set the source IP address hint here before we zero it. This moves out of SetFrame --ol
				pDPacket->SetDestinationIpWithSourceIp(mybuf.buf.Tcp.IP.DstIpAddress, mybuf.buf.Tcp.IP.SrcIpAddress);

				// Insert into the src IP a hint about which adapter to inject it when we get a packet back
				//uint32_t saddr = SetIPKey(nDevice, 0, 0); 
				// AIP 2.0 support: We no longer stuff the Device # into the address, 
				// but put it into a map.
				pDPacket->StoreInterfaceForAddress( mybuf.buf.Tcp.IP.SrcIpAddress, mybuf.dev );

				ZKVERIFY(zkIPHeaderSetSourceAddress((ZkIPHeader*)&mybuf.buf, 0));

				ZKCTRACE1(DPACKET_TRACE, "A TCP Packet for Adapter index [%d] will be sent through cloud\n", mybuf.dev);
				
				// We do not have to do anything to outgoing checksums
				// since the first aip will redoit
				// zkTCPHeaderChecksum((ZkTCPHeader*)&buf);

				// There's some padding added, so I'm ensuring here that we don't overflow the buffer
				ZKASSERT( (ntohs(mybuf.buf.Tcp.IP.TotalLength)-24) < FRPKT_BUF_MAX_SIZE);

				//Bug 1678, remove ip options
				zkIPHeaderStripOptions((ZkIPHeader*)&mybuf.buf);

				//Bug #767, zero out IP, TCP checksums
				mybuf.buf.Tcp.IP.HeaderChkSum = mybuf.buf.Tcp.TcpCheckSum = 0;

				//// Anonymize TCP traffic, which may be WWW at this point
				pDPacket->SendTroughCloud(mybuf.buf);
			}
		}
		catch( ... )
		{
			ZKCTRACE0( DPACKET_TRACE, "Exception caught in foo_MARIO_HACKINGAWAY_BUG666: not sending packet out.  Seee bug #7028" );
		}

		ResetEvent(pDPacket->m_hsomethingtosend); // set it back to non-signalled

	}while(1);

	return 1;
}
#endif

bool
DPacket::HandleOutGoingTCPPacket( 
	const int nDevice, 
	ZkBuffer *pbuf )
{
	bool			 bRet	  = ZK_TRUE ;

	FirewallAction   actionFirewall = m_pFirewall->GetActionForOutgoingTCPPacket( pbuf ) ;

#ifdef MARIO_HACKINGAWAY_BUG666

	if( actionFirewall == FirewallRouteThroughCloud )
	{
		if(NULL == m_hThread)
		{
			m_hsomethingtosend = CreateEvent(NULL, TRUE, FALSE, NULL);
			unsigned int val = 0;
			m_hThread = (HANDLE)_beginthreadex(NULL, 0, foo_MARIO_HACKINGAWAY_BUG666, this, 0, &val);
		}

		mybuffer mybuf = { 0 };
		memcpy( &mybuf.buf, pbuf, 1514 );
		mybuf.dev = nDevice;

		{
			ZKS::Lock_Guard<ZKCriticalSection> gate(m_gate);
			m_vector.PushBack(mybuf);
		}

		SetEvent(m_hsomethingtosend); // remember its manual!
		return false;
	}

#endif

	if ( actionFirewall == FirewallBlock )
	{
		ZKCTRACE0(DPACKET_TRACE, "Blocking packet\n");
		return false;
	}
	else if ( actionFirewall == FirewallRouteThroughCloud )
	{
		// No matter what we can't let this packet leave without going through the cloud, encrypted
		bRet = false;

		// Bug #1566, make sure we don't send D-H through the cloud
		ZKASSERT(FR_DFLT_AIP_TCP_PORT != ntohs(pbuf->Tcp.DstPort) );

		// You may think that anonymizing a remote host, which is equal to the first aip,
		// would lead to an endless loop, since although we block this outgoing packet
		// we send another out via Sendto() below, which would bring us here, and so on.
		// This is not the case since we communicate udp style to the first aip, so the SendTo()
		// below will lead back to the case statement above, not here, hence, no endless loop.

		
		// Sometimes, the OutputDebugString() call the following line ends up calling
		// from _CrtDbgReport() in dbgrpt.c, seems to freeze...may be related to Q173260 
		// since this happens under NT only
		ZKCTRACE1(DPACKET_TRACE, "Going to try and send a UDPacket with this TCP as its payload, pbuf->Tcp.IP.TotalLength is [%d]\n", ntohs(pbuf->Tcp.IP.TotalLength));

#ifdef SEE_OUTGOING_TCP
		ZKCTRACE0(DPACKET_TRACE, "!!!!!!!!!OUTGOING!!!!!!!!!!!!!\n");
		ZKS::DumptcpHeader((ZkTCPHeader*)&buf);
#endif

#ifdef COMPARE_CHKSUMS_TCP
		const uint16_t stack_check_sum = pbuf->Tcp.TcpCheckSum;

		ZkBuffer test;
		memcpy(&test, &buf, sizeof(buf));

		zkTCPHeaderChecksum((ZkTCPHeader*)&test);
		if(stack_check_sum == test.Tcp.TcpCheckSum)
		{
			ZKCTRACE0(DPACKET_TRACE, "Our TCP Checksum function worked fine\n");
		}
		else
		{
			ZKCTRACE0(DPACKET_TRACE, "Our TCP Checksum function ERROR\n");
			ZKASSERT(0);
		}
#endif

		// Set the source IP address hint here before we zero it. This moves out of SetFrame --ol
		SetDestinationIpWithSourceIp(pbuf->Tcp.IP.DstIpAddress, pbuf->Tcp.IP.SrcIpAddress);

		// Insert into the src IP a hint about which adapter to inject it when we get a packet back
		//uint32_t saddr = SetIPKey(nDevice, 0, 0); 
		// AIP 2.0 support: We no longer stuff the Device # into the address, 
		// but put it into a map.
		StoreInterfaceForAddress( pbuf->Tcp.IP.SrcIpAddress, nDevice );

		ZKVERIFY(zkIPHeaderSetSourceAddress((ZkIPHeader*)pbuf, 0));
		ZKCTRACE1(DPACKET_TRACE, "A TCP Packet for Adapter index [%d] will be sent through cloud\n", nDevice);
		
		// We do not have to do anything to outgoing checksums
		// since the first aip will redoit
		// zkTCPHeaderChecksum((ZkTCPHeader*)&buf);

		// There's some padding added, so I'm ensuring here that we don't overflow the buffer
		ZKASSERT( (ntohs(pbuf->Tcp.IP.TotalLength)-24) < FRPKT_BUF_MAX_SIZE);

		//Bug 1678, remove ip options
		zkIPHeaderStripOptions((ZkIPHeader*)pbuf);

		//Bug #767, zero out IP, TCP checksums
		pbuf->Tcp.IP.HeaderChkSum = pbuf->Tcp.TcpCheckSum = 0;

		//// Anonymize TCP traffic, which may be WWW at this point
		SendTroughCloud(*pbuf);
	}
	else 
	{
		// We will allow the packet out in the clear
		ZKASSERT( actionFirewall == FirewallAllowOut ) ;
		bRet = true ;
	}

	return bRet ;
}

bool
DPacket::HandleOutGoingUDPPacket( 
	const int nDevice, 
	ZkBuffer *pbuf )
{
	bool			bRet		= ZK_TRUE ;


	/* bug #1267
	Instead of letting out fragmented UDP we ensure it is actually going to an aip 
	which is the first hop in a route, or to the first aip in a route we're currently
	building
	*/

	FirewallAction   actionFirewall = m_pFirewall->GetActionForOutgoingUDPPacket( pbuf ) ;

	if ( actionFirewall == FirewallBlock )
	{
		ZKCTRACE0(DPACKET_TRACE, "Blocking packet\n");
		return false;
	}
	else if ( actionFirewall == FirewallRouteThroughCloud )
	{

		bRet = false;

		ZKCTRACE1(DPACKET_TRACE, "Going to try and send a UDPacket with this UDP as its payload, buf.Udp.IP.TotalLength is [%d]\n", ntohs(pbuf->Udp.IP.TotalLength));

#ifdef COMPARE_CHKSUMS_UDP
		const uint16_t stack_check_sum = pbuf->Udp.UdpCheckSum;

		ZkBuffer test;	

		memcpy(&test, &buf, sizeof(buf));

		zkUDPHeaderChecksum((ZkUDPHeader*)&test, 1);

		if(stack_check_sum == test.Udp.UdpCheckSum)
			ZKCTRACE0(DPACKET_TRACE, "Our UDP Checksum function worked fine\n");
		else{
			ZKASSERT(0);
			ZKCTRACE0(DPACKET_TRACE, "Our UDP Checksum function ERROR\n");
			ZKS::DumptcpHeader((ZkTCPHeader*)&buf);
			ZKCTRACE1(DPACKET_TRACE, "\r\n Phils checksum [%d]\n", (unsigned short)test.Udp.UdpCheckSum);
			ZKCTRACE1(DPACKET_TRACE, "\r\n orig UDP checksum [%d]\n", (unsigned short)stack_check_sum);
		}
#endif

		// For DNS we redirect the traffic to our own DNS server
		// This is the only case thus far where traffic is redirected (Destination IP modified) once it leaves the exit funnel

		if ( htons( DNS_PORT ) == pbuf->Udp.DstPort/* in NBO */ )
		{
			if ( !HandleOutGoingDNSLookup( pbuf ) )
			{
				return false ;
			}
		}

		// Set the source IP address hint here before we zero it. This moves out of SetFrame --ol
		SetDestinationIpWithSourceIp(pbuf->Tcp.IP.DstIpAddress, pbuf->Tcp.IP.SrcIpAddress);

		ZKASSERT( 0 != memcmp(&pbuf->Tcp.IP.DstIpAddress, &pbuf->Tcp.IP.SrcIpAddress, 4) );

		// Insert into the src IP a hint about which adapter to inject it when we get a packet back
		//uint32_t saddr = SetIPKey(nDevice, 0, 0); 
		// AIP 2.0 support: We no longer stuff the Device # into the address, 
		// but put it into a map.
		StoreInterfaceForAddress( pbuf->Tcp.IP.SrcIpAddress, nDevice );

		ZKVERIFY(zkIPHeaderSetSourceAddress((ZkIPHeader*)pbuf, 0));
		ZKCTRACE1(DPACKET_TRACE, "A UDP PAcket for Adapter index [%d] will be sent through cloud\n", nDevice);

		//Bug 1678, remove ip options
		zkIPHeaderStripOptions((ZkIPHeader*)pbuf);

		//Bug #767, zero out IP, UDP checksums
		pbuf->Udp.IP.HeaderChkSum = pbuf->Udp.UdpCheckSum = 0;

		// Anonymize UDP traffic, which may be DNS at this point, or a key query
		SendTroughCloud(*pbuf);
	}
	else 
	{
		// We will allow the packet out in the clear
		ZKASSERT( actionFirewall == FirewallAllowOut ) ;
		bRet = true ;
	}

	return bRet ;
}

bool
DPacket::To_receive_OR_not_to_receive(const ZkNetFrame* in_buf)
{
	if(!m_pFirewall) return true; // allow into the stack

	return m_pFirewall->HandleIncomingPacket( in_buf ) ;
}


// Returning true will send it out through the same adapter from which the packet came from
bool 
DPacket::To_send_OR_not_to_send(
	const int nDevice,
	ZkBuffer *pbuf,
	const int proto
	)
{
	// CSingleLock csl(&cs, TRUE);//Will call  CSingleLock::Lock() which simply calls ::EnterCriticalSection()

    // This is needed to debug on linux. Stepping in the debugger causes execution of other
    // threads to drop out of syscalls. There is a bug in the kernel code, I think, which is 
    // causing execution to fall through to here --ol (XXX Smoe this after the bug is fixed)
    if (!m_pFirewall) {
        ZKCTRACE0(DPACKET_TRACE, "Spurious call to DPacket::To_Send_OR_not_To_sEnD\n");
        return ZK_TRUE;
    }
	//ZKASSERT(m_pFirewall);

	bool			 bRet	  = ZK_TRUE ;

	// Sometimes, the OutputDebugString() call the following line ends up calling
	// from _CrtDbgReport() in dbgrpt.c, seems to freeze...may be related to Q173260 
	// since this happens under NT only
	ZKCTRACE0(DPACKET_TRACE, "in To_send_OR_not_to_send\n");

	try
	{
		switch(proto)
		{
			case IPPROTO_UDP:
				bRet = HandleOutGoingUDPPacket( nDevice, pbuf ) ;
				break ;

			case IPPROTO_TCP:
				bRet = HandleOutGoingTCPPacket( nDevice, pbuf ) ;
				break;

			case IPPROTO_ICMP:
				bRet = HandleOutGoingICMPPacket( nDevice, pbuf ) ;
				break;

			default:
				ZKCTRACE1(DPACKET_TRACE, "Unknown IP protocol [%d] in DPacket::To_send_OR_not_to_send()\r\n", proto);
		};
	}
	catch( ... )
	{
		ZKCTRACE0( DPACKET_TRACE, "Exception caught in DPacket::To_send_OR_not_to_send: not sending packet out. See bug #7028." );
		bRet = ZK_FALSE;
	}

	return bRet;
}

bool
DPacket::HandleOutGoingDNSLookup( ZkBuffer *io_pbuf )
{
	bool bRet = true ;

	ZKASSERT( htons( DNS_PORT ) == io_pbuf->Udp.DstPort ) ;

	// We get the IP of the last hop of our nym's authenticated route
	sockaddr_in exitAipHost ;
	if( m_pRoutePacket->GetExitAipIPForDNS( exitAipHost ) )
	{ 

		ZKCTRACE2(DPACKET_TRACE,  "Original DNS Destination IP Address [%s] to port [%d]\n",  
				  ZKS::Cvt_int32toString(io_pbuf->Udp.IP.DstIpAddress).c_str(), 
				  DNS_PORT );

		ZKCTRACE2(DPACKET_TRACE, "Redirecting DNS to MUST BE dot notation, must be exit funnel [%s] to port [%d]\n",  
			     ZKS::Cvt_int32toString(exitAipHost.sin_addr.s_addr), 
				 DNS_PORT );

		unsigned long addr = exitAipHost.sin_addr.s_addr ;


		if( INADDR_NONE != addr )
		{
			GetDnsFilter().ProcessDnsRequest(io_pbuf);
/*			bool    found = false; 
			for (out_dnsKey = 1 ; out_dnsKey <= m_nDnsKeys; out_dnsKey++)
			{
				if (m_dnsMap[out_dnsKey] == io_pbuf->Udp.IP.DstIpAddress) 
				{
					found = true;
					break;
				}
			}

			if (!found) 
			{
				out_dnsKey = ++m_nDnsKeys;
				ZKCTRACE1(DPACKET_TRACE, "DNS Server not found, DNS key is [%hu]\n", out_dnsKey);
				m_dnsMap[out_dnsKey] = io_pbuf->Udp.IP.DstIpAddress;
				ZKASSERT( io_pbuf->Udp.IP.DstIpAddress );
			}
*/
			SetDestinationIpWithSourceIp(io_pbuf->Udp.IP.DstIpAddress, io_pbuf->Udp.IP.SrcIpAddress);
			zkIPHeaderSetDestinationAddress((ZkIPHeader*)io_pbuf, addr);

			ZKCTRACE1(DPACKET_TRACE, "New DNS Destination IP Address  %s\n",  ZKS::Cvt_int32toString(io_pbuf->Udp.IP.DstIpAddress).c_str());
		}
		else
		{
			ZKCTRACE0(DPACKET_TRACE, "Got IP INADDR_NONE as EXIT AIP for DNS lookup, have to block this packet\n");
			ZKASSERT(0);

			bRet = false ;
		}					


		// Our DNS server may be on another port than 53
		// VERIFY(zkUDPHeaderSetDestinationPort((ZkUDPHeader*)&buf, sOurDnsPort /*must be little endian-host byte order*/));

		// We do not have to do anything to outgoing checksums
		// since the first aip will redoit
		// zkUDPHeaderChecksum((ZkUDPHeader*)&buf, 1);
	}

	return bRet ;
}

void 
DPacket::SendTroughCloud(ZkBuffer& buf)
{
	char				data[FRPKT_BUF_MAX_SIZE];	// New freedomized version of packet
	size_t				dlen = FRPKT_BUF_MAX_SIZE;
	sockaddr_in			firstHopAddr ;						// Address to send packet too
	memset( &firstHopAddr, 0 , sizeof( firstHopAddr ) ) ;

	if ( m_pRoutePacket == NULL )
	{
		ZKASSERT( false ) ;
		return ;
	}

	// We call up to get a packet ready to send to the
	// correct route.  If no route exists this may fail,
	// in which case we do not send anything out
	if ( m_pRoutePacket->CreateOutgoingPacket( 
							buf,
							data,
							dlen,
							firstHopAddr ) )
	{
		if(RCloudSocket::theRCloudSocket().SendTo(data, dlen, firstHopAddr))
		{
			ZKCTRACE0(DPACKET_TRACE, "RCloudSocket::SendTo() worked\n");
		}	
		else
		{
			ZKCTRACE0(DPACKET_TRACE, "RCloudSocket::SendTo() failed to work\n");
		}
	}
}


void DPacket::SetFirewallPtr(IPacketFirewall* in_pIPacketFirewall) 
{ 	
	ZKASSERT( in_pIPacketFirewall != NULL ) ;
	m_pFirewall = in_pIPacketFirewall; 
}

void DPacket::SetRoutePacketHandler(IRoutePacket* in_pRoutePacket)
{
	ZKASSERT( in_pRoutePacket != NULL ) ;
	m_pRoutePacket = in_pRoutePacket; 
}

  
void DPacket::SetDestinationIpWithSourceIp(const uint32_t & in_ulDestIp, const uint32_t& in_ulSrcIp)
{	
	ZKASSERT(in_ulDestIp);
	ZKASSERT(in_ulSrcIp);

	ZKCTRACE2(DPACKET_TRACE, "SetDestinationIpWithSourceIp() [%s] = [%s]\n", 
		ZKS::Cvt_int32toString(in_ulDestIp).c_str(), 
		ZKS::Cvt_int32toString(in_ulSrcIp).c_str());

	// Notice how we store [IBM]-> me
	// since when we'll need to determine our IP, we'll only have IBM available

	ZKS::Lock_Guard<WZkMutex> lock( m_ipMapMutex );
	m_lastKnownGoodIp = m_ipMap[in_ulDestIp] = in_ulSrcIp;
}

bool DPacket::GetDestinationIpFromSourceIp(uint32_t & out_ulDestIp, const uint32_t& in_ulSrcIp)
{	
	bool bRet = false;

	ZKASSERT(in_ulSrcIp);

	ZKCTRACE1(DPACKET_TRACE, "GetDestinationIpFromSourceIp() looking for [%d]\n", in_ulSrcIp );

	ZKS::Lock_Guard<WZkMutex> lock( m_ipMapMutex );

	ZKSTD::map<uint32_t, uint32_t>::const_iterator it = m_ipMap.find(in_ulSrcIp);

	if(it != m_ipMap.end())
	{
		out_ulDestIp = (*it).second;		
		bRet = true;
		ZKCTRACE1(DPACKET_TRACE, "[%d] would be your (this machine's) IP for this packet\n", out_ulDestIp );
		ZKASSERT(out_ulDestIp != 0);
	}

	return bRet;
}


void DPacket::StoreInterfaceForAddress( uint32_t addr, unsigned int nDevice )
{
	m_deviceMap[ addr ] = nDevice;
}

unsigned int DPacket::RetrieveInterfaceForAddress( uint32_t addr )
{
	return m_deviceMap[ addr ];
}

// XXX Remove that
#ifndef NDEBUG
void DPacket::DumpIPMapToFile( const ZkBuffer& in_rBuffer )
{
	ZKSTD::map<uint32_t, uint32_t>::const_iterator it = m_ipMap.begin();
	FILE* fp = fopen( "IPMap.txt", "a" );

	if( !fp )
	{
		ZKASSERT( fp );
		return;
	}

	in_addr bufAddr = { 0 };
	bufAddr.s_addr = in_rBuffer.Tcp.IP.SrcIpAddress;
	fprintf( fp, "The IP %s could not be found in the IP map containing those IPs: \n", inet_ntoa( bufAddr ) );
	
	for(; it != m_ipMap.end(); ++it )
	{
		in_addr addr = { 0 };
		addr.s_addr = it->first;

		fprintf( fp, "%s\n", inet_ntoa( addr ) );
	}

	fclose( fp );
}

#endif