// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// author="Mario Contestabile"
// DPacket.h: interface for the DPacket class.
//
//////////////////////////////////////////////////////////////////////


#if !defined(_DPACKET_H_INCLUDED)
#define _DPACKET_H_INCLUDED

#ifdef ZK_LINUX
#	include <sys/socket.h>
#	include <netinet/in.h>
#	include <arpa/inet.h>
#endif

#include <zkPacket.h>
#include <list>
#include <DZkError.h>
#include <zkThread.h>
#include <vector>
# include <map>
#ifdef WIN32
#  include <CircularVector.h>
#endif
#include <frAips.h>

#include "DnsFilter.h"
#include <WSyncTypes.h>

class IRoutePacket ;
class IPacketFirewall ;

typedef struct mybuffer
{
	ZkBuffer		buf;
	int				dev;
}mybuffer;

class DPacket
: public IDnsFilterOwner
{
protected:
	IPacketFirewall*				m_pFirewall;
	IRoutePacket*                   m_pRoutePacket;
	DZkError						m_err;
	IN_ADDR							myIP;			// myIP.S_addr is the IP Eg: 0x0F01A8C0
	ZKSTD::map<uint32_t, uint32_t>	m_ipMap;
	uint32_t						m_lastKnownGoodIp;
	WZkMutex						m_ipMapMutex;
	
	// outbound source IP addresses -> device #
	// on the way back in we use this to prepend
	// ether data-link headers onto IP datagrams pulled
	// off of the freedom cloud
	ZKSTD::map<uint32_t, unsigned int> m_deviceMap; 
						
//	ZKSTD::map<uint16_t, uint32_t>	m_dnsMap;	// DNS ip keys -> dns server addresses
//	uint16_t							m_nDnsKeys; // 16bits for key in the dns map, because we split the source IP in 2 (32bits/2 :) 16bits for the device handle and 16 bits for this indez in the dns map.

	// non-implemented functions
							DPacket(const DPacket&);			// never allow copying
	DPacket&				operator=(const DPacket&); // never allow assignment

	// private member functions

	// Send a packet over the cloud.  It will be 'freedomized' and send on the correct route
	void	SendTroughCloud(ZkBuffer&);
		
	// PURPOSE:	For the given IP 'in_ulDestIp' entry in the 'm_ipMap', set its value to 'in_ulSrcIp'
	// PROMISE: Not to throw
	// REQUIRE:	Non zero parameters
	void					SetDestinationIpWithSourceIp(const uint32_t& in_ulDestIp, const uint32_t& in_ulSrcIp) throw();


public:	

#ifdef WIN32
	friend unsigned __stdcall 
	foo_MARIO_HACKINGAWAY_BUG666( void * in_parm);

	ZKCriticalSection m_gate;
	HANDLE m_hsomethingtosend;
	HANDLE m_hThread;
	ZKS::CircularVector<mybuffer> m_vector;

#endif

	// PURPSOE: For the given 'in_ulSrcIp' entry in 'm_ipMap' return the value in 'out_ulDestIp'
	// PROMISE: Returns true if 'in_ulSrcIp' is in the map, false otherwise
	//			Does not touch 'out_ulDestIp' if return false ('in_ulSrcIp' is not in the map)
	// REQUIRE:	Non zero 'in_ulSrcIp' parameter
	bool					GetDestinationIpFromSourceIp(uint32_t& out_ulDestIp, const uint32_t& in_ulSrcIp) throw();


	// Pure virtual functions
	virtual int				Init() = 0;
	virtual bool			InjectPacketIntoProtocolStack(const ZkBuffer &buf, const unsigned int &nDevice) = 0;


	virtual void			Terminate() = 0;

	// Called by our udp server thread when it receives a FRPKTTYPE_AIP_ROUTE_CRTACK packet
	void					CreatePacketWasAcked(const ZkACI nAci);
	
	void					SetFirewallPtr(IPacketFirewall* in_pFirewall);
	void					SetRoutePacketHandler(IRoutePacket* in_pRoutePacket);

	const ZKSTD::string&	GetThisMachineName();

							DPacket();
	virtual					~DPacket();

	
	virtual bool			To_receive_OR_not_to_receive(const ZkNetFrame* in_buf);
	
	virtual bool			To_send_OR_not_to_send(const int nDevice, ZkBuffer *pbuf, const int proto);

	virtual bool			HandleOutGoingICMPPacket(	const int nDevice, 	ZkBuffer *pbuf );
	virtual bool			HandleOutGoingUDPPacket( const int nDevice, ZkBuffer *pbuf ) ;
	virtual bool			HandleOutGoingTCPPacket( const int nDevice, ZkBuffer *pbuf ) ;
	virtual	bool			HandleOutGoingDNSLookup( ZkBuffer *io_pbuf ) ;

	virtual bool			HandleIncomingCloudICMPPacket(ZkICMPHeader* in_icmp);
	virtual bool			ProcessIncomingCloudPacket(ZkBuffer &buf, unsigned int &nDevice);

	bool					GetMyIp(IN_ADDR&);
	

	// These functions replace the 'stuffing' cookies into IP address hack which will not work with
	// AIP 2.0. StoreInterfaceForAddress will create an association between an address and an interface 
	// number (by storing it in in m_deviceMap ). RetrieveInterfaceForAddress will retrieve an interface number
	// based on an IP address. The first function should be called with the source IP of sent packets,
	// the second with the destination IP of incoming packets. --ol
	void					StoreInterfaceForAddress( uint32_t addr, unsigned int nDevice );
	unsigned int			RetrieveInterfaceForAddress( uint32_t addr );

	// Get an instance of the dns filter class used to do additional processing on DNS requests.
	virtual DnsFilter&      GetDnsFilter() = 0;

// XXX Remove that
#ifndef NDEBUG
	void DumpIPMapToFile( const ZkBuffer& in_rBuffer );
#endif
};


#endif // !defined(_DPACKET_H_INCLUDED)
