// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom(R) code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "DnsFilter.h"
#include <RDnsMessage.h>


//---------------------------------------------------------------------------
//
DnsKey::DnsKey()
{
	SetTtl(DNSKEY_TTL);
}

//---------------------------------------------------------------------------
//
DnsKey::DnsKey(uint16_t dnsId, uint16_t SourcePort, unsigned short Ttl)
{
	SetDnsId(dnsId);
	SetSourcePort(SourcePort);
	SetTtl(Ttl);
}

//---------------------------------------------------------------------------
//
DnsKey::~DnsKey()
{
}

//---------------------------------------------------------------------------
//
bool DnsKey::lessthan::operator()(const DnsKey& c1, const DnsKey& c2) const
{
	return ( (uint32_t)c1 < (uint32_t)c2 );
}

//---------------------------------------------------------------------------
//
DnsFilter::DnsFilter()
	: m_DnsMap()
{
	DZkError derr;
	ZKVERIFY( zkMutexNew(&m_err, &m_mutex) );
}


//---------------------------------------------------------------------------
//
DnsFilter::~DnsFilter()
{
	ZKVERIFY( zkMutexFree(&m_err, &m_mutex) );
}


//---------------------------------------------------------------------------
//
bool DnsFilter::ProcessDnsRequest( ZkBuffer *pDnsBuffer )
{
	DnsKey requestKey;
	RDnsMessage::DnsHeader *pHeader;
	DnsMapType::iterator iMapPair;
	uint32_t dstIP;

	ZKASSERT( htons( DNS_PORT ) == pDnsBuffer->Udp.DstPort ) ;

	dstIP = pDnsBuffer->Udp.IP.DstIpAddress;
	requestKey.SetSourcePort( pDnsBuffer->Udp.SrcPort );	
	pHeader = reinterpret_cast<RDnsMessage::DnsHeader *>
		( reinterpret_cast<char*>( pDnsBuffer ) + sizeof( ZkUDPHeader ) );
	requestKey.SetDnsId( pHeader->GetID() );

	zkMutexLock( &m_err, m_mutex );
	// Check to see if the map already contains a entry with this key
	iMapPair = m_DnsMap.find( requestKey );
	if ( iMapPair != m_DnsMap.end() ) 
	{
        //ZKASSERT("DNS request map collision!" && 0 );
		// It does. Erase the old entry.
		m_DnsMap.erase( iMapPair );
	}
	// Insert a mapping for this dns request such that we'll be able to find it later.
	m_DnsMap[ requestKey ] = dstIP;
	zkMutexUnlock( &m_err, m_mutex );
	
	return true;
}

//---------------------------------------------------------------------------
//
bool DnsFilter::ProcessDnsResponse( ZkBuffer *pDnsBuffer )
{
	DnsKey requestKey;
	RDnsMessage::DnsHeader *pHeader;
	DnsMapType::iterator iMapPair;
	uint32_t dstIP;

	ZKASSERT( htons( DNS_PORT ) == pDnsBuffer->Udp.SrcPort ) ;

	requestKey.SetSourcePort( pDnsBuffer->Udp.DstPort );	
	pHeader = reinterpret_cast<RDnsMessage::DnsHeader *>
		( reinterpret_cast<char*>( pDnsBuffer ) + sizeof( ZkUDPHeader ) );
	requestKey.SetDnsId( pHeader->GetID() );

	zkMutexLock( &m_err, m_mutex );
	// Check to see if the map already contains a entry with this key
	iMapPair = m_DnsMap.find( requestKey );
	if ( iMapPair == m_DnsMap.end() ) 
	{
		ZKTRACE0("Martian DNS request (didn't exist in our map). Dropping");
		zkMutexUnlock( &m_err, m_mutex );
		return false;
	}
	// We found a good entry. Stuff the associated dns server's IP address
	// into the source IP of this packet
	dstIP = iMapPair->second;	
	pDnsBuffer->Udp.IP.SrcIpAddress = dstIP;

	zkMutexUnlock( &m_err, m_mutex );
	return true;
}

//---------------------------------------------------------------------------
//
void ZK_ITIMER_CALLBACK DnsFilter::CleanupTimer(void*, unsigned int, unsigned int, unsigned long)
{
    // XXX ToDo: implement
}


