// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// RCloudSocket.cpp

// RCloudSocket.cpp: implementation of the RCloudSocket class.
//
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#define __STDC_LIMIT_MACROS

#include "RCloudSocket.h"

#ifdef WIN32
#	include <winsock.h>
#else
#	include <sys/socket.h>
#   include <signal.h>
#endif

#include <sys/types.h>

#include "DPacket.h"

#ifdef ZK_LINUX
#	include <RLinuxPacketHandler.h>	// For FRAME_LENGTH
#	define SOCKET_ERROR -1
#   include <frNetworkDefaults.h>
#   include <netinet/in.h>
#   include <arpa/inet.h>
#   include <RShimInterface.h>
#endif

#include <RZkException.h>
#include <zkFragCache.h>
#include <DZkError.h>
#include <ErrorHandling.h>
#include <IRoutePacket.h>
#include <TraceControl.h>
#include <ZkNetMisc.h>		// for DumptcpHeader

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
/////////////////////////////////////////////////////////////////////

RCloudSocket::RCloudSocket() 
	: m_UDPThreadSocket( INVALID_SOCKET )
	, m_pDPacket( NULL )
	, m_fcache( NULL )
	, m_pUdpThread( NULL )
	, m_pRoutePacket( NULL )
{

}

RCloudSocket::~RCloudSocket()
{
}


// Singleton Pattern
RCloudSocket& RCloudSocket::theRCloudSocket(){
	static RCloudSocket d;          // static in function, not in class, so if never
	return d;                    // called, never created. We are also certain about its time of creation.
}


void    
RCloudSocket::Init( DPacket* in_pDPacket ) throw(RZkException)
{
	// For example, 'in_pDPacket' could be a RWin2kPacketHandler
	ZKASSERT( in_pDPacket != NULL ) ;

	m_pDPacket = in_pDPacket ;
	
	DZkError derr ;

	ZKVERIFY(zkFragCacheNew(&derr, &m_fcache));

	// Always completes immediately, default used of blocking mode
	m_UDPThreadSocket = ::socket(AF_INET,	// Address family
				SOCK_DGRAM,		// Socket type
				IPPROTO_UDP);	// Protocol

	if (m_UDPThreadSocket == INVALID_SOCKET)
	{
		ZKCTRACE0( TRACE_RCLOUDSOCKET,"UDPServer thread unable to create socket()\r\n");

		throw RZkException(
					"RCloudSocket::Init", 
					"Failed to Create Cloud UDP Socket", 
					IDS_FATAL_STARTUP_NETWORKING);	

	}
	unsigned long var = 1;

	int nRet = 0;

	// Fill in the address structure
	SOCKADDR_IN saServer;
	saServer.sin_family			= AF_INET;
	saServer.sin_addr.s_addr	= INADDR_ANY; // Let WinSock assign address
	saServer.sin_port			= htons(WINDOWS_TCP_AND_UDP_PORT);

	// bind the name to the socket
	//
	nRet = ::bind( m_UDPThreadSocket,	// Socket descriptor
				   (SOCKADDR *)&saServer,		// Address to bind to
				   sizeof(struct sockaddr)		// Size of address
				);

	if (nRet == SOCKET_ERROR)
	{
#ifdef WIN32	
		ZKCTRACE1( TRACE_RCLOUDSOCKET,"UDP Server thread unable to name the socket with bind() [%d]\n", WSAGetLastError());	
#else
		ZKTRACE1( "UDP Server thread unable to name the socket with bind() [%d]\n", errno);	
#endif				

		throw RZkException( 
					"RCloudSocket::Init", 
					"Failed to Bind Cloud UDP Socket", 
					IDS_FATAL_STARTUP_NETWORKING);	

	}

	// Start the thread which reads incoming data from the cloud

	if (!zkThreadCreate(&derr, &m_pUdpThread, &RCloudSocket::UDPServer, this)) {
		throw RZkException( derr,
							"RCloudSocket::Init", 
							"UDP Cloud Thread Creation failed", 
							IDS_FATAL_STARTUP_NETWORKING );	
	}
}

void
RCloudSocket::Terminate()
{

	if(m_UDPThreadSocket != INVALID_SOCKET)
	{
#ifdef WIN32
		::closesocket(m_UDPThreadSocket);
#else
        {
            DZkError derr;
            pthread_t *pptr;
            zkThreadGetThreadNP( &derr, m_pUdpThread, (void **)&pptr );
            ::close( m_UDPThreadSocket );
            ZKVERIFY( !pthread_cancel( *pptr ) );
        }
#endif
            m_UDPThreadSocket = INVALID_SOCKET ;
	}


	// Close down the UDP Thread
	if (m_pUdpThread)
	{
		DZkError derr ;
		void *rval;
		ZKVERIFY(zkThreadJoin(&derr, &m_pUdpThread, &rval));
	}

	if(m_fcache)
	{
		ZKVERIFY(zkFragCacheFree(NULL, &m_fcache));
	}

}


bool RCloudSocket::SendTo(const char *Buf, const int nLen2Write, const SOCKADDR_IN& Destination, int nWait /* =0 */) const
{

	if(Buf && (nLen2Write>0) )
	{
		ZKCTRACE1( TRACE_RCLOUDSOCKET,"Gonna send data of length [%d] to first hop\n", nLen2Write);
#ifdef VERIFY_MTU
			if(CheckMTU(m_UDPThreadSocket, nLen2Write))
			{
#endif
				if(m_UDPThreadSocket == INVALID_SOCKET)
				{
					ZKASSERT(0);
					return false;
				}

				const int ret = ::sendto(m_UDPThreadSocket,	// Protocol stack implicitly names my client socket on my sendto() call...no more bind()ing allowed.
						Buf,								// Data buffer
						nLen2Write,							// Length of data
						0,									// Flags
						(const SOCKADDR *)&Destination,		// Server address, SOCKADDR_IN
						sizeof(struct sockaddr));

				if(ret == nLen2Write)
				{
					return true;
				}

#ifdef VERIFY_MTU
			}else
			{ // if MTU okay
				TRACE0("nLen2Write via sendto() exceeds MTU, not sending this out\n");
			}
#endif
	}else
		ZKCTRACE1( TRACE_RCLOUDSOCKET,"buf is NULL or nLen2Write is <=0 [%d] in RCloudSocket::SendTo()\n", nLen2Write);

	return false;
}


// care must be taken not to exceed the maximum packet size of the underlying subnets
bool RCloudSocket::CheckMTU(const ZkSocketDescriptor s, const int nLen) const{

	bool bRet	= false;
	int nOptLen = sizeof(int);
	int nOptVal	= 0;

#ifdef WIN32	// No sock opt to do this on linux. Must be another waym
	const int nRet = ::getsockopt(s,
					  SOL_SOCKET,
					  SO_MAX_MSG_SIZE,
					  (char*)&nOptVal,
					  &nOptLen);
	
	if(nRet != -1){		
		bRet = nOptVal > nLen;
	}
#endif
	return true;
}

#ifdef DUMPINCOMING2FILE
	#include <fstream>
#	ifdef WIN32	
	ZKSTD::ofstream outfile("C:\\ethernet.bin", ZKSTD::ios::binary | ZKSTD::ios::app | ZKSTD::ios::ate);
#	else
	ZKSTD::ofstream outfile("~/netdump.bin", ZKSTD::ios::binary | ZKSTD::ios::app | ZKSTD::ios::ate);
#	endif
#endif

void * ZK_THREAD_CALLBACK 
RCloudSocket::UDPServer(void *in_pThis)
{
	DZkError			derr;
	RCloudSocket* pThis = (RCloudSocket*) in_pThis ;

	if ( pThis == NULL )
	{
		ZKASSERT( false ) ;
		return NULL ;
	}

	ZKASSERT( pThis->m_fcache != NULL ) ;


	ZKCTRACE0( TRACE_RCLOUDSOCKET,"UDPServer thread started, gonna try to listen on WINDOWS_TCP_AND_UDP_PORT if possible, make sure you checked Winsock 2 was installed\r\n");	

	DPacket* pPacket = pThis->m_pDPacket ;

	//
	// Wait for data from the client
	//
	SOCKADDR_IN saClient;

	char szBuf[FRPKT_BUF_MAX_SIZE];
	unsigned int nLen = sizeof(SOCKADDR);

	do
	{
		memset(szBuf, 0, sizeof(szBuf));


		const int nRet = ::recvfrom(pThis->m_UDPThreadSocket,			// Bound socket
							szBuf,								// Receive buffer
							FRPKT_BUF_MAX_SIZE,					// Size of buffer in bytes
							0,									// Flags
							(SOCKADDR *)&saClient,				// Buffer to receive client address 
							(int *)&nLen);								// Length of client address buffer


			if(nRet > 0)
			{
				//ZKCTRACE1( TRACE_RCLOUDSOCKET,"UDPServer thread received [%d] bytes\n", nRet);
			}
#ifdef _WIN32
			else if(0 == nRet || (WSAENOTSOCK == WSAGetLastError()) )
#else
			else if(0 == nRet || (ENOTSOCK == errno) )
#endif
			{
				ZKCTRACE0( TRACE_RCLOUDSOCKET,"UDPServer thread socket closed\n");			
				return NULL;
			}


		// Maybe for performance we might not want to zero this (large) structure
		ZkBuffer TCPwithData = { 0 } ;
		
		ZKASSERT( pThis->m_pRoutePacket != NULL ) ;
		FrPktType type = pThis->m_pRoutePacket->ProcessIncomingCloudPacket( saClient, szBuf, nRet, TCPwithData ) ;

#ifdef DUMP_INCOMING		
		ZKS::DumptcpHeader((ZkTCPHeader*)&TCPwithData);
#endif


		if ( type != FRPKTTYPE_AIP_DATA )
		{
			continue ;
		}

/*
	Packet_zkTCPHeaderChecksum() will crash if we send it garbage
*/
		if(TCPwithData.Tcp.IP.version != 4){	// simple test to ensure data read is something good
			ZKCTRACE0( TRACE_RCLOUDSOCKET,"UdpServer received bad ZkBuffer, [the ip version wasn't 4] continuing to listen\n");
			ZKS::DumptcpHeader((ZkTCPHeader*)&TCPwithData);
			continue; // go back to do{}
		}		


		size_t iph_sz = ZKBUFFER_LEN;

		if(ntohs(TCPwithData.Tcp.IP.Flags_FrameOffs) & 0x3FFF){ // is this a fragment?
			ZKVERIFY(zkFragCacheAddPacket(
						NULL, 
						pThis->m_fcache, 
						(ZkIPHeader*)&TCPwithData, 
						ntohs(TCPwithData.Tcp.IP.TotalLength),
						0, 
						PACKET_FROM_INTERNET, 
						(ZkIPHeader*)&TCPwithData, 
						&iph_sz));
			if(iph_sz < 1){
				ZKCTRACE0( TRACE_RCLOUDSOCKET,"Got a fragment\n");
				continue;
			}
		}

		unsigned int nDevice = 0 ;
		if ( !pPacket->ProcessIncomingCloudPacket(TCPwithData, nDevice)) {
			ZKCTRACE0( TRACE_RCLOUDSOCKET,"DPacket::ProcessIncomingCloudPacket() returned false in UDPServer\n");
			continue;
		}

		if(TCPwithData.Tcp.IP.Protocol == IPPROTO_TCP)
		{
			// Can't compare checksums, we know it is invalid when we get it
			zkTCPHeaderChecksum((ZkTCPHeader*)&TCPwithData);

		}else if(TCPwithData.Tcp.IP.Protocol == IPPROTO_UDP)
		{
			zkUDPHeaderChecksum((ZkUDPHeader*)&TCPwithData, 1);

		}else if(TCPwithData.Tcp.IP.Protocol == IPPROTO_ICMP)
		{

			ZkIPHeader* phdr = (ZkIPHeader*)&TCPwithData;
			const int iphl = phdr->ihl * 4; // ip header length in bytes

			zkICMPHeaderChecksum(  (ZkICMPHeader*)((char*)(&TCPwithData)+iphl), 8);

			zkIPHeaderChecksum( (ZkIPHeader*)&TCPwithData );
		}
		else{
			ZKCTRACE1( TRACE_RCLOUDSOCKET,"UDP Server received IP packet which wasn't UDP or TCP protocol, discarding it [%d](see IPPROTO_xxx in winsock2.h)\n", TCPwithData.Tcp.IP.Protocol);
			continue;	// go back to do{}
		}

		if ( !pPacket->InjectPacketIntoProtocolStack(TCPwithData, nDevice)) {
			ZKCTRACE0( TRACE_RCLOUDSOCKET,"DPacket::InjectPacketIntoProtocolStack() returned false in UDPServer\n") ;
		}

#ifdef DUMPINCOMING2FILE
		if(!outfile.is_open()){ ASSERT(0); }
		else{ 			
			// not 125 ... who cares it'll dump the headers
			outfile.write((const char*)TCPwithData, 125); 
			outfile.write((const char*)"OLIVER", 6 );
		}
#endif
	}while(1);

// Never get here

	return (void *)1;
}

void 
RCloudSocket::SetRoutePacketHandler(IRoutePacket* in_pRoutePacket)
{
	ZKASSERT( in_pRoutePacket != NULL ) ;
	m_pRoutePacket = in_pRoutePacket; 
}
