// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):



//===========================================================================
// EXTERNAL DECLARATIONS
//===========================================================================

#include <UiConsoleWindow.h>
#include <UiThreadId.h>

#include <ZkConsoleHint.h>
#include <ZkGladeLoader.h>
#include <ZKFileConfiguration.h>
#include <ZkGtkUtils.h>
#include <lockguard.h>
#include <IZkRegistry.h>

#include <clientdebug.h>

#include <cstdio>
#include <gnome.h>
#include <gdk/gdkx.h>

#include <time.h>

#define CONSOLE_NUM_COLUMNS 3
#define CONSOLE_PIXMAP 0
#define CONSOLE_TIME 1
#define CONSOLE_MSG 2

const int MAX_MESSAGES=2001;
const int yPosShimmy = 21;
const char CONSOLE_RECT_LEFT_KEY[] = { "Console Dlg Rect Left" };
const char CONSOLE_RECT_TOP_KEY[] = { "Console Dlg Rect Top" };
const char CONSOLE_X_SIZE[] = { "Console X Size" };
const char CONSOLE_Y_SIZE[] = { "Console Y Size" };

//===========================================================================
// LOCAL FUNCTIONS
//===========================================================================

namespace {



//---------------------------------------------------------------------------
//
void DestroyEventCb(
    GtkWidget* /*widget*/,
    gpointer   /*data  */
)
{

}

} // end of namespace for local functions

//===========================================================================
// CLASS UiConsoleWindow
//===========================================================================

//---------------------------------------------------------------------------
//
UiConsoleWindow::UiConsoleWindow()
    : m_msgCount( 0 ),
      m_window_x( 0 ),
      m_window_y( 0 ),
      m_window_height( 0 ),
      m_window_width( 0 ),
	  m_verbose( false )
{
    ZkGladeLoader loader( "console.xml", "ConsoleWindow" );

    m_window = loader.GetWidget( "ConsoleWindow" );
    m_list   = loader.GetWidget( "CW_List" );

    GtkWidget* scroll = loader.GetWidget( "CW_Scroll" );
    gtk_scrolled_window_set_placement( GTK_SCROLLED_WINDOW( scroll ),
                                       GTK_CORNER_BOTTOM_LEFT );

    // Realize, so we can have a window id.
    gtk_widget_realize( GTK_WIDGET( m_window ) );
    m_xid = GDK_WINDOW_XWINDOW( m_window->window );

    gtk_signal_connect( GTK_OBJECT( m_window ), "delete_event", 
                        GTK_SIGNAL_FUNC( DeleteEventCb ), this );
    gtk_signal_connect( GTK_OBJECT( m_window ), "destroy", 
                        GTK_SIGNAL_FUNC( DestroyEventCb ), 0 );
    gtk_signal_connect( GTK_OBJECT( m_window ), "configure_event", 
                        GTK_SIGNAL_FUNC( ConfigureCb ), this );
    LoadPixmaps();

}

//---------------------------------------------------------------------------
//
UiConsoleWindow::~UiConsoleWindow()
{
    gtk_widget_destroy( m_window );
}


//---------------------------------------------------------------------------
//
void UiConsoleWindow::AddMessage( 
    const ZKSTD::string& msg, 
    ConsImage t
)
{
    if( msg.empty() ) {
        ZKTRACE0( "UiConsoleWindow::AddMessage : Empty message... ignoring." );
        return;
    }

    int ind;
    gchar* cells[CONSOLE_NUM_COLUMNS];

    time_t currentTime;
    time( & currentTime );
    struct tm * currentTimeTm = localtime( & currentTime );
    char timeString[ 100 ] = "";
    strftime( timeString, 100, "%H:%M:%S", currentTimeTm );

    cells[CONSOLE_PIXMAP] = const_cast<gchar*>( "" );
    cells[CONSOLE_TIME] = const_cast<gchar*>( timeString );
    cells[CONSOLE_MSG] = const_cast<gchar*>( msg.c_str() );

    gtk_clist_freeze( GTK_CLIST( m_list ) ); 

    m_msgCount++;
    if( m_msgCount == MAX_MESSAGES )
    {
        gtk_clist_remove( GTK_CLIST( m_list ), 0 );
        m_msgCount--;
    }

    ind = gtk_clist_append( GTK_CLIST( m_list ), cells );
    gtk_clist_set_pixmap( GTK_CLIST( m_list ),
                          ind, CONSOLE_PIXMAP,
                          m_pixmaps[t].first, m_pixmaps[t].second );
    gtk_clist_thaw( GTK_CLIST( m_list ) );
    gtk_clist_moveto( GTK_CLIST( m_list ), 
                      ind, 0, 1, 0 );
}


//---------------------------------------------------------------------------
//
void UiConsoleWindow::LoadPixmaps()
{
    AddPixmap( CONSIMAGE_NORMAL,    "normal.xpm"  );
    AddPixmap( CONSIMAGE_WARNING,   "warning.xpm" );
    AddPixmap( CONSIMAGE_SERIOUS,   "serious.xpm" );
    AddPixmap( CONSIMAGE_QUESTION,  "question.xpm");
    AddPixmap( CONSIMAGE_COMP_TERM, "term.xpm"    );
    AddPixmap( CONSIMAGE_MAIL,      "mail.xpm"    );
    AddPixmap( CONSIMAGE_COOKIE,    "cookie.xpm"  );
    AddPixmap( CONSIMAGE_CONNECTION,"connect.xpm" );
}

//---------------------------------------------------------------------------
//
void UiConsoleWindow::AddPixmap( int index, const char* filename )
{
    GdkPixmap* pixmap;
    GdkPixmap* mask;

    ZKSTD::string fPath = ZKFileConfiguration::GetFreedomPixmapDir();
    fPath += filename;

    pixmap  = gdk_pixmap_create_from_xpm( m_window->window,
                                          &mask,
                                          NULL,
                                          fPath.c_str() );

    m_pixmaps[ index ] = pixmask( pixmap, mask );    

    // FIXME - check this out, what about leaks here ?
    // gdk_pixmap_unref( pixmap );
}

//---------------------------------------------------------------------------
//
void UiConsoleWindow::Show()
{
    m_verbose = IZkRegistry::GetFlag( CONSOLE_VERBOSE_REGISTRY_KEY ) ;

    if( (IZkRegistry::GetDWORD( CONSOLE_X_SIZE, 
           (unsigned long)m_window_width )) &&
        (IZkRegistry::GetDWORD( CONSOLE_Y_SIZE, 
           (unsigned long)m_window_height ))){
        gdk_window_resize( m_window->window, m_window_width, m_window_height );
    }

    if( (IZkRegistry::GetDWORD( CONSOLE_RECT_LEFT_KEY, 
           (unsigned long)m_window_x )) &&
        (IZkRegistry::GetDWORD( CONSOLE_RECT_TOP_KEY, 
           (unsigned long)m_window_y )))
    {
        gtk_widget_set_uposition( m_window , m_window_x, m_window_y );
    }

    ZkGtkUtils::GtkWidgetShow( m_window );
}

//---------------------------------------------------------------------------
//
void UiConsoleWindow::Hide()
{
    gtk_widget_hide( m_window );
}

//---------------------------------------------------------------------------
//
void UiConsoleWindow::update( ZkHint* hint )
{
    ZkConsoleHint* cHint = dynamic_cast< ZkConsoleHint* >( hint );

    // First acquire the lock on the data member.
    ZKS::Lock_Guard< WZkMutex > gate( m_lock );

    if( !cHint || (cHint->getSeverity() == CONS_UI_NOT_CONSOLE) )
    {
		return ;
	}

	if ( !m_verbose && ( cHint->getSeverity() == CONS_DEBUG || cHint->getSeverity() == CONS_VERBOSE ) )
	{
		return ;
	}

    m_HintQueue.push( new ZkConsoleHint( *cHint ));

    // Now if this is a background thread, defer the work otherwise 
    // do it rightaway. Ask not to wait on the UI thread.
    NotifyUi( false );
}

//---------------------------------------------------------------------------
//
void UiConsoleWindow::UiThreadCallback()
{

    // FIXME - hate to do this dynamic_cast buziness on both sides of the 
    //         descriptor. Will have to generalize the copy constructor
    //         of ZkHint.

    // First acquire the lock on the data member.

    ZKS::Lock_Guard< WZkMutex > gate( m_lock );

    ZkHint* pHint = m_HintQueue.front();
    m_HintQueue.pop();

    if( pHint != 0 )
    {
        ZkConsoleHint* cHint = dynamic_cast< ZkConsoleHint* >( pHint );

        if( cHint )
        {
            AddMessage( cHint->getMessage(), cHint->getImageId() );
        }
        delete pHint;
    }
}

void UiConsoleWindow::RegisterHints()
{
    gdk_window_get_deskrelative_origin( m_window->window, & m_window_x,
      & m_window_y );
    gdk_window_get_position( m_window->window, & m_window_x, 
      & m_window_y );
    gdk_window_get_size( m_window->window, & m_window_width, 
      & m_window_height );

    IZkRegistry::SetDWORD(CONSOLE_RECT_LEFT_KEY, 
          (long unsigned) m_window_x );
    IZkRegistry::SetDWORD(CONSOLE_RECT_TOP_KEY, 
          (long unsigned) m_window_y );
    IZkRegistry::SetDWORD(CONSOLE_X_SIZE, 
          (long unsigned) m_window_width );
    IZkRegistry::SetDWORD(CONSOLE_Y_SIZE, 
          (long unsigned) m_window_height );
}

//---------------------------------------------------------------------------
//
gint UiConsoleWindow::DeleteEventCb( 
    GtkWidget* /*widget*/,
    GdkEvent * /*event*/,
    UiConsoleWindow*  inst
)
{
    // This function gets called when the user clicks the corner "close" 
    // button in the title bar. Returning TRUE means we accept the close.
    // In the future we could do some tests here and refuse to close in 
    // certain circumstances.
    // In our case we simply want to hide the console, not close it.
    inst->RegisterHints();
    inst->Hide();
    return( ZK_TRUE );
}

//-------------------------------------------------------------------
//
void UiConsoleWindow::ConfigureCb(
    GtkWidget*,
    GdkEventConfigure*,
    UiConsoleWindow* inst)
{
    inst->RegisterHints();
}
