// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):


//===========================================================================
// EXTERNAL DECLARATIONS
//===========================================================================

#include <UiFreedomManager.h>
#include <UiLoginDlg.h>
#include <UiPopup.h>
#include <UiPrefInit.h>
#include <UiProxyDialogs.h>
#include <ZkGladeLoader.h>
#include <clientdebug.h>
#include <UiThreadSafeDlg.h>
#include <UiThreadSafeMessage.h>
#include <UiRouteDialogs.h>
#include <UiNIQSUpdate.h>


#include <glade/glade.h>
#include <gnome.h>

#include <iostream>

#include <UiMainWindow.h>

// Freedom App. Logic Includes

// zkbase
#include <IZkConsoleMsg.h>
#include <IZkPopup.h>
#include <ZKConfiguration.h>
#include <ZKFileConfiguration.h>
#include <ZkStringMgr.h>

// zkservice
#include <ZkService.h>
#include <INetworkPerformance.h>
#include <IClientPersist.h>
#include <IClientCentral.h>
#include <IFirewall.h>
#include <IRouteManager.h>
#include <IRoutePacket.h>
#include <IKeyQry.h>
#include <ICrypto.h>
#include <ITokenService.h>
#include <INiqs.h>
#include <INymCentral.h>
#include <IScriptServer.h>
#include <IWordScan.h>
#include <IClientInfo.h>

// clienttypes
#include <WClientInfo.h>

// trafficfilter
#include <TrafficFilter.h>

// proxies
#include <DInetD.h>

// error handling
#include <IZkException.h>
#include <IZKErrorReporter.h>
#include <ZkErrorHandler.h>
#include <RZkException.h>

#include <IZkRegistry.h>

#include <sys/stat.h>   // for mkdir
#include <sys/types.h>  // for mkdir
#include <fcntl.h>      // for mkdir
#include <unistd.h>     // for mkdir

#include <signal.h>     // for signal et al

#include <sys/ipc.h>    // for semget
#include <sys/sem.h>    // for semget


// don't ask. will fix this as soon as possible.
// those files only contain static string tables.
// Have to be included in the .cpp so that they get declared only once.
#include <UiMessages.cpp>
#include <ConsMessages.cpp>
#include <MiscMessages.cpp>
#include <ExceptionMessages.cpp>
#include <ZkDialogsStrings.cpp>
#include <ZkFreedomWizardsStrings.cpp>
#include <FreedomClientStrings.cpp>
#include <ClientVersions.h>

// Test Framework stuph
#ifdef INTERNAL_BUILD
	#include "testframework.h"
	#include <test_commands.h>
	#include <zkbase/libzktest/ZkTest.h>
#endif

//===========================================================================
// LOCAL DECLARATIONS
//===========================================================================

namespace {

/*----- variables -----*/

UiFreedomManager* ZkSingleton<UiFreedomManager>::m_instance = 0;

// Does NIQS updates in a background thread while spinning the nullset.
UiNIQSUpdate g_niqsUpdate;

const char g_strClientVersion[] = FREEDOM_CLIENT_NIQS_LINUX;

// Id of the semaphore created to ensure only one instance of Freedom
// is running.
int g_singleInstanceSema = -1;
}

#define PERMS S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH

#define INIT_SCRIPT_REGKEY "RanInitScript"
#define INIT_SCRIPT_FILENAME "inituser.sh"

//===========================================================================
// CLASS UiFreedomManager
//===========================================================================


//---------------------------------------------------------------------------
//
UiFreedomManager::UiFreedomManager()
    : m_bLaunchSetupWizard( true )
   , m_bCallTrafficFilterTerminate( false )
   , m_bCallStopDInetD( false )
   , m_bNeedSaveOnExit( false )
   , m_bCallStopScriptServerUpdate( false )
   , m_NIQSTimer( NULL )
   , m_TheRegistry( NULL )
   , m_UiMessages( 11, UIMSG_SIZE, UiMessages )
   , m_ConMessages( 1,  CONSMESSAGES_SIZE, ConsMessages )
   , m_MiscMessages( 2, MISCMESSAGES_SIZE, MiscMessages )
   , m_DlgMessages( 6,  ZKDIALOGSSTRINGS_SIZE, ZkDialogsStrings )
   , m_WzMessages( 13,  ZKFREEDOMWIZARDSSTRINGS_SIZE, ZkFreedomWizardsStrings )   
   , m_ExceptionMessages( 3,  EXCEPTIONMESSAGES_SIZE, ExceptionMessages )
   , m_ClientMessages( 12,  FREEDOMCLIENTSTRINGS_SIZE, FreedomClientStrings )
   , m_TheErrorReporter( ZKSTD::auto_ptr< ZkAbstractReporter >( new UiErrorReporter() ) )
   , m_ThePopupUtils( ZKSTD::auto_ptr< ZkPopupHandler >( new UiPopup() ) )
   , m_TheProxyDialogs( ZKSTD::auto_ptr< IProxyDialogs >( new UiProxyDialogs() ) )
   , m_TheSetupWizard()
   , m_TheUiThreadSafeMessage()
   , m_TheZKRouteDialogs( ZKSTD::auto_ptr<IRouteDialogs>(new UiRouteDialogs()) )
{
    // empty CTOR
}


//---------------------------------------------------------------------------
//
UiFreedomManager::~UiFreedomManager()
{
    Finalize();
}


//---------------------------------------------------------------------------
//

void * ZK_THREAD_CALLBACK UiFreedomManager::NetworkInitThread( void *arg )
{
    UiFreedomManager *theStartupThing = (UiFreedomManager *)arg;
    if ( !theStartupThing->InitializeNetworking() ) {
        // Get rid of our single instance semaphore.
        semctl( g_singleInstanceSema, 0, IPC_RMID );

        gtk_main_quit();
        exit(-2);
    }
    return NULL;
}

//---------------------------------------------------------------------------
//
bool UiFreedomManager::Initialize()
{
	ZkError err;

    if( IsAlreadyRunning() )
    {
        return false;
    }


    // Initialise Timers

	zkErrorNew(&err);  
    zkITimerInit(err);
    zkErrorFree(&err);

    ZKSTD::string registryPath = ZKFileConfiguration::GetFreedomUserDataDir();

    // Force the directory to exist, so that the registry can go in it.
    // Should this go elsewhere?
    // directory might well already exist, so don't report that error
    if (mkdir(registryPath.c_str(), 0700) < 0 && errno != EEXIST)
      perror("Creating ~/.freedom directory failed");

    registryPath += "freedom.reg";

    m_TheRegistry = new RZkRegistryHolder(registryPath.c_str());
    
    // Bug 7767. Run the $FREEDOM_HOME/inituser.sh script if it is present and
    // has never been run before. This provide a way to initialize default registry
    // settings and other fun stuff before we initialize everything. --ol
    unsigned long dwUserInitialized = 0;
    IZkRegistry::GetDWORD(INIT_SCRIPT_REGKEY, dwUserInitialized);

    if ( !dwUserInitialized ) {
        // We need to run the init script. Tear down the registry before hand
        // to make sure the script can safely modify it. The registry holder
        // is re-instantiated afterwards.
        delete m_TheRegistry;
        RunUserInitScript();
        m_TheRegistry = new RZkRegistryHolder(registryPath.c_str());
        IZkRegistry::SetDWORD(INIT_SCRIPT_REGKEY, 1);
    }


    if( ! InitializeCore() ) {
        // Get rid of our single instance semaphore.
        semctl( g_singleInstanceSema, 0, IPC_RMID );

        return false;
    }

    // Install signal handler so that we manage shutdown
    signal(SIGINT, terminationSignalHandler);
    signal(SIGTERM, terminationSignalHandler);

    m_TheMainWindow.Initialize();
    m_TheMainWindow.Show();

    m_bCallStopScriptServerUpdate = ZkService::GetIScriptServer()->UpdateAllInNonBlockingThread();
    ZKASSERT(m_bCallStopScriptServerUpdate);

    zkErrorNew(&err);
    ZKVERIFY(zkThreadCreateDetached(err, NetworkInitThread, (void *)this));
    zkErrorFree(&err);

    return true;
}


//---------------------------------------------------------------------------
//
bool UiFreedomManager::InitializeCore()
{
	// All actions which occur in this function should run
	// as fast as possible, otherwise there will be a long
	// pause before the UI appears and user will attempt to launch
	// multiple instances.  Longer initialization can occur
	// in InitializeNetworking()
	
	// Check whether or not Freedom is already running.
    // not implemented


    // Set wait cursor.
    // not implemented
    
    if (geteuid() == 0 || getuid() == 0) {
        // Running as root. Bad user, Baaaad user.
        IZkPopup::MessageBox( IDS_FREEDOM_NOROOT, ZK_ERROR ) ;
        exit(-1);
    }

    // Init Error handling API

    ZkErrorHandler::Init();


    // Verify that file configuration is correct - we want to exit if
    // it isn't

    try {
        ZKFileConfiguration::EnsureFileConfigurationValid() ;
    }
	catch(IZkException& ex )
	{
		ZkErrorHandler::GetIZkErrorReporter()->ShowError(ex);
        return false;
	}

	/*
	if( IZkFileHelp::FileExists(ZKFileConfiguration::GetFreedomDataFilePathProd().c_str()) )
    {        
        // FIXME - Beta specific code. Should we remove it ?
        // Ask user to choose between beta or production network.
        ZKSTD::string txt, title;
        ZkStringMgr::Load( txt, IDS_CHOOSE_NETWORK );
        ZkStringMgr::Load( title, IDS_CHOOSE_NETWORK_TITLE );

        GtkWidget* dialog = gnome_message_box_new( 
            txt.c_str(), 
            GNOME_MESSAGE_BOX_QUESTION,
            GNOME_STOCK_BUTTON_YES, 
            GNOME_STOCK_BUTTON_NO, 
            NULL );

        gtk_window_set_modal( GTK_WINDOW( dialog ), TRUE );
        gtk_window_set_title( GTK_WINDOW( dialog ), title.c_str() );

        if( gnome_dialog_run_and_close( GNOME_DIALOG( dialog ) ) )
        {
            ZKConfiguration::SetUseProdNetwork(true);
        }
    }
	*/

    // Init ZKService

    ZkService::Init();

    // Only start console logging after initializing ZkService because the console needs
	// a pointer to IWordScan
    StartConsoleLogging();

    // Bind our RouteObserver
    ZkService::GetIRouteManager()->connect(this);

    // Allow traffic out at startup (although the driver isn't active, so it will be anyway)

    ZkService::GetIFirewall()->EnableFirewall( false );

    // If it is the first time we are running or if setup
    // wasn't completed, open up wizard.

    if ( !ZkService::GetIClientPersist()->IsFirstTimeRun() )
    {
        m_bLaunchSetupWizard = false;
        // Try logging in
        if( !get().Login() ) 
        {
            return false;
        }

		// At this point the full client state inside ZkService (Nyms,
		// tokens, cookies etc, has been loaded into memory)
    }

    return true;
}


//---------------------------------------------------------------------------
//
bool UiFreedomManager::InitializeNetworking()
{
	// This function runs in a background thread so that UI can
	// appear and redraw itself properly during a potentially long
	// startup operation
    
	DZkError derr;
    bool returnValue = true;

    ZKTRACE0( "UiFreedomManager::InitializeNetworking" );

	// Show the basic information about the client that is loading
	IZkConsoleMsg::ShowMsgForceUpdate( IDS_CONS_STARTUP_INIT, g_strClientVersion ) ;
	IZkConsoleMsg::ShowMsg( CONS_NORMAL,  CONSIMAGE_NORMAL, IDS_CONS_STARTUP_NIQSINFO, ZKConfiguration::GetNetworkName().c_str() );

    UiMainWindow::get().SetDisabled( true );	

	try
	{

		// Carefull when Changing the order of these calls!
		
		// Start up the Traffic Filter, which will take care of packets coming in and out.
		// It compiles before ZkService, so it needs a few interface pointers provided to it.


		m_bCallTrafficFilterTerminate = TrafficFilter::theTrafficFilter().Init(
															ZkService::GetIPacketFirewall(), 
															ZkService::GetIRoutePacket() );


		ZkService::GetICrypto()->KeyCacheFromFile(ZKFileConfiguration::GetKeyCacheFilePath().c_str());


		// Initialize the NIQS service
		// This brings us up and running
		// Provided we have enough information caches locally this can succeed even
		// with NIQS server down.  If we don't have enough information then this
		// will throw and Freedom shuts down

		ZkService::GetINiqs()->Init( 
			false, 	// XXX FIXME --ol, andrew
					// This argument allows windows users to avoid their NIQS cache by holding down keys
					// when they start the client.  Because Linux users are expected to understand
					// the file system it is probably acceptable not to support this feature.
			g_strClientVersion ) ;

		// During the v2.1 release this _must_ occur after NIQS init to get
		// the information about what type of route to create #8395
		ZkService::GetIRouteManager()->Init() ;
		
		ZkService::GetIKeyQry()->DownLoadKeysIfRequired();

		if( ZkService::GetIClientInfo()->ClientInfoIsValid() )
		{
			// We know that we have contacted the NIQS, 
			// hence we should have seen some packets through
			// our Firewall.  If no then we are protentially dealing
			// with a configuration that might compromise the IP
			// of the user

			if( !ZkService::GetIFirewall()->IsDriverSendingUpPackets() )
			{
				// We risk compromising privacy if we continue without 
				// driver properly configured - we should have processed
				// some packets by now

				throw RZkException( "bool UiFreedomManager::InitializeNetworking", 
									"No packets have been sent to the Firewall from the driver", 
									IDS_FAILED_CONFIRM_DRIVER_WORKING ) ;
			}


			if ( false == CheckIfClientIsExpired() )
			{
				return false;
			}
		}

		ZkService::GetIFirewall()->Init();

        // This launches the ZkService unit test to create a nym.
		if ( m_bLaunchSetupWizard )
		{		

            if( !get().Setup() )
            {
                return false;
            }

            m_TheMainWindow.InitializeNymList() ;
		}

		// Verify the Life Cycle Condition of the nyms
		// Uses the time value coming from the NIQS so we
		// need the ClientInfo before calling this

		ZkService::GetINymCentral()->InitNymLifeCycle() ;

		// Verify if any nyms need warning messages because they
		// are almost expired or can be completely
		// removed from the client

		for( size_t nIndex = 1 ; 
			 nIndex < ZkService::GetINymCentral()->GetNumberOfNyms() ; 
			 nIndex++ ) 
		{
			ZKSTD::string strUserWarning ;
			ZkService::GetINymCentral()->HandleNymsLifeCycleIssues( nIndex, strUserWarning ) ;
            

			if ( !strUserWarning.empty() )
			{
				IZkPopup::MessageBox( strUserWarning ) ;
                ZKTRACE1( "We should have a Message Box now with message : %s", strUserWarning.c_str() ) ;
			}
		}


        // Explicitly set the current Nym to user before the proxies start. The kernel shim
        // is in set up for initialization (no proxies) up to this point. By chnaging the nym,
        // this also changes. --ol  (FIX - this is ugly reliance on side effect of a call, 
		// probably better to call directly the function to update the shim state - Andrew)
        ZkService::GetINymCentral()->SetActiveNym(USER_NYM_INDEX);

        // Start up DinetD (note, I moved this from InitializeCore, it belongs here
        // --ol
        CDInetDApp::theCDInetDApp().StartDInetD();
        m_bCallStopDInetD = true;

        // Ensure DisplayMOTD() gets called after CheckIfClientIsExpired()
        
        if( ZkService::GetIClientInfo()->ClientInfoIsValid() )
        {
            DisplayMOTD();
        }

        if(!ZkService::GetIClientCentral()->GetPreference(ZK_SHOW_AGAIN_PREF_SET, PREF_NO_SHOW_NFS_WARNING))
		{
			if(IZkPopup::MessageCheckBox(IDS_NFS_WARNING))
			{
				ZkService::GetIClientCentral()->SetPreference(ZK_SHOW_AGAIN_PREF_SET, PREF_NO_SHOW_NFS_WARNING, true);
			}
		}
                
        // TODO: Windows shows total startup time here --ol

        // Show initialization as completed
        IZkConsoleMsg::ShowImportantMsg(CONSIMAGE_NORMAL, IDS_CONS_APPINITCOMPLETE );

        m_bNeedSaveOnExit = true;

        // (If we failed to retrieve client information from the Niqs, we'll use the default update frequency)

		int niqsUpdateFrequence = ZkService::GetIClientInfo()->GetNiqsUpdateFrequency() ;
		ZKASSERT( niqsUpdateFrequence > 0 ) ;
		ZKTRACE1("Gonna do a network update in [%d] minutes\n", niqsUpdateFrequence / 60);
        ZKVERIFY(zkITimerNew(&derr, &m_NIQSTimer, NIQSTimerProc, niqsUpdateFrequence * 1000));
        
        ZkService::GetIFirewall()->EnableFirewall( true ) ;

#ifdef INTERNAL_BUILD
        // This is moved from freedom.cpp
        // (where it was in a wierd place) --ol
        InitForUnitTests();
#endif


	}
	catch(IZkException& ex )
	{
		ZkErrorHandler::GetIZkErrorReporter()->ShowError(ex);

        // will return false, but still needs to do the following lines.
//        returnValue = false;
        // ^^^ no we don't they make us crash. Bug #6291 --ol
        return false;
	}

    UiMainWindow::get().SetDisabled( false );

#warning These calls may cause thread problems.
/*
  This should probably go elsewhere, I think the nym pages can be created
  and added before all the networking is initialized.
  Problems may arise when the dialogs actually get shown, since it is at 
  this time they will try to read information from the servers (or am I wrong...)

  Next step is to separate InitializeNymList and SetActiveNym so that we can do 
  the first before Network initialization.
*/
    UiMainWindow::get().InitializeNymList();

    return returnValue;
}


//---------------------------------------------------------------------------
//
void UiFreedomManager::RunUserInitScript( void )
{
    struct stat stbuf;
    int rval;

    ZKSTD::string strScriptName = ZKFileConfiguration::GetFreedomInstallDir();
    strScriptName += "/" INIT_SCRIPT_FILENAME;
    
    if ( !stat(strScriptName.c_str(), &stbuf) ) {
        // This is generally a horrible call to use, but in this case is appropriate, methinks. --ol.
        rval = system(strScriptName.c_str());
        if ( rval < 0 ) {
            ZKTRACE0("Unable to execv() user initialization script.\n");
        } 
        else if ( rval == 127 ) { 
            ZKTRACE0("Unable to execv() command interpreter.\n");
        }
        // Beyond a bit of error reporting, if this failed, well, too bad.
    }
}

//---------------------------------------------------------------------------
//
bool UiFreedomManager::Run()
{
    m_TheMainWindow.Show();

    GDK_THREADS_LEAVE();
    GDK_THREADS_ENTER();
    gtk_main();
    GDK_THREADS_LEAVE();
    
    m_TheMainWindow.Finalize();

    return true;
}


//---------------------------------------------------------------------------
//
void UiFreedomManager::Shutdown()
{
    // Send extra stats to console log file (but not shown
    // on UI).  If console logging is not turned on this has
    // no ill effect 
    ZkService::GetINetworkPerformance()->ShowCurrentStats( CONS_UI_NOT_CONSOLE ) ;

    if (m_bCallStopScriptServerUpdate)
    {
        ZkService::GetIScriptServer()->AbortIfStillUpdating();
    }
    
    // These have problems shutting down. Temporarily disabled --ol
	if(m_bCallStopDInetD)
	{
		CDInetDApp::theCDInetDApp().StopDInetD();
	}	

	if(m_bCallTrafficFilterTerminate)
	{
		ZkService::GetIRouteManager()->Terminate() ;
		TrafficFilter::theTrafficFilter().Terminate();
	}


	if(m_NIQSTimer)
	{	
        DZkError derr;
        ZKTRACE0("Shutting down NIQS update timer.\n");
		ZKVERIFY(zkITimerFree(&derr, &m_NIQSTimer));
	}

    // Make sure any NIQS update is completed.
       g_niqsUpdate.JoinThread();

	SaveClientState();

    // Unbind our route observer
    ZkService::GetIRouteManager()->disconnect(this);
    
	// It is very important the console logging is stopped before ZkService is
	// terminated because the console holds a pointer to IWordScan
	StopConsoleLogging();
	ZkService::Term() ;

	ZkErrorHandler::Term();

/* Not yet functional --ol
#ifdef INTERNAL_BUILD

	TestFramework::CloseTestWindow();
	
	ZKSTD::string strFullFileName( TestFramework::GetUnitTestExportFileName( false ) );

	if( !strFullFileName.empty() )
		::unlink( strFullFileName.c_str() );
	else
		ZKASSERT( false );
#endif 
*/
    delete m_TheRegistry;

    // Get rid of our single instance semaphore.
    semctl( g_singleInstanceSema, 0, IPC_RMID );
}

//---------------------------------------------------------------------------
//
void UiFreedomManager::Finalize()
{
    // Obviously there isn't much to cleanup at this point.
}


//---------------------------------------------------------------------------
//
void 
UiFreedomManager::SaveClientState()
{
	// Only save the state if we loaded it up properly already
	if (!m_bNeedSaveOnExit)
	{
		return;
	}

    m_bNeedSaveOnExit = false ;

	// We do a complete save, although this is potentially redundant
	// because we also save at other times.
	try
	{
		if ( ZkService::GetIClientPersist() != NULL )
		{
			ZkService::GetIClientPersist()->SaveClientState() ;
		}
    }
	catch(IZkException& /*ex*/){
		// its too late in the game to show a message box with current ex.what() code
		ZKASSERT( false ) ;
	}
}


//---------------------------------------------------------------------------
//
bool UiFreedomManager::Setup()
{

    // Enable firewall so outbound connections to Freedom servers work
    ZkService::GetIFirewall()->EnableFirewall( true );

	if ( !ZkService::GetIClientInfo()->ClientInfoIsValid() )
	{
		// We haven't succeeded in talking with the NIQS.
		// We don't want to continue to allow a setup
		// One particular reason is that we don't know whether to offer
		// free version or not. (see #5170)
		// Another reason is that we don't really know if this 
		// software has expired
		
		IZkPopup::MessageBox( IDS_NO_NIQS_NO_SETUP_WIZ ) ;
		return false ;
	}

    bool ret = m_TheSetupWizard.Run();
    
    ZKTRACE1( "Exited Setup Wizard (%X)\n", ret );

    return ret;
}

//---------------------------------------------------------------------------
//
bool UiFreedomManager::Login()
{
    bool valid = false;

    try {
        UiLoginDlg loginDlg;
        valid = loginDlg.getPassphrase();
    }
    catch( ZkGladeLoaderEx& ){
        ZKTRACE0( "Caught exception: Could not create login dialog." );
    }
    
    return valid;
}

//---------------------------------------------------------------------------
//
bool UiFreedomManager::Splash()
{
    return false;
}


//---------------------------------------------------------------------------
//
void ZK_ITIMER_CALLBACK UiFreedomManager::NIQSTimerProc(void *, unsigned int,
					       unsigned int, unsigned long)
{
    g_niqsUpdate.StartThread();
}

//---------------------------------------------------------------------------
//
void *ZK_THREAD_CALLBACK UiFreedomManager::HandleAuthenticatedRouteFailure( void * )
{
    UiMainWindow::get().SetDisabled( true );
    ZkStatusSubject::UpdateStatus( ZK_STATUS_NETWORK_BUSY, true, true );    
    if ( !ZkService::GetIRouteManager()->AttemptRouteDeathRecovery() )
	{
		// Attempts to build a new route has failed so we must tell the user               
		ZKTRACE0("Authenticated Route died and we were unable to build another automatically\n");        
        IZkPopup::MessageBox(IDS_SECURITY_ISOFF, ZK_WARNING);

		// We wait until the user has "OK'd" the death before reselecting using nym
		// Bug # 862
        UiMainWindow::get().ReselectUserNym();
	}
    ZkStatusSubject::UpdateStatus(ZK_STATUS_NETWORK_BUSY, false, true);
    UiMainWindow::get().SetDisabled( false );
    return NULL;
}

//---------------------------------------------------------------------------
//
void UiFreedomManager::update( ZkHint *hint )
{
    try {
        IRouteManager::RouteHint *routeHint = dynamic_cast<IRouteManager::RouteHint *>(hint);
        
        switch (routeHint->GetStateChange()) {
            case IRouteManager::RouteHint::ROUTE_DEATH_NOTIFICATION :
                {
                    DZkError derr;

                    ZKVERIFY( zkThreadCreateDetached( &derr, HandleAuthenticatedRouteFailure, NULL ) );

                    break;
                }
            default:
                ZKTRACE0("WARNING Unknown notification sent by RouteManager\n");
                break;
        }
    } catch (...) {
        ZKTRACE0("UIFreedomManager::update caught, probably a bad ZkHint downcast\n");
    }
}

//---------------------------------------------------------------------------
//
void 
UiFreedomManager::DisplayMOTD()
{
	ZKSTD::string sMotd = ZkService::GetIClientInfo()->GetMessageOfTheDay();
    ZKSTD::string title;

	switch( ZkService::GetIClientInfo()->GetMessageOfTheDayUrgency() )
	{
	case 1:

		if( !sMotd.empty() ) 
		{ 
            ZkStringMgr::Load( title, IDS_MOTD_TITLE );
			IZkPopup::MessageBox( sMotd, ZK_INFORMATION, OVER_FREEDOM, title );
		}
		break;
	default:
		ZKTRACE0("This network does not support MOTD yet\n");
	};
}

//---------------------------------------------------------------------------
//
int 
UiFreedomManager::CheckIfClientIsExpired()
{
	int bFalseIfWeHaveToShutdown = true;

	if(ZkService::GetIClientInfo()->GetTimeToClientExpiry() <= TIME_TO_UPGRADE)
	{
		try
		{
			// Hide the main UI during the update process - it doesn't provide much
			// value and user might try to click it
//			m_TheMainWindow.Hide();
            bFalseIfWeHaveToShutdown = m_TheUpgradeNotification.Run(ZkService::GetIClientInfo()->GetTimeToClientExpiry());
		}
		
		catch(IZkException& ex)
		{
			ZkErrorHandler::GetIZkErrorReporter()->ShowError(ex);
		}
	}

	return bFalseIfWeHaveToShutdown;
}

//---------------------------------------------------------------------------
//
void UiFreedomManager::terminationSignalHandler(int)
{
    // Clear our single instance semaphore, otherwise Freedom will think
    // there is another instance running the next time we start it.
    semctl( g_singleInstanceSema, 0, IPC_RMID );

	// Make sure that we save our state before exiting
	// when we receive a termination signal
    UiFreedomManager::get().SaveClientState();
    exit(0);
}

//---------------------------------------------------------------------------
//
bool UiFreedomManager::IsAlreadyRunning()
{
    bool bIsRunning = false;

    // Create a semaphore to en sure only one instance 
    // of Freedom is running.
    // If create fails and the error is EEXIST, it is assume Freedom is 
    // already running.
    // FIXME - we may use a better file for the key, but at least this one
    //         will exist if the program has been installed properly.
    ZKSTD::string fPath = ZKFileConfiguration::GetFreedomPixmapDir();
    fPath += "freedom_icon.png";

    ZKVERIFY( IZkFileHelp::FileExists( fPath.c_str() ) );

    key_t semkey = ftok( fPath.c_str(), 1 );
    g_singleInstanceSema = semget( semkey, 1, PERMS | IPC_CREAT | IPC_EXCL );
    if( g_singleInstanceSema == -1 && errno == EEXIST )
    {
// FIXME - because of a dependency on ZkService introduced by 
//         ZkGtkUtils::GetAndSetTopPref, we cannot use this call at this time.
//         Maybe Always On Top should be revisited.
//        IZkPopup::MessageBox( "Freedom is already running.", ZK_ERROR );
        GtkWidget* dialog = gnome_message_box_new( 
            "Freedom is already running.\n\nIf it is not the case, it may be due to Freedom \nexiting abnormally in the past and you may ignore the error.", 
            GNOME_MESSAGE_BOX_ERROR,
            "Exit", 
            "Ignore", 
            NULL );

        int answer = gnome_dialog_run( GNOME_DIALOG( dialog ) );
        switch( answer )
        {
            case 0:
                bIsRunning = true;
                break;

            case 1:
                {
                    int semaID = semget( semkey, 1, PERMS | IPC_CREAT );
                    if( semaID != -1 )
                    {
                        semctl( semaID, 0, IPC_RMID );   
                        g_singleInstanceSema = semget( semkey, 1, PERMS | IPC_CREAT | IPC_EXCL );
                    }
                }
                break;

            default:
                bIsRunning = true;
                break;
        }
    }

    return bIsRunning;
}

//---------------------------------------------------------------------------
//

void UiFreedomManager::StartConsoleLogging( void )
{
    uint32_t doLog = 0;

    IZkRegistry::GetDWORD( CONSOLE_LOG_REGKEY, doLog );

    if (doLog) {
        string fname = ZKFileConfiguration::GetFreedomConsoleLogDir();
      
        // make a time suffix for the file
        time_t currentTime;
        time( &currentTime );
        struct tm *currentTimeTm = localtime( &currentTime );
        char timeString[ 100 ] = "";
        strftime( timeString, 100, "%m-%d-%u--%H-%M-%S", currentTimeTm );
  
        // build a filename (fname is assumed to have a trailing slash)
        fname += CONSOLE_LOG_FILENAME;
        fname += ".";
        fname += timeString;
        fname += ".txt";
        m_TheConsoleLogger.SetFileName(fname);

		IWordScan* pWordScan = NULL;
		pWordScan = ZkService::GetIWordScan();
		ZKASSERT(pWordScan);
		m_TheConsoleLogger.SetIWordScan(pWordScan);

        // Connect the logger to the Console subject
        IZkConsoleMsg::connectHack( &m_TheConsoleLogger );
    }
}


//---------------------------------------------------------------------------
//

void UiFreedomManager::StopConsoleLogging( void )
{
	m_TheConsoleLogger.SetIWordScan( NULL );
    IZkConsoleMsg::disconnectHack( &m_TheConsoleLogger );
}
