/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef ZKS_INCL_UIFREEDOMSTARTUP
#define ZKS_INCL_UIFREEDOMSTARTUP

/*===========================================================================
 * EXTERNAL DECLARATIONS
 *=========================================================================*/

#ifndef ZKS_INCL_ZKSINGLETON
#include <ZkSingleton.h>
#endif

#include <UiMainWindow.h>
#include <UiPrefDlg.h>
#include <UiThreadId.h>
#include <UiThreadSafeMessage.h>
#include <UiErrorReporter.h>
#include <UiAssertReporter.h>
#include <IRouteDialogs.h>

#include <UiUpgradeNotification.h>

#include <../linux/RZkPopup.h>
#include <IZkConsoleMsg.h>
#include <../linux/ZkCatalogRegistry.h>
#include <ZKFileConfiguration.h>
#include <ZkObserver.h>
#include <ZkConsoleLogger.h>

#include <ZkStatusSubject.h>
#include <ZkStatusId.h>
#include <ZkStatus.h>
#include <ZkCatalog.h>
#include <ZkAbstractReporter.h>
#include <ZkProxyDialogs.h>

#include <WzSetupDlg.h>

#include <../linux/RZkRegistryHolder.h>
#include <../linux/RZkFileHelp.h>

// FIXME - message catalogs will have to be registered by the 
//         libraries using them.
#include <UiMessages.h>
#include <ConsMessages.h>
#include <MiscMessages.h>
#include <ExceptionMessages.h>
#include <ZkDialogsStrings.h>

#include <errno.h>
#include <client.h>

#include <iostream>

#include <zkThread.h> // For interval timer support

/*===========================================================================
 * CLASS UiFreedomManager
 *=========================================================================*/

/*
  This file takes care of the client initialization. 
  It performs the differents steps such as login and splash screen before
  getting the UI up on the screen.

*/

class UiFreedomManager
    : public ZkSingleton< UiFreedomManager > 
    , public ZkObserver
{

public: 

    // Single instance management inherited from ZkSingleton.

    /*----- member functions -----*/

    // The constructor basically does nothing. All UI creation and init
    // takes place in the Initilialize function.
    UiFreedomManager();

    // The destructor will do any required cleanup by calling the Finalize
    // function.
    ~UiFreedomManager();

    // Top Level application initialization
    bool Initialize();

    // Run function (should be called right before going into main UI loop)
    bool Run();
  
	// Purpose: Called by main() when we need to shudown the freedom instance
    void Shutdown();

    // Purpose: Make a persistent copy of client state w/ backup
	// Promise: Not to throw.
    void SaveClientState();

	// Purpose:  Post notification when an authenticated route goes down, and reselect user nym.
	//           Used by RouteObserver
    static void* ZK_THREAD_CALLBACK  HandleAuthenticatedRouteFailure( void * );

    void update( ZkHint* );


    // Start and stop logging of console messages to a text file.
    void StartConsoleLogging( void );
    void StopConsoleLogging( void );

private: 

    /*----- member functions -----*/

    // initialize core app services
    bool InitializeCore();

    // initialize networking and driver. get freedom network configuration
    bool InitializeNetworking();

    // Show the setup wizard
    bool Setup();

    // Purpose:  Pops up the Login dialog and returns whether or not the password
    //           was valid on exit.
    bool Login();

	// Purpose:  Check if the client version is expired and initiate retrieving a new one
    //           if it has
    int CheckIfClientIsExpired();

    // Purpose:  Make sure only one instance of Freedom is running.
    //           Uses a semaphore as an ipc flag.
    bool IsAlreadyRunning();

    // Purpose:  Pops up the splash screen.
    //           Does not do any work by itself, the ZKSplash API is used to 
    //           display status and update messages.
    bool Splash();

    // Purpose:  Called by the destructor to do any required cleanup.
    // Promise:  Everything will be squeaky clean.
    void Finalize();

	// Purpose: Timer callback for background NIQS Update routine.
	// Promise: Not to block UI thread. 

    static void ZK_ITIMER_CALLBACK NIQSTimerProc(void *, unsigned int,
					       unsigned int, unsigned long);

	// Purpose: Thread which performs initialization of network while main()
    //   thread is within UI event loop.
	// Promise: Not to throw. Calls exit() on error.
    static void * ZK_THREAD_CALLBACK NetworkInitThread( void *arg );

	// Purpose:  Displays MOTD information retrieved from NIQS
	// Promise:  Not to throw
    void DisplayMOTD( void );

	// Purpose:  Signal handler for SIGTERM and SIGINT
	// Promise:  Not to throw (god only knows what would happen)
	// Requires: Freedom is running (?!)
	// Throws:   Nothing
    static void terminationSignalHandler(int);

	// Purpose:  Run the inituser.sh script if present in $FREEDOM_HOME. This script can
    //           be used to do installation tasks on a per-user basis after the initial 
    //           installation by root.
	// Promise:  Won't throw. Won't fail.
	// Requires: Nothing.
    void RunUserInitScript( void );

    /*----- Member Variables ------*/
    bool m_bLaunchSetupWizard;
    bool m_bCallTrafficFilterTerminate;
    bool m_bCallStopDInetD;
    bool m_bNeedSaveOnExit;
    bool m_bCallStopScriptServerUpdate;

    // Timer for background NIQS updates
    ZkITimer m_NIQSTimer;


    // Yuk. It would be really nice if this didn't have a ctor that needed
    // a prepared value pased in.
    RZkRegistryHolder     *m_TheRegistry;
    RZkFileHelp     m_TheFileHelp;


    // Temporary message catalog.
    // In the future each module should register its own catalog.
    // This can be done by putting this declaration in a .cpp and forcing the
    // linker to resolve all symbols (I thnik... ericb).
    // ------
    // Probably but --whole-archive might also bloat us up quite a lot and generate linkage
    // difficulties --ol
    // ------
    // I've put in a bug about this #8177
    // --ol
    ZkCatalog m_UiMessages;
    ZkCatalog m_ConMessages;
    ZkCatalog m_MiscMessages;
    ZkCatalog m_DlgMessages;
    ZkCatalog m_WzMessages;
    ZkCatalog m_ExceptionMessages;
    ZkCatalog m_ClientMessages;

    UiThreadId      m_TheThreadId; // This is valid only if the thread which constructs this objet
                                 // is also the ui thread.

    // Initialize a few needed singletons.
    // Those instances are to be used through their respective interfaces, not by using the
    // objects themselves. (i.e. do: IZkConsoleMsg::ShowMsg(), not: TheMessageConsole.ShowMsg())
    ZkReporterInstance m_TheErrorReporter;
    UiAssertReporter m_TheAssertReporter;
    RZkPopup        m_ThePopupUtils;
    IZkConsoleMsg   m_TheMessageConsole;
    ZkStatusSubject m_TheStatusSubject;
    ZkProxyDialogs  m_TheProxyDialogs;
    WzSetupDlg      m_TheSetupWizard;
    UiThreadSafeMessage m_TheUiThreadSafeMessage;
    ZKRouteDialogs m_TheZKRouteDialogs;

	// Create an instance of the text file console logger 
	ZkConsoleLogger m_TheConsoleLogger;

    // Create the Preferences singleton and initilize it.
    UiPrefDlg m_ThePrefs;

    // Create The Main Window
    UiMainWindow m_TheMainWindow;

    // Create the upgrade notification dialog
    UiUpgradeNotification m_TheUpgradeNotification;
};

#endif
