// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):



//===========================================================================
// EXTERNAL DECLARATIONS
//===========================================================================

#include <UiIconDisplay.h>
#include <ZkStatusSubject.h>
#include <ZKFileConfiguration.h>
#include <UiThreadId.h>
#include <RZkException.h>
#include <IZkPopup.h>
#include <ZkStringMgr.h>

#include <clientdebug.h>
#include <client.h>
#include <string>

//===========================================================================
// CLASS UiStatusDisplay
//===========================================================================


//---------------------------------------------------------------------------
//
UiIconDisplay::UiIconDisplay( GtkWidget* icon ) throw( IZkException )
    : m_icon( icon )
   , m_pixbuf( 0 )
   , m_anim( 0 )
   , m_frames( 0 )
   , m_curframe( 0 )
   , m_busy_refcount( 0 )
   , m_timeout_tag( -1 ) 
{
    ZKASSERT( m_icon != 0 );

    // Keep widget references around to avoid lifescope issues.
    gtk_widget_ref( m_icon );

    // Load a blank pixmap
    UpdateIcon( NULL );

    // Load the animation from the file
    ZKSTD::string path = ZKFileConfiguration::GetFreedomPixmapDir();
    path += "busy.gif";
    m_anim = gdk_pixbuf_animation_new_from_file(path.c_str());
    
    if (!m_anim) 
    {
        throw RZkException( "UiIconDisplay::UiIconDisplay", "",
                            IDS_ANIMATION_CANNOT_LOAD, path.c_str() );
    }
    
    // Now get the list of frames.
    m_frames = m_curframe = gdk_pixbuf_animation_get_frames( m_anim );
    
    if( !m_frames ) 
    {
        throw RZkException( "UiIconDisplay::UiIconDisplay", "",
                            IDS_ANIMATION_NO_FRAMES );
    }


    // If everything went fine, connect to the status subject, otherwise
    // dont even bother receiving updates.
    ZkStatusSubject::get().connect( this );
}

//---------------------------------------------------------------------------
//
UiIconDisplay::~UiIconDisplay( )
{
    // avoid leaks by releasing our widget references.
    gtk_widget_unref( m_icon );
    if (m_pixbuf) {
        gdk_pixbuf_unref(m_pixbuf);
    }
    if (m_anim) {
        gdk_pixbuf_animation_unref(m_anim);
    }
}


//---------------------------------------------------------------------------
//
// FIXME: double check for pixmap leaks here
void UiIconDisplay::Render( GdkPixbuf *pixbuf, bool overlap )
{
    GdkPixmap *pm = NULL;
    GdkGC *gc;

    if( overlap ) 
    {
        gtk_pixmap_get( GTK_PIXMAP(m_icon), &pm, NULL );
    }

    if( !pm ) 
    {
        // FIXME: there has _got_ to be a better way to make a solid black
        // pixmap.
        GdkColor color;
        gdk_color_black(gdk_rgb_get_cmap(), &color);
        gchar data[40*40];
        memset(data, 0, sizeof(data));
        pm = gdk_pixmap_create_from_data( NULL, data, 40, 40,
                                          gdk_rgb_get_visual()->depth,
                                          &color, &color );
//	    pm = gdk_pixmap_new( NULL, 40, 40,
//				 gdk_rgb_get_visual()->depth );
        gtk_pixmap_set(GTK_PIXMAP(m_icon), pm, NULL);
    }

    // When no pixbuf is given, use the first frame of the animation 
    // by default.    
    if( !pixbuf && m_frames )
    {
        pixbuf = gdk_pixbuf_frame_get_pixbuf((GdkPixbufFrame *)m_frames->data );
    }

    if( pixbuf ) 
    {
        /* Render it to the GtkPixmap */
        gc = gdk_gc_new( pm );

        // If the gdkpixbuf is smaller than the gtkpixmap, handle that case
        gint width = 40;
        gint height = 40;

        if ( gdk_pixbuf_get_width(pixbuf) < 40 ) {
            width = gdk_pixbuf_get_width(pixbuf);
        }
        if ( gdk_pixbuf_get_height(pixbuf) < 40 ) {
            height = gdk_pixbuf_get_height(pixbuf);
        }

        gdk_pixbuf_render_to_drawable_alpha( pixbuf, pm, 0, 0,
                                             (40 - width) / 2,
                                             (40 - height) / 2, width, height,
                                             GDK_PIXBUF_ALPHA_BILEVEL,
                                             100,
                                             GDK_RGB_DITHER_NORMAL, 0, 0 );
        gdk_gc_unref( gc );
        gtk_widget_show( m_icon );
        gtk_widget_queue_draw( m_icon );
    } 
    else 
    {
        gtk_widget_hide( m_icon );        
    }
}

//---------------------------------------------------------------------------
//
gint UiIconDisplay::NextBusyFrame( gpointer data )
{
    UiIconDisplay* inst = (UiIconDisplay *)data;
    GdkPixbufFrame *frame;
    GdkPixbuf *pixbuf;

    // Grab and scale the frame
    frame = (GdkPixbufFrame *)inst->m_curframe->data;
    pixbuf = gdk_pixbuf_frame_get_pixbuf( frame );
    // gdk_pixbuf_frame_get_pixbuf doesn't up the refcount --phil

    // Move to the next frame (or restart the animation loop)
    if (inst->m_curframe->next) {
        inst->m_curframe = inst->m_curframe->next;
    } else {
        inst->m_curframe = inst->m_frames;
    }

    inst->Render(pixbuf, true);

    // Disable and re-set the timer if the delay is different than last time
    // or there is no timer set
    gint old_delay = gdk_pixbuf_frame_get_delay_time(frame);
    frame = (GdkPixbufFrame *)inst->m_curframe->data;

    if (old_delay != gdk_pixbuf_frame_get_delay_time(frame) ||
        inst->m_timeout_tag == -1) {

        if (inst->m_timeout_tag != -1) {
            gtk_timeout_remove(inst->m_timeout_tag);
        }

        inst->m_timeout_tag =
            gtk_timeout_add(gdk_pixbuf_frame_get_delay_time(frame),
                            NextBusyFrame, inst);
    }

    return TRUE;
}

//---------------------------------------------------------------------------
//
void UiIconDisplay::update( ZkHint* hint )
{
    // This function may get called from any thread, so make sure we do not 
    // conflict with the UI thread.
    UiThreadGuard guard;

    ZkStatusHint* cHint = dynamic_cast< ZkStatusHint* >( hint );

    if (cHint == 0) {
        return;
    }

    if (cHint->GetStatusId() == ZK_STATUS_PROGRESS_BUSY ||
        cHint->GetStatusId() == ZK_STATUS_TASKBAR_BUSY ||
        cHint->GetStatusId() == ZK_STATUS_NETWORK_BUSY) 
    {
        if (cHint->GetStatus()) 
        {
            m_busy_refcount++;

            // If we just got to a refcount of 1, start the animation.
            // Higher ref counts will simply let the animation continue.      
            if( m_busy_refcount == 1 ) 
            {
                StartAnimation();
            } 
        } 
        else 
        {
            m_busy_refcount--;

            if( m_busy_refcount == 0 )
            {
                // If we just returned to zero, stop the animation.
                StopAnimation();
            }
            else if (m_busy_refcount < 0) 
            {
                ZKTRACE0("busy refcount < 0?\n");
                m_busy_refcount = 0;
            }
        }
    } 
}

//---------------------------------------------------------------------------
//
void UiIconDisplay::StartAnimation( )
{
    ZKASSERT( "Everything is initialized properly" && ( m_anim && m_frames ));
    
    // Set the animation to the first frame.
    m_curframe = m_frames;
    GdkPixbufFrame *frame = (GdkPixbufFrame *)m_curframe->data;
    int delay = gdk_pixbuf_frame_get_delay_time( frame );

    // Set the timeout to start the animation.
    m_timeout_tag = gtk_timeout_add( delay, NextBusyFrame, this );
}

//---------------------------------------------------------------------------
//
void UiIconDisplay::StopAnimation( )
{
    ZKASSERT( "Everything is initialized properly" && ( m_anim && m_frames ));

    // Turn off the timer and replace the
    // old pixbuf in the icon window.
    if ( m_timeout_tag > -1 ) 
    {
        gtk_timeout_remove( m_timeout_tag );
        m_timeout_tag = -1;
    }

    Render( m_pixbuf, false );
}

//---------------------------------------------------------------------------
//
void UiIconDisplay::UpdateIcon( const char* filename )
{
    GdkPixbuf *pb, *pbs;

    /* Free old pixbuf */
    if( m_pixbuf ) 
    {
        gdk_pixbuf_unref( m_pixbuf );
        m_pixbuf = NULL;
    }


    // If a filename was given, use it. Otherwise set the first frame of 
    // the animation as the default image. (as in windows)
    if( ( filename != 0 ) && ( strlen( filename ) ) )
    {
        // Try loading the file.
        pb = gdk_pixbuf_new_from_file( filename );
        
        if (!pb) 
        {
            ZKSTD::string msg;
            ZkStringMgr::Load( msg, IDS_CANNOT_LOAD_ICON, filename );
            IZkPopup::MessageBox( msg, ZK_ERROR );
        }
        else 
        {
            // At least keep the full sized one for drawing.
            m_pixbuf = pb;

            // Now try resizing the icon.
            pbs = gdk_pixbuf_scale_simple( pb, 40, 40, GDK_INTERP_NEAREST );

            if( pbs ) 
            {
                gdk_pixbuf_unref( pb );
                m_pixbuf = pbs;
            }
        }
    }

    // Draw only if we are not busy.
    if( !m_busy_refcount ) {
        Render( m_pixbuf, false );
    }
}
