// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):



//===========================================================================
// EXTERNAL DECLARATIONS
//===========================================================================

#include <UiErrorReporter.h>
#include <UiThreadId.h>
#include <lockguard.h>
#include <ZkGtkUtils.h>

#include <IZkException.h>
#include <../linux/RZKErrorReporter.h>

#include <gnome.h>
#include <strstream>
#include <string_hlp.h>


//===========================================================================
// CLASS UiErrorReporter
//===========================================================================

//---------------------------------------------------------------------------
//
UiErrorReporter::UiErrorReporter( )
   : m_answer( 0 ),
     m_strCallerFunctionName( "Unknown function" ),
     m_strDetail( "Unknown details" ),
     m_strDescription( "Unknown description" ),
     m_errNum( 0 )
{
}

//---------------------------------------------------------------------------
//
UiErrorReporter::~UiErrorReporter( )
{
}


//---------------------------------------------------------------------------
//
ZkAbstractReporter* UiErrorReporter::clone()
{
    return new UiErrorReporter();
}

//---------------------------------------------------------------------------
//
void UiErrorReporter::ShowError( const IZkException& in_except ) throw()
{
    ZKSTD::ostrstream s;

    // First acquire the lock on the data members so we can safely wait
    // in a thread while the ui does its job.
    ZKS::Lock_Guard< WZkMutex > gate( m_lock );
    
    try
    {
        if( in_except.GetDZkError().getfnc().size() > 0 )
        {
            m_strCallerFunctionName = in_except.GetDZkError().getfnc();
            s << " [" << m_strCallerFunctionName << "] ";
        }

        if( in_except.GetDZkError().geterrcode() > 0 )
        {
            m_errNum = in_except.GetDZkError().geterrcode();
            s << " [Error Code: " << m_errNum << "] ";
        }

        if( in_except.GetDZkError().getdetail().size() > 0 )
        {
            m_strDetail = in_except.GetDZkError().getdetail();
            s << " [" << m_strDetail << "] ";
        }
 
    }
    catch(...)
    {
        ZKASSERT( 0 );
    }


    m_strDescription = in_except.GetFormattedString();
    if( m_strDescription.size() > 60 ) 
    {
        ZKS::breakStringIntoLines( m_strDescription );
    }

    // Now if this is a background thread, defer the work otherwise 
    // open the dialog rightaway.
    NotifyUi();
}


//---------------------------------------------------------------------------
//
void UiErrorReporter::UiThreadCallback()
{

// In internal mode we still keep the ability to abort from the error dialog,
// while in release mode we only show an OK button to close the dialog.

#ifdef _DEBUG

    const gchar * strButtonsDetailsOff[] = 
       { "Abort", "Ignore", "Details", NULL };
    const gchar * strButtonsDetailsOn[] = 
       { "Abort", "Ignore", "No Details", NULL };

    bool detailsAnswer = true;
    bool showDetails = false;

    while( detailsAnswer ) {
       if( showDetails ) {
          ZKSTD::string strLongWinded = m_strDescription;
          strLongWinded += "\n\n<<< Details >>>"; 
          strLongWinded += "\nCalled from: ";
          strLongWinded += m_strCallerFunctionName ;
          strLongWinded += "\nDetails: ";
          strLongWinded += m_strDetail;
          ZKS::breakStringIntoLines( strLongWinded );
          m_dialog = gnome_message_box_newv( strLongWinded.c_str(),
                     GNOME_MESSAGE_BOX_WARNING, strButtonsDetailsOn );
       }
       else
       {
          m_dialog = gnome_message_box_newv( m_strDescription.c_str(),
                     GNOME_MESSAGE_BOX_WARNING, strButtonsDetailsOff );
       }
       gtk_window_set_position( GTK_WINDOW( m_dialog ), GTK_WIN_POS_CENTER );

       ZkGtkUtils::GtkWidgetShow( m_dialog );
       m_answer = gnome_dialog_run( GNOME_DIALOG( m_dialog ) );

       switch( m_answer ){
           case 0:
               abort();
               detailsAnswer = false;
               break;
  
           case 2: {
               showDetails = ! showDetails;
               detailsAnswer = true;
               break;
           }
           case 1:
           default:
               /* User closed dialog with window manager,
                * assume "ignore"
                */
               detailsAnswer = false;
               break;
       }
    }

#else

    m_dialog = gnome_message_box_new(
        m_strDescription.c_str(),
        GNOME_MESSAGE_BOX_ERROR,
        GNOME_STOCK_BUTTON_OK,
        NULL);

    ZkGtkUtils::GtkWidgetShow( m_dialog );
    gnome_dialog_run_and_close( GNOME_DIALOG( m_dialog ) );

#endif
}
