// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):



//===========================================================================
// EXTERNAL DECLARATIONS
//===========================================================================

#include <UiPopup.h>
#include <UiThreadId.h>
#include <ZkGtkUtils.h>
#include <lockguard.h>

#include <gnome.h>
#include <string_hlp.h>



//===========================================================================
// LOCAL DECLARATIONS
//===========================================================================

namespace {

const gchar* IconTypeToStockIcon( IconType icon )
{
    const gchar* ret;

    switch( icon ) 
    {
        case ZK_INFORMATION:
            ret = GNOME_MESSAGE_BOX_INFO;
            break;

        case ZK_WARNING:
            ret = GNOME_MESSAGE_BOX_WARNING;
            break;
 
        case ZK_MESSAGE:
            ret  = GNOME_MESSAGE_BOX_INFO;
            break;

        case ZK_ERROR:
            ret = GNOME_MESSAGE_BOX_ERROR;
            break;

        case ZK_QUESTION:
            ret = GNOME_MESSAGE_BOX_QUESTION;
            break;
    }        

    return ret;
}

} // end local namespace


//===========================================================================
// CLASS UiPopup
//===========================================================================

//---------------------------------------------------------------------------
//
UiPopup::UiPopup( )
    : m_messageCheckPopupDlg( new UiCheckBoxPopup() ),
      m_questionCheckBoxPopupDlg( new UiQuestionCheckBoxPopup() )
{
}

//---------------------------------------------------------------------------
//
UiPopup::~UiPopup( )
{
}

//---------------------------------------------------------------------------
//
void UiPopup::openMessageBox( const tstring& msg, const tstring& title, IconType iconType )
{
    // FIXME - For now it will compile as long as we do not use unicode defines.

    // First acquire the lock on the data members so we can safely wait
    // in a thread while the ui does its job.
    ZKS::Lock_Guard< WZkMutex > gate( m_lock );

    // Then create the dialog box, we need it whether or not this is a background
    // thread.
    { // scope for thread guard
        UiThreadGuard tguard;

        ZKS::breakStringIntoLines( msg, 60 );
        m_dialog = gnome_message_box_new( 
            msg.c_str(), 
            IconTypeToStockIcon( iconType ),
            GNOME_STOCK_BUTTON_OK, 
            NULL );
        
        gtk_window_set_modal( GTK_WINDOW( m_dialog ), TRUE );
        gtk_window_set_title( GTK_WINDOW( m_dialog ), title.c_str() );

    } // scope for thread guard

/* 
   if (parent != NULL) {
     gnome_dialog_set_parent(GNOME_DIALOG(m_dialog),parent);
   }
*/ 

    // Now if this is a background thread, defer the work otherwise 
    // open the dialog rightaway.
    NotifyUi();
}

//---------------------------------------------------------------------------
//
bool UiPopup::openQuestionBox( 
    const tstring& msg, 
    BtnType btnType, 
    IconType iconType, 
    bool defBtn )
{
    // FIXME - For now it will compile as long as we do not use unicode defines.

    // First acquire the lock on the data members so we can safely wait
    // in a thread while the ui does its job.
    ZKS::Lock_Guard< WZkMutex > gate( m_lock );

    { // scope for thread guard
        UiThreadGuard tguard;

        // Then create the dialog box, we need it whether or not this is a 
        // background thread.
        ZKS::breakStringIntoLines( msg, 60 );
        switch( btnType ) 
        {
            case ZK_OKCANCEL:
                m_dialog = gnome_message_box_new( 
                    msg.c_str(), 
                    IconTypeToStockIcon( iconType ),
                    GNOME_STOCK_BUTTON_OK, 
                    GNOME_STOCK_BUTTON_CANCEL, 
                    NULL );
                break;
                
            case ZK_YESNO:
                m_dialog = gnome_message_box_new( 
                    msg.c_str(), 
                    IconTypeToStockIcon( iconType ),
                    GNOME_STOCK_BUTTON_YES, 
                    GNOME_STOCK_BUTTON_NO, 
                    NULL );
                break;
        }
        
        gnome_dialog_set_default( GNOME_DIALOG( m_dialog ), defBtn );
        
        gtk_window_set_modal( GTK_WINDOW( m_dialog ), TRUE );

    } // scope for thread guard

/* 
   if (parent != NULL) {
     gnome_dialog_set_parent(GNOME_DIALOG(m_dialog),parent);
   }
*/ 

    // Now if this is a background thread, defer the work otherwise 
    // open the dialog rightaway.
    NotifyUi();

    return !m_answer;
}

//---------------------------------------------------------------------------
//
bool UiPopup::openMessageCheckBox(	
    const tstring& in_strMsg,
    const tstring& in_strCheckMsg,
    bool	  in_bDefaultChecked, 
    IconType  in_iconType
)
{
    return m_messageCheckPopupDlg->Run(in_strMsg, 
                                       in_strCheckMsg,
                                       in_bDefaultChecked,
                                       in_iconType);
}

//---------------------------------------------------------------------------
//
bool UiPopup::openQuestionCheckBox(
    const tstring& in_strQuestion,
    bool&       out_bChecked,
    uint32_t    in_btn1textId,
    uint32_t    in_btn2textId,
    uint32_t    in_checkMsgId,
    bool        in_bDefaultChecked,
    IconType    in_iconType
)
{
    return m_questionCheckBoxPopupDlg->Run(in_strQuestion,
                                           out_bChecked,
                                           in_btn1textId,
                                           in_btn2textId,
                                           in_checkMsgId,
                                           in_bDefaultChecked,
                                           in_iconType );
}

//---------------------------------------------------------------------------
//
tstring UiPopup::openRequestStringBox(	
	const tstring& /*in_strTitle*/,
    const tstring& /*in_strPrompt*/,
    const tstring& /*in_strInitialValue*/,
    int            /*in_nMinLength*/,
    PopupPosition /*in_popupPos*/ )
{
    ZKASSERT(false && "RequestStringBox not yet implemented on Linux");
    return("");
}
//---------------------------------------------------------------------------
//
void UiPopup::UiThreadCallback()
{
    ZkGtkUtils::GtkWidgetShow( m_dialog );
    m_answer = gnome_dialog_run_and_close( GNOME_DIALOG( m_dialog ) );
}
