/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */
  
#ifndef ZKS_INCL_UITHREADBASE
#define ZKS_INCL_UITHREADBASE

/*===========================================================================
 * EXTERNAL DECLARATIONS
 *=========================================================================*/

#include <zkThread.h>


/*===========================================================================
 * CLASS UiThreadBaseThread
 *=========================================================================*/
/*
  This is a class only so it can be friends with UiThreadBase and
  calls its private functions. Which are private mainly to prevent class users
  from shooting themselves in the foot.
*/
class UiThreadBaseThread
{

public:
    
    /*----- member functions -----*/

    // Purpose: Execute the nym creation activities from a background thread
    //          so we can animate the nullset while Freedom is busy.
    static void* ZK_THREAD_CALLBACK Execute( void* klass );
};



/*===========================================================================
 * CLASS UiThreadBase
 *=========================================================================*/

/*
  This class provides a common interface to all classes that wish to 
  support a background thread. It groups all tasks common to such support:
  A class that needs to start a thread, have the UI wait or it to end and
  do some work when it is finished can override the function described here:
*/

class UiThreadBase {

    friend class UiThreadBaseThread;

public: 

    /*----- member functions -----*/

    UiThreadBase();
    virtual ~UiThreadBase();

    // Purpose: Starts the background thread, which will execute
    //          DoThreadWork, then EndThread. The main difference between
    //          DoThreadWork and EndThread is that one is called by the new 
    //          thread while the other is called within the UI thread.
    void StartThread( );

    // Purpose: Allows to join the thread created by StartThread.
    void JoinThread();

protected: 

    /*----- member functions -----*/

    // Purpose: Called by StartThread to set states prior to staring the thread.
    //          Disables the given widget (usually the main container of the 
    //          window), starts the busy animation and changes the cursor..
    //          You can override this function to do some more work or change
    //          the behavior.
    virtual void OnThreadStart();

    // Purpose: Called on DoThreadWork exit. This funciton gets called from the
    //          UI thread so that states can be reset and widget values modified 
    //          safely. NOOP by default.
    virtual void OnThreadEnd();

    // Purpose: Even though the work is done in the background, it often needs
    //          to access data members. This function will get called from a
    //          thread other than the UI main thread, so make sure it is thread 
    //          safe.
    virtual void OnThreadExecute();

    // Purpose: Although the class catches exceptions, you can call this function
    //          to know whether or not the thread execution was successfull and
    //          take appropriate action.
    bool WasSuccessful() const;


private: 

    /*----- static member functions -----*/

    // Purpose: Does cleanup work and exits from the local message loop.
    //          Cannot be overriden mainly to protect future users against 
    //          themselves.
    static void EndThread( void* klass );

    // Purpose: A local message loop has to be started from the ui thread, so 
    //          this is really the only reason for such a function.
    static void MessageLoop( void* );

    /*----- member functions -----*/

    // Purpose: Called by the background thread to perform the actual work.
    //          Executes OnThreadExecute() in a try/catch block.
    void DoThreadWork();
      
    /*----- data members -----*/

    bool     m_bSuccess;
    ZkThread m_thread;
};


//---------------------------------------------------------------------------
//
inline bool UiThreadBase::WasSuccessful() const
{
    return m_bSuccess;
}

#endif
