/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef ZKS_INCL_UITHREADSAFEDLG
#define ZKS_INCL_UITHREADSAFEDLG

/*===========================================================================
 * EXTERNAL DECLARATIONS
 *=========================================================================*/

#include <WSyncTypes.h>
#include <gtk/gtk.h>

/*===========================================================================
 * CLASS UiThreadSafeDlg
 *=========================================================================*/

/*
  The purpose of this base class is to provide a framework to easily implement
  thread safe dialogs.
  Thread safe dialogs are pieces of ui that can be opened (and waited on) from
  background threads.

  The dialog has two phases, one in which the info to be displayed is prepared,
  the other in which the dialog is actually shown. This separation is due to 
  the fact that the background thread knows what needs to be displayed, while 
  the ui thread is the only one that can actually show it.

  You can develop a thread safe dialog by overriding the NotifyUi and 
  UiThreadCallback functions. The rest is taken care of by the base class.

  This class also provides a lock to protect data hwne it is ready to be 
  displayed. For example, let;s say yo uneed do display a string. You would:

  a) acquire the m_lock 
  b) set the content of the string (a member of your class)
  c) call NotifyUi to wake up the Ui thread.
     This call blocks, waiting for the UiThreadCallback function to return.
     You can use the blockAndWait argument to the function call to tell 
     NotifyUi not to wait.
  d) In the UiThreadCallback function DO NOT acquire the lock, it is held by
     the other thread. The reason for this lock is that if another thread
     tries to display a message, it will block wiating for the other message
     to be displayed. Otherwise you risk changing the content of the string
     before it gets displayed. (or worse, 'while' it gets diplayed.
  e) When you return from UiThreadCallback, the othe thread is unblocked.
     
  UiThreadSafeDlg uses a WZkEvent to do the synchronisation between threads.
  You will usually want to block when the Ui thread displays a message box
  and you need need to wait for an answer from the user.

  Checkout the UiSignedMail dialog for a simple and complete example.


  Additional Notes on Thread Issues:


  GDK_THREADS_ENTER/GDK_TREADS_LEAVE

  These macros should be used when you need to do some Ui work from a thread.
  In theory they should ensure no other thread accesses Xlib calls while you
  hold the lock. Unfortunately it turns out we often end up with the
  "Unexpected Async Reply" mesage, which indicates a thread conflict during
  calls to Xlib.

  The Ui lock is obtained automatically for you by GTK when you are within 
  an event callback (signal handler in GTK terms). So not only do you not need
  to obtain the lock, but you will also DEADLOCK if you try getting it.
  If you dont know whether a function is going to be called from a thread or
  from GTK, use the UiThreadGuard class. It will obtain the lock only if you
  are not in the ui thread.

  While using the UiThreadSafeDlg, you do not need to obtain the lock,
  the cass gets it for you before calling the UiThreadCallback function.
  
  Most of the thread UI work is now done through the UiThreadSafe class because
  the calls that are actually safe to make from a separate thread seem to be 
  limited at best. This class uses file descriptors to get the ui thread to 
  call our callback. It creates a pipe, then registers the read descriptor with 
  the GTK message loop (gtk_input_add). When GTK is idle, it will gt woken up by 
  a write to the file descriptor and will call the corresponding callback.

  All Proxy dialogs, as well as the Console are implemented using this class.
  
  


*/

class UiThreadSafeDlg {


protected: 


    /*----- member functions -----*/

    UiThreadSafeDlg();
    virtual ~UiThreadSafeDlg();

    virtual void UiThreadCallback() = 0;

    void NotifyUi( bool blockAndWait = true );

    /*----- data members -----*/

    WZkMutex  m_lock;

private: 

    /*----- static member functions -----*/

    static void Dispatch(
        gpointer /*data*/,
        gint /*source*/,
        GdkInputCondition /*condition*/
    );


    /*----- member functions -----*/

    /*----- data members -----*/

    WZkEvent  m_event;
    int       m_descr[2];
    int       m_input;

};
#endif
