// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include "UiThreadSafeMessage.h"
#include <UiThreadId.h>

#include <clientdebug.h>

#include <unistd.h>

UiThreadSafeMessage* ZkSingleton< UiThreadSafeMessage >::m_instance = 0;

//===========================================================================
// CLASS UiThreadSafeMessage
//===========================================================================

//---------------------------------------------------------------------------
//
UiThreadSafeMessage::UiThreadSafeMessage( )
{
    ZKVERIFY( "UiThreadSafeMessage - create pipe" &&
	      pipe( get().m_descr ) == 0 );

    get().m_input = gdk_input_add( get().m_descr[0],
				   GDK_INPUT_READ,
				   Dispatch, NULL);
}

//---------------------------------------------------------------------------
//
UiThreadSafeMessage::~UiThreadSafeMessage( )
{
    gdk_input_remove( get().m_input );
    close(get().m_descr[0]);
    close(get().m_descr[1]);
}

//---------------------------------------------------------------------------
//
void UiThreadSafeMessage::HandleMessage( gpointer func, gpointer data )
{
    if (!func) {
	ZKTRACE0("UiThreadSafeMessage::HandleMessage; func=NULL\n");
	return;
    }

    //ZKTRACE2("HandleMessage: about to call %p(%p)\n", func, data);

    (*(void (*)(gpointer))func)(data);
}

//---------------------------------------------------------------------------
//
void UiThreadSafeMessage::Dispatch( gpointer, gint, GdkInputCondition )
{
    ZKASSERT( "This is the UI." && UiThreadId::IsUiThread() );

    gpointer infunc, indata;

    // Read the pointers from the pipe
    while( read( get().m_descr[0], &infunc, sizeof(infunc) ) == -1 &&
	   errno == EINTR ) ;
    while( read( get().m_descr[0], &indata, sizeof(indata) ) == -1 &&
	   errno == EINTR ) ;

    //ZKTRACE2("Dispatch: I'm the UI thread!  I got %p,%p\n", infunc, indata);

    GDK_THREADS_ENTER();
    get().HandleMessage(infunc, indata);
    GDK_THREADS_LEAVE();

    // Unblock background thread waiting for the popup to close.
    get().m_event.Set();
}

//---------------------------------------------------------------------------
//
void UiThreadSafeMessage::NotifyUi( gpointer func, gpointer data,
				    bool blockAndWait )
{
    if( UiThreadId::IsUiThread() ) {
	//ZKTRACE0("NotifyUI: I -am- the input thread; handling.\n");
        get().HandleMessage(func, data);
    } else {
        get().m_event.Reset();

	//ZKTRACE2("NotifyUI: sending %p,%p to UI.\n", func, data);

	// Write the pointers to the pipe
	int retval;
	while ( (retval = write( get().m_descr[1], &func, sizeof(func) ))
		== -1 && errno == EINTR ) ;
	while ( (retval = write( get().m_descr[1], &data, sizeof(data) ))
		== -1 && errno == EINTR ) ;
	//ZKTRACE2("retval: %d, errno: %d\n", retval, errno);
        ZKVERIFY( retval != -1);
        if ( blockAndWait )
            get().m_event.Wait();
    }
}
