// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):



//===========================================================================
// EXTERNAL DECLARATIONS
//===========================================================================


#include <ZkGladeLoader.h>
#include <ZKFileConfiguration.h>
#include <clientdebug.h>

#include <string>

//===========================================================================
// LOCAL DECLARATIONS
//===========================================================================

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/*
  We use a custom function to create our pixmap widgets so that we can
  decide from which directory the pixmap will be taken.
  (unless, of course, someting escaped me in the libglade options...)
*/

 
static GtkWidget *
pixmap_build( GladeXML */*xml*/, GladeWidgetInfo *info )
{
	GtkWidget *pix;
	GList *tmp;
	GdkPixmap *pixmap;
	GdkBitmap *bitmap = NULL;
	char *filename = NULL;
  
	for (tmp = info->attributes; tmp; tmp = tmp->next) 
    {
		GladeAttribute *attr = (GladeAttribute *)(tmp->data);

		if (!strcmp(attr->name, "filename")) 
        {
			if( filename ) g_free( filename );

			/*filename = glade_xml_relative_file(xml, attr->value);*/
            if( g_path_is_absolute( attr->value )) 
            {
                filename = g_strdup( attr->value );
            }
            else 
            {
                /* prepend XML file's dir onto filename */
                std::string fPath = ZKFileConfiguration::GetFreedomPixmapDir();
                fPath += attr->value; 

                filename = g_strdup( fPath.c_str() );
            }
		}
	}
	
    pixmap = gdk_pixmap_colormap_create_from_xpm(
        NULL,
		gtk_widget_get_default_colormap(), 
        &bitmap, 
        NULL, 
        filename
    );

	if (filename)
    {
		g_free(filename);
    }
	if (pixmap)
    {
		pix = gtk_pixmap_new( pixmap, bitmap );
    }
	else
    {
		pix = (GtkWidget*)gtk_type_new( gtk_pixmap_get_type() );
    }

	if (pixmap) gdk_pixmap_unref(pixmap);
	if (bitmap) gdk_bitmap_unref(bitmap);

//  FIXME - check whether or not this will make us miss any alignments.
//	misc_set(GTK_MISC(pix), info);

	return pix;
}

static const GladeWidgetBuildData widget_data[] = {
    /* general widgets */
    {const_cast<char*>("GtkPixmap"),  pixmap_build,  NULL},
    {NULL, NULL, NULL}
};


#ifdef __cplusplus
}
#endif /* __cplusplus */


//===========================================================================
// CLASS ZkGladeLoader
//===========================================================================

bool ZkGladeLoader::m_initialized = false;
GladeWidgetBuildData ZkGladeLoader::m_data;

//---------------------------------------------------------------------------
//
ZkGladeLoader::ZkGladeLoader( )
{
    // I do hate those init things, but why do this repeatedly ?
    if( !m_initialized ) 
    {
        glade_register_widgets( widget_data );    
        m_initialized = true;
    }
}

//---------------------------------------------------------------------------
//
ZkGladeLoader::ZkGladeLoader( 
    const char* in_xmlfile, 
    const char* in_component 
) throw( ZkGladeLoaderEx )
{
    ZKASSERT( (in_xmlfile != 0) && (in_component != 0) );

    // I do hate those init things, but why do this repeatedly ?
    if( !m_initialized ) 
    {
        glade_register_widgets( widget_data );    
        m_initialized = true;
    }

    LoadFile( in_xmlfile, in_component );
}

//---------------------------------------------------------------------------
//
void ZkGladeLoader::LoadFile( 
    const char* in_xmlfile, 
    const char* in_component 
) throw( ZkGladeLoaderEx )
{
    std::string xmlFullPath( ZKFileConfiguration::GetFreedomXMLDir() );
    xmlFullPath += in_xmlfile;

    // FIXME - add some check here on the validity of the file.
    m_xml = glade_xml_new( xmlFullPath.c_str(), in_component );

    if( !m_xml ) {
        ZKTRACE1( "XML Loader: could not open file %s\n", xmlFullPath.c_str() );
        throw ZkXMLFileNotFoundEx();
    }
}

//---------------------------------------------------------------------------
//
ZkGladeLoader::~ZkGladeLoader()
{
    // Get rid of the parsed xml.
    if( m_xml ) {
        gtk_object_unref( GTK_OBJECT( m_xml ) );    
    }
}

//---------------------------------------------------------------------------
//
GtkWidget* ZkGladeLoader::GetWidget( 
    const char* in_widget 
) const 
{
    GtkWidget* widget = glade_xml_get_widget( m_xml, in_widget );

    if( !widget ) {
        ZKTRACE2( "XML Loader: could not find widget named %s in file %s\n", 
                  in_widget, m_xml->filename );
        throw ZkXMLWidgetNotFoundEx();
    }

    return widget;
}


