// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):



//===========================================================================
// EXTERNAL DECLARATIONS
//===========================================================================

#include "UiOutgoingMail.h"
#include <UiThreadId.h>
#include <ZkDialogsStrings.h>
#include <ZkDialogsHelpers.h>
#include <ZkGtkUtils.h>

#include <ZkStringMgr.h>
#include <ZkGladeLoader.h>
#include <lockguard.h>

#include <ZkService.h>
#include <INym.h>
#include <INymCentral.h>
#include <IClientCentral.h>
#include <IZkPopup.h>

#include <clientdebug.h>
#include <vector>

#include <gnome.h>
// FIXME - put this in a catalog.
static const char g_strYourTrueIdentity[] = "Your true identity";

//===========================================================================
// CLASS UiOutgoingMail.cpp
//===========================================================================

//---------------------------------------------------------------------------
//
UiOutgoingMail::UiOutgoingMail( )
   : m_pVecAddresses( 0 )
   , m_fSMTP( false )
   , m_pStrUserEmailAddr( NULL )
   , m_bShowCheckBox( true )
   , m_answer( 0 )
{
    // Call that a folk legend if you want, but it seems X really doesn't
    // appreciate top level windows being created in threads.
    // (which makes sense, given the fact X is not really thread safe)
    ZKASSERT( UiThreadId::IsUiThread() );
    
    ZkGladeLoader loader( "outgoing_mail.xml", "window" );

    m_window     = loader.GetWidget( "window" );
    m_nyms       = loader.GetWidget( "nyms" );
    m_recipients = loader.GetWidget( "recipients" );
    m_header     = loader.GetWidget( "msg" );
    m_dontshow   = loader.GetWidget( "dont_show" );

    gtk_widget_hide( GTK_WIDGET( m_window ) );
}

//---------------------------------------------------------------------------
//
UiOutgoingMail::~UiOutgoingMail( )
{
}

//---------------------------------------------------------------------------
//
long UiOutgoingMail::Run( 
    ZKSTD::vector<ZKSTD::string>* pVecAddresses, 
    bool in_fSMTP, /* whether or not this is News */
    const ZKSTD::string* in_pStrUserEmailAddr /* NULL if mail server is mail.freedom.net */
)
{
	ZKASSERT( !pVecAddresses->empty() );
    m_pStrUserEmailAddr = in_pStrUserEmailAddr;
    m_bShowCheckBox = true;

    long nSendFromNymIndex = -1;

    if (!ZkDialogsHelpers::ShouldShowOutgoingMailDlg(
        (m_pStrUserEmailAddr == NULL),
        m_fSMTP,
        nSendFromNymIndex,
        m_bShowCheckBox))
    {
        return nSendFromNymIndex;
    }

    // First acquire the lock on the data members so we can safely wait
    // in a thread while the ui does its job.
    ZKS::Lock_Guard< WZkMutex > gate( m_lock );

    m_pVecAddresses = pVecAddresses;
    m_fSMTP = in_fSMTP;
    m_answer = 0;

    // Now if this is a background thread, defer the work otherwise 
    // open the dialog rightaway.
    NotifyUi();

    return m_answer;
}

//---------------------------------------------------------------------------
//
void UiOutgoingMail::UiThreadCallback()
{
    // First get rid of everything in this menu.
    gtk_option_menu_remove_menu( GTK_OPTION_MENU( m_nyms ) );

    // Alway reset the toggle, if we dgot here it means the option is off.
    gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( m_dontshow ), FALSE );
    
    // Now get a new menu, fill it, and put it back.
    GtkWidget* newMenu = gtk_menu_new();

	size_t iActiveNym = ZkService::GetINymCentral()->GetActiveNymIndex() ;
	size_t cntNyms    = ZkService::GetINymCentral()->GetNumberOfNyms() ;

	ZKASSERT( iActiveNym < cntNyms ) ;

	ZKSTD::string strNymName;

    int nymCnt = 0;

    if(!m_fSMTP)
    {
        gtk_menu_append( GTK_MENU( newMenu ), 
                         GetNewMenuItem( const_cast<char*>( g_strYourTrueIdentity )
                                         , USER_NYM_INDEX ));
    }
    else if (m_pStrUserEmailAddr)
    {
        ZKSTD::string strUserNymLabel;
        if (*m_pStrUserEmailAddr != "")
        {
            strUserNymLabel += *m_pStrUserEmailAddr;
            strUserNymLabel += " - ";
        }
        strUserNymLabel += g_strYourTrueIdentity;

        gtk_menu_append( GTK_MENU( newMenu ), 
                         GetNewMenuItem( const_cast<char*>( strUserNymLabel.c_str() )
                                         , USER_NYM_INDEX ));
    }
    else 
    {
        // When there is no True Identity menu item, we have to offset nym count
        // by one in order to select the right menu item.
        nymCnt = -1;
    }

	size_t iSelectedNym = USER_NYM_INDEX ;

	INym* pNym = NULL ;

	for ( size_t i = 1; i < cntNyms; i++)
	{					 
		ZkService::GetINymCentral()->GetNym( IID_INYM, (void**)&pNym, i );

		unsigned long dwLifeStage = NYM_LIFE_STAGE_UNDEFINED ;
		pNym->GetYourLifeStage( dwLifeStage ) ;

		if ( ( dwLifeStage != NYM_LIFE_STAGE_ACTIVE ) ||
             ( !pNym->DoesNymUseMail()) )
        {
			// We don't add non active nyms
            // or nyms that dont use email.
			continue ;
		}

		pNym->GetNymName( strNymName );

		ZKSTD::string  strNymDomain ;

		// Method name is misleading - this is the nyms domain
		// name (eg trial.freedom.net)
		ZKVERIFY( pNym->GetNymEmailDomain( strNymDomain ) );

		strNymName.append( "@" );
		strNymName.append( strNymDomain );

        GtkWidget* item = GetNewMenuItem( const_cast< char* >( strNymName.c_str() ), i );
        gtk_menu_append( GTK_MENU( newMenu ), item );

        nymCnt++;

		if ( i == iActiveNym ) {
			// Remember what index was the active nym
			iSelectedNym = nymCnt ;
		}
	}

	ZKASSERT( iSelectedNym <= iActiveNym ) ;

    gtk_menu_set_active(GTK_MENU( newMenu ), iSelectedNym);
    gtk_option_menu_set_menu( GTK_OPTION_MENU( m_nyms ), newMenu );


	if ( !m_fSMTP )
	{
		// Change the UI for handling news
		ZKSTD::string strHeader;
		ZKSTD::string strTitle;
        ZkStringMgr::Load( strHeader, IDS_OUTGOING_NEWS_HEADER );
        ZkStringMgr::Load( strTitle, IDS_OUTGOING_NEWS_TITLE );

        gtk_window_set_title( GTK_WINDOW( m_window ), strTitle.c_str() );
        gtk_label_set( GTK_LABEL( m_header ), strHeader.c_str() );
    }

    // Show Recipients
	ZKSTD::string strRecipients;

    // Empty current list.
    gtk_list_clear_items( GTK_LIST( m_recipients ), 0, -1 );
    GtkWidget* li;

    ZKSTD::vector<ZKSTD::string>::iterator it;
	for	(it = m_pVecAddresses->begin(); it != m_pVecAddresses->end(); it++)
    {
        li = gtk_list_item_new_with_label (const_cast< gchar *>((*it).c_str()));
        gtk_widget_show( li );
        gtk_container_add( GTK_CONTAINER( m_recipients ), li );
    }

    ZkGtkUtils::GtkWidgetShowAll( m_window );

    if (!m_bShowCheckBox)
    {
        gtk_widget_hide_all(m_dontshow);
    }

    int answer = gnome_dialog_run( GNOME_DIALOG( m_window ) );

    switch( answer )
    {
        case 0: /* Send Message */
            gnome_dialog_close( GNOME_DIALOG( m_window ) );
            m_answer = (long)gtk_object_get_data( GTK_OBJECT(gtk_menu_get_active( GTK_MENU( newMenu) )), "nym_index" );
            break;
            
        case 1: /* Block Message */
            gnome_dialog_close( GNOME_DIALOG( m_window ) );
            m_answer = -1;
            break;
            
        case -1: /* ESC or wm close */
            m_answer = -1;
            break;                
    }        

    // Check whether or not the user wnats to be bothered all the time.
    bool noshow = gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( m_dontshow ));

    if( noshow ) 
    {
        if( m_fSMTP )
        {
            ZkService::GetIClientCentral()->SetPreference(
                ZK_SHOW_AGAIN_PREF_SET, 
                PREF_NO_SHOW_OUTGOING_MAIL, 
                true );
        }
        else
        {
            ZkService::GetIClientCentral()->SetPreference(
                ZK_SHOW_AGAIN_PREF_SET, 
                PREF_NO_SHOW_OUTGOING_NEWS, 
                true );
        }
    }

}

//---------------------------------------------------------------------------
//
GtkWidget* UiOutgoingMail::GetNewMenuItem( char* label, int index )
{
    GtkWidget* item = gtk_menu_item_new_with_label( label );
    
    // Set the given index as the item data.
    gtk_object_set_data( GTK_OBJECT( item ), "nym_index", (void*)index ); 

    gtk_widget_show( item );

    return item;
}
