// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):



//===========================================================================
// EXTERNAL DECLARATIONS
//===========================================================================

#include <UiPrefBase.h>
#include <UiPrefDlg.h>
#include <ZkGladeLoader.h>
#include <ZKFileConfiguration.h>

#include <clientdebug.h>
#include <clientstl.h>


//===========================================================================
// CLASS UiPrefBase
//===========================================================================

//---------------------------------------------------------------------------
//
UiPrefBase::UiPrefBase( 
    const ZKSTD::string& name,
    const ZKSTD::string& desc
)
    : m_ct( 0 )
   ,  m_subCt( 0 )
   ,  m_label( 0 )
   ,  m_name( name )
   ,  m_desc( desc )
   ,  m_bDelete( false )
   ,  m_bModified( true )
{
    ZKASSERT( (!name.empty()) && (!desc.empty()) );

    ZkGladeLoader loader( "prefbase.xml", "PrefBase_ct" );

    // Create base container.
    m_ct    = loader.GetWidget( "PrefBase_ct" );
    m_subCt = loader.GetWidget( "PrefBase_dlg" );
    m_label = loader.GetWidget( "PrefBase_label" );


    m_backCt = loader.GetWidget( "PrefBase_labelBg" );
    
    // Redefine a style for our label here, so that it uses a bigger font
    // and a darker background color.
    // FIXME - check how this can survive or react correctly to a theme change.
    //         also check whether or not I should unref this style.
    // FIXME - find a way to set the background color here !!&^@#%$

    GdkColor darkgray  = { 0, 0x0000, 0x0000, 0x0000 };
    GdkColor lightgray = { 0, 0xAAAA, 0xAAAA, 0xAAAA };
    gdk_color_alloc( gdk_colormap_get_system(), &darkgray );
    gdk_color_alloc( gdk_colormap_get_system(), &lightgray );

    GtkStyle* wStyle = gtk_widget_get_style( m_label );
    GtkStyle* newStyle = gtk_style_copy( wStyle );

    // this font seems to be more standard
    GdkFont* newFt = gdk_font_load( "-*-helvetica-bold-r-normal-*-18-*-*-*-*-*-*-*" );

    newStyle->fg[ GTK_STATE_NORMAL ]   = darkgray;
    newStyle->text[ GTK_STATE_NORMAL ] = darkgray;
    newStyle->bg[ GTK_STATE_NORMAL ]   = darkgray;
    newStyle->base[ GTK_STATE_NORMAL ]   = darkgray;
    if( newFt )
    {
        newStyle->font = newFt;
    }
  
    gtk_widget_set_style( m_backCt,  newStyle );
    gtk_widget_set_style( m_label, newStyle );
    
    // Keep our own instance so we control the container life scope.
    gtk_widget_ref( m_ct );
    gtk_widget_unparent( m_ct );

    gtk_label_set_text( GTK_LABEL( m_label ), desc.c_str() );
}

//---------------------------------------------------------------------------
//
UiPrefBase::~UiPrefBase()
{
    // well.. not cosher, but should get me the parent
    // gtk_widget_unparent has unpleasnat side effects....so do it from the parent
    if( m_ct->parent != 0 )
    {
        gtk_container_remove( GTK_CONTAINER( m_ct->parent ), m_ct );
    }
}

//---------------------------------------------------------------------------
//
void UiPrefBase::setParent( GtkWidget* parent )
{
    ZKASSERT( parent != 0 );
    ZKASSERT( GTK_IS_CONTAINER( parent ) );

    gtk_container_add( GTK_CONTAINER( parent ), m_ct );
    gtk_widget_unref( m_ct );
}

//---------------------------------------------------------------------------
//
void UiPrefBase::SetPixmap( GdkPixmap* gdkPixmap,
                            GdkBitmap* gdkBitmapMask)
{
    if( gdkPixmap )
    {
         GtkWidget* pixmap = gtk_pixmap_new (gdkPixmap,gdkBitmapMask);
         gtk_widget_show (pixmap);
         gtk_box_pack_start (GTK_BOX (m_backCt), pixmap, FALSE, FALSE, 0);
    }
}


//---------------------------------------------------------------------------
//
void UiPrefBase::SetTitle( const char* title )
{
    gtk_label_set_text( GTK_LABEL( m_label ), title );
}

//---------------------------------------------------------------------------
//
void UiPrefBase::hide() 
{
    ZKASSERT( m_ct );
    gtk_widget_hide( m_ct );
}

//---------------------------------------------------------------------------
//
void UiPrefBase::show() 
{
    ZKASSERT( m_ct );
    gtk_widget_show( m_ct );
}

//---------------------------------------------------------------------------
//
void UiPrefBase::SetTopContainer( GtkWidget* ct )
{
    gtk_widget_ref( ct );
    gtk_widget_unparent( ct );
    gtk_container_add( GTK_CONTAINER( m_subCt ), ct );
    gtk_widget_unref( ct );
}

//---------------------------------------------------------------------------
//
void UiPrefBase::Initialize()
{
}

//---------------------------------------------------------------------------
//
bool UiPrefBase::Finalize()
{
    return true;
}

//---------------------------------------------------------------------------
//
void UiPrefBase::Cancel()
{
}

//---------------------------------------------------------------------------
//
void UiPrefBase::MarkForDelete()
{
    m_bDelete = true;
    UiPrefDlg::DisablePage( m_name );
}

//---------------------------------------------------------------------------
//
bool UiPrefBase::IsMarkedForDelete() const
{
    return m_bDelete;
}

//---------------------------------------------------------------------------
//
void UiPrefBase::SetModified( bool done )
{
    m_bModified = done;
}

//---------------------------------------------------------------------------
//
bool UiPrefBase::WasModified() const
{
    return m_bModified;
}

//---------------------------------------------------------------------------
//
GtkWidget* UiPrefBase::GetContainer() const
{
    return m_subCt;
}


//---------------------------------------------------------------------------
//
const ZKSTD::string& UiPrefBase::GetName() const
{
    return m_name;
}

//---------------------------------------------------------------------------
//
const ZKSTD::string& UiPrefBase::GetDesc() const
{
    return m_desc;
}

