/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef ZKS_INCL_UIPREFBASE
#define ZKS_INCL_UIPREFBASE

/*===========================================================================
 * EXTERNAL DECLARATIONS
 *=========================================================================*/

#include <clientstl.h>

#include <string>
#include <gnome.h>

/*===========================================================================
 * CLASS UiPrefBase
 *=========================================================================*/

/*
  Base class to implement pages for the Preferences dialog.
  Provides the initial container in which to put the rest of the page UI.
  Also provides a standard label and could contain more default widgets in 
  the future.

  All UiPrefBases have the same default initial size (will decide size later)
  but they can be stretched to any size so they should handle a bigger size
  gracefully.
*/

class UiPrefBase {

public: 

    /*----- member functions -----*/

	// Purpose:  Creates a container in which the content of a preferences
    //           page can be put. The name is used by the preferences dialog
    //           to identify individual page.
    //           When the same page is used for multiple nyms, the name
    //           is NymName-PageName. 
    // FIXME -   make sure the preceding sentence is true and supported.
    UiPrefBase( 
        const ZKSTD::string& name,
        const ZKSTD::string& desc
    );

    virtual ~UiPrefBase();

	// Purpose:  Adds the dialog container to the children of the given widget.
	// Requires: The given widget obvioulsy has to be a container of some sort.
    void setParent( GtkWidget* );
    
    // Purpose:  Set the top-right image of the page.  Example : each nym page has 
    //           a lock icon under each page if the root is a pro nym.
    // Require:  no, you can pass null if you want
    void SetPixmap( GdkPixmap* lockPixmap,
                    GdkBitmap* lockBitmapMask);

	// Purpose:  Return the name and description strings respectively.
	// Promise:  String is not null, enforced by asserts in the constructor.
    const ZKSTD::string& GetName() const;    
    const ZKSTD::string& GetDesc() const;

	// Purpose:  Used by the parent Preferences dialog to show only one 
    //           dialog at time.
    void hide();
    void show();

    // Purpose:  This function gets called on every preference page whenever
    //           the Preferences dialog is opened. This way every page can reset
    //           its contents according to the most recent update of zkservices.
    // Requires: By the time this function gets called, zkservices should be
    //           initialized.
    virtual void Initialize();

    // Purpose:  This function gets called when OK gets clicked to close the
    //           Preferences dialog. Each dialog which has temporary changes
    //           can then save them permenently. 
    //           The boolean return value can be used by the page to interrupt
    //           the OK and keep the dialog opened. (false will stop the process)
    virtual bool Finalize();

	// Purpose:  This function gets called when the user cancels the changes
    //           made in the Preferences. All non-permanent changes will be
    //           discarded. The function can be used to reset any flags or
    //           values.
    virtual void Cancel();
    
	// Purpose:  Return a pointer to the main dialog container.
    // Promise:  Will not be null, since creating the dialog creates the 
    //           container.
    GtkWidget*   GetContainer() const;

	// Purpose:  Sets the text at the top of the page (in bold).
    void         SetTitle( const char* );

    // Purpose:  Marks a page for deletion by the prefs dialog.
    //           Page cannot delete themselves because they are referenced in 
    //           a vector in the prefs dialog, so they get marked for deletion
    //           and get destroyed when the dialog is closed. In the meantime
    //           the entire page gets disabled.
    void         MarkForDelete( );
    bool         IsMarkedForDelete() const;

    // Purpose:  Dialogs can use this flag to keep track of whether or not data
    //           has been modified in the dialog and not been applied.
    void         SetModified( bool );
    bool         WasModified() const;

protected: 

    /*----- static member functions -----*/


    /*----- member functions -----*/

	// Purpose:  Indicates to the base class that what widget is to be considered
    //           the top level container for this dialog. 
    //           It parents the given widget in the base m_subCt.
    void SetTopContainer( GtkWidget* ct );


private: 

    /*----- member functions -----*/

	// Prohibited funcs - not implemented.
	UiPrefBase( const UiPrefBase& );
	const UiPrefBase& operator=( const UiPrefBase& );

    /*----- data members -----*/
    
    GtkWidget*   m_ct;    // top container
    GtkWidget*   m_subCt; // sub container for derived classes contents.
    GtkWidget*   m_label;
    GtkWidget*   m_backCt;
    
    ZKSTD::string  m_name;
    ZKSTD::string  m_desc;

    bool m_bDelete;   // marks a page for deletion by the prefs dialog.
    bool m_bModified; // indicates whether or not a page is done modifying info.
};



#endif
