// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):



//===========================================================================
// EXTERNAL DECLARATIONS
//===========================================================================

#include <UiPrefDlgImp.h>
#include <UiPrefBase.h>
#include <UiPrefInit.h> // FIXME - remove this dependency

#include <clientdebug.h>
#include <ZkGtkUtils.h>
#include <ZkGladeLoader.h>
#include <glade/glade.h>

#include <vector>
#include <cstdio>

//===========================================================================
// LOCAL DECLARATIONS and FUNCTIONS
//===========================================================================

// This ID is used to stored a dialog associated with a tree entry right in
// the tree entry using gtk_object_set_data.
#define DATA_ID "dialog"

int offset;


//===========================================================================
// CLASS UiPrefDlgImp
//===========================================================================

//---------------------------------------------------------------------------
//
UiPrefDlgImp::UiPrefDlgImp( )  
    : m_currentDialog( 0 )
    , m_max_w( 0 )
    , m_max_h( 0 )
{
    ZkGladeLoader loader( "preferences.xml", "PrefsDialog" );

    m_dialog = loader.GetWidget( "PrefsDialog" );
    m_tree   = loader.GetWidget( "Prefs_tree" );
    m_ct     = loader.GetWidget( "Prefs_container" );
    
    gtk_signal_connect (GTK_OBJECT( m_dialog ), "close",
                        GTK_SIGNAL_FUNC( UiPrefDlgImp::CloseCb ), NULL );

    gtk_tree_set_selection_mode( GTK_TREE( m_tree ),
                                 GTK_SELECTION_SINGLE );
    gtk_tree_set_view_mode( GTK_TREE( m_tree ), GTK_TREE_VIEW_ITEM );

}

//---------------------------------------------------------------------------
//
UiPrefDlgImp::~UiPrefDlgImp()
{
}

//---------------------------------------------------------------------------
//
void UiPrefDlgImp::Open( const ZKSTD::string& page )
{
    bool closeIt = false;
    ZKSTD::string activePage( "preferences" );

    if( m_currentDialog != 0 )
    {
        activePage = m_currentDialog->GetName();
        m_currentDialog->hide();
        m_currentDialog = 0;
    }

    // Make sure we clear the selection before we actually delete the items
    // from the tree.
    GList* dlist = GTK_TREE( m_tree )->selection;

    if( dlist )
    {
        GtkObject* tree_item = GTK_OBJECT( dlist->data );
        gtk_item_toggle( GTK_ITEM( tree_item ) );
    }

    // These two calls are needed mostly because we do not know about
    // changes to the nym list and have to make sure it is up to date.
    // It could be improved by a subject/observer pattern in ZkService.
    UiPrefInit::RemoveNymPages();
    UiPrefInit::AddNymPages();

    // We have to make this call early mostly because gtl_tree will crash
    // if we try to set its selection before it gets realized.
    ZkGtkUtils::GtkWidgetShow( m_dialog );

    // Then set the active page. Either to the last active page, or to a new page
    // if one was given.
    if( !page.empty() )
        activePage = page;

    SetActiveDialog( activePage );


    // Make sure every dialog has read the updated info from ZkServices.
    // Initialization is separate from page construction because the ui
    // oten gets created long before ZkService has been initialized.
    InitializeAll( );

    do 
    {
        int answer = gnome_dialog_run( GNOME_DIALOG( m_dialog ));
        switch( answer )
        {
            case 0: /* OK */
                // Pages can refuse closing by returning false in their 
                // Finalize call.
                closeIt = FinalizeAll(  );
                break;
            
            case 1: /* Cancel */
            case -1: /* ESC or wm close */
                //CancelAll(); // not used at the moment.
                closeIt = true;
                break;

            default:
                ZKASSERT( "UiPrefDlgImp::Open - Unknown case." && 0 );
                break;
        }  
    }
    while( !closeIt );

    gnome_dialog_close( GNOME_DIALOG( m_dialog ) );
}

//---------------------------------------------------------------------------
//
gint UiPrefDlgImp::CloseCb( 
    GtkWidget */*widget*/,
    gpointer   /*data*/ )
{
    return( FALSE ); 
}


//---------------------------------------------------------------------------
//
void UiPrefDlgImp::DestroyCb( 
    GtkWidget *widget,
    UiPrefDlgImp* inst )
{
    GtkWidget* w = GTK_BIN( widget )->child;
    char* label;
    gtk_label_get( GTK_LABEL( w ), &label );
    ZKTRACE1( "Item [%s] is getting destroyed\n", label );

    UiPrefBase* dlg  = static_cast< UiPrefBase* > (gtk_object_get_data( 
        GTK_OBJECT( widget ), DATA_ID ) );        

    if( dlg ) {
        if( dlg == inst->m_currentDialog ) {
            inst->SetCurrentDialog( 0 );
        }

        gtk_object_remove_data( GTK_OBJECT( widget ), DATA_ID );

        // also remove it from the map.
        TreeMap::iterator iter = inst->m_treeMap.find( dlg->GetName() );
        ZKASSERT( iter != inst->m_treeMap.end() );
        if(  iter != inst->m_treeMap.end() )
        {
            inst->m_treeMap.erase( iter );
        }

        ZKTRACE1( "   deleting page [%s] as a result\n", dlg->GetName().c_str());
        delete dlg;
    }            
}


//---------------------------------------------------------------------------
//
void UiPrefDlgImp::AddDialog( 
    const ZKSTD::string& label,
    ZKSTD::auto_ptr<UiPrefBase> dialog, 
    const ZKSTD::string& nodeName
)
{
    ZKASSERT( !label.empty() );
    ZKASSERT( dialog.get() != NULL );

    UiPrefBase* dlg = dialog.release();

    dlg->setParent( m_ct );
    dlg->hide();

    // Create a new tree item, give it the label.
    // the cast here is really just to get rid of a compiler warning...
    GtkWidget* item = gtk_tree_item_new_with_label( 
        const_cast< gchar* >( label.c_str() ) );

    TreePair treePair( item, dlg );
    m_treeMap[ dlg->GetName() ] = treePair;

    // connect to its select signal.

    gtk_signal_connect( GTK_OBJECT( item ), "select",
                        GTK_SIGNAL_FUNC( UiPrefDlgImp::SelectCb ), this );
    gtk_signal_connect (GTK_OBJECT( item ), "destroy",
                        GTK_SIGNAL_FUNC( UiPrefDlgImp::DestroyCb ), this );

    if( nodeName.empty() ) {
        gtk_tree_append( GTK_TREE( m_tree ), item );
    }
    else {
        GtkWidget* node  = FindPage( nodeName );
        ZKASSERT( "UiPrefDlgImp: Parent page found." && ( node != 0 ));

        // FIXME - do something if not found.
        if( node != 0 ) {
        
            GtkWidget* subtree = GTK_TREE_ITEM_SUBTREE( node );
            
            if( !subtree ) {
                subtree = gtk_tree_new();
                gtk_tree_set_selection_mode( GTK_TREE(subtree),
                                             GTK_SELECTION_SINGLE );
                gtk_tree_set_view_mode( GTK_TREE(subtree), GTK_TREE_VIEW_ITEM );

                gtk_tree_item_set_subtree( 
                    GTK_TREE_ITEM( node ), 
                    subtree );
            }

            gtk_tree_append( GTK_TREE( subtree ), item );
            gtk_tree_item_expand( GTK_TREE_ITEM( node ) );
        }
    }

    gtk_widget_show( item );

    // Set the dialog pointer as the widget data.
    gtk_object_set_data( GTK_OBJECT( item ), DATA_ID, dlg ); 

    // Make sure we allocate enough space in our container to accomodate 
    // for all pages.
    GtkRequisition size_requisition;

    gtk_widget_size_request( 
        GTK_WIDGET( dlg->GetContainer() ), 
        &size_requisition );

    m_max_w = MAX( m_max_w, size_requisition.width );
    m_max_h = MAX( m_max_h, size_requisition.height );
         
    // Now set the size hint for our window.
    gtk_widget_set_usize( GTK_WIDGET( m_ct ), m_max_w, m_max_h );    
}

//---------------------------------------------------------------------------
//
void UiPrefDlgImp::DeleteSubPages( const ZKSTD::string& name )
{
    GtkWidget* w = FindPage( name );

    if( w != 0 ) 
    {
        GtkWidget* subtree = GTK_TREE_ITEM_SUBTREE( w );
        if( subtree != 0 )
        {
            // This call will, in turn, cause the DestroyCb to be
            // called on each tree leaf, and will delete the associated
            // page.
            gtk_tree_item_remove_subtree( GTK_TREE_ITEM( w ) );
        }
    }    
}

//---------------------------------------------------------------------------
//
void UiPrefDlgImp::DisablePage( const ZKSTD::string& name )
{
    GtkWidget* item = FindPage( name );

    if( item != 0 ) 
    {
        // Extract the dialog from the item.
        UiPrefBase* dlg  = static_cast< UiPrefBase* > (gtk_object_get_data( 
            GTK_OBJECT( item ), DATA_ID ) );

        gtk_widget_set_sensitive( dlg->GetContainer(), FALSE );


        GtkWidget* subtree = GTK_TREE_ITEM_SUBTREE( item );
        
        if( subtree != 0 ) 
        {
            GList* children = gtk_container_children( GTK_CONTAINER( subtree ));

            while( children ) {
                GtkWidget* item = GTK_WIDGET( children->data );
                dlg  = static_cast< UiPrefBase* > (gtk_object_get_data( 
                    GTK_OBJECT( item ), DATA_ID ) );
                
                gtk_widget_set_sensitive( dlg->GetContainer(), FALSE );
                children = g_list_next( children );
            }
        }
    }

}

//---------------------------------------------------------------------------
//
GtkWidget* UiPrefDlgImp::FindPage( const ZKSTD::string& pageName )
{
    GtkWidget* found = 0;

    TreeMap::iterator iter = m_treeMap.find( pageName );
    if( iter != m_treeMap.end() )
        found = (*iter).second.first;

    return found;
}

    
//---------------------------------------------------------------------------
//
void UiPrefDlgImp::SetActiveDialog( const ZKSTD::string& name )
{
    if( name.empty() )
    {
        return;
    }

    // Dont bother if the window is not yet mapped.
    if( !m_tree->window )
        return;

    // First find the dialog, then make it current.
    GtkWidget* page = FindPage( name );

    ZKASSERT( "UiPrefDlgImp: Active page found" && ( page != 0 ));

    if( page )
    {
        gtk_item_toggle( GTK_ITEM( page ) );

// this function call gives out a gtk critical warning.
//        gtk_tree_select_child( GTK_TREE( m_tree ), page );

        UiPrefBase* dlg  = static_cast< UiPrefBase* > (gtk_object_get_data( 
            GTK_OBJECT( page ), DATA_ID ) );
        SetCurrentDialog( dlg );
    }
}

//---------------------------------------------------------------------------
//
void UiPrefDlgImp::SetCurrentDialog( UiPrefBase* dlg )
{
//    ZKASSERT( dlg );

    if( dlg != m_currentDialog ) {
        if( m_currentDialog ) {
            m_currentDialog->hide();
        }

        m_currentDialog = dlg;

        if( m_currentDialog ) {
            m_currentDialog->show();
        }
    }    
}


//---------------------------------------------------------------------------
//
void UiPrefDlgImp::SelectCb (
    GtkWidget*    child, 
    UiPrefDlgImp* inst
)
{
    // Extract the dialog from the item.
    UiPrefBase* dlg  = static_cast< UiPrefBase* > (gtk_object_get_data( 
        GTK_OBJECT( child ), DATA_ID ) );

    inst->SetCurrentDialog( dlg );
}


//---------------------------------------------------------------------------
//
void UiPrefDlgImp::InitializeAll()
{
    TreeMap::iterator iter;

    for( iter = m_treeMap.begin(); iter != m_treeMap.end(); ++iter )
    {
        ((*iter).second.second)->Initialize();
    }
}

//---------------------------------------------------------------------------
//
bool UiPrefDlgImp::FinalizeAll( )
{
    bool result = true;
    TreeMap::iterator iter;

    for( iter = m_treeMap.begin(); (iter != m_treeMap.end()) && result; ++iter )
    {
        result = ((*iter).second.second)->Finalize();        
    }

    return result;
}


//---------------------------------------------------------------------------
//
void UiPrefDlgImp::CancelAll( )
{
    TreeMap::iterator iter;

    for( iter = m_treeMap.begin(); iter != m_treeMap.end(); ++iter )
    {
        ((*iter).second.second)->Cancel();
    }
}


