// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):



//===========================================================================
// EXTERNAL DECLARATIONS
//===========================================================================

#include <UiPrefPassphrase.h>
#include <UiPrefDlg.h>
#include <UiPasswordConfirm.h>
#include <UiMessages.h>
#include <ZkGladeLoader.h>
#include <IZkPopup.h>
#include <ZkDialogsStrings.h>


#include <ICrypto.h>
#include <ZkService.h>
#include <ZkErrorHandler.h>
#include <IZKErrorReporter.h>
#include <IZkException.h>
#include <clientdebug.h>

//===========================================================================
// CLASS UiPrefPassphrase
//===========================================================================

//---------------------------------------------------------------------------
//
UiPrefPassphrase::UiPrefPassphrase( 
    const ZKSTD::string& name,
    const ZKSTD::string& desc 
) 
    : UiPrefBase( name, desc )
{
    ZkGladeLoader loader( "prefpassphrase.xml", "topct" );

    SetTopContainer( loader.GetWidget( "topct" ) );

    m_frame     = loader.GetWidget( "passphrase_ct" );
    m_protect   = loader.GetWidget( "protect" );
    m_update    = loader.GetWidget( "update" );
    m_currentLb = loader.GetWidget( "currentLb" );
    m_current   = GTK_ENTRY( loader.GetWidget( "current" ));
    m_new       = GTK_ENTRY( loader.GetWidget( "new" ));
    m_confirm   = GTK_ENTRY( loader.GetWidget( "confirm" ));

    // Make sure users can't turn the *** off and see the atcual text...
    gtk_entry_set_visibility( m_current, FALSE );
    gtk_entry_set_visibility( m_new,     FALSE );
    gtk_entry_set_visibility( m_confirm, FALSE );

    gtk_signal_connect (GTK_OBJECT( m_update ), "clicked",
                        GTK_SIGNAL_FUNC( UpdateCb ), this);
    m_signalId = gtk_signal_connect (GTK_OBJECT( m_protect ), "toggled",
                        GTK_SIGNAL_FUNC( ProtectCb ), this);
}

//---------------------------------------------------------------------------
//
UiPrefPassphrase::~UiPrefPassphrase( )
{
}

//---------------------------------------------------------------------------
//
void UiPrefPassphrase::Initialize()
{
    bool hasPassphrase = UserHasPassphrase();

    gtk_signal_disconnect( GTK_OBJECT( m_protect ), m_signalId );
    gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( m_protect ), hasPassphrase );
    m_signalId = gtk_signal_connect (GTK_OBJECT( m_protect ), "toggled",
                        GTK_SIGNAL_FUNC( ProtectCb ), this);

    // We connect only after initializing the toggle's state, otherwise
    // we call the callback at the wrong time.

    ShowPassphraseWidgets( hasPassphrase, hasPassphrase );
}

//---------------------------------------------------------------------------
//
bool UiPrefPassphrase::Finalize()
{
    // Make sure the user has actually entered a passphrase if he chose 
    // to enable it.
    if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( m_protect ) ) &&
        !UserHasPassphrase() )
    { 
        if( IZkPopup::QuestionBox( 
            "You have chosen to enable passphrase protection\nbut you have not entered a passphrase.\n\nDo you wish to enter one now ?",
            ZK_YESNO,
            ZK_WARNING ) )
        {
            UiPrefDlg::SetActiveDialog( GetName() );
            return false;
        }
    }    

    return true;
}

//---------------------------------------------------------------------------
//
void UiPrefPassphrase::ProtectCb( GtkWidget*, UiPrefPassphrase* inst )
{
    if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( inst->m_protect ) ) )
    {
        inst->ShowPassphraseWidgets( true, inst->UserHasPassphrase( ) );
    }
    else if( inst->UserHasPassphrase( ) ) 
    {
		ZKSTD::string currentPP, defaultPP;

        UiPasswordConfirm passwd;

        if( passwd.Run() ) 
        {
            ZkService::GetICrypto()->GetPassphrase( currentPP );
            ZkService::GetICrypto()->GetDefaultPassphrase( defaultPP );

            try
            {
                ZkService::GetICrypto()->ChangePassphrase( currentPP, defaultPP ) ;
            } 
            catch(IZkException& ex)
            {
                ZkErrorHandler::GetIZkErrorReporter()->ShowError( ex );
            }
            
            inst->ShowPassphraseWidgets( false );
        }
        else 
        {
            gtk_signal_disconnect( GTK_OBJECT( inst->m_protect ), inst->m_signalId );

            gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( inst->m_protect ), true );

            inst->m_signalId = gtk_signal_connect (GTK_OBJECT( inst->m_protect ), "toggled",
                                             GTK_SIGNAL_FUNC( ProtectCb ), inst);
        }
    }
    else 
    {
        inst->ShowPassphraseWidgets( false );
    }
}

//---------------------------------------------------------------------------
//
void UiPrefPassphrase::UpdateCb( GtkWidget*, UiPrefPassphrase* inst )
{
    gchar* new_pwd     = gtk_entry_get_text( inst->m_new );
    gchar* confirm_pwd = gtk_entry_get_text( inst->m_confirm );

    if( strlen( new_pwd ) < MIN_PASSPHRASE ){
		
		IZkPopup::MessageBox( IDS_PASSPHRASE_ERR, ZK_WARNING );

        return;
	}
	
    // verify the password was entered the same way in the verify field.
    if( strcmp( new_pwd, confirm_pwd ) )
    {
		IZkPopup::MessageBox( IDS_NEW_PASSPHRASE_CONFIRM_ERROR, ZK_ERROR );

        return;
	}

    bool success = true;

	ZKSTD::string oldPP;
    if( inst->UserHasPassphrase() )
    {
        oldPP = gtk_entry_get_text( inst->m_current );
    }
    else 
    {
        ZkService::GetICrypto()->GetDefaultPassphrase( oldPP );
    }


	try
    {
		ZkService::GetICrypto()->ChangePassphrase( oldPP, new_pwd ) ;
	} 
    catch(IZkException& ex)
    {
		ZkErrorHandler::GetIZkErrorReporter()->ShowError( ex );
        success = false;
	}

    if( success  == true )
    {
        inst->ShowPassphraseWidgets( true, true );

        gtk_entry_set_text( inst->m_current, "" );
        gtk_entry_set_text( inst->m_new, "" );
        gtk_entry_set_text( inst->m_confirm, "" );

		IZkPopup::MessageBox( IDS_PASSPHRASE_CHANGED );
    }
}


//---------------------------------------------------------------------------
//
bool UiPrefPassphrase::UserHasPassphrase()
{
	ZKSTD::string strCurPP, strDefPP;
	ZkService::GetICrypto()->GetPassphrase( strCurPP );
	ZkService::GetICrypto()->GetDefaultPassphrase( strDefPP );
	return ( strCurPP != strDefPP );
}

//---------------------------------------------------------------------------
//
void UiPrefPassphrase::ShowPassphraseWidgets( 
    bool show,
    bool hasPassphrase 
)
{
    if( show )
    {
        gtk_widget_show( m_frame );

        if( hasPassphrase )
        {
            gtk_widget_show( m_currentLb );
            gtk_widget_show( GTK_WIDGET( m_current ));
        }
        else
        {
            gtk_widget_hide( m_currentLb );
            gtk_widget_hide( GTK_WIDGET( m_current ));
        }
    }
    else 
    {
        gtk_widget_hide( m_frame );
    }
}
