// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

//===========================================================================
// EXTERNAL DECLARATIONS
//===========================================================================

#include <WzBase.h>
#include <WzBasePage.h>
#include <ZkFreedomWizardsStrings.h>

#include <ZKFileConfiguration.h>
#include <UiThreadId.h>
#include <lockguard.h>
#include <ZkGladeLoader.h>
#include <ZkCatalogRegistry.h>
#include <IZkPopup.h>
#include <ZkGtkUtils.h>

#include <clientdebug.h>

#include <glade/glade.h>
#include <gnome.h>

#include <algorithm>

//===========================================================================
// CLASS WzBase
//===========================================================================

//---------------------------------------------------------------------------
//
WzBase::WzBase( )
    : m_currentPage( 0 )
    , m_mode( START )
    , m_max_w( 0 )
    , m_max_h( 0 )
    , m_done( false )
    , m_freeVersion( false )
    , m_built( false )
{
}

//---------------------------------------------------------------------------
//
WzBase::~WzBase()
{
    if ( m_built )
    {
        gtk_widget_unref( m_start );                               
        gtk_widget_unref( m_finish ); 
        gtk_widget_unref( m_previous );
        gtk_widget_unref( m_next );
        gtk_widget_unref( m_help );
        gtk_widget_unref( m_cancel );
        m_built = false;
    }
    // FIXME: we're leaking all of the pages
}

//---------------------------------------------------------------------------
//
void
WzBase::BuildUI()
{
    if ( m_built )
    {
        return;
    }

    {
	ZkGladeLoader loader( "wizard.xml", "wizard" );

	m_window       = loader.GetWidget( "wizard" );
	m_btnbox_left  = loader.GetWidget( "btnbox_left" );
	m_btnbox_right = loader.GetWidget( "btnbox_right" );
	m_pageCt       = loader.GetWidget( "pageCt" );

	gtk_widget_show_all( loader.GetWidget( "vbox" ) );
    }

    // Then create all buttons and reference them so that we can remove
    // them from the button box without losing them.
    m_start    = gnome_stock_or_ordinary_button( ZkCatalogRegistry::GetString( IDS_START ));
    m_finish   = gnome_stock_or_ordinary_button( ZkCatalogRegistry::GetString( IDS_FINISH ));
    m_previous = gnome_stock_or_ordinary_button( GNOME_STOCK_BUTTON_PREV );
    m_next     = gnome_stock_or_ordinary_button( GNOME_STOCK_BUTTON_NEXT );
    m_help     = gnome_stock_or_ordinary_button( GNOME_STOCK_BUTTON_HELP );
    m_cancel   = gnome_stock_or_ordinary_button( GNOME_STOCK_BUTTON_CANCEL );



    gtk_widget_ref( m_start );
    gtk_widget_ref( m_finish );
    gtk_widget_ref( m_previous );
    gtk_widget_ref( m_next );
    gtk_widget_ref( m_cancel );
    gtk_widget_ref( m_help );

    gtk_widget_show( m_cancel );
//    gtk_widget_show( m_help );
    // FIXME - temporary fix for bug 6664: hide the help button.
    //         This only work because at the moment the Route wizard
    //         is the only user of this class. ericb
    gtk_widget_hide( m_help );


    // Hook functions.
    gtk_signal_connect (GTK_OBJECT( m_start ), "clicked",
			GTK_SIGNAL_FUNC( WzBase::StartCb ), this );
    gtk_signal_connect (GTK_OBJECT( m_finish ), "clicked",
			GTK_SIGNAL_FUNC( WzBase::FinishCb ), this );
    gtk_signal_connect (GTK_OBJECT( m_previous ), "clicked",
			GTK_SIGNAL_FUNC( WzBase::PreviousCb ), this );
    gtk_signal_connect (GTK_OBJECT( m_next ), "clicked",
			GTK_SIGNAL_FUNC( WzBase::NextCb ), this );
    gtk_signal_connect (GTK_OBJECT( m_help ), "clicked",
			GTK_SIGNAL_FUNC( WzBase::HelpCb ), this );
    gtk_signal_connect (GTK_OBJECT( m_cancel ), "clicked",
			GTK_SIGNAL_FUNC( WzBase::CancelCb ), this );

    gtk_signal_connect (GTK_OBJECT( m_window ), "delete_event",
			GTK_SIGNAL_FUNC( WzBase::CloseCb ), NULL );

    // Pack the boxes full of buttons
    gtk_box_pack_start( GTK_BOX( m_btnbox_left ), m_previous, FALSE,
			FALSE, 0 );
    gtk_box_pack_start( GTK_BOX( m_btnbox_left ), m_next, FALSE, FALSE, 0 );
    gtk_box_pack_start( GTK_BOX( m_btnbox_left ), m_start, FALSE, FALSE, 0 );
    gtk_box_pack_start( GTK_BOX( m_btnbox_left ), m_finish, FALSE, FALSE, 0 );
    gtk_box_pack_start( GTK_BOX( m_btnbox_right ), m_cancel, FALSE, FALSE, 0 );
    gtk_box_pack_start( GTK_BOX( m_btnbox_right ), m_help, FALSE, FALSE, 0 );

    // Initial mode is START
    gtk_widget_show( m_start );

    // This function should be reimplemented for each derived class; the
    // base wizard does nothing.
    DoAddPages();

    m_built = true;
}

//---------------------------------------------------------------------------
//
bool
WzBase::Run( )
{
    // First acquire the lock on the data members so we can safely wait
    // in a thread while the ui does its job.
    //
    // If we're going to block here until the UI thread is done, who else would
    // have access to these data members? -phil
    //
    //ZKS::Lock_Guard< WZkMutex > gate( m_lock );

    // Have the UiThreadCallback function called from the proper thread.
    UiThreadSafeMessage::NotifyUi( (gpointer)&WzBase::UiThreadCallback,
				   (gpointer)this, true );

    return m_done;
}

//---------------------------------------------------------------------------
//
void
WzBase::AddPage( ZKSTD::auto_ptr< WzBasePage > page )
{
    ZKASSERT( page.get() != NULL );

    WzBasePage* dlg = page.release();

    dlg->SetParent( this, m_pageCt );
    m_pages.push_back( dlg );

    // First page is shown, others get hidden.
    if( m_pages.size() == 1 ) 
    {
        dlg->Show();
    }
    else 
    {
        dlg->Hide();
    }

    GtkRequisition size_requisition;

    // In an attempt to get this thing to look good, calculate max page size
    // so that we dont jump from one size to the other when we click 
    // next/previous.
    gtk_widget_size_request( GTK_WIDGET( dlg->GetContainer() ), 
			     &size_requisition );

    m_max_w = MAX( m_max_w, size_requisition.width );
    m_max_h = MAX( m_max_h, size_requisition.height );

    // Now set the size hint for our window.
    gtk_widget_set_usize( GTK_WIDGET( m_pageCt ), m_max_w, m_max_h );

    UpdateMode();
}

//---------------------------------------------------------------------------
//
void
WzBase::RemovePage( ZKSTD::auto_ptr< WzBasePage > page )
{
    ZKASSERT( page.get() != NULL );

    WzBasePage* pageInst = page.release();

    pageInst->UnParent( );

    // Remove the page from the list.
    // I hate iterators.  With a blinding passion.
    guint i;
    for ( i = 0; i < m_pages.size() ; i++ ) {
	if ( m_pages[ i ] == pageInst ) {
	    m_pages.erase( m_pages.begin() + i );
	}
    }

    // We assume that we're not removing the current page.  Hope that's ok.
    UpdateMode();
}

//---------------------------------------------------------------------------
//
void
WzBase::SetActivePage( const ZKSTD::string& name )
{
    ZKASSERT( !name.empty() );

    // Find the page by name.
    PageList::size_type size = m_pages.size();
    PageList::size_type index = 0;

    for( ; index < size; index++ ) {
        if( m_pages[ index ]->GetName() == name ) {
            SetCurrentPage( index );
            break;
        }
    }

    ZKASSERT( index < size );
}

//---------------------------------------------------------------------------
//
void
WzBase::SkipPage( const ZKSTD::string& name, bool skip )
{
    ZKASSERT( !name.empty() );

    // Find the page by name.
    PageList::size_type size = m_pages.size();
    PageList::size_type index = 0;

    for( ; index < size; index++ ) {
        if( m_pages[ index ]->GetName() == name ) {
            m_pages[ index ]->SkipPage( skip );
            break;
        }
    }

    ZKASSERT( index < size );
}

//---------------------------------------------------------------------------
//
void
WzBase::Finish()
{
    bool processed;
    processed = m_pages[ m_currentPage ]->Finish();

    if( !processed )
    {
	m_done = true;
	gtk_main_quit();
    }
}

//---------------------------------------------------------------------------
//
void
WzBase::NextPage()
{
    bool processed;
    processed = m_pages[ m_currentPage ]->NextPage();

    if( !processed )
    {
        int size = static_cast< int >( m_pages.size() );
        int current = m_currentPage + 1;

        // go to the next page not marked as skipped.
        for( ; current < size && m_pages[ current ]->SkipIt(); current++ )
	    ;

        // Make sure at least the last page is not skippable.
        ZKASSERT( current < size );

        SetCurrentPage( current );
    }
}

//---------------------------------------------------------------------------
//
void
WzBase::PrevPage()
{
    bool processed;
    processed = m_pages[ m_currentPage ]->PreviousPage();

    if( !processed )
    {
        int current = m_currentPage-1;

        // go to the next page not marked as skipped.
        for( ; current >= 0 && m_pages[ current ]->SkipIt(); current-- );

        SetCurrentPage( current );
    }
}

//---------------------------------------------------------------------------
//
void WzBase::ButtonSetActive( int btn, bool state )
{
    if( btn & WZ_START )
    {
        gtk_widget_set_sensitive( m_start, state );    
    }

    if( btn & WZ_PREVIOUS )
    {
        gtk_widget_set_sensitive( m_previous, state );    
    }

    if( btn & WZ_NEXT )
    {
        gtk_widget_set_sensitive( m_next, state );    
    }

    if( btn & WZ_CANCEL )
    {
        gtk_widget_set_sensitive( m_cancel, state );    
    }

    if( btn & WZ_FINISH )
    {
        gtk_widget_set_sensitive( m_finish, state );    
    }


}

//---------------------------------------------------------------------------
// As the name implies, this is only ever run as the UI thread
//
void
WzBase::UiThreadCallback( gpointer klass )
{
    WzBase *imp = (WzBase *)klass;

    imp->BuildUI();

    ZkGtkUtils::GtkWidgetShow( imp->m_window );
    gtk_main();
    gtk_widget_hide( imp->m_window );
}

//---------------------------------------------------------------------------
//
gint WzBase::CloseCb( 
    GtkWidget */*widget*/,
    gpointer   /*data*/ )
{
    gint ret = TRUE;

    if( IZkPopup::QuestionBox( IDS_CLOSE_ARE_YOU_SURE, ZK_YESNO ) ) 
    {
        gtk_main_quit();
        ret = FALSE;
    }

    return ret;
}

//---------------------------------------------------------------------------
//
void
WzBase::StartCb( GtkWidget *, WzBase* inst )
{
    inst->NextPage();
}

//---------------------------------------------------------------------------
//
void
WzBase::FinishCb( GtkWidget *,  WzBase* inst )
{
    inst->Finish();
}

//---------------------------------------------------------------------------
//
void
WzBase::PreviousCb( GtkWidget *,  WzBase* inst )
{
    inst->PrevPage();
}

//---------------------------------------------------------------------------
//
void
WzBase::NextCb( GtkWidget *,  WzBase* inst  )
{
    inst->NextPage();
}

//---------------------------------------------------------------------------
//
void
WzBase::CancelCb( GtkWidget *,  WzBase* /*inst*/  )
{
    if( IZkPopup::QuestionBox( IDS_CLOSE_ARE_YOU_SURE, ZK_YESNO ) ) 
    {
        // Get out of here and return false.
        gtk_main_quit();
    
    }
}

//---------------------------------------------------------------------------
//
void
WzBase::HelpCb( GtkWidget *,  WzBase* /*inst*/  )
{
    // FIXME - This work only because the Route Selection wizard is the only
    //         one using this base class at the moment.
    ZKSTD::string helpFile;
    helpFile = ZKFileConfiguration::GetFreedomRouteWizardHelp();

    gnome_url_show( helpFile.c_str() );
    
}

//---------------------------------------------------------------------------
//
void
WzBase::SetCurrentPage( int index )
{
    ZKASSERT( (index >= 0) && (index < static_cast< int >( m_pages.size() )) );

    if( index == m_currentPage ) {
        return;
    }

    m_pages[ m_currentPage ]->Hide();
    m_currentPage = index;
    m_pages[ m_currentPage ]->Show();

    UpdateMode();
}

//---------------------------------------------------------------------------
//
void
WzBase::StartMode()
{
    if( m_mode == START )
    {
        return;
    }
    
    if( m_mode == NORMAL )
    {
	gtk_widget_hide( m_previous );
	gtk_widget_hide( m_next );
    }
    else if( m_mode == FINISH )
    {
	gtk_widget_hide( m_previous );
	gtk_widget_hide( m_finish );
    }

    gtk_widget_show( m_start );
    m_mode = START;
}

//---------------------------------------------------------------------------
//
void
WzBase::NormalMode()
{
    if( m_mode == NORMAL )
    {
        return;
    }
    
    if( m_mode == START )
    {
	gtk_widget_hide( m_start );
    }
    else if( m_mode == FINISH )
    {
	gtk_widget_hide( m_previous );
	gtk_widget_hide( m_finish );
    }

    gtk_widget_show( m_previous );
    gtk_widget_show( m_next );
    m_mode = NORMAL;
}

//---------------------------------------------------------------------------
//
void
WzBase::FinishMode()
{
    if( m_mode == FINISH )
    {
        return;
    }
    
    if( m_mode == START )
    {
	gtk_widget_hide( m_start );
    }
    else if( m_mode == NORMAL )
    {
	gtk_widget_hide( m_next );
	gtk_widget_hide( m_previous );
    }

    // Previous doesn't make sense if it's the first page
    if ( m_currentPage != 0 ) {
	gtk_widget_show( m_previous );
    }

    gtk_widget_show( m_finish );
    m_mode = FINISH;
}

//---------------------------------------------------------------------------
//
void
WzBase::UpdateMode()
{
    int current = m_currentPage + 1;
    int size = m_pages.size();

    if ( current == size ) {
	FinishMode();
    } else if ( current == 1 ) {
	StartMode();
    } else {
	NormalMode();
    }
}

//---------------------------------------------------------------------------
//
void WzBase::DoAddPages()
{
    // This -must- be implemented by derived classes
    ZKASSERT("Someone forgot to implement DoAddPages" && 0);
}
