/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef ZKS_INCL_WZBASE
#define ZKS_INCL_WZBASE

/*===========================================================================
 * EXTERNAL DECLARATIONS
 *=========================================================================*/

#include <UiThreadSafeMessage.h>

#include <clientstl.h>

#include <gtk/gtk.h>
#include <vector>
#include <memory>
#include <string>

class WzBasePage;

/*===========================================================================
 * TYPE WzBase
 *=========================================================================*/

typedef enum {
    WZ_START    = 1 << 0,
    WZ_PREVIOUS = 1 << 1,
    WZ_NEXT     = 1 << 2,
    WZ_CANCEL   = 1 << 3,
    WZ_FINISH   = 1 << 4
} WzBaseBtnId;

/*===========================================================================
 * CLASS WzBase
 *=========================================================================*/

class WzBase {

public: 

    /*----- member functions -----*/

    WzBase( );
    virtual ~WzBase( );

    void BuildUI();

    bool Run();

    // Purpose:  Add a page to the wizard.  It automatically adjusts Next and
    //           Finish buttons
    void AddPage( ZKSTD::auto_ptr<WzBasePage> page );

    // Purpose:  Remove a page from the wizard.  It automatically adjusts Next
    //           and Finish buttons
    void RemovePage( ZKSTD::auto_ptr<WzBasePage> page );

    // Purpose:  Sets the currently displayed page.
    // Throws:   FIXME - throw if not in the wizard.
    void SetActivePage( const ZKSTD::string& name );


    // Purpose:  Sets the currently displayed page.
    //           FIXME - should be renamed SetActivePage or the other should be renamed....
    void SetCurrentPage( int dlg );

    // Purpose:  This function can be used ti=o indicate to the wizard that 
    //           some pages should be skipped.
    void SkipPage( const ZKSTD::string& name, bool skip );

    virtual void Finish();
    void NextPage();
    void PrevPage();

    // Purpose: Enables or disables the wizard buttons en masse
    void ButtonSetActive( int btn, bool state ); 

    // Purpose:  Indicates whether or not we are using the free or "pro"
    //           version. Influences the content of some pages.
    bool IsFreeVersion() const { return m_freeVersion; }

    // Purpose:  Well... sets the flag indicating whther or not we are
    //           doing Setup or Creating a nym for th efree version.
    void SetFreeFlag( bool flag ) { m_freeVersion = flag; }

    // Purpose: These two function set/get the token, to communicate 
    //          between pages.
    tstring GetToken( ) const { return m_token; }
    void SetToken( const tstring& in_token ) { m_token = in_token; }

protected:

    /*----- member functions -----*/

    static void UiThreadCallback( gpointer klass );

    /*----- types and enumerations ----*/

    typedef ZKSTD::vector< WzBasePage* > PageList;

    /*----- data members -----*/

    PageList   m_pages;
    int        m_currentPage;

private: 

    /*----- types and enumerations ----*/

    typedef enum  {
        START,
        NORMAL,
        FINISH
    } Mode;

    /*----- static member functions -----*/

    static gint CloseCb( GtkWidget *, gpointer   );

    static void StartCb( GtkWidget *,  WzBase* inst  );
    static void FinishCb( GtkWidget *,  WzBase* inst  );
    static void PreviousCb( GtkWidget *,  WzBase* inst  );
    static void NextCb( GtkWidget *,  WzBase* inst  );
    static void CancelCb( GtkWidget *,  WzBase* inst  );
    static void HelpCb( GtkWidget *,  WzBase* inst  );

    /*----- member functions -----*/

    void StartMode();
    void NormalMode();
    void FinishMode();
    void UpdateMode();

    // Purpose:  Facilitate the necessary adding of pages by derived classes
    //           Note that the base implementation just aborts.
    virtual void DoAddPages();

    // prohibited funcs
    WzBase( const WzBase& );
    const WzBase& operator=( const WzBase& );

    /*----- data members -----*/

    GtkWidget* m_window;
    GtkWidget* m_pageCt;     // Main container for pages.
    GtkWidget* m_vbox;
    GtkWidget* m_btnbox_left;
    GtkWidget* m_btnbox_right;
    GtkWidget* m_start;
    GtkWidget* m_finish;
    GtkWidget* m_previous;
    GtkWidget* m_next;
    GtkWidget* m_help;
    GtkWidget* m_cancel;

    Mode m_mode;

    int        m_max_w;
    int        m_max_h;

    bool       m_done;
    bool       m_freeVersion;
    bool       m_nymWizard;
    bool       m_built;
    tstring    m_token;
};
#endif
