/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef ZKS_INCL_WZBASEPAGE
#define ZKS_INCL_WZBASEPAGE

/*===========================================================================
 * EXTERNAL DECLARATIONS
 *=========================================================================*/

#include <clientstl.h>

#include <gtk/gtk.h>
#include <string>


class WzBase;

/*===========================================================================
 * CLASS WzBasePage
 *=========================================================================*/

/*
  As the name implies, this is a base class to create wizard pages.
  WzBase supports the API to plug into the wizard dialog box.

  Note: to create the glade xml UI description for a Wizard page, start from
  the wizard_template.xml file and put your ui in the Wizard_subCt widget.
*/

class WzBasePage {

public: 

    /*----- member functions -----*/

    // Purpose:  Initializes the WzBasePage class.
    //           If 'defaultUI' is true, the common default page setup 
    //           will be built. I is up to the derived class to parent its
    //           UI in this default setup.
    WzBasePage( const ZKSTD::string& name );
    virtual ~WzBasePage();
      
    // Purpose:  Adds the page container to the children of the given widget.
    //           Called by WzBase class, no need to do it by hand.
    // Requires: The given widget obvioulsy has to be a container of some sort.
    virtual void SetParent( WzBase*, GtkWidget* );

    // Purpose:  Removes the page container from the given widget.
    //           Called by WzBase class, no need to do it by hand.
    void UnParent( );

    // Purpose:  Parents the given widget (usually the page main container)
    //           into the base page sub container (m_subCt).
    void AddPageUi( GtkWidget* ct );

    // Purpose:  Used by the parent Wizard dialog to show only one 
    //           page at time.
    virtual void Hide();
    virtual void Show();

    virtual bool NextPage() { return false; }
    virtual bool PreviousPage() { return false; }
    virtual bool Finish() { return false; }

    // Purpose:  Return a pointer to the main dialog container.
    // Promise:  Will not be null, since creating the dialog creates the 
    //           container.
    GtkWidget*   GetContainer() const;

    // Purpose:  Returns the widget that should have the focus by default.
    //           (can be null, if not set)
    GtkWidget*   GetDefaultWidget() const;

    // Purpose:  Sets the flags indicating whether or not the page should be
    //           skipped when the wizard gets to it.
    // Requires: The page shoul dnot be the active page when the skip flag
    //           is changed.
    void SkipPage( bool );

    // Purpose:  Returns a bool stating whether or not the page should be
    // skiped.
    bool SkipIt() const;

    WzBase* GetParent() const;

    // Purpose:  Return the name...
    // Promise:  String is not null, enforced by asserts in the constructor.
    const ZKSTD::string& GetName() const;

    // Purpose: Hide the pixmap on the left side of the page
    void HidePixmap() const { gtk_widget_hide(m_pixmap); }

    // Purpose: Show the pixmap on the left side of the page
    void ShowPixmap() const { gtk_widget_show(m_pixmap); }

protected: 

    /*----- static member functions -----*/

    // Purpose:  This function is used by derived classes to connect to a
    //           signal which should trigger a NextPage() call.
    //           (i.e. if "enter" in a text field gets you to the next page)
    static void NextPageCb( GtkWidget*, WzBasePage* );
    
    /*----- member functions -----*/

    // Purpose:  Sets the widgt that should have th efocus by default when the
    //           page is shown.
    void SetDefaultWidget( GtkWidget* );

private: 

    /*----- member functions -----*/

    // Prohibited funcs - not implemented.
    WzBasePage( const WzBasePage& );
    const WzBasePage& operator=( const WzBasePage& );

    /*----- static data members -----*/

    /*----- data members -----*/

    WzBase         *m_parent; // the Wizard which contains this page

    GtkWidget*      m_pixmap; // the pixmap on the left

    bool            m_skip;   // whether or not this page should be skipped.
    GtkWidget*      m_topCt;  // top container
    GtkWidget*      m_subCt;  // sub container for derived classes contents.
    ZKSTD::string   m_name;

    // This widget pointer can be used to indicate to the wizard main dialog
    // that  widget should be given the focus when the page is shown.
    GtkWidget*      m_default;
};


//---------------------------------------------------------------------------
//
inline GtkWidget* WzBasePage::GetContainer() const
{
    return m_topCt;
}

//---------------------------------------------------------------------------
//
inline GtkWidget* WzBasePage::GetDefaultWidget() const
{
    return m_default;
}

//---------------------------------------------------------------------------
//
inline void WzBasePage::SetDefaultWidget( GtkWidget* w ) 
{
    m_default = w;
}


//---------------------------------------------------------------------------
//
inline const ZKSTD::string& WzBasePage::GetName() const
{
    return m_name;
}

//---------------------------------------------------------------------------
//
inline WzBase * WzBasePage::GetParent() const
{
    return m_parent;
}

//---------------------------------------------------------------------------
//
inline void WzBasePage::SkipPage( bool skipit )
{
    m_skip = skipit;
}

//---------------------------------------------------------------------------
//
inline bool WzBasePage::SkipIt() const
{
    return m_skip;
}
#endif
