// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

//===========================================================================
// EXTERNAL DECLARATIONS
//===========================================================================

#include "WzRoute.h"
#include "WzRoutePageFirst.h"
#include "WzRoutePage.h"

#include <ZkGladeLoader.h>
#include <ZkGtkUtils.h>
#include <ZKFileConfiguration.h>

#include <ZkService.h>
#include <INym.h>
#include <INymRoute.h>
#include <WNymIconInfo.h>

#include <IRouteManager.h>
#include <WRouteRules.h>

#include <guids.h>
#include <INymCentral.h>

#include <gdk-pixbuf/gdk-pixbuf.h>

//===========================================================================
// CLASS WzRoute
//===========================================================================

//---------------------------------------------------------------------------
//
WzRoute::WzRoute( WRouteRules& io_rules, WRoute& out_routes )
    : m_RouteRules( io_rules )
    , m_Route( out_routes )
{
    // Determine if we have a previous route for this nym
    INymRoute * pNymRoute = NULL;
    ZkService::GetINymCentral()->GetActiveNym( IID_INYMROUTE,
					       (void**)&pNymRoute);

    m_ratings[0] = NULL;
    pNymRoute->GetLastSuccessfulRoute( m_LastUsedRoute );
}

//---------------------------------------------------------------------------
//
WzRoute::~WzRoute()
{
    // Free ratings pixmaps
    if (m_ratings[0]) {
	gint i;

	for (i = 0; i < 6; i++) {
	    gdk_pixmap_unref( m_ratings[i] );
	}
    }
}

//---------------------------------------------------------------------------
//
void
WzRoute::RemoveHopPagesFromWizard()
{
    for ( size_t i = m_pages.size() - 1 ; i > 0 ; i-- ) {
	RemovePage( ZKSTD::auto_ptr<WzBasePage>(m_pages[i]) );
    }

    m_Route.Clear() ;
}

//---------------------------------------------------------------------------
//
void
WzRoute::AddHopPagesToWizard()
{
    if ( m_pages.size() > 1 )
    {
	// I guess user clicked on same radio button twice
	// so we don't have anything to do
	return ;
    }

    // Add one page in for each hop that the user needs to select
    for ( guint i = 0; i < m_RouteRules.GetHopCount(); i++ )
    {
	AddPage( ZKSTD::auto_ptr<WzBasePage>(new WzRoutePage(i)) );
    }

    // Load up the ratings pixmaps, if we haven't already
    if ( !m_ratings[0] )
    {
	ZKSTD::string filename = ZKFileConfiguration::GetFreedomPixmapDir();
	filename += "ratings.xpm";

	GdkPixbuf *pb = gdk_pixbuf_new_from_file( filename.c_str( ) );

	if (!pb)
	{
	    ZKTRACE2( "gdk_pixbuf_new_from_file(%s) failed: %d\n",
		      filename.c_str(), errno );
	    return;
	}

	gint width = gdk_pixbuf_get_width( pb );
	gint height = gdk_pixbuf_get_height( pb );
	gint single = width / 6;

	// The ratings image contains all 6 items
	gint i;
	for (i = 0; i < 6; i++)
	{
	    GdkGC *gc;

	    m_ratings[i] = gdk_pixmap_new( NULL, single, height,
					   gdk_rgb_get_visual()->depth );
	    gc = gdk_gc_new( m_ratings[i] );
	    gdk_pixbuf_render_to_drawable( pb, m_ratings[i], gc,
					   single * i, 0, 0, 0,
					   single, height,
					   GDK_RGB_DITHER_NORMAL, 0, 0 );
	    gdk_gc_unref( gc );
	}

	gdk_pixbuf_unref( pb );
    }
}

//---------------------------------------------------------------------------
//
void
WzRoute::Finish( )
{
    // We duplicate some of the base class functionality here because ordering
    // is particularly important for the Route wizard.

    // First, find out if the user has entered all of the data.  Finish() will
    // return true if it needs more data, and false if it does not.
    bool processed;
    processed = m_pages[ m_currentPage ]->Finish();

    if ( processed )
    {
        // It is Page::Finish's responsibility to notify the user
        return;
    }

    // By this point, the page is known to be filled in, which means that our
    // route should be valid.

    if ( static_cast<WzRoutePageFirst *>(m_pages[0])->IsInteractive() )
    {
        // At this point the route has already been selected and is set
        // in m_Route. Since we kept a reference to the route that was
        // passed in, there is no need to return or copy anything.
    }
    else if ( static_cast<WzRoutePageFirst *>(m_pages[0])->IsAutomatic() )
    {
        ZKASSERT( m_pages.size() == 1 );

        // Make sure we don't return any route at all
        // so that Freedom will find one for us.
        m_Route.Clear() ;

        if ( static_cast<WzRoutePageFirst *>( m_pages[0])->IsDontShow() )
        {
            // This will prevent the dialog from appearing again
            INym * pNym = NULL ;
            ZkService::GetINymCentral()->GetActiveNym( IID_INYM, (void**) &pNym ) ;

            pNym->SetNymPreferencesBitfield( 0, PREF_INTERACTIVE_ROUTES ) ;
        }
    }
    else if ( static_cast<WzRoutePageFirst *>(m_pages[0])->IsLastUsed() )
    {
        ZKASSERT( m_pages.size() == 1 );

        // Assign the last used route to be the 
        // route to try again
        m_Route = m_LastUsedRoute;
    }
    else
    {
        ZKASSERT( false ) ;
    }

    // Finally, allow the base class to do its thing.  It will trigger an
    // extra Page::Finish() call, but that is OK.
    WzBase::Finish();
}

//---------------------------------------------------------------------------
//
void
WzRoute::DoAddPages( )
{
    AddPage( ZKSTD::auto_ptr< WzBasePage >(new WzRoutePageFirst()) );
}
