// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

// Freedom includes
#include <IZkRegistry.h>
#include <../linux/RZkRegistry.h>
#include <../linux/RZkRegistryHolder.h>
#include <ZKConfiguration.h>
#include <ZKFileConfiguration.h>
#include <client.h>

// Linux includes
#include <getopt.h>
#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>

// C++
#include <iostream>

#define VERSION_STRING "zkreged v0.1"

#define FOO_BUF_SZ 4096

#define NONE 0
#define STRING 1
#define BYTES 2 
#define SHORTS 3
#define INTS 4


static RZkRegistryHolder *TheRegistry;

int hexFormat = 0;

void Usage( void )
{
    ZKSTD::cout 
        << "Freedom Client Registry Editor" << ZKSTD::endl
        << "(C) 2000 Zero Knowledge Systems inc." << ZKSTD::endl
        << ZKSTD::endl
        << "Usage: zkreged [OPTION] [NAME] [VALUE]" << ZKSTD::endl
        << "-h, --help:\t\tThis message" << ZKSTD::endl
        << "-v, --version:\t\tPrint version info." << ZKSTD::endl
        << "-a, --add=TYPE:\t\tAdd a entry NAME of TYPE with VALUE" << ZKSTD::endl
        << "\t\t\tTYPE must be one of (s)tring,(b)ytes,s(h)orts,(i)nts" << ZKSTD::endl
        << "-l, --list[=TYPE]:\tList registry entries." << ZKSTD::endl 
        << "\t\t\tIf TYPE is specified numerical conversions will " << ZKSTD::endl
        << "\t\t\tbe performed using that type." << ZKSTD::endl
        << "-d, --delete:\t\tDelete entry NAME." << ZKSTD::endl
        << "-x, --hex:\t\tPerform numerical conversions in haxadecimal" << ZKSTD::endl
        ;
}

void Version( void )
{
    ZKSTD::cout 
        << "Freedom Client Registry Editor" << ZKSTD::endl
        << "(C) 2000 Zero Knowledge Systems inc." << ZKSTD::endl
        << VERSION_STRING << ZKSTD::endl
        ;
}

void RegInit( void )
{
    ZKSTD::string registryPath = ZKFileConfiguration::GetFreedomUserDataDir();

    // Force the directory to exist, so that the registry can go in it.
    // Should this go elsewhere?
    // directory might well already exist, so don't report that error
    if (mkdir(registryPath.c_str(), 0700) < 0 && errno != EEXIST)
      perror("Creating ~/.freedom directory failed");

    registryPath += "freedom.reg";

    TheRegistry = new RZkRegistryHolder(registryPath.c_str());
}

void RegFini( void )
{
   delete TheRegistry;
}

void DumpInts(unsigned int *arr, int len)
{
    for (int i = 0; i < len; i++) {    
        if (hexFormat)
            printf("%c%x", (i == 0) ? ' ' : ',', arr[i]);
        else
            printf("%c%u", (i == 0) ? ' ' : ',', arr[i]);
    }
    printf("\n");
}

void ScanInts(char *src, unsigned int *dst, int *len)
{
    int i;
    char *start = src;
    char dpos = 0;
    int end = 0;
    for (i=0;; i++) {
        if (src[i] == ',' || src[i] == 0) {
            if (src[i] == 0) end = 1;
            src[i] = 0;
            if (hexFormat)
                sscanf(start, "%x", &dst[dpos++]);
            else
                sscanf(start, "%u", &dst[dpos++]);
            if (end) {
                *len = dpos;
                return;
            }
            i++;
            start = &src[i];
        } else {
            if ((( hexFormat && !isxdigit(src[i]) ) || (!hexFormat &&!isdigit(src[i]))) && !isspace(src[i])) 
            {
                puts("unable to parse VALUE string");
                exit( -1 );
            }
        }

    }
}

void DumpShorts(unsigned short *arr, int len)
{
    for (int i = 0; i < len; i++) {    
        if (hexFormat)
            printf("%c%hx", (i == 0) ? ' ' : ',', arr[i]);
        else
            printf("%c%hu", (i == 0) ? ' ' : ',', arr[i]);
    }
    printf("\n");
}

void ScanShorts(char *src, unsigned short int *dst, int *len)
{
    int i;
    char *start = src;
    char dpos = 0;
    int end = 0;
    for (i=0;; i++) {
        if (src[i] == ',' || src[i] == 0) {
            if (src[i] == 0) end = 1;
            src[i] = 0;
            if (hexFormat)
                sscanf(start, "%hx", &dst[dpos++]);
            else
                sscanf(start, "%hu", &dst[dpos++]);
            if (end) {
                *len = dpos;
                return;
            }
            i++;
            start = &src[i];
        } else {
            if ((( hexFormat && !isxdigit(src[i]) ) || (!hexFormat &&!isdigit(src[i]))) && !isspace(src[i])) {
                puts("unable to parse VALUE string");
                exit (-1);
            }
        }
    }
}

void DumpBytes(unsigned char *arr, int len)
{
    for (int i = 0; i < len; i++) {
        unsigned short int tmp = arr[i];
        if (hexFormat) 
            printf("%c%hx", (i == 0) ? ' ' : ',', tmp);
        else 
            printf("%c%hu", (i == 0) ? ' ' : ',', tmp);
    }
    printf("\n");
}

void ScanBytes(char *src, unsigned char *dst, int *len)
{
    int i;
    char *start = src;
    char dpos = 0;
    int end = 0;
    unsigned short int tmp;
    for (i=0;; i++) {
        if (src[i] == ',' || src[i] == 0) {
            if (src[i] == 0) end = 1;
            src[i] = 0;
            if (hexFormat)
                sscanf(start, "%hx", &tmp);
            else
                sscanf(start, "%hu", &tmp);
            dst[dpos++] = (unsigned char)tmp;
            if (end) {
                *len = dpos;
                return;
            }
            i++;
            start = &src[i];
        } else {
            if ((( hexFormat && !isxdigit(src[i]) ) || (!hexFormat && !isdigit(src[i]))) && !isspace(src[i])) {
                puts("unable to parse VALUE string");
                exit (-1);
            }
        }
    }
}


void List( int type )
{
    REGENUM state = 0;
    REGINFO info;
    REGERR err = REGERR_OK;
    char regBuf[FOO_BUF_SZ];
    RegInit();
    RZkRegistry &reg = (RZkRegistry&)IZkRegistry::get();

    info.size = sizeof(REGINFO);
    while ((err = NR_RegEnumEntries(reg.GetRegistry(), reg.GetFreedomKey(), 
                                   &state, regBuf, FOO_BUF_SZ, &info)) == REGERR_OK)
    {
        REGERR err2 = REGERR_OK;
        int len = FOO_BUF_SZ;
        printf("%s\n", regBuf);

        if (info.entryType == REGTYPE_ENTRY_BYTES) {            
            unsigned char bytes[FOO_BUF_SZ]; 
            len = FOO_BUF_SZ;
            err2 = NR_RegGetEntry(reg.GetRegistry(), reg.GetFreedomKey(), regBuf, (void *)bytes, (uint32_t *)&len);
            if (err2 == REGERR_OK) {
                printf("\tValue: ");
                if ((type == NONE && len % sizeof(int) == 0) || type == INTS)
                    DumpInts((unsigned int *)bytes, len / sizeof(int));
                else if ((type == NONE && len % sizeof(short int) == 0) || type == SHORTS)
                    DumpShorts((unsigned short *)bytes, len / sizeof(short int));
                else if (type == NONE || type == BYTES)
                    DumpBytes(bytes, len);
                else
                    printf("<no string representation>\n");
            }
        } else if (info.entryType == REGTYPE_ENTRY_STRING_UTF) {
            ZKSTD::string val;
            IZkRegistry::GetString(regBuf, val);
            printf("\tValue: %s\n", val.c_str());
            len = info.entryLength;
        } else {
            puts("\tValue: <unknown>");
        }

        printf("\tsize: %u\ttype: %u\tlength: %d\n", info.size, info.entryType, len);
        fflush(stdout);
    }

    RegFini();
}

void Delete( char *name )
{
    RegInit();
    IZkRegistry::DeleteValue(name);
    printf("%s deleted.\n", name);    
    RegFini();
}

void Add(int type, char *name, char *value)
{
    REGERR err = REGERR_OK;
    RegInit();
    if (type == STRING) {
        IZkRegistry::SetString(name, value);        
    } else {
        unsigned char buf[FOO_BUF_SZ]; int len = FOO_BUF_SZ;        
        switch (type) 
        {
            case BYTES:
                ScanBytes(value, buf, &len);
                break;
            case SHORTS:
                len /= sizeof(short int);
                ScanShorts(value, (unsigned short*)buf, &len);
                len *= sizeof(short int);
                break;
            case INTS:
                len /= sizeof(int);
                ScanInts(value, (unsigned int*)buf, &len);
                len *= sizeof(int);
                break;
            default:
                // Whoa! We be all f'd up now.
                RegFini();
                exit(-1);
        }
        RZkRegistry &reg = (RZkRegistry&)IZkRegistry::get();
        err = NR_RegSetEntry(reg.GetRegistry(), reg.GetFreedomKey(), name, REGTYPE_ENTRY_BYTES, (void *)buf, len);
    }
    if (err == REGERR_OK) printf("%s added to registry\n", name);
    RegFini();
}

static struct option long_options[] =
{
    {"help", 0, 0, 'h'},
    {"version", 0, 0, 'v'},
    {"list", 2, 0, 'l'},
    {"delete", 0, 0, 'd'},
    {"add", 1, 0, 'a'},
    {"hex", 0, 0, 'x'},
    {0, 0, 0, 0}
};

int main (int argc, char **argv)
{
    int f_list = 0, f_add = 0, f_delete = 0;
    char *type = NULL;
    opterr = 0;
    int parsedType = NONE;
    while (1)
    {
        char c;
        int option_index;
        
        c = getopt_long (argc, argv, "hvl::dxa:",
                         long_options, &option_index);
        if (c == -1)
            break;
        
        switch (c)
        {
            case 'h':
                Usage();
                exit(0);
                break;
            case 'v':
                Version();
                exit(0);
                break;
            case 'l':
                f_list = 1;
                if (optarg) type = optarg;
                break;
            case 'd':
                f_delete = 1;
                break;
            case 'x':
                hexFormat = 1;
                break;
            case 'a':
                f_add = 1;
                if (optarg) type = optarg;
                else {
                    puts("option -a requires a TYPE argument");
                    exit(-1);
                }
                break;
        }
    }

    if ( 1 != f_list + f_delete + f_add ) {
        if (0 == f_list + f_delete + f_add ) {
            Usage();
        } else {
            puts("must specify one and only one of --add --list --delete (-a -l -d)");
        }
        exit (-1);
    }

    if (type) {
        switch (type[0]) {
            case 's' :
                parsedType = STRING;
                break;
            case 'b' :
                parsedType = BYTES;
                break;
            case 'h' :
                parsedType = SHORTS;
                break;
            case 'w' :
            case 'i' :
                parsedType = INTS;
                break;
            default :
                puts("TYPE must be one of (s)tring,(b)ytes,s(h)orts,(i)nts");
                exit(-1);
        }
    }

    if (f_list) {
        List(parsedType);
        exit(0);
    }

    if (optind >= argc) {
        puts("option requires NAME argument");
        exit(-1);
    }
    
    if (f_delete) {
        if (!argv[optind]) {
            puts("option -d requires NAME argument");
            exit(-1);
        }
        Delete(argv[optind]);
        exit(0);
    }

    if (f_add) {
        if (parsedType == NONE) {
            puts("option -a requires TYPE parameter");
            exit(-1);
        }
        if (optind < argc - 1 && argv[optind] && argv[optind+1]) {
            Add(parsedType, argv[optind], argv[optind+1]);
            exit(0);
        }
        puts("option -a requires both NAME and VALUE arguments");
        exit(-1);
    }
}
