/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


/*

  makecat  - x-platform string catalog generator


  This program is a short term solution to the problem of string catalogs
  common to all platforms.

  It takes as input a file containing string IDs and string definitions, and
  creates as output a header file containing the string IDs defines, along
  with a file format specific to the target platform.

  If you need to add new strings to the client, DO NOT edit the platform 
  specific files directly. Instead go edit the .cat file and use makecat.

  The format is pretty simplistic:

  - empty lines are ignored.
  - lines with # as a first character are comment.
  - the first non-comment line must be the catalog index.
  - the second non-comment line is the basename of the output files.
    (usually the same as the catalog filename)


  Example:

  Let's say you give makecat the following catalog file as input:

--- Strings.cat

  # Some example catalog.

  20
  Strings

  STRING_ID_1 "Text content for string 1"
  STRING_ID_2 "Text content for string 2"
  STRING_ID_3 "Text content for string 3"
  STRING_ID_4 "Text content for string 4"

---  
  
  Then you type the following command:

       makecat Strings.cat

  This will create the files for all platforms. The files created will 
  look as follows:



--- Strings.h

//===========================================================================
// Catalog indices for strings.
// DO NOT MODIFY. Use makecat to generate (see ericb)
//===========================================================================

#define CATALOG_INDEX 0x00140000

#define STRING_ID_1                                        CATALOG_INDEX+0
#define STRING_ID_2                                        CATALOG_INDEX+1
#define STRING_ID_3                                        CATALOG_INDEX+2
#define STRING_ID_4                                        CATALOG_INDEX+3




--- Strings.rc ( for Windows )

//===========================================================================
// Resource file for Windows.
// DO NOT MODIFY. Use makecat to generate (see ericb)
//===========================================================================


STRINGTABLE DISCARDABLE 
BEGIN
    STRING_ID_1 "Text content for string 1"
    STRING_ID_2 "Text content for string 2"
    STRING_ID_3 "Text content for string 3"
    STRING_ID_4 "Text content for string 4"
END


--- Strings.cpp ( for Linux )

//===========================================================================
// String Catalog for Linux.
// DO NOT MODIFY. Use makecat to generate (see ericb)
//===========================================================================

#include "Strings.h"


const char* Strings[] = {
   "Text content for string 1",
   "Text content for string 2",
   "Text content for string 3",
   "Text content for string 4",
}






*/



#include <iostream>
#include <string>
#include <fstream>
#include <strstream>
#include <cstring>
#include <cstdlib>
#include <ctype.h>

#include <set>
#include <iomanip>

using namespace std;


//===========================================================================
// File Headers
//===========================================================================

#define MOZILLA_LICENSE \
"/*\n"\
" * The contents of this file are subject to the Mozilla Public\n"\
" * License Version 1.1 (the \"License\"); you may not use this file\n"\
" * except in compliance with the License. You may obtain a copy of\n"\
" * the License at http://www.mozilla.org/MPL/\n"\
" * \n"\
" * Software distributed under the License is distributed on an \"AS\n"\
" * IS\" basis, WITHOUT WARRANTY OF ANY KIND, either express or\n"\
" * implied. See the License for the specific language governing\n"\
" * rights and limitations under the License.\n"\
" * \n"\
" * The Original Code is Freedom code.\n"\
" * \n"\
" * The Initial Developer of the Original Code is Zero-Knowledge Systems,\n"\
" * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000\n"\
" * Zero-Knowledge Systems, Inc.  All Rights Reserved.\n"\
" * \n"\
" * Contributor(s):\n"\
"\n"\
" */\n"\
"\n\n"


// include file
#define INCLUDE_HEADER \
"\n\n"\
"//===========================================================================\n"\
"// Catalog indices for strings.\n"\
"// DO NOT MODIFY. Use makecat to generate (see ericb)\n"\
"//===========================================================================\n"\
"\n"\
"#define "

#define SOURCE_HEADER \
"\n\n"\
"//===========================================================================\n"\
"// String Catalog for Linux.\n"\
"// DO NOT MODIFY. Use makecat to generate (see ericb)\n"\
"//===========================================================================\n"\
"\n\n"

#define RESOURCE_HEADER \
"\n\n"\
"//===========================================================================\n"\
"// Resource file for Windows.\n"\
"// DO NOT MODIFY. Use makecat to generate (see ericb)\n"\
"//===========================================================================\n"\
"\n\n"


bool DiscardLine( char* inputLine )
{
    int i = 0;
    int len = strlen( inputLine );

    while( (i < len) && ((inputLine[i] == ' ') || (inputLine[i] == '\t')) )
    {
        ++i;
    }


    // If we have reached the end of the line, or if it is a # character, discard.
    return ( (i == len) ||
             (inputLine[i] == '#')  || 
             (inputLine[i] == '\n') || 
             (inputLine[i] == '\r') );
}


int main( int argc, char* argv[] )
{
    ofstream outCatalog;
    ofstream outResource;

    // Create a set to make sure we only have one instance of every 
    // message ID in the header.
    set< string > setOfIds;


    if( argc < 2 || argc > 3 ) {
        cout << endl << "Usage : ";
        cout << "  makecat catalog_file" << endl << endl;

        exit( -1 );
    }

    // Validate input file
    ifstream inFile( argv[ 1 ] );
    if( !inFile ) {
        cout << "Error: cannot open file <" << argv[ 1 ] << ">" << endl;
        exit( -1);
    }

    // Well, I am not pretending this is high quality code...
    char inputLine[ 5000 ];

    // Get the catalog index and output file name from the cat file.
    
    // Skip the crap.
    while( inFile.getline( inputLine, 5000, '\n' ) &&
           DiscardLine( inputLine ) ); 

    unsigned long catIndex = atol( inputLine ) << 10;
    cout << "Catalog Index: " << inputLine << ":" << catIndex << endl;

    // Now get the output file basename.
    // Skip the crap.
    while( inFile.getline( inputLine, 5000, '\n' ) &&
           DiscardLine( inputLine ) ); 

    string outFileName = inputLine;
    if( outFileName.empty() )
    {
        cout << "Invalid file format, no output file name." << endl << endl;
        exit( -1 );
    }

    // Catalog name is the same as the output file name, but capitalized.
    char* catName = strdup( inputLine );
    for( int i = strlen( catName )-1; i >= 0; i-- ) 
    {
        catName[i] = ::toupper( catName[i] );
    }
    cout << "Catalog Name: " << catName << endl;
    

    // Create the output files.
    // FIXME - should check whether or not to erase existing files !

    string outHeaderFile( outFileName );
    outHeaderFile += ".h";
    ofstream outHeader( outHeaderFile.c_str() );
    if( !outHeader ) {
        cout << "Error: cannot open file <" << outHeaderFile << "> for writing." << endl;
        exit( -1);
    }
    
    // Header header..
    outHeader << MOZILLA_LICENSE;
    outHeader << INCLUDE_HEADER << catName << " 0x" << hex;
    outHeader.width( 4 );
    outHeader.fill( '0' );
    outHeader << catIndex << endl << endl;
    outHeader.width( 0 );

    // Linux .cpp file ----------------------------------------------------
    string outCatalogFile( "../linux/" );
    outCatalogFile += outFileName;
    outCatalogFile += ".cpp";
    outCatalog.open( outCatalogFile.c_str() );
    if( !outCatalog ) {
        cout << "Error: cannot open file <" << outCatalogFile << "> for writing." << endl;
        exit( -1);
    }

    // Linux Header.
    outCatalog << MOZILLA_LICENSE
               << SOURCE_HEADER 
               << "#include \"" << outFileName << ".h\"" << endl << endl
               << "const char* " << outFileName << "[] = {" << endl;


    // Window .rc2 file ----------------------------------------------------
    string outResFile( "../win/" );
    outResFile += outFileName;
    outResFile += ".rc2";
    outResource.open( outResFile.c_str() );
    if( !outResource ) {
        cout << "Error: cannot open file <" << outResFile << "> for writing." << endl;
        exit( -1);
    }

    // Resource File Header.
    outResource << MOZILLA_LICENSE
                << RESOURCE_HEADER
				<< endl << "#include <..\\xp\\" << outHeaderFile << ">" << endl << endl
                << "STRINGTABLE DISCARDABLE " << endl
                << "BEGIN" << endl;



    int idCount = 0;
    int resCount = 0;
    int strCount = 0;

    outHeader.fill( ' ' );
    outHeader.setf( ios_base::left );

    bool bLinux, bWindows;
    int  read_offset;

    while( inFile.getline( inputLine, 5000, '\n' ) ) 
    {
        if( DiscardLine( inputLine ) ) 
        {
            continue;
        }

        // Reset the platform flags.
        bLinux = true;
        bWindows = true;

        // reset offset in the input buffer.
        read_offset = 0;

        // Now check whether or not we have a platform specific prefix 
        // on the line.
#ifndef _WIN32
		// onfortunately, this function does not exist on windows -bern
        if( !strncasecmp( "linux", inputLine, 5 ) )
#else
		if( !_strnicmp( "linux", inputLine, 5 ) )	 
#endif
        {
            bWindows = false;
            read_offset = 6;
        }
#ifndef _WIN32
		// onfortunately, this function does not exist on windows - bern
        else if( !strncasecmp( "windows", inputLine, 7 ) )
#else
		else if( !_strnicmp( "windows", inputLine, 7 ) )	 
#endif
        {
            bLinux = false;
            read_offset = 8;
        }


        // Accumulate the string ID.
        char strID[ 500 ];
        int cnt = 0;

        while( (inputLine[ cnt + read_offset ] != ' ') && 
               (inputLine[ cnt + read_offset ] != '\t'))
        {
            strID[ cnt ] = inputLine[ cnt + read_offset ];
            cnt++;
        }

        strID[ cnt ] = '\0';
        int offset = cnt + read_offset ;


        string newID( strID );
        pair< set< string >::iterator, bool > result = setOfIds.insert( newID );

        if( result.second == true )
        {
            // String ID is relative to the catalog index.
            outHeader << "#define " << setw( 50 ) << left << newID
                      << dec << catName << "+" << idCount << endl;
            ++idCount;
        }

        if( bLinux ) {
            outCatalog << "/* " << setw( 50 ) << left << newID << " */";
            ++strCount;
        }
        

        // Get rid of remaining white characters before 
        // outputting the string.
        while( (inputLine[offset] == ' ') || (inputLine[offset] == '\t') )
        {
            ++offset;
        }
        
        if( bLinux ) {
            outCatalog  << "    " << inputLine+offset << "," << endl;
        }

        // Windows resources. Group strings in tables of 16
        if( bWindows ) {
            if( (resCount != 0 ) && (resCount % 16) == 0 ) 
            {
                outResource << "END" << endl << endl
                            << "STRINGTABLE DISCARDABLE " << endl
                            << "BEGIN" << endl;
            }
        
            outResource << "    " << (inputLine + read_offset) << endl;
            ++resCount;
        }
    }

    outHeader  << "#define " << catName << "_SIZE " << idCount << endl;
    outCatalog << "};" << endl;
    outResource << "END" << endl;

    
    if( strCount != idCount ) {
        cout << endl << " ** ERROR: Platform Specific string defined for only one platform ** " << endl;
    }

	cout << "Hit enter to exit makecat." << endl;
	
	getchar();


	return 0;

}
