/*
 *  $Id: fr_elg.h,v 1.9.8.1 2000/10/11 13:17:46 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef FR_ELG_H_
#define FR_ELG_H_

/*
 *
 *
 *                             ElGamal Cryption
 *
 *
 */

#include "zks.h"
#include "zkError.h"
#include "zkIntTypes.h"
#include "zkClock.h"
#include "fr_pk.h"


/*
 *  For storage purposes the following ordering is defined on the cryption
 *  parameters: 0: y = g^x mod p; 1: x; 2: p; 3: g.
 *  
 *  We refer to y as the public key, and x as the private key.
 *  
 *  The numbers respresent the position of bits in a bitmask, which if set,
 *  indicate that parameter is stored.  Therefore, all possible combinations
 *  of the parameters _may_ be stored. In practice only some common subsets
 *  are stored, as defined below.
 */

#define FRCRYPT_CRYPTPARM_Y 0x01
#define FRCRYPT_CRYPTPARM_X 0x02
#define FRCRYPT_CRYPTPARM_P 0x04
#define FRCRYPT_CRYPTPARM_G 0x08

#define FRCRYPT_CRYPTPARMS_FIXED   (FRCRYPT_CRYPTPARM_P | FRCRYPT_CRYPTPARM_G)

#define FRCRYPT_CRYPTPARMS_ALL     (FRCRYPT_CRYPTPARM_Y | FRCRYPT_CRYPTPARM_X | \
                                    FRCRYPT_CRYPTPARM_P | FRCRYPT_CRYPTPARM_G)
#define FRCRYPT_CRYPTPARMS_PUB     FRCRYPT_CRYPTPARM_Y
#define FRCRYPT_CRYPTPARMS_PUBFX   (FRCRYPT_CRYPTPARMS_PUB | FRCRYPT_CRYPTPARMS_FIXED)
#define FRCRYPT_CRYPTPARMS_PRIV    FRCRYPT_CRYPTPARM_X
#define FRCRYPT_CRYPTPARMS_PRIVFX  (FRCRYPT_CRYPTPARMS_PRIV | FRCRYPT_CRYPTPARMS_FIXED)
#define FRCRYPT_CRYPTPARMS_PUBPRIV (FRCRYPT_CRYPTPARMS_PUB | FRCRYPT_CRYPTPARMS_PRIV)

/* Number of bytes needed to store a cryption private key. */
#define FRCRYPT_CRYPTPRIVKEY_SIZE  BITS_TO_BYTES(256)

/*
 *  Number of bytes needed to store the largest cryption public key.
 *  
 *  Currently, we limit the key length to 1024 bits.  Changing this would
 *  affect various buffer sizes, such as the pre-secret key material in
 *  route creation packets, etc.
 */
#define FRCRYPT_CRYPTPUBKEY_SIZE  BITS_TO_BYTES(1024)

/*
 *  The size of the shared secret in the ElGamal key agreement protocol.
 *  
 *  The preceding comment also applies to this value.
 */
#define FRCRYPT_SECRET_SIZE BITS_TO_BYTES(1024)


typedef FrCryptPubKey FrCryptCrypt;


/*
 *  These values indicate the size, in bits, of a key.  This refers only to
 *  one of the key's parameters, and should not be taken as an indication of
 *  how much storage is required to store the entire key.
 */
typedef enum {
  FRCRYPT_CRYPTKEYSIZE_UNUSED_,
  FRCRYPT_CRYPTKEYSIZE_1024 = 1024,
  FRCRYPT_CRYPTKEYSIZE_1536 = 1536,
  FRCRYPT_CRYPTKEYSIZE_2048 = 2048,
  FRCRYPT_CRYPTKEYSIZE_3072 = 3072,
  FRCRYPT_CRYPTKEYSIZE_4096 = 4096
} FrCryptCryptKeySize;


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

  extern DLLExport int frCryptCryptNew(ZkError err, FrCryptCrypt *key);
  extern DLLExport int frCryptCryptMakeCrypt(ZkError err, FrCryptCrypt cr);
  extern DLLExport int frCryptCryptNewFromPubKey(ZkError err, FrCryptCrypt *key, FrCryptPubKey *pk);
  extern DLLExport int frCryptCryptFree(ZkError err, FrCryptCrypt *key);
  extern DLLExport int frCryptCryptKeySizeFromInt(ZkError err, int n, FrCryptCryptKeySize *ksz);
  extern DLLExport int frCryptCryptGetPubKey(ZkError err, FrCryptCrypt key, FrCryptPubKey *pk);
  extern DLLExport int frCryptCryptSetVersion(ZkError err, FrCryptCrypt key, unsigned int vers);
  extern DLLExport int frCryptCryptGetVersion(ZkError err, FrCryptCrypt key, unsigned int *vers);
  extern DLLExport int frCryptCryptGetKeySize(ZkError err, FrCryptCrypt key, FrCryptCryptKeySize *ksz);
  extern DLLExport int frCryptCryptSetKeySize(ZkError err, FrCryptCrypt key, FrCryptCryptKeySize ksz);
  extern DLLExport int frCryptCryptGetExpiration(ZkError err, FrCryptCrypt key, ZkClock *exp);
  extern DLLExport int frCryptCryptSetExpiration(ZkError err, FrCryptCrypt key, ZkClock exp);
  extern DLLExport int fr_CryptCryptPrivKeyLoad(ZkError err, FrCryptCrypt key, void *data, size_t dsz);
  extern DLLExport int fr_CryptCryptPrivKeyStore(ZkError err, FrCryptCrypt key, void *data, size_t dsz);
  extern DLLExport int fr_CryptCryptPubKeyLoad(ZkError err, FrCryptCrypt key, void *data, size_t dsz);
  extern DLLExport int fr_CryptCryptPubKeyStore(ZkError err, FrCryptCrypt key, void *data, size_t dsz);
  extern DLLExport int frCryptCryptSetRandomKeys(ZkError err, FrCryptCrypt key);
  extern DLLExport int frCryptCryptSetRandomKeyWithPrivSize(ZkError err, FrCryptCrypt key, size_t nbits);
  extern DLLExport int frCryptCryptSetPreSecret(ZkError err, FrCryptCrypt key, void *preSec, size_t pssz, void *sec, size_t ssz);
  extern DLLExport int frCryptCryptSecretFromPrivKey(ZkError err, FrCryptCrypt key, void *data, size_t dsz, void *sec, size_t ssz);
#if 0
  extern DLLExport int frCryptCryptSecretFromPubKey(ZkError err, FrCryptCrypt key, void *data, size_t dsz, void *sec, size_t ssz);
#endif

  extern DLLExport int frCryptCryptEqual(ZkError err, FrCryptCrypt e0, FrCryptCrypt e1, int *equal);
  extern DLLExport int frCryptCryptStore(ZkError err, FrCryptCrypt elg, uint8_t mask, void *data, size_t *dsz);
  extern DLLExport int frCryptCryptNewFromData(ZkError err, FrCryptCrypt *elg, void *data, size_t *dsz);
  extern DLLExport int frCryptCryptNewFromFile(ZkError err, FrCryptCrypt *key, const char *file);

  extern DLLExport int fr_CryptCryptPrint(ZkError err, FrCryptCrypt key);
  extern DLLExport int fr_CryptCryptPrintAsC(ZkError err, FrCryptCrypt key, uint8_t mask, char *name, FILE *fp);

  extern DLLExport int frCryptCryptPublicValidate(ZkError err, FrCryptCrypt key);
  
#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* FR_ELG_H_ */
