/*
 *  $Id: frBf.c,v 1.12.8.4 2000/12/14 17:52:40 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#define CORRUPT_DEBUG 0

#include "zks.h"

#if CORRUPT_DEBUG
#include <assert.h>
#endif
#include <stdlib.h>
#include "frBf.h"
#include "zkrand.h"
#include "sym.h"
#include "zkError.h"


#define MYMOD ZKMFRCRYPT


static const size_t dflKeySize = BLOWFISH_DFL_KEY_SIZE;


/*
 *  These functions are internal to the crypto library; assume the sym
 *  argument is valid.
 */


int fr_BlowfishInit(ZkError err, FrCryptSym sym)
{
  zkErrorClear(err);
  
  sym->u.bfish.keySize = dflKeySize;
  return 1;
}


int fr_BlowfishGetKey(ZkError err, FrCryptSym sym, void *kbuf, size_t kbsz)
{
  const char *fn = "fr_BlowfishGetKey";
  int ret = 0;
  
  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    goto done;
  }
  if ( ! kbuf) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "kbuf is null");
    goto done;
  }
  if (kbsz < sym->u.bfish.keySize) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "buffer too small (%lu < %lu)",
                  (unsigned long)kbsz, (unsigned long)sym->u.bfish.keySize);
    goto done;
  }

  memcpy(kbuf, sym->u.bfish.key, sym->u.bfish.keySize);

  ret = 1;

 done:

  return ret;
}


int fr_BlowfishGetKeySize(ZkError err, FrCryptSym sym, size_t *sz)
{
  const char *fn = "fr_BlowfishGetKeySize";
  int ret = 0;
  
  if ( ! sym) {
    zkErrorSet(err, fn, ZKERROR, ZKMFRCRYPT, ZKENULLPTR, "sym is null");
    goto done;
  }
  if ( ! sz) {
    zkErrorSet(err, fn, ZKERROR, ZKMFRCRYPT, ZKENULLPTR, "sz is null");
    goto done;
  }
  
  *sz = sym->u.bfish.keySize;
  ret = 1;

 done:

  return ret;
}


int fr_BlowfishSetKeySize(ZkError err, FrCryptSym sym, size_t sz)
{
  zkErrorClear(err);
  
  /*
   *  Put a lower limit of 128 bits (16 bytes) on the key size.  The upper
   *  limit is 448 bits (56 bytes), as specified in the second edition of
   *  Schneir's Applied Cryptography.  The SSLeay documentation claims it can
   *  be up to 72 bytes, but we'll be conservative.
   */
  
  if (sz < 16 || sz > 56) {
    zkErrorSet_va(err, "fr_BlowfishSetKeySize", ZKWARN, ZKMFRCRYPT, ZKEPARMBAD,
                  "key (%lu) size must be between 16 and 56 bytes, inclusive",
                  (unsigned long)sz);
    return 0;
  }

  sym->u.bfish.keySize = sz;

  return 1;
}


int fr_BlowfishRandomKey(ZkError err, FrCryptSym sym)
{
  zkErrorClear(err);
  
  if ( ! zkRandGetData(err, sym->u.bfish.key, sym->u.bfish.keySize)) {
    return 0;
  }
  
  BF_set_key(&sym->u.bfish.expKey, sym->u.bfish.keySize, sym->u.bfish.key);

  return 1;
}


/*
 *  bug: this routine is horked.  Go over where and how it is used.
 */
int fr_BlowfishDecrypt(ZkError err, FrCryptSym sym, void *ivec, void *data, size_t sz)
{
  /* bug: check data, sz?  What will BF_ do if sz is wrong? */

  unsigned char *d;

  zkErrorClear(err);
  
  if ( ! data) {
    zkErrorSet(err, "fr_BlowfishDecrypt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
#if CORRUPT_DEBUG
  assert(sz % FRCRYPT_SIZE_MULT == 0);
  fprintf(stderr, "BlowfishDecrypt: sz is %lu.\n", (unsigned long)sz);
#endif
  if (sz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "fr_BlowfishDecrypt", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "sz (%lu) has an invalid value", (unsigned long)sz);
    return 0;
  }

  /*
   *  SSLeay guarantees that the pointers to the plaintext and ciphertext may
   *  be the same.
   */
  
  d = (unsigned char *)data;
  BF_cbc_encrypt(d, d, sz, &sym->u.bfish.expKey, ivec, 0);

  return 1;
}


/* Uses 2 buffers */
int fr_BlowfishDecryptToBuf(ZkError err, FrCryptSym sym, void *ivec, void *data, void *output, size_t sz)
{
  /* bug: check data, sz?  What will BF_ do if sz is wrong? */
  
  unsigned char *d;
  unsigned char *o;

  zkErrorClear(err);
  
  if ( ! data) {
    zkErrorSet(err, "fr_BlowfishDecrypt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! output) {
    zkErrorSet(err, "fr_BlowfishDecrypt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "output is null");
    return 0;
  }
#if CORRUPT_DEBUG
  assert(sz % FRCRYPT_SIZE_MULT == 0);
  fprintf(stderr, "BlowfishDecrypt: sz is %lu.\n", (unsigned long)sz);
#endif
  if (sz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "fr_BlowfishDecrypt", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "sz (%lu) has an invalid value", (unsigned long)sz);
    return 0;
  }

  d = (unsigned char *)data;
  o = (unsigned char *)output;
  BF_cbc_encrypt(d, o, sz, &sym->u.bfish.expKey, ivec, 0);

  return 1;
}


/*
 *  bug: this routine is horked.  Go over where and how it is used.
 */
int fr_BlowfishEncrypt(ZkError err, FrCryptSym sym, void *ivec, void *data, size_t sz)
{
  /* bug: check data, sz?  What will BF_ do if sz is wrong? */

  unsigned char *d;

  zkErrorClear(err);
  
  if ( ! data) {
    zkErrorSet(err, "fr_BlowfishEncrypt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
#if CORRUPT_DEBUG
  assert(sz % FRCRYPT_SIZE_MULT == 0);
  fprintf(stderr, "BlowfishEncrypt: sz is %lu.\n", (unsigned long)sz);
#endif
  if (sz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "fr_BlowfishEncrypt", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "sz (%lu) has an invalid value", (unsigned long)sz);
    return 0;
  }

  /*
   *  SSLeay guarantees that the pointers to the plaintext and ciphertext may
   *  be the same.
   */
  
  d = (unsigned char *)data;
  BF_cbc_encrypt(d, d, sz, &sym->u.bfish.expKey, ivec, 1);

  return 1;
}


/* Uses 2 buffers */
int fr_BlowfishEncryptToBuf(ZkError err, FrCryptSym sym, void *ivec, void *data, void *output, size_t sz)
{
  /* bug: check data, sz?  What will BF_ do if sz is wrong? */

  unsigned char *d;
  unsigned char *o;

  zkErrorClear(err);
  
  if ( ! data) {
    zkErrorSet(err, "fr_BlowfishEncrypt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! output) {
    zkErrorSet(err, "fr_BlowfishEncrypt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "output is null");
    return 0;
  }
#if CORRUPT_DEBUG
  assert(sz % FRCRYPT_SIZE_MULT == 0);
  fprintf(stderr, "BlowfishEncrypt: sz is %lu.\n", (unsigned long)sz);
#endif
  if (sz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "fr_BlowfishEncrypt", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "sz (%lu) has an invalid value", (unsigned long)sz);
    return 0;
  }

  d = (unsigned char *)data;
  o = (unsigned char *)output;
  BF_cbc_encrypt(d, o, sz, &sym->u.bfish.expKey, ivec, 1);

  return 1;
}


#if 0
int fr_BlowfishLoad(ZkError err, FrCryptSym sym, void *data, size_t sz)
{
  char *d = data;
  
  zkErrorClear(err);
  
  if ( ! sym) {
    zkErrorSet(err, "fr_BlowfishLoad", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sym is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "fr_BlowfishLoad", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }

  if (sz < BLOWFISH_PORTABLE_SIZE) {
    zkErrorSet(err, "fr_BlowfishLoad", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sz is smaller than required");
    return 0;
  }

  sym->alg = d[0]; /* bug? use INT8 macro? */
  sym->u.xor.key[0] = d[1];

  return 1;
}
#endif


int fr_BlowfishStore(ZkError err, FrCryptSym sym, void *data, size_t *sz)
{
  const char *fn = "fr_BlowfishStore";
  
  zkErrorClear(err);
  
  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "sym is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "data is null");
    return 0;
  }
  if ( ! sz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "sz is null");
    return 0;
  }

  if (*sz < sym->u.bfish.keySize) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "*sz is smaller than required");
    return 0;
  }

  memcpy(data, sym->u.bfish.key, sym->u.bfish.keySize);
  *sz = sym->u.bfish.keySize;
  
  return 1;
}
