/*
 *  $Id: frDes3.c,v 1.16.8.4 2000/12/14 17:52:40 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include <stdlib.h>
#include "frDes3.h"
#include "zkrand.h"
#include "sym.h"
#include "zkError.h"
#include "openssl/crypto.h"
#include "openssl/rand.h" /* des_random_key replacement for old SSLeay */


#define DES3_PORTABLE_SIZE 24
#define MYMOD ZKMFRCRYPT


/*
 *  These functions are internal to the crypto library; assume the sym
 *  argument is valid.
 */


int fr_Des3Init(ZkError err, FrCryptSym sym)
{
  zkErrorClear(err);
  (void)sym;

  return 1;
}


int fr_Des3GetKeySize(ZkError err, FrCryptSym sym, size_t *sz)
{
  const char *fn = "fr_Des3GetKeySize";
  int ret = 0;
  
  if ( ! sym) {
    zkErrorSet(err, fn, ZKERROR, ZKMFRCRYPT, ZKENULLPTR, "sym is null");
    goto done;
  }
  if ( ! sz) {
    zkErrorSet(err, fn, ZKERROR, ZKMFRCRYPT, ZKENULLPTR, "sz is null");
    goto done;
  }
  
  *sz = DES3_PORTABLE_SIZE;
  ret = 1;

 done:

  return ret;
}


int fr_Des3GetKey(ZkError err, FrCryptSym sym, void *kbuf, size_t kbsz)
{
  const char *fn = "fr_Des3GetKey";
  int ret = 0;
  
  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    goto done;
  }
  if ( ! kbuf) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "kbuf is null");
    goto done;
  }
  if (kbsz < DES3_PORTABLE_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "buffer too small (%lu < %d)",
                  (unsigned long)kbsz, DES3_PORTABLE_SIZE);
    goto done;
  }

  memcpy((char *)kbuf +  0, sym->u.des3.key1, sizeof sym->u.des3.key1);
  memcpy((char *)kbuf +  8, sym->u.des3.key2, sizeof sym->u.des3.key2);
  memcpy((char *)kbuf + 16, sym->u.des3.key3, sizeof sym->u.des3.key3);

  ret = 1;

 done:

  return ret;
}


int fr_Des3SetKeySize(ZkError err, FrCryptSym sym, size_t sz)
{
  zkErrorClear(err);
  zkErrorSet(err, "fr_Des3SetKeySize", ZKERR, ZKMFRCRYPT, ZKERROR,
             "DES3 keys have a fixed size");
  (void)sym; (void)sz;
  return 0;
}


int fr_des_random_key(ZkError err, des_cblock *ret)
{
  /* des_random_key is buggy in SSLeay */
  do {
    if (zkRandGetData(err, ret, sizeof(des_cblock)) == 0)
      return 0;
  } while (des_is_weak_key(ret));
  des_set_odd_parity(ret);
  return 1;
}


int fr_Des3RandomKey(ZkError err, FrCryptSym sym)
{
  zkErrorClear(err);

  if (fr_des_random_key(err, &sym->u.des3.key1) == 0 ||
      fr_des_random_key(err, &sym->u.des3.key2) == 0 ||
      fr_des_random_key(err, &sym->u.des3.key3) == 0)
      return 0;

  (void)des_set_key(&sym->u.des3.key1, sym->u.des3.ks1);
  (void)des_set_key(&sym->u.des3.key2, sym->u.des3.ks2);
  (void)des_set_key(&sym->u.des3.key3, sym->u.des3.ks3);

  return 1;
}


int fr_Des3Decrypt(ZkError err, FrCryptSym sym, void *ivec, void *data, size_t sz)
{
  des_cblock *c;

  zkErrorClear(err);

  if ( ! data) {
    zkErrorSet(err, "fr_Des3Decrypt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
#if CORRUPT_DEBUG
  assert(sz % FRCRYPT_SIZE_MULT == 0);
  fprintf(stderr, "Des3Decrypt: sz is %lu.\n", (unsigned long)sz);
#endif
  if (sz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "fr_Des3Decrypt", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "sz (%lu) has an invalid value", (unsigned long)sz);
    return 0;
  }

  /*
   *  SSLeay guarantees that the pointers to the plaintext and ciphertext may
   *  be the same.
   */
  
  c = data;

#if OPENSSL_VERSION_NUMBER >= 0x0922	/* OpenSSL 0.9.2b */
  /* some function prototypes are different between versions of SSLeay/OpenSSL */
  des_ede3_cbc_encrypt((unsigned char *) c, (unsigned char *) c, sz,
                       sym->u.des3.ks1, sym->u.des3.ks2, sym->u.des3.ks3,
                       ivec, 0);
#else
  des_ede3_cbc_encrypt(c, c, sz,
                       sym->u.des3.ks1, sym->u.des3.ks2, sym->u.des3.ks3,
                       ivec, 0);
#endif

  return 1;
}


/* Uses 2 buffers */
int fr_Des3DecryptToBuf(ZkError err, FrCryptSym sym, void *ivec, void *data, void *output, size_t sz)
{
  des_cblock *c;
  des_cblock *o;

  zkErrorClear(err);

  if ( ! data) {
    zkErrorSet(err, "fr_Des3Decrypt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! output) {
    zkErrorSet(err, "fr_Des3Decrypt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "ouput is null");
    return 0;
  }
#if CORRUPT_DEBUG
  assert(sz % FRCRYPT_SIZE_MULT == 0);
  fprintf(stderr, "Des3Decrypt: sz is %lu.\n", (unsigned long)sz);
#endif
  if (sz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "fr_Des3Decrypt", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "sz (%lu) has an invalid value", (unsigned long)sz);
    return 0;
  }

  c = data;
  o = output;

#if OPENSSL_VERSION_NUMBER >= 0x0922	/* OpenSSL 0.9.2b */
  /* some function prototypes are different between versions of SSLeay/OpenSSL */
  des_ede3_cbc_encrypt((unsigned char *) c, (unsigned char *) o, sz,
                       sym->u.des3.ks1, sym->u.des3.ks2, sym->u.des3.ks3,
                       ivec, 0);
#else
  des_ede3_cbc_encrypt(c, o, sz,
                       sym->u.des3.ks1, sym->u.des3.ks2, sym->u.des3.ks3,
                       ivec, 0);
#endif

  return 1;
}


int fr_Des3Encrypt(ZkError err, FrCryptSym sym, void *ivec, void *data, size_t sz)
{
  des_cblock *c;

  zkErrorClear(err);

  if ( ! data) {
    zkErrorSet(err, "fr_Des3Encrypt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
#if CORRUPT_DEBUG
  assert(sz % FRCRYPT_SIZE_MULT == 0);
  fprintf(stderr, "Des3Encrypt: sz is %lu.\n", (unsigned long)sz);
#endif
  if (sz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "fr_Des3Encrypt", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "sz (%lu) has an invalid value", (unsigned long)sz);
    return 0;
  }

  /*
   *  SSLeay guarantees that the pointers to the plaintext and ciphertext may
   *  be the same.
   */
  
  c = data;
#if OPENSSL_VERSION_NUMBER >= 0x0922	/* OpenSSL 0.9.2b */
  /* some function prototypes are different between versions of SSLeay/OpenSSL */
  des_ede3_cbc_encrypt((unsigned char *) c, (unsigned char *) c, sz,
                       sym->u.des3.ks1, sym->u.des3.ks2, sym->u.des3.ks3,
                       ivec, 1);
#else
  des_ede3_cbc_encrypt(c, c, sz,
                       sym->u.des3.ks1, sym->u.des3.ks2, sym->u.des3.ks3,
                       ivec, 1);
#endif

  return 1;
}

/* Uses 2 buffers */
int fr_Des3EncryptToBuf(ZkError err, FrCryptSym sym, void *ivec, void *data, void *output, size_t sz)
{
  des_cblock *c;
  des_cblock *o; 

  zkErrorClear(err);

  if ( ! data) {
    zkErrorSet(err, "fr_Des3Encrypt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! output) {
    zkErrorSet(err, "fr_Des3Encrypt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "ouput is null");
    return 0;
  }
#if CORRUPT_DEBUG
  assert(sz % FRCRYPT_SIZE_MULT == 0);
  fprintf(stderr, "Des3Encrypt: sz is %lu.\n", (unsigned long)sz);
#endif
  if (sz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "fr_Des3Encrypt", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "sz (%lu) has an invalid value", (unsigned long)sz);
    return 0;
  }
  
  c = data;
  o = output;
#if OPENSSL_VERSION_NUMBER >= 0x0922	/* OpenSSL 0.9.2b */
  /* some function prototypes are different between versions of SSLeay/OpenSSL */
  des_ede3_cbc_encrypt((unsigned char *) c, (unsigned char *) o, sz,
                       sym->u.des3.ks1, sym->u.des3.ks2, sym->u.des3.ks3,
                       ivec, 1);
#else
  des_ede3_cbc_encrypt(c, o, sz,
                       sym->u.des3.ks1, sym->u.des3.ks2, sym->u.des3.ks3,
                       ivec, 1);
#endif

  return 1;
}



#if 0
int fr_Des3Load(ZkError err, FrCryptSym sym, void *data, size_t sz)
{
  char *d = data;
  
  zkErrorClear(err);

  if ( ! sym) {
    zkErrorSet(err, "fr_Des3Load", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sym is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "fr_Des3Load", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }

  if (sz < DES3_PORTABLE_SIZE) {
    zkErrorSet(err, "fr_Des3Load", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sz is smaller than required");
    return 0;
  }

  sym->alg = d[0]; /* bug? use INT8 macro? */
  sym->u.xor.key[0] = d[1];

  return 1;
}
#endif


int fr_Des3Store(ZkError err, FrCryptSym sym, void *data, size_t *sz)
{
  char *d = data;
  const char *fn = "fr_Des3Store";
  
  zkErrorClear(err);

  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "sym is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "data is null");
    return 0;
  }
  if ( ! sz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "sz is null");
    return 0;
  }

  if (*sz < DES3_PORTABLE_SIZE) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "*sz is smaller than required");
    return 0;
  }

  /* bug: use sizeof, etc. */
  memcpy(d +  0, sym->u.des3.key1, sizeof sym->u.des3.key1);
  memcpy(d +  8, sym->u.des3.key2, sizeof sym->u.des3.key2);
  memcpy(d + 16, sym->u.des3.key3, sizeof sym->u.des3.key3);
  *sz = 24;

  return 1;
}
