/*
 *  $Id: frDesX.c,v 1.10.8.2 2000/12/14 17:52:40 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include <stdlib.h>
#include "frDesX.h"
#include "fr_rand.h"
#include "openssl/crypto.h"
#include "sym.h"
#include "zkError.h"


#define DESX_PORTABLE_SIZE 24
#define MYMOD ZKMFRCRYPT


static const size_t dflKeySize = DESX_DFL_KEY_SIZE;


/*
 *  These functions are internal to the crypto library; assume the sym
 *  argument is valid.
 */


int fr_DesxInit(ZkError err, FrCryptSym sym)
{
  zkErrorClear(err);
  (void)sym;
  return 1;
}


int fr_DesxGetKey(ZkError err, FrCryptSym sym, void *kbuf, size_t kbsz)
{
  const char *fn = "fr_DesxGetKey";
  int ret = 0;
  
  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    goto done;
  }
  if ( ! kbuf) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "kbuf is null");
    goto done;
  }
  if (kbsz < DESX_PORTABLE_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "buffer too small (%lu < %d)",
                  (unsigned long)kbsz, DESX_PORTABLE_SIZE);
    goto done;
  }

  memcpy((char *)kbuf +  0, sym->u.desx.key,  sizeof sym->u.desx.key);
  memcpy((char *)kbuf +  8, sym->u.desx.inw,  sizeof sym->u.desx.inw);
  memcpy((char *)kbuf + 16, sym->u.desx.outw, sizeof sym->u.desx.outw);

  ret = 1;

 done:

  return ret;
}


int fr_DesxGetKeySize(ZkError err, FrCryptSym sym, size_t *sz)
{
  const char *fn = "fr_DesxGetKeySize";
  int ret = 0;
  
  if ( ! sym) {
    zkErrorSet(err, fn, ZKERROR, ZKMFRCRYPT, ZKENULLPTR, "sym is null");
    goto done;
  }
  if ( ! sz) {
    zkErrorSet(err, fn, ZKERROR, ZKMFRCRYPT, ZKENULLPTR, "sz is null");
    goto done;
  }
  
  *sz = DESX_PORTABLE_SIZE;
  ret = 1;

 done:

  return ret;
}


int fr_DesxSetKeySize(ZkError err, FrCryptSym sym, size_t sz)
{
  zkErrorClear(err);
  (void)sym; (void)sz;
  zkErrorSet(err, "fr_DesxSetKeySize", ZKERR, ZKMFRCRYPT, ZKERROR,
             "DESX keys have a fixed size");
  return 0;
}


int fr_DesxRandomKey(ZkError err, FrCryptSym sym)
{
  zkErrorClear(err);

  if (fr_des_random_key(err, &sym->u.desx.key) == 0)
    return 0;

  (void)des_set_key(&sym->u.desx.key, sym->u.desx.ks);

  return 1;
}


int fr_DesxDecrypt(ZkError err, FrCryptSym sym, void *ivec, void *data, size_t sz)
{
  des_cblock *c;

  zkErrorClear(err);

  if ( ! data) {
    zkErrorSet(err, "fr_DesxDecrypt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
#if CORRUPT_DEBUG
  assert(sz % FRCRYPT_SIZE_MULT == 0);
  fprintf(stderr, "DesxDecrypt: sz is %lu.\n", (unsigned long)sz);
#endif
  if (sz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "fr_DesxDecrypt", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "sz (%lu) has an invalid value", (unsigned long)sz);
    return 0;
  }

  /*
   *  SSLeay guarantees that the pointers to the plaintext and ciphertext may
   *  be the same.
   */
  
  c = data;
#if OPENSSL_VERSION_NUMBER >= 0x0922	/* OpenSSL 0.9.2b */
  /* some function prototypes are different between versions of SSLeay/OpenSSL */
  des_xcbc_encrypt((unsigned char *) c, (unsigned char *) c, sz, sym->u.desx.ks, ivec,
                   &sym->u.desx.inw, &sym->u.desx.outw, 0);
#else
  des_xcbc_encrypt(c, c, sz, sym->u.desx.ks, ivec,
                   &sym->u.desx.inw, &sym->u.desx.outw, 0);
#endif

  return 1;
}

/* Uses 2 buffers */
int fr_DesxDecryptToBuf(ZkError err, FrCryptSym sym, void *ivec, void *data, void *output, size_t sz)
{
  des_cblock *c;
  des_cblock *o;

  zkErrorClear(err);

  if ( ! data) {
    zkErrorSet(err, "fr_DesxDecrypt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! output) {
    zkErrorSet(err, "fr_DesxDecrypt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "output is null");
    return 0;
  }
#if CORRUPT_DEBUG
  assert(sz % FRCRYPT_SIZE_MULT == 0);
  fprintf(stderr, "DesxDecrypt: sz is %lu.\n", (unsigned long)sz);
#endif
  if (sz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "fr_DesxDecrypt", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "sz (%lu) has an invalid value", (unsigned long)sz);
    return 0;
  }

  c = data;
  o = output;

#if OPENSSL_VERSION_NUMBER >= 0x0922	/* OpenSSL 0.9.2b */
  /* some function prototypes are different between versions of SSLeay/OpenSSL */
  des_xcbc_encrypt((unsigned char *) c, (unsigned char *) o, sz, sym->u.desx.ks, ivec,
                   &sym->u.desx.inw, &sym->u.desx.outw, 0);
#else
  des_xcbc_encrypt(c, o, sz, sym->u.des.ks, ivec,
                   &sym->u.desx.inw, &sym->u.desx.outw, 0);
#endif

  return 1;
}


int fr_DesxEncrypt(ZkError err, FrCryptSym sym, void *ivec, void *data, size_t sz)
{
  des_cblock *c;

  zkErrorClear(err);

  if ( ! data) {
    zkErrorSet(err, "fr_DesxEncrypt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
#if CORRUPT_DEBUG
  assert(sz % FRCRYPT_SIZE_MULT == 0);
  fprintf(stderr, "DesxEncrypt: sz is %lu.\n", (unsigned long)sz);
#endif
  if (sz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "fr_DesxEncrypt", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "sz (%lu) has an invalid value", (unsigned long)sz);
    return 0;
  }

  /*
   *  SSLeay guarantees that the pointers to the plaintext and ciphertext may
   *  be the same.
   */
  
  c = data;
#if OPENSSL_VERSION_NUMBER >= 0x0922	/* OpenSSL 0.9.2b */
  /* some function prototypes are different between versions of SSLeay/OpenSSL */
  des_xcbc_encrypt((unsigned char *) c, (unsigned char *) c, sz, sym->u.desx.ks, ivec,
                   &sym->u.desx.inw, &sym->u.desx.outw, 1);
#else
  des_xcbc_encrypt(c, c, sz, sym->u.desx.ks, ivec,
                   &sym->u.desx.inw, &sym->u.desx.outw, 1);
#endif

  return 1;
}


/* Uses 2 buffers */
int fr_DesxEncryptToBuf(ZkError err, FrCryptSym sym, void *ivec, void *data, void *output, size_t sz)
{
  des_cblock *c;
  des_cblock *o;

  zkErrorClear(err);

  if ( ! data) {
    zkErrorSet(err, "fr_DesxEncrypt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  } 
  if ( ! output) {
    zkErrorSet(err, "fr_DesxEncrypt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "output is null");
    return 0;
  }
#if CORRUPT_DEBUG
  assert(sz % FRCRYPT_SIZE_MULT == 0);
  fprintf(stderr, "DesxEncrypt: sz is %lu.\n", (unsigned long)sz);
#endif
  if (sz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "fr_DesxEncrypt", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "sz (%lu) has an invalid value", (unsigned long)sz);
    return 0;
  }

  c = data;
  o = output;

#if OPENSSL_VERSION_NUMBER >= 0x0922	/* OpenSSL 0.9.2b */
  /* some function prototypes are different between versions of SSLeay/OpenSSL */
  des_xcbc_encrypt((unsigned char *) c, (unsigned char *) o, sz, sym->u.desx.ks, ivec,
                   &sym->u.desx.inw, &sym->u.desx.outw, 1);
#else
  des_xcbc_encrypt(c, o, sz, sym->u.desx.ks, ivec,
                   &sym->u.desx.inw, &sym->u.desx.outw, 1);
#endif

  return 1;
}


#if 0
int fr_DesxLoad(ZkError err, FrCryptSym sym, void *data, size_t sz)
{
  char *d = data;
  
  zkErrorClear(err);

  if ( ! sym) {
    zkErrorSet(err, "fr_DesxLoad", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sym is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "fr_DesxLoad", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }

  if (sz < DESX_PORTABLE_SIZE) {
    zkErrorSet(err, "fr_DesxLoad", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sz is smaller than required");
    return 0;
  }

  sym->alg = d[0]; /* bug? use INT8 macro? */
  sym->u.xor.key[0] = d[1];

  return 1;
}
#endif


int fr_DesxStore(ZkError err, FrCryptSym sym, void *data, size_t *sz)
{
  char *d = data;
  const char *fn = "fr_DesxStore";
  
  zkErrorClear(err);

  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "sym is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "data is null");
    return 0;
  }
  if ( ! sz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "sz is null");
    return 0;
  }

  if (*sz < DESX_PORTABLE_SIZE) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "*sz is smaller than required");
    return 0;
  }

  /* bug: use sizeof, etc. */
  memcpy(d +  0, sym->u.desx.key, sizeof sym->u.desx.key);
  memcpy(d +  8, sym->u.desx.inw, sizeof sym->u.desx.inw);
  memcpy(d + 16, sym->u.desx.outw, sizeof sym->u.desx.outw);
  *sz = 24;

  return 1;
}
