/*
 *  $Id: frXor.c,v 1.15.8.3 2000/12/14 17:52:40 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#define DEBUG_NO_XOR 0

#include "zks.h"

#include <stdlib.h>
#include "frXor.h"
#include "zkrand.h"
#include "sym.h"
#include "zkError.h"

/*
 *  Size needed to store the cryption type and `key'.
 *  
 *  One byte for the type, and one byte for the `key'.
 */
#define XOR_PORTABLE_SIZE 1

/* Sanity check. */
#if XOR_PORTABLE_SIZE > FRCRYPT_SYM_PORTABLE_SIZE
#error "check the storage size"
#endif

/*
 *  These functions are internal to the crypto library; assume the sym
 *  argument is valid.
 */


int fr_XorInit(ZkError err, FrCryptSym sym)
{
  zkErrorClear(err);

#if DEBUG_NO_XOR

  fprintf(stderr, "\nfr_XorInit: *** no encryption ***\n\n");
  sym->u.xor.key[0] = '\0';

#else

  sym->u.xor.key[0] = 0x55;

#endif
  
  return 1;
}


int fr_XorSetKeySize(ZkError err, FrCryptSym sym, size_t sz)
{
  zkErrorClear(err);

  (void)sym;

  /*
   *  Key size must be 1.
   */
  
  if (sz != 1) {
    zkErrorSet_va(err, "fr_XorSetKeySize", ZKWARN, ZKMFRCRYPT, ZKEPARMBAD,
                  "key size (%lu) must be 1 byte", (unsigned long)sz);
    return 0;
  }

  return 1;
}


int fr_XorRandomKey(ZkError err, FrCryptSym sym)
{
  zkErrorClear(err);

#if DEBUG_NO_XOR

  fprintf(stderr, "\nfr_XorRandomKey: *** no encryption ***\n\n");
  sym->u.xor.key[0] = '\0';

#else

  if ( ! zkRandGetData(err, sym->u.xor.key, sizeof sym->u.xor.key)) {
    return 0;
  }

#endif
  
  return 1;
}


/*
 *  Assume all parameters are valid.
 */
static int xor_crypt(ZkError err, FrCryptSym sym, void *data, size_t sz)
{
  char *d = data;
  size_t i;

  zkErrorClear(err);

  if (sz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "xor_crypt", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "sz (%lu) has an invalid length", (unsigned long)sz);
    return 0;
  }

  for (i = 0; i < sz; i++) {
    d[i] ^= sym->u.xor.key[0];
  }

  return 1;
}

/*
 *  Assume all parameters are valid.
 *  Use 2 buffers
 */
static int xor_cryptToBuf(ZkError err, FrCryptSym sym, void *data, void *output, size_t sz)
{
  char *d = data;
  char *o = output;
  size_t i;

  zkErrorClear(err);

  if (sz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "xor_crypt", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "sz (%lu) has an invalid length", (unsigned long)sz);
    return 0;
  }

  for (i = 0; i < sz; i++) {
    o[i] = d[i] ^ sym->u.xor.key[0];
  }

  return 1;
}


/*
 *  Assume all parameters are valid.
 */
int fr_XorDecrypt(ZkError err, FrCryptSym sym, void *ivec, void *data, size_t sz)
{
  zkErrorClear(err);
  (void)ivec;
  return xor_crypt(err, sym, data, sz);
}

/*
 *  Assume all parameters are valid.
 *  Use 2 buffers
 */
int fr_XorDecryptToBuf(ZkError err, FrCryptSym sym, void *ivec, void *data, void *output, size_t sz)
{
  zkErrorClear(err);
  (void)ivec;
  return xor_cryptToBuf(err, sym, data, output, sz);
}


/*
 *  Assume all parameters are valid.
 */
int fr_XorEncrypt(ZkError err, FrCryptSym sym, void *ivec, void *data, size_t sz)
{
  zkErrorClear(err);
  (void)ivec;
  return xor_crypt(err, sym, data, sz);
}


/*
 *  Assume all parameters are valid.
 *  Use 2 buffers
 */
int fr_XorEncryptToBuf(ZkError err, FrCryptSym sym, void *ivec, void *data, void *output, size_t sz)
{
  zkErrorClear(err);
  (void)ivec;
  return xor_cryptToBuf(err, sym, data, output, sz);
}


int fr_XorLoad(ZkError err, FrCryptSym sym, void *data, size_t sz)
{
  char *d = data;
  
  zkErrorClear(err);

  if ( ! sym) {
    zkErrorSet(err, "fr_XorLoad", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sym is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "fr_XorLoad", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }

  if (sz < XOR_PORTABLE_SIZE) {
    zkErrorSet(err, "fr_XorLoad", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sz is smaller than required");
    return 0;
  }

  sym->alg = d[0]; /* bug? use INT8 macro? */
  sym->u.xor.key[0] = d[1];

  return 1;
}


int fr_XorStore(ZkError err, FrCryptSym sym, void *data, size_t *sz)
{
  char *d = data;
  const char *fn = "fr_XorStore";
  
  zkErrorClear(err);

  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "sym is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "data is null");
    return 0;
  }
  if ( ! sz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "sz is null");
    return 0;
  }

  if (*sz < XOR_PORTABLE_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
                  "*sz is smaller than required (%lu < %d)",
                  (unsigned long)*sz, XOR_PORTABLE_SIZE);
    return 0;
  }

  d[0] = sym->u.xor.key[0];
  *sz = 1;

  return 1;
}
