/*
 *  $Id: hash.c,v 1.8.10.1 2000/10/11 13:17:47 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include "openssl/sha.h"
#include "fr_hash.h"
#include "zkAlloc.h"


struct fr_crypt_hash {
  SHA_CTX sc;
};


/*
 *
 *
 *                            External functions
 *
 *
 */


DLLExport int frCryptHashNew(ZkError err, FrCryptHash *hash)
{
  FrCryptHash h = 0;
  const char *fn = "frCryptHashNew";
  int ret = 0;

  zkErrorClear(err);

  if ( ! hash) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hash is null");
    return 0;
  }

  if ( ! zkMemAlloc(err, (void **)&h, sizeof *h)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
               "can't allocate hash object");
    goto done;
  }

  SHA1_Init(&h->sc);

  *hash = h; h = 0; /* the caller now owns it */

  ret = 1;

 done:

  if (h) { zkMemFree(err, (void **)&h); }

  return ret;
}


DLLExport int frCryptHashFree(ZkError err, FrCryptHash *hash)
{
  FrCryptHash h;
  const char *fn = "frCryptHashFree";
  int ret = 0;

  zkErrorClear(err);

  if ( ! hash) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hash is null");
    goto done;
  }
  if ( ! (h = *hash)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "*hash is null");
    goto done;
  }

  if ( ! zkMemFree(err, (void **)&h)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
               "error freeing hash object");
    goto done;
  }

  *hash = 0;

  ret = 1;

 done:

  return ret;
}

DLLExport int frCryptHashReset(ZkError err, FrCryptHash hash)
{
  const char *fn = "frCryptHashNew";
  int ret = 0;

  zkErrorClear(err);

  if ( ! hash) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hash is null");
    return 0;
  }

  SHA1_Init(&hash->sc);

  ret = 1;
  return ret;
}



/*
 *  Update the hash over `dsz' bytes of `data'.
 *  
 *  Use this function when you need to create a hash over non-contiguous data
 *  buffers.  For example, when generating a hash over a large file, than only
 *  can be read one part at a time.
 */
DLLExport int frCryptHashUpdate(ZkError err, FrCryptHash hash, const void *data, size_t dsz)
{
  const char *fn = "frCryptHashUpdate";
  int ret = 0;

  zkErrorClear(err);

  if ( ! hash) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hash is null");
    goto done;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "data is null");
    goto done;
  }
  if (dsz == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD, "dsz is 0");
    goto done;
  }

  SHA1_Update(&hash->sc, data, dsz);

  ret = 1;

 done:

  return ret;
}


/*
 *  Get the hash that was generated by one or more calls to
 *  frCryptHashUpdate().
 */
DLLExport int frCryptHashGetHash(ZkError err, FrCryptHash hash, void *hbuf, size_t hsz)
{
  const char *fn = "frCryptHashGetHash";
  int ret = 0;

  zkErrorClear(err);

  if ( ! hash) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hash is null");
    goto done;
  }
  if ( ! hbuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hbuf is null");
    goto done;
  }
  if (hsz != FRCRYPT_HASH_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
                  "invalid hsz (%lu != %d)", (unsigned long)hsz, FRCRYPT_HASH_SIZE);
    goto done;
  }

  SHA1_Final(hbuf, &hash->sc);

  ret = 1;

 done:

  return ret;
}


/*
 *  Warning: this function was written before a hash object existed, therefore
 *  it doesn't take one as a parameter.
 */
DLLExport int frCryptHashGenerate(ZkError err, void *data, size_t dsz, void *hbuf, size_t hsz)
{
  SHA_CTX sha_ctx;
  const char *fn = "frCryptHashGenerate";

  zkErrorClear(err);

  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "data is null");
    return 0;
  }
  if (dsz == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD, "dsz is 0");
    return 0;
  }
  if ( ! hbuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hbuf is null");
    return 0;
  }
  if (hsz != FRCRYPT_HASH_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
                  "invalid hsz (%lu != %d)", (unsigned long)hsz, FRCRYPT_HASH_SIZE);
    return 0;
  }

  SHA1_Init(&sha_ctx);
  SHA1_Update(&sha_ctx, data, dsz);
  SHA1_Final(hbuf, &sha_ctx);

  return 1;
}


/*
 *  Warning: this function was written before a hash object existed, therefore
 *  it doesn't take one as a parameter.
 */
DLLExport int frCryptHashVerify(ZkError err, void *data, size_t dsz, void *hbuf, size_t hsz, int *valid)
{
  char newhash[FRCRYPT_HASH_SIZE];
  const char *fn = "frCryptHashVerify";

  zkErrorClear(err);

  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "data is null");
    return 0;
  }
  if (dsz == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD, "dsz is 0");
    return 0;
  }
  if ( ! hbuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hbuf is null");
    return 0;
  }
  if (hsz != FRCRYPT_HASH_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
                  "invalid hsz (%lu != %d)", (unsigned long)hsz, FRCRYPT_HASH_SIZE);
    return 0;
  }

  if ( ! frCryptHashGenerate(err, data, dsz, newhash, sizeof newhash)) {
    return 0;
  }

  *valid = memcmp(hbuf, newhash, hsz) == 0;

  return 1;
}
