/*
 *  $Id: init.c,v 1.19.10.3 2000/12/14 17:52:41 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include <errno.h>
#include <sys/types.h>
#define __USE_XOPEN_EXTENDED /* for fchmod() */
#include <sys/stat.h>
#undef __USE_XOPEN_EXTENDED
#include <fcntl.h>
#include <string.h>
#include <time.h>
#ifndef WIN32
#include <unistd.h>
#else
#include <assert.h>
#endif
#include "openssl/rand.h"
#include "elgamal.h"
#include "sig.h"
#include "fr_gencrypt.h"
#include "zkrand.h"
#include "zkLog.h"

#ifndef HAVE_YARROW
#  define USE_ENTROPY_FILE	1
#endif

static const int efMode = S_IRUSR | S_IWUSR; /* i.e. 0600 */
static const int permBits = S_IRWXU | S_IRWXG | S_IRWXO; /* i.e. 0777 */
static const size_t entSize = 2048;

static int cryptRefs = 0; /* ref. count for Init and Final */


#if USE_ENTROPY_FILE

/*
 *  Initialize the random number generator from the entropy pool.
 *  
 *  This routine is based on SSLeay's RAND_load_file(), but does more checking
 *  on file permissions.
 *  
 *  If the file is readable or writable by anyone other than us (except root,
 *  of course) try changing the permissions, then ignore the contents of the
 *  file.
 *  
 *  We assume the umask has been set appropriately.
 */
static int fr_CryptLoadEntropyFile(ZkError err, char *ef)
{
  int fd = -1, ret = 0;
  struct stat s;
  time_t tseed = time(0); /* always seed with this, at least */
  
  zkErrorClear(err);

  zkRandEntropyAdd(err, &tseed, sizeof tseed);

  if ( ! ef || ef[0] == '\0') {
    zkErrorSet(err, "fr_CryptLoadEntropyFile", ZKERR, ZKMFRCRYPT, ZKEDATABAD,
               "ef is null or empty string");
    goto done;
  }

  if ((fd = open(ef, O_CREAT | O_RDWR, efMode)) == -1) {
    zkErrorSet_va(err, "fr_CryptLoadEntropyFile", ZKERR, ZKMFRCRYPT, ZKEFILEOPEN,
                  "can't open `%s': %s", ef, strerror(errno));
    goto done;
  }
  if (fstat(fd, &s) == -1) {
    zkErrorSet_va(err, "fr_CryptLoadEntropyFile", ZKERR, ZKMFRCRYPT, ZKEFILEOPEN,
                  "can't fstat `%s': %s", ef, strerror(errno));
    goto done;
  }

  zkRandEntropyAdd(err, &s, sizeof s);

  /* It must be a regular file. */

  if ( ! S_ISREG(s.st_mode)) {
    zkErrorSet_va(err, "fr_CryptLoadEntropyFile", ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
                  "`%s' not a regular file", ef);
    goto done;
  }

  /*
   *  We must own it.
   *  
   *  Win32 Note: Win95 has a weak notion of security, so files don't "belong"
   *  to a particular user.  What I'll do here is assume that stat.st_uid == 0
   *  all the time, and simply #define geteuid() to 0 in zks.h, so that the
   *  file is always "owned" by us. This whole phile will have to be rewritten
   *  somehow to use Yarrow in the future.
   */

#ifdef WIN32
	assert(0 == s.st_uid && 0 == geteuid());
#endif

  {
    uid_t me = geteuid();
    if (s.st_uid != me) { 
      zkErrorSet_va(err, "fr_CryptLoadEntropyFile", ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
                    "file not owned by us (%d != %d)", s.st_uid, me);
      goto done;
    }
  }

  /* The permissions should deny access to anyone else. */

  if ((s.st_mode & permBits) & ~efMode) {
    /*
     *  Do what we can to fix the mode, but ignore the contents this time.
     */
    (void)fchmod(fd, efMode);
  } else {
    unsigned char seed[512];
    size_t tot = 0;
    ssize_t n;

    while ((n = read(fd, seed, sizeof seed)) > 0 && tot < entSize) {
      zkRandEntropyAdd(err, seed, sizeof seed); tot += n;
    }
    memset(seed, 0, sizeof seed);
  }

  ret = 1;

 done:

  if (fd != -1) { close(fd); }

  return ret;
}


static int fr_CryptStoreEntropyFile(ZkError err, char *ef)
{
  int fd;
  time_t tseed = time(0); /* always seed with this, at least */
  
  zkErrorClear(err);

  zkRandEntropyAdd(err, &tseed, sizeof tseed);

  if ( ! ef || ef[0] == '\0') {
    zkErrorSet(err, "fr_CryptStoreEntropyFile", ZKERR, ZKMFRCRYPT, ZKEDATABAD,
               "ef is null or empty string");
    return 0;
  }

  if ((fd = open(ef, O_CREAT | O_RDWR, efMode)) == -1) {
    zkErrorSet_va(err, "fr_CryptStoreEntropyFile", ZKERR, ZKMFRCRYPT, ZKEFILEOPEN,
                  "can't open `%s': %s", ef, strerror(errno));
    return 0;
  }

  /* bug? check ownership? */

  (void)fchmod(fd, efMode);

  {
    unsigned char seed[512];
    size_t tot = 0;
    ssize_t n = 1;

    while (tot < entSize && n > 0) {
      RAND_bytes(seed, sizeof seed);      
      if ((n = write(fd, seed, sizeof seed)) > 0) {
        tot += n;
      }
    }
    memset(seed, 0, sizeof seed);
  }

  close(fd);

  return 1;
}

#endif	/* USE_ENTROPY_FILE */


/*
 *  This should be called before using any crypto routines.
 *  
 *  A reference count is maintained to support calling Init and Final more
 *  than once: from constructors and destructors in the client, in particular.
 */
DLLExport int frCryptInit(ZkError err, char *entFile)
{
  int good;
  
  zkErrorClear(err);

  cryptRefs++;

  if (cryptRefs == 1) {
    
    good = zkRand_Init(err, entFile);
    
    if (good) {
    	good = fr_CryptCryptInit(err);
    	
    	if (good) {
    		good = fr_CryptSigInit(err);
    		
    		if (!good) {
    			fr_CryptCryptFinal(err);
    		}
    	}
    	
    	if (!good) {
    		zkRand_Final(err);
    	}
    }
    
    if (!good) {
      cryptRefs--;
      return 0;
    }

#if USE_ENTROPY_FILE
    if (entFile && ! fr_CryptLoadEntropyFile(err, entFile)) {
      fr_CryptSigFinal(err);
      fr_CryptCryptFinal(err);
      zkRand_Final(err);
      cryptRefs--;
      return 0;
    }
#endif	/* USE_ENTROPY_FILE */
  }

  return 1;
}


/*
 *  This should be called when no more crypto routines will be used.
 *  
 *  A reference count is maintained to support calling Init and Final more
 *  than once: from constructors and destructors in the client, in particular.
 */
DLLExport int frCryptFinal(ZkError err, char *entFile)
{
  const char *fn = "frCryptFinal";
  
  /* silence compiler */
  (void) entFile;

  zkErrorClear(err);

  cryptRefs--;
  
  if (cryptRefs < 0) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
                  "too many calls to %s", fn);
    return 0;
  } else if (cryptRefs == 0) {
    int ret = 1;
    
#if USE_ENTROPY_FILE
    if (entFile) {
      ret &= fr_CryptStoreEntropyFile(err, entFile);
    }
#endif	/* USE_ENTROPY_FILE */

    fr_CryptSigFinal(err);
    fr_CryptCryptFinal(err);
    zkRand_Final(err);
    
    return ret;
  }

  return 1;
}
