/*
 *  $Id: mac.c,v 1.14.10.4 2000/12/14 17:52:41 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "openssl/sha.h"
#include "fr_gencrypt.h"
#include "fr_mac.h"
#include "fr_sym.h"


#define DEBUG_MAC 0


#if DEBUG_MAC

static void prbuf(FILE *fp, const char *s, const void *buf, size_t blen)
{
  size_t i;
  const unsigned char *b = buf;

  fprintf(fp, "%s:", s);

  for (i = 0; i < blen; i++) {
    if (i % 20 == 0) { fputc('\n', fp); }
    fprintf(fp, " %02x", (unsigned)b[i]);
  }
  fputc('\n', fp);
}

#endif


struct fr_crypt_mac_key {
  int isSet;
  unsigned char key[FRCRYPT_MACKEY_SIZE];
};


/*
 *
 *
 *                            Internal functions
 *
 *
 */


/*
 *  Do HMAC-SHA1-n (n = 8*maclen).  Use the given key, of length keylen.  Put
 *  the output (at most maclen bytes) into mac.
 *  
 *  This code is from Ian Goldberg.
 */

static void hmac_sha1_n(unsigned char *key, unsigned int keylen,
                        unsigned char *data1, unsigned int len1,
                        unsigned char *data2, unsigned int len2,
                        unsigned char *mac, unsigned int maclen)
{
  SHA_CTX sha_ctx;
  unsigned char pad[64];
  unsigned char digest[20];
  int i;

  /* Construct the ipad */
  memset(pad, 0, 64);
  if (keylen <= 64) {
    /* Use the key material directly */
    memmove(pad, key, keylen);
  } else {
    /* We need to hash the key material first */
    SHA1_Init(&sha_ctx);
    SHA1_Update(&sha_ctx, key, keylen);
    SHA1_Final(digest, &sha_ctx);
    memmove(pad, digest, 20);
  }
  for(i=0;i<64;++i) pad[i] ^= 0x36;

  /* Do the inner hash */
  SHA1_Init(&sha_ctx);
  SHA1_Update(&sha_ctx, pad, 64);
  SHA1_Update(&sha_ctx, data1, len1);
  if (data2)
    SHA1_Update(&sha_ctx, data2, len2);
  SHA1_Final(digest, &sha_ctx);

  /* Convert the ipad to the opad */
  for(i=0;i<64;++i) pad[i] ^= (0x36 ^ 0x5c);

  /* Do the outer hash */
  SHA1_Init(&sha_ctx);
  SHA1_Update(&sha_ctx, pad, 64);
  SHA1_Update(&sha_ctx, digest, 20);
  SHA1_Final(digest, &sha_ctx);
  memset(&sha_ctx, 0, sizeof(sha_ctx));

  /* Copy the result into mac */
  if (maclen > 20) maclen = 20;
  memmove(mac, digest, maclen);

  /* Clean memory */
  memset(pad, 0, 64);
  memset(digest, 0, 20);
}


static void macInit(FrCryptMacKey k)
{
  k->isSet = 1;
}


/*
 *
 *
 *                            External functions
 *
 *
 */

DLLExport int frCryptMacKeyNew(ZkError err, FrCryptMacKey *key)
{
  FrCryptMacKey k;
  
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptMacKeyNew", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }

  if ( ! (k = malloc(sizeof *k))) {
    zkErrorSet_va(err, "frCryptMacKeyNew", ZKERR, ZKMFRCRYPT, ZKENOMEM,
                  "can't allocate %lu byte MAC object",
                  (unsigned long)sizeof *k);
    return 0;
  }

  macInit(k);

  *key = k;

  return 1;
}


DLLExport int frCryptMacKeyFree(ZkError err, FrCryptMacKey *key)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptMacKeyFree", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! *key) {
    zkErrorSet(err, "frCryptMacKeyFree", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
               "*key is null");
    return 0;
  }

  macInit(*key);
  memset(*key, 0, sizeof **key);
  free(*key); *key = 0;

  return 1;
}


/*
 *  Create a MAC key from an arbitrary amount of data.  In particular, we
 *  don't require a 20 byte key (as in Load), which is overkill.
 */
DLLExport int frCryptMacKeySetFromData(ZkError err, FrCryptMacKey key, void *data, size_t dsz)
{
  SHA_CTX sha_ctx;

  (void)err; /* unused */

  /*
   *  Hash the data to the right size.
   */

  SHA1_Init(&sha_ctx);
  SHA1_Update(&sha_ctx, data, dsz);
  SHA1_Final(key->key, &sha_ctx);
  memset(&sha_ctx, 0, sizeof sha_ctx);

  key->isSet = 1;

  return 1;
}


DLLExport int frCryptMacKeyLoad(ZkError err, FrCryptMacKey key, void *data, size_t dsz)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptMacKeyLoad", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "frCryptMacKeyLoad", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (dsz < FRCRYPT_MACKEY_SIZE) {
    zkErrorSet_va(err, "frCryptMacKeyLoad", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "dsz is too small (%lu < %d)",
                  (unsigned long)dsz, FRCRYPT_MACKEY_SIZE);
    return 0;
  }

  memcpy(key->key, data, FRCRYPT_MACKEY_SIZE);
  key->isSet = 1;

  return 1;
}


DLLExport int frCryptMacKeyStore(ZkError err, FrCryptMacKey key, void *data, size_t dsz)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptMacKeyStore", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "frCryptMacKeyStore", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (dsz < FRCRYPT_MACKEY_SIZE) {
    zkErrorSet_va(err, "frCryptMacKeyStore", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "dsz is too small (%lu < %d)",
                  (unsigned long)dsz, FRCRYPT_MACKEY_SIZE);
    return 0;
  }

  if ( ! key->isSet) {
    zkErrorSet(err, "frCryptMacKeyStore", ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "key has not been set");
    return 0;
  }

  memcpy(data, key->key, FRCRYPT_MACKEY_SIZE);

  return 1;
}


DLLExport int frCryptMacSetKeyFromSym(ZkError err, FrCryptMacKey mac, FrCryptSym sym)
{
  SHA_CTX sha_ctx;
  char t[128];
  const char *fn = "frCryptMacSetKeyFromSym";
  size_t tsz = sizeof t;

  zkErrorClear(err);

  if ( ! mac) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "mac is null");
    return 0;
  }
  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "sym is null");
    return 0;
  }

  if ( ! frCryptSymStore(err, sym, t, &tsz)) {
    return 0;
  }

  /*
   *  Hash the symmetric key just to ensure it's the right size for a MAC key.
   */

  SHA1_Init(&sha_ctx);
  SHA1_Update(&sha_ctx, t, tsz);
  SHA1_Final(mac->key, &sha_ctx);
  memset(&sha_ctx, 0, sizeof sha_ctx);

  mac->isSet = 1;

  return 1;
}


DLLExport int frCryptMacSetKeyZero(ZkError err, FrCryptMacKey key)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptMacSetKeyZero", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }

  memset(key->key, 0, FRCRYPT_MACKEY_SIZE);

  key->isSet = 1;

  return 1;
}


/*
 *  Generate a message authentication code (MAC) over `dsz' bytes of `data',
 *  storing the first `msz' bytes of the MAC into `mac'.
 *  
 *  Ensure the key size is at least 20 bytes, and that no less than 10 bytes
 *  of the MAC are stored.
 */
DLLExport int frCryptMacIntoBuffer(ZkError err, FrCryptMacKey key, void *data, size_t dsz, void *mac, size_t msz)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptMacIntoBuffer", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "frCryptMacIntoBuffer", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (dsz == 0) {
    zkErrorSet(err, "frCryptMacIntoBuffer", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
               "dsz is 0");
    return 0;
  }
  if ( ! mac) {
    zkErrorSet(err, "frCryptMacIntoBuffer", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "mac is null");
    return 0;
  }
  if (msz < 10 || msz > 20) {
    zkErrorSet_va(err, "frCryptMacIntoBuffer", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "msz (%lu) has an invalid value", (unsigned long)msz);
    return 0;
  }

  if ( ! key->isSet) {
    zkErrorSet(err, "frCryptMacIntoBuffer", ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "key has not been set");
    return 0;
  }

  hmac_sha1_n(key->key, sizeof key->key, data, dsz, 0, 0, mac, msz);

#if DEBUG_MAC
  prbuf(stderr, "mac key", key->key, sizeof key->key);
  prbuf(stderr, "mac", mac, msz);
  prbuf(stderr, "data", data, dsz);
  fputs("\n", stderr);
  fputs("\n", stderr);
#endif

  return 1;
}


DLLExport int frCryptMacVerifyFromBuffer(ZkError err, FrCryptMacKey key, void *data, size_t dsz, void *mac, size_t msz, int *valid)
{
  int ret;
  unsigned char m[FRCRYPT_MAC_SIZE];
  
  zkErrorClear(err);

  if ((ret = frCryptMacIntoBuffer(err, key, data, dsz, m, sizeof m))) {
    *valid = memcmp(m, mac, msz) == 0;
  }
  memset(m, 0, sizeof m);

  return ret;
}

/*
 *  Generate a message authentication code (MAC) over `dsz' bytes of `data',
 *  storing the first `msz' bytes of the MAC into `mac'.
 *  
 *  Ensure the key size is at least 20 bytes, and that no less than 10 bytes
 *  of the MAC are stored.
 */
DLLExport int frCryptMac2IntoBuffer(ZkError err, FrCryptMacKey key, void *data1, size_t dsz1, void *data2, size_t dsz2, void *mac, size_t msz)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptMac2IntoBuffer", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! data1 || ! data2) {
    zkErrorSet(err, "frCryptMac2IntoBuffer", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (dsz1 == 0 || dsz2 == 0) {
    zkErrorSet(err, "frCryptMac2IntoBuffer", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
               "dsz is 0");
    return 0;
  }
  if ( ! mac) {
    zkErrorSet(err, "frCryptMac2IntoBuffer", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "mac is null");
    return 0;
  }
  if (msz < 10 || msz > 20) {
    zkErrorSet_va(err, "frCryptMac2IntoBuffer", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "msz (%lu) has an invalid value", (unsigned long)msz);
    return 0;
  }

  if ( ! key->isSet) {
    zkErrorSet(err, "frCryptMac2IntoBuffer", ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "key has not been set");
    return 0;
  }

  hmac_sha1_n(key->key, sizeof key->key, data1, dsz1, data2, dsz2, mac, msz);

#if DEBUG_MAC
  prbuf(stderr, "mac key", key->key, sizeof key->key);
  prbuf(stderr, "mac", mac, msz);
  prbuf(stderr, "data", data, dsz);
  fputs("\n", stderr);
  fputs("\n", stderr);
#endif

#if 0
  memset(mac, msz, 0); // FNORD
#endif

  return 1;
}


DLLExport int frCryptMac2VerifyFromBuffer(ZkError err, FrCryptMacKey key, void *data1, size_t dsz1, void *data2, size_t dsz2, void *mac, size_t msz, int *valid)
{
  int ret;
  unsigned char m[FRCRYPT_MAC_SIZE];
  
  zkErrorClear(err);

  if ((ret = frCryptMac2IntoBuffer(err, key, data1, dsz1, data2, dsz2, m, sizeof m))) {
    *valid = memcmp(m, mac, msz) == 0;
  }
  memset(m, 0, sizeof m);

  return ret;
}
