/*
 *  $Id: pk.c,v 1.38.8.3 2000/12/14 17:52:41 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Generic public keys: cryption and signing.
 *  
 *  The intention is to have this API supplant that in pubkey.c, as well as
 *  take over functionality common to cryption and signing keys.
 */

#include "zks.h"

#include <errno.h>
#define _GNU_SOURCE /* for snprintf() */
#include <stdio.h>
#undef _GNU_SOURCE
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifndef WIN32
#include <netinet/in.h>
#else
#include <winsock2.h>
#endif

#include "fr_gencrypt.h"
#include "fr_pk.h"
#include "fr_blinding.h"
#include "frBn.h"
#include "pk.h"
#include "frEnt.h"
#include "frPktFmt.h"
#include "zkAlloc.h"
#include "zkCStr.h"
#include "zkCode.h"
#include "zkFile.h"
#include "zkIntTypes.h"
#include <zkClock.h>


const ZkClock minExpiration = ZKCLOCK_INIT(631152000); /* Mon Jan 01 00:00:00 GMT 1990 */
/* If a nym key has the following magic date it can't create routes */
const ZkClock CantCreateRoutesMagicExpiration = ZKCLOCK_INIT(631152001); 

const unsigned int minVersion = 1;
const unsigned int maxVersion = 255;

/*
 *  
 *  
 *                            Internal Routines
 *  
 *  
 */


static void pubkeyInit(FrCryptPubKey k)
{
  int i;

  k->vers = 0;
  k->type = FRCRYPT_KEYTYPE_UNUSED_;
  k->alg = FRCRYPT_PUBALG_UNUSED_;
  k->size = 0;
  k->expires = zkClockFromSeconds(0);
  memset(&k->ent, 0, sizeof k->ent);
  k->freeMask = 0;
  k->saveMask = 0;
  k->blinding_flag = 0;
  k->bl = 0;

  for (i = 0; i < NUM_PARMS; i++) { k->parm[i] = 0; }
}


static void resetKey(FrCryptPubKey k)
{
  int i;

  for (i = 0; i < NUM_PARMS; i++) {
    if (k->parm[i] && (k->freeMask & (1 << i))) {
      BN_clear_free(k->parm[i]);
    }
    k->parm[i] = 0;
  }

  pubkeyInit(k);
  
}


/*
 *  
 *  
 *                            External Routines
 *  
 *  
 */


DLLExport int frCryptPubKeyNew(ZkError err, FrCryptPubKey *key)
{
  const char *fn = "frCryptPubKeyNew";
  FrCryptPubKey k = 0;
  int ret = 0;

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    goto done;
  }

  if ( ! (k = malloc(sizeof *k))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKENOMEM,
                  "can't allocate new %lu generic public key object: %s",
                  (unsigned long)sizeof *k, strerror(errno));
    goto done;
  }

  pubkeyInit(k);

  if( ! frCryptPubKeySetBlinding(err, k, 1)) {
    goto done;
  }  

  *key = k;
  k = 0;   
  
  ret = 1;
 done:
   if (k) { free(k); }
   return ret;
}


/*
 *  Assume a key stored on disk will be smaller than this size.
 */
#define KEY_SIZE_MAX 1024


DLLExport int frCryptPubKeyNewFromFp(ZkError err, FrCryptPubKey *key, FILE *fp)
{
  FrCryptPubKey k = 0;
  int ret = 0;
  size_t bsz;
  void *buf = 0;

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeyNewFromFp", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! fp) {
    zkErrorSet(err, "frCryptPubKeyNewFromFp", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "fp is null");
    return 0;
  }

#if 1
  if ( ! zkFileReadFp(err, fp, KEY_SIZE_MAX + 1, &buf, &bsz)) {
    goto done;
  }

  if (bsz > KEY_SIZE_MAX) {
    zkErrorSet_va(err, "frCryptPubKeyNewFromFp", ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                  "key should be less than %d bytes", KEY_SIZE_MAX);
    goto done;
  }

#else

  if ( ! zkFileLengthFp(err, fp, &bsz)) {
    goto done;
  }

  if ( ! (buf = malloc(bsz))) {
    zkErrorSet_va(err, "frCryptPubKeyNewFromFp", ZKERR, ZKMFRCRYPT, errno,
               "can't allocate buffer to hold public key: %s", strerror(errno));
    goto done;
  }

  if (fread(buf, 1, bsz, fp) != bsz) {
    goto done;
  }
#endif

  if ( ! (frCryptPubKeyNew(err, &k) &&
          frCryptPubKeyReadFromBuf(err, k, buf, &bsz))) {
    goto done;
  }

  *key = k; k = 0;
  ret = 1;

 done:

  if (buf) { memset(buf, 0, bsz); free(buf); }
  if (k) { frCryptPubKeyFree(0, &k); }

  return ret;
}


DLLExport int frCryptPubKeyNewFromFile(ZkError err, FrCryptPubKey *key, const char *file)
{
  FILE *fp = 0;
  int ret = 0;

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeyNewFromFile", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! file) {
    zkErrorSet(err, "frCryptPubKeyNewFromFile", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "file is null");
    return 0;
  }

  if ( ! (fp = fopen(file, "rb"))) {
    zkErrorSet_va(err, "frCryptPubKeyNewFromFile", ZKERR, ZKMFRCRYPT, errno,
                  "can't open `%s' for reading: %s", file, strerror(errno));
    goto done;
  }

  ret = frCryptPubKeyNewFromFp(err, key, fp);

 done:

  if (fp) { fclose(fp); }

  return ret;
}


/*
 *  Make a `copy' of `key'.
 *  
 *  We make copies of those bignums whose bit is set in freeMask, but only
 *  copy the pointer to those bignums whose bit is not set.
 */
DLLExport int frCryptPubKeyDup(ZkError err, FrCryptPubKey *copy, FrCryptPubKey key)
{
  FrCryptPubKey k = 0;
  int i, ret = 0;
  FrCryptBlinding tmp = 0;


  zkErrorClear(err);

  if ( ! copy) {
    zkErrorSet(err, "frCryptPubKeyDup", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "copy is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeyDup", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }

  if ( ! frCryptPubKeyNew(err, &k)) {
    goto done;
  }

  tmp = k->bl;

  memcpy(k, key, sizeof *k);

  k->bl = tmp;

  for (i = 0; i < NUM_PARMS; i++) {
    if (k->freeMask & (1 << i) && ! (k->parm[i] = BN_dup(key->parm[i]))) {
      zkErrorSet(err, "frCryptPubKeyDup", ZKERR, ZKMFRCRYPT, ZKERROR,
                 "error from BN_dup()");
      goto done;
    }
  }

  *copy = k; k = 0;

  ret = 1;

 done:
  
  if (k) { frCryptPubKeyFree(0, &k); }

  return ret;
}


DLLExport int frCryptPubKeyFree(ZkError err, FrCryptPubKey *key)
{
  FrCryptPubKey k;

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeyFree", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! (k = *key)) {
    zkErrorSet(err, "frCryptPubKeyFree", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
               "*key is null");
    return 0;
  }
  
  /* free bliding structure if it exists */
  if (k->bl) {
    if ( ! fr_CryptBlindingFree(err, &(k->bl))) {
      return 0;
    }
  }

  resetKey(k);

  free(k);
  *key = 0;

  return 1;
}


DLLExport int frCryptPubKeyGetVersion(ZkError err, FrCryptPubKey key, unsigned int *vers)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeyGetVersion", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! vers) {
    zkErrorSet(err, "frCryptPubKeyGetVersion", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "vers is null");
    return 0;
  }

  *vers = key->vers;

  return 1;
}


DLLExport int frCryptPubKeySetVersion(ZkError err, FrCryptPubKey key, unsigned int vers)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeySetVersion", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if (vers < minVersion || vers > maxVersion) {
    zkErrorSet_va(err, "frCryptPubKeySetVersion", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "vers (%u) must be in the range [%u, %u]",
                  vers, minVersion, maxVersion);
    return 0;
  }

  key->vers = vers;

  return 1;
}


DLLExport unsigned int frCryptPubKeyMinVersion(void)
{
  return minVersion;
}


DLLExport int frCryptPubKeyNextVersion(ZkError err, FrCryptPubKey key, unsigned int *vers)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeyNextVersion", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! vers) {
    zkErrorSet(err, "frCryptPubKeyNextVersion", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "vers is null");
    return 0;
  }

  *vers = key->vers + 1;
  if (*vers > maxVersion) { *vers = minVersion; }

  return 1;
}


DLLExport int frCryptPubKeyGetType(ZkError err, FrCryptPubKey key, FrPubKeyType *type)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeyGetType", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! type) {
    zkErrorSet(err, "frCryptPubKeyGetType", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "type is null");
    return 0;
  }

  *type = key->type;

  return 1;
}


/*
 *  We force the setting of the algoritm field, as there is currently a
 *  one-to-one mapping between the key type and the algoritm.  Consequently,
 *  there is no function to independantly set the algorithm.
 */
DLLExport int frCryptPubKeySetType(ZkError err, FrCryptPubKey key, FrPubKeyType type)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeySetType", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }

  switch (type) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    key->alg = FRCRYPT_PUBALG_ELGAMAL;
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    key->alg = FRCRYPT_PUBALG_DSA;
    break;

  default:
    zkErrorSet_va(err, "frCryptPubKeySetType", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "key type has value %u", type);
    return 0;
  }

  key->type = type;

  return 1;
}


/*
 *  Copy the key's entity to `ent'.
 *  
 *  It's an error if the entity has not been set.
 */
DLLExport int frCryptPubKeyGetEntity(ZkError err, FrCryptPubKey key, FrEntIdExp *ent)
{
  zkErrorClear(err);
  
  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeyGetEntity", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! ent) {
    zkErrorSet(err, "frCryptPubKeyGetEntity", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "ent is null");
    return 0;
  }

  return frEntCopy(err, ent, &key->ent);
}


DLLExport int frCryptPubKeySetEntity(ZkError err, FrCryptPubKey key, FrEntIdExp *ent)
{
  zkErrorClear(err);
  
  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeySetEntity", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! ent) {
    zkErrorSet(err, "frCryptPubKeySetEntity", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "ent is null");
    return 0;
  }

  return frEntCopy(err, &key->ent, ent);
}


/*
 *  Return the "size" of the key, in bits.
 *  
 *  Typically, this is the maximum size of one or more of the key parameters.
 */
DLLExport int frCryptPubKeyGetSize(ZkError err, FrCryptPubKey key, unsigned int *nbits)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeyGetSize", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! nbits) {
    zkErrorSet(err, "frCryptPubKeyGetSize", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "nbits is null");
    return 0;
  }

  *nbits = key->size;

  return 1;
}


DLLExport int frCryptPubKeySetSize(ZkError err, FrCryptPubKey key, unsigned int nbits)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeySetSize", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }

  key->size = nbits;

  return 1;
}


DLLExport int frCryptPubKeyGetExpiration(ZkError err, FrCryptPubKey key, ZkClock *exp)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeyGetExpiration", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! exp) {
    zkErrorSet(err, "frCryptPubKeyGetExpiration", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "exp is null");
    return 0;
  }

  *exp = key->expires;

  return 1;
}


DLLExport int frCryptPubKeySetExpiration(ZkError err, FrCryptPubKey key, ZkClock exp)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeySetExpiration", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }

  /* Have to be careful to not overwrite the expires field if this is a magic
     free nym. */

  if (zkClockCmp(key->expires, CantCreateRoutesMagicExpiration) != 0) {
    key->expires = exp;
  } else {
    zkErrorSet(err, "frCryptPubKeySetExpiration", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "can't set expiration of this key, overloaded value");
    return 0;
  }

  return 1;
}


/*
 *  Set `*bn' to the value of the pointer at position `n' of our parameter
 *  array.  Set `*willFree' to 0 if this bignum will not be freed when the
 *  public key object is freed; set it to 1 if it will be freed.  Set
 *  `*willSave' to 0 if this bignum will not be stored when the key is written
 *  to a buffer; set it to 1 if it will be stored.
 *  
 *  An example of parameters that one would not want to free are those
 *  pointing to static storage, elsewhere in the program.
 */
DLLExport int frCryptPubKeyGetParm(ZkError err, FrCryptPubKey key, size_t n, BIGNUM **bn, int *willFree, int *willSave)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeyGetParm", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if (n > 7) {
    zkErrorSet_va(err, "frCryptPubKeyGetParm", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "key parameter number (%lu) must be less than 7", (unsigned long)n);
    return 0;
  }
  if ( ! bn) {
    zkErrorSet(err, "frCryptPubKeyGetParm", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "bn is null");
    return 0;
  }

  *bn = key->parm[n];
  *willFree = !! (key->freeMask & (1 << n));
  *willSave = !! (key->saveMask & (1 << n));
  
  return 1;
}


/*
 *  Directly copy the pointer `bn' into position `n' of our parameter array.
 *  
 *  If `freeIt' is 0 it will not be automatically freed when the public key
 *  object is freed, otherwise it will be freed.  If `saveIt' is 0 it will not
 *  be stored when the public key object is written to a buffer, otherwise it
 *  will be stored.
 *  
 *  It is an error to try to set a parameter that already has a value.
 */
DLLExport int frCryptPubKeySetParm(ZkError err, FrCryptPubKey key, size_t n, BIGNUM *bn, int freeIt, int saveIt)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeySetParm", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if (n > 7) {
    zkErrorSet_va(err, "frCryptPubKeySetParm", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "key parameter number (%lu) must be less than 7", (unsigned long)n);
    return 0;
  }
  if ( ! bn) {
    zkErrorSet(err, "frCryptPubKeySetParm", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "bn is null");
    return 0;
  }

  if (key->parm[n]) {
    zkErrorSet_va(err, "frCryptPubKeySetParm", ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
                  "key parameter %lu is already set", (unsigned long)n);
    return 0;
  }

  key->parm[n] = bn;
  key->freeMask |= (( !! freeIt) << n);
  key->saveMask |= (( !! saveIt) << n);

  return 1;
}


/*
 *  Get the `mask' indicating which bignum parameters will be freed by the
 *  `key' object.
 *  
 *  The n'th bit of `mask' corresponds to the n'th parameter.
 */
DLLExport int frCryptPubKeyGetFreeMask(ZkError err, FrCryptPubKey key, uint8_t *mask)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeyGetFreeMask", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }

  *mask = key->freeMask;

  return 1;
}


/*
 *  Each set bit in `mask' indicates that the `key' object "owns" the
 *  corresponding bignum parameter, and that it is responsible for freeing it.
 *  
 *  The n'th bit of `mask' corresponds to the n'th parameter.
 */
DLLExport int frCryptPubKeySetFreeMask(ZkError err, FrCryptPubKey key, uint8_t mask)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeySetFreeMask", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }

  key->freeMask = mask;

  return 1;
}


/*
 *  Get the `mask' indicating which bignum parameters will be stored when the
 *  `key' object is written to a buffer.
 *  
 *  The n'th bit of `mask' corresponds to the n'th parameter.
 */
DLLExport int frCryptPubKeyGetSaveMask(ZkError err, FrCryptPubKey key, uint8_t *mask)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeyGetSaveMask", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }

  *mask = key->saveMask;

  return 1;
}


/*
 *  Each set bit in `mask' indicates that the `key' object should store the
 *  corresponding bignum parameter when told to write the object to a buffer.
 *  
 *  The n'th bit of `mask' corresponds to the n'th parameter.
 */
DLLExport int frCryptPubKeySetSaveMask(ZkError err, FrCryptPubKey key, uint8_t mask)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeySetSaveMask", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }

  key->saveMask = mask;

  return 1;
}


#pragma pack(1)

/*
 *  The following structure is a packed representation of information relating
 *  to a key.  It's used for the portable transmission and storage of keys.
 *  
 *  The actual key parameters will immediately follow this structure.  Each is
 *  stored as a two byte length, followed by that many bytes representing the
 *  bignum.
 *  
 *  All values are stored in network byte order (i.e. big-endian).
 */

struct fr_crypt_key_pk {
  uint8_t vers;
  uint8_t type;
  uint8_t alg;
  uint16_t size;
  ZkClock expires;
  char ent[FRENT_PORTABLE_SIZE];
  uint8_t mask; /* bitmask indicating which of 8 possible bignums are stored */

  uint16_t totBytes; /* total # bytes used to store key */
};

#pragma pack()


/*
 *  Return a bit mask corresponding to the parameters set in `key'.
 */
static uint8_t parmMask(FrCryptPubKey key)
{
  int i;
  uint8_t mask = 0;
  
  for (i = 0; i < NUM_PARMS; i++) {
    if (key->parm[i]) {
      mask |= (1 << i);
    }
  }

  return mask;
}


/*
 *  Return the total number of bytes required to store the contents of `key'.
 */
static size_t bytesRequired(FrCryptPubKey key)
{
  size_t i, n;
  uint8_t mask = parmMask(key) & key->saveMask;
  
  n = sizeof(struct fr_crypt_key_pk);

  for (i = 0; i < NUM_PARMS; i++) {
    if (mask & (1 << i)) {
      n += 2 + BN_num_bytes(key->parm[i]);
    }
  }

  return n;
}


DLLExport int frCryptPubKeyStorageSize(ZkError err, FrCryptPubKey key, size_t *n)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeyStorageSize", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! n) {
    zkErrorSet(err, "frCryptPubKeyStorageSize", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "n is null");
    return 0;
  }

  *n = bytesRequired(key);

  return 1;
}


/*
 *  Initialize a `key' from the portable representation stored in `b'.
 *  
 *  Initially, `*bsz' is the total size of the buffer pointed to by `b'.  Upon
 *  a successful return it will be the actual number of bytes occupied by the
 *  stored key.
 */
DLLExport int frCryptPubKeyReadFromBuf(ZkError err, FrCryptPubKey key, char *b, size_t *bsz)
{
  const char *fn = "frCryptPubKeyReadFromBuf";
  int i, ret = 0;
  size_t totBytes, need;
  struct fr_crypt_key_pk *k = (void *)b;

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "key is null");
    return 0;
  }
  if ( ! b) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "b is null");
    return 0;
  }
  if ( ! bsz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "bsz is null");
    return 0;
  }
  
  if (*bsz < sizeof *k) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "*bsz is too small (%lu < %lu)",
                  (unsigned long)*bsz, (unsigned long)sizeof *k);
    return 0;
  }

  totBytes = ntohs(k->totBytes);

  if (*bsz < totBytes) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "*bsz is less than expected key size (%lu < %lu)",
                  (unsigned long)*bsz, (unsigned long)totBytes);
    return 0;
  }

  key->vers = k->vers;
  key->type = k->type;
  key->alg = k->alg;
  key->size = ntohs(k->size);
  {
    /*
     *  bug: necessary because k is packed.  We'll have to get rid of #pragma
     *  pack for portability reasons, anyway.
     */
    ZkClock c;
    memcpy(&c, &k->expires, sizeof c);
    key->expires = zkClockFromNbo(c);
  }
  fr_EntFromPacked(&key->ent, k->ent);
  key->freeMask = k->mask;
  key->saveMask = k->mask;

  need = sizeof *k;
  b += sizeof *k;

  /*
   *  Initialize paramter pointers to 0, in case we must find those that have
   *  been allocated (i.e. cleaning up on an error).
   */
  
  for (i = 0; i < NUM_PARMS; i++) { key->parm[i] = 0; }

  for (i = 0; i < NUM_PARMS; i++) {
    if (k->mask & (1 << i)) {
      size_t bnsz;
      
      need += 2; /* bignum byte count */
      if (need > *bsz) {
        zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                      "need (%lu) more bytes than buffer size (%lu)",
                      (unsigned long)need, (unsigned long)*bsz);
        goto done;
      }
      { uint16_t x; memcpy(&x, b, 2); bnsz = ntohs(x); b += 2; }

      need += bnsz;
      if (need > *bsz) {
        zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                      "need (%lu) more bytes than buffer size (%lu)",
                      (unsigned long)need, (unsigned long)*bsz);
        goto done;
      }
      if ( ! (key->parm[i] = BN_bin2bn((unsigned char *)b, bnsz, 0))) {
        zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
                      "can't create %hu byte bignum at position %d", bnsz, i);
        goto done;
      }
      b += bnsz;
    }
  }

  /*
   *  Final sanity check on the data.  Is the amount of data taken from the
   *  buffer equal to what was stored?
   */

  if (need != totBytes) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
                  "bytes used (%lu) doesn't equal expected number (%lu)",
                  (unsigned long)need, (unsigned long)totBytes);
    goto done;
  }

  *bsz = need;

  ret = 1;

 done:
  
  if ( ! ret) {
    for (i = 0; i < NUM_PARMS; i++) {
      if (key->parm[i]) {
        BN_clear_free(key->parm[i]);
        key->parm[i] = 0;
      }
    }
  }

  return ret;
}


DLLExport int frCryptPubKeyWriteToBuf(ZkError err, FrCryptPubKey key, char *b, size_t *bsz)
{
  int i;
  size_t totBytes;
  struct fr_crypt_key_pk *k = (void *)b;
  uint8_t realMask;
  
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeyWriteToBuf", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! b) {
    zkErrorSet(err, "frCryptPubKeyWriteToBuf", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "b is null");
    return 0;
  }
  if ( ! bsz) {
    zkErrorSet(err, "frCryptPubKeyWriteToBuf", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "bsz is null");
    return 0;
  }

  realMask = parmMask(key) & key->saveMask;
  totBytes = bytesRequired(key);

  if (*bsz < totBytes) {
    zkErrorSet_va(err, "frCryptPubKeyWriteToBuf", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "*bsz (%lu) is less than required (%lu)",
                  (unsigned long)*bsz, (unsigned long)totBytes);
    return 0;
  }

  k->vers = key->vers;
  k->type = key->type;
  k->alg = key->alg;
  k->size = htons(key->size);
  k->expires = zkClockToNbo(key->expires);
  fr_EntToPacked(&key->ent, k->ent);
  k->mask = realMask;
  k->totBytes = htons(totBytes);

  b += sizeof *k;

  /*
   *  For each key parameter that is set, append it to the buffer, preceded by
   *  its length.
   */
  
  for (i = 0; i < NUM_PARMS; i++) {
    if (realMask & (1 << i)) {
      uint16_t n = BN_num_bytes(key->parm[i]);
      { uint16_t x = htons(n); memcpy(b, &x, 2); } b += 2;
      (void)BN_bn2bin(key->parm[i], (unsigned char *)b); b += n;
    }
  }

  *bsz = totBytes;

  return 1;
}


/*
 *  Write a compact, portable representation of `key' to an allocated buffer.
 */
DLLExport int frCryptPubKeyWriteToMem(ZkError err, FrCryptPubKey key, void **mem, size_t *msz)
{
  int ret = 0;
  size_t n;
  void *m = 0;

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeyWriteToMem", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! mem) {
    zkErrorSet(err, "frCryptPubKeyWriteToMem", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "mem is null");
    return 0;
  }
  if ( ! msz) {
    zkErrorSet(err, "frCryptPubKeyWriteToMem", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "msz is null");
    return 0;
  }

  n = bytesRequired(key);

  if ( ! (m = malloc(n))) {
    zkErrorSet_va(err, "frCryptPubKeyWriteToMem", ZKERR, ZKMFRCRYPT, errno,
                  "can't allocate %lu bytes for key: %s",
                  (unsigned long)n, strerror(errno));
    goto done;
  }

  if ( ! frCryptPubKeyWriteToBuf(err, key, m, &n)) {
    goto done;
  }

  *mem = m; m = 0; /* the caller now owns it */
  *msz = n;

  ret = 1;

 done:

  if (m) { memset(m, 0, n); free(m); }
  
  return ret;
}


/*
 *  Return 1 if both entities are equal.  Two uninitialized entities are
 *  considered equal.
 */
static int equalEntities(ZkError err, FrCryptPubKey k0, FrCryptPubKey k1)
{
  int equal;

  return frEntEqualAllowUndef(err, &k0->ent, &k1->ent, &equal) && equal;
}


/*
 *  Return 1 if both keys have the same parameter positions set, and the
 *  bignums at those positions are equal.  Otherwise return 0.
 *  
 *  Only consider those parameters set in `mask'.
 */
static int equalBignums(FrCryptPubKey k0, FrCryptPubKey k1, uint8_t mask)
{
  int i;
  
  for (i = 0; i < NUM_PARMS; i++) {
    if (mask & (1 << i)) {
      if ( !! k0->parm[i] ^ !! k1->parm[i]) {
        return 0;
      }
      if (k0->parm[i] && k1->parm[i] &&
          BN_cmp(k0->parm[i], k1->parm[i]) != 0) {
        return 0;
      }
    }
  }

  return 1;
}


/*
 *  Get the version of the public key stored in `kbuf'.
 */
DLLExport int frCryptPubKeyBufGetVersion(ZkError err, void *kbuf, size_t ksz, unsigned int *vers)
{
  struct fr_crypt_key_pk *pk = kbuf;
  
  zkErrorClear(err);

  if ( ! kbuf) {
    zkErrorSet(err, "frCryptPubKeyBufGetVersion", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if (ksz < sizeof *pk) {
    zkErrorSet_va(err, "frCryptPubKeyBufGetVersion", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "ksz (%u) is too small", ksz);
    return 0;
  }
  if ( ! vers) {
    zkErrorSet(err, "frCryptPubKeyBufGetVersion", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "vers is null");
    return 0;
  }

  *vers = pk->vers;

  return 1;
}


/*
 *  Get the type of the public key stored in `kbuf'.
 */
DLLExport int frCryptPubKeyBufGetType(ZkError err, void *kbuf, size_t ksz, FrPubKeyType *type)
{
  struct fr_crypt_key_pk *pk = kbuf;
  
  zkErrorClear(err);

  if ( ! kbuf) {
    zkErrorSet(err, "frCryptPubKeyBufGetType", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if (ksz < sizeof *pk) {
    zkErrorSet_va(err, "frCryptPubKeyBufGetType", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "ksz (%u) is too small", ksz);
    return 0;
  }
  if ( ! type) {
    zkErrorSet(err, "frCryptPubKeyBufGetType", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "type is null");
    return 0;
  }

  *type = pk->type;

  return 1;
}


/*
 *  Get the type of the public key stored in `kbuf'.
 */
DLLExport int frCryptPubKeyBufGetExpiration(ZkError err, void *kbuf, size_t ksz, ZkClock *exp)
{
  struct fr_crypt_key_pk *pk = kbuf;
  
  zkErrorClear(err);

  if ( ! kbuf) {
    zkErrorSet(err, "frCryptPubKeyBufGetExpiration", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if (ksz < sizeof *pk) {
    zkErrorSet_va(err, "frCryptPubKeyBufGetExpiration", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "ksz (%u) is too small", ksz);
    return 0;
  }
  if ( ! exp) {
    zkErrorSet(err, "frCryptPubKeyBufGetExpiration", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "exp is null");
    return 0;
  }

  *exp = zkClockFromNbo(pk->expires);

  return 1;
}


/*
 *  Get the version of the public key stored in `kbuf'.
 *  
 *  The entity need not be valid for this routine to succeed.
 */
DLLExport int frCryptPubKeyBufGetEntity(ZkError err, void *kbuf, size_t ksz, FrEntIdExp *ent)
{
  struct fr_crypt_key_pk *pk = kbuf;
  
  zkErrorClear(err);

  if ( ! kbuf) {
    zkErrorSet(err, "frCryptPubKeyBufGetEntity", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if (ksz < sizeof *pk) {
    zkErrorSet_va(err, "frCryptPubKeyBufGetEntity", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "ksz (%u) is too small", ksz);
    return 0;
  }
  if ( ! ent) {
    zkErrorSet(err, "frCryptPubKeyBufGetEntity", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "ent is null");
    return 0;
  }

  fr_EntFromPacked(ent, pk->ent);

  return 1;
}


/*
 *  Set `*equal' to 0 if the two keys are different; set it to 1 if they are
 *  equal.
 *  
 *  Two keys are considered to be equal if they have the same version, type,
 *  algorithm, size and expiry time, as well as having parameters that are
 *  equal.  Parameters are equal if they point to identical bignums, not
 *  necessarily the "same" bignums.
 *  
 *  The free and save masks are not considered, as they may change when a key
 *  is stored and loaded.
 */
DLLExport int frCryptPubKeyEqual(ZkError err, FrCryptPubKey k0, FrCryptPubKey k1, int *equal)
{
  uint8_t mask = -1; /* all bits set */

  zkErrorClear(err);

  if ( ! k0) {
    zkErrorSet(err, "frCryptPubKeyEqual", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "k0 is null");
    return 0;
  }
  if ( ! k1) {
    zkErrorSet(err, "frCryptPubKeyEqual", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "k1 is null");
    return 0;
  }
  if ( ! equal) {
    zkErrorSet(err, "frCryptPubKeyEqual", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "equal is null");
    return 0;
  }

  *equal = (k0->vers == k1->vers &&
            k0->type == k1->type &&
            k0->alg == k1->alg &&
            k0->size == k1->size &&
            zkClockCmp(k0->expires, k1->expires) == 0 &&
            equalEntities(err, k0, k1) &&
            equalBignums(k0, k1, mask));

  return 1;
}


DLLExport int frCryptPubKeyMaskEqual(ZkError err, FrCryptPubKey k0, FrCryptPubKey k1, uint8_t mask, int *equal)
{
  zkErrorClear(err);

  if ( ! k0) {
    zkErrorSet(err, "frCryptPubKeyEqual", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "k0 is null");
    return 0;
  }
  if ( ! k1) {
    zkErrorSet(err, "frCryptPubKeyEqual", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "k1 is null");
    return 0;
  }
  if ( ! equal) {
    zkErrorSet(err, "frCryptPubKeyEqual", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "equal is null");
    return 0;
  }

  *equal = (k0->vers == k1->vers &&
            k0->type == k1->type &&
            k0->alg == k1->alg &&
            k0->size == k1->size &&
            zkClockCmp(k0->expires, k1->expires) == 0 &&
            equalEntities(err, k0, k1) &&
            equalBignums(k0, k1, mask));

  return 1;
}


/* bug: use Philippe's ISO routines. */
static int strFromTime(char *buf, size_t bsz, ZkClock t)
{
  return zkClockStrftime(buf, bsz, "%B %d %Y (local)", t) != 0;
}


/*
 *  Print a public key structure.
 *  
 *  This routine is mainly for debugging.
 */
DLLExport int frCryptPubKeyPrint(ZkError err, FrCryptPubKey key, FILE *fp)
{
  const char *fn = "frCryptPubKeyPrint";
  int i;

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "key is null");
    return 0;
  }
  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "fp is null");
    return 0;
  }

  fprintf(fp, "Public Key:\n");
  fprintf(fp, "    version: %u\n", (unsigned int)key->vers);
  fprintf(fp, "    type: %s (%u)\n",
          frCryptPubKeyTypeStr(key->type), (unsigned int)key->type);
  fprintf(fp, "    alg:  %s (%u)\n",
          frCryptPubKeyAlgStr(key->alg), (unsigned int)key->alg);
  fprintf(fp, "    size: %hu\n",   key->size);
  {
    char tbuf[256];
    (void)strFromTime(tbuf, sizeof tbuf, key->expires);
    fprintf(fp, "    expires: %s (" LONG_LONG_FTM ")\n",
            tbuf, (LONG_LONG) zkClockToSeconds(key->expires));
  }
  fprintf(fp, "    owner: ");
  frEntPrint(err, fp, &key->ent);
  fprintf(fp, "\n");
  fprintf(fp, "    free: 0x%02x\n", key->freeMask);
  fprintf(fp, "    save: 0x%02x\n", key->saveMask);

  for (i = 0; i < NUM_PARMS; i++) {
    if (key->parm[i]) {
      fprintf(fp, "parm %d: ", i);
      if ( ! BN_print_fp(fp, key->parm[i])) {
        return 0;
      }
      fputc('\n', fp);
    }
  }

  return 1;
}  


DLLExport const char *frCryptPubKeyTypeStr(FrPubKeyType type)
{
  const char *tstr;

  switch (type) {
  case FRCRYPT_KEYTYPE_CRYPTION: tstr = "cryption"; break;
  case FRCRYPT_KEYTYPE_SIGNING:  tstr = "signing";  break;
  default:                       tstr = "???";      break;
  }

  return tstr;
}


DLLExport const char *frCryptPubKeyAlgStr(FrCryptPubAlg alg)
{
  const char *astr;

  switch (alg) {
  case FRCRYPT_PUBALG_DSA:     astr = "dsa";     break;
  case FRCRYPT_PUBALG_ELGAMAL: astr = "elgamal"; break;
  default:                     astr = "???";     break;
  }

  return astr;
}


/*
 *  Print identifying information, about the key, to `s'.
 */
DLLExport char *frCryptPubKeyBufIdSnprint(ZkError err, void *k, size_t ksz, char *s, size_t ssz)
{
  char entbuf[128];
  const char *fn = "frCryptPubKeyBufIdSnprint";
  char *ret = 0;
  struct fr_crypt_key_pk *kbuf = k;
  
  zkErrorClear(err);

  if ( ! k) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "k is null");
    goto done;
  }
  if (ksz < sizeof *kbuf) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "ksz is too small (%lu < %lu)",
                  (unsigned long)ksz, (unsigned long)sizeof *kbuf);
    goto done;
  }
  if ( ! s) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "s is null");
    goto done;
  }
  if (ssz == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "ssz is 0");
    goto done;
  }

  if ( ! frEntBufSnprint(err, kbuf->ent, sizeof kbuf->ent, entbuf, sizeof entbuf)) {
    goto done;
  }

  snprintf(s, ssz, "[%u, %s (%u), %s]",
           kbuf->vers, frCryptPubKeyTypeStr(kbuf->type), kbuf->type, entbuf);

  ret = s;

 done:

  return ret;
}


/*
 *  Print a short summary of the information contained in a public key
 *  structure.
 */
DLLExport int frCryptPubKeyShortPrint(ZkError err, FrCryptPubKey key, int indent, FILE *fp)
{
  char tbuf[256];
  const char *fn = "frCryptPubKeyShortPrint";
  const char *tstr;

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "key is null");
    return 0;
  }
  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "fp is null");
    return 0;
  }

  switch (key->type) {
  case FRCRYPT_KEYTYPE_CRYPTION: tstr = "cryption"; break;
  case FRCRYPT_KEYTYPE_SIGNING:  tstr = "signing";  break;
  default:                       tstr = "???";      break;
  }

  fprintf(fp, "%*sOwner: ", indent, "");
  (void)frEntPrint(err, fp, &key->ent);
  fprintf(fp, "\n");
  (void)strFromTime(tbuf, sizeof tbuf, key->expires);
  fprintf(fp, "%*sVersion: %d      Type: %-8s      Size: %hu bits\n",
          indent, "", key->vers, tstr, key->size);
  
  fprintf(fp, "%*sExpires: %s (" LONG_LONG_FTM ")\n",
          indent, "", tbuf, (LONG_LONG) zkClockToSeconds(key->expires));

  return 1;
}  


DLLExport int frCryptPubKeyToCString(ZkError err, FrCryptPubKey key, char *aname, uint8_t mask, char **str)
{
  char *kstr = 0, *rstr = 0;
  int ret = 0;
  size_t kbsz, kssz;
  uint8_t omask;
  void *kbuf = 0;

  if ( ! key) {
    zkErrorSet(err, "frCryptPubKeyToCString", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! aname) {
    zkErrorSet(err, "frCryptPubKeyToCString", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "aname is null");
    return 0;
  }
  if ( ! str) {
    zkErrorSet(err, "frCryptPubKeyToCString", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "str is null");
    return 0;
  }

  if ( ! (frCryptPubKeyGetSaveMask(err, key, &omask) &&
          frCryptPubKeySetSaveMask(err, key, mask) &&
          (kbsz = bytesRequired(key)) &&
          zkMemAlloc(err, &kbuf, kbsz) &&
          frCryptPubKeyWriteToBuf(err, key, kbuf, &kbsz) &&
          frCryptPubKeySetSaveMask(err, key, omask) &&
          zkCodeUCharArrayFromData(err, kbuf, kbsz, &kstr, &kssz) &&
          zkCStrSprintf(&rstr, "unsigned char %s[] = %s;\n", aname, kstr) != -1)) {
    goto done;
  }

  *str = rstr; rstr = 0; /* the caller now owns it */

  ret = 1;
  
 done:

  if (kstr) { free(kstr); }
  if (kbuf) { free(kbuf); }
  if (rstr) { free(rstr); }

  return ret;
}


/*
 *  Print, to `fp', a public key object as a C structure.
 *  
 *  This routine is for generating hardwired keys.
 */
DLLExport int frCryptPubKeyPrintAsC(ZkError err, FrCryptPubKey key, uint8_t mask, char *name, FILE *fp)
{
  const char *fn = "frCryptPubKeyPrintAsC";
  char *str = 0;
  int ret = 0;

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "key is null");
    return 0;
  }
  if ( ! name) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "name is null");
    return 0;
  }
  if (strlen(name) > 32) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "name is too long (%lu > %d)",
                  (unsigned long)strlen(name), 32 /* bug */);
    return 0;
  }
  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "fp is null");
    return 0;
  }

  if ( ! frCryptPubKeyToCString(err, key, name, mask, &str)) {
    goto done;
  }

  if (fputs(str, fp) == EOF) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, errno,
                  "error writing string: %s", strerror(errno));
    goto done;
  }

  ret = 1;
  
 done:

  if (str) { free(str); }
  
  return ret;
}  


/*
 *  Set the shared parameters according to the key type.
 *  
 *  It's unfortunate that it needs to know the various key types the exist...
 */
DLLExport int frCryptPubKeySetSharedParms(ZkError err, FrCryptPubKey key)
{
  const char *fn = "frCryptPubKeySetSharedParms";
  int ret = 0;

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "key is null");
    return 0;
  }

  switch (key->type) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    ret = frCryptCryptMakeCrypt(err, key);
    break;

  case FRCRYPT_KEYTYPE_SIGNING:
    ret = frCryptSigMakeSig(err, key);
    break;

  default:
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
                  "unknown key type (%u)", key->type);
    break;
  }

  return ret;
}


/*
 *  Return 1 if `key' appears to be valid.
 */
DLLExport int frCryptPubKeyPublicValidate(ZkError err, FrCryptPubKey key)
{
  const char *fn = "frCryptPubKeyPublicValidate";
  int ret = 0;

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "key is null");
    return 0;
  }

  /*
   *  Check some shared parameters.
   */

  if (key->vers > maxVersion) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                  "key version too large (%u > %u)", key->vers, maxVersion);
    goto done;
  }

  switch (key->alg) {
  case FRCRYPT_PUBALG_DSA:
  case FRCRYPT_PUBALG_ELGAMAL:
    break;

  default:
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                  "unknown key algorithm (%u)", key->alg);
    goto done;
  }
  
  /*
   *  Expiration dates shouldn't be too far in the past.
   */

  if (zkClockCmp(key->expires, minExpiration) < 0) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                  "expiration date is too early ("
                  LONG_LONG_FTM " < " LONG_LONG_FTM ")",
                  (LONG_LONG) zkClockToSeconds(key->expires), 
                  (LONG_LONG) zkClockToSeconds(minExpiration));
    goto done;
  }

  /*
   *  The entity should be valid.
   */

  {
    int valid;
    
    if ( ! frEntIsValid(err, &key->ent, &valid)) {
      goto done;
    }
    if ( ! valid) {
      zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                 "invalid key owner entity");
      /* bug: print entity */
      goto done;
    }
  }

  switch (key->type) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    ret = frCryptCryptPublicValidate(err, key);
    break;

  case FRCRYPT_KEYTYPE_SIGNING:
    ret = frCryptSigPublicValidate(err, key);
    break;

  default:
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                  "unknown key type (%u)", key->type);
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 *  Return 1 if a valid key is contained in `buf'.  Among other properties it
 *  must possess, the amount of space occupied by the key must be _exactly_
 *  equal to `bsz'.
 *  
 *  The impetus for writing this routine is to track down the bug where short
 *  keys are appearing in the key database.
 */
DLLExport int frCryptPubKeyBufPublicValidate(ZkError err, void *buf, size_t bsz)
{
  FrCryptPubKey key = 0;
  const char *fn = "frCryptPubKeyBufPublicValidate";
  int ret = 0;
  size_t sz = bsz;

  zkErrorClear(err);

  if ( ! buf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "buf is null");
    return 0;
  }

  /*
   *  Try creating a key from the buffer.
   */

  if ( ! (frCryptPubKeyNew(err, &key) &&
          frCryptPubKeyReadFromBuf(err, key, buf, &sz))) {
    goto done;
  }
  
  if (sz != bsz) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                  "inconsistent key buffer sizes (%lu != %lu)",
                  (unsigned long)sz, (unsigned long)bsz);
    goto done;
  }

  if ( ! frCryptPubKeyPublicValidate(err, key)) {
    goto done;
  }

  ret = 1;

 done:

  if (key) { frCryptPubKeyFree(err, &key); }

  return ret;
}

DLLExport int frCryptPubKeySetBlinding(ZkError err, FrCryptPubKey key, int bool)
{
  const char *fn = "frCryptPubKeySetBlinding";

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "key is null");
    return 0;
  }

  if (bool == 1) {
    key->blinding_flag = 1;
    /* if a key->bl already exists, free it before creating a new one */
    if (key->bl) {
      if ( ! fr_CryptBlindingFree(err, &(key->bl)) ) {
        return 0;
      }
    }
    if ( ! fr_CryptBlindingNew(err, &(key->bl))) {
      return 0;
    }
  }
  else if (bool == 0) {
    key->blinding_flag = 0;
    /* free the blinding strucutre if it existed */
    if (key->bl) {
      if ( ! fr_CryptBlindingFree(err, &(key->bl)) ) {
        return 0;
      }
    }
  }
  else {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD, "wrong bool input: %d", bool);
    return 0;
  }

  return 1;

}

/* These two functions are used to set a flag indicating that a key can't be
   used for creating routes.  They overload the use of the expires field.
   Yuck!  */

DLLExport int frCryptPubKeySetCanCreateRoutes(ZkError err, FrCryptPubKey key, int canCreateRoutes)
{
  const char *fn = "frCryptPubKeySetCanCreateRoutes";

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "key is null");
    return 0;
  }

  if (canCreateRoutes) {
    key->expires = zkClockFromSeconds(0);
  } else {
    key->expires = CantCreateRoutesMagicExpiration;
  }

  return 1;
}

DLLExport int frCryptPubKeyGetCanCreateRoutes(ZkError err, FrCryptPubKey key, int * canCreateRoutes)
{
  const char *fn = "frCryptPubKeyGetCanCreateRoutes";

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "key is null");
    return 0;
  }
  if ( ! canCreateRoutes) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "canCreateRoutes is null");
    return 0;
  }

  if (zkClockCmp(key->expires, CantCreateRoutesMagicExpiration) == 0) {
    *canCreateRoutes = 0;
  } else {
    *canCreateRoutes = 1;
  }

  return 1;
}
