/*
 *  $Id: pubkey.c,v 1.7.10.1 2000/10/11 13:17:47 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#ifndef WIN32
#include <netinet/in.h> /* ntohs() */
#else
#include <winsock2.h>
#endif

#include <limits.h>
#include "pubkey.h"
#include "openssl/bn.h"


#pragma pack(1)

struct fr_pubkey_info_p {
  uint8_t vers; /* key version */
  uint8_t type; /* Key type: cryption, signing. */
  uint8_t alg;  /* Algorithm: DSA, ElGamal. */
  uint8_t mask; /* bitmask indicating which of 8 possible bignums are stored */
  uint8_t keySize; /* indication of key size */
};

#pragma pack()


DLLExport int fr_CryptPubKeyLoad(ZkError err, struct fr_pubkey_info *pki, void *data, size_t *dsz)
{
  unsigned char *d = data;
  size_t i, need = 0; /* space needed to store key */
  struct fr_pubkey_info_p pkip;

  zkErrorClear(err);

  if ( ! pki) {
    zkErrorSet(err, "fr_CryptPubKeyLoad", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "pki is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "fr_CryptPubKeyLoad", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, "fr_CryptPubKeyLoad", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "dsz is null");
    return 0;
  }

  need += sizeof pkip;

  if (*dsz < need) {
    zkErrorSet_va(err, "fr_CryptPubKeyLoad", ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                  "*dsz is too short (%lu < %lu)",
                  (unsigned long)*dsz, (unsigned long)need);
    return 0;
  }

  memcpy(&pkip, d, sizeof pkip); d += sizeof pkip;

  pki->vers = pkip.vers;
  pki->type = pkip.type;
  pki->alg = pkip.alg;
  pki->mask = pkip.mask;
  pki->keySize = pkip.keySize;

  for (i = 0; i < NUM_BN_MAX; i++) {
    uint16_t bnsz;
    
    /*
     *  If the i'th bit is set, set the i'th bignum.
     */
    
    if (pki->mask & (1 << i)) {
      need += 2; /* need two bytes for bignum size */
      if (need > *dsz) {
        zkErrorSet_va(err, "fr_CryptPubKeyLoad", ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                      "*dsz is too short (%lu < %lu) (i is %lu)",
                      (unsigned long)*dsz, (unsigned long)need, (unsigned long)i);
        return 0; /* bug: leak */
      }
      /* bug? put INT16_FROM_NBO() in a common include file? */
      memcpy(&bnsz, d, 2); bnsz = ntohs(bnsz); d += 2;

      need += bnsz;
      if (need > *dsz) {
        zkErrorSet_va(err, "fr_CryptPubKeyLoad", ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                      "*dsz is too short (%lu < %lu) (i is %lu)",
                      (unsigned long)*dsz, (unsigned long)need, (unsigned long)i);
        return 0; /* bug: leak */
      }
      if ( ! (pki->bn[i] = BN_bin2bn(d, bnsz, 0))) {
        zkErrorSet(err, "fr_CryptPubKeyLoad", ZKERR, ZKMFRCRYPT, ZKERROR,
                   "error from BN_bin2bn()");
        return 0; /* bug: leak */
      }
      d += bnsz;
    }
  }

  *dsz = need;

  return 1;
}

                                  
DLLExport int fr_CryptPubKeyStore(ZkError err, struct fr_pubkey_info *pki, void *data, size_t *dsz)
{
  unsigned char *d = data;
  size_t i, need = 0; /* space needed to store key */
  struct fr_pubkey_info_p pkip;
  
  zkErrorClear(err);

  if ( ! pki) {
    zkErrorSet(err, "fr_CryptPubKeyStore", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "pki is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "fr_CryptPubKeyStore", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, "fr_CryptPubKeyStore", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "dsz is null");
    return 0;
  }

  /* Sum up the space we'll need. */

  need += sizeof pkip;
  for (i = 0; i < NUM_BN_MAX; i++) {
    if (pki->mask & (1 << i)) {
      need += 2 + BN_num_bytes(pki->bn[i]);
    }
  }

  if (need > *dsz) {
    zkErrorSet_va(err, "fr_CryptPubKeyStore", ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                  "*dsz is too small (%lu < %lu)",
                  (unsigned long)*dsz, (unsigned long)need);
    return 0;
  }

  pkip.vers = pki->vers;
  pkip.type = pki->type;
  pkip.alg = pki->alg;
  pkip.mask = pki->mask;
  pkip.keySize = pki->keySize;

  memcpy(d, &pkip, sizeof pkip); d += sizeof pkip;
  for (i = 0; i < NUM_BN_MAX; i++) {
    if (pki->mask & (1 << i)) {
      uint16_t bnsz = BN_num_bytes(pki->bn[i]);
      /* bug? put NBO_FROM_INT16() in a common include file? */
      { uint16_t x = htons(bnsz); memcpy(d, &x, 2); d += 2; }
      (void)BN_bn2bin(pki->bn[i], d); d += bnsz;
    }
  }
  
  *dsz = need;

  return 1;
}
