/*
 *  $Id: randaccum.c,v 1.6.10.1 2000/10/11 13:17:47 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#define DEBUG_LG 0


/*
 *  We store our random data in SSLeay's entropy pool.  We merely keep track
 *  of how much we've added.
 *  
 *  This means that we should not extract any bits from it while we're in the
 *  process of seeding it.
 */

#include "zks.h"

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fr_gencrypt.h"
#include "fr_accum.h"
#include "openssl/rand.h"
#include "zkDefs.h"


/*
 *  As a sanity check, only keep the 12 least significant bits of the smallest
 *  delta.
 */
#define MASK ((1 << 12) - 1)

/*
 *  The number of elements in an array.
 */
#define NELTS(x) (sizeof(x) / sizeof((x)[0]))

/*
 *  Return the positive difference between `a' and `b'.
 */
#define POSDIFF(a, b) ((a) < (b) ? (b) - (a) : (a) - (b))


/*
 *  The `oDelta' values are the n'th order differences between successive
 *  values from the source.  The zero'th element is the last actual value, not
 *  a difference.
 *  
 *  `nbits' contains an estimate of the number of bits of entropy associated
 *  with the given source.
 */
struct fr_randaccum {
  int count;
  unsigned int nbits;
  unsigned int oDelta[4];
#if DEBUG_PGP_TOO
  unsigned int pgphist[3];
#endif
};


#if DEBUG_PGP_TOO

/*
 *  From PGP sources, for comparison purposes only.
 */
static unsigned int randEstimate(unsigned int x, unsigned int *history, unsigned h)
{
  unsigned int t, min = x;

  while (h--) {
    t = history[h];		 	/* Last delta */
    t = (x > t) ? x - t : t - x;	 /* |x - history[h]| */
    history[h] = x;
    x = t;
    if (min > x)
      min = x;
	}
  return min;
}

#endif


/*
 *  Return floor(log_2(n)).  If `n' is 0 return 0.
 */
static int lg(unsigned int n)
{
  int l = 0;
#if DEBUG_LG
  unsigned int on = n;
#endif
  
  do {
    n >>= 1; l++;
  } while (n);

#if DEBUG_LG
  fprintf(stdout, "lg(%u) = %u\n", on, l-1);
#endif

  return l - 1;
}


/*
 *
 *
 *                            External routines
 *
 *
 */


/*
 *  Assuming `a' is valid, we directly return the nubmer of bits that have
 *  been accumulated through it.
 */
DLLExport size_t fr_CryptRandAccumNumBits(FrCryptRandAccum a)
{
  return a->nbits;
}


DLLExport int frCryptRandAccumNew(ZkError err, FrCryptRandAccum *a)
{
  FrCryptRandAccum ra;
  
  zkErrorClear(err);

  if ( ! a) {
    zkErrorSet(err, "frCryptRandAccumNew", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "a is null");
    return 0;
  }

  if ((ra = malloc(sizeof *ra))) {
    memset(ra, 0, sizeof *ra);
    *a = ra;
  }

  return !! ra;
}


DLLExport int frCryptRandAccumFree(ZkError err, FrCryptRandAccum *a)
{
  zkErrorClear(err);

  if ( ! a) {
    zkErrorSet(err, "frCryptRandAccumFree", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "a is null");
    return 0;
  }
  if ( ! *a) {
    zkErrorSet(err, "frCryptRandAccumFree", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "*a is null");
    return 0;
  }

  memset(*a, 0, sizeof **a);
  free(*a); *a = 0;

  return 1;
}


/*
 *  Treat the contents of `val' as one datum in a sequence of random data,
 *  adding its contents to the SSLeay entropy pool.  Based on previous values,
 *  estimate how much actual randomness is being contributed.
 *  
 *  Initially, we want to build up some deltas in order to estimate the number
 *  of bits added, so we don't increase the estimated number of bits of
 *  entropy until these deltas have values.
 */
int frCryptRandAccumAdd(ZkError err, FrCryptRandAccum a, unsigned int val)
{
  size_t i;
  static size_t count = NELTS(a->oDelta);
  unsigned int delta[NELTS(a->oDelta) + 1], min;
#if DEBUG_PGP_TOO
  unsigned int pgpmin;
#endif

  zkErrorClear(err);

  RAND_seed((unsigned char *)&val, sizeof val);

  delta[0] = val;
  for (i = 0; i < NELTS(a->oDelta); i++) {
    delta[i+1] = POSDIFF(delta[i], a->oDelta[i]);
    a->oDelta[i] = delta[i];
  }

#if DEBUG_PGP_TOO
  pgpmin = randEstimate(val, &a->pgphist[0], NELTS(a->pgphist));
#endif

  if (count > 0) {
    /* All the oDeltas have not yet been initialized. */
    count--;
    return 1;
  }
  
  min = delta[1];
  for (i = 1; i < NELTS(delta); i++) {
    min = ZK_MIN(min, delta[i]);
  }
    
#if DEBUG_PGP_TOO
  fprintf(stdout, "Me: %u, PGP: %u  (%.3f)\n", min, pgpmin, (float)min/(float)pgpmin);
#endif

  /*
   *  bug? the shift may be unnecessary since lg() essentially does a floor on
   *  the result.  (But, it won't hurt.)
   */

  a->nbits += lg((min >> 1) & MASK);

  return 1;
}


/*
 *  Set `*n' to the number of random bits that have been accumulated so far.
 */
DLLExport int frCryptRandAccumNumBits(ZkError err, FrCryptRandAccum a, size_t *n)
{
  zkErrorClear(err);

  if ( ! a) {
    zkErrorSet(err, "frCryptRandAccumNumBits", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "a is null");
    return 0;
  }
  if ( ! n) {
    zkErrorSet(err, "frCryptRandAccumNumBits", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "n is null");
    return 0;
  }

  *n = a->nbits;
  
  return 1;
}
