/*
 *  $Id: sig.c,v 1.45.8.3 2000/12/14 17:52:41 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#define _GNU_SOURCE /* for snprintf() */

#define EXPORTMASTERKEY

#define DEBUG_ZKM_PARMS 0
#define DEBUG_VALID_PQG 0
#define DEBUG_SIG 0
#define DEBUG_LDST 0

#if DEBUG_VALID_PQG && ! defined(WIN32)
#include <assert.h>
#endif

#ifndef WIN32
#include <arpa/inet.h>
#else
#include <winsock2.h>
#endif

#include <errno.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "frBn.h"
#include "fr_gencrypt.h"

#include "fr_sig.h"

#include "pk.h"
#include "openssl/crypto.h"
#include "openssl/dsa.h"
#include "pubkey.h"
#include "openssl/sha.h"
#include "sig.h"
#include "_zksupdp.h"
#include "dsaparms.h"
#include "_zkmp.h"

/*
 *  Convenience macros defining the positions, within the FrCryptPubKey
 *  object, of our parameters.  These positions correspond to the bits defined
 *  by the FRCRYPT_SIGPARM_* macros (in frCrypt.h), so that we can directly
 *  use those bitmasks with the frCryptPubKey* routines.
 */

#define PUB  0
#define PRIV 1
#define P    2
#define Q    3
#define G    4

/*
 *  Besides computing the signature over the data provided by the caller, we
 *  usually include the tail of the packed signature structure.  These macros
 *  define the offset and length of the extra data to be signed.
 */

#define SIG_INIT_OFF (offsetof(struct fr_sig_pk, keyVersion))
#define SIG_INIT_LEN (sizeof(struct fr_sig_pk) - SIG_INIT_OFF)

/*
 *  We allow 29 bytes to store each of the r and s components of a DSA
 *  signature.  This value is the size of our largest q value (for 2048 bit
 *  keys).
 */

#define RS_LEN 29


#pragma pack(1)

struct fr_sig_pk {
  char r[RS_LEN];
  char s[RS_LEN];
  uint8_t keyVersion;
  uint16_t keySize; /* in bits */
  uint32_t dataSize; /* amount of signed data */
};

#pragma pack()


/*
 *  DSA shared parameters for each key size.  Currently, we support key sizes
 *  only up to 2048 bits.
 */

#define DSA_SHARED(n)     \
{ FRCRYPT_SIGKEYSIZE_##n, \
  fr_BnData_p_##n,        \
  sizeof fr_BnData_p_##n, \
  fr_BnData_q_##n,        \
  sizeof fr_BnData_q_##n, \
  fr_BnData_g_##n,        \
  sizeof fr_BnData_g_##n, \
  0, 0, 0 }

static struct dsaSharedParm {
  FrCryptSigKeySize nbits;

  const void *p_data;
  size_t p_size;

  const void *q_data;
  size_t q_size;

  const void *g_data;
  size_t g_size;

  BIGNUM *p;
  BIGNUM *q;
  BIGNUM *g;
} dsaSharedParm[] = {
  DSA_SHARED(1024),
  DSA_SHARED(1536),
  DSA_SHARED(2048),

#if 0
  DSA_SHARED(3072),
  DSA_SHARED(4096),
#endif
  
  { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }
};

#undef DSA_SHARED


/*
 *  This is the public part of the Zero Knowledge master signature key.
 */
FRCRYPT_API FrCryptSig frCryptSigZkMaster = 0;


/*
 *  This is the public part of the Zero Knowledge software update signature
 *  key.
 */
DLLExport FrCryptSig frCryptSigZkSupd = 0;

/*
 *  The flag indicates whether initialization has been done.
 */
static int sigIsInit = 0;


/*
 *  For now, we expect to be using DSA as our only algorithm.
 */


/*
 *
 *
 *                            Internal routines.
 *
 *
 */

/* some function prototypes are different between versions of SSLeay/OpenSSL */
static int bn_mul(BIGNUM *r, BIGNUM *a, BIGNUM *b, BN_CTX *ctx)
{
#if OPENSSL_VERSION_NUMBER >= 0x0922    /* OpenSSL 0.9.2b */
  return BN_mul(r, a, b, ctx);
#else
  (void)ctx;
  return BN_mul(r, a, b);
#endif
}

/*
 *  Original code from Ian Goldberg.
 */

/*
 *  bug? change BN_free()s to BN_clear_free()s?
 *
 *  Yup.  Fixed.  - Ian
 */

/* Generate a DSA signature on a hash */
static int zks_dsa_sign(ZkError err, FrCryptSig dsakey, unsigned char *hash,
                            unsigned char *r_bin, unsigned char *s_bin)
{
  BIGNUM *kinv = NULL;
  BIGNUM *m = NULL;
  BIGNUM *r = NULL;
  BIGNUM *s = NULL;
  BIGNUM *xr = NULL;
  BN_CTX *ctx = NULL;
  DSA *dsa = NULL;
  const char *fn = "zks_dsa_sign";
  int ret = 0;

  zkErrorClear(err);

  if ( ! dsakey) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "dsakey is null");
    goto done;
  }
  
  if ( ! dsakey->parm[PRIV]) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
               "private parameter is null");
    goto done;
  }

  if ( ! (dsa = DSA_new())) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR, "DSA_new() failed");
    goto done;
  }

  if ( ! (ctx = BN_CTX_new())) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR, "BN_CTX_new() failed");
    goto done;
  }

  dsa->p = BN_dup(dsakey->parm[P]);
  dsa->q = BN_dup(dsakey->parm[Q]);
  dsa->g = BN_dup(dsakey->parm[G]);
  dsa->pub_key = BN_dup(dsakey->parm[PUB]);
  dsa->priv_key = BN_dup(dsakey->parm[PRIV]);

  if ( ! (dsa->p && dsa->q && dsa->g && dsa->pub_key && dsa->priv_key)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
               "couldn't dup dsakey parameters");
    goto done;
  }

  if ( ! DSA_sign_setup(dsa, ctx, &kinv, &r)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
               "DSA_sign_setup() failed");
    goto done;
  }

  m = BN_new();
  xr = BN_new();
  s = BN_new();

  if ( ! m || ! xr || ! s) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
               "BN_new() failed for one of m, xr or s");
    goto done;
  }

  if ( ! (BN_bin2bn(hash, 20, m) &&
          bn_mul(xr, dsa->priv_key, r, ctx) &&
          BN_add(s, xr, m) &&
          BN_mod_mul(s,s,kinv,dsa->q,ctx))) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
               "bignum operation failed");
    goto done;
  }

  if (BN_cmp(r, BN_value_one()) < 0 || BN_cmp(s, BN_value_one()) < 0) {
    /* This error indicates we should try again. */
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "r or s is zero");
    goto done;
  }

  /* Store r and s with leading zeroes. */

  memset(r_bin, 0, RS_LEN);
  memset(s_bin, 0, RS_LEN);
  BN_bn2bin(r, r_bin+RS_LEN-BN_num_bytes(r));
  BN_bn2bin(s, s_bin+RS_LEN-BN_num_bytes(s));

#if DEBUG_SIG
  fprintf(stderr, "Sign\n");
  fprintf(stderr, "p "); BN_print_fp(stderr, dsakey->parm[P]); fprintf(stderr, "\n");
  fprintf(stderr, "q "); BN_print_fp(stderr, dsakey->parm[Q]); fprintf(stderr, "\n");
  fprintf(stderr, "g "); BN_print_fp(stderr, dsakey->parm[g]); fprintf(stderr, "\n");
  fprintf(stderr, "pub "); BN_print_fp(stderr, dsakey->parm[PUB]); fprintf(stderr, "\n");
  fprintf(stderr, "priv "); BN_print_fp(stderr, dsakey->parm[PRIV]); fprintf(stderr, "\n");
  fprintf(stderr, "r "); BN_print_fp(stderr, r); fprintf(stderr, "\n");
  fprintf(stderr, "s "); BN_print_fp(stderr, s); fprintf(stderr, "\n");
#endif

  ret = 1;

 done:
  
  if (kinv) BN_clear_free(kinv);
  if (s) BN_free(s);
  if (r) BN_free(r);
  if (xr) BN_clear_free(xr);
  if (m) BN_clear_free(m);
  if (ctx) BN_CTX_free(ctx);
  if (dsa) DSA_free(dsa);
  
  return ret;
}

/* Verify a DSA signature */
static int zks_dsa_verify(FrCryptSig dsakey, unsigned char *hash,
                              unsigned char *r_bin, unsigned char *s_bin,
                              int *valid)
{
  BIGNUM *p, *q, *g, *y, *H, *r, *s;
  BIGNUM *w = NULL;
  BN_CTX *ctx;
  int ret = 0;

  p = BN_dup(dsakey->parm[P]);
  q = BN_dup(dsakey->parm[Q]);
  g = BN_dup(dsakey->parm[G]);
  y = BN_dup(dsakey->parm[PUB]);

  H = BN_new();
  r = BN_new(); 
  s = BN_new(); 
  ctx = BN_CTX_new();
  if (!p || !q || !g || !y || !H || !r || !s || !ctx) {
    goto zks_dsa_verify_done;
  }

  BN_bin2bn(hash, 20, H);
  BN_bin2bn(r_bin, RS_LEN, r);
  BN_bin2bn(s_bin, RS_LEN, s);
    
#if DEBUG_SIG
  fprintf(stderr, "Verify\n");
  fprintf(stderr, "p "); BN_print_fp(stderr, dsakey->parm[P]); fprintf(stderr, "\n");
  fprintf(stderr, "q "); BN_print_fp(stderr, dsakey->parm[Q]); fprintf(stderr, "\n");
  fprintf(stderr, "g "); BN_print_fp(stderr, dsakey->parm[G]); fprintf(stderr, "\n");
  fprintf(stderr, "pub "); BN_print_fp(stderr, dsakey->parm[PUB]); fprintf(stderr, "\n");
#if 0
  fprintf(stderr, "priv "); BN_print_fp(stderr, dsakey->parm[PRIV]); fprintf(stderr, "\n");
#endif
  fprintf(stderr, "r "); BN_print_fp(stderr, r); fprintf(stderr, "\n");
  fprintf(stderr, "s "); BN_print_fp(stderr, s); fprintf(stderr, "\n");
#endif

#if OPENSSL_VERSION_NUMBER >= 0x0922   /* OpenSSL 0.9.2b */
/* some function prototypes are different between versions of SSLeay/OpenSSL */
  w = BN_mod_inverse(NULL, s, q, ctx);
#else
  w = BN_mod_inverse(s, q, ctx);
#endif
  BN_free(s); s = NULL;

  BN_mod_mul(H, H, w, q, ctx);
  BN_mod_mul(w, w, r, q, ctx);
  BN_mod_exp(g, g, H, p, ctx);
  BN_mod_exp(y, y, w, p, ctx);
  BN_mod_mul(g, g, y, p, ctx);
  BN_mod(w, g, q, ctx);

#if DEBUG_SIG
  fprintf(stderr, "w "); BN_print_fp(stderr, w); fprintf(stderr, "\n");
#endif

#if 1
  *valid = BN_cmp(w, r) == 0;
  ret = 1;
#else
  ret = BN_cmp(w, r) ? 1 : 0;
#endif

 zks_dsa_verify_done:
  if (ctx) BN_CTX_free(ctx);
  if (w) BN_free(w);
  if (s) BN_free(s);
  if (r) BN_free(r);
  if (H) BN_free(H);
  if (y) BN_free(y);
  if (g) BN_free(g);
  if (q) BN_free(q);
  if (p) BN_free(p);
  return ret;
}


static int intToSigKeySize(ZkError err, unsigned int n, FrCryptSigKeySize *ksz)
{
  zkErrorClear(err);

  switch (n) {
  case 1024: *ksz = FRCRYPT_SIGKEYSIZE_1024; break;
  case 1536: *ksz = FRCRYPT_SIGKEYSIZE_1536; break;
  case 2048: *ksz = FRCRYPT_SIGKEYSIZE_2048; break;

  default:
    zkErrorSet_va(err, "intToSigKeySize", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "key size (%u bits) is invalid", n);
    return 0;
  }

  return 1;
}


/*
 *
 *
 *                           Debugging routines.
 *
 *
 */


/*
 *  Print the signature to stderr.
 */
DLLExport int fr_CryptSigPrint(ZkError err, FrCryptSig sig)
{
  const char *fn = "fr_CryptSigPrint";

  zkErrorClear(err);

  if (sig->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", sig->type);
    return 0;
  }

  fr_BnPrint(err, "pub  ", sig->parm[PUB]);
  fr_BnPrint(err, "priv ", sig->parm[PRIV]);
  fr_BnPrint(err, "p    ", sig->parm[P]);
  fr_BnPrint(err, "q    ", sig->parm[Q]);
  fr_BnPrint(err, "g    ", sig->parm[G]);

  return 1;
}


/*
 *  We export this only for the benefit of the test program dsaParms.
 */
DLLExport int fr_CryptGetSigSharedParms(ZkError err, FrCryptSigKeySize ksz, BIGNUM **p, BIGNUM **q, BIGNUM **g)
{
  const char *fn = "fr_CryptGetSigSharedParms";
  size_t i;
  
  zkErrorClear(err);

  if ( ! sigIsInit) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "signature initialization has not been performed");
    return 0;
  }

  for (i = 0; dsaSharedParm[i].nbits > 0; i++) {
    if (ksz == dsaSharedParm[i].nbits) {
      *p = dsaSharedParm[i].p;
      *q = dsaSharedParm[i].q;
      *g = dsaSharedParm[i].g;
      return 1;
    }
  }

  zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                "invalid key size %d", ksz);

  return 0;
}


static int checkSigFields(ZkError err, FrCryptSig key)
{
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "checkSigFields", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
#if 0
  if (key->vers == 0) {
    zkErrorSet(err, "checkSigFields", ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "key version is 0");
    return 0;
  }
  if ( ! verifyKeySize(key->keySize)) {
    zkErrorSet(err, "checkSigFields", ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "invalid key size");
    return 0;
  }
#endif
  if ( ! (key->parm[P] && key->parm[Q] && key->parm[G])) {
    zkErrorSet(err, "checkSigFields", ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "fixed parameters are not all set");
    return 0;
  }

  return 1;
}


/*
 *  This routine exists solely for testing purposes.
 */
DLLExport int fr_CryptSigRandomDsaNew(ZkError err, DSA **d)
{
  DSA *dsa = 0;
  FrCryptSig sig;
  int ret = 0;
  
  zkErrorClear(err);

  if ( ! sigIsInit) {
    zkErrorSet(err, "fr_CryptSigRandomDsaNew", ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "signature initialization has not been performed");
    return 0;
  }

  if ( ! frCryptSigNew(err, &sig)) { goto done; }
  if ( ! frCryptSigSetKeySize(err, sig, FRCRYPT_SIGKEYSIZE_1024)) { goto done; }
  if ( ! frCryptSigSetRandomKeys(err, sig)) { goto done; }

  if ( ! (dsa = DSA_new())) { goto done; }
  if ( ! (dsa->p = BN_dup(sig->parm[P]))) { goto done; }
  if ( ! (dsa->q = BN_dup(sig->parm[Q]))) { goto done; }
  if ( ! (dsa->g = BN_dup(sig->parm[G]))) { goto done; }
  if ( ! (dsa->pub_key = BN_dup(sig->parm[PUB]))) { goto done; }
  if ( ! (dsa->priv_key = BN_dup(sig->parm[PRIV]))) { goto done; }

  ret = 1;

 done:

  if (ret) {
    *d = dsa;
  } else {
    if (dsa) { DSA_free(dsa); }
  }

  if (sig) { frCryptSigFree(0, &sig); }

  return ret;
}


/*
 *  Print, to `fp', a signature object as a C structure.
 *  
 *  This routine is for generating hardwired keys.
 *  
 *  Always print the p, q and g parameters if they are set.  Only check the
 *  mask specifically for the public and private parameters.
 */
DLLExport int fr_CryptSigPrintAsC(ZkError err, FrCryptSig sig, uint8_t mask, char *name, FILE *fp)
{
  const char *fn = "fr_CryptSigPrintAsC";

  zkErrorClear(err);

  if ( ! sig) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sig is null");
    return 0;
  }
  if ( ! name) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "name is null");
    return 0;
  }
  if (strlen(name) > 32 /* bug */) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "name is too long (%lu > %d)",
                  (unsigned long)strlen(name), 32);
    return 0;
  }
  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "fp is null");
    return 0;
  }

  if (sig->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", sig->type);
    return 0;
  }

  if ( ! checkSigFields(err, sig)) {
    return 0;
  }

  /*
   *  Ensure that, if set, the fixed parameters will be printed.
   */

  if (sig->parm[P]) { mask |= FRCRYPT_SIGPARM_P; }
  if (sig->parm[Q]) { mask |= FRCRYPT_SIGPARM_Q; }
  if (sig->parm[G]) { mask |= FRCRYPT_SIGPARM_G; }

  if ( ! frCryptPubKeyPrintAsC(err, sig, mask, name, fp)) {
    return 0;
  }
  fputc('\n', fp);

  return 1;
}  


/*
 *
 *
 *                            External routines.
 *
 *
 */


#if DEBUG_VALID_PQG && ! defined(WIN32)

static int valid_pqg(BIGNUM *p, BIGNUM *q, BIGNUM *g)
{
  (void)g; /* bug? run tests on g? */
  
  assert(DSA_is_prime(p, 0, 0));
  assert(DSA_is_prime(q, 0, 0));

  /*
   *  Check that q is a prime factor of p-1.
   *  
   *  We already know that q is prime, so check that it's a factor.  Just for
   *  the hell of it, check that t = (p-1)/q has no remainder, and that q * t
   *  == (p-1) * q.
   */

  {
    BIGNUM *p_1, *p_1_div_q, *p_1_mod_q, *trial_p_1;
    BN_CTX *ctx, *ctx2;
    
    assert(ctx = BN_CTX_new());
    assert(ctx2 = BN_CTX_new());
    p_1 = ctx->bn[0];
    p_1_div_q = ctx->bn[1];
    p_1_mod_q = ctx->bn[2];
    trial_p_1 = ctx->bn[3];
    assert(BN_sub(p_1, p, BN_value_one()));
    assert(BN_div(p_1_div_q, p_1_mod_q, p_1, q, ctx2));
    assert(BN_is_zero(p_1_mod_q));
    assert(bn_mul(trial_p_1, p_1_div_q, q, ctx2));
    assert(BN_cmp(trial_p_1, p_1) == 0);
    BN_CTX_free(ctx);
    BN_CTX_free(ctx2);
  }
  
  /*
   *  Verify that g is a generator of the unique cyclic group of order q, in
   *  Z_*_p.  Where Z_p is the set of integers modulo p, and Z_*_p is the
   *  multiplicative group of Z_p.  In the case where p is prime (i.e. here)
   *  Z_*_p is [1, p-1].
   *  
   *  In particular...
   */
  
  /* bug: check g properties */

  return 1;
}

#endif


/*
 *  Create the DSA shared parameter bignums from their hardcoded, portable
 *  representation.
 *  
 *  Also initialize the public master signature key and the public software
 *  update signature key.  These will use the shared DSA parameters, too.
 */
DLLExport int fr_CryptSigInit(ZkError err)
{
  int ret = 0;
  size_t i, sz;
  
  zkErrorClear(err);

  if (sigIsInit) {
    zkErrorSet(err, "fr_CryptSigInit", ZKWARN, ZKMFRCRYPT, ZKERROR,
               "already initialized");
    return 0;
  }

  if (FRCRYPT_SIG_SIZE != sizeof(struct fr_sig_pk)) {
    zkErrorSet_va(err, "fr_CryptSigInit", ZKWARN, ZKMFRCRYPT, ZKERROR,
                  "inconsistent buffer sizes (%d != %lu)",
                  FRCRYPT_SIG_SIZE, (unsigned long)sizeof(struct fr_sig_pk));
    return 0;
  }

  for (i = 0; dsaSharedParm[i].nbits > 0; i++) {
    if ( ! ((dsaSharedParm[i].p = fr_BnNewFromData(err,
                                                   dsaSharedParm[i].p_data,
                                                   dsaSharedParm[i].p_size)) &&
            (dsaSharedParm[i].q = fr_BnNewFromData(err,
                                                   dsaSharedParm[i].q_data,
                                                   dsaSharedParm[i].q_size)) &&
            (dsaSharedParm[i].g = fr_BnNewFromData(err,
                                                   dsaSharedParm[i].g_data,
                                                   dsaSharedParm[i].g_size)))) {
      goto done;
    }

#if DEBUG_VALID_PQG && ! defined(WIN32)
    if ( ! valid_pqg(dsaSharedParm[i].p, dsaSharedParm[i].q, dsaSharedParm[i].g)) {
      abort();
    }
#endif
  }

  /* Initialize the public master signing key. */

  sz = sizeof fr_zkmp;

  if ( ! (frCryptPubKeyNew(err, &frCryptSigZkMaster) &&
          frCryptPubKeyReadFromBuf(err, frCryptSigZkMaster, fr_zkmp, &sz))) {
    return 0;
  }
  
  /* Initialize the public software update signing key. */

  sz = sizeof fr_supdp;

  if ( ! (frCryptPubKeyNew(err, &frCryptSigZkSupd) &&
          frCryptPubKeyReadFromBuf(err, frCryptSigZkSupd, fr_supdp, &sz))) {
    return 0;
  }
  
  /*
   *  bug: master and software keys should use shared parms directly
   */

#if DEBUG_ZKM_PARMS
  if (BN_cmp(frCryptSigZkMaster->parm[2], dsaSharedParm[2].p) == 0 &&
      BN_cmp(frCryptSigZkMaster->parm[3], dsaSharedParm[2].q) == 0 &&
      BN_cmp(frCryptSigZkMaster->parm[4], dsaSharedParm[2].g) == 0) {
    fprintf(stderr, "**** Master key parameters are consistent. ****\n");
  } else {
    fprintf(stderr, "**** Master key parameters are *not* consistent! ****\n");
  }

  if (BN_cmp(frCryptSigZkSupd->parm[2], dsaSharedParm[2].p) == 0 &&
      BN_cmp(frCryptSigZkSupd->parm[3], dsaSharedParm[2].q) == 0 &&
      BN_cmp(frCryptSigZkSupd->parm[4], dsaSharedParm[2].g) == 0) {
    fprintf(stderr, "**** Supd key parameters are consistent. ****\n");
  } else {
    fprintf(stderr, "**** Supd key parameters are *not* consistent! ****\n");
  }
#endif

  sigIsInit = 1;
  
  ret = 1;

 done:
  
  if ( ! ret) {
    for (i = 0; dsaSharedParm[i].nbits > 0; i++) {
      if (dsaSharedParm[i].p) { BN_free(dsaSharedParm[i].p); dsaSharedParm[i].p = 0; }
      if (dsaSharedParm[i].q) { BN_free(dsaSharedParm[i].q); dsaSharedParm[i].q = 0; }
      if (dsaSharedParm[i].g) { BN_free(dsaSharedParm[i].g); dsaSharedParm[i].g = 0; }
    }
    if (frCryptSigZkMaster) { frCryptPubKeyFree(err, &frCryptSigZkMaster); }
  }

  return ret;
}


/*
 *  Free memory that was allocated in fr_CryptSigFinal().
 */
DLLExport int fr_CryptSigFinal(ZkError err)
{
  size_t i;
  
  for (i = 0; dsaSharedParm[i].nbits > 0; i++) {
    if (dsaSharedParm[i].p) { BN_free(dsaSharedParm[i].p); dsaSharedParm[i].p = 0; }
    if (dsaSharedParm[i].q) { BN_free(dsaSharedParm[i].q); dsaSharedParm[i].q = 0; }
    if (dsaSharedParm[i].g) { BN_free(dsaSharedParm[i].g); dsaSharedParm[i].g = 0; }
  }

  if (frCryptSigZkMaster) { frCryptPubKeyFree(err, &frCryptSigZkMaster); }
  if (frCryptSigZkSupd) { frCryptPubKeyFree(err, &frCryptSigZkSupd); }

  return 1;
}


DLLExport int frCryptSigNew(ZkError err, FrCryptSig *sig)
{
  FrCryptSig s = 0;
  int ret = 0;
  
  zkErrorClear(err);

  if ( ! sigIsInit) {
    zkErrorSet(err, "frCryptSigNew", ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "signature initialization has not been performed");
    return 0;
  }

  if ( ! sig) {
    zkErrorSet(err, "frCryptSigNew", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sig is null");
    return 0;
  }

  if ( ! (frCryptPubKeyNew(err, &s) &&
          frCryptPubKeySetType(err, s, FRCRYPT_KEYTYPE_SIGNING))) {
    goto done;
  }

  *sig = s; s = 0;
  ret = 1;

 done:

  if (s) { (void)frCryptSigFree(0, &s); }
  
  return ret;
}


DLLExport int frCryptSigDup(ZkError err, FrCryptSig *copy, FrCryptSig key)
{
  const char *fn = "frCryptSigDup";

  zkErrorClear(err);

  if ( ! copy) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "copy is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }

  if (key->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", key->type);
    return 0;
  }

  return frCryptPubKeyDup(err, copy, key);
}


DLLExport int frCryptSigFree(ZkError err, FrCryptSig *sig)
{
  const char *fn = "frCryptSigFree";

  zkErrorClear(err);

  if ( ! sig) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sig is null");
    return 0;
  }
  if ( ! *sig) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
               "*sig is null");
    return 0;
  }

  if ((*sig)->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", (*sig)->type);
    return 0;
  }

  return frCryptPubKeyFree(err, sig);
}


/*
 *  Convert an integer to a signature key size.
 */
DLLExport int frCryptSigKeySizeFromInt(ZkError err, int n, FrCryptSigKeySize *ksz)
{
  zkErrorClear(err);

  if ( ! ksz) {
    zkErrorSet(err, "frCryptSigKeySizeFromInt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "ksz is null");
    return 0;
  }

  switch (n) {
  case 1024: *ksz = FRCRYPT_SIGKEYSIZE_1024; break;
  case 1536: *ksz = FRCRYPT_SIGKEYSIZE_1536; break;
  case 2048: *ksz = FRCRYPT_SIGKEYSIZE_2048; break;

  default:
    zkErrorSet_va(err, "frCryptSigKeySizeFromInt", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key size (%d)", n);
    return 0;
  }

  return 1;
}


/*
 *  Get the version from the packed signature header.
 */
DLLExport int frCryptSigBufGetVersion(ZkError err, void *sbuf, size_t ssz, unsigned int *vers)
{
  struct fr_sig_pk *s = sbuf;
  
  zkErrorClear(err);

  if ( ! sbuf) {
    zkErrorSet(err, "frCryptSigBufGetVersion", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sbuf is null");
    return 0;
  }
  if (ssz < sizeof *s) {
    zkErrorSet_va(err, "frCryptSigBufGetVersion", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
                  "signature buffer is too small (%lu < %lu)",
                  (unsigned long)ssz, (unsigned long)sizeof *s);
    return 0;
  }
  if ( ! vers) {
    zkErrorSet(err, "frCryptSigBufGetVersion", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "vers is null");
    return 0;
  }

  *vers = s->keyVersion;

  return 1;
}


/*
 *  Print, in a human readable format, the contents of a signature buffer.
 */
DLLExport int frCryptSigBufPrint(ZkError err, FILE *fp, void *sbuf, size_t ssz)
{
  struct fr_sig_pk *s = sbuf;
  size_t i;

  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, "frCryptSigBufPrint", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "fp is null");
    return 0;
  }
  if ( ! sbuf) {
    zkErrorSet(err, "frCryptSigBufPrint", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sbuf is null");
    return 0;
  }
  if (ssz != sizeof *s) {
    zkErrorSet_va(err, "frCryptSigBufPrint", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "ssz is invalid (%lu != %lu)",
                  (unsigned long)ssz, (unsigned long)sizeof *s);
    return 0;
  }
  
  /*
   *  Print the fields in the order in which they appear.
   */
  
  fprintf(fp, "Signature:\n");
  fprintf(fp, "    r: ");
  for (i = 1; i < sizeof s->r; i++) {
    fprintf(fp, "%02x", (unsigned int)(unsigned char)s->r[i]);
    if (i % 20 == 0) { fprintf(fp, "\n       "); }
  }
  fputs("\n", fp);
  fprintf(fp, "    s: ");
  for (i = 1; i < sizeof s->s; i++) {
    fprintf(fp, "%02x", (unsigned int)(unsigned char)s->s[i]);
    if (i % 20 == 0) { fprintf(fp, "\n       "); }
  }
  fputs("\n", fp);
  fprintf(fp, "    key version: %u\n", (unsigned int)s->keyVersion);
  fprintf(fp, "    key size: %u\n", (unsigned int)ntohs(s->keySize));
  fprintf(fp, "    signed length: %lu\n", (unsigned long)ntohl(s->dataSize));

  return 1;
}


/*
 *  We only accept a limited set of key sizes.
 */
DLLExport int frCryptSigGetKeySize(ZkError err, FrCryptSig sig, FrCryptSigKeySize *ksz)
{
  const char *fn = "frCryptSigGetKeySize";
  unsigned int ks;

  zkErrorClear(err);

  if ( ! sig) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sig is null");
    return 0;
  }
  if ( ! ksz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "ksz is null");
    return 0;
  }

  if (sig->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", sig->type);
    return 0;
  }

  if ( ! (frCryptPubKeyGetSize(err, sig, &ks) &&
          intToSigKeySize(err, ks, ksz))) {
    return 0;
  }
  
  return 1;
}


/*
 *  We only accept a limited set of key sizes.
 */
DLLExport int frCryptSigSetKeySize(ZkError err, FrCryptSig sig, FrCryptSigKeySize ksz)
{
  BIGNUM *p, *q, *g;
  const char *fn = "frCryptSigSetKeySize";

  zkErrorClear(err);

  if ( ! sig) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sig is null");
    return 0;
  }

  if (sig->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", sig->type);
    return 0;
  }

  if (sig->parm[PRIV] || sig->parm[PUB]) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "error setting size with public and/or private key already set");
    return 0;
  }
  
  /*
   *  Set the fixed parameters, based on the key size.
   */

  if ( ! fr_CryptGetSigSharedParms(err, ksz, &p, &q, &g)) {
    return 0;
  }

  if ( ! (frCryptPubKeySetParm(err, sig, P, p, 0, 0) &&
          frCryptPubKeySetParm(err, sig, Q, q, 0, 0) &&
          frCryptPubKeySetParm(err, sig, G, g, 0, 0) &&
          frCryptPubKeySetSize(err, sig, ksz))) {
    return 0;
  }

  return 1;
}


/*
 *  Generate random public and private keys.
 */
DLLExport int frCryptSigSetRandomKeys(ZkError err, FrCryptSig sig)
{
  BIGNUM *priv = 0, *pub = 0;
  DSA *keys = 0;
  const char *fn = "frCryptSigSetRandomKeys";
  int ret = 0;
  size_t n;
  
  zkErrorClear(err);

  if ( ! sig) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sig is null");
    return 0;
  }
  
  if (sig->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", sig->type);
    return 0;
  }

  if ( ! (sig->parm[P] && sig->parm[Q] && sig->parm[G])) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "the key has not been initialized");
    return 0;
  }

  /*
   *  Set up a DSA structure so we can use SSLeay's DSA_generate_key().
   */

  if ( ! ((keys = DSA_new()) &&
          (keys->p = BN_dup(sig->parm[P])) &&
          (keys->q = BN_dup(sig->parm[Q])) &&
          (keys->g = BN_dup(sig->parm[G])))) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENOMEM,
               "can't initialize DSA parameters");
    goto done;
  }
    
  n = BN_num_bits(keys->q) / CHAR_BIT;

  if ( ! DSA_generate_key(keys)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
               "error generating DSA keys");
    goto done;
  }
  
  if ( ! ((priv = BN_dup(keys->priv_key)) &&
          (pub  = BN_dup(keys->pub_key)))) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENOMEM,
               "can't duplicate public and private keys");
    goto done;
  }

  /*
   *  We give "ownership" of these parameters to the PubKey object.  It will
   *  be responsible for freeing them.
   */

  if ( ! (frCryptPubKeySetParm(err, sig, PRIV, priv, 1, 1) &&
          frCryptPubKeySetParm(err, sig, PUB,  pub,  1, 1))) {
    goto done;
  }

#if DEBUG_SIG
  fprintf(stderr, "RandomKeys\n");
  fprintf(stderr, "  p is %d bits, %d bytes.\n",
          BN_num_bits(sig->parm[P]), BN_num_bytes(sig->parm[P]));
  fprintf(stderr, "  q is %d bits, %d bytes.\n",
          BN_num_bits(sig->parm[Q]), BN_num_bytes(sig->parm[Q]));
  fprintf(stderr, "  g is %d bits, %d bytes.\n",
          BN_num_bits(sig->parm[G]), BN_num_bytes(sig->parm[G]));
#endif

  priv = 0; pub = 0;
  ret = 1;

 done:
  
  if (priv) { BN_clear_free(priv); }
  if (pub)  { BN_clear_free(pub); }
  if (keys) { DSA_free(keys); }

  return ret;
}


/*
 *  In the case of signing an existing hash we don't include the length of the
 *  signed data in the signature, since we already know it.
 */
DLLExport int frCryptSigSignHash(ZkError err, FrCryptSig sig, void *hash, size_t hlen, void *sbuf, size_t slen)
{
  const char *fn = "frCryptSigSignHash";


  zkErrorClear(err);

  if ( ! sig) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sig is null");
    return 0;
  }
  if ( ! hash) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "hash is null");
    return 0;
  }
  if (hlen != 20 /* bug */) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
                  "hlen is invalid (%lu != %d)", (unsigned long)hlen, 20);
    return 0;
  }
  if ( ! sbuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sbuf is null");
    return 0;
  }
  if (slen != FRCRYPT_SIG_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
                  "slen is invalid (%lu != %d)",
                  (unsigned long)slen, FRCRYPT_SIG_SIZE);
    return 0;
  }

  if (sig->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", sig->type);
    return 0;
  }

  if ( ! checkSigFields(err, sig)) {
    return 0;
  }

  {
    FrCryptSigKeySize keySize;
    struct fr_sig_pk *s = sbuf;
    unsigned int vers;
    int ret = 0, i;
    

    if ( ! (frCryptPubKeyGetVersion(err, sig, &vers) &&
            frCryptPubKeyGetSize(err, sig, &keySize))) {
      return 0;
    }

    s->keyVersion = vers;
    s->keySize = htons(keySize);
    s->dataSize = htonl(hlen);
    
    /*
     * if the error returned is ZKEBADSTATE, we should try again.
     * There is a slight chance of it happening, calling the function
     * 4 times shoould be enough.
     */
    for (i = 0; i < 3; i++) {
      ret = zks_dsa_sign(err, sig, hash, s->r, s->s);

      if (ret || ( ! ret && zkErrorCode(err) != ZKEBADSTATE)) {
        break;
      }
    }

    return ret;
  }
}


/*
 *  In the case of verifying the signature over a hash, the signature doesn't
 *  include the length of the signed data, since we already know it.
 */
DLLExport int frCryptSigVerifyHash(ZkError err, FrCryptSig sig, void *hash, size_t hlen, void *sbuf, size_t slen, int *valid)
{
  const char *fn = "frCryptSigVerifyHash";

  zkErrorClear(err);

  if ( ! sig) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sig is null");
    return 0;
  }
  if ( ! hash) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "hash is null");
    return 0;
  }
  if (hlen != 20 /* bug */) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "hlen is invalid (%lu != %d)", (unsigned long)hlen, 20);
    return 0;
  }
  if ( ! sbuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sbug is null");
    return 0;
  }
  if (slen != FRCRYPT_SIG_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "slen is invalid (%lu != %d)",
                  (unsigned long)slen, FRCRYPT_SIG_SIZE);
    return 0;
  }
  if ( ! valid) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "valid is null");
    return 0;
  }

  if (sig->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", sig->type);
    return 0;
  }

  if ( ! checkSigFields(err, sig)) {
    return 0;
  }

  {
    FrCryptSigKeySize keySize;
    size_t ds;
    struct fr_sig_pk *s = sbuf;
    unsigned int ks, vers;
  
    if ( ! (frCryptPubKeyGetVersion(err, sig, &vers) &&
            frCryptPubKeyGetSize(err, sig, &keySize))) {
      return 0;
    }

    ks = ntohs(s->keySize);

    if (s->keyVersion != vers) {
      zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                    "inconsistent key versions (%u != %u)",
                    (unsigned int)s->keyVersion, vers);
      return 0;
    }

    if (ks != keySize) {
      zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                    "inconsistent key sizes (%u != %d)", ks, keySize);
      return 0;
    }

    ds = ntohl(s->dataSize);

    if (ds != hlen) {
      zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                    "inconsistent signed sizes (%lu != %lu)",
                    (unsigned long)ds, (unsigned long)hlen);
      return 0;
    }
    
    return zks_dsa_verify(sig, hash, s->r, s->s, valid);
  }
}


/*
 *  Similar to frCryptSigSignData(), but the data contained in one or more
 *  regions of possibly non-contiguous memory, pointed to by the elements of
 *  the null terminated array `data'.
 */
DLLExport int frCryptSigSignDataArray(ZkError err, FrCryptSig sig, void **data, size_t *dlen, void *sbuf, size_t slen)
{
  FrCryptSigKeySize keySize;
  const char *fn = "frCryptSigSignDataArray";
  size_t i, totSize;
  struct fr_sig_pk *s = sbuf;
  unsigned int vers;

  zkErrorClear(err);

  if ( ! sig) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sig is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! dlen) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "dlen is null");
    return 0;
  }
  if ( ! sbuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sbuf is null");
    return 0;
  }
  if (slen < FRCRYPT_SIG_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "the value of slen is too small (%lu < %d)",
                  (unsigned long)slen, FRCRYPT_SIG_SIZE);
    return 0;
  }

  if (sig->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", sig->type);
    return 0;
  }

#if 0
  if (sig->keySize == FRCRYPT_SIGKEYSIZE_UNUSED_) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "the key size has not been set");
    return 0;
  }
#endif

  if ( ! checkSigFields(err, sig)) {
    return 0;
  }

  if ( ! (frCryptPubKeyGetVersion(err, sig, &vers) &&
          frCryptPubKeyGetSize(err, sig, &keySize))) {
    return 0;
  }

  if (vers == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
               "key version is 0");
    return 0;
  }

  s->keyVersion = vers;
  s->keySize = htons(keySize);

  totSize = SIG_INIT_LEN;

  for (i = 0; data[i]; i++) {
    totSize += dlen[i];
  }

  s->dataSize = htonl(totSize);

  /*
   *  Generate a SHA-1 hash over the tail of the packed signature structure
   *  and the buffers pointed to by the elements of `data', then sign it.
   */
  
  {
    SHA_CTX sha;
    int ret = 0;
    unsigned char hash[20];
    
    SHA1_Init(&sha);
    SHA1_Update(&sha, (unsigned char *)sbuf + SIG_INIT_OFF, SIG_INIT_LEN);
    for (i = 0; data[i]; i++) {
      SHA1_Update(&sha, data[i], dlen[i]);
    }
    SHA1_Final(hash, &sha);

    /*
     *  Allow a few attempts to sign some data, on the remote chance that
     *  zks_dsa_sign() will generate bad parameters.
     */

    for (i = 0; i < 3; i++) {
      ret = zks_dsa_sign(err, sig, hash, s->r, s->s);

      if (ret || ( ! ret && zkErrorCode(err) != ZKEBADSTATE)) {
        break;
      }
    }

    memset(hash, 0, sizeof hash);

    return ret;
  }
}


/*
 *  Sign the `dlen' bytes in `data', putting the resulting signature into
 *  `sbuf'.  The initial value of `*slen' should be the number of bytes in
 *  `sbuf'; it will be updated to reflect the size of the resulting signature.
 */
DLLExport int frCryptSigSignData(ZkError err, FrCryptSig sig, void *data, size_t dlen, void *sbuf, size_t slen)
{
  const char *fn = "frCryptSigSignData";

  zkErrorClear(err);

  if ( ! sig) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sig is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (dlen == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
               "dlen is 0");
    return 0;
  }
  if ( ! sbuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sbuf is null");
    return 0;
  }
  if (slen < FRCRYPT_SIG_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "the value of slen is too small (%lu < %d)",
                  (unsigned long)slen, FRCRYPT_SIG_SIZE);
    return 0;
  }

  if (sig->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", sig->type);
    return 0;
  }

#if 0
  if (sig->keySize == FRCRYPT_SIGKEYSIZE_UNUSED_) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "the key size has not been set");
    return 0;
  }
#endif

  if ( ! checkSigFields(err, sig)) {
    return 0;
  }

  {
    size_t dsz[2];
    void *d[2];

    d[0] = data; dsz[0] = dlen;
    d[1] = 0;    dsz[1] = 0;
    return frCryptSigSignDataArray(err, sig, d, dsz, sbuf, slen);
  }
}


/*
 *  Verify the signature over the `dlen' bytes in `data'.  The signature is
 *  contained in `sbuf' and has length `slen'.  The value of `*valid' will be
 *  set to indicate whether the signature is valid.
 */
DLLExport int frCryptSigVerifyDataArray(ZkError err, FrCryptSig sig, void **data, size_t *dlen, void *sbuf, size_t slen, int *valid)
{
  FrCryptSigKeySize keySize;
  const char *fn = "frCryptSigVerifyDataArray";
  size_t ds, i, totSize;
  struct fr_sig_pk *s = sbuf;
  unsigned int ks, vers;
  
  zkErrorClear(err);

  if ( ! sig) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sig is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (dlen == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
               "dlen is 0");
    return 0;
  }
  if ( ! sbuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sbuf is null");
    return 0;
  }
  if (slen < FRCRYPT_SIG_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "the value of slen is too small (%lu < %d)",
                  (unsigned long)slen, FRCRYPT_SIG_SIZE);
    return 0;
  }
  if ( ! valid) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "valid is null");
    return 0;
  }
  
  if (sig->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", sig->type);
    return 0;
  }

  if ( ! checkSigFields(err, sig)) {
    return 0;
  }

  if ( ! (frCryptPubKeyGetVersion(err, sig, &vers) &&
          frCryptPubKeyGetSize(err, sig, &keySize))) {
    return 0;
  }
  
  ks = ntohs(s->keySize);

  if (vers == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
               "key version is 0");
    return 0;
  }

  if (s->keyVersion != vers) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
                  "key version does not match that of the signature (%u != %u)",
                  s->keyVersion, vers);
    return 0;
  }

  if (ks != keySize) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
                  "inconsistent key sizes (%u != %d)", ks, keySize);
    return 0;
  }

  totSize = SIG_INIT_LEN;

  for (i = 0; data[i]; i++) {
    totSize += dlen[i];
  }

  ds = ntohl(s->dataSize);

  if (ds != totSize) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
                  "inconsistent signed data sizes (%lu != %lu)",
                  (unsigned long)ds, (unsigned long)totSize);
    return 0;
  }

  {
    SHA_CTX sha;
    int ret;
    unsigned char hash[20];
    
    SHA1_Init(&sha);
    SHA1_Update(&sha, (unsigned char *)sbuf + SIG_INIT_OFF, SIG_INIT_LEN);
    for (i = 0; data[i]; i++) {
      SHA1_Update(&sha, data[i], dlen[i]);
    }
    SHA1_Final(hash, &sha);

    ret = zks_dsa_verify(sig, hash, s->r, s->s, valid);
    memset(hash, 0, sizeof hash);
    
    return ret;
  }
}


/*
 *  Verify the signature over the `dlen' bytes in `data'.  The signature is
 *  contained in `sbuf' and has length `slen'.  The value of `*valid' will be
 *  set to indicate whether the signature is valid.
 */
DLLExport int frCryptSigVerifyData(ZkError err, FrCryptSig sig, void *data, size_t dlen, void *sbuf, size_t slen, int *valid)
{
  const char *fn = "frCryptSigVerifyData";
  
  zkErrorClear(err);

  if ( ! sig) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sig is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (dlen == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
               "dlen is 0");
    return 0;
  }
  if ( ! sbuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sbuf is null");
    return 0;
  }
  if (slen < FRCRYPT_SIG_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "the value of slen is too small (%lu < %d)",
                  (unsigned long)slen, FRCRYPT_SIG_SIZE);
    return 0;
  }
  if ( ! valid) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "valid is null");
    return 0;
  }
  
  if (sig->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", sig->type);
    return 0;
  }

  if ( ! checkSigFields(err, sig)) {
    return 0;
  }

  {
    size_t dsz[2];
    void *d[2];

    d[0] = data; dsz[0] = dlen;
    d[1] = 0;    dsz[1] = 0;
    return frCryptSigVerifyDataArray(err, sig, d, dsz, sbuf, slen, valid);
  }
}


/*
 *  Create a new signature key from a portable representation of a signature
 *  public key.  `*dsz' should initially contain the size of the buffer
 *  holding the portable signature data.  After a successful return this value
 *  will be updated with the actual size of the signature data.
 *  
 *  The storage format is: 1 byte key type (here it is `signing'); 1 byte
 *  algorithm (currently DSA); 1 byte key size indicator; 1 byte bitmask
 *  representing the parameters that are stored (see the FRCRYPT_SIGPARMS_*
 *  defines in frCrypt.h); that many bignums, each stored as a two byte
 *  unsigned byte count, followed by that many bytes (both parts in network
 *  byte order).
 */
DLLExport int frCryptSigNewFromData(ZkError err, FrCryptSig *sig, void *data, size_t *dsz)
{
  FrCryptSig s = 0;
  int ret = 0;

  zkErrorClear(err);

  if ( ! sigIsInit) {
    zkErrorSet(err, "frCryptSigNewFromData", ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "signature initialization has not been performed");
    return 0;
  }

  if ( ! sig) {
    zkErrorSet(err, "frCryptSigNewFromData", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sig is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "frCryptSigNewFromData", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, "frCryptSigNewFromData", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "dsz is null");
    return 0;
  }

  if ( ! (frCryptSigNew(err, &s) &&
          frCryptPubKeyReadFromBuf(err, s, data, dsz))) {
    goto done;
  }

  /*
   *  Verify that this is, in fact, a signature key.  We'll assume that if the
   *  type is correct, so are the other fields (perhaps not a great
   *  assumption...).
   */

  {
    unsigned int type;

    if ( ! frCryptPubKeyGetType(err, s, &type)) {
      goto done;
    }
    if (type != FRCRYPT_KEYTYPE_SIGNING) {
      zkErrorSet_va(err, "frCryptSigNewFromData", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                    "not a signature key (type is %u)", type);
      goto done;
    }
  }

#if 0
  /*
   *  For convenience, point our named bignums at the key parameters.
   *  Normally p, q and g will not have been saved (as they usually point to
   *  static storage), but we'll keep them even if they do get set.
   */

  if ( ! (frCryptPubKeyGetParm(err, s->pk, P,    &s->p,       &freeIt, &saveIt) &&
          frCryptPubKeyGetParm(err, s->pk, Q,    &s->q,       &freeIt, &saveIt) &&
          frCryptPubKeyGetParm(err, s->pk, G,    &s->g,       &freeIt, &saveIt) &&
          frCryptPubKeyGetParm(err, s->pk, PUB,  &s->parm[PUB],  &freeIt, &saveIt) &&
          frCryptPubKeyGetParm(err, s->pk, PRIV, &s->parm[PRIV], &freeIt, &saveIt))) {
    goto done;
  }
#endif
  
  /*
   *  If one is unset assume they're all unset...
   */

  if ( ! (s->parm[P] && s->parm[Q] && s->parm[G])) {
    unsigned int ksz;

    if ( ! (frCryptPubKeyGetSize(err, s, &ksz) &&
            fr_CryptGetSigSharedParms(err, ksz, &s->parm[P],
                                      &s->parm[Q], &s->parm[G]))) {
      goto done;
    }
  
#if 0
    if ( ! (frCryptPubKeySetParm(err, s->pk, P, s->p, 0, 0) &&
            frCryptPubKeySetParm(err, s->pk, Q, s->q, 0, 0) &&
            frCryptPubKeySetParm(err, s->pk, G, s->g, 0, 0))) {
      goto done; /* bug */
    }
#endif
  }

  *sig = s; s = 0;
  ret = 1;

#if DEBUG_LDST
  fprintf(stderr, "Loaded\n"); fr_CryptSigPrint(err, s); fprintf(stderr, "\n");
#endif

 done:

  if (s) { (void)frCryptSigFree(0, &s); }

  return ret;
}


DLLExport int frCryptSigNewFromFile(ZkError err, FrCryptSig *sig, const char *file)
{
  FILE *fp = 0;
  char buf[3 * 1024]; /* bug: fixed size */
  int ret = 0;
  size_t n;

  zkErrorClear(err);

  if ( ! sig) {
    zkErrorSet(err, "frCryptSigNewFromFile", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sig is null");
    return 0;
  }
  if ( ! file) {
    zkErrorSet(err, "frCryptSigNewFromFile", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "file is null");
    return 0;
  }

  if ( ! (fp = fopen(file, "rb"))) {
    zkErrorSet_va(err, "frCryptSigNewFromFile", ZKERR, ZKMFRCRYPT, errno,
                  "can't open `%s' for reading: %s", file, strerror(errno));
    goto done;
  }

  /* bug: assumes we can get everything in one read. */

  if ((n = fread(buf, 1, sizeof buf, fp)) < 1) {
    zkErrorSet_va(err, "frCryptSigNewFromFile", ZKERR, ZKMFRCRYPT, errno,
                  "error reading %lu bytes from `%s': %s",
                  (unsigned long)sizeof buf, file, strerror(errno));
    goto done;
  }

  (void)fclose(fp); fp = 0;

  ret = frCryptSigNewFromData(err, sig, buf, &n);

 done:

  if (fp) { (void)fclose(fp); }

  return ret;
}


DLLExport int frCryptSigMakeSig(ZkError err, FrCryptSig sig)
{
  zkErrorClear(err);

  if ( ! sig) {
    zkErrorSet(err, "frCryptSigMakeSig", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sig is null");
    return 0;
  }
  
  if (sig->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, "frCryptSigMakeSig", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type (%d)", sig->type);
    return 0;
  }

  if ( ! (sig->parm[P] && sig->parm[Q] && sig->parm[G])) {
    unsigned int ksz;

    if ( ! frCryptPubKeyGetSize(err, sig, &ksz)) {
      return 0;
    }
    if ( ! fr_CryptGetSigSharedParms(err, ksz, &sig->parm[P],
                                     &sig->parm[Q], &sig->parm[G])) {
      return 0;
    }
  }

  return 1;
}


DLLExport int frCryptSigStore(ZkError err, FrCryptSig sig, uint8_t mask, void *data, size_t *dsz)
{
  const char *fn = "frCryptSigStore";

  zkErrorClear(err);

  if ( ! sig) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sig is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "dsz is null");
    return 0;
  }

  if (sig->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", sig->type);
    return 0;
  }

  if ( ! checkSigFields(err, sig)) {
    return 0;
  }

#if DEBUG_LDST
  fprintf(stderr, "Storing\n"); fr_CryptSigPrint(err, sig); fprintf(stderr, "\n");
#endif

  /*
   *  Currently, we only allow a subset of all possible combinations.
   */

  switch (mask) {
  case FRCRYPT_SIGPARMS_ALL:
  case FRCRYPT_SIGPARMS_PUB:
  case FRCRYPT_SIGPARMS_PUBFX:
  case FRCRYPT_SIGPARMS_PRIV:
  case FRCRYPT_SIGPARMS_PRIVFX:
  case FRCRYPT_SIGPARMS_PUBPRIV:
    break;

  default:
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                  "invalid parameter mask (0x%x)", mask);
    return 0;
  }

  if ( ! frCryptPubKeySetSaveMask(err, sig, mask)) {
    return 0;
  }
  
  return frCryptPubKeyWriteToBuf(err, sig, data, dsz);
}


DLLExport int frCryptSigEqual(ZkError err, FrCryptSig s0, FrCryptSig s1, int *equal)
{
  const char *fn = "frCryptSigEqual";

  zkErrorClear(err);

  if ( ! s0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "s0 is null");
    return 0;
  }
  if ( ! s1) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "s1 is null");
    return 0;
  }
  if ( ! equal) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "equal is null");
    return 0;
  }

  if (s0->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "s0 has an invalid key type `%u'", s0->type);
    return 0;
  }

  if (s1->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "s1 has an invalid key type `%u'", s1->type);
    return 0;
  }

  if ( ! (checkSigFields(err, s0) && checkSigFields(err, s1))) {
    return 0;
  }

  return frCryptPubKeyEqual(err, s0, s1, equal);
}


/*
 *  Set `*equal' to 1 if `s0' and `s1' form a public/private signature key
 *  pair.
 */
DLLExport int frCryptSigPubPrivPair(ZkError err, FrCryptSig s0, FrCryptSig s1, int *equal)
{
  const char *fn = "frCryptSigPubPrivPair";
  uint8_t mask = (FRCRYPT_SIGPARM_Y |
                  FRCRYPT_SIGPARM_P |
                  FRCRYPT_SIGPARM_Q |
                  FRCRYPT_SIGPARM_G);
  
  zkErrorClear(err);

  if ( ! s0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "s0 is null");
    return 0;
  }
  if ( ! s1) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "s1 is null");
    return 0;
  }
  if ( ! equal) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "equal is null");
    return 0;
  }

  if (s0->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "s0 has an invalid key type `%u'", s0->type);
    return 0;
  }

  if (s1->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "s1 has an invalid key type `%u'", s1->type);
    return 0;
  }

  if ( ! (checkSigFields(err, s0) && checkSigFields(err, s1))) {
    return 0;
  }

  return frCryptPubKeyMaskEqual(err, s0, s1, mask, equal);
}


/*
 *  Only validate parameters specific to cryption keys.
 */
DLLExport int frCryptSigPublicValidate(ZkError err, FrCryptCrypt key)
{
  const char *fn = "frCryptSigPublicValidate";
  
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "key is null");
    return 0;
  }

  if (key->type != FRCRYPT_KEYTYPE_SIGNING) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", key->type);
    return 0;
  }

  if ( ! key->parm[PUB]) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE, "no public parmameter");
    return 0;
  }

  if (key->parm[PRIV]) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE, "has private parmameter");
    return 0;
  }

  return 1;
}


/*
 *  This pack routine, and the following unpack routine were written for the
 *  specific purpose of compressing and decompressing the signature generated
 *  over certain email headers.  The goal is to minimize the amount of space
 *  consumed by the signature, and to avoid problems with overly long lines.
 *  
 *  Warning: do not use this code in general, since it may go away when header
 *           signatures are replaced by a better mechanism.
 */
DLLExport int frCryptPackSig( ZkError err, const void *sbuf, size_t slen, char* compact )
{
  const char *fn = "frCryptPackSig";
  const struct fr_sig_pk *sig_pkt;

  if ( ! sbuf || ! compact ) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "param is null ptr");
    return 0;
  }

  if ( slen != FRCRYPT_SIG_SIZE ) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD, "sig size wrong" );
    return 0;
  }

  sig_pkt = (const struct fr_sig_pk*) sbuf;

  if ( ntohs( sig_pkt->keySize ) != 1024 ) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD, "attempt to pack unsupported key size" );
    return 0;
  }

  *compact = sig_pkt->keyVersion;
  memcpy( compact+1, sig_pkt->r + RS_LEN - FRCRYPT_HASH_SIZE, FRCRYPT_HASH_SIZE );
  memcpy( compact+1+FRCRYPT_HASH_SIZE, sig_pkt->s + RS_LEN - FRCRYPT_HASH_SIZE, FRCRYPT_HASH_SIZE );

  return 1;
}


/*
 *  See comments for frCryptPackSig().
 *  
 *  Warning: do not use this code in general, since it may go away when header
 *           signatures are replaced by a better mechanism.
 */
DLLExport int frCryptUnpackSig( ZkError err, void *sbuf, size_t slen, const char* compact )
{
  const char *fn = "frCryptUnpackSig";
  struct fr_sig_pk *sig_pkt;  

  if ( ! sbuf || ! compact ) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "param is null ptr");
    return 0;
  }

  if ( slen != FRCRYPT_SIG_SIZE ) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD, "sig size wrong" );
    return 0;
  }

  sig_pkt = (struct fr_sig_pk*) sbuf;
  sig_pkt->keySize = htons( 1024 );
  sig_pkt->dataSize = htonl( FRCRYPT_HASH_SIZE );
  sig_pkt->keyVersion = *compact;

  memset( sig_pkt->r, 0, RS_LEN );
  memset( sig_pkt->s, 0, RS_LEN );

  memcpy( sig_pkt->r+RS_LEN-FRCRYPT_HASH_SIZE, compact+1, FRCRYPT_HASH_SIZE );
  memcpy( sig_pkt->s+RS_LEN-FRCRYPT_HASH_SIZE, compact+1+FRCRYPT_HASH_SIZE, FRCRYPT_HASH_SIZE );

  return 1;
}
