/*
 *  $Id: symkey.c,v 1.30.8.6 2000/12/14 17:52:41 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include <errno.h>
#include <stdlib.h>
#define _XOPEN_SOURCE          1 /* for strcasecmp() */
#define _XOPEN_SOURCE_EXTENDED 1 /* for strcasecmp() */
#include <string.h>
#undef _XOPEN_SOURCE
#undef _XOPEN_SOURCE_EXTENDED
#include "openssl/evp.h"
#include "frBf.h"
#include "frBn.h"
#include "frDes3.h"
#include "openssl/des.h"
#include "fr_gencrypt.h"
#include "fr_sym.h"
#include "fr_hash.h"
#include "sym.h"
#include "zkvheap.h"


#ifndef FREEDOM_PRODUCTION_BUILD
#ifndef WIN32
#warning "enabling insecure encryption algorithms!"
#endif
#endif


#define MYMOD ZKMFRCRYPT

/** The maximum size of a symmetric key. */
#define KEY_SIZE_MAX  BLOWFISH_KEY_SIZE_MAX

/** To be encrypted, all data blocks must be a multiple of 8 bytes. */
#define MIN_CRYPT_SIZE 8

/** Value used by the XOR `encryption' algorithm. */
#define XOR_VALUE 0x55


static const char xor_value = XOR_VALUE;


/*
 *
 *
 *                            Internal routines.
 *
 *
 */


/*
 *  For reporting errors, use the caller's function name.
 */
/*
 *  bug? use the higher level routines described in cipher.doc?
 */
static int symCrypt(ZkError err, const char *fname, FrCryptSym sym,
                    int encrypt, void *ivec, void *data, size_t len)
{
  int ret = 0;

  zkErrorClear(err);
  
  if ( ! sym) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    return 0;
  }
  if ( ! ivec) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR, "ivec is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR, "data is null");
    return 0;
  }

  if (len % MIN_CRYPT_SIZE) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKEPARMBAD,
               "length must be a multiple of 8");
    /* bug: add varargs to zkErrorSet */
    return 0;
  }

  /*
   *  bug? maybe we should set a function pointer to our encryption algorithm,
   *  so we don't need a bunch of switch statements.  On the other hand, it
   *  would still have to be a front-end, since there's no common arguments.
   */
  
  switch (sym->alg) {
#ifndef FREEDOM_PRODUCTION_BUILD
    
  case FRCRYPT_SYMALG_NONE:
    ret = 1;
    break;
    
  case FRCRYPT_SYMALG_XOR:
    ret = (encrypt ? fr_XorEncrypt : fr_XorDecrypt)(err, sym, ivec, data, len);
    break;

#endif

  case FRCRYPT_SYMALG_BLOWFISH:
    ret = (encrypt ? fr_BlowfishEncrypt : fr_BlowfishDecrypt)(err, sym, ivec, data, len);
    break;
    
  case FRCRYPT_SYMALG_DES3:
    ret = (encrypt ? fr_Des3Encrypt : fr_Des3Decrypt)(err, sym, ivec, data, len);
    break;
    
  case FRCRYPT_SYMALG_DESX:
    ret = (encrypt ? fr_DesxEncrypt : fr_DesxDecrypt)(err, sym, ivec, data, len);
    break;
    
  default:
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEPARMBAD,
                  "unknown encryption algoritm (%d)", sym->alg);
    return 0;
  }

  return ret;
}



/* Use Encrypt/Decrypt in a different buffer */
static int symCryptToBuf(ZkError err, const char *fname, FrCryptSym sym,
                        int encrypt, void *ivec, void *data, void *output, size_t len)
{
  int ret = 0;

  zkErrorClear(err);
  
  if ( ! sym) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    return 0;
  }
  if ( ! ivec) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR, "ivec is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR, "data is null");
    return 0;
  }
  if ( ! output) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR, "output is null");
    return 0;
  }

  if (len % MIN_CRYPT_SIZE) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "length must be a multiple of 8");
    /* bug: add varargs to zkErrorSet */
    return 0;
  }

  /*
   *  bug? maybe we should set a function pointer to our encryption algorithm,
   *  so we don't need a bunch of switch statements.  On the other hand, it
   *  would still have to be a front-end, since there's no common arguments.
   */
  
  switch (sym->alg) {
#ifndef FREEDOM_PRODUCTION_BUILD
    
  case FRCRYPT_SYMALG_NONE:
    ret = 1;
    break;
    
  case FRCRYPT_SYMALG_XOR:
    ret = (encrypt ? fr_XorEncryptToBuf : fr_XorDecryptToBuf)(err, sym, ivec, data, output, len);
    break;

#endif

  case FRCRYPT_SYMALG_BLOWFISH:
    ret = (encrypt ? fr_BlowfishEncryptToBuf : fr_BlowfishDecryptToBuf)(err, sym, ivec, data, output, len);
    break;
    
  case FRCRYPT_SYMALG_DES3:
    ret = (encrypt ? fr_Des3EncryptToBuf : fr_Des3DecryptToBuf)(err, sym, ivec, data, output, len);
    break;
    
  case FRCRYPT_SYMALG_DESX:
    ret = (encrypt ? fr_DesxEncryptToBuf : fr_DesxDecryptToBuf)(err, sym, ivec, data, output, len);
    break;
    
  default:
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEPARMBAD,
                  "unknown encryption algoritm (%d)", sym->alg);
    return 0;
  }

  return ret;
}


/*
 *
 *
 *                            External routines.
 *
 *
 */


DLLExport int frCryptSymNameToAlg(ZkError err, const char *name, FrCryptSymAlg *alg)
{
  FrCryptSymAlg a = FRCRYPT_SYMALG_UNUSED_;
  
  zkErrorClear(err);

  if (strcasecmp(name, "blowfish") == 0) {
    a = FRCRYPT_SYMALG_BLOWFISH;
  } else if (strcasecmp(name, "des3") == 0) {
    a = FRCRYPT_SYMALG_DES3;
  } else if (strcasecmp(name, "desx") == 0) {
    a = FRCRYPT_SYMALG_DESX;
#ifndef FREEDOM_PRODUCTION_BUILD
  } else if (strcasecmp(name, "xor") == 0) {
    a = FRCRYPT_SYMALG_XOR;
  } else if (strcasecmp(name, "idem") == 0) {
    a = FRCRYPT_SYMALG_NONE;
  } else if (strcasecmp(name, "null") == 0) {
    a = FRCRYPT_SYMALG_NONE;
#endif
  }

  if (a == FRCRYPT_SYMALG_UNUSED_) {
    zkErrorSet_va(err, "frCryptSymNameToAlg", ZKERR, MYMOD, ZKEPARMBAD,
                  "unknown symmetric algorithm `%s'", name);
    return 0;
  } else {
    *alg = a;
    return 1;
  }
}


DLLExport int frCryptSymDecrypt(ZkError err, FrCryptSym sym, void *ivec, void *data, size_t len)
{
  zkErrorClear(err);
  
  return symCrypt(err, "frCryptSymDecrypt", sym, 0, ivec, data, len);
}


/* Encrypt using two buffers */
DLLExport int frCryptSymDecryptToBuf(ZkError err, FrCryptSym sym, void *ivec, void *data, void *output, size_t len)
{
  zkErrorClear(err);
  
  return symCryptToBuf(err, "frCryptSymDecrypt", sym, 0, ivec, data, output, len);
}


DLLExport int frCryptSymEncrypt(ZkError err, FrCryptSym sym, void *ivec, void *data, size_t len)
{
  zkErrorClear(err);
  
  return symCrypt(err, "frCryptSymEncrypt", sym, 1, ivec, data, len);
}


/* Encrypt using two buffers */
DLLExport int frCryptSymEncryptToBuf(ZkError err, FrCryptSym sym, void *ivec, void *data, void *output, size_t len)
{
  zkErrorClear(err);

  return symCryptToBuf(err, "frCryptSymEncrypt", sym, 1, ivec, data, output, len);
}


DLLExport int frCryptSymFree(ZkError err, FrCryptSym *sym)
{
  zkErrorClear(err);
  
  if ( ! sym) {
    zkErrorSet(err, "frCryptSymFree", ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    return 0;
  }
  if ( ! *sym) {
    zkErrorSet(err, "frCryptSymFree", ZKERR, MYMOD, ZKEPARMBAD, "*sym is null");
    return 0;
  }

  memset(*sym, 0, sizeof **sym);
  free(*sym); *sym = 0;
  
  return 1;
}


int frCryptSymGetAlg(ZkError err, FrCryptSym sym, FrCryptSymAlg *alg)
{
  zkErrorClear(err);
  
  if ( ! sym) {
    zkErrorSet(err, "frCryptSymGetAlg", ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    return 0;
  }
  if ( ! alg) {
    zkErrorSet(err, "frCryptSymGetAlg", ZKERR, MYMOD, ZKENULLPTR, "alg is null");
    return 0;
  }

  if (sym->alg == FRCRYPT_SYMALG_UNUSED_) {
    zkErrorSet(err, "frCryptSymGetAlg", ZKERR, MYMOD, ZKEBADSTATE,
               "symmetric key is uninitialized");
    return 0;
  }

  *alg = sym->alg;

  return 1;
}


DLLExport int frCryptSymNew(ZkError err, FrCryptSym *sym)
{
  FrCryptSym s;
  
  zkErrorClear(err);
  
  if ( ! sym) {
    zkErrorSet(err, "frCryptSymNew", ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    return 0;
  }

  /* bug: secure memory */
  if ( ! (s = malloc(sizeof *s))) {
    zkErrorSet_va(err, "frCryptSymNew", ZKERR, MYMOD, ZKENOMEM,
                  "can't allocate %lu bytes for symmetric cryption object: %s",
                  (unsigned long)sizeof *s, strerror(errno));
    return 0;
  }

  memset(s, 0, sizeof *s);
  *sym = s;
  
  return 1;
}


/*
 *  Make a `copy' of the symmetric `key'.  This routine allocates the memory
 *  for the copy.
 */
DLLExport int frCryptSymCopy(ZkError err, FrCryptSym key, FrCryptSym *copy)
{
  FrCryptSym cp = 0;
  const char *fn = "frCryptSymCopy";
  int ret = 0;

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "key is null");
    goto done;
  }
  if ( ! copy) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "copy is null");
    goto done;
  }

  if ( ! frCryptSymNew(err, &cp)) {
    goto done;
  }
  cp->alg = key->alg;
  memcpy(&cp->u, &key->u, sizeof cp->u);

  *copy = cp; cp = 0; /* the caller now owns it */

  ret = 1;

 done:

  if (cp) { frCryptSymFree(err, &cp); }
  
  return ret;
}


DLLExport int frCryptSymGetKey(ZkError err, FrCryptSym sym, void *kbuf, size_t kbsz)
{
  const char *fn = "frCryptSymGetKey";
  int ret = 0;

  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    goto done;
  }
  if ( ! kbuf) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "kbuf is null");
    goto done;
  }

  switch (sym->alg) {
#ifndef FREEDOM_PRODUCTION_BUILD

  case FRCRYPT_SYMALG_NONE:
    break;
    
  case FRCRYPT_SYMALG_XOR:
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKESTUB, "can't yet get XOR key");
    goto done;
    break;
    
#endif
    
  case FRCRYPT_SYMALG_DES3:
    ret = fr_Des3GetKey(err, sym, kbuf, kbsz);
    break;

  case FRCRYPT_SYMALG_DESX:
    ret = fr_DesxGetKey(err, sym, kbuf, kbsz);
    break;
    
  case FRCRYPT_SYMALG_BLOWFISH:
    ret = fr_BlowfishGetKey(err, sym, kbuf, kbsz);
    break;

  default:
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "invalid symmetric cryption algorithm (%d)", sym->alg);
    goto done;
    break;
  }

  ret = 1;

 done:

  return ret;
}


DLLExport int frCryptSymGetKeySize(ZkError err, FrCryptSym sym, size_t *sz)
{
  const char *fn = "frCryptSymGetKeySize";
  int ret = 0;
  
  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    goto done;
  }

  switch (sym->alg) {
#ifndef FREEDOM_PRODUCTION_BUILD

  case FRCRYPT_SYMALG_NONE:
    *sz = 0;
    break;
    
  case FRCRYPT_SYMALG_XOR:
    *sz = 1;
    break;
    
#endif
    
  case FRCRYPT_SYMALG_DES3:
    ret = fr_Des3GetKeySize(err, sym, sz);
    break;

  case FRCRYPT_SYMALG_DESX:
    ret = fr_DesxGetKeySize(err, sym, sz);
    break;
    
  case FRCRYPT_SYMALG_BLOWFISH:
    ret = fr_BlowfishGetKeySize(err, sym, sz);
    break;

  default:
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "invalid symmetric cryption algorithm (%d)", sym->alg);
    goto done;
    break;
  }

  ret = 1;

 done:

  return ret;
}


int frCryptSymSetKeySize(ZkError err, FrCryptSym sym, size_t sz)
{
  zkErrorClear(err);
  
  if ( ! sym) {
    zkErrorSet(err, "frCryptSymSetKeySize", ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    return 0;
  }

  switch (sym->alg) {
#ifndef FREEDOM_PRODUCTION_BUILD
  case FRCRYPT_SYMALG_NONE:
  case FRCRYPT_SYMALG_XOR:
#endif
  case FRCRYPT_SYMALG_DES3:
  case FRCRYPT_SYMALG_DESX:
    zkErrorSet_va(err, "frCryptSymSetKeySize", ZKWARN, MYMOD, ZKEBADSTATE,
                  "this encryption algorithm (%d) has a fixed key size", sym->alg);
    return 0;
    
  case FRCRYPT_SYMALG_BLOWFISH:
    fr_BlowfishSetKeySize(err, sym, sz);
    break;

  default:
    zkErrorSet_va(err, "frCryptSymSetKeySize", ZKERR, MYMOD, ZKEPARMBAD,
                  "invalid symmetric cryption algorithm (%d)", sym->alg);
    return 0;
  }

  return 1;
}


DLLExport int frCryptSymRandomKey(ZkError err, FrCryptSym sym)
{
  zkErrorClear(err);
  
  if ( ! sym) {
    zkErrorSet(err, "frCryptSymRandomKey", ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    return 0;
  }

  switch (sym->alg) {
#ifndef FREEDOM_PRODUCTION_BUILD
    
  case FRCRYPT_SYMALG_NONE:
    break;
    
  case FRCRYPT_SYMALG_XOR:
    return fr_XorRandomKey(err, sym);
    
#endif
  
  case FRCRYPT_SYMALG_BLOWFISH:
    return fr_BlowfishRandomKey(err, sym);
    
  case FRCRYPT_SYMALG_DES3:
    return fr_Des3RandomKey(err, sym);
    
  case FRCRYPT_SYMALG_DESX:
    return fr_DesxRandomKey(err, sym);

  default:
    zkErrorSet_va(err, "frCryptSymRandomKey", ZKERR, MYMOD, ZKEPARMBAD,
                  "invalid symmetric cryption algorithm (%d)", sym->alg);
    return 0;
  }

  return 0;
}


DLLExport int frCryptSymKeyFromData(ZkError err, FrCryptSym sym, void *data, size_t len)
{
  /*
   *  bug? maybe this could be made cleaner by making better use of SSLeay's
   *  EVP_ routines.  I.e. putting the EVP_ stuff in the Crypt structure.
   *  
   *  This routine is crap.  It could be made cleaner by throwing out the EVP_
   *  stuff.
   */
  EVP_CIPHER *cipher = 0;
  EVP_MD *md = EVP_sha1();
  size_t keySize;
  unsigned char key[KEY_SIZE_MAX];
  
  zkErrorClear(err);

  if ( ! sym) {
    zkErrorSet(err, "frCryptSymKeyFromData", ZKERR, MYMOD, ZKENULLPTR,
               "sym is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "frCryptSymKeyFromData", ZKERR, MYMOD, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (len == 0) {
    zkErrorSet(err, "frCryptSymKeyFromData", ZKERR, MYMOD, ZKEPARMBAD,
               "len is zero");
    return 0;
  }

  switch (sym->alg) {
  case FRCRYPT_SYMALG_UNUSED_:
    zkErrorSet(err, "frCryptSymKeyFromData", ZKERR, MYMOD, ZKEBADSTATE,
               "symmetric cryption algorithm has not been set");
    return 0;
    
#ifndef FREEDOM_PRODUCTION_BUILD
  
  case FRCRYPT_SYMALG_NONE:
    /* This algorithm doesn't have a `key'. */
    return 1;
    
  case FRCRYPT_SYMALG_XOR:
    /* Set this `key' from the first byte of the data. */
    sym->u.xor.key[0] = *(char *)data;
    return 1;
    
#endif
  
  case FRCRYPT_SYMALG_BLOWFISH:
    cipher = EVP_bf_cbc();
    break;
    
  case FRCRYPT_SYMALG_DES3:
    cipher = EVP_des_ede3_cbc();
    break;
    
  case FRCRYPT_SYMALG_DESX:
    cipher = EVP_desx_cbc();
    break;
    
  default:
    zkErrorSet_va(err, "frCryptSymKeyFromData", ZKERR, MYMOD, ZKESTUB,
                  "unknown cryption algorithm (%d)", sym->alg);
    return 0;
  }

  /*
   *  Sanity check.  Verify that the key array is large enough to store the
   *  key for this algorithm.
   */
  
  if ((keySize = EVP_BytesToKey(cipher, md, 0, 0, 0, 1, 0, 0)) > sizeof key) {
    zkErrorSet_va(err, "frCryptSymKeyFromData", ZKERR, MYMOD, ZKESTUB,
                  "internal key size too small (%d > %lu)",
                  keySize, (unsigned long)sizeof key);
    return 0;
  }
  
  keySize = EVP_BytesToKey(cipher, md, 0 /* no salt */, data, len, 1, key, 0 /* no iv */);

  switch (sym->alg) {
  case FRCRYPT_SYMALG_UNUSED_:
    /* The previous switch should pick this up... */
    zkErrorSet(err, "frCryptSymKeyFromData", ZKERR, MYMOD, ZKEBADSTATE,
               "symmetric cryption algorithm has not been set!");
    return 0;
    
#ifndef FREEDOM_PRODUCTION_BUILD
  
  case FRCRYPT_SYMALG_NONE:
  case FRCRYPT_SYMALG_XOR:
    zkErrorSet(err, "frCryptSymKeyFromData", ZKERR, MYMOD, ZKEBADSTATE,
               "this statement should not be reached!");
    return 0;

#endif
    
  case FRCRYPT_SYMALG_BLOWFISH:
    memcpy(sym->u.bfish.key, key, keySize);
    BF_set_key(&sym->u.bfish.expKey, sym->u.bfish.keySize, sym->u.bfish.key);
    break;
    
  case FRCRYPT_SYMALG_DES3:
    if (keySize < 24) {
      zkErrorSet_va(err, "frCryptSymKeyFromData", ZKERR, MYMOD, ZKEPARMBAD,
                    "DES3 key data too short (%d < 24)", keySize); /* bug bug bug */
      return 0;
    }
    memcpy(sym->u.des3.key1, key,      8);
    memcpy(sym->u.des3.key2, key +  8, 8);
    memcpy(sym->u.des3.key3, key + 16, 8);
    (void)des_set_key(&sym->u.des3.key1, sym->u.des3.ks1);
    (void)des_set_key(&sym->u.des3.key2, sym->u.des3.ks2);
    (void)des_set_key(&sym->u.des3.key3, sym->u.des3.ks3);
    break;
    
  case FRCRYPT_SYMALG_DESX:
    if (keySize < 24) {
      zkErrorSet_va(err, "frCryptSymKeyFromData", ZKERR, MYMOD, ZKEPARMBAD,
                    "DESX key data too short (%d < 24)", keySize); /* bug bug bug */
      return 0;
    }
    memcpy(sym->u.desx.key,  key,      8);
    memcpy(sym->u.desx.inw,  key +  8, 8);
    memcpy(sym->u.desx.outw, key + 16, 8);
    (void)des_set_key(&sym->u.desx.key, sym->u.desx.ks);
    break;
    
  default:
    zkErrorSet_va(err, "frCryptSymKeyFromData", ZKERR, MYMOD, ZKESTUB,
                  "unknown cryption algorithm (%d)", sym->alg);
    return 0;
  }

  return 1;
}


DLLExport int frCryptSymSetAlg(ZkError err, FrCryptSym sym, FrCryptSymAlg alg)
{
  int ret = 1;

  zkErrorClear(err);
  
  if ( ! sym) {
    zkErrorSet(err, "frCryptSymSetAlg", ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    return 0;
  }

  switch (alg) {
#ifndef FREEDOM_PRODUCTION_BUILD
  
  case FRCRYPT_SYMALG_NONE:
    break;

  case FRCRYPT_SYMALG_XOR:
    ret = fr_XorInit(err, sym);
    break;

#endif
  
  case FRCRYPT_SYMALG_BLOWFISH:
    ret = fr_BlowfishInit(err, sym);
    break;
    
  case FRCRYPT_SYMALG_DES3:
    ret = fr_Des3Init(err, sym);
    break;
    
  case FRCRYPT_SYMALG_DESX:
    ret = fr_DesxInit(err, sym);
    break;
    
  default:
    zkErrorSet_va(err, "frCryptSymSetAlg", ZKERR, MYMOD, ZKEPARMBAD,
                  "invalid symmetric cryption algorithm (%d)", sym->alg);
    /* bug: stdargs addition to zkErrorSet */
    return 0;
  }

  sym->alg = alg;
  
  return ret;
}


/*
 *  The next two routines load and store compact, fixed size representations
 *  of specific symmetric encryption/decryption algorithms, with associated
 *  keys.
 */

/*
 *  Initialize `sym' from the information stored in `*data'.  `Len' specifies
 *  the number of valid bytes in `*data'.  If successful, `data' is changed to
 *  point to the next available location.
 */
DLLExport int frCryptSymLoad(ZkError err, FrCryptSym sym, void *data, size_t len)
{
  FrCryptSymAlg alg;
  int ret = 1;

  zkErrorClear(err);
  
  if ( ! sym) {
    zkErrorSet(err, "frCryptSymLoad", ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "frCryptSymLoad", ZKERR, MYMOD, ZKENULLPTR, "data is null");
    return 0;
  }
  
  if (len == 0) {
    zkErrorSet(err, "frCryptSymLoad", ZKERR, MYMOD, ZKEPARMBAD, "len is zero");
    return 0;
  }

  alg = *(char *)data;

  switch (alg) {
#ifndef FREEDOM_PRODUCTION_BUILD
  
  case FRCRYPT_SYMALG_NONE:
    break;
    
  case FRCRYPT_SYMALG_XOR:
    ret = fr_XorLoad(err, sym, data, len);
    break;

#endif
  
  case FRCRYPT_SYMALG_BLOWFISH:
#if 0
    ret = fr_BlowfishLoad(err, sym, data, len);
    break;
#endif    
  case FRCRYPT_SYMALG_DES3:
  case FRCRYPT_SYMALG_DESX:
    zkErrorSet_va(err, "frCryptSymLoad", ZKERR, MYMOD, ZKESTUB,
                  "this algorithm (%d) is not yet implemented", alg);
    return 0;

  default:
    zkErrorSet_va(err, "frCryptSymLoad", ZKERR, MYMOD, ZKEPARMBAD,
                  "unknown encryption algoritm (%d)", alg);
    return 0;
  }

  return ret;
}


/*
 *  Copy, to `data', a compact and portable representation of the symmetric
 *  key represented by `sym'.  `*len' is the initial length of the data
 *  buffer.  It's updated with the actual amount of data stored.
 *  
 *  bug: this routine is less general than it should be, since it doesn't
 *  allow for variable length keys.
 */
DLLExport int frCryptSymStore(ZkError err, FrCryptSym sym, void *data, size_t *len)
{
  char *d = data;
  const char *fn = "frCryptSymStore";
  int ret = 1;
  size_t l;

  zkErrorClear(err);
  
  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "data is null");
    return 0;
  }
  if ( ! len) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "len is null");
    return 0;
  }
  
  if ((l = *len) < 1) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "*len too short to store symmetric algorithm (%lu < %d)",
                  (unsigned long)*len, 1);
    return 0;
  }

  d[0] = sym->alg; l--;

  switch (sym->alg) {
#ifndef FREEDOM_PRODUCTION_BUILD

  case FRCRYPT_SYMALG_NONE:
    /* bug: make up some value */
    break;
    
  case FRCRYPT_SYMALG_XOR:
    ret = fr_XorStore(err, sym, d+1, &l);
    break;

#endif
  
  case FRCRYPT_SYMALG_BLOWFISH:
    ret = fr_BlowfishStore(err, sym, d+1, &l);
    break;

  case FRCRYPT_SYMALG_DES3:
    ret = fr_Des3Store(err, sym, d+1, &l);
    break;

  case FRCRYPT_SYMALG_DESX:
    ret = fr_DesxStore(err, sym, d+1, &l);
    break;

  default:
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "unknown encryption algoritm (%d)", sym->alg);
    return 0;
  }

  *len = l + 1;

  return ret;
}


/*
 *  Return the number of bytes required to store the portable, compact
 *  representation of a symmetric encryption/decryption algorithm with
 *  associated keys.
 */
DLLExport int frCryptSymPortableSize(ZkError err, FrCryptSym sym, size_t *sz)
{
  zkErrorClear(err);
  
  if ( ! sym) {
    zkErrorSet(err, "frCryptSymPortableSize", ZKERR, MYMOD, ZKENULLPTR,
               "sym is null");
    return 0;
  }
  if ( ! sz) {
    zkErrorSet(err, "frCryptSymPortableSize", ZKERR, MYMOD, ZKENULLPTR,
               "sz is null");
    return 0;
  }
  
  *sz = FRCRYPT_SYM_PORTABLE_SIZE;

  return 1;
}


/*
 *  Return, through `sz', the size of the smallest block that may be
 *  encrypted.  That is, the length of anything to be encrypted with a
 *  symmetric algorithm must be a multiple of this value.
 */
DLLExport int frCryptSymSizeMultiple(ZkError err, FrCryptSym sym, size_t *sz)
{
  zkErrorClear(err);
  
  if ( ! sym) {
    zkErrorSet(err, "frCryptSymSizeMultiple", ZKERR, MYMOD, ZKENULLPTR,
               "sym is null");
    return 0;
  }
  if ( ! sz) {
    zkErrorSet(err, "frCryptSymSizeMultiple", ZKERR, MYMOD, ZKENULLPTR,
               "sz is null");
    return 0;
  }
  
  *sz = MIN_CRYPT_SIZE;

  return 1;
}

#if defined( ZK_LINUX ) || defined( ZK_SOLARIS )
/*
 * Create a new Symetric key using a virtual heap as memory allocator.
 */
DLLExport int frCryptSymVHeapNew
 (
  ZkError     err,        /* [in]  ZkError */ 
  FrCryptSym *symP,       /* [out] return FrCryptSym key */
  ZkVHeap     vheap       /* [in]  a ZkVHeap allocator */
 )
{
  FrCryptSym s = NULL;
  
  zkErrorClear(err);
  
  if ( !symP ) {
    zkErrorSet(err, "frCryptSymVHeapNew", ZKERR, MYMOD, ZKENULLPTR, "symP is null");
    return 0;
  }
  else if ( !vheap ) {
    zkErrorSet(err, "frCryptSymVHeapNew", ZKERR, MYMOD, ZKENULLPTR, "vheap is null");
    return 0;
  }

  if( !zkVHeapMemoryAlloc(err, vheap, (void **)&s, sizeof *s) ) {
    zkErrorSet_va(err, "frCryptSymVHeapNew", ZKERR, MYMOD, ZKENOMEM,
                  "zkVHeapMemoryAlloc:can't allocate %lu bytes for symmetric encryption object",
                  (unsigned long)sizeof *s);
    return 0;
  }

  memset(s, 0, sizeof *s);

  *symP = s;
  
  return 1;
}

/*
 * Free a new Symetric key allocated from a virtual heap
 */
DLLExport int frCryptSymVHeapFree
 (
   ZkError     err,        /* [in]     ZkError */ 
   FrCryptSym *symP,       /* [in/out] a FrCryptSym key allocated with the folllowing vheap */
   ZkVHeap     vheap       /* [in]     a ZkVHeap allocator */
  )
{
  FrCryptSym s = NULL;

  zkErrorClear(err);
  
  if ( !symP ) {
    zkErrorSet(err, "frCryptSymVHeapFree", ZKERR, MYMOD, ZKENULLPTR, "symP is null");
    return 0;
  }
  else if ( ! *symP) {
    zkErrorSet(err, "frCryptSymVHeapFree", ZKERR, MYMOD, ZKEPARMBAD, "*symP is null");
    return 0;
  }
  else if ( !vheap ) {
    zkErrorSet(err, "frCryptSymVHeapFree", ZKERR, MYMOD, ZKENULLPTR, "vheap is null");
    return 0;
  }

  s = *symP;

  memset(s, 0, sizeof *s);

  if( !zkVHeapMemoryFree(err, vheap, (void **)&s) ) {
    zkErrorSet_va(err, "frCryptSymVHeapFree", ZKERR, MYMOD, ZKERROR,
                  "zkVHeapMemoryFree:can't free %p from symmetric encryption object", symP);
    return 0;
  }
  *symP = NULL;

  return 1;
}

#endif /*#if defined( ZK_LINUX ) || defined( ZK_SOLARIS )*/

/*
struct fr_crypt_sym {
  FrCryptSymAlg alg;
  union {
    struct desx desx;
    struct des3 des3;
    struct blowfish_v1 bfish;
    struct xor xor;
    struct idem idem;
  } u;
};
*/


/*
 * Dump a SymKey to a buffer in a literal form
 */
DLLExport int frCryptSymDebugDump
(
 ZkError     err,        /* [in]     ZkError */ 
 FrCryptSym  sym,        /* [in]     a FrCryptSym key */
 char       *buffer,     /* [in/out] output buffer */
 uint16_t    buffersz    /* [in]     length of the buffer (in bytes) */
 )
{
#ifndef FREEDOM_PRODUCTION_BUILD

  const char *algo                    = NULL;
  char       *key                     = NULL;
  uint32_t    keysz                   = 0;
  char        hash[FRCRYPT_HASH_SIZE] = { 0 };
  
  if ( !sym ) {
    zkErrorSet(err, "frCryptSymDebugDump", ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    return 0;
  }
  else {
    switch( sym->alg ) 
      {
      case FRCRYPT_SYMALG_UNUSED_: algo = "UNUSED";   break;
      case FRCRYPT_SYMALG_NONE:    algo = "NONE";     break;
      case FRCRYPT_SYMALG_XOR:     algo = "XOR";      key = (char *)&sym->u.xor,   keysz = sizeof(struct xor);         break;
      case FRCRYPT_SYMALG_BLOWFISH:algo = "BLOWFISH"; key = (char *)&sym->u.bfish, keysz = sizeof(struct blowfish_v1); break;
      case FRCRYPT_SYMALG_DES3:    algo = "DES3";     key = (char *)&sym->u.des3,  keysz = sizeof(struct des3);        break;
      case FRCRYPT_SYMALG_DESX:    algo = "DESX";     key = (char *)&sym->u.desx,  keysz = sizeof(struct desx);        break;
      }
    if( algo ) {
      if( key ) {
        if( !frCryptHashGenerate(err, (void *)key, (size_t)keysz, hash, FRCRYPT_HASH_SIZE) ) {
          zkErrorSet(err, "frCryptSymDebugDump", ZKERR, MYMOD, ZKENULLPTR, "frCryptHashGenerate failled");
          return 0;
        }
        
        snprintf(buffer, buffersz,
                 "{symkey: algo:[%s], "
                 "header:[%02x-%02x-%02x-%x-%02x-%02x-%02x-%02x-%02x-%02x], "
                 "hash:[%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x-%02x]}",
                 algo,
                 key[0]&0xff, key[1]&0xff, key[2]&0xff, key[3]&0xff, key[4]&0xff, 
                 key[5]&0xff, key[6]&0xff, key[7]&0xff, key[8]&0xff, key[9]&0xff,
                 hash[0]&0xff,hash[1]&0xff,hash[2]&0xff,hash[3]&0xff,hash[4]&0xff,
                 hash[5]&0xff,hash[6]&0xff,hash[7]&0xff,hash[8]&0xff,hash[9]&0xff,
                 hash[10]&0xff,hash[11]&0xff,hash[12]&0xff,hash[13]&0xff,hash[14]&0xff,
                 hash[15]&0xff,hash[16]&0xff,hash[17]&0xff,hash[18]&0xff,hash[19]&0xff);  
      }
      else {
        snprintf(buffer, buffersz, "{symkey: algo:[%s]}", algo);
      }
    }
    else {
      snprintf(buffer, buffersz, "{symkey: << INVALID ALGORITH >>}");
    }
  }
  
#else
  
  (void) err;
  (void) sym;

  if (buffersz > 0) {
    buffer[0] = '\0';
  }
  
#endif

  return 1;
}
