/*
 * FrAipStatus DBO Driver routines
 * $Id: frDBOTypeAipStatus.c,v 1.5.28.1 2000/10/11 13:17:53 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zks.net>

 *
 */
#ifndef WIN32
#include "zks.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkDefs.h>
#include <zkDump.h>
#include <zkDebug.h>
#include <zkDBO.h>
#include <zkDBOTypes.h>
#include <zkDBODefs.h>
#include <frDBOTypes.h>

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNDB

#define DBODS_REGID ZKDBO_REG_FRAIPSTATUS
#define DBODS_T T_FrDBOTypeAipStatus

typedef struct DBODS_T *DBODS;

static int DSnew (ZkError err, uint32_t version, void **ds, size_t *dsz)
{
  static const char *funcname = "DSnew";

  DBODS d;

  (void) version;

  if (!dsz || !ds) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers passed to function");
    return 0;
  }
  
  *dsz = sizeof(struct DBODS_T);
  zkMALLOC (d, DBODS, sizeof(struct DBODS_T));

  d->regID = DBODS_REGID;
  *ds = (void *)d;
  return 1;
}

static int DSfree (ZkError err, void **ds, size_t *dsz)
{
  static const char *funcname = "DSfree";
  
  if (!dsz || !ds || !*ds) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers passed to function");
    return 0;
  }
  
  memset (*ds, 0, *dsz);
  free (*ds);
  *ds = NULL;
  *dsz = 0;
  return 1;
}

/**
 * Passing a null pointer to *pobj will create the buffer you need.
 * On failure, you'll have to take care of cleaning it up.
 */
static int DSpack (ZkError err, void *ds, size_t dsz, void **pds, size_t *psz)
{
  static const char *funcname = "DSpack";
  DBODS d;
  size_t pos;
  
  if ( !ds || !dsz || !pds || !psz ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers or values passed to function");
    return 0;
  }
  d = ds;

  if (!zkDBOPackedInit(err, d->regID, pds, psz, &pos))
    return 0;

  if (!zkDBOPackField(err, d->regID, "regID", (void *) & d->regID,
		      sizeof(d->regID), pds, psz, &pos)) {
    return 0;
  }

  if (!zkDBOPackField(err, d->regID, "status", (void *) &d->status,
		      sizeof(d->status), pds, psz, &pos)) {
    return 0;
  }

  return 1;
}

/**
 * Passing a null pointer to *ds will create the buffer you need.
 * On failure, you'll have to take care of cleaning it up.
 */
static int DSunpack(ZkError err, void **ds, size_t *dsz, void *pds, size_t psz)
{
  static const char *funcname = "DSunpack";
  size_t s, pos=0;
  int regID;
  void *pd;
  struct DBODS_T *d;
  
  if ( !ds || !dsz || !pds || !psz) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
		  "Null pointers or values passed to function");
    return 0;
  }

  if (!zkDBOUnpackedInit(err, &regID, ds, dsz, pds, psz, &pos))
    return 0;

  d = *ds;
    
  s = sizeof(d->regID);
  pd = &d->regID;
  if (!zkDBOUnpackField(err, regID, "regID", &pd,
		 &s, pds, psz, &pos))
    return 0;

  s = sizeof(d->status);
  pd = &d->status;
  if (!zkDBOUnpackField(err, regID, "status", &pd,
		 &s, pds, psz, &pos))
    return 0;

  return 1;
}

#ifndef DSdump
static int DSdump (ZkError err, void *ds)
{
  static const char *funcname = "DSdump";
  DBODS d;
  
  if ( !ds ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers or values passed to function");
    return 0;
  }
  d = (DBODS) ds;
  zkDebug("**Dumping Registered Datastructure for AipStatus***");
  zkDebug("regID : %lu", (unsigned long) d->regID);
  zkDebug("status: %lu", (unsigned long) d->status);
  
  return 1;
}
#endif

DLLExport
int frDBODriverInitAipStatus(ZkError err)
{
  ZkDBODesc dboDesc;
  ZkDBODriver dboDrv;
  size_t sz, psz;
  
  if (!zkDBODescNew(err, 2, &dboDesc))
    return 0;

  if (!zkDBODescSetField(err, dboDesc, 0, "regID", FT_UINT32, &sz, &psz)
      || !zkDBODescSetField(err, dboDesc, 1, "status", FT_UINT32, &sz, &psz))
    return 0;

  if (!zkDBODriverNew (err, &dboDrv))
    return 0;

  if (!zkDBODriverInit (err, DBODS_REGID, dboDrv,
			DSnew, DSfree, DSpack, DSunpack,
			NULL, NULL, DSdump, NULL))
    return 0;

  if (!zkDBODescRegister(err, dboDesc, dboDrv))
    return 0;

# ifndef NDEBUG
  if (!zkDBODescDump(err, dboDesc))
    return 0;
# endif  
  return 1;
}

#endif 
