/*
 * NTDB Detailed Neighbour DBO Driver routines
 *
 * $Id: frDBOTypeNtdbDetailedNeighbours.c,v 1.11.16.3 2000/12/14 17:52:44 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zeroknowledge.com>

 *
 *
 */

#include "zks.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "zkCStr.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkDefs.h"
#include "zkAlloc.h"
#include "zkDump.h"
#include "zkDebug.h"
#include "zkFile.h"
#include "frDBOTypes.h"

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNDB

#define DBODS_REGID ZKDBO_REG_FRNTDB_DETAILED_NEIBR
#define DBODS_T     T_FrDBOTypeNtdbDetailedNeighbours
#define DBODS_AT    struct neibr_detail

typedef struct DBODS_T *DBODS;
/**
 * Warning - Overloaded *dsz. Holds number of elements
 */
static int DSnew (ZkError err, uint32_t version, void **ds, size_t *dsz)
{
  static const char *funcname = "DSnew";
  /* Special handing of DS */
  size_t num = *dsz;

  DBODS d;

  (void) version;

  if (!dsz || !ds) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers passed to function");
    return 0;
  }
  if (num == 0) num = 1;
  
  *dsz = sizeof(struct DBODS_T) +
    (sizeof(struct neibr_detail) * (num-1));
  zkMALLOC (d, DBODS, *dsz);

  d->regID = DBODS_REGID;
  *ds = (void *)d;
  return 1;
}

static int DSfree (ZkError err, void **ds, size_t *dsz)
{
  static const char *funcname = "DSfree";
  
  if (!dsz || !ds || !*ds) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers passed to function");
    return 0;
  }
  
  memset (*ds, 0, *dsz);
  free (*ds);
  *ds = NULL;
  *dsz = 0;
  return 1;
}

/**
 * Passing a null pointer to *pobj will create the buffer you need.
 * On failure, you'll have to take care of cleaning it up.
 */
static int DSpack (ZkError err, void *ds, size_t dsz, void **pds, size_t *psz)
{
  static const char *funcname = "DSpack";
  DBODS d = NULL;
  size_t pos = 0;
  size_t i = 0, min_size = 0;
  
#define DBO_PACK_ENT2(w,x) \
{ \
  void *_buf; size_t _sz = FRENT_PORTABLE_SIZE; \
  zkMALLOC (_buf, void *, _sz); \
  if (!frEntStore(err, &d->x, _buf, _sz)) { \
    free (_buf); \
    return 0; \
  } \
  if (!zkDBOPackField(err, d->regID, #w, _buf, _sz, pds, psz, &pos)) { \
    free (_buf); \
    return 0; \
  } \
  free(_buf); \
}  

#define DBO_PACK(x) \
  if (!zkDBOPackField(err, d->regID, #x, (void *) & d->x, \
		      sizeof(d->x), pds, psz, &pos)) { \
    return 0; \
  }

#define DBO_PACK2(w, x) \
  if (!zkDBOPackField(err, d->regID, #w, (void *) & d->x, \
		      sizeof(d->x), pds, psz, &pos)) { \
    return 0; \
  }

  if ( !ds || !dsz || !pds || !psz ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers or values passed to function");
    return 0;
  }
  d = ds;

  /* Now we allocate additional space for us */
  if (!zkDBOPackedReqSpace(err, d->regID, &min_size))
    return 0;

  *psz = min_size + ((d->num_neighbours - 1) * (sizeof(DBODS_AT)+4));

  zkDebug ("Adjusted packing size is now %d", *psz);
  
  if (!zkDBOPackedInit(err, d->regID, pds, psz, &pos))
    return 0;

  DBO_PACK(regID);
  DBO_PACK(num_neighbours);
  DBO_PACK(dh_port);
  DBO_PACK(aip_port);
  DBO_PACK_ENT2(ent, ent);
  for (i = 0; i < d->num_neighbours; i++) {
    DBO_PACK_ENT2(n_ent, neighbour[i].ent);
    DBO_PACK2(n_ipaddr, neighbour[i].ipaddr);
    DBO_PACK2(n_dh_port, neighbour[i].dh_port);
    DBO_PACK2(n_aip_port, neighbour[i].aip_port);
    DBO_PACK2(n_dh_timeout, neighbour[i].dh_timeout);
    DBO_PACK2(n_dh_negotiation, neighbour[i].dh_negotiation);
    DBO_PACK2(n_ts_delay, neighbour[i].ts_delay);
    DBO_PACK2(n_ts_pkts, neighbour[i].ts_pkts);
    DBO_PACK2(n_ct_on_flag, neighbour[i].ct_on_flag);
    DBO_PACK2(n_spare1, neighbour[i].spare2);
    DBO_PACK2(n_spare2, neighbour[i].spare1);
  }
  return 1;
}


/**
 * Passing a null pointer to *ds will create the buffer you need.
 * On failure, you'll have to take care of cleaning it up.
 */
static int DSunpack(ZkError err, void **ds, size_t *dsz, void *pds, size_t psz)
{
  static const char *funcname = "DSunpack";
  size_t s = 0, pos=0, i = 0;
  int regID = 0;
  void *pd = NULL;
  struct DBODS_T *d = NULL;
  
#define DBO_UNPACK(x) \
  s = sizeof(d->x); \
  pd = &(d->x); \
  if (!zkDBOUnpackField(err, regID, #x, &pd, \
		 &s, pds, psz, &pos)) \
    return 0;

#define DBO_UNPACK2(w,x) \
  s = sizeof(d->x); \
  pd = &(d->x); \
  if (!zkDBOUnpackField(err, regID, #w, &pd, \
		 &s, pds, psz, &pos)) \
    return 0;

#define DBO_UNPACK_ENT2(w,x) \
{ \
  void *_ebuf = NULL; \
  size_t _esz = FRENT_PORTABLE_SIZE; \
  \
  if (!zkDBOUnpackField(err, regID, #w, &_ebuf, \
		 &_esz, pds, psz, &pos)) \
    return 0; \
  if (zkDebugIsOn()) \
    zkDumpBuffer(err,_ebuf,FRENT_PORTABLE_SIZE,zkDumpToStream,stderr); \
  if (!frEntLoad(err, &d->##x, _ebuf, _esz)) { \
    free(_ebuf); \
    return 0; \
  } \
  free(_ebuf); \
} 

  if ( !ds || !dsz || !pds || !psz) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
		  "Null pointers or values passed to function");
    return 0;
  }

  if (!zkDBOUnpackedInit(err, &regID, ds, dsz, pds, psz, &pos))
    return 0;

  d = *ds;

  DBO_UNPACK(regID);
  DBO_UNPACK(num_neighbours);
  DBO_UNPACK(dh_port);
  DBO_UNPACK(aip_port);
  DBO_UNPACK_ENT2(ent,ent);
  if (d->num_neighbours > 1) {
    /* Need to reallocate space */
    *dsz += ( d->num_neighbours - 1 ) * sizeof (DBODS_AT);
    if (! zkMemRealloc (err, ds, *dsz)) {
      free (*ds);
      return 0;
    }
    zkDebug("Realloced space to %d bytes", *dsz);
    d = *ds; /* Just in case */
  }
  
  for (i = 0; i < d->num_neighbours; i++) {
    DBO_UNPACK_ENT2(n_ent, neighbour[i].ent);
    DBO_UNPACK2(n_ipaddr, neighbour[i].ipaddr);
#if 0    
#ifdef ZK_SOLARIS
#warning DO NOT BUILD A SOLARIS NIDB
#warning REMOVE THIS HACK ASAP
    {
      uint32_t x = d->neighbour[i].ipaddr;
      x = ((x & 0xFF000000)/0x1000000) + ((x & 0xFF0000)/0x100)
        + ((x&0xFF00)*0x100) + ((x&0xFF)*0x1000000);
      d->neighbour[i].ipaddr = x;
    }
#endif
#endif
    
    DBO_UNPACK2(n_dh_port, neighbour[i].dh_port);
    DBO_UNPACK2(n_aip_port, neighbour[i].aip_port);
    DBO_UNPACK2(n_dh_timeout, neighbour[i].dh_timeout);
    DBO_UNPACK2(n_dh_negotiation, neighbour[i].dh_negotiation);
    DBO_UNPACK2(n_ts_delay, neighbour[i].ts_delay);
    DBO_UNPACK2(n_ts_pkts, neighbour[i].ts_pkts);
    DBO_UNPACK2(n_ct_on_flag, neighbour[i].ct_on_flag);
    DBO_UNPACK2(n_spare1, neighbour[i].spare2);
    DBO_UNPACK2(n_spare2, neighbour[i].spare1);
  }
  
  return 1;
}

#ifndef DSdump
static int DSdump (ZkError err, void *ds)
{
  static const char *funcname = "DSdump";
  DBODS d;
  size_t i; 
  
  if ( !ds ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers or values passed to function");
    return 0;
  }
  d = (DBODS) ds;
  zkDebug("**Dumping Registered Datastructure for Extened Topology Configuraion ***");

#define DBO_DUMP(x)  zkDebug("%-16s: %lu", #x, (unsigned long) d->x);
#define DBO_DUMP_ENT2(w, x)  zkDebug("%-16s: %s/%d", #w, d->##x##.name,d->##x##.type);
#define DBO_DUMP_ENT3(i, w, x)  zkDebug("%2d-> %-16s: %s/%d", i, #w, d->##x##.name,d->##x##.type);
#define DBO_DUMP2(w, x)  zkDebug("%-16s: %lu", #w, (unsigned long) d->x);
#define DBO_DUMP3(i, w, x)  zkDebug("%2d-> %-16s: %lu", i, #w, (unsigned long) d->x);
  
  DBO_DUMP(regID);
  DBO_DUMP(num_neighbours);
  DBO_DUMP(dh_port);
  DBO_DUMP(aip_port);
  DBO_DUMP_ENT2(ent, ent);
  for (i = 0; i < d->num_neighbours; i++) {
    DBO_DUMP_ENT3(i,n_ent, neighbour[i].ent);
    DBO_DUMP3(i,n_ipaddr, neighbour[i].ipaddr);
    DBO_DUMP3(i,n_dh_port, neighbour[i].dh_port);
    DBO_DUMP3(i,n_aip_port, neighbour[i].aip_port);
    DBO_DUMP3(i,n_dh_timeout, neighbour[i].dh_timeout);
    DBO_DUMP3(i,n_dh_negotiation, neighbour[i].dh_negotiation);
    DBO_DUMP3(i,n_ts_delay, neighbour[i].ts_delay);
    DBO_DUMP3(i,n_ts_pkts, neighbour[i].ts_pkts);
    DBO_DUMP3(i,n_ct_on_flag, neighbour[i].ct_on_flag);
    DBO_DUMP3(i,n_spare1, neighbour[i].spare2);
    DBO_DUMP3(i,n_spare2, neighbour[i].spare1);
  }

  return 1;
}

static int DStest (ZkError err)
{
  DBODS d1 = NULL;
  DBODS d2 = NULL;
  void *pds = NULL , *ds2_p = NULL;
  size_t dsz1 = 0 , dsz2 = 0 , pdsz = 0 ;
  int retcode = 0;

  dsz1 = sizeof(struct DBODS_T);
  dsz1 = 3 ; /* Neighbours */ 
  zkDBONew(err, DBODS_REGID, (void **) &d1, &dsz1);
  zkDBONew(err, DBODS_REGID, (void **) &d2, &dsz2); /*Let's see it adjust itself*/

  /*
   * Do any special initialization here
   */
  d1->num_neighbours = 3;
  d1->dh_port = 0x4444;
  d1->aip_port = 0x4445;
  d1->ent.type = FRENT_TYPE_AIP;
  zkCStrlcpy(d1->ent.name,"Hellspawn",sizeof d1->ent.name);

  d1->neighbour[0].ent.type = FRENT_TYPE_AIP;
  zkCStrlcpy(d1->neighbour[0].ent.name, "Dis",sizeof d1->neighbour[0].ent.name);
  d1->neighbour[0].ipaddr = 11223344;
  d1->neighbour[0].dh_port = 6661;
  d1->neighbour[0].aip_port = 6662;
  d1->neighbour[0].dh_timeout = 0x1111;
  d1->neighbour[0].dh_negotiation = 0x1111;
  d1->neighbour[0].ts_delay = 0x1111;
  d1->neighbour[0].ts_pkts = 0x1111;
  d1->neighbour[0].ct_on_flag = 1;
  d1->neighbour[0].spare1 = 0x2222;
  d1->neighbour[0].spare2 = 0x3333;

  d1->neighbour[1].ent.type = FRENT_TYPE_AIP;
  zkCStrlcpy(d1->neighbour[1].ent.name, "Gehenna",sizeof d1->neighbour[1].ent.name);
  d1->neighbour[1].ipaddr = 11221122;
  d1->neighbour[1].dh_port = 6663;
  d1->neighbour[1].aip_port = 6664;
  d1->neighbour[1].dh_timeout = 0x2222;
  d1->neighbour[1].dh_negotiation = 0x2222;
  d1->neighbour[1].ts_delay = 0x2222;
  d1->neighbour[1].ts_pkts = 0x2222;
  d1->neighbour[1].ct_on_flag = 1;
  d1->neighbour[1].spare1 = 0x2222;
  d1->neighbour[1].spare2 = 0x3333;

  d1->neighbour[2].ent.type = FRENT_TYPE_AIP;
  zkCStrlcpy(d1->neighbour[2].ent.name, "Abyss",sizeof d1->neighbour[2].ent.name);
  d1->neighbour[2].ipaddr = 44445555;
  d1->neighbour[2].dh_port = 6665;
  d1->neighbour[2].aip_port = 6666;
  d1->neighbour[2].dh_timeout = 0x3333;
  d1->neighbour[2].dh_negotiation = 0x3333;
  d1->neighbour[2].ts_delay = 0x3333;
  d1->neighbour[2].ts_pkts = 0x3333;
  d1->neighbour[2].ct_on_flag = 0;
  d1->neighbour[2].spare1 = 0x2222;
  d1->neighbour[2].spare2 = 0x3333;

  /*
   * The rest should be the same
   */
  DSdump(err, d1);
  	 
  zkDebug("Packing datastructure");
  if (!zkDBOPack (err, d1, dsz1, &pds, &pdsz))
    goto failed;

  /* dsz2 = sizeof(struct DBODS_T); */
  ds2_p = (void *) d2;

  zkDebug("Unpacking datastructure");
  if (!zkDBOUnpack (err, &ds2_p, &dsz2, pds, pdsz))
    goto failed;

  /* Because a realloc might happen, let's track it.. */
  d2 = (DBODS) ds2_p;
     

  if (dsz1 != dsz2) {
    zkDebug("Data structure sizes have changed (%u -> %u)", dsz1, dsz2);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d1,dsz1,zkDumpToStream,stderr);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d2,dsz2,zkDumpToStream,stderr);
    goto failed;

  }

  if (memcmp(d1, d2, dsz1) != 0) {
    zkDebug("Data structure improperly copied");
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d1,dsz1,zkDumpToStream,stderr);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d2,dsz2,zkDumpToStream,stderr);
    goto failed;
  }

  retcode = 1;
  
 failed:
  if (retcode) {
    zkDebug("Self test passed");
  } else {
    zkDebug("Self test FAILED");
  }
  FREEIT (d1);
  FREEIT (d2);
  FREEIT (pds);
  return retcode;
  
}

#else
#define DSdump NULL
#define DStest NULL
#endif


DLLExport
int frDBODriverInitNtdbDetailedNeighbours(ZkError err)
{
  ZkDBODesc dboDesc;
  ZkDBODriver dboDrv;
  size_t sz, psz;
  size_t entsz = sizeof(FrEntIdExp);
  size_t entpsz = FRENT_PORTABLE_SIZE;
  size_t entsz2 = sizeof(FrEntIdExp);
  size_t entpsz2 = FRENT_PORTABLE_SIZE;

#define NUMFIELDS 16

  if (!zkDBODescNew(err, NUMFIELDS, &dboDesc))
    return 0;

  /* Bug: Fix the userdefs */
  /* Bug: Problem with recording sizes of BUF and STR */
  if (!zkDBODescSetField(err, dboDesc, 0, "regID", FT_UINT32, &sz, &psz)
      ||!zkDBODescSetField(err, dboDesc, 1, "num_neighbours",FT_UINT32, &sz, &psz)
      ||!zkDBODescSetField(err, dboDesc, 2, "dh_port", FT_UINT32, &sz, &psz)
      ||!zkDBODescSetField(err, dboDesc, 3, "aip_port", FT_UINT32,&sz, &psz)
      ||!zkDBODescSetField(err, dboDesc, 4, "ent", FT_USRDEF,  &entsz, &entpsz)
      
      ||!zkDBODescSetField(err, dboDesc, 5, "n_ent", FT_USRDEF,  &entsz2, &entpsz2)
      ||!zkDBODescSetField(err, dboDesc, 6, "n_ipaddr", FT_IPADDR, &sz, &psz)
      ||!zkDBODescSetField(err, dboDesc, 7, "n_dh_port", FT_UINT32, &sz, &psz)
      ||!zkDBODescSetField(err, dboDesc, 8, "n_aip_port", FT_UINT32, &sz, &psz)
      ||!zkDBODescSetField(err, dboDesc, 9, "n_dh_timeout", FT_UINT32, &sz, &psz)
      ||!zkDBODescSetField(err, dboDesc,10, "n_dh_negotiation",FT_UINT32,&sz,&psz)
      ||!zkDBODescSetField(err, dboDesc,11, "n_ts_delay", FT_UINT32, &sz, &psz)
      ||!zkDBODescSetField(err, dboDesc,12, "n_ts_pkts", FT_UINT32, &sz, &psz)
      ||!zkDBODescSetField(err, dboDesc,13, "n_ct_on_flag", FT_UINT32, &sz, &psz)
      ||!zkDBODescSetField(err, dboDesc,14, "n_spare1", FT_UINT32, &sz, &psz)
      ||!zkDBODescSetField(err, dboDesc,15, "n_spare2", FT_UINT32, &sz, &psz)
      )
    return 0;

  if (!zkDBODriverNew (err, &dboDrv))
    return 0;

  if (!zkDBODriverInit (err, DBODS_REGID, dboDrv,
			DSnew, DSfree, DSpack, DSunpack,
			NULL, NULL, DSdump, DStest))
    return 0;

  if (!zkDBODescRegister(err, dboDesc, dboDrv))
    return 0;

# ifndef NDEBUG
  if (!zkDBODescDump(err, dboDesc))
    return 0;
# endif  
  return 1;
}
