/*
 * Database Object Core Functions
 * $Id: zkDBO.c,v 1.12.28.2 2000/10/17 15:58:51 danielf Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zeroknowledge.com>

 *
 */
#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zkError.h>
#include <zkDebug.h>
#include <zkDump.h>
#include <zkDBO.h>
#include <zkDBOTypes.h>
#include "zkDBOpriv.h"

#ifdef ZKMOD
#undef ZKMOD
#endif
#define ZKMOD ZKMDB

/* Bug: Not thread safe */
struct T_ZkDBODescRegistry DBOReg = {0};

DLLExport
int zkDBODescRegister(ZkError err, ZkDBODesc dbod, ZkDBODriver dbd)
{
  static const char *funcname = "zkDBODescRegister";

  if (!dbod) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	       "ZkDBODesc is an invalid pointer");
    return 0;
  }

  dbod->descId = DBO_GET_REG(dbd->regID);
  DBOReg.desc[DBOReg.numreg] = dbod;
  DBOReg.driver[DBOReg.numreg] = dbd;
  DBOReg.map[DBOReg.numreg] = DBO_GET_REG(dbd->regID);  /* Bug: check for duplicates */
  DBOReg.numreg ++;
  return 1;
}


DLLExport
int zkDBODescUnRegisterAll(ZkError err)
{
  static const char *funcname = "zkDBODescUnRegisterAll";
  int i;

  for ( i = 0; i < DBOReg.numreg; i++ ) {
    if ( ! zkDBODriverFree(err,  & DBOReg.driver[i]) ) {
      zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
                 "Error freeing DBO driver");
      return 0;
    }

    if ( ! zkDBODescFree(err,  & DBOReg.desc[i]) ) {
      zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
                 "Error freeing DBO Driver Description.");
      return 0;
    }
    
    DBOReg.driver[i] = 0;
    DBOReg.desc[i] = 0;    
  }

  DBOReg.numreg  = 0;
  return 1;
}




DLLExport
int locateInRegMap (ZkDBODescRegistry reg, int descID)
{
  int i;

  for (i = 0; i < reg->numreg; i++) {
    if (reg->map[i] == DBO_GET_REG(descID))
      return i;
  }
  return -1;
}

DLLExport
int locateFieldName (ZkDBODesc dbod, const char *fieldname)
{
  size_t i;

  for (i = 0; i < dbod->numFields; i++) {
    /* Bug: AA required here */
    if (strcmp(fieldname, dbod->field[i].name) == 0)
      return i;
  }
  return -1;
}

DLLExport
int zkDBOExtractSig (ZkError err, void *pds, size_t pdsz, uint32_t *sig, size_t *pos)
{
  static const char *funcname = "zkDBOExtractSig";

  if (!zkDBOUnpackUint32 (err, funcname, (void **) &sig, pos, pds, &pdsz))
    return 0;
  
  return 1;
}

/**
 * Create a new DBO Data structure
 */
DLLExport
int zkDBONew (ZkError err, uint32_t regID, void **ds, size_t *dsz)
{
  static const char *funcname = "zkDBOPack";
  int i;
  ZkDBOGenDS gds;

  /* Bug: on non-null, don't reallocate */
  if (!ds) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "packing data structure is empty");
    return 0;
  }

  gds = ( ZkDBOGenDS ) *ds;
  
  if ((i = locateInRegMap(&DBOReg,regID)) < 0) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "Registration ID (%d) is unknown", DBO_GET_REG(regID));
    return 0;
  }

  if (!DBOReg.driver[i]->new) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "Pack driver for %d is not set ", DBO_GET_REG(regID));
    return 0;
  }
    
  if (!DBOReg.driver[i]->new(err, DBO_GET_VERS(regID), ds, dsz))
    return 0;
  
  return 1;
}

DLLExport
int zkDBODup (ZkError err, void *origDBO, size_t osz, void **dupDBO)
{
  static const char *funcname = "zkDBODup";

  /* A quick hack for the first pass */
  if ( ! origDBO || ! osz || ! dupDBO || ! *dupDBO ) {

    if ((*dupDBO = (void *) malloc (osz)) == NULL) { 
      zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENOMEM,"Can't malloc() memory"); 
      return 0;
    }
    memcpy (*dupDBO, origDBO, osz);
  }
  return 1;
  
}


DLLExport
int zkDBOPack(ZkError err, void *ds, size_t dsz, void **pds, size_t *pdsz)
{
  static const char *funcname = "zkDBOPack";
  int i;
  ZkDBOGenDS gds;

  gds = (ZkDBOGenDS) ds;
  
  if (!ds) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "packing data structure is empty");
    return 0;
  }
  
  if ((i = locateInRegMap(&DBOReg, gds->regID)) < 0) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "Registration ID (%d) is unknown", DBO_GET_REG(gds->regID));
    return 0;
  }

  if (!DBOReg.driver[i]->pack) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "Pack driver for %d is not set ", DBO_GET_REG(gds->regID));
    return 0;
  }
    
  if (!DBOReg.driver[i]->pack(err, ds, dsz, pds, pdsz))
    return 0;
  
  return 1;
}


DLLExport
int zkDBOFindPackedSize(ZkError err, void *ds, size_t dsz, void *pds, size_t *pdsz, void *userdata)
{
  static const char *funcname = "zkDBOPack";
  int i;
  ZkDBOGenDS gds;

  if (!ds || !pdsz) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "data structures are unavailable");
    return 0;
  }
  
  gds = (ZkDBOGenDS) ds;
  
  if (!ds) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "packing data structure is empty");
    return 0;
  }
  
  if ((i = locateInRegMap(&DBOReg, gds->regID)) < 0) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "Registration ID (%d) is unknown", DBO_GET_REG(gds->regID));
    return 0;
  }

  if (DBOReg.driver[i]->packedsize) {
    return (DBOReg.driver[i]->packedsize(err, ds, dsz, pds, pdsz, userdata));
  }
  
  return 1;
}

DLLExport
int zkDBOUnpack(ZkError err, void **ds, size_t *dsz, void *pds, size_t pdsz)
{
  static const char *funcname = "zkDBOUnpack";
  uint32_t sig = 0;
  size_t pos = sizeof(sig);
  int i;
  
  if (!pds || !pdsz) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "packed data structure is unavailable");
    return 0;
  }

  if (!zkDBOExtractSig (err, pds, pdsz, &sig, &pos))
    return 0;
  
  if ((i = locateInRegMap(&DBOReg, sig)) < 0) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "Registration ID (%d) is unknown", sig);
    return 0;
  }

  if (!DBOReg.driver[i]->unpack) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "Pack driver for %d is not set ", sig);
    return 0;
  }

  if (!DBOReg.driver[i]->unpack(err, ds, dsz, pds, pdsz))
    return 0;
  
  return 1;
}

DLLExport
int zkDBOFindUnpackedSize(ZkError err, void *ds, size_t *dsz, void *pds, size_t pdsz, void *userdata)
{
  static const char *funcname = "zkDBOUnpack";
  uint32_t sig = 0;
  size_t pos = sizeof(sig);
  int i;
  
  if (!pds || !pdsz || !dsz) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "packed data structure is unavailable");
    return 0;
  }

  if (!zkDBOExtractSig (err, pds, pdsz, &sig, &pos))
    return 0;
  
  if ((i = locateInRegMap(&DBOReg, sig)) < 0) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "Registration ID (%d) is unknown", sig);
    return 0;
  }

  if (DBOReg.driver[i]->unpackedsize) 
    return (DBOReg.driver[i]->unpackedsize(err, ds, dsz, pds, pdsz, userdata));
  return 1;
}

#ifndef NDEBUG
DLLExport
int zkDBODump(ZkError err, void *ds, void *pds, size_t pdsz) 
{ 
  static const char *funcname = "zkDBODump";
  ZkDBOGenDS gds;
  int i;

  gds = (ZkDBOGenDS) ds;
  
  if (ds) { 
    if ((i = locateInRegMap(&DBOReg, gds->regID)) < 0) { 
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD, 
		    "Registration ID (%d) is unknown", DBO_GET_REG(gds->regID)); 
      return 0; 
    } 
    if (!DBOReg.driver[i]->dump(err, ds)) 
      return 0; 
  } 

  if (pds){  
    if (!zkDumpBuffer(err,(const char *) pds, pdsz, zkDumpToStream, stderr)) 
      return 0; 
  } 

  return 1;
}

DLLExport
int zkDBOTestAll(ZkError err) 
{ 
  static const char *funcname = "zkDBOTestAll";
  int i;
  int retcode = 1;
  
  (void) funcname;
  /* Bug: should work on an object, not the Class variable */
  for (i = 0; i < DBOReg.numreg; i++) {
    if (!DBOReg.driver[i]->test) {
      zkDebug("Test driver unavailable");
    } else {
      retcode &= DBOReg.driver[i]->test(err);
    }
  }
  return retcode;
}

#else /* Release code*/

DLLExport
int zkDBODump(ZkError err, void *ds, void *pds, size_t pdsz) {
        (void)err;
        (void)ds;
        (void)pds;
        (void)pdsz;
	return 1;
}

DLLExport
int zkDBOTestAll(ZkError err) {
        (void)err;
	return 1;
}

#endif
