/**
 * DBO Utilities
 *
 * Pack and unpack an arbritary buffer
 *
 * $Id: zkDBOTypeBuf.c,v 1.6.26.1 2000/10/11 13:17:54 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zks.net>

 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zkDebug.h>
#include <zkDump.h>
#include <zkDBO.h>
#include <zkDBOTypes.h>

/*
      +----+---------+
      |SIZE| BUFFER  |
      +----+---------+

      size is sizof(buffer)+sizeof(size);
*/

#ifdef ZKMOD
#undef ZKMOD
#endif
#define ZKMOD ZKMDB

/**
 * Given a buffer size, returned what it's packed size would be
 */
DLLExport
int zkDBOSizeOfPackedBuf (
  ZkError err,
  const char *func,
  size_t bufsize,
  size_t *pbufsize) 
{
  static const char *callee = "zkDBOGetBufPackedSize";
  if (!func) func = callee;

  if (!pbufsize) {
    zkErrorSet_va(err,func,ZKERR,ZKMOD,ZKERROR,
		  "Data objects have null pointers");
    return 0;
  }
  *pbufsize = bufsize + sizeof(size_t);
  return 1;
}


DLLExport
int zkDBOPackBuf (ZkError err, const char *funcname, void *data, size_t ds,
	      void **packed, size_t *psize) 
{
  static const char *callee = "zkDBOPackBuf";
  uint32_t obj ;
  unsigned char *p;
    
  if (!funcname) funcname = callee;
  if (!data) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		  "Data objects have null pointers");
    return 0;
  }

  if (!ds) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		  "Original data has 0 length");
    return 0;
  }

  if (!*packed) {
    zkMALLOC (p, unsigned char *, ds + sizeof(ds));
  } else {
    if (*psize < (ds + sizeof(uint32_t))) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
           "Insufficient space (%d) allocated for packed data object (%d)",
		    *psize, ds);
      return 0;
    }
    p = (unsigned char *)*packed;
  }
  /* Copy the size of the buffer, followed by the buffer */
  obj = ds;
  
  p[0] = (obj & 0xff000000) >> 24;
  p[1] = (obj & 0xff0000) >> 16;
  p[2] = (obj & 0xff00) >> 8;
  p[3] = (obj & 0xff);
  
  memcpy(&p[4], data, ds);
  
  *packed = (void *) p;
  *psize = ds + sizeof(ds);

  return 1;
}

DLLExport
int zkDBOUnpackBuf (ZkError err, const char *funcname, void **data, size_t *ds,
	      void *packed, size_t *psize) 
{
  static const char *callee = "zkDBOUnpackBuf";
  uint32_t size;
  unsigned char *buf, *obj;
  unsigned char *p ;

  
  if (!funcname) funcname = callee;

  if (!packed) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		  "Data objects have null pointers");
    return 0;
  }

  p = (unsigned char *)packed;
  
  size = p[0]*0x1000000 + p[1]*0x10000 + p[2]*0x100 + p[3];

  *psize = size;

  obj = &p[4];
  
  if (!*data) {
    zkMALLOC (buf, unsigned char *, size);
    zkDebug("malloc'd %lu bytes", (unsigned long)size);
    *data = (void *)buf;
  } else {
    if (*ds < size) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
	     "Insufficient space [%d] allocated for unpacked data object(%lu)",
		    *ds, (unsigned long) size);
      return 0;
    }
    buf = (unsigned char *)*data;
  }

  memcpy(buf, obj, size);
  *ds = size;

  return 1;
}

#ifndef NDEBUG

DLLExport
int zkDBOTestBuf (ZkError err, char *buf, size_t len)
{
  static const char *funcname = "zkDBOTestBuf";
  char *dbuf = NULL;
  char *rbuf = NULL;
  size_t dlen, rlen;

  zkDebug ("%s: testing TypeBuf", funcname);
  
  if (!zkDBOPackBuf (err, funcname, buf, len,
		    (void **) &dbuf, &dlen))
    return 0;

  zkDebug ("%s: (dynbuf) packed contents of %d bytes", funcname, dlen);
  zkDumpBuffer(err, dbuf, dlen, zkDumpToStream, stderr);  

  if (!zkDBOUnpackBuf (err, funcname, (void **) &rbuf, &rlen,
	      (void *)dbuf, &len))
    return 0;

  zkDebug ("%s: (dynbuf) unpacked value using %d bytes (returned %d bytes)",
	   funcname,  (int) rlen, (int) len);
  zkDumpBuffer(err, rbuf, rlen, zkDumpToStream, stderr);  
  return 1;
}

#endif
