/**
 * DBO Utilities
 *
 * Pack and unpack uint32_t
 *
 * $Id: zkDBOTypeUint64.c,v 1.1.26.1 2000/10/11 13:17:54 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zks.net>

 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include "zknbo.h"
#include "zkDebug.h"
#include "zkDump.h"
#include "zkDBO.h"
#include "zkDBOTypes.h"


#ifdef ZKMOD
#undef ZKMOD
#endif
#define ZKMOD ZKMDB

#ifndef WIN32
#define ZK_PACKEDSIZE 8
DLLExport
int zkDBOPackUint64 (ZkError err, const char *funcname, void *data, size_t ds,
	      void **packed, size_t *psize) 
{
  static const char *callee = "zkPackUint64";
  unsigned char *ptr = NULL;

  (void) ds;
    
  if (!funcname) funcname = callee;

  if (!data || !packed || !psize ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		  "Data objects have null pointers");
    return 0;
  }

  if (!*packed) {
    zkMALLOC (ptr, unsigned char *, ZK_PACKEDSIZE);
    *psize = ZK_PACKEDSIZE;
  } else {
    if (*psize < ZK_PACKEDSIZE) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		    "Insufficient space allocated (%d) for packed data object (%d)", *psize,
                    ZK_PACKEDSIZE);
      return 0;
    }
    ptr = (unsigned char *) *packed;
  }

  zkNboFromHostUint64 (ptr, *(uint64_t *)data);

  *packed = ptr;
  *psize = ZK_PACKEDSIZE;

  return 1;
}

DLLExport
int zkDBOUnpackUint64 (ZkError err, const char *funcname, void **data, size_t *ds,
	      void *packed, size_t *psize) 
{
  static const char *callee = "zkUnpackUint64";
  uint64_t obj;
  unsigned char *ptr = NULL;

  if (!funcname) funcname = callee;

  if ( !packed || !psize || !data || !ds ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		  "Data objects have null pointers");
    return 0;
  }

  if (*psize < ZK_PACKEDSIZE) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		  "Incorrect buffer size for holding uint64_t");
    return 0;
  }
  ptr = (unsigned char *) packed;

  obj = zkNboToHostUint64(ptr);

  if (!*data) {
    zkMALLOC (*data, void *, sizeof (uint64_t));
  } else {
    if (*ds < sizeof(uint64_t)) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
	    "Insufficient space [%d] allocated for unpacked data object", *ds);
      return 0;
    }
  }

  *ds = sizeof(uint64_t);
  ** (uint64_t ** )data = obj;
  *psize = ZK_PACKEDSIZE;

  return 1;
}

#ifndef NDEBUG

DLLExport
int zkDBOTestUint64 (ZkError err, UINT64_T num)
{
  static const char *funcname = "zkTestUint64";
  char *dbuf = NULL;
  char sbuf[32] = {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
                   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
                   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};
  uint64_t recovered, *pu;
  size_t s = 32, ps = 8;

  dbuf = sbuf;
  
  zkDebug ("%s: --------- Uint64 Test ----------", funcname);
  zkDebug ("%s: testing " U_LONG_LONG_FMT " [0x%Lx]", funcname,  num, num);
  
  if (!zkDBOPackUint64 (err, funcname, &num, sizeof(num),
		    (void **) &dbuf, &s))
    return 0;

  zkDebug ("%s: static buf has packed size of %u bytes)",
           funcname,  s);
  zkDumpBuffer(err, dbuf, s, zkDumpToStream, stderr);  

  pu = &recovered;
  
  if (!zkDBOUnpackUint64 (err, funcname, (void **) &pu, &ps,
	      (void *)dbuf, &s))
    return 0;

  zkDebug ("%s: (static buf) unpacked value is "
           U_LONG_LONG_FMT "/%d (0x%Lx) using %d bytes\n-------------",
	   funcname, (UINT64_T)*pu,
	   ps, (UINT64_T) *pu,
	   (int) s);
  return 1;
}

#endif
#endif
