/*
 * Database Object Functions
 *
 * Unpacking handlers
 *
 * $Id: zkDBOUnpacked.c,v 1.10.20.1 2000/10/11 13:17:54 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zeroknowledge.com>

 *
 */
#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zkError.h>
#include <zkDebug.h>
#include <zkDBO.h>
#include <zkDBOTypes.h>
#include "zkDBOpriv.h"

#ifdef ZKMOD
#undef ZKMOD
#endif
#define ZKMOD ZKMDB

/*
 * WARNING!!! It is implied that unpacking is done in a sequential manner
 */
DLLExport
int zkDBOUnpackField(ZkError err, int regID, const char *fieldname,
		void **data, size_t *ds, void *pdata, size_t psz, size_t *pos)
{
  static const char *funcname = "zkDBOUnpackField";
  int i;
  ZkDBODesc dbod;
  void *pd;
  size_t p, overhead = 0;
  
  if ( !data || !ds || !pdata || !psz || !fieldname || !*fieldname) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
		  "Null pointers or values passed to function");
    return 0;
  }
  /*
   * Bug: add this when I switch to dynamic data structures
  if (DBOReg){
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
		  "DBO Registration is not initialized");
    return 0;
  }
  */

  if ((i = locateInRegMap(&DBOReg, regID)) < 0) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "Registration ID (%d) is unknown", DBO_GET_REG(regID));
    return 0;
  }

  if (!(dbod = DBOReg.desc[i])) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "DBO [%d/%d] is unavailable", DBO_GET_REG(regID), i );
    return 0;
  }
  
  if ((i = locateFieldName(dbod, fieldname)) < 0) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "No (%s) fieldname located (case?)", fieldname);
    return 0;
  }

  pd = (char *) pdata + *pos;
  p = psz - *pos;

  switch (dbod->field[i].type) {

#ifndef WIN32
  case FT_UINT64: 
    if (!zkDBOUnpackUint64 (err, funcname, data, ds, pd, &p))
      return 0;
    break;
#endif

  case FT_UINT32: 
    if (!zkDBOUnpackUint32 (err, funcname, data, ds, pd, &p))
      return 0;
    break;
  case FT_IPADDR: 
    if (!zkDBOUnpackIpAddr (err, funcname, data, ds, pd, &p))
      return 0;
    break;
  case FT_UINT16: 
    if (!zkDBOUnpackUint16 (err, funcname, data, ds, pd, &p))
      return 0;
    break;
  case FT_STR:
  case FT_BUF:
  case FT_USRDEF:
    overhead = sizeof(uint32_t);
    if (!zkDBOUnpackBuf (err, funcname, data, ds, pd, &p))
      return 0;
    break;
  default: 
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKESTUB,
		  "Datatype (%s) not implemented yet",
		  printFieldType(dbod->field[i].type));
    return 0;
  }
  *pos += (p + overhead);
  return 1;
}


/**
 * Default mechanism to determine the unpacked spaced based on the types of all
 * packed objects. No consideration to the DBO's version is made.
 * Install a  `ZkDBODriverUnPkSz_f unpackedsize' driver entry to
 * override this.
 */

DLLExport
int zkDBOUnpackedReqSpace(ZkError err, int regID, size_t *size)
{
  static const char *funcname = "zkDBOUnpackedReqSpace";
  int i;
  size_t s, sum;
  ZkDBODesc dbod;

  if ( !size ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
		  "Null pointers or values passed to function");
    return 0;
  }
  if ((i = locateInRegMap(&DBOReg, regID)) < 0) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "Registration ID (%d) is unknown", DBO_GET_REG(regID));
    return 0;
  }

  if (!(dbod = DBOReg.desc[i])) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "DBO [%d/%d] is unavailable", DBO_GET_REG(regID), i );
    return 0;
  }

  if (DBOReg.driver[i]->unpackedsize) {
    return (DBOReg.driver[i]->unpackedsize(err, NULL, size, NULL, 0, NULL));
  } else {

    for (s = 0, sum = 0; s < dbod->numFields ; s++) {
      sum += dbod->field[s].size;
    }

    *size = sum;
  }
  
  zkDebug("Unpacked (ie: datastructure) size for %d is %u", DBO_GET_REG(regID), *size);
  return 1;
}

/* Set the *size to non-zero for a specific datablock */
static int zkDBOUnpackedNew(ZkError err, int regID, void **data, size_t *size)
{
  static const char *funcname = "zkDBOUnpackedNew";
  size_t tsize = 0;

  if ( !data || !size ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
		  "Null pointers or values passed to function");
    return 0;
  }

  if (!zkDBOUnpackedReqSpace(err, regID, &tsize))
    return 0;
#if 0
  /* This used to be the old way of doing it */
  if (size && *size && (tsize < *size))
    /* Allocate more space if we ask for it */
    tsize = *size;
#endif  

  /* If the size is set, use the size supplied */
  if (*size)
    /* Allocate more space if we ask for it */
    tsize = *size;

  if (*data)
    /* Allocate more space if we ask for it */
    zkErrorSet_va(err,funcname,ZKWARN,ZKMOD,ZKENULLPTR,
		  "Potential memory leak detected");

  /* Complain if data is not NULL */
  
  zkMALLOC(*data, void *, tsize);

  *size = tsize;
 
  return 1;
}

DLLExport
int zkDBOUnpackedInit(ZkError err, int *regID, void **data, size_t *size, void *pdata, size_t psz, size_t *pos)
{
  static const char *funcname = "zkDBOUnpackedInit";
  size_t tsize = 0, ssz = 0;
  uint32_t sig = 0, *psig = NULL;

  /*
   * First we need to examine the incoming data object
   * to get it's type
   */

  psig = & sig;
  ssz = sizeof(sig);
  tsize = psz;
  if (!zkDBOUnpackUint32 (err, funcname, (void **)&psig, &ssz, pdata, &tsize))
    return 0;
  *pos = tsize;
  *regID = sig;

  tsize = *size;
  
  if (!*data) {
    /* Need to alloc space */
    if (!zkDBOUnpackedNew(err, *regID, data, size)) 
      return 0;
    tsize = *size;
  } else {
    /* Check space and sign the datablock */
    if (!zkDBOUnpackedReqSpace(err, *regID, &tsize)) 
      return 0;/* 
    if (!zkDBOFindUnpackedSize(err, *data, size, pdata, psz, &tsize))
    return 0;*/
    if (*size < tsize) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEFULL,
		    "Insufficient space[%lu] to hold datastructure[%lu]",
		    (unsigned long) *size, (unsigned long) tsize);
      return 0;
    }
  }
  return 1;
  
}
