/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "frKeyQry.h"
#include "zkNet.h"


/** 
 * Fills in a FrKeyQryQueryExp structure and tries to retrieve the key for
 * that query from the key query server.  This function will "block" until the
 * key is found and authenticated, key is not found, or the timeout (in
 * seconds).
 *
 *   NOTE: This interface was added to provide an API that distinguish a key
 *         not found condition from a failure. 
 * 
 * @param err
 * @param kq                    Key cache to store the keys
 * @param keyVers               Version of the key to look up
 * @param keyType               Type of key (signing or cryption)
 * @param entid                 Entity of the key to look up
 * @param keyHost               A sockaddr_in and the descriptor of 
 *                              the open unconnected UDP socket
 * @param timeout               Max time to wait for all keys
 * @param retries               Number of times to retry
 * @param key                   Public key (if found)
 * @param found                 Indicates if the key was found.
 * @return                      1 on success, 0 on failure
 */
DLLExport 
extern int frKeyQryFetchPubKeyWithStatus(ZkError err,
                                         FrKeyQry kq,
                                         unsigned int keyVers,
                                         FrPubKeyType keyType,
                                         FrEntIdExp entid, 
                                         FrKeyQryHost keyHost,
                                         int retries,
                                         int timeout,
                                         FrCryptPubKey *key,
                                         ZkBool *found,
                                         ZkBool *valid)
{
  const char *func = "frKeyQryFetchPubKeyWithStatus";
  FrKeyQryQueryExp kqry;
  FrKeyQryState state;
  fd_set rset;
  struct timeval timer;

  zkErrorClear(err);

  if (!kq) {
    zkErrorSet(err, func, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "Key cache is NULL");
    return 0;
  }

  if (!found) {
    zkErrorSet(err, func, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "found boolean is NULL");
    return 0;
  }
  *found = 0;

  if (!valid) {
    zkErrorSet(err, func, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "valid boolean is NULL");
    return 0;
  }
  *valid = 0;

  if (retries < 0) {
    retries = 0;
  }
  retries++;

  if ((keyType != FRCRYPT_KEYTYPE_SIGNING) && 
      (keyType != FRCRYPT_KEYTYPE_CRYPTION)) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRKEYQRY, ZKEPARMBAD, 
                  "Key type [%d] is not recognized", keyType);
    return 0;
  }

  memset(&timer, 0, sizeof(timer));
  timer.tv_sec = timeout;

  /* Reset the key query structure */
  memset(&kqry, 0, sizeof(kqry));
  kqry.keyVersion = keyVers;
  kqry.keyType = keyType;
  memcpy(&kqry.keyOwner, &entid, sizeof(entid));

  do {
    if (!frKeyQryFindKey(err, kq, keyHost.fd, 
                         (struct sockaddr *) &keyHost.host, &kqry, &state,
                         key)) {
      return 0;
    }

    switch (state) {
    case FRKEYQRY_ST_AWAITING_RESPONSE: /* Nothing yet */
    case FRKEYQRY_ST_AWAITING_VALIDATION: /* Waiting for higher level key */
      break;

    case FRKEYQRY_ST_VALID:     /* Got a key which I requested */
      *valid = 1;
      *found = 1;
      break;

    case FRKEYQRY_ST_INVALID:   /* Invalid key qry */
      *valid = 0;
      *found = 1;
      return 1;

    case FRKEYQRY_ST_NOT_FOUND: /* Key not found */
      /* The key not found condition is not considered an error. */      
      return 1;

    default:                    /* Unknown state returned */
      zkErrorSet(err, func, ZKERR, ZKMFRKEYQRY, ZKEDATABAD, "unknown state");
      return 0;
    }

    if (!*found) {
      int nsel;
      int nbytes;
      char buf[BUFSIZ];
      size_t bsz = BUFSIZ;
      struct sockaddr_in from;
      size_t fromLen = sizeof(from);

      FD_ZERO(&rset);
      FD_SET(keyHost.fd, &rset);

      if ((nsel = zkSelectUpdateTimeout(err, keyHost.fd + 1, &rset, 0, 0, 
                                        &timer)) == -1) {
        zkErrorSet_va(err, func, ZKERR, ZKMFRKEYQRY, errno, "select() failed: %s",
                      strerror(errno));
        return 0;
      }

      if (nsel > 0) {
        if ((nbytes = recvfrom(keyHost.fd, buf, bsz, 0, &from, 
                               &fromLen)) == -1) {
          zkErrorSet_va(err, func, ZKERR, ZKMFRKEYQRY, errno, 
                        "recvfrom(%s) failed", strerror(errno));
          return 0;
        }

        /* A key was received drop it into the cache */
        if (!frKeyQryRecvKey(err, kq, keyHost.fd, 
                             (struct sockaddr *) &keyHost.host, buf, nbytes)) {
          return 0;
        }
      } else {
        retries--;

        if (!retries) {
          zkErrorSet(err, func, ZKWARN, ZKMFRKEYQRY, ZKETIME,
                     "timeout before finding the key");
          key = NULL;
          return 0;
        }
        /* Reset the timer */
        memset(&timer, 0, sizeof(timer));
        timer.tv_sec = timeout;
      }
    }
  } while (!*found);

  return 1;
}


/** 
 * Fills in a FrKeyQryQueryExp structure and tries to retrieve the key for
 * that query from the key query server.  This function will "block" until the
 * key is found and authenticated, key is not found, or the timeout (in
 * seconds).
 *
 *   NOTE: With this function, a "key not found" condition is considered an
 *         error.
 *
 * @param err
 * @param kq                    Key cache to store the keys
 * @param keyVers               Version of the key to look up
 * @param keyType               Type of key (signing or cryption)
 * @param entid                 Entity of the key to look up
 * @param keyHost               A sockaddr_in and the descriptor of 
 *                              the open unconnected UDP socket
 * @param timeout               Max time to wait for all keys
 * @param retries               Number of times to retry
 * @param key                   Public key (if found)
 * @return                      1 on success, 0 on failure
 */
DLLExport 
extern int frKeyQryFetchPubKey(ZkError err, FrKeyQry kq, unsigned int keyVers,
                               FrPubKeyType keyType, FrEntIdExp entid, 
                               FrKeyQryHost keyHost, int retries, int timeout,
                               FrCryptPubKey *key)
{
  const char *func = "frKeyQryFetchPubKey";
  ZkBool found = 0;
  ZkBool valid = 0;

  if (frKeyQryFetchPubKeyWithStatus(err, kq, keyVers, keyType, entid, keyHost, 
                                    retries, timeout, key, &found, &valid)) {
    /* In the current function "not found"  and "not valid" are errors. */
    if (!found) {
      zkErrorSet(err, func, ZKWARN, ZKMFRKEYQRY, ZKENOTFOUND, "not found");
      return 0;
    }
    if (!valid) {
      zkErrorSet(err, func, ZKWARN, ZKMFRKEYQRY, ZKEINVAL, "invalid");
      return 0;
    }
    return 1;
  }

  return 0;
}



/*
 *  Perform multiple key queries in parallel.  No more than `tries' queries
 *  will be made for any particular key.  Also, `timeout' gives the
 *  approximate total time to wait for all the responses.
 *  
 *  `*key' will be set to an array (allocated by this function) of pointers to
 *  retrieved keys.  The pointer will be null if the key was not found.  In
 *  any case, it will be a copy of a pointer in the key cache: thus it should
 *  not be freed.
 *  
 *  It is not considered an error if some (or even all) keys are not
 *  retrieved.
 *  
 *  bug: add trace information for debugging (esp. for Mario).
 */
DLLExport int frKeyQryMultiFetch(ZkError err, FrKeyQry kq, FrKeyQryHost keyHost, unsigned int retries, struct timeval timeout, FrKeyQryQueryExp *qry, size_t nqrys, FrCryptPubKey **key)
{
  FrCryptPubKey *k = 0;
  const char *fn = "frKeyQryMultiFetch";
  int ret = 0;
  size_t q;
  struct timeval iqto; /* inter-query timeout */
  struct timeval to;
  unsigned int r = 0; /* retry count */

  zkErrorClear(err);

  if ( ! kq) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "kq is null");
    return 0;
  }
  if ( ! qry) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "qry is null");
    return 0;
  }
  if (nqrys == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKEPARMBAD, "nqrys is 0");
    return 0;
  }

  /*
   *  Set the `inter-query' timeout to a fraction of the total timeout.
   */

  iqto = zkTvDivide(timeout, retries + 1);
  to = iqto;

  /*
   *  Allocate memory for the public keys.
   */
  
  if ( ! (k = malloc(nqrys * sizeof *k))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, errno,
                  "can't allocate %lu bytes for key pointers: %s",
                  (unsigned long)nqrys * sizeof *k, strerror(errno));
    goto done;
  }

  for (q = 0; q < nqrys; q++) {
    k[q] = 0;
  }
  
  do {
    FrKeyQryState state;
    int moreKeys = 0;
    int n;
    fd_set rdset;
    
    /*
     *  Send queries for keys that aren't in the local cache.
     */
    
    for (q = 0; q < nqrys; q++) {
      int noResp;
#ifdef WIN32
		Sleep(200);
#endif

      noResp = ! frKeyQryCacheFindKey(err, kq, &qry[q], &state, &k[q]);
      
      if (noResp || (state != FRKEYQRY_ST_INVALID && state != FRKEYQRY_ST_VALID &&
                     state != FRKEYQRY_ST_NOT_FOUND)) {
        if ( ! frKeyQryFindKey(err, kq, keyHost.fd, (struct sockaddr *)&keyHost.host,
                               &qry[q], &state, &k[q])) {
          goto done;
        }

        moreKeys = 1; /* there are more keys to receive */
      }
    }

    if ( ! moreKeys) {
      break;
    }
    
    /*
     *  Wait for responses.
     */

    FD_ZERO(&rdset);
    FD_SET(keyHost.fd, &rdset);
    
    if ((n = zkSelectUpdateTimeout(err, keyHost.fd + 1, &rdset, 0, 0, &to)) == -1) {
      zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, errno,
                    "error waiting for responses: %s", strerror(errno));
      goto done;
    } else if (n != 0) {
      int maxLoops = nqrys;
      
      /*
       *  Read and process responses.
       *  
       *  Limit the number of times through the loop to prevent a
       *  denial-of-service attack.
       */

      while (maxLoops--) {
        char rpkt[3 * 1024];
        int rpsz;
        struct timeval rdyto = { 0, 0 };
        
        /*
         *  Use select() to determine if the file descriptor is still
         *  readable, to avoid blocking in rcvfrom().
         */
        
        if (select(keyHost.fd + 1, &rdset, 0, 0, &rdyto) < 1) {
          break;
        }

        if ((rpsz = recvfrom(keyHost.fd, rpkt, sizeof rpkt, 0, 0, 0)) == -1) {
          zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, errno,
                        "error reading key query response: %s", strerror(errno));
          goto done;
        }

        /*
         *  Ignore errors so that one bad packet won't stop us.
         *  
         *  bug: maybe we shouldn't ignore _all_ errors.
         */

        (void)frKeyQryRecvKey(err, kq, keyHost.fd, (struct sockaddr *)&keyHost.host,
                              rpkt, rpsz);
      }
    }

    /*
     *  Any time left in the timeout will be given to the next select.
     */

    to = zkTvAdd(to, iqto);
  } while (r++ < retries);

  *key = k; k = 0; /* caller now owns it */

  ret = 1;

 done:

  if (k) { free(k); }

  return ret;
}
