/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "frKeyQry.h"
#include "zkNet.h"


/*
 *  Perform multiple key queries in parallel.  `timeout' gives the approximate
 *  total time to wait for all the responses.
 *  
 *  `*key' will be set to an array (allocated by this function) of pointers to
 *  retrieved keys.  The pointer will be null if the key was not found.  In
 *  any case, it will be a copy of a pointer in the key cache: thus it should
 *  not be freed.
 *  
 *  It is not considered an error if some (or even all) keys are not
 *  retrieved.
 */
DLLExport int frKeyQryMultiFetch2(ZkError err, FrKeyQry kq, FrKeyQryHost keyHost, unsigned int retries, struct timeval timeout, FrKeyQryQueryExp *qry, size_t nqrys, FrCryptPubKey **key)
{
  FrCryptPubKey *k = 0;
  const char *fn = "frKeyQryMultiFetch";
  int ret = 0;
  size_t i;
  size_t nq = 0; /* index of next query to make */
  struct timeval iqto; /* inter-query timeout */
  struct timeval to;

  zkErrorClear(err);

  if ( ! kq) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "kq is null");
    return 0;
  }
  if ( ! qry) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "qry is null");
    return 0;
  }
  if (nqrys == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKEPARMBAD, "nqrys is 0");
    return 0;
  }

  /*
   *  Set the `inter-query' timeout to a fraction of the total timeout.
   *  
   *  The fraction is decided by the number of queries, to which is added a
   *  bit of slop to cover higher-level keys that may also have to be queried.
   *  (Assuming a query is for a nym cryption key, we add enough to include
   *  the nym signature key as well as the nym server, monthly and yearly
   *  signing keys.)
   *  
   *  bug? have a minimum timeout?
   */

  {
    int slop = 4;
    
    iqto = zkTvDivide(timeout, nqrys + slop);
  }

  /*
   *  Allocate memory for the public keys.
   */
  
  if ( ! (k = malloc(nqrys * sizeof *k))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, errno,
                  "can't allocate %lu bytes for key pointers: %s",
                  (unsigned long)nqrys * sizeof *k, strerror(errno));
    goto done;
  }

  for (i = 0; i < nqrys; i++) {
    k[i] = 0;
  }
  
  do {
    FrKeyQryState state;
    int n;
    fd_set rdset, wrset;
    
    /*
     *  Find the index, nq, of the next entry for which to query.
     */

    {
      int mustQuery = 0;
      size_t t;
      
      for (i = 0; i < nqrys; i++) {
        t = (nq + i + 1) % nqrys;
        if ( ! k[t]) {
          if ( ! frKeyQryCacheFindKey(err, kq, &qry[t], &state, &k[t]) ||
               (state != FRKEYQRY_ST_INVALID && state != FRKEYQRY_ST_VALID &&
                state != FRKEYQRY_ST_NOT_FOUND)) {
            mustQuery = 1;
            break;  /* n is index of next query to make */
          }
        }
      }

      if (mustQuery) {
        nq = t;
      } else {
        break;  /* all entries have been found */
      }
    }
    
    FD_ZERO(&rdset);
    FD_SET(keyHost.fd, &rdset);
    
    FD_ZERO(&wrset);
    FD_SET(keyHost.fd, &wrset);

    to = iqto;

    n = select(keyHost.fd + 1, &rdset, &wrset, 0, &to);

    if (n == -1) {
      zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, errno,
                    "error from select(): %s", strerror(errno));
      goto done;
    } else if (n != 0) {
      /*
       *  See if the file descriptor is readable and writable.
       */
      
      if (FD_ISSET(keyHost.fd, &rdset)) {
        char rpkt[3 * 1024];
        int rpsz;
        size_t maxRecv = nqrys;
        struct timeval noblock = { 0, 0 };
          
        /*
         *  Read packets until there are no more.
         *  
         *  Use select() to determine if the file descriptor is still
         *  readable, to avoid blocking in rcvfrom().
         */

        do {
          if ((rpsz = recvfrom(keyHost.fd, rpkt, sizeof rpkt, 0, 0, 0)) == -1) {
            zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, errno,
                          "error reading key query response: %s", strerror(errno));
            goto done;
          }

          /*
           *  Ignore errors so that one bad packet won't stop us.
           *  
           *  bug: maybe we shouldn't ignore _all_ errors.
           */

          (void)frKeyQryRecvKey(err, kq, keyHost.fd, (struct sockaddr *)&keyHost.host,
                                rpkt, rpsz);
        } while (maxRecv-- && (n = select(keyHost.fd + 1, &rdset, 0, 0, &noblock)));

        if (n < 0) {
          zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, errno,
                        "error from select() on readable descriptor (%d): %s",
                        keyHost.fd, strerror(errno));
          goto done;
        }
      }

      if (FD_ISSET(keyHost.fd, &wrset)) {
        /*
         *  Make one query.
         *  
         *  In fact, we first query our local cache, which may be able to
         *  satisfy this request.  We make a real query if the response is not
         *  in the cache, or if it has not yet been fully resolved.
         */

        {
          struct timeval t = { 0, 50000 }; /* 1/20 sec */
          (void)select(0, 0, 0, 0, &t);
        }

        if ( ! frKeyQryFindKey(err, kq, keyHost.fd, (struct sockaddr *)&keyHost.host,
                               &qry[nq], &state, &k[nq])) {
          goto done;
        }
      }
    }
  } while (1);

  *key = k; k = 0; /* caller now owns it */

  ret = 1;

 done:

  if (k) { free(k); }

  return ret;
}
