/*
 *  $Id: mkdb.c,v 1.20.20.1 2000/10/11 13:18:10 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/*
 *  The key is exported from the crypto DLL and used (imported) by the other,
 *  keyqry.DLL in this case.
 */
#undef EXPORTMASTERKEY

#define DEBUG_SIGN 0

#include "zks.h"

#include <stdlib.h>
#include <time.h>
#include "fr_keyqry_wire.h"
#include "frCrypt.h"
#include "frKeyQry.h"
#include "zkError.h"
#include "zkClock.h"


/*
 *  Create a key server database query, and its corresponding datum, from key
 *  meta data and the key itself.
 *  
 *  If `curr' is 0 we create a query that will find the specified version of
 *  the key; if 1 it will find the current version.
 *  
 *  This is primarily used in the generation of database entries for testing.
 */
DLLExport int fr_KeyQryDbFromOwnerKey(ZkError err, int curr, void *q, size_t *qsz, void *d, size_t *dsz, FrCryptPubKey pk)
{
  FrEntIdExp keyOwner;
  struct fr_keyqry_qry *db_query = q;
  struct fr_keyqry_auth *db_datum = d;

  zkErrorClear(err);
  
  if ( ! q) {
    zkErrorSet(err, "fr_KeyQryDbFromOwnerKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "q is null");
    return 0;
  }
  if ( ! qsz) {
    zkErrorSet(err, "fr_KeyQryDbFromOwnerKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "qsz is null");
    return 0;
  }
  if (*qsz < sizeof *db_query) {
    zkErrorSet_va(err, "fr_KeyQryDbFromOwnerKey", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "*qsz is too small (%lu < %lu)",
                  (unsigned long)*qsz, (unsigned long)sizeof *db_query);
    return 0;
  }
  if ( ! d) {
    zkErrorSet(err, "fr_KeyQryDbFromOwnerKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "d is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, "fr_KeyQryDbFromOwnerKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "dsz is null");
    return 0;
  }
  if (*dsz < sizeof *db_datum) {
    zkErrorSet_va(err, "fr_KeyQryDbFromOwnerKey", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "*dsz is too small (%lu < %lu)",
                  (unsigned long)*dsz, (unsigned long)sizeof *db_datum);
    return 0;
  }
  if ( ! pk) {
    zkErrorSet(err, "fr_KeyQryDbFromOwnerKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "pk is null");
    return 0;
  }
  
  /*
   *  Create the query that will find the key.
   */

  {
    unsigned int t, v;

    if ( ! (frCryptPubKeyGetVersion(err, pk, &v) &&
            frCryptPubKeyGetType(err, pk, &t))) {
      return 0;
    }

    db_query->keyVersion = v;
    db_query->keyType = t;
  }
    
  if (curr) {
    db_query->keyVersion = 0; /* bug: use #define */
  }

  if ( ! frCryptPubKeyGetEntity(err, pk, &keyOwner)) {
    return 0;
  }

  switch (keyOwner.type) {
  case FRENT_TYPE_AIP:
  case FRENT_TYPE_ANON:
  case FRENT_TYPE_FMG:
  case FRENT_TYPE_FREEDOM_MONTH:
  case FRENT_TYPE_FREEDOM_YEAR:
  case FRENT_TYPE_KEYQRY_SRV:
  case FRENT_TYPE_KEYUPD_SRV:
  case FRENT_TYPE_MAIP:
  case FRENT_TYPE_NIDB:
  case FRENT_TYPE_NIQS:
  case FRENT_TYPE_NISS:
  case FRENT_TYPE_NNTP_SRV:
  case FRENT_TYPE_NYM:
  case FRENT_TYPE_NYM_SRV:
  case FRENT_TYPE_SER_SRV:
  case FRENT_TYPE_TOK_SRV:
    break;

  default:
    zkErrorSet_va(err, "fr_KeyQryDbFromOwnerKey", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "invalid key type (%u)", (unsigned int)keyOwner.type);
    return 0;
  }
  db_query->keyOwner[0] = keyOwner.type; /* bug */
  memcpy(&db_query->keyOwner[1], keyOwner.name, sizeof db_query->keyOwner - 1);

  *qsz = sizeof *db_query;

  /*
   *  Create the datum.
   */

  memcpy(db_datum->keyOwnerEntity, db_query->keyOwner, sizeof db_datum->keyOwnerEntity);

  /*
   *  We don't fill in the `signer' field or the `signerSig' field; that will
   *  be done only when we're ready to sign the response.
   */

  {
    size_t offset = sizeof *db_datum; /* offset at which to store the key */
    size_t remain = *dsz - offset; /* how much is left to store the key */
    
    if ( ! frCryptPubKeyWriteToBuf(err, pk, (char *)db_datum + offset, &remain)) {
      return 0;
    }
    *dsz = offset + remain;
  }

  return 1;
}


/*
 *  Create a key server database datum from a key.
 *  
 *  This routine was written to help implement the --createDatum option of
 *  fr_datum.
 *  
 *  bug: we assume that the key contains only those parameters that we want to
 *  store.  I.e. we don't mask out any unwanted parameters.
 */
DLLExport int fr_KeyQryDbDatumFromKey(ZkError err, void *d, size_t *dsz, FrCryptPubKey key)
{
  FrEntIdExp owner;
  int ret = 0;
  size_t sz;
  struct fr_keyqry_auth *db_datum = d;

  zkErrorClear(err);
  
  if ( ! d) {
    zkErrorSet(err, "fr_KeyQryDbDatumFromKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "d is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, "fr_KeyQryDbDatumFromKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "dsz is null");
    return 0;
  }
  if (*dsz < sizeof *db_datum) {
    zkErrorSet_va(err, "fr_KeyQryDbDatumFromKey", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "*dsz is too small (%lu < %lu)",
                  (unsigned long)*dsz, (unsigned long)sizeof *db_datum);
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, "fr_KeyQryDbDatumFromKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "key is null");
    return 0;
  }
  
  sz = *dsz - sizeof *db_datum;

  memset(db_datum, 0, sizeof *db_datum);
  if ( ! (frCryptPubKeyGetEntity(err, key, &owner) &&
          frEntStore(err, &owner, db_datum->keyOwnerEntity,
                     sizeof db_datum->keyOwnerEntity) &&
          frCryptPubKeyWriteToBuf(err, key, (char *)(db_datum + 1), &sz))) {
    goto done;
  }

  *dsz = sizeof *db_datum + sz;

  ret = 1;

 done:
  
  return ret;
}


/*
 *  Create a key server database query from a key.
 *  
 *  If `curr' is 0 we create a query that will find the specified version of
 *  the key; if 1 it will find the current version.
 *  
 *  This routine was written to help implement the --createQuery option of
 *  fr_datum.
 */
DLLExport int fr_KeyQryDbQueryFromKey(ZkError err, int curr, void *q, size_t *qsz, FrCryptPubKey key)
{
  FrEntIdExp keyOwner;
  int ret = 0;
  struct fr_keyqry_qry *db_query = q;

  zkErrorClear(err);
  
  if ( ! q) {
    zkErrorSet(err, "fr_KeyQryDbQueryFromKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "q is null");
    return 0;
  }
  if ( ! qsz) {
    zkErrorSet(err, "fr_KeyQryDbQueryFromKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "qsz is null");
    return 0;
  }
  if (*qsz < sizeof *db_query) {
    zkErrorSet_va(err, "fr_KeyQryDbQueryFromKey", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "*qsz is too small (%lu < %lu)",
                  (unsigned long)*qsz, (unsigned long)sizeof *db_query);
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, "fr_KeyQryDbQueryFromKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "key is null");
    return 0;
  }
  
  /*
   *  Create the query that will find the key.
   */

  {
    unsigned int t, v;

    if ( ! (frCryptPubKeyGetVersion(err, key, &v) &&
            frCryptPubKeyGetType(err, key, &t))) {
      goto done;
    }

    db_query->keyVersion = v;
    db_query->keyType = t;
  }
    
  if (curr) {
    db_query->keyVersion = 0; /* bug: use #define */
  }

  if ( ! (frCryptPubKeyGetEntity(err, key, &keyOwner) &&
          frEntStore(err, &keyOwner, db_query->keyOwner,
                     sizeof db_query->keyOwner))) {
    goto done;
  }

  *qsz = sizeof *db_query;

  ret = 1;

 done:
  
  return ret;
}


/*
 *  Given a database datum, set `**keydata' and `*ksz' to point to the start
 *  of the key data and its length, respectively.
 *  
 *  This routine was written to support the --extractKey option of fr_datum.
 *  
 *  bug: doesn't do error checking on the key data.  (Write routine to
 *  validate portable key?)
 */
DLLExport int fr_KeyQryGetKeyData(ZkError err, void *d, size_t dsz, void **keydata, size_t *ksz)
{
  struct fr_keyqry_auth *db_datum = d;

  zkErrorClear(err);
  
  if ( ! d) {
    zkErrorSet(err, "fr_KeyQryGetKeyData", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "d is null");
    return 0;
  }
  if (dsz < sizeof *db_datum) {
    zkErrorSet_va(err, "fr_KeyQryGetKeyData", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "dsz is too small (%lu < %lu)",
                  (unsigned long)dsz, (unsigned long)sizeof *db_datum);
    return 0;
  }
  if ( ! keydata) {
    zkErrorSet(err, "fr_KeyQryGetKeyData", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "keydata is null");
    return 0;
  }
  if ( ! ksz) {
    zkErrorSet(err, "fr_KeyQryGetKeyData", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "ksz is null");
    return 0;
  }

  *keydata = db_datum + 1;
  *ksz = dsz - sizeof *db_datum;

  return 1;
}


/*
 *  Warning: this routine performs no checks on the validity of the key owner.
 *  That is, we don't check that the public key parameters are actually
 *  correct.
 *  
 *  bug: we probably ought to do some checks on what we're signing.
 *  
 *  bug: verify that the signature includes all the key parameters.
 */
DLLExport int fr_KeyQrySignDbDatum(ZkError err, FrCryptSig sig, void *d, size_t dsz)
{
  FrEntIdExp keyOwner, sigOwner;
  FrPubKeyType keyType;
  int ret = 0, valid;
  size_t keyDataSz;
  struct fr_keyqry_auth *ka = d;
  void *keyDataPtr;
  
  zkErrorClear(err);
 
  if ( ! sig) {
    zkErrorSet(err, "fr_KeyQrySignDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "sig is null");
    return 0;
  }
  if ( ! d) {
    zkErrorSet(err, "fr_KeyQrySignDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "d is null");
    return 0;
  }
  if (dsz < sizeof *ka) {
    zkErrorSet_va(err, "fr_KeyQrySignDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "dsz is too small (%u < %u)", dsz, sizeof *ka);
    return 0;
  }

  keyDataSz = dsz - sizeof *ka;
  keyDataPtr = (char *)ka + sizeof *ka;

  /*
   *  Get the key signer's entity.
   */
  
  if ( ! frCryptPubKeyGetEntity(err, sig, &sigOwner)) {
    goto done;
  }

  /*
   *  Check the relationship between the key owner and the key signer.
   */

  fr_EntFromPacked(&keyOwner, ka->keyOwnerEntity);
  if ( ! (frCryptPubKeyBufGetType(err, keyDataPtr, keyDataSz, &keyType) &&
          frEntIsKeySigningParent(err, &keyOwner,
                                  keyType == FRCRYPT_KEYTYPE_CRYPTION, &sigOwner))) {
    goto done;
  }
  
  /*
   *  We won't sign a key that's already expired.
   */

  {
#ifdef ZK_KEY_EXPIRATION_CHECK
    ZkClock now = zkClockNow();
#endif
    ZkClock exp;

    if ( ! frCryptPubKeyBufGetExpiration(err, keyDataPtr, keyDataSz, &exp)) {
      goto done;
    }

#ifdef ZK_KEY_EXPIRATION_CHECK
    if (zkClockCmp(exp, now) < 0) {
      zkErrorSet_va(err, "fr_KeyQrySignDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                    "refusing to sign expired key (%Ld < %Ld)", zkClockToSeconds(exp), zkClockToSeconds(now));
      goto done;
    }
#else
#ifndef WIN32
#warning "NOT CHECKING KEY EXPIRATION DATE"
#endif
#endif
  }

  if ( ! frEntTypeIsValid(err, sigOwner.type, &valid)) {
    goto done;
  }
  if ( ! valid) {
    zkErrorSet(err, "fr_KeyQrySignDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "invalid signer for key");
    goto done;
  }
  
  if ( ! frEntStore(err, &sigOwner, &ka->keySignerEntity, sizeof ka->keySignerEntity)) {
    goto done;
  }

  {
    size_t ssz = dsz - sizeof ka->signerSig;
    void *sdat = (char *)d + sizeof ka->signerSig;
    
#if DEBUG_SIGN
    fprintf(stderr, "**** %s to sign %u bytes\n", sigOwner->name, ssz);
#endif

    if ( ! frCryptSigSignData(err, sig, sdat, ssz,
                              ka->signerSig, sizeof ka->signerSig)) {
      goto done;
    }
  }

  ret = 1;

 done:
  
  return ret;
}


DLLExport int fr_KeyQryVerifyDbDatum(ZkError err, FrCryptSig sig, void *d, size_t dsz)
{
  FrEntIdExp keyOwner, keySigner;
  int ret = 0;
  size_t keyDataSz;
  struct fr_keyqry_auth *ka = d;
  unsigned int keyType;
  void *keyDataPtr;
  
  if ( ! sig) {
    zkErrorSet(err, "fr_KeyQryVerifyDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "sig is null");
    return 0;
  }
  if ( ! d) {
    zkErrorSet(err, "fr_KeyQryVerifyDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "d is null");
    return 0;
  }
  if (dsz < sizeof *ka) {
    zkErrorSet_va(err, "fr_KeyQryVerifyDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "sz is too small (%lu < %lu)",
                  (unsigned long)dsz, (unsigned long)sizeof *ka);
    return 0;
  }

  keyDataSz = dsz - sizeof *ka;
  keyDataPtr = (char *)ka + sizeof *ka;

  /*
   *  Check the relationship between the key owner and the key signer.
   */

  if ( ! frCryptPubKeyGetEntity(err, sig, &keySigner)) {
    goto done;
  }
  fr_EntFromPacked(&keyOwner, ka->keyOwnerEntity);
  if ( ! (frCryptPubKeyBufGetType(err, keyDataPtr, keyDataSz, &keyType) &&
          frEntIsKeySigningParent(err, &keyOwner,
                                  keyType == FRCRYPT_KEYTYPE_CRYPTION, &keySigner))) {
    goto done;
  }
  
  /*
   *  A key that has already expired will fail verification.  So too will a
   *  key whose signer's key has expired.
   */

  {
#ifdef ZK_KEY_EXPIRATION_CHECK
    ZkClock now = zkClockNow();
#endif
    ZkClock exp;

    if ( ! frCryptPubKeyBufGetExpiration(err, keyDataPtr, keyDataSz, &exp)) {
      goto done;
    }

#ifdef ZK_KEY_EXPIRATION_CHECK
    if (zkClockCmp(exp, now) < 0) {
      zkErrorSet_va(err, "fr_KeyQryVerifyDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                    "key has expired (%Ld < %Ld)", zkClockToSeconds(exp), zkClockToSeconds(now));
      return 0;
    }
#else
#ifndef WIN32
#warning "NOT CHECKING KEY EXPIRATION DATE"
#endif
#endif

    if ( ! frCryptPubKeyGetExpiration(err, sig, &exp)) {
      goto done;
    }

#ifdef ZK_KEY_EXPIRATION_CHECK
    if (zkClockCmp(exp, now) < 0) {
      zkErrorSet_va(err, "fr_KeyQryVerifyDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                    "signer's key has expired (%Ld < %Ld)", zkClockToSeconds(exp), zkClockToSeconds(now));
      return 0;
    }
#else
#ifndef WIN32
#warning "NOT CHECKING KEY EXPIRATION DATE"
#endif
#endif
  }

  {
    int valid;
    size_t ssz = dsz - sizeof ka->signerSig;
    void *sdat = (char *)d + sizeof ka->signerSig;
    
#if 0
#if DEBUG_SIGN
    fprintf(stderr, "**** will verify %u bytes signed by %s\n", ssz, signer->name);
#endif
#endif

    if ( ! frCryptSigVerifyData(err, sig, sdat, ssz,
                                ka->signerSig, sizeof ka->signerSig, &valid)) {
      goto done;
    }
    if ( ! valid) {
      zkErrorSet(err, "fr_KeyQryVerifyDbDatum", ZKERR, ZKMFRKEYQRY, ZKERROR,
                 "invalid signature");
      goto done;
    }
  }

  ret = 1;

 done:
  
  return ret;
}


DLLExport int fr_KeyQryDbDatumPrint(ZkError err, FILE *fp, void *datum, size_t dsz)
{
  FrCryptPubKey pk = 0;
  int ret = 0;
  struct fr_keyqry_auth *ka = datum;

  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, "fr_KeyQryDbDatumPrint", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "fp is null");
    return 0;
  }
  if ( ! datum) {
    zkErrorSet(err, "fr_KeyQryDbDatumPrint", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "datum is null");
    return 0;
  }
  if (dsz < sizeof *ka) {
    zkErrorSet_va(err, "fr_KeyQryDbDatumPrint", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "datum size is incorrect (%u < %u)", dsz, sizeof *ka);
    return 0;
  }

  if ( ! frCryptSigBufPrint(err, fp, ka->signerSig, sizeof ka->signerSig)) {
    goto done;
  }

  fprintf(fp, "Owner: ");
  (void)frEntBufPrint(err, fp, ka->keyOwnerEntity, sizeof ka->keyOwnerEntity);
  fprintf(fp, "\n");

  fprintf(fp, "Signer: ");
  (void)frEntBufPrint(err, fp, ka->keySignerEntity, sizeof ka->keySignerEntity);
  fprintf(fp, "\n");

  {
    size_t dlen = dsz - sizeof *ka;
    
    if ( ! (frCryptPubKeyNew(err, &pk) &&
            frCryptPubKeyReadFromBuf(err, pk, (char *)datum + sizeof *ka, &dlen) &&
            frCryptPubKeyPrint(err, pk, fp))) {
      goto done;
    }
  }
  fputs("\n", fp);

  ret = 1;

 done:

  if (pk) { frCryptPubKeyFree(0, &pk); }

  return ret;
}


/*
 *  Print a condensed summary of the key contained in the datum.
 */
DLLExport int fr_KeyQryDbDatumKeyShortPrint(ZkError err, FILE *fp, void *datum, size_t dsz)
{
  FrCryptPubKey pk = 0;
  const char *fn = "fr_KeyQryDbDatumShortPrint";
  int ret = 0;
  struct fr_keyqry_auth *ka = datum;

  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "fp is null");
    return 0;
  }
  if ( ! datum) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "datum is null");
    return 0;
  }
  if (dsz < sizeof *ka) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "datum size is incorrect (%u < %u)", dsz, sizeof *ka);
    return 0;
  }

  {
    size_t dlen = dsz - sizeof *ka;
    
    if ( ! (frCryptPubKeyNew(err, &pk) &&
            frCryptPubKeyReadFromBuf(err, pk, (char *)datum + sizeof *ka, &dlen) &&
            frCryptPubKeyShortPrint(err, pk, 0, fp))) {
      goto done;
    }
  }

  ret = 1;

 done:

  if (pk) { frCryptPubKeyFree(0, &pk); }

  return ret;
}


DLLExport int fr_KeyQryDbKeyPrint(ZkError err, FILE *fp, void *key, size_t ksz)
{
  const char *fn = "fr_KeyQryDbKeyPrint";
  struct fr_keyqry_qry *q = key;
  
  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "fp is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "key is null");
    return 0;
  }
  if (ksz != sizeof *q) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "key size is incorrect (%u != %u)", ksz, sizeof *q);
    return 0;
  }

  fprintf(fp, "Version: %u\n", (unsigned int)q->keyVersion);
  fprintf(fp, "Type: %s (%u)\n",
          frCryptPubKeyTypeStr(q->keyType), (unsigned int)q->keyType);
  fprintf(fp, "Owner: ");
  frEntBufPrint(err, fp, q->keyOwner, sizeof q->keyOwner);
  fprintf(fp, "\n");

  return 1;
}
