/*
 *  $Id: qry.c,v 1.50.10.3 2000/12/14 17:52:47 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Routines to send key queries, receive responses and manage a local key
 *  cache.
 *  
 *  bug: responses indicating a key was not found, the query was invalid, etc.
 *  should be signed by the key server, to ensure they can't be spoofed.
 *  Therefore, requests should contain a random number that the server should
 *  include in its responses (in order to prevent replay attacks).
 *  
 *  bug: it would be useful to have a routine, that given a public key object,
 *  will insert it into the cache (potentially with the option to never
 *  expire).  This could be used to insert old nym keys, for example.
 */

#define DEBUG_AIP_PRUNE 0 /* pruning seems connected with AIP core dump */
#define DEBUG_CORRUPT_QUERY 0 /* set for debugging kqd */
#define DEBUG_NEG_CACHE 0 /* don't do negative caching */


/* The key is exported from the crypto DLL and used (imported) by the other, keyqry.DLL in this case*/
#undef EXPORTMASTERKEY

#ifdef WIN32
	#ifdef _DEBUG
		#pragma comment(lib, "fcryptd.lib")
	#else
		#pragma comment(lib, "fcryptr.lib")
	#endif
#endif

#include "zks.h"

#include <fr_sig.h>
#include <errno.h>
#include <stdarg.h>
#include <stddef.h>
#include <string.h>
#include <time.h>
#include "fr_keyqry_wire.h"
#include "frCrypt.h"
#include "frKeyQry.h"
#include "frPktFmt.h"
#include "frSec.h"
#include "zkAlloc.h"
#include "zkClock.h"
#include "zkHash.h"


/*
 *  This is the cache lifetime of keys that have already expired.
 */
#define FRKQ_EXPKEY_EXPTIME (3 * 60 * 60)


/*
 *  This structure maintains a local cache of keys, indexed on the keys'
 *  owners.
 */
struct fr_keyqry {
  ZkError cbErr; /* error object for callback functions */
  ZkHash cache;
};


/*
 *  A local cache datum.  Only when the state is FRKEYQRY_ST_VALID will
 *  keyInfo be a valid pointer.  The pointer `reply' will be valid when the
 *  state is FRKEYQRY_ST_AWAITING_VALIDATION, otherwise it will be null.
 *  
 *  bug: it may be worthwhile to have separate keyInfo and reply fields.  If a
 *  reply comes in to replace an existing key we wouldn't want to do the
 *  replacement if it's invalid (i.e. throw out a good key).  But then we'd
 *  also need to save the key's state as well as that of the new reply, and be
 *  able to back out of a failed update.  Also, the union is pretty ugly.
 */
struct cache_entry {
  FrKeyQryState state;
  size_t replyLen; /* full length of data, when `reply' is in use. */
  uint32_t expires; /* expiration date of key (for convenience) */
  ZkClock expTime;  /* cache entry expiration time */
  struct fr_keyqry_reply_pkt *reply;
  struct cache_key_info *keyInfo;
};


static const size_t dflCacheSize = 10000;
static int traceQry = 0;


static int findKey(ZkError err, FrKeyQry kq, int fd, struct sockaddr *addr, FrKeyQryQueryExp *qry, struct cache_entry **ce);


/*
 *
 *
 *                            Internal routines.
 *
 *
 */


/*
 *  bug: this is a hack to allow queries to be traced, both on Windows and
 *  UNIX.
 */
static void traceQuery(const char *fmt, ...)
{
  char buf[512];
  int ret;
  va_list ap;

  va_start(ap, fmt);
  ret = vsnprintf(buf, sizeof buf, fmt, ap);
  va_end(ap);

  ClientTrace("", buf);

  fputs(buf, stderr);
}


static int keyReqEqual(const FrKeyQryQueryExp *q0, const FrKeyQryQueryExp *q1)
{
  /* bug: ensure unused characters in entity name are nul. */
  
  return (q0->keyVersion == q1->keyVersion &&
          q0->keyType == q1->keyType &&
          q0->keyOwner.type == q1->keyOwner.type &&
          memcmp(q0->keyOwner.name, q1->keyOwner.name, sizeof q0->keyOwner.name) == 0);
}


/*
 *  Our return value follows the same pattern as memcmp(), et. al.
 */
static int cacheKeyEqual(const void *a, const void *b, size_t sz)
{
  const FrKeyQryQueryExp *qa = a, *qb = b;

  (void)sz; /* unused */
  return ! keyReqEqual(qa, qb);
}


/*
 *  Free a cache entry key, which is a fr_key_req structure.
 */
static int cacheKeyFree(ZkHash h, void *key, void *userData)
{
  (void)h; (void)userData; /* remove compiler warnings */

  free(key);
  return 1;
}


/*
 *  This is an internal routine; we assume that `ki' is valid.
 */
static int keyInfoFree(ZkError err, struct cache_key_info *ki)
{
  zkErrorClear(err);

  if (ki->key) {
    if (traceQry) {
      traceQuery("***** kif: freeing %s's key\n", ki->keyOwner.name);
    }
    frCryptPubKeyFree(err, &ki->key);
  }
  memset(ki, 0, sizeof *ki);
  free(ki);
  
  return 1;
}


/*
 *  Free a cache datum, which is a cache_entry structure.
 *  
 *  We free the public key, itself.  If the caller needs it, they should make
 *  a copy before freeing the cache.
 */
static int cacheDatumFree(ZkHash h, void *datum, void *userData)
{
  ZkError err = userData;
  int ret = 1;
  struct cache_entry *ce = datum;

  (void)h; /* remove compiler warning */

#if DEBUG_AIP_PRUNE
  if ( ! ce->keyInfo) {
    fprintf(stderr, "%s: pruning entry without key.\n",
            "cacheDatumFree");
  } else {
    void *p = ce->keyInfo ? ce->keyInfo->key : 0;
    char buf[100];

    buf[0] = '\0';
    (void)frEntSnprint(0, buf, sizeof(buf) - 1, &ce->keyInfo->keyOwner);
    fprintf(stderr, "%s: pruning `%s', key ptr: %p.\n",
            "cacheDatumFree", buf, p);
  }
#endif

  if (ce->reply) { free(ce->reply); }
  if (ce->keyInfo) { ret = keyInfoFree(err, ce->keyInfo); }

  memset(ce, 0, sizeof *ce);
  free(datum);

  return ret;
}


int makeQueryPkt(ZkError err, struct fr_keyqry_qry_pkt *pkt, FrKeyQryQueryExp *qry)
{
  zkErrorClear(err);

  memset(pkt->rand,
         FRCRYPT_RAND_SIZE - (sizeof pkt->rand % FRCRYPT_RAND_SIZE),
         sizeof pkt->rand);
  pkt->vers = 1; /* bug */
  pkt->type = FRPKTTYPE_KEYQRY;
  memset(pkt->spec,
         FRCRYPT_RAND_SIZE - (sizeof pkt->spec % FRCRYPT_RAND_SIZE),
         sizeof pkt->spec);

  pkt->query.keyVersion = qry->keyVersion;
  pkt->query.keyType = qry->keyType;

  if ( ! frEntStore(err, &qry->keyOwner,
                    pkt->query.keyOwner, sizeof pkt->query.keyOwner)) {
    return 0;
  }

#if DEBUG_CORRUPT_QUERY
  {
    /*
     *  Corrupt a random byte in the body of the query. (Not the header, since
     *  that's too easily rejected by the server.)
     */
    char *c = (void *)&pkt->query;
    uint32_t o, v; /* random offset and value for corruption */

    fprintf(stderr, "***** corrupting query *****\n");
    if ( ! (zkRandGetUInt32InRange(err, 0, sizeof pkt->query - 1, &o) &&
            zkRandGetUInt32InRange(err, 0, 255, &v))) {
      return 0;
    }
    c[o] = v;
  }
#endif

  return 1;
}


/*
 *  The following table defines the time-to-live (TTL) for various key types.
 *  
 *  Exceptions to this are the master key, which should never be expired, and
 *  cache entries with no associated key, which should have a very short TTL.
 */

#if 1

#define MINUTE(n) ((n) * 60)
#define HOUR(n) (MINUTE(n) * 60)

#else

#warning "***** using bogus cache expiration times for testing *****"
#define MINUTE(n) (n)
#define HOUR(n) (n)

#endif


static struct {
  FrEntIdType type;
  ZkClock exp;
} expTab[] = {
  { FRENT_TYPE_AIP,           ZKCLOCK_INIT(HOUR(6)) },
  { FRENT_TYPE_FMG,           ZKCLOCK_INIT(HOUR(6)) },
  { FRENT_TYPE_FREEDOM_MONTH, ZKCLOCK_INIT(HOUR(24)) },
  { FRENT_TYPE_FREEDOM_YEAR,  ZKCLOCK_INIT(HOUR(48)) },
  { FRENT_TYPE_KEYQRY_SRV,    ZKCLOCK_INIT(HOUR(6)) },
  { FRENT_TYPE_KEYUPD_SRV,    ZKCLOCK_INIT(HOUR(6)) },
  { FRENT_TYPE_MAIP,          ZKCLOCK_INIT(HOUR(6)) },
  { FRENT_TYPE_NIDB,          ZKCLOCK_INIT(HOUR(6)) },
  { FRENT_TYPE_NIQS,          ZKCLOCK_INIT(HOUR(6)) },
  { FRENT_TYPE_NISS,          ZKCLOCK_INIT(HOUR(6)) },
  { FRENT_TYPE_NNTP_SRV,      ZKCLOCK_INIT(HOUR(6)) },
  { FRENT_TYPE_NYM,           ZKCLOCK_INIT(HOUR(3)) },
  { FRENT_TYPE_NYM_SRV,       ZKCLOCK_INIT(HOUR(6)) },
  { FRENT_TYPE_SER_SRV,       ZKCLOCK_INIT(HOUR(6)) },
  { FRENT_TYPE_TOK_SRV,       ZKCLOCK_INIT(HOUR(6)) },
  
  { 0, ZKCLOCK_INIT(0) } /* must be last */
};


/*
 *  Given a pointer to a cache entry, set its expiration time.
 */
static int setCacheExpTime(ZkError err, struct cache_entry *ce)
{
  ZkClock now = zkClockNow();
  const char *fn = "setCacheExpTime";
  int ret = 0;
  
  zkErrorClear(err);

  if ( ! ce) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "ce is null");
    goto done;
  }

#if 0 /* NRDEBUG */
  ce->expTime = zkClockAdd(now, zkClockFromSeconds(MINUTE(1))); /* default, if not found in table */
#else
  if (ce->state != FRKEYQRY_ST_VALID) {
    ce->expTime = zkClockFromSeconds(MINUTE(5));
  } else {
    FrEntIdType kot = ce->keyInfo->keyOwner.type;

    if (kot == FRENT_TYPE_MASTER) {
      ce->expTime = zkClockMaxPos(); /* never expire from cache */
    } else {
      int i;
    
      ce->expTime = zkClockAdd(now, zkClockFromSeconds(MINUTE(30))); /* default, if not found in table */

      for (i = 0; zkClockToSeconds(expTab[i].exp) != 0; i++) {
        if (kot == expTab[i].type) {
          ce->expTime = zkClockAdd(now, expTab[i].exp);
          break;
        }
      }
    }
  }
#endif

  ret = 1;

 done:

  return ret;
}


#undef MINUTE
#undef HOUR


static int makeCacheEntry(ZkError err, FrKeyQry kq, FrKeyQryQueryExp *qry, struct cache_entry **cacheEntry)
{
  FrKeyQryQueryExp *ck = 0;
  int ret = 0;
  struct cache_entry *ce = 0;
  
  zkErrorClear(err);

  if ( ! ((ck = malloc(sizeof *ck)) &&
          (ce = malloc(sizeof *ce)))) {
    zkErrorSet_va(err, "makeCacheEntry", ZKERR, ZKMFRKEYQRY, errno,
                  "can't allocate byte cache entry: %s", strerror(errno));
    goto done;
  }

  *ck = *qry;

  ce->state = FRKEYQRY_ST_AWAITING_RESPONSE;
  ce->replyLen = 0;
  ce->expires = 0;
  ce->reply = 0;
  ce->keyInfo = 0;

  if ( ! zkHashInsert(err, kq->cache, ck, ce)) {
    goto done;
  }

  if ( ! setCacheExpTime(err, ce)) {
    goto done;
  }

  if (traceQry) {
    traceQuery("***** mce: made new cache entry for %s, cache entry %p.\n",
            ck->keyOwner.name, (void *)ce);
  }

  *cacheEntry = ce;
  ck = 0; ce = 0; /* the cache now `owns' them */

  ret = 1;

 done:

  if (ck) { free(ck); }
  if (ce) { free(ce); }

  return ret;
}


/*
 *  Return 1 if the signature over `reply' is validated with the key in `pki'
 *  (parent key info).
 */
DLLExport int valid_parent_sig(ZkError err, struct cache_key_info *pki, 
                               struct fr_keyqry_reply_pkt *reply, 
                               size_t replySz)
{
  FrEntIdExp childEntity;
  FrPubKeyType keyType;
  struct fr_keyqry_auth *auth;
  size_t keyDataSz, sigSz;
  void *keyDataPtr, *sigPtr;
  
  zkErrorClear(err);

  /*
   *  Set up some pointers and sizes for convenience.
   */

  /* authentication information */

  auth = (void *)((char *)reply + sizeof *reply);

  /* key data */
  
  keyDataPtr = (char *)auth + sizeof *auth;
  keyDataSz = replySz - (sizeof *reply + sizeof *auth);

  /* signed data */
  
  sigPtr = (char *)auth + offsetof(struct fr_keyqry_auth, keyOwnerEntity);
  sigSz = replySz - ((char *)sigPtr - (char *)reply);

  /*
   *  Check the entity type of the signer.
   */

  fr_EntFromPacked(&childEntity, auth->keyOwnerEntity);

  if ( ! (frCryptPubKeyBufGetType(err, keyDataPtr, keyDataSz, &keyType) &&
          frEntIsKeySigningParent(err, &childEntity,
                                  keyType == FRCRYPT_KEYTYPE_CRYPTION,
                                  &pki->keyOwner))) {
    return 0;
  }

  {
    int valid;
    
    if (traceQry) {
      traceQuery("***** vps: verify %u bytes of signed data\n", sigSz);
      traceQuery("           using %s's signature key\n", pki->keyOwner.name);
      traceQuery("           on %s's key data\n", &auth->keyOwnerEntity[1]);
    }

    if ( ! frCryptSigVerifyData(err, pki->key, sigPtr, sigSz,
                                auth->signerSig, sizeof auth->signerSig, &valid)) {
      return 0;
    }

    if (traceQry) {
      traceQuery("           it's %svalid.\n", valid ? "" : "not ");
    }

    if ( ! valid) {
      zkErrorSet(err, "valid_parent_sig", ZKWARN, ZKMFRKEYQRY, ZKEDATABAD,
                 "invalid signature");
      return 0;
    }
  }

  return 1;
}


static int query(ZkError err, int fd, struct sockaddr *addr, FrKeyQryQueryExp *qry)
{
  const char *fn = "query";
  int ret = 0;
  struct fr_keyqry_qry_pkt qpkt;

  zkErrorClear(err);

  ret = makeQueryPkt(err, &qpkt, qry);

  if (ret) {
    errno = 0;
    if (sendto(fd, &qpkt, sizeof qpkt, 0, addr, sizeof *addr) != sizeof qpkt) {
      zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, errno,
                    "sendto() failed on %lu byte query: %s",
                    (unsigned long)sizeof qpkt, strerror(errno));
      ret = 0;
    }
  }

  return ret;
}


/*
 *  bug: the procedure for converting the key data to an actual key is ugly.
 *  In order to set the appropriate shared, fixed key parameters we need to
 *  know the key type.
 */
DLLExport int key_from_buf(ZkError err, FrCryptPubKey *pk, 
                           void *buf, size_t bsz)
{
  FrPubKeyType keyType;
  int ret = 0;
  
  zkErrorClear(err);
  
  if ( ! frCryptPubKeyBufGetType(err, buf, bsz, &keyType)) {
    goto done;
  }

  switch (keyType) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    if ( ! frCryptCryptNewFromData(err, pk, buf, &bsz)) {
      goto done;
    }
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    if ( ! frCryptSigNewFromData(err, pk, buf, &bsz)) {
      goto done;
    }
    break;

  default:
    zkErrorSet_va(err, "key_from_buf", ZKERR, ZKMFRKEYQRY, ZKEDATABAD,
               "invalid key type (%d)", keyType);
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 *  A reply has been validated with respect to its parent signing key.  We
 *  convert the reply to the actual key, and set the state accordingly.
 */
static int key_from_reply(ZkError err, struct cache_entry *ce)
{
  int ret = 0;
  size_t keySz = ce->replyLen - (sizeof(struct fr_keyqry_reply_pkt) +
                                 sizeof(struct fr_keyqry_auth));
  struct cache_key_info *ki = 0;
  struct fr_keyqry_auth *auth = (void *)((char *)ce->reply + sizeof *ce->reply);
  void *keyData = (char *)ce->reply + (sizeof(struct fr_keyqry_reply_pkt) +
                                       sizeof(struct fr_keyqry_auth));
  
  zkErrorClear(err);

  if ( ! (ki = malloc(sizeof *ki))) {
    zkErrorSet_va(err, "key_from_reply", ZKERR, ZKMFRKEYQRY, errno,
                  "can't allocate memory for cache key info: %s",
                  strerror(errno));
    goto done;
  }
  if ( ! key_from_buf(err, &ki->key, keyData, keySz)) {
    goto done;
  }
  fr_EntFromPacked(&ki->keyOwner, auth->keyOwnerEntity);

  if (traceQry) {
    traceQuery("***** kfr: new key from reply (%s).\n", ki->keyOwner.name);
  }

  ce->keyInfo = ki; ki = 0;
  ce->state = FRKEYQRY_ST_VALID;

  if ( ! setCacheExpTime(err, ce)) {
    goto done;
  }

  ret = 1;

 done:

  if (ki) { keyInfoFree(0, ki); }

  return ret;
}


/*
 *  In order to ascertain the status of a key, query for its parent.  Update
 *  the child key's cache entry if appropriate.
 */
static int query_parent(ZkError err, FrKeyQry kq, int fd, struct sockaddr *addr, FrKeyQryQueryExp *pqry, struct cache_entry *cce)
{
  struct cache_entry *pce; /* parent cache entry */
  int ret = 0;
  
  zkErrorClear(err);

  if ( ! findKey(err, kq, fd, addr, pqry, &pce)) {
    goto done;
  }

  switch (pce->state) {
  case FRKEYQRY_ST_VALID:
    if (valid_parent_sig(err, pce->keyInfo, cce->reply, cce->replyLen)) {
      if (traceQry) {
        traceQuery("***** qp: %s is validated.\n",
                &cce->reply->origQuery.keyOwner[1]);
      }
      if ( ! key_from_reply(err, cce)) {
        goto done;
      }
    } else {
      cce->state = FRKEYQRY_ST_INVALID; /* bug? free child key info? */
    }
    break;
      
  case FRKEYQRY_ST_INVALID:
  case FRKEYQRY_ST_NOT_FOUND:
    cce->state = FRKEYQRY_ST_INVALID; /* bug? free child key info? */
    break;

  case FRKEYQRY_ST_AWAITING_RESPONSE:
  case FRKEYQRY_ST_AWAITING_VALIDATION:
    /* We've received nothing to change the child's state. */
    break;
  }

  ret = 1;

 done:

  return ret;
}


static int findKey(ZkError err, FrKeyQry kq, int fd, struct sockaddr *addr, FrKeyQryQueryExp *qry, struct cache_entry **ce)
{
  int ret = 0;
  struct cache_entry *cacheEntry;
  
  zkErrorClear(err);

  if ( ! kq) {
    zkErrorSet(err, "findKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "kq is null");
    return 0;
  }
  if (fd < 0) {
    zkErrorSet_va(err, "findKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "fd (%d) is invalid", fd);
    return 0;
  }
  if ( ! addr) {
    zkErrorSet(err, "findKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "addr is null");
    return 0;
  }
  if ( ! qry) {
    zkErrorSet(err, "findKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "qry is null");
    return 0;
  }
  if ( ! ce) {
    zkErrorSet(err, "findKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "ce is null");
    return 0;
  }

  *ce = 0;

  if ( ! zkHashSearch(err, kq->cache, qry, (void **)&cacheEntry)) {
    /*
     *  Insert an entry into the local cache, then create a query packet and
     *  send it to the query server.
     */
    
    if (traceQry) {
      traceQuery("***** fk: didn't find %s in cache -- querying.\n",
              qry->keyOwner.name);
    }
    
    if ( ! (query(err, fd, addr, qry) &&
            makeCacheEntry(err, kq, qry, ce))) {
      goto done;
    }
  } else {
    if (traceQry) {
      traceQuery("***** fk: found %s in cache (%p, %p).\n",
              qry->keyOwner.name,
              (void *)cacheEntry, (void *)cacheEntry->reply);
    }
    
    switch (cacheEntry->state) {
    case FRKEYQRY_ST_VALID:
      *ce = cacheEntry;
      if (traceQry) {
        traceQuery("          it's valid.\n");
      }
      break;
      
    case FRKEYQRY_ST_INVALID:
    case FRKEYQRY_ST_NOT_FOUND:
      *ce = cacheEntry;
      break;

    case FRKEYQRY_ST_AWAITING_RESPONSE:
      /*
       *  Send out another query for the key.
       */
      
      if (traceQry) {
        traceQuery("***** fk: awaiting response -- send another query.\n");
      }
    
      if ( ! query(err, fd, addr, qry)) {
        goto done;
      }
      *ce = cacheEntry;
      break;
      
    case FRKEYQRY_ST_AWAITING_VALIDATION:
      /*
       *  Check if the parent key has come in and has itself been validated.
       */
      {
        FrKeyQryQueryExp pqry; /* parent key query */
        struct fr_keyqry_auth *auth = (void *)((char *)cacheEntry->reply +
                                               sizeof *cacheEntry->reply);
        unsigned int pkv; /* parent's key version */

        if ( ! frCryptSigBufGetVersion(err, auth->signerSig,
                                       sizeof auth->signerSig, &pkv)) {
          goto done;
        }
        
        pqry.keyVersion = pkv;
        pqry.keyType = FRCRYPT_KEYTYPE_SIGNING;
        pqry.keyOwner.type = auth->keySignerEntity[0];
        memcpy(pqry.keyOwner.name,
               &auth->keySignerEntity[1],
               sizeof pqry.keyOwner.name);

        if (traceQry) {
          traceQuery("***** fk: awaiting validation -- query parent %s.\n",
                  pqry.keyOwner.name);
          traceQuery("          where the child is %s (%p, %p).\n",
                  &cacheEntry->reply->origQuery.keyOwner[1],
                  (void *)cacheEntry, (void *)cacheEntry->reply);
        }
    
        if ( ! query_parent(err, kq, fd, addr, &pqry, cacheEntry)) {
          goto done;
        }

        if (traceQry) {
          traceQuery("***** fk: after querying parent %s.\n",
                  pqry.keyOwner.name);
          traceQuery("          the child is %s (%p, %p).\n",
                  &cacheEntry->reply->origQuery.keyOwner[1],
                  (void *)cacheEntry, (void *)cacheEntry->reply);
        }

        *ce = cacheEntry;
        break;
      }
    }
  }

  ret = 1;
  
 done:

  if (traceQry) {
    traceQuery("\n");
  }

  return ret;
}


/*
 *  Mark an entry as invalid if it has no associated key, otherwise leave it
 *  alone.
 *  
 *  bug: to mark an existing key as invalid we probably want the key query
 *  server's signature on the response, otherwise we have a possible
 *  denial-of-service attack.
 */
static int maybe_invalidate_entry(struct cache_entry *ce)
{
  if ( ! (ce->reply || ce->keyInfo)) {
    ce->state = FRKEYQRY_ST_INVALID;
  }

  return 1;
}


static int free_cache_entry_key(ZkError err, struct cache_entry *ce)
{
  zkErrorClear(err);

  if (ce->reply) {
    free(ce->reply); ce->reply = 0;
  }
  
  if (ce->keyInfo) {
    int ret = keyInfoFree(err, ce->keyInfo);
    ce->keyInfo = 0;
    return ret;
  } else {
    return 1;
  }
}


/*
 *  If there is no key or reply associated with the cache entry, then insert
 *  the new one, marking it as awaiting validation.  Otherwise, only remove
 *  the existing key or reply if the expiration date on the new one is greater
 *  than that of the existing entry, in which case it is marked as awaiting
 *  validation.
 */
static int maybe_insert_reply(ZkError err, struct cache_entry *ce, struct fr_keyqry_reply_pkt *reply, size_t rsz)
{
  int ret = 0;
  size_t keyDataSz;
#ifdef ZK_KEY_EXPIRATION_CHECK
  ZkClock now = zkClockNow();
#endif
  ZkClock newExp;
  void *keyDataPtr, *replyCopy;

  zkErrorClear(err);

  keyDataPtr = (char *)reply + sizeof *reply + sizeof(struct fr_keyqry_auth);
  keyDataSz = rsz - (sizeof *reply + sizeof(struct fr_keyqry_auth));
    
  /*
   *  Reject keys whose version, type and owner entity do not match that in
   *  the original query field of the response.
   */

  {
    FrEntIdExp kown, qown;
    FrPubKeyType ktype;
    int equal;
    unsigned int kvers;

    if ( ! (frEntLoad(err, &qown,
                      reply->origQuery.keyOwner,
                      sizeof reply->origQuery.keyOwner) &&
            frCryptPubKeyBufGetVersion(err, keyDataPtr, keyDataSz, &kvers) &&
            frCryptPubKeyBufGetType(err, keyDataPtr, keyDataSz, &ktype) &&
            frCryptPubKeyBufGetEntity(err, keyDataPtr, keyDataSz, &kown))) {
      goto done;
    }

    /* bug: is checking for 0 okay? */
    if ( ! ((reply->origQuery.keyVersion == kvers || reply->origQuery.keyVersion == 0) &&
            reply->origQuery.keyType    == ktype &&
            frEntEqual(err, &qown, &kown, &equal) && equal)) {
      zkErrorSet(err, "maybe_insert_reply", ZKERR, ZKMFRKEYQRY, ZKERROR,
                 "key does not match original query");
      goto done;
    }
  }
  
  /*
   *  Grab the expiry date from the response.
   */
  
  if ( ! frCryptPubKeyBufGetExpiration(err, keyDataPtr, keyDataSz, &newExp)) {
    goto done;
  }

  /*
   *  Reject keys that have already expired.
   */
  
#ifdef ZK_KEY_EXPIRATION_CHECK
#warning "We probably want to accept expired keys in the cache."
#warning "Handling of expired keys needs to be reviewed"
  if (zkClockCmp(newExp, now) < 0) {
    zkErrorSet_va(err, "maybe_insert_reply", ZKWARN, ZKMFRKEYQRY, ZKEDATABAD,
                  "rejecting expired response (%ld < %ld)", (long) zkClockToSeconds(newExp), (long) zkClockToSeconds(now));
    goto done;
  }
#else
#ifndef WIN32
#warning "NOT CHECKING KEY EXPIRATION DATE"
#endif
#endif

  if (ce->reply || ce->keyInfo /* bug? */) {
    /*
     *  Replace the existing entry only if the new response has a greater
     *  expiry date.
     */
    
    if (zkClockCmp(newExp, zkClockFromSeconds(ce->expires)) <= 0) {
      return 1; /* ignore the new response */
    }

    ce->state = FRKEYQRY_ST_AWAITING_RESPONSE; /* in case of later failure */
    if ( ! free_cache_entry_key(err, ce)) {
      goto done;
    }
  }

  /*
   *  Either a new response is to be inserted, or the old one is to be
   *  replaced.  Regardless of which, the steps are the same.
   */

  if ( ! (replyCopy = malloc(rsz))) {
    zkErrorSet_va(err, "maybe_insert_reply", ZKERR, ZKMFRKEYQRY, errno,
                  "can't allocate %u bytes for copy of response: %s",
                  rsz, strerror(errno));
    goto done;
  }
  ce->state = FRKEYQRY_ST_AWAITING_VALIDATION;
  memcpy(replyCopy, reply, rsz);
  ce->reply = replyCopy; /* bug? check pkt for validity? */
  ce->replyLen = rsz;
  ce->expires = zkClockToSeconds(newExp);

  if (traceQry) {
    traceQuery("***** mir: set cache entry (%p, %p) to reply (%p).\n",
            (void *)ce, (void *)ce->reply, (void *)reply);
  }

  ret = 1;

 done:

  return ret;
}

/*
 *  Insert, into the local key cache, an entry for the master signing key.
 *  This lets us treat it (mostly) just like any other key.
 */
static int insert_master(ZkError err, FrKeyQry kq)
{
  FrCryptPubKey mk = frCryptSigZkMaster;
  FrKeyQryQueryExp mq; /* master query */
  char mastTypeStr[] = "master", mastNameStr[] = "ZkMaster"; /* bug */
  int ret = 0;
  struct cache_entry *mce = 0; /* master cache entry */
  unsigned int type, vers;

  zkErrorClear(err);

  if ( ! (frCryptPubKeyGetType(err, mk, &type) &&
          frCryptPubKeyGetVersion(err, mk, &vers))) {
    return 0;
  }

  mq.keyType = type;
  mq.keyVersion = vers;
  /* bug: need define for ZkMaster string */
  if ( ! (frEntSetFromStr2(err, &mq.keyOwner, mastTypeStr, mastNameStr) &&
          makeCacheEntry(err, kq, &mq, &mce))) {
    return 0;
  }
  
  mce->state = FRKEYQRY_ST_VALID;
  if ( ! (mce->keyInfo = malloc(sizeof *mce->keyInfo))) {
    zkErrorSet_va(err, "insert_master", ZKERR, ZKMFRKEYQRY, errno,
                  "error inserting master key into cache: %s", strerror(errno));
    goto done;
  }
  memset(mce->keyInfo, 0, sizeof *mce->keyInfo);
  mce->keyInfo->key = 0;
  
  if ( ! frCryptPubKeyDup(err, &mce->keyInfo->key, mk)) {
    goto done;
  }
  mce->keyInfo->keyOwner = mq.keyOwner;
  
  if ( ! setCacheExpTime(err, mce)) {
    goto done;
  }

  mce = 0; /* the cache now `owns' it */
  ret = 1;

 done:

  if (mce) { cacheDatumFree(kq->cache, mce, err); }

  return ret;
}


/*
 *  bug: currently doesn't print the datum.  Add a flag (passed via `arg')
 *  that indicates what to print?
 */
static int cachePrint(ZkError err, void *key, void *datum, void *arg)
{
  FILE *fp = arg;
  FrKeyQryQueryExp *ck = key;
  int r;
  
  (void)datum; /* unused */

  zkErrorClear(err);

  r = fprintf(fp, "query -- version %u, type %d, [%u, %s]\n",
              ck->keyVersion, ck->keyType,
              (unsigned int)ck->keyOwner.type, ck->keyOwner.name);
  
  if (r < 0) {
    zkErrorSet_va(err, "cachePrint", ZKERR, ZKMFRKEYQRY, errno,
                  "fprintf() failed: %s", strerror(errno));
    return 0;
  }
  
  return 1;
}


/*
 *
 *
 *                            External routines.
 *
 *
 */


DLLExport int frKeyQryNew(ZkError err, FrKeyQry *kq)
{
  FrKeyQry q = 0;
  int ret = 0;

  zkErrorClear(err);

  if ( ! kq) {
    zkErrorSet(err, "frKeyQryNew", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "kq is null");
    return 0;
  }

  if ( ! (q = malloc(sizeof *q))) {
    zkErrorSet_va(err, "frKeyQryNew", ZKERR, ZKMFRKEYQRY, ZKENOMEM,
                  "can't allocate %lu bytes for key query object: %s",
                  (unsigned long)sizeof *q, strerror(errno));
    goto done;
  }

  if ( ! zkErrorNew(&q->cbErr)) {
    goto done;
  }
  
  if ( ! (zkHashNew(err, &q->cache) &&
          zkHashSetTableSize(err, q->cache, dflCacheSize) &&
          zkHashSetKeySize(err, q->cache, sizeof(FrKeyQryQueryExp)) &&
          zkHashSetKeyEqualFn(err, q->cache, cacheKeyEqual) &&
          zkHashSetDataFreeFn(err, q->cache, cacheDatumFree) &&
          zkHashSetKeyFreeFn(err, q->cache, cacheKeyFree) &&
          zkHashSetUserData(err, q->cache, q->cbErr))) {
    goto done;
  }

  if ( ! insert_master(err, q)) {
    goto done;
  }

  *kq = q; q = 0;
  ret = 1;

 done:

  if (q) { frKeyQryFree(0, &q); }  


  return ret;
}


DLLExport int frKeyQryFree(ZkError err, FrKeyQry *kq)
{
  FrKeyQry q = 0;
  int ret = 1;
  
  zkErrorClear(err);

  if ( ! kq) {
    zkErrorSet(err, "frKeyQryFree", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "kq is null");
    return 0;
  }
  if ( ! (q = *kq)) {
    zkErrorSet(err, "frKeyQryFree", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "*kq is null");
    return 0;
  }

  /*
   *  Note that the cache must be freed before the callback error object,
   *  since it's used when freeing the hash table elements.
   */

  if (q->cache) { ret &= zkHashFree(err, &q->cache); }
  if (q->cbErr) { ret &= zkErrorFree(&q->cbErr); }

  free(q); *kq = 0;

  return ret;
}


/*
 *  Look for the key specified by `qry'.  If necessary, send requests to the
 *  key query server.
 */
DLLExport int frKeyQryFindKey(ZkError err, FrKeyQry kq, int fd, struct sockaddr *addr, FrKeyQryQueryExp *qry, FrKeyQryState *state, FrCryptPubKey *key)
{
  struct cache_entry *ce;
  
  zkErrorClear(err);

  *key = 0;
  
  if ( ! kq) {
    zkErrorSet(err, "frKeyQryFindKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "kq is null");
    return 0;
  }
  if (fd < 0) {
    zkErrorSet_va(err, "frKeyQryFindKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "fd (%d) is invalid", fd);
    return 0;
  }
  if ( ! addr) {
    zkErrorSet(err, "frKeyQryFindKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "addr is null");
    return 0;
  }
  if ( ! qry) {
    zkErrorSet(err, "frKeyQryFindKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "qry is null");
    return 0;
  }
  if ( ! state) {
    zkErrorSet(err, "frKeyQryFindKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "state is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, "frKeyQryFindKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "key is null");
    return 0;
  }

  if (findKey(err, kq, fd, addr, qry, &ce)) {
    *state = ce->state;
    if (ce->state == FRKEYQRY_ST_VALID) {
      *key = ce->keyInfo->key;
    }

#if ! DEBUG_NEG_CACHE
    /*
     *  bug: don't do negative caching.  If the cache entry is marked as
     *  invalid or not found, then remove it from the cache.  This is intended
     *  to be a temporary hack until we have a method to periodically prune
     *  the cache.
     */

    if ((ce->state == FRKEYQRY_ST_INVALID ||
         ce->state == FRKEYQRY_ST_NOT_FOUND) &&
        ! zkHashDelete(err, kq->cache, qry)) {
      return 0;
    }
#endif
    
    return 1;
  }

  return 0;
}


/*
 *  Look for the key specified by `qry'.  Only search the local cache, don't
 *  send any requests to the key query server.
 *  
 *  Return 1 if there is a cache entry for the requested item.  (This doesn't
 *  mean there is an associated key; it may be in a non-valid state
 *  [e.g. awaiting a response].)  Return 0 if there is no cache entry
 *  whatsoever.
 *  
 *  The `state' parameter will be valid only if the function is successful
 *  (i.e. returns 1), while the `key' parameter will be valid when `state' has
 *  the value FRKEYQRY_ST_VALID.
 */
DLLExport int frKeyQryCacheFindKey(ZkError err, FrKeyQry kq, FrKeyQryQueryExp *qry, FrKeyQryState *state, FrCryptPubKey *key)
{
  struct cache_entry *cacheEntry;
  
  zkErrorClear(err);

  if ( ! kq) {
    zkErrorSet(err, "frKeyQryCacheFindKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "kq is null");
    return 0;
  }
  if ( ! qry) {
    zkErrorSet(err, "frKeyQryCacheFindKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "qry is null");
    return 0;
  }
  if ( ! state) {
    zkErrorSet(err, "frKeyQryCacheFindKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "state is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, "frKeyQryCacheFindKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "key is null");
    return 0;
  }

  if (zkHashSearch(err, kq->cache, qry, (void **)&cacheEntry)) {
    *state = cacheEntry->state;
    if (cacheEntry->state == FRKEYQRY_ST_VALID) {
      *key = cacheEntry->keyInfo->key;
    }
    return 1;
  }

  return 0;
}


/*
 *  Process a response from the key query server.
 */
DLLExport int frKeyQryRecvKey(ZkError err, FrKeyQry kq, int fd, struct sockaddr *addr, void *pkt, size_t pktLen)
{
  int ret = 0;
  FrKeyQryQueryExp cacheQuery;
  struct cache_entry *cacheEntry = 0;
  struct fr_keyqry_reply_pkt *reply = pkt;

  zkErrorClear(err);

  if ( ! kq) {
    zkErrorSet(err, "frKeyQryRecvKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "kq is null");
    return 0;
  }
  if (fd < 0) {
    zkErrorSet_va(err, "frKeyQryRecvKey", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "fd (%d) is invalid", fd);
    return 0;
  }
  if ( ! addr) {
    zkErrorSet(err, "frKeyQryRecvKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "addr is null");
    return 0;
  }
  if ( ! pkt) {
    zkErrorSet(err, "frKeyQryRecvKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "pkt is null");
    return 0;
  }
  
  if (pktLen < sizeof *reply) {
    zkErrorSet_va(err, "frKeyQryRecvKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "reply packet is too short (%u < %u)", pktLen, sizeof *reply);
    return 0;
  }

  if ( ! (reply->vers == 1 /* bug */ &&
          reply->type == FRPKTTYPE_KEYRESP)) {
    zkErrorSet_va(err, "frKeyQryRecvKey", ZKERR, ZKMFRKEYQRY, ZKEDATABAD,
                  "received an invalid packet (%u, %u)",
                  reply->vers, reply->type);
    return 0;
  }

  /* Look up the key in the local cache. */

  cacheQuery.keyVersion = reply->origQuery.keyVersion;
  cacheQuery.keyType = reply->origQuery.keyType;
  fr_EntFromPacked(&cacheQuery.keyOwner, reply->origQuery.keyOwner);

  if (traceQry) {
    traceQuery("***** rk: got response for %s.\n", cacheQuery.keyOwner.name);
  }

  if (zkHashSearch(err, kq->cache, &cacheQuery, (void **)&cacheEntry)) {
    if (cacheQuery.keyOwner.type == FRENT_TYPE_MASTER) {
      zkErrorSet(err, "frKeyQryRecvKey", ZKERR, ZKMFRKEYQRY, ZKEDATABAD,
                 "rejecting packet containing master key");
      goto done;
    }
    if (traceQry) {
      traceQuery("          it has a cache entry (%p, %p).\n",
              (void *)cacheEntry, (void *)cacheEntry->reply);
    }
  } else {
    if (traceQry) {
      traceQuery("          it has no cache entry (%p, %p).\n",
              (void *)cacheEntry, (void *)cacheEntry->reply);
    }
  }

  if ( ! cacheEntry) {
    zkErrorSet_va(err, "frKeyQryRecvKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "spurious response for [%u, %u, %u, %s]",
                  reply->origQuery.keyVersion, reply->origQuery.keyType,
                  reply->origQuery.keyOwner[0], &reply->origQuery.keyOwner[1]);
    goto done;
  }

  /*
   *  bug: In general, consider what should happen, in each case, if there is
   *  already an existing key or reply.
   *  
   *  bug: For any responses that would remove an existing entry from the
   *  table (e.g. not found, invalid?) we probably want the key server's
   *  signature on the response.
   */

  switch (reply->respCode) {
  case FR_KEYQRY_CODE_BADQUERY:
  case FR_KEYQRY_CODE_SRVERROR:
    if ( ! maybe_invalidate_entry(cacheEntry)) {
      goto done;
    }
    break;
    
  case FR_KEYQRY_CODE_FOUND:
    if ( ! maybe_insert_reply(err, cacheEntry, reply, pktLen)) {
      goto done;
    }
    break;
    
  case FR_KEYQRY_CODE_NOTFOUND:
    cacheEntry->state = FRKEYQRY_ST_NOT_FOUND;
    /* bug: existing entry? */
    break;
    
  default:
    zkErrorSet_va(err, "frKeyQryRecvKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "unknown response code (%u) from server",
                  (unsigned int)reply->respCode);
    goto done;
  }

  ret = 1;

  done:

  if (traceQry) {
    traceQuery("\n");
  }
  
  return ret;
}


/*
 *  Remove all entries from the cache.
 *  
 *  bug: code taken from Prune.
 */
DLLExport int frKeyQryCacheClear(ZkError err, FrKeyQry kq)
{
  ZkHashEnum e = 0;
  const char *fn = "frKeyQryCacheClear";
  int ret = 0;

  zkErrorClear(err);
  
  if ( ! kq) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "kq is null");
    goto done;
  }

  if ( ! (zkHashEnumNew(err, &e) &&
          zkHashEnumSetHash(err, e, kq->cache))) {
    goto done;
  }

  while (zkHashEnumNext(err, e)) {
    FrKeyQryQueryExp *qry;
    struct cache_entry *ce;

    if ( ! (zkHashEnumGetKey(err, e, (void **)&qry) &&
            zkHashEnumGetData(err, e, (void **)&ce) &&
            zkHashDelete(err, kq->cache, qry))) {
      goto done;
    }
  }

  ret = 1;

 done:

  if (e) { zkHashEnumFree(err, &e); }

  return ret;
}


/*
 *  Remove from the cache those entries that have expired.  (That is, those
 *  keys that have expired from the cache.)
 */
DLLExport int frKeyQryCachePrune(ZkError err, FrKeyQry kq)
{
  ZkClock now = zkClockNow();
  ZkHashEnum e = 0;
  const char *fn = "frKeyQryCachePrune";
  int ret = 0;

  zkErrorClear(err);
  
  if ( ! kq) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "kq is null");
    goto done;
  }

  if ( ! (zkHashEnumNew(err, &e) &&
          zkHashEnumSetHash(err, e, kq->cache))) {
    goto done;
  }

  while (zkHashEnumNext(err, e)) {
    FrKeyQryQueryExp *qry;
    struct cache_entry *ce;

    if ( ! (zkHashEnumGetKey(err, e, (void **)&qry) &&
            zkHashEnumGetData(err, e, (void **)&ce))) {
      continue; /* skip this one */
    }

    /*
    ** Never prune the Master Key, Our model consider it persistant.
    */
    if (qry->keyOwner.type != FRENT_TYPE_MASTER) {
      if (zkClockCmp(ce->expTime, now) < 0) {
        char buf[101];

        if ( ! frEntSnprint(err, buf, sizeof(buf) - 1, &qry->keyOwner)) {
          goto done;
        }
        zkErrorSet_va(err, fn, ZKINFO, ZKMFRKEYQRY, 0,
                      "expiring `%s' from local cache", buf);

        if ( ! zkHashDelete(err, kq->cache, qry)) {
          goto done;
        }
      }
    }
  }

  ret = 1;

 done:

  if (e) { zkHashEnumFree(err, &e); }

  return ret;
}


/*
 *  Set the maximum number of cache entries to allow.
 */
DLLExport int frKeyQryCacheSetMaxEntries(ZkError err, FrKeyQry kq, size_t nelts)
{
  (void) kq;
  (void) nelts;

  zkErrorClear(err);

  return 0;
}


/*
 *  Initialize the cache from an external source.
 *  
 *  We specifically ignore any key claiming to be the master key (having
 *  already inserted it ourselves), and mark all entries as having been
 *  validated.
 *  
 *  We set the states of the entries as FRKEYQRY_ST_AWAITING_VALIDATION
 *  so that a query will validate it (for security against file tampering).
 */
DLLExport int frKeyQryCacheRead(ZkError err, FrKeyQry kq, FrCryptSym sym, FILE *fp)
{
  FrKeyQryQueryExp qry;
  char *data;
  char *refData;
  size_t refDataSz;
  const char *fn = "frKeyQryCacheRead";
  int ret = 0;
  size_t dataSz;
  void *blockBuf = 0;
  unsigned int bEncrypted = 0;

  zkErrorClear(err);
  
  if ( ! kq) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "kq is null");
    return 0;
  }
  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "key is null");
    return 0;
  }
  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "fp is null");
    return 0;
  }
  

  /* decrypt the file, save info in 'data' */
//  if ( ! frSecDecryptFromFp(err, fp, sym, (void **)&data, &dataSz)) {

  if ( ! frSecReadFromFp(err, fp, sym, (void **)&data, &dataSz, &bEncrypted)) {
    return 0;
  }

  /*
   * The buffer is composed of blocks, each one composed of a
   *   unbol 'totsz' 
   *   + a fr_keyqry_reply_pkt 
   *   + a fr_keyqry_auth structure 
   *   + serialized key.
   * We treat each block of the buffer.
   */ 

  refData = data;
  refDataSz = dataSz;

  while (data < (refData + dataSz)) {
    FrEntIdExp keyOwner;
    ZkClock keyExp;
    size_t blockSz;
    struct cache_entry *ce = 0;  /* cache entry for the block */
    struct fr_keyqry_auth *authPtr = 0;
    struct fr_keyqry_reply_pkt *replyPtr = 0;
    uint16_t blockNbolSz;  /* size of unsigned, network byte order length (unbol) */
    void* keyDataPtr;

    /* get the size of the "block" */
    memcpy(&blockNbolSz, data, sizeof blockNbolSz);
    blockSz = ntohs(blockNbolSz);

    /* cp the block into 'blockBuf' */
    if ( ! (blockBuf = malloc(blockSz))) {
      zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, errno, "can't allocate buffer");
      goto done;
    }
    memcpy(blockBuf, data + (sizeof blockNbolSz), blockSz);

    /* point replyPtr to reply section in blockBuf */
    replyPtr = (struct fr_keyqry_reply_pkt *) blockBuf;

    /* point authPtr to auth section in blockBuf */
    authPtr = (struct fr_keyqry_auth *)(replyPtr + 1);
    
    /* point keyDataPtr to key data section in blockBuf */
    keyDataPtr = authPtr + 1;   /* note to self:  same as   keyDataPtr = buf + (sizeof *authPtr);   */
    
    /* create FrKeyQryQueryExp 'qry' for call to makeCacheEntry */
    qry.keyVersion = (replyPtr->origQuery).keyVersion;
    qry.keyType = (replyPtr->origQuery).keyType;
    if ( ! frEntLoad(err, &keyOwner, &(replyPtr->origQuery).keyOwner[0], FRENT_PORTABLE_SIZE)) {
      goto done;
    }
    if ( (keyOwner.type == FRENT_TYPE_MASTER)) {
      zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENODATA, "keyOwner.type == FRENT_TYPE_MASTER");
      goto done;
    }

    qry.keyOwner = keyOwner;
    
    /* create cache entry 'ce' */
    if ( ! makeCacheEntry(err, kq, &qry, &ce)) {
      data += (sizeof blockNbolSz) + blockSz;
      continue;
    }

	if (bEncrypted)
	{
		ce->state = FRKEYQRY_ST_VALID;
		ce->reply = blockBuf; 
		blockBuf = 0;
		ce->replyLen = blockSz;
		if ( ! key_from_reply(err, ce)) { goto done; }
	}
	else
	{

		/* modify necessary parameters of 'ce' */
		ce->state =  FRKEYQRY_ST_AWAITING_VALIDATION;
		ce->replyLen = blockSz;
		if ( ! frCryptPubKeyBufGetExpiration(err, keyDataPtr,
											 blockSz - (sizeof replyPtr) - (sizeof authPtr),
											 &keyExp)) {
		  goto done;
		}
		ce->expires = zkClockToSeconds(keyExp);
		if ( ! setCacheExpTime(err, ce)) {
		  goto done;
		}
		ce->reply = blockBuf; blockBuf = 0; /* now owned by the cache */

	}

    data += (sizeof blockNbolSz) + blockSz;
  }

  ret = 1;

 done:

  if (blockBuf) { free(blockBuf); }
  if (refData) { memset(refData, 0 , refDataSz); zkMemFree(err, (void **)&refData);  }

  return ret;
}


/*
 *  Write an encrypted copy of the cache to a file.
 *  
 *  The actual data that is stored is a sequence of entries, each of which is
 *  a 16 bit unsigned length, followed by a public key query server response.
 *  The length is in network byte order.  The entire collection of entries is
 *  stored in an encrypted file format.
 *  
 *  We only save those entries that have been validated.  As a special case,
 *  we don't bother saving the master signing key, as we'll ignore it when
 *  reading it back.
 */
DLLExport int frKeyQryCacheWrite(ZkError err, FrKeyQry kq, FrCryptSym sym, FILE *fp)
{
  ZkHashEnum e = 0;
  char *buf = 0;
  const char *fn = "frKeyQryCacheWrite";
  int ret = 0;
  size_t bsz = 0;

  zkErrorClear(err);

  if ( ! kq) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "kq is null");
    return 0;
  }
  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "sym is null");
    return 0;
  }
  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "fp is null");
    return 0;
  }

  /*
   *  Just allocate a single byte, so we can pass a valid pointer to
   *  zkMemAlloc().
   */
  
  if ( ! zkMemAlloc(err,(void **) &buf, 1)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, errno,
               "can't allocate memory for cache entries");
    goto done;
  }

  if ( ! (zkHashEnumNew(err, &e) &&
          zkHashEnumSetHash(err, e, kq->cache))) {
    goto done;
  }

  while (zkHashEnumNext(err, e)) {
    FrKeyQryQueryExp *qry;
    struct cache_entry *ce;

    if ( ! (zkHashEnumGetKey(err, e, (void **)&qry) &&
            zkHashEnumGetData(err, e, (void **)&ce))) {
      continue; /* skip this one */
    }

    if ((ce->state == FRKEYQRY_ST_VALID &&
         ce->keyInfo->keyOwner.type != FRENT_TYPE_MASTER) ||
        ce->state == FRKEYQRY_ST_AWAITING_VALIDATION) {
      size_t obsz;
      uint16_t nbrsz; /* reply length in network byte order */

      obsz = bsz;

      /*
       *  Format the data as a two byte length (in network byte order),
       *  followed by the `certificate' (i.e. database datum).
       */
      
      bsz += 2 + ce->replyLen;
      
      if ( ! zkMemRealloc(err, (void **) &buf, bsz)) {
        zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, errno,
                   "can't reallocate memory for cache entries");
        goto done;
      }

      nbrsz = htons(ce->replyLen);

      memcpy(buf + obsz, &nbrsz, sizeof nbrsz);
      memcpy(buf + obsz + sizeof nbrsz, ce->reply, ce->replyLen);
    }
  }

  if ( ! frSecEncryptToFp(err, fp, sym, buf, bsz)) {
    goto done;
  }

  ret = 1;

 done:

  if (e) { zkHashEnumFree(err, &e); }
  if (buf) { memset(buf, 0, bsz); free(buf); }

  return ret;
}


/*
 *  Print the contents of the key cache.
 */
DLLExport int frKeyQryCachePrint(ZkError err, FrKeyQry kq, FILE *fp)
{
  zkErrorClear(err);

  if ( ! kq) {
    zkErrorSet(err, "frKeyQryCachePrint", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "kq is null");
    return 0;
  }
  if ( ! fp) {
    zkErrorSet(err, "frKeyQryCachePrint", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "fp is null");
    return 0;
  }

  return zkHashTraverse(err, kq->cache, cachePrint, fp);
}


/*
 *  For each entry in the cache that is marked as awaiting validation, try to
 *  validate it.
 *  
 *  bug: this routine was written in support of bug #1433.
 */
DLLExport int frKeyQryValidateCache(ZkError err, FrKeyQry kq, int fd, struct sockaddr *addr)
{
  ZkHashEnum e = 0;
  FrKeyQryHost keysrv;
  const char *fn = "frKeyQryValidateCache";
  int ret = 0;
  int retries = 4, timeout = 1;
  
  zkErrorClear(err);

  if ( ! kq) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "kq is null");
    goto done;
  }
  if (fd < 0) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "fd (%d) is invalid", fd);
    goto done;
  }
  if ( ! addr) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "addr is null");
    goto done;
  }

  keysrv.fd = fd;
  memcpy(&keysrv.host, addr, sizeof keysrv.host);

  if ( ! (zkHashEnumNew(err, &e) &&
          zkHashEnumSetHash(err, e, kq->cache))) {
    goto done;
  }

  while (zkHashEnumNext(err, e)) {
    FrKeyQryQueryExp *qry;
    struct cache_entry *ce;

    if ( ! (zkHashEnumGetKey(err, e, (void **)&qry) &&
            zkHashEnumGetData(err, e, (void **)&ce))) {
      continue; /* skip this one */
    }

    if (ce->state == FRKEYQRY_ST_AWAITING_VALIDATION) {
      FrCryptPubKey key; /* will not be used */
      FrEntIdExp owner;
      FrPubKeyType kt = ce->reply->origQuery.keyType;
      unsigned int kv = ce->reply->origQuery.keyVersion;

      if ( ! (frEntLoad(err, &owner,
                        ce->reply->origQuery.keyOwner,
                        sizeof ce->reply->origQuery.keyOwner) &&
              frKeyQryFetchPubKey(err, kq, kv, kt, owner,
                                  keysrv, retries, timeout, &key))) {
        goto done;
      }
    }
  }

  ret = 1;

 done:

  if (e) { zkHashEnumFree(err, &e); }
  
  return ret;
}
