/*
 *  $Id: srv.c,v 1.5.32.2 2000/10/11 13:18:10 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#define DEBUG_SHOW_QRY 0


/*
 *
 *
 *                    External server specific routines.
 *
 *
 */


#include "zks.h"

#include <stddef.h>
#include <string.h>
#include "fr_keyqry_wire.h"
#include "frKeyQry.h"
#include "zkError.h"
#include "zkLog.h"


/*
 *  Extract the actual query from a key query packet.  The packet is contained
 *  in `pkt' and has size `psz'; `*q' is set to point to the actual query and
 *  `*qsz' is set to its size.
 */
DLLExport int frKeyQryRecvQuery(ZkError err, void *pkt, size_t psz, void **q, size_t *qsz)
{
  const char * const func = "frKeyQryRecvQuery";

  int ret = 0;
  struct fr_keyqry_qry_pkt *qp = pkt;
  
  zkErrorClear(err);

  if ( ! pkt) {
    zkErrorSet(err, func, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "pkt is null");
    return 0;
  }
  if (psz < sizeof *qp) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "psz is too small (%u < %u)", psz, sizeof *qp);
    return 0;
  }
  if ( ! q) {
    zkErrorSet(err, func, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "q is null");
    return 0;
  }
  if ( ! qsz) {
    zkErrorSet(err, func, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "qsz is null");
    return 0;
  }
  
#if DEBUG_SHOW_QRY
  {
    ZkLog log;

    zkErrorLog(err, &log);
    zkLogWrite_va(log, "query: %u, %u, %u, %s",
                  (unsigned int)qp->query.keyVersion,
                  (unsigned int)qp->query.keyType,
                  (unsigned int)qp->query.keyOwner[0],
                  qp->query.keyOwner + 1);
  }
#endif

  *q = &qp->query;
  *qsz = sizeof qp->query;

  /* Sanity check on the query. */

  if (qp->vers != 1 /* bug */) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRKEYQRY, ZKEDATABAD,
                  "invalid query packet version (%u)", (unsigned int)qp->vers);
    goto done;
  }
  if (qp->type != FRPKTTYPE_KEYQRY) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRKEYQRY, ZKEDATABAD,
                  "invalid query packet type (%u)", (unsigned int)qp->type);
    goto done;
  }
  if (qp->query.keyVersion > FR_KEYQRY_MAXKEYVERSION) {
    zkErrorSet(err, func, ZKERR, ZKMFRKEYQRY, ZKEDATABAD, "bad key version");
    goto done;
  }
  if (   (qp->query.keyType != FRCRYPT_KEYTYPE_CRYPTION)
      && (qp->query.keyType != FRCRYPT_KEYTYPE_SIGNING)) {
    zkErrorSet(err, func, ZKERR, ZKMFRKEYQRY, ZKEDATABAD, "bad key type");
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 *  Create a key query response packet.
 *  
 *  `code' indicates, to the client performing the query, whether or not its
 *  query was successful.  If unsuccessful, it provides an extremely limited
 *  indication of what may have gone wrong.
 *  
 *  `k' points to data that is an fr_keyqry_qry structure, containing the
 *  original query.  `ksz' is its size.
 *  
 *  `d' points to data containing the database entry corresponding to the
 *  query given by `k'.  It is an fr_keyqry_auth structure, followed by key
 *  parameters (i.e. bignums, and other information).  `dsz' is its size.
 *  (Thus, the total size of the key parameters is given by rsz -
 *  sizeof(struct fr_keyqry_auth)).  Unless `code' equals
 *  FR_KEYQRY_CODE_FOUND, `d' and `dsz' are not guaranteed to be valid.
 *  
 *  `pkt' points to a buffer in which the ultimate response will be copied.
 *  `*psz' is the size of the buffer; it will be set to the actual amount
 *  required by the response packet.
 */
DLLExport int frKeyQryMakeResp(ZkError err, int code, void *k, size_t ksz, void *d, size_t dsz, void *pkt, size_t *psz)
{
  struct fr_keyqry_qry *dbkey = k;
  struct fr_keyqry_auth *dbdatum = d;
  struct fr_keyqry_reply_pkt *reply = pkt;

  zkErrorClear(err);

  /* Check the `code' parameter later. */
  
  if ( ! k) {
    zkErrorSet(err, "frKeyQryMakeResp", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "k is null");
    return 0;
  }
  if (ksz != sizeof *dbkey) {
    zkErrorSet_va(err, "frKeyQryMakeResp", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "ksz is invalid (%u != %u)", ksz, sizeof *dbkey);
    return 0;
  }
  /*
   *  We check `d' and `dsz' only if `code' is FR_KEYQRY_CODE_FOUND.  See
   *  below.
   */
  if ( ! pkt) {
    zkErrorSet(err, "frKeyQryMakeResp", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "pkt is null");
    return 0;
  }
  if ( ! psz) {
    zkErrorSet(err, "frKeyQryMakeResp", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "psz is null");
    return 0;
  }
  if (*psz < sizeof *reply) {
    zkErrorSet_va(err, "frKeyQryMakeResp", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "*psz is too small (%u < %u)", *psz, sizeof *reply);
    return 0;
  }
  
  /*
   *  Initialize the reply.
   */

  memset(reply->rand,
         FRCRYPT_RAND_SIZE - (sizeof reply->rand % FRCRYPT_RAND_SIZE),
         sizeof reply->rand);
  reply->vers = 1; /* bug */
  reply->type = FRPKTTYPE_KEYRESP;
  memset(reply->spec,
         FRCRYPT_RAND_SIZE - (sizeof reply->spec % FRCRYPT_RAND_SIZE),
         sizeof reply->spec);

  reply->respCode = code;
  memcpy(&reply->origQuery, dbkey, sizeof reply->origQuery);

  switch (code) {
  case FR_KEYQRY_CODE_BADQUERY:
  case FR_KEYQRY_CODE_NOTFOUND:
  case FR_KEYQRY_CODE_SRVERROR:
    *psz = sizeof *reply;
    break;

  default:
    /* bug: log an error */
    reply->respCode = FR_KEYQRY_CODE_SRVERROR;
    *psz = sizeof *reply;
    zkErrorSet_va(err, "frKeyQryMakeResp", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "code (%d) is invalid", code);
    return 0;

  case FR_KEYQRY_CODE_FOUND:
    if ( ! d) {
      zkErrorSet(err, "frKeyQryMakeResp", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                 "d is null");
      return 0;
    }
    if (dsz < sizeof *dbdatum) {
      zkErrorSet_va(err, "frKeyQryMakeResp", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                    "dsz is too small (%u < %u)", dsz, sizeof *dbdatum);
      return 0;
    }
    if (*psz < sizeof *reply + dsz) {
      zkErrorSet_va(err, "frKeyQryMakeResp", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                    "*psz is too small (%u < %u)", *psz, sizeof *reply + dsz);
      reply->respCode = FR_KEYQRY_CODE_SRVERROR;
      return 0;
    }
      
    memcpy((char *)reply + sizeof *reply, dbdatum, dsz);
    *psz = sizeof *reply + dsz;
  }

  return 1;
}
