/*
 *  $Id: updcom.c,v 1.10.26.1 2000/10/11 13:18:10 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Common key update routines.
 */

#include "zks.h"

#include <sys/types.h>
#include <stddef.h>
#include "upd.h"
#include "frKeyQry.h"
#include "zkError.h"
#include "zkFile.h"


static const size_t dbDatumSizeMin = 4 * 1024; /* bug: better value */
static const size_t dbQuerySizeMin = 256; /* bug: better value */


/*
 *  Return the minimum size required to store a key database datum.
 */
DLLExport size_t frKeyUpdMinDbDatumSize(void)
{
  return dbDatumSizeMin;
}


/*
 *  Return the minimum size required to store a key database query.
 */
DLLExport size_t frKeyUpdMinDbQuerySize(void)
{
  return dbQuerySizeMin;
}


/*
 *  From an update request, `upd', set `*kbuf' to point to the start of the
 *  key parameters, and set *ksz to the size of the remainder of the buffer
 *  (not necessarily the size of the key data, itself).
 *  
 *  Note: error checking is not performed on the key parameter data.  It may
 *  be too much, or too little data, or it may be complete garbage.  You won't
 *  know until you try to construct a key from it.
 */
DLLExport int frKeyUpdGetKeyParms(ZkError err, void *upd, size_t usz, void **kbuf, size_t *ksz)
{
  const char *fn = "frKeyUpdGetKeyParms";
  struct fr_keyupd_update_pk *ku = upd;

  zkErrorClear(err);

  if ( ! upd) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "upd is null");
    return 0;
  }
  if (usz < sizeof *ku) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "usz (%lu) is too small", (unsigned long)usz);
    return 0;
  }
  if ( ! kbuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "kbuf is null");
    return 0;
  }
  if ( ! ksz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "ksz is null");
    return 0;
  }

  *kbuf = ku + 1;
  *ksz = usz - sizeof *ku;

  return 1;
}


/*
 *  From an update request, `upd', extract into `data' that part containing
 *  the key parameters.
 *  
 *  Note: error checking is not performed on the key parameter data.  It may
 *  be too much, or too little data, or it may be complete garbage.  You won't
 *  know until you try to construct a key from it.
 */
DLLExport int frKeyUpdCopyKeyParms(ZkError err, void *upd, size_t usz, void *data, size_t *dsz)
{
  size_t n;
  struct fr_keyupd_update_pk *ku = upd;

  zkErrorClear(err);

  if ( ! upd) {
    zkErrorSet(err, "frKeyUpdCopyKeyParms", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "upd is null");
    return 0;
  }
  if (usz < sizeof *ku) {
    zkErrorSet_va(err, "frKeyUpdCopyKeyParms", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "usz (%u) is too small", usz);
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "frKeyUpdCopyKeyParms", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, "frKeyUpdCopyKeyParms", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "dsz is null");
    return 0;
  }

  n = usz - sizeof *ku;

  if (*dsz < n) {
    zkErrorSet_va(err, "frKeyUpdCopyKeyParms", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "*dsz (%u) is smaller than required (%u)", *dsz, n);
    return 0;
  }

  memcpy(data, (char *)upd + sizeof *ku, n);
  *dsz = n;

  return 1;
}


DLLExport int frKeyQryLogQueryPkt(ZkError err, ZkLog log, void *pkt, size_t psz)
{
  FrEntIdExp keyOwner;
  const char *entTypeStr;
  struct fr_keyqry_qry_pkt *q = pkt;

  if ( ! log) {
    zkErrorSet(err, "frKeyQryLogQueryPkt", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "log is null");
    return 0;
  }
  if ( ! pkt) {
    zkErrorSet(err, "frKeyQryLogQueryPkt", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "pkt is null");
    return 0;
  }
  if (psz < sizeof *q) {
    zkErrorSet_va(err, "frKeyQryLogQueryPkt", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "psz is too small (%lu < %lu)",
                  (unsigned long)psz, (unsigned long)sizeof *q);
    return 0;
  }
  
  if ( ! (frEntLoad(err, &keyOwner, q->query.keyOwner,
                    sizeof q->query.keyOwner) &&
          frEntGetTypeStr(err, &keyOwner, &entTypeStr))) {
    return 0;
  }
  
  zkLogWrite_va(log, "key vers %u, key type %c (%u), key owner: %s (%u), %64s",
                (unsigned int)q->query.keyVersion,
                fr_CryptKeyTypeChar(q->query.keyType),
                (unsigned int)q->query.keyType,
                entTypeStr, keyOwner.type, keyOwner.name);
  return 1;
}


DLLExport int frKeyQryLogRespPkt(ZkError err, ZkLog log, void *pkt, size_t psz)
{
  FrEntIdExp keyOwner;
  const char *codeStr, *entTypeStr;
  struct fr_keyqry_reply_pkt *r = pkt;

  if ( ! log) {
    zkErrorSet(err, "frKeyQryLogRespPkt", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "log is null");
    return 0;
  }
  if ( ! pkt) {
    zkErrorSet(err, "frKeyQryLogRespPkt", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "pkt is null");
    return 0;
  }
  if (psz < sizeof *r) {
    zkErrorSet_va(err, "frKeyQryLogRespPkt", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "psz is too small (%lu < %lu)",
                  (unsigned long)psz, (unsigned long)sizeof *r);
    return 0;
  }

  switch (r->respCode) {
  case FR_KEYQRY_CODE_BADQUERY: codeStr = "bad query";    break;
  case FR_KEYQRY_CODE_FOUND:    codeStr = "found";        break;
  case FR_KEYQRY_CODE_NOTFOUND: codeStr = "not found";    break;
  case FR_KEYQRY_CODE_SRVERROR: codeStr = "server error"; break;
  default:                      codeStr = "???";          break;
  }

  zkLogWrite_va(log, "response is: %s", codeStr);
  
  switch (r->respCode) {
  case FR_KEYQRY_CODE_FOUND:
  case FR_KEYQRY_CODE_NOTFOUND:
    if ( ! (frEntLoad(err, &keyOwner, r->origQuery.keyOwner,
                      sizeof r->origQuery.keyOwner) &&
            frEntGetTypeStr(err, &keyOwner, &entTypeStr))) {
      return 0;
    }
  
    zkLogWrite_va(log, "key vers %u, key type %c (%u), key owner: %s (%u), %s",
                  (unsigned int)r->origQuery.keyVersion,
                  fr_CryptKeyTypeChar(r->origQuery.keyType),
                  (unsigned int)r->origQuery.keyType,
                  entTypeStr, keyOwner.type, keyOwner.name);
    break;
    
  default:
    break; 
  }

  return 1;
}


/*
 *  Print a short description of the public key update request.
 */
DLLExport int frKeyUpdShortPrint(ZkError err, FILE *fp, int indent, void *data, size_t dsz)
{
  FrCryptPubKey pk = 0;
  const char *fn = "frKeyUpdShortPrint";
  int flags, ret = 0;
  size_t asz, ksz;
  struct fr_keyupd_update_pk *upd = data;
  struct fr_keyqry_auth *auth;
  void *kbuf;

  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "fp is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "data is null");
    return 0;
  }
  if (dsz < sizeof *upd) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "dsz (%u) is too small", dsz);
    return 0;
  }

  auth = (void *)((char *)upd + offsetof(struct fr_keyupd_update_pk, auth));
  asz = dsz - offsetof(struct fr_keyupd_update_pk, auth);

  kbuf = upd + 1;
  ksz = dsz - sizeof *upd;

  flags = ntohs(upd->flags);

  fprintf(fp, "%.*sKey update request:\n\n", indent, "");

  fprintf(fp, "%.*s  Flags: [", indent, "");
  if (flags & FRKU_FLAG_CURRENT_KEY) {
    fprintf(fp, "curr");
  }
  fprintf(fp, "]\n");

  if ( ! (zkFprintf(err, fp, "%.*s  Key owner: ", indent, "") &&
          frEntBufPrint(err, fp, auth->keyOwnerEntity,
                        sizeof auth->keyOwnerEntity) &&
          zkFprintf(err, fp, "\n"))) {
    goto done;
  }
  
  if (auth->keySignerEntity[0]) { /* bug! HACK */
    if ( ! (zkFprintf(err, fp, "%.*s  Key signer: ", indent, "") &&
            frEntBufPrint(err, fp, auth->keySignerEntity,
                          sizeof auth->keySignerEntity) &&
            zkFprintf(err, fp, "\n\n"))) {
      goto done;
    }
  } else {
    if ( ! (zkFprintf(err, fp, "%.*s  Key signer: %s", indent, "", "<unset>") &&
            zkFprintf(err, fp, "\n\n"))) {
      goto done;
    }
  }

  if ( ! (frCryptPubKeyNew(err, &pk) &&
          frCryptPubKeyReadFromBuf(err, pk, kbuf, &ksz) &&
          frCryptPubKeyShortPrint(err, pk, indent + 2, fp))) {
    goto done;
  }

  ret = 1;

 done:

  if (pk) { frCryptPubKeyFree(0, &pk); }

  return ret;
}


#if 0
DLLExport int frKeyUpdDbDatumPrint(ZkError err, FILE *fp, void *datum, size_t dsz)
{
  struct fr_kq_keyinfo_pk *ki = datum;

  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, "frKeyUpdDbDatumPrint", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "fp is null");
    return 0;
  }
  if ( ! datum) {
    zkErrorSet(err, "frKeyUpdDbDatumPrint", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "datum is null");
    return 0;
  }
  if (dsz < sizeof *ki) {
    zkErrorSet_va(err, "frKeyUpdDbDatumPrint", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "datum size is incorrect (%u < %u)", dsz, sizeof *ki);
    return 0;
  }

  fprintf(fp, "Owner type: %u\n", (unsigned int)ki->owner[0]);
  fprintf(fp, "Owner name: %s\n", &ki->owner[1]); /* bug: assumes string */
  fprintf(fp, "Signer type: %u\n", (unsigned int)ki->signer[0]);
  fprintf(fp, "Signer name: %s\n", &ki->signer[1]); /* bug: assumes string */
  fprintf(fp, "Signer signature version: %u\n", (unsigned int)ki->signerSig[0]); /* bug */
  fprintf(fp, "Signer signature: \n\n");

  {
    size_t i;
    
    fprintf(fp, "    "); /* indent */
    for (i = 1; i < sizeof ki->signerSig; i++) {
      fprintf(fp, "%02x", (unsigned int)(unsigned char)ki->signerSig[i]);
      if (i % 20 == 0) { fprintf(fp, "\n    "); }
    }
    fputs(i % 20 == 0 ? "\n" : "\n\n", fp);
  }

  return 1;
}


DLLExport int frKeyUpdDbKeyPrint(ZkError err, FILE *fp, void *key, size_t ksz)
{
  struct fr_kq_dbkey_pk *k = key;
  
  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, "frKeyUpdDbKeyPrint", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "fp is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, "frKeyUpdDbKeyPrint", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if (ksz != sizeof *k) {
    zkErrorSet_va(err, "frKeyUpdDbKeyPrint", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "key size is incorrect (%u != %u)", ksz, sizeof *k);
    return 0;
  }

  fprintf(fp, "Version: %u\n", k->vers);
  fprintf(fp, "Type: %u\n", k->type);
  fprintf(fp, "Owner type: %u\n", (unsigned int)k->owner[0]);
  fprintf(fp, "Owner name: %s\n", &k->owner[1]); /* bug: assumes string */

  return 1;
}
#endif

/*
 *
 */

DLLExport int
frKeyUpdGetKey (
  ZK_INOUT ZkError err ,
  ZK_IN    void * b ,
  ZK_IN    size_t bsz ,
  ZK_OUT   FrCryptPubKey * pk )
{
  static const char * func = "frKeyUpdGetKey";
  struct fr_keyupd_update_pk * upd = b;
  void * key = NULL;
  size_t key_sz = 0;
  int ret = 0;

  if (!b) {
    zkErrorSet(err,func,ZKERR,ZKMFRKEYQRY,ZKENULLPTR,
               "update buffer pointer is NULL");
  } else if (!pk) {
    zkErrorSet(err,func,ZKERR,ZKMFRKEYQRY,ZKENULLPTR,
               "PubKey pointer is NULL");
  } else {
    key = upd + 1;
    key_sz = bsz - sizeof(*upd);
    if (!(frCryptPubKeyNew(err,pk) &&
          frCryptPubKeyReadFromBuf(err,*pk,key,&key_sz))) {
      ret = 0;
    } else {
      ret = 1;
    }
  }
  return ret;
} 
     
