/*
 *  $Id: vrfy.c,v 1.3.32.1 2000/10/11 13:18:10 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Routines to support verification (i.e. sanity checks) on public key
 *  database datums.
 */


#include <errno.h>
#include <stdlib.h>
#include "fr_datumvrfy.h"
#include "fr_keyqry_wire.h"
#include "frCrypt.h"
#include "frEnt.h"
#include "zkCStr.h"
#include "zkError.h"
#include "zkLog.h"


#ifdef ZKMOD
#undef ZKMOD
#endif
#define ZKMOD ZKMFRKEYQRY


/*
 *  Check the validity (as much as possible) of the public key query/datum
 *  pair given by `q' and `d'.
 */
DLLExport int frKeyUpdPublicValidate(ZkError err, void *q, size_t qsz, void *d, size_t dsz)
{
  FrCryptPubKey pk = 0;
  FrEntIdExp kown;
  FrPubKeyType kt;
  ZkLog log;
  char *dstr = 0;
  const char *fn = "frKeyUpdValidate";
  int ret = 0;
  size_t ksz; /* size of key buffer */
  struct fr_keyqry_qry *qry = q;
  struct fr_keyqry_auth *datum = d;
  unsigned int kv;
  void *kdat;

  zkErrorClear(err);

  if ( ! q) {
    zkErrorSet(err, fn, ZKERR, ZKMOD, ZKENULLPTR, "q is null");
    return 0;
  }
  if (qsz < sizeof *qry) {
    zkErrorSet_va(err, fn, ZKERR, ZKMOD, ZKEPARMBAD,
                  "qsz is too small (%lu < %lu)",
                  (unsigned long)qsz, (unsigned long)sizeof *qry);
    return 0;
  }
  if ( ! d) {
    zkErrorSet(err, fn, ZKERR, ZKMOD, ZKENULLPTR, "d is null");
    return 0;
  }
  if (dsz < sizeof *datum) {
    zkErrorSet_va(err, fn, ZKERR, ZKMOD, ZKEPARMBAD,
                  "dsz is too small (%lu < %lu)",
                  (unsigned long)dsz, (unsigned long)sizeof *datum);
    return 0;
  }

  if ( ! zkErrorLog(err, &log)) {
    goto done;
  }

  /*
   *  Get the key data.
   */

  kdat = datum + 1;
  ksz = dsz - sizeof *datum;

  if ( ! (frCryptPubKeyNew(err, &pk) &&
          frCryptPubKeyReadFromBuf(err, pk, kdat, &ksz) &&
          frCryptPubKeyGetVersion(err, pk, &kv) &&
          frCryptPubKeyGetType(err, pk, &kt) &&
          frCryptPubKeyGetEntity(err, pk, &kown))) {
    goto done;
  }

  /*
   *  Create a string to identify this datum.
   */

  {
    char estr[256];
    
    if ( ! frEntSnprint(err, estr, sizeof estr, &kown)) {
      goto done;
    }
    if (zkCStrSprintf(&dstr, "%u, %u, %s", kv, kt, estr) == -1) {
      zkErrorSet_va(err, fn, ZKERR, ZKMOD, errno,
                    "can't allocate entity string: %s", strerror(errno));
      goto done;
    }
  }

  if (dsz != sizeof *datum + ksz) {
    zkErrorSet_va(err, fn, ZKERR, ZKMOD, ZKEDATABAD,
                  "datum `%s' contains %lu bytes of trailing garbage",
                  dstr, (unsigned long)(dsz - (sizeof *datum + ksz)));
    goto done;
  }

  /*
   *  Check the query against the datum.
   *  
   *  If the query is not for the `current' key, then its version should match
   *  that of the key.
   */

  if (qry->keyVersion != 0 && qry->keyVersion != kv) {
    zkErrorSet_va(err, fn, ZKERR, ZKMOD, ZKEDATABAD,
                  "query version doesn't match that in key (%u != %u) [%s]",
                  qry->keyVersion, kv, dstr);
    goto done;
  }

  if (qry->keyType != kt) {
    zkErrorSet_va(err, fn, ZKERR, ZKMOD, ZKEDATABAD,
                  "query type doesn't match that in key (%u != %u) [%s]",
                  qry->keyType, kt, dstr);
    goto done;
  }

  /* The various entities should match. */

  {
    FrEntIdExp down, qown; /* key owners from datum and query */
    int equal;
    
    if ( ! frEntLoad(err, &down, datum->keyOwnerEntity,
                     sizeof datum->keyOwnerEntity)) {
      /* bug: include datum entity string */
      zkLogWrite_va(log, "%s: bad entity in datum: entity from key data is `%s'",
                    fn, dstr);
      goto done;
    }
    if ( ! frEntLoad(err, &qown, qry->keyOwner, sizeof qry->keyOwner)) {
      zkLogWrite_va(log, "%s: bad entity in query: entity from key data is `%s'",
                    fn, dstr);
      goto done;
    }
         
    if ( ! (frEntEqual(err, &down, &qown, &equal) && equal)) {
      /* bug: include entity strings */
      zkLogWrite_va(log, "%s: query entity doesn't equal datum entity", fn);
      goto done;
    }

    if ( ! (frEntEqual(err, &down, &kown, &equal) && equal)) {
      /* bug: include entity strings */
      zkLogWrite_va(log, "%s: query entity doesn't equal key owner in `%s'",
                    fn, dstr);
      goto done;
    }
  }

  /* The signer should be correct. */

  {
    FrEntIdExp signer;
    
    if ( ! frEntLoad(err, &signer, datum->keySignerEntity,
                     sizeof datum->keySignerEntity)) {
      /* bug: include signer entity string */
      zkLogWrite_va(log, "%s: bad signer for `%s'.", fn, dstr);
      goto done;
    }

    if ( ! frEntIsKeySigningParent(err, &kown,
                                   kt == FRCRYPT_KEYTYPE_CRYPTION, &signer)) {
      zkLogWrite_va(log, "%s: bad signer for `%s'.", fn, dstr);
      goto done;
    }
  }

  /*
   *  Check the validity of the key.
   */
  
  if ( ! frCryptPubKeyPublicValidate(err, pk)) {
    zkLogWrite_va(log, "%s: key (%s) is invalid.", fn, dstr);
    goto done;
  }
  
  ret = 1;

 done:

  if (pk) { frCryptPubKeyFree(0, &pk); }
  if (dstr) { free(dstr); }
  
  return ret;
}
