/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef _FRLOCALAIP_H
#define _FRLOCALAIP_H

#ifndef ZKS_H_
#include <zks.h>
#endif

#ifndef _STDIO_H
#include <stdio.h>
#endif

#ifndef _ZK_ERROR_H
#include <zkError.h>
#endif

#ifndef _ZKLOG_H
#include <zkLog.h>
#endif

#ifndef FRAIPS_H_
#include <frAips.h>
#endif

#ifndef FRPKTFMT_H_
#include <frPktFmt.h>
#endif

#ifndef FRCRYPT_H_
#include <frCrypt.h>
#endif

#ifndef _ZKINTTYPES_H
#include <zkIntTypes.h>
#endif

#ifndef FRKEYQRY_H_
#include <frKeyQry.h>
#endif

#ifndef FRAIPS_H_
#include <frAips.h>
#endif

#ifndef FRNETINFO_H_
#include "frNetInfo.h"
#endif FRNETINFO_H_

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

  typedef struct T_FrLocalAipCtx *FrLocalAipCtx;

  /* DATA STRUCTURE ROUTINES */
  /**
   * Initialize a new local AIP context.  Expects an ZkError object,
   * addr to the local AIP context, maximum number of routes allowed,
   * maximum number of routers in a route, IP port we are listening on.
   */ 
  DLLExport extern 
  int frLocalAipCtxNew(ZkError, 
                       FrLocalAipCtx *, /* Context */
                       unsigned long data_packet_version, /* Current data packet version */    
                       unsigned long route_packet_version,/* Current route packet version */    
                       size_t,          /* max routes */
                       size_t,          /* max hops/route */
                       int);            /* local port (for computing ACI) */

  /**
   * Associate a key cache to the FrLocalAipCtx structure.  This key cache
   * will be used to look up public encryption and signature keys for AIPs.
   * No changes will be made to the cache by the FrLocalAipCtx.
   */
  DLLExport extern
  int frLocalAipCtxSetKeyCache(ZkError,       /* ZkError */
                               FrLocalAipCtx, /* FrLocalAipCtx object */
                               const FrKeyQry); /* Key cache */

  /**
   * Associate a FrNetInfo object to the FrLocalAipCtx object so it can look
   * up network information as needed.  The object is used as read-only memory,
   * never making any changes to it.
   */
  DLLExport extern
  int frLocalAipCtxSetNetInfo(ZkError err,
                              FrLocalAipCtx aip,
                              const FrNetInfo ni);

  /** Free the structure. */
  DLLExport extern int frLocalAipCtxFree(ZkError, FrLocalAipCtx *);


  /* ROUTE CONFIGURATION AND SETUP */
  /**
   * Get a new route, given the nym name.  Returns the sockaddr of the first
   * hop and the "ACI" to reference this new route
   *
   * @param err
   * @param aip
   * @param duration            Duration of the route
   * @param hops                Number of hops in the selected route
   * @param route               Array of aip IDs (hops elements)
   * @param firstHop            IP addr/port for the FH
   * @param aci                 ACI for this route
   */ 
  DLLExport extern 
  int frLocalAipStartNewRoute(ZkError err,
                              FrLocalAipCtx aip,
                              uint32_t duration,
                              size_t hops, 
                              uint32_t *route, 
                              struct sockaddr_in *firstHop,
                              ZkACI *aci);

  DLLExport extern
  int frLocalAipGetAipNames(ZkError, FrLocalAipCtx,
                            ZkACI, char [][FRENT_NAME_SIZE], 
                            size_t *);

  DLLExport extern
  int frLocalAipGetAipIds(ZkError err, 
                          FrLocalAipCtx aip,
                          ZkACI aci, 
                          uint32_t *aipIds,
                          size_t *numAips);

  DLLExport extern
  int frLocalAipGetOutgoingAci(ZkError err, FrLocalAipCtx aip,
                               ZkACI aci, ZkACI *outgoingaci) ;

   /**
   * Return the exit funnel address in struct sockaddr_in and dotted quad
   * formats
   */
  DLLExport extern
  int frLocalAipGetExit(ZkError err,
                        FrLocalAipCtx aip,
                        ZkACI aci,
                        struct sockaddr_in *addr,
                        char **dottedquad);

  /** 
   * Reuse the exit funnel corresponding to the ACI.  Works the same way as
   * frLocalAipStartNewRoute() except the exit funnel must correspond with a
   * previously used route.
   *
   * @param err
   * @param aip
   * @param duration            Duration of the route
   * @param hops                Number of hops in the selected route
   * @param route               Array of aip IDs (hops elements)
   * @param firstHop            IP addr/port for the FH
   * @param aci                 ACI for this route
   * @param resueAci            ACI for the same exit hop
   */ 
  DLLExport extern 
  int frLocalAipNewRouteReuseExit(ZkError err, 
                                  FrLocalAipCtx aip,
                                  uint32_t duration,
                                  size_t hops,
                                  uint32_t *route,
                                  struct sockaddr_in *firstHop,
                                  ZkACI *aci,
                                  ZkACI reuseAci);
  /** 
   * Set the Nym to route ACI.  Last parameter is the data directory where I
   * can find the private signing key pair.
   */ 
  DLLExport extern 
  int frLocalAipSetRouteNym(ZkError,
                            FrLocalAipCtx, 
                            ZkACI,          /* Route ACI */
                            const char *,   /* Nym name */
                            FrCryptSig);    /* Private sig key for nym */

  /*
	Primarily Used for debugging.
	Returns the public signing key and nym name associated with an authenticated route.
	Returns 1 on success..
	Returns 0 on error.
  */
  DLLExport int frLocalAipGetRouteNym(	ZkError in_err, 
										FrLocalAipCtx in_aip, 
										ZkACI in_aci,
										char *out_nymName, 
										FrCryptSig* out_sigKey);

  /*
    Primarily Used for debugging.
	Returns all the symmetric keys for all the aips in a route.
	Returns 1 on success.
	Returns 0 on error.
  */
  DLLExport int frLocalAipGetRouteSymKeys(	ZkError in_err, 
											FrLocalAipCtx in_aip, 
											ZkACI in_aci, 
											size_t* out_nAipsInRoute, 
											FrCryptSym* out_symkeys);


  /*
	Primarily Used for debugging.
	Returns the first hop's public signing key.
	Returns 1 on success.
	Returns 0 on error.
  */
  DLLExport int frLocalAipGetFirstHopPubSignKey(	ZkError in_err, 
													FrLocalAipCtx in_aip, 
													ZkACI in_aci, 
													FrCryptSig* out_sigKey);


  /**
   * Delete the ACI route entry from the routing table.
   */
  DLLExport extern 
  int frLocalAipDeleteRoute(ZkError,
                            FrLocalAipCtx, 
                            ZkACI);        /* Route ACI to delete */

  
  /* PACKET FUNCTIONS */

  /**
   * Link encrypt the buffer.  Using the route ACI, get the symmetric key
   * from the routing table for link encryption.
   */
  DLLExport extern 
  int frLocalAipLinkEncrypt(ZkError,
                            FrLocalAipCtx, 
                            ZkACI,          /* Route ACI */
                            FrPktType,      /* Type of packet */
                            char *,         /* Buffer */
                            size_t);        /* Size of packet */

  /**
   * Link decrypt the buffer.  Using the route ACI, get the symmetric key
   * from the routing table for link decryption.
   */
  DLLExport extern 
  int frLocalAipLinkDecrypt(ZkError, 
                            FrLocalAipCtx,
                            char *,             /* Buffer */
                            int,                /* Data size in buffer */
                            struct sockaddr_in, /* Where the packet was rcvd */
                            FrPktType *,        /* Packet type (returned) */
                            unsigned int *,     /* Packet version (returned) */
                            ZkACI *);           /* Route ACI (returned) */

  /**
   * Make a ROUTE CREATE packet.
   */
  DLLExport extern 
  int frLocalAipMakeCreatePkt(ZkError, 
                              FrLocalAipCtx,
                              char *,        /* Buffer to fill */
                              size_t *,      /* Buffer size, packet size (r) */
                              ZkACI,         /* Route ACI */
                              int,           /* Reuse bit */
                              FrCryptSymAlg);/* Crypt algorithm */

  /**
   * Parse a ROUTE CREATE ACK packet.
   */
  DLLExport extern 
  int frLocalAipReadCreateAckPkt(ZkError,
                                 FrLocalAipCtx,
                                 ZkACI,       /* Route ACI */
                                 char *,      /* Buffer */
                                 size_t,      /* Packet size */
                                 int);        /* Reuse bit */

  /**
   * Parse DATA packet.  Remove multiple layers of encryption.
   */
  DLLExport extern 
  int frLocalAipReadDataPkt(ZkError, 
                            FrLocalAipCtx, 
                            ZkACI,         /* Route ACI */
                            char *,        /* Incoming packet */
                            size_t *,      /* Size of packet */
                            char *,        /* IP packet buffer */
                            size_t *);     /* Buffer size, size of data (r) */
  /**
   * Make A DATA packet.  Put on the multiple layers of encryption.
   */
  DLLExport extern
  int frLocalAipMakeDataPkt(ZkError,
                            FrLocalAipCtx,
                            ZkACI,         /* Route ACI */
                            char *,        /* IP packet buffer */
                            size_t,        /* Packet size */
                            char *,        /* Packet buffer (returned) */
                            size_t *);     /* Buffer size, Data size (ret) */

  DLLExport extern int frLocalAipReadNdbUpdatePkt(ZkError, FrLocalAipCtx ,
                                                  ZkACI, char *, size_t,
                                                  char *, size_t *);

  DLLExport extern int frLocalAipMakeNdbUpdatePkt(ZkError, FrLocalAipCtx,
                                                  ZkACI, char *, size_t,
                                                  char *, size_t *);
  /**
   * Make a DESTROY packet.
   */
  DLLExport extern
  int frLocalAipMakeDestroyPkt(ZkError,
                               FrLocalAipCtx,
                               char *,     /* Buffer */
                               size_t *,   /* Buffer size, packet size (r) */
                               ZkACI);     /* Route ACI */

  DLLExport 
  int frLocalAipMakeKeepAlivePkt(ZkError,
                                 FrLocalAipCtx,
                                 char *,
                                 size_t *);


  /**
   * Make a garbage packet.  This packet is randomly filled with data except
   * the last byte which has the symmetric key version.
   */
  DLLExport extern 
  int frLocalAipMakeGarbagePkt(ZkError,
                               FrLocalAipCtx,
                               ZkACI,       /* Route ACI */
                               char *,      /* Buffer */
                               size_t *);   /* Buffer size, packet size (r) */

  DLLExport extern int frLAReadFirstHopPubSigKey(ZkError, FrLocalAipCtx,
                                                 ZkACI);

  DLLExport extern int frLocalAipSetRoutePriority(ZkError, FrLocalAipCtx,
                                                  ZkACI, unsigned int);

  DLLExport int frLocalAipGetSymKey(ZkError, FrLocalAipCtx, ZkACI,
                                    uint8_t, FrCryptSym *);




  /**
   * Store the current UTC time of the Freedom network.
   */
  DLLExport extern int frLocalAipCtxSetFrTime(ZkError err, 
                                              FrLocalAipCtx aip, 
                                              const time_t freedomTime);

  DLLExport extern int frLocalAipGetDataProtocolVersion(ZkError err, 
                                                        const FrLocalAipCtx aip, 
                                                        uint32_t * version);

  DLLExport extern int frLocalAipGetRteProtocolVersion(ZkError err, 
                                                       const FrLocalAipCtx aip, 
                                                       uint32_t * version);

  DLLExport extern int frLocalAipGetMaxDataPacketSize(ZkError err,
                                                      const FrLocalAipCtx aip,
                                                      size_t * max_size);

  DLLExport extern int frLocalAipGetMaxRtePacketSize(ZkError err, 
                                                     const FrLocalAipCtx aip, 
                                                     size_t * max_size);

  DLLExport void frLocalAipSetMss(char* buf, uint16_t mss,
                                  uint16_t pktlen);


  /*
   *  The following functions were written to allow an AIP data packet object
   *  to be used in place of the local AIP object's internal one.
   *  
   *  The local AIP object seems to correspond to one anonymous route (since
   *  it contains an ACI), but then doesn't allow for multiple TCP streams
   *  (for example) each handled by a separate thread.
   *  
   *  A possible solution was to add locking around uses of the local AIP
   *  object, but this could cause an unnecessary performance loss, since it
   *  would force all TCP and UDP traffic through the single data packet
   *  object.
   *  
   *  See bug #8251.
   */

  DLLExport int frLocalAipLinkDecryptUsingAipData(ZkError err, FrLocalAipCtx aip,
                                                  FrPktAD aipData,
                                                  char *udpBuf, int udpSz, 
                                                  struct sockaddr_in host,
                                                  FrPktType *type, unsigned int *vers,
                                                  ZkACI *aci);
  
  DLLExport int frLocalAipMakeDataPktUsingAipData(ZkError err, FrLocalAipCtx aip,
                                                  FrPktAD aipData,
                                                  ZkACI aci,
                                                  char *ipBuf, size_t ipSize,
                                                  char *udpBuf, size_t *udpSize);
  
  DLLExport int frLocalAipReadDataPktUsingAipData(ZkError err, FrLocalAipCtx aip,
                                                  FrPktAD aipData, ZkACI aci,
                                                  char *udpBuf, size_t *udpSize, 
                                                  char *ipBuf,  size_t *ipSize);
  
  DLLExport int frLocalAipMakeDestroyPktUsingAipData(ZkError err, FrLocalAipCtx aip,
                                                     FrPktAD aipData,
                                                     char *udpBuf, size_t *udpSz,
                                                     ZkACI aci);

  DLLExport int frLocalAipReadCreateAckPktUsingAipData(ZkError err, FrLocalAipCtx aip,
                                                       FrPktAD aipData,
                                                       ZkACI aci, char *udpBuf,
                                                       size_t udpSz, int reuse);

  DLLExport int frLocalAipMakeGarbagePktUsingAipData(ZkError err, FrLocalAipCtx aip,
                                                     FrPktAD aipData,
                                                     ZkACI aci, char *udpBuf, 
                                                     size_t *udpSize);

  DLLExport int frLocalAipMakeNdbUpdatePktUsingAipData(ZkError err, FrLocalAipCtx aip,
                                                       FrPktAD aipData,
                                                       ZkACI aci, char *payload, size_t psz,
                                                       char *udpBuf, size_t *udpSize);

  DLLExport int frLocalAipReadNdbUpdatePktUsingAipData(ZkError err, FrLocalAipCtx aip,
                                                       FrPktAD aipData,
                                                       ZkACI aci, char *udpBuf,
                                                       size_t udpSize, char *payBuf,
                                                       size_t *paySz);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* _FRLOCALAIP_H */
