/* $Id: frLocalAipPkts.c,v 1.40.20.10 2001/01/29 23:01:13 oliver Exp $ */
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#ifdef ZK_LINUX
#define _GNU_SOURCE
#endif
#include <stdio.h>

#include <zkPacket.h>

#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>
#else
#include <winsock2.h>
#include <time.h>
#endif

#include <string.h>
#include "localAipStructs.h"
#include <frLocalAip.h>
#include "localAipRouterTable.h"
#include <frNetworkDefaults.h> /* for MAX_MSS */




/**
 * Link decrypt the packet in the udpBuf.  Check the link cryption key for 
 * host because the data may have come from any one of our routes.  Return 
 * the type, version, and aci so that we know which route was used to 
 * decrypt this. 
 *  
 * @param err
 * @param aip
 * @param udpBuf                Buffer to link decrypt 
 * @param udpSz                 Size of the data in udpBuf 
 * @param host                  Sending host IP/port 
 * @param type                  Packet type (returned)
 * @param vers                  Packet version (returned)
 * @param aci                   Route table Mapping (returned)
 */ 
DLLExport int frLocalAipLinkDecrypt(ZkError err, FrLocalAipCtx aip,
                                    char *udpBuf, int udpSz, 
                                    struct sockaddr_in host,
                                    FrPktType *type, unsigned int *vers,
                                    ZkACI *aci)
{
  return frLocalAipLinkDecryptUsingAipData(err, aip, aip->data.packet.handler,
                                           udpBuf, udpSz, host, type, vers, aci);
}

/**
 * Link decrypt the packet in the udpBuf.  Check the link cryption key for 
 * host because the data may have come from any one of our routes.  Return 
 * the type, version, and aci so that we know which route was used to 
 * decrypt this. 
 *  
 * @param err
 * @param aip
 * @param aipData               data packet object to use
 * @param udpBuf                Buffer to link decrypt 
 * @param udpSz                 Size of the data in udpBuf 
 * @param host                  Sending host IP/port 
 * @param type                  Packet type (returned)
 * @param vers                  Packet version (returned)
 * @param aci                   Route table Mapping (returned)
 */ 
DLLExport int frLocalAipLinkDecryptUsingAipData(ZkError err, FrLocalAipCtx aip,
                                                FrPktAD aipData,
                                                char *udpBuf, int udpSz, 
                                                struct sockaddr_in host,
                                                FrPktType *type, unsigned int *vers,
                                                ZkACI *aci)
{
  size_t i, j;
  int decrypt = 0;
  ZkACI aciIn;
  uint8_t cur;
  const char *func = "frLocalAipLinkDecryptUsingAipData";
  char * frPkt;
  size_t frPktSz;

  if (udpSz < 0) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
               "Size of pkt is negative");
    return 0;                   
  }

  if( !frPktGetLekv(err, udpBuf, udpSz, &cur) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktSetLekv() failed");
    return 0;
  }
  if ((cur != 0) && (cur != 1)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD,
               "Current sym key number is not valid");
    return 0;
  }

  if (aip) {
    /* Check all valid routes */
    for (i = 0; i < aip->route.context->maxRoutes; i++) {
      if (aip->route.context->route[i].valid) {
        if ((aip->route.context->route[i].router[0].host.sin_addr.s_addr == 
             host.sin_addr.s_addr) && 
            (aip->route.context->route[i].router[0].host.sin_port == host.sin_port)) {

          if (!frPktLinkDecrypt(err, aip->route.context->route[i].symKey[cur], udpBuf,
                               udpSz)) {
            /* Link decryption failed */
            zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD,
                       "Link decryption failed");
            return 0;
          }
          decrypt = 1;          /* Link Decrypted */
          break;
        }
      }
    }
    if (decrypt) {
      /* Decrypted, check out the ACIs */
      if (!frPktGetVersion(err, udpBuf, udpSz, vers)) {
        zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                       "frPktGetVersion() failed");
        return 0;
      }
          
      if (!frPktGetType(err, udpBuf, udpSz, type)) {
        zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                       "frPktGetType() failed");
        return 0;
      }

      switch (*type) {
      case FRPKTTYPE_AIP_DATA:
      case FRPKTTYPE_AIP_ROUTE_DESTROY:
      case FRPKTTYPE_AIP_ROUTE_CRTACK:
      case FRPKTTYPE_NDB_ROUTE_UPDATE:
        if( !frPktGetEmbeddedPacket(err, udpBuf, udpSz, (void **)&frPkt, &frPktSz) ) {
          zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                     "frPktGetEmbeddedPacket failed");
          return 0;
        }

        if( !frPktAipDataReceiveInBuffer(err, 
                                         aipData,
                                         frPkt, 
                                         frPktSz)) {
          zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                     "frPktAipDataReceiveInBuffer failed");
          return 0;
        }

        if (!frPktAipDataGetAci(err, aipData, &aciIn)) {
          zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                     "frPktAipDataGetAci() failed");
          return 0;
        }
        break;
      case FRPKTTYPE_AIP_ROUTE_CREATE:
        if( !frPktGetEmbeddedPacket(err, udpBuf, udpSz, (void **)&frPkt, &frPktSz) ) {
          zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                     "frPktGetEmbeddedPacket failed");
          return 0;
        }

        if( !frPktAipRteCrtReceiveInBuffer(err, 
                                           aip->route.packet.handler,
                                           frPkt, 
                                           frPktSz)) {
          zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                     "frPktAipRteCrtReceiveInBuffer failed");
          return 0;
        }
        if (!frPktAipRteCrtGetAci(err, aip->route.packet.handler, &aciIn)) {
          zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                     "frPktAipRteCrtGetAci() failed");
          return 0;
        }
        break;

      default:
        zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                      "Unknown packet type [%d]", *type);
        break;
      }

      if (*type != FRPKTTYPE_NDB_ROUTE_UPDATE) {
        /* Only check the ACI on packets which set the ACI */
        if (aip->route.context->route[i].aciOut == aciIn) {
          *aci = i;
          return 1;
        }
      } else {
        *aci = i;
        return 1;
      }

      /* Hmm, must have another route with first hop for this ACI */
      for (j = i + 1; j < aip->route.context->maxRoutes; j++) {
        if (aip->route.context->route[j].valid) {
          if ((aip->route.context->route[j].router[0].host.sin_addr.s_addr == 
               host.sin_addr.s_addr) && 
              (aip->route.context->route[j].router[0].host.sin_port ==
               host.sin_port)) {
            /* Found another one */
            if (aip->route.context->route[j].aciOut == aciIn) {
              *aci = j;
              return 1;
            }
          }
        }
      }
    }
  } else {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "FrLocalAipCtx is NULL");
  }
  return 0;
}

/**
 * Set up the fr_rte_set_exp structure and append to the ROUTE CREATE packet
 * that is being made.  Reads the public cryptography key for the AIP from
 * the filename which is passed.
 * 
 * @param err
 * @param aip
 * @param n                     Router number
 * @param aci                   Route number
 * @param reuse                 Reuse flag (nym prolongation)
 * @param numRoutes             Number of hops in this route
 * @param alg                   Crypto algorithm to use at the AIP
 * @param filename              Filename of the AIP's public key
 */ 
static int setNextHostInfo(ZkError err, FrLocalAipCtx aip, size_t n, ZkACI aci,
                           int reuse, size_t numRoutes, FrCryptSymAlg alg)
{
  struct fr_rte_set_exp route;
  FrCryptCrypt aipPubKey = NULL;
  struct timeval now;
  const char *func = "setNextHostInfo";

  if (!aip) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, 
               "FrLocalAipCtx is NULL");
    return 0;
  }

  if (!aip->kq) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,
               "Key query cache was never initialized");
    return 0;
  }

  if (n > numRoutes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Router number [%d] is greater than number of routes [%d]",
                  n, numRoutes);
    return 0;
  }

  route.fwdCrypt = aip->route.context->route[aci].router[n].forwSymKey;
  route.bckCrypt = aip->route.context->route[aci].router[n].backSymKey;
  route.rteAlg = alg;
  route.macKey = aip->route.context->route[aci].macKey;
  if (n != (numRoutes - 1)) {
    memcpy(&route.hostEnt, &aip->route.context->route[aci].router[n + 1].aipEntId, 
           sizeof(FrEntIdExp));
    route.flags = 0;
  } else {
    /* Last hop - mark it a 0 */
    memset(&route.hostEnt, 0, sizeof(FrEntIdExp));
    if (aip->route.context->route[aci].na.nymID.type == FRENT_TYPE_UNUSED_) {
      /* Mark it to be an unauthenticated route */
      route.hostEnt.type = FRENT_TYPE_ANON;
    } else {
      route.hostEnt.type = FRENT_TYPE_AIP;
    }
    route.flags = FRPKT_RTECRT_FL_LAST_RTE;

    if (reuse) {
      route.specAci = aip->route.context->route[aci].aciExit;
      route.flags = route.flags | FRPKT_RTECRT_FL_SPEC_ACI;
    }
  }

  if (gettimeofday(&now, 0) == -1) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, "gettimeofday() failed");
    return 0;
  }
    
  route.expTime = now.tv_sec + aip->freedomDelta + 
    aip->route.context->route[aci].expiry;

  {
    FrKeyQryState state;
    FrKeyQryQueryExp kqry;

    memset(&kqry, 0, sizeof(kqry));
    kqry.keyVersion = 0;
    kqry.keyType = FRCRYPT_KEYTYPE_CRYPTION;
    memcpy(&kqry.keyOwner, &aip->route.context->route[aci].router[n].aipEntId,
           sizeof(FrEntIdExp));
    
    if (!frKeyQryCacheFindKey(err, aip->kq, &kqry, &state, &aipPubKey)) {
      return 0;
    }

    if (state != FRKEYQRY_ST_VALID) {
      /* Not found in the cache */
      return 0;
    }
  }

  if (!frPktAipRteCrtAppendRouteExp(err, aip->route.packet.handler, aipPubKey, &route)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, 
               "frPktAipRteCrtAppendRouteExp() failed");
    return 0;
  }

  return 1;
}

/*
 * Make a ROUTE CREATE packet.
 *
 * @param err
 * @param aip
 * @param udpBuf                Buffer to be filled
 * @param udpSz                 Size of buffer/size of data (returned)
 * @param aci                   Route number
 * @param reuse                 Whether or not to reuse the exit funnel
 * @param alg                   Crypto algorithm to use
 */
DLLExport int frLocalAipMakeCreatePkt(ZkError err, FrLocalAipCtx aip,
                                      char *udpBuf, size_t *udpSz,
                                      ZkACI aci, int reuse, FrCryptSymAlg alg)
{
  size_t i;
  size_t xmitSz;
  size_t maxsz;
  const char *func = "frLocalAipMakeCreatePkt";
  char  *frPkt;
  size_t frPktSz;

  /* param checking */
  if (!aip ) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, 
               "FrLocalAipCtx object is NULL");
    return 0;
  }

  if( !frPktAipDataPayloadMaxSize(err, aip->data.protocol.version, &maxsz) ) {
    zkErrorSet(err, func, ZKWARN, ZKMFRLAIP, ZKEPARMBAD, 
               "frPktAipDataPayloadMaxSize failed");
    return 0;
  }

  /* Bounds check */
  if( *udpSz < maxsz ) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD, 
                  "Buffer size (%d) is too small (should be > %d)", 
                  *udpSz, maxsz);
    return 0;
  }

  if (aci > aip->route.context->maxRoutes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route ACI [%d] is out of bounds", aci);
    return 0;
  }

  /* Check if this route is valid */ 
  if (!aip->route.context->route[aci].valid) {
    /* Not a valid route */
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route ACI [%d] is not valid", aci);
    return 0;
  }

  if( !frPktGetEmbeddedPacket(err, udpBuf, *udpSz, (void **)&frPkt, &frPktSz) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktGetEmbeddedPacket failed");
    return 0;
  }

  if (!frPktAipRteCrtCreateInBuffer(err, aip->route.packet.handler, frPkt, frPktSz)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, 
               "Unable to set the packet buffer");
    return 0;
  }
 
  for (i = 0; i < aip->route.context->route[aci].numRouters; i++) {
    if (!setNextHostInfo(err, aip, i, aci, reuse, 
                         aip->route.context->route[aci].numRouters,
                         alg)) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKERROR, 
                    "Error setting hop %d", i);
      return 0;
    }
  }

  if (aip->route.context->route[aci].na.nymID.type != FRENT_TYPE_UNUSED_) {
    /* No Nym associated with the route / force localhost */
    if (!frPktAipRteCrtAppendNymAuthExp(err, aip->route.packet.handler,
                                        &(aip->route.context->route[aci].na))) {
      zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, 
                 "Error appending nym authentication");
      return 0;
    }
  }

  if (!frPktAipRteCrtEncryptAllRoutes(err, aip->route.packet.handler)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, 
               "Error encrypting the ROUTE CREATE packet");
    return 0;
  }

  /* Set ACI */
  if (!frPktAipRteCrtSetAci(err, aip->route.packet.handler,
                            aip->route.context->route[aci].aciOut)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, "Error setting ACI");
    return 0;
  }

  if( !frPktAipRteCrtGetCurrentSize(err, aip->route.packet.handler, &xmitSz)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktAipRteCtrGetCurrentSize() failed");
    return 0;
  }

  *udpSz = frPktGetSize(xmitSz);

  /* Add the version and type */
  if (!frPktSetVersion(err, udpBuf, *udpSz, aip->route.protocol.version)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktSetVersion() failed");
    return 0;
  }
  if (!frPktSetType(err, udpBuf, *udpSz, FRPKTTYPE_AIP_ROUTE_CREATE)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktSetType() failed");
    return 0;
  }
  if (!frPktSetPriority(err, udpBuf, *udpSz, aip->route.context->route[aci].prio)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktSetPriority() failed");
    return 0;
  }

  return 1;
}

/**
 * Link encrypt a buffer based on the packet type.
 * 
 * @param err
 * @param aip
 * @param aci                   Route number
 * @param type                  Packet type
 * @param udpBuf                Buffer to encrypt
 * @param udpSz                 Size of the data
 */
DLLExport int frLocalAipLinkEncrypt(ZkError err, FrLocalAipCtx aip, ZkACI aci,
                                    FrPktType type, char *udpBuf, size_t udpSz)
{
  uint8_t cur;
  const char *func = "frLocalAipLinkEncrypt";

  if (!aip) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, 
               "FrLocalAipCtx object is NULL");
    return 0;
  }

  if (aci > aip->route.context->maxRoutes) {
    /* out of bounds */
    return 0;                   /* bug */
  }
  /* Check if this route is valid */ 
  if (!aip->route.context->route[aci].valid) {
    /* Not a valid route */
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD, 
                  "Route for ACI %d is not valid", aci);
    return 0;                   /* bug */
  }

  cur = aip->route.context->route[aci].current;
  if (!frPktLinkEncrypt(err, aip->route.context->route[aci].symKey[cur], udpBuf, 
                        udpSz)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktLinkEncrypt() failed");
    return 0;
  }

  if( !frPktSetLekv(err, udpBuf, udpSz, cur) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktSetLekv() failed");
    return 0;
  }

  return 1;
}

/**
 * Take the ipBuf and convert it into a anonymous packet.  Multiple levels 
 * of encryption to be performed on the ipBuf contents.  Version, type aci
 * added to the udpBuf payload and link encrypted.
 *
 * @param err
 * @param aip
 * @param aci                   Route number 
 * @param ipBuf                 Buffer contains an IP pkt to be anonymized
 * @param ipSize                Length of the IP pkt
 * @param udpBuf                Buffer for the anonymized pkt
 * @param udpSize               Maximum size, on return, size of anon pkt
 */
DLLExport int frLocalAipMakeDataPkt(ZkError err, FrLocalAipCtx aip, ZkACI aci,
                                    char *ipBuf, size_t ipSize,
                                    char *udpBuf, size_t *udpSize)
{
  return frLocalAipMakeDataPktUsingAipData(err, aip, aip->data.packet.handler,
                                           aci, ipBuf, ipSize, udpBuf, udpSize);
}

/**
 * Take the ipBuf and convert it into a anonymous packet.  Multiple levels 
 * of encryption to be performed on the ipBuf contents.  Version, type aci
 * added to the udpBuf payload and link encrypted.
 *
 * @param err
 * @param aip
 * @param aipData               data packet object to use
 * @param aci                   Route number 
 * @param ipBuf                 Buffer contains an IP pkt to be anonymized
 * @param ipSize                Length of the IP pkt
 * @param udpBuf                Buffer for the anonymized pkt
 * @param udpSize               Maximum size, on return, size of anon pkt
 */
DLLExport int frLocalAipMakeDataPktUsingAipData(ZkError err, FrLocalAipCtx aip,
                                                FrPktAD aipData,
                                                ZkACI aci,
                                                char *ipBuf, size_t ipSize,
                                                char *udpBuf, size_t *udpSize)
{
  char payload[FRPKT_BUF_MAX_SIZE];
  size_t i;
  size_t max, maxsz;
  const char *func = "frLocalAipMakeDataPktUsingAipData";
  char  *frPkt;
  size_t frPktSz;
  uint16_t mss;
  char debugout[120];
  ZkTCPHeader* ph;

  zkErrorClear(err);

  if (!aip) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, 
               "FrLocalAipCtx is NULL");
    return 0;
  }

  if (aci > aip->route.context->maxRoutes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD, 
                  "ACI %d is out of range", aci);
    return 0;
  }

  if (!aip->route.context->route[aci].valid) {
    /* Not a valid route */
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD, 
                  "Route for ACI %d is not valid", aci);
    return 0;
  }

 
  if( !frPktAipDataPayloadMaxSize(err, aip->data.protocol.version, &maxsz) ) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD, 
                  "frPktAipDataPayloadMaxSize failed");
    return 0;
  }    

  /* Bounds check */
  if( *udpSize < maxsz ) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD, 
                  "Buffer size (%d) is too small (should be >= %d)", 
                  *udpSize, maxsz);
    return 0;
  }

  if( (ipSize + 2) > maxsz ) {
    /* Payload too big */
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD,
                  "Size of IP packet (%d) is too large", ipSize);
    return 0;
  }
  
  /* Get the length of the IP payload */
  NBO_FROM_INT16(payload, ipSize);
  memcpy(&payload[2], ipBuf, ipSize); /* Copy the payload */

  
  /* We don't call zkTCPHeaderSetMSS() if there is no MSS set.
     Do we want to artificially add an MSS? */

	if( aip->data.protocol.version == FRPKT_DATA_VERSION_1 )
	{
#ifndef NDEBUG
	  ClientTrace("frLocalAipMakeDataPkt()", "Using v1 freedom protocol, not checking MSS size");
#endif
	}
	else
	{
		ph = (ZkTCPHeader*)&payload[2];

		/* Do we want to verify if it is a syn packet and proceed only if so ? */

		if( (ph->IP.Protocol == IPPROTO_TCP) && zkTCPHeaderGetMSS( ph, &mss) )
		{
#ifndef NDEBUG
			snprintf(debugout, 120, "This outgoing TCP packet has an MSS of [%d]. If it is larger than MAX_MSS [%d] it will be reduced\n", mss, MAX_MSS);
			ClientTrace("frLocalAipMakeDataPkt", debugout);
#endif
			if(mss > MAX_MSS)
			{
				/* Do we want to check for a fragment before reducing the mss? */
				zkTCPHeaderModifyMSS ( ph, MAX_MSS);
#ifndef NDEBUG				
				zkTCPHeaderGetMSS( ph, &mss);
				snprintf(debugout, 120, "The MSS is now [%d] mss\n", mss);
				ClientTrace("frLocalAipMakeDataPkt", debugout);
#endif
			}
		}
		else
		{
			ClientTrace("frLocalAipMakeDataPkt", "Not a TCP packet or no MSS to manipulate");
		}
	}

  if( !frPktGetEmbeddedPacket(err, udpBuf, *udpSize, (void **)&frPkt, &frPktSz) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktGetEmbeddedPacket failed");
    return 0;
  }

  if( !frPktAipDataCreateInBuffer(err, 
                                  aipData,
                                  frPkt, 
                                  frPktSz)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, 
               "frPktAipDataCreateInBuffer() failed");
    return 0;
  }

  if( !frPktAipDataSetVersion(err, aipData, aip->data.protocol.version) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, 
               "frPktAipDataSetVersion() failed");
    return 0;
  }


  if (!frPktAipDataSetPayload(err, aipData, aip->route.context->route[aci].macKey,
                              payload, ipSize + 2)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, 
               "frPktAipDataSetPayload() failed");
    return 0;
  }

  max = aip->route.context->route[aci].numRouters;
  for (i = 0; i < max; i++) {
    /* Last router first */
    if (!frPktAipDataEncryptPayload(err, aipData,
                                    aip->route.context->route[aci].router[max - 1 - i].backSymKey)) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKERROR, 
                    "Payload encryption for %s:%d failed", 
                    inet_ntoa(aip->route.context->route[aci].router[max - 1 - i].host.sin_addr),
                    ntohs(aip->route.context->route[aci].router[max - 1 - i].host.sin_port));
      return 0;                 
    }
  }

  /* Set the ACI */
  if (!frPktAipDataSetAci(err, aipData, aip->route.context->route[aci].aciOut)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktAipDataSetAci() failed");
    return 0;
  }

  {
    size_t adsz;

    if( !frPktAipDataGetCurrentSize(err, aipData, &adsz)) {
      zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                 "frPktAipDataGetCurrentSize() failed");
      return 0;
    }

    *udpSize = frPktGetSize(adsz);
  }

  /* Add the version and type */
  if (!frPktSetVersion(err, udpBuf, *udpSize, aip->data.protocol.version)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktSetVersion() failed");
    return 0;
  }
  if (!frPktSetType(err, udpBuf, *udpSize, FRPKTTYPE_AIP_DATA)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktSetType() failed");
    return 0;
  }
  if (!frPktSetPriority(err, udpBuf, *udpSize, aip->route.context->route[aci].prio)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktSetPriority() failed");
    return 0;                   
  }

  return 1;
}

/**
 * Extract the payload from the data packet.
 * 
 * @param err
 * @param aip
 * @param aci                   Route number
 * @param udpBuf                Incoming buffer
 * @param udpSize               Size of the incoming packet
 * @param ipBuf                 Extracted payload buffer
 * @param ipSize                Size of the buffer/size of payload (returned)
 */
DLLExport int frLocalAipReadDataPkt(ZkError err, FrLocalAipCtx aip, ZkACI aci,
                                    char *udpBuf, size_t *udpSize, 
                                    char *ipBuf,  size_t *ipSize)
{
  return frLocalAipReadDataPktUsingAipData(err, aip, aip->data.packet.handler,
                                           aci, udpBuf, udpSize, ipBuf, ipSize);
}

/**
 * Extract the payload from the data packet.
 * 
 * @param err
 * @param aip
 * @param aipData               data packet object to use
 * @param aci                   Route number
 * @param udpBuf                Incoming buffer
 * @param udpSize               Size of the incoming packet
 * @param ipBuf                 Extracted payload buffer
 * @param ipSize                Size of the buffer/size of payload (returned)
 */
DLLExport int frLocalAipReadDataPktUsingAipData(ZkError err, FrLocalAipCtx aip,
                                                FrPktAD aipData, ZkACI aci,
                                                char *udpBuf, size_t *udpSize, 
                                                char *ipBuf,  size_t *ipSize)
{
  size_t i;
  size_t max;
  size_t maxsz;
  char *tmp;
  const char *func = "frLocalAipReadDataPktUsingAipData";
  char  *frPkt;
  size_t frPktSz;
  uint16_t mss;
  char debugout[120];
  ZkTCPHeader* ph;

  /* aip->data.packet.handler should have been set in the LinkDecrypt function */
  if (!aip) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, 
               "FrLocalAipCtx object is NULL");
    return 0;
  }

  if (aci > aip->route.context->maxRoutes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD,
                  "Route aci [%d] is out of range [%d]", aci, 
                  aip->route.context->maxRoutes);
    return 0;
  }

  if( !frPktAipDataPayloadMaxSize(err, aip->data.protocol.version, &maxsz) ) {
    zkErrorSet(err, func, ZKWARN, ZKMFRLAIP, ZKEPARMBAD, 
               "frPktAipDataPayloadMaxSize failed");
    return 0;
  }

  if( aip->data.protocol.version == FRPKT_DATA_VERSION_1 )
  {
#ifndef NDEBUG
	  ClientTrace("frLocalAipReadDataPkt()", "Using v1 freedom protocol, not checking packet size");
#endif
  }
  else
  {

	  /* Bounds check */
	  if (*udpSize > maxsz) {
		zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD, 
					  "Buffer size (%d) is too big (should be <= %d)", *udpSize, 
					  maxsz);
		return 0;
	  }
  }

  if (!aip->route.context->route[aci].valid) {
    /* Not a valid route */
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD, 
                  "Route for ACI %d is not valid", aci);
    return 0;
  }

  if( !frPktGetEmbeddedPacket(err, udpBuf, *udpSize, (void **)&frPkt, &frPktSz) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktGetEmbeddedPacket failed");
    return 0;
  }
 
  if( !frPktAipDataReceiveInBuffer(err, 
                                   aipData,
                                   frPkt, 
                                   frPktSz) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktAipDataReceiveInBuffer failed");
    return 0;
  }


  /* Decrypt the multi-encrypted payload */
  max = aip->route.context->route[aci].numRouters;
  for (i = 0; i < max; i++) {
    if (!frPktAipDataDecryptPayload(err, aipData,
                                    aip->route.context->route[aci].router[i].forwSymKey)) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                    "Decryption for router %d failed", i);
      return 0;
    }
  }

  /* Done with decryption, get the payload */
  if (!frPktAipDataGetPayload(err, aipData, aip->route.context->route[aci].macKey,
                              (void **) &tmp, ipSize)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktAipDataGetPayload() failed");
    return 0;
  }

  {
    size_t adsz;

    if( !frPktAipDataGetCurrentSize(err, aipData, &adsz)) {
      zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                 "frPktAipDataGetCurrentSize() failed");
      return 0;
    }

    *udpSize = frPktGetSize(adsz);
  }

  /* Now I should have LEN + IP-Payload */
  /* Strip off the LEN and Garbage after LEN */
  /* Set *ipSize */
  INT16_FROM_NBO(*ipSize, tmp);

  if( *ipSize > *udpSize) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "Magic 2-byte payload size greater than freedom header payload size");
    return 0;
  }
  memcpy(ipBuf, tmp + 2, *ipSize);


  /* We don't call zkTCPHeaderSetMSS() if there is no MSS set.
	 Do we want to artificially add an MSS? */

	if( aip->data.protocol.version == FRPKT_DATA_VERSION_1 )
	{
#ifndef NDEBUG
	  ClientTrace("frLocalAipReadDataPkt()", "Using v1 freedom protocol, not checking MSS size");
#endif
	}
	else
	{
		ph = (ZkTCPHeader*)ipBuf;

		if( (ph->IP.Protocol == IPPROTO_TCP) && zkTCPHeaderGetMSS( ph, &mss) )
		{
#ifndef NDEBUG
			snprintf(debugout, 120, "This incoming TCP packet has an MSS of [%d]. If it is larger than MAX_MSS [%d] it will be reduced\n", mss, MAX_MSS);
			ClientTrace("frLocalAipReadDataPkt", debugout);
#endif
			if(mss > MAX_MSS)
			{
				/* Do we want to check for a fragment before reducing the mss? */
				zkTCPHeaderModifyMSS ( ph, MAX_MSS);
#ifndef NDEBUG				
				zkTCPHeaderGetMSS( ph, &mss);
				snprintf(debugout, 120, "The MSS is now [%d] mss\n", mss);
				ClientTrace("frLocalAipMakeDataPkt", debugout);
#endif
			}
		}
		else
		{
			ClientTrace("frLocalAipReadDataPkt", "Not a TCP packet or no MSS to manipulate");
		}
	}

  return 1;
}

/**
 * Make a DESTROY packet for the route "aci".
 *
 * @param err
 * @param aip
 * @param udpBuf                Buffer to store the packet
 * @param udpSz                 Size of the buffer/size of packet (returned)
 * @param aci                   Route ACI number
 */
DLLExport int frLocalAipMakeDestroyPkt(ZkError err, FrLocalAipCtx aip,
                                       char *udpBuf, size_t *udpSz,
                                       ZkACI aci)
{
  return frLocalAipMakeDestroyPktUsingAipData(err, aip, aip->data.packet.handler,
                                              udpBuf, udpSz, aci);
}

/**
 * Make a DESTROY packet for the route "aci".
 *
 * @param err
 * @param aip
 * @param aipData               data packet object to use
 * @param udpBuf                Buffer to store the packet
 * @param udpSz                 Size of the buffer/size of packet (returned)
 * @param aci                   Route ACI number
 */
DLLExport int frLocalAipMakeDestroyPktUsingAipData(ZkError err, FrLocalAipCtx aip,
                                                   FrPktAD aipData,
                                                   char *udpBuf, size_t *udpSz,
                                                   ZkACI aci)
{
  const char *func = "frLocalAipMakeDestroyPktUsingAipData";
  char  *frPkt;
  size_t frPktSz;

  if (!aip) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, 
               "FrLocalAipCtx object is NULL");
    return 0;
  }
  if (*udpSz < FRPKT_AIP_ROUTE_DESTROY_SIZE) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD,
                  "Data buffer size %d is too small, need at least %d",
                  *udpSz, FRPKT_AIP_ROUTE_DESTROY_SIZE);
    return 0;
  }

  if (aci > aip->route.context->maxRoutes) {
    /* out of bounds */
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD,
                  "Route aci [%d] is out of range [%d]", aci, 
                  aip->route.context->maxRoutes);
    return 0;
  }

  /* Check if this route is valid */ 
  if (!aip->route.context->route[aci].valid) {
    /* Not a valid route */
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD, 
                  "Route for ACI %d is not valid", aci);
    return 0;
  }

  if( !frPktGetEmbeddedPacket(err, udpBuf, *udpSz, (void **)&frPkt, &frPktSz) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktGetEmbeddedPacket failed");
    return 0;
  }

  if( !frPktAipDataCreateInBuffer(err, 
                                  aipData,
                                  frPkt, 
                                  frPktSz)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktAipDataCreateInBuffer() failed");    
    return 0;
  }


  /* Set the ACI */
  if (!frPktAipDataSetAci(err, aipData, aip->route.context->route[aci].aciOut)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktAipDataSetAci() failed");
    return 0;
  }

  /* Add the version and type */
  if (!frPktSetVersion(err, udpBuf, *udpSz, aip->data.protocol.version)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktSetVersion() failed");
    return 0;
  }

  if (!frPktSetType(err, udpBuf, *udpSz, FRPKTTYPE_AIP_ROUTE_DESTROY)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktSetType() failed");
    return 0;
  }

  if (!frPktSetPriority(err, udpBuf, *udpSz, aip->route.context->route[aci].prio)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktSetPriority() failed");
    return 0;
  }

  /* We need to call this function in order to set _some_ sort of payload
   * such that the packets size is properly calculated and returned to the caller --ol
   */
  if (!frPktAipDataSetRandomFiller(err, aipData)) {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR,
                 "frPktAipDataSetRandomFiller() failed");
      return 0;
  }

  {
    size_t xmitSz = 0;

    if( !frPktAipDataGetCurrentSize(err, aipData, &xmitSz)) {
      zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                 "frPktAipRteCtrGetCurrentSize() failed");
      return 0;
    }

    *udpSz = frPktGetSize(xmitSz);
  }
  
  return 1;
}

/**
 * Incoming packet is determined to be a ROUTE CREATE ACK packet.  Store the
 * ACI from the payload and/or check to see if it is the correct one if reuse 
 * is set.
 *
 * @param err
 * @param aip
 * @param aci                   Route aci
 * @param udpBuf                Buffer holding the packet
 * @param udpSz                 Size of the packet
 * @parma reuse                 Reuse bit
 */
DLLExport int frLocalAipReadCreateAckPkt(ZkError err, FrLocalAipCtx aip,
                                         ZkACI aci, char *udpBuf, size_t udpSz,
                                         int reuse)
{
  return frLocalAipReadCreateAckPktUsingAipData(err, aip, aip->data.packet.handler,
                                                aci, udpBuf, udpSz, reuse);
}

/**
 * Incoming packet is determined to be a ROUTE CREATE ACK packet.  Store the
 * ACI from the payload and/or check to see if it is the correct one if reuse 
 * is set.
 *
 * @param err
 * @param aip
 * @param aipData               data packet object to use
 * @param aci                   Route aci
 * @param udpBuf                Buffer holding the packet
 * @param udpSz                 Size of the packet
 * @parma reuse                 Reuse bit
 */
DLLExport int frLocalAipReadCreateAckPktUsingAipData(ZkError err, FrLocalAipCtx aip,
                                                     FrPktAD aipData,
                                                     ZkACI aci, char *udpBuf,
                                                     size_t udpSz, int reuse)
{
  const char *func = "frLocalAipReadCreateAckPktUsingAipData";
  size_t i, max, maxsz;
  char *tmp;
  ZkACI aciExit;
  size_t psz;
  char  *frPkt;
  size_t frPktSz;

  if (!aip) {
    zkErrorSet(err, func, ZKWARN, ZKMFRLAIP, ZKENULLPTR,
               "FrLocalAipCtx object is NULL");
    return 0;
  }

  if( !frPktAipDataPayloadMaxSize(err, aip->data.protocol.version, &maxsz) ) {
    zkErrorSet(err, func, ZKWARN, ZKMFRLAIP, ZKEPARMBAD, 
               "frPktAipDataPayloadMaxSize failed");
    return 0;
  }


  if( aip->route.protocol.version == FRPKT_ROUTE_VERSION_1 )
  {
#ifndef NDEBUG
	  ClientTrace("frLocalAipReadCreateAckPkt()", "Using v1 freedom protocol, not checking packet size");
#endif
  }
  else
  {
	  if( udpSz > maxsz ) {
		zkErrorSet_va(err, func, ZKWARN, ZKMFRLAIP, ZKEPARMBAD, 
               "ROUTE_CREATE_ACK packet is too big: %d > %d", udpSz, maxsz);
		return 0;
	  }
  }

  if (aci > aip->route.context->maxRoutes) {
    zkErrorSet(err, func, ZKWARN, ZKMFRLAIP, ZKEPARMBAD,
               "ACI value given is out of range");
    return 0;                   
  }

  if (!aip->route.context->route[aci].valid) {
    zkErrorSet(err, func, ZKWARN, ZKMFRLAIP, ZKEDATABAD,
               "Route for ACI is not valid");
    return 0;
  }

  if( !frPktGetEmbeddedPacket(err, udpBuf, udpSz, (void **)&frPkt, &frPktSz) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktGetEmbeddedPacket failed");
    return 0;
  }

  if( !frPktAipDataReceiveInBuffer(err, 
                                   aipData,
                                   frPkt, 
                                   frPktSz)) {
    zkErrorSet(err, func, ZKWARN, ZKMFRLAIP, ZKERROR,
               "frPktAipDataReceiveInBuffer() failed");
    return 0;
  }

  max = aip->route.context->route[aci].numRouters;
  for (i = 0; i < max; i++) {
    if (!frPktAipDataDecryptPayload(err, aipData,
                                    aip->route.context->route[aci].router[i].forwSymKey)) {
      zkErrorSet(err, func, ZKWARN, ZKMFRLAIP, ZKERROR, 
                 "Failed to remove all layers of the encryption");
      return 0;
    }
  }

  /* Done with decryption, get the payload */
  if (!frPktAipDataGetPayload(err, aipData, aip->route.context->route[aci].macKey,
                              (void **) &tmp, &psz)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktAipDataGetPayload() failed");    
    return 0;
  }

  if( psz > maxsz ) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                  "Size of payload [%d] is too large, should be at most [%d]",
                  psz, maxsz);
    return 0;
  }

  INT16_FROM_NBO(aciExit, tmp);
  if (reuse) {
    if (aip->route.context->route[aci].aciExit != aciExit) {
      zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD,
                 "Unexpected payload");
      return 0;                 /* Did not confirm the reuse */
    }
  } else {
    aip->route.context->route[aci].aciExit = aciExit;
  }
  return 1;
}


/* 
 *  Make a keep-alive packet.  This is to be sent, by a client, to the first
 *  AIP to let it know that it's still alive.  The AIP doesn't need to look at
 *  the contents, so the packet can be complete garbage.
 *
 *  @param err                   ZkError Object
 *  @param aip                   Unused; kept for consistency with other calls
 *  @param pbuf                  Buffer to hold the outgoing packet
 *  @param psz                   Size of the buffer, the usage will be returned
 *  @return                      1 on success, 0 on failure
 *  @version                     1.0
 *  @author                      Bill Heelan
 */ 
DLLExport int frLocalAipMakeKeepAlivePkt(ZkError err, FrLocalAipCtx aip, char *pbuf, size_t *psz)
{
  const char *fn = "frLocalAipMakeKeepAlivePkt";
  int ret = 0;
  size_t n;

  if ( ! aip) {
    zkErrorSet(err, fn, ZKERR, ZKMFRLAIP, ZKENULLPTR, "aip is NULL");
    goto done;
  }
  if ( ! pbuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRLAIP, ZKENULLPTR, "pbuf is NULL");
    goto done;
  }
  if ( ! psz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRLAIP, ZKENULLPTR, "psz is NULL");
    goto done;
  }

  /*
   *  Set `n' to the size of the buffer needed to hold a version 1 AIP data
   *  packet, including link layer overhead.
   */

  if ( ! frPktAipDataGetMaxSize(err, 1, &n)) {
    goto done;
  }
  n = frPktGetSize(n);

  if (*psz < n) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRLAIP, ZKEPARMBAD,
                  "*psz too small (%lu < %lu)",
                  (unsigned long)*psz, (unsigned long)n);
    goto done;
  }
  
  /*
   *  Put random data in the packet.  We set the LEKV to a fixed value.  It
   *  doesn't matter whether or not it's correct, since the AIP won't be able
   *  to decrypt the packet anyway.
   */
  
  if ( ! (zkRandGetData(err, pbuf, n-1) &&
          frPktSetLekv(err, pbuf, n, 0))) {
    goto done;
  }

  *psz = n;
  
  ret = 1;

 done:

  return ret;
}


/*********** 
 *  XXX: These *GarbagePkt* functions are deprecated
 *   use the MakeKeepAlive function instead --ol
 ***********/

/**
 * Make a garbage packet for the route `aci'.  The buffer is at least 
 * FRPKT_BUF_MAX_SIZE and this is filled with random data except for the last
 * byte which is the current version of the symKey.
 *
 * @param err                   ZkError Object
 * @param aip                   FrLocalAipCtx
 * @param aci                   Route ACI
 * @param udpBuf                Buffer to hold the outgoing packet
 * @param udpSize               Size of the buffer, the usage will be returned
 * @return                      1 on success, 0 on failure
 * @version                     1.0
 * @author                      Kent Tse
 */
DLLExport int frLocalAipMakeGarbagePkt(ZkError err, FrLocalAipCtx aip,
                                       ZkACI aci, char *udpBuf, 
                                       size_t *udpSize)
{
  return frLocalAipMakeGarbagePktUsingAipData(err, aip, aip->data.packet.handler,
                                              aci, udpBuf, udpSize);
}

/**
 * Make a garbage packet for the route `aci'.  The buffer is at least 
 * FRPKT_BUF_MAX_SIZE and this is filled with random data except for the last
 * byte which is the current version of the symKey.
 *
 * @param err                   ZkError Object
 * @param aip                   FrLocalAipCtx
 * @param aipData               data packet object to use
 * @param aci                   Route ACI
 * @param udpBuf                Buffer to hold the outgoing packet
 * @param udpSize               Size of the buffer, the usage will be returned
 * @return                      1 on success, 0 on failure
 * @version                     1.0
 * @author                      Kent Tse
 */
DLLExport int frLocalAipMakeGarbagePktUsingAipData(ZkError err, FrLocalAipCtx aip,
                                                   FrPktAD aipData,
                                                   ZkACI aci, char *udpBuf, 
                                                   size_t *udpSize)
{
  const char *func = "frLocalAipMakeGarbagePktUsingAipData";  
  size_t maxsz;
  char  *frPkt;
  size_t frPktSz;

  if (!aip) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, 
               "FrLocalAipCtx is NULL");
    return 0;
  }

  if (aci > aip->route.context->maxRoutes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD, 
                  "ACI %d is out of range", aci);
    return 0;
  }

  if (!aip->route.context->route[aci].valid) {
    /* Not a valid route */
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD, 
                  "Route for ACI %d is not valid", aci);
    return 0;
  }

  if( !frPktAipDataPayloadMaxSize(err, aip->data.protocol.version, &maxsz) ) {
    zkErrorSet(err, func, ZKWARN, ZKMFRLAIP, ZKEPARMBAD, 
               "frPktAipDataPayloadMaxSize failed");
    return 0;
  }

  /* Bounds check */
  if (*udpSize < maxsz) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD, 
                  "Buffer size (%d) is not large enough (should be >= %d)", *udpSize, 
                  maxsz);
    return 0;
  }

  if( !frPktGetEmbeddedPacket(err, udpBuf, *udpSize, (void **)&frPkt, &frPktSz) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktGetEmbeddedPacket failed");
    return 0;
  }

  if( !frPktAipDataCreateInBuffer(err, 
                                  aipData,
                                  frPkt, 
                                  frPktSz)) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR,
               "frPktAipDataCreateInBuffer() failed");
    return 0;
  }

  if (!frPktAipDataSetVersion(err, aipData, aip->data.protocol.version)) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR,
               "frPktAipDataSetVersion() failed");
    return 0;
  }


  if( aip->data.protocol.version == FRPKT_DATA_VERSION_1 ) {
    if (!frPktAipDataSetRandomFiller(err, aipData)) {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR,
                 "frPktAipDataSetRandomFiller() failed");
      return 0;
    }
  }

  {
    size_t adsz;

    if( !frPktAipDataGetCurrentSize(err, aipData, &adsz)) {
      zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                 "frPktAipDataGetCurrentSize() failed");
      return 0;
    }

    *udpSize = frPktGetSize(adsz);
  }

  /* Set the last byte to the current version of the symKey */
  udpBuf[*udpSize - 1] = aip->route.context->route[aci].current;

  return 1;
}

/* 
 * Make a NDB ROUTE update packet. 
 *
 * @param err 
 * @param aip 
 * @param aci                   Route ACI
 * @param payload               Payload buffer
 * @param psz                   Size of payload
 * @param udpBuf                Buffer to hold packet
 * @param udpSize               Size of the buffer, size of packet (returned)
 */ 
DLLExport int frLocalAipMakeNdbUpdatePkt(ZkError err, FrLocalAipCtx aip,
                                         ZkACI aci, char *payload, size_t psz,
                                         char *udpBuf, size_t *udpSize)
{
  return frLocalAipMakeNdbUpdatePktUsingAipData(err, aip, aip->data.packet.handler,
                                                aci, payload, psz, udpBuf, udpSize);
}

/* 
 * Make a NDB ROUTE update packet. 
 *
 * @param err 
 * @param aip 
 * @param aipData               data packet object to use
 * @param aci                   Route ACI
 * @param payload               Payload buffer
 * @param psz                   Size of payload
 * @param udpBuf                Buffer to hold packet
 * @param udpSize               Size of the buffer, size of packet (returned)
 */ 
DLLExport int frLocalAipMakeNdbUpdatePktUsingAipData(ZkError err, FrLocalAipCtx aip,
                                                     FrPktAD aipData,
                                                     ZkACI aci, char *payload, size_t psz,
                                                     char *udpBuf, size_t *udpSize)
{
  const char *func = "frLocalAipMakeNdbUpdatePktUsingAipData";
  char  *frPkt;
  size_t frPktSz;

  if (!aip) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,
               "FrLocalAipCtx object is NULL");
    return 0;
  }

  if (aci > aip->route.context->maxRoutes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD, 
                  "ACI %d is out of range", aci);
    return 0;
  }

  if (!aip->route.context->route[aci].valid) {
    /* Not a valid route */
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD, 
                  "Route for ACI %d is not valid", aci);
    return 0;
  }

  if (psz > *udpSize) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD,
                  "Payload size is larger than buffer size");
    return 0;
  }

  if( !frPktGetEmbeddedPacket(err, udpBuf, *udpSize, (void **)&frPkt, &frPktSz) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktGetEmbeddedPacket failed");
    return 0;
  }

  if( !frPktAipDataCreateInBuffer(err, 
                                  aipData,
                                  frPkt, 
                                  frPktSz)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktAipDataCreateInBuffer() failed");
    return 0;
  }

  {
    size_t adsz;

    if( !frPktAipDataGetCurrentSize(err, aipData, &adsz)) {
      zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                 "frPktAipDataGetCurrentSize() failed");
      return 0;
    }

    *udpSize = frPktGetSize(adsz);
  }

    
  /* Turn MAC usage off */
  if (!frPktAipDataSetUseMac(err, aipData, 0)) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, 
               "Unable to turn off MAC key usage for payload");
    return 0;
  }

  if (!frPktAipDataSetPayload(err, aipData, NULL, payload, psz)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, 
               "frPktAipDataSetPayload() failed");
    return 0;
  }

  /* Turn MAC usage on */
  if (!frPktAipDataSetUseMac(err, aipData, 1)) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, 
               "Unable to turn on MAC key usage for payload");
    return 0;
  }

  /* Add the version and type */
  if (!frPktSetVersion(err, udpBuf, *udpSize, aip->data.protocol.version)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktSetVersion() failed");
    return 0;
  }
  if (!frPktSetType(err, udpBuf, *udpSize, FRPKTTYPE_NDB_ROUTE_UPDATE)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktSetType() failed");
    return 0;
  }
  if (!frPktSetPriority(err, udpBuf, *udpSize, aip->route.context->route[aci].prio)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frPktSetPriority() failed");
    return 0;                   
  }
  return 1;
}

/**
 * Upon receiving a NDB ROUTE UPDATE packet this function should be called.
 *
 * @param err
 * @param aip
 * @param aci
 * @param udpBuf
 * @param udpSize
 * @param payBuf
 * @param paySz
 */
DLLExport int frLocalAipReadNdbUpdatePkt(ZkError err, FrLocalAipCtx aip,
                                         ZkACI aci, char *udpBuf,
                                         size_t udpSize, char *payBuf,
                                         size_t *paySz)
{
  return frLocalAipReadNdbUpdatePktUsingAipData(err, aip, aip->data.packet.handler,
                                                aci, udpBuf, udpSize, payBuf, paySz);
}

/**
 * Upon receiving a NDB ROUTE UPDATE packet this function should be called.
 *
 * @param err
 * @param aip
 * @param aipData               data packet object to use
 * @param aci
 * @param udpBuf
 * @param udpSize
 * @param payBuf
 * @param paySz
 */
DLLExport int frLocalAipReadNdbUpdatePktUsingAipData(ZkError err, FrLocalAipCtx aip,
                                                     FrPktAD aipData,
                                                     ZkACI aci, char *udpBuf,
                                                     size_t udpSize, char *payBuf,
                                                     size_t *paySz)
{
  const char *func = "frLocalAipReadNdbUpdatePktUsingAipData";
  char *payload = NULL;
  size_t psz;
  size_t maxsz;

  (void) udpBuf;                /* Not used */

  /* aip->data.packet.handler should have been set in the LinkDecrypt function */
  if (!aip) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, 
               "FrLocalAipCtx object is NULL");
    return 0;
  }

  if (aci > aip->route.context->maxRoutes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD,
                  "Route aci [%d] is out of range [%d]", aci, 
                  aip->route.context->maxRoutes);
    return 0;
  }

  if (!aip->route.context->route[aci].valid) {
    /* Not a valid route */
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD, 
                  "Route for ACI %d is not valid", aci);
    return 0;
  }
 
  if( !frPktAipDataPayloadMaxSize(err, aip->data.protocol.version, &maxsz) ) {
    zkErrorSet(err, func, ZKWARN, ZKMFRLAIP, ZKEPARMBAD, 
               "frPktAipDataPayloadMaxSize failed");
    return 0;
  }

  /* Bounds check */
  if( udpSize >  maxsz ) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                  "Size of incoming packet [%d] is incorrect, should be less than [%d]",
                  udpSize, maxsz);
    return 0;
  }

  /* Turn MAC usage off */
  if (!frPktAipDataSetUseMac(err, aipData, 0)) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, 
               "Unable to turn off MAC key usage for payload");
    return 0;
  }

  if (!frPktAipDataGetPayload(err, aipData, NULL, (void **) &payload, &psz)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, 
               "frPktAipDataSetPayload() failed");
    return 0;
  }

  /* Turn MAC usage on */
  if (!frPktAipDataSetUseMac(err, aipData, 1)) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, 
               "Unable to turn on MAC key usage for payload");
    return 0;
  }

  if (psz > *paySz) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Payload buffer size [%d] is too small", *paySz);
    return 0;
  }

  *paySz = psz;
  memcpy(payBuf, payload, psz);

  return 1;
}

