/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <frLocalAip.h>
#include "localAipStructs.h"
#include "localAipRouterTable.h"

#define MINROUTERS 2

/**
 * Create a new RouterInfo object.  Initialize elements to 0.
 *
 * @param err
 * @param r                     RouterInfo object
 * @param max                   Number of routers in route
 */
static int routerNew(ZkError err, RouterInfo *r, size_t max)
{
  size_t i;
  const char *func = "routerNew";

  if (r) {
    *r = (RouterInfo) malloc(max * sizeof(struct T_RouterInfo));
    for (i = 0; i < max; i++) {
      (*r)[i].host.sin_addr.s_addr = 0;
      (*r)[i].host.sin_port = 0;
      (*r)[i].forwSymKey = NULL;
      (*r)[i].backSymKey = NULL;
      memset(&((*r)[i].aipEntId), 0, sizeof((*r)[i].aipEntId));
      (*r)[i].aipId = 0;
    }
    return 1;
  }
  zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "RouterInfo is NULL");
  return 0;
}

/**
 * Create a new RouterTableEntry object.  A static array is allocated for 
 * router table entries.  Each entry contains a maximum number of routers.
 * 
 * @param err
 * @param e                     RouterTableEntry object
 * @param tsz                   Router Table size
 * @param max                   Maximum number of routers
 */
static int routeNew(ZkError err, RouterTableEntry *e, size_t tsz, size_t max)
{
  size_t i;

  if (e) {
    *e = (RouterTableEntry) malloc(sizeof(struct T_RouterTableEntry) * tsz);
    if (*e) {
      memset(*e, 0, tsz * sizeof(struct T_RouterTableEntry));
      for (i = 0; i < tsz; i++) {
        (*e)[i].aciIn = 0;
        (*e)[i].macKey = NULL;

        (*e)[i].na.nymID.type = FRENT_TYPE_UNUSED_;
        (*e)[i].na.nymSig = NULL;

        (*e)[i].current = FRAIPS_SYMKEY_INITIAL;
        (*e)[i].symKey[0] = NULL;
        (*e)[i].symKey[1] = NULL;
        (*e)[i].dh.dh = NULL;
        (*e)[i].dh.send = -1;
        (*e)[i].expiry = 0;
        (*e)[i].valid = 0;
        (*e)[i].aciOut = 0;
        (*e)[i].prio = 0;
        
        (*e)[i].numRouters = 0;        
        (*e)[i].maxRouters = max;

        (*e)[i].fhSigKey = NULL;
        if (!routerNew(err, &((*e)[i].router), max)) {
          return 0;             /* bug */
        }
      }
      return 1;
    }
  }
  return 0;
}

/**
 * Create a FirstHops structure.  Initialize everything to be 0.
 *
 * @param err
 * @param fh                    FirstHops Object
 * @param max                   Maximum size of the router table-->first hops
 * @return                      1 on success, 0 on failure
 */
static int fhNew(ZkError err, FirstHops *fh, size_t max)
{
  const char *func = "fhNew";
  
  if (!fh) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "FirstHops is NULL");
    return 0;
  }

  *fh = (FirstHops) malloc(max * sizeof(struct T_FirstHops));
  if (*fh == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENOMEM, "malloc() failed");
    return 0;
  }
  memset(*fh, 0, max * sizeof(struct T_FirstHops));
  
  return 1;
}

int rtCtxNew(ZkError err, RouterTable *rt, size_t tsz, size_t maxRouters)
{
  const char *func = "rtCtxNew";

  if (maxRouters < MINROUTERS) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
               "Value of maxRouters is too small");
    return 0;
  }
  if (rt) {
    *rt = (RouterTable) malloc(sizeof(struct T_RouterTable));
    if (*rt) {
      (*rt)->curRoute = 0;
      (*rt)->maxRoutes = tsz;

      if (!fhNew(err, &(*rt)->fh, tsz)) {
        return 0;
      }
      return(routeNew(err, &((*rt)->route), tsz, maxRouters));
    }
  }
  return 0;
}

static int routerFree(ZkError err, RouterInfo *r, size_t max)
{
  size_t i;

  if (r) {
    if (*r) {
      for (i = 0; i < max; i++) {
        (*r)[i].host.sin_addr.s_addr = 0;
        (*r)[i].host.sin_port = 0;
        if ((*r)[i].forwSymKey) {
          frCryptSymFree(err, &((*r)[i].forwSymKey));
        }
        if ((*r)[i].backSymKey) {
          frCryptSymFree(err, &((*r)[i].backSymKey));
        }
      }
      free(*r);
      *r = NULL;
      return 1;
    }
  }
  return 0;
}

static int fhFree(ZkError err, FirstHops *fh, size_t max)
{
  const char *func = "fhFree";
  size_t i;
  
  if (!fh || !*fh) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "FirstHops is NULL");
    return 0;
  }

  for (i = 0; i < max; i++) {
    if ((*fh)[i].symKey[0]) {
      frCryptSymFree(err, &((*fh)[i].symKey[0]));
    }

    if ((*fh)[i].symKey[1]) {
      frCryptSymFree(err, &((*fh)[i].symKey[1]));
    }
    if ((*fh)[i].aciCtx) {
      (void) zkACIContextFree(err, &((*fh)[i].aciCtx));
    }
  }

  free(*fh);
  *fh = NULL;
  return 1;
}

static int routeFree(ZkError err, RouterTableEntry *e, size_t tsz)
{
  size_t i;

  if (e) {
    if (*e) {
      for (i = 0; i < tsz; i++) {
        if ((*e)[i].macKey) {
          frCryptMacKeyFree(err, &((*e)[i].macKey));
        }
        if ((*e)[i].dh.dh) {
          frPktDifHelFree(err, &((*e)[i].dh.dh));
        }
        (*e)[i].dh.send = -1;
        (*e)[i].symKey[0] = NULL;
        (*e)[i].symKey[1] = NULL;
        
        (*e)[i].expiry = 0;
        (*e)[i].aciIn = 0;
        (*e)[i].aciOut = 0;

        (*e)[i].na.nymID.type = FRENT_TYPE_UNUSED_;
        if ((*e)[i].na.nymSig) {
          frCryptSigFree(err, &((*e)[i].na.nymSig));
        }

        (*e)[i].numRouters = 0;
        if ((*e)[i].router) {
          routerFree(err, &((*e)[i].router), (*e)[i].maxRouters);
        }
        (*e)[i].maxRouters = 0;
        (*e)[i].valid = 0;
      }
      free(*e);
      *e = NULL;
      return 1;
    }

  }
  return 0;
}

int rtCtxFree(ZkError err, RouterTable *rt)
{
  if (rt) {
    if (*rt) {
      (*rt)->curRoute = 0;
      if ((*rt)->route) {
        routeFree(err, &((*rt)->route), (*rt)->maxRoutes);
      }
      if ((*rt)->fh) {
        fhFree(err, &((*rt)->fh), (*rt)->maxRoutes);
      }

      (*rt)->maxRoutes = 0;
      free(*rt);
      *rt = NULL;
      return 1;
    }
  }

  return 0;
}
