/* ==========================================================================
 *
 * $Id: frMailAccount.c,v 1.3.10.3 2000/12/14 17:52:51 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *
 * This file defines functions which facilitate the creation of the entries
 * which are used in the mail system block-list.
 *
 * Copyright (c) 2000 Zero-Knowledge Systems.  All rights reserved.
 */

# include "frMailAccount.h"
# include "frMailAccount_utils.h"
# include <ctype.h>
# include <stdio.h>

# define MAX_ACCOUNT_NAME_LENGTH (64)

/* --------------------------------------------------------------------------
 *
 * This function calculates a 24-bit hash of the given name.  It uses a symbol
 * table hash function iterated over the name 3 times.
 */
static uint32_t calc_hash( const char * const s )
{
    const char * p = NULL;
    uint32_t     h = 0;
    uint32_t     g = 0;
    int          i = 3; /* this number is pretty arbitrary; but don't change it! */

    /*
     * Iteratively put data into the hash
     */

    while (i--)
    {
        /*
         * Algorithm taken from "Compilers: Principles, Techniques, and Tools"
         * by Aho, Sethi, and Ullman page 436.
         */

        for (p = s; *p != '\0'; ++p)
        {
            h = ((h << 4) + (*p));

            if (0 != (g = h & 0xF0000000))
            {
                h = h ^ (g >> 24);
                h = h ^ g;
            }
        }
    }

    /*
     * The original algorithm returns the hash mod a prime, but we're working
     * in a 20 bit space so lets just return the bits we want to use.  In this
     * case, we want the middle 20 bits.
     *
     * 15-08-2000   Roger - Changed to 20 bits (from 24)
     */

    return ((h >> 6) & 0x000FFFFF);
}


/* --------------------------------------------------------------------------
 *
 * Copy at most 'len' character from the 'src' string to the 'dest' buffer,
 * changing all letters to lowercase as it goes along.
 *
 * Notes:
 *
 *    1. This function will ensure that the dest buffer is NUL terminated.
 *
 *    2. 'dest' is assumed to point to a buffer with sufficient space for
 *       at least 'len'+1 characters.
 *
 */
static const char *
strncpy_tolower(
    const char * src,
    char *       dest,
    size_t       len )
{
    char * result_ptr = NULL;

    if (src && dest && len)
    {
        result_ptr = dest;

        while ( (--len > 0)
                && (0 != (*dest++ = ((char)tolower(*src++)))))
        {
            /* do nothing */
        }
        
        *dest = (char)0;
    }

    return (result_ptr);
}


/* --------------------------------------------------------------------------
 *
 * Given a root directory and an account name, calculate the path to the
 * account.  The path to the account is defined by the middle 24 bits of
 * a the hash returned by calc_hash(name) as following:
 *
 *    [/]root/i/j/k/l/m/name
 *
 * where:
 *
 *    h1 is bits 16-19 of the 24 bit hash in hexadecimal
 *    h2 is bits 12-15 of the 24 bit hash in hexadecimal
 *    h3 is bits  8-11 of the 24 bit hash in hexadecimal
 *    h4 is bits  8-11 of the 24 bit hash in hexadecimal
 *    h5 is bits  8-11 of the 24 bit hash in hexadecimal
 *
 * The hexadecimal representations use lowercase letters.
 */
int
frMailAccount_GetPath(
    ZK_INOUT ZkError            error,
    ZK_IN    const char * const root,
    ZK_IN    const char * const name,
    ZK_OUT   char *             buffer,
    ZK_IN    size_t             buffer_size )
{
    const char * const func       = "frMailAccount_GetPath";
    uint32_t           hash       = 0;
    size_t             min_size   = 0;
    size_t             name_size  = 0;
    size_t             root_size  = 0;
    char *             safe_name  = NULL;
    int                result     = ZK_FAILURE_CODE;

    ZK_CHECK_CONSTRAINT( root   != NULL );
    ZK_CHECK_CONSTRAINT( *root  == '/' );
    ZK_CHECK_CONSTRAINT( name   != NULL );
    ZK_CHECK_CONSTRAINT( buffer != NULL );

    name_size = strlen(name);
    root_size = strlen(root);

    /*
     * Minimum size is the length of the root plus the length of
     * the hash directories [ (5 * (1 chars + 1 slash)) = 10 ]
     * plus the length of the name, plus the trailing NUL. 
     */

    min_size  = ( root_size + 10 + name_size + 1 );

    /*
     * We might also have to add a slash after the root
     */

    if (root[root_size -1] != '/')
    {
        ++min_size;
    }

    ZK_CHECK_CONSTRAINT( buffer_size > min_size );

    safe_name = (char *) alloca(name_size + 1);

    ZK_CHECK_CONSTRAINT( safe_name != NULL );
    
    ZK_ATTEMPT(frMailAccount_NormalizeName(error, name, safe_name, name_size + 1));

    hash = calc_hash(safe_name);

    snprintf( buffer, buffer_size,
              "%s%s%01x/%01x/%01x/%01x/%01x/%s",
              root,
              ((root[root_size-1] != '/') ? "/" : ""),
              ((hash >> 16) & 0xF),
              ((hash >> 12) & 0xF),
              ((hash >>  8) & 0xF),
              ((hash >>  4) & 0xF),
              (hash & 0xF),
              safe_name );

    buffer[buffer_size-1] = '\0';

    result = ZK_SUCCESS_CODE;

ZK_ATTEMPT_FAILED:

    return (result);
}


/* --------------------------------------------------------------------------
 *
 * Makes sure that a prospective account name only contains the legal number
 * and type of characters.
 */
int
frMailAccount_NameIsValid(
    ZK_INOUT  ZkError       error,
    ZK_IN     const char *  name,
    ZK_OUT    ZkBool *      is_valid )
{
    const char * const func   = "frMailAccount_NameIsValid";
    size_t             length = 0;
    int                result = ZK_FAILURE_CODE;

    ZK_CHECK_CONSTRAINT( name     != NULL );
    ZK_CHECK_CONSTRAINT( is_valid != NULL );

    while ( (*name != '\0')
            && ( isalnum(*name)  /* letters and numbers */
                 || (*name == '-')  /* hyphen              */
                 || (*name == '_')  /* underscore          */
                 || (*name == '.')  /* periods             */
                ) )
    {
        ++name;
        ++length;
    }

    /*
     * If we made it to the end of the name then the name must
     * have had only valid characters.
     */

    (*is_valid) = (*name == '\0') && (length < MAX_ACCOUNT_NAME_LENGTH);
    result      = ZK_SUCCESS_CODE;

ZK_ATTEMPT_FAILED:

    return (result);
}


/*
 * 
 */

int
frMailAccount_NormalizeName(
    ZK_INOUT  ZkError       error,
    ZK_IN     const char *  name,
    ZK_OUT    char *        tgt_buffer,
    ZK_IN     size_t        tgt_buffer_size )
{
    const char * const func          = "frMailAccount_NormalizeName";
    int                result        = ZK_FAILURE_CODE;
    ZkBool             name_is_valid = ZK_FALSE;

    ZK_CHECK_CONSTRAINT( name       != NULL );
    ZK_CHECK_CONSTRAINT( tgt_buffer != NULL );

    ZK_ATTEMPT( frMailAccount_NameIsValid(error, name, &name_is_valid) );

    ZK_CHECK_CONSTRAINT( name_is_valid );

    strncpy_tolower( name, tgt_buffer, tgt_buffer_size );

    result = ZK_SUCCESS_CODE;

ZK_ATTEMPT_FAILED:

    return (result);
}
