// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom(R) code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <pwd.h>
#include <errno.h>
#include <unistd.h>
#include <grp.h>

#include <sys/types.h> // for umask and getpwnam
#include <sys/stat.h> // for umask
#include <sys/wait.h> // for waitpid

#include <string>
#include <iostream>
#include <fstream>

#include "auth-common.hh"
#include "entry.hh"
#include "zkCStr.h"

using namespace std;

static int changeUserAndGroup(const uid_t uid, const gid_t gid) 
{
  // It is not possible for normal users to change their group, only root
  // can do that so we have to go back to root each time we want to change
  // group.
  if (seteuid(0) == -1) return -1;
  if (setegid(gid) == -1) return -1;
  return seteuid(uid);
}  

// the purpose of this program is to ensure that, on its completion, the user
// named on its command line has a mail account. as such it defines policy for
// what goes in a user's home directory in the mail system, and where that home
// directory is. it must be run as root, and must have access to the config
// files mailroot, tmpmailroot, and mailowner (all presumably in
// /var/qmail/control)
//

static void
createUserLimitFile(
    const string &      user_dir,
    const string &      setting,
    const string &      env_variable,
    const string &      suffix = "" )
{
    const char *       value_str = getenv(env_variable.c_str()); /* audit: ok */
    char *             end_ptr   = NULL;
    unsigned long int  value     = 0;

    if (value_str == NULL)
    {
        die(string(setting + " value is not configured").c_str());
    }

    value = strtoul( value_str, &end_ptr, 10);

    if (*end_ptr != '\0')
    {
        die(string(setting + " is not valid!").c_str());
    }
    
    const string filename(user_dir + '.' + setting);
    ofstream file(filename.c_str());

    if (!file)
    {
        die(string(setting + " cannot be initialized!").c_str());
    }

    file << value << suffix;

    file.close();

    if (!Entry::exists(filename))
    {
        die(string(setting + " cannot be saved!").c_str());
    }
}

static void
createUserLimits( const string & user_dir)
{
    struct Field {
        const char * name;
        const char * env_var;
        const char * suffix;
    };

    static const struct Field field_table[] = {
        { FRMAILPARM_NYM_TYPE,         FRMAILENV_NYM_TYPE,         "" },
        { FRMAILPARM_SEND_LIMIT,       FRMAILENV_SEND_LIMIT,       "" },
        { FRMAILPARM_CROSSPOST_LIMIT,  FRMAILENV_CROSSPOST_LIMIT,  "" },
        { FRMAILPARM_STORAGE_QUOTA,    FRMAILENV_STORAGE_QUOTA,    "k"},
        { FRMAILPARM_VOLUME_LIMIT,     FRMAILENV_VOLUME_LIMIT,     "k"},
        { 0, 0, 0 } // terminator
    };

    for (const struct Field *f = field_table; f->name; ++f)
    {
        createUserLimitFile(user_dir, f->name, f->env_var, f->suffix);
    }
}

int mailWelcomeMessage(string &nymstr) {
  string cmdstr(string(qmailBinDirDefault) + "/send-welcome-message.sh");
  char cmd[cmdstr.size() + 1];
  char nym[nymstr.size() + 1];
  strncpy(cmd,cmdstr.c_str(),cmdstr.size() + 1);
  strncpy(nym,nymstr.c_str(),nymstr.size() + 1);
  
  int pid = fork();
  if (pid == -1)
    return -1;
  if (pid == 0) {
    char *argv[3];
    argv[0] = cmd;
    argv[1] = nym;
    argv[2] = 0;
    execve(cmd, argv, NULL);
    exit(127);
  }
  while(1) {
    int status;
    if (waitpid(pid, &status, 0) == -1) {      
      if (errno != EINTR)
	return -1;
    } else {
      if (WIFEXITED(status)) {
	return WEXITSTATUS(status);
      } else {
	return -1;
      }
    }
  }
}

int main(int argc, char **argv)
{
    const string progname = "ensure-user";
    ZkError err_; 
    zkErrorNew (&err_);
    uid_t owner_uid, qmailr_uid;
    gid_t owner_gid, qmail_gid;

    if (argc != 2) {
        die(string(progname + " usage: " + progname + " <nymname>").c_str());
    }

    char* qmailCtlDir = (char*) malloc(sizeof(char) * (strlen(qmailCtlDirDefault)+1));
    zkCStrlcpy(qmailCtlDir, qmailCtlDirDefault, strlen(qmailCtlDirDefault)+1);
    
    /* Do not rely on umask but instead use proper permissions 
       when creating directories. */
    umask(000);

    string mailrootConfigFile( string(qmailCtlDir) +  "/mailroot");
    string mailownerConfigFile( string(qmailCtlDir) + "/mailowner");

    string mailroot = slurpConfigFile(mailrootConfigFile.c_str());
    string mailowner = slurpConfigFile(mailownerConfigFile.c_str());

    struct passwd *pw = getpwnam(mailowner.c_str());
    if (!pw) {
        die(string(progname + ": couldn't get mail owner passwd info").c_str());
    }                   

    /* Need to save the passwd info otherwise it gets whiped out on
       the next call to getpwnam... */
    owner_uid = pw->pw_uid;
    owner_gid = pw->pw_gid;

    if (!(pw = getpwnam("qmailr"))) {
        die(string(progname + ": couldn't get qmailr passwd info").c_str());
    }                   
    qmailr_uid = pw->pw_uid;
    qmail_gid = pw->pw_gid;

    string nym(*++argv);
    string nympath = calculateUserDir(err_,mailroot,nym);
  
    int limit = 100;
    int i;
    for (i = 0; i < limit && ! Entry::exists(nympath); ++i) {

        if (changeUserAndGroup(owner_uid, owner_gid) == -1) {
            die(string(progname + ": couldn't seteuid or setegid" ).c_str());
        }

        for (unsigned int j = 0; j < nympath.size(); ++j) {
            if ( (j != 0) && (nympath[j] == '/') ) {
               if ( !Entry::exists(nympath.substr(0,j).c_str()) ) {
                   int rc = mkdir(nympath.substr(0,j).c_str(),0711);
                   if (rc != 0) {
                      die(string(progname + ": couldn't create directory " + nympath.substr(0,j).c_str() +
                                            " : " + strerror(errno)).c_str() );
                   }
               }
            }
        }

        /* Note: qmail down-cases all email addresses, so this is never
           going to be anyone's email address. */

        string tmproot = nympath.substr(0,nympath.rfind('/')) + "/.TMP";
        mkdir(tmproot.c_str(), 0711);

        // LOG AUDIT: The creation date of a directory can be correlated
        // LOG AUDIT:  with the nym's creation date. We must randomize or
        // LOG AUDIT:  standardize the date. Here or on the shell?
       
        string tmp(guid(nym));

        /* NOTE: We have to give group write permission so that the Stat
           directory can be created by someone else than the mail owner.
           We will change the group permission afterward. */
        Dir *tree = (new Dir(tmp.c_str(), 0770)) 
            ->add( (new Dir("Maildir",0700))
                   ->add(new Dir("tmp"))
                   ->add(new Dir("new"))
                   ->add(new Dir("cur")));
        tree->createIn(tmproot);
        if (!tree->move(tmproot, nympath))
        {
            tree->eraseFrom(tmproot);
        }
        else
        {
            if (nympath.at(nympath.size() -1) != '/')
            {
                nympath += '/';
            }

            /* The limit files should be readable by all (for qmailr when
               checking the limits). */
            umask(033);
            createUserLimits(nympath);

            /* To allow both the qmailr and the mail owner to update the
               counter file, the Stat directory is created with the following
               user/group and permissions:
               
               drwxrwx---   qmailr   <mailOwnerGroup>
               
               The only potentiel problem is that all members of <mailOwnerGroup>
               can also write in the direcrtory but we have control over this.
            */
            string mailStatDir(nympath + "Stat");
            if (changeUserAndGroup(qmailr_uid, owner_gid) == -1) {
                die(string(progname + ": couldn't change user and group for Stat" ).c_str());
            }
            umask(000);
            if (mkdir(mailStatDir.c_str(), 0770) == -1) {
                die(string(progname 
                           + ": couldn't create stat directory: " 
                           + strerror(errno)).c_str());
            }
            
            /* We need to go back to the mail owner uid/gid so that we
               can remove the group write permission (and thus only allow
               the mail owner to write in the directory. */
            if (changeUserAndGroup(owner_uid, owner_gid) == -1) {
                die(string(progname + ": couldn't change user and group for Stat" ).c_str());
            }
            
            if (chmod(nympath.c_str(), 0711)) {
                die(string(progname
                           + ": couldn't set proper stat directory permissions: " 
                           + strerror(errno)).c_str());
            }

	    /* now create the tagline file with default on */
            umask(036);
            if (changeUserAndGroup(owner_uid, qmail_gid) == -1) {
                die(string(progname + ": couldn't change user and group for tagline" ).c_str());
            }
            
	    const string filename(nympath + "/.tagline"  );
            ofstream file(filename.c_str());

            if (!file)
            {
              die(string("tagline cannot be initialized!").c_str());
            }

            file << "on";

            file.close();

            if (!Entry::exists(filename))
            {
              die(string("tagline cannot be saved!").c_str());
            }

/*            if (chmod(filename.c_str(), 610)) {
                die(string(progname
                           + ": couldn't set proper stat directory permissions: " 
                           + strerror(errno)).c_str());
            }
*/

	    /* change the uid back to root, this fixes the send welcome message to not have permission denied, /dev/fd/4  */
	    if (changeUserAndGroup(0,0) == -1) {
                die(string(progname + ": couldn't change user and group for Stat" ).c_str());
            }
	    
	    /* mail welcome message */
	    if (mailWelcomeMessage(nym)) {
	      die(string(progname
                           + ": couldn't send welcome message").c_str());
	    }
        }   
    
        delete tree;
    }

    if (i >= limit) {
        die(string(progname + ": max iterations exhausted").c_str());
    }

    return 0;
}

