/*
 *  $Id: frMailAuth.c,v 1.6.4.7 2000/12/14 17:52:53 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *                  Cacophonix's Mail Authentication library 
 * 
 *  
 */

#include <errno.h>
#include <string.h>
#include "frMailAuth.h"
#include "zknbo.h"

/* Fixed for the protocol. Allowed value are between 10 and 20 */
#define MAC_LENGTH 10

/* Size of packed mail cert (in byte) w/o the nym name. Should change if struct fr_mail_auth change.
 * It's 2 more (for packet version and packet type) then the size of constant lenght cert fields:
 * 39 = 2 (pkt vers & type) + 2 (vers) + 8 (exp) + 1 (type) + 4 (mail) + 4 (vol) + 4 (size) + 4 (cross) + 10 (mac);
 */
#define CTE_CERT_SIZE 39

/* packet format version */
#define PKT_VERS_1 1
#define SYM_MAIL_CERT 1


/*
 *   The mail certificate object used for authentication.
 *   Any modifications to the struct should be reflected on CTE_CERT_SIZE.
 */
struct fr_mail_auth {
  char *ent;      /* ent name, [ type || name ],  _not_ NULL terminated */
  size_t entsize; /* size of the ent ( = length of name + 1)*/
  uint16_t vers;  /* certificate version */
  ZkClock exp;    /* certificate expiration date */
  uint8_t type;   /* nym type: free, pro, trial, ... */
  uint32_t mail;  /* limit on the number of mail sent */
  uint32_t vol;   /* limit on the volume of mail sent (in bytes) */
  uint32_t size;  /* size limit of mail box (in bytes) */
  uint32_t cross; /* the cross-post limit */
  char mac[MAC_LENGTH];
};


/*
 *
 *
 *                            Internal routines.
 *
 *
 */


/*
 *   Initialize the certificate.
 */
static void mailCertInit(FrMailAuth cert)
{
  memset(cert, 0, sizeof *cert);
  cert->ent = 0;
}

/*
 *   Convert the mail certificate 'cert' object to a buffer 'buf'.
 *   It's assumed that the arguments are ok, since the fct is internal.
 *   'buf' will be packed in network byte order.
 */ 
static int packCertToBuf(FrMailAuth cert, char *buf)
{
  int64_t e;

  int offset = 0;

  buf[offset++] = PKT_VERS_1;      /* the packet version */
  buf[offset++] = SYM_MAIL_CERT;   /* the packet type */

  memcpy(&buf[offset], cert->ent, cert->entsize);  
  offset += cert->entsize;
  
  zkNboFromHostInt16(&buf[offset], cert->vers);
  offset += sizeof cert->vers;
  
  e = zkClockToSeconds(cert->exp);
  
  zkNboFromHostInt64(&buf[offset], e);
  offset += sizeof e;

  buf[offset] = cert->type;
  offset++;

  zkNboFromHostInt32(&buf[offset], cert->mail);
  offset += sizeof cert->mail;

  zkNboFromHostInt32(&buf[offset], cert->vol);
  offset += sizeof cert->vol;

  zkNboFromHostInt32(&buf[offset], cert->size);
  offset += sizeof cert->size;

  zkNboFromHostInt32(&buf[offset], cert->cross);
  offset += sizeof cert->cross;
  
  memcpy(&buf[offset], cert->mac, MAC_LENGTH);
  
  return 1;
}


/*
 *
 *
 *                            External routines.
 *
 *
 */


/*
 *   Return the size of the mac (constant).
 */
DLLExport size_t frMailAuthMacLength(void)
{
  return MAC_LENGTH;
}


/*
 *   Return the size of the cert, in byte, w/o the ent name (constant).
 */
DLLExport extern size_t frMailAuthCteSize(void)
{
  return CTE_CERT_SIZE;
}


/*
 *   Return the size of the cert, in byte, w/ the ent name.
 */
DLLExport extern int frMailAuthSize(ZkError err, FrMailAuth cert, size_t *csz)
{
  const char *fn = "frMailAuthSize";
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }

  if ( ! csz) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "csz is null");
#endif
    return 0;
  }
  
  *csz = CTE_CERT_SIZE + cert->entsize;
  
  return 1;
}


/*
 *   Create a new mail certificate object.
 */
DLLExport extern int frMailAuthNew(ZkError err, FrMailAuth *cert)
{
  const char *fn = "frMailAuthNew";
  FrMailAuth c = 0;
  int ret = 0;
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    goto done;
  }
    
  if ( ! (c = malloc(sizeof *c))) {
#ifndef NDEBUG
    zkErrorSet_va(err, fn, ZKERR, ZKMALLOC, ZKENOMEM,
		  "can't allocate %lu bytes for mail cert: %s",
		  (unsigned long)sizeof *c, strerror(errno));    
#endif
    goto done;
  }
  
  mailCertInit(c);
  *cert = c; 
  c = 0;
  
  ret = 1;

 done:
  
  if (c) { free(c); }
  
  return ret;
}


/*   
 *   Create a new mail certificate from a buffer.
 */
DLLExport extern int frMailAuthNewFromBuf(ZkError err, FrMailAuth *cert, const void *buf, size_t bsz)
{
  const char *fn = "frMailAuthNewFromBuf";
  FrMailAuth c = 0;
  int ret = 0;
  const char *b = 0;
  
  size_t offset = 0; 
  char *ent = 0;
  size_t esz = 0;
  
  int64_t exp = 0;
  
  if ( ! cert) {

#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }
  
  if ( ! buf) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "buf is null");
#endif
    return 0;
  }
  
  if (bsz <= CTE_CERT_SIZE || (bsz > FRENT_PORTABLE_SIZE + CTE_CERT_SIZE)) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, 
		  "buf cannot be a mail certificate");
#endif
    return 0;
  }
  
  if ( ! (c = malloc(sizeof *c))) {
#ifndef NDEBUG
    zkErrorSet_va(err, fn, ZKERR, ZKMALLOC, ZKENOMEM,
		  "can't allocate %lu bytes for mail cert: %s",
		  (unsigned long)sizeof *c, strerror(errno));    
#endif
    goto done;
  }
  
  mailCertInit(c);
  
  /* 
   * unpack buf (from network byte order)
   */
  
  esz = bsz - CTE_CERT_SIZE; /* size of ent name */
  
  /* contains the ent name of size esz */
  if ( ! (ent = malloc(esz))) {
#ifndef NDEBUG
    zkErrorSet_va(err, fn, ZKERR, ZKMALLOC, ZKENOMEM,
		  "can't allocate %lu bytes for ent name: %s",
		  (unsigned long)(esz), strerror(errno));    
#endif
    goto done;
  }
  
  b = (const char *)buf;

  /* No need to check the version number now, only in further version */
  offset += 2; /* pass the version number and packet type */

  memcpy(ent, &b[offset], esz);
  c->ent = ent;
  c->entsize = esz;
  offset += esz;
  ent = 0;
  
  c->vers = zkNboToHostInt16(&b[offset]);
  offset += sizeof c->vers;
  
  exp = zkNboToHostInt64(&b[offset]);
  offset += sizeof exp;
  
  c->exp = zkClockFromSeconds(exp);

  c->type = b[offset];
  offset++;
  
  c->mail = zkNboToHostInt32(&b[offset]);
  offset += sizeof c->mail;
   
  c->vol = zkNboToHostInt32(&b[offset]);
  offset += sizeof c->vol;
   
  c->size = zkNboToHostInt32(&b[offset]);
  offset += sizeof c->size;

  c->cross = zkNboToHostInt32(&b[offset]);
  offset += sizeof c->cross;

  memcpy(c->mac, &b[offset], MAC_LENGTH);
  
  ret = 1;
  
  *cert = c; 
  c = 0;
  
 done:
  
  if (ent) { free(ent); }
  if (c) { free(c); }
  
  return ret;  
}


/*
 *   Free a mail certificate.
 */
DLLExport extern int frMailAuthFree(ZkError err, FrMailAuth *cert)
{
  const char *fn = "frMailAuthFree";  
  FrMailAuth c = 0;
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }
  
  if ( ! *cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "*cert is null");

#endif
    return 0;
  }

  c = *cert;
  
  if (c->ent) {
    free(c->ent);
  }
  
  mailCertInit(c);
  
  free(c);
  
  *cert = 0;

  return 1;
}


/* 
 *   Get the owner of the certificate in 'ent'.
 */
DLLExport extern int frMailAuthGetEnt(ZkError err, FrMailAuth cert, FrEntIdExp *ent)
{
  const char *fn = "frMailAuthGetEnt";
  char buf[FRENT_PORTABLE_SIZE];
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  } 
  
  if ( ! ent) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "ent is null");
#endif
    return 0;
  }
  
  if ( ! cert->entsize) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKEPARMBAD, "ent is not set");
#endif
    return 0;
  }
  
  if ( cert->entsize > FRENT_PORTABLE_SIZE) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKEPARMBAD, "entsize is to big");
#endif
    return 0;
  }

  memset(buf,0,sizeof buf);
  memcpy(buf, cert->ent, cert->entsize);
  frEntLoad(err, ent, buf, sizeof buf); /* load buf in ent */
  
  return 1;
}


/*
 *   Set the owner of the certificate to be the entity 'ent'.
 */
DLLExport extern int frMailAuthSetEnt(ZkError err, FrMailAuth cert, FrEntIdExp *ent)
{
  const char *fn = "frMailAuthSetEnt";
  char buf[FRENT_PORTABLE_SIZE + 1]; /* +1 to use strlen */
  char *e = 0;
  size_t esz = 0;
  int ret = 0;
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  } 
  
  if ( ! ent) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "ent is null");
#endif
    return 0;
  }
  
  memset(buf,0,sizeof buf);
  frEntStore(err, ent, buf, sizeof buf); /* store ent in buf */
  
  buf[FRENT_PORTABLE_SIZE] = 0; /* should be 0 already */
  esz = strlen(&buf[1]) + 1; /* in case buf[0] == 0 */
  
  if (cert->ent) { free(cert->ent);}
  
  if ( ! (e = malloc(esz))) {
#ifndef NDEBUG
    zkErrorSet_va(err, fn, ZKERR, ZKMALLOC, ZKENOMEM,
		  "can't allocate %lu bytes for ent name: %s",
		  (unsigned long)(esz), strerror(errno));    
#endif
    goto done;
  }
  
  memcpy(e, buf, esz);
  cert->ent = e;
  cert->entsize = esz;
  e = 0;

  ret = 1;
  
 done:
  
  if (e) { free(e); }
  return ret;
}


/*
 *   Return the size of entity (name length + 1) of the owner of the certificate.
 */
DLLExport extern int frMailAuthGetEntSize(ZkError err, FrMailAuth cert, size_t *esz)
{
  const char *fn = "frMailAuthGetEntSize";
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }
  
  if ( ! esz) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "esz is null");
#endif
    return 0;
  }

  if ( ! cert->entsize) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKEPARMBAD, "ent is not set");
#endif
    return 0;
  }
  
  *esz = cert->entsize;
  
  return 1;
}


/*
 *   Get the version of the mail certificate.
 */
DLLExport extern int frMailAuthGetVersion(ZkError err, FrMailAuth cert, uint16_t *vers)
{
  const char *fn = "frMailAuthGetVersion";
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }

  if ( ! vers) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "vers is null");
#endif
    return 0;
  }

  *vers = cert->vers;
  
  return 1;
}


/*
 *   Set the version of the mail certificate.
 */
DLLExport extern int frMailAuthSetVersion(ZkError err, FrMailAuth cert, uint16_t vers)
{
  const char *fn = "frMailAuthSetVersion";

  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }

  cert->vers = vers;
  
  return 1;
}


/*
 *   Get the expiration date of the mail certificate.
 */
DLLExport extern int frMailAuthGetExpiration(ZkError err, FrMailAuth cert, ZkClock *exp)
{
  const char *fn = "frMailAuthGetExpiration";
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }

  if ( ! exp) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "exp is null");
#endif
    return 0;
  }
 
  *exp = cert->exp;
  
  return 1;
}


/*
 *   Set the expiration date of the mail certificate.
 */
DLLExport extern int frMailAuthSetExpiration(ZkError err, FrMailAuth cert, ZkClock exp)
{
  const char *fn = "frMailAuthSetExpiration";

  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }

  cert->exp = exp;
  
  return 1;
}


/*
 *   Get the type of the owner of the mail certificate.
 */
DLLExport extern int frMailAuthGetNymType(ZkError err, FrMailAuth cert , uint8_t *type)
{
  const char *fn = "frMailAuthGetNymType";
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }
  
  *type = cert->type;
  
  return 1;
}


/*
 *   Set the type of the owner of the mail certificate.
 */
DLLExport extern int frMailAuthSetNymType(ZkError err, FrMailAuth cert, uint8_t type)
{
  const char *fn = "frMailAuthSetNymType";
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }

  cert->type = type;
  
  return 1;
}


/*
 *   Get the mail limit that can be sent by the owner.
 */
DLLExport extern int frMailAuthGetMailLimit(ZkError err, FrMailAuth cert, uint32_t *limit)
{
  const char *fn = "frMailAuthGetMailLimit";
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }
  
  if ( ! limit) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "limit is null");
#endif
    return 0;
  }
  
  *limit = cert->mail;
  
  return 1;
}


/*
 *   Set the mail limit that can be sent by the owner.
 */
DLLExport extern int frMailAuthSetMailLimit(ZkError err, FrMailAuth cert, uint32_t limit)
{
  const char *fn = "frMailAuthSetMailLimit";
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }
 
  cert->mail = limit;
  
  return 1;
}


/*
 *   Get the volume limit that can be sent by the owner.
 */
DLLExport extern int frMailAuthGetVolLimit(ZkError err, FrMailAuth cert, uint32_t *limit)
{
  const char *fn = "frMailAuthGetVolLimit";
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }
  
  if ( ! limit) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "limit is null");
#endif
    return 0;
  }
  
  *limit = cert->vol;
  
  return 1;
}


/*
 *   Set the volume limit that can be sent by the owner.
 */
DLLExport extern int frMailAuthSetVolLimit(ZkError err, FrMailAuth cert, uint32_t limit)
{
  const char *fn = "frMailAuthSetVolLimit";
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }

  cert->vol = limit;
  
  return 1;
}


/*
 *   Get the size limit of the mailbox of the owner.
 */
DLLExport extern int frMailAuthGetSizeLimit(ZkError err, FrMailAuth cert, uint32_t *limit)
{
  const char *fn = "frMailAuthGetSizeLimit";
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }
  
  if ( ! limit) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "limit is null");
#endif
    return 0;
  }
  
  *limit = cert->size;
  
  return 1;
}


/*
 *   Set the size limit of the mailbox of the owner.
 */
DLLExport extern int frMailAuthSetSizeLimit(ZkError err, FrMailAuth cert, uint32_t limit)
{
  const char *fn = "frMailAuthSetSizeLimit";
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }

  cert->size = limit;

  return 1;
}


/*
 *   Get the cross-post limit of the user.
 */
DLLExport extern int frMailAuthGetCrossLimit(ZkError err, FrMailAuth cert, uint32_t *limit)
{
  const char *fn = "frMailAuthGetCrossLimit";
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }
  
  if ( ! limit) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "limit is null");
#endif
    return 0;
  }
  
  *limit = cert->cross;
  
  return 1;
}


/*
 *   Set the cross-post limit of user.
 */
DLLExport extern int frMailAuthSetCrossLimit(ZkError err, FrMailAuth cert, uint32_t limit)
{
  const char *fn = "frMailAuthSetCrossLimit";
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }

  cert->cross = limit;

  return 1;
}


/*
 *   Transfer the mail certificate to 'buf'. 
 *   The memory for 'buf' is allocated here.
 *   It's the user responability to free it.
 */
DLLExport extern int frMailAuthToBuf(ZkError err, FrMailAuth cert, void **buf, size_t *bsz)
{
  const char *fn = "frMailAuthToBuf";
  char *b = 0;
  size_t csz = 0;
  int ret = 0;
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }

  if ( ! buf) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "buf is null");
#endif
    return 0;
  }
  
  if ( ! cert->entsize) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKEPARMBAD, "ent name is not set");
#endif
    return 0;
  }
  
  csz = cert->entsize + CTE_CERT_SIZE;
  
  if ( ! (b = malloc(csz))) {
#ifndef NDEBUG
    zkErrorSet_va(err, fn, ZKERR, ZKMALLOC, ZKENOMEM,
		  "can't allocate %lu bytes for b: %s",
		  (unsigned long)(csz), strerror(errno));    
#endif
    goto done;
  }
  
  if ( ! packCertToBuf(cert, b)) {
    goto done;
  }
  
  *buf = b;
  *bsz = csz;
  
  b = 0;
  ret = 1;
  
 done:
  
  if (b) { free(b); }
  
  return ret;
}


/*
 *   Mac a mail certificate, using the mac key 'key'.
 *   The mac will be placed in the certificate.
 */
DLLExport extern int frMailAuthMac(ZkError err, FrMailAuth cert, FrCryptMacKey key)
{
  const char *fn = "frMailAuthMac";
  char *buf = 0;
  size_t csz = 0;
  int ret = 0;
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  } 
  
  if ( ! key) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "key is null");
#endif
    return 0;
  }

  if ( ! cert->entsize) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKEPARMBAD, "ent name is not set");
#endif
    return 0;
  }
  
  csz = cert->entsize + CTE_CERT_SIZE;
  
  /* 
   * convert cert to buf 
   */
  if ( ! (buf = malloc(csz))) {
#ifndef NDEBUG
    zkErrorSet_va(err, fn, ZKERR, ZKMALLOC, ZKENOMEM,
		  "can't allocate %lu bytes for buf: %s",
		  (unsigned long)(csz), strerror(errno));    
#endif
    goto done;
  }
  
  if ( ! packCertToBuf(cert, buf)) {
    goto done;
  }
  
  /* mac 'csz - MAC_LENGTH' bytes of 'buf' and place the result ('MAC_LENGTH' bytes) in 'cert->mac' */
  if ( ! frCryptMacIntoBuffer(err, key, buf, csz - MAC_LENGTH, cert->mac, MAC_LENGTH)) {
    goto done;
  }
  
  ret = 1;
  
 done:
  
  if (buf) { free(buf); }
  
  return ret;
}


/*
 *   Verify the mail certificate (check the mac, using the key).
 *   'ok' is set to 1 if the mac is valid. If not, it's set to 0.
 */
DLLExport extern int frMailAuthCheckMac(ZkError err, FrMailAuth cert, FrCryptMacKey key, int *ok)
{
  const char *fn = "frMailAuthMac";
  char mac[MAC_LENGTH];
  char *buf = 0;
  size_t csz = 0;
  int ret = 0;
  
  *ok = 0;

  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }
  
  if ( ! key) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "key is null");
#endif
    return 0;
  }
  

  if ( ! cert->entsize) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKEPARMBAD, "ent name is not set");
#endif
    return 0;
  }
  
  csz = cert->entsize + CTE_CERT_SIZE;
  
  /* 
   * convert cert to buf 
   */
  if ( ! (buf = malloc(csz))) {
#ifndef NDEBUG
    zkErrorSet_va(err, fn, ZKERR, ZKMALLOC, ZKENOMEM,
		  "can't allocate %lu bytes for buf: %s",
		  (unsigned long)(csz), strerror(errno));    
#endif
    goto done;
  }
  
  if ( ! packCertToBuf(cert, buf)) {
    goto done;
  }
  
  if ( ! frCryptMacIntoBuffer(err, key, buf, csz - MAC_LENGTH, mac, MAC_LENGTH)) {
    goto done;
  }
  
  *ok = !memcmp(cert->mac, mac, MAC_LENGTH);
  
  ret = 1;
  
 done:
  
  if (buf) { free(buf); }
  
  return ret;
}


/*
 *   Checks if 'ent' is the owner of 'cert'.
 *   'ok' is set to 1 if this is the case. If not, it's set to 0.
 */
DLLExport extern int frMailAuthCheckOwner(ZkError err, FrMailAuth cert, FrEntIdExp *ent, int *ok)
{
  const char *fn = "frMailAuthCheckOwner";
  char buf[FRENT_PORTABLE_SIZE + 1]; /* +1 to use strlen */
  size_t esz = 0;

  *ok = 0;
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }
  
  if ( ! ent) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "ent is null");
#endif
    return 0;
  }
  
  if ( ! cert->entsize) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKEPARMBAD, "ent name is not set");
#endif
    return 0;
  }
  
  /* put entity 'ent' in 'buf' */
  memset(buf,0,sizeof buf);
  frEntStore(err, ent, buf, sizeof buf); /* store ent in buf */

  buf[FRENT_PORTABLE_SIZE] = 0; /* should be 0 already */
  esz = strlen(&buf[1]) + 1; /* in case buf[0] == 0 */
  
  *ok = ( (esz == cert->entsize) &&
	  (!memcmp(buf, cert->ent, esz)) );
  
  return 1;
}

/*
 *   Check if the cert is expired. 
 *   'ok' is set to 1 if the cert is valid, or to 0 if it's expired.
 */
DLLExport extern int frMailAuthCheckExpiration(ZkError err, FrMailAuth cert, ZkClock clk, int *ok)
{
  const char *fn = "frMailAuthCheckExpiration";

  *ok = 0;
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }
  
  *ok = ( zkClockCmp(clk,cert->exp) <= 0  );
  
  return 1;
}


/*
 *   Checks if 'ent' is the owner of 'cert' and verify the mac.
 *   'ok' is set to 1 if this is the case. If not, it's set to 0.
 */
DLLExport extern int frMailAuthCheckOwnerAndMac(ZkError err, FrMailAuth cert, FrEntIdExp *ent, FrCryptMacKey key, int *ok)
{
  const char *fn = "frMailAuthCheckOwnerAndMac";
  int okCheck = 0;
  int okMac = 0;

  *ok = 0;
  
  if ( ! cert) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "cert is null");
#endif
    return 0;
  }
  
  if ( ! ent) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "ent is null");
#endif
    return 0;
  }

  if ( ! key) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "key is null");
#endif
    return 0;
  }
  

  if ( ! (frMailAuthCheckOwner(err, cert, ent, &okCheck) && 
	  frMailAuthCheckMac(err, cert, key, &okMac))) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKERROR, "mac verification failed");
#endif
    return 0;
  }
  
  *ok = ( okCheck && okMac );

  return 1;
}

/*
 *  Generates a mac key `mkey' by hashing `ssz' bytes 
 *  of `secret' concatenated with the nym name in `ent'.
 */
DLLExport extern int frGenSharedNymMailMacKey(ZkError err, void *secret, size_t ssz, FrEntIdExp *ent, FrCryptMacKey mkey) 
{
  const char *fn = "frGenSharedNymMailMacKey";
  int ret = 0;
  FrCryptHash h = 0;
  char nymname[64];
  char hash[FRCRYPT_MACKEY_SIZE];
  size_t len = 0;

  if ( ! secret) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "secret is null");
#endif
    return 0;
  } 
  
  if ( ! ent) {
#ifndef NDEBUG
    zkErrorSet(err, fn, ZKERR, ZKMMAILAUTH, ZKENULLPTR, "ent is null");
#endif
    return 0;
  }
  
  /* put entity name in nymname */
  strncpy(nymname, ent->name, 64);
  nymname[63] = '\0';   /* just to make sure */

  
  /*
   *   The Shared Nym-Mail System Mac Key is:
   *   mkey = SHA-1(secret||nym name)
   */
  if ( ! frCryptHashNew(err, &h)) {
    goto done;
  }
  if ( ! frCryptHashUpdate(err, h, secret, ssz)) {
    goto done;
  }

  len = strlen(nymname);
  len = len > 64 ? 64 : len;
  if ( ! frCryptHashUpdate(err, h, nymname, len )) {
    goto done;
  }

  if ( ! frCryptHashGetHash(err, h, hash, FRCRYPT_MACKEY_SIZE)) {
    goto done;
  }
  
  /* put the hash in the mac key */
  /* equivalent to frCryptMacKeySetFromData ? */
  if ( !frCryptMacKeyLoad(err, mkey, hash, FRCRYPT_MACKEY_SIZE )) {
    goto done;
  }

  ret = 1;

 done:

   if (h) { frCryptHashFree(NULL, &h); }

  return ret;

}

