// $Id: imep-encrypt.cpp,v 1.1.2.16 2000/10/17 17:26:23 serge Exp $
// 
// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):


#include <unistd.h>
#include <stdlib.h>
#include <string>
#include <ostream>
#include <istream>
#include <streambuf>
#include <fstream>
#include <iostream>

#include <strings.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <arpa/inet.h> 
#include <netdb.h>

#include <zks.h>

#include <zkTypes.h>
#include <zkIntTypes.h>
#include <zkErrdef.h>
#include <zkError.h>
#include <zkDNS.h>
#include <fr_keyqry.h>
#include <fr_keyqry_fetchkey.h>
#include <fr_rand.h>

#include "frmailmsg.h"
#include "frMailAuth.h"

#define KEY_QUERY_HOST "KEY_QUERY_HOST"
#define KEY_QUERY_SERVICE "KEY_QUERY_SERVICE"
#define KEY_QUERY_TIMEOUT "KEY_QUERY_TIMEOUT"
#define KEY_QUERY_TIMEOUT_DEFAULT 120
#define KEY_QUERY_RETRY "KEY_QUERY_RETRY"
#define KEY_QUERY_RETRY_DEFAULT 5
#define ZKMODULE ZKMFMG

using namespace std;

/*
 *
 * Setup the key query host from the environment variables
 *
 */
static int frKeyQryHostSetup(ZkError err,
			     const char * host, 
			     const char * service,
			     FrKeyQryHost *kh)
{
  const char * func = "frKeyQryHostSetup";
  
  if(kh == NULL || host == NULL || service == NULL ) {
    zkErrorSet(err,func,ZKERR,ZKMODULE,ZKENULLPTR,NULL);
    return 0;
  }
  
  kh->host.sin_family = AF_INET;
  
  
  /*  
   *  Fill in the remote address.
   */
  
  {
    struct hostent *hent;
    unsigned long inaddr;
    
    /*  
     *  Initially, we assume that `host' contains a dotted decimal address.
     *  If the conversion fails it's probably a host name.
     */
    
    if ((inaddr = inet_addr(host)) != INADDR_NONE) {
      memcpy((void *)&kh->host.sin_addr, (void *)&inaddr, sizeof inaddr);
    } else if (zkGetHostByName(err, host, &hent)) {
      kh->host.sin_addr = *(struct in_addr *)(hent->h_addr);
    } else {
      return 0;
    }
  }
  
  /*  
   *  Fill in the remote port.
   */
  
  {
    struct servent *sent;
    unsigned short port;        /* port number in network byte order */
    
    /*  
     *  Initially, we assume that `service' is a port number.  If the
     *  conversion fails it's probably a service name.
     */
    
    if ((port = atoi(service)) != 0) {
      kh->host.sin_port = htons(port);
    } else if ((sent = getservbyname(service, "udp"))) {
      kh->host.sin_port = sent->s_port;
    } else {
      return 0;
    }
  }
  
  if((kh->fd = socket(AF_INET,SOCK_DGRAM,0)) == -1) {
    return 0;
    
  }  
  return 1;
}

static int frKeyQryHostSetupFromEnv(ZkError err, FrKeyQryHost *kh, int *timeout,int *retries)
{
  const char * func = "frKeyQryHostSetupFromEnv";
  char * host = getenv(KEY_QUERY_HOST);          /* audit: ok */
  char * port = getenv(KEY_QUERY_SERVICE);       /* audit: ok */
  char * env_time = getenv(KEY_QUERY_TIMEOUT);   /* audit: ok */
  char * szRetry = getenv(KEY_QUERY_RETRY);      /* audit: ok */

  if (host == NULL) { 
    cerr << "must specify environment variable KEY_QUERY_HOST" << endl;
    zkErrorSet(err,func,ZKERR,ZKMODULE,ZKENULLPTR,"Must specify environment variable KEY_QUERY_HOST");
    return 0;
  }
  if (port == NULL) { 
    cerr << "must specify environment variable KEY_QUERY_SERVICE" << endl;
    zkErrorSet(err,func,ZKERR,ZKMODULE,ZKENULLPTR,"Must specify environment variable KEY_QUERY_SERVICE");
    return 0;
  }
  
  if(env_time == NULL) {
    *timeout = KEY_QUERY_TIMEOUT_DEFAULT;
  } else {
    *timeout = atoi(env_time);
  }

  if(szRetry == NULL) {
    *retries = KEY_QUERY_RETRY_DEFAULT;
  } else {
    *retries = atoi(szRetry);
  }

  return frKeyQryHostSetup(err,host,port,kh);
}


#define ERROR_MESSAGE_BUFFER_SIZE 8192

static int frNymPubKeyLookup(ZkError err,const char * Name,FrCryptPubKey *key)
{
  FrKeyQry kq;     /* the key cache */
  FrKeyQryHost kh; /* the key query host */
  FrEntIdExp    keyEnt;
  int retries = -1;
  int timeout = 0;
  ZkBool found,valid;
  char error_message_buffer[ERROR_MESSAGE_BUFFER_SIZE];

  /* setup the key query host */
  if(!frKeyQryHostSetupFromEnv(err,&kh,&timeout,&retries)) {
    zkErrorMessage(err, error_message_buffer, ERROR_MESSAGE_BUFFER_SIZE);
    fprintf(stderr,"Failed creating key query host: `%s'\n", error_message_buffer);
    return 111;
  }
  
  /* create a new key cache */
  if (!frKeyQryNew(err, &kq)) {
    zkErrorMessage(err, error_message_buffer, ERROR_MESSAGE_BUFFER_SIZE);
    fprintf(stderr,"Failed creating new key query: `%s'\n", error_message_buffer);
    return 111;
  }
  
  /* set the entity type */
  keyEnt.type = FRENT_TYPE_NYM;
  strncpy(keyEnt.name,Name,FRENT_NAME_SIZE);
  
  /* get the key */
  if(!frKeyQryFetchPubKeyWithStatus(err,kq, 0, FRCRYPT_KEYTYPE_CRYPTION,
			  keyEnt,kh,retries,timeout,key,&found,&valid)) {
    zkErrorMessage(err, error_message_buffer, ERROR_MESSAGE_BUFFER_SIZE);
    fprintf(stderr,"Fetching Pub Key Failed: `%s'\n", error_message_buffer);
    return 111;
  }
  
  if(!found || !valid) {
    /* the key for the nym wasn't found, so permanent failure */
    fprintf(stderr,"This nym does not exist (public key not found).");
    return 100;
  }

  /* everything is fine in freedomland */
  return 0;
  
}


// this just wraps a message up in 70-character lines, including continued headers,
// and prints it to stdout.
void dump_normalized_message(ostream &out, char *msg, int mlen) {
  char c, last = '\0';
  bool inheader = true;
  char lf = '\n';
  unsigned long len = 0;
  while(mlen--) {    
    c = *msg++;
    // decide if we're still in a header
    if (c == lf) {
      len = 0;
      if (last == lf) inheader = false;
    }
    last = c;

    // wrap lines
    if (len == 70) {
      if (inheader) {
	out << lf << ' ' << c; len = 2;
      } else {
	out << lf << c; len = 1;
      }
    } else {
      out.put(c); len += (c == lf ? 0 : 1); 
    }
  }
}


int fork_subprocess_and_feed_message(char **argv, char *msg, int mlen) {
  int fds[2];
  if (pipe(fds)) {
    cerr << "pipe failed" << endl;
    return -1;
  }
  pid_t res = fork();
  int status;
  switch(res) {
  case -1:
    cerr << "fork failed" << endl;
    return -1;
    break;
  case 0:
    close(fds[1]);
    dup2(fds[0],fileno(stdin)); 
    execv(argv[0], argv); 
    cerr << "exec " << argv[0] << " failed" << endl;
    return -1;
    break;
  default:
    {
      close(fds[0]);
      ostream out(new filebuf(fds[1],"subprocess",ios_base::out));
      dump_normalized_message(out,msg,mlen);
      out.flush();
      close(fds[1]);
    }    
      
    if((wait(&status) == -1) || (WIFEXITED(status) == 0)) {
      cerr << "child " << argv[0] << " exited abnormally" << endl;
      return -1;
    }
    return WEXITSTATUS(status);
    break;
  }
  return -1;
}

#define READ_SIZE 65536

// An instance of a Stream_Slurper will read an entire input stream
// from a file descriptor, into a contiguous in-core buffer.

class Stream_Slurper
{
public:
  Stream_Slurper() 
    : _buf(0),
      _buffer_size(0),
      _bytes_read(0)
  {
  }

  ~Stream_Slurper() 
  {
    delete _buf;
  }
           
  // return the contiguous buffer
  char *buf() 
  { 
    return _buf; 
  }

  // return the buffer's size
  unsigned size() 
  { 
    return _bytes_read; 
  }

  // read from fd into our buffer
  int slurp(int fd)
  {
    for(;;) 
      {
	int read_size = 0;

	// Do we need to grow the buffer?
	if (_buffer_size <= _bytes_read) 
	  {
	    grow();
	  }
	do {
	  // Read into the rest of the buffer.
	  // The above test and grow() guarantee that _buffer_size - bytes_read != 0.

	  read_size = read(fd, _buf + _bytes_read, _buffer_size - _bytes_read);

	  // ignore EINTR
	} while (read_size == -1 && errno == EINTR);

	if (read_size == 0) {
	  // finished reading
	  return 1;
	} else if (read_size < 0) {
	  // permanent failure
	  perror("read(0, ...) failed");
	  return 0;
	} else {
	  // continue reading: update count
	  _bytes_read += read_size;
	}
      }
  }
  
private:
  // Grows the internal buffer.
  void grow()
  {
    if (_buffer_size == 0) {
	_buffer_size = READ_SIZE;
      } else {
	_buffer_size *= 2;
      }
    // Create a new buffer
    char *new_buf = new char [_buffer_size];
    // Copy the old buffer to the new
    memcpy(new_buf, _buf, _bytes_read);
    // Delete the old
    delete _buf;
    // Set the new
    _buf = new_buf;
  }

  char *_buf;
  unsigned _buffer_size;
  unsigned _bytes_read;
};


//
// this gets a recipient public key, reads  a signing mac key on fd 3, and a
// message on stdin, encrypts the message with the pubkey, signs the result and
// prints it on stdout. 
//
// if it sees the environment variable TESTING is set, it will read a pubkey from
// fd 4, just to save you having to have a KQD setup all the time.
//
// if given a command line option, forks and execs the command line arguments, 
// with stdin reading from imep-encrypt's stdout. fails if the subprocess fails, 
// succeeds otherwise.
//

ZkError err_; 

#define ensure(x) do { if (! x) { char buf[1024]; zkErrorMessage(err_,buf,1024); \
cerr << "failure " #x " " << endl << buf << endl; exit(100); }} while (0)

int main(int argc, char **argv) {
  
  zkErrorNew (&err_);
  ensure(frCryptInit(err_, 0));
  zkRand_Debug();
  char c;
  int iRet = 0;
  
  const char *sender_ch = getenv("SENDER");   /* audit: ok */
  if (sender_ch == NULL) { cerr << "must specify environment variable SENDER" << endl;
  return 100; 
  }
  string sender_email(sender_ch);
  
  FrCryptCrypt renc = 0; 


  //work out nym
  const char * local_ch = getenv("LOCAL"); /* audit: ok */
  if(local_ch == NULL) { cerr << "must specify environment variable LOCAL" << endl;
  return 100;
  }
  FrEntIdExp nym;
  ensure(frEntSetFromStr2(err_, &nym, "NYM", local_ch));

  const char * testing = getenv("TESTING"); /* audit: ok */
  if (testing != NULL) {
    // load pubkey from fd 4 if TESTING env variable is set
    istream pk(new filebuf(4,"pubkey",ios_base::in));
    string pubkey;
    while(pk.get(c)) pubkey += c;
    size_t pk_sz = pubkey.size();
    ensure(frCryptPubKeyNew(err_, &renc));
    ensure(frCryptPubKeyReadFromBuf(err_, renc, const_cast<char *>(pubkey.c_str()), &pk_sz));
    ensure(frCryptPubKeySetSharedParms(err_,renc));

  } else {
    // load pubkey from kqd
    string recipient_user(local_ch);
    /* if the key succeeded, returns 0 which isn't doesn't let us in */
    if((iRet = frNymPubKeyLookup(err_,recipient_user.c_str(),&renc))) {
      // try to fetch and reencrypt message later.
      return iRet;
    }
  }
    
  // load shared secret from fd 3
  istream sk(new filebuf(3,"sharedsecret",ios_base::in));
  string secret;
  while(sk.get(c)) secret += c; 

  // derive mac key
  FrCryptMacKey mac = 0; 
  ensure(frCryptMacKeyNew(err_, &mac));
  ensure(frGenSharedNymMailMacKey(err_, const_cast<char *>(secret.c_str()), secret.size(), &nym, mac));
  
  // load email body from fd 0
  Stream_Slurper slurper;
  if (!slurper.slurp(0)) {
    fprintf(stderr, "Failed to slurp message from stdin");
    /* permanent failure */
    return 100;
  }
  
  // construct ourselves
  FrEntIdExp imep;
  ensure(frEntSetFromStr2(err_, &imep, "imep", "IMEP"));
  
  // build, sign and encrypt message
  FrMailMsg mm = 0;
  ensure(frMailMsgNew(err_, &mm));  
  ensure(frMailMsgSetSenderEmail(err_, mm, sender_email.c_str(), sender_email.size()));
  ensure(frMailMsgCreateWithMac(err_, mm, renc, mac, &imep, NULL, 0, slurper.buf(), slurper.size()));
  
  // dump to stdout
  char *mp; size_t mlen;
  ensure(frMailMsgGive(err_, mm, &mp, &mlen));
  ensure(frMailMsgFree(err_, &mm));  

  int status;
  if (argc > 1) {
    if ((status = fork_subprocess_and_feed_message(++argv,mp,mlen)) == -1) {
      return 111; // temporary failure
    } else {
      return status;
    }
  } else {
    dump_normalized_message(cout,mp,mlen);
    return 0;
  }
}
