/* ===========================================================================
 *
 * $Id: verify-envsig.cpp,v 1.1.2.5 2001/01/05 16:24:56 marcp Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *
 * Verify the enveloppe signature of a email.
 * 
 * Description:
 *
 * In the Cacophonix mail system, each email from a nym to an internet
 * user has an envelope signature field of this form:
 *
 *    X-Freedom-Envelope-Sig: <recipient email> <base 64 signature>
 *
 * where the signature is computed as:
 *
 *    <base 64 sig> = Base64( Sig_[Snym] (recipient,sender) )
 *
 *    where Snym is the nym's signature key.
 * 
 * This program fetches the nyms public key and verifies the signature.
 * The result of the verification is printed on stdout.
 *
 * Note: the program is based on the old program: frmaip/prog/frmailverify.c
 *
 */

#include "zks.h"
#include "zkError.h"
#include "frCrypt.h"
#include "fr_rand.h"
#include "frKeyQry.h"
#include "frNetworkDefaults.h"
#include "frmailmsg.h"
#include "zkDNS.h"

#include <ctype.h>
#include <netdb.h>
#include <stdio.h>
#include <arpa/inet.h>
#include <string.h>

#define EXIT_SIG_OK      (0)
#define EXIT_SIG_ERROR   (1)
#define EXIT_HOST_ERROR  (2)
#define EXIT_ERROR       (3)
#define EXIT_USAGE_OK    (4)
#define EXIT_USAGE_ERROR (5)

#define STR(n) (#n)
#define STR2(n) STR(n)

#define DEBUG 0
#define debugOut if (DEBUG) printf


/*
 *  Fill in a sockaddr structure, given the host, service and protocol.
 *  
 *  Host may be an alphanumeric hostname, or a dotted decimal address; service
 *  may be an textual service name (e.g. http), or a port number.  The
 *  protocol is either "tcp" or "udp".
 *
 *  Note: stolen from zk/freedom/libs/frmaip/prog/frmailverify.c
 *  Note2: modify: added ZkError.
 *
 *  Bug: revert to using gethostbyname instead of zkGetHostByName. Didn't 
 *       work when using the host name.
 */
static int makeSaddr(ZkError err, const char *host, const char *service, const char *proto, struct sockaddr *addr)
{
  struct hostent *hent;
  struct sockaddr_in a;
  unsigned long inaddr;

  memset(&a, 0, sizeof a);
  a.sin_family = AF_INET;

  /*  
   *  Initially, we assume that `host' contains a dotted decimal address.
   *  If the conversion fails it's probably a host name.
   */
    
  if ((inaddr = inet_addr(host)) != INADDR_NONE) {
    memcpy((void *)&a.sin_addr, (void *)&inaddr, sizeof inaddr);
  } else if ((hent = gethostbyname(host))) {
    a.sin_addr = *(struct in_addr *)(hent->h_addr);
  } else {
    return 0;
  }

  /*  
   *  Fill in the remote port.
   */

  {
    struct servent *sent;
    unsigned short port;        /* port number in network byte order */

    /*  
     *  Initially, we assume that `service' is a port number.  If the
     *  conversion fails it's probably a service name.
     */
    
    if ((port = atoi(service)) != 0) {
      a.sin_port = htons(port);
    } else if ((sent = getservbyname(service, proto))) {
      a.sin_port = sent->s_port;
    } else {
      return 0;
    }
  }

  memcpy(addr, &a, sizeof *addr);

  return 1;
}

/*  
 *  Open a TCP connection to a `service' on `host'.  `host' may be a domain
 *  name or an IP address, and `service' may be a service name or a port
 *  number.
 *  
 *  Return a file descriptor open for reading and writing, unless an error
 *  occurs, in which case return -1.
 *
 *  Note: stolen from zk/freedom/libs/frmaip/prog/frmailverify.c
 *
 *  Bug: revert to using gethostbyname instead of zkGetHostByName. Didn't 
 *       work when using the host name.
 */
static int bhOpenUDP(ZkError err, const char *host, const char *service)
{
  int fd;
  struct sockaddr_in server_addr;

  memset(&server_addr, 0, sizeof server_addr);
  server_addr.sin_family = AF_INET;

  /*  
   *  Fill in the remote address.
   */

  {
    struct hostent *hent;
    unsigned long inaddr;

    /*  
     *  Initially, we assume that `host' contains a dotted decimal address.
     *  If the conversion fails it's probably a host name.
     */
    
    if ((inaddr = inet_addr(host)) != INADDR_NONE) {
      memcpy((void *)&server_addr.sin_addr, (void *)&inaddr, sizeof inaddr);
    } else if ((hent = gethostbyname(host))) {
      server_addr.sin_addr = *(struct in_addr *)(hent->h_addr);
    } else {
      return -1;
    }
  }

  /*  
   *  Fill in the remote port.
   */

  {
    struct servent *sent;
    unsigned short port;        /* port number in network byte order */

    /*  
     *  Initially, we assume that `service' is a port number.  If the
     *  conversion fails it's probably a service name.
     */
    
    if ((port = atoi(service)) != 0) {
      server_addr.sin_port = htons(port);
    } else if ((sent = getservbyname(service, "udp"))) {
      server_addr.sin_port = sent->s_port;
    } else {
      return -1;
    }
  }
  
  if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    return -1;
  }

#if 0
  if (connect(fd, (struct sockaddr *)&server_addr, sizeof server_addr) < 0) {
    close(fd);
    return -1;
  }
#endif
  
  return fd;
}


/*
 *  May `c' appear in a nym name?
 *  
 *  Note: stolen from zk/freedom/libs/frmaip/prog/frmailverify.c
 */
static int isValidNymChar(int c)
{
  return (islower(c) ||
          isupper(c) ||
          isdigit(c) ||
          c == '-'   ||
          c == '_'   ||
          c == '.');
}


/*
 *  Extract the first part of an email address (i.e. what's before @).
 *
 *  Note: inspired from zk/freedom/libs/frmaip/prog/frmailverify.c
 */
static int getNymName(const char *email, char name[FRENT_NAME_SIZE + 1])
{
  const char *b, *e; /* beginning and end of nym name */
  int ret = 0;

  /*
   *  bug? How much serious parsing must we do?
   */
  
  if ( ! (e = strchr(email, '@'))) {
    printf("no `@' in email address `%s'\n", email);
    goto done;
  }

  if (e == email) {
    printf("no nym name before `@' in email address `%s'\n", email);
    goto done;
  }

  for (b = e - 1; b > email; b--) {
    if (isValidNymChar(*b)) {
      continue;
    }
  }

  if ( ! isValidNymChar(*b)) {
    b++;
  }

  if (b == e) {
    printf("empty nym name in email address `%s'\n", email);
    goto done;
  }

  memcpy(name, b, e - b + 1);
  name[e - b] = '\0';

  ret = 1;

 done:

  return ret;
}


/* print the program usage */
static void usage(int code, const char *prog)
{
  printf( "Usage: %s"
          " [-h|-help|--help]"
          " [--host <hostname>]"
          " [--service <servicename>]"
          " --sender <sender's email>"
          " --recip <recipient's email>"
          " --sig <base64 signature>"
          "\n",
          prog);
  exit(code);
}


static const char *tail(const char *path)
{
  const char *p = strrchr(path, '/');
  return p ? p + 1 : path;
}


int main (int ac, char **av) {
  const char *prog;
  int ret = EXIT_ERROR;
  ZkError err = 0;
  const char *host    = FR_DFLT_KEYQRY_SERVER_HOST;
  const char *service = STR2(FR_DFLT_KEYQRY_SERVER_PORT);
  char *sender = 0;
  char *recip  = 0;
  char *sig    = 0;
  FrKeyQryHost kqhost;
  FrKeyQry kq = 0;
  unsigned int kv = 0; /* signer's key version */
  FrEntIdExp kown;     /* key owner */
  char nymname[FRENT_NAME_SIZE+1];
  int retries = 5;
  int timeout = 3;
  FrCryptSig sigKey; /* nym's signing key */
  int fd = -1;
  int valid = 0; /* is the sig valid */
  int ok = 0;

  prog = tail(av[0]);  

  if ( ! zkErrorNew(&err)) {
    printf("%s: can't allocate error object.\n", prog);
    goto done;
  }

  if ( ! frCryptInit(err, 0)) {
    printf("error initialising the crypto library\n");
    goto done;
  }
  
  
  /*
   *  read arguments
   */

  while (av++, --ac) {
    if (strcasecmp("-h",        av[0]) == 0 ||
        strcasecmp("-help",     av[0]) == 0 ||
        strcasecmp("--help",    av[0]) == 0) {
      usage(EXIT_USAGE_OK,prog); goto done;
    } else if (strcasecmp("--host", av[0]) == 0) {
      if (av++, --ac < 1) { usage(EXIT_USAGE_ERROR,prog); }
      host = av[0];
    } else if (strcasecmp("--service", av[0]) == 0) {
      if (av++, --ac < 1) { usage(EXIT_USAGE_ERROR,prog); }
      service = av[0];
    } else if (strcasecmp("--sender", av[0]) == 0) {
      if (av++, --ac < 1) { usage(EXIT_USAGE_ERROR,prog); }
      sender = av[0];
    } else if (strcasecmp("--recip", av[0]) == 0) {
      if (av++, --ac < 1) { usage(EXIT_USAGE_ERROR,prog); }
      recip = av[0];
    } else if (strcasecmp("--sig", av[0]) == 0) {
      if (av++, --ac < 1) { usage(EXIT_USAGE_ERROR,prog); }
      sig = av[0];
    } else {
      printf("%s: unexpected program argument `%s'.\n",
              prog, av[0]);
      usage(EXIT_USAGE_ERROR,prog); goto done;
    }
  }
  if ( sender == 0 || recip == 0 || sig == 0 ) {
    /* those fields are necessary */
    usage(EXIT_USAGE_ERROR,prog); goto done;
  }

  debugOut("sender: %s\nrecip: %s\nsig: %s\n", sender, recip, sig);  
  debugOut("host: %s\nservice: %s\n", host, service);

   
  /*
   *  get sender's key from key server
   */

  if ( ! makeSaddr(err, host, service, "udp", (struct sockaddr *)&kqhost.host)) {
    printf("error creating sockaddr structure for `%s/%s'\n",
	    host, service);
    ret = EXIT_HOST_ERROR; goto done;
  }
    
  if ( ! (fd = bhOpenUDP(err, host, service))) {
    printf("error creating socket to `%s/%s'\n",
	    host, service);
    ret = EXIT_HOST_ERROR; goto done;
  }
    
  kqhost.fd = fd;
    
  if ( ! (getNymName(sender, nymname) &&
	  frEntSetFromStr2(err, &kown, "nym", nymname)) ) { 
    printf("error setting sender entity\n");
    ret = EXIT_ERROR; goto done; 
  }
  debugOut("nymname: %s\n", nymname);  

  if ( ! frKeyQryNew(err, &kq)) { 
    printf("error creating query\n");
    ret = EXIT_HOST_ERROR; goto done; 
  }

  if ( ! frMailEnvelopeGetSigVersion(err, sig, &kv)) {
    printf("error extracting key version from signature\n");
    ok = 1; ret = EXIT_SIG_ERROR; goto done;
  }
  debugOut("key version: %d\n",kv); 

  if ( ! frKeyQryFetchPubKey(err, kq, kv, FRCRYPT_KEYTYPE_SIGNING,
			     kown, kqhost, retries, timeout, &sigKey)) { 
    printf("error fetching the public key\n");
    ret = EXIT_HOST_ERROR; goto done; 
  }
  
  
  /*
   *  verify envelope signature
   */
  
  if ( ! frMailEnvelopeAddrSigVerify( err, sigKey, sender, recip, sig, &valid )) {
    ret = EXIT_ERROR; goto done;
  }  
  
  ret = valid ? EXIT_SIG_OK : EXIT_SIG_ERROR;
    
  ok = 1;

 done:
    
  frCryptFinal(err, 0);
  if (fd != -1) { close(fd); }
  if (err)      { zkErrorFree(&err); }
  if (kq)       { frKeyQryFree(err, &kq); }

  if (ok) {
    printf("Signature is %s\n", valid ? "valid." : "*** invalid ***");
  }
  
  return ret;
}
