/*
 *  $Id: frMaipRb.c,v 1.23.28.3 2000/12/14 17:52:57 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Reply blocks.
 *  
 *  Private reply blocks are those stored on the client, used to decrypt
 *  incoming Freedom mail.
 *  
 *  Public reply blocks are those used by MAIPs to route mail to a Freedom
 *  user.
 */


#define DEBUG_MAC 0


#include "zks.h"

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "frCrypt.h"
#include "frPktFmt.h"
#include "frSec.h"
#include "zkAlloc.h"
#include "zkError.h"
#include "frmaip_repblk.h"


/*
 *  When data is being encrypted with the symmetric key we append part of a
 *  MAC.  Make it a multiple of 8 bytes.
 */
#define FRMAIP_MAC_LEN 16


#pragma pack(1)

/*
 *  We overlay the following structure on the key seed material, in order to
 *  generate various keys.  When doing so, be sure to skip the first two
 *  bytes, which may be predictable.
 *  
 *  The MAC key is used only by the FMG to create a MAC, and by the client to
 *  verify it.
 *  
 *  Warning!  The length of the structure had better not be longer than the
 *  seed material!
 */

struct fr_maip_keyseed_pk {
  char encrKeySeed[FRCRYPT_SYMKEY_SEED_SIZE];
  char replyKeySeed[FRCRYPT_SYMKEY_SEED_SIZE];
  char replyMacKey[FRCRYPT_MAC_SIZE];
};

#pragma pack()


/*
 *  The internal (i.e. in core) representation of a single hop.
 *  
 *  Note that certain fields may not contain useful data, depending on how
 *  this structure was initiailized.  That is, if it was loaded from a private
 *  reply block file, then the contents of `keySeed' will be meaningless: it
 *  is assumed that a public reply block has already been generated.
 *  
 *  For simplicity, each hop will have an associated MAC key, but only the FMG
 *  (i.e. first hop) will use it.  We don't want to append a MAC at each hop,
 *  since that would change the message size in a predictable manner, allowing
 *  attackers to track the route of the message.
 *  
 *  bug: various parts of this structure probably ought to be locked in
 *  memory.
 */
struct fr_maip_hop {
  FrCryptSym encrSym;
  FrCryptCrypt encrPubKey;
  FrCryptSym replySym; /* encrypt/decrypt mail body */
  FrCryptMacKey replyMacKey;
  struct fr_maip_keyseed_pk ksm;
  FrEntIdExp maip;
  uint16_t port; /* 0 implies this is last hop */
  uint32_t latency;
};

/*
 *  This structure is the internal (i.e. in core) representation of a private
 *  reply block.
 *  
 *  If `decryptOnly' is 1, then this reply block may only be used to decrypt
 *  incoming mail; it can't generate a public reply block, since the
 *  nescessary information has not been saved.  In practice, this flag will be
 *  set when the reply block is initialized from its external representation.
 */
struct frmaip_rbpriv {
  int decryptOnly;
  char rbName[FRMAIP_RBNAME_LEN];
  FrEntIdExp nym;
  unsigned int nHops; /* number of valid hops */
  struct fr_maip_hop hop[FRMAIP_HOPS_MAX];
  char email[FRMAIP_EMAIL_LEN];
  void *userData;
};


DLLExport int frMaipRbPrivNew(ZkError err, FrMaipRbPriv *repblk)
{
  int i, ret = 0;
  FrMaipRbPriv rb = 0;

  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivNew", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }

  if ( ! (rb = malloc(sizeof *rb))) {
    zkErrorSet_va(err, "frMaipRbPrivNew", ZKERR, ZKMFRMAIP, errno,
                  "can't allocate %lu bytes for private reply block: %s",
                  (unsigned long)sizeof *rb, strerror(errno));
    goto done;
  }

  memset(rb, 0, sizeof *rb);

  /* Explicitly set pointers to 0. */

  for (i = 0; i < FRMAIP_HOPS_MAX; i++) {
    rb->hop[i].encrSym  = 0;
    rb->hop[i].encrPubKey  = 0;
    rb->hop[i].replySym = 0;
    rb->hop[i].replyMacKey = 0;
  }
  rb->userData = 0;

  *repblk = rb; rb = 0; /* caller now owns it */
  ret = 1;

 done:

  if (rb) { free(rb); }

  return ret;
}


DLLExport int frMaipRbPrivFree(ZkError err, FrMaipRbPriv *repblk)
{
  FrMaipRbPriv rb = 0;
  int i;

  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivFree", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! (rb = *repblk)) {
    zkErrorSet(err, "frMaipRbPrivFree", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "*repblk is null");
    return 0;
  }

  for (i = 0; i < FRMAIP_HOPS_MAX; i++) {
    struct fr_maip_hop *h = &rb->hop[i];
    
    if (h->encrSym)     { (void)frCryptSymFree(0,    &h->encrSym); }
    if (h->encrPubKey)  { (void)frCryptPubKeyFree(0, &h->encrPubKey); }
    if (h->replySym)    { (void)frCryptSymFree(0,    &h->replySym); }
    if (h->replyMacKey) { (void)frCryptMacKeyFree(0, &h->replyMacKey); }
  }

  memset(rb, 0, sizeof *rb);
  free(rb); *repblk = 0;

  return 1;
}


DLLExport int frMaipRbPrivGetEmail(ZkError err, FrMaipRbPriv repblk, const char **email)
{
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivGetEmail", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! email) {
    zkErrorSet(err, "frMaipRbPrivGetEmail", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "email is null");
    return 0;
  }

  *email = repblk->email;

  return 1;
}


DLLExport int frMaipRbPrivSetEmail(ZkError err, FrMaipRbPriv repblk, const char *email)
{
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivSetEmail", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! email) {
    zkErrorSet(err, "frMaipRbPrivSetEmail", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "email is null");
    return 0;
  }

  if (strlen(email) >= sizeof repblk->email) {
    zkErrorSet_va(err, "frMaipRbPrivSetEmail", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "email address too long (%lu > %lu)",
                  (unsigned long)strlen(email), (unsigned long)repblk->email);
    return 0;
  }

  strncpy(repblk->email, email, sizeof repblk->email);

  return 1;
}


DLLExport int frMaipRbPrivGetName(ZkError err, FrMaipRbPriv repblk, const char **name)
{
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivGetName", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! name) {
    zkErrorSet(err, "frMaipRbPrivGetName", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "email is null");
    return 0;
  }

  *name = repblk->rbName;

  return 1;
}


DLLExport int frMaipRbPrivSetName(ZkError err, FrMaipRbPriv repblk, const char *name)
{
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivSetName", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! name) {
    zkErrorSet(err, "frMaipRbPrivSetName", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "email is null");
    return 0;
  }

  if (strlen(name) >= sizeof repblk->rbName) {
    zkErrorSet_va(err, "frMaipRbPrivSetName", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "reply block name is too long (%lu > %lu)",
                  (unsigned long)strlen(name), (unsigned long)repblk->rbName);
    return 0;
  }

  strncpy(repblk->rbName, name, sizeof repblk->rbName);

  return 1;
}


DLLExport int frMaipRbPrivGetNym(ZkError err, FrMaipRbPriv repblk, const FrEntIdExp **nym)
{
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivGetNym", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! nym) {
    zkErrorSet(err, "frMaipRbPrivGetNym", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "nym is null");
    return 0;
  }

  *nym = &repblk->nym;

  return 1;
}


DLLExport int frMaipRbPrivSetNym(ZkError err, FrMaipRbPriv repblk, FrEntIdExp *nym)
{
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivSetNym", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! nym) {
    zkErrorSet(err, "frMaipRbPrivSetNym", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "nym is null");
    return 0;
  }

  /* bug: check validity of nym */

  repblk->nym = *nym;

  return 1;
}


DLLExport int frMaipRbPrivGetUserData(ZkError err, FrMaipRbPriv repblk, void **ud)
{
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivGetUserData", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! ud) {
    zkErrorSet(err, "frMaipRbPrivGetUserData", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "ud is null");
    return 0;
  }

  *ud = repblk->userData;

  return 1;
}


DLLExport int frMaipRbPrivSetUserData(ZkError err, FrMaipRbPriv repblk, void *ud)
{
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivSetUserData", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! ud) {
    zkErrorSet(err, "frMaipRbPrivSetUserData", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "ud is null");
    return 0;
  }

  repblk->userData = ud;

  return 1;
}


/*
 *  Get the n'th hop of the private reply block.
 *  
 *  When `n' corresponds to the last hop `lastHop' will be set to 1, otherwise
 *  it will be 0.
 *  
 *  The reply block is assumed to be fixed when this routine is called.  In
 *  particular, the nHops field specifies the number of hops contained within,
 *  while the parameter `n' is the index of the request hop (from 0 to
 *  nHops-1 inclusive).
 */
DLLExport int frMaipRbPrivGetHop(ZkError err, FrMaipRbPriv repblk, unsigned int n, int *lastHop, struct frmaip_rbpriv_set_hop *h)
{
  int ret = 0;

  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivGetHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if (n >= repblk->nHops) {
    zkErrorSet_va(err, "frMaipRbPrivGetHop", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "hop number is too large (%d >= %u)",
                  n, (unsigned int)repblk->nHops);
    return 0;
  }
  if ( ! lastHop) {
    zkErrorSet(err, "frMaipRbPrivGetHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "lastHop is null");
    return 0;
  }
  if ( ! h) {
    zkErrorSet(err, "frMaipRbPrivGetHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "h is null");
    return 0;
  }

  if (repblk->nHops > FRMAIP_HOPS_MAX) {
    zkErrorSet_va(err, "frMaipRbPrivGetHop", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "hop number is too large (%u > %d)",
                  (unsigned int)repblk->nHops, FRMAIP_HOPS_MAX - 1);
    return 0;
  }

  {
    FrCryptSymAlg encrSymAlg, replySymAlg;
    struct fr_maip_hop *privHop = &repblk->hop[n];
    
    h->encrPubKey = 0; /* we don't even try to get this... */
    if ( ! (frCryptSymGetAlg(err, privHop->encrSym, &encrSymAlg) &&
            frCryptSymGetAlg(err, privHop->replySym, &replySymAlg))) {
      goto done;
    }
    h->encrSymAlg = encrSymAlg;
    h->replySymAlg = replySymAlg;
    h->maip = privHop->maip;
    h->port = privHop->port;
    h->latency = privHop->latency;

    *lastHop = privHop->port == 0;
  }

  ret = 1;

 done:

  return ret;
}


/*
 *  Set the next hop of the reply block.
 *  
 *  Note: we don't generate any random keys here (although we do _create_ the
 *  keys).  We wait until the creation of a public reply block to generate the
 *  pre-key seed material, etc.
 *  
 *  The reply block is still "under construction" when this function is
 *  called, so the `nHops' field should be interpreted as the index of the
 *  next reply block to use.  Therefore, when called, `nHops' should be in the
 *  range [0, FRMAIP_HOPS_MAX - 1].
 */
DLLExport int frMaipRbPrivSetNextHop(ZkError err, FrMaipRbPriv repblk, int firstHop, int lastHop, struct frmaip_rbpriv_set_hop *h)
{
  int ret = 0;
  struct fr_maip_hop *privHop;

  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivSetNextHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! h) {
    zkErrorSet(err, "frMaipRbPrivSetNextHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "h is null");
    return 0;
  }

  if (repblk->nHops >= FRMAIP_HOPS_MAX) {
    zkErrorSet_va(err, "frMaipRbPrivSetNextHop", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "hop number is too large (%u > %d)",
                  (unsigned int)repblk->nHops, FRMAIP_HOPS_MAX - 1);
    return 0;
  }

  /* bug: more error checking on `hop' argument. */

  if ( ! h->encrPubKey) {
    zkErrorSet(err, "frMaipRbPrivSetNextHop", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
               "encryption public key not set");
    return 0;
  }

  /*
   *  Point to the next available hop element.
   */

  privHop = &repblk->hop[repblk->nHops];

  /*
   *  Duplicate the key so that the caller can reuse the copy passed (should
   *  they so desire).
   */
  
  if ( ! frCryptPubKeyDup(err, &privHop->encrPubKey, h->encrPubKey)) {
    goto done;
  }

  if ( ! (frCryptSymNew(err, &privHop->encrSym) &&
          frCryptSymSetAlg(err, privHop->encrSym, h->encrSymAlg) &&
          frCryptSymNew(err, &privHop->replySym) &&
          frCryptSymSetAlg(err, privHop->replySym, h->replySymAlg))) {
    goto done;
  }

  /*
   *  Only the first hop (i.e. Freedom Mail Gateway) should use a MAC.
   */

#if DEBUG_MAC
  fprintf(stderr, "frMaipRbPrivSetNextHop: 1st hop: %d\n", firstHop);
#endif

  if (firstHop && ! frCryptMacKeyNew(err, &privHop->replyMacKey)) {
    goto done;
  }

  if (lastHop) {
    memset(&privHop->maip, 0, sizeof privHop->maip);
    privHop->port = 0;
  } else {
    /* bug: check validity of maip. */
    privHop->maip = h->maip;
    if ((privHop->port = h->port) == 0) {
      zkErrorSet(err, "frMaipRbPrivSetNextHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
                 "port is 0");
      goto done;
    }
  }
  
  privHop->latency = h->latency;
  repblk->nHops++;

  ret = 1;

 done:

  if ( ! ret) {
    if (privHop->encrSym)     { frCryptSymFree(0,    &privHop->encrSym);  }
    if (privHop->replySym)    { frCryptSymFree(0,    &privHop->replySym); }
    if (privHop->replyMacKey) { frCryptMacKeyFree(0, &privHop->replyMacKey); }
  }

  return ret;
}


/*
 *  These packed structures are used to externally store the client's copy of
 *  its reply blocks.
 */

#pragma pack(1)

/*
 *  The external storage representation of each hop in the private reply
 *  block.
 */
struct fr_maip_rbpriv_hop_pk {
  uint8_t encrSymAlg;
  uint8_t encrPubKeyVers;
  uint8_t replySymAlg;
  struct fr_maip_keyseed_pk ksm;
  char maip[FRENT_PORTABLE_SIZE];
  uint16_t port;
  uint32_t latency;
};

/*
 *  This structure defines the external storage representation of the private
 *  reply block.
 */
struct fr_maip_rbpriv_pk {
  uint8_t vers;
  uint8_t type;
  char rbName[FRMAIP_RBNAME_LEN]; /* a name given to identify this reply block */
  char nym[FRENT_PORTABLE_SIZE]; /* the associated nym */
  uint8_t nHops; /* number of valid hops */
  struct fr_maip_rbpriv_hop_pk hop[FRMAIP_HOPS_MAX];
  char email[FRMAIP_EMAIL_LEN];
};

#pragma pack()


/*
 *  Initialize a private reply block from the contents of a buffer.
 */
DLLExport int frMaipRbPrivReadFromBuf(ZkError err, FrMaipRbPriv repblk, void *rbuf, size_t rsz)
{
  int ret = 0;
  size_t i;
  struct fr_maip_rbpriv_pk *rb = rbuf;

  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivReadFromBuf", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! rbuf) {
    zkErrorSet(err, "frMaipRbPrivReadFromBuf", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rbuf is null");
    return 0;
  }
  if (rsz < sizeof *rb) {
    zkErrorSet_va(err, "frMaipRbPrivReadFromBuf", ZKERR, ZKMFRMAIP, ZKENULLPTR,
                  "rsz is too small (%lu < %lu)",
                  (unsigned long)rsz, (unsigned long)sizeof *rb);
    goto done;
  }

  if (rb->vers != 1) { /* bug */
    zkErrorSet_va(err, "frMaipRbPrivReadFromBuf", ZKERR, ZKMFRMAIP, ZKENULLPTR,
                  "invalid version number (%u)", (unsigned int)rb->vers);
    goto done;
  }
  if (rb->type != FRPKTTYPE_REPLYBLOCK_PRIV) {
    zkErrorSet_va(err, "frMaipRbPrivReadFromBuf", ZKERR, ZKMFRMAIP, ZKENULLPTR,
                  "invalid type (%u)", (unsigned int)rb->type);
    goto done;
  }
  if (rb->nHops > FRMAIP_HOPS_MAX) {
    zkErrorSet(err, "frMaipRbPrivReadFromBuf", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "too many hops");
    goto done;
  }

  memset(repblk, 0, sizeof *repblk);

  repblk->decryptOnly = 1;
  strncpy(repblk->rbName, rb->rbName, sizeof repblk->rbName);

  repblk->nym.type = rb->nym[0]; /* bug */
  memcpy(repblk->nym.name, &rb->nym[1], sizeof repblk->nym.name); /* bug */

  repblk->nHops = rb->nHops;

  for (i = 0; i < repblk->nHops; i++) {
    struct fr_maip_hop *privHop = &repblk->hop[i];
    struct fr_maip_rbpriv_hop_pk *bufHop = &rb->hop[i];
    
    if ( ! (frCryptSymNew(err, &privHop->encrSym) &&
            frCryptSymSetAlg(err, privHop->encrSym, bufHop->encrSymAlg) &&
            frCryptSymKeyFromData(err, privHop->encrSym, privHop->ksm.encrKeySeed,
                                  sizeof privHop->ksm.encrKeySeed) &&
            frCryptSymNew(err, &privHop->replySym) &&
            frCryptSymSetAlg(err, privHop->replySym, bufHop->replySymAlg) &&
            frCryptSymKeyFromData(err, privHop->replySym, bufHop->ksm.replyKeySeed,
                                  sizeof bufHop->ksm.replyKeySeed))) {
      goto done;
    }

    /*
     *  We need a MAC key only for the first hop.
     */
    
#if DEBUG_MAC
    if (i == 0) { fprintf(stderr, "frMaipRbPrivReadFromBuf: loading mac key\n"); }
#endif

    if (i == 0 && ! (frCryptMacKeyNew(err, &privHop->replyMacKey) &&
                     frCryptMacKeyLoad(err, privHop->replyMacKey,
                                       bufHop->ksm.replyMacKey,
                                       sizeof bufHop->ksm.replyMacKey))) {
      goto done;
    }

    privHop->maip.type = bufHop->maip[0]; /* bug */
    memcpy(privHop->maip.name, &bufHop->maip[1], sizeof privHop->maip.name); /* bug */

    privHop->port = ntohs(bufHop->port);
    privHop->latency = ntohl(bufHop->latency);
  }

  strncpy(repblk->email, rb->email, sizeof repblk->email);

  ret = 1;

 done:

  return ret;
}


/*
 *  Write a representation of the private reply block suitable for external
 *  storage.  This function will allocate the necessary memory, returning it
 *  and its size via `*rbuf' and `rsz' respectively.
 *  
 *  bug: just to be safe, we should check the value of `nHops'.
 */
DLLExport int frMaipRbPrivWriteToBuf(ZkError err, FrMaipRbPriv repblk, void **rbuf, size_t *rsz)
{
  int ret = 0;
  size_t i;
  struct fr_maip_rbpriv_pk *rb = 0;
  
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivWriteToBuf", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! rbuf) {
    zkErrorSet(err, "frMaipRbPrivWriteToBuf", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rbuf is null");
    return 0;
  }
  if ( ! rsz) {
    zkErrorSet(err, "frMaipRbPrivWriteToBuf", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rsz is null");
    return 0;
  }

  if ( ! (rb = malloc(sizeof *rb))) {
    zkErrorSet_va(err, "frMaipRbPrivWriteToBuf", ZKERR, ZKMFRMAIP, errno,
                  "can't allocate %lu bytes for reply block buffer: %s",
                  (unsigned long)sizeof *rb, strerror(errno));
    goto done;
  }
  
  memset(rb, 0, sizeof *rb);

  rb->vers = 1; /* bug */
  rb->type = FRPKTTYPE_REPLYBLOCK_PRIV;
  rb->nHops = repblk->nHops;
  strncpy(rb->rbName, repblk->rbName, sizeof rb->rbName);
  if ( ! frEntStore(err, &repblk->nym, rb->nym, sizeof rb->nym)) {
    goto done;
  }

  for (i = 0; i < repblk->nHops; i++) {
    FrCryptSymAlg encrSymAlg, replySymAlg;
    struct fr_maip_hop *hop = &repblk->hop[i];
    struct fr_maip_rbpriv_hop_pk *bufHop = &rb->hop[i];
    unsigned int epkv; /* encryption public key version */

    if ( ! (frCryptSymGetAlg(err, hop->encrSym, &encrSymAlg) &&
            frCryptSymGetAlg(err, hop->replySym, &replySymAlg) &&
            frCryptPubKeyGetVersion(err, hop->encrPubKey, &epkv))) {
      goto done;
    }
    
    bufHop->encrSymAlg = encrSymAlg;
    bufHop->replySymAlg = replySymAlg;
    bufHop->encrPubKeyVers = epkv;

    memcpy(&bufHop->ksm, &hop->ksm, sizeof bufHop->ksm);

    bufHop->maip[0] = hop->maip.type; /* bug */
    memcpy(&bufHop->maip[1], hop->maip.name, sizeof bufHop->maip - 1); /* bug */

    bufHop->port = htons(hop->port);
    bufHop->latency = htonl(hop->latency);
  }
  
  strncpy(rb->email, repblk->email, sizeof rb->email);

  *rbuf = rb; rb = 0; /* the caller now owns it */
  *rsz = sizeof *rb;
  
  ret = 1;

 done:

  if (rb) { memset(rb, 0, sizeof *rb); free(rb); }

  return ret;
}


/*
 *  Allocate memory into which `encr' is multiply decrypted, based on the
 *  symmetric keys in the reply block.  The amount of encrypted data is given
 *  by `esz'.  Upon successful decryption the cleartext will be returned via
 *  `*clr', while its size will be returned via `csz'.
 *  
 *  Each layer of encryption has a MAC appended to it.
 *  
 *  The decryptions are performed in the reverse order of the calls to
 *  frMaipRbPrivSetHop().  That is, the symmetric key corresponding to the
 *  last hop set will be the first applied to the data.
 */
DLLExport int frMaipRbPrivDecrypt(ZkError err, FrMaipRbPriv repblk, void *encr, size_t esz, void **clr, size_t *csz)
{
  char iv[FRCRYPT_IVEC_SIZE];
  int ret = 0;
  size_t i;
  void *c = 0; /* cleartext data */
    
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivDecrypt", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! encr) {
    zkErrorSet(err, "frMaipRbPrivDecrypt", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "encr is null");
    return 0;
  }
  if (esz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "frMaipRbPrivDecrypt", ZKERR, ZKMFRMAIP, ZKENULLPTR,
                  "esz (%lu) is not a multiple of %d",
                  (unsigned long)esz, FRCRYPT_SIZE_MULT);
    return 0;
  }
  if ( ! clr) {
    zkErrorSet(err, "frMaipRbPrivDecrypt", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "clr is null");
    return 0;
  }
  if ( ! csz) {
    zkErrorSet(err, "frMaipRbPrivDecrypt", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "csz is null");
    return 0;
  }

  if (repblk->nHops == 0) {
    zkErrorSet(err, "frMaipRbPrivDecrypt", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
               "no hops in reply block");
    return 0;
  }

  if ( ! (c = malloc(esz))) {
    zkErrorSet_va(err, "frMaipRbPrivDecrypt", ZKERR, ZKMFRMAIP, errno,
                  "can't allocate %lu bytes for cleartext: %s",
                  (unsigned long)esz, strerror(errno));
    goto done;
  }

  memcpy(c, encr, esz);

  {
    int valid;
    struct fr_maip_hop *h = 0; /* remove warning */

    for (i = 0; i < repblk->nHops; i++) {
      h = &repblk->hop[repblk->nHops-1-i];

      memset(iv, 0, sizeof iv);
      if ( ! frCryptSymDecrypt(err, h->replySym, iv, c, esz)) {
        goto done;
      }
    }

    if (esz < FRMAIP_MAC_LEN) {
      zkErrorSet_va(err, "frMaipRbPrivDecrypt", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                    "hop %lu data too short to contain MAC (%lu < %d)",
                    (unsigned long)i, (unsigned long)esz, FRMAIP_MAC_LEN);
      goto done;
    }

#if DEBUG_MAC
    fprintf(stderr, "frMaipRbPrivDecrypt: verifying mac...\n");
#endif

    if ( ! (frCryptMacVerifyFromBuffer(err, h->replyMacKey, c, esz - FRMAIP_MAC_LEN,
                                       (char *)c + esz - FRMAIP_MAC_LEN,
                                       FRMAIP_MAC_LEN, &valid))) {
      goto done;
    }
    
    if ( ! valid) {
      zkErrorSet_va(err, "frMaipRbPrivDecrypt", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                    "invalid MAC on hop %lu data", (unsigned long)i);
      goto done;
    }
  }
  
  
  *clr = c; c = 0; /* the caller now owns it */
  *csz = esz - FRMAIP_MAC_LEN;

  ret = 1;

 done:

  if (c) { memset(c, 0, esz); free(c); }

  return ret;
}


#pragma pack(1)

/*
 *  The external storage representation of each hop in the public reply block.
 *  
 *  Note that the layered encryption of the public reply block requires this
 *  structure, as a whole, and that part following the `rand' field, to be a
 *  multiple of 8 bytes in length.
 */
struct fr_maip_rbpub_hop_pk {
  uint8_t encrSymAlg;
  uint8_t encrPubKeyVers;
  char pad[6];
  char preKeySeed[FRCRYPT_SECRET_SIZE];
  char rand[8];
  uint8_t replySymAlg;
  char maip[FRENT_PORTABLE_SIZE];
  uint16_t port;
  uint32_t latency;
};

/*
 *  This structure defines the external storage representation of the public
 *  reply block.
 */
struct fr_maip_rbpub_pk {
  uint8_t vers;
  uint8_t type;
  struct fr_maip_rbpub_hop_pk hop[FRMAIP_HOPS_MAX];
  char email[FRMAIP_EMAIL_LEN];
};

#pragma pack()


/*
 *  Construct a public reply block from the internal, private one.
 *  
 *  This routine allocates the memory for the public reply block, returning it
 *  via `*rbpub'; it's size is returned via `rsz'.
 */
DLLExport int frMaipRbPrivMakePublic(ZkError err, FrMaipRbPriv repblk, void **rbpub, size_t *rsz)
{
  char iv[FRCRYPT_IVEC_SIZE];
  char ksm[FRCRYPT_SECRET_SIZE];
  int ret = 0;
  size_t i;
  size_t nHops;
  struct fr_maip_rbpub_pk *rbp = 0;

  zkErrorClear(err);
  
  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivMakePublic", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! rbpub) {
    zkErrorSet(err, "frMaipRbPrivMakePublic", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rbpub is null");
    return 0;
  }
  if ( ! rsz) {
    zkErrorSet(err, "frMaipRbPrivMakePublic", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rsz is null");
    return 0;
  }

  if (repblk->nHops == 0 || repblk->nHops > FRMAIP_HOPS_MAX) {
    zkErrorSet_va(err, "frMaipRbPrivMakePublic", ZKERR, ZKMFRMAIP, ZKEDATABAD,
                  "invalid number of hops (%u)", repblk->nHops);
    goto done;
  }
    
  if ( ! (rbp = malloc(sizeof *rbp))) {
    zkErrorSet_va(err, "frMaipRbPrivMakePublic", ZKERR, ZKMFRMAIP, errno,
                  "can't allocate public reply block: %s", strerror(errno));
    goto done;
  }

  rbp->vers = 1; /* bug */
  rbp->type = FRPKTTYPE_REPLYBLOCK_PUB;

  nHops = repblk->nHops;

  /*
   *  Put the email address just after the last hop.
   */

  memcpy(&rbp->hop[nHops], repblk->email, sizeof repblk->email);

  /*
   *  Fill the unused portion of the public reply block with random data.
   *  Using a fixed value would allow a MAIP to determine how many hops
   *  preceded it.
   */

  {
    char *p0 = (char *)&rbp->hop[nHops] + sizeof rbp->email;
    char *p1 = (char *)(rbp + 1);
    
    if ( ! zkRandGetData(err, p0, p1 - p0)) {
      goto done;
    }
  }
  
  /*
   *  For each hop, from last to first, set an encryption header and fill in
   *  the hop information.
   */
  
  for (i = 0; i < repblk->nHops; i++) {
    FrCryptSymAlg encrAlg, replyAlg;
    struct fr_maip_hop *privHop = &repblk->hop[nHops-1-i];
    struct fr_maip_keyseed_pk *ks;
    struct fr_maip_rbpub_hop_pk *pubHop = &rbp->hop[nHops-1-i];
    unsigned int epkv;
    
    ks = (void *)&ksm[2]; /* skip first two bytes */

    if ( ! (frCryptSymGetAlg(err, privHop->encrSym, &encrAlg) &&
            frCryptSymGetAlg(err, privHop->replySym, &replyAlg) &&
            frCryptPubKeyGetVersion(err, privHop->encrPubKey, &epkv) &&
            zkRandGetData(err, pubHop->preKeySeed, sizeof pubHop->preKeySeed) &&
            zkRandGetData(err, pubHop->rand, sizeof pubHop->rand) &&
            frCryptCryptSetPreSecret(err, privHop->encrPubKey,
                                     pubHop->preKeySeed, sizeof pubHop->preKeySeed,
                                     ksm, sizeof ksm) &&
            frCryptSymKeyFromData(err, privHop->encrSym,
                                  ks->encrKeySeed, sizeof ks->encrKeySeed) &&
            frCryptSymKeyFromData(err, privHop->replySym,
                                  ks->replyKeySeed, sizeof ks->replyKeySeed))) {
      goto done;
    }
    
#if DEBUG_MAC
    if (i == repblk->nHops - 1) {
      fprintf(stderr, "frMaipRbPrivMakePublic: loading a mac key.\n");
    }
#endif
    
    if (i == repblk->nHops - 1 &&
        ! frCryptMacKeyLoad(err, privHop->replyMacKey,
                            ks->replyMacKey, sizeof ks->replyMacKey)) {
      goto done;
    }

    /* bug: a bit ugly that we have to save this */
    memcpy(&privHop->ksm, ks, sizeof privHop->ksm);

    pubHop->encrSymAlg = encrAlg;
    pubHop->encrPubKeyVers = epkv;
    fr_CryptSetPad(pubHop->pad, sizeof pubHop->pad);
    pubHop->replySymAlg = replyAlg;
    pubHop->maip[0] = privHop->maip.type;
    memcpy(&pubHop->maip[1], privHop->maip.name, sizeof privHop->maip.name);
    pubHop->port = htons(privHop->port);
    pubHop->latency = htonl(privHop->latency);

    /*
     *  Encrypt from the `rand' field of the current hop until the end of the
     *  reply block.
     */

    memset(iv, 0, sizeof iv);
    {
      char *p0 = pubHop->rand, *p1 = (char *)(rbp + 1);

      if ( ! frCryptSymEncrypt(err, privHop->encrSym, iv, p0, p1 - p0)) {
        goto done;
      }
    }
  }

  *rbpub = rbp; rbp = 0; /* the caller now owns it */
  *rsz = sizeof *rbp;
  
  ret = 1;
  
 done:

  memset(iv, 0, sizeof iv);
  memset(ksm, 0, sizeof ksm);
  if (rbp) { memset(rbp, 0, sizeof *rbp); free(rbp); }

  return ret;
}


/*
 *  Get the version of the public key used to encrypt the first hop.
 *  
 *  We only care about the first hop, since frMaipRbPubGetNextHop() shift them
 *  to the beginning for us.
 */
DLLExport int frMaipRbPubGetHopCryptKeyVersion(ZkError err, void *rbpub, size_t rsz, unsigned int *vers)
{
  int ret = 0;
  struct fr_maip_rbpub_hop_pk *pubHop;
  struct fr_maip_rbpub_pk *rbp = rbpub;

  zkErrorClear(err);

  if ( ! rbpub) {
    zkErrorSet(err, "frMaipRbPubGetHopCryptKeyVersion", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rbpub is null");
    return 0;
  }
  if (rsz < sizeof *rbp) {
    zkErrorSet_va(err, "frMaipRbPubGetHopCryptKeyVersion", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "rsz is too small (%lu < %lu)",
                  (unsigned long)rsz, (unsigned long)sizeof *rbp);
    return 0;
  }
  if ( ! vers) {
    zkErrorSet(err, "frMaipRbPubGetHopCryptKeyVersion", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "vers is null");
    return 0;
  }

  /*
   *  Check the public reply block header.
   */

  if (rbp->vers != 1 /* bug */) {
    zkErrorSet_va(err, "frMaipRbPubGetHopCryptKeyVersion", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has incorrect version (%u)",
                  (unsigned int)rbp->vers);
    goto done;
  }
  if ( (rbp->type != FRPKTTYPE_REPLYBLOCK_PUB)
       && (rbp->type != FRPKTTYPE_REPLYBLOCK_PUB_INACTIVE) ){
    zkErrorSet_va(err, "frMaipRbPubGetHopCryptKeyVersion", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has an invalid type (%u)",
                  (unsigned int)rbp->type);
    goto done;
  }

  pubHop = &rbp->hop[0]; /* for convenience */

  *vers = pubHop->encrPubKeyVers;

  ret = 1;

 done:

  return ret;
}


/*
 *  Check if the reply block header indicates that the reply block is active.
 */
DLLExport int
frMaipRbPubIsActive(
  ZK_INOUT   ZkError        err,
  ZK_IN      void *         rbpub,
  ZK_IN      size_t         rsz,
  ZK_OUT     unsigned int * is_active)
{
  static const char * func = "frMaipRbPubIsActive";
  struct fr_maip_rbpub_pk *rbp = rbpub;
  int ret = 0;

  zkErrorClear(err);

  if ( ! rbpub) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rbpub is null");
    return 0;
  }
  if (rsz < sizeof *rbp) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "rsz is too small (%lu < %lu)",
                  (unsigned long)rsz, (unsigned long)sizeof *rbp);
    return 0;
  }
  if ( ! is_active) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "is_active is null");
    return 0;
  }

  (*is_active) = 0;

  /*
   *  Check the public reply block header.
   */

  if (rbp->vers != 1 /* bug */) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has incorrect version (%u)",
                  (unsigned int)rbp->vers);
    goto done;
  }

  if (rbp->type == FRPKTTYPE_REPLYBLOCK_PUB) {
    (*is_active) = 1;
  } else if (rbp->type != FRPKTTYPE_REPLYBLOCK_PUB_INACTIVE) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has an invalid type (%u)",
                  (unsigned int)rbp->type);
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 * Mark a reply block as active (i.e., as a normal public reply block)
 */
DLLExport int
frMaipRbPubActivate(
  ZK_INOUT   ZkError        err,
  ZK_IN      void *         rbpub,
  ZK_IN      size_t         rsz )
{
  static const char * func = "frMaipRbPubActivate";
  struct fr_maip_rbpub_pk *rbp = rbpub;
  int ret = 0;

  zkErrorClear(err);

  if ( ! rbpub) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rbpub is null");
    return 0;
  }
  if (rsz < sizeof *rbp) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "rsz is too small (%lu < %lu)",
                  (unsigned long)rsz, (unsigned long)sizeof *rbp);
    return 0;
  }

  /*
   *  Check the public reply block header.
   */

  if (rbp->vers != 1 /* bug */) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has incorrect version (%u)",
                  (unsigned int)rbp->vers);
    goto done;
  }

  if (rbp->type == FRPKTTYPE_REPLYBLOCK_PUB_INACTIVE) {
    rbp->type = FRPKTTYPE_REPLYBLOCK_PUB;
  } else if (rbp->type != FRPKTTYPE_REPLYBLOCK_PUB) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has an invalid type (%u)",
                  (unsigned int)rbp->type);
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 * Mark a reply block as active (i.e., as a normal public reply block)
 */
DLLExport int
frMaipRbPubDeactivate(
  ZK_INOUT   ZkError        err,
  ZK_IN      void *         rbpub,
  ZK_IN      size_t         rsz )
{
  static const char * func = "frMaipRbPubDeactivate";
  struct fr_maip_rbpub_pk *rbp = rbpub;
  int ret = 0;

  zkErrorClear(err);

  if ( ! rbpub) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rbpub is null");
    return 0;
  }
  if (rsz < sizeof *rbp) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "rsz is too small (%lu < %lu)",
                  (unsigned long)rsz, (unsigned long)sizeof *rbp);
    return 0;
  }

  /*
   *  Check the public reply block header.
   */

  if (rbp->vers != 1 /* bug */) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has incorrect version (%u)",
                  (unsigned int)rbp->vers);
    goto done;
  }

  if (rbp->type == FRPKTTYPE_REPLYBLOCK_PUB) {
    rbp->type = FRPKTTYPE_REPLYBLOCK_PUB_INACTIVE;
  } else if (rbp->type != FRPKTTYPE_REPLYBLOCK_PUB_INACTIVE) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has an invalid type (%u)",
                  (unsigned int)rbp->type);
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 *  Decrypt the body of the public reply block using the private key
 *  `maipPriv'.  Extract the next hop information, then `shift' the hop
 *  information.  If the current hop is that last one, set `*lastHop' to 1,
 *  otherwise set it to 0.
 *  
 *  If `firstHop' is true we also extract a MAC key.  It's up to the caller to
 *  indicate whether or not this is the first hop.
 *  
 *  
 *  Shifting the contents of the reply block will be reflected in the
 *  modification of the contents of `rbpub'.
 *  
 *  The caller is responsible for freeing the symmetric key in `nextHop->sym'.
 */
DLLExport int frMaipRbPubGetNextHop(ZkError err, void *rbpub, size_t rsz, FrCryptCrypt maipPriv, int firstHop, int *lastHop, struct frmaip_rbpub_hop *nextHop)
{
  FrCryptSym sym = 0;
  char iv[FRCRYPT_IVEC_SIZE];
  char ksm[FRCRYPT_SECRET_SIZE];
  int ret = 0;
  struct fr_maip_keyseed_pk *ks;
  struct fr_maip_rbpub_hop_pk *pubHop;
  struct fr_maip_rbpub_pk *rbp = rbpub;
  char *p0, *p1;

  zkErrorClear(err);

  if ( ! rbpub) {
    zkErrorSet(err, "frMaipRbPubGetNextHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rbpub is null");
    return 0;
  }
  if (rsz < sizeof *rbp) {
    zkErrorSet_va(err, "frMaipRbPubGetNextHop", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "rsz is too small (%lu < %lu)",
                  (unsigned long)rsz, (unsigned long)sizeof *rbp);
    return 0;
  }
  if ( ! maipPriv) {
    zkErrorSet(err, "frMaipRbPubGetNextHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "maipPriv is null");
    return 0;
  }
  if ( ! lastHop) {
    zkErrorSet(err, "frMaipRbPubGetNextHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "lastHop is null");
    return 0;
  }
  if ( ! nextHop) {
    zkErrorSet(err, "frMaipRbPubGetNextHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "nextHop is null");
    return 0;
  }

  /*
   *  Check the public reply block header.
   */

  if (rbp->vers != 1 /* bug */) {
    zkErrorSet_va(err, "frMaipRbPubGetNextHop", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has incorrect version (%u)",
                  (unsigned int)rbp->vers);
    goto done;
  }
  if ( (rbp->type != FRPKTTYPE_REPLYBLOCK_PUB)
       && (rbp->type != FRPKTTYPE_REPLYBLOCK_PUB_INACTIVE) ){
    zkErrorSet_va(err, "frMaipRbPubGetNextHop", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has an invalid type (%u)",
                  (unsigned int)rbp->type);
    goto done;
  }

  pubHop = &rbp->hop[0]; /* for convenience */
  ks = (void *)&ksm[2]; /* skip first two bytes */

  /*
   *  Decrypt the current hop.
   */

  memset(iv, 0, sizeof iv);

  if ( ! (frCryptCryptSecretFromPrivKey(err, maipPriv,
                                        pubHop->preKeySeed, sizeof pubHop->preKeySeed,
                                        ksm, sizeof ksm) &&
          frCryptSymNew(err, &sym) &&
          frCryptSymSetAlg(err, sym, pubHop->encrSymAlg) &&
          frCryptSymKeyFromData(err, sym, ks->encrKeySeed, sizeof ks->encrKeySeed) &&
          (p0 = pubHop->rand, p1 = (char *)(rbp + 1),
           frCryptSymDecrypt(err, sym, iv, p0, p1 - p0)))) {
    goto done;
  }

  /*
   *  Extract the hop information.
   */

  memset(nextHop, 0, sizeof *nextHop);
    
  if (pubHop->port == 0) {
    /*
     *  Fields particular to the last hop.
     */
    memcpy(nextHop->email, pubHop + 1, sizeof nextHop->email);
    *lastHop = 1;
  } else {
    /*
     *  Fields particular to intermediate hops.
     */
    nextHop->maip.type = pubHop->maip[0]; /* bug */
    memcpy(nextHop->maip.name, &pubHop->maip[1], sizeof nextHop->maip.name); /* bug */
    nextHop->port = ntohs(pubHop->port);
    *lastHop = 0;
  }

  /*
   *  Fields common to intermediate hops and the last hop.
   */
    
  if ( ! (frCryptSymNew(err, &nextHop->replySym) &&
          frCryptSymSetAlg(err, nextHop->replySym, pubHop->replySymAlg) &&
          frCryptSymKeyFromData(err, nextHop->replySym,
                                ks->replyKeySeed, sizeof ks->replyKeySeed))) {
    goto done;
  }

  /*
   *  Only at the first hop do we look for a MAC key.
   */
  
#if DEBUG_MAC
  if (firstHop) { fprintf(stderr, "frMaipRbPubGetNextHop: load a mac key\n"); }
#endif

  if (firstHop && ! (frCryptMacKeyNew(err, &nextHop->replyMacKey) &&
                     frCryptMacKeyLoad(err, nextHop->replyMacKey,
                                       ks->replyMacKey, sizeof ks->replyMacKey))) {
    goto done;
  }
            
  nextHop->latency = ntohl(pubHop->latency);

  /*
   *  Shift the next hop into position in the original reply block.
   *  
   *  Copy random data to the unused tail of the reply block?
   */

  if ( ! *lastHop) {
    char *e = (void *)(rbp + 1), *h0 = (void *)&rbp->hop[0], *h1 = (void *)&rbp->hop[1];
    size_t szDiff = h1 - h0;  /* size of block that we have to pad */
    size_t szBlockToMv = e - h1;
    /* move */
    memmove(&rbp->hop[0], h1, szBlockToMv);
    /* pad leftover space, that is starting from b + szBlockToMv untill the end */
    if ( ! zkRandGetData(err, (void *)(h0 + szBlockToMv), szDiff)) {
      goto done;
    };

  }




  ret = 1;
  
 done:

  if (sym) { frCryptSymFree(0, &sym); }
  memset(iv, 0, sizeof iv);
  memset(ksm, 0, sizeof ksm);

  return ret;
}


/*
 *  Perform any necessary clean-up on the hop structure returned set by
 *  frMaipRbPubGetNextHop().
 */
DLLExport int frMaipRbPubClearHop(ZkError err, struct frmaip_rbpub_hop *hop)
{
  zkErrorClear(err);

  if ( ! hop) {
    zkErrorSet(err, "frMaipRbPubClearHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "hop is null");
    return 0;
  }

  if (hop->replySym && ! frCryptSymFree(err, &hop->replySym)) {
    return 0;
  }
  if (hop->replyMacKey && ! frCryptMacKeyFree(0, &hop->replyMacKey)) {
    return 0;
  }

  memset(hop, 0, sizeof *hop);

  return 1;
}


DLLExport int frMaipRbPubHopEncrypt(ZkError err, struct frmaip_rbpub_hop *nextHop, int firstHop, void *clr, size_t csz, void **encr, size_t *esz)
{
  char iv[FRCRYPT_IVEC_SIZE];
  int ret = 0;
  size_t sz;
  void *e = 0;
  
  zkErrorClear(err);

  if ( ! nextHop) {
    zkErrorSet(err, "frMaipRbPubHopEncrypt", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "nextHop is null");
    return 0;
  }
  if ( ! clr) {
    zkErrorSet(err, "frMaipRbPubHopEncrypt", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "clr is null");
    return 0;
  }
  if (csz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "frMaipRbPubHopEncrypt", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "csz (%lu) is not a multiple of %d",
                  (unsigned long)csz, FRCRYPT_SIZE_MULT);
    return 0;
  }
  if ( ! encr) {
    zkErrorSet(err, "frMaipRbPubHopEncrypt", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "encr is null");
    return 0;
  }
  if ( ! esz) {
    zkErrorSet(err, "frMaipRbPubHopEncrypt", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "esz is null");
    return 0;
  }

  if (firstHop) {
    sz = csz + FRMAIP_MAC_LEN;
  } else {
    sz = csz;
  }

  if ( ! (e = malloc(sz))) {
    zkErrorSet_va(err, "frMaipRbPubHopEncrypt", ZKERR, ZKMFRMAIP, errno,
                  "can't allocate %lu bytes: %s", (unsigned long)sz, strerror(errno));
    goto done;
  }

  memcpy(e, clr, csz);
  memset(iv, 0, sizeof iv);
  
#if DEBUG_MAC
  fprintf(stderr, "frMaipRbPubHopEncrypt: 1st hop: %d, elen: %d\n",
          firstHop, csz);
#endif

  if ((firstHop && ! frCryptMacIntoBuffer(err, nextHop->replyMacKey,
                                          e, csz, (char *)e + csz, FRMAIP_MAC_LEN)) ||
      ! frCryptSymEncrypt(err, nextHop->replySym, iv, e, sz)) {
    goto done;
  }

  *encr = e; e = 0; /* the caller now owns it */
  *esz = sz;
  ret = 1;

 done:

  if (e) { memset(e, 0, sz); free(e); }

  return ret;
}
