/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/* InfoObjClient.c 
 * 
 * Collection of routines to implement a generic information server.
 * see also: InfoObj.c, InfoObjServer.c, InfoObjGetData.c 
 */

#include "InfoObj.h"


/*
** Client: Try to connect socket to server 
*/
int InfoObjClientConnect
(
 ZkError       err,                      /* [in]   ZkError */
 P_InfoObjMain info                      /* [in]   InfoObj */
)
{
  const char *func     = "InfoObjClientConnect";
  int ok = 1;

  /*
  ** Verify input pointers
  */
  if( !info )
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "info is null");
    ok = 0;
  }
  else
  {
    if (info->s) {
      zkSocketFree(err, &info->s);
    }

    if (!zkSocketNew( err, AF_UNIX, SOCK_STREAM, IPPROTO_IP, &info->s)) {
      ok = 0;
    } 
    else {
      ok = zkSocketConnect( err, info->s, info->InfoObjSocketName);
      if (ok) {
        info->connected = ZK_TRUE;
      }
    }
  }
   
  return ok; 
}

/*
** Client: disconnect socket to server 
*/

int InfoObjClientDisconnect
(
 ZkError       err,                      /* [in]   ZkError */
 P_InfoObjMain info                      /* [in]   InfoObj */
)
{
  const char *func     = "InfoObjClientConnect";
  int ok = 1;

  char command;

  /*
  ** Verify input pointers
  */
  if( !info )
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "info is null");
    ok = 0;
  }
  else
  {
    info->connected = ZK_FALSE;

    command = InfoObjCmdDisconnect;
    zkSocketWriteAll( err, info->s, &command, 1);
    zkSocketFree(err, &info->s);
  }
  return ok;
    
}

/*
** Client: return if client is connected
*/

int InfoObjClientConnected
(
 ZkError        err,                      /* [in]   ZkError */
 P_InfoObjMain  info,                     /* [in]   InfoObj */
 ZkBool*        connected                 /* [out]  is client connected */
)
{
  const char *func     = "InfoObjClientConnected";
  int ok = 1;

  /*
  ** Verify input pointers
  */
  if( !info || !connected)
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "parameter(s) are null");
    ok = 0;
  }
  else {
    *connected = info->connected;
  }
  return ok;
    
}

/*
** send command to server 
*/
int InfoObjClientSendCommand
(
 ZkError       err,                      /* [in]   pointer on a ZkError */
 P_InfoObjMain info,                     /* [in]   info obj */
 void          *buf                      /* [in]   command */
)
{
  const char *func = "InfoObjClientSendCommand";
  int ok = 1;

  /*
  ** Verify input pointers
  */
  if( !info || !buf)
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "parameter(s) are null");
    ok = 0;
  }
  else {
    ok = zkSocketWriteAll( err, info->s, buf, 1);
  }

  return ok;

}

/*
** send client request data to server
*/
int InfoObjClientSendData
(
 ZkError       err,                      /* [in]   pointer on a ZkError */
 P_InfoObjMain info,                     /* [in]   info obj */
 void          *buf,                     /* [in]   data */
 size_t        buflen                    /* [in]   no. of bytes of data */
)
{
  const char *func = "InfoObjClientSendData";
  int ok = 1;

  /*
  ** Verify input pointers
  */
  if( !info || !buf)
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "parameter(s) are null");
    ok = 0;
  }
  else {
    ok = zkSocketWriteAll( err, info->s, &buflen, sizeof( buflen));
    if (ok) {
      ok = zkSocketWriteAll( err, info->s, buf, buflen);
    }
  }

  return ok;

}

/*
**  receive data from server, in format: command, time, size, and data 
*
*   IMPORTANT NOTE: caller is reponsible for freeing allocated memory
*/
int InfoObjClientReceive
(
 ZkError       err,                      /* [in]   pointer on a ZkError */
 P_InfoObjMain info,                     /* [in]   info obj */
 void          *command,                 /* [out]  return command */
 void          **buf,                    /* [out]  return pointer to allocated buffer */    
 int           *size,                    /* [out]  return size */
 time_t        *tim                      /* [out]  return time */
)
{
  const char *func = "InfoObjClientReceive";
  int ok = 1;

  /*
  ** Verify input pointers
  */
  if( !info || !command || !buf || !size || !tim )
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "parameter(s) are null");
    ok = 0;
  }
  else {
    *buf = NULL;
    ok = zkSocketReadAll( err, info->s, command, 1);
    if (ok) {
      ok = zkSocketReadAll( err, info->s, tim, sizeof(*tim));
    }
    if (ok) {
      ok = zkSocketReadAll( err, info->s, size, sizeof(*size));
    }
    if (ok) {
      *buf = malloc( *size);
      if (*buf == NULL) {
        zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "cannot allocate data buffer");
        ok = 0;
      }
      else {
        ok = zkSocketReadAll( err, info->s, *buf, *size);
        if (!ok) {
          free(*buf);
          *buf = NULL;
	}
      }
    }
  }

  return ok;

}

/*
**  read data from cache file 
**
**  IMPORTANT NOTE: caller is responsible for freeing allocated buffer memory 
**
*/
int InfoObjClientReadCache
(
 ZkError       err,                      /* [in]   pointer on a ZkError */
 P_InfoObjMain info,                     /* [in]   info obj */
 char          command,                  /* [in]   command  */
 void          **buf,                    /* [out]  return pointer to allocated buffer */    
 int           *size,                    /* [out]  return size */
 time_t        *tim                      /* [out]  return time */
)
{
  const char *func = "InfoObjClientReadCache";
  int ok = 1;

  P_InfoObj infoobj;

  /*
  ** Verify input pointers
  */
  if( !info || !buf || !size || !tim)
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "parameter(s) are null");
    ok = 0;
  }
  else 
  {
    ok = InfoObjFindByCommand(err, info, command, &infoobj);
    if (ok) {
      ok = CacheObjRead( err, infoobj->cache, buf, size, tim);    
    }
  }

  return ok;

}

/*
** return FD of socket
*/
int InfoObjClientGetFD
(
 ZkError       err,                      /* [in]   pointer on a ZkError */
 P_InfoObjMain info,                     /* [in]   info obj */
 int           *fd                       /* [out]  return fd */
)
{
  const char *func = "InfoObjClientGetFD";
  int ok = 0;

  /*
  ** Verify input pointers
  */
  if( !info || !fd)
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "parameter(s) are null");
    ok = 0;
  }

  ok = zkSocketGetFD( err, info->s, fd);
   
  return ok;

}

