/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/*
 *  InfoObjGetData.c
 *
 *  collection of routines that implement "GetData" component of InfoObj server
 *
 *  see also: InfoObj.c, InfoObjServer.c, InfoObjClient.c
 *
 */

#include "zkserr.h"
#include "frNiqs.h"
#include "InfoObj.h"

/* 
 * InfoObjGetNetInfo - get netinfo data from niqs and store in buffer for
 *                     transmission to client
 *
 * @err       - ZkError code
 * @niqshost  - niqs host to connect to
 * @niqsport  - port to connect to niqs
 * @timeout   - timeout for connecting to niqs
 * @retbuf    - return pointer to allocated buffer of data
 * @retbufsize - return size of data
 *
 * created Mar. 2000, F. Foster
 */ 
int InfoObjGetNetInfo(
  ZkError     err,
  const char* niqsHost,
  uint16_t    niqsPort,
  uint32_t    timeout,
  char**      retbuf, 
  int*        retbufsize)
{
  DECL_FUNC("InfoObjGetNetInfo");

  static FrNetInfo ni = NULL;
  static FrKeyQry kq = NULL;
  
  ZkLog log = NULL;

  int ok = 1;


  if (!niqsHost || !retbuf || !retbufsize) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "parameter(s) are null");
    ok = 0;
  }    
  else if (!ni && !frNetInfoNew(err, &ni)) {
    ok = 0;
  }
  else if (!kq && !frKeyQryNew(err, &kq)) {
    ok = 0;
  }
  else if (!frNetInfoPopulateSequence(err, ni, 
				      niqsHost, niqsPort, timeout, 
				      0 /*nocache*/, kq)) {
    ok = 0;
  }
  else if (!frNetInfoStore(err, ni, (void *) retbuf, retbufsize)) {
    ok = 0;
  }

  /* Cleanup */

  if (!ok && retbuf) {
    if (*retbuf) {
      free(*retbuf);
    }
  }

  if (!ok) {
    if (zkErrorLog(err, &log)) {
      zkLogWriteContext(log, err, func, "failed");
    }
  }

  return ok;
}


/* 
 * InfoObjCreateNetInfo - construct FrNetInfo object from data buffer obtained from server
 *
 * @err       - ZkError code
 * @ni        - return FrNetInfo object
 * @kq        - Key Query object
 * @buf       - pointer to buffer of data
 * @bufsize   - size of data buffer
 *
 * created Mar. 2000, F. Foster
 */ 
int InfoObjCreateNetInfo(ZkError err, FrNetInfo *ni, FrKeyQry kq, char* buf, int bufsize)
{
  const char *func = "InfoObjCreateNetInfo";

  int ok = 1;

  if (!ni || !buf) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "parameter(s) are null");
    ok = 0;
  }    
  else if (!frNetInfoNew(err, ni)) {
    ok = 0;
  }
  else if (!frNetInfoLoad(err, *ni, kq, buf, bufsize)) {
    ok = 0;
  }

  return ok;

}


/* 
 * InfoObjGetTopology - get Topology data from niqs and store in buffer for
 *               transmission to client
 *
 * @err       - ZkError code
 * @niqshost  - niqs host to connect to
 * @niqsport  - port to connect to niqs
 * @timeout   - timeout for connecting to niqs
 * @entId     - data containing entity info to query NIQS for
 * @buf       - return pointer to allocated buffer of data
 * @size      - return size of data
 *
 * created Mar. 2000, F. Foster
 */ 
int InfoObjGetTopology(
  ZkError        err,
  const char*    niqshost,
  unsigned short niqsport,
  uint32_t       timeout,
  FrEntIdExp*    entId,
  char**         buf,
  int*           size)
{
  DECL_FUNC("InfoObjGetTopology");

  ZkipClient niqs = NULL;
  int ok = 1;
  ZkLog log = NULL;

  if (!niqshost || !entId || !buf || !size) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "parameter(s) are null");
    ok = 0;
  }    
  else if (!frNiqsConnect(err, &niqs, niqshost, niqsport, timeout)) {
    ok = 0;
  }
  else if (!frNiqsQuery(err, niqs, FRNIQ_ATTR_TOPOLOGY, 
			entId, (void **)buf, size)) {
    ok = 0;
  }

  if (niqs != NULL) {
    if (!frNiqsDisconnect(err, &niqs)) {
      if (zkErrorLog(err, &log)) {
        zkLogWriteContext(log, err, func, "frNiqsDisconnect failed");
      }
    }
  }

  if (!ok) {
    if (zkErrorLog(err, &log)) {
      zkLogWriteContext(log, err, func, "failed");
    }
  }

  return ok;
}


/* 
 * InfoObjGetACL     - get ACL data from niqs and store in buffer for
 *                     transmission to client
 *
 * @err       - ZkError code
 * @niqshost  - niqs host to connect to
 * @niqsport  - port to connect to niqs
 * @timeout   - timeout for connecting to niqs
 * @retbuf    - return pointer to allocated buffer of data
 * @retbufsize - return size of data
 *
 * created Mar. 2000, F. Foster
 */ 
int InfoObjGetACL(
  ZkError        err,
  const char*    niqshost,
  unsigned short niqsport,
  uint32_t       timeout,
  char**         buf,
  int*           size)
{
  DECL_FUNC("InfoObjGetACL");

  ZkipClient niqs = NULL;
  ZkLog log = NULL;

  int ok = 1;

  if (!niqshost || !buf || !size) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "parameter(s) are null");
    ok = 0;
  }    
  else if (!frNiqsConnect(err, &niqs, niqshost, niqsport, timeout)) {
    ok = 0;
  }
  else if (!frNiqsAccessCtrlList(err, niqs, ( void ** )buf, size)) {
    ok = 0;
  }

  if (niqs != NULL) {
    if( !frNiqsDisconnect(err, &niqs) ) {
      if (zkErrorLog(err, &log)) {
        zkLogWriteContext(log, err, func, "frNiqsDisconnect failed");
      }
    }
  }

  if (!ok) {
    if (zkErrorLog(err, &log)) {
      zkLogWriteContext(log, err, func, "failed");
    }
  }

  return ok;

}




