/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/* InfoObjServer.c
 *
 * Collection of routines to implement a generic information server.
 * see also: InfoObj.c, InfoObjClient.c, InfoObjGetData.c 
 */


#include "zkserr.h"  /* For DECL_FUNC */
#include "InfoObj.h"



static volatile sig_atomic_t running = 1;
static volatile sig_atomic_t sigchild = 0;
static volatile sig_atomic_t logReload = 0;

static void
signalHandler ( 
  int sig )
{
  switch (sig) {
  case SIGPIPE:
    break;
  case SIGHUP:
    break;
  case SIGURG:
    break;
  case SIGUSR1:
    break;

  case SIGCHLD:
    sigchild++;
    break;

  case SIGUSR2:
    logReload = 1;
    break;

  case SIGTERM:
  case SIGINT:
    running = 0;
  default:                      /* All other signals should terminate */
    break;
  }
  return;
}




/*
** Server Get Data  
*/
static int InfoObjServerGetData
(
 ZkError      err,                      /* [in]   pointer on a ZkError */
 char         *name,                    /* [in]   data name */
 P_InfoObj    info,                     /* [in]   info obj */
 T_NiqsInfo   *niqsInfo,                /* [in]   niqs info */
 T_InfoObjBuf *buf                      /* [out]  data buffer */
)
{
  DECL_FUNC("InfoObjServerGetData");
  int ok = 1;

  /*
  ** Verify input pointers
  */

  /* #define TESTING */ 

  if( !info || !name || !niqsInfo || !buf)
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "parameter(s) are null");
    ok = 0;
  }
  else
  {
    buf->buffer = NULL;

      /* call actual routine to get data */
      
    if (strcmp(name,INFO_OBJ_NETINFO) == 0) 
    {
#ifdef TESTING
#warning "using simNetInfo"
      simNetInfo(err, niqsInfo->niqshost, niqsInfo->niqsport, niqsInfo->timeout, 
                        &buf->buffer, &buf->size);
#else
      InfoObjGetNetInfo(err, niqsInfo->niqshost, niqsInfo->niqsport, niqsInfo->timeout, 
                        &buf->buffer, &buf->size);
#endif
    }
    else if (strcmp(name,INFO_OBJ_TOPOLOGY) == 0) 
    {
#ifdef TESTING
      /* for debugging */
#warning "using simTopology"
      simTopology(err, niqsInfo->niqshost, niqsInfo->niqsport, niqsInfo->timeout, 
		  info->clientdata.u.entId,
		  &buf->buffer, &buf->size);
#else
      InfoObjGetTopology(err, niqsInfo->niqshost, niqsInfo->niqsport, niqsInfo->timeout, 
                         &(info->clientdata),
                         &buf->buffer, &buf->size);
#endif      
    }
    else if (strcmp(name,INFO_OBJ_ACL) == 0) 
    {
#ifdef TESTING
      /* for debugging */
#warning "using simACL"
      simACL(err, niqsInfo->niqshost, niqsInfo->niqsport, niqsInfo->timeout, 
                    &buf->buffer, &buf->size);
#else
      InfoObjGetACL(err, niqsInfo->niqshost, niqsInfo->niqsport, niqsInfo->timeout, 
                    &buf->buffer, &buf->size);
#endif
    }
    else {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "name not recognized");
      ok = 0;
    }

    if (buf->size == 0) {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "buffer loaded is empty");
      ok = 0;
    }

    if (ok) {

      buf->time = time(NULL);

      if (!CacheObjWrite( err, info->cache, buf->buffer, buf->size)) {
        ok = 0;
      }  
    }
  }

  return ok;

}


/*
** Process Getdata  
*/
static void InfoObjServerProcessGetData
(
 P_InfoClientCtx  clientCtx                     /* [out]  pointer to InfoCLientCtx */
)
{
  DECL_FUNC("InfoObjServerProcessGetData");
  int ok = 1;


  if (clientCtx) {
    T_InfoObjBuf buf;
    ZkLog log;

    zkErrorLog(clientCtx->err, &log);

    buf.buffer = NULL;

    while (ok && running)
    {
      if (InfoObjServerGetData(clientCtx->err, 
                               clientCtx->infoobj->cache->idname, 
                               clientCtx->infoobj, clientCtx->niqsInfo,
			       &buf)) {
	zkLogWrite_va(log, "%s:  \"%s\" reloaded", 
		      func,
                      clientCtx->infoobj->cache->idname);
      }
      else {
	/* Why is this disabled? */
        /* zkLogWriteContext(log, clientCtx->err, func, "reload failed"); */
      }

      if (buf.buffer) {
        free(buf.buffer);
        buf.buffer = NULL;
      }
      if (clientCtx->infoobj->reloadtime <= 0) {
	break;
      }

      sleep( clientCtx->infoobj->reloadtime);
    }
  }

  return;
}




/*
** Server Process Event
*/
static int InfoObjServerProcessEvent
(
 ZkError             err,                    /* [in]   pointer to ZkError     */
 P_InfoClientCtx     clientCtx               /* [in]   info client ctx */
)
{
  const char *func = "InfoObjServerProcessEvent";
  int ok = 1;

  int gotdata;

  T_InfoObjBuf buf;

  (void)func;

  /*
  ** Verify input pointers
  */
  if( !clientCtx )
  {
    ok = 0;
  }
  else {
    switch (clientCtx->command) {

      /* get topology request data from client */

    case InfoObjCmdSendTopologyData: 
      {
	FrEntIdExp  newEnt;
	size_t      newSize;

	if (!zkSocketReadAll(err, clientCtx->c, &newSize, sizeof(newSize)) ||
	    (sizeof(newEnt) != newSize)) {
	  ok = 0;
	}
	else if (!zkSocketReadAll(err, clientCtx->c, &newEnt, sizeof(newEnt))) {
	  ok = 0;
	}
	else {
	  memcpy(&(clientCtx->infoobj->clientdata), &newEnt, sizeof(newEnt));
	}
      } 
      break;

    case InfoObjCmdUpdateTopology: 
    case InfoObjCmdUpdateACL:  
    case InfoObjCmdUpdateNetInfo:
      /* only makes sense to do this if auto-periodic updates are disabled */  
      if (clientCtx->infoobj->reloadtime == 0) {
        ok = InfoObjServerGetData(err, clientCtx->infoobj->cache->idname, 
				  clientCtx->infoobj, clientCtx->niqsInfo, &buf);
        if (buf.buffer) { 
          free(buf.buffer);
	}
      }
      break;

    case InfoObjCmdGetTopology: 
    case InfoObjCmdGetACL: 
    case InfoObjCmdGetNetInfo:
      /* if auto-periodic updates of data are enabled, try to get data from cache */ 
      gotdata = 0;
      if (clientCtx->infoobj->reloadtime > 0) {
        gotdata = CacheObjRead(err, clientCtx->infoobj->cache, 
			       (void **) &buf.buffer, &buf.size, &buf.time);
        
      }
      /* if not, get data right now */
      if (!gotdata) {
        gotdata = InfoObjServerGetData(err, clientCtx->infoobj->cache->idname, 
				       clientCtx->infoobj, clientCtx->niqsInfo, 
				       &buf);
      }  

      /* send data to client, including: command, time, size and actual data */
      if (gotdata) {
        if (!zkSocketWriteAll(err, clientCtx->c, 
                               &clientCtx->command, sizeof(clientCtx->command))) {
          ok = 0;
        }
        else if (!zkSocketWriteAll(err, clientCtx->c, 
				   &buf.time, sizeof(buf.time))) 
        {
          ok = 0;
        }
        else if (!zkSocketWriteAll(err, clientCtx->c, 
				   &buf.size, sizeof(buf.size))) {
          ok = 0;
        }
        else if (!zkSocketWriteAll(err, clientCtx->c, 
				   buf.buffer, buf.size)) {
          ok = 0;
        }

        free(buf.buffer);
      } 
      break;
    }
  }

  return ok;
}


/*
** Server Event (read client command)
*/
static int InfoObjServerEvent
(
  ZkError       err, 
  P_InfoObjMain info,
  ZkSocket      c
)
{
  const char *func = "InfoObjServerEvent";

  int ok = 1;

  P_InfoObj       infoobj;
  char            command;

  /* get command from client */

  if( !info)
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "parameter(s) are null");
    ok = 0;
  }
  else if (!zkSocketReadAll( err, c, &command, 1)) {
    ok = 0;
  }
  else if (command == InfoObjCmdDisconnect) {
    ok = 0;
  }
  else if (!InfoObjFindByCommand(err, info, command, &infoobj)) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "unrecognized command");
    ok = 0;
  } 
  else {
    /* process command */
    P_InfoClientCtx clientCtx = malloc(sizeof (*clientCtx));

    if (!clientCtx) {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "error in allocating memory ");
      ok = 0;
    }
    else
    { 
      clientCtx->infoobj = infoobj;
      clientCtx->command = command; 
      clientCtx->c = c;
      clientCtx->niqsInfo = &info->niqsInfo;

      ok = InfoObjServerProcessEvent(err, clientCtx);
      free(clientCtx);
    }
  }
  
  return ok;
}



/*
** Start server data update child 
*/

int InfoObjServerDataUpdateChild
(
  ZkError       err,                      /* [in]  ZkError */
  P_InfoObjMain info,                     /* [in]  pointer to InfoObj */
  const char*   name,                     /* [in]  name of infoobj */
  time_t        reloadtime                /* [in]  time to auto do reload */
)
{
  DECL_FUNC("InfoObjSErverDataUpdateChild");

  P_InfoObj infoobj;

  int ok = 1;

  /*
  ** Verify input pointers
  */
  if (!info || !name)
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "parameter(s) are null");
    ok = 0;
  }
  else if (!InfoObjFindByName(err, info->InfoObjList, name, &infoobj)) 
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "cant find name");
    ok = 0;
  }
  else {
    infoobj->reloadtime = reloadtime;

    if (info->server && (infoobj->reloadtime > 0)) {
      ZkLog log;

      zkLogWrite_va(log, "%s:  Forking reload child \"%s\".", 
		    func, name);

      switch ((infoobj->reloadChildPid = fork())) {
      case -1: /* Error */
	zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, 
		   "Forking error.");
	ok = 0;
	break;

      case 0:  /* Child */
	{
	  P_InfoClientCtx clientCtx = malloc(sizeof (*clientCtx));

	  if (!clientCtx) {
	    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "error in allocating memory");
	    ok = 0;
	  }
	  else {
	    clientCtx->err      = err;
	    clientCtx->infoobj  = infoobj;
	    clientCtx->niqsInfo = &info->niqsInfo;

	    zkErrorLog(clientCtx->err, &log);

	    InfoObjServerProcessGetData(clientCtx);
	    free(clientCtx);
	    clientCtx = NULL;
	  }

	  exit(ok);
	}
	break;

      default: /* Parent */
	break;
      }
    }
  }

  return ok;
}



/*
** Server:  Set the termination signals.  
** System calls are _not_ restarted.
*/

int InfoObjServerSignalHandler(void)
{
  zkSignalIntr(SIGINT, signalHandler);
  zkSignalIntr(SIGTERM,signalHandler);
  zkSignalIntr(SIGPIPE,signalHandler);
  zkSignalIntr(SIGCHLD,signalHandler);
  zkSignalIntr(SIGUSR2,signalHandler);

  return 1;
}


/*
** Server: set NIQS info
*/

int InfoObjServerSetNiqsInfo
(
  ZkError        err,                      /* [in]   ZkError */
  P_InfoObjMain  info,                     /* [in]   InfoObj */
  const char*    niqshost,
  unsigned short niqsport,
  uint32_t       timeout
)
{
  DECL_FUNC("InfoObjServerSetNiqsInfo");

  int ok = 1;  

  /*
  ** Verify input pointers
  */
  if (!info || !niqshost)
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "parameter(s) are null");
    ok = 0;
  }
  else 
  {
    info->niqsInfo.niqshost    = niqshost;
    info->niqsInfo.niqsport    = niqsport;
    info->niqsInfo.timeout     = timeout;
  }
  return ok;
    
}



/*
** Server Wait For Connect 
*/
int InfoObjServerWaitForConnect
(
  ZkError        err,                     /* [in]   pointer on a ZkError */
  const char*    logfile, 
  P_InfoObjMain  info                     /* [in]   info obj */
)
{
  DECL_FUNC("InfoObjServerWaitForConnect");

  int ok = 1;

  P_InfoObjClient client;
  P_InfoObjClient prev;

  P_InfoObj infoobj;

  fd_set rset;
  int    nsel;

  int status;
  int found;
  pid_t reaped;

  ZkLog log;

  /*
  ** Verify input pointers
  */

  if( !info )
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "info is null");
    ok = 0;
  }

  zkErrorLog(err, &log);

  while (ok && running) 
  {

    /* wait for something to happen */

    rset = info->fdset;
    nsel = select( info->maxfd, &rset, NULL, NULL, NULL);

    if (!running) break;

    /* restart the Data Update Children if they stopped */ 
    while (sigchild && ((reaped = waitpid(-1, &status, WNOHANG)) > 0)) {
      /* Decrease the child count */
      sigchild--;
      infoobj = info->InfoObjList;
      found = 0;
      while (!found && infoobj)
      {
        if (reaped == infoobj->reloadChildPid) {
          InfoObjServerDataUpdateChild( err, info, infoobj->cache->idname, 
                                        infoobj->reloadtime);
          found = 1;
	}
        infoobj = infoobj->next;
      }
    }

    if (logReload) {            /* Rotate the log file */
      zkLogWrite(log, "Caught signal, Reopening logfile.");
      zkLogReopenFile(log, logfile);
      logReload = 0;
    }

    if (nsel > 0) {

      if (FD_ISSET( info->sfd, &rset))
      {

        /* got a connect request */

        client = malloc( sizeof (*client));
        if (!client) {
          zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, 
                     "error in allocating memory ");
          ok = 0;
          break;
        }           
        else if (!zkSocketAccept( err, info->s, &client->c)) {
          ok = 0;
          free(client);
          break;
        }
        else if (!InfoObjFDset( err, client->c, &client->cfd, &info->fdset)) {
          ok = 0;
          free(client);
          break;
        }
        else {

          /* add client socket to our list */

          prev = info->InfoObjClientList;
          client->next = prev;
          info->InfoObjClientList = client;
          InfoObjSetMaxFD( err, info);
	}
      }

      prev = NULL;
      client = info->InfoObjClientList;
      while (client)
      {
        if (FD_ISSET( client->cfd, &rset))
	{

          /* got a client request for service */

          if (!InfoObjServerEvent( err, info, client->c))
	  {
      
            /* if there was an error (or got Disconnect command)
               free socket and remove from client list */

            FD_CLR( client->cfd, &info->fdset);
            zkSocketFree(err, &client->c);
            if (prev) { 
              prev->next = client->next;
              free(client);
              client = prev->next;
            }
            else {
              info->InfoObjClientList = client->next;
              free(client);
              client = info->InfoObjClientList;
            }
            InfoObjSetMaxFD( err, info);

	  }
          else {
            client = client->next;
	  }            
	}
        else {
          client = client->next;
	}
      } 
    }
  }

  /* cleanup children when exiting */

  infoobj = info->InfoObjList;

  while (infoobj)
  {
    if (infoobj->reloadChildPid) {
      kill(infoobj->reloadChildPid, SIGTERM);
      waitpid(-1, &status, WNOHANG);
    }
    infoobj = infoobj->next;
  }

  return ok;

}

