/*
 * FIPS10 interface
 *
 * $Id: frNidbFips10.c,v 1.4.28.4 2000/10/11 13:18:52 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zks.net>

 */

#include "zks.h"

#include <stdlib.h>
#include "fips.h"
#include "zkCStr.h"
#include "zkError.h"
#include "frNidbObj.h"


#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNIDB

/**
 * Description: frNidbVerifyFip10CountryRegionAbbreviation takes
 *              a Fip10 country region code in parameter and verifies
 *              if it exists in the FIPS10 list. (*pExists) will be set
 *              to 1 if the abbreviation exists in the table. Only 'abbrev_sz'
 *              caracters will be checked.
 *
 * Parameters: - [ZkError] error: current ZK error object
 *             - [const char *] abbrev: Abbreviation string to be verified
 *             - [size_t] abbrev_sz: number of caracters in abbrev (should be 2 or 4)
 *             - [int *] pExists: Pointer to the exist flag
 */
DLLExport
int frNidbVerifyFip10CountryRegionAbbreviation (
  ZK_INOUT ZkError err,
  ZK_IN    const char * abbrev,
  ZK_IN    size_t abbrev_sz,
  ZK_OUT   int *pExists )
{
  static const char *funcname = "frNidbVerifyFip10CountryRegionAbbreviation";
  struct FrFips10_T *pFips10 = NULL;

  if ( !abbrev || !pExists ) /* Check parameters for null pointers */
  {
    zkErrorSet( err, funcname, ZKERR, ZKMOD, ZKENULLPTR, "Null parameter passed" );
    return 0;
  }

  *pExists = 0; /* Does not exist by default */
  pFips10 = Fips10List;
  
  while( pFips10->code && (strcmp(pFips10->code,"")!=0) /* Last element of the table */ )
  {
    if( strncasecmp(abbrev,pFips10->code,abbrev_sz)==0 )
      *pExists = 1;

    pFips10++;
  }

  return 1;
}


/**
 * Finds the representation of the region in the
 * FIPS10 list. Countryregion points to a dynamically allocated
 * piece of memory that has to be freed.
 */
DLLExport
int frNidbLocateFip10CountryRegion (
  ZK_INOUT ZkError err,
  ZK_IN    const char * abbrev,
  ZK_OUT   char **countryregion)
{
  static const char *funcname = "frNidbLocateFip10CountryRegion";
  struct FrFips10_T *fips10_p = NULL;
  const char *country = NULL;
  size_t len = 0;
  char *p = NULL;

  static const char *unknownReg = "(unknown region)";

  if ( !abbrev || ! countryregion ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  fips10_p = Fips10List;
  
  while (fips10_p->code && (strncasecmp (abbrev, fips10_p->code, 2) != 0) ) {
    fips10_p ++;
  }
  if ( fips10_p->code ) {
    country = fips10_p->name;
    while (fips10_p->code && (strncasecmp (abbrev, fips10_p->code, 4) > 0) ) {
      fips10_p ++;
    }
    if ( fips10_p->code ) {
      if ( ( fips10_p->name != country ) &&
           (strncasecmp (abbrev, fips10_p->code, 2) == 0) ) {
        if ( (len = zkCStrSprintf(&p, "%s, %s", fips10_p->name, country )) == 0 ) {
          zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENOMEM,"Error in zkCStrSprintf()");
          return 0;
        }
      } else { 
        if ( (len = zkCStrSprintf(&p, "%s", country)) == 0 ) {
          zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENOMEM,"Error in zkCStrSprintf()");
          return 0;
        }
      }
    } else { 
      if ( (len = zkCStrSprintf(&p, "%s", country)) == 0 ) {
        zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENOMEM,"Error in zkCStrSprintf()");
        return 0;
      }
    }
  } else {
    zkErrorSet_va(err,funcname,ZKWARN,ZKMOD,ZKENOTFOUND,
               "Unable to locate region string for %s", abbrev);
  }
  if (len >= 1024) {
    zkErrorSet(err,funcname,ZKWARN,ZKMOD,ZKENOTFOUND,
               "Internal buffer overflow - truncating output");
  }


  /*
  ** Give a warning string in case the region code is invalid.
  */
  if(!p) {
      if ( (len = zkCStrSprintf(&p, "%s", unknownReg)) == 0 ) {
        zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENOMEM,"Error in zkCStrSprintf()");
        return 0;
      }
  }

  *countryregion = p;

  return 1;
}      

