/*
 * $Id: frNiqs.c,v 1.25.8.4 2000/12/14 17:53:03 jean Exp $
 */

static const char * module_id_ = "$Id: frNiqs.c,v 1.25.8.4 2000/12/14 17:53:03 jean Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */
#include <string.h>
#include <zkError.h>
#include <zkCStr.h>
#include <ZKIP.h>
#include <zkipClient.h>
#include <frNiqs.h>
#include "frNiqsTokenInfo.h"
#include "frCrypt.h"
#include "frSec.h"

/*
 * For generating ZKIP command strings
 */

#define CMD_SZ 512

/*
 * The command type which controls the behaviour of frNiqsListOrStatus()
 */

typedef enum {
  FrNiqsCommandList ,          /* 7.1.0 list command for entities */
  FrNiqsCommandStatus ,        /* 7.1.0 status commands to see what's up doc */
  FrNiqsCommandAvailable       /* 7.1.0 available command (list active ents) */
} FrNiqsCommand;

/*
 * Internal function prototypes
 */

static int frNiqsListOrStatus ( ZkError , ZkipClient , FrNiqsCommand , time_t ,
                                FrEntIdType * , size_t , void ** , size_t *);
/*
 *
 */

DLLExport int
frNiqsConnect (
  ZkError err ,
  ZkipClient * client ,
  const char * host ,
  uint16_t port ,
  uint16_t timeout )
{
  static const char * func = "frNiqConnect";
  struct ZkipCommandResponse * response = NULL;
  char * extra_data = NULL;
  int return_code = 0;
  int fd = 0;

  (void) module_id_; /* Gets rid of compile time warning */

  if (host && client) {
    *client = NULL;
    if (zkipClientNew(err,client) &&
        zkipClientMakeBlocking(err,*client,timeout) &&
        zkipClientConnectToHost(err,*client,host,port) &&
        zkipClientFd(err,*client,&fd)) {
      
      /*
       * Succeeded in connecting, now check the server banner and make sure
       * we can do business...
       */

      if (!zkipClientReadResponse(err,*client,&response,&extra_data)) {
        zkErrorSet_va(err,func,ZKERR,ZKMNIQS,ZKERROR,
                      "Unable to read welcome banner from %s:%d",host,port);
      } else if (response->response != _SERVICE_READY) {
        zkErrorSet_va(err,func,ZKERR,ZKMNIQS,ZKERROR,
                      "Connection refused by %s:%d: %s",host,port,
                      response->message);
      } else {
        return_code = 1;
      }
      
      if (return_code != 1) {
        zkipClientDisconnect(NULL,*client);
        zkipClientFree(NULL,client);
        *client = NULL;
      }
    } else if (*client != NULL) {
      zkipClientFree(NULL,client);
      *client = NULL;
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,NULL);
  }
  return return_code;
}

/**
 *  When disconnecting a ZKIP client, a few things could go wrong such as
 *  a server crashing, quit command not being recognized, etc etc (of
 *  course these should almost never happen).  In such an event the function
 *  will return a 0 and set the ZkError, however the ZkipClient will still 
 *  be disconnected and the memory for that structure freed().
 */

DLLExport int
frNiqsDisconnect (
  ZkError err ,
  ZkipClient * client )
{
  char quit_command[] = "0 0.1.0 quit nodata binary 0\015\012";
  static const char * func = "frNiqDisconnect";
  struct ZkipCommandResponse * response = NULL;
  char * extra_data = NULL;
  int return_code = 0;


  if (client) {
    if (*client) {
      return_code = 1;
      if (!zkipClientWrite(err,*client,quit_command,strlen(quit_command))) {
        return_code = 0;
      }
      if (!zkipClientReadResponse(err,*client,&response,&extra_data)) {
        return_code = 0;
      }
      if (!zkipClientDisconnect(err,*client)) { 
        return_code = 0;
      }
      zkipClientFree(NULL,client);
      *client = NULL;
    } else {
      zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,
                 "ZkipClient pointer dereferences to NULL");
    }
  } else {
     zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,NULL);
  }
  return return_code;
}

/**
 *  Send a command string to the NIQS server and get back the signed response
 *  buffer.
 *
 *  @param err Return the error if one occured
 *  @param client The ZKIP client structure which is managing the session
 *  @param cmd A NUL terminated command string
 *  @param data Allocate space for and returns the response data
 *  @param sz Returns the size of the response buffer
 *  @version 1.0
 *  @author Philippe Boucher
 *  @returns Zero on failure, non-zero on success
 */

DLLExport int
frNiqsCommand (
  ZkError err ,
  ZkipClient client ,
  char * cmd ,
  void ** data ,
  size_t * sz )
{
  static const char * func = "frNiqsCommand";
  int return_code = 0;
  struct ZkipCommandResponse * response = NULL;
  char * extra_data = NULL;


  if (client && cmd && data && sz) {
    if (!zkipClientWrite(err,client,cmd,strlen(cmd))) {
      /* Failed to write data, error is already set by function call */ 
    } else if (!zkipClientReadResponse(err,client,&response,&extra_data)) {
      /* Failed to read response, error is already set by function call */ 
    } else if (response->response != _SERVICE_READY) {
      zkErrorSet_va(err,func,ZKERR,ZKMNIQS,ZKERROR,"Query failed:  [%d] [%s] cmd=>[%s]",
                    response->response,response->message,cmd); 
    } else {
      *sz = (size_t) response->datalen;
      if ((extra_data) && (*sz)) {
        *data = malloc(*sz + 1);
        memcpy(*data,extra_data,*sz);
        ((char *)*data)[*sz] = (char) 0;
      }
      return_code = 1;
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,NULL);
  }
  return return_code;
}

/** 
 * Return a signed buffer containing a ZkConfig parsable set of information
 * for the specified client.  This library does not specify the parameter
 * names or what the values mean.  That is entirely up to the client.
 * You must strip the signature before trying to parse the buffer using
 * zkConfigLoadromBuffer()
 *
 * @param err Return the error if one occured
 * @param client The ZKIP client connected to an NIQS
 * @param clientid A string token identifying the client type
 * @param data Allocates and returns a signed buffer containing the client info.
 *             You must free this when it is no longer needed.
 * @param sz Returns the size of the data buffer
 * @version 1.0
 * @author Philippe Boucher (philippe@zks.net)
 * @return A non-zero value on success, a 0 on failure
 *
 */

DLLExport int
frNiqsClientInfo (
  ZkError err ,
  ZkipClient client ,
  const char * clientid ,
  void ** data ,
  size_t * sz ) 
{
  static const char * func = "frNiqsClientInfo";
  int return_code = 0;
  char cmd[CMD_SZ] = "";


  if (client && clientid && data && sz) {
    if (snprintf(cmd,CMD_SZ,"1 7.1.0 clientinfo nodata binary 0 %s\015\012",
                 clientid) < 0) {
      zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,
                 "Unable to formulate 7.1.0 clientinfo command");
    } else {
      return_code = frNiqsCommand(err,client,cmd,data,sz);
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,NULL);
  }
  return return_code;
}


/** 
 * Return a signed buffer containing a ZkConfig parsable set of the access 
 * control list.  This library does not specify the parameter
 * names or what the values mean.  That is entirely up to the client.
 * You must strip the signature before trying to parse the buffer using
 * zkConfigLoadromBuffer()
 *
 * @param err Return the error if one occured
 * @param client The ZKIP client connected to an NIQS
 * @param data Allocates and returns a signed buffer containing the access
 *              control list.
 *             You must free this when it is no longer needed.
 * @param sz Returns the size of the data buffer
 * @version 1.0
 * @author Philippe Desaulniers (filou@zks.net)
 * @return A non-zero value on success, a 0 on failure
 *
 */

DLLExport int
frNiqsAccessCtrlList (
  ZkError err ,
  ZkipClient client ,
  void ** data ,
  size_t * sz ) 
{
  static const char * func = "frNiqsAccessCtrlList";
  int return_code = 0;
  char cmd[CMD_SZ] = "";


  if (client && data && sz) {
    if (snprintf(cmd,CMD_SZ,"1 7.1.0 accessctrllist nodata binary 0 \015\012") < 0) {
      zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,
                 "Unable to formulate 7.1.0 accessctrllist command");
    } else {
      return_code = frNiqsCommand(err,client,cmd,data,sz);
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,NULL);
  }
  return return_code;
}

/*
 *
 */

DLLExport int
frNiqsCacheDataVersion (
  ZkError err ,
  ZkipClient client ,
  void ** data ,
  size_t * sz )
{
  static const char * func = "frNiqsCacheDataVersion";
  int ret = 0;
  char cmd[CMD_SZ] = "";

  if (client && data && sz) {
    if (snprintf(cmd,CMD_SZ,"1 7.1.0 cacheversion nodata binary 0\015\012") < 0) {
      zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,
                 "Unable to formulate 7.1.0 cacheversion command");
    } else {
      ret = frNiqsCommand(err,client,cmd,data,sz);
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,NULL);
  }
  return ret;
}

/*
 *
 */

DLLExport int
frNiqsGetCacheDataVersion (
  ZkError err ,
  ZkipClient client ,
  FrKeyQryHost kh ,
  FrKeyQry kq ,
  uint32_t * cacheDataVersion )
{
  static const char * func = "frNiqsGetCacheDataVersion";
  int ret = 0;
  FrEntIdExp signer;
  FrCryptPubKey sigkey = NULL;
  unsigned int kv = 0; /* Key version */
  unsigned char * buf = NULL;        /* Buffer from the NIQS */
  size_t sz = 0;
  int valid = 0;
  int retries = 5;
  int timeout = 3;

  if (client && kq && cacheDataVersion) {
    *cacheDataVersion = 0;
    if (frNiqsCacheDataVersion(err,client,(void **) &buf,&sz) &&
        frSecSigGetEntity(err,buf,sz,&signer) &&
        frSecSigGetVersion(err,buf,sz,&kv) &&
        (signer.type == FRENT_TYPE_NIQS) &&
        frKeyQryFetchPubKey(err,kq,kv,FRCRYPT_KEYTYPE_SIGNING,
                            signer,kh,retries,timeout,&sigkey) &&
        frSecSigVerify(err,sigkey,buf,sz,&valid) &&
        (valid) &&
        frSecSigRemoveInSitu(err,buf,&sz)) { 
      *cacheDataVersion = buf[0] * 0x1000000 + buf[1] * 0x10000 + 
                          buf[2] * 0x100 + buf[3];
      ret = 1;
    }

    /*
     * cleanup
     */

    if (buf) {
      free(buf);
    }


  } else {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,NULL);
  }
  return ret;
}


/*
 *
 */

DLLExport int
frNiqsTokenInfo (
  ZkError err ,
  ZkipClient client ,
  void ** data ,
  size_t * sz )
{
  static const char * func = "frNiqsTokenInfo";
  int return_code = 0;
  char cmd[CMD_SZ];
  

  if (client && data && sz) {
    if (snprintf(cmd,CMD_SZ,"1 7.1.0 tokeninfo nodata binary 0\015\012") < 0) {
      zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,
                 "Unable to formulate 7.1.0 clientinfo command");
    } else {
      return_code = frNiqsCommand(err,client,cmd,data,sz);
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,NULL);
  }
  return return_code;
}

/*
 * Make an NIQS query
 */

DLLExport int
frNiqsQuery (
  ZkError err ,
  ZkipClient client ,
  FrNiqAttr attr ,
  FrEntIdExp * entity ,
  void ** data ,
  size_t * sz )
{
  static const char * func = "frNiqsQuery";
  int return_code = 0;
  const char * ent_type = NULL;
  const char * attribute = NULL;
  char cmd[CMD_SZ] = "";
 

  if (client && entity && data && sz) {
    *data = NULL;
    *sz = 0;
    if (!frNiqsAttributeToToken(err,attr,&attribute)) {
      zkErrorSet_va(err,func,ZKERR,ZKMNIQS,ZKEPARMBAD,
                 "Requested attribute %d is invalid", (int) attr);
    } else if (!frEntGetTypeStr(err,entity,&ent_type)) {
      zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKEPARMBAD, 
                 "Unable to determine entity type");
    } else {

      if( snprintf(cmd,CMD_SZ,"1 7.2.0 query nodata binary 0 %s %s %s\015\012",
                   attribute,ent_type,entity->name) < 0) {
          zkErrorSet(err,func,ZKERR,ZKMNIQS, ZKERROR,
                     "snprintf copy truncated");
      }
      else {
          return_code = frNiqsCommand(err,client,cmd,data,sz);
      }

    }
    if (return_code == 0) {
      if (*data) {
        free(*data);
      }
      *data = NULL;
      *sz = 0;
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,NULL);
  }
  return return_code; 
}

/**
 *  Return a set of status information for the requested entity types
 *
 *  @param err Returns the error if one occured
 *  @param client The ZKIP client structure which is connected to an NIQS
 *  @param since Request new information only since this time_t value
 *  @param type An array of FrEntIdType (see frEnt.h for the list)
 *  @param types The number of elements in the type array 
 *  @param buf Allocates and returns a signed data buffer.  You must free this
 *             when you're done with it
 *  @param sz Returns the size of the buffer
 *  @version 1.0
 *  @author Philippe Boucher (philippe@zks.net)
 *  @return A non-zero value on success, a zero on failure
 *
 */

DLLExport int
frNiqsStatus (
  ZkError err ,
  ZkipClient client ,
  time_t since ,
  FrEntIdType * type ,
  size_t types ,
  void ** data ,
  size_t * sz )
{
  return frNiqsListOrStatus(err,client,FrNiqsCommandStatus,since,
                            type,types,data,sz);
}

/**
 *  Returns a signed buffer containing a list of entities of the requested 
 *  types.  The signature of this buffer should be verified using the
 *  frSecSigVerify() function.  It must be removed prior to converting the
 *  buffer to a ZkList of FrNiqsEntity structures (see frNiqs.h) using the
 *  frNiqsListFromBuffer() function. 
 *
 *  @param err Returns the error if one occured
 *  @param client The ZKIP client structure which is connected to an NIQS
 *  @param since Request new information only since this time_t value
 *  @param type An array of FrEntIdType (see frEnt.h for the list)
 *  @param types The number of elements in the type array 
 *  @param buf Allocates and returns a signed data buffer.  You must free this
 *             when you're done with it
 *  @param sz Returns the size of the buffer
 *  @version 1.0
 *  @author Philippe Boucher (philippe@zks.net)
 *  @return A non-zero value on success, a zero on failure
 *
 */

DLLExport int
frNiqsList (
  ZkError err ,
  ZkipClient client ,
  time_t since ,
  FrEntIdType * type ,
  size_t types ,
  void ** data ,
  size_t *sz )
{
  return frNiqsListOrStatus(err,client,FrNiqsCommandList,since,
                            type,types,data,sz);
}

/*
 *
 */


DLLExport int
frNiqsAvailable (
  ZkError err ,
  ZkipClient client ,
  FrEntIdType * type ,
  size_t types ,
  void ** data ,
  size_t * sz )
{
  return frNiqsListOrStatus(err,client,FrNiqsCommandAvailable,0,
                            type,types,data,sz);
}

DLLExport int
frNiqsGetAvailable (
  ZkError err ,
  ZkipClient client ,
  FrKeyQryHost kh ,
  FrKeyQry kq ,
  FrEntIdType * type ,
  size_t types ,
  uint32_t ** entid ,
  size_t * entid_sz )
{
  static const char * func = "frNiqsGetAvailable";
  int ret = 0;
  FrEntIdExp signer;
  FrCryptPubKey sigkey = NULL;
  unsigned int kv = 0; /* Key version */
  unsigned char * buf = NULL;        /* Buffer from the NIQS */
  size_t sz = 0;
  size_t es = 0;   /* Number of entity IDs in the array */
  size_t n = 0;    /* Loop index to convert the buffer to entity ID array */
  uint32_t *entid_p = NULL;
  int valid = 0;
  int retries = 5;
  int timeout = 3;

  if (client && kq && entid && entid_sz) {
    *entid = NULL;
    *entid_sz = 0;
    if (frNiqsAvailable(err,client,type,types,(void **) &buf,&sz) &&
        frSecSigGetEntity(err,buf,sz,&signer) &&
        frSecSigGetVersion(err,buf,sz,&kv) &&
        (signer.type == FRENT_TYPE_NIQS) &&
        frKeyQryFetchPubKey(err,kq,kv,FRCRYPT_KEYTYPE_SIGNING,
                            signer,kh,retries,timeout,&sigkey) &&
        frSecSigVerify(err,sigkey,buf,sz,&valid) &&
        (valid) &&
        frSecSigRemoveInSitu(err,buf,&sz)) { 

      es = sz / 4;
      if ((sz % 4) != 0) {
        zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKEDATABAD,
                   "Returned buffer is incomplete");
      } else if ((entid_p = (uint32_t *) 
                            malloc(es * sizeof(uint32_t))) == NULL) {
        zkErrorSet_va(err,func,ZKERR,ZKMNIQS,ZKENOMEM,
                      "malloc():  %s",strerror(errno));
      } else {
        for (n = 0; n < es; n++) {
          entid_p[n] = buf[n * 4] * 0x1000000 + buf[n * 4 + 1] * 0x10000 +
                       buf[n * 4 + 2] * 0x100 + buf[n * 4 + 3];
        }
        *entid = entid_p; 
        *entid_sz = es;
        ret = 1;
      }
    }

    /*
     * cleanup
     */

    if (buf) {
      free(buf);
    }


  } else {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,NULL);
  }
  return ret;
}


static const char * list_proto = "7.1.0";

/*  
 *  %%Bug:  This is a hack.  It MUST be removed.
 *
 */

DLLExport int
frNiqsSetListProtocol (
  ZkError err ,
  int proto )
{
  static const char * func = "frNiqsSetListProtocol";
  int ret = 1;

  switch (proto) {
  case 1:  list_proto = "7.1.0"; break;
  case 2:  list_proto = "7.2.0"; break;
  case 3:  list_proto = "7.3.0"; break;
  default:
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,"Unknown protocol version");
    ret = 0;
    break;
  }
  return ret;
}
    
/*
 *  This function implements the 7.1.0 list and 7.1.0 status commands.
 *  It is wrapped by the functions frNiqsList() and frNiqsStatus().  
 *  It also will handle the 7.1.0 available command.
 */

static int
frNiqsListOrStatus (
  ZkError err ,
  ZkipClient client ,
  FrNiqsCommand command ,
  time_t since ,
  FrEntIdType * type ,
  size_t types ,
  void ** data ,
  size_t *sz )
{
  static const char * func = "frNiqsListOrStatus";
  int return_code = 0;
  int bytes = 0;  /* Number of bytes written to the cmd buffer */
  int base = 0;  /* Next write position in the cmd buffer */
  size_t n = 0; /* Loop index */
  char cmd[CMD_SZ + 2] = "";  /* The command string, the + 2 is to make sure
                               * I always have space enough to add the final
                               * CRLF in the command string.
                               */
  const char * ent_str = "";  /* Holds the entity type ID string token */
  const char * cmd_tok = NULL;  /* List or status command token */

/*
 * %%BUG:
 *  This function is now using version 1 (instead of 0) of the list and
 *  status commands.  This allows you to specify a negative value for the
 *  time since the last lookup.  This will act as an offset from the current
 *  NIQS server UTC
 *
 */


  if (client && data && sz && type) {

    /* 
     * Get the command token first
     */
    
    if (command == FrNiqsCommandList) {
      cmd_tok = "list";
      bytes = snprintf(cmd,CMD_SZ,"0 %s %s nodata binary 0 %d",list_proto,
                       cmd_tok, ( int )since);
    } else if (command == FrNiqsCommandStatus) {
      cmd_tok = "status";
      bytes = snprintf(cmd,CMD_SZ,"0 7.3.0 %s nodata binary 0 %d",cmd_tok, 
                       ( int )since);
    } else if (command == FrNiqsCommandAvailable) {
      cmd_tok = "available";
      bytes = snprintf(cmd,CMD_SZ,"0 7.1.0 %s nodata binary 0 ",cmd_tok);
    }

    /*
     * Now create the ZKIP 7.1.0 list command string
     */
    
    if (cmd_tok == NULL) {
      zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKEPARMBAD,"Unrecognized command");
    } else if (types < 1) {
      zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKEPARMBAD,
                 "You must have at least one FrEntIdType specified");
    } else if (bytes < 0) {
      zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,
                 "Unable to create ZKIP 7.1.0 list command");
    } else {
      base = bytes;
      return_code = 1;
      for (n = 0; n < types; n++) {

        /*
         * Get the entity type string token
         */

        if ((ent_str = frEntTypeGetStr(type[n])) == NULL) {
          zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKEPARMBAD,
                     "Unable to identify requested entity type");
          return_code = 0;
          break;
 
        /*
         * And add it as a parameter in the list command buffer
         */

        } else if ((bytes = snprintf(&(cmd[base]),CMD_SZ - base,
                                     " %s",ent_str)) < 0) { 
          zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,
                     "Unable to create ZKIP 7.1.0 list command");
          return_code = 0;
          break;
        } else {
          base += bytes;
        }
      }
      zkCStrlcat(cmd,"\015\012", CMD_SZ);

      /*
       * Only send the command if we succeeded in creating the cmd buffer
       */

      if (return_code) {
        if (!frNiqsCommand(err,client,cmd,data,sz)) {
          return_code = 0;
        }
      }
    }

    /*
     * Do the cleanup if we failed.
     */

    if (return_code == 0) {
      if (*data) {
        free(*data);
      }
      *data = NULL;
      *sz = 0;
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,NULL);
  }
  return return_code;
}

/*
**
*/

DLLExport int
frNiqsMxQuery (
    ZkError err,       /* eturns the error if one occured */
    ZkipClient client, /* The zkip client structure connected to the NIQS */
    FrNiqAttr attr,    /* The requested attribute */
    time_t since,      /*new data since this time (should always be set to 0)*/
    uint16_t encoding,
    FrEntIdType *type, /* (array) Type of entity for which we want to query */
    size_t types,      /* the number of entity in the array */
    void **data,       /* Returns a buffer containing the list */
    size_t *sz         /* size of the buffer */
)
{

    static const char * func = "frNiqsMxQuery";
    int return_code = 0;

    const char * attribute = NULL;

    int bytes = 0;  /* Number of bytes written to the cmd buffer */
    int base = 0;  /* Next write position in the cmd buffer */
    size_t n = 0; /* Loop index */
    const char * ent_str = "";  /* Holds the entity type ID string token */
    const char * encoding_str = "NONE";

    char cmd[CMD_SZ + 2] = "";
 
    if( encoding == ZKIP_ENCODING_BZIP2 ) {
      encoding_str = "BZIP2";
    }

    if ( ! client || ! attr || ! type || ! data || ! sz) {
        zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,NULL);
    }
    else {

        *data = NULL;
        *sz = 0;

        if (!frNiqsAttributeToToken(err,attr,&attribute)) {
            zkErrorSet_va(err,func,ZKERR,ZKMNIQS,ZKEPARMBAD,
                          "Requested attribute %d is invalid", (int) attr);
        }
        else if( (bytes = snprintf(cmd,CMD_SZ,"0 7.1.0 mxquery nodata binary 0 %s %d %s", attribute, ( int )since, encoding_str)) < 0) {
            zkErrorSet(err,func,ZKERR,ZKMNIQS, ZKERROR,
                       "snprintf copy truncated");
        }
        else {
            base = bytes;
            
            return_code = 1;

            for (n = 0; n < types; n++) {
                
                /*
                 * Get the entity type string token
                 */
                
                if ((ent_str = frEntTypeGetStr(type[n])) == NULL) {
                    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKEPARMBAD,
                               "Unable to identify requested entity type");
                    return_code = 0;
                    break;
                    
                    /*
                     * And add it as a parameter in the list command buffer
                     */
                    
                }
                else if ((bytes = snprintf(&(cmd[base]),CMD_SZ - base,
                                           " %s",ent_str)) < 0) { 
                    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,
                               "Unable to create ZKIP 7.1.0 mxquery command");
                    return_code = 0;
                    break;
                } else {
                    base += bytes;
                }
            }

            zkCStrlcat(cmd,"\015\012", CMD_SZ);
            
        }
        
        /*
         * Only send the command if we succeeded in creating the cmd buffer
         */
        
        if (return_code) {
            if ( ! frNiqsCommand(err,client,cmd,data,sz)) {
                return_code = 0;
            }
        }
        
        
        /*
         * Do the cleanup if we failed.
         */

        if (return_code == 0) {
            if (*data) {
                free(*data);
            }
            *data = NULL;
            *sz = 0;
        }
    }

    return return_code;
}
