/*
 * $Id: frNymClient.c,v 1.19.8.9 2001/01/15 23:37:13 corey Exp $
 */

static const char * module_id_ = "$Id: frNymClient.c,v 1.19.8.9 2001/01/15 23:37:13 corey Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */



#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#else
#include <winsock2.h>
#endif

#include <zkIntTypes.h>
#include <zkError.h>
#include <zkAlloc.h>
#include <ZKIP.h>
#include <zkipClient.h>
#include <frNym.h>
#include <frToken.h>
#include <frNymClient.h>
#include <frSec.h>
#include "zkCStr.h"
#include "zkFile.h"
#include "zkList.h"
#include "frNymUtil.h"

/*
 * This is the command string which must be issued to properly quit from
 * the connection to the nymserver
 */

#define QUIT_COMMAND "1 0.1.0 quit nodata none 0\015\012"

#define CMD_SZ (256 + TOKEN_BASE64_SZ) /* Maximum length of a nym server  */
                                       /* command string */

#define COMMAND_GET        1
#define COMMAND_CREATE     2
#define COMMAND_MODIFY     3
#define COMMAND_DISABLE    4
#define COMMAND_ADDTOKEN   5
#define COMMAND_AVAILABLENAMES 6
#define COMMAND_RBACTIVATE 7
#define COMMAND_MAILCERT   8

static int frRemoteNymServerCommand (
  ZkError  ,
  ZkipClient,
  FrCryptSig,
  NymServerResponse * ,  /* Returns the relevant nym server response */
  int ,                  /* The command code */
  char * ,               /* The command string */
  char * ,               /* binary nym data for creating a nym (optional) */
  int ,                  /* Length of the nym data */
  void * ,               /* The content depends on the command */
  void * );              /* The content depends on the command */

static int getConnectedNymServer(ZkError err, NymServer *nymserver, ZkipClient* pClient);

static int getConnectedNymServer(ZkError err, NymServer *nymserver, ZkipClient* pClient)
{
  int ret = 0;
  int fd = 0;

  if (nymserver
	  && pClient
      && zkipClientNew(err, pClient)
	  && zkipClientMakeBlocking(err, *pClient, nymserver->timeout)
	  && zkipClientConnectToHost(err, *pClient, nymserver->nymserver, nymserver->port)
	  && zkipClientFd(err, *pClient, &fd))
  {
    ret = 1;
  }
  else if (pClient && *pClient)
  {
    zkipClientDisconnect(NULL, *pClient);
    zkipClientFree(NULL, pClient);
  }
  return ret;
}

/*
 * Basic function to issue a command to the nym server and quit immediatly.
 * The extra_parm parameter is used to return a pointer to a newly created
 * FrRemoteNym in the case of the CREATE, MODIFY, DISABLE, and ADDTOKEN
 * commands, or to return a new list of names in the case of an AVAILABLENAMES
 * command.
 */

static int
frRemoteNymServerCommand (
  ZkError err ,
  ZkipClient client,
  FrCryptSig nymServSigKey,
  NymServerResponse * nsresponse,
  int command ,
  char * cmdstr ,
  char * nymdata, /* binary nym data for creating a nym */
  int nymlen , /* Length of the nym data */
  void * extra_parm1,  /* The content depends on the command */
  void * extra_parm2 )  /* The content depends on the command */
{
  int retcode = 0;
  struct ZkipCommandResponse * response = NULL;
  char * extra_data = NULL;
  FrRemoteNym * nym = NULL;
  ZkList * name_list = NULL;
  FrMailAuth *cert = NULL;
  FrCryptMacKey *macKey = NULL;

  (void) module_id_; /* Gets rid of compile time warning */
  zkErrorClear(err);
  if (client && cmdstr && nsresponse) {

    /* 
     * Sanity checks...
     */

    switch (command) {
    case COMMAND_GET:
    case COMMAND_CREATE:
    case COMMAND_MODIFY:
    case COMMAND_DISABLE:
    case COMMAND_ADDTOKEN:
      nym = (FrRemoteNym *) extra_parm1;
      break;
    case COMMAND_AVAILABLENAMES:
      name_list = (ZkList *) extra_parm1;
      break;
    case COMMAND_RBACTIVATE:
      break;
    case COMMAND_MAILCERT:
      cert = (FrMailAuth *) extra_parm1;
      macKey = (FrCryptMacKey *) extra_parm2;
      break;
    default:
      zkErrorSet(err,"frRemoteNymCommand",ZKERR,ZKMRNYM,ZKEPARMBAD,
                 "Unrecognized nym command");
      return 0;
    }


    /*
     * Make sure that if the user is expecting a nym back, he wont get anything
     * if the function fails but he does not check the return code.
     */

    if (nym) {
      *nym = NULL;
    }

    nsresponse->code = 0;
    zkCStrlcpy(nsresponse->message,"Did not connect to nym server",sizeof nsresponse->message);

    /*
     * Now we need to read the banner response and make sure we got
     * a SERVICE_READY (220)
     */

    if (!zkipClientReadResponse(err,client,&response,&extra_data)) {
      zkCStrlcpy(nsresponse->message,"Failed to read welcome banner",sizeof nsresponse->message);
      goto done;
    }
    if (response->response != _SERVICE_READY) {
      nsresponse->code = response->response;
      strncpy(nsresponse->message,response->message,NSRMESG_SZ);
      nsresponse->message[NSRMESG_SZ] = (char) 0;
      zkErrorSet(err,"frRemoteNymServerCommand",ZKERR,ZKMRNYM,ZKERROR,
                 response->message);
      goto done;
    }

    /*
     * Now send the command
     */
  
    if (!zkipClientWrite(err,client,cmdstr,strlen(cmdstr))) {
      zkCStrlcpy(nsresponse->message,"Failed to write data to nym server",sizeof nsresponse->message);
      goto done;
    } 
  
    /*
     * And read the first response
     */

    if (!zkipClientReadResponse(err,client,&response,&extra_data)) {
      zkCStrlcpy(nsresponse->message,"Failed to read data from nym server",sizeof nsresponse->message);
      goto done;
    }
    
    /*
     * Depending on the specific command, we either need to check that
     * we got back the data we wanted or we have to send some data over
     */

    switch (command) {

    case COMMAND_GET:
      if (response->response == _SERVICE_READY) { /* Nym found */
        *nym = NULL;
        if (frRemoteNymFromBinary(err,nym,nymServSigKey,
                                  (unsigned char *)extra_data,
                                  response->datalen)) {
          retcode = 1;
        } else {
          nsresponse->code = _COMMAND_FAILED;
          if (zkErrorCode(err) == ZKESIGCK) {
            zkCStrlcpy(nsresponse->message,"Nymserver signature check failed",sizeof nsresponse->message);
          } else {
            /* Have to fudge the response here a bit since it isnt a real
             * ZKIP error 
             */
            zkCStrlcpy(nsresponse->message,"Incoming nym data corrupted",sizeof nsresponse->message);
          }
        }
      } else { /* Nym not found , or something unexpected */
        nsresponse->code = response->response;
        strncpy(nsresponse->message,response->message,NSRMESG_SZ); 
        nsresponse->message[NSRMESG_SZ] = (char) 0;
      }
      break;

    case COMMAND_CREATE:
    case COMMAND_ADDTOKEN:
    case COMMAND_MODIFY:
    case COMMAND_DISABLE:
    case COMMAND_AVAILABLENAMES:
    case COMMAND_RBACTIVATE:
    case COMMAND_MAILCERT:
      if (response->response == _DATA_STARTINPUT) { /* Can send data */
        if (!zkipClientWrite(err,client,nymdata,nymlen)) {
          zkCStrlcpy(nsresponse->message,"Failed to write data to nym server",sizeof nsresponse->message);
          goto done;
        }
        if (!zkipClientWrite(err,client,"\015\012.\015\012",5)) {
          zkCStrlcpy(nsresponse->message,"Failed to write data to nym server",sizeof nsresponse->message);
          goto done;
        }
        if (!zkipClientReadResponse(err,client,&response,&extra_data)) {
          zkCStrlcpy(nsresponse->message,"Failed to read data from nym server",sizeof nsresponse->message);
          goto done;
        }

        /* The create and modify commands will return a copy of the 
         * nym which was inserted in the database.  This allows the client to
         * know under which name the nym was stored, when the expiry date
         * is, etc...
         *
         * I am creating a temporary copy of the nym structure from the 
         * returned binary nym information regardless of wether or not the
         * caller of this funciton wants it.  This acts as an extra sanity
         * check.
         */

        if (response->response == _SERVICE_READY) {
          if ((command == COMMAND_CREATE) || 
              (command == COMMAND_MODIFY) ||
              (command == COMMAND_ADDTOKEN)) {
            FrRemoteNym nym_t = NULL;
            if (frRemoteNymFromBinary(err,&nym_t,nymServSigKey,
                                      (unsigned char *) extra_data, 
                                      response->datalen)) {
              retcode = 1;
            }
            if (nym) {
              *nym = nym_t;
            } else {
              frRemoteNymFree(NULL,&nym_t);
            }

          /*
           * If the command was list availablenames, convert the buffer we 
           * got back to a stringlist.
           */

          } else if (command == COMMAND_AVAILABLENAMES) {
            FrRemoteNym lh = NULL;
            if (!(frRemoteNymNew(err,&lh) && 
                  frRemoteNymAddParameter(err,lh,NymFriends,extra_data,
                                          response->datalen) &&
                  frRemoteNymGetStringListParameter(err,lh,NymFriends,
                                                    name_list) &&
                  frRemoteNymFree(err,&lh))) {
              retcode = 0;
            } else {
              retcode = 1;
            }

          } else if (command == COMMAND_MAILCERT) {
            /* 
             * The buffer contains both a Mail certificate (variable size)
             * and a MAC key (fixed size).
             */
            size_t mailCertSize  = (response->datalen - FRCRYPT_MACKEY_SIZE);
            
            /* Reconstruct the Mail Certificate and the MAC key */
            if (frMailAuthNewFromBuf(err,cert,extra_data,mailCertSize)) {
              if (frCryptMacKeyNew(err,macKey)) {
                if (frCryptMacKeyLoad(err,*macKey,extra_data+mailCertSize,
                                      FRCRYPT_MACKEY_SIZE)) {
                  /* Got the Mail Cert and MAC key !!! */
                  retcode = 1;
                } else {
                  /* Release the MAC key. */
                  frCryptMacKeyFree(err,macKey);
                }
              }

              /* If failed to get MAC key we need to free the Mail Cert. */
              if (!retcode)
                frMailAuthFree(err,cert);
            }
            
            
            /*
             * For any other command which does not expect returned data 
             */

          } else {
            retcode = 1;
          }
        }
      }
      /* The command failed */
      nsresponse->code = response->response;
      strncpy(nsresponse->message,response->message,NSRMESG_SZ); 
      nsresponse->message[NSRMESG_SZ] = (char) 0;
      break;
    }

    /*
     * Issue the quit command and get the reponse.  Note that there is
     * no error structure passed because I dont care if this fails or
     * not.  It's mostly being done to be polite to the remote server
     */

    zkipClientWrite(NULL,client,QUIT_COMMAND,strlen(QUIT_COMMAND));
    zkipClientReadResponse(NULL,client,&response,&extra_data);

done:
    return retcode; 
  } 
  zkErrorSet(err,"frRemoteNymCommand",ZKERR,ZKMRNYM,ZKENULLPTR,NULL);
  return 0;
}

/*
 *
 */

DLLExport int
frRemoteNymGet (
  ZkError err ,
  FrRemoteNym * nym ,
  char * nymname ,
  NymServer * nymserver,
  NymServerResponse * response)
{
  ZkipClient client = NULL;
  int ret = 0;

  zkErrorClear(err);

  if (nymserver && getConnectedNymServer(err, nymserver, &client))
  {
    ret = frRemoteNymGetConnected(err, nym, nymname, client,
				  nymserver->sigkey, response);
    zkipClientDisconnect(NULL,client);
    zkipClientFree(NULL,&client);
  }
  return ret;
}

DLLExport int
frRemoteNymGetConnected (
  ZkError err ,
  FrRemoteNym * nym ,
  char * nymname ,
  ZkipClient client,
  FrCryptSig nymServSigKey,
  NymServerResponse * response)
{
  const char * const func = "frRemoteNymGet";
  char cmdstr[CMD_SZ];

  /*
   * build the ZKIP nym server get command
   */

  zkErrorClear(err);
  if (nym && nymname && client && response) {
    int spf_res = 0;

    spf_res = snprintf(cmdstr,
                       sizeof cmdstr,
                       "1 1.1.0 get nodata binary 0 \"%s\"\r\n",
                       nymname);
    
    if ((spf_res >= (int) sizeof cmdstr) || (spf_res < 0)) {
      zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKEPARMBAD,"snprintf failed");
      return 0;
    } else { 
      if (!frRemoteNymServerCommand(err,client, nymServSigKey, response,COMMAND_GET,cmdstr,
                                    NULL,0,nym,NULL)) {
        return 0;
      } else {
        return 1;
      }
    }
  }
  zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,"Null param");
  return 0;
}

DLLExport int
frRemoteNymCreate (
  ZkError err ,
  FrRemoteNym nym ,
  char * nymname ,
  char * nymtoken ,
  NymServer * nymserver,
  NymServerResponse * response,
  FrRemoteNym * newnym )
{
  ZkipClient client = NULL;
  int ret = 0;

  zkErrorClear(err);

  if (getConnectedNymServer(err, nymserver, &client))
  {
    ret = frRemoteNymCreateConnected(err, nym, nymname, nymtoken,
				     client, nymserver->sigkey,
				     response, newnym);
    zkipClientDisconnect(NULL,client);
    zkipClientFree(NULL,&client);
  }
  return ret;
}

DLLExport int
frRemoteNymCreateConnected (
  ZkError err ,
  FrRemoteNym nym ,
  char * nymname ,
  char * nymtoken ,
  ZkipClient client,
  FrCryptSig nymServSigKey,
  NymServerResponse * response,
  FrRemoteNym * newnym )
{
  const char * const func = "frRemoteNymCreate";
  char cmdstr[CMD_SZ];
  unsigned char * nymdata = NULL;
  size_t nymlen = 0;
  int ret = 0;   /* Function return code */
  int spf_res = 0;  /* snprintf call results */

  zkErrorClear(err);
  if (nym && nymname && nymtoken && client && response && newnym) {
    response->code = 0;
    zkCStrlcpy(response->message,"Error converting nym data",sizeof response->message);
    if (!frRemoteNymToBinary(err,nym,&nymdata,&nymlen)) {
      return 0;
    }

    spf_res = snprintf(cmdstr,
                       sizeof cmdstr,
                       "1 1.1.0 create remotenym binary %d \"%s\" \"%s\"\r\n",
                       nymlen,
                       nymname,
                       nymtoken);

    if ((spf_res >= (int) sizeof cmdstr) || (spf_res < 0)) {
      zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKEPARMBAD,"snprintf failed");
      ret = 0;

    } else if (!frRemoteNymServerCommand(err,
                                         client,
										 nymServSigKey,
                                         response,
                                         COMMAND_CREATE,
                                         cmdstr,
                                         (char *) nymdata,
                                         nymlen,
                                         newnym,
                                         NULL)) {
      ret = 0;

    } else {
        ret = 1;      
    }

    if (nymdata) {
      free(nymdata);
    } 
  
    return ret;
  }
  zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,"Null pointer");
  return 0;
}

/**
 *  This function will upgrade the specified nym with the given token.  The
 *  token must be a base64 string (this is what is returned when using the
 *  token client functions from frtoken).  If the newnym parameter is specified
 *  the current version of the nym as stored on the nym server will be
 *  returned on a succesfull upgrade.
 *
 *  @param err Returns the error if one occured
 *  @param nym The remote nym to which you wish to add the token
 *  @param token The base64 token string which you wish to send to the 
 *               nymserver.  This string must be NUL terminated
 *  @param nymserver The nym server connection structure
 *  @param newnym Returns the remote nym image which is stored on the 
 *                nym server after the token has been successfully added.
 *                You can set this to NULL if you dont care, however you should
 *                seriously consider looking at this data since it will hold
 *                the new modified token controlled nym parameters (these
 *                have the FL_TOKEN flag)
 *  @author Philippe Boucher (philippe@zks.net)
 *  @version 1.1
 *  @return A non-zero value on success, 0 on failure
 */

DLLExport int
frRemoteNymAddToken (
  ZkError err ,
  FrRemoteNym nym ,
  char * token ,
  NymServer * nymserver,
  NymServerResponse * response,
  FrRemoteNym * newnym )
{
  ZkipClient client = NULL;
  int ret = 0;

  zkErrorClear(err);

  if (getConnectedNymServer(err, nymserver, &client))
  {
    ret = frRemoteNymAddTokenConnected(err, nym, token, client,
				       nymserver->sigkey,
				       response, newnym);
    zkipClientDisconnect(NULL,client);
    zkipClientFree(NULL,&client);
  }
  return ret;
}

DLLExport int
frRemoteNymAddTokenConnected (
  ZkError err ,
  FrRemoteNym nym ,
  char * token ,
  ZkipClient client,
  FrCryptSig nymServSigKey,
  NymServerResponse * response,
  FrRemoteNym * newnym )
{
  static const char * func = "frRemoteNymAddToken";
  const unsigned char * name = NULL;  /* Name string extracted from the nym */
  size_t name_sz = 0;  /* Number of bytes in the name string (excludes NUL) */
  char cmd[CMD_SZ];    /* For the add token command string */
  FrCryptSig prvsig;   /* Nym's private signature key to sign the token */
  void * sgndtok = NULL; /* Signed token buffer */
  size_t sgndtok_sz = 0; /* Size of the signed token buffer */
  int return_code = 0;

  zkErrorClear(err);

  /*
   * Sanity checks.  Note that newnym is allowed to be NULL so it is not
   * checked here.
   */

  if (!nym) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,"nym pointer is NULL");
  } else if (!token) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,"token pointer is NULL");
  } else if (!client) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,"client pointer is NULL");
  } else if (!response) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,"response pointer is NULL");
  } else {

    /*
     * Get the name and private signature key of the nym and sign the token.
     * If this is all successfull, we can send the information over to the 
     * nymserver.
     */

    frRemoteNymResetParameterSearch(err,nym);
    if (frRemoteNymGetParameter(err,nym,NymName,&name,&name_sz) &&
        frRemoteNymPrivateSignatureKey(err,nym,&prvsig) &&
        frSecSignData(err,prvsig,token,strlen(token),&sgndtok,&sgndtok_sz)) {

      int spf_res = 0;

      spf_res = snprintf(cmd,
                         sizeof cmd,
                         "0 1.1.0 addtoken signedtok binary %d \"%s\"\015\012",
                         sgndtok_sz,
                         name);

      if ((spf_res >= (int) sizeof cmd) || (spf_res < 0)) {
        zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKEINCOMPLETE,
                   "Unable to create 1.1.0 addtoken command buffer");
        
        /*
         * Send the command with the signed token data to the nymserver
         * On success, the nym will have been upgraded and if newnym is
         * not NULL, it will contain the current nym parameters from the 
         * nym server.
         */

      } else if (frRemoteNymServerCommand(err,client,nymServSigKey,response,
                                          COMMAND_ADDTOKEN,cmd,
                                          sgndtok,sgndtok_sz,newnym,NULL)) {
        return_code = 1;
      }
    }

    /*
     * Cleanup 
     */

    if (sgndtok) {
      free(sgndtok);
    }
  }

  return return_code;
}

DLLExport int 
frRemoteNymModify (
  ZkError err ,
  FrRemoteNym nym ,
  char * nymname ,
  NymServer * nymserver,
  NymServerResponse * response,
  FrRemoteNym * newnym )
{
  ZkipClient client = NULL;
  int ret = 0;

  zkErrorClear(err);

  if (getConnectedNymServer(err, nymserver, &client))
  {
    ret = frRemoteNymModifyConnected(err, nym, nymname, client,
				     nymserver->sigkey, response,
				     newnym);
    zkipClientDisconnect(NULL,client);
    zkipClientFree(NULL,&client);
  }
  return ret;
}

DLLExport int 
frRemoteNymModifyConnected (
  ZkError err ,
  FrRemoteNym nym ,
  char * nymname ,
  ZkipClient client,
  FrCryptSig nymServSigKey,
  NymServerResponse * response,
  FrRemoteNym * newnym )
{
  const char * func = "frRemoteNymModify";
  int bytes = 0;
  char cmdstr[CMD_SZ + 1];
  unsigned char * nymdata = NULL;
  size_t nymlen = 0;
  int return_code = 0; /* Function return code */

  zkErrorClear(err);
  if (nym && nymname && client && response && newnym) {
    response->code = 0;
    zkCStrlcpy(response->message,"Error converting nym data",sizeof response->message);
    if (!frRemoteNymToBinary(err,nym,&nymdata,&nymlen)) {
      return 0;
    }
    bytes = snprintf(cmdstr, sizeof cmdstr,
                     "1 1.1.0 modify remotenym binary %d \"%s\"\015\012",
                     nymlen,nymname);
    if ((bytes < 0) || (bytes >= (int) sizeof cmdstr))  {
      zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKEINCOMPLETE,
                 "Not enough space to store nym 1.1.0 modify command");
      return_code = 0;
    } else if (!frRemoteNymServerCommand(err,client,nymServSigKey,response,
                                         COMMAND_MODIFY,cmdstr,
                                         (char *) nymdata,nymlen,newnym,NULL)) {
      return_code = 0;
    } else {
      return_code = 1;
    }
   
    /*
     * Do the cleanup
     */

    if (nymdata) {
      free(nymdata);
    }

    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,NULL);
  return 0;
}
/*
 *
 */

DLLExport int
frRemoteNymDisable (
  ZkError err ,
  FrRemoteNym nym ,
  char * nymname ,
  NymServer * nymserver,
  NymServerResponse * response )
{
  ZkipClient client = NULL;
  int ret = 0;

  zkErrorClear(err);

  if (getConnectedNymServer(err, nymserver, &client))
  {
    ret = frRemoteNymDisableConnected(err, nym, nymname, client,
				      nymserver->sigkey,
				      response);
    zkipClientDisconnect(NULL,client);
    zkipClientFree(NULL,&client);
  }
  return ret;
}

DLLExport int
frRemoteNymDisableConnected (
  ZkError err ,
  FrRemoteNym nym ,
  char * nymname ,
  ZkipClient client,
  FrCryptSig nymServSigKey,
  NymServerResponse * response )
{
  const char * func = "frRemoteNymDisable";
  int bytes = 0;
  char cmdstr[CMD_SZ + 1];
  unsigned char * nymdata = NULL;
  size_t nymlen = 0;
  int return_code = 0;

  zkErrorClear(err);
  if (nym && nymname && client && response) {
    response->code = 0;
    zkCStrlcpy(response->message,"Error converting nym data",sizeof response->message);
    if (!frRemoteNymToBinary(err,nym,&nymdata,&nymlen)) {
      return 0;
    }
    bytes = snprintf(cmdstr, sizeof cmdstr,
                     "1 1.1.0 disable remotenym binary %d \"%s\"\015\012",
                     nymlen,nymname);
    if ((bytes < 0) || (bytes >= (int) sizeof cmdstr))  {
      zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKEINCOMPLETE,
                 "Not enough space to store nym 1.1.0 disable command");
      return_code = 0;
    } else if (!frRemoteNymServerCommand(err,client,nymServSigKey,response,
                                         COMMAND_DISABLE,cmdstr,
                                         (char *) nymdata,nymlen,NULL,NULL)) {
      return_code = 0;
    } else {
      return_code = 1;
    }
 
    /*
     * Cleanup 
     */

    if (nymdata) {
      free(nymdata);
    }

    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,NULL);
  return 0;
}

/*
 *
 */

DLLExport int
frRemoteNymAvailableNames (
  ZkError err ,
  ZkList names ,
  NymServer * nymserver ,
  NymServerResponse * response ,
  ZkList * anames)
{
  ZkipClient client = NULL;
  int ret = 0;

  zkErrorClear(err);

  if (getConnectedNymServer(err, nymserver, &client))
  {
    ret = frRemoteNymAvailableNamesConnected(err, names, client,
					     nymserver->sigkey,
					     response, anames);
    zkipClientDisconnect(NULL,client);
    zkipClientFree(NULL,&client);
  }
  return ret;
}

DLLExport int
frRemoteNymAvailableNamesConnected (
  ZkError err ,
  ZkList names ,
  ZkipClient client,
  FrCryptSig nymServSigKey,
  NymServerResponse * response ,
  ZkList * anames)
{
  static const char * func = "frRemoteNymAvailableNames";
  int ret = 1;                 /* function return code */
  FrRemoteNym lh = NULL;       /* Hack to make list buffer */
  char * data = NULL;          /* The list as a buffer */
  char cmdstr[CMD_SZ];         /* The availablenames command string */
  int data_len = 0;            /* Size of the list buffer */

  /*
   * Verify the function parameters
   */
  
  if (!client) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,
               "client is NULL");
    ret = 0;
  } else if (!response) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,
               "Nym server response structure is NULL");
    ret = 0;
  } else if (!names) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,
               "Original name list (names) is NULL");
    ret = 0;
  } else if (!anames) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,
               "Available name list (anames) is NULL");
    ret = 0;

  /*
   * Create a temporary nym object so that we can use its string list
   * capabilities.
   */

  } else if (!(frRemoteNymNew(err,&lh) &&
               frRemoteNymAddStringListParameter(err,lh,NymFriends,names) &&
               frRemoteNymGetParameter(err,lh,NymFriends,
                                       (const unsigned char **) &data,
                                       &data_len))) {
    ret = 0;

  /*
   * Set up the availble names command 
   */

  } else {

    int spf_res = 0;

    spf_res = snprintf(cmdstr, sizeof cmdstr,
                       "0 1.1.0 availablenames list binary %d\015\012",
                       data_len);

    if ((spf_res >= (int) sizeof cmdstr) || (spf_res < 0)) {
      zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKEINCOMPLETE,
                 "Unable to create 1.1.0 availablenames command buffer");
      ret = 0;

      /*
       * Send the command and get back the available names list
       */

    } else if (!frRemoteNymServerCommand(err,client,nymServSigKey,response,
                                         COMMAND_AVAILABLENAMES,cmdstr,
                                         (char *) data,data_len,
                                         (void *) anames,NULL)) {
      ret = 0;
    } else {
   
      /*
       * Success! the anames list now contains the list of currently available
       * names on the nym server
       */
 
      ret = 1;
    }
  }

  /*
   * Do our cleanup
   */

  if (lh) {
    frRemoteNymFree(NULL,&lh);
  }

  return ret;
}
  

/*
 *
 */

int
frRemoteNymRbActivate(
  ZkError err ,
  FrRemoteNym nym ,
  NymServer * nymserver,
  const void * request,
  size_t request_sz,
  NymServerResponse * response )
{
  ZkipClient client = NULL;
  int ret = 0;

  zkErrorClear(err);

  if (getConnectedNymServer(err, nymserver, &client))
  {
    ret = frRemoteNymRbActivateConnected(err, nym, client,
					 nymserver->sigkey,
					 request, request_sz,
					 response);
    zkipClientDisconnect(NULL,client);
    zkipClientFree(NULL,&client);
  }
  return ret;
}

int
frRemoteNymRbActivateConnected(
  ZkError err ,
  FrRemoteNym nym ,
  ZkipClient client,
  FrCryptSig nymServSigKey,
  const void * request,
  size_t request_sz,
  NymServerResponse * response )
{
  static const char * func = "frRemoteNymRbActivate";
  int                 is_ok = 0;
  void *              signed_request = NULL;
  size_t              signed_request_sz = 0;
  char                command[CMD_SZ]; /* The reply block activate command string */
  FrCryptSig          sigkey;
  zkErrorClear(err);

  if (!nym) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,
               "nym is null");
  } else if (!client) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,
			   "client is null");
  } else if (!request) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,
               "request is null");
  } else if (!response) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,
               "response is null");
  /*
   * 1. Get the signature key
   * 2. Sign the activation request buffer
   * 3. Send the command and request
   */

  } else if (frRemoteNymPrivateSignatureKey(err,nym,&sigkey)) {

    if (!sigkey) {
      zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,
                 "private signature key has not been set");
    } else if (!frSecSignData(err,sigkey,request,request_sz,
                              &signed_request,&signed_request_sz)) {
      is_ok = 0;
     /*
      * Set up the reply block activate command
      */
    } else {

      int spf_res = 0;

      spf_res = snprintf(command, sizeof command,
                         "0 1.1.0 rbactivate token binary %d\015\012",
                         signed_request_sz);

      if ((spf_res >= (int) sizeof command) || (spf_res < 0)) {
        zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKEINCOMPLETE,
                   "Unable to create 1.1.0 rbactivate command buffer");
      } else if (frRemoteNymServerCommand(err,client,nymServSigKey,response,
                                          COMMAND_RBACTIVATE,command,
                                          signed_request,
                                          signed_request_sz,
                                          NULL,NULL) ) {
        is_ok = 1;
      }
    }
  }

  if (signed_request) {
    zkMemFree(NULL,&signed_request);
  }

  return (is_ok);     ;
}

/*
 *  This function will query the Nym server for a Mail certificate for the
 *  given Nym with the given version.
 *
 *  @param err    Returns the error if one occured
 *  @param nym    The remote nym to which you wish to add the token
 *  @param srv    The nym server connection structure
 *  @param vers   Mail Certificate version
 *  @param resp   Nym server response
 *  @param cert   New Mail certificate 
 *  @param macKey New Nym specific MAC key 
 *  @return A non-zero value on success, 0 on failure
 */
DLLExport int frRemoteNymMailCert(
  ZK_INOUT ZkError           err,
  ZK_IN    FrRemoteNym       nym,
  ZK_IN    NymServer         *srv,
  ZK_IN    uint16_t          vers,
  ZK_OUT   NymServerResponse *resp,
  ZK_OUT   FrMailAuth        *cert, 
  ZK_OUT   FrCryptMacKey     *macKey)
{
  ZkipClient client = NULL;
  int ret = 0;

  zkErrorClear(err);

  if (getConnectedNymServer(err, srv, &client))
  {
    ret = frRemoteNymMailCertConnected(err, nym, client,
				       srv->sigkey, vers, resp,
				       cert, macKey);
    zkipClientDisconnect(NULL,client);
    zkipClientFree(NULL,&client);
  }
  return ret;
}
	
DLLExport int frRemoteNymMailCertConnected(
  ZK_INOUT ZkError           err,
  ZK_IN    FrRemoteNym       nym,
  ZkipClient                 client,
  FrCryptSig nymServSigKey,
  ZK_IN    uint16_t          vers,
  ZK_OUT   NymServerResponse *resp,
  ZK_OUT   FrMailAuth        *cert, 
  ZK_OUT   FrCryptMacKey     *macKey)
{
  static const char * func = "frRemoteNymMailCert";
  FrCryptSig prvsig;   /* Nym's private signature key to sign the version */
  char cmdstr[CMD_SZ + 1];
  const unsigned char *name = NULL;
  size_t name_sz = 0;
  void *sgndvers = NULL;
  size_t sgndvers_sz = 0;
  int res = 0;
  int spf_res = 0;  /* snprintf call results */

  if (!nym) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,
               "nym is null");
    goto done;
  }
  if (!client) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,
			   "client is null");
    goto done;
  }
  if (!resp) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,
               "resp is null");
    goto done;
  }
  if (!cert) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,
               "cert is null");
    goto done;
  }
  if (!macKey) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,
               "macKey is null");
    goto done;
  }
  
  /*
   * Get the name and private signature key of the nym and sign the version.
   */
  
  if (!frRemoteNymResetParameterSearch(err,nym) ||
      !frRemoteNymGetParameter(err,nym,NymName,&name,&name_sz)) {
    goto done;
  }

  if (!frRemoteNymPrivateSignatureKey(err,nym,&prvsig) ||
      !frSecSignData(err,prvsig,&vers,sizeof(vers),&sgndvers,&sgndvers_sz)) {
    goto done;
  }

  spf_res = snprintf(cmdstr, sizeof cmdstr,
                     "0 1.1.0 mailcert signedvers binary %d \"%s\" \015\012",
                     sgndvers_sz,name);

  if ((spf_res >= (int) sizeof cmdstr) || (spf_res < 0)) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKEINCOMPLETE,
               "Unable to create 1.1.0 mailcert command buffer");
    goto done;
  }
  
  /*
   * Send the command with the signed version data to the nymserver
   * On success, the encrypted mail certificate and the shared secret
   * will be returned.
   */
  if (!frRemoteNymServerCommand(err,client,nymServSigKey,resp,COMMAND_MAILCERT,cmdstr,
                                sgndvers,sgndvers_sz,cert,macKey)) {
    goto done;
  }

  res = 1;

 done:
  if (sgndvers) {
    zkMemFree(NULL,&sgndvers);
  }

  return res;
}
