/*
 *  $Id: fr_genpkt.h,v 1.18.8.2 2000/12/14 17:53:09 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef FR_GENPKT_H_
#define FR_GENPKT_H_


#include "zkError.h"
#include "frCrypt.h"


/*
 *  Defines, declarations, etc. relating to general packets, not any
 *  specific  type.
 *
 *  20 Sept. 1999
 *  --------------
 *
 *      Well ... the above comment was true "once upon a time".
 *
 *      XXX - Hack Alert! - XXX
 *
 *      FRPKTTYPE_REPLYBLOCK_PUB_INACTIVE breaks this nifty comment.  It
 *      represents the "lesser of several HACKS" to facilitate client
 *      activation of new reply blocks.  The nym server will control the
 *      active/inactive status of a public reply block.  A new reply
 *      block will be marked inactive by the nym server.  Its type will
 *      be changed to active (i.e., just a normal public reply block)
 *      following receipt of an activation request from the client.
 *
 *      - Roger McFarlane <roger@zks.net>
 */

typedef enum {
  FRPKTTYPE_UNUSED_,
  FRPKTTYPE_AIP_DATA,
  FRPKTTYPE_AIP_ROUTE_CREATE,
  FRPKTTYPE_AIP_ROUTE_DESTROY,
  FRPKTTYPE_AIP_ALIVE,
  FRPKTTYPE_AIP_ROUTE_CRTACK,
  FRPKTTYPE_DIFHEL,
  FRPKTTYPE_KEYQRY,
  FRPKTTYPE_KEYRESP,
  FRPKTTYPE_STAT,
  FRPKTTYPE_NDB_ROUTE_UPDATE,
  FRPKTTYPE_PKENCR_DATA,
  FRPKTTYPE_SIGNED_DATA,
  FRPKTTYPE_REPLYBLOCK_PRIV,
  FRPKTTYPE_REPLYBLOCK_PUB,
  FRPKTTYPE_NIDB_UPDATE,        /* Client sending requesting an update */
  FRPKTTYPE_NIDB_UPD_REQ,       /* Server sending update request */
  FRPKTTYPE_NIDB_UPD_REGTOP,    /* Server sending Topology Update request */
  FRPKTTYPE_NIDB_UPD_URGTOP,    /* Server sending URG Topology Update req */
  FRPKTTYPE_SYMENCR_DATA,       /* Symmetrically encrypted data */
  FRPKTTYPE_CLRTXT_DATA,        /* No encryption Clear Text */
  FRPKTTYPE_STRETCH_COUNT,      /* Type of header containing log iterations for key stretching */
  FRPKTTYPE_PASSINFO,           /* Passphrase info pkt */   
  FRPKTTYPE_REPLYBLOCK_ACTIVATION_REQ,
  FRPKTTYPE_REPLYBLOCK_PUB_INACTIVE,
  FRPKTTYPE_END_
} FrPktType;

/* The priority of a packet (or route): lower values mean higher priority */
typedef unsigned char FrPktPrio;

/*
 *  Packet sizes.
 *  
 *  All packet sizes should be the same, except for route creation.  Ideally,
 *  it too should be the same size.  Maybe in a future version.
 */

/* bug: better value for following */
#define FRPKT_AIP_GENERIC_SIZE       FRCRYPT_SIZE_FLOOR(288)

#define FRPKT_AIP_ALIVE_SIZE         FRPKT_AIP_GENERIC_SIZE
#define FRPKT_AIP_DATA_SIZE          FRPKT_AIP_GENERIC_SIZE
/* bug: better value for following */
#define FRPKT_AIP_ROUTE_CREATE_SIZE  FRCRYPT_SIZE_FLOOR(1240) /* 5 routes */
#define FRPKT_AIP_ROUTE_CRTACK_SIZE  FRPKT_AIP_GENERIC_SIZE
#define FRPKT_AIP_ROUTE_DESTROY_SIZE FRPKT_AIP_GENERIC_SIZE
#define FRPKT_STAT_SIZE              FRPKT_AIP_GENERIC_SIZE
#define FRPKT_NDB_ROUTE_UPDATE_SIZE  FRPKT_AIP_GENERIC_SIZE

/*
 *  So that we can detect cover traffic packets that consist entirely of
 *  random data, we append 8 bytes, each of which contains the number 8.  Upon
 *  link decryption we immediately reject those packets not having these
 *  values.
 *  
 *  Pass the packet size (defined above) to FRPKT_EIGHTS_OFFSET() to get the
 *  offset at which the bytes should be appended.
 */

#define FRPKT_EIGHTS_OFFSET(n) (n)
#define FRPKT_EIGHTS_SIZE      8
#define FRPKT_EIGHTS_VALUE     8


/*
 *  After the trailing eight bytes, we store a link encryption key version
 *  (LEKV) that tells us with which version of a key the packet is encrypted.
 *  
 *  This is not indended only to allow us to continue processing packet while
 *  negotiating a new link level encryption key.
 */

#define FRPKT_LEKV_SIZE      1


/*
 *  Return the size of a buffer to store a particular packet.
 *  
 *  Eight extra bytes store special padding for cover traffic rejection, and
 *  one extra byte is used to store the link encryption key version, which
 *  must appear as the last byte of all packets.  Having a second extra byte
 *  may allow you to determine whether an illegally sized packet was read.
 *  (I.e.  Try to read enough bytes to fill an entire buffer.  The number of
 *  bytes read should be less than the size of the buffer.)
 */

#define FRPKT_BUFFER_SIZE(n) ((n) + FRPKT_EIGHTS_SIZE + 2)

/*
 *  The size of the buffer needed to store the largest valid packet that might
 *  be received.
 */

#define FRPKT_BUF_MAX_SIZE (FRPKT_BUFFER_SIZE(UINT16_MAX))

/*
 *  Packet version number.
 */
#define FRPKT_VERSION_CURRENT 1

/*
 *  Macro definitions.
 */

/*
 *  Macros to copy data from a buffer to a variable, converting from network
 *  byte order to host byte order, and vice versa.
 */

#define BYTE_CPY(buf, n) ((unsigned int)*((unsigned char *)(buf) + (n)))

#define INT8_FROM_NBO(var, buf) do { var = BYTE_CPY(buf, 0); } while (0)
#define NBO_FROM_INT8(buf, var) do { *(char *)(buf) = var; } while (0)

#define INT16_FROM_NBO(var, buf) do { var = (BYTE_CPY(buf, 0) << 8) | BYTE_CPY(buf, 1); } while (0)
#define NBO_FROM_INT16(buf, var) do { uint16_t x = htons(var); \
                                      memcpy(buf, &x, 2); } while (0)

#define INT32_FROM_NBO(var, buf) do { var = (BYTE_CPY(buf, 0) << 24) | \
                                            (BYTE_CPY(buf, 1) << 16) | \
                                            (BYTE_CPY(buf, 2) <<  8) | \
                                             BYTE_CPY(buf, 3);         \
                                    } while (0)
#define NBO_FROM_INT32(buf, var) do { uint32_t x = htonl(var); \
                                      memcpy(buf, &x, 4); } while (0)


#ifdef DEBUG_CKSUM
extern unsigned short g_cksum;
#endif


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

  typedef unsigned int FrAci;

  /*
   *  Routines for general packets.
   */
  
  DLLExport extern int frPktGetVersion(ZkError err, void *data, size_t len, unsigned int *vers);
  DLLExport extern int frPktSetVersion(ZkError err, void *data, size_t len, unsigned int vers);

  DLLExport extern int frPktGetType(ZkError err, void *data, size_t len, FrPktType *type);
  DLLExport extern int frPktSetType(ZkError err, void *data, size_t len, FrPktType type);

  DLLExport extern int frPktGetPriority(ZkError err, void *data, size_t len, unsigned int *pri);
  DLLExport extern int frPktSetPriority(ZkError err, void *data, size_t len, unsigned int pri);

  DLLExport extern int frPktLinkEncrypt(ZkError err, FrCryptSym sym, void *data, size_t len);
  DLLExport extern int frPktLinkDecrypt(ZkError err, FrCryptSym sym, void *data, size_t len);
  DLLExport extern int frPktLinkSmartDecrypt(ZkError err, FrCryptSym sym, void *data, void *output, size_t len);
  DLLExport extern int frPktGetEmbeddedPacket(ZkError err, void *buf, size_t bsz, void **sp, size_t *spsz);
  DLLExport extern int frPktGetLekv(ZkError err, void *data, size_t dsz, uint8_t *lekv);
  DLLExport extern int frPktSetLekv(ZkError err, void *data, size_t dsz, uint8_t lekv);
  DLLExport extern size_t frPktGetSize(size_t epsz);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* FR_PKTFMT_H_ */
