/*
 *  $Id: aipRteCrt.c,v 1.34.4.4 2000/12/14 17:53:09 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  AIP route creation packets.
 */

#define DEBUG_BOUNDS 0
#define DEBUG_USING_PURIFY 0
#define DEBUG_SANITY 0

#include "zks.h"

#include <assert.h>
#include <errno.h>

#ifndef WIN32
#include <netinet/in.h>
#else
#include <winsock2.h>
#endif

#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include "aci.h"
#include "frCrypt.h"
#include "frPktFmt.h"
#include "nymAuth.h"
#include "pkt.h"
#include "openssl/sha.h"
#include "zkError.h"


#define MYMOD ZKMFRPKTFMT

/*
 *  Currently, all symmetric keys are initialized with the same amount of seed
 *  material.  The material may be broken up, depending on the algorithm. bug?
 *  put this in frCrypt.h?
 */

#define SYM_KEY_SEED_SIZE 24


#pragma pack(1)

/*
 *  This defines the route field of a route creation packet.
 *  
 *  The structure's size must be a multiple of eight bytes.
 */

struct fr_route_p {
  char keySeed[FRCRYPT_SECRET_SIZE];
  uint8_t keyAlg;
  uint8_t cryptKeyVers;
  char pad0[6];

  /*
   *  Layered symmetric encryption begins here.  This offset (within the
   *  structure) should be a mutliple of eight bytes.
   */
  
  uint8_t bckKeyAlg;
  uint8_t fwdKeyAlg;
  uint8_t flags;
  char hostEnt[FRENT_PORTABLE_SIZE];
  uint16_t specAci;
  uint32_t expTime;
  char pad1[6];
};

struct fr_nymauth_p {
  char serialNum[8];
  uint8_t nymType;
  char nymName[64];
  char nymSig[FRCRYPT_SIG_SIZE];
  char pad0[6];
};

/*
 *  This defines a route creation packet.
 */
struct fr_rtecrt_p {
  /*
   *  The standard packet header.
   */
  
  uint8_t vers;
  uint8_t type;
  uint8_t pri;
  char pktspec[3];

  uint16_t aci;

  /*
   *  The routes, followed by the nym authentication field.
   *  
   *  We can't define this as an array of route structures, followed by a nym
   *  authentication structure because each time we extract a route we need to
   *  shift the whole thing forward by the size of a route structure.
   */
  char rteStart[FRPKT_NUM_ROUTES_MAX * sizeof(struct fr_route_p) +
               sizeof(struct fr_nymauth_p)];
};

#pragma pack()


struct fr_pkt_rte_crt {
  size_t nextRoute; /* index of the next route to be appended */
  FrCryptSym rteCrypt[FRPKT_NUM_ROUTES_MAX]; /* encryption keys for the routes portion of the packet */
  char ivec[FRPKT_NUM_ROUTES_MAX][FRCRYPT_IVEC_SIZE]; /* IVs for onionizing */
  SHA_CTX shaCtx;
  unsigned char hash[20];
  char *buf; /* pointer to memory holding the AIP route creation packet */
  size_t bsz; /* size of the buffer into which the packet is written */
  struct fr_rtecrt_p *pkt; /* buffer interpreted as a packed route creation structure */
};


/*
 *
 *
 *                            Internal Routines
 *
 *
 */


#if DEBUG_SANITY

static void checkSizesOffsets(void)
{
  int ok = 1;
  static int checked = 0;
  struct fr_rtecrt_p rc;
  struct fr_route_p rte;
  struct fr_nymauth_p na;
  
  if ( ! checked) {
    fprintf(stderr, "struct fr_nymauth_p: %u bytes %s.\n",
            sizeof na, sizeof na % 8 ? "bad" : "ok");
    fprintf(stderr, "struct fr_rout_p:    %u bytes %s.\n",
            sizeof rte, sizeof rte % 8 ? "bad" : "ok");
    fprintf(stderr, "struct fr_rtecrt_p:  %u bytes %s.\n",
            sizeof rc, sizeof rc % 8 ? "bad" : "ok");
    
    if (sizeof rc % 8 != 0) {
      fprintf(stderr, "Route creation packet has bad size.\n"); ok = 0;
    }
    if (offsetof(struct fr_rtecrt_p, rteStart) % 8 != 0) {
      fprintf(stderr, "Bad route start offset.\n"); ok = 0;
    }
    if (sizeof rc.rteStart % 8 != 0) {
      fprintf(stderr, "Bad route start size.\n"); ok = 0;
    }
    if (sizeof rte != 168) {
      fprintf(stderr, "Mistake in calculating the route size.\n"); ok = 0;
    }
    if (sizeof rte % 8 != 0) {
      fprintf(stderr, "Route has bad size.\n"); ok = 0;
    }
    if (offsetof(struct fr_route_p, bckKeyAlg) % 8 != 0) {
      fprintf(stderr, "Bad bckKeyOffset.\n"); ok = 0;
    }
    if (sizeof na % 8 != 0) {
      fprintf(stderr, "Nym authentication has bad size.\n"); ok = 0;
    }

    if (ok) {
      fprintf(stderr, "AIP route creation packet sanity check passed.\n");
    } else {
      fprintf(stderr, "AIP route creation packet sanity check failed -- aborting.\n");
      abort();
    }

    checked = 1;
  }
}

#endif


static void rteCrtInit(FrPktRC r)
{
  size_t i;
  
  r->nextRoute = 0;
  for (i = 0; i < FRPKT_NUM_ROUTES_MAX; i++) {
    r->rteCrypt[i] = 0;
    memset(r->ivec[i], 0, sizeof r->ivec[i]);
  }
  r->buf = 0;
  r->bsz = 0;
  r->pkt = 0;
}


static int rteCrtReset(ZkError err, FrPktRC pkt)
{
  int ret = 0;
  size_t i;
  
  for (i = 0; i < pkt->nextRoute; i++) {
    if (pkt->rteCrypt[i] && ! frCryptSymFree(err, &pkt->rteCrypt[i])) {
      goto done;
    }
  }

  rteCrtInit(pkt);
  ret = 1;

 done:
  
  return ret;
}


/*
 *  Generate the symmetric key to crypt the remainder of a route.
 *  
 *  Use the last portion of the seed as an initialization vector.
 */
static int keyFromSeed(ZkError err, FrCryptSym key, void *seed, size_t sz)
{
  const char *fn = "keyFromSeed";
  int ret = 0;
  
  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "key is null");
    goto done;
  }
  if ( ! seed) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "seed is null");
    goto done;
  }
  if (sz != FRCRYPT_SYMKEY_SEED_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "sz has an invalid value (%lu != %d)",
                  (unsigned long)sz, FRCRYPT_SYMKEY_SEED_SIZE);
    goto done;
  }
  
  if ( ! frCryptSymKeyFromData(err, key, seed, sz)) {
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 *  Generate key seed material from the pre-key seed material and our private
 *  key.
 *  
 *  We use this to generate the same key seed material as the sender, so that
 *  we can generate the same keys that they did.
 */
static int keySeedFromPrivKey(ZkError err, FrCryptCrypt privKey, void *pks, size_t sz, void *ks, size_t ksz)
{
  return frCryptCryptSecretFromPrivKey(err, privKey, pks, sz, ks, ksz);
}


/*
 *  Generate two symmetric keys (backward and forward) from the key seed
 *  material.
 *  
 *  `ks' contains the key seed material, and has length `ksSz'.
 */
static int keysFromSeed(ZkError err, unsigned char *ks, size_t ksSz, FrCryptSym k0, FrCryptSym k1)
{
  const char *fn = "keysFromSeed";
  int ret = 0;
  
  if ( ! ks) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "ks is null");
    goto done;
  }
  if (ksSz < 2 * FRCRYPT_SYMKEY_SEED_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "ksSz is too small (%lu < %d)",
                  (unsigned long)ksSz, 2 * FRCRYPT_SYMKEY_SEED_SIZE);
    goto done;
  }
  if ( ! k0) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "k0 is null");
    goto done;
  }
  if ( ! k1) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "k1 is null");
    goto done;
  }

  /*
   *  Convert the pre-key seed material into key seed material.
   */

  if ( ! frCryptSymKeyFromData(err, k0, ks, FRCRYPT_SYMKEY_SEED_SIZE)) {
    goto done;
  }

  ks += FRCRYPT_SYMKEY_SEED_SIZE;

  if ( ! frCryptSymKeyFromData(err, k1, ks, FRCRYPT_SYMKEY_SEED_SIZE)) {
    goto done;
  }

  ret = 1;
  
 done:
  
  return ret;
}


/*
 *  Shift the routes and the nym authentication field up by one. We assume
 *  that all route creation packets have space for FRPKT_NUM_ROUTES_MAX
 *  routes, even if it is not all used.
 *  
 *  We don't random data to the new, `empty' route field at the end.
 */
static int shiftRoutes(ZkError err, void *routes)
{
  char *off = (char *)routes + sizeof(struct fr_route_p);
  int ret = 0;
  size_t sz = FRCRYPT_SIZE_CEIL((FRPKT_NUM_ROUTES_MAX - 1) * sizeof(struct fr_route_p) +
                                sizeof(struct fr_nymauth_p));

  /* shift the data */
  memmove(routes, off, sz);

  /* 
   *  replace the space no longer used (of sizeof(struct fr_route_p) at the
   *  end of 'routes' by random data.  Note: the rest has already been padded
   *  by random data at the creation of routes.
   */

  if ( ! zkRandGetData(err, (void *)((char *)routes + sz),
                              sizeof(struct fr_route_p))) {
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 *
 *
 *                            External Routines
 *
 *
 */


/*
 *  Return the maximum size of an AIP route create packet having version
 *  `vers'.
 *  
 *  This size is for the route create packet only; it doesn't include anything
 *  belonging to the link layer, such as leading random bytes, trailing 8s or
 *  lekv.
 *  
 *  Note: this routine replaces frPktAipRteCrtBufferSize().
 */
DLLExport int frPktAipRteCrtGetMaxSize(ZkError err, unsigned int vers, size_t *sz)
{
  const char *fn = "frPktAipRteCrtGetMaxSize";
  int ret = 0;
  
  if (vers != 1 && vers != 2) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "invalid version %u", vers);
    goto done;
  }
  if ( ! sz) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sz is null");
    goto done;
  }

  *sz = sizeof(struct fr_rtecrt_p);
  ret = 1;
  
 done:

  return ret;
}


DLLExport int frPktAipRteCrtNew(ZkError err, FrPktRC *pkt)
{
  FrPktRC p;
  const char *fn = "frPktAipRteCrtNew";
  int ret = 0;
  
#if DEBUG_SANITY
  checkSizesOffsets();
#endif
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }

  if ( ! (p = malloc(sizeof *p))) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKENOMEM,
               "can't allocate new %lu byte route creation object: %s",
               (unsigned long)sizeof *p, strerror(errno));
    goto done;
  }

  rteCrtInit(p);

  *pkt = p;
  ret = 1;

 done:
  
  return ret;
}


DLLExport int frPktAipRteCrtFree(ZkError err, FrPktRC *pkt)
{
  FrPktRC p;
  const char *fn = "frPktAipRteCrtFree";
  int ret = 0;
  size_t i;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! (p = *pkt)) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKEPARMBAD, "*pkt is null");
    goto done;
  }

  for (i = 0; i < p->nextRoute; i++) {
    if ( ! frCryptSymFree(err, &p->rteCrypt[i])) {
      goto done;
    }
  }

  rteCrtInit(p);
  free(p); *pkt = 0;

  ret = 1;

 done:
  
  return ret;
}


/*
 *  Specify the buffer, and its size, to be used to create new route create
 *  packets.
 */
DLLExport int frPktAipRteCrtCreateInBuffer(ZkError err, FrPktRC pkt, void *buf, size_t bsz)
{
  const char *fn = "frPktAipRteCrtCreateInBuffer";
  int ret = 0;
  size_t minsz;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! buf) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "buf is null");
    goto done;
  }

  /*
   *  Use the size of a version 2 route create packet.  (Version 1 is the same
   *  size.)
   */
  
  if ( ! frPktAipRteCrtGetMaxSize(err, 2, &minsz)) {
    goto done;
  }
  
  if (bsz < minsz) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "buffer is too small (%lu < %lu)",
                  (unsigned long)bsz, (unsigned long)minsz);
    goto done;
  }
  
  if ( ! rteCrtReset(err, pkt)) {
    goto done;
  }

  pkt->buf = buf;
  pkt->bsz = bsz;
  pkt->pkt = (void *)pkt->buf;

  ret = 1;

 done:

  return ret;
}


DLLExport int frPktAipRteCrtReceiveInBuffer(ZkError err, FrPktRC pkt, void *buf, size_t bsz)
{
  const char *fn = "frPktAipRteCrtReceiveInBuffer";
  int ret = 0;
  size_t minsz;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }

  /*
   *  Use the size of a version 2 route create packet.  (Version 1 is the same
   *  size.)
   */
  
  if ( ! frPktAipRteCrtGetMaxSize(err, 2, &minsz)) {
    goto done;
  }

  if (bsz < minsz) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "buffer is too small (%lu < %lu)",
                  (unsigned long)bsz, (unsigned long)minsz);
    goto done;
  }

  if ( ! rteCrtReset(err, pkt)) {
    goto done;
  }

  pkt->buf = buf;
  pkt->bsz = bsz;
  pkt->pkt = (void *)pkt->buf;

  ret = 1;

 done:
  
  return ret;
}


DLLExport int frPktAipRteCrtSetAci(ZkError err, FrPktRC pkt, FrAci aci)
{
  const char *fn = "frPktAipRteCrtSetAci";
  int ret = 0;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }

  if ( ! pkt->buf) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKEBADSTATE,
               "the AIP route creation packet has not been properly initialized");
    goto done;
  }

  pkt->pkt->aci = htons(aci);
  ret = 1;

 done:
  
  return ret;
}


DLLExport int frPktAipRteCrtGetAci(ZkError err, FrPktRC pkt, FrAci *aci)
{
  const char *fn = "frPktAipRteCrtGetAci";
  int ret = 0;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! aci) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "aci is null");
    goto done;
  }

  if ( ! pkt->buf) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKEBADSTATE,
               "the AIP route creation packet has not been properly initialized");
    goto done;
  }

  *aci = ntohs(pkt->pkt->aci);
  ret = 1;

 done:

  return ret;
}

 
/*
 *  Add the route given by `*r' to the route creation packet.  Route numbering
 *  begins at 0.
 */
DLLExport int frPktAipRteCrtAppendRouteExp(ZkError err, FrPktRC pkt, FrCryptCrypt aipPubKey, struct fr_rte_set_exp *r)
{
  const char *fn = "frPktAipRteCrtAppendRouteExp";
  int ret = 0;
  struct fr_route_p *rte; /* pointer, into the packet buffer, to the next route */
  unsigned char ksm[FRCRYPT_SECRET_SIZE], *k = ksm;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! aipPubKey) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "aipPubKey is null");
    goto done;
  }
  if ( ! r) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "r is null");
    goto done;
  }

  if (pkt->nextRoute >= FRPKT_NUM_ROUTES_MAX) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEBADSTATE,
                  "number of routes exceeds maximum (%d >= %d)",
                  pkt->nextRoute, FRPKT_NUM_ROUTES_MAX);
    goto done;
  }

  rte = (void *)(pkt->pkt->rteStart + pkt->nextRoute * sizeof *rte);

  /*
   *  Copy the cryption key version into the route header.
   */

  {
    unsigned int kv;
    
    if ( ! frCryptPubKeyGetVersion(err, aipPubKey, &kv)) {
      goto done;
    }
    rte->cryptKeyVers = kv;
  }

  /*
   *  Generate some pre-key seed material, then pack in the contents of the
   *  route structure.  At the same time the key seed material is stored in
   *  ksm, so we can generate the same keys as the receiving end.
   */

  if ( ! frCryptCryptSetPreSecret(err, aipPubKey,
                                  rte->keySeed, sizeof rte->keySeed,
                                  ksm, sizeof ksm)) {
    goto done;
  }

  rte->keyAlg = r->rteAlg;

  /* bug: assumes < 8 */
  memset(rte->pad0, sizeof rte->pad0, sizeof rte->pad0);

  {
    FrCryptSymAlg ba, fa;
    
    if ( ! (frCryptSymGetAlg(err, r->bckCrypt, &ba) &&
            frCryptSymGetAlg(err, r->fwdCrypt, &fa))) {
      goto done;
    }
    rte->bckKeyAlg = ba; rte->fwdKeyAlg = fa;
  }

  /*
   *  Create a new encryption key to encrypt the rest of the packet.  Also set
   *  the backward and forward route encryption keys.
   */
  
  if ( ! (frCryptSymNew(err, &pkt->rteCrypt[pkt->nextRoute]) &&
          frCryptSymSetAlg(err, pkt->rteCrypt[pkt->nextRoute], r->rteAlg))) {
    goto done;
  }

  /* Skip over the first 2 bytes of the key seed material. */
  
  k += 2;

  if ( ! keyFromSeed(err, pkt->rteCrypt[pkt->nextRoute], k, FRCRYPT_SYMKEY_SEED_SIZE)) {
    goto done;
  }

  k += FRCRYPT_SYMKEY_SEED_SIZE;

  /*
   *  Save the next few bytes, after the initial seed, should the cryption
   *  algorithm require an initialization vector.
   */

  memcpy(pkt->ivec[pkt->nextRoute], k, FRCRYPT_IVEC_SIZE);

  k += FRCRYPT_IVEC_SIZE;

  if ( ! keysFromSeed(err, k, sizeof ksm - (k - &ksm[0]), r->bckCrypt, r->fwdCrypt)) {
    goto done;
  }

  k += 2 * FRCRYPT_SYMKEY_SEED_SIZE;

  /* Secret MAC key */

  if ( ! frCryptMacKeyLoad(err, r->macKey, k, FRCRYPT_MACKEY_SIZE)) {
    goto done;
  }

#if DEBUG_BOUNDS
  assert(k < &ksm[sizeof ksm]);
#endif

  if ( ! frEntStore(err, &r->hostEnt, rte->hostEnt, sizeof rte->hostEnt)) {
    goto done;
  }

  /* Expiration time. */

  rte->expTime = htonl(r->expTime);

  /*
   *  Perform a sanity check on the flags field.
   */
  
  if (r->flags & ~(FRPKT_RTECRT_FL_LAST_RTE | FRPKT_RTECRT_FL_SPEC_ACI)) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEDATABAD,
                  "invalid flag (0x%x)", r->flags);
    goto done;
  }

  /*
   *  Flags indicating if this is the last route and/or whether to use a
   *  specific ACI.
   */

  rte->flags = r->flags;

  if (r->flags & FRPKT_RTECRT_FL_SPEC_ACI) {
    rte->specAci = htons(r->specAci);
  } else {
    /* Set each of the two bytes to 2. */
    memset(&rte->specAci, sizeof rte->specAci, sizeof rte->specAci);
  }

  memset(rte->pad1, sizeof rte->pad1, sizeof rte->pad1); /* bug: assumes < 8 */

  /*
   *  bug: if we're the last route begin the generation of a hash over the
   *  route, but substituting the final key seed material for the pre-key seed
   *  material.  The hash will be continued when we append the nym
   *  authentication, and it will then be signed.
   */

  if (r->flags & FRPKT_RTECRT_FL_LAST_RTE) {
    /* bug: we should hide this in some frCryptSig function. */
    SHA1_Init(&pkt->shaCtx);
    SHA1_Update(&pkt->shaCtx, (unsigned char *)ksm, sizeof ksm);
    SHA1_Update(&pkt->shaCtx,
                (unsigned char *)&rte->keyAlg,
                sizeof *rte - offsetof(struct fr_route_p, keyAlg));
  }

  /* Trash the key seed material. */
  
  memset(ksm, 0, sizeof ksm);

  pkt->nextRoute++;

  ret = 1;

 done:
  
  return ret;
}


/*
 *  Before extracting the next route, the caller needs to know which version
 *  of its cryption key was used in creating the route.
 */
DLLExport int frPktAipRteCrtGetCryptKeyVersion(ZkError err, FrPktRC pkt, unsigned int *vers)
{
  const char *fn = "frPktAipRteCrtGetCryptKeyVersion";
  int ret = 0;
  struct fr_route_p *rte;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! vers) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "vers is null");
    goto done;
  }

  rte = (void *)pkt->pkt->rteStart;
  
  *vers = rte->cryptKeyVers;
  ret = 1;

 done:

  return ret;
}


/*
 *  bug: we should assume the maximum number of routes, extract ours, shift
 *  the others up by one route (not including the NymAuth, then scribble
 *  random data in the last route position.
 */
DLLExport int frPktAipRteCrtGetNextRouteExp(ZkError err, FrPktRC pkt, FrCryptCrypt aipPrivKey, struct fr_rte_get_exp *r)
{
  FrCryptSymAlg symAlg;
  const char *fn = "frPktAipRteCrtGetNextRouteExp";
  int ret = 0;
  struct fr_route_p *rte;
  unsigned char ksm[FRCRYPT_SECRET_SIZE];
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! aipPrivKey) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "aipPrivKey is null");
    goto done;
  }
  if ( ! r) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "r is null");
    goto done;
  }

  /*
   *  Verify that the unused padding is valid.
   */

  if ( ! fr_CryptCheckPad(pkt->pkt->pktspec, sizeof pkt->pkt->pktspec)) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKEBADSTATE,
               "pad0 field is invalid");
    return 0;
  }

  /*
   *  Assume the next route we want is first in the list of routes.  After
   *  we've extracted it, we'll shift all the other routes up by one location.
   */

  rte = (void *)pkt->pkt->rteStart;
  
  /*
   *  Generate the key seed material with our private key, create a
   *  symmetric key then decrypt from the backward symmetric key to the end of
   *  the entire route creation packet.
   */

  {
    FrCryptSym rteCrypt;
    char ivec[FRCRYPT_IVEC_SIZE];
    unsigned char *k = &ksm[0];
    
    /* bug? trash parts of this as we use it up? */

    if ( ! keySeedFromPrivKey(err, aipPrivKey,
                              rte->keySeed, sizeof rte->keySeed,
                              k, sizeof ksm)) {
      goto done;
    }

    /* Skip over the first 2 bytes of the key seed material. */
    
    k += 2;

    if ( ! frCryptSymNew(err, &rteCrypt)) {
      goto done;
    }

    symAlg = rte->keyAlg;
    if ( ! frCryptSymSetAlg(err, rteCrypt, symAlg)) {
      goto done;
    }

    if ( ! keyFromSeed(err, rteCrypt, k, FRCRYPT_SYMKEY_SEED_SIZE)) {
      goto done;
    }

    k += FRCRYPT_SYMKEY_SEED_SIZE;

    /*
     *  Save the next few bytes, after the initial seed, should the cryption
     *  algorithm require an initialization vector.
     *  
     *  bug: we could avoid the copy and just store a pointer to the ivec.
     */

    memcpy(ivec, k, FRCRYPT_IVEC_SIZE);

    k += FRCRYPT_IVEC_SIZE;

    if ( ! frCryptSymDecrypt(err, rteCrypt, ivec,
                             &rte->bckKeyAlg, /* bug: double check next line */
                             sizeof *(pkt->pkt)
                             - ((char *)&rte->bckKeyAlg - pkt->buf))) {
      goto done;
    }

    if ( ! frCryptSymFree(err, &rteCrypt)) {
      goto done;
    }

    symAlg = rte->bckKeyAlg;
    if ( ! frCryptSymSetAlg(err, r->bckCrypt, symAlg)) {
      goto done;
    }

    symAlg = rte->fwdKeyAlg;
    if ( ! frCryptSymSetAlg(err, r->fwdCrypt, symAlg)) {
      goto done;
    }

    if ( ! keysFromSeed(err, k, sizeof ksm - (k - &ksm[0]), r->bckCrypt, r->fwdCrypt)) {
      goto done;
    }

    k += 2 * FRCRYPT_SYMKEY_SEED_SIZE;

    /* Secret MAC key */

    if ( ! frCryptMacKeyLoad(err, r->macKey, k, FRCRYPT_MACKEY_SIZE)) {
      goto done;
    }

#if DEBUG_BOUNDS
    assert(k < &ksm[sizeof ksm]);
#endif
  }

  /* Do some checks before proceeding. */

  if ( ! (fr_PktMemCheck(rte->pad0, sizeof rte->pad0, sizeof rte->pad0) &&
          fr_PktMemCheck(rte->pad1, sizeof rte->pad1, sizeof rte->pad1))) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKEBADSTATE,
               "padding has invalid value");
    goto done;
  }

  if ( ! frEntLoad(err, &r->hostEnt, rte->hostEnt, sizeof rte->hostEnt)) {
    goto done;
  }

  r->expTime = ntohl(rte->expTime);

  r->flags = rte->flags;

  if (r->flags & ~(FRPKT_RTECRT_FL_LAST_RTE | FRPKT_RTECRT_FL_SPEC_ACI)) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "invalid flags (0x%x)", (unsigned int)r->flags);
    goto done;
  }

  if (r->flags & FRPKT_RTECRT_FL_SPEC_ACI) {
    r->specAci = ntohs(rte->specAci);
  }

  /*
   *  As when appending, if we're the last route begin the generation of a
   *  hash over the route, but substituting the final key seed material for
   *  the pre-key seed material.  The hash will be continued when we extract
   *  the nym authentication, and the signature will then be verified.
   */

  if (r->flags & FRPKT_RTECRT_FL_LAST_RTE) {
    /* bug: we should hide this in some frCryptSig function. */
    SHA1_Init(&pkt->shaCtx);
    SHA1_Update(&pkt->shaCtx, ksm, sizeof ksm);
    SHA1_Update(&pkt->shaCtx,
                (unsigned char *)&rte->keyAlg,
                sizeof *rte - offsetof(struct fr_route_p, keyAlg));
  }

  /* Trash the key seed material. */
  
  memset(ksm, 0, sizeof ksm);

  if ( ! shiftRoutes(err, rte)) {
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 *  Put the nym authentication after the last route.
 *  
 *  The caller must fill in the nymID and nymSig fields of the nymAuth
 *  argument.  For nymID, both the enity type (typically FRENT_TYPE_NYM) and
 *  entity name must be set.  The nymSig field is that nym's signing key.
 *  
 *  No other fields need be set by the caller.
 *  
 *  We generate a random serial number, then sign the last route (including
 *  the full shared secret, _not_ the pre-secret), the serial number and nym
 *  ID.
 */
DLLExport int frPktAipRteCrtAppendNymAuthExp(ZkError err, FrPktRC pkt, struct fr_nym_auth_exp *nymAuth)
{
  const char *fn = "frPktAipRteCrtAppendNymAuthExp";
  struct fr_nymauth_p *na;
  int ret = 0;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! nymAuth) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "nymAuth is null");
    goto done;
  }

  /*
   *  The nym authentication field starts just beyond the current route.
   */

  na = (void *)(pkt->pkt->rteStart + pkt->nextRoute * sizeof(struct fr_route_p));

  /*
   *  Generate a random serial number to be used at the exit funnel to detect
   *  replay attacks.
   */

  if ( ! zkRandGetData(err, na->serialNum, sizeof na->serialNum)) {
    goto done;
  }

  /* Copy the nym identifier. */

  na->nymType = nymAuth->nymID.type;
  memcpy(na->nymName, nymAuth->nymID.name, sizeof na->nymName);

  /*
   *  Complete the hash over the last route and nym authentication, then sign
   *  it.
   */

  {
    unsigned char hash[20];
    
    SHA1_Update(&pkt->shaCtx, (unsigned char *)na,
                offsetof(struct fr_nymauth_p, nymSig));
    SHA1_Final(hash, &pkt->shaCtx);
  
    if ( ! frCryptSigSignHash(err, nymAuth->nymSig, hash, sizeof hash,
                              na->nymSig, sizeof na->nymSig)) {
      memset(hash, 0, sizeof hash);
      goto done;
    }
    memset(hash, 0, sizeof hash);
  }
  
  memset(na->pad0, sizeof na->pad0, sizeof na->pad0);

  ret = 1;

 done:

  return ret;
}


/*
 *  Get the nym authentication field from the route creation packet.  This
 *  should be called when an AIP has extracted its route, and determined that
 *  it is the exit funnel.
 *  
 *  This routine sets the serialNum, nymID, hash and hashsig fields of
 *  nymAuth.
 *  
 *  The serialNum field contains a random serial number that should be unique
 *  for the lifetime of this route.  That is, it is used to prevent a replay
 *  attack, in which a copy of the route creation packet (or a copy of the
 *  last route and nym authentication field) is again sent to the exit funnel.
 *  The exit funnel should reject route creation packets, for this nym, with
 *  duplicate serial numbers.
 *  
 *  The hash and hashsig fields are used internally, for authenticating the
 *  nym.
 *  
 *  After calling this, frPktAipRteCrtAuthenticateNym() should be called, to
 *  perform the actual authentication of the nym.
 */
DLLExport int frPktAipRteCrtGetNymAuthExp(ZkError err, FrPktRC pkt, struct fr_nym_auth_exp *nymAuth)
{
  const char *fn = "frPktAipRteCrtGetNymAuthExp";
  int ret = 0;
  struct fr_nymauth_p *na;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! nymAuth) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "nymAuth is null");
    goto done;
  }

  /*
   *  The last route will have been shifted out, so the nym authentication
   *  structure is at the beginning of the routes area.
   */
  na = (void *)pkt->pkt->rteStart;

  /* Copy the random serial number. */

  memcpy(nymAuth->serialNum, na->serialNum, sizeof na->serialNum);

  /* Copy the nym identifier. */

  nymAuth->nymID.type = na->nymType;
  memcpy(nymAuth->nymID.name, na->nymName, sizeof nymAuth->nymID.name);

  /* Get the version of the nym's signature key. */

  if ( ! frCryptSigBufGetVersion(err, na->nymSig, sizeof na->nymSig,
                                 &nymAuth->sigKeyVersion)) {
    goto done;
  }

  /* Verify the pad0 field. */

  if ( ! fr_PktMemCheck(na->pad0, sizeof na->pad0, sizeof na->pad0)) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKEDATABAD,
               "invalid nym auth. pad0 field");
    goto done;
  }

  /*
   *  Complete the hash over the last route and nym authentication.
   */

  SHA1_Update(&pkt->shaCtx, (unsigned char *)na, offsetof(struct fr_nymauth_p, nymSig));
  SHA1_Final(pkt->hash, &pkt->shaCtx);

  /*
   *  Copy the hash and signature over the hash, so that we need only
   *  `nymAuth' to perform the nym authentication.  That is, we can safely
   *  trash, or reuse, the route creation packet object, as well as its
   *  buffer.
   */

  memcpy(nymAuth->hash, pkt->hash, sizeof nymAuth->hash);
  memcpy(nymAuth->hashsig, na->nymSig, sizeof nymAuth->hashsig);

  memset(pkt->hash, 0, sizeof pkt->hash);

  ret = 1;

 done:
  
  return ret;
}


/*
 *  Authenticate the nym appearing in the nymAuth argument.
 *  
 *  The value of `*isAuth' will be set to 1 if the nym should be allowed to
 *  create the route otherwise it will be set to 0.
 *  
 *  Note that this routine does not take an FrPktRC argument.  Typically, this
 *  routine will be called after waiting for the results of a key query, and
 *  we want to be able to reuse the route creation packet object, and its
 *  buffer, in the meantime.
 */
DLLExport int frPktAipRteCrtAuthenticateNym(ZkError err, struct fr_nym_auth_exp *nymAuth, int *isAuth)
{
  const char *fn = "frPktAipRteCrtAuthenticateNym";
  int ret = 0;

  if ( ! nymAuth) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "nymAuth is null");
    goto done;
  }
  if ( ! isAuth) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "isAuth is null");
    goto done;
  }

#ifndef WIN32
#warning "Wed Aug 18 10:58:12 EDT 1999: we should check that the signature key"
#warning "hasn't expired, but due to the short expiration date on Beta-2 nyms"
#warning "we won't do this now."
#endif

  if ( ! frCryptSigVerifyHash(err, nymAuth->nymSig, nymAuth->hash, sizeof nymAuth->hash,
                              nymAuth->hashsig, sizeof nymAuth->hashsig, isAuth)) {
    goto done;
  }

  ret = 1;

 done:
  
  /*
   *  Trash the buffers on general principles.
   */
  
  memset(nymAuth->hash, 0, sizeof nymAuth->hash);
  memset(nymAuth->hashsig, 0, sizeof nymAuth->hashsig);

  return ret;
}


/*
 *  Add layers of encryption to the packet so that each AIP can access only
 *  the information it needs. Pad with randomness at the end.
 */
DLLExport int frPktAipRteCrtEncryptAllRoutes(ZkError err, FrPktRC pkt)
{
  const char *fn = "frPktAipRteCrtEncryptAllRoutes";
  int ret = 0;
  size_t i;
  size_t end;
  size_t lPad;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }

  if (pkt->nextRoute > FRPKT_NUM_ROUTES_MAX) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEBADSTATE,
                  "nextRoute field has too large a value (%d > %d)",
                  pkt->nextRoute, FRPKT_NUM_ROUTES_MAX);
    goto done;
  }

  /*
   *  Initialize the unused `pktspec' field.
   */

  fr_CryptSetPad(pkt->pkt->pktspec, sizeof pkt->pkt->pktspec);

  /*
   *  Random pad starts at the end of fr_nymauth_p.
   */
  
  end = offsetof(struct fr_rtecrt_p, rteStart) +
           ((pkt->nextRoute)*sizeof(struct fr_route_p)) +
           (sizeof(struct fr_nymauth_p));
  
  lPad = FRCRYPT_SIZE_CEIL(sizeof(struct fr_rtecrt_p) - end);

  if ( ! zkRandGetData(err, (void *)(pkt->buf + end), lPad)) {
    goto done;
  }

  
  /*
   *  Encrypt the routes, from last to first.
   *  
   *  Encryption for a particular route begins at the backward key algorithm
   *  field, and continues to the end of the packet.
   */

  for (i = 0; i < pkt->nextRoute; i++) {
    size_t j = pkt->nextRoute - 1 - i; /* index the routes, from last to first */
    size_t o = (offsetof(struct fr_rtecrt_p, rteStart) +
                j * sizeof(struct fr_route_p) + offsetof(struct fr_route_p, bckKeyAlg));
    size_t s = sizeof(struct fr_rtecrt_p) - o;

    /*
     *  Encryption begins at `o' bytes into pkt->buf and continues until the
     *  end of the packet (i.e. for `s' bytes).
     */

    if ( ! frCryptSymEncrypt(err, pkt->rteCrypt[j], pkt->ivec[j], pkt->buf + o, s)) {
      goto done;
    }
  }

  ret = 1;

 done:
  
  return ret;
}


/*
 *  Return the size of an AIP route create packet, not including the leading
 *  random bytes, the trailing eights and LEKV.
 *  
 *  Since version 2 packets haven't changed from version 1, and they're a
 *  fixed size, we just call the existing function frPktAipRteCrtGetMaxSize().
 */
DLLExport int frPktAipRteCrtGetCurrentSize(ZkError err, FrPktRC pkt, size_t *sz)
{
  (void)pkt;
  return frPktAipRteCrtGetMaxSize(err, 1, sz);
}
