/*
 *  $Id: pkt.c,v 1.22.8.5 2000/12/14 17:53:10 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#define DEBUG_TREAT_DESTROY_AS_DATA 1
#define MARIO_HACK 0

#include "zks.h"

#include <stdio.h>
#include <string.h>
#include "frPktFmt.h"
#include "zkIntTypes.h"
#include "pkt.h"
#include "zkrand.h"


#define PKTFMT_MAJOR 1
#define PKTFMT_MINOR 0


/*
 *  The amount of data used by link overhead: the leading random data,
 *  trailing 8s and the link encryption key version (lekv).
 */
const size_t linkOverhead = FRCRYPT_RAND_SIZE + FRPKT_EIGHTS_SIZE + FRPKT_LEKV_SIZE;


/*
 *
 *
 *                        Internal library routines.
 *
 *
 */

/*
 *  Return 1 if all `sz' bytes, starting at `data', contain `value'.
 */
DLLExport int fr_PktMemCheck(void *data, char value, size_t sz)
{
  char *d = data;
  size_t i;
  
  for (i = 0; i < sz; i++) {
    if (d[i] != value) { return 0; }
  }
  return 1;
}


/*
 *
 *
 *                            External routines.
 *
 *
 */

DLLExport int frPktGetVersion(ZkError err, void *data, size_t len, unsigned int *vers)
{
  uint8_t v;

  zkErrorClear(err);
  
  if ( ! data) {
    zkErrorSet(err, "frPktGetVersion", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (len < STD_HDR_OFFSET + STD_HDR_SIZE) {
    zkErrorSet_va(err, "frPktGetVersion", ZKERR, ZKMFRPKTFMT, ZKEPARMBAD,
                  "len too small to include standard header (%lu < %d)",
                  (unsigned long)len, STD_HDR_OFFSET + STD_HDR_SIZE);
    return 0;
  }
  if ( ! vers) {
    zkErrorSet(err, "frPktGetVersion", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "vers is null");
    return 0;
  }

  INT8_FROM_NBO(v, (char *)data + STD_HDR_VERSION_OFFSET);
  *vers = v;

  return 1;
}


DLLExport  int frPktSetVersion(ZkError err, void *data, size_t len, unsigned int vers)
{
  uint8_t v;

  zkErrorClear(err);
  
  if ( ! data) {
    zkErrorSet(err, "frPktSetVersion", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (len < STD_HDR_OFFSET + STD_HDR_SIZE) {
    zkErrorSet_va(err, "frPktSetVersion", ZKERR, ZKMFRPKTFMT, ZKEPARMBAD,
                  "len too small to include standard header (%lu < %d)",
                  (unsigned long)len, STD_HDR_OFFSET + STD_HDR_SIZE);
    return 0;
  }

  v = vers;
  NBO_FROM_INT8((char *)data + STD_HDR_VERSION_OFFSET, v);

  return 1;
}


DLLExport int frPktGetType(ZkError err, void *data, size_t len, FrPktType *type)
{
  uint8_t t;

  zkErrorClear(err);
  
  if ( ! data) {
    zkErrorSet(err, "frPktGetType", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (len < STD_HDR_OFFSET + STD_HDR_SIZE) {
    zkErrorSet_va(err, "frPktGetType", ZKERR, ZKMFRPKTFMT, ZKEPARMBAD,
                  "len too small to include standard header (%lu < %d)",
                  (unsigned long)len, STD_HDR_OFFSET + STD_HDR_SIZE);
    return 0;
  }
  if ( ! type) {
    zkErrorSet(err, "frPktGetType", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "type is null");
    return 0;
  }

  INT8_FROM_NBO(t, (char *)data + STD_HDR_TYPE_OFFSET);
  *type = t;

  return 1;
}


DLLExport int frPktSetType(ZkError err, void *data, size_t len, FrPktType type)
{
  uint8_t t;

  zkErrorClear(err);
  
  if ( ! data) {
    zkErrorSet(err, "frPktSetType", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (len < STD_HDR_OFFSET + STD_HDR_SIZE) {
    zkErrorSet_va(err, "frPktSetType", ZKERR, ZKMFRPKTFMT, ZKEPARMBAD,
                  "len too small to include standard header (%lu < %d)",
                  (unsigned long)len, STD_HDR_OFFSET + STD_HDR_SIZE);
    return 0;
  }

  t = type;
  NBO_FROM_INT8((char *)data + STD_HDR_TYPE_OFFSET, t);

  return 1;
}


DLLExport int frPktGetPriority(ZkError err, void *data, size_t len, unsigned int *pri)
{
  uint8_t p;

  zkErrorClear(err);
  
  if ( ! data) {
    zkErrorSet(err, "frPktGetPriority", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (len < STD_HDR_OFFSET + STD_HDR_SIZE) {
    zkErrorSet_va(err, "frPktGetPriority", ZKERR, ZKMFRPKTFMT, ZKEPARMBAD,
                  "len too small to include standard header (%lu < %d)",
                  (unsigned long)len, STD_HDR_OFFSET + STD_HDR_SIZE);
    return 0;
  }

  INT8_FROM_NBO(p, (char *)data + STD_HDR_PRI_OFFSET);
  *pri = p;

  return 1;
}


DLLExport int frPktSetPriority(ZkError err, void *data, size_t len, unsigned int pri)
{
  uint8_t p;

  zkErrorClear(err);
  
  if ( ! data) {
    zkErrorSet(err, "frPktSetPriority", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (len < STD_HDR_OFFSET + STD_HDR_SIZE) {
    zkErrorSet_va(err, "frPktSetPriority", ZKERR, ZKMFRPKTFMT, ZKEPARMBAD,
                  "len too small to include standard header (%lu < %d)",
                  (unsigned long)len, STD_HDR_OFFSET + STD_HDR_SIZE);
    return 0;
  }

  p = pri;
  NBO_FROM_INT8((char *)data + STD_HDR_PRI_OFFSET, p);

  return 1;
}


/*
 *  This routine link encrypts a buffer.
 *  
 *  It sets the first eight bytes of the buffer to random data, sets the last
 *  eight bytes to a known value (to allow rejection of cover traffic), then
 *  encrypts the buffer.  The length is assumed to be the size of the full
 *  packet, including the trailing 8s and lekv.
 */
DLLExport int frPktLinkEncrypt(ZkError err, FrCryptSym sym, void *data, size_t len)
{
  FrPktType t;
  const char *fn = "frPktLinkEncrypt";
  int ret = 0;
  unsigned char ivec[FRCRYPT_IVEC_SIZE];
  unsigned int v;

  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, ZKMFRPKTFMT, ZKENULLPTR, "sym is null");
    goto done;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRPKTFMT, ZKENULLPTR, "data is null");
    goto done;
  }
  if (len < linkOverhead) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
                  "packet is too short (%lu < %lu)",
                  (unsigned long)len, (unsigned long)linkOverhead);
    goto done;
  }
  if (len % FRCRYPT_SIZE_MULT != FRPKT_LEKV_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
                  "packet has incorrect length ((%lu %% %d) != %d)",
                  (unsigned long)len, FRCRYPT_SIZE_MULT, FRPKT_LEKV_SIZE);
    goto done;
  }

  if ( ! frPktGetVersion(err, data, len, &v)) {
    goto done;
  }

  if ( ! frPktGetType(err, data, len, &t)) {
    goto done;
  }

  /*
   *  Set random data in the first eight bytes of the packet.
   */

  if ( ! zkRandGetData(err, data, PKT_RAND_SIZE)) {
    goto done;
  }

#ifndef NDEBUG
  {
    unsigned char* tmp;
    char buf[300];

    tmp = (unsigned char*)data;
    snprintf(buf, sizeof buf,
             "frPktLinkEncrypt() received data0-7"
             " [%u] [%u] [%u] [%u] [%u] [%u] [%u] [%u] aci [%u] [%u]",
             tmp[0], tmp[1], tmp[2], tmp[3], tmp[4], tmp[5], tmp[6],
             tmp[7], tmp[14], tmp[15]);
    ClientTrace("frPktLinkEncrypt()", buf);
  }
#endif

#if MARIO_HACK
  {
    unsigned int a;
    int i;
    
    for (i = 0; i < 8; i++) {
      a = ((unsigned int) *((unsigned char *) (data) + i));
      fprintf(stderr, "E - MARIO DEBUG: %u\n", a);
    }
    a = ((unsigned int) *((unsigned char *) (data) + 14));
    fprintf(stderr, "E - ACI DEBUG: %u\n", a);
    a = ((unsigned int) *((unsigned char *) (data) + 15));
    fprintf(stderr, "E - ACI DEBUG: %u\n", a);
  }
#endif

  /*
   *  Set a fixed value in the eight bytes after the packet.  Their purpose is
   *  to detect garbage packets.
   */

  {
    size_t off = len - (FRPKT_EIGHTS_SIZE + FRPKT_LEKV_SIZE);
    size_t sz = FRPKT_EIGHTS_SIZE;
    
    memset((char *)data + off, FRPKT_EIGHTS_VALUE, sz);
  }

  /*
   *  Encrypt up to, but not including the last byte.  The last byte should be
   *  the link encryption key version (lekv).
   */

  memset(ivec, 0, sizeof ivec); /* Use a zero ivec. */
  if ( ! frCryptSymEncrypt(err, sym, ivec, data, len-1)) {
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 *  This routine link decrypts a buffer.
 *  
 *  The length is assumed to be the size of the full packet, including the
 *  trailing 8s and lekv.
 */
DLLExport int frPktLinkDecrypt(ZkError err, FrCryptSym sym, void *data, size_t len)
{
  FrPktType t;    /* packet type */
  const char *fn = "frPktLinkDecrypt";
  int ret = 0;
  size_t off, sz;
  unsigned char ivec[FRCRYPT_IVEC_SIZE];
  unsigned int v; /* packet version */

  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, ZKMFRPKTFMT, ZKENULLPTR, "sym is null");
    goto done;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRPKTFMT, ZKENULLPTR, "data is null");
    goto done;
  }
  if (len < linkOverhead) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
                  "packet is too short (%lu < %lu)",
                  (unsigned long)len, (unsigned long)linkOverhead);
    goto done;
  }
  if (len % FRCRYPT_SIZE_MULT != FRPKT_LEKV_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
                  "packet has incorrect length ((%lu %% %d) != %d)",
                  (unsigned long)len, FRCRYPT_SIZE_MULT, FRPKT_LEKV_SIZE);
    goto done;
  }

  /*
   *  Decrypt enough to uncover the standard header.
   */

  off = 0;
  sz = FRCRYPT_SIZE_CEIL(STD_HDR_OFFSET + STD_HDR_SIZE);

  memset(ivec, 0, sizeof ivec); /* Use a zero ivec. */

  if ( ! frCryptSymDecrypt(err, sym, ivec, data, sz)) {
    goto done;
  }

  if ( ! frPktGetVersion(err, data, len, &v)) {
    goto done;
  }

  if ( ! frPktGetType(err, data, len, &t)) {
    goto done;
  }

  /*
   *  Decrypt up to, but not including the last byte.  The last byte should be
   *  the link encryption key version (lekv).
   */

  off += sz;
  sz = len - (off + FRPKT_LEKV_SIZE);

  if ( ! frCryptSymDecrypt(err, sym, ivec, (char *)data + off, sz)) {
    goto done;
  }

  /* Verify the trailing eight bytes. */

  off = len - (FRPKT_EIGHTS_SIZE + FRPKT_LEKV_SIZE);
  sz = FRPKT_EIGHTS_SIZE;

  if ( ! fr_PktMemCheck((char *)data + off, FRPKT_EIGHTS_VALUE, sz)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRPKTFMT, ZKEDATABAD,
               "invalid trailing eights field");
    goto done;
  }

#if MARIO_HACK
  {
    unsigned int a;
    int i;
    
    for (i = 0; i < 8; i++) {
      a = ((unsigned int) *((unsigned char *) (data) + i));
      fprintf(stderr, "D - MARIO DEBUG: %u\n", a);
    }
    a = ((unsigned int) *((unsigned char *) (data) + 14));
    fprintf(stderr, "D - ACI DEBUG: %u\n", a);
    a = ((unsigned int) *((unsigned char *) (data) + 15));
    fprintf(stderr, "D - ACI DEBUG: %u\n", a);
  }
#endif

  ret = 1;

 done:

  return ret;
}




/* 
 *  Like frPktLinkDecrypt, but decrypt on buffer output, so if the sym key is not good, 
 *  we can start again on the same input data without making copies each time.
 */
DLLExport int frPktLinkSmartDecrypt(ZkError err, FrCryptSym sym, void *data, void *output, size_t len)
{
  FrPktType t;    /* packet type */
  size_t decrLen; /* how much to decrypt */
  size_t decrOff; /* offset at which to begin decryption */
  size_t eightOff;
  unsigned char ivec[FRCRYPT_IVEC_SIZE];
  unsigned int v; /* packet version */
  
  zkErrorClear(err);
  
  if ( ! sym) {
    zkErrorSet(err, "frPktLinkDecrypt", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "sym is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "frPktLinkDecrypt", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! output) {
    zkErrorSet(err, "frPktLinkDecrypt", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "output is null");
    return 0;
  }

  /*
   *  Decrypt enough to uncover the standard header.
   */

  decrOff = 0;
  decrLen = FRCRYPT_SIZE_CEIL(STD_HDR_OFFSET + STD_HDR_SIZE);

  if (len < decrLen) {
    zkErrorSet_va(err, "frPktLinkDecrypt", ZKERR, ZKMFRPKTFMT, ZKEBADSTATE,
                  "packet is too short (%lu < %lu)",
                  (unsigned long)len, (unsigned long)decrLen);
    return 0;
  }

  memset(ivec, 0, sizeof ivec); /* Use a zero ivec. */

  if ( ! frCryptSymDecryptToBuf(err, sym, ivec,
				(char *)data + PKT_RAND_OFFSET,
				(char *)output + PKT_RAND_OFFSET, decrLen)) {
    return 0;
  }
  
  if ( ! frPktGetVersion(err, output, len, &v)) {
    return 0;
  }
  
  if ( ! frPktGetType(err, output, len, &t)) {
    return 0;
  }

  /*
   *  Based on the packet type and version we know how much of it to decrypt.
   *  
   *  We have to decrypt the remainder of the packet, if for no other reason
   *  than to verify the trailing eight btyes.
   */

  decrOff += decrLen;

  
  /* check if thepacket version is ok */
  if (v != FRPKT_VERSION_CURRENT) {
    zkErrorSet(err, "frPktLinkDecrypt", ZKERR, ZKMFRPKTFMT, ZKWARN,
               "wrong packet version");
    return 0;
  }
 

  switch (t) {
  case FRPKTTYPE_AIP_ALIVE:
    eightOff = FRPKT_EIGHTS_OFFSET(FRPKT_AIP_ALIVE_SIZE);
    break;

  case FRPKTTYPE_AIP_ROUTE_CRTACK:
    eightOff = FRPKT_EIGHTS_OFFSET(FRPKT_AIP_ROUTE_CRTACK_SIZE);
    break;

  case FRPKTTYPE_AIP_DATA:
    eightOff = FRPKT_EIGHTS_OFFSET(FRPKT_AIP_DATA_SIZE);
    break;
    
  case FRPKTTYPE_AIP_ROUTE_CREATE:
    eightOff = FRPKT_EIGHTS_OFFSET(FRPKT_AIP_ROUTE_CREATE_SIZE);
    break;
    
  case FRPKTTYPE_AIP_ROUTE_DESTROY:
#if DEBUG_TREAT_DESTROY_AS_DATA
    eightOff = FRPKT_EIGHTS_OFFSET(FRPKT_AIP_DATA_SIZE);
    break;
#else
    zkErrorSet(err, "frPktLinkDecrypt", ZKERR, ZKMFRPKTFMT, ZKESTUB,
               "can't yet handle route destruction packets");
    return 0;
#endif

  case FRPKTTYPE_STAT:
#if DEBUG_TREAT_DESTROY_AS_DATA
    eightOff = FRPKT_EIGHTS_OFFSET(FRPKT_AIP_DATA_SIZE);
    break;
#else
    zkErrorSet(err, "frPktLinkDecrypt", ZKERR, ZKMFRPKTFMT, ZKESTUB,
               "can't yet handle route destruction packets");
    return 0;
#endif

  case FRPKTTYPE_NDB_ROUTE_UPDATE:
    eightOff = FRPKT_EIGHTS_OFFSET(FRPKT_AIP_DATA_SIZE);
    break;

  default:
    zkErrorSet_va(err, "frPktLinkDecrypt", ZKERR, ZKMFRPKTFMT, ZKEPARMBAD,
                  "packet has an invalid type (%d)", t);
    return 0;
  }

  decrLen = eightOff + FRPKT_EIGHTS_SIZE;
  
  if (len < decrLen) {
    zkErrorSet_va(err, "frPktLinkDecrypt", ZKERR, ZKMFRPKTFMT, ZKEPARMBAD,
                  "packet is shorter than decryption length (%lu < %lu)",
                  (unsigned long)len, (unsigned long)decrLen);
    return 0;
  }

  decrLen -= decrOff; /* less what we've already decrypted */

  if ( ! frCryptSymDecryptToBuf(err, sym, ivec,
				(char *)data + decrOff, 
				(char *)output + decrOff, decrLen)) {
    return 0;
  }

  /* Verify the trailing eight bytes. */

  if ( ! fr_PktMemCheck((char *)output + eightOff, FRPKT_EIGHTS_VALUE, FRPKT_EIGHTS_SIZE)) {
    zkErrorSet(err, "frPktLinkDecrypt", ZKERR, ZKMFRPKTFMT, ZKEDATABAD,
               "invalid trailing eights field");
    return 0;
  }

#if MARIO_HACK
  {
    unsigned int a;
    int i;
    
    for (i = 0; i < 8; i++) {
      a = ((unsigned int) *((unsigned char *) (output) + i));
      fprintf(stderr, "D - MARIO DEBUG: %u\n", a);
    }
    a = ((unsigned int) *((unsigned char *) (output) + 14));
    fprintf(stderr, "D - ACI DEBUG: %u\n", a);
    a = ((unsigned int) *((unsigned char *) (output) + 15));
    fprintf(stderr, "D - ACI DEBUG: %u\n", a);
  }
#endif

  /*
   *  Add the leading 8 bytes of the packet to the entropy pool.
   *  
   *  bug? Presumably a compromised neighbour could flood us with bogus data
   *  packets, to try to influence our randomness pool.  How serious a threat
   *  is this?
   */

#ifndef WIN32

  // This function is extremely expensive to do per packet on Windows
  (void)zkRandEntropyAdd(err, (char *)output + PKT_RAND_OFFSET, PKT_RAND_SIZE);
#endif

  return 1;
}


/*
 *  Set a pointer to the subpacket (e.g. data, route create) contained in the
 *  buffer.  Essentially, we set the pointer and size to ignore the leading
 *  random bytes, as well as the trailing 8s and link encryption key version
 *  (lekv).
 *  
 *  This routine allows functions, dealing with specific packet types, to
 *  ingore parts that are irrelevant to them.
 */
DLLExport int frPktGetEmbeddedPacket(ZkError err, void *buf, size_t bsz, void **sp, size_t *spsz)
{
  const char *fn = "frPktGetSubPacket";
  int ret = 0;
  
  if ( ! buf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRPKTFMT, ZKENULLPTR, "buf is null");
    goto done;
  }
  if (bsz < linkOverhead) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRPKTFMT, ZKEPARMBAD,
                  "buffer is too small (%lu < %lu)",
                  (unsigned long)bsz, (unsigned long)linkOverhead);
    goto done;
  }
  if ( ! sp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRPKTFMT, ZKENULLPTR, "sp is null");
    goto done;
  }
  if ( ! spsz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRPKTFMT, ZKENULLPTR, "spsz is null");
    goto done;
  }

  *sp = (char *)buf + FRCRYPT_RAND_SIZE;
  *spsz = bsz - linkOverhead;

  ret = 1;

 done:

  return ret;
}


/*
 *  Get the link encryption key version (LEKV) of the packet.
 */
DLLExport int frPktGetLekv(ZkError err, void *data, size_t dsz, uint8_t *lekv)
{
  const char *fn = "frPktGetLekv";
  int ret = 0;

  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRPKTFMT, ZKENULLPTR, "data is null");
    goto done;
  }
  if ( ! lekv) {
    zkErrorSet(err, fn, ZKERR, ZKMFRPKTFMT, ZKENULLPTR, "vers is null");
    goto done;
  }

  *lekv = ((char *)data)[dsz-1];
  
  ret = 1;

 done:

  return ret;
}


/*
 *  Set the link encryption key version (LEKV) of the packet.
 */
DLLExport int frPktSetLekv(ZkError err, void *data, size_t dsz, uint8_t lekv)
{
  const char *fn = "frPktSetLekv";
  int ret = 0;

  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRPKTFMT, ZKENULLPTR, "data is null");
    goto done;
  }

  ((char *)data)[dsz-1] = lekv;

  ret = 1;

 done:

  return ret;
}


/*
 *  Given the size of an embedded packet, return via `sz' the size of the full
 *  packet, including leading random data, trailing 8s and LEKV.
 */
DLLExport size_t frPktGetSize(size_t epsz)
{
  return epsz + linkOverhead;
}
