// The contents of this file are subject to the Mozilla Public
// License Version 1.1 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
// 
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
// 
// The Original Code is Freedom(R) code.
// 
// The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
// Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
// Zero-Knowledge Systems, Inc.  All Rights Reserved.
// 
// Contributor(s):

#undef NDEBUG
#include <assert.h>
#include "zksexception.h"
#include "cRouteCreateAckPkt.h"


cRouteCreateAckPkt::cRouteCreateAckPkt()
{
    ackCtx = NULL;
    pktbuffer = NULL;
    pktbufferlen = 0;
    freePkt = ZK_FALSE;
}

cRouteCreateAckPkt::~cRouteCreateAckPkt()
{
   if (freePkt) {
      destroy();

   }
}

void cRouteCreateAckPkt::create(uint32_t version, uint8_t prio,
                                FrAci exitAci, FrAci nextHopAci, 
                                FrCryptMacKey macKey)
{
    const char    * func = "cRouteCreateAckPkt::create";
    CZKS::Exception e;

    /* Validate params */
    assert((version == FRPKT_DATA_VERSION_1) || (version == FRPKT_DATA_VERSION_2));
    assert(ackCtx == NULL);
    assert(!freePkt);
            
    try
    {
        uint8_t       * pkt = NULL;
        size_t          pktSize = 0;
        size_t          maxsize;
        
        uint16_t        payload = htons( exitAci);

        if (frPktAipDataGetMaxSize(e.zkError(), version, &maxsize) == 
            ZKS_FAILURE) 
        {
            CZKS::Throw(e);
        }

        pktbufferlen = frPktGetSize( maxsize)+1;
        pktbuffer = new uint8_t[pktbufferlen]; 
      
        // Allocate Route Create ACK context
        if (frPktAipDataNew(e.zkError(), &ackCtx) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
        else if (frPktSetVersion(e.zkError(), pktbuffer, pktbufferlen, 
                                 version) == ZKS_FAILURE) 
        {
            CZKS::Throw(e);
        }
        else if (frPktSetPriority(e.zkError(), pktbuffer, pktbufferlen, 
                                  prio) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
	}
        else if (frPktSetType(e.zkError(), pktbuffer, pktbufferlen, 
                              FRPKTTYPE_AIP_ROUTE_CRTACK) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
        else if (frPktGetEmbeddedPacket(e.zkError(), 
                                        pktbuffer, pktbufferlen, 
                                        (void**)&pkt, &pktSize) == ZKS_FAILURE) 
        {
            CZKS::Throw(e);
        }
        else if (frPktAipDataCreateInBuffer(e.zkError(), ackCtx, pkt, 
                                            pktSize) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
        // Set exit hop ACI as payload (for route prolongation)
        else if (frPktAipDataSetPayload(e.zkError(), ackCtx, macKey, 
                                        &payload, sizeof(payload)) == 
                 ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
        else if (!frPktAipDataSetAci(e.zkError(), ackCtx, nextHopAci))
        {
            CZKS::Throw(e);
        }
        freePkt = ZK_TRUE;
    }
    catch (...)
    {
        delete pktbuffer;
        pktbuffer = NULL;

        if (ackCtx != NULL)
        {
            frPktAipDataFree(e.zkError(), &ackCtx);
            ackCtx = NULL;
        }
        
        CZKS::Rethrow(func, "Failed");
    }
}

void cRouteCreateAckPkt::destroy(void)
{
    const char * func = "cRouteCreateAckPkt::destroy";

    assert(ackCtx != NULL);
    assert(freePkt);
    
    try
    {
        CZKS::Exception e;
        
        delete pktbuffer;
        pktbuffer = NULL;
 
        if (frPktAipDataFree(e.zkError(), &ackCtx) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
        freePkt = ZK_FALSE;
        ackCtx = NULL;
    }
    
    catch (CZKS::Exception &e)
    {
        CZKS::Rethrow(func, "Failed");
    }
}

void cRouteCreateAckPkt::attachPkt
(
    uint8_t * pkt,         /* buffer containing RC pkt */
    size_t    pktSize      /* pkt size */
)
{
    const char    * func = "cRouteCreateAckPkt::attachPkt";
    CZKS::Exception e;

    /* Validate params */
    assert(pkt != NULL);
    assert(pktSize > 0);
    assert(ackCtx == NULL);

    try
    {
        uint8_t *epkt;
        size_t epktlen;        

        if (frPktAipDataNew(e.zkError(), &ackCtx) == ZKS_FAILURE) 
        {
            CZKS::Throw(e);
        }
        else if (frPktGetEmbeddedPacket(e.zkError(), 
                                        pkt, pktSize, 
                                        (void**)&epkt, &epktlen) == ZKS_FAILURE) 
        {
            CZKS::Throw(e);
        }
        /* Associate pkt with packet context */
        else if (frPktAipDataReceiveInBuffer(e.zkError(), ackCtx, 
                                             epkt, epktlen) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }

        //save pointer to pkt
        pktbuffer = pkt;
        pktbufferlen = pktSize;

    } /* end of try */

    catch (...)
    {
        if (ackCtx != NULL)
        {
            frPktAipDataFree(e.zkError(), &ackCtx);
        }
        
        CZKS::Rethrow(func, "failed");
    }
}

void cRouteCreateAckPkt::detachPkt(void)
{
    const char * func = "cRouteCreateAckPkt::detachPkt";

    /* Validate params */
    assert(ackCtx != NULL);

    try
    {
        CZKS::Exception e;
        
        if (frPktAipDataFree(e.zkError(), &ackCtx) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
        else
        {
            ackCtx = NULL;
        }
    } /* end of try */

    catch(...)
    {
        CZKS::Rethrow(func, "failed");
    }
}

void cRouteCreateAckPkt::formatBackwardHop( FrAci aci)
{
    const char    * func = "cRouteCreateAckPkt::formatBackwardHop";
    CZKS::Exception e;

    /* Validate params */
    assert(ackCtx != NULL);
            
    try
    {
        if (!frPktAipDataSetAci(e.zkError(), ackCtx, aci))
        {
            CZKS::Throw(e);
        }
    }
    catch (...)
    {
        CZKS::Rethrow(func, "Failed");
    }
}

FrAci cRouteCreateAckPkt::getAci()
{
    const char * func = "cRouteCreateAckPkt::getAci";

    assert(ackCtx != NULL);

    FrAci aci;

    try
    {
        CZKS::Exception e;
        
        if (frPktAipDataGetAci(e.zkError(), ackCtx, &aci) == 
            ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
    } /* end of throw */

    catch (...)
    {
        CZKS::Rethrow(func, "cRouteCreatePkt failed");
    }

    return aci;

} 

FrAci cRouteCreateAckPkt::getExitAci(FrCryptMacKey macKey)
{
    const char * func = "cRouteCreateAckPkt::getExitAci";

    assert(ackCtx != NULL);

    FrAci aci;

    uint16_t temp;

    try
    {
        CZKS::Exception e;
        
        uint8_t* payload;
        size_t psz;
        if (!frPktAipDataGetPayload(e.zkError(), ackCtx, macKey, 
                                    (void**)&payload, &psz))
        {
            CZKS::Throw(e);
        }
        if (psz >= sizeof( temp)) {
           memcpy(&temp, payload, sizeof(temp));
           aci = ntohs(temp);
	}
    } 

    catch (...)
    {
        CZKS::Rethrow(func, "cRouteCreateAckPkt failed");
    }

    return aci;

} 

void cRouteCreateAckPkt::encrypt
(
    FrCryptSym cryptKey /* cryption key (for hop info decryption) */
)
{
    const char    * func = "cRouteCreatePkt::decrypt";
    CZKS::Exception e;

    /* Validate params */
    assert(cryptKey != NULL);
    assert(ackCtx != NULL);

    try
    {
        if (frPktAipDataEncryptPayload(e.zkError(), ackCtx, cryptKey) == 
            ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }

    } /* end of try */

    catch (...)
    {
        CZKS::Rethrow(func, "cRouteCreatePkt failed");
    }
}

void cRouteCreateAckPkt::decrypt
(
    FrCryptSym cryptKey /* cryption key (for hop info decryption) */
)
{
    const char    * func = "cRouteCreatePkt::decrypt";
    CZKS::Exception e;

    /* Validate params */
    assert(cryptKey != NULL);
    assert(ackCtx != NULL);

    try
    {
        if (frPktAipDataDecryptPayload(e.zkError(), ackCtx, cryptKey) == 
            ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }

    } /* end of try */

    catch (...)
    {
        CZKS::Rethrow(func, "cRouteCreatePkt failed");
    }

}

void cRouteCreateAckPkt::getBuffer
(
    uint8_t ** buf,    // out: pointer to buffer
    size_t   * bufSize // out: buffer size
)
{
    const char * func = "cRouteCreateAckPkt::getBuffer";

    size_t pktsize;

    // Validate params
    assert(buf != NULL);
    assert(bufSize != NULL);

    // A packet MUST currently be attached
    assert(ackCtx != NULL);

    try
    {
        CZKS::Exception e;
        
        /* Get size of data packet. */
        if (frPktAipDataGetCurrentSize(e.zkError(), ackCtx, 
                                       &pktsize) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
        *bufSize = frPktGetSize(pktsize);
        *buf = (uint8_t*)pktbuffer;
 
        // frPktAipData library allocates an extra byte on creation,
        // don't report it
        //*bufSize -= 1; 
    }
    catch (CZKS::Exception &e)
    {
        CZKS::Rethrow(func, "Failed");
    }
}
