/*
 *  $Id: buf.c,v 1.12.10.4 2000/12/14 17:53:13 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Simple decryption of data read from a file and encryption of data written
 *  to a memory buffer.
 *  
 *  The buffer will have a small header prepended to it, and the portion
 *  containing the data will be a multiple of 8 bytes in length.
 *  
 *  bug: add information to the header, such as version, so we can change the
 *  format of the buffer (e.g. the header).
 *  
 *  bug: add a hash over the data; put it in the header.
 */

#include "zks.h"

#ifndef WIN32
#include <arpa/inet.h>
#else
#include <winsock2.h>
#endif

#include <errno.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include "fr_buf.h"
#include "frCrypt.h"
#include "frPktFmt.h"
#include "hdr.h"
#include "zkAlloc.h"
#include "zkError.h"

/*** t.b.d: take out define ***/
/* #define FRPKTTYPE_CLRTXT_DATA 7 */


static int checkMac(ZkError err, FrCryptSym sym, void *buf, size_t bsz, int *valid)
{
  FrCryptMacKey mackey = 0;
  int ret = 0;
  size_t msz;
  struct buf_hdr_v1 *hdr = buf;

  msz = bsz - offsetof(struct buf_hdr_v1, dsz);
  
  if ( ! (frCryptMacKeyNew(err, &mackey) &&
          frCryptMacSetKeyFromSym(err, mackey, sym) &&
          frCryptMacVerifyFromBuffer(err, mackey, &hdr->dsz, msz,
                                     hdr->mac, sizeof hdr->mac, valid))) {
    goto done;
  }

  ret = 1;

 done:

  if (mackey) { frCryptMacKeyFree(0, &mackey); }

  return ret;
}

static int checkZeroKeyMac(ZkError err, void *buf, size_t bsz, int *valid)
{
  FrCryptMacKey mackey = 0;
  int ret = 0;
  size_t msz;
  struct buf_hdr_v1 *hdr = buf;

  msz = bsz - offsetof(struct buf_hdr_v1, dsz);
  
  if ( ! (frCryptMacKeyNew(err, &mackey) &&
          frCryptMacSetKeyZero(err, mackey) &&
          frCryptMacVerifyFromBuffer(err, mackey, &hdr->dsz, msz,
                                     hdr->mac, sizeof hdr->mac, valid))) {
    goto done;
  }

  ret = 1;

 done:

  if (mackey) { frCryptMacKeyFree(0, &mackey); }

  return ret;
}


static int genMac(ZkError err, FrCryptSym sym, void *buf, size_t bsz)
{
  FrCryptMacKey mackey = 0;
  int ret = 0;
  size_t msz;
  struct buf_hdr_v1 *hdr = buf;

  msz = bsz - offsetof(struct buf_hdr_v1, dsz);
  
  if ( ! (frCryptMacKeyNew(err, &mackey) &&
          frCryptMacSetKeyFromSym(err, mackey, sym) &&
          frCryptMacIntoBuffer(err, mackey, &hdr->dsz, msz,
                               hdr->mac, sizeof hdr->mac))) {
    goto done;
  }

  ret = 1;

 done:

  if (mackey) { frCryptMacKeyFree(0, &mackey); }

  return ret;
}


/* Generates a MAC with a 0 key. */
static int genZeroKeyMac(ZkError err, void *buf, size_t bsz)
{
  FrCryptMacKey mackey = 0;
  int ret = 0;
  size_t msz;
  struct buf_hdr_v1 *hdr = buf;

  msz = bsz - offsetof(struct buf_hdr_v1, dsz);
  
  if ( ! (frCryptMacKeyNew(err, &mackey) &&
          frCryptMacSetKeyZero(err, mackey) &&
          frCryptMacIntoBuffer(err, mackey, &hdr->dsz, msz,
                               hdr->mac, sizeof hdr->mac))) {
    goto done;
  }

  ret = 1;

 done:

  if (mackey) { frCryptMacKeyFree(0, &mackey); }

  return ret;
}


/*
 *  Decrypt the contents of `src', allocating memory for the result, then
 *  setting `*dst' to that buffer.  The size of the buffer will be returned
 *  via `*dsz'.
 */
static int decryptBuf_v1(ZkError err, void *src, size_t ssz, FrCryptSym sym, void **dst, size_t *dsz)
{
  const char *fn = "decryptBuf_v1";
  int ret = 0;
  size_t psz; /* plaintext size */
  struct buf_hdr_v1 *hdr = src;
  void *buf = 0;

  zkErrorClear(err);

  if ( ! src) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "src is null");
    return 0;
  }

  /* We suppose that this function will always be called with a non null 'sym'
   * parameter, even when decrypting a cleartext file.
   */
  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "sym is null");
    return 0;
  }
  if ( ! dst) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "dst is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "dsz is null");
    return 0;
  }

  if (ssz < sizeof *hdr) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "buffer too short (%lu < %lu)",
                  (unsigned long)ssz, (unsigned long)sizeof *hdr);
    goto done;
  }

  if ((ssz - sizeof *hdr) % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "buffer size (%lu) not a multiple of %lu",
                  (unsigned long)(ssz - sizeof *hdr),
                  (unsigned long)FRCRYPT_SIZE_MULT);
    goto done;
  }
  
  if (hdr->vers != 1) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "unexpected version %u", hdr->vers);
    goto done;
  }

  if ( ! (hdr->type == FRPKTTYPE_SYMENCR_DATA || hdr->type ==  FRPKTTYPE_CLRTXT_DATA)) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "unexpected type: %u", hdr->type);
    goto done;
  }

  /*
   *  Copy the encrypted buffer to a malloc()ed chunk of memory.  Decrypt it
   *  there, check any padding, verify the MAC, then later resize the memory
   *  to hold just the actual data.
   */
  
  /* bug: locked memory alloc. */
  if ( ! (buf = malloc(ssz))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, errno,
                  "can't malloc %lu bytes for data: %s",
                  (unsigned long)ssz, strerror(errno));
    goto done;
  }

  memcpy(buf, src, ssz);
  hdr = buf; /* hdr now points to copy of the data */

  /* If the file was encrypted, Decrypt it and check the MAC */
  if (hdr->type == FRPKTTYPE_SYMENCR_DATA) {

    /*Decrypt */
    if ( ! frCryptSymDecrypt(err, sym, hdr->iv, &hdr->mac,
                             ssz - offsetof(struct buf_hdr_v1, mac))) {
      goto done;
    }

    /* Check the MAC. */
    {
      int valid;
      
      if ( ! checkMac(err, sym, buf, ssz, &valid)) {
        goto done;
      }
      if ( ! valid) {
        zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD, "invalid MAC");
        goto done;
      }
    }
  }

  /* If the file was _not_ encrypted, just check the Zero Key MAC */
  else if(hdr->type == FRPKTTYPE_CLRTXT_DATA) {
     /* Check the Zero Key MAC */
     int valid;

     if ( ! checkZeroKeyMac(err, buf, ssz, &valid)) {  
       zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD, "invalid MAC");
       goto done;
     }
  }
    

  psz = ntohl(hdr->dsz); /* amount of `actual' data in buffer */

  /* Check size consistency. */
  
  if (psz > ssz - sizeof *hdr) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "inconsistent header (%lu > %lu)",
                  (unsigned long)psz, (unsigned long)(ssz - sizeof *hdr));
    goto done;
  }

  /* Check any trailing padding. */
  
  if (psz < ssz - sizeof *hdr) {
    size_t s = ssz - sizeof *hdr - psz;
    if ( ! fr_CryptCheckPad((char *)buf + sizeof *hdr + psz, s)) {
      zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD,
                 "invalid trailing padding");
      goto done;
    }
  }

  /* Reallocate space for the decrypted data. */
  
  memset(buf, 0, sizeof *hdr);
  memmove(buf, (char *)buf + sizeof *hdr, psz);
  if ( ! zkMemRealloc(err, &buf, psz)) {
    goto done;
  }
  
  *dst = buf; buf = 0; /* the caller now owns the buffer */
  *dsz = psz;
  
  ret = 1;

 done:

  if (buf) { memset(buf, 0, ssz); free(buf); }

  return ret;
}


/*
 *  Encrypt the contents of `src', allocating memory for the result, then
 *  setting `*dst' to that buffer.  The size of the buffer will be returned
 *  via `*dsz'.
 *  
 *  Extra flag `encrypt': if 0 we don't encrypt (and we use a Zero Key MAC)
 *  else we do encrypt and MAC with `sym' key.
 */
static int encryptBuf_v1(ZkError err, void *src, size_t ssz, FrCryptSym sym, void **dst, size_t *dsz, int encrypt)
{
  char iv[FRCRYPT_IVEC_SIZE];
  const char *fn = "encryptBuf_v1";
  int ret = 0;
  size_t bsz; /* size of malloc()ed buffer */
  struct buf_hdr_v1 *hdr;
  void *buf;

  zkErrorClear(err);

  if ( ! src) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "src is null");
    return 0;
  }
  if (encrypt && ! sym) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "sym is null");
    return 0;
  }
  if ( ! dst) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "dst is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "dsz is null");
    return 0;
  }

  bsz = FRCRYPT_SIZE_CEIL(ssz + sizeof *hdr);
  /* bug: should allocate locked memory */
  if ( ! (buf = malloc(bsz))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, errno,
                  "can't malloc %lu bytes for data: %s",
                  (unsigned long)bsz, strerror(errno));
    goto done;
  }

  /* Fill in the header. */

  hdr = (void *)buf;

  hdr->vers = 1;

  if (encrypt) {
    hdr->type = FRPKTTYPE_SYMENCR_DATA; 
  }
  else {
    hdr->type = FRPKTTYPE_CLRTXT_DATA;
  }

  fr_CryptSetPad(hdr->spec, sizeof hdr->spec); /* bug: check on decrypt */
  
  if ( ! zkRandGetData(err, hdr->iv, sizeof hdr->iv)) {
    goto done;
  }
  memcpy(iv, hdr->iv, sizeof iv);
  hdr->dsz = htonl(ssz);

  /* Copy the plaintext. */

  memcpy((char *)buf + sizeof *hdr, src, ssz);
  
  /* Set any trailing padding to the appropriate value. */

  if (bsz > ssz + sizeof *hdr) {
    size_t s = sizeof *hdr + ssz;
    fr_CryptSetPad((char *)buf + s, bsz - s);
  }

  if (encrypt) {

    /* Set the MAC. */

    if ( ! genMac(err, sym, hdr, bsz)) {
      goto done;
    }

    /*
     *  Encrypt the entire buffer following the IV.  The IV in the header
     *  retains its original value; our copy in `iv' is changed.
     */

    if ( ! frCryptSymEncrypt(err, sym, iv, &hdr->mac,
                          bsz - (offsetof(struct buf_hdr_v1, mac)))) {
      goto done;
    }
  }
  else {
   
    /* set MAC with key = 0 */

    if ( ! genZeroKeyMac(err, hdr, bsz)) {
      goto done;
    }

  }


  *dst = buf; buf = 0; /* the caller now owns the buffer */
  *dsz = bsz;

  ret = 1;

 done:

  if (buf) { memset(buf, 0, bsz); free(buf); }

  return ret;
}


DLLExport int frSecDecryptFromBuf(ZkError err, void *src, size_t ssz, FrCryptSym sym, void **dst, size_t *dsz)
{
  return decryptBuf_v1(err, src, ssz, sym, dst, dsz);
}


/*
 *  bug: should be named frSecEncryptToMem(), since it allocates memory.
 */
DLLExport int frSecEncryptToBuf(ZkError err, void *src, size_t ssz, FrCryptSym sym, void **dst, size_t *dsz)
{
  return encryptBuf_v1(err, src, ssz, sym, dst, dsz, 1);
}


DLLExport int frSecCleartextToBuf(ZkError err, void *src, size_t ssz, FrCryptSym sym, void **dst, size_t *dsz)
{
  return encryptBuf_v1(err, src, ssz, sym, dst, dsz, 0);
}
