/*
 *  $Id: buf_v0.c,v 1.3.16.3 2000/12/14 17:53:13 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  WARNING: this is the first, deprecated version of the encrypted buffer
 *  code.  It is being kept only while conversion to the new format is
 *  necessary.
 *  
 *  Simple decryption of data read from a file and encryption of data written
 *  to a memory buffer.
 *  
 *  The buffer will have a small header prepended to it, and the portion
 *  containing the data will be a multiple of 8 bytes in length.
 *  
 *  bug: add information to the header, such as version, so we can change the
 *  format of the buffer (e.g. the header).
 *  
 *  bug: add a hash over the data; put it in the header.
 */


#include "zks.h"

#ifndef WIN32
#include <arpa/inet.h>
#else
#include <winsock2.h>
#endif

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include "fr_buf_v0.h"
#include "frCrypt.h"
#include "hdr_v0.h"
#include "zkAlloc.h"
#include "zkError.h"


/*
 *  Decrypt the contents of `src', allocating memory for the result, then
 *  setting `*dst' to that buffer.  The size of the buffer will be returned
 *  via `*dsz'.
 */
DLLExport int fr_SecDecryptFromBuf_v0(ZkError err, void *src, size_t ssz, FrCryptSym sym, void **dst, size_t *dsz)
{
  int ret = 0;
  size_t psz; /* plaintext size */
  struct buf_hdr_v0 *hdr;
  void *buf;

  zkErrorClear(err);

  if ( ! src) {
    zkErrorSet(err, "fr_SecDecryptFromBuf_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "src is null");
    return 0;
  }
  if ( ! sym) {
    zkErrorSet(err, "fr_SecDecryptFromBuf_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "sym is null");
    return 0;
  }
  if ( ! dst) {
    zkErrorSet(err, "fr_SecDecryptFromBuf_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "dst is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, "fr_SecDecryptFromBuf_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "dsz is null");
    return 0;
  }

  if (ssz < sizeof *hdr) {
    /* Invalid buffer size */
    zkErrorSet_va(err, "fr_SecDecryptFromBuf_v0", ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "buffer too short (%lu < %lu)",
                  (unsigned long)ssz, (unsigned long)sizeof *hdr);
    goto done;
  }

  if ((ssz - sizeof *hdr) % FRCRYPT_SIZE_MULT != 0) {
    /* Invalid buffer size. */
    zkErrorSet_va(err, "fr_SecDecryptFromBuf_v0", ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "buffer size (%lu) not a multiple of %lu",
                  (unsigned long)(ssz - sizeof *hdr),
                  (unsigned long)FRCRYPT_SIZE_MULT);
    goto done;
  }
  
  /*
   *  Copy the encrypted buffer to a malloc()ed chunk of memory.  Decrypt it
   *  there, then later resize the memory to hold just the actual data.
   */
  
  /* bug: locked memory alloc. */
  if ( ! (buf = malloc(ssz))) {
    zkErrorSet_va(err, "fr_SecDecryptFromBuf_v0", ZKERR, ZKMFRSEC, errno,
                  "can't malloc %lu bytes for data: %s",
                  (unsigned long)ssz, strerror(errno));
    goto done;
  }

  memcpy(buf, src, ssz);
  hdr = buf;

  /* Decrypt the buffer following the IV (which wasn't encrypted). */

  if ( ! frCryptSymDecrypt(err, sym, hdr->iv, &hdr->dsz, ssz - sizeof hdr->iv)) {
    goto done;
  }

  psz = ntohl(hdr->dsz); /* amount of `actual' data in buffer */

  if (psz > ssz - sizeof *hdr) {
    /* Inconsistent header. */
    zkErrorSet_va(err, "fr_SecDecryptFromBuf_v0", ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "inconsistent header (%lu > %lu)",
                  (unsigned long)psz, (unsigned long)(ssz - sizeof *hdr));
    goto done;
  }

  if ( ! fr_CryptCheckPad(hdr->pad, sizeof hdr->pad)) {
    zkErrorSet(err, "fr_SecDecryptFromBuf_v0", ZKERR, ZKMFRSEC, ZKEDATABAD,
               "invalid padding in header");
    goto done;
  }

  if (psz < ssz - sizeof *hdr) {
    size_t s = ssz - sizeof *hdr - psz;
    if ( ! fr_CryptCheckPad((char *)buf + sizeof *hdr + psz, s)) {
      zkErrorSet(err, "fr_SecDecryptFromBuf_v0", ZKERR, ZKMFRSEC, ZKEDATABAD,
                 "invalid trailing padding");
      goto done;
    }
  }

  /* Reallocate space for the decrypted data. */
  
  memset(buf, 0, sizeof *hdr);
  memmove(buf, (char *)buf + sizeof *hdr, psz);
  if ( ! zkMemRealloc(err, &buf, psz)) {
    goto done;
  }
  
  *dst = buf; buf = 0; /* the caller now owns the buffer */
  *dsz = psz;
  
  ret = 1;

 done:

  if (buf) { memset(buf, 0, ssz); free(buf); }

  return ret;
}


/*
 *  Encrypt the contents of `src', allocating memory for the result, then
 *  setting `*dst' to that buffer.  The size of the buffer will be returned
 *  via `*dsz'.
 */
DLLExport int fr_SecEncryptToBuf_v0(ZkError err, void *src, size_t ssz, FrCryptSym sym, void **dst, size_t *dsz)
{
  char iv[FRCRYPT_IVEC_SIZE];
  int ret = 0;
  size_t bsz; /* size of malloc()ed buffer */
  struct buf_hdr_v0 *hdr;
  void *buf;

  zkErrorClear(err);

  if ( ! src) {
    zkErrorSet(err, "fr_SecEncryptToBuf_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "src is null");
    return 0;
  }
  if ( ! sym) {
    zkErrorSet(err, "fr_SecEncryptToBuf_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "sym is null");
    return 0;
  }
  if ( ! dst) {
    zkErrorSet(err, "fr_SecEncryptToBuf_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "dst is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, "fr_SecEncryptToBuf_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "dsz is null");
    return 0;
  }

  bsz = FRCRYPT_SIZE_CEIL(ssz + sizeof *hdr);
  /* bug: should allocate locked memory */
  if ( ! (buf = malloc(bsz))) {
    zkErrorSet_va(err, "fr_SecEncryptToBuf_v0", ZKERR, ZKMFRSEC, errno,
                  "can't malloc %lu bytes for data: %s",
                  (unsigned long)bsz, strerror(errno));
    goto done;
  }

  /* Fill in the header. */

  hdr = (void *)buf;
  
  if ( ! zkRandGetData(err, hdr->iv, sizeof hdr->iv)) {
    goto done;
  }
  memcpy(iv, hdr->iv, sizeof iv);
  hdr->dsz = htonl(ssz);
  fr_CryptSetPad(hdr->pad, sizeof hdr->pad);

  /* Copy the plaintext. */

  memcpy((char *)buf + sizeof *hdr, src, ssz);
  
  /* Set any trailing padding to the appropriate value. */

  if (bsz > ssz + sizeof *hdr) {
    size_t s = sizeof *hdr + ssz;
    fr_CryptSetPad((char *)buf + s, bsz - s);
  }

  /*
   *  Encrypt the entire buffer following the IV.  The IV in the header
   *  retains its original value; our copy in `iv' is changed.
   */

  if ( ! frCryptSymEncrypt(err, sym, iv, &hdr->dsz, bsz - sizeof hdr->iv)) {
    goto done;
  }

  *dst = buf; buf = 0; /* the caller now owns the buffer */
  *dsz = bsz;

  ret = 1;

 done:

  if (buf) { memset(buf, 0, bsz); free(buf); }

  return ret;
}
