/*
 *  $Id: file_v0.c,v 1.2.20.3 2000/12/14 17:53:13 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  WARNING: this is the first, deprecated version of the encrypted buffer
 *  code.  It is being kept only while conversion to the new format is
 *  necessary.
 *  
 *  Simple decryption of data read from a file and encryption of data written
 *  to a file.
 *  
 *  The file will have a small header prepended to it, and the portion
 *  containing the data will be a multiple of 8 bytes in length.
 *  
 *  bug: add information to the header, such as version, so we can change the
 *  format of the file (e.g. the header).
 *  
 *  bug: add a hash over the data; put it in the header.
 */

#define _BSD_SOURCE 1 /* S_ISLNK(), fstat() */
#if 0
#define _POSIX_SOURCE 1 /* fileno */
#endif

#include "zks.h"

#include <sys/stat.h>
#ifndef WIN32
#include <arpa/inet.h>
#include <unistd.h>
#else
#include <winsock2.h>
#endif

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include "frCrypt.h"
#include "fr_file_v0.h"
#include "hdr_v0.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkIntTypes.h"
#include "zkLog.h"



static int secMemCheck(void *d, size_t dsz)
{
  unsigned char *p = d;
  size_t i;

  for (i = 0; i < dsz; i++) {
    if (p[i] != dsz) {
      return 0;
    }
  }

  return 1;
}


static void secMemFill(void *d, size_t dsz)
{
  memset(d, dsz, dsz);
}


/*
 *  bug: this is very similar to dirIsSecure() in frsec's keystore.c.  [Make
 *  common routine taking a stat structure?]
 */
static int fileIsSecure(ZkError err, FILE *fp)
{
  struct stat sbuf;
  uid_t owner = geteuid();

  zkErrorClear(err);
  
#ifdef WIN32
#pragma message(Reminder "some of this may be problematic for Windows")
#endif
  
  if (fstat(fileno(fp), &sbuf) == -1) {
    zkErrorSet_va(err, "fileIsSecure", ZKERR, ZKMFRSEC, errno,
                  "can't fstat file: %s", strerror(errno));
    return 0;
  }

#if 0
  /*
   *  bug: this test is useless; one the file is open any information about
   *  how we got to it is gone.
   */

  /* Ensure it's not a symbolic link. */

  if (S_ISLNK(sbuf.st_mode)) {
    zkErrorSet(err, "fileIsSecure", ZKERR, ZKMFRSEC, ZKEMLINK,
               "file is a symbolic link");
    return 0;
  }
#endif

  /* It must be owned by us. Unless you are root, than you have access to 
   * the file anyway.
   */
  
  if( owner != 0 ) {
    if (sbuf.st_uid != owner) {
      zkErrorSet_va(err, "fileIsSecure", ZKERR, ZKMFRSEC, ZKEACCES,
                    "the owner (%lu) of the file is not me (%lu)",
                    (unsigned long)sbuf.st_uid, (unsigned long)owner);
      return 0;
    }
  }

  /* No one else should be able to read or write to the file. */

  if (sbuf.st_mode & (S_IRWXG | S_IRWXO)) {
    zkErrorSet_va(err, "fileIsSecure", ZKERR, ZKMFRSEC, ZKEACCES,
                  "*only* the owner (%lu) should be able to access the file",
                  (unsigned long)owner);
    return 0;
  }

  return 1;
}


DLLExport int fr_SecDecryptFromFp_v0(ZkError err, FILE *fp, FrCryptSym key, void **data, size_t *dataSz)
{
  char *d = 0;
  char iv[FRCRYPT_IVEC_SIZE];
  int ret = 0;
  size_t dsz = 0; /* # bytes of encrypted data, including padding, but not the header */
  size_t fsz; /* size of the file */
  struct buf_hdr_v0 fh;

  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, "fr_SecDecryptFromFp_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "file is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, "fr_SecDecryptFromFp_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "fr_SecDecryptFromFp_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! dataSz) {
    zkErrorSet(err, "fr_SecDecryptFromFp_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "dataSz is null");
    return 0;
  }

  if ( ! (fileIsSecure(err, fp) &&
          zkFileLengthFp(err, fp, &fsz))) {
    goto done;
  }

  if (fsz < sizeof fh) {
    /* Invalid file size */
    zkErrorSet_va(err, "fr_SecDecryptFromFp_v0", ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "file too short (%lu < %lu)",
                  (unsigned long)fsz, (unsigned long)sizeof fh);
    goto done;
  }

  dsz = fsz - sizeof fh;

  if (dsz % FRCRYPT_SIZE_MULT != 0) {
    /* Invalid file size. */
    zkErrorSet_va(err, "fr_SecDecryptFromFp_v0", ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "data size (%lu) not a multiple of %lu",
                  (unsigned long)dsz, (unsigned long)FRCRYPT_SIZE_MULT);
    goto done;
  }
  
  /* Assume the file won't change size on us... */

  if (fread(&fh, 1, sizeof fh, fp) != sizeof fh) {
    zkErrorSet_va(err, "fr_SecDecryptFromFp_v0", ZKERR, ZKMFRSEC, errno,
                  "error reading header (%lu bytes) from file: %s",
                  (unsigned long)sizeof fh, strerror(errno));
    goto done;
  }

  /* Decrypt that part of the header following the IV. */

  memcpy(iv, fh.iv, sizeof iv);
  if ( ! frCryptSymDecrypt(err, key, iv, &fh.dsz, sizeof fh - sizeof fh.iv)) {
    goto done;
  }

  if ( ! secMemCheck(fh.pad, sizeof fh.pad)) {
    zkErrorSet(err, "fr_SecDecryptFromFp_v0", ZKERR, ZKMFRSEC, ZKEDATABAD,
               "invalid padding in file");
    goto done;
  }

  fh.dsz = ntohl(fh.dsz);

  if (fh.dsz > dsz) {
    /* Inconsistent header. */
    zkErrorSet_va(err, "fr_SecDecryptFromFp_v0", ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "inconsistent header (%lu > %lu)",
                  (unsigned long)fh.dsz, (unsigned long)dsz);
    goto done;
  }

  if ( ! (d = malloc(dsz))) {
    zkErrorSet_va(err, "fr_SecDecryptFromFp_v0", ZKERR, ZKMFRSEC, errno,
                  "can't malloc %lu bytes for data: %s",
                  (unsigned long)dsz, strerror(errno));
    goto done;
  }

  if (fread(d, 1, dsz, fp) != dsz) {
    zkErrorSet_va(err, "fr_SecDecryptFromFp_v0", ZKERR, ZKMFRSEC, errno,
                  "can't read %lu bytes of data from file: %s",
                  (unsigned long)dsz, strerror(errno));
    goto done;
  }

  if ( ! frCryptSymDecrypt(err, key, iv, d, dsz)) {
    goto done;
  }

  if (fh.dsz < dsz && ! secMemCheck(d + fh.dsz, dsz - fh.dsz)) {
    zkErrorSet(err, "fr_SecDecryptFromFp_v0", ZKERR, ZKMFRSEC, errno,
               "invalid trailing padding in file");
    goto done;
  }

  *data = d; d = 0;
  *dataSz = fh.dsz;
  
  ret = 1;

 done:

  if (d) {
    memset(d, 0, dsz);
    free(d);
  }

  return ret;
}


DLLExport int fr_SecDecryptFromFile_v0(ZkError err, const char *file, FrCryptSym key, void **data, size_t *dataSz)
{
  FILE *fp = 0;
  ZkLog log;
  int ret = 0;

  zkErrorClear(err);
  zkErrorLog(err, &log);

  if ( ! file) {
    zkErrorSet(err, "fr_SecDecryptFromFile_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "file is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, "fr_SecDecryptFromFile_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "fr_SecDecryptFromFile_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! dataSz) {
    zkErrorSet(err, "fr_SecDecryptFromFile_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "dataSz is null");
    return 0;
  }

  if ( ! (fp = fopen(file, "rb"))) {
    zkErrorSet_va(err, "fr_SecDecryptFromFile_v0", ZKERR, ZKMFRSEC, errno,
                  "can't open `%s' for reading: %s", file, strerror(errno));
    goto done;
  }

  if ( ! (ret = fr_SecDecryptFromFp_v0(err, fp, key, data, dataSz))) {
    zkLogWrite_va(log, "can't decrypt from `%s'", file);
  }

 done:

  if (fp) { fclose(fp); }

  return ret;
}


DLLExport int fr_SecEncryptToFp_v0(ZkError err, FILE *fp, FrCryptSym key, void *data, size_t dataSz)
{
  char *d = 0;
  char iv[FRCRYPT_IVEC_SIZE];
  int ret = 0;
  size_t dsz;
  struct buf_hdr_v0 fh;

  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, "fr_SecEncryptToFp_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "fp is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, "fr_SecEncryptToFp_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "fr_SecEncryptToFp_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (dataSz == 0) {
    zkErrorSet(err, "fr_SecEncryptToFp_v0", ZKERR, ZKMFRSEC, ZKEPARMBAD,
               "dataSz is 0");
    return 0;
  }

  dsz = FRCRYPT_SIZE_CEIL(dataSz); /* amount to allocate and encrypt */

  /* Set up the file header. */

  if ( ! zkRandGetData(err, fh.iv, sizeof fh.iv)) {
    goto done;
  }
  memcpy(iv, fh.iv, sizeof iv);
  fh.dsz = htonl(dataSz);
  secMemFill(fh.pad, sizeof fh.pad);

  if ( ! (d = malloc(dsz))) {
    zkErrorSet_va(err, "fr_SecEncryptToFp_v0", ZKERR, ZKMFRSEC, errno,
                  "can't malloc %lu bytes for data: %s",
                  (unsigned long)dsz, strerror(errno));
    goto done;
  }

  /* Encrypt that part of the header following the IV. */

  if ( ! frCryptSymEncrypt(err, key, iv,
                           &fh.dsz, sizeof fh - sizeof fh.iv)) {
    goto done;
  }

  memcpy(d, data, dataSz);
  if (dataSz < dsz) {
    secMemFill(d + dataSz, dsz - dataSz);
  }

  /* Continue encrypting the actual data to be stored. */

  if ( ! frCryptSymEncrypt(err, key, iv, d, dsz)) {
    goto done;
  }

  if ( ! (fwrite(&fh, 1, sizeof fh, fp) == sizeof fh &&
          fwrite(d, 1, dsz, fp) == dsz)) {
    zkErrorSet_va(err, "fr_SecEncryptToFp_v0", ZKERR, ZKMFRSEC, errno,
                  "error writing to file: %s", strerror(errno));
    goto done;
  }

  ret = 1;

 done:

  if (d) {
    memset(d, 0, dsz);
    free(d);
  }

  return ret;
}


/*
 *  Store an encrypted copy of the contents of `data' in `file'.  Any existing
 *  contents of `file' are overwritten.
 */
DLLExport int fr_SecEncryptToFile_v0(ZkError err, const char *file, FrCryptSym key, void *data, size_t dataSz)
{
  FILE *fp = 0;
  ZkLog log;
  int ret = 0, omask;

  zkErrorClear(err);
  zkErrorLog(err, &log);

  if ( ! file) {
    zkErrorSet(err, "fr_SecEncryptToFile_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "file is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, "fr_SecEncryptToFile_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "fr_SecEncryptToFile_v0", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (dataSz == 0) {
    zkErrorSet(err, "fr_SecEncryptToFile_v0", ZKERR, ZKMFRSEC, ZKEPARMBAD,
               "dataSz is 0");
    return 0;
  }

  omask = umask(0077);

  /* Open the file, then write everything. */

  if ( ! (fp = fopen(file, "wb"))) {
    zkErrorSet_va(err, "fr_SecEncryptToFile_v0", ZKERR, ZKMFRSEC, errno,
                  "can't open `%s' for writing: %s", file, strerror(errno));
    goto done;
  }

  /* bug: set permissions, check links, etc. */
  fprintf(stderr, "**** set permissions, check links, etc. ****\n");

  if ( ! (ret = fr_SecEncryptToFp_v0(err, fp, key, data, dataSz))) {
    zkLogWrite_va(log, "can't decrypt from `%s'", file);
  }

  ret = 1;

 done:

  umask(omask);
  if (fp && fclose(fp) == EOF) {
    zkLogWrite_va(log, "error closing `%s': %s", file, strerror(errno));
    ret = 0;
  }

  return ret;
}
