/*
 *  $Id: frsec_passinfo.c,v 1.13.20.4 2000/12/14 17:53:13 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Routines used for the verification of a passphrase with salt usage. The
 *  algorithm used is as follows...
 */

#include <zks.h>

#ifndef WIN32
#include <unistd.h>
#include <termios.h>
#else
#include <winsock2.h>
#endif

#include <string.h>
#include <errno.h>

#include "zkFile.h"
#include "frSec.h"
#include "frPktFmt.h"

struct fr_sec_passinfo {
  char salt_key[FRCRYPT_SALT_SIZE];
  char salt_ver[FRCRYPT_SALT_SIZE];
  char hash_ver[FRCRYPT_HASH_SIZE];
  uint32_t logIters;
};

#pragma pack(1)
/* This is for external storage. */
struct fr_sec_passinfo_pk {
  uint8_t vers;
  uint8_t type;
  char salt_key[FRCRYPT_SALT_SIZE];
  char salt_ver[FRCRYPT_SALT_SIZE];
  char hash_ver[FRCRYPT_HASH_SIZE];
  uint32_t logIters; /* in network byte order (big-endian) */
};
#pragma pack()


/*
 * Reads from fp a passphrase. If a prompt is passed, it is displayed.
 * It returns the password in *pass and its length in *plen
 */

static int readPassPhrase(ZkError err, FILE *fp, const char *prompt, char **pass, int *plen )
{
  int ret = 0;
  
  if ( prompt ) {
    fprintf(fp, prompt);
  }
  
  fflush(fp);
  
  ret = zkFileGetLineFp(err, fp, pass, plen );
  
  fflush(fp);
  
  if ( prompt ) {
    fprintf(fp, "\n");
    fflush(fp);
  }
  
  return ret;
}

/*
 * Creates a new FrSecPassInfo object. It fills in the salt values with 
 * random data.
 */

DLLExport int frSecPassInfoNew(ZkError err, FrSecPassInfo *pinfo)
{
  int ret = 0;
  const char *func = "frSecPassInfoNew";
  
  zkErrorClear(err);
  if ( !pinfo ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"pfino is null" ); 
    goto done;
  }
   

  if ( (*pinfo = malloc(sizeof **pinfo)) == NULL ) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                  "can't allocate %lu bytes for structure pinfo: %s",
                  (unsigned long)(sizeof **pinfo), strerror(errno));
    goto done;
  }
 
  memset((*pinfo), 0, sizeof **pinfo);
  if ( ! zkRandGetData(err, (*pinfo)->salt_key, FRCRYPT_SALT_SIZE) ||
       ! zkRandGetData(err, (*pinfo)->salt_ver, FRCRYPT_SALT_SIZE) ) {
    goto done;
  }
 
  ret = 1;
 
 done:
  return ret;
}

/*
 * Frees the memory occupied by the FrSecPassInfo object pinfo. *pinfo is set to NULL.
 *
 */

DLLExport int frSecPassInfoFree(ZkError err, FrSecPassInfo *pinfo)
{
  const char *fn = "frSecPassInfoFree"; 
  int ret = 0;

  zkErrorClear(err);
  if ( !pinfo ) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR,"pinfo is null" );    
    goto done;
  }

  if ( !*pinfo) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKEPARMBAD,"*pinfo is null" );    
    goto done;
  }

  memset((*pinfo), 0, sizeof **pinfo);

  free(*pinfo);
  *pinfo = NULL;
 
  ret = 1;
 
 done:
  return ret;
}

/*
 * Given a FrSecPassInfo object, frSecPassInfoToMem() copies the content to a memory buffer.
 * The calling function can then use that buffer in operations such as writing to a file.
 * 
 * The necessary memory space is allocated by the function and is returned in *mem with
 * its size in *msz.
 *
 * The calling function is responsible for freeing the allocated memory.
 *
 */

DLLExport int frSecPassInfoToMem(ZkError err, FrSecPassInfo pinfo,
                                 void **mem, size_t *msz)
{
  int ret = 0;
  const char *func = "frSecPassInfoToMem";
  struct fr_sec_passinfo_pk *pk;

  zkErrorClear(err);
  if ( ! pinfo ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"pfino is null" );    
    goto done;
  }

  if ( (pk = malloc(sizeof *pk)) == NULL ) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                  "can't allocate %lu bytes for packed data:: %s",
                  (unsigned long)(sizeof *pk), strerror(errno));
    goto done;
  }

  pk->vers = 1;
  pk->type = FRPKTTYPE_PASSINFO ;
  memcpy( pk->salt_key, pinfo->salt_key, sizeof pk->salt_key );
  memcpy( pk->salt_ver, pinfo->salt_ver, sizeof pk->salt_ver );
  memcpy( pk->hash_ver, pinfo->hash_ver, sizeof pk->hash_ver );  
  pk->logIters = htonl(pinfo->logIters);

  *mem = pk;
  *msz = sizeof *pk;
  ret = 1;
  
 done:
  return ret;
}

/*
 * Given a buffer (buf) of size (bsz) and a FrSecPassInfo (pinfo), the function
 * fills in the pinfo object with the buffer content.
 *
 */

DLLExport int frSecPassInfoSetFromBuf(ZkError err, FrSecPassInfo pinfo,
                                      void *buf, size_t bsz)
{
  int ret = 0;
  const char *func = "frSecPassInfoSetFromBuf";
  struct fr_sec_passinfo_pk *pk;

  zkErrorClear(err);
  if ( !pinfo ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"pfino is null" );    
    goto done;
  }
  
  if ( !buf ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"buf is null" );    
    goto done;
  }
  
  if ( bsz != sizeof(struct fr_sec_passinfo_pk) ) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,
                  "unexpected buffer size %lu bytes, expecting %lu bytes",
                  (unsigned long)bsz, (unsigned long)(sizeof(struct fr_sec_passinfo_pk))  );    
    goto done;
  }

  pk = buf;
  memcpy( pinfo->salt_key, pk->salt_key, sizeof pk->salt_key );
  memcpy( pinfo->salt_ver, pk->salt_ver, sizeof pk->salt_ver );
  memcpy( pinfo->hash_ver, pk->hash_ver, sizeof pk->hash_ver );  
  pinfo->logIters = ntohl(pk->logIters);

  ret = 1;
  
 done:
  return ret;
}

#ifndef WIN32
/*
 * Reads a passphrase from a stream (fp). If the input is a TTY, the echo on
 * the TTY is turned off so that no entered passphrase is displayed.
 *
 * 2 prompts can be supplied. The 2nd prompt is used when one asks for
 * confirmation of the passphrase. The use of these prompts depends on the
 * input device used. If fp is a file.. then the prompts are not used. If
 * on the other hand it's a tty then the prompts are used as follows.
 *
 * prompt1 = NULL && prompt2 = NULL -> Use default prompt for prompt1
 * prompt1 = NULL && prompt2 != NULL -> Error
 * prompt1 != NULL && prompt2 = NULL -> Use prompt1
 * prompt1 != NULL && prompt2 != NULL -> Use prompt1 and prompt2
 *
 * The passphrase is returned in the *pass and its length in *plen.
 *
 */
#define PROMPT_DEFAULT "Enter passphrase: "

int frSecPassReadFromFp(ZkError err, FILE *fp,
                                  const char *prompt1, const char *prompt2,
                                  char **pass, size_t *plen)
{
  int ret = 0;
  const char *func = "frSecPassReadFromFp";
  int fd;
  char *p1 = 0, *p2 = 0;
  int p1len =0, p2len = 0;
  struct termios term, newterm;
  
  zkErrorClear(err);

#ifndef WIN32
  
  if ( ! fp ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"fp is null" );
    goto done;
  }

  if ( ! pass ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"pass is null" );
    goto done;
  }

  fd = fileno(fp);
  if ( isatty( fd ) ) { /* This is a tty. so remove echo and display prompts */
    
    if ( ! prompt1 ) {
      if ( prompt2 ) {
        zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"prompt1 is null while prompt2 is not null.");
        goto done;        
      }
      
      if ((prompt1 = strdup(PROMPT_DEFAULT)) == NULL ) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,"strdup: Unable to duplicate default prompt: %s",strerror(errno));
      goto done;        
      }
      
    }

    /* Need to turn off echo on the file descriptor fd */

    if ( tcgetattr(fd, &term) < 0) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,"tcgetattr: %s",strerror(errno));
      goto done;
    }
    memcpy(&newterm, &term, sizeof term);
    newterm.c_lflag &= ~ECHO;
    if ( tcsetattr(fd, TCSANOW, &newterm) < 0 ) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,"tcsetattr: %s",strerror(errno));
      goto done;
    }
    
    if ( ! readPassPhrase( err, fp, prompt1, &p1, &p1len ) ||
         (prompt2 && ! readPassPhrase( err, fp, prompt2, &p2, &p2len )) ) {
      tcsetattr(fd, TCSANOW, &term);
      goto done;
    }

    /* Reset it to the original value */
    if ( tcsetattr(fd, TCSANOW, &term) < 0 ) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,"tcsetattr: %s",strerror(errno));
      goto done;
    }

    /* Probably need to restore the state of the fd */
    
    if ( prompt2 && strcmp(p1,p2) ) {
      goto done;
    }

    if ( p2 ) {
      memset(p2, 0, strlen(p2));
      free(p2); /* Free p2 and give back p1 */
      p2 = 0;
    }
    
  } else { /* reading from a file */
    if ( ! readPassPhrase( err, fp, NULL, &p1, &p1len ) ) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,"readPassPhrase() returned %d", ret);
      goto done;
    }
  }

  *pass = p1; p1 = 0;
  *plen = p1len;
  
  ret = 1;
  
 done:
  if ( p1 ) {memset(p1,0,strlen(p1));free(p1);}
  if ( p2 ) {memset(p2,0,strlen(p2));free(p2);}
  
  return ret;

#else

  /*
   *  This function is disabled under Windows because it uses functions that
   *  are probably UNIX specific, such as isatty(), tcgetattr(), etc.
   */

  zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKESTUB, "not implemented under Windows" );
  return 0;

#endif
}
#endif /* WIN32 */

#ifndef WIN32
/*
 * Reads the passphrase from a file
 *
 */

int frSecPassReadFromFile(ZkError err, const char *file,
                                    const char *prompt1, const char *prompt2,
                                    char **pass, size_t *plen)
{
  int ret = 0;
  const char *func = "frSecPassReadFromFile";
  FILE *fp;
  int fd = 0;

  zkErrorClear(err);
  if ( ! file ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"file is null" );
    goto done;
  }

  if ( ! pass ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"pass is null" );
    goto done;
  }

  if ( ! (fp = fopen(file, "r")) ) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                  "can't open `%s': %s.\n", file, strerror(errno));
    goto done;
  }

  fd = fileno(fp);
  if ( isatty( fd ) ) { /* This is a tty. so reopen the file r+, one cannot open r+ a pipe
                           where the other end is opened for write only */
    if ( fclose(fp) < 0 ) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                    "can't fclose `%s': %s.\n", file, strerror(errno));
      goto done;
    }
    if ( ! (fp = fopen(file, "r+")) ) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                    "can't open `%s': %s.\n", file, strerror(errno));
      goto done;
    }
  }

  
  ret = frSecPassReadFromFp(err,fp, prompt1, prompt2, pass, plen);
  
  if ( fclose(fp) < 0 ) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                  "can't fclose `%s': %s.\n", file, strerror(errno));
  }
  
 done:
  return ret;
}
#endif /* WIN32*/


/*
 * Performs the creation of the hash values by assembling the salt and the data
 *
 */

static int hashCreate( ZkError err, char salt[], const char *data, size_t dlen, char **hash, size_t *hlen, unsigned int logiters)
{
  char *h = 0;
  const char *func = "hashCreate";
  int ret = 0;
  size_t sz;
  
  sz = FRCRYPT_SALT_SIZE+dlen;
  if ( ! (h = malloc(sz)) ) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                  "can't allocate %lu bytes for salt and data: %s",
                  (unsigned long)sz, strerror(errno));
    goto done;
  }
  memcpy(h, salt, FRCRYPT_SALT_SIZE);
  memcpy(h+FRCRYPT_SALT_SIZE, data, dlen);

  if ( ! frSecStretchDataToMem(err, logiters, h, sz, (void *)hash, hlen)) {
    goto done;
  }
  
  if ( *hlen != FRCRYPT_HASH_SIZE ) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, ZKEBADSTATE,
                  "unexpected stretched data length (%lu != %lu)",
                  (unsigned long)*hlen, (unsigned long)FRCRYPT_HASH_SIZE);
    goto done;
  }  

  ret = 1;
  
 done:
  if (h) free(h);
  
  return ret;
}

/*
 * This function given a passphrase, generates a symmetric key and fills in
 * the FrSecPassInfo object pinfo with the generated intermediate data as per
 * the following algo...
 *
 * Let
 *
 *    P = <new passphrase>
 *
 *  Generate
 *
 *   S_key = <new random salt for the key>
 *   S_ver = <new random salt for the verification>
 *
 *   H_key = Stretch(Concat(S_key, P))
 *   H_ver = Stretch(Concat(S_ver, H_key))
 *
 *  then
 *
 *    K = KeyFromData(H_key)
 *
 *  where K is the new symmetric key used to encrypt and decrypt data.
 *
 *  Let
 *
 *    C = Concat(S_key, S_ver, H_ver)
 *
 *  Store C in the clear.
 *
 *
 * Note that S_key, S_ver and H_ver are stored in pinfo along with
 * the parameter logiters. This info is required for the verification.
 *
 * 
*/
 

DLLExport int frSecPassSetSymKey(ZkError err, const char *pass,
                                 unsigned int logiters, FrCryptSym key,
                                 FrSecPassInfo pinfo)
{
  int ret = 0;
  const char *func = "frSecPassSetSymKey";
  char *h_key = 0, *h_ver = 0;
  size_t h_key_len, h_ver_len;

  zkErrorClear(err);

  if ( !pass ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"pass is null" );
    goto done;
  }
  if ( !key ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"key is null" );
    goto done;
  }
  if ( !pinfo ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"pinfo is null" );
    goto done;
  }
  
  /* Compute Hash key */
  if ( ! hashCreate( err, pinfo->salt_key, pass, strlen(pass), &h_key, &h_key_len, logiters ) ) {
    goto done;
  }

  /* Compute Hash verify */
  if ( ! hashCreate( err, pinfo->salt_ver, h_key, h_key_len, &h_ver, &h_ver_len, logiters ) ) {
    goto done;
  }
  memcpy(pinfo->hash_ver, h_ver, sizeof pinfo->hash_ver);

  if ( ! frCryptSymKeyFromData(err, key, h_key, h_key_len) ) {
    goto done;
  }
  
  pinfo->logIters = logiters;
  ret = 1;
  
 done:
  
  if ( h_key ) { memset(h_key, 0, h_key_len); free(h_key); }
  if ( h_ver ) { memset(h_ver, 0, h_ver_len); free(h_ver); }
  
  return ret;
}



/*
 *
 * Verify that a given passphrase is valid. The passinfo object must be
 * filled with existing valid information.
 * If the function is successful, the parameter *valid is filled with 1
 * on success, 0 otherwise.
 *
 * The algorithm used to do the verification is as follows
 *
 *  Let
 *    P'     = <supposed passphrase>
 *    S_ver  = <current verification salt from C>
 *    H_ver  = <current verification hash from C>
 *
 *   Generate:
 *
 *    H_key' = Stretch(Concat(S_key, P'))
 *    H_ver' = Stretch(Concat(S_ver, H_key'))
 *
 *   We consider P' to be the correct passphrase if and only if H_ver' = H_ver.
 *   
 */


DLLExport int frSecPassVerify(ZkError err, const char *pass, FrSecPassInfo pinfo, int *valid)
{
  int ret = 0;
  const char *func = "frSecPassVerify";
  char *h_key, *h_ver;
  int h_key_len, h_ver_len;

  zkErrorClear(err);
  if ( !pass ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"pass is null" );
    goto done;
  }
  if ( !pinfo ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"pinfo is null" );    
    goto done;
  }
  if ( !valid ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"valid is null" );    
    goto done;
  }

  /* Compute Hash key ' */
  if ( ! hashCreate( err, pinfo->salt_key, pass, strlen(pass), &h_key, &h_key_len, pinfo->logIters ) ) {
    goto done;
  }

  /* Compute Hash verify ' */
  if ( ! hashCreate( err, pinfo->salt_ver, h_key, h_key_len, &h_ver, &h_ver_len, pinfo->logIters ) ) {
    goto done;
  }
  
  *valid =  ( memcmp(pinfo->hash_ver, h_ver, FRCRYPT_HASH_SIZE) == 0 ) ? 1 : 0 ;

  ret = 1;
  
 done:
  if ( h_key ) free(h_key);
  if ( h_ver ) free(h_ver);  
  
  return ret;
}

/*
 * This function is a mixture of the functions frSecPassSetSymKey() and
 * frSecPassVerify().
 *
 * It verifies that the given passphrase (pass) along with FrSecPassInfo object
 * (pinfo) are valid, in which case it builds the symmetric key and stores
 * the info in (key).
 *
 * If the passphrase is invalid, the code of the error object (err) is set to
 * ZKEDATABAD.
 *
 * Function returns 0 on failure and 1 on success.. i.e. if everything works.
 *
 * See the functions frSecPassSetSymKey() and frSecPassVerify() for the
 * algorithms.
 *
 */

DLLExport int frSecPassVerifySetSymKey(ZkError err, const char *pass,
                                       FrCryptSym key, FrSecPassInfo pinfo)
{
  int ret = 0;
  const char *func = "frSecPassVerifySetSymKey";
  char *h_key = 0, *h_ver = 0;
  size_t h_key_len, h_ver_len;

  zkErrorClear(err);

  if ( !pass ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"pass is null" );
    goto done;
  }
  if ( !key ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"key is null" );
    goto done;
  }
  if ( !pinfo ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"pinfo is null" );
    goto done;
  }
  
  /* Compute Hash key */
  if ( ! hashCreate( err, pinfo->salt_key, pass, strlen(pass), &h_key, &h_key_len, pinfo->logIters ) ) {
    goto done;
  }

  /* Compute Hash verify */
  if ( ! hashCreate( err, pinfo->salt_ver, h_key, h_key_len, &h_ver, &h_ver_len, pinfo->logIters ) ) {
    goto done;
  }

  if ( memcmp(pinfo->hash_ver, h_ver, FRCRYPT_HASH_SIZE) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKEDATABAD, "passphrase is invalid");
    goto done;
  }

  if ( ! frCryptSymKeyFromData(err, key, h_key, h_key_len) ) {
    goto done;
  }
  
  ret = 1;
  
 done:
  
  if ( h_key ) { memset(h_key, 0, h_key_len); free(h_key); }
  if ( h_ver ) { memset(h_ver, 0, h_ver_len); free(h_ver); }
  
  return ret;
}

DLLExport int frSecGetLogiters(ZkError err, FrSecPassInfo pinfo, int *logiters )
{
  int ret = 0;
  const char *func = "frSecGetLogiters";

  if ( !pinfo) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"pinfo is null" );
    goto done;
  }

  *logiters = pinfo->logIters;

  ret = 1;

 done:

  return ret;
}



/*
  Often, it is necessary to retrieve a symmetric key without having to handle
  aspects like reading the passphrase, verifying the passphrase, 
  handling a passinfo structure. This function provides this behavior.
  The function returns 1 only if the passphrase is valid and the key
  was retrieved succesfully.
  
  pfile   is the location of the passinfo file.
  file   tells where to read the passphrase from.
  prompt is a string that will be used when the source (file) is not a
         regular file.
  key    is where the key will be stored
  alg    is the type of algorithm to use when creating the key.
         
 */
#ifndef WIN32
int frSecPassSymKey( ZkError err, const char *pfile,
                               const char *file, const char *prompt,
                               FrCryptSym *key, FrCryptSymAlg alg  )
{
  const char *func = "frSecPassSymKey";
  int ret = 0;
  FILE *fp = 0;
  FrSecPassInfo pinfo = 0;
  char *pass = 0;
  char *buf = 0;
  int plen, bsz;

  zkErrorClear(err);
  
  if ( !key ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"key is null" ); 
    goto done;
  }
  if ( !pfile ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR,"pfile is null" ); 
    goto done;
  }
  
  /* Create Sym Key */
  if ( ! frCryptSymNew(err, key) ||
       ! frCryptSymSetAlg(err, *key, alg ) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,
               "can't allocate symmetric key object.");
    goto done;
  }


  /* Create new pinfo */
  if ( ! frSecPassInfoNew(err, &pinfo) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,
               "can't allocate symmetric key object.");
    goto done;
  }

  /* Read the passphrase */
  if ( ! frSecPassReadFromFile(err, file, prompt, NULL, &pass, &plen) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,        
               "Cannot read passphrase.");
    goto done;
  }

  /* Read passinfo file */ 
  if ( (fp = fopen(pfile, "rb")) == NULL ) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                  "can't open `%s': %s.\n", pfile, strerror(errno));
    goto done;
  }
  
  /* Let get the passinfo file */
  if ( ! zkFileReadFp( err, fp, 0, (void*)&buf, &bsz ) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,
               "Cannot read pass info from file.");
    goto done;
  }

  if ( ! frSecPassInfoSetFromBuf(err, pinfo, buf, bsz ) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,        
               "Cannot convert passinfo.");
    goto done;
  }

  /* Free everything */
  if ( ! frSecPassVerifySetSymKey( err, pass, *key, pinfo) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,        
               "Cannot verify/set key.");
    goto done;
  }

  ret = 1;
  
 done:
  
  if (pinfo) { frSecPassInfoFree(err, &pinfo); }
  if (pass) { memset(pass, 0, strlen(pass)); free(pass); }
  if (buf) { memset(buf, 0, bsz); free(buf); }
  if ( ret == 0 && *key ) { frCryptSymFree(err, key); }
  if (fp) {fclose(fp);}

  return ret;
}
#endif /* Win32 */
