/*
 *  $Id: frsec_sig.c,v 1.11.20.1 2000/10/11 13:19:16 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Sign arbitrary data with a signature key.
 *  
 *  We prepend the data with a special header containing information about the
 *  version of the signature key being used, the exact amount of signed data,
 *  etc.
 */

#include "zks.h"

#ifndef WIN32
#include <arpa/inet.h>
#else
#include <winsock2.h>
#endif

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include "frsec_sig.h"
#include "frPktFmt.h"
#include "zkLog.h"
#include "frsec_sig_v2.h"

#define MYMOD ZKMFRSEC

#pragma pack(1)

/*
 *  The header prepended to the signed data.
 *  
 *  This header should be a multiple of eight (8) bytes in length, to avoid
 *  messing up any other encryption that may be involved.
 */
struct frsec_sig_pk {
  uint8_t vers;
  uint8_t type;
  char sig[FRCRYPT_SIG_SIZE];
  char signer[FRENT_PORTABLE_SIZE];
  uint32_t dataLen;
};

#pragma pack()


static int sigCheckHdr(ZkError err, const char *fname, struct frsec_sig_pk *sigHdr, size_t dsz)
{
  zkErrorClear(err);

  switch ( sigHdr->vers ) {
  case 1: 
    return sigCheckHdr_v1(err, fname, sigHdr, dsz);
  case 2:
    return sigCheckHdr_v2(err, fname, sigHdr, dsz);
  default:    /* bug */
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEDATABAD,
                  "invalid version (%u) in signature header",
                  (unsigned int)sigHdr->vers);
    return 0;
  }
}


DLLExport int frSecSigGetVersion(ZkError err, void *buf, size_t bsz, unsigned int *vers)
{
  const char *fname = "frSecSigGetVersion";
  struct frsec_sig_pk *sigHdr = buf;

  zkErrorClear(err);

  if ( ! buf) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "buf is null");
    return 0;
  }
  if (bsz < sizeof *sigHdr) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEPARMBAD,
                  "buffer size is too small (%lu < %lu)",
                  (unsigned long)bsz, (unsigned long)sizeof *sigHdr);
    return 0;
  }
  if ( ! vers) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "vers is null");
    return 0;
  }

  if ( ! (sigCheckHdr(err, fname, buf, bsz) &&
          frCryptSigBufGetVersion(err, sigHdr->sig, sizeof sigHdr->sig, vers))) {
    return 0;
  }

  return 1;
}


DLLExport int frSecSigGetEntity(ZkError err, void *buf, size_t bsz, FrEntIdExp *ent)
{
  const char *fname = "frSecSigGetEntity";
  struct frsec_sig_pk *sigHdr = buf;

  zkErrorClear(err);

  if ( ! buf) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "buf is null");
    return 0;
  }
  if (bsz < sizeof *sigHdr) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEPARMBAD,
                  "buffer size is too small (%lu < %lu)",
                  (unsigned long)bsz, (unsigned long)sizeof *sigHdr);
    return 0;
  }
  if ( ! ent) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "ent is null");
    return 0;
  }

  if ( ! sigCheckHdr(err, fname, buf, bsz)) {
    return 0;
  }

  if ( ! frEntLoad(err, ent, sigHdr->signer, sizeof sigHdr->signer)) {
    return 0;
  }

  return 1;
}


/*
 *  Allocate memory for a signed version of `src'.
 */
DLLExport int frSecSignData(ZkError err, FrCryptSig sigKey, void *src, size_t ssz, void **dst, size_t *dsz)
{
  const char *fname = "frSecSignData";
  int ret = 0;
  size_t sz;
  struct frsec_sig_pk *sigHdr;
  void *d = 0;

  zkErrorClear(err);

  if ( ! sigKey) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "sigKey is null");
    return 0;
  }
  if ( ! src) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "src is null");
    return 0;
  }
  if ( ! dst) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "dst is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "dsz is null");
    return 0;
  }

  sz = ssz + sizeof *sigHdr;

  if ( ! (d = malloc(sz))) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, errno,
                  "can't allocate %lu bytes for signed data: %s",
                  (unsigned long)sz, strerror(errno));
    goto done;
  }

  sigHdr = d;

  sigHdr->vers = 1; /* bug */
  sigHdr->type = FRPKTTYPE_SIGNED_DATA;
  {
    FrEntIdExp sigEnt;
    
    if ( ! frCryptPubKeyGetEntity(err, sigKey, &sigEnt)) {
      goto done;
    }
    fr_EntToPacked(&sigEnt, sigHdr->signer);
  }
  sigHdr->dataLen = htonl(ssz);

  memcpy(sigHdr + 1, src, ssz);

  {
    char *sdstart = sigHdr->sig + sizeof sigHdr->sig;
    char *sdend = (char *)(sigHdr + 1) + ssz;

    if ( ! frCryptSigSignData(err, sigKey, sdstart, sdend - sdstart,
                              sigHdr->sig, sizeof sigHdr->sig)) {
      goto done;
    }
  }

  *dst = d; d = 0; /* the caller owns it now */
  *dsz = sz;

  ret = 1;

 done:

  if (d) { memset(d, 0, sz); free(d); }

  return ret;
}


DLLExport int frSecSigVerify(ZkError err, FrCryptSig sigKey, void *data, size_t dsz, int *valid)
{
  const char *fname = "frSecSigVerify";
  int ret = 0;
  struct frsec_sig_pk *sigHdr;

  zkErrorClear(err);

  if ( ! sigKey) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "sigKey is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (dsz < sizeof *sigHdr) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEPARMBAD,
                  "dsz is too small (%lu < %lu)",
                  (unsigned long)dsz, (unsigned long)sizeof *sigHdr);
    return 0;
  }
  if ( ! valid) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "valid is null");
    return 0;
  }

  sigHdr = data;

  if ( ! sigCheckHdr(err, fname, sigHdr, dsz)) {
    goto done;
  }

  {
    char *sdstart = sigHdr->sig + sizeof sigHdr->sig;
    char *sdend = (char *)sigHdr + dsz;

    if ( ! frCryptSigVerifyData(err, sigKey, sdstart, sdend - sdstart,
                                sigHdr->sig, sizeof sigHdr->sig, valid)) {
      goto done;
    }
  }

  /*
   *  Now check whether the entity in the signature key matches that in the
   *  header.  It's only a warning if they don't, since the signature has
   *  still been verified.  (Of course, we still return 0 on failure.)
   */

  {
    ZkLog log;
    FrEntIdExp hdrEnt, keyEnt;
    int equal;

    (void)zkErrorLog(err, &log);

    if ( ! (frEntLoad(err, &hdrEnt, sigHdr->signer, sizeof sigHdr->signer) &&
            frCryptPubKeyGetEntity(err, sigKey, &keyEnt) &&
            frEntEqual(err, &hdrEnt, &keyEnt, &equal) && equal)) {
      zkLogError(log, err);
      zkErrorSet(err, fname, ZKWARN, MYMOD, ZKEDATABAD,
                 "entities are not the same");
      /* bug? what if they're both null? */
      goto done;
    }
  }
  
  ret = 1;

 done:

  return ret;
}


/*
 *  Remove the leading signature header, returning only the data that was
 *  signed.  Memory is allocated for the data, and returned via `*dst'.  The
 *  length of the data is returned via `dsz'.
 *  
 *  The signed data is not verified.
 */
DLLExport int frSecSigRemove(ZkError err, void *src, size_t ssz, void **dst, size_t *dsz)
{
  const char *fname = "frSecSigRemove";
  struct frsec_sig_pk *sigHdr;

  zkErrorClear(err);

  if ( ! src) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "src is null");
    return 0;
  }
  if (ssz < sizeof *sigHdr) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEPARMBAD,
                  "ssz is too small (%lu < %lu)",
                  (unsigned long)ssz, (unsigned long)sizeof *sigHdr);
    return 0;
  }

  sigHdr = src;

  if ( ! sigCheckHdr(err, fname, sigHdr, ssz)) {
    return 0;
  }

  switch ( sigHdr->vers ) {
  case 0:  /* bug */
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEDATABAD,
                  "invalid version (%u) in signature header",
                  (unsigned int)sigHdr->vers);
    return 0;
  case 1: 
    return frSecSigRemove_v1(err, src, ssz, dst, dsz);
  case 2:
  default:
    return frSecSigRemove_v2(err, src, ssz, dst, dsz);
  }
}


/*
 *  Shift the data to the beginning of the buffer, overwriting the signature
 *  header.  Initially `*dsz' should be the size of the buffer, after shifting
 *  the data it will be the previous value less the size of the header.
 *  
 *  The signed data is not verified.
 */
DLLExport int frSecSigRemoveInSitu(ZkError err, void *data, size_t *dsz)
{
  const char *fname = "frSecSigRemoveInSitu";
  struct frsec_sig_pk *sigHdr;

  zkErrorClear(err);

  if ( ! data) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "dsz is null");
    return 0;
  }
  if (*dsz < sizeof *sigHdr) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEPARMBAD,
                  "*dsz is too small (%lu < %lu)",
                  (unsigned long)dsz, (unsigned long)sizeof *sigHdr);
    return 0;
  }

  sigHdr = data;

  if ( ! sigCheckHdr(err, fname, sigHdr, *dsz)) {
    return 0;
  }

  switch ( sigHdr->vers ) {
  case 0:  /* bug */
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEDATABAD,
                  "invalid version (%u) in signature header",
                  (unsigned int)sigHdr->vers);
    return 0;
  case 1: 
    return frSecSigRemoveInSitu_v1(err, data, dsz);
  case 2:
  default:
    return frSecSigRemoveInSitu_v2(err, data, dsz);
  }
}
