/*
 *  $Id: frsec_sig_v1.c,v 1.1.22.1 2000/10/11 13:19:16 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Sign arbitrary data with a signature key.
 *  
 *  We prepend the data with a special header containing information about the
 *  version of the signature key being used, the exact amount of signed data,
 *  etc.
 */

#include "zks.h"

#ifndef WIN32
#include <arpa/inet.h>
#else
#include <winsock2.h>
#endif

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include "frsec_sig.h"
#include "frPktFmt.h"
#include "zkLog.h"
#include "frsec_sig_v2.h"

#define MYMOD ZKMFRSEC

#pragma pack(1)

/*
 *  The header prepended to the signed data.
 *  
 *  This header should be a multiple of eight (8) bytes in length, to avoid
 *  messing up any other encryption that may be involved.
 */
struct frsec_sig_pk {
  uint8_t vers;
  uint8_t type;
  char sig[FRCRYPT_SIG_SIZE];
  char signer[FRENT_PORTABLE_SIZE];
  uint32_t dataLen;
};

#pragma pack()


DLLExport int sigCheckHdr_v1(ZkError err, const char *fname, struct frsec_sig_pk *sigHdr, size_t dsz)
{
  FrEntIdExp sigEnt;
  int ret = 0, valid;
  size_t dataLen;
  
  zkErrorClear(err);

  if (sigHdr->vers != 1 /* bug */) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEDATABAD,
                  "invalid version (%u) in signature header",
                  (unsigned int)sigHdr->vers);
    goto done;
  }
  if (sigHdr->type != FRPKTTYPE_SIGNED_DATA) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEDATABAD,
                  "invalid type (%u) in signature header",
                  (unsigned int)sigHdr->type);
    goto done;
  }
  fr_EntFromPacked(&sigEnt, sigHdr->signer);
  if ( ! (frEntIsValid(err, &sigEnt, &valid) && valid)) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEDATABAD,
                  "invalid signing entity (%u, %.64s)", /* bug */
                  (unsigned int)sigEnt.type, sigEnt.name);
    goto done;
  }
  
  dataLen = ntohl(sigHdr->dataLen);

  if (dataLen + sizeof *sigHdr != dsz) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEDATABAD,
                  "inconsistent data lengths (%lu + %lu != %lu)",
                  (unsigned long)dataLen, (unsigned long)sizeof *sigHdr,
                  (unsigned long)dsz);
    goto done;
  }

  ret = 1;
  
 done:

  return ret;
}


/*
 *  Remove the leading signature header, returning only the data that was
 *  signed.  Memory is allocated for the data, and returned via `*dst'.  The
 *  length of the data is returned via `dsz'.
 *  
 *  The signed data is not verified.
 */
DLLExport int frSecSigRemove_v1(ZkError err, void *src, size_t ssz, void **dst, size_t *dsz)
{
  const char *fname = "frSecSigRemove_v1";
  int ret = 0;
  size_t sz = 0;
  struct frsec_sig_pk *sigHdr;
  void *d = 0;

  zkErrorClear(err);

  if ( ! src) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "src is null");
    return 0;
  }
  if (ssz < sizeof *sigHdr) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEPARMBAD,
                  "ssz is too small (%lu < %lu)",
                  (unsigned long)ssz, (unsigned long)sizeof *sigHdr);
    return 0;
  }
  if ( ! dst) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "dst is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "dsz is null");
    return 0;
  }

  sigHdr = src;

  if ( ! sigCheckHdr_v1(err, fname, sigHdr, ssz)) {
    goto done;
  }

  sz = ssz - sizeof *sigHdr;

  if ( ! (d = malloc(sz))) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, errno,
                  "can't allocate %lu bytes for data: %s",
                  (unsigned long)sz, strerror(errno));
    goto done;
  }

  memcpy(d, sigHdr + 1, sz);

  *dst = d; d = 0; /* the caller now owns it */
  *dsz = sz;
  
  ret = 1;

 done:

  if (d) { memset(d, 0, sz); free(d); }

  return ret;
}


/*
 *  Shift the data to the beginning of the buffer, overwriting the signature
 *  header.  Initially `*dsz' should be the size of the buffer, after shifting
 *  the data it will be the previous value less the size of the header.
 *  
 *  The signed data is not verified.
 */
DLLExport int frSecSigRemoveInSitu_v1(ZkError err, void *data, size_t *dsz)
{
  const char *fname = "frSecSigRemoveInSitu_v1";
  int ret = 0;
  size_t sz;
  struct frsec_sig_pk *sigHdr;

  zkErrorClear(err);

  if ( ! data) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "dsz is null");
    return 0;
  }
  if (*dsz < sizeof *sigHdr) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEPARMBAD,
                  "*dsz is too small (%lu < %lu)",
                  (unsigned long)dsz, (unsigned long)sizeof *sigHdr);
    return 0;
  }

  sigHdr = data;

  if ( ! sigCheckHdr_v1(err, fname, sigHdr, *dsz)) {
    goto done;
  }

  sz = *dsz - sizeof *sigHdr;

  memmove(sigHdr, sigHdr + 1, sz);

  *dsz = sz;

  ret = 1;

 done:

  return ret;
}
