/*
 *  $Id: frsec_sig_v2.c,v 1.1.22.1 2000/10/11 13:19:16 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Sign arbitrary data with a signature key.
 *  
 *  We prepend the data with a special header containing information about the
 *  version of the signature key being used, the exact amount of signed data,
 *  etc.
 *
 *  This file deals with a new extension mechanism.  In version 1.1 the
 *  code will just ignore the contents of the extension.  This just allows
 *  v1.1 deployed code to function with v1.2 code which is generating and 
 *  using the extension.
 *
 */

#include "zks.h"

#ifndef WIN32
#include <arpa/inet.h>
#else
#include <winsock2.h>
#endif

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include "frsec_sig.h"
#include "frPktFmt.h"
#include "zkLog.h"
#include "frsec_sig_v2.h"

#define MYMOD ZKMFRSEC

#pragma pack(1)

/*
 *  The header prepended to the signed data.
 *  
 *  This header should be a multiple of eight (8) bytes in length, to avoid
 *  messing up any other encryption that may be involved.
 * 
 *  With v2 the header is no longer a multiple of 8 bytes in length, but
 *  it doesn't matter.
 *
 */

struct frsec_sig_pk {
  uint8_t vers;
  uint8_t type;
  char sig[FRCRYPT_SIG_SIZE];
  char signer[FRENT_PORTABLE_SIZE];
  uint32_t dataLen;
  uint32_t extLen;
};

#pragma pack()


DLLExport int sigCheckHdr_v2(ZkError err, const char *fname, struct frsec_sig_pk *sigHdr, size_t dsz)
{
  FrEntIdExp sigEnt;
  int ret = 0, valid;
  size_t dataLen;
  
  zkErrorClear(err);

  if (sigHdr->vers != 2 /* bug */) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEDATABAD,
                  "invalid version (%u) in signature header",
                  (unsigned int)sigHdr->vers);
    goto done;
  }
  if (sigHdr->type != FRPKTTYPE_SIGNED_DATA) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEDATABAD,
                  "invalid type (%u) in signature header",
                  (unsigned int)sigHdr->type);
    goto done;
  }
  fr_EntFromPacked(&sigEnt, sigHdr->signer);
  if ( ! (frEntIsValid(err, &sigEnt, &valid) && valid)) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEDATABAD,
                  "invalid signing entity (%u, %.64s)", /* bug */
                  (unsigned int)sigEnt.type, sigEnt.name);
    goto done;
  }
  
  dataLen = ntohl(sigHdr->dataLen);
  if (dataLen + sizeof *sigHdr != dsz) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEDATABAD,
                  "inconsistent data lengths (%lu + %lu != %lu)",
                  (unsigned long)dataLen, (unsigned long)sizeof *sigHdr,
                  (unsigned long)dsz);
    goto done;
  }

  ret = 1;
  
 done:

  return ret;
}


/*
 *  Remove the leading signature header, returning only the data that was
 *  signed.  Memory is allocated for the data, and returned via `*dst'.  The
 *  length of the data is returned via `dsz'.
 *  
 *  The signed data is not verified.
 *
 *  The extension data is removed as well.
 */
DLLExport int frSecSigRemove_v2(ZkError err, void *src, size_t ssz, void **dst, size_t *dsz)
{
  const char *fname = "frSecSigRemove_v2";
  int ret = 0;
  size_t sz = 0;
  struct frsec_sig_pk *sigHdr;
  uint32_t extLen;
  void *d = 0;

  zkErrorClear(err);

  if ( ! src) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "src is null");
    return 0;
  }
  if (ssz < sizeof *sigHdr) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEPARMBAD,
                  "ssz is too small (%lu < %lu)",
                  (unsigned long)ssz, (unsigned long)sizeof *sigHdr);
    return 0;
  }
  if ( ! dst) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "dst is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "dsz is null");
    return 0;
  }

  sigHdr = src;

  if ( ! sigCheckHdr_v2(err, fname, sigHdr, ssz)) {
    goto done;
  }

  extLen = ntohl(sigHdr->extLen);
  sz = ssz - sizeof *sigHdr - extLen;

  if ( ! (d = malloc(sz))) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, errno,
                  "can't allocate %lu bytes for data: %s",
                  (unsigned long)sz, strerror(errno));
    goto done;
  }

  memcpy(d, (char*)(sigHdr + 1) + extLen, sz);

  *dst = d; d = 0; /* the caller now owns it */
  *dsz = sz;
  
  ret = 1;

 done:

  if (d) { memset(d, 0, sz); free(d); }

  return ret;
}


/*
 *  Shift the data to the beginning of the buffer, overwriting the signature
 *  header.  Initially `*dsz' should be the size of the buffer, after shifting
 *  the data it will be the previous value less the size of the header.
 *  
 *  The signed data is not verified.
 */
DLLExport int frSecSigRemoveInSitu_v2(ZkError err, void *data, size_t *dsz)
{
  const char *fname = "frSecSigRemoveInSitu_v2";
  int ret = 0;
  size_t sz;
  struct frsec_sig_pk *sigHdr;
  uint32_t extLen;

  zkErrorClear(err);

  if ( ! data) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "dsz is null");
    return 0;
  }
  if (*dsz < sizeof *sigHdr) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEPARMBAD,
                  "*dsz is too small (%lu < %lu)",
                  (unsigned long)dsz, (unsigned long)sizeof *sigHdr);
    return 0;
  }

  sigHdr = data;

  if ( ! sigCheckHdr_v2(err, fname, sigHdr, *dsz)) {
    goto done;
  }

  extLen = ntohl(sigHdr->extLen);
  sz = *dsz - sizeof *sigHdr - extLen;

  memmove(sigHdr, (char*)(sigHdr + 1) + extLen, sz);

  *dsz = sz;

  ret = 1;

 done:

  return ret;
}

/*
 *  Allocate memory for a signed version of `src'.
 *
 *  This is a test function which should never be called in any v1.1
 *  software.  It just generates an extension block, filled with 
 *  garbage to verify that the v1.1 code successfully ignores it.
 *
 *  It can be modified for the v1.2 release to actually
 *  do something useful with the extension mechanism.
 *
 *  With v1.2 this function should be removed and replaced with a _v3
 *  version.
 */
DLLExport int frSecSignData_v2(ZkError err, FrCryptSig sigKey, void *src, size_t ssz, void **dst, size_t *dsz, size_t extLen)
{
  int ret = 0;
  size_t sz;
  struct frsec_sig_pk *sigHdr;
  void *d = 0;
  uint8_t junk = 0;
  uint32_t i;
  uint8_t *extension;

  zkErrorClear(err);

  if ( ! sigKey) {
    zkErrorSet(err, "frSecSignData", ZKERR, MYMOD, ZKENULLPTR,
               "sigKey is null");
    return 0;
  }
  if ( ! src) {
    zkErrorSet(err, "frSecSignData", ZKERR, MYMOD, ZKENULLPTR,
               "src is null");
    return 0;
  }
  if ( ! dst) {
    zkErrorSet(err, "frSecSignData", ZKERR, MYMOD, ZKENULLPTR,
               "dst is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, "frSecSignData", ZKERR, MYMOD, ZKENULLPTR,
               "dsz is null");
    return 0;
  }

  sz = ssz + extLen + sizeof *sigHdr;

  if ( ! (d = malloc(sz))) {
    zkErrorSet_va(err, "frSecSignData", ZKERR, MYMOD, errno,
                  "can't allocate %lu bytes for signed data: %s",
                  (unsigned long)sz, strerror(errno));
    goto done;
  }

  sigHdr = d;

  sigHdr->vers = 2;
  sigHdr->type = FRPKTTYPE_SIGNED_DATA;
  {
    FrEntIdExp sigEnt;
    
    if ( ! frCryptPubKeyGetEntity(err, sigKey, &sigEnt)) {
      goto done;
    }
    fr_EntToPacked(&sigEnt, sigHdr->signer);
  }
  sigHdr->dataLen = htonl(ssz+extLen);
  sigHdr->extLen = htonl(extLen);

  extension = (uint8_t*)sigHdr + sizeof( *sigHdr );
  for (i=0; i < extLen; i++, junk++) {
    extension[i] = junk;
  }

  memcpy((char*)(sigHdr + 1)+extLen, src, ssz);

  {
    char *sdstart = sigHdr->sig + sizeof sigHdr->sig;
    char *sdend = (char *)(sigHdr + 1) + extLen + ssz;

    if ( ! frCryptSigSignData(err, sigKey, sdstart, sdend - sdstart,
                              sigHdr->sig, sizeof sigHdr->sig)) {
      goto done;
    }
  }

  *dst = d; d = 0; /* the caller owns it now */
  *dsz = sz;

  ret = 1;

 done:

  if (d) { memset(d, 0, sz); free(d); }

  return ret;
}
