/*
 *  $Id: keystore.c,v 1.23.12.3 2000/12/14 17:53:13 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Routines to encrcypt keys and store them in a file, and read encrypted
 *  keys from a file.
 */

#define _BSD_SOURCE /* S_ISLNK(), lstat() */

#include "zks.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>
#ifndef WIN32
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#endif
#include "frCrypt.h"
#include "fr_buf.h"
#include "fr_file.h"
#include "fr_file_v0.h"
#include "fr_keystore.h"
#include "zkAlloc.h"
#include "zkCStr.h"
#include "zkDir.h"
#include "zkClock.h"


/*
 *  Try to open a file, locking it either for reading or writing, depending
 *  upon the value of `mode'.  The lock will go away when the file is closed.
 *  
 *  A special failure case is when the file is already locked, in which case a
 *  warning is returned, and the error code is set to EACCES.  In this case
 *  the caller may want to retry the open.
 */
static int lockedFopen(ZkError err, const char *path, const char *mode, FILE **fp)
{
  FILE *f;
  int ret = 0;
#ifndef WIN32
  mode_t oumask;
#endif

  zkErrorClear(err);

#ifndef WIN32
  oumask = umask(0177);
#endif
  
  if ( ! (f = fopen(path, mode))) {
    zkErrorSet_va(err, "lockedFopen", ZKERR, ZKMFRSEC, errno,
                  "can't open `%s' with mode %s: %s",
                  path, mode, strerror(errno));
    goto done;
  }

#ifndef WIN32
  {
    struct flock lck;

    switch (mode[0]) {
    case 'a': lck.l_type = F_WRLCK; break;
    case 'r': lck.l_type = F_RDLCK; break;
    case 'w': lck.l_type = F_WRLCK; break;
    default:
      zkErrorSet_va(err, "lockedFopen", ZKERR, ZKMFRSEC, ZKEPARMBAD,
                    "unexpected first character in mode `%s'", mode);
      goto done;
    }

    if (strchr(mode, '+')) { lck.l_type = F_WRLCK; }

    lck.l_start = 0;
    lck.l_whence = SEEK_SET;
    lck.l_len = 0;

    if (fcntl(fileno(f), F_SETLK, &lck) == -1) {
      unsigned short errorCode;
      
      if (errno == EACCES || errno == EAGAIN) {
        errno = EAGAIN;
        errorCode = ZKWARN;
      } else {
        errorCode = ZKERR;
      }
      zkErrorSet_va(err, "lockedFopen", errorCode, ZKMFRSEC, errno,
                    "can't lock `%s' for %s: %s",
                    path, lck.l_type == F_RDLCK ? "reading" : "writing",
                    strerror(errno));
      goto done;
    }
  }
#endif

  *fp = f; f = 0;
  ret = 1;

 done:

#ifndef WIN32
  (void)umask(oumask);
#endif
  
  if (f) { fclose(f); }

  return ret;
}


/*
 *  Create a new public key from the encrypted contents of a buffer.  The
 *  contents of the buffer remain unchanged.
 *  
 *  bug: we should use a temporary, locked buffer to store the key prior to
 *  encryption.
 */
DLLExport int frSecPubKeyNewDecryptFromBuf(ZkError err, FrCryptPubKey *pk, FrCryptSym sym, void *buf, size_t bsz)
{
  FrCryptPubKey pubkey = 0;
  FrPubKeyType kt;
  int ret = 0;
  size_t ksz, s;
  void *kbuf = 0;

  zkErrorClear(err);

  if ( ! pk) {
    zkErrorSet(err, "frSecPubKeyNewDecryptFromBuf", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "pk is null");
    return 0;
  }
  if ( ! sym) {
    zkErrorSet(err, "frSecPubKeyNewDecryptFromBuf", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "sym is null");
    return 0;
  }
  if ( ! buf) {
    zkErrorSet(err, "frSecPubKeyNewDecryptFromBuf", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "buf is null");
    return 0;
  }
  if (bsz == 0) {
    zkErrorSet(err, "frSecPubKeyNewDecryptFromBuf", ZKERR, ZKMFRSEC, ZKEPARMBAD,
               "bsz is 0");
    return 0;
  }

  if ( ! (frSecDecryptFromBuf(err, buf, bsz, sym, &kbuf, &ksz) &&
          frCryptPubKeyNew(err, &pubkey) &&
          ((s = ksz), 1) &&
          frCryptPubKeyReadFromBuf(err, pubkey, kbuf, &s) &&
          frCryptPubKeyGetType(err, pubkey, &kt))) {
    goto done;
  }

  switch (kt) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    if ( ! (frCryptCryptMakeCrypt(err, pubkey))) {
      goto done;
    }
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    if ( ! (frCryptSigMakeSig(err, pubkey))) {
      goto done;
    }
    break;

  default:
    zkErrorSet_va(err, "frSecPubKeyNewDecryptFromBuf", ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "invalid key type %d", kt);
    goto done;
  }

  *pk = pubkey; pubkey = 0; /* the caller now owns it */

  ret = 1;

 done:

  if (pubkey) { frCryptPubKeyFree(0, &pubkey); }
  if (kbuf) { memset(kbuf, 0, ksz); free(kbuf); }

  return ret;
}


/*
 *  Allocate a buffer containing an encrypted copy of the public key.  The
 *  buffer, encrypted with `sym', is returned via `*buf' and has the size
 *  `*bsz'.
 *  
 *  bug: we should use a temporary, locked buffer to store the key prior to
 *  encryption.
 */
DLLExport int frSecPubKeyEncryptToBuf(ZkError err, FrCryptPubKey pk, FrCryptSym sym, void **buf, size_t *bsz)
{
  int ret = 0;
  size_t ksz;
  void *kbuf = 0;

  zkErrorClear(err);

  if ( ! pk) {
    zkErrorSet(err, "frSecPubKeyEncryptToBuf", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "pk is null");
    return 0;
  }
  if ( ! sym) {
    zkErrorSet(err, "frSecPubKeyEncryptToBuf", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "sym is null");
    return 0;
  }
  if ( ! buf) {
    zkErrorSet(err, "frSecPubKeyEncryptToBuf", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "buf is null");
    return 0;
  }
  if (bsz == 0) {
    zkErrorSet(err, "frSecPubKeyEncryptToBuf", ZKERR, ZKMFRSEC, ZKEPARMBAD,
               "bsz is 0");
    return 0;
  }

  if ( ! frCryptPubKeyStorageSize(err, pk, &ksz)) {
    goto done;
  }
  
  if ( ! (kbuf = malloc(ksz))) { /* bug: locked memory */
    zkErrorSet_va(err, "frSecPubKeyEncryptToBuf", ZKERR, ZKMFRSEC, errno,
                  "can't allocate %lu bytes for key buffer: %s",
                  (unsigned long)ksz, strerror(errno));
    goto done;
  }
    
  if ( ! (frCryptPubKeyWriteToBuf(err, pk, kbuf, &ksz) &&
          frSecEncryptToBuf(err, kbuf, ksz, sym, buf, bsz))) {
    goto done;
  }

  ret = 1;

 done:

  if (kbuf) { memset(kbuf, 0, ksz); free(kbuf); }

  return ret;
}


/*
 *  This function is provided only for the conversion program fr_cvtencrfmt0.
 *  
 *  WARNING: This function should not be used in any new code.
 */
DLLExport int fr_SecPubKeyNewFromFile_v0(ZkError err, FrCryptPubKey *pk, FrCryptSym key, const char *file)
{
  FILE *fp = 0;
  FrCryptPubKey p = 0;
  int ret = 0;
  size_t dsz;
  void *data = 0;

  zkErrorClear(err);

  if ( ! (lockedFopen(err, file, "rb", &fp) &&
          fr_SecDecryptFromFp_v0(err, fp, key, &data, &dsz) &&
          frCryptPubKeyNew(err, &p) &&
          frCryptPubKeyReadFromBuf(err, p, data, &dsz))) {
    goto done;
  }

  *pk = p; p = 0;
  ret = 1;

 done:

  if (fp) { fclose(fp); }
  if (p) { frCryptPubKeyFree(0, &p); }
  if (data) { memset(data, 0, dsz); free(data); }

  return ret;
}


/*
 *  Create a new public key from the encrypted contents of a file.  In order
 *  to open the file we must be able to lock it for reading.  See the comments
 *  on lockedFopen() for how this may fail.
 */
DLLExport int frSecPubKeyNewFromFile(ZkError err, FrCryptPubKey *pk, FrCryptSym key, const char *file)
{
  FILE *fp = 0;
  FrCryptPubKey p = 0;
  int ret = 0;
  size_t dsz;
  void *data = 0;

  zkErrorClear(err);

  if ( ! (lockedFopen(err, file, "rb", &fp) &&
          frSecDecryptFromFp(err, fp, key, &data, &dsz) &&
          frCryptPubKeyNew(err, &p) &&
          frCryptPubKeyReadFromBuf(err, p, data, &dsz))) {
    goto done;
  }

  *pk = p; p = 0;
  ret = 1;

 done:

  if (fp) { fclose(fp); }
  if (p) { frCryptPubKeyFree(0, &p); }
  if (data) { memset(data, 0, dsz); free(data); }

  return ret;
}


/*
 *  Store, to a file, an encrypted copy of a public key.  In order to open the
 *  file we must be able to lock it for writing.  See the comments on
 *  lockedFopen() for how this may fail.
 */
DLLExport int frSecPubKeyWriteToFile(ZkError err, const char *file, FrCryptSym key, FrCryptPubKey pk)
{
  FILE *fp = 0;
  char data[4 * 1024]; /* bug: fixed size */
  int ret = 0;
  size_t dsz = sizeof data;
  
  zkErrorClear(err);

  if ( ! (frCryptPubKeyWriteToBuf(err, pk, data, &dsz) &&
          lockedFopen(err, file, "wb", &fp) &&
          frSecEncryptToFp(err, fp, key, data, dsz))) {
    goto done;
  }

  ret = 1;
  
 done:

  memset(data, 0, sizeof data);
  if (fp) { fclose(fp); }
  
  return ret;
}


/*
 *  Return 1 if the directory is secure.  That is, it is owned by the current
 *  user, and is readable and writable only by this user, and it is not a
 *  symbolic link.
 *  
 *  bug: there's a race condition with the opening of the directory.
 */
static int dirIsSecure(ZkError err, const char *dir)
{
  struct stat sbuf;
  uid_t owner = geteuid();

  zkErrorClear(err);
  
#ifdef WIN32
#pragma message(Reminder "some of this may be problematic for Windows")
#endif
  
  if (lstat(dir, &sbuf) == -1) {
    zkErrorSet_va(err, "dirIsSecure", ZKERR, ZKMFRSEC, errno,
                  "can't lstat `%s': %s", dir, strerror(errno));
    return 0;
  }

  /* Ensure it's not a symbolic link. */

  if (S_ISLNK(sbuf.st_mode)) {
    zkErrorSet_va(err, "dirIsSecure", ZKERR, ZKMFRSEC, ZKEMLINK,
                  "`%s' is a symbolic link", dir);
    return 0;
  }

  /* It must be owned by us. Unless you are root, than you have access to 
   * the file anyway.
   */
  
  if( owner != 0 ) {
    if (sbuf.st_uid != owner) {
      zkErrorSet_va(err, "dirIsSecure", ZKERR, ZKMFRSEC, ZKEACCES,
                    "the owner (%lu) of `%s' is not me (%lu)",
                    (unsigned long)sbuf.st_uid, dir, (unsigned long)owner);
      return 0;
    }
  }

  /* No one else should be able to read or write to the directory. */

  if (sbuf.st_mode & (S_IRWXG | S_IRWXO)) {
    zkErrorSet_va(err, "dirIsSecure", ZKERR, ZKMFRSEC, ZKEACCES,
                  "*only* the owner (%lu) should be able to access `%s'",
                  (unsigned long)owner, dir);
    return 0;
  }

  return 1;
}


/*
 *  Allocate a new PubKey object corresponding to the most recent (by
 *  expiration date) key stored in the files in `dir'.  Each file is assumed
 *  to contain a public key encrypted with `sym'.  Nonetheless, errors are
 *  ignored and 0 is returned only if no key could be found, or if the most
 *  recent expiration date has already passed.
 *  
 *  Files whose names begin with a period are ignored.
 */
DLLExport int frSecPubKeyMostRecentInDir(ZkError err, const char *dir, FrCryptSym sym, FrCryptPubKey *pk)
{  
  FrCryptPubKey recent = 0, current = 0;
  ZkDir d = 0;
  char *path = 0;
  const char *file;
  ZkClock recentExp = ZKCLOCK_INIT(0), currentExp;

  zkErrorClear(err);

  if ( ! (zkDirNew(err, &d) &&
          zkDirOpen(err, d, dir))) {
    goto done;
  }

  if ( ! dirIsSecure(err, dir)) {
    goto done;
  }
  
  /* bug: check for links, permissions, etc. */
  fprintf(stderr, "**** check for links, permissions, etc. ****\n");

  while (zkDirNext(err, d) && zkDirGetFileName(err, d, &file)) {
    if (file[0] == '.') {
      continue;
    }

    if (zkCStrSprintf(&path, "%s/%s", dir, file) == -1) {
      goto done;
    }

    if (frSecPubKeyNewFromFile(err, &current, sym, path) &&
        frCryptPubKeyGetExpiration(err, current, &currentExp) &&
        zkClockCmp(currentExp, recentExp) > 0) {
      if (recent) { frCryptPubKeyFree(err, &recent); }
      recent = current; current = 0;
      recentExp = currentExp;
    }

    free(path); path = 0;
  }

 done:

  if (current) { frCryptPubKeyFree(0, &current); }
  if (d) { zkDirClose(0, d); zkDirFree(0, &d); }
  if (path) { free(path); }

  if ( ! recent) {
    zkErrorSet_va(err, "frSecPubKeyMostRecentInDir", ZKWARN, ZKMFRSEC, ZKENOTFOUND,
                  "no valid keys found in `%s'", dir);
    return 0;
  } else {
    ZkClock now = zkClockNow();

    if (zkClockCmp(recentExp, now) < 0) {
      frCryptPubKeyFree(err, &recent);
      zkErrorSet_va(err, "frSecPubKeyMostRecentInDir", ZKWARN, ZKMFRSEC, ZKENOTFOUND,
                    "all keys in `%s' have expired", file);
      return 0;
    } else {
      FrPubKeyType kt;

      if ( ! frCryptPubKeyGetType(err, recent, &kt)) {
        return 0;
      }
      switch (kt) {
      case FRCRYPT_KEYTYPE_CRYPTION:
        if ( ! frCryptCryptMakeCrypt(err, recent)) {
          return 0;
        }
        break;
      case FRCRYPT_KEYTYPE_SIGNING:
        if ( ! frCryptSigMakeSig(err, recent)) {
          return 0;
        }
        break;

      default:
        zkErrorSet_va(err, "frSecPubKeyMostRecentInDir", ZKERR, ZKMFRSEC, ZKERROR,
                      "most recent key in `%s' has invalid type (%d)", dir, kt);
        return 0;
      }

      *pk = recent;
      return 1;
    }
  }
}


/*
 *  Allocate and return, via `pk', an array of the public keys found in the
 *  directory `dir'.  The number of elements in the array will be returned via
 *  `nKeys'.
 *  
 *  The array is indexed by the key version number.  That is, pk[2] refers to
 *  the version 2 of the public key.  The 0'th element refers to the current
 *  version of the keys.  It will be a duplicate of another key in the array.
 *  
 *  Thus, the actual keys will be the non-null elements from `keys[1]' to
 *  `keys[nKeys-1]' inclusive.
 *  
 *  If no key corresponds to a particular version, or there was an error
 *  reading that version, the corresponding array element will be 0.
 *  
 *  The current key is that with the latest expiration date (even if it has
 *  already expired).
 *  
 *  Files whose names begin with a period are ignored.
 */
DLLExport int frSecPubKeyAllInDir(ZkError err, const char *dir, FrCryptSym sym, FrCryptPubKey **keys, size_t *nKeys)
{  
  char *path = 0;
  const char *file;
  FrCryptPubKey *dirKeys = 0, *retKeys = 0;
  FrCryptPubKey currKey = 0;
  int ret = 0;
  size_t i, n = 0 /* remove warning */;
  unsigned int recentVersion = 0, maxVersion = 0;
  ZkClock recentExp = ZKCLOCK_INIT(0);
  ZkDir d = 0;

  zkErrorClear(err);

  if ( ! (zkDirNew(err, &d) &&
          zkDirOpen(err, d, dir) &&
          dirIsSecure(err, dir) &&
          zkMemAlloc(err, (void **)&dirKeys, sizeof *dirKeys))) {
    goto done;
  }

  dirKeys[0] = 0;

  /* bug: check for links, permissions, etc. */

  while (zkDirNext(err, d) && zkDirGetFileName(err, d, &file)) {
    ZkClock cke; /* current key expiration date */
    unsigned int ckv; /* current key version */
    
    if (file[0] == '.') {
      continue;
    }

    if (zkCStrSprintf(&path, "%s/%s", dir, file) == -1) {
      goto done;
    }

    if ( ! (frSecPubKeyNewFromFile(err, &currKey, sym, path))) {
      continue;
    }

    if ( ! (frCryptPubKeyGetVersion(err, currKey, &ckv) &&
            frCryptPubKeyGetExpiration(err, currKey, &cke) &&
            frCryptPubKeySetSharedParms(err, currKey))) {
      goto done;
    }
    
    if ( ! zkMemRealloc(err, (void **)&dirKeys, (n+1) * sizeof *dirKeys)) {
      goto done;
    }
    dirKeys[n] = 0;

    if (ckv > maxVersion) {
      maxVersion = ckv;
    }
    if (zkClockCmp(cke, recentExp) > 0) {
      recentVersion = ckv;
    }

    dirKeys[n++] = currKey; currKey = 0;

    free(path); path = 0;
  }

  if (maxVersion == 0) {
    zkErrorSet_va(err, "frSecPubKeyAllInDir", ZKWARN, ZKMFRSEC, ZKERROR,
                  "no valid keys found in `%s'", dir);
    goto done;
  }

  /*
   *  Copy the keys to a new array, leaving gaps where necessary, and
   *  duplicate the current key in element 0.
   */

  if ( ! zkMemAlloc(err, (void **)&retKeys, (maxVersion + 1) * 
                    sizeof *retKeys)) {
    goto done;
  }
  
  for (i = 0; i <= maxVersion; i++) { retKeys[i] = 0; }

  for (i = 0; i < n; i++) {
    unsigned int kv;

    if ( ! frCryptPubKeyGetVersion(err, dirKeys[i], &kv)) {
      goto done;
    }
    if (retKeys[kv]) {
      zkErrorSet_va(err, "frSecPubKeyAllInDir", ZKERR, ZKMFRSEC, ZKERROR,
                    "duplicate key version %u found", kv);
      goto done;
    }
    retKeys[kv] = dirKeys[i]; dirKeys[i] = 0;
  }

  if ( ! frCryptPubKeyDup(err, &retKeys[0], retKeys[maxVersion])) {
    goto done;
  }

  *keys = retKeys; retKeys = 0;
  *nKeys = maxVersion + 1;

  ret = 1;

 done:

  if (path) { free(path); }
  if (dirKeys) {
    for (i = 0; i < n; i++) {
      if (dirKeys[i]) { frCryptPubKeyFree(0, &dirKeys[i]); }
    }
    free(dirKeys);
  }
  if (retKeys) {
    for (i = 0; i < maxVersion; i++) {
      if (retKeys[i]) { frCryptPubKeyFree(0, &retKeys[i]); }
    }
    free(retKeys);
  }
  if (d) { zkDirClose(0, d); zkDirFree(0, &d); }

  return ret;
}


DLLExport int frSecCryptNewFromFile(ZkError err, const char *file, FrCryptSym key, FrCryptCrypt *priv)
{
  FrCryptPubKey pk = 0;
  
  zkErrorClear(err);

  if (frSecPubKeyNewFromFile(err, &pk, key, file) &&
      frCryptCryptMakeCrypt(err, pk)) {
    *priv = pk;
    return 1;
  }

  if (pk) { frCryptPubKeyFree(err, &pk); }

  return 0;
}


/*
 *  Write the private signature key to an encrypted file.
 *  
 *  Save all the key's parameters, just in case we're given a key not using
 *  the built-in, fixed parameters.
 */
DLLExport int frSecCryptPrivWriteToFile(ZkError err, const char *file, FrCryptSym key, FrCryptCrypt priv)
{
  char data[4 * 1024]; /* bug: fixed size */
  int ret = 0;
  size_t dsz = sizeof data;
  uint8_t mask = FRCRYPT_CRYPTPARMS_ALL;
  
  zkErrorClear(err);

  if ( ! frCryptCryptStore(err, priv, mask, data, &dsz)) {
    goto done;
  }

  if ( ! frSecEncryptToFile(err, file, key, data, dsz)) {
    goto done;
  }

  ret = 1;
  
 done:

  memset(data, 0, sizeof data);
  
  return ret;
}


DLLExport int frSecSigNewFromFile(ZkError err, const char *file, FrCryptSym key, FrCryptSig *priv)
{
  FrCryptPubKey pk = 0;
  
  zkErrorClear(err);

  if (frSecPubKeyNewFromFile(err, &pk, key, file) &&
      frCryptSigMakeSig(err, pk)) {
    *priv = pk;
    return 1;
  }

  if (pk) { frCryptPubKeyFree(err, &pk); }

  return 0;
}


/*
 *  Write the private signature key to an encrypted file.
 *  
 *  Assume the key uses the shared, static parameters, saving only the key's
 *  private and public parameters.
 */
DLLExport int frSecSigPrivWriteToFile(ZkError err, const char *file, FrCryptSym key, FrCryptSig priv)
{
  char data[4 * 1024]; /* bug: fixed size */
  int ret = 0;
  size_t dsz = sizeof data;
  uint8_t mask = FRCRYPT_SIGPARMS_PUBPRIV;
  
  zkErrorClear(err);

  if ( ! frCryptSigStore(err, priv, mask, data, &dsz)) {
    goto done;
  }

  if ( ! frSecEncryptToFile(err, file, key, data, dsz)) {
    goto done;
  }

  ret = 1;
  
 done:

  memset(data, 0, sizeof data);
  
  return ret;
}
