/*
 *  $Id: pass.c,v 1.31.8.4 2000/12/14 17:53:13 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Passphrase routines.
 *  
 *  bug: there's too much overlap between the various routines.
 */

#include "zks.h"

#include <errno.h>
#include <string.h>
#ifndef WIN32
#include <unistd.h>
#endif
/* #include <math.h> */
#include "openssl/des.h"
#include "frFilenames.h"
#include "frSec.h"
#include "zkFile.h"
#include "pass.h"
#include "zkAlloc.h"

static const char *passHashFile = FREEDOM_PASSHASH_FILE;

/*
 *  The value of checkPass determines if we try to compare the passphrase with
 *  a hash stored in a file.
 *  
 *  Under Windows we never try.
 */
#ifdef WIN32
static const int checkPass = 0;
#else
static const int checkPass = 1;
#endif

/*
 *  bug: should we impose a minimum passphrase length?
 */
static const size_t pwMinSize = 8;
static const size_t pwWarnSize = 24;


/*
 *  bug: this function is taken from fr_pass.c.
 */
static int verifyPass(ZkError err, const char *pass, const char *sanfile)
{
  FILE *sfp = 0;
  char hash[FRCRYPT_HASH_SIZE];
  int ret = 0;
  size_t max = FRCRYPT_RAND_SIZE + sizeof hash + 1;
  size_t dlen, plen, slen;
  void *d = 0, *sbuf = 0;
    
  plen = strlen(pass);

  if ( ! (sfp = fopen(sanfile, "rb"))) {
    zkErrorSet_va(err, "verifyPass", ZKERR, ZKMKEYUTILS, errno,
                  "can't open `%s' for reading: %s",
                  sanfile, strerror(errno));
    goto done;
  }

  if ( ! zkFileReadFp(err, sfp, max, &sbuf, &slen)) {
    goto done;
  }

  if (slen != max - 1) {
    zkErrorSet_va(err, "verifyPass", ZKERR, ZKMKEYUTILS, ZKEDATABAD,
                  "sanity file has an invalid length (%lu != %lu)",
                  (unsigned long)slen, (unsigned long)max - 1);
    goto done;
  }

  dlen = FRCRYPT_RAND_SIZE + plen + 1;
  if ( ! (d = malloc(dlen))) {
    zkErrorSet_va(err, "verifyPass", ZKERR, ZKMKEYUTILS, errno,
                  "can't allocate %lu bytes for passphrase and salt: %s",
                  (unsigned long)dlen, strerror(errno));
    goto done;
  }

  memcpy(d, sbuf, FRCRYPT_RAND_SIZE);
  memcpy((char *)d + FRCRYPT_RAND_SIZE, pass, plen + 1);

  if ( ! frCryptHashGenerate(err, d, dlen, hash, sizeof hash)) {
    goto done;
  }

  if (memcmp(hash, (char *)sbuf + FRCRYPT_RAND_SIZE, sizeof hash) != 0) {
    zkErrorSet(err, "verifyPass", ZKWARN, ZKMKEYUTILS, ZKEDATABAD,
               "current hash does not match that in the sanity file");
    goto done;
  }
  
  ret = 1;

 done:

  if (sfp) { fclose(sfp); }
  if (d) { memset(d, 0, dlen); free(d); }
  if (sbuf) { free(sbuf); }

  return ret;
}


/*
 *  Return 1 if `bool' indicates true, return 0 if it indicates false.  If
 *  `bool' is unrecognizable as a boolean value, then return `huh'.
 *  
 *  bug: should put in libzk.
 */
static int boolFromStr(const char *bool, int huh)
{
  if (strcasecmp(bool, "1") == 0 ||
      strcasecmp(bool, "yes") == 0 ||
      strcasecmp(bool, "true") == 0) {
    return 1;
  }

  if (strcasecmp(bool, "0") == 0 ||
      strcasecmp(bool, "no") == 0 ||
      strcasecmp(bool, "false") == 0) {
    return 0;
  }

  return huh;
}


/*
 *  Return 1 if `pass' seems to be the same as the current system passphrase.
 *  
 *  bug: the name of the `sanity file' is hardwired.  Also, for debugging we
 *  always return 1 if the environment variable ZK_CHECK_PASS is set to "no".
 */
DLLExport int frSecPassphraseIsCurrent_v0(ZkError err, const char *pass)
{
  char *p;
  
  zkErrorClear(err);
  
  if ( ! checkPass) {
    return 1;
  }
  
#ifndef FREEDOM_PRODUCTION_BUILD
  /*
   *  Only allow the passphrase hash file to be bypassed in the development
   *  build.  I.e. for debugging purposes.
   */
  
  if ((p = getenv("FR_CHECK_PASS_HASH")) && ! boolFromStr(p, 1)) { /* audit: ok */
    return 1;
  }
#endif

  return verifyPass(err, pass, passHashFile);
}


/*
 *  Perform checks on the validity of the passphrase.
 */
DLLExport int frSecPassphraseValid(ZkError err, const char *pass)
{
  zkErrorClear(err);

  if ( ! pass) {
    zkErrorSet(err, "frSecPassphraseValid", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "pass is null");
    return 0;
  }

  if (strlen(pass) < pwMinSize) {
    zkErrorSet_va(err, "frSecPassphraseValid", ZKERR, ZKMFRSEC, ZKEPARMBAD,
                  "passphrase must be at least %lu characters",
                  (unsigned long)pwMinSize);
    return 0;
  }

  return 1;
}


/*
 *  The length returned via `plen' does not include the terminating nul.
 */
DLLExport int frSecGetPassphrase(ZkError err, const char *prompt, char **pass, size_t *plen)
{
  char pw[1024];
  char *p = 0;
  int ret = 0;
  size_t pwlen;
  
  zkErrorClear(err);

  if ( ! prompt) {
    zkErrorSet(err, "frSecGetPassphrase", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "prompt is null");
    return 0;
  }
  if ( ! pass) {
    zkErrorSet(err, "frSecGetPassphrase", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "pass is null");
    return 0;
  }

  /* This call also removes a trailing newline. */
  if (des_read_pw_string(pw, sizeof pw, (char *)prompt, 0) != 0) {
    zkErrorSet(err, "frSecNewSymFromPass", ZKERR, ZKMFRSEC, ZKEBADSTATE,
               "error from des_read_pw_string");
    goto done;
  }

  if ((pwlen = strlen(pw)) < pwMinSize) {
    /* bug: log to stderr (e.g.) so user knows */
    zkErrorSet_va(err, "frSecGetPassphrase", ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "passphrase is too short (%lu < %lu)",
                  (unsigned long)pwlen, (unsigned long)pwMinSize);
    goto done;
  }

  if ( ! (p = malloc(pwlen + 1))) {
    zkErrorSet_va(err, "frSecNewSymFromPass", ZKERR, ZKMFRSEC, errno,
                  "can't allocate %lu bytes: %s",
                  (unsigned long)pwlen + 1, strerror(errno));
    goto done;
  }

  memcpy(p, pw, pwlen + 1);

  *pass = p; p = 0; /* the caller now owns it */
  *plen = pwlen;
  
  ret = 1;

 done:

  memset(pw, 0, sizeof pw);
  if (p) { memset(p, 0, pwlen); free(p); }
  
  return ret;
}


/*
 *  Read a passphrase from the user, then use it to generate a symmetric key.
 *  
 *  bug: ask for passphrase twice, also loop if passphrase fails min length
 *  test.
 *  
 *  bug: provide alternate routine to get passphrase from callback routine.
 *  (I.e.  if we need to get the passphrase from some stored location.)
 *  Callback routine should have an argument telling it whether it's the first
 *  or second time called (i.e. for confirmation), so it can decide what
 *  prompt to display.
 */
DLLExport int frSecNewSymFromPass_v0(ZkError err, FrCryptSym *sym, FrCryptSymAlg alg, const char *prompt, int checkExisting)
{
  FrCryptSym s;
#define MAX_PWD 1024
  char pw[MAX_PWD];
  int ret = 0;
  size_t pwlen;
  
  zkErrorClear(err);

  if ( ! (frCryptSymNew(err, &s) &&
          frCryptSymSetAlg(err, s, alg))) {
    goto done;
  }

#ifndef WIN32
  if (des_read_pw_string(pw, sizeof pw, (char *)prompt, 0) != 0) {
    zkErrorSet(err, "frSecNewSymFromPass", ZKERR, ZKMFRSEC, ZKEBADSTATE,
               "error from des_read_pw_string");
    goto done;
  }
#else
  strncpy(pw, prompt, MAX_PWD);
#endif

  if (checkExisting && ! frSecPassphraseIsCurrent_v0(err, pw)) {
    goto done;
  }

  if ((pwlen = strlen(pw)) < pwMinSize) {
    /* bug: log to stderr (e.g.) so user knows */
    zkErrorSet_va(err, "frSecNewSymFromPass", ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "passphrase is too short (%lu < %lu)",
                  (unsigned long)pwlen, (unsigned long)pwMinSize);
    goto done;
  }

  /* bug: log warning if passphrase is below recommended length. */

  if ( ! frCryptSymKeyFromData(err, s, pw, pwlen)) {
    goto done;
  }

  ret = 1;
  *sym = s; s = 0;

 done:

  memset(pw, 0, sizeof pw);
  if (s) { frCryptSymFree(0, &s); }
  return ret;
}


/*
 *  Create a new symmetric key from the passphrase in `pw'.
 */
DLLExport int frSecNewSymFromPassPhrase_v0(ZkError err, FrCryptSym *sym, FrCryptSymAlg alg, char *pw)
{
  FrCryptSym s;
#define MAX_PWD 1024
  int ret = 0;
  size_t pwlen;
  
  zkErrorClear(err);

  if ( ! (frCryptSymNew(err, &s) &&
          frCryptSymSetAlg(err, s, alg))) {
    goto done;
  }

  if ((pwlen = strlen(pw)) < pwMinSize) {
    zkErrorSet_va(err, "frSecNewSymFromPassPhrase",
                  ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "passphrase is too short (%lu < %lu)",
                  (unsigned long)pwlen, (unsigned long)pwMinSize);
    goto done;
  }

  if ( ! frCryptSymKeyFromData(err, s, pw, pwlen)) {
    goto done;
  }

  ret = 1;
  *sym = s; s = 0;

 done:

  if (s) { frCryptSymFree(0, &s); }
  return ret;
}


/*
 *  Record the time it takes to stretch some data.
 *  
 *  For each n, from 0 to maxlogiters, record the time it takes to perform 2^n
 *  stretches.
 *  
 *  This function is for testing only.
 */
int frSec_StretchTime(ZkError err, unsigned int maxlogiters, clock_t *tm, size_t tsz)
{
  FrCryptHash hash = 0;
  char data[FRCRYPT_HASH_SIZE];
  clock_t base;
  const char *fn = "frSec_SecStretchDataTime";
  int ret = 0;
  unsigned int beg, i;

  zkErrorClear(err);
  
  if ( ! tm) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "tm is null");
    goto done;
  }

  if (tsz < maxlogiters + 1) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR,
                  "tsz must have at least %u elements", maxlogiters + 1);
    goto done;
  }

  for (i = 0; i < sizeof data; i++) {
    data[i] = i;
  }

  if ( ! frCryptHashNew(err, &hash)) {
    goto done;
  }

  base = clock();

  beg = 0;
  for (i = 1; i <= maxlogiters + 1; i++) {
    unsigned int end = 1 << (i-1), j;
    
    for (j = beg; j < end; j++) {
      if ( ! (frCryptHashReset(err, hash) &&
              frCryptHashUpdate(err, hash, data, sizeof data) &&
              frCryptHashGetHash(err, hash, data, sizeof data))) {
        goto done;
      }
    }

    tm[i-1] = clock();
    beg = end;
  }

  for (i = 0; i <= maxlogiters; i++) {
    tm[i] -= base;
  }

  ret = 1;

 done:

  if (hash) { frCryptHashFree(err, &hash); }

  return ret;
}


/*
 *
 * Stretches pw to pwLong, then uses this to generate Sym key.
 *
 */

DLLExport int frSecNewSymFromStretchedPassPhrase(ZkError err, FrCryptSym *sym, FrCryptSymAlg alg, unsigned int logiter , const char *pw)
{

 
  int ret = 0;
  FrCryptSym s;
  size_t szPwLong = 0;
  size_t szPw = 0;
  char pwLong[FRCRYPT_HASH_SIZE];
  
  zkErrorClear(err);

  if ( ! (frCryptSymNew(err, &s) &&
          frCryptSymSetAlg(err, s, alg))) {
    goto done;
  }

  if ((szPw = strlen(pw)) < pwMinSize) {
    zkErrorSet_va(err, "frSecNewSymFromStrechedPassPhrase",
                  ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "passphrase is too short (%lu < %lu)",
                  (unsigned long)szPw, (unsigned long)pwMinSize);
    goto done;
  }

  szPwLong = FRCRYPT_HASH_SIZE;
  if ( ! frSec_StretchDataToBuf(err, logiter, pw, szPw, (void *)pwLong, szPwLong)) {
    goto done;
  }

  if ( ! frCryptSymKeyFromData(err, s, pwLong, szPwLong)) {
    goto done;
  }

  ret = 1;
  *sym = s; s = 0;

 done:

  if (s) { frCryptSymFree(0, &s); }
  return ret;    
}


/*
 *  Stretches pw to pwLong, using logiter log iterations of password
 *  stretching algorithm from article "Secure Applications of Low-Entropy
 *  Keys", from Kelsey, Schneier, Hall, and Wagner.
 */
int frSec_StretchDataToBuf(ZkError err,  unsigned int logiter, const void *pw, size_t szPw, void* pwLong, size_t szPwLong )
{
  int ret = 0;
  uint32_t i = 0;
  uint32_t iterations = 0;
  FrCryptHash hash = 0;

  zkErrorClear(err);

  if ( ! pw) {
    zkErrorSet(err, "frSec_StretchDataToBuf", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "pw is null");
    goto done;
  }    
  if (logiter < 2) {
    zkErrorSet_va(err, "frSec_StretchDataToBuf",
                  ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "can't have %d < 2 log_2 iterations.",
                  logiter);
    goto done;
  }
  
  if ( szPwLong != FRCRYPT_HASH_SIZE ) {
    zkErrorSet_va(err, "frSec_StretchDataToBuf", ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "wrong size of buffer (%lu bytes).", (unsigned long)szPwLong);
    goto done;
  }

  if ( ! frCryptHashNew(err, &hash)) {
    goto done;
  }

  /* Key stretching */
  if ( ! frCryptHashUpdate(err, hash, pw, szPw)) {
    goto done;
  }
  if ( ! frCryptHashGetHash(err, hash, pwLong, szPwLong)) {
    goto done;
  }
  
  iterations = 2 << (logiter - 1);    /* iterations = 2^logiter */
  for (i = 1; i <= iterations; i ++) {
    if ( ! (frCryptHashReset(err, hash) &&
            frCryptHashUpdate(err, hash, pwLong, szPwLong) &&
            frCryptHashGetHash(err, hash, pwLong, szPwLong))) {
      goto done;
    }
  }

  ret = 1;

 done:

  if(hash) {frCryptHashFree(err, &hash);} 

  return ret;
}



DLLExport int frSecStretchDataToMem(ZkError err, unsigned int logiter, const void *pw, size_t szPw, void** pwLong, size_t* szPwLong )
{
  int ret = 0;
  char *stretchPw = NULL;
  int sz;

  zkErrorClear(err);

  if ( ! pw) {
    zkErrorSet(err, "frSecStretchDataToMem", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "pw is null");
    goto done;
  }    
  if (logiter < 2) {
    zkErrorSet_va(err, "frSecStretchDataToMem",
                  ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "can't have %d < 2 log_2 iterations.",
                  logiter);
    goto done;
  }

  sz = FRCRYPT_HASH_SIZE;
  if ( ! (stretchPw = malloc(sz)) ) {
    zkErrorSet_va(err, "frSecStretchDataToMem", ZKERR, ZKMFRSEC, errno,
                  "can't allocate %lu bytes: %s",
                  (unsigned long)sz, strerror(errno));
    goto done;
  }

  if ( ! frSec_StretchDataToBuf( err, logiter, pw, szPw, stretchPw, sz ) ) {
    goto done;
  }

  *szPwLong = FRCRYPT_HASH_SIZE;
  *pwLong = stretchPw;
  stretchPw = 0;

  ret = 1;

 done:

  if (stretchPw) free(stretchPw);

  return ret;
}



#ifndef WIN32

DLLExport int frSecPassPhraseFromFd(ZkError err, char **pass, int fd)
{
  FILE *fp = 0;
  int newfd = -1, ret = 0;
  size_t plen; /* unused */

  zkErrorClear(err);

  if ( ! pass) {
    zkErrorSet(err, "frSecPassPhraseFromFd", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "pass is null");
    return 0;
  }
  if (fd < 0) {
    zkErrorSet_va(err, "frSecPassPhraseFromFd", ZKERR, ZKMFRSEC, ZKEPARMBAD,
                  "fd (%d) is invalid", fd);
    return 0;
  }

  /*
   *  We dup() the file descriptor so our later fclose() won't also close the
   *  original file descriptor.
   */

  if ((newfd = dup(fd)) == -1) {
    zkErrorSet_va(err, "frSecPassPhraseFromFd", ZKERR, ZKMFRSEC, errno,
                  "can't dup fd (%d): %s", fd, strerror(errno));
    goto done;
  }

  if ( ! (fp = fdopen(newfd, "rb"))) {
    zkErrorSet_va(err, "frSecPassPhraseFromFd", ZKERR, ZKMFRSEC, errno,
                  "can't create FILE pointer from fd (%d): %s",
                  newfd, strerror(errno));
    return 0;
  }

  if ( ! zkFileGetLineFp(err, fp, pass, &plen)) {
    goto done;
  }
  
  ret = 1;
  
 done:

  if (fp) { fclose(fp); }
  if (newfd != -1) { close(newfd); }

  return ret;
}


#if 1

/*
 *  Read a passphrase from a file.
 *  
 *  This routine should replace frSecNewSymFromPassPhraseFd(), whose
 *  functionality it subsumes.
 */
DLLExport int frSecNewSymFromPassPhraseFile(ZkError err, FrCryptSym *sym, FrCryptSymAlg alg, const char *file, int checkExisting)
{
  char *pass = 0;
  int fd = -1, ret = 0;
  
  zkErrorClear(err);

  if ( ! sym) {
    zkErrorSet(err, "frSecNewSymFromPassPhraseFile", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "sym is null");
    return 0;
  }
  if ( ! file) {
    zkErrorSet(err, "frSecNewSymFromPassPhraseFile", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "file is null");
    return 0;
  }

  if ((fd = open(file, O_RDONLY)) == -1) {
    zkErrorSet_va(err, "frSecNewSymFromPassPhraseFile", ZKERR, ZKMFRSEC, errno,
                  "error opening `%s': %s", file, strerror(errno));
    goto done;
  }

  /*
   *  bug: check permissions, ownnership of file.
   */

  if ( ! frSecPassPhraseFromFd(err, &pass, fd)) {
    goto done;
  }

  if (checkExisting && ! frSecPassphraseIsCurrent(err, 0, pass)) {
    goto done;
  }

  if ( ! frSecNewSymFromPassPhrase(err, sym, alg, 0, pass)) {
    goto done;
  }

  ret = 1;

 done:

  if (pass) {
    memset(pass, 0, strlen(pass));
    free(pass);
  }
  if (fd != -1) { close(fd); }

  return ret;
}

#else

/*
 *  Read a passphrase from a file descriptor, converting it to a symmetric
 *  key.
 */
DLLExport int frSecNewSymFromPassPhraseFd(ZkError err, FrCryptSym *sym, FrCryptSymAlg alg, int fd)
{
  char *pass = 0;
  int ret = 0;
  
  zkErrorClear(err);

  if ( ! (frSecPassPhraseFromFd(err, &pass, fd) &&
          frSecNewSymFromPassPhrase(err, sym, alg, 0, pass))) {
    goto done;
  }

  ret = 1;

 done:

  if (pass) {
    memset(pass, 0, strlen(pass));
    free(pass);
  }

  return ret;
}

#endif
#endif

/* New version of the functions */

/*
 *  Return 1 if `pass' seems to be the same as the current system passphrase.
 *  
 *  If `pifile' is null then use the default passphrase info file.
 *  
 *  bug: For debugging we always return 1 if the environment variable
 *  ZK_CHECK_PASS is set to "no".
 */
DLLExport int frSecPassphraseIsCurrent(ZkError err, const char *pifile, const char *pass)
{
  FILE *fp = 0;
  FrSecPassInfo pinfo = 0;
  char *buf = 0;
  size_t bsz = 0;
  int valid = 0;

  zkErrorClear(err);

  if ( ! pifile) {
    pifile = FREEDOM_PASSINFO_FILE;
  }
  
  if ( ! (fp = fopen(pifile, "r"))) {
    zkErrorSet_va(err, "frSecPassphraseIsCurrent", ZKERR, ZKMFRSEC, errno,
                  "can't open `%s': %s.\n", pifile, strerror(errno));
    goto done;
  }

  /* Let get the passinfo file */
  if ( ! zkFileReadFp( err, fp, 0, (void*)&buf, &bsz ) ) {
    goto done;
  }

  if ( ! frSecPassInfoNew(err, &pinfo) ) {
    /*    zkLogWrite_va(log,"%s: can't allocate passinfo object.\n", prog); */
    goto done;
  }

  if ( ! frSecPassInfoSetFromBuf(err, pinfo, buf, bsz ) ) {
    goto done;
  }

  if ( ! frSecPassVerify( err, pass, pinfo, &valid ) ) {
    goto done;
  }
  
 done:

  if ( fp ) {
    if ( fclose(fp) < 0 ) {
      zkErrorSet_va(err, "frSecPassphraseIsCurrent", ZKERR, ZKMFRSEC, errno,
                    "can't fclose `%s': %s.\n", pifile, strerror(errno));
    }
  }

  if ( pinfo ) { frSecPassInfoFree(err, &pinfo); }
  if ( buf ) {memset(buf,0,bsz); zkMemFree(err, (void**)&buf);} 
  
  return valid;
}





/*
 *  Read a passphrase from the user, then use it to generate a symmetric key.
 *  
 *  bug: ask for passphrase twice, also loop if passphrase fails min length
 *  test.
 *  
 *  bug: provide alternate routine to get passphrase from callback routine.
 *  (I.e.  if we need to get the passphrase from some stored location.)
 *  Callback routine should have an argument telling it whether it's the first
 *  or second time called (i.e. for confirmation), so it can decide what
 *  prompt to display.
 */
DLLExport int frSecNewSymFromPass(ZkError err, FrCryptSym *sym, FrCryptSymAlg alg, const char *pifile, const char *prompt, int checkExisting)
{
  FILE *fp = 0;
  FrSecPassInfo pinfo = 0;
  char *buf = 0;
  size_t bsz = 0;
  FrCryptSym s;
#define MAX_PWD 1024
  char pw[MAX_PWD];
  int ret = 0;
  
  zkErrorClear(err);

  if ( ! (frCryptSymNew(err, &s) &&
          frCryptSymSetAlg(err, s, alg))) {
    goto done;
  }

#ifndef WIN32
  if (des_read_pw_string(pw, sizeof pw, (char *)prompt, 0) != 0) {
    zkErrorSet(err, "frSecNewSymFromPass", ZKERR, ZKMFRSEC, ZKEBADSTATE,
               "error from des_read_pw_string");
    goto done;
  }
#else
  strncpy(pw, prompt, MAX_PWD);
#endif

  if ( ! pifile) {
    pifile = FREEDOM_PASSINFO_FILE;
  }
  
  if (checkExisting && ! frSecPassphraseIsCurrent(err, pifile, pw)) {
    goto done;
  }

  if ( ! (fp = fopen(pifile, "r"))) {
    zkErrorSet_va(err, "frSecPassphraseIsCurrent", ZKERR, ZKMFRSEC, errno,
                  "can't open `%s': %s.\n", pifile, strerror(errno));
    goto done;
  }

  /* Let get the passinfo file */
  if ( ! zkFileReadFp( err, fp, 0, (void*)&buf, &bsz ) ) {
    goto done;
  }

  if ( ! frSecPassInfoNew(err, &pinfo) ) {
    /*    zkLogWrite_va(log,"%s: can't allocate passinfo object.\n", prog); */
    goto done;
  }
  
  if ( ! frSecPassInfoSetFromBuf(err, pinfo, buf, bsz ) ) {
    goto done;
  }

  if ( ! frSecPassVerifySetSymKey( err, pw, s, pinfo) ) {
    goto done;
  }
  

  ret = 1;
  *sym = s; s = 0;

 done:

  if ( fp ) {
    if ( fclose(fp) < 0 ) {
      zkErrorSet_va(err, "frSecPassphraseIsCurrent", ZKERR, ZKMFRSEC, errno,
                    "can't fclose `%s': %s.\n", pifile, strerror(errno));
    }
  }

  if ( pinfo ) { frSecPassInfoFree(err, &pinfo); }
  if ( buf ) {memset(buf,0,bsz); zkMemFree(err, (void**)&buf);} 
  
  memset(pw, 0, sizeof pw);
  if (s) { frCryptSymFree(0, &s); }
  return ret;
}


/*
 *  Create a new symmetric key from the passphrase in `pw'.
 */
DLLExport int frSecNewSymFromPassPhrase(ZkError err, FrCryptSym *sym, FrCryptSymAlg alg, const char *pifile, char *pw)
{
  FILE *fp = 0;
  FrSecPassInfo pinfo = 0;
  char *buf = 0;
  size_t bsz = 0;
  FrCryptSym s;
#define MAX_PWD 1024
  int ret = 0;
  
  zkErrorClear(err);

  if ( ! (frCryptSymNew(err, &s) &&
          frCryptSymSetAlg(err, s, alg))) {
    goto done;
  }

  if ( ! pifile) {
    pifile = FREEDOM_PASSINFO_FILE;
  }
  
  if ( ! (fp = fopen(pifile, "r"))) {
    zkErrorSet_va(err, "frSecPassphraseIsCurrent", ZKERR, ZKMFRSEC, errno,
                  "can't open `%s': %s.\n", pifile, strerror(errno));
    goto done;
  }

  /* Let get the passinfo file */
  if ( ! zkFileReadFp( err, fp, 0, (void*)&buf, &bsz ) ) {
    goto done;
  }

  if ( ! frSecPassInfoNew(err, &pinfo) ) {
    /*    zkLogWrite_va(log,"%s: can't allocate passinfo object.\n", prog); */
    goto done;
  }

  if ( ! frSecPassInfoSetFromBuf(err, pinfo, buf, bsz ) ) {
    goto done;
  }

  if ( ! frSecPassVerifySetSymKey( err, pw, s, pinfo) ) {
    goto done;
  }

  ret = 1;
  *sym = s; s = 0;

 done:

  if ( fp ) {
    if ( fclose(fp) < 0 ) {
      zkErrorSet_va(err, "frSecPassphraseIsCurrent", ZKERR, ZKMFRSEC, errno,
                    "can't fclose `%s': %s.\n", pifile, strerror(errno));
    }
  }

  if ( pinfo ) { frSecPassInfoFree(err, &pinfo); }
  if ( buf ) {memset(buf,0,bsz); zkMemFree(err, (void**)&buf);} 
  
  memset(pw, 0, sizeof pw);
  if (s) { frCryptSymFree(0, &s); }
  return ret;
}





