/*
 * $Id: frSerial.c,v 1.8.32.3 2000/12/14 17:53:16 jean Exp $
 */

static const char * module_id_ = "$Id: frSerial.c,v 1.8.32.3 2000/12/14 17:53:16 jean Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* !WIN32 */
#include <ctype.h>
#include <string.h>
#include <frCrypt.h>
#include <zkError.h>
#include <frSerial.h>

#define FRSERIALNUMBER_SZ     12  /* Size of serial number (digits) */
#define FRSNCKSUM_SZ           2  /* Size of the checksum (digits) */

/*
 * digit characters for each 5 bit value (0 to 31)
 */

const char sn_digit[32] = {
  'A', 'B', '2', '3', '4', '5', '6', '7', '8', '9', 'C', 'D', 'E', 'F', 
  'G', 'H', 'J', 'K', 'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 
  'W', 'X', 'Y', 'Z' 
};

/*
 * 5 bit value for each valid ascii digit character
 */

const unsigned char sn_value[256] = {  
/*        0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F */
/* 0 */ 128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
/* 1 */ 128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
/* 2 */ 128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
/* 3 */ 128,128,  2,  3,  4,  5,  6,  7,  8,  9,128,128,128,128,128,128,
/* 4 */ 128, 0 ,  1, 10, 11, 12, 13, 14, 15,128, 16, 17, 18, 19, 20,128,
/* 5 */  21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,128,128,128,128,128,
/* 6 */ 128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
/* 7 */ 128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
/* 8 */ 128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
/* 9 */ 128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
/* A */ 128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
/* B */ 128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
/* C */ 128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
/* D */ 128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
/* E */ 128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
/* F */ 128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128
};

struct T_FrSerialNumber {
  char digit[FRSERIALNUMBER_SZ + 1];
  char cksum[FRSNCKSUM_SZ + 1];
  unsigned short cksum_i;
  size_t size;
};


DLLExport int
frSerialNumberNew (
  ZkError err ,
  FrSerialNumber * sn )
{
  (void) module_id_; /* Gets rid of compile time warning */
  zkErrorClear(err);
  if (sn) {
    *sn = (FrSerialNumber) malloc(sizeof(struct T_FrSerialNumber));
    memset(*sn,0,sizeof(struct T_FrSerialNumber));
    return 1;
  }
  zkErrorSet(err,"frSerialNumberNew",ZKERR,ZKMTOKEN,ZKENULLPTR,NULL);
  return 0;
}


/**
 *  Destroy the specified serial number.  The memory which holds the serial
 *  number will be zeroed out and the pointer will be reset to NULL.
 *
 *  @param err Returns the error if any occured
 *  @param sn Pointer to the serial number to destroy
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A zero on failure, a non-zero value on success
 */

DLLExport int
frSerialNumberFree (
  ZkError err ,
  FrSerialNumber * sn )
{
  zkErrorClear(err);
  if (sn) {
    if (*sn) {
      memset(*sn,0,sizeof(struct T_FrSerialNumber));
      free(*sn);
      *sn = NULL;
      return 1;
    }
  }
  zkErrorSet(err,"frSerialNumberFree",ZKERR,ZKMTOKEN,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int 
frSerialNumberComputeChecksum (
  ZkError err ,
  FrSerialNumber sn ,
  unsigned short * cksum )
{
  unsigned char a = 0;
  unsigned char b = 0;
  int i = 0;

  zkErrorClear(err);
  if (sn && cksum) {
    *cksum = 0;
    for (i = 0; i < FRSERIALNUMBER_SZ; i += 2) {
      if (((a = sn_value[(int) sn->digit[i]]) & 128) || 
          ((b = sn_value[(int) sn->digit[i + 1]]) & 128)) {
        zkErrorSet(err,"frSerialNumberComputeChecksum",ZKERR,ZKMTOKEN,
                   ZKEDATABAD,"Serial number contains an illegal character");
        return 0;
      }
      *cksum += a * 32 + b;
    }
    *cksum = (*cksum >> 10) + (*cksum & 0x03FF);
    *cksum += *cksum >> 10;
    *cksum = ~(*cksum) & 0x03FF;
    return 1;
  }
  zkErrorSet(err,"frSerialNumberComputeChecksum",ZKERR,ZKMTOKEN,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
frSerialNumberSetChecksum (
  ZkError err,
  FrSerialNumber sn ,
  unsigned short cksum ) 
{
  zkErrorClear(err);
  if (sn) {
    sn->cksum_i = cksum & 0x03FF; 
    sn->cksum[0] = sn_digit[(sn->cksum_i & 0x03E0) >> 5];
    sn->cksum[1] = sn_digit[sn->cksum_i & 0x001F];
    sn->cksum[2] = (char) 0;
    return 1;
  }
  zkErrorSet(err,"frSerialNumberSetChecksum",ZKERR,ZKMTOKEN,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
frSerialNumberVerifyChecksum (
  ZkError err ,
  FrSerialNumber sn )
{
  unsigned short cksum = 0;
  
  zkErrorClear(err);
  if (sn) {
    if (!frSerialNumberComputeChecksum(err,sn,&cksum)) {
      return 0;
    }
    if (cksum != sn->cksum_i) {
      zkErrorSet(err,"frSerialNumberComputeChecksum",ZKERR,ZKMTOKEN,ZKEDATABAD,
                 "Serial number checksum is incorrect");
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"frSerialNumberVerifyChecksum",ZKERR,ZKMTOKEN,ZKENULLPTR,NULL);
  return 0;
} 
  


/**
 *  Use this function to create a randomly generated serial number and place
 *  it inside the serial number structure.
 *
 */
#ifndef WIN32
#warning  Taking modulo 32 of every byte in random buffer, should be ok, but ask ian
#else
#pragma message(Reminder "Taking modulo 32 of every byte in random buffer, should be ok, but ask ian")
#endif

DLLExport int
frSerialNumberGenerate (
  ZkError err ,
  FrSerialNumber sn )
{
  int i = 0;
  unsigned char buf[FRSERIALNUMBER_SZ];
  unsigned short cksum = 0;

  zkErrorClear(err);
  if (sn) {
    if (!zkRandGetData(err,buf,FRSERIALNUMBER_SZ)) {
      return 0;
    }
    for (i = 0; i < FRSERIALNUMBER_SZ; i++) {
      sn->digit[i] = sn_digit[buf[i]&0x1F];
    }
    if (!frSerialNumberComputeChecksum(err,sn,&cksum)) {
      return 0;
    }
    if (!frSerialNumberSetChecksum(err,sn,cksum)) {
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"frSerialNumberGenerate",ZKERR,ZKMTOKEN,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
frSerialNumberNewFromRandom (
  ZkError err ,
  FrSerialNumber * sn )
{
  zkErrorClear(err);
  if (sn) {
    if (!frSerialNumberNew(err,sn)) {
      return 0;
    }
    if (!frSerialNumberGenerate(err,*sn)) {
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"frSerialNumberNewFromRandom",ZKERR,ZKMTOKEN,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
frSerialNumberGet (
  ZkError err ,
  FrSerialNumber sn ,
  char * s ,
  size_t size )
{
  zkErrorClear(err);
  if (sn && s) {
    if (size < (FRSERIALNUMBER_SZ + FRSNCKSUM_SZ + 1)) {
      zkErrorSet(err,"frSerialNumberGet",ZKERR,ZKMTOKEN,ZKEPARMBAD,
                 "Character string too short to hold serial number");
      return 0;
    }
    memcpy(s,sn->digit,FRSERIALNUMBER_SZ);
    memcpy(&(s[FRSERIALNUMBER_SZ]),sn->cksum,FRSNCKSUM_SZ);
    s[FRSERIALNUMBER_SZ + FRSNCKSUM_SZ] = (char) 0;
    return 1;
  }
  zkErrorSet(err,"frSerialNumberGet",ZKERR,ZKMTOKEN,ZKENULLPTR,NULL);
  return 0;
}


DLLExport int
frSerialNumberParse (
  ZkError err ,
  FrSerialNumber sn ,
  char * s, 
  size_t size )
{
  size_t i = 0;
  int tail = 0;

  zkErrorClear(err);
  if (sn && s) {
    for (i = 0; i < size; i++) {
      if (isspace(s[i]) || (s[i] == '-')) {
        continue;
      }
      if (sn_value[(int) s[i]] & 128) {
        zkErrorSet(err,"frSerialNumberParse",ZKERR,ZKMTOKEN,ZKEDATABAD,
                   "Serial number contains an illegal character");
        return 0;
      }
      if (tail < FRSERIALNUMBER_SZ) {
        sn->digit[tail] = s[i];
      } else if ((tail - FRSERIALNUMBER_SZ) < FRSNCKSUM_SZ) {
        sn->cksum[tail - FRSERIALNUMBER_SZ] = s[i];
      } else {
        zkErrorSet(err,"frSerialNumberParse",ZKERR,ZKMTOKEN,ZKEDATABAD,
                   "Serial number is too long");
        return 0;
      }
      tail++;
    }
    if (tail != (FRSERIALNUMBER_SZ + FRSNCKSUM_SZ)) {
      zkErrorSet(err,"frSerialNumberParse",ZKERR,ZKMTOKEN,ZKEDATABAD,
                 "Serial number is too short");
      return 0;
    }
    sn->cksum_i = sn_value[(int)sn->cksum[0]]*32 + sn_value[(int)sn->cksum[1]];
    sn->cksum[FRSNCKSUM_SZ] = (char) 0;
    sn->digit[FRSERIALNUMBER_SZ] = (char) 0;
    if (!frSerialNumberVerifyChecksum(err,sn)) {
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"frSerialNumberParse",ZKERR,ZKMTOKEN,ZKENULLPTR,NULL);
  return 0;
}

  
  
