/*
 * $Id: frTokenCache.c,v 1.5.32.1 2000/10/11 13:19:26 phil Exp $
 */

static const char * module_id_ = "$Id: frTokenCache.c,v 1.5.32.1 2000/10/11 13:19:26 phil Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* !WIN32 */
#include <string.h>
#include <errno.h>
#include <zkError.h>
#include <zkFile.h>
#include <frTokenCache.h>


#define CURRENT_TOKEN_CACHE_VERSION 1

#define DEFAULT_TSZ  64  /* Default size of the token array (can increase) */

struct T_FrTokenCache {
  uint32_t flags;
  size_t tokens;          /* Number of tokens in the cache */
  size_t tsz;             /* Size of the token array */
  size_t tail;            /* Next available entry */
  char ** token;          /* The token array */
  uint32_t * tok_flags;   /* Set of flags for each token in the cache */
};

/*
 * 
 */

DLLExport int
frTokenCacheNew (
  ZkError err ,
  FrTokenCache * tk )
{
  const char * func = "frFrTokenCacheNew";

  (void) module_id_; /* Gets rid of compile warning */
  zkErrorClear(err);
  if (tk) {
    *tk = (FrTokenCache) malloc(sizeof(struct T_FrTokenCache));
    if ((*tk) == NULL) {
      zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENOMEM,"malloc");
      return 0;
    }
    memset(*tk,0,sizeof(struct T_FrTokenCache));
    (*tk)->flags = 0;
    (*tk)->tokens = 0;
    (*tk)->tsz = DEFAULT_TSZ;
    (*tk)->tail = 0;
    (*tk)->token = (char **) malloc(sizeof(char *) * (*tk)->tsz);
    (*tk)->tok_flags = (uint32_t *) malloc(sizeof(uint32_t) * (*tk)->tsz);
    if (((*tk)->token == NULL) || ((*tk)->tok_flags == NULL)) {
      zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENOMEM,"malloc");
      free(*tk);
      *tk = NULL;
      return 0;
    }
    memset((*tk)->token,0,sizeof(char *) * (*tk)->tsz);
    memset((*tk)->tok_flags,0,sizeof(uint32_t) * (*tk)->tsz);
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */

DLLExport int
frTokenCacheFree (
  ZkError err ,
  FrTokenCache * tk )
{
  const char * func = "frTokenCacheFree";
  char * s = NULL; 
  size_t n = 0; 
 
  zkErrorClear(err);
  if (tk) {
    if (*tk) {
      if ((*tk)->token) {
        for (n = 0; n < (*tk)->tsz; n++) {
          if ((*tk)->token[n]) {
            s = (*tk)->token[n];
            while (*s) *s++ = (char) 0;
            free((*tk)->token[n]);
          } 
        }
        free((*tk)->token);
      }
      if ((*tk)->tok_flags) {
        free((*tk)->tok_flags);
      }
      memset(*tk,0,sizeof(struct T_FrTokenCache));
      free(*tk);
      *tk = NULL;
      return 1;
    }
  }
  zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */

DLLExport int
frTokenCacheFromBuffer (
  ZkError err ,
  FrTokenCache * tk ,
  void * buf ,
  size_t bsz )
{
  const char * func = "frTokenCacheFromBuffer";
  unsigned short tlen = 0;
  unsigned short version = 0; /* version of the token cache buffer */
  size_t base = 0;
  uint32_t tok_flags = 0;
  unsigned char * b = (unsigned char *) buf;
  size_t bufsz = 0; 
  size_t n = 0;
  int ndx = 0;
  size_t tokens = 0;

  zkErrorClear(err);
  if (b && tk) {
    if (bsz < 6) {
      zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKEINCOMPLETE,
                 "Buffer is too short");
      return 0;
    }
 
    /* Do we have a version code, if so read it */
 
    if (b[0] == 255) {
      version = b[1] * 0x100 + b[2];
      base = 3;
      if (version != CURRENT_TOKEN_CACHE_VERSION) {
        zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKERROR,
                   "Incorrect token cache version");
      } 
    }

    /* Compute the buffer size of the token cache to make sure we have all 
     * the data in the buffer.
     */

    bufsz = (b[base]*0x1000000) + (b[base + 1]*0x10000) + 
             (b[base + 2]*0x100) + b[base + 3]; 
    if (bufsz != bsz) {
      zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKEDATABAD,
                 "Buffer size mismatch");
      return 0;
    }
    if (!frTokenCacheNew(err,tk)) {
      return 0;
    }
    tokens = (b[base + 4]*0x100) + b[base + 5];
    base += 6;
    while (base < (bsz - 3)) {
      n++;  /* To keep track of how many tokens added */
      tlen = (b[base]*0x100) + b[base+1];

      /*
       * If tlen is 0, then we've got a set of flags for the token.  Read
       * the flags in then read the token.
       */

      if (tlen == 0) {
        tok_flags = (b[base + 2] * 0x1000000) + (b[base + 3] * 0x10000)
                    + (b[base + 4] * 0x100) + b[base + 5];
        base += 6;
        tlen = (b[base]*0x100) + b[base+1];
      } else {
        tok_flags = 0;
      }

      if ((tlen + base + 3) <= bsz) {
        if (b[tlen + base + 2] == (char) 0) {
          if (!frTokenCacheAddToken(err,*tk,(const char *)&(b[base + 2]),&ndx)) {
            return 0;
          }
          if (!frTokenCacheSetTokenFlags(err,*tk,ndx,tok_flags)) {
            return 0;
          }
        } else {
          zkErrorSet_va(err,func,ZKERR,ZKMTOKENCACHE,ZKEDATABAD,
                        "Token %d is not NUL terminated",n); 
          return 0;
        }
      } else {
        zkErrorSet_va(err,func,ZKERR,ZKMTOKENCACHE,ZKEDATABAD,
                     "Token %d extends beyond buffer",n);
        return 0;
      }
      base += tlen + 3;
    }
    if (n != tokens) {
      zkErrorSet_va(err,func,ZKERR,ZKMTOKENCACHE,ZKEDATABAD,
                    "Token count mismatch.  Expected %d, got %d",tokens,n);
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,func,ZKENULLPTR,ZKMTOKENCACHE,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 *
 * Buffer format:
 * byte 0
 *      1
 *      2
 *      3   Total size of the buffer, most significant byte first
 *      4
 *      5   Number of tokens, most significant byte first
 *      6... Token data
 *      Each piece of token data is  2 bytes for the length (msb first)
 *      followed by the token string and a NUL.  The NUL is not counted in
 *      the length.
 */

DLLExport int
frTokenCacheToBuffer (
  ZkError err ,
  FrTokenCache tk ,
  void ** b ,
  size_t * bsz )
{
  const char * func = "frTokenCacheToBuffer";
  unsigned char * buf = NULL;
  unsigned short tlen = 0;
  int base = 0;
  size_t sz = 0;
  size_t n = 0;

  zkErrorClear(err);
  if (tk && b && bsz) {
    sz = 9; /* For size of buffer + number of tokens + version ID */
    for (n = 0; n < tk->tsz; n++) {
      if (tk->token[n]) {
        sz += strlen(tk->token[n]) + 3;
      } 
      /* Only store the token flags if they are non-zero */
      if (tk->tok_flags[n]) {  
        sz += 6;
      }
    }
    if ((buf = (unsigned char *) malloc(sz)) == NULL) {
      zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENOMEM,"malloc");
      return 0;
    }
    buf[0] = 255;  /* First bytes is always 255 */
    buf[1] = (CURRENT_TOKEN_CACHE_VERSION & 0xFF00) / 0x100;
    buf[2] = (CURRENT_TOKEN_CACHE_VERSION & 0xFF);
    buf[3] = (sz & 0xFF000000) / 0x1000000;
    buf[4] = (sz & 0x00FF0000) / 0x10000;
    buf[5] = (sz & 0x0000FF00) / 0x100;
    buf[6] = (sz & 0x000000FF);
    buf[7] = (tk->tokens & 0xFF00) / 0x100;
    buf[8] = (tk->tokens & 0x00FF);
    base = 9;
    for (n = 0; n < tk->tsz; n++) {
      if (tk->token[n]) {

        /*
         * If this token has flags set for it, save them.  I'm overloading
         * the token size field here by setting it to 0 to indicate the
         * presence of flags (normally, it would be an error to have a size 0
         * token).  yes, yes, I know this is evil and I should have put a 
         * version number in right from the beginning.
         */

        if (tk->tok_flags[n]) {
          buf[base++] = 0;
          buf[base++] = 0;
          buf[base++] = (tk->tok_flags[n] & 0xFF000000) / 0x1000000;
          buf[base++] = (tk->tok_flags[n] & 0x00FF0000) / 0x10000;
          buf[base++] = (tk->tok_flags[n] & 0x0000FF00) / 0x100;
          buf[base++] = (tk->tok_flags[n] & 0x000000FF);
        }

        /* Store the token and its size */
        tlen = (unsigned short) strlen(tk->token[n]);
        buf[base++] = (tlen & 0xFF00) / 0x100;
        buf[base++] = (tlen & 0x00FF);
        memcpy(&(buf[base]),tk->token[n],tlen);
        base += tlen;
        buf[base++] = (char) 0;
      }
    }
    *b = buf;
    *bsz = sz;
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */

DLLExport int
frTokenCacheLoadFromFile (
  ZkError err ,
  FrTokenCache * tk ,
  const char * fn )
{
  const char * func = "frTokenCacheLoadFromFile";
  char * b = NULL;
  size_t bsz = 0;
  int return_code = 0;

  zkErrorClear(err);
  if (tk && fn) {
    if (!zkFileLoad(err,fn,&b,&bsz)) {
      return 0;
    }
    if (frTokenCacheFromBuffer(err,tk,b,bsz)) {
      return_code = 1;
    }
    free(b);
    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */

DLLExport int 
frTokenCacheSaveToFile (
  ZkError err ,
  FrTokenCache tk ,
  const char * fn )
{
  const char * func = "frTokenCacheSaveToFile";
  char * b = NULL;
  size_t bsz = 0;
  int return_code = 0;
  
  zkErrorClear(err);
  if (tk && fn) {
    if (!frTokenCacheToBuffer(err,tk,(void **)&b,&bsz)) {
      return 0;
    }
    if (zkFileSave(err,fn,b,bsz,O_CREAT|O_TRUNC,0600)) {
      return_code = 1;
    }
    free(b);
    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */

DLLExport int
frTokenCacheTokens (
  ZkError err ,
  FrTokenCache tk ,
  size_t * tokens )
{
  const char * func = "frTokenCacheTokens";

  zkErrorClear(err);
  if (tk && tokens) {
    *tokens = tk->tokens;
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */

DLLExport int
frTokenCacheAddToken (
  ZkError err ,
  FrTokenCache tk ,
  const char * tok ,
  unsigned int * ndx )
{
  const char * func = "frTokenCacheAddToken";
  char ** token = NULL;
  uint32_t * tok_flags = NULL;
 
  zkErrorClear(err);
  if (tk && tok && ndx) {
    if (tk->tail >= tk->tsz) {
      token = (char **) realloc(tk->token,
                                (size_t)(sizeof(char*)*(tk->tsz+DEFAULT_TSZ)));
      if (token == NULL) {
        zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENOMEM,
                   "realloc for token array");
        return 0;
      }      
      tok_flags = (uint32_t *) realloc(tk->tok_flags,
                             (size_t)(sizeof(uint32_t)*(tk->tsz+DEFAULT_TSZ)));
      if (tok_flags == NULL) {
        free(token);
        zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENOMEM,
                   "realloc for token flag array");
        return 0;
      }
      memset(&(token[tk->tsz]),0,DEFAULT_TSZ * sizeof(char *));
      memset(&(tok_flags[tk->tsz]),0,DEFAULT_TSZ * sizeof(uint32_t));
      tk->tsz += DEFAULT_TSZ;
      tk->token = token;
      tk->tok_flags = tok_flags;
    }
    if ((tk->token[tk->tail] = strdup(tok)) == NULL) {
      zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENOMEM,"strdup token");
      return 0;
    }
    *ndx = tk->tail;
    tk->tok_flags[tk->tail] = 0;   /* There are no default flags */
    tk->tail += 1;
    tk->tokens += 1;
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */


DLLExport int
frTokenCacheGetToken (
  ZkError err ,
  FrTokenCache tk ,
  unsigned int ndx ,
  const char ** token )
{
  const char * func = "frTokenCacheGetToken";

  zkErrorClear(err);
  if (tk && token) {
    if (ndx < tk->tail) {
      if (tk->token[ndx]) {
        *token = tk->token[ndx];
        return 1;
      }
    }
    *token = NULL;
    zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENOTFOUND,"token not found"); 
    return 0;
  }
  zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENULLPTR,NULL);
  return 0;
} 

/*
 *
 */

DLLExport int
frTokenCacheGetTokenFlags (
  ZkError err ,
  FrTokenCache tk ,
  unsigned int ndx ,
  uint32_t * flags )
{
  static const char * func = "frTokenCacheGetTokenFlags";
  int ret = 0;

  if (!tk) {
    zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENULLPTR, 
               "Token cache pointer is NULL");
  } else if (!flags) {
    zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENULLPTR,
               "flag pointer is NULL");
  } else {
    if (ndx < tk->tail) {
      if (tk->token[ndx]) {
        *flags = tk->tok_flags[ndx];
        ret = 1;
      }
    }
    if (!ret) {
      zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENOTFOUND,"token not found");
    }
  }

  return ret;
}

/*
 *
 */

DLLExport int
frTokenCacheSetTokenFlags (
  ZkError err ,
  FrTokenCache tk ,
  unsigned int ndx ,
  uint32_t flags )
{
  static const char * func = "frTokenCacheSetTokenFlags";
  int ret = 0;

  if (!tk) {
    zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENULLPTR,
               "Token cache pointer is NULL");
    ret = 0;
  } else {
    if (ndx < tk->tail) {
      if (tk->token[ndx]) {
        tk->tok_flags[ndx] = flags;
        ret = 1;
      }
    }
    if (!ret) {
      zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENOTFOUND,"token not found");
    }
  }

  return ret;
}

/*
 * 
 */


DLLExport int
frTokenCacheDeleteToken (
  ZkError err ,
  FrTokenCache tk ,
  unsigned int ndx )
{
  const char * func = "frTokenCacheDeleteToken";
  char * s = NULL;

  zkErrorClear(err);
  if (tk) {
    if (ndx < tk->tail) {
      if (tk->token[ndx]) {
        s = tk->token[ndx];
        while (*s) *s++ = (char) 0;
        free(tk->token[ndx]);
        tk->tokens -= 1;

        while (ndx < tk->tail) {
          tk->token[ndx] = tk->token[ndx + 1];
          tk->tok_flags[ndx] = tk->tok_flags[ndx + 1];
          ndx++;
        }
        tk->tail -= 1;
        return 1;
      }
    }
    zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENOTFOUND,"token not found"); 
    return 0;
  }
  zkErrorSet(err,func,ZKERR,ZKMTOKENCACHE,ZKENULLPTR,NULL);
  return 0;
}
  
  

