/*
 * $Id: frTokenClient.c,v 1.12.32.6 2001/01/15 23:38:12 corey Exp $
 */

static const char * module_id_ = "$Id: frTokenClient.c,v 1.12.32.6 2001/01/15 23:38:12 corey Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* !WIN32 */
#include <string.h>
#include <zkError.h>
#include "zkCStr.h"
#include "zkFile.h"
#include <zkBase64.h>
#include <ZKIP.h>
#include <zkipClient.h>
#include <frSerial.h>
#include <frToken.h>
#include <frTokenClient.h>
#include "frCrypt.h"
#include "frSec.h"

/*
 * Some string sizes
 */

#define CMDSTR_SZ 128 /* Size of the ZKIP command string */
#define SNSTR_SZ 32 /* Size of buffer to hold a serial number string */

/*
 * standard quit commmand for ZKIP 
 */

#define QUIT_COMMAND "1 0.1.0 quit nodata none 0\015\012"

static int getConnectedTokenServer(ZkError err, TokenServer *tokenserv, ZkipClient* pClient);

static int getConnectedTokenServer(ZkError err, TokenServer *tokenserv, ZkipClient* pClient)
{
  int ret = 0;
  int fd = 0;

  if (tokenserv
	  && pClient
      && zkipClientNew(err, pClient)
	  && zkipClientMakeBlocking(err, *pClient, tokenserv->timeout)
	  && zkipClientConnectToHost(err, *pClient, tokenserv->tokenserver, tokenserv->port)
	  && zkipClientFd(err, *pClient, &fd))
  {
    ret = 1;
  }
  else if (pClient && *pClient)
  {
    zkipClientDisconnect(NULL, *pClient);
    zkipClientFree(NULL, pClient);
  }
  return ret;
}

/*
 * Get the capability code string from the server for a specific serial number
 * and verify the signatures on it.
 */

DLLExport int
frSerialNumberGetCapability (
  ZK_INOUT ZkError err ,
  ZK_IN    FrKeyQryHost kh ,
  ZK_INOUT FrKeyQry kq ,
  ZK_IN    FrSerialNumber sn ,
  ZK_IN    TokenServer * tokenserver ,
  ZK_OUT   TokenServerResponse * tsresp ,
  ZK_OUT   char * cap ,
  ZK_INOUT size_t * cap_sz )
{
  ZkipClient client = NULL;
  int ret = 0;

  zkErrorClear(err);

  if (tokenserver && getConnectedTokenServer(err, tokenserver, &client))
  {
	ret = frSerialNumberGetCapabilityConnected(err, kh, kq, sn, client, tsresp, cap, cap_sz);
    zkipClientDisconnect(NULL,client);
	zkipClientFree(NULL,&client);
  }
  return ret;
}

DLLExport int
frSerialNumberGetCapabilityConnected (
  ZK_INOUT ZkError err ,
  ZK_IN    FrKeyQryHost kh ,
  ZK_INOUT FrKeyQry kq ,
  ZK_IN    FrSerialNumber sn ,
  ZK_INOUT ZkipClient client,
  ZK_OUT   TokenServerResponse * tsresp ,
  ZK_OUT   char * cap ,
  ZK_INOUT size_t * cap_sz )
{
  static const char * func = "frSerialNumberGetCapability";
  int ret = 0;                       /* Function return code */
  char * scap = NULL;                /* Signed capability buffer */
  size_t scap_sz = 0;                /* Size of scap buffer */
  FrCryptSig sigkey = NULL;          /* Token server public sig key */
  FrEntIdExp signer;        /* Should be the token server */
  unsigned int kv = 0;      /* Key version of the token server signature */
  int valid = 0;            /* Is the key valid? */

#ifndef WIN32
#warning hardwired key timeout and retry parameters
#endif
  int timeout = 3;          /* Some hardwired key lookup defaults */
  int retries = 5; 

  /*
   * Parameter check 
   */

  if (!kq) {
    zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,"kq is NULL");
  } else if (!sn) {
    zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,"sn is NULL");
  } else if (!client) {
    zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,"client is NULL");
  } else if (!tsresp) {
    zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,"tsresp is NULL");
  } else if (!cap) {
    zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,"cap is NULL");
  } else if (!cap_sz) {
    zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,"cap_sz is NULL");

  /*
   * Get the signed capability code from the buffer 
   */

	} else if (!frSerialNumberCapabilityConnected(err,sn,client,
                                       tsresp,&scap,&scap_sz)) {
    ret = 0;
  } else {

    /*
     * Verify the signature on the buffer
     */

    if (!(frSecSigGetEntity(err,scap,scap_sz,&signer) &&
          frSecSigGetVersion(err,scap,scap_sz,&kv) &&
          (signer.type == FRENT_TYPE_TOK_SRV) &&
          frKeyQryFetchPubKey(err,kq,kv,FRCRYPT_KEYTYPE_SIGNING,
                              signer,kh,retries,timeout,&sigkey) &&
          frSecSigVerify(err,sigkey,scap,scap_sz,&valid) &&
          (valid))) {
      tsresp->code = 0;
      snprintf(tsresp->message,TSRMESG_SZ,"Response signature check failed");

    /*
     * Remove it
     */

    } else if (!frSecSigRemoveInSitu(err,scap,&scap_sz)) {
      tsresp->code = 0;
      snprintf(tsresp->message,TSRMESG_SZ,"Failed to remove signature");
    
    /*
     * Make sure we got enough space to copy it
     */

    } else if (scap_sz >=  *cap_sz) {
      zkErrorSet_va(err,func,ZKERR,ZKMTOKEN,ZKERROR,
                    "cability code buffer is too small (need %d bytes)",
                    scap_sz + 1);
      tsresp->code = 0;
      snprintf(tsresp->message,TSRMESG_SZ,
               "Failed to copy capability code buffer");
   
    /*
     * do the copy
     */

    } else {
      memcpy(cap,scap,scap_sz);
      cap[scap_sz] = (char) 0;
      *cap_sz = scap_sz;
      ret = 1;
    }
  }
  
  /*
   * Cleanup
   */

  if (scap) {
    free(scap);
  }

  return ret;
}

/*
 * Get the signed capability code from the token server.
 */


DLLExport int
frSerialNumberCapabilityConnected (
  ZK_INOUT ZkError err ,
  ZK_IN    FrSerialNumber sn ,
  ZK_INOUT ZkipClient client ,
  ZK_OUT   TokenServerResponse * tsresp ,
  ZK_OUT   char ** cap ,
  ZK_INOUT size_t * cap_sz )
{
  static const char * func = "frSerialNumberCapability";
  int ret = 0;           /* Function return code */
  struct ZkipCommandResponse * response = NULL;  /* Server response */
  char cmdstr[CMDSTR_SZ]; /* The ZKIP command string for checking sn */
  char * extra_data = NULL; /* Holds the signed capability code */
  char sns[SNSTR_SZ] ; /* String representation of the serial number */

  /*
   * Parameter check 
   */

  if (!sn) {
    zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,"sn is NULL");
  } else if (!client) {
    zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,"client is NULL");
  } else if (!tsresp) {
    zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,"tsresp is NULL");
  } else if (!cap) {
    zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,"cap is NULL");

  /*
   * Convert the serial number to a string token
   */

  } else if (!frSerialNumberGet(err,sn,sns,SNSTR_SZ)) {
    tsresp->code = 0;
    snprintf(tsresp->message,TSRMESG_SZ,
             "Failed to convert serial number to string");

  } else if (!zkipClientReadResponse(err,client,&response,&extra_data)) {
    tsresp->code = 0;
    snprintf(tsresp->message,TSRMESG_SZ,"Failed to read response from token server");

  /*
   * Make sure that the welcome banner said the server is ready
   */

  } else if (response->response != _SERVICE_READY) {
    tsresp->code = response->response;
    snprintf(tsresp->message,TSRMESG_SZ,"%s",response->message);

  /*
   * Create the capability command and send it.
   */

  } else if (snprintf(cmdstr,CMDSTR_SZ,
                      "1 5.1.0 capability SERIALNUMBER BINARY 0 %s\015\012",
                      sns) < 0) {
    tsresp->code = 0;
    zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKERROR,
               "Unable to create command string");
  } else if (!zkipClientWrite(err,client,cmdstr,strlen(cmdstr))) {
    snprintf(tsresp->message,TSRMESG_SZ,"Failed to write data to token server");
    tsresp->code = 0;

  /*
   * Read the response
   */

  } else if (!(zkipClientReadResponse(err,client,&response,&extra_data))) {
    tsresp->code = 0;
    snprintf(tsresp->message,TSRMESG_SZ,
             "Failed to read data from token server");
  } else if (response->response != _SERVICE_READY) {
    tsresp->code = response->response;
    snprintf(tsresp->message,TSRMESG_SZ,"%s",response->message);
 
  /*
   * We got a capability code.  Save the information and send it back up
   */

  } else if ((*cap = malloc(response->datalen * sizeof(char))) == NULL) {
    zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENOMEM,"Failed to allocate memory");
  } else {
    memcpy(*cap,extra_data,response->datalen);
    *cap_sz = (size_t) response->datalen;
    zkipClientWrite(NULL,client,QUIT_COMMAND,strlen(QUIT_COMMAND));
    zkipClientReadResponse(NULL,client,&response,&extra_data);
    ret = 1;
  } 

  /*
   * cleanup
   */

  if (!ret) {
    if (cap) {
      *cap = NULL;
    }
  }
    
  return ret;
}

  
DLLExport int
frSerialNumberRedeem (
  ZkError err ,
  FrSerialNumber sn ,
  TokenServer * tokenserver ,
  TokenServerResponse * tsresp ,
  char *** tokarr ,
  size_t * tokens )
{
  ZkipClient client = NULL;
  int ret = 0;

  zkErrorClear(err);

  if (tokenserver && getConnectedTokenServer(err, tokenserver, &client))
  {
	ret = frSerialNumberRedeemConnected(err, sn, client, tokenserver->sigkey, tsresp, tokarr, tokens);
    zkipClientDisconnect(NULL,client);
	zkipClientFree(NULL,&client);
  }
  return ret;
}

DLLExport int
frSerialNumberRedeemConnected (
  ZkError err ,
  FrSerialNumber sn ,
  ZkipClient client ,
  FrCryptSig tokenServSigKey ,
  TokenServerResponse * tsresp ,
  char *** tokarr ,
  size_t * tokens )
{
  int return_code = 0;
  char cmdstr[CMDSTR_SZ]; /* The ZKIP command string for redeeming sn */
  char sns[SNSTR_SZ] ; /* String representation of the serial number */
  int sns_sz = 0; /* Size of the serial number string */
  struct ZkipCommandResponse * response = NULL;
  FrToken token = NULL; /* Builds the token for verification */
  char * extra_data = NULL;
  char * extra_data2 = NULL;
  int token_size; 
  size_t i = 0; /* Loop index for building tokens */
  
  (void) module_id_; /* Gets rid of compile time warning */
  zkErrorClear(err);
  if (sn && tsresp && tokarr && tokens && client) {

    *tokarr = NULL;
    tsresp->code = 0;
    zkCStrlcpy(tsresp->message,"Did not connect to token server",sizeof tsresp->message);


    if (!frSerialNumberGet(err,sn,sns,SNSTR_SZ)) {
      tsresp->code = 0;
      zkCStrlcpy(tsresp->message,"Unable to get serial number",sizeof tsresp->message);
      goto finish;
    }
    sns_sz = strlen(sns);

	if (!zkipClientReadResponse(err,client,&response,&extra_data))
	{
      zkCStrlcpy(tsresp->message,"Failed to read response from token server",sizeof tsresp->message);
      goto finish;
    }
  
    /*
     * Did it tell us it is ready for processing a serial number?
     */

    if (response->response != _SERVICE_READY) {
      tsresp->code = response->response;
      strncpy(tsresp->message,response->message,TSRMESG_SZ);
      tsresp->message[TSRMESG_SZ] = (char) 0;
      zkErrorSet(err,"frSerialNumberRedeem",ZKERR,ZKMTOKEN,ZKERROR,
                 response->message);
      goto finish;
    }

    /*
     * Create and send the command to redeem a serial number
     */

    snprintf(cmdstr,sizeof cmdstr,"1 5.1.0 generate SERIALNUMBER BINARY %d\015\012",sns_sz);
    if (!zkipClientWrite(err,client,cmdstr,strlen(cmdstr))) {
      zkCStrlcpy(tsresp->message,"Failed to write data to token server",sizeof tsresp->message);
      tsresp->code = 0;
      goto finish;
    }

    /*
     * Read the response and write the serial number if the server has
     * accepted the command.
     */

    tsresp->code = 0;
    if (!zkipClientReadResponse(err,client,&response,&extra_data)) {
      zkCStrlcpy(tsresp->message,"Failed to read data from token server",sizeof tsresp->message);
      goto finish;
    } else if (response->response != _DATA_STARTINPUT) { /* Cant send the sn */
      tsresp->code = response->response;
      strncpy(tsresp->message,response->message,TSRMESG_SZ);
      goto finish;
    } else if (!(zkipClientWrite(err,client,sns,sns_sz) &&
                 zkipClientWrite(err,client,"\015\012.\015\012",5))) {
      zkCStrlcpy(tsresp->message,"Failed to write data to token server",sizeof tsresp->message);
      goto finish;
    }

    /* 
     * Read the response
     */

    if (!(zkipClientReadResponse(err,client,&response,&extra_data))) {
      zkCStrlcpy(tsresp->message,"Failed to read data from token server",sizeof tsresp->message);
      goto finish;
    } else if (response->response != _SERVICE_READY) {
      tsresp->code = response->response;
      strncpy(tsresp->message,response->message,TSRMESG_SZ);
      goto finish;
    }   
   
    /*
     * If we get to this point, we have a set of valid tokens inside the
     * buffered which is referenced by "extra_data"  The first byte must
     * be the number of tokens, the rest are the tokens themselves (these are
     * all supposed to be the same size)
     */
   
    tsresp->code = 0;
    if (response->datalen < 1) {
      zkErrorSet(err,"frSerialNumberRedeem",ZKERR,ZKMTOKEN,ZKEDATABAD,
                 "No tokens generated");
      zkCStrlcpy(tsresp->message,"No tokens generated for serial number",sizeof tsresp->message);
      goto finish;
    }
    *tokens = (size_t) extra_data[0];
    token_size = (response->datalen - 1) / *tokens;
    if (token_size < 1) {
      zkErrorSet(err,"frSerialNumberRedeem",ZKERR,ZKMTOKEN,ZKEDATABAD,
                 "Tokens are incorrect size");
      zkCStrlcpy(tsresp->message,"No tokens generated for serial number",sizeof tsresp->message);
      goto finish;
    } 
    *tokarr = (char **) malloc(sizeof(char *) * (*tokens + 1));
    if (*tokarr == NULL) {
      zkErrorSet(err,"frSerialNumberRedeem",ZKERR,ZKMTOKEN,ZKENOMEM,"malloc");
      zkCStrlcpy(tsresp->message,"Unable to create token list",sizeof tsresp->message);
      goto finish;
    }
   


    /*
     * Turn the tokens to base64 string
     */

    for (i = 0; i < *tokens; i++) {
      if (!frTokenNewFromBinary(err,&token,
                                (unsigned char *)&(extra_data[i*token_size+1]),
                                token_size, tokenServSigKey)) {
        zkCStrlcpy(tsresp->message,"Received token data was invalid",sizeof tsresp->message);
        frTokenFree(err,&token);
        goto finish;
      }
      frTokenFree(err,&token);
      (*tokarr)[i] = (char *)malloc(sizeof(char) * ((token_size * 4)/3 + 5));
      if ((*tokarr)[i] == NULL) {
        zkCStrlcpy(tsresp->message,"Unable to allocate token string",sizeof tsresp->message);
        zkErrorSet(err,"frSerialNumberRedeem",ZKERR,ZKMTOKEN,ZKENOMEM,"malloc");
        goto finish;
      }
      if (!zkConvertToBase64(err,(unsigned char *)&(extra_data[i*token_size+1]),
                             (unsigned char *)(*tokarr)[i],token_size)) {
        zkCStrlcpy(tsresp->message,"Unable to convert token to base64 string",sizeof tsresp->message);
        goto finish;
      }
    }
    (*tokarr)[*tokens] = NULL;
    
    /*
     * SUCCESS DOOD!
     */

    return_code = 1; 

    /*
     * Issue the quit command and get the reponse.  Note that there is
     * no error structure passed because I dont care if this fails or
     * not.  It's mostly being done to be polite to the remote server
     */

finish:
    zkipClientWrite(NULL,client,QUIT_COMMAND,strlen(QUIT_COMMAND));
    zkipClientReadResponse(NULL,client,&response,&extra_data2);


    if (return_code == 0) {
      if (*tokarr) {
        for (i = 0; i < *tokens; i++) {
          if ((*tokarr)[i]) {
            free((*tokarr)[i]);
          }
        }
        free(*tokarr);
        *tokarr = NULL;
      }
    }
    return return_code;
  }
  zkErrorSet(err,"frSerialNumberRedeem",ZKERR,ZKMTOKEN,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
frTokenLoadTypeDescriptions (
  ZkError err ,
  FrTokenType * tt ,
  NymServer * server ,
  NymServerResponse * nsresp )
{
  const char * func = "frTokenLoadTypeDescriptions";
  int return_code = 0;
  ZkipClient client = NULL;
  struct ZkipCommandResponse * response = NULL;
  char cmdstr[CMDSTR_SZ]; /* The ZKIP command string buffer */
  char * extra_data = NULL;
  int fd = 0;

  zkErrorClear(err);
  if (tt && server && nsresp) {
    *tt = NULL;
    if (!(zkipClientNew(err,&client) &&
          zkipClientMakeBlocking(err,client,server->timeout) &&
          zkipClientConnectToHost(err,client,server->nymserver,server->port) &&
          zkipClientFd(err,client,&fd) &&
          zkipClientReadResponse(err,client,&response,&extra_data))) {
      nsresp->code = 0;      
      zkCStrlcpy(nsresp->message,"Failed connection to nymserver",sizeof nsresp->message);
      goto finish;
    }
  
    /*
     * Is the server ready?
     */

    if (response->response != _SERVICE_READY) {
      nsresp->code = response->response;
      strncpy(nsresp->message,response->message,NSRMESG_SZ);
      nsresp->message[NSRMESG_SZ] = (char) 0;
      zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKERROR,
                 response->message);
      goto finish;
    }

    /*
     * Create and send the command to get the token description
     */

    snprintf(cmdstr,sizeof cmdstr,"1 1.1.0 tokendescription TOKENDESCR BINARY 0\015\012");
    if (!zkipClientWrite(err,client,cmdstr,strlen(cmdstr))) {
      zkCStrlcpy(nsresp->message,"Failed to write data to nymserver",sizeof nsresp->message);
      nsresp->code = 0;
      goto finish;
    }

    /* 
     * Read the response
     */

    if (!(zkipClientReadResponse(err,client,&response,&extra_data))) {
      zkCStrlcpy(nsresp->message,"Failed to read data from nymserver",sizeof nsresp->message);
      goto finish;
    } else if (response->response != _SERVICE_READY) {
      nsresp->code = response->response;
      strncpy(nsresp->message,response->message,TSRMESG_SZ);
      goto finish;
    }   
   
    /*
     * Extra data now contains the token description stuff
     */

    if (!frTokenTypeFromBinary(err,tt,extra_data,response->datalen)) {
      nsresp->code = 0;
      zkCStrlcpy(nsresp->message,"Unable to reconstruct token type descriptions",sizeof nsresp->message);
      goto finish;
    } 
    return_code = 1;

finish:
#if 0
    if (extra_data) {
      free(extra_data);
      extra_data = NULL;
    }
#endif
    zkipClientWrite(NULL,client,QUIT_COMMAND,strlen(QUIT_COMMAND));
    zkipClientReadResponse(NULL,client,&response,&extra_data);

    if (client) {
      zkipClientDisconnect(NULL,client);
      zkipClientFree(NULL,&client);
    }

    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,NULL);
  return 0;
}
