/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef ZKFILE__H_
#define ZKFILE__H_


#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include "zkError.h"
#include "zkClock.h"

#include "zkfile_replace.h"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

  typedef enum { 
    ZKFILE_UNUSED_,
    ZKFILE_HASH_NOHASH_NOFN,    /* No hash value, no hash function */
    ZKFILE_HASH_VALID,          /* Valid hash value */
    ZKFILE_HASH_INVALID,        /* Invalid hash value */
    ZKFILE_HASH_NOHASH,         /* Have a hash fn, but no hash value stored */
    ZKFILE_HASH_FN_MISSING,     /* Missing hash fn, but have a hash */
    ZKFILE_EOF                  /* EOF before reading record */
  } ZkFileHashStatus;

  /* Hash function */
  typedef int ZkFileHashFn(ZkError err, void *buffer, uint32_t bsz, void *hash, uint8_t *hsz, void *userData);

  extern DLLExport int zkFileLengthFp(ZkError err, FILE *fp, size_t *len);
  extern DLLExport FILE *zkNullFp(void);
  extern DLLExport int zkFileFopenLocked(ZkError err, const char *path, const char *mode, FILE **fp);
  extern DLLExport int zkFileGetLineFp(ZkError err, FILE *fp, char **ln, size_t *l);
  extern DLLExport int zkFileReadFp(ZkError err, FILE *fp, size_t max, void **buf, size_t *bsz);
  DLLExport int zkFileWriteFp(ZkError err, FILE *fp, void *data, size_t dsz);

DLLExport int zkFileLoad (
  ZkError ,               /* Returns the error if any occured */
  const char * ,          /* Filename to load into the buffer */
  char ** ,               /* Returns a newly allocated buffer with file data */
  size_t * );             /* Returns the number of bytes in the buffer */

/*#ifndef WIN32*/
DLLExport int zkFileSave (
  ZkError ,               /* Returns the error if any occured */
  const char * ,          /* Filename to save from the buffer */
  const char * ,          /* The buffer */
  size_t ,                /* Returns the number of bytes in the buffer */
  int,
  mode_t );               /* Some flags to consider */
/*#endif*/

DLLExport extern int zkFileWritePid(ZkError err, 
                                    const char *pidFile, /* PID file name */
                                    FILE **fptr); /* 
                                                   * FILE ptr to the open
                                                   * locked PID file.
                                                   */ 

  /*
   *  A simple wrapper for fprintf().
   */
  extern DLLExport int zkFprintf(ZkError err, FILE *fp, const char *fmt, ...);
  extern DLLExport int zkFileTempFp(ZkError err, char *tmpl, FILE **tempFp);

  extern DLLExport
  int zkFileRecordData(ZkError err, 
                       const char *rfile, /* File to append */
                       void *data, /* Data to write to file */
                       size_t dsz, /* Size of the data */
                       ZkFileHashFn hashFn, /* Callback to compute hash
                                               over data, could be NULL */
                       void *userData); /* Extra user data for hashFn */

  extern DLLExport 
  int zkFileRecordGetX(ZkError err,
                       const char *rfile, /* File to read */
                       void **data, /* Data from the file */
                       size_t *dsz, /* Size of the data */
                       size_t recordno, /* Record number in the file */
                       ZkFileHashStatus *status, /* hash status */
                       ZkFileHashFn hashFn, /* Callback to compute the hash 
                                               over the data, could be NULL */
                       void *userData); /* User data for hashFn */


  /*
   *  The next function reads record in sequence.
   *  hashFn, userData, sec, usec can be NULL.
   */
  extern DLLExport
  int zkFileRecordGetNext(ZkError             err,
                          FILE               *file,
                          void              **data,
                          size_t             *datasz,
                          ZkFileHashStatus   *status,
                          ZkFileHashFn        hashFn,
                          void               *userData,
                          ZkClock            *sec,   /* Second portion of timestamp */
                          int32_t            *usec); /* microsecond portion of timestamp
                                                       (-1 if only second portion is valid) */

struct zkfile_rec_hdr_pk {
  ZkClock time;                 /* 64 bit signed time (second portion) */
  int32_t usec;                 /* microsecond portion of the timestamp
                                   (must be 0 <= usec < 999999. If not,
                                   only second portion is considered valid) */
  uint32_t size;                /* number of bytes of data */
  uint8_t hashsz;               /* Hash length */
};

  typedef struct zkfile_rec_hdr_pk *ZkFileRecHdr;
  
  extern DLLExport 
  int zkFileRecordHeaderSerialIn (ZkError err,
                                  void *inbuf,
                                  size_t insz,
                                  ZkFileRecHdr hdr,
                                  size_t *outsz);
  extern DLLExport
  int zkFileRecordHeaderSerialSize (ZkError err, size_t *shsz);

  extern DLLExport
  int zkFileRecordHeaderSerialOut (ZkError err,
                                   ZkFileRecHdr hdr,
                                   void **outbuf, size_t *outsz);


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* ZKFILE__H_ */









