/*
 * Basic graph datastructures and algorithms
 *
 * $Id: zkGraph.h,v 1.8.32.1 2000/10/11 13:19:59 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zks.net>

 */
#ifndef ZKGRAPH_H_
#define ZKGRAPH_H_

#include <zkError.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

  /* Currently unused */
  typedef struct T_ZkGraph *ZkGraph;

  /* Create a new graph */
  DLLExport extern
  int zkGraphNew(ZkError err,
                 ZkGraph *graph);

  /* Free a graph */
  DLLExport extern
  int zkGraphFree(ZkError err, 
                  ZkGraph *graph);

  /* Get the number of nodes in the graph */
  DLLExport extern
  int zkGraphTotalNodes(ZkError err, 
                        ZkGraph graph, 
                        size_t *numnodes);

  /* Add a new node to the graph.  If it already exists, it will fail */
  DLLExport extern 
  int zkGraphAddNode(ZkError err, 
                     ZkGraph graph, 
                     uint32_t id);

  /* Only add an edge in the s -> d direction */
  DLLExport extern 
  int zkGraphDirectedEdgeAdd(ZkError err, 
                             ZkGraph graph, 
                             uint32_t s, 
                             uint32_t d, 
                             int w);

  /* Add an edge for s -> d and d -> s */
  DLLExport extern
  int zkGraphEdgeAdd(ZkError err, 
                     ZkGraph graph, 
                     uint32_t s, 
                     uint32_t d, 
                     int w);

  /* Check if s -> d */
  DLLExport extern
  int zkGraphNodesDirectedConnected(ZkError err,
                                    ZkGraph graph, 
                                    uint32_t s, /* source node */
                                    uint32_t d, /* destination node */
                                    int *connected);

  /* Check if s -> d and d -> s */
  DLLExport extern
  int zkGraphNodesConnected(ZkError err,
                            ZkGraph graph, 
                            uint32_t s,
                            uint32_t d,
                            int *connected);

  /* Compute a predecessor list in a breadth-first search. */
  DLLExport extern
  int zkGraphBFS(ZkError err,
                 ZkGraph graph, 
                 uint32_t s, 
                 uint32_t **p);

  /* Check if a path exists between s and d using the predecessor list. */
  DLLExport extern
  int zkGraphPathExists(ZkError err, 
                        ZkGraph graph, 
                        uint32_t s,   /* Source node */
                        uint32_t d,   /* Destination node */
                        uint32_t *pi, /* Array of num nodes in graph size */
                        int *yes);    /* Set to 1 if the path exists */

  /* 
   * Construct the path between the node s and d.  Return nodes in an array
   * and the number of nodes that make up the path.
   */ 
  DLLExport extern
  int zkGraphConstructPath(ZkError err,
                           ZkGraph graph,
                           uint32_t s,       /* Source node */
                           uint32_t d,       /* Destination node */
                           uint32_t *pi,     /* Predecessor list */
                           uint32_t **path,  /* Path of the traversal */
                           size_t *psz);     /* Num of nodes in traversal */
#ifndef NDEBUG
  /* Dump out the graph. */
  DLLExport extern 
  int zkGraphDump(ZkError err,
                  ZkGraph graph);
#endif

  /* 
   * Enumerate all paths between node s and d.  Return it in the array
   * paths which must be freed by the calling program.  Each path is 
   * seperated by a 0.  Return the size of the array in np.
   */
  DLLExport extern
  int zkGraphEnumerateAllPaths(ZkError err,
                               ZkGraph graph,
                               uint32_t s, 
                               uint32_t d, 
                               uint32_t **paths, 
                               size_t *np);
  DLLExport extern
  int zkGraphEnumerateXPaths(ZkError err,
                             ZkGraph graph,
                             uint32_t s,
                             uint32_t d, 
                             int xpaths,
                             uint32_t **paths,
                             size_t *np);

  /* Delete the node n and all edges associated with it. */
  DLLExport extern
  int zkGraphDeleteNode(ZkError err,
                        ZkGraph graph, 
                        uint32_t n);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif ZKGRAPH_H_
