/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/*
 * zkSecure.h
 * 
 * Security checking functions.
 */
#ifndef ZKSECURE_H_
#define ZKSECURE_H_

#ifndef _ZKERROR_H
#include <zkError.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

  typedef 
  enum E_ZkSecureParanoidLevel {
    ZKSECURE_LAX, 
    ZKSECURE_SAFE,
    ZKSECURE_PARANOID
  } ZkSecureParanoidLevel;

  /* 
   * Typical paranoid settings.  Change working directory to /, set core
   * size limit to 0, set a good umask (only user can rwx)
   */ 
  extern int zkSecureSetParanoid(ZkError);

  /* 
   * Typical safe settings, not as tough as the paranoid settings, but still
   * cautious.  Change working directory to /, set
   * a okay umask (user and group can rwx)
   */  
  extern int zkSecureSetSafe(ZkError);

  /* Change the core size limit.  Only useful for paranoid level */
  extern int zkSecureCoreSizeLimit(ZkError, ZkSecureParanoidLevel);

  /* Check if I'm running this program as root */
  extern int zkSecureRunAsRoot(void);

  /* Set my umask */
  extern int zkSecureSetUmask(ZkError, ZkSecureParanoidLevel);

  /* 
   * Check to see if the file exists and has the correct permissions.  Full 
   * pathnames for paranoid level.
   */ 
  extern int zkSecureCheckFile(ZkError, const char *, ZkSecureParanoidLevel);

  /* 
   * Check the directory and it's permissions.  Must be a full pathname in 
   * paranoid level. 
   */
  extern int zkSecureCheckDir(ZkError, const char *, ZkSecureParanoidLevel);

  /*
   * Open a file, then check fstat().  The file descriptor is returned.
   */
  extern int zkSecureOpenFile(ZkError, const char *, int *, 
                              ZkSecureParanoidLevel);

  /* 
   * Open a file for writing, then check fstat().  The file stream is 
   * returned. 
   */ 
  extern int zkSecureOpenDebugFileStream(ZkError, const char *, FILE **,
                                         ZkSecureParanoidLevel);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* ZKSECURE_H_ */
