/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#if defined( ZK_LINUX )
#ifndef STATISTIC_H_
#define STATISTIC_H_

#include "zkError.h"
#include "zkClock.h"
#include "zkMutex.h"
#include "zkshm.h"


/*
 * Stats collecting library
 *
 * Daniel Faguy <danielf@zeroknowledge.com>
 * $Id: zkStatistic.h,v 1.3.8.4 2000/12/14 17:54:27 jean Exp $
 */


// ---- Statistic Pool Creation Flags ----

enum
{
  SPF_USE_SHARED_MEM = 0x00000001,
  SPF_PTHREAD_SAFE   = 0x00000002
} ;


#ifdef __cplusplus

#include <string>
#include <hash_map>
#include <list>
#include <hash_set>


//  Welcome to the wonderful world of collecting data


//  Class that you might want some other classes to inherit from.
//  Protection is relative since you can provide a NULL pointer in the
//  constructor.  And guess what: If you do so, you get no protection at all
//  (niet, nada, kaput, ...).
//  However, if you feel you need some protection, feel comfortable to provide
//  the constructor with a RMutex-derived class object.
//
class RProtected
{
  RMutex  *m_pMutex ;

protected:
  RProtected( RMutex *pMutex = NULL ) ;
  virtual ~RProtected( ) ;

public:
  bool  lock( )   const ;
  bool  unlock( ) const ;
} ;


//  Class that you might want some other classes to inherit from.
//

class RStatisticItem : protected RProtected
{
  friend class RStatisticPool ;

  const char  *m_pName ;

protected:
  RStatisticItem( const char *pName, RMutex *pMutex = NULL ) ;
  virtual bool  absorb( RStatisticItem *pItem ) = NULL ;  // See RStatisticPool comment

public:
  virtual bool  reset( ) = NULL ;
  const char   *getName( ) const ;
} ;


//  Here are the few classes that you can put in a statistic pool.  I sincerely
//  wish I had made the whole shabang a little more polymorphic.  I think some
//  feature could be tweaked in that direction (absorb, reset, etc.) You want
//  to disscuss about it? Come see me.
//
//  However, here it is, laying there and waiting for you to use it.
//
//  You cannot create statistic item directly.  You must use the
//  RStatisticPoolDesc class to describe the item to create.
//
//  Once the description of a pool is completed, you use the RStatisticPool
//  class which will create all the items.  Then, for each item, you ask the
//  pool object for its address.
//
//  Then you can use the public member-functions of each item.
//


//  Counters are values meant to count something.  As such it holds a positive
//  value.  This value can only be incremented.
//
class RCounterItem : public RStatisticItem
{
  friend class RStatisticPool ;

  unsigned int  m_value ;

protected:
  RCounterItem( const char *pName, RMutex *pMutex = NULL ) ;
  virtual bool  absorb( RStatisticItem *pItem ) ; // Will only work with RCounterItem objects

public:
  bool          increment( ) ;
  virtual bool  reset( ) ;
  unsigned int  operator++( ) ;          // prefix
  unsigned int  operator++( int ) ;      // postfix

  unsigned int  getCount( ) const ;
} ;


//  Values are values (??) meant to follow the variation of something.  As such
//  it can hold a positive or negative value.  This value can be incremented and
//  decremented.  It also memorizes the peak value (min & max) along with the
//  time it occured.
//
//  The KeepFirstMinMax parameter is meant to control which timestamp we keep
//  when we have two peaks with the same value.
//

template< class val_t, class sum_t, class avg_t > class TValueItem : public RStatisticItem
{
  friend class RStatisticPool ;

  val_t         m_lastValue ;     // actual value
  int64_t       m_sum ;           // Overall sum (integral over time)
  unsigned int  m_nbLastDt ;      // Number of changes in the last quantum of time (sec for now)
                                  // IMPORTANT: it should never be zero.
  val_t         m_sumLastDt ;     // Sum of the changes that occured in the last quantum of time
  val_t         m_min ;
  val_t         m_max ;
  ZkClock       m_timeLast ;      // Time of the last update
  ZkClock       m_timeFirst ;     // Time when collecting started
  ZkClock       m_timeMin ;       // TIme of peak
  ZkClock       m_timeMax ;       // TIme of peak
  const bool    m_bKeepFirstMinMax ;

protected:
  TValueItem< val_t, sum_t, avg_t > ( const char  *pName,
                                      val_t        initialValue     = 0,
                                      bool         bKeepFirstMinMax = true,
                                      RMutex      *pMutex           = NULL ) : RStatisticItem( pName, pMutex ),
                                                                               m_lastValue( initialValue ),
                                                                               m_sum( 0 ),
                                                                               m_nbLastDt( 1 ),
                                                                               m_sumLastDt( initialValue ),
                                                                               m_min( initialValue ),
                                                                               m_max( initialValue ),
                                                                               m_bKeepFirstMinMax( bKeepFirstMinMax )
  {
    m_timeFirst =
    m_timeLast  =
    m_timeMin   =
    m_timeMax   = ::zkClockNow( ) ;
  } ;

  virtual bool  absorb( RStatisticItem *p ) // Will only work with TValueItem< > objects
  {
    TValueItem< val_t, sum_t, avg_t >  *pItem = dynamic_cast< TValueItem< val_t, sum_t, avg_t > * >( p ) ;
    bool                                bRc   = false ; // assume failure;

    if ( this == NULL || pItem == NULL )
      return false ; 

    pItem->lock( ) ;
    lock( ) ;

    if ( ::zkClockCmp( m_timeLast, pItem->m_timeFirst ) != 1 )  // don't mess with me 
    {
      if ( pItem->m_min < m_min || ! m_bKeepFirstMinMax && pItem->m_min == m_min ) 
      {
        m_min     = pItem->m_min ;
        m_timeMin = pItem->m_timeMin ;
      }
      if ( m_max < pItem->m_max || ! m_bKeepFirstMinMax && m_max == pItem->m_max )
      {
        m_max     = pItem->m_max ;
        m_timeMax = pItem->m_timeMax ;
      }

      int64_t  diff = ::zkClockToSeconds( ::zkClockDiffSat( pItem->m_timeFirst, m_timeLast ) ) ;

      if ( diff )
        m_sum += ( m_sumLastDt / m_nbLastDt + ( diff - 1 ) * m_lastValue ) ;

      m_sum       += pItem->m_sum ;
      m_lastValue  = pItem->m_lastValue ;
      m_nbLastDt   = pItem->m_nbLastDt ;
      m_sumLastDt  = pItem->m_sumLastDt ;
      m_timeLast   = pItem->m_timeLast ;

      pItem->reset( ) ;

      bRc = true ;
    }

    unlock( ) ;
    pItem->unlock( ) ;

    return bRc ;
  } ;


public:
  bool  increment( )
  {
    if ( this == NULL )
      return false ;

    lock( ) ;
    set( m_lastValue + 1 ) ;
    unlock( ) ;

    return true ;
  } ;

  bool  decrement( )
  {
    if ( this == NULL )
      return false ;

    lock( ) ;
    set( m_lastValue - 1 ) ;
    unlock( ) ;

    return true ;
  } ;

  bool  set( int v )
  {
    if ( this == NULL )
      return false ;

    lock( ) ;

    ZkClock  now  = ::zkClockNow( ) ;
    int64_t  diff = ::zkClockToSeconds( ::zkClockDiffSat( now, m_timeLast ) ) ;

    if ( diff ) // time has changed, update sum
    {
      m_sum       += ( m_sumLastDt / m_nbLastDt + ( diff - 1 ) * m_lastValue ) ;
      m_nbLastDt   = 0 ;
      m_sumLastDt  = 0 ;
    }

    m_nbLastDt++ ;
    m_lastValue  = v ;
    m_sumLastDt += v ;
    m_timeLast   = now ;

    if ( v < m_min || ! m_bKeepFirstMinMax && v == m_min )
    {
      m_min     = v ;
      m_timeMin = now ;
    }
    else if ( m_max < v || ! m_bKeepFirstMinMax && m_max == v )
    {
      m_max     = v ;
      m_timeMax = now ;
    }

    unlock( ) ;

    return true ;
  } ;

  virtual bool  reset( )
  {
    if ( this == NULL )
      return false ;

    m_min       =
    m_max       =
    m_sumLastDt = m_lastValue ;
    m_sum       = 0 ;
    m_nbLastDt  = 1 ;
    m_timeFirst =
    m_timeLast  =
    m_timeMin   =
    m_timeMax   = ::zkClockNow( ) ;

    return true ;
  } ;

  val_t  operator++( )  // prefix
  {
    if ( this == NULL )
      return 0 ;

    lock( ) ;
    increment( ) ;
    const val_t  v = m_lastValue ;
    unlock( ) ;

    return v ; // return the new value
  } ;

  val_t  operator++( int )  // postfix
  {
    if ( this == NULL )
      return 0 ;

    lock( ) ;
    const val_t  v = m_lastValue ;
    increment( ) ;
    unlock( ) ;

    return v ; // return the old value
  } ;

  val_t  operator--( )  // prefix
  {
    if ( this == NULL )
      return 0 ;

    lock( ) ;
    decrement( ) ;
    const val_t  v = m_lastValue ;
    unlock( ) ;

    return v ; // return the new value
  } ;

  val_t  operator--( int )  // postfix
  {
    if ( this == NULL )
      return 0 ;

    lock( ) ;
    const val_t  v = m_lastValue ;
    decrement( ) ;
    unlock( ) ;

    return v ; // return the old value
  } ;

  avg_t  getAverage( ) const
  {
    if ( this == NULL )
      return false ;

    lock( ) ;

    ZkClock  now    = ::zkClockNow( ) ;
    int64_t  diff   = ::zkClockToSeconds( ::zkClockDiffSat( now, m_timeLast ) ) ;
    int64_t  period = ::zkClockToSeconds( ::zkClockDiffSat( now, m_timeFirst ) ) + 1 ;
    avg_t    sum    = m_sum + m_sumLastDt / m_nbLastDt ;

    if ( 1 < diff ) // time has changed, update sum
      sum += ( diff * m_lastValue ) ;

    unlock( ) ;

    if ( period < 1 ) // This could happen if the time is changed (ntp, ...) during the execution
      period = 1 ;    // This should prevent the program to crash. (returned value might look strange)

// period should never be 0
    return ( period <= 0 ) ? 0 : ( sum / period ) ;
  } ;

  val_t  getValue( ) const
  {
    if ( this == NULL )
      return 0 ;

    lock( ) ;
    const val_t  v = m_lastValue ;
    unlock( ) ;

    return v ;
  } ;

  val_t  getMin( ZkClock *pTime = NULL ) const
  {
    if ( this == NULL )
      return 0 ;

    lock( ) ;
    const val_t  v = m_min ;
    if ( pTime )
      *pTime = m_timeMin ;
    unlock( ) ;

    return v ;
  } ;

  val_t  getMax( ZkClock *pTime = NULL ) const
  {
    if ( this == NULL )
      return 0 ;

    lock( ) ;
    const val_t  v = m_max ;
    if ( pTime )
      *pTime = m_timeMax ;
    unlock( ) ;

    return v ;
  } ;
} ;


typedef TValueItem< int, int64_t, double >  RIntValueItem ;


//  IntSamples are meant to keep track of integer value samples.  It holds the
//  sum of all the values submitted to the object, the number of values, and
//  memorizes the maximum and minimum values, along with the time they occured.
//
//  This class should be re-done using the same interface but a different
//  hierarchy.  I believe there should be a common class for samples or may be
//  a template.
//
//  The KeepFirstMinMax parameter is meant to control which timestamp we keep
//  when we have two peaks (or two minimum) with the same value.
//

template< class val_t, class sum_t, class avg_t > class TSamplesItem : public RStatisticItem
{
  friend class RStatisticPool ;

  sum_t         m_sum ;
  val_t         m_min ;
  val_t         m_max ;
  unsigned int  m_nb ;
  ZkClock       m_timeMin ;
  ZkClock       m_timeMax ;
  const bool    m_bKeepFirstMinMax ;

protected:
  TSamplesItem< val_t, sum_t, avg_t >( const char  *pName,
                                       bool         bKeepFirstMinMax = true,
                                       RMutex      *pMutex           = NULL ) : RStatisticItem( pName, pMutex ),
                                                                                m_sum( 0 ),
                                                                                m_min( 0 ),  // arbitrary value (safe)
                                                                                m_max( 0 ),  // arbitrary value (safe)
                                                                                m_nb( 0 ),
                                                                                m_timeMin( s_timeInvalid ),
                                                                                m_timeMax( s_timeInvalid ),
                                                                                m_bKeepFirstMinMax( bKeepFirstMinMax ) { } ;

  virtual bool  absorb( RStatisticItem *p ) // Will only work with TSamplesItem objects
  {
    TSamplesItem< val_t, sum_t, avg_t >  *pItem = dynamic_cast< TSamplesItem< val_t, sum_t, avg_t > * >( p ) ;

    if ( this == NULL || pItem == NULL )
      return false ;

    pItem->lock( ) ;

    if ( pItem->m_nb )
    {
      lock( ) ;

      if ( m_nb )
      {
        if ( pItem->m_min < m_min || ! m_bKeepFirstMinMax && pItem->m_min == m_min )
        {
          m_min     = pItem->m_min ;
          m_timeMin = pItem->m_timeMin ;
        }
        if ( m_max < pItem->m_max || ! m_bKeepFirstMinMax && m_max == pItem->m_max )
        {
          m_max     = pItem->m_max ;
          m_timeMax = pItem->m_timeMax ;
        }
      }
      else  // First value(s)
      {
        m_min     = pItem->m_min ;
        m_max     = pItem->m_max ;
        m_timeMin = pItem->m_timeMin ;
        m_timeMax = pItem->m_timeMax ;
      }

      m_sum += pItem->m_sum ;
      m_nb  += pItem->m_nb ;

      pItem->reset( ) ;

      unlock( ) ;
    }

    pItem->unlock( ) ;

    return true ;
  } ;


public:
  bool  add( val_t n )
  {
    if ( this == NULL )
      return false ;

    lock( ) ;

    if ( m_nb )
    {
      if ( n < m_min || ! m_bKeepFirstMinMax && n == m_min )
      {
        m_min     = n ;
        m_timeMin = ::zkClockNow( ) ;
      }
      if ( m_max < n || ! m_bKeepFirstMinMax && m_max == n )
      {
        m_max     = n ;
        m_timeMax = ::zkClockNow( ) ;
      }
    }
    else  // First value
    {
      m_min     =
      m_max     = n ;
      m_timeMin =
      m_timeMax = ::zkClockNow( ) ;
    }

    m_nb++ ;
    m_sum += n ;

    unlock( ) ;

    return true ;
  } ;

  virtual bool  reset( )
  {
    if ( this == NULL )
      return false ;

    lock( ) ;

    m_sum     =
    m_nb      = 0 ;
    m_min     =
    m_max     = 0 ;
    m_timeMin =
    m_timeMax = s_timeInvalid ;

    unlock( ) ;

    return true ;
  } ;

  unsigned int  getCount( ) const
  {
    if ( this == NULL )
      return 0 ;

    lock( ) ;
    const unsigned int  n = m_nb ;
    unlock( ) ;

    return n ;
  } ;

  avg_t  getAverage( ) const // returns an invalid value when no smaples in the set
  {
    if ( this == NULL )
      return 0 ;

    lock( ) ;
    const avg_t  a = m_nb == 0 ? 0 : ( ( avg_t )m_sum / m_nb ) ;
    unlock( ) ;

    return a ;
  } ;

  sum_t  getSum( ) const
  {
    if ( this == NULL )
      return 0 ;

    lock( ) ;
    const sum_t  s = m_sum ;
    unlock( ) ;

    return s ;
  } ;

  val_t  getMin( ZkClock *pTime = NULL ) const // returns an invalid value when no smaples in the set
  {
    if ( this == NULL )
      return 0 ;

    lock( ) ;
    const val_t  val = m_min ;
    if ( pTime )
      *pTime = m_timeMin ;
    unlock( ) ;

    return val ;
  } ;

  val_t  getMax( ZkClock *pTime = NULL ) const // returns an invalid value when no smaples in the set
  {
    if ( this == NULL )
      return 0 ;

    lock( ) ;
    const val_t  val = m_max ;
    if ( pTime )
      *pTime = m_timeMax ;
    unlock( ) ;

    return val ;
  } ;
} ;


typedef  TSamplesItem< int, int64_t, double >  RIntSamplesItem ;


//  States are meant to keep track of boolean values.  They hold the last
//  state, and also the number of switches (true->false & false->true).
//
class RStateItem : public RStatisticItem
{
  friend class RStatisticPool ;

  bool          m_bState ;
  unsigned int  m_nbChanges ;

protected:
  RStateItem( const char *pName, bool bInitialState = false, RMutex *pMutex = NULL ) ;
  virtual bool  absorb( RStatisticItem *pItem ) ; // Will only work with RStateItem objects

public:
  bool          set( ) ;
  bool          unset( ) ;
  virtual bool  reset( ) ;

  unsigned int  getCount( ) const ;
  bool          getState( ) const ;
} ;


//  The next four classes, the template function and the typedefs are
//  exclusively used by the RStatisticPoolDesc class.
//
struct RCounterItemDesc : public std::string
{
  RCounterItemDesc( const char  *pName ) : std::string( pName ) { } ;
  int  memsz8( ) const ;
} ;


struct RIntValueItemDesc : public std::string
{
  const int   m_initialValue ;
  const bool  m_bKeepFirstMinMax ;

  RIntValueItemDesc( const char  *pName,
                     int          initialValue,
                     bool         bKeepFirstMinMax ) : std::string( pName ),
                                                       m_initialValue( initialValue ),
                                                       m_bKeepFirstMinMax( bKeepFirstMinMax ) { } ;
  int  memsz8( ) const ;
} ;


struct RIntSamplesItemDesc : public std::string
{
  const bool  m_bKeepFirstMinMax ;

  RIntSamplesItemDesc( const char  *pName,
                       bool         bKeepFirstMinMax ) : std::string( pName ),
                                                         m_bKeepFirstMinMax( bKeepFirstMinMax ) { } ;
  int  memsz8( ) const ;
} ;


struct RStateItemDesc : public std::string
{
  const bool  m_bInitialState ;

  RStateItemDesc( const char  *pName,
                  bool         bInitialState ) : std::string( pName ),
                                                 m_bInitialState( bInitialState ) { } ;
  int  memsz8( ) const ;
} ;


template< > class std::hash< std::string >
{
public:
  size_t  operator( )( string const &str ) const
  {
    hash< char const * > h ;

    return h( str.c_str( ) ) ;
  }
} ;


typedef  std::list< class RCounterItemDesc >     RCounterItemDescList ;
typedef  std::list< class RIntValueItemDesc >    RIntValueItemDescList ;
typedef  std::list< class RIntSamplesItemDesc >  RIntSamplesItemDescList ;
typedef  std::list< class RStateItemDesc >       RStateItemDescList ;
typedef  std::hash_set< std::string >            RStringSet ;


//  This class is used to construct a template for pool of statistic objects.
//  Using that description, you will be able to construct pools of statistic
//  objects.
//
//  You construct such a template by adding items using the public members of
//  the class.
//
//  Item's name must be unique.
//
class RStatisticPoolDesc
{
  friend class RStatisticPool ;

  RCounterItemDescList     m_counters ;
  RIntValueItemDescList    m_values ;
  RIntSamplesItemDescList  m_samples ;
  RStateItemDescList       m_states ;

  RStringSet             m_names ;  // to prevent name clash

  unsigned int           m_size ;

  int  size( ) const { return this ? m_size : 0 ; } ;

public:
  RStatisticPoolDesc( ) ;
  virtual ~RStatisticPoolDesc( ) ;

  bool  addCounter(    const char *pName ) ;
  bool  addIntValue(   const char *pName, int initialValue, bool bKeepFirstMinMax ) ;
  bool  addIntSamples( const char *pName, bool bKeepFirstMinMax ) ;
  bool  addState(      const char *pName, bool bInititalState ) ;
} ;


//  The next four typedefs and the structure are exclusively used by the
//  RStatisticPool class.
//
typedef  std::hash_map< std::string, RCounterItem    * >  RCounters ;
typedef  std::hash_map< std::string, RIntValueItem   * >  RIntValues ;
typedef  std::hash_map< std::string, RIntSamplesItem * >  RIntSamples ;
typedef  std::hash_map< std::string, RStateItem      * >  RStates ;


struct StatisticPoolHeader
{
  unsigned long   m_version ;
  unsigned long   m_size ;
  void           *m_pStartOfItems ;
} ;


//  Here's the pool.
//  A pool is meant to contain statistic items.  You construct a pool by
//  providing a pool description (RStatisticPoolDesc) and some flags.  The
//  flags determine how the pool will be allocated.  Using a single description
//  you can create multiple pools, each of them with different (or same) flags.
//
//  Once created, you can retrieve the pointers to the statistic items you
//  previously put in the description.
//
//  A pool can absorb another pool.  When this happens, (Do not be afraid, it
//  does not happen suddenly, you must call absorb for this to occur), data
//  collected in pool A is combined with data from pool B.  Once completed,
//  pool B contains the merged data, and the data of pool A is reset.
//
class RStatisticPool
{
  RMutex               *m_pMutex ;
  char                 *m_pName ;
  const uint32_t        m_flags ;
  StatisticPoolHeader  *m_pHdr ;
  ZkShm                 m_shm ;  // When shared memory is used
  RCounters             m_counters ;
  RIntValues            m_values ;
  RIntSamples           m_samples ;
  RStates               m_states ;

  void  setupMemBuffer( const RStatisticPoolDesc *pDesc ) ;

public:
  RStatisticPool( const char *pName, RStatisticPoolDesc *pDesc, uint32_t flags ) ;
  virtual ~RStatisticPool( ) ;

  RCounterItem     *getCounterItem( const char *pName ) ;
  RIntValueItem    *getIntValueItem( const char *pName ) ;
  RIntSamplesItem  *getIntSamplesItem( const char *pName ) ;
  RStateItem       *getStateItem( const char *pName ) ;
  const char       *getName( ) const ;
  bool              absorb( RStatisticPool *pPool ) ;
  bool              reset( ) ;

  bool              lock( )   const ;
  bool              unlock( ) const ;
} ;


extern "C"
{

# endif // __cplusplus

// =======================================
//  C   L A N G U A G E   W R A P P E R S
// =======================================

  typedef  struct S_zkStatisticIntSamples_t  *ZkStatisticIntSamples ;
  typedef  struct S_zkStatisticState_t       *ZkStatisticState ;
  typedef  struct S_zkStatisticCounter_t     *ZkStatisticCounter ;
  typedef  struct S_zkStatisticIntValue_t    *ZkStatisticIntValue ;
  typedef  struct S_zkStatisticPool_t        *ZkStatisticPool ;
  typedef  struct S_zkStatisticPoolDesc_t    *ZkStatisticPoolDesc ;


  DLLExport ZkBool               zkStatisticPoolDescNew( ZkStatisticPoolDesc *pDesc ) ;

  DLLExport ZkBool               zkStatisticPoolDescAddCounter( ZkStatisticPoolDesc   desc,
                                                                const char           *pName ) ;

  DLLExport ZkBool               zkStatisticPoolDescAddIntValue( ZkStatisticPoolDesc   desc,
                                                                 const char           *pName,
                                                                 int                   initialValue,
                                                                 ZkBool                bKeepFirstMinMax ) ;

  DLLExport ZkBool               zkStatisticPoolDescAddIntSamples( ZkStatisticPoolDesc   desc,
                                                                   const char           *pName,
                                                                   ZkBool                bKeepFirstMinMax ) ;

  DLLExport ZkBool               zkStatisticPoolDescAddState( ZkStatisticPoolDesc   desc,
                                                              const char           *pName,
                                                              ZkBool                bInitialState ) ;

  DLLExport ZkBool               zkStatisticPoolDescFree( ZkStatisticPoolDesc *pDesc ) ;



  DLLExport ZkBool               zkStatisticPoolNew( ZkStatisticPool      *pPool,
                                                     ZkStatisticPoolDesc   desc,
                                                     const char           *pName,
                                                     uint32_t              flags ) ;

  DLLExport ZkBool               zkStatisticPoolLock( ZkStatisticPool pool ) ;
  DLLExport ZkBool               zkStatisticPoolUnlock( ZkStatisticPool pool ) ;

  DLLExport ZkBool               zkStatisticPoolAbsorb( ZkStatisticPool pool1, ZkStatisticPool pool2 ) ;

  DLLExport ZkBool               zkStatisticPoolReset( ZkStatisticPool pool ) ;

  DLLExport const char          *zkStatisticPoolGetName( ZkStatisticPool pool ) ;

  DLLExport ZkStatisticCounter   zkStatisticPoolGetCounterItem( ZkStatisticPool   pool,
                                                                const char       *pName ) ;

  DLLExport ZkStatisticIntValue  zkStatisticPoolGetIntValueItem( ZkStatisticPool   pool,
                                                                 const char       *pName ) ;

  DLLExport ZkStatisticIntSamples  zkStatisticPoolGetIntSamplesItem( ZkStatisticPool   pool,
                                                                     const char       *pName ) ;

  DLLExport ZkStatisticState     zkStatisticPoolGetStateItem( ZkStatisticPool   pool,
                                                              const char       *pName ) ;

  DLLExport ZkBool               zkStatisticPoolFree( ZkStatisticPool *pPool ) ;



  DLLExport const char          *zkStatisticCounterGetName( ZkStatisticCounter item ) ;

  DLLExport ZkBool               zkStatisticCounterIncrement( ZkStatisticCounter item ) ;

  DLLExport ZkBool               zkStatisticCounterReset( ZkStatisticCounter item ) ;

  DLLExport unsigned int         zkStatisticCounterGetCount( ZkStatisticCounter item ) ;



  DLLExport const char          *zkStatisticIntValueGetName( ZkStatisticIntValue item ) ;

  DLLExport ZkBool               zkStatisticIntValueIncrement( ZkStatisticIntValue item ) ;

  DLLExport ZkBool               zkStatisticIntValueDecrement( ZkStatisticIntValue item ) ;

  DLLExport ZkBool               zkStatisticIntValueSet( ZkStatisticIntValue item, int v ) ;

  DLLExport ZkBool               zkStatisticIntValueReset( ZkStatisticIntValue item ) ;

  DLLExport int                  zkStatisticIntValueGetValue( ZkStatisticIntValue item ) ;

  DLLExport double               zkStatisticIntValueGetAverage( ZkStatisticIntValue item ) ;

  DLLExport int                  zkStatisticIntValueGetMin( ZkStatisticIntValue   item,
                                                            ZkClock              *pTime ) ;

  DLLExport int                  zkStatisticIntValueGetMax( ZkStatisticIntValue   item,
                                                            ZkClock              *pTime ) ;



  DLLExport const char          *zkStatisticIntSamplesGetName( ZkStatisticIntSamples item ) ;

  DLLExport ZkBool               zkStatisticIntSamplesAdd( ZkStatisticIntSamples  item,
                                                           int                  n ) ;

  DLLExport ZkBool               zkStatisticIntSamplesReset( ZkStatisticIntSamples item ) ;

  DLLExport unsigned int         zkStatisticIntSamplesGetCount( ZkStatisticIntSamples item ) ;

  DLLExport double               zkStatisticIntSamplesGetAverage( ZkStatisticIntSamples item ) ;

  DLLExport int64_t              zkStatisticIntSamplesGetSum( ZkStatisticIntSamples item ) ;

  DLLExport int                  zkStatisticIntSamplesGetMin( ZkStatisticIntSamples   item,
                                                              ZkClock              *pTime ) ;

  DLLExport int                  zkStatisticIntSamplesGetMax( ZkStatisticIntSamples   item,
                                                              ZkClock              *pTime ) ;



  DLLExport const char          *zkStatisticStateGetName( ZkStatisticState item ) ;

  DLLExport ZkBool               zkStatisticStateSet( ZkStatisticState item ) ;

  DLLExport ZkBool               zkStatisticStateUnset( ZkStatisticState item ) ;

  DLLExport ZkBool               zkStatisticStateReset( ZkStatisticState item ) ;

  DLLExport unsigned int         zkStatisticStateGetCount( ZkStatisticState item ) ;

  DLLExport ZkBool               zkStatisticStateGetState( ZkStatisticState item ) ;

#ifdef __cplusplus

} ;

#endif // ifdef __cplusplus
#endif // ifndef STATISTIC_H_
#endif // if defined( LINUX )
