/*
 *  $Id: net.c,v 1.10.10.1 2000/10/11 13:20:01 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#ifndef WIN32
#include <netdb.h>
#include <netinet/in.h>
#include <sys/time.h>
#else
#include <winsock2.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include "zkNet.h"


/*
 *  Perform: res = a - b.
 *  
 *  bug? check for overflow/underflow conditions?
 */
static void zk_TvSubtract(struct timeval *res, struct timeval *a, struct timeval *b)
{
  res->tv_usec = a->tv_usec - b->tv_usec;
  res->tv_sec = a->tv_sec - b->tv_sec;
  
  if (res->tv_usec < 0) {
    res->tv_usec += 1000000;
    res->tv_sec--;
  }
}


/*
 *  Return a + b.
 *  
 *  bug: we assume that the fields of the initial values are within range.
 */
DLLExport struct timeval zkTvAdd(struct timeval a, struct timeval b)
{
  struct timeval r;

  r.tv_usec = a.tv_usec + b.tv_usec;
  r.tv_sec = a.tv_sec + b.tv_sec;

  if (r.tv_usec >= 1000000) {
    r.tv_usec -= 1000000;
    r.tv_sec++;
  }

  return r;
}


/*
 *  Return a - b.
 */
DLLExport struct timeval zkTvSubtract(struct timeval a, struct timeval b)
{
  struct timeval r;

  zk_TvSubtract(&r, &a, &b);
  return r;
}


/*
 *  Return a / b;
 */
DLLExport struct timeval zkTvDivide(struct timeval a, double b)
{
  double t;
  struct timeval r;
    
  t = (a.tv_sec * 1000000.0 + a.tv_usec) / b;
  
  r.tv_sec = t / 1000000.0;
  r.tv_usec = t - r.tv_sec * 1000000.0;

  return r;
}


/*
 *  Call select(), updating `*timeout' with the amount of time (approximately)
 *  actually spent in select().
 *  
 *  bug? update timeout even when select() returns -1?  What if a timer
 *  interrupts select()?
 *  
 *  bug: two extra system calls is a lot of overhead.
 */
DLLExport int zkSelectUpdateTimeout(ZkError err, int n, fd_set  *readfds,  
                                    fd_set  *writefds, fd_set *exceptfds, 
                                    struct timeval *timeout)
{
#ifdef SELECT_UPDATES_TIMEOUT
  int ret;
  int e;

  zkErrorClear(err);
  if ((ret = select(n, readfds, writefds, exceptfds, timeout)) < 0 ) {
#ifndef WIN32
    e = errno;
    zkErrorSet_va(err, "zkSelectUpdateTimeout", ZKERR, ZKMNET, errno, 
                  "Error from select() [%s]",strerror(errno));
    errno = e;
#endif
  }
  
  return ret;

#else

  int ret, e = 0;
  struct timeval slept, tv0, tv1;

  zkErrorClear(err);
  if (gettimeofday(&tv0, 0) == -1) {
    return -1;
  }
  
  ret = select(n, readfds, writefds, exceptfds, timeout);
  if(ret < 0){
    e = errno;
    zkErrorSet_va(err, "zkSelectUpdateTimeout", ZKERR, ZKMNET, errno, 
                  "Error from select() [%s]",strerror(errno));
    errno = e;
  }
  
  if (ret != -1) {
    if (gettimeofday(&tv1, 0) == -1) {
      return -1;
    } else {
      zk_TvSubtract(&slept, &tv1, &tv0);
      zk_TvSubtract(timeout, timeout, &slept);
    }
#ifdef WIN32
  } else if(SOCKET_ERROR == ret){
    char buf[50];
    snprintf(buf, sizeof buf, "error code [%d]", WSAGetLastError());
    ClientTrace("Select Failed", buf);
#endif
  }
    
  return ret;
  
#endif
}


/*
 *  Given a service name and protocol string, return via `port' the equivalent
 *  port number in network byte order.
 */
DLLExport int zkNetPortFromService(ZkError err, unsigned short *port, const char *service, const char *proto)
{
  struct servent *pse;
  unsigned short p;

  zkErrorClear(err);

  if ( ! port) {
    zkErrorSet(err, "zkNetPortFromService", ZKERR, ZKMNET, ZKENULLPTR,
               "port is null");
    return 0;
  }
  if ( ! service) {
    zkErrorSet(err, "zkNetPortFromService", ZKERR, ZKMNET, ZKENULLPTR,
               "service is null");
    return 0;
  }

  if ((pse = getservbyname(service, proto))) {
    p = pse->s_port;
  } else if ((p = htons(atoi(service))) == 0) {
    zkErrorSet_va(err, "zkNetPortFromService", ZKERR, ZKMNET, ZKERROR,
                  "unknown or invalid service `%s'", service);
    return 0;
  }

  *port = p;

  return 1;
}
