/*
 *  $Id: zkAlloc.c,v 1.12.16.1 2000/10/11 13:20:01 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include "zkAlloc.h"


DLLExport int zkMemAlloc(ZkError err, void **mem, size_t msz)
{
  void *m;

  zkErrorClear(err);

  if ( ! mem) {
    zkErrorSet(err, "zkMemAlloc", ZKERR, ZKMALLOC, ZKENULLPTR,
               "mem is null");
    return 0;
  }
  if (msz == 0) {
    zkErrorSet(err, "zkMemAlloc", ZKERR, ZKMALLOC, ZKEPARMBAD,
               "msz is 0");
    return 0;
  }

  if ( ! (m = malloc(msz))) {
    zkErrorSet_va(err, "zkMemAlloc", ZKERR, ZKMALLOC, errno,
                  "can't allocate %lu bytes: %s",
                  (unsigned long)msz, strerror(errno));
    return 0;
  }
  
  *mem = m;

  return 1;
}


DLLExport int zkMemCalloc(ZkError err, void **mem, size_t nmemb, size_t msz)
{
  void *m;

  zkErrorClear(err);

  if ( ! mem) {
    zkErrorSet(err, "zkMemCalloc", ZKERR, ZKMALLOC, ZKENULLPTR,
               "mem is null");
    return 0;
  }
  if (msz == 0) {
    zkErrorSet(err, "zkMemCalloc", ZKERR, ZKMALLOC, ZKEPARMBAD,
               "msz is 0");
    return 0;
  }

  if (nmemb == 0) {
    zkErrorSet(err, "zkMemCalloc", ZKERR, ZKMALLOC, ZKEPARMBAD,
               "msz is 0");
    return 0;
  }

  if ( ! (m = calloc(nmemb, msz))) {
    zkErrorSet_va(err, "zkMemCalloc", ZKERR, ZKMALLOC, errno,
                  "can't allocate %lu block(s) of %lu bytes: %s",
                  (unsigned long)nmemb, (unsigned long)msz, strerror(errno));
    return 0;
  }
  
  *mem = m;

  return 1;  
}


/*
 *  Append `asz' byte of `app' to the location pointed to by `*dst'.  It is
 *  assumed that latter points to dynamically allocated memory.
 *  
 *  If successful, 1 is returned and `*dst' is the new pointer to the memory,
 *  and `*dsz' is its new size.
 */
DLLExport int zkMemAppend(ZkError err, void **dst, size_t *dsz, const void *app, size_t asz)
{
  zkErrorClear(err);

  if ( ! dst) {
    zkErrorSet(err, "zkMemAppend", ZKERR, ZKMALLOC, ZKENULLPTR,
               "dst is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, "zkMemAppend", ZKERR, ZKMALLOC, ZKENULLPTR,
               "dsz is null");
    return 0;
  }
  if ( ! app) {
    zkErrorSet(err, "zkMemAppend", ZKERR, ZKMALLOC, ZKENULLPTR,
               "app is null");
    return 0;
  }

  if ( ! zkMemRealloc(err, dst, *dsz + asz)) {
    return 0;
  }

  memcpy((char *)*dst + *dsz, app, asz);

  *dsz += asz;

  return 1;
}


/*
 *  Create a copy of a chunk of memory.
 */
DLLExport int zkMemDup(ZkError err, void **dst, const void *src, size_t n)
{
  void *newp;

  zkErrorClear(err);

  if ( ! dst) {
    zkErrorSet(err, "zkMemDup", ZKERR, ZKMALLOC, ZKENULLPTR,
               "dst is null");
    return 0;
  }
  if ( ! src) {
    zkErrorSet(err, "zkMemDup", ZKERR, ZKMALLOC, ZKENULLPTR,
               "src is null");
    return 0;
  }
  if (n == 0) {
    zkErrorSet(err, "zkMemDup", ZKERR, ZKMALLOC, ZKENULLPTR,
               "n is 0");
    return 0;
  }

  if ( ! (newp = malloc(n))) {
    zkErrorSet_va(err, "zkMemDup", ZKERR, ZKMALLOC, errno,
                  "can't allocate %lu bytes: %s",
                  (unsigned long)n, strerror(errno));
    return 0;
  }

  memcpy(newp, src, n);
  *dst = newp; newp = 0;

  return 1;
}


/*
 *  If we can perform the memory reallocation, set `*p' to the new pointer and
 *  return 1.  If unsuccessful, `*p' remains unchanged and we return 0.
 *  
 *  ANSI semantics are explicitly _not_ supported for a null `*p'.  That is,
 *  we'd rather consider it an error than treat it as a malloc().
 *  
 *  Similarly, it is an error to pass a size of 0.
 *  
 *  This function merely provides a sane interface to realloc(), in which you
 *  don't need to keep a temporary pointer for fear of leaking memory on a
 *  failed reallocation.
 *  
 *  bug: it should be renamed to zkMemRealloc().
 */
DLLExport int zkMemRealloc(ZkError err, void **p, size_t n)
{
  void *newp;

  zkErrorClear(err);

  if ( ! p) {
    zkErrorSet(err, "zkMemRealloc", ZKERR, ZKMALLOC, ZKENULLPTR,
               "p is null");
    return 0;
  }
  if ( ! *p) {
    zkErrorSet(err, "zkMemRealloc", ZKERR, ZKMALLOC, ZKEPARMBAD,
               "*p is null");
    return 0;
  }
  if (n == 0) {
    zkErrorSet(err, "zkMemRealloc", ZKERR, ZKMALLOC, ZKEPARMBAD,
               "n is zero");
    return 0;
  }

  if ( ! (newp = realloc(*p, n))) {
    zkErrorSet_va(err, "zkMemRealloc", ZKERR, ZKMALLOC, ZKENOMEM,
                  "%lu byte realloc failed: %s",
                  (unsigned long)n, strerror(errno));
    return 0;
  }

  *p = newp;

  return 1;
}


/*
 * Free a previously alloated block of memory
 */
DLLExport int zkMemFree(ZkError err, void **mem)
{
  zkErrorClear(err);

  if ( ! mem) {
    zkErrorSet(err, "zkMemFree", ZKERR, ZKMALLOC, ZKENULLPTR,
               "mem is null");
    return 0;
  }
  free(*mem);

  *mem = NULL;

  return 1;
}

