/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <zkBase64.h>

#define	__pad64 	'='

static const char * __digit64 = 
  "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

static const unsigned char __reverse64[] = 
  { 255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255, 
    255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255, 
    255,255,255,255,255,255,255,255,255,255,255, 62,255,255,255, 63, 
     52, 53, 54, 55, 56, 57, 58, 59, 60, 61,255,255,255,100,255,255, 
    255,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 
     15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25,255,255,255,255,255, 
    255, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 
     41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51,255,255,255,255,255 };
    

/*
 * %%Name:
 *   ERRCODE COFFIN_toBase64(src,dest,size)
 * %%Parameters:
 *   char * src		The source (binary) string
 *   char * dest	Holds the base 64 encoding of src
 *   size_t size	The number of characters to encode from src
 * %%Description:
 *   This is a primitive base 64 converter.  You must make sure that the
 *   destination char string is large enough to hold the base64 encoding
 *   of the binary string (4 char for every 3 in the source, plus a NUL
 *   character to terminate it).  If you are encoding source data in segments,
 *   make sure that you encode a multiple of 3 bytes at a time to avoid
 *   padding.
 * %%Return Value:
 *   A non-zero value on success, a 0 on failure
 */

DLLExport int
zkConvertToBase64 (
  ZkError err ,
  const unsigned char * src ,
  unsigned char * dest ,
  size_t size )
{
  size_t	i = 0;		/* Loop index */
  size_t	k = 0;
  size_t	len = 0;

  /*
   * Encode the bulk of the binary data
   */

  zkErrorClear(err);
  len = (size / 3) * 3;
  for (i = 0; i < len; i += 3) {
    dest[k++] = __digit64[(src[i] & 0xFC) >> 2];
    dest[k++] = __digit64[((src[i] & 0x03) << 4) + ((src[i+1] & 0xF0) >> 4)];
    dest[k++] = __digit64[((src[i+1] & 0x0F) << 2) + ((src[i+2] & 0xC0) >> 6)];
    dest[k++] = __digit64[src[i+2] & 0x3F];
  }

  /*
   * take care of the padding
   */

  switch (size - len) {
  case 0:
    break;
  case 1:
    dest[k++] = __digit64[(src[i] & 0xFC) >> 2];
    dest[k++] = __digit64[(src[i] & 0x03) << 4];
    dest[k++] = __pad64;
    dest[k++] = __pad64;
    break;
  case 2:
    dest[k++] = __digit64[(src[i] & 0xFC) >> 2];
    dest[k++] = __digit64[((src[i] & 0x03) << 4) + ((src[i+1] & 0xF0) >> 4)];
    dest[k++] = __digit64[(src[i+1] & 0x0F) << 2];
    dest[k++] = __pad64;
    break;
  }

  /*
   * NUL terminate the string
   */

  dest[k] = (char) 0x0;

  return 1;
}

DLLExport    
int
zkConvertFromBase64 (
  ZkError err ,
  const unsigned char * src ,
  unsigned char * dest ,
  size_t * size )
{
  size_t i = 0;
  size_t j = 0;
  unsigned char digit;

  zkErrorClear(err);
  for (;;) {
    while ((digit = __reverse64[src[i] &0x7F]) == 255) {
      if (src[i] == 0) {
        goto done;
      }
      i++;
    }  
    if (digit == 100) {
      goto done;
    }
    i++;
      
    dest[j] = digit << 2;

    while ((digit = __reverse64[src[i] &0x7F]) == 255) {
      if (src[i] == 0) {
        goto done;
      }
      i++;
    }  
    if (digit == 100) {
      goto done;
    }
    i++;

    dest[j++] += digit >> 4;
    dest[j] = (digit & 0x0F) << 4;
    
    while ((digit = __reverse64[src[i] &0x7F]) == 255) {
      if (src[i] == 0) {
        goto done;
      }
      i++;
    }  
    if (digit == 100) {
      goto done;
    }
    i++;

    dest[j++] += digit >> 2;
    dest[j] = (digit & 0x03) << 6;

    while ((digit = __reverse64[src[i] &0x7F]) == 255) {
      if (src[i] == 0) {
        goto done;
      }
      i++;
    }  
    if (digit == 100) {
      goto done;
    }
    i++;

    dest[j++] += digit;
  }

done:
  dest[j] = (char) 0;
  *size = j;

  return 1;
}
