/*
 *  $Id: zkClock.c,v 1.6.20.1 2000/10/11 13:20:01 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  A ZkClock is a quasi-opaque type, representing the number of 
 *  seconds elapsed since January 1, 1970, UTC.  The range of a 
 *  ZkClock is that of a signed 64-bit integer.
 *  
 *  Important: NEVER assign a time_t directly to a ZkClock or vice-
 *  versa.  The encoding of time_t is implementation-defined, so 
 *  such an assignment is inherently non-portable.  Use the 
 *  conversion functions zkClockToTime() and zkClockFromTime() 
 *  instead.
 */

#include "zks.h"

#include <limits.h>
#include <time.h>
#include "zkClock.h"

#if defined(macintosh)
#	include <time.mac.h>
#	include <DateTimeUtils.h>
#	include <Errors.h>
#	include <Events.h>
#	include <OSUtils.h>
#	include <UTCUtils.h>

#	define MAC_TO_UNIX_EPOCH_DELTA		(-2082844800L)
#	define MAC_SYNC_TICKS_DELTA			60
	
	static long	GetUTCDelta(Boolean forceReadLoc);

#endif


static union order {
  char bytes[sizeof(ZkClock)];
  ZkClock clock;
} order = {
  { 0x07, 0x06, 0x05, 0x04, 0x03, 0x02, 0x01, 0x00 }
};


/*
 *  Return the value of `n' with its byte order reversed.
 */
static ZkClock reverse(ZkClock n)
{
  size_t i;
  union order from, to;

  from.clock = n;
  for (i = 0; i < sizeof from.bytes; i++) {
    to.bytes[i] = from.bytes[sizeof from.bytes - 1 - i];
  }
  return to.clock;
}


/*
 *  Convert a value from network byte order to host order.
 */
DLLExport ZkClock zkClockFromNbo(ZkClock n)
{
  if ((order.clock.rep & 0xff) == 0) {
    return n; /* big endian */
  } else {
    return reverse(n); /* little endian */
  }
}


/*
 *  Convert a value from host order to network byte order.
 */
DLLExport ZkClock zkClockToNbo(ZkClock n)
{
  if ((order.clock.rep & 0xff) == 0) {
    return n; /* big endian */
  } else {
    return reverse(n); /* little endian */
  }
}


/*
 *  Return the current time.
 */
DLLExport ZkClock zkClockNow(void)
{
	ZkClock	now;
	
#if !defined(macintosh)
	
	now.rep = time(NULL);
	
#else

	OSStatus	err	= paramErr;
	
	/* retrieve the current date & time, in UTC */
	
#	if TARGET_API_MAC_OS8 && TARGET_CPU_PPC
	
	if (GetUTCDateTime != NULL)
	{
		/* Mac OS 9 -- UTCUtils are available */
		UTCDateTime	dt;
		
		err = GetUTCDateTime(&dt, kUTCDefaultOptions);
		
		if (err == noErr)
		{
			now.rep		= dt.highSeconds;
			now.rep		<<= 32;
			now.rep		|= dt.lowSeconds;
		}
	}
	
#	endif	/* TARGET_API_MAC_OS8 && TARGET_CPU_PPC */
	
	if (err != noErr)
	{
		/* No UTCUtils -- do it the old way */
		unsigned long	localTime;
		
		GetDateTime(&localTime);
		
		now.rep = localTime - GetUTCDelta(false);
	}
	
	/* convert the date to use the UNIX epoch (1/1/1970) */
	now.rep += MAC_TO_UNIX_EPOCH_DELTA;
	
#endif
	
	return (now);
}


/*
 *  Return the saturated time difference between a and b.  Calcuate the
 *  difference as a - b.
 *  
 *  bug: currently doesn't do any checks, and doesn't do the saturation.
 */
DLLExport ZkClock zkClockDiffSat(ZkClock a, ZkClock b)
{
#if !defined(WIN32) && !defined(macintosh)
#warning "no checks, no saturation"
#endif

  ZkClock diff;
  
  diff.rep = a.rep - b.rep;
  
  return diff;
}


/*
 *  Return the largest positive ZkClock value.
 *  
 *  bug: this may not be entirely portable, even among two's complement
 *  machines.
 */
DLLExport ZkClock zkClockMaxPos(void)
{
  ZkClock maxClock;
  
  maxClock.rep = (LONG_LONG)-1 ^ (LONG_LONG)1 << (sizeof(LONG_LONG) * CHAR_BIT - 1);
  
  return (maxClock);
}


/*
 *  bug: assumes time_t is signed long.
 */
#if defined(macintosh)
# define TIME_T_MAX ULONG_MAX
# define TIME_T_MIN 0
#else
# define TIME_T_MAX LONG_MAX
# define TIME_T_MIN LONG_MIN
#endif


/*
 *  Perform the function of strftime() for ZkClock values.
 *  
 *  bug: converts ZkClocks to time_t, so we can't represent values outside
 *  that range.
 */
DLLExport size_t zkClockStrftime(char *buf, size_t bsz, const char *fmt, ZkClock n)
{
  if (bsz == 0) {
    return 0;
  }

  if (n.rep < TIME_T_MIN) {
    strncpy(buf, "<too long ago>", bsz - 1); buf[bsz-1] = '\0';
  } else {
    if (n.rep > TIME_T_MAX) {
      strncpy(buf, "<too far ahead>", bsz - 1); buf[bsz-1] = '\0';
    } else {
      time_t t = (time_t) n.rep;
      struct tm *tm = localtime(&t);

      return strftime(buf, bsz, fmt, tm) != 0;
    }
  }
  
  return strlen(buf);
}


/*
 *  Convert a ZkClock into a time_t.  ZkClocks are always encoded as 
 *  seconds since January 1 1970, UTC.  The encoding of time_t is 
 *  implementation-defined.
 */
DLLExport time_t zkClockToTime(ZkClock n)
{
#if defined(macintosh)
	
	OSStatus	err	= paramErr;
	
	/* convert from UNIX epoch (1/1/1970) to Mac OS epoch. */
	
	n.rep -= MAC_TO_UNIX_EPOCH_DELTA;
	
#	if TARGET_API_MAC_OS8 && TARGET_CPU_PPC
	
	/* convert from UTC to local time */
	
	if (ConvertUTCToLocalDateTime != NULL)
	{
		/* Mac OS 9 -- UTCUtils are available */
		LocalDateTime	ldt;
		UTCDateTime		udt;
		
		udt.highSeconds	= (UInt16) (n.rep >> 32);
		udt.lowSeconds	= (UInt32) n.rep;
		udt.fraction	= 0;
		
		err = ConvertUTCToLocalDateTime(&udt, &ldt);
		
		if (err == noErr)
		{
			n.rep	= ldt.highSeconds;
			n.rep	<<= 32;
			n.rep	|= ldt.lowSeconds;
		}
	}
	
#	endif	/* TARGET_API_MAC_OS8 && TARGET_CPU_PPC */
	
	if (err != noErr)
	{
		/* No UTCUtils -- do it the old way */
		n.rep += GetUTCDelta(false);
	}
	
	n.rep += _mac_msl_epoch_offset_;

#endif
	
	return ((time_t) n.rep);
}


/*
 *  Convert a time_t into a ZkClock .  ZkClocks are always encoded as 
 *  seconds since January 1 1970, UTC.  The encoding of time_t is 
 *  implementation-defined.
 */
DLLExport ZkClock zkClockFromTime(time_t t)
{
	ZkClock	n;
	
#if !defined(macintosh)
	
	n.rep = t;
	
#else
	
	OSStatus	err		= paramErr;
	
	/* time_t's are relative to the MSL epoch, which is January 1, 1900, 
	local time.  We need to convert first to the Mac OS epoch (January 1, 
	1904), perform any local-to-UTC conversion, then convert the epoch 
	again to that defined for ZkClock (January 1, 1970). */
	
	n.rep = t;
	n.rep -= _mac_msl_epoch_offset_;
	
#	if TARGET_API_MAC_OS8 && TARGET_CPU_PPC
	
	if (ConvertLocalToUTCDateTime != NULL)
	{
		/* Mac OS 9 -- UTCUtils are available */
		
		LocalDateTime	ldt;
		UTCDateTime		udt;
		
		ldt.highSeconds	= (UInt16) (n.rep >> 32);
		ldt.lowSeconds	= (UInt32) n.rep;
		ldt.fraction	= 0;
		
		err = ConvertLocalToUTCDateTime(&ldt, &udt);
		
		if (err == noErr)
		{
			n.rep	= udt.highSeconds;
			n.rep	<<= 32;
			n.rep	|= udt.lowSeconds;
		}
	}
	
#	endif	/* TARGET_API_MAC_OS8 && TARGET_CPU_PPC */
	
	if (err != noErr)
	{
		/* No UTCUtils -- do it the old way */
		n.rep -= GetUTCDelta(false);
	}
	
	n.rep += MAC_TO_UNIX_EPOCH_DELTA;
	
#endif
	
	return (n);
}


/*
 *  Returns a ZkClock set to a given number of seconds.
 *  
 *  IMPORTANT:  do NOT pass in a time_t to this function, as the 
 *  encoding of time_t is platform-dependent.  Use zkClockFromTime
 *  instead.
 */
DLLExport ZkClock zkClockFromSeconds(int64_t secs)
{
  ZkClock n;
  n.rep = secs;
  return (n);
}


/*
 *  Returns a 64-bit integer representing the number of seconds 
 *  contained in the given ZkClock.
 *  
 *  IMPORTANT:  do NOT assign the function result to a time_t, as the 
 *  encoding of time_t is platform-dependent.  Use zkClockToTime
 *  instead.
 */
DLLExport int64_t zkClockToSeconds(ZkClock n)
{
  return (n.rep);
}


/*
 *  Compare two ZkClock values.  Returns a value that is greater than, 
 *  less than, or equal to zero depending on whether c1 is greater than, 
 *  less than, or equal to c2.
 */
DLLExport int zkClockCmp(ZkClock c1, ZkClock c2)
{
  int result;
  
  if (c1.rep > c2.rep)
    result = 1;
  else if (c1.rep < c2.rep)
    result = -1;
  else
    result = 0;
  
  return (result);
}


/*
 *  Adds two ZkClocks together, returning the result.
 */
DLLExport ZkClock zkClockAdd(ZkClock a, ZkClock b)
{
  ZkClock result;
  
  result.rep = a.rep + b.rep;
  
  return result;
}


#if defined(macintosh)

/*
 *  Force re-reading of UTC delta info from parameter RAM.
 */
DLLExport void zkClockFlushUTCCache(void)
{
	GetUTCDelta(true);
}

/*
 *  Return the number of seconds' difference between the local 
 *  system time and UTC.  The delta isn't re-read every time (it 
 *  can be an expensive operation), rather it is cached internally.
 */
static long	GetUTCDelta(Boolean forceReadLoc)
{
	static UInt32	sLastSyncTicks	= 0;
	static long		sLastUTCDelta;
	UInt32			ticksNow;
	
	if (forceReadLoc || 
		((ticksNow = TickCount()) > sLastSyncTicks + MAC_SYNC_TICKS_DELTA))
	{
		MachineLocation	location;
		
		ReadLocation(&location);
		
		sLastSyncTicks = ticksNow;
		sLastUTCDelta = location.u.gmtDelta;
		
		sLastUTCDelta &= 0x00FFFFFF;
		if (sLastUTCDelta & 0x00800000)
			sLastUTCDelta |= 0xFF000000;
	}
	
	return (sLastUTCDelta);
}

#endif
