/*
 * $Id: zkConfig.c,v 1.27.32.3 2000/12/14 17:54:28 jean Exp $
 */

static const char * module_id_ = "$Id: zkConfig.c,v 1.27.32.3 2000/12/14 17:54:28 jean Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include <zkIntTypes.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>                   /* malloc(), free() */
#include <zkLog.h>
#include <string.h>
#include <ctype.h>
#include <zkError.h>
#include <zkConfig.h>

#define	__MAX_PARAMETERS        1024
#define	__MAX_SECTIONS           128

#define	__SECNAME_SZ             128
#define __LINE_SZ                512
#define	__ERRMSG_SZ               64

struct T_ZkSectionParameter {
  char * comment;
  char * name;
  char * value;
};

typedef struct T_ZkSectionParameter ZkSectionParameter;

struct T_ZkSection {
  char * name;
  size_t parameters;
  struct T_ZkSectionParameter * parameter[__MAX_PARAMETERS];  
  int last_parm;  /* Keeps track of last parameter to get the next one */ 
  int loop_parm;  /* For getting the attribute value pairs one after another */
};

typedef struct T_ZkSection ZkSection;

struct T_ZkConfig {
  int argc;            /* Number of command line parameters */
  const char ** argv;  /* Command line parameters */
  size_t sections;
  struct T_ZkSection * section[__MAX_SECTIONS];
  int last_sec;
};

/*
 * Internal function prototypes
 */

static int zkConfigSetSection ( ZkError , ZkConfig , const char * );

/**
 *  Set the command line environment in the zkConfig
 */

DLLExport int
zkConfigSetCmdParms (
  ZkError err,
  ZkConfig config ,
  int argc ,
  const char ** argv )
{
  (void) module_id_;
  zkErrorClear(err);
  if (config && argv) {
    config->argc = argc;
    config->argv = argv;
    return 1;
  }
  zkErrorSet(err,"zkConfigSetCmdParms",ZKERR,ZKMCONFIG,ZKENULLPTR,NULL);
  return 0;
}

/**
 *
 *
 */

DLLExport int
zkConfigGetParameter (
  ZkError err ,
  ZkConfig config ,
  uint32_t flags ,
  const char * section ,
  const char * parm ,
  void * val )
{
  int return_code = 0;
  int opt = 1;
  const char * rawval = NULL;  /* Raw value */
  const char * pptr = NULL;  /* Command line parameter name pointer */

  zkErrorClear(err);
  if (config && section && parm && val) {
   
    /*
     * First step is to find the value, look in the command line parameters
     */

    rawval = NULL;
    if (config->argv && (config->argc > 1)) {
      while (opt < config->argc) {
        pptr = config->argv[opt];
        if ((pptr[0] == '-') && (pptr[1] == '-')) {
          if (strcasecmp(&(pptr[2]),parm) == 0) {
            if ((flags & 0xFF) == TYPE_BOOLEAN) {
              rawval = (const void *) "yes";
              if (++opt < config->argc) {
                if (strncmp(config->argv[opt],"--",2) != 0) {
                  rawval = config->argv[opt];
                }
              } 
            } else if (++opt < config->argc) {
              rawval = config->argv[opt]; 
            }
            break;
          }
        }       
        opt++;
      } 
    }

    /*
     * If we didnt find it, look in the configuration structure itself
     */

    if (rawval == NULL) {
      if (!zkConfigGet(err,config,section,parm,&rawval)) {
        return 0;
      }
    }

    /*
     * Now procress the value
     */

    switch (flags & 0xFF) {
    case TYPE_BOOLEAN:
      if ((strcasecmp((const char *)rawval,"yes") == 0) ||
          (strcasecmp((const char *)rawval,"true") == 0) ||
          (strcasecmp((const char *)rawval,"on") == 0)) {
        *((uint8_t *)val) = 1;
        return_code = 1;
      } else {
        *((uint8_t *)val) = 0;
        return_code = 1;
      }
      break;
    case TYPE_UINT8: 
      *((uint8_t *)val) = (uint8_t) atoi(rawval);
      return_code = 1;
      break; 
    case TYPE_UINT16: 
      *((uint16_t *)val) = (uint16_t) atoi(rawval);
      return_code = 1;
      break; 
    case TYPE_UINT32: 
      *((uint32_t *)val) = (uint32_t) atoi(rawval);
      return_code = 1;
      break; 
    case TYPE_INT8: 
      *((int8_t *)val) = (int8_t) atoi(rawval);
      return_code = 1;
      break; 
    case TYPE_INT16: 
      *((int16_t *)val) = (int16_t) atoi(rawval);
      return_code = 1;
      break; 
    case TYPE_INT32: 
      *((int32_t *)val) = (int32_t) atoi(rawval);
      return_code = 1;
      break; 
    case TYPE_STRING: 
      *((const char **)val) = rawval;
      return_code = 1;
      break; 
    default:
      zkErrorSet(err,"zkConfigGetParameter",ZKERR,ZKMCONFIG,ZKEDATABAD,
                 "Unrecognized value type");
    }  
     
    return return_code;
  }
  zkErrorSet(err,"zkConfigGetParameter",ZKERR,ZKMCONFIG,ZKENULLPTR,NULL);
  return 0;
}

/** 
 *  Creates a new ZkConfig object.  ZkConfig objects are used to manipulate
 *  configuration files which are in the format of a windows .INI file.  The
 *  file is divided into sections, denoted by [SECTIONNAME] and consists of
 *  parameter / value pairs and comments.  A comment line is denoted by a ';'
 *  as the first non-whitespace character.
 *  Possible error codes are ZKENULLPTR and ZKENOMEM.
 *
 *  @param err             Used to return the error
 *  @param conf            Used to return a new ZkConfig object
 *  @version               1.0
 *  @author                Philippe Boucher
 *  @return                a non-zero value on success, a 0 on failure
 */

DLLExport int
zkConfigNew (
  ZkError err,
  ZkConfig * conf )
{
  zkErrorClear(err);
  if (conf) {
    *conf = (ZkConfig) malloc(sizeof(struct T_ZkConfig));
    if (*conf == NULL) {
      zkErrorSet(err,"zkConfigNew",ZKERR,ZKMCONFIG,ZKENOMEM,"malloc");
      return 0;
    }
    memset(*conf,0,sizeof(struct T_ZkConfig));
    (*conf)->argv = NULL;
    (*conf)->argc = 0;
    (*conf)->sections = 0;
    (*conf)->last_sec = -1;
    return 1;
  }
  zkErrorSet(err,"zkConfigNew",ZKERR,ZKMCONFIG,ZKENULLPTR,"conf is NULL");
  return 0;
}

/** 
 *  Used to destroy a ZkConfig object.  This function will completely free
 *  up the configuration object, including the section and parameter lists.
 *  On success, the pointer to the ZkConf object is reset to NULL.
 *  Possible error codes are ZKENULLPTR.
 *
 *  @param err              Used to return the error
 *  @param conf             Points to ZkConfig pointer to destroy
 *  @author                 Philippe Boucher
 *  @version                1.0
 *  @return                 A non-zero value on success, a 0 on failure
 */

DLLExport int
zkConfigFree (
  ZkError err,
  ZkConfig * conf )
{
  size_t i = 0;
  size_t j = 0;
  ZkSection * secptr = NULL;
  ZkSectionParameter * parmptr = NULL;

  zkErrorClear(err);
  if (conf) {
    if (*conf) {
      for (i = 0; i < (*conf)->sections; i++) {
        secptr = (*conf)->section[i];
        if (secptr) {
          for (j = 0; j < secptr->parameters; j++) {
            parmptr = secptr->parameter[j];
            if (parmptr) {
              if (parmptr->name) {
                free(parmptr->name);
              }
              if (parmptr->comment) {
                free(parmptr->comment);
              }
              if (parmptr->value) {
                free(parmptr->value);
              }
              free(parmptr);
            }
          }
          if (secptr->name) {
            free(secptr->name);
          }
          free(secptr);
        }
      }
      memset(*conf,0,sizeof(struct T_ZkConfig));
      free(*conf);
      *conf = NULL;
      return 1;
    }  
  }
  zkErrorSet(err,"zkConfigFree",ZKERR,ZKMCONFIG,ZKENULLPTR,"Null pointer passed as config");
  return 0;
}

/** 
 *  Used to load a .INI file into the ZkConfig object.  See the zkConfigNew()
 *  function for details of the file format.  The function will try to load in
 *  the contents of the configuration file, up to the first error.
 *  Possible error codes are ZKENULLPTR, ZKENOMEM, ZKEFILEOPEN, and
 *  ZKEPARSE.
 * 
 *  @param err                Used to return the error code
 *  @param conf               ZkConfig which will hold the configuration file
 *  @param filename           Name of configuration file to load
 *  @version                  1.0
 *  @author                   Philippe Boucher
 *  @return                   A non-zero value on success, a 0 on failure
 */

DLLExport int 
zkConfigLoad (
  ZkError err ,
  ZkConfig conf ,
  const char * filename )
{
  FILE * f = NULL;
  char line[__LINE_SZ + 1];
  int return_code = 1;
  int lineno = 0;
  char secname[__SECNAME_SZ + 1];

  if (conf && filename) {
    memset(secname,0,__SECNAME_SZ + 1);
    if ((f = fopen(filename,"r")) == NULL) {
      zkErrorSet(err,"zkConfigLoad",ZKERR,ZKMCONFIG,ZKEFILEOPEN,filename);
      return 0;
    }
    lineno = 0;
    while (!feof(f)) {
      lineno += 1;
      line[0] = (char) 0;
      fgets(line,__LINE_SZ,f);
      line[__LINE_SZ] = (char) 0;
      if (!zkConfigParseLine(err,conf,line,secname,lineno)) {
        return_code = 0;
        break;
      }
    }
    fclose(f);
    return return_code;
  }
  zkErrorSet(err,"zkConfigLoad",ZKERR,ZKMCONFIG,ZKENULLPTR,NULL);
  return 0;
}


DLLExport int
zkConfigLoadFromBuffer(
  ZkError err ,
  ZkConfig conf ,
  const char * buffer ,
  size_t sz )
{
  size_t start = 0;
  size_t offset = 0;
  char line[__LINE_SZ + 1];
  int parse_line = 0;
  int lineno = 0;
  int return_code = 1;
  char secname[__SECNAME_SZ + 1];

  zkErrorClear(err);
  if (conf && buffer) {
    offset = 0;
    memset(secname,0,__SECNAME_SZ + 1);
    while (start < sz) {
      if ((buffer[start] != (char) 13) && (buffer[start] != (char) 10)) {
        line[offset++] = buffer[start++];
      } else { 
        parse_line = 1;
        while ((buffer[start] == (char) 13) || (buffer[start] == (char) 10)) {
          start++; 
        }
      }
      if ((parse_line) || (offset >= __LINE_SZ)) {
        parse_line = 0;
        line[offset] = (char) 0;
        if (!zkConfigParseLine(err,conf,line,secname,lineno)) {
          return_code = 0;
          break;
        }
        offset = 0; 
      }
    }
    return return_code;
  }
  zkErrorSet(err,"zkConfigLoadFromBuffer",ZKERR,ZKMCONFIG,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int 
zkConfigParseLine (
  ZkError err ,
  ZkConfig conf ,
  char * line ,
  char * secname,
  int lineno )
{
  char * start = NULL;
  char * end = NULL;
  char * parm = NULL;
  char * value = NULL;
  int i = 0;
  char errmsg[__ERRMSG_SZ + 1];
  

  zkErrorClear(err);
  if (conf && line) {
    start = line;
    while (isspace(*start)) {
      start++;
    }
    switch (*start) {
    case '\0':  /* Empty line */
      break;
    case ';':  /* Comment line */
      break;
    case '[':  /* Section Name */
      start++;
      while (isspace(*start)) {
        start++;
      }
      i = strlen(start) - 1;
      while ((isspace(start[i])) && (i > 0)) {
        i -= 1;
      }
      if (start[i] != ']') {
        snprintf(errmsg,sizeof errmsg,"Parse error in line %d",lineno);
        zkErrorSet(err,"zkConfigLoad",ZKERR,ZKMCONFIG,ZKEPARSE,errmsg);
        return 0;        
      }
      while ((i > 1) && (isspace(start[i - 1]))) {
        i -= 1;
      }
      start[i] = (char) 0;
      strncpy(secname,start,__SECNAME_SZ);
      break;
    default:  /* Parameter = value */
      parm = start;
      start++;
      while ((!isspace(*start)) && (*start != '=')) {
        start++;
      }
      end = start;
      while (isspace(*start)) {
        start++;
      }
      if (*start != '=') {
        snprintf(errmsg,sizeof errmsg,"Parse error in line %d",lineno);
        zkErrorSet(err,"zkConfigLoad",ZKERR,ZKMCONFIG,ZKEPARSE,errmsg);
        return 0;
      }
      *end = (char) 0;
      start++;
      while (isspace(*start)) {
        start++;
      }
      value = start;
      i = strlen(value) - 1;
      while ((isspace(value[i])) && (i > 0)) {
        i--;
      } 
      value[i + 1] = (char) 0;
      
      zkConfigSet(err,conf,secname,NULL,parm,value);
      break;
    } 
    return 1;
  }
  zkErrorSet(err,"zkConfigLoad",ZKERR,ZKMCONFIG,ZKENULLPTR,"config is NULL");
  return 0;
}

/**
 *  This function is used to save a configuration as a windows .INI file format.
 *  See the zkConfigNew() function for details on the file format.
 *  Possible error codes are ZKEFILEOPEN, ZKENULLPTR
 *
 *  @param err                  Used to return the error code
 *  @param conf                 ZkConfig object to dump to the file.
 *  @param filename             Name of the configuration file to create 
 *  @author                     Philippe Boucher
 *  @version                    1.0
 *  @return                     A non-zero value on success, a 0 on failure
 */

DLLExport int
zkConfigSave (
  ZkError err, 
  ZkConfig conf ,
  const char * filename )
{
  size_t i = 0;
  size_t j = 0;
  ZkSection * secptr = NULL;
  ZkSectionParameter * parmptr = NULL;
  FILE * f = NULL;

  zkErrorClear(err);
  if ((conf) && (filename)) {
    f = fopen(filename,"w");
    if (f == NULL) {
      zkErrorSet(err,"zkConfigSave",ZKERR,ZKMCONFIG,ZKEFILEOPEN,filename);
      return 0;
    }
    for (i = 0; i < conf->sections; i++) {
      secptr = conf->section[i];
      if (secptr) {
        fprintf(f,"[ %s ]\n",secptr->name); 
        for (j = 0; j < secptr->parameters; j++) {
          parmptr = secptr->parameter[j];
          if (parmptr) {
            if (parmptr->comment) {
              fprintf(f,"; %s\n",parmptr->comment);
            }
            if ((parmptr->name) && (parmptr->value)) {
              fprintf(f,"%s = %s\n",parmptr->name,parmptr->value);
            }
          }
        }
        fprintf(f,"\n");
      }
    }
    fclose(f);
    return 1;
  }
  zkErrorSet(err,"zkConfigSave",ZKERR,ZKMCONFIG,ZKENULLPTR,"config is NULL");
  return 0;
}

/**
 *  Take the requested section from the configuration file and write it
 *  into a buffer.  This function will allocate a buffer of the appropriate
 *  size
 *
 *  @param err Returns the error if one occured
 *  @param conf The configuration object from which you wish to get a section
 *  @param section The requested section name
 *  @param buf Allocates and returns a buffer containing a text version of 
 *             the requested section
 *  @param sz Returns the size of the section
 *  @author Philippe Boucher (philippe@zks.net)
 *  @version 1.0
 *  @Return Non-zero on success, 0 on failure.
 */

DLLExport int
zkConfigSectionToBuffer (
  ZkError err ,
  ZkConfig conf ,
  const char * section ,
  void ** buf ,
  size_t * sz )
{
  static const char * func = "zkConfigSectionToBuffer";
  int return_code = 0;
  char * b = NULL; /* Allocates the text buffer */
  ZkSection * secptr = NULL;  /* Section pointer to dump */
  ZkSectionParameter * parmptr = NULL; /* Section parameter pointer */
  size_t ndx = 0; /* requested section's array index in the config object */
  size_t n = 0; /* Loop index */
  int base = 0;  /* Current write position in the buffer */
  int bytes = 0;  /* Number of bytes written to buffer */

  zkErrorClear(err);
  if (conf && section && buf && sz) {
    *buf = NULL;
    *sz = 0;
 
    /* Find the section */

    if (!zkConfigSectionIndexByName(err,conf,section,&ndx)) {

    } else if ((secptr = conf->section[ndx]) == NULL) {
      zkErrorSet(err,func,ZKERR,ZKMCONFIG,ZKEDATABAD,
                 "Internal ZkConfig error, section pointer is NULL");
    } else {  /* Dump the section to a buffer */
      
      /*
       * First figure out the exact size of the buffer 
       */

      *sz = strlen(secptr->name) + 7;
      for (n = 0; n < secptr->parameters; n++) {
        parmptr = secptr->parameter[n];
        if (parmptr) {
          if ((parmptr->name) && (parmptr->value)) {
            *sz += strlen(parmptr->name) + strlen(parmptr->value) + 4;
          }
        }
      }
      
      /*
       * Now create the buffer and load it
       */
      
      if ((b = (char *) malloc(*sz * sizeof(char))) == NULL) {
        zkErrorSet(err,func,ZKERR,ZKMCONFIG,ZKENOMEM,"malloc");
      } else {
        base = snprintf(b,*sz,"[ %s ]\n",secptr->name);
        for (n = 0; n < secptr->parameters; n++) {
          parmptr = secptr->parameter[n];
          if (parmptr) {
            if (parmptr->value && parmptr->name) {
              bytes = snprintf(&(b[base]),*sz - base,"%s = %s\n",
                               parmptr->name,parmptr->value);
              base += bytes;
            } 
          }
        }
        *buf = b;
        return_code = 1;
      }
    }

    /*
     * Clean up if we failed
     */

    if (return_code == 0) {
      if (b) {
        free(b);
      }
      *buf = NULL;
      *sz = 0;
    } 
        
  } else {
    zkErrorSet(err,func,ZKERR,ZKMCONFIG,ZKENULLPTR,NULL);
  }
  return return_code;
}

/**
 *  Take the configuration object and write it into a buffer.  
 *  This function will allocate a buffer of the appropriate size
 *
 *  @param err Returns the error if one occured
 *  @param conf The configuration object you wish to write to a buffer
 *  @param buf Allocates and returns a buffer containing a text version of 
 *             the given configuration object
 *  @param sz Returns the size of the buffer
 *  @author Philippe Desaulniers (philipped@zks.net)
 *  @version 1.0
 *  @Return Non-zero on success, 0 on failure.
 */

DLLExport int
zkConfigToBuffer (
  ZkError err ,
  ZkConfig conf ,
  void ** buf ,
  size_t * sz )
{
  static const char * func = "zkConfigToBuffer";
  int return_code = 1;
  char * b = NULL; /* Allocates the text buffer */
  ZkSection * secptr = NULL;  /* Section pointer to dump */
  ZkSectionParameter * parmptr = NULL; /* Section parameter pointer */
  size_t ndx = 0; /* requested section's array index in the config object */
  size_t n = 0; /* Loop index */
  int base = 0;  /* Current write position in the buffer */

  zkErrorClear(err);
  if (conf && buf && sz) {
    *buf = NULL;
    *sz = 0;


    /*
     * First figure out the exact size of the buffer 
     */
    for (ndx = 0; ndx < conf->sections; ndx++) {
      if ((secptr = conf->section[ndx]) == NULL) {
        zkErrorSet(err,func,ZKERR,ZKMCONFIG,ZKEDATABAD,
                   "Internal ZkConfig error, section pointer is NULL");
        return_code = 0;
        /* break out of for loop */
        break;
      } else {  /* Get section buffer size */      
        *sz += strlen(secptr->name) + 7;
        for (n = 0; n < secptr->parameters; n++) {
          parmptr = secptr->parameter[n];
          if (parmptr) {
            if ((parmptr->name) && (parmptr->value)) {
              *sz += strlen(parmptr->name) + strlen(parmptr->value) + 4;
            }
          }
        }
      }
    }
      
    /*
     * Now create the buffer and load it
     */
    if (return_code == 1) {
      if ((b = (char *) malloc(*sz * sizeof(char))) == NULL) {
        zkErrorSet(err,func,ZKERR,ZKMCONFIG,ZKENOMEM,"malloc");
        return_code = 0;
      } else {

        base = 0;

        for (ndx = 0; ndx < conf->sections; ndx++) {
          secptr = conf->section[ndx];

          base += snprintf(&b[base],*sz,"[ %s ]\n",secptr->name);
          for (n = 0; n < secptr->parameters; n++) {
            parmptr = secptr->parameter[n];
            if (parmptr) {
              if (parmptr->value && parmptr->name) {
                base += snprintf(&(b[base]),*sz - base,"%s = %s\n",
                                 parmptr->name,parmptr->value);
              } 
            }
          }
        }

        *buf = b;
        return_code = 1;
      }
    }

    /*
     * Clean up if we failed
     */

    if (return_code == 0) {
      if (b) {
        free(b);
      }
      *buf = NULL;
      *sz = 0;
    } 
        
  } else {
    zkErrorSet(err,func,ZKERR,ZKMCONFIG,ZKENULLPTR,NULL);
  }
  return return_code;
}


/**
 *  Set a section parameter in the ZkConfig object.  This function requires
 *  a section name, a parameter name and a value. The comment is optional and
 *  can be set to NULL.  The function will add the parameter to the specified
 *  section (the section name is required for later retreival as well).
 *  Possible error codes are ZKENULLPTR, ZKENOMEM.
 * 
 *  @param err              Used to return the error
 *  @param conf             ZkConfig object to which to add the parameter
 *  @param section          Name of the section to which to add the parameter
 *  @param comment          An optional comment for the parameter (can be NULL)
 *  @param parm             The name of the parameter (must be unique)
 *  @param value            The value of the parameter 
 *  @version                1.0
 *  @author                 Philippe Boucher
 *  @return                 A non-zero value on success, a 0 on failure
 */

DLLExport int
zkConfigSet (
  ZkError err ,
  ZkConfig conf ,
  const char * section ,
  const char * comment ,
  const char * parm ,
  const char * value )
{
  size_t i = 0;                               /* Loop index */
  ZkSection * secptr = NULL;
  ZkSectionParameter * parmptr = NULL; 

  zkErrorClear(err);
  if ((conf) && (section) && (parm) && (value)) {
    for (i = 0; i < conf->sections; i++) {
      secptr = conf->section[i];
      if (strcasecmp(secptr->name,section) == 0) {
        break;
      }
    }

    /*
     * We dont have that section yet, so add it to the configuration...
     */
  
    if (i == conf->sections) {
      secptr = (ZkSection *) malloc(sizeof(ZkSection));
      if (secptr == NULL) {
        zkErrorSet(err,"zkConfigSet",ZKERR,ZKMCONFIG,ZKENOMEM,"malloc");
        return 0;
      }
      memset(secptr,0,sizeof(ZkSection));
      secptr->name = (char *) strdup(section);
      secptr->parameters = 0;
      secptr->last_parm = -1;
      secptr->loop_parm = 0;
      conf->section[conf->sections] = secptr;
      conf->sections += 1;
    }
    parmptr = (ZkSectionParameter *) malloc(sizeof(ZkSectionParameter));
    if (parmptr == NULL) {
      zkErrorSet(err,"zkConfigSet",ZKERR,ZKMCONFIG,ZKENOMEM,"malloc");
      return 0;
    }
    memset(parmptr,0,sizeof(ZkSectionParameter));
    if (comment) {
      parmptr->comment = (char *) strdup(comment);
    }
    parmptr->name = (char *) strdup(parm);
    parmptr->value = (char *) strdup(value); 
    secptr->parameter[secptr->parameters] = parmptr;
    secptr->parameters += 1;
    
    return 1; 
  }
  zkErrorSet(err,"zkConfigSet",ZKERR,ZKMCONFIG,ZKENULLPTR,"config is NULL");
  return 0;
}


static int
zkConfigSetSection (
  ZkError err ,
  ZkConfig conf ,
  const char * section )
{
  ZkSection * secptr = NULL; /* Section pointer */
  int i = 0;                /* Loop index */

  if (conf && section) {
    if ((conf->last_sec > -1) && (conf->last_sec < (int) conf->sections)) {
      secptr = conf->section[conf->last_sec];
      if (secptr == NULL) {
        conf->last_sec = -1;
      } else if (secptr->name == NULL) {
        conf->last_sec = -1;
      } else if (strcasecmp(secptr->name,section) != 0) {
        conf->last_sec = -1;
      }
    } else {
      conf->last_sec = -1;
    }
    if (conf->last_sec == -1) {
      for (i = 0; i < (int) conf->sections; i++) {
        secptr = conf->section[i];
        if (secptr) {
          if (secptr->name) {
            if (strcasecmp(secptr->name,section) == 0) {
              conf->last_sec = i;
              return 1;
            }
          }
        }
      }
      zkErrorSet_va(err,"zkConfigSetSection",ZKERR,ZKMCONFIG,ZKENOTFOUND,
                    "Section %s not found",section);
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkConfigSetSection",ZKERR,ZKMCONFIG,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkConfigGetPairReset (
  ZkError err ,
  ZkConfig conf,
  const char * section )
{
  ZkSection * secptr = NULL;
  zkErrorClear(err);
  if (section && conf) {
    if (!zkConfigSetSection(err,conf,section)) {
      return 0;
    } 
    if ((secptr = conf->section[conf->last_sec]) == NULL) {
      zkErrorSet(err,"zkConfigGetPairReset",ZKERR,ZKMCONFIG,ZKERANGE,
                 "Internal section pointer is corrupted");
      return 0;
    }
    secptr->loop_parm = 0;
    return 1;
  }
  zkErrorSet(err,"zkConfigGetPairReset",ZKERR,ZKMCONFIG,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkConfigGetPair (
  ZkError err,
  ZkConfig conf,
  const char * section ,
  const char ** name ,
  const char ** value )
{
  ZkSection * secptr = NULL;    /* Section pointer */
  ZkSectionParameter * parmptr = NULL; /* Parameter struct pointer */

  zkErrorClear(err);
  if (conf && name && value) {
    if (section) {
      if (!zkConfigSetSection(err,conf,section)) {
        return 0;
      }
    } else if ((conf->last_sec < 0) || (conf->last_sec >= (int) conf->sections)) {
      zkErrorSet(err,"zkConfigGetPair",ZKERR,ZKMCONFIG,ZKERANGE,
                 "Internal section pointer is corrupted");
      return 0;
    } 
    if ((secptr = conf->section[conf->last_sec]) == NULL) {
      zkErrorSet(err,"zkConfigGetPair",ZKERR,ZKMCONFIG,ZKERANGE,
                 "That section is not allocated");
      return 0;
    }
    if (secptr->loop_parm < (int) secptr->parameters) {
      parmptr = secptr->parameter[secptr->loop_parm];
      if (parmptr == NULL) {
        zkErrorSet(err,"zkConfigGetPair",ZKERR,ZKMCONFIG,ZKERROR,
                   "Corrupted parameter");
        return 0;
      }
      *name = parmptr->name;
      *value = parmptr->value;
      secptr->loop_parm += 1;
      return 1;
    } 
    zkErrorSet(err,"zkConfigGetPair",ZKERR,ZKMCONFIG,ZKENOTFOUND,
               "No more parameters in section");
    return 0;
  }
  zkErrorSet(err,"zkConfigGetPair",ZKERR,ZKMCONFIG,ZKENULLPTR,NULL);
  return 0;
}

/*
 * Not implemented yet
 */

#if 0
DLLExport int
zkConfigUnset (
  ZkError err, 
  ZkConfig conf ,
  char * section ,
  char * parm )
{
  zkErrorSet(err,"zkConfigUnset",ZKWARN,ZKMCONFIG,ZKENOSYS,"Unset");
  return 0;
}
#endif

/**
 *  This function is used to reset a parameter value from a given section.
 *  The section and parameter names are case insensitive.
 *  Possible error codes are ZKENOTFOUND and ZKENULLPTR.
 * 
 *  @param err                     Used to return the error
 *  @param conf                    ZkConfig object to search
 *  @param section                 Name of the section where the parameter is
 *  @param parm                    Name of the desired parameter
 *  @param value                   Returns the parameter value string
 *  @version                       1.0
 *  @author                        Philippe Boucher
 *  @return                        A non-zero value on success, a 0 on failure
 */

DLLExport int
zkConfigReset (
  ZkError err ,
  ZkConfig conf ,
  const char * section ,
  const char * parm ,
  const char * value )
{
  size_t i = 0;
  ZkSection * secptr = NULL;
  ZkSectionParameter * parmptr = NULL;

  zkErrorClear(err);
  if ((section) && (parm) && (value) && (conf)) {
 
    for (i = 0; i < conf->sections; i++) {
      if ((secptr = conf->section[i]) == NULL) {
        continue;
      }
      if (secptr->name == NULL) {
        continue;
      }
      if (strcasecmp(secptr->name,section) == 0) {
        conf->last_sec = i;
        for (i = 0; i < secptr->parameters; i++) {
          if ((parmptr = secptr->parameter[i]) == NULL) {
            continue;
          }
          if (parmptr->name == NULL) {
            continue;
          }
          if (strcasecmp(parmptr->name,parm) == 0) {
            secptr->last_parm = i;

	    if(parmptr->value) {
		free(parmptr->value);
	    }
	    parmptr->value = (char *) strdup(value);

            return 1;
          }
        }
        zkErrorSet(err,"zkConfigGet",ZKWARN,ZKMCONFIG,ZKENOTFOUND,parm);
        return 0;
      }
    }
    zkErrorSet(err,"zkConfigGet",ZKWARN,ZKMCONFIG,ZKENOTFOUND,section);
    return 0;
  }
  zkErrorSet(err,"zkConfigGet",ZKERR,ZKMCONFIG,ZKENULLPTR,"config is NULL");
  return 0;
}


/**
 *  This function is used to retreive a parameter value from a given section.
 *  The section and parameter names are case insensitive.
 *  Possible error codes are ZKENOTFOUND and ZKENULLPTR.
 * 
 *  @param err                     Used to return the error
 *  @param conf                    ZkConfig object to search
 *  @param section                 Name of the section where the parameter is
 *  @param parm                    Name of the desired parameter
 *  @param value                   Returns the parameter value string
 *  @version                       1.0
 *  @author                        Philippe Boucher
 *  @return                        A non-zero value on success, a 0 on failure
 */

DLLExport int
zkConfigGet (
  ZkError err ,
  ZkConfig conf ,
  const char * section ,
  const char * parm ,
  const char ** value )
{
  size_t i = 0;
  ZkSection * secptr = NULL;
  ZkSectionParameter * parmptr = NULL;

  zkErrorClear(err);
  if ((section) && (parm) && (value) && (conf)) {
    *value = NULL;
    for (i = 0; i < conf->sections; i++) {
      if ((secptr = conf->section[i]) == NULL) {
        continue;
      }
      if (secptr->name == NULL) {
        continue;
      }
      if (strcasecmp(secptr->name,section) == 0) {
        conf->last_sec = i;
        for (i = 0; i < secptr->parameters; i++) {
          if ((parmptr = secptr->parameter[i]) == NULL) {
            continue;
          }
          if (parmptr->name == NULL) {
            continue;
          }
          if (strcasecmp(parmptr->name,parm) == 0) {
            secptr->last_parm = i;
            *value = parmptr->value;
            return 1;
          }
        }
        zkErrorSet(err,"zkConfigGet",ZKWARN,ZKMCONFIG,ZKENOTFOUND,parm);
        return 0;
      }
    }
    zkErrorSet(err,"zkConfigGet",ZKWARN,ZKMCONFIG,ZKENOTFOUND,section);
    return 0;
  }
  zkErrorSet(err,"zkConfigGet",ZKERR,ZKMCONFIG,ZKENULLPTR,"config is NULL");
  return 0;
}

/**
 *  Get the next parameter of the same section and name.  Once you have
 *  called zkConfigGet(), you can call zkConfigGetNext() to get the other
 *  parameters in the same section which have the same name.  The function 
 *  will fail if there are no more matching parameters.  This function will
 *  always be reset on a call to zkConfigGet().
 * 
 *  @param err Returns the error if any occured.
 *  @param conf The ZkConfig object to scan
 *  @param value Returns a pointer to a NUL terminated string inside the
 *               configuration object
 *  @return A non-zero value on success, a 0 on failure (no more parameters 
 *          to be found)
 *  @version 1.0
 *  @author Philippe Boucher
 */

DLLExport int
zkConfigGetNext (
  ZkError err ,
  ZkConfig conf ,
  const char ** value )
{
  ZkSection * secptr = NULL;
  ZkSectionParameter * parmptr = NULL;
  ZkSectionParameter * parmptr2 = NULL;
  int i = 0;

  zkErrorClear(err);
  if ((conf == NULL) || (value == NULL)) {
    zkErrorSet(err,"zkConfigGetNext",ZKERR,ZKMCONFIG,ZKENULLPTR,
               "conf or value is NULL");
    return 0;
  }
  *value = NULL;
  if (conf->last_sec >= 0) {
    secptr = conf->section[conf->last_sec];
    if (secptr != NULL) {
      if (secptr->last_parm >= 0) {
        parmptr = secptr->parameter[secptr->last_parm]; 
        if (parmptr != NULL) {
          for (i = secptr->last_parm + 1; i < (int) secptr->parameters; i++) {
            parmptr2 = secptr->parameter[i];
            if (strcasecmp(parmptr2->name,parmptr->name) == 0) {
     
              /*
               * Found the next parameter.  update and return
               */

              secptr->last_parm = i;
              *value = parmptr2->value;
              return 1;
            }
          }

          /*
           * Failed to find the next one, return a 0
           */

          zkErrorSet_va(err,"zkConfigGetNext",ZKWARN,ZKMCONFIG,ZKENOTFOUND,
                     "Could not locate [%s]", parmptr->name);
          return 0;
        }
      }
 
      /*
       * No last parameter...
       */

      zkErrorSet(err,"zkConfigGetNext",ZKERR,ZKMCONFIG,ZKENOTFOUND,
                 "last section");
      return 0;
    }
  }

  /*
   * No last section....
   */

  zkErrorSet(err,"zkConfigGetNext",ZKERR,ZKMCONFIG,ZKENOTFOUND,"last section");
  return 0;
}

/**
 *  This function returns the number of sections currently in the ZkConfig
 *  object.
 *  Possible error codes are ZKENULLPTR.
 * 
 *  @param err                  Used to return the error
 *  @param conf                 ZkConfig object to analyze
 *  @param count                Returns the number of sections in the object
 *  @version                    1.0
 *  @author                     Philippe Boucher
 *  @return                     A non-zero value on success, a 0 on failure
 */

DLLExport int
zkConfigSections (
  ZkError err, 
  ZkConfig conf ,
  size_t * count )
{
  zkErrorClear(err);
  if ((conf) && (count)) {
    *count = conf->sections;
    return 1;
  }
  if (count) {
    *count = 0;
  }
  zkErrorSet(err,"zkConfigSections",ZKERR,ZKMCONFIG,ZKENULLPTR,
             "config or count = NULL");
  return 0; 
}

/** 
 *  Used to retrieve a section name using an index.  This function will return
 *  the name of the specified section (These names are required in order to
 *  retrieve parameters)
 *  Possible error codes are ZKEPARMBAD, and ZKENULLPTR.
 *
 *  @param err                   Used to return the error
 *  @param conf                  ZkConfig object to analyze
 *  @param section               The desired section number
 *  @param name                  Used to return the section name
 *  @version                     1.0
 *  @author                      Philippe Boucher
 *  @return                      A non-zero value on success, a 0 on failure
 */

DLLExport int
zkConfigSection (
  ZkError err ,
  ZkConfig conf ,
  size_t section ,
  const char ** name )
{
  ZkSection * secptr = NULL;

  zkErrorClear(err);
  if ((conf) && (name)) {
    if (section < conf->sections) {
      secptr = conf->section[section];
      if (secptr) {
        *name =  secptr->name;
        return 1;
      }
    }
    zkErrorSet(err,"zkConfigSection",ZKERR,ZKMCONFIG,ZKEPARMBAD,"section");
    return 0;
  }
  zkErrorSet(err,"zkConfigSection",ZKERR,ZKMCONFIG,ZKENULLPTR,"config = NULL");
  return 0;
}

/** 
 *  Used to retrieve a section index using an section name.
 *  This function will return
 *  the index of the specified section (These names are required in order to
 *  retrieve parameters)
 *  Possible error codes are ZKEPARMBAD, and ZKENULLPTR.
 *
 *  @param err                   Used to return the error
 *  @param conf                  ZkConfig object to analyze
 *  @param section               Used to return the section index
 *  @param name                  The desired section number
 *  @version                     1.0
 *  @author                      Luc Boulianne
 *  @return                      A non-zero value on success, a 0 on failure
 */

DLLExport int
zkConfigSectionIndexByName (
  ZkError err ,
  ZkConfig conf ,
  const char * name ,
  size_t *section )
{
  static const char *funcname = "zkConfigSectionIndexByName";
  ZkSection * secptr = NULL;
  size_t i;

  zkErrorClear(err);
  if ((conf) && (section)) {
    if (name && *name) {
      for (i = 0; i < conf->sections; i++) {
	secptr = conf->section[i];
	if (secptr) {
	  if (strcasecmp (secptr->name, name) == 0) {
	    *section = i;
	    return 1;
	  }
	} else {
          zkErrorSet(err,funcname,ZKWARN,ZKMCONFIG,ZKENULLPTR,
                     "ZkConfig internal section pointer is NULL");
	}
      }
      *section = 0;
      zkErrorSet_va(err,funcname,ZKWARN,ZKMCONFIG,ZKENOTFOUND,
		    "section name [%s] not found", name);
    }
    zkErrorSet(err,funcname,ZKERR,ZKMCONFIG,ZKEPARMBAD,"empty section name");
    return 0;
  }
  zkErrorSet(err,funcname,ZKERR,ZKMCONFIG,ZKENULLPTR,"NULL parameters");
  return 0;
}
