/*
 * $Id: zkDebug.c,v 1.21.10.1 2000/10/11 13:20:02 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */ 

#include "zks.h"
#include "zkIntTypes.h"

#include <stdio.h>
#include <stdlib.h> 
#include <stdarg.h>
#include <string.h>
#include <errno.h>

#ifndef WIN32
#include <unistd.h>
#else
#include <winsock2.h>
#endif

#include <time.h>
#include <zkDebug.h>

/* Only active when NDEBUG is defined */

#ifndef WIN32
#ifndef stderr
#define StdDebug _IO_stderr
#else 
#define StdDebug stderr
#endif /* stderr */
#endif

#define	__TIME_SZ	                     24
#define  ZK_DEBUG_MAX_STREAM           10

#ifndef WIN32

static struct {
  FILE    *list[ZK_DEBUG_MAX_STREAM];
  uint16_t size;
  uint16_t limit;
}ZkDebugStreamList = { { NULL }, 1, ZK_DEBUG_MAX_STREAM };

#endif

static int ZkDebugFlag = 0;

#define __DebugMax 8*1024
static char DebugBuf[__DebugMax];

#ifdef __linux
extern int vsnprintf ( char *str, size_t n,
		       const char *format, va_list ap );
#endif

DLLExport 
int zkDebugOn (void)
{
#ifndef WIN32    
	/* the static struct can't be initialized with a non constant w/ gcc 2.95 */
  if (ZkDebugStreamList.list[0] == NULL) ZkDebugStreamList.list[0] = StdDebug;
#endif
  ZkDebugFlag = 1;
  return 1;
}

#ifndef NDEBUG
DLLExport 
int zkDebugIsOn (void)
{
  return (ZkDebugFlag);
}
#endif

DLLExport 
int zkDebugOff (void)
{
  ZkDebugFlag = 0;
  return 1;
}

DLLExport 
int zkDebugToggle (void)
{
#ifndef WIN32    
  if (ZkDebugStreamList.list[0] == NULL) ZkDebugStreamList.list[0] = StdDebug;
#endif
  ZkDebugFlag = !ZkDebugFlag;
  return 1;
}


#ifndef WIN32
int zkDebugAddStream(FILE * aStream)
{
  int i;

  if( !aStream ) {
    return 0;
  }

  if( ZkDebugStreamList.size == ZkDebugStreamList.limit ) {    
    return 0;
  }

  for( i=0; i<ZkDebugStreamList.size; i++ ) {
    if( ZkDebugStreamList.list[i] == aStream ) {
      return 1;
    }
  }

  ZkDebugStreamList.list[ZkDebugStreamList.size++] = aStream;
 
  return 1;
}
#endif


    

/*
 * zkDebug_va: a debugging tool
 *
 *	zkDebug (format, arg1, arg2 ...);
 *
 */

DLLExport int
zkDebug (
   const char * msg ,
  ... )
{
#ifndef NDEBUG
  const char *func = "Debug> ";
  char *buf = DebugBuf, *p;
  int len=0;
  int err = errno; /* for safekeeping */
  va_list ap;
  struct tm * utc;
  char timestr[__TIME_SZ + 1] = "-";
  time_t timep = 0;
  int pid = 0;
  int i;


  if (ZkDebugFlag == 0)
    return 1;
  
  if (msg == NULL) {
    msg = "[NULL]";
  }

  timep = time(NULL);
  utc = gmtime(&timep);
  if (utc != NULL) {
    strftime(timestr,__TIME_SZ,"%Y/%m/%d-%H:%M:%S.0",utc);
  }

  pid = getpid(); 

  va_start(ap, msg);
  len = vsnprintf(buf, __DebugMax, msg, ap);
  va_end(ap);
  if (len < 0) goto full;

  /* Look for %m messages */
  if ((p = strstr (buf, "%m")) != NULL) {

    ClientTrace(func, buf);
    ClientTrace(func,strerror(err));
#ifndef WIN32
    *(p+1) = 's'; /* convert it to a string */
    for( i = 0; i<ZkDebugStreamList.size; i++ ) {
      fprintf(ZkDebugStreamList.list[i], func);
      fprintf(ZkDebugStreamList.list[i], buf, strerror(err));
      fprintf(ZkDebugStreamList.list[i], "\n");
      fflush(ZkDebugStreamList.list[i]);
    }
#endif
  } else {

    ClientTrace(func,buf);
#ifndef WIN32
    for( i = 0; i<ZkDebugStreamList.size; i++ ) {
      fprintf(ZkDebugStreamList.list[i],
              "[ DEBUG ] %s %d \"%s\"\n",
              timestr,
              (int) pid,            
              buf);
      fflush(ZkDebugStreamList.list[i]);
    }
# endif /* WIN32 */

  }
  return 1;
 full:

  ClientTrace (func, "INTERNAL MESSAGE OVERFLOWED\n");

# ifndef WIN32
  fprintf(StdDebug, "INTERNAL MESSAGE OVERFLOWED [>%d]\n", __DebugMax);
# endif /* WIN32 */
  
#endif /* NDEBUG */
  return 1;
}

