/*
 *  $Id: zkDir.c,v 1.5.20.1 2000/10/11 13:20:02 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"
#include "zkIntTypes.h"
#include "zkTypes.h"

#include <sys/types.h>

#ifndef WIN32
#include <dirent.h>
#else
#include <io.h>
#endif

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include "zkList.h"
#include "zkDir.h"

#ifdef WIN32
#pragma message(Reminder "These are the new directory routines.")
#endif


struct zk_dir {
#ifdef WIN32
  int got_first; /* 0: first entry has not been seen; 1: it has */
  long d; /* directory handle */
  struct _finddata_t file; /* current directory entry */
#else
  DIR *d; /* directory handle */
  struct dirent *dent; /* current directory entry */
#endif
};


static int dirInUse(ZkDir dir)
{
#ifdef WIN32

  return dir->d != -1;

#else

  return !! dir->d;
  
#endif
}


static void dirInit(ZkDir dir)
{
#ifdef WIN32

  memset(dir, 0, sizeof *dir);

#else

  dir->d = 0;
  dir->dent = 0;
  
#endif
}


DLLExport int zkDirNew(ZkError err, ZkDir *dir)
{
  struct zk_dir *d;

  zkErrorClear(err);

  if ( ! dir) {
    zkErrorSet(err, "zkDirNew", ZKERR, ZKMDIR, ZKENULLPTR,
               "dir is null");
    return 0;
  }

  if ( ! (d = malloc(sizeof *d))) {
    zkErrorSet_va(err, "zkDirNew", ZKERR, ZKMDIR, errno,
                  "can't allocate directory object: %s",
                  strerror(errno));
    return 0;
  }

  dirInit(d);
  *dir = d;

  return 1;
}


DLLExport int zkDirFree(ZkError err, ZkDir *dir)
{
  zkErrorClear(err);

  if ( ! dir) {
    zkErrorSet(err, "zkDirFree", ZKERR, ZKMDIR, ZKENULLPTR,
               "dir is null");
    return 0;
  }
  if ( ! *dir) {
    zkErrorSet(err, "zkDirFree", ZKERR, ZKMDIR, ZKENULLPTR,
               "*dir is null");
    return 0;
  }

  if (dirInUse(*dir)) {
    (void)zkDirClose(err, *dir);
  }

  free(*dir); *dir = 0;

  return 1;
}


DLLExport int zkDirOpen(ZkError err, ZkDir dir, const char *path)
{
  char *p = 0;
  int ret = 0;

  zkErrorClear(err);

  if ( ! dir) {
    zkErrorSet(err, "zkDirOpen", ZKERR, ZKMDIR, ZKENULLPTR,
               "dir is null");
    return 0;
  }

  if (dirInUse(dir)) {
    zkErrorSet(err, "zkDirOpen", ZKERR, ZKMDIR, ZKEBADSTATE,
               "directory is already open");
    return 0;
  }

#ifdef WIN32
  
  if (zkCStrSprintf(&p, "%s\\*.*", dir) == -1 ||
      (dir->d = _findfirst(p, &dir->file)) == -1) {
    zkErrorSet(err, "zkDirOpen", ZKERR, ZKMDIR, ZKERROR,
               "can't get first directory entry");
    goto done;
  }

#else

  if ( ! (dir->d = opendir(path))) {
    zkErrorSet_va(err, "zkDirOpen", ZKERR, ZKMDIR, errno,
                  "can't open `%s': %s", path, strerror(errno));
    goto done;
  }

#endif

  ret = 1;

 done:

  if (p) { free(p); }

  return ret;
}


DLLExport int zkDirClose(ZkError err, ZkDir dir)
{
  zkErrorClear(err);

  if ( ! dir) {
    zkErrorSet(err, "zkDirClose", ZKERR, ZKMDIR, ZKENULLPTR,
               "dir is null");
    return 0;
  }

  if ( ! dirInUse(dir)) {
    zkErrorSet(err, "zkDirClose", ZKERR, ZKMDIR, ZKENULLPTR,
               "dir has not been opened");
    return 0;
  }

#ifdef WIN32

  _findclose(dir->d);

#else

  (void)closedir(dir->d);
  
#endif

  dirInit(dir);

  return 1;
}

DLLExport int zkDirNext(ZkError err, ZkDir dir)
{
  zkErrorClear(err);

  if ( ! dir) {
    zkErrorSet(err, "zkDirNext", ZKERR, ZKMDIR, ZKENULLPTR,
               "dir is null");
    return 0;
  }

  if ( ! dirInUse(dir)) {
    zkErrorSet(err, "zkDirNext", ZKERR, ZKMDIR, ZKENULLPTR,
               "directory has not been opened");
    return 0;
  }

#ifdef WIN32

  if ( ! dir->got_first) {
    /*
     *  We got the first one when we opened the directory.
     */
    dir->got_first = 1;
    return 1;
  }
  
  return _findnext(dir->d, &dir->file) == 0;

#else

  return !! (dir->dent = readdir(dir->d));
  
#endif
}


DLLExport int zkDirGetFileName(ZkError err, ZkDir dir, const char **name)
{
  zkErrorClear(err);

  if ( ! dir) {
    zkErrorSet(err, "zkDirGetFileName", ZKERR, ZKMDIR, ZKENULLPTR,
               "dir is null");
    return 0;
  }

  if ( ! dirInUse(dir)) {
    zkErrorSet(err, "zkDirGetFileName", ZKERR, ZKMDIR, ZKEBADSTATE,
               "directory has not been opened");
    return 0;
  }

#ifdef WIN32

  *name = dir->file.name;
  
#else

  *name = dir->dent->d_name;

#endif

  return 1;
}


/*
 * Scan a directory and push the list of filename found 
 *
 * REM: The list datas MUST be always Free using zkDirListEntriesFree
 */
DLLExport int zkDirListEntries
 (
  ZkError      err,                    /* [in]     ZkError*/
  const char * dirName,                /* [in]     name of the directory to be scan */
  ZkList     * listP                   /* [out]    list of entries */
 )
{
  const char * func    = "zkDirGetFileName";
  ZkBool       success = ZK_TRUE; 
  ZkDir        dir     = NULL;
  ZkBool       eol     = ZK_FALSE; 

  zkErrorClear(err);

  /* check input values */
  if( !dirName ) {
    success = ZK_FALSE;
    zkErrorSet(err, func, ZKERR, ZKMDIR, ZKENULLPTR,
               "dirName is null");
  }
  else if( !dirName ) {
    success = ZK_FALSE;
    zkErrorSet(err, func, ZKERR, ZKMDIR, ZKENULLPTR,
               "dirName is null");
  }

  if( success ) {
    /* Create the list use to list the directory */
    if( !(success = zkListNew(err, listP)) ) {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "zkListNew failed");
    }
    else if( !(success = zkListSetDataFreeFunction(err, *listP, zkListDataFree, NULL)) ) {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "zkListSetDataFreeFunction failed");
    }
  }

  /* Create the directory object*/
  if( success ) {
    if( !(success = zkDirNew(err, &dir)) ) {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "zkDirNew failed");
    }
  }

  if( success ){
    /* Open the directory */
    if( !(success = zkDirOpen(err, dir, dirName)) ) {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "zkDirOpen failed");
    }
    else {
      /* Scan the entries */
      while( success && !eol ) {
        
        if( !zkDirNext(err, dir) ) {
          eol = ZK_TRUE;
        }
        else {
          const char * filename;

          if( !(success = zkDirGetFileName(err, dir, &filename)) ) {
            zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "zkDirGetFileName failed");
          }
          else {
            char * dup = NULL;

            /* Added the entry to the list */
            if( (dup = strdup(filename)) != NULL ) {
              if( !(success = zkListInsert(err, *listP, dup, LIST_TAIL)) ) {
                zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "zkListInsert failed");
              }
            }
            if( !success && dup ) {
              free(dup);
            }
          }
        }
        /* >> End of list is not an error */
        if( eol ) {
          success = ZK_TRUE;
        }
      }
      zkDirClose(NULL, dir);      
    }    
    zkDirFree(NULL, &dir);
  }
  return success;
}



DLLExport int zkDirListEntriesFree
 (
  ZkError      err,                    /* [in]     ZkError*/
  ZkList     * listP                   /* [in] list to append the entries */
 )
{
  return zkListFree(err, listP);
}

