/*
 *  $Id: zkFile.c,v 1.22.16.5 2000/12/14 17:54:28 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#define DEBUG_READ 0

#include "zks.h"

#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "zkAlloc.h"
#include "zkCStr.h"
#include "zkFile.h"


#ifdef WIN32
#define DEV_NULL "NUL:"
#else
#define DEV_NULL "/dev/null"
#include <sys/types.h>
#endif


/*
 *  Set `*len' to the length of the file specified by `fp'.
 *  
 *  Note that values may be misleading if the file is simultaneously being
 *  written to.  It will probably also fail if `fp' doesn't represent a plain
 *  file.
 */
DLLExport int zkFileLengthFp(ZkError err, FILE *fp, size_t *len)
{
  long here, l;

  zkErrorClear(err);

  if ((here = ftell(fp)) == -1 ||
      fseek(fp, 0, SEEK_END) == -1 ||
      (l = ftell(fp)) == -1 ||
      fseek(fp, here, SEEK_SET) == -1) {
    zkErrorSet_va(err, "zkFileLengthFp", ZKERR, ZKMFILE, errno,
               "can't get file size: %s", strerror(errno));
    return 0;
  }

  *len = l;
  return 1;
}

/**
 *  Load a file.  The function will allocate a buffer large enough to 
 *  accomodate the contents the of the file (be sure to free() it when you
 *  no longer need it).  The number of bytes in the buffer is returned by
 *  size.  There is always a NUL byte appended to the end of the buffer (right
 *  after the last real byte) This NUL is not part of the data.
 *
 *  @param err Returns the error if any occured
 *  @param filename The file to load
 *  @param buffer Returns a buffer containing the file data
 *  @param size Returns the number of bytes in the file (not counting the 
 *              last NUL)
 *  @return A non-zero value on success, a 0 on failure
 *  @author Philippe Boucher
 *  @version 1.0
 */

DLLExport int
zkFileLoad (
  ZkError err ,
  const char * filename ,
  char ** buffer ,
  size_t * size )
{
  int fd = 0 ;  /* File descriptor */
  int bytes = 0; /* Number of bytes read */
  struct stat sbuf;

  zkErrorClear(err);
  if (filename && buffer && size) {
    if (stat(filename,&sbuf) != 0) {
      zkErrorSet_va(err,"zkFileLoad",ZKERR,ZKMFILE,errno,"%s: %s",filename,
                 strerror(errno));
      return 0;
    }
    *buffer = (char *) malloc(sbuf.st_size + 1);
    if (*buffer == NULL) {
      zkErrorSet(err,"zkFileLoad",ZKERR,ZKMFILE,ZKENOMEM,"malloc");
      return 0; 
    }
    if ((fd = open(filename,O_RDONLY)) < 0) {
      zkErrorSet_va(err,"zkFileLoad",ZKERR,ZKMFILE,errno,"%s: %s",filename,
                   strerror(errno));
      free(*buffer);
      *buffer = NULL;
      return 0;
    }
    *size = 0;
    while (*size < (size_t) sbuf.st_size) {
      bytes = read(fd,&((*buffer)[*size]),sbuf.st_size - *size);
      if (bytes == 0) {
	break;
      }
      if (bytes < 0) {
	zkErrorSet_va(err,"zkFileLoad",ZKERR,ZKMFILE,errno,"reading %s: %s",
		      filename, strerror(errno));
	free(*buffer);
	close (fd);
	return 0;
        break;
      }
      *size += bytes;
    }
    (*buffer)[*size] = (char) 0;
    close(fd);
    return 1;
  }
  zkErrorSet(err,"zkFileLoad",ZKERR,ZKMFILE,ZKENULLPTR,NULL);
  return 0;
}

/*#ifndef WIN32*/

/**
 *  Save a file.
 *
 *  @param err       Returns the error if any occured
 *  @param filename  The file to save
 *  @param buffer    a buffer containing the file data
 *  @param size      the size of the data to save
 *  @param flags     flags as used by open(2)
 *  @return A non-zero value on success, a 0 on failure
 *  @author Luc Boulianne
 *  @version 1.0
 */

DLLExport int
zkFileSave (
  ZkError err ,
  const char * filename ,
  const char * buffer ,
  size_t  size ,
  int flags,
  mode_t mode)
{
  static const char *funcname = "zkFileSave";

  int fd = 0 ;  /* File descriptor */
  int bytes = 0; /* Number of bytes read */
  size_t count = 0;

  zkErrorClear(err);
  if (filename && buffer && size) {
    if ((fd = open(filename,O_WRONLY|flags, mode)) < 0) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMFILE,errno,"open: %s: %s",filename,
                   strerror(errno));
      return 0;
    }
    while (count < size) {
      bytes = write(fd,buffer+count,size-count);
      if (bytes == 0) {
	break;
      }
      if (bytes < 0) {
	zkErrorSet_va(err,funcname,ZKERR,ZKMFILE,errno,"writing %s: %s",
		      filename, strerror(errno));
	close (fd);
	return 0;
        break;
      }
      count += bytes;
    }
    close(fd);
    return 1;
  }
  zkErrorSet(err,"`zkFileLoad",ZKERR,ZKMFILE,ZKENULLPTR,NULL);
  return 0;
}

/*#endif*/

/*  
 *  Set `*null' to a file pointer openned on the the null device.  This
 *  function allows us to open it many times while only using one file
 *  pointer.
 *  
 *  Bugs: we keep returning the same pointer, even after the user closes the
 *  file.
 */
DLLExport FILE *zkNullFp(void)
{
  static FILE *fp = 0;

  if ( ! fp) {
    fp = fopen(DEV_NULL, "wb");
  }

  return fp;
}


static int alloc(ZkError err, char **p, size_t n)
{
  zkErrorClear(err);

  if (*p) {
    return zkMemRealloc(err, (void **)p, n);
  } else if ( ! (*p = malloc(n))) {
    zkErrorSet_va(err, "alloc", ZKERR, ZKMFILE, errno,
                  "can't allocate %lu bytes: %s",
                  (unsigned long)n, strerror(errno));
    return 0;
  }

  return 1;
}


/*
 *  Try to open a file, locking it either for reading or writing, depending
 *  upon the value of `mode'.  The lock will go away when the file is closed.
 *  
 *  A special failure case is when the file is already locked, in which case a
 *  warning is returned, and the error code is set to EAGAIN.  In this case
 *  the caller may want to retry the open.
 */
DLLExport int zkFileFopenLocked(ZkError err, const char *path, const char *mode, FILE **fp)
{
  FILE *f;
  int ret = 0;
#ifndef WIN32
  mode_t oumask;
#endif

  zkErrorClear(err);

#ifndef WIN32
  oumask = umask(0177);
#endif
  
  if ( ! (f = fopen(path, mode))) {
    zkErrorSet_va(err, "zkFileFopenLocked", ZKERR, ZKMFRSEC, errno,
                  "can't open `%s' with mode `%s': %s",
                  path, mode, strerror(errno));
    goto done;
  }

#ifndef WIN32
  {
    struct flock lck;

    switch (mode[0]) {
    case 'a': lck.l_type = F_WRLCK; break;
    case 'r': lck.l_type = F_RDLCK; break;
    case 'w': lck.l_type = F_WRLCK; break;
    default:
      zkErrorSet_va(err, "zkFileFopenLocked", ZKERR, ZKMFRSEC, ZKEPARMBAD,
                    "unexpected first character in mode `%s'", mode);
      goto done;
    }

    if (strchr(mode, '+')) { lck.l_type = F_WRLCK; }

    lck.l_start = 0;
    lck.l_whence = SEEK_SET;
    lck.l_len = 0;

    if (fcntl(fileno(f), F_SETLK, &lck) == -1) {
      unsigned short errorCode;
      
      if (errno == EACCES || errno == EAGAIN) {
        errno = EAGAIN;
        errorCode = ZKWARN;
      } else {
        errorCode = ZKERR;
      }
      zkErrorSet_va(err, "zkFileFopenLocked", errorCode, ZKMFRSEC, errno,
                    "can't lock `%s' for %s: %s",
                    path, lck.l_type == F_RDLCK ? "reading" : "writing",
                    strerror(errno));
      goto done;
    }
  }
#endif

  *fp = f; f = 0;
  ret = 1;

 done:

#ifndef WIN32
  (void)umask(oumask);
#endif
  
  if (f) { fclose(f); }

  return ret;
}


/*  
 *  Set `*ln' to point to memory containing the next input line.  The buffer
 *  will *not* contain the terminating newline.
 *  
 *  If `*ln' is non-null it is assumed to point to a malloc()ed region of
 *  memory to which the line is copied.  If necessary, the memory will
 *  realloc()ed, causing the value of `*ln' to change.  If `*ln' is null, the
 *  original memory buffer will be allocated by this routine.  In any case, it
 *  is up to the caller to free the memory.
 *  
 *  Upon a successful return, `*l' will be set to the length of the line,
 *  which may contain embedded nuls.
 *  
 *  This routine returns a warning if EOF is immediately true, and an error if
 *  enough memory can't be allocated for the line, or if an error occurs while
 *  reading.
 *  
 *  The return value is 0 in the case of an error or a warning, 1 if
 *  successful.
 *  
 *  Note: Since we read (and realloc()) one character at a time, this is not
 *  the most efficient routine in the world.  However, it allows us to easily
 *  keep track of how many bytes we've read (in the presence of nuls), without
 *  doing our own buffered I/O.
 */
DLLExport int zkFileGetLineFp(ZkError err, FILE *fp, char **ln, size_t *l)
{
  int c;
  char *line = *ln;
  size_t len = 0; /* current line length (not incl. nul) */

  zkErrorClear(err);

  if ( ! ln) {
    zkErrorSet(err, "zkFileGetLineFp", ZKERR, ZKMFILE, ZKENULLPTR,
               "ln is null");
    return 0;
  }
  if ( ! l) {
    zkErrorSet(err, "zkFileGetLineFp", ZKERR, ZKMFILE, ZKENULLPTR,
               "l is null");
    return 0;
  }
  if ( ! fp) {
    zkErrorSet(err, "zkFileGetLineFp", ZKERR, ZKMFILE, ZKENULLPTR,
               "fp is null");
    return 0;
  }

  do { 
    c = getc(fp); 
    if ( ! alloc(err, &line, len + 1)) { goto fail; } 
    line[len++] = c; 
  } while ( ! (c == EOF || c == '\n')); 
   
  if (c == EOF && ferror(fp)) { 
    zkErrorSet_va(err, "zkFileGetLineFp", ZKERR, ZKMFILE, errno, 
                  "error reading line: %s", strerror(errno)); 
    goto fail; 
  } else if (c == EOF && len == 1) { 
    zkErrorSet(err, "zkFileGetLineFp", ZKWARN, ZKMFILE, ZKEEMPTY, 
               "immediate end of file"); 
    goto fail; 
  } 
 
  line[len-1] = '\0'; 
  *ln = line; *l = len - 1; 
  return 1; 
 
 fail: 
 
  /*
   *  Even though we've failed, the caller may have given us the buffer to
   *  fill.  If so, it may have been already freed by the call to alloc(), in
   *  which case we need to update it with `line'.
   */

  if (line) { 
    if (*ln) { *ln = line; }
    else { free(line); } 
  } 
  return 0; 
}


/*
 *  Read from `fp' until `max' bytes have been read, or until end-of-file.  If
 *  `max' is 0, it is considered to be unlimited.
 *  
 *  The data read, and its size, are returned via `*buf' and `*bsz'
 *  respectively.
 *  
 *  If end-of-file is encountered immediately `*bsz' will be 0, and `*buf'
 *  will be a valid pointer, although the size of the allocated memory, and
 *  its contents, are undefined.
 */
DLLExport int zkFileReadFp(ZkError err, FILE *fp, size_t max, void **buf, size_t *bsz)
{
  char tmp[1024];
  int ret = 0;
  size_t n, psz;
  void *p = 0;
  
  zkErrorClear(err);

#if DEBUG_READ
  fprintf(stderr, "zkFileReadFp: enter.\n");
#endif

  if ( ! fp) {
    zkErrorSet(err, "zkFileReadFp", ZKERR, ZKMFILE, ZKENULLPTR,
               "fp is null");
    return 0;
  }
  if ( ! buf) {
    zkErrorSet(err, "zkFileReadFp", ZKERR, ZKMFILE, ZKENULLPTR,
               "buf is null");
    return 0;
  }
  if ( ! bsz) {
    zkErrorSet(err, "zkFileReadFp", ZKERR, ZKMFILE, ZKENULLPTR,
               "bsz is null");
    return 0;
  }

  psz = 0;

  if ( ! (p = malloc(1))) {
    zkErrorSet_va(err, "zkFileReadFp", ZKERR, ZKMFILE, errno,
                  "can't allocate 1 byte buffer: %s", strerror(errno));
    goto done;
  }

  while ((max ? psz < max : 1) &&
         (n = fread(tmp, 1, sizeof tmp, fp)) > 0) {
#if DEBUG_READ
    fprintf(stderr, "zkFileReadFp: read %lu bytes.\n", (unsigned long)n);
#endif
    if ( ! zkMemRealloc(err, &p, psz + n)) {
      goto done;
    }
    memcpy((char *)p + psz, tmp, n);
    psz += n;
  }

  if (n == 0 && ferror(fp)) {
    zkErrorSet_va(err, "zkFileReadFp", ZKERR, ZKMFILE, errno,
                  "error reading: %s", strerror(errno));
    goto done;
  }

  *bsz = psz;
  *buf = p; p = 0; /* the caller now owns it */

  ret = 1;

 done:

  if (p) { memset(p, 0, psz); free(p); }
  memset(tmp, 0, sizeof tmp);
  
#if DEBUG_READ
  fprintf(stderr, "zkFileReadFp: leave.\n");
#endif

  return ret;
}


/*
 *  Write `dsz' bytes of `data' to `fp'.  Return only when all the data has
 *  been written, or an when error occurs.
 */
DLLExport int zkFileWriteFp(ZkError err, FILE *fp, void *data, size_t dsz)
{
  char *r = data; /* remainder of data */
  const char *fn = "zkFileWriteFp";
  size_t rem = dsz; /* remainder to be written */

  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFILE, ZKENULLPTR, "fp is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFILE, ZKENULLPTR, "data is null");
    return 0;
  }

  while (rem > 0) {
    size_t n;

    if ((n = fwrite(r, 1, rem, fp)) == 0) {
      return 0;
    }

    rem -= n;
    r += n;
  }

  return 1;
}


/*
 *  Try to open a PID file with a write lock.  If we fail, it may be due to
 *  another instance of this server running.  If successful, we write our
 *  process ID to the file and we don't close the file pointer, in order to
 *  preserve our lock.
 */
DLLExport int zkFileWritePid(ZkError err, const char *pidFile, FILE **fptr)
{ 
  const char *func = "zkFileWritePid";
  FILE *fp = 0; 
  int ret = 0;

  zkErrorClear(err);

  if ( ! pidFile) {
    zkErrorSet(err, func, ZKERR, ZKMFILE, ZKENULLPTR, "pidFile is null");
    return 0;
  }
  if ( ! fptr) {
    zkErrorSet(err, func, ZKERR, ZKMFILE, ZKENULLPTR, "fptr is null");
    return 0;
  }

  if ( ! zkFileFopenLocked(err, pidFile, "w", &fp)) {
    if (zkErrorType(err) == ZKWARN && zkErrorCode(err) == EAGAIN) {
      ZkLog log; 

      zkErrorLog(err, &log); 
      zkErrorSet(err, func, ZKERR, ZKMFILE, ZKEBADSTATE,
                 "an instance of this server is already running");
    }
    goto done;
  }

  if (fprintf(fp, "%ld\n", (long)getpid()) < 0 || fflush(fp) == EOF) {
    zkErrorSet_va(err, func, ZKERR, ZKMFILE, errno,
                  "couldn't write to PID file: %s", strerror(errno));
    goto done;
  }

  *fptr = fp; fp = 0;
  ret = 1;

 done:

  if (fp) { fclose(fp); }

  return ret;
}


DLLExport int zkFprintf(ZkError err, FILE *fp, const char *fmt, ...)
{
  const char *fn = "zkFprintf";
  int r;
  va_list ap;

  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFILE, ZKENULLPTR, "fp is null");
    return 0;
  }
  if ( ! fmt) {
    zkErrorSet(err, fn, ZKERR, ZKMFILE, ZKENULLPTR, "fmt is null");
    return 0;
  }

  va_start(ap, fmt);
  errno = 0;
  if ((r = vfprintf(fp, fmt, ap)) == -1) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFILE, errno, "%s", strerror(errno));
  }
  va_end(ap);

  return r != -1;
}


/*
 *  Return a file pointer, opened for reading and writing, on a temporary
 *  file.  The actual file name is based on the `tmpl' argument, which
 *  will be rewritten.  The template must contain six sequential Xs, which are
 *  placeholders for the distinguishing characters of the name.  It is up to
 *  the caller to remove the file.
 *  
 *  Because `tmpl' will be rewritten it cannot be a string constant.
 */
DLLExport int zkFileTempFp(ZkError err, char *tmpl, FILE **tempFp)
{
#ifndef __unix__
  const char *fn = "zkFileTempFp";

  zkErrorSet(err, fn, ZKERR, ZKMFILE, ZKESTUB, "not implemented");
  return 0;

#else

#define NO_MASK (0777 + 1)

  FILE *fp = 0;
  const char *fn = "zkFileTempFp";
  int fd = -1;
  int ret = 0;
  mode_t omask = NO_MASK;

  if ( ! tmpl) {
    zkErrorSet(err, fn, ZKERR, ZKMFILE, ZKENULLPTR, "template is null");
    goto done;
  }

  if ( ! tempFp) {
    zkErrorSet(err, fn, ZKERR, ZKMFILE, ZKENULLPTR, "tempFp is null");
    goto done;
  }

  /*
   *  Set the umask in case our version of mkstemp() doesn't do it.
   */
  
  omask = umask(600);
  
  if ((fd = mkstemp(tmpl)) == -1) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFILE, errno,
                  "error creating temporary file using template `%s': %s",
                  tmpl, strerror(errno));
    goto done;
  }

  if ( ! (fp = fdopen(fd, "r+b"))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFILE, errno,
                  "error making file pointer from descriptor on `%s': %s",
                  tmpl, strerror(errno));
    goto done;
  }    

  *tempFp = fp; fp = 0; fd = -1; /* caller now owns it */

  ret = 1;

 done:

  if (fp) { fclose(fp); }
  if (fd != -1) { close(fd); }
  if (omask != NO_MASK) { umask(omask); }

  return ret;

#define NO_MASK (0777 + 1)

#endif
}
