/*
 * Basic graph datastructures and algorithms
 *
 * $Id: zkGraph.c,v 1.12.16.1 2000/10/11 13:20:02 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zks.net>

 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "zkIntTypes.h"
#include "zkIntLimits.h"
#include "zkError.h"
#include "zkDebug.h"
#include "zkGraph.h"
#include "zkHash.h"
#include "zkAlloc.h"
#include "zkList.h"

#define HASH_PRIME 1001

typedef 
struct T_ZkGraphNode {
  uint32_t id;                  /* ID of this node */
  int weight;                   /* Weight of the edge to previous node */
  struct T_ZkGraphNode *next;   /* Next node connected edge */
} *ZkGraphNode;

struct T_ZkGraph {
  size_t numnodes;              /* Number of nodes in this graph */
  ZkHash h;                     /* Hash table of starting node */
  ZkGraphNode *node;            /* Array of nodes */
};

typedef
enum {
  ZKGRAPH_COLOUR_NONE,
  ZKGRAPH_COLOUR_WHITE,
  ZKGRAPH_COLOUR_GREY,
  ZKGRAPH_COLOUR_BLACK
} ZkGraphColour;

#define ZKGRAPH_DIST_INFINITY   -1

/*
 * Create a graph node.
 */
static int zkGraphNodeNew(ZkError err, ZkGraphNode *node)
{
  const char *func = "zkGraphNodeNew";

  zkErrorClear(err);
  if (!node) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL param");
    return 0;
  }

  if ((*node = (ZkGraphNode) malloc(sizeof(struct T_ZkGraphNode))) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENOMEM, "malloc() failure");
    return 0;
  }

  (*node)->id = 0;
  (*node)->weight = 0;
  (*node)->next = NULL;

  return 1;
}

static size_t hash_mod(ZkHash h, void *key, void *userData)
{
  uint32_t *p = key;
  size_t v;

  (void) userData;
  (void) h;

  v = *p % HASH_PRIME;
  
  return v;  
}

/*
 * Create a new graph.
 */
DLLExport int zkGraphNew(ZkError err, ZkGraph *graph)
{
  static const char *func = "zkGraphNew";

  zkErrorClear(err);
  if (!graph) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if ((*graph = malloc(sizeof(struct T_ZkGraph))) == NULL) {
    zkErrorSet_va(err, func, ZKERR, ZKMGRAPH, ZKENOMEM,
		  "calloc returs errno [%d - %s]", errno, strerror(errno));
    return 0;
  }

  /* Set up a hash table */
  if (!zkHashNew(err, &((*graph)->h))) {
    return 0;
  }

  if (!zkHashSetHashFn(err, (*graph)->h, hash_mod)) {
    return 0;
  }
  
  if (!zkHashSetKeySize(err, (*graph)->h, sizeof(uint32_t))) {
    return 0;
  }
  if (!zkHashSetTableSize(err, (*graph)->h, HASH_PRIME)) {
    return 0;
  }
  if (!zkHashSetUserData(err, (*graph)->h, NULL)) {
    return 0;
  }

  (*graph)->numnodes = 0;
  (*graph)->node = NULL;

  return 1;
}

/* 
 * Starting a node, recursively find the next node and free it.
 */
static int zkGraphNodeFree(ZkError err, ZkGraphNode *node)
{
  const char *func = "zkGraphFreeNodes";
  
  zkErrorClear(err);
  if (!node || !*node) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL param");
    return 0;
  }

  if ((*node)->next) {
    (void) zkGraphNodeFree(err, &((*node)->next));
    (*node)->next = NULL;
  }

  free(*node);
  *node = NULL;
  
  return 1;
}

/**
 * Free the graph object.
 */
DLLExport int zkGraphFree(ZkError err, ZkGraph *graph)
{
  const char *func = "zkGraphFree";
  size_t i;

  zkErrorClear(err);
  if (!graph || !*graph) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL param");
    return 0;
  }

  for (i = 0; i < (*graph)->numnodes; i++) {
    if (((*graph)->node[i])) {
      (void) zkGraphNodeFree(err, &((*graph)->node[i]));
    }
  }
  if ((*graph)->node) {
    free((*graph)->node);
    (*graph)->node = NULL;
  }

  (void) zkHashFree(err, &((*graph)->h));

  free(*graph);
  *graph = NULL;
  
  return 1;
}

/*
 * Return the number of nodes currently in the adjecency list.
 */
DLLExport int zkGraphTotalNodes(ZkError err, ZkGraph graph, size_t *numnodes)
{
  const char *func = "zkGraphTotalNodes";

  zkErrorClear(err);
  if (!graph || !numnodes) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  *numnodes = graph->numnodes;

  return 1;
}

/**
 * Add a node to the graph
 */
DLLExport int zkGraphAddNode(ZkError err, ZkGraph graph, uint32_t id)
{
  const char *func = "zkGraphAddNode";
  void *data = NULL;
  
  zkErrorClear(err);
  if (!graph) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (zkHashSearch(NULL, graph->h, &id, &data)) {
    /* Node already in my hash table */
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKEEXIST, "Node already exists");
    return 0;
  }

  if (!graph->node) {           /* Allocate 1 element array */
    if ((graph->node = calloc(1, sizeof(struct T_ZkGraphNode))) == NULL) {
      zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENOMEM, "malloc() failed");
      return 0;
    }
  } else {                      /* Reallocate a new element */
    if (!zkMemRealloc(err, (void *) &(graph->node), 
                   sizeof(struct T_ZkGraphNode) * (graph->numnodes + 1))) {
      return 0;
    }
  }

  if (!zkGraphNodeNew(err, &(graph->node[graph->numnodes++]))) {
    return 0;
  }

  graph->node[graph->numnodes - 1]->id = id;
  /* Overload the weight to indicated the array index for this node */
  graph->node[graph->numnodes - 1]->weight = graph->numnodes - 1;

  /* Shove this into the hash table */
  if (!zkHashInsert(err, graph->h, &(graph->node[graph->numnodes - 1]->id),
                    graph->node[graph->numnodes - 1])) {
    return 0;
  }

  return 1;
}

/**
 * Add an edge to the graph.  Use the ID for the source and destination of the
 * edge and assign the given weight to the edge.  The source and destination
 * nodes must exist otherwise it is an error.
 */
DLLExport int zkGraphDirectedEdgeAdd(ZkError err, ZkGraph graph, uint32_t s, 
                                     uint32_t d, int w)
{
  const char *func = "zkGraphDirectedEdgeAdd";
  ZkGraphNode node = NULL;
  int found = 0;

  zkErrorClear(err);
  if (!graph) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL param");
    return 0;
  }

  /* Node not found */
  if (!zkHashSearch(err, graph->h, &s, (void *) &node)) {
    return 0;
  }

  while (node->next) {
    if (node->next->id == d) {
      found = 1;
      break;
    }
    node = node->next;
  }

  if (!found) {
    if (!zkGraphNodeNew(err, &node->next)) {
      return 0;
    }
    node->next->weight = w;
    node->next->id = d;
  }

  return 1;
}

DLLExport int zkGraphEdgeAdd(ZkError err, ZkGraph graph, uint32_t s, 
                             uint32_t d, int w)
{
  const char *func = "zkGraphEdgeAdd";
  void *node = NULL;            /* dummy variable */

  if (!graph) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (!zkHashSearch(err, graph->h, &s, (void *) &node) ||
      !zkHashSearch(err, graph->h, &d, (void *) &node)) {
    return 0;
  }

  return(zkGraphDirectedEdgeAdd(err, graph, s, d, w) && 
         zkGraphDirectedEdgeAdd(err, graph, d, s, w));
}

/**
 * Look at the directed graph and check if s -> d.  Set the variable
 * connected to 1 if an edge exists, 0 otherwise.
 */
DLLExport int zkGraphNodesDirectedConnected(ZkError err, ZkGraph graph, 
                                            uint32_t s, uint32_t d, 
                                            int *connected)
{
  const char *func = "zkGraphNodesDirectedConnected";
  ZkGraphNode node = NULL;

  zkErrorClear(err);
  if (!graph || !connected) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  *connected = 0;

  if (!zkHashSearch(err, graph->h, &s, (void *) &node)) {
    return 0;
  }

  while (node->next) {
    if (node->next->id == d) {
      *connected = 1;
      break;
    }
    node = node->next;
  }

  return 1;
}
/*
 * Check the graph to see if an edge exists between nodes s and d.  Set
 * connected to 1 if there is an edge, 0 otherwise.
 */
DLLExport int zkGraphNodesConnected(ZkError err, ZkGraph graph, uint32_t s,
                                    uint32_t d, int *connected)
{
  const char *func = "zkGraphNodesConnected";
  void *node = NULL;            /* dummy variable */

  zkErrorClear(err);
  if (!graph || !connected) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  *connected = 0;
  if (!zkHashSearch(err, graph->h, &s, (void *) &node) ||
      !zkHashSearch(err, graph->h, &d, (void *) &node)) {
    return 0;                   /* Nodes not found */
  }

  if (!zkGraphNodesDirectedConnected(err, graph, s, d, connected)) {
    return 0;
  }
  if (*connected) {
    if (!zkGraphNodesDirectedConnected(err, graph, d, s, connected)) {
      return 0;
    }
  }

  return 1;
}

/*
 * Given a graph and the predicate list, check if there is a path between 
 * the s node and the d node.  
 */
DLLExport int zkGraphPathExists(ZkError err, ZkGraph graph, uint32_t s,
                                uint32_t d, uint32_t *pi, int *yes)
{
  const char *func = "zkGraphPathExists";
  ZkGraphNode node = NULL;

  zkErrorClear(err);
  if (!graph || !yes || !pi) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (d == s) {
    *yes = 1;
  } else {
    if (!zkHashSearch(err, graph->h, &d, (void *) &node)) {
      return 0; 
    }
    if (pi[node->weight] == (graph->numnodes + 1)) {
      *yes = 0;
    } else {
      if (!zkGraphPathExists(err, graph, s, pi[node->weight], pi, yes)) {
        return 0;
      }
    }
  }
  return 1;
}

/* 
 * Breadth first search of a graph given a starting point, return all paths
 * in an array.
 */ 
DLLExport int zkGraphBFS(ZkError err, ZkGraph graph, uint32_t s, uint32_t **p)
{
  const char *func = "zkGraphBFS";
  ZkList q = NULL;              /* Treat the list like a queue */
  uint32_t *colour = NULL;
  int *d = NULL;
  uint32_t *pi = NULL;
  int ret = 0;
  size_t i;
  size_t sz;
  ZkGraphNode node = NULL;
  ZkGraphNode u = NULL;
  ZkGraphNode v = NULL;

  if (!graph || !p) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL ptr(s)");
    return 0;
  }

  if ((colour = calloc(graph->numnodes, sizeof(uint32_t))) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENOMEM, "No memory for colour");
    return 0;
  }
  if ((d = calloc(graph->numnodes, sizeof(uint32_t))) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENOMEM, "No memory for d");
    goto colourcleanup;
  }
  if ((pi = calloc(graph->numnodes, sizeof(uint32_t))) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENOMEM, "No memory for pi");
    goto colourcleanup;
  }

  for (i = 0; i < graph->numnodes; i++) {
    colour[i] = ZKGRAPH_COLOUR_WHITE;
    d[i] = ZKGRAPH_DIST_INFINITY;
    pi[i] = graph->numnodes + 1; /* Infinity */
  }

  /* Get my source ID */
  if (!zkHashSearch(err, graph->h, &s, (void *) &node)) {
    goto colourcleanup;
  }

  /* Set up my starting node as visited */
  colour[node->weight] = ZKGRAPH_COLOUR_GREY;
  d[node->weight] = 0;
  pi[node->weight] = graph->numnodes + 1; /* Infinity */

  /* Set up my queue of nodes */
  if (!zkListNew(err, &q)) {
    goto colourcleanup;
  }
  /* Don't allow the list to clean up my data */
  if (!zkListSetDataFreeFunction(err, q, NULL, NULL)) {
    goto colourcleanup;
  }

  if (!zkListInsert(err, q, node, LIST_TAIL)) {
    goto colourcleanup;
  }

  while (zkListSize(err, q, &sz) && (sz != 0)) {
    if (!zkListMoveCursor(err, q, LIST_HEAD) ||
	!zkListGet(err, q, (void *) &u, LIST_HEAD)) {
      goto colourcleanup;
    }

    v = u;
    while (v->next) {
      if (!zkHashSearch(err, graph->h, &(v->next->id), (void *) &node)) {
        goto colourcleanup;
      }
      
      if (colour[node->weight] == ZKGRAPH_COLOUR_WHITE) {
        colour[node->weight] = ZKGRAPH_COLOUR_GREY;
        d[node->weight] = d[u->weight] + 1;
        pi[node->weight] = u->id;
        if (!zkListInsert(err, q, node, LIST_TAIL)) {
          goto colourcleanup;
        }
      }
      
      v = v->next;
    }

    /* Dequeue */
    if (!zkListMoveCursor(err, q, LIST_HEAD) || 
        !zkListDelete(err, q, LIST_HEAD)) {
      goto colourcleanup;
    }
    colour[u->weight] = ZKGRAPH_COLOUR_BLACK;
  }
  
  if (sz != 0) {                /* zkListSize failed */
    return 0;
  }

  *p = pi;
  pi = NULL;

  ret = 1;

 colourcleanup:
  if (colour) { free(colour); colour = NULL; }
  if (d) { free(d); d = NULL; }
  if (pi) { free(pi); pi = NULL; }
  if (q) { zkListFree(err, &q); }
      
  return ret;
}

/**
 * Construct the path using the predicate list in the array pi (of size
 * total nodes in graph) and store the path in *path allocating space as
 * needed.  Set the array size *psz.  The path variable needs to be
 * freed from the calling function.
 */
DLLExport int zkGraphConstructPath(ZkError err, ZkGraph graph,
                                   uint32_t s, uint32_t d, uint32_t *pi,
                                   uint32_t **path, size_t *psz)
{
  const char *func = "zkGraphConstructPath";
  ZkGraphNode node = NULL;

  zkErrorClear(err);
  if (!graph || !pi || !path || !psz) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (d == s) {
    *psz = 0;
    if ((*path = malloc(sizeof(uint32_t))) == NULL) {
      zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENOMEM, "malloc() failed"); 
      return 0;
    }
    (*path)[(*psz)++] = s;
  } else {
    if (!zkHashSearch(err, graph->h, &d, (void *) &node)) {
      goto cleanpath;
    }
    if (pi[node->weight] == (graph->numnodes + 1)) {
      goto cleanpath;
    } else {
      if (!zkGraphConstructPath(err, graph, s, pi[node->weight], pi, path,
                                psz)) {
        goto cleanpath;
      }

      if (!zkMemRealloc(err, (void *) path, sizeof(uint32_t) * (*psz + 1))) {
        goto cleanpath;
      }
      (*path)[(*psz)++] = d;
    }
  }

  return 1;

 cleanpath:
  if (*path) {
    free(*path);
    *path = NULL;
    *psz = 0;
  }
  return 0;
}

#ifndef NDEBUG 
DLLExport int zkGraphDump(ZkError err, ZkGraph graph)
{
  const char *func = "zkGraphDump";
  size_t i;
  ZkGraphNode node;

  zkErrorClear(err);
  if (!graph) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  for (i = 0; i < graph->numnodes; i++) {
    if (graph->node[i]) {
      node = graph->node[i];
      fprintf(stderr, "%6d", graph->node[i]->id);
      while (node->next) {
        fprintf(stderr, " -> %6d (w=%02d)\n", node->next->id, 
                node->next->weight);
        node = node->next;
        fprintf(stderr, "      ");
      }
      fprintf(stderr, "\n");
    }
  }
  return 1;
}
#endif

static int inSet(ZkError err, uint32_t *partial, int k, uint32_t id, 
                 int *in)
{
  const char *func = "inSet";
  int i;

  if (!partial || !in) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL param");
    return 0;
  }
  *in = 0;

  for (i = 0; i < k; i++) {
    if (partial[i] == id) {
      *in = 1;
      break;
    }
  }

  return 1;
}

static int enum_paths(ZkError err, ZkGraph graph, uint32_t s, uint32_t d,
                      uint32_t *partial, int k, uint32_t **paths,
                      size_t *np)
{
  const char *func = "enum_paths";
  ZkGraphNode node = NULL;
  int i;
  int in = 0;

  if (!graph || !partial || !paths || !np) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (s == d) {
    partial[++k] = d;
    if (!*np) {
      if ((*paths = calloc(k + 2, sizeof(uint32_t))) == NULL) {
        zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENOMEM, "NULL param(s)");
        return 0;
      }
    } else {
      if (!zkMemRealloc(err, (void *) paths, sizeof(uint32_t) * 
                     (*np + 1 + k + 2))) {
        return 0;
      }
    }
    for (i = 0; i <= k; i++) {
      (*paths)[(*np)++] = partial[i];
    }
    (*paths)[(*np)++] = 0;           /* Seperate them using a 0 */
  } else {
    k++;

    if (!zkHashSearch(err, graph->h, &s, (void *) &node)) {
      return 0;
    }

    while (node->next) {
      if (!inSet(err, partial, k, node->next->id, &in)) {
        return 0;
      }

      if (!in) {
        partial[k] = s;
        if (!enum_paths(err, graph, node->next->id, d, partial, k, paths,
                        np)) {
          return 0;
        }
      }
      node = node->next;
    }
  }

  return 1;
}

/**
 * Delete the edge from the source node s to the destination node d.
 */
static int zkGraphDirectedEdgeDelete(ZkError err, ZkGraph graph, uint32_t s,
                                     uint32_t d)
{
  const char *func = "zkGraphDirectedEdgeDelete";
  ZkGraphNode node = NULL;
  ZkGraphNode temp = NULL;      /* Node to delete */

  if (!graph) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL param");
    return 0;
  }

  /* Search for the node to delete */
  if (!zkHashSearch(err, graph->h, &s, (void *) &node)) {
    zkErrorSet_va(err, func, ZKERR, ZKMGRAPH, ZKENOTFOUND,
                  "Node %d not found", s);
    return 0;
  }

  while (node->next) {
    if (node->next->id == d) {
      temp = node->next;
      node->next = node->next->next;
      temp->next = NULL;
      (void) zkGraphNodeFree(err, &temp);
      break;
    }
  }

  return 1;
}

/**
 * Delete a node.
 */
DLLExport int zkGraphDeleteNode(ZkError err, ZkGraph graph, uint32_t n)
{
  const char *func = "zkGraphDeleteNode";
  ZkGraphNode node = NULL;
  ZkGraphNode temp = NULL;      /* Node to delete */
  uint32_t id;

  if (!graph) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL param");
    return 0;
  }

  /* Search for the node to delete */
  if (!zkHashSearch(err, graph->h, &n, (void *) &node)) {
    zkErrorSet_va(err, func, ZKERR, ZKMGRAPH, ZKENOTFOUND,
                  "Node %d not found", n);
    return 0;
  }

  id = node->weight;            /* Index into the array of nodes */

  while (node->next) {
    temp = node->next;
    node->next = node->next->next;
    temp->next = NULL;

    /* Delete the edge that may be associated with this node */
    if (!zkGraphDirectedEdgeDelete(err, graph, temp->id, n)) {
      return 0;
    }

    (void) zkGraphNodeFree(err, &temp);
  }

  (void) zkHashDelete(err, graph->h, &n);

  temp = graph->node[id];
  graph->node[id] = graph->node[graph->numnodes - 1];
  graph->node[graph->numnodes - 1] = temp;
  (void) zkGraphNodeFree(err, &graph->node[graph->numnodes - 1]);
  graph->numnodes--;
  
  return 1;
}

/**
 * Recursive function to enumerate all paths from start point s to d.  The
 * enumeration stops when the path will be greater than xpaths.
 */
static int enum_xpaths(ZkError err, ZkGraph graph, uint32_t s, uint32_t d,
                      uint32_t *partial, int k, int xpaths, uint32_t **paths,
                      size_t *np)
{
  const char *func = "enum_xpaths";
  ZkGraphNode node = NULL;
  int i;
  int in = 0;

  if (!graph || !partial || !paths || !np) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (k == xpaths - 1) {
    /* Do nothing, break out of it */
  } else if (s == d) {
    partial[++k] = d;
    if (!*np) {
      if ((*paths = calloc(k + 2, sizeof(uint32_t))) == NULL) {
        zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENOMEM, "NULL param(s)");
        return 0;
      }
    } else {
      if (!zkMemRealloc(err, (void *) paths, sizeof(uint32_t) * 
                     (*np + 1 + k + 2))) {
        return 0;
      }
    }
    for (i = 0; i <= k; i++) {
      (*paths)[(*np)++] = partial[i];
    }
    (*paths)[(*np)++] = 0;           /* Seperate them using a 0 */
  } else {
    k++;

    if (!zkHashSearch(err, graph->h, &s, (void *) &node)) {
      return 0;
    }

    while (node->next) {
      if (!inSet(err, partial, k, node->next->id, &in)) {
        return 0;
      }

      if (!in) {
        partial[k] = s;
        if (!enum_xpaths(err, graph, node->next->id, d, partial, k, xpaths,
                         paths, np)) {
          return 0;
        }
      }
      node = node->next;
    }
  }

  return 1;
}

/*
 * Enumerate all available paths between nodes s and d of length xpaths and
 * return the number of paths np in the array paths.
 */ 
DLLExport int zkGraphEnumerateXPaths(ZkError err, ZkGraph graph,
                                     uint32_t s, uint32_t d, int xpaths,
                                     uint32_t **paths, size_t *np)
{
  const char *func = "zkGraphEnumerateXPaths";
  uint32_t *partial = NULL;
  int k = -1;

  if (!graph || !paths || !np) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  *np = 0;

  /* The partial solution should have no more than xpaths nodes */
  if ((partial = calloc(xpaths, sizeof(uint32_t))) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENOMEM, 
               "No memory for partial solution");
    return 0;
  }

  if (!enum_xpaths(err, graph, s, d, partial, k, xpaths, paths, np)) {
    return 0;
  }

  if (partial) {
    free(partial);
    partial = NULL;
  }
  return 1;
}

/**
 * Enumerate all available paths between nodes s and d and return
 * the number of paths np in the array paths.
 */
DLLExport int zkGraphEnumerateAllPaths(ZkError err, ZkGraph graph,
                                       uint32_t s, uint32_t d, 
                                       uint32_t **paths, size_t *np)
{
  const char *func = "zkGraphEnumerateAllPaths";

  if (!graph || !paths || !np) {
    zkErrorSet(err, func, ZKERR, ZKMGRAPH, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  return(zkGraphEnumerateXPaths(err, graph, s, d, graph->numnodes, paths, np));
}

