/*
 *  $Id: zkHash.c,v 1.14.32.2 2000/10/11 13:20:02 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/**
 *  
 *        Pointers to keys and data are copied directly into the hash table;
 *        copies of these objects are not made.
 */


#include "zks.h"

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include "zkHash.h"


struct helt {
  void *key;
  void *data;
  size_t hval; /* cached hash value for quick rejection */
  struct helt *next;
};

struct zk_hash {
  size_t nchains;
  size_t nused;
  size_t keySize;
  struct helt **elt;
  void *userData;
  
  /** Return 0 if equal, otherwise non-zero! */
  ZkHashKeyEqualFn *keyEqual;
  ZkHashKeyFreeFn *kfree;
  ZkHashDataFreeFn *dfree;
  ZkHashHashFn *hash;
};

struct zk_hash_enum {
  /* The hash table whose elements are being enumerated. */
  ZkHash htab;
  /* The current element chain at which we're looking. */
  size_t chain;
  /* The current element at which we're looking. */
  struct helt *elt;
  /* The next element. */
  struct helt *nextElt;
};


static const size_t dflTableSize = 1000;


/*
 *
 *
 *                            Internal Functions
 *
 *
 */


/*
 *  Assume h is valid.
 */
static size_t dflHashFn(ZkHash h, void *key, void *userData)
{
  char *p = key;
  size_t v = 0, i;

  (void)userData;

  for (i = 0; i < h->keySize; i++) {
    v = v * 33 + p[i];
  }

  return v;
}


static struct helt *initHelt(ZkError err, void *key, void *data, size_t hval)
{
  struct helt *e;

  zkErrorClear(err);

  if ( ! (e = malloc(sizeof *e))) {
    zkErrorSet_va(err, "initHelt", ZKERR, ZKMHASH, ZKENOMEM,
                  "can't allocate new %lu byte hash table element: %s",
                  (unsigned long)sizeof *e, strerror(errno));
    return 0;
  }

  e->key = key;
  e->data = data;
  e->hval = hval;
  e->next = 0;

  return e;
}


/*
 *  Assume h is valid.
 */
static void init(ZkHash h)
{
  h->nchains = 0;
  h->nused = 0;
  h->keySize = 0;
  h->elt = 0;
  h->keyEqual = memcmp;
  h->dfree = 0;
  h->kfree = 0;
  h->hash = dflHashFn;
}


/*
 *  Assume h is valid.
 */
static void reset(ZkHash h)
{
  h->nchains = 0;
  h->nused = 0;
  /* Don't reset the keySize field */
  h->elt = 0;
  
  /* Don't reset the keyEqual field */
  /* Don't reset the dfree field */
  /* Don't reset the kfree field */
  /* Don't reset the hash field */
}


/*
 *  Assume h and e are valid.
 */
static void freeChain(ZkHash h, struct helt *e)
{
  struct helt *ne;
  
  for (; e; e = ne) {
    ne = e->next;
    if (e->data && h->dfree) {
      h->dfree(h, e->data, h->userData); e->data = 0;
    }
    if (e->key && h->kfree) {
      h->kfree(h, e->key, h->userData);  e->key = 0;
    }
    free(e); e = 0;
  }
}


/*
 *  Assume h is valid.
 */
static void freeTable(ZkHash h)
{
  size_t i;
  
  for (i = 0; i < h->nchains; i++) {
    if (h->elt[i]) {
      freeChain(h, h->elt[i]);
    }
  }
}

/* To distinguish between an error and a key not found use find() */

static int search(ZkError err, ZkHash h, void *key, void **okey, void **data,
                  struct helt **elt, struct helt **prev, size_t *hval)
{
  size_t hv, pos;
  struct helt *e, *p;

  zkErrorClear(err);

  if ( ! h) {
    zkErrorSet(err, "search", ZKERR, ZKMHASH, ZKENULLPTR, "h is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, "search", ZKERR, ZKMHASH, ZKENULLPTR, "key is null");
    return 0;
  }

  hv = h->hash(h, key, h->userData);
  if (hval) { *hval = hv; }
  pos = hv % h->nchains;
  
  for (p = 0, e = h->elt[pos]; e; p = e, e = e->next) {
    /*
     *  It seems odd requiring the key equality function to return 0 if the
     *  keys are equal, but it allows us to directly use memcmp(), strncmp(),
     *  etc., which have this behaviour.  We just don't care about the
     *  relative ordering of the keys.
     */
    if (e->hval == hv && h->keyEqual(key, e->key, h->keySize) == 0) {
      if (okey) { *okey = e->key; }
      if (data) { *data = e->data; }
      if (elt)  { *elt  = e; }
      if (prev) { *prev = p; }
      return 1;
    }
  }

  return 0;
}

static int find(ZkError err, ZkHash h, void *key, void **okey, void **data,
                  struct helt **elt, struct helt **prev, size_t *hval, int *found)
{

  size_t hv, pos;
  struct helt *e, *p;
  int ret = 0;

  zkErrorClear(err);
  *found = 0;

  if ( ! h) {
    zkErrorSet(err, "find", ZKERR, ZKMHASH, ZKENULLPTR, "h is null");
    goto done;
  }
  if ( ! key) {
    zkErrorSet(err, "find", ZKERR, ZKMHASH, ZKENULLPTR, "key is null");
    goto done;
  }

  hv = h->hash(h, key, h->userData);
  if (hval) { *hval = hv; }
  pos = hv % h->nchains;
  
  for (p = 0, e = h->elt[pos]; e; p = e, e = e->next) {
    /*
     *  It seems odd requiring the key equality function to return 0 if the
     *  keys are equal, but it allows us to directly use memcmp(), strncmp(),
     *  etc., which have this behaviour.  We just don't care about the
     *  relative ordering of the keys.
     */
    if (e->hval == hv && h->keyEqual(key, e->key, h->keySize) == 0) {
      if (okey) { *okey = e->key; }
      if (data) { *data = e->data; }
      if (elt)  { *elt  = e; }
      if (prev) { *prev = p; }
      /* set the 'found' argument to 'true' */
      *found = 1;
      ret = 1;
      goto done;
    }
  }

  /* if we are here, we didn't find it ('found' is 0) */
  ret = 1;
 done:
  return ret;

}


/*
 *
 *
 *                            External Functions
 *
 *
 */

DLLExport int zkHashNew(ZkError err, ZkHash *h)
{
  ZkHash htab;
  
  zkErrorClear(err);
  
  if ( ! h) {
    zkErrorSet(err, "zkHashNew", ZKERR, ZKMHASH, ZKENULLPTR,
               "h is null");
    return 0;
  }

  if ( ! (htab = malloc(sizeof *htab))) {
    zkErrorSet_va(err, "zkHashNew", ZKERR, ZKMHASH, ZKENOMEM,
                  "can't allocate %lu byte hash table: %s",
                  (unsigned long)sizeof *htab, strerror(errno));
    return 0;
  } else {
    init(htab);
    *h = htab;

    return 1;
  }
}


DLLExport int zkHashFree(ZkError err, ZkHash *h)
{
  ZkHash htab;

  zkErrorClear(err);
  
  if ( ! h) {
    zkErrorSet(err, "zkHashFree", ZKERR, ZKMHASH, ZKENULLPTR,
               "h is null");
    return 0;
  } else if ( ! (htab = *h)) {
    zkErrorSet(err, "zkHashFree", ZKERR, ZKMHASH, ZKEPARMBAD,
               "*h is null");
    return 0;
  }

  if (htab->elt) {
    freeTable(htab);
    free(htab->elt);
    htab->elt = 0;
  }

  free(htab); *h = 0;

  return 1;
}


DLLExport int zkHashInsert(ZkError err, ZkHash h, void *key, void *data)
{
  size_t hval;
  struct helt *elt;
  void *d;
  
  zkErrorClear(err);
  
  if ( ! h) {
    zkErrorSet(err, "zkHashInsert", ZKERR, ZKMHASH, ZKENULLPTR, "h is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, "zkHashInsert", ZKERR, ZKMHASH, ZKENULLPTR, "key is null");
    return 0;
  }

  /*
   *  The data pointer is allowed to be null.
   */

  if (search(err, h, key, 0, &d, &elt, 0, &hval)) {
    /* Item already exists. */
    zkErrorSet(err, "zkHashInsert", ZKWARN, ZKMHASH, ZKEEXIST,
               "key found in table");
    return 0;
  } else {
    size_t pos;
    
    if ( ! (elt = initHelt(err, key, data, hval))) {
      return 0;
    }
    
    pos = hval % h->nchains;
    
    /*
     *  Insert the element at the beginning of the chain.
     */
    
    elt->next = h->elt[pos];
    h->elt[pos] = elt;

    h->nused++;

    return 1;
  }
}



DLLExport int zkHashDelete(ZkError err, ZkHash h, void *key)
{
  size_t hval, pos;
  struct helt *elt, *prev;
  void *d;
  
  zkErrorClear(err);
  
  if ( ! h) {
    zkErrorSet(err, "zkHashDelete", ZKERR, ZKMHASH, ZKENULLPTR,
               "h is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, "zkHashDelete", ZKERR, ZKMHASH, ZKENULLPTR,
               "key is null");
    return 0;
  }
  
  if ( ! search(err, h, key, 0, &d, &elt, &prev, &hval)) {
    zkErrorSet(err, "zkHashDelete", ZKWARN, ZKMHASH, ZKENOTFOUND,
               "item not found in hash table");
    return 0;
  }

  pos = hval % h->nchains;

  if (prev) {
    prev->next = elt->next;
  } else {
    h->elt[pos] = elt->next;
  }

  if (h->kfree) { h->kfree(h, elt->key, h->userData);  elt->key  = 0; }
  if (h->dfree) { h->dfree(h, elt->data, h->userData); elt->data = 0; }

  free(elt);

  return 1;
}


DLLExport int zkHashSetKeySize(ZkError err, ZkHash h, size_t sz)
{
  zkErrorClear(err);
  
  if ( ! h) {
    zkErrorSet(err, "zkHashSetKeySize", ZKERR, ZKMHASH, ZKENULLPTR,
               "h is null");
    return 0;
  }
  if (sz == 0) {
    zkErrorSet(err, "zkHashSetKeySize", ZKERR, ZKMHASH, ZKEPARMRANGE,
               "key size is 0");
    return 0;
  }

  h->keySize = sz;
  
  return 1;
}


DLLExport int zkHashSetTableSize(ZkError err, ZkHash h, size_t sz)
{
  zkErrorClear(err);
  
  if ( ! h) {
    zkErrorSet(err, "zkHashSetTableSize", ZKERR, ZKMHASH, ZKENOMEM,
               "h is null");
    return 0;
  }

  if (sz == 0) {
    /*
     *  Toast the existing table.
     */
    
    if (h->elt) {
      freeTable(h);
      free(h->elt); h->elt = 0;
    }
  } else {
    /*
     *  Transfer the elements from the current table to a new table, then
     *  destroy the old table.
     */

    size_t i;
    struct zk_hash old;

    old = *h;
    reset(h);

    if ( ! (h->elt = malloc(sz * sizeof *(h->elt)))) {
      zkErrorSet_va(err, "zkHashSetTableSize", ZKERR, ZKMHASH, ZKENOMEM,
                    "can't allocate %lu new hash table elements: %s",
                    (unsigned long)sz, strerror(errno));
      *h = old; /* restore it upon an error */
      return 0;
    }

    /*
     *  Pointers to hash table elements must be initialized; we'll need to
     *  know whether a chain exists in this slot.
     */
    
    for (i = 0; i < sz; i++) {
      h->elt[i] = 0;
    }
    
    h->nchains = sz;
    
    /*
     *  Rehash each element in the old table into the new table.
     */
    
    for (i = 0; i < old.nchains; i++) {
      if (old.elt[i]) {
        struct helt *e;

        for (e = old.elt[i]; e; e = old.elt[i]->next) {
          size_t pos = e->hval % h->nchains;

          old.elt[i] = e->next;
          e->next = h->elt[pos];
          h->elt[pos] = e;
        }
      }
    }

    if (old.elt) {
      free(old.elt); old.elt = 0;
    }
  }

  return 1;
}

/* 
 * this function is deprecated, use zkHashFind() instead. 
 */

DLLExport int zkHashSearch(ZkError err, ZkHash h, void *key, void **data)
{
  zkErrorClear(err);
  
  if ( ! h) {
    zkErrorSet(err, "zkHashSearch", ZKERR, ZKMHASH, ZKENULLPTR,
               "h is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, "zkHashSearch", ZKERR, ZKMHASH, ZKENULLPTR,
               "key is null");
    return 0;
  }

  /*
   *  Search allows data to be null; it will just tell us whether or not the
   *  item was found.
   */

  return search(err, h, key, 0, data, 0, 0, 0);
}

/*
 * 'found' is set to 1 if 'key' is found, 0 if not.
 */

DLLExport int zkHashFind(ZkError err, ZkHash h, void *key, void **data, int *found)
{
  zkErrorClear(err);
  
  if ( ! h) {
    zkErrorSet(err, "zkHashFind", ZKERR, ZKMHASH, ZKENULLPTR,
               "h is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, "zkHashFind", ZKERR, ZKMHASH, ZKENULLPTR,
               "key is null");
    return 0;
  }

  /* found is set to 1 if key is found, 0 if not. */
  if ( ! find(err, h, key, 0, data, 0, 0, 0, found)) {
     return 0;
  }

  return 1;
}

/*
 *  Identical to zkHashSearch(), but we also return a pointer to the key
 *  stored in the hash table.
 */
DLLExport int zkHashSearchGivingKeyData(ZkError err, ZkHash h, void *key, void **kptr, void **data)
{
  zkErrorClear(err);
  
  if ( ! h) {
    zkErrorSet(err, "zkHashSearchGivingKeyData", ZKERR, ZKMHASH, ZKENULLPTR,
               "h is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, "zkHashSearchGivingKeyData", ZKERR, ZKMHASH, ZKENULLPTR,
               "key is null");
    return 0;
  }

  return search(err, h, key, kptr, data, 0, 0, 0);
}


DLLExport int zkHashSetDataFreeFn(ZkError err, ZkHash h, ZkHashDataFreeFn fn)
{
  zkErrorClear(err);
  
  if ( ! h) {
    zkErrorSet(err, "zkHashSetDataFreeFn", ZKERR, ZKMHASH, ZKENULLPTR,
               "h is null");
    return 0;
  }
  if ( ! fn) {
    zkErrorSet(err, "zkHashSetDataFreeFn", ZKERR, ZKMHASH, ZKENULLPTR,
               "function pointer is null");
    return 0;
  }

  if (h->nused > 0) {
    /*
     *  You shouldn't change functions while there's data in the table.
     */
    zkErrorSet(err, "zkHashSetDataFreeFn", ZKWARN, ZKMHASH, ZKEBADSTATE,
               "can't change function with data in table");
    return 0;
  }
  
  h->dfree = fn;
  return 1;
}


DLLExport int zkHashSetKeyFreeFn(ZkError err, ZkHash h, ZkHashKeyFreeFn fn)
{
  zkErrorClear(err);
  
  if ( ! h) {
    zkErrorSet(err, "zkHashSetKeyFreeFn", ZKERR, ZKMHASH, ZKENULLPTR,
               "h is null");
    return 0;
  }
  if ( ! fn) {
    zkErrorSet(err, "zkHashSetKeyFreeFn", ZKERR, ZKMHASH, ZKENULLPTR,
               "function pointer is null");
    return 0;
  }
  
  if (h->nused > 0) {
    /*
     *  You shouldn't change functions while there's data in the table.
     */
    zkErrorSet(err, "zkHashSetKeyFreeFn", ZKWARN, ZKMHASH, ZKEBADSTATE,
               "can't change function with data in table");
    return 0;
  }
  
  h->kfree = fn;
  return 1;
}


DLLExport int zkHashSetHashFn(ZkError err, ZkHash h, ZkHashHashFn fn)
{
  zkErrorClear(err);
  
  if ( ! h) {
    zkErrorSet(err, "zkHashSetHashFn", ZKERR, ZKMHASH, ZKENULLPTR,
               "h is null");
    return 0;
  }
  if ( ! fn) {
    zkErrorSet(err, "zkHashSetHashFn", ZKERR, ZKMHASH, ZKENULLPTR,
               "function pointer is null");
    return 0;
  }

  if (h->nused > 0) {
    /*
     *  You shouldn't change functions while there's data in the table.
     */
    zkErrorSet(err, "zkHashSetHashFn", ZKWARN, ZKMHASH, ZKEBADSTATE,
               "can't change function with data in table");
    return 0;
  }
  
  h->hash = fn;
  return 1;
}


DLLExport int zkHashSetKeyEqualFn(ZkError err, ZkHash h, ZkHashKeyEqualFn *fn)
{
  zkErrorClear(err);
  
  if ( ! h) {
    zkErrorSet(err, "zkHashSetKeyEqualFn", ZKERR, ZKMHASH, ZKENULLPTR,
               "h is null");
    return 0;
  }
  if ( ! fn) {
    zkErrorSet(err, "zkHashSetKeyEqualFn", ZKERR, ZKMHASH, ZKENULLPTR,
               "function pointer is null");
    return 0;
  }

  if (h->nused > 0) {
    /*
     *  You shouldn't change functions while there's data in the table.
     */
    zkErrorSet(err, "zkHashSetKeyEqualFn", ZKWARN, ZKMHASH, ZKEBADSTATE,
               "can't change function with data in table");
    return 0;
  }
  
  h->keyEqual = fn;
  return 1;
}


DLLExport int zkHashSetUserData(ZkError err, ZkHash h, void *userData)
{
  zkErrorClear(err);
  
  if ( ! h) {
    zkErrorSet(err, "zkHashSetUserData", ZKERR, ZKMHASH, ZKENULLPTR,
               "h is null");
    return 0;
  }

  /*
   *  It's okay if userData is null.
   */

  h->userData = userData;
  return 1;
}


/*
 *  Traverse the hash table, calling `cb' for each element.
 */
DLLExport int zkHashTraverse(ZkError err, ZkHash h, ZkHashTravFn cb, void *userData)
{
  ZkHashEnum en = 0;
  int ret = 0;
  
  zkErrorClear(err);

  if ( ! (zkHashEnumNew(err,&en) &&
          zkHashEnumSetHash(err, en, h))) {
    goto done;
  }

  while (zkHashEnumNext(err, en)) {
    void *d, *k;

    if ( ! (zkHashEnumGetData(err, en, &d) &&
            zkHashEnumGetKey(err, en, &k))) {
      goto done;
    }

    if ( ! cb(err, k, d, userData)) {
      goto done;
    }
  }

  ret = 1;

 done:
  
  if (en) { zkHashEnumFree(0, &en); }

  return ret;
}


DLLExport int zkHashEnumFree(ZkError err, ZkHashEnum *e)
{
  ZkHashEnum en;

  zkErrorClear(err);
  
  if ( ! e) {
    zkErrorSet(err, "zkHashEnumFree", ZKERR, ZKMHASH, ZKENULLPTR, "e is null");
    return 0;
  }
  if ( ! *e) {
    zkErrorSet(err, "zkHashEnumFree", ZKERR, ZKMHASH, ZKEPARMBAD, "*e is null");
    return 0;
  }

  en = *e;

  en->htab = 0;
  en->chain = 0;
  en->elt = 0;

  free(en); *e = 0;
  
  return 1;
}


DLLExport int zkHashEnumGetData(ZkError err, ZkHashEnum e, void **data)
{
  zkErrorClear(err);
  
  if ( ! e) {
    zkErrorSet(err, "zkHashEnumGetData", ZKERR, ZKMHASH, ZKENULLPTR, "e is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "zkHashEnumGetData", ZKERR, ZKMHASH, ZKENULLPTR, "data is null");
    return 0;
  }

  *data = e->elt->data;

  return 1;
}


DLLExport int zkHashEnumGetKey(ZkError err, ZkHashEnum e, void **key)
{
  zkErrorClear(err);
  
  if ( ! e) {
    zkErrorSet(err, "zkHashEnumGetKey", ZKERR, ZKMHASH, ZKENULLPTR, "e is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, "zkHashEnumGetKey", ZKERR, ZKMHASH, ZKENULLPTR, "key is null");
    return 0;
  }

  *key = e->elt->key;

  return 1;
}


DLLExport int zkHashEnumNew(ZkError err, ZkHashEnum *e)
{
  ZkHashEnum en;

  zkErrorClear(err);
  
  if ( ! e) {
    zkErrorSet(err, "zkHashEnumNew", ZKERR, ZKMHASH, ZKENULLPTR, "e is null");
    return 0;
  }

  if ( ! (en = malloc(sizeof *en))) {
    zkErrorSet_va(err, "zkHashEnumNew", ZKERR, ZKMHASH, ZKENOMEM,
                  "can't allocate %lu byte hash table enumerator: %s",
                  (unsigned long)sizeof *en, strerror(errno));
    return 0;
  }
  
  en->htab = 0;
  en->chain = 0;
  en->elt = 0;
  en->nextElt = 0;

  *e = en;

  return 1;
}


DLLExport int zkHashEnumNext(ZkError err, ZkHashEnum e)
{
  zkErrorClear(err);
  
  if ( ! e) {
    zkErrorSet(err, "zkHashEnumNext", ZKERR, ZKMHASH, ZKENULLPTR, "e is null");
    return 0;
  }

  if ( ! e->nextElt) {
    /*
     *  The elements of the hash table have been enumerated.
     */
    return 0;
  }
  
  e->elt = e->nextElt;

  /*
   *  Find the next element by looking for a successor in the current chain,
   *  or by looking for the next non-empty chain.
   */

  if ( ! (e->nextElt = e->elt->next)) {
    size_t i;
    
    for (i = e->chain + 1; i < e->htab->nchains; i++) {
      if ((e->nextElt = e->htab->elt[i])) {
        break;
      }
    }
    
    e->chain = i;
  }

  return 1;
}


DLLExport int zkHashEnumSetHash(ZkError err, ZkHashEnum e, ZkHash h)
{
  size_t i;

  zkErrorClear(err);
  
  if ( ! e) {
    zkErrorSet(err, "zkHashEnumSetHash", ZKERR, ZKMHASH, ZKENULLPTR, "e is null");
    return 0;
  }
  if ( ! h) {
    zkErrorSet(err, "zkHashEnumSetHash", ZKERR, ZKMHASH, ZKENULLPTR, "h is null");
    return 0;
  }

  e->htab = h;
  e->chain = 0;
  e->elt = 0;

  /*
   *  Find the first element in the table.
   */

  for (i = 0; i < e->htab->nchains; i++) {
    if ((e->nextElt = e->htab->elt[i])) {
      break;
    }
  }

  /*
   *  If there are no elements in the hash table `i' will be invalid, but
   *  e->nextElt will be null so zkHashEnumNext() will return an indication
   *  that the table has been completely enumerated.
   */

  e->chain = i;

  return 1;
}
