/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */
#include <zkError.h>
#include <zkList.h>

struct T_ZkListNode {
  struct T_ZkListNode * prev;  /* Previous node in list */
  struct T_ZkListNode * next;  /* Next Node in list */
  void * data;                 /* Data pointer for the node */
};

#define HEAD 1  /* These are used when the cursor is NULL to indicate wether */
#define TAIL 2  /* it is at the head or tail of the list using cend */

struct T_ZkList {
  size_t elements;                /* Number of elements in list */
  ZkListDataFreeFn dataFree;  /* Function used to free element structure */
  void * dataFreeData;          /* Extra data pointer passed to free function */
  struct T_ZkListNode * cursor; /* Current cursor */
  struct T_ZkListNode * head;  /* First node in list */
  struct T_ZkListNode * tail;  /* Last node in list */
  int cend;                    /* When cursor is NULL, indicates HEAD or TAIL */
};

DLLExport int
zkListDataFree (
  ZkError err ,
  ZkList list ,
  void * data ,
  void * extra )
{
  zkErrorClear(err);
  if (data && list) {
    if (extra) extra = NULL; /* Have to do something to get rid of warning */
    free(data);
    return 1;
  }
  zkErrorSet(err,"zkListDataFree",ZKERR,ZKMLIST,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Create a new zkList structure.
 *
 *  @param err Returns the error if any occured 
 *  @param list Returns the new list
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkListNew (
  ZkError err ,
  ZkList * list )
{
  zkErrorClear(err);
  if (list == NULL) {
    zkErrorSet(err,"zkListNew",ZKERR,ZKMLIST,ZKENULLPTR,NULL);
    return 0;
  }
  *list = (ZkList) malloc(sizeof(struct T_ZkList));
  if (*list == NULL) {
    zkErrorSet(err,"zkListNew",ZKERR,ZKMLIST,ZKENOMEM,"malloc");
    return 0;
  }
  memset(*list,0,sizeof(struct T_ZkList));
  (*list)->cend = HEAD;
#if 0
  (*list)->dataFree = zkListDataFree;
#else
  (*list)->dataFree = NULL;
#endif
  return 1;
}


/**
 *  Set the function and user data used to free pieces of data inserted into
 *  the list when elements are deleted or the whole list is destroyed.  By
 *  default, the zkListDataFree function is used as the free function.  This
 *  is just a wrapper for free().  You can reset this function to NULL and
 *  the data will not be freed when the node containing it is deleted or the
 *  list is destroyed.
 *
 *  @param err Returns the error if any occured
 *  @param list The ZkList you wish to modify 
 *  @param func The data deallocation function you wish to use when a node is
 *              deleted or the list is destroyed.   See the prototype for a
 *              free function in zkList.h
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkListSetDataFreeFunction (
  ZkError err ,
  ZkList list ,
  ZkListDataFreeFn func ,
  void * data )
{
  zkErrorClear(err);
  if (list) {
    list->dataFree = func;
    list->dataFreeData = data;
    return 1;
  }
  zkErrorSet(err,"zkListSetDataFreeFunction",ZKERR,ZKMLIST,ZKENULLPTR,NULL);
  return 0;
}


/**
 *  Get the function and user data used to free pieces of data inserted into
 *  the list when elements are deleted or the whole list is destroyed.
 *
 *  @param err      Returns the error if any occured
 *  @param list     The ZkList you wish to retreive the info from 
 *  @param freeFunc Storage for the data deallocation function 
 *  @param data     Storage for user data
 *
 *  @author Yannick Dubuc
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkListGetDataFreeFunction (
  ZkError err ,
  ZkList list ,
  ZkListDataFreeFn * freeFunc ,
  void ** data )
{
  if (list) {
    *freeFunc = list->dataFree;
    *data = (void *)list->dataFreeData;
    return 1;
  }
  zkErrorSet(err,"zkListGetDataFreeFunction",ZKERR,ZKMLIST,ZKENULLPTR,NULL);
  return 0;
}


/**
 *  Destroy the list.  If you specified a datafree function, the data inserted
 *  into the list will be destroyed as well.  If you did not specify such
 *  a function, you are responsible for cleaning up the memory yourself.
 * 
 *  @param err Returns the error if any occured
 *  @param list The list to destroy.  The pointer value is reset to NULL
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure.
 */

DLLExport int
zkListFree (
  ZkError err ,
  ZkList * list )
{
  ZkListNode current = NULL;
  ZkListNode next = NULL;

  if (list) {
    if (*list) {
      current = (*list)->head;
      while (current) {
        next = current->next;
        if ((*list)->dataFree && current->data) {
          if (!((*list)->dataFree(err,*list,current->data,
                                  (*list)->dataFreeData))) {
            return 0;
          }
        }
        memset(current,0,sizeof(struct T_ZkListNode));
        free(current);
        current = next;
      }
      memset(*list,0,sizeof(struct T_ZkList));
      free(*list);
      *list = NULL;
      return 1;
    }
  }
  zkErrorSet(err,"zkListFree",ZKERR,ZKMLIST,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Return the current cursor position inside the list.  The value returned
 *  is a list node pointer.  If you wish to get the data at the cursor 
 *  position, use the zkListGet() function.  Note that the cursor pointer
 *  can be NULL when it is just passed the head or tail of the list.
 * 
 *  @param err Return the error if any occured
 *  @param list The list whose cursor pointer you wish to retreive
 *  @param cursor Returns the pointer to the node where the cursor is
 *                currently located.
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkListCursor (
  ZkError err ,
  ZkList list ,
  ZkListNode * cursor )
{
  zkErrorClear(err);
  if (list && cursor) {
    *cursor = list->cursor;
    return 1;
  }
  zkErrorSet(err,"zkListCursor",ZKERR,ZKMLIST,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Set the list cursor to an explicit pointer position.   Do not use this
 *  function unless you're sure of what you're doing (you can really screw
 *  yourself up with this one).  If you wish to move the cursor forwards or
 *  backwards or to the head or tail of the list, use the zkListMoveCursor()
 *  function.  You should only ever set the cursor to value which were obtained
 *  using the zkListCursor() function.  
 * 
 *  @param err Return the error if any occured
 *  @param list The list for which you wish to set the cursor
 *  @param cursor The cursor pointer you wish to set
 *  @version 1.0 
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkListSetCursor (
  ZkError err ,
  ZkList list ,
  ZkListNode cursor )
{
  zkErrorClear(err);
  if (list && cursor) {
    list->cursor = cursor;
    return 1;
  }
  zkErrorSet(err,"zkListSetCursor",ZKERR,ZKMLIST,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Return the current number of elements in the list.
 *
 *  @param err Return the error if any occured
 *  @param list The list to analyze
 *  @param size Returns the number of elements (nodes) in the list
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkListSize (
  ZkError err ,
  ZkList list ,
  size_t * size )
{
  zkErrorClear(err);
  if (list && size) {
    *size = list->elements;
    return 1;
  }
  zkErrorSet(err,"zkListSize",ZKERR,ZKMLIST,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Move the list cursor.  The allowable cursor positions are
 *  LIST_HEAD       Move the cursor to the first node of the list
 *  LIST_TAIL       Move the cursor to the last node of the list
 *  LIST_FORWARDS   Move the cursor one position towards the start of the list
 *  LIST_BACKWARDS  Move the cursor one position towards the end of the list
 *  LIST_STAY       The cursor stays where it is
 *
 *  @param err Returns the error if any occured
 *  @param list The list whose cursor you wish to move
 *  @param cursorpos Where do you wish to move the cursor? see the list
 *  @version 1.0
 *  @author Philippe Boucher
 *  @Return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkListMoveCursor (
  ZkError err ,
  ZkList list ,
  int cursorpos )
{
  zkErrorClear(err);
  if (list == NULL) {
    zkErrorSet(err,"zkListMoveCursor",ZKERR,ZKMLIST,ZKENULLPTR,NULL);
    return 0;
  }
  switch (cursorpos) {
  case LIST_HEAD:
    if (list->head) {
      list->cursor = list->head;
    } else {
      list->cursor = NULL;
      list->cend = HEAD;
    }
    break;
  case LIST_TAIL:
    if (list->tail) {
      list->cursor = list->tail;
    } else {
      list->cursor = NULL;
      list->cend = TAIL;
    }
    break;
  case LIST_BACKWARDS:
    if (list->cursor) {
      if ((list->cursor)->prev) {
        list->cursor = (list->cursor)->prev;
      } else {
        list->cursor = NULL;
        list->cend = HEAD;
      }
    } else {
      if (list->cend == TAIL) {
        list->cursor = list->tail;
      } else {
        zkErrorSet(err,"zkListMoveCursor",ZKERR,ZKMLIST,ZKERANGE,
                   "Cursor is at the head of the list");
        return 0;
      }
    }  
    break;
  case LIST_FORWARDS:
    if (list->cursor) {
      if ((list->cursor)->next) {
        list->cursor = (list->cursor)->next;
      } else {
        list->cursor = NULL;
        list->cend = TAIL;
      }
    } else {
      if (list->cend == HEAD) {
        list->cursor = list->head;
      } else {
        zkErrorSet(err,"zkListMoveCursor",ZKERR,ZKMLIST,ZKERANGE,
                   "Cursor is at the end of the list");
        return 0;
      }
    }
    break;
  case LIST_STAY:
    break;
  default:
    zkErrorSet(err,"zkListMoveCursor",ZKERR,ZKMLIST,ZKEPARMBAD,
               "Unrecognized move position");
    return 0;
  }
  return 1;
}

/**
 *  Get the data at the cursor position and update the cursor position.  After
 *  retreival, the list cursor will be moved according to the cursorpos 
 *  parameter as follows:
 *  LIST_HEAD       Move the cursor to the first node of the list
 *  LIST_TAIL       Move the cursor to the last node of the list
 *  LIST_FORWARDS   Move the cursor one position towards the start of the list
 *  LIST_BACKWARDS  Move the cursor one position towards the end of the list
 *  LIST_STAY       The cursor stays where it is
 *
 *  @param err Returns the error if any occured
 *  @param list The list from which you wish to retreive data
 *  @param data Returns a pointer to the data held at the cursor position
 *  @param cursorpos The position where you wish to move the cursor after
 *                   having retreived the data
 */

DLLExport int
zkListGet (
  ZkError err ,
  ZkList list ,
  void ** data ,
  int cursorpos )
{
  zkErrorClear(err);
  if (list && data) {
    if (list->cursor) {
      *data = (list->cursor)->data;
    } else {
      zkErrorSet(err,"zkListGet",ZKWARN,ZKMLIST,ZKERANGE,
                 "Cursor points past head or tail of list");
      return 0;
    }
    if (!zkListMoveCursor(err,list,cursorpos)) {
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkListGet",ZKERR,ZKMLIST,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Set the data at the cursor position.
 *
 *  @param err Returns the error if any occured
 *  @param list The list from which you wish to retreive data
 *  @param data Pointer to the data to be set
 */

DLLExport int
zkListSet (
  ZkError err ,
  ZkList list ,
  void * data )
{
  zkErrorClear(err);
  if (list && data) {
    if (list->cursor) {
      (list->cursor)->data = data;
    } else {
      zkErrorSet(err,"zkListSet",ZKWARN,ZKMLIST,ZKERANGE,
                 "Cursor points past head or tail of list");
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkListSet",ZKERR,ZKMLIST,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Delete the cursor node and update its position.  See zkListGet for the
 *  cursor positions.  For this particular function LIST_STAY will cause
 *  the cursor to move forward (it does not make sense to stay on a node 
 *  which is no longer there!)  If the list data free function has been
 *  defined (it is defined to a wrapper around free() by default) The data
 *  associated with the deleted node will be deallocated.
 * 
 *  @param err Returns the error if any occured
 *  @param list The list from which you wish to delete a node
 *  @param cursorpos Where you wish the cursor positioned after the node has
 *                   been deleted.
 *  @return A non-zero value on success, a 0 on failure
 *  @version 1.0
 *  @author Philippe Boucher
 */

DLLExport int
zkListDelete (
  ZkError err ,
  ZkList list ,
  int cursorpos ) 
{
  ZkListNode delete = NULL;

  zkErrorClear(err);
  if (list) {
    if (list->elements == 0) {
      zkErrorSet(err,"zkListDelete",ZKERR,ZKMLIST,ZKEEMPTY,
                 "Data structure is empty");
      return 0;
    }
    if (list->cursor == NULL) {
      zkErrorSet(err,"zkListDelete",ZKERR,ZKMLIST,ZKERANGE,
                 "Cursor points past the head or tail of the list");
      return 0;
    }
    delete = list->cursor;
    if ((delete != list->head) && (delete != list->tail)) {
      if (cursorpos == LIST_STAY) {
        cursorpos = LIST_FORWARDS;
      }
      if (!zkListMoveCursor(err,list,cursorpos)) {
        return 0;
      }
      (delete->prev)->next = delete->next;
      (delete->next)->prev = delete->prev;
    } else if ((delete == list->head) && (delete == list->tail)) {
      list->head = NULL;
      list->tail = NULL;
      list->cursor = NULL;
      list->cend = HEAD;
    } else if (delete == list->head) {
      list->head = delete->next;
      (list->head)->prev = NULL;
      list->cursor = list->head;
      list->cend = HEAD;
    } else {
      list->tail = delete->prev;
      (list->tail)->next = NULL;
      list->cursor = list->tail;
      list->cend = TAIL;
    }
    list->elements -= 1; 
    if (list->dataFree) {
      if (!(list->dataFree(err,list,delete->data,list->dataFreeData))) {
        return 0;
      }
    } 
    memset(delete,0,sizeof(struct T_ZkListNode));
    free(delete);
    return 1;
  }
  zkErrorSet(err,"zkListDelete",ZKERR,ZKMLIST,ZKENULLPTR,NULL);
  return 0;
}  

/**
 *  Insert data into the list at the position specified.  After an insertion,
 *  the cursor will point to the newly inserted data.
 *
 *  @param err Return the error if any occured
 *  @param list The list to which you wish to add data
 *  @param data The data pointer you wish to add.  Note only the pointer is
 *              added to the list, The data itself is not copied so you must
 *              make sure it does not get accidentaly destroyed.
 *  @param where The position where you wish the cursor to be placed after
 *               the insert.
 *  @return A non-zero value on success, a 0 on failure
 *  @version 1.0
 *  @author Philippe Boucher
 */

  
DLLExport int
zkListInsert (
  ZkError err ,
  ZkList list ,
  void * data ,
  int where )
{
  struct T_ZkListNode * node = NULL;

  zkErrorClear(err);
  if ((list == NULL) || (data == NULL)) {
    zkErrorSet(err,"zkListInsert",ZKERR,ZKMLIST,ZKENULLPTR,NULL);
    return 0;
  }
  node = (struct T_ZkListNode *) malloc(sizeof(struct T_ZkListNode));
  if (node == NULL) {
    zkErrorSet(err,"zkListInsert",ZKERR,ZKMLIST,ZKENOMEM,
               "Could not create a list node for the specified data");
    return 0;
  }  
  memset(node,0,sizeof(struct T_ZkListNode));
  node->data = data;
  switch (where) {
  case LIST_HEAD:
    if (list->tail == NULL) {
      list->tail = node;
    }
    if (list->head != NULL) {
      (list->head)->prev = node;
    }
    node->next = list->head;
    node->prev = NULL;
    list->head = node;
    break;
  case LIST_TAIL:
    if (list->head == NULL) {
      list->head = node;
    } 
    if (list->tail != NULL) {
      (list->tail)->next = node;
    }
    node->prev = list->tail;
    node->next = NULL;
    list->tail = node;
    break;
  case LIST_BEFORECURSOR:
  case LIST_AFTERCURSOR:
    zkErrorSet(err,"zkListInsert",ZKERR,ZKMLIST,ZKESTUB,
               "This position is not yet implemented");
    free(node);
    return 0;
  }
  list->elements += 1;
  list->cursor = node;
  return 1;
}
