/*
 *  $Id: zkLock.c,v 1.1.12.3 2000/12/14 17:54:28 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */
 
#include <unistd.h>
#include <sys/types.h>
#include <sys/file.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include "zks.h"
#include "zkError.h"

#include "zkLock.h"

/* Lock handle definition */


struct zk_lock {
    FILE * fLockFile;
    char * szLockFile;
    struct flock * lck;
};


/* Private function prototypes */

int zkLockRequestNoWait(
    ZK_OUT      ZkError err,
    ZK_IN       ZkLock  lock,
    ZK_IN       short   lockType,
    ZK_OUT      int    *granted );



/*--------------------------------------------------------------------------------
** Creation of a lock handle. This function allocate a new lock handle and return it
** to the caller.
**--------------------------------------------------------------------------------
** err,        Contain an error if this function fails. 
** lockHandle  Pointer to a lock handle 
** path        Path where to put the lock file          
** lockname    Optional, Used to give a special name to 
**             the lock file instead of the default.    
**
** return ZK_FALSE and an explanation in err if the operation fails
*/
DLLExport int zkLockNew(
    ZK_OUT      ZkError  err,
    ZK_OUT      ZkLock * lockHandle, /* Pointer to a lock handle                 */
    ZK_IN       char   *  path,
    ZK_IN       char   *  lockname) {

    ZkLock      pLock = 0;
    const char *fn    = "zkLockNew";
    
    
    if (!err ) {
        goto abort;
    }

    zkErrorClear(err);
    
    if (!lockHandle || !path) {
        zkErrorSet(err,fn,ZKERR,ZKMLOCK,ZKENULLPTR,"invalid parameter.");
        goto abort;
    }
    

    /* We allocate the lock handle */
    if ( ! (pLock = malloc(sizeof (*pLock)))) {
        zkErrorSet_va(err, fn, ZKERR, ZKMLOCK, errno,
                    "error %d allocating %lu bytes of memory for lock handle : %s",
                    errno, (unsigned long)sizeof *pLock, strerror(errno));
        goto abort;
    }
    
    memset(pLock,0,sizeof(*pLock));

    /* We construct the file path and open the lock file */
    if (!lockname) {
        /* We use the default lock file name */
        pLock->szLockFile = malloc(strlen(path)+strlen(ZKLOCK_DEFLT_FILENAME)+2);
        if (!pLock) {
            zkErrorSet_va(err, fn, ZKERR, ZKMLOCK, errno,
                        "error %d allocating %lu bytes of memory for lock file name : %s",
                        errno, (unsigned long)(strlen(path)+strlen(ZKLOCK_DEFLT_FILENAME)+2), strerror(errno));
            goto abort;
        }
        pLock->szLockFile[0] = 0;
        snprintf(pLock->szLockFile, (strlen(path)+strlen(ZKLOCK_DEFLT_FILENAME)+2), 
                 "%s/%s", path, ZKLOCK_DEFLT_FILENAME);
    } else {
        /* we use the file name given by the caller */
        pLock->szLockFile = malloc(strlen(path)+strlen(lockname)+2);
        if (!pLock) {
            zkErrorSet_va(err, fn, ZKERR, ZKMLOCK, errno,
                        "error %d allocating %lu bytes of memory for lock file name : %s",
                        errno, (unsigned long)(strlen(path)+strlen(lockname)+2), strerror(errno));
            goto abort;
        }
        pLock->szLockFile[0] = 0;
        snprintf(pLock->szLockFile, (strlen(path)+strlen(lockname)+2), 
                 "%s/%s", path, lockname);
    }
    
    /* We open the lock file and create it if it does not already exists */
    
    if (!(pLock->fLockFile = fopen(pLock->szLockFile,ZKLOCK_MODE))) {
        zkErrorSet_va(err, fn, ZKERR, ZKMLOCK, errno,
                    "error %d opening file %s : %s",
                    errno,pLock->szLockFile, strerror(errno));
        goto abort;
    }

    *lockHandle = pLock;
    
    return ZK_TRUE;
    
abort:

    if (pLock) {
        if (pLock->szLockFile) { 
            free(pLock->szLockFile);
        }
        free(pLock);
        pLock = NULL;
    }

    return ZK_FALSE;
}

/*--------------------------------------------------------------------------------
** Try to aquire a shared lock on the ressource, return immediately if the
** lock can not be granted.
** Multiples processs can share the same ressource at the same time with
** shared locks. If there is already an exclusive lock, the shared lock
** will not be granted.
**--------------------------------------------------------------------------------
** err         Contain an error if this function fails.
** lock        Lock handle
** lockType    Lock type {F_RDLCK|F_WRLCK}
** granted     Will be set to ZK_TRUE if the lock has been granted.                   
**
** return ZK_FALSE and an explanation in err if the operation fails.
*/
int zkLockRequestNoWait(
    ZK_OUT      ZkError err,
    ZK_IN       ZkLock  lock,
    ZK_IN       short   lockType,
    ZK_OUT      int    *granted ) {

    const char *fn  = "zkLockRequestNoWait";
    int         ret = 0;
    
    *granted = ZK_FALSE;
    
    if (!err) {
        goto abort;
    }

    zkErrorClear(err);
    
    if (!lock) {
        zkErrorSet(err,fn,ZKERR,ZKMLOCK,ZKENULLPTR,"invalid parameter.");
        goto abort;
    }

    if(lockType != F_RDLCK && lockType != F_WRLCK) {
        zkErrorSet(err,fn,ZKERR,ZKMLOCK,ZKERROR,"invalid parameter value.");
        goto abort;
    }

    if (lock->lck) {
        zkErrorSet(err,fn,ZKERR,ZKMLOCK,ZKEALREADY,"The previous lock has not been released.");
        return ZK_FALSE; /* do not go to abort here, we want to keep the current lock ! */
    }
    
    /* We allocate and configure the flock structure used to manage this lock */
    if ( ! (lock->lck = malloc(sizeof(*lock->lck)))) {
        zkErrorSet_va(err, fn, ZKERR, ZKMLOCK, errno,
                    "error %d allocating %lu bytes of memory for lock handle : %s",
                    errno, (unsigned long)sizeof *lock->lck, strerror(errno));
        goto abort;
    }
    
    memset(lock->lck,0,sizeof(*lock->lck));
    
    lock->lck->l_type   = lockType;
    lock->lck->l_start  = 0;
    lock->lck->l_whence = SEEK_SET;
    lock->lck->l_len    = 0;

    
    /* we request the lock. If it can not be granted, we do not wait and return immediately */

    ret = fcntl(fileno(lock->fLockFile), F_SETLK, lock->lck);

    if (ret==-1) {
        /* Lock not granted */
        *granted = ZK_FALSE;
        free(lock->lck);
        lock->lck = NULL;
    } else if (ret != 0) {
      zkErrorSet_va(err, fn, ZKERR, ZKMLOCK, errno,
                  "error %d when trying to lock %s : %s",
                  errno, lock->szLockFile, strerror(errno));
      *granted = ZK_FALSE;
      goto abort;
    } else {
        *granted = ZK_TRUE;
    }

    return ZK_TRUE;

abort:
    if (lock->lck) {
        free(lock->lck);
        lock->lck = NULL;
    }

    return ZK_FALSE;
}


/*--------------------------------------------------------------------------------
** Try to aquire a shared lock on the ressource, return immediately if the
** lock can not be granted.
** Multiples processs can share the same ressource at the same time with
** shared locks. If there is already an exclusive lock, the shared lock
** will not be granted.
**--------------------------------------------------------------------------------
** err         Contain an error if this function fails.
** lock        Lock handle                   
** granted     Will be set to ZK_TRUE if the lock has been granted.                   
**
** return ZK_FALSE and an explanation in err if the operation fails.
*/
DLLExport int zkLockRequestSharNoWait(
    ZK_OUT      ZkError err,
    ZK_IN       ZkLock  lock,
    ZK_OUT      int    *granted ) {

    return zkLockRequestNoWait(err,lock,F_RDLCK,granted );
}

/*--------------------------------------------------------------------------------
** Try to aquire an exclusive lock on the ressource, return immediately if the
** lock can not be granted.
** Only one process can access the locked ressource with an exclusive
** lock. If there is already an exclusive lock or shared locks, this exclusive
** lock will not be granted.
**--------------------------------------------------------------------------------
** err         Contain an error if this function fails.
** lock        Lock handle                   
** granted     Will be set to ZK_TRUE if the lock has been granted.                   
**
** return ZK_FALSE and an explanation in err if the operation fails.
*/
DLLExport int zkLockRequestExclNoWait(
    ZK_OUT      ZkError err,
    ZK_IN       ZkLock  lock,
    ZK_OUT      int    *granted ) {

    return zkLockRequestNoWait(err,lock,F_WRLCK,granted );
}
    
/*--------------------------------------------------------------------------------
** Release any granted lock on the shared ressource.
**--------------------------------------------------------------------------------
** err         Contain an error if this function fails.
** lock        Lock handle                   
**
** return ZK_FALSE and an explanation in err if the operation fails.
*/
DLLExport int zkLockRelease(
    ZK_OUT      ZkError err,
    ZK_IN       ZkLock  lock) {

    const char *fn  = "zkLockRelease";
    int         ret = 0;
    
    if (!err) {
        goto done;
    }

    zkErrorClear(err);
    
    if (!lock) {
        zkErrorSet(err,fn,ZKERR,ZKMLOCK,ZKENULLPTR,"invalid parameter.");
        goto done;
    }

    if (lock->lck) {
        /* we release the current lock */

        lock->lck->l_type   = F_UNLCK;

        if (fcntl(fileno(lock->fLockFile), F_SETLK, lock->lck)== -1) {
          zkErrorSet_va(err, fn, ZKERR, ZKMLOCK, errno,
                      "error %d when trying to unlock %s : %s",
                      errno, lock->szLockFile, strerror(errno));
          goto done;
        } 
    }
   
    ret = ZK_TRUE;

done:
    if (lock->lck) {
        free(lock->lck);
        lock->lck = NULL;
    }

    return ret;
}
    
/*--------------------------------------------------------------------------------
** Free the lock handle. The handle is not usable after this call.
** Another one can be allocated by calling zkLockNew().
**--------------------------------------------------------------------------------
** err         Contain an error if this function fails.
** lock        Lock handle                   
**
** return ZK_FALSE and an explanation in err if the operation fails.
*/
DLLExport int zkLockDelete(
    ZK_OUT      ZkError err,
    ZK_INOUT    ZkLock *lock) {

    const char *fn  = "zkLockDelete";
    
    if (!err) {
        return ZK_FALSE;
    }

    zkErrorClear(err);
    
    if (!lock) {
        zkErrorSet(err,fn,ZKERR,ZKMLOCK,ZKENULLPTR,"invalid parameter.");
        return ZK_FALSE;
    }
    
    if (!(*lock)) {
        zkErrorSet(err,fn,ZKERR,ZKMLOCK,ZKENULLPTR,"invalid parameter.");
        return ZK_FALSE;
    }
    
    /* If we still have a lock, we release it */
    if ((*lock)->lck) {
        zkLockRelease(err,*lock);
    }
    
    /* And we can free the rest of the handle and the handle itself */
    if ((*lock)->fLockFile) {
        fclose((*lock)->fLockFile); 
    }
    
    if ((*lock)->szLockFile) {
        free((*lock)->szLockFile);
    }
    
    free(*lock);
    
    *lock = NULL;
    
    return ZK_TRUE;
}
