/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"
#include "zkIntTypes.h"
#include "zkTypes.h"

#include <stdio.h>
#include <stdlib.h>          /* malloc(), free() */
#include <string.h>
#include <stdarg.h>
#include <zkLog.h>
#include "zkCStr.h"

#ifndef WIN32
#define __USE_XOPEN /* for gethostname() */
#include <unistd.h>
#undef __USE_XOPEN
#else
#include <winsock2.h>
#endif


#include <time.h>

#define	FLAG_ENABLED	  0x1
#define	FLAG_LOGFILE      0x2
#define	FLAG_LOGSTREAM    0x4

#define	__HOSTNAME_SZ	64
#define	__TIME_SZ	24
#define	__MESG_SZ	256
#define	__PROGNAME_SZ	64

struct T_ZkLog {
  int flags;
  FILE * f;
  char progname[__PROGNAME_SZ + 1];
};


/**
 *  This function allows a program to directly write a message to the log.
 *  
 *  @param zklog            The ZkLog object to which to write the message
 *  @param msg              A NUL terminated message string to send to the log
 *  @param vargs            A varaible argement list retrive using va_start.
 *  @version                1.0
 *  @author                 Philippe Boucher
 */

static int zkLogWrite_valist (
  ZkLog        zklog,
  const char * msg,
  va_list      vargs)
{
  struct tm  * utc = NULL;
  char         hostname[__HOSTNAME_SZ + 1] = "-";
  char         timestr[__TIME_SZ + 1] = "-";
  time_t       timep = 0;
  int          pid = 0;
  int          err = 0;
  const char * ostr;

  if (zklog == NULL) {
    return 0;
  }

  if ((!(zklog->flags & FLAG_ENABLED)) || (zklog->f == NULL)) {
    return 0;
  }

  ostr = (msg) ? msg : "";

  timep = time(NULL);
  utc = gmtime(&timep);
  if (utc != NULL) {
    strftime(timestr,__TIME_SZ,"%Y/%m/%d-%H:%M:%S.0",utc);
  }

  err = gethostname(hostname,__HOSTNAME_SZ);
  if (err != 0) {
    zkCStrlcpy(hostname,"-",sizeof hostname);
  }

  pid = getpid(); 

  fprintf(zklog->f,
          "%s %s %d %s \"",
          timestr,
          hostname,
          (int) pid,
          zklog->progname);

  vfprintf(zklog->f,
           ostr,
           vargs);

  fprintf(zklog->f,
          "\"\n");

  fflush(zklog->f);

  return 1;
}

/**
 *  This function creates a new ZkLog struture.
 *
 *  @param zklog       A Pointer to a ZkLog object
 *  @return            A non-zero value on success, a 0 on failure
 *  @version           1.0
 *  @author            Philippe Boucher
 */

DLLExport int
zkLogNew (
  ZkLog * zklog)
{
  if (zklog) {
    *zklog = (ZkLog) malloc(sizeof(struct T_ZkLog));
    if (*zklog == NULL) {
      return 0;
    }
    memset(*zklog,0,sizeof(struct T_ZkLog));
    (*zklog)->flags = 0;
    (*zklog)->f = NULL; 
    zkCStrlcpy((*zklog)->progname,"-",sizeof (*zklog)->progname);
    return 1;
  }
  return 0;
}

/**
 *  This function destroys an existing ZkLog object.  The pointer will be reset
 *  to NULL.
 *
 *  @param zklog            Pointer to ZkLog object to destroy
 *  @return                 A non-zero integer on success, a 0 on failure.
 *  @version                1.0
 *  @author                 Philippe Boucher
 */

DLLExport int 
zkLogFree (
  ZkLog * zklog)
{
  if (zklog) {
    if (*zklog) {
      zkLogDisable(*zklog);
      if (((*zklog)->flags & FLAG_LOGFILE) && ((*zklog)->f)) {
        fclose((*zklog)->f);
      }
      memset(*zklog,0,sizeof(struct T_ZkLog));
      free(*zklog);
      *zklog = NULL;
      return 1;
    }
  }
  return 0;
}

/**
 *  This function allows a program to directly write a message to the log.
 *  
 *  @param zklog            The ZkLog object to which to write the message
 *  @param msg              A NUL terminated message string to send to the log
 *  @version                1.0
 *  @author                 Philippe Boucher
 */

DLLExport int
zkLogWrite_va (
  ZkLog zklog ,
  const char * msg ,
  ... )
{
  int ret = 0;
  va_list ap;

  va_start(ap, msg);
  ret = zkLogWrite_valist(zklog, msg, ap);
  va_end(ap);

  return ret;
}

/**
 *  This function allows a program to directly write a message to the log.
 *  
 *  @param zklog            The ZkLog object to which to write the message
 *  @param msg              A NUL terminated message string to send to the log
 *  @version                1.0
 *  @author                 Philippe Boucher
 */

DLLExport int
zkLogWrite (
  ZkLog zklog ,
  const char * msg )
{
  return zkLogWrite_va(zklog, "%s", msg);
}

/**
 *  This function logs a ZkError object to the given log.
 *
 *  @param zklog           The target ZkLog object
 *  @param error           The ZkError object to log
 *  @return                a non-zero value on success, a 0 on failure
 *  @author                Philippe Boucher
 */

DLLExport int
zkLogError (
  ZkLog zklog,
  ZkError error )
{
  char mesg[__MESG_SZ + 1];

  if ((zklog == NULL) || (error == NULL)) {
    return 0;
  }
  if (!(zklog->flags & FLAG_ENABLED)) {
    return 0;
  }

  mesg[0] = (char) 0;
  zkErrorMessage(error,mesg,__MESG_SZ);
  zkErrorSetLoggedFlag(error, ZK_TRUE);
  return zkLogWrite(zklog,mesg);
}

/**
 *  This function is used to set the program name for the given log.  If this
 *  function is not called, the program name will default to "-".
 * 
 *  @param zklog              The target ZkLog object
 *  @param progname           The name of the current program / process
 *  @return                   A non-zero value on success, a 0 on failure
 *  @version                  1.0
 *  @author                   Philippe Boucher
 */
DLLExport int
zkLogSetProgname (
  ZkLog zklog ,
  const char * progname )
{
  char * tmp = NULL;
  if (zklog == NULL) {
    return 0;
  }
  if (progname == NULL) {
    progname = "-";
  }

  /*
   * If the program name was a full pathname, strip everything out up to and
   * including the last '/' character.  This makes the log easier to read
   */

  tmp = strrchr(progname,'/');
  if (tmp != NULL) {
    progname = &(tmp[1]); 
  }
  strncpy(zklog->progname,progname,__PROGNAME_SZ);
  return 1;
}

/**
 *  This function attaches a log file to the current ZkLog object.  If a log
 *  file was already being used by the object, it will be closed before the
 *  new one is opened.  If the file already exists, it will be appended to,
 *  it will not be truncated.
 * 
 *  @param zklog              The target ZkLog object
 *  @param filename           The full pathname of the logfile to open 
 *  @return                   A non-zero value on success, a 0 on failure
 *  @version                  1.0
 *  @author                   Philippe Boucher
 */

int
zkLogSetFile (
  ZkLog zklog ,
  const char * filename )
{
  if (zklog == NULL) {
    return 0;
  }
  if ((zklog->flags & FLAG_LOGFILE) && (zklog->f)) {
    fclose(zklog->f);
    zklog->flags &= ~FLAG_LOGFILE;
  }
  zklog->f = fopen(filename,"a");
  if (zklog->f == NULL) {
    zkLogDisable(zklog);
    return 0;
  }
  zklog->flags |= FLAG_LOGFILE;
  return 1;
}

/**
 *  Associate the specified file with the current log object.  If the file
 *  cannot be opened, the existing log file will not be closed.
 *  
 *  @param zklog              The target ZkLog object
 *  @param filename           The full pathname of the logfile to open 
 *  @return                   A non-zero value on success, a 0 on failure
 *  @version                  1.0
 *  @author                   Bill Heelan
 */

int
zkLogReopenFile (
  ZkLog zklog ,
  const char * filename)
{
  FILE *f;

  if (zklog == NULL || filename == NULL) {
    return 0;
  }
  f = fopen(filename, "a");
  if (f == NULL) {
    return 0;
  }
  if ((zklog->flags & FLAG_LOGFILE) && (zklog->f)) {
    fclose(zklog->f);
  }
  zklog->f = f;
  zklog->flags |= FLAG_LOGFILE;
  return 1;
}

/**
 *  This function is used to attach an output stream to the ZkLog object.  This
 *  stream can be something like stdout or stderr, or an already open file.
 *  If the ZkLog object already had an open logfile, that old logfile is first
 *  closed.
 * 
 *  @param zklog             The target ZkLog object
 *  @param stream            The new logging IO stream
 *  @return                  A non-zero value on success, a 0 on failure
 *  @version                 1.0
 *  @author                  Philippe Boucher
 */

DLLExport int 
zkLogSetStream (
  ZkLog zklog ,
  FILE * stream )
{
  if ((zklog == NULL) || (stream == NULL)) {
    return 0;
  }
  if ((zklog->flags & FLAG_LOGFILE) && (zklog->f)) {
    zklog->flags &= ~FLAG_LOGFILE;
    fclose(zklog->f);
  }
  zklog->f = stream;
  zklog->flags |= FLAG_LOGSTREAM; 
  return 1;
}

int
zkLogStream (
  ZkLog zklog ,
  FILE ** stream )
{
  if ((zklog) || (stream)) {
    *stream = zklog->f;
    return 1;
  }
  return 0;
}
 

/**
 *  This function turns on the logging for the given object.  By default, the
 *  logging capability of a ZkLog is turned off when it is first created.
 *
 *  @param zklog               The target ZkLog object
 *  @return                    A non-zero value on success, a 0 on failure
 *  @version                   1.0
 *  @author                    Philippe Boucher
 */

DLLExport int 
zkLogEnable (
  ZkLog zklog )
{
  if (zklog) {
    if (zklog->f) {
      zklog->flags |= FLAG_ENABLED;
#if 0
      zkLogWrite(zklog,"Logging turned on");
#endif
      return 1;
    }
  }
  return 0;
}

/**
 *  This function is used to turn off a log.  This does not close the logging
 *  stream however.
 * 
 *  @param zklog             The target ZkLog object
 *  @return                  A non-zero value on success, a 0 on failure
 *  @version                 1.0
 *  @author                  Philippe Boucher
 */

DLLExport int zkLogDisable (
  ZkLog zklog )
{
  if (zklog) {
#if 0
    zkLogWrite(zklog,"Logging turned off");
#endif
    zklog->flags &= ~FLAG_ENABLED;
    return 1;
  }
  return 0;
}


/**
 *  This function dump a context to the log, it as to be use to
 *  dump the current function an a string to represent the context 
 *  of the logging.
 *
 *  @param zklog            ZkLog object to which to write the message
 *  @param err              ZkError object
 *  @param func             Name of the corrent function
 *  @param msg              A NULL terminated message string to send to the log
 *  @version                1.0
 *  @author                 Nicoals Roy
 */

DLLExport int zkLogWriteContext (
  ZkLog        zklog,
  ZkError      err,
  const char * func,
  const char * msg,
  ...
)
{
  ZkBool   isLogged            = ZK_FALSE;
  va_list  vargs               = NULL;
  int      ret                 = 1;
  char     ostr[__MESG_SZ + 1] = { 0 };
  uint16_t module;
  uint16_t errcode;

  if( !zklog ) {
    ret = 0;

  } else {
    module  = zkErrorModule(err);
    errcode = zkErrorCode(err);

    /* Log the current error if it has not been logged yet */
    if( errcode || zkErrorDetail(err)) {
      if( zkErrorGetLoggedFlag(err, &isLogged) && !isLogged) {
        zkLogError(zklog, err);
      }
    }

    /* Build the context oputput string */
    snprintf(ostr, __MESG_SZ, "%02X:%04X %s: %s", 
             module, 
             errcode, 
             ((func) ? func : "-"), 
             ((msg)  ? msg  : "-"));

    /* Log the current context */
    va_start(vargs, msg);
    ret = zkLogWrite_valist(zklog, ostr, vargs);
    va_end(vargs);
  }

  return ret;
}  
