/*
 *  $Id: zkMemBuf.c,v 1.13.10.1 2000/10/11 13:20:02 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *
 *  These functions manage a buffer that grows automatically as you append to
 *  it.
 *  
 *  bug: the routines are not currently designed to support simultaneous reads
 *  and write on the buffer.  I.e. it's not a general stream.  You should
 *  first fill the buffer, then empty it, or vice versa.
 */


#define DEBUG_WRITE 0


#include "zks.h"

#include <errno.h>
#if DEBUG_WRITE
#include <stdio.h>
#endif
#include <string.h>
#include <sys/types.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include "zkAlloc.h"
#include "zkMemBuf.h"
#include "zkDefs.h"


/*
 *  Note: the buffer is empty if firstUsed == nextAvail.
 */

struct zk_membuf {
  char *buf;   /* pointer to the buffer */
  size_t bsz;  /* total size of the buffer */
  size_t firstUsed; /* offset into buffer of the first used byte */
  size_t nextAvail; /* next offset into buffer available for appending */
  int clean;   /* 0: don't memset() buffer before free()  1: do memset() */
};


/* Initial buffer size. */
static const size_t dflBufSz = 1024;

/* Minimum rate at which the buffer grows when more space is needed. */
static const float dflGrowRate = 2.0;


/*
 *
 *
 *                            Internal routines.
 *
 *
 */

static void initMemBuf(ZkMemBuf b)
{
  b->buf = 0;
  b->bsz = 0;
  b->firstUsed = 0;
  b->nextAvail = 0;
  b->clean = 1; /* default to paranoia */
}


/*
 *  Increase the size of the buffer enough to hold `more' bytes.
 *  
 *  This function should only be called if the buffer is currently too small
 *  to hold `more' bytes.
 */
static int incrBufSize(ZkError err, ZkMemBuf b, size_t more)
{
  size_t n = ZK_MAX(dflGrowRate * b->bsz, b->nextAvail + more);

  zkErrorClear(err);

  if ( ! zkMemRealloc(err, (void **)&b->buf, n)) {
    return 0;
  }

  b->bsz = n;

  return 1;
}


static int moveToFront(ZkMemBuf buf)
{
  size_t n = buf->nextAvail - buf->firstUsed;
  
  memmove(buf->buf, buf->buf + buf->firstUsed, n);
  buf->firstUsed = 0;
  buf->nextAvail = n;

  return 1;
}


/*
 *
 *
 *                            External routines.
 *
 *
 */


DLLExport int zkMemBufNew(ZkError err, ZkMemBuf *buf)
{
  ZkMemBuf b;
  
  zkErrorClear(err);

  if ( ! buf) {
    zkErrorSet(err, "zkMemBufNew", ZKMMEMBUF, ZKERR, ZKENULLPTR,
               "buf is null");
    return 0;
  }

  if ( ! ((b = malloc(sizeof *b)))) {
    zkErrorSet(err, "zkMemBufNew", ZKMMEMBUF, ZKERR, ZKENOMEM,
               "can't allocate space for memory buffer object");
    goto fail;
  }

  initMemBuf(b);

  b->bsz = dflBufSz;
  
  if ( ! (b->buf = malloc(b->bsz))) {
    zkErrorSet(err, "zkMemBufNew", ZKMMEMBUF, ZKERR, ZKENOMEM,
               "can't allocate space for memory buffer");
    goto fail;
  }

  *buf = b;

  return 1;

 fail:

  if (b) {
    if (b->buf) { free(b->buf); }
    initMemBuf(b);
    free(b);
  }

  return 0;
}


DLLExport int zkMemBufFree(ZkError err, ZkMemBuf *buf)
{
  ZkMemBuf b;
  
  zkErrorClear(err);

  if ( ! buf) {
    zkErrorSet(err, "zkMemBufFree", ZKMMEMBUF, ZKERR, ZKENULLPTR,
               "buf is null");
    return 0;
  }
  if ( ! *buf) {
    zkErrorSet(err, "zkMemBufFree", ZKMMEMBUF, ZKERR, ZKEPARMBAD,
               "*buf is null");
    return 0;
  }

  b = *buf;

  if (b) {
    if (b->buf) {
      if (b->clean) { memset(b->buf, 0, b->bsz); }
      free(b->buf);
    }
    initMemBuf(b);
    free(b);
    *buf = 0;
  }

  return 1;
}


DLLExport int zkMemBufReset(ZkError err, ZkMemBuf buf)
{
  zkErrorClear(err);

  if ( ! buf) {
    zkErrorSet(err, "zkMemBufReset", ZKMMEMBUF, ZKERR, ZKENULLPTR,
               "buf is null");
    return 0;
  }

  if (buf->clean) { memset(buf->buf, 0, buf->nextAvail); }
  buf->firstUsed = buf->nextAvail = 0;

  return 1;
}


DLLExport int zkMemBufAppend(ZkError err, ZkMemBuf buf, const void *d, size_t dsz)
{
  zkErrorClear(err);

  if ( ! buf) {
    zkErrorSet(err, "zkMemBufAppend", ZKMMEMBUF, ZKERR, ZKENULLPTR,
               "buf is null");
    return 0;
  }
  if ( ! d) {
    zkErrorSet(err, "zkMemBufAppend", ZKMMEMBUF, ZKERR, ZKENULLPTR,
               "d is null");
    return 0;
  }

  if (dsz + buf->nextAvail < buf->bsz) {
    memcpy(buf->buf + buf->nextAvail, d, dsz);
    buf->nextAvail += dsz;
  } else if ( ! incrBufSize(err, buf, dsz)) {
    return 0;
  } else {
    memcpy(buf->buf + buf->nextAvail, d, dsz);
    buf->nextAvail += dsz;
  }

  return 1;
}


/*
 *  Remove the first `dsz' byte from the buffer, shifting down the remainder.
 *  
 *  bug? use a circular buffer?
 */
DLLExport int zkMemBufRemoveLeading(ZkError err, ZkMemBuf buf, size_t dsz)
{
  zkErrorClear(err);

  if ( ! buf) {
    zkErrorSet(err, "zkMemBufRemoveLeading", ZKMMEMBUF, ZKERR, ZKENULLPTR,
               "buf is null");
    return 0;
  }

  if (dsz > buf->nextAvail - buf->firstUsed) {
    zkErrorSet(err, "zkMemBufRemoveLeading", ZKMMEMBUF, ZKERR, ZKEBADSTATE,
               "attempt to remove too much data");
    return 0;
  }

  buf->firstUsed += dsz;

  return moveToFront(buf);
}


DLLExport int zkMemBufGetData(ZkError err, ZkMemBuf buf, void **d, size_t *dsz)
{
  zkErrorClear(err);

  if ( ! buf) {
    zkErrorSet(err, "zkMemBufGetData", ZKMMEMBUF, ZKERR, ZKENULLPTR,
               "buf is null");
    return 0;
  }
  if ( ! d) {
    zkErrorSet(err, "zkMemBufGetData", ZKMMEMBUF, ZKERR, ZKENULLPTR,
               "d is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, "zkMemBufGetData", ZKMMEMBUF, ZKERR, ZKENULLPTR,
               "dsz is null");
    return 0;
  }

  *d = buf->buf + buf->firstUsed;
  *dsz = buf->nextAvail - buf->firstUsed;

  return 1;
}


DLLExport int zkMemBufGetSize(ZkError err, ZkMemBuf buf, size_t *dsz)
{
  zkErrorClear(err);

  if ( ! buf) {
    zkErrorSet(err, "zkMemBufGetSize", ZKMMEMBUF, ZKERR, ZKENULLPTR,
               "buf is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, "zkMemBufGetSize", ZKMMEMBUF, ZKERR, ZKENULLPTR,
               "dsz is null");
    return 0;
  }

  *dsz = buf->nextAvail - buf->firstUsed;

  return 1;
}


#if 0
DLLExport int zkMemBufSetSize(ZkError err, ZkMemBuf buf, size_t dsz)
{
  zkErrorClear(err);

  if ( ! buf) {
    zkErrorSet(err, "zkMemBufSetSize", ZKMMEMBUF, ZKERR, ZKENULLPTR,
               "buf is null");
    return 0;
  }

  /*
   *  bug: if buf->clean is set then we should be paranoid: malloc() a new
   *  buffer, copy what we want, then wipe the old one.  We can't depend on
   *  realloc() not copying the data.
   */

}
#endif

/*
 *  Try to write as much of the buffer contents as possible to the file
 *  descriptor `fd'.  Set `*empty' to 0 there remains more to write; set it to
 *  1 if the entire buffer contents have been written.
 *  
 *  We assume a non-blocking descriptor is being used, so we treat EAGAIN as
 *  equivalent to a write of 0 bytes.
 */
DLLExport int zkMemBufWriteToFd(ZkError err, ZkMemBuf buf, int fd, int *empty)
{
  ssize_t n;
#ifdef WIN32
  DWORD dWord;
#endif
  
  zkErrorClear(err);

  if ( ! buf) {
    zkErrorSet(err, "zkMemBufWriteToFd", ZKMMEMBUF, ZKERR, ZKENULLPTR,
               "buf is null");
    return 0;
  }
  if (fd < 0) {
    zkErrorSet(err, "zkMemBufWriteToFd", ZKMMEMBUF, ZKERR, ZKENULLPTR,
               "fd is null");
    return 0;
  }
  if ( ! empty) {
    zkErrorSet(err, "zkMemBufWriteToFd", ZKMMEMBUF, ZKERR, ZKENULLPTR,
               "empty is null");
    return 0;
  }

#ifdef WIN32

  n= send(fd, buf->buf + buf->firstUsed, buf->nextAvail - buf->firstUsed, 0);

  if(SOCKET_ERROR == n){
    zkErrorSet(err, "zkMemBufWriteToFd", ZKMMEMBUF, ZKERR, ZKENULLPTR,
               "send() failed");
	return 0;
	}
#else
  if ((n = write(fd, buf->buf + buf->firstUsed,
                 buf->nextAvail - buf->firstUsed)) < 0) {
    if (errno == EAGAIN) {
      n = 0;
    } else {
      zkErrorSet_va(err, "zkMemBufWriteToFd", ZKMMEMBUF, ZKERR, errno,
                    "error %d writing %lu bytes: %s",
                    errno, (unsigned long)(buf->nextAvail - buf->firstUsed),
                    strerror(errno));
      return 0;
    }
  }

#if DEBUG_WRITE
  fprintf(stderr, "zkMemBufWriteToFd: wrote %ld bytes.\n", (long)n);
#endif
#endif

  buf->firstUsed += n;
  *empty = buf->firstUsed == buf->nextAvail;

  return 1;
}

