#include <stdexcept>
#include <new>
#include "zkMutex.h"

/*
 * Mutex manipulation
 *
 * $Id: zkMutex.cpp,v 1.2.8.5 2000/12/19 15:13:43 danielf Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 * 
 */


RInterProcessMutex::RInterProcessMutex( const char *pName ) : m_count( 0 ),
                                                              m_sem( NULL )
{
  uint32_t  flags = ( ZKSEM_USER_READ_PMODE | ZKSEM_USER_WRITE_PMODE ) ;

  if ( pName == NULL )
    flags |= ZKSEM_PRIVATE_KEY ;

  ZkError  err ;

  if ( ! zkErrorNew( &err )                       ||
       ! zkSemNew( err, &m_sem, pName, 1, flags )   )
    throw std::bad_alloc( ) ;

  zkErrorFree( &err ) ;
}


RInterProcessMutex::~RInterProcessMutex( )
{
  if ( m_sem )
    zkSemFree( NULL, &m_sem, false ) ;
}


bool  RInterProcessMutex::take( )
{
  if ( this == NULL || m_sem == NULL || m_count < 0 )
    return false ;

  if ( m_count == 0 && ! zkSemWait( NULL, m_sem ) )
    return false ;

  m_count++ ;

  return true ;
}


bool  RInterProcessMutex::tryTake( bool *pbGotIt )
{
  if ( this == NULL || m_sem == NULL || m_count < 0 || pbGotIt == NULL )
    return false ;

  ZkBool  bGotIt = true ;

  if ( m_count == 0 && ! zkSemTryWait( NULL, m_sem, &bGotIt ) )
  {
    return false ;
  }

  if ( bGotIt )
  {
    m_count++ ;
  }

  *pbGotIt = bGotIt ;

  return true ;
}


bool  RInterProcessMutex::give( )
{
  if ( this == NULL || m_sem == NULL || m_count < 1 )
    return false ;

  if ( --m_count == 0 && ! zkSemRelease( NULL, m_sem ) )
  {
    m_count = 1 ; // It did not work --> restore count
    return false ;
  }
  return true ;
}


bool  RInterProcessMutex::cleanup( const char *pName )
{
  if ( pName == NULL )
    return false ;

  return zkSemCleanup( NULL, pName ) ;
}


// =======================================
//  C   L A N G U A G E   W R A P P E R S
// =======================================

extern "C"
{
  DLLExport ZkBool  zkInterProcessMutexNew( ZkMtx       *pMtx,   // [out]  pointer on a ZkMtx
                                            const char  *pName ) // [in]   unique string to identify the mutex
  {
    if ( pMtx == NULL )
      return false ;

    *pMtx = ( ZkMtx )( new RInterProcessMutex( pName ) ) ;

    return *pMtx != NULL ;
  }


  DLLExport ZkBool  zkInterProcessMutexFree( ZkMtx  *pMtx ) // [out]  pointer on a ZkMtx
  {
    if ( pMtx == NULL || *pMtx == NULL )
      return false ;

    delete ( *( RInterProcessMutex ** )pMtx ) ;
    *pMtx = NULL ;

    return true ;
  }


  DLLExport ZkBool  zkMtxTake( ZkMtx  mtx ) // [in]  ZkMtx
  {
    if ( mtx == NULL )
      return ZK_FALSE ;

    return ( ( RMutex * )mtx )->take( ) ;
  }


  DLLExport ZkBool  zkMtxTryTake( ZkMtx   mtx,     // [in]     ZkMtx
                                  ZkBool   *pGotIt ) // [in/out] pointer on a ZkBool to represent if the waiting success
  {
    if ( mtx == NULL || pGotIt == NULL )
      return false ;

    bool        bGotIt ;
    const bool  bRc = ( ( RMutex * )mtx )->tryTake( &bGotIt ) ;

    if ( bRc )
      *pGotIt = bGotIt ;
 
    return bRc ;
  }


  DLLExport ZkBool  zkMtxGive( ZkMtx  mtx ) // [in]  ZkMtx
  {
    if ( mtx == NULL )
      return ZK_FALSE ;

    return ( ( RMutex * )mtx )->give( ) ;
  }


  DLLExport ZkBool  zkInterProcessMutexCleanup( const char  *pName ) // [in]   unique string to identify the mutex
  {
    return  RInterProcessMutex::cleanup( pName ) ;
  }
} ;
