/*
 *  $Id: zkNameTable.c,v 1.2.20.1 2000/10/11 13:20:02 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <zkError.h>
#include <zkList.h>
#include <zkCStr.h>
#include <zkNameTable.h>

struct T_ZkNameTableElement {
  char * name;    /* Previous node in list */
};

typedef struct T_ZkNameTableElement * ZkNameTableElement;

struct T_ZkNameTable {
  ZkList element_list;
};


static int ZkNameTableElementNew(ZkError , ZkNameTableElement *);  
static int ZkNameTableElementFree(ZkError , ZkNameTableElement);  

static int ZkNameTableElementFreeFn(ZkError , ZkList , void * , void *);  

static int 
ZkNameTableElementNew (
  ZkError err, 
  ZkNameTableElement * element)
{
  const char * func = "ZkNameTableElementNew";
  int ret = 1;

  zkErrorClear(err);

  *element = (ZkNameTableElement) malloc (sizeof (struct T_ZkNameTableElement));

  if (!(*element)) {
    zkErrorSystemSet (err, func, ZKERROR, ZKMNAMETABLE, "Error allocating ZkNameTableElement");
    /* Log error... */
    /* fprintf (stderr, "[%s]: Error allocating ZkNameTableElement\n", func); */
    ret = 0;
  }

  else {
    (*element)->name = NULL;
  }
  
  return ret;
}


static int 
ZkNameTableElementFree (
  ZkError err, 
  ZkNameTableElement element)
{
  const char * func = "ZkNameTableElementFree";
  int ret = 1;

  (void)func;

  zkErrorClear(err);
  
  if (element->name != NULL) {
    free (element->name);
    element->name = NULL;
  }

  free (element);

  return ret;
}


static int 
ZkNameTableElementFreeFn  (
  ZkError err, 
  ZkList list, 
  void * table_element, 
  void * not_used)
{
  const char * func = "ZkNameTableElementFree";
  int ret = 1;

  (void)func;
  (void)list;
  (void)not_used;

  zkErrorClear(err);
  
  ret = ZkNameTableElementFree (err, (ZkNameTableElement)table_element);

  return ret;
}
 

/**
 *  Create a new zkNameTable structure.
 *
 *  @author Philippe Desaulniers
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int 
zkNameTableNew (
  ZkError err,              /* Returns the error if any occured */
  ZkNameTable * table )     /* Returns a new zk name table */
{
  const char * func = "zkNameTableNew";
  int ret = 1;

  zkErrorClear(err);
  
  *table = (ZkNameTable) malloc (sizeof(struct T_ZkNameTable));

  if (!(*table)) {
    zkErrorSystemSet (err, func, ZKERROR, ZKMNAMETABLE, "Error allocating ZkNameTable");
    /* Log error... */
    /* fprintf (stderr, "[%s]: Error allocating ZkNameTable\n", func); */
    ret = 0;
  }

  else if (!zkListNew (err, &(*table)->element_list)) {
    free(*table);
    *table = NULL;
    /* Log error... */
    /* fprintf (stderr, "[%s]: zkListNew failed\n", func); */
    ret = 0;
  }

  else if (!zkListSetDataFreeFunction (err, (*table)->element_list, 
                                       ZkNameTableElementFreeFn, NULL)) { 
    free(*table);
    *table = NULL;
    /* Log error... */
    /* fprintf (stderr, "[%s]: zkListSetDataFreeFunction failed\n", func); */
    ret = 0;
  }

  return ret;
}


/**
 *  Destroy a zkNameTable structure.
 *
 *  @author Philippe Desaulniers
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int 
zkNameTableFree (
  ZkError err,              /* Returns the error if any occured */
  ZkNameTable table )       /* Name table to destroy */
{
  const char * func = "zkNameTableFree";
  int ret = 1;

  (void)func;

  zkErrorClear(err);

  if (!zkListFree (err, &(table->element_list))) {
    /* Log error... */
    /* fprintf (stderr, "[%s]: zkListFree failed\n", func); */
    ret = 0;
  }

  else {
    free (table);
  }

  return ret;
}


/**
 *  Add an element to zkNameTable structure.
 *
 *  @author Philippe Desaulniers
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int 
zkNameTableAddElement (
  ZkError err,               /* Returns the error if any occured */
  ZkNameTable table,         /* Table to which we want to add an item */
  const char * element_name) /* Name of the item to add */ 
{
  const char * func = "zkNameTableAddElement";
  int ret = 1;
  int elementInListFlag = 0;

  ZkNameTableElement new_element = NULL;

  zkErrorClear(err);
  
  if (!element_name) {
    zkErrorSet(err,func,ZKERR,ZKMNAMETABLE,ZKENULLPTR,"NULL parameter");
    ret = 0;
  }

  /* Verify element is not yet in list */
  else if (!zkNameTableIsElement (err, table, element_name, &elementInListFlag)) {
    /* Log error... */
    /* fprintf (stderr, "[%s]: zkNameTableIsElement failed\n", func); */
    ret = 0;
  }

  else if (elementInListFlag) {
    zkErrorSet_va(err,func,ZKERR,ZKMNAMETABLE,ZKEDATABAD,
                  "Element %s already in table", element_name);
    /* Log error... */
    /* fprintf (stderr, "[%s]: Element %s already in table\n", func, element_name); */
    ret = 0;
  }

  /* Create new element */
  else if (!ZkNameTableElementNew(err, &new_element)){
    /* Log error... */
    /* fprintf (stderr, "[%s]: ZkNameTableElementNew failed\n", func); */
    ret = 0;
  }

  /* Duplicate name string */
  else if (!(new_element->name = zkCStrDup(element_name))) {
    zkErrorSet_va(err,func,ZKERR,ZKMNAMETABLE,ZKMALLOC,
                  "Could not allocate string for element %s", element_name);
    ZkNameTableElementFree (err, new_element);
    new_element = NULL;
    /* Log error... */
    /* fprintf (stderr, "[%s]: zkCStrDup failed\n", func); */
    ret = 0;
  }

  /* Add to list */
  else if (!zkListInsert (err, table->element_list, new_element, LIST_TAIL)) {
    ZkNameTableElementFree (err, new_element);
    new_element = NULL;
    /* Log error... */
    /* fprintf (stderr, "[%s]: zkListInsert failed\n", func); */
    ret = 0;
  }

  return ret;
}


/**
 *  Verify an element belongs to a zkNameTable structure.
 *
 *  @author Philippe Desaulniers
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int 
zkNameTableIsElement (
  ZkError err,               /* Returns the error if any occured */
  ZkNameTable table,         /* Table which we want to check for existing element */
  const char * element_name, /* Name of the item to verify */ 
  int * result)              /* Boolean, verification result */
{
  const char * func = "zkNameTableIsElement";
  int ret = 1;
  int size = 0;
  ZkNameTableElement currentElement = NULL;

  zkErrorClear(err);

  *result = 0;

  if (!element_name || !result) {
    zkErrorSet(err,func,ZKERR,ZKMNAMETABLE,ZKENULLPTR,"NULL parameter");
    ret = 0;
  }
  
  else if (!zkListSize (err, table->element_list, &size)) {
    /* Log error... */
    /* fprintf (stderr, "[%s]: zkListSize failed\n", func); */
    ret = 0;
  }

  else if (size > 0) {
    if (!zkListMoveCursor (err, table->element_list, LIST_HEAD)) {
      /* Log error... */
      /* fprintf (stderr, "[%s]: zkListMoveCursor failed\n", func); */
      ret = 0;
    }

    if (zkListGet (err, table->element_list, (void **)&currentElement, LIST_STAY)) {
      while ((*result == 0) && (currentElement != NULL)) {
        if (strcmp (currentElement->name, element_name) == 0) {
          *result = 1;
        }

        else if (!zkListMoveCursor (err, table->element_list, LIST_FORWARDS)) {
          /* Log error... */
          /* fprintf (stderr, "[%s]: zkListMoveCursor failed\n", func); */
          ret = 0;
        }

        else if (!zkListGet (err, table->element_list, (void **)&currentElement, LIST_STAY)) {
          currentElement = NULL;
        }
      }
    }
  }

  return ret;
}


/**
 *  Deletes an element from a zkNameTable structure.
 *
 *  @author Philippe Desaulniers
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int 
zkNameTableDeleteElement (
  ZkError err,               /* Returns the error if any occured */
  ZkNameTable table,         /* Name table from which we want to delete */
  const char * element_name) /* Name of element to delete */
{
  const char * func = "zkNameTableDeleteElement";
  int ret = 1;
  int found = 0;
  ZkNameTableElement currentElement = NULL;

  zkErrorClear(err);

  if (!element_name) {
    zkErrorSet(err,func,ZKERR,ZKMNAMETABLE,ZKENULLPTR,"NULL parameter");
    ret = 0;
  }
  
  else if (!zkListMoveCursor (err, table->element_list, LIST_HEAD)) {
    /* Log error... */
    /* fprintf (stderr, "[%s]: zkListMoveCursor failed\n", func); */
    ret = 0;
  }

  else if (zkListGet (err, table->element_list, (void **)&currentElement, LIST_STAY)) {
    while ((ret == 1) && (found == 0) && (currentElement != NULL)) {
      if (strcmp (currentElement->name, element_name) == 0) {
        found = 1;
      }

      else if (!zkListMoveCursor (err, table->element_list, LIST_FORWARDS)) {
        /* Log error... */
        /* fprintf (stderr, "[%s]: zkListMoveCursor failed\n", func); */
        ret = 0;
      }

      else if (!zkListGet (err, table->element_list, (void **)&currentElement, LIST_STAY)) {
        currentElement = NULL;
      }
    }
  }

  if (found) {
    if (!zkListDelete (err, table->element_list, LIST_HEAD)) {
      /* Log error... */
      /* fprintf (stderr, "[%s]: zkListDelete failed\n", func); */
      ret = 0;
    }
  }

  else {
    zkErrorSet_va(err,func,ZKERR,ZKMNAMETABLE,ZKEDATABAD,
                  "Element %s not found in table", element_name);
    /* Log error */
    /* fprintf (stderr, "[%s]: Element %s not found in table\n", func, element_name); */
    ret = 0;
  }

  return ret;
}



