/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkQueue.h>

/**
 * Generic Queue.  Holds pointers to real structures.
 */
struct T_ZkQueue {
                                /** Capacity of the queue */ 
    int maxElements;
                                /** Number of elements currently in queue */
    int numElements;
                                /** Front of the queue */
    int front;
                                /** Rear of the queue */
    int rear;           
                                /** Pointer to an element in the queue */
    void **ptr;         
};

/**
 * Initializes a ZkQueue with a maximum number of elements (max).
 * 
 * @param err        ZkError object
 * @param Q          Pointer to ZkQueue object
 * @param max        Maximum number of elements for this queue
 * @version          1.0
 * @return           Non-zero on success, 0 on failure
 * @author           Kent Tse
 */
DLLExport int zkQueueNew(ZkError err, ZkQueue *Q, int max)
{
    int i;

    if (Q) {
        *Q = (ZkQueue) malloc(sizeof(struct T_ZkQueue));
        if (*Q == NULL) {
            zkErrorSet(err, "zkQueueNew", ZKERR, ZKMQUEUE, ZKENULLPTR, "ZkQueue malloc failed");
            return 0;
        }
        (*Q)->ptr = (void **) malloc(max * sizeof(void *));
        if ((*Q)->ptr == NULL) {
            zkErrorSet(err, "zkQueueNew", ZKERR, ZKMQUEUE, ZKENULLPTR, "void malloc failed");
            return 0;
        }
        (*Q)->maxElements = max;
        (*Q)->front = 1;
        (*Q)->rear = 0;
        (*Q)->numElements = 0;

        memset((*Q)->ptr, 0, max * sizeof(void *));
        for (i = 0; i < max; i++) {
            (*Q)->ptr[i] = 0;
        }
        return 1;
    }
    zkErrorSet(err, "zkQueueNew", ZKERR, ZKMQUEUE, ZKENULLPTR, "ZkQueue * NULL");
    return 0;
}

/**
 * Returns 1 on a empty Queue, 
 *        -1 on a non-empty queue,
 *         0 on error.
 * 
 * @param err        ZkError object
 * @param Q          Pointer to ZkQueue object
 * @version          1.0
 * @return           1 on an empty queue, -1 on non-empty queue, 0 on failure
 * @author           Kent Tse
 */
int zkQueueIsEmpty(ZkError err, ZkQueue Q)
{
    if (Q == NULL) {
        zkErrorSet(err, "zkQueueIsEmpty", ZKERR, ZKMQUEUE, ZKENULLPTR, "ZkQueue * NULL");
        return 0;
    }
    if (Q->numElements == 0) {
        return 1;
    }
    return -1;
}

/**
 * Returns a pointer pointer to the element removed from the queue.
 * Returns NULL on error.
 * 
 * @param err        ZkError object
 * @param Q          ZkQueue object
 * @version          1.0
 * @return           1 on an empty queue, -1 on non-empty queue, 0 on failure
 * @author           Kent Tse
 */
DLLExport void **zkQueueDequeue(ZkError err, ZkQueue Q)
{
    int empty;
    void **retPtr;

    if (Q == NULL) {
        zkErrorSet(err, "zkQueueDequeue", ZKERR, ZKMQUEUE, ZKENULLPTR, "ZkQueue * NULL");
        return NULL;
    }

    if (!(empty = zkQueueIsEmpty(err, Q))) {
        zkErrorSet(err, "zkQueueDequeue", ZKERR, ZKMQUEUE, ZKENULLPTR, "zkQueueIsEmpty failed");
        return NULL;            /* Error */
    } else if (empty == 1) {
        zkErrorSet(err, "zkQueueDequeue", ZKERR, ZKMQUEUE, ZKENULLPTR, "zkQueueIsEmpty");
        return NULL;            /* Empty Queue */
    }

    retPtr = Q->ptr[Q->front];
    if (Q->front == Q->maxElements - 1) {
        Q->front = 0;        /* Wrap around */
    } else {
        Q->front = Q->front + 1;
    }
    Q->numElements = Q->numElements - 1;
    
    return(retPtr);
}

/**
 * Insert element into the queue.
 * 
 * @param err        ZkError object
 * @param Q          ZkQueue object
 * @param ptr        Pointer pointer of element to insert
 * @version          1.0
 * @return           non-zero on success, 0 on failure
 * @author           Kent Tse
 */
DLLExport int zkQueueEnqueue(ZkError err, ZkQueue Q, void **ptr)
{
    if (Q == NULL) {
        zkErrorSet(err, "zkQueueEnqueue", ZKERR, ZKMQUEUE, ZKENULLPTR, "ZkQueue * NULL");
        return 0;
    }

    if (Q->numElements == Q->maxElements) {
        zkErrorSet(err, "zkQueueEnqueue", ZKERR, ZKMQUEUE, ZKENOMEM, "ZkQueue full");
        return 0;               /* At capacity */
    }

    /* Increment number of elements in the queue */
    Q->numElements = Q->numElements + 1;

    /* Update rear pointer / watch wraparound */
    if (Q->rear == (Q->maxElements - 1)) {
        Q->rear = 0;         /* Wrap around */
    } else {
        Q->rear = Q->rear + 1; /* Just increment it */
    }

    /* Set the pointer */
    Q->ptr[Q->rear] = ptr;
        
    return 1;
}

/**
 * Free the ZkQueue structure.
 * 
 * @param err        ZkError object
 * @param Q          Pointer to ZkQueue object
 * @version          1.0
 * @return           non-zero on success, 0 on failure
 * @author           Kent Tse
 */
DLLExport int zkQueueFree(ZkError err, ZkQueue *Q)
{
    if (Q == NULL) {
        zkErrorSet(err, "zkQueueFree", ZKERR, ZKMQUEUE, ZKENOMEM, "ZkQueue * NULL");
        return 0;
    }
    if (*Q == NULL) {
        zkErrorSet(err, "zkQueueFree", ZKERR, ZKMQUEUE, ZKENOMEM, "ZkQueue NULL");
        return 0;
    }
    if ((*Q)->ptr) {
        free((*Q)->ptr);
    }

    free(*Q);
    *Q = NULL;
    return 1;
}

/**
 * Return the capacity of the queue.
 *
 * @param err        ZkError object
 * @param Q          ZkQueue object
 * @param *cap       Capacity of the queue
 * @version          1.0
 * @return           non-zero on success, 0 on failure
 * @author           Kent Tse
 */
DLLExport int zkQueueCapacity(ZkError err, ZkQueue Q, unsigned int *cap)
{
    if (Q) {
        *cap = Q->maxElements - Q->numElements;
        return 1;
    }
    zkErrorSet(err, "zkQueueCapacity", ZKERR, ZKMQUEUE, ZKENOMEM, "ZkQueue NULL");
    return 0;
}

/**
 * Print out the entire queue elements.  Uses a user-supplied
 * callback to get the actual void ** part of the structure.
 *
 * @param err        ZkError object
 * @param Q          ZkQueue object
 * @param callback   Callback function to print the void structure
 * @version          1.0
 * @return           non-zero on success, 0 on failure
 * @author           Kent Tse
 */
int zkQueueDump(ZkError err, ZkQueue Q, ZkQueueDumpCallback callback)
{
    int i, front;

    if (Q) {
        fprintf(stdout, "ZkQueue maxElement:  %d\n", Q->maxElements);
        fprintf(stdout, "ZkQueue numElement:  %d\n", Q->numElements);
        fprintf(stdout, "ZkQueue front index: %d\n", Q->front);
        fprintf(stdout, "ZkQueue rear index:  %d\n", Q->rear);

        front = Q->front;
        for (i = 0; i < Q->numElements; i++) {
            callback(err, Q->ptr[front]);
            if (front == (Q->maxElements - 1)) {
                front = 0;
            } else {
                front = front + 1;
            }
        }
    }
    return 1;
}
