/*
** **************************************************************************
** $Id: zkSharedObject.c,v 1.1.32.1 2000/10/11 13:20:02 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

**
** Copyright (C) 1999 Zero-Knowledge Systems Inc.  All Rights Reserved
** **************************************************************************
*/

# include <zks.h>
# include "zkAlloc.h"
# include "zkError.h"
# include "zkSharedObject.h"

# ifndef WIN32
# include <dlfcn.h>
# endif

/*
 * Private definitions of opaque data structures
 */

struct t_ZkSharedObject {
# ifdef WIN32
    HINSTANCE           handle;
# else
    void *              handle;
# endif
};


/* --------------------------------------------------------------------------
 * zkSharedObjectLoad
 * --------------------------------------------------------------------------
 *
 */
DLLExport int
zkSharedObjectLoad(
    ZkError             err,
    const char *        file,
    ZkSharedObject *    object )
{
    static const char * func = "zkSharedObjectLoad";
    if (!(object && file))
    {
        zkErrorSet(err,func,ZKERR,ZKMSHAREDOBJECT,ZKENULLPTR,
                   "input params are null");
    }
    else if (zkMemAlloc(err,(void**)object, sizeof(**object)))
    {
#       ifdef WIN32
        (*object)->handle = LoadLibrary(file);
#       else
        (*object)->handle = dlopen(file, RTLD_NOW);
# endif
        
        if ((*object)->handle != NULL)
        {
            return 1;
        }

#       ifdef WIN32
        zkErrorSet_va(err,func,ZKERR,ZKMSHAREDOBJECT,GetLastError(),
                   "could not load shared object: %s",file);
#       else
        zkErrorSet(err,func,ZKERR,ZKMSHAREDOBJECT,ZKELIBACC,dlerror());
#       endif
        free(*object);
        *object = NULL;
    }

    return 0;
}



/* --------------------------------------------------------------------------
 * zkSharedObjectUnload
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkSharedObjectUnload(
    ZkError              err,
    ZkSharedObject *     object )
{
    static const char * func = "zkSharedObjectUnload";
    if (!(object && *object))
    {
        zkErrorSet(err,func,ZKERR,ZKMSHAREDOBJECT,ZKENULLPTR,
                   "input params are null");
        return 0;
    }

# ifdef WIN32
    FreeLibrary((*object)->handle);
# else
    dlclose((*object)->handle);
# endif

    free(*object);
    *object = NULL;

    return 1;
}



/* --------------------------------------------------------------------------
 * zkSharedObjectFindSymbol
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkSharedObjectFindSymbol(
    ZkError              err,
    ZkSharedObject       object,
    const char *         symbol,
    void **              location )
{
    static const char * func = "zkSharedObjectFindSymbol";

    if (!(object && symbol && location))
    {
        zkErrorSet(err,func,ZKERR,ZKMSHAREDOBJECT,ZKENULLPTR,
                   "input params are null");
    }
    else
    {
#   ifdef WIN32
        *location = GetProcAddress(object->handle, symbol);
#   else
        char * errormsg = NULL;
        *location = dlsym(object->handle, symbol);
#   endif
        
        if (*location)
        {
            return 1;
        }
        
#   ifdef WIN32
        zkErrorSet_va(err,func,ZKERR,ZKMSHAREDOBJECT,GetLastError(),
                      "cannot find symbol: %s",symbol);
#   else
        errormsg = dlerror();
        if (errormsg)
        {
            zkErrorSet(err,func,ZKERR,ZKMSHAREDOBJECT,ZKELIBACC,errormsg);
        }
#   endif
    }

    return 0;
}
