/*
 * $Id: zkShell.c,v 1.12.32.2 2000/10/11 13:20:02 phil Exp $
 */

static const char * module_id_  = "$Id: zkShell.c,v 1.12.32.2 2000/10/11 13:20:02 phil Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkHash.h>
#include <zkShell.h>

#define CHAR_DOLLAR '$'
  
#define MAX_TOKENS          32
#define CMDTOKEN_SZ         32
#define PROMPT_SZ           32

struct T_ZkShellCommand {
  char cmdtoken[CMDTOKEN_SZ + 1];
  char * usage;
  char * helpstr;
  ZkShellCommandCallback callback;
  uint32_t pmask;
  struct T_ZkShellCommand * link;
};

typedef struct T_ZkShellCommand ZkShellCommand;

#ifdef ZKSHELL_USE_GNU_READLINE
#warning Compiling with readline

#include <stdio.h>
#include <readline/readline.h>
#include <readline/history.h>

/* **************************************************************** */
/*                                                                  */
/*                    How to emulate the gets function.             */
/*                                                                  */
/* **************************************************************** */

/* A static variable for holding the line. */
static char *line_read = (char *)NULL;

/* Read a string, and return a pointer to it.  Returns NULL on EOF. */
   char *fgets(char *s, int size, FILE *stream);
static char *rl_fgets(char *buf,int size,FILE *stream, char *prompt)
{
  rl_instream = stream;

  /* If the buffer has already been allocated, return the memory
     to the free pool. */

  if (line_read)
    {
      free (line_read);
      line_read = (char *)NULL;
    }

  /* Get a line from the user. */
  line_read = readline (prompt);

  /* If the line has any text in it, save it on the history. */
  if (line_read && *line_read) {
    if (strlen (line_read) > (unsigned int) size) {
      line_read[size] = '\0';
    }
    add_history (line_read);
    strncpy(buf, line_read, size);
  } else {
    *buf = '\0';
  }
  return (line_read);
}
#endif
     
struct T_ZkShell {
  uint32_t flags;
  ZkLog log; 
  FILE * input;
  FILE * output;
  char prompt[PROMPT_SZ + 1];
  ZkShellCommand * command;
  size_t commands;
  void * context;
  ZkHash variables;         /* Holds variables for the shell session */
};

static const struct T_StandardCommand {
  const char * token;
  const char * usage;
  const char * hlpst; 
  ZkShellCommandCallback callback;
  uint32_t pmask;
} StandardCommand[] = {
  {  "quit",
     "quit",
     "Terminate the current shell session",
     zkShellCommandQuit ,
     ZKSHELL_PARMS_NONE },
  {  "help" ,
     "help [ command ]",
     "When the \"command\" parameter is not specified, help will list all "
     "commands which are currently available in the shell.  If a command is "
     "specified, a complete description of the command will be displayed" ,
      zkShellCommandHelp ,
     ZKSHELL_PARMS_NONE | ZKSHELL_PARMS_1 },
  {  "?" ,
     "? [ command ] " ,
     "This is an alias for the help command.",
     zkShellCommandHelp ,
     ZKSHELL_PARMS_NONE | ZKSHELL_PARMS_1 },
  {  "prompt",
     "prompt string",
     "This resets the current shell prompt to the specified string.",
     zkShellCommandPrompt,
     ZKSHELL_PARMS_1 }, 
  {  "echo" ,
     "echo [ parm1 ... parmN ]",
     "Echo a line to the screen",
     zkShellCommandEcho ,
     0xFFFFFFFF },
  {  "source",
     "source filename",
     "Read a file of commands and execute them",
     zkShellCommandSource ,
     ZKSHELL_PARMS_1 },
  {  "set" ,
     "set [ name value ]",
     "Set a variable called \"name\" to the specified value string.  If no "
     "name and value pair is specified, all currently set variables will be "
     "listed.",
     zkShellCommandSet ,
     ZKSHELL_PARMS_2 | ZKSHELL_PARMS_NONE },
  {  "unset",
     "unset name",
     "Unset a variable from the shell environment.\n",
     zkShellCommandUnset ,
     ZKSHELL_PARMS_1 } 
};
  
#define STDCMDS (sizeof(StandardCommand) / sizeof(struct T_StandardCommand))

/*
 * Internal function prototypes.
 */

DLLExport size_t zkShellVariableHash ( ZkHash , void * , void * );
DLLExport int zkShellVariableFree ( ZkHash , void * , void * );
DLLExport int zkShellVariableTraverse ( ZkError , void * , void *, void * );
static int zkShellTokenizeLine ( ZkError, char *, size_t, char **, size_t *);


/**
 *  Hash function modified from the default hash function in zkHash.  It
 *  assumes a string and hashes until the NUL termination.
 */

DLLExport size_t 
zkShellVariableHash (
  ZkHash h ,
  void * key ,
  void * userdata )
{
  char *p = key;
  size_t v = 0, i;

  h = NULL; /* Gets rid of compile time warnings */
  userdata = NULL;
  for (i = 0; p[i]; i++) {
    v = v * 33 + p[i];
  }
  return v;
}

DLLExport int 
zkShellVariableFree (
  ZkHash h ,
  void * data ,
  void * userdata )
{
  userdata = NULL; /* Gets rid of compile time warnings */
  if (h && data) {
    free(data);
    return 1;
  }
  return 0;
}

DLLExport int
zkShellVariableTraverse (
  ZkError err ,
  void * key ,
  void * data ,
  void * userdata )
{
  ZkShell shell = NULL;

  shell = (ZkShell) userdata;
  zkShellWrite_va(err,shell,"%s=\"%s\"\n",(char *) key,(char *) data);
  return 1;
}  

/**
 *   Add a context to the shell structure.  This pointer is passed to 
 *   all shell commands so that application specific features can be
 *   implemented.
 */

DLLExport int
zkShellSetContext (
  ZkError err ,
  ZkShell shell ,
  void * context )
{
  char *echocmd_str = strdup (ZKSHELL_ECHO_CMD);
  char *no_str = strdup (ZKSHELL_NO_STR);
  int retcode = 0;
  
  zkErrorClear(err);
  if (shell && context) {
    shell->context = context;
    /* Slight hack - should have a shell init routine */
    if ( ! zkShellVariableSet (err, shell, echocmd_str, no_str) ) {
      retcode = 0;
    }
    retcode = 1;
  } else {
    zkErrorSet(err,"zkShellSetContext",ZKERR,ZKMSHELL,ZKENULLPTR,NULL);
    retcode = 0;
  }
  free(echocmd_str);
  free(no_str);
  return retcode;
}

/**
 *  This sets the shell's input, output and error streams.  By default these
 *  are stdin, stdout, and error.  Eventually, a I/O callback mechanism must
 *  be added in here.
 */

DLLExport int
zkShellSetStreams (
  ZkError err ,
  ZkShell shell ,
  FILE * input ,
  FILE * output )
{
  zkErrorClear(err);
  if (shell && input && output) {
    shell->input = input;
    shell->output = output;
    return 1;
  }
  zkErrorSet(err,"zkShellSetStreams",ZKERR,ZKMSHELL,ZKENULLPTR,NULL);
  return 0;
}



/**
 *  Set the shell prompt string.  
 */

DLLExport int
zkShellSetPrompt (
  ZkError err ,
  ZkShell shell ,
  const char * prompt )
{ 
  zkErrorClear(err);
  if (shell && prompt) {
    strncpy(shell->prompt,prompt,PROMPT_SZ);
    shell->prompt[PROMPT_SZ] = (char) 0;
    return 1;
  }
  zkErrorSet(err,"zkShellSetPrompt",ZKERR,ZKMSHELL,ZKENULLPTR,NULL);
  return 0;
} 

/**
 *
 */

DLLExport int
zkShellAddCommand (
  ZkError err ,
  ZkShell shell ,
  ZkShellCommandCallback command ,
  const char * cmdtoken ,
  const char * usage ,
  const char * helpstr ,
  uint32_t pmask )
{
  ZkShellCommand * cmd = NULL;

  zkErrorClear(err);
  if (shell && command && cmdtoken) {
    if ((cmd = (ZkShellCommand *) malloc(sizeof(ZkShellCommand))) == NULL) {
      zkErrorSet(err,"zkShellAddCommand",ZKERR,ZKMSHELL,ZKENOMEM,"malloc");
      return 0;
    }
    strncpy(cmd->cmdtoken,cmdtoken,CMDTOKEN_SZ);
    cmd->cmdtoken[CMDTOKEN_SZ] = (char) 0;
    cmd->callback = command;
    cmd->pmask = pmask;
    if (usage) {
      cmd->usage = (char *) strdup(usage);
    }
    if (helpstr) {
      cmd->helpstr = (char *) strdup(helpstr);
    }
    cmd->link = shell->command;
    shell->command = cmd;
    shell->commands += 1;
    return 1;
  }
  zkErrorSet(err,"zkShellAddCommand",ZKERR,ZKMSHELL,ZKENULLPTR,NULL);
  return 0;
}
    
/**
 *
 */  

DLLExport int
zkShellNew (
  ZkError err ,
  ZkShell * shell )
{
  int i = 0; /* Loop index for adding in the standard commands */

  (void) module_id_; /* Get rid of compile warning */
  zkErrorClear(err);
  if (shell) {
    *shell = (ZkShell) malloc(sizeof(struct T_ZkShell));
    if (*shell == NULL) {
      zkErrorSet(err,"zkShellNew",ZKERR,ZKMSHELL,ZKENOMEM,"malloc");
      return 0;
    }
    memset(*shell,0,sizeof(struct T_ZkShell));
    zkShellSetPrompt(err,*shell,ZKSHELL_DEFAULT_PROMPT);
    zkShellSetStreams(err,*shell,ZKSHELL_DEFAULT_INPUT,ZKSHELL_DEFAULT_OUTPUT);
    (*shell)->flags = 0;
    (*shell)->command = NULL;
    (*shell)->commands = 0;
    (*shell)->context = NULL;
    
    /* 
     * Initialize the variable hash table
     */

#define ZKSHELL_TABLE_SZ 37
    if (!(zkHashNew(err,&((*shell)->variables)) &&
          zkHashSetHashFn(err,(*shell)->variables,zkShellVariableHash) &&
          zkHashSetUserData(err,(*shell)->variables,NULL) &&
          zkHashSetKeyFreeFn(err,(*shell)->variables,zkShellVariableFree) &&
          zkHashSetDataFreeFn(err,(*shell)->variables,zkShellVariableFree) &&
          zkHashSetTableSize(err,(*shell)->variables,ZKSHELL_TABLE_SZ))) {
      free(*shell);
      *shell = NULL;
      return 0;
    }
  
    /*
     * Load the commands 
     */

    for (i = 0; i < (int) STDCMDS; i++) {
      zkShellAddCommand(err,*shell,StandardCommand[i].callback,
                        StandardCommand[i].token,StandardCommand[i].usage,
                        StandardCommand[i].hlpst,StandardCommand[i].pmask);
    }
    return 1;
  }
  zkErrorSet(err,"zkShellNew",ZKERR,ZKMSHELL,ZKENULLPTR,NULL);
  return 0;
}
  

DLLExport int
zkShellFree (
  ZkError err ,
  ZkShell * shell ) 
{
  ZkShellCommand * command = NULL;
  ZkShellCommand * next = NULL;

  zkErrorClear(err);
  if (shell) {
    if (*shell) {
      command = (*shell)->command;
      while (command) {
        next = command->link;
        if (command->helpstr) {
          free(command->helpstr);
        }
        if (command->usage) {
          free(command->usage);
        }
        memset(command,0,sizeof(ZkShellCommand));
        free(command);
        command = next;
      }
      if ((*shell)->variables) {
        zkHashFree(err,&((*shell)->variables));
      }
      memset(*shell,0,sizeof(struct T_ZkShell));
      free(*shell);
      *shell = NULL;
      return 1;
    }
  }
  zkErrorSet(err,"zkShellFree",ZKERR,ZKMSHELL,ZKENULLPTR,NULL);
  return 0;
}


DLLExport int
zkShellWrite_va (
  ZkError err ,
  ZkShell shell ,
  const char * fmt ,
  ... )
{
  va_list ap;

  zkErrorClear(err);
  if (shell && fmt) {
    va_start(ap,fmt);
    vfprintf(shell->output,fmt,ap);
    va_end(ap);
    fflush(shell->output);
    return 1;
  }
  zkErrorSet(err,"zkShellWrite_va",ZKERR,ZKMSHELL,ZKENULLPTR,NULL);
  return 0;
} 


/*
 * This token parser is adapted from the zkipSessionTokenizeLine() function
 * it is a bit more reusable now.  At some point, this should be exported from
 * the libzk library and zkipSession should be made to use it.
 */
/**
 * You must pass the size of the the token array in tokens.  It will then 
 * return the number of tokens actually in the array. the line buffer is
 * also going to get mangled (the token array points inside the modified 
 * contents of line)
 */

#define CHAR_CR             0x0D        /* Carriage Return (ASCII 13) */
#define CHAR_LF             0x0A        /* Line Feed (ASCII 10) */
#define CHAR_NUL            0x00        /* NUL string terminator (ASCII 0) */
#define CHAR_SPACE          0x20        /* Blank space (ASCII 32) */
#define CHAR_HTAB           0x09        /* Horizontal Tab (ASCII 9) */
#define CHAR_DQUOTE         0x22        /* Double Quote (ASCII 34) */
#define CHAR_SOLIDUS        0x5C        /* Slash (ASCII 92) */
#define CHAR_ZERO           0x30        /* Decimal Zero (ASCII 48) */
#define CHAR_PERIOD         0x2E        /* Period (ASCII 46) */


static int
zkShellTokenizeLine (
  ZkError err ,
  char * line ,
  size_t linesz ,
  char ** token ,
  size_t * tokens )
{
  int i = 0;
  int j = 0;
  int state = 0;
  char * block = NULL;
  size_t maxtokens = 0;

  zkErrorClear(err);
  if (line && token && tokens) {
    if ((block = (char *) malloc(sizeof(char) * (linesz + 1))) == NULL) {
      zkErrorSet(err,"zkShellTokenizeLine",ZKERR,ZKMSHELL,ZKENOMEM,"malloc");
      return 0;
    }
    memcpy(block,line,linesz);
    block[linesz] = (char) 0; 
    maxtokens = *tokens;
    *tokens = 0;
    while ((block[i]) && (state < 6)) {
      switch (state) {
      case 0:
        switch (block[i]) {
        case CHAR_SPACE:
        case CHAR_HTAB:
        case CHAR_LF:
        case CHAR_CR:
          break;
        case CHAR_DQUOTE:
          state = 2;
          break;
        default:
          state = 1;
          line[j] = block[i]; 
          token[*tokens] = &(line[j]);
          *tokens += 1;
          j++;
        }
        break;
      case 1:
        switch (block[i]) {
        case CHAR_SPACE:
        case CHAR_HTAB:
        case CHAR_LF:
        case CHAR_CR:
          state = 0;
          line[j] = (char) 0;
          j++;
          break;
        default:
          line[j] = line[i];
          j++;
          break;
        }
        break;
      case 2:
        switch (block[i]) {
        case CHAR_DQUOTE:
          state = 3;
          line[j] = (char) 0;
          j++;
          break;
        default:
          state = 5;
          line[j] = block[i];
          token[*tokens] = &(line[j]);
          *tokens += 1;
          j++;
          break;
        }
        break;
      case 3:
        switch (block[i]) {
        case CHAR_HTAB:
        case CHAR_SPACE:
        case CHAR_LF:
        case CHAR_CR:
          state = 0;
          break;
        default:
          state = 6;
          break;
        }
        break;
      case 4:
        line[j] = line[i];
        j++;
        state = 5;
        break;
      case 5:
        switch (block[i]) {
        case CHAR_SOLIDUS:
          state = 4;
          break;
        case CHAR_DQUOTE:
          state = 3;
          line[j] = (char) 0;
          j++;
          break;
        default:
          line[j] = block[i];
          j++;
          break;
        }
        break;
      default:
      case 6:
        /* Ick! error state, how did we get here? */
        break;
      }
      i++;
    }
    line[j] = (char) 0;
    free(block);
    if (state > 5) {
      zkErrorSet(err,"zkShellTokenizeLine",ZKERR,ZKMSHELL,ZKEPARSE,NULL);
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkShellTokenizeLine",ZKERR,ZKMSHELL,ZKENULLPTR,NULL);
  return 0;
}

static int
zkShellProcessCommand (
  ZkError err ,
  ZkShell shell ,
  char * cmd )
{
  size_t tokens = 0;
  size_t n = 0;
  ZkShellCommand * command = NULL;
  ZkShellCommandCallback callback = NULL; /* Command callback pointer */
  char * token[MAX_TOKENS];
  char * value = NULL;                  /* To replace token w/ var value */

  /* 
   * Tokenize the command and resolve any variables specified
   */

  tokens = MAX_TOKENS;
  if (!zkShellTokenizeLine(err,cmd,strlen(cmd),token,&tokens)) {
    zkShellWrite_va(err,shell,"Error:  unable to parse command line\n");
    return ZKSHELL_SESSION_CONTINUE;
  }
  for (n = 0; n < tokens; n++) {
    if (token[n][0] == CHAR_DOLLAR)  {
      if (!zkShellVariableGet(err,shell,&(token[n][1]),&value)) {
        zkShellWrite_va(err,shell,"Error:  unknown variable \"%s\"\n",
                        &(token[n][1]));
        return ZKSHELL_SESSION_CONTINUE;
      }
      token[n] = value;
    }
  }

#if 0
  while ((token[tokens] = strtok(start," \t\r\n")) != NULL) {
    if (token[tokens][0] == CHAR_DOLLAR) {
      if (!zkShellVariableGet(err,shell,&(token[tokens][1]),&value)) {
        zkShellWrite_va(err,shell,"Error:  unknown variable \"%s\"\n",
                        &(token[tokens][1]));
        return ZKSHELL_SESSION_CONTINUE;
      }
      token[tokens] = value;
    }
    tokens++;
    start = NULL;
    if (tokens >= MAX_TOKENS) {
      break;
    }
  }
#endif

  if (tokens < 1) {
    return ZKSHELL_SESSION_CONTINUE;
  }

  /*
   * Find the command in the current shell
   */

  command = shell->command;
  while (command) {
    if (strcasecmp(token[0],command->cmdtoken) == 0) {
      callback = command->callback;
      break;
    }
    command = command->link;
  }
  if (command == NULL) {
    zkShellWrite_va(err,shell,"Error:  unrecognized command: %s\n",token[0]);
    return ZKSHELL_SESSION_CONTINUE;
  }
 
  /*
   * Check the parameter mask...
   */

  if (!((1 << (tokens - 1)) & command->pmask)) {
    zkShellWrite_va(err,shell,"Error:  incorrect number of parameters\n");
    return ZKSHELL_SESSION_CONTINUE;
  }

  /*
   * Call the command callback
   */

  return command->callback(err,shell,token,tokens,shell->context);
}

DLLExport int
zkShellVariableSet (
  ZkError err ,
  ZkShell shell ,
  char * name ,
  char * value )
{
  char * n = NULL;
  char * v = NULL;
  zkErrorClear(err);
  if (shell && name && value) {
    n = strdup(name);
    v = strdup(value);
    if (!(n && v)) {
      zkErrorSet(err,"zkShellVariableSet",ZKERR,ZKMSHELL,ZKENOMEM,"strdup");
      return 0; 
    }
    if (!zkHashInsert(err,shell->variables,n,v)) {
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkShellVariableSet",ZKERR,ZKMSHELL,ZKENULLPTR,NULL);
  return 0;
}
  
DLLExport int
zkShellVariableGet (
  ZkError err ,
  ZkShell shell ,
  char * name ,
  char ** value )
{
  zkErrorClear(err);
  if (shell && name && value) {
    if (!zkHashSearch(err,shell->variables,(void *)name,(void **)value)) {
      *value = 0;
      zkErrorSet_va(err,"zkShellVariableGet",ZKINFO,ZKMSHELL,ZKENOTFOUND,
                    "Could not find variable %s",name);
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkShellVariableGet",ZKERR,ZKMSHELL,ZKENULLPTR,NULL);
  return 0;
}
  
DLLExport int 
zkShellVariableUnset (
  ZkError err ,
  ZkShell shell ,
  char * name )
{
  zkErrorClear(err);
  if (shell && name) {
    if (!zkHashDelete(err,shell->variables,name)) {
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkShellVariableUnset",ZKERR,ZKMSHELL,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkShellRun (
  ZkError err ,
  ZkShell shell )
{
#define CMD_SZ 1024 /* Size of the command buffer */
  static const char * func = "zkShellRun";
  char cmd[CMD_SZ + 1];
  char *echocmd = NULL;
  char *echocmd_str = strdup (ZKSHELL_ECHO_CMD);
  char *yes_str = strdup (ZKSHELL_YES_STR);
  int return_code = 1; /* Function return code */

  zkErrorClear(err);
  if (shell) {
    for (;;) {
      memset(cmd,0,CMD_SZ);
#ifdef ZKSHELL_USE_GNU_READLINE      
      if (rl_fgets(cmd,CMD_SZ,shell->input,shell->prompt) == NULL) {
#else	
      zkShellWrite_va(err,shell,"%s",shell->prompt);
      if (fgets(cmd,CMD_SZ,shell->input) == NULL) {
#endif	
        break;
      }
      
      zkShellVariableGet (err,shell, echocmd_str, &echocmd);
      if (echocmd && strncmp (echocmd,yes_str,4) == 0) {
	zkShellWrite_va(err,shell,"%s",cmd);
      }
      
      switch (zkShellProcessCommand(err,shell,cmd)) {
      case ZKSHELL_SESSION_QUIT:
        goto done;
        break; 
      case ZKSHELL_SESSION_FATAL:
        return_code = 0;
        goto done;
        break;
      default:
      case ZKSHELL_SESSION_CONTINUE:
        break;
      }
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMSHELL,ZKENULLPTR,NULL);
    return_code = 0;
  }

done:
  /*
   * Cleanup
   */

  if (echocmd_str) {
    free(echocmd_str);
  }
  if (yes_str) {
    free(yes_str);
  }

  return return_code;
}

/*
 *
 */

DLLExport int
zkShellCommandSource (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * context )
{
  int return_code = ZKSHELL_SESSION_CONTINUE;
  static const char * func = "zkShellCommandSource";
  char cmd[CMD_SZ + 1];           /* Command buffer, read from file */
  char *echocmd = NULL;
  char * p = NULL;
  FILE * f = NULL;   /* File stream to read in the commands */


  (void) context;  /* Get rid of compile warning */

  zkErrorClear(err);

  /*
   * Check the function parameters
   */

  if (!shell) {
    zkErrorSet(err,func,ZKERR,ZKMSHELL,ZKENULLPTR,"shell is NULL");
  } else if (!token) {
    zkErrorSet(err,func,ZKERR,ZKMSHELL,ZKENULLPTR,"token array is NULL");

  /*
   * We only want one argument token, the filename to source
   */

  } else if (tokens != 2) {
    zkErrorSet_va(err,func,ZKERR,ZKMSHELL,ZKEPARMBAD,
                  "Got %d tokens, expected 2",tokens);
    zkShellWrite_va(err,shell,"Error:  You must specify a file to source.\n");
  } else if ((p = strdup(ZKSHELL_ECHO_CMD)) == NULL) {
    zkErrorSet_va(err,func,ZKERR,ZKMSHELL,ZKENOMEM,
                  "strdup():  %s",strerror(errno));

  /*
   * Try to open the source file for reading.
   */

  } else if ((f = fopen(token[1],"r")) == NULL) {
    zkErrorSet_va(err,func,ZKERR,ZKMSHELL,ZKEFILEOPEN,
                  "Failed to open \"%s\" for reading:  %s.\n",
                  token[1],strerror(errno));
    zkShellWrite_va(err,shell,"Failed to open \"%s\" for reading:  %s.\n",
                    token[1],strerror(errno));

  /*
   * File was succesfully opened, so while there are commands in it, read
   * them and execute them
   */

  } else while (!feof(f)) {
    memset(cmd,0,CMD_SZ);
    if (fgets(cmd,CMD_SZ,f) == NULL) {
      break;
    }

    /*
     * If echocmd is set, display the command as it is being read
     */

    zkShellVariableGet (err,shell, p, &echocmd);
    if (echocmd && strncmp(echocmd,ZKSHELL_YES_STR,4)) {
      zkShellWrite_va(err,shell,"%s",cmd);
    }
      
    return_code = zkShellProcessCommand(err,shell,cmd);
    if (return_code != ZKSHELL_SESSION_CONTINUE) {
      break;
    }
  }

  /*
   * Cleanup 
   */

  if (p) {
    free(p);
  }
  if (f) {
    fclose(f);
  }

  return return_code;
}

/*
 *
 */

DLLExport int 
zkShellCommandSet (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * context )
{
  (void) context; /* Get rid of compile warning */
  if (tokens == 1) {
    zkHashTraverse(err,shell->variables,zkShellVariableTraverse,shell);
  } else if (tokens == 3) { 
    if (!zkShellVariableSet(err,shell,token[1],token[2])) {
      zkShellWrite_va(err,shell,"Error:  could not set variable \"%s\"\n",
                      token[1]);
    }
  }
  return ZKSHELL_SESSION_CONTINUE;
}

DLLExport int
zkShellCommandUnset (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * context )
{
  (void) context; /* Get rid of compile warning */
  if (tokens == 2) {
    if (!zkShellVariableUnset(err,shell,token[1])) {
      zkShellWrite_va(err,shell,"Error:  unable to unset \"%s\"\n",token[1]);
    }
  }
  return ZKSHELL_SESSION_CONTINUE;
}

DLLExport int 
zkShellCommandEcho (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * context )
{
  int i = 1;
  (void) context; /* Get rid of compile warning */
  while (i < (int)tokens) {
    zkShellWrite_va(err,shell,"%s ",token[i++]);
  }
  zkShellWrite_va(err,shell,"\n");
  return ZKSHELL_SESSION_CONTINUE;
}

DLLExport int
zkShellCommandQuit (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * context )
{
  token = NULL;  /* These are reset to get rid of the compilation warnings */
  tokens = 0;
  context = NULL;
  zkShellWrite_va(err,shell,"Goodbye...\n");
  return ZKSHELL_SESSION_QUIT;
}

DLLExport int 
zkShellCommandPrompt (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * context )
{
  context = NULL; /* Gets rid of compile time warning */
  if (tokens == 2) {
    zkShellSetPrompt(err,shell,token[1]); 
  }
  return ZKSHELL_SESSION_CONTINUE;
}

DLLExport int
zkShellCommandHelp (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * context )
{
  ZkShellCommand * command = NULL;

  context = NULL;  /* These are reset to get rid of the compilation warnings */
  command = shell->command;
  if (tokens == 2) {
    while (command) {
      if (strcmp(token[1],command->cmdtoken) == 0) {
        zkShellWrite_va(err,shell,"Command: %s\nUsage: %s\nDescription: %s\n",
                        command->cmdtoken,command->usage,command->helpstr);
        break;
      }
      command = command->link;
    }
    if (command == NULL) {
      zkShellWrite_va(err,shell,"Error:  Command \"%s\" not found.\n",token[1]);
    }
  } else {
    command = shell->command;
    while (command) {
      zkShellWrite_va(err,shell,"%-15s  %s\n",command->cmdtoken,
                      command->usage);
      command = command->link;
    }
  }
  return ZKSHELL_SESSION_CONTINUE;
}
