/*
 *  $Id: zkSignal.c,v 1.6.32.1 2000/10/11 13:20:02 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#define _GNU_SOURCE 1 /* sigaction, strcasecmp */

#include "zks.h"

#include <stdarg.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include "zkSignal.h"


/*  
 *  A signal(3) replacement with reasonable semantics.
 *  
 *  Restart all interrupted system calls, except when the signal is SIGALRM.
 *  (Allowing us to use it as a timeout.)
 *  
 *  The code is taken from Stevens' "Advanced Programming in the UNIX
 *  Environment", page 298.
 */
DLLExport ZkSigFunc *zkSignal(int sig, ZkSigFunc *fn)
{
  struct sigaction act, oact;

  act.sa_handler = fn;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;

  if (sig == SIGALRM) {
#ifdef SA_INTERRUPT
    act.sa_flags |= SA_INTERRUPT; /* SunOS */
#endif
  } else {
#ifdef SA_RESTART
    act.sa_flags |= SA_RESTART; /* SVR4, 4.3+BSD */
#endif
  }

  if (sigaction(sig, &act, &oact) < 0) {
    return SIG_ERR;
  }

  return oact.sa_handler;
}


/*
 *  Like zkSignal(), but `slow' system calls are not restarted.
 */
DLLExport ZkSigFunc *zkSignalIntr(int sig, ZkSigFunc *fn)
{
  struct sigaction act, oact;

  act.sa_handler = fn;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;

  if (sig == SIGALRM) {
#ifdef SA_INTERRUPT
    act.sa_flags |= SA_INTERRUPT; /* SunOS */
#endif
  } else {
#ifdef SA_RESTART
    act.sa_flags &= ~SA_RESTART; /* SVR4, 4.3+BSD */
#endif
  }

  if (sigaction(sig, &act, &oact) < 0) {
    return SIG_ERR;
  }

  return oact.sa_handler;
}


/*
 *  Add `sig' to `*set'.
 */
DLLExport int zkSignalSetAdd(ZkSignalSet *set, int sig)
{
  return sigaddset(set, sig) == 0;
}


/*
 *  Initialize `*set' to the list of signals.  The list is terminated by 0,
 *  which is ignored.
 */
DLLExport int zkSignalSetInit(ZkSignalSet *set, int sig, ...)
{
  int ret = 0, s;
  va_list ap;
  
  if (sigemptyset(set) < 0) {
    return 0;
  }

  va_start(ap, sig);
  for (s = sig; s != 0; s = va_arg(ap, int)) {
    if (sigaddset(set, s) < 0) {
      goto done;
    }
  }

  ret = 1;
  
 done:

  va_end(ap);
  return ret;
}


/*
 *  Set `*old' to the current signal mask, then block the signals listed in
 *  `block'.
 */
DLLExport int zkSignalProtectBegin(ZkSignalSet block, ZkSignalSet *restore)
{
  return sigprocmask(SIG_BLOCK, &block, restore) == 0;
}


/*
 *  Set the signal mask to that in `restore'.
 */
DLLExport int zkSignalProtectEnd(ZkSignalSet restore)
{
  return sigprocmask(SIG_SETMASK, &restore, 0) == 0;
}


/*
 *  Map between signal names and numbers.
 */

#define S(n) { #n, n }

static struct signal_map {
  const char *name;
  const int num;
} sigMap[] = {
  S(SIGHUP),
  S(SIGINT),
  S(SIGQUIT),
  S(SIGILL),
  S(SIGTRAP),
  S(SIGIOT),
  S(SIGBUS),
  S(SIGFPE),
  S(SIGKILL),
  S(SIGUSR1),
  S(SIGSEGV),
  S(SIGUSR2),
  S(SIGPIPE),
  S(SIGALRM),
  S(SIGTERM),
  S(SIGCHLD),
  S(SIGCONT),
  S(SIGSTOP),
  S(SIGTSTP),
  S(SIGTTIN),
  S(SIGTTOU),
  S(SIGURG),
  S(SIGXCPU),
  S(SIGXFSZ),
  S(SIGVTALRM),
  S(SIGPROF),
  S(SIGWINCH),
  S(SIGIO),
  S(SIGPWR),

  { 0, 0 }
};

#undef S


static size_t sigIdxFromName(const char *name)
{
  size_t i;

  for (i = 0; sigMap[i].name; i++) {
    if (strcasecmp(name, sigMap[i].name) == 0 ||
        strcasecmp(name, sigMap[i].name + 3) == 0) {
      break;
    }
  }

  return i;
}


static size_t sigIdxFromNum(int n)
{
  size_t i;

  for (i = 0; sigMap[i].name; i++) {
    if (n == sigMap[i].num) {
      break;
    }
  }

  return i;
}


/*
 *  Return 0 if the signal having `name' can't be found, otherwise return the
 *  corresponding signal number.
 */
DLLExport int zkSignalNumFromName(const char *name)
{
  return sigMap[sigIdxFromName(name)].num;
}


/*
 *  Return a pointer to the name of signal number `signo'.  If this signal is
 *  unknown then return a null pointer.
 */
DLLExport const char *zkSignalNameFromNum(int signo)
{
  return sigMap[sigIdxFromNum(signo)].name;
}
