/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zkIntTypes.h>
#include <zkString.h>

/*
 * This is a rough initial pass at a string object. The API is fine, the
 * internal representation of the string needs work.
 *
 */

struct T_ZkString {
  unsigned char * data;
  uint32_t datalen;
};

/**
 *  This function is used to create a new string object.  The object will 
 *  contain absolutely nothing.
 * 
 *  @param err                Used to return the error
 *  @param s                  Returns a pointer to a new ZkString
 *  @version                  1.0
 *  @author                   Philippe Boucher
 *  @return                   a non-zero value on success, a 0 on failure
 */

DLLExport int
zkStringNew (
  ZkError err ,
  ZkString * s )
{
  zkErrorClear(err);
  if (s) {
    *s = (ZkString) malloc(sizeof(struct T_ZkString));
    if (*s == NULL) {
      zkErrorSet(err,"zkStringNew",ZKERR,ZKMSTRING,ZKENOMEM,NULL); 
      return 0;
    }
    memset(*s,0,sizeof(struct T_ZkString)); 
    (*s)->data = NULL;
    (*s)->datalen = 0;
    return 1;   
  }
  zkErrorSet(err,"zkStringNew",ZKERR,ZKMSTRING,ZKENULLPTR,NULL); 
  return 0;
}
  
/**
 *  This function is used to destroy a ZkString object.  The pointer to the
 *  string will be reset to NULL and the entire contents of the string will
 *  be destroyed as well.
 *
 *  @param err                used to return an error
 *  @param s                  points to the ZkString object to destroy
 *  @version                  1.0
 *  @author                   Philippe Boucher
 *  @return                   A non-zero value on success, a 0 on failure
 */
 
DLLExport int
zkStringFree (
  ZkError err ,
  ZkString * s )
{
  zkErrorClear(err);
  if (s) {
    if (*s) {
      if ((*s)->data) {
        free((*s)->data);
      }
      memset(*s,0,sizeof(struct T_ZkString));
      free(*s);
      *s = NULL;
      return 1;
    }
  }
  zkErrorSet(err,"zkStringFree",ZKERR,ZKMSTRING,ZKENULLPTR,NULL); 
  return 0;
}

/**
 *  Set a value for the string.  Any previous value which the string may have
 *  had before this call will be destroyed.  The data string must be NUL
 *  terminated (a la standard C string).
 * 
 *  @param err             Used to return the error
 *  @param s               The ZkString to manipulate
 *  @param data            NUL terminated text to which to set the string 
 *  @version               1.0
 *  @return                A non-zero value on success, a 0 on failure
 *  @author                Philippe Boucher
 */

int
zkStringSet (
  ZkError err ,
  ZkString s ,
  const unsigned char * data )
{
  zkErrorClear(err);
  if ((s) && (data)) {
    if (s->data != NULL) {
      free(s->data);
    }
    s->data = (unsigned char *) strdup((const char *)data);
    if (s->data == NULL) { 
      zkErrorSet(err,"zkStringSet",ZKERR,ZKMSTRING,ZKENOMEM,NULL); 
      return 0;
    }
    s->datalen = strlen((char *)s->data);
    return 1;
  }
  zkErrorSet(err,"zkStringSet",ZKERR,ZKMSTRING,ZKENULLPTR,NULL); 
  return 0;
}

/**
 *  Adds more text to an existing string.  This function will append the
 *  given NUL terminated text to the current contents of the string.  If the
 *  string does not contain any data, this function behaves exactly like
 *  zkStringSet().
 *
 *  @param err             Used to return the error
 *  @param s               The ZkString to manipulate
 *  @param data            NUL terminated text to append to the string 
 *  @version               1.0
 *  @return                A non-zero value on success, a 0 on failure
 *  @author                Philippe Boucher
 */

DLLExport int
zkStringAdd (
  ZkError err ,
  ZkString s ,
  const unsigned char * data )

{
  unsigned char * tmp = NULL;

  zkErrorClear(err);
  if ((s) && (data)) {
    if (s->data != NULL) {
      tmp = (unsigned char *) malloc(strlen((const char *)data) + s->datalen + 1);
      memcpy(tmp,s->data,s->datalen);
      memcpy(&(tmp[s->datalen]),data,strlen((const char *)data));
      tmp[strlen((const char *)data) + s->datalen] = (char) 0;
      free(s->data);
      s->data = tmp;
      s->datalen = strlen((const char *)s->data);
      return 1;
    } else {
      return zkStringSet(err,s,data);
    }
  }
  zkErrorSet(err,"zkStringAdd",ZKERR,ZKMSTRING,ZKENULLPTR,NULL); 
  return 0;
}

/**
 *  Return the length (in characters) of the string.
 *
 *  @param err           Used to return the error
 *  @param s             ZkString to analyze
 *  @param length        Returns the length of the string in characters
 *  @author              Philippe Boucher
 *  @version             1.0
 *  @return              a non-zero value on success, a 0 on failure
 */
DLLExport
int
zkStringLength (
  ZkError err,
  ZkString s ,
  unsigned int * length )
{
  zkErrorClear(err);
  if (s) {
    *length = s->datalen;
    return 1;
  }
  zkErrorSet(err,"zkStringLength",ZKERR,ZKMSTRING,ZKENULLPTR,NULL); 
  return 0;
}

/**
 *  Retreives the complete string data.  The returned string is a pointer to
 *  data within the string.  Do not try to modify this string itself, nor should
 *  you free it yourself (the call to zkStringFree() will take care of it)
 *  The return string is NUL terminated.
 *
 *  @param err              Used to return the error
 *  @param s                ZkString object to analyze
 *  @param data             Returns the string data (NUL terminated)
 *  @author                 Philippe Boucher
 *  @version                1.0
 *  @return                 a non-zero value on success, a 0 on failure
 */

DLLExport int
zkStringData (
  ZkError err ,
  ZkString s ,
  const unsigned char ** data )
{
  zkErrorClear(err);
  if ((s) && (data)) {
    *data = s->data;
    return 1; 
  }
  zkErrorSet(err,"zkStringData",ZKERR,ZKMSTRING,ZKENULLPTR,NULL); 
  return 0;
}

/**
 *  Retreives the complete string data and returns a pointer to the data.
 *  Do not try to modify this string itself, nor should
 *  you free it yourself (the call to zkStringFree() will take care of it)
 *  The return string is NUL terminated.
 *
 *  @param err              Used to return the error
 *  @param s                ZkString object to analyze
 *  @param data             Returns the string data (NUL terminated)
 *  @author                 Luc Boulianne
 *  @version                1.0
 *  @return                 NULL on err, or a pointer to a string             
 */

DLLExport const unsigned char *zkStringGetData (ZkError err, ZkString s)
{
  zkErrorClear(err);
  if ((s)) {
    return s->data;
  }
  zkErrorSet(err,"zkStringData",ZKERR,ZKMSTRING,ZKENULLPTR,NULL); 
  return NULL;
}

