/* 
 * $Id: zkserr.c,v 1.1.4.1 2000/12/14 17:54:29 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>          
#include <stdlib.h>

#ifdef WIN32
# include <time.h>
#else
# include <sys/time.h>
# include <unistd.h>
#endif

#include <zkError.h>

#include "zkserr_p.h"
#include "zkserr.h"
#include "zksmodule.h"




/*
**
*/
ZkBool zksErrGetLastSysError(uint32_t * syserrP)
{
  if( !syserrP ) { return ZKS_FAILURE; }

#if defined( WIN32 )
  *syserrP = (uint32_t)GetLastError();

#elif defined( ZK_LINUX )
  *syserrP = (uint32_t)errno;
#endif
  return ZKS_SUCCESS;
}



/*
**
*/
ZkBool zksErrGetSysErrorMessage(uint32_t syserrcode, char * mesg, uint32_t  mesgsz)
{
  ZkBool ret = ZKS_SUCCESS;

  if( !mesg || !mesgsz ) { return ZKS_FAILURE; }

  mesg[0] = '\0';

#if defined( WIN32 )
  if( !FormatMessage(FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM,
                    NULL,
                    syserrcode,
                    MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL),
                    mesg,
                    (DWORD)mesgsz-1,
                    NULL) ) {
    ret = ZKS_FAILURE;
  }


#elif defined( ZK_LINUX )
# if defined( ZK_POSIX ) 
  if( !strerror_r((int)syserrcode, mesg, mesgsz) ) {
    ret = ZKS_FAILURE;
  }
# else  
  snprintf(mesg, mesgsz, "%s", strerror((int)syserrcode));
# endif
#endif  
  /*
   * Remove any traling end of line (usefull on windows)
   */
  if( ret ) {    
    uint32_t pos = strlen(mesg);
    while( pos >= 1 && ((mesg[pos-1]=='\n') || (mesg[pos-1] == '\r')) ) {
      mesg[--pos] = '\\';
    }
  }
  return ret;
}






/* ============================================================================
 *
 *   Local Calls
 *
 */

static void   errMsgDestroy                  (ZksErr         err);

static void   errMsgCtxFill_valist           (ZksErrMsgCtx   ctx,
                                              struct timeval errTime,
                                              const char   * func,
                                              const char   * mesg,
                                              va_list        args);

static ZkBool errSetEx_valist                (ZksErr         err, 
                                              uint16_t       category,
                                              ZksErrCode     errcode,
                                              const char   * func, 
                                              const char   * mesg, 
                                              va_list        vargs);

static ZkBool errMsgPushTail_valist          (ZksErr         err,
                                              struct timeval errTime,
                                              const char   * func,
                                              const char   * mesg, 
                                              va_list        args);


static ZkBool errStatus                      (ZksErr err);



static void errMsgDestroy(ZksErr err)
{
  uint32_t idx;

  /* Cleanup the Message stack
   */
  for( idx=0; idx<err->stacksz; idx++ ) {
    if( err->stack[idx] ) {
      free(err->stack[idx]);
      err->stack[idx] = NULL;
    }
  }
  err->stacksz = 0;
}



/*
 *
 */ 
static void errMsgCtxFill_valist(ZksErrMsgCtx   ctx,
                                 struct timeval errTime,
                                 const char   * func,
                                 const char   * mesg,
                                 va_list        args)
{
  ctx->errTime = errTime;
  snprintf(ctx->func, sizeof(ctx->func), "%s", (func) ? func : "(nul)");
  vsnprintf(ctx->mesg, sizeof(ctx->mesg), (mesg) ? mesg : "(nul)", args);
}




/*
 *
 */ 
static ZkBool errSetEx_valist(ZksErr       err, 
                              uint16_t     category,
                              ZksErrCode   errcode,
                              const char * func, 
                              const char * mesg, 
                              va_list      vargs)
{
  struct timeval curtime = { 0, 0 };
  uint32_t  syserr;
  T_ZksPid  pid;
  uint16_t  xflags = ZKS_ERR_NONE_FLAG;

  if( err ) {

    /* Get Current system error
     */
    zksErrGetLastSysError(&syserr);

    /* Get Current system UTC time
     */
    gettimeofday(&curtime, NULL);

 
    /* Get Current Process Id
     */
    zksPidInit(&pid);


    /* Check for resetting that would append 
       before the error has beeing logged. It
       would result the lost of an error.
    */
    if( (err->flags & ZKS_ERR_SETTED_FLAG) ||
        (err->flags & ZKS_ERR_RESETTED_FLAG) ) {    
      xflags |= ZKS_ERR_RESETTED_FLAG;
    }

    /* Cleanup current context
     */
    zksErrClear(err);     

    /* Set the RESETTED flags to advised that
       errors has been lost 
    */
    err->flags |= xflags;

    /* Fill the error code struncture 
     */
    err->category   = category;
    err->code       = errcode;
    err->sysPid     = pid;
    err->sysErrCode = syserr;

    errMsgCtxFill_valist(&err->detail, curtime, func, mesg, vargs);

    /* Enable the Setted flag 
     */
    err->flags |= ZKS_ERR_SETTED_FLAG;
  }
  return errStatus(err);
}



/*
 *
 */
static ZkBool errMsgPushTail_valist(ZksErr         err,
                                    struct timeval errTime,
                                    const char   * func,
                                    const char   * mesg, 
                                    va_list        args)
{
  if( err ) {
    /* Check for abnormal situation */
    if( err->flags & ZKS_ERR_LOWMEMORY_FLAG )      { return ZKS_FAILURE; }
    if( err->flags & ZKS_ERR_STACK_OVERFLOW_FLAG ) { return ZKS_FAILURE; }


    /* Allocate the new message context */
    if( (err->stack[err->stacksz] = (ZksErrMsgCtx) malloc(sizeof(struct S_ZksErrMsgCtx))) == NULL ) {
      err->flags |= ZKS_ERR_LOWMEMORY_FLAG;
      return ZKS_FAILURE;
    } 

    errMsgCtxFill_valist(err->stack[err->stacksz], errTime, func, mesg, args);
    err->stacksz++;

    if( err->stacksz == ZKS_ERR_STACK_MAX ) { err->flags |= ZKS_ERR_STACK_OVERFLOW_FLAG; }
  }

  return ZKS_SUCCESS;
}


static ZkBool errStatus(ZksErr err)
{
  return ( err && !err->code ) ? ZKS_SUCCESS : ZKS_FAILURE;
}



/* ============================================================================
 *
 * Exported Calls
 *
 * - ZksErrCode API
 *
 */

ZksErrCode zksErrCodeBuild(const char *  modName, 
                           uint16_t      modErr)
{
  uint16_t modId = zksModuleId(modName);

  return ZKS_ERRCODE_SET(modId, modErr);
}






/* ============================================================================
 *
 * Exported Calls
 *
 * - ZksErr API
 *
 */


/*
 *
 */ 
ZkBool zksErrNew(ZksErr * errP)
{
  ZkBool  ret  = ZKS_SUCCESS;
  ZkError zerr = NULL;
  ZksErr  err  = NULL;

  if( !errP ) { return ZKS_FAILURE; }

  /* Allocate an Old ZkError Handler use 
   * contain the ZksErr context.
   */
  if( !zkErrorNew(&zerr) ) {
    ret = ZKS_FAILURE;

  } else {
    /*
     * Retreive the handler from its content
     */
    err = zksErr(zerr);
    
    /* Allocate the inbeded temporary error handler
     */
    if( (err->tmpErr = (ZksErr) calloc(1, sizeof(struct S_ZksErr))) == NULL ) {
      ret = ZKS_FAILURE;
    }
  }
  if( !ret ) {
    zkErrorFree(&zerr);

  } else {
    *errP = err;
  }

  return ret;
}





/*
 *
 */ 
void zksErrFree(ZksErr * errP)
{
  if(  !errP ) { return; }
  
  if( *errP ) {
    /*
     * Liberate imbeded content
     */
    if( (*errP)->tmpErr ) {
      zksErrFree(&(*errP)->tmpErr);
    }

    /*
     * Do not Free imbeded component
     */
    if( !((*errP)->hook) ) {
      free(*errP);
      *errP = NULL;
    }
  }
}




/*
 *
 */ 
ZkBool zksErrSet(ZksErr       err, 
                 const char * modName, 
                 uint16_t     modErr, 
                 const char * func, 
                 const char * mesg,
                 ...)
{
  ZkBool  ret = ZKS_SUCCESS;
  va_list args;

  va_start(args, mesg);
  ret = errSetEx_valist(err, 
                        ZKS_ERR_CATEGORY_NONE, 
                        zksErrCodeBuild(modName, modErr), 
                        func, 
                        mesg,
                        args);
  va_end(args);
  return ret;
}




/*
 *
 */ 
ZkBool zksErrSetEx(ZksErr       err, 
                   uint16_t     category,
                   const char * modName, 
                   uint16_t     modErr, 
                   const char * func, 
                   const char * mesg, 
                   ...)
{
  ZkBool  ret = ZKS_SUCCESS;
  va_list args;

  va_start(args, mesg);
  ret = errSetEx_valist(err, 
                        category, 
                        zksErrCodeBuild(modName, modErr),
                        func,
                        mesg,
                        args);
  va_end(args);
  return ret;
}




/*
 *
 */ 
ZkBool zksErrPush(ZksErr       err,
                  const char * func,
                  const char * mesg,
                  ...)
{
  struct timeval        curtime   = { 0, 0 };
  va_list               args;
  ZkBool                overwrite = ZK_FALSE;
  struct S_ZksErrMsgCtx msgctx;

  if( err ) { 
    /* Get Current system GMT time
     */
    gettimeofday(&curtime, NULL);

    /* 
     * Merge the temporary error context to the current one
     */
    if( err->tmpErr && err->tmpErr->code ) {
      /* If the error is not set than look at the temporary error 
       * context.  If this one is set that use it to fill the current
       * error context.
       */
      if( !err->code ) {
        overwrite        = ZK_TRUE;
        err->category    = err->tmpErr->category;
        err->code        = err->tmpErr->code;
        err->sysPid      = err->tmpErr->sysPid;
        err->sysErrCode  = err->tmpErr->sysErrCode;
        err->flags       = err->tmpErr->flags;
        err->detail      = err->tmpErr->detail;

      } else {
        /*
         * Push the temporary error detail into the message queue 
         */
        zksErrMsgPushTail(err, 
                          err->tmpErr->detail.errTime,
                          func,
                          "(ERR=%08X %s: %s)",
                          err->tmpErr->detail.func,
                          err->tmpErr->code,
                          err->tmpErr->detail.mesg);
      }

      /* 
       * Join the temporary error message queue to the current context
       */
      while( zksErrMsgPopHead(err->tmpErr, 
                              &msgctx.errTime,
                              msgctx.func,
                              sizeof(msgctx.func),
                              msgctx.mesg,
                              sizeof(msgctx.mesg)) ) {
        if( overwrite ) {
          zksErrMsgPushTail(err, 
                            msgctx.errTime,
                            msgctx.func,
                            "%s",
                            msgctx.mesg);

        } else {
          zksErrMsgPushTail(err, 
                            msgctx.errTime,
                            func,
                            "(ERR=%08X: %s: %s)",
                            msgctx.func,
                            err->tmpErr->code,
                            msgctx.mesg);
        }
      }
    }

    /* If the error is now set than push the message
     */
    if( err->code ) {
      va_start(args, mesg);
      errMsgPushTail_valist(err, curtime, func, mesg, args);
      va_end(args);
    }
    /* Wipe the temporary error content
     */
    zksErrClear(err->tmpErr);
  }
  return errStatus(err);
}


/*
 *
 */
ZkBool zksErrInfo(const ZksErr     err,
                  struct timeval * errtimeP,
                  T_ZksPid       * syspidP,
                  ZksErrCode     * errcodeP,
                  uint32_t       * syserrcodeP,
                  uint16_t       * flagsP,
                  char           * func,
                  uint32_t         funcsz,
                  char           * mesg,
                  uint32_t         mesgsz)
{
  return zksErrInfoEx(err, 
                      NULL, 
                      errtimeP, 
                      syspidP, 
                      errcodeP, 
                      syserrcodeP, 
                      flagsP, 
                      func, 
                      funcsz, 
                      mesg, 
                      mesgsz);
}



/*
 *
 */
ZkBool zksErrInfoEx(const ZksErr    err,
                    uint16_t      * categoryP,
                    struct timeval     * errtimeP,
                    T_ZksPid      * syspidP,
                    ZksErrCode    * errcodeP,
                    uint32_t      * syserrcodeP,
                    uint16_t      * flagsP,
                    char          * func,
                    uint32_t        funcsz,
                    char          * mesg,
                    uint32_t        mesgsz)
{
  if( !err ) { return ZKS_FAILURE; }
  
  /* Filled output buffer
   */
  if( categoryP )   { *categoryP   = err->category; }
  if( errtimeP )    { *errtimeP    = err->detail.errTime; }
  if( syspidP )     { *syspidP     = err->sysPid; }
  if( syserrcodeP ) { *syserrcodeP = err->sysErrCode; }
  if( errcodeP )    { *errcodeP    = err->code; }
  if( flagsP )      { *flagsP      = err->flags; }
  if( func )        {  snprintf(func, funcsz, "%s", err->detail.func); }
  if( mesg )        {  snprintf(mesg, mesgsz, "%s", err->detail.mesg); }

  return ZKS_SUCCESS;
}
  


   
/*
 *
 */ 
void zksErrCategorySet(const ZksErr err, uint16_t category)
{
  if( !err ) { return; }
  err->category = category;
}





/*
 *
 */
ZksErrCode zksErrCode(const ZksErr err)
{
  if( !err ) { return ZKS_ERR_CODE_NONE; }
  return err->code;
}



/*
 *
 */
const char * zksErrMessage(const ZksErr err)
{
  if( !err ) { return NULL; }
  return err->detail.mesg;
}



/*
 *
 */
uint16_t zksErrFlags(const ZksErr err)
{
  if( !err ) { return ZKS_ERR_NONE_FLAG; }
  return err->flags;
}




/*
 *
 */
void zksErrClear(ZksErr err)
{
  if( err ) {
    /* Cleanup the Error Message List
     */
    errMsgDestroy(err);

    /* Cleanup the Error Context
     */
    err->category   = ZKS_ERR_CATEGORY_NONE;
    err->code       = ZKS_ERR_CODE_NONE;
    err->sysErrCode = ZKS_ERR_SYSERRCODE_NONE;
    err->flags      = ZKS_ERR_NONE_FLAG;

    /* Cleanup the Error Detail
     */  
    memset(&err->detail, 0, sizeof(struct S_ZksErrMsgCtx));

    /* Cleanup the temporary error context
     */
    zksErrClear(err->tmpErr);
  }
}



/*
 *
 */
ZksErr zksErrTmp(const ZksErr err)
{
  if( !err ) { return NULL; }

  return (err->tmpErr) ? err->tmpErr : err;
}



/*
 *
 */
ZkBool zksErrTranslate(ZksErr           err,
                       const char     * CriteriaModule,
                       uint16_t         CriteriaCode,
                       const char     * toModule,
                       uint16_t         toErrcode,
                       const char     * toFunc, 
                       const char     * toMesg,
                       ...)
{
  va_list args;
  ZkBool  match = ZK_FALSE;

  /* Check for matching criteria 
   */
  if( err && (zksErrCode(err) == zksErrCodeBuild(CriteriaModule, CriteriaCode))  ) {
    /* Clear cuirrent context 
     */
    zksErrClear(err);

    /* Reassign the Error content
     */
    va_start(args, toMesg);
    errSetEx_valist(err, 
                    err->category, 
                    zksErrCodeBuild(toModule, toErrcode), 
                    toFunc, 
                    toMesg, 
                    args);
    va_end(args);

    /* Enable Translated flag 
     */
    err->flags |=  ZKS_ERR_TRANSLATED_FLAG;

    match = ZK_TRUE;
  }
  return match;
}



/*
 *
 */
ZkBool zksErrTranslateEx(ZksErr           err, 
                         const char     * CriteriaModule,
                         uint16_t         CriteriaCode,
                         uint16_t         toCategory,
                         const char     * toModule,
                         uint16_t         toErrcode,
                         const char     * toFunc, 
                         const char     * toMesg,
                         ...)
{
  va_list args;
  ZkBool  match = ZK_FALSE;
  

  /* Check for matching criteria 
   */
  if( zksErrCode(err) == zksErrCodeBuild(CriteriaModule, CriteriaCode) ) {
    /* Clear cuirrent context 
     */
    zksErrClear(err);

    /* Reassign the Error content
     */
    va_start(args, toMesg);
    errSetEx_valist(err, 
                    toCategory,
                    zksErrCodeBuild(toModule, toErrcode), 
                    toFunc, 
                    toMesg, 
                    args);
    va_end(args);

    /* Enable Translated flag 
     */
    err->flags |=  ZKS_ERR_TRANSLATED_FLAG;
    match = ZK_TRUE;
  }
  return match;
}





/* ============================================================================
 *
 * Exported Calls
 *
 * - ZksErr Message Stack API
 *
 */
 

/*
 *
 */
ZkBool zksErrMsgEmpty(ZksErr err)
{
  return ( !err ) ? ZK_TRUE : ( (err->stacksz) ? ZK_FALSE : ZK_TRUE);
}



/*
 *
 */
ZkBool zksErrMsgPopHead(ZksErr           err,
                        struct timeval * errTimeP,
                        char           * func,
                        uint32_t         funcsz,
                        char           * mesg,
                        uint32_t         mesgsz)
{
  if( !err || !err->stacksz ) { return ZKS_FAILURE; }

  /* Unconditionaly remove the overflow flag */
  err->flags &= ~ZKS_ERR_STACK_OVERFLOW_FLAG;

  if( errTimeP ) { *errTimeP = err->stack[0]->errTime; }
  if( func )     {  snprintf(func, funcsz, "%s", err->stack[0]->func); }
  if( mesg )     {  snprintf(mesg, mesgsz, "%s", err->stack[0]->mesg); }

  /* Realign the queue
   */
  free(err->stack[0]);
  memmove(&err->stack[0], &err->stack[1], (err->stacksz-1) * sizeof(err->stack[0]));
  err->stacksz--;
  
  return ZKS_SUCCESS;
}




/*
 *
 */
ZkBool zksErrMsgPushTail(ZksErr         err,
                         struct timeval errTime,
                         const char   * func,
                         const char   * mesg, 
                         ...)
{
  ZkBool  ret = ZKS_SUCCESS;
  va_list args;

  if( !err ) { return ZKS_SUCCESS; } 

  va_start(args, mesg);
  ret = errMsgPushTail_valist(err, errTime, func, mesg, args);
  va_end(args);

  return ret;
}




/* ============================================================================
 *
 * Exported Calls
 *
 * - ZksErr Message Stack API
 *
 */



/*
 *
 */
ZkBool zksErrCtxClone(ZksErr       err,
                      const ZksErr from)
{
  uint32_t idx;
  ZkBool   ret = ZK_TRUE;

  if( err && from ) {
    /* Clone the error context
     */
    (*err) = (*from);

    /* Clone it's error message stack
     */
    err->stacksz = 0;

    for( idx=0; ret && idx<from->stacksz; idx++ ) {    
      ret = zksErrMsgPushTail(err, 
                              from->stack[idx]->errTime,
                              from->stack[idx]->func,
                              "%s",
                              from->stack[idx]->mesg);
    }

    /* Clone the temporary error context as well
     */

    if( ret && from->tmpErr ) {
      ret = zksErrCtxClone(err->tmpErr, from->tmpErr);
    }
  }
  return ret;
}




/*
 *
 */ 
ZkBool zksErrCtxSet_valist(ZksErr            err,
                           ZksErrCode        errcode,
                           const char      * func, 
                           const char      * mesg, 
                           va_list           vargs)
{
  ZkBool   ret = ZK_TRUE;

  if( err ) {
    ret = errSetEx_valist(err, 
                          ZKS_ERR_CATEGORY_NONE, 
                          errcode, 
                          func, 
                          mesg,
                          vargs);
  }
  return ret;
}



/* ============================================================================
 *
 * Exported Calls
 *
 * - Error handling coexistance calls
 *
 */

/*
 *  This function return a reference to a ZksErr context
 */
ZksErr zksErr (const ZkError err)
{  
  return err ? &err->context : NULL;
}



/*
 *  This function return a reference to a ZkError context
 */
ZkError zkError(const ZksErr  err)
{
  if( err ) {
    return (ZkError) err->hook;
  }
  return NULL;
}

