/* 
 * $Id: zksmodule.c,v 1.1.4.2 2001/01/04 20:55:42 nicolas Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>          
#include <stdlib.h>

#if defined( WIN32 )
# include <time.h>

#elif defined( ZK_LINUX )
# include <sys/time.h>
# include <unistd.h>
#endif

#include "zksmodule.h"
#include "zkserr.h"


/*
** Definition
*/
typedef struct S_Module {
  const char * name;
  uint16_t     id;
} ModuleList;


/* ============================================================================
 *
 * Global MODULE list.
 *
 * - Increase this list by adding your module name and it's associate Id
 *
 *
 * ie: { MY_FIRST_MOD,  ZKS_MODULE_ID_BASE + 0x0000}
 *     { MY_SECOND_MOD, ZKS_MODULE_ID_BASE + 0x0001}
 *
 *
 *     ...
 *
 * WARNING: *NEVER REMOVE OR SWAP* any entry from this list. It would corrupt
 *          our Error Number list an prevent us to discover errors based on
 *          their Error Code.
 *          
 */


/*
** Declaration
*/
static ModuleList s_moduleList[]
= { { ZKS_GENERIC_UNKNOWN_MOD,            ZKS_MODULE_ID_BASE + 0x0000 },
    { ZKS_GENERIC_TEST_MOD,               ZKS_MODULE_ID_BASE + 0x0001 },
    { "ZksModule",                        ZKS_MODULE_ID_BASE + 0x0002 },
    { "ZksErr",                           ZKS_MODULE_ID_BASE + 0x0003 },
    { "ZksLogMng",                        ZKS_MODULE_ID_BASE + 0x0004 },
    { "ZksLog",                           ZKS_MODULE_ID_BASE + 0x0005 },
    { "ZksLogAgent",                      ZKS_MODULE_ID_BASE + 0x0006 },
    { "ZksPid",                           ZKS_MODULE_ID_BASE + 0x0007 },
    { "CZKS::Exception",                  ZKS_MODULE_ID_BASE + 0x0008 },
    { "CZKS::CmdLine",                    ZKS_MODULE_ID_BASE + 0x0009 },
    { "HtmlGet",                          ZKS_MODULE_ID_BASE + 0x000A },
    { "AIP",                              ZKS_MODULE_ID_BASE + 0x000B },  
    { "AipKernel",                        ZKS_MODULE_ID_BASE + 0x000C },
    { "ConfigProvider",                   ZKS_MODULE_ID_BASE + 0x000D },  
    { "PrivateKeys",                      ZKS_MODULE_ID_BASE + 0x000E },
    { "NetInfoProvider",                  ZKS_MODULE_ID_BASE + 0x000F },
    { "KeyProvider",                      ZKS_MODULE_ID_BASE + 0x0010 },
    { "StatsHarvestor",                   ZKS_MODULE_ID_BASE + 0x0011 },
    { "StatsPackager",                    ZKS_MODULE_ID_BASE + 0x0012 },
    { "LinkManager",                      ZKS_MODULE_ID_BASE + 0x0013 },
    { "LinkMaintainer",                   ZKS_MODULE_ID_BASE + 0x0014 },
    { "DHNegotiator",                     ZKS_MODULE_ID_BASE + 0x0015 },
    { "RouteCreator",                     ZKS_MODULE_ID_BASE + 0x0016 },
    { "PacketDispatcher",                 ZKS_MODULE_ID_BASE + 0x0017 },
    { "ComponentManager",                 ZKS_MODULE_ID_BASE + 0x0018 },
    { "CZKS::LogWrapper",                 ZKS_MODULE_ID_BASE + 0x0019 }
};





/* ============================================================================
 *
 *   MODULE handler
 *
 */

#define MODULE_LIST_SIZE           (sizeof(s_moduleList)/(sizeof(s_moduleList[0])))

#if defined( WIN32 )
# define cmpModName(a, b)          ((int)strnicmp(a ? a : "", b ? b : "", ZKS_MODULE_NAME_MAX))
#else
# define cmpModName(a, b)          ((int)strncasecmp(a ? a : "", b ? b : "", ZKS_MODULE_NAME_MAX))
#endif 

#define cmpModId(a, b)             ((int)((a < b) ? -1 : ((a == b) ? 0 : 1)))


typedef struct S_ZksModuleHandler {
  ZkBool            isInit;
  struct S_Module   namelist[MODULE_LIST_SIZE];
  struct S_Module   idlist[MODULE_LIST_SIZE];
  uint32_t          listsz;
} * ZksModuleHandler;


static struct S_ZksModuleHandler s_modulehdl;


/*
* Call use to compare two entry from the namelist
*/
static int comp_ModuleName(const void * parmA, const void * parmB)
{
  const struct S_Module * modA = (const struct S_Module *) parmA;
  const struct S_Module * modB = (const struct S_Module *) parmB;

  if( modA && modB ) {
    return cmpModName(modA->name, modB->name);
  }
  return 0;
}



/*
 * Call use to compare two entry from the idlist
 */
static int comp_ModuleId(const void * parmA, const void * parmB)
{
  const struct S_Module * modA = (const struct S_Module *) parmA;
  const struct S_Module * modB = (const struct S_Module *) parmB;

  if( modA && modB ) {
    return cmpModId(modA->id, modB->id);
  }
  return 0;
}



/*
 * Initialysation call
 */
static void ModuleHandlerInit(ZksModuleHandler hdl) 
{
  if( !hdl->isInit ) {
    /* 
    ** Per thread automatic buffer use to avoid thread sorting collision
    */
    struct S_Module   tmpNamelist[MODULE_LIST_SIZE];
    struct S_Module   tmpIdlist[MODULE_LIST_SIZE];
  
  /*
  ** Set up the list size
  */
    hdl->listsz = MODULE_LIST_SIZE;

    /*
  ** Take a local copy of the module list  
  */
    memcpy(tmpNamelist, s_moduleList, sizeof(s_moduleList));
    memcpy(tmpIdlist,   s_moduleList, sizeof(s_moduleList));

  /*
  ** Sort the entirer name and id list 
  */
    qsort(tmpNamelist, hdl->listsz, sizeof(struct S_Module), comp_ModuleName);
    qsort(tmpIdlist,   hdl->listsz, sizeof(struct S_Module), comp_ModuleId);

  /* 
  ** Fill back the handler structure with the sorted result
  */
    memcpy(hdl->namelist, tmpNamelist,  sizeof(s_moduleList));
    memcpy(hdl->idlist,   tmpIdlist,    sizeof(s_moduleList));

  /* 
  ** Initialysation done
  */
    hdl->isInit = ZK_TRUE;
  }
}





/*
 * Return the ModuleId associate to a given ModuleName
 */
static uint16_t ModuleId(ZksModuleHandler hdl, const char * modName)
{
  uint32_t min;
  uint32_t max;
  uint32_t mid;
  ZkBool   done;
  int      cmp;

  ModuleHandlerInit(hdl);

  min  = 0;
  max  = hdl->listsz - 1;
  mid  = (max + min) >> 1;
  done = ZK_FALSE;
  cmp  = -1;

  /* 
   * Dycotomic scan of the list
   */
  while( !done ) {
    mid = (max + min) >> 1;
    cmp = cmpModName(modName, hdl->namelist[mid].name);

    if( cmp < 0 ) {
      max = mid;

    } else if( cmp > 0 ) {
      min = mid;

    } else if( cmp == 0 ) {
      done = ZK_TRUE;
    }
    if( !done && ((max - min) < 2) ) {
      if( (cmp = cmpModName(modName, hdl->namelist[mid = min].name)) != 0 ) {
        cmp = cmpModName(modName, hdl->namelist[mid = max].name);
      }
      done = ZK_TRUE;
    }
  }

  /* 
   * If the module has been found
   */
  if( !cmp ) { return hdl->namelist[mid].id; }

  return ZKS_MODULE_ID_UNKNOWN;
}



/*
 * Return the ModuleName associate to a given ModuleId
 */
static const char * ModuleName(ZksModuleHandler hdl, uint16_t modId)
{
  uint32_t min;
  uint32_t max;
  uint32_t mid;
  ZkBool   done;
  int      cmp;

  ModuleHandlerInit(hdl);

  min   = 0;
  max   = hdl->listsz - 1;
  mid   = (max + min) >> 1;
  done  = ZK_FALSE;
  cmp   = -1;

  /* 
   * Dycotomic scan of the list
   */
  while( !done ) {
    mid = (max + min) >> 1;
    cmp = cmpModId(modId, hdl->idlist[mid].id);

    if( cmp < 0 ) {
      max = mid;

    } else if( cmp > 0 ) {
      min = mid;        

    } else if( cmp == 0 ) {
      done = ZK_TRUE;
    }
    if( !done && ((max - min) < 2) ) {
      if( (cmp = cmpModId(modId, hdl->idlist[mid = min].id)) != 0 ) {
        cmp = cmpModId(modId, hdl->idlist[mid = max].id);
      }
      done = ZK_TRUE;
    }
  }

  /* 
   * If the module has been found
   */
  if( !cmp ) { 
    return (hdl->idlist[mid].name) ? (hdl->idlist[mid].name) : ZKS_MODULE_NAME_UNKNOWN;
  } 

  return ZKS_MODULE_NAME_UNKNOWN;  
}



/* ============================================================================
 *
 *   MODULE API
 *
 */

void zksModuleHandlerInit(void) 
{
  /* Invoke the local call to initialyse the global module handling context. */
  ModuleHandlerInit(&s_modulehdl);
}





/*
 * Return the ModuleId associate to a given Module Name
 */
uint16_t zksModuleId(const char * modName)
{
  return ModuleId(&s_modulehdl, modName);
}



/*
 * Return the ModuleName associate to a given Module Id
 */
const char * zksModuleName(uint16_t modId)
{
  return ModuleName(&s_modulehdl, modId);
}
