/* **************************************************************************
** $Id: zksocket.c,v 1.28.8.9 2001/01/09 20:33:30 oliver Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

** Copyright (C) 2000 Zero-Knowledge Systems Inc.  All rights reserved.
** --------------------------------------------------------------------------
**
** This file contains definitions and declarations for a socket wrapper
** object and its associated API.
**
** **************************************************************************
*/


# include "zks.h"
# include "zksocket.h"
# include "zkAlloc.h"
# include "zkNet.h"
# include "zknbo.h"
# include "zkDNS.h"

# if defined (WIN32)
#   include <time.h>
# else  /* !defined (WIN32) */
#   include <sys/time.h>
#   include <unistd.h>
#   include <netdb.h>
#   include <arpa/inet.h>
#   include <sys/types.h>
#   include <sys/socket.h>
#   include <sys/un.h>
#   include <fcntl.h>
#endif

# include <ctype.h>
# include <limits.h>
# include <errno.h>
# include <stdio.h>


# if defined (WIN32)
#   define ZKSOCKET_INVALID_DESCRIPTOR   (INVALID_SOCKET)
# else
#   define ZKSOCKET_INVALID_DESCRIPTOR   (-1)
# endif

# ifndef AF_LOCAL
# define AF_LOCAL AF_UNIX
# endif

#include "zksockettypes.h"


/* ------------------------------------------------------------------------
 *
 * Close the given file descriptor in a somewhat portable manner.
 *
 */
static int
safe_close(
    ZK_INOUT ZkSocketDescriptor * descriptor )
{
    int result;

# if defined WIN32
    shutdown(*descriptor, SD_BOTH);
    result = !closesocket(*descriptor);
# else
    result =  !close(*descriptor);
# endif

    (*descriptor) = ZKSOCKET_INVALID_DESCRIPTOR;

    return (result);
}


static void
trim_whitespace( char * text )
{
    char * source = text;
    char * target = text;

    while (((*source) != '\0') && isspace(*source))
    {
        ++source;
    }

    while ((*(target++) = *(source++)) != '\0')
    {
        /* do nothing */
    }

    while ((--target != text) && isspace(*target))
    {
        (*target) = '\0';
    }
}



/* --------------------------------------------------------------------------
 *
 * Groks the remote address structure and generates a textual
 * representation of the address for the general consumption of
 * users of the library.
 *
 */
static void
save_remote_path( ZkSocket s )
{
    uint32_t ip_addr = 0;
    uint16_t ip_port = 0;

    memset(s->remote_path, 0, sizeof(s->remote_path));

    switch (s->family)
    {
    case AF_INET:
        ip_addr = zkNboToHostUint32(
                     (char*) &s->remote_addr.m_inet.sa.sin_addr.s_addr );
        ip_port = zkNboToHostUint16(
                     (char*) &s->remote_addr.m_inet.sa.sin_port );

        snprintf(s->remote_path, sizeof s->remote_path,
                "%d.%d.%d.%d:%d",
                (ip_addr >> 24) & 0xFF,
                (ip_addr >> 16) & 0xFF,
                (ip_addr >>  8) & 0xFF,
                (ip_addr & 0xFF),
                ip_port);
        break;
    case AF_LOCAL:
        strncpy( s->remote_path,
                 s->local_addr.m_unix.sun_path,
                 sizeof(s->remote_path) - 1);
        break;
    }
}


/* --------------------------------------------------------------------------
 *
 * Reads a character string and interprets it an an unix socket address
 *
 */
static int
parse_unix_addr(
    ZK_INOUT ZkError               error,
    ZK_IN    const char *          path,
    ZK_OUT   struct sockaddr_un *  addr )
{
    const char * const func         = "parse_unix_addr";
    int                result       = 0;
    char *             trimmed_path = NULL;

    if (path == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "'path' is NULL");
    }
    else if (addr == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "'addr' is NULL");
    }
    else if (zkMemDup(error,(void**)&trimmed_path, path, strlen(path)+1))
    {
        trim_whitespace(trimmed_path);

        memset(addr,0,sizeof(*addr));

        addr->sun_family = AF_LOCAL;
        
        strncpy( addr->sun_path,
                 trimmed_path,
                 sizeof(addr->sun_path) - 1 );

# ifdef HAVE_SOCKADDR_SA_LEN
        addr->sun_len = SUN_LEN(addr);
# endif


        result = 1;
    }

    if (trimmed_path != NULL)
    {
        zkMemFree(NULL,(void**)&trimmed_path);
    }

    return (result);
}


/* --------------------------------------------------------------------------
 *
 * Reads a character string and interprets it an an IP subnet mask.  This
 * function understands both dot notation (eg., 255.255.0.0) and mask bit
 * field length (i.e., a number denoting the top most number of bits to
 * use in the mask).
 *
 * Note that path is assumed to be a string which the function is able
 * and allowed to modify.
 */
static int
parse_inet_mask(
    ZK_INOUT ZkError               error,
    ZK_IN    char *                path,
    ZK_OUT   uint32_t *            mask )
{
    const char * const func       = "parse_inet_mask";
    uint32_t           mask_value = 0;
    int                result     = 0;

    if (path == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "'path' is NULL");
    }
    else if (mask == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "'mask' is NULL");
    }
    else
    {
        trim_whitespace(path);
    
        if (strchr(path,'.') != NULL)
        {
            mask_value = inet_addr(path);
            if ( (mask_value == 0xFFFFFFFF)
                 && (strcmp(path,"255.255.255.255") != 0))
            {
                zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                              "invalid subnet mask (\"%s\")",
                              path);            
            }
            else
            {
                result = 1;
            }
        }
        else
        {
            char *   end_ptr  = NULL;
            uint32_t bit_count = strtoul(path,&end_ptr,10);
            
            if ((*end_ptr != '\0') || (bit_count > 32))
            {
                zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                              "invalid subnet mask size(\"%s\")",
                              path);            
            }
            else
            {
                /*
                 * HACK! - There's GOTTA be a better way to do this!
                 *
                 * We want the mask to have the same byte ordering as the
                 * parsed inet address to which it will be applier.  We
                 * DON'T want to assume (unless somebody can track down
                 * some hard docs) that this byte ordering MUST be any
                 * certain ordering across ALL implementations.  So, we
                 * convert the mask to an dot-format address and use
                 * the same parsing function to convert it back to an
                 * ordinal value which can be used for a mask.
                 */

                char buffer[12 + 3 + 1];

                mask_value = 0xFFFFFFFF << (32 - bit_count);
                
                snprintf(buffer,sizeof buffer,"%d.%d.%d.%d",
                         (mask_value >> 24) & 0xFF,
                         (mask_value >> 16) & 0xFF,
                         (mask_value >>  8) & 0xFF,
                         (mask_value >>  0) & 0xFF );
                
                mask_value = inet_addr(buffer);

                result = 1;
            }
        }

        if (result)
        {
            memcpy( mask,
                    &mask_value,
                    sizeof(mask_value) );
        }
    }

    return (result);
}    


/* --------------------------------------------------------------------------
 *
 * Reads a character string and interprets it an an internet address
 *
 */
int
parse_inet_addr(
    ZK_INOUT ZkError               error,
    ZK_IN    const char *          path,
    ZK_OUT   struct T_ZkInetAddr *   addr )
{
    const char * const func     = "parse_inet_addr";
    int                result   = 0;
    char *             host_str = NULL;

    if (path == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "'path' is NULL");
    }
    else if (addr == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "'addr' is NULL");
    }
    else if (zkMemDup(error,(void**)&host_str, path, strlen(path)+1))
    {
        char *    port_str = NULL;
        char *    mask_str = NULL;
        uint32_t  mask     = 0xFFFFFFFF;
        uint16_t  port     = 0;   

        result = 1;

        if (result
            && ((mask_str = strchr(host_str,'/')) != NULL))
        {
            *(mask_str++) = '\0';
            result = parse_inet_mask(error,mask_str,&mask);
        }

        if (result
            && ((port_str= strchr(host_str, ':')) != NULL))
        {
            char *             end_ptr    = NULL;
            unsigned long int  port_value = 0;

            *(port_str++) = '\0';

            trim_whitespace(port_str);

            port_value = strtoul(port_str, &end_ptr, 10);
            if ( (port_value > USHRT_MAX)
                 || (port_value < 1)
                 || (*end_ptr != '\0'))
            {
                result = 0;
                zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                              "invalid port (\"%s\")",
                              port_str);
            }
            else
            {
                port = (uint16_t) port_value;
            }
        }

        if (result != 0)
        {
            trim_whitespace(host_str);

            addr->mask          = mask;
            addr->sa.sin_family = AF_INET;
            addr->sa.sin_port   = htons(port);

# ifdef HAVE_SOCKADDR_SA_LEN
            addr->sa.sin_len = sizeof(*addr);
# endif

            if (host_str[0] == '\0')
            {
                addr->sa.sin_addr.s_addr = htonl(INADDR_ANY);
            }
            else
            {
                unsigned long       ip_addr  = 0;
                struct hostent *    host     = NULL;
                
                /*
                 * If we have a dot notation IP address instead of an text
                 * based hostname, convert it to a proper address and don't
                 * do the DNS lookup.
                 */
                
                ip_addr = inet_addr(host_str);
                
                if (ip_addr != 0xFFFFFFFF)
                {
                    memcpy( &(addr->sa.sin_addr.s_addr),
                            &ip_addr,
                            sizeof(ip_addr) );
                }
                else
                {
                    /*
                     * Figure out the internet address based on the hostname
                     */
/*
  Calling zkGetHostName was causing problems with reverse DNS lookup.
  This change has been approved by Kent. ( ericb 00/07/20 )
*/
/*                     if (!zkGetHostByName(error, host_str, &host)) */
                    if (! ( host = gethostbyname(host_str) ) )
                    {
                        result = 0;
                    }
                    else if (host == NULL)
                    {
                        result = 0;
                        zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKENOTFOUND,
                                      "host not found (%s)", host_str);
                    }
                    else if (host->h_length != 4)
                    {
                        result = 0;
                        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKEDATABAD,
                                   "Unknown format of host address");
                    }
                    else
                    {
                        memcpy( &(addr->sa.sin_addr.s_addr),
                                host->h_addr_list[0],
                                sizeof(uint32_t));
                    }
                }
            }
        }
    }

    if (host_str != NULL)
    {
        zkMemFree(NULL,(void**)&host_str);
    }

    return (result);
}


/* ------------------------------------------------------------------------
 *
 * Reads a charactes string and interprets it as a socket address
 *
 */
static int
parse_address(
    ZK_INOUT ZkError            error,
    ZK_IN    const char *       path,
    ZK_OUT   int *              out_family,
    ZK_OUT   ZkSocketAddress *  out_address )
{
    const char * const func   = "parse_address";
    int                result = 0;

    if (path == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "'path' is NULL");
    }
    else if (out_family == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "'out_family' is NULL");
    }
    else if (out_address == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "'out_address' is NULL");
    }
    else
    {
        memset(out_address, 0, sizeof(*out_address));

        /*
         * Unix domain sockets have absolute paths which means that they
         * must begin with a slash ('/') character.
         */

        if ((path[0] == '/')
            && parse_unix_addr(error,path,&(out_address->m_unix)))
        {
            (*out_family) = AF_LOCAL;
            result        = 1;
        }

        /*
         * Currently the only other option we support is Internet domain
         * sockets.
         */

        else if (parse_inet_addr(error, path, &(out_address->m_inet)))
        {
            (*out_family) = AF_INET;
            result        = 1;
        }
    }

    return (result);
}


/*
 * Taken from zkipClient.c
 */
static int
connect_with_timeout (
  ZkError err ,
  int sockfd ,
  const struct sockaddr * serv_addr ,
  int addrlen ,
  int tout )
{
  static const char * func = "ConnectWithTimeout";
  int    r         = 0;           /* Function return value (fail by default). */

#ifdef ZK_LINUX
  ZkBool bNonBlock = ZK_TRUE;     /* Try non-blocking connect by default */
  int fd_flags = 0;               /* Stores the old socket flags */

  /*
   * First, try to get the current flags on the socket descriptor then
   * set it to non-blocking mode.  Revert to normal connect if these
   * operations fail.
   */
  if ((fd_flags = fcntl(sockfd, F_GETFL)) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKIP, ZKESOCKET,
		  "fcntl() failed to retrieve current flags:  %s",
		  strerror(errno));
    bNonBlock = ZK_FALSE;
  }
  else if ((fcntl(sockfd, F_SETFL, (O_NONBLOCK | fd_flags)) < 0)) {
    zkErrorSet_va(err, func, ZKERR, ZKIP, ZKESOCKET,
		  "fcntl() failed to make socket non-blocking:  %s",
		  strerror(errno));
    bNonBlock = ZK_FALSE;
  } 
#endif /* ZK_LINUX */

  if (0 == connect(sockfd, serv_addr, addrlen)) {
    /* Immediate success */
    r = 1;
  }
#ifdef ZK_LINUX
  else if (bNonBlock && (errno == EINPROGRESS)) {
    /*
     * We should be waiting for the connection to happen now...
     */        
    struct timeval timeout;    /* Connection timeout */
    fd_set wfds;               /* read to write descriptor set */
    int s = 0;                 /* select() return code */

    timeout.tv_sec = tout;
    timeout.tv_usec = 0;
    FD_ZERO(&wfds);
    FD_SET(sockfd, &wfds);
    while ((s = zkSelectUpdateTimeout(err, sockfd + 1, NULL,
				      &wfds, NULL, &timeout)) < 0) {
      if (EINTR == errno) {
	/* An error occurred; interruptions not fatal. */
	continue;
      }
      else {
	zkErrorSet_va(err, func, ZKERR, ZKIP, ZKESOCKET,
		      "select() failed while waiting for connection:  %s",
		      strerror(errno));
	break;
      }
    }


    /*
     * If there is something for me to read, the connection attempt is
     * complete (s > 0).  Check error status.
     *
     * Otherwise, we have timed out (s == 0), or some other error has
     * occurred (s < 0; error code already set). 
     */
            
    if (0 < s) {
      int sock_err = 0;          /* delayed connect() return code */
      socklen_t sock_err_sz = 0; /* Size of the socket option response */

      sock_err_sz = sizeof(sock_err);
      if (getsockopt(sockfd, SOL_SOCKET, SO_ERROR,
		     (void *) &sock_err, &sock_err_sz) < 0) {
	zkErrorSet_va(err,func,ZKERR,ZKIP,ZKESOCKET,
		      "getsockopt() failed to get connect() error code: %s",
		      strerror(errno));
      } 
      else if (sock_err != 0) {
	zkErrorSet_va(err,func,ZKERR,ZKIP,ZKESOCKET,
		      "Delayed connect() failed:  %s", strerror(sock_err));
      } 
      else {  /* SUCCESS!!! */
	r = 1;
      }
    }
    else if (0 == s) {
      /* Timed out */
      zkErrorSet(err, func, ZKERR, ZKIP, ZKESOCKET,
		 "Connect() attempt has timed out");
    }           
  } 
#endif /* ZK_LINUX */
  else {
    /*
     * The connection failed for some other reason...
     */
            
    zkErrorSet_va(err, func, ZKERR, ZKIP, ZKESOCKET,
		  "connect() failed: %s", strerror(errno));
  }

    

#ifdef ZK_LINUX
  /*
   * Restore the socket flags to their original values.  
   */
    
  if (bNonBlock) {
    if (fcntl(sockfd,F_SETFL,fd_flags) < 0) {
      zkErrorSet_va(err,func,ZKERR,ZKIP,ZKESOCKET,
		    "fcntl() failed to restore socket flags:  %s",
		    strerror(errno));
      r = 0;
    }
  }
#endif /* ZK_LINUX */


  return r;
}

/* --------------------------------------------------------------------------
 *
 * Compares two address paths and returns whether or not they are the
 * same.  Takes flags (bitwise ORed together to determine exactly what
 * constitutes "sameness".
 *
 * ZKSOCKET_CMP_INETADDR - if the addresses given are internet domain
 *                         socket addresses, the IP addresses must match
 *                         for the addresses to be the evaluated as the
 *                         same.  Has no effect if the addresses are not
 *                         both internet domain addresses.
 *
 * ZKSOCKET_CMP_INETPORT - if the addresses given are internet domain
 *                         socket addresses, the ports must match for the
 *                         addresses to be evaluated as the same.  Has no
 *                         effect if the addresses are not both internet
 *                         domain addresses.
 */
int
zkSocketAddressCompare(
    ZK_INOUT ZkError       error,
    ZK_IN    uint32_t      flags,
    ZK_IN    const char *  path1,
    ZK_IN    const char *  path2,
    ZK_OUT   ZkBool *      is_same )
{
    const char * const func   = "zkSocketAddressCompare";
    int                result = 0;
    ZkSocketAddress    addr1;
    ZkSocketAddress    addr2;
    int                family1;
    int                family2;

    if (path1 == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "'path1' is NULL");
    }
    else if (path2 == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "'path2' is NULL");
    }
    else if (is_same == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "'is_same' is NULL");
    }
    else if (parse_address(error,path1,&family1,&addr1)
             && parse_address(error,path2,&family2,&addr2))
    {
        if (family1 != family2)
        {
            (*is_same) = ZK_FALSE;
            result = 1;
        }
        else if (family1 == AF_INET)
        {
            ZkBool same_port = ZK_FALSE;
            ZkBool same_addr = ZK_FALSE;

            addr1.m_inet.sa.sin_addr.s_addr &= addr1.m_inet.mask;
            addr2.m_inet.sa.sin_addr.s_addr &= addr1.m_inet.mask;

            if (((flags & ZKSOCKET_CMP_INETADDR) == 0)
                || ( addr1.m_inet.sa.sin_addr.s_addr == addr2.m_inet.sa.sin_addr.s_addr))
            {
                same_addr = ZK_TRUE;
            }

            if (((flags & ZKSOCKET_CMP_INETPORT) == 0)
                || (addr1.m_inet.sa.sin_port == addr2.m_inet.sa.sin_port))
            {
                same_port = ZK_TRUE;
            }

            (*is_same) = (same_addr && same_port);
            result = 1;
        }
        else if (family1 == AF_LOCAL)
        {
            (*is_same) = !strcmp(addr1.m_unix.sun_path,addr2.m_unix.sun_path);
            result = 1;
        }
        else
        {
            zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                          "unrecognized socket family (%d)",
                          family1);
            (*is_same) = ZK_FALSE;
        }
    }

    return (result);
}


/* --------------------------------------------------------------------------
 *
 * Creates a new disconnected socket.
 *
 */
int
zkSocketNew(
    ZK_INOUT ZkError       error,
    ZK_IN    int           family,  /* AF_INET, AF_LOCAL */
    ZK_IN    int           type,    /* SOCK_STREAM, SOCK_DGRAM, SOCK_RAW */
	ZK_IN	 int		   protocol, /* IPPROTO_xxx */
    ZK_OUT   ZkSocket *    out_socket )
{
    const char * const func       = "zkSocketNew";
    int                result     = 0;
    int                descriptor = ZKSOCKET_INVALID_DESCRIPTOR;

    if (out_socket == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "'out_socket' is NULL");
    }
    else if ((family != AF_INET) && (family != AF_LOCAL))
    {
        zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKEDATABAD,
                      "invalid socket family (0x%08x)", type);
    }
    else if ((type != SOCK_STREAM) && (type != SOCK_DGRAM) && (type != SOCK_RAW) )
    {
        zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKEDATABAD,
                      "invalid socket type (0x%08x)", type);
    }
    else
    {
        descriptor = socket(family, type, protocol);

        if (descriptor == ZKSOCKET_INVALID_DESCRIPTOR)
        {
            zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                       "failed to create socket");
        }
        else if (!zkSocketNewFromDescriptor( error,
                                             family,
                                             descriptor,
                                             out_socket ))
        {
            safe_close(&descriptor);
        }
        else
        {
            (*out_socket)->close_descriptor_on_free = CLOSE_ON;
            result = 1;
        }
    }

    return (result);
}


/* --------------------------------------------------------------------------
 *
 * This function allows existing code which makes use of file/socket
 * descriptors to be gradually converted to use the ZkSocket type.
 *
 */
int
zkSocketNewFromDescriptor(
    ZK_INOUT ZkError                error,
    ZK_IN    int                    family,  /* AF_INET, AF_LOCAL */
    ZK_IN    ZkSocketDescriptor     descriptor,
    ZK_OUT   ZkSocket *             out_socket )
{
    const char * const func   = "zkSocketNewFromDescriptor";
    int                result = 0;

    zkErrorClear(error);

    if (out_socket == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "'out_socket' is NULL");
    }
    else if ((family != AF_INET) && (family != AF_LOCAL))
    {
        zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKEDATABAD,
                      "invalid socket family (0x%08x)", family);
    }
    else if ( descriptor == ZKSOCKET_INVALID_DESCRIPTOR )
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKEDATABAD,
                   "invalid descriptor");
    }

    else if (zkMemCalloc(error, (void**)out_socket, sizeof(**out_socket),1))
    {
        (*out_socket)->descriptor      = descriptor;
        (*out_socket)->family          = family;
        (*out_socket)->timeout_accept  = ZKSOCKET_DFLT_TIMEOUT_ACCEPT;
        (*out_socket)->timeout_connect = ZKSOCKET_DFLT_TIMEOUT_CONNECT;
        (*out_socket)->timeout_read    = ZKSOCKET_DFLT_TIMEOUT_READ;
        (*out_socket)->timeout_write   = ZKSOCKET_DFLT_TIMEOUT_WRITE;
        (*out_socket)->block_size      = ZKSOCKET_DFLT_BLOCK_SIZE;
        (*out_socket)->queue_depth     = ZKSOCKET_DFLT_QUEUE_DEPTH;
        result = 1;
    }

    return (result);
}



/* --------------------------------------------------------------------------
 *
 * Frees the memory associated with this socket.  If the socket is open
 * then the connection is closed.
 *
 */
int
zkSocketFree(
    ZK_INOUT ZkError       error,
    ZK_INOUT ZkSocket *    io_socket )
{
    const char * const func   = "zkSocketFree";
    int                result = 0;

    zkErrorClear(error);

    if (io_socket == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "'io_socket' is NULL");
    }
    else if (*io_socket == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "'io_socket' doesn't point to anything");
    }
    else
    {
        if ( ((*io_socket)->descriptor != -1)
             && ((*io_socket)->close_descriptor_on_free))
        {
	    if (((*io_socket)->close_descriptor_on_free == CLOSE_ON) &&
                ((*io_socket)->family == AF_LOCAL))
	    {
		
                unlink((*io_socket)->local_addr.m_unix.sun_path);
	    }

            safe_close(&((*io_socket)->descriptor));
        }

        zkMemFree(NULL, (void**)io_socket);

        result = 1;
    }

    return (result);
}



/* ------------------------------------------------------------------------
 * Retrieves a string denoting the socket address of the remote entity
 */
int
zkSocketGetRemoteAddress(
    ZK_INOUT ZkError          error,
    ZK_IN    ZkSocket         s,
    ZK_OUT   const char * *   out_path )
{
    const char * const func   = "zkSocketGetRemoteAddress";
    int                result = 0;

    if (s == NULL)
    {
        result = 0;
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "socket pointer is NULL");
    }
    else if (out_path == NULL)
    {
        result = 0;
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "'out_path' is NULL");
    }
    else
    {
        (*out_path) = &s->remote_path[0];
        result = 1;
    }

    return (result);
}


/* ------------------------------------------------------------------------
 *
 * This function sets some high level options for this socket.
 *
 */
int
zkSocketSetOption(
    ZK_INOUT ZkError          error,
    ZK_IN    ZkSocket         s,
    ZK_IN    ZkSocketOption   option,
    ZK_IN    uint32_t         value )
{
    const char * const func   = "zkSocketSetOption";
    int                result = 1;
    int                optVal = 1;

    if (s == NULL)
    {
        result = 0;
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "socket pointer is NULL");
    }
    else
    {
        switch (option)
        {		
		case ZKSOCKET_OPTION_RecvTimeout:
#ifndef ZK_LINUX
			if( setsockopt( s->descriptor, SOL_SOCKET, SO_RCVTIMEO, (char *)&value, sizeof(value) ) < 0)
			{
				zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKERROR, 
                              "Failed to set SO_RCVTIMEO option: %s", strerror(errno));
                result = 0;
			}
            s->timeout_read    = value / 1000;
#endif
			break;
		case ZKSOCKET_OPTION_SendTimeout:
#ifndef ZK_LINUX
			if( setsockopt( s->descriptor, SOL_SOCKET, SO_SNDTIMEO, (char *)&value, sizeof(value) ) < 0)
			{
				zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKERROR, 
                              "Failed to set SO_SNDTIMEO option: %s", strerror(errno));
                result = 0;
			}
            s->timeout_write   = value / 1000;
#endif
			break;
        case ZKSOCKET_OPTION_BlockSize:
            s->block_size = value;
            break;
        case ZKSOCKET_OPTION_Timeout:
            s->timeout_accept  = value;
            s->timeout_connect = value;
            s->timeout_read    = value;
            s->timeout_write   = value;
            break;
        case ZKSOCKET_OPTION_TimeoutAccept:
            s->timeout_accept = value;
            break;
        case ZKSOCKET_OPTION_TimeoutConnect:
            s->timeout_connect = value;
            break;
        case ZKSOCKET_OPTION_TimeoutRead:
            s->timeout_read = value;
            break;
        case ZKSOCKET_OPTION_TimeoutWrite:
            s->timeout_write = value;
            break;
        case ZKSOCKET_OPTION_QueueDepth:
            if (value < 1)
            {
                result = 0;
                zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                              "invalid  queue depth (%u)", value); 
            }
            else
            {
                s->queue_depth = value;
            }
            break;
        case ZKSOCKET_OPTION_ReuseAddress:
            if (setsockopt(s->descriptor, SOL_SOCKET, 
                           SO_REUSEADDR, (char *)&optVal, sizeof(int)) < 0) 
            {
                zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKERROR, 
                              "Failed to set SO_REUSEADDR option: %s", strerror(errno));
                result = 0;
            }
            break;
	case ZKSOCKET_OPTION_NoTransmitDelay:
  	    if (setsockopt(s->descriptor, SOL_TCP,
                           TCP_NODELAY, (char *)&optVal, sizeof(int)) < 0) 
            {
                zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKESOCKET, 
                              "Failed to set TCP_NODELAY option: %s", strerror(errno));
                result = 0;
            }
            break;
        default:
            result = 0;
            zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                          "Unrecognized option (0x%08x)", option); 
        }
    }

    return (result);
}



/* --------------------------------------------------------------------------
 *
 * Binds this socket to a port.
 *
 * TODO: Do we want to be able to bind to a particular host address?
 */
int
zkSocketBind(
    ZK_INOUT ZkError       error,
    ZK_IN    ZkSocket      s,
    ZK_IN    const char *  path )
{
    const char * const  func       = "zkSocketBind";
    int                 family     = -1;
    int                 result     = 0;
    ZkSocketAddress     local_addr;

    if (s == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "socket pointer is NULL");
    }
    else if (path == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "path pointer is NULL");
    }
    else if (s->descriptor == ZKSOCKET_INVALID_DESCRIPTOR)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                   "invalid socket descriptor");
    }
    else if (parse_address(error, path, &family, &local_addr))
    {
        struct sockaddr * addr   = NULL;
        int               size   = 0;

        if (family != s->family)
        {
            zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                          "mismatched address family! (socket=%d; address=%d)",
                          s->family,
                          family);
        }
        else switch (family)
        {
        case AF_INET:
            addr = (struct sockaddr *) &local_addr.m_inet.sa;
            size = sizeof(local_addr.m_inet.sa);
            break;
        case AF_LOCAL:
            addr = (struct sockaddr *) &local_addr.m_unix;
            size = sizeof(local_addr.m_unix);
            break;
        default:
            zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                          "unrecognized address family (%d)",
                          family);
        }

        if ((addr != NULL) && (size != 0))
        {
            if (bind(s->descriptor, addr, size))
            {
                zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                              "cannot bind to path (%s)", path);
            }
            else
            {
                s->local_addr = local_addr;
                result        = 1;
            }
        }
    }

    return (result);
}



/* --------------------------------------------------------------------------
 *
 * Puts this socket into the listen state so that it can accept
 * connections.
 *
 */
int
zkSocketListen(
    ZK_INOUT ZkError       error,
    ZK_IN    ZkSocket      s )
{
    const char * const  func    = "zkSocketBind";
    int                 result  = 0;

    if (s == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "socket pointer is NULL");
    }
    else if (s->descriptor == ZKSOCKET_INVALID_DESCRIPTOR)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                   "invalid socket descriptor");
    }
    else if (listen(s->descriptor,s->queue_depth))
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                   "cannot listen using this socket");
    }
    else
    {
        result = 1;
    }

    return (result);
}



/* --------------------------------------------------------------------------
 *
 * Accept a connection from a remote client
 *
 */
int
zkSocketAccept(
    ZK_INOUT ZkError       error,
    ZK_IN    ZkSocket      in_socket,
    ZK_OUT   ZkSocket *    out_socket )
{
    const char * const  func    = "zkSocketAccept";
    int                 result  = 0;

    if (in_socket == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "socket pointer is NULL");
    }
    else if (out_socket == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "socket pointer is NULL");
    }
    else if (in_socket->descriptor == ZKSOCKET_INVALID_DESCRIPTOR)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                   "invalid socket descriptor");
    }
    else
    {
        fd_set             descriptor_set;
        int                select_result;
        struct timeval     timeout;
        struct timeval *   timeout_ptr = NULL;

        FD_ZERO(&descriptor_set);
        FD_SET(in_socket->descriptor, &descriptor_set);

        (*out_socket)   = NULL;
        timeout.tv_sec  = in_socket->timeout_accept;
        timeout.tv_usec = 0;

        if (in_socket->timeout_accept != ZKSOCKET_TIMEOUT_INFINITE)
        {
            timeout_ptr = &timeout;
        }

        select_result = select( in_socket->descriptor + 1,
                                &descriptor_set,
                                NULL, NULL,
                                timeout_ptr );

        if (select_result < 0)
        {
            /*
             * Any failure other than interuption is evil!
             */
# if defined (WIN32)
            if (WSAEINTR == WSAGetLastError())
# else
            if (EINTR == errno)
# endif
            {
                result = 1;
                zkErrorSet(error,func,ZKINFO,ZKMSOCKET,ZKESOCKET,
                           "interrupted");
            }
            
        }
        else if (select_result == 0)
        {
            /*
             * We've timed out
             */

            result = 1;
            zkErrorSet(error,func,ZKINFO,ZKMSOCKET,ZKESOCKET,
                       "timeout");
        }
        else
        {
            /*
             * There's a connection waiting to be accepted
             */
            struct sockaddr *    addr       = NULL;
            int                  size       = 0;
            ZkSocketDescriptor   descriptor = ZKSOCKET_INVALID_DESCRIPTOR;
            ZkSocketAddress      remote_addr;
            
            memset(&remote_addr,0,sizeof(remote_addr));

            switch (in_socket->family)
            {
            case AF_INET:
                addr = (struct sockaddr *) &remote_addr.m_inet.sa;
                size = sizeof(remote_addr.m_inet.sa);
                break;
            case AF_LOCAL:
                addr = (struct sockaddr *) &remote_addr.m_unix;
                size = sizeof(remote_addr.m_unix);
                break;
            default:
                zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                              "unrecognized address family (%d)",
                              in_socket->family);
            }
            
            if ((addr != NULL) && (size != 0))
            {
                descriptor = accept(in_socket->descriptor, addr, &size);
                
                if (descriptor == ZKSOCKET_INVALID_DESCRIPTOR)
                {
                    zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                               "failed to accept connection");
                }
                else if (!zkSocketNewFromDescriptor( error,
                                                     in_socket->family,
                                                     descriptor,
                                                     out_socket ))
                {
                    safe_close(&descriptor);
                }
                else
                {
                    (*out_socket)->local_addr  = in_socket->local_addr;
                    (*out_socket)->remote_addr = remote_addr;
                    /* keep the link open if this socket is closed */
                    (*out_socket)->close_descriptor_on_free = CLOSE_ON_KEEP_LINK;
                    
                    save_remote_path(*out_socket);
                    result = 1;
                }
            }
        }
    }

    return (result);
}



/* --------------------------------------------------------------------------
 *
 * Accept a connection from a remote client
 *
 */
int
zkSocketAcceptWithTimeoutUpdate(
    ZK_INOUT ZkError            error,
    ZK_IN    ZkSocket           in_socket,
    ZK_INOUT uint32_t *         io_timeout_in_secs,
    ZK_OUT   ZkSocket *         out_socket )
{
    const char * const  func    = "zkSocketAcceptWithTimeoutUpdate";
    int                 result  = 0;

    if (in_socket == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "socket pointer is NULL");
    }
    else if (out_socket == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                  "socket pointer is NULL");
    }
    else if (in_socket->descriptor == ZKSOCKET_INVALID_DESCRIPTOR)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                   "invalid socket descriptor");
    }
    else
    {
        fd_set             descriptor_set;
        struct timeval     timeout;
        int                select_result;

        FD_ZERO(&descriptor_set);
        FD_SET(in_socket->descriptor, &descriptor_set);

        timeout.tv_sec  = (*io_timeout_in_secs);
        timeout.tv_usec = 0;

        select_result = zkSelectUpdateTimeout( error,
                                               in_socket->descriptor + 1,
                                               &descriptor_set,
                                               NULL, NULL,
                                               &timeout );

        (*io_timeout_in_secs) = timeout.tv_sec;
        (*out_socket)         = NULL;

        if (select_result < 0)
        {
            /*
             * Any failure other than interuption is evil!
             */
# if defined (WIN32)
            if (WSAEINTR == WSAGetLastError())
# else
            if (EINTR == errno)
# endif
            {
                result = 1;
                zkErrorSet(error,func,ZKINFO,ZKMSOCKET,ZKESOCKET,
                           "interrupted");
            }
            
        }
        else if (select_result == 0)
        {
            /*
             * We've timed out
             */

            result = 1;
            zkErrorSet(error,func,ZKINFO,ZKMSOCKET,ZKESOCKET,
                       "timeout");
        }
        else
        {
            /*
             * There's a connection waiting to be accepted
             */
            struct sockaddr *    addr       = NULL;
            int                  size       = 0;
            ZkSocketDescriptor   descriptor = ZKSOCKET_INVALID_DESCRIPTOR;
            ZkSocketAddress      remote_addr;
            
            memset(&remote_addr,0,sizeof(remote_addr));
            
            switch (in_socket->family)
            {
            case AF_INET:
                addr = (struct sockaddr *) &remote_addr.m_inet.sa;
                size = sizeof(remote_addr.m_inet.sa);
                break;
            case AF_LOCAL:
                addr = (struct sockaddr *) &remote_addr.m_unix;
                size = sizeof(remote_addr.m_unix);
                break;
            default:
                zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                              "unrecognized address family (%d)",
                              in_socket->family);
            }
            
            if ((addr != NULL) && (size != 0))
            {
                descriptor = accept(in_socket->descriptor, addr, &size);
                
                if (descriptor == ZKSOCKET_INVALID_DESCRIPTOR)
                {
                    zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                               "failed to accept connection");
                }
                else if (!zkSocketNewFromDescriptor( error,
                                                     in_socket->family,
                                                     descriptor,
                                                     out_socket ))
                {
                    safe_close(&descriptor);
                }
                else
                {
                    (*out_socket)->local_addr  = in_socket->local_addr;
                    (*out_socket)->remote_addr = remote_addr;
                    /* keep the link open if this socket is closed */
                    (*out_socket)->close_descriptor_on_free = CLOSE_ON_KEEP_LINK;
                    
                    save_remote_path(*out_socket);
                    result = 1;
                }
            }
        }
    }

    return (result);
}

/*
Connect this socket to a remote host at ZkSocketAddress
*/
int
zkSocketConnectAddr(
	ZK_INOUT	ZkError					error,
	ZK_IN		ZkSocket				s,
	ZK_IN		const ZkSocketAddress	address)
{
    const char * const  func = "zkSocketConnectAddr";
    const struct sockaddr *	addr = (const struct sockaddr *) &address.m_inet.sa;
    int                 result =  0;


    if (s == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "socket pointer is NULL");
    }
    else if (s->descriptor ==  ZKSOCKET_INVALID_DESCRIPTOR)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                   "invalid socket descriptor");
    }
	else if (addr == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "addr is NULL");
    }
	else if ( connect_with_timeout(error, s->descriptor, addr, sizeof(address.m_inet.sa), s->timeout_connect) != ZK_TRUE )
    {        
		zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                          "connect call failed (%s)", strerror(errno));

        safe_close(&(s->descriptor));
    }
    else
    {        
        result = 1;
    }

	return result;
}


/* --------------------------------------------------------------------------
 *
 * Connect this socket object to a remote host.
 *
 * TODO : this function should do a non-blocking connect and then do a
 *        select to allow the connect to timeout using the user specified
 *        timeout value.  See "Unix Network Programming - Volume 1" by
 *        Stephens (section 15.4).
 */
int
zkSocketConnect(
    ZK_INOUT ZkError       error,
    ZK_IN    ZkSocket      s,
    ZK_IN    const char *  path )
{
    const char * const  func   = "zkSocketConnect";
    int                 family = -1;
    int                 result =  0;
    ZkSocketAddress     remote_addr;

    memset(&remote_addr,0,sizeof(remote_addr));

    if (s == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "socket pointer is NULL");
    }
    else if (s->descriptor ==  ZKSOCKET_INVALID_DESCRIPTOR)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                   "invalid socket descriptor");
    }
    else if (path == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "path is NULL");
    }
    else if (parse_address(error, path, &family, &remote_addr))
    {
        struct sockaddr * addr   = NULL;
        int               size   = 0;

        if (family != s->family)
        {
            zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                          "mismatched address family! (socket=%d; address=%d)",
                          s->family,
                          family);
        }
        else switch (family)
        {
        case AF_INET:
            addr = (struct sockaddr *) &remote_addr.m_inet.sa;
            size = sizeof(remote_addr.m_inet.sa);
            break;
        case AF_LOCAL:
            addr = (struct sockaddr *) &remote_addr.m_unix;
            size = sizeof(remote_addr.m_unix);
            break;
        default:
            zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                          "unrecognized address family (%d)",
                          family);
        }

        if ((addr != NULL) && (size != 0))
        {
            /*
             * TODO : add select to support timeout
             */
            if ( connect_with_timeout(error, s->descriptor, addr, size, s->timeout_connect) != ZK_TRUE )
            {
                zkErrorSet_va(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                           "connect to %s failed", path);
                safe_close(&(s->descriptor));
            }
            else
            {
                save_remote_path(s);
                result = 1;
            }
        }
    }

    return (result);
}


/* --------------------------------------------------------------------------
 *
 * Wait for data to be available on any of a list of sockets, or
 * for one of those sockets to be closed.
 *
 * This function will attempt to perform a select on the list of
 * sockets, waiting for readability. On completion whichSocket will
 * be the index of the first readable socket, or -1.
 *
 * The return value is 0 on failure, nonzero otherwise.
 *
 */
int
zkSocketWaitReadableOrClosed(
	ZK_INOUT ZkError				error,
	ZK_IN	 ZkList					list_of_sockets,
	ZK_IN    struct timeval			*timeout,
	ZK_OUT   int					*whichSocket,
    ZK_IN    int                    signal_fd )
{
    const char * const	func        = "zkSocketWaitReadableOrClosed";
    int					result      = 0;
	size_t				nSockets	= 0;
	size_t				i			= 0;
	ZkSocket			curSocket	= NULL;
	fd_set	readset;
    struct timeval      tv_backup;

    zkErrorClear(error);
	FD_ZERO(&readset);

	if(!whichSocket)
	{
	    zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
		           "whichSocket pointer is NULL");
	}
	else if(!list_of_sockets)
	{
	    zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
		           "list pointer is NULL");
	}
	else if(!zkListSize( error, list_of_sockets, &nSockets))
	{
		// zkListSize failed... We'll return 0.
	}
	else if(nSockets < 1)
	{
	    zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
		           "list_of_sockets is empty");
		// Passed us zero sockets... We'll return 0.
	}
	else if(!zkListMoveCursor(error, list_of_sockets, LIST_HEAD))
	{
		// zkListMoveCursor failed... We'll return 0.
	}
	else
	{
		int iRet = 0;
        int high_descriptor = -1;

		*whichSocket = -1;


		for(i = 0; i < nSockets; i++)
		{
			zkListGet(error, list_of_sockets, (void **) &curSocket, LIST_FORWARDS);
            if ( curSocket->descriptor < 0 ) continue; // Do not set invalid sockets (causes SEGV) --ol
			FD_SET(curSocket->descriptor, &readset);
            if ((int)curSocket->descriptor + 1 > high_descriptor)
			{
				high_descriptor = curSocket->descriptor + 1;
			}
		}

#ifdef ZK_LINUX
        if (signal_fd > 0) {
            FD_SET(signal_fd, &readset);
            if (signal_fd >= high_descriptor) high_descriptor = signal_fd + 1;
        }

DoSelect:
#endif
        memcpy(&tv_backup, (void *)timeout, sizeof(struct timeval));
        if (high_descriptor <= 0) 
        {
            zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                       "list_of_sockets is empty");
			iRet = -1;
        }
		else if(!(iRet = select(high_descriptor, &readset, NULL, NULL, timeout) ))
		{
			// Timed out.. This is not a failure.
			result = 1;
		}
        memcpy((void *)timeout, &tv_backup, sizeof(struct timeval));
#ifdef ZK_LINUX
	    if(iRet == -1 && errno == EINTR)
		{
			goto DoSelect;
		}
#endif
		if(iRet == -1)
		{
			
            zkErrorSet(error,func,ZKINFO,ZKMSOCKET,ZKESOCKET,
                       "select call failed");
			// select failed... We'll return 0.
		}
		else if(!zkListMoveCursor(error, list_of_sockets, LIST_HEAD))
		{
			// zkListMoveCursor failed... We'll return 0.
		}
		else
		{
			for(i = 0; i < nSockets; i++)
			{
				zkListGet(error, list_of_sockets, (void **) &curSocket, LIST_FORWARDS);
                // We need to make sure the descriptor is valid on Linux or
                // FD_ISSET will segfault. This can happen if it got closed while we
                // were in the select and somebody set the value to indicate so, --ol
				if(curSocket->descriptor >= 0 && FD_ISSET(curSocket->descriptor, &readset))
				{
					*whichSocket = i;
					result = 1;
					break;
				}
			}
#ifdef ZK_LINUX         
            if (signal_fd > 0 && FD_ISSET(signal_fd, &readset))
            {
                *whichSocket = -2;
                result = 1;
            }
#endif
		}
	}

	return (result);
}

/* --------------------------------------------------------------------------
 *
 * Attempt to retrieve data from the connected socket.
 *
 * This function will attempt to read 'max_bytes_to_read' octets from the
 * socket and will put the retrieved data into the supplied target_buffer.
 * It will return successfully after reading some non-zero quantity of
 * data into the target buffer.
 */
int
zkSocketRead(
    ZK_INOUT ZkError       error,
    ZK_IN    ZkSocket      s,
    ZK_INOUT void *        target_buffer,
    ZK_IN    uint32_t      max_bytes_to_read,
    ZK_OUT   uint32_t *    out_bytes_read )
{
    const char * const func         = "zkSocketRead";
    int                result       = 0;

    zkErrorClear(error);

    if (s == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "socket pointer is NULL");
    }
    else if (target_buffer == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "target buffer is NULL");
    }
    else if (out_bytes_read == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "'out_bytes_read' is NULL");
    }
    else if (s->descriptor == ZKSOCKET_INVALID_DESCRIPTOR)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                   "socket descriptor is invalid");
    }
    else
    {
        fd_set             readfds;           /* read descriptor field */
        struct timeval     timeout;           /* Timeout for select */
        struct timeval *   timeout_ptr     = NULL;
        int                select_result   = 0;
        
        memset(target_buffer,0,max_bytes_to_read);

        FD_ZERO(&readfds);
        FD_SET(s->descriptor,&readfds);
        
        timeout.tv_sec  = s->timeout_read;
        timeout.tv_usec = 0;

        if (s->timeout_read != ZKSOCKET_TIMEOUT_INFINITE)
        {
            timeout_ptr = &timeout;
        }

        select_result = select( s->descriptor + 1,
                                &readfds,
                                NULL,
                                NULL,
                                timeout_ptr );
        if (select_result > 0)
        {
            int bytes_read = recv( s->descriptor,
                                   target_buffer,
                                   max_bytes_to_read,
                                   0 );
            
            if (bytes_read > 0)
            {
                (*out_bytes_read) = bytes_read;
                result = 1;
            }
            /*
             * All errors except for interuptions ...
             */
            else if (bytes_read < 0)
            {
#if defined(WIN32)
                if (WSAEINTR == WSAGetLastError())
#else
                if (EINTR == errno)
#endif
                {
                    zkErrorSet(error,func,ZKINFO,ZKMSOCKET,ZKESOCKET,
                               "read operation was interrupted (1)");
                    result = 1;
                }
                else
                {
                    zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                               "No data to read");
                }
            }
            /*
             * bytes_read == zero
             */
            else
            {
                zkErrorSet(error,func,ZKWARN,ZKMSOCKET,ZKENODATA,
                           "select returned with 0 bytes to read");
            }
        }
        else if (select_result < 0)
        {
#if defined(WIN32)
            if (WSAGetLastError() == WSAEINTR)
#else
            if (errno == EINTR)
#endif
            {
                zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                           "read operation was interrupted (2)");  
                result = 1;
            }
            else
            {
                zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                           "Error reading from remote entity");  
            }
        }
        else /* (select_result == 0) -> timeout */
        {
            zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                       "Connection to remote entity has timed out");
        }
    }

    return (result);
}



/* --------------------------------------------------------------------------
 *
 * Attempt to retrieve data from the connected socket.
 *
 * This function will attempt to read exactly 'bytes_to_read' octets from
 * the socket and will put the retrieved data into the supplied target_buffer.
 * To do this read, the function divides the read operation into blocks and
 * continues reading blocks until the entirety of the requested data has been
 * retrieved or until an error or timeout occurs.
 *
 * This function will timeout if it fails to retrieve a block within the
 * currently specified timeout period.
 *
 * Note that this functionality could have been implemented in terms of
 * zkSocketRead; however, this would have incurred unnecessary function
 * overhead in terms of parameter validation for each iterative call.
 */
int
zkSocketReadAll(
    ZK_INOUT ZkError       error,
    ZK_IN    ZkSocket      s,
    ZK_INOUT void *        target_buffer,
    ZK_IN    uint32_t      bytes_to_read )
{
    const char * const func         = "zkSocketFillBuffer";
    int                result       = 0;

    zkErrorClear(error);

    if (s == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "socket pointer is NULL");
    }
    else if (target_buffer == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "target buffer is NULL");
    }
    else if (s->descriptor == ZKSOCKET_INVALID_DESCRIPTOR)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                   "socket descriptor is invalid");
    }
    else
    {
        fd_set             readfds;           /* read descriptor field */
        struct timeval     timeout;           /* Timeout for select */
        struct timeval *   timeout_ptr     = NULL;
        uint8_t *          buffer          = (uint8_t *) target_buffer;
        uint32_t           bytes_remaining = bytes_to_read;
        int                failed          = ZK_FALSE;
        uint32_t           chunk_size      = ( (bytes_remaining > s->block_size)
                                            ? s->block_size
                                            : bytes_remaining );

        if (s->timeout_write != ZKSOCKET_TIMEOUT_INFINITE)
        {
            timeout_ptr = &timeout;
        }

        memset(buffer,0,bytes_remaining);
        
        while ((!failed) && (bytes_remaining > 0))
        {
            int select_result = 0;
            
            FD_ZERO(&readfds);
            FD_SET(s->descriptor,&readfds);
            
            timeout.tv_sec  = s->timeout_read;
            timeout.tv_usec = 0;
            
            select_result = select( s->descriptor + 1,
                                    &readfds,
                                    NULL,
                                    NULL,
                                    timeout_ptr );
            if (select_result > 0)
            {
                int bytes_read = recv( s->descriptor,
                                       buffer + bytes_to_read - bytes_remaining,
                                       chunk_size,
                                       0 );
                
                if (bytes_read > 0)
                {
                    bytes_remaining -= bytes_read;
                    if (bytes_remaining < chunk_size)
                    {
                        chunk_size = bytes_remaining;
                    }
                }
                /*
                 * All errors except for interuptions ...
                 */
                else if (bytes_read < 0)
                {
#if defined(WIN32)
                    if (WSAEINTR != WSAGetLastError())
#else
                    if (EINTR != errno)
#endif
                    {
                        failed = ZK_TRUE;
                        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                                   "No data to read");
                    }
                }
                /*
                 * bytes_read == zero
                 */
                else
                {
                    failed = ZK_TRUE;
                    zkErrorSet(error,func,ZKWARN,ZKMSOCKET,ZKENODATA,
                               "select returned with 0 bytes to read");
                }
            }
            else if (select_result < 0)
            {
#if defined(WIN32)
                if (WSAGetLastError() != WSAEINTR)
#else
                if (errno != EINTR)
#endif
                {
                    failed = ZK_TRUE;
                    zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                               "Error reading from remote entity");  
                }
            }
            else /* (select_result == 0) -> timeout */
            {
                failed = ZK_TRUE;
                zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                           "Connection to remote entity has timed out");
            }
        }
        
        if (!failed)
        {
            if (bytes_remaining != 0)
            {
                zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                           "Failed to receive requested number of bytes");  
            }
            else
            {
                result = 1;
            }
        }
    }

    return (result);
}



/* --------------------------------------------------------------------------
 *
 * Write data to this socket.
 *
 * Writes the number of octets denoted by 'bytes_to_write' to the socket
 * from the given 'source_buffer'.
 *
 */
int zkSocketWrite(
    ZK_INOUT ZkError                error,
    ZK_IN    ZkSocket               s,
    ZK_INOUT const void *           source_buffer,
    ZK_IN    uint32_t               max_bytes_to_write,
    ZK_OUT   uint32_t *             out_bytes_written )
{
    const char * const func   = "zkSocketWrite";
    int                result = 0;

    zkErrorClear(error);

    if (s == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "socket pointer is NULL");
    }
    else if (source_buffer == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "source buffer is NULL");
    }
    else if (out_bytes_written == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "'out_bytes_written' is NULL");
    }
    else if (s->descriptor == ZKSOCKET_INVALID_DESCRIPTOR)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                   "socket descriptor is invalid");
    }
    else
    {
        fd_set             writefds;          /* write descriptor set */
        struct timeval     timeout;           /* timeout for select */
        struct timeval *   timeout_ptr     = NULL;
        int                select_result   = 0;
        
        FD_ZERO(&writefds);
        FD_SET(s->descriptor,&writefds);
        
        timeout.tv_sec  = s->timeout_write;
        timeout.tv_usec = 0;

        (*out_bytes_written) = 0;

        if (s->timeout_write != ZKSOCKET_TIMEOUT_INFINITE)
        {
            timeout_ptr = &timeout;
        }

        select_result = select( s->descriptor + 1,
                                NULL,
                                &writefds,
                                NULL,
                                timeout_ptr );
        if (select_result > 0)
        {
            int bytes_written = send( s->descriptor,
                                      source_buffer,
                                      max_bytes_to_write,
                                      MSG_NOSIGNAL );
            
            if (bytes_written > 0)
            {
                (*out_bytes_written) = bytes_written;
                result = 1;
            }
            /*
             * All errors except for interuptions ...
             */
            else if (bytes_written <= 0)
            {
#if defined(WIN32)
                if (WSAEINTR == WSAGetLastError())
#else
                if (EINTR == errno)
#endif
                {
                    zkErrorSet(error,func,ZKINFO,ZKMSOCKET,ZKESOCKET,
                               "write operation was interrupted (1)");
                    result = 1;
                }
                else
                {
                    zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                               "could not write data to remote entity");
                }
            }
        }
        else if (select_result < 0)
        {
#if defined(WIN32)
            if (WSAGetLastError() == WSAEINTR)
#else
            if (errno == EINTR)
#endif
            {
                zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKESOCKET,
                           "write operation was interrupted (2)");  
                result = 1;
            }
            else
            {
                zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                           "Error writing to remote entity");  
            }
        }
        else /* (select_result == 0) -> timeout */
        {
            zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                       "Connection to remote entity has timed out");
        }
    }

    return (result);
}



/* --------------------------------------------------------------------------
 *
 * Write data to this socket.
 *
 * Writes the number of octets denoted by 'bytes_to_write' to the socket
 * from the given 'source_buffer'.  It performs the write in block_size
 * increments and fails if it cannot write a block within the given
 * timeout period.
 *
 */
int
zkSocketWriteAll(
    ZK_INOUT ZkError       error,
    ZK_IN    ZkSocket      s,
    ZK_INOUT const void *  source_buffer,
    ZK_IN    uint32_t      bytes_to_write )
{
    const char * const func         = "zkSocketWrite";
    int                result       = 0;

    zkErrorClear(error);

    if (s == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "socket pointer is NULL");
    }
    else if (s->descriptor < 0)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                   "socket descriptor is invalid");
    }
    else
    {
        fd_set            writefds;       /* read descriptor field */
        struct timeval    timeout;        /* Timeout value for select */
        struct timeval *  timeout_ptr     = NULL;
        const uint8_t *   buffer          = (const uint8_t *) source_buffer;
        uint32_t          bytes_remaining = bytes_to_write;
        int               failed          = ZK_FALSE;
        uint32_t          chunk_size      = ( (bytes_remaining > s->block_size)
                                               ? s->block_size
                                               : bytes_remaining );

        if (s->timeout_write != ZKSOCKET_TIMEOUT_INFINITE)
        {
            timeout_ptr = &timeout;
        }

        while ((!failed) && (bytes_remaining > 0))
        {
            int select_result = 0;

            FD_ZERO(&writefds);
            FD_SET(s->descriptor,&writefds);

            timeout.tv_sec = s->timeout_write;
            timeout.tv_usec = 0;

            select_result = select( s->descriptor + 1,
                                    NULL,
                                    &writefds,
                                    NULL,
                                    timeout_ptr );

            if (select_result > 0)
            {
                int bytes_written = send( s->descriptor,
                                          buffer + bytes_to_write - bytes_remaining,
                                          chunk_size,
                                          MSG_NOSIGNAL );
                
                if (bytes_written > 0)
                {
                    bytes_remaining -= bytes_written;
                    if (bytes_remaining < chunk_size)
                    {
                        chunk_size = bytes_remaining;
                    }
                }
                /*
                 * Everything except for interuptions are errors
                 */
                else if (bytes_written <= 0)
                {
#if defined(WIN32)
                    if  (WSAEINTR != WSAGetLastError())
#else
                    if (EINTR != errno)
#endif
                    {
                        failed = ZK_TRUE;
                        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKEWRITE,
                                   "Could not write data to remote entity");
                    }
                }
            }
            else if (select_result < 0)
            {
#if defined(WIN32)
                if (WSAGetLastError() != WSAEINTR)
#else
                if (errno != EINTR)
#endif
                {
                    failed = ZK_TRUE;
                    zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                               "failed to write data");
                }
            }
            else /* (select_result == 0) -> timeout */
            {
                failed = ZK_TRUE;
                zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                           "Connection to remote entity has timed out");
            }
        }

        if (!failed)
        {
            if (bytes_remaining != 0)
            {
                zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                           "Failed to write requested number of bytes");  
            }
            else
            {
                result = 1;
            }
        }
    }

    return (result);
}

/* --------------------------------------------------------------------------
 *
 * return File Descriptor of this socket (for use in select statement).
 *
 * added by Frank Foster, Apr. 2000
 *
 */
int
zkSocketGetFD(
    ZK_INOUT ZkError              error,
    ZK_IN    ZkSocket             s,
    ZK_OUT   ZkSocketDescriptor*  fd)
{
    const char * const func         = "zkSocketGetFD";
    int                result       = 0;

    zkErrorClear(error); 

    if (s == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKENULLPTR,
                   "socket pointer is NULL");
    }
    else if (s->descriptor < 0)
    {
        zkErrorSet(error,func,ZKERR,ZKMSOCKET,ZKERROR,
                   "socket descriptor is invalid");
    }
    else
    {
        *fd = s->descriptor;
        result = 1;
    }

    return result;
}


