#if defined( ZK_LINUX ) || defined( ZK_SOLARIS )

/*
 * Timer manipulation
 *
 * $Id: zktimer.c,v 1.4.18.1 2000/10/11 13:20:03 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Nicolas Roy <nicolas@zeroknowledge.com>

 */

#include "zks.h"
#include "zkIntTypes.h"
#include "zkTypes.h"
#include "zkDefs.h"
#include "zkError.h"
#include "zkallocator.h"

#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <sys/times.h>
 
#include "zkAlloc.h"
#include "zktimer.h"


struct S_Timer {
  /*
   * is it started
   */
  ZkBool isStarted;
  /*
   * Allocator object
   */
  ZkAllocator allocator;
  /*
   * Real time
   */
  clock_t    realStartTime;
  clock_t    realStopTime;
  /* 
   * Cpu consumming time
   */
  struct tms cpuStartTime;
  struct tms cpuStopTime;
  /* 
   * Total time spent into pause state
   */
  struct S_Pause {
    ZkBool     state;                  /* State (ZK_TRUE) is in pause state */
    clock_t    real_cummulator;        /* Total real time spent in pause state */
    clock_t    real_last;              /* Last  real time pause was invoke */
    struct tms tms_cummulator;         /* Total tms time spent in pause state */
    struct tms tms_last;               /* Last  tms time pause was invoke */
  } pause;
};
 

/*
** Create a new Timer
*/
DLLExport int zkTimerNew
(
 ZkError     err,                      /* [in]   ZkError */                   
 ZkTimer    *timerP,                   /* [out]  pointer on a ZkTimer */
 ZkAllocator allocator                 /* Allocator for share mem usage */
 )
{
  ZkBool          success = ZK_TRUE;
  const char     *fn      = "zkTimerNew"; 

  zkErrorClear(err);

  if( !timerP )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "timerP is null");
      success = ZK_FALSE;
    }
  else
    {
      if (allocator) {
        if ( ! zkAllocatorMemCalloc(err, allocator, 1,
                                    (uint32_t)sizeof(struct S_Timer),
                                    (void**)timerP) ) {
          success = ZK_FALSE;
        }
      }
      else {
        if( !zkMemCalloc(err, (void **)timerP, 1, (size_t)sizeof(struct S_Timer)) )
          {
            success = ZK_FALSE;
          }
      }
      (*timerP)->allocator = allocator;
    }
  
  return success;
}


/*
** Free the Timer
*/
DLLExport int zkTimerFree
(
 ZkError     err,                     /* [in]   ZkError */                   
 ZkTimer    *timerP                   /* [out]  pointer on a ZkTimer */
 )
{
  ZkBool          success = ZK_TRUE;
  const char     *fn      = "zkTimerFree"; 
  ZkAllocator allocator = NULL;
  
  zkErrorClear(err);

  if( !timerP )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "timerP is null");
      success = ZK_FALSE;
    }
  else
    {
      allocator = (*timerP)->allocator;
      if ( allocator ) {
        if ( !zkAllocatorMemFree(err, allocator, (void**)timerP) ) {
          success = ZK_FALSE;
        }
      } else {
        if( !zkMemFree(err, (void **)timerP) )
          {
            success = ZK_FALSE;
          }
      }
    }
  return success;
}

/*
** Start the timer
*/
DLLExport int zkTimerStart
(
 ZkError     err,                     /* [in]  ZkError */
 ZkTimer     timer                    /* [in]  ZkTimer */
 )
{
  const char   *fn          = "zkTimerStart"; 
  clock_t      real_current = { 0 };
  struct tms   tms_current  = { 0 };

  /* First thing to do before any other time consuming */
  real_current = times(&tms_current);

  if (!timer )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "timer is null");
      return 0;
    }

  memset((void *)timer, 0, (size_t)sizeof(struct S_Timer));

  timer->realStartTime = real_current;
  timer->cpuStartTime  = tms_current;
  timer->isStarted = 1;
  
  return 1;
}
 

/*
** Stop the timer
*/
DLLExport int zkTimerStop
(
 ZkError     err,                     /* [in]  ZkError */
 ZkTimer     timer                    /* [in]  ZkTimer */
 )
{
  const char  *fn           = "zkTimerStop"; 
  clock_t      real_current = { 0 };
  struct tms   tms_current  = { 0 };

  /* First thing to do before any other time consuming */
  real_current = times(&tms_current);

  if (!timer )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "timer is null");
      return 0;
    }

  timer->realStopTime = real_current;
  timer->cpuStopTime  = tms_current;
  timer->isStarted = 0;
  
  return 1;
}
 

/*
** Togle pause state to On for this timer
*/
DLLExport int zkTimerPause
(
 ZkError     err,                     /* [in]  ZkError */
 ZkTimer     timer                    /* [in]  ZkTimer */
 )
{
  const  char *fn      = "zkTimerPause"; 
  clock_t      real_current = { 0 };
  struct tms   tms_current  = { 0 };

  /* First thing to do before any other time consuming */
  real_current = times(&tms_current);

  if (!timer )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "timer is null");
      return 0;
    }

  /* If not started complain */
  if ( ! timer->isStarted ) {
    zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKERROR, "timer is not started");
    return 0;
  }
  
  if (timer->pause.state )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKERROR, "timer already in pause state");
      return 0;
    }

  /* Mark current time */
  timer->pause.real_last = real_current;
  timer->pause.tms_last  = tms_current;

  /* Toggle current pause state */
  timer->pause.state     = ZK_TRUE;

  return 1;
}


/*
** Togle pause state to Off for this timer
*/
DLLExport int zkTimerResume
(
 ZkError     err,                     /* [in]  ZkError */
 ZkTimer     timer                    /* [in]  ZkTimer */
 )
{
  const  char *fn      = "zkTimerResume"; 
  clock_t      real_current = { 0 };
  struct tms   tms_current  = { 0 };

  /* First thing to do before any other time consuming */
  real_current = times(&tms_current);
  
  if (!timer )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "timer is null");
      return 0;
    }

  /* If not started, silently start it */
  if ( ! timer->isStarted ) {
    return zkTimerStart(err, timer);
  }
  
  if (!timer->pause.state )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKERROR, "timer not in pause state");
      return 0;
    }

  /* End of pause */
  timer->pause.real_cummulator           += (real_current          - timer->pause.real_last);
  timer->pause.tms_cummulator.tms_utime  += (tms_current.tms_utime - timer->pause.tms_last.tms_utime);
  timer->pause.tms_cummulator.tms_stime  += (tms_current.tms_stime - timer->pause.tms_last.tms_stime);  

  /* Mark current time */
  timer->pause.real_last = real_current;
  timer->pause.tms_last  = tms_current;

  /* Toggle current pause state */
  timer->pause.state     = ZK_FALSE;

  return 1;
}


/*
** Reset the timer to zero
*/
DLLExport int zkTimerReset
(
 ZkError     err,                     /* [in]  ZkError */
 ZkTimer     timer                    /* [in]  ZkTimer */
 )
{
  const  char *fn      = "zkTimerResume"; 
  clock_t      real_current = { 0 };
  struct tms   tms_current  = { 0 };

  /* First thing to do before any other time consuming */
  real_current = times(&tms_current);
  
  if (!timer )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "timer is null");
      return 0;
    }

  memset((void *)timer, 0, (size_t)sizeof(struct S_Timer));

  timer->realStartTime = real_current;
  timer->cpuStartTime  = tms_current;

  /* Toggle current pause state */
  timer->pause.state     = ZK_FALSE;

  return 1;
}


/*
** Return the ventilation of the elapses times of the current process during this timer lifetime
*/
DLLExport int zkTimerProcessTimesGet
(
 ZkError  err,                       /* [in]  ZkError */
 ZkTimer  timer,                     /* [in]  ZkTimer */
 double  *realTimeP,                 /* [out] return Real elapsed time (uptional) */
 double  *userTimeP,                 /* [out] return User CPU elapsed time (uptional) */
 double  *sysTimeP                   /* [out] return System CPU elapsed time (uptional) */
 )
{
  const char *fn  = "zkTimerProcessTimesGet"; 
  double realTime = 0;
  double userTime = 0;
  double sysTime  = 0;
  double realClockTick;
  double cpuClockTick;
  long   tmp;

  if (!timer )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "timer is null");
      return 0;
    }

  /* 
   * Double assignation to resolve an assignment warning on 'gcc'
   */
  tmp           = sysconf(_SC_CLK_TCK);
  realClockTick = (double) tmp;
 
  if (realClockTick > 0.0)
    {
      cpuClockTick = realClockTick;

      if( timer->pause.state == ZK_TRUE )
        {
          realTime = (double) (timer->pause.real_last          - timer->realStartTime          - timer->pause.real_cummulator)          / realClockTick;
          userTime = (double) (timer->pause.tms_last.tms_utime - timer->cpuStartTime.tms_utime - timer->pause.tms_cummulator.tms_utime) / cpuClockTick;
          sysTime  = (double) (timer->pause.tms_last.tms_stime - timer->cpuStartTime.tms_stime - timer->pause.tms_cummulator.tms_stime) / cpuClockTick;
        }
      else
        {
          realTime = (double) (timer->realStopTime          - timer->realStartTime          - timer->pause.real_cummulator)          / realClockTick;
          userTime = (double) (timer->cpuStopTime.tms_utime - timer->cpuStartTime.tms_utime - timer->pause.tms_cummulator.tms_utime) / cpuClockTick;
          sysTime  = (double) (timer->cpuStopTime.tms_stime - timer->cpuStartTime.tms_stime - timer->pause.tms_cummulator.tms_stime) / cpuClockTick;
        }

    }
  if( realTimeP )
    {
      *realTimeP = realTime;
    }
  if( userTimeP )
    {
      *userTimeP = userTime;
    }
  if( sysTimeP )
    {
      *sysTimeP = sysTime;
    }
  return 1;
}



/*
** Return the ventilation of the elapses times of the current process during this timer lifetime
*/
DLLExport int zkTimerProcessChildTimesGet
(
 ZkError  err,                       /* [in]  ZkError */
 ZkTimer  timer,                     /* [in]  ZkTimer */
 double  *realTimeP,                 /* [out] return Real elapsed time (uptional) */
 double  *userTimeP,                 /* [out] return User CPU elapsed time (uptional) */
 double  *sysTimeP                   /* [out] return System CPU elapsed time (uptional) */
 )
{
  const char     *fn = "zkTimerProcessChildTimesGet"; 
  double realTime = 0;
  double userTime = 0;
  double sysTime  = 0;
  double realClockTick;
  double cpuClockTick;
  long   tmp;

  if (!timer )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "timer is null");
      return 0;
    }

  /* 
   * Double assignation to resolve an assignment warning on 'gcc'
   */
  tmp           = sysconf(_SC_CLK_TCK);
  realClockTick = (double) tmp;
 
  if (realClockTick > 0.0)
    {
      cpuClockTick = realClockTick;
 
      realTime = (double) (timer->realStopTime           - timer->realStartTime)           / realClockTick;
      userTime = (double) (timer->cpuStopTime.tms_cutime - timer->cpuStartTime.tms_cutime) / cpuClockTick;
      sysTime  = (double) (timer->cpuStopTime.tms_cstime - timer->cpuStartTime.tms_cstime) / cpuClockTick;
    }
  if( realTimeP )
    {
      *realTimeP = realTime;
    }
  if( userTimeP )
    {
      *userTimeP = userTime;
    }
  if( sysTimeP )
    {
      *sysTimeP = sysTime;
    }
  return 1;
}


 
/*
** Return the ventilation of the elapses times of the current process during this timer pause
*/
DLLExport int zkTimerPauseTimesGet
(
 ZkError  err,                       /* [in]  ZkError */
 ZkTimer  timer,                     /* [in]  ZkTimer */
 double  *realTimeP,                 /* [out] return Real elapsed time (uptional) */
 double  *userTimeP,                 /* [out] return User CPU elapsed time (uptional) */
 double  *sysTimeP                   /* [out] return System CPU elapsed time (uptional) */
 )
{
  const char *fn  = "zkTimerProcessTimesGet"; 
  double realTime = 0;
  double userTime = 0;
  double sysTime  = 0;
  double realClockTick;
  double cpuClockTick;
  long   tmp;

  if (!timer )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "timer is null");
      return 0;
    }

  /* 
   * Double assignation to resolve an assignment warning on 'gcc'
   */
  tmp           = sysconf(_SC_CLK_TCK);
  realClockTick = (double) tmp;
 
  if (realClockTick > 0.0)
    {
      cpuClockTick = realClockTick;

      realTime = (double) (timer->pause.real_cummulator)          / realClockTick;
      userTime = (double) (timer->pause.tms_cummulator.tms_utime) / cpuClockTick;
      sysTime  = (double) (timer->pause.tms_cummulator.tms_stime) / cpuClockTick;
    }
  if( realTimeP )
    {
      *realTimeP = realTime;
    }
  if( userTimeP )
    {
      *userTimeP = userTime;
    }
  if( sysTimeP )
    {
      *sysTimeP = sysTime;
    }
  return 1;
}

 
 
#endif /* ZK_LINUX || ZK_SOLARIS */
