/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#if defined( ZK_LINUX ) || defined( ZK_SOLARIS )

/*
**  The purpose of this module is to create a virtual allocation heap over a given address space.
**  It vill partition it into sub portions and allowed memory allocation/desallocation of those sub portions.
**
**  It may be usefull for Shared Memory or Memory Maped Files.
**
**
**  TODO: Modify the read/write procedure (add marcheling) of the header 
**        to allowed talking between process compiled with different alignment.
**
**
**  Nicolas Roy <nicolas@zeroknowledge.com>
**  $Id: zkvheap.c,v 1.8.26.4 2000/12/14 17:54:29 jean Exp $
*/

#include "zks.h"
#include "zkIntTypes.h"

#include <errno.h>
#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <stdio.h>

#include "zkError.h"
#include "zkLog.h"
#include "zkvheap.h"
#include "zkAlloc.h"
#include "zkDebug.h"

#include "zksem.h"



/* **********************************************************************
** **********************************************************************
**
**
** Integrity checking
**
**
** **********************************************************************
** **********************************************************************
*/

#define VHEAP_CHECKSUM_ENABLE                   1

/*
#define VHEAP_MPROTECT_LOCKING                  1 
*/

#define VHEAP_INTEGRITY_CHECK_QUIET_MODE        ZK_FALSE


#ifdef VHEAP_MPROTECT_LOCKING
# include <sys/elf.h>
# ifndef PAGESIZE
#  define PAGESIZE      ELF_EXEC_PAGESIZE
# endif
#endif



/* **********************************************************************
** **********************************************************************
**
**
** Local Error handle
**
**
** **********************************************************************
** **********************************************************************
*/

#define ERROR_STACK_LIMIT 100

typedef struct {
  const char *circ_stack_fnc[ERROR_STACK_LIMIT];
  const char *circ_stack_msg[ERROR_STACK_LIMIT];
  uint16_t    size;
  uint16_t    limit;
} ZkVheapErrStack;


static ZkVheapErrStack ERRSTACK = { {0}, {0}, 0, ERROR_STACK_LIMIT };


static void ErrorStack( const char * func, const char * msg )
{
  if( ERRSTACK.size < ERRSTACK.limit ) 
    { 
      ERRSTACK.circ_stack_fnc[ERRSTACK.size] = func;
      ERRSTACK.circ_stack_msg[ERRSTACK.size] = msg;
      ERRSTACK.size++;      
    }
}

static void ErrorClear(void)
{
  ERRSTACK.size = 0;
}


static void ErrorDump(ZkError err, const char * func) 
{
  ZkLog log = NULL;

  if( zkErrorLog(err, &log) )
    {
      uint16_t size = ERRSTACK.size;
      uint16_t idx;

      for( idx=0; idx<size; idx++ )
        {
          zkLogWrite_va(log, "[VHEAP] %s:: %s: %s", func, ERRSTACK.circ_stack_fnc[idx], ERRSTACK.circ_stack_msg[idx]);
        }
    }
  ERRSTACK.size = 0;
}


/* **********************************************************************
** **********************************************************************
**
**
** Debug calls
**
**
** **********************************************************************
** **********************************************************************
*/

/*
#define XDEBUG                                  1 
#define XDEBUG_INTEGRITY_CHECK_FREQUENCY        10000 
#define XDEBUG_INTEGRITY_CHECK_QUIET_MODE       ZK_TRUE 
*/


#ifdef XDEBUG 
static volatile uint32_t freq = 0;

static int PeriodicalIntegrityCheck(void)
{
  int ret = 1;
  
  if( !(freq++ % XDEBUG_INTEGRITY_CHECK_FREQUENCY) )
    {
      ret = checkIntegrity(err, vheap->hdr, XDEBUG_INTEGRITY_CHECK_QUIET_MODE);
    }
  return ret;
}

# define PERIODICAL_INTEGRITY_CHECK()           PeriodicalIntegrityCheck()
#else
# define PERIODICAL_INTEGRITY_CHECK()
#endif


/* **********************************************************************
** **********************************************************************
**
**
** MACRO 
**
**
** **********************************************************************
** **********************************************************************
*/

/*
 *  CRC 0102010
 */
static unsigned short crc_table[256] = {
  0x0000, 0x1189, 0x2312, 0x329b, 0x4624, 0x57ad, 0x6536, 0x74bf,
  0x8c48, 0x9dc1, 0xaf5a, 0xbed3, 0xca6c, 0xdbe5, 0xe97e, 0xf8f7,
  0x1081, 0x0108, 0x3393, 0x221a, 0x56a5, 0x472c, 0x75b7, 0x643e,
  0x9cc9, 0x8d40, 0xbfdb, 0xae52, 0xdaed, 0xcb64, 0xf9ff, 0xe876,
  0x2102, 0x308b, 0x0210, 0x1399, 0x6726, 0x76af, 0x4434, 0x55bd,
  0xad4a, 0xbcc3, 0x8e58, 0x9fd1, 0xeb6e, 0xfae7, 0xc87c, 0xd9f5,
  0x3183, 0x200a, 0x1291, 0x0318, 0x77a7, 0x662e, 0x54b5, 0x453c,
  0xbdcb, 0xac42, 0x9ed9, 0x8f50, 0xfbef, 0xea66, 0xd8fd, 0xc974,
  0x4204, 0x538d, 0x6116, 0x709f, 0x0420, 0x15a9, 0x2732, 0x36bb,
  0xce4c, 0xdfc5, 0xed5e, 0xfcd7, 0x8868, 0x99e1, 0xab7a, 0xbaf3,
  0x5285, 0x430c, 0x7197, 0x601e, 0x14a1, 0x0528, 0x37b3, 0x263a,
  0xdecd, 0xcf44, 0xfddf, 0xec56, 0x98e9, 0x8960, 0xbbfb, 0xaa72,
  0x6306, 0x728f, 0x4014, 0x519d, 0x2522, 0x34ab, 0x0630, 0x17b9,
  0xef4e, 0xfec7, 0xcc5c, 0xddd5, 0xa96a, 0xb8e3, 0x8a78, 0x9bf1,
  0x7387, 0x620e, 0x5095, 0x411c, 0x35a3, 0x242a, 0x16b1, 0x0738,
  0xffcf, 0xee46, 0xdcdd, 0xcd54, 0xb9eb, 0xa862, 0x9af9, 0x8b70,
  0x8408, 0x9581, 0xa71a, 0xb693, 0xc22c, 0xd3a5, 0xe13e, 0xf0b7,
  0x0840, 0x19c9, 0x2b52, 0x3adb, 0x4e64, 0x5fed, 0x6d76, 0x7cff,
  0x9489, 0x8500, 0xb79b, 0xa612, 0xd2ad, 0xc324, 0xf1bf, 0xe036,
  0x18c1, 0x0948, 0x3bd3, 0x2a5a, 0x5ee5, 0x4f6c, 0x7df7, 0x6c7e,
  0xa50a, 0xb483, 0x8618, 0x9791, 0xe32e, 0xf2a7, 0xc03c, 0xd1b5,
  0x2942, 0x38cb, 0x0a50, 0x1bd9, 0x6f66, 0x7eef, 0x4c74, 0x5dfd,
  0xb58b, 0xa402, 0x9699, 0x8710, 0xf3af, 0xe226, 0xd0bd, 0xc134,
  0x39c3, 0x284a, 0x1ad1, 0x0b58, 0x7fe7, 0x6e6e, 0x5cf5, 0x4d7c,
  0xc60c, 0xd785, 0xe51e, 0xf497, 0x8028, 0x91a1, 0xa33a, 0xb2b3,
  0x4a44, 0x5bcd, 0x6956, 0x78df, 0x0c60, 0x1de9, 0x2f72, 0x3efb,
  0xd68d, 0xc704, 0xf59f, 0xe416, 0x90a9, 0x8120, 0xb3bb, 0xa232,
  0x5ac5, 0x4b4c, 0x79d7, 0x685e, 0x1ce1, 0x0d68, 0x3ff3, 0x2e7a,
  0xe70e, 0xf687, 0xc41c, 0xd595, 0xa12a, 0xb0a3, 0x8238, 0x93b1,
  0x6b46, 0x7acf, 0x4854, 0x59dd, 0x2d62, 0x3ceb, 0x0e70, 0x1ff9,
  0xf78f, 0xe606, 0xd49d, 0xc514, 0xb1ab, 0xa022, 0x92b9, 0x8330,
  0x7bc7, 0x6a4e, 0x58d5, 0x495c, 0x3de3, 0x2c6a, 0x1ef1, 0x0f78
};


static unsigned short cksum(unsigned char *s, int n)
{
  unsigned int crc;

  for (crc = 0; n > 0; n--) {
    crc = crc_table[(crc ^ *s++) & 0xff] ^ (crc >> 8);
  }

  return crc;
}



/* **********************************************************************
** **********************************************************************
**
**
** MACRO 
**
**
** **********************************************************************
** **********************************************************************
*/

/*
** Granularity of the heap (SHOULD NEVER BE SMALLER THAN 16 BYTES)
*/
#ifdef  VHEAP_MPROTECT_LOCKING
# define ZK_VHEAP_GRANULARITY_DEFAULT                      PAGESIZE
#else
# define ZK_VHEAP_GRANULARITY_DEFAULT                      16
#endif

/*
** Return the granularity of a given vheap
*/
#define GRANULARITY(vheapP)                                (uint32_t)((*vheapP).heapGranularity)

/*
** Calculate the offset (in bytes) of a given position (in blocks)
*/
#define BLOCK2BYTE(vheapP, pos)                            (size_t)((pos) * GRANULARITY(vheapP))

/*
** Calculate the offset (in blocks) of a given position (in bytes)
*/
#define BYTE2BLOCK(vheapP, size)                           (uint32_t)( ((uint32_t)(size) / GRANULARITY(vheapP)) + (((uint32_t)(size) % GRANULARITY(vheapP)) ? 1 : 0) )

/*
** Calculate the address of a given position (in blocks)
*/
#define BLOCK2ADDRESS(vheapP, pos)                         (void *)( (ptrdiff_t)(*vheapP).heapAddress + (ptrdiff_t)BLOCK2BYTE(vheapP,pos) )

/*
** Calculate the block position of a given address (in bytes)
*/
#define ADDRESS2BLOCK(vheapP, addr)                        (uint32_t)BYTE2BLOCK(vheapP, (ptrdiff_t) (addr) - (ptrdiff_t)(*vheapP).heapAddress )

/*
** Calculate the address of a given position (in blocks)
*/
#define IS_A_BLOCK(vheapP, addr)                           ((((ptrdiff_t) (addr) - (ptrdiff_t)(*vheapP).heapAddress) % GRANULARITY(vheapP)) ? 0 : 1)

/*
** Free Table List offsets
*/ 
static const uint32_t freeTableChunkSize[] = {0x00000001UL, 0x00000002UL, 0x00000004UL, 0x00000008UL,  
                                              0x00000010UL, 0x00000020UL, 0x00000040UL, 0x00000080UL,  
                                              0x00000100UL, 0x00000200UL, 0x00000400UL, 0x00000800UL,  
                                              0x00001000UL, 0x00002000UL, 0x00004000UL, 0x00008000UL,  
                                              0xffffffff};

/*
** Length of the free chunks table
*/
#define ZK_VHEAP_FREETABLE_SIZE                            (size_t)(sizeof(freeTableChunkSize) / sizeof(freeTableChunkSize[0]))

/*
** Return the size (in bytes) of a bitmap (length in bits)
*/
#define BIT2BYTE(size)                                     (size_t)(((size)>>3) + (((size) & 3) ? 1 : 0))

#define BIT2BLOCK(vheapP, size)                            (uint32_t)BYTE2BLOCK(vheapP, BIT2BYTE(size))


/*
** Memory protection 
*/
# ifdef VHEAP_MPROTECT_LOCKING
#  define MPROTECT(addr, size, prot)                       mprotect(addr, size, prot)
# else
#  define MPROTECT(addr, size, prot)
# endif

# define LOCK_RO_BLOCK(vheapP, pos, size)                  MPROTECT(BLOCK2ADDRESS(vheapP, pos), BLOCK2BYTE(vheapP, size), PROT_READ)
# define LOCK_WO_BLOCK(vheapP, pos, size)                  MPROTECT(BLOCK2ADDRESS(vheapP, pos), BLOCK2BYTE(vheapP, size), PROT_WRITE)
# define LOCK_RW_BLOCK(vheapP, pos, size)                  MPROTECT(BLOCK2ADDRESS(vheapP, pos), BLOCK2BYTE(vheapP, size), PROT_NONE)
# define UNLOCK_BLOCK(vheapP, pos, size)                   MPROTECT(BLOCK2ADDRESS(vheapP, pos), BLOCK2BYTE(vheapP, size), PROT_READ | PROT_WRITE)



/* **********************************************************************
** **********************************************************************
**
**
** STRUCTURES
**
**
** **********************************************************************
** **********************************************************************
*/


/*
** Heap Header Handle
*/
struct  S_ZkVHeapHdr {
  /*
  ** heap specific attributs
  */
  void     *heapAddress;                                   /* Address of memory space */
  size_t    heapGranularity;                               /* Size (in bytes) of the blocks (Granularity of the memory) */
  size_t    heapByteSize;                                  /* Maximum space (in bytes) of memory */
  uint32_t  heapBlockSize;                                 /* Maximum space (in blocks) of memory */
  uint32_t  heapFreeSpace;                                 /* Free space (in blocks) of memory */
  uint32_t  heapFlags;                                     /* Internal flags */
  /*
  ** allocated blocks's bitmap specific attributs
  */
  uint32_t  allocatedBmBitSize;                            /* Size (in bits) of the allocated's bitmap */
  uint32_t  allocatedBmBegOffset;                          /* Offset of the allocated's begin bitmap (in blocks) */
  uint32_t  allocatedBmEndOffset;                          /* Offset of the allocated's end bitmap (in blocks) */
  /*
  ** free chunk's specific attributs
  */
  uint32_t  freeTable[ZK_VHEAP_FREETABLE_SIZE];            /* Free Chunk's Table */
  /*
  ** other attributs
  */
  uint32_t  firstBlockOffset;                              /* Offset of the first block (in blocks) */
  /*
  ** user define hook
  */
  char     *hook;                                          /* extra data hook (user defined) */
};

/*
** ZkVHeapHdr handle
*/
typedef struct S_ZkVHeapHdr * ZkVHeapHdr;


/*
** Heap Handle
*/
struct S_ZkVHeap {
  ZkVHeapHdr hdr;                                          /* pointer to the virtual heap header */
  ZkSem      mutex;                                        /* Mutex use to synchronyze critical section */
};



/*
** Return begin address of bitmaps
*/
#define BEGIN_BITMAP_ADDR(vheapP)                          BLOCK2ADDRESS(vheapP, (*vheapP).allocatedBmBegOffset)
#define END_BITMAP_ADDR(vheapP)                            BLOCK2ADDRESS(vheapP, (*vheapP).allocatedBmEndOffset)


/*
** Return size (in blocks) of bitmaps
*/
#define BITMAP_BIT_SIZE(vheapP)                          (uint32_t)((*vheapP).allocatedBmBitSize)


/*
** Wipe size blocks from position 'pos' (in blocks)
*/
#define WIPE_BLOCKS(vheapP, pos, size)                   memset(BLOCK2ADDRESS(vheapP, pos+1), 0x00, BLOCK2BYTE(vheapP, size-1))


/*
** Copy size blocks from position 'src' (in blocks) to position 'dst' (in blocks)
*/
#define COPY_BLOCKS(vheapP, dst, src, size)              memcpy(BLOCK2ADDRESS(vheapP, dst), BLOCK2ADDRESS(vheapP, src), BLOCK2BYTE(vheapP, size))


/*
** Advance locking
*/
#define LOCK_VHEAP_HEADER(vheapP)                        LOCK_RO_BLOCK(vheapP, 0, (*vheapP).firstBlockOffset)
#define UNLOCK_VHEAP_HEADER(vheapP)                      UNLOCK_BLOCK(vheapP,  0, (*vheapP).firstBlockOffset)

#define LOCK_RO_FREE_CHUNCK(vheapP, pos, size)           LOCK_RO_BLOCK(vheapP,   pos, size)
#define LOCK_RW_FREE_CHUNCK(vheapP, pos, size)           LOCK_RW_BLOCK(vheapP,   pos, size)
#define UNLOCK_FREE_CHUNCK(vheapP, pos, size)            UNLOCK_BLOCK(vheapP,    pos, size)

#define LOCK_RO_ALLOC_CHUNCK(vheapP, pos, size)          LOCK_RO_BLOCK(vheapP,   pos, size)
#define LOCK_RW_ALLOC_CHUNCK(vheapP, pos, size)          LOCK_RW_BLOCK(vheapP,   pos, size)
#define UNLOCK_ALLOC_CHUNCK(vheapP, pos, size)           UNLOCK_BLOCK(vheapP,    pos, size)


/* **********************************************************************
** **********************************************************************
**
**
** Local calls
**
**
** **********************************************************************
** **********************************************************************
*/

static int checkIntegrity(ZkError err, ZkVHeapHdr vheap, ZkBool quiet);



/* **********************************************************************
** **********************************************************************
**
**
** BITMAP MANIPULATION
**
**
** **********************************************************************
** **********************************************************************
*/

static const uint32_t bitIdx[32] = {
    0x00000001UL, 0x00000002UL, 0x00000004UL, 0x00000008UL,
    0x00000010UL, 0x00000020UL, 0x00000040UL, 0x00000080UL,
    0x00000100UL, 0x00000200UL, 0x00000400UL, 0x00000800UL,
    0x00001000UL, 0x00002000UL, 0x00004000UL, 0x00008000UL,
    0x00010000UL, 0x00020000UL, 0x00040000UL, 0x00080000UL,
    0x00100000UL, 0x00200000UL, 0x00400000UL, 0x00800000UL,
    0x01000000UL, 0x02000000UL, 0x04000000UL, 0x08000000UL,
    0x10000000UL, 0x20000000UL, 0x40000000UL, 0x80000000UL
  };


typedef uint32_t bitmap_t[];




/*
** Set the state of a bit
*/
static void setBit
  (
    bitmap_t bm,   /* [in] bitmap */
    uint32_t pos,  /* [in] bit position */
    uint8_t  state /* [in] On(1) | Off(0) */
  )
{
  uint32_t idx = pos >> 5; 
  uint32_t bit = pos & 0x1f;
  
  if( state )
    {
      bm[idx] |= bitIdx[bit];
    }
  else
    {
      bm[idx] &= ~bitIdx[bit];
    }
}




/*
** Return the state of a bit
*/
static int getBit
  (
    bitmap_t bm, /* [in] bitmap */
    uint32_t pos /* [in] bit position */
  )
{
  uint32_t idx = pos >> 0x05; 
  uint32_t bit = pos &  0x1f;
  
  return bm[idx] & bitIdx[bit];
}




/*
** Foward scan a bitmap to find the next bit ON
** - return '1' if a bit has been found, '0' otherwise
*/
static int findBit
  (
    bitmap_t  bm,       /* [in]     bitmap */
    uint32_t *findposP, /* [in/out] found position */
    uint32_t  fromPos,  /* [in]     Starting position (included) */
    uint32_t  stopPos   /* [in]     Stopping position (included) */
  )
{
  const char * func = "findBit";
  uint32_t pos = fromPos;

  while( pos <= stopPos ) 
    {
      if( bm[pos>>5] ) 
        {
          do
            {
              if( getBit(bm, pos) ) 
                {
                  if( findposP )
                    {
                      *findposP = pos;
                    }
                  return 1;
                }
              pos++;
              if( pos > stopPos )
                {
                  ErrorStack(func, "pos > stopPos");
                  return 0;
                }
            }
          while( pos & 0x1f );
        }          
      while( !bm[pos>>5] && pos <= stopPos ) 
        {
          pos = ((pos>>5)<<5) + 0x20;   
        }
    }
  ErrorStack(func, "failed");
  return 0;
}




/*
** Backward scan a bitmap to find the previous bit ON
** - return '1' if a bit has been found, '0' otherwise
*/
static int rfindBit
  (
    bitmap_t  bm,       /* [in]     bitmap */
    uint32_t *findposP, /* [in/out] found position */
    uint32_t  fromPos,  /* [in]     Starting position (included) */
    uint32_t  stopPos   /* [in]     Stopping position (included) */
  )
{
  const char * func = "rfindBit";
  uint32_t pos = fromPos;
  uint32_t idx;

  while( (pos != (uint32_t) -1) && (pos >= stopPos) )
    {
      while( ((pos>>5) > 1) && !bm[pos>>5] && (pos >= stopPos) )
        {
          pos = (((pos>>5)<<5) - 1);
        }
      if( pos >= stopPos )
        {
          idx = pos>>5;
          if( bm[idx] ) 
            {
              while( ((pos>>5) == idx) && (pos >= stopPos) )
                {
                  if( getBit(bm, pos) ) 
                    {
                      if( findposP )
                        {
                          *findposP = pos;
                        }
                      return 1;
                    }
                  pos--;
                }
            }
          else
            {
              pos--;
            }
        }
    } 
  ErrorStack(func, "failed");
  return 0;
}





/* **********************************************************************
** **********************************************************************
**
**
** ALLOCATE CHUNK CALLS
**
**
** **********************************************************************
** **********************************************************************
*/

#define ALLOC_CHUNK_NONE_FLAGS                             0x00  /* No flags */

typedef struct S_AllocChunk {
  uint32_t size;                                           /* size (in blocks) of the chunk */
  uint32_t flag;                                           /* internal flag */
  uint32_t beg;                                            /* offset (in block) of the beginning block of the chunk */
  uint32_t end;                                            /* offset (in block) of the ending block of the chunk (redondant info) */
}AllocChunk;





/*
** Foward scan to find the next allocated chunk
** - return '1' if found, '0' otherwise
*/
static int findAllocChunk
 (
    ZkVHeapHdr  vheap,         /* [in]     ZkVHeapHdr */
    AllocChunk *foundP,        /* [in/out] pointer on a found AllocChunk */
    uint32_t    from           /* [in]     Starting position (in blocks) (included) */
 )
{
  const char * func = "findAllocChunk";
  AllocChunk found  = { 0 };
  bitmap_t  *begbmP = (bitmap_t *) BEGIN_BITMAP_ADDR(vheap);
  bitmap_t  *endbmP = (bitmap_t *) END_BITMAP_ADDR(vheap);
  uint32_t   bmsize = (uint32_t)   BITMAP_BIT_SIZE(vheap) - 1;

  if( findBit(*begbmP, &found.beg, from, bmsize) && findBit(*endbmP, &found.end, found.beg, bmsize) )
    {
      found.flag = ALLOC_CHUNK_NONE_FLAGS;
      found.size = found.end - found.beg + 1; /* Add one because hedge are include inside of the section */
      if( foundP )
        {
          *foundP = found;
        }
      return 1;
    }
  ErrorStack(func, "failed");
  return 0;
}




/*
** Backward scan to find the previous allocated chunk
** - return '1' if found, '0' otherwise
*/
static int rfindAllocChunk
 (
    ZkVHeapHdr  vheap,         /* [in]     ZkVHeapHdr */
    AllocChunk *foundP,        /* [in/out] pointer on a found AllocChunk */
    uint32_t    from           /* [in]     Starting position (in blocks) (included) */
 )
{
  const char * func = "rfindAllocChunk";
  AllocChunk found  = { 0 };
  bitmap_t  *begbmP = (bitmap_t *) BEGIN_BITMAP_ADDR(vheap);
  bitmap_t  *endbmP = (bitmap_t *) END_BITMAP_ADDR(vheap);
  uint32_t   bmsize = (uint32_t)   BITMAP_BIT_SIZE(vheap) - 1;

  if( rfindBit(*begbmP, &found.beg, from, 0) && findBit(*endbmP, &found.end, found.beg, bmsize) )
    {
      found.flag = ALLOC_CHUNK_NONE_FLAGS;
      found.size = found.end - found.beg + 1; /* Add one because hedge are include inside of the section */
      if( foundP )
        {
          *foundP = found;
        }
      return 1;
    }
  ErrorStack(func, "failed");
  return 0;
}




/*
** Fill a AllocChunk structure
*/
static void fillAllocChunk
  (
    AllocChunk *chunkP,        /* [in/out] pointer on a AllocChunk */
    uint32_t    beg,           /* offset (in block) of the beginning block of the chunk */
    uint32_t    size          /* size (in blocks) of the chunk */
  )
{
  chunkP->size = size;
  chunkP->flag = ALLOC_CHUNK_NONE_FLAGS;
  chunkP->beg  = beg;
  chunkP->end = ( size ) ? beg + size - 1 : beg;           /* Sub one because hedge are include inside of the section */  
}



/*
** Look if a given chunk exist 
** - return '1' if it exist, '0' otherwise
*/
static int validateAllocChunk
  (
    ZkVHeapHdr  vheap,         /* [in] ZkVHeapHdr */
    AllocChunk *chunkP         /* [in] pointer on a chunk */
  )
{
  const char * func = "validateAllocChunk";
  AllocChunk found;

  if( !findAllocChunk(vheap, &found, chunkP->beg) || 
      (found.beg != chunkP->beg) || 
      (found.end != chunkP->end) )
    {
      ErrorStack(func, "failed");
      return 0;
    }
  return 1;
}




/*
** Allocate a new chunk if chunk space is not already use
** - return '1' on success, '0' otherwise
*/
static int addAllocChunk
  (
    ZkVHeapHdr  vheap,         /* [in]     ZkVHeapHdr */
    AllocChunk *chunkP,        /* [in/out] pointer on a new AllocChunk */
    uint32_t    begin,         /* [in]     beginning block */
    uint32_t    size           /* [in]     size (in block) of the chunk */
  )
{
  bitmap_t  *begbmP  = (bitmap_t *) BEGIN_BITMAP_ADDR(vheap);
  bitmap_t  *endbmP  = (bitmap_t *) END_BITMAP_ADDR(vheap);
  AllocChunk chunk  = { 0 };
           
  fillAllocChunk(&chunk, begin, size);

  /* Set bitmaps */
  setBit(*begbmP, chunk.beg, 1);
  setBit(*endbmP, chunk.end, 1);
  if( chunkP )
    {
      *chunkP = chunk;
    }
  return 1;
}




/*
** Desallocate a chunk if it's a valid one
** - return '1' on success, '0' otherwise
*/
static int remAllocChunk
  (
    ZkVHeapHdr  vheap,         /* [in] ZkVHeapHdr */
    AllocChunk *chunkP         /* [in] pointer on a chunk  */
  )
{
  const char * func = "remAllocChunk";
  bitmap_t  *begbmP  = (bitmap_t *) BEGIN_BITMAP_ADDR(vheap);
  bitmap_t  *endbmP  = (bitmap_t *) END_BITMAP_ADDR(vheap);

  /* Look if this chunk exist */
  if( !validateAllocChunk(vheap, chunkP) )
    {
      ErrorStack(func, "validateAllocChunk failed");
      return 0;
    }
  /* Unset bitmaps */
  setBit(*begbmP, chunkP->beg, 0);
  setBit(*endbmP, chunkP->end, 0);
  return 1;
}




/*
** Return the chunk allocate at a given position
** - return '1' on success, '0' otherwise
*/
static int getAllocChunk
  (
    ZkVHeapHdr  vheap,         /* [in]     ZkVHeapHdr */
    AllocChunk *chunkP,        /* [in/out] pointer on a new AllocChunk */
    uint32_t    pos            /* [in]     postion (in blocks) */
  )
{
  const char * func = "getAllocChunk";
  AllocChunk  chunk;

  /* Look if a chunk at that position */
  if( !findAllocChunk(vheap, &chunk, pos) || (chunk.beg != pos) )
    {
      ErrorStack(func, "findAllocChunk failed");
      return 0;
    }
  if( chunkP )
    {
      *chunkP = chunk;
    }
  return 1;
}





/* **********************************************************************
** **********************************************************************
**
**
** FREE CHUNK CALLS
**
**
** **********************************************************************
** **********************************************************************
*/


/*
** Free chunk handle
*/
typedef struct S_FreeChunk {
  uint32_t pos;                                            /* offset (in block) of this chunk */
  uint32_t size;                                           /* size (in blocks) of the chunk */
  uint32_t chksum;                                         /* internal chksum */
  uint32_t next;                                           /* offset (in block) of the next chunk (0 at the beginning of the list) */
  uint32_t prev;                                           /* offset (in block) of the previous chunk (0 at the end of the list) */
}FreeChunk;



static uint8_t chksumFreeChunk
(
 uint32_t    pos,           /* offset (in block) of the beginning block of the chunk */
 uint32_t    size,          /* size (in blocks) of the chunk */
 uint32_t    next,          /* position (in blocks) of the next chunk */
 uint32_t    prev           /* position (in blocks) of the previous chunk */
)
{
  uint32_t chksum  = (uint32_t)74371;

  (void) pos;
#ifdef FREEDOM_DEVELOPMENT_BUILD
# if defined( VHEAP_CHECKSUM_ENABLE ) && !defined( VHEAP_MPROTECT_LOCKING )
  chksum += cksum((unsigned char *)&size, 4);
  chksum += cksum((unsigned char *)&next, 4);
  chksum += cksum((unsigned char *)&prev, 4);
# else
  (void) size;
  (void) next;
  (void) prev;  
# endif
#endif
  return chksum;
}


/*
** Fill a FreeChunk structure
*/
static void fillFreeChunk
  (
    FreeChunk  *chunkP,        /* [in/out] pointer on a FreeChunk */
    uint32_t    pos,           /* offset (in block) of the beginning block of the chunk */
    uint32_t    size,          /* size (in blocks) of the chunk */
    uint32_t    next,          /* position (in blocks) of the next chunk */
    uint32_t    prev           /* position (in blocks) of the previous chunk */
  )
{
  chunkP->pos    = pos;
  chunkP->size   = size;
  chunkP->chksum = chksumFreeChunk(pos, size, next, prev);
  chunkP->next   = next;
  chunkP->prev   = prev;
}




/*
** Write the FreeChunk header into the chunk at position 'pos'
*/
static int writeFreeChunk
  (
    ZkVHeapHdr vheap,         /* [in]     ZkVHeapHdr */
    FreeChunk *fchunkP        /* [in]     pointer on a FreeChunk */
  )
{
  const char * func = "writeFreeChunk";
  uint8_t * addr;

  if( (!fchunkP->pos) || (fchunkP->pos >= vheap->heapBlockSize) )
    {
      ErrorStack(func, "1");       
      return 0;
    }

  addr = (uint8_t *)BLOCK2ADDRESS(vheap, fchunkP->pos);

  /* Add integrity sentinel to the free chumck */
  fchunkP->chksum = chksumFreeChunk(fchunkP->pos, fchunkP->size, fchunkP->next, fchunkP->prev);

  UNLOCK_FREE_CHUNCK(vheap, fchunkP->pos, 1);
  memcpy(&addr[0],  &fchunkP->size,   4);
  memcpy(&addr[4],  &fchunkP->chksum, 4);
  memcpy(&addr[8],  &fchunkP->next,   4);
  memcpy(&addr[12], &fchunkP->prev,   4);
  LOCK_RW_FREE_CHUNCK(vheap, fchunkP->pos, 1);
  return 1;
}


/*
** Read the FreeChunk header from the chunk at position 'pos'
*/
static int readFreeChunk
  (
    ZkVHeapHdr       vheap,         /* [in]     ZkVHeapHdr */
    FreeChunk       *fchunkP,       /* [in/out] pointer on a FreeChunk */
    uint32_t         pos,           /* [in]     postion (in blocks) of the free chunk */
    ZkBool          *badChksumP     /* [out]    chksum failure */
  )
{
  const char * func = "readFreeChunk";
  uint8_t * addr;
  uint8_t   chksum = 0;

  *badChksumP = ZK_FALSE;

  if( (!pos) || (pos >= vheap->heapBlockSize) )
    {
      ErrorStack(func, "1");       
      return 0;
    }

  fchunkP->pos = pos;
  addr         = (uint8_t *)BLOCK2ADDRESS(vheap, pos);  

  LOCK_RO_FREE_CHUNCK(vheap, fchunkP->pos, 1);
  memcpy(&fchunkP->size,   &addr[0],  4);
  memcpy(&fchunkP->chksum, &addr[4],  4);
  memcpy(&fchunkP->next,   &addr[8],  4);
  memcpy(&fchunkP->prev,   &addr[12], 4);
  LOCK_RW_FREE_CHUNCK(vheap, fchunkP->pos, 1);

  /* verify the integrity of the free chumck */
  chksum = chksumFreeChunk(fchunkP->pos, fchunkP->size, fchunkP->next, fchunkP->prev);
  if( chksum != fchunkP->chksum )
    {
      *badChksumP = ZK_TRUE;
      return 0;
    }
  return 1;
}




/*
** Wipe the memory occupied by a given chunk
*/
static int wipeFreeChunk
  (
    ZkVHeapHdr       vheap,         /* [in]     ZkVHeapHdr */
    FreeChunk       *fchunkP        /* [in/out] pointer on a FreeChunk */
  )
{  
  WIPE_BLOCKS(vheap, fchunkP->pos, fchunkP->size);
  return 1;
}




/*
** Find the minimal index of a given chunk size into the freeTable
**
** REM: this implementation should be replace by a binary search.
*/
static uint16_t freeChunkIndex
  (
    uint32_t size                   /* [in]     size */
  )
{
  /* do not use binary searh for small chunk */  
  if( size <= freeTableChunkSize[6] )
    {
      uint16_t idx = 0;

      while( (freeTableChunkSize[idx] < size) && (idx < ZK_VHEAP_FREETABLE_SIZE) )
        {
         idx++;
        }  
      return (idx == ZK_VHEAP_FREETABLE_SIZE) ? ZK_VHEAP_FREETABLE_SIZE - 1 : idx;
    }
  else
    {
      uint16_t pmin = 0;
      uint16_t pmax = ZK_VHEAP_FREETABLE_SIZE-1;
      uint16_t pmid = (pmin+pmax)>>1;
  

      do
        {
          if( freeTableChunkSize[pmid] > size )
            {
              if( !pmid || (freeTableChunkSize[pmid-1] <= size) )
                {
                  break;
                }
              else
                {
                  pmax = pmid;
                }
            }
          else if( freeTableChunkSize[pmid] < size )
            {
              if( pmid == pmin )
                {
                  pmid++;
                  break;
                }
              else
                {
                  pmin = pmid;
                }
            }
          else
            {
              break;
            }
          pmid = (pmin+pmax)>>1;      
        }
      while( 1 );
      return (pmid == ZK_VHEAP_FREETABLE_SIZE) ? ZK_VHEAP_FREETABLE_SIZE - 1 : pmid;
  }
}




/*
** Find a free chunk for a given size
** - return '1' on success, '0' otherwise
*/
static int findFreeChunk
  (
    ZkVHeapHdr       vheap,         /* [in]     ZkVHeapHdr */
    FreeChunk       *fchunkP,       /* [in/out] pointer on a FreeChunk */
    uint32_t         size           /* [in]     desirer size */
  )
{
  const char * func = "findFreeChunk";
  FreeChunk next;
  uint16_t  idx;
  uint32_t  pos = 0;
  ZkBool    badchksum = ZK_FALSE;

  /* find minimal index */
  idx = freeChunkIndex(size);  
  /* find effective index (this index may not point on a chunk list) */
  while( (idx < ZK_VHEAP_FREETABLE_SIZE) && (!(pos = vheap->freeTable[idx])) )
    {
      idx++;
    }
  if( idx >= ZK_VHEAP_FREETABLE_SIZE )
    {
      ErrorStack(func, "idx >= ZK_VHEAP_FREETABLE_SIZE");
      return 0;
    }
  /* if no more block of this size or bigger size are available */
  if( !pos )
    {
      ErrorStack(func, "!pos <<< 1");
      return 0;
    }
  /* travel across the chain list to find a block large enough */
  while( pos && readFreeChunk(vheap, &next, pos, &badchksum) )
    {
      if( next.size >= size)
        {
          pos = next.pos;
          break;
        }
      else 
        {
          pos = next.next;
        } 
    }
  if( badchksum )
    {
      ErrorStack(func, "bad chksum");
      return 0;      
    }
  /* Check the next slot if it doesn't success with this one */
  if( !pos )
    {
      do
        {
          idx++;
        }
      while( (idx < ZK_VHEAP_FREETABLE_SIZE) && (!(pos = vheap->freeTable[idx])) );
      if( idx == ZK_VHEAP_FREETABLE_SIZE )
        {
          ErrorStack(func, "idx == ZK_VHEAP_FREETABLE_SIZE");
          return 0;
        }
      /* if no more block of this size or bigger size are available */
      if( !pos )
        {
          ErrorStack(func, "!pos <<< 2");
          return 0;
        }
      /* travel across the chain list to find a block large enough */
      while( pos && readFreeChunk(vheap, &next, pos, &badchksum) )
        {
          if( next.size >= size)
            {
              pos = next.pos;
              break;
            }
          else 
            {
              pos = next.next;
            } 
        }
      if( badchksum )
        {
          ErrorStack(func, "bad chksum");
          return 0;      
        }
    }
  /* if no more block of this size or bigger size are available */
  if( !pos )
    {
      /*findFreeChunk(vheap, fchunkP, size); THIS CODE should be remove ??? */
      ErrorStack(func, "!pos <<< 3");
      return 0;
    }
  if( fchunkP )
    {
      *fchunkP = next;
    }
  return 1;
}




/*
** Get the free chunk at position 'pos'
** - return '1' if this chunk exist, '0' otherwise
*/
static int getFreeChunk
  (
    ZkVHeapHdr       vheap,         /* [in]     ZkVHeapHdr */
    FreeChunk       *fchunkP,       /* [in/out] pointer on a FreeChunk */
    uint32_t         pos            /* [in]     postion (in blocks) of the free chunk */
  )
{
  const char * func = "getFreeChunk";
  FreeChunk  cur;
  uint32_t   cur_pos;
  uint8_t    found = 0;
  uint16_t   idx;
  ZkBool     badchksum = ZK_FALSE;


  /*
  ** Test boundary conditions
  */
  if( pos >= vheap->heapBlockSize )
    {
      ErrorStack(func, "pos >= vheap->heapBlockSize");
      return 0;
    }
  /*
  ** Read this address
  */
  if( !readFreeChunk(vheap, &cur, pos, &badchksum) || !cur.size )
    {
      if( badchksum )
        {
          ErrorStack(func, "readFreeChunk: bad chksum");
        }
      else
        {
          ErrorStack(func, "readFreeChunk");
        }
      return 0;
    }
  /* 
  ** find the index on the indirection table 
  */
  idx = freeChunkIndex(cur.size);  
  /* 
  ** scan the list to found if it is really a chained block.
  */
  cur_pos  = vheap->freeTable[idx];
  while( cur_pos && !found )
    {
      /* read next block info */
      if( !readFreeChunk(vheap, &cur, cur_pos, &badchksum) )
        {
          if( badchksum )
            {
              ErrorStack(func, "readFreeChunk: bad chksum");
            }
          else
            {
              ErrorStack(func, "readFreeChunk");
            }
          return 0;
        }
      if( cur.pos == pos )
        {
          found = 1;
        }
      else
        {
          cur_pos = cur.next;
        }
    }
  if( !found )
    {
      ErrorStack(func, "!found");
      return 0;
    }
  if( fchunkP )
    {
      *fchunkP = cur;
    }
  return 1;    
}




/*
** Add a free chunk to the FreeChunk table
** - return '1' on success, '0' otherwise
*/
static int addFreeChunk
  (
    ZkVHeapHdr       vheap,         /* [in]     ZkVHeapHdr */
    FreeChunk       *fchunkP,       /* [in/out] pointer on the new FreeChunk */
    uint32_t         pos,           /* [in]     postion (in blocks) of the free chunk */
    uint32_t         size           /* [in]     size */
  )
{
  const char * func = "addFreeChunk";
  FreeChunk cur; 
  FreeChunk toadd = { 0 };
  uint16_t  idx;
  uint32_t  cur_pos;
  uint8_t   done = 0;
  ZkBool    badchksum = ZK_FALSE;

  /*
  ** Test boundary conditions
  */
  if( (pos + size) > vheap->heapBlockSize )
    {
      ErrorStack(func, "(pos + size) > vheap->heapBlockSize");
      return 0;
    }
  /*
  ** Initialyze the new free chunk 
  */
  toadd.pos    = pos;
  toadd.size   = size;
  toadd.prev   = 0;
  toadd.next   = 0;
  toadd.chksum = chksumFreeChunk(toadd.pos, toadd.size, toadd.prev, toadd.next);

  /* 
  ** find the index on the indirection table 
  */
  idx = freeChunkIndex(size);  
  /* 
  ** place it into the chain list 
  */
  cur_pos  = vheap->freeTable[idx];
  /* check for first insertion */
  if( !cur_pos )
    {      
      if( !writeFreeChunk(vheap, &toadd) ) /* add the fisrt block at this index */
        {
          ErrorStack(func, "writeFreeChunk[1]");
          return 0;
        }
      done = 1;
    }
  while( !done )
    {
      /* read current block info */
      if( !readFreeChunk(vheap, &cur, cur_pos, &badchksum) )
        {
          if( badchksum )
            {
              ErrorStack(func, "readFreeChunk[1]: bad chksum");
            }
          else
            {
              ErrorStack(func, "readFreeChunk[1]");
            }
          return 0;
        }
      if( (cur.size >= toadd.size) && cur.next ) /* Insert before the cur block (list from small to big)*/
        {
          FreeChunk next = { 0 };
              
          /* read next block info */
          if( !readFreeChunk(vheap, &next, cur.next, &badchksum) )
            {
              if( badchksum )
                {
                  ErrorStack(func, "readFreeChunk[2]: bad chksum");
                }
              else
                {
                  ErrorStack(func, "readFreeChunk[2]");
                }
              return 0;
            }              
          cur.next    = toadd.pos;
          toadd.next  = next.pos;
          next.prev   = toadd.pos;
          toadd.prev  = cur.pos;
          /* applied changes to the heap */
          if( !writeFreeChunk(vheap, &toadd) ) 
            {
              ErrorStack(func, "writeFreeChunk[2]");
              return 0;
            }
          else if( !writeFreeChunk(vheap, &cur) ) 
            {
              ErrorStack(func, "writeFreeChunk[3]");
              return 0;
            }          
          else if( !writeFreeChunk(vheap, &next) ) 
            {
              ErrorStack(func, "writeFreeChunk[4]");
              return 0;
            }
          done = 1;
        }
      else
        {
          if( !cur.next ) /* Add after */
            {
              cur.next    = toadd.pos;
              toadd.prev  = cur.pos;
              /* applied changes to the heap */
              if( !writeFreeChunk(vheap, &toadd) ) 
                {
                  ErrorStack(func, "writeFreeChunk[5]");
                  return 0;
                }
              else if( !writeFreeChunk(vheap, &cur) ) 
                {
                  ErrorStack(func, "writeFreeChunk[6]");
                  return 0;
                }

              done = 1;
            }
          else /* go to next block */
           {
             cur_pos = cur.next;
           }
        }
    }
  /* if it is the first block of the list */
  if( !toadd.prev )
    {
      vheap->freeTable[idx] = toadd.pos;
    }
  if( fchunkP )
    {
      *fchunkP = toadd;
    }
  return 1;
}



/*
** Remove a free chunk from the FreeChunk table
** - return '1' on success, '0' otherwise
*/
static int remFreeChunk
  (
    ZkVHeapHdr       vheap,         /* [in]     ZkVHeapHdr */
    FreeChunk       *fchunkP        /* [in/out] pointer on a FreeChunk */
  )
{
  const char * func = "remFreeChunk";
  uint16_t  idx;
  FreeChunk next;
  FreeChunk prev; 
  FreeChunk wipe = {0}; 
  ZkBool    badchksum = ZK_FALSE;

  if( !readFreeChunk(vheap, fchunkP, fchunkP->pos, &badchksum) )
    {
      if( badchksum )
        {
          ErrorStack(func, "readFreeChunk[1]: bad chksum");
        }
      else
        {
          ErrorStack(func, "readFreeChunk[1]");
        }
      return 0;
    }

  /* 
  ** find the index on the indirection table 
  */
  idx = freeChunkIndex(fchunkP->size);  
  /*
  ** Ajust the integrity of the free block index table
  */
  if( !fchunkP->prev )
    {
      vheap->freeTable[idx] = fchunkP->next;
    }
  /*
  ** Wipe free chunck header
  */
  wipe.pos = fchunkP->pos;

  if( !writeFreeChunk(vheap, &wipe) ) 
    {
      ErrorStack(func, "writeFreeChunk[1]");
      return 0;
    }
  /*
  ** Ajust the next element
  */
  if( fchunkP->next )
    {
      if( !readFreeChunk(vheap, &next, fchunkP->next, &badchksum) )
        {
          if( badchksum )
            {
              ErrorStack(func, "readFreeChunk[2]: bad chksum");
            }
          else
            {
              ErrorStack(func, "readFreeChunk[2]");
            }
          return 0;
        }
      next.prev = fchunkP->prev;
      if( !writeFreeChunk(vheap, &next) ) /* applied changes to the heap */
        {
          ErrorStack(func, "writeFreeChunk[2]");
          return 0;
        }
    }
  /*
  ** Ajust the previous element
  */
  if( fchunkP->prev )
    {
      if( !readFreeChunk(vheap, &prev, fchunkP->prev, &badchksum) )
        {
          if( badchksum )
            {
              ErrorStack(func, "readFreeChunk[3]: bad chksum");
            }
          else
            {
              ErrorStack(func, "readFreeChunk[3]");
            }          
          return 0;
        }
      prev.next = fchunkP->next;
      if( !writeFreeChunk(vheap, &prev) ) /* applied changes to the heap */
        {
          ErrorStack(func, "writeFreeChunk[3]");
          return 0;
        }
    }
  return 1;
}




/*
** Return the largest free Chunk available
*/
static int largestFreeChunk
  (
    ZkVHeapHdr       vheap,         /* [in]     ZkVHeapHdr */
    uint32_t        *sizeP          /* [in]     size (in blocks) */
  )
{
  const char * func = "largestFreeChunk";
  uint16_t  idx = ZK_VHEAP_FREETABLE_SIZE;
  uint32_t  cur_pos;
  FreeChunk next;
  ZkBool    badchksum = ZK_FALSE;

  *sizeP = 0;
  while( idx )
    {
      if( (cur_pos = vheap->freeTable[idx-1]) != 0 )
        {
          while( readFreeChunk(vheap, &next, cur_pos, &badchksum) )
            {
              if( next.size > *sizeP )
                {          
                  *sizeP = next.size;
                }
              if( (cur_pos = next.next) == 0 )
                {
                  return 1;
                }
            }
          if( badchksum )
            {
              ErrorStack(func, "readFreeChunk: bad chksum");
            }
          else
            {
              ErrorStack(func, "readFreeChunk");
            }
          return 0;
        }
      else
        {
          idx--;
        }
    }
  return 1;
}





/* **********************************************************************
** **********************************************************************
**
**
** CHUNK MANAGEMENT CALLS
**
**
** **********************************************************************
** **********************************************************************
*/

/*
** Verify if this chunk is marked as free
** - return '1' on success, '0' otherwise
*/
static int isChunkMarkAsFree
  (
    ZkVHeapHdr       vheap,         /* [in]     ZkVHeapHdr */
    FreeChunk       *chunkP,        /* [in/out] pointer on the founded FreeChunk */
    uint32_t         pos            /* [in]     postion (in blocks) of the free chunk */
  )
{
  const char * func = "isChunkMarkAsFree";
  FreeChunk chunk;

  if( !getFreeChunk(vheap, &chunk, pos) )
    {
      ErrorStack(func, "getFreeChunk");
      return 0;
    }
  if( chunkP )
    {
      *chunkP = chunk;
    }
  return 1;
}




/*
** Verify if this chunk is marked as alloc
** - return '1' on success, '0' otherwise
*/
static int isChunkMarkAsAlloc
  (
    ZkVHeapHdr       vheap,         /* [in]     ZkVHeapHdr */
    AllocChunk      *chunkP,        /* [in/out] pointer on the founded AllocChunk */
    uint32_t         pos            /* [in]     postion (in blocks) of the free chunk */
  )
{
  const char * func = "isChunkMarkAsAlloc";
  AllocChunk chunk;

  if( !getAllocChunk(vheap, &chunk, pos) )
    {
      ErrorStack(func, "getAllocChunk");
      return 0;
    }
  if( chunkP )
    {
      *chunkP = chunk;
    }
  return 1;
}




/*
** Mark a chunk as free
** - return '1' on success, '0' otherwise
*/
static int markChunkAsFree
  (
    ZkVHeapHdr       vheap,         /* [in]     ZkVHeapHdr */
    uint32_t         pos,           /* [in]     postion (in blocks) of the free chunk */
    uint32_t         size           /* [in]     size (in blocks) of the free chunk */
  )
{
  const char * func = "markChunkAsFree";

  if( !addFreeChunk(vheap, NULL, pos, size) )
    {
      ErrorStack(func, "addFreeChunk");
      return 0;
    }
  return 1;
}



/*
** Mark a chunk as free
** - return '1' on success, '0' otherwise
*/
static int unmarkChunkAsFree
  (
    ZkVHeapHdr       vheap,         /* [in]     ZkVHeapHdr */
    FreeChunk       *chunkP         /* [in]     pointer on a chunk */
  )
{
  const char * func = "unmarkChunkAsFree";

  if( !remFreeChunk(vheap, chunkP) )
    {
      ErrorStack(func, "remFreeChunk");
      return 0;
    }
  return 1;    
}



/*
** Mark a chunk as allocated
** - return '1' on success, '0' otherwise
*/
static int markChunkAsAlloc
  (
    ZkVHeapHdr       vheap,         /* [in]     ZkVHeapHdr */
    uint32_t         pos,           /* [in]     postion (in blocks) of the free chunk */
    uint32_t         size           /* [in]     size (in blocks) of the free chunk */
  )
{
  const char * func = "markChunkAsAlloc";

  if( !addAllocChunk(vheap, NULL, pos, size) )
    {
      ErrorStack(func, "addAllocChunk");
      return 0;
    }
  return 1;    
}



/*
** Mark a chunk as allocated
** - return '1' on success, '0' otherwise
*/
static int unmarkChunkAsAlloc
  (
    ZkVHeapHdr       vheap,         /* [in]     ZkVHeapHdr */
    AllocChunk      *chunkP         /* [in]     pointer on a chunk */
  )
{
  const char * func = "unmarkChunkAsAlloc";

  if( !remAllocChunk(vheap, chunkP) )
    {
      ErrorStack(func, "remAllocChunk");
      return 0;
    }
  return 1;    
}




/*
** Return the chunk allocated before 'pos'
** - return '1' if found, '0' otherwise
*/
static int prevAllocChunk
  (
    ZkVHeapHdr        vheap,        /* [in]     ZkVHeapHdr */
    AllocChunk       *chunkP,       /* [in/out] pointer on a found AllocChunk */
    uint32_t          pos           /* [in]     starting position (in blocks) (included) */
  )
{
  const char * func = "prevAllocChunk";
  AllocChunk chunk;

  if( !pos || !rfindAllocChunk(vheap, &chunk, pos) )
    {
      ErrorStack(func, "1");       
      return 0;
    }
  if( chunkP )
    {
      *chunkP = chunk;
    }
  return 1;
}




/*
** Return the chunk free before 'pos'
** - return '1' if found, '0' otherwise
*/
static int prevFreeChunk
  (
    ZkVHeapHdr       vheap,         /* [in]     ZkVHeapHdr */
    FreeChunk       *chunkP,        /* [in/out] pointer on the founded FreeChunk */
    uint32_t         pos            /* [in]     starting position (in blocks) (included) */
  )
{
  const char * func = "prevFreeChunk";
  AllocChunk achunk;
  FreeChunk  fchunk;

  /*
  ** find the previous allocated block
  */
  if( !prevAllocChunk(vheap, &achunk, pos) )
    {
      ErrorStack(func, "prevAllocChunk");
      return 0;
    }
  /*
  ** is there a gap between it an the startin position?
  */
  if( (achunk.beg + achunk.size) > pos )
    {
      ErrorStack(func, "(achunk.beg + achunk.size) > pos");
      return 0;
    }
  /*
  ** is this gap marked as free 
  */
  if( !isChunkMarkAsFree(vheap, &fchunk, achunk.beg + achunk.size) )
    {
      ErrorStack(func, "isChunkMarkAsFree");
      return 0;
    }
  /*
  ** is there a gap between this free gap and the startin position?
  */
  if( (fchunk.pos + fchunk.size - 1) > pos )
    {
      ErrorStack(func, " (fchunk.pos + fchunk.size - 1) > pos ");
      return 0;
    }
  if( chunkP )
    {
      *chunkP = fchunk;
    }
  return 1;
}




/*
** Return the chunk free after 'pos'
** - return '1' if found, '0' otherwise
*/
static int nextFreeChunk
  (
    ZkVHeapHdr       vheap,         /* [in]     ZkVHeapHdr */
    FreeChunk       *chunkP,        /* [in/out] pointer on the founded FreeChunk */
    uint32_t         pos            /* [in]     postion (in blocks) of the current free chunk */
  )
{
  const char * func = "nextFreeChunk";
  FreeChunk  fchunk;

  if( !isChunkMarkAsFree(vheap, &fchunk, pos) )
    {
      ErrorStack(func, "isChunkMarkAsFree");
      return 0;
    }
  if( chunkP )
    {
      *chunkP = fchunk;
    }
  return 1;
}




/*
** Scan blocks before a free chunk and try to create a larger chunk.
** - return '1' on success, '0' on error
*/
static int expandFreeChunkWithPreviousChunck
  (
    ZkVHeapHdr       vheap,       /* [in] ZkVHeapHdr */
    FreeChunk       *newP,        /* [in/out] pointer on the new founded FreeChunk */
    FreeChunk       *curP         /* [in] pointer on the current FreeChunk */
  )
{
  const char * func = "expandFreeChunkWithPreviousChunck";
  FreeChunk tojoin;
  FreeChunk result;
  ZkBool    didjoin   = ZK_FALSE;
  ZkBool    badchksum = ZK_FALSE;

  result = *curP;
  /*
  ** join previous free block (if there is)
  */
  if( result.pos ) 
    {
      if( prevFreeChunk(vheap, &tojoin, result.pos-1) )
        {
          if( !unmarkChunkAsFree(vheap, &tojoin) )
            {
              ErrorStack(func, "unmarkChunkAsFree[1]");
              return 0;
            }
          result.pos   = tojoin.pos;
          result.size += tojoin.size;
          didjoin      = ZK_TRUE;
        }
    }
  /*
  ** create this expanded free block
  */
  if( didjoin )
    {
      if( !unmarkChunkAsFree(vheap, curP) )
        {
          ErrorStack(func, "unmarkChunkAsFree[2]");
          return 0;
        }
      else if( !markChunkAsFree(vheap, result.pos, result.size) )
        {
          ErrorStack(func, "markChunkAsFree");
          return 0;
        }
      else if( !readFreeChunk(vheap, &result, result.pos, &badchksum) )
        {
          if( badchksum )
            {
              ErrorStack(func, "readFreeChunk: bad chksum");
            }
          else
            {
              ErrorStack(func, "readFreeChunk");
            }
          return 0;
        }
    }     
  /*
  ** return the new block
  */
  if( newP )
    {
      *newP = result;
    }
  return 1;
}



/*
** Scan blocks after a free chunk and try to create a larger chunk.
** - return '1' on success, '0' on error
*/
static int expandFreeChunkWithNextChunck
  (
    ZkVHeapHdr       vheap,       /* [in] ZkVHeapHdr */
    FreeChunk       *newP,        /* [in/out] pointer on the new founded FreeChunk */
    FreeChunk       *curP         /* [in] pointer on the current FreeChunk */
  )
{
  const char * func = "expandFreeChunkWithNextChunck";
  FreeChunk tojoin;
  FreeChunk result;
  ZkBool    didjoin = ZK_FALSE;
  ZkBool    badchksum = ZK_FALSE;

  result = *curP;
  /*
  ** join next free block (if there is)
  */
  if( (result.pos + result.size) < vheap->heapBlockSize ) 
    {
      if( nextFreeChunk(vheap, &tojoin, result.pos + result.size) )
        {
          if( !unmarkChunkAsFree(vheap, &tojoin) )
            {
              ErrorStack(func, "unmarkChunkAsFree[1]");
              return 0;
            }
          result.size += tojoin.size;
          didjoin      = ZK_TRUE;
        }
    }
  /*
  ** create this expanded free block
  */
  if( didjoin )
    {
      if( !unmarkChunkAsFree(vheap, curP) )
        {
          ErrorStack(func, "unmarkChunkAsFree[1]");
          return 0;
        }
      if( !markChunkAsFree(vheap, result.pos, result.size) )
        {
          ErrorStack(func, "markChunkAsFree");
          return 0;
        }
      else if( !readFreeChunk(vheap, &result, result.pos, &badchksum) )
        {
          if( badchksum )
            {
              ErrorStack(func, "readFreeChunk: bad chksum");
            }
          else
            {
              ErrorStack(func, "readFreeChunk");
            }
          return 0;
        }
    }     
  /*
  ** return the new block
  */
  if( newP )
    {
      *newP = result;
    }
  return 1;
}




/*
** Scan blocks before and after a free chunk and try to create a larger chunk.
** - return '1' on success, '0' on error
*/
static int expandFreeChunk
  (
    ZkVHeapHdr       vheap,       /* [in] ZkVHeapHdr */
    FreeChunk       *newP,        /* [in/out] pointer on the new founded FreeChunk */
    FreeChunk       *curP         /* [in] pointer on the current FreeChunk */
  )
{
  const char * func = "expandFreeChunk";
  FreeChunk withprev = {0};
  FreeChunk withnext = {0};
  FreeChunk new      = {0};
  
  new = *curP;

  if( !expandFreeChunkWithPreviousChunck(vheap, &withprev, &new) )
    {
      ErrorStack(func, "expandFreeChunkWithPreviousChunck");
      return 0;
    }
  else
    {
      new = withprev;
      if( !expandFreeChunkWithNextChunck(vheap, &withnext, &new) )
        {
          ErrorStack(func, "expandFreeChunkWithNextChunck");
          return 0;
        }
      else
        {
          new = withnext;
        }
    }
  if( newP )
    {
      *newP = new;
    }
  return 1;
}
     



/*
** Split a free chunk to keep size block only.
** - return '1' on success, '0' on error
*/
static int splitFreeChunk
  (
    ZkVHeapHdr       vheap,         /* [in]     ZkVHeapHdr */
    FreeChunk       *chunkP,        /* [in/out] pointer on the resulting FreeChunk */
    uint32_t         pos,           /* [in]     postion (in blocks) of free chunk to be split */
    uint32_t         size           /* [in]     size (in blocks) of the resulting chunk */
  )
{
  const char * func = "splitFreeChunk";
  FreeChunk chunk;

  /* check if this chunk is large enough */
  if( !getFreeChunk(vheap, &chunk, pos) || (chunk.size < size) )
    {
      ErrorStack(func, "getFreeChunk");
      return 0;
    }
  /* remove this chunk of the free list */
  if( !unmarkChunkAsFree(vheap, &chunk) )
    {
      ErrorStack(func, "unmarkChunkAsFree");
      return 0;
    }
  /* mark the new chunk as free */
  if( !markChunkAsFree(vheap, pos, size) )
    {
      ErrorStack(func, "markChunkAsFree[1]");
      return 0;
    }
  /* put back the rest of the chunk into the free list */
  if( !markChunkAsFree(vheap, chunk.pos+size, chunk.size - size) )
    {
      ErrorStack(func, "markChunkAsFree[2]");
      return 0;
    }
  if( chunkP )
    {
      if( !getFreeChunk(vheap, chunkP, pos) )
        {
          ErrorStack(func, "getFreeChunk");
          return 0;
        }
    }
  return 1;
}




/*
** Fill a FreeChunk structure from a AllocChunk
*/
static void Alloc2FreeChunk
  (
    FreeChunk  *fchunkP,         /* [in/out] pointer on the resulting FreeChunk */
    AllocChunk *achunkP          /* [in] pointer to an AllocChunk */
  )
{
  fillFreeChunk(fchunkP, achunkP->beg, achunkP->size, 0, 0);
}





/* **********************************************************************
** **********************************************************************
**
**
** ALLOCATION CALLS
**
**
** **********************************************************************
** **********************************************************************
*/

/*
** Try to allocate a chunk of memory
*/
static int allocMemory
  (
    ZkVHeapHdr       vheap,         /* [in]     ZkVHeapHdr */    
    uint32_t        *blockP,        /* [in/out] position (in blocks) of the resulting chunk */
    uint32_t         size           /* [in]     size (in blocks) of the resulting chunk */
  )
{
  const char * func = "allocMemory";
  FreeChunk  fchunk;

  if( !vheap || !blockP ) /* pointer checking */
    {
      ErrorStack(func, "1");       
      return 0;
    }
  if( !size || (size > vheap->heapFreeSpace) ) /* is space available */
    {
      ErrorStack(func, "2");
      return 0;
    }
  if( !findFreeChunk(vheap, &fchunk, size) ) /* find a free chunk */
    {
      ErrorStack(func, "findFreeChunk");
      return 0;
    }  
  if( fchunk.size > size ) /* split it if it is to big */
    {
      if( !splitFreeChunk(vheap, &fchunk, fchunk.pos, size) ) 
        {
          ErrorStack(func, "splitFreeChunk");
          return 0;
        }
    }
  /* 
  ** ajust free and alloc tables 
  */
  if( !markChunkAsAlloc(vheap, fchunk.pos, size) ) 
    {
      ErrorStack(func, "markChunkAsAlloc");
      return 0;
    }
  if( !unmarkChunkAsFree(vheap, &fchunk) )
    {
      ErrorStack(func, "unmarkChunkAsFree");
      return 0;
    }  

  UNLOCK_ALLOC_CHUNCK(vheap, fchunk.pos, size);

  vheap->heapFreeSpace -= size; /* ajust vitual heap counter */  
  *blockP = fchunk.pos; /* return the allocated pointer */
  return 1;
}




/*
** Free a chunk of memory
*/
static int freeMemory
  (
    ZkVHeapHdr       vheap,         /* [in] ZkVHeapHdr */    
    uint32_t         block          /* [in] position (in blocks) of the resulting chunk */
  )
{
  const char * func = "freeMemory";
  FreeChunk  fchunk;
  AllocChunk achunk;
  uint32_t   pos = 0;
  uint32_t   size = 0;

  if( !vheap ) /* pointer checking */
    {
      ErrorStack(func, "!vheap");
      return 0;
    }
  /* check if this chunk is allocate */
  if( !isChunkMarkAsAlloc(vheap, &achunk, block) )
    {
      ErrorStack(func, "isChunkMarkAsAlloc");
      return 0;
    }
  /* unallocate it */
  if( !unmarkChunkAsAlloc(vheap, &achunk) )
    {
      ErrorStack(func, "unmarkChunkAsAlloc");
      return 0;
    }
  /* mark it as free */
  Alloc2FreeChunk(&fchunk, &achunk);
  if( !markChunkAsFree(vheap, fchunk.pos, fchunk.size) )
    {
      ErrorStack(func, "markChunkAsFree");
      return 0;
    }

  /* Keep this information for locking this memory space */
  pos  =  fchunk.pos;
  size =  fchunk.size;

  /* Wipe the chunk memory space. */
  if( vheap->heapFlags & ZKVHEAP_WIPE_FREE_SPACE_CREATION_FLAGS )
    {
      wipeFreeChunk(vheap, &fchunk);
    }
  /* try to join it with neighbur chunk */
  if( !expandFreeChunk(vheap, NULL, &fchunk) )
    {
      ErrorStack(func, "expandFreeChunk");
      return 0;
    }  

  LOCK_RW_ALLOC_CHUNCK(vheap, pos, size);

  vheap->heapFreeSpace += achunk.size; /* ajust vitual heap counter */
  return 1;
}




/*
** Try to reallocate a chunk of memory
*/
static int reallocMemory
  (
    ZkVHeapHdr       vheap,         /* [in]     ZkVHeapHdr */    
    uint32_t        *blockP,        /* [in/out] position (in blocks) of the resulting chunk */
    uint32_t         size           /* [in]     size (in blocks) of the resulting chunk */
  )
{
  const char * func = "reallocMemory";
  AllocChunk cur;
  uint32_t   newblock;
  uint32_t   orig_pos = 0;
  uint32_t   orig_size = 0;

  if( !vheap || !blockP ) /* pointer checking */
    {
      ErrorStack(func, "1");
      return 0;
    }
  if( !size ) /* size checking */
    {
      ErrorStack(func, "!size");
      return 0;
    }
  /*
  ** Find the current chunk
  */
  if( !findAllocChunk(vheap, &cur, *blockP) || (cur.beg != *blockP) )
    {
      ErrorStack(func, "findAllocChunk");
      return 0;
    }

  /*
  ** If the chunck stay the same size
  */
  if( cur.size == size ) 
    {
      return 1;
    }  

  orig_pos  = cur.beg;
  orig_size = cur.size;

  /*
  ** Lazy way to do it, always reallocate a new chunck instead of trying to
  ** join it with neighbur chunks
  */
  if( !allocMemory(vheap, &newblock, size) )
    {
      ErrorStack(func, "allocMemory");
      return 0;
    }
  else
    {     
      if( size > cur.size ) /* If the chunck grow */
        {          
          /*
          ** Copy the original chunk to it new location
          */
          COPY_BLOCKS(vheap, newblock, cur.beg, cur.size);
          /*
          ** Wipe the rest of the chunk memory space.
          */
          if( vheap->heapFlags & ZKVHEAP_WIPE_FREE_SPACE_CREATION_FLAGS )
            {
              WIPE_BLOCKS(vheap, cur.end+1, size-cur.size);
            }          
        }
      else /* If the chunck become smaller */
        {
          /*
          ** Copy the original chunk to it new location
          */
          COPY_BLOCKS(vheap, newblock, cur.beg, size);
        }
      /*
      ** Liberate the original chunk
      */
      if( !freeMemory(vheap, cur.beg) )
        {
          ErrorStack(func, "freeMemory");
          return 0;
        }
    }

  LOCK_RW_ALLOC_CHUNCK(vheap, orig_pos, orig_size);
  UNLOCK_ALLOC_CHUNCK(vheap, newblock, size);

  *blockP = newblock;
  return 1;
}




/*
** Enter into a critical section
*/
static int enterCriticalSection
 (
  ZkError   err,                        /* [in]  ZkError */                   
  ZkSem     mutex                       /* [in]  ZkSem */
 )
{
  int ret;

  if( mutex )
    {      
      ret = zkSemWait(err, mutex);
      return ret;
    }
  return 1;
}




/*
** Leave a critical section
*/
static int leaveCriticalSection
 (
  ZkError   err,                        /* [in]  ZkError */                   
  ZkSem     mutex                       /* [in]  ZkSem */
 )
{
  if( mutex )
    {
      return zkSemRelease(err, mutex);
    }
  return 1;
}





/* **********************************************************************
** **********************************************************************
**
**
** EXPORTED CALLS
**
**
** **********************************************************************
** **********************************************************************
*/

/*
** Create a new virtual heap handler
*/
DLLExport int zkVHeapNew
(
  ZkError     err,           /* [in]     pointer on a ZkError */
  ZkVHeap    *vheapP,        /* [in/out] pointer on a ZkVHeap */
  void       *memP,          /* [in]     pointer on a memory block to be partition */
  size_t      msz,           /* [in]     size (in bytes) of the memory block */
  ZkSem       mutex,         /* [in]     mutex use to synchronyze critical section (optional) */
  uint32_t    flags          /* [in]     optional flags */
)
{
  static const char   *func       = "zkVHeapNew";
  int                  error      = 0;
  struct S_ZkVHeapHdr  vhh        = { 0 };
  struct S_ZkVHeap    *vhP        = NULL;
  uint32_t             headersize;
  ZkBool               critical   = ZK_FALSE;
  void                *alignMemP  = 0;
  uint32_t             alignMsz   = 0;
  uint32_t             alignDelta = 0;
  
  zkErrorClear(err);

  ErrorClear();


  /*
  ** check for invalid pointers
  */
  if( ! vheapP )
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKENULLPTR, "vheapP is null");
      error = 1;
      goto end;
    }

  if( ! memP ) 
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKENULLPTR, "memP is null");
      error = 1;
      goto end;
    }

#ifdef VHEAP_MPROTECT_LOCKING
  /*
  ** Allign the memory block to the granulariry
  */
  if( (uint32_t)memP % (uint32_t)ZK_VHEAP_GRANULARITY_DEFAULT )
    {
      alignDelta = ZK_VHEAP_GRANULARITY_DEFAULT - ((uint32_t)memP % (uint32_t)ZK_VHEAP_GRANULARITY_DEFAULT);
    }
  else
    {
      alignDelta = 0;
    }
  if( msz < alignDelta )
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKEPARMBAD, "msz is too small");
      error = 1;
      goto end;
    }
  alignMemP = &(((char *)memP)[alignDelta]);
  alignMsz  = msz - alignDelta;  
#else
  alignMemP  = memP;
  alignMsz   = msz;
  alignDelta = 0;
#endif
  /*
  ** validate the memory space size
  */
  vhh.heapGranularity      = ZK_VHEAP_GRANULARITY_DEFAULT;
  vhh.allocatedBmBitSize   = BYTE2BLOCK(&vhh, alignMsz) - 1;
  vhh.allocatedBmBegOffset = BYTE2BLOCK(&vhh, sizeof(vhh));
  vhh.allocatedBmEndOffset = vhh.allocatedBmBegOffset + BIT2BLOCK(&vhh, vhh.allocatedBmBitSize);
  /* calculate the size of the header (in blocks) */
  headersize              = vhh.allocatedBmEndOffset + BIT2BLOCK(&vhh, vhh.allocatedBmBitSize);

  /*
  ** Verify if this address space is long enough
  */  
  if( vhh.allocatedBmBitSize < (headersize+1U) )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMVHEAP, ZKEPARMBAD,"msz is too small: should be at least %u bytes", BLOCK2BYTE(&vhh,headersize+1)+alignDelta);
      error = 1;
      goto end;
    }

  /*
  ** set up the heap structure 
  */
  vhh.heapAddress      = alignMemP;
  vhh.heapByteSize     = alignMsz; 
  vhh.heapBlockSize    = vhh.allocatedBmBitSize;
  vhh.heapFreeSpace    = vhh.heapBlockSize - headersize;
  vhh.heapFlags        = flags;
  vhh.firstBlockOffset = headersize;

  if( !zkMemAlloc(err, (void **)&vhP, sizeof(struct S_ZkVHeap)) )
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKEAGAIN,"zkMemAlloc failure");
      error = 1;
      goto end;
    }

  /*
  ** Enter the critical section
  */
  if( !enterCriticalSection(err, mutex) )
    {
      error = 1;
      goto end;
    }
  else
    {
      critical = ZK_TRUE;
    }

  /*
  ** Initialyze the entire memory block
  */
  memset(alignMemP, 0x00, BLOCK2BYTE(&vhh,vhh.heapBlockSize));

  /*
  ** Copy the header over the given memory
  */
  memcpy(alignMemP, &vhh,  (size_t)sizeof(vhh));
  
  /*
  ** Reserve 'headersize' first block for the header
  */
  if( !markChunkAsAlloc((ZkVHeapHdr)alignMemP, 0, headersize) )
    {
      error = 1;
      goto end;
    }
  /*
  ** Mark the rest of the space as free
  */
  if( !markChunkAsFree((ZkVHeapHdr)alignMemP, vhh.firstBlockOffset,  vhh.heapFreeSpace) )
    {
      error = 1;
      goto end;
    }

  /*
  ** Build the handler
  */
  vhP->hdr   = (ZkVHeapHdr)alignMemP;
  vhP->mutex = mutex;

end:
  /*
  ** Leave the critical section
  */
  if( !leaveCriticalSection(err, mutex) )
    {
      error = 1;
    }
  
  /*
  ** cleanup if error
  */
  if( error )
    {
      if( vhP )
        {
          zkMemFree(err, (void **)&vhP);
        }
    }
  else
    {
      LOCK_RW_FREE_CHUNCK(vhP->hdr, vhh.firstBlockOffset, vhh.heapFreeSpace);
      LOCK_VHEAP_HEADER(vhP->hdr);
    }

  /*
  ** Assign the return handle
  */
  *vheapP = vhP;

  return !error;
}




/*
** Open an already initialyzed virtual heap handler
*/
int zkVHeapOpen
(
  ZkError  err,           /* [in]     pointer on a ZkError */
  ZkVHeap *vheapP,        /* [in/out] pointer on a ZkVHeap */
  void    *memP,          /* [in]     pointer on a memory block already partition */
  ZkSem    mutex          /* [in]     mutex use to synchronyze critical section (optional) */
)
{
  static const char *func  = "zkVHeapOpen";
  struct S_ZkVHeap  *vhP = NULL;
  int error = 0;
  void              *alignMemP  = 0;
  uint32_t           alignDelta = 0;

  zkErrorClear(err);

  ErrorClear();

  /*
  ** check for invalid pointers
  */
  if( ! vheapP )
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKENULLPTR, "vheapP is null");
      ErrorDump(err, func);
      error = 1;
    }
  else if( ! memP ) 
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKENULLPTR, "memP is null");
      ErrorDump(err, func);
      error = 1;
    }


#ifdef VHEAP_MPROTECT_LOCKING
  /*
  ** Allign the memory block to the granulariry
  */ 
  if( (uint32_t)memP % (uint32_t)ZK_VHEAP_GRANULARITY_DEFAULT )
    {
      alignDelta = ZK_VHEAP_GRANULARITY_DEFAULT - ((uint32_t)memP % (uint32_t)ZK_VHEAP_GRANULARITY_DEFAULT);
    }
  else
    {
      alignDelta = 0;
    }

  alignMemP = &(((char *)memP)[alignDelta]);
#else
  alignMemP  = memP;
  alignDelta = 0;
#endif

  if( !error )
    {
      if( !zkMemAlloc(err, (void **)&vhP, sizeof(struct S_ZkVHeap)) )
        {
          ErrorDump(err, func);
          error = 0;
        }
      else
        {
          /* Build the handler */
          vhP->hdr   = (ZkVHeapHdr)alignMemP;
          vhP->mutex = mutex;
        }
    }

  *vheapP = vhP;

  return 1;
}




/*
** Free a virtual heap handler
*/
DLLExport int zkVHeapFree 
(
  ZkError  err,           /* [in]     pointer on a ZkError */
  ZkVHeap *vheapP         /* [in/out] pointer on a ZkVHeap */
)
{
  static const char *func = "zkVHeapFree";
  int error = 0;

  zkErrorClear(err);

  ErrorClear();

  /*
  ** check for invalid pointers
  */
  if( !vheapP )
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKENULLPTR, "vheapP is null");
      error = 1;
    }
  else 
    {
      UNLOCK_VHEAP_HEADER((*vheapP)->hdr);
      if( !zkMemFree(err, (void **)vheapP) )
        {
          error = 1;
        }
      *vheapP = NULL;
    }

  if( error )
    {
      ErrorDump(err, func);
    }
  return !error;
}




/*
** Allocate a chunk of memory
*/
DLLExport int zkVHeapMemoryAlloc
(
  ZkError  err,           /* [in]     pointer on a ZkError */
  ZkVHeap  vheap,         /* [in]     ZkVHeap */
  void   **addrP,         /* [in/out] pointer on the allocated address space */
  uint32_t size           /* [in]     size (in bytes) to be allocate */
)
{
  static const char *func   = "zkVHeapMemoryAlloc";
  int      error = 0;
  uint32_t largest_size;
  uint32_t block_size;
  uint32_t block_pos;
  ZkBool   critical = ZK_FALSE;

  zkErrorClear(err);

  ErrorClear();

  /*
  ** check for invalid pointers
  */
  if( ! vheap )
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKENULLPTR, "vheap is null");
      error = 1;
      goto error;
    }
  if( !addrP )
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKENULLPTR, "addrP is null");
      error = 1;
      goto error;
    }
  if( !size )
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKENULLPTR, "size must be greather than zero");
      error = 1;
      goto error;
    }

  UNLOCK_VHEAP_HEADER(vheap->hdr);

  block_size = BYTE2BLOCK(vheap->hdr, size);

  /*
  ** Enter the critical section
  */
  if( !enterCriticalSection(err, vheap->mutex) )
    {
      error = 1;
      goto error;
    }
  else
    {
      critical = ZK_TRUE;
      if( !allocMemory(vheap->hdr, &block_pos, block_size) )
        {
          if( !largestFreeChunk(vheap->hdr, &largest_size) )
            {
              zkErrorSet_va(err, func, ZKERR, ZKMVHEAP, ZKEAGAIN, "internal error(largestFreeChunk)");
              error = 1;
              goto error;
            }
          zkErrorSet_va(err, func, ZKERR, ZKMVHEAP, ZKEAGAIN, "can't allocate %u bytes, %u available", size, BLOCK2BYTE(vheap->hdr, largest_size));
          error = 1;
          goto error;
        }
      *addrP = BLOCK2ADDRESS(vheap->hdr, block_pos);
    }

 error:
  if( critical )
    {
      /*
      ** Leave the critical section
      */
      if( !leaveCriticalSection(err, vheap->mutex) )
        {
          error = 1;
        }
    }
  if( error )
    {
      ErrorDump(err, func);
      checkIntegrity(err, vheap->hdr, VHEAP_INTEGRITY_CHECK_QUIET_MODE);
    }
  else
    {
      PERIODICAL_INTEGRITY_CHECK();
    }

  LOCK_VHEAP_HEADER(vheap->hdr);

  return !error;
}


/*
** Try to reallocate a chunk of memory
*/
DLLExport int zkVHeapMemoryRealloc
(
  ZkError  err,           /* [in]     pointer on a ZkError */
  ZkVHeap  vheap,         /* [in]     ZkVHeap */
  void   **addrP,         /* [in/out] pointer on the allocated address space (may change) */
  uint32_t size           /* [in]     new size (in bytes) */
)
{
  static const char *func = "zkVHeapMemoryRealloc";
  int      error        = 0;
  uint32_t block_size;
  uint32_t block_pos;
  ZkBool   critical = ZK_FALSE;

  zkErrorClear(err);

  ErrorClear();

  /*
  ** check for invalid pointers
  */
  if( ! vheap )
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKENULLPTR, "vheap is null");
      error = 1;
      goto error;
    }
  else if( !addrP )
    {
      return zkVHeapMemoryAlloc(err, vheap, addrP, size);
    }
  else if( !size )
    {
      return zkVHeapMemoryFree(err, vheap, addrP);
    }

  UNLOCK_VHEAP_HEADER(vheap->hdr);

  block_size = BYTE2BLOCK(vheap->hdr, size);
  block_pos  = ADDRESS2BLOCK(vheap->hdr, *addrP);

  /*
  ** Enter the critical section
  */
  if( !enterCriticalSection(err, vheap->mutex) )
    {
      error = 1;
      goto error;      
    }
  else
    {
      critical = ZK_TRUE;
      if( !reallocMemory(vheap->hdr, &block_pos, block_size) )
        {
          zkErrorSet_va(err, func, ZKERR, ZKMVHEAP, ZKEAGAIN, "can't reallocate this chunck");
          error = 1;
          goto error;
        }
      *addrP = BLOCK2ADDRESS(vheap->hdr, block_pos);
    }
 error:
  if( critical )
    {
      /*
      ** Leave the critical section
      */
      if( !leaveCriticalSection(err, vheap->mutex) )
        {
          error = 1;
        }
    }
  if( error )
    {
      ErrorDump(err, func);
      checkIntegrity(err, vheap->hdr, VHEAP_INTEGRITY_CHECK_QUIET_MODE);
    }
  else
    {
      PERIODICAL_INTEGRITY_CHECK();
    }

  LOCK_VHEAP_HEADER(vheap->hdr);

  return !error;  
}



/*
** Free a chunk of memory
*/
DLLExport int zkVHeapMemoryFree
(
  ZkError  err,           /* [in]     pointer on a ZkError */
  ZkVHeap  vheap,         /* [in]     ZkVHeap */
  void   **addrP          /* [in]     pointer on the allocated address space */
)
{
  static const char *func = "zkVHeapMemoryFree";
  int      error        = 0;
  uint32_t block_pos;
  ZkBool   critical = ZK_FALSE;
  
  zkErrorClear(err);

  ErrorClear();

  /*
  ** check for invalid pointers
  */
  if( ! vheap )
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKENULLPTR, "vheap is null");
      error = 1;
      goto error;
    }
  else if( !addrP )
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKENULLPTR, "addrP is null");
      error = 1;
      goto error;
    }

  UNLOCK_VHEAP_HEADER(vheap->hdr);

  if( !IS_A_BLOCK(vheap->hdr, *addrP) )
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKENULLPTR, "addrP is not a valid pointer");
      error = 1;
      goto error;
    }

  block_pos = ADDRESS2BLOCK(vheap->hdr, *addrP);

  /*
  ** Enter the critical section
  */
  if( !enterCriticalSection(err, vheap->mutex) )
    {
      error = 1;
      goto error;
    }
  else
    {
      critical = ZK_TRUE;
      if( !freeMemory(vheap->hdr, block_pos) )
        {
          zkErrorSet_va(err, func, ZKERR, ZKMVHEAP, ZKEAGAIN, "free fail to release address %p", *addrP);
          error = 1;
          goto error;
        }
      *addrP = NULL;
    }
 error:
  if( critical )
    {
      /*
      ** Leave the critical section
      */
      if( !leaveCriticalSection(err, vheap->mutex) )
        {
          error = 1;
        }
    }
  if( error )
    {
      ErrorDump(err, func);
      checkIntegrity(err, vheap->hdr, VHEAP_INTEGRITY_CHECK_QUIET_MODE);
    }
  else
    {
      PERIODICAL_INTEGRITY_CHECK();
    }

  LOCK_VHEAP_HEADER(vheap->hdr);

  return !error;
}




/*
** Return the largest continious block of memory available
*/
DLLExport int zkVHeapMemoryAvailable
(
  ZkError   err,           /* [in]     pointer on a ZkError */
  ZkVHeap   vheap,         /* [in]     ZkVHeap */
  uint32_t *sizeP          /* [in]     size (in bytes) available */
)
{
  static const char *func = "zkVHeapMemoryAvailable";
  int      error        = 0;
  uint32_t block_size;
  ZkBool   critical = ZK_FALSE;

  zkErrorClear(err);

  ErrorClear();

  /*
  ** check for invalid pointers
  */
  if( ! vheap )
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKENULLPTR, "vheap is null");
      error = 1;
      goto error;
    }
  if( !sizeP )
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKENULLPTR, "sizeP is null");
      error = 1;
      goto error;
    }

  UNLOCK_VHEAP_HEADER(vheap->hdr);

  /*
  ** Enter the critical section
  */
  if( !enterCriticalSection(err, vheap->mutex) )
    {
      error = 1;
      goto error;
    }
  else
    {
      critical = ZK_TRUE;
      if( !largestFreeChunk(vheap->hdr, &block_size) )
        {
          zkErrorSet_va(err, func, ZKERR, ZKMVHEAP, ZKEAGAIN, "internal error(largestFreeChunk)");
          error = 1;
          goto error;
        }
      *sizeP = BLOCK2BYTE(vheap->hdr, block_size);
    }
 error:
  if( critical )
    {
      /*
      ** Leave the critical section
      */
      if( !leaveCriticalSection(err, vheap->mutex) )
        {
          error = 1;
        }
    }
  if( error )
    {
      ErrorDump(err, func);
      checkIntegrity(err, vheap->hdr, VHEAP_INTEGRITY_CHECK_QUIET_MODE);
    }

  LOCK_VHEAP_HEADER(vheap->hdr);

  return !error;
}




/*
** Set a user define hook to the virtual heap
*/
DLLExport int zkVHeapSetHook
(
  ZkError  err,           /* [in]     pointer on a ZkError */
  ZkVHeap  vheap,         /* [in]     pointer on a ZkVHeap */
  char    *hook           /* [in]     pointer the extra data (may be NULL) */
)
{
  static const char *func = "zkVHeapSetHook";
  int error = 0;

  zkErrorClear(err);

  ErrorClear();

  /*
  ** check for invalid pointers
  */
  if( ! vheap )
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKENULLPTR, "vheap is null");
      return 0;
    } 

  UNLOCK_VHEAP_HEADER(vheap->hdr);

  /*
  ** Enter the critical section
  */
  if( !enterCriticalSection(err, vheap->mutex) )
    {      
      error = 1;
    }
  else
    {
      vheap->hdr->hook = hook; 

      /*
      ** Leave the critical section
      */
      if( !leaveCriticalSection(err, vheap->mutex) )
        {
          error = 1;
        }
    }
  if( error )
    {
      ErrorDump(err, func);
    }

  LOCK_VHEAP_HEADER(vheap->hdr);

  return !error;
}




/*
** Set a user define hook to the virtual heap
*/
DLLExport int zkVHeapGetHook
(
  ZkError  err,           /* [in] pointer on a ZkError */
  ZkVHeap  vheap,         /* [in] pointer on a ZkVHeap */
  char   **hookP          /* [out]  pointer the extra data */
)
{
  static const char *func = "zkVHeapGetHook";
  int error = 0;

  zkErrorClear(err);

  ErrorClear();

  /*
  ** check for invalid pointers
  */
  if( !vheap )
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKENULLPTR, "vheap is null");
      return 0;
    } 
  else if( !hookP )
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKENULLPTR, "hookP is null");
      return 0;
    }

  UNLOCK_VHEAP_HEADER(vheap->hdr);

  /*
  ** Enter the critical section
  */
  if( !enterCriticalSection(err, vheap->mutex) )
    {
      error = 1;
    }
  else
    {
      *hookP = vheap->hdr->hook;

      /*
      ** Leave the critical section
      */
      if( !leaveCriticalSection(err, vheap->mutex) )
        {          
          error = 1;
        }
    }

  LOCK_VHEAP_HEADER(vheap->hdr);

  return 1;
}


/*
** Set a user define hook to the virtual heap
*/
DLLExport int zkVHeapCheckIntegrity
(
  ZkError  err,           /* [in] pointer on a ZkError */
  ZkVHeap  vheap,         /* [in] pointer on a ZkVHeap */
  ZkBool   quiet          
)
{
  static const char *func  = "zkVHeapCheckIntegrity";
  int                error = 0;

  zkErrorClear(err);

  ErrorClear();

  /*
  ** check for invalid pointers
  */
  if( !vheap )
    {
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKENULLPTR, "vheap is null");
      return 0;
    } 

  UNLOCK_VHEAP_HEADER(vheap->hdr);

  /*
  ** Enter the critical section
  */
  if( !enterCriticalSection(err, vheap->mutex) )
    {
      error = 1;
    }
  else
    {
      error = !checkIntegrity(err, vheap->hdr, quiet); 

      /*
      ** Leave the critical section
      */
      if( !leaveCriticalSection((!error) ? err : NULL, vheap->mutex) )
        {
          error = 1;
        }
    }
  
  LOCK_VHEAP_HEADER(vheap->hdr);

  return !error;
}


/*
** Verify the integrity of the vheap
*/
static int checkIntegrity
(  
 ZkError    err,
 ZkVHeapHdr vheap,        /* [in]     ZkVHeapHdr */    
 ZkBool     quiet
)
{               
  const char * func = "checkIntegrity";
  char      *map = NULL;
  char       state;
  uint32_t   idx;
  int        error = 0;
  bitmap_t  *begbmP = (bitmap_t *) BEGIN_BITMAP_ADDR(vheap);
  bitmap_t  *endbmP = (bitmap_t *) END_BITMAP_ADDR(vheap);
  uint32_t   bmsize = (uint32_t)   BITMAP_BIT_SIZE(vheap);
  uint32_t   stop   = bmsize;
  char       output[64];
  char       buff[20];
  ZkLog      log = NULL;
  ZkBool     badchksum = ZK_FALSE;

  zkErrorLog(err, &log);   
  
  if( !quiet )
    {
      err = NULL; /* Do not overload the error in non quiet mode */
    }
  
  map = (char *)calloc(1, (size_t)bmsize);
  if( !map )
    {
      error = 1;
      zkErrorSet(err, func, ZKERR, ZKMVHEAP, ZKENULLPTR, "calloc fail");
    }
  else
    {
      /*
      ** Scan the allocation bitmap
      */
      state = 0;
      for( idx=0; !error && idx<bmsize; idx++ )
        {
          int bset = getBit(*begbmP, idx);
          int eset = getBit(*endbmP, idx);

          if( bset && eset )
            {
              map[idx] = 1;
              state    = 0;
            }
          else
            {
              if( bset ) 
                {
                  if( state == 1 )
                    {
                      zkErrorSet_va(err, func, ZKERR, ZKMVHEAP, ZKERROR, "[VHEAP] bitmap integrity error 1 [pos %u]", idx);
                      if (!quiet) zkLogError(log, err);
                      error = 1; 
                      stop  = idx;
                    }
                  else
                    {
                      state    = 1;
                      map[idx] = 1;
                    }
                }
              else if( eset )
                {
                  if( state != 1 )
                    {
                      zkErrorSet_va(err, func, ZKERR, ZKMVHEAP, ZKERROR, "[VHEAP] bitmap integrity error 2 [pos %u]", idx);
                      if (!quiet) zkLogError(log, err);
                      error = 1; 
                      stop  = idx;
                    }
                  else
                    {
                      state    = 0;
                      map[idx] = 1;
                    }
                }
              else
                {
                  if( state )
                    {
                      map[idx] = 1;
                    }
                }
            }
        }
      /*
      ** Scan the free list
      */
      for( idx=0; !error && idx<ZK_VHEAP_FREETABLE_SIZE; idx++ )
        {
          FreeChunk cur;
          uint32_t  pos = vheap->freeTable[idx];
          uint32_t  midx;
          
          /*
          ** travel across the chain list 
          */
          while( !error && pos && readFreeChunk(vheap, &cur, pos, &badchksum) )
            {
              for( midx=0; midx<cur.size; midx++ )
                {
                  if( map[cur.pos+midx] != 0 )
                    {
                      switch(map[cur.pos+midx])
                        {
                        case 1:   
                          {
                            zkErrorSet_va(err, func, ZKERR, ZKMVHEAP, ZKERROR, "[VHEAP] free list integrity error 1 [pos %u], already allocated", cur.pos); 
                            if (!quiet) zkLogError(log, err);
                            break;
                          }
                        case 2:
                          {
                            zkErrorSet_va(err, func, ZKERR, ZKMVHEAP, ZKERROR, "[VHEAP] free list integrity error 1 [pos %u], already freed", cur.pos);
                            if (!quiet) zkLogError(log, err);
                            break;
                          }
                        default: 
                          {
                            zkErrorSet_va(err, func, ZKERR, ZKMVHEAP, ZKERROR, "[VHEAP] free list integrity error 1 [pos %u], unknown state", cur.pos);
                            if (!quiet) zkLogError(log, err);
                            break;
                          }
                        }
                      error = 1; 
                    }
                  else
                    {
                      map[cur.pos+midx] = 2; /* free */
                    }
                }
              pos = cur.next;
            }
          if( badchksum )
            {
              ErrorStack(func, "readFreeChunk: bad chksum");
              error = 1;
            }
        }
      if( !error )
        {
          /*
          ** Scan the entired list to check for global state integrity
          ** everithing should be mark 1 or 2;
          */
          for( idx=0; !error && idx<bmsize; idx++ )
            {
              if( map[idx] == 0 )
                {
                  zkErrorSet_va(err, func, ZKERR, ZKMVHEAP, ZKERROR, "[VHEAP] global integrity error 1 [pos %u]", idx);
                  if (!quiet) zkLogError(log, err);
                  error = 1; 
                }
            }
        }

      if( !quiet || error )
        {
          zkLogWrite_va(log, "[VHEAP] ------------ ALLOCATION REPORT ------------");
          state = map[0];
          switch( state )
            {
            case 0: snprintf(output,sizeof output,"[VHEAP] state[UNDEF] from 0 to "); break;
            case 1: snprintf(output,sizeof output,"[VHEAP] state[ALLOC] from 0 to "); break;
            case 2: snprintf(output,sizeof output,"[VHEAP] state[FREE]  from 0 to "); break;
            }
          for( idx=0; idx<stop; idx++ ) 
            {
              if( map[idx] != state )
                {
                  snprintf(buff,sizeof buff,"%u", idx-1);
                  zkCStrlcat(output, buff, sizeof(output));
                  zkLogWrite_va(log,  output);
                  state = map[idx];
                  switch( state )
                    {
                    case 0: snprintf(output,sizeof output,"[VHEAP] state[UNDEF] from %u to ",idx); break;
                    case 1: snprintf(output,sizeof output,"[VHEAP] state[ALLOC] from %u to ",idx); break;
                    case 2: snprintf(output,sizeof output,"[VHEAP] state[FREE]  from %u to ",idx); break;
                    }
                }
            }      
          snprintf(buff,sizeof buff,"%u", idx-1);
          zkCStrlcat(output, buff, sizeof(output));
          zkLogWrite_va(log, output);
          zkLogWrite_va(log,  "[VHEAP] -------------------------------------------");
        }
			free(map);
    }
  if( error )
    {
      return 0;
    }
  else
    {
      return 1;
    }
}

#endif /* ZK_LINUX || ZK_SOLARIS */
