/*
** **************************************************************************
** $Id: zkAuditHelper.c,v 1.2.30.2 2000/10/11 13:20:08 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

**
** Copyright (C) 1999 Zero-Knowledge Systems Inc.  All Rights Reserved
** **************************************************************************
*/

# include "zks.h"

#include <string.h>

# include "zkAlloc.h"
# include "zkBase64.h"
# include "zkAudit.h"
# include "zkAuditHelper.h"

# include <errno.h>
# include <unistd.h>


/* --------------------------------------------------------------------------
 * zkaudit_EntryCreate
 * --------------------------------------------------------------------------
 *
 * Creates a new ZkAuditEntry structure using the supplied parameters.
 *
 * @param err              the error handling/tracking structure
 * @param user             the identity of the user who is responsible for
 *                         the action being audited
 * @param host             the machine on which the action is taking place
 * @param session_id       the local identifier for the log entry
 * @param service          the module being asked to perform the action
 * @param request          the action being audited
 * @param comment          an optional description, provided by the user, of
 *                         the action being audited (may be NULL)
 * @param ticket_id        an identifier linking the action to the support
 *                         database
 * @param client_id        some identifier which specifies the namespace from
 *                         which the transaction id originates
 * @param transaction_id   a unique identifer linking different actions on
 *                         different hosts
 * @param result_code      a status value denoting the outcome of the action
 * @param result_text      a message explaining the outcome of the action
 * @param audit_entry_ptr the output structure
 */
int
zkaudit_EntryCreate(
    ZkError             err,
    const ZkDateTime *  timeStamp,
    const char *        user,
    const char *        host,
    uint32_t            session_id,
    const char *        service,
    const char *        request,
    const char *        comment,           /* may be NULL */
    const char *        ticket_id,         /* may be NULL */
    const char *        client_id,         /* may be NULL */
    const char *        trans_id,          /* may be NULL */
    uint32_t            result_code,
    const char *        res_text,          /* may be NULL */
    ZkAuditEntry *      audit_entry_ptr )  /* may be NULL */
{
    static const char * func          = "zkaudit_EntryCreate";
    static const char * blank_str     = "<none>";
    ZkAuditEntry        new_entry     = NULL;
    const char *        tmp_user      = NULL;
    const char *        tmp_host      = NULL;
    const char *        tmp_service   = NULL;
    const char *        tmp_request   = NULL;
    const char *        tmp_comment   = NULL;
    const char *        tmp_ticket_id = NULL;
    const char *        tmp_client_id = NULL;
    const char *        tmp_trans_id  = NULL;
    const char *        tmp_res_text  = NULL;

    zkErrorClear(err);

    /*
     * Validate the input parameters
     */

    if ( !(timeStamp && user && host && service && request && audit_entry_ptr) )
    {
        zkErrorSet(err,func,ZKERR,ZKMAUDIT,ZKENULLPTR,
                   "input params are null");
    }
    else
    {
        (*audit_entry_ptr) = NULL;

        /*
         * The variables which were allowed to be NULL on input must be made
         * non-NULL before we try to do anything with them.  In particular,
         * they can't be NULL when we try to duplicate them below.
         */
        
        if (!comment)
        {
            comment = blank_str;
        }
        
        if (!ticket_id)
        {
            ticket_id = blank_str;
        }
        
        if (!client_id)
        {
            client_id = blank_str;
        }
        
        if (!trans_id)
        {
            trans_id = blank_str;
        }
        
        if (!res_text)
        {
            res_text = blank_str;
        }


        /*
         * Allocate and fill in the audit entry structure
         */

        if ( zkMemAlloc(err,(void**)&new_entry,sizeof(*new_entry))
             && zkMemDup(err,(void**)&tmp_user,user,strlen(user)+1)
             && zkMemDup(err,(void**)&tmp_host,host,strlen(host)+1)
             && zkMemDup(err,(void**)&tmp_service,service,strlen(service)+1)
             && zkMemDup(err,(void**)&tmp_request,request,strlen(request)+1)
             && zkMemDup(err,(void**)&tmp_comment,comment,strlen(comment)+1)
             && zkMemDup(err,(void**)&tmp_ticket_id,ticket_id,strlen(ticket_id)+1)
             && zkMemDup(err,(void**)&tmp_client_id,client_id,strlen(client_id)+1)
             && zkMemDup(err,(void**)&tmp_trans_id,trans_id,strlen(trans_id)+1)
             && zkMemDup(err,(void**)&tmp_res_text,res_text,strlen(res_text)+1) )
        {
            new_entry->timestamp      = *timeStamp;
            new_entry->session_id     = session_id;
            new_entry->result_code    = result_code;
            new_entry->user           = tmp_user;
            new_entry->host           = tmp_host;
            new_entry->service        = tmp_service;
            new_entry->request        = tmp_request;
            new_entry->comment        = tmp_comment;
            new_entry->ticket_id      = tmp_ticket_id;
            new_entry->client_id      = tmp_client_id;
            new_entry->transaction_id = tmp_trans_id;
            new_entry->result_text    = tmp_res_text;

            (*audit_entry_ptr) = new_entry;
            
            return (1);
        }
    }

    /*
     * If we reach here then an error has occurred.  Clean up and return
     * an indication of the failure.
     */

    if (tmp_user)
    {
        zkMemFree(NULL, (void**)&tmp_user);
    }

    if (tmp_host)
    {
        zkMemFree(NULL, (void**)&tmp_host);
    }

    if (tmp_service)
    {
        zkMemFree(NULL, (void**)&tmp_service);
    }

    if (tmp_request)
    {
        zkMemFree(NULL, (void**)&tmp_request);
    }

    if (tmp_comment)
    {
        zkMemFree(NULL, (void**)&tmp_comment);
    }

    if (tmp_ticket_id)
    {
        zkMemFree(NULL, (void**)&tmp_ticket_id);
    }

    if (tmp_trans_id)
    {
        zkMemFree(NULL, (void**)&tmp_trans_id);
    }

    if (tmp_res_text)
    {
        zkMemFree(NULL, (void**)&tmp_res_text);
    }

    return (0);
}


/* --------------------------------------------------------------------------
 * zkaudit_ParseQueryStr
 * --------------------------------------------------------------------------
 *
 *
 * @param err the error logging/handling tool
 * @param query_str the search string for which to generate a audit key
 * @param audit_entry_ptr the output structure
 */
int
zkaudit_ParseQueryStr(
    ZK_INOUT  ZkError            err,
    ZK_IN     const char *       query_str,
    ZK_OUT    ZkAuditEntry *     audit_entry_ptr )
{
    static const char * func = "zkaudit_ParseQueryStr";

    if (!(query_str && audit_entry_ptr))
    {
        zkErrorSet(err,func,ZKERR,ZKMAUDIT,ZKENULLPTR,
                   "input params are null");
    }
    else
    {
        zkErrorSet(err,func,ZKERR,ZKMAUDIT,ZKESTUB,
                   "not yet implemented");
    }
    
    return 0;
}


/* --------------------------------------------------------------------------
 * zkaudit_EntryCmp
 * --------------------------------------------------------------------------
 *
 *
 */
int
zkaudit_EntryCmp(
    ZK_INOUT  ZkError            err,
    ZK_IN     ZkAuditEntry       key,
    ZK_IN     ZkAuditEntry       entry,
    ZK_OUT    ZkBool*            match )
{
    static const char * func = "zkaudit_EntryCmp";

    int is_ok = 0;

    if (!(key && entry && match))
    {
        zkErrorSet(err,func,ZKERR,ZKMAUDIT,ZKENULLPTR,
                   "input params are null");
    }
    else
    {
        (*match) = ((key->session_id==0) || (key->session_id==entry->session_id))
            && (strncasecmp(key->user,entry->user,strlen(key->user))==0)
            && (strncasecmp(key->host,entry->host,strlen(key->host))==0)
            && (strncasecmp(key->service,entry->service,strlen(key->service))==0)
            && (strncasecmp(key->request,entry->request,strlen(key->request))==0)
            && (strncasecmp(key->ticket_id,entry->ticket_id,strlen(key->ticket_id))==0)
            && (strncasecmp(key->client_id,entry->client_id,strlen(key->client_id))==0)
            && (strncasecmp(key->transaction_id,entry->transaction_id,strlen(key->transaction_id))==0);

        is_ok = 1;
    }
    
    return (is_ok);
}



/* --------------------------------------------------------------------------
 * zkaudit_FopenLocked
 * --------------------------------------------------------------------------
 *
 * Opens a file in locked mode
 *
 * @param err the error logging/handling tool
 * @param filename the file to be opened
 * @mode the manner in which the file should be opened (eg. "r", "w", "r+")
 * @fp the output file pointer
 */
int
zkaudit_FopenLocked(
    ZkError           err,
    const char *      filename,
    const char *      mode,
    FILE * *          file_ptr )
{
    int result;


    zkErrorClear(err);

    do
    {
        result = zkFileFopenLocked(err, filename, mode, file_ptr);
    }
    while ((result == ZKWARN) && (errno == EAGAIN));

    return (result);
}


/* --------------------------------------------------------------------------
 * zkaudit_ConvertToBase64
 * --------------------------------------------------------------------------
 *
 * Converts a string to base64.  Allocates a buffer for the resulting string.
 */
int
zkaudit_ConvertToBase64(
  ZkError             err,
  const char *        source,
  char **             dest )
{
    size_t      source_len = strlen(source);
    size_t      dest_len = ((source_len+2) * 4) / 3;

    zkErrorClear(err);

    *dest = NULL;

    if ( zkMemAlloc(err, (void**)dest, dest_len + 1)
         && zkConvertToBase64(err, source, *dest, source_len) )
    {
        return (1);
    }

    if (*dest)
    {
        free(*dest);
        *dest = NULL;
    }

    return (0);
}

/* --------------------------------------------------------------------------
 * zkaudit_ConvertFromBase64
 * --------------------------------------------------------------------------
 *
 * Converts a string from base64.  Allocates a buffer for the resulting string.
 */
int
zkaudit_ConvertFromBase64(
  ZkError             err,
  const char *        source,
  char **             dest )
{
    size_t      source_len = strlen(source);
    size_t      dest_len = ((source_len+2) * 3) / 4;
    size_t      rtnLen;

    zkErrorClear(err);

    *dest = NULL;

    if ( zkMemAlloc(err, (void**)dest, dest_len + 1)
         && zkConvertFromBase64(err, source, *dest, &rtnLen) )
    {
        if(rtnLen <= dest_len) {
            return 1;
        }
    }

    if (*dest)
    {
        free(*dest);
        *dest = NULL;
    }

    return 0;
}


/* --------------------------------------------------------------------------
 * Format Strings
 * --------------------------------------------------------------------------
 *
 * These constants are used to control the printing and parsing of audit
 * data to the database file.
 *
 */

# define FIELD_DATE                     "%04d/%02d/%02d"
# define FIELD_TIME                     "%02d:%02d:%02d"
# define FIELD_USER                     "%s"
# define FIELD_HOST                     "%s"
# define FIELD_SESSION_ID               "0x%08X"
# define FIELD_SERVICE                  "%s"
# define FIELD_REQUEST_64               "%s"
# define FIELD_COMMENT_64               "%s"
# define FIELD_TICKET_ID_64             "%s"
# define FIELD_CLIENT_ID_64             "%s"
# define FIELD_TRANSACTION_ID_64        "%s"
# define FIELD_RESULT_CODE              "0x%08X"
# define FIELD_RESULT_TEXT_64           "%s"

/* some format need to be redefined for reading */
# define FIELD_READ_DATE                "%04d/%02d/%02d"
# define FIELD_READ_TIME                "%02d:%02d:%02d"
# define FIELD_READ_SESSION_ID          "0x%X"
# define FIELD_READ_RESULT_CODE         "0x%X"


# define DELIMITER                      "\t"
# define TERMINATOR                     "\015\012"

/* hope an entry will never get THAT big ... */
# define MAX_ENTRY_LENGTH               128*1024


/* --------------------------------------------------------------------------
 * zkaudit_WriteEntry
 * --------------------------------------------------------------------------
 *
 *
 */
int
zkaudit_WriteEntry(
    ZkError             err,
    ZkAuditSession      audit_session,
    ZkAuditEntry        audit_entry )
{
    static const char * func                  = "zkaudit_WriteEntry";
    char *              base64_request        = NULL;
    char *              base64_comment        = NULL;
    char *              base64_ticket_id      = NULL;
    char *              base64_client_id      = NULL;
    char *              base64_transaction_id = NULL;
    char *              base64_result_text    = NULL;
    FILE *              fp                    = NULL;
    int                 is_successful         = 0;

    zkErrorClear(err);

    if (!(audit_session && audit_entry && audit_session->audit_log))
    {
        zkErrorSet(err,func,ZKERR,ZKMAUDIT,ZKENULLPTR,
                   "input params are null");
    } else
        if ( zkaudit_ConvertToBase64(err,audit_entry->request,&base64_request)
             && zkaudit_ConvertToBase64(err,audit_entry->comment,&base64_comment)
             && zkaudit_ConvertToBase64(err,audit_entry->ticket_id,&base64_ticket_id)
             && zkaudit_ConvertToBase64(err,audit_entry->client_id,&base64_client_id)
             && zkaudit_ConvertToBase64(err,audit_entry->transaction_id,&base64_transaction_id)
             && zkaudit_ConvertToBase64(err,audit_entry->result_text,&base64_result_text)
             && zkaudit_FopenLocked(err,audit_session->audit_log->audit_file,"a",&fp))
    {
        fprintf( fp,
                 FIELD_DATE                DELIMITER \
                 FIELD_TIME                DELIMITER \
                 FIELD_USER                DELIMITER \
                 FIELD_HOST                DELIMITER \
                 FIELD_SESSION_ID          DELIMITER \
                 FIELD_SERVICE             DELIMITER \
                 FIELD_REQUEST_64          DELIMITER \
                 FIELD_COMMENT_64          DELIMITER \
                 FIELD_TICKET_ID_64        DELIMITER \
                 FIELD_CLIENT_ID_64        DELIMITER \
                 FIELD_TRANSACTION_ID_64   DELIMITER \
                 FIELD_RESULT_CODE         DELIMITER \
                 FIELD_RESULT_TEXT_64      TERMINATOR,

                 /* values for FIELD_DATE */
                 audit_entry->timestamp.date.year,
                 audit_entry->timestamp.date.month,
                 audit_entry->timestamp.date.day,

                 /* values for FIELD_TIME */
                 audit_entry->timestamp.time.hour,
                 audit_entry->timestamp.time.minute,
                 audit_entry->timestamp.time.second,

                 /* all the other fields */
                 audit_entry->user,
                 audit_entry->host,
                 audit_entry->session_id,
                 audit_entry->service,
                 base64_request,
                 base64_comment,
                 base64_ticket_id,
                 base64_client_id,
                 base64_transaction_id,
                 audit_entry->result_code,
                 base64_result_text );

        fclose(fp);

        is_successful = 1;
    }

    if (base64_request)
    {
        zkMemFree(NULL,(void**)&base64_request);
    }

    if (base64_comment)
    {
        zkMemFree(NULL,(void**)&base64_comment);
    }

    if (base64_ticket_id)
    {
        zkMemFree(NULL,(void**)&base64_ticket_id);
    }

    if (base64_client_id)
    {
        zkMemFree(NULL,(void**)&base64_client_id);
    }

    if (base64_transaction_id)
    {
        zkMemFree(NULL,(void**)&base64_transaction_id);
    }

    if (base64_result_text)
    {
        zkMemFree(NULL,(void**)&base64_result_text);
    }

    return (is_successful);
}



/* --------------------------------------------------------------------------
 * zkaudit_ReadEntry
 * --------------------------------------------------------------------------
 *
 *
 */
int
zkaudit_ReadEntry(
    ZkError             err,
    FILE *              fp,
    ZkAuditEntry *      audit_entry)
{
    static const char * func = "zkaudit_ReadEntry";

    ZkDateTime timestamp;
    char *     entry_buffer    = NULL;
    char *     field           = NULL;
    uint32_t   readSessionId   = 0;
    uint32_t   readResultCode  = 0;
    char *     user         = NULL;
    char *     host         = NULL;
    char *     service      = NULL;
    char *     request      = NULL;
    char *     comment      = NULL;
    char *     ticket       = NULL;
    char *     client       = NULL;
    char *     trans        = NULL;
    char *     result       = NULL;
    int        is_successful   = 0;


    zkErrorClear(err);

    if (!audit_entry || !fp)
    {
        zkErrorSet(err,func,ZKERR,ZKMAUDIT,ZKENULLPTR,
                   "input params are null");
    }
    else if ( !zkMemAlloc(err, (void **)&entry_buffer, MAX_ENTRY_LENGTH) )
    {
        /*
         * the allocation failed!  zkMemAlloc takes care of setting the
         * error status properly for us ... so nothing further to do here
         */
    }
    else if (fgets(entry_buffer, MAX_ENTRY_LENGTH, fp) == NULL)
    {
        zkErrorSet(err,func,ZKERR,ZKMAUDIT,ZKERROR,"file input failed");
    }
    else
    {
        int is_good = 1;

        /*
         * parse the date portion of the timestamp
         */
        
        if (is_good)
        {
            if ((field = strtok(entry_buffer, "\t")) == NULL)
            {
                is_good = 0;
            }
            else
            {
                int year=0, month=0, day=0;

                sscanf(field, FIELD_READ_DATE, &year, &month, &day); /* audit: ok */
                
                timestamp.date.year  = year;
                timestamp.date.month = month;
                timestamp.date.day   = day;
            }
        }
        
        /*
         * parse the time portion of the timestamp
         */
        
        if (is_good)
        {
            if ((field = strtok(NULL, "\t")) == NULL)
            {
                is_good = 0;
            }
            else
            {
                int hour=0, minute=0, second=0;

                sscanf(field, FIELD_READ_TIME, &hour, &minute, &second); /* audit: ok */
                
                timestamp.time.hour    = hour;
                timestamp.time.minute  = minute;
                timestamp.time.second  = second;
            }
        }
        
        /* user */
        if (is_good)
        {
            user = strtok(NULL, "\t");
            is_good = (user != NULL);
        }
        
        /* host */
        if (is_good)
        {
            host = strtok(NULL, "\t");
            is_good = (host != NULL);
        }
        
        /* session id */
        if (is_good)
        {
            if ((field = strtok(NULL, "\t")) == NULL)
            {
                is_good = 0;
            }
            else
            {
                sscanf(field, FIELD_READ_SESSION_ID, &readSessionId); /* audit: ok */
            }
        }
        
        /* service */
        if (is_good)
        {
            service = strtok(NULL, "\t");
            is_good = (service != NULL);
        }
        
        /* request */
        if (is_good)
        {
            if ((field = strtok(NULL, "\t")) == NULL)
            {
                is_good = 0;
            }
            else
            {
                zkaudit_ConvertFromBase64(err, field, (char**)(&request));
            }
        }
        
        /* comment */
        if (is_good)
        {
            if ((field = strtok(NULL, "\t")) == NULL)
            {
                is_good = 0;
            }
            else
            {
                zkaudit_ConvertFromBase64(err, field, (char**)(&comment));
            }
        }
        
        /* ticket id */
        if (is_good)
        {
            if ((field = strtok(NULL, "\t")) == NULL)
            {
                is_good = 0;
            }
            else
            {
                zkaudit_ConvertFromBase64(err, field, (char**)(&ticket));
            }
        }
        
        /* client id */
        if (is_good)
        {
            if ((field = strtok(NULL, "\t")) == NULL)
            {
                is_good = 0;
            }
            else
            {
                zkaudit_ConvertFromBase64(err, field, (char**)(&client));
            }
        }
        
        /* transaction id */
        if (is_good)
        {
            if ((field = strtok(NULL, "\t")) == NULL)
            {
                is_good = 0;
            }
            else
            {
                zkaudit_ConvertFromBase64(err, field, (char**)(&trans));
            }
        }
        
        /* result code */
        if (is_good)
        {
            if ((field = strtok(NULL, "\t")) == NULL)
            {
                is_good = 0;
            }
            else
            {
                sscanf(field, FIELD_READ_RESULT_CODE, &readResultCode); /* audit: ok */
            }
        }
        
        /* result text */
        if (is_good)
        {
            if ((field = strtok(NULL, "\t")) == NULL)
            {
                is_good = 0;
            }
            else
            {
                zkaudit_ConvertFromBase64(err, field, (char**)(&result));
            }
        }
        
        
        /* create the entry */
        if (is_good)
        {
            is_successful = zkaudit_EntryCreate( err,&timestamp,user,
                                                 host,readSessionId,
                                                 service,request,
                                                 comment,ticket,
                                                 client,trans,
                                                 readResultCode,result,
                                                 audit_entry );
        }
    }

    /*
     * Clean up memory allocated for base64-to-text translation buffers
     * and for the read buffer
     */

    if (request)
    {
        zkMemFree(NULL,(void**)&request);
    }

    if (comment)
    {
        zkMemFree(NULL,(void**)&comment);
    }

    if (ticket)
    {
        zkMemFree(NULL,(void**)&ticket);
    }

    if (client)
    {
        zkMemFree(NULL,(void**)&client);
    }

    if (trans)
    {
        zkMemFree(NULL,(void**)&trans);
    }

    if (result)
    {
        zkMemFree(NULL,(void**)&result);
    }

    if (entry_buffer)
    {
        zkMemFree(NULL,(void**)&entry_buffer);
    }

    /*
     * Lets go home ...
     */

    return (is_successful);
}



/* --------------------------------------------------------------------------
 * zkaudit_GetHostName
 * --------------------------------------------------------------------------
 *
 *
 */
const char *
zkaudit_GetHostName(void)
{
    static char hostname[256];
    static int  retrieved = 0;

    /*
     * XXX - This is NOT thread safe!
     */

    if (retrieved == 0)
    {
         gethostname(hostname, sizeof(hostname));
         retrieved = 1;
    }

    return (hostname);
}



/* --------------------------------------------------------------------------
 * zkaudit_GetNewSessionId
 * --------------------------------------------------------------------------
 *
 *
 */
int
zkaudit_GetNewSessionId(
    ZkError           err,
    ZkAuditLog        audit_log,
    uint32_t *        sid_ptr )
{
    static const char * func = "zkaudit_GetNewSessionId";
    FILE * fp = NULL;


    zkErrorClear(err);

    if ( !(audit_log && audit_log->sid_file) )
    {
        zkErrorSet(err,func,ZKERR,ZKMAUDIT,ZKENULLPTR,
                   "input params are null");
    }
    else if ( zkaudit_FopenLocked(err, audit_log->sid_file, "r+", &fp) )
    {
        rewind(fp);
        fread(sid_ptr, sizeof(*sid_ptr), 1, fp);
        ++(*sid_ptr);
        rewind(fp);
        fwrite(sid_ptr, sizeof(*sid_ptr), 1, fp);
        fclose(fp);

        return (1);
    }

    return (0);
}
