/*
** **************************************************************************
** $Id: zkAuditSession.c,v 1.2.30.1 2000/10/11 13:20:08 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

**
** Copyright (C) 1999 Zero-Knowledge Systems Inc.  All Rights Reserved
** **************************************************************************
*/

# include <zks.h>
# include <zkAlloc.h>
# include <zkAudit.h>
# include "zkAuditHelper.h"



/* --------------------------------------------------------------------------
 * zkAuditSessionBegin
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkAuditSessionBegin(
    ZkError          err,
    ZkAuditLog       audit_log,
    const char *     user,
    const char *     ticket_id,        /* may be NULL */
    const char *     client_id,        /* may be NULL */
    const char *     trans_id,         /* may be NULL */
    ZkAuditSession * audit_session_ptr )
{
    static const char * func = "zkAuditSessionBegin";
    ZkAuditSession   new_session = NULL;
    const char *     tmp_user    = NULL;
    const char *     tmp_ticket  = NULL;
    const char *     tmp_client  = NULL;
    const char *     tmp_trans   = NULL;
    uint32_t         tmp_sid     = 0;


    zkErrorClear(err);

    if ( !audit_log || !user )
    {
        zkErrorSet(err,func,ZKERR,ZKMAUDIT,ZKENULLPTR,
                   "input params are null");
    }
    else
    {
        if ( zkMemAlloc(err,(void**)&new_session,sizeof(*new_session))
             && zkMemDup(err,(void**)&tmp_user,user,strlen(user)+1)
             && zkaudit_GetNewSessionId(err, audit_log, &tmp_sid)
             && ( !ticket_id
                  || zkMemDup(err,(void**)&tmp_ticket,ticket_id,strlen(ticket_id)+1) )
             && ( !client_id
                  || zkMemDup(err,(void**)&tmp_client,client_id,strlen(client_id)+1) )
             && ( !trans_id
                  || zkMemDup(err,(void**)&tmp_trans,trans_id,strlen(trans_id)+1) ) )
        {
            new_session->audit_log      = audit_log;
            new_session->session_id     = tmp_sid;
            new_session->user           = tmp_user;
            new_session->host           = zkaudit_GetHostName();
            new_session->ticket_id      = tmp_ticket;
            new_session->client_id      = tmp_client;
            new_session->transaction_id = tmp_trans;

            (*audit_session_ptr) = new_session;

            return (1);
        }
    }

    /*
     * If we reach this point, then an error has occurred. Let's
     * clean up and return a failure.  Note that we're using a
     * static array for the host name ... so it doesn't have to
     * be freed.
     */

    if (tmp_user)
    {
        zkMemFree(err, (void**)&tmp_user);
    }

    if (tmp_ticket)
    {
        zkMemFree(err, (void**)&tmp_ticket);
    }

    if (tmp_client)
    {
        zkMemFree(err, (void**)&tmp_client);
    }

    if (tmp_trans)
    {
        zkMemFree(err, (void**)&tmp_trans);
    }

    if (new_session)
    {
        zkMemFree(err, (void**)&new_session);
    }

    return (0);
}



/* --------------------------------------------------------------------------
 * zkAuditSessionEnd
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkAuditSessionEnd(
    ZkError          err,
    ZkAuditSession * audit_session_ptr )
{
    static const char * func = "zkAuditSessionEnd";


    zkErrorClear(err);

    if ( !(audit_session_ptr && *audit_session_ptr) )
    {
        zkErrorSet(err,func,ZKERR,ZKMAUDIT,ZKENULLPTR,
                   "input params are null");
        return (0);
    }

    /*
     * Note that the audit log may be shared by multiple sessions and the
     * host name is a static buffer.  Neither of these fields should be
     * freed here.
     */

    zkMemFree(err, (void**)&((*audit_session_ptr)->user));
    zkMemFree(err, (void**)&((*audit_session_ptr)->ticket_id));
    zkMemFree(err, (void**)&((*audit_session_ptr)->client_id));
    zkMemFree(err, (void**)&((*audit_session_ptr)->transaction_id));

    zkMemFree(err, (void**)audit_session_ptr);

    return (1);
}



/* --------------------------------------------------------------------------
 * zkAuditSessionAddEntry
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkAuditSessionAddEntry(
    ZkError          err,
    ZkAuditSession   audit_session,
    const char *     service,
    const char *     request,
    const char *     comment,           /* may be NULL */
    uint32_t         result_code,
    const char *     result_text,       /* may be NULL */
    ZkAuditEntry *   audit_entry_ptr )  /* may be NULL */
{
    static const char * func = "zkAuditSessionAddEntry";
    ZkAuditEntry        new_entry = NULL;

    zkErrorClear(err);

    /*
     * Make sure that all the required params are non-NULL
     */
    if (!audit_session || !service || !request)
    {
        zkErrorSet(err,func,ZKERR,ZKMAUDIT,ZKENULLPTR,
                   "input params are null");
    }
    else
    {
        /*
         * These all point into the session structure and should
         * not be freed.
         */

	ZkDateTime timeStamp;

        const char * tmp_user    = NULL;
        const char * tmp_host    = NULL;
        const char * tmp_ticket  = NULL;
        const char * tmp_client  = NULL;
        const char * tmp_trans   = NULL;
        uint32_t     tmp_sid     = 0;

        if ( zkDateTimeSetCurrent (err, &timeStamp, ZKTIME_FL_UTC)
	     && zkAuditSessionGetUser(err,audit_session,&tmp_user)
             && zkAuditSessionGetHost(err,audit_session,&tmp_host)
             && zkAuditSessionGetTicketId(err,audit_session,&tmp_ticket)
             && zkAuditSessionGetClientId(err,audit_session,&tmp_client)
             && zkAuditSessionGetTransactionId(err,audit_session,&tmp_trans)
             && zkAuditSessionGetSessionId(err,audit_session,&tmp_sid)
             && zkaudit_EntryCreate(err,&timeStamp, tmp_user,tmp_host,tmp_sid,service,
                                   request,comment,tmp_ticket,tmp_client,
                                   tmp_trans,result_code,result_text,
                                   &new_entry)
             && zkaudit_WriteEntry(err,audit_session,new_entry))
        {
            if (audit_entry_ptr)
            {
                (*audit_entry_ptr) = new_entry;
            }
            else
            {
                zkAuditEntryFree(err, &new_entry);
            }

            return (1);
        }
    }

    return (0);
}



/* --------------------------------------------------------------------------
 * zkAuditSessionGetSessionId
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkAuditSessionGetSessionId(
    ZkError             err,
    ZkAuditSession      audit_session,
    uint32_t*           session_id_ptr )
{
    static const char * func = "zkAuditSessionGetSessionId";


    zkErrorClear(err);

    if ( !audit_session || !session_id_ptr )
    {
        zkErrorSet(err,func,ZKERR,ZKMAUDIT,ZKENULLPTR,
                   "input params are null");
        return (0);
    }

    *session_id_ptr = audit_session->session_id;

    return (1);
}



/* --------------------------------------------------------------------------
 * zkAuditSessionGetHost
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkAuditSessionGetHost(
    ZkError             err,
    ZkAuditSession      audit_session,
    const char **       host_ptr )
{
    static const char * func = "zkAuditSessionGetHost";


    zkErrorClear(err);

    if ( !audit_session || !host_ptr )
    {
        zkErrorSet(err,func,ZKERR,ZKMAUDIT,ZKENULLPTR,
                   "input params are null");
        return (0);
    }

    *host_ptr = audit_session->host;

    return (1);
}



/* --------------------------------------------------------------------------
 * zkAuditSessionGetUser
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkAuditSessionGetUser(
    ZkError             err,
    ZkAuditSession      audit_session,
    const char **       user_ptr )
{
    static const char * func = "zkAuditSessionGetUser";


    zkErrorClear(err);

    if ( !audit_session || !user_ptr )
    {
        zkErrorSet(err,func,ZKERR,ZKMAUDIT,ZKENULLPTR,
                   "input params are null");
        return (0);
    }

    *user_ptr = audit_session->user;

    return (1);
}



/* --------------------------------------------------------------------------
 * zkAuditSessionGetTicketId
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkAuditSessionGetTicketId(
    ZkError             err,
    ZkAuditSession      audit_session,
    const char **       ticket_id_ptr )
{
    static const char * func = "zkAuditSessionGetTicketId";


    zkErrorClear(err);

    if ( !audit_session || !ticket_id_ptr )
    {
        zkErrorSet(err,func,ZKERR,ZKMAUDIT,ZKENULLPTR,
                   "input params are null");
        return (0);
    }

    *ticket_id_ptr = audit_session->ticket_id;

    return (1);
}



/* --------------------------------------------------------------------------
 * zkAuditSessionGetClientId
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkAuditSessionGetClientId(
    ZkError             err,
    ZkAuditSession      audit_session,
    const char **       client_id_ptr )
{
    static const char * func = "zkAuditSessionGetClientId";


    zkErrorClear(err);

    if ( !audit_session || !client_id_ptr )
    {
        zkErrorSet(err,func,ZKERR,ZKMAUDIT,ZKENULLPTR,
                   "input params are null");
        return (0);
    }

    *client_id_ptr = audit_session->client_id;

    return (1);
}



/* --------------------------------------------------------------------------
 * zkAuditSessionGetTransactionId
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkAuditSessionGetTransactionId(
    ZkError             err,
    ZkAuditSession      audit_session,
    const char **       transaction_id_ptr )
{
    static const char * func = "zkAuditSessionGetTransactionId";


    zkErrorClear(err);

    if ( !audit_session || !transaction_id_ptr )
    {
        zkErrorSet(err,func,ZKERR,ZKMAUDIT,ZKENULLPTR,
                   "input params are null");
        return (0);
    }

    *transaction_id_ptr = audit_session->transaction_id;

    return (1);
}

