/*
 * $Id: zkauth.c,v 1.8.10.4 2000/12/14 17:54:31 jean Exp $
 */

static const char * module_id_ = "$Id: zkauth.c,v 1.8.10.4 2000/12/14 17:54:31 jean Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include "zkIntTypes.h"
#include "zkTypes.h"
#ifndef WIN32
#define _XOPEN_SOURCE
#include <unistd.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <time.h>
#include "zkrand.h"
#include "zkString.h"
#include "zkAlloc.h"


#include "zkConfig.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkCStr.h"

#include "zkauth.h"


/*
** private structure
*/
typedef struct t_ZkAuthContext{
    char *  dbFilename;
    char *  lockFilename;
} T_ZkAuthContext;

typedef struct t_ZkAuthSession{
    char *            logUsername;
    T_ZkAuthContext * context;
} T_ZkAuthSession;


/*
** private functions prototypes
*/
static int
validatePassphrase(
		   const char * refPassphrase,
		   const char * passphrase
);

static int
operationInList(
		const char *operList,
		const char *reqOperation
);

static int
generateSaltString(
                   ZkError err,
                   char * salt
);

static int
zkauth_lockLoadConfig(
		      ZkError   err,
		      char *    lockFilename,
		      char *    configFilename,
		      ZkConfig  config
);

static int
zkauth_releaseFreeConfig(
			 ZkError    err,
			 char *     lockFilename,
			 ZkConfig * config
);

static int
generateUniqueFilename(
		       const char * lockFile,
		       char **      uniqueFile
);

static int
createLock(
	   char *lockFile,
	   char *uniqueFile
);

static int
removeLock(
	   char *lockFile
);


/*
**------------------------------------------------------------------------------
** Debugging interfaces
**------------------------------------------------------------------------------
*/

/*
 *  The sole purpose of this function is to make generateSaltString()
 *  available for testing.
 */
int
zkAuth_generateSaltString(ZkError err, char *salt)
{
  return generateSaltString(err, salt);
}


/*
**------------------------------------------------------------------------------
** Init operation
**------------------------------------------------------------------------------
*/
DLLExport int
zkAuthCreate(
	     ZkError          err,
	     const char *     dbFilename,
	     ZkAuthContext *  authContext
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr;

    struct stat fileStat;

    char *localDbFilename     = NULL;
    char *localLockFilename   = NULL;


    (void) module_id_; /* to avoid warning */


    zkErrorClear(err);

    /*
    ** param validation
    */
    if(!authContext || !dbFilename) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }

    /*
    ** verify the existence of the passphrase file
    */
    if(errCode == ZKENOERROR &&
       stat(dbFilename, &fileStat) != 0) {
	errCode = ZKERROR;
	errStr = "passphrase file does not exist";
    }

    /*
    ** allocate a context instance and get info to fill it
    */
    if(errCode == ZKENOERROR &&
       zkMemAlloc(err, (void**)authContext, sizeof(T_ZkAuthContext)) ) {

	memset(*authContext,0,sizeof(T_ZkAuthContext));

	localDbFilename = malloc(strlen(dbFilename)+1); /* bug: audit -- not checked */
	zkCStrlcpy(localDbFilename, dbFilename, strlen(dbFilename)+1);

	generateUniqueFilename(dbFilename, &localLockFilename);
    }

    if(errCode != ZKENOERROR){
	rtnCode = 0;
	zkErrorSet(err, "zkAuthCreate", ZKERR, ZKMAUTH, errCode, errStr);

	/* memory management */
	if(localDbFilename)     free(localDbFilename);
	if(localLockFilename)   free(localLockFilename);
	if(*authContext != NULL){
	    zkMemFree(err, (void**)authContext);
	    (*authContext) = NULL;
	}
    } else {
	/* update authContext structure */
	(*authContext)->dbFilename     = localDbFilename;
	(*authContext)->lockFilename   = localLockFilename;
    }

    return rtnCode;
}


DLLExport int
zkAuthFree(
	   ZkError         err,
	   ZkAuthContext * authContext
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr;


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!authContext || !(*authContext)) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }

    if(errCode == ZKENOERROR) {

	if((*authContext)->dbFilename) free((*authContext)->dbFilename);
	if((*authContext)->lockFilename) free((*authContext)->lockFilename);

	zkMemFree(err, (void**)authContext);
	(*authContext) = NULL;
    }


    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, "zkAuthFree", ZKERR, ZKMAUTH, errCode, errStr);
    }

    return rtnCode;
}


/*
**------------------------------------------------------------------------------
** Authentication operation
**------------------------------------------------------------------------------
*/

DLLExport int
zkAuthLogin(
	    ZkError         err,
	    ZkAuthContext   authContext,
	    const char *    username,
	    const char *    userPassphrase,
	    ZkAuthSession* authSessionPtr
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr;

    ZkConfig  config = NULL;
    char *localUsername = NULL;
    const char *dbPassphrase = NULL;

    ZkBool validUser = ZK_FALSE;


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!authContext || !username || !userPassphrase || !authSessionPtr){
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }

    /*
    ** get/generate data for session instance
    */
    if(errCode == ZKENOERROR &&
       zkMemAlloc(err, (void**)authSessionPtr, sizeof(T_ZkAuthSession))) {

      zkMemAlloc(err, (void**)&localUsername, strlen(username)+1); /* bug: audit -- not checked */
      zkCStrlcpy(localUsername, username, strlen(username)+1);

    }
    
    if(errCode == ZKENOERROR) {
	
	if(zkConfigNew(err, &config) == 0) {

	    errCode = ZKERROR;
	    errStr = "problem with zkConfigNew";

	} else {

	    /* lock and load config file */
	    if(zkauth_lockLoadConfig(err,
				     authContext->lockFilename,
				     authContext->dbFilename,
				     config) == 0) {
		errCode = ZKERROR;
		errStr = "can't load passphrase file";
	    } else {

		/* validate username and passphrase */
		if(zkConfigGet(err, config, username, ZKAUTH_TOKEN_PASSPHRASE, (const char**)&dbPassphrase) &&
		   validatePassphrase(dbPassphrase, userPassphrase)) {
		    validUser = ZK_TRUE;
		}

		/* release and free config */
		zkauth_releaseFreeConfig(err,
					 authContext->lockFilename,
					 &config);
	    }
	}
    }


    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, "zkAuthLogin", ZKERR, ZKMAUTH, errCode, errStr);
    }

    if(!validUser) {
	/* memory management */
	if(localUsername)       free(localUsername);
	if(config)              zkConfigFree(err, &config);
	if(*authSessionPtr)     zkMemFree(err, (void**)authSessionPtr);
	*authSessionPtr = NULL;

    } else {
	/* fill the session instance */
	(*authSessionPtr)->logUsername  = localUsername;
	(*authSessionPtr)->context      = authContext;
    }

    return rtnCode;
}


DLLExport int
zkAuthCheckAccess(
		  ZkError       err,
		  ZkAuthSession authSession,
		  const char *  reqService,
		  const char *  reqOperation,
		  ZkBool *      accessGranted
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr;

    ZkAuthContext  context;
    ZkConfig       config;

    const char *availOperations;


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!authSession || !reqService || !reqOperation){
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }

    /*
    ** proceed with the request
    */
    if(errCode == ZKENOERROR){

	context = authSession->context;

	if(zkConfigNew(err, &config) == 0) {

	    errCode = ZKERROR;
	    errStr = "problem with zkConfigNew";

	} else {

	    /* lock and load config file */
	    if(zkauth_lockLoadConfig(err,
				     context->lockFilename,
				     context->dbFilename,
				     config) == 0) {
		errCode = ZKERROR;
		errStr = "can't load passphrase file";
	    } else {
	    
		/* check to see if access to service/operation is granted */
		if(zkConfigGet(err, config,
			       authSession->logUsername,
			       (const char*)reqService,
			       (const char**)&availOperations) &&
		   operationInList(availOperations, reqOperation)) {
	    
		    *accessGranted = ZK_TRUE;
		} else {
		    *accessGranted = ZK_FALSE;
		}

		/* release and free config */
		zkauth_releaseFreeConfig(err,
					 context->lockFilename,
					 &config);
	    }
	}
    }


    if(errCode != ZKENOERROR){
	rtnCode = 0;
	zkErrorSet(err, "zkAuthCheckAccess", ZKERR, ZKMAUTH, errCode, errStr);
    }

    return rtnCode;
}


DLLExport int
zkAuthLogout(
	     ZkError         err,
	     ZkAuthSession * authSession
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr;


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!authSession || !(*authSession)){
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }

    /*
    ** proceed with the request
    */
    if(errCode == ZKENOERROR) {

	if((*authSession)->logUsername) {
	    free((*authSession)->logUsername);
	    (*authSession)->logUsername = NULL;
	}

	zkMemFree(err, (void**)authSession);
	*authSession = NULL;
    }

    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, "zkAuthLogout", ZKERR, ZKMAUTH, errCode, errStr);
    }

    return rtnCode;
}


/*
**------------------------------------------------------------------------------
** Management operation
**------------------------------------------------------------------------------
*/

DLLExport int
zkAuthChangePassphrase(
		       ZkError        err,
		       ZkAuthSession  authSession,
		       char *	    newPassphrase
)
{
  int rtnCode = 1;
  unsigned short errCode = ZKENOERROR;
  const char *errStr;

  ZkConfig config;

  ZkAuthContext context;
    
  char *salt;
  char *crypted;


  zkErrorClear(err);
  /*
  ** param validation
  */
  if(!authSession || !newPassphrase) {
    errCode = ZKENULLPTR;
    errStr = "invalid parameters";
  }

  /*
  ** proceed with the request
  */
  if(errCode == ZKENOERROR) {
	
    context = authSession->context;

    salt = (char*)malloc(2);

    /* generate a random salt string */
    if ( ! generateSaltString(err, salt)) {

	    errCode = ZKERROR;
	    errStr = "problem with generateSaltString";

    } else {
      /* encrypt new passphrase with it */
      crypted = crypt(newPassphrase, salt);


      /*
      ** update config file with new passphrase
      */
      if(zkConfigNew(err, &config) == 0) {

        errCode = ZKERROR;
        errStr = "problem with zkConfigNew";

      } else {

        /* lock and load config file */
        if(zkauth_lockLoadConfig(err,
                                 context->lockFilename,
                                 context->dbFilename,
                                 config) == 0) {
          errCode = ZKERROR;
          errStr = "can't load passphrase file";
        } else {

          /* update field in config data */
          if(zkConfigReset(err,
                           config,
                           authSession->logUsername,
                           ZKAUTH_TOKEN_PASSPHRASE,
                           crypted) == 0) {
            errCode = ZKERROR;
            errStr = "error with zkConfigReset";
          }
	       
          /* save it */
          if(errCode == ZKENOERROR &&
             zkConfigSave(err,
                          config,
                          context->dbFilename) == 0) {
            /* remap to new convention */
            errCode = ZKERROR;
            errStr = "error with zkConfigSave";
          }

          /* release and free config */
          zkauth_releaseFreeConfig(err,
                                   context->lockFilename,
                                   &config);
        }
      }
    }

    /* memory management */
    free(salt);
  }

  if(errCode != ZKENOERROR) {
    rtnCode = 0;
    zkErrorSet(err, "zkAuthChangePassphrase", ZKERR, ZKMAUTH, errCode, errStr);
  }

  return rtnCode;
}


DLLExport int
zkAuthAdd(
	  ZkError        err,
	  ZkAuthSession	 authSession,
	  ZkAuthEntry *  authEntry
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr;


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!authSession || !authEntry) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }


    if(errCode == ZKENOERROR) {

	/* stub it */
	errCode = ZKESTUB;
	errStr = "call to stub function";
    }


    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, "zkAuthAdd", ZKERR, ZKMAUTH, errCode, errStr);
    }

    return rtnCode;
}



DLLExport int
zkAuthDelete(
	     ZkError        err,
	     ZkAuthSession  authSession,
	     ZkAuthEntry *  authEntry
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr;


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!authSession || !authEntry) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }


    if(errCode == ZKENOERROR) {

	/* stub it */
	errCode = ZKESTUB;
	errStr = "call to stub function";
    }

    /*
     * log error if any
     */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, "zkAuthDelete", ZKERR, ZKMAUTH, errCode, errStr);
    }

    return rtnCode;
}


DLLExport int
zkAuthQuery(
	    ZkError        err,
	    ZkAuthSession  authSession,
	    ZkAuthEntry *  authEntry
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr;


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!authSession || !authEntry) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }


    if(errCode == ZKENOERROR) {

	/* stub it */
	errCode = ZKESTUB;
	errStr = "call to stub function";
    }


    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, "zkAuthQuery", ZKERR, ZKMAUTH, errCode, errStr);
    }

    return rtnCode;
}


/*
**------------------------------------------------------------------------------
** private functions
**------------------------------------------------------------------------------
*/

static int
validatePassphrase(
		   const char * refPassphrase,
		   const char * passphrase
)
{
    int rtnCode = 1;

    char *cryptedPassphrase;
    char *salt;
    const char *onlyPassphrase;
    
    ZkBool validated = ZK_FALSE;


    /*
    ** first check if one of the two passphrase is missing
    */
    if(refPassphrase == NULL || refPassphrase[0] == '\0') {
	if(passphrase != NULL && passphrase[0] != '\0') {
	    /* there is a ref passphrase and NO passphrase to compare */
	    rtnCode = 0;
	}
	/* in either cases the validation is done */
	validated = ZK_TRUE;

    } else if(passphrase == NULL || passphrase[0] == '\0') {
	/* there is NO ref passphrase to compare with */
	rtnCode = 0;
	validated = ZK_TRUE;
    }

    /*
    ** if not in one of the trivial cases proceed with encryption
    */
    if(!validated) {

	salt = (char*)malloc(2);

	/* get salt and passphrase string */
	strncpy(salt, refPassphrase, 2);
	onlyPassphrase = &refPassphrase[2];

	/* encrypt passphrase */
	cryptedPassphrase = crypt(passphrase, salt);
	
	/*
	** compare it with refPassphrase
	*/
	if(strcmp(cryptedPassphrase,refPassphrase) != 0) {
	    rtnCode = 0;
	}

	free(salt);
    }

    return rtnCode;
}


static int
operationInList(
		const char *operList,
		const char *reqOperation
)
{
    int rtnCode = 1;

    char *localList;
    char *operationToken;


    /* make local copy of list */
    localList = malloc(strlen(operList)+1); /* bug: audit -- not checked */
    zkCStrlcpy(localList, operList, strlen(operList)+1);

    /*
    ** we assume that operation are separated by white spaces
    */
    operationToken = strtok(localList, " ");

    while(operationToken != NULL){

	if(strcmp(operationToken, reqOperation) == 0) {
	    break;
	}

	operationToken = strtok(NULL, " ");
    }
    
    if(operationToken == NULL){
	rtnCode = 0;
    }

    /* memory management */
    free(localList);

    return rtnCode;
}


static int
generateSaltString(
                   ZkError err,
                   char * salt
)
{
  int i;
  int rv;

  /* init rand with srand */
    
  for(i=0; i<2 ;i++) {
    uint32_t x;

    /* get the random value */

    if ( ! zkRandGetUInt32InRange(err, 0, RAND_MAX, &x)) {
      return 0;
    }
    rv = (int)(64.0*x/(RAND_MAX+1.0));

    /* map it to corresponding character */
    if(rv>=0 && rv<=25) {
	    salt[i] = 'a' + rv;
    } else if(rv>=26 && rv<=51) {
	    salt[i] = 'A' + rv - 26;
    } else if(rv>=52 && rv<=61) {
	    salt[i] = '0' + rv - 52;
    } else if(rv==62){
	    salt[i] = '.';
    } else {
	    salt[i] = '/';
    }
  }

  return 1;
}

static int
zkauth_lockLoadConfig(
		      ZkError   err,
		      char *    lockFilename,
		      char *    configFilename,
		      ZkConfig  config
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr;
	

    /*
    ** create advisory lockfile before reading the config file
    ** (tries for a certain time if already lock)
    */
    if(createLock(configFilename, lockFilename) == 0) {
	errCode = ZKERROR;
	errStr = "can't get lock";
    } else {

	if(zkConfigLoad(err, config, configFilename) == 0) {
	    /* remap to new convention */
	    errCode = ZKERROR;
	    errStr = "could not load passphrase file";
	}
    }

    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, "zkauth_lockLoadConfig", ZKERR, ZKMAUTH, errCode, errStr);
    }

    return rtnCode;
}


static int
zkauth_releaseFreeConfig(
			 ZkError    err,
			 char *     lockFilename,
			 ZkConfig * config
)
{
    int rtnCode = 1;


    if( removeLock(lockFilename) == 0) {
	rtnCode = 0;
	zkErrorSet(err, "zkAuthInit", ZKWARN, ZKMAUTH, ZKERROR, "can't remove unique lockfile");
    }

    zkConfigFree(err, config);

    return rtnCode;
}


static int
generateUniqueFilename(
		       const char * lockFile,
		       char **      uniqueFile
)
{
    pid_t pid;

    static char *tmpString;


    pid = getpid();

    tmpString = (char*)malloc(1024);

    snprintf(tmpString, 1024, "%s.%ld", lockFile, pid);

    (*uniqueFile) = (char*)malloc(strlen(tmpString)+1); /* bug: audit -- not checked */
    zkCStrlcpy((*uniqueFile), tmpString, strlen(tmpString)+1);

    free(tmpString);

    return 1;
}


static int
createLock(
	   char *lockFile,
	   char *uniqueFile
)
{
    int rtnCode = 1;

    int numIter = 5;
    int i;

    ZkBool linkOK = ZK_FALSE;
    ZkBool lockOK = ZK_FALSE;

    struct stat statBuf;


    /* make a link to lockFile */
    for(i=0; i<numIter ;i++){
	if(link(lockFile, uniqueFile) == 0) {
	    linkOK = ZK_TRUE;
	    break;
	}
	sleep(1);
    }

    /* get stat on uniqueFile */
    if(linkOK == ZK_TRUE) {

	for(i=0; i<numIter ;i++){
	    if(stat(uniqueFile, &statBuf) == 0) {
		if(statBuf.st_nlink == 2) {
		    /* lock succeed !!! */
		    lockOK = ZK_TRUE;
		    break;
		}
	    }
	    sleep(1);
	}
    }

    if(lockOK == ZK_FALSE) {
	if(linkOK) {
	    remove(uniqueFile);
	}
	rtnCode = 0;
    }

    return rtnCode;
}


static int
removeLock(
	   char *uniqueFile
)
{
    int rtnCode = 1;

    if( remove(uniqueFile) != 0) {
	rtnCode = 0;
    }

    return rtnCode;
}

