/* **************************************************************************
** $Id: zkcpCommand.c,v 1.5.18.1 2000/10/11 13:20:15 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

** Copyright (C) 2000 Zero-Knowledge Systems Inc.  All rights reserved.
** --------------------------------------------------------------------------
**
** This file contains definitions and declarations for the Zero-Knowledge
** Command Protocol.
**
** **************************************************************************
*/

# include "zkcp_priv.h"
# include "zkAlloc.h"

static const char * const
_module_id = "$Id: zkcpCommand.c,v 1.5.18.1 2000/10/11 13:20:15 phil Exp $";


/* ------------------------------------------------------------------------
 * Contants and Type Definitions
 */

# define ZKCP_PARAM_TABLE_GROW_SIZE (10)

typedef struct T_ZkcpParamTableEntry * ZkcpParamTableEntry;

struct T_ZkcpCommand {
    ZkcpContext                 context;
    ZkcpCommandId               id;
    uint32_t                    param_table_size;
    ZkcpParamTableEntry         param_table;
    uint32_t                    param_count;
    
};

struct T_ZkcpParamTableEntry {
    uint16_t                    id;
    ZkcpValue                   value;
};



/* ------------------------------------------------------------------------
 * Creates a new command object.
 */
int
zkcpCommand_Create(
    ZK_INOUT ZkError            error,
    ZK_IN    ZkcpCommandId      command_id,
    ZK_OUT   ZkcpCommand *      out_command )
{
    const char * const          func    = "zkcpCommand_Create";
    ZkcpCommand                 command = NULL;
    int                         result  = 0;
    
    (void)_module_id; /* compiler complains about unreferenced variable */

    zkErrorClear(error);
    
    if (out_command == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'out_command' is NULL");
    }
    else
    {
        (*out_command) = NULL;
        
        if (!zkMemCalloc(error, (void**)&command, sizeof(*command), 1))
        {
            zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKERROR,
                       "can't allocate mem for command");        
        }
        else if (!zkMemCalloc(error, (void**)&(command->param_table),
                              sizeof(command->param_table[0]),
                              ZKCP_PARAM_TABLE_GROW_SIZE))
        {
            zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKERROR,
                       "can't allocate parameter table");        
        }
        else
        {
            command->id               = command_id;
            command->param_table_size = ZKCP_PARAM_TABLE_GROW_SIZE;
            (*out_command)            = command;
            command                   = NULL;
            result                    = 1;
        }
    }
    
    if (command != NULL)
    {
        /*
         * On success, the newly allocated command is assigned to the
         * out_command variable and command is set to null.  If we get
         * here, then it means that the function was unsuccessful.
         *
         * Clean-up time.
         */

        zkcpCommand_Destroy(NULL, &command);
    }

    return (result);
}



/* ------------------------------------------------------------------------
 * Destroys a command object.
 */
int
zkcpCommand_Destroy(
    ZK_INOUT ZkError            error,
    ZK_OUT   ZkcpCommand *      command )
{
    const char * const func = "zkcpCommand_Destroy";
    int result = 0;

    zkErrorClear(error);
    
    if (command == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'command' is NULL");
    }
    else if (*command == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'command' doesn't point to anything");
    }
    else

    {
        if ((*command)->param_table != NULL)
        {
            uint16_t i;
            
            for (i = 0 ; i < (*command)->param_count; ++i)
            {
                zkcpValue_Destroy(NULL,&((*command)->param_table[i].value));
            }

            zkMemFree(NULL, (void**)(&((*command)->param_table)));
        }
        
        zkMemFree(NULL,(void**)command);
        result = 1;
    }

    return (result);
}



/* ------------------------------------------------------------------------
 * Makes sure that there is enough room in the parameter table of the
 * given command to insert at least one more parameter.
 */
static int
make_sure_table_is_big_enough( ZkcpCommand command )
{
    int result = 1; /* yes, the 1 is intentional */

    if (command->param_count >= command->param_table_size)
    {
        /*
         * If the parameter table is currently full, extend the buffer
         * to hold the current number of elements plus the 'growth size'
         */
        size_t size = command->param_table_size + ZKCP_PARAM_TABLE_GROW_SIZE;

        if (!zkMemRealloc( NULL, (void**)&command->param_table,
                        (size * sizeof(command->param_table[0]))))
        {
            result = 0;
        }
        else
        {
            command->param_table_size = size;
        }
    }

    return (result);
}
  

/* ------------------------------------------------------------------------
 * Adds the supplied parameter to the given command.  The parameter ID
 * for the new parameter must be unique from those of the previously added
 * parameters.  Ownership of the parameter object is transferred to the
 * command object.  Do not destroy the parameter after adding it to the
 * command.
 */
int
zkcpCommand_AddParam(
    ZK_INOUT ZkError            error,
    ZK_IN    ZkcpCommand        command,
    ZK_IN    ZkcpParameterId    id,
    ZK_IN    ZkcpValue          value )
{
    const char * const func   = "zkcpCommand_AddParam";
    int                result = 0;

    zkErrorClear(error);

    if (command == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'command' is NULL");
    }
    else if (id == ZKCP_PARAM_ID_MAC)
    {
        zkErrorSet_va(error,func,ZKERR,ZKMZKCP,ZKERROR,
                      "Parameter ID 0x%02X is reserved", id);      
    }
    else if (value == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'value' is NULL");
    }
    else if (!make_sure_table_is_big_enough(command))
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKERROR,
                   "can't adjust table size");
    }
    else
    {
        int       found = 0;
        uint16_t  slot  = 0;
        
        while ((slot < command->param_count) && (!found))
        {
            if (id == command->param_table[slot].id)
            {
                found = 1;
            }

            ++slot;
        }
        
        if (found)
        {
            zkErrorSet_va(error,func,ZKERR,ZKMZKCP,ZKERROR,
                          "can't add duplicate type id (0x%02X)", id);
        }
        else
        {
            /*
             * (slot = command->param_count) is the slot that we want
             * to put the new parameter in
             */

            command->param_table[slot].id    = id;
            command->param_table[slot].value = value;

            ++(command->param_count);

            result = 1;
        }
    }
    
    return (result);
}



/* ------------------------------------------------------------------------
 * Retrieves the command identifier associated with the given command
 *  object.
 */
int
zkcpCommand_GetId(
    ZK_INOUT ZkError            error,
    ZK_IN    ZkcpCommand        command,
    ZK_OUT   ZkcpCommandId *    command_id )
{
    const char * const func   = "zkcpCommand_GetId";
    int                result = 0;
    
    zkErrorClear(error);
    
    if (command == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'command' is NULL");
    }
    if (command_id == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'command_id' is NULL");
    }
    else
    {
        (*command_id) = command->id;
        result        = 1;
    }

    return (result);
}



/* ------------------------------------------------------------------------
 * Retrieves the parameter object associated with the given parameter id.
 *
 * The pointer that is returned parameter object refers to memory that
 * is owned by the command object.  The parameter to which it refers
 * should not be freed by the recipient.  Note that this also means that
 * the lifetime of the object referred to does not exceed that of the
 * command from which it was retrieved.
 */
int
zkcpCommand_GetParamById(
    ZK_INOUT ZkError            error,
    ZK_IN    ZkcpCommand        command,
    ZK_IN    ZkcpParameterId    param_id,
    ZK_OUT   ZkcpValue *        value )
{
    const char * const func   = "zkcpCommand_GetParamById";
    int                result = 0;

    zkErrorClear(error);

    if (command == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'command' is NULL");
    }
    if (value == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'value' is NULL");
    }
    else
    {
        size_t slot = 0;

        for ( (*value) = NULL;
              ((*value) == NULL) && (slot < command->param_count);
              ++slot )
        {
            if (param_id == command->param_table[slot].id)
            {
                (*value) = command->param_table[slot].value;
                result   = 1;
            }
        }

        if ((*value) == NULL)
        {
            zkErrorSet_va(error,func,ZKERR,ZKMZKCP,ZKERROR,
                          "can't find parameter with ID=0x%02X",
                          param_id);
        }
    }

    return (result);
}



/* ------------------------------------------------------------------------
 * Retrieves the parameter object associated with the given parameter id.
 *
 * The pointer that is returned parameter object refers to memory that
 * is owned by the command object.  The parameter to which it refers
 * should not be freed by the recipient.  Note that this also means that
 * the lifetime of the object referred to does not exceed that of the
 * command from which it was retrieved.
 */
int
zkcpCommand_GetParamByIndex(
    ZK_INOUT ZkError            error,
    ZK_IN    ZkcpCommand        command,
    ZK_IN    uint32_t           param_index,
    ZK_OUT   ZkcpParameterId *  param_id,
    ZK_OUT   ZkcpValue *        value )
{
    const char * const func   = "zkcpCommand_GetParamByIndex";
    int                result = 0;

    zkErrorClear(error);

    if (command == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'command' is NULL");
    }
    if (param_id == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'param_id' is NULL");
    }
    if (value == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'value' is NULL");
    }
    else if (param_index >= command->param_count)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKERROR,
                   "invalid parameter index");
    }
    else
    {
        (*param_id) = command->param_table[param_index].id;
        (*value)    = command->param_table[param_index].value;
        result      = 1;
    }

    return (result);
}



/* ------------------------------------------------------------------------
 * Retrieves the number of parameters associated with this command object.
 */
int
zkcpCommand_GetParamCount(
    ZK_INOUT ZkError            error,
    ZK_IN    ZkcpCommand        command,
    ZK_OUT   uint32_t *         count )
{
    const char * const func   = "zkcpCommand_GetParamCount";
    int                result = 0;
    
    zkErrorClear(error);
    
    if (command == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'command' is NULL");
    }
    if (count == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'count' is NULL");
    }
    else
    {
        (*count) = command->param_count;
        result   = 1;
    }

    return (result);
}
