/* **************************************************************************
** $Id: zkCmdSrv.c,v 1.16.10.1 2000/10/11 13:20:18 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

** Copyright (C) 2000 Zero-Knowledge Systems Inc.  All rights reserved.
** --------------------------------------------------------------------------
**
** This file contains the implementation of interface function for the 
** ZK Command Server.
**
** **************************************************************************
*/

#include <stdio.h>
#include <zks.h>
#include <netinet/in.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkDebug.h>
#include <zkFile.h>
#include <zkParms.h>
#include <zkCmdSrv.h>
#include "zkCmdSrvCore.h"
#include "zksocket.h"
#include "zkIpTable.h"


/*  zkCmdSrv_DisplayUsage - Display zk command server usage
 *
 *  This procedure displays information about the available command line options
 *  for the zk command server.
 *
 *  @param progName - Name of the program running
 *  @return Nothing
 */
void zkCmdSrv_DisplayUsage (const char *progName)
{
  fprintf(stdout,VERSION_COMMAND_SERVER "\n\n"
  "Usage:  %s [--options ...]\n\n"
  "  Where the options are:\n"
  "    " P_HELP "\n"
  "       Displays this information.\n\n"
  "    " P_CONFIG " <filename>\n"
  "       Specify which configuration file you wish to use.\n\n"
  "    " P_LOGFILE " <filename>\n"
  "       Specifies the server logFile.\n\n"
  "    " P_PIDFILE " <filename>\n"
  "       Specifies the server pid file.\n\n"
  "    " P_ACCESS_FILE " <filename>\n"
  "       Specifies the name of the file containing IP addresses allowed to\n"
  "       connect.\n\n"
  "    " P_AUTHENTICATION " <string>\n"
  "       Specifies the key for packet authentication.\n\n"
  "    " P_ADDRESS_PATH " <address>\n"
  "       Specifies the address path on which the server will listen.\n\n"
  "    " P_LISTEN_QUEUE_LENGTH " <length>\n" 
  "       Maximum number of incoming pending connections.\n\n"
  "    " P_MAX_PACKET_SIZE " <kilo-Bytes>\n"
  "       This is the maximum size (in kB) of packets that can be transmited.\n\n"
  "    " P_MAX_IDLE_TIME " <seconds>\n"
  "       This is the maximum number of seconds that a connection can be idle.\n"
  "       When the limit is exceeded, the connection is automatically closed.\n\n"
  "    " P_MAX_CONNECTIONS " <connections>\n"
  "       This specifies the number of connections to handle (i.e.number of processes\n"
  "       to fork.\n\n"
  "    " P_PROCESS_LIFE " <connections>\n"
  "       This specifies the maximum connections that a process should handled. After\n"
  "       that the process should terminate and be replaced by a newly forked process.\n"
  "       Note that special value 0 means that processes are immortal.\n\n"
  "    " P_TCP_DELAY "\n"
  "       This specifies that TCP transmit delay should be used when sending packets.\n\n"
  "    " P_VERBOSE " <level>\n"
  "       Specifies the logging verbose level.  This is an integer value between \n"
  "       0 and 4, where 0 provides almost no logging and 4 logs all input.\n\n"
  "  Some debugging options are also available:\n"
  "    " P_DEBUG "\n"
  "       This turns debugging message on or off.\n\n"
  "    " P_DUMP_OPTIONS "\n"
  "       This option causes all configuration options to be displayed at startup.\n\n"
  "    " P_DONTFORK "\n"
  "       When specified, the server will not background itself automatically when\n"
  "       launched.\n\n"
  "    " P_NOCHILD "\n"
  "       When specified, the server process will not fork processes to handle new\n"
  "       connections but will handle them itself.\n\n",
          progName );
}

/*  zkCmdSrv_Create - Allocate and initialize zk command server
 *
 *  This procedure initializes the server context with default values and
 *  creates the error and log objects.
 *
 *  @param err        - Used to report errors (if any)
 *  @param userConfig - User configurable options
 *  @param cmdsrv     - New command server to be allocated
 *  @return 0 (failure) or 1 (success)
 */
int zkCmdSrv_Create(ZK_OUT ZkError err, 
                    ZK_IN  ZkCmdSrvConfig_s *userConfig, 
                    ZK_OUT ZkCmdSrv *cmdsrv)
{
  const char *func = "zkCmdSrv_Create";
  int result = 0;

  if (!cmdsrv) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Null pointer to Command Server object");
    goto done;
  }

  if ((*cmdsrv = malloc(sizeof(**cmdsrv))) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENOMEM, "malloc() failed");
    goto done;
  }

  /* Reset memory so all memory pointers in the structures are NULL. */
  memset(*cmdsrv, 0, sizeof(**cmdsrv));

  if (((*cmdsrv)->callbackTable = malloc(userConfig->maxCommands * sizeof(ZkCmdSrvCallback)))
      == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENOMEM, "malloc() failed");
    goto done;
  }

  if (!zkErrorNew(&((*cmdsrv)->err))) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENOMEM, "zkErrorNew() failed");
    goto done;
  }

  if (!zkLogNew(&((*cmdsrv)->log))) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENOMEM, "zkLogNew() failed");
    goto done;
  }
  if (!zkSocketNew(err, AF_INET, SOCK_STREAM, IPPROTO_IP, &((*cmdsrv)->socket))) {
    goto done;
  }

  /* Set config fields to their default values. Note that since memset was
     called all values not explicit set are 0 (or NULL). */
  (*cmdsrv)->userConfig     = *userConfig;
  (*cmdsrv)->idleConnTimer  = DEFAULT_CMDSRV_MAX_IDLE_CONN_TIME;
  (*cmdsrv)->maxConn        = DEFAULT_CMDSRV_MAX_CONNECTIONS;
  (*cmdsrv)->processLife    = DEFAULT_CMDSRV_PROCESS_LIFE;
  (*cmdsrv)->inQueueLength  = DEFAULT_CMDSRV_INQUEUE_LENGTH;
  (*cmdsrv)->maxPacketSize  = DEFAULT_CMDSRV_MAX_PACKET_SIZE;
  (*cmdsrv)->verbosityLevel = 1;

  /* Everything is fine */
  result = 1;

 done:
  /* If we failed, make sure to deallocate the memory. */
  if (!result) {
    if (cmdsrv) {
      zkCmdSrv_Destroy(err, cmdsrv);
    }
  }
  
  return result;
}

/*  zkCmdSrv_LoadConfig - Load command line and config file options into the
 *                        command server.
 *
 *  This procedure loads the command line and config file options into 
 *  the given command server. It also enables the logging system.
 *
 *  @param err    - Used to report errors (if any)
 *  @param config - Contains config file and command line options.
 *  @param cmdsrv - Command server to load the config in.
 *  @return 0 (failure) or 1 (success)
 */
int zkCmdSrv_LoadConfig(ZK_OUT ZkError err, 
                        ZK_IN  ZkConfig config,
                        ZK_OUT ZkCmdSrv cmdsrv)
{
  const char *func = "zkCmdSrv_LoadConfig";
  uint16_t dumpOptions = 0;
  const char *configFile = NULL;

  /* Make sure we don't have NULL objects. */

  if (!config) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Null Config object");
    return 0;
  }

  if (!cmdsrv) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Null Command Server");
    return 0;
  }

  /* Set the IP table to NULL (Needed in error condition). */
  cmdsrv->ipTable = NULL;

  /*
    Parameter look up
  */

  zkConfigGetParameter(err, config, TYPE_BOOLEAN, S_COMMAND_SERVER,
                       P_DEBUG, &(cmdsrv->debug));
  
  zkConfigGetParameter(err, config, TYPE_BOOLEAN, S_COMMAND_SERVER,
                       P_DONTFORK, &(cmdsrv->dontFork));
  
  zkConfigGetParameter(err, config, TYPE_BOOLEAN, S_COMMAND_SERVER,
                       P_NOCHILD, &(cmdsrv->userConfig.noChild));
  
  zkConfigGetParameter(err, config, TYPE_BOOLEAN, S_COMMAND_SERVER,
                       P_TCP_DELAY, &(cmdsrv->tcpDelay));
  
  zkConfigGetParameter(err, config, TYPE_UINT16, S_COMMAND_SERVER,
                       P_VERBOSE, &(cmdsrv->verbosityLevel));

  zkConfigGetParameter(err, config, TYPE_STRING, S_COMMAND_SERVER,
                       P_PIDFILE, &(cmdsrv->userConfig.pidFile));

  zkConfigGetParameter(err, config, TYPE_STRING, S_COMMAND_SERVER,
                       P_LOGFILE, &(cmdsrv->userConfig.logFile));

  zkConfigGetParameter(err, config, TYPE_STRING, S_COMMAND_SERVER,
                       P_ACCESS_FILE, &(cmdsrv->userConfig.accessFile));

  zkConfigGetParameter(err, config, TYPE_STRING, S_COMMAND_SERVER,
                       P_ADDRESS_PATH, &(cmdsrv->userConfig.addressPath));

  zkConfigGetParameter(err, config, TYPE_UINT16, S_COMMAND_SERVER,
                       P_MAX_IDLE_TIME, &(cmdsrv->idleConnTimer));

  zkConfigGetParameter(err, config, TYPE_UINT16, S_COMMAND_SERVER,
                       P_MAX_CONNECTIONS, &(cmdsrv->maxConn));

  zkConfigGetParameter(err, config, TYPE_UINT32, S_COMMAND_SERVER,
                       P_PROCESS_LIFE, &(cmdsrv->processLife));

  zkConfigGetParameter(err, config, TYPE_UINT16, S_COMMAND_SERVER,
                       P_LISTEN_QUEUE_LENGTH, &(cmdsrv->inQueueLength));

  zkConfigGetParameter(err, config, TYPE_UINT16, S_COMMAND_SERVER,
                       P_MAX_PACKET_SIZE, &(cmdsrv->maxPacketSize));

  zkConfigGetParameter(err, config, TYPE_STRING, S_COMMAND_SERVER,
                       P_AUTHENTICATION, &(cmdsrv->passphrase));

  /* 
     Validation
  */

  if (!cmdsrv->passphrase) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKERROR,
               "No passphrase supplied for HMAC encryption");
    goto error;
  }

  if (!zkcpContext_SetKey(err, cmdsrv->userConfig.zkcpCtx, 
                          cmdsrv->passphrase, strlen(cmdsrv->passphrase)) || 
      !zkcpContext_SetMaxCommandSize(err, cmdsrv->userConfig.zkcpCtx,
                                     1024 * cmdsrv->maxPacketSize)) {
    goto error;
  }
  
  /* Load the access table */
  if (!zkLoadIpTableFromFile(err, cmdsrv->userConfig.accessFile, 
                             (const char ***) &(cmdsrv->ipTable),
                             &(cmdsrv->ipTableSize))) {
    goto error;
  }
  
  /* Ignore the log file if debug is set */
  if (cmdsrv->debug) {
    zkLogSetStream(cmdsrv->log, stderr); 
    zkLogSetProgname(cmdsrv->log, cmdsrv->userConfig.progName);
    zkLogEnable(cmdsrv->log);
    zkDebugOn();
  } 
  else if (cmdsrv->userConfig.logFile) {
    if (!zkLogSetFile(cmdsrv->log, cmdsrv->userConfig.logFile)) {
      zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKERROR, 
                    "Unable to open logFile %s", cmdsrv->userConfig.logFile);
      goto error;
    }
    zkLogEnable(cmdsrv->log);
  } 
  else {
    zkLogDisable(cmdsrv->log);
  }

  /* Check if option configuration should be displayed */
  zkConfigGetParameter(err, config, TYPE_BOOLEAN, S_COMMAND_SERVER,
                       P_DUMP_OPTIONS, &dumpOptions);

  if (dumpOptions) {
    zkConfigGetParameter(err, config, TYPE_STRING, S_COMMAND_SERVER,
                         P_CONFIG, &configFile);

    if (!configFile) {
      configFile = "Config file not specified (using application default)";
    }

    printf("Server configuration options:\n"
           "  " P_CONFIG " = %s\n"
           "  " P_LOGFILE " = %s\n"
           "  " P_PIDFILE " = %s\n"
           "  " P_ACCESS_FILE " = %s\n"
           "  " P_AUTHENTICATION " = %s\n"
           "  " P_ADDRESS_PATH " = %s\n"
           "  " P_LISTEN_QUEUE_LENGTH " = %d\n" 
           "  " P_MAX_PACKET_SIZE " = %d KB\n"
           "  " P_MAX_IDLE_TIME " = %d seconds\n"
           "  " P_MAX_CONNECTIONS " = %d\n"
           "  " P_PROCESS_LIFE " = %d\n"
           "  " P_TCP_DELAY "= %d\n"
           "  " P_VERBOSE " = %d\n"
           "  " P_DEBUG " = %d\n"
           "  " P_DONTFORK "= %d\n"
           "  " P_NOCHILD " = %d\n"
           "  " P_DUMP_OPTIONS " = 1\n", 
           configFile,
           cmdsrv->userConfig.logFile,
           cmdsrv->userConfig.pidFile,
           cmdsrv->userConfig.accessFile,
           cmdsrv->passphrase,
           cmdsrv->userConfig.addressPath,
           cmdsrv->inQueueLength,
           cmdsrv->maxPacketSize,
           cmdsrv->idleConnTimer,
           cmdsrv->maxConn,
           cmdsrv->processLife,
           cmdsrv->tcpDelay,
           cmdsrv->verbosityLevel,
           cmdsrv->debug, 
           cmdsrv->dontFork,
           cmdsrv->userConfig.noChild);
  }

  return 1;

 error:
  /* Deallocate the IP table. */
  if (cmdsrv->ipTable) {
    int count = cmdsrv->ipTableSize;
    while (count--){
      free(cmdsrv->ipTable[count]);
    }
    free(cmdsrv->ipTable);
    cmdsrv->ipTable = NULL;
  }

  return 0;
}

/*  zkCmdSrv_AddCommand - Add command to a Command Server
 *
 *  This procedure adds a new command and its callback function to a
 *  Command Server.
 *
 *  @param err      - Used to report errors (if any)
 *  @param commId   - Id of the new command
 *  @param function - Callback function for the new command
 *  @param cmdsrv   - Command server
 *  @return 0 (failure) or 1 (success)
 */
int zkCmdSrv_AddCommand(ZK_OUT ZkError err,
                        ZK_IN  ZkCmdSrv cmdsrv,
                        ZK_IN  uint16_t commId,
                        ZK_IN  ZkCmdSrvCallback function)
{
  const char *func = "zkCmdSrv_AddCommand";

  if (!function) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Null callback function");
    return 0;
  }
  
  if (!cmdsrv) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Null Command Server");
    return 0;
  }

  if (!cmdsrv->callbackTable) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKERROR, "Null callback table");
    return 0;
  }

  if (commId >= cmdsrv->userConfig.maxCommands) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKEPARMRANGE, 
                  "Invalid command ID (%d) - must be between 0 and %d", 
                  commId, cmdsrv->userConfig.maxCommands - 1);
    return 0;
  }
  
  /* Everything is fine so let's store the callback function in the table. */
  cmdsrv->callbackTable[commId] = function;

  return 1;
}

/*  zkCmdSrv_SetInitSession - Set the session init callback function 
 *
 *  This function set the callback function to be called when a new
 *  connection (or session) is made.
 *
 *  @param err      - Used to report errors (if any)
 *  @param cmdsrv   - Command server
 *  @param function - Callback function
 *  @return 0 (failure) or 1 (success)
 */
int zkCmdSrv_SetInitSession(ZK_OUT ZkError err,
                            ZK_OUT ZkCmdSrv cmdsrv,
                            ZK_IN  ZkCmdSrvInitSession function)
{
  const char *func = "zkCmdSrv_SetInitSession";

  if (!cmdsrv) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Null Command Server");
    return 0;
  }

  if (!function) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Null callback function");
    return 0;
  }

  /* Everything is fine so let's store the callback function in the table. */
  cmdsrv->initSession = function;

  return 1;
}

/*  zkCmdSrv_SetTermSession - Set the session termination callback function 
 *
 *  This function set the callback function to be called when a connection
 *  (or session) is about to be closed.
 *
 *  @param err      - Used to report errors (if any)
 *  @param cmdsrv   - Command server
 *  @param function - Callback function
 *  @return 0 (failure) or 1 (success)
 */
int zkCmdSrv_SetTermSession(ZK_OUT ZkError err,
                            ZK_OUT ZkCmdSrv cmdsrv,
                            ZK_IN  ZkCmdSrvTermSession function)
{
  const char *func = "zkCmdSrv_SetTermSession";
  
  if (!cmdsrv) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Null Command Server");
    return 0;
  }

  if (!function) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Null callback function");
    return 0;
  }

  /* Everything is fine so let's store the callback function in the table. */
  cmdsrv->termSession = function;

  return 1;
}

/*  zkCmdSrv_SetInitProcess - Set the process init callback function 
 *
 *  This function set the callback function to be called when a new
 *  process is forked.
 *
 *  @param err      - Used to report errors (if any)
 *  @param cmdsrv   - Command server
 *  @param function - Callback function
 *  @return 0 (failure) or 1 (success)
 */
int zkCmdSrv_SetInitProcess(ZK_OUT ZkError err,
                            ZK_OUT ZkCmdSrv cmdsrv,
                            ZK_IN  ZkCmdSrvInitProcess function)
{
  const char *func = "zkCmdSrv_SetInitProcess";

  if (!cmdsrv) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Null Command Server");
    return 0;
  }

  if (!function) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Null callback function");
    return 0;
  }

  /* Everything is fine so let's store the callback function in the table. */
  cmdsrv->initProcess = function;

  return 1;
}

/*  zkCmdSrv_SetTermProcess - Set the process termination callback function 
 *
 *  This function set the callback function to be called when a process
 *  terminates.
 *
 *  @param err      - Used to report errors (if any)
 *  @param cmdsrv   - Command server
 *  @param function - Callback function
 *  @return 0 (failure) or 1 (success)
 */
int zkCmdSrv_SetTermProcess(ZK_OUT ZkError err,
                            ZK_OUT ZkCmdSrv cmdsrv,
                            ZK_IN  ZkCmdSrvTermProcess function)
{
  const char *func = "zkCmdSrv_SetTermProcess";
  
  if (!cmdsrv) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Null Command Server");
    return 0;
  }

  if (!function) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Null callback function");
    return 0;
  }

  /* Everything is fine so let's store the callback function in the table. */
  cmdsrv->termProcess = function;

  return 1;
}

/*  zkCmdSrv_SetSignal - Set the SIGURG signal callback
 *
 *  This function set the callback function to be called when the SIGURG signal
 *  is raised.
 *
 *  @param err      - Used to report errors (if any)
 *  @param cmdsrv   - Command server
 *  @param function - Callback function
 *  @return 0 (failure) or 1 (success)
 */
int zkCmdSrv_SetSignal(ZK_OUT ZkError        err,
                       ZK_OUT ZkCmdSrv       cmdsrv,
                       ZK_IN  ZkCmdSrvSignal function)
{
  const char *func = "zkCmdSrv_SetSignal";
  
  if (!cmdsrv) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Null Command Server");
    return 0;
  }

  if (!function) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Null callback function");
    return 0;
  }

  /* Everything is fine so let's store the callback function in the table. */
  cmdsrv->signalCallback = function;

  return 1;
}

/*  zkCmdSrv_Start - Start the command server
 *
 *  This procedure forks a new process that will run as the Command Server.
 *
 *
 *  @param err       - Used to report errors (if any)
 *  @param cmdsrv    - Command server to run
 *  @param processId - Same as fork: -1 fork error
 *                                   0 for child process
 *                                   > 0 for parent process
 *                     If no fork is performed (i.e. dontFork), 0 is returned.
 *  @return 0 (failure) or 1 (success)
 */
int zkCmdSrv_Start(ZK_OUT ZkError err, 
                   ZK_IN  ZkCmdSrv cmdsrv,
                   ZK_OUT pid_t *processId)
{
  const char *func = "zkCmdSrv_Start";
  int result = 0;

  /* Make sure we don't have NULL objects. */
  if (!cmdsrv) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Null Command Server");
    goto done;
  }

  if (!processId) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Null process ID pointer");
    goto done;
  }

  /* Should we run as daemon ? */
  if (cmdsrv->dontFork) {
    *processId = 0;
  } else {
    switch (*processId = fork()) {
    case -1:
      /* Failure to fork new process */
      zkErrorSystemSet(err, func, ZKERR, ZKMOD, "fork() failed");
      goto done;
      
    case 0: /* Child */
      break;
      
    default:
      /* Parent process exits. */
      result = 1;
      goto done;
    }
  }

  if (!cmdsrv->tcpDelay)
    if (!zkSocketSetOption(err, cmdsrv->socket, 
			   ZKSOCKET_OPTION_NoTransmitDelay, 0)) {
      goto logError;
    }

  /* Need to configure and start listening on socket */
  if (!zkSocketSetOption(err, cmdsrv->socket, 
                         ZKSOCKET_OPTION_QueueDepth, cmdsrv->inQueueLength) ||
      !zkSocketSetOption(err, cmdsrv->socket, 
                         ZKSOCKET_OPTION_ReuseAddress, 0) ||
      !zkSocketBind(err, cmdsrv->socket, cmdsrv->userConfig.addressPath) ||
      !zkSocketListen(err, cmdsrv->socket) ) {
    goto logError;
  }
  
  /* We need to have a PID file to prevent multiple server instances so
     exit if we fail to create the PID file. */
  if (!zkFileWritePid(err, cmdsrv->userConfig.pidFile, &(cmdsrv->pidFptr))) {
    cmdsrv->pidFptr = NULL;
    goto logError;
  }

  /* Start the daemon loop */
  if (!zkCmdSrv_RunDaemon(err, cmdsrv)) {
    goto logError;
  }

  result = 1;

 logError:
  if (!result)
    zkLogError(cmdsrv->log, err);

 done:
  if (cmdsrv->pidFptr) {
    fclose(cmdsrv->pidFptr);
    cmdsrv->pidFptr = NULL;
    unlink(cmdsrv->userConfig.pidFile);
  }

  return result;
}

/*  zkCmdSrv_Destroy - Deallocate all resources associated with a Command Server
 *
 *  This procedure frees memory allocated for the command server and close all 
 *  open descriptors. 
 *
 *  @param err    - Used to report errors (if any).
 *  @param cmdsrv - Pointer to the command server 
 *  @return Nothing
 */
int zkCmdSrv_Destroy(ZK_OUT ZkError err, ZK_IN ZkCmdSrv *cmdsrv)
{
  const char *func = "zkCmdSrv_Destroy";

  if (!cmdsrv || !*cmdsrv) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Null Command Server");
    return 0;
  }

  if ((*cmdsrv)->log)     zkLogFree(&((*cmdsrv)->log));
  if ((*cmdsrv)->err)     zkErrorFree(&((*cmdsrv)->err));
  if ((*cmdsrv)->socket)  zkSocketFree(err, &((*cmdsrv)->socket));

  if ((*cmdsrv)->ipTable) {
    int count = (*cmdsrv)->ipTableSize;
    while (count--){
      free((*cmdsrv)->ipTable[count]);
    }
    free((*cmdsrv)->ipTable);
  }
  if ((*cmdsrv)->callbackTable) free((*cmdsrv)->callbackTable);

  free(*cmdsrv);
  *cmdsrv = NULL;

  return 1;
}
