/* **************************************************************************
** $Id: zkIpTable.c,v 1.8.10.1 2000/10/11 13:20:18 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

** Copyright (C) 2000 Zero-Knowledge Systems Inc.  All rights reserved.
** --------------------------------------------------------------------------
**
** This file contains implementation of function for manipulating an IP
** table.
**
** **************************************************************************
*/

#include "zks.h"
#include <stdlib.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <errno.h>
#include <zkError.h>
#include "zkCmdSrvCore.h"
#include "zkIpTable.h"


/*  zkLoadIpTableFromFile - Create an IP address table from a test file.
 *
 *  This function allocates and fills an IP address table from a file.
 *
 *  @param err      - Used to report errors (if any)
 *  @param filename - Name of the file to read
 *  @param ipTable  - Pointer to the new table of IP addresses.
 *                    Only updated on success.
 *  @param size     - Number of IP address in the table.
 *  @return 1 (Success) or 0 (Failure)
 */
int zkLoadIpTableFromFile(ZK_OUT ZkError       err,
                          ZK_IN  const char*   filename,
                          ZK_OUT const char*** ipTable,
                          ZK_OUT uint16_t*     size)
{
  static const char *func = "zkLoadIpTableFromFile";
  char **tmpTable = NULL;
  FILE *file;
  uint16_t count = 0,tmpTableSize = ZKIPTABLE_GROW_SIZE;
  int exitCode = 0;
  char line[MAX_LINE_SIZE];
  char *ipAddr, *end;
  int lineCount = 0; 
  (void) func;

  if (!filename || !ipTable || !size) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "NULL pointer passed in");
    return 0;
  }

  if ( !(file = fopen(filename, "r")) ) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKEFILEOPEN,
                  "Failed to open file \"%s\": %s", filename, strerror(errno) );
    goto cleanup;
  }
  
  if ( !(tmpTable = malloc(tmpTableSize * sizeof(char *)))) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENOMEM, "IP table allocation failed");
    goto cleanup;
  }
  
  /* This makes sure we don't go out of range when searching the line. */
  line[MAX_LINE_SIZE-1] = '\n';
  
  while (fgets(line, MAX_LINE_SIZE-1, file)) {
    lineCount++;
    
    /* Move to first non-space character */
    ipAddr = line;
    for (; *ipAddr == ' '; ipAddr++);
    
    /* Check for comment or empty line */
    if ((*ipAddr == ';') || (*ipAddr == '\n')) {
      continue;
    }
    
    /* Get end of address and make sure address is terminated by NULL */
    end = ipAddr;
    for (; (*end != ' ') && (*end != '\n') ; end++);
    *end = '\0';
    
    /* Make sure the IP address onyl contains allowed characters */
    if (strspn(ipAddr, "0123456789./:") != strlen(ipAddr)) {
      /* The entry is not an IP address... Spit out the invalid symbol. */
      zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKEFILEOPEN, 
                    "Invalid token in file \"%s\" at line %d: \"%s\"", 
                    filename, lineCount,ipAddr);
      goto cleanup;
    }
    
    /* Check if the table is full. */
    if (count >= tmpTableSize) {
      char **tmpPtr = realloc(tmpTable, (tmpTableSize + ZKIPTABLE_GROW_SIZE) 
                                    * sizeof(char *));
      
      if (tmpPtr == NULL ) {
        
        zkErrorSet(err, func, ZKERR, ZKMOD, ZKENOMEM,
                   "Failed to allocate IP table");
        goto cleanup;
      }
      else {
        tmpTable = tmpPtr;
        tmpTableSize += ZKIPTABLE_GROW_SIZE;
      }
    }
    
    tmpTable[count++] = strdup(ipAddr);
  }
  
  if (count == 0) {
    /* Empty file */
    zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKEFILEOPEN,
                  "No valid IP address in \"%s\"", filename);
    goto cleanup;
  }
  else {
    /* Everything went fine */
    *ipTable = (const char **) tmpTable;
    *size = count;
    exitCode = 1;
  }

 cleanup:
  if (file) 
    fclose(file);

  /* Deallocate the table if we failed... */
  if (exitCode == 0)
    if (tmpTable) {
      while (count--) {
        free(tmpTable[count]);
      }
      free(tmpTable);
    }
    
  return exitCode;
}

/*  zkHostInTable - Check if an IP address is in the IP table
 *  
 *  This function searches for an IP address through a given IP table.
 *
 *  @param err     - Used to report errors (if any)
 *  @param ipTable - Pointer to the IP table
 *  @param size    - Number of addresses in the table
 *  @param addr    - Address to validate
 *  @param found   - True if the addr is found
 *  @return 1 (Success) or 0 (Failure)
 */
int zkHostInTable(ZK_OUT ZkError      err, 
                  ZK_OUT const char** ipTable,
                  ZK_IN  uint16_t     size,
                  ZK_IN  const char*  addrPath,
                  ZK_OUT int*         found)
{
  static const char *func = "zkHostInTable";
  int i;
  (void) func;
  
  if (!addrPath || !found) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR,
               "NULL pointer passed in");
    return 0;
  }

  /* Default is not found */
  *found = 0;

  if (!ipTable) {
    /* Empty table. Nothing to look up... */
    return 1;
  }

  /* Let's search through the table */
  for(i=0; i<size && !*found; i++) {
    /* Only look for address match (not port) */
    if (!zkSocketAddressCompare(err, ZKSOCKET_CMP_INETADDR,
				ipTable[i], addrPath, found)) {
      return 0;
    }
  } 
  return 1;
}
