/*
 * $Id: ZKIP.h,v 1.17.8.1 2000/10/11 13:20:22 phil Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef _ZKIP_H
#define _ZKIP_H

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#ifndef _ZKLOG_H
#include <zkLog.h>
#endif /* _ZKLOG_H */

/*
 * Basic ZKIP command return codes 
 *
 * Notes:
 * Some of these codes are specific to particular applications of ZKIP, for
 * instance _COMMAND_FAILED_NYM_EXISTS (549) is particular to NSCP and will
 * be returned for a 1.1.0 create command which fails because the nym was
 * already in the database.
 *
 * These response codes have a comment which indicates which protocol they
 * belong to.  You should not use them outside of that context.
 */


#define _SERVICE_READY              220
#define _SERVICE_CLOSING            221
#define _DATA_STARTINPUT            344
#define _SERVICE_UNAUTHORIZED       401
#define _SERVICE_FORBIDDEN          403
#define _SERVICE_NOTFOUND           404
#define _SERVICE_UNAVAILABLE        421
#define _COMMAND_UNRECOGNIZED       500
#define _COMMAND_ERROR              501
#define _COMMAND_NOTIMPLEMENTED     502
#define _COMMAND_BADDATA            541
#define _COMMAND_FAILED             544
#define _COMMAND_DATA_NOTFOUND      545
#define _COMMAND_NOTEXECUTED        550

/*
 * Special ZKIP response codes
 *
 * These codes were added in to support specific aspects of the Freedom Client
 * GUI.  Where possible, these extra codes should be kept to a minimum.  
 *
 * NSCP/548 is returned by the nym server when a user makes a request to disable
 *          a nym which has already been disabled
 * NSCP/549 is returned when a user tries to create a nym whose name is already
 *          in use by another nym
 * NSCP/547 Returned when a user tries to respend a token which has already 
 *          been used.
 * NSCP/546 Returned when a token is refused.  This happens when the specified
 *          token is inappropriate for the particular command for which it
 *          was applied.
 * SNAP/546 Serial number was refused because it is not yet active (it will
 *          be valid later)
 */

#define _COMMAND_FAILED_TOKEN_REFUSED 546    /* NSCP */
#define _COMMAND_FAILED_NYM_DISABLED  548    /* NSCP */
#define _COMMAND_FAILED_TOKEN_SPENT   547    /* NSCP */
#define _COMMAND_FAILED_NYM_EXISTS    549    /* NSCP */


#define _COMMAND_FAILED_SERIAL_NUMBER_REFUSED 546    /* SNAP */

/*
 * Protocol ID codes (first number of the command ID)
 */

#define ZKIP_PROTO_ZKIP  0 /* Basic ZKIP commands.  Must be on all servers */
#define ZKIP_PROTO_NSCP  1 /* Nym Server Command Protocol */
#define ZKIP_PROTO_AMTP  2 /* Anonymous Mail Transfer Protocol */
#define ZKIP_PROTO_TAP   3 /* Token Authentication Protocol (currently unused)*/
#define ZKIP_PROTO_KSCP  4 /* Key Server Command Protocol */
#define ZKIP_PROTO_SNAP  5 /* Serial Number Access Protocol */
#define ZKIP_PROTO_XFER  6 /* Data transfer protocol */
#define ZKIP_PROTO_NTRP  7 /* Network Topology Requests Protocol */
#define ZKIP_PROTO_LODE  9 /* Support administration (Lode server) */

/*
 * Data encoding formats
 */

#define ZKIP_ENCODING_NONE          0
#define	ZKIP_ENCODING_UNKNOWN       6
#define ZKIP_ENCODING_BASE64        1
#define ZKIP_ENCODING_7BIT          2
#define ZKIP_ENCODING_8BIT          3
#define	ZKIP_ENCODING_BINARY        4
#define ZKIP_ENCODING_BZIP2         5

/*
 * These return codes are used by callback functions to continue or terminate
 * a ZKIP session.
 */

#define	_SESSION_CONTINUE           1
#define	_SESSION_END                0
#define _SESSION_ERROR              2 /* May or may not terminate based on 
                                       * flags 
                                       */

/*
 * A rather quick way to get around a circular reference, it's in zkipSession.h
 * as well since the zkipCommand, zkipEngine, and zkipSession objects are all
 * interelated when it comes to the ZkipCommandCallback.  --Philippe
 */

#ifndef _T_ZKIPSESSION
#define _T_ZKIPSESSION
typedef struct T_ZkipSession * ZkipSession;
#endif /* _T_ZKIPSESSION */

typedef int (*ZkipCommandCallback)(ZkError,ZkipSession,void *);

#define	ZKIP_CALLBACK(_callback) _callback(\
  ZkError err, \
  ZkipSession session, \
  void * data )

/*
 * CMDID() builds up a command identifier number (an unsigned 32 bit value)
 * from the the protocol, major version, and minor version numbers.  The
 * other 3 macros are used to pull out the protocol, major and minor version
 * numbers.
 */

#define	CMDID(_proto,_major,_minor) ((_proto)*0x10000+(_major)*0x100+(_minor))
#define CMDID_PROTO(_cmdid) (((_cmdid) & 0x00FF0000) >> 16)
#define CMDID_MAJVER(_cmdid) (((_cmdid) & 0x0000FF00) >> 8)
#define CMDID_MINVER(_cmdid) ((_cmdid) & 0x000000FF)

#ifndef _ZKIPENGINE_H
#include <zkipEngine.h>
#endif /* _ZKIPENGINE_H */

#ifndef _ZKIPSESSION_H
#include <zkipSession.h>
#endif /* _ZKIPSESSION_H */

/*
 * Predefine callback functions
 */

int zkipCallbackQuit (
  ZkError ,
  ZkipSession ,
  void * );

int zkipCallbackList (
  ZkError ,
  ZkipSession ,
  void * );

int zkipCallbackEcho (
  ZkError ,
  ZkipSession ,
  void * );

int zkipCallbackPulse (
  ZkError ,
  ZkipSession ,
  void * );

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* _ZKIP_H */
