/*
 * $Id: zkipClient.c,v 1.34.8.4 2000/10/11 13:20:22 phil Exp $
 */

static const char * module_id_ = "$Id: zkipClient.c,v 1.34.8.4 2000/10/11 13:20:22 phil Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>

#ifndef WIN32
#include <sys/time.h>
#include <unistd.h>
#include <netdb.h>
#include <arpa/inet.h>
#else
#include <time.h>
#include <winsock2.h>
#endif /* !WIN32 */

#include <fcntl.h>
#include <zkIntTypes.h>
#include <zkError.h>
#include <zkMemBuf.h>
#include <ZKIP.h>
#include "zkipParse.h"
#include <zkipClient.h>
#include <zkNet.h>

#define ZKIP_RESPONSE_TOKENS  7 /* A Zkip response has 7 tokens in it */
#define TOKEN_TRANSID         0 /* The transaction id */
#define TOKEN_CMDID           1 /* The Command identifier code */
#define TOKEN_RESPONSE        2 /* The response code */
#define TOKEN_DATATYPE        3 /* The extra data datatype */
#define TOKEN_ENCODING        4 /* The extra data encoding format */
#define TOKEN_DATALEN         5 /* The number of octets in the extra data */
#define TOKEN_MESSAGE         6 /* The response message string */


#define DEFAULT_TIMEOUT     30 /* Default timeout is 30 seconds */

#define FLAG_BLOCKING       0x1 /* Make the ZKIP client blocking */

#define PARSE_MODE_ERROR      1 /* Some error occured in the incoming data */
#define PARSE_MODE_COMMAND    2 /* Currently reading a command */
#define PARSE_MODE_DATA       3 /* Currently reading extra data */
#define PARSE_MODE_BANNER     4 /* Currently reading a welcome banner */
#define PARSE_MODE_RESPONSE   5 /* Currently reading a command response */ 

#define RESPONSE_MODE_INCOMING  1 /* Further response data is expected */
#define RESPONSE_MODE_COMPLETE  2 /* The response is ready to use */
#define RESPONSE_MODE_ERROR     3 /* The complete response was broken */

#define CONNECT_STATUS_CONNECTED    1
#define CONNECT_STATUS_NOTCONNECTED 2
#define CONNECT_STATUS_ERROR        3
#define CONNECT_STATUS_WAITING      4


#define HOSTNAME_SZ 128       /* Maximum length of the hostname */

struct T_ZkipClient {
  uint32_t flags;            /* Behaviour flags */
  int fd;                    /* The socket descriptor used to read or write */
  size_t response_len;       /* Length of the response */
  ZkMemBuf response;         /* Response data buffer */
  size_t data_len;           /* Length of the data */
  ZkMemBuf data;             /* Buffer of extradata being read in */ 
  int parse_mode;            /* can be ERROR, BANNER, RESPONSE, or DATA */
  int state;                 /* used to keep track of crlf and crlf.crlf */
  int response_mode;         /* can be INCOMING, COMPLETE, ERROR */
  int connect_status;        /* can be NOTCONNECTED, CONNECTED, or ERROR */
  struct ZkipCommandResponse cmdresp; /* Stores the command response */
  char hostname[HOSTNAME_SZ + 1]; /* hostname name of remote ZKIP server */
  uint16_t port;             /* remote ZKIP server port */  
  char * extra_data;         /* Holds the extra data when complete */
  int timeout;               /* Timeout for blocking operations */
};

/*
 * Internal function prototypes
 */

static int SafeClose ( const int );

#ifdef FREEDOM_LINUX_CLIENT
static int zkConnectWithTimeout ( ZkError , int , const struct sockaddr * , 
                                  socklen_t , int );
#endif /* FREEDOM_LINUX_CLIENT */

/*
 * Close a file descriptor function
 */
static int SafeClose(const int fd){

#ifdef WIN32
	shutdown(fd, SD_BOTH);
	return closesocket(fd);
#else
	return close(fd);
#endif
}


/**
 *  Create a new zkipClient structure and initialize it to default 
 *  parameters.
 *
 */

DLLExport int 
zkipClientNew (
  ZkError err,
  ZkipClient * client )
{
  (void) module_id_;
  zkErrorClear(err);
  if (client) {
    *client = (ZkipClient) malloc(sizeof(struct T_ZkipClient));
    if (*client == NULL) {
      zkErrorSet(err,"zkipClientNew",ZKERR,ZKIP,ZKENOMEM,"malloc");
      return 0;
    }
    memset(*client,0,sizeof(struct T_ZkipClient));
    if (!zkMemBufNew(err,&((*client)->response))) {
      free(*client);
      return 0;
    } 
    if (!zkMemBufNew(err,&((*client)->data))) {
      free(*client);
      return 0;
    } 
    (*client)->parse_mode = PARSE_MODE_BANNER;
    (*client)->response_mode = RESPONSE_MODE_INCOMING;
    (*client)->connect_status = CONNECT_STATUS_NOTCONNECTED;
    (*client)->fd = -1;
    (*client)->timeout = DEFAULT_TIMEOUT;
    (*client)->flags = 0;
    return 1;
  }
  zkErrorSet(err,"zkipClientNew",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}
    
/**
 *  Destroy a zkipClient structure.
 */

DLLExport int
zkipClientFree (
  ZkError err,  
  ZkipClient * client )
{
  zkErrorClear(err);
  if (client) {
    if (*client) {
      if ((*client)->data) {
        if (!zkMemBufFree(err,&((*client)->data))) {
          return 0;
        }
      }
      if ((*client)->response) {
        if (!zkMemBufFree(err,&((*client)->response))) {
          return 0;
        }
      }
      if ((*client)->extra_data) {
        free((*client)->extra_data);
      }
    
      memset(*client,0,sizeof(struct T_ZkipClient));
      free(*client);
      *client = NULL;
      return 1;
    }
  }
  zkErrorSet(err,"zkipClientFree",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Make the zkip client object blocking.  This will modify the behaviour
 *  of certain zkipClient calls so that they will block instead of immediately
 *  returning.  If you wish this option, you must also specify a timeout
 *  in seconds.
 */

DLLExport int
zkipClientMakeBlocking (
  ZkError err ,
  ZkipClient client ,
  int timeout )
{
  zkErrorClear(err);
  if (client) {
    client->flags |= FLAG_BLOCKING;
    client->timeout = timeout;
    return 1;
  }
  zkErrorSet(err,"zkipClientMakeBlocking",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Make the zkipClient object nonblocking
 */

DLLExport int
zkipClientMakeNonBlocking (
  ZkError err ,
  ZkipClient client )
{
  zkErrorClear(err);
  if (client) {
    client->flags &= ~FLAG_BLOCKING;
    return 1;
  }
  zkErrorSet(err,"zkipClientMakeNonBlocking",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}


DLLExport int
zkipClientFd (
  ZkError err ,
  ZkipClient client ,
  int * fd )
{
  int sel = 0; /* Return code for the select() */
  zkErrorClear(err);
  if (fd && client) {
    if (client->fd < 0) {
      zkErrorSet(err,"zkipClientFd",ZKERR,ZKIP,ZKERROR,
                 "Socket descriptor is invalid");
      return 0;
    } else if (client->connect_status == CONNECT_STATUS_WAITING) {
      struct timeval timeval;
      fd_set writefds;
      timeval.tv_sec = 0;
      timeval.tv_usec = 0;
      FD_ZERO(&writefds);
      FD_SET(client->fd,&writefds);
      sel = select(client->fd + 1,NULL,&writefds,NULL,&timeval);
      if (sel > 0) {
        *fd = client->fd;
        return 1;
#if defined(WIN32)
      } else if ((sel < 0) && (WSAGetLastError() != WSAEINTR)) {
#else /* Some flavour of unix */
      } else if ((sel < 0) && (errno != EINTR)) {
#endif
        SafeClose(client->fd);
        client->fd = -1;
        client->connect_status = CONNECT_STATUS_ERROR;
        return 0;
      }
      zkErrorSet(err,"zkipClientFd",ZKWARN,ZKIP,ZKERROR,
                 "Still waiting for connection");
      return 0;
    } else if (client->connect_status != CONNECT_STATUS_ERROR) {
      *fd = client->fd;
      return 1;
    } else {
      zkErrorSet(err,"zkipClientFd",ZKERR,ZKIP,ZKERROR,
                 "Socket descriptor is invalid");
      return 0;
    }
  }
  zkErrorSet(err,"zkipClientFd",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}


/**
 *  Connect the ZKIP client structure to a remote ZKIP server.  If you 
 *  wish to use the client structure in non-blocing mode, use
 *  zkipClientFd() in order to get the socket descriptor to read and write
 *  to.
 */

DLLExport int
zkipClientConnectToHost (
  ZkError err ,
  ZkipClient client ,
  const char * hostname ,
  uint16_t port )
{
  struct hostent * host = NULL;
  struct sockaddr_in addr;
  unsigned long ip_addr = 0;

  zkErrorClear(err);
  if ((client == NULL) || (hostname == NULL)) {
    zkErrorSet(err,"zkipClientConnectToHost",ZKERR,ZKIP,ZKENULLPTR,NULL);
    return 0;
  }
  if (client->connect_status == CONNECT_STATUS_CONNECTED) {
    zkErrorSet(err,"zkipClientConnectToHost",ZKERR,ZKIP,ZKERROR,
               "ZkipClient structure is already connected to a server");
    return 0;
  }

  /*
   * If we have a dot notation IP address instead of a proper hostname, 
   * convert it to a proper address and dont do the DNS lookup
   */
  ip_addr = inet_addr(hostname);
  if (ip_addr != 0xFFFFFFFF) {
    memcpy(&(addr.sin_addr.s_addr),&ip_addr,sizeof(ip_addr));
  } else {

  /*
   * Figure out the internet address based on the hostname and connect using
   * the zkipClientConnectToAddress() function.
   */


    host = gethostbyname(hostname);         /* audit: ok */
    if (host == NULL) {
      zkErrorSet_va(err,"zkipClientConnectToHost",ZKERR,ZKIP,ZKENOTFOUND,
                    "Unable to resolve host: %s", hostname);
      return 0;
    }
    if (host->h_length != 4) {
      zkErrorSet(err,"zkipClientConnectToHost",ZKERR,ZKIP,ZKEDATABAD,
                 "Host address was not 4 bytes long (unknown format)");
      return 0;
    }
    memcpy(&(addr.sin_addr.s_addr), host->h_addr_list[0],sizeof(uint32_t));
  }

  strncpy(client->hostname,hostname,HOSTNAME_SZ);
  client->hostname[HOSTNAME_SZ] = (char) 0;
  client->port = port;
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);

  return zkipClientConnectToAddress(err,client,addr);
}

#ifdef FREEDOM_LINUX_CLIENT

#warning "This code is using zkConnectWithTimeout()"

static int
zkConnectWithTimeout (
  ZkError err ,
  int sockfd ,
  const struct sockaddr * serv_addr ,
  socklen_t addrlen ,
  int tout )
{
  static const char * func = "zkConnectWithTimeout";
  int r = 0;                 /* Function return value */
  int fd_flags = 0;          /* Stores the old flags */
  struct timeval timeout;    /* Connection timeout */
  fd_set wfds;               /* read to write descriptor set */
  int s = 0;                 /* select() return code */
  int sock_err = 0;          /* delayed connect() return code */
  socklen_t sock_err_sz = 0; /* Size of the socket option response */

  /*
   * First, try to get the current flags on the socket descriptor then
   * set it to non-blocking mode.
   */

  if ((fd_flags = fcntl(sockfd,F_GETFL)) < 0) {
     zkErrorSet_va(err,func,ZKERR,ZKIP,ZKESOCKET,
                   "fcntl() failed to retreive current flags:  %s",
                    strerror(errno));
  } else if ((fcntl(sockfd,F_SETFL,O_NONBLOCK) < 0)) {
     zkErrorSet_va(err,func,ZKERR,ZKIP,ZKESOCKET,
                   "fcntl() failed to make socket non-blocking:  %s",
                    strerror(errno));
  } else if (connect(sockfd,serv_addr,addrlen) != 0) {

    /*
     * We should be waiting for the connection to happen now...
     */

    if (errno == EINPROGRESS) {
      timeout.tv_sec = tout;
      timeout.tv_usec = 0;
      FD_ZERO(&wfds);
      FD_SET(sockfd,&wfds);
      while ((s = zkSelectUpdateTimeout(err,sockfd + 1,NULL,
                                        &wfds,NULL,&timeout)) <= 0) {
        /*
         * The connection to the ZKIP server has timed out.
         */

        if (s == 0) {
           zkErrorSet(err,func,ZKERR,ZKIP,ZKESOCKET,
                      "Connect() attempt has timed out");

        /*
         * The select has returned an error.  If it was interrupted, we'll just
         * go back into it until we either time out or get an error again.
         */

        } else if (errno == EINTR) {
          continue;
        } else {
          zkErrorSet_va(err,func,ZKERR,ZKIP,ZKESOCKET,
                        "select() failed while waiting for connection:  %s",
                        strerror(errno));
        }
        break;
      }

      /*
       * If there is something for me to read, the connection attempt is
       * complete.  If s is 0 or less, the error response will have been
       * set in the while loop.
       */

      if (s > 0) {
        sock_err_sz = sizeof(sock_err);
        if (getsockopt(sockfd,SOL_SOCKET,SO_ERROR,
                       (void *) &sock_err,&sock_err_sz) < 0) {
          zkErrorSet_va(err,func,ZKERR,ZKIP,ZKESOCKET,
                        "getsockopt() failed to get connect() error code: %s",
                        strerror(errno));
        } else if (sock_err != 0) {
          zkErrorSet_va(err,func,ZKERR,ZKIP,ZKESOCKET,
                        "Delayed connect() failed:  %s", strerror(sock_err));
        } else {  /* SUCCESS!!! */
          r = 1;
        }
      }

    /*
     * The connection failed for some other reason...
     */

    } else {
      zkErrorSet_va(err,func,ZKERR,ZKIP,ZKESOCKET,
                    "connect() failed: %s",strerror(errno));
    }
  }

  /*
   * Restore the socket flags to their original values.  
   */

  if (fd_flags >= 0) {
    if (fcntl(sockfd,F_SETFL,fd_flags) < 0) {
      zkErrorSet_va(err,func,ZKERR,ZKIP,ZKESOCKET,
                    "fcntl() failed to restore socket flags:  %s",
                    strerror(errno));
      r = 0;
    }
  }

  return r;
}
#endif /* FREEDOM_LINUX_CLIENT */

DLLExport int
zkipClientConnectToAddress (
  ZkError err ,
  ZkipClient client ,
  struct sockaddr_in addr )
{
#ifdef WIN32
  int win32err = 0; /* For keeping track of WSAGetLastError() */
  ULONG ulB = TRUE; /* Set socket to non-blocking mode */
#endif

  zkErrorClear(err);
  if (client == NULL) {
    zkErrorSet(err,"zkipClientConnectToAddress",ZKERR,ZKIP,ZKENULLPTR,NULL);
    return 0;
  }
  if (client->connect_status == CONNECT_STATUS_CONNECTED) {
    zkErrorSet(err,"zkipClientConnectToAddress",ZKERR,ZKIP,ZKERROR,
               "ZkipClient structure is already connected to a server");
    return 0;
  }
  client->fd = socket(AF_INET,SOCK_STREAM,0);
  if (client->fd < 0) {
    zkErrorSet_va(err,"zkipClientConnectToAddress",ZKERR,ZKIP,ZKESOCKET,
                  "Unable to create socket -> %s", strerror(errno));
    client->connect_status = CONNECT_STATUS_ERROR;    
    client->fd = -1;
    return 0;
  }

/*
Always block
*/

  client->flags |= FLAG_BLOCKING;

#ifdef FREEDOM_LINUX_CLIENT
  if (!zkConnectWithTimeout(err,client->fd,(struct sockaddr *) &addr, 
                           sizeof(struct sockaddr_in),client->timeout)) {
#else
  if (connect(client->fd,(struct sockaddr *) &addr, sizeof(struct sockaddr_in)) != 0) {
#endif /* ! FREEDOM_LINUX_CLIENT */
    zkErrorSet_va(err,"zkipClientConnectToAddress",ZKERR,ZKIP,ZKESOCKET, 
                  "connect() failed to %s:%d -> %s", inet_ntoa(addr.sin_addr), 
                  ntohs(addr.sin_port), strerror(errno));
    client->connect_status = CONNECT_STATUS_ERROR;
    SafeClose(client->fd);
    client->fd = -1;
    return 0;

  }else{
    client->connect_status = CONNECT_STATUS_CONNECTED;
    return 1;
  }
}

DLLExport int
zkipClientDisconnect (
  ZkError err ,
  ZkipClient client )
{
  zkErrorClear(err);
  if (client) {
    if (client->fd < 0) {
      zkErrorSet(err,"zkipClientDisconnect",ZKERR,ZKIP,ZKERROR,
                 "Client socket descriptor is invalid");
      return 0;
    }
    if (client->connect_status != CONNECT_STATUS_CONNECTED) {
      zkErrorSet(err,"zkipClientDisconnect",ZKERR,ZKIP,ZKERROR,
                 "Client is not connected");
      client->connect_status = CONNECT_STATUS_NOTCONNECTED;
      SafeClose(client->fd);
      client->fd = 0;
      return 0;
    }
    client->connect_status = CONNECT_STATUS_NOTCONNECTED;
    SafeClose(client->fd);
    client->fd = 0;
    return 1;
  }
  zkErrorSet(err,"zkipClientDisconnect",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}


/**
 *  Grok the response line and find out how much data is expected.  
 *
 */

DLLExport int
zkipClientVerifyResponse (
  ZkError err ,
  ZkipClient client )
{
  char * data = NULL;
  int retcode = 0; /* The function return code */
  size_t dlen = 0;  
  char * token[ZKIP_RESPONSE_TOKENS];  /* See top of the file */
  int i = 0; /* loop index */
  char * tmp = NULL; /* Used for stripping the trailing quote in message */

  zkErrorClear(err);
  if (client) {

    /*
     * Make sure the response line is NUL terminated
     */

    if (!zkMemBufAppend(err,client->response,(const void *)"\0",1)) {
      return 0;
    }
   
    if (!zkMemBufGetData(err,client->response,(void **) &data,&dlen)) {
      return 0;
    }

    /*
     * Get the response tokens 
     */

    for (i = 0; i < ZKIP_RESPONSE_TOKENS; i++) {
      if (i == TOKEN_MESSAGE) {
        token[i] = strtok(data,"\015\12");
      } else {
        token[i] = strtok(data," \t\015\12");
      }
      if (token[i] == NULL) {
        zkErrorSet(err,"zkipClientVerifyResponse",ZKERR,ZKIP,ZKEPARSE,
                   "Command response line is missing tokens");
        goto done;
      }
      data = NULL;
    }   

    /*
     * Parse the individual tokens.
     */

    if (!zkipParseInteger(err,(unsigned char *)token[TOKEN_TRANSID],
                          &(client->cmdresp.transid))) {
      zkErrorSetDetail(err,"Error parsing TRANSID");
      goto done;
    }
    if (!zkipParseCmdid(err,(unsigned char *)token[TOKEN_CMDID],
                        &(client->cmdresp.cmdid))) {
      zkErrorSetDetail(err,"Error parsing CMDID");
      goto done;
    }
    if (!zkipParseInteger(err,(unsigned char *)token[TOKEN_RESPONSE],
                          &(client->cmdresp.response))) {
      zkErrorSetDetail(err,"Error parsing TRANSID");
      goto done;
    }
    if (!zkipParseEncoding(err,(unsigned char *)token[TOKEN_ENCODING],
                          &(client->cmdresp.encoding))) {
      zkErrorSetDetail(err,"Error parsing ENCODING");
      goto done;
    }
    if (!zkipParseInteger(err,(unsigned char *)token[TOKEN_DATALEN],
                          &(client->cmdresp.datalen))) {
      zkErrorSetDetail(err,"Error parsing DATALEN");
      goto done;
    }
    strncpy(client->cmdresp.datatype,token[TOKEN_DATATYPE],DATATYPE_SZ);
    client->cmdresp.datatype[DATATYPE_SZ] = (char) 0;
  
    /*
     * Have to strip the quotes out of the response message
     */

    if ((token[TOKEN_MESSAGE] = strchr(token[TOKEN_MESSAGE],'"')) == NULL) {
      zkErrorSet(err,"zkipClientVerifyResponse",ZKERR,ZKIP,ZKEPARSE,
                 "Response message is not enclosed in double quotes");
      goto done;
    } 
    token[TOKEN_MESSAGE]++;
    if ((tmp = strrchr(token[TOKEN_MESSAGE],'"')) == NULL) {
      zkErrorSet(err,"zkipClientVerifyResponse",ZKERR,ZKIP,ZKEPARSE,
                 "Response message is not enclosed in double quotes");
      goto done;
    } 
    *tmp = (char) 0;
    strncpy(client->cmdresp.message,token[TOKEN_MESSAGE],MESSAGE_SZ);
    client->cmdresp.message[MESSAGE_SZ] = (char) 0;

    /*
     * do we have to read extra data? if so, set up the client structure
     * to suck it in, otherwise, advertise ourselves as being ready.
     */ 

    if (client->cmdresp.datalen > 0) {
      client->response_mode = RESPONSE_MODE_INCOMING;
      client->parse_mode = PARSE_MODE_DATA;
      client->state = 2;
    } else {
      client->response_mode = RESPONSE_MODE_COMPLETE;
      client->parse_mode = PARSE_MODE_RESPONSE;
      client->state = 0;
    }
    
    /*
     * Clean the extra data currently assigned to the client session 
     */
    if (client->extra_data) {
      free(client->extra_data);
      client->extra_data = NULL;
    } 
    retcode = 1;
done:
    if (!zkMemBufReset(err,client->response)) {
      return 0;
    }
    return retcode;
  }
  zkErrorSet(err,"zkipClientVerifyResponse",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

/**
 * Grok the incoming extra data.
 */

DLLExport int 
zkipClientVerifyData (
  ZkError err,
  ZkipClient client )
{
  char * extra_data = NULL;
  int retcode = 0; /* Function return code */
  size_t dlen = 0; 

  zkErrorClear(err);
  if (client) {
    if (client->extra_data) {
      free(client->extra_data);
    }
    if (!zkMemBufGetData(err,client->data,(void **)&extra_data,&dlen)) {
      return 0;
    }
    if (dlen > 4) {
      dlen -= 5; /* Ignore the trailing CRLF.CRLF sequence */
    } else {
      dlen = 0;  /* Obviously we didnt even get a full CRLF.CRLF */
    }
    if (dlen != client->cmdresp.datalen) {
      zkErrorSet_va(err,"zkipClientVerifyData",ZKERR,ZKIP,ZKEPARSE,
                   "The ZKIP client received %d bytes, expected %d instead",
                    dlen, client->cmdresp.datalen);
      goto done;
    }
    client->extra_data = (char *) malloc(dlen + 1);
    if (client->extra_data == NULL) {
      zkErrorSet(err,"zkipClientVerifyData",ZKERR,ZKIP,ZKENOMEM,"malloc");
      return 0;
    }
    memcpy(client->extra_data,extra_data,dlen);
    client->extra_data[dlen] = (char) 0;
    client->parse_mode = PARSE_MODE_RESPONSE;
    client->response_mode = RESPONSE_MODE_COMPLETE;
    client->state = 0;
    retcode = 1;
done:
    if (!zkMemBufReset(err,client->data)) {
      return 0;
    }
    return retcode;
  }
  zkErrorSet(err,"zkipClientVerifyData",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkipClientIsReady (
  ZkError err ,
  ZkipClient client ,
  struct ZkipCommandResponse ** response ,
  char ** extra_data )
{
  zkErrorClear(err);
  if (client && response && extra_data) {
    if (client->response_mode == RESPONSE_MODE_COMPLETE) {
      *response = &(client->cmdresp);
      *extra_data = client->extra_data;
      client->response_mode = RESPONSE_MODE_INCOMING;
      client->state = 0;
      client->parse_mode = PARSE_MODE_RESPONSE;
      return 1;
    }
    *response = NULL;
    *extra_data = NULL;
    return 0;   
  }
  zkErrorSet(err,"zkipClientIsReady",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}



/**
 *  Add some data into the zkipClient structure.
 */

DLLExport int
zkipClientAddData (
  ZkError err ,
  ZkipClient client ,
  const char * data ,
  size_t len ,
  size_t * used )
{
  static const char * response_term = "\015\012";
  static const char * data_term = "XX\015\012.\015\012";
  size_t i = 0; /* Loop index */

  zkErrorClear(err);
  if (client && data && used) {
    *used = 0;
    switch (client->parse_mode) {
    case PARSE_MODE_BANNER:
    case PARSE_MODE_RESPONSE:
      for (i = 0; i < len; i++) {
        if (client->state >= 2) {
           break;
        }
        if (data[i] == response_term[client->state]) {
          client->state += 1;
        }
      }
      *used += i;
      client->response_len += i;
      if (!zkMemBufAppend(err,client->response,(const void *)data,i)) {
        client->parse_mode = PARSE_MODE_ERROR;
        return 0;
      }
      if (client->state >= 2) {
        if (!zkipClientVerifyResponse(err,client)) {
          client->parse_mode = PARSE_MODE_ERROR;
          client->response_mode = RESPONSE_MODE_ERROR;
          return 0;
        }
      }
  
      /*
       * If we have a chunk of the extra data in the buffer, make sure
       * we grab that as well...
       */
      if ((i < len) && (client->cmdresp.datalen > 0) && 
         (client->parse_mode == PARSE_MODE_DATA)) {
        zkErrorSet(err,"zkipClientAddData",ZKWARN,ZKIP,0,
                   "Extra data in the same buffer as command response");
        data = &(data[i]);
        len -= i;
      } else {
        break;
      }
    case PARSE_MODE_DATA:
      for (i = 0; i < len; i++) {
        if (client->state >= 7) {
          break;
        }
        if (data[i] == data_term[client->state]) {
          client->state += 1;
        } else if (data[i] == (char) 0x0D) {
          client->state = 3;
        } else {
          client->state = 2;
        }
      }
      *used += i;
      client->data_len += i;
      if (!zkMemBufAppend(err,client->data,(const void *)data,i)) {
        client->parse_mode = PARSE_MODE_ERROR;
        return 0;
      }
      if (client->state >= 7) {
        if (!zkipClientVerifyData(err,client)) {
          client->parse_mode = PARSE_MODE_ERROR;
          client->response_mode = RESPONSE_MODE_ERROR;
          return 0;
        }
      }  
      break;
    case PARSE_MODE_ERROR:
      zkErrorSet(err,"zkipClientAddData",ZKERR,ZKIP,ZKEPARSE,
                 "ZKIP client is in error state");
      return 0;
    default:
      zkErrorSet(err,"zkipClientAddData",ZKERR,ZKIP,ZKEPARSE,
                 "ZKIP client is in an unknown state");
      client->response_mode = RESPONSE_MODE_ERROR;
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkipClientAddData",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

#define BUFFER_SZ 1024

DLLExport int
zkipClientWrite (
  ZkError err ,
  ZkipClient client ,
  const char * data ,
  int datalen )
{
  fd_set writefds;
  struct timeval timeout;
  int bytes = 0;
  int sel = 0;

  zkErrorClear(err);
  if (client && data) {
    if (client->connect_status != CONNECT_STATUS_CONNECTED) {
      zkErrorSet(err,"zkipClientWrite",ZKERR,ZKIP,ZKERROR,
                 "ZKIP client is not in connected status");
      return 0;
    }
    for (;;) {
      FD_ZERO(&writefds);
      FD_SET(client->fd,&writefds);
      timeout.tv_sec = client->timeout;
      timeout.tv_usec = 0;
      sel = select(client->fd + 1,NULL,&writefds,NULL,&timeout);
      if (sel > 0) {

/*#ifdef WIN32		
		if(0 == WriteFile(client->fd,
				data,
				datalen,
				&bytes,
				NULL)){

			bytes = 0;
		}
#else*/
		bytes = send(client->fd,data,datalen,0);
//#endif

        if (bytes <= 0) {
#if defined(WIN32) 
          if ((bytes == 0) && (WSAGetLastError() == WSAEINTR)) {
#else
          if ((bytes == 0) && (errno == EINTR)) {
#endif
            /* interupted system call */
            continue;
          }
          zkErrorSet(err,"zkipClientWrite",ZKERR,ZKIP,ZKEWRITE,
                     "Could not write data to remote server");
          return 0;
        } 
        if (bytes == datalen) {
          return 1;
        }
        data = &(data[bytes]);
        datalen -= bytes;
        continue;
#if defined(WIN32)
      } else if ((sel < 0) && (WSAGetLastError() == WSAEINTR)) {
#else
      } else if ((sel < 0) && (errno == EINTR)) {
#endif
        /* interupted system call */
        continue;
      } 
      /* Timeout */
      zkErrorSet(err,"zkipClientWrite",ZKERR,ZKIP,ZKERROR,
                 "Connection to remote server has timed out");
      return 0;
    }
  }
  zkErrorSet(err,"zkipClientWrite",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}


DLLExport int
zkipClientReadResponse (
  ZkError err ,
  ZkipClient client ,
  struct ZkipCommandResponse ** response ,
  char ** extra_data )
{
#define READ_BUFFER_SZ 4096
  fd_set readfds; /* read descriptor field */
  struct timeval timeout; /* Timeout for select */
  int bytes = 0; /* Bytes read */
  char buffer[READ_BUFFER_SZ]; /* Read buffer */
  size_t used = 0; /* Number of bytes used from input buffer */
  int sel = 0;  /* Return code from select() */
#if defined(WIN32)
  int win32err = 0; /* For keeping track of the WIN32 last error */
#endif

  zkErrorClear(err);
  if (client && response && extra_data) {
    *response = NULL;
    *extra_data = NULL;
    if (client->connect_status != CONNECT_STATUS_CONNECTED) {
      zkErrorSet(err,"zkipCLientReadResponse",ZKERR,ZKIP,ZKERROR,
                 "ZKIP client is not in connected status");
      return 0;
    } 
    for (;;) {
      FD_ZERO(&readfds);
      FD_SET(client->fd,&readfds);
      timeout.tv_sec = client->timeout;
      timeout.tv_usec = 0;
      sel = select(client->fd + 1,&readfds,NULL,NULL,&timeout);
      if (sel > 0) {
        
/*#ifdef WIN32
		   if(0 == ReadFile(client->fd,
                     buffer,
                     READ_BUFFER_SZ,
                     &bytes,
                     NULL)){
			   bytes = 0;
		   }
#else*/
			bytes = recv(client->fd,buffer,READ_BUFFER_SZ,0);
//#endif


#ifdef WIN32
#ifndef NDEBUG
		{
		char buf[100] = "";
		_snprintf(buf, 95, "bytes == [%d] WSAGetLastError(%d)\n", bytes, WSAGetLastError());
		ClientTrace(buf, "zkipClientReadResponse()");
		}
#endif /* NDEBUG */
#endif /* WIN32 */

        if (bytes > 0) {
          if (!zkipClientAddData(err,client,buffer,(size_t)bytes,&used)) {
            return 0;
          }
          if (zkipClientIsReady(err,client,response,extra_data)) {
            return 1;
          }
#if defined(WIN32)
        } else if (bytes < 0) {
          win32err = WSAGetLastError();
          switch (win32err) {
          case WSAEINTR:
            zkErrorSet_va(err,"zkipClientReadResponse",ZKWARN,ZKIP,ZKENODATA,
                          "recv: Trapped and ignored WSA error %d",win32err);
            break;
          default:
            zkErrorSet_va(err,"zkipClientReadResponse",ZKERR,ZKIP,ZKERROR,
                          "No data to read (WSA error is %d)",win32err);
            return 0;
          }
          continue;  
#else /* error handling for unix */
        } else if ((bytes < 0) && (errno == EINTR)) {
          continue;
        } else if (bytes < 0) {
          zkErrorSet(err,"zkipClientReadResponse",ZKERR,ZKIP,ZKERROR,
                     "No data to read");
          return 0;
#endif
        } else {  /* Bytes equals zero */
          zkErrorSet(err,"zkipClientReadResponse",ZKWARN,ZKIP,ZKENODATA,
                     "Select returned with 0 bytes to read");
          return 0;
        }
      } else if (sel == 0) {
        zkErrorSet(err,"zkipClientReadResponse",ZKERR,ZKIP,ZKERROR,
                   "Connection to remote server has timed out");
        return 0;
      } else if (sel < 0) {
#if defined(WIN32)
        if (WSAGetLastError() == WSAEINTR) {
#else
        if (errno == EINTR) {
#endif
          /* Interrupted system call */
        } else {
          zkErrorSet(err,"zkipClientResponse",ZKERR,ZKIP,ZKERROR,
                     "Error reading from remote server");  
          return 0;
        }
      }
    }
  }
  zkErrorSet(err,"zkipClientReadResponse",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}
