/*
 * $Id: zkChunk.c,v 1.5.32.1 2000/10/11 13:20:28 phil Exp $
 */

static const char * module_id_ = "$Id: zkChunk.c,v 1.5.32.1 2000/10/11 13:20:28 phil Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* !WIN32 */
#include <string.h>
#include <zkError.h>
#include <zkLog.h>
#include "zkChunk.h"

struct T_Chunk {
  char * data;
  size_t tail;
  size_t size;
  struct T_Chunk * link;
};

typedef struct T_Chunk Chunk;

#define DEFAULT_CHUNKSZ      4096

struct T_ZkChunk {
  size_t size;               /* Total number of bytes in the chunks */
  size_t chunk_size;         /* Size of newly generated chunks */ 
  Chunk * head;              /* Head of the chunk list */
  Chunk * tail;              /* Tail (insertion) pointer of the chunk list */
};

/*
 * Internal function prototypes 
 */

static int chunkNew ( ZkError , Chunk ** , size_t );

/*
 * Functions
 */

static int
chunkNew (
  ZkError err ,
  Chunk ** chunk ,
  size_t size )
{
  (void) module_id_;
  if ((*chunk = (Chunk *) malloc(sizeof(struct T_Chunk))) == NULL) {
    zkErrorSet(err,"chunkNew (internal function)",ZKERR,ZKMHTTP,ZKENOMEM,
               "malloc");
    return 0;
  }
  memset(*chunk,0,sizeof(struct T_Chunk));
  if (((*chunk)->data = (void *) malloc(size + 1)) == NULL) {
    zkErrorSet(err,"chunkNew (internal function)",ZKERR,ZKMHTTP,ZKENOMEM,
               "malloc (for data component)");
    free(*chunk);
    *chunk = NULL;
    return 0;
  }  
  memset((*chunk)->data,0,size + 1);
  (*chunk)->size = size;
  (*chunk)->tail = 0;
  (*chunk)->link = NULL;
  return 1;
} 

DLLExport int
zkChunkNew (
  ZkError err ,
  ZkChunk * chunk )
{
  zkErrorClear(err);
  if (chunk) {
    *chunk = (struct T_ZkChunk *) malloc(sizeof(struct T_ZkChunk));
    if (*chunk == NULL) {
      zkErrorSet(err,"zkChunkNew",ZKERR,ZKMHTTP,ZKENOMEM,"malloc");
      return 0;
    }
    memset(*chunk,0,sizeof(struct T_ZkChunk));
    (*chunk)->chunk_size = DEFAULT_CHUNKSZ;
    if (!chunkNew(err,&((*chunk)->head),(*chunk)->chunk_size)) {
      free(*chunk);
      *chunk = NULL;
      return 0;
    }
    (*chunk)->size = 0;
    (*chunk)->tail = (*chunk)->head; 
    return 1;
  }
  zkErrorSet(err,"zkChunkNew",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkChunkFree (
  ZkError err ,
  ZkChunk * chunk )
{
  Chunk * current = NULL;
  Chunk * next = NULL;

  zkErrorClear(err);
  if (chunk) {
    if (*chunk) {
    
      /*
       * Nuke the list of mini chunks
       */

      current = (*chunk)->head;
      while (current) {
        next = current->link;
        if (current->data) {
          memset(current->data,0,current->tail);
          free(current->data);
        }
        free(current);
        current = next;
      }
   
      /*
       * Nuke the chunk itself
       */

      memset(*chunk,0,sizeof(struct T_ZkChunk));
      free(*chunk);
      *chunk = NULL;
      return 1;
    }
  }
  zkErrorSet(err,"zkChunkFree",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkChunkSetBlockSize (
  ZkError err ,
  ZkChunk chunk ,
  size_t bsz )
{
  zkErrorClear(err);
  if (chunk) {
    chunk->chunk_size = bsz;
    return 1;
  }
  zkErrorSet(err,"zkChunkSetBlockSize",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkChunkAppend (
  ZkError err ,
  ZkChunk chunk ,
  const char * data ,
  size_t size )
{
  size_t offset = 0;
  size_t sz = 0;
  Chunk * c = NULL;

  zkErrorClear(err);
  if (chunk && data) {
    do {
      c = chunk->tail;
      sz = ((c->size-c->tail)<(size-offset))?(c->size-c->tail):(size-offset);
      memcpy(&(c->data[c->tail]),&(data[offset]),sz);
      c->tail += sz;
      offset += sz;
      chunk->size += sz;
      if (c->tail >= c->size) {
        if (!chunkNew(err,&(c->link),chunk->chunk_size)) {
          return 0;
        }
        chunk->tail = c->link;
      }
    } while (offset < size); 
    return 1;
  }
  zkErrorSet(err,"zkChunkAppend",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
  return 0;
}

/**
 * This function will essentially reallocate the first mini chunk of data
 * to hold the complete string up to now and return a pointer to that data.
 */

DLLExport int
zkChunkGetData (
  ZkError err ,
  ZkChunk chunk ,
  char ** data ,
  size_t * size )
{
  char * d = NULL;
  Chunk * c = NULL;
  Chunk * p = NULL;
  size_t offset = 0;
  

  zkErrorClear(err);
  if (chunk && data && size) {
    *data = NULL;
    *size = 0;
    if ((d = (void *) malloc(chunk->size + 1)) == NULL) {
      zkErrorSet(err,"zkChunkGetData",ZKERR,ZKMHTTP,ZKENOMEM,"malloc");
      return 0;
    }
    c = chunk->head;
    while (c) {
      memcpy(&(d[offset]),c->data,c->tail);
      offset += c->tail;
      p = c->link;
      memset(c->data,0,c->tail);
      free(c->data);
      if (c != chunk->head) {
        memset(c,0,sizeof(Chunk));
        free(c);
      }
      c = p;
    }
    d[offset] = (char) 0;
    c = chunk->head;
    chunk->tail = c;
    c->data = (void *) d;
    c->size = offset;
    c->tail = offset;
    c->link = NULL; 
    *data = d;
    *size = offset;
    return 1;
  }
  zkErrorSet(err,"zkChunkGetData",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  This gets back a data pointer for the contents of the chunk list but also
 *  removes the data from the chunk list so you can destroy it without maiming
 *  the data buffer.
 */

DLLExport int
zkChunkExtractData (
  ZkError err ,
  ZkChunk chunk ,
  char ** data ,
  size_t * size )
{
  Chunk * c = NULL;
  zkErrorClear(err);
  if (chunk && data && size) {
    if (zkChunkGetData(err,chunk,data,size)) {
      c = chunk->head;
      *data = c->data;
      *size = c->size;
      free(c);
      c = NULL;
      if (!chunkNew(err,&c,chunk->chunk_size)) {
        return 0;
      }
      chunk->head = c;
      chunk->tail = c;
      chunk->size = 0;
      return 1;
    }
    return 0;
  }
  zkErrorSet(err,"zkChunkExtractData",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}
