/*
 * $Id: zkHTTPField.c,v 1.6.32.1 2000/10/11 13:20:28 phil Exp $
 */

static const char * module_id_ = "$Id: zkHTTPField.c,v 1.6.32.1 2000/10/11 13:20:28 phil Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zkIntTypes.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkMIME.h>

/**
 *  Create a new ZkHTTPField structure.
 *
 *  @param err Returns the error if one occured
 *  @param field Returns a new ZkHTTPField object
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */


DLLExport int
zkHTTPFieldNew (
  ZkError err ,
  ZkHTTPField * field )
{
  (void) module_id_; 
  zkErrorClear(err);
  if (!zkEmailFieldNew(err,(ZkEmailField *) field)) {
    return 0;
  }
  if (!zkEmailFieldDontStripComments(err,*field)) {
    return 0;
  }
  return 1;
}

/**
 *  Destroy a ZkHTTPField structure.
 *
 *  @param err Returns the error if one occured
 *  @param field The ZkHTTPField object to destroy.  The pointer will be
 *               reset to NULL on success
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkHTTPFieldFree (
  ZkError err ,
  ZkHTTPField * field )
{
  zkErrorClear(err);
  if (!zkEmailFieldFree(err,(ZkEmailField *) field )) {
    return 0;
  }
  return 1;
}

/**
 *  Retrieve a MIME parameter from the given field object.
 *
 *  @param err Returns the error if one occured
 *  @param field The ZkHTTPField to analyze
 *  @param name The name of the parameter you want (NUL terminated string)
 *  @param value Returns the value of that parameter (A pointer to a NUL
 *               terminated string contained within the ZkHTTPField.  dont
 *               mess with it)
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success ( the parameter was found) or a 0
 *          on failure (the parameter was not found)
 */

DLLExport int
zkHTTPFieldGetParameter (
  ZkError err ,
  ZkHTTPField field ,
  const unsigned char * name ,
  const unsigned char ** value )
{
  zkErrorClear(err);
  if (!zkEmailFieldGetParameter(err,(ZkEmailField) field,name,value)) {
    return 0;
  }
  return 1;
}

/**
 *  This returns the link to the next field in a field list (ZkHTTPFields are
 *  stored as lists).  Do not use this function directly.  In order to search
 *  fields in a mail header use the zkHTTPHeaderFirstField(),
 *  zkHTTPHeaderNextField(), and zkHTTPHeaderGetField() functions.
 *
 *  @param err Returns the error if one occured
 *  @param field The field object to modify
 *  @param link Returns the pointer to the next field in the field list
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure
 */


DLLExport int
zkHTTPFieldLink (
  ZkError err ,
  ZkHTTPField field ,
  ZkHTTPField * link )
{
  zkErrorClear(err);
  if (!zkEmailFieldLink(err,(ZkEmailField) field,(ZkEmailField *) link)) {
    return 0;
  }
  return 1;
}

/**
 *  This function is used to set the link to the next field in the field list.
 *  Do not use this function directly, instead, to add a field to a mail
 *  header, use the zkHTTPHeaderAddField() or zkHTTPHeaderAddFieldData()
 *  functions.
 *
 *  @param err Returns the error if one occured
 *  @param field The ZkHTTPField to modify
 *  @param link The next email field in the list
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkHTTPFieldSetLink (
  ZkError err ,
  ZkHTTPField field ,
  ZkHTTPField link )
{
  zkErrorClear(err);
  if (!zkEmailFieldSetLink(err,(ZkEmailField)field,(ZkEmailField) link)) {
    return 0;
  }
  return 1;
}

/**
 *  This deletes fields.  It doesnt work so dont use it, 'nuff said
 */


DLLExport int
zkHTTPFieldDelete (
  ZkError err ,
  ZkHTTPField * field )
{
  zkErrorClear(err);
  if (!zkEmailFieldDelete(err,(ZkEmailField *) field)) {
    return 0;
  }
  return 1;
}

/**
 * Force the value of specific fields within the HTTP header.
 * See the zkEmailFieldForceValue function for more details.
 *
 */

DLLExport int
zkHTTPFieldForceValue (
  ZkError err ,
  ZkHTTPField field ,
  enum ZkMIMEFieldId id ,
  const char * value ,
  size_t size )
{
  zkErrorClear(err);
  if (!zkEmailFieldForceValue(err,(ZkEmailField) field,id,value,size)) {
    return 0;
  }
  return 1;
}

/**
 *  Remove fields which have been marked as potentially compromising.
 *  Use the function zkHTTPHeaderStripField() instead in order to clean
 *  up HTTP headers.  The flags can be set to 0 or DONT_STRIP_UNKNOWN
 *
 *  @param err Returns the error if one occured
 *  @param head The first field in the ZkHTTPField list to strip
 *  @param flags Strip flags (see the zkMIME.h file for the list)
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkHTTPFieldStrip (
  ZkError err,
  ZkHTTPField * head ,
  uint32_t flags)
{
  zkErrorClear(err);
  if (!zkEmailFieldStrip(err,(ZkEmailField *) head,flags)) {
    return 0;
  }
  return 1;
}

/**
 *  This function is used to set the name string of the field.  The name string
 *  must be NUL terminated.  When the name is set, the function will try to
 *  determine the matching ID for the field from the list in zkMIME.h
 *
 *  @param err Returns the error if one occured
 *  @param field The ZkHTTPField for which to set the name
 *  @param name The NUL terminated name string
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkHTTPFieldSetName (
  ZkError err ,
  ZkHTTPField field ,
  const unsigned char * name )
{
  zkErrorClear(err);
  if (!zkEmailFieldSetName(err,(ZkEmailField)field,name)) {
    return 0;
  }
  return 1;
}

/**
 *  Returns the NUL terminated name string for the specified field.
 *
 *  @param err Returns the error if one occured
 *  @param field The ZkHTTPField for which you wish to have the name string
 *  @param name Returns a pointer to a NUL terminated string contained inside
 *              The field object.  Do not mess with this string.
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkHTTPFieldName (
  ZkError err ,
  ZkHTTPField field ,
  const unsigned char ** name )
{
  zkErrorClear(err);
  if (!zkEmailFieldName(err,(ZkEmailField)field,name)) {
    return 0;
  }
  return 1;
}


DLLExport int
zkHTTPFieldSetValue (
  ZkError err,
  ZkHTTPField field ,
  const unsigned char * value )
{
  zkErrorClear(err);
  if (!zkEmailFieldSetValue(err,(ZkEmailField)field,value)) {
    return 0;
  }
  return 1;
}

/**
 *  Set the HTTP field value.  
 *
 *  @param err Returns the error if one occured
 *  @param field The ZkHTTPField for which to set the value string.
 *  @param value The NUL terminated value string
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure
 */


DLLExport int
zkHTTPFieldValue (
  ZkError err ,
  ZkHTTPField field ,
  const unsigned char ** value )
{
  zkErrorClear(err);
  if (!zkEmailFieldValue(err,(ZkEmailField) field,value)) {
    return 0;
  }
  return 1;
}

/**
 *  Return the field ID for the ZkHTTPField.  This id is one of the elements
 *  in the list described in the file zkMIME.h
 *
 *  @param err Returns the error if one occured
 *  @param field The ZkHTTPField to analyze
 *  @param id Returns the field id
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */


DLLExport int
zkHTTPFieldId (
  ZkError err ,
  ZkHTTPField field,
  enum ZkMIMEFieldId * id )
{
  zkErrorClear(err);
  if (!zkEmailFieldId(err,(ZkEmailField) field,id)) {
    return 0;
  }
  return 1;
}

/**
 *  This sets the field id.  You must use one of the elements from the list in
 *  zkMIME.h.  When the field ID is set, the name string will be updated to
 *  match the new id.
 *
 *  @param err Returns the error if one occured.
 *  @param field The ZkHTTPField to modify
 *  @param id The new id for the field object
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkHTTPFieldSetId (
  ZkError err,
  ZkHTTPField field ,
  enum ZkMIMEFieldId id )
{
  zkErrorClear(err);
  if (!zkEmailFieldSetId(err,(ZkEmailField)field,id)) {
    return 0;
  }
  return 1;
}

/**
 *  Add a parameter to the given HTTP field.  You should only do this to MIME
 *  fields. NOTE:  Cookie field value are currently not parsed!!! 
 *
 *  @param err Returns the error if one occured
 *  @param field The ZkEmailField to which to add the parameter
 *  @param name The NUL terminated string which is the name of the parameter
 *  @param value The NUL terminated string which is the value of the parameter
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */
 
DLLExport int
zkHTTPFieldAddParameter (
  ZkError err ,
  ZkHTTPField field ,
  const unsigned char * name ,
  const unsigned char * value )
{
  zkErrorClear(err);
  if (!zkEmailFieldAddParameter(err,(ZkEmailField) field,name,value)) {
    return 0;
  }
  return 1;
}

/**
 *  Parse the string into an RFC822/MIME/HTTP field.  You probably don need to
 *  call this yourself as all of the parsing is handled by the 
 *  zkHTTPHeaderParse() function.
 *
 *  @param err  Returns the error if one occured
 *  @param field The ZkHTTPField object which will hold the result of the parse
 *  @param line The NUL terminated line of text which will be parsed in order
 *              to load the ZkHTTPField object.  Note that it is assumed here
 *              that the line has already been unfolded.
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkHTTPFieldParse (
  ZkError err ,
  ZkHTTPField field ,
  const unsigned char * line )
{
  zkErrorClear(err);
  if (!zkEmailFieldParse(err,(ZkEmailField)field,line)) {
    return 0;
  }
  return 1;
}

/**
 *  Dump the HTTP field to the given callback.  See the zkEmailDump() function
 *  for more details
 *
 *  @param err Returns the error if one occured
 *  @param field The ZkHTTPField object to dump
 *  @param callback The callback function to use for dumping the field
 *  @param cdata The extra data pointer for the callback function
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkHTTPFieldDump (
  ZkError err ,
  ZkHTTPField field ,
  ZkEmailDumpCallback callback ,
  void * cdata )
{
  zkErrorClear(err);
  if (!zkEmailFieldDump(err,(ZkEmailField) field,callback,cdata)) {
    return 0;
  }
  return 1;
}

/**
 *  This function makes a complete copy of a ZkHTTPField.  Note that if you
 *  wish to make a copy of a a specific field (recall that fields are actually
 *  lists) you should get the link of the field using the zkHTTPFieldLink()
 *  function, then reset the link to NULL using a call like:
 *  zkHTTpFieldSetLink(err,field,NULL)
 *  make a duplicate of the field using this function then reset the link of
 *  the original to what it previously was.
 *  Otherwise, the field list beginning with the specified field is copied
 *  completely.
 *
 *  @param err Returns the error if one occured
 *  @param field The ZkHTTPField to copy
 *  @param copy Returns a new copy of the field
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkHTTPFieldDup (
  ZkError err ,
  ZkHTTPField field ,
  ZkHTTPField * copy )
{
  zkErrorClear(err);
  if (!zkEmailFieldDup(err,(ZkEmailField) field,(ZkEmailField *) copy)) {
    return 0;
  }
  return 1;
}

/**
 * Returns the flags which identify the type of field.  These flags are
 * defined in zkEmailField.h:
 *   FL_DELETE       Field which should be thrown away
 *   FL_RFC822       A Mail field described in RFC822
 *   FL_RFC1327      A Mail extension field described in RFC1327
 *   FL_MIME         A MIME field
 *   FL_ZKS_FREEDOM  A Zero-Knowledge Freedom product field
 *   FL_ZKS          Some other Zero-Knowledge mail field
 *   FL_EXTENSION    Some other extension field
 *   FL_UNKNOWN      Unknown email header field
 *   FL_RFC1945      an HTTP 1.0 field
 *   FL_COOKIE       an HTTP cookie field (rfc2109)
 *
 *
 * @param err Returns the error if one occured
 * @param field The ZkHTTPField to analyze
 * @param flags Returns the flags associated with this field
 * @author Philippe Boucher
 * @version 1.0
 * @return A non-zero value on success, a 0 on failure
 */


DLLExport int
zkHTTPFieldIdFlags (
  ZkError err ,
  ZkHTTPField field ,
  uint32_t * flags )
{
  zkErrorClear(err);
  if (!zkEmailFieldIdFlags(err,(ZkEmailField) field,flags)) {
    return 0;
  }
  return 1;
}
