/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>

#ifdef WIN32
#include <winsock2.h>
#else
#include <netinet/in.h>
#endif

#include <zkIntTypes.h>
#include <zkPacket.h>

/**
 *  Compute the checksum of an ICMP datagram.  The checksum will be written 
 *  inside the packet itself
 * 
 *  @param icmph The ICMP header pointer for which to return the check sum.
 *  @param datalen The number of bytes for which to compute the checksum
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return The header checksum
 */

DLLExport uint16_t
zkICMPHeaderChecksum (
  ZkICMPHeader * icmph ,
  size_t datalen ) /* Number of bytes */
{
  unsigned long cksum = 0;
  unsigned short * data = NULL;
  size_t i = 0;

  if (!icmph) {
    return 0;
  }
  icmph->CheckSum = 0;
  data = (unsigned short *) icmph;
  datalen /= 2;
  cksum = 0;
  for (i = 0; i < datalen; i++) {
    cksum += *data++;
  }
  cksum = (cksum >> 16) + (cksum & 0xFFFF);
  cksum += (cksum >> 16);
  icmph->CheckSum = ~cksum;

  return icmph->CheckSum;
}

/*
 * The following ICMP support functions are hacks to do some network magic
 *
 */

/**
 * This function creates an ICMP fragmentation needed message which is can
 * be sent back to a remote host.
 *
 * @param iph The offending IP packet you wish to get fragmented
 * @param mesg Buffer to hold the message
 * @param size Holds the original size of buffer, returns bytes used
 * @author Philippe Boucher
 * @version 1.0
 * @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkICMPCreateFragmentationNeeded (
  ZkIPHeader * iph ,
  unsigned char * mesg ,
  size_t * size )
{
  return zkICMPCreateDestinationUnreachable(iph,
                                            ZK_ICMP_FRAGMENTATION_NEEDED,
                                            mesg,size);
}

/**
 * Function to create a generic time exceeded message
 *
 */

DLLExport int
zkICMPCreateTimeExceeded (
  ZkIPHeader * iph ,
  uint8_t code ,
  unsigned char * mesg ,
  size_t * size )
{
  uint16_t totlen = 0;
  ZkIPHeader * header = NULL;

  header = (ZkIPHeader *) mesg;

  /*
   * ip header (20) + icmp header (8) + returned iph (ihl * 4) + 8 bytes
   */

  totlen = iph->ihl * 4 + 36;  
  if (*size < (size_t) totlen) {
    return 0;
  }
  *size = (size_t) totlen;

  /*
   * Make the return IP datagram
   */

  mesg[0] = 0x45; /* Version and header length */
  mesg[1] = 0; /* Type of service */
  mesg[2] = totlen / 0x100; /* First byte of total packet length */
  mesg[3] = totlen % 0x100; /* Second byte of total packet length */
  mesg[4] = 0xfe; /* First half of identification */
  mesg[5] = 0xed; /* Second half of identification */
  mesg[6] = 0; /* Fragmentation flags and first half of frag offset */
  mesg[7] = 0; /* Second half of frag offset */
  mesg[8] = 255; /* Time to live */
  mesg[9] = ZK_IPPROTO_ICMP; /* IP Protocol type */
  mesg[10] = 0; /*First half of checksum */
  mesg[11] = 0; /* Second half of checksum */
  memcpy((void *)&(mesg[12]),(void *)&(iph->DstIpAddress),4);
  memcpy((void *)&(mesg[16]),(void *)&(iph->SrcIpAddress),4);
  
  /*
   * ICMP Header  
   */

  mesg[20] = 11; /* Type = Time exceeded */
  mesg[21] = code; /* Code (see zkPacket.h for list) */
  mesg[22] = 0; /* First half of ICMP checksum */
  mesg[23] = 0; /* Second half of ICMP checksum */
  mesg[24] = 0; /* unused, must be 0 */
  mesg[25] = 0; /* unused, must be 0 */
  mesg[26] = 0; /* unused, must be 0 */
  mesg[27] = 0; /* unused, must be 0 */
 
  /*
   * IP Header + first 64 bits of the datagram
   */

  memcpy((void *) &(mesg[28]),(void *) iph,iph->ihl * 4 + 8); 

  /*
   * Compute the ICMP header checksum
   */

  zkICMPHeaderChecksum((ZkICMPHeader *)&(mesg[20]),iph->ihl * 4 + 16);

  /*
   * Compute the IP header checksum 
   */

  zkIPHeaderChecksum(header);

  /*
   * All done.  the buffer can be dropped onto the network (with a raw socket)
   */

  return 1;
}

/**
 * Function to generate generic Destination Unreachable ICMP messages.
 * 
 *  @param iph The offending IP header
 *  @param code The message code (see the list in zkPacket.h)
 *  @param mesg Buffer to hold the ICMP message
 *  @param size The size of the buffer, returns the number of bytes used
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkICMPCreateDestinationUnreachable (
  ZkIPHeader * iph ,
  uint8_t code ,
  unsigned char * mesg ,
  size_t * size )
{
  uint16_t totlen = 0;
  ZkIPHeader * header = NULL;

  header = (ZkIPHeader *) mesg;

  /*
   * ip header (20) + icmp header (8) + returned iph (ihl * 4) + 8 bytes
   */

  totlen = iph->ihl * 4 + 36;  
  if (*size < (size_t) totlen) {
    return 0;
  }
  *size = (size_t) totlen;

  /*
   * Make the return IP datagram
   */

  mesg[0] = 0x45; /* Version and header length */
  mesg[1] = 0; /* Type of service */
  mesg[2] = totlen / 0x100; /* First byte of total packet length */
  mesg[3] = totlen % 0x100; /* Second byte of total packet length */
  mesg[4] = 0xfe; /* First half of identification */
  mesg[5] = 0xed; /* Second half of identification */
  mesg[6] = 0; /* Fragmentation flags and first half of frag offset */
  mesg[7] = 0; /* Second half of frag offset */
  mesg[8] = 255; /* Time to live */
  mesg[9] = ZK_IPPROTO_ICMP; /* IP Protocol type */
  mesg[10] = 0; /*First half of checksum */
  mesg[11] = 0; /* Second half of checksum */
  memcpy((void *)&(mesg[12]),(void *)&(iph->DstIpAddress),4);
  memcpy((void *)&(mesg[16]),(void *)&(iph->SrcIpAddress),4);
  
  /*
   * ICMP Header  
   */

  mesg[20] = 3; /* Type = destination unreachable */
  mesg[21] = code; /* Code (see zkPacket.h for list) */
  mesg[22] = 0; /* First half of ICMP checksum */
  mesg[23] = 0; /* Second half of ICMP checksum */
  mesg[24] = 0; /* unused, must be 0 */
  mesg[25] = 0; /* unused, must be 0 */
  mesg[26] = 0; /* unused, must be 0 */
  mesg[27] = 0; /* unused, must be 0 */
 
  /*
   * IP Header + first 64 bits of the datagram
   */

  memcpy((void *) &(mesg[28]),(void *) iph,iph->ihl * 4 + 8); 

  /*
   * Compute the ICMP header checksum
   */

  zkICMPHeaderChecksum((ZkICMPHeader *)&(mesg[20]),iph->ihl * 4 + 16);

  /*
   * Compute the IP header checksum 
   */

  zkIPHeaderChecksum(header);

  /*
   * All done.  the buffer can be dropped onto the network (with a raw socket)
   */

  return 1;
}

/**
 *  Create a source quench to stop the remote host from sending the 
 *  offending packet.
 *
 * @param iph The offending IP packet you wish to get fragmented
 * @param mesg Buffer to hold the message
 * @param size Holds the original size of buffer, returns bytes used
 * @author Philippe Boucher
 * @version 1.0
 * @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkICMPCreateSourceQuench (
  ZkIPHeader * iph ,
  unsigned char * mesg ,
  size_t * size )
{
  uint16_t totlen = 0;
  ZkIPHeader * header = NULL;

  header = (ZkIPHeader *) mesg;

  /*
   * ip header (20) + icmp header (8) + returned iph (ihl * 4) + 8 bytes
   */

  totlen = iph->ihl * 4 + 36;  
  if (*size < (size_t) totlen) {
    return 0;
  }
  *size = (size_t) totlen;

  /*
   * Make the return IP datagram
   */

  mesg[0] = 0x45; /* Version and header length */
  mesg[1] = 0; /* Type of service */
  mesg[2] = totlen / 0x100; /* First byte of total packet length */
  mesg[3] = totlen % 0x100; /* Second byte of total packet length */
  mesg[4] = 0xfe; /* First half of identification */
  mesg[5] = 0xed; /* Second half of identification */
  mesg[6] = 0; /* Fragmentation flags and first half of frag offset */
  mesg[7] = 0; /* Second half of frag offset */
  mesg[8] = 255; /* Time to live */
  mesg[9] = ZK_IPPROTO_ICMP; /* IP Protocol type */
  mesg[10] = 0; /*First half of checksum */
  mesg[11] = 0; /* Second half of checksum */
  memcpy((void *)&(mesg[12]),(void *)&(iph->DstIpAddress),4);
  memcpy((void *)&(mesg[16]),(void *)&(iph->SrcIpAddress),4);
  
  /*
   * ICMP Header  
   */

  mesg[20] = 4; /* Type = Source quench */
  mesg[21] = 0; /* Code  */
  mesg[22] = 0; /* First half of ICMP checksum */
  mesg[23] = 0; /* Second half of ICMP checksum */
  mesg[24] = 0; /* unused, must be 0 */
  mesg[25] = 0; /* unused, must be 0 */
  mesg[26] = 0; /* unused, must be 0 */
  mesg[27] = 0; /* unused, must be 0 */
 
  /*
   * IP Header + first 64 bits of the datagram
   */

  memcpy((void *) &(mesg[28]),(void *) iph,iph->ihl * 4 + 8); 

  /*
   * Compute the ICMP header checksum
   */

  zkICMPHeaderChecksum((ZkICMPHeader *)&(mesg[20]),iph->ihl * 4 + 16);

  /*
   * Compute the IP header checksum 
   */

  zkIPHeaderChecksum(header);

  /*
   * All done.  the buffer can be dropped onto the network (with a raw socket)
   */

  return 1;
}

