/*
** **************************************************************************
** $Id: zkPlugin.c,v 1.4.12.3 2000/12/14 17:54:34 jean Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom(R) code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge(R) Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

**
** Copyright (C) 1999 Zero-Knowledge Systems Inc.  All Rights Reserved
** **************************************************************************
*/

# include <zks.h>
# include <zkAlloc.h>
# include <zkSharedObject.h>
# include "zkPlugin.h"
# include "zkCStr.h"

# define ZKPLUGIN_GETINFOCALLBACK_SYMBOL "zkPluginGetInfo"

typedef int (*zkPluginGetInfoCallback)(
    ZkError                    err,
    uint32_t                   desired_version,
    struct t_ZkPluginInfo *    plugin_info_ptr );



/* --------------------------------------------------------------------------
 * zkPluginLoad
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkPluginLoad(
    ZkError                    err,
    const char *               plugin_file,
    ZkPluginInfo *             plugin_info_ptr,
    const char *               plugin_conf)
{
    static const char * func = "zkPluginLoad";
    ZkSharedObject object  = NULL;
    ZkPluginInfo   plugin  = NULL;
    void *         address = NULL;
    char *         dupconf = NULL;

    if ( !(plugin_file && plugin_info_ptr) )
    {
        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKENULLPTR,
                   "input params are null");        
    }
    else if ( zkSharedObjectLoad(err, plugin_file, &object)
              && zkSharedObjectFindSymbol(err, object,
                                          ZKPLUGIN_GETINFOCALLBACK_SYMBOL,
                                          &address)
              && zkMemAlloc(err, (void**)&plugin, sizeof(*plugin))

              /* Allocate space for a context string if we were given one. */

              && (   (plugin_conf == NULL)
                  || (zkMemAlloc(err,
                                 (void**)&(dupconf),
                                 strlen(plugin_conf) + 1))))
    {
        memset(plugin, 0, sizeof(*plugin));

        if (plugin_conf != NULL) 
        {
          zkCStrlcpy(dupconf, plugin_conf, strlen(plugin_conf) + 1);
        }
        plugin->conf_file_name = dupconf;
        plugin->object_handle = object;
        object = NULL;
        
        if ((*((zkPluginGetInfoCallback)address)) ( err,
                                                    ZKPLUGIN_APIVERSION,
                                                    plugin) )
        {
            *plugin_info_ptr = plugin;
            plugin = NULL;
            return 1;
        }
    }

    /*
     * Reaching here means that an error has occurred.
     */

    if (object)
    {
        zkSharedObjectUnload(NULL, &object);
    }

    if (plugin)
    {
        zkMemFree(NULL, (void**)(&plugin));
    }

    if (dupconf)
    {
        zkMemFree(NULL, (void**)(&dupconf));
    }

    return 0;
}



/* --------------------------------------------------------------------------
 * zkPluginUnload
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkPluginUnload(
    ZkError                    err,
    ZkPluginInfo *             plugin_info_ptr )
{
    static const char * func = "zkPluginUnload";
    int result = 0;

    if ( (plugin_info_ptr == NULL) || (*plugin_info_ptr == NULL) )
    {
        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKENULLPTR,
                   "input params are null");        
    }
    else
    {
        void * address = NULL;

        switch ((*plugin_info_ptr)->api_version)
        {
        case ZKPLUGIN_APIVERSION_1:
            address = ((ZkPlugin_1)((*plugin_info_ptr)->vtable))->destructor;

            if (address == NULL)
            {
                zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKENULLPTR,
                           "destructor callback is NULL");        
            }
            else
            {
                result = (*((ZkPluginDestructor_1)address))(
                    err,*plugin_info_ptr);

                if ((*plugin_info_ptr)->conf_file_name != NULL) {
                  zkMemFree(err, (void**)&((*plugin_info_ptr)->conf_file_name));
                }
            }
            break;
            
        default:
            zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKENULLPTR,
                       "unrecognized api version");        
        }
    }
        
    if ( result
         && zkSharedObjectUnload(err,(ZkSharedObject*)&((*plugin_info_ptr)->object_handle))
         && zkMemFree(err, (void**)plugin_info_ptr))
    {
        result = 1;
    }

    return (result);
}



/* --------------------------------------------------------------------------
 * zkPluginGetName
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkPluginGetName(
    ZkError                    err,
    ZkPluginInfo               plugin_info,
    const char **              name )
{
    if ( (plugin_info == NULL) || (name == NULL) )
    {
        zkErrorSet(err,"zkSharedObjectGetNamee",ZKERR,ZKMPLUGIN,ZKENULLPTR,
                   "input params are null");        
        return 0;
    }    

    (*name) = plugin_info->name;
    return 1;
}



/* --------------------------------------------------------------------------
 * zkPluginGetVersion
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkPluginGetVersion(
    ZkError                    err,
    ZkPluginInfo               plugin_info,
    uint32_t *                 version )
{
    if ( (plugin_info == NULL) || (version == NULL) )
    {
        zkErrorSet(err,"zkSharedObjectGetVersion",ZKERR,ZKMPLUGIN,ZKENULLPTR,
                   "input params are null");        
        return 0;
    }    

    (*version) = plugin_info->version;

    return 1;
}



/* --------------------------------------------------------------------------
 * zkPluginGetApiVersion
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkPluginGetApiVersion(
    ZkError                    err,
    ZkPluginInfo               plugin_info,
    uint32_t *                 api_version )
{
    static const char * func = "zkPluginGetApiVersion";

    if ( (plugin_info == NULL) || (api_version == NULL) )
    {
        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKENULLPTR,
                   "input params are null");        
        return 0;
    }    

    (*api_version) = plugin_info->api_version;

    return 1;
}



/* --------------------------------------------------------------------------
 * zkPluginMemFree
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkPluginMemFree(
    ZkError                    err,
    ZkPluginInfo               plugin_info,
    void **                    mem_to_free )
{
    static const char * func = "zkPluginMemFree";
    int result = 0;

    if ( (plugin_info == NULL) || (mem_to_free == NULL) )
    {
        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKENULLPTR,
                   "input params are null");        
    }
    else
    {
        void * address = NULL;

        switch (plugin_info->api_version)
        {
        case ZKPLUGIN_APIVERSION_1:
            address = ((ZkPlugin_1)(plugin_info->vtable))->free_func;

            if (address == NULL)
            {
                zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKENULLPTR,
                           "destructor callback is NULL");        
            }
            else
            {
                result = (*((ZkPluginDataFreeFunc_1)address))(
                    err,plugin_info->private_data,mem_to_free);
            }
            break;
            
        default:
            zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKENULLPTR,
                       "unrecognized api version");        
        }
    }

    return (result);
}



/* --------------------------------------------------------------------------
 * zkPluginCommand_1
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkPluginCommand_1(
    ZkError                    err,
    ZkPluginInfo               plugin_info,
    const char *               argv[],
    size_t                     argc,
    const void *               in_data,
    size_t                     in_data_size,
    void **                    out_data,
    size_t *                   out_data_size )
{
    static const char * func = "zkPluginCommand_1";
    int result = 0;

    if ( !(plugin_info && argv && out_data && out_data_size) )
    {
        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKENULLPTR,
                   "input params are null");        
    }
    else if (in_data_size && !in_data)
    {
        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKERROR,
                   "invalid input data");        
    }
    else if (strcasecmp(argv[0],plugin_info->name) != 0)
    {
        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKERROR,
                   "plugin mismatch");        
    }
    else if (argc < 2)
    {
        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKERROR,
                   "too few parameters");        
    }
    else
    {
        void * address = NULL;

        switch (plugin_info->api_version)
        {
        case ZKPLUGIN_APIVERSION_1:
            address = ((ZkPlugin_1)(plugin_info->vtable))->handler_func;

            if (address == NULL)
            {
                zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKENULLPTR,
                           "destructor callback is NULL");        
            }
            else
            {
                result = (*((ZkPluginCommandHandler_1)address))(err,
                                                                plugin_info->private_data,
                                                                argv,
                                                                argc,
                                                                in_data,
                                                                in_data_size,
                                                                out_data,
                                                                out_data_size);
            }
            break;
            
        default:
            zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKENULLPTR,
                       "unrecognized api version");        
        }
    }

    return (result);
}    
