/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * -- Oliver Lavery <oliver@zks.net>

 */

/*
 * $id$ 
 * Linux implementation of (semi) portable interval timer API
 */
#include <zkThread.h>
#include <zkAlloc.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>

/* The client assert macros are in a client only header. 
 * I want an assert in here to catch nasty errors in debug
 * builds. --ol 
 */
#if defined(DEBUG) || defined (_DEBUG)
#define ASSERT(expr) \
	do { if (!(expr)) { \
    char foo; \
	fprintf(stderr, "ASSERTION FAILED at %s:%d (%s)\n", __FILE__, __LINE__, #expr);\
	printf("a) to abort and dump core\nq) to exit\nAny other key to continue.\n");\
	switch(getchar()) {\
		case 'a':\
		case 'A':\
			foo = *(char*)NULL;\
			break;\
		case 'q':\
		case 'Q':\
			exit(-1);\
	 }}\
	 } while (0)
#endif


/* NOTE: I have no idea how well this will work.
 * it's largely an experiment, since we don't have enough
 * itimers on linux to suit the client's needs. 
 * Also I wanted to avoid using signal based timers
 * since they have a nasty habit of interrupting system calls
 * and windows code tends not to look for these cases
 * (yes SA_RESTART can sort of fix this, but not reliably
 *  and timers firing causing system calls to bail can be the
 *  source of incredably hard to find bugs)
 * Crudely measuring the accuracy of this implementation, it
 * appears to be at least as accurate as the win32 SetTimer
 * API, and the client doesn't need incredably high accuracy.
 */

 /* Quick and dirty doubly linked list manipulation stuff based on linux/list.h
  */

#define INIT_LIST(ptr) do {(ptr)->next = (ptr); (ptr)->prev = (ptr); } while (0)

#define RUNNING_STATE 0
#define SHUTDOWN_STATE 1

/*
 * delays shorter than this amount (in millis) will be executed as
 * busy waits. BEWARE: this devoures CPU.
 */
#define TIMER_FUDGE 10

typedef struct __tag_ZkITimerQueue
{
    ZkMutex queueMutex;
    ZkITimer timerQueue;
    ZkEvent notifyEvent;
    unsigned short int state;
    ZkITimer current;
} ZkITimerQueue;

struct __tag_ZkITimer
{
  struct __tag_ZkITimer *next, *prev;
  struct timeval fire;
  struct timeval interval;
  void (ZK_ITIMER_CALLBACK *callback) (void *, unsigned int, unsigned int, unsigned long);
};

static unsigned int timerInitialised = ZK_FALSE;
static ZkThread timerThread = NULL;
static ZkITimerQueue g_Queue;

/* sort order insertion */
static inline void
_timer_add (ZkITimer head, ZkITimer node)
{
    ZkITimer iter;
    iter = head->next;
    /* This just traverses the list until it finds the first node who's firing time is later than ours */
    /* and then puts us before that node */
    while (iter != head && (node->fire.tv_sec > iter->fire.tv_sec ||
                            (node->fire.tv_sec == iter->fire.tv_sec
                             && node->fire.tv_usec >= iter->fire.tv_usec)))
    {
        iter = iter->next;
    }
    node->next = iter;
    node->prev = iter->prev;
    iter->prev->next = node;
    iter->prev = node;
    
#if defined(DEBUG) || defined (_DEBUG)
    iter = head->next;
    while (iter != head) 
    {
        ASSERT( "Queue is in order" && (
            iter->prev->fire.tv_sec < iter->fire.tv_sec 
            || ( iter->prev->fire.tv_sec == iter->fire.tv_sec 
                 && iter->prev->fire.tv_sec <= iter->fire.tv_sec) )
        );
        iter = iter->next;
    }
#endif   
}

/* Pull node out of the doubly linked list */
static inline void
_timer_del (ZkITimer node)
{
  /* Must do NULL checks here. look at comments in zkITimerShutdown */
  if (node->prev)
    node->prev->next = node->next;
  if (node->next)
    node->next->prev = node->prev;
  node->prev = NULL;
  node->next = NULL;
}

/* Recalculate firing time from the current time and an interval timer's interval time.
 */
static inline void
_timer_reset (ZkITimer node, struct timeval *now)
{
  node->fire.tv_sec = now->tv_sec + node->interval.tv_sec;
  node->fire.tv_usec = now->tv_usec + node->interval.tv_usec;
}

static void *
_timer_thread (void *context)
{
  ZkITimerQueue *queue = (ZkITimerQueue *) context;
  ZkITimer queueHead = queue->timerQueue;
  ZkITimer node;
  ZkError err;
  struct timeval now;
  long idleTime = 0;
  zkErrorNew (&err);
  while (1)
    {
      gettimeofday (&now, NULL);
      if (!zkMutexLock (err, queue->queueMutex))
	goto die;
      for (node = queueHead->next;
	   node != queueHead
	   && (node->fire.tv_sec < now.tv_sec
	       || (node->fire.tv_sec == now.tv_sec
		   && node->fire.tv_usec <= now.tv_usec)); node = node->next)
	{
	  unsigned long fireTime = now.tv_sec * 1000L + now.tv_usec / 1000L;

	  /* Unlock queue around callbacks. In the case that the callback takes a while,
	   * we don't want to keep other threads from adding / removing timers
	   * If a timer is added which is supposed to fire prior to now, this could cause 
	   * problems (which shouldn't happen, obviously)
	   */
      queue->current = node;
	  if (!zkMutexUnlock (err, queue->queueMutex))
	    goto die;
	  if (node->callback)
	    node->callback (NULL, 0, 0, fireTime);
	  if (!zkMutexLock (err, queue->queueMutex))
	    goto die;
      
      if (queue->current) {
          node = queue->current;
          _timer_del (node);
          _timer_reset (node, &now);
          _timer_add (queueHead, node);
      }

      queue->current = NULL;

	  /* reset the for loop and refetch the current time since the callback may 
	   * have consumed a measurable amount. This is ok since eventually we
	   * will bail out of the for loop as expired timers are being pushed to the back of
	   * the queue. In the case where we have a lot of short interval timers, we may
	   * get stuck in this loop for a long time. Timers _SHOULD_ be removed prior to 
	   * shutdown of this thread to prevent that situation (which could end up in deadlock).
	   */
	  node = queueHead;
	  gettimeofday (&now, NULL);
	}
      if (!zkMutexUnlock (err, queue->queueMutex))
	goto die;
      /* Ok, were idle. Now we need to sit around and do nothing for a while 
       * based on our ordered insertion of nodes to the list, the next timer to fire
       * should be queueHead->next. 
       */
      gettimeofday (&now, NULL);
      if (queueHead->next->fire.tv_sec)
	{
	  idleTime = ((queueHead->next->fire.tv_sec - now.tv_sec) * 1000L
		      + (queueHead->next->fire.tv_usec -
			 now.tv_usec) / 1000L) - TIMER_FUDGE;
	  if (idleTime < 0)
	    idleTime = 0;
	}
      else
	{
	  idleTime = ZK_INFINITE;
	}
      zkErrorClear (err);

      if (!zkEventWait (err, queue->notifyEvent, idleTime)
	  && zkErrorCode (err) != ETIMEDOUT)
	{
	  /* A miscellaneous error occured. Not a good thing. */
	  goto die;
	}
      else if (queue->state == SHUTDOWN_STATE)
	goto die;
      /* we either timed out and need to process timers in the queue, 
       * or we were notified of a change to the queue and need to reschedule ...
       * so we loop.
       */
    }
die:
  return (void *) err;
}

/*********************************
 * Initialize Interval Timer thread
 */
DLLExport int
zkITimerInit (ZkError err	/* [in/out]   ZkError */
  )
{
  int ret = ZK_FALSE;

  zkErrorClear (err);
  if (timerInitialised)
    goto die;

  else if (!zkMutexNew (err, &g_Queue.queueMutex))
    goto die;
  else if (!zkEventNew (err, &g_Queue.notifyEvent, ZK_FALSE))
    goto die;
  else
    if (!zkMemAlloc
	(err, (void **) &g_Queue.timerQueue,
	 (size_t) sizeof (struct __tag_ZkITimer))) goto die;
  memset (g_Queue.timerQueue, 0, sizeof (struct __tag_ZkITimer));
  INIT_LIST (g_Queue.timerQueue);
  g_Queue.state = RUNNING_STATE;
  g_Queue.current = NULL;
  if (!zkThreadCreate (err, &timerThread, &_timer_thread, (void *) &g_Queue))
    goto die;
  ret = timerInitialised = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Kill interval timer thread
 */
DLLExport int
zkITimerShutdown (ZkError err	/* [in/out]   ZkError */
  )
{
  int ret = ZK_FALSE;
  void *threadReturn;
  ZkITimer iter;

  zkErrorClear (err);
  if (!timerInitialised)
    goto die;			/* XXX should we return false in this case? */
  g_Queue.state = SHUTDOWN_STATE;
  if (!zkEventSet (err, g_Queue.notifyEvent))
    goto die;			/* signal shutdown */
  else if (!zkThreadJoin (err, &timerThread, &threadReturn))
    goto die;			/* stop itimer thread */
  else if (!zkMutexLock (err, g_Queue.queueMutex))
    goto die;
  timerInitialised = ZK_FALSE;

  /* Now the timer thread is stopped, but we don't know if there are any timers which remain in the queue
   * For now we blow away the queue anyway and leak any memory. In other words, empty the queue BEFORE calling
   * shutdown  (although we're probably exiting anyway so the leak won't matter)
   * (n.b. we could just free all the ZkITimers in the queue, but this would invalidate a whole bunch of pointers
   * dangling in code that uses this library)
   */

  /* In case clients try to free nodes after calling shutdown (i.e. _way_ strange code) we NULL all the next/prevs
   * so we can detect that the timers are invalid.
   */
  for (iter = g_Queue.timerQueue->next; iter != g_Queue.timerQueue;)
    {
      iter = iter->next;
      iter->prev->next = NULL;
      iter->prev->prev = NULL;
    }
  if (!zkEventFree (err, &g_Queue.notifyEvent))
    goto die;
  else if (!zkMutexUnlock (err, g_Queue.queueMutex))
    goto die;
  else if (!zkMutexFree (err, &g_Queue.queueMutex))
    goto die;
  else if (!zkMemFree (err, (void **) &g_Queue.timerQueue))
    goto die;
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Create a new interval timer
 */
DLLExport int
zkITimerNew (ZkError err,	/* [in/out]   ZkError */
	     ZkITimer * pTimer,	/* [in/out] indirect pointer to the timer */
	     void (ZK_ITIMER_CALLBACK *callback) (void *, unsigned int, unsigned int, unsigned long),	/*[in] timer callback fn */
	     unsigned long millis	/* [in] interval for the new timer */
  )
{
  const char *fn = "zkITimerNew";
  struct __tag_ZkITimer *l_pTimer = NULL;
  int ret = ZK_FALSE;
  struct timeval now;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pTimer)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pTimer is null");	/* XXX not MSEM */
      goto die;
    }
  else if (millis < TIMER_FUDGE << 1) 
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Cannot set an itimer with an interval less than 20ms");     /* XXX not MSEM */
      goto die;
    }
  else if (!timerInitialised)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL,
		     "Interval timer library not initialized.");	/* XXX not MSEM */
      goto die;
    }
  /*
   * Lock queue mutex
   */
  else if (!zkMutexLock (err, g_Queue.queueMutex))
    {
      goto die;
    }
  /* 
   * Allocate system resources 
   */
  else if (!zkMemAlloc
	   (err, (void **) &l_pTimer,
	    (size_t) sizeof (struct __tag_ZkITimer)))
    {
      zkMutexUnlock (err, g_Queue.queueMutex);
      goto die;
    }
  /* 
   * OK. Do insertion into queue
   */
  l_pTimer->callback = callback;
  l_pTimer->interval.tv_sec = millis / 1000L;
  l_pTimer->interval.tv_usec = (millis % 1000L) * 1000L;
  gettimeofday (&now, NULL);
  _timer_reset (l_pTimer, &now);
  _timer_add (g_Queue.timerQueue, l_pTimer);

  /* If either of these fail than something BAD will happen */
  if (!zkMutexUnlock (err, g_Queue.queueMutex))
    goto die;
  else if (!zkEventSet (err, g_Queue.notifyEvent))
    goto die;

  /* Good to go. Return the timer to the client and keep truckin' */
  *pTimer = l_pTimer;
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Release an interval timer
 */
DLLExport int
zkITimerFree (ZkError err,	/* [in/out]   ZkError */
	      ZkITimer * pTimer	/* [in/out] indirect pointer to the timer */
  )
{
  const char *fn = "zkITimerFree";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pTimer || !*pTimer)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pTimer is null");	/* XXX not MSEM */
      goto die;
    }
  else if (!timerInitialised)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL,
		     "Interval timer library not initialized.");	/* XXX not MSEM */
      goto die;
    }
  /*
   * Lock queue mutex
   */
  else if (!zkMutexLock (err, g_Queue.queueMutex))
    {
      goto die;
    }

  /* 
   * Slice 'em out of the queue 
   */
  _timer_del (*pTimer);
  if (g_Queue.current == *pTimer) g_Queue.current = NULL;

  /* 
   * Free system resources 
   */
  if (!zkMemFree (err, (void **) pTimer))
    {
      goto die;
    }

  /* If either of these fail than something BAD will happen */
  else if (!zkMutexUnlock (err, g_Queue.queueMutex))
    goto die;
  else if (!zkEventSet (err, g_Queue.notifyEvent))
    goto die;

  /* Good to go. Null out the client's ptr */
  *pTimer = NULL;
  ret = ZK_TRUE;
die:
  return ret;
}
