#!/usr/local/ActivePerl-5.6/bin/perl -w
# 
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
# 
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
# 
# The Original Code is Freedom code.
# 
# The Initial Developer of the Original Code is Zero-Knowledge Systems,
# Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
# Zero-Knowledge Systems, Inc.  All Rights Reserved.
# 
# Contributor(s):
#

BEGIN {
    $GscriptDir = $0;
    $GscriptDir =~ s,\\,/,g;
    $GscriptDir =~ s,/?[^/]+$,,;
    $GscriptDir = "." if ($GscriptDir eq "");
    unshift(@INC, $GscriptDir);
}

use Config;
use ZKFTP;
use File::Path;
use ZKUtils;
use POSIX 'strftime';
use ZKLog;
use ZKBoard;
use Sys::Hostname;


$Gbn = {host => exists($ENV{BNSERVER}) ? "$ENV{BNSERVER}" : "marcp.cia.zks.net", 
	port => exists($ENV{BNPORT}) ? "$ENV{BNPORT}" : "8666"};

$Gconfig = {boardServer => 'zdev.dev.zks.net',
	    boardPort => 8080,
	    branch => undef, 
	    build => "true",
	    buildnum => 0,
	    cfgs => => ["Release"],
	    clean => undef,
	    clientPrefix => undef,
	    date => undef,
	    emailRecipients => => ["build-team"],
	    its4 => undef,
	    label => undef, 
	    localdir => undef, 
	    log => new ZKLog(),
	    product => undef,
	    postdir => undef,
	    rtc => undef,
	    starttime => time,
	    sync => undef,
	    warnCount => "true",
	    zkboard => undef,
	   };

$Gstatus = 0;

# Setup some environment variable that affect behavior of Makefiles
# but not the result
$ENV{ZK_NODEPLIBS} = "true"; # Do not perform libs dependencies

#
# Display Usage of this script
#
sub Usage {
    print("Usage: $0 [-h]\n");
    print("   [-h]: print this message\n");
    print("   [-branch <branch>]: Branch to sync with\n");
    print("   [-cfg <cfg,...>]: Configs to build\n");
    print("   [-clean]: To clean all trees before compiling\n");
    print("   [-clientprefix <prefix>]: Prefix for client setup (must be defined is using -postdir on win32)\n");
    print("   <-project <client,server>>: In order to build the client, the server or both\n");
    print("   [-date <date>]: Date to sync with. Useless with -label\n");
    print("   [-email <recipient,...>]: Distribution list of people to be notified of build failures\n");
    print("   [-its4]: Will run the ITS4 reporting process on the source files\n");
    print("   [-label <label>]: Label to sync with\n");
    print("   [-logonly]: Will only output into logfile\n");
    print("   [-nobuild]: Will skip the build process\n");
    print("   [-rtc]: Indicates this is a RTC build\n");
    print("   [-sync <module:tag,...>]: override default sync module:tags\n");
    print("   <-localdir>: Path where to sync and build\n");
    print("   [-postdir <directory>]: Directory where to post results\n");
    print("   [-nowarn]: Will skip the warning counting process\n");
    print("   [-zkboard <version>]: Will post the results to the zkboard builds page\n");
    exit(1);
}

#
# Parse the command line option
#
sub Getopts {
    my ($config, @args) = @_;
    
    while(@args) {
	$_ = shift(@args);
	
	# Print Usage
	if (m,^-h(elp)?$,) {
	    &Usage;
	    exit(1);
	}
	
	# Project to build
	if (m,^-project$,) {
	    if (@args) {
		$config->{projects} = [split(m/,/, lc(shift(@args)))];
	    } else {
		print("Error: missing arg to -project\n");
		&Usage;
	    }
	    next;
	}
	
	# Module:Tags
	if (m,^-sync$,) {
	    if (@args) {
		$config->{sync} = shift(@args);
	    } else {
		print("Error: missing arg to -sync\n");
		&Usage;
	    }
	    next;
	}
	
	# Client prefix
	if (m,^-c(lient)?p(refix)?$,) {
	    if (@args) {
		$config->{clientPrefix} = &strftime(shift(@args), localtime(time));
	    } else {
		print("Error: missing arg to -clientprefix\n");
		&Usage;
	    }
	    next;
	}
	
	# Path where to sync and build
	if (m,^-localdir$,) {
	    if (@args) {
		$config->{localdir} = &ZKUtils::FixOsPath(shift(@args));
		$ENV{ZK_ROOT} = $config->{localdir};
				# Make sur to remove trailing /. for ZK_ROOT
		$ENV{ZK_ROOT} =~ s,(/\.)*$,,;
	    } else {
		print("Error: missing arg to -localdir\n");
		&Usage;
	    }
	    next;
	}
	
	# Path where to post results
	if (m,^-postdir$,) {
	    if (@args) {
		$config->{postdir} = &strftime(shift(@args), localtime(time));
	    } else {
		print("Error: missing arg to -postdir\n");
		&Usage;
	    }
	    next;
	}
	
	# Specify the branch
	if (m,^-b(ranch)?$,) {
	    if (@args) {
		$config->{branch} = shift(@args);
		next;
	    } else {
		print("Error: missing arg to option -branch <branch>\n");
		&Usage;
	    }
	}
	
	# Specify the label
	if (m,^-l(abel)?$,) {
	    if (@args) {
		$config->{label} = shift(@args);
		next;
	    } else {
		print("Error: missing arg to option -label <label>\n");
		&Usage;
	    }
	}
	
	# Specify the date
	if (m,^-d(ate)?$,) {
	    if (@args) {
		$config->{date} = shift(@args);
		next;
	    } else {
		print("Error: missing arg to option -date <cvs date | now>\n");
		&Usage;
	    }
	}
	
	# Specify the configs
	if (m,^-cfg?$,) {
	    if (@args) {
		$config->{cfgs} = [split(m/,/, shift(@args))];
		next;
	    } else {
		print("Error: missing arg to option -cfg <cfgs,...>\n");
		&Usage;
	    }
	}
	
	# If we need to clean
	if (m,^-clean$,) {
	    $config->{clean} = "true";
	    next;
	}
	
	# Will only output in logfile
	if (m,^-logonly$,) {
	    $config->{log}->logonly(1);
	    next;
	}
	
	# If we need to sync
	if (m,^-nosync$,) {
	    $config->{sync} = undef;
	    next;
	}
	
	# If we want to skip the build section
	if (m,^-nobuild$,) {
	    $config->{build} = undef;
	    next;
	}
	
	# Test for -rtc
	if ( m,^-rtc$,i) {
	    $config->{rtc} = "RTC";
	    next;
	}
	
	# If we want to skip the warning counting section
	if (m,^-nowarn$,) {
	    $config->{warnCount} = undef;
	    next;
	}
	
	# If we want the results to be posted to the zkboard
	if (m,^-zkboard$,) {

	    if (@args) {
		#Keep the version for now
		$config->{zkboard} = shift(@args);	    
	    } else {
		print("Error: missing arg to -zkboard\n");
		&Usage;
	    }
	    
	    next;
	}
	
	# If we want to run the ITS4 reporting section
	if (m,^-its4$,) {
	    $config->{its4} = "true";
	    next;
	}
	
	# Specify the distribution list for build failures
	if (m,^-e(mail)?$,) {
	    if (@args) {
		$config->{emailRecipients} = [split(m/,/, shift(@args))];
		next;
	    } else {
		print("Error: missing arg to option -email <recipient,...>\n");
		&Usage;
	    }
	}

	print("Ignoring option '$_'\n");
    }
}

sub SyncSources {
    my ($localdir, $branch, $label, $date, $moduleTag, $log) = @_;
    my $status = 0;
    my $dateopt = "";
    my $tagopt = "";
    my $cwd = &POSIX::getcwd();

    # Check for label and if not defined then check for branch
    if (defined($label)) {
	$tagopt = "-r $label";
    } elsif (defined($branch)) {
	$tagopt = "-r $branch";
    }
    
    # Check for date specification only if these is no label
    if (!defined($label)) {
	# Special case for date
	if (!defined($date) || (defined($date) && (lc($date) eq "now"))) {
	    $dateopt = &strftime("-D \"%Y-%m-%d %H:%M\"", localtime(time));
	} else {
	    $dateopt = "-D \"$date\"";
	}
    }
    
    # Chdir to the source directory
    if (chdir($localdir)) {
	# For all moduleTag pair, perform a checkout. If tag is empty
	# and no label or branch specified use -A (reset sticky tag) 
	# Format: module:tag,...,module:tag
	foreach (split(m/,/, $moduleTag)) {
	    if (m,^([^:]+)(:([\w-]+))?,) {
		my $module = $1;
		my $cmd = undef;
		my $tagopt2 = undef;
		
		# If a tag was specified
		if (defined($3)) {
		    $tagopt2 = "-r $3";
		} elsif ( $tagopt eq "") {
		    $tagopt2 = "-A"
		} else {
		    if (!defined($label) && defined($branch)) {
			# On a branch use branch tag append date
			$tagopt2 = "$tagopt $dateopt";
		    } else {
			# Don't know if tag is a branch or
			# label. Ignore date
			$tagopt2 = $tagopt;
		    }
		}
		
		# Build the sync command
		$cmd = "cvs -d :pserver:build\@cvs.zks.net:/export/cvs -r co $tagopt2 $module";

		# Add a trace of the sync command
		$log->print("$cmd\n");

		# Execute it
		if ($log->system($cmd)) {
		    $log->print("Error: sync '$tagopt2' '$module' exit with error code\n");
		    $status = 1;
		    last;
		}
	    }
	}
	
	# Go back to the previous directory
	if (!chdir($cwd)) {
	    $log->print("Error: can't chdir back to '$cwd'\n");
	    $status = 1;
	}
    } else {
	$log->print("Error: can't chdir to '$localdir'\n");
	$status = 1;
    }
    
    # Log sync status
    $log->print("Sync status=$status\n");
    return($status);
}

sub BuildUnix {
    my ($localdir, $cfg, $log) = @_;
    my $cwd = &POSIX::getcwd();
    my $status1 = 0;
    
    # Build the tree
    if (chdir($localdir)) {
	my $targets = $Config::Config{osname} eq "solaris" ? "libs solpkg" : "libs package";
	
	# Build it
	if ($log->system("gmake -f makefiles/Makefile.top promote $targets ")) {
	    $log->print("Error: build exited with an error code\n");
	    $status1 = 1;
	}
	
	if (!chdir($cwd)) {
	    $log->print("Error: can't chdir back to '$cwd'\n");
	    $status1 = 1;
	}
    } else {
	$log->print("Error: can't chdir to '$localdir'\n");
	$status1 = 1;
    }
    
    return($status1 ? 1 : 0);
}

sub BuildWin32 {
    my ($srcdir, $cfg, $log) = @_;
    my $clientdir = "${srcdir}freedom/client";
    my $vxddir = "$clientdir/zkshim/win32/zksHook";
    my $ntdrvdir = "$clientdir/zkshim/win32/NT2000";
    my $cwd = &POSIX::getcwd();
    my $status1 = 0;
    my $status2 = 0;
    my $status3 = 0;
    
    # Build the dll and client executable
    if (chdir($clientdir)) {
	if ($log->system("msdev WindowClient.dsw /USEENV /MAKE \"FreedomClient - Win32 $cfg\"")) {
	    $log->print("Error: build exited with an error code\n");
	    $status1 = 1;
	}
	
	if (!chdir($cwd)) {
	    $log->print("Error: can't chdir back to '$cwd'\n");
	    $status1 = 1;
	}
    } else {
	$log->print("Error: can't chdir to '$clientdir'\n");
	$status1 = 1;
    }
    
    # Build VXD
    if (chdir($vxddir)) {
	if ($log->system("msdev zksHook_9x.dsp /USEENV /MAKE \"zksHook_9x - Win32 $cfg\"")) {
	    $log->print("Error: build exited with an error code\n");
	    $status2 = 1;
	}
	
	if (!chdir($cwd)) {
	    $log->print("Error: can't chdir back to '$cwd'\n");
	    $status2 = 1;
	}
    } else {
	$log->print("Error: can't chdir to '$vxddir'\n");
	$status2 = 1;
    }
    
    # NT2000 driver
    if (chdir($ntdrvdir)) {
	if ($log->system(($cfg eq "Debug") ? "DebugBuild.bat" : "RetailBuild.bat")) {
	    $log->print("Error: building nt2000 driver\n");
	    $status3 = 1;
	}
	
	if (!chdir($cwd)) {
	    $log->print("Error: can't chdir back to '$cwd'\n");
	    $status3 = 1;
	}
    } else {
	$log->print("Error: can't chdir to '$ntdrvdir'\n");
	$status3 = 1;
    }

    # Make sure the NT2000 driver was generated properly
    my $objDir = "";
    if ($cfg eq "Debug") {
      $objDir = "objchk/i386";
    } else {
      $objDir = "objfre/i386";
    }
    
    my $drvFile = "$ntdrvdir/Driver/$objDir/freedom.sys";
    my $notifObjFile = "$ntdrvdir/Install/Install_2000/NotifyObject/$objDir/freebind.dll";
    my $instFile = "$ntdrvdir/Install/Install_2000/zkInstallDriver/$objDir/zkInstallDriver.exe";

    if ((! -e $drvFile) || (! -e $notifObjFile) || (! -e $instFile)) {
      $status3 = 1;
    }

    $log->print("Status of WindowClient=$status1\n");
    $log->print("Status of Vxd=$status2\n");
    $log->print("Status of NT2000 Driver=$status3\n");
    return($status1 | $status2 | $status3 ? 1 : 0);
}

sub PostRcp {
    my ($log, $host, $postdir, @instfiles) = @_;
    my $status = 0;
    
    
    return(1) if ($log->system("rsh -n $host mkdir -p $postdir"));
    return(1) if ($log->system("rcp " . join(" ", @instfiles) . " $host:$postdir\n"));
    return(0)
}

sub PostFtp {
    my ($host, $postdir, $subdirs, @instfiles) = @_;
    my $ftp = ZKFTP->new($host, "build", "CiaBld");
    my $dir = $postdir;
    my $status = 0;
    
    # Create the postdir directory
    return(1) if ($ftp->mkdir($postdir));
    
    # Create all subdirs under the postdir
    foreach (@{$subdirs}) {
	$dir .= "/$_";
	$status |= $ftp->mkdir($dir);
    }
    
    # Post the files
    if (!$status) {
	$ftp->remotedir($dir);
	$status |= $ftp->mput(\@instfiles);
    }
    
    return($status);
}

sub ZkboardUpdate {
    my ($status,$refproject,$refconfig,$platform) = @_;

    # Set the zkboard fields to update
    if (defined ($Gconfig->{zkboard})) {
	my @projects = undef;
	my @configs = undef;
	my $project = undef;
	my $config = undef;

	# Fields hash
	my %fields =();

	# Check if we have a list of project and/or config
	# change the ref to array
	if (ref($refproject) eq 'ARRAY') {
	    @projects = @$refproject;
	}elsif (ref($refproject) eq 'SCALAR') {
	    $projects[0] = $$refproject;
	}
	if (ref($refconfig) eq 'ARRAY') {
	    @configs = @$refconfig
	    }elsif (ref($refconfig) eq 'SCALAR') {
		$configs[0] = $$refconfig;
	    }

	if (defined ($Gconfig->{rtc})) {
	    #Set the update keys which allow us to get a unique board
	    $Gconfig->{zkboard}->setUpdKeys ('PLATFORM', $platform);
	    $Gconfig->{zkboard}->setUpdKeys ('BUILDNO', $Gconfig->{buildnum});

	    #Set the common fields
	    $fields {'BUILDNO'} = $Gconfig->{buildnum};
	    $fields {'PLATFORM'} = $platform;
	    $fields {'START_TIME'} = &strftime("%Hh%M.%S", localtime($Gconfig->{starttime}));
	    $fields {'STATUS'} = $status;

	    #Build the fields array
	    foreach $project (@projects) {
		foreach $config (@configs) {
		    $fields {'PROJECT'} = $project;
		    $fields {'CONFIG'} = $config;

		    $Gconfig->{zkboard}->setUpdKeys ('PROJECT', $project);
		    $Gconfig->{zkboard}->setUpdKeys ('CONFIG', $config);

		    #Perform the update for this combination

		    my $zkstatus = $Gconfig->{zkboard}->set(\%fields);

		    if ($zkstatus eq '0'){
			$Gconfig->{log}->print("[zkboard] RTC Update successful for status <$status>\n");
		    }	    
		    else{
			$Gconfig->{log}->print("[zkboard] Error: board update fail <$status> \n $zkstatus\n");
		    }
		    
		}
	    }
	}
	else {
	    my $zkstatus = "";

	    # Check that we have the right platform
	    $platform = "WIN" if ($platform =~ /win32/i);

	    # Get the projects and configs
	    foreach $project (@projects) {
		foreach $config (@configs) {
		    $fields{'STATUS_'.uc($platform).'_'.uc($config).'_'.uc($project)} = $status;
		}
	    }

	    #Perform only one update for all fields since they are all in the same page
	    $zkstatus = $Gconfig->{zkboard}->set(\%fields);
	    if ($zkstatus eq '0') {
		$Gconfig->{log}->print("[zkboard] Update successful for status <$status>\n");
	    }
	    else{
		$Gconfig->{log}->print("[zkboard] Unable to perform update for status <$status>: $zkstatus\n");
	    }
	}
    }
}



sub UpdateBuildNumber {
    my ($localdir, $buildnum, $log) = @_;
    my $cwd = &POSIX::getcwd();
    my $status = 0;

    if ( chdir($localdir)) {
	if ( &ZKUtils::GetOS() eq "Win32") {
	    my $rcfile = "include/version.rc";
	    my $fh = new FileHandle;
	    
	    # Make it writeable
	    chmod(0644, $rcfile);

	    # Modify it
	    if ( defined($fh) && $fh->open("+<$rcfile")) {
		my @lines = $fh->getlines();
		
		# Set the new build number
		foreach ( @lines) {
		    s/^(\s+VALUE\s+\"BuildNumber\"\s*,\s*)\"[^\"]*\"/$1\"$buildnum\\0\"/;
		}
		
		# Save the file back
		$fh->truncate(0);
		seek($fh, 0, 0);
		$fh->print(@lines);
		$fh->close();
	    } else {
		$log->print("Error: can't read '$rcfile'\n");
		$status = 1;
	    }
	} else {
	    my $buildnofile = "makefiles/.buildno";
	    my $fh = new FileHandle(">$buildnofile");
	    
	    # Update the $buildnofile file
	    if ( defined($fh)) {
		$fh->print("override ZK_BUILDNO := $buildnum\n");
		$fh->close();
	    } else {
		$log->print("Error: can't update '$buildnofile'\n");
		$status = 1;
	    }
	}
	
	# Change back to previous working dir
	if ( !chdir($cwd)) {
	    $log->print("Error: can't chdir back to '$cwd'\n");
	    $status = 1;
	}
    } else {
	$log->print("Error: can't chdir to '$localdir'\n");
	$status = 1;
    }

    return($status);
}

sub CountWarnings {
    my ($localdir, $project, $config, $log) = @_;
    my $status = 0;
    my $osname = &ZKUtils::GetOS();
    my $cmd = "";
    my $logdir = ($log->change(undef))[0];
    my $logPrefix = "$project-$config";
    my $tagopt = "";
    my $logonly = undef;
    my $updateWrn = 1;
    
    # Add a trace for start of warning count
    $log->timeTrace("Warning count of started at:");
    
    # Make sure we have write access to the warnings file before updating it
    if (!chmod(0644, "$localdir/scripts/WARNFILES/$osname.wrn")) {
	$log->print("Warning: couldn't make $osname.wrn writeable, warnings update skipped");
	$updateWrn = undef;
    }
    
    if (defined($updateWrn)) {
	# Build the command to update the warnings file
	$cmd = "$localdir/scripts/UpdateWarnings.pl $logdir/$logPrefix.log $localdir/scripts/WARNFILES/$osname.wrn";
	if ($osname eq "Win32") {
	    $cmd =~ s,/,\\,g;
	}
	
	# Update the warnings file
	if ($log->system($cmd)) {
	    $log->print("Error: warnings update exited with an error code\n");
	    $status = 1;
	}
    }
    
    # Count the warnings
    if (!$status) {
	$cmd = "$localdir/scripts/CountWarning.pl $localdir $localdir/scripts/WARNFILES/$osname.wrn $logdir/$logPrefix-wrn.log $logdir/$logPrefix-wrn.lst $logdir/$logPrefix.log";
	if ($osname eq "Win32") {
	    $cmd =~ s,/,\\,g;
	}
	if ($log->system($cmd)) {
	    $log->print("Error: warnings count exited with an error code\n");
	    $status = 1;
	}
    }
    
    # Add a trace to the $project-$config.log for end of warning count
    $log->timeTrace("Warning count ended at:");
    $log->print("Status = $status\n\n");
    
    return($status);
}

sub GetCurrentBuildNumber {
    my ($localdir, $newBuildNum, $os, $log) = @_;
    my $cwd = &POSIX::getcwd();
    my $buildnofile = undef;

    if (!defined($newBuildNum) || (ref($newBuildNum) ne "SCALAR")) {
	$log->print("Error: invalid newBuildNum arg to GetCurrentBuildNumber\n");
	return(1);
    }

    if ( chdir($localdir)) {
	if ( $os eq "Win32") {
	    $buildnofile = "include/version.rc";
	    my $fh = new FileHandle("<$buildnofile");
	    
	    # Read it
	    if (defined($fh)) {
		my @lines = $fh->getlines();
		$fh->close();

		# Set the new build number
		foreach ( @lines) {
		    if (m/^\s+VALUE\s+\"BuildNumber\"\s*,\s*\"([^\"]+)\"/) {
			my $buildnum = $1;
			
			# Remove trailing \0
			$buildnum =~ s,\\0$,,;
			next if ($buildnum eq "");
			$$newBuildNum = $buildnum;
			return(0);
		    }
		}
	    } else {
		$log->print("Error: can't read '$buildnofile'\n");
	    }
	} else {
	    $buildnofile = "makefiles/.buildno";
	    my $fh = new FileHandle("<$buildnofile");
	    
	    # Read the $buildnofile file
	    if ( defined($fh)) {
		my @lines = $fh->getlines();
		$fh->close();

		foreach (@lines) {
		    if (m,^\s*(override)?\s*ZK_BUILDNO\s*:=\s*(\d+)\s*,) {
			$$newBuildNum = $2;
			return(0);
		    }
		}
	    } else {
		$log->print("Error: can't read '$buildnofile'\n");
	    }
	}
	
	# Change back to previous working dir
	if ( !chdir($cwd)) {
	    $log->print("Error: can't chdir back to '$cwd'\n");
	}
    } else {
	$log->print("Error: can't chdir to '$localdir'\n");
    }

	    
    # Build number was not found
    $log->print("Error: could not find a build number in '$buildnofile'\n");
    return(1);
}

sub BuildEmailContent {
  my ($failedBuilds) = @_;

  # Build details for the build that failed
  my $buildDetails = "";

  # Specify the OS
  my $osname = &ZKUtils::GetOS();
  $buildDetails = "Operating System:\t$osname";

  # Specify the project-config pairs that failed
  $buildDetails = "$buildDetails\nProject-Config:\t$failedBuilds";

  # Specify the label/branch/date of the build
  if (defined($Gconfig->{label})) {
    $buildDetails = "$buildDetails\nBranch/Label:\t$Gconfig->{label}";
  }
  elsif (defined($Gconfig->{branch})) {
    $buildDetails = "$buildDetails\nBranch/Label:\t$Gconfig->{branch}";
  }
  else {
    $buildDetails = "$buildDetails\nBranch/Date:\t$Gconfig->{date}";
  }
  
  # Specify whether it was an RTC build or not
  if (defined($Gconfig->{rtc})) {
    $buildDetails = "$buildDetails\nRTC:\t\t\tYes";
  }
  else {
    $buildDetails = "$buildDetails\nRTC:\t\t\tNo";
  }
  
  # Specify the build number
  $buildDetails = "$buildDetails\nBuild number:\t$Gconfig->{buildnum}";
  
  # Specify the localdir
  $buildDetails = "$buildDetails\nLocaldir:\t\t$Gconfig->{localdir}";
  
  # Specify the start time of the build
  my $buildStartTime = &strftime("%Y%m%d-%Hh%M.%S", localtime($Gconfig->{starttime}));
  $buildDetails = "$buildDetails\nStart Time:\t\t$buildStartTime";
  
  # Specify the build machine name
  my $hostName = hostname;
  $buildDetails = "$buildDetails\nBuild machine:\t$hostName";

  # Specify message subject and other message info
  my $subject = "Build failed";
  my $messageHeader = "The following build is broken:\n\n";  
  my $messageFooter = "\n\nPlease proceed to fix the build as quickly as possible.  Thank you for your cooperation :-)";
  $buildDetails = "$messageHeader$buildDetails$messageFooter\n";

  # Return the message built
  return($buildDetails);
}


# Get the options
&Getopts($Gconfig, @ARGV);

# Make sure a localdir is set
if (!defined($Gconfig->{localdir})) {
    print("Error: must specify -localdir option\n");
    &Usage();
    exit(1);
}
# Make sure a project is selected
if (!defined($Gconfig->{projects})) {
    print("Error: must specify the -project option\n");
    &Usage();
    exit(1);
}

# Get a build number from the build number server
if ( defined($Gconfig->{sync})) {
    $Gconfig->{buildnum} = &ZKUtils::GetBuildNumber($Gbn->{host}, $Gbn->{port}, &ZKUtils::GetOS());
} else {
    # The build number will come from the current source tree
    $Gstatus |= &GetCurrentBuildNumber($Gconfig->{localdir}, \$Gconfig->{buildnum}, &ZKUtils::GetOS(), $Gconfig->{log});
}

# Setup directory and file
$Gconfig->{log}->change(&strftime("$Gconfig->{localdir}/Logs/%Y%m%d-%Hh%M", localtime(time)), "General.log");

# Set the tag for the zkboard
if (defined($Gconfig->{zkboard})) {

    #Keep the version temporarily
    my $boardversion = $Gconfig->{zkboard};

    # Create the zkboard instance
    $Gconfig->{zkboard} = new ZKBoard(defined($Gconfig->{rtc})?'rtc':'builds',
				      $Gconfig->{boardServer}, $Gconfig->{boardPort});

    #Set the version criteria
    $Gconfig->{zkboard}->setCriteria('version',&strftime($boardversion, localtime(time)));

    # Check for label and if not defined then check for branch
    if (defined($Gconfig->{label})) {
	$Gconfig->{zkboard}-> setCriteria ('tag', $Gconfig->{label});
    }
    elsif (defined($Gconfig->{branch})) {
	$Gconfig->{zkboard}-> setCriteria ('tag',$Gconfig->{branch});
    }
    else {
	# Use the main branch
	$Gconfig->{zkboard}-> setCriteria ('tag','main');
    }
}

# Cleaning up
if (defined($Gconfig->{clean})) {
    $Gconfig->{log}->timeTrace("Clean started at:");

    # Cleanup the compilation tree
    if (&ZKUtils::DeleteTree([$Gconfig->{localdir}], undef, "^\Q$Gconfig->{localdir}/Logs\E")) {
	$Gstatus = 1;
	$Gconfig->{log}->print("Error: cleaning the '$Gconfig->{localdir}' directory\n");
	&ZkboardUpdate ('Cleaning error', $Gconfig->{projects}, $Gconfig->{cfgs}, &ZKUtils::GetOS());
    }

    $Gconfig->{log}->timeTrace("Clean ended at:");
}

# Create localdir
&mkpath($Gconfig->{localdir}, 1, 0755) if ($Gstatus == 0);

# Syncing sources
if (($Gstatus == 0) && defined($Gconfig->{sync})) {
    
    $Gconfig->{log}->timeTrace("Sync started at:");
    
    # Set the log to the sync.log file
    $Gconfig->{log}->change(undef, "sync.log");
    
    #Perform the update for status <syncing>
    &ZkboardUpdate ('Syncing', $Gconfig->{projects}, $Gconfig->{cfgs}, &ZKUtils::GetOS());
    
    $Gstatus |= &SyncSources($Gconfig->{localdir}, $Gconfig->{branch}, 
			     $Gconfig->{label}, $Gconfig->{date}, $Gconfig->{sync},
			     $Gconfig->{log});
    
    # Update the build number files
    $Gstatus |= &UpdateBuildNumber($Gconfig->{localdir}, $Gconfig->{buildnum}, $Gconfig->{log});
    
    # Set the log to the General.log file for end of sync trace
    $Gconfig->{log}->change(undef, "General.log");
    $Gconfig->{log}->timeTrace("Sync ended at:");
    
    # Perform the update for status <Waiting to compile>
    &ZkboardUpdate ('Waiting to compile', $Gconfig->{projects}, $Gconfig->{cfgs}, &ZKUtils::GetOS());
}

# Building the projects for all requested configuations
if ($Gstatus == 0) {
    my $config = undef;
    my $status = 0;
    my $failedBuilds = "";
    
    foreach $config (@{$Gconfig->{cfgs}}) {
	my $project = undef;
	my %ext = ("Debug" => "D", Release => "R", "Release-dev" => "R");
	my %ext2 = ("Debug" => "D", Release => "", "Release-dev" => "");
	
	$ENV{ZK_CONFIG} = "$config";
	$ENV{ZK_CONFIG_EXTENSION} = "$ext{$config}";
	
	foreach $project (@{$Gconfig->{projects}}) {
	    
	    $project = ucfirst(lc($project));
	    $ENV{ZK_PROJECT} = "$project";
	    
	    
	    # Set the status of this config status to 0
	    $status = 0;
	    
	    if (defined($Gconfig->{build})) {
		# Add a trace to the General log for start of build
		$Gconfig->{log}->change(undef, "General.log");
		$Gconfig->{log}->timeTrace("Build of '$project' in '$config' started at:");
		
		# Set the log to the $config.log file
		$Gconfig->{log}->change(undef, "$project-$config.log");
		
		# Add trace for start of build
		$Gconfig->{log}->timeTrace("Build started at:");
		
		# Update zkboard
                &ZkboardUpdate ('Compiling', \$project, \$config, &ZKUtils::GetOS());
		
		# Build it
		if (&ZKUtils::GetOS() eq "Win32") {
		    if (&BuildWin32($Gconfig->{localdir}, $config, $Gconfig->{log})) {
			# There was an error
			$status = 1;
		    }
		} else {
		    if (&BuildUnix($Gconfig->{localdir}, $config, $Gconfig->{log})) {
			# There was an error
			$status = 1;
		    }
		}
	    }
	    
	    #Update zkboard
            &ZkboardUpdate ($status ? 'Broken':'Ready', \$project, \$config, &ZKUtils::GetOS());
	    
	    # Redirect the output to ITS4.log
	    $Gconfig->{log}->change(undef, "ITS4.log");
	    
	    # Generate ITS4 report "manually" for Win32
	    if (($status == 0) && (defined($Gconfig->{its4})) && (&ZKUtils::GetOS eq "Win32")) {
		my ($logdir, $logfile) = $Gconfig->{log}->change(undef, undef);
		if ($Gconfig->{log}->system("$Gconfig->{localdir}\\scripts\\RunWin32ITS4.pl $Gconfig->{localdir}\\freedom\\client\\WindowClient.dsw $logdir\\ITS4.log")) {
		    $Gconfig->{log}->print("Error: generating ITS4 report\n");
		}
	    }
	    
	    # Redirect the output back to $project-$config.log
	    $Gconfig->{log}->change(undef, "$project-$config.log");
	    
	    # Count warnings generated from build
	    if (($status == 0) && (defined($Gconfig->{build})) && (defined($Gconfig->{warnCount}))) {
		# Counting warnings should not influence build
		# result. Ignoring return code
		&CountWarnings($Gconfig->{localdir}, $project, $config, $Gconfig->{log});
	    }
	    
	    # Rebase and generate setups
	    if ((&ZKUtils::GetOS eq "Win32") && ($status == 0)) {
		my $cwd = &POSIX::getcwd();
		my $bindir = "$Gconfig->{localdir}\\arch\\Win32\\$config\\bin";
		my $status = 0;

		# Copy sxl dll
		system("copy /y \\lib3rd\\win32\\SXL\\bin\\sxlrt232.dll $bindir");

		# Rebase the dlls
		if (chdir($bindir)) {
		    my $cmd = "$Gconfig->{localdir}\\scripts\\freedom_rebase.bat";
		    my $result = "freedom_rebase_results.txt";

		    # rebase the dlls
		    if ($Gconfig->{log}->system($cmd) == 0) {
			my $fh = new FileHandle("<$result");

			if (defined($fh)) {
			    while($_ = $fh->getline()) {
				chomp;
				if (m,0xFFFFFFFF,) {
				    $Gconfig->{log}->print("Error: rebased failed\n");
				    $status = 1;
				    $fh->getlines();
				    last;
				}
			    }
			    $fh->close();
			} else {
			    $Gconfig->{log}->print("Error: can't open '$result'\n");
			    $status = 1;
			}
		    } else {
			$Gconfig->{log}->print("Error: running '$cmd'\n");
			$status = 1;
		    }
		    
		    # Go back to previous directory
		    if (!chdir($cwd)) {
			$Gconfig->{log}->print("Error: can't chdir into '$cwd'\n");
			$status = 1;
		    }
		    
		    $Gstatus |= $status;

		    # Build the Installers
		    if ($status == 0) {
			my $wrun = "wrun.exe";
			my $wscript = "C:\\Program Files\\Mercury Interactive\\WinRunner\\arch\\Mindvision_Driver";
			my $prefix = "$Gconfig->{clientPrefix}-$Gconfig->{buildnum}-";
			
			$prefix =~ s,\\,/,g;
			&mkpath(&dirname($prefix), 1, 0755);
			$ENV{ZK_SETUP_PREFIX}= "$Gconfig->{clientPrefix}-$Gconfig->{buildnum}-";
			$ENV{ZK_VCT_FILE} = "$Gconfig->{localdir}\\freedom\\client\\zkinstall\\win32\\Vise\\FreedomClientWin32.VCT";
			if ($Gconfig->{log}->system("wrun.exe -t \"$wscript\" -run")) {
			    $Gconfig->{log}->print("Error: generating setups\n");
			    $Gstatus = 1;
			}
		    }
		} else {
		    $Gconfig->{log}->print("Error: can't chdir into '$bindir'\n");
		    $Gstatus = 1;
		}
	    }
	    
	    # Post the results
	    if (defined($Gconfig->{postdir})) {
		my $os = &ZKUtils::GetOS();
		my $dstdir = &ZKUtils::FixOsPath("$Gconfig->{postdir}/$project/$os/$config");
		my $cwd = &POSIX::getcwd();
		my ($logdir, $logfile) = $Gconfig->{log}->change(undef, undef);
		my $promLogCmd = "$Gconfig->{localdir}/scripts/promoteFiles.pl -force $logdir $dstdir";
		
		if (&ZKUtils::GetOS() ne "Win32") {
		    my $archdir = `$Gconfig->{localdir}/bin/zk_arch`;
		    chomp($archdir);
		    my %instdir = (Linux => "RPMS-$project/RPMS/i386", Solaris => "solpkg",);
		    my %instext = (Linux => ".rpm", Solaris => ".pkg",);
		    my $srcdir = "$Gconfig->{localdir}/arch/$archdir/$config/$instdir{$os}";
		    
		    if (chdir($srcdir)) {
			my @instfiles = glob("*$instext{$os}");
			
			if (@instfiles) {
			    # Check for ftp post
			    if ($Gconfig->{postdir} =~ m,^ftp://([^/]+)(/.+)$,) {
				# $1 stores the hostname and $2 the dstdir
				$status |= &PostFtp($1, $2, [$project, $os, $config], @instfiles);
			    } elsif ($Gconfig->{postdir} =~ m,^rcp://([^/]+)(/.+)$,) {
				$status |= &PostRcp($Gconfig->{log}, $1, "$2/$project/$os/$config", @instfiles);
			    } else {
				# Regular copy post
				my $instfiles = join(" ", @instfiles);
				
				$status = 1 if ($Gconfig->{log}->system("$Gconfig->{localdir}/scripts/promoteFiles.pl -force $srcdir $dstdir $instfiles"));
			    }
			    
			    # Go back to previous working directory
			    if (!chdir($cwd)) {
				$Gconfig->{log}->print("Error: can't chdir back to '$cwd'\n");
				$status = 1;
			    }
			} else {
			    $Gconfig->{log}->print("Warning: no files to post from '$srcdir'\n");
			}
		    } else {
			$Gconfig->{log}->print("Error: can't chdir to '$srcdir'\n");
			$status = 1;
		    }
		} else {
		    # Replace the '/' by '\\' in promote logfile command
		    $promLogCmd =~ s,/,\\,g;
		    
		    # On Win32
		    if (defined($Gconfig->{clientPrefix})) {
			my $prefix = "$Gconfig->{clientPrefix}-$Gconfig->{buildnum}-";
			$prefix =~ s,\\,/,g;
			my $dirname = &dirname($prefix);
			my $basename = &basename($prefix);
			
			if (chdir($dirname)) {
			    my $files = join(" ", glob("$basename*$ENV{ZK_CONFIG_EXTENSION}.exe"));
			    my $bindir = "$Gconfig->{localdir}\\arch\\Win32\\$config\\bin";
			    
			    if ($Gconfig->{log}->system("$Gconfig->{localdir}\\scripts\\promoteFiles.pl . $dstdir $files")) {
				$Gconfig->{log}->print("Error: promoting setups to '$dstdir'\n");
				$Gstatus = 1;
			    }

			    # Promote the freedomcfg executable
			    if ($Gconfig->{log}->system("$Gconfig->{localdir}\\scripts\\promoteFiles.pl $bindir $dstdir freedomcfg$ext2{$config}.exe")) {
				$Gconfig->{log}->print("Error: promoting freedomcfg$ext2{$config} to '$dstdir'\n");
				$Gstatus = 1;
			    }

			    # Promote the NukeNym executable
			    if ($Gconfig->{log}->system("$Gconfig->{localdir}\\scripts\\promoteFiles.pl $bindir $dstdir NukeNym$ext2{$config}.exe")) {
				$Gconfig->{log}->print("Error: promoting NukeNym$ext2{$config} to '$dstdir'\n");
				$Gstatus = 1;
			    }

			    if (!chdir($cwd)) {
				$Gconfig->{log}->print("Error: can't chdir back to '$cwd'\n");
				$status = 1;
			    }
			} else {
			    $Gconfig->{log}->print("Error: can't chdir to '$dirname'\n");
			    $status = 1;
			}
		    } else {
			$Gconfig->{log}->print("Error: must use -clientPrefix with -postdir on Win32\n");
			$status = 1;
		    }
		}
		
				# Promote the logfile if build is defined
		if (defined($Gconfig->{build})) {
		    $status = 1 if ($Gconfig->{log}->system("$promLogCmd $logfile"));
		}
		
				# Promote the ITS4.log file if its4 is defined and the log file exists
		if ((defined($Gconfig->{its4})) && (-e "$logdir/ITS4.log")) {
		    $status = 1 if ($Gconfig->{log}->system("$promLogCmd ITS4.log"));
		}
		
				# Promote the warnings files if warnCount is defined
		if (defined($Gconfig->{warnCount})) {
		    if (-e "$logdir/$project-$config-wrn.log") {
			$status = 1 if ($Gconfig->{log}->system("$promLogCmd $project-$config-wrn.log"));
		    }
		    if (-e "$logdir/$project-$config-wrn.lst") {
			$status = 1 if ($Gconfig->{log}->system("$promLogCmd $project-$config-wrn.lst"));
		    }
		}
	    }
	} continue {
	    # Force logging to General logfile and STDOUT 
	    my $logonly = ($Gconfig->{log}->change(undef, "General.log", undef))[2];
	    
	    # Add status
	    $Gconfig->{log}->print("$project-$config Status = $status\n");
	    
	    # Change back previous logonly setting
	    $Gconfig->{log}->change(undef, undef, $logonly);
	    
	    $Gconfig->{log}->timeTrace("Build of '$project' in '$config' ended at:");
	    $Gconfig->{log}->print("Status = $status\n\n");
	    
	    # Set global status
	    $Gstatus |= $status;

	    # If the build failed, save the build project-config info
	    $failedBuilds = "$failedBuilds$project-$config " if ($status);
	}
    }

    # Send an e-mail to inform us of the failed build(s)
    if ($Gstatus) {
      my $messageContent = &BuildEmailContent($failedBuilds);
      &ZKUtils::SendMail("build\@zks.net", $Gconfig->{emailRecipients}, "Build Failed", $messageContent);
  }
}

# Force logging to STDOUT and General.log
$Gconfig->{log}->change(undef, "General.log", undef);
$Gconfig->{log}->print("\n");

# Log some usefull information
if (defined($Gconfig->{sync})) {
    $Gconfig->{log}->print("Sync=$Gconfig->{sync}\n");
    $Gconfig->{log}->print("Branch=", defined($Gconfig->{branch}) ? $Gconfig->{branch} : "undef", "\n");
    $Gconfig->{log}->print("Label=", defined($Gconfig->{label}) ? $Gconfig->{label} : "undef", "\n"); 
    $Gconfig->{log}->print("Date=", defined($Gconfig->{date}) ? $Gconfig->{date} : "undef", "\n");
}
$Gconfig->{log}->print("BuildNumber = $Gconfig->{buildnum}\n");
$Gconfig->{log}->print("Postdir = ", defined($Gconfig->{postdir}) ? $Gconfig->{postdir} : "undef", "\n");
$Gconfig->{log}->print("Localdir = ", defined($Gconfig->{localdir}) ? $Gconfig->{localdir} : "undef", "\n");
$Gconfig->{log}->print("Hostname = ", &Sys::Hostname::hostname(), " (", &ZKUtils::GetOS(), ")\n\n");

$Gconfig->{log}->print(&strftime("Started on %Y%m%d-%Hh%M.%S\n", localtime($Gconfig->{starttime})));
$Gconfig->{log}->print(&strftime("Ended on %Y%m%d-%Hh%M.%S\n", localtime(time)));
$Gconfig->{log}->print("Global status = $Gstatus\n");

exit($Gstatus ? 1 : 0);
