#!/usr/bin/perl -w
# 
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
# 
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
# 
# The Original Code is Freedom code.
# 
# The Initial Developer of the Original Code is Zero-Knowledge Systems,
# Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
# Zero-Knowledge Systems, Inc.  All Rights Reserved.
# 
# Contributor(s):

BEGIN {
    $GscriptDir = $0;
    $GscriptDir =~ s,\\,/,g;
    $GscriptDir =~ s,/?[^/]+$,,;
    $GscriptDir = "." if ($GscriptDir eq "");
    unshift(@INC, $GscriptDir);
}

use File::Path;
use ZKUtils;
use POSIX 'strftime';
use ZKLog;

$Gconfig = {srcdir => "/", 
	    clean => undef,
	    cfgs => => ["Release"],
	    branch => undef, 
	    label => undef, 
	    date => undef,
	    sync => "true",
	    log => new ZKLog(&strftime("/Logs/%Y%m%d-%Hh%M", localtime(time)), "General.log"),
	};
$Gstatus = 0;

#
# Display Usage of this script
#
sub Usage {
    print("Usage: $0 [-h]\n");
    print("   [-h]: print this message\n");
    print("   [-branch <branch>]: Branch to sync with\n");
    print("   [-cfg <cfg,...>]: Configs to build\n");
    print("   [-clean]: To clean all trees before compiling\n");
    print("   [-date <date>]: Date to sync with\n");
    print("   [-label <label>]: Label to sync with\n");
    print("   [-nosync]: do not snc the sources\n");
    exit(1);
}

#
# Parse the command line option
#
sub Getopts {
    my ($config, @args) = @_;

    while(@args) {
	$_ = shift(@args);

	# Print Usage
	if (m,^-h(elp)?$,) {
	    &Usage;
	    exit(1);
	}

	# Specify the branch
	if (m,^-b(ranch)?$,) {
	    if (@args) {
		$config->{branch} = shift(@args);
		next;
	    } else {
		print("Error: missing arg to option -branch <branch>\n");
		&Usage;
	    }
	}

	# Specify the label
	if (m,^-l(abel)?$,) {
	    if (@args) {
		$config->{label} = shift(@args);
		next;
	    } else {
		print("Error: missing arg to option -label <label>\n");
		&Usage;
	    }
	}
	
	# Specify the date
	if (m,^-d(ate)?$,) {
	    if (@args) {
		$config->{date} = shift(@args);
		next;
	    } else {
		print("Error: missing arg to option -date <cvs date | now>\n");
		&Usage;
	    }
	}

	# Specify the configs
	if (m,^-cfg?$,) {
	    if (@args) {
		$config->{cfgs} = [split(m/,/, shift(@args))];
		next;
	    } else {
		print("Error: missing arg to option -cfg <cfgs,...>\n");
		&Usage;
	    }
	}
	
	# If we need to clean
	if (m,^-clean$,) {
	    $config->{clean} = "true";
	    next;
	}

	# If we need to sync
	if (m,^-nosync$,) {
	    $config->{sync} = undef;
	    next;
	}

	print("Ignoring option '$_'\n");
    }
}

sub SyncSources {
    my ($srcdir, $branch, $label, $date, $log, @modules) = @_;
    my $status = 0;
    my $dateopt = "";
    my $tagopt = "";
    my $cwd = &POSIX::getcwd();

    # Add trace for start of sync
    $log->timeTrace("Sync started at:");

    # Check for label and if not defined then check for branch
    if (defined($label)) {
	$tagopt = "-r $label";
    } elsif (defined($branch)) {
	$tagopt = "-r $branch";
    }

    # Check for date specification only if these is no label
    if (defined($date) && !defined($label)) {
	# Special case for date
	if (lc($date) eq "now") {
	    $dateopt = &strftime("-D \"%Y-%m-%d %H:%M\"", localtime(time));
	} else {
	    $dateopt = "-D \"$date\"";
	}
    }

    # Chdir to the source directory
    if (chdir($srcdir)) {
	# Sync the sources based on options
	if ($log->system("cvs -r co $tagopt $dateopt ", join(" ", @modules))) {
	    $log->print("Error: sync exit with error code\n");
	    $status = 1;
	}
	if (!chdir($cwd)) {
	    $log->print("Error: can't chdir back to '$cwd'\n");
	    $status = 1;
	}
    } else {
	$log->print("Error: can't chdir to '$srcdir'\n");
	$status = 1;
    }

    # Add trace for end of sync
    $log->timeTrace("Sync ended at:");

    $log->print("Sync status=$status\n");
    return($status);
}

sub BuildClient {
    my ($srcdir, $cfg, $log) = @_;
    my $clientdir = "${srcdir}freedom/client";
    my $vxddir = "$clientdir/zkshim/win32/zksHook";
    my $ntdrvdir = "$clientdir/zkshim/win32/NT2000";
    my $cwd = &POSIX::getcwd();
    my $status1 = 0;
    my $status2 = 0;
    my $status3 = 0;
    
    # Build the dll and client executable
    if (chdir($clientdir)) {
	if ($log->system("msdev WindowClient.dsw /MAKE \"FreedomClient - Win32 $cfg\"")) {
	    $log->print("Error: build exited with an error code\n");
	    $status1 = 1;
	}
	
	if (!chdir($cwd)) {
	    $log->print("Error: can't chdir back to '$cwd'\n");
	    $status1 = 1;
	}
    } else {
	$log->print("Error: can't chdir to '$clientdir'\n");
	$status1 = 1;
    }

    # Build VXD
    if (chdir($vxddir)) {
	if ($log->system("msdev zksHook_9x.dsp /MAKE \"zksHook_9x - Win32 $cfg\"")) {
	    $log->print("Error: build exited with an error code\n");
	    $status2 = 1;
	}
	
	if ($log->system("msdev zksHook_me.dsp /MAKE \"zksHook_Me - Win32 $cfg\"")) {
	    $log->print("Error: build exited with an error code\n");
	    $status2 = 1;
	}

	if (!chdir($cwd)) {
	    $log->print("Error: can't chdir back to '$cwd'\n");
	    $status2 = 1;
	}
    } else {
	$log->print("Error: can't chdir to '$vxddir'\n");
	$status2 = 1;
    }

    # NT2000 driver
    if (chdir($ntdrvdir)) {
	if ($log->system(($cfg eq "Debug") ? "DebugBuild.bat" : "RetailBuild.bat")) {
	    $log->print("Error: building nt2000 driver\n");
	    $status3 = 1;
	}
	
	if (!chdir($cwd)) {
	    $log->print("Error: can't chdir back to '$cwd'\n");
	    $status3 = 1;
	}
    } else {
	$log->print("Error: can't chdir to '$ntdrvdir'\n");
	$status3 = 1;
    }

    $log->print("Status of WindowClient=$status1\n");
    $log->print("Status of Vxd=$status2\n");
    $log->print("Status of NT2000 Driver=$status3\n");
    return($status1 | $status2 | $status3 ? 1 : 0);
}

sub Kill {
    my ($log, @processes) = @_;
    my $proc = undef;

    foreach $proc (@processes) {
	$log->system("kill -f $proc");
    }
}


# Get the options
&Getopts($Gconfig, @ARGV);

# Cleaning up
if (defined($Gconfig->{clean})) {
    # First we need to kill some application in order to cleanup properly
    &Kill($Gconfig->{log}, "msdev", "wincvs", "notepad");

    if (&ZKUtils::DeleteTree($Gconfig->{srcdir})) {
	$Gstatus = 1;
	$Gconfig->{log}->print("Error: cleaning the '$Gconfig->{srcdir}' directory\n");
    }
}

# Create srcdir
&mkpath($Gconfig->{srcdir}, 1, 0755) if ($Gstatus == 0);

# Syncing sources
if (defined($Gconfig->{sync}) && ($Gstatus == 0)) {
    # Set the log to the sync.log file
    $Gconfig->{log}->change(undef, "sync.log");

    $Gstatus |= &SyncSources($Gconfig->{srcdir}, $Gconfig->{branch}, 
			     $Gconfig->{label}, $Gconfig->{date}, $Gconfig->{log},
			     "freedom-client-win32");
}

# Building the client
if ($Gstatus == 0) {
    my $status = 0;

    foreach $config (@{$Gconfig->{cfgs}}) {
	# First set the status of this config to 0
	$status = 0;

	# Set the log to the $config.log file
	$Gconfig->{log}->change(undef, "$config.log");

	# Add trace for start of build
	$Gconfig->{log}->timeTrace("Build started at:");

	# Build the client
	if (&BuildClient($Gconfig->{srcdir}, $config, $Gconfig->{log})) {
	    # There was an error
	    $status = 1;
	}

    } continue {
	$Gstatus |= $status;
	
	# Add trace for start of build
	$Gconfig->{log}->timeTrace("Build ended at:");
	$Gconfig->{log}->print("Status = $status\n");
    }
}

$Gconfig->{log}->change(undef, "General.log");
$Gconfig->{log}->print("Global status = $Gstatus\n");
exit($Gstatus ? 1 : 0);
