use FileHandle;
use File::Basename;
use strict 'refs';

autoflush STDOUT 1;

package PRFile;

# GetFileList
#	Parameters:
#	 @_: messages to print
#	Return Value: TRUE on success
sub Trace {
    print(@_);
}

# GetFileList
#	Parameters:
#	 $in_item: directory to search or file
#	 $out_files: ref ARRAY holding matching files
#	 $in_match: if defined, files need to match this regular expression
#	 $in_noMatch: if defined, files need to not match this regular expression.
#	              This regular expression is evaluated after the match one.
#	Return Value: 0 on success, !0 on error
sub GetFileList {
    my ($in_item, $out_files, $in_match, $in_noMatch) = @_;
    my $status = 0;
    
    # Make sure $out_files is a list ref
    if (!defined($in_item) || (ref($out_files) ne "ARRAY")) {
	print(STDERR "Error: invalid args to GetFileList\n");
	return(1);
    }
    
    # Item does not exist
    return($status) if (! -e $in_item);
    
    # Item is not a directory
    if (! -d $in_item ) {
	my $filename = &File::Basename::basename($in_item);
	
	# Test for the paterns
	if (!((defined($in_match) && ($filename !~ m/$in_match/)) ||
	      (defined($in_noMatch) && ($filename =~ m/$in_noMatch/)))) {
	    push(@{$out_files}, $in_item);
	}
	return($status);
    }
    
    # Item is a directory
    if (opendir(DIR, "$in_item")) {
	my @subitems = ();
	my $subitem = undef;
	my @dirs = ();
	my $dir = undef;
	
	# Read content of this directory
	@subitems = readdir(DIR);
	closedir(DIR);
	
	foreach $subitem (@subitems) {
	    # Skip . and ..
	    next if ($subitem =~ m/^\.\.?$/);
	    
	    my $newitem = "$in_item/$subitem";
	    
	    # This is a directory
	    if (-d $newitem) {
		push(@dirs, $newitem);
		next;
	    }
	    
	    # This is not a directory, test the paterns
	    next if ((defined($in_match) && ($subitem !~ m/$in_match/)) ||
		     (defined($in_noMatch) && ($subitem =~ m/$in_noMatch/)));
	    
	    # Add the item to the list
	    push(@{$out_files}, $newitem);
	}
	
	foreach $dir (@dirs) {
	    $status |= &GetFileList($dir, $out_files, $in_match, $in_noMatch);
	}
	
    } else {
	print(STDERR "Error: can't get list '$in_item'\n");
	$status = 1;
    }
    
    return($status);
}

# GetFileList
#	Parameters:
#	 $in_filename: filename (relative or absolute) to process
#	 $in_oper: ref to subroutine to call
#	Return Value: 0 on success, !0 on error
sub Process {
    my ($in_filename, $in_oper) = @_;
    my @lines = ();
    my $status = 0;
    my $modified = 0;
    my ($mode, $mtime);
    
    Trace("$in_filename: ");
    # Read the file
    if (open(PRFILE, $in_filename)) {
	Trace("r");
	binmode(PRFILE);
	@lines = <PRFILE>;
	close(PRFILE);
    } else {
	print(STDERR "\nError: can't read file '$in_filename'\n");
	return(1);
    }
    
    ($mode, $mtime) = (stat($in_filename))[2,9];
    if (!defined($mode) || !defined($mtime)) {
	print(STDERR "\nError: can't stat '$in_filename'\n");
	return(1);
    }
    
    $status = &{$in_oper}($in_filename, \@lines, \$modified);

    # Was there any modification
    if ($modified) {
	if ($status) {
	    print(STDERR "\nError: skipping writing '$in_filename'\n");
	    return(1);
	} else {
	    my $bkfile = "$in_filename.prfile$$";
	    
	    # Backup current file
	    if (!rename($in_filename, $bkfile)) {
		print(STDERR "\nError: can't rename '$in_filename' to '$bkfile'\n");
		return(1);
	    }
	    
	    # Trace backup
	    Trace("b");
	    
	    # Save the modified file
	    if (open(PRFILE, ">$in_filename")) {
		binmode(PRFILE);
		$status = 1 if (!print(PRFILE @lines));
		$status = 1 if (!close(PRFILE));
	    } else {
		print(STDERR "\nError: can't write file '$in_filename'\n");
		$status = 1;
	    }
	    
	    # There was an error writing the file
	    if ($status) {
		unlink($in_filename);
		if (!rename($bkfile, $in_filename)) {
		    print(STDERR "\nError: restoring backup file '$bkfile'\n");
		    # We want to return here and leave the backup file
		    return(1);
		}
	    }
	    
	    # Trace the write operation
	    Trace("w");
	    
	    # Removed the backup file
	    if (!unlink($bkfile)) {
		print(STDERR "\nError: can't delete backup file\n");
		$status = 1;
	    } else {
		Trace("c");
	    }
	    
	    # Restore permissions of the file
	    if (!chmod($mode, $in_filename)) {
		print(STDERR "\nError: can't restore permission on '$in_filename'\n");
		$status = 1;
	    } else {
		Trace("p");
	    }
	}
    } else {
	# File not modified
	Trace("s");
    }
    
    # Add a newline in the trace
    Trace("\n");
    return($status);
}

sub ProcessMatchFiles {
    my ($in_dir, $in_oper, $in_match, $in_noMatch) = @_;
    my @files = ();
    my $status = 0;
    
    if (!defined($in_dir) || (ref($in_oper) ne "CODE")) {
	print(STDERR "Error: invalid args to ProcessMatchFiles\n");
	return(1);
    }
    
    $status = &GetFileList($in_dir, \@files, $in_match, $in_noMatch);
    
    # Process all files
    foreach  (@files) {
	$status |= &Process($_, $in_oper);
    }
    
    return($status);
}

1;
