#!/usr/bin/perl -w
# 
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
# 
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
# 
# The Original Code is Freedom code.
# 
# The Initial Developer of the Original Code is Zero-Knowledge Systems,
# Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
# Zero-Knowledge Systems, Inc.  All Rights Reserved.
# 
# Contributor(s):


#
# Purpose: Updates FILEVERSION information in resource files
#

use FileHandle;
use File::Path;
use File::Basename;
use strict 'refs';
use POSIX qw(strftime);

autoflush STDOUT 1;

$Gconfig = {branchname => undef, cvsroot => undef, files => [], message => "Updating version"};
$GtmpDir = "/tmp/UpdateVersion.$$";
$Gstatus = 0;

#
# Display Usage of this script
#
sub Usage {
    print("Usage: $0 [-help] [-cvsroot cvsroot] <files>\n");
    print("   [-branch branchtag]: Spefiy the branch to use\n");
    print("   [-cvsroot cvsroot]: where to set CVSROOT to\n");
    print("   [-help]: Display this message\n");
    print("   [-message message]: Use message as commit log for the new version\n");
    print("   files: List of files separated by spaces to update\n\n");
    exit(1);
}

#
# Parse the command line option
#
sub Getopts {
    my ($config, @args) = @_;

    while(@args) {
	$_ = shift(@args);

	if (m,^-h(elp)?$,) {
	    &Usage();
	    exit(1);
	}

	if (m,^-cvsroot$,i) {
	    if (@args) {
		$config->{cvsroot} = shift(@args);
		$ENV{CVSROOT} = $config->{cvsroot};

		# Modify path in order to gain access to cvs
		$ENV{PATH} = "/usr/local/bin/gnu:$ENV{PATH}";
	    } else {
		print(STDERR "Error: missing argument to -cvsroot option\n");
		exit(1);
	    }
	    next;
	}

	if (m,^-branch$,i) {
	    if (@args) {
		$config->{branchname} = shift(@args);
	    } else {
		print(STDERR "Error: missing argument to -branch option\n");
		exit(1);
	    }
	    next;
	}

	if (m,^-m(essage)?$,i) {
	    if (@args) {
		$config->{message} = shift(@args);
	    } else {
		print(STDERR "Error: missing argument to -message option\n");
		exit(1);
	    }
	    next;
	}

	# Add it to the list of files to process
	push(@{$config->{files}}, $_);
    }
}

#
# Perform cleanup
#
sub Cleanup {
    print("Cleaning $GtmpDir\n");
    chdir(&dirname($GtmpDir));
    system("/bin/rm -rf $GtmpDir");
}

#
# Perform checkout on list of files for a specific branch. 
#
sub CvsCheckout {
    my ($branchname, @files) = @_;
    my $branchopt = defined($branchname) ? "-r $branchname" : "";
    my $file = undef;
    my $status = 0;
    my %dirs = ();

    # Make temporary directory
    &mkpath($GtmpDir, 1, 0755);

    # Change directory to the working directory
    if (!chdir($GtmpDir)) {
	print(STDERR "Error: can't chdir  into '$GtmpDir'\n");
	return(1);
    }

    foreach $file (@files) {
	my $dir = &dirname($file);
	
	# Only checkout the directory once
	if (!exists($dirs{$dir})) {
	    # Checkout the module for this file
	    if (system("cvs co $branchopt \"$dir\"")) {
		print(STDERR "Error: checking out '$dir'\n");
		$status = 1;
		unlink($file) if (-e $file);
		next;
	    }
	    $dirs{$dir} = 1;
	}
    }
    return($status);
}

#
# Commit list of files with specified message
#
sub CvsCommit {
    my ($message, @files) = @_;
    my $file = undef;
    my $status = 0;

    foreach $file (@files) {
	# Commit the file
	if (system("cvs ci -m '$message' \"$file\"")) {
	    print(STDERR "Error: committing '$file'\n");
	    $status = 1;
	    next;
	}
    }
    return($status);
}

#
# Update the version information for each files
#
sub UpdateVersion {
    my (@files) = @_;
    my $file = undef;
    my $status = 0;
    
    foreach $file (@files) {
	# Update the file
	if (open(FILE, "$file")) {
	    my @lines = <FILE>;
	    
	    close(FILE);

	    foreach (@lines) {
		if (m/^(\s*)(FILEVERSION|PRODUCTVERSION)(\s+\d+,\d+,)\d+,\d+/) {
		    $_ = "$1$2$3" . &strftime("%Y,%m%d\n", localtime(time()));
		    next;
		}
		
		if (m/^(\s*)(VALUE\s+\"(FileVersion|PrivateBuild|ProductVersion)\"\s*,\s*\")(\d+\.\d+\.)\d+\.\d+((\\0)?\")/) {
		    $_ = "$1$2$4" . &strftime("%Y.%m%d", localtime(time())) . "$5\n";
		    next;
		}
	    }
	    
	    # Make sure the file is writable
	    chmod(0644, $file);

	    # Save the modified lines
	    if (open(FILE, ">$file")) {
		print(FILE @lines);
		close(FILE);
	    } else {
		print(STDERR "Error: can't update '$file'\n");
		$status = 1;
		next;
	    }
	} else {
	    print(STDERR "Error: can't read file '$file'\n");
	    $status = 1;
	    next;
	}
    }
    return($status);
}


&Usage() if (!@ARGV);
&Getopts($Gconfig, @ARGV);

$Gstatus |= &CvsCheckout($Gconfig->{branchname}, @{$Gconfig->{files}}) if (defined($Gconfig->{cvsroot}));
$Gstatus |= &UpdateVersion(@{$Gconfig->{files}});
$Gstatus |= &CvsCommit($Gconfig->{message}, @{$Gconfig->{files}}) if (defined($Gconfig->{cvsroot}));
&Cleanup();
exit($Gstatus);

