#
# File: zkboard.pm
#
# Desc:   This is a package require to update the zkboards application
#         using perl.
#
# Author: Marc Desfosses
#
# Date:   october 2000
# 

#Modules to include
use strict 'ref';
use IO::Socket;
use IO::Select;

@ISA=(Exporter);
@EXPORT=qw();
@EXPORT_OK=qw();

package ZKBoard;

# Sub:  new
#
# Desc: Create a new ZKBoard instance
#
# Params:
#	(in) $boardID:   The name of the board to be created (no default value)
#       (in) %fields:    FieldID->Value list of fields to change (no default value)
#       (in) %criterias: CriteriaID->Value list of criterias specifying a unique instance
#	(in) $webserver: The webserver that contains the zkboard application (default: ciaweb)
#       (in) %listboard: Reference to the hash describing the index of the listboard
#
# Return: New blessed object
#
sub new {
    my ($class, $boardid,$updserver,$port,$fields, $criterias, $updkeys) = @_;
    my $self = {
		boardid => (defined($boardid) && !ref($boardid)) ? "$boardid" : undef ,		
		fields => (defined($fields) && ref($fields) eq "HASH") ? $fields : undef ,
		criterias => (defined($criterias) && ref($criterias) eq "HASH") ? $criterias : undef,
		updkeys => (defined($updkeys) && ref($updkeys) eq "HASH") ? $updkeys : undef,
		updserver => (defined($updserver) && !ref($updserver)) ? $updserver : undef,		
		port => (defined($port) && !ref($port)) ? $port : undef,		
	       };
    
    #return the blessed obect
    bless $self;
}


# 
# Sub:  setCriteria
#
# Desc: Will set the criteria/value in the criteria hash
#      
# Params:
#      (in) $critID :   CriteriaID to set
#      (in) $critValue: Criteria value that has criteriaID as key
#
sub setCriteria {
    
    my ($self, $critID, $critValue) = @_;
    
    #Set the value in the hash
    $self->{criterias}->{$critID} = $critValue;
}


# 
# Sub:  setField
#
# Desc: Will take the FieldIDs and Values and update the form for
#       the update.
# Params:
#      (in) $fieldID :   fieldID to set
#      (in) $fieldValue: Field value that has fieldID as key
#
sub setField {
    
    my ($self, $fieldID, $fieldValue) = @_;
    
    #Set the value in the hash
    $self->{fields}->{$fieldID} = $fieldValue;
    
}


# 
# Sub:  setUpdKeys
#
# Desc: Will take the FieldIDs and Values and update the form for
#       the update.
# Params:
#      (in) $key :   key to set
#      (in) $value:  value that has key
#
sub setUpdKeys {
    
    my ($self, $key, $value) = @_;
    
    #Set the value in the hash
    $self->{updkeys}->{$key} = $value;
}

# 
# Sub:  saveAsPerlData
#
# Desc: 
#       
# Params:
#      (in) 
#
sub saveAsPerlData {
    my (@args) = @_;
    my $out = "";

    while (@args) {
	my $arg = shift(@args);

	if ( ! defined($arg)) {
	    $out .= "undef,";
	    next;
	}
	if ( !ref($arg)) {
	    $arg =~ s,\',\\\',g;
	    $out .= "'$arg',";
	    next;
	}
	if ( ref($arg) eq "ARRAY") {
	    $out .= "[";
	    $out .= saveAsPerlData(@{$arg});
	    $out .= "]";
	    next;
	}
	if ( ref($arg) eq "HASH") {
	    my $key = undef;

	    $out .= "{";
	    foreach $key ( keys(%{$arg})) {
		my $value = $arg->{$key};
		my $newkey = $key;

		$newkey =~ s,\',\\\',g;
		$out .= "'$newkey' => ";
		if ( !defined($value) || !ref($value)) {
		    $value =~ s,\',\\\',g if (defined($value));
		    $out .= defined($value) ? "'$value'," : "undef,";
		} else {
		    $out .= saveAsPerlData($value);
		    $out .= ",";
		}
	    }
	    $out .= "}";
	    next;
	}

    }
    return($out);
}

# 
# Sub:  set
#
# Desc: Will take the FieldIDs and Values and update the form for
#       the update.
# Params:
#      (in) %field :    reference to the fields hash
#      (in) %criterias: reference to the criterias hash
#
sub set {
    
    local $self = shift;
    
    #Get the parameters and verify them
    my ($fieldref, $criteriasref) = @_;
    
    if ((defined ($fieldref)) && (ref($fieldref) eq 'HASH')) {
	$self->{fields} =  $fieldref;
    }
    elsif (! defined ($self->{fields})) {
	return "Invalid or absent reference to a fields hash\n";
    }
    
    if ((defined ($criteriasref)) && (ref($criteriasref) eq 'HASH')) {
	$self->{criterias} =  $criteriasref;
    }
    elsif (! defined ($self->{criterias})) {
	return "Invalid or absent reference to a criteria hash\n";
    }   
    
    my $sock = IO::Socket::INET->new(PeerAddr => $self->{updserver},
				     PeerPort => $self->{port},
				     Proto    => 'tcp');

    return "Error: Cannot create socket\n" if !defined ($sock);

    my $sel = new IO::Select( $sock );

    return "Error: Cannot create select\n" if !defined ($sel);

    my $string = saveAsPerlData ({
				  boardid => $self->{boardid}, 
				  fields => $self->{fields},
				  criterias => $self->{criterias},
				  updkeys => $self->{updkeys},
				 });

    my $length = length($string);
    
    $sock->send(pack ("N",$length));
    $sock->send(pack("a$length", $string));
    
    while(my @ready = IO::Select::select($sel, undef, $sel, 30)) {
	
	# Read response from update server
	foreach my $fh (@{$ready[0]}) {
	    my $ack = <$fh>;
	    $sel->remove($fh);
	    $fh->close;
	    return $ack;
	}
	
	# Error occured on socket
	foreach my $fh (@{$ready[2]}) {
	    $sel->remove($fh);
	    $fh->close;
	    return "Error: exception on socket\n";
	}
    }

    return "Error: Error during select call\n";
    
}


1;
