#
# Logging package
#
#
use FileHandle;
use File::Basename;
use File::Path;
use POSIX 'strftime';

@ISA=(Exporter);
@EXPORT=qw();
@EXPORT_OK=qw();

package ZKLog;

#
# Create a new ZKLog instance
# Params:
#	(in) $logdir: the directory where to store the logfile (default: .)
#	(in) $logfile: the log filename (default: default.log)
#	(in) $logonly: Only send info to logfile (default: undef)
# Return: New blessed object
#
sub new {
    my ($class, $logdir, $logfile, $logonly) = @_;
    my $self = {fh => undef,
		logdir => defined($logdir) ? "$logdir" : ".",
		logfile => defined($logfile) ? "$logfile" : "default.log",
		logonly => $logonly
		};
    
    bless $self;
}

#
# Return current logfile path the ZKLog instance is set to
# Params:
# Return: the completes logdir/logfile path
#
sub getlogfile {
    my ($self) = @_;

    # Return the c
    return "$self->{logdir}/$self->{logfile}";
}

#
# Toggle display of log in STDOUT for this instance. Set to true if you want 
# to log only in the logfile
# Params:
#	(in) $logonly: undef to send to logfile and STDOUT otherwise only send to logfile
# Return: none
sub logonly {
    my ($self, $logonly) = @_;
    $self->{logonly} = $logonly;
}

#
# Change this instance internal parameters
# Params:
#	(in) $logdir: the directory where to store the logfile (default: .)
#	(in) $logfile: the log filename (default: default.log)
#	(in) $logonly: Only send info to logfile (default: undef)
# Return: the old values for logdir, logfile and logonly
#
sub change {
    my ($self, $logdir, $logfile, $logonly) = @_;
    my ($ologdir, $ologfile, $ologonly) = ($self->{logdir}, $self->{logfile}, $self->{logonly});

    $self->close();
    $self->{logdir} = defined($logdir) ? $logdir : $self->{logdir};
    $self->{logfile} = defined($logfile) ? $logfile : $self->{logfile};
    $self->{logonly} = $logonly if (@_ > 3);
    return($ologdir, $ologfile, $ologonly);
}

#
# Delete the file this instance is related to
# Params:
# Return: status of the deletion
#
sub delete {
    my ($self) = @_;
    my $status = 0;

    # First close it since it will be destroyed anyway
    $self->close();

    # Destroy it
    if (-f "$self->{logdir}/$self->{logfile}") {
	$status = 1 if (!unlink("$self->{logdir}/$self->{logfile}"));
    }
    return($status);
}

#
# Print to the log and possibly also to SDTOUT
# Params:
#	(in) @msg: list of messages to print
# Return: none
#
sub print {
    my ($self, @msg) = @_;

    # Perform the open once
    if (! defined($self->{fh})) {
	my $logpath = "$self->{logdir}/$self->{logfile}";
	# Create the log directory
	&File::Path::mkpath(&File::Basename::dirname($logpath, 1, 0755));

	$self->{fh} = new FileHandle;
	if (!$self->{fh}->open(">>$logpath")) {
	    print("Error: can't open logfile '$logpath'\n");
	    return 1;
	}
	$self->{fh}->autoflush();
    }

    # To Log
    $self->{fh}->print(@msg);

    # To STDOUT
    print STDOUT @msg if (!defined($self->{logonly}));;
}

#
# Run a command and log the output
# Params: 
#	(in) $cmd: command to run
#	(in) @args: list of argument for $cmd
# Return: status of the open (if failed) or the command executed
#
sub system {
    my ($self, $cmd, @args) = @_;
    my $status = 0;
    my $fh = new FileHandle;

    $cmd .= " " . join(" ", @args);
    if ($fh->open("$cmd 2>&1 |")) {
	while(<$fh>) {
	    $self->print($_);
	}
	$fh->close();
	$status = $?;
    } else {
	print("Error: can't run '$cmd'\n");
	return(1<<8);
    }
    return($status);
}

#
# Adds a time trace to the log
# Params:
#	(in) $prefix: text to be display before the trace (default: \nDate:)
# Return: none
#
sub timeTrace {
    my ($self, $prefix) = @_;
    
    $prefix = "\nDate:" if (!defined($prefix));
    $self->print(&POSIX::strftime("$prefix %Y%m%d-%Hh%M.%S\n", localtime(time)));
}

#
# Close the current ZKLog instance. You can still log to it.
# It will reopen and append.
# Params: 
# Return: none
#
sub close {
    my ($self) = @_;

    # Close
    undef($self->{fh}) if (defined($self->{fh}));
}

1;
