use FileHandle;
use IO::Socket;
use File::Basename;
use strict 'refs';
use Config;
use Net::SMTP;

package ZKUtils;


# GetListIfFilesAndDirs: build a list of directories and files
# Parameters
#          $in (in): scalar or list ref of path to search in
#          $files (out): list ref containing files found
#          $dirs (out): list ref containing dirs found
# Return
#       (Scalar) 0 on success, !0 on error
#
sub GetListOfFilesAndDirs {
    my ($in, $files, $dirs) = @_;
    my $item = undef;
    my $status = 0;
    my $items = undef;
    
    # Check the input args
    if (ref($in) eq "ARRAY") {
	$items = $in;
    } elsif (!ref($in)) {
	$items = [$in];
    } else {
	print("Error: arg1 of GetListOfFilesAndDirs must be scalar or ARRAY\n");
	return(1);
    }
    
    foreach $item (@{$items}) {
	# If item is a file or symbolic link
	if (-f $item || -l $item || -p $item || -S $item || -b $item || -c $item) {
	    push(@{$files}, $item);
	    next;
	}
	
	# If item is a directory
	if (-d $item) {
	    if (opendir(ZKUTILS_DIR, $item)) {
		my @subitems = grep(! m/^\.\.?$/, readdir(ZKUTILS_DIR));
		
		map {s,^,$item/,} @subitems;
		closedir(ZKUTILS_DIR);
		$status |= &GetListOfFilesAndDirs(\@subitems, $files, $dirs);
		push(@{$dirs}, $item);
	    } else {
		print("Error: can't list directory '$item'\n");
		$status = 1;
	    }
	    next;
	}
	
	# Check for existence of the item
	next if (! -e $item);
	
	# Unsupported: not a directory 
	print("Error: unsupported file type '$item'\n");
	$status =  1;
    }
    return($status);
}

# DeleteTree: Delete of list of items maching or not regexp
# Parameters 
#     $items (in): reference of list of items to delete
#
#     $matchre (in): if defined, item must matched regexp in order to
#     be deleted
#
#     $nomatchre (in): if defined, items must not match the regexp in
#     order to be deleted 
#
# Return
#    (Scalar) 0 on success, !0 on error
#
sub DeleteTree {
    my ($items, $matchre, $nomatchre) = @_;
    my @files = ();
    my @dirs = ();
    my $status = 0;
    my @skipped = ();
    
    # Get the list of files and directories to delete
    $status = &GetListOfFilesAndDirs($items, \@files, \@dirs);
    
    
    # Delete the files
    @files = sort(@files);
    foreach (sort(@files)) {
	# Regexp matching
	if ((defined($matchre) && ! m,$matchre,) || (defined($nomatchre) && m,$nomatchre,)) {
	    push(@skipped, $_);
	    next;
	}
	if (unlink($_) != 1) {
	    print("Error: can't delete file '$_'\n");
	    $status = 1;
	    next;
	}
    }
    
    # Sort by reverse length order and delete the diectories
    @dirs = sort {length($b) <=> length($a)} @dirs;
  DIR: foreach (@dirs) {
	next if (m,^([a-z]:)?[/\\]$,i);
	
	# Do not erase the directory if it ends in /.
	next if (m,(/|\\)\.$,);
	
	# Regexp matching
	if ((defined($matchre) && ! m,$matchre,) || (defined($nomatchre) && m,$nomatchre,)) {
	    push(@skipped, $_);
	    next;
	}
	
	foreach $skipped (@skipped) {
	    if ($skipped =~ m,^$_,) {
		push(@skipped, $_);
		next DIR;
	    }
	}
	
	if (!rmdir($_)) {
	    print("Error: can't delete directory '$_'\n");
	    $status = 1;
	    next;
	}
    }
    return($status);
}

# GetOS: Return OS name currently running on
# Parameters
#
# Return
#     (Scalar) The osname
#
sub GetOS {
    my $os = $Config::Config{osname};

    if ($os eq "linux") {
	return("Linux");
    } elsif ($os eq "solaris") {
	return("Solaris");
    } elsif ($os eq "MSWin32") {
        return("Win32");
    } elsif ($os eq "MacOS") {
        return("Mac");
    }
    return("undef");
}

# FixOsPath: adjust the path in order to respect OS specifics
# Parameters
#     $path (in): scalar path to fix
#     $os   (in): optional, specify for which OS
# Return
#     (Scalar) The fixed path
#
sub FixOsPath {
    my ($path, $os) = @_;

    return($path) if (!defined($path));
    
    $os = $Config::Config{osname} if (!defined($os));

    if ($os eq "MSWin32") {
	# Change all / to \ 
	$path =~ s,/,\\,g;
	
	# If there is only \ in the path
	$path =~ s,^\\+$,\\,;
	$path =~ s,^\\\\+,\\\\,;
	
	# Check for UNC path
	if ($path =~ m,^\\\\(.+),) {
	    my $tmp = $1;
	    
	    $tmp =~ s,\\+,\\,g;
	    $path = "\\\\$tmp";
	} else {
	    $path =~ s,\\+,\\,g;
	}
    } else {
	# Assume Unix
	$path =~ s,/+,/,g;
    }
    
    return($path);
}

# GetBuildNumber: Get a new build number
# Parameters
#     $server (in): server to connect to
#     $port   (in): port to connect to
#     $os     (in): name of OS we're building on
#     $type   (in): type of build we're doing (RTC, Daily, Release)
#
# Return
#     (Scalar) build number
#
sub GetBuildNumber {
    my ($server, $port, $os, $type) = @_;
    my $buildnum = 0;	# Default build number    

    # If one of the args is undefined, just returf the default build
    # number
    if (defined($server) || defined($port) || defined($os)) {
	my $sock = IO::Socket::INET->new(PeerAddr => $server, PeerPort => $port, Proto => 'tcp');
	
	# Socket created 
	if ( defined($sock)) {
	    my $oldInputRecordSeparator = $/;

	    $sock->autoflush(1);
	    
            # Change the input record separator on Mac so the server reads the request properly
	    $/ = "\r\n" if ($os eq "Mac");

	    # Send $os and read back build number
	    $sock->print("$os", defined($type) ? "-$type" : "", $/);
	    $buildnum = $sock->getline();

            # Change the input record separator on Mac so chomps works properly
	    $/ = "\r" if ($os eq "Mac");

	    # Remove any trailing newline characters
	    chomp($buildnum);

	    # Reset the input record separator on Mac
	    $/ = $oldInputRecordSeparator;

	    # Close the socket
	    $sock->close();
	}
    }
    
    return($buildnum);
}


# SendMail: Send a build failure e-mail
# Parameters
#     $from    (in): Sender of the e-mail
#     $refto   (in): Recipient(s) of the e-mail
#     $subject (in): Subject of the e-mail to send
#     $message (in): Message to appear in the e-mail body
#
# Return
#     (Scalar) 0 on success, !0 on error
#
sub SendMail {
  my ($from, $refto, $subject, $message) = @_;
  my $status = 0;
  
  # Connect to the mail server
  $smtp = Net::SMTP->new('mailhost', Hello => 'smartypants.zks.net');

  if (defined($smtp)) {
    # Set sender and recipient info
    $smtp->mail($from);

    my $recipient = undef;
    my @recipients = undef;

    # Check if we have a list of recipients
    # change the ref to array
    if (ref($refto) eq 'ARRAY') {
      @recipients = @$refto
    }elsif (ref($refto) eq 'SCALAR') {
      $recipient[0] = $$refto;
    }

    # Send to all the recipients
    foreach $recipient (@recipients) {
      $smtp->to($recipient);
    }

    # Send the message
    $smtp->data();
    $smtp->datasend("To: ");

    # Add each recipient to the "To:" list
    foreach $recipient (@recipients) {
      $smtp->datasend("$recipient; ");
    }
    $smtp->datasend("\n");

    $smtp->datasend("Subject: $subject\n");
    $smtp->datasend("\n");
    $status |= $smtp->datasend("$message\n");
    $smtp->dataend();

    # Disconnect from the mail server
    $smtp->quit;
  } else {
    $status = 1;
  }

  return($status);
}

1;
