/* -*- Mode: C; c-file-style: "bsd" -*- */

#ifndef YARROW_H
#define YARROW_H

#if 1 /* zks build */
# include "zks.h"
# define YARROW_SAVE_STATE
# ifdef WIN32
#  define USE_LIBZKTHREAD
# endif
#endif

/* These error codes are returned by the functions below. */

#define    YARROW_OK                0  /* All is well */
#define    YARROW_NOT_INIT          1  /* YarrowInit hasn't been called */
#define    YARROW_ALREADY_INIT      2  /* YarrowInit has already been called */
#define    YARROW_NO_DRIVER         3  /* driver doesn't exist */
#define    YARROW_CANT_OPEN         4  /* can't open driver */
#define    YARROW_BAD_SOURCE        5  /* invalid source id */
#define    YARROW_TOO_MANY_SOURCES  6  /* can't create any more source ids */
#define    YARROW_BAD_ARG           7  /* invalid argument */
#define    YARROW_ACCESS            8  /* insufficient privileges */
#define    YARROW_NOMEM             9  /* out of memory */
#define    YARROW_NORSRC           10  /* a resource (apart from memory) is exhausted */
#define    YARROW_GENERIC_ERROR    11  /* catch-all */
#define    YARROW_NOT_SEEDED       12  /* not enough entropy to generate output */
#define    YARROW_LOCKING          13  /* locking error */
#define    YARROW_STATE_ERROR      14  /* error with state load or save */

/* pool identification */
#define YARROW_FAST_POOL 0
#define YARROW_SLOW_POOL 1


#ifdef __cplusplus
extern "C" {
#endif

/* portability stuff */

#if defined(macintosh) && YARROW_DRIVER && TARGET_CPU_PPC
# define mem_zero(p, n)       BlockZero((p), (n))
# define mem_copy(d, s, n)    BlockMoveData((s), (d), (n))
#else
# define mem_zero(p, n)       memset((p), 0, (n))
# define mem_copy(d, s, n)    memcpy((d), (s), (n))
#endif

#ifndef WIN32
# define min(x, y) ((x) < (y) ? (x) : (y))
# define max(x, y) ((x) > (y) ? (x) : (y))
#endif

/* end portability stuff */

#if !YARROW_USE_DRIVER

/* "Normal", all-in-user-space implementation of Yarrow */
#include "ytypes.h"
#include "yhash.h"
#include "ycipher.h"

#ifdef WIN32
# ifdef YARROW_IMPL
#  define YARROW_DLL __declspec(dllexport) 
# else
#  define YARROW_DLL __declspec(dllimport) 
# endif
#else
# define YARROW_DLL
#endif

#define YARROW_MAX_SOURCES 20
#define YARROW_ENTROPY_MULTIPLIER 0.5

typedef size_t estimator_fn(const void* sample, size_t size);

typedef struct
{
    int pool;
    size_t entropy[2];
    int slow_thresh;
    estimator_fn* estimator;
} Source;

typedef struct
{
    /* state */
    int seeded;
#ifdef YARROW_DETECT_FORK
    pid_t pid;
#endif
    Source source[YARROW_MAX_SOURCES];
    unsigned int num_sources;
    HASH_CTX pool[2];
    byte out[CIPHER_BLOCK_SIZE];
    unsigned int out_left;
    COUNTER out_count;
    COUNTER gate_count;
    COUNTER gates_limit;
    byte C[CIPHER_BLOCK_SIZE];
    CIPHER_CTX cipher;
    byte k[CIPHER_KEY_SIZE];

    char entropyfile[256];

    /* parameters */
    COUNTER Pt[2];
    COUNTER Pg;
    int slow_k_of_n;
    int slow_k_of_n_thresh;
} Yarrow_CTX;

YARROW_DLL
int Yarrow_Init( Yarrow_CTX* y, const char *filename );

YARROW_DLL
int Yarrow_Poll( Yarrow_CTX *y );

YARROW_DLL
int Yarrow_Input( Yarrow_CTX* y, unsigned source_id,
		  const void* sample, 
		  size_t size, size_t entropy_bits );

YARROW_DLL
int Yarrow_Status( Yarrow_CTX* y, int *num_sources, unsigned *source_id,
		   size_t *entropy_bits, size_t *entropy_max );

YARROW_DLL
int Yarrow_Output( Yarrow_CTX* y, void* out, size_t size );

YARROW_DLL
int Yarrow_New_Source( Yarrow_CTX* y, unsigned* source_id );

YARROW_DLL
int Yarrow_Register_Source_Estimator( Yarrow_CTX* y, unsigned source_id, 
				      estimator_fn* fptr );

YARROW_DLL
int Yarrow_Stretch( const byte* m, size_t size, byte* out, size_t out_size );

YARROW_DLL
int Yarrow_Reseed( Yarrow_CTX* y, int pool );
YARROW_DLL
int Yarrow_Gate( Yarrow_CTX* y );

YARROW_DLL
int Yarrow_Final( Yarrow_CTX* y );

#else    /* !YARROW_USE_DRIVER */

#include <stddef.h>

/* Driver implementation of Yarrow.  Obviously, the implementation of these 
   functions is completely platform-specific. */

int Yarrow_Open(void);
int Yarrow_Close(void);

/* Synchronous interfaces, suitable for use from applications. */

#if !defined(YARROW_ASYNC)

int Yarrow_Read(void* out, size_t size);
int Yarrow_Write(const void* sample, size_t size, size_t entropy_bits);
#if 0
int Yarrow_Stretch(const void* m, size_t size, void* out, size_t out_size);
#endif
int Yarrow_Reseed(int pool);
int Yarrow_Gate(void);
int Yarrow_Status(int* num_sources, unsigned* source_id,
		  size_t* entropy_bits, size_t* entropy_max);

#if defined(macintosh)

int	Yarrow_Driver_Stats(unsigned short* numBuiltInSources, unsigned short* userSource);

#if !defined(NDEBUG)

int    Yarrow_Source_Stats(unsigned short source, long* numDevices, 
			   long* packetsReceived, long* bytesReceived, long* packetsOverflowed, 
			   long* bytesOverflowed, long* packetsNoBufferse, long* bytesNoBuffers, 
			   long* bytesWritten, long* buffersWritten, double* avgPacketSize, 
			   double* totalWriteTime, double* minWriteTime, double* maxWriteTime, 
			   double* avgWriteTime, char* description);

#endif	/*!defined(NDEBUG) */

#endif	/* defined(macintosh) */

#endif    /* !defined(YARROW_ASYNC) */


/* Asynchronous interfaces, suitable for use from interrupt-time code, such as 
   an Open Transport STREAMS module. */

#if defined(macintosh)
# include <Devices.h>
#endif

typedef struct Yarrow_Async_Ctxt    Yarrow_Async_Ctxt;

typedef void    (*Yarrow_Async_Callback)(Yarrow_Async_Ctxt* ctxt);

struct Yarrow_Async_Ctxt {
    Yarrow_Async_Callback  cb;        /* in:    callback                   */
    void*                  arg;       /* in:    argument to callback       */
    int                    result;    /* out:   return value from call     */
    union {
	struct {
	    void*              out;       /* in:    buffer to fill             */
	    size_t             size;      /* in:    size of buffer             */
	} rd;
	struct {
	    const void*        sample;    /* in:    buffer to write out        */
	    size_t             size;      /* in:    size of buffer             */
	    size_t             entropy;   /* in:    bits of entropy in buffer  */
	} wr;
	struct {
	    int                pool;      /* in:   the pool to reseed          */
	} rs;
	struct {
	    int                nsources;  /* out:  # of non-seeded sources     */
	    unsigned           source_id; /* out:  source closes to threshold  */
	    size_t             entropy;   /* out:  entropy in source_id        */
	    size_t             entrmax;   /* out:  threshold for source_id     */
	} st;
    } u;
#if defined(macintosh)
    ParamBlockRec          pb;
#endif
};

int Yarrow_Read_Async(Yarrow_Async_Ctxt* ctxt);
int Yarrow_Write_Async(Yarrow_Async_Ctxt* ctxt);
int Yarrow_Reseed_Async(Yarrow_Async_Ctxt* ctxt);
int Yarrow_Gate_Async(Yarrow_Async_Ctxt* ctxt);
int Yarrow_Status_Async(Yarrow_Async_Ctxt* ctxt);

#endif /* !YARROW_USE_DRIVER */

#ifdef __cplusplus
}
#endif

#endif /* YARROW_H */
